variable _isisscripts_version_string = "e61ca32ec6f8db127f610c81700a6673d0cee730 (built on 2024-03-08 18:09)";

%
% Initialization routines for the isisscripts
% Executed BEFORE the functions are defined
%

% give the isisscripts a verbose option
#ifeval __get_reference("_Isisscripts_Verbose")==NULL
variable _Isisscripts_Verbose=Isis_Verbose;
#endif

if (_Isisscripts_Verbose>0) {
    vmessage("loading %s", __FILE__);
}

$1 = path_sans_extname(__FILE__)+".txt";
ifnot(path_is_absolute($1))  $1 = path_concat(getcwd(), $1);
if(stat_file($1)!=NULL)
  set_doc_files([$1, get_doc_files()]);
else if (_Isisscripts_Verbose>0)
  vmessage("*** warning: Unable to locate help file '%s'", $1);

if (_Isisscripts_Verbose>0) {
message(`
If you use results obtained with the ISISscripts in a paper please cite as:
"This research has made use of ISIS functions (ISISscripts) provided by
ECAP/Remeis observatory and MIT (http://www.sternwarte.uni-erlangen.de/isis/)."
`);
}

provide("isisscripts");

%
% define local paths
%   Let users define where the local models are found!

%   We default to nothing, let user either set environment
%   variables, the local_paths variable, or use the
%   set_local_paths function

#ifeval __get_reference("local_paths")==NULL
variable local_paths = struct { @NULL };
#endif

private define __report_local_paths()
{
  if (_Isisscripts_Verbose>1) {
    variable path_name, path;
    variable path_names = get_struct_field_names(local_paths);
    if (length(path_names)) {
      message("   Setting local paths:");
      foreach path_name (path_names) {
	path = get_struct_field(local_paths, path_name);
	if (Array_Type == typeof(path))
	  path = strjoin(path, char(path_get_delimiter()));
	vmessage("     %s = %s", path_name, path);
      }
    } else {
      message("   No local model paths set");
    }
    message("");
  }
}

define set_local_paths ()
% This function sets the local paths to find instrument related data and local models
% It can either be done via qualifiers or environment variables
% The following settings exit:
% Data                  Qualifier            Environment
% localmodels           localmodels          XSPEC_LOCAL_MODELS,XSPEC_TABLE_MODELS     % semi-colon seperated list of paths
% RXTE data             rxte_data            RXTE_DATA_PATH                            % path to RXTE/FD
% RXTE ASM data         rxte_asm_data        RXTE_ASM_PATH                             % path to RXTE/ASM
% RXTE obscat           rxte_obscat          RXTE_OBSCAT_PATH                          % path to RXTE obscat
% MAXI lightcurves      maxi_lightcurves     MAXI_LIGHTCURVES_PATH                     % path to MAXI lightcurves
% FERMI catalog         fermi_catalog        FERMI_CATALOG_PATH                        % path to FERMI catalog
% FERMI catalog 3fgl    fermi_catalog_3fgl   FERMI_CATALOG_3FGL_PATH                   % path to FERMI catalog 3fgl
{
  % we combine model libraries and table models here for now. But keep the
  % two environment variables in case we need to change it
  variable local_tab_models = getenv("XSPEC_TABLE_MODELS"); % ';' separated list of paths to table models
  variable local_xsp_models = getenv("XSPEC_LOCAL_MODELS"); % ';' separated list of paths to compiled models
  variable localmodels = NULL;
  
  if (local_tab_models != NULL)
    localmodels = local_tab_models;
  if (local_xsp_models != NULL)
    localmodels = (localmodels != NULL) ? sprintf("%s%c%s", localmodels, path_get_delimiter(), local_xsp_models)
      : local_xsp_models;

  % overwrite with qualifiers
           localmodels        = qualifier("localmodels",        localmodels);
  variable rxte_data          = qualifier("rxte_data",          getenv("RXTE_DATA_PATH"));
  variable rxte_asm_data      = qualifier("rxte_asm_data",      getenv("RXTE_ASM_PATH"));
  variable rxte_obscat        = qualifier("rxte_obscat",        getenv("RXTE_OBSCAT"));
  variable maxi_lightcurves   = qualifier("maxi_lightcurves",   getenv("MAXI_LIGHTCURVES_PATH"));
  variable fermi_catalog      = qualifier("fermi_catalog",      getenv("FERMI_CATALOG_PATH"));
  variable fermi_catalog_3fgl = qualifier("fermi_catalog_3fgl", getenv("FERMI_CATALOG_3FGL_PATH"));

  % use HOME for backwards comp.
  if (localmodels != NULL)
    localmodels += char(path_get_delimiter()) + path_concat(getenv("HOME"),"share/software/localmodels");
  else
    localmodels = path_concat(getenv("HOME"),"share/software/localmodels");

  % split localmodels path

  variable loc = struct { localmodels = strchop(localmodels, ':', 0) };
  if (rxte_data != NULL)
    loc = struct { @loc, RXTE_data = [rxte_data] };
  if (rxte_asm_data != NULL)
    loc = struct { @loc, RXTE_ASM_data = [rxte_asm_data] };
  if (rxte_obscat != NULL)
    loc = struct { @loc, RXTE_obscat = [rxte_obscat] };
  if (maxi_lightcurves != NULL)
    loc = struct { @loc, MAXI_lightcurves = [maxi_lightcurves] };
  if (fermi_catalog != NULL)
    loc = struct { @loc, fermi_catalog = [fermi_catalog] };
  if (fermi_catalog_3fgl != NULL)
    loc = struct { @loc, fermi_catalog_3fgl = [fermi_catalog_3fgl] };

  local_paths = struct { @loc, @local_paths };
  __report_local_paths(); % show settings
}

% execute once on startup
set_local_paths();

%
% preload XSPEC local models
%

try {require("xspec");} catch AnyError: {;};

% only activate if module xspec could be loaded
#ifeval __get_reference("load_xspec_local_models")!=NULL

private define __load_localmodel (model, verbose)
{
  variable base,path,searchpath;
  if (path_is_absolute(model))
    searchpath = [model];
  else
    searchpath = [local_paths.localmodels, model];

  foreach base (searchpath) {
    if (model == base) % first try specified path
      path = base;
    else if (any(strchop(base, '/', 0)==model)) % then try if path contains this exact name
      path = base;
    else % else search for subfolders
      path = path_concat(base, model);

    variable e;
    variable gLmod = glob(path_concat(path, "lmodel.dat"), path_concat(path, "model.dat"));
    variable f;
    if (verbose>3) vmessage("%% Searching in '%s'", path);

    foreach f (gLmod) { % compiled models
      if (verbose) vmessage("%% loading xspec model from '%s'", path);
      try (e) {
	load_xspec_local_models(path);
      } catch AnyError, sprintf("%s: an error was caught: %s\n%s", _function_name(), e.descr, e.message);
      return;
    }

    variable gTmod = glob(path_concat(path, "*.fits"),
			  path_concat(path, "*.FITS"),
			  path_concat(path, "*.Fits"),
			  path_concat(path, "*.mod")
			 );

    variable t, tname, ttype, isModel=0;
    foreach t (gTmod) { % table model
      % get model name and type from primary extension
      tname = fits_read_key(t+"[0]", "MODLNAME");
      ttype = fits_read_key(t+"[0]", "ADDMODEL");
      if (NULL != tname && NULL != ttype) {
	try (e) {
	  if (ttype) add_atable_model(t, tname);
	  else add_mtable_model(t, tname);
	} catch AnyError, sprintf("%s: an error was caught: %s\n%s", _function_name(), e.descr, e.message);
	isModel |= 1;
      }
    }

    if (isModel) {
      if (verbose) vmessage("%% loading table model from '%s'", path);
      return;
    }
  }

  if (verbose>-1)
    vmessage("*** Warning: Unable to find a candidate for '%s'", model);
}

%%%%%%%%%%%%%%%%%%%%%
define use_localmodel()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{use_localmodel}
%\synopsis{load (an) xspec local (table) model[s] from local paths}
%\usage{use_localmodel(String_Type model[, model2, ...]);}
%\qualifiers{
%\qualifier{verbose}{: show the isis command[s] to load the xspec local model}
%}
%\description
%  This functions searches the defined search paths for local models
%  and tries to load the best matching model from a folder named
%  'model'. It is assumed that all models are installed with a default
%  lmodel.dat description file (compiled models) or follow the OGIP
%  table model specifications (table models).
%
%  An attempt is made to load table models correctly as additive or
%  multiplicative. Exponential models are also loaded as multiplicative
%  models.
%!%-
{
  if(_NARGS==0) { help(_function_name()); return; }

  variable model, models = ();
  loop(_NARGS-1)
  { model = ();
    models = [models, model];
  }
  variable verbose = qualifier("verbose", _Isisscripts_Verbose);

  foreach model ([models]) {
    if(verbose>2) vmessage("%% searching for %s", model);
    __load_localmodel(model, verbose);
  }
}
 
%
% use local models
%   use ISISCRIPTS_USE_MODELS to load set of default models
$1 = getenv("ISISSCRIPTS_USE_MODELS");
if (NULL != $1)
  foreach $2 (strchop($1, path_get_delimiter(), 0)) use_localmodel($2);

#endif
%
% Including external modules
%
try {require("curl");} catch AnyError: {;};
try {require("expat");} catch AnyError: {;};
try {require("fitswcs");} catch AnyError: {;};
try {require("fork");} catch AnyError: {;};
try {require("gcontour");} catch AnyError: {;};
try {require("gsl","gsl");} catch AnyError: {;};
try {require("gslfft","gsl");} catch AnyError: {;};
try {require("gslinterp");} catch AnyError: {;};
try {require("pcre");} catch AnyError: {;};
try {require("pgplot");} catch AnyError: {;};
try {require("png");} catch AnyError: {;};
try {require("pvm_ms");} catch AnyError: {;};
try {require("rand");} catch AnyError: {;};
try {require("select");} catch AnyError: {;};
try {require("socket");} catch AnyError: {;};
try {require("stats");} catch AnyError: {;};
try {require("vector");} catch AnyError: {;};
try {require("wcsfuns");} catch AnyError: {;};
try {require("xfig");} catch AnyError: {;};
try {require("xspec");} catch AnyError: {;};

%
% Loading external libraries
%
try { import("slmpi"); } catch AnyError: {;};

#ifnexists beta_inc
public define beta_inc();
#endif
#ifnexists help
define help();
#endif
#ifnexists lngamma
public define lngamma();
#endif
#ifnexists mean
define mean();
#endif
#ifnexists sincos
public define sincos();
#endif
#ifnexists wherefirst
define wherefirst();
#endif
#ifnexists wherefirstmax
public define wherefirstmax();
#endif
#ifnexists wherefirstmin
public define wherefirstmin();
#endif
#ifnexists wherelast
define wherelast();
#endif
#ifnexists wherelast_le
public define wherelast_le();
#endif
%
% Functions defined in this module
% (a total of 1084 functions)
%
define __push_array();
define __xy();
define _fsf_add_string();
define _maximum();
define _minimum();
define _sso_client();
define _sso_server();
define _sso_split_input();
define abkn_plaw_defaults();
define abkn_plaw_fit();
define ad_colormap_gauss();
define ad_colormap_norm_col();
define ad_create_colormap();
define ad_init_raw();
define ad_init_raw_timing();
define ad_interpol_image();
define ad_make_image();
define ad_read_image();
define ad_write_image();
define adbkn_plaw_defaults();
define adbkn_plaw_fit();
define add_color_palette();
define add_hist();
define aglc();
define aglc_filter();
define aglc_get_ephem();
define aglc_get_regs();
define aglc_get_trange();
define aglc_jd_to_phase();
define aglc_jd_to_rotnum();
define aglc_phased();
define aglc_pr_ephem();
define aglc_pr_trange();
define aglc_read_events();
define aglc_reset_regs();
define aglc_set_regs();
define aglc_set_status_filter();
define aglc_set_trange();
define aglc_tobs_to_jd();
define aglc_tobs_to_phase();
define aglc_truncate_events();
define aglc_truncate_wave();
define aglc_write_curve();
define air_to_vacuum();
define Aitoff_projection();
define alpha_ff();
define angle2string();
define angle_to_rad();
define angular_separation();
define ansi_escape_code();
define aplaw_defaults();
define aplaw_fit();
define append_chain();
define append_struct_arrays();
define aproposer();
define aqpo_defaults();
define aqpo_fit();
define ARFconstruct();
define ARFfromMirrorArea();
define ARFreadMirrorArea();
define array2image();
define array2matrix();
define array_copy();
define array_density();
define array_fit_gauss();
define array_flatten();
define array_insert();
define array_map_index();
define array_map_qualifiers();
define array_permutation_matrix();
define array_permute();
define array_remove();
define array_unique();
define array_zip();
define arrtimes_default_hook();
define arrtimes_fit();
define AS();
define ascii_read_table();
define asinwave_defaults();
define asinwave_fit();
define asqrwave_defaults();
define asqrwave_fit();
define assoc_copy();
define assoc_struct_combine();
define associate_arrays();
define associate_coords();
define aStar();
define atime_calc_beta();
define atime_dataind();
define atime_det();
define atime_det_beta();
define atime_get_ephemeris();
define atime_get_t0();
define atime_load_ephemeris();
define atime_merge();
define atime_merge_beta();
define atime_metavalid();
define atime_save_ephemeris();
define atime_set_ephemeris();
define atime_shiftref();
define atime_sim();
define atime_sim_orbitimpact();
define atime_sim_residuals();
define atime_useref();
define atime_xinclude();
define atom_name();
define attitude_lissajous_pattern();
define AVLTree();
define AzEl_from_RAdec();
define azfc_defaults();
define azfc_fit();
define B_lambda();
define B_nu();
define backshift();
define barycen();
define bayes_nconf();
define bayes_sconf();
define bayesian_blocks();
define bayesian_blocks_cperror();
define bbootstrap();
define behr();
define bezier();
define bezNp();
define bgsubHEXTElc();
define bin_average();
define binary_search();
define BinaryCor();
define binarydigits();
define BinaryPos();
define BinarySearchTree();
define bkn_plaw_defaults();
define bkn_plaw_fit();
define bknpowerlaw_xyfit();
define bknpowerlaw_xyfit_default();
define blocks_between_gaps();
define BoundingBox();
define brew();
define cart2cyl();
define cart2sphere();
define CCF_1d();
define CCF_2d();
define ceil2();
define cel2gal();
define cerf();
define cerfc();
define Chandra_display_mask();
define change_plot_options();
define cheby_sum();
define chebyshev_lagrange_weights();
define chebyshev_nodes();
define chebyshev_poly();
define check_parameter_hit();
define check_pulseperiod_orbit_struct();
define checkrwlc();
define chip();
define cholesky_decomposition();
define circle_from_points();
define cl_save();
define clip_points_polygon();
define clip_points_rectangle();
define clip_polyline_polygon();
define clip_polyline_rectangle();
define close_aa1col_xlabel();
define close_aa2col_xlabel();
define close_match();
define close_plot_ps2eps();
define cohen_sutherland();
define colacal();
define color_blackbody();
define color_color_data();
define color_complex();
define color_desaturate();
define color_mix();
define color_wavelength();
define combine_chain();
define compare_modelfits();
define compare_par();
define compare_pars();
define complementary_array();
define compressedArrayStr();
define Compton_crosssection();
define continued_fraction();
define continued_fraction_expansion();
define contour_borders();
define contour_trq();
define convert_units();
define coords_in_box();
define coordX_in_box();
define coordY_in_box();
define COPY();
define correlation_coefficient();
define cosmo_param();
define cov_matrix();
define covariance_correlation_matrix();
define create_basic_simputfile();
define create_struct_field();
define crossing_number_polygon();
define cstat_goodness();
define cut_dataset_range();
define cutoffpl2_defaults();
define cutoffpl2_fit();
define cyl2cart();
define data_map_function();
define DateOfJD();
define DateOfMJD();
define dateOfMJD();
define dayOfWeek();
define daysPerMonth();
define dbkn_plaw_defaults();
define dbkn_plaw_fit();
define dcf();
define define_atime();
define define_counts_2d();
define define_xydata();
define deg2dms();
define detconst_init();
define diff();
define difmap_restore();
define disk_map();
define diskline2_fit();
define distribution_matrix();
define dms2deg();
define dof();
define dont_use_line();
define Doppler_velocity();
define douglas_peucker();
define draw_plot_commands();
define draw_progress_bar();
define dummy_default();
define dummy_fit();
define e_bv();
define EasterSunday();
define EchoMap();
define EchoMap_binary();
define EchoMap_makelc();
define ecliptical2equatorial();
define edit_var();
define element2Z();
define element_name();
define element_symbol();
define ellipse();
define ellipse_xy_xyfit();
define ellipse_xy_xyfit_default();
define emcee_hammer();
define empiric_cdf_inverse();
define empty_struct();
define enclosing_ellipse();
define energy_resolution();
define energyflux();
define energyfluxdensity();
define enlarge_image();
define epatplot();
define epferror();
define epfold();
define epfoldpdot();
define EpochofJD();
define equation_equinoxes();
define equatorial2ecliptical();
define equatorial2galactic();
define equatorial2horizon();
define eqwFit_fit();
define eqwFit_init();
define erfinv();
define erg2keV();
define err_map_gaussian();
define escapedParameterName();
define escapeStr();
define eval_array();
define eval_fun2_keV();
define eval_fun_keV();
define eval_simputfile();
define eval_xyfun();
define eval_xyfun2();
define exponential_xyfit();
define exponential_xyfit_default();
define ext_info_string();
define ext_info_string_PGPLOT();
define ext_line_info();
define ext_line_type();
define F_fit();
define factorial();
define factorized_arf_rmf();
define factorset();
define Faddeeva();
define Faddeeva_dz();
define fake_pulsar_lightcurve();
define fcorback_default();
define fcorfile_default();
define fermi2MJD();
define filter_gti();
define find_correlations();
define find_directories();
define find_files();
define find_function_maximum();
define find_function_value();
define find_multiargumentfunction_value();
define find_peak();
define first_valid_digit();
define fit_brute_force();
define fit_gauss_to_img_noise();
define fit_interactive();
define fit_line();
define fit_pars();
define fit_pars_read_stdoutlogfile();
define fit_steppar();
define fitfun_cache_enabled();
define fitfun_disable_cache();
define fitfun_enable_cache();
define fitfun_get_cache();
define fits_add_fit();
define fits_append_binary_table();
define fits_append_extension();
define fits_append_tmp_extension();
define fits_column_unit_struct();
define fits_conv_to_legal_char();
define fits_get_hdu_names();
define fits_lc_exposure();
define fits_list_fit_pars();
define fits_load_fit();
define fits_load_fit_struct();
define fits_modify_header();
define fits_nr_extensions();
define fits_num_hdus();
define fits_plot_rmf();
define fits_read_header();
define fits_read_key_int_frac();
define fits_read_lc();
define fits_read_rmf();
define fits_read_unsigned_img();
define fits_save_data_model_flux();
define fits_save_fit();
define fits_save_fit_struct();
define fits_save_fit_write();
define fits_wcs_struct();
define fits_write_arf();
define fits_write_arf_diff();
define fits_write_binary_table_extensions();
define fits_write_gti();
define fits_write_image();
define fits_write_pha_file();
define fits_write_rmf();
define fits_write_sixte_vignetting();
define fits_write_tex_table();
define floor2();
define flux2lum();
define flux_err();
define fm_unred();
define foucalc();
define foufreq();
define four_chip_defaults();
define four_chip_fit();
define fplot();
define fread_struct();
define freeParameters();
define frequency2velocity();
define FriendCastor1982_dMdot_dOmega_CygX1();
define ftest_xspec();
define g_earth();
define gal2cel();
define galactic2equatorial();
define galactic_rotation_velocity();
define galaxy_elliptical_fit();
define galLB_from_RAdec();
define gapsync();
define gapsync2();
define garf_gaps();
define gauss_2d_defaults();
define gauss_2d_fit();
define gauss_2d_integrated_defaults();
define gauss_2d_integrated_fit();
define Gauss_complex();
define gauss_xyfit();
define gauss_xyfit_default();
define gcrs2j2000_matrix();
define gehrels_confidence();
define gehrels_error();
define gehrels_lower_confidence();
define gehrels_upper_confidence();
define generate_iauname();
define geographic2vector();
define get1match();
define get2matches();
define get_2d_data_grid();
define get_2d_model();
define get_all_data();
define get_arg_struct();
define get_bin_corr_factor();
define get_blocks_data();
define get_color_map();
define get_color_palette();
define get_color_palette_names();
define get_column_density_from_line();
define get_combined_data_model_residuals();
define get_component();
define get_confmap();
define get_confmap_collect_results();
define get_contour_lines();
define get_coordinatearrays_of_image();
define get_count_rate();
define get_data_counts_with_tot_err();
define get_dataset_response();
define get_ephemeris();
define get_flux_corrected_convolved_model_flux();
define get_grid();
define get_instrument_resolution_from_data();
define get_intersection();
define get_lambda_parameters_of_all_lines();
define get_lambda_parameters_of_lines();
define get_lambda_parameters_of_lines_from_one_ion();
define get_line_ids();
define get_line_labels();
define get_line_lambdas();
define get_line_transition();
define get_line_velocity();
define get_map_value_at_position();
define get_modelcomponents();
define get_par_combinations();
define get_param_from_filename();
define get_params_from_file();
define get_params_table_from_files();
define get_rainbow_col();
define get_ratio();
define get_residuals();
define get_selected_data_flux_en();
define get_simple_gpile_info();
define get_simputfile_struct();
define get_sixte_eventfile_statistic();
define get_sixte_xml_data();
define get_source_counts();
define get_struct_fields();
define get_unpiled_data_flux();
define get_unpiled_fit_fun();
define get_variable_name();
define get_warmabs_model();
define get_xydata();
define get_xymodel();
define getCommonGTIs();
define getVarPeriod();
define GhoshLamb79();
define GMST();
define gravitational_radius();
define greatcircle_coordinates();
define greatcircle_distance();
define GreenwichSiderealTime_from_MJD();
define greiner_hormann();
define group_data_keV();
define group_noticed_data();
define group_pha();
define GTIoverlap();
define Hammer_projection();
define haPSFcos_fit();
define hardnessratio();
define hardnessratio_error_prop();
define hardnessratio_from_dataset();
define hardnessratio_from_spec();
define hardnessratio_simulate_grid();
define hardnessratio_simulate_grid_load();
define hardnessratio_simulate_grid_save();
define HeTriplet_defaults();
define HeTriplet_fit();
define hex2rgb();
define hist1d_confidence();
define hist2d_confidence();
define hist_fwhm_index();
define histogram2d_min_max();
define histogram3d();
define histogram_gaussian_probability();
define histogram_min_max();
define history();
define hms2deg();
define horizon2equatorial();
define hplot_filled();
define hplot_with_err();
define hpluv2rgb();
define hquantile();
define hsl2rgb();
define hsluv2rgb();
define hsv2rgb();
define ignore_large_bins();
define ignore_xy();
define image2rgb();
define ImageMagick_text();
define indices_of_2Darray();
define init_Apec();
define init_histo();
define init_jet_fit();
define init_plot();
define init_time_structure();
define inner_product();
define inner_products();
define integral_cat2ds9();
define integrate2d();
define integrate2d_test();
define integrate_trapez();
define integrateAB5();
define integrateGK();
define integrateRK4();
define integrateRK45_adaptive();
define integrateRKF45();
define interpol2D();
define interpol_2d_table();
define interpol_params();
define interpol_polynomial();
define inversemapping();
define ionabs_default();
define ionabs_fit();
define ionized_phabs_sigma();
define j_ff();
define jd2mjd();
define JD2MJD();
define jd2year();
define JDofDate();
define JDofEpoch();
define jpl_bineph2fits();
define jpl_eph();
define jpl_eph_vec();
define jpl_initeph();
define kaastra_cstat_goodness();
define kendall_tau_censored();
define KeplerEquation();
define kerr_lp_redshift();
define kerr_rms();
define kerr_rplus();
define keV2erg();
define keyinput();
define KS_test();
define lag_energy_spectrum();
define lagrange_poly();
define lagrange_poly_deriv();
define lagrange_weights();
define Lambert_Equal_Area_projection();
define lc_from_events();
define lcseg();
define leapyear();
define line_intersect();
define linear_fit_xerr_yerr_data();
define linear_regression();
define linear_xyfit();
define linear_xyfit_default();
define lines_defaults();
define lines_fit();
define lissajous_pattern();
define list_all_lines();
define list_copy();
define list_free_bounds();
define list_lines();
define List_par();
define list_Par();
define list_par_bounds();
define list_xypar();
define load_1storderHETGS_datasets();
define load_atime();
define load_data_combined();
define load_data_integral();
define load_fermi();
define load_fermi_catalog();
define load_fermi_spectrum();
define load_grouped_xmm_data();
define load_HETGS_datasets();
define load_par_from_FITS_header();
define load_par_tbnew();
define load_pha();
define load_piled1storderHETGS_datasets();
define load_piledHETGS_datasets();
define load_radio2();
define load_xypar();
define loadDataset();
define loadHETGlc_sum();
define loadHETGlcs();
define loadHETGspec();
define log_grid();
define log_par_defaults();
define log_par_en_defaults();
define log_par_en_fit();
define log_par_fit();
define logging();
define Lorentz_complex();
define lorentz_trafo();
define lorentz_xyfit();
define lorentz_xyfit_default();
define lorentzmb_default();
define lorentzmb_fit();
define lorentzmb_old_default();
define lorentzmb_old_fit();
define lumdist();
define luminosity();
define make_conlines();
define make_fine_rmf();
define make_movie_from_evtList();
define make_path();
define make_spix();
define makepsd();
define mass_function();
define matrix2array();
define matrix33_adjoint();
define matrix33_as_array();
define matrix33_cofactors();
define matrix33_determinant();
define matrix33_diag();
define matrix33_get_diag();
define matrix33_get_trace();
define matrix33_identity();
define matrix33_inverse();
define matrix33_new();
define matrix33_null();
define matrix33_reflect();
define matrix33_rot();
define matrix33_scalar();
define matrix33_transpose();
define matrixmul();
define max_struct_field();
define max_time();
define MAXI_lightcurve();
define mc_sig();
define median_2d();
define merge_struct_arrays();
define message_system();
define mgauss_defaults();
define mgauss_fit();
define midas_bary_helio_corr();
define min_max();
define min_struct_field();
define min_time();
define minus_defaults();
define missing_contour_trq();
define MJD2fermi();
define mjd2jd();
define MJD2JD();
define MJD2UNIXtime();
define MJDofDate();
define MJDofDateString();
define MJDref_satellite();
define MJDstrftime();
define mkdir_rec();
define MN_NFW();
define MN_TF();
define modelComponents();
define mpi_fit_pars();
define mpi_fit_pars_trqscript_gen();
define mpi_master_only();
define mpi_mc_sig();
define multibknpower_fit();
define multiple_string();
define multiplot_data_m_lam_counts();
define multiplot_flux_counts_res();
define N_body_simulation();
define N_body_simulation_MW_kernel();
define N_body_simulation_std_kernel();
define narrowline_default();
define narrowline_fit();
define new_printplot();
define new_table_model();
define normalized_in_0_1();
define normpsd();
define notice_human2isis();
define notice_isis2human();
define notice_listFromString();
define notice_listToString();
define notice_xy();
define ntree();
define ntree_type();
define nuFnu();
define num_bin();
define nutation_angles();
define nutation_matrix();
define obfuscate_string();
define object_visibility_satellite();
define obliquity();
define ohplot_filled();
define ohplot_with_err();
define opdf();
define open_plot_aa1col();
define open_plot_aa2col();
define open_plot_ps2eps();
define oplot_contour_lines();
define oplot_rect();
define oplot_struct_arrays();
define oplot_with_err();
define oplot_xline();
define oplot_yline();
define oplotGTIs();
define opmc();
define opmf();
define orbit_calculator();
define orbitalphase();
define outer_product();
define outer_products();
define overplot_clean_model();
define pack_obj();
define panstarrsImageDownload();
define par2tex();
define par_bounds();
define par_list();
define par_load();
define par_names_to_tex();
define par_save();
define parseDateString();
define partial_correlation();
define path_realpath();
define pdc();
define pdf();
define period_search_via_string_length();
define pfold();
define phasebin_GTIs();
define phaseOfTime();
define photon_flux();
define physical_quantity();
define physical_quantity_current_in_A();
define physical_quantity_length_in_m();
define physical_quantity_mass_in_kg();
define physical_quantity_temperature_in_K();
define physical_quantity_time_in_s();
define physical_quantity_unit();
define planetpos();
define plaw_defaults();
define plaw_fit();
define plot_atime();
define plot_combined_data_model_residuals();
define plot_component();
define plot_contour_trq();
define plot_jet_speed();
define plot_spix();
define plot_struct_arrays();
define plot_table_columns();
define plot_vlbi_map();
define plot_vlbi_pol_map();
define plot_with_err();
define plot_with_y2axis();
define plot_xyfit();
define plotGTI();
define plt_garf_gaps();
define plummer_interaction_kernel();
define plummer_MW();
define pmc();
define pmf();
define png_read_curve();
define point_distance2_from_line();
define point_in_polygon();
define point_is_left_of_line();
define polint();
define polytropic_standard_model();
define pos_modulo();
define position_angle();
define powerlaw_noise();
define powerlaw_xyfit();
define powerlaw_xyfit_default();
define pr_garf_gaps();
define precess();
define precession_matrix();
define precession_matrix2();
define primefactors();
define principal_components();
define print_array();
define print_statistics();
define print_struct();
define printhplot();
define printplot();
define printplot_out();
define probmap();
define properties_satellite_galaxies();
define psd_lc();
define psdcorr_zhang();
define pulsar_GTI();
define pulsarorbit_fit_constraint();
define pulsarorbit_fluxerror_mc();
define pulsartaylor_set_order();
define pulse2pulse_flux_lc();
define pulse_time();
define pulseperiod();
define pulseperiod2phase();
define pulseperiod_epfold();
define pulseperiod_magic();
define pulseperiod_search();
define pulseperiod_transform();
define pulseprofile_compose();
define pulseprofile_decompose();
define pulseprofile_energy_interpolate();
define pulseprofile_energy_lag();
define pulseprofile_energy_map();
define pulseprofile_energy_normalize();
define pulseprofile_phase_connect();
define pushKommaString();
define pushString();
define pushStructFieldArray();
define pvm_fit_pars();
define pvm_fit_pars_txt2fits();
define qpo_defaults();
define qpo_fit();
define qromb();
define qsimp();
define quadratic_xyfit();
define quadratic_xyfit_default();
define quantile();
define quasar_accr_fit();
define query_user();
define rad2RD();
define RAdec_from_AzEl();
define RAdec_from_galLB();
define radial_velocity_binary();
define radial_velocity_LSR();
define radio_mod2img();
define radius_to_unit();
define rate2mdot();
define ratio_error_prop();
define RD2rad();
define read_data_from_write_plot();
define read_difmap_fits();
define read_ep();
define read_histo();
define read_par();
define read_sixte_lc();
define read_spix();
define rebin_combined_optimal();
define rebin_dataset_optimal();
define rebin_error_hook();
define rebin_fouquan();
define rebin_human2isis();
define rebin_isis2human();
define rebin_lc();
define rebin_mean();
define rebin_to_energy_grid();
define rebin_to_instrument();
define rebinDensity();
define rebinGroup();
define rectangle_where();
define rectangles_overlap();
define redden_fit();
define reduce_image_resolution();
define reduce_struct();
define reflection_fraction_relxill();
define refraction();
define refractive_index_air();
define removeLeadingZeros();
define rename_struct_fields();
define renormalizeTo();
define replicate_table();
define require_atoms();
define rescale_range();
define residual_runs();
define resolution_rebin();
define rgb2hex();
define rgb2hpluv();
define rgb2hsl();
define rgb2hsluv();
define rgb2hsv();
define ridge_line();
define rms_gauss_defaults();
define rms_gauss_fit();
define rndbknpwrlc();
define rndpwrlc();
define Roche_critical();
define Roche_equipotential_volume();
define Roche_lobe_dimension();
define Roche_lobe_surface();
define Roche_potential();
define Roman();
define roman();
define roman2int();
define round2();
define round_conf();
define round_err();
define runs_test();
define RXTE_ASM_countrate();
define RXTE_ASM_lightcurve();
define RXTE_filter_file_info();
define RXTE_nr_PCUs_from_filename();
define RXTE_nr_PCUs_from_filterfile();
define RXTE_obscat_info();
define RXTE_PCA_info();
define RXTE_PCA_modes();
define RXTE_plot_PCA_info();
define SaturationVaporPressure();
define save_atime();
define save_atime_beta();
define save_par_to_FITS_header_struct();
define save_plot();
define save_slang_variable();
define save_statistics();
define save_xypar();
define savitzky_golay_coefficients();
define savitzky_golay_smoothing();
define scale_hist();
define scargle();
define segment_lc_for_psd();
define set_2d_data_grid();
define set_bin_corr_factor();
define set_gauss_line_par();
define set_lines_par_fun();
define set_par_from_confmap_table();
define set_simputfile_model_grid();
define set_xyfit_qualifier();
define set_xyfit_sys_err_frac();
define setPGPLOTenv();
define shift_2d();
define shift_2d_old();
define shift_hist();
define shift_intpol();
define show_slang_code();
define simfit_namespace();
define simple_gpile2_defaults();
define simple_gpile2_fit();
define simple_gpile3_defaults();
define simple_gpile3_fit();
define simple_gpile_defaults();
define simple_gpile_fit();
define simplify_polygon();
define simput_athenacrab();
define simulate_data();
define simultaneous_fit();
define sinwave2_default();
define sinwave2_fit();
define sinwave_defaults();
define sinwave_fit();
define SIprefix();
define sixte_create_img_wfi();
define sixte_create_img_xifu();
define sixte_psfgen();
define sltable();
define sltableBuildBody();
define socket_get_object();
define socket_send_object();
define solarAbund();
define solve_2d_system_of_equations();
define solve_Eddington_quartic_equation();
define solve_Lane_Emden_equation();
define solveODEbyIntegrate();
define sort_arrays();
define sort_struct_arrays();
define sort_struct_fields();
define spearmanrho();
define sphere_urand();
define spherical_volume();
define split_and_epfold_lc();
define split_lc_at_gaps();
define split_struct();
define split_struct_cols();
define sqrsinc_xyfit();
define SRT_get_mw_spectra();
define SRT_image();
define SRT_plot_mw_spectra();
define SRT_read();
define SRT_show_chunks();
define SRT_spectrum();
define stacking_vlbi_images();
define step1_fit();
define step2_fit();
define steppar();
define steppar_get_bestparams();
define steppar_get_conf();
define steppar_load();
define steppar_plot();
define steppar_save();
define storevar();
define stretch_hist();
define strftime_MJD();
define string2sys_err_array();
define string_intersection();
define string_match_perl();
define strjoin_wrap();
define strreplace1();
define strreplaces();
define strsplit();
define struct_array();
define struct_array_2_struct_of_arrays();
define struct_copy();
define struct_field_or_default();
define struct_fields_list_to_array();
define struct_of_arrays_2_struct_array();
define struct_print_field_statistics();
define struct_to_assoc_array();
define sum_lc();
define sunpos();
define switch_tbnew_rel_abund();
define sys_err_array2string();
define table_add_row();
define table_copy();
define table_print_TeX();
define tai2tt();
define tai2utc();
define taylor();
define taylorcoeff_from_struct();
define tbnew_metal_fit();
define tbnew_simple_fit();
define tbnew_simple_z_fit();
define tcg2tt();
define tcp_client();
define tcp_get_server_handle();
define tcp_server();
define tdb2tt();
define TeX_value_pm_error();
define thawedParameters();
define three_chip_defaults();
define three_chip_fit();
define tilt_defaults();
define time_array();
define timegap();
define timeOfPhase();
define timeseg();
define timesync();
define transpose_table();
define try_or_timeout();
define tt2tai();
define tt2tcg();
define tt2tdb();
define tt2ut1();
define unique_n();
define UNIXtime2MJD();
define unpack_obj();
define unset_lines_par_fun();
define use_HeTriplet();
define use_simple_gpile();
define ut12tt();
define utc2tai();
define UTC2UNIXtime();
define v_relativistic_to_optical();
define vacuum_to_air();
define vector_astro();
define vector_to_spherical();
define Voigt_complex();
define Voigt_fit();
define voigt_profile();
define vradbinary_default();
define vradbinary_fit();
define vsop87();
define wcs_axes_plot();
define Weibull_default();
define Weibull_fit();
define weighted_mean();
define where_max();
define where_min();
define winding_number_polygon();
define WorldCoastLine();
define write_simput_dummy_bkg();
define write_slurm_script();
define write_spix();
define x1axis();
define x2axis();
define x2axis_with_tics();
define x2label();
define xfig_3d_orbit_on_cube();
define xfig_compress_eps();
define xfig_draw_orbit();
define xfig_draw_orbit_3D();
define xfig_get_normalized_position();
define xfig_get_normalized_x_position();
define xfig_get_normalized_y_position();
define xfig_mix_colors();
define xfig_multibox();
define xfig_plot_allbezier();
define xfig_plot_colormap();
define xfig_plot_conf();
define xfig_plot_confmap();
define xfig_plot_data();
define xfig_plot_distribution_matrix();
define xfig_plot_epfpd();
define xfig_plot_image();
define xfig_plot_ionization_structure();
define xfig_plot_model();
define xfig_plot_params();
define xfig_plot_res();
define xfig_plot_sky_img();
define xfig_plot_spectrum();
define xfig_plot_unfold();
define xfig_polarplot_new();
define xfigplot_hardnessratio_grid();
define xfigplot_pulseprofile_energy_map();
define xnotice_atime();
define xspec_to_isis();
define XSTAR_read_pops();
define XTEtoRealtime();
define xyfit_fun();
define xyfit_residuals();
define xylabel_in_box();
define y1axis();
define y2axis();
define y2label();
define yearOfMJD();
define z2fold();
define zams();
define zang();
define zfc_defaults();
define zfc_fit();
define zpc_tbnew_simple_fit();
define zsize();
define zsquare();
%
% Functions explicitly defined as public
% (a total of 110 functions)
%
public define add_cursor_line();
public define add_line();
public define add_plot_unit();
public define apj_size();
public define apply_i2x_grp();
public define apply_x2i_grp();
public define calc_eqw();
public define calc_flux();
public define cash();
public define chain_confidence();
public define chain_hist();
public define chain_hist2d();
public define chain_plot();
public define chain_set_params();
public define clear_all();
public define close_print();
public define conf_map();
public define conf_map_fail_hook();
public define corback();
public define corfile();
public define data_flux();
public define data_list();
public define delete_line();
public define emcee();
public define emcee_chain_hist_collect();
public define emcee_merge();
public define eqw();
public define fancy_plot_unit();
public define fcorback_fit();
public define fcorfile_fit();
public define find_line();
public define fitfun_init_cache();
public define fstat_gehr();
public define fstat_mod();
public define fstat_norm();
public define get_xyfit_fun();
public define gridmapping();
public define i2x_grp();
public define init_line();
public define keynote_size();
public define line_model_prefix();
public define list_all();
public define lmdif();
public define load_radio();
public define load_scaled_data();
public define log_grid_hook();
public define marq();
public define minim();
public define minus();
public define minus_fit();
public define model_flux();
public define mpfit();
public define mpi_emcee();
public define mpi_new_best_fit_hook();
public define mpi_read_chain();
public define mrg_grid();
public define msg();
public define new_plot_labels();
public define nice_width();
public define old_read_chain();
public define old_to_new_chain();
public define open_print();
public define pg_color();
public define pg_info();
public define pha2_bkg();
public define plot_comps();
public define plot_counts();
public define plot_covar();
public define plot_data();
public define plot_double();
public define plot_fit_model();
public define plot_residuals();
public define plot_unfold();
public define plotxy();
public define read_chain();
public define read_optical();
public define read_radio();
public define reset_plot_defaults();
public define rolling_fit();
public define save_line_models();
public define sep_grid();
public define set_plot_labels();
public define set_plot_widths();
public define set_power_scale();
public define simbad2ds9();
public define sitar_avg_cpd();
public define sitar_avg_psd();
public define sitar_bin_events();
public define sitar_define_psd();
public define sitar_epfold_rate();
public define sitar_global_optimum();
public define sitar_glvary();
public define sitar_lags();
public define sitar_lbin_cpd();
public define sitar_lbin_psd();
public define sitar_make_data_cells();
public define sitar_pfold_event();
public define sitar_pfold_rate();
public define sitar_readasm();
public define sitar_rebin_rate();
public define sov();
public define subplex();
public define tilt_fit();
public define usr_grid();
public define vec_hist();
public define vec_hist2d();
public define write_chain();
public define write_plot();
public define write_x_grp();
public define x2i_grp();
public define xfig_chain_plot();
%
% Functions explicitly defined as private
% (a total of 667 functions)
%
private define __aindex();
private define __alarm_signal_handler();
private define __assemblePanstarrsUrl();
private define __bounds_lines();
private define __calc_gehrels_beta();
private define __calc_gehrels_gamma();
private define __CL_to_S();
private define __counts_from_dataset();
private define __emcee_params();
private define __emceeFitsReadChecks();
private define __emceeFitsWriteChecks();
private define __emceeFitsWriteT1();
private define __emceeFitsWriteT2();
private define __emceeFitsWriteT3();
private define __emceeMikeWriteChecks();
private define __f_access();
private define __Faddeeva();
private define __findex();
private define __get();
private define __get_background_counts();
private define __panstarrsDownloadCallback();
private define __panstarrsDownloadFile();
private define __panstarrsDownloadImages();
private define __panstarrsDownloadStacks();
private define __panstarrsDownloadStacksAux();
private define __panstarrsDownloadWarps();
private define __panstarrsDownloadWarpsAux();
private define __panstarrsDownloadWriteCallback();
private define __panstarrsGetOptions();
private define __panstarrsInsertStackAuxMatch();
private define __panstarrsInsertStackMatch();
private define __panstarrsInsertWarpAuxMatch();
private define __panstarrsInsertWarpMatch();
private define __panstarrsRetrieveSite();
private define __panstarrsSortUrls();
private define __panstarrsXMLCharHandler();
private define __panstarrsXMLStartElementHandler();
private define __parsePanstarrsPage();
private define __print_xfig_plot_spectrum_version();
private define __runs_info();
private define __runs_p();
private define _aglc();
private define _aglc2();
private define _aglc_phased();
private define _aglc_phased2();
private define _client_get();
private define _client_greet();
private define _client_send();
private define _epferror_pget();
private define _fits_save_fit_check_values();
private define _fits_save_fit_extname();
private define _fits_save_fit_make_struct();
private define _get_lambda_parameters_of_lines();
private define _is_primitive_datatype();
private define _lc_from_event_single();
private define _ode_torque();
private define _pack_obj_fmt();
private define _print_apropos();
private define _print_aproposer();
private define _print_err();
private define _pulseprofile_decompose_fitwrapper();
private define _pulseprofile_phase_connect_ccfint();
private define _server_get();
private define _server_greet();
private define _server_send();
private define _simultaneous_fit_comb_chi_sqr();
private define _simultaneous_fit_get_instances();
private define _simultaneous_fit_help();
private define _simultaneous_fit_map_global();
private define _simultaneous_fit_map_groups();
private define _simultaneous_fit_parname();
private define _simultaneous_fit_steppar_fit();
private define _sitar_pfold_event();
private define _sitar_pfold_event_profile();
private define _sitar_pfold_profile();
private define _sitar_pfold_rate();
private define _slang_primitive_datatype_format_string();
private define _slang_variable_to_file();
private define _steppar_get_bestparams();
private define _steppar_get_params();
private define _tcp_disconnect();
private define _tcp_receive();
private define _tcp_receive_msg();
private define _tcp_receive_obj();
private define _tcp_send();
private define _tcp_send_msg();
private define _tcp_send_obj();
private define _tcp_server_cleanclients();
private define _tcp_server_handle();
private define _tcp_server_shutdown();
private define _unpack_obj_fmt();
private define _xfig_draw_orbit_3D_ellipse();
private define _xfig_draw_orbit_3D_proj();
private define _xfig_plot_sky_img();
private define aglc_expno_to_phase();
private define aglc_expno_to_time();
private define aglc_filter_gti();
private define aglc_get_expno_filt_range();
private define aglc_par_check();
private define aglc_phased_par_check();
private define aglc_phased_usage();
private define aglc_set_ephem();
private define aglc_time_to_expno();
private define aglc_usage();
private define an_akritas();
private define append_interpol_table();
private define apply_transform_coord();
private define apply_transform_vector();
private define apropos_slang();
private define ARFmultiplyDetQE();
private define ARFmultiplyFactor();
private define ARFmultiplyFilter();
private define ARFmultiplySupport();
private define ARFreadHenkeTransm();
private define array2image_pixel();
private define array_split_at_extrema();
private define assoc_has_default_value();
private define aStar_constructPath();
private define aStar_estimateCost();
private define aStar_pathLength();
private define atime_sim_residuals_maxdif();
private define atmosphere_hohenkerk();
private define AVLNodeBalanceFactor();
private define AVLNodeHeight();
private define AVLNodeUpdateHeight();
private define AVLTreeDeleteKey();
private define AVLTreeDeleteKey_helper();
private define AVLTreeDeleteNode();
private define AVLTreeInsert();
private define AVLTreeInsert_helper();
private define AVLTreeRebalance();
private define AVLTreeRotate();
private define back_fun_string();
private define BinarySearchPrintNode();
private define BinarySearchTraversal();
private define BinarySearchTreeDeleteKey();
private define BinarySearchTreeDeleteKey_helper_iterative();
private define BinarySearchTreeDeleteKey_helper_recursive();
private define BinarySearchTreeDeleteNode();
private define BinarySearchTreeInsert();
private define BinarySearchTreeInsert_helper();
private define BinarySearchTreeKeyExists();
private define BinarySearchTreeMaximum();
private define BinarySearchTreeMinimum();
private define BinarySearchTreeNodePredecessor();
private define BinarySearchTreeNodePrint();
private define BinarySearchTreeNodeSuccessor();
private define BinarySearchTreePrint();
private define BinarySearchTreeSearch();
private define binWidthKaastraBleeker();
private define blend_palette();
private define buildOptimalBins();
private define buildOptimalBinsGroup();
private define calc_backscales();
private define calc_pattern_fractions_of_grid();
private define calculateSummedReferenceLightcurve();
private define call_func_with_qualifiers();
private define capitalize();
private define ccode();
private define check_chain_headers();
private define check_parameters_headers();
private define clean_messages();
private define client_main();
private define colormap_palette();
private define compare_confmap_data_indices();
private define compare_header();
private define compare_stat();
private define computed_roll_angle();
private define confmap_save_hook();
private define cons_event_flags();
private define convert_GTI();
private define convert_negative_height_to_rg();
private define convert_negative_radius_to_rg();
private define convMAS();
private define correct_model();
private define countsInResolver();
private define create_back();
private define create_backscale();
private define cross_power_density();
private define cs_outcode();
private define cubehelix_palette();
private define dark_palette();
private define datatype_to_string();
private define datplot_err();
private define dcode();
private define define_generic();
private define define_xyfun();
private define deg2dms_rounded();
private define dimi();
private define distM_array();
private define distM_struct();
private define distribute_runs();
private define diverging_palette();
private define dopplerorbit_defaults();
private define dopplerorbit_fit();
private define dorDFT();
private define EchoMap_redist();
private define EchoMap_respon();
private define elementType();
private define emceeDefaultFile();
private define emceeDrawSet();
private define emceeFileFits();
private define emceeFileFitsClose();
private define emceeFileFitsCreate();
private define emceeFileFitsOpen();
private define emceeFileFitsRead();
private define emceeFileFitsWrite();
private define emceeFileMike();
private define emceeFileMikeCreate();
private define emceeFileMikeOpen();
private define emceeFileMikeWrite();
private define emceeFilePar();
private define emceeFileParRead();
private define emceeForkAbort();
private define emceeForkEnterHarbor();
private define emceeForkLeaderReceive();
private define emceeForkLeaderSend();
private define emceeForkMemberReceive();
private define emceeForkMemberSend();
private define emceeForkSetSail();
private define emceeInitChain();
private define emceeInitChainPick();
private define emceeInitFile();
private define emceeInitFilePick();
private define emceeInitGauss();
private define emceeInitGaussPick();
private define emceeInitUniform();
private define emceeInitUniformPick();
private define emceeLoop();
private define emceeMPIAbort();
private define emceeMPIEnterHarbor();
private define emceeMPILeaderReceive();
private define emceeMPILeaderSend();
private define emceeMPIMemberReceive();
private define emceeMPIMemberSend();
private define emceeMPISetSail();
private define emceeOption();
private define emceeProgressNone();
private define emceeProgressReport();
private define emceeProgressReportFinish();
private define emceeProgressReportReporter();
private define emceeSerialInit();
private define emceeSetup();
private define emceeSetupEngine();
private define emceeSetupGears();
private define emceeSetupLeader();
private define emceeSetupWriteBuffer();
private define emceeShipFork();
private define emceeShipMPI();
private define emceeShipSerial();
private define emceeSignalHandler();
private define emceeStepStretch();
private define emceeStepStretchMove();
private define enflux_default();
private define enflux_fit();
private define epfold_rate_use();
private define erfinv_kernel();
private define error_estimate();
private define exit_all_slaves();
private define f1();
private define f2();
private define fermi_new_centered_energy_grid();
private define file2string();
private define file_exists();
private define filter_ionization_states();
private define filterFouQuantForFrequencyRange();
private define fisher_yates();
private define fork_probe_status();
private define forked_conf();
private define free_par();
private define from_linear();
private define fstat();
private define galaxy_elliptical_defaults();
private define galridge_default();
private define galridge_fit();
private define garf_gaps_usage();
private define gauss_prior();
private define gehrels_error_lower_11();
private define gehrels_error_lower_12();
private define gehrels_error_lower_13();
private define gehrels_error_lower_14();
private define gehrels_error_upper_10();
private define gehrels_error_upper_7();
private define gehrels_error_upper_9();
private define get_arcmin_area();
private define get_bkg_src_flux();
private define get_bounds();
private define get_col_map();
private define get_corr_factor();
private define get_fit_options_from_qualifiers();
private define get_index();
private define get_ions();
private define get_mod_num();
private define get_params_and_stepsize();
private define get_qualifier();
private define get_reflfrac_agrid();
private define get_reflfrac_hgrid();
private define get_reflfrac_struct();
private define get_simputfile_empty_struct();
private define get_string_starts();
private define get_tic_pos_dezimal();
private define get_vignet_data();
private define get_vignet_keys();
private define get_xyfit_metadata();
private define getLightcurve();
private define getPCUgti();
private define gliding_average();
private define glob_opt_use();
private define greiner_insert_between();
private define gti_flag();
private define h_akritas();
private define handle_physical_quantity_units();
private define handle_user_message();
private define hp_loop();
private define hp_loopit();
private define hpluv2lch();
private define hpluv_palette();
private define hsl_palette();
private define hsluv2lch();
private define hsluv_palette();
private define I();
private define icdf();
private define improvement_by_interpolation();
private define indarray();
private define init_setup();
private define init_walker();
private define integrate_diff_flux();
private define interpol_frac();
private define intersect_line_line();
private define ipol_reflFracStruct();
private define is_free();
private define iterateParameterVector();
private define jet_speed_default();
private define jpl_name2jplid();
private define jplinter();
private define kill_all_slaves();
private define l2y();
private define Lane_Emden_eom();
private define lch2hpluv();
private define lch2hsluv();
private define lch2luv();
private define ldist();
private define light_palette();
private define list_par_string_pattern();
private define list_xypar_string();
private define local_eval_simputfile();
private define log_interpol();
private define log_post_bin();
private define log_post_evt();
private define log_post_evt_alpha();
private define log_post_evt_mean();
private define log_post_evt_rate();
private define log_post_max();
private define logpn();
private define luv2lch();
private define luv2xyz();
private define m2w();
private define make_array_a_list();
private define make_data();
private define make_default();
private define make_exp_and_back();
private define make_flux();
private define make_indices();
private define map_hexstr();
private define map_hsl();
private define map_hue();
private define map_max();
private define map_rhsl();
private define massage_param_structs();
private define matrix33_as_string();
private define matrix33_chs();
private define matrix33_mat_add_mat();
private define matrix33_mat_div_mat();
private define matrix33_mat_div_real();
private define matrix33_mat_mul_mat();
private define matrix33_mat_mul_real();
private define matrix33_mat_mul_vector();
private define matrix33_mat_sub_mat();
private define matrix33_real_div_mat();
private define matrix33_real_mul_mat();
private define max_chroma_for_lh();
private define max_safe_chroma_for_l();
private define maybe_finished();
private define mb_gauss_2d();
private define mdc_use();
private define message_param_structs();
private define model_start();
private define mpi_gather_results();
private define mpi_init_walker();
private define mpi_slave_fun();
private define mpi_update_walker();
private define mpi_write_conf_file();
private define multibknpower_default();
private define multiplotit();
private define new_AVLNode();
private define new_BinarySearchTreeNode();
private define new_greiner_vertex();
private define new_greiner_vertexlist();
private define ntree_get();
private define ntree_get_method();
private define ntree_getAll();
private define ntree_getAll_method();
private define ntree_getChilds();
private define ntree_getChilds_method();
private define ntree_insert();
private define ntree_insert_method();
private define ntree_remove();
private define ntree_remove_method();
private define ntree_type_mapQuad_insert();
private define ntree_type_mapQuad_lookup();
private define ntree_type_mapQuad_lookup_method();
private define ntree_type_mapQuad_new();
private define null_it();
private define p_range_label();
private define p_reset();
private define par_array_idx();
private define par_cmp();
private define par_format_number();
private define par_format_text();
private define par_read();
private define param_default_structure();
private define param_string();
private define parse_unit_power();
private define pd_set();
private define periodFoldEvents();
private define periodFoldLC();
private define pfold_event_use();
private define pfold_rate_use();
private define phasebin_to_gti();
private define phflux_default();
private define phflux_fit();
private define physical_quantity_add();
private define physical_quantity_chs();
private define physical_quantity_div();
private define physical_quantity_div_qs();
private define physical_quantity_div_sq();
private define physical_quantity_mul();
private define physical_quantity_mul_qs();
private define physical_quantity_mul_sq();
private define physical_quantity_power();
private define physical_quantity_string();
private define physical_quantity_sub();
private define plot_info_field();
private define plot_single_ion();
private define plot_xyresiduals();
private define power_scale();
private define prepare_data_for_pp_plot();
private define print_ephem_data();
private define print_info();
private define print_log();
private define print_stdout_and_file();
private define probmap_2dgauss();
private define probmap_add();
private define probmap_add_method();
private define probmap_calcmap();
private define probmap_calcmap_maxNorm();
private define probmap_calcpath();
private define probmap_calcpath_estimate();
private define probmap_calcpath_method();
private define probmap_delete();
private define probmap_delete_method();
private define probmap_getmap();
private define probmap_getmap_method();
private define probmap_getpoint();
private define probmap_getpoint_method();
private define probmap_worlddefined();
private define probmap_worlddefined_method();
private define probmap_xrange();
private define probmap_xrange_method();
private define probmap_yrange();
private define probmap_yrange_method();
private define projected_angle_for_eye();
private define psd_lc_normalized_lightcurve();
private define pulsar_fit();
private define pulsarorbit_defaults();
private define pulsarorbit_fit();
private define pulsartaylor_defaults();
private define pulsartaylor_fit();
private define pulsartorque_defaults();
private define pulsartorque_fit();
private define qphcol();
private define quality_function();
private define quasar_accr_defaults();
private define r_e();
private define rand_gauss_cut();
private define ray_length_until_intersect();
private define read_ascl();
private define read_leapfile();
private define read_xspec();
private define readArray();
private define readasm_use();
private define reap_childs();
private define reb_evt_use();
private define reb_rat_use();
private define receive_spectrum();
private define recv_slave_result();
private define redden_defaults();
private define reform_revidx();
private define refract_integrand();
private define refract_mu();
private define relat_aberration();
private define res_set();
private define resplot();
private define responseEval();
private define responseEvalCleanup();
private define responseEvalSetup();
private define restart_slaves_with_new_params();
private define rgb2xyz();
private define rmf_profile();
private define save_results();
private define search();
private define searchkey();
private define send_next_task();
private define send_spectrum();
private define set_pars();
private define set_plot_parms();
private define set_plot_parms_struct();
private define set_simputfile_flux();
private define setup_xyfun();
private define sigma_akritas();
private define simultaneous_fit_add_data();
private define simultaneous_fit_add_data_old();
private define simultaneous_fit_apply_logic();
private define simultaneous_fit_copy_par();
private define simultaneous_fit_delete_data();
private define simultaneous_fit_eval_global();
private define simultaneous_fit_eval_groups();
private define simultaneous_fit_filter_groups();
private define simultaneous_fit_fit_fun();
private define simultaneous_fit_fit_global();
private define simultaneous_fit_fit_groups();
private define simultaneous_fit_fit_pars();
private define simultaneous_fit_fit_pars_collect();
private define simultaneous_fit_fit_pars_jobfile();
private define simultaneous_fit_fit_pars_run_job();
private define simultaneous_fit_fit_smart();
private define simultaneous_fit_freeze();
private define simultaneous_fit_get_par();
private define simultaneous_fit_group_stats();
private define simultaneous_fit_list_data();
private define simultaneous_fit_list_global();
private define simultaneous_fit_list_groups();
private define simultaneous_fit_load();
private define simultaneous_fit_load_model();
private define simultaneous_fit_mpi_fit_pars();
private define simultaneous_fit_mpi_fit_pars_jobfiles();
private define simultaneous_fit_plot_group();
private define simultaneous_fit_restore();
private define simultaneous_fit_save();
private define simultaneous_fit_select_groups();
private define simultaneous_fit_set_global();
private define simultaneous_fit_set_par();
private define simultaneous_fit_set_par_fun();
private define simultaneous_fit_set_par_fun_history();
private define simultaneous_fit_setrestore();
private define simultaneous_fit_sort_params();
private define simultaneous_fit_steppar();
private define simultaneous_fit_thaw();
private define simultaneous_fit_unset_global();
private define slave_handler();
private define slave_next_task();
private define slave_spawned_callback();
private define slave_task();
private define slog();
private define sltableBuildColHead();
private define sltableBuildFooter();
private define sltableBuildHeader();
private define sltableCaption();
private define sltableColHead();
private define sltableColHeads();
private define sltableColJust();
private define sltableEOL();
private define sltableGetType();
private define sltableNoteMark();
private define sltableNoteText();
private define sltableRound();
private define sltableRoundToDigit();
private define sltableSetType();
private define sltableTabs();
private define space_converter();
private define sron_palette();
private define start_plot();
private define stretchInverseCDF();
private define struct_fieldnames_to_lowercase();
private define t_m();
private define tableModelAddAdditionalParameter();
private define tableModelAddParameter();
private define tableModelClose();
private define tableModelInvalid();
private define tableModelOpenFile();
private define tableModelParameterList();
private define tableModelReset();
private define tableModelSetAdditionalFunction();
private define tableModelSetEnergyGrid();
private define tableModelSetFunction();
private define tableModelSummary();
private define tableModelUniqueParameter();
private define tableModelWriteGrid();
private define tableModelWriteHeader();
private define tableModelWriteParameter();
private define tableModelWriteParameterHeader();
private define tableModelWriteParameterInitials();
private define tableModelWriteParameterMethods();
private define tableModelWriteParameterNames();
private define tableModelWriteParameterUnits();
private define tableModelWriteParameterValues();
private define tableModelWriteSpectra();
private define tableModelWriteSpectraHeader();
private define tableModelWriteTable();
private define tau_kendall();
private define tau_kendall_partial();
private define tbnew_metal_defaults();
private define tbnew_simple_defaults();
private define tbnew_simple_z_defaults();
private define term_process_list();
private define tfun();
private define to_linear();
private define to_rgb();
private define tokenize();
private define tout_poly();
private define trans_strct();
private define trans_strct_and_array();
private define tree_key_compare();
private define try_fits_read_key_or_fill();
private define unit_qualifiers();
private define unit_to_baseunit();
private define unity();
private define unshift_default_hook();
private define unshift_fit();
private define update_walker();
private define vaxis();
private define vector_div_scalar();
private define verbexec();
private define verify_frequency_binning();
private define verifyLightcurveTime();
private define void();
private define Voigt_defaults();
private define w2m();
private define wait_for_slave_to_exit();
private define write_bintable();
private define write_cpd_to_ascii();
private define write_dummy_bkg_ctsmap();
private define write_movie_img_png();
private define write_plot_head();
private define write_to_outfile();
private define writeArray();
private define x_axis_up();
private define xerplot();
private define xfig_prepare_dm();
private define xfig_prepare_dm_xpp();
private define xfig_prepare_dm_ypp();
private define xfig_prepare_dm_zpp();
private define xfig_set_eye_with_computed_roll_angle();
private define xfps_create_string_array();
private define xfps_create_xfig_object();
private define xfps_labels();
private define xfps_plot_data();
private define xfps_plot_model();
private define xfps_plot_res();
private define xfps_read_args();
private define xfps_setup_gaps();
private define xyz2luv();
private define xyz2rgb();
private define y1();
private define y2();
private define y2l();
private define zhangfunction();
private define zpc_tbnew_simple_defaults();

%
% Code from src/slang/math/Aitoff_projection.sl
%
define Aitoff_projection()
%!%+
%\function{Aitoff_projection}
%\synopsis{computes an Aitoff projection}
%\usage{(Double_Type x, y) = Aitoff_projection(Double_Type l, b);}
%\qualifiers{
%\qualifier{deg}{\code{l} and \code{b} are in degrees, not in radian}
%\qualifier{normalized}{\code{x} and \code{y} are normalized (by \code{PI/2})
%                 such that \code{abs(x) <= 2} and \code{abs(y) <= 1}.}
%}
%\description
%    NOTE: for astronomical purposes you probably want to use the
%    equal area Hammer-Aitoff projection rather than an Aitoff-projection.
%    Erroneously many astronomers call the Hammer-Aitoff projection an
%    Aitoff projection
%
%    \code{x = 2 * cos(b) * sin(l/2) / sinc(alpha);}\n
%    \code{y = sin(b) / sinc(alpha);}\n
%    where
%      \code{cos(alpha) = cos(b) * cos(l/2)}\n
%    and
%      \code{sinc(alpha) = sin(alpha)/alpha}
%\seealso{Hammer_projection, Lambert_Equal_Area_projection}
%!%-
{
  variable l, b;
  switch(_NARGS)
  { case 2: (l, b) = (); }
  { help(_function_name()); return; }

  vmessage("WARNING: Aitoff_projection is a map projection that is very rarely used");
  vmessage("   You probably want to use the Hammer-Aitoff projection, which is");
  vmessage("   implemented in the function Hammer_projection!");

  if(qualifier_exists("inverse")) {
      throw RunTimeError, "Error: inverse of Aitoff_projection is not yet implemented";
  }

  if(qualifier_exists("deg")) { 
      l *= PI/180;
      b *= PI/180;
  }
  if(typeof(l)==Array_Type) {
      while(any(abs(l)>PI)) { 
	  l[where(l> PI)] -= 2*PI;
	  l[where(l<-PI)] += 2*PI;
      }
  } else {
      while(l> PI)  l-= 2*PI;
      while(l<-PI)  l+= 2*PI;
  }



  variable alpha = acos(cos(b)*cos(l/2.));
  variable inv_sinc_alpha = alpha/sin(alpha);
  if(typeof(alpha)==Array_Type) {
    inv_sinc_alpha[where(alpha==0)] = 1.;
  } else {
    if(alpha==0)  inv_sinc_alpha = 1;
  }
  if(qualifier_exists("normalized"))  inv_sinc_alpha /= PI/2;

  variable sig=+2.;
  if(qualifier_exists("astronomical")) {
      sig=-2.;
  }

  return sig*cos(b)*sin(l/2.)*inv_sinc_alpha,
         sin(b)*inv_sinc_alpha;
}

%
% Code from src/slang/astronomy/AzEl_from_RAdec.sl
%
define AzEl_from_RAdec()
%!%+
%\function{AzEl_from_RAdec}
%\synopsis{computes horizontal (azimut, elevation) coordinates from a point's (RA, dec) at a time MJD}
%\usage{(az, el) = AzEl_from_RAdec(RA, dec, MJD,  longw, lat);}
%\description
%    This function is deprecated. Please use equatorial2horizon instead.
%\seealso{equatorial2horizon,horizon2equatorial}
%!%-
{
  variable RA, dec, MJD, longw, lat;
  switch(_NARGS)
  { case 5: (RA, dec, MJD, longw, lat) = (); }
  { help(_function_name()); return; }

  RA = angle_to_rad(RA; unit="hms");
  dec = angle_to_rad(dec; unit="deg");

  variable az,el;

  (az,el)=equatorial2horizon(RA,dec;lat=lat*PI/180.,lon=longw*PI/180.,JD=MJD,mjd);
  
  return az*180./PI, el*180./PI;
}

%
% Code from src/slang/physics/B_lambda.sl
%
define B_lambda() {
%!%+
%\function{B_lambda}
%\synopsis{Calculates the spectral density for black body radiation (wavelength space)}
%\usage{Double_Type Blambda = B_lambda(lambda,T);}
%
%\qualifiers{
%\qualifier{keV}{T argument is kT in keV}
%\qualifier{A}{Wavelength is given in Angstroms}
%\qualifier{Angstrom}{Wavelength is given in Angstroms}
%}
%\description
%    This function returns the spectral energy density of a black body,
%    in units erg/cm^2/s/cm/sr 
%    Either lambda or T can be an array.
%    
%!%-
    variable lambda,kT;

    switch(_NARGS)
    { case 2: (lambda,kT)=(); }
    { help(_function_name()); return; }

    variable h=6.62606885e-27;% erg s
    variable c=2.99792457e10; % cm/s
    variable k=1.3806504e-16; % erg/K

    variable keV2erg=1.60218e-9; % keV -> erg
    variable A2cm  =1e-8;

    if (qualifier_exists("Angstrom") or qualifier_exists("A")) {
	lambda*=A2cm;
    }
    if (qualifier_exists("keV")) {
	kT*=keV2erg;
    }
	    
    variable Blambda=2.*h*(c/lambda)*(c/lambda)/lambda/lambda/lambda/(exp(h*c/lambda/kT)-1.);

    return Blambda;
}

%
% Code from src/slang/physics/B_nu.sl
%
define B_nu() {
%!%+
%\function{B_nu}
%\synopsis{Calculates the spectral density for black body radiation (frequency space)}
%\usage{Double_Type Bnu = B_nu(nu,T);}
%
%\qualifiers{
%\qualifier{keV}{1st argument is energy in keV,
%                  T argument is kT in keV}
%\qualifier{MHz}{Frequency is given in MHz}
%\qualifier{GHz}{Frequency is given in GHz}
%}
%\description
%    This function returns the spectral energy density of a black body,
%    in units erg/cm^2/s/Hz/sr 
%    Either nu or T can be an array.
%    
%!%-
    variable par1,par2;

    switch(_NARGS)
    { case 2: (par1,par2)=(); }
    { help(_function_name()); return; }
    
    variable h=6.62606885e-27;% erg s
    variable c=2.99792457e10; % cm/s
    variable k=1.3806504e-16; % erg/K

    variable keV2erg=1.60218e-9; % keV -> erg
    variable MHz2Hz=1e6;      % MHz -> Hz
    variable GHz2Hz=1e9;      % GHz -> Hz
    variable A2cm  =1e-8;

    variable kT=par2;
    variable nu=par1;
    if (qualifier_exists("keV")) {
	% temperature argument is in keV; convert to erg
	kT*=keV2erg;
	% frequency argument is in keV -> convert to Hz
	nu*=keV2erg/h;
    } else {
	kT=k*kT;
	if (qualifier_exists("MHz")) {
	    nu*=MHz2Hz;
	} else {
	    if (qualifier_exists("GHz")) {
		nu*=GHz2Hz;
	    }
	}
    }
	
    variable Bnu=2*h*(nu/c)*(nu/c)*nu/(exp(h*nu/kT)-1.);

    return Bnu;
}

%
% Code from src/slang/astronomy/BinaryCor.sl
%
define BinaryCor()
%!%+
%\function{BinaryCor}
%\synopsis{Removes the influence of the doublestar motion for circular or eliptical orbits.}
%\usage{Array_Type t = BinaryCor(Array_Type OR Double_Type time), time in MJD}
%\qualifiers{
%\qualifier{asini}{Projected semi-major axis [lt-secs], Mandatory}
%\qualifier{porb}{Orbital period at the epoch [days], Mandatory}
%\qualifier{eccentricity}{Eccentricity, (0<=e<1)}
%\qualifier{omega}{Longitude of periastron [degrees], mandatory}
%\qualifier{t0}{epoch for mean longitude of 0 degrees (periastron, MJD)}
%\qualifier{t90}{epoch for mean longitude of 90 degrees (MJD)}
%\qualifier{pporb}{rate of change of the orbital period (s/s) (default 0)}
%\qualifier{limit}{absolute precision of the correction of the  computation (days, default: 1D-6)}
%\qualifier{maxiter}{stop Newton-Raphson iteration after maxiter steps if limit is not reached (default: 20)}
%}
%\description
%    (transcribed from IDL-programm BinaryCor.pro)
%
%    For each time, the z-position of the emitting object is computed
%    and the time is adjusted accordingly. This is iterated until
%    convergence is reached (usually only one iteration is necessary,
%    even in high elliptic cases).
%
%    Follows equations from Hilditch's book and has also been
%    checked against fasebin/axBary. All codes give identical results,
%    (to better than 1d-7s) as checked by a Monte Carlo search using
%    1d7 different orbits.
%
%    qualifiers t90 and t0 have to be in days and in the same time
%    system as time (e.g. JD or MJD)
%
%    Circular orbits:
%         * if time of lower conjunction Tlow is known, set
%           t0=Tlow and omega=0
%         * if time of ascending node is known, Tasc, set
%           t90=Tasc and omega=0
%         * if time of mid eclipse is known, Tecl, set
%           t0=Tecl-0.25*porb and omega=0
%!%-
{
  variable time;
  switch(_NARGS)
  { case 1: time = (); }
  { help(_function_name()); return; }
  %%make sure that time is an array
  time = [time];

  %% qualifiers
  variable asini   = qualifier("asini");
  variable porb    = qualifier("porb");
  variable ecc     = qualifier("eccentricity");
  variable omega_d = qualifier("omega");
  variable t90     = qualifier("t90");
  variable t0      = qualifier("t0");
  variable pporb   = qualifier("pporb",0.0);
  variable limit   = qualifier("limit",1e-6);
  variable maxiter = qualifier("maxiter",20);

  %% check if required qualifiers assini, poen and eccentricity
  %% are assigned, die else
  ifnot(qualifier_exists("asini"))
  { message("error: asini value not assigned");
    return;
  }

  ifnot(qualifier_exists("porb"))
  { message("error: porb value not assigned");
    return;
  }

  ifnot(qualifier_exists("eccentricity"))
  { message("error: eccentricity value not assigned");
    return;
  }

  %% look that one and only one of the t0 and t90 values is assigned
  ifnot(qualifier_exists("t0") or qualifier_exists("t90"))
  { message("error: need t90 or t0 value");
    return;
  }

  if(qualifier_exists("t90")  and qualifier_exists("t0"))
  { message("error: Only one of the t90 and t0 arguments is allowed");
    return;
  }

  %% check that omega is set for noncircular orbits
  if(ecc != 0)
    ifnot(qualifier_exists("omega"))
    { message("error: Need omega keyword for noncircular orbits");
      return;
    }
  else  %% if omega not-set for circular orbits, set to 0.
    ifnot(qualifier_exists("omega"))
      omega_d=0.;

  %%check if eccentricity is 0<=ecc<1
  if(ecc < 0)
  { message("error: eccentricity must be positive!");
    return;
  }
  if(ecc >=1)
  { message("error: Orbit correction has only been implemented for circular and elliptic orbits");
    return;
  }

  %% compute t0 from t90 if necessary
  %% note that this assumes pporb<<porb, which can be generally assumed
  if(qualifier_exists("t90"))
   t0 = t90+(omega_d-90.)/360. * porb;

  %% set default values
  if (maxiter <=0)
    maxiter=20;

  variable asini_d = double(asini)/86400.;
  variable t = time;

  %%Corrections for eccentricity 0<=ecc<1
  variable omega = omega_d * PI/180.0;
  variable sinw = sin(omega);
  variable cosw = cos(omega);
  variable sq = sqrt((1.-ecc)*(1.+ecc));
  variable cor =2.*limit;

  %%start with number of iterations = zero
  variable numiter=0;
  while((abs(max(cor)) > limit) and (numiter < maxiter))
  {
    variable tper = (t-t0)/porb;
    variable m = 2*PI*(tper*(1.-0.5*pporb*tper));
    variable eanom = KeplerEquation(m,ecc);
    variable sin_e = sin(eanom);
    variable cos_e = cos(eanom);
    variable z = asini_d*(sinw*(cos_e-ecc)+sq*cosw*sin_e);
    variable f = (t-time)+z;
    variable df = (sq*cosw*cos_e - sinw*sin_e)*(2*PI*asini_d/(porb*(double(1.0)-ecc*cos_e)));
    cor = f/(double(1)+df);
    t -= cor;
    numiter++;
  }

  if(numiter >= maxiter)
    message("Exceeded maxiter iterations and did not reach convergence");
  return t;
}

%
% Code from src/slang/astronomy/BinaryPos.sl
%
%%%%%%%%%%%%%%%%%%%%%
define BinaryPos()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{BinaryPos}
%\synopsis{computes the position of a star in a binary system}
%\usage{Double_Type[] BinaryPos(Double_Type[] time)}
%\qualifiers{
%\qualifier{porb}{Orbital period at the epoch [same units as time argument], mandatory}
%\qualifier{eccentricity}{Eccentricity, (0<=e<1)}
%\qualifier{omega}{Longitude of periastron [degrees], mandatory}
%\qualifier{asini}{Projected semi-major axis (a sin i), mandatory for spectroscopic binaries, 
%                    give it in light seconds (a sini/c) if you want to compute a time correction}
%\qualifier{semi}{semimajor axis, mandatory for visual binaries}
%\qualifier{bigomega}{Longitude of the ascending node [degrees], mandatory for visual binaries}
%\qualifier{incl}{inclination [degrees], mandatory for visual binaries}
%\qualifier{t0}{epoch for mean longitude of 0 degrees (periastron, MJD)}
%\qualifier{t90}{epoch for mean longitude of 90 degrees (MJD)}
%\qualifier{pporb}{rate of change of the orbital period (s/s) (default 0)}
%}
%\description
%    (transcribed from IDL-programm BinaryPos.pro)
%
%    For spectroscopic binaries, the z-position of the emitting object
%    is computed (negative: closer to Earth). In case of visual binaries
%    the output is a 3d array containing the xyz position of the object
%    in a right handed coordinate system where the z axis points away from
%    Earth, the x axis points towards the North Celestial Pole, and where
%    the y-axis increases towards East.
%
%    Follows equations from Hilditch's book.
%
%    Qualifiers t90 and t0 have to be in days and in the same time
%    system as time (e.g. JD or MJD)
%
%    Circular orbits:
%         * if time of lower conjunction Tlow is known, set
%           t0=Tlow and omega=0
%         * if time of ascending node is known, Tasc, set
%           t90=Tasc and omega=0
%         * if time of mid eclipse is known, Tecl, set
%           t0=Tecl-0.25*porb and omega=0
%!%-
{
  switch(_NARGS)
  { case 1: variable t = (); }
  { help(_function_name()); return; }
  % make sure that time is an array
  t = [t];
 
  if (not qualifier_exists("porb") or not qualifier_exists("eccentricity")
      or (not qualifier_exists("t90") and not qualifier_exists("t0")))
     throw RunTimeError, "Need porb, eccentricity, and omega values, and one of t90 or t0";

  if (qualifier_exists("asini") + qualifier_exists("semi") != 1)
     throw RunTimeError, "Need exactly one of asini and semi";

  if (qualifier_exists("semi"))
  {
     if (not qualifier_exists("bigomega") or not qualifier_exists("incl"))
       throw RunTimeError, "Visual binaries: Need bigomega and incl";
  }

  if (qualifier_exists("t90") and qualifier_exists("t0"))
     throw RunTimeError, "Only one of the t90 and t0 arguments is allowed";

  if (qualifier("eccentricity",0.0) != 0.0 and not qualifier_exists("omega"))
     throw RunTimeError, "Need omega keyword for noncircular orbits";

  if (qualifier("eccentricity",0.0) < 0.0) 
     throw RunTimeError, "Eccentricity must be positive";

  if (qualifier("eccentricity",0.0) >= 1.0) 
     throw RunTimeError, "Binarypos only works for 0<=ecc<1";

  % qualifiers
  variable porb = qualifier("porb");
  variable ecc = qualifier("eccentricity",0.0);
  variable omega = qualifier("omega");
  variable asini = qualifier("asini");
  variable semi = qualifier("semi");
  variable bigomega = qualifier("bigomega");
  variable incl = qualifier("incl");
  variable t90 = qualifier("t90");
  variable pporb = qualifier("pporb",0.0);
  % compute t0 from t90 if necessary
  % note that this assumes pporb<<porb, which can be generally assumed
  variable t0 = qualifier("t0");
  if (t0 == NULL) t0 = qualifier("t0", t90+(omega-90.)/360. * porb);

  variable theta, pos;
  variable tper = (t-t0) / porb;
  variable m = 2.0 * PI * (tper * (1.0 - 0.5 * pporb * tper));
  variable eanom = KeplerEquation(m,ecc);

  % spectroscopic binaries
  if (asini != NULL)
  {
    omega = omega * PI / 180.0;
    variable sinw = sin(omega);
    variable cosw = cos(omega);
    variable sq = sqrt((1.0 - ecc)*(1.0 + ecc));
    variable sin_e, cos_e;
    
%%    pos = Double_Type[t];

    sin_e = sin(eanom);
    cos_e = cos(eanom);

    pos = asini * (sinw * (cos_e - ecc) + sq * cosw * sin_e);
%%    theta = 2.0 * atan(sqrt ((1.0 + ecc) / (1.0 - ecc)) * tan(eanom / 2.0));
%%    pos = asini * (1.0 - ecc^2.0) * sin(theta + omega) / (1.0 + ecc * cos(theta));
  }
  % visual binaries
  else
  {
    bigomega = bigomega * PI / 180.0;
    incl = incl * PI / 180.0;
    variable r, sin_to, cos_to;
    variable sin_o = sin(bigomega);
    variable cos_o = cos(bigomega);
    variable sin_i = sin(incl);
    variable cos_i = cos(incl);

    pos = Double_Type[3,length(t)];

    % true anomaly
    theta = 2.0 * atan(sqrt ((1.0 + ecc) / (1.0 - ecc)) * tan(eanom / 2.0));
    % radius vector
    r = semi * (1.0 - ecc * cos(eanom));
    % helpers
    sin_to = sin(theta + omega);
    cos_to = cos(theta + omega);
    % Hilditch, eq. 2.61
    pos[0,*] = r * (cos_o * cos_to - sin_o * sin_to * cos_i);
    pos[1,*] = r * (sin_o * cos_to + cos_o * sin_to * cos_i);
    pos[2,*] = r * sin_to * sin_i;
  }

  return pos;
}

%
% Code from src/plot/BoundingBox.sl
%
define BoundingBox()
%!%+
%\function{BoundingBox}
%\synopsis{retrieves the Bounding Box of a postscript file}
%\usage{Double_Type (x1, y1, x2, y2) = BoundingBox(String_Type psfile);
%\altusage{Double_Type (w, h) = BoundingBox(String_Type psfile; size);}
%}
%\description
%    \code{x1, y1, x2, y2} are the values from the last line in \code{psfile}
%    of the form \code{"%%BoundingBox: x1 y1 x2 y2"}.
%    \code{x1, y1, x2, y2} are \code{NULL} if no such line is found.
%\qualifiers{
%\qualifier{first}{Take the first matching %%BoundingBox line instead of the last.}
%\qualifier{size}{If this qualifier is set, \code{w = x2-x1} and \code{h = y2-y1} are returned.}
%}
%!%-
{
  variable psfile;
  switch(_NARGS)
  { case 1: psfile = (); }
  { help(_function_name()); return; }

  variable x1=NULL, y1=NULL, x2=NULL, y2=NULL;
  variable take_first = qualifier_exists("first");
  variable F = fopen(psfile, "r");
  while(not feof(F))
  {
    variable line = fgetslines(F, 1);
    if(line!=NULL && length(line)>0 && sscanf(line[0], "%%%%BoundingBox: %f %f %f %f", &$1, &$2, &$3, &$4)==4)
    {
      (x1, y1, x2, y2) = ($1, $2, $3, $4);
      if(take_first)  break;
    }
  }
  ()=fclose(F);
  if(qualifier_exists("size"))
    return (x2-x1, y2-y1);
  else
    return (x1, y1, x2, y2);
}

%
% Code from src/slang/arrays/CCF_1d.sl
%
%%%%%%%%%%%%%
define CCF_1d()
%%%%%%%%%%%%%
%!%+
%\function{CCF_1d}
%\synopsis{calculates the cross correlation function of two (1d) arrays}
%\usage{Double_Type CCF = CCF_1d(a1, a2, Integer_Type dx);
%\altusage{Array_Type CCF = CCF_1d(a1, a2);}
%}
%\qualifiers{
%\qualifier{notperiodic}{arrays are not extended periodically.}
%}
%\description
%    The arrays \code{a1} and \code{a2} have to have the same length \code{n}.
%    They can be passed directly or as references to arrays.
%
%    By default (unless the \code{notperiodic} qualifier is given), both arrays
%    are considered to be extended with periodic boundary conditions.
%    When using \code{notperiodic}, chose \code{|dx| < n-1}, since the larger \code{|dx|},
%    the smaller are the parts of the arrays that can actually be compared.
%
%    This function computes the correlation of \code{a1[x]} and \code{a2[x-dx]}:
%       \code{CCF = sum_x { (a1[x]-<a1>)/|a1| * (a2[x-dx]-<a2>)/|a2| }}
%    For periodic boundary conditions, the entire array is considered,
%    else the index \code{x} takes only values such that \code{x} and \code{x-dx} are contained.
%    The norm \code{|a| = sqrt{ sum_x a[x]^2 }} ensures that for any array \code{a},
%       \code{CCF_1d(a, a, 0) = +1}   and   \code{CCF_1d(a, -a, 0) = -1}.\n
%    \n
%    If only the two arrays \code{a1} and \code{a2} are given, the function returns
%    an array with all possible correlations between them, i.e.:
%    \code{[ CCF_1d(a1, a2, 0), CCF_1d(a1, a2, 1), ..., CCF_1d(a1, a2, n-1) ]}
%    This is most reasonable for periodic boundary conditions.
%\notes
%    The function is vectorized on \code{dx}, i.e., can handle an array \code{dx}.
%\example
%    variable a1 = Double_Type[n];  % some data
%    variable a2 = Double_Type[n];  % some other data
%
%    variable dx = [-n/2 : n/2];
%    variable CCF_np = CCF_1d(a1, a2, dx; notperiodic);
%
%    variable CCF_p = CCF_1d(a1, a2);  % This is equivalent to:
%    variable CCF_p = array_map(Double_Type, &CCF_1d, &a1, &a2, [0:n-1]);
%\seealso{CCF_2d}
%!%-
{
   variable arr1, arr2, dxs = NULL, periodic = not qualifier_exists("notperiodic");
   switch(_NARGS)
   { case 3: (arr1, arr2, dxs) = (); }
   { case 2: (arr1, arr2) = (); }
   { return help(_function_name()); }

   if (typeof(arr1) == Ref_Type)  arr1 = @arr1;
   if (typeof(arr2) == Ref_Type)  arr2 = @arr2;
   variable n = length(arr1);
   if (length(arr2) != n)  throw InvalidParmError;

   dxs = dxs == NULL ? [0 : n-1] : int(dxs);
   if (periodic)  dxs = ((dxs mod n) + n) mod n;  % => 0 <= dxs < n

   variable X1 = [0 : n-1];
   variable res = Double_Type[length(dxs)];
   variable dx, i = 0;
   foreach dx (dxs) {
      ifnot (periodic) {
         X1 = [int(_max(0,dx)) : n+int(_min(0,dx))-1];
      }
      variable a1 = arr1[X1   ];  a1 -= mean(a1);  a1 /= sqrt(sumsq(a1));
      variable a2 = arr2[X1-dx];  a2 -= mean(a2);  a2 /= sqrt(sumsq(a2));
      res[i] = sum(a1*a2);
      i++;
   }

   return typeof(dxs) != Array_Type ? res[0] : res;
}

%
% Code from src/slang/arrays/2d/CCF_2d.sl
%
define CCF_2d()
%!%+
%\function{CCF_2d}
%\synopsis{calculates the cross correlation function of two 2d-arrays (e.g., images)}
%\usage{Double_Type CCF = CCF_2d(img1, img2, Integer_Type dx, Integer_Type dy);}
%\description
%     The 2d-arrays img1 and img2 from which the CCF is to be computed can
%     either be passed directly or by a reference to the arrays.
%     This function computes the correlation of \code{img1[y, x]} and \code{img2[y-dy, x-dx]},
%     where the indices (x,y) take all values such that (x,y) and (x-dx,y-dy) are contained:
%        \code{CCF = sum_{x,y} { (img1[y,x]-<img1>)/|img1| * (img2[y-dy,x-dx]-<img2>)/|img2| }}
%     The norm \code{|img| = sqrt{ sum_{x,y} img[y,x]^2 }} ensures that \code{CCF_2d(img, +-img, 0, 0) = +-1}.
%     for any image \code{img}.\n
%     If \code{dx} or \code{dy} are arrays, \code{CCF} will be a 2d array.
%\qualifiers{
%\qualifier{verbose}{turns on verbosity}
%\qualifier{savebest}{=&best: if \code{dx} or \code{dy} are arrays, the best combination will be saved in \code{best}.}
%}
%\seealso{CCF_1d}
%!%-
{
  variable img1, img2, dx, dy;
  switch(_NARGS)
  { case 4: (img1, img2, dx, dy) = (); }
  { help(_function_name()); return; }

%  if(typeof(img1)==Ref_Type) { img1 = @img1; };
%  if(typeof(img2)==Ref_Type) { img2 = @img2; };
  if(typeof(dx)==Array_Type || typeof(dy)==Array_Type)
  {
    dx = [dx];
    dy = [dy];
    variable CCF_best = 0;
    variable dx_best = NULL;
    variable dy_best = NULL;
    variable CCF = Double_Type[length(dy), length(dx)];
    variable idx, idy;
    _for idy (0, length(dy)-1, 1)
      _for idx (0, length(dx)-1, 1)
      {
        variable c = CCF_2d(img1, img2, dx[idx], dy[idy]);
	if(qualifier_exists("verbose"))
	  vmessage("dx=%d, dy=%d  =>  CCF = %.3f", int(dx[idx]), int(dy[idy]), c);
        if(abs(c)>abs(CCF_best))
	{ CCF_best = c;
	  dx_best = int(dx[idx]);
	  dy_best = int(dy[idy]);
	}
        CCF[idy, idx] = c;
      }
    vmessage("best: CCF(dx=%d, dy=%d) = %.3f", dx_best, dy_best, CCF_best);
    if(qualifier_exists("savebest"))
    { variable b = qualifier("savebest");
      @b = [dx_best, dy_best];
    }
    return CCF;
  }

  if(typeof(dx)!=Integer_Type)  dx = int(dx);
  if(typeof(dy)!=Integer_Type)  dy = int(dy);

  variable w = length(img1[0,*]);
  variable h = length(img1[*,0]);
  variable X1 = [int(_max(0,dx)) : w+int(_min(0,dx))-1];
  variable Y1 = [int(_max(0,dy)) : h+int(_min(0,dy))-1];

  img1 = img1[Y1   , X1   ];  img1 -= mean(img1); img1 /= sqrt(sum(img1^2));
  img2 = img2[Y1-dy, X1-dx];  img2 -= mean(img2); img2 /= sqrt(sum(img2^2));

  return sum(img1*img2);
}

%
% Code from src/slang/COPY.sl
%
% -*- mode: slang; mode: fold -*-

define COPY() %{{{
%!%+
%\function{COPY}
%\synopsis{makes a copy of a nested Struct/Array/List_Type}
%   \usage{Struct_Type copy = COPY(Struct_Type copyme);}
%\altusage{Array_Type  copy = COPY(Array_Type  copyme);}
%\altusage{List_Type   copy = COPY(List_Type   copyme);}
%\altusage{Assoc_Type  copy = COPY(Assoc_Type   copyme);}
%\description
%    \code{COPY} returns a properly dereferenced copy of an
%    arbitrarily nested \code{Struct/Array or List_Type} with
%    all its entries. 
%    Depending on the Data_Type of the initial given
%    copyme and the Data_Type of its entries, \code{COPY}
%    iteratively calls the  according sub-function
%    (\code{struct_copy}, \code{array_copy} or \code{list_copy}).
%    If the given Data_Type or one of its entries is not
%    one of previously mentioned ones, \code{COPY} returns
%    @(Data_Type) or respectively @(entry).
%    
%    NOTE: * Ref_Type entries will not be dereferenced, which
%            allows to copy XFIG-OBJECTS !!!
%          * Double_Type[] is an Array_Type
%
%\example
%    Array_Type:
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9];
%     copy = COPY(s); copy[0].a[[0]] = ["modified"];
%     print(s[0].a);
%
%    List_Type:
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9];
%     l = [{ array_copy(s), 1., [1:10] }, {"abs"}];
%     copy = COPY(l); copy[0][0][0].a[[0]] = ["modified"];
%     print(l[0][0][0].a);
%
%    Struct_Type:
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9]; S = struct{ f=s };     
%     C = struct_copy(S); C.f[0].a[[0]] = ["modified"];
%     print(S.f[0].a);
%
%\seealso{struct_copy, array_copy, list_copy, assoc_copy}
%!%-
{
  variable copyme;
  switch(_NARGS)
  { case 1: copyme = (); }
  { help(_function_name()); return; }

  variable type = typeof(copyme);
  if( type  == Struct_Type )
    return struct_copy(copyme);
  else if( type == Array_Type )
    return array_copy(copyme);
  else if( type == List_Type )
    return array_copy(copyme);
  else if( type == Assoc_Type )
    return assoc_copy(copyme);
  else{
    vmessage("WARNING: <%s>: Unsupported Data_Type=%S, retruning @copyme",
	     _function_name,type);
    return @copyme;
  }
}
%}}}

define struct_copy() %{{{
%!%+
%\function{struct_copy}
%\synopsis{makes a copy of a struct with all its fields}
%\usage{Struct_Type copy = struct_copy(Struct_Type struct);}
%\description
%    \code{struct_copy} copies a \code{Struct_Type} with all its
%    fields, which can be of any Type! If a field is an Array_Type[]
%    \code{struct_copy} calls \code{array_copy} and if a field is
%    \code{Struct_Type}, \code{struct_copy} calls itself.
%    Further it copies Ref_Type, which allows copying XFIG-Objects.
%\example
%    s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%    s[0].a[[0]]=[0:9]; S = struct{ f=s };     
%    C = struct_copy(S); C.f[0].a[[0]] = ["modified"];
%    print(S.f[0].a);
%\seealso{array_copy, list_copy, COPY, assoc_copy}
%!%-
{
  variable table;
  switch(_NARGS)
  { case 1: table = (); }
  { help(_function_name()); return; }
  
  variable f, fieldnames = get_struct_field_names(table);
  variable field , ftype;
  variable copy = @Struct_Type(fieldnames);
  foreach f (fieldnames){
    field = get_struct_field(table, f);
    ftype = typeof(field);
    if( ftype == Struct_Type )
      set_struct_field(copy, f, struct_copy(field) );
    else if( ftype == Array_Type )
      set_struct_field(copy, f, array_copy(field) );
    else if( ftype == List_Type )
      set_struct_field(copy, f, list_copy(field) );
    else if( ftype == Assoc_Type )
      set_struct_field(copy, f, assoc_copy(field) );
    else if ( (ftype == String_Type) ||  (ftype == Ref_Type) )
      set_struct_field(copy, f, field );	
    else
      set_struct_field(copy, f, @field );
  }
  return copy;
}
%}}}
define array_copy() %{{{
%!%+
%\function{array_copy}
%\synopsis{makes a copy of a nested array}
%\usage{Array_Type copy = array_copy(Array_Type[] array);}
%\description
%    \code{array_copy} copies an \code{Array_Type[]} with all its
%    entries, which can be of any Type! If an entry is a Struct_Type
%    \code{array_copy} calls \code{struct_copy} and if an entrie is
%    an \code{Array_Type[]}, \code{array_copy} calls itself.
%\example
%    s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%    s[0].a[[0]]=[0:9];
%    copy = array_copy(s); copy[0].a[[0]] = ["modified"];
%    print(s[0].a);
%\seealso{struct_copy}
%!%-
{
  variable array;
  switch(_NARGS)
  { case 1: array = (); }
  { help(_function_name()); return; }
  
  variable type = typeof(array);
  ifnot( type == Array_Type ){
    if( type == Struct_Type )
      return struct_copy( array );
    if( type == List_Type )
      return list_copy( array );
    else if( type == Ref_Type || type == String_Type )
      return array;
    else
      return @array;
  }
  
  variable len  = length(array);
  variable dim  = array_shape(array);
  type = _typeof(array);
  variable copy = type[len];
  reshape( copy, dim );
  
  variable i;
  _for i ( 0, len-1, 1 ){
    copy[[i]] = array_copy( array[[i]][0] );
  }
  return copy;
}
%}}}

define list_copy() %{{{
%!%+
%\function{list_copy}
%\synopsis{makes a copy of a nested list}
%\usage{List_Type copy = list_copy(List_Type list);}
%\description
%    \code{list_copy} returns a properly dereferenced copy of an
%    arbitrarily nested \code{List_Type} with all its entries. 
%    Depending on the Data_Type of the entries, \code{list_copy}
%    iteratively calls the according sub-function
%    (\code{struct_copy}, \code{array_copy} or \code{list_copy}).
%    If the Data_Type of one of its entries is not
%    one of privouse mentioned ones, \code{list_copy} returns
%    @(Data_Type) or respectively @(entry).
%\example
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9];
%     l = [{ array_copy(s), 1., [1:10] }, {"abs"}];
%     copy = COPY(l); copy[0][0][0].a[[0]] = ["modified"];
%     print(l[0][0][0].a);
%\seealso{COPY, struct_copy, array_copy, assoc_copy}
%!%-
{
  variable list;
  switch(_NARGS)
  { case 1: list = (); }
  { help(_function_name()); return; }
  
  variable type = typeof(list);
  ifnot( type == List_Type ){
    if( type == Struct_Type )
      return struct_copy( list );
    else if( type == Array_Type )
      return array_copy( list );
    else if( type == Assoc_Type )
      return assoc_copy( list );
    else if( type == Ref_Type || type == String_Type )
      return list;
    else
      return @list;
  }
  
  variable len  = length(list);
  variable copy = {};
  
  variable i;
  _for i ( 0, len-1, 1 ){
    list_append( copy, list_copy(list[i]) );
  }
  return copy;
}
%}}}

define table_copy() %{{{
%!%+
%\function{table_copy}
%\synopsis{makes a copy of all columns of a table}
%\usage{Struct_Type copy = table_copy(Struct_Type table);}
%\description
%    \code{copy.field = @(table.field);  %} for every \code{field}
%    NOTE: This also works for XFIG-PLOT-OBJECTS
%!%-
{
  variable table;
  switch(_NARGS)
  { case 1: table = (); }
  { help(_function_name()); return; }
  
  vmessage("<%s> is deprecated! Use <struct_copy> instead!",_function_name);
  return struct_copy(table);
}
%}}}

define assoc_copy()
%!%+
%\function{assoc_copy}
%\synopsis{makes a copy of an associated array}
%\usage{Assoc_Type copy = assoc_copy(Assoc_Type list);}
%\description
%    \code{assoc_copy} returns a properly dereferenced copy of an
%    arbitrarily nested \code{Assoc_Type} with all its entries. 
%    Depending on the Data_Type of the entries, \code{assoc_copy}
%    iteratively calls the according sub-function
%    (\code{struct_copy}, \code{array_copy} or \code{list_copy}).
%    If the Data_Type of one of its entries is not
%    one of previously mentioned ones, \code{assoc_copy} returns
%    @(Data_Type) or respectively @(entry).
%\example
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9];
%     l = [{ array_copy(s), 1., [1:10] }, {"abs"}];
%     copy = COPY(l); copy[0][0][0].a[[0]] = ["modified"];
%     print(l[0][0][0].a);
%\seealso{COPY, struct_copy, array_copy, list_copy}
%!%-
{
    variable copyme;
    switch(_NARGS)
    { case 1: copyme = (); }
    { help(_function_name()); return; }

    variable new=Assoc_Type[];
    
    variable s=assoc_get_values(copyme);
    variable key;
    foreach key ( s ) {
        new[key]=COPY(s[key]);
    }
    return new;
}




%
% Code from src/misc/satellites/Chandra/Chandra_display_mask.sl
%
define Chandra_display_mask()
%!%+
%\function{Chandra_display_mask}
%\synopsis{shows the content of a Chandra mask file (msk1.fits)}
%\usage{Chandra_display_mask([String_Type filename]);}
%\description
%    \code{filename} can be a globbing expression.
%    If it is omitted, \code{acis*msk1.fits*} is assumed.
%
%    The mask file contains the valid part of the CCD,
%    i.e., the portion for which events can be telemetered.
%!%-
{
  variable pattern = "acis*msk1.fits*";
  switch(_NARGS)
  { case 0: ; }
  { case 1: pattern = (); }
  { return help(_function_name()); }

  variable files = glob(pattern);
  if(length(files)==0)
    return vmessage("error (%s): no mask file %s found", _function_name(), pattern);

  variable filename;
  foreach filename (files[array_sort(files)])
  {
    if(filename!=pattern)
      vmessage("\nisis> %s(\"%s\");", _function_name(), filename);
    variable F = fits_open_file(files[0], "r");
    while(_fits_movrel_hdu(F, 1)==0)
    {
      variable mask = fits_read_table(F);
      variable i;
      vmessage("\n%s:", fits_read_key(F, "HDUNAME"));
      _for i (0, length(mask.component)-1, 1)
	vmessage("#%d: %10s, CHIPX x CHIPY = [%4d:%4d] x [%4d:%4d],  samp_cyc = %2d,  PHA = [%d:%d]",
		 mask.component[i], mask.shape[i], mask.chipx[i,0], mask.chipx[i,1], mask.chipy[i,0], mask.chipy[i,1],
		 mask.samp_cyc[i], mask.phamin[i], mask.phamax[i]);
    }
    fits_close_file(F);
    message("");
  }
}

%
% Code from src/slang/astronomy/Compton_crosssection.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define Compton_crosssection()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Compton_crosssection}
%\usage{Double_Type sigma = Compton_crosssection(Double_Type E);}
%\description
%    \code{E} is the energy in keV.
%    \code{sigma} is the total Klein-Nishina crosssection in cm^2:
%    \code{sigma = 3/4 sigma[Th] * [  (1+y)/y^3 * { 2y*(1+y)/(1+2y) - log(1+2*y) }  +  log(1+2*y)/(2*y)  +  (1+3y)/(1+2y)^2  ]}
%    where \code{y = E/(511 keV)}.
%\seealso{Rybicki & Lightman (1979)}
%!%-
{
  variable E;
  switch(_NARGS)
  { case 1: E = (); }
  { help(_function_name()); return; }

  variable y = E/510.998903;
  variable twoy = 2*y;
  variable onep2y = 1. + twoy;
  variable log1p2y =log1p(twoy);
  return 3/4. * 6.6524586e-25 * ( (1+y)/y^3*( twoy*(1+y)/onep2y - log1p2y ) + log1p2y/twoy - (onep2y+y)/onep2y^2 );
}

%
% Code from src/slang/date/DateOfJD.sl
%
define DateOfJD() {
%!%+
%\function{DateOfJD}
%\synopsis{Calculates the date corresponding to a JD value}
%\usage{Struct_Type date = DateOfJD(JD);}
%\description
%    The return value is a structure of the following form:\n
%       \code{date = struct { year, month, day, hour, minute, second };}
%    The algorithm used is that by Meeus, Astronomical Formulae for
%    Calculators. This routine is array safe.
%\qualifiers{
%\qualifier{julianswitch}{For JD smaller than this date, return the date
%        in the Julian calendar rather than in the Gregorian
%        calendar. The default is JD2299161 (15 October 1582),
%        which is appropriate for most of catholic Europe.
%        Note, however, that there are countries that switched
%        only much later (see https://en.wikipedia.org/wiki/Julian_calendar
%        and https://de.wikipedia.org/wiki/Gregorianischer_Kalender).    
%        For example, Russia only switched on 1 February 1918 (Gregorian).}
%\qualifier{julian_calendar}{force returning the date in the Julian Calendar
%        (ignore julianswitch)}
%\qualifier{gregorian_calendar}{force returning the date in the Gregorian Calendar
%        (ignore julianswitch)}
%}
%\seealso{JDofDate, MJDofDate}
%!%-
    variable JD;
    switch(_NARGS)
    { case 1: JD=(); }
    { help(_function_name()); return;}

    variable ndx=where(JD<0.);
    if (length(ndx)>0) {
	throw UsageError,sprintf("%s: only positive JD are allowed\n",_function_name());
    }

    variable retval=struct{year,month,day,hour,minute,second};

    if (qualifier_exists("julian_calendar") or qualifier_exists("gregorian_calendar")) {
	if (qualifier_exists("julianswitch")) {
	    throw UsageError,sprintf("%s: julianswitch and julian_calendar or gregorian_calendar qualifiers are not allowed at the same time\n",_function_name());
	}
    }

    if (qualifier_exists("julian_calendar") and qualifier_exists("gregorian_calendar")) {
	    throw UsageError,sprintf("%s: julian_calendar or gregorian_calendar qualifiers are mutually exclusive\n",_function_name());
    }

    variable jswitch=qualifier("julianswitch",2299161);
    if (qualifier_exists("gregorian_calendar")) {
	jswitch=-1.; % force all dates to gregorian (since JD<0 is not allowed)
    }
    if (qualifier_exists("julian_calendar")) {
	jswitch=1.+max(JD); % force all dates to julian
    }
    
    
    %
    % implement the Algorithm of Meeus
    % Note that this algorithm depends crucially on the peculiarities
    % of floating point arithmetics vs. integer arithmetics, so please
    % change the following lines only if you REALLY know what you are
    % doing and have a copy of Meeus next to you.
    %
    variable t=JD+0.5;
    variable z=int(t);
    variable f=t-z;
    variable a;
    variable x;

    if (length(JD)>1) {
	variable ndx1,ndx2;
	ndx1=where(JD<jswitch,&ndx2);
	a=Double_Type[length(JD)];
	if (length(ndx1)>0) {
	    a[ndx1]=z[ndx1];
	}
	if (length(ndx2)>0) {
	    x=int((z[ndx2]-1867216.25)/36524.25);
	    a[ndx2]=z[ndx2]+1.+x-int(x/4.);
	}
    } else {
	if (JD<jswitch) {
	    a=z;
	} else {
	    x=int((z-1867216.25)/36524.25);
	    a=z+1.+x-int(x/4.);
	}
    }
    variable b=a+1524.;
    variable c=int((b-122.1)/365.25);
    variable d=int(365.25*c);
    variable e=int((b-d)/30.6001);
    retval.day=int(b-d-int(30.6001*e));
    if (length(JD)>1) {
	retval.month=Integer_Type[length(JD)];
	retval.year=Integer_Type[length(JD)];
	ndx1=where(e<13.5,&ndx2);
	if (length(ndx1)>0) {
	    retval.month[ndx1]=int(e[ndx1]-1.);
	}
	if (length(ndx2)>0) {
	    retval.month[ndx2]=int(e[ndx2]-13.);
	}
	retval.year=int(c-4715);
	ndx1=where(retval.month>=3);
	if (length(ndx1)>0) {
	    retval.year[ndx1]=retval.year[ndx1]-1;
	}
    } else {
	retval.month=int ( (e<13.5) ? e-1. : e-13. );
	retval.year=int ( (retval.month<3) ? c-4715. : c-4716.);
    }
    f=f*24.;
    retval.hour=int(f);
    f=(f-retval.hour)*60.;
    retval.minute=int(f);
    retval.second=(f-retval.minute)*60.;
    return retval;
}

%
% Code from src/slang/date/EasterSunday.sl
%
define EasterSunday()
%!%+
%\function{EasterSunday}
%\synopsis{Calculate the JD or date of Easter Sunday for the Gregorian or Julian calendar}
%\usage{jd=EasterSunday(year)}
%\description
%This function calculates the JD of Easter Sunday for the Gregorian or
%Julian (=orthodox) calendar using the algorithms given by J. Meeus, 1998,
%Astronomical Algorithms, 2nd ed, William-Bell.
%
%The function either returns the JD (more precise: the JD at midnight
%of Easter Sunday), or a struct with the same format as DateOfJD.
%
%This function is array safe, i.e., an array of years can be given.
%
%\qualifiers{
%\qualifier{mjd}{Return dates in  MJD, not in JD.}
%\qualifier{orthodox}{Return the JD of Easter Sunday using the rules
%                  appropriate for Eastern Christianity such as the
%                  Russian Orthodox church.}
%\qualifier{get_date}{Return date as a struct (year, month, date, hour,
%                  minute, second). If the orthodox keyword is given,
%                  the date will be in the Julian calendar, otherwise
%                  it will be in the Gregorian calendar.}
%\qualifier{backward}{An undocumented function with the same name
%                  as this one existed in the isisscripts until
%                  17 January 2020. It returned the date of Easter Sunday
%                  in the Gregorian calendar as the number of days
%                  since -1 March (sic!) of the given year. This
%                  qualifier switches the current function to this
%                  backward compatible behavior.}
%}
%\seealso{DateOfJD}
%!%-
{
    variable year=();

    variable x=int(year);

    variable a,b,c,d,e,f,g,h,i,k,l,m,n,p,t,tt;
    
    if (qualifier_exists("orthodox")) {
	% Meeus, p.69
	a=x mod 4;
	b=x mod 7;
	c=x mod 19;
	d=(19*c+15) mod 30;
	e=(2*a+4*b-d+34) mod 7;
	tt=(d+e+114);
	n=tt/31; % month
	p=1+(tt mod 31); % day
    } else {
	% Meeus, p. 67
	a=x mod 19;
	b=x/100;
	c=x mod 100;
	d=b/4;
	e=b mod 4;
	f=(b+8)/25;
	g=(b-f+1)/3;
	h=(19*a+b-d-g+15) mod 30;
	i=c/4;
	k=c mod 4;
	l=(32+2*e+2*i-h-k) mod 7;
	m=(a+11*h+22*l)/451;
	tt=h+l-7*m+114;
	n=tt/31;    % month
	p=1+(tt mod 31); % day
	if (qualifier_exists("backward")) {
	    variable jd=JDofDate(x,n,p);
	    variable jd0=JDofDate(x,3,-1);
	    return jd-jd0;
	}
    }

    if (qualifier_exists("get_date")) {
	if (typeof(x)==Integer_Type) {
	    return struct{year=x,month=n,day=p,hour=0,minute=0,second=0.};
	} else {
	    variable hh=Integer_Type[length(x)];
	    variable mm=Integer_Type[length(x)];
	    variable ss=Double_Type[length(x)];
	    return struct{year=x,month=n,day=p,hour=hh,minute=mm,second=ss};
	}
    }

    if (qualifier_exists("orthodox")) {
	if (qualifier_exists("mjd")) {
	    return JDofDate(x,n,p;julian_calendar,mjd);
	}
	return JDofDate(x,n,p;julian_calendar);
    }

    if (qualifier_exists("mjd")) {
	return JDofDate(x,n,p;mjd);
    }
    return JDofDate(x,n,p);
}

%
% Code from src/misc/astronomy/EchoMap.sl
%

% only activate if module vector could be loaded
#ifeval __get_reference("vector")!=NULL

% mirror-like redistribution function
private define EchoMap_redist(power, mu, surf, pcase) { return power*mu; }
% mirror-like reponse function
private define EchoMap_respon() { return struct { time = [0.], power = [1.] }; }

%%%%%%%%%%%%%%%%%%%%%
define EchoMap_makelc()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{EchoMap_makelc}
%\synopsis{takes reprocessing information to produce a response signal}
%\usage{Struct_Type EchoMap_makelc(
%      Struct_Type reproment, Struct_Type signal[, Double_Type length]
%    );}
%\qualifiers{
%    \qualifier{noDoppler}{disable Doppler Shift}
%    \qualifier{response}{reference to a function returning the
%                relative emissivity of a surface element
%                as response to an incoming power peak at
%                t = 0. For details see the description
%                (default: delta peak at t=0)}
%    \qualifier{surfsigs}{reference to a variable, which will get the
%                output signal for each element (2dim-array
%                defined as [element,power]; use the time
%                from the returned total signal)}
%}
%\description
%    Takes the 'reproment' structure of a former run of
%    EchoMap and returnes the output signal produced by
%    reprocessing the input 'signal' on the surface used
%    to caluclate the reproment (i.e., the geometrical
%    effect of the reprocessing). By default, the returned
%    power vs. time is the total response to the incomming
%    signal. However, if the optional argument 'length' is
%    provided, the input signal is periodically repeated
%    until the output signal has at least the given length.
%    Furthermore, the signal is repeated backwards in time
%    as well such that no rising or declining parts of the
%    echo is visible in the output.
%
%    By default, it is assumed that each surface element
%    mirrors the incoming power instantaneously. More complex
%    behaviour can be implemented easily using the 'response'
%    qualifier, which defines a function returning the
%    relative 'power' emissivity over 'time' (in seconds) of
%    a surface element. Furthermore, the response has to
%    fullfil energy conservation such that the integrated
%    power does not exceed 1, but may be less, i.e. power is
%    reprocessed via non-radiative processes. The time
%    resolution of the response has to be equal or better
%    than the input signal!
%\seealso{EchoMap}
%!%-
{
  variable repro, sig, T = NULL, cyc = 0;
  switch(_NARGS)
    { case 2: (repro, sig) = (); }
    { case 3: (repro, sig, T) = (); cyc = 1; }
    { help(_function_name()); return; }

  % parameter validity checks
  ifnot (typeof(sig) == Struct_Type) { vmessage("error (%s): 'signal' must be of Struct_Type", _function_name); return; }
  ifnot (struct_field_exists(sig, "time") and struct_field_exists(sig, "power")) { vmessage("error (%s): 'signal' does not have the required fields", _function_name); return; }
  ifnot (length(sig.time) > 1 and length(sig.power > 1)) { vmessage("error (%s): 'signal' arrays have to have more than one element", _function_name); return; }
  ifnot (length(sig.time) == length(sig.power > 1)) { vmessage("error (%s): field lengths of 'signal' must be equal", _function_name); return; }
  ifnot (typeof(repro) == Struct_Type) { vmessage("error (%s): 'reproment' must be of Struct_Type", _function_name); return; }
  variable reprofields = ["visible_obs", "visible_sig", "lt_surf", "lt_obs", "doppler", "flux_rec", "flux_emi", "flux_int"];
  ifnot (sum(array_map(Integer_Type, &struct_field_exists, repro, reprofields)) == length(reprofields)) { vmessage("error (%s): 'reproment' does not have the required fields", _function_name); return; }
  
  % only use surface elements visible by observer
  repro = struct_filter(repro, where(repro.visible_obs); copy);

  % get response function
  variable response = @(qualifier("response", &EchoMap_respon))();
  ifnot (struct_field_exists(response, "time") and struct_field_exists(response, "power")) { vmessage("error (%s): 'response' does not have the required fields", _function_name); return; }
  % interpolate it to match the signal time resolution
  variable dt = (sig.time[-1]-sig.time[0]) / (length(sig.time)-1); % time resolution of input signal
  variable iresp = struct { time, power };
  if (length(response.time) > 1) {
    iresp.time = [response.time[0]:response.time[-1]:dt];
    iresp.power = interpol(iresp.time, response.time, response.power);
    iresp.power /= sum(iresp.power)/sum(response.power); % still the same amount of energy
    response = iresp; % caution: both variables point to the same structure
  }

  %%% calculate the response of an element in its co-moving frame
    % (still independent from any specific element)
  variable resig = COPY(sig), i, l = length(response.time);
  if (l == 0) { vmessage("error (%s): length of interpolated 'response' is zero", _function_name); return; } % should not happen...
  % if the response is delayed shift the time
  if (response.time[0] > 0) {
    resig.time += response.time[0];
    response.time -= response.time[0]; % acutally is not necessary since response.time is not used from now on
  }
  % in case of a one bin response just scale the power
  if (l == 1) { resig.power *= response.power[0]; }
  % otherwise fold the signal with the reponse and
  % increase time and power array in case of a none periodic signal
  else {
    ifnot (cyc) {
      resig.time = [resig.time, dt*[1:l-1] + resig.time[-1]];
      resig.power = Double_Type[length(resig.power) + l-1];
    }
    % fold
    _for i (0, length(sig.power)-1, 1) { % modulo only affects periodic signals here
      resig.power[i+[0:l-1] mod length(resig.power)] += sig.power[i] * response.power;
    }
  }

  %%% initialize output lightcurve
  % if a length of the output lightcurve is not given a non-periodic
  % signal is assumed -> calculate the required length (upper limit)
  ifnot (cyc) {
    T = (sig.time[-1] - sig.time[0] % length of the input signal
       + max(repro.lt_surf + repro.lt_obs) % largest time delay
       + response.time[-1] % length of the response function
	) * (qualifier_exists("noDoppler") ? 1. : max(repro.doppler)); % highest Doppler factor
  }
  % on the other hand, if it is a periodic signal the response signal
  % has to be repeated until its length matches at least the given length
  else {
    % let the periodic signal start as early as possible such that after
    % the transformation into the observing frame the first time bin < 0
    l = max(repro.lt_surf + repro.lt_obs) * max(repro.doppler); % upper limit for latest possible signal in output lightcurve
    i = sig.time[-1] - sig.time[0]; % length of the input signal
    resig.time -= ceil(l/i) * i; % move l/i signal loops back in time
    % copy signal periodically until its length matches the required one
    l = COPY(resig.power); % this copy will be inserted into the signal
    while (resig.time[-1] < T) {
      resig.time = [resig.time, resig.time[-1] + dt*[1:length(l)]];
      resig.power = [resig.power, l];
    }
  }

  %%% build echo -> the fun part
  variable outsig = struct { time = [0:T:dt] + sig.time[0], power };
           outsig.power = Double_Type[length(outsig.time)];
  variable surfsig, surfsigs = qualifier("surfsigs", NULL);
  if (surfsigs != NULL) { @surfsigs = Double_Type[length(repro.visible_obs), length(outsig.power)]; }
  % loop over visible surface
  _for l (0, length(repro.visible_obs)-1, 1) {
    % transform the time into the observing frame and apply flux projection
    % black-magic one-liner: interpolate power on the output time grid, but DON'T extrapolate
    if (repro.visible_sig[l]) {
      surfsig = interpol(outsig.time, % output time grid
			 (@(resig.time) + repro.lt_surf[l] + repro.lt_obs[l])       % add time delay and
			 * (qualifier_exists("noDoppler") ? 1. : repro.doppler[l]), % Doppler factor to co-moving frame -> observing frame
			 @(resig.power) * repro.flux_emi[l]; % apply flux projection on signal power
			 extrapolate = "none", null_value = 0.); % no flux outside of the time signal
    }
    % othwerwise a "dark" surface element if not visible from signal origin
    else { surfsig = Double_Type[length(outsig.time)]; }
    % add instrinsic power
    surfsig += repro.flux_int[l];
    % add to output lightcurve
    outsig.power += surfsig;
    if (surfsigs != NULL) { surfsigs[l,*] = surfsig; }
  }

  return outsig;
}


%%%%%%%%%%%%%%%%%%%%%
define EchoMap()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{EchoMap}
%\synopsis{reprocesses a signal on a given surface}
%\usage{Struct_Type EchoMap(
%      Struct_Type surface, Struct_Type signal
%      Vector_Type observer[, Double_Type length]
%    );}
%\qualifiers{
%    \qualifier{c}{defines the speed of light (default: 1)}
%    \qualifier{noFiniteC}{infinite speed of light}
%    \qualifier{noDoppler}{disable Doppler Shift}
%    \qualifier{powRedist}{reference to a function returning the re-
%                distributed power of all surface elements
%                (i.e., all parameters are arrays!) in
%                direction to the observer. The parameters
%                passed are
%                  * the power to be redistributed
%                  * the cosine of the angle between the
%                    element and the observer
%                  * the surface element as structure with
%                    the corresponding fields as below
%                  * wether the passed power is the surface's
%                    intrinsic one (1) or incident power (2)
%                  * a reference to a variable, which may be
%                    set to 1 to trigger re-calculating the
%                    powers (i.e. the given function is called
%                    again for the two cases *after* it was
%                    called for both cases in the first place;
%                    might be useful if, e.g., the incident
%                    power changes properties of the surface)
%                This function is called twice: (1) for the
%                intrinsic power of a surface element and
%                (2) the reprocessed relative (!!) power. The
%                last parameter specifies these two cases.
%                If you need to pass additional qualifiers
%                you may set the powRedistQual-qualifier
%                to a structure holding the qualifiers
%                (default: power*cosine_to_observer)}
%    \qualifier{response}{response function of a surface element,
%                see EchoMap_makelc for details}
%    \qualifier{reproment}{instead of returning the echo signal the
%                reproment structure is returned (see text)}
%}
%\description
%    This function propagates the given 'signal' on an object
%    defined by a 'surface'. There the signal is absorbed and
%    re-emitted (reprocessed) by each surface element. The
%    resulting response signal, as seen in direction to an
%    'observer', is returned by calling EchoMap_makelc after
%    all geometrical have been calculated (called reproment
%    here). If the corresponding qualifier is set, the latter
%    call is skipped and the reproment structure is returned
%    instead. This structure has the following fields:
%      visible_obs - boolean value if the surface is visible
%      visible_sig   by the observer and the signal origin
%      lt_surf - light travel times from the signal to the
%      lt_obs    surface and from there to the observing plane
%      doppler - Doppler factor in direction to the observer
%      flux_int - intrinsic flux of the surface as seen by
%                 the observer
%      flux_rec - received signal flux of the surface
%      flux_emi - emitted relative flux of the surface as seen
%                 by the observer
%
%    The following effects are taken into account during the
%    calculation of the response signal:
%    - light travel time between each surface element to the
%      signal source and to the observing plane
%    - Doppler shift by a moving surface
%    - effective areas and projection effects
%    - power redistribution function at the surface (qualifier
%      'powRedist')
%    - power response function of the surface (qualifier
%      'response', see EchoMap_makelc for details)
%
%    Restrictions:
%    - sqrt(surface.A) << vector_norm(surface.r-signal.origin)
%      (small surface elements relative to distance to source)
%    - surface.v << c
%    - vector_norm(observer) != 1
%    - signal fields have to be sorted by time
%                
%    all qualifiers are passed to EchoMap_makelc
%    
%    The 'surface' structure contains the surface elements,
%    defined by the fields
%      Vector_Type[] r - position vector to each element
%      Double_Type[] A - area of each element
%      Vector_Type[] n - normal vector of each element
%      Vector_Type[] v - velocity vector of each element
%                        (optional)
%      Double_Type[] L - intrinsic power of each element,
%                        which will be added to the response
%                        signal (optional, in energy/time)
%    The 'signal' is treated as power (energy/time) as a
%    function of time and defined by the fields
%      Double_Type[] time  - lower time bin (in seconds)
%      Double_Type[] power - power in each time bin
%      Vector_Type origin  - position vector to the signal's
%                            source (default: (0,0,0))
%    Any length is considered to be in units of the speed of
%    light (lt-s). If another unit is used, the internal speed
%    of light has to be changed via the 'c' qualifier.
%\seealso{EchoMap_makelc, EchoMap_binary, Vectory_Type}
%!%-
{
  variable surf, sig, obs, T = NULL;
  switch(_NARGS)
    { case 3: (surf, sig, obs) = (); }
    { case 4: (surf, sig, obs, T) = (); }
    { help(_function_name()); return; }

  % parameter validity checks
  ifnot (typeof(surf) == Struct_Type) { vmessage("error (%s): 'surface' must be of Struct_Type", _function_name); return; }
  ifnot (typeof(sig) == Struct_Type) { vmessage("error (%s): 'signal' must be of Struct_Type", _function_name); return; }
  ifnot (typeof(obs) == Vector_Type) { vmessage("error (%s): 'observer' must be of Vector_Type", _function_name); return; }
  ifnot (struct_field_exists(surf, "r") and struct_field_exists(surf, "n") and struct_field_exists(surf, "A")) { vmessage("error (%s): 'surface' does not have the required fields", _function_name); return; }
  ifnot (length(surf.r) == length(surf.A) and length(surf.r) == length(surf.n)) { vmessage("error (%s): field lengths of 'surface' must be equal", _function_name); return; }
  ifnot (struct_field_exists(sig, "time") and struct_field_exists(sig, "power")) { vmessage("error (%s): 'signal' does not have the required fields", _function_name); return; }
  ifnot (length(sig.time) > 1 and length(sig.power > 1)) { vmessage("error (%s): 'signal' arrays have to have more than one element", _function_name); return; }
  ifnot (length(sig.time) == length(sig.power > 1)) { vmessage("error (%s): field lengths of 'signal' must be equal", _function_name); return; }

  % input functions
  variable powerRedist = qualifier("powRedist", &EchoMap_redist);
  
  % some variable definitions
  normalize_vector(obs); % make sure vector to observer is normalized
  variable rsig = (struct_field_exists(sig, "origin") ? sig.origin : vector(0,0,0)); % vector to signal origin
  variable c = qualifier("c", 1.); % speed of light (default 1.0)
  if (qualifier_exists("noFiniteC")) c = _Inf;
  % initialize 'reproment' variable
  variable repro =  struct {
    visible_sig, % element visible from signal origin
    visible_obs, % element visible by observer
    lt_surf = Double_Type[length(surf.r)], % light travel time from signal origin to element
    lt_obs = Double_Type[length(surf.r)], % light travel time from element to observation plane
    doppler = Double_Type[length(surf.r)], % Doppler factor
    flux_rec = Double_Type[length(surf.r)], % received input flux by element
    flux_emi = Double_Type[length(surf.r)], % reprocessed flux by element to observer
    flux_int = Double_Type[length(surf.r)] % intrinsic flux emitted by element to observer
  };

  %%% calculate reproment (geometry parameters of the echo signal) for each surface element
    % - the light travel times (signal origin to element, element to observing plane
    % - Doppler factor
    % - received and emitted total power
  variable rsurfsig = surf.r - rsig; % vectors from signal origin to each surface element
  variable lsurfsig = array_map(Double_Type, &vector_norm, rsurfsig); % length of these vectors
  variable sigproj = -surf.n * rsurfsig / lsurfsig; % cosine of angle between surface elements and signal origin
  variable obsproj = surf.n * obs; % cosine of angle between surface elements and observer
  variable rsurfobs = rsurfsig * obs; % distance of surface to observing plane (at signal origin)
  % surface elements visible from signal origin and by observer
  repro.visible_sig = sigproj > 0;
  repro.visible_obs = obsproj > 0;
  % proceed with elements visible by observer only
  variable n = where(repro.visible_obs), m, redo = 0, r = 0, surfcopy;
  if (length(n) > 0) {
    % calculate light travel times
    repro.lt_surf[n] = lsurfsig[n]/c;
    repro.lt_obs[n] = - rsurfobs[n]/c; % note minus sign here (rsurfobs = rsurfsig * obs)
    % calculate Doppler factor
    repro.doppler[n] = struct_field_exists(surf, "v") and not qualifier_exists("noDoppler")
                     ? 1.-(surf.v[n]*obs)/c
                     : 1. + repro.doppler[n]; % here, repro.doppler is zero as initialized

    do {
      redo = 0;
      surfcopy = struct_filter(surf, n; copy);
      % calculate intrinsic flux of each element as seen by observer
      if (struct_field_exists(surf, "L")) {
        repro.flux_int[n] = (@powerRedist)(
          surf.L[n], obsproj[n], surfcopy, 1, &redo
	  ;; qualifier("powRedistQual", NULL)
        );
      }
      % proceed with elements visible from signal origin as well
      m = n[where(repro.visible_sig[n])];
      if (length(m) > 0) {
        % calculate received relative power
        repro.flux_rec[m] = 1./4/PI/lsurfsig[m]/lsurfsig[m] * surf.A[m] * sigproj[m];
        % calculate emitted relative flux
        repro.flux_emi[m] = (@powerRedist)(
	  repro.flux_rec[m], obsproj[m], struct_filter(surf, m; copy), 2, &redo
	  ;; qualifier("powRedistQual", NULL)
        );
      }
      r++;
    } while (redo and r<10); % re-calculate if redistribution function triggered it
    if (r == 10) { vmessage("warning (%s): re-calculation aborted after 10 iterations", _function_name); }
  }

  % eventually return reproment only
  if (qualifier_exists("reproment")) { return repro; }

  % proceed and calculate echo signal
  return T == NULL
    ? EchoMap_makelc(repro, sig;; __qualifiers)
    : EchoMap_makelc(repro, sig, T;; __qualifiers);
}

  
%%%%%%%%%%%%%%%%%%%%%
define EchoMap_binary()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{EchoMap_binary}
%\synopsis{reprocesses a signal on the companion of a binary}
%\usage{Struct_Type EchoMap_binary(
%      Struct_Type surface, Double_Type lum, Struct_Type signal,
%      Struct_Type orb, Double_Type phiorb, Double_Type mq
%      [, Double_Type length]
%    );}
%\qualifiers{
%    \qualifier{remark: }{any qualifiers are passed to 'EchoMap'}
%}
%\description
%    Uses the 'EchoMap' function to reprocess an input 'signal'
%    on the 'surface' of the secondary star with total luminosity
%    'lum'. The fields required for the 'surface' and 'signal'
%    structure are described in the 'EchoMap' help. The
%    'Roche_lobe_surface' function might be useful to calculate
%    the deformed surface of the star within the Roche potential.
%    Note, that the velocities of the surface elements are
%    calculated automatically if no 'v' field is specified. If
%    not specified, the luminosity field 'L' for each surface
%    element is calculated as well. Thereby, it is assumed that
%    each surface element has the same area luminosity, such
%    that the total luminosity still is 'lum'.
%    The orbit of the binary is described by the 'orb'ital
%    structure an has to contain the fields as described in the
%    'check_ephemeris' help. In addition, the inclination can
%    be specified optionally via the 'i' field. To complete the
%    description of the binary, the mass ratio
%      mq = M_primary / M_companion
%    has to be given at last.
%    As default, the source of the signal is the position of
%    the primary star. It can be changed by setting the 'origin'
%    field of the signal structur to the position vector.
%    Using the 'phiorb' parameter the orbital phase is defined,
%    which corresponds to the viewing angle on the binary.
%
%    As a conclusion and for further details the reference
%    frame is shown in the following:
%#v+
%                                   position of M2 depending
%      omega (right handed)         on orbital phase
%      observer (i = 0)
%        z                                 0.50
%        |  y                              ---
%        | /                              |   |
%        |/                         0.75 |  x  | 0.25
%    M1  o ---- x                         |- -|
%       /                                   | observer
%      o M2 (phiorb = 0)                  0.00
%#v-
%    - orbital plane = x-y-plane (fix)
%    - inclination and orbital phase is realized by rotating
%      the observer accordingly
%    - distance M_1 to M_2 = 1 (fix)
%
%    WARNING: at the moment circular orbits (ecc = 0) are
%             are implemented only!
%    NOTE:    for elliptical orbits, the shape of the stars
%             depends on the orbital phase
%\seealso{EchoMap, Roche_lobe_surface}
%!%-
{
  variable surf, lum, sig, orb, phi, mq, T = NULL;
  switch(_NARGS)
    { case 6: (surf, lum, sig, orb, phi, mq) = (); }
    { case 7: (surf, lum, sig, orb, phi, mq, T) = (); }
    { help(_function_name()); return; }
  surf = COPY(surf);

  % check on inclination
  orb = COPY(orb);
  ifnot (struct_field_exists(orb, "i")) orb = struct_combine(orb, struct { i = 90 });
  
  % parameter validity checks
  ifnot (mq > 0) { vmessage("error (%s): mass ratio 'mq' must be greater zero", _function_name); return; }

  % some variables, note reference frame definition in description above
  variable r2 = vector(0,-1,0); % vector to center of companion (for phi = 0)
  variable ez = vector(0,0,1); % unity vector in z-direction (= omega direction)
  variable obs = vector(0,0,1); % unity vector to observer (for i = 0)
  
  % incline binary by rotating observer
  obs = vector_rotate(obs, vector(1,0,0), PI/180*orb.i);
  % rotate binary by rotating observer in backwards direction
  obs = vector_rotate(obs, ez, -2*PI*phi);

  % calculate surface luminosity
  ifnot (struct_field_exists(surf, "L")) {
    surf = struct_combine(surf, struct {
      L = lum/sum(surf.A)*surf.A
    });
  }

  % make sure M1 is at (0,0,0)
  sig = COPY(sig);
  ifnot (struct_field_exists(sig, "origin")) sig = struct_combine(sig, struct { origin });
  sig.origin = vector(0,0,0);

  surf.r += r2;

  % calculate velocity vector of each surface element (for circular orbits!)
  ifnot (struct_field_exists(surf, "v")) {
    variable xcen = vector(0, -1./(mq + 1), 0); % center of mass (for phi = 0)
    surf = struct_combine(surf, struct {
      v = array_map(Vector_Type, &crossprod, ez, surf.r-xcen) * 2*PI *(1. / orb.porb / 86400)
    });
  }

  % convert unit of length from binary displacement to lt-s via speed of light
  variable newc = 1./radius_to_unit(1, "disp", "lts"; asini = orb.asini, i = orb.i, mq = mq);
  
  variable qual = struct_combine(__qualifiers, struct { c = newc });
  
  return T == NULL
    ? EchoMap(surf, sig, obs;; qual)
    : EchoMap(surf, sig, obs, T;; qual);
}

#endif

%
% Code from src/slang/date/EpochofJD.sl
%
define EpochofJD()
%!%+
%\function{EpochofJD}
%\synopsis{Returns the Julian or Besselian epoch of a (M)JD}
%\usage{epoch=EpochofJD(jd;qualifiers)}
%\qualifiers{
%\qualifier{mjd}{argument is a modified julian date, not a JD}
%\qualifier{besselian}{return the Besselian epoch}
%\qualifier{julian}{return the Julian epoch (the default)}
%}
%\description
%   This function converts a (modified) julian date into the corresponding
%   Julian or Besselian epoch (e.g., J2000, B1950.0 etc.) using the formulae
%   given by Lieske (1979, A&A 73, 282) for Besselian dates and the standard
%   definition of the Julian Date. 
%
%   If high precision is relevant, note that JD is assumed to be in TT,
%   note that the epoch will slowly drift with respect to Gregorian year
%   fractions such as those calculated by jd2year, but contrary to that
%   function the Julian/Besselian epoch linearly maps to TT.
%
%   This routine is array safe.
%
%\seealso{JDofEpoch,jd2year}
%!%-
{
    variable jd;

    switch(_NARGS)
    { case 1: jd=(); }
    { help(_function_name());return;}

    % Lieske (1979, A&A 73, 282)
    % Lieske: B = 1900.0 + (Julian date-2415020.31352) / 365.242198781
    %J = 2000.0 + (Julian date-2451545.0)/365.25 .
    
    if(qualifier_exists("besselian")) {
	if (qualifier_exists("mjd")) {
	    return 1900.+(jd-15019.81352)/365.242198781; 
	} else {
	    return 1900.+(jd-2415020.31352)/365.242198781;
	}
    }

    % Julian epoch
    if (qualifier_exists("mjd")) {
	return 2000.0+(jd-51544.5)/365.25;
    }
    return 2000.0+(jd-2451545.0)/365.25; 
}


%
% Code from src/slang/math/Faddeeva_function.sl
%
% Implementation of the Faddeeva function as described in
% https://dl.acm.org/citation.cfm?doid=3132683.3119904.
%
% Other routines use this to compute erf, erfc, real and imaginary voigt functions
%

private define __Faddeeva (z)
%{{{%
{
  % takes real or complex array returns Faddeyeva function w(z) = exp((-iz)^2)erfc(-iz)

  % often used numbers
  variable spi = 0.5641895835477563; % 1/sqrt(pi)
  variable ispi   = 1i*spi;     % i/sqrt(pi)
  variable zabs2 = (Real(z)^2 + Imag(z)^2);

  variable t; % integration variable
  variable t2;
  variable wz = @Complex_Type[length(z)]*_NaN; % return value, initialize as NaN
  variable wis;
  % Region 1 approximation
  variable ir1 = (zabs2 >= 38000.0);
  if (any(ir1)) {
    wis = where(ir1);
    t = z[wis];
    wz[wis] = ispi/t;
  }
  % Region 2
  variable ir2 = ( (38000.0>zabs2) and (zabs2>=256.0) );
  if (any(ir2)) {
    wis = where(ir2);
    t = z[wis];
    wz[wis] = ispi*t/(t*t-.5);
  }
  % Region 3
  variable ir3 = ( (256.0>zabs2) and (zabs2>=62.0) );
  if (any(ir3)) {
    wis = where(ir3);
    t=z[wis];
    wz[wis] = (ispi/t)*(1+.5/(t*t-1.5));
  }
  % Region 4
  variable ir4 = ( (62.0>zabs2) and (zabs2>=30.0) and (Imag(z)^2>=1e-13) );
  if (any(ir4)) {
    wis = where(ir4);
    t = z[wis];
    t2 = t*t;
    wz[wis] = ispi*t*(t2-2.5)/(t2*(t2-3.0)+.75);
  }
  % Region 5
  variable ir5 = ( (62.0>zabs2) and (not ir4) and ( (zabs2>2.5) and (Imag(z)^2<.072) ) );
  if (any(ir5)) {
    wis = where(ir5);
    t = z[wis];
    t2 = -t*t;
    wz[wis] = exp(t2)+1i*t*(36183.31-t2*(3321.99-t2*(1540.787-t2*(219.031-t2*(35.7668-t2*(
                    1.320522-t2*spi))))))/
                    (32066.6-t2*(24322.84-t2*(9022.228-t2*(2186.181-t2*(364.2191-t2*(
		    61.57037-t2*(1.841439-t2)))))));
  }
  % Region 6
  variable ir6 = ( (30.0>zabs2) and (not ir5) );
  if (any(ir6)) {
    wis = where(ir6);
    t = -1i*z[wis];
    wz[wis] = (122.60793+t*(214.38239+t*(181.92853+t*(93.15558+t*(30.180142+t*(
		   spi*t+5.9126262))))))/
                   (122.60793+t*(352.73063+t*(457.33448+t*(348.70392+t*(170.35400+t*(
		   53.992907+t*(t+10.479857)))))));
  }

  return wz;
}
%}}}%

%%%%%%%%%%%%%%%%%%%
define Faddeeva (z)
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Faddeeva}
%\synopsis{Compute w(z) = exp((-iz)^2 erfc(-iz) for complex z}
%\usage{Complex_Type[] Faddeeva(Complex_Type[]);}
%#c%{{{%
%\description
%    This function uses continued fraction expansion and the algorithms described by
%    Humlicek () and Hui () to compute an approximation to the Faddeeva function
%
%    The algorithm used only allows to compute w(z) for Im(z)>=0, but using the relation
%    w(-z) = 2*exp(-z^2)-w(-z) gives the remaining half.
%
%    The derivative is given via dw/dz = 2i/sqrt(pi) - 2*z*w(z)
%    \seealso{Faddeeva_dz}
%
%    It is claimed that the algorithm has an accuracy of <4e-4.
%!%-
{
  variable ini = Imag(z)<0; % Is Negative Imaginary
  variable w = Complex_Type[length(z)]; % Where Is Negative Imaginary
  variable wini = where(ini);
  variable wipi = where(not ini); % Where Is Positive Imaginary
  w[wini] = 2*exp(-z[wini]^2)-__Faddeeva(-z[wini]);
  w[wipi] = __Faddeeva(z[wipi]);
  return (length(w)>1) ? w : w[0];
}
%}}}%

%%%%%%%%%%%%%%%%%%%%%%
define Faddeeva_dz (z)
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Faddeeva_dz}
%\synopsis{Compute derivative of Faddeeva function}
%\usage{Complex_Type[] Faddeeva_dz (Complex_Type[]);}
%\description
%    Computes complex derivative of Faddeeva function
%    \seealso{Faddeeva}
%!%-
{
  return 1.12837916709551*1i-2*z*Faddeeva(z);
}

%%%%%%%%%%%%%%%%
define cerfc (z)
%%%%%%%%%%%%%%%%
%!%+
%\function{cerfc}
%\synopsis{Complex error function complement}
%\usage{Complex_Type[] = cerfc(Complex_Type[]);}
%\description
%    Compute complex error function complement. Only useful in a region
%    with abs(z)<10.
%\seealso{Faddeeva, cerf}
%!%-
{
  return exp(-z^2)*Faddeeva(1i*z);
}

%%%%%%%%%%%%%%%
define cerf (z)
%%%%%%%%%%%%%%%
%!%+
%\function{cerf}
%\synopsis{Complex error function}
%\usage{Complex_Type[] = cerf(Complex_Type[]);}
%\description
%    Compute complex error function. Only useful in a region
%    with abs(z)<10.
%\seealso{Faddeeva, cerfc}
%!%-
{
  return 1-cerfc(z);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define Lorentz_complex (z, z0, gamma)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Lorentz_complex}
%\synopsis{Compute complex Lorentz profile}
%\usage{Complex_Type[] = Lorentz_complex(z, z0, gamma);}
%!%-
{
  variable g2 = gamma^2;
  return 1./(PI*gamma)*(g2/((z-z0)^2+g2));
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define Gauss_complex (z, z0, sig)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Gauss_complex}
%\synopsis{Compute complex Gauss profile}
%\usage{Complex_Type = Gauss_complex(z, z0, sigma);}
%!%-
{
  variable s2 = sig^2;
  return 1./sqrt(2*PI*s2)*exp(-.5*(z-z0)^2/s2);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define Voigt_complex (z, z0, sig, gamma)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Voigt_complex}
%\synopsis{Compute complex Voigt profile}
%\usage{Complex_Type[] = Voigt_complex(z, z0, sigma, gamma);}
%!%-
{
  if (sig<0 || gamma<0) return z*0;
  else if (sig==0 && gamma==0) return z*0;
  else if (sig==0 && gamma>0) return Lorentz_complex(z,z0,gamma);
  else if (sig>0 && gamma==0) return Gauss_complex(z,z0,sig);
  variable zz = (z-z0+1i*gamma)/(sig*sqrt(2));
  return Faddeeva(zz)/(sig*sqrt(2*PI));
}

%
% Code from src/misc/wind_model/FriendCastor1982_dMdot_dOmega_CygX1.sl
%
define FriendCastor1982_dMdot_dOmega_CygX1()
%!%+
%\function{FriendCastor1982_dMdot_dOmega_CygX1}
%\synopsis{calculates the mass loss rate per solid angle for Cyg X-1}
%\usage{Double_Type FriendCastor1982_dMdot_dOmega_CygX1(Double_Type theta)}
%\description
%    \code{theta} is the angle from the binary axis in degrees.
%    The return value is in units of 1e-6 solar masses / year / sr.
%    (The total mass loss rate \code{2 pi int dMdot/dOmega(theta) sin(theta) dtheta}
%     is 2e-6 solar masses / year.)
%\seealso{Friend & Castor (1982), Fig. 4}
%!%-
{
  variable theta;
  switch(_NARGS)
  { case 1: theta = (); }
  { help(_function_name()); return; }

  return interpol_points(theta,
  [  0.000000,   0.214286,   0.428571,   0.642857,   0.857143,   1.071429,   1.285714,   1.500000,   1.714286,   1.928571,
     2.142857,   2.357143,   2.571429,   2.785714,   3.000000,   3.214286,   3.428571,   3.642857,   3.857143,   4.071429,
     4.285714,   4.500000,   4.714286,   4.928571,   5.142857,   5.357143,   5.571429,   5.785714,   6.000000,   6.214286,
     6.428571,   6.642857,   6.857143,   7.071429,   7.285714,   7.500000,   7.714286,   7.928571,   8.142857,   8.357143,
     8.571429,   8.785714,   9.000000,   9.214286,   9.428571,   9.642857,   9.857143,  10.071429,  10.285714,  10.500000,
    10.714286,  10.928571,  11.142857,  11.357143,  11.571429,  11.785714,  12.000000,  12.214286,  12.428571,  12.642857,
    12.857143,  13.071429,  13.285714,  13.500000,  13.714286,  13.928571,  14.142857,  14.357143,  14.571429,  14.785714,
    15.000000,  15.214286,  15.428571,  15.642857,  15.857143,  16.071429,  16.285714,  16.500000,  16.714286,  16.928571,
    17.142857,  17.357143,  17.571429,  17.785714,  18.000000,  18.214286,  18.428571,  18.642857,  18.857143,  19.071429,
    19.285714,  19.500000,  19.714286,  19.928571,  20.142857,  20.357143,  20.571429,  20.785714,  21.000000,  21.214286,
    21.428571,  21.642857,  21.857143,  22.071429,  22.285714,  22.500000,  22.714286,  22.928571,  23.142857,  23.357143,
    23.571429,  23.785714,  24.000000,  24.214286,  24.428571,  24.642857,  24.857143,  25.071429,  25.285714,  25.500000,
    25.714286,  25.928571,  26.142857,  26.357143,  26.571429,  26.785714,  27.000000,  27.214286,  27.428571,  27.642857,
    27.857143,  28.071429,  28.285714,  28.500000,  28.714286,  28.928571,  29.142857,  29.357143,  29.571429,  29.785714,
    30.000000,  30.214286,  30.428571,  30.642857,  30.857143,  31.071429,  31.285714,  31.500000,  31.714286,  31.928571,
    32.142857,  32.357143,  32.571429,  32.785714,  33.000000,  33.214286,  33.428571,  33.642857,  33.857143,  34.071429,
    34.285714,  34.500000,  34.714286,  34.928571,  35.142857,  35.357143,  35.571429,  35.785714,  36.000000,  36.214286,
    36.428571,  36.642857,  36.857143,  37.071429,  37.285714,  37.500000,  37.714286,  37.928571,  38.142857,  38.357143,
    38.571429,  38.785714,  39.000000,  39.214286,  39.428571,  39.642857,  39.857143,  40.071429,  40.285714,  40.500000,
    40.714286,  40.928571,  41.142857,  41.357143,  41.571429,  41.785714,  42.000000,  42.214286,  42.428571,  42.642857,
    42.857143,  43.071429,  43.285714,  43.500000,  43.714286,  43.928571,  44.142857,  44.357143,  44.571429,  44.785714,
    45.000000,  45.214286,  45.428571,  45.642857,  45.857143,  46.071429,  46.285714,  46.500000,  46.714286,  46.928571,
    47.142857,  47.357143,  47.571429,  47.785714,  48.000000,  48.214286,  48.428571,  48.642857,  48.857143,  49.071429,
    49.285714,  49.500000,  49.714286,  49.928571,  50.142857,  50.357143,  50.571429,  50.785714,  51.000000,  51.214286,
    51.428571,  51.642857,  51.857143,  52.071429,  52.285714,  52.500000,  52.714286,  52.928571,  53.142857,  53.357143,
    53.571429,  53.785714,  54.000000,  54.214286,  54.428571,  54.642857,  54.857143,  55.071429,  55.285714,  55.500000,
    55.714286,  55.928571,  56.142857,  56.357143,  56.571429,  56.785714,  57.000000,  57.214286,  57.428571,  57.642857,
    57.857143,  58.071429,  58.285714,  58.500000,  58.714286,  58.928571,  59.142857,  59.357143,  59.571429,  59.785714,
    60.000000,  60.214286,  60.428571,  60.642857,  60.857143,  61.071429,  61.285714,  61.500000,  61.714286,  61.928571,
    62.142857,  62.357143,  62.571429,  62.785714,  63.000000,  63.214286,  63.428571,  63.642857,  63.857143,  64.071429,
    64.285714,  64.500000,  64.714286,  64.928571,  65.142857,  65.357143,  65.571429,  65.785714,  66.000000,  66.214286,
    66.428571,  66.642857,  66.857143,  67.071429,  67.285714,  67.500000,  67.714286,  67.928571,  68.142857,  68.357143,
    68.571429,  68.785714,  69.000000,  69.214286,  69.428571,  69.642857,  69.857143,  70.071429,  70.285714,  70.500000,
    70.714286,  70.928571,  71.142857,  71.357143,  71.571429,  71.785714,  72.000000,  72.214286,  72.428571,  72.642857,
    72.857143,  73.071429,  73.285714,  73.500000,  73.714286,  73.928571,  74.142857,  74.357143,  74.571429,  74.785714,
    75.000000,  75.214286,  75.428571,  75.642857,  75.857143,  76.071429,  76.285714,  76.500000,  76.714286,  76.928571,
    77.142857,  77.357143,  77.571429,  77.785714,  78.000000,  78.214286,  78.428571,  78.642857,  78.857143,  79.071429,
    79.285714,  79.500000,  79.714286,  79.928571,  80.142857,  80.357143,  80.571429,  80.785714,  81.000000,  81.214286,
    81.428571,  81.642857,  81.857143,  82.071429,  82.285714,  82.500000,  82.714286,  82.928571,  83.142857,  83.357143,
    83.571429,  83.785714,  84.000000,  84.214286,  84.428571,  84.642857,  84.857143,  85.071429,  85.285714,  85.500000,
    85.714286,  85.928571,  86.142857,  86.357143,  86.571429,  86.785714,  87.000000,  87.214286,  87.428571,  87.642857,
    87.857143,  88.071429,  88.285714,  88.500000,  88.714286,  88.928571,  89.142857,  89.357143,  89.571429,  89.785714,
    90.000000,  90.214286,  90.428571,  90.642857,  90.857143,  91.071429,  91.285714,  91.500000,  91.714286,  91.928571,
    92.142857,  92.357143,  92.571429,  92.785714,  93.000000,  93.214286,  93.428571,  93.642857,  93.857143,  94.071429,
    94.285714,  94.500000,  94.714286,  94.928571,  95.142857,  95.357143,  95.571429,  95.785714,  96.000000,  96.214286,
    96.428571,  96.642857,  96.857143,  97.071429,  97.285714,  97.500000,  97.714286,  97.928571,  98.142857,  98.357143,
    98.571429,  98.785714,  99.000000,  99.214286,  99.428571,  99.642857,  99.857143, 100.071429, 100.285714, 100.500000,
   100.714286, 100.928571, 101.142857, 101.357143, 101.571429, 101.785714, 102.000000, 102.214286, 102.428571, 102.642857,
   102.857143, 103.071429, 103.285714, 103.500000, 103.714286, 103.928571, 104.142857, 104.357143, 104.571429, 104.785714,
   105.000000, 105.214286, 105.428571, 105.642857, 105.857143, 106.071429, 106.285714, 106.500000, 106.714286, 106.928571,
   107.142857, 107.357143, 107.571429, 107.785714, 108.000000, 108.214286, 108.428571, 108.642857, 108.857143, 109.071429,
   109.285714, 109.500000, 109.714286, 109.928571, 110.142857, 110.357143, 110.571429, 110.785714, 111.000000, 111.214286,
   111.428571, 111.642857, 111.857143, 112.071429, 112.285714, 112.500000, 112.714286, 112.928571, 113.142857, 113.357143,
   113.571429, 113.785714, 114.000000, 114.214286, 114.428571, 114.642857, 114.857143, 115.071429, 115.285714, 115.500000,
   115.714286, 115.928571, 116.142857, 116.357143, 116.571429, 116.785714, 117.000000, 117.214286, 117.428571, 117.642857,
   117.857143, 118.071429, 118.285714, 118.500000, 118.714286, 118.928571, 119.142857, 119.357143, 119.571429, 119.785714,
   120.000000, 120.214286, 120.428571, 120.642857, 120.857143, 121.071429, 121.285714, 121.500000, 121.714286, 121.928571,
   122.142857, 122.357143, 122.571429, 122.785714, 123.000000, 123.214286, 123.428571, 123.642857, 123.857143, 124.071429,
   124.285714, 124.500000, 124.714286, 124.928571, 125.142857, 125.357143, 125.571429, 125.785714, 126.000000, 126.214286,
   126.428571, 126.642857, 126.857143, 127.071429, 127.285714, 127.500000, 127.714286, 127.928571, 128.142857, 128.357143,
   128.571429, 128.785714, 129.000000, 129.214286, 129.428571, 129.642857, 129.857143, 130.071429, 130.285714, 130.500000,
   130.714286, 130.928571, 131.142857, 131.357143, 131.571429, 131.785714, 132.000000, 132.214286, 132.428571, 132.642857,
   132.857143, 133.071429, 133.285714, 133.500000, 133.714286, 133.928571, 134.142857, 134.357143, 134.571429, 134.785714,
   135.000000, 135.214286, 135.428571, 135.642857, 135.857143, 136.071429, 136.285714, 136.500000, 136.714286, 136.928571,
   137.142857, 137.357143, 137.571429, 137.785714, 138.000000, 138.214286, 138.428571, 138.642857, 138.857143, 139.071429,
   139.285714, 139.500000, 139.714286, 139.928571, 140.142857, 140.357143, 140.571429, 140.785714, 141.000000, 141.214286,
   141.428571, 141.642857, 141.857143, 142.071429, 142.285714, 142.500000, 142.714286, 142.928571, 143.142857, 143.357143,
   143.571429, 143.785714, 144.000000, 144.214286, 144.428571, 144.642857, 144.857143, 145.071429, 145.285714, 145.500000,
   145.714286, 145.928571, 146.142857, 146.357143, 146.571429, 146.785714, 147.000000, 147.214286, 147.428571, 147.642857,
   147.857143, 148.071429, 148.285714, 148.500000, 148.714286, 148.928571, 149.142857, 149.357143, 149.571429, 149.785714,
   150.000000, 150.214286, 150.428571, 150.642857, 150.857143, 151.071429, 151.285714, 151.500000, 151.714286, 151.928571,
   152.142857, 152.357143, 152.571429, 152.785714, 153.000000, 153.214286, 153.428571, 153.642857, 153.857143, 154.071429,
   154.285714, 154.500000, 154.714286, 154.928571, 155.142857, 155.357143, 155.571429, 155.785714, 156.000000, 156.214286,
   156.428571, 156.642857, 156.857143, 157.071429, 157.285714, 157.500000, 157.714286, 157.928571, 158.142857, 158.357143,
   158.571429, 158.785714, 159.000000, 159.214286, 159.428571, 159.642857, 159.857143, 160.071429, 160.285714, 160.500000,
   160.714286, 160.928571, 161.142857, 161.357143, 161.571429, 161.785714, 162.000000, 162.214286, 162.428571, 162.642857,
   162.857143, 163.071429, 163.285714, 163.500000, 163.714286, 163.928571, 164.142857, 164.357143, 164.571429, 164.785714,
   165.000000, 165.214286, 165.428571, 165.642857, 165.857143, 166.071429, 166.285714, 166.500000, 166.714286, 166.928571,
   167.142857, 167.357143, 167.571429, 167.785714, 168.000000, 168.214286, 168.428571, 168.642857, 168.857143, 169.071429,
   169.285714, 169.500000, 169.714286, 169.928571, 170.142857, 170.357143, 170.571429, 170.785714, 171.000000, 171.214286,
   171.428571, 171.642857, 171.857143, 172.071429, 172.285714, 172.500000, 172.714286, 172.928571, 173.142857, 173.357143,
   173.571429, 173.785714, 174.000000, 174.214286, 174.428571, 174.642857, 174.857143, 175.071429, 175.285714, 175.500000,
   175.714286, 175.928571, 176.142857, 176.357143, 176.571429, 176.785714, 177.000000, 177.214286, 177.428571, 177.642857,
   177.857143, 178.071429, 178.285714, 178.500000, 178.714286, 178.928571, 179.142857, 179.357143, 179.571429, 180.000000
  ],
  [0.632799, 0.632799, 0.632799, 0.632799, 0.632799, 0.632799, 0.632799, 0.632799, 0.627897, 0.617380,
   0.606506, 0.600267, 0.580315, 0.563280, 0.543226, 0.524955, 0.513815, 0.499846, 0.484403, 0.470627,
   0.456898, 0.443939, 0.430481, 0.424269, 0.413469, 0.401515, 0.392602, 0.382353, 0.376560, 0.367201,
   0.358734, 0.350499, 0.343137, 0.338235, 0.331996, 0.324866, 0.319123, 0.311943, 0.307431, 0.301248,
   0.297683, 0.292168, 0.287433, 0.282977, 0.278075, 0.275401, 0.271527, 0.266934, 0.262849, 0.259358,
   0.256239, 0.252593, 0.248663, 0.245900, 0.242424, 0.239750, 0.236185, 0.234281, 0.230838, 0.228610,
   0.225738, 0.223262, 0.221925, 0.219598, 0.217023, 0.216231, 0.213012, 0.211676, 0.209447, 0.208111,
   0.206201, 0.204545, 0.202763, 0.200980, 0.200089, 0.198307, 0.196524, 0.195187, 0.194296, 0.193405,
   0.191622, 0.190285, 0.189840, 0.188057, 0.187611, 0.186720, 0.185383, 0.183601, 0.182709, 0.182264,
   0.180927, 0.179590, 0.179144, 0.177807, 0.177807, 0.176025, 0.175579, 0.175134, 0.173797, 0.172906,
   0.172460, 0.171569, 0.170677, 0.170232, 0.169340, 0.168895, 0.168449, 0.167112, 0.167558, 0.166667,
   0.165775, 0.165330, 0.164884, 0.164439, 0.163993, 0.163102, 0.163102, 0.162210, 0.161765, 0.160873,
   0.161319, 0.160428, 0.159537, 0.159091, 0.158645, 0.158200, 0.157754, 0.157308, 0.156863, 0.156417,
   0.156863, 0.155971, 0.155526, 0.155080, 0.154189, 0.154189, 0.153743, 0.154189, 0.153298, 0.152852,
   0.152406, 0.152406, 0.151961, 0.151961, 0.151515, 0.150624, 0.150624, 0.150624, 0.150178, 0.150178,
   0.149733, 0.149733, 0.149287, 0.148841, 0.148841, 0.148396, 0.147950, 0.147950, 0.147504, 0.147950,
   0.147059, 0.147059, 0.147059, 0.146613, 0.146613, 0.146168, 0.146168, 0.145722, 0.146168, 0.145276,
   0.145276, 0.145276, 0.145276, 0.144831, 0.144385, 0.144385, 0.144385, 0.144385, 0.143494, 0.143939,
   0.143494, 0.143939, 0.143494, 0.143048, 0.143494, 0.143048, 0.143048, 0.142602, 0.143048, 0.142602,
   0.142602, 0.142602, 0.142602, 0.142157, 0.142602, 0.142157, 0.142602, 0.141711, 0.141711, 0.141711,
   0.141711, 0.141711, 0.141711, 0.141711, 0.141266, 0.141266, 0.141266, 0.140820, 0.141266, 0.140820,
   0.141266, 0.140820, 0.141266, 0.140374, 0.140820, 0.140374, 0.140374, 0.140374, 0.140374, 0.140374,
   0.140374, 0.140374, 0.140374, 0.139929, 0.139929, 0.139483, 0.139929, 0.139483, 0.139929, 0.139483,
   0.139929, 0.139483, 0.139929, 0.139483, 0.139929, 0.139483, 0.139483, 0.139483, 0.139483, 0.139483,
   0.139483, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037,
   0.139037, 0.139037, 0.139037, 0.139037, 0.138592, 0.138592, 0.138592, 0.139037, 0.138146, 0.138592,
   0.138146, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592, 0.138146, 0.138592,
   0.138592, 0.138592, 0.138592, 0.138146, 0.138592, 0.138146, 0.138592, 0.138146, 0.138592, 0.138592,
   0.138592, 0.138592, 0.138146, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592,
   0.139037, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592, 0.138592,
   0.138592, 0.138592, 0.138592, 0.138592, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037,
   0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037, 0.139037,
   0.139037, 0.139037, 0.139037, 0.139483, 0.139483, 0.139483, 0.139483, 0.139483, 0.139483, 0.139483,
   0.139929, 0.139483, 0.139929, 0.139483, 0.139929, 0.139483, 0.139929, 0.139929, 0.140374, 0.139929,
   0.140374, 0.139929, 0.140374, 0.139929, 0.140374, 0.139929, 0.140374, 0.140374, 0.140374, 0.140374,
   0.140374, 0.140374, 0.140374, 0.140374, 0.140374, 0.140374, 0.140374, 0.140820, 0.140374, 0.140820,
   0.140820, 0.141266, 0.140820, 0.141266, 0.140820, 0.141266, 0.140820, 0.141266, 0.140820, 0.141266,
   0.140820, 0.141266, 0.141266, 0.141266, 0.141266, 0.141711, 0.141711, 0.141711, 0.141711, 0.141711,
   0.141711, 0.141711, 0.141711, 0.141711, 0.141711, 0.142157, 0.142157, 0.142157, 0.142602, 0.142157,
   0.142602, 0.142157, 0.142602, 0.142157, 0.142602, 0.143048, 0.142602, 0.143048, 0.142602, 0.143048,
   0.142602, 0.143048, 0.142602, 0.143048, 0.142602, 0.143048, 0.143048, 0.143048, 0.143494, 0.143494,
   0.143494, 0.143494, 0.143494, 0.143494, 0.143494, 0.143494, 0.143494, 0.143939, 0.144385, 0.143939,
   0.144385, 0.143939, 0.144385, 0.143939, 0.144385, 0.143939, 0.144385, 0.144385, 0.144831, 0.144831,
   0.144831, 0.144831, 0.144831, 0.144831, 0.144831, 0.144831, 0.145276, 0.145276, 0.145276, 0.145722,
   0.145276, 0.145722, 0.145276, 0.145722, 0.145276, 0.146168, 0.146168, 0.146168, 0.146168, 0.146168,
   0.146168, 0.146168, 0.146168, 0.146168, 0.146168, 0.146613, 0.146168, 0.146613, 0.146613, 0.147059,
   0.147059, 0.147059, 0.147059, 0.147504, 0.147059, 0.147504, 0.147059, 0.147504, 0.147059, 0.147950,
   0.147950, 0.147950, 0.147950, 0.147950, 0.147950, 0.147950, 0.148841, 0.148396, 0.148841, 0.148396,
   0.148841, 0.148396, 0.149287, 0.149287, 0.149287, 0.149287, 0.149287, 0.149287, 0.149733, 0.149287,
   0.150178, 0.149733, 0.150178, 0.149733, 0.150178, 0.150178, 0.150624, 0.150178, 0.150624, 0.150624,
   0.151070, 0.150624, 0.151070, 0.151070, 0.151070, 0.151070, 0.151070, 0.151961, 0.151515, 0.151961,
   0.151515, 0.151961, 0.151961, 0.151961, 0.152406, 0.152406, 0.152406, 0.152406, 0.152852, 0.152852,
   0.153298, 0.152852, 0.153298, 0.152852, 0.153743, 0.153298, 0.153743, 0.153743, 0.153743, 0.154189,
   0.154189, 0.154189, 0.154635, 0.154635, 0.155080, 0.154635, 0.155526, 0.155080, 0.155526, 0.155526,
   0.155526, 0.155971, 0.155971, 0.156417, 0.155971, 0.156417, 0.156417, 0.156417, 0.156863, 0.157308,
   0.157308, 0.157754, 0.157308, 0.158200, 0.157308, 0.158200, 0.157754, 0.158200, 0.158645, 0.158645,
   0.158645, 0.158645, 0.158645, 0.159537, 0.159091, 0.159982, 0.159537, 0.159982, 0.159982, 0.159982,
   0.160428, 0.160428, 0.160873, 0.160428, 0.161319, 0.161319, 0.161319, 0.161319, 0.161765, 0.162210,
   0.161765, 0.162210, 0.162210, 0.162656, 0.162210, 0.163102, 0.163102, 0.163102, 0.163547, 0.163993,
   0.163547, 0.164439, 0.164439, 0.164439, 0.164439, 0.165330, 0.164884, 0.165330, 0.164884, 0.165775,
   0.165330, 0.165775, 0.166221, 0.166221, 0.166667, 0.166667, 0.167112, 0.166667, 0.167558, 0.167558,
   0.167558, 0.168449, 0.168004, 0.168449, 0.168895, 0.168449, 0.169340, 0.169340, 0.169340, 0.169786,
   0.169786, 0.170232, 0.170232, 0.170677, 0.171123, 0.171123, 0.172014, 0.172014, 0.172014, 0.172014,
   0.172906, 0.172460, 0.173351, 0.172906, 0.173797, 0.173797, 0.173797, 0.174242, 0.174688, 0.175134,
   0.175134, 0.175579, 0.176025, 0.175579, 0.176025, 0.176471, 0.176471, 0.176916, 0.177362, 0.177362,
   0.177807, 0.178253, 0.178253, 0.179144, 0.178699, 0.179144, 0.179590, 0.179144, 0.180036, 0.180481,
   0.180481, 0.180927, 0.181373, 0.181373, 0.181373, 0.181818, 0.182709, 0.182709, 0.182709, 0.183155,
   0.183601, 0.183601, 0.184492, 0.184492, 0.184492, 0.184938, 0.185383, 0.185829, 0.185829, 0.186720,
   0.186275, 0.186720, 0.187166, 0.187166, 0.188057, 0.187611, 0.188503, 0.188057, 0.188948, 0.188948,
   0.189394, 0.189840, 0.190285, 0.190285, 0.190731, 0.191176, 0.191622, 0.191176, 0.192068, 0.192513,
   0.192513, 0.193405, 0.193405, 0.193405, 0.194296, 0.194296, 0.194742, 0.195187, 0.195187, 0.196078,
   0.196078, 0.196524, 0.196970, 0.197415, 0.197415, 0.197861, 0.197861, 0.198752, 0.199643, 0.199198,
   0.199643, 0.200535, 0.200535, 0.201426, 0.201426, 0.201426, 0.202317, 0.203209, 0.202763, 0.203209,
   0.204100, 0.204100, 0.204991, 0.204991, 0.205437, 0.205882, 0.206328, 0.206774, 0.206774, 0.206774,
   0.207665, 0.208111, 0.208111, 0.209002, 0.208556, 0.209447, 0.209893, 0.209893, 0.210339, 0.211230,
   0.210784, 0.211230, 0.211676, 0.212121, 0.212567, 0.213012, 0.213012, 0.213904, 0.213904, 0.213904,
   0.214349, 0.214349, 0.215241, 0.215241, 0.215686, 0.215686, 0.216132, 0.216578, 0.217023, 0.217023,
   0.217469, 0.217469, 0.217914, 0.218360, 0.218806, 0.218360, 0.219251, 0.219251, 0.219251, 0.219251,
   0.220143, 0.220143, 0.220588, 0.220588, 0.221034, 0.221034, 0.221480, 0.221480, 0.221480, 0.221925,
   0.221925, 0.222371, 0.222816, 0.222816, 0.222816, 0.223262, 0.223708, 0.223708, 0.224153, 0.223708,
   0.224599, 0.224153, 0.224599, 0.224599, 0.224599, 0.225045, 0.224599, 0.225490, 0.225490, 0.225490,
   0.225936, 0.225936, 0.226381, 0.226381, 0.226381, 0.226827, 0.226827, 0.226827, 0.226827, 0.227718,
   0.227273, 0.228164, 0.227718, 0.228164, 0.228164, 0.228610, 0.228164, 0.228610, 0.228610, 0.229055,
   0.229055, 0.229055, 0.229055, 0.229501, 0.229055, 0.229947, 0.229501, 0.229947, 0.229947, 0.229947
  ]);
}

%
% Code from src/slang/date/GMST.sl
%
define GMST()
%!%+
%\function{GMST}
%\synopsis{Calculates the Greenwich sidereal time and the local siderial time}
%\usage{Double_Type GST = GMST(Double_Type JD);}
%
%\qualifiers{
%\qualifier{hour}{Return the GMST in hours, not in seconds.}
%\qualifier{dut1}{The difference UT1-UTC in seconds (see IERS Bull. A).}
%\qualifier{mjd}{Argument is in MJD, not in JD.}
%\qualifier{gast}{Return the Greenwhich Apparent Siderial Time, GAST.}
%\qualifier{era}{Return the Earth rotation angle (in rotations).}
%\qualifier{lst}{Return the local mean siderial time LST.}
%\qualifier{lon}{longitude for the calculation of the LST (east is positive;
%               in geodetic [ITRS] coordinates, for
%               practical purposes these are identical to the
%               WGS-84 coordinates given by GPS).}
%\qualifier{lat}{latitude for the calculation of the LST (north is positive,
%                  only needed if xp and yp are given)}
%\qualifier{deg}{lon and lat are in degrees, not in radian.}
%\qualifier{xp}{x-coordinate of the Celestial Intermediate Pole (CIP) 
%               ALWAYS in arcseconds, as given by the IERS bulletin.}
%\qualifier{yp}{y-coordinate of the CIP, ALWAYS in arcseconds.}
%}
%\description
%    This code calculates the Greenwich Mean Siderial Time and the
%    Greenwhich Apparent Siderial Time in seconds.
%
%    The argument \code{JD} is the Julian Date in the UT1 time system.
%    For most applications, one can take UT1=UTC. For the most precise
%    applications, a correction term  DUT1=UT1-UTC can be looked up in
%    IERS Bulletin A
%    (see https://www.iers.org/IERS/EN/Publications/Bulletins/bulletins.html).
%
%    The algorithms used are described by Kaplan (2005, USNO Circular
%    179, Sect. 2.6.2).
%
%    This routine is array safe.
%
%    Note that contrary to the coordinate transform routines the
%    deg qualifier only affects the interpretation of the lon, lat
%    qualifiers and does NOT affect the other arguments or return
%    arguments!
%\seealso{equation_equinoxes}
%!%-
{
  variable JD;
  switch(_NARGS)
  { case 1: JD = (); }
  { help(_function_name()); return; }

  if (qualifier_exists("mjd")) {
      JD=mjd2jd(JD);
  }
  JD+=qualifier("dut1",0.)/86400.;

  variable fra=JD mod 1.;
  variable du=JD-2451545.0;
  
  % Kaplan, eq. 2.11.
  % Do NOT change the parentheses!
  variable era=(0.7790572732640+fra)+0.00273781191135448*du;

  if (qualifier_exists("era")) {
      return era;
  }


  variable T;
  if (typeof(JD)==Array_Type) {
      T=Double_Type[length(JD)];
      variable ndx1;
      variable ndx=where(JD>=2443144.5,&ndx1);
      if (length(ndx)>0) {
	  T[ndx]=(tai2tt(utc2tai(JD[ndx]))-2451545.0)/36525.;
      }
      %FIXME: we're assuming that ut=tt before 1977 Jan 1.
      %this is not entirely correct, check Meeus here!
      if (length(ndx1)>0) {
	  T[ndx1]=(tai2tt(JD[ndx1])-2451545.0)/36525.;
      }
  } else {
      if (JD>=2443144.5) {
	  T=(tai2tt(utc2tai(JD))-2451545.0)/36525.;
      } else {
	  T=(tai2tt(JD)-2451545.0)/36525.;
      }
  }
  
  % Kaplan, eq. 2.12
  variable gmst=86400.0*era+
                (((((-0.0000000368*T-0.000029956)*T
                -0.00000044)*T+1.3915817)*T+4612.156534)*T+0.014506)/15.;

  if (qualifier_exists("lst")) {
      variable lon=qualifier("lon",0.);
      variable lat=qualifier("lat",0.);
      if (qualifier_exists("deg")) {
	  lon*=PI/180.;
	  lat*=PI/180.;
      }

      variable xp=qualifier("xp",0.);
      variable yp=qualifier("yp",0.);
      if (xp!=0. or yp!=0.) {
	  variable sl,cl;
	  (sl,cl)=sincos(lon);
	  lon+=(xp*sl+yp*cl)*tan(lat)/3600.;
      }
      % ...equation for longitude in rad
      % gmst=gmst+(3600./15.)*lon*180./PI;
      gmst=gmst+43200*lon/PI;
  }

  if (qualifier_exists("gast")) {
      variable equ=equation_equinoxes(JD;seconds);
      gmst+=equ;
  }

  
  gmst=pos_modulo(gmst,86400.);

  if(qualifier_exists("hour")) {
      return gmst/3600.;
  }
  return gmst;
}

%
% Code from src/data/GTI/GTIoverlap.sl
%
define GTIoverlap()
%!%+
%\function{GTIoverlap}
%\synopsis{computes the overlap of an interval [t1, t2] with a set of good time invervals}
%\usage{Double_Type GTIopverlap(Double_Type t1, Double_Type t2, Struct_Type GTI}
%\seealso{getCommonGTIs}
%!%-
{
  variable t1, t2, GTI;
  switch(_NARGS)
  { case 3: (t1, t2, GTI) = (); }
  { help(_function_name()); return; }

  variable i, overlap=0;
  _for i (0, length(GTI.start)-1, 1)
  { variable diff = _min(GTI.stop[i], t2) - _max(GTI.start[i], t1);
    if(diff>0)  overlap += diff;
  }
  return overlap;
}

%
% Code from src/slang/astronomy/GhoshLamb79.sl
%
%%%%%%%%%%%%%%%%%%%%%
define GhoshLamb79()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{GhoshLamb79}
%\synopsis{calculates the spin-up of a neutron star after Ghosh & Lamb (1979)}
%\usage{Double_Type GhoshLamb79(Double_Type pulse_period; disk);
% or Double_Type GhoshLamb79(Double_Type pulse_period; wind);}
%\qualifiers{
%    \qualifier{fastness}{hook into the issue of a fastness parameter >=0.9. If
%               this qualifier is set to "limit" then the fastness is
%               set to 0.9 in this case. If set to a reference
%               function (Ref_Type) then a hook function of the form
%                 Double_Type fastness_hook(fastness, P, L, M, R, B)
%               is called, which is given the computed 'fastness' and
%               all neutrons star parameters. It has to return the new
%               value for the fastness.}
%    \qualifier{neutron star parameters:}{The qualifiers L, M, R, B, mu, V0,
%      cs, Vorb, a, xi (see the description for details)}
%}
%\description
%    This implements the accretion torque theory by
%      Ghosh & Lamb, 1979, ApJ 234, 296
%    (here after GL79) which calculates the spin-up of a neutron star
%    spinning with a period of 'pulse_period' (given in seconds). The
%    returned Pdot will be in seconds per seconds.
%
%    The theory depends on the physical parameters of the neutron star
%    and the accretion mode. All of these are controlled by qualifiers.
%
%    Two accretion modes are available (set as corresponding qualifier):
%      disk - accretion from an accretion disk (Eq. 15 of GL79)
%      wind - accretion from a stellar wind (Eq. 24 of GL79)
%
%    Three neutron star parameters are used independently from the mode:
%      L - the X-ray luminosity (in 10^37 erg/s; default: 1)
%      M - the neutron star mass (in solar masses; default: 1.4)
%      R - the neutron star radius (in km; default: 11.5)
%
%    For the disk-mode the magnetic field strength is needed. One of
%    the following quantities has to be given:
%      mu - magnetic moment (in 10^30 G cm^3; default: 1)
%      B  - surface magnetic field strength (in 10^12 G; overwrites mu)
%    For the wind-mode several parameters are needed:
%      V0   - the capture velocity (see Eq. 20 of GL79) (in km/s;
%             default: 1000)
%      cs   - the sound speed at the capture radius (in km/s; default: 0)
%      Vorb - the orbital velocity (in km/s; default related to:
%             V0^2 = Vorb^2 + Vwind^2 + cs^2 assuming Vwind = Vorb)
%      a    - the binary separation (in lt-s; default: 100)
%      xi   - "on the order of unity" (see Eq. 22 of GL79; default: 1)
%
%    The function should be vectorized, i.e., an array of pulse
%    periods can be given or the qualifiers can be arrays.
%
%    In case of disk accretion, the so-called fastness parameter is
%    computed after Eq. 16 of GL79. This equation is accurate by 5% for
%    fastnesses <0.9. For larger values a warning message is displayed.
%    For fastnesses >=1.0 the solution is a complex number and therefore
%    a _NaN value is returned and an error message is displayed. The
%    'fastness' qualifier can be used to hook into this issue.
%\seealso{pulsarGL79}
%!%-
{
  % pulse period
  if (_NARGS != 1) { help(_function_name); return; }
  variable P = ();
    
  % accretion type (wind vs. disk)
  variable type;
  if (qualifier_exists("disk")) { type = 1; }
  else if (qualifier_exists("wind")) { type = 2; }
  else {
    help(_function_name);
    return;
  }

  % neutron star parameters
  variable L = qualifier("L", 1.); % in 1e37 erg/s
  variable M = qualifier("M", 1.4);  % in solar masses
  variable R = .1 * qualifier("R", 11.5); % in 1e6 cm

  % moment of intertia
  variable I = 2./5 * (M*1.989) * R^2; % in 1e45 g cm^2

  % disk accretion
  if (type == 1) {
    % magnetic moment
    variable mu = qualifier("mu", 1.);
    if (qualifier_exists("B")) {
      variable B = qualifier("B"); % in 1e12 G
      mu = .5 * B * R^3; % in 1e30 G/cm^3
    }

    % first structure function (eq. 17)
    variable S1 = R^(6./7)*M^(-3./7)/I;

    % second structure function (eq. 18)
    variable S2 = R^(-3./7)*M^(-2./7);

    % fastness parameter (eq. 16)
    variable omegas = 1.35*mu^(6./7)*S2*(P*L^(3./7))^(-1);
    if (any(omegas >= 0.9)) {
      % hook
      variable fastness = qualifier("fastness", NULL);
      if (typeof(fastness) == Ref_Type) {
	omegas = (@fastness)(omegas, P, L, M, R, mu);
      }
      else if (typeof(fastness) == String_Type && fastness == "limit") {
	omegas = _min(omegas, 0.9);
      }
      % print warnings/errors
      else {
	if (any(omegas >= 1.0)) {
	  vmessage("error (%s): fastness parameter >= 1.0", _function_name);
	} else {
	  vmessage("warning (%s): fastness parameter >= 0.9", _function_name);
	}
      }
    }
    
    % accretion torque (eq. 10)
    variable n = 1.39*(1.-omegas*(4.03*(1-omegas)^0.173-0.878))/(1.-omegas);

    % spin up (eq. 15)
    return -5e-5*mu^(2./7)*n*S1*(P*L^(3./7))^2 / 365/86400;
  }

  % wind accretion
  if (type == 2) {
    variable V0 = qualifier("V0", 1000); % in km/s
    variable xi = qualifier("xi", 1.); % "on the order of unity" (GL79)
    % capture velocity (text after eq. 20): V^2 = Vorb^2 + Vwind^2 + cs^2
    variable cs = qualifier("cs", 0); % sound speed at capture radius
    % assuming that Vorb and Vwind are equal:
    variable vorb = qualifier("Vorb", sqrt(.5*(V0^2-cs^2))); % in km/s
    variable a = qualifier("a", 100)*(Const_c*1e-10); % in 10^10 cm

    % capture radius (eq. 22)
    variable ra = xi*2*(Const_G*1e8)*(M*1.989)/V0^2*1e5; % in 10^10 cm

    % specific angular moment (eq. 21)
    variable la = .5*vorb*ra^2/a * 1e-2; % in 1e17 cm^2/s

    % spin up (eq. 24)
    return -3.8e-5/R/I*la*P^2*L / 365/86400;
  }
}

%
% Code from src/slang/date/GreenwichSiderealTime_from_MJD.sl
%

define GreenwichSiderealTime_from_MJD()
%!%+
%\function{GreenwichSiderealTime_from_MJD}
%\synopsis{computes the Greenwich sidereal time}
%\usage{Double_Type GST = GreenwichSiderealTime_from_MJD(Double_Type MJD);}
%\description
%    \code{MJD} is the Modified Julian Date (*in UT*).
%    This routine is a wrapper around GMST(JD), and only
%    kept for compatibility reasons.
%    This function is DEPRECATED. Please do not use for future work.
%\seealso{GMST}
%!%-
{
  variable MJD;
  switch(_NARGS)
  { case 1: MJD = (); }
  { help(_function_name()); return; }

  return(GMST(MJD;mjd,hour));
}

%
% Code from src/slang/math/Hammer_projection.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define Hammer_projection()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Hammer_projection}
%\synopsis{Computes the Hammer-Aitoff projection}
%\usage{(Double_Type x, y) = Hammer_projection(Double_Type l, b);}
%\qualifiers{
%\qualifier{deg}{\code{l} and \code{b} are in degrees, not in radian}
%\qualifier{normalized}{\code{x} and \code{y} are normalized (by \code{sqrt(2)})
%                 such that \code{abs(x) <= 2} and \code{abs(y) <= 1}.}
%\qualifier{astronomical}{flip x-axis for astronomical maps, where east is
%                 to the left}
%\qualifier{inverse}{calculate the inverse projection, interpreting l as the x- and
%               b as the y-coordinate; all other qualifiers are 
%               also interpreted as expected.}
%}
%\description
% This is the projection erroneously called the Aitoff projection by many
% astronomers. It is an equal area projection. The projection equations are
%
%    \code{x = 2 * sqrt(2) * cos(b) * sin(l/2) / sqrt(1 + cos(b)*cos(l/2));}\n
%    \code{y = sqrt(2) * sin(b) / sqrt(1 + cos(b)*cos(l/2));}\n
%
% The inverse function returns nan if the arguments given are not possible.
%
%\seealso{Aitoff_projection, Lambert_Equal_Area_projection}
%!%-
{
  variable l, b;
  switch(_NARGS)
  { case 2: (l, b) = (); }
  { help(_function_name()); return; }

  if (qualifier_exists("inverse")) {

      % inverse projection
      variable x=l; 
      variable y=b;
      if (qualifier_exists("astronomical")) {
	  x=-x;
      }
      if (qualifier_exists("normalized")) {
	  x*=sqrt(2.0);
	  y*=sqrt(2.0);
      }

      variable z2=1.-(x/4.)^2.-(y/2.)^2.;
      variable z = sqrt(z2);

      l=2.*atan2(z*x/2.0, 2.*z*z-1.);
      b=asin(y*z);

      if (qualifier_exists("astronomical")) {
       	  if(typeof(l)==Array_Type) {
       	      l[where(l<0)] += 2*PI;
       	  } else { 
	      if (l<0.0) {
		  l+=2*PI;
	      }
	  }
      }

      if (qualifier_exists("deg")) {
	  l*=180./PI;
	  b*=180./PI;
      }

      return l,b;
  }

  if (qualifier_exists("deg") ) {
      l *= PI/180;
      b *= PI/180;
  }

  if(typeof(l)==Array_Type) {
      while(any(abs(l)>PI)) {
   	  l[where(l> PI)] -= 2*PI;
   	  l[where(l<-PI)] += 2*PI;
      }
  } else { 
      while(l> PI)  l-= 2*PI;
      while(l<-PI)  l+= 2*PI;
  }

  variable sinb,cosb;
  (sinb,cosb)=sincos(b);

  variable sinl2,cosl2;
  (sinl2,cosl2)=sincos(l/2.0);

  variable w = 1./sqrt(0.5+0.5*cosb*cosl2);
  if (qualifier_exists("normalized"))  {
      w = w/sqrt(2.0);
  }

  variable astro=2.;
  if (qualifier_exists("astronomical")) {
      astro=-2.;
  }

  return (astro*cosb*sinl2*w,sinb*w);

}

%
% Code from src/plot/ImageMagick_text.sl
%

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

define ImageMagick_text(txt)
{
  variable tmpfile = "/tmp/ImageMagick_txt_tmp.png";
  variable size = qualifier("size", 12)*4/3;
  variable cmd = "convert";
  cmd += sprintf(" -size %dx%d", 8192, 3*size);
  cmd += " xc:white";
  cmd += sprintf(" -pointsize %d", size);
  cmd += sprintf(` -draw "text %d,%d '%s'"`, size, 2*size, txt);
  cmd += " -colors 2";
  cmd += " "+tmpfile;
  ()=system(cmd);
  variable i = png_read_flipped(tmpfile) < 0xFFFF;
  remove(tmpfile);
  return i[rectangle_where(i)];
}


#endif

%
% Code from src/slang/date/JDofDate.sl
%
define JDofDate()
%!%+
%\function{JDofDate}
%\synopsis{Calculates the (modified) Julian Date}
%\usage{Double_Type JD = JDofDate( struct { year, month, day, hour, minute, second } );
%\altusage{Double_Type JD = JDofDate(year, month, day[, hour[, minute[, second]]]);}
%}
%\qualifiers{
%\qualifier{mjd}{return the MJD, i.e., JD-2400000.5}
%\qualifier{julian_calendar}{return the JD/MJD assuming the
%       Julian calendar rather than the Gegorian one}
%}
%\description
%    This routine calculates the Julian Date for a given Gregorian date
%    following the routine by Fliegel and van Flandern (1968, Comm. ACM 11, 657)
%    and further optimized as discussed by Pulkkinen and van Flandern (1979,
%    ApJ Suppl 41: 391). With the julian_calendar qualifier the date is
%    calculated for the Julian calendar (should be done for dates before
%    15 October 1582 in Germany, see
%    https://en.wikipedia.org/wiki/Gregorian_calendar
%    for a rough table and
%    https://de.wikipedia.org/wiki/Gregorianischer_Kalender
%    for a table containing the exact dates for the switch over).
%
%    The date is either given by a structure or by at least 3 integer arguments.
%    Fractional parts for the year and month are ignored. The default values for
%    hour, minute and second are 0, fractional days, hours, etc. are properly
%    taken into account.
%
%    The year must be larger than -4716 (4717 BC).
%    This function is array safe.
%
%\seealso{MJDofDate,dateOfMJD}
%!%-
{
  variable year, month, day, hour=0, minute=0, second=0.;
  switch(_NARGS)
  { case 1:
      variable date = ();
      if(typeof(date)==Struct_Type)
      { year   = date.year;
        month  = date.month;
        day    = date.day;
        hour   = date.hour;
        minute = date.minute;
        second = date.second;
      }
      else
      { help(_function_name());
        return;
      }
  }
  { case 3: (year, month, day) = (); }
  { case 4: (year, month, day, hour) = (); }
  { case 5: (year, month, day, hour, minute) = (); }
  { case 6: (year, month, day, hour, minute, second) = (); }
  { help(_function_name()); return; }

  variable Y=typecast(year,Integer_Type);
  variable M=typecast(month,Integer_Type);
  variable D=day;

  variable jd;
  if (qualifier_exists("julian_calendar")) {
      if (M<3) {
	  Y+=-1;
	  M+=12;
      }
      variable A=int(Y/100);
      jd=int(365.25*(Y+4716))+int(30.6001*(M+1));
      jd+=( (qualifier_exists("mjd")) ? -2401525.5 : -1524 );
  } else {
      %jw do not even THINK about modifying the following lines.
      %jw the expression HAS to be done in INTEGER arithmetic, so we
      %jw better force this here
      jd=367*Y-7*(Y+(M+9)/12)/4-3*((Y+(M-9)/7)/100+1)/4+275*M/9;
      jd+=(  (qualifier_exists("mjd")) ? -678971.5 : 1721029);
  }
  
  % do this last to minimize round off errors
  jd=jd-0.5+D+(hour+(minute+second/60.)/60.)/24.;
  
  return jd;
}

%
% Code from src/slang/date/JDofEpoch.sl
%
define JDofEpoch()
%!%+
%\function{JDofEpoch}
%\synopsis{Returns the (M)JD of a Julian or Besselian epoch such as J2000.0}
%\usage{jd=JDofEpoch(epochstring;qualifiers)}
%\qualifiers{
%\qualifier{mjd}{Return the modified julian date (default: JD)}
%}
%\description
%   This function returns the julian date corresponding to a
%   Julian or Besselian epoch (e.g., J2000, B1950.0 etc.). For highest
%   precision, the most commonly used values are hardcoded, other values
%   are calculated using the formulae given by Lieske (1979, A&A 73, 282)
%   for Besselian dates and the standard definition of the Julian epoch
%   (based on Julian years with 365.25 days).
%
%   The epoch string must start with a "B" for a Besselian epoch and
%   "J" for a Julian epoch, followed by a number that designates the year
%   fractional values such as 2017.25 are possible.
%
%   Note that the time system in which the dates are defined is TT, for most
%   applications that do not require microsecond accuracy for the conversion
%   of positions the precise time system won't matter.
%
%   If the routine is called with a floating point number, it is assumed
%   that this number is the epoch and this number is returned. This greatly
%   simplifies functions in which the epoch is either a string or
%   a number.
%
%\seealso{EpochofJD,tt2tai,tt2tdb}
%!%-
{
    variable ep;

    switch(_NARGS)
    { case 1: ep=(); }
    { help(_function_name());return;}

    % if called with a number: do nothing
    if (__is_numeric(ep)) {
	return typecast(ep,Double_Type);
    }

    % deal with the most common cases first
    if (qualifier_exists("mjd")) {
	switch(ep)
	{ case "J2000.0": return 51544.5;}
	{ case "B1950.0": return 33281.9235;}
	{ case "B1900.0": return 15019.81352;}
    } else {
	switch(ep)
	{ case "J2000.0": return 2451545.0;}
	{ case "B1950.0": return 2433282.4235;}
	{ case "B1900.0": return 2415020.31352;}
    }
    
    %
    % general epoch string -> decode the string
    variable type=substr(ep,1,1);
    variable year=atof(substr(ep,2,strlen(ep)));

    if(type!="J" && type !="B") {
	throw UsageError,sprintf("%s: epoch must start with J or B\n",_function_name());
    }
    
    if(type=="J") {
	% Julian epoch
	if (qualifier_exists("mjd")) {
	    return (year-2000.0)*365.25+51544.5;
	} else {
	    return (year-2000.0)*365.25+2451545.0;
	}
    } else {
	if (qualifier_exists("mjd")) {
	    return (year-1900.) * 365.242198781 + 15019.81352;
	} else {
	    return (year-1900.) * 365.242198781 + 2415020.31352;
	}
    }
}

%
% Code from src/misc/math/KS_test.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define KS_test(x1,x2)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{KS_test}
%\synopsis{computes the test statistics of a two sample Kolmogorov-Smirnov test}
%\usage{test_statistics = KS_test(Double_Type x1, Double_Type x2);}
%\description
% The null hypothesis of the KS test is that two samples are
% distributed according to the same distribution. It is rejected if
% the test statistic D=max(F_1(x),F_2(x)) is greater than a certain
% value
%!%-
{
   % sort distributions
   variable y1=x1[array_sort(x1)];
   variable y2=x2[array_sort(x2)];
   variable n1=length(x1);
   variable n2=length(x2);   
   % get empirical distribution function (step function)
   variable F1=1.0*[1:n1:1]/n1;
   variable F2=1.0*[1:n2:1]/n2;
   % sort the two distributions with respect to each other
   variable s=array_sort([y1,y2]);
   variable Fe1=Double_Type[n1+n2+1];
   variable Fe2=Double_Type[n1+n2+1];
   Fe1[*]=-1;
   Fe2[*]=-1;
   Fe1[0]=0;
   Fe2[0]=0;
   variable int=[1:n1+n2:1];
   Fe1[int[where(s<n1)]]=F1;
   Fe2[int[where(s>n1-1)]]=F2;
   % there is a -1  everywhere where the other function changes, there
   % the value should be the last not -1 value (because this function
   % stays constant):
   variable m=where(Fe1==-1);
   while (length(m)>0)
     {
	Fe1[m]=Fe1[m-1];
	m=where(Fe1==-1);
     }
   m=where(Fe2==-1);
   while (length(m)>0)
     {
         Fe2[m]=Fe2[m-1];
	 m=where(Fe2==-1);
     }
   % return the maximum of the difference
   return max(abs(Fe1-Fe2));
}
%%%%%%%%%%%%%%%%%%%%

%
% Code from src/slang/astronomy/KeplerEquation.sl
%
define KeplerEquation()
%!%+
%\function{KeplerEquation}
%\synopsis{solves Kepler's Equation  E - e sin E  =  M }
%\usage{Double_Type E = KeplerEquation(Double_Type M, e)}
%\description
%    Solve Kepler's Equation by using the result of the method
%    by S. Mikkola (1987) Celestial Mechanics, 40, 329-334
%    as starting value for a Newton-Raphson iteration
%    to extend the applicability of this function to higher eccentricities.
%\qualifiers{
%\qualifier{threshold}{stopping criterion for the Newton Raphson iteration, default = 1e-5}
%}
%!%-
{
  variable M, e;
  switch(_NARGS)
  { case 2: (M, e) = (); }
  { help(_function_name()); return; }
  M = [M];

  variable threshold = qualifier("threshold", 1e-5);
  if(e<0 || e>=1)
  { message("error (%s): eccentricity must be between 0 and 1");
    return;
  }

  % range reduction  =>  -pi < M < pi
  M = M mod (2*PI);
  M[where(M<PI)] += 2*PI;
  M[where(M>PI)] -= 2*PI;

  % bail out for circular orbits
  if(e==0)  return M;

  variable aux = 4*e + 0.5;
  variable alpha = (1-e)/aux;
  variable beta = M/(2*aux);
  aux = sqrt(beta^2 + alpha^3);
  variable z = beta + aux;
  variable i = where(z <= 0.0);
  z[i] = beta[i] - aux[i];
  z = abs(z)^(1./3);
  variable s0 = z - alpha/z;
  variable s1 = s0 - (0.078*s0^5)/(1.+e);
  variable e0 = M + e*(3.*s1 - 4.*s1^3);
  variable se0 = sin(e0);
  variable ce0 = cos(e0);
  variable f = e0 - e*se0 - M;
  variable f1 = 1. - e*ce0;
  variable f2 = e*se0;
  variable f3 = e*ce0;
  variable f4 = -f2;
  variable u1 = -f/f1;
  variable u2 = -f/(f1 + 0.5*f2*u1);
  variable u3 = -f/(f1 + 0.5*f2*u2 + f3*u2*u2/6.);
  variable u4 = -f/(f1 + 0.5*f2*u3 + f3*u3*u3/6. + f4*u3^3/24.);
  variable eccanom = e0 + u4;
  eccanom[where(eccanom >= 2*PI)] -= 2*PI;
  eccanom[where(eccanom < 0)]     += 2*PI;

  % Now get more precise solution using Newton Raphson method
  % for those times when the Kepler equation is not yet solved
  % to better than 1e-10 (modification J. Wilms)
  M[where(M<0.)] += 2.*PI;
  foreach i (where(M != 0 and abs(eccanom - e*sin(eccanom) - M) > 1e-10))
  { if(qualifier_exists("verbose"))
      message("Newton Raphson");
    do
    {
      variable fe = eccanom[i] - e*sin(eccanom[i]) - M[i];  % f  = E - e sin E - M
      variable fs = 1. - e*cos(eccanom[i]);              ;  % f' = 1 - e cos E
      variable oldval = eccanom[i];
      eccanom[i] = oldval - fe/fs;
    } while(abs(oldval-eccanom[i]) > threshold);
    % the following should be coded more intelligently ;-)
    % (similar to range reduction of mx...)
    while(eccanom[i] >= PI)  eccanom[i] -= 2*PI;
    while(eccanom[i] <  PI)  eccanom[i] += 2*PI;
  }

  if(length(eccanom)>1)
    return eccanom;
  else
    return eccanom[0];
}

%
% Code from src/slang/math/Lambert_Equal_Area_projection.sl
%
define Lambert_Equal_Area_projection(phi,theta) {
%!%+
%\function{Lamberg_Equal_Area_projection}
%\synopsis{Computes the Lambert_Equal_Area_projection}
%\usage{(Double_Type x, y) = Lambert_Equal_Area_projection(Double_Type phi,theta);}
%\qualifiers{
%\qualifier{deg}{\code{phi} and \code{theta} are in degrees, not in radian}
%\qualifier{inverse}{calculate the inverse projection, interpreting phi as the x- and
%               theta as the y-coordinate; all other qualifiers are 
%               also interpreted as expected.}
%}
%\description
% This is an equal area projection centered on the North Pole (when plotting
% southern objects, flip the sign of theta!). phi is the angle along the 
% equator, theta is measured from the equator.
%
%    \code{R =sin((pi/2 - theta)/2);}\n
%    \code{x = R * sin(phi);}\n
%    \code{y = R * cos(phi);}\n
%
% The inverse function returns nan if the arguments given are not possible.
%
%\seealso{Aitoff_projection, Hammer_projection}
%!%-

    if (qualifier_exists("inverse")) {
	variable x=phi;
	variable y=theta;

	phi=atan2(x,-y);
	variable rtheta=sqrt(x*x+y*y);
	theta=PI/2 - 2*asin(rtheta/2.);

	if (qualifier_exists("deg")) {
	    phi*=180./PI;
	    theta*=180./PI;
	}
	return(phi,theta);
    }


    if (qualifier_exists("deg")) {
	phi  *=PI/180.0;
	theta*=PI/180.0;
    }

    if(typeof(phi)==Array_Type) {
	while(any(abs(phi)>PI)) {
	    phi[where(phi> PI)] -= 2*PI;
	    phi[where(phi<-PI)] += 2*PI;
	}
    } else { 
	while(phi> PI)  phi-= 2*PI;
	while(phi<-PI)  phi+= 2*PI;
    }

    rtheta=sin((PI/2.-theta)/2.);
    
    variable sinp,cosp;
    (sinp,cosp)=sincos(phi);

    return (rtheta*sinp, rtheta*cosp);

}

%
% Code from src/fitting/List_par.sl
%
% -*- mode:slang; mode:fold -*-

private variable Fmt_Colheads, Fmt_Params;
Fmt_Colheads = " idx%-*stie-to  freeze         value         min         max";
Fmt_Params   = "%3u  %-*s  %2d    %2d   %14.7g  %10.7g  %10.7g  %s%s";
private define param_string (p, len) %{{{
{
  return sprintf (Fmt_Params, p.index, len, p.name,
		  p.tie, p.freeze, p.value, p.min, p.max, p.units, p.fun);
}
%}}}
private define back_fun_string (i) %{{{
{
  variable s = _isis->_get_instrumental_background_hook_name(i);
  if (s == NULL) return "";
  return sprintf ("# [bgd %d] = %s", i, strtrim(s));
}
%}}}
private define massage_param_structs (pars) %{{{
{
  variable len = 0;

  foreach (pars)
  {
    variable p = ();

    len = max([len, strlen(p.name)]);

    if (p.tie != NULL)
    {
      p.tie = _get_index(p.tie);
    }
    else p.tie = 0;

    if (p.fun != NULL)
    {
      p.fun = sprintf ("\n#=>  %s", strtrim(p.fun));
    }
    else p.fun = "";

    if (andelse
      {p.min == -_isis->DBL_MAX}
	{p.max ==  _isis->DBL_MAX})
    {
      p.min = 0;
      p.max = 0;
    }
  }

  return len;
}

%}}}
private define is_free (p) %{{{
{
  return (p.tie == NULL and p.freeze == 0 and p.fun == NULL);
}
%}}}

private define list_par_string_pattern (pat) %{{{
{
  variable s = get_fit_fun();
  s = (s == NULL) ? String_Type[0] : [strtrim(s)];

  variable datasets = all_data();
  if (datasets != NULL)
  {
    variable bs = array_map (String_Type, &back_fun_string, datasets);
    bs = bs[where(bs != "")];
    s = [s, bs];
  }

  variable p, _i, i = Integer_Type[0];
  foreach p ([pat])
  {
    _i = _get_index(p);
    if( typeof(_i) == Array_Type )
      i = [i, _i ];
  }
  
  variable ps = String_Type[0];
  
  variable pars = get_params(i);
  if (pars != NULL)
  {
    variable len = massage_param_structs (pars);
    
    variable colheads;
    colheads = sprintf (Fmt_Colheads, len, "  param");
    ps = array_map (String_Type, &param_string, pars, len);
    
    s = strjoin ([s, colheads, ps], "\n");
  }
  
  if (length(s) == 0)
    return NULL;
  else if (length(s) == 1 && typeof(s) == Array_Type)
    s = s[0];

  return s;
}
%}}}

define List_par()
%!%+
%\function{List_par}
%\synopsis{list current fit function and parameters}
%\usage{List_par ([arg])}
%\altusage{List_par ( String_Type[] pattern )}
%\description
%   The optional argument is used to redirect the output.  If arg
%   is omitted, the output goes to stdout.  If arg is of type
%   Ref_Type, it the output string is stored in the referenced
%   variable.  If arg is a file name, the output is stored in that
%   file.  If arg is a file pointer (File_Type) the output is
%   written to the corresponding file.
%
%   If the given argument is a String pattern, only parameter matching
%   this pattern are listed!
%
%   The parameter listing looks like this:
%
%   gauss(1) + poly(1)
%     idx  param        tie-to  freeze    value      min          max
%      1  gauss(1).area     0     0       103.6        0            0
%      2  gauss(1).center   0     0        12.1       10           13
%      3  gauss(1).sigma    0     0       0.022    0.001          0.1
%      4  poly(1).a0        0     0       1.2e4        0            0
%      5  poly(1).a1        0     1           0        0            0
%      6  poly(1).a2        0     1           0        0            0
%
%    The first line defines the form of the fit-function. The
%    parameter index idx may be used to refer to individual fit
%    parameters (see set_par).  freeze = 1 (0) indicates that the
%    corresponding parameter value is frozen (variable).  If two
%    parameter values are tied together, the connection is indicated
%    in the tie-to column.  For example, if parameter 1 has tie-to =
%    5, that means the value of parameter 1 is tied to the value of
%    parameter 5; if parameter 5 changes, parameter 1 will follow
%    the change exactly.  If min=max=0, the corresponding parameter
%    value is unconstrained.
%
%    In input parameter files (see load_par), lines beginning with a
%    '#' are mostly ignored and may be used to include comments.
%    Exceptions to this rule are "special" comment lines which are
%    used to support additional functionality such as, e.g. writing
%    some parameters as functions of other parameters (see
%    set_par_fun).  Note that, aside from these special cases,
%    comment lines are not loaded by load_par and will not be
%    preserved if file is later overwritten by save_par.
%
%\seealso{list_free, edit_par, set_par, get_par, save_par, set_par_fun}
%!%-
{
  variable pat;
  switch( _NARGS )
  { case 0 : list_par;return; }
  { case 1 : pat = ();}
  { help(_function_name()); return; }

  if ( _typeof(pat) == String_Type ){
    _isis->put_string (NULL, _function_name, list_par_string_pattern (pat) );

    % % get par list as string array (lines)
    % variable l;
    % list_par(&l);
    % l = strtok( l, "\n" );
    % % get line indices matching the 'pat'
    % variable pnames = String_Type[0];
    % variable p;
    % _for p ( 2, length(l)-1, 1 ){
    %   pnames = [ pnames, strtok(l[p])[1] ];
    % }
    % variable idx = where( is_substr( pnames, pat ) != 0 ) + 2;

    % sprintf( strjoin( l[[0,1,idx]], "\n" ) );
  }
  else
    list_par( pat );
  return;
}

%
% Code from src/misc/satellites/MAXI_lightcurve.sl
%
define MAXI_lightcurve()
%!%+
%\function{MAXI_lightcurve}
%\usage{Struct_Type MAXI_lightcurve(String_Type source);}
%\description
%    The returned structure has the following fields:\n
%    - \code{time}: Modified Julian Date\n
%    - \code{rate}  , \code{err}  :  2-20 keV light curve [c/s/cm^2]\n
%    - \code{rate_a}, \code{err_a}:  2- 4 keV light curve [c/s/cm^2]\n
%    - \code{rate_b}, \code{err_b}:  4-10 keV light curve [c/s/cm^2]\n
%    - \code{rate_c}, \code{err_c}: 10-20 keV light curve [c/s/cm^2]\n
%\seealso{http://maxi.riken.jp/top/index.php?cid=000000000036, http://ads.nao.ac.jp/abs/2009PASJ...61..999M}
%!%-
{
  variable source;
  switch(_NARGS)
  { case 1: source = (); }
  { help(_function_name()); return; }

  variable path_to_data = NULL;
  try {
  foreach path_to_data ([local_paths.MAXI_lightcurves])
    if(stat_file(path_to_data)!=NULL) { break; }
  } catch NotImplementedError: { vmessage("warning (%s): local_paths.MAXI_lightcurves not defined",_function_name()); };
  if(_isnull(path_to_data) || stat_file(path_to_data)==NULL)
  { vmessage("error (%s): no path to MAXI lightcurves found", _function_name());
    return NULL;
  }

  variable lc = struct { time, rate, err, rate_a, err_a, rate_b, err_b, rate_c, err_c };
  set_struct_fields(lc, readcol(path_to_data+source+".txt", 1, 2,3, 4,5, 6,7, 8,9));
  return lc;
}

%
% Code from src/slang/date/MJD2UNIXtime.sl
%
%%%%%%%%%%%%%%%%%%%
define MJD2UNIXtime()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{MJD2UNIXtime}
%\synopsis{converts modified Julian date to UNIX time}
%\usage{Long_Type MJD2UNIXtime(Double_Type MJD)}
%\description
%    Time formats and conversion functions:
%      Long_Type secs = MJD2UNIXtime(Double_Type MJD);
%      Double_Type MJD = UNIXtime2MJD(Long_Type secs);
%
%      Struct_Type tm = localtime(Long_Type secs);  % or gmtime
%      Long_Type secs = mktime(Struct_Type tm);
%
%      String_Type str = strftime(String_Type fmt, Struct_Type tm);
%
%    Current time and date:
%      Long_Type secs = _time();
%      String_Type str = time();
%\seealso{UNIXtime2MJD, localtime, gmtime, mktime}
%!%-
{
  variable MJD;
  switch(_NARGS)
  { case 1: MJD = (); }
  { help(_function_name()); return; }

  % note: we need to first typecast MJD to a 64bit floating point
  % number (Double_Type == Float64_Type) in order to have the
  % calculation performed in 64bit. If, e.g., MJD is given of
  % Integer_Type == Int32_Type, then some dates might exceed
  % INT32_MAX because the calculation would be performed in 32bit.
  % Finally, because UNIXtime is an integer number, we need to
  % convert the result to an integer using 64bit, i.e., typecast
  % to Long_Type == Int64_Type
  return typecast((typecast(MJD, Float64_Type)-40587)*86400, Int64_Type);
}

%
% Code from src/slang/date/MJD2fermi.sl
%
%%%%%%%%%%%%%%%%
define MJD2fermi()
%%%%%%%%%%%%%%%%
%!%+
%\function{MJD2fermi}
%\synopsis{calculate Fermi MET used for LC and spectral analysis}
%\usage{Double_Type = MJD2fermi (Double_Type);}
%\description
%	Calculates the Fermi Mission Elapsed Time (MET)
%	in seconds for a given MJD(UTC).
%\example
%	variable m = 54900.345;
%	variable fermi_s = MJD2fermi(m);
%\seealso{fermi2MJD, MJDref_satellite}
%!%-
{ 
  variable mjd_in;
  
  switch(_NARGS)
  { case 1: mjd_in  = (); }
  { help(_function_name()); return; }
  
  variable mjd_ref=51910.0; % MJD of 00:00:00 UTC 2001 January 1
  variable diff1=utc2tai(mjd_ref;mjd,deltat);
  variable diff2=utc2tai(mjd_in;mjd,deltat);
  return ((mjd_in-mjd_ref)*86400.)+diff2-diff1;
}


%
% Code from src/slang/date/MJDofDate.sl
%
define MJDofDate()
%!%+
%\function{MJDofDate}
%\synopsis{calculates the modified Julian date}
%\usage{Double_Type MJD = MJDofDate( struct { year, month, day, hour, minute, second } );
%\altusage{Double_Type MJD = MJDofDate(year, month, day[, hour[, minute[, second]]]);}
%}
%\description
%    The Gregorian or Julian date is either given by a structure or by at
%    least 3 integer arguments. Fractional parts are ignored. 
%    The default values for hour, minute and second are 0, fractional
%    hours etc. are properly taken into account.
%
%    This function is array safe and equivalent to calling JDofDate with the
%    mjd qualifier.
%\qualifiers{
%\qualifier{julian_calendar}{return the JD/MJD for the date in the Julian calendar rather
%       the Gegorian date. This does not make much sense for the MJD, which was negative
%       for the time when the Julian calendar was used.}
%}
%\seealso{JDofDate,dateOfMJD}
%!%-
{
  variable year, month, day, hour=0, minute=0, second=0;
  switch(_NARGS)
  { case 1:
      variable date = ();
      if(typeof(date)==Struct_Type)
      { year   = date.year;
        month  = date.month;
        day    = date.day;
        hour   = date.hour;
        minute = date.minute;
        second = date.second;
      }
      else
      { help(_function_name());
        return;
      }
  }
  { case 3: (year, month, day) = (); }
  { case 4: (year, month, day, hour) = (); }
  { case 5: (year, month, day, hour, minute) = (); }
  { case 6: (year, month, day, hour, minute, second) = (); }
  { help(_function_name()); return; }
  if (qualifier_exists("julian_calendar")) {
      return JDofDate(year,month,day,hour,minute,second;mjd,julian_calendar);
  } 
  return JDofDate(year,month,day,hour,minute,second;mjd);
}

%
% Code from src/slang/date/MJDofDateString.sl
%
define MJDofDateString ()
%!%+
%\function{MJDofDateString}                                                                                         
%\synopsis{calculates the MJD from a string including year, month, day [, hour [, minute [, second]]]}
%\usage{Double_Type \code{mjd} = MJDofDateString(String_Type \code{date_string});}
%\qualifiers{
%\qualifier{verbose}{prints the found values for YYYY-MM-DD hh:mm:ss.sss}
%\qualifier{no_2digit_correction}{see `parseDateString`}
%}
%\description
%    This function calculates the Modified Julian Date from a String
%    using the `parseDateString` function.
%\example
%    MJDofDateString("69/07/21");
%    MJDofDateString("1969-07-21 02:56:30.44");
%    MJDofDateString("sdf1969--X07..21QQ_02/56((30,,44xyz");
%\seealso{parseDateString}
%!%-
{
  variable str;
  switch(_NARGS)
    { case 1: str = (); }
    { help(_function_name()); return;}

  % parse string
  variable tm = parseDateString(str);

  % correct month and year
  tm.tm_mon += 1;
  tm.tm_year += 1900;
  
  if (qualifier_exists("verbose")) {
    vmessage("%s uses YYYY-MM-DD: %04d-%02d-%02d  %02d:%02d:%f",
      _function_name(), tm.tm_year, tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec
    );
  }
  return MJDofDate(tm.tm_year, tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);
}

%
% Code from src/misc/satellites/MJDref_satellite.sl
%
define MJDref_satellite()
%!%+
%\function{MJDref_satellite}
%\synopsis{returns the reference date (MJD) of a satellite's time}
%\usage{Double_Type MJDref_satellite(String_Type satellite)}
%!%-
{
  variable satellite;
  switch(_NARGS)
  { case 1: satellite = (); }
  { help(_function_name()); }

  switch( strlow(satellite) )
  { case "chandra" : return 50814.; }
  { case "xmm"     : return 50814.; }
  { case "integral": return 51544.; }
  { case "suzaku"  : return 51544.00074287037037037; }
  { case "xte"
    or case "rxte" : return 49353.000696574074; }
  { case "swift"   : return 51910.00074287038; }   % at least for a BAT lightcurve
  { case "rhessi"  : return 43874.0; }   % calculated from position data text file
  { case "fermi"   : return 51910.; } % UTC 00:00:00 on 1. January 2001 
  % (fermi, see: http://adsabs.harvard.edu/abs/2010arXiv1002.2280T)
  { case "nustar"  : return 55197.00076601852; } % Jan 1, 2010, in TT time system 
  vmessage(`warning (%s): MJDref of satellite "%s" not known`, _function_name(), satellite);
  return -1;
}

%
% Code from src/slang/date/MJDstrftime.sl
%
define MJDstrftime()
%!%+
%\function{MJDstrftime}
%\synopsis{formats a modified Julian date}
%\usage{String_Type MJDstrftime(String_Type fmt, Double_Type MJD);}
%\seealso{strftime, MJD2UNIXtime, localtime}
%!%-
{
  variable fmt, MJD;
  switch(_NARGS)
  { case 2: (fmt, MJD) = (); }
  { help(_function_name()); return; }

  return strftime(fmt, localtime(MJD2UNIXtime(MJD)));
}

%
% Code from src/Mike/Mike.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

% only activate if module xspec could be loaded
#ifeval __get_reference("load_xspec_local_models")!=NULL
% Last Updated: Aug. 25, 2005

% Force the Appending of semi-colons
% Isis_Append_Semicolon=0;

% Make sure filenames and background file names get printed in list functions
% Isis_List_Filenames=1;

% For reading in radio data.  Make sure minimum allowable error damned
% near 0, since I *think* `Ideal ARFs' do *not* take input from set_data_exposure.
% Minimum_Stat_Err = 1.e-30;

try { require("xspec"); require( "gsl", "gsl" ); }
catch AnyError: { };

static variable path = "";

% A message function I'll use a lot below

public define msg(str_array)
{
   () = printf("\n");
   foreach(str_array)
   {
      variable str = ();
      () = printf(" %s\n", str);
   }
   () = printf("\n");
   return;
}


#endif

#endif

%
% Code from src/slang/physics/N_body_simulation.sl
%
define N_body_simulation_std_kernel()
%!%+
%\function{N_body_simulation_std_kernel}
%\synopsis{Default interaction kernel for the function 'N_body_simulation'}
%\usage{Double_Type r[6,N] = N_body_simulation_std_kernel(Double_Types t, m[6,N]; qualifiers)}
%\description
%   This function is the default interaction kernel of the function 'N_body_simulation'.
%   It computes the equations of motion of N interacting particles at time t.
%   The input parameter 'm' is a [6,N]-matrix with
%      m[0,j] = x_j;
%      m[1,j] = y_j;
%      m[2,j] = z_j;
%      m[3,j] = vx_j;
%      m[4,j] = vy_j;
%      m[5,j] = vz_j;
%
%   For each particle, the potential of particle i (at position (x_i,y_i,z_i)) exerted
%   on particle j (at position (x_j,y_j,z_j)) is a Plummer sphere of the form
%      Phi(x_i,y_i,z_i,x_j,y_j,z_j) = -psa_i*(psb_i+dis^2)^(-1/2)
%   with
%      dis^2 = (x_i-x_j)^2+(y_i-y_j)^2+(z_i-z_j)^2
%   and psa_i and psb_i model constants (see qualifiers).
%   The resulting acceleration of particle j is then
%      d^2/dt^2 x_j = -d/dx_j Phi = sum( psa_i*(psb_i+dis^2)^(-3/2)*(x_i-x_j), i!=j)
%      d^2/dt^2 y_j = -d/dy_j Phi = sum( psa_i*(psb_i+dis^2)^(-3/2)*(y_i-y_j), i!=j)
%      d^2/dt^2 z_j = -d/dz_j Phi = sum( psa_i*(psb_i+dis^2)^(-3/2)*(z_i-z_j), i!=j)
%   yielding the following system of first-order differential equations, which are the
%   equations of motion of this system and which are returned as a [6,N]-matrix 'r'
%      d/dt x_j  = r[0,j] = vx_j
%      d/dt y_j  = r[1,j] = vy_j
%      d/dt z_j  = r[2,j] = vz_j
%      d/dt vx_j = r[3,j] = -d/dx_j Phi = sum( psa_i*(psb_i+dis^2)^(-3/2)*(x_i-x_j), i!=j)
%      d/dt vy_j = r[4,j] = -d/dy_j Phi = sum( psa_i*(psb_i+dis^2)^(-3/2)*(y_i-y_j), i!=j)
%      d/dt vz_j = r[5,j] = -d/dz_j Phi = sum( psa_i*(psb_i+dis^2)^(-3/2)*(z_i-z_j), i!=j)
%\qualifiers{
%\qualifier{psa}{[\code{=Double_Type[N]+1}] Parameter used to parametrize the interaction potential.}
%\qualifier{psb}{[\code{=Double_Type[N]+0}] Parameter used to parametrize the interaction potential.}
%}
%\example
%    N = 4;
%    m = Double_Type[6,N];
%    r = N_body_simulation_std_kernel(0,m; psa=Double_Type[N]+1, psb=Double_Type[N]+1);
%\seealso{N_body_simulation}
%!%-
{
  if(_NARGS!=2)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': Wrong number of arguments.", _function_name());
  }
  variable t, m;
  (t, m)= ();
  variable N = length(m[0,*]);
  variable psa = qualifier("psa", Double_Type[N]+1);
  if(length(psa)!=N)
    throw UsageError, sprintf("Usage error in '%s': Qualifier 'psa' invalid. It is supposed to be an array of length %d.", _function_name(), N);
  variable psb = qualifier("psb", Double_Type[N]+0);
  if(length(psb)!=N)
    throw UsageError, sprintf("Usage error in '%s': Qualifier 'psb' invalid. It is supposed to be an array of length %d.", _function_name(), N);
  %
  variable r = Double_Type[6,N];
  r[0,*] = m[3,*];
  r[1,*] = m[4,*];
  r[2,*] = m[5,*];
  variable i, j;
  _for i(0,N-2,1)
  {
    _for j(i+1,N-1,1)
    {
      variable dx = m[0,j]-m[0,i];
      variable dy = m[1,j]-m[1,i];
      variable dz = m[2,j]-m[2,i];
      variable dis2 = dx^2 + dy^2 + dz^2;
      variable interaction_ij = psa[j] * (psb[j]+dis2)^(-1.5);
      variable interaction_ji = psa[i] * (psb[i]+dis2)^(-1.5);
      r[3,i] += interaction_ij*dx;
      r[3,j] -= interaction_ji*dx;
      r[4,i] += interaction_ij*dy;
      r[4,j] -= interaction_ji*dy;
      r[5,i] += interaction_ij*dz;
      r[5,j] -= interaction_ji*dz;
    }
  }
  return r;
}

define N_body_simulation()
%!%+
%\function{N_body_simulation}
%\synopsis{Compute orbits for N interacting particles}
%\usage{Struct_Type ret = N_body_simulation(Struct_Type in, Double_Type t_end; qualifiers)}
%\description
%    By numerical integration and without any simplifying approximations, this function
%    directly solves the equations of motion of a system of N particles under the
%    influence of their mutual forces (The interaction is specified in a separate
%    function, see qualifier 'kernel'.) from time t=0 to time t=t_end. Negative values
%    for t_end imply backward integration. Cartesian coordinates are used throughout.
%
%    The input structure 'in' has to contain the fields "x", "y", "z", "vx", "vy", "vz".
%    Each of these fields has to be an array of length N. The corresponding index gives
%    the particle id, i.e., particle 0 refers to x[0], y[0], z[0], vx[0], vy[0], vz[0].
%    The return structure 'ret' contains for each particle a field, e.g., for particle 0
%    a field named "o0". This field is again a structure with fields "t", "x", "y", "z",
%    "vx", "vy", and "vz" giving the time-dependent coordinates of the respective particle.
%\notes
%    An adaptive Runge-Kutta-Fehlberg method of fourth/fifth order is applied to solve the
%    coupled system of first-order differential equations. The stepsize is hereby controlled
%    such that for each step an absolute accuracy in coordinates and velocity components is
%    achieved that is smaller than given by the qualifier 'tolerance'.
%\qualifiers{
%\qualifier{kernel}{[\code{="N_body_simulation_std_kernel"}] Name of the function which describes the
%      mutual interaction. Note that all qualifiers are passed to this function as well.}
%\qualifier{threshold}{[\code{=0}] Lower limit on the time difference of two consecutive moments of time
%      that will be saved.}
%\qualifier{tolerance}{[\code{=1e-10}] Absolut error control tolerance; lower limit: 1e-15.}
%\qualifier{verbose}{Show intermediate times t.}
%}
%\example
%    % Four interacting particles:
%    s = struct{x, y, z, vx, vy, vz};
%    s.x = [-10,0,10,0];
%    s.y = [0,10,0,-10];
%    s.z = [0,0,0,0];
%    s.vx = [0,-1,0,1];
%    s.vy = [-1,0,1,0];
%    s.vz = [0,0,0,0];
%    r = N_body_simulation(s, 50; kernel="N_body_simulation_std_kernel", psa=10*[1,1,1,1], psb=0.1*[1,1,1,1]);
%    xrange(min_max([r.o0.x,r.o1.x,r.o2.x,r.o3.x]));
%    yrange(min_max([r.o0.y,r.o1.y,r.o2.y,r.o3.y]));
%    plot(r.o0.x,r.o0.y); oplot(r.o1.x,r.o1.y); oplot(r.o2.x,r.o2.y); oplot(r.o3.x,r.o3.y);
%\seealso{N_body_simulation_std_kernel, N_body_simulation_MW_kernel}
%!%-
{
  if(_NARGS!=2)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': Wrong number of arguments.", _function_name());
  }
  variable in, t_end;
  (in, t_end) = ();
  if(_is_struct_type(in)!=1)
    throw UsageError, sprintf("Usage error in '%s': First input parameter has to be Struct_Type.", _function_name());
  if(struct_field_exists(in, "x")==0 or struct_field_exists(in, "y")==0 or struct_field_exists(in, "z")==0 or
     struct_field_exists(in, "vx")==0 or struct_field_exists(in, "vy")==0 or struct_field_exists(in, "vz")==0)
    throw UsageError, sprintf("Usage error in '%s': First input parameter has to be Struct_Type with fields 'x', 'y', 'z', 'vx', 'vy', and 'vz'.", _function_name());
  variable N = length(in.x);
  if(length(in.y)!=N or length(in.z)!=N or length(in.vx)!=N or length(in.vy)!=N or length(in.vz)!=N)
    throw UsageError, sprintf("Usage error in '%s': Fields of first input parameter have to be arrays of the same length.", _function_name());
  %
  variable kernel = __get_reference(qualifier("kernel"));
  variable eps = _max( qualifier("tolerance", 1e-10), 1e-15); % error control tolerance
  variable threshold = qualifier("threshold", 0);
  variable verbose = qualifier_exists("verbose");
  %
  variable s = sign(t_end);
  if(s==0) s=1; % to account for t_end==0
  t_end *= s; % for backward integration, i.e. t_end < 0, reverse t_end and the sign of the velocity components
  %
  variable m = Double_Type[6,N]; m[0,*] = in.x, m[1,*] = in.y, m[2,*] = in.z, m[3,*] = s*in.vx, m[4,*] = s*in.vy, m[5,*] = s*in.vz;
  variable cols = String_Type[N];
  variable i;
  _for i(0, N-1, 1)
    cols[i] = sprintf("o%d", i);
  variable ret = @Struct_Type(cols);
  variable temp = struct{ t, x, y, z, vx, vy, vz };
  _for i(0, N-1, 1)
  {
    temp.t = {0};
    temp.x = {m[0,i]};
    temp.y = {m[1,i]};
    temp.z = {m[2,i]};
    temp.vx = {m[3,i]};
    temp.vy = {m[3,i]};
    temp.vz = {m[5,i]};
    set_struct_field(ret, cols[i], @temp);
  }
  % ============================
  % initializing the ODE-solver:
  variable k1, k2, k3, k4, k5, k6, dx_c, dx_e; % dx_c = dx continue, dx_e = dx error estimation
  variable coef = Double_Type[7,7]; % modified Butcher tableau for Runge-Kutta-Fehlberg method
  coef[0,0]= 1./4;   coef[0,1]=  1./4;
  coef[1,0]= 3./8;   coef[1,1]=  3./32;      coef[1,2]=  9./32;
  coef[2,0]= 12./13; coef[2,1]=  1932./2197; coef[2,2]= -7200./2197;  coef[2,3]=  7296./2197;
  coef[3,0]= 1.;     coef[3,1]=  439./216;   coef[3,2]= -8.;          coef[3,3]=  3680./513;    coef[3,4]= -845./4104;
  coef[4,0]= 1./2;   coef[4,1]= -8./27;      coef[4,2]=  2.;          coef[4,3]= -3544./2565;   coef[4,4]=  1859./4104; coef[4,5]= -11./40;
  coef[5,0]= 0;      coef[5,1]=  25./216;    coef[5,2]=  1408./2565;  coef[5,3]=  2197./4104;   coef[5,4]= -1./5;       coef[5,5]= 0;       % fourth order solution
  coef[6,0]= 0;      coef[6,1]=  16./135;    coef[6,2]=  6656./12825; coef[6,3]=  28561./56430; coef[6,4]= -9./50;      coef[6,5]= 2./55;   % fifth order solution
  % ============================
  % =================================
  % solving the equations of motions:
  variable t = 0;
  variable t_last_save = 0;
  variable dt = 0.001;
  while(t < t+dt <= t_end)
  {
    k1 = dt * (@kernel)(s*t, m;; __qualifiers);
    k2 = dt * (@kernel)(s*(t + coef[0,0]*dt), m + coef[0,1]*k1;; __qualifiers);
    k3 = dt * (@kernel)(s*(t + coef[1,0]*dt), m + coef[1,1]*k1 + coef[1,2]*k2;; __qualifiers);
    k4 = dt * (@kernel)(s*(t + coef[2,0]*dt), m + coef[2,1]*k1 + coef[2,2]*k2 + coef[2,3]*k3;; __qualifiers);
    k5 = dt * (@kernel)(s*(t + coef[3,0]*dt), m + coef[3,1]*k1 + coef[3,2]*k2 + coef[3,3]*k3 + coef[3,4]*k4;; __qualifiers);
    k6 = dt * (@kernel)(s*(t + coef[4,0]*dt), m + coef[4,1]*k1 + coef[4,2]*k2 + coef[4,3]*k3 + coef[4,4]*k4 + coef[4,5]*k5;; __qualifiers);
    dx_c = coef[5,1]*k1 + coef[5,2]*k3 + coef[5,3]*k4 + coef[5,4]*k5 + coef[5,5]*k6; % fourth order solution
    dx_e = coef[6,1]*k1 + coef[6,2]*k3 + coef[6,3]*k4 + coef[6,4]*k5 + coef[6,5]*k6; % fifth order solution
    variable scale = 0.9*(eps/max(abs(dx_c-dx_e)))^(0.2); % 0.9 is safety factor; derivation see Numerical Recipes, Section "Adaptive Stepsize Control for Runge-Kutta"
    if(scale>=1 or dt<=1e-14*t) % scale>1: accept stepsize only if correction factor is equal/larger than 1
    {
      m += dx_c;
      t += dt;
      if(verbose) vmessage("t=%g", s*t);
      if(t-t_last_save > threshold or t==t_end)
      {
	t_last_save = t;
	list_append((get_struct_field(ret, cols[0])).t, t);
	_for i(0, N-1, 1)
	{
	  list_append((get_struct_field(ret, cols[i])).x, m[0,i]);
	  list_append((get_struct_field(ret, cols[i])).y, m[1,i]);
	  list_append((get_struct_field(ret, cols[i])).z, m[2,i]);
	  list_append((get_struct_field(ret, cols[i])).vx, m[3,i]);
	  list_append((get_struct_field(ret, cols[i])).vy, m[4,i]);
	  list_append((get_struct_field(ret, cols[i])).vz, m[5,i]);
	}
      }
    }
    dt = _max(dt*scale, t*1e-15); % abs(1e-15*t) in order to avoid infinite loops due to Double_Type precision limit at dt=1e-16*t
    if(t+dt > t_end) dt = t_end-t;
  }
  % =================================
  temp = struct{ t, x, y, z, vx, vy, vz };
  temp.t = list_to_array((get_struct_field(ret, cols[0])).t, Double_Type);
  temp.t *= s; % for backward integration, i.e. t_end < 0, re-reverse t
  _for i(0, N-1, 1)
  {
    temp.x = list_to_array((get_struct_field(ret, cols[i])).x, Double_Type);
    temp.y = list_to_array((get_struct_field(ret, cols[i])).y, Double_Type);
    temp.z = list_to_array((get_struct_field(ret, cols[i])).z, Double_Type);
    temp.vx = list_to_array((get_struct_field(ret, cols[i])).vx, Double_Type);
    temp.vy = list_to_array((get_struct_field(ret, cols[i])).vy, Double_Type);
    temp.vz = list_to_array((get_struct_field(ret, cols[i])).vz, Double_Type);
    temp.vx *= s; temp.vy *= s; temp.vz *= s; % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
    set_struct_field(ret, cols[i], @temp);
  }
  return ret;
}

%
% Code from src/misc/astronomy/N_body_simulation_MW_kernel.sl
%
define N_body_simulation_MW_kernel()
%!%+
%\function{N_body_simulation_MW_kernel}
%\synopsis{Alternative interaction kernel for the function 'N_body_simulation'}
%\usage{Double_Type r[6,N] = N_body_simulation_MW_kernel(Double_Types t, m[6,N]; qualifiers)}
%\description
%    This function is an alternative interaction kernel for the function 'N_body_simulation'.
%    It combines the mutual N-nody interactions of the standard interaction kernel
%    'N_body_simulation_std_kernel' with the external forces stemming from an analytical
%    model for the gravitational potential of the Milky Way (see qualifier 'model').
%\notes
%    Because of the unit convention used for the potentials of the Milky Way, 'psa', which
%    is a qualifier of the function 'N_body_simulation_std_kernel' and which is assumed to
%    be the mass of the Plummer spheres in solar masses, has to be converted to Galactic
%    mass units and then multiplied with a constant accounting for the remaining unit
%    conversions (see example below). The units of 'psb' have to be kpc^2.
%\qualifiers{
%\qualifier{model}{[\code{="AS"}]: Function ("AS", "MN_NFW", or "MN_TF"), which evaluates the equations of
%      motion that result from a model for the gravitational potential of the Milky Way.}
%\qualifier{All qualifiers from the model potential function except 'coords'.}{}
%\qualifier{All qualifiers from the function 'N_body_simulation_std_kernel'.}{}
%}
%\example
%    % Interacting satellite galaxies in the Milky Way:
%    s = properties_satellite_galaxies();
%    i = struct{ x, y, z, vx, vy, vz, psa, psb };
%    model = "AS"; % Milky Way mass model
%    SunGCDist = (@(__get_reference(model)))(; eomecd="sgcd"); % Sun-GC distance of chosen mass model
%    temp = [SunGCDist,0,0,0,0,0]; reshape(temp, [6,1]);
%    vlsr = (@(__get_reference(model)))(0, temp; eomecd="circ")[0]; % Local standard of rest velocity of chosen mass model
%    (i.x, i.y, i.z, i.vx, i.vy, i.vz) = cel2gal(s.ah, s.am, s.as, s.dd, s.dm, s.ds, s.dist, s.vrad, s.pma_cos_d, s.pmd; SunGCDist=SunGCDist, vlsr=vlsr);
%    kpcmyr_to_kms = 977.7736364875057; % = conversion factor from kpc/myr to km/s = 3.0856775975*10^16 / (10^6 * 3.15582 * 10^7);
%    i.vx /= kpcmyr_to_kms;
%    i.vy /= kpcmyr_to_kms;
%    i.vz /= kpcmyr_to_kms;
%    i.psa = s.Pl_mass/2.325131802556774e+07; % conversion from solar masses to Galactic mass units Mgal to have G=1
%    % Mgal = 2.325131802556774e+07 = 1e8*3.0856775975*1e19/6.6742/1e-11/1.9884/1e30, see Irrgang et al., 2013, A&A, 549, A137
%    const = 100./kpcmyr_to_kms^2; % factor 100 because potential is given in 100 km^2/s^2, see Irrgang et al. 2013
%    i.psa *= const;
%    i.psb = (s.Pl_radius)^2;
%    r = N_body_simulation(i, -1000; kernel="N_body_simulation_MW_kernel", psa=i.psa, psb=i.psb, model=model);
%    xrange(min_max([r.o0.x,r.o1.x])); yrange(min_max([r.o0.y,r.o1.y]));
%    plot(r.o0.x,r.o0.y); oplot(r.o1.x,r.o1.y);
%\seealso{N_body_simulation, N_body_simulation_std_kernel, AS, MN_NFW, MN_TF}
%!%-
{
  if(_NARGS!=2)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': Wrong number of arguments.", _function_name());
  }
  variable model = qualifier("model", "AS"); % function evaluating the equations of motions of the Milky Way potential
  if(__get_reference(model)==NULL or (model!="AS" and model!="MN_NFW" and model!="MN_TF"))
  {
    _pop_n(_NARGS); % remove function arguments from stack
    throw UsageError, sprintf("Usage error in '%s': Invalid input for qualifier 'model'.", _function_name());
  }
  model = __get_reference(model);
  variable t, m;
  (t, m) = ();
  variable N = length(m[0,*]);
  variable forces_only = Double_Type[6,N]; % to avoid summing over velocities
  forces_only[3,*] += 1;
  forces_only[4,*] += 1;
  forces_only[5,*] += 1;
  variable qualies = __qualifiers;
  qualies = struct_combine(qualies, "coords");
  set_struct_field(qualies, "coords", "cart");
  return ( N_body_simulation_std_kernel(t,m;; __qualifiers) + forces_only * (@model)(t,m;; qualies) );
}

%
% Code from src/fitting/statistics/PG.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
static define PG_function (S, m, weight)
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{PG_function}
%\synopsis{sets fitting statistic for poisson source counts and already subtracted gaussian background}
%\usage{set_fit_statistic("PG");}
%\description
%    	This fitting statistic can be used when dealing with
%	 	low count spectra that have already been subtracted
%		by the instrumental background. Cash statistics are
%		not suited for this task since purely poisson distri-
%		buted counts are assumed in that case. 
%		This version of the fitting statistic has been derived
%		from the profile likelihood statistic for zero back-
%		ground counts from:
%
%		https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/XSappendixStatistics.html
%
%		The statistic should be used when fitting low count 
%		spectra of the Swift/BAT instrument.
%		IMPORTANT: the statistic can only be used when fitting
%		count rate spectra since the exposure times for each
%		channel have dummy values of 1s.
%\seealso{set_fit_statistic}
%!%-
{
	variable ts = 1;
	variable tb = 1;
	variable sig = 1.0/sqrt(weight);
	variable d = sqrt((ts*sig^2+tb^2*m)^2 - 4*tb^2*(ts*sig^2*m-S*sig^2));
	variable f = (-(ts*sig^2+tb^2*m) + d)/(2*tb^2);
	
       	variable PG = ts*(m+f) - S*log(ts*m+ts*f) + 1.0*(tb^2*f^2)/(2*sig^2) - S*(1-log(S));
       	return (PG, 2*sum(PG));
}

static define PG_report (stat, npts, nvpars)
{
    variable s = sprintf ("  PG = %0.4g\n", stat);
    return s;
}

add_slang_statistic ("PG", &PG_function, &PG_report);


%
% Code from src/plot/PLOT_ENV.sl
%
% This is an extra file for setting global environment variables when the isisscripts are required
% The new build process put these two lines inside of two #ifeval statements which often was not evaluated

putenv("PGPLOT_BACKGROUND=white"); % invert default for better view.
putenv("PGPLOT_FOREGROUND=black");

%
% Code from src/slang/astronomy/RAdec_from_AzEl.sl
%
define RAdec_from_AzEl()
%!%+
%\function{RAdec_from_AzEl}
%\synopsis{computes equatorial (RA, dec) coordinates from a point's (azimut, elevation) at a time MJD}
%\usage{(RA, dec) = RAdec_from_AzEl(az, el, MJD,  longw, lat);}
%\description
%    This function is deprecated. Please use horizon2equatorial instead.
%\seealso{horizon2equatorial,equatorial2horizon}
%!%-
{
  variable az, el, MJD, longw, lat;
  switch(_NARGS)
  { case 5: (az, el, MJD, longw, lat) = (); }
  { help(_function_name()); return; }

  az = angle_to_rad(az; unit="deg");
  el = angle_to_rad(el; unit="deg");

  variable alpha,delta;

  (alpha,delta)=horizon2equatorial(az,el;JD=MJD,mjd,lat=lat*PI/180,lon=longw*PI/180);

  alpha=alpha*180./PI * 24./360.;
  delta=delta*180./PI;
  
  return alpha,delta;
}

%
% Code from src/slang/astronomy/RAdec_from_galLB.sl
%
define RAdec_from_galLB()
%!%+
%\function{RAdec_from_galLB}
%\synopsis{convert galactic (l, b) coordinates to equatorial (RA, dec) coordinates}
%\usage{(RA, dec) = RAdec_from_galLB(l, b);}
%\qualifiers{
%\qualifier{l_unit}{ [\code{="deg"}]: set the unit of the galactic longitude
%                       \code{l_unit="rad"} \code{=>}  l in rad
%                       \code{l_unit="hms"} \code{=>}  l in hours as a scalar
%                       or an array of the form \code{[H, M]} or \code{[H, M, S]} }
%\qualifier{b_unit}{ [\code{="deg"}]: set the unit of the galactic lattitude}
%}
%\description
%    This function is deprecated. Please use galactic2equatorial instead.
%\seealso{galLB_from_RAdec, AzEl_from_RAdec, RAdec_from_AzEl, angle_to_rad}
%!%-
{
  variable l, b;
  switch(_NARGS)
  { case 2: (l, b) = (); }
  { help(_function_name()); return; }

    variable l_unit = qualifier("l_unit", "deg");
    variable b_unit = qualifier("b_unit", "deg");
  
    l  = angle_to_rad(l ; unit= l_unit);
    b  = angle_to_rad(b ; unit= b_unit);
  
  variable alpha,delta;

  (alpha,delta)=galactic2equatorial(l,b);

  return alpha*180./PI, delta*180./PI;

}

%
% Code from src/slang/math/RD2rad.sl
%
define RD2rad()
%!%+
%\function{RD2rad}
%\synopsis{Convert right ascension (h, m, s) and declination (deg, arcmin, arcsec) to radians}
%\usage{RD2rad(Double_Types ah[], am[], as[], dd[], dm[], ds[])}
%\description
%    Given right ascension in hours, minutes, seconds and declination in degrees,
%    minutes of arc, seconds of arc, the function converts them to radians. Always
%    provide six numbers, i.e., fill up with zeros where necessary. Right ascension
%    is assumed to be positive. If the degrees argument of declination is negative
%    the minus sign is automatically applied to the minutes and seconds parameters.
%    If the degrees argument of declination is zero, the sign of the minutes argument
%    (if nonzero) is applied to the seconds parameter.
%\example
%    (raInRad, declInRad) = RD2rad(01, 45, 12.54, 87, 55, 45.34);
%    (raInRad, declInRad) = RD2rad(01, 45.209, 0, -87, 55, 45.34);
%    (raInRad, declInRad) = RD2rad(01.753472, 0, 0, 0, -55, 45.34);
%    (raInRad, declInRad) = RD2rad(01, 45, 12.54, 0, 0, 45.34);
%    (raInRad, declInRad) = RD2rad([01, 01], [45,45.209], [12.54,0], [87,-87], [55,55], [45.34,45]);
%    (raInRad, declInRad) = RD2rad(rad2RD(0,-PI/4.));
%\seealso{hms2deg,dms2deg,rad2RD}
%!%-
{
  if(_NARGS != 6)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable  ah, am, as, dd, dm, ds;
    (ah, am, as, dd, dm, ds) = ();
    ah = [1.*ah]; am = [1.*am]; as = [1.*as]; dd = [1.*dd]; dm = [1.*dm]; ds = [1.*ds];
    % convert right ascention to radians:
    variable raInRad = PI*(ah/12. + am/720. + as/43200.); % right ascension in radians = ah*(2.*PI/24.) + am*(2.*PI/(24.*60.)) + as*(2.*PI/(24.*3600.));
    % convert declination to radians with the proper sign conversions:
    variable ind;
    ind = where(dd != 0);
    dm[ind] *= sign(dd[ind]);
    ds[ind] *= sign(dd[ind]);
    ind = where(dd == 0 and dm != 0);
    ds[ind] *= sign(dm[ind]);
    variable declInRad = PI*(dd/180. + dm/10800. + ds/648000.); % declination in radians = dd * (2.*PI/360.) + dm * (2.*PI/(360.*60.)) + ds * (2.*PI/(360.*60.^2.));
    if(length(raInRad)==1) {return (raInRad[0], declInRad[0]);}
    else return (raInRad, declInRad);
  }
}

define rad2RD()
%!%+
%\function{rad2RD}
%\synopsis{Convert right ascension and declination from radians to (h, m, s) and (deg, arcmin, arcsec)}
%\usage{rad2RD(Double_Types RA[], D[])}
%\description
%    Given right ascension and declination in radians, the function converts them to
%    hours, minutes, seconds and degrees, minutes of arc, seconds of arc. In case of
%    negative declinations, the minus sign is assigned only to the highest non-vanishing
%    term, i.e., to degrees if degrees is nonzero, to minutes if minutes is nonzero but
%    degrees is zero, ... . Negative right ascensions are not expected.
%\example
%    (ah, am, as, dd, dm, ds) = rad2RD(0,-PI/4.);
%    (ah, am, as, dd, dm, ds) = rad2RD(0,-PI/181.);
%    (ah, am, as, dd, dm, ds) = rad2RD(3/4.*PI,0);
%    (ah, am, as, dd, dm, ds) = rad2RD([0,3/4.*PI],[-PI/4,0]);
%    (ah, am, as, dd, dm, ds) = rad2RD(RD2rad(01, 45, 12.54, 87, 55, 45.34));
%\seealso{RD2rad}
%!%-
{
  if(_NARGS != 2)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable raInRad, declInRad;
    (raInRad, declInRad) = ();
    raInRad = [1.*raInRad]; declInRad = [1.*declInRad];
    % convert right ascention to (h, m, s):
    variable ah, am, as;
    raInRad *= (12./PI);
    ah = floor(raInRad);
    raInRad = (raInRad-ah)*60.;
    am = floor(raInRad);
    as = (raInRad-am)*60.;
    % convert declination to (deg, arcmin, arcsec) with the proper sign conversions:
    variable s = sign(declInRad);
    declInRad *= s; % all declinations positive; s contains information which ones were negative
    variable dd, dm, ds;
    declInRad *= (180./PI);
    dd = floor(declInRad);
    declInRad = (declInRad-dd)*60.;
    dm = floor(declInRad);
    ds = (declInRad-dm)*60.;
    % recover sign of declination:
    dd *= s;
    variable ind = where( dd==0 );
    dm[ind] *= s[ind];
    ind = where( dd==0 and dm==0 );
    ds[ind] *= s[ind];
    % return output:
    if(length(raInRad)==1) {return (ah[0], am[0], as[0], dd[0], dm[0], ds[0]);}
    else return (ah, am, as, dd, dm, ds);
  }
}

%
% Code from src/misc/satellites/RXTE/RXTE_ASM_countrate.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%
define RXTE_ASM_countrate()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{RXTE_ASM_countrate}
%\synopsis{estimates the RXTE-ASM countrate from the current model}
%\usage{(A, B, C) = RXTE_ASM_countrate();}
%\description
%     Zdziarski et al. (http://adsabs.harvard.edu/abs/2002ApJ...578..357Z)
%     provide a "response matrix" for the RXTE-ASM.
%     The inverse of this matrix is applied to the energy flux
%     derived from the current fit-function and its parameters,
%     in order to estimate the RXTE-ASM count rates \code{A}, \code{B} and \code{C} (cps).
%
%     Note that these numbers may only give a rough estimate!
%\seealso{energyflux}
%!%-
{
  variable Fa = energyflux(1.5, 3);
  variable Fb = energyflux(3,   5);
  variable Fc = energyflux(5,  12);
  return  62675e2/1570549.   * Fa +     81e4/1570549.    * Fb,
           8925e2/1570549.   * Fa +    638e4/1570549.    * Fb,
         226525e2/268563879. * Fa + 340054e4/5639841459. * Fb + 1e4/3591. * Fc;
}

%
% Code from src/misc/satellites/RXTE/RXTE_ASM_lightcurve.sl
%
define RXTE_ASM_lightcurve()
%!%+
%\function{RXTE_ASM_lightcurve}
%\synopsis{retrieves ASM lightcurves for a given source}
%\usage{Struct_Type RXTE_ASM_lightcurve(String_Type sourcename);}
%\qualifiers{
%\qualifier{MJDmin}{earliest MJD to be used}
%\qualifier{MJDmax}{latest MJD to be used}
%\qualifier{dt}{if specified, time resolution in MJD for rebinning}
%\qualifier{no_filter_nan}{do not remove empty bins after rebinning (only with \code{dt})}
%\qualifier{list}{lists available sources; \code{sourcename} may be omitted but can be a regular expression}
%\qualifier{get_list}{as list, but the list of sources is returned as an array of strings}
%\qualifier{save}{saves the light curve data in a local FITS file}
%\qualifier{verbose}{}
%}
%!%-
{
  variable src=NULL;
  switch(_NARGS)
  { case 0: ifnot(qualifier_exists("list")) { help(_function_name()); return; } }
  { case 1: src = (); }
  { help(_function_name()); return; }

  variable filename = "./RXTE-ASM_" + src + ".fits";
  if(qualifier_exists("save"))
  { vmessage("%s: writing %s", _function_name(), filename);
    fits_write_binary_table(filename, src, RXTE_ASM_lightcurve(src));
    return;
  }

  variable path_to_data = NULL;
  try {
    foreach path_to_data ([local_paths.RXTE_ASM_data])
      if(stat_file(path_to_data)!=NULL)  break;
  } catch NotImplementedError: { vmessage("warning (%s): local_paths.RXTE_ASM_data not defined",_function_name()); };

  if(qualifier_exists("list") || qualifier_exists("get_list"))
  {
    if(_isnull(path_to_data) || stat_file(path_to_data)==NULL)
    { vmessage("error (%s): no path to RXTE data found", _function_name());
      return;
    }
    variable source, sources = String_Type[0];
    foreach source (glob(path_to_data+"lightcurves/xa_*_d1.lc"))
    {
      (source,) = strreplace(source, path_to_data+"lightcurves/xa_", "", 1);
      (source,) = strreplace(source, "_d1.lc", "", 1);
      if(src==NULL || string_match(source, src, 1))
	sources = [sources, source];
    }
    sources = sources[array_sort(sources)];
    if(qualifier_exists("get_list"))
      return sources;
    foreach source (sources)
      message(source);
    return;
  }

  variable asm, asm_sum, asm_col, n, col_a, col_b, col_c;
  if(_isnull(path_to_data) || stat_file(path_to_data)==NULL)
  {
    if(stat_file(filename)==NULL)
    { vmessage("error (%s): no path to RXTE data found", _function_name());
      return;
    }
    vmessage("warning (%s): no path to RXTE data found, but using %s", _function_name(), filename);
    asm = fits_read_table(filename);
    asm_sum = struct {
      time = asm.time,
      timedel = asm.timedel,
      rate = asm.rate,
      error = asm.err,
      rdchi_sq = asm.chi2red
    };
    n = length(asm.time);
    asm_col = struct {
      time = [asm.time, asm.time, asm.time],
      timedel = [asm.timedel, asm.timedel, asm.timedel],
      band = String_Type[3*n],
      rate = [asm.rate_a, asm.rate_b, asm.rate_c],
      error = [asm.err_a, asm.err_b, asm.err_c],
      rdchi_sq = [asm.chi2red_a, asm.chi2red_b, asm.chi2red_c],
    };
    asm_col.band[ [  0 :   n-1] ] = "A";
    asm_col.band[ [  n : 2*n-1] ] = "B";
    asm_col.band[ [2*n : 3*n-1] ] = "C";
  }
  else
  {
    filename = path_to_data+"lightcurves/xa_"+src+"_d1.lc";
    if(stat_file(filename)==NULL)
    { vmessage("error (%s): %s does not exist\nAre you sure that the source is called '%s'?", _function_name(), filename, src);
      vmessage("use %s(\"\";list); to see the list of source names", _function_name());
      return NULL;
    }
    asm_sum = fits_read_table(filename, ["time", "rate", "error", "timedel", "rdchi_sq"]);
    %MJDREFF wrong in ASM headers (=0), thus not used.
    %for reference see
    %https://heasarc.gsfc.nasa.gov/docs/xte/recipes/asm_recipe.html
    asm_sum.time = asm_sum.time + fits_read_key(filename, "MJDREFI")+0.000696574074;;

    filename = path_to_data+"colors/xa_"+src+"_d1.col";
    asm_col = fits_read_table(filename, ["time", "band", "rate", "error", "timedel", "rdchi_sq"]);
    %MJDREFF wrong in ASM headers (=0), thus not used.
    %for reference see
    %https://heasarc.gsfc.nasa.gov/docs/xte/recipes/asm_recipe.html
    asm_col.time = asm_col.time + fits_read_key(filename, "MJDREFI")+0.000696574074;
  }

  variable MJDmin = qualifier("MJDmin", min([asm_sum.time, asm_col.time]));
  variable MJDmax = qualifier("MJDmax", max([asm_sum.time, asm_col.time]));
  struct_filter(asm_sum, where(MJDmin <= asm_sum.time <= MJDmax));
  struct_filter(asm_col, where(MJDmin <= asm_col.time <= MJDmax));

  col_a = where(asm_col.band=="A");
  col_b = where(asm_col.band=="B");
  col_c = where(asm_col.band=="C");

  variable i;
  if(length(col_a) != length(col_b) || length(col_b) != length(col_c) || length(col_c) != length(asm_sum.time))
  { vmessage("warning (%s): lightcurves A, B, C, SUM do not have the same length", _function_name());
    vmessage("  len(A)=%d, len(B)=%d, len(C)=%d, len(SUM)=%d", length(col_a), length(col_b), length(col_c), length(asm_sum.time));
    i = min([length(col_a), length(col_b), length(col_c), length(asm_sum.time)]);
    vmessage("  trying to cut light curves to common length %d", i);
    i = [0:i-1];
    col_a = col_a[i];
    col_b = col_b[i];
    col_c = col_c[i];
    struct_filter(asm_sum, i);
  }
  if(length(col_a)==0)
  {
    vmessage("warning (%s): no data on %s found for %.1f <= MJD <= %.1f", _function_name(), src, MJDmin, MJDmax);
    return NULL;
  }
  variable maxdiff_time = maxabs([asm_col.time[col_a]-asm_col.time[col_b],
				  asm_col.time[col_b]-asm_col.time[col_c],
				  asm_col.time[col_c]-asm_sum.time ]);
  variable maxdiff_timedel = maxabs([asm_col.timedel[col_a]-asm_col.timedel[col_b],
				     asm_col.timedel[col_b]-asm_col.timedel[col_c],
				     asm_col.timedel[col_c]-asm_sum.timedel ]);
  if(qualifier_exists("verbose") || maxdiff_time > 1e-6 || maxdiff_timedel > 1e-6)
  {
    vmessage("max |time(A) - time(B)|   = %f", maxabs(asm_col.time[col_a]-asm_col.time[col_b]) );
    vmessage("max |time(B) - time(C)|   = %f", maxabs(asm_col.time[col_b]-asm_col.time[col_c]) );
    vmessage("max |time(C) - time(SUM)| = %f", maxabs(asm_col.time[col_c]-asm_sum.time       ) );

    vmessage("max |timedel(A) - timedel(B)|   = %f", maxabs(asm_col.timedel[col_a]-asm_col.timedel[col_b]) );
    vmessage("max |timedel(B) - timedel(C)|   = %f", maxabs(asm_col.timedel[col_b]-asm_col.timedel[col_c]) );
    vmessage("max |timedel(C) - timedel(SUM)| = %f", maxabs(asm_col.timedel[col_c]-asm_sum.timedel       ) );
  }

  if(qualifier_exists("dt"))
  {
    variable time_lo = [MJDmin : MJDmax : qualifier("dt")];
    variable time_hi = make_hi_grid(time_lo);
    n = length(time_lo);
    variable rev;
    asm = struct {
      time_lo = time_lo,
      time_hi = time_hi,
      time    = (time_lo+time_hi)*0.5,
      n       = histogram(asm_sum.time, time_lo, time_hi, &rev),
      rate    = Float_Type[n],
      err     = Float_Type[n],
      rate_a  = Float_Type[n],
      err_a   = Float_Type[n],
      rate_b  = Float_Type[n],
      err_b   = Float_Type[n],
      rate_c  = Float_Type[n],
      err_c   = Float_Type[n]
    };
    _for i (0, n-1, 1)
    { variable one_over_n = 1./asm.n[i];
      asm.rate[i]   =       sum(asm_sum.rate [      rev[i] ]  )   * one_over_n;
      asm.err[i]    = sqrt( sum(asm_sum.error[      rev[i] ]^2) ) * one_over_n;
      asm.rate_a[i] =       sum(asm_col.rate [col_a[rev[i]]]  )   * one_over_n;
      asm.err_a[i]  = sqrt( sum(asm_col.error[col_a[rev[i]]]^2) ) * one_over_n;
      asm.rate_b[i] =       sum(asm_col.rate [col_b[rev[i]]]  )   * one_over_n;
      asm.err_b[i]  = sqrt( sum(asm_col.error[col_b[rev[i]]]^2) ) * one_over_n;
      asm.rate_c[i] =       sum(asm_col.rate [col_c[rev[i]]]  )   * one_over_n;
      asm.err_c[i]  = sqrt( sum(asm_col.error[col_c[rev[i]]]^2) ) * one_over_n;
    }
    ifnot(qualifier_exists("no_filter_nan"))
      struct_filter(asm, where(asm.n>0));
    return asm;
  }
  else
    return struct {
      time      = asm_sum.time,
      timedel   = asm_sum.timedel,
      rate      = asm_sum.rate,
      err       = asm_sum.error,
      chi2red   = asm_sum.rdchi_sq,
      rate_a    = asm_col.rate    [col_a],
      err_a     = asm_col.error   [col_a],
      chi2red_a = asm_col.rdchi_sq[col_a],
      rate_b    = asm_col.rate    [col_b],
      err_b     = asm_col.error   [col_b],
      chi2red_b = asm_col.rdchi_sq[col_a],
      rate_c    = asm_col.rate    [col_c],
      err_c     = asm_col.error   [col_c],
      chi2red_c = asm_col.rdchi_sq[col_a],
    };
}

%
% Code from src/misc/satellites/RXTE/RXTE_PCA_info.sl
%
define RXTE_PCA_info()
%!%+
%\function{RXTE_PCA_info}
%\synopsis{retrieves light curves for a given RXTE-PCA observation}
%\usage{Struct_Type RXTE_PCA_info(String_Type dirs[])}
%\qualifiers{
%  \qualifier{binning}{how much datapoints are to be summed up into a single point}
%  \qualifier{path}{path to the observation}
%  \qualifier{dirs}{subdirectories to the individual observing blocks}
%
%  \qualifier{noback}{set if no bkg subtraction is to be performed}
%  \qualifier{earthvle}{set if EarthVLE background model is to be used}
%  \qualifier{faint}{set if Faint background model is to be used}
%  \qualifier{q6}{set if Q6 background model is to be used
%                (default is to test for earthvle,faint,q6)}
%  \qualifier{skyvle}{set if SkyVLE background model is to be used
%             (default is 0 for noback,earthvle,faint,q6,skyvle)}
%  \qualifier{exclusive}{set to search for data that was extracted
%                with the exclusive keyword to pca_standard being set.}
%  \qualifier{top}{set to read top-layer data}
%  \qualifier{nopcu0}{set to search for data that was extracted
%                ignoring PCU0}
%  \qualifier{fivepcu}{plot count-rates wrt to whole PCA, i.e.,
%                normalizing to five PCU; default is to plot the average
%                countrate per PCU}
%  \qualifier{bary}{Try to use barycenter time column in data. Must be
%                   created with fxbary before into file with postfix _bary.}
%  \qualifier{t}{time array of data in MJD.}
%  \qualifier{c}{count array of data}
%  \qualifier{err}{Estimated error by applying Poisson statistic.
%            Binning/background subtraction will be acknowledged.}
%\description
%    The elements of \code{dirs} may contain globbing expressions.
%
%    The returned structure has the following fields:\n
%    \code{obstime = struct { start, stop}} with the times for each observation in \code{dirs}\n
%    \code{lc = struct { time, rate, error }} with the PCA light curve\n
%    \code{obscat = struct} with information from the observation catalogue (occultation, saa, good time)\n
%    \code{gti}\n
%    \code{xfl = struct} with information from the filter file
%\seealso{aitlib/rxte/readxtedata.pro}
%}
%!%-
%
{
  variable dirs;
  switch(_NARGS)
  { case 1: dirs = (); }
  { help(_function_name()); return; }

  dirs = glob(dirs);
  variable verbose = qualifier_exists("verbose");

  variable nobs = length(dirs);
  if(verbose)
  { vmessage("retrieving PCA info from %d directories:", length(dirs));
    array_map(Void_Type, &vmessage, "- %s", dirs);
    message("");
  }

  variable obstime = struct {
    start = Double_Type[nobs],
    stop  = Double_Type[nobs],
  };
  obstime.start[*] = 1e38;
  obstime.stop[*] = -1e38;
  variable lc = struct {
    time  = Double_Type[0],
    rate  = Double_Type[0],
    error = Double_Type[0],
  };

  % given the state of the PCA, here are ALL possible combinations of PCUs where at least one PCU is still on
  variable detoff = ["",
   "0", "1", "2", "3", "4",
   "01", "02", "03", "04", "12", "13", "14", "23", "24", "34",
   "012", "013", "014", "023", "024", "034", "123", "124", "134", "234",
   "0123", "0124", "0134", "0234", "1234"];
  variable numpcuon = 5 - array_map(Integer_Type, &strlen, detoff);

  if(qualifier_exists("nopcu0"))
    numpcuon[where(array_map(Integer_Type, &string_match, detoff, "0", 1) > 0)]--;

  % conversion of measured countrate to 1PCU
  variable factor = 1./numpcuon;

  if(qualifier_exists("fivepcu"))  factor *= 5;
  variable binning = qualifier("binning", 1);

  % prepare filename
  detoff = "_" + detoff + "off";
  detoff[0] = "";

  variable topins  = ""; if(qualifier_exists("top"))  topins = "_top";
  variable exclins = ""; if(qualifier_exists("exclusive"))  exclins = "_excl";
  variable ignore  = ""; if(qualifier_exists("nopcu0"))  ignore = "_ign0";
  variable baryins = ""; if(qualifier_exists("bary"))  baryins = "_bary";

  variable lcfiles = "/standard2f" + detoff + exclins + ignore + topins + "/standard2f" + detoff + exclins + ignore + topins + baryins + ".lc";
  variable deadfi = "/pcadead/RemainingCnt.lc";
  variable backfaint = "/pcabackest/standard2f_back_Faint"    + topins + "_good" + detoff + exclins + ignore + baryins + ".lc";
  variable backq6    = "/pcabackest/standard2f_back_Q6"       + topins + "_good" + detoff + exclins + ignore + baryins + ".lc";
  variable backsky   = "/pcabackest/standard2f_back_SkyVLE"   + topins + "_good" + detoff + exclins + ignore + baryins + ".lc";
  variable backearth = "/pcabackest/standard2f_back_EarthVLE" + topins + "_good" + detoff + exclins + ignore + baryins + ".lc";

  variable i, j;
  _for i (0, nobs-1, 1)
  {
    _for j (0, length(lcfiles)-1, 1)
    {
      variable path_dir_file = dirs[i] + lcfiles[j];
      if(stat_file(path_dir_file) != NULL)
      {
	variable tt, cc, ee;
	ifnot(qualifier_exists("back"))
	{
	  if(verbose)  message("reading " + path_dir_file);
	  variable srclc = fits_read_lc(path_dir_file);  %  readlc,tt,cc,path+dirs(i)+file(j),/mjd,bary=bary
	  tt = srclc.time;
	  cc = srclc.rate;
	  % readlc,t_err,c_err,path+dirs(i)+file(j),/mjd,bary=bary
	  % ee = sqrt(c_err/double(binning)) * factor[j]
          ee = sqrt(cc/binning) * factor[j];
	}
%         IF (keyword_set(deadcorr)) THEN BEGIN
%           IF (file_exist(path+dirs(i)+deadfi)) THEN BEGIN
%             pcadeadlc,tt,cc,path+dirs(i),/mjd
%           ENDIF
%         ELSE BEGIN
%           print,'READXTEDATA: deadtime correction failed: "/pcadead/RemainingCnt.lc" does not exist'
%         ENDELSE
%
	ifnot(qualifier_exists("noback"))
	{
	  variable backfi = "";
	  if(qualifier_exists("faint")    && backfi=="" && stat_file(dirs[i] + backfaint)!=NULL)  backfi = backfaint[j];
	  if(qualifier_exists("q6")       && backfi=="" && stat_file(dirs[i] + backq6   )!=NULL)  backfi = backq6[j];
	  if(qualifier_exists("skyvle")   && backfi=="" && stat_file(dirs[i] + backsky  )!=NULL)  backfi = backsky[j];
	  if(qualifier_exists("earthvle") && backfi=="" && stat_file(dirs[i] + backearth)!=NULL)  backfi = backearth[j];

	  if(backfi=="" && stat_file(dirs[i] + backsky[j])!=NULL)
	  { backfi = backsky[j];
	    if(verbose)  message(" No background lightcurve has been specified, using 'SkyVLE'");
	  }
	  if(backfi=="")
	    message("Specified background has not been found.");

	  backfi = dirs[i] + backfi;
	  if(verbose)  message(" reading " + backfi);
  	  variable bkglc = fits_read_lc(backfi);  % readlc,tb,cb,path+dirs(i)+backfi,/mjd,bary=bary
	  if(qualifier_exists("back"))
	  { tt = bkglc.time;
	    cc = bkglc.rate;
	  }
	  else
          { if(any(tt!=bkglc.time)) vmessage("error (%s): bkglc.time != srclc.time", _function_name());
	    cc -= bkglc.rate;
	  }
	}
        cc = cc * factor[j];

%	if(binning>1)  % rebin data
%	{ % skip data at end of read set not fitting into binning:
%	  variable len = length(cc)-1 - length mod binning;
%         tt = tt[0:len:binning];
%         cc = cc[0:len];
%         ee = ee[0:len];
%	  variable ccc, eee;
%  %%%%%  cc = rebin(cc,n_elements(cc) / BINNING)
%       }
%

	% begin and end of the observation
	obstime.start[i] = _min(obstime.start[i], tt);
	obstime.stop[i]  = _max(obstime.stop [i], tt);

	lc.time  = [lc.time,  tt];
	lc.rate  = [lc.rate,  cc];
	lc.error = [lc.error, ee];
      }
    }
  }
  % sort light curve
  struct_filter(lc, array_sort(lc.time));
  if(verbose)
    if(length(lc.time)<=1)
      message("no files found");
    else
      message("");

  % read timeline to get occultation-times, saa-times, ...
  variable obscat = NULL;
  variable obscat_days = Integer_Type[0];
  _for i (0, nobs-1, 1)
    obscat_days = [ obscat_days, [int(obstime.start[i]-49354) : int(obstime.stop[i]-49352)] ];
  % produce unique list of timelines
  if(any(obscat_days>=0))  obscat = RXTE_obscat_info( obscat_days[unique(obscat_days)];; __qualifiers() );
  if(verbose and obscat!=NULL)  message("");

  % read GTI-Files used for preparing the spectra
  variable gti = struct { start = Double_Type[0], stop = Double_Type[0] };
  variable gtfi;
  _for i (0, length(dirs)-1, 1)
    foreach gtfi ( dirs[i] + "/filter/good" + detoff + exclins + ignore + ".gti" )
      if(stat_file(gtfi)!=NULL)
      {
	if(verbose)  message("reading " + gtfi);
	variable MJDREF = fits_read_key(gtfi, "MJDREFI") + fits_read_key(gtfi, "MJDREFF");
	gtfi = fits_read_table(gtfi);
        gti.start = [gti.start, MJDREF + gtfi.start/86400.];
        gti.stop  = [gti.stop,  MJDREF + gtfi.stop /86400.];
      }
  struct_filter(gti, array_sort(gti.start));
  if(verbose)  message("");

  % read electron ratios and times when the PCUs were on from the filter files
  variable xfl = RXTE_filter_file_info( dirs + "/filter/*xfl" ;; __qualifiers());

  return struct {
    obstime = obstime,
    lc = lc,
    obscat = obscat,
    gti = gti,
    xfl = xfl,
  };
}

%
% Code from src/misc/satellites/RXTE/RXTE_PCA_modes.sl
%
define RXTE_PCA_modes()
%!%+
%\function{RXTE_PCA_modes}
%\usage{RXTE_PCA_modes(String_Type obsids);}
%\qualifiers{
%\qualifier{compact}{only one row per ObsID, omit TSTART, TSTOP and duration}
%\qualifier{get_struct}{return information on PCA modes in a structure, too}
%\qualifier{quiet}{do not show information, implies \code{get_struct}}
%}
%\description
%    Reads RXTE-PCA data modes from the PCA FITS-index (FIPC) file.
%    The location of the RXTE data archive is determined from the
%    \code{local_paths.RXTE_data} variable (defined within the isisscripts).
%!%-
{
  variable path_to_data = NULL;
  try {
    foreach path_to_data ([local_paths.RXTE_data])
      if(stat_file(path_to_data)!=NULL)
        break;
  } catch NotImplementedError: { vmessage("warning (%s): local_paths.RXTE_PCA_modes not defined",_function_name()); };
  if(path_to_data == NULL)
  { vmessage("error (%s): no path to RXTE data found", _function_name());
    return;
  }

  variable dir=NULL, obsid;
  switch(_NARGS)
  { case 0: dir = getcwd(); }
  { case 1: obsid = (); }

  variable quiet = qualifier_exists("quiet");
  variable get_struct = qualifier_exists("get_struct") or quiet;

  variable PCAmodes;
  if(typeof(obsid)==Array_Type)
  {
    variable i, n = length(obsid);
    PCAmodes = Struct_Type[n];
    _for i (0, n-1, 1)
      PCAmodes[i] = RXTE_PCA_modes(obsid[i];; struct_combine(__qualifiers(), "quiet", "get_struct"));
    PCAmodes = merge_struct_arrays(PCAmodes);
    ifnot(quiet)  print_struct(PCAmodes);
    if(get_struct)
      return PCAmodes;
    else
      return;
  }

  if(dir==NULL)
  { variable AO = obsid[[0]];
    variable prop = string_matches(obsid, `\([0-9]*\)-`, 1)[1];
    if(AO=="9")  AO = string( 9 + integer(obsid[[1]]) );
    dir = path_to_data + "/AO" + AO + "/P" + prop + "/" + obsid;
  }

  variable FIPCfiles = glob(dir+"/FIPC_*");
  PCAmodes = struct { ObsID, TSTART, TSTOP, duration, EA1mode, EA2mode, EA3mode, EA5mode, EA6mode, EA7mode };
  if(length(FIPCfiles)==0)
  { message("warning ("+_function_name()+"): no FIPC file in "+dir);
    set_struct_fields(PCAmodes, String_Type[0], Integer_Type[0], dup(), dup(), String_Type[0], dup(), dup(), dup(), dup(), dup());
  }
  else
  {
    variable PCAindex = fits_read_table(FIPCfiles[0]);
    set_struct_fields(PCAmodes, PCAindex.obsid, PCAindex.beginmet, PCAindex.endmet,
                                PCAindex.endmet-PCAindex.beginmet,
		                PCAindex.ea1modenm, PCAindex.ea2modenm, PCAindex.ea3modenm, PCAindex.ea5modenm, PCAindex.ea6modenm, PCAindex.ea7modenm);
    if(qualifier_exists("compact"))
    {
      variable i1 = wherefirst(PCAmodes.EA1mode!=" ");  i1 = (i1 ? i1 : 0);   % if wherefirst == NULL (== 0 in a boolean context), set i to 0
      variable i2 = wherefirst(PCAmodes.EA2mode!=" ");  i2 = (i2 ? i2 : 0);
      variable i3 = wherefirst(PCAmodes.EA3mode!=" ");  i3 = (i3 ? i3 : 0);
      variable i5 = wherefirst(PCAmodes.EA5mode!=" ");  i5 = (i5 ? i5 : 0);
      variable i6 = wherefirst(PCAmodes.EA6mode!=" ");  i6 = (i6 ? i6 : 0);
      variable i7 = wherefirst(PCAmodes.EA7mode!=" ");  i7 = (i7 ? i7 : 0);
      PCAmodes = struct {
        ObsID   = [PCAmodes.ObsID[0]],
        EA1mode = [PCAmodes.EA1mode[i1]],
        EA2mode = [PCAmodes.EA2mode[i2]],
        EA3mode = [PCAmodes.EA3mode[i3]],
        EA5mode = [PCAmodes.EA5mode[i5]],
        EA6mode = [PCAmodes.EA6mode[i6]],
        EA7mode = [PCAmodes.EA7mode[i7]],
      };
    }
    ifnot(quiet)  print_struct(PCAmodes);
  }

  if(get_struct)
    return PCAmodes;
}

%
% Code from src/misc/satellites/RXTE/RXTE_filter_file_info.sl
%
private define getPCUgti(time, p)
{
  % intelligent padding for values that aren't known
  % ... not so fast but safe
  variable ndx = where(p==255);
  if(length(ndx) > 0)
  {
    variable start = 0;
    if(ndx[0] == 0)
    { p[0] = p[wherefirst(p!=255)];
      start = 1;
    }
    variable i;
    _for i (start, length(ndx)-1, 1)
      p[ndx[i]] = p[ndx[i]-1];
  }

  variable p1 = shift(p, 1);
  p1[-1] = p1[-2];

  % where det switches on
  variable switch_on = where(p==0 and p1==1);
  % where the det turns off
  variable switch_off = where(p==1 and p1==0);

  variable gti = struct { start, stop };

  if(length(switch_on)==0 || length(switch_off)==0)  % special case that detector does nothing or turns on/off exactly once
  {
    if(length(switch_on)==0 && length(switch_off)==0) % nothing happens
    {
      if(p[0]==1) % always on
      { gti.start = [ time[0] ];
        gti.stop = [ time[-1] ];
      }
      else  % always off
      { % gti.start = Double_Type[0];
	% gti.stop = Double_Type[0];
	gti.start = [ time[0] ];
        gti.stop = [ time[0] ];
      }
      return gti;
    }
    if(length(switch_on)==0) % only a switch off observed
    { gti.start = [ time[0] ];
      gti.stop = [ time[switch_off[0]] ];
      return gti;
    }
    % last case: only switch on seen
    gti.start = [ time[switch_on[0]] ];
    gti.stop = [ time[-1] ];
    return gti;
  }

  % general case of multiple on/off events
  switch_on += 1;  % add one since we start in the next bin
  % take care of det being on at beginning or end
  if(switch_on[0] > switch_off[0])  switch_on = [0, switch_on];
  if(length(switch_on) > length(switch_off))  switch_off = [switch_off, length(p)-1];

  % and return the values
  gti.start = time[switch_on];
  gti.stop = time[switch_off];
  return gti;
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define RXTE_filter_file_info()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{RXTE_filter_file_info}
%\synopsis{returns RXTE filter file information}
%\usage{Struct_Type RXTE_filter_file_info(String_Type xflfiles[])}
%\seealso{aitlib/rxte/readxfl.pro}
%!%-
{
  variable xflfiles;
  switch(_NARGS)
  { case 1: xflfiles = (); xflfiles = glob(xflfiles); }
  { help(_function_name()); return; }

  variable not_nopcu0 = not qualifier_exists("nopcu0");

  variable xfl_info = struct {
    time = Double_Type[0],
    electron = Double_Type[0],
    pcu0 = struct { start = Double_Type[0], stop = Double_Type[0] },
    pcu1 = struct { start = Double_Type[0], stop = Double_Type[0] },
    pcu2 = struct { start = Double_Type[0], stop = Double_Type[0] },
    pcu3 = struct { start = Double_Type[0], stop = Double_Type[0] },
    pcu4 = struct { start = Double_Type[0], stop = Double_Type[0] },
  };

  variable i;
  _for i (0, length(xflfiles)-1, 1)
  { variable tab = fits_read_table(xflfiles[i], ["time", "electron0", "electron1", "electron2", "electron3", "electron4", "pcu0_on", "pcu1_on", "pcu2_on", "pcu3_on", "pcu4_on"]);
    tab.time = fits_read_key(xflfiles[i], "MJDREFI") + fits_read_key(xflfiles[i], "MJDREFF") + tab.time/86400.;
    xfl_info.time = [xfl_info.time, tab.time];

    % read the electron ratios.
    variable use_el0;
    if(not_nopcu0)
      % do not use PCU0 to gauge background since the Xenon layer is damaged
      % check whether all finite el0 values are 0;
      % if so, el0 cannot be used because the detector does not work
      use_el0 = (sum(tab.electron0[wherenot(isnan(tab.electron0) or isinf(tab.electron0))]) > 0);

    % number of valid measurements in each bin
    variable num_PCUs =   not(isnan(tab.electron1) or isinf(tab.electron1))
			+ not(isnan(tab.electron2) or isinf(tab.electron2))
			+ not(isnan(tab.electron3) or isinf(tab.electron3))
			+ not(isnan(tab.electron4) or isinf(tab.electron4));
    if(not_nopcu0 && use_el0)  num_PCUs += not(isnan(tab.electron0) or isinf(tab.electron0));

    % where the ratio is not defined (e.g. detector is switched off), set the value to 0.
    % this way the averaging procedure isn't screwed up.
    if(not_nopcu0)  tab.electron0[where(isnan(tab.electron0) or isinf(tab.electron0))] = 0;
    tab.electron1[where(isnan(tab.electron1) or isinf(tab.electron1))] = 0;
    tab.electron2[where(isnan(tab.electron2) or isinf(tab.electron2))] = 0;
    tab.electron3[where(isnan(tab.electron3) or isinf(tab.electron3))] = 0;
    tab.electron4[where(isnan(tab.electron4) or isinf(tab.electron4))] = 0;

    % average electron ratio
    variable electron = tab.electron1 + tab.electron2 + tab.electron3 + tab.electron4;
    if(not_nopcu0 && use_el0)  electron += tab.electron0;
    electron[where(num_PCUs!=0)] /= num_PCUs[where(num_PCUs!=0)];  % el is 0 where it hasn't been measured
    xfl_info.electron = [xfl_info.electron, electron];

    append_struct_arrays(&xfl_info.pcu0, getPCUgti(tab.time, tab.pcu0_on));
    append_struct_arrays(&xfl_info.pcu1, getPCUgti(tab.time, tab.pcu1_on));
    append_struct_arrays(&xfl_info.pcu2, getPCUgti(tab.time, tab.pcu2_on));
    append_struct_arrays(&xfl_info.pcu3, getPCUgti(tab.time, tab.pcu3_on));
    append_struct_arrays(&xfl_info.pcu4, getPCUgti(tab.time, tab.pcu4_on));
  }
  return xfl_info;
}

%
% Code from src/misc/satellites/RXTE/RXTE_nr_PCUs_from_filename.sl
%
define RXTE_nr_PCUs_from_filename()
%!%+
%\function{RXTE_nr_PCUs_from_filename}
%\synopsis{counts how many PCUs were off from an *_xyoff_excl_* filename}
%\usage{Integer_Type RXTE_nr_PCUs_from_filename(String_Type filename)}
%!%-
{
  variable filename;
  switch(_NARGS)
  { case 1: filename = (); }
  { help(_function_name()); return; }

  variable part = strtok(filename, "_\.");
  return min([5, 6-string_match(part[wherefirst(part=="excl")-1], "off", 1)]);
  % string_match("<p1><p2>...<pn>off", "off", 1) == n+1
}

%
% Code from src/misc/satellites/RXTE/RXTE_nr_PCUs_from_filterfile.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define RXTE_nr_PCUs_from_filterfile()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{RXTE_nr_PCUs_from_filterfile}
%\synopsis{returns the number of PCUs switched on over time}
%\usage{Integer_Type[] RXTE_nr_PCUs_from_filterfile(String_Type filterfile[, Double_Type[] time])}
%\description
%     The number of PCUs switched on during an observation
%     may vary, which results in jumps in the lightcurve.
%     The filter file (your_extraction/filter/*.xfl) provides
%     time resolved information about the operating PCUs,
%     which is read out and returned.
%
%     ATTENTION:
%     If no time array is given the number of PCUs is returned
%     for the full length of the observation (no GTIs applied!).
%     In the other case the given time array HAS TO be in the
%     SATELLITE TIME SYSTEM and in SECONDS since RXTE started
%     operating (see 'MJDref_satellite').
%\seealso{MJDref_satellite, RXTE_nr_PCUs_from_filename}
%!%-
{
  variable filt, t;
  switch(_NARGS)
  { case 1: % return full list
    filt = ();
    return fits_read_table(filt).num_pcu_on;
  } { case 2: % return time matched list
    (filt,t) = ();
    filt = fits_read_table(filt);
    variable i, num = Integer_Type[length(t)];
    _for i (0, length(num)-1, 1) % match times
      num[i] = filt.num_pcu_on[where_min(abs(filt.time - t[i]))];
    return num;
  } {
    help(_function_name()); return;
  }
}

%
% Code from src/misc/satellites/RXTE/RXTE_obscat_info.sl
%
%%%%%%%%%%%%%%%%%%%
define RXTE_obscat_info()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{RXTE_obscat_info}
%\synopsis{reads an RXTE human-readable obs(ervation)cat(alogue)}
%\usage{Struct_Type RXTE_obscat_info(Integer_Type obscat_days[]);}
%\seealso{aitlib/rxte/readobscat.pro}
%!%-
{
  variable days;
  switch(_NARGS)
  { case 1: days = (); }
  { help(_function_name()); return; }

  variable verbose = qualifier_exists("verbose");

  % find local path
  variable path = NULL;
  try {
    foreach path ([local_paths.RXTE_obscat])
      if(stat_file(path)!=NULL)  break;
  } catch NotImplementedError: { vmessage("error (%s): local_paths.RXTE_obscat not defined",_function_name()); return NULL; };

  % return structure
  variable info = struct {
%   start_slew = Double_Type[0],
%     end_slew = Double_Type[0],
     in_occult = Double_Type[0],
    out_occult = Double_Type[0],
        in_saa = Double_Type[0],
       out_saa = Double_Type[0],
    start_good = Double_Type[0],
      end_good = Double_Type[0],
  };

  variable MJDREF = 49353.000696574074;
  variable mode = -1;
  variable day;
  foreach day ([days])
  { variable file = sprintf("%s/OCday%04d", path, day);
    if(stat_file(file)==NULL)
      vmessage("warning (%s): %s does not exist", _function_name(), file);
    else
    { if(verbose)  vmessage("reading %s", file);
      variable unit = fopen(file, "r");
      variable line;
      if(unit!=NULL)
      { while(not feof(unit) && fgets(&line, unit)>0)
	{ variable modstr = line[[0:8]];
	  if(modstr != "         ")  mode = modstr;
	  switch( mode )
          { case "  EVENTS:":
	      variable when = MJDREF + atof(line[[34:43]])/86400.;
  	      switch( line[[10:19]] )
%	      { case "start_slew":
%		  if(length(info.start_slew) > length(info.end_slew))
%		    info.start_slew[-1] = when;
%		  else
%		    info.start_slew = [info.start_slew, when];
%	      }
%	      { case "end_slew  ":
%		  info.end_slew = [info.end_slew, when];
%	      }
	      { case "in_occult ":
		  if(length(info.in_occult) > length(info.out_occult))
		    info.in_occult[-1] = when;
		  else
		    info.in_occult = [info.in_occult, when];
	      }
	      { case "out_occult":
		  info.out_occult = [info.out_occult, when];
	      }
	      { case "in_saa    ":
		  if(length(info.in_saa) > length(info.in_saa))
		    info.in_saa[-1] = when;
		  else
		    info.in_saa = [info.in_saa, when];
	      }
	      { case "out_saa   ":
		  info.out_saa = [info.out_saa, when];
		  % hack for missing 'in_saa'-tag in some obscats
		  if(length(info.in_saa)<length(info.out_saa))
		    info.out_saa = [info.out_saa, when];
	      }
	  }
	  { case "GOODTIME:":
	      info.start_good = [info.start_good, MJDREF + atof(line[[20:29]])/86400.];
	      info.end_good   = [info.end_good,   MJDREF + atof(line[[45:54]])/86400.];
	  }
	}
	()=fclose(unit);
      }
    }
  }

% info.start_slew = info.start_slew[[:length(info.end_slew)-1]];
  info.in_occult = info.in_occult[[:length(info.out_occult)-1]];
  info.in_saa = info.in_saa[[:length(info.out_saa)-1]];
  return info;
}

%
% Code from src/misc/satellites/RXTE/RXTE_plot_PCA_info.sl
%
define RXTE_plot_PCA_info()
%!%+
%\function{RXTE_plot_PCA_info}
%\synopsis{plots an overview of an RXTE observation (lc, GTI, SAA, bkg)}
%\usage{RXTE_plot_PCA_info(String_Type dirs[]);
%\altusage{Struct_Type RXTE_plot_PCA_info(String_Type dirs[]; get_info)}
%}
%\qualifiers{
%\qualifier{electron}{set to electron ratio threshold for data extraction}
%\qualifier{noback}{set if no bkg subtraction is to be performed}
%\qualifier{earthvle}{set if EarthVLE background model is to be used}
%\qualifier{faint}{set if Faint background model is to be used}
%\qualifier{q6}{set if Q6 background model is to be used (default is to test for earthvle,faint,q6)}
%\qualifier{skyvle}{set if SkyVLE background model is to be used (default)}
%\qualifier{exclusive}{set to search for data that was extracted
%                 with the exclusive keyword to pca_standard being set.}
%\qualifier{top}{set to read top-layer data}
%\qualifier{nopcu0}{set to search for data that was extracted ignoring PCU0}
%\qualifier{fivepcu}{plot count-rates wrt to whole PCA, i.e., normalizing to 5 PCUs.
%               Default is to plot the average countrate per PCU.}
%\qualifier{charsize_obsid}{}
%\qualifier{get_info}{returns the info structure}
%}
%\description
%    The elements of \code{dirs} may contain globbing expressions.
%\seealso{aitlib/rxte/rxtescreen.pro}
%!%-
{
  variable dirs;
  switch(_NARGS)
  { case 1: dirs = (); }
  { help(_function_name()); return; }

  variable dir, real_dirs = String_Type[0];
  foreach dir (glob(dirs))
    if(stat_is("dir", stat_file(dir).st_mode))
      real_dirs = [real_dirs, dir];
  dirs = real_dirs;

  % call RXTE_PCA_info
  variable qualifiers = __qualifiers();
  variable info = RXTE_PCA_info(dirs;; qualifiers);

  % check info (should top and exclusive have been set?)
  if(length(info.lc.time)==0)
  {
    vmessage("warning (%s): no data found", _function_name());
    ifnot(qualifier_exists("exclusive") || qualifier_exists("top"))
    { message("trying options 'exclusive' and 'top':\n");
      info = RXTE_PCA_info(dirs;; struct_combine(qualifiers, struct { exclusive, top }));
      if(length(info.lc.time)==0)
        vmessage("warning (%s): still no data found", _function_name());
    }
    if(length(info.lc.time)==0)
      if(qualifier_exists("get_info"))
        return info;
      else
	return;
  }

  % prepare plotting
  variable time0 = round(info.lc.time[0]);
  variable tmin = info.lc.time[ 0] - time0;
  variable tmax = info.lc.time[-1] - time0;
  variable dt = 0.05*(tmax-tmin);
  tmin -= dt;
  tmax += dt;

  % plot
  multiplot([2,1,3]);
  xrange(tmin, tmax);

  % lightcurve
  yrange(0.9*min(info.lc.rate), 1.1*max(info.lc.rate));
  ylabel("PCA counts/s/PCU");
  if(qualifier_exists("fivepcu"))  ylabel("PCA counts/s");
  plot_with_err(info.lc.time-time0, info.lc.rate, info.lc.error);

  % electron ratio
  ylabel("El. ratio");
  yrange(0, _max(0.25, 1.02*info.xfl.electron));
  init_plot;
  variable i1 = 0;
  variable i2, len = length(info.xfl.time);
  _for i2 (1, len, 1)
    if(i2==len || info.xfl.time[i2]>info.xfl.time[i2-1]+32/86400.)
    { color(1); oplot(info.xfl.time[[i1:i2-1]]-time0, info.xfl.electron[[i1:i2-1]]);
      i1 = i2;
    }
  oplot([tmin, tmax], qualifier("electron", 0.1)*[1, 1]);

  % multiple horizontal lines (=> no numerical y-axis)
  variable i, X, y=11;
  mpane(3);
  _pgsci(1);
  _pgswin(tmin, tmax, 0, y+1);
  _pgbox("BCNST", 0, 1, "BC", 0, 1);

  % ObsID
  _pgptxt(tmin-dt, y-0.5, 0, 0.5, "ObsID");
  _pgsch(qualifier("charsize_obsid", 0.8));
  _for i (0, length(info.obstime.start)-1, 1)
  { X = [info.obstime.start[i], info.obstime.stop[i]]-time0;
    ()=_pgline(2, X, [y, y]);
    _pgptxt((X[0]+X[1])/2, y-0.25-0.5*(-1)^i, 0, 0.5, strtok(dirs[i], "/")[-1]);
  }
  y -= 2;
  _pgsch(1);

  % all others are done in the same way
  variable label_start_stop;
  foreach label_start_stop ([
    { "SAA",    info.obscat.in_saa,     info.obscat.out_saa },
    { "occult", info.obscat.in_occult,  info.obscat.out_occult },
    { "Good",  info.obscat.start_good, info.obscat.end_good },
    { "GTI",   info.gti.start,         info.gti.stop },
    { "PCU 0", info.xfl.pcu0.start,    info.xfl.pcu0.stop },
    { "PCU 1", info.xfl.pcu1.start,    info.xfl.pcu1.stop },
    { "PCU 2", info.xfl.pcu2.start,    info.xfl.pcu2.stop },
    { "PCU 3", info.xfl.pcu3.start,    info.xfl.pcu3.stop },
    { "PCU 4", info.xfl.pcu4.start,    info.xfl.pcu4.stop },
  ])
  { _pgptxt(tmin-dt, y-0.2, 0, 0.5, label_start_stop[0]);
    _for i (0, length(label_start_stop[1])-1, 1)
    { X = [label_start_stop[1][i], label_start_stop[2][i]]-time0;
      if(X[0]>tmin || tmax<X[1])
        ()=_pgline(2, X, [y, y]);
    }
    y--;
  }
  % FIXME:
  _pgptxt((tmin+tmax)/2, y-2, 0, 0.5, sprintf("days since MJD %.0f", time0));

  multiplot(1);

  % return info if requested
  if(qualifier_exists("get_info"))  return info;
}

%
% Code from src/slang/astronomy/Roche_critical.sl
%
define Roche_critical()
%!%+
%\function{Roche_critical}
%\synopsis{calculates the critical potential at L1 and the corresponding radius in z-direction}
%\usage{Struct_Type crit = Roche_critical(Double_Type mq);}
%\description
%  The Roche potential describes the motion of a binary consis-
%  ting of two objects, M1 and M2. The mass ratio mq = M1/M2
%  must be given. From the position of the first Lagrange point
%  'xL1', which is a saddle point in the Roche potential, the
%  value of the critical potential 'critpot' at this point is
%  calculated. The corresponding equipotential surface describes
%  the inner most closed surface, which is also known as 'Roche
%  lobe'. Since the Roche potential does not depend on the
%  binary rotation, the radius 'rcrit' of the star in z-direction
%  (perpendicular to the orbital plane) is often used as the
%  star's radius.
%  
%  All three values are returned by the following structure:
%    xL1     - distance of M1 to L1 in units of the binary
%              displacement
%    critpot - value of the critical potential at L1
%    rcrit   - radius of M1 in z-direction in units of the
%              binary displacement
%\seealso{Roche_potential}
%!%-
{
  variable mq;
  switch(_NARGS)
    { case 1: mq = (); }
    { help(_function_name()); return; }

  % position of L1
  variable xL1 = find_function_maximum(&Roche_potential, 1e-8, 1.-1e-8; qualifiers = struct{ q=mq });
  % value of critical potential at L1
  variable critpot = Roche_potential(xL1; q=mq);
  % critical radius in z-direction
  variable rcrit = find_multiargumentfunction_value(&Roche_potential, critpot, 0.0, 0.0, [1e-8, xL1]; qualifiers = struct{ q=mq });

  return struct { xL1 = xL1, critpot = critpot, rcrit = rcrit };
}

%
% Code from src/slang/astronomy/Roche_equipotential_volume.sl
%
define Roche_equipotential_volume()
%!%+
%\function{Roche_equipotential_volume}
%\synopsis{computes the volume inside an Roche equipotential surface}
%\usage{Double_Type Roche_equipotential_volume}
%\qualifiers{
%\qualifier{phi}{[= critical potential] the surface's Roche potential}
%\qualifier{N}{[= 100] number of subdivisions in x and y for numerical integration}
%}
%!%-
{ variable q;
  switch(_NARGS)
  { case 1: q = (); }
  { help(_function_name()); return; }

  variable RocheLobe = qualifier("RocheLobe", NULL);
  if(RocheLobe==NULL)  RocheLobe = Roche_lobe_dimension(q);
  variable phi = qualifier("phi", RocheLobe.potential);
  variable N = qualifier("N", 100);
  variable x, X = [RocheLobe.xmin:RocheLobe.xmax:#N];
  variable y, Y = [0:RocheLobe.ymax:#N];
  variable V = 0;
  variable wx = 1;
  foreach x (X)
  { if(x==X[-1])  wx = 1;
    variable wy = 1;
    foreach y (Y)
    { if(y==Y[-1])  wy = 1;
      if(x!=0 || y!=0)
      { if(Roche_potential(x, y, 0; q=q) < phi)
          V += wx*wy*find_multiargumentfunction_value(&Roche_potential, phi, x, y, [0, 1.01*RocheLobe.zmax]; qualifiers=struct{q=q});
      }
      else
        V += wx*wy*find_multiargumentfunction_value(&Roche_potential, phi, x, y, [0.99, 1.01]*RocheLobe.zmax; qualifiers=struct{q=q});
      if(wy==1||wy==2)  wy = 4;  else  wy = 2;
    }
    if(wx==1||wx==2)  wx = 4;  else  wx = 2;
  }
  return 4*V/9.*(X[1]-X[0])*(Y[1]-Y[0]);  % 4 due to symmetry in y and z, 9 is from the 2D Simpson rule
}

%
% Code from src/slang/astronomy/Roche_lobe_dimension.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define Roche_lobe_dimension()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Roche_lobe_dimension}
%\synopsis{determines the size of the Roche lobe}
%\usage{Struct_Type Roche_lobe_dimension(Double_Type q)}
%\description
%    The structure has the following fields:\n
%    \code{potential}: the critical Roche potential of the Roche lobe\n
%    {\code{x}/\code{y}/\code{z}}{\code{min}/\code{max}}: the coordinates of the envelopping box
%!%-
{
  variable q;
  switch(_NARGS)
  { case 1: q = (); }
  { help(_function_name()); return; }

  variable eps = 1e-8;
  variable qualifiers = struct { q = q };  % __qualifiers();
  variable xmax = find_function_maximum(&Roche_potential, eps, 1.-eps; qualifiers=qualifiers);
  variable phi = Roche_potential(xmax;; qualifiers);
  variable xmin = find_function_value(&Roche_potential, phi, -xmax, -eps; qualifiers=qualifiers);
  variable ymax = find_multiargumentfunction_value(&Roche_potential, phi, 0, [eps, xmax]; qualifiers=qualifiers);
  variable zmax = find_multiargumentfunction_value(&Roche_potential, phi, 0, 0, [eps, xmax]; qualifiers=qualifiers);
  return struct {
    potential = phi,
    xmin =  xmin,
    xmax =  xmax,
    ymin = -ymax,
    ymax =  ymax,
    zmin = -zmax,
    zmax =  zmax,
  };
}

%
% Code from src/slang/astronomy/Roche_lobe_surface.sl
%

% only activate if module vector could be loaded
#ifeval __get_reference("vector")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%%
define Roche_lobe_surface()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Roche_lobe_surface}
%\synopsis{calculates surface elements of a binary star}
%\usage{Struct_Type = Roche_lobe_surface(Double_Type mq, Double_Type fill, Integer_Type ntheta);}
%\qualifiers{
%  \qualifier{noRoche}{neglect the Roche potential -> spherical star}
%  \qualifier{noNorm}{do not return the normal vectors}
%  \qualifier{noArea}{do not return the areas}
%  \qualifier{poly}{return the polygons defining the elements}
%  \qualifier{rotate}{rotate the star around its (z-)axis by the
%            given degrees (default: 0)}
%  \qualifier{eps}{numerical precision (default: 1e-8)}
%  \qualifier{chatty}{be chatty}
%}
%\description
%  This function calculates the surface of the primary star
%  in a binary, which is deformed by the Roche potential.
%  Since this potential can not be solved analytically the
%  surface must be divided into elements. The number of
%  elements into theta direction can be specified by the
%  'ntheta' argument. The total number of surface elements
%  taking into account the singularity at the poles and the
%  symmetry in theta is here
%    2*ntheta * (ntheta-2) + 2
%  The Roche potential for a spherical orbit depends on the
%  mass ratio mq = M_primary / M_secondary
%  To determine the size of the star, the undeformed radius
%  Rz in direction of the angular velocity vector (omega)
%  can be used via the so-called fill factor:
%    fill = Rz / R_crit
%  where R_crit is the critical undeformed radius for a
%  star, which fills its Roche lobe completely. In that
%  case, the potential value at the surface is equal to
%  that at the first Lagrange point, which might lead to
%  mass transfer onto the companion (Roche lobe overflow).
%
%  The returned structure contains the following fields:
%    Vector_Type[] r - position vector to each element
%                      starting at the stars center
%    Vector_Type[] n - normal vector of each element
%    Vector_Type[] A - area of each element
%    Vector_Type[] p - if the 'poly' qualifier is given,
%                      it contains the polygons defining
%                      each surface element
%
%  NOTE: The origin of the reference frame is in the center
%        of the star and the first Lagrange point is on
%        the x-axis, while the rotation axis of the star is
%        along the z-axis!
%        If you want to rotate the star around its rotation
%        axis (z-axis), use the 'rotate'-qualifier!
%\seealso{Roche_critical, Roche_potential, radius_to_unit, Vector_Type}
%!%-
{
 variable mq, fill, ntheta;
 switch(_NARGS)
   { case 3: (mq, fill, ntheta) = (); mq = 1./mq; }
   { help(_function_name()); return; }

 if (qualifier_exists("chatty")) vmessage("initializing", _function_name);
 % surface polygons expressed in spherical coordinates
 variable pphi    = 2.*PI*[0:2*ntheta-2]/(2*ntheta-2); % phi (one more for phi=2PI)
 variable ptheta  = 1.*PI*[0:ntheta-1]/(ntheta-1);     % theta
 variable pradius = Double_Type[2*ntheta-1,ntheta];    % radius as function of [phi, theta]

 % convert fill factor into radius in units of binary displacement
 variable radius = radius_to_unit(fill, "fill", "disp"; mq = mq);
  
 if (qualifier_exists("chatty")) vmessage("calculating radii", _function_name);
 % calculate radii
 if (qualifier_exists("noRoche")) pradius = pradius + radius;
 % by numerically solve Roche potential for given radius
 else
 {
   % critical potential on surface
   variable surfpot = Roche_potential(0.0, 0.0, radius; q = mq);
   variable p, t;
   _for p (0, 2*ntheta-2, 1)
   {
     _for t (0, ntheta-1, 1)
     {
       % singularity on poles and meridian
       if (t == 0 && p > 0) pradius[p,t] = pradius[0,0]; % theta=0
       else if (t == ntheta-1 && p > 0) pradius[p,t] = pradius[0,ntheta-1]; % theta=PI
       else if (p == 2*ntheta-2) pradius[p,t] = pradius[0,t]; % meridian
       % find radius corresponding to potential value
       else
       {
         variable tmprad = 0., dr = .1, pot = -_Inf, tmppot, j = 0;
         variable eps = qualifier("eps", 1e-8);
	 % start at radius 0 (=infinite negative potential) and increase radius
	 % in steps until corresponding potential matches surface potential
         while (surfpot-pot > eps)
         {
	   % increase radius and calculate potential at this point
           tmprad += dr;
           tmppot = Roche_potential(tmprad, pphi[p], ptheta[t]; q = mq, sph);
           % acutal potential to large -> step back and decrease stepsize to
	   % approach surface potential slower than before
           if (tmppot > surfpot)
           {
             tmprad -= dr;
             dr *= 0.1;
	     j=0;
           }
           % if the derivative negative, we are "over the hill" and outside of
	   % the Roche lobe (fill=1), and have to step back
           else if ((tmppot - pot) / dr < 0.)
           {
             if (j==0) { tmprad -= dr; dr *= 0.1; }
             else tmprad -= 10.*dr + j*dr; % go back faster if we're here again
             j++;
             pot = Roche_potential(tmprad, pphi[p], ptheta[t]; q = mq, sph);
           }
           % on the right way -> overwrite actual potential value and proceed
           else
           {
             pot = tmppot;
             j=0;
           }
         }
	 % set found radius
         pradius[p,t] = tmprad;
       }
     }
   }
 }

 % build structure
 if (qualifier_exists("chatty")) vmessage("building structure", _function_name);
 variable np = 2*ntheta*(ntheta-2)+2; % number of surface elements (rectangles)
 variable surface = struct { r = Vector_Type[np] };
 ifnot (qualifier_exists("noNorm")) surface = struct_combine(surface, struct { n = Vector_Type[np] });
 ifnot (qualifier_exists("noArea")) surface = struct_combine(surface, struct { A = Double_Type[np] });
 if (qualifier_exists("poly")) surface = struct_combine(surface, struct { p = Vector_Type[np] });
 variable rot = qualifier("rotate", 0.);
 % loop
 variable tto;
 _for t (1, ntheta-1, 1)
 {
   tto = (t-1)*(2*ntheta-2);
   _for p (0, 2*ntheta-3, 1) % exclude P = 2PI
   {
     % polygons
     if (qualifier_exists("poly"))
       surface.p[tto+p] = vector([pradius[p,t-1], pradius[p,t], t < ntheta-1 ? pradius[p+1,t] : Double_Type[0], pradius[p+1,t-1], t > 1 ? pradius[p,t-1] : Double_Type[0]],
				 [pphi[p], pphi[p], pphi[p+1], t < ntheta-1 ? pphi[p+1] : Double_Type[0], t > 1 ? pphi[p] : Double_Type[0]],
				 [ptheta[t-1], ptheta[t], t < ntheta-1 ? ptheta[t] : Double_Type[0], ptheta[t-1], t > 1 ? ptheta[t-1] : Double_Type[0]]; sph);
     % position vector
     surface.r[tto+p] = 1./(3+(t>1 && t<ntheta-1))
                        * ((t > 1 ? vector(pradius[p,t-1], pphi[p],  ptheta[t-1]; sph) : vector(0.,0.,0.)) % element including north pole is a triangle
      		           + vector(pradius[p,t], pphi[p], ptheta[t]; sph)
			   + (t < ntheta-1 ? vector(pradius[p+1,t], pphi[p+1], ptheta[t]; sph) : vector(0., 0., 0.)) % element including south pole is a triangle
			   + vector(pradius[p+1,t-1], pphi[p+1], ptheta[t-1]; sph)
			  );
     variable a, b;
     ifnot (qualifier_exists("noNorm") && qualifier_exists("noArea"))
     {
       a = vector(pradius[p,t], pphi[p], ptheta[t]; sph) - vector(pradius[p,t-1], pphi[p], ptheta[t-1]; sph);
       b = (t < ntheta-1 ? vector(pradius[p+1,t], pphi[p+1], ptheta[t]; sph) : vector(pradius[p+1,t-1], pphi[p+1], ptheta[t-1]; sph)) - vector(pradius[p,t], pphi[p], ptheta[t]; sph);
     }
     % normal vector
     ifnot (qualifier_exists("noNorm"))
     {
       surface.n[tto+p] = crossprod(a,b);
       surface.n[tto+p] = 1./vector_norm(surface.n[tto+p])*surface.n[tto+p];
     }
     % area
     ifnot (qualifier_exists("noArea"))
     {
       variable na = vector_norm(a), nb = vector_norm(b);
       surface.A[tto+p] = na*nb*sqrt(1.-(a*b)/na/nb) / ((t == 1 || t == ntheta-1) + 1.); % divide by two (area of triangle)
     }
     % evntually rotate
     if (rot > 0 && rot < 360) {
       surface.p[tto+p] = vector_rotate(surface.p[tto+p], vector(0, 0, 1), PI/180*rot);
       surface.r[tto+p] = vector_rotate(surface.r[tto+p], vector(0, 0, 1), PI/180*rot);
       surface.n[tto+p] = vector_rotate(surface.n[tto+p], vector(0, 0, 1), PI/180*rot);
     }
   }
 }

 return surface;
}

#endif

%
% Code from src/slang/astronomy/Roche_potential.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define Roche_potential()
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Roche_potential}
%\usage{Double_Type Roche_potential(Double_Type x[, y[, z]]);}
%\qualifiers{
%\qualifier{q}{mass ratio}
%\qualifier{sph}{coordinates are considered as spherical coordinates r, phi, theta}
%}
%\description
%   The default values for y and z is 0.
%
%   The Roche potential has the following form:
%   \code{-1/sqrt(x^2 + y^2 + z^2 ) - q/sqrt( (x-1.)^2 + y^2 + z^2 ) - (1+q)/2.*( (x-q/(1.+q))^2 + y^2 )}
%!%-
{
  variable x=0, y=0, z=0;
  switch(_NARGS)
  { case 1:  x        = (); }
  { case 2: (x, y)    = (); }
  { case 3: (x, y, z) = (); }
  { help(_function_name()); return; }

  variable q = qualifier("q", .567);
%  if(qualifier_exists("verbose"))
%    vmessage("q=%f", q);
  if (qualifier_exists("sph"))
  {
    variable siny, cosy, sinz = sin(z);
    (siny, cosy) = sincos(y);

    return -1./x - q/sqrt(x^2 - 2.*x*cosy*sinz + 1) - (1+q)/2.*( (x*sinz*cosy-q/(1.+q))^2 + (x*sinz*siny)^2 );
  }
  else return - 1./sqrt(x^2 + y^2 + z^2 ) - q/sqrt( (x-1.)^2 + y^2 + z^2 ) - (1+q)/2.*( (x-q/(1.+q))^2 + y^2 );
}

%
% Code from src/slang/strings/Roman.sl
%
%%%%%%%%%%%%
define Roman(n) {
%!%+
%\function{Roman}
%\synopsis{translates n to upper-case string with roman numeral}
%\usage{String_Type res = Roman(Integer_Type n)}
%\qualifiers{
%\qualifier{latex}{typeset minus sign ("$-$"R)}
%\qualifier{toobig}{[=""] string that is returned of n is larger
%                    than the largest known Roman numeral (3999) }
%}
%\description
% Converts an integer into a uppercase roman numeral.
% Even though not known in Roman times, negative numbers are allowed.
%
% Algorithm based on
% https://www.geeksforgeeks.org/converting-decimal-number-lying-between-1-to-3999-to-roman-numerals/
%
%\seealso{roman}
%!%-
   variable minus = "-";
   if(qualifier_exists("latex")) {minus = "$-$"R;}
   variable oob = qualifier("toobig", "");
   variable nums,w1,w2;
   variable array = 1;
   
   if (typeof(n)==Array_Type) {
       return array_map(String_Type, &Roman, n);
   }

   if (typeof(n)!=Integer_Type) {
       throw UsageError,sprintf("%s: argument must be of Integer type",_function_name());
   }

   if (n==0) {
       return "0";
   }
   
   variable num= [1,4,5,9,10,40,50,90,100,400,500,900,1000]; 
   variable sym = ["I","IV","V","IX","X","XL","L","XC","C","CD","D","CM","M"];

   variable i=12;
   variable res="";
   if (n<0) {
       res=minus;
       n=-n;
   } 
   
   while(n>0) { 
       variable div = n/num[i]; 
       n = n mod num[i];
       while(div>0) {
	   div--;
	   res+=sym[i];
       } 
       i--; 
   }

   return(res);
}

%%%%%%%%%%%%
define roman(n)
%%%%%%%%%%%%
%!%+
%\function{roman}
%\synopsis{replaces an integer with lowercase Roman numeral strings}
%\usage{String_Type rom = roman(Integer_Type);}
%\synopsis{translates n to lower-case string with roman numeral}
%\usage{String_Type roman = romann(Integer_Type n)}
%\qualifiers{
%\qualifier{latex}{typeset minus sign ("$-$"R)}
%\qualifier{toobig [=""]}{string that is returned of n is larger
%                    than the largest known Roman numeral}
%}
%\seealso{Roman}
%!%-
{
  return strlow(Roman(n;; __qualifiers));
}

define roman2int()
%!%+
%\function{roman2int}
%\usage{Integer_Type=roman2int(String_Type)}
%\synopsis{translates a roman numeral into an integer}
%\description
% This function converts roman numerals into integers.
% The function is case insensitive and array safe.
%\seealso{roman,Roman}
%!%-
{
    variable r=();

    if (typeof(r)==Array_Type) {
	return array_map(Integer_Type,&roman2int,r);
    }
    
    r=strup(r);

    % deal with negative numbers
    variable sn=+1;
    if (substr(r,1,1)=="-") {
	sn=-1;
	r=substr(r,2,strlen(r));
    }
    
    variable vals=Assoc_Type[Integer_Type];
    vals["0"]=0;
    vals["I"]=1;
    vals["V"]=5;
    vals["X"]=10;
    vals["L"]=50;
    vals["C"]=100;
    vals["D"]=500;
    vals["M"]=1000;

    variable res=0;
    variable i;
    for (i=1; i<=strlen(r); i++) {
	variable s1=vals[substr(r,i,1)];
	if (i<strlen(r)) {
	    variable s2=vals[substr(r,i+1,1)];
	    if (s1>=s2) {
		res+=s1;
	    } else {
		res+=s2-s1;
		i++;
	    }
	} else {
	    res+=s1;
	}
    }

    res=sn*res;

    % sanity check (otherwise we silently return erroneous results for
    % illegal roman numerals)
    if (r!=Roman(res)) {
	throw UsageError,sprintf("%s: %s is not a valid roman numeral",_function_name(),r);
    }

    
    return sn*res;
}

%
% Code from src/misc/SRT/SRT_image.sl
%
define SRT_image()
%!%+
%\function{SRT_image}
%\synopsis{extracts an image from a SRT data structure containing an npoinscan}
%\usage{Float_Type img[] = SRT_image(Struct_Type data);}
%\qualifiers{
%\qualifier{plot}{plots the image}
%}
%\seealso{SRT_read}
%!%-
{
  variable data;
  switch(_NARGS)
  { case 1: data = (); }
  { help(_function_name()); return; }

  variable BEAMWIDTH = 7.;
  variable img = Float_Type[5,5];
  variable y;
  _for y (0, 4, 1)
  {
    variable eloff = (y-2)*BEAMWIDTH/2;
    variable irow = where(data.eloff == eloff);
    variable n = length(irow);
    switch(n)
    { case 5: img[y,*] = data.avflux[irow]; }
    { vmessage("warning (%s): row #%d at eloff=%.1f deg contains %d != 5 elements", _function_name(), y, eloff, n);
      variable azoff = data.azoff[irow];
      variable maxAbsAzoff = max(abs(azoff));
      variable i;
      i = where(-maxAbsAzoff == azoff                            ); if(length(i)>0)  img[y,0] = mean(data.avflux[irow[i]]);
      i = where(-maxAbsAzoff <  azoff <  0                       ); if(length(i)>0)  img[y,1] = mean(data.avflux[irow[i]]);
      i = where(                azoff == 0                       ); if(length(i)>0)  img[y,2] = mean(data.avflux[irow[i]]);
      i = where(                         0 < azoff <  maxAbsAzoff); if(length(i)>0)  img[y,3] = mean(data.avflux[irow[i]]);
      i = where(                             azoff == maxAbsAzoff); if(length(i)>0)  img[y,4] = mean(data.avflux[irow[i]]);
    }
  }

  if(qualifier_exists("plot"))
  {
    xlin; % xrange(-1.25*BEAMWIDTH, 1.25*BEAMWIDTH);
    ylin; % yrange(-1.25*BEAMWIDTH, 1.25*BEAMWIDTH);
    plot_image(img, 0, [-BEAMWIDTH:BEAMWIDTH:#5], [-BEAMWIDTH:BEAMWIDTH:#5]);
  }

  return img;
}

%
% Code from src/misc/SRT/SRT_mw_plots.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

%%%%%%%%%%%%%%
define SRT_get_mw_spectra() {
%%%%%%%%%%%%%%
%!%+
%\function{SRT_get_mw_spectra}
%\synopsis{computes Milky Way spectrum from SRT data}
%\usage{Struct_Type[] spectra = SRT_get_mw_spectra(Struct_Type data, Integer_Type[] chunks}
%\description
%	This function requires as input the SRT data in the format
%	given by SRT_read() and an array of integer numbers which are the chunks
%	with the spectrum data. If an array of chunk numbers is given then all
%	corresponding spectra are calculated and returned.
%	The spectra are determined with SRT_spectrum and all qualifiers of this
%	function can be used.
%\example
%	variable data = SRT_read("milkyway.rad");
%	variable spectra = SRT_get_mw_spectra(data,[1,4,7]);
%\seealso{SRT_spectrum, SRT_read}
%!%-

	variable data,chunks;
	switch(_NARGS)
	{ case 2: (data, chunks) = (); }
	{ help(_function_name()); return; }

	%find out how many spectra should be plotted
	variable spectra = Struct_Type[length(chunks)];
	variable i;

	%get data of the spectra
	for (i=0; i<length(chunks); i++) {
		spectra[i] = SRT_spectrum(data[chunks[i]];;__qualifiers);
	}

	return spectra;
}

%%%%%%%%%%%%%%
define SRT_plot_mw_spectra() {
%%%%%%%%%%%%%%
%!%+
%\function{SRT_plot_mw_spectra}
%\synopsis{plots Milky Way spectra from SRT data}
%\usage{SRT_plot_mw_spectra(Struct_Type[] spectrum);}
%\altusage{SRT_plot_mw_spectra(Struct_Type[] data, Integer_Type[] chunks);}
%\qualifiers{
%\qualifier{pad}{additional space between plot frame and data limits}
%\qualifier{fMax}{give the frequency of the cloud to be plotted, has to be of same length as spectra/chunks}
%\qualifier{fConfMax}{upper confidence level of the maximal frequency of the cloud to be plotted, has to be of same length as spectra/chunks}
%\qualifier{fConfMin}{lower confidence level of the maximal frequency of the cloud to be plotted, has to be of same length as spectra/chunks}
%\qualifier{fMaxLin}{type of line used to plot fMax}
%\qualifier{fConfLine}{type of line used to plot fConfMin and fConfMax}
%\qualifier{xlabel}{label for x-axis}
%\qualifier{ylabel}{label for y-axis}
%\qualifier{title}{title for plot, either as single title or for each plot individually as an array of strings}
%\qualifier{name}{file name of plot, either a single one or for each plot as a an array of string, specify file format, i.e., mw_scan.pdf}
%\qualifier{return_xfig}{return xfig-data instead of plotting}
%}
%\description
%	The function automatically plots Milky Way spectra obtained with the SRT.
%	It can be given a single or an array of spectra or the data and an array of chunk numbers. The required data format is based on the output of SRT_get_mw_spectrum and all of its qualifiers can be used.
%	It is possible to give the previously determined maximal frequency of the cloud and its upper and lower confidence interval as qualifier. This will create lines in the plot at the provided frequency.
%\seealso{SRT_get_mw_spectra, SRT_read, SRT_spectrum}
%!%-
%}

	variable spectra, data, chunks;
	switch(_NARGS)
	{ case 1: spectra = (); }
	{ case 2: (data, chunks) = (); }
	{ help(_function_name()); return; }

	%get spectra data
	if(_NARGS==2)
		spectra = SRT_get_mw_spectra(data,chunks;;__qualifiers);

	%create as many plots as spectra
	variable nSpectra = length(spectra);
	variable pl = Struct_Type[nSpectra];

	variable i, xmin, xmax, ymin, ymax, title, name;
	variable world_qualis = struct {
		padx = qualifier("pad",0.05),
		pady = qualifier("pad",0.05),
	};

	%allow plotting of max frequency and confidence interval
	%have to be of same length as number of spectra
	variable fMax, fConfMin, fConfMax;
	variable plot_fMax, plot_fConfMin, plot_fConfMax;
	if(qualifier_exists("fMax") && length(qualifier("fMax")) == nSpectra) {
		fMax = qualifier("fMax");
		plot_fMax = 1;
	}
	else
		plot_fMax = 0;
	
	if(qualifier_exists("fConfMin") && length(qualifier("fConfMin")) == nSpectra) {
		fConfMin = qualifier("fConfMin");
		plot_fConfMin = 1;
	}
	else
		plot_fConfMin = 0;
	
	if(qualifier_exists("fConfMax") && length(qualifier("fConfMax")) == nSpectra) {
		fConfMax = qualifier("fConfMax");
		plot_fConfMax = 1;
	}
	else
		plot_fConfMax = 0;	

	%do the plotting
	for(i=0; i<nSpectra; i++) {
		xmin = min(spectra[i].bin_lo);
		xmax = max(spectra[i].bin_lo);
		ymin = min(spectra[i].value);
		ymax = max(spectra[i].value);

		pl[i] = xfig_plot_new();
		pl[i].world(xmin,xmax,ymin,ymax;;struct_combine(__qualifiers,world_qualis));
		pl[i].hplot(spectra[i].bin_lo,spectra[i].value;;__qualifiers);

		if(plot_fMax == 1)
			pl[i].plot([fMax[i],fMax[i]],[0.,1.];line = qualifier("fConfLine",0),world10);
		if(plot_fConfMax == 1)
			pl[i].plot([fConfMax[i],fConfMax[i]],[0.,1.];line = qualifier("fConfLine",1),world10);
		if(plot_fConfMin == 1)
			pl[i].plot([fConfMin[i],fConfMin[i]],[0.,1.];line = qualifier("fConfLine",1),world10);
		
		pl[i].x2axis(;ticlabels=0);
		pl[i].y2axis(;ticlabels=0);
		if(qualifier_exists("vLSR"))
			pl[i].xlabel("$v$ [km/s]"R);
		else if(qualifier_exists("xlabel"))
			pl[i].xlabel(qualifier("xlabel"));
		else
			pl[i].xlabel("$\nu$ [MHz]"R);

		if(qualifier_exists("ylabel"))
			pl[i].ylabel(qualifier("ylabel"));
		else
			pl[i].ylabel("$T_\mathrm{Ant}$ [K]"R);

		
		title = qualifier("title");
		if(qualifier_exists("title") && length(title)==1)
			pl[i].title(title);
		else if(qualifier_exists("title") && length(title) == nSpectra)
			pl[i].title(title[i]);

		name = qualifier("name");
		ifnot(qualifier_exists("return_xfig")){
			if(qualifier_exists("name") && length(name)==1) {
				if(i<10)
					pl[i].render("0"+string(i+1) + "_" + name);
				else 
					pl[i].render(string(i+1) + "_" + name);
			}
			else if(qualifier_exists("name") && length(name)==nSpectra)
				pl[i].render(name[i]);
			else {
				if(i<9)
					pl[i].render("mw_spec_0" + string(i+1) + ".pdf");
				else
					pl[i].render("mw_spec_" + string(i+1) + ".pdf");
			}
		}
	}

	if(qualifier_exists("return_xfig"))
		return pl;
}


#endif

%
% Code from src/misc/SRT/SRT_read.sl
%
define SRT_read()
%!%+
%\function{SRT_read}
%\synopsis{reads SRT data structures from a SRT .rad file}
%\usage{Struct_Type data[] = SRT_read(String_Type filename);}
%\qualifiers{
%\qualifier{verbose}{}
%\qualifier{onechunk}{read all data into one structure instead of an array
%                (default if file has no comments)}
%\qualifier{bins_to_cut}{[=8]: number of bad bins at high and low frequencies}
%\qualifier{position}{[lat,longw] position of the SRT
%                (default: lat=49.90 and longw=349.10)}
%}
%\seealso{SRT_spectrum, SRT_image}
%!%-
{
  variable filename = ();

  variable verbose = qualifier_exists("verbose");
  variable onechunk = qualifier_exists("onechunk");

  variable F = fopen(filename, "r");
  variable lines = fgetslines(F);
  ()=fclose(F);

  lines = lines[ onechunk ? [0, wherenot(array_map(Integer_Type, &string_match, lines, "^*"))]
		          : where(lines!="* ERROR communicating with radio\x0D\n")
	       ];
  variable nlines = length(lines);

  variable position = qualifier("position", [49.9, 349.1]);  % [lat, longw] @ Remeis

  variable lat, longw;
  if(   2!=sscanf(lines[0], "* STATION LAT= %f DEG LONGW= %f", &lat, &longw)
     || qualifier_exists("position") )
  {
    (lat, longw) = (position[0], position[1]);
  }

  variable i, nchunks=0;
  _for i (0, length(lines)-1, 1)
    if(lines[i][0] == '*')
      nchunks++;

  variable data = Struct_Type[max([1,nchunks])];
  variable ichunk = -1,  % index of current chunk
           n = 0;  % number of lines in current chunk
  variable tsys = NULL;
  _for i (0, nlines, 1)
    if(i<nlines && lines[i][0]!='*')
      n++;  % one more line belonging to current chunk
    else
    {
      % read tsys for this chunk or set it to previous
      if ((i < nlines) && string_match(lines[i], "tsys \\([0-9]+[.0-9]*\\)")) {
	variable pos, len;
	(pos, len) = string_match_nth (1);
	tsys = atof(lines[i][[pos:pos+len]]);
      }
      
      ichunk++;
      if(i<nlines || nchunks==0)  % start a new chunk
        data[ichunk] = struct { description=(i<nlines ? lines[i][[2:strlen(lines[i])-3]] : "no comment"),
				lat=lat, longw=longw,
				Y, D, H, M, S, MJD,
				az, el, azoff, eloff, RA, dec, glon, glat,
				f0, df, mode, nbins, spec, avflux,
				vLSR, tsys=tsys
			      };
      ifnot(nchunks)  ichunk++;  % pretend that another chunk was finished, since first chunk will never finish if file does not contain comments
      if(ichunk>0)
      { % allocate arrays for previous chunk (now that we know its length)
	data[ichunk-1].Y     = Short_Type [n];
	data[ichunk-1].D     = Short_Type [n];
	data[ichunk-1].H     = Char_Type  [n];
	data[ichunk-1].M     = Char_Type  [n];
	data[ichunk-1].S     = Char_Type  [n];
	data[ichunk-1].MJD   = Double_Type[n];

	data[ichunk-1].az    = Float_Type [n];
	data[ichunk-1].el    = Float_Type [n];
	data[ichunk-1].azoff = Float_Type [n];
	data[ichunk-1].eloff = Float_Type [n];
	data[ichunk-1].RA    = Float_Type [n];
	data[ichunk-1].dec   = Float_Type [n];
	data[ichunk-1].glon  = Float_Type [n];
	data[ichunk-1].glat  = Float_Type [n];

	data[ichunk-1].f0    = Float_Type [n];
	data[ichunk-1].df    = Float_Type [n];
	data[ichunk-1].mode  = Char_Type  [n];
	data[ichunk-1].nbins = UChar_Type [n];
	data[ichunk-1].spec  = Array_Type [n];
	data[ichunk-1].avflux= Float_Type [n];

	data[ichunk-1].vLSR  = Float_Type [n];
	if(verbose)  vmessage("chunk %2d: %4d lines after '%s'", ichunk-1, n, data[ichunk-1].description);
      }
      n = 0;
    }

  ichunk = nchunks>0 ? -1 : 0; % The previous version 'ichunk = -(nchunk>0)' caused an un-intended typecast to Char_Type
  i = 0;
  variable line, Y,D,H,M,S, az, el, azoff, eloff, f0, df, mode, nbins;
  foreach line (lines)
    ifnot(sscanf(line, "%d:%d:%d:%d:%d %f %f %f %f %f %f %d %d", &Y,&D,&H,&M,&S, &az, &el, &azoff, &eloff, &f0, &df, &mode, &nbins))
    {
      ichunk++;
      i = 0;
    }
    else
    {
      data[ichunk].Y     [i] = Y;
      data[ichunk].D     [i] = D;
      data[ichunk].H     [i] = H;
      data[ichunk].M     [i] = M;
      data[ichunk].S     [i] = S;
      variable MJD = 40587 + UTC2UNIXtime(Y, 1, 0, H, M, S)/86400. + D-1;  % 1970-01-01 = MJD 40587
      data[ichunk].MJD   [i] = MJD;

      data[ichunk].az    [i] = az;
      data[ichunk].el    [i] = el;
      data[ichunk].azoff [i] = azoff;
      data[ichunk].eloff [i] = eloff;

      variable RA, dec; (RA, dec) = RAdec_from_AzEl(az, el, MJD, longw, lat);
      data[ichunk].RA    [i] = RA;
      data[ichunk].dec   [i] = dec;
     (data[ichunk].glon  [i],
      data[ichunk].glat  [i]) = galLB_from_RAdec(RA, dec);

      data[ichunk].f0    [i] = f0;
      data[ichunk].df    [i] = df;
      data[ichunk].mode  [i] = mode;
      data[ichunk].nbins [i] = nbins;
      variable spec = array_map(Float_Type, &atof, strtok(line, " \x0D\n")[[9:]]);
      if(length(spec)<nbins)
	vmessage("spectrum has only %d bins, but expecting %d. Ignoring line:\n'%s'", length(spec), nbins, line);
      else
      {
	variable s = spec[[0:nbins-1]];
	data[ichunk].spec  [i] = s;
	variable bins_to_cut = qualifier("bins_to_cut", 8);
	data[ichunk].avflux[i] = mean(s[[bins_to_cut:nbins-1-bins_to_cut]]);

	data[ichunk].vLSR[i] = length(spec)>nbins ? spec[-1]
				                  : _NaN;
	i++;
      }
    }

  return onechunk || nchunks==0 ? data[0] : data;
}

%
% Code from src/misc/SRT/SRT_show_chunks.sl
%
define SRT_show_chunks()
%!%+
%\function{SRT_show_chunks}
%\synopsis{prints information for each chunk in an array of SRT data structures}
%\usage{SRT_show_chunks(Struct_Type data[]);}
%\seealso{SRT_read}
%!%-
{
  variable chunks = ();
  variable i;
  _for i (0, length(chunks)-1, 1)
    vmessage("chunk %2d: %4d lines after '%s'", i, length(chunks[i].Y), chunks[i].description);
}

%
% Code from src/misc/SRT/SRT_spectrum.sl
%
define SRT_spectrum()
%!%+
%\function{SRT_spectrum}
%\synopsis{extracts a spectrum from a SRT data structure}
%\usage{Struct_Type spec = SRT_spectrum(Struct_Type data);}
%\qualifiers{
%\qualifier{verbose}{}
%\qualifier{bins_to_cut}{[\code{=8}] number of bins at both sides of the spectrum that are discarded}
%\qualifier{bins_cut_low}{[\code{=0}] number of bins to additionaly cut from lower side}
%\qualifier{bins_cut_high}{[\code{=0}] number of bins to additionaly cut from upper side}
%\qualifier{normalize}{\code{spec.value} is normalized between 0 and 1.}
%\qualifier{vLSR}{transform the frequency grid to a velocity grid}
%\qualifier{filename=name}{plots the spectrum as name.ps in the
%     current directory}
%}
%\description
%\seealso{SRT_read}
%!%-
{
  variable data;
  switch(_NARGS)
  { case 1: data = (); }
  { help(_function_name()); return; }

  variable verbose = qualifier_exists("verbose");

  variable s = struct { bin_lo, bin_hi, value };
  variable f0 = data.f0[0];
  variable df = data.df[0];
  variable nbins = data.nbins[0];
  if(any(data.f0!=f0) || any(data.df!=df) || any(data.nbins!=nbins))
  {
    vmessage("error (%s): f0, df or nbins not constant", _function_name());
    return s;
  }

  s.bin_lo = f0 + [0 : nbins-1] * df;
  s.bin_hi = make_hi_grid(s.bin_lo);
  s.value = Float_Type[nbins];

  variable bins_cut_low=qualifier("bins_cut_low", 0), bins_cut_high=qualifier("bins_cut_high", 0);
  variable i, n, bins_to_cut=qualifier("bins_to_cut", 8);
  _for n (0, length(data.spec)-1, 1)
    _for i (bins_to_cut+bins_cut_low, nbins-1-bins_to_cut-bins_cut_high, 1)
      s.value[i] += data.spec[n][i];
  struct_filter(s, [bins_to_cut+bins_cut_low:nbins-1-bins_to_cut-bins_cut_high]);
  if(qualifier_exists("normalize"))
  {
    variable mn = min(s.value);
    variable mx = max(s.value);
    s.value = (s.value-mn)/(mx-mn);
  }
  else
    s.value /= length(data.spec);

  if(qualifier_exists("vLSR"))
  {
     
     if(string_match(getenv("USER"),"prakti",1) == 1)
     {
	message("\e[101m\e[34m\e[5m\e[25m Das sollt ihr schon selbst machen!!!\e[5m\e[m");
     } else
     {
	
	variable m = moment(data.vLSR);
	variable vcorr = m.ave;
	if(isnan(vcorr))
	{
	   vmessage("warning (%s): correction vLSR unknown", _function_name());
	   vcorr = 0;
	}
	else
	  if(verbose)
	    vmessage("<vcorr> = %f;  sdev(vcorr) = %f", vcorr, m.sdev);
	f0 = 1420.40575177;
	variable c = 299792.458;
	(s.bin_lo, s.bin_hi) = ( reverse((f0-s.bin_hi)/f0 * c - vcorr),
				 reverse((f0-s.bin_lo)/f0 * c - vcorr) );
     }
     
     xlabel("v [km/s]");
  } else
   {
      xlabel("Frequency [MHz]");
   }
   ylabel("T\\dAnt\\u [K]");
   
   if (qualifier_exists("filename"))
   {
      () = open_plot(qualifier("filename")+".ps/cps");
      hplot(s);
      close_plot; 
   }
   
   return s;
}

%
% Code from src/slang/physics/SaturationVaporPressure.sl
%
define SaturationVaporPressure() {
%!%+
%\function{SaturationVaporPressure}
%\synopsis{calculate the saturation vapor pressure of water in air}
%\usage{Double_Type[] SaturationWaterPressure(T;qualifiers)}
%\qualifiers{
%    \qualifier{centigrade}{temperature argument is in centigrade (default: K)}
%    \qualifier{kPa}{Return saturation vapor pressure in kilopascals}
%    \qualifier{hPa}{Return saturation vapor pressure in hectopascals (=mbar)}
%    \qualifier{water}{Calculate saturation vapor pressure over water}
%    \qualifier{ice}{Calculate saturation vapor pressure over ice}
%    \qualifier{iapws}{Use the more precise IAPWS prescription (see below)}
%}
%\description
% This function calculates the saturation vapor pressure (svp) of water over
% ice and water for a given temperature (default: K, but see the
% centigrade qualifier), the routine returns the svp in Pa unless
% the kPa or hPa keywords are given. By default, for T>0C (273.15K)
% the svp over water is returned, for smaller temperatures that for
% ice. Use the "water" and "ice" qualifiers if this is not what you want.
%
% As a default, for the svp over water the equation of
% Davis (1992), Metrologia 29, 67-70 is used, while for ice the
% prescription given by Marti and Mauersberger (1993), Geophys. Res.
% Lett. 20, 363-366 (1993) is applied. For the default settings,
% the relative deviation between these equations and the IAPWS recommended
% procedure is very small. It does not exceed 0.03% in the 0-100C range,
% and is <1.5% in the -50-0C range and less than 2.5% between -100 and -50C,
% while the absolute difference in the range below 50C does not exceed
% 1 Pa.
%
% If the iapws qualifier is given, the routine uses the recommendations
% of the International Association for the Properties of Water and Steam
% (IAPWS), which was originally given by Peter H. Huang, "New equations
% for water vapor pressure in the temperature range -100 deg C to 100 deg C
% for use with the 1997 NIST/ASME steam tables," in: Papers and abstracts
% from the third international symposium on humidity and moisture, Vol. 1,
% p. 69-76, National Physical Laboratory, Teddington, Middlesex, UK,
% April 1998. The algorithm used here is as described by 
% http://emtoolbox.nist.gov/Wavelength/Documentation.asp
% The uncertainty is 20kPa at 100 C, less than 2 kPa at 45 C and
% 0.7 kPa at 20C. 
%
% Note that formally the calculation is only valid in the range from
% -100 to +100 centigrade. The function returns "NaN" for T>100C and
% extrapolates formalism smoothly to lower temperatures (the values
% are very small in this regime anyway).
%
%
% This function is array safe. 
%
%!%-

variable T; % temperature in K

switch(_NARGS)
{ case 1: T=(); }
{ help(_function_name()); return; }

if (qualifier_exists("centigrade")) {
    T+=273.15;
}

if (qualifier_exists("water") && qualifier_exists("ice")) {
    throw RunTimeError,printf("%s: only one of the water and ice qualifiers is allowed!",_function_name());
}

% force array type
variable tarray=0;
if (Array_Type!=typeof(T)) {
    tarray=1;
    T=[T];
}

%
% calculate saturation water pressure over water or over ice
% default if no qualifier is set: over water for T>0C, over ice below that

variable waterndx=Int_Type[0];
variable icendx=Int_Type[0];

if (qualifier_exists("ice")) {
    icendx=[0:length(T)-1:1];
} else {
    if (qualifier_exists("water")) {
	waterndx=[0:length(T)-1:1];
    } else {
	waterndx=where(T>=273.15);
	icendx=where(T<273.15);
    }
}

variable Psv=Double_Type[length(T)];

if (qualifier_exists("iapws")) {
    % IAPWS  formula (see Stone and Zimmerman, NIST www pages and
    % Peter H. Huang, "New equations for water vapor pressure in the
    % temperature range -100 C to 100 C for use with the 1997 NIST/ASME
    % steam tables," in Papers and abstracts from the third international
    % symposium on humidity and moisture, Vol. 1, p. 69-76, National Physical
    % Laboratory, Teddington, Middlesex, UK, April 1998.)1
    %
    if (length(waterndx)>0) {
	variable K=[0.,1.16705214528E+03,-7.24213167032E+05,
 	           -1.70738469401E+01, 1.20208247025E+04,-3.23255503223E+06,
	            1.49151086135E+01,-4.82326573616E+03, 4.05113405421E+05,
	           -2.38555575678E-01, 6.50175348448E+02];

	variable omega=T[waterndx]+K[9]/(T[waterndx]-K[10]);
	variable A=(     omega+K[1])*omega+K[2];
	variable B=(K[3]*omega+K[4])*omega+K[5];
	variable C=(K[6]*omega+K[7])*omega+K[8];
	variable X=-B+sqrt(B*B-4.*A*C);
	Psv[waterndx]=1e6*(2.*C/X)^4.;
    }

    if (length(icendx)>0) {
	%
	% saturation over ice
	%
	variable A1=-13.927169;
	variable A2= 34.7078238;
	variable Theta=T[icendx]/273.16;
	variable Y=A1*(1.-Theta^-1.5)+A2*(1.-Theta^-1.25);
	Psv[icendx]=611.657*exp(Y);
    }
} else {
    if (length(waterndx)>0) {
	% equation of Davis, quoted by Ciddor (
	variable AA=1.2378847e-5;
	variable BB=-1.9121316e-2;
	variable CC=33.93711047;
	variable DD=-6.3431645e3;
	Psv[waterndx]=exp(((AA*T[waterndx]+BB)*T[waterndx])+CC+DD/T[waterndx]);
    }
    
    if (length(icendx)>0) {
	% equation of Marti and Mauersberger, quoted by
	% Ciddor (1996; appendix C);
	Psv[icendx]=10.^(-2663.5/T[icendx] + 12.537);
    }
} 

variable ndx=where(T>373.15);
if (length(ndx)>0) {
    Psv[ndx]=_NaN;
}


if (qualifier_exists("kPa")) {
    Psv=Psv/1000.;
}
if (qualifier_exists("hPa")) {
    Psv=Psv/100.;
}

if (tarray==1) {
    return Psv[0];
}
return Psv;
}

%
% Code from src/slang/strings/TeX_value_pm_error.sl
%
define TeX_value_pm_error()
%!%+
%\function{TeX_value_pm_error}
%\synopsis{formats a \code{(value, min, max)} triple as TeX string}
%\usage{String_Type TeX_value_pm_error(Double_Type value, Double_Type min, Double_Type max)}
%\altusage{String_Type TeX_value_pm_error(Struct_Type output_of__round_conf)}
%\qualifiers{
%\qualifier{factor}{[=\code{1}]: factor to multiply \code{value, min, max}, will not be shown,
%                   works only if the function is called with three arguments}
%\qualifier{neg}{factorize overall minus sign}
%\qualifier{sci}{[=\code{2}]: exponent to switch to scienficic notation}
%\qualifier{scismall}{[=\code{-sci}]: exponent for small numbers}
%\qualifier{scilarge}{[=\code{+sci}]: exponent for large numbers}
%}
%\description
%    \code{TeX_value_pm_error} tries to produce a reasonable TeX output
%    from a \code{(value, min, max)} tripel or the output of the function
%    \code{round_conf}.
%    Rounding and obtaining the number of digits is done with the
%    function \code{round_conf}.
%\seealso{round_conf,round_err}
%!%-
{
  variable s,val, mn, mx;
  switch(_NARGS)
  { case 3: (val, mn, mx) = ();
    variable factor = qualifier("factor", 1);
    val *= factor;
    mn  *= factor;
    mx  *= factor;
    s = round_conf(mn,val,mx;;__qualifiers);}
  { case 1: s = (); }
  { help(_function_name()); return; }

  val    = s.value;
  mn     = val-s.err_lo;
  mx     = val+s.err_hi;
  variable dig = s.digit;
  
  variable neg = qualifier_exists("neg") && (val<0);
  if(neg)
    (val, mn, mx) = (-val, -mx, -mn);

  variable max_abs = maxabs([mn, mx, val]);
  if(max_abs==0)  % mn == val == mx == 0
    return `$\equiv0$`;

  variable sci      = qualifier("sci",         2);
  variable scismall = qualifier("scismall", -sci);
  variable scilarge = qualifier("scilarge",  sci);

  variable power_str = "";
  variable power = 0;
  variable pow10 = 0;
  variable val4power = (val != 0 ? abs(val) : max_abs); % avoid problems with upper limits
  ifnot(10^scismall <= val4power < 10^scilarge)
  { power = int(log10(val4power)); % leading value >= 1
    if(power<0)
      power--;
    pow10 = 10^(-power);
    val *= pow10;
    mn  *= pow10;
    mx  *= pow10;
    power_str = sprintf(`\times10^{%d}`, power);
  }

  variable n = nint(power - dig);
  if (n<0)
  {
    pow10 = 10^(n);
    val *= pow10;
    mn  *= pow10;
    mx  *= pow10;
    power_str = sprintf(`\times10^{%d}`, power-n);
    n = 0;
  }
  
  variable fmt = "%.${n}f"$;
  variable m_err = sprintf(fmt, val - mn);
  variable p_err = sprintf(fmt, mx - val);
  variable err = (p_err==m_err ? `\pm`+p_err : "^{+$p_err}_{-$m_err}"$);
  val = sprintf(fmt, val);
  variable null = sprintf(fmt, 0);

  if(mn>=0 && val==null && m_err==null) % only upper limit found
    return sprintf(`$\le`+fmt+`%s$`, mx, power_str);
  else
  {
    variable paren = neg || power_str!="";
    return "$" + (neg ? "-" : "") + (paren ? `\left(` : "") + val + err + (paren ? `\right)` : "" ) + power_str + "$";
  }
}

%
% Code from src/slang/date/UNIXtime2MJD.sl
%
define UNIXtime2MJD()
%!%+
%\function{UNIXtime2MJD2}
%\synopsis{converts UNIX time to modified Julian date}
%\usage{Double_Type UNIXtime2MJD2(Long_Type secs)}
%\seealso{MJD2UNIXtime}
%!%-
{
  variable secs;
  switch(_NARGS)
  { case 1: secs = (); }
  { help(_function_name()); return; }

  return 40587 + secs/86400.;
}

%
% Code from src/slang/date/UTC2UNIXtime.sl
%
%%%%%%%%%%%%%%%%%%%
define UTC2UNIXtime()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{UTC2UNIXtime}
%\synopsis{converts Universal Time (GMT) to seconds since 1970-01-01, 0:00 UTC}
%\usage{Long_Type secs = UTC2UNIXtime(Struct_Type tm);
%\altusage{Long_Type secs = UTC2UNIXtime(Integer_Type Y[, m[, d[, H[, M[, S]]]]]);}
%}
%\description
%    \code{UTC2UNIXtime} can be seen as the inverse of the \code{gmtime} function,
%    operating on Coordinated Universal Time instead of dates in a local time zone,
%    as \code{mktime} does, which is the inverse of \code{localtime}.
%
%    \code{UTC2UNIXtime} can also take the components of the date \code{Y}-\code{m}-\code{d}, \code{H}:\code{M}:\code{S} UTC
%    as separate arguments. The year \code{Y} is then given directly (unlike \code{tm.tm_year = Y - 1900}),
%    and likewise the month \code{m} (unlike \code{tm.tm_mon = m-1}).
%    For omitted arguments, \code{m=1}, \code{d=1}, \code{H=0}, \code{M=0}, and \code{S=0} are assumed.
%\seealso{mktime, gmtime, localtime}
%!%-
{
  variable tm=NULL, Y, m=1, d=1, H=0, M=0, S=0;
  switch(_NARGS)
  { case 1:  Y = ();  if(typeof(Y)==Struct_Type)  tm = Y; }
  { case 2: (Y, m            ) = (); }
  { case 3: (Y, m, d         ) = (); }
  { case 4: (Y, m, d, H      ) = (); }
  { case 5: (Y, m, d, H, M   ) = (); }
  { case 6: (Y, m, d, H, M, S) = (); }
  { help(_function_name()); return; }

  if(tm==NULL) tm = struct { tm_sec=S, tm_min=M, tm_hour=H, tm_mday=d, tm_mon=m-1, tm_year=Y-1900, tm_wday=0, tm_yday=0, tm_isdst=0 };
  variable tm_test = gmtime(mktime(struct { tm_sec=0, tm_min=0, tm_hour=12, tm_mday=2, tm_mon=0, tm_year=100, tm_wday=0, tm_yday=0, tm_isdst=0 }));
  variable deltaH = 12-tm_test.tm_hour + 24*(2-tm_test.tm_mday);
  return mktime(tm) + 3600*deltaH;
}

%
% Code from src/fitting/fit-functions/Voigt.sl
%
%%%%%%%%%%%%%%%%
define Voigt_fit(lo, hi, par)
%%%%%%%%%%%%%%%%
%!%+
%\function{Voigt (fit-function)}
%\synopsis{implements the bare convolution of a Gaussian and a Lorentzian}
%\description
%    The \code{Voigt} fit-function uses the bare parameters of the
%    convolved Gaussian and Lorentzian profiles G and L:
%
%    \code{G(x; sigma) = 1/[sqrt(2 pi)*sigma] * exp(-(x/sigma)^2/2)}\n
%    \code{L(x; gamma) = gamma/pi * 1 / [x^2 + gamma^2]}\n
%    \code{V(x; sigma, gamma) = int dt G(t; sigma) * L(x-t; gamma)}
%
%    Unlike ISIS' internal \code{voigt} fit-function,
%    which implements a Voigt profile in astrophysical context
%    associating enerigies with \code{x} in the above formula,
%    \code{Voigt} uses \code{x} in wavelength units or any unit
%    associated with the "spectrum" through \code{define_counts}.
%\seealso{voigt, voigt_profile}
%!%-
{
  variable norm   = par[0];
  variable center = par[1];
  variable sigma  = par[2];
  variable gamma  = par[3];

  variable IVIN = Isis_Voigt_Is_Normalized;
  Isis_Voigt_Is_Normalized = 1;
%   variable dx = 1-_min(min(lo), center);
%   center += dx;
%   variable y = reverse(eval_fun2("voigt", _A(lo+dx, hi+dx), [N, center, 4*PI*gamma, sqrt(2)*Const_c/1e5*sigma/center]));
   variable y = reverse(eval_fun2("voigt", _A(lo, hi), [norm, center, 4*PI*gamma, sqrt(2)*Const_c/1e5*sigma/center]));
  Isis_Voigt_Is_Normalized = IVIN;
  return y;
}
add_slang_function("Voigt", ["norm", "center", "sigma", "gamma"], [0]);

        %%%%%%%%%%%%%%%%%%%%%
private define Voigt_defaults(i)
        %%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
  { case 0: struct {value=1, freeze=0, min=0,    max=10, hard_min=-_Inf, hard_max=_Inf, step=0, relstep=1e-4}; } % norm
  { case 1: struct {value=1, freeze=0, min=0,    max=10, hard_min=-_Inf, hard_max=_Inf, step=0, relstep=1e-4}; } % center
  { case 2: struct {value=1, freeze=0, min=1e-3, max=10, hard_min=1e-38, hard_max=_Inf, step=0, relstep=1e-4}; } % sigma
  { case 3: struct {value=1, freeze=0, min=0,    max=10, hard_min=0,     hard_max=_Inf, step=0, relstep=1e-4}; } % gamma
}

set_param_default_hook("Voigt", &Voigt_defaults);

%
% Code from src/fitting/fit-functions/Weibull.sl
%
%%%%%%%%%%%%%%%%%%
define Weibull_fit(bin_lo, bin_hi, pars)
%%%%%%%%%%%%%%%%%%
%!%+
%\function{Weibull (fit-function)}
%\description
%    \code{W(x) = norm * k/lambda * ((x-x0)/lam)^(k-1) * exp(-((x-x0)/lam)^k);}\n
%    \code{W( x0 + lambda * (1-1/k)^(1/k) )  =  max.}
%!%-
{
  variable x = (bin_lo + bin_hi) / 2. ;
  variable k      = pars[0];
  variable lambda = pars[1];
  variable x0     = pars[2];
  variable norm   = pars[3];
  variable y = norm * k/lambda * ((x-x0)/lambda)^(k-1) * exp(-((x-x0)/lambda)^k);
  y[where(x<=x0)] = 0;
  return y;
}

%%%%%%%%%%%%%%%%%%%%%%
define Weibull_default(i)
%%%%%%%%%%%%%%%%%%%%%%
{ switch(i)
  { case 0: return (4.9, 0, 0, 10 ); }
  { case 1: return (3.0, 0, 0, 20 ); }
  { case 2: return (1  , 0, 0, 0  ); }
  { case 3: return (1e3, 0, 0, 1e6); }
}

add_slang_function("Weibull", ["k", "lambda", "x0", "norm"]);
set_param_default_hook("Weibull", "Weibull_default");

%
% Code from src/slang/misc/WorldCoastLine.sl
%
define WorldCoastLine()
%!%+
%\function{WorldCoastLine}
%\usage{Struct_Type WCL[] = WorldCoastLine();}
%\description
%    \code{WCL} is an array of structures with \code{x} and \code{y} fields
%    which specify the coordinates (longitude east and lattitude north)
%    in degrees of the World's coast lines.
%\example
%    xrange(-2, 2); yrange(-1, 1);
%    variable cl;
%    foreach cl (WorldCoastLine())
%      oplot( Hammer_projection(cl.x, cl.y; deg, normalized) );
%!%-
{
  variable WCL = Struct_Type[1865];
  WCL[0] = struct {
   x = [0.19244, -0.114995, -0.391921, -0.657114, -0.887104, -1.2626, -1.41749, -1.56065, -1.73666, -1.93614, -2.19898, -2.4102, -2.5956, -2.77161],
   y = [5.66995, 5.51506, 5.36017, 5.19589, 5.1513, 5.06212, 5.00815, 4.87438, 4.77581, 4.68663, 4.78754, 4.87438, 4.9307, 4.95182] };
  WCL[1] = struct {
   x = [-2.77161, -2.86079, -3.00395, -3.302, -3.61178, -3.83238, -4.23838, -4.67959, -5.19824, -5.3414, -5.43058, -5.58547, -5.84831, -6.08065, -6.22381, -6.44441, -6.64389, -6.78705, -6.87388, -7.10622,
        -7.29396, -7.4911],
   y = [4.95182, 4.96356, 4.97529, 5.05274, 5.10671, 5.18416, 5.16304, 5.10671, 5.041, 5.00815, 5.01753, 4.98468, 4.91897, 4.79693, 4.75469, 4.67724, 4.5998, 4.54347, 4.4895, 4.38858, 4.27828, 4.25716] };
  WCL[2] = struct {
   x = [-7.4911, -7.52395, -7.73517, -8.16464, -8.48381, -8.83583, -9.11276, -9.46479, -9.80742],
   y = [4.25716, 4.24543, 4.3346, 4.4895, 4.58806, 4.80867, 5.00815, 5.33905, 5.68169] };
  WCL[3] = struct {
   x = [-9.80742, -10.0046, -10.204, -10.4129, -10.5232, -10.6452, -10.8001, -10.887, -11.009, -11.0958, -11.1522, -11.185, -11.1967, -11.1522, -11.131, -11.1639, -11.3071, -11.3516],
   y = [5.68169, 5.84597, 5.98912, 6.0783, 6.13228, 6.25432, 6.36462, 6.38574, 6.40921, 6.49604, 6.50777, 6.56175, 6.59461, 6.60634, 6.6392, 6.67205, 6.76123, 6.80347] };
  WCL[4] = struct {
   x = [-11.3516, -11.4619, -11.7905, -12.1214, -12.2998, -12.342, -12.3655, -12.5626, -12.7058, -12.7738, -12.7949, -12.9052, -13.0367, -13.0695, -13.0484, -12.971, -12.9827, -13.0813, -13.1141, -13.093,
        -13.1141, -13.1704, -13.1798],
   y = [6.80347, 6.89266, 7.04755, 7.17897, 7.25641, 7.32213, 7.45355, 7.6413, 7.72813, 7.80557, 8.03556, 8.1224, 8.1224, 8.26555, 8.36412, 8.47442, 8.49554, 8.55187, 8.69268, 8.80298, 8.85696, 8.92267,
        9.01185] };
  WCL[5] = struct {
   x = [-13.1798, -13.1916, -13.2479, -13.3253, -13.391, -13.5225, -13.5553, -13.6539, -13.7548, -13.9191, -14.0294],
   y = [9.01185, 9.02123, 9.17613, 9.26296, 9.36153, 9.48121, 9.61264, 9.67835, 9.79804, 9.93885, 9.9717] };
  WCL[6] = struct {
   x = [-12.4312, -12.4523, -12.5415, -12.6518, -12.7621, -12.7058, -12.5861, -12.4852, -12.4312],
   y = [7.49814, 7.4864, 7.53099, 7.4864, 7.4864, 7.40896, 7.36672, 7.3761, 7.49814] };
  WCL[7] = struct {
   x = [-49.9172, -49.774, -49.7623, -49.8632],
   y = [-0.018775, 0.004694, 0.136116, 0.225296] };
  WCL[8] = struct {
   x = [-50.0275, -49.95, -49.9735, -49.9618, -49.8843, -49.8632],
   y = [0.957509, 1.05608, 1.21331, 1.30015, 1.47851, 1.73196] };
  WCL[9] = struct {
   x = [-59.7082, -59.7739, -59.9053],
   y = [8.27729, 8.31014, 8.46269] };
  WCL[10] = struct {
   x = [-57.0609, -57.1173, -57.1619, -57.2722, -57.3919, -57.5467, -57.657, -57.7673, -57.9879, -58.2086, -58.3963, -58.5066, -58.5183, -58.4831, -58.462, -58.462, -58.4949, -58.5723, -58.6498, -58.7694,
        -58.9478, -59.1684, -59.3562, -59.5204, -59.6425, -59.6425, -59.6425, -59.6964, -59.7082],
   y = [5.79199, 5.90229, 6.06657, 6.26605, 6.30829, 6.44206, 6.58522, 6.70491, 6.83633, 6.87153, 6.83633, 6.80347, 6.80347, 6.90439, 7.01469, 7.1555, 7.35498, 7.50753, 7.60609, 7.71639, 7.84782, 8.02383,
        8.22096, 8.31953, 8.37585, 8.31953, 8.26555, 8.29841, 8.27729] };
  WCL[11] = struct {
   x = [-54.0288, -54.0288, -53.9749, -53.9866, -54.3832, -54.7798, -55.0004, -55.045, -55.1107, -55.3079, -55.6176, -55.8054, -55.904, -55.9251, -56.2795, -56.5001, -56.7746, -56.8638, -56.9295, -56.9741,
        -57.0609],
   y = [5.48221, 5.54792, 5.73567, 5.79199, 5.93515, 5.968, 5.968, 5.95627, 5.86944, 5.89056, 5.91168, 5.8577, 5.80138, 5.82484, 5.83423, 5.90229, 5.95627, 6.01259, 5.95627, 5.8577, 5.79199] };
  WCL[12] = struct {
   x = [-51.5928, -51.5811, -51.5717, -51.7594, -51.8252, -52.034, -52.1655, -52.2335, -52.4541, -52.7404, -53.0479, -53.3013, -53.5125, -53.5994, -53.7425, -53.7988, -53.9091, -54.0288],
   y = [4.06941, 4.13512, 4.2994, 4.53174, 4.51062, 4.73122, 4.84152, 4.9307, 5.10671, 5.2616, 5.38364, 5.48221, 5.53853, 5.58077, 5.71454, 5.72393, 5.54792, 5.48221] };
  WCL[13] = struct {
   x = [-50.1941, -50.2481, -50.3373, -50.3373, -50.2152, -50.0486],
   y = [0.457633, 0.413043, 0.314476, 0.204175, 0.058671, 0.037549] };
  WCL[14] = struct {
   x = [-49.8632, -50.0064, -50.1261, -50.1941],
   y = [0.225296, 0.248765, 0.335597, 0.457633] };
  WCL[15] = struct {
   x = [-50.931, -50.7339, -50.579, -50.4358, -50.2809, -50.1589, -50.0275],
   y = [-0.007041, 0.213562, 0.32621, 0.525691, 0.701704, 0.769762, 0.957509] };
  WCL[16] = struct {
   x = [-49.8632, -50.0838, -50.3467, -50.4452, -50.4687, -50.5555, -50.6658, -50.8207, -50.931, -50.9639, -50.9639, -51.0531, -51.1962, -51.2948, -51.3628, -51.384, -51.4051, -51.4825, -51.5928],
   y = [1.73196, 1.80002, 1.80941, 1.91032, 2.02062, 2.28582, 2.61907, 2.99456, 3.3372, 3.62586, 3.78075, 3.93564, 4.13512, 4.23369, 4.27828, 4.29002, 4.1891, 4.05768, 4.06941] };
  WCL[17] = struct {
   x = [-59.9053, -60.0273, -60.1822, -60.2808, -60.5014, -60.6328, -60.7783, -60.9426, -60.9543, -60.8206, -60.6774, -60.6446, -60.7337, -60.7431, -60.8206, -60.9215, -61.1632, -61.3721, -61.3956, -61.473,
        -61.5598, -61.649, -61.7265, -61.7358, -61.7922, -61.8039, -61.8579, -62.0222, -62.1325, -62.2662, -62.4094],
   y = [8.46269, 8.56125, 8.59411, 8.58472, 8.40871, 8.46269, 8.48616, 8.50728, 8.58472, 8.58472, 8.67156, 8.78186, 8.75839, 8.92267, 9.23949, 9.4155, 9.54692, 9.59151, 9.6009, 9.67835, 9.69947, 9.76518,
        9.79804, 9.80977, 9.85201, 9.85201, 9.84263, 9.87548, 9.87548, 9.79804, 9.8966] };
  WCL[18] = struct {
   x = [-80.0388, -79.9942, -79.8839, -79.8511, -79.8628, -79.7196, -79.4873, -79.255, -79.1564, -79.1118, -78.9358, -78.7269, -78.5275, -78.2975, -78.098, -77.9548, -77.8328, -77.7131, -77.5699],
   y = [9.06582, 9.08695, 9.11041, 9.23949, 9.46948, 9.59151, 9.61264, 9.61264, 9.6009, 9.54692, 9.50234, 9.46948, 9.4155, 9.36153, 9.25122, 9.13153, 9.02123, 8.9344, 8.78186] };
  WCL[19] = struct {
   x = [-77.9008, -78.0651, -78.2294, -78.3632, -78.3397, -78.2975, -78.1965, -78.0651, -78.1872, -78.4406, -78.5838, -78.8372, -79.079, -79.3441, -79.6422, -79.7408, -79.9074],
   y = [7.24468, 7.40896, 7.59671, 7.90414, 8.06842, 8.29841, 8.37585, 8.37585, 8.48616, 8.42983, 8.62697, 8.79124, 8.92267, 8.96726, 8.85696, 8.68329, 8.42983] };
  WCL[20] = struct {
   x = [-77.9008, -77.9008, -77.8445, -77.746, -77.7225, -77.6357, -77.4808, -77.415, -77.415, -77.4244, -77.4808, -77.4244, -77.3705, -77.3705, -77.4362, -77.5254, -77.5019, -77.469, -77.4479, -77.3916,
        -77.3587, -77.3705, -77.4596, -77.4925, -77.4596, -77.415, -77.3259, -77.0841, -76.9504, -77.0607, -77.1381, -77.2273, -77.3705, -77.4808, -77.5347, -77.6357, -77.6802, -77.7671, -77.8563,
        -78.0323, -78.1543, -78.3843, -78.4617, -78.5838, -78.6378, -78.6823, -78.6612, -78.6284, -78.6284, -78.7152, -78.8255, -78.9687, -78.9921, -78.9921, -78.9241, -78.8818],
   y = [7.24468, 7.20009, 7.10152, 6.93724, 6.8598, 6.71664, 6.67205, 6.58522, 6.40921, 6.26605, 6.14401, 5.93515, 5.79199, 5.66995, 5.61363, 5.5268, 5.51506, 5.33905, 5.00815, 4.76408, 4.63265, 4.47776,
        4.29002, 4.10227, 3.93564, 3.8582, 3.8582, 3.90279, 3.95676, 3.7479, 3.68218, 3.54841, 3.40526, 3.29496, 3.16119, 3.06262, 2.90773, 2.71764, 2.66366, 2.60733, 2.61907, 2.57448, 2.56274,
        2.39847, 2.25296, 2.12154, 1.93145, 1.83288, 1.80002, 1.80002, 1.7437, 1.6334, 1.61228, 1.47851, 1.4128, 1.37759] };
  WCL[21] = struct {
   x = [-78.8701, -78.8912, -79.0015, -79.1893, -79.4333, -79.5859, -79.6422, -79.8299],
   y = [1.37759, 1.31188, 1.23443, 1.1124, 1.01383, 0.901185, 0.812005, 0.823739] };
  WCL[22] = struct {
   x = [-75.6714, -75.6831, -75.6831, -75.6831, -75.6948, -75.8262, -75.9811, -76.1478, -76.2463, -76.3449, -76.5444, -76.7204, -76.8847, -76.9081, -76.8401, -76.765, -76.765, -76.765, -76.7532, -76.7861,
        -76.8964, -76.9621, -76.9856, -77.0513, -77.1381, -77.2484, -77.3493, -77.415, -77.4479, -77.5699],
   y = [10.0163, 9.90834, 9.80977, 9.66661, 9.54692, 9.43662, 9.4155, 9.34979, 9.19725, 8.98838, 8.84757, 8.72553, 8.65043, 8.54013, 8.4533, 8.42044, 8.38524, 8.21158, 8.05668, 7.98159, 7.937, 7.98159,
        8.14587, 8.21158, 8.343, 8.46269, 8.54013, 8.65043, 8.65982, 8.78186] };
  WCL[23] = struct {
   x = [-71.318, -71.1959, -71.1091, -71.0645, -71.0645, -71.0856, -71.2616, -71.494, -71.616, -71.67, -71.8038, -71.9352, -71.9798, -72.0126],
   y = [10.1149, 9.93885, 9.7112, 9.48121, 9.32867, 9.21837, 9.06582, 9.02123, 9.0447, 9.25122, 9.49295, 9.6009, 9.72059, 9.87548] };
  WCL[24] = struct {
   x = [-81.8036, -81.759, -81.7684, -81.8787, -81.8905, -81.7473, -81.7262, -81.7145, -81.8036],
   y = [7.70701, 7.69527, 7.70701, 7.58497, 7.43243, 7.29866, 7.4864, 7.68354, 7.70701] };
  WCL[25] = struct {
   x = [-83.2352, -83.0381, -82.8175, -82.7189],
   y = [10.1149, 9.8966, 9.73232, 9.67835] };
  WCL[26] = struct {
   x = [-82.7189, -82.6415, -82.4748, -82.3763, -82.2542, -82.2214, -82.1557, -82.1557, -82.1674, -81.9351, -81.7356, -81.5267, -81.285, -81.0409, -80.8978, -80.7335, -80.534, -80.3251, -80.0928, -80.0388,
        -80.0388],
   y = [9.67835, 9.65723, 9.53519, 9.32867, 9.26296, 9.31694, 9.32867, 9.17613, 9.01185, 9.00011, 9.01185, 8.86869, 8.83583, 8.86869, 8.95552, 9.0447, 9.08695, 9.17613, 9.32867, 9.25122, 9.06582] };
  WCL[27] = struct {
   x = [-79.9074, -80.1608, -80.3369, -80.4354, -80.4683, -80.3908, -80.2594, -80.1374, -80.0834, -80.2031, -80.358, -80.5575, -80.7663, -80.9212, -80.9306, -80.9541, -80.9541, -81.0198, -81.0644, -81.1747,
        -81.2075, -81.2287, -81.3624, -81.4821, -81.5924, -81.6699, -81.6816, -81.6933, -81.7684, -81.9562, -82.1885, -82.3645, -82.4631, -82.618, -82.7518, -82.8832, -82.916, -82.9278, -82.9724,
        -83.0263],
   y = [8.42983, 8.29841, 8.25382, 8.16699, 8.00271, 7.87129, 7.72813, 7.62956, 7.4864, 7.43243, 7.31039, 7.22356, 7.21182, 7.22356, 7.31039, 7.42069, 7.51926, 7.62956, 7.77272, 7.79384, 7.68354, 7.59671,
        7.6413, 7.77272, 7.87129, 8.01444, 8.01444, 8.03556, 8.14587, 8.18811, 8.26555, 8.27729, 8.33127, 8.31953, 8.343, 8.31953, 8.2327, 8.13413, 8.11301, 8.31014] };
  WCL[28] = struct {
   x = [-83.0263, -83.0827, -83.1249, -83.1578, -83.3455, -83.4793, -83.3901, -83.369, -83.545, -83.721, -83.6999, -83.721, -83.9862, -84.2396, -84.3945, -84.5588, -84.648, -84.702, -84.7912],
   y = [8.31014, 8.35473, 8.40871, 8.60584, 8.70441, 8.71615, 8.56125, 8.47442, 8.49554, 8.6387, 8.8241, 9.11041, 9.29581, 9.43662, 9.5258, 9.57978, 9.67835, 9.87548, 9.93885] };
  WCL[29] = struct {
   x = [-85.176, -85.0446, -84.9672, -85.0211, -85.0986, -85.2324, -85.3966, -85.5515, -85.6947],
   y = [10.0163, 9.98344, 9.86375, 9.75345, 9.64549, 9.76518, 9.86375, 9.87548, 9.96232] };
  WCL[30] = struct {
   x = [-79.8299, -80.0717, -80.1045, -80.0717, -80.0599],
   y = [0.823739, 0.746294, 0.570281, 0.359066, 0.171319] };
  WCL[31] = struct {
   x = [-91.3834, -91.3717],
   y = [0.147851, 0.103261] };
  WCL[32] = struct {
   x = [-91.5266, -91.4937, -91.4069, -91.3834],
   y = [-0.05163, 0.037549, 0.114995, 0.147851] };
  WCL[33] = struct {
   x = [-157.344, -157.308, -157.287, -157.254, -157.198, -157.189, -157.144, -157.046, -157.01, -156.989, -157.111, -157.266, -157.365, -157.421, -157.43, -157.421, -157.332, -157.287, -157.276, -157.299,
        -157.344, -157.344, -157.344],
   y = [2.01358, 2.03471, 1.97838, 1.969, 1.91267, 1.8141, 1.76951, 1.71319, 1.65687, 1.59115, 1.60289, 1.62401, 1.68033, 1.70145, 1.76951, 1.7789, 1.76951, 1.8141, 1.86808, 1.9244, 1.969, 1.99012, 2.01358] };
  WCL[34] = struct {
   x = [-162.079, -162.068, -162.056, -162.079, -162.079],
   y = [5.93749, 5.95861, 5.94923, 5.8929, 5.93749] };
  WCL[35] = struct {
   x = [-176.564, -176.552, -176.552, -176.564, -176.552, -176.552, -176.564],
   y = [0.227643, 0.173666, 0.183053, 0.227643, 0.173666, 0.183053, 0.227643] };
  WCL[36] = struct {
   x = [170.885, 170.951, 170.951, 170.962, 171.007, 171.073, 171.117, 171.171, 171.171, 171.15, 171.105, 171.061, 170.995, 170.939, 170.894, 170.861, 170.84, 170.84, 170.828, 170.852, 170.885],
   y = [8.86165, 8.84991, 8.82879, 8.80767, 8.79594, 8.76308, 8.71849, 8.64339, 8.55421, 8.51197, 8.47912, 8.50024, 8.55421, 8.58707, 8.65278, 8.71849, 8.76308, 8.81706, 8.84991, 8.86165, 8.86165] };
  WCL[37] = struct {
   x = [172.861, 172.882, 172.882, 172.927, 172.959, 172.959, 172.971, 173.016, 173.049, 173.049, 172.959, 172.905, 172.905, 172.905, 172.905, 172.882, 172.861, 172.861],
   y = [1.89155, 1.87982, 1.86808, 1.83523, 1.72492, 1.60289, 1.50197, 1.41514, 1.32596, 1.28137, 1.26964, 1.35882, 1.448, 1.53483, 1.63574, 1.76951, 1.86808, 1.89155] };
  WCL[38] = struct {
   x = [163.112, 163.133, 163.133, 163.058, 163.002, 162.99, 163.014, 163.067, 163.112],
   y = [5.36486, 5.35313, 5.33201, 5.30854, 5.29681, 5.3414, 5.36486, 5.36486, 5.36486] };
  WCL[39] = struct {
   x = [166.909, 166.942, 166.954, 167.031, 167.088, 167.163, 167.24, 167.318, 167.395, 167.44, 167.505, 167.583, 167.639, 167.66, 167.672, 167.627, 167.562, 167.538, 167.505, 167.463, 167.362, 167.24,
        167.142, 167.088, 166.998, 166.9, 166.888, 166.909],
   y = [9.28877, 9.30989, 9.27704, 9.27704, 9.2653, 9.2653, 9.25592, 9.30989, 9.33336, 9.30051, 9.16909, 9.07052, 8.99307, 8.89451, 8.7842, 8.69737, 8.74196, 8.84991, 8.9391, 8.99307, 9.07991, 9.15735,
        9.15735, 9.16909, 9.19021, 9.25592, 9.30051, 9.28877] };
  WCL[40] = struct {
   x = [169.549, 169.538, 169.493, 169.493, 169.549, 169.549],
   y = [5.99382, 6.0032, 5.99382, 6.01494, 5.99382, 5.99382] };
  WCL[41] = struct {
   x = [152.127, 152.171, 152.15, 152.138, 152.127],
   y = [7.1156, 7.13907, 7.12734, 7.1156, 7.1156] };
  WCL[42] = struct {
   x = [153.661, 153.64, 153.685, 153.718, 153.706, 153.673, 153.661],
   y = [5.60659, 5.65118, 5.63945, 5.60659, 5.58547, 5.58547, 5.60659] };
  WCL[43] = struct {
   x = [158.287, 158.332, 158.388, 158.397, 158.388, 158.343, 158.278, 158.254, 158.266, 158.287],
   y = [7.0053, 6.99592, 6.93959, 6.87388, 6.84102, 6.84102, 6.87388, 6.93959, 6.98418, 7.0053] };
  WCL[44] = struct {
   x = [134.729, 134.739, 134.739, 134.739, 134.706, 134.629, 134.596, 134.607, 134.629, 134.694, 134.729],
   y = [7.66476, 7.69997, 7.62252, 7.51222, 7.39018, 7.35733, 7.45589, 7.57793, 7.61079, 7.65538, 7.66476] };
  WCL[45] = struct {
   x = [138.24, 138.24, 138.273, 138.283, 138.24, 138.217, 138.24],
   y = [9.58447, 9.57274, 9.57274, 9.52815, 9.50703, 9.53988, 9.58447] };
  WCL[46] = struct {
   x = [123.368, 123.246, 123.192, 123.181, 123.169, 123.192, 123.237, 123.225, 123.059, 122.904, 122.793, 122.674, 122.54, 122.43, 122.486, 122.563],
   y = [10.2158, 9.98813, 9.84497, 9.72528, 9.6056, 9.46244, 9.34275, 9.13623, 9.03766, 9.19021, 9.35449, 9.40846, 9.52815, 9.70416, 9.94354, 9.96466] };
  WCL[47] = struct {
   x = [125.687, 125.699, 125.72, 125.687, 125.654],
   y = [10.1618, 9.99752, 9.85671, 9.81212, 9.94354] };
  WCL[48] = struct {
   x = [126.018, 126.072, 126.107, 126.117, 126.072, 126.006, 126.018],
   y = [9.99752, 9.9013, 9.86844, 9.791, 9.81212, 9.88956, 9.99752] };
  WCL[49] = struct {
   x = [123.699, 123.634, 123.544, 123.49, 123.413, 123.368, 123.368],
   y = [10.021, 9.88956, 9.69243, 9.51876, 9.46244, 9.65957, 9.85671] };
  WCL[50] = struct {
   x = [123.833, 123.833, 123.887, 123.986],
   y = [9.84497, 9.86844, 9.87783, 9.95528] };
  WCL[51] = struct {
   x = [124.572, 124.572, 124.54, 124.549, 124.504, 124.394, 124.251, 124.108, 123.976, 123.908, 123.854, 123.833],
   y = [10.021, 9.9013, 9.84497, 9.73702, 9.67131, 9.63845, 9.59386, 9.6056, 9.61733, 9.68304, 9.77926, 9.84497] };
  WCL[52] = struct {
   x = [125.422, 125.41, 125.455, 125.499, 125.598, 125.654, 125.72, 125.851, 125.929, 126.03, 126.128, 126.206, 126.316, 126.328, 126.239, 126.161, 126.194, 126.337, 126.349, 126.426, 126.438, 126.447,
        126.516, 126.581, 126.581, 126.57, 126.483, 126.37, 126.283, 126.271, 126.271, 126.149, 126.084, 126.072, 125.973, 125.896, 125.819, 125.732, 125.675, 125.577, 125.466, 125.422, 125.544,
        125.643, 125.708, 125.699, 125.565, 125.476, 125.344, 125.288, 125.279, 125.124, 124.936, 124.769, 124.572, 124.373, 124.23, 124.131, 124.108, 124.054, 124.009, 124.087, 124.218, 124.218,
        124.075, 123.866, 123.699, 123.568, 123.512, 123.425, 123.413, 123.39, 123.291, 123.192, 123.192, 123.082, 122.948, 122.927, 122.915, 122.85, 122.805, 122.662, 122.528, 122.451, 122.385,
        122.319, 122.23, 122.12, 122.01, 121.944, 122, 122.078, 122.111, 122.153, 122.331, 122.519, 122.674, 122.904, 122.96, 123.092, 123.279, 123.39, 123.479, 123.744, 123.833, 123.744, 123.732,
        123.908, 124.131, 124.274, 124.462, 124.593, 124.692, 124.727, 124.814, 124.936, 125.025, 125.112, 125.246, 125.377, 125.466, 125.488, 125.488, 125.476, 125.455, 125.422, 125.422, 125.422],
   y = [9.71355, 9.74641, 9.74641, 9.74641, 9.73702, 9.63845, 9.56101, 9.4953, 9.39673, 9.27704, 9.2653, 9.12449, 8.89451, 8.79594, 8.69737, 8.58707, 8.55421, 8.47912, 8.25851, 8.18107, 7.99567, 7.85251,
        7.72109, 7.4911, 7.32447, 7.14846, 7.01704, 6.96306, 6.8199, 6.58757, 6.36696, 6.47727, 6.7096, 6.99592, 7.13907, 7.31508, 7.35733, 7.2705, 7.19305, 7.03816, 6.85276, 6.61104, 6.489, 6.30125,
        6.08065, 5.97035, 5.76148, 5.59485, 5.6723, 5.8929, 6.01494, 5.90464, 5.87178, 5.91637, 6.01494, 6.12524, 6.28952, 6.44441, 6.63216, 6.78705, 7.01704, 7.10622, 7.30335, 7.44651, 7.62252,
        7.7868, 7.85251, 7.84078, 7.72109, 7.5662, 7.46763, 7.46763, 7.53334, 7.53334, 7.66476, 7.5662, 7.44651, 7.39018, 7.33621, 7.50049, 7.70935, 7.70935, 7.5662, 7.47936, 7.30335, 7.16019, 6.93959,
        6.90674, 6.9302, 7.12734, 7.28223, 7.53334, 7.68823, 7.84078, 8.04026, 8.0825, 8.12709, 8.1928, 8.35708, 8.4885, 8.58707, 8.7091, 8.67625, 8.57768, 8.34535, 8.06138, 8.01679, 8.09423, 8.1928,
        8.4885, 8.58707, 8.53309, 8.65278, 8.84991, 8.96022, 8.92736, 8.88277, 8.9391, 9.00481, 8.94848, 9.02593, 9.16909, 9.27704, 9.33336, 9.46244, 9.6056, 9.67131, 9.71355] };
  WCL[53] = struct {
   x = [121.834, 121.932, 122.066, 122.254, 122.308, 122.176, 122, 121.899, 121.834],
   y = [6.63216, 6.7096, 6.7096, 6.64389, 6.58757, 6.45614, 6.43267, 6.53359, 6.63216] };
  WCL[54] = struct {
   x = [120.951, 120.951, 120.984, 121.106, 121.205, 121.348, 121.425, 121.348, 121.184, 121.05, 120.951],
   y = [5.87178, 5.93749, 6.0032, 6.03606, 6.01494, 5.97035, 5.94923, 5.81546, 5.8929, 5.86005, 5.87178] };
  WCL[55] = struct {
   x = [119.968, 120.09, 120.221, 120.254, 120.134],
   y = [5.18651, 5.25456, 5.22171, 5.14426, 5.09967] };
  WCL[56] = struct {
   x = [121.115, 121.205, 121.315, 121.446, 121.482, 121.601, 121.735, 121.899, 122.01, 122.153, 122.263, 122.418, 122.585, 122.662, 122.749, 122.862, 122.96, 123.092, 123.246, 123.458, 123.622, 123.789,
        123.908, 124.164, 124.274, 124.406, 124.549, 124.692, 124.838, 125.025, 125.136, 125.178, 125.213, 125.19, 125.103, 124.957, 124.805, 124.626, 124.45, 124.263, 124.021, 123.798, 123.556,
        123.39, 123.192, 122.904, 122.761, 122.596, 122.364, 122.153, 121.923, 121.768, 121.625, 121.547, 121.369, 121.282, 121.127, 120.907, 120.698, 120.498, 120.332, 120.188],
   y = [1.32596, 1.31423, 1.31423, 1.30249, 1.20393, 1.10301, 1.08189, 1.11475, 1.11475, 1.10301, 1.07016, 1.07016, 1.02557, 0.948121, 0.927, 0.88241, 0.894144, 0.959856, 0.980977, 0.948121, 0.927, 0.903532,
        0.894144, 0.959856, 1.05842, 1.21331, 1.34708, 1.43626, 1.62401, 1.72492, 1.71319, 1.65687, 1.51371, 1.37055, 1.20393, 1.0373, 0.88241, 0.638339, 0.483448, 0.361413, 0.361413, 0.316823,
        0.305089, 0.305089, 0.438858, 0.492835, 0.50457, 0.50457, 0.483448, 0.45998, 0.41539, 0.492835, 0.50457, 0.45998, 0.516304, 0.50457, 0.41539, 0.427124, 0.45998, 0.45998, 0.272233, 0.028162] };
  WCL[57] = struct {
   x = [119.956, 120.057, 120.134, 120.266, 120.343, 120.475, 120.609, 120.773, 120.817, 120.928, 121.05, 121.115],
   y = [0.471714, 0.682929, 0.760375, 0.903532, 0.903532, 0.79323, 0.816699, 1.00445, 1.23678, 1.30249, 1.32596, 1.32596] };
  WCL[58] = struct {
   x = [125.434, 125.455, 125.52, 125.577, 125.598, 125.643, 125.544, 125.466, 125.434],
   y = [3.71739, 3.71739, 3.66106, 3.54137, 3.47332, 3.35363, 3.54137, 3.63994, 3.71739] };
  WCL[59] = struct {
   x = [126.769, 126.835, 126.868, 126.868, 126.891, 126.879, 126.823, 126.814, 126.781, 126.769],
   y = [4.53643, 4.53643, 4.44725, 4.34868, 4.24777, 4.07176, 4.02717, 4.20318, 4.40266, 4.53643] };
  WCL[60] = struct {
   x = [128.557, 128.59, 128.644, 128.668, 128.656, 128.656, 128.611, 128.524, 128.337, 128.304, 128.28, 128.292, 128.381, 128.501],
   y = [2.65427, 2.58856, 2.52285, 2.43367, 2.2999, 2.2999, 2.20133, 2.05583, 2.05583, 2.1896, 2.31163, 2.4454, 2.5557, 2.65427] };
  WCL[61] = struct {
   x = [127.938, 127.961, 128.015, 128.027, 127.949, 127.872, 127.863, 127.938, 128.006, 128.027, 128.006, 127.949, 127.785, 127.64, 127.651, 127.795, 127.905, 128.06, 128.149, 128.238, 128.402, 128.602,
        128.656, 128.724, 128.724, 128.602, 128.447, 128.325, 128.325, 128.513, 128.635, 128.668, 128.724, 128.834, 128.668, 128.501, 128.226, 128.039, 127.949, 127.928],
   y = [2.22245, 2.23419, 2.21307, 2.11215, 1.97838, 1.91267, 1.84696, 1.7789, 1.57003, 1.41514, 1.26964, 1.19219, 1.02557, 0.927, 0.88241, 0.849554, 0.97159, 1.08189, 1.24852, 1.41514, 1.51371, 1.52544,
        1.48085, 1.32596, 1.09363, 0.992711, 0.88241, 0.804964, 0.736906, 0.614871, 0.549159, 0.41539, 0.337944, 0.260499, 0.251111, 0.349678, 0.406002, 0.471714, 0.361413, 0.061018] };
  WCL[62] = struct {
   x = [127.75, 127.717, 127.696, 127.597, 127.597, 127.553, 127.475, 127.377, 127.398, 127.419, 127.464, 127.553, 127.651, 127.827, 127.938],
   y = [-0.269886, 0.093873, 0.293354, 0.537425, 0.760375, 0.927, 0.992711, 1.1476, 1.33535, 1.34708, 1.51371, 1.7789, 1.95726, 2.13327, 2.22245] };
  WCL[63] = struct {
   x = [109.966, 110.109, 110.219, 110.329, 110.451, 110.541, 110.66, 110.815, 110.916, 111.08, 111.125, 111.137, 111.158, 111.214, 111.224, 111.224, 111.247, 111.292, 111.268, 111.247, 111.334, 111.456,
        111.479, 111.479, 111.479, 111.512, 111.578, 111.644, 111.721, 111.831, 111.921, 112.019, 112.097, 112.085, 112.085, 112.118, 112.186, 112.251, 112.374, 112.526, 112.648, 112.77, 112.869,
        112.991, 113.068, 113.146, 113.2, 113.268, 113.322, 113.378, 113.444, 113.509, 113.587, 113.641, 113.73, 113.807, 113.864, 113.906, 113.941, 113.962, 113.974, 113.983, 113.983, 114.007],
   y = [1.83523, 1.7789, 1.75778, 1.76951, 1.79063, 1.75778, 1.68033, 1.63574, 1.63574, 1.60289, 1.60289, 1.61462, 1.73666, 1.85635, 1.9244, 1.95726, 2.03471, 2.13327, 2.24592, 2.32337, 2.38908, 2.35622,
        2.32337, 2.53224, 2.58856, 2.66601, 2.76457, 2.84202, 2.89834, 2.91008, 2.91946, 2.91946, 2.95466, 2.96405, 2.96405, 2.97579, 2.98752, 2.98752, 2.98752, 3.02038, 3.02038, 3.03211, 3.08609,
        3.20812, 3.28557, 3.39587, 3.46393, 3.52964, 3.59535, 3.70565, 3.80657, 3.88401, 3.94972, 3.98258, 4.06003, 4.13747, 4.23838, 4.28063, 4.34868, 4.42613, 4.48011, 4.54582, 4.56929, 4.59041] };
  WCL[64] = struct {
   x = [114.007, 114.019, 114.04, 114.105, 114.204, 114.293, 114.382, 114.469, 114.603, 114.725, 114.812, 114.901, 114.955, 114.978, 115, 115.056, 115.089, 115.133, 115.176, 115.22, 115.244],
   y = [4.59041, 4.61388, 4.62326, 4.635, 4.67959, 4.67959, 4.69132, 4.71244, 4.78989, 4.88846, 4.9659, 5.03162, 5.0762, 5.02223, 4.94478, 4.93305, 4.93305, 4.94478, 4.94478, 4.93305, 4.93305] };
  WCL[65] = struct {
   x = [115.244, 115.277, 115.354, 115.441, 115.53, 115.563, 115.464, 115.42, 115.497, 115.584, 115.661, 115.772, 115.884, 116.004, 116.06, 116.105, 116.114, 116.215, 116.293, 116.368, 116.436, 116.513,
        116.567, 116.633, 116.666, 116.743, 116.788, 116.8, 116.8, 116.8, 116.767, 116.767, 116.844, 116.91, 116.964, 117.032, 117.163, 117.208, 117.253, 117.297, 117.351, 117.417, 117.494, 117.572,
        117.637, 117.673, 117.706, 117.715, 117.661, 117.661, 117.673, 117.706, 117.771, 117.837, 117.936, 118.036, 118.09, 118.081, 118.004, 118.046, 118.135, 118.245, 118.344, 118.466, 118.609,
        118.755, 118.907, 119.02, 119.139, 119.25, 119.273, 119.229, 119.053, 118.931, 118.743, 118.642, 118.543, 118.4, 118.269, 118.245, 118.257, 118.29, 118.4, 118.478, 118.576, 118.6, 118.489,
        118.356, 118.245, 118.147, 118.046, 117.926, 117.837, 117.727, 117.706, 117.673, 117.661],
   y = [4.93305, 4.93305, 4.94478, 5.01049, 5.14426, 5.18651, 5.24283, 5.35313, 5.47517, 5.50802, 5.49629, 5.50802, 5.63945, 5.7826, 5.90464, 5.99382, 6.09238, 6.21442, 6.28952, 6.39043, 6.44441, 6.52186,
        6.5993, 6.71899, 6.84102, 6.95133, 6.93959, 6.895, 6.82929, 6.69787, 6.5993, 6.54298, 6.63216, 6.7096, 6.78705, 6.9302, 6.9302, 6.87388, 6.75419, 6.62042, 6.57583, 6.54298, 6.5993, 6.53359,
        6.47727, 6.44441, 6.32472, 6.23554, 6.1581, 6.02667, 5.91637, 5.8929, 5.92576, 5.91637, 5.97035, 5.97035, 5.88352, 5.7826, 5.71689, 5.6723, 5.6723, 5.77321, 5.74975, 5.69577, 5.58547, 5.48455,
        5.43058, 5.40946, 5.37425, 5.29681, 5.20997, 5.12079, 5.05508, 4.97764, 4.90019, 4.88846, 4.86734, 4.94478, 4.87907, 4.75704, 4.70071, 4.64673, 4.54582, 4.5247, 4.45899, 4.39093, 4.36981,
        4.34868, 4.29236, 4.27124, 4.22665, 4.23838, 4.27124, 4.34868, 4.33695, 4.24777, 4.19379] };
  WCL[66] = struct {
   x = [117.661, 117.661, 117.637, 117.551, 117.45, 117.595, 117.694, 117.825, 117.825, 117.727, 117.595, 117.562, 117.473, 117.318, 117.241, 117.297, 117.384, 117.461, 117.473, 117.407, 117.45, 117.572,
        117.649, 117.715, 117.816, 117.959, 118.046, 117.992, 117.971, 117.926, 117.914, 117.947, 118.114, 118.278, 118.478, 118.677, 118.865, 118.975, 118.975, 118.809, 118.588, 118.466, 118.311,
        118.191, 118.102, 118.046, 118.004, 118.069, 118.046, 117.849, 117.715, 117.628, 117.583, 117.583],
   y = [4.19379, 4.16094, 4.12574, 4.07176, 4.04829, 3.98258, 3.88401, 3.85116, 3.69626, 3.63994, 3.6728, 3.61882, 3.60709, 3.60709, 3.54137, 3.45219, 3.35363, 3.30904, 3.24098, 3.187, 3.10956, 3.02038,
        2.91946, 2.78804, 2.63315, 2.41255, 2.27878, 2.1896, 2.11215, 2.06756, 1.99012, 1.89155, 1.74605, 1.59115, 1.42453, 1.2579, 1.10301, 0.992711, 0.903532, 0.849554, 0.83782, 0.826086, 0.870676,
        0.894144, 1.01383, 1.07016, 0.980977, 0.83782, 0.79323, 0.704051, 0.528038, 0.3708, 0.215909, 0.007041] };
  WCL[67] = struct {
   x = [119.13, 118.952, 118.82, 118.755, 118.543, 118.344, 118.147, 117.926, 117.759, 117.605, 117.494, 117.375, 117.241, 117.297, 117.429, 117.562, 117.825, 117.936, 118.102, 118.269, 118.51],
   y = [10.0093, 9.94354, 9.85671, 9.59386, 9.36387, 9.21133, 9.03766, 8.80767, 8.67625, 8.64339, 8.51197, 8.41106, 8.43453, 8.66451, 8.89451, 9.02593, 9.22306, 9.24418, 9.4202, 9.6056, 9.84497] };
  WCL[68] = struct {
   x = [119.879, 119.846, 119.902, 119.968],
   y = [4.98937, 5.03162, 5.13253, 5.18651] };
  WCL[69] = struct {
   x = [120.134, 119.98, 119.879],
   y = [5.09967, 5.04335, 4.98937] };
  WCL[70] = struct {
   x = [119.792, 119.914, 119.914, 119.956],
   y = [-0.049284, 0.150197, 0.260499, 0.471714] };
  WCL[71] = struct {
   x = [104.049, 104.049, 104.049, 104.038, 104.038, 104.026, 104.016, 104.016, 103.939, 103.817, 103.751, 103.751, 103.871, 103.993, 104.049],
   y = [1.43626, 1.52544, 1.52544, 1.52544, 1.54656, 1.54656, 1.54656, 1.5583, 1.5583, 1.50197, 1.37055, 1.33535, 1.32596, 1.35882, 1.43626] };
  WCL[72] = struct {
   x = [100.208, 100.074],
   y = [6.44441, 6.67675] };
  WCL[73] = struct {
   x = [102.216, 102.437, 102.559, 102.679, 102.789, 102.923, 103.087, 103.242, 103.385, 103.495, 103.528, 103.528, 103.507, 103.474, 103.495, 103.507, 103.519, 103.629, 103.761, 103.939, 104.038, 104.214,
        104.324, 104.279, 104.202, 104.082, 103.939, 103.805, 103.695, 103.65, 103.573, 103.519, 103.674, 103.65, 103.585, 103.519, 103.507, 103.441, 103.23, 102.989, 102.812, 102.547, 102.294,
        102.139, 101.907, 101.609, 101.442, 101.409, 101.299, 101.135, 100.947, 100.804, 100.78, 100.67, 100.693, 100.682, 100.538, 100.47, 100.449, 100.438, 100.372, 100.327, 100.25, 100.208],
   y = [6.19095, 6.20268, 6.05953, 5.92576, 5.83893, 5.73801, 5.61832, 5.40946, 5.18651, 4.97764, 4.7453, 4.4144, 4.19379, 3.97085, 3.68453, 3.49678, 3.23159, 2.88661, 2.79977, 2.65427, 2.40081, 2.05583,
        1.86808, 1.57942, 1.448, 1.51371, 1.57942, 1.64748, 1.57942, 1.46912, 1.40341, 1.43626, 1.39167, 1.39167, 1.40341, 1.48085, 1.50197, 1.45738, 1.65687, 1.80237, 1.87982, 2.09103, 2.1896,
        2.32337, 2.52285, 2.73172, 2.91008, 3.14241, 3.37475, 3.5625, 3.8183, 4.00605, 4.13747, 4.33695, 4.60214, 4.82275, 4.93305, 5.12079, 5.41884, 5.70516, 5.94923, 6.2684, 6.43267, 6.44441] };
  WCL[74] = struct {
   x = [99.9095, 100.008, 100.041, 100.13, 100.294, 100.405, 100.47, 100.527, 100.616, 100.625, 100.714, 100.947, 101.266, 101.576, 101.841, 102.062, 102.216],
   y = [9.24418, 9.05878, 8.76308, 8.50024, 8.32422, 8.14821, 7.81966, 7.51222, 7.32447, 7.31508, 7.25876, 6.98418, 6.9302, 6.895, 6.68613, 6.43267, 6.19095] };
  WCL[75] = struct {
   x = [106.631, 106.61],
   y = [10.0867, 9.98813] };
  WCL[76] = struct {
   x = [106.499, 106.443, 106.544, 106.478, 106.3, 106.145, 106.014, 106.014, 106.136, 106.145, 106.103, 106.079, 105.903, 105.694, 105.528, 105.34, 105.087, 104.92, 104.887, 104.976, 104.887, 104.843,
        104.833, 104.843, 104.866, 104.997],
   y = [10.0093, 9.91069, 9.72528, 9.57274, 9.63845, 9.82385, 9.94354, 9.73702, 9.56101, 9.42958, 9.37561, 9.35449, 9.28877, 9.20194, 9.03766, 8.7842, 8.5988, 8.58707, 8.67625, 8.75135, 8.97195, 9.19021,
        9.39673, 9.56101, 9.69243, 9.86844] };
  WCL[77] = struct {
   x = [99.886, 100.053, 100.107, 100.229, 100.351, 100.461, 100.604, 100.714, 100.825, 100.923, 100.935, 100.98, 101.156, 101.278, 101.365, 101.4, 101.442, 101.477, 101.519, 101.609, 101.698, 101.796,
        101.862, 102.005, 102.094, 102.172, 102.216, 102.282, 102.36, 102.371, 102.371, 102.446, 102.536, 102.669, 102.822, 103.021, 103.165, 103.209, 103.188, 103.176, 103.122, 103.042, 103.066,
        103.099, 103.111, 103.132, 103.132, 103.01, 102.911, 102.89, 102.923, 103.042, 103.165, 103.287, 103.409, 103.507, 103.596, 103.74, 103.817, 103.838],
   y = [3.13068, 2.94293, 2.75518, 2.73172, 2.69886, 2.47826, 2.32337, 2.21307, 2.13327, 2.09103, 2.16848, 2.25531, 2.25531, 2.16848, 2.01358, 1.91267, 1.82349, 1.76951, 1.73666, 1.69207, 1.65687, 1.64748,
        1.61462, 1.52544, 1.49259, 1.40341, 1.32596, 1.20393, 1.15934, 1.29311, 1.35882, 1.31423, 1.18046, 1.08189, 1.09363, 1.10301, 1.02557, 0.97159, 0.927, 0.870676, 0.804964, 0.804964, 0.682929,
        0.605483, 0.605483, 0.560894, 0.438858, 0.361413, 0.305089, 0.260499, 0.239377, 0.293354, 0.349678, 0.438858, 0.492835, 0.528038, 0.471714, 0.328557, 0.173666, 0.072752] };
  WCL[78] = struct {
   x = [101.817, 101.731, 101.731, 101.698, 101.641, 101.597, 101.552, 101.543, 101.552, 101.62, 101.698, 101.785, 101.841, 101.895, 101.895, 101.895, 101.874, 101.841, 101.817],
   y = [1.79063, 1.80237, 1.80237, 1.83523, 1.85635, 1.91267, 1.969, 2.02297, 2.06756, 2.06756, 2.11215, 2.13327, 2.0793, 2.04644, 1.94553, 1.87982, 1.82349, 1.80237, 1.79063] };
  WCL[79] = struct {
   x = [102.16, 102.16, 102.205, 102.303, 102.371, 102.458, 102.524, 102.58, 102.601, 102.58, 102.547, 102.47, 102.371, 102.27, 102.193, 102.16],
   y = [1.6686, 1.6686, 1.6686, 1.62401, 1.63574, 1.62401, 1.61462, 1.448, 1.37055, 1.34708, 1.39167, 1.448, 1.48085, 1.53483, 1.61462, 1.6686] };
  WCL[80] = struct {
   x = [104.666, 104.744, 104.765, 104.732, 104.645, 104.601, 104.544, 104.458, 104.425, 104.446, 104.535, 104.622, 104.666],
   y = [1.17107, 1.11475, 1.00445, 0.894144, 0.88241, 0.992711, 1.07016, 1.07016, 1.18046, 1.21331, 1.22505, 1.20393, 1.17107] };
  WCL[81] = struct {
   x = [109.623, 109.679, 109.724, 109.745, 109.778, 109.844, 109.966],
   y = [2.05583, 2.10042, 2.12389, 2.05583, 1.95726, 1.87982, 1.83523] };
  WCL[82] = struct {
   x = [109.203, 109.194, 109.093, 109.027, 109.006, 108.994, 108.994, 109.006, 109.018, 109.06, 109.128, 109.203, 109.337, 109.459, 109.557, 109.623],
   y = [-0.082139, 0.138463, 0.316823, 0.382534, 0.483448, 0.638339, 0.816699, 0.927, 1.05842, 1.22505, 1.41514, 1.57942, 1.72492, 1.94553, 2.00185, 2.05583] };
  WCL[83] = struct {
   x = [100.074, 99.9306, 99.7874, 99.6865, 99.4776, 99.3134, 99.2125, 99.1022, 98.9261, 98.7384, 98.5506, 98.384, 98.33, 98.33, 98.3629, 98.4403, 98.5272, 98.6164, 98.6398],
   y = [6.67675, 6.8199, 6.97245, 7.22591, 7.41365, 7.65538, 7.80792, 7.86425, 8.04026, 8.25851, 8.24678, 8.22566, 8.38994, 8.66451, 8.98134, 9.25592, 9.55162, 9.81212, 9.96466] };
  WCL[84] = struct {
   x = [99.1913, 99.2242, 99.2688, 99.3228, 99.5011, 99.7311, 99.9095],
   y = [10.0093, 9.70416, 9.39673, 9.24418, 9.20194, 9.28877, 9.24418] };
  WCL[85] = struct {
   x = [98.384, 98.3394, 98.3394, 98.384, 98.4849, 98.4943, 98.4192, 98.384],
   y = [8.11536, 8.10597, 7.93934, 7.84078, 7.92996, 8.04964, 8.14821, 8.11536] };
  WCL[86] = struct {
   x = [93.8476, 93.7584, 93.681, 93.681, 93.749, 93.8593, 93.8922, 93.8593, 93.7584],
   y = [7.28223, 7.30335, 7.20478, 7.02877, 6.82929, 6.95133, 7.18131, 7.31508, 7.30335] };
  WCL[87] = struct {
   x = [95.3261, 95.3261, 95.3707, 95.4364, 95.4927, 95.5045, 95.5256, 95.5373, 95.5467, 95.6148, 95.7251, 95.8354, 95.9339, 96.0114, 96.0771, 96.2109, 96.3212, 96.4315, 96.5629, 96.6614, 96.8163, 96.983,
        97.1379, 97.3139, 97.5345, 97.6894, 97.8208, 97.9757, 98.0531, 98.0649, 98.1306, 98.2643, 98.3629, 98.3958, 98.4403, 98.5272, 98.6727, 98.7595, 98.8816, 99.0904, 99.2899, 99.5105, 99.6443,
        99.886],
   y = [5.60659, 5.59485, 5.60659, 5.60659, 5.63006, 5.63006, 5.63006, 5.63006, 5.63006, 5.63006, 5.63006, 5.60659, 5.55261, 5.44231, 5.3414, 5.33201, 5.32027, 5.26395, 5.26395, 5.25456, 5.27569, 5.28742,
        5.27569, 5.18651, 5.19824, 5.16538, 5.05508, 4.93305, 4.80162, 4.69132, 4.59041, 4.51296, 4.34868, 4.29236, 4.17033, 4.09288, 3.98258, 3.87228, 3.75024, 3.66106, 3.5625, 3.46393, 3.33016,
        3.13068] };
  WCL[88] = struct {
   x = [99.9634, 99.8649, 99.7217, 99.5434, 99.4119, 99.3791, 99.3556, 99.3016, 99.2688, 99.2125, 99.1467, 99.0576, 98.9379, 98.9379, 98.7924, 98.6046, 98.4192, 98.2526, 98.0766, 97.856, 97.7997, 97.7551,
        97.7011, 97.5673, 97.5579, 97.5133, 97.4242, 97.335, 97.2153, 97.0276, 96.8844, 96.6966, 96.5089, 96.354, 96.232, 96.0653, 95.9011, 95.7908, 95.6476, 95.5467, 95.4599, 95.3824, 95.3707,
        95.3707, 95.359, 95.3261],
   y = [-0.147851, 0.028162, 0.138463, 0.251111, 0.251111, 0.260499, 0.293354, 0.483448, 0.671195, 0.849554, 1.07016, 1.28137, 1.50197, 1.69207, 1.86808, 2.04644, 2.14501, 2.23419, 2.2999, 2.33275, 2.52285,
        2.7106, 2.89834, 2.94293, 2.98752, 3.05323, 3.14241, 3.2973, 3.47332, 3.65168, 3.74085, 3.74085, 3.91687, 4.10461, 4.21492, 4.31583, 4.49184, 4.61388, 4.76877, 4.90019, 5.06682, 5.2311,
        5.33201, 5.41884, 5.55261, 5.60659] };
  WCL[89] = struct {
   x = [95.8682, 95.8447, 95.8799, 95.9222, 96.056, 96.1991, 96.3423, 96.476, 96.53, 96.5511, 96.53, 96.408, 96.232, 96.0771, 95.9457, 95.8682],
   y = [2.80916, 2.86549, 2.87722, 2.91946, 2.83263, 2.73172, 2.64488, 2.54397, 2.4454, 2.37734, 2.36796, 2.47826, 2.53224, 2.62141, 2.69886, 2.80916] };
  WCL[90] = struct {
   x = [97.3021, 97.2927, 97.335, 97.4242, 97.5016, 97.6448, 97.7551, 97.856, 98.032, 98.0414, 98.0414, 98.0531, 98.0531, 98.0085, 97.91, 97.8208, 97.7433, 97.6448, 97.5345, 97.4242, 97.3702, 97.335,
        97.3021],
   y = [1.40341, 1.45738, 1.48085, 1.51371, 1.54656, 1.41514, 1.2579, 1.17107, 1.02557, 0.97159, 0.870676, 0.760375, 0.694663, 0.614871, 0.614871, 0.79323, 0.903532, 0.980977, 1.08189, 1.24852, 1.31423,
        1.37055, 1.40341] };
  WCL[91] = struct {
   x = [80.0247, 80.0247],
   y = [9.81212, 9.81212] };
  WCL[92] = struct {
   x = [79.9262, 80.0482, 80.2031, 80.2782, 80.1914, 80.0811, 80.0482],
   y = [9.74641, 9.65019, 9.58447, 9.46244, 9.47417, 9.35449, 9.11276] };
  WCL[93] = struct {
   x = [79.8933, 80.0482, 80.2125, 80.5105, 80.7546, 80.9517, 81.2404, 81.5267, 81.7356, 81.7919, 81.813, 81.7802, 81.6581, 81.5713, 81.4821, 81.3718, 81.2404, 81.1958, 81.1958, 81.1301, 81.1066, 81.0409,
        80.9189, 80.8532, 80.6654, 80.5011, 80.3345, 80.1796, 80.0247],
   y = [6.51012, 6.20268, 6.01494, 5.95861, 5.95861, 6.05953, 6.22381, 6.42329, 6.71899, 7.16019, 7.53334, 7.68823, 7.73282, 7.95108, 8.22566, 8.45565, 8.56595, 8.67625, 8.69737, 8.80767, 8.81706, 8.89451,
        9.00481, 9.19021, 9.40846, 9.58447, 9.73702, 9.80273, 9.81212] };
  WCL[94] = struct {
   x = [79.1869, 79.1541, 79.032, 78.9006, 78.8771, 78.8889, 78.9006, 78.7879, 78.5908, 78.3913, 78.1825, 78.1496, 78.1261, 78.0487, 78.0722, 78.0276, 77.9173, 77.8398, 77.7952, 77.7952, 77.6403, 77.3306,
        77.1006, 76.8354, 76.603, 76.481, 76.3379],
   y = [10.0961, 9.96466, 9.80273, 9.61733, 9.44132, 9.32163, 9.2653, 9.24418, 9.22306, 9.13623, 9.02593, 8.96022, 8.84991, 8.65278, 8.57768, 8.47912, 8.36881, 8.32422, 8.3031, 8.25851, 8.11536, 8.10597,
        8.3031, 8.63166, 8.98134, 9.23245, 9.74641] };
  WCL[95] = struct {
   x = [80.0247, 79.9262],
   y = [9.81212, 9.74641] };
  WCL[96] = struct {
   x = [80.0482, 79.9262, 79.905, 79.8698, 79.8276, 79.8041, 79.7267, 79.6727, 79.6727, 79.7267, 79.7713, 79.8159, 79.8933],
   y = [9.11276, 8.94848, 8.74196, 8.54483, 8.32422, 8.09423, 8.14821, 8.09423, 7.80792, 7.41365, 7.09448, 6.79643, 6.51012] };
  WCL[97] = struct {
   x = [50.8324, 50.755, 50.7104, 50.7221, 50.6564, 50.5555, 50.403, 50.3255, 50.1706, 50.0721],
   y = [10.1594, 9.88487, 9.63376, 9.40377, 9.20663, 8.98838, 8.85696, 8.62697, 8.4533, 8.21158] };
  WCL[98] = struct {
   x = [50.0721, 49.9946, 49.8726, 49.8069, 49.7623, 49.6637, 49.5534, 49.3891, 49.2882, 49.1685, 49.0911, 49.0911, 49.0676, 49.023, 48.9479, 48.8048, 48.5513, 48.3518, 48.143, 48.0655, 47.967, 47.7675,
        47.6454, 47.4366, 47.1949, 47.0846, 46.9179, 46.8194, 46.6879, 46.587, 46.4227, 46.289, 46.1787, 46.0473, 45.9158, 45.7046, 45.5075, 45.308, 45.0991, 44.8457, 44.7354, 44.6133, 44.4162,
        44.3388, 44.205, 44.0079, 43.8201, 43.6324, 43.5009, 43.3906, 43.2475, 43.1372],
   y = [8.21158, 8.16699, 8.03556, 7.88067, 7.6413, 7.47467, 7.24468, 6.9701, 6.80347, 6.60634, 6.43033, 6.50777, 6.34115, 6.05483, 5.83423, 5.61363, 5.4165, 5.20528, 4.98468, 4.74295, 4.51062, 4.26655,
        4.09053, 3.8582, 3.63759, 3.43811, 3.30434, 3.20578, 3.09548, 3.03915, 2.91712, 2.68478, 2.57448, 2.45244, 2.33041, 2.16613, 2.03236, 1.93145, 1.8892, 1.7437, 1.66625, 1.56769, 1.4128, 1.3682,
        1.21331, 1.06781, 0.880063, 0.701704, 0.570281, 0.481101, 0.32621, 0.19244] };
  WCL[99] = struct {
   x = [36.1014, 36.092, 36.2117, 36.3008, 36.3008, 36.3666, 36.2891, 36.3126, 36.3783, 36.5214, 36.6317, 36.7209, 36.7421, 36.7538, 36.6646, 36.6317, 36.566, 36.4229, 36.2117, 36.092, 36.0474, 35.9699,
        35.8925, 35.916, 35.9582, 35.9911],
   y = [4.52235, 4.52235, 4.36746, 4.17971, 3.93564, 3.73616, 3.63759, 3.36067, 3.08374, 2.86314, 2.76222, 2.67539, 2.50877, 2.35388, 2.36326, 2.5205, 2.71764, 2.85141, 2.98517, 3.21751, 3.43811, 3.593,
        3.89105, 4.17971, 4.38858, 4.52235] };
  WCL[100] = struct {
   x = [34.0291, 34.0197],
   y = [0.23703, 0.114995] };
  WCL[101] = struct {
   x = [34.0737, 33.9751, 33.8085, 33.6653, 33.4236, 33.3016, 33.203, 33.081, 32.9378, 32.7407, 32.6515, 32.6187, 32.5295, 32.4192],
   y = [0.23703, 0.204175, 0.248765, 0.19244, 0.28162, 0.359066, 0.23703, 0.058671, 0.126729, 0.147851, 0.204175, 0.093873, 0.058671, 0.103261] };
  WCL[102] = struct {
   x = [5.00111, 4.9659, 4.82275, 4.59276, 4.4496, 4.27124, 4.07411, 3.86524, 3.54372, 3.24802, 2.98283, 2.87253, 2.79508],
   y = [5.82484, 5.91168, 6.09942, 6.22146, 6.30829, 6.35288, 6.36462, 6.38574, 6.37635, 6.35288, 6.33176, 6.33176, 6.32003] };
  WCL[103] = struct {
   x = [2.79508, 2.72937, 2.57448, 2.33275, 2.12154, 1.97838],
   y = [6.32003, 6.32003, 6.30829, 6.27544, 6.26605, 6.25432] };
  WCL[104] = struct {
   x = [1.97838, 1.82584, 1.60523, 1.3729, 1.27433],
   y = [6.25432, 6.22146, 6.16514, 6.09942, 6.05483] };
  WCL[105] = struct {
   x = [1.27433, 1.21801, 1.09597, 1.00914, 0.842514, 0.767415, 0.621911, 0.403656, 0.19244],
   y = [6.05483, 6.03371, 5.86944, 5.7474, 5.71454, 5.78026, 5.7474, 5.72393, 5.66995] };
  WCL[106] = struct {
   x = [5.00111, 5.0762, 5.12079, 5.18651, 5.25456, 5.38599, 5.4517, 5.35313, 5.36486, 5.4517, 5.562, 5.6723, 5.76148, 5.98208, 6.16748, 6.21207, 6.41155, 6.53359, 6.56644, 6.63216, 6.73072, 6.78705,
        6.82929, 6.89735, 6.96306, 7.02877, 7.10622, 7.24937, 7.52395, 7.73517, 7.9323, 8.15291, 8.17637, 8.24208, 8.34065, 8.39698],
   y = [5.82484, 5.81311, 5.70281, 5.6371, 5.44935, 5.40476, 5.40476, 5.33905, 5.13018, 4.79693, 4.69836, 4.57633, 4.42144, 4.35573, 4.37919, 4.21257, 4.31114, 4.37919, 4.34399, 4.38858, 4.3346, 4.37919,
        4.37919, 4.41205, 4.49888, 4.53174, 4.65377, 4.5998, 4.47776, 4.49888, 4.49888, 4.4895, 4.69836, 4.86499, 4.82979, 4.74295] };
  WCL[107] = struct {
   x = [9.23245, 9.25592, 9.33336, 9.4202, 9.57509, 9.79569, 9.84028, 9.78396, 9.61968, 9.47652, 9.37795, 9.30989, 9.34275, 9.47652, 9.5305, 9.52111, 9.52111],
   y = [-0.194787, 0.248765, 0.23703, 0.147851, 0.093873, 0.082139, 0.136116, 0.180706, 0.213562, 0.314476, 0.403656, 0.513957, 0.603137, 0.612524, 0.701704, 0.880063, 0.97863] };
  WCL[108] = struct {
   x = [9.52111, 9.5305, 9.57509, 9.56335, 9.30989, 9.34275, 9.44366, 9.50938, 9.55397, 9.67365, 9.67365],
   y = [0.97863, 0.957509, 1.02322, 1.06781, 1.2227, 1.29076, 1.43392, 1.54422, 1.75543, 1.98777, 2.16613] };
  WCL[109] = struct {
   x = [9.67365, 9.68539, 9.70886, 9.71824, 9.76283, 9.84028, 9.85201, 9.88487, 9.78396, 9.66427, 9.63141, 9.54223, 9.54223, 9.60794, 9.6408, 9.63141, 9.57509, 9.45305, 9.43193, 9.34275, 9.13388, 8.98134,
        8.89216, 8.83583, 8.79359, 8.72553, 8.61523, 8.5284, 8.43922, 8.43922, 8.45095, 8.39698],
   y = [2.16613, 2.23184, 2.32102, 2.54162, 2.79508, 3.08374, 3.19404, 3.23863, 3.3724, 3.49444, 3.56015, 3.68218, 3.7479, 3.76902, 3.84646, 4.00135, 4.04594, 4.03421, 3.9685, 3.94738, 4.02482, 4.10227,
        4.26655, 4.42144, 4.54347, 4.54347, 4.54347, 4.51062, 4.56694, 4.60918, 4.67724, 4.74295] };
  WCL[110] = struct {
   x = [8.60584, 8.6387, 8.72553, 8.81471, 8.88042, 8.8593, 8.79359, 8.65982, 8.51666, 8.4181, 8.43922, 8.50728, 8.58238, 8.60584],
   y = [3.81361, 3.79249, 3.7479, 3.70331, 3.62586, 3.51556, 3.3372, 3.2621, 3.29496, 3.39352, 3.44985, 3.58127, 3.7479, 3.81361] };
  WCL[111] = struct {
   x = [6.489, 6.44441, 6.44441, 6.44441, 6.489, 6.56644, 6.63216, 6.56644, 6.489],
   y = [0.314476, 0.32621, 0.23703, 0.082139, 0.082139, 0.171319, 0.302742, 0.302742, 0.314476] };
  WCL[112] = struct {
   x = [-14.0294, -14.25, -14.3485, -14.4377, -14.4823, -14.5034, -14.4917, -14.5809, -14.6701, -14.6701, -14.7029, -14.6912, -14.7475, -14.8226, -14.8789, -14.9681, -15.0549],
   y = [9.9717, 10.1149, 10.1688, 10.312, 10.4199, 10.5725, 10.6593, 10.6593, 10.6922, 10.7462, 10.8565, 10.9644, 10.9527, 10.8565, 10.7696, 10.7462, 10.833] };
  WCL[113] = struct {
   x = [-15.0549, -15.0878, -15.0549, -15.0221, -15.177, -15.2756, -15.353, -15.3859, -15.353, -15.3319, -15.3976, -15.4304, -15.2967, -15.1981, -15.0432, -14.8578, -14.8343, -15.0432, -15.2756, -15.4187,
        -15.6722, -15.8153, -15.881, -15.8928, -16.0031, -16.1908, -16.2565, -16.2683, -16.2683, -16.4325],
   y = [10.833, 10.8776, 11.0512, 11.0958, 11.0724, 11.0958, 11.2038, 11.3352, 11.4314, 11.5089, 11.6168, 11.7037, 11.7999, 11.8327, 11.8445, 11.8539, 11.943, 11.9196, 11.8985, 11.8539, 11.7365, 11.7131,
        11.7365, 11.8773, 11.8656, 11.8985, 11.9853, 12.0839, 12.1472, 12.1918] };
  WCL[114] = struct {
   x = [-16.4325, -16.6438, -16.7541, -16.7189, -16.7189, -16.7095, -16.6977],
   y = [12.1918, 12.2552, 12.396, 12.5485, 12.7644, 12.8935, 13.0132] };
  WCL[115] = struct {
   x = [-16.6977, -16.686, -16.7189, -16.6766, -16.6203, -16.5428, -16.4443, -16.334, -16.3011, -16.3786, -16.3997],
   y = [13.0132, 13.0789, 13.208, 13.391, 13.4333, 13.3371, 13.262, 13.2385, 13.2714, 13.3582, 13.4544] };
  WCL[116] = struct {
   x = [-16.3997, -16.4325, -16.5006, -16.5006, -16.6531, -16.7634, -16.9066, -17.0732, -17.1835, -17.2046, -17.2046, -17.1507, -17.0075, -16.7541, -16.5217, -16.4114, -16.3997],
   y = [13.4544, 13.5412, 13.6914, 13.7571, 13.9496, 14.1866, 14.433, 14.595, 14.6795, 14.724, 14.7452, 14.8414, 14.9705, 15.2896, 15.6862, 15.867, 15.9092] };
  WCL[117] = struct {
   x = [-16.3997, -16.3997, -16.3997, -16.3997, -16.334, -16.1345, -16.0242, -15.9913, -15.9819, -16.0359, -16.1016, -16.2237, -16.3457, -16.4443, -16.4654, -16.3997, -16.3457, -16.2565, -16.2025],
   y = [15.9092, 15.9421, 16.1321, 16.5569, 16.9395, 17.3854, 17.7022, 18.1223, 18.418, 18.9014, 19.0164, 19.1314, 19.2886, 19.413, 19.4647, 19.4553, 19.528, 19.6946, 19.934] };
  WCL[118] = struct {
   x = [-15.881, -16.0031, -16.1345, -16.0922, -15.9374, -15.881],
   y = [11.0958, 11.1052, 11.0512, 10.9433, 11.0301, 11.0958] };
  WCL[119] = struct {
   x = [-24.9891, -25.1886, -25.2989, -25.2754, -25.2426, -25.0783, -24.968, -24.9445, -24.9891],
   y = [17.1413, 17.1202, 17.1202, 16.9512, 16.9089, 16.9606, 17.0357, 17.0873, 17.1413] };
  WCL[120] = struct {
   x = [-23.9307, -23.9753, -24.095, -24.2498, -24.3156, -24.3273, -24.2827, -24.217, -24.163, -24.0292, -23.9307],
   y = [16.5569, 16.5804, 16.5898, 16.6109, 16.6226, 16.5264, 16.4724, 16.5264, 16.5358, 16.5358, 16.5569] };
  WCL[121] = struct {
   x = [-22.8934, -22.9051, -22.9262, -22.9262, -22.8511, -22.8159, -22.884, -22.8934],
   y = [16.8878, 16.8761, 16.7376, 16.632, 16.632, 16.7799, 16.8667, 16.8878] };
  WCL[122] = struct {
   x = [-22.8723, -22.7408, -22.6634, -22.6399, -22.7408, -22.8277, -22.884, -22.9051, -22.8723],
   y = [16.1228, 16.1439, 16.1016, 15.9843, 15.8975, 15.8975, 15.9726, 16.0359, 16.1228] };
  WCL[123] = struct {
   x = [-23.6983, -23.7218, -23.7218, -23.7218, -23.6115, -23.4355, -23.3909, -23.4449, -23.5787, -23.6983],
   y = [15.2474, 15.2474, 15.0456, 14.8836, 14.8508, 14.9047, 15.001, 15.0761, 15.184, 15.2474] };
  WCL[124] = struct {
   x = [-24.339, -24.3484, -24.4376, -24.4587, -24.4259, -24.339, -24.2733, -24.339],
   y = [14.9165, 14.9165, 14.9047, 14.8203, 14.7663, 14.7546, 14.8297, 14.9165] };
  WCL[125] = struct {
   x = [-70.0601, -69.8841, -69.7292, -69.532, -69.3208, -69.1237, -68.9242, -68.8139, -68.6614, -68.5276, -68.4173, -68.3844, -68.5065, -68.605, -68.7036, -68.8256, -69.0134, -69.1565, -69.2011, -69.2551,
        -69.3983, -69.5977, -69.5977, -69.3771, -69.234, -69.3771, -69.6095, -69.8395, -69.9498],
   y = [18.3851, 18.4696, 18.4602, 18.4391, 18.4391, 18.4391, 18.3851, 18.2701, 18.249, 18.3546, 18.5118, 18.6808, 18.7442, 18.8169, 18.9319, 19.0047, 19.0469, 19.0586, 19.0586, 19.068, 19.0892, 19.1619,
        19.1924, 19.1736, 19.2558, 19.3614, 19.3285, 19.4647, 19.643] };
  WCL[126] = struct {
   x = [-66.047, -66.0587, -65.8921, -65.7607, -65.6715, -65.8053, -65.9813, -66.2465, -66.4553, -66.622, -66.8308, -67.0514, -67.204, -67.1946, -67.2392, -67.2486, -67.1383, -66.9505, -66.6971, -66.4765,
        -66.2465, -66.047],
   y = [18.5024, 18.5236, 18.4696, 18.3851, 18.2818, 18.1645, 18.0073, 17.9744, 17.9439, 17.9955, 17.9955, 17.9744, 18.0073, 18.1974, 18.3335, 18.4391, 18.5236, 18.533, 18.533, 18.533, 18.5236, 18.5024] };
  WCL[127] = struct {
   x = [-64.9346, -65.0214, -65.012, -64.9346, -64.9346],
   y = [18.4062, 18.3429, 18.2912, 18.3124, 18.4062] };
  WCL[128] = struct {
   x = [-64.8454, -64.8689, -64.8454, -64.7351, -64.8454],
   y = [17.7655, 17.7327, 17.6811, 17.7867, 17.7655] };
  WCL[129] = struct {
   x = [-62.4094, -62.4634, -62.508, -62.5971, -62.6628, -62.7286, -62.8717, -62.9609, -62.9609, -62.7168, -62.5854, -62.508, -62.3531, -62.2334, -62.1442, -61.9799, -61.9564, -62.2099, -62.2874, -62.2874,
        -62.3765, -62.6183, -62.7731, -62.928, -63.1252, -63.2918, -63.5781, -63.8316, -63.9747, -64.1296, -64.2282, -64.2493, -64.2493, -64.139, -64.0404, -63.8527, -63.7213, -63.6978, -63.8973,
        -64.1836, -64.3596, -64.4934, -64.5474, -64.714, -65.0331, -65.2749, -65.3969, -65.4086, -65.5401, -65.7161, -65.9601, -66.0259, -66.0352, -66.047, -66.047, -66.047, -66.1808, -66.3662,
        -66.7651, -66.974, -67.2931, -67.6123, -67.8001, -67.8892, -68.077, -68.1967, -68.2647, -68.2741, -68.2741, -68.2413, -68.2741, -68.3633, -68.5604, -68.6825, -68.7599, -68.9923, -69.0134,
        -69.1237, -69.1894, -69.2105, -69.2222, -69.2668, -69.356, -69.4311, -69.4874, -69.586, -69.6095, -69.6306, -69.6635, -69.708, -69.7292, -69.762, -69.8301, -69.9404],
   y = [9.8966, 10.1594, 10.1806, 10.2228, 10.3331, 10.366, 10.366, 10.3871, 10.5068, 10.5514, 10.5514, 10.5279, 10.5279, 10.6382, 10.6382, 10.6804, 10.725, 10.7368, 10.7462, 10.7133, 10.6922, 10.7133,
        10.7462, 10.7462, 10.7579, 10.7696, 10.7133, 10.6922, 10.6922, 10.6804, 10.6711, 10.5936, 10.5608, 10.5725, 10.5725, 10.5608, 10.5396, 10.5068, 10.4317, 10.4739, 10.3754, 10.2885, 10.2463,
        10.2017, 10.1477, 10.1477, 10.1688, 10.1688, 10.1688, 10.2557, 10.3988, 10.4645, 10.4739, 10.4739, 10.5185, 10.6382, 10.6593, 10.6711, 10.6382, 10.6382, 10.5725, 10.5279, 10.495, 10.4857,
        10.5279, 10.6476, 10.8001, 10.8119, 10.8776, 10.8893, 11.0841, 11.2484, 11.3024, 11.3657, 11.4197, 11.4314, 11.4314, 11.4854, 11.5089, 11.5089, 11.5089, 11.5183, 11.5089, 11.53, 11.53, 11.53,
        11.5722, 11.5957, 11.638, 11.7037, 11.7576, 11.9313, 12.1472, 12.2458] };
  WCL[130] = struct {
   x = [-70.0178, -69.9286, -69.8723, -69.8183, -69.8183, -69.9826],
   y = [11.638, 11.6708, 11.6825, 11.5511, 11.4432, 11.4314] };
  WCL[131] = struct {
   x = [-69.0462, -69.1237, -69.1237, -69.1119, -69.0462, -68.9359, -68.8045, -68.8045, -68.9359, -69.0462],
   y = [12.2998, 12.4195, 12.3866, 12.288, 12.1918, 12.1472, 12.0932, 12.1472, 12.2012, 12.2998] };
  WCL[132] = struct {
   x = [-68.375, -68.4408, -68.3962, -68.3187, -68.2741, -68.2413, -68.2202, -68.2741, -68.375],
   y = [12.288, 12.342, 12.2552, 12.1801, 12.0839, 12.0721, 12.2247, 12.2669, 12.288] };
  WCL[133] = struct {
   x = [-63.8973, -63.8316, -63.787, -63.8973, -64.0076, -64.0968, -64.1625, -64.3268, -64.3713, -64.2939, -64.1953, -64.1061, -63.9747, -63.8973],
   y = [11.1709, 11.117, 10.9973, 10.9433, 10.8776, 10.8987, 10.9433, 10.9433, 11.0395, 11.0958, 11.0958, 11.0301, 11.1263, 11.1709] };
  WCL[134] = struct {
   x = [-61.5481, -61.5927, -61.649, -61.703, -61.5833, -61.5481],
   y = [12.1801, 12.159, 12.1378, 11.9853, 12.051, 12.1801] };
  WCL[135] = struct {
   x = [-61.0318, -61.074, -61.0858, -61.074, -61.0318, -61.0318, -61.0318],
   y = [13.3042, 13.2831, 13.154, 13.0672, 13.1212, 13.262, 13.3042] };
  WCL[136] = struct {
   x = [-60.9215, -60.9426, -60.9755, -60.9989, -60.9215, -60.8534, -60.8534, -60.8769, -60.9215],
   y = [14.0364, 13.9402, 13.8111, 13.736, 13.7126, 13.7454, 13.8416, 13.9731, 14.0364] };
  WCL[137] = struct {
   x = [-61.1961, -61.1069, -60.9989, -60.9309, -60.8534, -60.8769, -60.9989, -61.0529, -61.1186, -61.1843, -61.2078, -61.1961],
   y = [14.9047, 14.8836, 14.7757, 14.6278, 14.4776, 14.4542, 14.433, 14.4776, 14.6372, 14.7546, 14.8297, 14.9047] };
  WCL[138] = struct {
   x = [-61.3181, -61.3721, -61.3721, -61.3721, -61.3275, -61.2407, -61.2078, -61.2172, -61.2172, -61.3181],
   y = [15.6205, 15.6205, 15.5149, 15.3553, 15.238, 15.238, 15.3436, 15.4398, 15.5572, 15.6205] };
  WCL[139] = struct {
   x = [-61.4167, -61.4941, -61.5152, -61.5481, -61.649, -61.7476, -61.8039, -61.7804, -61.6373, -61.5833, -61.5387, -61.3838, -61.2524, -61.2735, -61.3721, -61.4167],
   y = [16.5053, 16.4936, 16.4302, 16.3457, 16.3457, 16.3668, 16.1861, 16.0148, 16.0054, 16.1861, 16.219, 16.2284, 16.2823, 16.3457, 16.388, 16.5053] };
  WCL[140] = struct {
   x = [-61.825, -61.8461, -61.879, -61.8696, -61.8039, -61.7147, -61.703, -61.7687, -61.825],
   y = [17.153, 17.1413, 17.0662, 17.0028, 16.9934, 17.0451, 17.0873, 17.1202, 17.153] };
  WCL[141] = struct {
   x = [-62.7849, -62.8952, -62.86, -62.7943, -62.7497, -62.7497, -62.7849],
   y = [17.4276, 17.437, 17.3431, 17.2891, 17.3103, 17.4065, 17.4276] };
  WCL[142] = struct {
   x = [-60.5131, -60.5225, -60.6234, -60.722, -60.7877, -60.7431, -60.5906, -60.5131],
   y = [11.3446, 11.3446, 11.3117, 11.2249, 11.1827, 11.1498, 11.2484, 11.3446] };
  WCL[143] = struct {
   x = [-60.9426, -61.1186, -61.4941, -61.6936, -61.5716, -61.703, -61.7593, -61.4824, -61.2524, -61.0318, -60.9637, -60.9872, -60.9872, -60.9637, -60.9426, -60.9426],
   y = [10.833, 10.8119, 10.7579, 10.6711, 10.2017, 10.1266, 10.0937, 10.0703, 10.0609, 10.1594, 10.3214, 10.5396, 10.6476, 10.7579, 10.8119, 10.833] };
  WCL[144] = struct {
   x = [-74.9556, -75.2747, -75.486, -75.7817, -76.0468, -76.3003, -76.5984, -76.9175, -77.1381, -77.3376, -77.6568],
   y = [20.0068, 19.8824, 19.8824, 19.9458, 19.9551, 19.9551, 19.9763, 19.9246, 19.8918, 19.8613, 19.8401] };
  WCL[145] = struct {
   x = [-77.1499, -77.0724, -77.0396, -76.9081, -76.7978, -76.6429, -76.4223, -76.3449, -76.2909, -76.4669, -76.7321, -76.9293, -77.0959, -77.1616, -77.3705, -77.469, -77.7225, -77.8774, -78.0205, -78.22,
        -78.3397, -78.3068, -78.1214, -77.9666, -77.9008, -77.7225, -77.5911, -77.4479, -77.293, -77.1499],
   y = [18.4696, 18.4602, 18.4391, 18.3757, 18.2607, 18.249, 18.249, 18.0589, 17.9228, 17.8923, 17.965, 17.8923, 17.8923, 17.7749, 17.7867, 17.8712, 17.8923, 18.0167, 18.1551, 18.2185, 18.3429, 18.4273,
        18.4813, 18.5447, 18.5658, 18.5541, 18.5236, 18.5024, 18.4813, 18.4696] };
  WCL[146] = struct {
   x = [-73.0476, -73.0147, -72.895, -72.8622, -72.9936, -73.2353, -73.3245, -73.2259, -73.0476],
   y = [18.9624, 18.9413, 18.8897, 18.7324, 18.7958, 18.8169, 18.9413, 18.9835, 18.9624] };
  WCL[147] = struct {
   x = [-71.8038, -71.9117, -72.1323, -72.3647, -72.6299, -72.895, -73.2025, -73.4794, -73.4113, -73.2588, -73.0382, -72.8387, -72.7941, -72.8059, -72.8176, -72.7613, -72.5853, -72.4421, -72.4538, -72.6956,
        -72.949, -73.2588, -73.5545, -73.5897, -73.8197, -74.0966, -74.3595, -74.4275, -74.2844, -74.0403, -73.9089, -73.8197, -73.6437, -73.3785, -73.1814, -72.9373, -72.7847, -72.6416, -72.5524,
        -72.3201, -72.1323, -71.9469, -71.7686],
   y = [19.7979, 19.7674, 19.7369, 19.758, 19.8918, 19.9551, 19.9458, 19.7791, 19.6313, 19.6524, 19.6524, 19.5186, 19.3825, 19.2042, 19.0892, 18.9953, 18.838, 18.7324, 18.5752, 18.5236, 18.4696, 18.4813,
        18.5236, 18.5846, 18.5846, 18.6268, 18.6174, 18.4391, 18.2701, 18.2279, 18.0589, 18.1645, 18.2607, 18.2607, 18.2279, 18.1856, 18.1434, 18.1551, 18.2067, 18.2701, 18.2607, 18.1223, 18.0495] };
  WCL[148] = struct {
   x = [-71.7686, -71.7029, -71.6935, -71.616, -71.5503, -71.4165, -71.2523, -71.1302, -71.0317, -70.8768, -70.7008, -70.6562, -70.5247, -70.2689, -70.0601],
   y = [18.0495, 17.9533, 17.8383, 17.7655, 17.6599, 17.7233, 17.9439, 18.1645, 18.303, 18.364, 18.4391, 18.3124, 18.2067, 18.249, 18.3851] };
  WCL[149] = struct {
   x = [-69.9498, -70.1821, -70.4356, -70.6679, -70.9214, -71.1185, -71.3062, -71.5057, -71.7029, -71.8038],
   y = [19.643, 19.6641, 19.8096, 19.8308, 19.9035, 19.9129, 19.873, 19.9246, 19.873, 19.7979] };
  WCL[150] = struct {
   x = [-71.3062, -71.2616, -71.1959, -71.1959, -71.2288, -71.318, -71.3626, -71.4611, -71.5057, -71.6935, -71.7686, -71.8366, -71.9117, -71.9117, -71.9352, -72.0666, -72.1229, -72.2004, -72.2872, -72.4304,
        -72.6064, -72.7284, -72.8059, -72.895, -72.9608, -73.0593, -73.1579, -73.2353, -73.4231, -73.5897, -73.754, -73.9206, -73.9863, -74.0074, -74.2069, -74.3266, -74.3712, -74.5144, -74.681,
        -74.8007, -74.8453, -75.0447, -75.1879, -75.385, -75.5845, -75.6291, -75.6291, -75.6385, -75.6714],
   y = [11.8234, 11.9642, 12.0932, 12.1472, 12.2247, 12.3326, 12.3749, 12.4195, 12.4288, 12.5039, 12.5039, 12.3749, 12.3655, 12.2669, 12.2247, 12.2129, 12.159, 12.051, 11.8773, 11.7459, 11.6825, 11.6497,
        11.6262, 11.5629, 11.5183, 11.4643, 11.3775, 11.3657, 11.2906, 11.2695, 11.2789, 11.3117, 11.3235, 11.3235, 11.2906, 11.1827, 10.9973, 10.9527, 10.9644, 10.9855, 11.009, 10.9761, 10.8776,
        10.7368, 10.6171, 10.5185, 10.3542, 10.2134, 10.0163] };
  WCL[151] = struct {
   x = [-69.9404, -70.0835, -70.1938, -70.2807, -70.2807, -70.2032, -70.0178],
   y = [12.2458, 12.1683, 12.0064, 11.8867, 11.7131, 11.638, 11.638] };
  WCL[152] = struct {
   x = [-69.9826, -70.2596, -70.4684, -70.6233, -70.9988, -71.1959, -71.3508, -71.3954, -71.44, -71.4611, -71.4283, -71.318],
   y = [11.4314, 11.3986, 11.3235, 11.2789, 11.1381, 11.0512, 10.9527, 10.8236, 10.7462, 10.5279, 10.3542, 10.1149] };
  WCL[153] = struct {
   x = [-72.0126, -71.9023, -71.8695, -71.7146, -71.6043, -71.5503, -71.5503, -71.6043, -71.6043, -71.6254, -71.7592, -71.8577, -71.8906, -71.7803, -71.6043, -71.494, -71.3954, -71.3297, -71.3062],
   y = [9.87548, 10.1031, 10.2228, 10.3871, 10.5279, 10.6171, 10.6922, 10.8776, 10.9433, 11.0184, 11.2578, 11.4103, 11.53, 11.6051, 11.6497, 11.6708, 11.7037, 11.7905, 11.8234] };
  WCL[154] = struct {
   x = [-87.4924, -87.5792, -87.6895, -87.5463, -87.537, -87.6895, -87.6027, -87.5581, -87.6027, -87.6895, -87.7576, -87.7787, -87.8327, -87.8772, -87.943, -87.9875, -87.9664, -88.011, -88.0861, -88.1213,
        -88.0978, -88.065, -88.1964, -88.2856, -88.3959],
   y = [20.0185, 19.8519, 19.7064, 19.643, 19.528, 19.4224, 19.3614, 19.2886, 19.1103, 18.8592, 18.6902, 18.4907, 18.2912, 18.1223, 18.0378, 18.2067, 18.4602, 18.5236, 18.533, 18.5447, 18.6268, 18.8169,
        18.8592, 18.7324, 18.5752] };
  WCL[155] = struct {
   x = [-88.3959, -88.417, -88.363, -88.2856, -88.1636, -88.1753, -88.2645, -88.3067, -88.3184, -88.3184, -88.3396, -88.4499, -88.5602, -88.7033, -88.8371, -88.9239, -88.9591],
   y = [18.5752, 18.4062, 18.3757, 18.3968, 18.249, 17.9861, 17.7749, 17.4792, 17.1741, 16.8456, 16.6531, 16.4724, 16.388, 16.2706, 16.1533, 15.9092, 15.8975] };
  WCL[156] = struct {
   x = [-88.9591, -88.8911, -88.8042, -88.6939, -88.5719, -88.5179, -88.6493, -88.6376, -88.6165, -88.4945, -88.3396],
   y = [15.8975, 15.9209, 15.813, 15.7285, 15.7707, 15.8341, 15.7379, 15.7379, 15.8341, 15.8552, 15.7285] };
  WCL[157] = struct {
   x = [-88.2856, -88.2739, -88.1542, -88.0204, -87.8327, -87.612, -87.3375, -87.0957, -86.786, -86.5442, -86.3002, -86.0467, -85.9927, -85.8261, -85.6618, -85.5398, -85.3427, -85.1549, -85, -84.8451,
        -84.6245, -84.4602, -84.3382, -84.2068, -84.0965, -83.9862, -83.8196, -83.6999, -83.5661, -83.4112, -83.3033, -83.2798],
   y = [15.7285, 15.7285, 15.7168, 15.7825, 15.8552, 15.8458, 15.7919, 15.7707, 15.8036, 15.8458, 15.8763, 15.9209, 15.9937, 15.9843, 15.8763, 15.8763, 15.9092, 15.9514, 15.9421, 15.8763, 15.8341, 15.8341,
        15.813, 15.7285, 15.6323, 15.5149, 15.4398, 15.3553, 15.3108, 15.2263, 15.0761, 14.8954] };
  WCL[158] = struct {
   x = [-83.2798, -83.2915, -83.2681, -83.5215, -83.5661, -83.5778, -83.5661, -83.5896, -83.6553, -83.6553, -83.7093, -83.6999, -83.6999, -83.775, -83.8642, -83.843, -83.6764],
   y = [14.8954, 14.6701, 14.3697, 13.7665, 13.3371, 13.0461, 12.8513, 12.6776, 12.4828, 12.2129, 12.051, 11.8327, 11.6919, 11.53, 11.3446, 11.1592, 10.8776] };
  WCL[159] = struct {
   x = [-83.6764, -83.6553, -83.5661, -83.4347, -83.2352],
   y = [10.8776, 10.7696, 10.5725, 10.366, 10.1149] };
  WCL[160] = struct {
   x = [-84.7912, -84.8897, -85.0211, -85.1549, -85.2206, -85.176],
   y = [9.93885, 10.028, 10.1031, 10.1594, 10.1149, 10.0163] };
  WCL[161] = struct {
   x = [-85.6947, -85.8167, -85.8942, -85.8942, -85.7933, -85.7721, -85.805, -85.8942, -85.859, -85.7933],
   y = [9.96232, 10.1688, 10.3214, 10.5279, 10.5608, 10.6593, 10.8565, 10.8987, 10.9644, 11.0512] };
  WCL[162] = struct {
   x = [-85.7933, -85.7933, -85.9036, -86.0796, -86.3002, -86.5208, -86.6968, -86.8306, -87.0394, -87.2483, -87.3914, -87.537, -87.6238, -87.6355, -87.5463, -87.4032, -87.3375],
   y = [11.0512, 11.1381, 11.2695, 11.4643, 11.6051, 11.7905, 12.0604, 12.2012, 12.3655, 12.5603, 12.6987, 12.8607, 12.9381, 13.0226, 13.0132, 12.9592, 13.0226] };
  WCL[163] = struct {
   x = [-87.3375, -87.3821, -87.5017, -87.4924, -87.5463, -87.6684, -87.7787],
   y = [13.0226, 13.0789, 13.2385, 13.3699, 13.4122, 13.4004, 13.4239] };
  WCL[164] = struct {
   x = [-87.7787, -87.8772, -87.9758, -88.1964, -88.5273, -88.8488, -89.046, -89.2994, -89.5857, -89.8063],
   y = [13.4239, 13.3253, 13.2385, 13.1963, 13.1963, 13.262, 13.3371, 13.4661, 13.5201, 13.5624] };
  WCL[165] = struct {
   x = [-85.5515, -85.5844, -85.7393, -85.8167, -85.9036, -85.9481, -85.8496, -85.6501, -85.5187, -85.3966, -85.2863, -85.176, -85.0775, -84.9789, -84.9226, -84.8569],
   y = [11.2484, 11.2484, 11.3986, 11.5957, 11.8656, 12.0604, 12.159, 12.0275, 11.9313, 11.8656, 11.8116, 11.7037, 11.5957, 11.5089, 11.3235, 11.1263] };
  WCL[166] = struct {
   x = [-96.4573, -96.4009, -96.3681, -96.3258, -96.2695, -96.2249, -96.171, -96.1475, -96.1381, -96.1381, -96.1146, -96.1029, -96.07, -96.0161, -95.9597, -95.9058, -95.8401, -95.7955, -95.7391, -95.6406,
        -95.4974, -95.3754, -95.2862, -95.143, -94.9999],
   y = [20.0021, 19.8988, 19.8261, 19.7205, 19.575, 19.4717, 19.3778, 19.3145, 19.2628, 19.2206, 19.1995, 19.1995, 19.1995, 19.1572, 19.1056, 19.0328, 18.9906, 18.9272, 18.8639, 18.8005, 18.7583, 18.7066,
        18.6644, 18.655, 18.648] };
  WCL[167] = struct {
   x = [-100.046, -99.879, -99.7593, -99.6818, -99.5481, -99.426, -99.2735, -99.1068, -98.9848, -98.8299, -98.7642, -98.6656, -98.6328, -98.5553, -98.4779, -98.3559, -98.2127, -98.0578, -97.9921, -97.9475,
        -97.8701, -97.7809, -97.6706, -97.5955, -97.4734, -97.3725, -97.2528, -97.0651, -97.0534, -96.9642, -96.8656, -96.7764, -96.6568, -96.6004, -96.4573, -96.3352, -96.1921, -96.07, -95.9597,
        -95.8823, -95.7626, -95.6852, -95.5185, -95.3754, -95.2533, -95.143, -94.9999, -95.0445],
   y = [16.9442, 16.8925, 16.8174, 16.7752, 16.7235, 16.6907, 16.6484, 16.5945, 16.5428, 16.4889, 16.4889, 16.51, 16.4889, 16.4138, 16.341, 16.2659, 16.1908, 16.0946, 16.0735, 16.0641, 16.0641, 16.0007,
        15.9561, 15.935, 15.8928, 15.8928, 15.8716, 15.8623, 15.8505, 15.8388, 15.7872, 15.7449, 15.7003, 15.7003, 15.6252, 15.6158, 15.6158, 15.6252, 15.6792, 15.7449, 15.7872, 15.8294, 15.8623,
        15.9467, 16.043, 16.1157, 16.1861, 16.1486] };
  WCL[168] = struct {
   x = [-94.9999, -94.9999, -94.9224, -94.7253, -94.4812, -94.2395, -94.0635, -93.8312, -93.5894, -93.3125, -93.16, -92.9605, -92.7188, -92.4301, -92.1227, -91.9678, -91.956, -91.9021, -91.7918, -91.6158,
        -91.4491, -91.2966, -91.3294, -91.3388, -91.076, -90.8765, -90.7873, -90.7451, -90.6348],
   y = [18.648, 18.6808, 18.6268, 18.3757, 18.2067, 18.2067, 18.3124, 18.4391, 18.4813, 18.5236, 18.5024, 18.533, 18.6691, 18.7536, 18.7536, 18.7113, 18.6057, 18.5236, 18.5024, 18.5236, 18.6174, 18.6902,
        18.8592, 19.0258, 19.1314, 19.3191, 19.528, 19.7463, 19.8918] };
  WCL[169] = struct {
   x = [-89.8063, -90.0176, -90.2593],
   y = [13.5624, 13.7032, 13.8416] };
  WCL[170] = struct {
   x = [-90.2593, -90.4353, -90.7662, -90.9211, -91.0314, -91.3294, -91.6158, -91.8457, -92.0781, -92.2095],
   y = [13.8416, 13.8956, 13.9402, 13.9402, 13.9402, 13.9613, 14.1115, 14.304, 14.4542, 14.5621] };
  WCL[171] = struct {
   x = [-92.2095, -92.2447, -92.3315, -92.5521, -92.6953, -92.883, -93.0591, -93.2022, -93.3571, -93.5331, -93.7655, -93.9743, -94.1738, -94.3709, -94.5915, -94.7347, -94.9224, -94.9999],
   y = [14.5621, 14.595, 14.6701, 14.8085, 14.9705, 15.1418, 15.3014, 15.461, 15.6111, 15.7707, 15.8881, 15.9843, 16.0594, 16.1321, 16.1955, 16.2072, 16.1955, 16.1861] };
  WCL[172] = struct {
   x = [-105.356, -105.333, -105.288, -105.279, -105.112, -105.014, -104.925, -104.847, -104.749, -104.615, -104.516, -104.329, -104.197, -104.12, -104.019, -103.899, -103.744, -103.655, -103.601, -103.566,
        -103.578, -103.566, -103.5, -103.446, -103.434, -103.39, -103.291, -103.181, -103.115, -103.047, -102.937, -102.838, -102.805, -102.749, -102.695, -102.597, -102.409, -102.254, -102.111,
        -101.965, -101.888, -101.822, -101.702, -101.569, -101.515, -101.479, -101.47, -101.393, -101.336, -101.217, -101.095, -100.951, -100.94, -100.919, -100.829, -100.764, -100.707, -100.642,
        -100.552, -100.454, -100.355, -100.189, -100.046],
   y = [20.0443, 19.9082, 19.7956, 19.7533, 19.6594, 19.5444, 19.4506, 19.3661, 19.2839, 19.1783, 19.1267, 19.0727, 19.0727, 19.0328, 18.9366, 18.8334, 18.7794, 18.6972, 18.655, 18.5799, 18.5494, 18.5283,
        18.5189, 18.4766, 18.4133, 18.3288, 18.2654, 18.2443, 18.2443, 18.2349, 18.1715, 18.1598, 18.0964, 18.0871, 18.0753, 18.0448, 17.9908, 17.9486, 17.9064, 17.897, 17.9275, 17.9064, 17.843,
        17.7702, 17.6975, 17.6435, 17.5919, 17.5708, 17.5168, 17.4628, 17.3783, 17.3267, 17.2422, 17.2211, 17.1671, 17.1882, 17.1671, 17.1155, 17.0826, 17.0521, 17.0192, 16.9981, 16.9442] };
  WCL[173] = struct {
   x = [-113.15, -113.16, -113.15, -113.15, -113.15],
   y = [18.1504, 18.1387, 18.1504, 18.1293, 18.1504] };
  WCL[174] = struct {
   x = [-155.752, -155.666, -155.565, -155.478, -155.344, -155.234, -155.189, -155.135, -155.091, -155.069, -155.025, -154.98, -154.924, -154.882, -154.814, -154.781, -154.793, -154.87, -154.959, -155.123,
        -155.257, -155.377, -155.389, -155.487, -155.52, -155.609, -155.743, -155.818, -155.853, -155.886, -155.94, -155.952, -155.842, -155.83],
   y = [20.0443, 19.981, 19.981, 19.9504, 19.8988, 19.8566, 19.8261, 19.7627, 19.6899, 19.6594, 19.6383, 19.6172, 19.5444, 19.4811, 19.4294, 19.3238, 19.2839, 19.1361, 19.1572, 19.1689, 19.1361, 19.0516,
        19.0422, 18.9272, 18.8427, 18.77, 18.8427, 18.9789, 19.1267, 19.3567, 19.4928, 19.6172, 19.8049, 19.9199] };
  WCL[175] = struct {
   x = [166.489, 166.489, 166.501, 166.501, 166.489, 166.489, 166.501, 166.489, 166.489],
   y = [19.3661, 19.3567, 19.345, 19.3238, 19.345, 19.3661, 19.3238, 19.345, 19.3661] };
  WCL[176] = struct {
   x = [166.888, 166.921, 166.954, 167.02, 167.064, 167.031, 166.966, 166.921, 166.855, 166.754, 166.7, 166.7, 166.722, 166.787, 166.844, 166.888],
   y = [11.415, 11.4361, 11.4361, 11.4479, 11.4033, 11.3704, 11.3164, 11.2179, 11.1428, 11.1545, 11.2085, 11.2953, 11.361, 11.415, 11.415, 11.415] };
  WCL[177] = struct {
   x = [145.361, 145.349, 145.316, 145.293],
   y = [20.0138, 19.9716, 19.9504, 19.9927] };
  WCL[178] = struct {
   x = [145.889, 145.924, 145.933, 145.9, 145.868, 145.856, 145.856, 145.856, 145.879, 145.879, 145.889, 145.889],
   y = [18.1387, 18.1715, 18.1598, 18.0871, 18.0753, 18.0871, 18.0964, 18.1082, 18.1176, 18.1293, 18.1293, 18.1387] };
  WCL[179] = struct {
   x = [145.79, 145.823, 145.823, 145.769, 145.724, 145.713, 145.757, 145.79, 145.79],
   y = [15.0174, 15.0503, 14.9963, 14.9752, 14.9752, 15.0174, 15.0503, 15.0385, 15.0174] };
  WCL[180] = struct {
   x = [144.952, 145.006, 144.985, 144.908, 144.896, 144.884, 144.807, 144.753, 144.765, 144.819, 144.875, 144.929, 144.952],
   y = [13.5131, 13.5131, 13.4497, 13.4051, 13.3746, 13.2878, 13.2549, 13.33, 13.3746, 13.438, 13.4708, 13.5131, 13.5131] };
  WCL[181] = struct {
   x = [119.947, 120.001, 120.057, 120.045, 120.001],
   y = [11.8914, 11.8586, 11.7952, 11.7412, 11.6966] };
  WCL[182] = struct {
   x = [119.968, 120.045, 120.102, 120.177, 120.245, 120.254, 120.2, 120.111, 120.012],
   y = [12.227, 12.1636, 12.1519, 12.1636, 12.0768, 12.0322, 12.0322, 12.0228, 12.0862] };
  WCL[183] = struct {
   x = [120.376, 120.355, 120.421, 120.543, 120.698, 120.85, 120.928, 121.029, 121.193, 121.348, 121.425, 121.491, 121.514, 121.503, 121.536, 121.557, 121.491, 121.482, 121.392, 121.226, 121.148, 121.005,
        120.928, 120.817, 120.796, 120.698, 120.576, 120.519, 120.433, 120.376],
   y = [13.4169, 13.4708, 13.4802, 13.4802, 13.4708, 13.4708, 13.4802, 13.4591, 13.4169, 13.2878, 13.1798, 13.1681, 12.9639, 12.7809, 12.6823, 12.5743, 12.5086, 12.3678, 12.2927, 12.1848, 12.3256, 12.4453,
        12.6072, 12.8231, 13.039, 13.1892, 13.2549, 13.3957, 13.4286, 13.4169] };
  WCL[184] = struct {
   x = [122.254, 122.319, 122.331, 122.263, 122.188, 122.153, 122.153, 122.165, 122.254, 122.341, 122.385, 122.474, 122.474, 122.385, 122.275, 122.23, 122.132, 122.033, 122.01, 121.845, 121.712, 121.613,
        121.601, 121.524, 121.404, 121.425, 121.503, 121.592, 121.679, 121.625, 121.634, 121.712, 121.735, 121.845, 122, 122.099, 122.242, 122.209, 122.23, 122.287, 122.441, 122.629, 122.716, 122.85,
        122.972, 123.059, 123.082, 123.169, 123.279, 123.258, 123.357, 123.413, 123.577, 123.711, 123.842, 123.887, 123.732, 123.589, 123.622, 123.744, 123.854, 123.798, 123.821, 123.908, 123.908,
        123.92, 123.932, 124.03, 124.108, 124.108, 124.108, 124.021, 123.887, 123.833, 123.908, 123.965, 123.887, 123.666, 123.479, 123.345, 123.279, 123.169, 122.972, 122.871, 122.838, 122.674,
        122.486, 122.507, 122.606, 122.695, 122.695, 122.596, 122.486, 122.319, 122.176, 122.078, 121.923, 121.801, 121.592, 121.491, 121.348, 121.184, 121.094, 120.961, 120.829, 120.731, 120.663,
        120.641, 120.698, 120.829, 120.928, 120.918, 120.808, 120.63, 120.609, 120.564, 120.421, 120.322, 120.2, 120.102, 120.057],
   y = [18.4227, 18.371, 18.3077, 18.2138, 18.012, 17.8336, 17.6224, 17.4417, 17.3267, 17.2844, 17.1882, 17.0826, 16.9136, 16.6907, 16.4795, 16.2988, 16.1063, 16.0312, 16.0735, 15.9773, 15.8716, 15.7121,
        15.5525, 15.3694, 15.2967, 15.1347, 14.954, 14.7921, 14.6959, 14.6114, 14.4166, 14.2664, 14.1162, 14.02, 13.8909, 13.8792, 13.9332, 14.0083, 14.1819, 14.1491, 14.2875, 14.2664, 14.3321,
        14.2875, 14.1702, 13.9778, 13.8041, 13.729, 13.8369, 14.0294, 13.9989, 13.9543, 13.912, 13.8463, 13.7947, 13.7079, 13.6867, 13.5671, 13.4286, 13.33, 13.2972, 13.1587, 13.093, 13.093, 13.0813,
        13.0813, 13.039, 13.0179, 12.9428, 12.7903, 12.6072, 12.5086, 12.6283, 12.7691, 12.8231, 12.8653, 12.8888, 12.8771, 13.0179, 13.0601, 13.2667, 13.4169, 13.5131, 13.6116, 13.675, 13.8369,
        13.8909, 13.6633, 13.4591, 13.2667, 13.2009, 13.2127, 13.4169, 13.5459, 13.6961, 13.7947, 13.8792, 13.912, 13.8909, 13.7079, 13.621, 13.6116, 13.7079, 13.7407, 13.8463, 13.783, 13.9449,
        14.1279, 14.2664, 14.3955, 14.5363, 14.6536, 14.7381, 14.7499, 14.4917, 14.4072, 14.5668, 14.7499, 14.7592, 15.0503, 15.3061] };
  WCL[185] = struct {
   x = [119.968, 120.134, 120.355, 120.343, 120.322, 120.355, 120.433, 120.433, 120.409, 120.433, 120.475, 120.498, 120.597, 120.719, 120.829, 121.038, 121.27, 121.482, 121.667, 121.878, 122.01, 122.042,
        122.176, 122.23, 122.254],
   y = [16.2119, 16.043, 16.1063, 16.3926, 16.6367, 16.9019, 17.1577, 17.4628, 17.6764, 17.9064, 18.1176, 18.2959, 18.4649, 18.601, 18.5799, 18.5799, 18.5283, 18.4015, 18.3077, 18.2959, 18.3382, 18.5283,
        18.5377, 18.4766, 18.4227] };
  WCL[186] = struct {
   x = [121.89, 121.899, 122, 122.01, 122.021, 122.021, 121.932, 121.911, 121.89, 121.89, 122],
   y = [14.9963, 15.0714, 15.0385, 14.9212, 14.7499, 14.6841, 14.663, 14.8132, 14.9306, 14.9963, 15.0385] };
  WCL[187] = struct {
   x = [121.866, 121.822, 121.822, 121.834, 121.932, 122.021, 122.087, 122.12, 122.099, 122.078, 121.965, 121.89, 121.866],
   y = [13.3206, 13.3629, 13.384, 13.5037, 13.5037, 13.5248, 13.4708, 13.3746, 13.2667, 13.2009, 13.2009, 13.2878, 13.3206] };
  WCL[188] = struct {
   x = [121.988, 122, 122.01, 122.099, 122.099, 122.078, 122.021, 121.977, 121.988],
   y = [12.4453, 12.5086, 12.6072, 12.6072, 12.4124, 12.2059, 12.1308, 12.2387, 12.4453] };
  WCL[189] = struct {
   x = [122.441, 122.451, 122.474, 122.552, 122.629, 122.662, 122.662, 122.563, 122.451, 122.441],
   y = [12.4453, 12.4781, 12.4781, 12.4664, 12.4124, 12.3584, 12.3045, 12.3373, 12.3584, 12.4453] };
  WCL[190] = struct {
   x = [123.005, 123.038, 123.059, 123.103, 123.169, 123.258, 123.291, 123.324, 123.336, 123.303, 123.225, 123.136, 123.059, 123.005, 123.005],
   y = [13.1141, 13.1352, 13.0719, 13.0273, 12.9428, 12.8982, 12.856, 12.802, 12.7363, 12.7809, 12.7903, 12.9099, 12.9968, 13.0601, 13.1141] };
  WCL[191] = struct {
   x = [123.655, 123.678, 123.699, 123.744, 123.744, 123.723, 123.678, 123.655],
   y = [12.6283, 12.6612, 12.6189, 12.5086, 12.4241, 12.4124, 12.5204, 12.6283] };
  WCL[192] = struct {
   x = [123.258, 123.324, 123.345, 123.39, 123.479, 123.589, 123.744, 123.866, 123.986, 124.03, 124.042, 124.042, 123.986, 123.866, 123.711, 123.601, 123.523, 123.39, 123.27, 123.214, 123.246, 123.258,
        123.246, 123.258, 123.246, 123.258, 123.258],
   y = [12.5626, 12.5321, 12.4875, 12.4664, 12.4664, 12.3678, 12.2505, 12.1519, 12.044, 11.9243, 11.8492, 11.7623, 11.7835, 11.8163, 11.9243, 12.1097, 12.1636, 12.0862, 11.9689, 12.0017, 12.1636, 12.2387,
        12.3467, 12.4335, 12.5086, 12.5532, 12.5626] };
  WCL[193] = struct {
   x = [124.319, 124.319, 124.34, 124.373, 124.483, 124.528, 124.593, 124.781, 124.903, 125.002, 125.112, 125.255, 125.323, 125.476, 125.476, 125.488, 125.466, 125.488, 125.553, 125.598, 125.631, 125.732,
        125.631, 125.455, 125.255, 125.178, 125.068, 124.948, 124.936, 124.99, 124.957, 124.805, 124.671, 124.549, 124.429, 124.352, 124.319, 124.319],
   y = [12.5204, 12.5321, 12.5321, 12.5321, 12.5204, 12.5086, 12.5086, 12.4992, 12.5532, 12.5861, 12.5626, 12.4781, 12.3373, 12.2176, 12.1519, 12.0228, 11.7506, 11.5676, 11.415, 11.2953, 11.2085, 11.0559,
        11.0677, 11.0559, 11.1099, 11.2953, 11.3376, 11.4479, 11.5347, 11.6755, 11.7295, 11.8914, 12.0111, 12.0651, 12.2176, 12.3913, 12.4781, 12.5204] };
  WCL[194] = struct {
   x = [121.932, 121.977, 122.042, 122.143, 122.23, 122.376, 122.418, 122.528, 122.707, 122.826, 122.862, 123.026, 123.092, 123.103, 123.038, 122.894, 122.749, 122.606, 122.519, 122.364, 122.143, 122.01,
        121.956, 121.956, 122.01, 122.033, 122.054, 122.033, 121.932, 121.932],
   y = [11.8703, 11.8914, 11.8703, 11.8492, 11.7952, 11.6849, 11.5769, 11.5558, 11.5441, 11.5441, 11.4807, 11.523, 11.5112, 11.2953, 11.0888, 11.002, 10.8166, 10.7086, 10.664, 10.6429, 10.5561, 10.3918,
        10.5983, 10.8283, 11.0677, 11.3164, 11.5676, 11.6755, 11.7717, 11.8703] };
  WCL[195] = struct {
   x = [122.65, 122.695, 122.695, 122.683, 122.606, 122.552, 122.552, 122.585, 122.65],
   y = [10.664, 10.6429, 10.5655, 10.4904, 10.434, 10.434, 10.5561, 10.6312, 10.664] };
  WCL[196] = struct {
   x = [122.96, 123.026, 123.148, 123.214, 123.324, 123.5, 123.512, 123.479, 123.39, 123.368],
   y = [10.8283, 10.8823, 10.9574, 10.9245, 10.9151, 10.8705, 10.718, 10.4997, 10.4246, 10.2158] };
  WCL[197] = struct {
   x = [122.563, 122.761, 122.826, 122.826, 122.894, 122.927, 122.96],
   y = [9.96466, 10.0421, 10.1947, 10.4997, 10.664, 10.7837, 10.8283] };
  WCL[198] = struct {
   x = [124.406, 124.417, 124.471, 124.528, 124.549, 124.582, 124.516, 124.483, 124.439, 124.406],
   y = [11.6544, 11.7177, 11.6849, 11.6544, 11.5769, 11.5347, 11.4901, 11.4901, 11.5769, 11.6544] };
  WCL[199] = struct {
   x = [124.319, 124.319, 124.361, 124.429, 124.516, 124.605, 124.748, 124.859, 124.99, 125.002, 125.002, 125.091, 125.178, 125.246, 125.124, 125.014, 124.99, 124.859, 124.769, 124.76, 124.769, 124.671,
        124.549, 124.471, 124.417, 124.385, 124.34, 124.319, 124.319],
   y = [11.4901, 11.5018, 11.4033, 11.3376, 11.3821, 11.3282, 11.3493, 11.3493, 11.1639, 10.9574, 10.718, 10.5561, 10.3378, 10.2392, 10.1853, 10.2275, 10.0304, 10.075, 10.2604, 10.4904, 10.718, 10.9245,
        10.8823, 10.8705, 11.0771, 11.3047, 11.3821, 11.4807, 11.4901] };
  WCL[200] = struct {
   x = [125.553, 125.544, 125.544, 125.621, 125.675, 125.687],
   y = [10.1618, 10.1853, 10.305, 10.4669, 10.3801, 10.1618] };
  WCL[201] = struct {
   x = [125.654, 125.586, 125.553],
   y = [9.94354, 10.0421, 10.1618] };
  WCL[202] = struct {
   x = [123.997, 124.042, 124.054, 124.03, 124.042, 124.042, 124.021, 123.943, 123.798, 123.699],
   y = [11.2179, 11.2413, 11.1639, 11.0559, 10.8494, 10.6429, 10.3918, 10.3143, 10.2064, 10.021] };
  WCL[203] = struct {
   x = [123.368, 123.434, 123.5, 123.589, 123.688, 123.777, 123.833, 123.887, 123.932, 123.997],
   y = [9.85671, 10.0421, 10.2064, 10.3378, 10.5115, 10.718, 10.8588, 11.0114, 11.1639, 11.2179] };
  WCL[204] = struct {
   x = [123.986, 124.075, 124.206, 124.307, 124.429, 124.572, 124.572],
   y = [9.95528, 10.075, 10.1618, 10.1196, 10.075, 10.1196, 10.021] };
  WCL[205] = struct {
   x = [110.883, 110.949, 110.958, 110.782, 110.651, 110.55, 110.496, 110.451, 110.308, 110.142, 110.064],
   y = [20.0349, 19.8143, 19.6806, 19.5233, 19.2839, 19.1056, 18.9155, 18.8005, 18.6972, 18.5377, 18.4555] };
  WCL[206] = struct {
   x = [119.449, 119.503, 119.527, 119.527, 119.538, 119.571, 119.515, 119.527, 119.571, 119.604, 119.56, 119.393, 119.283, 119.13],
   y = [11.3376, 11.3821, 11.3493, 11.2179, 11.1099, 11.0442, 10.948, 10.8048, 10.664, 10.5443, 10.4246, 10.3143, 10.1524, 10.0093] };
  WCL[207] = struct {
   x = [118.51, 118.665, 118.764, 118.874, 119.02, 119.151, 119.217, 119.283, 119.273, 119.273, 119.318, 119.384, 119.36, 119.283, 119.327, 119.384, 119.405, 119.449],
   y = [9.84497, 10.0304, 10.0867, 10.2604, 10.3707, 10.4246, 10.5326, 10.664, 10.772, 10.8494, 10.7626, 10.8166, 10.8823, 10.948, 11.002, 11.1428, 11.2953, 11.3376] };
  WCL[208] = struct {
   x = [119.879, 119.89, 119.947],
   y = [11.9243, 11.9243, 11.8914] };
  WCL[209] = struct {
   x = [120.001, 119.947, 119.914, 119.879, 119.879],
   y = [11.6966, 11.7412, 11.8163, 11.882, 11.9243] };
  WCL[210] = struct {
   x = [119.914, 119.89, 119.968],
   y = [12.2927, 12.2833, 12.227] };
  WCL[211] = struct {
   x = [120.012, 119.947, 119.914],
   y = [12.0862, 12.2505, 12.2927] };
  WCL[212] = struct {
   x = [120.057, 119.989, 119.923, 119.858, 119.804, 119.813, 119.968],
   y = [15.3061, 15.4656, 15.7121, 15.8928, 16.2025, 16.2659, 16.2119] };
  WCL[213] = struct {
   x = [102.989, 102.834, 102.78, 102.69, 102.601, 102.568, 102.47, 102.336, 102.216, 102.083, 101.895, 101.707, 101.531, 101.343, 101.102, 101.001, 101.057, 100.98, 100.923, 100.935, 100.947, 100.968,
        100.914, 100.714, 100.527, 100.351, 100.24, 100.008, 100.053, 100.074, 100.029],
   y = [11.6426, 11.7177, 11.828, 12.0651, 12.1191, 12.0768, 12.1519, 12.2059, 12.3678, 12.4664, 12.6401, 12.6823, 12.6494, 12.6706, 12.6823, 12.6706, 12.6072, 12.6189, 12.748, 12.985, 13.1798, 13.384,
        13.4802, 13.5342, 13.5577, 13.5342, 13.5342, 13.4497, 13.2432, 13.039, 12.7691] };
  WCL[214] = struct {
   x = [104.49, 104.39, 104.336, 104.237, 104.092, 103.927, 103.784, 103.606, 103.606, 103.695, 103.662, 103.552, 103.463, 103.376, 103.209, 103.132, 103.054, 103.054, 103.122, 103.111, 103.033, 102.956,
        102.956, 102.956, 102.989],
   y = [10.4669, 10.4904, 10.5443, 10.5983, 10.61, 10.6312, 10.5561, 10.5655, 10.7626, 10.8823, 11.0559, 11.1217, 11.0771, 10.9245, 10.8917, 11.1099, 11.3493, 11.4033, 11.3376, 11.4244, 11.5347, 11.6098,
        11.6755, 11.6638, 11.6426] };
  WCL[215] = struct {
   x = [107.051, 107.128, 107.262, 107.426, 107.581, 107.713, 107.912, 108.022, 108.1, 108.123, 108.123, 108.133, 108.243, 108.243, 108.255, 108.255, 108.288, 108.398, 108.565, 108.684, 108.806, 108.872,
        108.961, 109.039, 109.104, 109.137, 109.17, 109.203, 109.203, 109.203, 109.238, 109.226, 109.226, 109.248, 109.248, 109.316, 109.337, 109.348, 109.348, 109.348, 109.271, 109.203, 109.215,
        109.215, 109.149, 109.128, 109.17, 109.203, 109.203, 109.161, 109.128, 109.182, 109.17, 109.116, 109.06, 109.039, 108.994, 108.827, 108.708, 108.574, 108.276, 108.145, 108.001, 107.891,
        107.746, 107.57, 107.438, 107.304, 107.173, 107.051, 106.896, 106.797, 106.741, 106.675, 106.699, 106.631],
   y = [17.031, 16.9676, 16.7963, 16.6813, 16.5428, 16.3504, 16.3082, 16.2988, 16.1908, 16.1486, 16.1157, 16.0641, 16.0735, 16.1392, 16.1392, 16.0852, 15.8928, 15.7214, 15.5736, 15.4563, 15.3272, 15.0925,
        14.8578, 14.6325, 14.4072, 14.1491, 13.9003, 13.8369, 13.8158, 13.6633, 13.5342, 13.438, 13.3206, 13.1798, 13.039, 12.9968, 12.8348, 12.6823, 12.6401, 12.7034, 12.7269, 12.6283, 12.4992,
        12.4241, 12.4781, 12.4124, 12.227, 12.0017, 11.9149, 11.9243, 11.8586, 11.7952, 11.6849, 11.6544, 11.5981, 11.4361, 11.3282, 11.3164, 11.1967, 11.1428, 11.002, 10.9363, 10.7391, 10.718, 10.664,
        10.5655, 10.4786, 10.4012, 10.4129, 10.4669, 10.434, 10.4012, 10.2815, 10.2275, 10.075, 10.0867] };
  WCL[216] = struct {
   x = [106.61, 106.499],
   y = [9.98813, 10.0093] };
  WCL[217] = struct {
   x = [104.997, 105.063, 104.953, 104.81, 104.622, 104.49],
   y = [9.86844, 10.0304, 10.1407, 10.1947, 10.2604, 10.4669] };
  WCL[218] = struct {
   x = [106.168, 105.969, 105.859, 105.76, 105.727, 105.659, 105.626, 105.727, 105.892, 106.091, 106.3, 106.401, 106.434, 106.643, 106.797, 106.962, 107.051],
   y = [20.0537, 19.8777, 19.6477, 19.4177, 19.1995, 19.0328, 18.8005, 18.5799, 18.4133, 18.2349, 18.1082, 17.8547, 17.6341, 17.4745, 17.2844, 17.1671, 17.031] };
  WCL[219] = struct {
   x = [110.064, 109.855, 109.757, 109.712, 109.614, 109.524, 109.37, 109.161, 108.839, 108.684, 108.663, 108.651, 108.651, 108.762, 108.994, 109.17, 109.194, 109.316, 109.536, 109.724, 109.933],
   y = [18.4555, 18.4015, 18.2654, 18.1927, 18.2138, 18.2748, 18.317, 18.3804, 18.4649, 18.6761, 18.9577, 19.1267, 19.3027, 19.3872, 19.6172, 19.7416, 19.8049, 19.8871, 19.981, 19.9927, 19.9927] };
  WCL[220] = struct {
   x = [104.005, 104.092, 104.103, 104.082, 103.981, 103.927, 103.927, 104.005],
   y = [10.4458, 10.3472, 10.2064, 10.1078, 10.2158, 10.3261, 10.3918, 10.4458] };
  WCL[221] = struct {
   x = [98.5718, 98.5389, 98.5624, 98.6281, 98.8041, 98.8041, 98.8487, 98.8698, 98.7924, 98.6727, 98.5389, 98.5624, 98.7173, 98.6821, 98.7173, 98.6821, 98.6398, 98.5718, 98.4286, 98.3512, 98.2737, 98.2291,
        98.1423, 98.0531, 97.9428, 97.8443, 97.7668, 97.7339, 97.7011, 97.7105, 97.7105, 97.7011, 97.7105, 97.7105, 97.5791, 97.3585, 97.1824, 97.0369, 96.9172, 96.8281, 96.7389, 96.7389, 96.6521,
        96.3986, 96.1217, 95.9222, 95.7251, 95.5467, 95.359, 95.2721, 95.1501, 95.0609, 94.9412, 94.7629, 94.5751, 94.4883, 94.554, 94.378, 94.2677, 94.2677, 94.2677, 94.2888, 94.3123, 94.3874,
        94.4437, 94.554, 94.554, 94.5, 94.4648, 94.3874, 94.378, 94.2677, 94.1456, 94.0025, 93.9368, 93.8804, 93.681, 93.5613, 93.6153, 93.8241, 93.7913, 93.6481, 93.7044, 93.681, 93.4721],
   y = [10.1735, 10.3143, 10.5443, 10.7837, 11.2179, 11.415, 11.5769, 11.7295, 11.7717, 11.8374, 11.936, 12.0651, 12.0322, 12.2176, 12.4664, 12.694, 12.9428, 13.093, 13.4169, 13.7501, 13.6867, 13.6116,
        13.8909, 14.3087, 14.5879, 14.954, 15.3905, 15.7543, 16.0641, 16.2119, 16.4138, 16.4138, 16.4255, 16.51, 16.5428, 16.6484, 16.7963, 16.9559, 16.9019, 17.2422, 17.4323, 17.0943, 16.7024,
        16.5522, 16.4255, 16.2542, 16.1275, 15.9773, 15.766, 15.8388, 15.8294, 15.9256, 15.7543, 15.8388, 15.9045, 15.9773, 16.1275, 16.0101, 16.0524, 16.0735, 16.0312, 16.1697, 16.3621, 16.6062,
        16.8925, 17.1788, 17.5379, 17.8031, 18.0659, 18.2021, 18.3288, 18.5377, 18.7277, 18.9366, 19.1267, 18.9061, 19.0117, 19.2417, 19.3872, 19.3027, 19.3989, 19.4928, 19.5961, 19.7017, 19.8049] };
  WCL[222] = struct {
   x = [98.6398, 98.5718],
   y = [9.96466, 10.1735] };
  WCL[223] = struct {
   x = [100.029, 99.9846, 99.9634, 99.8086, 99.6771, 99.5762, 99.5222, 99.5222, 99.4331, 99.3462, 99.3228, 99.2242, 99.1913, 99.1913],
   y = [12.7691, 12.4547, 12.1636, 11.9149, 11.6544, 11.3493, 11.1756, 10.9785, 10.7954, 10.718, 10.5326, 10.3707, 10.2604, 10.0093] };
  WCL[224] = struct {
   x = [98.5272, 98.5506, 98.4849, 98.3394, 98.4943, 98.5718, 98.5952, 98.5272],
   y = [11.8046, 11.7717, 11.828, 11.5887, 11.5558, 11.5769, 11.7177, 11.8046] };
  WCL[225] = struct {
   x = [98.3394, 98.3066, 98.2855, 98.2737, 98.3394, 98.3746, 98.3394],
   y = [11.7623, 11.7412, 11.5887, 11.4901, 11.5887, 11.7717, 11.7623] };
  WCL[226] = struct {
   x = [98.3746, 98.3629, 98.3629, 98.384, 98.4615, 98.3746],
   y = [12.6706, 12.6283, 12.4241, 12.3584, 12.4241, 12.6706] };
  WCL[227] = struct {
   x = [92.93, 92.9652, 92.8971, 92.8431, 92.7868, 92.7211, 92.7, 92.7094, 92.7094, 92.6671, 92.6108, 92.5568, 92.5662, 92.578, 92.6671, 92.754, 92.7211, 92.754, 92.7774, 92.8431, 92.8431, 92.8971, 92.9417,
        92.8971, 92.8971, 92.9863, 92.998, 93.0074, 93.0074, 92.93],
   y = [13.5248, 13.5671, 13.4708, 13.2878, 13.0813, 12.6823, 12.4875, 12.3373, 12.227, 12.0979, 11.9782, 11.8914, 11.7084, 11.6426, 11.5347, 11.5676, 11.6849, 11.9149, 12.0768, 12.2599, 12.3796, 12.5415,
        12.7903, 12.9099, 13.0273, 13.093, 13.2549, 13.3629, 13.5248, 13.5248] };
  WCL[228] = struct {
   x = [86.5724, 86.2744, 86.042, 85.7322, 85.4577, 85.169, 84.939, 84.6949, 84.4978, 84.2983, 84.1106, 84.066, 84.0214, 83.8219, 83.5591, 83.3267, 83.0052, 82.564, 82.3223, 82.212, 82.212, 82.1017, 81.7473,
        81.6699, 81.4281, 81.1958, 81.1629, 81.0738, 80.9423, 80.7992, 80.6208, 80.4565, 80.2782, 80.1256],
   y = [20.1992, 19.9411, 19.8472, 19.7627, 19.6172, 19.46, 19.2839, 19.0117, 18.77, 18.5283, 18.3593, 18.3499, 18.2865, 18.1715, 17.9815, 17.728, 17.5168, 17.1671, 17.0099, 16.8385, 16.6367, 16.5006,
        16.3715, 16.341, 16.3082, 16.2448, 16.2237, 16.0101, 15.7872, 15.7214, 15.8177, 15.8083, 15.7121, 15.4656] };
  WCL[229] = struct {
   x = [79.9801, 80.0153, 80.0365, 80.0693, 80.0811, 80.0928, 80.0928, 80.0482, 80.0365, 80.0693, 80.1139, 80.1468, 80.1679, 80.1139, 80.0576, 80.1139, 80.135, 80.2359, 80.2688, 80.2688, 80.2453, 80.2242,
        80.1679, 80.1022, 80.0036],
   y = [14.8883, 14.6748, 14.6325, 14.5879, 14.5879, 14.5457, 14.3744, 14.257, 14.1819, 14.02, 13.8698, 13.7173, 13.621, 13.6867, 13.621, 13.5788, 13.492, 13.4051, 13.2338, 13.1587, 13.0179, 12.9639,
        12.7574, 12.5321, 12.2716] };
  WCL[230] = struct {
   x = [73.578, 73.5123, 73.4466, 73.4113, 73.402, 73.3456, 73.2471, 73.2025, 73.1814, 73.1696, 73.1485, 73.1039, 73.1039, 73.0476, 72.9819, 72.9162, 72.9044, 72.8833, 72.8598, 72.8505, 72.8598, 72.827,
        72.7613, 72.7613, 72.7941, 72.827, 72.827, 72.8598, 72.8505, 72.8153, 72.7495, 72.6956, 72.6956, 72.6956, 72.6838, 72.6392, 72.6275, 72.5947, 72.5853, 72.5853, 72.5947],
   y = [15.7543, 15.8716, 15.935, 16.0312, 16.1275, 16.1697, 16.3621, 16.51, 16.7329, 16.8808, 17.031, 17.146, 17.3149, 17.4323, 17.6341, 17.8242, 17.8852, 17.8852, 18.0331, 18.1927, 18.2537, 18.2865,
        18.4015, 18.4978, 18.7489, 18.8427, 18.9061, 18.9695, 19.0845, 19.1478, 19.1056, 19.0422, 19.1056, 19.1995, 19.3238, 19.3989, 19.5139, 19.6172, 19.7017, 19.8871, 19.9504] };
  WCL[231] = struct {
   x = [73.6108, 73.6437, 73.6765, 73.7094, 73.754, 73.8103, 73.8995, 73.8878, 73.8314, 73.8197, 73.8643, 73.8878, 73.9206, 73.9417],
   y = [15.7449, 15.7121, 15.5736, 15.4868, 15.4563, 15.5102, 15.4985, 15.4656, 15.4023, 15.3178, 15.2098, 15.0925, 15.0057, 14.954] };
  WCL[232] = struct {
   x = [80.1256, 79.9801, 79.9708, 79.9801],
   y = [15.4656, 15.2849, 15.1465, 14.8883] };
  WCL[233] = struct {
   x = [80.0036, 79.9144, 79.7947, 79.7502, 79.7384, 79.7384, 79.7502, 79.7595, 79.783, 79.783, 79.783, 79.783, 79.783, 79.7267, 79.5953, 79.5295, 79.4732, 79.2855, 79.1869],
   y = [12.2716, 12.0862, 11.828, 11.6309, 11.4033, 11.1099, 10.9574, 10.772, 10.5866, 10.4669, 10.3472, 10.3378, 10.2721, 10.2721, 10.2721, 10.3378, 10.3378, 10.2932, 10.0961] };
  WCL[234] = struct {
   x = [76.3379, 76.1173, 75.9412, 75.8192, 75.6972, 75.6315, 75.378, 75.1903, 75.0471, 75.0354, 75.0354, 74.9696, 74.8148, 74.7373, 74.7045, 74.5942, 74.4604, 74.4064, 74.3172, 74.2304, 74.0966, 73.9417],
   y = [9.74641, 10.2604, 10.6851, 11.1428, 11.4901, 11.5676, 11.882, 12.044, 12.2833, 12.2927, 12.3796, 12.5532, 13.0508, 13.3957, 13.5882, 13.8463, 14.02, 14.2781, 14.4706, 14.663, 14.7287, 14.954] };
  WCL[235] = struct {
   x = [49.9054, 50.1378, 50.3255, 50.4804, 50.6236, 50.6564, 50.6658, 50.7996, 50.9216, 50.9756, 51.0976, 51.1516, 51.0531, 51.0085, 51.0319, 51.0859, 51.0859, 51.0859, 51.0648, 51.1751, 51.2619, 51.1188,
        50.931, 50.8324],
   y = [11.4971, 11.5394, 11.6825, 11.9313, 11.8985, 11.9642, 11.9642, 11.9102, 11.8985, 11.8773, 11.8234, 11.6591, 11.53, 11.3024, 11.1921, 11.0184, 10.7696, 10.5936, 10.4857, 10.495, 10.3988, 10.3871,
        10.3331, 10.1594] };
  WCL[236] = struct {
   x = [54.4043, 54.5146, 54.4935, 54.3597, 54.1955, 54.0077, 53.82, 53.5876, 53.4116, 53.4797, 53.5876, 53.6885, 53.7308, 53.7754, 53.8857, 54.0406, 54.2166, 54.4043],
   y = [12.5696, 12.5603, 12.5039, 12.3749, 12.342, 12.2552, 12.288, 12.3326, 12.4617, 12.6565, 12.6893, 12.6025, 12.5696, 12.5696, 12.5696, 12.6354, 12.6025, 12.5696] };
  WCL[237] = struct {
   x = [57.6899, 57.7016, 57.6899, 57.6688, 57.6688, 57.6688, 57.6782, 57.6899, 57.7345, 57.7885, 57.7228, 57.5796, 57.4482, 57.2276, 57.1173, 57.0516, 56.8967, 56.7207, 56.6197, 56.5986, 56.5446, 56.3898,
        56.2349, 56.0706, 55.8828, 55.7162, 55.5402, 55.451, 55.3642, 55.2539, 55.2422, 55.2656, 55.1319, 54.977, 54.923, 54.7916, 54.6461, 54.47, 54.294, 54.0969, 53.942, 53.6979, 53.4891, 53.2239,
        53.0033, 52.8061, 52.6184, 52.4424, 52.3086, 52.2429, 52.1983, 52.2546, 52.2758, 52.1772, 52.034, 51.9237, 51.7688, 51.6491, 51.4943, 51.2408, 51.0202, 50.7879, 50.6001, 50.3912, 50.227,
        50.0486],
   y = [20.0701, 19.9458, 19.8308, 19.7369, 19.6219, 19.5703, 19.5186, 19.4036, 19.3191, 19.1408, 19.0375, 18.9953, 18.9953, 18.953, 18.9835, 18.9014, 18.8592, 18.7324, 18.533, 18.3757, 18.1856, 17.9955,
        17.9228, 17.9322, 17.9017, 17.8712, 17.8383, 17.7444, 17.6388, 17.5426, 17.4487, 17.3103, 17.1084, 17.0028, 16.9395, 16.9512, 16.9934, 17.0239, 17.0146, 17.0028, 16.8878, 16.7916, 16.7705,
        16.686, 16.6109, 16.5264, 16.4302, 16.3129, 16.165, 15.9843, 15.9092, 15.7707, 15.6323, 15.5689, 15.59, 15.536, 15.4938, 15.461, 15.2802, 15.1934, 15.1723, 15.1089, 15.0127, 14.9587, 14.8836,
        14.8203] };
  WCL[238] = struct {
   x = [39.9948, 40.0933, 40.2154, 40.3796, 40.5674, 40.6777, 40.7012, 40.8654, 41.086, 41.2621, 41.3958, 41.4944, 41.593, 41.7267, 41.9356, 42.067, 42.1656, 42.2759, 42.3862, 42.4871, 42.6303, 42.7406,
        42.8063, 43.0386],
   y = [15.3014, 15.0972, 15.001, 15.001, 14.947, 14.8625, 14.7663, 14.724, 14.6583, 14.5527, 14.3697, 14.2289, 14.0904, 13.9285, 13.7994, 13.6914, 13.6069, 13.499, 13.3253, 13.2291, 13.1305, 12.9803,
        12.8513, 12.8067] };
  WCL[239] = struct {
   x = [43.0386, 43.0597, 43.1254, 43.224, 43.3343, 43.4118, 43.4024, 43.1818, 43.0034, 42.818, 42.7406, 42.6303, 42.6303, 42.7828, 42.9706, 43.1043, 43.224],
   y = [12.8067, 12.7973, 12.6987, 12.5274, 12.3866, 12.1801, 12.0604, 11.8985, 11.7694, 11.6825, 11.6051, 11.5394, 11.4432, 11.4854, 11.53, 11.5394, 11.4526] };
  WCL[240] = struct {
   x = [43.224, 43.2029, 43.2357, 43.3906, 43.5455, 43.6652, 43.8318, 44.0407, 44.2942, 44.6908, 44.956, 45.2423, 45.4629, 45.6272, 45.7609, 46.0144, 46.289, 46.6973, 46.9414, 47.2371, 47.4155, 47.7229,
        48.0327, 48.2087, 48.2087, 48.5513, 48.7367, 49.0019, 49.2671, 49.4431, 49.7083, 49.9054],
   y = [11.4526, 11.4971, 11.4526, 11.3563, 11.1827, 10.9855, 10.7696, 10.5514, 10.4199, 10.3331, 10.366, 10.5396, 10.6711, 10.779, 10.8565, 10.779, 10.6922, 10.6804, 10.8776, 11.1498, 11.2484, 11.1498,
        11.117, 11.1263, 10.779, 11.2695, 11.2366, 11.2484, 11.2906, 11.3775, 11.4643, 11.4971] };
  WCL[241] = struct {
   x = [42.6842, 42.6725, 42.6396, 42.6396, 42.6396, 42.5857, 42.5411, 42.5082, 42.4636, 42.419, 42.3533, 42.3087, 42.2759, 42.2665, 42.2665, 42.2336, 42.1984, 42.1445, 42.0342, 41.8699, 41.715, 41.6258,
        41.5061, 41.4381, 41.4287, 41.3841, 41.2949, 41.2292, 41.1963, 41.1518, 41.1189, 41.1072, 41.0978, 41.0649, 41.0532, 40.9875, 40.9194, 40.8772, 40.8209, 40.7434, 40.6894, 40.612, 40.5463,
        40.4571],
   y = [16.4091, 16.463, 16.5264, 16.6226, 16.6954, 16.7799, 16.855, 16.9183, 17.0028, 17.0662, 17.1084, 17.1741, 17.2164, 17.2375, 17.2586, 17.3854, 17.5215, 17.5848, 17.6693, 17.8289, 17.9017, 17.9955,
        18.1551, 18.2607, 18.2701, 18.3757, 18.5447, 18.6597, 18.6902, 18.7113, 18.7629, 18.838, 18.8592, 18.9624, 19.1619, 19.3191, 19.413, 19.4975, 19.5491, 19.6125, 19.7251, 19.7885, 19.8824,
        19.9974] };
  WCL[242] = struct {
   x = [43.4681, 43.4563, 43.4446, 43.4118, 43.4024, 43.3906, 43.346, 43.2921, 43.2146, 43.2146, 43.2146, 43.2146, 43.2146, 43.1912, 43.0926, 43.0386, 43.0034, 42.9706, 42.9612, 42.9494, 42.9494, 42.9377,
        42.9283, 42.8837, 42.8274, 42.8274, 42.8274, 42.7828, 42.7406, 42.6725, 42.6185, 42.6631, 42.7077, 42.7077, 42.6725, 42.6725, 42.6396, 42.6631, 42.7499, 42.7734, 42.7734, 42.7499, 42.7171,
        42.6842],
   y = [12.6987, 12.7644, 12.8395, 12.9052, 12.9475, 12.9592, 13.0015, 13.0883, 13.1751, 13.2831, 13.3371, 13.5201, 13.6492, 13.8111, 13.9073, 14.0364, 14.1984, 14.3274, 14.4448, 14.5293, 14.5527, 14.6583,
        14.7006, 14.7991, 14.8836, 14.9798, 15.0338, 15.0972, 15.1629, 15.2474, 15.2896, 15.3342, 15.3225, 15.3859, 15.461, 15.6205, 15.7168, 15.7496, 15.813, 15.8881, 16.0148, 16.1744, 16.3668,
        16.4091] };
  WCL[243] = struct {
   x = [50.0486, 49.9172, 49.7529, 49.598, 49.3657, 49.1897, 49.0676, 49.0019, 48.9034, 48.7602, 48.5607, 48.3964, 48.2204, 48.0209, 47.8449, 47.6689, 47.6009, 47.568, 47.4577, 47.2606, 47.0728, 46.8968,
        46.6879, 46.4885, 46.4227, 46.289, 46.1693, 45.9815, 45.7727, 45.5849, 45.484, 45.2751, 45.0663, 44.8903, 44.7142, 44.4491, 44.1839, 44.0618, 43.8858, 43.7309, 43.4681],
   y = [14.8203, 14.7991, 14.7757, 14.7546, 14.6489, 14.5832, 14.4987, 14.3697, 14.2406, 14.0904, 13.9402, 13.9402, 13.9613, 13.9496, 13.9402, 13.8322, 13.79, 13.7126, 13.6281, 13.5201, 13.4755, 13.4333,
        13.3699, 13.391, 13.4122, 13.4122, 13.391, 13.391, 13.3793, 13.2502, 13.0461, 12.9475, 12.8607, 12.7433, 12.7316, 12.7433, 12.6565, 12.6682, 12.6236, 12.7105, 12.6987] };
  WCL[244] = struct {
   x = [37.1833, 37.1833, 37.2044, 37.2278, 37.2818, 37.3475, 37.4484, 37.5681, 37.7113, 37.8216, 37.9554, 38.0657, 38.1314, 38.1525, 38.2534, 38.3074, 38.3966, 38.4177, 38.4505],
   y = [20.1523, 19.8613, 19.6735, 19.4553, 19.2042, 18.9319, 18.7841, 18.7113, 18.6597, 18.6057, 18.5752, 18.4813, 18.303, 18.2701, 18.303, 18.249, 18.249, 18.1551, 18.0918] };
  WCL[245] = struct {
   x = [38.4505, 38.5163, 38.5937, 38.7157, 38.8589, 38.9481, 39.0138, 39.1241, 39.1781, 39.2344, 39.3001, 39.4104, 39.4878, 39.6639, 39.7742, 39.807, 39.807, 39.8845, 39.9948],
   y = [18.0918, 18.0918, 17.9955, 17.7655, 17.5215, 17.268, 16.9817, 16.7376, 16.4091, 16.1228, 15.8763, 15.6956, 15.536, 15.2896, 15.1089, 15.2263, 15.3976, 15.5149, 15.3014] };
  WCL[246] = struct {
   x = [13.6116, 13.6656, 13.7548, 13.8862, 13.9965, 14.0294, 14.0622, 14.1725, 14.372, 14.5926, 14.7123, 14.8461, 14.9775, 15.0432, 15.1207, 15.0667, 15.0338, 14.8578, 14.5809],
   y = [14.4119, 14.2946, 14.1984, 14.1233, 13.9824, 13.7571, 13.5952, 13.5835, 13.5412, 13.5741, 13.5084, 13.4755, 13.5201, 13.4239, 13.262, 13.1751, 13.0343, 12.8935, 12.8184] };
  WCL[247] = struct {
   x = [13.6116, 13.5882, 13.5013, 13.215, 13.093, 13.0155, 13.0155, 13.0695, 13.1704],
   y = [14.4119, 14.3485, 14.3368, 14.3274, 14.25, 14.1233, 14.0247, 13.9191, 13.7126] };
  WCL[248] = struct {
   x = [13.1704, 13.2362, 13.2901, 13.4239, 13.6328, 13.7313, 13.7877, 13.8745, 14.0083],
   y = [13.7126, 13.6492, 13.499, 13.3159, 13.208, 13.0672, 12.8935, 12.6682, 12.4828] };
  WCL[249] = struct {
   x = [14.0083, 14.0528, 14.1397, 14.2289, 14.3063, 14.4495, 14.5692],
   y = [12.4828, 12.5157, 12.5368, 12.6447, 12.7644, 12.8067, 12.8067] };
  WCL[250] = struct {
   x = [-10.0398, -9.95058, -9.91773],
   y = [29.3871, 29.5607, 29.5983] };
  WCL[251] = struct {
   x = [-9.91773, -9.89426, -9.81916, -9.72998],
   y = [29.5983, 29.6382, 29.7532, 29.8494] };
  WCL[252] = struct {
   x = [-16.1791, -16.2448, -16.4114, -16.5992, -16.7963, -16.7541, -16.5992, -16.4114, -16.3457, -16.1697, -16.1016, -16.1251, -16.1791],
   y = [28.5539, 28.5633, 28.4483, 28.3685, 28.3216, 28.1456, 28.0095, 28.0775, 28.3005, 28.4178, 28.4953, 28.5352, 28.5539] };
  WCL[253] = struct {
   x = [-15.6065, -15.6604, -15.7379, -15.7379, -15.7285, -15.651, -15.4962, -15.3859, -15.3413, -15.3201, -15.3201, -15.3647, -15.5173, -15.6065],
   y = [28.1456, 28.0775, 27.9391, 27.8522, 27.7631, 27.7138, 27.7255, 27.7724, 27.8311, 27.8921, 28.0376, 28.0869, 28.1245, 28.1456] };
  WCL[254] = struct {
   x = [-13.8088, -13.8205, -13.9097, -13.9965, -14.0528, -14.1397, -14.196, -14.1843, -14.3157, -14.196, -13.9637, -13.8651, -13.8088, -13.8088, -13.8088],
   y = [28.6901, 28.7182, 28.7088, 28.5727, 28.4483, 28.3404, 28.2418, 28.155, 28.0775, 28.1151, 28.2043, 28.3099, 28.4366, 28.5539, 28.6901] };
  WCL[255] = struct {
   x = [-13.3793, -13.445, -13.4896, -13.5788, -13.6891, -13.7548, -13.7759, -13.6985, -13.5788, -13.5013, -13.4333, -13.3793],
   y = [29.1852, 29.1946, 29.1172, 29.0867, 29.0304, 28.9529, 28.8637, 28.8848, 28.913, 28.9623, 29.0491, 29.1852] };
  WCL[256] = struct {
   x = [-16.2025, -16.2025, -16.158, -16.1462, -16.158, -16.2237, -16.3222, -16.3997, -16.4325, -16.5006, -16.5874, -16.6438, -16.7095, -16.7423, -16.7634, -16.7752, -16.9183],
   y = [19.934, 20.0913, 20.1945, 20.2368, 20.3189, 20.4433, 20.5888, 20.6827, 20.6193, 20.5794, 20.807, 20.9103, 21.0042, 21.0652, 21.1168, 21.1168, 21.0229] };
  WCL[257] = struct {
   x = [-16.9183, -16.9301, -16.9301, -16.8972, -16.8644, -16.8409, -16.7541, -16.6438, -16.4232, -16.3129, -16.1908, -16.0477, -16.0922, -15.9819, -15.7379, -15.6722, -15.6604, -15.6182, -15.5407, -15.3976,
        -15.2098, -14.9892, -14.8672, -14.724, -14.6795, -14.5034, -14.4049, -14.196, -14.1186, -13.7548, -13.445, -13.3465, -13.2033, -13.0367, -12.9381],
   y = [21.0229, 21.0957, 21.199, 21.4172, 21.6003, 21.8279, 22.0133, 22.2691, 22.3724, 22.5766, 22.8417, 23.0459, 23.1891, 23.3416, 23.6772, 23.7875, 23.8579, 23.98, 24.0903, 24.2428, 24.4352, 24.6559,
        24.8577, 25.0689, 25.3576, 25.7565, 26.1062, 26.3456, 26.4441, 26.6014, 26.8196, 27.1153, 27.3805, 27.6457, 27.7349] };
  WCL[258] = struct {
   x = [-12.9381, -12.9052, -12.8607, -12.6846, -12.4523, -12.1003, -11.8468, -11.5605, -11.361, -11.1639, -10.8987, -10.5889, -10.3801, -10.3143, -10.2369],
   y = [27.7349, 27.8217, 27.9203, 27.9696, 27.9696, 28.047, 28.1456, 28.2324, 28.3803, 28.5727, 28.7581, 28.8755, 29.0397, 29.136, 29.1735] };
  WCL[259] = struct {
   x = [-10.2369, -10.1383, -10.0398],
   y = [29.1735, 29.251, 29.3871] };
  WCL[260] = struct {
   x = [-78.0769, -78.1097, -78.0862, -78.0534, -77.9666, -77.9337, -78.0111, -78.1308, -78.274, -78.4735, -78.6941, -78.9358, -79.0015, -78.8701, -78.7269, -78.5063, -78.3397, -78.1872, -78.0769],
   y = [26.8196, 26.829, 26.7797, 26.7797, 26.7117, 26.5732, 26.6507, 26.7117, 26.6999, 26.6319, 26.5732, 26.5826, 26.6999, 26.6999, 26.7211, 26.8009, 26.8501, 26.8501, 26.8196] };
  WCL[261] = struct {
   x = [-77.1287, -77.0959, -77.0959, -77.0959, -77.1381, -77.2038, -77.2484, -77.3705, -77.3259, -77.3047, -77.3141, -77.2602, -77.239, -77.2038, -77.1381, -77.1287],
   y = [26.3456, 26.5615, 26.5122, 26.4042, 26.2939, 26.1649, 25.9161, 26.0851, 26.125, 26.1954, 26.3151, 26.484, 26.5615, 26.5826, 26.4629, 26.3456] };
  WCL[262] = struct {
   x = [-77.3822, -77.3259, -77.3259, -77.4033, -77.5347, -77.5582, -77.3822],
   y = [25.0384, 25.0877, 25.1088, 25.1088, 25.0971, 25.0267, 25.0384] };
  WCL[263] = struct {
   x = [-78.1097, -78.0651, -78.044, -77.9431, -77.8445, -77.8328, -77.8445, -77.8656, -77.9548, -78.0862, -78.1426, -78.22, -78.3749, -78.4078, -78.3749, -78.3186, -78.2975, -78.3068, -78.2294, -78.1543,
        -78.1097],
   y = [25.0783, 25.0689, 24.9774, 24.8577, 24.7356, 24.5948, 24.5057, 24.4352, 24.3742, 24.3625, 24.4141, 24.5057, 24.5948, 24.7263, 24.8272, 24.968, 25.0971, 25.1886, 25.198, 25.2074, 25.0783] };
  WCL[264] = struct {
   x = [-77.7905, -77.7788, -77.8445, -77.922, -77.9102, -77.7905, -77.7014, -77.7905],
   y = [24.3132, 24.3226, 24.2522, 24.1513, 24.0809, 24.1419, 24.2827, 24.3132] };
  WCL[265] = struct {
   x = [-77.8328, -77.6802, -77.645, -77.6122, -77.645, -77.7788, -77.8328],
   y = [24.0105, 24.2123, 24.0598, 23.8697, 23.7875, 23.8392, 24.0105] };
  WCL[266] = struct {
   x = [-76.2463, -76.2463, -76.2229, -76.2017, -76.2017, -76.2017, -76.2135, -76.2135, -76.3332, -76.3684, -76.3003, -76.2792, -76.2463],
   y = [25.198, 25.259, 25.2074, 25.1276, 25.0384, 24.9469, 24.806, 24.6957, 24.7873, 24.907, 24.968, 25.1088, 25.198] };
  WCL[267] = struct {
   x = [-75.4414, -75.385, -75.3639, -75.3428, -75.4179, -75.5282, -75.486, -75.5517, -75.6291, -75.7488, -75.8051, -75.7277, -75.6385, -75.5728, -75.4953, -75.4414],
   y = [24.3437, 24.3531, 24.2733, 24.1724, 24.1513, 24.1818, 24.3132, 24.4235, 24.5455, 24.6864, 24.7662, 24.7568, 24.6652, 24.5643, 24.4352, 24.3437] };
  WCL[268] = struct {
   x = [-74.5261, -74.5801, -74.5918, -74.4815, -74.4275, -74.5144],
   y = [24.163, 24.0809, 24.0011, 24.0105, 24.163, 24.163] };
  WCL[269] = struct {
   x = [-75.2982, -75.2653, -75.2208, -75.1762, -75.1222, -75.0002, -74.9227, -74.9227, -74.9884, -75.1316, -75.1879, -75.209, -75.2208, -75.2982],
   y = [23.5951, 23.5434, 23.4331, 23.29, 23.1985, 23.1586, 22.9943, 22.9145, 23.0365, 23.1163, 23.2501, 23.3604, 23.4637, 23.5951] };
  WCL[270] = struct {
   x = [-74.0966, -74.0966, -74.0966, -74.0966, -74.0966, -74.1295, -74.1741, -74.2492, -74.3172, -74.4041, -74.4158, -74.3829, -74.2726, -74.1506, -74.0192, -73.9535, -73.9417, -74.0849, -74.2609, -74.3055,
        -74.1741, -74.0638, -74.0966],
   y = [22.6587, 22.7103, 22.7197, 22.7291, 22.7103, 22.7291, 22.7291, 22.7197, 22.7291, 22.7901, 22.8324, 22.8723, 22.8324, 22.7502, 22.7502, 22.6892, 22.5343, 22.3606, 22.2292, 22.3301, 22.4545, 22.5766,
        22.6587] };
  WCL[271] = struct {
   x = [-72.773, -72.827, -72.9725, -73.1485, -72.9819, -72.8622, -72.773],
   y = [22.3419, 22.4334, 22.4428, 22.3724, 22.3912, 22.3419, 22.3419] };
  WCL[272] = struct {
   x = [-72.2215, -72.0995, -72.0126, -71.8906, -71.7592, -71.6043, -71.5832, -71.7029, -71.9023, -72.0572, -72.2426, -72.3764, -72.3318, -72.2215],
   y = [21.9711, 22.0344, 21.9617, 21.9101, 21.849, 21.7552, 21.7247, 21.7458, 21.8795, 21.9312, 21.8185, 21.8068, 21.8889, 21.9711] };
  WCL[273] = struct {
   x = [-73.0476, -73.0711, -73.0711, -73.1039, -73.2353, -73.4559, -73.6554, -73.6437, -73.4559, -73.2917, -73.1485, -73.0476],
   y = [21.3961, 21.3961, 21.2295, 21.0863, 21.0042, 20.9713, 20.9619, 21.1685, 21.2506, 21.199, 21.3022, 21.3961] };
  WCL[274] = struct {
   x = [-80.0717, -79.9074, -79.7854, -79.6633, -79.499, -79.3559, -79.201, -79.0461, -78.8818, -78.7809, -78.7152, -78.5838, -78.7715, -78.7035, -78.5063, -78.3749, -78.2083, -78.098, -78.0111, -77.9431,
        -77.7788, -77.6568, -77.5347, -77.4362, -77.4033, -77.3047, -77.1944, -77.0841, -77.0067, -76.9175, -76.765, -76.6875, -76.6218, -76.5984, -76.4881, -76.2909, -76.0914, -75.9154, -75.7723,
        -75.7066, -75.7159, -75.7934, -75.7488, -75.6291, -75.4953, -75.3193, -75.0776, -74.857, -74.7913, -74.6904, -74.5261, -74.3712, -74.2609, -74.2069, -74.228, -74.3947, -74.6035, -74.9556],
   y = [22.9333, 22.8112, 22.7713, 22.6376, 22.485, 22.4217, 22.4029, 22.4545, 22.4029, 22.3419, 22.4217, 22.3207, 22.3606, 22.3301, 22.2292, 22.187, 22.0743, 21.9711, 21.8795, 21.8678, 21.8279, 21.8279,
        21.7364, 21.6941, 21.7552, 21.6848, 21.6003, 21.5393, 21.4571, 21.3961, 21.3327, 21.2717, 21.2412, 21.2717, 21.2717, 21.1685, 21.1262, 21.1379, 21.1379, 21.0558, 20.9619, 20.8587, 20.7648,
        20.7132, 20.7554, 20.7554, 20.7554, 20.6827, 20.5677, 20.4855, 20.3611, 20.3494, 20.34, 20.2673, 20.1523, 20.1312, 20.0913, 20.0068] };
  WCL[275] = struct {
   x = [-77.6568, -77.7131, -77.4033, -77.1499, -77.2273, -77.5019, -77.7788, -78.0205, -78.2294, -78.45, -78.5181, -78.5932, -78.7152, -78.9475, -79.2456, -79.4756, -79.729, -79.9285],
   y = [19.8401, 20.0068, 20.2673, 20.4949, 20.631, 20.7132, 20.7038, 20.7226, 20.8375, 20.9619, 21.1473, 21.3656, 21.5698, 21.612, 21.5698, 21.612, 21.6941, 21.6941] };
  WCL[276] = struct {
   x = [-78.0534, -78.1637, -78.1872, -78.2857, -78.1426, -78.0862, -77.9759, -77.9008, -77.7553, -77.6896, -77.7014, -77.7788, -77.922, -78.0534],
   y = [22.2902, 22.3419, 22.2902, 22.3912, 22.2386, 22.1142, 22.0438, 21.9922, 21.9194, 21.9194, 22.0133, 22.126, 22.2175, 22.2902] };
  WCL[277] = struct {
   x = [-81.4493, -81.3953, -81.3507, -81.2732, -81.2075, -81.1418, -81.0198, -80.8766, -80.7452, -80.7006, -80.6772, -80.6889, -80.6443, -80.5457, -80.4683, -80.3815, -80.2711, -80.182, -80.182, -80.1608,
        -80.1608, -80.1702, -80.2031, -80.2805, -80.3462, -80.3815, -80.3815, -80.3908, -80.5246, -80.6678, -80.8649, -81.1418, -81.2404, -81.2521, -81.2287, -81.3296, -81.4164, -81.4281, -81.4938,
        -81.6933, -81.813, -81.8905, -82.0008, -82.0571, -81.9679, -81.9796, -82.0899, -82.1322, -82.1439, -82.1885, -82.2871, -82.3199, -82.3763, -82.4748, -82.5734, -82.6837, -82.7072, -82.6626,
        -82.5851, -82.6415, -82.74, -82.7518, -82.8386, -82.8832, -82.8503, -82.7846, -82.7846, -82.8057, -82.8714, -83.0263, -83.2023, -83.3244, -83.4464, -83.4675, -83.4793, -83.4793, -83.599,
        -83.7421],
   y = [30.1075, 29.9245, 29.7626, 29.5795, 29.4058, 29.251, 29.0773, 28.8848, 28.6713, 28.5633, 28.4272, 28.223, 27.9696, 27.8429, 27.6551, 27.4392, 27.2045, 27.0168, 26.8501, 26.6413, 26.4042, 26.1954,
        26.0076, 25.8081, 25.5875, 25.4373, 25.4491, 25.4186, 25.3083, 25.2473, 25.1792, 25.1487, 25.2379, 25.3083, 25.428, 25.6767, 25.8081, 25.8363, 25.9067, 25.9278, 26.0475, 26.3151, 26.4629,
        26.5615, 26.6507, 26.6906, 26.6319, 26.8009, 26.9581, 26.9182, 26.8994, 26.9276, 26.998, 27.1059, 27.2444, 27.4392, 27.5472, 27.7349, 27.9203, 27.979, 28.0189, 27.8616, 27.8522, 28.0963,
        28.3592, 28.5938, 28.7487, 29.0186, 29.1735, 29.2134, 29.3284, 29.4551, 29.5795, 29.5701, 29.5983, 29.7227, 29.7813, 29.9644] };
  WCL[278] = struct {
   x = [-84.4157, -84.4485, -84.526, -84.6245, -84.7583, -84.824, -84.9554, -85.0657, -85.1878, -85.3427, -85.453, -85.4952],
   y = [30.0794, 29.9644, 29.9738, 29.955, 29.8682, 29.84, 29.8095, 29.8095, 29.772, 29.7438, 29.8776, 29.9644] };
  WCL[279] = struct {
   x = [-89.8627, -89.7524],
   y = [30.0019, 29.9926] };
  WCL[280] = struct {
   x = [-89.3886, -89.4097, -89.466, -89.5857, -89.6538, -89.7195, -89.6092, -89.466, -89.3323, -89.2126, -89.2337, -89.2994, -89.4215, -89.4754, -89.6186, -89.8181, -89.8955, -89.9612],
   y = [30.0207, 29.9151, 29.8494, 29.7907, 29.7227, 29.5983, 29.4645, 29.3965, 29.3472, 29.2228, 29.136, 29.0961, 29.0491, 29.1172, 29.3096, 29.3871, 29.4457, 29.4833] };
  WCL[281] = struct {
   x = [-82.3105, -82.2542, -82.1228, -81.9351, -81.759, -81.6042, -81.4727, -81.3178, -81.1864, -81.0526, -80.8884, -80.71, -80.5786, -80.3815, -80.2359, -80.0717],
   y = [23.1773, 23.1985, 23.2196, 23.1773, 23.1891, 23.1374, 23.1281, 23.1069, 23.0764, 23.0764, 23.1069, 23.0764, 22.9943, 22.9638, 23.0248, 22.9333] };
  WCL[282] = struct {
   x = [-79.9285, -80.1256, -80.3251, -80.4894, -80.6114, -80.7452, -80.8109, -80.9869, -81.1629, -81.2193, -81.3296, -81.4493, -81.6159, -81.8365, -82.0008, -82.1674, -82.1322, -81.8905, -81.7262, -81.8365,
        -82.0993, -82.3199, -82.5734, -82.7729, -82.9935, -83.1812, -83.3033, -83.4887, -83.6999, -83.9205, -84.0402, -84.0965, -84.2279, -84.4485, -84.5377, -84.5588, -84.5799, -84.6034, -84.824, -85,
        -84.8005, -84.6902, -84.4602, -84.4485, -84.4157, -84.2725, -84.0965, -83.8524, -83.6436, -83.4793, -83.2469, -83.0475, -82.8057, -82.6297, -82.4748, -82.3105],
   y = [21.6941, 21.7669, 21.8889, 22.0532, 22.0955, 22.0743, 22.0743, 22.0837, 22.1565, 22.2081, 22.0649, 22.1142, 22.187, 22.1964, 22.248, 22.3207, 22.4123, 22.4334, 22.5155, 22.6282, 22.6587, 22.6681,
        22.6986, 22.6986, 22.6071, 22.4639, 22.2996, 22.2292, 22.1964, 22.1776, 22.0532, 21.9711, 21.9406, 21.8185, 21.7458, 21.7552, 21.8068, 21.9194, 21.8584, 21.8678, 21.9312, 22.0344, 22.0133,
        22.1776, 22.5038, 22.6071, 22.7291, 22.8019, 22.8816, 22.945, 22.9755, 23.0154, 23.0764, 23.1163, 23.1985, 23.1773] };
  WCL[283] = struct {
   x = [-83.0146, -82.9395, -82.7729, -82.6837, -82.9278, -83.1578, -83.1484, -83.1155, -83.1249, -83.0146],
   y = [21.9007, 21.9194, 21.8068, 21.5393, 21.4571, 21.5299, 21.5815, 21.6848, 21.8279, 21.9007] };
  WCL[284] = struct {
   x = [-86.7977, -86.9291, -87.0605, -87.0512, -86.908, -86.7977],
   y = [20.6099, 20.6005, 20.4222, 20.3283, 20.4339, 20.6099] };
  WCL[285] = struct {
   x = [-90.0176, -89.8063, -89.6303, -89.3557, -89.1891, -88.9145, -88.7151, -88.5179, -88.3396, -88.1213, -87.7998, -87.7998, -87.5698, -87.2929, -87.0183, -86.8963, -86.8517, -86.8751, -86.9291, -87.0277,
        -87.1497, -87.3046, -87.3821, -87.4924, -87.4924],
   y = [21.3022, 21.3844, 21.4172, 21.4172, 21.4266, 21.4477, 21.5604, 21.612, 21.6636, 21.6848, 21.6214, 21.4172, 21.6003, 21.5698, 21.5393, 21.3961, 21.199, 21.0229, 20.8587, 20.7554, 20.631, 20.516,
        20.3705, 20.164, 20.0185] };
  WCL[286] = struct {
   x = [-94.9999, -95.0327, -95.0445, -95.0773, -95.0891, -95.0985, -95.1219, -95.2322, -95.2979, -95.3871, -95.4739, -95.5631, -95.6288, -95.6734, -95.7626, -95.8494, -95.9269, -96.0372, -96.1146, -96.1264,
        -96.1264, -96.0607, -95.9715, -95.8166, -95.7391, -95.7063, -95.6852, -95.7063, -95.7509, -95.8729, -95.9175, -96.0043, -96.0278, -96.0372, -96.0489, -96.07, -96.0818, -96.1029, -96.1146,
        -96.1264, -96.1264, -96.1475, -96.171, -96.1921, -96.2038, -96.2249, -96.3024, -96.3352, -96.3258, -96.3352, -96.3587, -96.4455, -96.5019, -96.5136, -96.5019, -96.4455, -96.3916, -96.4009,
        -96.4361, -96.5019, -96.5793, -96.645, -96.6896, -96.7013, -96.7342, -96.7764, -96.8116, -96.8656, -96.9219, -96.9759, -97.0651, -97.0651, -97.0651, -97.0651, -97.0322, -96.9548, -96.9431,
        -96.9994, -97.0322, -97.0651, -97.1308, -97.2411, -97.2622, -97.2974, -97.274, -97.2294, -97.2083, -97.1965, -97.1965, -97.2083, -97.2294, -97.2622, -97.3397, -97.3514, -97.3514, -97.396,
        -97.4946, -97.5274, -97.5063, -97.45, -97.3843, -97.3397, -97.3397, -97.3514, -97.3514, -97.3631, -97.4077, -97.4171, -97.4077, -97.396, -97.3631, -97.3514, -97.3514, -97.3397, -97.3397,
        -97.3303, -97.3186, -97.2857, -97.2411, -97.22, -97.1637, -97.1637, -97.1191, -97.0862, -97.0862, -97.0862],
   y = [29.2416, 29.143, 29.1148, 29.0374, 28.9975, 28.9599, 28.92, 28.852, 28.8051, 28.737, 28.6783, 28.629, 28.5798, 28.561, 28.5211, 28.4929, 28.4648, 28.4343, 28.4249, 28.4249, 28.4436, 28.4929, 28.5328,
        28.6009, 28.629, 28.6572, 28.6783, 28.6877, 28.6877, 28.6384, 28.629, 28.6009, 28.5891, 28.5798, 28.6689, 28.737, 28.737, 28.7159, 28.7159, 28.6971, 28.6572, 28.6478, 28.6384, 28.6197, 28.6478,
        28.6971, 28.7065, 28.6572, 28.6103, 28.5798, 28.629, 28.6877, 28.6572, 28.6103, 28.5422, 28.5117, 28.4742, 28.4249, 28.385, 28.3568, 28.3756, 28.3756, 28.3169, 28.2606, 28.2207, 28.1714,
        28.1526, 28.1714, 28.1315, 28.1221, 28.0846, 28.0635, 28.0259, 27.9954, 28.0048, 28.0353, 27.9672, 27.8781, 27.8194, 27.8006, 27.81, 27.8006, 27.8006, 27.7795, 27.7607, 27.6833, 27.6528,
        27.5542, 27.4955, 27.3876, 27.289, 27.2514, 27.2303, 27.2514, 27.289, 27.2702, 27.2796, 27.2303, 27.1904, 27.1904, 27.1717, 27.1529, 27.0637, 26.9745, 26.9252, 26.9159, 26.9041, 26.8361,
        26.7563, 26.6788, 26.5896, 26.5005, 26.47, 26.47, 26.4394, 26.3714, 26.2916, 26.2822, 26.1414, 26.0921, 25.9748, 25.9443, 25.9325, 25.9138, 25.9044, 25.895] };
  WCL[287] = struct {
   x = [-97.0862, -97.0862, -97.0862, -97.1425, -97.1848, -97.1965, -97.1754, -97.1425, -97.0862, -97.0534, -97.0322, -97.0322, -97.0416, -97.0862, -97.1425, -97.1965, -97.2622, -97.3303, -97.4171, -97.4946,
        -97.5509, -97.5509, -97.5509, -97.5603, -97.5838, -97.6283, -97.6941, -97.7386, -97.748, -97.7715, -97.7926, -97.7926, -97.748, -97.748, -97.748, -97.748, -97.7386, -97.7386, -97.7269,
        -97.7269, -97.6612, -97.6166, -97.6049, -97.6049, -97.6049, -97.5955, -97.5955, -97.6377, -97.6823, -97.6941, -97.6941, -97.6823, -97.6377, -97.572, -97.5392, -97.4828, -97.4406, -97.3514,
        -97.274, -97.1965, -97.1754, -97.1754, -97.22, -97.2622, -97.2857, -97.2622, -97.22, -97.1965, -97.1848, -97.1754, -97.1308, -97.0322, -96.9642, -96.8774, -96.7764, -96.6568, -96.5464,
        -96.4573],
   y = [25.895, 25.8832, 25.8645, 25.7753, 25.665, 25.5758, 25.5641, 25.5852, 25.6439, 25.7448, 25.7354, 25.6439, 25.5641, 25.4749, 25.3951, 25.3458, 25.3247, 25.3341, 25.3646, 25.374, 25.3153, 25.2754,
        25.1745, 25.0853, 24.9844, 24.9445, 24.9046, 24.8225, 24.7638, 24.7333, 24.6723, 24.6019, 24.5409, 24.4306, 24.2897, 24.1489, 24.0269, 23.9448, 23.9049, 23.8556, 23.7734, 23.6021, 23.3886,
        23.2454, 23.0623, 22.8981, 22.776, 22.5414, 22.3771, 22.2855, 22.2762, 22.194, 22.0814, 21.9359, 21.8748, 21.8138, 21.7833, 21.6683, 21.5979, 21.504, 21.443, 21.3398, 21.2788, 21.2154, 21.1544,
        21.091, 20.9784, 20.9478, 20.9361, 20.9056, 20.8329, 20.6686, 20.5536, 20.4386, 20.3447, 20.2203, 20.1382, 20.0021] };
  WCL[288] = struct {
   x = [-97.3186, -97.274, -97.2622, -97.2622, -97.2622, -97.2974, -97.3303, -97.3397, -97.396, -97.4077, -97.396, -97.3514, -97.3186],
   y = [25.2754, 25.2848, 25.2449, 25.1839, 25.1135, 25.0548, 25.0337, 25.0243, 25.0642, 25.144, 25.1839, 25.2449, 25.2754] };
  WCL[289] = struct {
   x = [-96.8656, -96.8445, -96.8328, -96.8116, -96.7553, -96.7013, -96.6661, -96.6239, -96.5464, -96.4901, -96.4244, -96.4009, -96.4127, -96.4784, -96.5793, -96.6661, -96.7553, -96.8656, -96.9313, -96.9759,
        -96.9994, -96.9642, -96.8985, -96.8774, -96.8656, -96.8656],
   y = [27.9367, 27.9367, 27.9766, 28.0142, 28.0541, 28.1034, 28.1315, 28.1315, 28.1714, 28.2395, 28.27, 28.2489, 28.1808, 28.1221, 28.0353, 27.9555, 27.8687, 27.7513, 27.6833, 27.6739, 27.7114, 27.7607,
        27.8405, 27.8781, 27.8992, 27.9367] };
  WCL[290] = struct {
   x = [-89.9612, -90.0715, -90.0715, -90.1372, -90.2264, -90.3578, -90.501, -90.5573, -90.6676, -90.8225, -90.9868, -91.0408, -91.2191, -91.3388, -91.4069, -91.5829, -91.7026, -91.8457, -91.956, -92.1344,
        -92.2095, -92.2001, -92.2869, -92.3644, -92.3644, -92.531, -92.7727, -92.9276, -93.1154, -93.3454, -93.5448, -93.8312, -94.0518, -94.2935, -94.5141, -94.6244, -94.6925, -94.791, -94.8239,
        -94.7018, -94.6807, -94.8121, -94.8121, -94.8896, -94.9788, -94.9882, -94.9999, -94.9999, -94.9999],
   y = [29.4833, 29.4645, 29.3777, 29.2908, 29.1852, 29.2322, 29.2322, 29.204, 29.2603, 29.1946, 29.2228, 29.2908, 29.2721, 29.4058, 29.5513, 29.5701, 29.6945, 29.7626, 29.8189, 29.8283, 29.7626, 29.6757,
        29.5983, 29.5889, 29.5983, 29.5889, 29.657, 29.7039, 29.7438, 29.8095, 29.8095, 29.7907, 29.7626, 29.6757, 29.5889, 29.5513, 29.5208, 29.4833, 29.5208, 29.5513, 29.6382, 29.6382, 29.8095,
        29.8095, 29.7626, 29.5983, 29.5115, 29.2416, 29.3965] };
  WCL[291] = struct {
   x = [-90.6348, -90.5667, -90.501, -90.4916, -90.4564, -90.3813, -90.271, -90.1818, -90.0176],
   y = [19.8918, 20.1124, 20.4034, 20.6521, 20.9103, 21.1262, 21.2107, 21.2717, 21.3022] };
  WCL[292] = struct {
   x = [-110.114, -109.949, -109.783, -109.717, -109.705, -109.628, -109.562, -109.496, -109.386, -109.341, -109.363, -109.44, -109.541, -109.705, -109.804, -109.893, -109.991],
   y = [24.2381, 24.0973, 23.987, 23.9448, 23.865, 23.7641, 23.6936, 23.5716, 23.4801, 23.3674, 23.236, 23.1234, 23.0107, 22.9403, 22.8371, 22.8488, 22.9708] };
  WCL[293] = struct {
   x = [-110.036, -109.959, -109.893, -109.827, -109.783, -109.75, -109.672, -109.649, -109.574, -109.461, -109.395, -109.264, -109.219, -109.175, -109.064, -109.064, -109.121, -109.175, -109.231, -109.264,
        -109.285, -109.297, -109.276, -109.187, -109.043, -108.966, -108.9, -108.844, -108.823, -108.79, -108.811, -108.844, -108.811, -108.689, -108.513, -108.382, -108.269, -108.182, -108.126,
        -108.039, -107.938, -107.839, -107.818, -107.818, -107.828, -107.783, -107.696, -107.553, -107.41, -107.264, -107.178, -107.077, -106.978, -106.889, -106.868, -106.847, -106.823, -106.769,
        -106.736, -106.736, -106.68, -106.558, -106.504, -106.448, -106.403, -106.382, -106.349, -106.272, -106.183, -106.063, -105.941, -105.852, -105.708, -105.577, -105.544, -105.565, -105.565,
        -105.509, -105.443, -105.422, -105.333, -105.312, -105.288, -105.178, -105.112, -105.068, -105.068, -105.112, -105.202, -105.288, -105.3, -105.288, -105.223, -105.145, -105.136, -105.267,
        -105.399, -105.488, -105.5, -105.455, -105.399, -105.356],
   y = [27.0543, 27.0144, 26.9557, 26.876, 26.7774, 26.7281, 26.6788, 26.6295, 26.6295, 26.6389, 26.6389, 26.5591, 26.4394, 26.3714, 26.3315, 26.2611, 26.1719, 26.0921, 26.064, 25.9748, 25.8645, 25.7448,
        25.665, 25.5547, 25.5453, 25.5547, 25.5852, 25.6251, 25.6439, 25.5852, 25.5148, 25.435, 25.3646, 25.3036, 25.2144, 25.1651, 25.144, 25.1933, 25.144, 25.0548, 24.9633, 24.9046, 24.7732, 24.6629,
        24.5925, 24.5221, 24.4517, 24.3296, 24.2287, 24.1184, 24.0574, 23.9659, 23.8861, 23.8556, 23.8345, 23.8251, 23.7946, 23.7429, 23.6631, 23.5599, 23.5106, 23.4895, 23.4191, 23.3463, 23.2665,
        23.2055, 23.1445, 23.053, 22.9802, 22.8981, 22.776, 22.6634, 22.5719, 22.5015, 22.3888, 22.3677, 22.2762, 22.1119, 21.9476, 21.8748, 21.7622, 21.7012, 21.5862, 21.4829, 21.3609, 21.2154,
        21.0605, 20.9666, 20.8845, 20.7906, 20.7601, 20.7085, 20.6686, 20.5958, 20.5231, 20.4808, 20.4292, 20.3658, 20.3048, 20.2532, 20.1476, 20.0443] };
  WCL[294] = struct {
   x = [-114.375, -114.298, -114.232, -114.209, -114.143, -114.066, -114.012, -113.967, -113.878, -113.779, -113.681, -113.526, -113.514, -113.502, -113.481, -113.437, -113.416, -113.347, -113.261, -113.16,
        -113.106, -113.073, -113.017, -112.984, -112.897, -112.862, -112.808, -112.763, -112.742, -112.709, -112.686, -112.632, -112.62, -112.599, -112.564, -112.486, -112.355, -112.212, -112.113,
        -112.09, -112.09, -112.068, -112.012, -111.935, -111.881, -111.881, -111.824, -111.747, -111.738, -111.627, -111.604, -111.681, -111.705, -111.705, -111.681, -111.615, -111.538, -111.449,
        -111.395, -111.35, -111.317, -111.294, -111.24, -111.228, -111.219, -111.195, -111.151, -111.13, -111.041, -111.008, -110.921, -110.853, -110.853, -110.853, -110.853, -110.832, -110.766,
        -110.688, -110.623, -110.623, -110.632, -110.611, -110.534, -110.477, -110.379, -110.28, -110.17, -110.114],
   y = [30.0278, 29.9128, 29.7978, 29.7109, 29.7015, 29.7015, 29.664, 29.5865, 29.4997, 29.4434, 29.3659, 29.2486, 29.1923, 29.1524, 29.0749, 28.9975, 28.9599, 28.852, 28.8731, 28.7839, 28.7652, 28.7933,
        28.6689, 28.5516, 28.4249, 28.3967, 28.4061, 28.385, 28.338, 28.3286, 28.2207, 28.1034, 27.9672, 27.8593, 27.7795, 27.6927, 27.5941, 27.4462, 27.2702, 27.1411, 27.1224, 27.113, 27.0825,
        27.0543, 26.9839, 26.944, 26.8267, 26.707, 26.6084, 26.5005, 26.5005, 26.5685, 26.6976, 26.7774, 26.7774, 26.7774, 26.6976, 26.6295, 26.5005, 26.4207, 26.3714, 26.2517, 26.2118, 26.1132,
        26.0123, 25.8832, 25.7354, 25.6955, 25.6157, 25.5453, 25.4843, 25.4444, 25.3153, 25.205, 25.144, 25.1534, 25.0642, 24.9539, 24.8225, 24.6629, 24.5315, 24.4517, 24.339, 24.2498, 24.1583,
        24.1067, 24.2076, 24.2381] };
  WCL[295] = struct {
   x = [-109.991, -110.081, -110.158, -110.28, -110.367, -110.489, -110.599, -110.742, -110.876, -110.975, -111.106, -111.252, -111.329, -111.404, -111.505, -111.517, -111.505, -111.559, -111.615, -111.693,
        -111.738, -111.759, -111.857, -111.925, -111.979, -112.024, -112.068, -112.113, -112.113, -112.045, -112.003, -111.979, -111.979, -111.979, -111.991, -112.024, -112.113, -112.2, -112.266,
        -112.366, -112.486, -112.587, -112.709, -112.784, -112.873, -112.93, -113.007, -112.995, -112.984, -113.017, -113.073, -113.15, -113.204, -113.326, -113.448, -113.568, -113.613, -113.657,
        -113.723, -113.812, -113.878, -113.911, -113.955, -114.054, -114.143, -114.22, -114.298, -114.331, -114.352, -114.396, -114.474, -114.563, -114.617, -114.662, -114.727, -114.772, -114.84,
        -114.882, -114.873, -114.695, -114.474, -114.286, -114.164, -114.044, -114.033, -114.066, -114.066, -114.012, -113.922, -113.934, -113.911, -113.946, -113.946, -113.934, -113.955, -114.033,
        -114.11, -114.164, -114.244, -114.331, -114.396, -114.453, -114.519, -114.584, -114.662, -114.718, -114.772, -114.817, -114.861, -114.939, -115.025, -115.093, -115.171, -115.281, -115.347,
        -115.523, -115.579, -115.612, -115.645, -115.657],
   y = [22.9708, 23.1938, 23.3768, 23.52, 23.6021, 23.642, 23.7429, 23.8556, 23.9565, 24.0668, 24.1489, 24.1982, 24.2498, 24.3109, 24.4, 24.4611, 24.4822, 24.4822, 24.4916, 24.5103, 24.562, 24.6723, 24.7239,
        24.7122, 24.7028, 24.792, 24.7732, 24.6934, 24.792, 24.914, 25.0642, 25.1933, 25.3458, 25.4045, 25.5758, 25.6955, 25.8832, 26.064, 26.1625, 26.1625, 26.2118, 26.2822, 26.3409, 26.4512, 26.5591,
        26.6084, 26.707, 26.7774, 26.8267, 26.9041, 26.8361, 26.768, 26.6976, 26.7187, 26.7375, 26.7962, 26.8361, 26.9041, 26.9252, 26.9745, 26.9933, 26.9933, 27.0919, 27.1411, 27.1411, 27.1529,
        27.1904, 27.2702, 27.3782, 27.4462, 27.4767, 27.4955, 27.5354, 27.5847, 27.6246, 27.6739, 27.6246, 27.6434, 27.7701, 27.7607, 27.7326, 27.7795, 27.8194, 27.7419, 27.8194, 27.8405, 27.918,
        27.9766, 27.9954, 28.0846, 28.1432, 28.1808, 28.27, 28.3075, 28.4343, 28.5117, 28.6009, 28.6689, 28.737, 28.8332, 28.8614, 28.9107, 28.9975, 29.0562, 29.0843, 29.1805, 29.2697, 29.3073,
        29.3753, 29.3941, 29.3941, 29.4809, 29.549, 29.5772, 29.6264, 29.6828, 29.7414, 29.7978, 29.9034, 29.9714] };
  WCL[296] = struct {
   x = [-112.686, -112.653, -112.632, -112.554, -112.486, -112.432, -112.42, -112.388, -112.366, -112.322, -112.277, -112.167, -112.101, -112.045, -112.012, -111.935, -111.869, -111.824, -111.78, -111.681,
        -111.615, -111.526, -111.472, -111.428, -111.416, -111.362, -111.273, -111.174, -111.085, -110.987, -110.897, -110.775, -110.677, -110.566, -110.51, -110.501, -110.501, -110.501, -110.51,
        -110.534, -110.501, -110.489, -110.456, -110.423, -110.39, -110.325, -110.247, -110.137, -110.036],
   y = [30.1521, 29.9714, 29.894, 29.7884, 29.779, 29.7203, 29.6452, 29.5185, 29.4129, 29.3354, 29.2885, 29.2298, 29.143, 29.0163, 28.9107, 28.8614, 28.8051, 28.737, 28.6197, 28.5117, 28.4436, 28.4061,
        28.385, 28.3662, 28.3662, 28.3075, 28.1808, 28.0635, 28.0048, 27.9367, 27.918, 27.9086, 27.8874, 27.8593, 27.8194, 27.7701, 27.6833, 27.5941, 27.5941, 27.5166, 27.4462, 27.3594, 27.3289,
        27.289, 27.2514, 27.1904, 27.1411, 27.0919, 27.0543] };
  WCL[297] = struct {
   x = [-118.229, -118.238, -118.238, -118.238, -118.238, -118.205, -118.173, -118.151, -118.151, -118.14, -118.14, -118.14, -118.161, -118.196, -118.229, -118.229],
   y = [28.9599, 28.9693, 28.9881, 29.0069, 29.0374, 29.0843, 29.0937, 29.0374, 28.9975, 28.9013, 28.8238, 28.8051, 28.8238, 28.8825, 28.9294, 28.9599] };
  WCL[298] = struct {
   x = [-115.082, -115.115, -115.126, -115.159, -115.192, -115.204, -115.159, -115.138, -115.115, -115.061, -115.049, -115.049, -115.049, -115.037, -115.049, -115.049, -115.061, -115.082],
   y = [27.9766, 27.9766, 28.0142, 28.0142, 28.0541, 28.0846, 28.1315, 28.2395, 28.2981, 28.3075, 28.2207, 28.1714, 28.1034, 28.0541, 28.0353, 27.9954, 27.9672, 27.9766] };
  WCL[299] = struct {
   x = [-112.984, -113.04, -113.073, -113.106, -113.15, -113.204, -113.27, -113.326, -113.371, -113.425, -113.448, -113.448, -113.383, -113.305, -113.237, -113.183, -113.082, -113.073, -113.04, -112.984,
        -112.984, -112.984],
   y = [28.9693, 28.9506, 28.9506, 28.9975, 29.0749, 29.143, 29.2017, 29.2697, 29.3354, 29.4129, 29.4716, 29.4903, 29.4528, 29.3753, 29.2979, 29.2204, 29.1618, 29.0656, 29.0374, 29.0069, 28.9787, 28.9693] };
  WCL[300] = struct {
   x = [-112.366, -112.366, -112.355, -112.343, -112.277, -112.223, -112.188, -112.155, -112.134, -112.146, -112.146, -112.146, -112.188, -112.233, -112.322, -112.376, -112.42, -112.432, -112.411, -112.376,
        -112.366],
   y = [28.9787, 29.0163, 29.0937, 29.143, 29.143, 29.1148, 29.0656, 28.9975, 28.9294, 28.8614, 28.7933, 28.7464, 28.737, 28.737, 28.7464, 28.7745, 28.8332, 28.8731, 28.9013, 28.9388, 28.9787] };
  WCL[301] = struct {
   x = [-159.618, -159.606, -159.573, -159.528, -159.451, -159.385, -159.352, -159.287, -159.287, -159.374, -159.474, -159.55, -159.618, -159.695, -159.728, -159.695, -159.662, -159.618],
   y = [22.0908, 22.0908, 22.0908, 22.0814, 22.0814, 22.1001, 22.1001, 21.9781, 21.9054, 21.8021, 21.72, 21.7317, 21.8021, 21.8232, 21.9054, 21.9781, 22.0297, 22.0908] };
  WCL[302] = struct {
   x = [-158.116, -158.116, -158.038, -157.907, -157.773, -157.684, -157.728, -157.829, -157.949, -158.059, -158.16, -158.149, -158.116],
   y = [21.3797, 21.4524, 21.4947, 21.4219, 21.3093, 21.2271, 21.206, 21.1943, 21.1755, 21.2365, 21.3797, 21.4219, 21.3797] };
  WCL[303] = struct {
   x = [-157.243, -157.276, -157.243, -157.156, -156.968, -156.846, -156.747, -156.724, -156.724, -156.846, -157.001, -157.099, -157.165, -157.233, -157.266, -157.243],
   y = [21.0816, 20.9877, 21.03, 21.0511, 21.0394, 21.0605, 21.0605, 21.0206, 20.9572, 20.915, 20.9478, 20.9666, 20.9361, 20.9267, 20.9995, 21.0816] };
  WCL[304] = struct {
   x = [-156.405, -156.381, -156.36, -156.316, -156.238, -156.095, -156.006, -156.006, -156.161, -156.238, -156.339, -156.414, -156.482, -156.548, -156.625, -156.647, -156.614, -156.527, -156.459, -156.405],
   y = [20.7906, 20.7906, 20.7906, 20.8235, 20.7601, 20.678, 20.6263, 20.5442, 20.4292, 20.4175, 20.4503, 20.5958, 20.6874, 20.7202, 20.8023, 20.8235, 20.8845, 20.8751, 20.8235, 20.7906] };
  WCL[305] = struct {
   x = [-155.874, -155.853, -155.853, -155.83, -155.752],
   y = [20.0443, 20.0232, 20.0537, 20.0537, 20.0443] };
  WCL[306] = struct {
   x = [-155.83, -155.863, -155.863, -155.874],
   y = [19.9199, 20.0349, 20.0443, 20.0443] };
  WCL[307] = struct {
   x = [-156.977, -157.001, -157.001, -156.968, -156.891, -156.825, -156.801, -156.834, -156.891, -156.935, -156.977, -156.977, -156.935, -156.977, -156.977],
   y = [20.7507, 20.7718, 20.7812, 20.7812, 20.7812, 20.7202, 20.678, 20.6263, 20.6146, 20.6475, 20.6991, 20.7507, 20.6475, 20.6991, 20.7507] };
  WCL[308] = struct {
   x = [-177.435, -177.435],
   y = [28.162, 28.162] };
  WCL[309] = struct {
   x = [153.95, 153.938, 153.905, 153.882, 153.938, 153.95, 153.873, 153.882, 153.927, 153.938],
   y = [24.3508, 24.3296, 24.3296, 24.3296, 24.3907, 24.3813, 24.339, 24.3813, 24.3907, 24.3907] };
  WCL[310] = struct {
   x = [145.316, 145.361],
   y = [20.0021, 20.0138] };
  WCL[311] = struct {
   x = [145.293, 145.304, 145.316],
   y = [19.9927, 20.0021, 20.0021] };
  WCL[312] = struct {
   x = [119.914, 120.034, 120.066, 120.078, 120.134, 120.254, 120.355, 120.409, 120.465, 120.543, 120.552, 120.597, 120.663, 120.74, 120.829, 120.951, 121.083, 121.184, 121.27, 121.348, 121.437, 121.47,
        121.458, 121.458, 121.536, 121.568, 121.58, 121.557, 121.524, 121.482, 121.491, 121.702, 121.801, 121.834, 121.845, 121.855, 121.855, 121.855, 121.845, 121.756, 121.557, 121.491, 121.658,
        121.801, 121.923, 122.021, 121.944, 121.801, 121.667, 121.568],
   y = [26.5192, 26.6295, 26.7469, 26.8572, 26.8947, 27.0238, 27.0825, 27.1623, 27.2209, 27.3876, 27.4955, 27.6129, 27.7114, 27.7795, 27.9086, 28.0353, 28.2207, 28.2794, 28.1714, 28.1315, 28.2794, 28.3662,
        28.4835, 28.6478, 28.7159, 28.8825, 28.9388, 28.9975, 29.0374, 29.1031, 29.1524, 29.2017, 29.258, 29.2017, 29.258, 29.3542, 29.4317, 29.5091, 29.5772, 29.549, 29.4622, 29.4716, 29.549, 29.664,
        29.7884, 29.8564, 29.8846, 29.8846, 29.9315, 29.9808] };
  WCL[313] = struct {
   x = [121.446, 121.503, 121.58, 121.658, 121.744, 121.801, 121.855, 121.845, 121.789, 121.78, 121.801, 121.756, 121.658, 121.601, 121.568, 121.547, 121.536, 121.503, 121.458, 121.414, 121.381, 121.327,
        121.226, 121.139, 121.017, 120.939, 120.883, 120.85, 120.817, 120.74, 120.674, 120.552, 120.364, 120.188, 120.144, 120.102, 120.078, 120.078, 120.111, 120.254, 120.364, 120.519, 120.686,
        120.841, 120.951, 121.127, 121.259, 121.336, 121.404],
   y = [25.2754, 25.2543, 25.2144, 25.144, 25.144, 25.1252, 25.0431, 24.975, 24.8436, 24.7427, 24.6113, 24.4517, 24.2592, 24.1372, 23.9753, 23.8345, 23.7124, 23.5411, 23.3768, 23.2759, 23.2055, 23.0717,
        22.9098, 22.7666, 22.6235, 22.4592, 22.2762, 22.0696, 21.9476, 21.9992, 22.2339, 22.4287, 22.5531, 22.8065, 23.0107, 23.2548, 23.4496, 23.581, 23.7523, 24.0175, 24.1583, 24.4, 24.6113, 24.8225,
        24.9938, 25.1041, 25.1651, 25.2449, 25.2637] };
  WCL[314] = struct {
   x = [123.732, 123.81, 123.866, 123.866, 123.833, 123.732, 123.711, 123.732],
   y = [24.4212, 24.4094, 24.3695, 24.3109, 24.2804, 24.2686, 24.339, 24.4212] };
  WCL[315] = struct {
   x = [124.108, 124.075, 124.108, 124.185, 124.23, 124.251, 124.218, 124.131, 124.108],
   y = [24.4, 24.4212, 24.4399, 24.4704, 24.5103, 24.4611, 24.3695, 24.3203, 24.4] };
  WCL[316] = struct {
   x = [127.795, 127.872, 127.872, 127.884, 127.973, 128.137, 128.194, 128.203, 128.137, 128.006, 127.917, 127.863, 127.863, 127.785, 127.785, 127.741, 127.675, 127.696, 127.717, 127.75, 127.773, 127.795],
   y = [26.5192, 26.5896, 26.6295, 26.707, 26.6882, 26.7868, 26.8666, 26.7469, 26.5896, 26.5286, 26.4512, 26.3902, 26.3503, 26.2728, 26.193, 26.1414, 26.2611, 26.3503, 26.4301, 26.4606, 26.4793, 26.5192] };
  WCL[317] = struct {
   x = [128.933, 128.942, 128.942, 128.954, 128.909, 128.876, 128.867, 128.876, 128.888, 128.9, 128.909, 128.909, 128.921, 128.933, 128.933],
   y = [27.7912, 27.7701, 27.7701, 27.7208, 27.702, 27.7607, 27.8405, 27.9086, 27.918, 27.8781, 27.8499, 27.81, 27.8006, 27.8006, 27.7912] };
  WCL[318] = struct {
   x = [129.395, 129.473, 129.583, 129.618, 129.595, 129.505, 129.463, 129.43, 129.353, 129.32, 129.275, 129.264, 129.297, 129.353, 129.395],
   y = [28.4155, 28.4436, 28.5117, 28.4835, 28.3756, 28.3286, 28.2794, 28.2301, 28.1808, 28.1902, 28.2395, 28.3286, 28.3662, 28.4155, 28.4155] };
  WCL[319] = struct {
   x = [125.312, 125.267, 125.267, 125.279, 125.333, 125.41, 125.434, 125.356, 125.312, 125.312],
   y = [24.8225, 24.853, 24.853, 24.8342, 24.792, 24.7638, 24.7239, 24.6817, 24.7427, 24.8225] };
  WCL[320] = struct {
   x = [109.912, 110.099, 110.386, 110.397, 110.329, 110.186, 110.285, 110.397, 110.562, 110.827, 110.994, 111.036, 111.202, 111.444, 111.644, 111.831, 111.998, 112.019, 112.075, 112.075, 112.24, 112.362,
        112.449, 112.561, 112.672, 112.836, 112.913, 112.97, 112.97, 112.946, 112.958, 113.024, 113.068, 113.113, 113.244, 113.343, 113.399, 113.455, 113.521, 113.53, 113.509, 113.509, 113.509,
        113.521, 113.542, 113.542, 113.599, 113.742, 113.873, 113.929, 113.983, 113.962, 113.983, 114.138, 114.204, 114.293, 114.272, 114.228, 114.204, 114.204, 114.204, 114.216, 114.228, 114.281,
        114.359, 114.436, 114.526, 114.57, 114.58, 114.57, 114.58, 114.648, 114.734, 114.734, 114.791, 114.913, 115.011, 115.176, 115.321, 115.331, 115.431, 115.551, 115.563, 115.64, 115.739, 115.894,
        116.049, 116.105, 116.126, 116.192, 116.314, 116.403, 116.513, 116.567, 116.624, 116.722, 116.788, 116.844, 116.922, 117.053, 117.131, 117.229, 117.307, 117.375, 117.44, 117.45, 117.461,
        117.562, 117.605, 117.628, 117.661, 117.748, 117.86, 118.004, 118.069, 118.004, 117.992, 118.025, 118.135, 118.278, 118.424, 118.555, 118.588, 118.6, 118.642, 118.686, 118.71, 118.722, 118.797,
        118.853, 118.907, 118.952, 119.062, 119.13, 119.184, 119.205, 119.163, 119.24, 119.351, 119.437, 119.47, 119.437, 119.449, 119.503, 119.548, 119.604, 119.625, 119.625, 119.616, 119.616,
        119.616, 119.637, 119.703, 119.726, 119.714, 119.682, 119.616, 119.581, 119.56, 119.703, 119.858, 119.879, 119.879, 119.914],
   y = [20.387, 20.272, 20.3987, 20.5958, 20.7296, 20.9267, 21.1426, 21.3797, 21.328, 21.4008, 21.4735, 21.5346, 21.5158, 21.5463, 21.5979, 21.7622, 21.7927, 21.7927, 21.7528, 21.7528, 21.7411, 21.8138,
        21.8138, 21.835, 21.8748, 21.957, 21.9265, 21.9781, 22.1213, 22.2034, 22.3372, 22.532, 22.6329, 22.5202, 22.4287, 22.3278, 22.255, 22.2457, 22.2762, 22.3278, 22.4498, 22.5625, 22.6329, 22.7056,
        22.8371, 22.8183, 22.7361, 22.6845, 22.593, 22.5531, 22.4709, 22.4287, 22.3466, 22.3677, 22.3771, 22.4076, 22.4381, 22.4709, 22.5108, 22.532, 22.5414, 22.5414, 22.5531, 22.5531, 22.5625,
        22.5625, 22.5531, 22.5414, 22.5531, 22.6235, 22.7244, 22.7455, 22.776, 22.7056, 22.6024, 22.6141, 22.7455, 22.8488, 22.7972, 22.715, 22.6751, 22.7056, 22.776, 22.8488, 22.8065, 22.776, 22.8488,
        22.8793, 22.8793, 22.8981, 22.9497, 22.9802, 23.0412, 23.0835, 23.1844, 23.2759, 23.3979, 23.459, 23.5716, 23.6115, 23.581, 23.642, 23.7124, 23.7124, 23.6725, 23.703, 23.8133, 23.8861, 23.7523,
        23.8744, 23.9565, 23.9565, 24.0668, 24.1794, 24.3296, 24.4212, 24.4704, 24.4916, 24.5831, 24.5526, 24.562, 24.5831, 24.5925, 24.6019, 24.6512, 24.6817, 24.7638, 24.8436, 24.9046, 24.9328,
        25.0548, 25.1839, 25.205, 25.1651, 25.2355, 25.2543, 25.374, 25.4655, 25.4538, 25.4139, 25.4655, 25.5946, 25.6955, 25.7143, 25.7636, 25.8739, 25.963, 25.9935, 26.0241, 26.0428, 26.0522,
        26.1226, 26.2611, 26.3104, 26.3808, 26.4394, 26.5404, 26.6178, 26.6483, 26.7563, 26.7774, 26.6389, 26.5497, 26.5192] };
  WCL[321] = struct {
   x = [110.496, 110.463, 110.55, 110.684, 110.883],
   y = [20.0654, 20.0748, 20.1593, 20.0654, 20.0349] };
  WCL[322] = struct {
   x = [109.933, 110.055, 110.186, 110.329, 110.419, 110.496],
   y = [19.9927, 20.0021, 20.0959, 20.0748, 20.0866, 20.0654] };
  WCL[323] = struct {
   x = [113.951, 113.906, 113.873, 113.864, 113.929, 113.951],
   y = [22.2644, 22.3278, 22.2762, 22.2339, 22.2457, 22.2644] };
  WCL[324] = struct {
   x = [107.945, 108.1, 108.198, 108.255, 108.255, 108.421, 108.497, 108.574, 108.708, 109.06, 109.083, 109.17, 109.358, 109.536, 109.656, 109.733, 109.668, 109.602, 109.59, 109.623, 109.778, 109.912],
   y = [21.4735, 21.5346, 21.5979, 21.4219, 21.5346, 21.5674, 21.7106, 21.72, 21.6683, 21.5768, 21.4947, 21.4125, 21.4219, 21.5346, 21.5862, 21.4524, 21.267, 21.091, 21.03, 20.9056, 20.6146, 20.387] };
  WCL[325] = struct {
   x = [107.945, 107.933, 107.879, 107.724, 107.57, 107.382, 107.349, 107.173, 106.997, 106.83, 106.753, 106.621, 106.553, 106.356, 106.168],
   y = [21.4735, 21.4219, 21.3797, 21.4336, 21.3492, 21.2576, 21.1027, 20.9666, 20.9056, 20.8329, 20.739, 20.6263, 20.3776, 20.2203, 20.0537] };
  WCL[326] = struct {
   x = [92.2682, 92.2236, 92.0922, 92.0264, 91.9607, 91.9373, 91.8716, 91.7378, 91.5618, 91.4397, 91.2426, 91.0103, 90.8671, 90.7662, 90.6911, 90.623, 90.4916, 90.5362, 90.5456, 90.5456, 90.4588, 90.2804,
        90.1936, 90.1161],
   y = [20.8845, 20.8634, 21.1544, 21.3914, 21.6496, 21.896, 22.2339, 22.4498, 22.6939, 22.776, 22.715, 22.6939, 22.7666, 22.9403, 22.8981, 22.7244, 22.6329, 22.4193, 22.2762, 22.2339, 22.1119, 21.9171,
        21.8021, 21.8443] };
  WCL[327] = struct {
   x = [93.4721, 92.9534, 92.9417, 92.7868, 92.6765, 92.6437, 92.5216, 92.4442, 92.3456, 92.2682],
   y = [19.8049, 20.0537, 20.2415, 20.272, 20.4175, 20.3658, 20.4714, 20.6357, 20.7601, 20.8845] };
  WCL[328] = struct {
   x = [90.6137, 90.6137, 90.623, 90.623, 90.7662, 90.8554, 90.8554, 90.7662, 90.6911, 90.6137],
   y = [22.6845, 22.6141, 22.4897, 22.3278, 22.2855, 22.4076, 22.5414, 22.715, 22.7455, 22.6845] };
  WCL[329] = struct {
   x = [88.8911, 88.7808, 88.7479, 88.6916, 88.6376, 88.593, 88.5813, 88.5038, 88.4053, 88.2621, 88.1941, 88.1401, 88.0955, 88.0838, 88.0838, 88.0838, 88.0838, 88.1072, 88.1166, 88.1166, 88.1166, 88.1166,
        88.0626, 87.9641, 87.9289, 87.842, 87.72, 87.544, 87.1239, 86.9596, 86.847, 86.9244, 86.9596, 86.7719, 86.7273, 86.5724],
   y = [21.6284, 21.5979, 21.5862, 21.6496, 21.8138, 21.7411, 21.6073, 21.6284, 21.5557, 21.5346, 21.6683, 21.6895, 21.6801, 21.7106, 21.8232, 21.8655, 21.9265, 22.0297, 22.0814, 22.0908, 22.1119, 22.1213,
        22.133, 22.018, 21.9171, 21.7716, 21.6895, 21.6284, 21.443, 21.3609, 21.091, 20.8329, 20.678, 20.4902, 20.3776, 20.1992] };
  WCL[330] = struct {
   x = [90.1161, 89.9401, 89.8509, 89.8181, 89.7078, 89.6421, 89.5857, 89.4872, 89.3769, 89.2877, 89.276, 89.2102, 89.1445, 89.0882, 89.0131, 88.8911],
   y = [21.8443, 22.0602, 21.9476, 21.8537, 21.7833, 21.72, 21.8443, 21.7622, 21.7716, 21.7833, 21.7012, 21.7317, 21.72, 21.6378, 21.5979, 21.6284] };
  WCL[331] = struct {
   x = [72.5947, 72.6275, 72.6721, 72.7284, 72.7284, 72.705, 72.6275, 72.6064, 72.5947, 72.5853, 72.5618, 72.5618, 72.4961, 72.4844, 72.4844, 72.4961, 72.5407, 72.5947, 72.4844, 72.4398, 72.4515, 72.4632,
        72.475, 72.4186, 72.4304, 72.475, 72.4961, 72.5735, 72.6956, 72.7378, 72.7941, 72.7824, 72.651, 72.5078, 72.3858, 72.2544, 72.1863, 72.1417, 72.1417, 72.1323, 72.1089, 72.076, 72.0549, 72.0549,
        72.0431, 72.0314, 71.9446, 71.8343, 71.6794, 71.5245, 71.4142, 71.3579, 71.271, 71.2382, 71.2264, 71.1161, 70.9167, 70.6867, 70.5858, 70.4426, 70.344, 70.2783, 70.2666, 70.1328],
   y = [19.9504, 20.0866, 20.1898, 20.3564, 20.5113, 20.7085, 20.915, 20.9784, 20.9877, 21.03, 21.1027, 21.1544, 21.1544, 21.206, 21.2975, 21.4008, 21.5252, 21.6284, 21.6801, 21.7317, 21.8021, 21.896,
        21.9265, 21.9476, 21.9875, 22.0908, 22.1635, 22.1635, 22.1635, 22.2128, 22.2644, 22.2949, 22.3067, 22.3067, 22.3067, 22.3067, 22.3372, 22.4076, 22.3888, 22.3067, 22.1424, 22.0391, 21.9359,
        21.8232, 21.6284, 21.4829, 21.3093, 21.1544, 21.0722, 20.9784, 20.9666, 20.9361, 20.8845, 20.8634, 20.854, 20.8235, 20.7812, 20.8023, 20.8023, 20.8023, 20.8235, 20.8845, 20.8845, 20.9877] };
  WCL[332] = struct {
   x = [69.8254, 70.0131, 70.2009, 70.2666, 70.344, 70.2877, 70.2009, 70.1328, 70.0131],
   y = [22.4287, 22.5531, 22.7056, 22.8793, 23.053, 23.1445, 23.114, 22.9802, 22.9403] };
  WCL[333] = struct {
   x = [68.0254, 68.0793, 67.9925, 67.8587, 67.7602, 67.6945, 67.5513, 67.495, 67.4621, 67.3401, 67.2533, 67.1852, 67.1406, 67.1312, 67.0538, 66.866, 66.7323, 66.6994, 66.6783, 66.6665, 66.6126, 66.4788,
        66.3568, 66.1808, 65.993, 65.8053, 65.5635, 65.4298, 65.2983, 65.1552, 65.0003],
   y = [23.7734, 23.7641, 23.7523, 23.8039, 23.8345, 23.8438, 23.8438, 23.8955, 23.9964, 24.1278, 24.3203, 24.501, 24.6113, 24.7122, 24.7122, 24.7732, 24.8436, 24.975, 25.1252, 25.1839, 25.2637, 25.3552,
        25.4139, 25.4444, 25.4045, 25.374, 25.3646, 25.374, 25.3341, 25.2754, 25.3153] };
  WCL[334] = struct {
   x = [70.1328, 69.9451, 69.7151, 69.2621, 69.173, 68.9641, 68.9195, 68.997, 69.0838, 69.1518, 69.2621, 69.3607, 69.5039, 69.6048, 69.8254],
   y = [20.9877, 21.1426, 21.3914, 21.8138, 21.8748, 22.0814, 22.3466, 22.4287, 22.3771, 22.3161, 22.3466, 22.3677, 22.3888, 22.3888, 22.4287] };
  WCL[335] = struct {
   x = [70.0131, 69.9685, 69.9451, 69.6799, 69.4264, 69.2833, 69.1284, 68.9195, 68.7529, 68.5651, 68.4337, 68.2999, 68.2342, 68.1896, 68.1357, 68.1028, 68.0911, 68.0793, 68.0254],
   y = [22.9403, 22.9192, 22.9192, 22.8488, 22.8183, 22.8183, 22.8277, 22.8582, 22.9403, 23.1234, 23.2665, 23.4285, 23.5716, 23.5716, 23.5716, 23.581, 23.6514, 23.7242, 23.7734] };
  WCL[336] = struct {
   x = [61.6162, 61.5716, 61.4378, 61.175, 60.9426, 60.6328, 60.5014, 60.3794, 60.0156],
   y = [25.1581, 25.1675, 25.1182, 25.137, 25.1886, 25.2778, 25.4092, 25.3177, 25.3388] };
  WCL[337] = struct {
   x = [65.0003, 64.8243, 64.7234, 64.5919, 64.5028, 64.2822, 64.1179, 63.8644, 63.6227, 63.4795, 63.2355, 62.9703, 62.7168, 62.5197, 62.3648, 62.2428, 62.1113, 61.9682, 61.825, 61.7265, 61.6819, 61.6162],
   y = [25.3153, 25.3388, 25.3482, 25.1487, 25.2473, 25.2473, 25.3669, 25.3388, 25.3669, 25.2379, 25.2074, 25.2473, 25.2684, 25.259, 25.1886, 25.137, 25.2074, 25.0971, 25.0877, 25.0384, 25.137, 25.1581] };
  WCL[338] = struct {
   x = [49.9383, 50.0275, 50.0157, 50.0064, 50.0721, 50.1495, 50.1589, 50.1378, 50.0603, 50.0838, 50.2035, 50.3584, 50.4804, 50.5133, 50.6236, 50.7433, 50.8536, 50.931, 50.9756, 51.0742, 51.1751, 51.2737,
        51.3628, 51.4614, 51.4708, 51.4614, 51.4708, 51.5271, 51.5388, 51.4825, 51.4168, 51.3182, 51.3276, 51.4051, 51.3511, 51.3182, 51.4825, 51.5717, 51.6585, 51.7031, 51.7266, 51.7688, 51.9683,
        52.2218, 52.3649, 52.4635, 52.6067, 52.8507, 53.0479, 53.1816, 53.3342, 53.4679, 53.6768, 53.8974, 54.0734, 54.118, 54.1837, 54.2612, 54.3715, 54.3597, 54.5146, 54.7352, 55.0004, 55.2093,
        55.4088, 55.5613, 55.7491, 55.9485, 56.0354, 56.1457, 56.202, 56.3006, 56.378, 56.3663, 56.3123, 56.3006, 56.3452, 56.3452, 56.2466, 56.2231, 56.2677, 56.2677, 56.256, 56.2795, 56.3663,
        56.5658, 56.7207, 56.8192, 56.9295, 57.0727, 57.1619, 57.0398, 57.0281, 57.0609, 57.183, 57.3261, 57.5022, 57.5913, 57.7134, 57.8894, 58.0537, 58.1968, 58.3071, 58.4409, 58.6052, 58.6591,
        58.7694, 58.814, 58.9149, 58.9572, 59.0581, 59.1449, 59.1778, 59.267, 59.3655, 59.4993, 59.5861, 59.619, 59.6847, 59.7622, 59.7199, 59.6519, 59.6307, 59.4876, 59.3444, 59.2107, 59.0792,
        58.9149, 58.7694, 58.7155, 58.638, 58.5395, 58.4174, 58.2977, 58.2297, 58.1874, 58.0983, 57.9551, 57.7673, 57.6899],
   y = [26.8102, 26.7117, 26.6624, 26.5333, 26.4441, 26.3456, 26.1766, 26.1156, 26.0475, 25.848, 25.6767, 25.5476, 25.4186, 25.198, 25.1675, 25.4679, 25.5875, 25.7682, 25.9959, 26.125, 26.1954, 26.0757,
        25.9466, 25.8175, 25.6673, 25.5078, 25.3669, 25.2989, 25.0079, 24.8858, 24.7263, 24.6253, 24.5549, 24.5244, 24.4634, 24.3038, 24.3343, 24.2733, 24.3132, 24.1325, 24.1208, 24.0715, 24.0316,
        24.0105, 24.0316, 24.1325, 24.2217, 24.2123, 24.2029, 24.1419, 24.1419, 24.0903, 24.102, 24.1114, 24.2123, 24.2827, 24.3038, 24.2827, 24.3531, 24.5643, 24.7356, 24.8858, 25.0971, 25.3083,
        25.4984, 25.6368, 25.7377, 25.9372, 26.1367, 26.254, 26.254, 26.3151, 26.2752, 26.1954, 26.1555, 26.0851, 26.0475, 25.8879, 25.7377, 25.6673, 25.5969, 25.4186, 25.1675, 24.907, 24.6957, 24.515,
        24.3625, 24.2217, 24.1912, 24.1114, 24.041, 24.1208, 24.1208, 24.0809, 24.0199, 23.9495, 23.8884, 23.8579, 23.818, 23.7875, 23.7875, 23.7171, 23.6561, 23.6655, 23.6045, 23.5434, 23.4331,
        23.351, 23.2595, 23.1374, 23.0459, 22.9943, 22.8324, 22.7596, 22.6775, 22.647, 22.647, 22.5977, 22.546, 22.4639, 22.3301, 22.1776, 22.0438, 21.7974, 21.612, 21.4172, 21.314, 21.199, 20.9924,
        20.8375, 20.7648, 20.5677, 20.4222, 20.4433, 20.4949, 20.6193, 20.5794, 20.4034, 20.1523, 20.0701] };
  WCL[339] = struct {
   x = [50.5555, 50.4804, 50.4241, 50.4241, 50.457, 50.5461, 50.5555],
   y = [26.2353, 26.254, 26.2846, 26.1062, 25.9677, 26.0076, 26.2353] };
  WCL[340] = struct {
   x = [60.0156, 59.7528, 59.4219, 59.2107, 59.1238, 59.091, 58.8704, 58.6263, 58.3071, 57.9316, 57.7345, 57.4153, 57.2604, 57.1173, 57.0516, 57.0281, 57.0164, 56.9835, 56.8967, 56.6197, 56.4437, 56.3006,
        56.1692, 55.9251, 55.7279, 55.5731, 55.4628, 55.3079, 55.1319, 54.8338, 54.6249, 54.3597, 54.2401, 53.9631, 53.7097, 53.5876, 53.4116, 53.4797, 53.4679, 53.3342, 53.0267, 52.7498, 52.5973,
        52.5973, 52.4306, 52.1655, 51.9237, 51.6257, 51.4614, 51.3276, 51.3065, 51.2619, 51.1751, 51.0085, 50.9427, 50.755, 50.579, 50.5015, 50.4241, 50.2692, 50.1706, 50.0721, 50.0486],
   y = [25.3388, 25.3482, 25.4092, 25.4772, 25.4491, 25.3787, 25.5078, 25.5782, 25.5875, 25.6274, 25.6885, 25.7072, 25.9278, 26.2353, 26.5122, 26.6413, 26.6413, 26.7117, 26.9088, 27.0661, 27.1247, 27.1341,
        27.0848, 26.998, 26.9487, 26.8994, 26.7703, 26.6999, 26.6906, 26.5333, 26.4934, 26.5521, 26.6507, 26.6999, 26.6624, 26.6624, 26.8384, 26.8102, 26.8877, 26.9581, 27.0567, 27.1341, 27.3312,
        27.4016, 27.5589, 27.6856, 27.7537, 27.7724, 27.8522, 27.9977, 28.0095, 28.0963, 28.3685, 28.6525, 28.8848, 29.0186, 29.1547, 29.3871, 29.542, 29.6171, 29.7532, 29.8682, 29.9832] };
  WCL[341] = struct {
   x = [56.1692, 56.0917, 55.9251, 55.7608, 55.6834, 55.4745, 55.2539, 55.275, 55.5402, 55.7608, 56.0354, 56.1692],
   y = [26.9276, 26.9487, 26.9276, 26.9182, 26.7797, 26.6624, 26.5826, 26.5122, 26.5732, 26.6319, 26.829, 26.9276] };
  WCL[342] = struct {
   x = [40.4571, 40.3468, 40.2811, 40.1825, 40.0839],
   y = [19.9974, 20.0913, 20.1218, 20.2039, 20.2673] };
  WCL[343] = struct {
   x = [48.495, 48.5067, 48.5396, 48.5842, 48.6499, 48.7273, 48.8799, 48.9808, 49.1122, 49.2343, 49.2671, 49.2014, 49.2554, 49.2882, 49.3891, 49.4431, 49.5206, 49.6755, 49.8069, 49.9383],
   y = [28.3005, 28.3005, 28.2043, 28.0869, 27.9884, 27.871, 27.8217, 27.6363, 27.5589, 27.5002, 27.4509, 27.411, 27.343, 27.1834, 27.1341, 27.1458, 27.0848, 26.998, 26.9276, 26.8102] };
  WCL[344] = struct {
   x = [48.2627, 48.2744, 48.2955, 48.319, 48.4058, 48.4739, 48.495],
   y = [28.8074, 28.7487, 28.6713, 28.5727, 28.4953, 28.3685, 28.3005] };
  WCL[345] = struct {
   x = [48.373, 48.373],
   y = [29.955, 29.9245] };
  WCL[346] = struct {
   x = [48.0655, 47.967, 48.0092, 48.2087, 48.2744, 48.1524, 47.9106, 47.7464, 47.69, 47.7675, 47.8895, 48.0421, 48.0984, 48.143, 48.2204],
   y = [30.0794, 29.9926, 29.9245, 29.8095, 29.6945, 29.5983, 29.5513, 29.4645, 29.3378, 29.3378, 29.3378, 29.2815, 29.1172, 28.9529, 28.8168] };
  WCL[347] = struct {
   x = [48.5513, 48.4739],
   y = [30.0982, 29.9738] };
  WCL[348] = struct {
   x = [49.6637, 49.5323, 49.4877],
   y = [30.07, 29.9456, 29.9926] };
  WCL[349] = struct {
   x = [32.3652, 32.3863, 32.3863, 32.3535, 32.3652, 32.4192, 32.5295, 32.6304, 32.6961, 32.7947, 32.9167, 33.0927, 33.2922, 33.4119, 33.4565, 33.4776, 33.4776, 33.4893, 33.5222, 33.656, 33.7874, 33.8531,
        33.9094, 33.9517, 34.0854, 34.2169, 34.3272, 34.4938, 34.5924, 34.7027, 34.813, 34.9233, 35.0453, 35.1204, 35.2424, 35.3738, 35.5616, 35.7165, 35.7282, 35.7282, 35.7165, 35.7165, 35.6625,
        35.4841, 35.4184, 35.4748, 35.4748, 35.4959, 35.5522],
   y = [30.1639, 29.9832, 29.8869, 29.7227, 29.5889, 29.434, 29.1735, 28.9529, 28.798, 28.6314, 28.4366, 28.2535, 28.0564, 27.9602, 27.9508, 27.8804, 27.7724, 27.6269, 27.519, 27.343, 27.2233, 26.998,
        26.8009, 26.5826, 26.3244, 26.125, 25.8668, 25.6673, 25.4679, 25.2074, 25.0173, 24.8765, 24.6253, 24.4845, 24.3437, 24.2334, 24.0809, 23.9588, 23.9283, 23.9096, 23.879, 23.8697, 23.9096,
        23.9495, 23.8579, 23.7265, 23.5434, 23.3205, 23.1891] };
  WCL[350] = struct {
   x = [36.8312, 36.8524, 36.8524, 36.8969, 37.0401, 37.1598, 37.2044, 37.1504, 37.1598, 37.1833, 37.1833],
   y = [22.0016, 21.8678, 21.7247, 21.5181, 21.3445, 21.2201, 21.1074, 21.1168, 20.7765, 20.4222, 20.1523] };
  WCL[351] = struct {
   x = [35.5522, 35.5733, 35.6625, 35.7939, 36.0591, 36.2563, 36.4675, 36.6317, 36.8312],
   y = [23.1891, 23.1069, 22.9755, 22.7901, 22.6986, 22.5859, 22.3606, 22.2785, 22.0016] };
  WCL[352] = struct {
   x = [34.8904, 34.935, 34.9115, 34.8012, 34.7566, 34.7355, 34.7355, 34.6792, 34.6463, 34.5924, 34.5478, 34.482, 34.3929, 34.3835, 34.3929, 34.36, 34.2943, 34.1394, 33.9423, 33.8202, 33.6325, 33.4776,
        33.379, 33.2241, 33.1913, 33.1045, 32.8932, 32.8275, 32.7055, 32.6726, 32.6633, 32.6515, 32.6304, 32.5741, 32.553],
   y = [29.4927, 29.3871, 29.4457, 29.3683, 29.2721, 29.251, 29.2322, 29.1078, 28.9811, 28.8262, 28.6525, 28.4859, 28.3099, 28.1644, 28.0283, 27.9297, 27.8804, 27.7936, 27.8429, 27.9696, 28.1738, 28.331,
        28.4366, 28.622, 28.8755, 29.0397, 29.251, 29.3777, 29.5208, 29.657, 29.7532, 29.772, 29.84, 29.9151, 29.9926] };
  WCL[353] = struct {
   x = [37.9108, 37.8779, 37.8545, 37.8216, 37.8005, 37.7348, 37.6573, 37.547, 37.4813, 37.4039, 37.371, 37.2818, 37.1598, 37.1598, 37.1833, 37.1269, 37.073, 37.0072, 36.9063, 36.7866, 36.6317, 36.5778,
        36.5426, 36.4229, 36.3337, 36.2563, 36.1694, 36.0591, 35.9254, 35.7728, 35.6954, 35.6719, 35.6179, 35.5405, 35.517, 35.4748, 35.409, 35.3199, 35.2542, 35.1978, 35.1556, 35.0547, 34.8998,
        34.7472, 34.7144, 34.6252, 34.5806, 34.6041, 34.7027, 34.7895, 34.8012, 34.8012, 34.8341, 34.8575, 34.9115, 34.8904],
   y = [24.1724, 24.2123, 24.2123, 24.2217, 24.2733, 24.3038, 24.3437, 24.3437, 24.3226, 24.3531, 24.4751, 24.6066, 24.7356, 24.8976, 24.9281, 25.0783, 25.259, 25.4186, 25.5476, 25.7565, 25.8175, 25.8879,
        26.0076, 26.1461, 26.2447, 26.423, 26.6413, 26.7703, 26.8595, 27.0262, 27.1341, 27.2233, 27.282, 27.3711, 27.4298, 27.519, 27.6363, 27.7724, 27.8921, 27.9696, 28.0564, 28.1245, 28.1245,
        28.1456, 28.1644, 28.1456, 28.0658, 28.1362, 28.3216, 28.5352, 28.7393, 28.8848, 29.0186, 29.0585, 29.1852, 29.4927] };
  WCL[354] = struct {
   x = [40.0839, 39.9736, 39.9384, 39.9384, 39.8962, 39.7859, 39.685, 39.5324, 39.3987, 39.3212, 39.2555, 39.2015, 39.1452, 39.1241, 39.1006, 39.0912, 39.0795, 39.0678, 39.0584, 39.0466, 39.0232, 38.9903,
        38.9575, 38.9481, 38.9903, 39.0349, 39.0584, 39.0584, 39.0232, 38.9809, 38.9246, 38.8472, 38.7815, 38.7486, 38.7369, 38.6711, 38.5937, 38.4834, 38.3637, 38.2206, 38.0868, 37.9554, 37.9108],
   y = [20.2673, 20.3189, 20.34, 20.3283, 20.3189, 20.3611, 20.4222, 20.5066, 20.6099, 20.7038, 20.8164, 20.9314, 21.1074, 21.2412, 21.375, 21.4477, 21.5087, 21.5393, 21.6214, 21.6941, 21.8068, 21.849,
        21.9194, 22.0133, 22.0532, 22.1776, 22.2996, 22.4029, 22.5038, 22.647, 22.7596, 22.8417, 22.9849, 23.0976, 23.1586, 23.2595, 23.4425, 23.635, 23.818, 23.9283, 24.0598, 24.1419, 24.1724] };
  WCL[355] = struct {
   x = [0.072752, -0.105608, -0.171319, -0.215909, -0.225296, -0.225296, -0.159585, -0.126729, -0.082139],
   y = [39.8164, 39.6287, 39.4503, 39.3893, 39.3635, 39.3306, 39.2532, 39.1077, 39.0138] };
  WCL[356] = struct {
   x = [0.225296, -0.004694, -0.183053, -0.302742, -0.391921, -0.478754, -0.523344, -0.612524, -0.657114, -0.621911, -0.788537, -0.997405, -1.19689, -1.36116, -1.57003, -1.62636, -1.69207, -1.76951,
        -1.86808, -2.01124, -2.23184, -2.50877, -2.73876, -3.02742, -3.32312, -3.55545, -3.87462, -4.05298, -4.21726, -4.42613, -4.65847, -4.86734, -5.02223, -5.0762, -5.16538, -5.20997, -5.24283,
        -5.30854, -5.40711, -5.61832, -5.83893, -5.95861, -6.03606, -6.05718, -6.19095, -6.20268, -6.27778, -6.43267, -6.57583, -6.7096, -6.84102, -6.98418, -7.23764],
   y = [38.6524, 38.5045, 38.4177, 38.3144, 38.183, 38.0797, 37.9389, 37.8263, 37.7043, 37.6174, 37.5893, 37.5822, 37.547, 37.432, 37.2842, 37.1504, 37.0096, 36.8946, 36.7984, 36.7796, 36.8148, 36.7256,
        36.735, 36.789, 36.7538, 36.7796, 36.7538, 36.7256, 36.7092, 36.5308, 36.5144, 36.4417, 36.3102, 36.2211, 36.1483, 36.1389, 36.1131, 36.0849, 36.0943, 36.1295, 36.2563, 36.3971, 36.505,
        36.6294, 36.7186, 36.8688, 36.9932, 37.0988, 37.1598, 37.2114, 37.2396, 37.2396, 37.2208] };
  WCL[357] = struct {
   x = [-8.72553, -8.77012, -8.84757, -8.88042, -8.99072, -9.13388, -9.21133, -9.21133, -9.21133, -9.25592, -9.28877, -9.27704, -9.12449, -9.00246, -8.94614, -8.81471, -8.65043, -8.54952, -8.56125, -8.61523,
        -8.60584, -8.60584, -8.65043, -8.61523, -8.62697, -8.62697, -8.62697, -8.61523, -8.65043, -8.749, -8.76074, -8.70441, -8.50728, -8.28668, -8.05434, -7.88771, -7.69058, -7.52395, -7.39253,
        -7.2705, -7.23764],
   y = [40.0863, 39.9431, 39.7906, 39.6873, 39.5184, 39.3729, 39.2696, 39.1663, 39.0138, 38.9011, 38.8073, 38.7627, 38.6852, 38.5468, 38.4623, 38.5045, 38.5303, 38.5397, 38.4787, 38.4177, 38.3496, 38.1924,
        38.0187, 37.9225, 37.8005, 37.7817, 37.7301, 37.5986, 37.4672, 37.2842, 37.1246, 37.0706, 37.1152, 37.1152, 37.1058, 37.0636, 37.0636, 37.1058, 37.1786, 37.2044, 37.2208] };
  WCL[358] = struct {
   x = [-1.94553, -1.8141, -1.57003, -1.28372, -1.14056, -0.964549, -0.755681, -0.579668, -0.469367, -0.347332, -0.258152, -0.061018],
   y = [35.0969, 35.1063, 35.1791, 35.3128, 35.4583, 35.6649, 35.7728, 35.7916, 35.7728, 35.8456, 35.916, 35.8456] };
  WCL[359] = struct {
   x = [-9.72998, -9.61968, -9.60794, -9.60794, -9.63141, -9.69712, -9.80742, -9.76283, -9.76283, -9.7511, -9.71824, -9.61968, -9.50938, -9.39907, -9.39907, -9.16674, -9.16674, -9.21133, -9.19021, -9.12449,
        -9.03532, -9.06817, -8.90389, -8.65982, -8.49554, -8.31953, -8.02148, -7.73517, -7.4911, -7.22825, -7.00765, -6.80817, -6.60869, -6.47727, -6.27778, -6.16748, -6.08065, -5.98208, -5.86005,
        -5.7826, -5.68403, -5.51741, -5.29681, -5.2311, -5.13253, -5.01049, -4.87907, -4.81336, -4.635, -4.48245, -4.29471, -4.07411, -3.8535, -3.72208, -3.56719, -3.37944, -3.24802, -3.0814, -2.94997,
        -2.80681, -2.66366, -2.47591, -2.27643, -1.94553],
   y = [29.8494, 30.0207, 30.0982, 30.2319, 30.3845, 30.4995, 30.6708, 30.8421, 31.0604, 31.2669, 31.3913, 31.5415, 31.6917, 31.8536, 31.9099, 32.0273, 32.0273, 31.9991, 32.046, 32.1681, 32.3558, 32.5342,
        32.7008, 32.9331, 33.1467, 33.2852, 33.4049, 33.4987, 33.5809, 33.7006, 33.8109, 33.9305, 34.123, 34.3788, 34.597, 34.8528, 35.0969, 35.3222, 35.5663, 35.754, 35.8174, 35.8526, 35.916, 35.8268,
        35.6555, 35.5381, 35.4302, 35.395, 35.3058, 35.2049, 35.1697, 35.2236, 35.287, 35.2589, 35.2682, 35.2589, 35.2682, 35.2776, 35.3222, 35.2589, 35.1697, 35.1227, 35.1415, 35.0969] };
  WCL[360] = struct {
   x = [-16.7752, -16.7963, -16.8737, -17.0075, -17.0404, -16.9512, -16.7541, -16.6086, -16.632, -16.7752],
   y = [32.8041, 32.8134, 32.7853, 32.8134, 32.7571, 32.6445, 32.6069, 32.6445, 32.7665, 32.8041] };
  WCL[361] = struct {
   x = [-16.1908, -16.2683, -16.2894, -16.3011, -16.2354, -16.2237],
   y = [33.1185, 33.0998, 33.0528, 32.9895, 32.9895, 33.1091] };
  WCL[362] = struct {
   x = [-28.5727, -28.6596, -28.6267, -28.561, -28.5727],
   y = [38.6172, 38.582, 38.5139, 38.5397, 38.6172] };
  WCL[363] = struct {
   x = [-28.0423, -27.9555, -28.0541, -28.209, -28.3404, -28.4061, -28.3967, -28.2864, -28.1644, -28.0423],
   y = [38.4013, 38.4271, 38.4787, 38.521, 38.5468, 38.521, 38.4529, 38.4083, 38.3919, 38.4013] };
  WCL[364] = struct {
   x = [-27.7232, -27.6786, -27.6786, -27.7349, -27.8452, -27.9766, -28.0658, -28.108, -28.0423, -27.8992, -27.7795, -27.7232],
   y = [38.5726, 38.5984, 38.6078, 38.6266, 38.6336, 38.6782, 38.7204, 38.7204, 38.6594, 38.582, 38.5397, 38.5726] };
  WCL[365] = struct {
   x = [-27.052, -27.0942, -27.228, -27.2608, -27.1834, -27.0614, -26.9839, -27.052],
   y = [38.7721, 38.7979, 38.7627, 38.704, 38.6594, 38.6336, 38.6852, 38.7721] };
  WCL[366] = struct {
   x = [-25.0548, -25.0877, -25.2308, -25.5735, -25.7049, -25.6955, -25.5289, -25.3317, -25.1205, -25.0548],
   y = [37.8357, 37.8521, 37.8521, 37.8427, 37.8615, 37.8005, 37.7395, 37.7043, 37.7395, 37.8357] };
  WCL[367] = struct {
   x = [-24.9328, -24.9234, -24.9445, -25.0337, -24.9891, -24.9328],
   y = [36.9744, 37.0096, 37.019, 36.9744, 36.9392, 36.9744] };
  WCL[368] = struct {
   x = [-31.0533, -31.0651, -31.119, -31.1636, -31.119, -31.0533],
   y = [39.4409, 39.509, 39.5184, 39.4151, 39.3729, 39.4409] };
  WCL[369] = struct {
   x = [-64.6131, -64.6037, -64.6131, -64.6248, -64.6694, -64.714, -64.7351, -64.7351, -64.714, -64.6483, -64.6248, -64.6131],
   y = [32.3652, 32.4309, 32.4121, 32.3746, 32.3464, 32.3089, 32.2713, 32.2526, 32.2526, 32.2713, 32.3183, 32.3652] };
  WCL[370] = struct {
   x = [-74.0192, -74.0403],
   y = [40.0018, 39.9854] };
  WCL[371] = struct {
   x = [-74.0732, -74.106, -74.1389, -74.1741, -74.1835, -74.1952, -74.2069, -74.2609, -74.3266, -74.3384, -74.3712, -74.3947, -74.4604, -74.559, -74.6035, -74.6904, -74.7584, -74.857, -74.8781, -74.8781,
        -74.8687, -74.911, -75.0002, -75.0659, -75.1433, -75.2982, -75.3639, -75.4179, -75.4625, -75.486, -75.4953, -75.5071, -75.5188, -75.5188, -75.4742, -75.3968, -75.3639, -75.3076, -75.1973,
        -75.1222, -75.0776, -75.0213, -75.0541, -75.0893, -75.0987, -75.1105, -75.1973, -75.2536, -75.2865, -75.2982, -75.2982, -75.3522, -75.4179, -75.4625, -75.5188, -75.5728, -75.5963, -75.6385,
        -75.662, -75.6831, -75.7394, -75.8262, -75.8826, -75.9272, -75.96, -75.9811, -75.9929, -75.9929, -75.96, -75.892, -75.838, -75.7934, -75.7605, -75.7394, -75.6948, -75.6831, -75.662, -75.7394,
        -75.8051, -75.7934, -75.7934, -75.838, -75.8497, -75.9154, -76.0023, -76.1032, -76.1571, -76.2135, -76.2229, -76.2463, -76.2581, -76.2346, -76.19, -76.1126, -76.1032, -76.136, -76.1243,
        -76.1243, -76.1571, -76.2017, -76.2792, -76.2909, -76.2135, -76.1571, -76.1243, -76.136, -76.1571, -76.1571, -76.1689, -76.1243, -76.0703, -75.9811, -75.9365, -75.9154, -75.9929, -76.014,
        -76.0468, -76.1126, -76.19, -76.2346, -76.3003, -76.3566, -76.4012, -76.4435, -76.5326, -76.5209, -76.4552, -76.4552, -76.4552, -76.4998, -76.5115, -76.5115, -76.5115, -76.4998, -76.4552,
        -76.4435, -76.4881, -76.5655, -76.4998, -76.4341, -76.4012, -76.3684, -76.3566, -76.3449, -76.3895, -76.4435, -76.4998, -76.5444, -76.6218, -76.6429, -76.6969, -76.7532, -76.7978, -76.819,
        -76.8964, -76.9504, -76.8635, -76.7321, -76.6312, -76.5538, -76.4435, -76.4106, -76.3684, -76.3003, -76.2909, -76.312, -76.3449, -76.4223, -76.4106, -76.3895, -76.4012, -76.312, -76.2581,
        -76.2463, -76.2909, -76.3895, -76.3895, -76.3777, -76.3449, -76.2674, -76.2346, -76.2346, -76.2909, -76.3895, -76.4881, -76.5115, -76.4669, -76.4012, -76.3332, -76.3003, -76.2674, -76.2581,
        -76.1806, -76.1243, -76.0375, -75.9929, -75.9694, -75.96, -75.9154, -75.9365, -75.9365, -75.9365, -75.9365, -75.9365, -75.9272, -75.9154, -75.9154, -75.9154, -75.8826, -75.8262, -75.7934,
        -75.7934, -75.8051, -75.8497, -75.96, -76.0703, -76.0703, -76.1478, -76.1806, -76.2909, -76.3777, -76.4669, -76.5538, -76.6101, -76.6218, -76.6547, -76.6758, -76.6547, -76.4998, -76.3777,
        -76.1571, -76.0703, -76.0586, -76.014, -75.9037, -75.8169, -75.7488, -75.7159, -75.7277, -75.8051, -75.9365, -76.0797, -76.2017, -76.2909, -76.3332, -76.4106, -76.4223, -76.4341, -76.5209,
        -76.5772, -76.7978, -76.6641, -76.5209, -76.4552, -76.5326, -76.5326, -76.6218, -76.6312, -76.6429, -76.6547, -76.6641, -76.6429, -76.5326, -76.3777, -76.4012, -76.5326, -76.6875, -76.8401,
        -77.0513, -77.3259, -77.4244, -77.5465, -77.5582, -77.6028, -77.7131, -77.8117, -77.8328, -78.0111, -78.1637, -78.3068, -78.5063, -78.6378, -78.8044, -78.8584, -79.025, -79.0578, -79.1118,
        -79.1447, -79.2221, -79.1775, -79.2456, -79.3324, -79.4333, -79.5202, -79.5436, -79.6305, -79.6422, -79.6751, -79.7619, -79.9285],
   y = [40.0018, 39.976, 39.9009, 39.8, 39.7319, 39.6967, 39.6803, 39.6287, 39.6122, 39.5184, 39.4925, 39.4503, 39.3893, 39.3376, 39.2626, 39.1663, 39.0654, 38.988, 39.0138, 39.1499, 39.2367, 39.2438,
        39.2438, 39.2532, 39.3212, 39.3564, 39.4081, 39.4573, 39.5184, 39.5442, 39.57, 39.6193, 39.6028, 39.4667, 39.3376, 39.1921, 39.0983, 38.9692, 38.8753, 38.7885, 38.6524, 38.5655, 38.4951,
        38.3919, 38.3754, 38.3567, 38.2886, 38.2088, 38.1666, 38.1572, 38.0704, 38.0281, 37.8967, 37.8169, 37.7559, 37.6597, 37.5822, 37.5024, 37.4156, 37.3522, 37.2396, 37.1786, 37.1786, 37.2044,
        37.2748, 37.3522, 37.4156, 37.493, 37.5564, 37.6597, 37.7488, 37.8521, 37.8779, 37.8779, 37.8873, 37.9061, 37.9835, 37.9929, 38.0352, 38.1056, 38.1478, 38.1924, 38.2698, 38.3215, 38.305,
        38.305, 38.3754, 38.4013, 38.4271, 38.4881, 38.521, 38.582, 38.643, 38.6688, 38.7204, 38.7979, 38.8753, 38.9105, 38.9176, 38.9176, 38.9105, 38.9692, 39.0138, 39.0654, 39.1335, 39.1405, 39.1851,
        39.2273, 39.2626, 39.3564, 39.3893, 39.4151, 39.4925, 39.5254, 39.5864, 39.5864, 39.5442, 39.4409, 39.4151, 39.4081, 39.3564, 39.3048, 39.2696, 39.2626, 39.2438, 39.2109, 39.1499, 39.117,
        39.0138, 38.9434, 38.7979, 38.6852, 38.6266, 38.5303, 38.4529, 38.4435, 38.4623, 38.4693, 38.4177, 38.3754, 38.2698, 38.1666, 38.1478, 38.1149, 38.1666, 38.176, 38.2276, 38.2792, 38.2792,
        38.2792, 38.2886, 38.3496, 38.3661, 38.3496, 38.3661, 38.2792, 38.2018, 38.1572, 38.1478, 38.1149, 38.0797, 38.0539, 37.9929, 37.9483, 37.8521, 37.7488, 37.6691, 37.6691, 37.6245, 37.5822,
        37.5282, 37.5024, 37.4062, 37.3522, 37.371, 37.3452, 37.3194, 37.2748, 37.1856, 37.1786, 37.1246, 37.0988, 37.08, 37.0636, 37.0706, 37.0096, 36.965, 36.9486, 36.9204, 36.9204, 36.9204, 36.9744,
        36.958, 36.9298, 36.9298, 36.8852, 36.7984, 36.6646, 36.62, 36.5496, 36.5402, 36.4862, 36.4792, 36.451, 36.4417, 36.4417, 36.4065, 36.3525, 36.2563, 36.2023, 36.1295, 36.0849, 36.1295, 36.2023,
        36.1835, 36.2375, 36.1671, 36.1741, 36.1483, 36.1131, 36.0591, 36.0497, 36.0685, 36.1741, 36.3008, 36.2727, 36.0685, 35.9347, 35.916, 35.9347, 35.9699, 35.9699, 35.7986, 35.7094, 35.8808,
        35.9606, 35.8901, 35.7188, 35.6203, 35.5921, 35.4748, 35.3598, 35.3598, 35.3504, 35.4302, 35.4137, 35.4935, 35.5381, 35.4489, 35.4396, 35.4302, 35.3762, 35.3504, 35.2776, 35.2236, 35.1861,
        35.0969, 35.0875, 35.0781, 35.0781, 34.9608, 34.9984, 35.0523, 35.0054, 34.8153, 34.7519, 34.6698, 34.6698, 34.6604, 34.5337, 34.5055, 34.4234, 34.4234, 34.3506, 34.2779, 34.1511, 34.0854,
        34.0033, 33.9399, 33.9305, 33.893, 33.8109, 33.7475, 33.7287, 33.5621, 33.5081, 33.4142, 33.3955, 33.3509, 33.2664, 33.1749, 33.0998, 33.0716, 33.0364, 32.9519, 32.8956, 32.8862, 32.8956,
        32.8134, 32.682] };
  WCL[372] = struct {
   x = [-79.9285, -80.0505, -80.1256, -80.2359, -80.3134, -80.4354, -80.534, -80.5997, -80.6349, -80.71, -80.7452, -80.7217, -80.6678, -80.6232, -80.6772, -80.7335, -80.8109, -80.9423, -80.9306, -80.8978,
        -80.8978, -81.0081, -81.0198, -81.0855, -81.1418, -81.1864, -81.2732, -81.3718, -81.3953, -81.4281, -81.461, -81.461, -81.5267, -81.5924, -81.6042, -81.6159, -81.5713, -81.5502, -81.5056,
        -81.4493],
   y = [32.682, 32.6257, 32.5717, 32.56, 32.5717, 32.5788, 32.5975, 32.5975, 32.5975, 32.5975, 32.56, 32.4779, 32.4121, 32.2995, 32.2338, 32.2526, 32.215, 32.1587, 32.1024, 32.046, 32.0366, 32.0507, 32.0226,
        31.9569, 31.8348, 31.7785, 31.6072, 31.41, 31.3443, 31.2481, 31.1542, 31.1261, 31.0604, 30.9172, 30.7647, 30.6614, 30.584, 30.4032, 30.2507, 30.1075] };
  WCL[373] = struct {
   x = [-83.7421, -83.897, -84.1059, -84.2514, -84.3265, -84.4157],
   y = [29.9644, 30.0489, 30.1263, 30.1357, 30.1263, 30.0794] };
  WCL[374] = struct {
   x = [-85.4952, -85.5515, -85.6501, -85.805, -85.9481, -86.1476, -86.3565, -86.4011, -86.333, -86.3236, -86.4996, -86.7203, -86.962, -87.1497, -87.0957, -87.0723, -87.138, -87.1943, -87.2811, -87.4266,
        -87.6566, -87.7998, -87.9993, -88.0321, -87.9664, -88.011, -88.0204, -88.065, -88.1307, -88.2081, -88.2199, -88.3067, -88.4827, -88.6493, -88.8488, -89.0342, -89.2666, -89.3769, -89.52,
        -89.5857, -89.6538, -89.797, -89.8627],
   y = [29.9644, 30.0301, 30.0794, 30.1639, 30.2413, 30.2976, 30.3845, 30.4314, 30.4408, 30.5182, 30.5464, 30.5182, 30.4713, 30.4408, 30.5089, 30.5933, 30.5746, 30.537, 30.537, 30.4032, 30.3469, 30.2976,
        30.2601, 30.3188, 30.4807, 30.6614, 30.8609, 30.9454, 30.7271, 30.537, 30.422, 30.3939, 30.4032, 30.4032, 30.422, 30.4408, 30.3375, 30.3751, 30.2976, 30.2789, 30.2601, 30.2883, 30.2976] };
  WCL[375] = struct {
   x = [-90.1372, -89.973, -89.8744, -89.8838, -89.8627],
   y = [30.1075, 30.1944, 30.185, 30.0888, 30.0019] };
  WCL[376] = struct {
   x = [-89.7524, -89.6092, -89.4332, -89.3651, -89.3886],
   y = [29.9926, 30.0982, 30.1639, 30.1169, 30.0207] };
  WCL[377] = struct {
   x = [-89.8627, -90.0058, -90.1818, -90.4024, -90.4799, -90.325, -90.1372],
   y = [30.2976, 30.3188, 30.4314, 30.3751, 30.1944, 30.0794, 30.1075] };
  WCL[378] = struct {
   x = [-117.058, -117.058, -117.114, -117.135, -117.135, -117.156, -117.224, -117.29, -117.377, -117.487, -117.6, -117.72, -117.863, -118.008, -118.107, -118.205, -118.316, -118.327, -118.36, -118.482,
        -118.569, -118.67, -118.682, -118.703, -118.78, -118.902, -118.989, -119.078, -119.177, -119.222, -119.278, -119.365, -119.442, -119.585, -119.696, -119.764, -119.895],
   y = [32.4497, 32.5248, 32.6562, 32.7008, 32.8604, 32.9707, 33.0833, 33.1749, 33.2312, 33.3415, 33.4541, 33.5363, 33.609, 33.7029, 33.7287, 33.7287, 33.7475, 33.839, 33.9305, 34.0315, 34.0784, 34.0596,
        34.015, 33.9681, 34.0408, 34.0972, 34.1159, 34.1418, 34.1887, 34.2708, 34.299, 34.3248, 34.3624, 34.4163, 34.4257, 34.3717, 34.4351] };
  WCL[379] = struct {
   x = [-114.861, -114.828, -114.739, -114.718, -114.727, -114.751, -114.751, -114.727, -114.706, -114.662, -114.563, -114.542, -114.53, -114.53, -114.486, -114.453, -114.441, -114.375],
   y = [31.9639, 31.87, 31.7761, 31.6635, 31.6072, 31.5579, 31.4077, 31.2857, 31.1143, 30.9923, 30.8773, 30.7342, 30.5816, 30.4291, 30.2859, 30.1052, 30.0958, 30.0278] };
  WCL[380] = struct {
   x = [-115.657, -115.678, -115.711, -115.755, -115.809, -115.877, -115.898, -115.898, -115.964, -116.02, -116.086, -116.119, -116.164, -116.164, -116.208, -116.262, -116.351, -116.429, -116.518, -116.527,
        -116.527, -116.539, -116.593, -116.661, -116.682, -116.739, -116.793, -116.816, -116.849, -116.926, -116.926, -116.936, -117.058],
   y = [29.9714, 30.0864, 30.239, 30.3727, 30.3915, 30.3446, 30.3821, 30.5347, 30.6872, 30.8022, 30.8491, 30.9242, 30.943, 31.0205, 31.1331, 31.2199, 31.3232, 31.4265, 31.5297, 31.5861, 31.6917, 31.7855,
        31.8606, 31.87, 31.9357, 32.0014, 32.039, 32.1047, 32.1798, 32.2737, 32.3206, 32.4028, 32.4497] };
  WCL[381] = struct {
   x = [-114.861, -114.882, -114.882, -114.873, -114.828, -114.739, -114.673, -114.596, -114.53, -114.462, -114.364, -114.343, -114.319, -114.298, -114.22, -114.122, -114.012, -113.901, -113.824, -113.735,
        -113.645, -113.592, -113.58, -113.502, -113.371, -113.261, -113.16, -113.073, -113.007, -113.007, -113.007, -112.995, -112.963, -112.897, -112.819, -112.796, -112.751, -112.686],
   y = [31.9639, 32.0014, 31.9921, 31.9451, 31.8794, 31.8325, 31.7949, 31.748, 31.7292, 31.6729, 31.6072, 31.6259, 31.6259, 31.5579, 31.5203, 31.4734, 31.4452, 31.4546, 31.4734, 31.4828, 31.4359, 31.3983,
        31.3044, 31.2669, 31.2481, 31.2106, 31.1894, 31.1237, 31.0392, 30.9735, 30.8491, 30.7248, 30.6309, 30.5441, 30.4103, 30.2953, 30.2577, 30.1521] };
  WCL[382] = struct {
   x = [-120.071, -119.994, -119.961, -119.916, -119.895, -119.928],
   y = [33.9963, 34.0244, 33.9963, 33.9587, 33.9141, 33.8953] };
  WCL[383] = struct {
   x = [-119.785, -119.818, -119.806, -119.74, -119.686, -119.63, -119.531, -119.454, -119.454, -119.52, -119.585, -119.653, -119.74, -119.773, -119.785],
   y = [34.069, 34.0972, 34.0878, 34.0408, 34.0056, 33.9869, 34.0056, 34.015, 33.9775, 33.9681, 33.9399, 33.9141, 33.9587, 34.015, 34.069] };
  WCL[384] = struct {
   x = [-118.393, -118.438, -118.417, -118.393, -118.349, -118.306, -118.271, -118.25, -118.262, -118.306, -118.306, -118.327, -118.372, -118.393],
   y = [33.426, 33.4166, 33.4166, 33.4166, 33.3696, 33.3696, 33.3603, 33.2969, 33.2781, 33.2781, 33.3063, 33.3509, 33.379, 33.426] };
  WCL[385] = struct {
   x = [-118.503, -118.482, -118.482, -118.438, -118.384, -118.316, -118.271, -118.238, -118.295, -118.339, -118.384, -118.426, -118.482, -118.503, -118.503],
   y = [33.0176, 33.0364, 33.0082, 32.9519, 32.8792, 32.8604, 32.8134, 32.7853, 32.7665, 32.7665, 32.8134, 32.8885, 32.9519, 32.9895, 33.0176] };
  WCL[386] = struct {
   x = [-119.895, -120.017, -120.116, -120.193, -120.292, -120.425, -120.491, -120.512, -120.536, -120.536, -120.548, -120.548, -120.548, -120.548, -120.512, -120.512, -120.512, -120.512, -120.524, -120.524,
        -120.512, -120.536, -120.557, -120.613, -120.691, -120.768, -120.778, -120.778, -120.778, -120.778, -120.867, -120.944, -121.033, -121.087, -121.198, -121.308, -121.418, -121.54, -121.683,
        -121.827, -121.85, -121.794, -121.728, -121.74, -121.838, -121.993, -122.08, -122.181, -122.279, -122.303, -122.345, -122.423, -122.446, -122.402, -122.312, -122.279, -122.157, -122.014,
        -122.071, -122.226, -122.268, -122.19, -122.026, -121.904, -122.115, -122.226, -122.345, -122.434, -122.434, -122.434, -122.5, -122.566, -122.61, -122.667, -122.744, -122.798, -122.876,
        -122.887, -122.887, -122.887, -122.887, -122.941, -123.03, -123.129, -123.251, -123.34, -123.406, -123.537, -123.582, -123.594, -123.648, -123.692, -123.692, -123.683, -123.648, -123.692,
        -123.758, -123.847, -123.913],
   y = [34.4351, 34.4797, 34.4985, 34.5079, 34.5079, 34.536, 34.59, 34.6628, 34.7449, 34.8176, 34.8716, 34.9162, 34.9444, 34.9796, 35.0101, 35.0007, 35.0101, 35.0007, 35.0171, 35.0359, 35.0641, 35.118,
        35.1439, 35.2072, 35.2448, 35.2706, 35.28, 35.2894, 35.3621, 35.4349, 35.4607, 35.5147, 35.5968, 35.632, 35.74, 35.8549, 36.0005, 36.0896, 36.214, 36.3924, 36.5707, 36.6059, 36.6505, 36.8265,
        36.9767, 36.9955, 36.9955, 37.0847, 37.2325, 37.3475, 37.479, 37.6033, 37.7348, 37.8122, 37.777, 37.6996, 37.5846, 37.5048, 37.6362, 37.838, 37.9436, 38.0211, 38.0469, 38.0915, 38.1619,
        38.1783, 38.1783, 38.1877, 38.0751, 37.9342, 37.892, 37.9084, 37.953, 37.953, 37.9859, 38.0117, 38.0915, 38.1079, 38.1525, 38.2206, 38.2393, 38.3356, 38.4036, 38.4904, 38.5937, 38.6805, 38.758,
        38.8612, 39.0067, 39.103, 39.2391, 39.3682, 39.4878, 39.5137, 39.6146, 39.7084, 39.7929, 39.8962, 39.9807] };
  WCL[387] = struct {
   x = [-120.116, -120.16, -120.139, -120.071],
   y = [33.9681, 33.9869, 33.9869, 33.9963] };
  WCL[388] = struct {
   x = [-119.928, -120.005, -120.062, -120.116, -120.005, -120.062, -120.116],
   y = [33.8953, 33.8953, 33.9141, 33.9681, 33.8953, 33.9141, 33.9681] };
  WCL[389] = struct {
   x = [141.96, 142.059, 142.103, 142.157, 142.136, 142.07, 142.047, 141.927, 141.805, 141.728, 141.639, 141.639, 141.629, 141.397, 141.143, 141.054, 141.087, 141.143, 141.143, 141.164, 140.977, 140.855,
        140.812, 140.756, 140.744, 140.888, 140.923, 140.768, 140.59, 140.479, 140.237],
   y = [40.0229, 39.8188, 39.6404, 39.4691, 39.333, 39.2297, 39.0748, 38.9551, 38.9481, 38.765, 38.5421, 38.3684, 38.3168, 38.3332, 38.1947, 37.9342, 37.6714, 37.4531, 37.1434, 36.958, 36.8265, 36.6036,
        36.5332, 36.2586, 35.9441, 35.7916, 35.6836, 35.6038, 35.3598, 35.1626, 35.0077] };
  WCL[390] = struct {
   x = [139.984, 140.017, 140.148, 140.193],
   y = [35.233, 35.3692, 35.4959, 35.6226] };
  WCL[391] = struct {
   x = [139.972, 140.071, 140.139, 140.148, 140.017],
   y = [39.1616, 39.3236, 39.5301, 39.7507, 39.8446] };
  WCL[392] = struct {
   x = [140.237, 139.993, 139.951, 139.984],
   y = [35.0077, 34.8622, 35.0077, 35.233] };
  WCL[393] = struct {
   x = [140.193, 139.951, 139.895, 139.806, 139.785, 139.653, 139.409, 139.299, 139.254, 139.21, 139.09, 138.956, 138.912, 138.947, 138.792, 138.625, 138.461, 138.348, 138.315, 138.196, 137.996, 137.731,
        137.478, 137.257, 137.212, 137.278, 137.388, 137.266, 137.168, 137.114, 137.058, 136.935, 136.926, 136.837, 136.715, 136.628, 136.771, 136.968, 136.968, 136.837, 136.715, 136.583, 136.45,
        136.372, 136.351, 136.229, 136.065, 135.854, 135.701, 135.579, 135.478, 135.29, 135.236, 135.248, 135.248, 135.335, 135.502, 135.523, 135.478, 135.326, 135.147, 134.927, 134.685, 134.464,
        134.265, 134.143, 133.978, 133.955, 133.901, 133.835, 133.68, 133.58, 133.403, 133.281, 133.162, 133.016, 132.852, 132.664, 132.53, 132.42, 132.3, 132.244, 132.155, 131.925, 131.758, 131.571,
        131.427, 131.338, 131.162, 130.986, 130.975, 130.998, 131.195, 131.46, 131.636, 131.815, 132.122, 132.322, 132.563, 132.718, 132.918, 133.138, 133.305, 133.46, 133.8, 134.077, 134.309, 134.584,
        134.795, 134.959, 135.138, 135.326, 135.314, 135.412, 135.734, 135.821, 135.91, 136.032, 136.142, 136.142, 136.152, 136.44, 136.682, 136.837, 136.849, 136.893, 137.234, 137.388, 137.356,
        137.212, 137.058, 137.036, 137.114, 137.114, 137.224, 137.433, 137.609, 137.853, 138.151, 138.47, 138.682, 138.836, 139.012, 139.233, 139.245, 139.254, 139.355, 139.508, 139.531, 139.576,
        139.695, 139.85, 139.928, 139.972],
   y = [35.6226, 35.6226, 35.4513, 35.3058, 35.1532, 35.233, 35.2612, 35.1157, 34.9186, 34.7261, 34.6182, 34.6369, 34.8716, 35.0359, 35.0523, 34.9256, 34.7637, 34.5806, 34.5548, 34.5712, 34.6088, 34.6182,
        34.6182, 34.5548, 34.5548, 34.6369, 34.7097, 34.7261, 34.7449, 34.7989, 34.7097, 34.8176, 35.0171, 34.9819, 34.7543, 34.6182, 34.536, 34.4163, 34.1793, 34.2708, 34.2169, 34.1981, 34.1347,
        34.0408, 33.9775, 33.8132, 33.555, 33.4541, 33.4987, 33.555, 33.7029, 33.8038, 33.9775, 34.1347, 34.2614, 34.3342, 34.5642, 34.6369, 34.6628, 34.644, 34.6088, 34.6815, 34.7261, 34.7167,
        34.5806, 34.5266, 34.4351, 34.4257, 34.4633, 34.4891, 34.4445, 34.4351, 34.3717, 34.3624, 34.3084, 34.2802, 34.2333, 34.1981, 34.299, 34.2708, 34.015, 33.886, 33.8672, 33.9329, 34.0056,
        33.9493, 33.9493, 33.9141, 33.9681, 33.9963, 34.2075, 34.3178, 34.3624, 34.4163, 34.5712, 34.6815, 34.827, 35.0453, 35.1791, 35.3879, 35.4865, 35.5499, 35.5217, 35.4771, 35.5053, 35.4959,
        35.5405, 35.5944, 35.632, 35.639, 35.693, 35.7024, 35.5592, 35.5311, 35.4771, 35.4959, 35.5686, 35.632, 35.693, 35.9441, 36.1765, 36.383, 36.6317, 36.9321, 37.1434, 37.3123, 37.4696, 37.4531,
        37.3123, 37.249, 37.1269, 37.0823, 36.9157, 36.7374, 36.728, 36.7561, 36.9063, 37.0119, 37.0917, 37.2842, 37.4531, 37.655, 37.8474, 37.9601, 37.9694, 37.9765, 37.9765, 38.1525, 38.3074,
        38.4388, 38.6289, 38.8354, 38.9974, 39.1616] };
  WCL[394] = struct {
   x = [140.017, 139.806, 139.907],
   y = [39.8446, 39.9126, 39.9807] };
  WCL[395] = struct {
   x = [138.604, 138.581, 138.625, 138.515, 138.36, 138.327, 138.449, 138.604],
   y = [38.2722, 38.0985, 38.0563, 37.8122, 37.8028, 37.9507, 38.2041, 38.2722] };
  WCL[396] = struct {
   x = [134.277, 134.065, 133.8, 133.57, 133.392, 133.293, 133.105, 133.084, 132.906, 132.751, 132.685, 132.554, 132.575, 132.509, 132.387, 132.223, 132.488, 132.697, 132.897, 133.105, 133.239, 133.436,
        133.624, 133.69, 133.824, 133.934, 134.065, 134.265, 134.453, 134.619, 134.661, 134.718, 134.762, 134.772, 134.762, 134.652, 134.551, 134.453, 134.363, 134.277],
   y = [33.2687, 33.379, 33.4541, 33.3978, 33.2969, 33.1115, 32.8792, 32.7125, 32.6938, 32.8041, 32.8792, 33.0833, 33.1678, 33.2781, 33.3603, 33.3603, 33.5363, 33.7569, 33.9775, 33.9493, 33.8953, 33.9329,
        34.0056, 34.1347, 34.2262, 34.3178, 34.3342, 34.2896, 34.1981, 34.1887, 33.9963, 33.8226, 33.8038, 33.7944, 33.7663, 33.6653, 33.6184, 33.5175, 33.3696, 33.2687] };
  WCL[397] = struct {
   x = [135.06, 135.028, 134.873, 134.762, 134.873, 134.983, 135.004, 135.06],
   y = [34.5712, 34.5266, 34.3717, 34.2075, 34.1793, 34.2262, 34.3905, 34.5712] };
  WCL[398] = struct {
   x = [130.864, 130.921, 130.831, 130.676, 130.522, 130.423, 130.28, 130.179, 130.036],
   y = [33.8766, 33.886, 33.8953, 33.8484, 33.7217, 33.5644, 33.5644, 33.4706, 33.4166] };
  WCL[399] = struct {
   x = [129.916, 130.146, 130.324, 130.411, 130.289, 130.235, 130.235, 130.334, 130.444, 130.578, 130.578, 130.622, 130.632, 130.622, 130.522, 130.402, 130.301, 130.289, 130.346, 130.289, 130.402, 130.655,
        130.709, 130.632, 130.688, 130.852, 130.775, 130.754, 130.81, 130.81, 130.942, 131.097, 131.108, 131.251, 131.362, 131.484, 131.549, 131.615, 131.726, 131.836, 131.967, 132.024, 132.035,
        132.024, 131.979, 131.925, 131.681, 131.648, 131.747, 131.693, 131.538, 131.24, 131.052, 131.04, 130.963, 130.864],
   y = [32.6562, 32.7125, 32.628, 32.7595, 32.8134, 32.8979, 33.0833, 33.149, 32.9449, 32.7665, 32.5999, 32.5811, 32.5811, 32.3863, 32.1798, 32.0953, 31.9357, 31.7104, 31.464, 31.3232, 31.2293, 31.1237,
        31.342, 31.5203, 31.6635, 31.6259, 31.5696, 31.4452, 31.2481, 31.0111, 31.0486, 31.1613, 31.3044, 31.4359, 31.3889, 31.6259, 31.8794, 32.1141, 32.3863, 32.553, 32.7125, 32.7947, 32.8416,
        32.9988, 33.0739, 33.1749, 33.2218, 33.3063, 33.5175, 33.6278, 33.5832, 33.5738, 33.7663, 33.8578, 33.8578, 33.8766] };
  WCL[400] = struct {
   x = [130.146, 130.092, 130.08, 130.113, 130.247, 130.146],
   y = [32.4779, 32.3582, 32.2174, 32.1423, 32.4215, 32.4779] };
  WCL[401] = struct {
   x = [131.118, 131.085, 131.031, 130.963, 131.031, 131.118],
   y = [30.6309, 30.7647, 30.6309, 30.4103, 30.3446, 30.6309] };
  WCL[402] = struct {
   x = [130.545, 130.512, 130.456, 130.512, 130.676, 130.676, 130.545],
   y = [30.354, 30.4103, 30.3258, 30.2108, 30.192, 30.3258, 30.354] };
  WCL[403] = struct {
   x = [130.897, 130.843, 130.843, 130.885, 130.975, 130.897],
   y = [37.54, 37.4954, 37.479, 37.4531, 37.5048, 37.54] };
  WCL[404] = struct {
   x = [130.036, 129.97, 129.86, 129.771, 129.66, 129.771, 129.839, 129.881, 129.949, 129.958, 129.872, 129.749, 129.782, 129.815, 129.916],
   y = [33.4166, 33.4706, 33.3978, 33.3063, 33.1749, 33.0927, 33.0552, 33.0082, 32.9355, 32.8134, 32.9636, 32.973, 32.7665, 32.5623, 32.6562] };
  WCL[405] = struct {
   x = [128.855, 128.712, 128.712, 128.867, 128.855],
   y = [32.7219, 32.6656, 32.5342, 32.5811, 32.7219] };
  WCL[406] = struct {
   x = [129.219, 129.198, 129.132, 129.055, 129.097, 129.219, 129.219],
   y = [32.9543, 33.027, 33.027, 32.8885, 32.7947, 32.8885, 32.9543] };
  WCL[407] = struct {
   x = [129.749, 129.728, 129.749, 129.815, 129.749],
   y = [33.839, 33.7663, 33.7029, 33.7569, 33.839] };
  WCL[408] = struct {
   x = [129.285, 129.198, 129.231, 129.32, 129.285],
   y = [34.2896, 34.1511, 34.069, 34.1347, 34.2896] };
  WCL[409] = struct {
   x = [129.407, 129.484, 129.484, 129.484, 129.463, 129.44, 129.386, 129.374, 129.407],
   y = [34.6369, 34.5712, 34.4985, 34.4257, 34.3342, 34.299, 34.3905, 34.5454, 34.6369] };
  WCL[410] = struct {
   x = [128.435, 128.513, 128.656, 128.811, 128.987, 129.165, 129.308, 129.43, 129.496, 129.484, 129.473, 129.505, 129.618, 129.55, 129.463, 129.329, 129.132, 128.942, 128.757, 128.602, 128.557, 128.536,
        128.492, 128.492, 128.48, 128.426, 128.346, 128.28, 128.182, 128.071, 127.994, 127.895, 127.773, 127.675, 127.75, 127.708, 127.607, 127.541, 127.464, 127.452, 127.452, 127.342, 127.255,
        127.198, 127.332, 127.342, 127.231, 127.079, 126.99, 126.945, 126.933, 126.9, 126.835, 126.79, 126.724, 126.548, 126.525, 126.361, 126.295, 126.239, 126.316, 126.349, 126.447, 126.537, 126.525,
        126.581, 126.492, 126.394, 126.337, 126.394, 126.438, 126.438, 126.525, 126.626, 126.57, 126.736, 126.781, 126.713, 126.713, 126.593, 126.57, 126.516, 126.492, 126.382, 126.426, 126.337,
        126.239, 126.14, 126.161, 126.295, 126.447, 126.57, 126.692, 126.856, 126.924, 126.835, 126.835, 126.769, 126.79, 126.835, 126.736, 126.68, 126.602, 126.593, 126.647, 126.614, 126.537, 126.504,
        126.504, 126.504],
   y = [38.5069, 38.4388, 38.2206, 37.9953, 37.8638, 37.6714, 37.4602, 37.1879, 36.8711, 36.613, 36.3384, 36.0333, 36.0263, 35.8549, 35.578, 35.3527, 35.1884, 35.0805, 35.1087, 35.1532, 35.0899, 35.0711,
        34.989, 34.9819, 34.9538, 34.8904, 34.8083, 34.9186, 34.9186, 34.9538, 34.9984, 34.9819, 34.935, 34.8998, 34.8083, 34.7097, 34.6534, 34.7637, 34.8176, 34.7261, 34.5712, 34.5079, 34.4985,
        34.5642, 34.6534, 34.7355, 34.7097, 34.6276, 34.5266, 34.4891, 34.4891, 34.4539, 34.4539, 34.5548, 34.4797, 34.3436, 34.4727, 34.4351, 34.3084, 34.3717, 34.4891, 34.644, 34.6276, 34.59,
        34.6909, 34.7989, 34.8528, 34.7543, 34.881, 35.0077, 35.0805, 35.226, 35.4137, 35.5405, 35.5944, 35.7024, 35.8104, 35.8995, 35.9887, 36.0802, 36.2656, 36.437, 36.6317, 36.6763, 36.498, 36.4534,
        36.5965, 36.5965, 36.7374, 36.8524, 36.8617, 36.9063, 36.9321, 36.8265, 36.8078, 36.9509, 37.073, 37.1434, 37.2325, 37.2677, 37.3381, 37.3288, 37.4179, 37.479, 37.4883, 37.5306, 37.5846,
        37.6714, 37.7512, 37.7676] };
  WCL[411] = struct {
   x = [126.328, 126.227, 126.239, 126.426, 126.692, 126.933, 126.9, 126.614, 126.328],
   y = [33.3603, 33.2593, 33.1678, 33.149, 33.1749, 33.3509, 33.4635, 33.4635, 33.3603] };
  WCL[412] = struct {
   x = [128.006, 127.928, 127.75, 127.553, 127.541, 127.553, 127.565, 127.541, 127.529, 127.452, 127.497, 127.729, 127.928, 128.182, 128.381, 128.435],
   y = [40.0229, 39.9032, 39.8258, 39.7765, 39.6826, 39.5981, 39.4174, 39.3165, 39.4362, 39.2649, 39.171, 39.1006, 38.9223, 38.7744, 38.6359, 38.5069] };
  WCL[413] = struct {
   x = [126.57, 126.426, 126.239, 126.161, 126.039, 125.908, 125.786, 125.664, 125.72, 125.654, 125.544, 125.466, 125.333, 125.288, 125.234, 125.112, 125.091, 125.19, 125.255, 125.267, 125.091, 124.87,
        124.805, 124.948, 125.035, 125.157, 125.178, 125.344, 125.455, 125.312, 125.267, 125.377, 125.443, 125.511, 125.41, 125.246, 125.136, 124.903, 124.727, 124.671, 124.593, 124.462, 124.361,
        124.394],
   y = [37.723, 37.838, 37.784, 37.723, 37.8028, 37.838, 37.9601, 37.9765, 37.9155, 37.777, 37.7418, 37.662, 37.662, 37.6714, 37.8286, 37.8122, 37.899, 37.899, 37.9859, 38.0563, 38.0211, 38.0211, 38.0305,
        38.2112, 38.4459, 38.4975, 38.5585, 38.6101, 38.6618, 38.6618, 38.765, 39.0936, 39.3658, 39.5043, 39.5536, 39.5629, 39.5794, 39.6474, 39.5723, 39.5465, 39.6897, 39.7765, 39.8352, 39.9032] };
  WCL[414] = struct {
   x = [121.568, 121.514, 121.514, 121.348, 121.193, 121.017, 120.817, 120.609, 120.388, 120.212, 120.364, 120.564, 120.85, 121.061, 121.127, 121.148, 121.27, 121.47, 121.658, 121.923, 121.932, 121.801,
        121.646, 121.392, 121.205, 120.972, 120.773, 120.719, 120.63, 120.409, 120.156, 120.156, 120.409, 120.641, 120.874, 121.106, 121.315, 121.568, 121.768, 121.911, 121.923, 121.834, 121.712,
        121.634, 121.514, 121.392, 121.369, 121.127, 120.961, 120.883, 120.763, 120.707, 120.653, 120.597, 120.465, 120.376, 120.254, 120.078],
   y = [29.9808, 30.077, 30.077, 30.2014, 30.239, 30.1733, 30.0958, 30.1826, 30.239, 30.124, 30.307, 30.3352, 30.3352, 30.4502, 30.5347, 30.5441, 30.6121, 30.6684, 30.8304, 30.8398, 30.9149, 31.1237,
        31.2857, 31.4077, 31.5978, 31.7292, 31.748, 31.87, 31.9263, 31.8794, 31.8419, 31.9357, 31.9827, 32.0202, 31.9545, 31.8231, 31.8137, 31.748, 31.6729, 31.6166, 31.6635, 31.7855, 31.9451, 31.9921,
        32.0953, 32.1986, 32.3582, 32.4497, 32.6562, 32.9167, 33.1397, 33.3251, 33.5269, 33.7381, 33.886, 34.1347, 34.2802, 34.3624] };
  WCL[415] = struct {
   x = [119.846, 120.045, 120.177, 120.254, 120.111, 120.245, 120.421, 120.475, 120.663, 120.796, 120.883, 120.939, 120.907, 120.895, 121.106, 121.327, 121.601, 121.723, 121.845, 121.977, 122.087, 122.197,
        122.242, 122.263, 122.341, 122.519, 122.507, 122.606, 122.639, 122.65, 122.585, 122.352, 122.176, 122.12, 121.965, 121.744, 121.592, 121.458, 121.27, 121.106, 120.996, 120.907, 120.773,
        120.543, 120.364, 120.233, 120.024],
   y = [35.578, 35.6672, 35.801, 35.9277, 36.0521, 36.1859, 36.1859, 36.0263, 36.0427, 36.2938, 36.3736, 36.3736, 36.5168, 36.5613, 36.5613, 36.6576, 36.6834, 36.7022, 36.8359, 36.9228, 36.8876, 36.9509,
        36.8711, 36.8007, 36.7561, 36.9321, 36.9955, 37.1269, 37.2325, 37.3217, 37.3475, 37.3733, 37.4344, 37.4602, 37.3898, 37.364, 37.3733, 37.4954, 37.5142, 37.5916, 37.6808, 37.7582, 37.7418,
        37.655, 37.5846, 37.4344, 37.3123] };
  WCL[416] = struct {
   x = [119.804, 120.012],
   y = [39.8704, 39.9807] };
  WCL[417] = struct {
   x = [121.956, 121.756, 121.557, 121.524, 121.348, 121.47, 121.625, 121.756, 121.702, 121.625, 121.36, 121.193, 121.303, 121.646, 121.756, 121.923, 121.977, 122.12, 122.242, 122.409, 122.739, 123.047,
        123.324, 123.401, 123.512, 123.655, 123.908, 124.108, 124.251, 124.373, 124.394],
   y = [40.0464, 39.8258, 39.7249, 39.5536, 39.4785, 39.4526, 39.3588, 39.3165, 39.1523, 38.9974, 38.9387, 38.765, 38.697, 38.8354, 38.9223, 38.9223, 38.9551, 39.0419, 39.2133, 39.3423, 39.4433, 39.5465,
        39.6897, 39.7577, 39.6662, 39.7084, 39.7577, 39.7671, 39.8023, 39.861, 39.9032] };
  WCL[418] = struct {
   x = [120.078, 119.89, 119.649, 119.482, 119.283, 119.25, 119.393, 119.548, 119.616, 119.726, 119.846],
   y = [34.3624, 34.4351, 34.5266, 34.644, 34.6909, 34.8176, 35.0523, 35.233, 35.3152, 35.5147, 35.578] };
  WCL[419] = struct {
   x = [120.024, 119.935, 119.813, 119.56, 119.327, 119.062, 118.919, 118.907, 118.874, 118.797, 118.6, 118.445, 118.356, 118.224, 118.069, 117.992, 117.825, 117.673, 117.727, 117.837, 118.013, 118.234,
        118.356, 118.555, 118.764, 119.02, 119.196, 119.351, 119.405, 119.527, 119.682, 119.804],
   y = [37.3123, 37.2067, 37.073, 37.0471, 37.1175, 37.1973, 37.364, 37.5048, 37.6198, 37.7676, 37.899, 37.899, 38.0023, 38.0023, 38.0633, 38.1854, 38.3168, 38.4646, 38.7064, 38.9903, 39.1264, 39.11, 39.049,
        39.0584, 39.11, 39.1358, 39.2297, 39.3588, 39.5536, 39.7413, 39.7929, 39.8704] };
  WCL[420] = struct {
   x = [49.9735, 50.1706, 50.3044, 50.4358, 50.6893, 51.0413, 51.3722, 51.682, 51.9026, 52.0458, 52.21, 52.4306, 52.6747, 52.9164, 53.2356, 53.4891, 53.6557, 53.5782, 53.3342, 53.4797, 53.6557, 53.8857,
        54.0734, 54.0617, 53.9514, 53.9185],
   y = [37.4649, 37.4555, 37.3757, 37.1481, 36.9885, 36.8195, 36.7233, 36.6529, 36.6341, 36.6787, 36.7045, 36.7585, 36.8195, 36.8922, 36.9275, 36.9791, 37.0072, 36.9439, 36.8993, 36.8829, 36.8735, 36.9087,
        36.9345, 37.2091, 37.4367, 37.5165] };
  WCL[421] = struct {
   x = [53.9185, 53.9303, 53.8857, 53.82, 53.7988, 53.8763, 54.0406, 53.8411, 53.6557, 53.5219, 53.3788, 53.3788, 53.5548, 53.7543, 53.8763, 53.6768, 53.5337, 53.3905, 53.5876, 53.7097, 53.8857],
   y = [37.5165, 37.5775, 37.8239, 38.1619, 38.535, 38.8542, 39.0607, 39.1124, 39.2414, 39.3869, 39.4127, 39.455, 39.523, 39.5817, 39.6498, 39.7014, 39.7624, 39.854, 39.9995, 39.922, 39.8633] };
  WCL[422] = struct {
   x = [53.0925, 52.961, 52.8507],
   y = [40.0839, 39.9572, 39.9572] };
  WCL[423] = struct {
   x = [48.8799, 48.8916, 48.9245, 49.0019, 49.1005, 49.3211, 49.5206, 49.7083, 49.9735],
   y = [38.5702, 38.3801, 38.1455, 37.8662, 37.6737, 37.5353, 37.5071, 37.4461, 37.4649] };
  WCL[424] = struct {
   x = [49.5088, 49.476, 49.4431, 49.3328, 49.2788, 49.2554, 49.1122, 49.023, 48.9808, 48.8916, 48.847, 48.8376, 48.8588, 48.8799],
   y = [40.0675, 39.9478, 39.6944, 39.4714, 39.2156, 39.1124, 39.2156, 39.3001, 39.2743, 39.225, 39.103, 38.8026, 38.6735, 38.5702] };
  WCL[425] = struct {
   x = [48.373, 48.2533, 48.0655],
   y = [29.9245, 30.0019, 30.0794] };
  WCL[426] = struct {
   x = [48.5607, 48.5513],
   y = [30.1169, 30.0982] };
  WCL[427] = struct {
   x = [50.0486, 49.9829, 49.8843, 49.6637],
   y = [29.9832, 30.0888, 30.1451, 30.07] };
  WCL[428] = struct {
   x = [49.4877, 49.3891, 49.3328, 49.2014, 49.1122, 49.0911, 49.0019, 48.847, 48.7602, 48.8048, 48.7602, 48.7696, 48.6264, 48.5936],
   y = [29.9926, 30.0207, 30.0982, 30.1756, 30.307, 30.4619, 30.537, 30.5652, 30.5182, 30.4126, 30.3282, 30.2132, 30.2132, 30.0794] };
  WCL[429] = struct {
   x = [29.8611, 30.0935, 30.2366, 30.4572, 30.6778, 30.7107, 30.469, 30.4126, 30.469, 30.6121, 30.8093, 30.9313, 30.9876, 30.9759, 31.2082, 31.5485, 31.8137, 31.9357, 32.1, 32.1117],
   y = [31.2575, 31.3537, 31.457, 31.4664, 31.5227, 31.4945, 31.41, 31.3349, 31.3256, 31.3537, 31.3819, 31.4664, 31.5039, 31.5415, 31.4945, 31.4851, 31.4664, 31.3443, 31.288, 31.288] };
  WCL[430] = struct {
   x = [31.7245, 31.7245, 31.748, 31.7151, 31.6823, 31.6142, 31.6823, 31.7597, 31.7808, 31.7808, 31.9357, 32.0226, 32.0883, 32.1446, 32.1775, 32.1986, 32.1986, 32.1892, 32.1986, 32.2878, 32.3652],
   y = [31.5415, 31.4945, 31.4288, 31.3725, 31.2669, 31.1824, 31.1824, 31.1824, 31.1261, 31.0322, 30.9947, 30.9266, 30.8609, 30.6426, 30.4995, 30.4525, 30.4408, 30.4032, 30.3282, 30.2601, 30.1639] };
  WCL[431] = struct {
   x = [35.9371, 35.9488, 35.9911, 35.916, 35.9911, 36.1694, 36.2469, 36.0802, 35.9817, 35.7611, 35.7282, 35.5851, 35.341, 35.1556, 34.935, 34.8012, 34.6909, 34.9444, 34.7566, 34.5478, 34.4046, 34.2169,
        34.1394, 33.9517, 33.8085, 33.7217, 33.6231, 33.4236, 33.2359, 33.0716, 32.905, 32.7407, 32.5623, 32.3863, 32.3089, 32.1986, 32.0226, 31.87, 31.6823, 31.5602, 31.4945, 31.3396, 31.173, 31.0299,
        30.8538, 30.6989, 30.645, 30.645, 30.5793, 30.5558, 30.4032, 30.192],
   y = [35.8151, 35.9582, 36.1014, 36.2962, 36.5379, 36.6247, 36.7585, 36.8829, 36.8383, 36.7773, 36.6247, 36.5801, 36.6083, 36.7327, 36.7937, 36.8195, 36.8383, 36.8031, 36.8289, 36.7585, 36.6341, 36.5003,
        36.3689, 36.3407, 36.2422, 36.2164, 36.1812, 36.1976, 36.153, 36.1272, 36.0638, 36.1014, 36.153, 36.3055, 36.4205, 36.5003, 36.5895, 36.6529, 36.7233, 36.7679, 36.8125, 36.8829, 36.8993,
        36.9181, 36.8922, 36.8922, 36.7679, 36.5801, 36.5191, 36.3501, 36.3243, 36.2797] };
  WCL[432] = struct {
   x = [34.0408, 34.0948, 34.2826, 34.4046, 34.5806, 34.6581, 34.6041, 34.4492, 34.3154, 34.2169, 34.1183, 34.1511, 34.0408, 33.8202, 33.6114, 33.3251, 33.1373, 33.027, 32.8158, 32.5858, 32.4426, 32.3535,
        32.4309, 32.5858, 32.7501, 32.9261, 33.0481, 33.0599, 33.0716, 33.1373, 33.2359, 33.3579, 33.5011, 33.6888, 33.8531, 34.0408],
   y = [35.4724, 35.4818, 35.5522, 35.6343, 35.7071, 35.7141, 35.5898, 35.4536, 35.409, 35.2823, 35.1556, 35.0664, 34.9491, 34.9115, 34.7496, 34.7027, 34.5947, 34.5947, 34.6205, 34.712, 34.8669, 35.0289,
        35.0476, 35.0922, 35.165, 35.1838, 35.219, 35.3363, 35.3551, 35.3269, 35.3269, 35.3363, 35.3457, 35.3903, 35.4372, 35.4724] };
  WCL[433] = struct {
   x = [35.9911, 35.9488, 36.0028, 35.9911, 35.9042, 35.8479, 35.8831, 35.9371],
   y = [34.7402, 34.9021, 35.0922, 35.3645, 35.5616, 35.6156, 35.6695, 35.8151] };
  WCL[434] = struct {
   x = [35.4748, 35.4748, 35.4748, 35.4748, 35.463, 35.4419, 35.3856, 35.341, 35.2987, 35.287, 35.2542, 35.1884, 35.165, 35.1321, 35.111, 35.0875, 35.0664, 35.0453, 35.0007, 34.989, 34.9772, 35.4748, 35.578,
        35.6602, 35.7212, 35.8432, 36.0263, 35.9911],
   y = [33.9704, 33.9611, 33.9517, 33.9235, 33.9235, 33.9141, 33.8507, 33.778, 33.6935, 33.6395, 33.5926, 33.5105, 33.4541, 33.4072, 33.3532, 33.3063, 33.2687, 33.2335, 33.1866, 33.1397, 33.1209, 33.9704,
        34.0009, 34.238, 34.3741, 34.4844, 34.5853, 34.7402] };
  WCL[435] = struct {
   x = [32.1446, 32.1775, 32.2314, 32.3323, 32.3769, 32.4192, 32.4966, 32.729, 32.8721, 32.9378, 33.0153, 33.081, 33.1045, 33.2148, 33.3579, 33.5222, 33.6888, 33.8531, 34.0291, 34.13, 34.13, 34.1629,
        34.2614, 34.36, 34.4375, 34.4938, 34.5149, 34.5149],
   y = [31.2786, 31.2481, 31.2481, 31.2012, 31.1448, 31.0885, 31.0228, 31.0134, 31.0322, 31.0885, 31.1073, 31.0322, 31.0134, 31.0228, 31.0416, 31.0604, 31.1167, 31.1824, 31.2974, 31.3725, 31.3819, 31.3913,
        31.457, 31.5415, 31.6259, 31.6917, 31.7386, 31.7503] };
  WCL[436] = struct {
   x = [34.5149, 34.5595, 34.5806, 34.6041, 34.6369, 34.6463, 34.6792, 34.7027, 34.7238, 34.7472, 34.7801, 34.7801, 34.8012, 34.8575, 34.8998, 34.935, 34.9444, 34.9444, 34.9561],
   y = [31.7503, 31.8724, 31.9475, 32.0038, 32.0789, 32.1446, 32.2455, 32.3112, 32.3957, 32.4708, 32.5459, 32.6022, 32.6656, 32.7595, 32.8252, 32.8909, 32.9824, 33.0481, 33.0669] };
  WCL[437] = struct {
   x = [32.553, 32.5201, 32.4755, 32.4638, 32.4309, 32.4098, 32.3417, 32.3089, 32.222, 32.1986, 32.1986, 32.1892, 32.1892, 32.1892, 32.1892, 32.1657],
   y = [29.9926, 30.0207, 30.0606, 30.0606, 30.0794, 30.1545, 30.2601, 30.2976, 30.4408, 30.6332, 30.7365, 30.8515, 31.004, 31.1167, 31.173, 31.2293] };
  WCL[438] = struct {
   x = [25.3646, 25.3083, 25.3317, 25.32, 25.2097, 25.0665, 25.022, 25.0877, 25.144, 25.1886, 25.2754, 25.2989, 25.3646],
   y = [39.922, 39.7953, 39.922, 39.983, 39.9901, 39.9643, 39.8305, 39.7953, 39.8469, 39.8375, 39.8047, 39.8305, 39.922] };
  WCL[439] = struct {
   x = [26.5333, 26.6436, 26.5544, 26.4559, 26.3667, 26.2235, 26.0804, 25.9161, 25.8715, 25.9912, 26.1367, 26.247, 26.1578, 26.3127, 26.4653, 26.4653, 26.5333],
   y = [39.0795, 39.0607, 39.225, 39.3611, 39.455, 39.394, 39.3517, 39.2931, 39.164, 39.1476, 39.2062, 39.2156, 39.0701, 39.0091, 39.0795, 39.1288, 39.0795] };
  WCL[440] = struct {
   x = [25.9701, 26.0475, 26.0921, 26.125, 26.1367, 26.0921, 25.8809, 25.8809, 25.9912, 25.9044, 25.9044, 25.9701],
   y = [38.2417, 38.2675, 38.3027, 38.4482, 38.5608, 38.6219, 38.6383, 38.528, 38.4153, 38.3191, 38.2487, 38.2417] };
  WCL[441] = struct {
   x = [28.0987, 28.1761, 28.1972, 28.2747, 28.3193, 28.2183, 28.0541, 27.8663, 27.7889, 27.7678, 27.7795, 27.9226, 28.0987],
   y = [36.0732, 36.1014, 36.1812, 36.2868, 36.4581, 36.4299, 36.3689, 36.2351, 36.1812, 36.1014, 35.93, 35.984, 36.0732] };
  WCL[442] = struct {
   x = [25.0665, 24.9234, 24.7474, 24.569, 24.3719, 24.2498, 24.1513, 24.1959, 24.1184, 24.0081, 23.8086, 23.7429, 23.6983, 23.5998, 23.5787, 23.5458, 23.6444, 23.7993, 23.9753, 24.1959, 24.3484, 24.4704,
        24.569, 24.7239, 24.8131, 25.1111, 25.2989, 25.4632, 25.6063, 25.8058, 25.9701, 26.1578, 26.3221, 26.2892, 26.1696, 26.0686, 25.9372, 25.7823, 25.6955, 25.496, 25.374, 25.1886, 25.0665,
        25.0665],
   y = [35.3738, 35.4184, 35.3997, 35.3738, 35.3457, 35.4278, 35.4372, 35.5264, 35.5991, 35.517, 35.6062, 35.6156, 35.5006, 35.5264, 35.3832, 35.1931, 35.2119, 35.2025, 35.219, 35.2119, 35.1838, 35.1556,
        35.0922, 35.0476, 34.8927, 34.9115, 34.9655, 34.9561, 34.9843, 34.9937, 35.0289, 34.9561, 35.1838, 35.2565, 35.2025, 35.1931, 35.1486, 35.1931, 35.3105, 35.2917, 35.3011, 35.3011, 35.3269,
        35.3738] };
  WCL[443] = struct {
   x = [24.9445, 25.0008, 25.0008, 24.968, 24.8459, 24.7122, 24.7028, 24.7896, 24.8459, 24.8905, 24.9445],
   y = [37.6573, 37.7089, 37.8146, 37.8756, 37.946, 37.9718, 37.9014, 37.8239, 37.7348, 37.6667, 37.6573] };
  WCL[444] = struct {
   x = [25.4303, 25.5523, 25.5735, 25.5852, 25.5171, 25.4749, 25.3857, 25.3857, 25.4303],
   y = [36.9181, 36.9275, 37.0683, 37.1551, 37.1551, 37.1387, 37.0589, 36.9345, 36.9181] };
  WCL[445] = struct {
   x = [22.0344, 21.9124, 21.7927, 21.5275, 21.3844, 21.1731, 21.1309, 21.2506, 21.5275, 21.6589, 21.5721, 21.6144, 21.7153, 21.8584, 21.903, 22.0227, 22.1001, 22.2433, 22.3325, 22.3536, 22.4428, 22.4756,
        22.5202, 22.6845, 22.7948, 22.938, 23.0929, 23.0694, 23.06, 23.0154, 22.9262, 22.7948, 22.7408, 22.8159, 22.938, 23.0694, 23.1046, 23.1914, 23.3135, 23.4684, 23.4355, 23.3346, 23.2149, 23.1374,
        23.0811, 23.0929, 22.9262, 22.4756, 22.255, 22.1565, 22.0884, 22.0344],
   y = [38.3097, 38.3027, 38.2839, 38.1455, 38.19, 37.9178, 37.7887, 37.5963, 37.4907, 37.3053, 37.0683, 36.8735, 36.7843, 36.7397, 36.8993, 37.0237, 36.9181, 36.8195, 36.5801, 36.4205, 36.4111, 36.5003,
        36.7045, 36.7679, 36.6177, 36.5097, 36.4393, 36.6693, 36.7679, 36.9181, 37.1997, 37.3593, 37.5353, 37.5071, 37.4461, 37.3687, 37.2795, 37.2959, 37.3851, 37.4015, 37.5071, 37.5517, 37.6221,
        37.7606, 37.8404, 37.953, 38.0234, 38.1455, 38.1971, 38.1971, 38.2675, 38.3097] };
  WCL[446] = struct {
   x = [20.7319, 20.7765, 20.6451, 20.6005, 20.502, 20.3705, 20.3588, 20.4456, 20.502, 20.6334, 20.7319],
   y = [38.068, 38.19, 38.2675, 38.406, 38.3614, 38.2675, 38.2487, 38.19, 38.129, 38.1103, 38.068] };
  WCL[447] = struct {
   x = [24.569, 24.6136, 24.6042, 24.4165, 24.3156, 24.2053, 24.2053, 24.1067, 23.8321, 23.6444, 23.4777, 23.358, 23.3252, 23.2149, 23.0365, 22.9614, 23.0037, 23.1468, 23.2689, 23.358, 23.5341, 23.5669,
        23.7218, 23.9307, 23.9964, 24.1278, 24.1959, 24.2381, 24.339, 24.4493, 24.5362, 24.569, 24.569, 24.569],
   y = [38.0751, 38.0328, 38.1548, 38.19, 38.2229, 38.4318, 38.5538, 38.6383, 38.6829, 38.7509, 38.8636, 38.9669, 39.0255, 39.0255, 38.9152, 38.8378, 38.8284, 38.8378, 38.7509, 38.6829, 38.5796, 38.467,
        38.3966, 38.3801, 38.3356, 38.2581, 38.1713, 38.0399, 37.9718, 37.953, 37.946, 37.946, 37.9788, 38.0751] };
  WCL[448] = struct {
   x = [20.049, 20.0607],
   y = [39.394, 39.455] };
  WCL[449] = struct {
   x = [19.9387, 20.049],
   y = [39.3869, 39.394] };
  WCL[450] = struct {
   x = [23.9424, 23.9635],
   y = [40.1426, 39.9643] };
  WCL[451] = struct {
   x = [23.5012, 23.6655, 23.5669],
   y = [40.1004, 39.9995, 39.9408] };
  WCL[452] = struct {
   x = [22.6071, 22.6399, 22.762, 22.8394, 23.0271, 23.1586, 23.2477, 23.29, 23.2032, 23.1257, 23.1586, 23.0811, 22.9168, 22.8723, 22.9497, 22.9708, 22.9168, 22.8277, 22.7174, 22.5414, 22.6188, 22.8277,
        22.9943, 23.1257, 23.29, 23.3792, 23.5012, 23.6326, 23.7758, 23.8744, 24.0081, 24.0175, 24.0175, 24.0175, 24.0292, 23.7875, 23.5998, 23.358, 23.0929, 22.9262, 23.114, 23.1468, 22.938, 22.8065,
        22.6845, 22.5742, 22.3982, 22.1659, 21.9453, 21.6261, 21.3726, 21.3069, 21.2741, 21.1638, 21.0863, 20.9877, 20.8774, 20.7554, 20.9432, 21.1192, 21.0535, 20.7883, 20.7437, 20.6451, 20.4691,
        20.2931, 20.2602, 20.2156, 20.1828, 20.1593],
   y = [40.152, 39.9736, 39.8727, 39.6592, 39.5324, 39.4198, 39.2743, 39.164, 39.1217, 39.1382, 39.232, 39.3189, 39.3353, 39.2156, 39.0701, 39.0185, 38.9762, 38.9575, 38.9669, 38.941, 38.8472, 38.7697,
        38.7345, 38.6735, 38.6641, 38.4998, 38.4998, 38.3872, 38.3027, 38.2933, 38.2229, 38.0844, 37.9272, 37.7794, 37.6831, 37.8052, 37.9882, 37.953, 37.953, 38.0234, 38.068, 38.1807, 38.2065,
        38.2933, 38.3966, 38.3191, 38.3708, 38.3449, 38.3801, 38.3449, 38.4224, 38.4224, 38.3708, 38.3449, 38.4412, 38.6219, 38.7509, 38.8894, 38.873, 38.8988, 39.0185, 39.103, 38.9762, 39.103,
        39.3353, 39.4386, 39.5395, 39.6686, 39.7014, 39.7014] };
  WCL[453] = struct {
   x = [20.0819, 20.0819],
   y = [39.685, 39.685] };
  WCL[454] = struct {
   x = [19.9387, 20.0725, 20.2156, 20.3799, 20.5559, 20.854, 21.152, 21.3726, 21.6472, 21.9359, 22.1659, 22.4639, 22.5953, 22.7408, 22.9168, 22.9262, 22.9051, 22.8605, 22.8934, 23.0694, 23.1586, 23.2477,
        23.5106, 23.7547, 23.865, 24.0175, 24.3156, 24.5596, 24.7685, 24.9234, 25.0008, 25.022],
   y = [32.1728, 32.3018, 32.4145, 32.5271, 32.6468, 32.7782, 32.8158, 32.8909, 32.9003, 32.9003, 32.8346, 32.7501, 32.6938, 32.6844, 32.6022, 32.452, 32.4239, 32.4333, 32.3206, 32.1986, 32.1728, 32.1892,
        32.1516, 32.0977, 32.0507, 31.9944, 31.9569, 31.9569, 31.9381, 31.863, 31.6635, 31.6729] };
  WCL[455] = struct {
   x = [25.0665, 25.1886, 25.5078, 25.7823, 26.0147, 26.2235, 26.4887, 26.7633, 27.1271, 27.3594, 27.6246, 27.7889, 27.932, 28.2301, 28.4624, 28.7276, 28.913, 29.2017, 29.488, 29.6523, 29.8611],
   y = [31.5602, 31.5227, 31.5884, 31.6072, 31.5696, 31.5321, 31.5133, 31.4758, 31.3725, 31.2387, 31.2012, 31.1636, 31.0697, 31.0697, 31.0134, 30.936, 30.8421, 30.8984, 31.0322, 31.1448, 31.2575] };
  WCL[456] = struct {
   x = [30.192, 29.9386, 29.7297, 29.5537, 29.3777, 29.2228, 29.1782, 29.1571, 29.1571, 29.0233, 28.9811, 28.9905, 28.9247, 28.7699, 28.6713, 28.6267, 28.4742, 28.3521, 28.3075, 28.1315, 28.0869, 28.1526,
        28.1198, 27.8992, 27.6786, 27.5026, 27.6011, 27.7795, 28.0212, 28.1432, 28.1761, 28.3286, 28.385, 28.2418, 28.0329, 27.8663, 27.6692, 27.4368, 27.3923, 27.4908, 27.6129, 27.5472, 27.4814,
        27.4134, 27.3148, 27.3383, 27.4368, 27.4697, 27.3477, 27.2702, 27.2162, 27.1271, 27.2491, 27.282, 27.1951, 27.0848, 26.9299, 26.8408, 26.7093, 26.6202, 26.5427, 26.4559, 26.355, 26.3456,
        26.4653, 26.4887, 26.4653, 26.5099, 26.5756, 26.6859, 26.7962, 26.9511, 27.1177, 27.1717, 27.1271, 26.9299, 26.8408, 26.8736, 27.0168, 26.9745, 26.8736, 26.8736, 26.7633, 26.7868, 26.9511,
        26.9065, 26.7422, 26.6202, 26.4324, 26.2892, 26.1789, 26.1696, 26.1907, 26.2235],
   y = [36.2797, 36.1976, 36.1976, 36.2351, 36.3149, 36.4487, 36.5895, 36.6623, 36.7233, 36.7773, 36.7139, 36.6623, 36.6623, 36.7397, 36.8477, 36.857, 36.8477, 36.8289, 36.7491, 36.6529, 36.6529, 36.7585,
        36.8125, 36.7585, 36.7045, 36.6693, 36.7843, 36.8641, 36.8641, 36.9181, 36.9533, 37.0143, 37.0941, 37.1035, 37.0683, 37.0331, 37.0331, 37.0331, 37.1129, 37.1739, 37.2701, 37.3147, 37.3851,
        37.4203, 37.4367, 37.5517, 37.6033, 37.6737, 37.6667, 37.6315, 37.6221, 37.7277, 37.7887, 37.9108, 38.0492, 38.1103, 38.129, 38.1548, 38.2417, 38.2159, 38.2065, 38.2487, 38.3285, 38.3801,
        38.4153, 38.5444, 38.6219, 38.7087, 38.6829, 38.4576, 38.4318, 38.4576, 38.4576, 38.4576, 38.4998, 38.5608, 38.6923, 38.7955, 38.8894, 38.9762, 39.0607, 39.1476, 39.2579, 39.4127, 39.5395,
        39.6005, 39.5747, 39.5489, 39.5489, 39.5395, 39.5489, 39.6169, 39.7437, 39.9056] };
  WCL[457] = struct {
   x = [20.0607, 19.9716, 19.8941, 19.7862, 19.6524, 19.751, 19.819, 19.9387],
   y = [39.455, 39.5747, 39.7108, 39.7789, 39.7437, 39.5911, 39.4878, 39.3869] };
  WCL[458] = struct {
   x = [20.0819, 19.995, 19.9622],
   y = [39.685, 39.7624, 39.915] };
  WCL[459] = struct {
   x = [15.3319, 15.4187],
   y = [40.1004, 39.9995] };
  WCL[460] = struct {
   x = [15.6839, 15.7285, 15.7379, 15.8036, 15.9045, 16.0031, 16.0359, 16.0805, 16.1345, 16.1791, 16.2237, 15.9702, 15.8271, 15.8928, 15.7707, 15.6956, 15.705, 15.7496, 15.7707, 15.7613, 15.8928, 16.0359,
        16.0922, 16.1345, 16.1791, 16.2354, 16.2894, 16.4114, 16.5335, 16.5757, 16.5874, 16.5874, 16.5874, 16.6438, 16.7869, 16.9747, 17.0615, 17.0943, 17.085, 17.0943, 17.0732, 17.0732, 17.0615,
        16.984, 16.808, 16.7189, 16.5992, 16.5217, 16.5335],
   y = [40.0253, 39.9643, 39.8727, 39.685, 39.5395, 39.4127, 39.3095, 39.1546, 39.0607, 38.9316, 38.8542, 38.7603, 38.6829, 38.528, 38.3801, 38.2323, 38.0586, 37.953, 37.9788, 37.946, 37.9366, 37.953,
        37.9882, 38.0751, 38.1455, 38.2323, 38.2933, 38.3801, 38.406, 38.4928, 38.6641, 38.7603, 38.812, 38.88, 38.9152, 38.9927, 38.9927, 39.0513, 39.1288, 39.2508, 39.3095, 39.4456, 39.4644, 39.5324,
        39.6075, 39.7014, 39.7437, 39.7695, 39.915] };
  WCL[461] = struct {
   x = [17.9979, 18.1316, 18.3077, 18.3288, 18.3288],
   y = [40.1262, 39.9643, 39.8962, 39.8798, 39.915] };
  WCL[462] = struct {
   x = [15.6182, 15.5525, 15.5619, 15.5407, 15.4304, 15.2873, 15.1652, 15.0995, 15.0878, 15.1652, 15.177, 15.2216, 15.2638, 15.2098, 15.1535, 15.1324, 15.0104, 14.8578, 14.6466, 14.4495, 14.2946, 14.0411,
        13.8745, 13.7313, 13.5999, 13.391, 13.1704, 12.9498, 12.7832, 12.6401, 12.5415, 12.4852, 12.518, 12.5532, 12.5955, 12.6729, 12.7386, 12.8395, 12.9498, 13.0813, 13.2479, 13.3253, 13.4122,
        13.4896, 13.621, 13.6985, 13.8322, 13.9519, 14.1397, 14.2946, 14.3814, 14.3931, 14.5152, 14.6583, 14.8226, 15.0761, 15.1535, 15.2638, 15.3084, 15.475, 15.6182],
   y = [38.3191, 38.3027, 38.2933, 38.1361, 37.9882, 37.8239, 37.5963, 37.4461, 37.3687, 37.3053, 37.2349, 37.1293, 37.0589, 36.9627, 36.8195, 36.7139, 36.7045, 36.7139, 36.7585, 36.8829, 37.0424, 37.1129,
        37.1199, 37.1833, 37.2889, 37.3687, 37.5001, 37.5775, 37.6033, 37.6315, 37.7348, 37.9366, 38.0586, 38.1103, 38.1713, 38.1971, 38.1713, 38.1009, 38.1103, 38.1807, 38.2417, 38.2065, 38.1713,
        38.1455, 38.1009, 37.9976, 38.007, 38.0399, 38.0586, 38.068, 38.0751, 38.0751, 38.0844, 38.1196, 38.19, 38.19, 38.19, 38.2675, 38.2581, 38.2745, 38.3191] };
  WCL[463] = struct {
   x = [14.372, 14.372, 14.4166, 14.4917, 14.5152, 14.5809, 14.5598, 14.4495, 14.372, 14.372],
   y = [35.9488, 35.9934, 35.9746, 35.9394, 35.8855, 35.8503, 35.8151, 35.8315, 35.9136, 35.9488] };
  WCL[464] = struct {
   x = [9.87313, 10.028, 10.1923, 10.1712, 10.2252, 10.3026, 10.2815, 10.2815, 10.4692, 10.5467, 10.6781, 10.7884, 10.887, 11.0207, 11.1193, 11.0207, 10.9104, 10.6898, 10.5467, 10.4904, 10.5561, 10.6452,
        10.833, 10.887, 11.0301, 11.0536, 11.0747, 11.131, 11.0536, 10.9644, 10.8541, 10.7227, 10.6124, 10.4904, 10.3918, 10.2815, 10.1266, 10.0937, 10.082, 10.082],
   y = [37.3311, 37.2889, 37.2443, 37.2091, 37.1129, 36.9979, 36.8922, 36.8383, 36.7773, 36.8383, 36.8735, 36.9533, 37.0683, 37.0683, 36.8922, 36.7491, 36.5543, 36.4839, 36.3853, 36.1812, 35.9582, 35.8503,
        35.8221, 35.7141, 35.6343, 35.5264, 35.3997, 35.2659, 35.1392, 34.9491, 34.7848, 34.6674, 34.5853, 34.5032, 34.4656, 34.3741, 34.3295, 34.2192, 34.1371, 34.0925] };
  WCL[465] = struct {
   x = [9.92946, 10.028, 10.1266, 10.2369, 10.3472, 10.4458, 10.5021, 10.5795, 10.7438, 10.833, 10.8987, 10.9644, 10.9644, 11.0301, 11.1522, 11.2179, 11.3164],
   y = [33.7944, 33.7498, 33.6841, 33.6489, 33.6489, 33.6747, 33.5832, 33.5105, 33.5926, 33.5926, 33.4729, 33.3157, 33.2617, 33.196, 33.196, 33.2148, 33.1584] };
  WCL[466] = struct {
   x = [10.7767, 10.7438, 10.7884, 10.8213, 10.8213, 10.7321, 10.6124, 10.5889, 10.6335, 10.711, 10.7767],
   y = [33.8789, 33.9047, 33.8507, 33.7944, 33.7029, 33.6935, 33.7029, 33.832, 33.8695, 33.9047, 33.8789] };
  WCL[467] = struct {
   x = [11.3164, 11.5159, 11.7483, 12.0463, 12.2434, 12.6283, 12.9944, 13.269, 13.4779, 13.7219, 14.0294, 14.1843, 14.5692, 14.8015, 15.0432, 15.1535, 15.2216, 15.231, 15.407, 15.6839, 16.0805, 16.334,
        16.7189, 17.0075, 17.2938, 17.5356, 17.667, 17.8007, 17.9439, 18.0988, 18.2091, 18.3523, 18.5283, 18.6386, 18.8028, 19.0469, 19.2229, 19.3989, 19.5656, 19.751, 19.8519, 19.9716, 19.9833,
        19.8941, 19.8847, 19.7744, 19.7627, 19.7627, 19.7956, 19.8284, 19.9387],
   y = [33.1584, 33.1115, 33.0012, 32.9003, 32.8252, 32.8064, 32.8909, 32.8721, 32.7876, 32.7782, 32.6938, 32.6116, 32.4614, 32.4145, 32.33, 32.2455, 32.0507, 31.8536, 31.6166, 31.3631, 31.2293, 31.2293,
        31.2012, 31.1355, 31.051, 30.9735, 30.9078, 30.8233, 30.8046, 30.7459, 30.6239, 30.5089, 30.3751, 30.3375, 30.2789, 30.2789, 30.3469, 30.4032, 30.4995, 30.6708, 30.8421, 30.9947, 31.1261,
        31.3162, 31.4758, 31.6259, 31.7597, 31.8724, 32.0038, 32.0695, 32.1728] };
  WCL[468] = struct {
   x = [0.403656, 0.258152, 0.072752],
   y = [40.1802, 39.9854, 39.8164] };
  WCL[469] = struct {
   x = [-0.082139, 0.004694, 0.105608, 0.248765, 0.314476, 0.225296],
   y = [39.0138, 38.9363, 38.8918, 38.8659, 38.7815, 38.6524] };
  WCL[470] = struct {
   x = [1.5935, 1.62636, 1.63809, 1.68033, 1.57003, 1.49493, 1.36116, 1.42688, 1.5935],
   y = [39.1077, 39.0983, 39.117, 39.0044, 38.9105, 38.8659, 38.927, 39.0302, 39.1077] };
  WCL[471] = struct {
   x = [3.29026, 3.29026, 3.28088, 3.24802, 3.25741, 3.4123, 3.50148, 3.4123, 3.21282, 3.09313, 2.9265, 2.86079, 2.71764, 2.61907, 2.50877, 2.68478, 2.97109, 3.17057, 3.28088, 3.302, 3.29026],
   y = [39.9596, 39.9337, 39.9173, 39.8328, 39.7742, 39.7742, 39.6639, 39.4573, 39.2954, 39.3564, 39.4339, 39.5442, 39.5348, 39.4925, 39.5958, 39.7554, 39.8751, 39.9337, 39.9854, 39.9854, 39.9596] };
  WCL[472] = struct {
   x = [4.32756, 4.3393, 4.34868, 4.34868, 4.25012, 4.08584],
   y = [40.0276, 39.9854, 39.9009, 39.8586, 39.8586, 39.9103] };
  WCL[473] = struct {
   x = [9.67365, 9.67365, 9.67365, 9.66427, 9.65253, 9.65253, 9.59856, 9.58682, 9.48825, 9.37795, 9.24418, 9.14562, 9.06817, 8.94614, 8.87104, 8.77012, 8.68329, 8.65043, 8.62697, 8.56125, 8.48381, 8.45095,
        8.46269, 8.40636, 8.42983, 8.43922, 8.45095, 8.46269, 8.5284, 8.54013, 8.49554],
   y = [40.1426, 39.9901, 39.7859, 39.6263, 39.4878, 39.4034, 39.3095, 39.2414, 39.1734, 39.1898, 39.225, 39.1546, 39.0185, 38.941, 38.9152, 38.8988, 38.9669, 39.0185, 39.0349, 39.103, 39.1898, 39.3095,
        39.394, 39.5066, 39.6169, 39.6427, 39.7272, 39.7953, 39.8469, 39.9314, 39.9736] };
  WCL[474] = struct {
   x = [-0.061018, 0.138463, 0.248765, 0.380187, 0.490489, 0.64538, 0.776802, 0.833126, 1.00914, 1.32831, 1.54891, 1.83523, 2.24357, 2.50877, 2.73876, 2.91712, 3.09313, 3.21282, 3.29026, 3.40056, 3.43342,
        3.48974, 3.53433, 3.69861, 3.86524, 3.98493, 4.12808, 4.28297, 4.42613, 4.54817, 4.70306, 4.84621, 4.9659, 5.00111],
   y = [35.8456, 35.916, 36.0591, 36.1741, 36.2281, 36.2563, 36.3454, 36.39, 36.4417, 36.5308, 36.5848, 36.5942, 36.62, 36.6458, 36.6388, 36.7186, 36.8336, 36.7984, 36.7984, 36.8148, 36.8148, 36.7984,
        36.789, 36.8148, 36.8852, 36.9298, 36.9204, 36.9204, 36.9298, 36.9392, 36.9204, 36.9204, 36.9134, 36.8993] };
  WCL[475] = struct {
   x = [5.00111, 5.04335, 5.09967, 5.16538, 5.36486, 5.562, 5.72863, 6.02433, 6.22381, 6.3787, 6.5993, 6.75419, 6.85276, 7.01704, 7.22825, 7.32682, 7.55681, 7.69997, 7.85486, 8.07546, 8.21862, 8.34065,
        8.48381, 8.58238, 8.71615],
   y = [36.8993, 36.857, 36.8289, 36.7773, 36.7233, 36.7679, 36.8383, 36.8829, 36.9533, 37.1293, 37.0683, 36.9885, 36.9885, 36.9345, 36.9979, 37.1129, 37.0424, 37.0072, 36.8993, 36.9275, 36.9791, 36.9533,
        36.9439, 36.9533, 36.9791] };
  WCL[476] = struct {
   x = [8.71615, 8.79359, 8.84757, 9.00246, 9.23245, 9.43193, 9.67365, 9.87313],
   y = [36.9791, 37.0072, 37.0683, 37.1551, 37.2537, 37.3147, 37.3405, 37.3311] };
  WCL[477] = struct {
   x = [10.082, 9.8614, 9.8614, 9.8614, 9.8614, 9.88487, 9.92946],
   y = [34.0925, 34.1371, 34.0714, 34.0714, 33.9704, 33.9141, 33.7944] };
  WCL[478] = struct {
   x = [-5.50802, -5.50802],
   y = [50.0439, 49.9946] };
  WCL[479] = struct {
   x = [-5.09967, -5.04335],
   y = [50.0298, 49.9876] };
  WCL[480] = struct {
   x = [-8.6387, -8.59411, -8.54013, -8.48381, -8.49554, -8.51666, -8.54952, -8.59411, -8.65043, -8.70441, -8.72553],
   y = [41.8229, 41.6493, 41.3935, 41.2104, 41.086, 40.9617, 40.8091, 40.6519, 40.5251, 40.2975, 40.0863] };
  WCL[481] = struct {
   x = [-1.69207, -1.86808, -2.19898, -2.45244, -2.82794, -2.99221, -3.34659, -3.55545, -3.79718, -4.15155, -4.47072, -4.86734, -5.0762, -5.28742, -5.62771, -5.84831, -6.03606, -6.41155, -6.78705, -7.13907,
        -7.41365, -7.6014, -7.77741, -8.02148, -8.14117, -8.13178, -8.27494, -8.61523, -8.82645, -9.03532, -9.12449, -8.9696, -8.83583, -8.8593, -8.73727, -8.65043, -8.77012, -8.60584, -8.65982,
        -8.50728, -8.62697, -8.69268, -8.68329, -8.6387],
   y = [43.3766, 43.3367, 43.3437, 43.4094, 43.4, 43.393, 43.4329, 43.4798, 43.4329, 43.4164, 43.4329, 43.4892, 43.5291, 43.5361, 43.6089, 43.6324, 43.562, 43.5526, 43.5526, 43.6018, 43.6957, 43.7286,
        43.7051, 43.6558, 43.5197, 43.4235, 43.3202, 43.3108, 43.2639, 43.1513, 42.9729, 42.8673, 42.7945, 42.6819, 42.6256, 42.6326, 42.4542, 42.3956, 42.2829, 42.2994, 42.175, 42.013, 41.8887,
        41.8229] };
  WCL[482] = struct {
   x = [-1.7038, -1.51605, -1.32831, -1.25086, -1.20627, -1.12883, -1.05373, -0.919959, -0.87537, -1.00914, -1.02087, -0.997405, -0.943428, -0.931694, -0.87537, -0.722825, -0.612524, -0.5562, -0.612524,
        -0.755681, -0.943428, -1.00914, -0.910572, -0.910572, -0.985671, -0.985671, -1.02087, -1.03026, -1.19689, -1.42688, -1.64748, -1.8141, -1.94553, -1.91267, -1.93614, -1.99012, -1.75778,
        -1.62636, -1.8141, -2.12154, -2.34214, -2.33275, -2.30929, -2.36561, -2.69651, -2.60733, -2.73876, -2.95936, -3.06027, -3.0814, -3.11425, -3.28088, -3.45689, -3.73147, -3.84177, -4.00839,
        -4.20553, -4.30644, -4.45899, -4.53643, -4.4144, -4.26185, -4.29471, -4.43787, -4.47072, -4.39328, -4.29471, -4.26185, -4.43787, -4.635, -4.635, -4.56929, -4.36042, -4.16094, -4.04125,
        -3.86524, -3.71035, -3.58831, -3.44515, -3.42403, -3.32312, -3.24802, -3.13772, -2.9265, -2.8162, -2.65192, -2.47591, -2.2999, -2.14501, -2.10042, -1.99012, -1.85869, -1.71554, -1.63809,
        -1.57003, -1.36116, -1.28372, -1.40575, -1.45973, -1.45973, -1.51605, -1.56065, -1.68033, -1.73666, -1.75778, -1.76951, -1.5935, -1.34943, -1.20627, -1.14056, -1.1523, -1.16403, -1.05373,
        -0.943428, -0.842514, -0.5562, -0.302742, -0.105608],
   y = [43.3672, 43.3601, 43.4493, 43.6018, 43.8013, 44.0712, 44.388, 44.6157, 44.6556, 44.7424, 44.7729, 44.9536, 45.2024, 45.4206, 45.5216, 45.4042, 45.3127, 45.2493, 45.4206, 45.5521, 45.6366, 45.7609,
        45.8994, 46.0684, 46.174, 46.2984, 46.289, 46.2984, 46.3195, 46.404, 46.5096, 46.6551, 46.8217, 46.9578, 47.101, 47.1902, 47.2207, 47.2348, 47.2887, 47.2793, 47.2887, 47.399, 47.4976, 47.5187,
        47.5187, 47.5633, 47.5938, 47.5422, 47.5633, 47.6009, 47.6314, 47.7135, 47.7581, 47.8168, 47.8543, 47.8473, 47.7863, 47.8919, 47.9881, 48.0327, 48.0679, 48.1054, 48.1641, 48.1571, 48.2157,
        48.2885, 48.2955, 48.3565, 48.3401, 48.326, 48.3917, 48.5091, 48.5818, 48.6335, 48.6475, 48.671, 48.671, 48.6992, 48.7203, 48.786, 48.8236, 48.8001, 48.8165, 48.8165, 48.7649, 48.6475, 48.5607,
        48.6053, 48.6335, 48.6194, 48.6123, 48.6546, 48.6781, 48.6546, 48.6053, 48.6123, 48.671, 48.7203, 48.8658, 49.0113, 49.1333, 49.2929, 49.3516, 49.4525, 49.5957, 49.6942, 49.659, 49.6661,
        49.6731, 49.652, 49.6097, 49.544, 49.4666, 49.3586, 49.3657, 49.3375, 49.2929, 49.2929] };
  WCL[483] = struct {
   x = [-60.0719, -59.9405, -59.8396, -59.7739, -59.7622, -59.8631, -59.8396, -59.8396, -59.9499],
   y = [45.8595, 45.937, 45.9768, 45.998, 46.0449, 46.0684, 46.1599, 46.228, 46.2655] };
  WCL[484] = struct {
   x = [-56.2888, -56.3334, -56.3334, -56.3006, -56.2888, -56.2795, -56.2137, -56.2231, -56.2137, -56.2888],
   y = [47.1385, 47.1221, 47.0846, 46.9578, 46.8663, 46.8053, 46.8522, 46.9883, 47.0846, 47.1385] };
  WCL[485] = struct {
   x = [-57.6359, -57.7228, -57.7556, -57.8448, -57.8565, -57.8237, -57.8237, -57.9668, -58.1311, -58.1968, -58.1522, -58.0208, -58.0208, -58.0654, -58.0208, -58.0865, -58.2531, -58.3517, -58.4503, -58.6052,
        -58.7366, -58.814, -58.9689, -59.1238, -59.1238, -58.9361, -58.7037, -58.5183, -58.4409, -58.6826, -58.9807, -59.2881, -59.3116, -59.2224, -58.9807, -58.6169, -58.2743, -58.0537, -57.8894,
        -57.657, -57.4247, -57.2604, -57.0164, -56.73, -56.4766, -56.324, -56.1457, -56.026, -55.8828, -55.7397, -55.7819, -55.7819, -56.0354, -56.0917, -55.8922, -55.8594, -55.7819, -55.6622,
        -55.4628, -55.3736, -55.275, -55.1225, -54.8573, -54.7798, -55.0215, -55.2093, -55.3313, -55.4299, -55.6834, -55.8265, -55.8594, -55.6622, -55.4088, -55.221, -55.1107, -55.045, -54.923,
        -54.7798, -54.6132, -54.4935, -54.4607, -54.4372, -54.3832, -54.2823, -54.1837, -54.0288, -54.0077, -53.9303, -53.8857, -54.0406, -54.0852, -53.942, -53.7543, -53.5876, -53.5008, -53.5782,
        -53.5548, -53.5665, -53.4679, -53.2802, -53.0479, -52.9164, -52.8273, -52.6958, -52.6958, -52.7944, -52.9164, -53.0385, -53.1253, -53.137, -53.0479, -52.961, -52.8718, -52.8718, -53.0479,
        -53.2473, -53.4116, -53.5008, -53.6111, -53.7097, -53.766, -53.7097, -53.6557, -53.6885, -53.7214, -53.8646, -53.7097, -53.4891, -53.3576, -53.1816, -53.0713, -53.1042, -53.2802, -53.4234,
        -53.5876, -53.6885, -53.9514, -54.0077, -53.8857, -53.7543, -53.8317, -53.942, -53.82, -53.6228, -53.4891, -53.5125, -53.4351, -53.4234, -53.6111, -53.9631, -54.294, -54.4372, -54.571,
        -54.8127, -55.1319, -55.3196, -55.1553, -55.1319, -55.3196, -55.3642, -55.4745, -55.6294, -55.8594, -55.85, -55.9368, -56.0354, -55.8148, -55.4839, -55.6294, -55.85, -56.08],
   y = [50.0369, 49.9289, 49.8515, 49.7459, 49.6238, 49.5792, 49.53, 49.53, 49.4009, 49.2929, 49.2436, 49.2272, 49.192, 49.1427, 49.0629, 49.0629, 49.0841, 49.0911, 48.9315, 48.6616, 48.6053, 48.6851,
        48.6335, 48.5537, 48.5091, 48.5232, 48.5396, 48.5161, 48.4293, 48.2885, 48.1125, 47.8919, 47.7135, 47.5938, 47.5633, 47.6243, 47.6525, 47.683, 47.6619, 47.6454, 47.6454, 47.5938, 47.5868,
        47.5797, 47.6243, 47.6314, 47.6525, 47.7417, 47.8027, 47.8473, 47.7135, 47.6314, 47.5563, 47.4906, 47.4976, 47.453, 47.46, 47.4741, 47.4812, 47.5797, 47.6689, 47.6243, 47.6243, 47.5868,
        47.4741, 47.3544, 47.2207, 47.1291, 47.0846, 47.04, 46.9203, 46.8827, 46.9203, 46.9719, 47.0775, 47.1667, 47.2582, 47.3709, 47.4155, 47.446, 47.4295, 47.4225, 47.5351, 47.7065, 47.9271,
        47.9271, 47.7863, 47.5633, 47.3544, 47.101, 46.8733, 46.8358, 47.0329, 47.1831, 47.1221, 46.9719, 46.8522, 46.7302, 46.6175, 46.7302, 46.6926, 46.9273, 47.2207, 47.446, 47.7135, 47.7511,
        47.6079, 47.46, 47.5351, 47.7205, 47.9059, 48.0327, 48.0984, 48.1946, 48.1054, 48.0233, 47.8238, 47.6619, 47.6525, 47.6994, 47.7863, 47.8754, 48.0327, 48.0468, 48.1125, 48.2157, 48.2533,
        48.3049, 48.4293, 48.4645, 48.5959, 48.671, 48.6053, 48.5748, 48.5161, 48.5232, 48.4504, 48.502, 48.6053, 48.7508, 48.8165, 48.8306, 48.9972, 49.1333, 49.3, 49.3586, 49.2718, 49.2859, 49.3727,
        49.4361, 49.4079, 49.5088, 49.4736, 49.3798, 49.2788, 49.2718, 49.4948, 49.5722, 49.5018, 49.4361, 49.4807, 49.4877, 49.5018, 49.652, 49.6802, 49.6942, 49.8163, 49.9219, 49.9665, 49.9805,
        49.9876] };
  WCL[486] = struct {
   x = [-56.1692, -56.3663, -56.5775, -56.7207, -56.7746, -56.7981, -56.7207],
   y = [50.1284, 49.936, 49.8022, 49.6379, 49.7224, 49.7811, 49.9148] };
  WCL[487] = struct {
   x = [-66.7323, -66.8519, -67.0186, -67.1171, -67.1828, -67.2603, -67.3143, -67.5583, -67.7883, -68.0089, -68.1638, -68.2202, -68.2413, -68.3305, -68.5393, -68.6707, -68.8468, -68.9571, -69.0462, -69.1894,
        -69.3654, -69.5649, -69.6517, -69.762, -69.7409, -69.8958],
   y = [50.0721, 49.9805, 49.8233, 49.6168, 49.5018, 49.3798, 49.3586, 49.3516, 49.3446, 49.3281, 49.2272, 49.192, 49.1333, 49.1333, 49.0043, 48.9174, 48.8588, 48.7438, 48.6546, 48.5091, 48.326, 48.2157,
        48.2016, 48.2016, 48.0022, 47.7863] };
  WCL[488] = struct {
   x = [-70.1164, -69.9286, -69.762, -69.6306, -69.532, -69.356, -69.1354, -68.9477, -68.8468, -68.7717, -68.6707, -68.5393, -68.3398, -68.1192, -67.9338, -67.7226, -67.5349, -67.272, -67.0725, -66.7416,
        -66.5445, -66.3122, -65.993, -65.7607, -65.5518, -65.2983, -65.0214, -64.8337, -64.5474, -64.3056, -64.2071, -64.1742, -64.2493, -64.3056, -64.3385, -64.2282, -64.2282, -64.4159, -64.6577,
        -64.8783, -65.1763, -65.5072, -65.7935, -65.9601, -66.1808, -66.4553, -66.521, -66.2558, -65.9813, -65.7724, -65.6832, -65.5964, -65.3406, -65.0543, -64.9886, -65.0003, -64.8454, -64.7468,
        -64.7586, -64.8454, -64.9228, -65.066, -65.2209, -65.2209, -65.066, -64.9346, -64.8008, -64.8125, -64.8337, -64.8008, -64.714, -64.6577, -64.5708, -64.4488, -64.4699, -64.3713, -64.1625,
        -63.9747, -63.7752, -63.6978, -63.8644, -63.8855, -63.733, -63.5218, -63.2355, -63.0595, -62.8717, -62.7286, -62.6628, -62.6183, -62.5314, -62.2662, -62.0667, -61.9025, -61.8133, -61.8461,
        -61.7358, -61.5716, -61.3838, -61.2524, -61.2524, -61.1843, -61.2946, -61.3181, -61.1961, -61.0318, -60.9543, -61.074, -61.2407, -61.4049, -61.5481, -61.6701, -61.7804, -61.9564, -62.1559,
        -62.3202, -62.4094, -62.454, -62.6183, -62.7403, -62.806, -62.928, -63.0055, -63.0383, -63.1041, -63.1698, -63.2144, -63.3012, -63.435, -63.5335, -63.5007, -63.4795, -63.5007, -63.7213,
        -63.8433, -63.8527, -63.9419, -64.0076, -64.0639, -64.1625, -64.1742, -64.1836, -64.2071, -64.2939, -64.3925, -64.5145, -64.6248, -64.6811, -64.7586, -64.8783, -64.9886, -65.0989, -65.1317,
        -65.2326, -65.2538, -65.2655, -65.3641, -65.4415, -65.6292, -65.7161, -65.8264, -65.9695, -66.0587, -66.0798, -66.0352, -65.9249, -65.817, -65.9038, -66.1033, -66.0587, -65.7818, -65.6504,
        -65.4744, -65.4861, -65.2983, -65.0543, -64.8243, -64.6365, -64.3831, -64.3056, -64.2728, -64.1061, -64.0522, -63.9301, -63.7541, -63.5124, -63.3247, -63.3012, -63.5781, -63.8433, -64.0968,
        -64.3056, -64.538, -64.7351, -64.768, -64.5262, -64.3056, -64.3596, -64.4488, -64.5028, -64.5708, -64.6694, -64.8243, -64.8783, -64.89, -64.9886, -65.242, -65.4626, -65.6386, -65.8264,
        -65.9601, -66.1033, -66.2347, -66.4014, -66.5117, -66.6313, -66.7416, -66.8402, -66.9622, -67.028],
   y = [47.446, 47.5563, 47.683, 47.8473, 47.9951, 48.0913, 48.2298, 48.3636, 48.4152, 48.4293, 48.4363, 48.5326, 48.6194, 48.6781, 48.7578, 48.8517, 48.9174, 48.9831, 49.0395, 49.1498, 49.192, 49.2366,
        49.2788, 49.2718, 49.3, 49.2859, 49.2648, 49.2202, 49.1263, 49.0113, 48.8893, 48.8306, 48.8588, 48.8728, 48.8447, 48.7508, 48.5607, 48.4363, 48.319, 48.1946, 48.0913, 48.1054, 48.2016, 48.1782,
        48.15, 48.0843, 48.0468, 48.0538, 47.9951, 47.9435, 47.7863, 47.7276, 47.8168, 47.92, 47.8989, 47.8473, 47.8308, 47.7722, 47.6384, 47.4906, 47.3404, 47.2348, 47.1456, 47.1151, 47.1291, 47.1151,
        47.101, 46.9719, 46.8428, 46.7466, 46.6316, 46.5096, 46.4181, 46.35, 46.2984, 46.2819, 46.2749, 46.2209, 46.1599, 46.113, 46.0449, 45.9675, 45.9299, 45.9135, 45.8384, 45.822, 45.8055, 45.7844,
        45.6999, 45.6999, 45.6999, 45.7304, 45.829, 45.9064, 45.7844, 45.6671, 45.6835, 45.6671, 45.6905, 45.5826, 45.7304, 45.5216, 45.4676, 45.3972, 45.3901, 45.3808, 45.3432, 45.2798, 45.2892,
        45.2188, 45.2493, 45.1883, 45.132, 45.071, 45.024, 44.9466, 44.9372, 44.8668, 44.8738, 44.8128, 44.7729, 44.7729, 44.7565, 44.78, 44.7565, 44.7424, 44.7096, 44.672, 44.7189, 44.7424, 44.6626,
        44.5687, 44.5054, 44.5054, 44.6157, 44.679, 44.5852, 44.5054, 44.6157, 44.4913, 44.4185, 44.3787, 44.3082, 44.2684, 44.205, 44.0947, 43.9914, 43.921, 43.8882, 43.8483, 43.6957, 43.7356, 43.745,
        43.7051, 43.6159, 43.569, 43.5056, 43.5056, 43.5526, 43.7286, 43.8154, 43.8248, 44.0149, 44.205, 44.3646, 44.4983, 44.6086, 44.5382, 44.442, 44.5523, 44.7025, 44.6861, 44.7565, 44.8269,
        44.9137, 45.0475, 45.1484, 45.2329, 45.3197, 45.2658, 45.1719, 45.1085, 45.1648, 45.2423, 45.3127, 45.3432, 45.3502, 45.4277, 45.4347, 45.3972, 45.4277, 45.4582, 45.3808, 45.3573, 45.5051,
        45.6366, 45.8384, 45.8454, 45.829, 45.9299, 45.8384, 45.707, 45.6765, 45.6671, 45.66, 45.599, 45.4746, 45.3502, 45.2892, 45.2728, 45.2963, 45.1883, 45.1484, 45.132, 45.1249, 45.1085, 45.1179,
        45.1648, 45.1954, 45.1954] };
  WCL[489] = struct {
   x = [-63.9747, -64.0076, -63.8433, -63.5335, -63.2683, -63.0055, -62.7614, -62.5525, -62.3648, -62.2334, -62.1231, -62.0456, -61.9799, -61.879, -61.825, -61.8039, -61.7358, -61.703, -61.8368, -62.0667,
        -62.2662, -62.5314, -62.7731, -62.9703, -63.1698, -63.3904, -63.5453, -63.611, -63.7541, -63.9184, -64.0639, -64.1953, -64.3174, -64.4159, -64.4277, -64.1953, -63.9747],
   y = [49.9594, 49.943, 49.936, 49.8444, 49.8233, 49.7529, 49.6942, 49.6379, 49.53, 49.4431, 49.422, 49.422, 49.3938, 49.4009, 49.3516, 49.2929, 49.2366, 49.1638, 49.1122, 49.1122, 49.1052, 49.1568,
        49.1709, 49.2436, 49.2788, 49.3516, 49.4596, 49.5957, 49.645, 49.6942, 49.7529, 49.7881, 49.8304, 49.8796, 49.95, 49.9805, 49.9594] };
  WCL[490] = struct {
   x = [-63.9536, -63.9184, -63.9301, -63.9419, -63.9301, -63.9301, -63.9958, -64.0287, -63.9419, -63.8855, -63.8198, -63.7095, -63.611, -63.4115, -63.2589, -63.0923, -62.9046, -62.6628, -62.5971, -62.4634,
        -62.2428, -62.0222, -62.0128, -62.2216, -62.2662, -62.3108, -62.4094, -62.4211, -62.4211, -62.3977, -62.5971, -62.8177, -62.8717, -62.8834, -63.0149, -63.0806, -63.0923, -63.2589, -63.4561,
        -63.6556, -63.6978, -63.7213, -63.9184, -64.0522, -64.0287, -64.1625, -64.2822, -64.2939, -64.1061, -63.963, -63.9536],
   y = [46.9883, 47.101, 47.101, 47.101, 47.0329, 46.9344, 46.8593, 46.8053, 46.7372, 46.648, 46.58, 46.533, 46.5706, 46.5495, 46.5096, 46.4814, 46.4181, 46.4486, 46.465, 46.4955, 46.4885, 46.4814, 46.411,
        46.3805, 46.3289, 46.289, 46.2514, 46.1224, 46.0825, 46.0449, 46.0074, 46.0684, 46.1294, 46.2209, 46.2819, 46.2655, 46.1975, 46.2045, 46.2514, 46.3124, 46.3805, 46.4181, 46.4345, 46.5401,
        46.6175, 46.648, 46.6926, 46.8287, 46.9578, 47.054, 46.9883] };
  WCL[491] = struct {
   x = [-60.546, -60.6, -60.6234, -60.7783, -60.8769, -60.9872, -61.074, -61.175, -61.3181, -61.3275, -61.4049, -61.4049, -61.3181, -61.2524, -61.074, -60.9872, -60.8652, -60.8112, -60.6, -60.4686, -60.3583,
        -60.248, -60.1494, -60.0719],
   y = [46.9954, 47.04, 46.9203, 46.7842, 46.6856, 46.4885, 46.3899, 46.3054, 46.174, 46.0825, 45.9135, 45.8149, 45.7375, 45.6436, 45.599, 45.6131, 45.66, 45.6671, 45.653, 45.66, 45.6999, 45.7375, 45.7844,
        45.8595] };
  WCL[492] = struct {
   x = [-59.9499, -60.0062, -60.0508, -60.1377, -60.1705, -60.194, -60.3137, -60.4028, -60.5225, -60.6, -60.5014, -60.4122, -60.4357, -60.5131, -60.3794, -60.3254, -60.3043, -60.3254, -60.3465, -60.3911,
        -60.5014, -60.546],
   y = [46.2655, 46.2984, 46.2514, 46.1364, 46.2514, 46.2514, 46.2045, 46.1599, 46.1224, 46.1435, 46.2514, 46.3289, 46.3359, 46.3594, 46.5706, 46.7372, 46.8593, 46.9038, 46.9789, 47.0095, 47.04, 46.9954] };
  WCL[493] = struct {
   x = [-67.028, -67.028, -67.0397, -67.028, -67.028, -67.028, -67.0725, -67.0937, -67.1054, -67.0514, -66.974, -66.9505, -66.9622, -67.028, -67.1383, -67.2157, -67.2931, -67.4152, -67.4809, -67.6029,
        -67.6569, -67.7132, -67.7883, -67.8235, -67.9221, -67.9761, -68.0207, -68.0653, -68.131, -68.1638, -68.2202, -68.3187, -68.3398, -68.3844, -68.483, -68.5182, -68.6379, -68.6707, -68.6942,
        -68.7153, -68.8045, -68.9031, -68.9359, -69.0016, -69.0251, -69.058, -69.0908, -69.1448, -69.1777, -69.2668, -69.3443, -69.41, -69.4546, -69.5414, -69.5649, -69.586, -69.6306, -69.6517,
        -69.6869, -69.7198, -69.7526, -69.7738, -69.7972, -69.8395, -69.9075],
   y = [45.1954, 45.1719, 45.1484, 45.078, 45.024, 44.9771, 44.9372, 44.9231, 44.8903, 44.8903, 44.8903, 44.8527, 44.7729, 44.726, 44.679, 44.6861, 44.6485, 44.6392, 44.5852, 44.5617, 44.5687, 44.5687,
        44.5453, 44.4819, 44.442, 44.435, 44.4514, 44.4279, 44.3716, 44.3646, 44.2918, 44.2777, 44.3012, 44.3317, 44.3787, 44.3411, 44.3552, 44.3787, 44.442, 44.5288, 44.5148, 44.4584, 44.3247,
        44.2378, 44.1416, 44.0876, 44.0313, 44.0149, 44.0313, 44.0477, 43.9914, 43.8882, 43.8882, 43.8717, 43.8952, 43.9445, 43.9351, 43.8483, 43.8013, 43.752, 43.7356, 43.7849, 43.8412, 43.8717,
        43.8717] };
  WCL[494] = struct {
   x = [-70.0272, -69.9944, -69.9826, -69.9826, -69.9826, -69.9944, -69.9826, -69.9286, -69.8958, -69.8723, -69.9075, -69.9826],
   y = [41.7643, 41.7971, 41.8136, 41.8652, 41.9215, 41.9708, 41.9966, 41.9544, 41.8394, 41.7408, 41.715, 41.6915] };
  WCL[495] = struct {
   x = [-69.8958, -70.1375, -70.2267, -70.4027, -70.6233, -70.8111, -70.9542, -71.0974, -71.0739, -70.865, -70.6773, -70.6679, -70.6562, -70.5576, -70.4567, -70.2924, -70.1164],
   y = [47.7863, 47.4976, 47.4906, 47.3638, 47.2207, 47.0939, 46.9789, 46.8898, 46.8593, 46.9038, 46.9883, 46.9883, 47.0329, 47.0705, 47.1597, 47.2723, 47.446] };
  WCL[496] = struct {
   x = [-69.9075, -70.0601, -70.1704, -70.2032, -70.215, -70.2267, -70.2689, -70.2924, -70.3041, -70.3253, -70.4027, -70.4802, -70.4895, -70.513, -70.5459, -70.6116, -70.6679, -70.7336, -70.7665, -70.7782,
        -70.7782, -70.7782, -70.7782, -70.7547, -70.7101, -70.6773, -70.7547, -70.7993, -70.8204, -70.865, -70.9425, -70.9425, -70.8533, -70.7782, -70.7219, -70.7008, -70.689, -70.6679, -70.6679,
        -70.6116, -70.5787, -70.5459, -70.5247, -70.4473, -70.3464, -70.2267, -70.1164, -70.0272],
   y = [43.8717, 43.8248, 43.7286, 43.6723, 43.6253, 43.5854, 43.5291, 43.5197, 43.5197, 43.4962, 43.393, 43.3108, 43.3108, 43.2874, 43.2076, 43.1184, 43.0222, 42.9166, 42.7945, 42.7781, 42.7382, 42.7312,
        42.7147, 42.6725, 42.6561, 42.6162, 42.5998, 42.5106, 42.4613, 42.4707, 42.3815, 42.3228, 42.2829, 42.2571, 42.1914, 42.1844, 42.1445, 42.0952, 42.0295, 42.0037, 41.9872, 41.9145, 41.8394,
        41.7971, 41.7643, 41.7478, 41.7408, 41.7643] };
  WCL[497] = struct {
   x = [-69.9826, -70.0929, -70.161, -70.2267, -70.3464, -70.4567, -70.5576, -70.5905, -70.5905, -70.5998, -70.6444, -70.743, -70.8439, -70.8885, -70.9636, -71.0199, -71.0856, -71.142, -71.142, -71.1631,
        -71.2194, -71.2734, -71.318, -71.3508, -71.3837, -71.4823, -71.6254, -71.7263, -71.7474, -71.7474, -71.846, -72.0126, -72.2215, -72.5407, -72.7167, -72.827, -73.0147, -73.0711, -73.1814,
        -73.3245, -73.3785, -73.4888, -73.6226, -73.7211, -73.7868, -73.8103, -73.8314, -73.8432, -73.8971, -73.9746, -74.0192, -74.0849, -74.1177, -74.1835, -74.2163, -74.2163, -74.1741, -74.0638,
        -73.9746, -73.9417, -73.9863, -74.0192],
   y = [41.6915, 41.6422, 41.6258, 41.6422, 41.6094, 41.5836, 41.5836, 41.6094, 41.6422, 41.6821, 41.6915, 41.6587, 41.6094, 41.5836, 41.5578, 41.5578, 41.5578, 41.5578, 41.6258, 41.6986, 41.7408, 41.6986,
        41.6422, 41.5671, 41.485, 41.4263, 41.3935, 41.377, 41.377, 41.3677, 41.3512, 41.3512, 41.3348, 41.302, 41.2926, 41.3278, 41.2597, 41.1846, 41.1682, 41.1353, 41.1025, 41.086, 41.0109, 40.9429,
        40.91, 40.903, 40.8936, 40.91, 40.91, 40.8443, 40.7856, 40.7434, 40.7105, 40.6261, 40.558, 40.4993, 40.4759, 40.4571, 40.3491, 40.2482, 40.0863, 40.0018] };
  WCL[498] = struct {
   x = [-74.0403, -74.0732],
   y = [39.9854, 40.0018] };
  WCL[499] = struct {
   x = [-75.6948, -75.6502, -75.6385, -75.7817, -75.9811, -76.2135, -76.4435, -76.6547, -76.7744, -76.9081, -76.941, -76.9621, -76.9856, -77.0607, -77.2156, -77.2484, -77.1616, -77.117, -77.0396, -77.0184,
        -77.1827, -77.2602, -77.4596, -77.6802, -77.7225, -77.746, -77.8234, -78.1308, -78.3632, -78.572, -78.8044, -78.9475, -79.1681, -79.2784, -79.4333, -79.4873, -79.5436, -79.5976, -79.6633,
        -79.7525, -79.8065, -79.7736, -79.6633, -79.5319, -79.3887, -79.2338, -79.0132, -78.8818, -78.6941, -78.5063, -78.1754, -78.1097, -77.9877, -77.8445, -77.7553, -77.7225, -77.6568, -77.5699,
        -77.4596, -77.4033, -77.3705, -77.3047, -77.1381, -76.941, -76.7744, -76.6101, -76.4787, -76.312, -76.2792, -76.1689, -76.136, -76.1689, -76.2017, -76.2581, -76.2674, -76.2463, -76.1806,
        -76.1126, -76.1032, -76.1032, -76.1571, -76.2017, -76.2909, -76.312, -76.2463, -76.136, -76.014, -75.8708, -75.7605, -75.6948],
   y = [44.5523, 44.5758, 44.5758, 44.4983, 44.3716, 44.3411, 44.2684, 44.1979, 44.1581, 44.1017, 44.0876, 44.1416, 44.1745, 44.1815, 44.1581, 44.1182, 44.1182, 44.0477, 43.968, 43.9046, 43.8811, 43.9445,
        43.9515, 43.9844, 43.9844, 43.9844, 43.9985, 43.9844, 43.9281, 43.8811, 43.8483, 43.8154, 43.7614, 43.7051, 43.6159, 43.569, 43.5361, 43.4563, 43.4094, 43.3367, 43.2874, 43.2709, 43.2475,
        43.2146, 43.1912, 43.224, 43.2569, 43.2968, 43.3273, 43.3672, 43.393, 43.4, 43.4, 43.3672, 43.3601, 43.3531, 43.3038, 43.2639, 43.2874, 43.3038, 43.3108, 43.2968, 43.2968, 43.3038, 43.3437,
        43.4399, 43.5127, 43.5455, 43.5455, 43.5526, 43.6417, 43.7685, 43.7919, 43.8483, 43.8882, 43.921, 43.9116, 43.9445, 43.9515, 43.9985, 44.0477, 44.0384, 44.0477, 44.1182, 44.1745, 44.2214,
        44.2777, 44.3552, 44.4913, 44.5523] };
  WCL[500] = struct {
   x = [-78.8818, -78.8701, -78.8912, -79.079, -79.4099, -79.7079],
   y = [42.8673, 42.8368, 42.8274, 42.8509, 42.8837, 42.8767] };
  WCL[501] = struct {
   x = [-80.1608, -79.9614, -79.8182, -79.6868, -79.4873, -79.2996, -79.2338, -79.1681, -79.1235, -79.1024, -79.0578, -78.9475, -78.9241, -78.9147, -78.8701, -78.8818],
   y = [42.175, 42.1914, 42.2665, 42.3322, 42.4284, 42.5599, 42.5669, 42.5998, 42.6561, 42.6819, 42.7054, 42.7781, 42.7875, 42.7875, 42.8204, 42.8673] };
  WCL[502] = struct {
   x = [-80.0388, -79.9942, -79.9285, -79.7854, -79.7079, -79.7079, -79.7619, -79.8299, -79.9285],
   y = [44.8738, 44.8363, 44.7964, 44.7964, 44.8269, 44.8433, 44.9231, 44.9536, 44.9771] };
  WCL[503] = struct {
   x = [-72.2004, -72.2872, -72.3975, -72.5735, -72.7402, -72.949, -73.0593, -73.1814, -73.2917, -73.3785, -73.4677, -73.5451, -73.5991, -73.6554, -73.6765, -73.754, -73.7751, -73.8197, -73.8314, -73.8643,
        -73.9206, -73.93, -73.9417, -73.93, -73.9206, -73.8432, -73.8314, -73.754, -73.6554, -73.5662, -73.4677, -73.4231, -73.3363, -73.2259, -73.1368, -73.0053, -72.8833, -72.7941, -72.651, -72.5735,
        -72.4538, -72.4304, -72.3858, -72.3107, -72.1769, -72.0455, -71.9563, -71.9352, -71.9798, -72.0783, -72.1441, -72.2332, -72.299, -72.3858, -72.5078, -72.4867, -72.4093, -72.3318, -72.2544,
        -72.2098, -72.2004],
   y = [41.194, 41.1447, 41.0696, 40.9945, 40.9781, 40.9781, 40.9617, 40.9429, 40.9265, 40.91, 40.91, 40.8866, 40.8607, 40.8514, 40.8349, 40.8185, 40.8091, 40.7598, 40.7434, 40.7434, 40.7176, 40.6683,
        40.6167, 40.6002, 40.6425, 40.6589, 40.6096, 40.5932, 40.6261, 40.6683, 40.6941, 40.6941, 40.7105, 40.727, 40.7434, 40.7763, 40.7927, 40.8021, 40.8279, 40.8279, 40.8678, 40.8772, 40.903, 40.91,
        40.9523, 41.0016, 41.0368, 41.0438, 41.0438, 41.0602, 41.0602, 41.0274, 41.0109, 40.9523, 40.9358, 40.9781, 41.0368, 41.086, 41.1283, 41.1776, 41.194] };
  WCL[504] = struct {
   x = [-82.1322, -82.1768, -82.3528, -82.5194, -82.6508, -82.794, -82.9606, -83.0475, -83.1812, -83.1366, -83.0052, -82.794, -82.6626, -82.5734, -82.388, -82.2425, -82.0219, -81.9468, -81.8248, -81.813,
        -81.7473, -81.7027, -81.6605, -81.6699, -81.759, -81.7919, -81.7919, -81.8459, -81.9022, -81.8693, -81.8693, -81.8693, -81.9233, -82.0454, -82.1111, -82.1322],
   y = [45.8924, 45.8759, 45.9064, 45.8994, 45.8525, 45.8924, 45.9064, 45.9064, 45.8924, 45.822, 45.7844, 45.7445, 45.714, 45.66, 45.653, 45.592, 45.5591, 45.545, 45.5685, 45.5826, 45.653, 45.7304, 45.7914,
        45.8595, 45.8525, 45.7304, 45.6765, 45.6131, 45.6225, 45.6999, 45.7914, 45.8595, 45.9463, 45.9463, 45.8994, 45.8924] };
  WCL[505] = struct {
   x = [-79.7079, -80.0505, -80.2594, -80.3815, -80.5457, -80.7546, -81.1301, -81.4727, -81.5713, -81.6487, -81.7473, -81.7919, -81.8459, -82.0571, -82.1674, -82.3528, -82.4537, -82.4631, -82.5523, -82.74,
        -82.8386, -83.1038, -83.1366, -83.2587, -83.3784, -83.4558, -83.193, -83.0592, -82.9724, -82.8386, -82.794, -82.6743, -82.4537, -82.266, -82.1111, -81.9562, -81.813, -81.6816, -81.4938,
        -81.339, -81.1747, -81.0315, -80.8203, -80.656, -80.4565, -80.3134, -80.1608],
   y = [42.8767, 42.8274, 42.7617, 42.6819, 42.5763, 42.6256, 42.6725, 42.5927, 42.5528, 42.5035, 42.3815, 42.3064, 42.3228, 42.2665, 42.1844, 42.1022, 42.0295, 41.9638, 41.9802, 42.0037, 41.9802, 42.0037,
        41.9966, 41.8887, 41.8065, 41.7244, 41.6422, 41.5578, 41.5343, 41.5413, 41.492, 41.4756, 41.4521, 41.492, 41.5343, 41.5249, 41.5343, 41.593, 41.6821, 41.7572, 41.83, 41.8816, 41.9473, 41.9802,
        42.0365, 42.0788, 42.175] };
  WCL[506] = struct {
   x = [-82.5734, -82.6626, -82.794, -82.9043, -82.9395, -82.9395, -82.9043, -82.8503, -82.8057, -82.7189, -82.6415, -82.5523, -82.4866, -82.4537, -82.442, -82.4631, -82.5077, -82.5523, -82.5734, -82.5734],
   y = [42.6256, 42.6256, 42.6256, 42.5434, 42.4707, 42.3815, 42.3322, 42.3158, 42.3064, 42.2994, 42.3064, 42.3158, 42.3393, 42.405, 42.4449, 42.4871, 42.4941, 42.527, 42.5927, 42.6256] };
  WCL[507] = struct {
   x = [-82.4631, -82.4209, -82.4209, -82.3974, -82.266, -82.1768, -82.1322, -81.9679, -81.813, -81.7473, -81.759, -81.8036, -81.7919, -81.7027, -81.5596, -81.4727, -81.4047, -81.3953, -81.3953, -81.4281,
        -81.461, -81.5056, -81.5502, -81.5596, -81.6253, -81.6699, -81.637, -81.5384, -81.4281, -81.3507, -81.339, -81.3061, -81.2404, -81.2193, -81.1958, -81.1864, -81.1512, -81.0972, -81.0526,
        -81.0315, -81.0409, -81.0526, -81.0644, -81.0644, -81.0315, -80.9869, -80.9541, -80.9541, -80.9306, -80.8438, -80.8109, -80.7663, -80.71, -80.6772, -80.5786, -80.358, -80.182, -80.0928,
        -80.0177, -80.0177, -80.0388, -80.1045, -80.0834, -80.0388],
   y = [43.0621, 43.0949, 43.102, 43.102, 43.0855, 43.1513, 43.224, 43.2311, 43.3601, 43.576, 43.7849, 43.9985, 44.1581, 44.2613, 44.4279, 44.5288, 44.6392, 44.733, 44.8269, 44.9137, 44.9912, 45.0545,
        45.1249, 45.1648, 45.1883, 45.2564, 45.2798, 45.2658, 45.2423, 45.2423, 45.1648, 45.071, 44.9771, 44.9302, 44.9067, 44.9067, 44.8997, 44.9536, 44.9771, 44.9841, 44.9372, 44.8738, 44.8199,
        44.7964, 44.7964, 44.7964, 44.6955, 44.6157, 44.6392, 44.6955, 44.726, 44.7424, 44.6861, 44.6321, 44.5922, 44.5382, 44.5288, 44.5382, 44.6157, 44.679, 44.726, 44.8034, 44.8598, 44.8738] };
  WCL[508] = struct {
   x = [-79.9285, -80.006, -80.0717, -80.0177, -80.0717, -80.0599, -80.1702, -80.2923, -80.3815, -80.5575, -80.6232, -80.656, -80.71, -80.832, -81.0315, -81.1864, -81.3624, -81.4821, -81.583, -81.6159,
        -81.6159, -81.6605, -81.6816, -81.8459, -82.0993, -82.2871, -82.4209, -82.5194, -82.5851, -82.7072, -82.794, -82.9043, -83.0146, -83.0592, -83.0827, -83.193, -83.3784, -83.5661, -83.6553,
        -83.7984, -83.9416, -84.0308, -84.1059, -84.1293, -84.1411, -84.1951, -84.2279, -84.2396, -84.2514, -84.2936, -84.3499, -84.3711, -84.3828, -84.3945, -84.3828, -84.3945, -84.3828, -84.3382,
        -84.3265, -84.2842, -84.2842, -84.2725, -84.2725, -84.2279, -84.1622, -84.1293, -84.0965, -84.0073, -83.9862, -83.9956, -84.1176, -84.1505, -84.1622, -84.2608, -84.3828, -84.4485, -84.526,
        -84.6151, -84.6574, -84.6809, -84.824, -84.9015, -84.9554, -84.9672, -85.1432, -85.3638, -85.5844, -85.7064, -85.9927, -86.2462, -86.3565, -86.4433, -86.5654, -86.6639, -86.6639, -86.6545,
        -86.786, -86.9291, -86.9854, -87.1708, -87.3163, -87.4924, -87.5909, -87.7012, -87.8233, -87.8655, -87.9218, -87.9993, -87.9993, -87.9101, -87.7787, -87.5792, -87.4478, -87.4032, -87.3914,
        -87.2811, -87.138, -87.0957, -87.1615, -87.2389, -87.3163, -87.4032, -87.5017, -87.5463, -87.5463, -87.6027, -87.6566, -87.7012, -87.713, -87.7998, -87.8655, -87.9218, -87.9101, -87.889,
        -87.8444, -87.8327, -87.8561, -87.8444, -87.8444, -87.7787, -87.6895, -87.6238, -87.5581, -87.4806, -87.3586, -87.206, -87.0512, -86.8517, -86.786, -86.7085, -86.6757, -86.5771, -86.4785,
        -86.3682, -86.3236, -86.3002, -86.279, -86.2673, -86.3236, -86.4339, -86.5114, -86.5536, -86.5654, -86.5114, -86.4668, -86.4222, -86.3236, -86.3448, -86.3236, -86.2908, -86.2908, -86.2908,
        -86.2908, -86.2908, -86.1899, -86.157, -86.0467, -85.9153, -85.8167, -85.7275, -85.683, -85.6736, -85.6618, -85.6384, -85.5726, -85.4858, -85.453, -85.453, -85.4412, -85.2746, -85.0329, -85,
        -85.0892, -85.1103, -85.1103, -85.0446, -85.0118, -84.9226, -84.8005, -84.702, -84.5917, -84.5588, -84.4368, -84.3382, -84.2396, -84.1059, -84.019, -83.9087, -83.8853, -83.8759, -83.7867,
        -83.6318, -83.5121, -83.4347, -83.3901, -83.4229, -83.4793, -83.4675, -83.4112, -83.369, -83.3361, -83.3361, -83.3784, -83.4887, -83.5661, -83.6764, -83.8196, -83.9087, -83.9533, -83.9087,
        -83.7867, -83.6764, -83.545, -83.4558, -83.4229, -83.3901, -83.2915, -83.092, -82.9395, -82.8292, -82.7189, -82.6415, -82.6297, -82.6086, -82.5734, -82.5194, -82.4631],
   y = [44.9771, 45.071, 45.132, 45.1719, 45.2798, 45.3737, 45.3808, 45.4206, 45.5122, 45.6131, 45.7445, 45.8525, 45.8759, 45.9229, 45.9463, 45.9768, 45.9909, 45.9839, 45.9604, 46.0214, 46.0825, 46.0825,
        46.0684, 46.0613, 46.0825, 46.113, 46.1294, 46.1599, 46.1435, 46.1599, 46.2139, 46.1975, 46.1834, 46.1599, 46.1529, 46.1669, 46.2209, 46.2819, 46.2984, 46.3359, 46.3289, 46.3195, 46.2984,
        46.289, 46.3899, 46.4415, 46.4814, 46.5096, 46.533, 46.5706, 46.6011, 46.587, 46.5565, 46.5495, 46.5495, 46.533, 46.4885, 46.4415, 46.3805, 46.3195, 46.2655, 46.2139, 46.174, 46.1529, 46.1294,
        46.0918, 46.0613, 46.0214, 45.9839, 45.9768, 45.9768, 45.9675, 45.9675, 45.9534, 45.937, 45.9135, 45.9839, 46.0214, 45.9675, 45.8759, 45.8759, 45.9463, 45.9675, 45.9675, 46.0379, 46.0754,
        46.0449, 45.937, 45.9135, 45.8595, 45.7914, 45.714, 45.6225, 45.6131, 45.7445, 45.8595, 45.7844, 45.7375, 45.7445, 45.599, 45.4136, 45.2259, 45.0851, 44.9536, 44.8738, 44.8128, 44.726, 44.6086,
        44.5453, 44.6157, 44.6955, 44.8199, 44.8363, 44.9466, 45.0545, 45.1555, 45.2564, 45.1414, 45.0311, 44.8903, 44.7893, 44.6955, 44.5288, 44.3552, 44.2378, 44.1111, 43.9914, 43.8483, 43.6723,
        43.5291, 43.4, 43.224, 42.9987, 42.8837, 42.7617, 42.649, 42.4941, 42.3486, 42.2829, 42.1351, 41.9802, 41.8065, 41.7314, 41.6986, 41.6657, 41.6587, 41.7408, 41.83, 41.8558, 41.898, 42.013,
        42.1515, 42.29, 42.5035, 42.5763, 42.642, 42.8603, 43.102, 43.224, 43.3531, 43.5291, 43.6324, 43.6652, 43.7919, 43.9515, 44.1979, 44.2543, 44.3012, 44.4655, 44.5523, 44.5617, 44.5852, 44.6321,
        44.6861, 44.7729, 44.8903, 44.9466, 45.0006, 45.078, 45.1179, 45.0475, 44.9701, 44.8738, 44.8363, 44.9137, 44.9372, 45.078, 45.2188, 45.2564, 45.2798, 45.3502, 45.3901, 45.4746, 45.545,
        45.6131, 45.6671, 45.6999, 45.714, 45.714, 45.6905, 45.6671, 45.6436, 45.66, 45.6295, 45.5755, 45.4981, 45.4746, 45.4136, 45.4042, 45.4042, 45.3901, 45.3362, 45.2564, 45.1648, 45.0851, 45.078,
        45.024, 44.9231, 44.8363, 44.6955, 44.5617, 44.4279, 44.3481, 44.2613, 44.1346, 44.0642, 43.9844, 43.8553, 43.7356, 43.6816, 43.6488, 43.6816, 43.745, 43.8882, 43.9515, 43.975, 44.0079,
        44.0642, 44.0783, 43.968, 43.8248, 43.6723, 43.5197, 43.3836, 43.2404, 43.1278, 43.0621] };
  WCL[509] = struct {
   x = [-84.526, -84.4814, -84.4814, -84.5377, -84.5588, -84.4931, -84.5471, -84.5706, -84.4602, -84.4602, -84.702, -84.7677, -84.6809, -84.9554, -85.0118, -84.9789, -84.9226, -85.0446, -85.176, -85.352,
        -85.5069, -85.7487, -85.9599, -86.1476, -86.2345, -86.3448, -86.4339, -86.4668, -86.4996, -86.5888, -86.8188, -87.0957, -87.0957, -87.4478, -87.8327, -88.0861, -88.1753, -88.1964, -88.1964,
        -88.1753, -88.2081, -88.2973, -88.363, -88.4499, -88.5179, -88.5602, -88.4733, -88.417, -88.4405, -88.5273, -88.6048, -88.7033, -88.7479, -88.8582, -88.9028, -88.8911, -89.0014, -89.1445,
        -89.2548, -89.3323, -89.3651, -89.4097, -89.52, -89.6092, -89.6186, -89.6303, -89.7524, -89.9284],
   y = [46.4885, 46.4955, 46.5096, 46.526, 46.6246, 46.7231, 46.7067, 46.7748, 46.8593, 46.8733, 46.9649, 47.1291, 47.3333, 47.5563, 47.6454, 47.8614, 47.9951, 47.9717, 47.9646, 47.9435, 47.9435, 47.9365,
        48.0397, 48.3119, 48.4363, 48.6335, 48.7273, 48.7273, 48.7367, 48.7273, 48.786, 48.786, 48.3119, 48.8517, 48.9456, 48.9972, 48.9597, 48.9245, 48.8447, 48.6851, 48.6405, 48.6053, 48.5748,
        48.4645, 48.4363, 48.5232, 48.6405, 48.7508, 48.8165, 48.7133, 48.5818, 48.4222, 48.3777, 48.3471, 48.4504, 48.5818, 48.502, 48.3847, 48.1946, 48.1571, 48.0984, 48.0843, 48.0608, 48.0162,
        48.0162, 48.0162, 47.981, 47.8754] };
  WCL[510] = struct {
   x = [-90.0715, -89.8627, -89.6421, -89.3651, -89.168, -88.9685, -88.7479, -88.4733, -88.2645, -88.065, -87.8655, -87.8233, -87.8233, -87.8327, -87.9993, -88.1307, -88.2973, -88.417, -88.5062, -88.5273,
        -88.5179, -88.3748, -88.2856, -88.187, -87.9993, -87.8655, -87.7787, -87.6684, -87.5581, -87.4266, -87.3492, -87.1826, -87.0066, -86.8517, -86.7648, -86.7414, -86.6428, -86.5208, -86.3565,
        -86.1805, -85.9693, -85.7604, -85.6172, -85.4623, -85.2863, -85.1314, -85.0892, -85.0892, -84.9672, -84.9226, -84.7912, -84.6809, -84.6151, -84.5799, -84.526],
   y = [46.6621, 46.7748, 46.8287, 46.8663, 46.9719, 47.047, 47.2207, 47.3474, 47.4225, 47.4671, 47.4741, 47.4671, 47.446, 47.4084, 47.3263, 47.2277, 47.0846, 47.04, 46.9578, 46.8053, 46.7536, 46.8522,
        46.8358, 46.8428, 46.8733, 46.8287, 46.7982, 46.7466, 46.6246, 46.526, 46.4814, 46.5025, 46.4814, 46.4579, 46.4486, 46.4415, 46.465, 46.5636, 46.6011, 46.6786, 46.6692, 46.6621, 46.6692,
        46.6926, 46.7607, 46.7536, 46.6856, 46.5495, 46.5025, 46.4885, 46.465, 46.4814, 46.4814, 46.4814, 46.4885] };
  WCL[511] = struct {
   x = [-89.9284, -90.1701, -90.5245, -90.8436, -91.076, -91.2966, -91.4726, -91.7261, -91.9795, -92.1109, -92.0335, -91.8786, -91.6486, -91.428, -91.252, -91.0971, -90.9751, -90.8648, -90.8554, -90.9094,
        -90.8882, -90.8108, -90.6559, -90.4799, -90.2147, -90.0715],
   y = [47.8754, 47.7276, 47.6384, 47.4741, 47.3474, 47.2207, 47.101, 46.9508, 46.7982, 46.6997, 46.6175, 46.6011, 46.6387, 46.7372, 46.8147, 46.8733, 46.9273, 46.8522, 46.7372, 46.5941, 46.5636, 46.6175,
        46.5941, 46.58, 46.6011, 46.6621] };
  WCL[512] = struct {
   x = [-113.082, -113.061, -113.061, -112.972, -112.906, -112.897, -112.841, -112.784, -112.697, -112.641, -112.665, -112.718, -112.697, -112.599, -112.477, -112.388, -112.376, -112.376, -112.343, -112.31,
        -112.244, -112.179, -112.122, -112.101, -112.012, -111.979, -112.003, -112.078, -112.167, -112.155, -112.078, -111.946, -111.869, -111.914, -111.958, -111.946, -112.012, -112.09, -112.2,
        -112.289, -112.343, -112.343, -112.388, -112.444, -112.554, -112.632, -112.763, -112.796, -112.784, -112.862, -112.939, -112.995, -113.049, -113.073, -113.082, -113.082],
   y = [41.7455, 41.7361, 41.7455, 41.7103, 41.6704, 41.7619, 41.7455, 41.6868, 41.6939, 41.6704, 41.5624, 41.4639, 41.4146, 41.4216, 41.3301, 41.3465, 41.4545, 41.5226, 41.5296, 41.5718, 41.5624, 41.5718,
        41.5554, 41.5038, 41.4874, 41.431, 41.3958, 41.3724, 41.3231, 41.1987, 41.0978, 41.0649, 40.9804, 40.964, 40.8725, 40.8044, 40.7645, 40.7129, 40.6965, 40.781, 40.8302, 40.8819, 40.9053,
        40.8467, 40.9804, 41.0063, 41.14, 41.1987, 41.3231, 41.3559, 41.4545, 41.5296, 41.5883, 41.661, 41.7197, 41.7455] };
  WCL[513] = struct {
   x = [-125.305, -125.173, -125.039, -124.908, -124.884, -124.908, -124.875, -124.807, -124.73, -124.586, -124.422, -124.256, -124.166, -124.047, -123.946, -123.859, -123.781, -123.737, -123.725, -123.648,
        -123.594, -123.582, -123.528, -123.472, -123.451, -123.427, -123.361, -123.317, -123.361, -123.394, -123.516, -123.648, -123.716, -123.847, -123.969, -124.101, -124.256, -124.288, -124.288,
        -124.378, -124.488, -124.654, -124.753, -124.753, -124.687, -124.654, -124.708, -124.753, -124.863, -125.028, -125.063, -125.072, -124.974, -124.875, -124.974, -125.096, -125.194, -125.316,
        -125.537, -125.725, -125.647, -125.471, -125.537, -125.626, -125.725, -125.802, -125.924, -125.966, -126.154, -126.222, -126.354, -126.443, -126.508, -126.52, -126.431, -126.199, -125.99,
        -126.178, -126.342, -126.386, -126.419, -126.53, -126.607, -126.717, -126.783, -126.839, -126.893, -127.027, -127.159],
   y = [50.0416, 49.9571, 49.8562, 49.7576, 49.72, 49.6637, 49.5933, 49.5417, 49.5135, 49.4337, 49.3845, 49.3774, 49.3187, 49.2906, 49.2319, 49.1756, 49.1169, 49.0535, 49.016, 48.9292, 48.8494, 48.7837,
        48.6687, 48.61, 48.6311, 48.6827, 48.6006, 48.5208, 48.5067, 48.4269, 48.3589, 48.3683, 48.3824, 48.4105, 48.4481, 48.4997, 48.5513, 48.5654, 48.5936, 48.617, 48.5795, 48.6382, 48.6593, 48.718,
        48.7766, 48.8423, 48.8283, 48.7555, 48.732, 48.8212, 48.8705, 48.9362, 48.9503, 49.023, 49.0301, 49.0747, 49.0019, 49.023, 48.9878, 49.0887, 49.0817, 49.1028, 49.1897, 49.2319, 49.3047,
        49.3117, 49.3187, 49.3845, 49.4408, 49.4337, 49.4337, 49.4478, 49.5628, 49.5863, 49.6285, 49.6215, 49.6778, 49.7341, 49.706, 49.7435, 49.8069, 49.7506, 49.8632, 49.9195, 49.943, 49.9195,
        49.9712, 49.8984, 49.9712] };
  WCL[514] = struct {
   x = [-122.786, -122.831, -122.864, -122.908, -122.998, -123.019, -122.998, -122.92, -122.798, -122.688, -122.578, -122.601, -122.655, -122.676, -122.711, -122.765, -122.831, -122.941, -123.042, -123.153,
        -123.174, -123.162, -123.174, -123.206, -123.206, -123.162, -123.162, -123.153, -123.153, -123.239, -123.251, -123.34, -123.406, -123.46, -123.516, -123.615, -123.659, -123.659, -123.627,
        -123.605, -123.683, -123.737, -123.737, -123.781, -123.793, -123.826, -123.814, -123.904],
   y = [49.0442, 49.0606, 49.0747, 49.0887, 49.0747, 49.1028, 49.1474, 49.1615, 49.1967, 49.2037, 49.1897, 49.2483, 49.2483, 49.2835, 49.2976, 49.2765, 49.2624, 49.2483, 49.2483, 49.2695, 49.3047, 49.3187,
        49.3328, 49.3774, 49.4126, 49.4337, 49.4994, 49.5628, 49.713, 49.7341, 49.6708, 49.6074, 49.5769, 49.5206, 49.4478, 49.4642, 49.5487, 49.5933, 49.6637, 49.713, 49.713, 49.713, 49.7787, 49.7787,
        49.8069, 49.8421, 49.8914, 49.9571] };
  WCL[515] = struct {
   x = [-123.981, -123.936, -123.946, -124.023, -124.056, -124.068, -124.08, -124.124, -124.19, -124.279, -124.345, -124.41, -124.488, -124.577, -124.654],
   y = [50.0275, 49.9266, 49.8702, 49.9266, 49.9641, 49.9336, 49.9195, 49.8562, 49.821, 49.8139, 49.821, 49.828, 49.8632, 49.9054, 49.9641] };
  WCL[516] = struct {
   x = [-123.913, -123.969, -124.002, -124.091, -124.19, -124.267, -124.267, -124.246, -124.178, -124.091, -124.014, -124.023, -124.056, -124.023, -124.014, -124.014, -123.981, -123.981, -124.047, -124.112,
        -124.133, -124.19, -124.234, -124.312, -124.321, -124.345, -124.399, -124.399, -124.345, -124.3, -124.211, -124.101, -124.124, -124.166, -124.145, -124.101, -124.056, -124.035, -124.023,
        -124.014, -123.981, -123.981, -123.99, -123.99, -123.99, -123.99, -123.925, -123.913, -123.904, -123.904, -123.892, -123.835, -123.859, -123.859, -123.826, -123.88, -123.904, -123.892, -123.88,
        -123.868, -123.793, -123.638, -123.483, -123.34, -123.23, -123.174, -123.206, -123.296, -123.385, -123.451, -123.549, -123.683, -123.803, -123.88, -123.946, -123.99, -124.002, -124.002,
        -123.925, -123.913, -123.913, -123.859, -123.814, -123.803, -123.826, -123.835, -123.814, -123.835, -123.925, -124.002, -124.014, -123.936, -123.925, -123.99, -124.047, -124.091, -124.124,
        -124.157, -124.178, -124.211, -124.246, -124.267, -124.3, -124.356, -124.443, -124.544, -124.619, -124.664, -124.664, -124.619, -124.598, -124.544, -124.455, -124.345, -124.223, -124.145,
        -124.023, -123.904, -123.793, -123.648, -123.516, -123.361, -123.239, -123.206, -123.153, -123.009, -122.899, -122.765, -122.667, -122.634, -122.676, -122.798, -122.908, -123.009, -122.986,
        -122.854, -122.831, -122.908, -122.899, -122.798, -122.7, -122.578, -122.524, -122.434, -122.357, -122.324, -122.291, -122.279, -122.258, -122.258, -122.258, -122.247, -122.279, -122.291,
        -122.303, -122.336, -122.357, -122.369, -122.378, -122.413, -122.467, -122.488, -122.5, -122.556, -122.556, -122.488, -122.479, -122.456, -122.423, -122.423, -122.413, -122.345, -122.268,
        -122.268, -122.324, -122.413, -122.524, -122.601, -122.643, -122.7, -122.744, -122.765, -122.744, -122.754, -122.798, -122.798],
   y = [39.9807, 40.0487, 40.098, 40.2013, 40.2928, 40.328, 40.4102, 40.5369, 40.612, 40.7293, 40.8302, 40.9405, 40.9312, 40.9734, 41.0485, 41.1635, 41.3395, 41.4803, 41.6375, 41.7948, 41.9661, 42.0647,
        42.1562, 42.3017, 42.4589, 42.6443, 42.7992, 42.9119, 43.0574, 43.2357, 43.3812, 43.4282, 43.5338, 43.5338, 43.5972, 43.684, 43.8201, 43.9328, 44.0524, 44.1698, 44.3129, 44.3833, 44.4467,
        44.55, 44.6908, 44.8316, 44.9888, 45.0358, 45.1672, 45.308, 45.4324, 45.5403, 45.5943, 45.6952, 45.7797, 45.7891, 45.8501, 45.9651, 46.0332, 46.1341, 46.1881, 46.1881, 46.2021, 46.2021,
        46.1576, 46.1881, 46.2327, 46.2397, 46.2866, 46.3242, 46.3171, 46.3171, 46.3101, 46.3242, 46.3617, 46.4227, 46.5682, 46.6668, 46.6222, 46.5377, 46.4697, 46.4533, 46.4767, 46.4838, 46.5448,
        46.6433, 46.7278, 46.7724, 46.7654, 46.7959, 46.878, 46.9836, 47.0376, 47.1057, 47.0752, 47.0141, 47.0892, 47.1573, 47.2629, 47.3286, 47.4202, 47.5234, 47.6126, 47.7112, 47.8285, 47.9177,
        47.9998, 48.096, 48.2275, 48.3378, 48.3824, 48.4199, 48.3964, 48.3448, 48.3237, 48.3096, 48.2721, 48.2415, 48.2134, 48.2204, 48.1993, 48.1923, 48.1923, 48.1993, 48.1993, 48.1547, 48.1172,
        48.1547, 48.1383, 48.0209, 47.9177, 47.8449, 47.7018, 47.568, 47.4037, 47.4342, 47.5093, 47.5093, 47.5234, 47.6126, 47.6572, 47.7557, 47.859, 47.8872, 47.8801, 47.8144, 47.791, 47.8731,
        47.9247, 48.0139, 48.0585, 48.1101, 48.1829, 48.2204, 48.2275, 48.2861, 48.3166, 48.3448, 48.3753, 48.4035, 48.4105, 48.4199, 48.434, 48.4621, 48.4856, 48.5138, 48.5208, 48.5865, 48.5795,
        48.6522, 48.718, 48.732, 48.6968, 48.7109, 48.7555, 48.7766, 48.7907, 48.7696, 48.8212, 48.8635, 48.9292, 48.9667, 49.0019, 49.023, 49.0301, 49.0442] };
  WCL[517] = struct {
   x = [154.832, 154.8, 154.722, 154.645, 154.612, 154.743, 154.8, 154.832, 154.832],
   y = [49.6426, 49.591, 49.4994, 49.3915, 49.3117, 49.3539, 49.476, 49.591, 49.6426] };
  WCL[518] = struct {
   x = [154.225, 154.236, 154.192, 154.126, 154.138, 154.248, 154.281, 154.225],
   y = [48.8987, 48.9057, 48.8353, 48.7461, 48.7625, 48.7977, 48.8705, 48.8987] };
  WCL[519] = struct {
   x = [152.293, 152.281, 152.183, 152.073, 151.918, 151.828, 151.939, 152.094, 152.204, 152.293],
   y = [47.1409, 47.0447, 46.9602, 46.878, 46.7959, 46.817, 46.9297, 47.0071, 47.1409, 47.1409] };
  WCL[520] = struct {
   x = [150.571, 150.538, 150.383, 150.228, 150.052],
   y = [46.1716, 46.1106, 46.0027, 45.8712, 45.7562] };
  WCL[521] = struct {
   x = [149.843, 150.085, 150.294, 150.46, 150.549, 150.571],
   y = [45.8572, 46.0402, 46.1552, 46.2327, 46.2467, 46.1716] };
  WCL[522] = struct {
   x = [140.702, 140.557, 140.58, 140.59, 140.613, 140.59, 140.503, 140.446, 140.413, 140.392, 140.425, 140.36, 140.327, 140.303, 140.148, 140.017],
   y = [50.0322, 49.9054, 49.7928, 49.6496, 49.5065, 49.3915, 49.253, 49.1592, 49.0512, 49.023, 48.8987, 48.7532, 48.5936, 48.4621, 48.3518, 48.3002] };
  WCL[523] = struct {
   x = [144.068, 144.168, 144.222, 144.323, 144.422, 144.565, 144.765, 144.797, 144.708, 144.553, 144.398, 144.157, 143.87, 143.593, 143.373, 143.141, 143.051, 143.009, 142.875, 142.777, 142.699, 142.655,
        142.666, 142.81, 142.953, 143.108, 143.141, 143.206, 143.307, 143.462, 143.516, 143.572, 143.626, 143.671, 143.65, 143.549, 143.495, 143.417, 143.173, 142.943, 142.699, 142.523, 142.434,
        142.357, 142.258, 142.026, 141.969, 141.96, 142.026, 142.07, 142.091, 142.124, 142.157, 142.279, 142.258, 142.124, 142.005, 142.005, 142.091, 142.157, 142.202, 142.214, 142.214, 142.246,
        142.267],
   y = [50.0111, 49.8069, 49.591, 49.3539, 49.1756, 49.009, 48.8564, 48.732, 48.8189, 48.9644, 49.1099, 49.2249, 49.3187, 49.3469, 49.3258, 49.2037, 48.9878, 48.7977, 48.5865, 48.3378, 48.1031, 47.8801,
        47.6501, 47.4483, 47.3357, 47.1949, 47.0282, 46.9156, 46.8405, 46.7724, 46.8029, 46.756, 46.5823, 46.4157, 46.2256, 46.0707, 46.2702, 46.5377, 46.6058, 46.6879, 46.6973, 46.5589, 46.3617,
        46.1482, 45.9557, 46.0496, 46.3547, 46.6574, 46.9836, 47.1949, 47.4413, 47.6713, 47.8942, 48.035, 48.2204, 48.4316, 48.7109, 48.9503, 49.131, 49.361, 49.591, 49.7764, 49.9547, 49.9759, 49.9759] };
  WCL[524] = struct {
   x = [150.052, 149.798, 149.644, 149.688, 149.843],
   y = [45.7562, 45.6248, 45.5873, 45.7398, 45.8572] };
  WCL[525] = struct {
   x = [148.958, 148.97, 148.958, 148.871, 148.693, 148.496, 148.318, 148.142, 148.031, 147.909, 147.79, 147.679, 147.567, 147.435, 147.313, 147.215, 147.071, 147.125, 147.236, 147.28, 147.381, 147.59,
        147.733, 147.865, 147.954, 148.097, 148.153, 148.198, 148.329, 148.529, 148.693, 148.86, 148.949, 148.958],
   y = [45.3925, 45.4394, 45.369, 45.3455, 45.2681, 45.1742, 45.0827, 44.9959, 44.9489, 44.9489, 44.8856, 44.7541, 44.6978, 44.6204, 44.5242, 44.4467, 44.4467, 44.5406, 44.6509, 44.7682, 44.8011, 45.0029,
        45.1132, 45.1601, 45.2916, 45.3385, 45.2986, 45.2306, 45.2611, 45.369, 45.5004, 45.5239, 45.4864, 45.3925] };
  WCL[526] = struct {
   x = [146.938, 147.015, 146.928, 146.762, 146.663, 146.762, 146.938],
   y = [43.8436, 43.8201, 43.7567, 43.691, 43.7309, 43.7966, 43.8436] };
  WCL[527] = struct {
   x = [146.21, 146.231, 146.288, 146.365, 146.419, 146.32, 146.121, 145.912, 145.724, 145.614, 145.537, 145.703, 145.9, 146.067, 146.144, 146.21],
   y = [44.4467, 44.4866, 44.4467, 44.3998, 44.3036, 44.2496, 44.097, 43.9398, 43.7966, 43.6676, 43.7966, 43.9468, 44.1135, 44.2801, 44.4303, 44.4467] };
  WCL[528] = struct {
   x = [145.084, 145.149, 145.37, 145.447, 145.349, 145.182, 145.25, 145.426, 145.426, 145.647, 145.912, 145.814, 145.492, 145.227, 145.051, 144.952, 144.842, 144.509, 144.234, 143.957, 143.692, 143.516,
        143.45, 143.328, 143.051, 142.842, 142.479, 142.181, 141.96, 141.65, 141.418, 141.254, 141.054, 140.845, 140.557, 140.404, 140.503, 140.702, 140.923, 141.111, 141.242, 141.176, 140.956,
        140.789, 140.625, 140.524, 140.27, 140.139, 140.083, 140.16, 140.216, 140.029],
   y = [44.0665, 44.1135, 44.2566, 44.2097, 44.0032, 43.7802, 43.6206, 43.5408, 43.339, 43.332, 43.4047, 43.2498, 43.1466, 43.0011, 42.9776, 43.0175, 42.9048, 42.9353, 42.9119, 42.7992, 42.5387, 42.2853,
        42.0482, 41.9239, 42.0811, 42.1961, 42.2947, 42.4402, 42.5293, 42.5552, 42.473, 42.3674, 42.3838, 42.5059, 42.4894, 42.3111, 42.1539, 42.0975, 42.0154, 41.9004, 41.8089, 41.7103, 41.7267,
        41.7197, 41.6352, 41.5202, 41.3888, 41.4216, 41.5624, 41.8018, 41.9825, 42.121] };
  WCL[529] = struct {
   x = [139.928, 140.094, 140.327, 140.514, 140.568, 140.47, 140.514, 140.744, 140.965, 141.153, 141.364, 141.495, 141.463, 141.463, 141.596, 141.728, 141.739, 141.805, 141.871, 141.805, 141.728, 141.683,
        141.695, 141.793, 141.904, 141.993, 142.124, 142.136, 142.181, 142.347, 142.49, 142.61, 142.765, 142.932, 143.141, 143.34, 143.516, 143.737, 143.936, 144.145, 144.333, 144.443, 144.654,
        144.863, 145.084],
   y = [42.5223, 42.6608, 42.7499, 42.9048, 43.0574, 43.163, 43.2827, 43.2428, 43.1794, 43.1794, 43.1865, 43.3789, 43.5573, 43.7168, 43.8107, 44.0431, 44.3036, 44.5476, 44.7213, 44.8621, 45.0358, 45.1601,
        45.2845, 45.3549, 45.4324, 45.4629, 45.3761, 45.362, 45.2845, 45.1907, 45.0428, 44.9184, 44.7682, 44.6438, 44.4936, 44.3833, 44.2801, 44.1933, 44.1299, 44.083, 44.0595, 43.9703, 43.9069,
        43.9398, 44.0665] };
  WCL[530] = struct {
   x = [141.242, 141.132, 141.242, 141.34, 141.364, 141.242],
   y = [45.2306, 45.1672, 45.0968, 45.0897, 45.2306, 45.2306] };
  WCL[531] = struct {
   x = [141.209, 141.23, 141.364, 141.486, 141.528, 141.519, 141.495, 141.519, 141.606, 141.793, 141.904, 141.96],
   y = [41.3724, 41.3794, 41.3724, 41.3794, 41.2808, 41.1377, 40.9476, 40.7223, 40.5017, 40.3257, 40.1309, 40.0229] };
  WCL[532] = struct {
   x = [139.907, 140.083, 140.071, 140.061, 140.148, 140.282, 140.315, 140.336, 140.348, 140.413, 140.547, 140.658, 140.744, 140.822, 140.944, 141.075, 141.197, 141.319, 141.331, 141.186, 140.944, 140.878,
        140.944, 141.042, 141.209],
   y = [39.9807, 40.1825, 40.436, 40.612, 40.7293, 40.788, 40.788, 40.8631, 41.0227, 41.1306, 41.1799, 41.072, 40.8795, 40.7974, 40.9312, 40.8889, 40.8959, 41.0884, 41.2456, 41.1564, 41.1048, 41.2644,
        41.4287, 41.4545, 41.3724] };
  WCL[533] = struct {
   x = [132.73, 132.786, 132.829, 132.742, 132.521, 132.178, 132.089, 132.2, 132.2, 132.3, 132.476, 132.697, 132.73],
   y = [44.8386, 44.909, 45.1132, 45.2446, 45.2845, 45.2376, 45.1203, 44.9419, 44.7776, 44.6673, 44.5969, 44.6814, 44.8386] };
  WCL[534] = struct {
   x = [140.017, 139.972, 139.785, 139.62, 139.421, 139.266, 139.123, 138.935, 138.747, 138.592, 138.538, 138.461, 138.405, 138.261, 138.139, 138.261, 138.172, 138.085, 137.886, 137.787, 137.588, 137.379,
        137.212, 137.013, 136.87, 136.804, 136.682, 136.539, 136.396, 136.297, 136.098, 135.922, 135.776, 135.755, 135.633, 135.513, 135.347, 135.093, 134.873, 134.629, 134.33, 134.065, 133.878,
        133.746, 133.547, 133.326, 133.105, 132.995, 132.929, 132.763, 132.563, 132.411, 132.378, 132.366, 132.399, 132.399, 132.289, 132.068, 132.002, 132.089, 131.934, 131.913, 131.848, 131.782,
        131.648, 131.559, 131.505, 131.383, 131.284, 131.073, 130.909, 130.754, 130.81, 130.843, 130.622],
   y = [48.3002, 48.2275, 48.0726, 47.9623, 47.8144, 47.6196, 47.4413, 47.2911, 47.1409, 46.9602, 46.817, 46.6433, 46.5213, 46.3242, 46.1622, 46.3922, 46.2561, 46.1106, 45.9792, 45.8173, 45.6553, 45.5098,
        45.362, 45.2306, 45.1132, 45.0358, 44.902, 44.8081, 44.6438, 44.5007, 44.4139, 44.2895, 44.1463, 44.0337, 43.8999, 43.7638, 43.5972, 43.4047, 43.2827, 43.163, 43.0081, 42.8462, 42.7828,
        42.7828, 42.7406, 42.6772, 42.7007, 42.7499, 42.7406, 42.8297, 42.8556, 42.8462, 42.8955, 43.0339, 43.1372, 43.2428, 43.2099, 43.0902, 43.1466, 43.2827, 43.2991, 43.4188, 43.2921, 43.1067,
        42.9846, 42.9283, 42.8227, 42.7406, 42.6115, 42.6115, 42.6209, 42.5951, 42.5293, 42.4237, 42.5622] };
  WCL[535] = struct {
   x = [140.029, 139.883, 139.895, 139.928],
   y = [42.121, 42.2125, 42.3346, 42.5223] };
  WCL[536] = struct {
   x = [130.7, 130.733, 130.611, 130.477, 130.324, 130.104],
   y = [42.3745, 42.2782, 42.3017, 42.2688, 42.1304, 41.9567] };
  WCL[537] = struct {
   x = [130.104, 129.982, 129.904, 129.749, 129.705, 129.771, 129.761, 129.749, 129.639, 129.374, 129.231, 129.064, 128.834, 128.712, 128.656, 128.435, 128.17, 128.006],
   y = [41.9567, 41.8511, 41.7361, 41.5624, 41.4123, 41.2879, 41.072, 40.8889, 40.8138, 40.7129, 40.5791, 40.4524, 40.3351, 40.2576, 40.1731, 40.0558, 40.03, 40.0229] };
  WCL[538] = struct {
   x = [120.012, 120.266, 120.475, 120.552, 120.663, 120.773, 120.883, 121.038, 121.184, 121.348, 121.547, 121.78, 121.866, 121.965, 122.033, 122.153, 122.188, 122.242, 122.287, 122.153, 121.956],
   y = [39.9807, 40.0722, 40.1403, 40.2154, 40.3843, 40.4782, 40.5956, 40.7035, 40.8302, 40.8795, 40.8795, 40.8959, 40.9875, 40.8959, 40.7293, 40.6026, 40.5791, 40.5369, 40.4266, 40.2412, 40.0464] };
  WCL[539] = struct {
   x = [74.9039, 74.979, 75.1785, 75.4437, 75.73, 76.0727, 76.4153, 76.7251, 77.0231, 77.2555, 77.5418, 77.8516, 78.0722, 78.2036, 78.3913, 78.5791, 78.7879, 78.9874, 79.086, 79.0977, 78.8889, 78.7011,
        78.5228, 78.3256, 77.9501, 77.6403, 77.3869, 77.0443, 76.7133, 76.4036, 76.1173, 75.8755, 75.5869, 75.3991, 75.3663, 75.2888, 75.256, 75.1785, 74.979, 74.803, 74.6481, 74.451, 74.2398, 74.1741,
        74.1295, 74.052, 74.0638, 74.1412, 74.329, 74.0966, 73.9206, 73.7094, 73.4888, 73.402, 73.4348, 73.5451, 73.7657, 73.8995, 74.0849, 74.3407, 74.5496, 74.7162, 74.9039],
   y = [46.8334, 46.81, 46.749, 46.7865, 46.817, 46.7795, 46.7044, 46.6879, 46.6128, 46.5988, 46.6574, 46.6739, 46.5894, 46.5213, 46.6504, 46.7724, 46.81, 46.81, 46.749, 46.5589, 46.3993, 46.3922, 46.4533,
        46.3852, 46.3476, 46.4392, 46.4908, 46.4603, 46.4908, 46.5072, 46.5448, 46.5448, 46.5072, 46.5448, 46.6199, 46.6574, 46.5072, 46.4462, 46.4298, 46.3242, 46.1482, 46.0566, 46.0027, 45.8173,
        45.6412, 45.4934, 45.2306, 45.0029, 44.8152, 44.8386, 45.0827, 45.3385, 45.4629, 45.6788, 45.9182, 46.1247, 46.1622, 46.3242, 46.4603, 46.5988, 46.7349, 46.817, 46.8334] };
  WCL[540] = struct {
   x = [59.9734, 60.1705, 60.424, 60.4897, 60.6117, 60.7103, 60.8534, 60.9989, 61.0412, 61.175, 61.4824, 61.6373, 61.7358, 61.649, 61.5598, 61.4613, 61.351, 61.1632, 61.0529, 61.0083, 61.0858, 61.2618,
        61.4495, 61.6044, 61.6162, 61.4613, 61.2407, 61.0647, 60.8886, 60.7783, 60.6563, 60.5789, 60.5671, 60.4568, 60.3794, 60.2808, 60.1494, 60.1611, 60.2691, 60.4686, 60.6328, 60.7995, 60.8206,
        60.546, 60.4803, 60.3043, 60.194],
   y = [43.5972, 43.6535, 43.6934, 43.8131, 43.8858, 44.1158, 44.1956, 44.3364, 44.4866, 44.6837, 44.7471, 44.8175, 45.0287, 45.1226, 45.3174, 45.4723, 45.5638, 45.6483, 45.8337, 46.127, 46.1951, 46.4321,
        46.5753, 46.7349, 46.81, 46.7889, 46.6903, 46.5682, 46.5002, 46.7419, 46.7278, 46.6128, 46.5612, 46.5237, 46.6058, 46.6739, 46.4626, 46.3547, 46.3101, 46.2866, 46.3031, 46.3101, 46.1951,
        46.1177, 46.1787, 46.2561, 46.2021] };
  WCL[541] = struct {
   x = [53.8857, 53.7754, 53.4022, 53.0925],
   y = [39.8633, 40.0323, 40.1262, 40.0839] };
  WCL[542] = struct {
   x = [52.8507, 52.785, 52.7287, 52.7615, 52.8953, 52.9047, 52.8953, 52.8836, 52.8718, 52.9047, 52.9704, 53.0596, 53.2802, 53.4445, 53.6228, 53.8082, 53.996, 54.2823, 54.4489, 54.2729, 54.2494, 54.5264,
        54.7564, 54.3504, 54.0523, 53.942, 53.7543, 53.7214, 53.5008, 53.1699, 52.8953, 52.7287, 52.7287, 52.8179, 52.8179, 52.8273, 52.785, 52.5409, 52.4635, 52.4635, 52.4541, 52.6184, 52.6512,
        52.4424, 52.2218, 51.858, 51.682, 51.5388, 51.3182, 51.2854, 51.2619, 51.0531, 50.9099, 50.7667, 50.3584, 50.1495, 50.2809, 50.6118, 50.931, 51.107, 51.2525, 51.506, 51.4168, 51.3511, 51.3065,
        51.4051, 51.5811, 51.5811, 51.7806, 51.9894, 52.034, 52.21, 52.5292, 52.8179, 53.1042, 53.3905, 53.5548, 53.5782, 53.644, 53.7308, 54.0617, 54.2823, 54.47, 54.7024, 54.7564, 54.6249, 54.8127,
        54.6813, 54.47, 54.294, 54.2072, 54.0077, 53.7988, 53.6979, 53.5665, 53.5125, 53.3788, 53.2802, 53.2473, 52.9939, 52.8953, 52.6841, 52.6841, 52.4424, 52.2429, 52.0669, 51.8698, 51.8134, 51.682,
        51.5482, 51.4614, 51.2173, 50.9639, 50.6447, 50.3373],
   y = [39.9572, 40.1849, 40.3796, 40.5885, 40.7411, 40.7997, 40.9405, 41.0485, 41.1236, 41.0649, 40.9147, 40.8068, 40.7997, 40.8068, 40.7645, 40.7152, 40.7317, 40.673, 40.7411, 40.7645, 40.8396, 40.8913,
        41.14, 41.3724, 41.6211, 41.8605, 42.0506, 42.0741, 42.0999, 42.0248, 41.9027, 41.7455, 41.6211, 41.5883, 41.3724, 41.2315, 41.2644, 41.6141, 41.9098, 42.0905, 42.3932, 42.5411, 42.8086,
        42.872, 42.8884, 43.0104, 43.1888, 43.1794, 43.2052, 43.4446, 43.7168, 43.9797, 44.0595, 44.259, 44.3763, 44.5899, 44.6767, 44.6837, 44.6204, 44.557, 44.5969, 44.5734, 44.6767, 44.8175,
        45.0522, 45.2, 45.2704, 45.1766, 45.1601, 45.2071, 45.3315, 45.4394, 45.4488, 45.3315, 45.254, 45.2939, 45.362, 45.3854, 45.3714, 45.3174, 45.2376, 45.1601, 45.1836, 45.1531, 45.254, 45.4019,
        45.4019, 45.5732, 45.7422, 45.8267, 45.9651, 45.9722, 46.1881, 46.4838, 46.6293, 46.5542, 46.7583, 46.7583, 46.9085, 47.0587, 47.1127, 47.1268, 47.2019, 47.2183, 47.1573, 47.1432, 47.0752,
        47.2019, 47.2089, 47.2019, 47.2324, 47.27, 47.277, 47.0822, 46.871] };
  WCL[543] = struct {
   x = [49.95, 50.0838, 50.1941, 50.2809, 50.3373, 50.1378],
   y = [40.6988, 40.6214, 40.5885, 40.4712, 40.3444, 40.4712] };
  WCL[544] = struct {
   x = [59.9734, 59.9734],
   y = [43.5972, 43.5972] };
  WCL[545] = struct {
   x = [60.194, 59.9828, 59.917, 59.8302, 59.6847, 59.6519, 59.6425, 59.5416, 59.3444, 59.3444, 59.3655, 59.1238, 58.8375, 58.7155, 58.5277, 58.2203, 58.1968, 58.2414, 58.2203, 58.286, 58.3517, 58.3728,
        58.408, 58.5395, 58.638, 58.7037, 58.8586, 58.99, 59.1449, 59.267, 59.4219, 59.619, 59.6964, 59.7856, 59.8959, 59.9734],
   y = [46.2021, 46.1576, 46.3101, 46.3852, 46.3242, 46.1787, 46.0426, 45.9041, 45.8572, 45.9815, 46.0332, 45.9815, 45.8877, 45.6483, 45.3174, 45.0358, 44.8011, 44.6837, 44.503, 44.3599, 44.0994, 43.9093,
        43.7497, 43.7004, 43.7661, 43.7661, 43.7966, 43.7497, 43.7004, 43.7966, 43.806, 43.7403, 43.6699, 43.6065, 43.5972, 43.5972] };
  WCL[546] = struct {
   x = [41.5061, 41.593, 41.7033, 41.7361, 41.715, 41.5718, 41.539, 41.5061, 41.4381, 41.3184, 41.1306, 41.0203, 40.91, 40.7997, 40.6448, 40.4806, 40.2928, 40.1497],
   y = [41.6469, 41.6798, 41.7713, 41.9591, 42.1163, 42.2454, 42.3932, 42.5481, 42.7265, 42.8649, 42.9612, 43.0503, 43.1161, 43.1395, 43.1724, 43.1958, 43.2686, 43.3977] };
  WCL[547] = struct {
   x = [39.9948, 40.3796, 40.7997, 41.1189, 41.4498, 41.5061],
   y = [41.025, 41.0415, 41.2574, 41.424, 41.5718, 41.6469] };
  WCL[548] = struct {
   x = [50.3373, 49.9735, 49.5088, 49.0348, 48.617, 48.7602, 49.0348, 48.7156, 48.6053, 48.4621, 48.373, 48.2955, 48.197, 48.0538, 47.967, 47.866, 47.8778, 47.69, 47.4694, 47.6454, 47.6243, 47.5469, 47.4037,
        47.3474, 47.2934, 47.1174, 47.0846, 47.0517, 46.9954, 46.8522, 46.8311, 46.6762, 46.7536, 46.9508, 47.1714, 47.3474, 47.392, 47.4483, 47.6126, 47.6572, 47.5258, 47.5258, 47.5915, 47.7557,
        47.8989, 48.0867, 48.2415, 48.319, 48.5279, 48.7367, 48.9573, 49.1005, 49.2671, 49.4314, 49.5745, 49.72, 49.95],
   y = [46.871, 46.8475, 46.7513, 46.6973, 46.6222, 46.5542, 46.4321, 46.3476, 46.3031, 46.0637, 45.9346, 45.9346, 45.9041, 45.9111, 46.0871, 46.2327, 46.3242, 46.3336, 46.3782, 46.1787, 46.0261, 45.7727,
        45.8032, 45.6037, 45.3479, 45.2376, 45.1132, 45.0827, 45.0827, 44.9489, 44.8011, 44.6204, 44.4326, 44.4397, 44.2097, 43.9891, 43.7567, 43.6769, 43.9562, 43.7826, 43.4282, 43.1395, 42.9612,
        42.7992, 42.6044, 42.3768, 42.1398, 42.0013, 41.9098, 41.7197, 41.5061, 41.2808, 41.0415, 40.8326, 40.6894, 40.6308, 40.6988] };
  WCL[549] = struct {
   x = [50.1378, 49.8726, 49.6848, 49.5417, 49.5088],
   y = [40.4712, 40.4641, 40.3609, 40.2177, 40.0675] };
  WCL[550] = struct {
   x = [40.1497, 39.9948, 39.8633, 39.6756, 39.5324, 39.4104, 39.2766, 39.0584, 38.8143, 38.528, 38.2745, 38.0328, 37.9108, 37.7676, 37.6691, 37.5916, 37.4039, 37.2278, 37.1269, 36.9063, 36.6763, 36.8524,
        36.8524, 36.9181, 37.195, 37.4367, 37.6127, 37.7887, 38, 38.1103, 38.2299, 38.474, 38.528, 38.3402, 38.176, 38.0751, 37.9436, 37.7794, 37.8216, 38.0093, 38.2863, 38.474, 38.5608, 38.4834,
        38.474, 38.6829, 38.9809, 39.1898, 39.3658, 39.3987, 39.3447, 39.2555, 39.2109, 39.2015, 39.1898, 39.0795, 38.9246, 38.826, 38.6946, 38.5843, 38.528, 38.4294, 38.1971, 38.0328, 37.8451, 37.601,
        37.425, 37.2936, 37.1387, 36.9509, 36.8969, 36.7749, 36.6763, 36.5003, 36.39, 36.268, 36.2117, 36.1905, 35.9911, 35.8479, 35.7282, 35.5616, 35.4959, 35.3973, 35.2542, 35.1204, 35.165, 35.287,
        35.3645, 35.3316, 35.287, 35.1978, 35.1767, 35.111, 35.0101, 34.935, 34.9772, 34.9772, 34.9772, 34.9561, 34.8904, 34.8904, 34.8458, 34.7895, 34.813, 34.7684, 34.6463, 34.5689, 34.536, 34.5806,
        34.536, 34.4046, 34.2943, 34.2732, 34.2614, 34.1957, 34.1629, 34.1183, 34.0291, 33.9423, 33.8414, 33.7874, 33.8742, 33.9634, 34.0408, 34.0948, 34.1394, 34.1511, 34.2169, 34.2614, 34.2732,
        34.36, 34.4492, 34.536, 34.6369, 34.6135, 34.5478, 34.5924, 34.7144, 34.7566, 34.7472, 34.7238, 34.8575, 34.9561, 35.0664, 35.0781, 35.1204, 35.1767, 35.4184, 35.5076, 35.5287, 35.7282,
        35.9371, 36.0145, 36.0591, 36.2023, 36.444, 36.6529, 36.62, 36.5426, 36.5214, 36.5332, 36.5214, 36.4675, 36.322, 36.1905, 36.1577, 36.1248, 35.9371, 35.7939, 35.5522, 35.3856, 35.2213, 35.0664,
        34.8458, 34.6792, 34.5806, 34.6792, 34.5595, 34.4703, 34.4492, 34.3929, 34.2732, 34.1629, 34.062, 33.9305, 33.7663, 33.7099, 33.6442, 33.5011, 33.555, 33.5785, 33.5785, 33.5785, 33.6114,
        33.5879, 33.5011, 33.379, 33.2476, 33.1138, 32.905, 32.7947, 32.6187, 32.6069, 32.5975, 32.6633, 32.8393, 32.8932, 33.0716, 33.2359, 33.3251, 33.5668, 33.7428, 33.8202, 33.7428, 33.7311,
        33.6114, 33.5128, 33.4025, 33.3344, 33.3251, 33.2359, 33.1584, 33.1373, 33.0153, 32.851, 32.6844, 32.6069, 32.4309, 32.3089, 32.2666, 32.1986, 32.0883, 32.0132, 31.8794, 31.8583, 32.046, 32.1,
        32.0672, 31.9569, 31.7808, 31.6823, 31.8254, 32.0883, 32.276, 32.276, 32.1657, 32.046, 31.9569, 31.9123, 31.7691, 31.6048, 31.6142, 31.6377, 31.5274, 31.3185, 31.173, 31.0205, 30.8984, 30.767,
        30.7553, 30.7435, 30.6239, 30.5347, 30.469, 30.4032, 30.3258, 30.2484, 30.1826, 30.2272, 30.3915, 30.4244, 30.3141, 30.2272, 30.1052],
   y = [43.3977, 43.4775, 43.5807, 43.7004, 43.7333, 43.8365, 43.9961, 44.1158, 44.266, 44.3693, 44.4467, 44.5195, 44.6039, 44.6837, 44.7706, 44.7142, 44.7307, 44.8715, 45.0522, 45.1766, 45.254, 45.308,
        45.3854, 45.4488, 45.4253, 45.4324, 45.6647, 45.8267, 46.0566, 46.0426, 46.1787, 46.127, 46.1482, 46.2725, 46.4321, 46.4626, 46.4932, 46.5917, 46.7208, 46.7349, 46.7654, 46.7278, 46.7583,
        46.8264, 46.8874, 46.939, 47.0587, 47.1127, 47.1198, 47.1714, 47.2089, 47.2089, 47.2254, 47.2324, 47.3005, 47.3286, 47.3075, 47.2911, 47.2324, 47.2324, 47.2465, 47.2324, 47.1808, 47.1714,
        47.1503, 47.1057, 46.9836, 46.993, 46.939, 46.8945, 46.7513, 46.8029, 46.8405, 46.7795, 46.7419, 46.6973, 46.6433, 46.7044, 46.6598, 46.6527, 46.5847, 46.4697, 46.4626, 46.3242, 46.1881,
        46.1411, 46.2092, 46.2561, 46.3171, 46.4157, 46.5002, 46.5002, 46.4321, 46.3171, 46.2561, 46.1881, 46.0027, 45.8947, 45.8877, 45.8877, 45.8877, 45.9346, 45.9956, 46.0801, 46.127, 46.1787,
        46.1177, 46.0332, 46.0566, 46.1341, 46.1787, 46.1716, 46.1787, 46.2092, 46.3171, 46.3782, 46.2021, 46.1881, 46.1951, 46.2256, 46.2561, 46.2561, 46.2092, 46.1341, 46.1482, 46.127, 46.0566,
        45.9815, 45.9276, 45.9581, 46.0426, 46.0191, 46.0191, 45.9722, 45.9722, 45.9346, 45.8196, 45.7961, 45.8337, 45.8102, 45.7727, 45.7023, 45.7187, 45.7187, 45.6178, 45.4629, 45.3479, 45.4019,
        45.3479, 45.3409, 45.4183, 45.308, 45.4793, 45.4864, 45.4183, 45.4723, 45.5098, 45.4957, 45.3854, 45.3409, 45.2376, 45.1531, 45.0991, 45.0897, 45.0897, 45.0897, 45.0757, 45.0428, 45.0757,
        45.1226, 45.1695, 45.0522, 44.9489, 44.8246, 44.8551, 44.8246, 44.8011, 44.7612, 44.7471, 44.6532, 44.5899, 44.55, 44.5335, 44.4866, 44.4397, 44.4467, 44.4561, 44.4702, 44.5101, 44.5969,
        44.6603, 44.6908, 44.7612, 44.8246, 44.9278, 45.0522, 45.193, 45.2235, 45.2071, 45.308, 45.4019, 45.3948, 45.4019, 45.4019, 45.4324, 45.5192, 45.5802, 45.6342, 45.7492, 45.8337, 45.8572,
        45.8947, 45.9346, 45.9722, 46.0027, 46.1036, 46.1787, 46.0965, 46.1341, 46.1646, 46.2186, 46.1951, 46.1881, 46.1482, 46.1482, 46.127, 46.1036, 46.0801, 46.1177, 46.1576, 46.2021, 46.2397,
        46.2632, 46.2561, 46.2491, 46.3031, 46.3922, 46.4227, 46.4626, 46.4697, 46.4838, 46.5143, 46.5612, 46.5377, 46.5143, 46.5753, 46.5917, 46.7044, 46.8264, 46.7349, 46.6903, 46.6527, 46.6903,
        46.7795, 46.7208, 46.6433, 46.6293, 46.6058, 46.5612, 46.5377, 46.4932, 46.4321, 46.2937, 46.2092, 46.2256, 46.3242, 46.3476, 46.3476, 46.4016, 46.3171, 46.2186, 46.0801, 45.9416, 45.8736,
        45.8642] };
  WCL[551] = struct {
   x = [29.9292, 30.1826, 30.3915, 30.5793, 30.7107, 30.7881, 30.9876, 31.173, 31.4288, 31.5485, 31.7691, 31.9803, 32.2314, 32.4192, 32.6515, 32.8604, 33.1256, 33.4354, 33.7311, 34.1394, 34.4938, 34.813,
        35.0336, 35.165, 35.2659, 35.2659, 35.409, 35.6625, 35.8925, 36.1366, 36.2563, 36.3994, 36.6763, 36.8969, 37.0401, 37.3147, 37.5353, 37.6456, 37.7559, 38.1314, 38.528, 38.826, 39.0912, 39.3658,
        39.6967, 39.9948],
   y = [41.1729, 41.1729, 41.1823, 41.14, 41.0907, 41.1072, 41.0907, 41.0649, 41.1564, 41.3395, 41.424, 41.5648, 41.6305, 41.7525, 41.8535, 41.9192, 42.0013, 42.0342, 42.0412, 42.0083, 42.0177, 42.0177,
        42.0999, 42.1163, 42.0834, 41.9755, 41.8206, 41.7197, 41.7197, 41.7619, 41.6047, 41.3231, 41.316, 41.2902, 41.2245, 41.1166, 41.0156, 41.0743, 41.1072, 40.9664, 40.9898, 41.0743, 41.1001,
        41.133, 41.0743, 41.025] };
  WCL[552] = struct {
   x = [29.6405, 29.6969, 29.6851, 29.6851, 29.6523, 29.5091, 29.3214, 29.1336, 28.9905, 28.859, 28.8144, 28.9365, 29.0468, 29.035, 29.0468, 28.9036, 28.8708, 28.8379, 28.7487, 28.7159, 28.6596, 28.6596,
        28.6596, 28.6384, 28.6596],
   y = [45.3854, 45.3315, 45.254, 45.1062, 44.9654, 44.8246, 44.794, 44.7002, 44.5664, 44.4796, 44.557, 44.6532, 44.8175, 44.8879, 44.9959, 44.9959, 44.8644, 44.7002, 44.5335, 44.3927, 44.2824, 44.1956,
        44.0431, 43.9398, 43.7896] };
  WCL[553] = struct {
   x = [28.6596, 28.6173, 28.5164, 28.1761, 27.9555, 27.9438, 27.9438, 27.7889, 27.7349, 27.6575, 27.5354, 27.5917, 27.6786, 27.756, 27.8217, 27.8992, 28.0212, 28.0329],
   y = [43.7896, 43.5807, 43.4282, 43.3648, 43.1724, 42.994, 42.7758, 42.7359, 42.6631, 42.6044, 42.5317, 42.4754, 42.4496, 42.3604, 42.2219, 42.1327, 42.0248, 41.9356] };
  WCL[554] = struct {
   x = [28.0329, 28.0095, 28.0658, 28.1972, 28.3404, 28.5938, 28.9036, 29.0796, 29.0914, 28.9811, 28.7159, 28.5727, 28.3193, 28.0329, 27.8335, 27.6129, 27.458, 27.1271, 26.8853, 26.6647, 26.4887, 26.3667,
        26.2235, 26.2235, 26.247, 26.4441, 26.7093, 26.8408, 26.6319, 26.355, 26.2353, 26.1015, 26.0593],
   y = [41.9356, 41.8276, 41.6798, 41.5554, 41.4733, 41.3817, 41.2996, 41.2902, 41.1658, 41.0485, 40.9898, 41.0415, 41.0579, 41.0485, 41.0156, 41.0086, 40.8654, 40.6566, 40.5392, 40.4219, 40.3116, 40.159,
        40.1098, 40.3116, 40.4031, 40.497, 40.5721, 40.673, 40.6566, 40.6308, 40.6214, 40.673, 40.7317] };
  WCL[555] = struct {
   x = [26.355, 26.3995, 26.5662, 26.5427, 26.3456, 26.3127, 26.247, 26.0593],
   y = [41.8112, 41.7291, 41.6211, 41.4639, 41.3489, 41.14, 40.9312, 40.7317] };
  WCL[556] = struct {
   x = [24.7122, 24.8131, 24.7685, 24.7568, 24.6699, 24.515, 24.5268, 24.6371, 24.7122],
   y = [40.6143, 40.6472, 40.6824, 40.781, 40.8068, 40.6894, 40.6308, 40.5885, 40.6143] };
  WCL[557] = struct {
   x = [26.0593, 25.8926, 25.6063, 25.3317, 25.198, 25.0337, 24.8577, 24.7568, 24.6253, 24.4259, 24.2381, 23.9424, 23.6655, 23.8204, 23.9424, 24.0856, 24.1841, 24.2616, 24.3062, 24.2287, 24.1067, 23.9307,
        23.8204, 23.7429, 23.9424],
   y = [40.7317, 40.8396, 40.8983, 40.957, 40.9734, 41.0086, 40.9241, 40.8983, 40.8983, 40.957, 40.8068, 40.7575, 40.6214, 40.504, 40.4712, 40.4289, 40.3538, 40.2787, 40.2013, 40.2435, 40.328, 40.3703,
        40.3796, 40.2529, 40.1426] };
  WCL[558] = struct {
   x = [23.9635, 23.8204, 23.7101, 23.5341, 23.4566, 23.5012],
   y = [39.9643, 40.0839, 40.2177, 40.2693, 40.2529, 40.1004] };
  WCL[559] = struct {
   x = [23.5669, 23.3792, 23.2806, 23.0929, 22.8605, 22.8511, 22.9051, 22.8159, 22.6517, 22.6071, 22.6071],
   y = [39.9408, 40.0839, 40.26, 40.3609, 40.4712, 40.5228, 40.6214, 40.5885, 40.5557, 40.3538, 40.152] };
  WCL[560] = struct {
   x = [30.1052, 29.9714, 29.9292, 29.8189, 29.8072, 29.718, 29.6969, 29.664, 29.664, 29.664, 29.6757, 29.6969, 29.6405],
   y = [45.8642, 45.8642, 45.8032, 45.7656, 45.6718, 45.7023, 45.8267, 45.8806, 45.7586, 45.6342, 45.5192, 45.4394, 45.3854] };
  WCL[561] = struct {
   x = [26.4113, 26.5333, 26.6765, 26.8079, 26.9299, 27.0402, 27.1388, 27.3148, 27.4134, 27.6246, 27.8335, 27.7889, 27.7678, 27.9226, 28.0329, 28.0212, 28.1432, 28.385, 28.6056, 28.8802, 29.1125, 29.0914,
        28.9365, 28.913, 29.0914, 29.3331, 29.5208, 29.7086, 29.8963, 29.8846, 29.7414, 29.6311, 29.4645, 29.3871, 29.3331, 29.2556, 29.2228, 29.1125, 29.1571, 29.2345, 29.2674, 29.2885, 29.4551,
        29.6851, 29.9292],
   y = [40.0675, 40.2107, 40.3116, 40.4383, 40.4289, 40.4289, 40.4547, 40.497, 40.4219, 40.3703, 40.3796, 40.497, 40.5721, 40.5557, 40.5134, 40.4383, 40.4289, 40.4641, 40.4547, 40.4454, 40.4712, 40.565,
        40.5885, 40.6401, 40.6988, 40.7059, 40.7223, 40.7317, 40.7411, 40.8068, 40.8326, 40.8068, 40.8162, 40.8232, 40.8819, 40.8983, 40.9828, 41.0321, 41.1236, 41.2315, 41.2738, 41.2245, 41.2245,
        41.1658, 41.1729] };
  WCL[562] = struct {
   x = [26.2235, 26.3127, 26.3221, 26.4113],
   y = [39.9056, 40.0417, 40.0417, 40.0675] };
  WCL[563] = struct {
   x = [19.9622, 19.8401, 19.521, 19.3543, 19.2769, 19.4107, 19.4107, 19.345, 19.4107, 19.4435, 19.4318, 19.4647, 19.3872, 19.4107, 19.4435, 19.4764, 19.5538, 19.4647, 19.3098],
   y = [39.915, 40.0487, 40.1426, 40.3444, 40.4712, 40.4547, 40.5721, 40.7645, 40.9405, 40.9734, 41.1072, 41.2245, 41.3724, 41.4568, 41.5296, 41.6047, 41.7948, 41.8699, 41.8934] };
  WCL[564] = struct {
   x = [19.3098, 19.2112, 19.1009, 18.946, 18.7371, 18.6268, 18.5494, 18.3968, 18.2302, 18.0542, 17.8876, 17.7233, 17.5919, 17.3032, 17.0732, 17.1389, 17.3384, 17.437, 17.5356, 17.5144, 17.3595, 17.1272,
        16.8644, 16.5428, 16.3222, 16.0688, 15.9913, 15.9585, 15.881, 15.6839, 15.5619, 15.3741, 15.177, 15.1652, 15.2216, 15.407, 15.2098, 15.0338, 14.8907, 14.8907, 14.8343, 14.7569, 14.7029,
        14.6583, 14.548, 14.3274, 14.2054, 14.1068, 14.0083, 13.9097, 13.8205, 13.6656, 13.5788, 13.5342, 13.5107, 13.6539, 13.6985, 13.6328],
   y = [41.8934, 41.952, 42.1327, 42.2125, 42.3604, 42.419, 42.466, 42.5411, 42.6373, 42.7359, 42.825, 42.8321, 42.8579, 42.987, 43.0269, 43.0832, 43.0503, 42.9612, 42.9541, 43.0175, 43.0996, 43.285,
        43.4282, 43.5244, 43.5408, 43.5244, 43.5338, 43.63, 43.7497, 43.8201, 43.9093, 43.9961, 44.1862, 44.259, 44.3129, 44.3458, 44.4631, 44.557, 44.7706, 44.9583, 45.0897, 45.1695, 45.1695, 45.2071,
        45.2939, 45.3479, 45.1132, 44.9818, 44.8551, 44.841, 44.8879, 45.1132, 45.2306, 45.4183, 45.5028, 45.5802, 45.6952, 45.5568] };
  WCL[565] = struct {
   x = [14.5034, 14.5152, 14.5269, 14.5363, 14.5598, 14.6255, 14.6912, 14.7123, 14.7123, 14.6583, 14.5809, 14.4706, 14.4166, 14.4377, 14.4589, 14.5034],
   y = [45.2376, 45.2376, 45.2071, 45.1766, 45.1132, 45.0827, 45.0428, 45.0053, 44.9888, 45.0053, 45.0358, 45.0428, 45.0757, 45.1461, 45.2, 45.2376] };
  WCL[566] = struct {
   x = [14.2828, 14.2711, 14.3063, 14.3157, 14.3392, 14.3392, 14.3392, 14.3392, 14.3603, 14.372, 14.372, 14.372, 14.372, 14.3931, 14.426, 14.4377, 14.4166, 14.3157, 14.2617, 14.2617, 14.2828, 14.3392,
        14.3392, 14.3063, 14.2828, 14.2828],
   y = [45.1601, 45.1836, 45.193, 45.1836, 45.1461, 45.1296, 45.0991, 45.0592, 45.0217, 44.9747, 44.9489, 44.8879, 44.848, 44.794, 44.7142, 44.6837, 44.7002, 44.8246, 44.8879, 44.9114, 44.9184, 44.9747,
        45.0358, 45.0897, 45.1461, 45.1601] };
  WCL[567] = struct {
   x = [16.3903, 16.4114, 16.4232, 16.4325, 16.4771, 16.5335, 16.5992, 16.6649, 16.7423, 16.808, 16.8292, 16.8198, 16.808, 16.7869, 16.7306, 16.6203, 16.5335, 16.4325, 16.3997, 16.3903],
   y = [43.4141, 43.4282, 43.4282, 43.4282, 43.4211, 43.4141, 43.4047, 43.3977, 43.3648, 43.3578, 43.332, 43.3249, 43.3179, 43.3179, 43.3179, 43.3015, 43.3015, 43.3179, 43.3648, 43.4141] };
  WCL[568] = struct {
   x = [16.6438, 16.6766, 16.7306, 16.8198, 16.9301, 17.0075, 17.0732, 17.0943, 17.0404, 16.984, 16.9066, 16.7963, 16.686, 16.6531, 16.6438, 16.6438],
   y = [42.9776, 43.0104, 43.0104, 43.0104, 43.0104, 43.0104, 42.994, 42.9447, 42.9447, 42.9447, 42.9377, 42.9377, 42.9377, 42.9377, 42.9612, 42.9776] };
  WCL[569] = struct {
   x = [9.8614, 10.1712, 10.2252, 10.258, 10.3589, 10.5021, 10.5561, 10.5467, 10.6007, 10.8001, 10.9198, 11.0958, 11.131, 11.1639, 11.3071, 11.4385, 11.6051, 11.7365, 11.8586, 12.0346, 12.1331, 12.3655,
        12.5955, 12.8607, 13.0484, 13.147, 13.4568, 13.6328, 13.7431, 13.8745, 14.0083, 14.0857, 14.25, 14.3814, 14.4049, 14.3814, 14.5598, 14.724, 14.8226, 14.9001, 14.9681, 14.9564, 14.9329, 15.0221,
        15.231, 15.3319],
   y = [44.083, 43.9891, 43.9257, 43.7661, 43.5573, 43.3085, 43.156, 43.0597, 43.0104, 42.872, 42.7828, 42.6209, 42.5247, 42.419, 42.4589, 42.4589, 42.3604, 42.2055, 42.0834, 42.067, 41.9591, 41.7361,
        41.5554, 41.4076, 41.2808, 41.3066, 41.2996, 41.2902, 41.2644, 41.1564, 41.0321, 40.8913, 40.8396, 40.7997, 40.7575, 40.6472, 40.673, 40.6894, 40.673, 40.5885, 40.4383, 40.3796, 40.3374,
        40.2365, 40.1684, 40.1004] };
  WCL[570] = struct {
   x = [15.4187, 15.5079, 15.6065, 15.6276, 15.6839],
   y = [39.9995, 40.1262, 40.1262, 40.0839, 40.0253] };
  WCL[571] = struct {
   x = [16.5335, 16.5874, 16.6203, 16.7752, 16.984, 17.1835, 17.2821, 17.4816, 17.735, 17.9322, 17.9979],
   y = [39.915, 40.0675, 40.1849, 40.3538, 40.5298, 40.5557, 40.4876, 40.3867, 40.3374, 40.2529, 40.1262] };
  WCL[572] = struct {
   x = [18.3288, 18.3851, 18.4719, 18.418, 18.3077, 18.1316, 18.0213, 17.9322, 17.7444, 17.4816, 17.2821, 17.1389, 16.9183, 16.6977, 16.4889, 16.2683, 16.1251, 16.0148, 15.9374, 15.881, 15.881, 15.9913,
        16.1251, 16.1462, 16.0922, 16.0242, 15.9256, 15.8388, 15.7613, 15.9045, 15.7707, 15.6065, 15.4304, 15.2098, 15.1981, 15.1981, 15.1981, 14.9681, 14.8907, 14.7804, 14.724, 14.6912, 14.6466,
        14.5692, 14.5152, 14.3274, 14.196, 14.0951, 13.9637, 13.898, 13.8534, 13.7877, 13.7313, 13.6774, 13.6116, 13.5342, 13.4568, 13.3465, 13.3253, 13.3136, 13.2362, 13.147, 13.0601, 12.9381,
        12.7949, 12.6401, 12.5532, 12.4429, 12.342, 12.288, 12.288, 12.288, 12.2998, 12.3209, 12.4312, 12.4523, 12.3983, 12.2669, 12.2434, 12.2223, 12.2552, 12.342, 12.4523, 12.518, 12.5415, 12.6283,
        12.8161, 13.093, 13.2244, 13.391, 13.5553, 13.6891, 13.6985, 13.6656, 13.6328, 13.6891],
   y = [39.915, 40.0065, 40.1332, 40.2365, 40.3609, 40.504, 40.6566, 40.7645, 40.8068, 40.8913, 40.9499, 41.0743, 41.1658, 41.2315, 41.2996, 41.3724, 41.4404, 41.4733, 41.4733, 41.5296, 41.6047, 41.6962,
        41.7619, 41.8863, 41.9262, 42.0013, 42.0013, 41.9849, 41.952, 41.9849, 41.952, 41.9426, 41.952, 41.952, 41.9262, 41.9192, 41.8441, 42.0248, 42.067, 42.1233, 42.1398, 42.1327, 42.2383, 42.2383,
        42.2876, 42.419, 42.5082, 42.6209, 42.8086, 42.9307, 43.0503, 43.1161, 43.2921, 43.4282, 43.5103, 43.63, 43.6699, 43.6699, 43.6863, 43.7168, 43.7497, 43.7966, 43.8459, 43.9328, 43.9727, 44.029,
        44.0689, 44.1627, 44.2824, 44.3998, 44.5429, 44.6368, 44.7706, 44.8551, 44.9184, 45.0053, 45.0897, 45.1226, 45.2071, 45.3174, 45.4089, 45.4488, 45.5192, 45.5192, 45.4957, 45.5098, 45.5802,
        45.6718, 45.7351, 45.7023, 45.7656, 45.7492, 45.6718, 45.6178, 45.5802, 45.7257] };
  WCL[573] = struct {
   x = [5.13253, 5.12079, 5.06682, 4.9659, 4.93305, 4.7805, 4.62561, 4.58102, 4.5247, 4.39328, 4.25012, 4.18441, 4.08584, 3.91921, 3.82065, 3.76432, 3.62117, 3.48974, 3.26914, 3.15884, 3.11425, 3.13772,
        3.17996, 3.22455],
   y = [43.3836, 43.3836, 43.393, 43.393, 43.3437, 43.3367, 43.3766, 43.4235, 43.4493, 43.4657, 43.4798, 43.5291, 43.5291, 43.4657, 43.393, 43.3672, 43.2968, 43.2475, 43.1841, 43.0386, 42.8438, 42.8204,
        42.7875, 42.3956] };
  WCL[574] = struct {
   x = [3.22455, 3.35832, 3.35832, 3.39118, 3.23629, 3.25741, 3.32312, 3.15884, 2.82794, 2.46418, 2.22245, 1.90094, 1.50432, 1.14056, 0.919959, 0.919959, 0.767415, 0.546813, 0.403656],
   y = [42.3956, 42.3557, 42.3322, 42.2571, 42.2008, 42.0694, 41.9473, 41.7901, 41.6258, 41.4686, 41.2926, 41.2104, 41.0931, 40.9945, 40.7856, 40.6941, 40.5744, 40.3656, 40.1802] };
  WCL[575] = struct {
   x = [4.30644, 4.32756],
   y = [40.0276, 40.0276] };
  WCL[576] = struct {
   x = [4.08584, 3.95207, 4.02952, 4.19614, 4.26185, 4.30644],
   y = [39.9103, 40.0112, 40.0699, 40.0793, 40.0628, 40.0276] };
  WCL[577] = struct {
   x = [-0.105608, 0.049284, 0.23703, 0.457633, 0.612524, 0.64538, 0.535078, 0.269886, 0.258152, 0.380187, 0.657114, 0.919959, 1.23913],
   y = [49.2929, 49.314, 49.4009, 49.415, 49.429, 49.4666, 49.4877, 49.4948, 49.5581, 49.7083, 49.8233, 49.8726, 49.943] };
  WCL[578] = struct {
   x = [7.54742, 7.52395, 7.42539, 7.29396, 7.18366, 7.16019, 7.04989, 6.95133, 6.84102, 6.73072, 6.67675, 6.66501, 6.62042, 6.489, 6.35523, 6.25666, 6.19095, 6.16748, 6.03606, 5.90464, 5.84831, 5.79434,
        5.73801, 5.61832, 5.52914, 5.46343, 5.43058, 5.32027, 5.18651, 5.09967, 5.04335, 5.13253],
   y = [43.8365, 43.8131, 43.7732, 43.7262, 43.6863, 43.63, 43.5807, 43.5103, 43.4376, 43.3742, 43.3484, 43.2686, 43.2287, 43.1888, 43.1489, 43.1231, 43.0902, 43.0996, 43.1067, 43.1067, 43.1489, 43.1724,
        43.2193, 43.2193, 43.2287, 43.2522, 43.3179, 43.3484, 43.3578, 43.3883, 43.4211, 43.3836] };
  WCL[579] = struct {
   x = [7.54742, 7.69997, 7.86659, 8.06607, 8.15291, 8.18576, 8.26321, 8.35239, 8.43922, 8.47208, 8.56125, 8.62697, 8.79359, 8.94614, 9.12449, 9.30989, 9.65253, 9.8614],
   y = [43.8365, 43.8295, 43.8459, 43.8999, 43.9398, 44.036, 44.1158, 44.2097, 44.2425, 44.3223, 44.3528, 44.3763, 44.4232, 44.4232, 44.3927, 44.3223, 44.1698, 44.083] };
  WCL[580] = struct {
   x = [9.13388, 9.19959, 9.21133, 9.26765, 9.34275, 9.37795, 9.43193, 9.47652, 9.49764, 9.5305, 9.56335, 9.59856, 9.67365, 9.70886, 9.70886, 9.69712, 9.66427, 9.67365],
   y = [41.316, 41.3066, 41.2996, 41.2738, 41.2574, 41.2151, 41.1917, 41.1564, 41.1236, 41.0743, 41.0086, 40.9077, 40.7223, 40.5228, 40.4641, 40.3444, 40.2271, 40.1426] };
  WCL[581] = struct {
   x = [8.49554, 8.46269, 8.48381, 8.47208, 8.43922, 8.38524, 8.35239, 8.26321, 8.20923, 8.20923, 8.20923, 8.21862, 8.20923, 8.21862, 8.29606, 8.38524, 8.56125, 8.68329, 8.80298, 8.92501, 9.03532, 9.08929,
        9.13388],
   y = [39.9736, 40.1004, 40.2107, 40.3538, 40.4641, 40.5721, 40.5885, 40.5885, 40.673, 40.7411, 40.8326, 40.8983, 40.9664, 40.9828, 40.9147, 40.8654, 40.8913, 40.9828, 41.0743, 41.1564, 41.1917, 41.2574,
        41.316] };
  WCL[582] = struct {
   x = [9.34275, 9.43193, 9.44366, 9.45305, 9.45305, 9.49764, 9.55397, 9.52111, 9.45305, 9.4202, 9.37795, 9.34275, 9.30989, 9.27704, 9.22306, 9.21133, 9.12449, 9.02358, 8.90389, 8.84757, 8.8593, 8.73727,
        8.70441, 8.69268, 8.69268, 8.68329, 8.61523, 8.60584, 8.62697, 8.61523, 8.61523, 8.61523, 8.62697, 8.6387, 8.65982, 8.70441, 8.78186, 8.90389, 9.01419, 9.10103, 9.26765, 9.30989, 9.32163,
        9.34275, 9.34275],
   y = [43.0339, 43.0034, 42.8579, 42.7194, 42.6044, 42.4425, 42.2383, 42.0905, 41.9685, 41.8276, 41.7455, 41.6375, 41.5883, 41.5648, 41.4733, 41.3911, 41.4967, 41.5132, 41.5648, 41.6375, 41.7525, 41.7877,
        41.8699, 41.9849, 42.0905, 42.1562, 42.229, 42.2947, 42.3533, 42.4097, 42.4331, 42.4425, 42.4589, 42.4754, 42.5575, 42.6209, 42.6044, 42.6373, 42.6936, 42.7523, 42.7828, 42.8485, 42.9307,
        42.9776, 43.0339] };
  WCL[583] = struct {
   x = [-1.27433, -1.30719, -1.25086, -1.16403],
   y = [60.0039, 59.8889, 59.91, 59.9828] };
  WCL[584] = struct {
   x = [-6.14636, -6.16748, -6.25666, -6.45614, -6.57583, -6.7096, -6.7096, -6.67675, -6.78705, -6.91847, -6.97245, -6.97245, -6.97245, -6.91847, -6.86449, -6.93959, -6.95133, -6.91847, -6.79643, -6.73072,
        -6.71899, -6.64389, -6.57583, -6.45614, -6.38809, -6.31299, -6.28952, -6.28952, -6.35523, -6.35523, -6.34584, -6.32237, -6.27778, -6.17922, -6.09238, -6.13463, -6.19095, -6.16748, -6.1581,
        -6.14636, -6.14636],
   y = [58.4456, 58.4526, 58.4362, 58.3775, 58.3142, 58.2672, 58.2273, 58.1733, 58.1569, 58.2039, 58.11, 58.0466, 57.9809, 57.941, 57.8941, 57.8354, 57.8072, 57.7462, 57.7697, 57.8002, 57.8542, 57.8776,
        57.9293, 57.9222, 57.941, 57.9879, 58.0396, 58.0584, 58.063, 58.1217, 58.1288, 58.1522, 58.1733, 58.1804, 58.2203, 58.2156, 58.2555, 58.3071, 58.3611, 58.4127, 58.4456] };
  WCL[585] = struct {
   x = [-7.02877, -6.97245, -6.98418, -7.08275, -7.22825, -7.32682, -7.39253, -7.40426, -7.34794, -7.30335, -7.24937, -7.16019, -7.09448, -7.06163, -7.06163, -7.04989, -7.02877],
   y = [57.6125, 57.6289, 57.6359, 57.6477, 57.6359, 57.6289, 57.6242, 57.589, 57.5773, 57.5397, 57.5233, 57.5162, 57.5162, 57.5162, 57.5467, 57.5819, 57.6125] };
  WCL[586] = struct {
   x = [-2.62846, -2.60733, -2.62846, -2.66366, -2.77161, -2.87253, -2.97109, -3.00395, -3.01568, -3.11425, -3.1917, -3.22455, -3.22455, -3.12598, -3.00395, -2.87253, -2.79508, -2.75049, -2.68478, -2.62846],
   y = [58.9455, 58.9689, 58.9736, 58.9971, 58.99, 59.0018, 59.0018, 59.0464, 59.0933, 59.1543, 59.1543, 59.0651, 58.99, 58.9572, 58.9502, 58.9619, 58.9502, 58.9408, 58.929, 58.9455] };
  WCL[587] = struct {
   x = [-3.10252, -3.09313, -3.14711, -3.22455, -3.29026, -3.302, -3.23629, -3.13772, -3.10252],
   y = [58.8305, 58.8657, 58.9009, 58.922, 58.9103, 58.8539, 58.814, 58.8211, 58.8305] };
  WCL[588] = struct {
   x = [-5.25456, -5.24283, -5.24283, -5.24283, -5.24283, -5.3414, -5.30854, -5.2311, -5.16538, -5.16538, -5.24283, -5.32966, -5.41884, -5.49629, -5.58547, -5.66057, -5.65118, -5.66057, -5.65118, -5.65118,
        -5.59485, -5.52914, -5.562, -5.68403, -5.73801, -5.73801, -5.73801, -5.73801, -5.73801, -5.72863, -5.60659, -5.4517, -5.39772, -5.41884, -5.51741, -5.48455, -5.40711, -5.41884, -5.50802,
        -5.51741, -5.52914, -5.51741, -5.50802, -5.55026, -5.61832, -5.61832, -5.61832, -5.68403, -5.72863, -5.73801, -5.72863, -5.72863, -5.73801, -5.82719, -5.92576, -6.02433, -6.01494, -5.88117,
        -5.84831, -5.83893, -5.82719, -5.81546, -5.72863, -5.63945, -5.59485, -5.58547, -5.562, -5.562, -5.49629, -5.40711, -5.32027, -5.27569, -5.24283, -5.20997, -5.19824, -5.18651, -5.18651,
        -5.26395, -5.30854, -5.36486, -5.39772, -5.41884, -5.43996, -5.46343, -5.46343, -5.50802, -5.52914, -5.54088, -5.54088, -5.54088, -5.43996, -5.39772, -5.40711, -5.47517, -5.562, -5.60659,
        -5.63945, -5.6723, -5.6723, -5.55026, -5.49629, -5.43996, -5.3414, -5.29681, -5.24283, -5.24283, -5.24283, -5.25456, -5.30854, -5.32027, -5.29681, -5.26395, -5.19824, -5.06682, -5.04335,
        -5.03396, -5.04335, -5.11141, -5.17712, -5.18651, -5.17712, -5.13253, -5.08794, -5.0762, -5.05508, -5.01049, -4.9659, -4.95652, -4.91193, -4.8908, -4.87907, -4.86734, -4.8556, -4.8556, -4.8556,
        -4.78989, -4.69132, -4.59276, -4.58102, -4.67959, -4.72418, -4.75704, -4.76877, -4.76877, -4.7805, -4.72418, -4.61388, -4.58102, -4.67959, -4.76877, -4.8908, -4.93305, -4.92366, -5.04335,
        -5.0762, -5.01049, -4.93305, -4.8556, -4.81336, -4.83448, -4.81336, -4.7453, -4.62561, -4.43787, -4.3393, -4.28297, -4.28297, -4.229, -4.09523, -4.01778, -3.8535, -3.75494, -3.66576, -3.65402,
        -3.55545, -3.5226, -3.39118, -3.28088, -3.12598, -3.00395, -3.06027, -3.13772, -3.1917, -3.29026, -3.34659, -3.43342, -3.51086, -3.54372, -3.50148, -3.42403, -3.36771, -3.31373, -3.29026,
        -3.25741, -3.24802, -3.1917, -3.13772, -3.02742, -2.88191, -2.72937, -2.65192, -2.72937, -2.72937, -2.75049, -2.84906, -2.91712, -2.83967, -2.73876, -2.84906, -2.91712, -2.93824, -2.88191,
        -2.79508, -2.86079, -2.94997, -3.11425, -3.35832, -3.67749, -3.91921, -4.16094, -4.36042, -4.58102, -4.49184, -4.229, -4.02952, -3.99666, -3.99666, -3.98493, -3.9638, -3.8535, -3.93095,
        -3.95207, -4.04125, -4.20553, -4.4144, -4.59276, -4.80162, -5.00111, -5.15365, -5.11141, -5.0762, -5.02223, -4.97764, -4.94478, -4.93305, -4.82275, -4.69132, -4.5247, -4.36042, -4.29471,
        -4.27124, -4.25012, -4.18441, -3.9638, -3.83238, -3.77606, -3.72208, -3.65402, -3.54372, -3.44515, -3.33485, -3.17996, -3.06966, -2.98283, -2.86079, -2.75049, -2.58621, -2.47591, -2.46418,
        -2.55336, -2.7059, -2.83967, -2.88191, -2.93824, -3.00395, -3.17996, -3.48974, -3.75494, -4.01778, -4.02952, -4.1187, -4.29471, -4.4144, -4.65847, -4.86734, -5.04335, -5.16538, -5.32027,
        -5.4517, -5.50802],
   y = [58.2156, 58.2086, 58.1569, 58.1053, 58.0818, 58.0513, 58.0044, 57.9692, 57.9363, 57.9222, 57.8941, 57.8706, 57.8589, 57.8166, 57.8119, 57.7814, 57.711, 57.6758, 57.6171, 57.5773, 57.5585, 57.535,
        57.5162, 57.528, 57.5115, 57.481, 57.4646, 57.4576, 57.3919, 57.3613, 57.3567, 57.3754, 57.3801, 57.3332, 57.3027, 57.2909, 57.2862, 57.2675, 57.2018, 57.1713, 57.1595, 57.1173, 57.0938,
        57.0868, 57.0398, 57.0281, 56.9976, 56.9624, 56.8967, 56.8709, 56.8286, 56.7699, 56.7512, 56.7394, 56.7254, 56.6901, 56.6596, 56.6549, 56.6479, 56.6479, 56.6174, 56.5564, 56.5259, 56.4766,
        56.4531, 56.4531, 56.4602, 56.4649, 56.4907, 56.5212, 56.5681, 56.5939, 56.6104, 56.6244, 56.6174, 56.6057, 56.5517, 56.4649, 56.3921, 56.324, 56.2818, 56.2396, 56.1645, 56.1222, 56.08,
        56.0236, 55.9509, 55.9063, 55.8453, 55.8195, 55.8336, 55.796, 55.7326, 55.6716, 55.5778, 55.4909, 55.4158, 55.3337, 55.3032, 55.3032, 55.3454, 55.4721, 55.6083, 55.6763, 55.7467, 55.7819,
        55.8195, 55.8828, 55.9556, 56.0002, 56.0119, 56.0307, 56.08, 56.1785, 56.1832, 56.1903, 56.1222, 56.0354, 55.9509, 55.8758, 55.8382, 55.8758, 55.8828, 55.8828, 55.9251, 55.9439, 55.9321,
        55.9321, 55.918, 55.9016, 55.9016, 55.9063, 55.9133, 55.9368, 56.0495, 56.0236, 56.0002, 55.9556, 55.9368, 55.9251, 55.8758, 55.8195, 55.7631, 55.7139, 55.6646, 55.6153, 55.566, 55.4581,
        55.3642, 55.2257, 55.0873, 55.0004, 54.9042, 54.9605, 54.9042, 54.7892, 54.6625, 54.6249, 54.6367, 54.7775, 54.815, 54.8291, 54.7845, 54.6953, 54.6695, 54.7329, 54.815, 54.8221, 54.7963,
        54.7963, 54.7892, 54.8221, 54.8479, 54.8526, 54.8737, 54.9042, 54.9488, 54.9418, 54.9371, 54.923, 54.8925, 54.8925, 54.8596, 54.815, 54.7329, 54.6367, 54.5404, 54.4583, 54.3809, 54.3222,
        54.2846, 54.2002, 54.1884, 54.2189, 54.2002, 54.1556, 54.0781, 54.1039, 54.1298, 54.2189, 54.2002, 54.1368, 54.0523, 53.949, 53.8904, 53.766, 53.7519, 53.7331, 53.6487, 53.543, 53.4773,
        53.4257, 53.3999, 53.3717, 53.2872, 53.2943, 53.2872, 53.299, 53.2145, 53.0878, 52.9634, 52.8226, 52.7897, 52.8953, 52.9235, 52.8226, 52.7029, 52.602, 52.555, 52.5081, 52.4213, 52.3861,
        52.2664, 52.1983, 52.1091, 52.0364, 51.973, 51.9003, 51.851, 51.8181, 51.7501, 51.6937, 51.6327, 51.6186, 51.6046, 51.5928, 51.6257, 51.7219, 51.7688, 51.7501, 51.7078, 51.6046, 51.5647,
        51.5576, 51.6327, 51.6398, 51.6186, 51.5576, 51.4684, 51.398, 51.3722, 51.3793, 51.4262, 51.5083, 51.5647, 51.5928, 51.6679, 51.7219, 51.6468, 51.5576, 51.4332, 51.337, 51.2549, 51.2126,
        51.1915, 51.1774, 51.1845, 51.1774, 51.1446, 51.1164, 51.0249, 50.9498, 50.7902, 50.6212, 50.4734, 50.3537, 50.2622, 50.1988, 50.1354, 50.0439] };
  WCL[589] = struct {
   x = [-5.50802, -5.4517, -5.38599, -5.26395, -5.18651, -5.09967, -5.09967],
   y = [49.9946, 50.0721, 50.1002, 50.065, 50.0228, 50.0369, 50.0298] };
  WCL[590] = struct {
   x = [-5.04335, -4.9659, -4.91193, -4.71244, -4.60214, -4.4496, -4.29471, -4.17267, -4.06237, -3.94268, -3.80891, -3.71035, -3.61178, -3.56719, -3.5226, -3.44515, -3.40056, -3.36771, -3.31373, -3.26914,
        -3.17057, -2.98283, -2.77161, -2.69651, -2.60733, -2.50877, -2.44306, -2.39847],
   y = [49.9876, 50.0369, 50.1354, 50.2058, 50.3326, 50.3326, 50.3467, 50.3396, 50.3396, 50.3185, 50.2903, 50.2551, 50.241, 50.2481, 50.2833, 50.3467, 50.5015, 50.6142, 50.6283, 50.6212, 50.6212, 50.6776,
        50.7128, 50.7339, 50.6776, 50.6142, 50.5649, 50.5227] };
  WCL[591] = struct {
   x = [-2.39847, -2.375, -2.375, -2.375, -2.32102, -2.1544, -1.969, -1.86808, -1.9244, -1.7038, -1.45973, -1.40575, -1.28372, -1.2626, -1.2626, -1.1523, -1.03026, -0.87537, -0.833126, -0.732213, -0.546813,
        -0.19244, -0.037549],
   y = [50.5227, 50.5156, 50.5227, 50.5297, 50.579, 50.6072, 50.572, 50.6001, 50.6846, 50.6846, 50.6916, 50.7268, 50.762, 50.816, 50.8794, 50.8231, 50.8019, 50.8231, 50.7832, 50.7339, 50.755, 50.7972,
        50.8019] };
  WCL[592] = struct {
   x = [0.004694, -0.114995, -0.248765, -0.171319, -0.061018],
   y = [53.5571, 53.6228, 53.7003, 53.7261, 53.6557] };
  WCL[593] = struct {
   x = [0.093873, -0.072752, -0.138463, -0.248765, -0.258152, -0.258152, -0.269886, -0.359066, -0.523344, -0.732213, -0.985671, -1.05373, -1.12883, -1.2626, -1.3729, -1.42688, -1.43861, -1.45034, -1.45973,
        -1.49493, -1.65921, -1.85869, -1.97838, -2.14501, -2.35388, -2.49703, -2.5956, -2.64019, -2.72937, -2.90538, -3.04854, -3.17996, -3.1917, -3.14711, -3.03681, -2.89365, -2.69651, -2.52989,
        -2.61907, -2.72937, -2.8162, -2.84906, -2.67305, -2.56274, -2.45244, -2.41959, -2.30929, -2.2647, -2.23184, -2.19898, -2.17786, -2.16613, -2.14501, -2.06756, -2.04644, -2.01124, -1.9244,
        -1.83523, -1.75778, -1.75778, -1.75778, -1.82584, -1.9244, -2.13327, -2.375, -2.57448, -2.78335, -2.99221, -3.21282, -3.36771, -3.69861, -3.87462, -4.00839, -4.08584, -4.04125, -4.04125,
        -4.19614, -4.19614, -3.93095, -3.76432, -3.82065, -4.07411, -4.04125, -4.01778, -3.84177, -3.47801, -3.33485, -3.09313, -3.06966, -3.03681, -3.25741, -3.33485, -3.36771, -3.51086, -3.76432,
        -3.97554, -4.09523, -4.27124, -4.29471, -4.4496, -4.50358, -4.53643, -4.64673, -4.71244, -4.84621, -4.97764, -5.03396, -5.04335, -5.04335, -5.06682, -5.18651, -5.24283, -5.25456],
   y = [53.8364, 53.9749, 54.1298, 54.2447, 54.2447, 54.2518, 54.2518, 54.3621, 54.5029, 54.5545, 54.5803, 54.6179, 54.7587, 54.8784, 55.0192, 55.1319, 55.2445, 55.3266, 55.397, 55.5285, 55.5965, 55.7209,
        55.8382, 55.9016, 55.9556, 56.0002, 56.0119, 56.0049, 55.9861, 55.9439, 55.9556, 55.9861, 56.0049, 56.0847, 56.148, 56.1715, 56.1785, 56.2208, 56.2818, 56.3686, 56.3921, 56.4461, 56.4414,
        56.4836, 56.5681, 56.6549, 56.7394, 56.7699, 56.7981, 56.8403, 56.8474, 56.8709, 56.9436, 57.0586, 57.136, 57.2205, 57.2792, 57.3261, 57.3801, 57.4763, 57.5702, 57.6524, 57.6641, 57.6641,
        57.6477, 57.6711, 57.6758, 57.6594, 57.6946, 57.6641, 57.6054, 57.5702, 57.5045, 57.5397, 57.6125, 57.6406, 57.6171, 57.6641, 57.7063, 57.7955, 57.8072, 57.8119, 57.8894, 57.8776, 57.9926,
        58.117, 58.2555, 58.3376, 58.4761, 58.584, 58.6263, 58.6146, 58.6028, 58.5794, 58.5512, 58.5512, 58.5277, 58.5113, 58.5277, 58.5512, 58.5441, 58.5512, 58.5676, 58.5512, 58.5395, 58.4292,
        58.3728, 58.2907, 58.2273, 58.2203, 58.2156, 58.2203, 58.2156] };
  WCL[594] = struct {
   x = [-1.09597, -1.12883, -1.17342, -1.18515, -1.22974, -1.23913, -1.25086, -1.31658, -1.34943, -1.31658, -1.17342, -1.06312, -1.00914, -0.976283, -0.985671, -1.09597],
   y = [50.6987, 50.7268, 50.7128, 50.7128, 50.6987, 50.6987, 50.6987, 50.6705, 50.6353, 50.6072, 50.579, 50.5931, 50.6353, 50.6564, 50.6776, 50.6987] };
  WCL[595] = struct {
   x = [-4.09523, -3.99666, -4.09523, -4.21726, -4.30644, -4.43787],
   y = [53.2215, 53.2802, 53.3389, 53.3858, 53.3928, 53.3788] };
  WCL[596] = struct {
   x = [-4.30644, -4.31583, -4.40501, -4.51531, -4.62561, -4.67959, -4.5247, -4.32756, -4.23838, -4.229, -4.229, -4.30644],
   y = [54.3738, 54.3691, 54.3105, 54.233, 54.1039, 54.0453, 54.064, 54.1298, 54.2189, 54.2776, 54.348, 54.3738] };
  WCL[597] = struct {
   x = [-5.2311, -5.24283, -5.25456, -5.24283, -5.16538, -5.06682, -5.04335, -5.02223, -5.04335, -5.08794, -5.17712, -5.20997, -5.22171, -5.2311],
   y = [55.6575, 55.6646, 55.5895, 55.4768, 55.4017, 55.4205, 55.4909, 55.566, 55.6646, 55.6716, 55.6646, 55.6575, 55.6575, 55.6575] };
  WCL[598] = struct {
   x = [-6.08065, -6.09238, -6.11351, -6.16748, -6.32237, -6.39982, -6.38809, -6.2684, -6.23554, -6.19095, -6.16748, -6.0032, -5.99147, -6.05718, -6.08065, -6.08065],
   y = [55.8641, 55.8758, 55.8641, 55.8382, 55.8077, 55.6904, 55.6646, 55.7514, 55.7139, 55.6388, 55.6036, 55.6528, 55.7397, 55.8195, 55.857, 55.8641] };
  WCL[599] = struct {
   x = [-5.65118, -5.62771, -5.66057, -5.7826, -5.83893, -5.84831, -5.93749, -5.94923, -5.84831, -5.80372, -5.69577, -5.65118, -5.65118],
   y = [56.0729, 56.0917, 56.08, 56.0119, 55.9626, 55.9063, 55.8453, 55.7819, 55.8007, 55.8758, 55.9814, 56.0542, 56.0729] };
  WCL[600] = struct {
   x = [-6.24493, -6.27778, -6.2684, -6.16748, -5.98208, -5.86005, -5.81546, -5.73801, -5.65118, -5.6723, -5.74975, -5.77087, -5.84831, -5.94923, -5.95861, -5.97035, -5.99147, -6.11351, -6.17922, -6.16748,
        -6.05718, -6.01494, -6.03606, -6.03606, -6.0032, -6.03606, -6.12524, -6.20268, -6.24493],
   y = [56.2888, 56.3123, 56.2748, 56.2443, 56.2443, 56.2888, 56.3193, 56.3193, 56.3381, 56.3991, 56.4343, 56.4414, 56.4836, 56.5446, 56.5517, 56.5564, 56.5986, 56.5869, 56.5446, 56.4907, 56.4602, 56.4414,
        56.3921, 56.3616, 56.3499, 56.3123, 56.2888, 56.2888, 56.2888] };
  WCL[601] = struct {
   x = [-6.24493, -6.27778, -6.27778, -6.30125, -6.32237, -6.33411, -6.47727, -6.54298, -6.58757, -6.63216, -6.55471, -6.42329, -6.34584, -6.24493, -6.24493, -6.16748, -6.03606, -5.88117, -5.88117, -5.87178,
        -5.76148, -5.70516, -5.65118, -5.7826, -5.8929, -5.95861, -5.95861, -5.99147, -6.09238, -6.09238, -6.12524, -6.22381, -6.24493],
   y = [57.6711, 57.6758, 57.6641, 57.6007, 57.5115, 57.4693, 57.5467, 57.4998, 57.4693, 57.3801, 57.352, 57.352, 57.237, 57.183, 57.1243, 57.1243, 57.129, 57.1478, 57.1008, 57.0469, 57.0703, 57.1548,
        57.2135, 57.2205, 57.2323, 57.2675, 57.3684, 57.3919, 57.4646, 57.528, 57.5937, 57.6594, 57.6711] };
  WCL[602] = struct {
   x = [-7.44651, -7.44651, -7.43712, -7.45824, -7.46998, -7.4911, -7.6014, -7.69058, -7.77741, -7.84547, -8.0426, -8.18576, -8.27494, -8.32892, -8.28668, -8.3078, -8.42983, -8.59411, -8.62697, -8.6387,
        -8.59411, -8.43922, -8.34065, -8.24208, -8.07546, -8.06607, -8.13178, -8.20923, -8.35239, -8.45095, -8.43922, -8.45095, -8.46269, -8.73727, -8.90389, -9.06817, -9.13388, -9.32163, -9.58682,
        -9.69712, -9.78396, -9.7511, -9.71824, -9.82855, -9.99517, -9.8614, -9.58682, -9.47652, -9.44366, -9.59856, -9.72998, -9.79569, -9.93885, -9.95058, -9.82855, -9.74171, -9.74171, -9.61968,
        -9.50938, -9.47652, -9.30051, -9.02358, -8.88042, -8.82645, -8.90389, -9.13388, -9.26765, -9.28877, -9.30051, -9.37795, -9.47652, -9.59856, -9.63141, -9.6408, -9.66427, -9.6408, -9.52111,
        -9.49764, -9.39907, -9.23245, -9.06817, -8.92501, -8.83583, -8.82645, -8.78186, -8.77012, -8.78186, -8.77012, -8.61523, -8.60584, -8.76074, -8.99072, -9.21133, -9.49764, -9.56335, -9.69712,
        -9.70886, -9.76283, -9.87313, -9.98344],
   y = [54.9418, 54.9863, 54.9863, 55.0685, 55.2023, 55.2257, 55.214, 55.1764, 55.1577, 55.1882, 55.1131, 55.0309, 55.0122, 54.9112, 54.8667, 54.8338, 54.7517, 54.6883, 54.6742, 54.6695, 54.6062, 54.6062,
        54.585, 54.6062, 54.6062, 54.5592, 54.47, 54.4513, 54.3926, 54.3433, 54.2964, 54.2518, 54.233, 54.2401, 54.2189, 54.2072, 54.2706, 54.2846, 54.2964, 54.2189, 54.1157, 54.0382, 53.949, 53.9749,
        53.9537, 53.8693, 53.8434, 53.8434, 53.7777, 53.7191, 53.6745, 53.59, 53.5313, 53.4515, 53.3788, 53.3647, 53.306, 53.299, 53.3131, 53.2403, 53.2286, 53.2286, 53.2286, 53.1675, 53.1277, 53.069,
        52.9821, 52.9094, 52.8226, 52.7217, 52.6559, 52.5808, 52.5761, 52.5621, 52.555, 52.548, 52.5621, 52.5761, 52.5761, 52.5879, 52.609, 52.6888, 52.7615, 52.7686, 52.7897, 52.7756, 52.7029, 52.663,
        52.6489, 52.6278, 52.5949, 52.5761, 52.5339, 52.5222, 52.4471, 52.3321, 52.2593, 52.2241, 52.2382, 52.2124] };
  WCL[603] = struct {
   x = [-10.0398, -9.80742, -9.77457],
   y = [52.0833, 52.1021, 52.0481] };
  WCL[604] = struct {
   x = [-10.0046, -9.9717, -9.87313, -9.72998, -9.56335, -9.55397, -9.71824, -9.85201, -9.92946, -9.96232, -9.88487, -9.71824, -9.60794, -9.45305, -9.4202, -9.38734, -9.43193, -9.54223, -9.65253, -9.61968,
        -9.61968, -9.45305, -9.30989, -9.24418, -9.15735, -9.12449, -9.11276, -9.10103, -9.08929, -9.03532, -9.02358, -8.91328, -8.79359, -8.67156, -8.5284, -8.40636, -8.39698, -8.35239, -8.26321,
        -8.24208, -8.25147, -8.18576, -8.15291, -8.13178, -8.07546, -7.96516, -7.84547, -7.77741, -7.72343, -7.7117, -7.61313, -7.54742, -7.50283, -7.45824, -7.24937, -7.01704, -6.91847, -6.80817,
        -6.57583, -6.44441, -6.31299, -6.25666, -6.32237, -6.2684, -6.20268, -6.14636, -6.12524, -6.03606, -5.99147, -5.99147, -6.03606, -6.03606, -6.05718, -6.08065, -6.10177, -6.03606, -6.01494,
        -6.0032, -6.06892, -6.14636, -6.20268, -6.24493, -6.23554, -6.20268, -6.08065, -6.08065],
   y = [51.7501, 51.7501, 51.7688, 51.797, 51.8322, 51.7618, 51.7149, 51.6468, 51.6046, 51.5858, 51.5858, 51.6327, 51.6398, 51.675, 51.6867, 51.6538, 51.5858, 51.5787, 51.5365, 51.5083, 51.4614, 51.5083,
        51.5224, 51.4966, 51.4896, 51.4966, 51.4966, 51.5036, 51.5154, 51.5365, 51.5435, 51.5365, 51.5717, 51.5787, 51.6186, 51.6468, 51.6468, 51.675, 51.7149, 51.79, 51.8721, 51.858, 51.797, 51.79,
        51.7759, 51.8111, 51.8439, 51.919, 51.919, 51.919, 51.905, 51.9542, 52.0364, 52.0763, 52.1091, 52.1302, 52.1631, 52.1772, 52.1772, 52.1443, 52.1443, 52.2171, 52.325, 52.3931, 52.5222, 52.6818,
        52.6958, 52.8085, 52.9024, 52.9282, 53.076, 53.1206, 53.2004, 53.2802, 53.3271, 53.3459, 53.4116, 53.5172, 53.5947, 53.7191, 53.8059, 53.8693, 53.9537, 53.9608, 53.9796, 53.9866] };
  WCL[605] = struct {
   x = [-6.08065, -6.06892, -6.0032, -5.83893, -5.79434, -5.562, -5.46343, -5.46343, -5.562, -5.59485, -5.562, -5.4517, -5.41884, -5.50802, -5.59485, -5.84831, -5.74975, -5.76148, -5.93749, -6.04779,
        -6.2684, -6.35523, -6.62042, -6.79643, -6.91847, -7.04989, -7.08275, -6.97245, -6.89735, -6.93959, -7.08275, -7.18366, -7.24937, -7.24937, -7.2705, -7.32682, -7.35968, -7.33855, -7.33855,
        -7.37141, -7.42539, -7.44651],
   y = [53.9866, 53.9866, 54.0124, 54.0781, 54.1884, 54.2189, 54.2917, 54.2964, 54.4184, 54.4959, 54.4771, 54.3691, 54.4912, 54.6062, 54.6367, 54.6109, 54.7212, 54.9112, 55.038, 55.1506, 55.1882, 55.214,
        55.1577, 55.1319, 55.0685, 55.038, 55.0755, 55.1506, 55.2069, 55.2445, 55.3032, 55.3219, 55.3149, 55.2961, 55.2633, 55.2398, 55.1764, 55.1131, 55.0309, 54.9676, 54.9418, 54.9418] };
  WCL[606] = struct {
   x = [-9.98344, -10.1055, -10.2158, -10.258, -10.0398],
   y = [52.2124, 52.2241, 52.1631, 52.0692, 52.0833] };
  WCL[607] = struct {
   x = [-9.77457, -10.0046, -10.1594, -10.1829, -10.0609, -10.0046, -10.0046],
   y = [52.0481, 51.9801, 51.8721, 51.804, 51.8111, 51.7618, 51.7501] };
  WCL[608] = struct {
   x = [-43.9633, -43.9633, -43.9844, -43.9961, -44.029, -44.029, -43.9961, -43.8858, -43.7661, -43.6652, -43.6558, -43.6887, -43.8647, -43.9093, -43.9633],
   y = [59.9476, 59.7504, 59.9476, 59.9147, 59.8607, 59.8326, 59.8161, 59.8161, 59.8279, 59.8607, 59.8936, 59.91, 59.9476, 59.9476, 59.9476] };
  WCL[609] = struct {
   x = [-43.9187, -43.8318, -43.6769, -43.5455, -43.5127],
   y = [60.0203, 59.971, 59.9663, 59.971, 59.971] };
  WCL[610] = struct {
   x = [-60.0062, -59.9288, -59.8396, -59.6636, -59.6425, -59.5533, -59.4547, -59.5416, -59.6753, -59.5322, -59.3116, -59.1356, -59.1567, -59.1003, -58.9807, -58.9478, -58.8915, -58.7694, -58.5512, -58.4831,
        -58.3071, -58.1757, -58.1874, -58.11, -58.0208, -57.9222, -57.8776, -57.9222, -57.8331, -57.7228, -57.6031, -57.481, -57.3825, -57.481, -57.4693, -57.4928, -57.657, -57.7791, -57.9105,
        -58.0208, -58.1522, -58.2203, -58.2414, -58.2414, -58.2977, -58.5277, -58.7601, -59.0675, -59.3773, -59.5322, -59.2998, -59.0464, -58.7812, -58.6826, -58.99, -59.2881, -59.5744, -59.8396],
   y = [55.2844, 55.2398, 55.2891, 55.3407, 55.2445, 55.2328, 55.214, 55.0755, 54.9042, 54.9746, 55.1506, 55.221, 55.0873, 55.0873, 55.1577, 55.0309, 54.9371, 54.8784, 54.8479, 54.8291, 54.7775, 54.8103,
        54.8854, 54.923, 54.9605, 54.9042, 54.8596, 54.7775, 54.7517, 54.7446, 54.6953, 54.6813, 54.5803, 54.5146, 54.4583, 54.4395, 54.4137, 54.4325, 54.4255, 54.3996, 54.3621, 54.3222, 54.3222,
        54.3034, 54.2917, 54.2072, 54.1814, 54.1556, 54.0899, 54.0007, 54.0054, 54.0593, 54.1415, 54.1227, 54.0265, 53.949, 53.8951, 53.8951] };
  WCL[611] = struct {
   x = [-60.1048, -59.8513, -59.7622, -59.5533, -59.3327, -59.0252, -58.8375, -58.6263, -58.3517, -58.0983, -57.8237, -57.7016, -57.8659, -58.1311, -58.3846, -58.4292, -58.2743, -57.9434, -57.6688, -57.4576,
        -57.2604, -57.2041, -57.237, -57.1947, -57.1713, -57.305, -57.3144, -57.2816, -57.2722, -57.237, -57.1384, -57.007, -56.8521, -56.6104, -56.4766, -56.5329, -56.4109, -56.1903, -55.9697,
        -55.8594, -55.904, -55.8922, -55.8922, -56.0025, -55.8922, -55.7819, -55.7608, -55.6834, -55.6834, -55.7819, -55.8594, -56.0589, -56.2795, -56.2349, -56.1363, -56.0354, -55.8922, -55.8265,
        -55.8148, -55.8922, -55.8922, -55.8148, -55.7397],
   y = [53.3342, 53.4374, 53.5243, 53.6088, 53.7261, 53.8059, 53.942, 54.0265, 54.0781, 54.1039, 54.0711, 54.0899, 54.1368, 54.1743, 54.1415, 54.1368, 54.1814, 54.226, 54.2072, 54.2002, 54.2072, 54.1485,
        54.0382, 53.935, 53.8059, 53.6932, 53.5947, 53.4656, 53.4116, 53.4187, 53.543, 53.7073, 53.7777, 53.7918, 53.766, 53.7073, 53.6815, 53.6416, 53.6017, 53.5689, 53.59, 53.5829, 53.5313, 53.4914,
        53.3928, 53.3529, 53.2802, 53.2614, 53.1746, 53.1488, 53.0878, 53.1347, 53.1887, 53.1417, 53.0549, 53.0291, 53.0291, 52.9235, 52.8836, 52.8366, 52.8014, 52.7099, 52.6419] };
  WCL[612] = struct {
   x = [-55.7397, -55.7279, -55.7937, -55.904, -55.9697, -55.8594, -55.6716, -55.5848, -55.85, -55.9814, -55.8922, -55.7045, -55.6059, -55.7819, -55.9697, -56.3006, -56.4437, -56.5775, -56.7864, -56.8638,
        -57.1947, -57.535, -57.8002, -58.0983, -58.3517, -58.5723, -58.7694, -58.8704, -59.0018, -59.091, -59.2552, -59.3984, -59.619, -59.8185, -59.9828],
   y = [52.6419, 52.6231, 52.6231, 52.6348, 52.5691, 52.548, 52.5081, 52.379, 52.3931, 52.4213, 52.3462, 52.2593, 52.1772, 52.0622, 51.905, 51.7688, 51.6468, 51.5365, 51.5224, 51.4684, 51.4755, 51.4825,
        51.398, 51.3511, 51.33, 51.2338, 51.0531, 51.0671, 50.9146, 50.8231, 50.6846, 50.5649, 50.4734, 50.3255, 50.2622] };
  WCL[613] = struct {
   x = [-56.202, -56.3334, -56.3334, -56.4109, -56.4883, -56.6315, -56.6643, -56.7418, -56.7746, -56.8638, -56.8638, -56.9741, -56.9835, -57.0164, -57.1173, -57.2722, -57.359, -57.305, -57.2041, -57.237,
        -57.3261, -57.3825, -57.4364, -57.4247, -57.359, -57.2933, -57.3473, -57.4364, -57.6125, -57.6359],
   y = [51.5083, 51.4755, 51.4684, 51.4544, 51.4192, 51.3652, 51.269, 51.2056, 51.1446, 51.1235, 51.0883, 51.0531, 50.9498, 50.8653, 50.7902, 50.755, 50.7057, 50.6705, 50.6283, 50.5931, 50.5438, 50.403,
        50.2762, 50.2481, 50.2199, 50.1706, 50.1495, 50.1495, 50.1002, 50.0369] };
  WCL[614] = struct {
   x = [-56.08, -56.0354, -56.1692],
   y = [49.9876, 50.1214, 50.1284] };
  WCL[615] = struct {
   x = [-56.7207, -56.7207, -56.5094, -56.378, -56.1903, -56.1692, -56.1903, -56.1692, -56.0917, -55.8922, -55.8148, -55.7608, -55.8148, -55.9251, -55.904, -55.6716, -55.5191, -55.5285, -55.6951, -55.7819,
        -55.9697, -56.202],
   y = [49.9148, 50.065, 50.241, 50.4382, 50.6001, 50.7128, 50.7902, 50.9146, 50.8301, 50.9075, 51.039, 51.1704, 51.2126, 51.2619, 51.3511, 51.3652, 51.5506, 51.6398, 51.5647, 51.5928, 51.5858, 51.5083] };
  WCL[616] = struct {
   x = [-70.0835, -69.9169, -69.7738, -69.6869, -69.586, -69.5532, -69.4663, -69.4311, -69.4663, -69.4217, -69.4874, -69.5766, -69.6095, -69.6752, -69.6189, -69.4663, -69.3208, -69.2222, -69.2668, -69.4874,
        -69.4663, -69.3889, -69.2786, -69.2105, -69.2222, -69.2457, -69.2551, -69.3443, -69.3654, -69.4663, -69.6869, -69.7292, -69.8395],
   y = [59.9264, 59.9382, 59.9147, 59.8983, 59.8607, 59.8326, 59.7997, 59.7105, 59.6824, 59.6331, 59.5721, 59.4711, 59.4078, 59.3468, 59.3139, 59.3421, 59.3515, 59.3139, 59.267, 59.1825, 59.1332, 59.1215,
        59.1332, 59.0698, 59.0182, 59.0135, 58.99, 58.9971, 59.0252, 58.9572, 58.8539, 58.9502, 59.0088] };
  WCL[617] = struct {
   x = [-70.0061, -69.9075, -69.9169],
   y = [58.9689, 58.8774, 58.814] };
  WCL[618] = struct {
   x = [-70.0061, -69.8958, -69.7855, -69.6189, -69.4663, -69.3208, -69.0791, -68.882, -68.6262, -68.4619, -68.4079, -68.3398, -68.2647, -68.2295, -68.2741, -68.3187, -68.5839, -68.9923, -69.2786, -69.0251,
        -68.7036, -68.3187, -68.1967, -68.077, -67.9338, -67.9432, -67.9761, -67.8118, -67.7672, -67.8001, -67.7789, -67.7344, -67.6123, -67.4363, -67.2931, -67.0608, -66.9177, -66.6759, -66.6313,
        -66.521, -66.4225, -66.2558, -66.047, -65.9813, -66.0704, -65.993, -65.9038, -65.8381, -65.9601, -66.0798, -66.1455, -65.9601, -65.871, -65.7724, -65.8498, -65.817, -65.6621, -65.6058,
        -65.7067, -65.5964, -65.5612, -65.5072, -65.4744, -65.3758, -65.3758, -65.2092, -65.3523, -65.4861, -65.5729, -65.4744, -65.3312],
   y = [58.7342, 58.7225, 58.6544, 58.7272, 58.814, 58.8586, 58.8586, 58.9009, 58.9103, 58.8704, 58.746, 58.6427, 58.5911, 58.4456, 58.3024, 58.1616, 58.0396, 57.9175, 57.8354, 57.8894, 57.9574, 58.1053,
        58.2438, 58.5043, 58.6146, 58.4174, 58.3541, 58.4573, 58.3658, 58.2907, 58.1452, 58.1053, 58.2672, 58.3306, 58.3611, 58.4456, 58.4761, 58.5043, 58.631, 58.7624, 58.8492, 58.7624, 58.7108,
        58.6146, 58.4409, 58.4996, 58.6075, 58.699, 58.7906, 58.8539, 58.922, 58.8492, 58.8704, 58.929, 58.9572, 59.0417, 59.037, 59.1661, 59.2975, 59.2294, 59.267, 59.3139, 59.3303, 59.3585, 59.4313,
        59.4805, 59.5932, 59.6659, 59.7504, 59.8607, 59.9875] };
  WCL[619] = struct {
   x = [-64.3713, -64.3713, -64.4699, -64.3174, -64.2164, -64.2164, -64.2282, -64.0193, -64.1061, -64.0968, -63.909, -63.733, -63.8198, -63.963, -63.7659, -63.7095, -63.6438, -63.5453, -63.5453, -63.4467,
        -63.3458, -63.3692, -63.5781, -63.787, -63.9536, -64.1836, -64.2282, -64.0522, -63.9536, -63.8198, -63.6767, -63.5335, -63.4021, -63.2472, -63.1815, -63.2261, -63.2472, -63.1909, -63.1909,
        -63.2355, -63.1486, -63.0266, -62.9374, -62.982, -62.9492, -62.8177, -62.8506, -63.0149, -63.2589, -63.4021, -63.4467, -63.4678, -63.5007, -63.5124, -63.4232, -63.2355, -63.0923, -62.9046,
        -62.7168, -62.5854, -62.5408, -62.6628, -62.7286, -62.7074, -62.63, -62.7403, -62.982, -63.1815, -63.0266, -62.8717, -62.6746, -62.5314, -62.4751, -62.4094, -62.3437, -62.4422, -62.5408,
        -62.5525, -62.4305, -62.3531, -62.2099, -62.2099, -62.0456, -61.8907, -61.8461, -62.0222, -62.1771, -62.3859, -62.3977, -62.2005, -61.9025, -61.7687, -61.8461, -61.7593, -61.5716, -61.4378,
        -61.3721, -61.5716, -61.7593, -61.6936, -61.825, -62.1113, -62.3437, -62.3977, -62.1325, -62.1325, -62.3202, -62.6957, -62.8717, -62.5737, -62.2756, -61.9799, -61.7476, -61.6936, -61.8579,
        -61.9119, -61.9799, -61.9564, -61.825, -61.703, -61.6701, -61.8461, -61.9119, -61.703, -61.4824, -61.3181, -61.2946, -61.2524, -61.175, -61.2735, -61.1632, -60.9098, -60.722, -60.6, -60.5789,
        -60.4686, -60.2925, -60.2925, -60.3465, -60.2597, -60.2151, -60.424, -60.4122, -60.2925, -60.194, -60.1048, -60.0062],
   y = [60.0367, 59.9311, 59.8725, 59.9147, 59.8725, 59.7434, 59.6659, 59.6542, 59.626, 59.5533, 59.5768, 59.5322, 59.4641, 59.3914, 59.3749, 59.2975, 59.3749, 59.3585, 59.2459, 59.2787, 59.2459, 59.1496,
        59.0863, 59.0933, 59.0933, 59.0299, 58.9783, 59.0464, 59.0135, 59.037, 59.0417, 59.037, 59.0651, 59.0816, 59.0698, 59.0135, 58.9689, 58.9572, 58.8774, 58.8023, 58.8211, 58.814, 58.8093,
        58.7108, 58.6779, 58.6544, 58.6028, 58.5676, 58.5441, 58.584, 58.5747, 58.5629, 58.5559, 58.5043, 58.4761, 58.4996, 58.4996, 58.4878, 58.5043, 58.4996, 58.3893, 58.3423, 58.2907, 58.279,
        58.2672, 58.2086, 58.1804, 58.0396, 58.0466, 58.1217, 58.1288, 58.1687, 58.1217, 58.0818, 58.0278, 58.0161, 57.9527, 57.8894, 57.9175, 57.9481, 57.9058, 57.8776, 57.7697, 57.7181, 57.6289,
        57.5467, 57.4881, 57.4881, 57.4224, 57.4341, 57.4341, 57.3379, 57.2487, 57.2252, 57.1947, 57.1666, 57.0328, 56.9436, 56.8286, 56.6972, 56.6831, 56.7629, 56.8357, 56.7746, 56.7324, 56.7019,
        56.6831, 56.6714, 56.6409, 56.6057, 56.6057, 56.6174, 56.5939, 56.5517, 56.5329, 56.4954, 56.4719, 56.4414, 56.3921, 56.3428, 56.2818, 56.3006, 56.202, 56.202, 56.2208, 56.2278, 56.1222,
        56.0049, 55.9626, 55.9063, 55.9016, 55.8758, 55.8758, 55.8148, 55.7139, 55.7585, 55.7631, 55.6528, 55.5824, 55.5332, 55.4393, 55.3219, 55.2774, 55.3149, 55.3525, 55.3219, 55.2844] };
  WCL[620] = struct {
   x = [-59.8396, -60.0391, -60.1259, -60.3465, -60.5577, -60.7877, -61.0529, -61.0318, -60.7103, -60.4474, -60.248, -60.2691, -60.2691, -60.3043, -60.1048],
   y = [53.8951, 53.7261, 53.6416, 53.6815, 53.773, 53.8106, 53.8059, 53.7144, 53.6885, 53.6228, 53.543, 53.4773, 53.3999, 53.2732, 53.3342] };
  WCL[621] = struct {
   x = [-59.9828, -60.1048, -60.1822, -60.1048, -60.2151, -60.2925, -60.3254, -60.5577, -60.8534, -61.074, -61.3275, -61.5481, -61.6584, -61.6819, -61.8368, -62.1559, -62.454, -62.7286, -62.9703, -63.2589,
        -63.5007, -63.733, -63.8433, -63.8433, -63.8644, -63.9747, -64.1836, -64.4277, -64.7022, -64.9228, -65.0989, -65.3312, -65.6386, -65.8827, -66.0259, -66.1362, -66.2558, -66.345, -66.3779,
        -66.5539, -66.7323],
   y = [50.2622, 50.2622, 50.3185, 50.4523, 50.4241, 50.3396, 50.2903, 50.2762, 50.2481, 50.227, 50.1777, 50.1143, 50.1354, 50.2058, 50.2692, 50.2903, 50.2974, 50.2903, 50.3044, 50.2551, 50.2762, 50.3044,
        50.3255, 50.3185, 50.3185, 50.3185, 50.3396, 50.3467, 50.3115, 50.2692, 50.2974, 50.3044, 50.2762, 50.2974, 50.2551, 50.2551, 50.2551, 50.2199, 50.1354, 50.1566, 50.0721] };
  WCL[622] = struct {
   x = [-80.0388, -79.8957, -79.7736, -79.7079, -79.729, -79.7619, -79.7196, -79.5976, -79.5648, -79.5648, -79.729, -79.7408, -79.6633, -79.6305, -79.553, -79.4099, -79.2784, -79.2784, -79.0461, -79.0015,
        -78.9921, -78.8818, -78.7481, -78.7481, -78.6823, -78.6612, -78.8044, -78.9147, -79.0132, -79.0461, -78.9358, -78.8138, -78.8044, -78.6612, -78.5838, -78.5063, -78.4617, -78.396, -78.2646,
        -78.0534, -77.922, -77.9431, -78.0111, -78.0862, -78.22, -78.3632, -78.4946, -78.5392, -78.5838, -78.6049, -78.7152, -78.7809, -78.8255, -78.7715, -78.7481, -78.8255, -78.8466, -78.8701,
        -78.8912, -78.9029, -78.9475, -78.9569, -78.9569, -78.9687, -78.9475, -78.9687, -78.9921, -79.0461, -79.1893, -79.2667, -79.3653, -79.4099, -79.499, -79.6422],
   y = [51.1845, 51.1235, 51.0531, 51.0742, 51.2338, 51.3933, 51.4544, 51.4544, 51.5295, 51.5787, 51.4332, 51.4403, 51.4544, 51.5036, 51.6046, 51.6327, 51.6257, 51.5717, 51.4755, 51.337, 51.2478, 51.269,
        51.33, 51.398, 51.4473, 51.5224, 51.5647, 51.6538, 51.7501, 51.8322, 51.8369, 51.851, 51.905, 51.9683, 52.0833, 52.1561, 52.1842, 52.2124, 52.2053, 52.1913, 52.2241, 52.2711, 52.2664, 52.2593,
        52.2382, 52.2523, 52.3133, 52.4213, 52.4752, 52.4823, 52.5081, 52.6348, 52.7099, 52.7498, 52.8085, 52.8413, 52.9892, 53.0878, 53.2145, 53.3271, 53.4327, 53.5102, 53.6228, 53.7144, 53.7988,
        53.8575, 53.942, 54.0523, 54.1556, 54.2706, 54.3738, 54.4395, 54.5029, 54.5287] };
  WCL[623] = struct {
   x = [-79.6422, -79.729, -79.7196, -79.5976, -79.3887, -79.1564, -78.9029, -78.6378, -78.4078, -78.1426, -77.9102, -77.645, -77.3916, -77.2813, -77.4033, -77.645, -77.5254, -77.3587, -77.1827, -76.9621,
        -76.8518, -76.765, -76.6101, -76.4552, -76.4012, -76.2674, -76.2135, -76.312, -76.3684, -76.3777, -76.4106, -76.4787, -76.5444, -76.5866, -76.6312, -76.7087, -76.7532, -76.7532, -76.7532,
        -76.765, -76.7861, -76.8072, -76.8401, -76.8753, -76.9504, -76.9856, -76.995, -77.0067, -77.0607, -77.1499, -77.2038, -77.2602, -77.3047, -77.3493, -77.3822, -77.4596, -77.5136, -77.5136,
        -77.5347, -77.7014, -77.9008, -78.0534, -78.1426, -78.2294, -78.3843, -78.5181, -78.6166, -78.7809, -78.8255, -78.8044, -78.7598, -78.8044, -78.8255, -78.7926, -78.6823, -78.5392, -78.4406,
        -78.22, -78.098, -78.0323, -77.922, -77.746, -77.7342, -77.746, -77.8445, -77.8563, -77.8328, -77.7553, -77.6802, -77.5136, -77.4362, -77.239, -77.0724, -77.171, -77.3587, -77.4479, -77.4362,
        -77.4033, -77.3259, -77.239, -77.2038, -77.3259, -77.4362, -77.5699, -77.746],
   y = [54.5287, 54.5545, 54.5545, 54.5991, 54.6437, 54.7212, 54.8291, 54.9042, 55.0192, 55.1319, 55.2069, 55.3525, 55.5097, 55.5965, 55.5097, 55.4088, 55.5144, 55.627, 55.8077, 56.0049, 56.148, 56.195,
        56.1832, 56.1175, 56.1293, 56.1598, 56.2443, 56.2583, 56.3616, 56.4297, 56.4954, 56.5634, 56.4836, 56.4038, 56.3428, 56.263, 56.3006, 56.3546, 56.4531, 56.5259, 56.6291, 56.7699, 56.8826,
        56.9741, 57.0328, 57.1008, 57.2018, 57.298, 57.3989, 57.5467, 57.6406, 57.7767, 57.8589, 57.9481, 58.0584, 58.1288, 58.1968, 58.2438, 58.279, 58.2719, 58.3306, 58.3775, 58.4362, 58.4644,
        58.5395, 58.5911, 58.6544, 58.7389, 58.8023, 58.8023, 58.8375, 58.9103, 59.0088, 59.0863, 59.105, 59.1496, 59.1895, 59.2013, 59.1895, 59.2505, 59.2505, 59.2623, 59.3139, 59.3867, 59.4195,
        59.5157, 59.5369, 59.626, 59.6167, 59.5885, 59.5486, 59.5721, 59.5932, 59.6167, 59.6378, 59.6941, 59.7997, 59.8537, 59.8537, 59.9382, 59.9875, 59.9476, 59.9476, 59.9663, 59.9828] };
  WCL[624] = struct {
   x = [-70.9096, -70.7101, -70.5787, -70.4473, -70.2924, -70.0835],
   y = [60.032, 59.9828, 59.9311, 59.9264, 59.9382, 59.9264] };
  WCL[625] = struct {
   x = [-69.8395, -70.0272, -70.0061],
   y = [59.0088, 58.9971, 58.9689] };
  WCL[626] = struct {
   x = [-69.9169, -70.0929, -70.2596, -70.4684, -70.6562, -70.8111, -70.6444, -70.3792, -70.1821, -70.0061],
   y = [58.814, 58.8023, 58.7859, 58.6662, 58.5559, 58.4456, 58.5395, 58.699, 58.7342, 58.7342] };
  WCL[627] = struct {
   x = [-90.0269, -89.7852, -89.5857, -89.3769, -89.2126, -89.0342, -88.9028, -88.7597, -88.6165, -88.417, -88.3067, -88.1636, -88.0321, -87.889, -87.7787, -87.713, -87.7224, -87.6238, -87.537, -87.4266,
        -87.3586, -87.2389, -87.0723, -86.9409, -86.7977, -86.7085, -86.5888, -86.4879, -86.4339, -86.2673, -86.0702, -85.9599, -85.8261, -85.7275, -85.683, -85.5398, -85.3966, -85.3755, -85.4858,
        -85.5398, -85.6618, -85.7275, -85.6172, -85.4741, -85.3966, -85.3075, -85.0211, -84.7348, -84.4931, -84.2842, -84.0965, -83.9627, -83.7867, -83.5661, -83.4112, -83.2915, -83.193, -82.9935,
        -82.9489, -82.8057, -82.618, -82.4983, -82.4302, -82.3763, -82.3528, -82.3199, -82.2871, -82.2988, -82.3317, -82.4091, -82.4302, -82.3317, -82.2777, -82.2777, -82.2425, -82.2214, -82.3199,
        -82.3434, -82.2542, -82.1674, -82.1322, -81.9796, -81.8248, -81.6699, -81.5924, -81.4493, -81.4493, -81.6933, -81.8693, -82.0571, -82.1557, -82.1439, -82.0336, -81.9233, -81.7473, -81.6933,
        -81.5502, -81.3718, -81.2404, -81.1184, -81.0081, -80.8203, -80.71, -80.6232, -80.5903, -80.5457, -80.6114, -80.7452, -80.8766, -81.0526, -81.1418, -81.0315, -80.8978, -80.7217, -80.5997,
        -80.5129, -80.4143, -80.2923, -80.182, -80.0388],
   y = [56.9319, 56.8544, 56.8051, 56.7934, 56.7981, 56.7324, 56.6714, 56.6244, 56.5681, 56.5024, 56.4907, 56.4461, 56.3428, 56.2325, 56.1339, 56.0682, 56.0002, 55.9509, 55.9814, 55.9556, 55.9133, 55.8946,
        55.9133, 55.8946, 55.8265, 55.789, 55.7702, 55.7467, 55.7397, 55.7021, 55.6716, 55.6528, 55.6083, 55.559, 55.4839, 55.4276, 55.383, 55.3149, 55.1647, 55.0309, 54.9042, 54.815, 54.923, 55.038,
        55.1694, 55.2633, 55.2774, 55.2703, 55.2703, 55.2703, 55.2515, 55.221, 55.2069, 55.2257, 55.2328, 55.221, 55.2586, 55.2398, 55.1577, 55.1764, 55.1389, 55.0826, 55.1131, 55.1248, 55.045,
        54.9558, 54.8667, 54.7, 54.5099, 54.3105, 54.1157, 54.0054, 53.8763, 53.6346, 53.4257, 53.2732, 53.1816, 52.9892, 52.8883, 52.8554, 52.7827, 52.67, 52.616, 52.5292, 52.4072, 52.2851, 52.2664,
        52.2452, 52.1842, 52.1162, 52.0551, 52.0411, 52.0622, 52.0904, 52.1021, 52.1091, 52.1443, 52.1302, 52.0833, 52.0082, 51.9331, 51.8322, 51.7548, 51.6538, 51.5295, 51.4403, 51.3511, 51.269,
        51.1845, 51.0742, 51.0108, 51.0531, 51.1235, 51.1915, 51.2478, 51.3159, 51.3229, 51.2619, 51.1986, 51.1845] };
  WCL[628] = struct {
   x = [-82.0219, -81.8576, -81.7145, -81.6699, -81.5384, -81.3953, -81.2967, -81.0972, -80.9752, -80.8978, -80.8649, -80.7781, -80.7992, -80.9541, -81.1418, -81.339, -81.5267, -81.6816, -81.8576, -81.9562,
        -82.0454, -82.1228, -82.0782, -82.0219],
   y = [53.1089, 53.1277, 53.1417, 53.1558, 53.1746, 53.1347, 53.1159, 53.1089, 53.0361, 52.9282, 52.8155, 52.6958, 52.6489, 52.717, 52.717, 52.8014, 52.8883, 52.9164, 52.9235, 52.9563, 52.9751, 52.9962,
        53.0408, 53.1089] };
  WCL[629] = struct {
   x = [-98.0038, -97.9804, -97.9264, -97.8818, -97.8818, -97.8372, -97.8372, -97.9592, -98.0461, -97.9358, -97.8161, -97.6377, -97.5603, -97.5955, -97.7058, -97.7809, -97.7809, -97.7715, -97.7386, -97.7386,
        -97.6612, -97.5955, -97.5838, -97.572, -97.4734, -97.3725, -97.2622, -97.2857, -97.3068, -97.274, -97.2411, -97.1848, -97.1308, -97.0651, -97.0416, -97.0322, -97.0322, -97.0416, -96.9642,
        -96.8539, -96.7764, -96.7671, -96.7671, -96.7436, -96.6661, -96.5793, -96.5136, -96.4455, -96.4455, -96.4127, -96.3141, -96.2249, -96.1381, -96.1381, -96.171, -96.2484, -96.2695, -96.2367,
        -96.2813, -96.3141, -96.3352, -96.5019, -96.5464, -96.5887, -96.7107, -96.8116, -96.8656, -96.9219, -96.9102, -96.8774, -96.821, -96.6896, -96.5793, -96.6004, -96.6333, -96.7013, -96.7999,
        -96.7999, -96.6896, -96.6779, -96.7553, -96.8539, -96.9759, -97.0322, -97.0534, -97.1848, -97.3303, -97.3186, -97.274, -97.5838, -97.5838, -97.5603, -97.5392, -97.572, -97.7058, -97.7152,
        -97.7269, -97.7269, -97.7269, -97.8161, -98.0578, -98.1024, -98.1024, -98.1235, -98.2573, -98.3676, -98.4122, -98.4216, -98.4216, -98.4896, -98.6211, -98.7314, -98.7431, -98.8088, -98.7431,
        -98.5319, -98.4662, -98.6328, -98.8088, -98.9402, -98.9754, -98.9848, -99.1608, -99.1843, -99.2054, -99.1843, -99.1608, -99.0857, -99.0294, -98.9965, -98.9965, -98.9308, -98.8862, -98.7525,
        -98.6774, -98.7102, -98.5765, -98.445, -98.2573, -98.1024, -97.9686, -97.9264, -97.9475, -98.0813, -98.1798, -98.2127, -98.1681, -98.1353, -98.0907, -98.1564, -98.201, -98.1916, -98.201,
        -98.3887, -98.5647, -98.5882, -98.4779, -98.3887, -98.3019, -98.1916, -98.1141, -98.0367, -98.0367, -98.0367, -98.0038, -97.9592],
   y = [54.3152, 54.3152, 54.348, 54.341, 54.2964, 54.2518, 54.2119, 54.1227, 53.9796, 53.9608, 54.0711, 54.0899, 53.9725, 53.9655, 53.9608, 53.9138, 53.8294, 53.7449, 53.6721, 53.6346, 53.5618, 53.4304,
        53.3248, 53.2051, 53.1535, 53.0925, 52.9939, 52.9751, 52.9141, 52.7874, 52.6747, 52.5926, 52.5339, 52.4588, 52.379, 52.311, 52.2499, 52.1748, 52.1091, 52.0481, 51.9801, 51.9378, 51.8838,
        51.7947, 51.7266, 51.6726, 51.5975, 51.5412, 51.5342, 51.5083, 51.4379, 51.3417, 51.2455, 51.1892, 51.1352, 51.0789, 51.0038, 50.87, 50.7386, 50.64, 50.5837, 50.64, 50.5133, 50.4077, 50.3725,
        50.3795, 50.5203, 50.6963, 50.809, 50.9263, 51.0507, 51.1493, 51.2596, 51.2948, 51.2455, 51.2173, 51.1962, 51.3769, 51.5342, 51.6233, 51.6914, 51.7477, 51.7196, 51.6304, 51.4802, 51.4661,
        51.6445, 51.7618, 51.8439, 52.0129, 52.0199, 52.0411, 52.088, 52.1349, 52.1279, 52.0598, 51.9918, 51.9589, 51.905, 51.8768, 51.9237, 52.034, 52.1889, 52.311, 52.325, 52.379, 52.3861, 52.3861,
        52.4447, 52.5057, 52.5597, 52.6278, 52.7615, 52.9071, 52.954, 52.9751, 53.0549, 53.0596, 53.0596, 53.0549, 53.0666, 53.1394, 53.1675, 53.2849, 53.3975, 53.4163, 53.4891, 53.529, 53.5806,
        53.6674, 53.7519, 53.7707, 53.8176, 53.8552, 53.8434, 53.7848, 53.7848, 53.7895, 53.7636, 53.7191, 53.6862, 53.7003, 53.7965, 53.8434, 53.9068, 54.0124, 54.0899, 54.1485, 54.2189, 54.179,
        54.2002, 54.2635, 54.3081, 54.2706, 54.2706, 54.3339, 54.3597, 54.3856, 54.4043, 54.4043, 54.4184, 54.3996, 54.355, 54.3081, 54.3152, 54.348] };
  WCL[630] = struct {
   x = [-94.7347, -94.7582, -94.791, -94.8121, -94.8239, -94.8239, -94.8239, -94.8028, -94.8121, -94.9013, -94.9882, -94.8239, -94.7582, -94.7136, -94.6244, -94.5047, -94.4812, -94.4366, -94.3827, -94.3498,
        -94.3498, -94.3381, -94.2724, -94.2606, -94.2606, -94.1949, -94.0306, -93.864, -93.6997, -93.5894, -93.4557, -93.3782, -93.2679, -93.2257, -93.2257, -93.2257, -93.2022, -93.1694, -93.1365,
        -93.1036, -93.0379, -92.9488, -92.8502, -92.7727, -92.7727, -92.7516, -92.6413, -92.5193, -92.4982, -92.4864, -92.5075, -92.5756, -92.6953, -92.8291, -92.8502, -92.7845, -92.707, -92.5756,
        -92.4653, -92.4207, -92.409, -92.409, -92.409, -92.3433, -92.233, -92.0452, -91.8692, -91.6486, -91.428, -91.1745, -90.9657, -90.7779, -90.6559, -90.447, -90.3461, -90.2147, -90.0269],
   y = [60.0931, 59.9875, 59.8818, 59.7715, 59.6941, 59.5814, 59.4477, 59.2904, 59.2013, 59.1215, 59.037, 59.0135, 58.922, 58.8891, 58.8093, 58.7577, 58.7507, 58.7694, 58.7906, 58.7272, 58.5629, 58.5043,
        58.5629, 58.7061, 58.7694, 58.7859, 58.7859, 58.7976, 58.7906, 58.7577, 58.7507, 58.7741, 58.746, 58.7272, 58.7225, 58.7178, 58.7061, 58.6662, 58.5629, 58.4174, 58.3259, 58.2391, 58.11,
        57.9926, 57.7955, 57.6524, 57.5585, 57.4693, 57.3144, 57.2135, 57.1243, 57.0398, 56.9741, 56.9389, 56.9201, 56.8967, 56.9201, 56.9671, 57.0093, 57.0164, 56.9624, 56.9389, 56.9319, 56.9319,
        56.9671, 57.0469, 57.0868, 57.1431, 57.1947, 57.2675, 57.2323, 57.2135, 57.1478, 57.0821, 56.9976, 56.9671, 56.9319] };
  WCL[631] = struct {
   x = [-110.137, -109.959, -109.804, -109.761, -109.738, -109.628, -109.407, -109.252, -109.252, -109.341, -109.461, -109.363, -109.285, -109.142, -108.987, -108.888, -108.701, -108.612, -108.513, -108.525,
        -108.668, -108.734, -108.877, -109.032, -109.163, -109.198, -109.163, -109.142, -109.01, -108.966, -108.811, -108.701, -108.612, -108.734, -108.823, -108.823, -108.689, -108.567, -108.403,
        -108.215, -108.083, -107.938, -107.828, -107.729, -107.663, -107.708, -107.663, -107.53, -107.53, -107.53, -107.452, -107.309, -107.178, -107.034, -106.945, -106.779, -106.647, -106.492,
        -106.37, -106.492, -106.68, -106.814, -106.924, -107.044, -107.112, -107.222, -107.342, -107.476, -107.586, -107.717, -107.783, -107.896, -108.083, -108.269, -108.447, -108.635, -108.778,
        -108.921, -109.064, -109.187, -109.351, -109.541, -109.661, -109.848, -109.982],
   y = [59.267, 59.3961, 59.4758, 59.5533, 59.6495, 59.6824, 59.727, 59.727, 59.7551, 59.8114, 59.8607, 59.8725, 59.9217, 59.91, 59.9147, 59.8889, 59.9053, 59.9053, 59.8889, 59.849, 59.8208, 59.7715, 59.788,
        59.7387, 59.6777, 59.6096, 59.5533, 59.5087, 59.5087, 59.5486, 59.5697, 59.5814, 59.5369, 59.4923, 59.3961, 59.3632, 59.3702, 59.4078, 59.4524, 59.4524, 59.4524, 59.4477, 59.4477, 59.4758,
        59.4524, 59.4078, 59.4148, 59.4359, 59.4148, 59.3632, 59.3468, 59.3421, 59.3421, 59.335, 59.335, 59.3468, 59.3303, 59.3233, 59.3022, 59.267, 59.2623, 59.267, 59.2904, 59.2904, 59.2904, 59.2857,
        59.2857, 59.267, 59.2623, 59.2951, 59.2341, 59.1895, 59.1543, 59.1262, 59.1379, 59.1332, 59.1215, 59.1215, 59.1215, 59.1426, 59.105, 59.0581, 59.0417, 58.99, 58.9502] };
  WCL[632] = struct {
   x = [-111.195, -111.219, -111.207, -111.163, -111.151, -111.106, -111.031, -110.843, -110.688, -110.599, -110.477, -110.313, -110.137],
   y = [58.7178, 58.7225, 58.7272, 58.7577, 58.814, 58.8422, 58.8258, 58.8774, 58.9736, 59.0863, 59.1661, 59.1778, 59.267] };
  WCL[633] = struct {
   x = [-109.982, -110.092, -110.236, -110.236, -110.247, -110.203, -110.17, -110.268, -110.4, -110.477, -110.611, -110.733, -110.876, -110.954, -111.019, -111.097, -111.174, -111.184, -111.195],
   y = [58.9502, 58.9384, 58.929, 58.8539, 58.7976, 58.7671, 58.7178, 58.6826, 58.6591, 58.6756, 58.6544, 58.6591, 58.6474, 58.6427, 58.6638, 58.7061, 58.7108, 58.7178, 58.7178] };
  WCL[634] = struct {
   x = [-130.052, -129.963, -129.897, -129.843, -129.832, -129.897, -129.975],
   y = [53.5102, 53.4492, 53.3506, 53.2732, 53.2004, 53.2004, 53.2732] };
  WCL[635] = struct {
   x = [-129.82, -129.787, -129.775, -129.764, -129.742, -129.698, -129.632, -129.566, -129.522, -129.522, -129.522, -129.522, -129.588, -129.656, -129.71, -129.754, -129.799, -129.832, -129.864, -129.918,
        -129.975],
   y = [53.6604, 53.6674, 53.6674, 53.6463, 53.6205, 53.5876, 53.5548, 53.4961, 53.4562, 53.4046, 53.3389, 53.2849, 53.2192, 53.2919, 53.3647, 53.4116, 53.4703, 53.5219, 53.5548, 53.6017, 53.6346] };
  WCL[636] = struct {
   x = [-130.019, -129.987, -129.954, -129.918, -129.876, -129.843, -129.82],
   y = [53.8294, 53.8176, 53.7895, 53.7519, 53.705, 53.6862, 53.6604] };
  WCL[637] = struct {
   x = [-129.236, -129.224, -129.203, -129.179, -129.146, -129.102, -129.06, -129.015, -129.003, -129.003, -129.069, -129.146, -129.179, -129.257, -129.257, -129.247, -129.247, -129.236],
   y = [52.8672, 52.8672, 52.8672, 52.8531, 52.7944, 52.7475, 52.7146, 52.6536, 52.5926, 52.5597, 52.5808, 52.6395, 52.6935, 52.7733, 52.8085, 52.8413, 52.8601, 52.8672] };
  WCL[638] = struct {
   x = [-127.513, -127.48, -127.48, -127.468, -127.513, -127.513, -127.513, -127.734, -127.888, -127.987, -128.064, -128.097, -128.24, -128.341, -128.341, -128.297, -128.198, -128.02, -127.823, -127.656,
        -127.546, -127.489, -127.325, -127.105, -126.905, -126.717, -126.53, -126.464, -126.354, -126.166, -126.044, -125.912, -125.769, -125.626, -125.537, -125.471, -125.459, -125.448, -125.305],
   y = [50.4218, 50.4288, 50.4147, 50.4288, 50.5485, 50.5907, 50.647, 50.647, 50.6611, 50.7315, 50.6541, 50.5837, 50.6259, 50.6893, 50.7667, 50.8231, 50.8841, 50.8911, 50.8559, 50.8019, 50.7808, 50.7808,
        50.6963, 50.633, 50.5837, 50.5414, 50.5062, 50.5062, 50.4851, 50.457, 50.4288, 50.3936, 50.3725, 50.3654, 50.3161, 50.2457, 50.2105, 50.1401, 50.0416] };
  WCL[639] = struct {
   x = [-127.159, -127.126, -127.072, -127.126, -127.259, -127.304, -127.313, -127.403, -127.447, -127.525, -127.612, -127.712, -127.799, -127.844, -127.799, -127.799, -127.788, -127.9, -127.91, -127.745,
        -127.59, -127.534, -127.513],
   y = [49.9712, 50.0768, 50.1401, 50.1401, 50.1965, 50.1049, 50.0768, 50.1542, 50.2035, 50.1894, 50.2176, 50.1331, 50.1472, 50.1894, 50.2457, 50.3021, 50.3373, 50.3795, 50.464, 50.4992, 50.4851, 50.4499,
        50.4218] };
  WCL[640] = struct {
   x = [-123.904, -123.946, -123.981],
   y = [49.9571, 50.0486, 50.0275] };
  WCL[641] = struct {
   x = [-124.654, -124.753, -124.786, -124.708, -124.643, -124.619, -124.631, -124.676, -124.654, -124.631, -124.697, -124.762, -124.807, -124.852, -124.917, -124.974, -125.007, -125.018, -125.039, -125.063,
        -125.129, -125.194, -125.272, -125.37, -125.471, -125.537, -125.603, -125.668, -125.701, -125.769, -125.847, -125.924, -126.044, -126.154, -126.255, -126.386, -126.452, -126.332, -126.222,
        -126.121, -125.978, -125.934, -126.044, -126.145, -126.121, -126.11, -126.187, -126.276, -126.398, -126.386, -126.264, -126.178, -126.264, -126.419, -126.562, -126.619, -126.685, -126.75,
        -126.851, -126.95, -127.072, -127.093, -127.048, -127.137, -127.259, -127.379, -127.468, -127.534, -127.546, -127.602, -127.689, -127.734, -127.734, -127.635, -127.513, -127.414, -127.447,
        -127.489, -127.534, -127.635, -127.734, -127.734, -127.712, -127.612, -127.546, -127.513, -127.403, -127.325, -127.259, -127.313, -127.457, -127.489, -127.435, -127.37, -127.435, -127.447,
        -127.489, -127.567, -127.612, -127.656, -127.755, -127.811, -127.844, -127.856, -127.856, -127.844, -127.811, -127.811, -127.856, -127.877, -127.877, -127.844, -127.823, -127.856, -127.91,
        -127.91, -127.942, -128.01, -128.142, -128.219, -128.264, -128.297, -128.273, -128.24, -128.24, -128.285, -128.395, -128.496, -128.571, -128.661, -128.759, -128.794, -128.804, -128.837,
        -128.893, -128.947, -128.982, -129.024, -129.069, -129.081, -129.113, -129.158, -129.146, -129.102, -129.048, -129.036, -129.036, -129.036, -129.036, -129.048, -129.048, -129.069, -129.113,
        -129.135, -129.167, -129.224, -129.247, -129.257, -129.301, -129.322, -129.346, -129.358, -129.433, -129.489, -129.555, -129.632, -129.688, -129.721, -129.764, -129.808, -129.864, -129.942],
   y = [49.9641, 50.0416, 50.0908, 50.1331, 50.1683, 50.2105, 50.288, 50.3443, 50.3865, 50.4358, 50.4499, 50.4218, 50.3936, 50.3865, 50.3865, 50.4358, 50.5133, 50.5555, 50.5555, 50.5555, 50.5344, 50.4992,
        50.4851, 50.5062, 50.5274, 50.4922, 50.5203, 50.5696, 50.5133, 50.4851, 50.5203, 50.5274, 50.5344, 50.5414, 50.5485, 50.5485, 50.5907, 50.6118, 50.6189, 50.6541, 50.6682, 50.7104, 50.7175,
        50.7175, 50.7667, 50.8231, 50.8418, 50.87, 50.8841, 50.8982, 50.8982, 50.9263, 50.9967, 51.0038, 51.0249, 50.9826, 50.9404, 50.9826, 50.9897, 50.9897, 51.0038, 50.9474, 50.9122, 50.8982,
        50.9334, 50.9686, 51.0108, 51.0249, 51.0319, 51.0648, 51.1352, 51.2103, 51.2737, 51.3065, 51.3018, 51.3276, 51.3769, 51.384, 51.398, 51.391, 51.3769, 51.4051, 51.4309, 51.452, 51.4661, 51.4872,
        51.4731, 51.4731, 51.5013, 51.5201, 51.5482, 51.5905, 51.6374, 51.6797, 51.7196, 51.7548, 51.7735, 51.7196, 51.6797, 51.6304, 51.6093, 51.6304, 51.7125, 51.7688, 51.8439, 51.912, 51.9589,
        51.9918, 52.0059, 52.0411, 52.1021, 52.1631, 52.2241, 52.2969, 52.318, 52.3978, 52.4518, 52.426, 52.3861, 52.379, 52.372, 52.372, 52.426, 52.4518, 52.5057, 52.5597, 52.5597, 52.5456, 52.5269,
        52.5339, 52.5386, 52.5738, 52.6466, 52.7263, 52.7334, 52.7545, 52.7874, 52.8273, 52.8742, 52.9352, 52.9751, 53.015, 53.0596, 53.1323, 53.1934, 53.252, 53.2919, 53.3107, 53.3318, 53.3717,
        53.4445, 53.4703, 53.5031, 53.5102, 53.4891, 53.4703, 53.4562, 53.4304, 53.4445, 53.4773, 53.4773, 53.4773, 53.5219, 53.5736, 53.6064, 53.6533, 53.705, 53.7519, 53.7848, 53.8176, 53.8481,
        53.881] };
  WCL[642] = struct {
   x = [-130.008, -129.918, -129.942],
   y = [54.1415, 54.1673, 54.2307] };
  WCL[643] = struct {
   x = [-130.118, -130.163, -130.216, -130.273, -130.273, -130.252, -130.294, -130.339, -130.383, -130.416, -130.505, -130.583, -130.669, -130.693, -130.747, -130.824, -130.836, -130.857, -130.935, -130.967,
        -130.967, -130.967, -130.857, -130.78, -130.714, -130.669, -130.637, -130.559, -130.604, -130.637, -130.714, -130.77, -130.836, -130.881, -130.89, -130.89, -130.89, -130.869, -130.925,
        -130.935, -130.991, -131.024, -131.078, -131.134, -131.167, -131.146, -131.09, -131.134, -131.223, -131.277, -131.343, -131.432, -131.486, -131.554, -131.632, -131.707, -131.763, -131.819,
        -131.796, -131.829, -131.862, -131.862, -131.939, -131.995, -132.005, -132.082, -132.16, -132.237, -132.293, -132.261, -132.171, -132.106, -132.082, -132.061, -132.04, -132.017, -131.995,
        -131.995, -131.995, -131.906, -131.775, -131.608, -131.596, -131.674, -131.784, -131.852, -131.93, -131.995, -132.049, -132.117, -132.204, -132.27, -132.315, -132.326, -132.326, -132.326,
        -132.326, -132.347, -132.38, -132.502, -132.535, -132.535, -132.591, -132.702, -132.812, -132.845, -132.878, -132.89, -132.934, -132.976, -133.033, -133.054, -132.857, -132.779, -132.768,
        -132.845, -132.845, -132.922, -133.012, -133.087, -133.176, -133.188, -133.209, -133.286, -133.31, -133.385, -133.507, -133.53, -133.507, -133.364, -133.242, -133.176, -133.232, -133.396,
        -133.474, -133.486, -133.429, -133.453, -133.54, -133.54, -133.441, -133.453, -133.53, -133.65, -133.695, -133.673, -133.641, -133.551, -133.453, -133.352, -133.242, -133.188, -133.143,
        -133.098, -133.044, -133.087, -133.11, -133.188, -133.209, -133.274, -133.286, -133.375, -133.453, -133.53, -133.563, -133.507, -133.298, -133.242, -133.342, -133.497, -133.608, -133.673,
        -133.685, -133.685, -133.763, -133.817, -133.871, -133.861, -133.739, -133.673, -133.551, -133.584, -133.662, -133.751, -133.763, -133.718, -133.65, -133.763, -133.838, -133.882, -133.96,
        -134.07, -134.093, -134.103, -134.136, -134.136, -134.192, -134.246, -134.323, -134.401, -134.49, -134.556, -134.633, -134.776, -134.854, -134.91, -134.988, -135.065, -135.02, -135.065,
        -135.131, -135.164, -135.196, -135.253, -135.351, -135.473, -135.516, -135.551, -135.593, -135.56, -135.495, -135.506, -135.584, -135.527, -135.483, -135.483, -135.417, -135.34, -135.295,
        -135.286, -135.286, -135.175, -135.164, -135.253, -135.295, -135.45, -135.495, -135.539, -135.626, -135.638, -135.703, -135.858, -135.969, -135.959, -135.969, -135.903, -136.001, -136.112,
        -136.123, -136.079, -136.189, -136.246, -136.278, -136.377, -136.522, -136.633, -136.675, -136.743, -136.731, -136.776, -136.907, -137.006, -136.952, -136.997, -137.095, -137.107, -137.062,
        -137.041, -136.907, -136.743, -136.698, -136.72, -136.621, -136.522, -136.522, -136.478, -136.422, -136.344, -136.267, -136.201, -136.189, -136.213, -136.335, -136.433, -136.522, -136.565,
        -136.544, -136.555, -136.654, -136.743, -136.82, -136.952, -137.095, -137.229, -137.339, -137.449, -137.569, -137.703, -137.647, -137.614, -137.715, -137.813, -137.834, -137.891, -137.989,
        -138.132, -138.2, -138.287, -138.311, -138.266, -138.243, -138.222, -138.144, -138.001, -137.825, -137.858, -137.956, -138.088, -138.156, -138.046, -137.945, -138.001, -138.088, -138.2,
        -138.276, -138.311, -138.398, -138.442, -138.386, -138.398, -138.421, -138.508, -138.674, -138.728, -138.728, -138.785, -138.895, -138.994, -139.137, -139.315, -139.303, -139.348, -139.458,
        -139.548, -139.557, -139.634, -139.724, -139.834, -139.9, -139.822, -139.768, -139.691, -139.646, -139.613, -139.613, -139.667, -139.679, -139.667, -139.634, -139.548, -139.503, -139.447,
        -139.435, -139.458, -139.447, -139.435, -139.348, -139.303, -139.315, -139.348, -139.348, -139.282, -139.205, -139.137, -139.214, -139.325, -139.435],
   y = [55.0614, 55.0614, 55.0614, 55.0732, 55.0802, 55.0943, 55.0427, 55.0122, 54.9723, 54.9535, 54.8972, 54.8714, 54.8714, 54.8455, 54.8197, 54.9042, 54.8596, 54.7892, 54.8009, 54.8972, 54.9535, 55.0356,
        55.0685, 55.0427, 55.0802, 55.0685, 55.1248, 55.1365, 55.1999, 55.3384, 55.3571, 55.3196, 55.3196, 55.3384, 55.3759, 55.39, 55.4956, 55.5707, 55.627, 55.7373, 55.7749, 55.8242, 55.8758,
        55.9931, 55.9439, 56.0236, 56.0847, 56.1339, 56.1222, 56.0847, 56.0283, 55.9978, 55.9368, 55.9603, 55.9744, 55.9298, 55.911, 55.8993, 55.857, 55.7819, 55.7514, 55.7068, 55.6505, 55.6646,
        55.5895, 55.5402, 55.5824, 55.6505, 55.7373, 55.7819, 55.7937, 55.8429, 55.8429, 55.8993, 55.9063, 55.9673, 56.0119, 56.0964, 56.1527, 56.195, 56.2137, 56.2137, 56.2443, 56.256, 56.2748,
        56.2865, 56.3053, 56.3358, 56.378, 56.3921, 56.4226, 56.4695, 56.5259, 56.5916, 56.6174, 56.6831, 56.7019, 56.7066, 56.7136, 56.7207, 56.7559, 56.8051, 56.8403, 56.8474, 56.8591, 56.8779,
        56.8943, 56.9131, 56.9483, 56.9741, 57.0281, 57.0703, 56.9859, 56.9483, 56.9178, 57.0398, 57.1243, 57.1243, 57.0821, 57.0821, 57.1173, 57.1595, 57.1713, 57.1783, 57.1478, 57.129, 57.2018,
        57.237, 57.2909, 57.3027, 57.3027, 57.3567, 57.3613, 57.3496, 57.3801, 57.4036, 57.4388, 57.4623, 57.4693, 57.5397, 57.5749, 57.6054, 57.6054, 57.5749, 57.6054, 57.6594, 57.7063, 57.6946,
        57.6594, 57.6171, 57.5984, 57.5702, 57.5233, 57.5045, 57.5045, 57.5162, 57.5397, 57.5866, 57.6758, 57.7063, 57.6641, 57.6993, 57.7533, 57.8049, 57.8636, 57.9105, 57.887, 57.9058, 57.9293,
        57.934, 57.934, 57.934, 57.9222, 57.8636, 57.8166, 57.8589, 57.934, 57.9879, 57.9926, 58.0278, 58.0748, 58.11, 58.0748, 58.0513, 58.0865, 58.1381, 58.1968, 58.1921, 58.1616, 58.0865, 58.0278,
        58.0584, 58.0865, 58.1335, 58.1687, 58.232, 58.2672, 58.2672, 58.232, 58.2837, 58.3259, 58.347, 58.3893, 58.401, 58.4174, 58.4691, 58.5723, 58.6239, 58.6873, 58.8093, 58.7976, 58.7788, 58.8657,
        58.9572, 59.1426, 59.2904, 59.4078, 59.3702, 59.3303, 59.2951, 59.1942, 59.1543, 59.1942, 59.1215, 59.037, 58.9854, 58.9103, 58.7976, 58.6709, 58.6028, 58.6192, 58.4526, 58.3822, 58.2719,
        58.232, 58.2484, 58.3376, 58.4526, 58.5043, 58.4057, 58.4127, 58.4127, 58.4174, 58.4761, 58.584, 58.5911, 58.6357, 58.7507, 58.8539, 58.8774, 58.9384, 58.9337, 58.8093, 58.7859, 58.8539,
        58.9337, 58.9854, 58.9971, 58.9455, 58.922, 58.9689, 59.0252, 58.9689, 58.929, 58.922, 58.9056, 58.9056, 58.9056, 58.8985, 58.8422, 58.7906, 58.7788, 58.7389, 58.7061, 58.6544, 58.631, 58.6544,
        58.6544, 58.5911, 58.5207, 58.4644, 58.3893, 58.347, 58.3376, 58.3306, 58.394, 58.3822, 58.2954, 58.2719, 58.2602, 58.2954, 58.3658, 58.4245, 58.4338, 58.4573, 58.5277, 58.5723, 58.5958,
        58.6591, 58.7108, 58.7178, 58.6943, 58.7108, 58.7342, 58.7976, 58.8985, 58.9572, 59.0182, 59.105, 59.1731, 59.2576, 59.3421, 59.3867, 59.4359, 59.5087, 59.5486, 59.5768, 59.6213, 59.6425,
        59.5768, 59.5369, 59.4923, 59.4923, 59.5157, 59.4477, 59.3961, 59.4195, 59.4078, 59.3303, 59.2341, 59.2107, 59.2177, 59.2224, 59.1825, 59.1426, 59.1825, 59.2341, 59.2787, 59.3186, 59.3468,
        59.3749, 59.3914, 59.3867, 59.4148, 59.443, 59.4477, 59.4758, 59.4876, 59.5251, 59.5603, 59.5932, 59.6049, 59.6378, 59.6871, 59.7387, 59.7715, 59.8279, 59.8936, 59.9429, 59.9875, 59.971, 59.91,
        59.8443, 59.7927, 59.7316, 59.6213, 59.5697, 59.6049, 59.6777, 59.7316, 59.7997, 59.8279, 59.8607, 59.8654, 59.8772, 59.9053, 59.9546] };
  WCL[644] = struct {
   x = [-139.712, -139.768, -139.855, -139.956],
   y = [60.0579, 59.9828, 59.9382, 59.8725] };
  WCL[645] = struct {
   x = [-134.943, -134.92, -134.91, -134.887, -134.866, -134.866, -134.91, -134.887, -134.833, -134.776, -134.765, -134.809, -134.809, -134.788, -134.788, -134.755, -134.711, -134.699, -134.711, -134.722,
        -134.722, -134.722, -134.722, -134.722, -134.722, -134.776, -134.866, -134.877, -134.877, -134.91, -134.997, -135.065, -135.053, -134.976, -134.92, -134.92, -134.955, -134.976, -135.03,
        -135.107, -135.164, -135.218, -135.307, -135.384, -135.34, -135.295, -135.351, -135.286, -135.274, -135.351, -135.396, -135.45, -135.506, -135.617, -135.661, -135.617, -135.626, -135.584,
        -135.516, -135.462, -135.429, -135.295, -135.229, -135.143, -135.119, -135.053, -134.988, -134.943],
   y = [57.2839, 57.3144, 57.3074, 57.2487, 57.183, 57.136, 56.9976, 57.0868, 57.0563, 56.9976, 56.9483, 56.9178, 56.8779, 56.8169, 56.7864, 56.7559, 56.6033, 56.5986, 56.5001, 56.4954, 56.4578, 56.4273,
        56.3851, 56.3546, 56.2818, 56.2255, 56.2255, 56.3006, 56.3616, 56.3663, 56.4649, 56.5869, 56.6479, 56.7136, 56.7911, 56.8521, 56.8051, 56.7441, 56.7207, 56.7207, 56.7512, 56.8286, 56.8474,
        56.8943, 56.8943, 56.9553, 56.9976, 57.0281, 57.0938, 57.129, 57.1713, 57.2557, 57.2792, 57.3144, 57.3567, 57.3989, 57.4458, 57.5115, 57.5467, 57.4928, 57.474, 57.4576, 57.4341, 57.3919,
        57.3449, 57.3027, 57.3027, 57.2839] };
  WCL[646] = struct {
   x = [-134.633, -134.666, -134.678, -134.732, -134.744, -134.8, -134.821, -134.821, -134.833, -134.833, -134.8, -134.744, -134.645, -134.535, -134.434, -134.368, -134.281, -134.237, -134.147, -134.07,
        -134.004, -133.96, -133.939, -133.906, -133.861, -133.849, -133.871, -133.915, -133.96, -134.004, -134.049, -134.093, -134.147, -134.192, -134.213, -134.258, -134.302, -134.314, -134.359,
        -134.359, -134.314, -134.258, -134.18, -134.136, -134.136, -134.037, -133.983, -133.948, -134.025, -134.082, -134.093, -134.037, -133.971, -133.971, -134.061, -134.126, -134.192, -134.192,
        -134.18, -134.204, -134.302, -134.347, -134.392, -134.434, -134.467, -134.568, -134.622, -134.666, -134.645, -134.645, -134.622, -134.633, -134.622, -134.612, -134.568, -134.457, -134.413,
        -134.467, -134.535, -134.568, -134.511, -134.445, -134.392, -134.467, -134.568, -134.612, -134.633],
   y = [57.6171, 57.6054, 57.6289, 57.7462, 57.758, 57.8284, 57.8988, 57.9692, 58.0161, 58.063, 58.1499, 58.1733, 58.1804, 58.1804, 58.1804, 58.1733, 58.1616, 58.0936, 58.0232, 57.9574, 57.8988, 57.8049,
        57.7298, 57.6993, 57.6218, 57.5866, 57.6054, 57.6524, 57.6993, 57.7533, 57.8119, 57.8471, 57.9105, 57.9809, 58.0161, 58.0748, 58.0701, 58.0161, 58.0161, 57.9457, 57.8518, 57.7814, 57.7533,
        57.7063, 57.6876, 57.6289, 57.5632, 57.5233, 57.5397, 57.5115, 57.4857, 57.4623, 57.4341, 57.3684, 57.3613, 57.3801, 57.3989, 57.3332, 57.2487, 57.2252, 57.2839, 57.2252, 57.1783, 57.1243,
        57.0868, 57.0821, 57.0938, 57.136, 57.1713, 57.237, 57.2675, 57.2909, 57.3214, 57.3684, 57.3848, 57.3801, 57.4153, 57.4271, 57.4576, 57.5045, 57.528, 57.5467, 57.5937, 57.5866, 57.5514,
        57.5819, 57.6171] };
  WCL[647] = struct {
   x = [-134.204, -134.237, -134.314, -134.359, -134.445, -134.467, -134.434, -134.38, -134.291, -134.246, -134.136, -134.049, -133.993, -133.871, -133.838, -133.882, -134.025, -133.971, -133.96, -133.948,
        -133.948, -133.96, -133.971, -133.993, -134.061, -134.136, -134.159, -134.246, -134.258, -134.281, -134.169, -134.082, -134.093, -134.136, -134.204, -134.246, -134.281, -134.281, -134.204],
   y = [56.6479, 56.6761, 56.6714, 56.7324, 56.7559, 56.8357, 56.8779, 56.8896, 56.8943, 56.9248, 56.8638, 56.7864, 56.7911, 56.8098, 56.6479, 56.6479, 56.6596, 56.5493, 56.4649, 56.3733, 56.3428, 56.2935,
        56.2137, 56.141, 56.2677, 56.3053, 56.2137, 56.2935, 56.3663, 56.4578, 56.4461, 56.5493, 56.5869, 56.5376, 56.5564, 56.5986, 56.6174, 56.6221, 56.6479] };
  WCL[648] = struct {
   x = [-133.795, -133.817, -133.838, -133.894, -133.96, -134.025, -134.016, -133.948, -133.795, -133.608, -133.474, -133.429, -133.319, -133.209, -133.155, -133.077, -133.054, -133.044, -133.054, -133.066,
        -133.054, -133.054, -133.087, -133.131, -133.199, -133.253, -133.352, -133.385, -133.396, -133.319, -133.265, -133.209, -133.199, -133.188, -133.188, -133.253, -133.375, -133.462, -133.596,
        -133.65, -133.695, -133.718, -133.718, -133.718, -133.718, -133.727, -133.763, -133.828, -133.882, -133.882, -133.849, -133.795],
   y = [56.9248, 56.9553, 56.9671, 56.9905, 57.0164, 57.0703, 57.0985, 57.1173, 57.0938, 57.0633, 57.0328, 57.0328, 57.0328, 57.0164, 57.0093, 56.9483, 56.9061, 56.8638, 56.8403, 56.7676, 56.7066, 56.6714,
        56.6643, 56.7066, 56.7559, 56.8098, 56.8591, 56.8521, 56.7746, 56.7324, 56.6761, 56.6362, 56.6033, 56.5611, 56.4954, 56.4883, 56.5141, 56.4836, 56.4766, 56.5798, 56.7019, 56.7981, 56.8521,
        56.8779, 56.8826, 56.8943, 56.8779, 56.8779, 56.9061, 56.9248, 56.9248, 56.9248] };
  WCL[649] = struct {
   x = [-132.857, -132.845, -132.845, -132.824, -132.779, -132.735, -132.645, -132.657, -132.714, -132.756, -132.768, -132.845, -132.911, -132.934, -132.934, -132.944, -132.944, -132.911, -132.857],
   y = [56.8051, 56.8098, 56.8051, 56.7793, 56.7324, 56.6831, 56.6104, 56.5869, 56.5752, 56.5916, 56.5446, 56.5446, 56.6033, 56.6643, 56.7324, 56.7911, 56.8286, 56.8286, 56.8051] };
  WCL[650] = struct {
   x = [-132.768, -132.702, -132.723, -132.723, -132.756, -132.8, -132.857, -132.944, -133.033, -133.087, -133.044, -132.934, -132.857, -132.789, -132.779, -132.768],
   y = [56.4343, 56.4273, 56.3546, 56.317, 56.3053, 56.2935, 56.317, 56.3358, 56.3428, 56.3968, 56.4461, 56.4649, 56.4695, 56.4578, 56.4531, 56.4343] };
  WCL[651] = struct {
   x = [-132.106, -132.082, -132.04, -131.995, -132.028, -132.094, -132.082, -132.106, -132.16, -132.228, -132.282, -132.326, -132.338, -132.347, -132.347, -132.315, -132.261, -132.216, -132.15, -132.106,
        -132.106],
   y = [56.3358, 56.3123, 56.2865, 56.256, 56.2513, 56.2255, 56.2137, 56.2067, 56.2443, 56.2677, 56.2935, 56.3428, 56.3851, 56.4343, 56.4461, 56.4226, 56.3921, 56.3733, 56.3663, 56.3428, 56.3358] };
  WCL[652] = struct {
   x = [-135.131, -135.086, -135.074, -135.042, -135.042, -135.042, -135.131, -135.253, -135.196, -135.119, -135.119, -135.218, -135.363, -135.506, -135.661, -135.682, -135.715, -135.682, -135.649, -135.572,
        -135.462, -135.363, -135.307, -135.241, -135.218, -135.143, -135.042, -135.02, -134.997, -135.03, -135.02, -135.175, -135.307, -135.396, -135.506, -135.649, -135.694, -135.804, -135.825,
        -135.825, -135.76, -135.715, -135.694, -135.694, -135.715, -135.825, -135.936, -136.037, -136.037, -135.959, -136.025, -136.123, -136.135, -136.189, -136.267, -136.356, -136.4, -136.389,
        -136.267, -136.257, -136.323, -136.41, -136.422, -136.4, -136.389, -136.311, -136.234, -136.213, -136.112, -136.058, -135.959, -135.882, -135.739, -135.661, -135.605, -135.506, -135.483,
        -135.372, -135.253, -135.196, -135.164, -135.131],
   y = [58.0818, 58.0865, 58.0818, 58.0513, 57.9879, 57.941, 57.9527, 57.9645, 57.9222, 57.8941, 57.8166, 57.8002, 57.8401, 57.8988, 57.9293, 57.9105, 57.8823, 57.8589, 57.8518, 57.8237, 57.7885, 57.7814,
        57.7697, 57.7767, 57.7767, 57.7767, 57.7462, 57.6336, 57.5514, 57.5397, 57.481, 57.5233, 57.5585, 57.5937, 57.6453, 57.7063, 57.7767, 57.7932, 57.7228, 57.6829, 57.6289, 57.5937, 57.5467,
        57.4975, 57.4271, 57.4153, 57.4341, 57.4857, 57.5162, 57.535, 57.5937, 57.6336, 57.6946, 57.7533, 57.7697, 57.8284, 57.9058, 57.9762, 57.9105, 57.9645, 58.0396, 58.11, 58.1616, 58.2156,
        58.2273, 58.1217, 58.1499, 58.2273, 58.232, 58.2484, 58.3189, 58.2837, 58.2391, 58.2086, 58.1381, 58.1217, 58.1452, 58.1335, 58.1335, 58.11, 58.0865, 58.0818] };
  WCL[653] = struct {
   x = [-132.812, -132.789, -132.746, -132.645, -132.547, -132.514, -132.502, -132.603, -132.58, -132.502, -132.458, -132.371, -132.27, -132.228, -132.27, -132.38, -132.502, -132.568, -132.559, -132.469,
        -132.437, -132.448, -132.38, -132.315, -132.183, -132.183, -132.228, -132.193, -132.106, -132.04, -132.04, -132.094, -132.15, -132.171, -132.106, -132.04, -132.028, -132.061, -132.04, -132.049,
        -132.106, -132.16, -132.193, -132.293, -132.326, -132.338, -132.347, -132.392, -132.425, -132.458, -132.535, -132.568, -132.568, -132.591, -132.678, -132.812, -132.89, -132.967, -133.012,
        -133.155, -133.242, -133.122, -133.012, -133.044, -133.066, -133.033, -132.976, -132.988, -133.044, -133.164, -133.298, -133.42, -133.364, -133.265, -133.253, -133.286, -133.286, -133.274,
        -133.31, -133.342, -133.364, -133.408, -133.507, -133.584, -133.662, -133.784, -133.739, -133.65, -133.617, -133.551, -133.662, -133.695, -133.65, -133.584, -133.474, -133.429, -133.364,
        -133.31, -133.242, -133.11, -133.11, -133.143, -133.131, -133.066, -132.976, -132.901, -132.857, -132.812],
   y = [56.0354, 56.0236, 55.9603, 55.8993, 55.8382, 55.8007, 55.7256, 55.7186, 55.6951, 55.6834, 55.6951, 55.627, 55.5402, 55.4839, 55.4956, 55.5402, 55.5942, 55.613, 55.5332, 55.4956, 55.4698, 55.4135,
        55.39, 55.4088, 55.3525, 55.2938, 55.282, 55.2257, 55.2515, 55.2187, 55.1365, 55.1178, 55.0497, 54.9863, 55.0122, 54.991, 54.9277, 54.9042, 54.8268, 54.7446, 54.7822, 54.7704, 54.7634, 54.7376,
        54.8338, 54.9042, 54.9418, 55.0239, 54.991, 54.9676, 54.9981, 55.0732, 55.1365, 55.1553, 55.214, 55.2633, 55.282, 55.275, 55.3126, 55.3384, 55.3689, 55.383, 55.4088, 55.4698, 55.5402, 55.5566,
        55.5895, 55.6317, 55.6646, 55.627, 55.62, 55.6763, 55.7631, 55.8312, 55.918, 56.0283, 56.0964, 56.1598, 56.1715, 56.0847, 56.0589, 56.0471, 56.0542, 55.9931, 55.9556, 55.9556, 56.0542, 56.1152,
        56.1527, 56.2208, 56.2818, 56.3358, 56.3663, 56.3663, 56.3663, 56.3053, 56.3311, 56.3546, 56.3428, 56.2748, 56.1903, 56.1457, 56.0847, 56.0847, 56.0847, 56.0659, 56.0589, 56.0354] };
  WCL[654] = struct {
   x = [-131.42, -131.42, -131.376, -131.322, -131.298, -131.244, -131.244, -131.223, -131.167, -131.101, -130.991, -130.946, -130.958, -131.024, -131.045, -131.045, -131.024, -131, -131.045, -131.09,
        -131.111, -131.146, -131.212, -131.289, -131.355, -131.409, -131.498, -131.542, -131.62, -131.674, -131.596, -131.608, -131.674, -131.653, -131.718, -131.697, -131.707, -131.796, -131.73,
        -131.608, -131.542, -131.554, -131.554, -131.521, -131.477, -131.399, -131.376, -131.388, -131.376, -131.334, -131.322, -131.334, -131.388, -131.432, -131.42],
   y = [55.3126, 55.3079, 55.3008, 55.3454, 55.3947, 55.3525, 55.2633, 55.2187, 55.1882, 55.1553, 55.1624, 55.1811, 55.2375, 55.3126, 55.4017, 55.4463, 55.5144, 55.62, 55.7256, 55.789, 55.85, 55.8922,
        55.9439, 55.9603, 55.9485, 55.9368, 55.911, 55.8688, 55.85, 55.8312, 55.789, 55.7514, 55.7373, 55.6881, 55.5942, 55.5707, 55.5332, 55.4581, 55.3759, 55.3454, 55.39, 55.4839, 55.5214, 55.4768,
        55.4322, 55.4651, 55.5707, 55.6388, 55.62, 55.5261, 55.4651, 55.4017, 55.3571, 55.3126, 55.3126] };
  WCL[655] = struct {
   x = [-133.054, -133.033, -133, -132.955, -132.911, -132.857, -132.8, -132.866, -132.988, -133.044, -133.131, -133.209, -133.253, -133.265, -133.274, -133.253, -133.176, -133.11, -133.066, -133.066,
        -133.054],
   y = [55.0356, 55.0427, 55.0239, 54.9605, 54.8972, 54.8338, 54.7704, 54.7258, 54.7963, 54.8901, 54.9723, 55.0239, 55.0873, 55.1365, 55.214, 55.2562, 55.2328, 55.1952, 55.1248, 55.0685, 55.0356] };
  WCL[656] = struct {
   x = [-131.632, -131.596, -131.542, -131.498, -131.444, -131.409, -131.399, -131.399, -131.486, -131.542, -131.596, -131.575, -131.587, -131.62, -131.632],
   y = [55.3126, 55.2938, 55.275, 55.2257, 55.2069, 55.1248, 55.0732, 55.0544, 55.0356, 55.1131, 55.1553, 55.1764, 55.2375, 55.2891, 55.3126] };
  WCL[657] = struct {
   x = [-132.061, -131.995, -131.885, -131.819, -131.763, -131.73, -131.808, -131.906, -131.951, -131.951, -131.951, -132.005, -132.073, -132.127, -132.139, -132.171, -132.249, -132.282, -132.315, -132.404,
        -132.437, -132.526, -132.547, -132.568, -132.535, -132.415, -132.448, -132.568, -132.636, -132.756, -132.866, -132.934, -132.967, -133.033, -133.131, -133.131, -133.122, -133.11, -133,
        -132.833, -132.678, -132.636, -132.678, -132.678, -132.613, -132.502, -132.371, -132.326, -132.282, -132.261, -132.261, -132.359, -132.448, -132.568, -132.657, -132.559, -132.458, -132.481,
        -132.491, -132.371, -132.315, -132.237, -132.183, -132.139, -132.094, -132.094, -132.094, -132.094, -132.094, -132.094, -132.061],
   y = [54.05, 54.064, 54.0969, 54.1532, 54.1344, 53.9983, 53.8223, 53.705, 53.5736, 53.4773, 53.3788, 53.299, 53.2333, 53.2779, 53.2849, 53.2849, 53.299, 53.306, 53.2849, 53.1605, 53.2333, 53.2732, 53.3506,
        53.3905, 53.3647, 53.4116, 53.4961, 53.5149, 53.5219, 53.5618, 53.6275, 53.7378, 53.8223, 53.9397, 53.9725, 54.057, 54.1415, 54.179, 54.179, 54.1743, 54.1743, 54.1532, 54.0429, 53.9796,
        54.0429, 54.1086, 54.1344, 54.0828, 54.0124, 53.8951, 53.8434, 53.8223, 53.7848, 53.7378, 53.6932, 53.6932, 53.7003, 53.6533, 53.6275, 53.6275, 53.6674, 53.7191, 53.7848, 53.8693, 53.9138,
        53.9537, 53.9796, 54.0124, 54.0382, 54.0429, 54.05] };
  WCL[658] = struct {
   x = [-132.15, -132.15, -132.16, -132.193, -132.171, -132.15, -132.261, -132.326, -132.305, -132.228, -132.326, -132.425, -132.359, -132.261, -132.193, -132.106, -132.017, -131.951, -131.894, -131.841,
        -131.763, -131.707, -131.775, -131.93, -131.939, -131.841, -131.707, -131.674, -131.751, -131.862, -131.918, -131.906, -131.972, -131.939, -131.885, -131.852, -131.829, -131.707, -131.608,
        -131.575, -131.575, -131.498, -131.432, -131.334, -131.266, -131.266, -131.31, -131.409, -131.542, -131.641, -131.751, -131.775, -131.763, -131.751, -131.751, -131.829, -131.951, -132.04,
        -132.117, -132.15],
   y = [52.7404, 52.7545, 52.7615, 52.8132, 52.8953, 52.9751, 52.8953, 52.9352, 52.9798, 53.0549, 53.0807, 53.0878, 53.1136, 53.1394, 53.1535, 53.1863, 53.2262, 53.2591, 53.2919, 53.2732, 53.1863, 53.1136,
        53.1206, 53.0995, 53.0737, 53.0596, 53.0338, 52.961, 52.9, 52.9282, 52.9, 52.7944, 52.7733, 52.7076, 52.6466, 52.5996, 52.5926, 52.5808, 52.5668, 52.5128, 52.4917, 52.487, 52.3861, 52.2969,
        52.2171, 52.1631, 52.1561, 52.2241, 52.3039, 52.3509, 52.4119, 52.4447, 52.4518, 52.4588, 52.5057, 52.5128, 52.5879, 52.6677, 52.7334, 52.7404] };
  WCL[659] = struct {
   x = [-130.66, -130.66, -130.637, -130.604, -130.571, -130.526, -130.461, -130.404, -130.327, -130.294, -130.339, -130.395, -130.472, -130.515, -130.571, -130.625, -130.66],
   y = [53.9796, 53.9913, 54.0171, 54.0429, 54.064, 54.0828, 54.1016, 54.0758, 54.0171, 53.9467, 53.9021, 53.8693, 53.9021, 53.9326, 53.9608, 53.9655, 53.9796] };
  WCL[660] = struct {
   x = [-130.538, -130.559, -130.55, -130.493, -130.395, -130.273, -130.163, -130.052],
   y = [53.6604, 53.6721, 53.6674, 53.6463, 53.6275, 53.6017, 53.5548, 53.5102] };
  WCL[661] = struct {
   x = [-129.975, -130.097, -130.24, -130.306, -130.428, -130.493, -130.526, -130.538, -130.538],
   y = [53.2732, 53.3647, 53.4163, 53.482, 53.543, 53.6064, 53.6463, 53.6533, 53.6604] };
  WCL[662] = struct {
   x = [-129.975, -130.029, -130.052, -130.052, -130.052, -130.097, -130.151, -130.195, -130.252, -130.252, -130.252, -130.207, -130.151, -130.106, -130.064, -130.019],
   y = [53.6346, 53.6721, 53.705, 53.7378, 53.7707, 53.7848, 53.7848, 53.8223, 53.8693, 53.881, 53.9068, 53.9138, 53.9021, 53.8763, 53.8481, 53.8294] };
  WCL[663] = struct {
   x = [-129.942, -130.008, -130.085, -130.118, -130.085, -130.008],
   y = [53.881, 53.9138, 53.9608, 54.0054, 54.0899, 54.1415] };
  WCL[664] = struct {
   x = [-129.942, -130.04, -130.118, -130.184, -130.273, -130.317, -130.371, -130.404, -130.461, -130.472, -130.461, -130.461, -130.404, -130.383, -130.339, -130.24, -130.118, -130.097, -130.163, -130.163,
        -130.141, -130.106, -130.029, -130.019, -130.04, -130.052, -130.064, -130.118],
   y = [54.2307, 54.226, 54.2189, 54.2189, 54.2377, 54.2964, 54.348, 54.3856, 54.4114, 54.4513, 54.4888, 54.5592, 54.6109, 54.6226, 54.693, 54.7258, 54.6625, 54.693, 54.7564, 54.808, 54.8526, 54.9159,
        54.9488, 54.991, 55.0122, 55.0427, 55.0544, 55.0614] };
  WCL[665] = struct {
   x = [-139.956, -140.054, -140.209, -140.385, -140.552, -140.716, -140.916, -141.103, -141.258, -141.455, -141.533, -141.512, -141.39, -141.402],
   y = [59.8725, 59.8114, 59.7668, 59.7316, 59.7434, 59.7551, 59.7762, 59.7997, 59.8373, 59.8818, 59.91, 59.9382, 59.9663, 59.9921] };
  WCL[666] = struct {
   x = [-141.643, -141.786, -141.932, -142.084],
   y = [60.0086, 59.971, 59.9757, 59.9875] };
  WCL[667] = struct {
   x = [-147.485, -147.506, -147.583, -147.607, -147.639, -147.75, -147.914, -147.991, -147.97, -147.926, -147.837],
   y = [60.0039, 59.9546, 59.9311, 59.8818, 59.8818, 59.8373, 59.7927, 59.8161, 59.8983, 59.9476, 59.9757] };
  WCL[668] = struct {
   x = [-148.742, -148.886, -149.052, -149.195],
   y = [60.0156, 59.9593, 59.9757, 59.9757] };
  WCL[669] = struct {
   x = [-149.395, -149.472],
   y = [60.0086, 59.9875] };
  WCL[670] = struct {
   x = [-149.547, -149.637, -149.648, -149.702, -149.77, -149.813, -149.881, -149.89, -149.857, -149.935],
   y = [60.0156, 59.9593, 59.849, 59.7152, 59.8279, 59.9593, 59.8983, 59.8654, 59.7762, 59.6988] };
  WCL[671] = struct {
   x = [-158.569, -158.623, -158.601, -158.569, -158.479, -158.447, -158.39, -158.447, -158.402, -158.447, -158.491, -158.524, -158.557, -158.569, -158.557, -158.557, -158.578, -158.569],
   y = [56.3663, 56.324, 56.3428, 56.3428, 56.3475, 56.3311, 56.3053, 56.2677, 56.2443, 56.1762, 56.0964, 56.0424, 56.0847, 56.1715, 56.256, 56.2818, 56.3123, 56.3663] };
  WCL[672] = struct {
   x = [-154.771, -154.76, -154.736, -154.694, -154.671, -154.626, -154.605, -154.605, -154.671, -154.727, -154.781, -154.793, -154.771],
   y = [56.5564, 56.5681, 56.5752, 56.5798, 56.5916, 56.6033, 56.5798, 56.5259, 56.4695, 56.4156, 56.4226, 56.4836, 56.5564] };
  WCL[673] = struct {
   x = [-154.241, -154.295, -154.34, -154.372, -154.363, -154.307, -154.241, -154.185, -154.185, -154.185, -154.241],
   y = [56.5446, 56.5446, 56.5446, 56.5564, 56.5869, 56.6174, 56.6291, 56.6174, 56.5916, 56.5681, 56.5446] };
  WCL[674] = struct {
   x = [-153.192, -153.204, -153.237, -153.246, -153.347, -153.389, -153.434, -153.347, -153.258, -153.204, -153.082, -152.96, -152.948, -153.049, -153.169, -153.192],
   y = [57.129, 57.1173, 57.0938, 57.0868, 57.0398, 57.0703, 57.1173, 57.1947, 57.2252, 57.2065, 57.2065, 57.2135, 57.19, 57.1525, 57.136, 57.129] };
  WCL[675] = struct {
   x = [-154.837, -154.903, -154.903, -154.849, -154.748, -154.736, -154.661, -154.516, -154.45, -154.372, -154.175, -154.074, -154.131, -154.175, -154.098, -154.053, -154.02, -153.952, -153.866, -153.842,
        -153.842, -153.809, -153.854, -153.767, -153.866, -153.976, -153.997, -153.931, -153.788, -153.687, -153.666, -153.612, -153.523, -153.434, -153.347, -153.258, -153.192, -153.169, -153.258,
        -153.159, -153.091, -152.981, -152.981, -152.915, -152.861, -152.76, -152.608, -152.572, -152.596, -152.617, -152.617, -152.584, -152.572, -152.486, -152.385, -152.286, -152.242, -152.331,
        -152.385, -152.54, -152.641, -152.751, -152.87, -152.993, -153.049, -152.96, -152.861, -152.739, -152.683, -152.683, -152.683, -152.772, -152.849, -152.861, -152.927, -153.016, -153.192,
        -153.159, -153.082, -153.169, -153.38, -153.457, -153.556, -153.678, -153.711, -153.678, -153.678, -153.767, -153.898, -154.009, -154.119, -154.042, -153.931, -153.943, -154.053, -154.098,
        -154.107, -154.086, -154.185, -154.34, -154.396, -154.363, -154.241, -154.229, -154.229, -154.351, -154.527, -154.572, -154.593, -154.649, -154.771, -154.837],
   y = [57.3731, 57.3684, 57.3989, 57.4036, 57.4036, 57.4928, 57.5514, 57.6289, 57.6218, 57.6594, 57.6711, 57.6641, 57.6218, 57.5819, 57.5702, 57.5233, 57.4458, 57.4271, 57.4271, 57.481, 57.535, 57.5702,
        57.6218, 57.6406, 57.7228, 57.8002, 57.8636, 57.8988, 57.9105, 57.8166, 57.711, 57.6993, 57.7932, 57.8354, 57.8166, 57.8166, 57.8589, 57.8988, 57.9574, 57.934, 57.9457, 57.934, 57.8049,
        57.7767, 57.8354, 57.9058, 57.9222, 57.8753, 57.7533, 57.7298, 57.6711, 57.6711, 57.6171, 57.6101, 57.6171, 57.6218, 57.6171, 57.5162, 57.474, 57.474, 57.5115, 57.535, 57.5702, 57.5585, 57.474,
        57.4693, 57.4623, 57.4388, 57.3989, 57.3496, 57.3144, 57.3214, 57.3332, 57.4036, 57.3989, 57.3496, 57.3261, 57.3261, 57.2792, 57.2839, 57.2417, 57.1947, 57.1102, 57.0985, 57.0633, 57.0211,
        56.9483, 56.8779, 56.8521, 56.8169, 56.7911, 56.8943, 56.9671, 57.0093, 57.0164, 57.0328, 57.0938, 57.1407, 57.1783, 57.1947, 57.1713, 57.1478, 57.136, 57.0398, 56.9483, 56.9061, 56.9859,
        57.0821, 57.183, 57.2792, 57.3074, 57.3731] };
  WCL[676] = struct {
   x = [-153.478, -153.502, -153.478, -153.445, -153.38, -153.291, -153.204, -153.159, -153.115, -153.237, -153.314, -153.38, -153.457, -153.478],
   y = [58.11, 58.11, 58.1217, 58.1452, 58.1569, 58.1452, 58.1264, 58.0865, 58.0396, 58.0748, 58.1053, 58.0936, 58.0865, 58.11] };
  WCL[677] = struct {
   x = [-152.793, -152.772, -152.772, -152.894, -152.981, -152.906, -152.882, -152.828, -152.793, -152.694, -152.608, -152.507, -152.385, -152.385, -152.486, -152.462, -152.331, -152.242, -152.232, -152.242,
        -152.164, -152.143, -152.307, -152.385, -152.551, -152.65, -152.718, -152.849, -152.861, -152.906, -152.971, -153.004, -153.049, -153.082, -153.169, -153.225, -153.147, -153.126, -153.082,
        -152.981, -152.906, -152.828, -152.793],
   y = [58.2907, 58.2954, 58.3142, 58.3189, 58.3541, 58.3658, 58.4057, 58.4456, 58.4526, 58.4761, 58.4644, 58.4644, 58.4691, 58.4245, 58.394, 58.3541, 58.3658, 58.4127, 58.3376, 58.2837, 58.3071, 58.2273,
        58.2273, 58.1569, 58.1452, 58.1687, 58.1147, 58.0983, 58.0748, 58.063, 58.0701, 58.0983, 58.1335, 58.1733, 58.1921, 58.2086, 58.2086, 58.2719, 58.3024, 58.2907, 58.279, 58.279, 58.2907] };
  WCL[678] = struct {
   x = [-152.683, -152.662, -152.641, -152.629, -152.608, -152.551, -152.495, -152.453, -152.462, -152.495, -152.584, -152.662, -152.683, -152.683],
   y = [58.6474, 58.6192, 58.6192, 58.6028, 58.6028, 58.6357, 58.6357, 58.6192, 58.5559, 58.5324, 58.5324, 58.5606, 58.6075, 58.6474] };
  WCL[679] = struct {
   x = [-160.003, -159.916, -159.892, -159.815, -159.662, -159.54, -159.43, -159.32, -159.186, -159.054, -158.977, -158.9, -158.789, -158.733, -158.688, -158.613, -158.569, -158.491, -158.414, -158.315,
        -158.28, -158.193, -158.127, -157.916, -157.85, -157.794, -157.752, -157.728, -157.696, -157.674, -157.642, -157.475, -157.398, -157.463, -157.552, -157.585, -157.541, -157.487, -157.299,
        -157.132, -157.144, -157.21, -157.088, -157.001, -156.9, -156.813, -156.846, -156.945, -157.078, -157.233, -157.308, -157.376, -157.463, -157.597, -157.684, -157.794, -157.895, -158.017,
        -158.149, -158.259, -158.325, -158.402, -158.524, -158.613, -158.601, -158.569, -158.536, -158.545, -158.613, -158.667, -158.745, -158.822, -158.9, -158.944, -158.953, -158.855, -158.843,
        -158.855, -158.998, -159.054, -159.066, -159.176, -159.242, -159.341, -159.439, -159.55, -159.662, -159.65, -159.728, -159.782, -159.827, -159.925],
   y = [56.5376, 56.5916, 56.5916, 56.5869, 56.6526, 56.7019, 56.7066, 56.7324, 56.8051, 56.8709, 56.8357, 56.8638, 56.8474, 56.9131, 57.0328, 57.129, 57.129, 57.1947, 57.237, 57.3074, 57.3731, 57.3848,
        57.3496, 57.5702, 57.6054, 57.6594, 57.7415, 57.8753, 57.9997, 58.0748, 58.1733, 58.2719, 58.3259, 58.3306, 58.3259, 58.3376, 58.4808, 58.5958, 58.6638, 58.7741, 58.7976, 58.8093, 58.8586,
        58.9455, 59.0182, 59.098, 59.1332, 59.1097, 59.0698, 58.9971, 58.9384, 58.9056, 58.8657, 58.8093, 58.7741, 58.746, 58.7272, 58.6943, 58.6638, 58.6638, 58.6826, 58.7342, 58.7859, 58.8375,
        58.8539, 58.9056, 58.9689, 59.0299, 59.0182, 58.9689, 58.929, 58.8938, 58.8023, 58.7671, 58.7671, 58.6709, 58.5606, 58.5207, 58.4338, 58.4338, 58.4245, 58.4691, 58.5324, 58.6427, 58.746,
        58.8258, 58.8586, 58.9173, 58.922, 58.9502, 58.9056, 58.8704] };
  WCL[680] = struct {
   x = [-149.935, -150.033, -150.212, -150.256, -150.266, -150.399, -150.486, -150.432, -150.432, -150.564, -150.674, -150.763, -150.84, -150.963, -151.028, -151.139, -151.282, -151.315, -151.303, -151.448,
        -151.601, -151.8, -151.878, -152.021, -152.021, -151.922, -151.934, -151.866, -151.767, -151.735, -151.69, -151.636, -151.526, -151.425, -151.502, -151.481, -151.371, -151.249, -151.16,
        -151.073, -151.16, -151.338, -151.547, -151.636, -151.723, -151.866, -151.934, -151.899, -151.8],
   y = [59.6988, 59.7316, 59.6824, 59.6659, 59.565, 59.4641, 59.3749, 59.4876, 59.5885, 59.5439, 59.5486, 59.4923, 59.443, 59.3421, 59.2787, 59.3233, 59.3515, 59.335, 59.2388, 59.2459, 59.2341, 59.1895,
        59.2388, 59.267, 59.3421, 59.3632, 59.4195, 59.4805, 59.4195, 59.4758, 59.4993, 59.504, 59.4758, 59.4641, 59.504, 59.5439, 59.5439, 59.5979, 59.6706, 59.7833, 59.7927, 59.727, 59.6612, 59.6425,
        59.6612, 59.7058, 59.7668, 59.8537, 59.9476] };
  WCL[681] = struct {
   x = [-152.772, -152.849, -152.915, -153.082, -153.213, -153.323, -153.159, -153.147, -153.258, -153.457, -153.457, -153.589, -153.633, -153.744, -153.699, -153.711, -153.821, -153.821, -153.943, -154.185,
        -154.185, -154.25, -154.295, -154.34, -154.25, -153.988, -153.821, -153.788, -153.687, -153.6, -153.523, -153.469, -153.469, -153.49, -153.544, -153.612, -153.687, -153.776, -153.931, -154.02,
        -154.053, -154.164, -154.185, -154.152, -154.152, -154.328, -154.262, -154.208, -154.218, -154.218, -154.417, -154.527, -154.617, -154.661, -154.781, -154.814, -154.891, -155.112, -155.147,
        -155.189, -155.234, -155.332, -155.422, -155.445, -155.478, -155.511, -155.565, -155.588, -155.666, -155.785, -155.809, -155.863, -155.886, -156.05, -156.116, -156.194, -156.25, -156.36,
        -156.503, -156.56, -156.637, -156.593, -156.482, -156.492, -156.492, -156.581, -156.703, -156.736, -156.846, -156.858, -156.9, -156.989, -157.022, -157.111, -157.189, -157.276, -157.353,
        -157.487, -157.597, -157.618, -157.609, -157.597, -157.752, -157.839, -157.907, -158.071, -158.149, -157.973, -157.928, -157.982, -158.071, -158.238, -158.315, -158.503, -158.623, -158.768,
        -158.953, -159.031, -159.099, -159.209, -159.153, -159.054, -158.9, -158.745, -158.756, -158.733, -158.789, -158.834, -158.843, -158.9, -159.022, -159.108, -159.198, -159.263, -159.32, -159.43,
        -159.507, -159.55, -159.561, -159.561, -159.606, -159.65, -159.695, -159.749, -159.782, -159.749, -159.749, -159.749, -159.805, -159.925, -159.993],
   y = [60.0203, 59.9663, 59.91, 59.9053, 59.8983, 59.8654, 59.8373, 59.7668, 59.7152, 59.6988, 59.7833, 59.7668, 59.6871, 59.7152, 59.6425, 59.5697, 59.5322, 59.4758, 59.4195, 59.3961, 59.3585, 59.2177,
        59.1262, 59.0651, 59.0417, 59.0863, 59.0581, 59.0065, 58.9854, 58.9689, 58.9455, 58.9056, 58.8375, 58.7671, 58.7272, 58.6943, 58.6709, 58.6544, 58.631, 58.5794, 58.5441, 58.5324, 58.4761,
        58.4292, 58.394, 58.347, 58.3142, 58.3024, 58.2391, 58.2086, 58.1499, 58.1733, 58.1217, 58.0865, 58.0748, 58.063, 58.063, 58.0161, 57.9645, 57.934, 57.9058, 57.8589, 57.8518, 57.8284, 57.7885,
        57.7415, 57.7533, 57.758, 57.7814, 57.7885, 57.6993, 57.6524, 57.6054, 57.5984, 57.6101, 57.4857, 57.5233, 57.474, 57.4153, 57.4036, 57.3449, 57.3144, 57.3214, 57.237, 57.19, 57.1055, 57.0563,
        57.0281, 57.0563, 57.0093, 56.9859, 57.0023, 56.9483, 56.9061, 56.8779, 56.8357, 56.8403, 56.8779, 56.8521, 56.7864, 56.7207, 56.6596, 56.6596, 56.7066, 56.6948, 56.6362, 56.5869, 56.6033,
        56.5916, 56.5141, 56.5141, 56.5141, 56.5071, 56.4695, 56.4109, 56.3311, 56.3428, 56.4109, 56.5188, 56.5259, 56.4273, 56.3428, 56.2748, 56.2513, 56.1832, 56.0847, 56.0354, 56.0354, 56.0354,
        56.0354, 55.9744, 55.9439, 55.9368, 55.9368, 55.9251, 55.911, 55.911, 55.9368, 55.9251, 55.8758, 55.789, 55.7256, 55.6646, 55.6012, 55.613, 55.6951, 55.7373, 55.8077, 55.8758, 55.8758, 55.8242] };
  WCL[682] = struct {
   x = [-169.331, -169.322, -169.322, -169.277, -169.188, -169.111, -169, -168.913, -168.803, -168.735, -168.592, -168.583, -168.547, -168.36, -168.207, -168.106, -168.085, -168.118, -168.261, -168.437,
        -168.559, -168.615, -168.726, -168.759, -168.869, -169.101, -169.179, -169.298, -169.331],
   y = [52.8014, 52.8343, 52.8742, 52.9071, 52.954, 53.0737, 53.1793, 53.2849, 53.3248, 53.3248, 53.3436, 53.4116, 53.482, 53.5219, 53.5618, 53.536, 53.4562, 53.3788, 53.3436, 53.2779, 53.2121, 53.1722,
        53.0995, 53.0666, 53.008, 52.8531, 52.8413, 52.7874, 52.8014] };
  WCL[683] = struct {
   x = [-167.841, -167.918, -167.909, -167.841, -167.764, -167.721, -167.587, -167.489, -167.346, -167.311, -167.257, -167.212, -167.09, -167.003, -167.057, -167.179, -167.278, -167.257, -167.125, -166.97,
        -166.86, -166.792, -166.682, -166.616, -166.55, -166.484, -166.494, -166.538, -166.682, -166.714, -166.616, -166.595, -166.693, -166.792, -166.926, -166.98, -167.135, -167.289, -167.433,
        -167.534, -167.677, -167.82, -167.841],
   y = [53.3177, 53.3717, 53.3835, 53.4046, 53.4163, 53.4163, 53.4304, 53.4562, 53.529, 53.6275, 53.6721, 53.7308, 53.6792, 53.7378, 53.7848, 53.7848, 53.881, 53.9537, 53.9913, 54.0054, 53.9655, 53.888,
        53.9608, 54.0241, 54.0171, 53.9655, 53.9397, 53.881, 53.8035, 53.7308, 53.7707, 53.6932, 53.6275, 53.6017, 53.5618, 53.4891, 53.4773, 53.4304, 53.3506, 53.3389, 53.3248, 53.3177, 53.3177] };
  WCL[684] = struct {
   x = [-166.13, -166.109, -166.076, -165.975, -165.931, -165.898, -166.02, -166.109, -166.208, -166.24, -166.229, -166.13],
   y = [54.2776, 54.2706, 54.2377, 54.2307, 54.179, 54.1344, 54.064, 54.0758, 54.0828, 54.1532, 54.2377, 54.2776] };
  WCL[685] = struct {
   x = [-165.048, -165.048, -164.905, -164.828, -164.696, -164.652, -164.595, -164.497, -164.33, -164.133, -163.987, -163.889, -163.811, -163.746, -163.668, -163.647, -163.624, -163.591, -163.57, -163.546,
        -163.513, -163.469, -163.427, -163.349, -163.305, -163.281, -163.316, -163.459, -163.581, -163.802, -164.056, -164.243, -164.44, -164.706, -164.828, -164.938, -164.983, -165.027, -165.048],
   y = [54.6484, 54.6554, 54.7071, 54.7564, 54.8455, 54.9159, 54.9863, 54.9793, 54.9723, 54.9981, 55.0427, 55.0943, 55.1131, 55.1131, 55.0943, 55.0427, 54.9793, 54.9277, 54.8784, 54.8643, 54.8455, 54.8338,
        54.808, 54.808, 54.7822, 54.7376, 54.7188, 54.7517, 54.686, 54.6554, 54.6625, 54.6625, 54.5334, 54.3926, 54.4513, 54.4959, 54.5475, 54.6109, 54.6484] };
  WCL[686] = struct {
   x = [-162.896, -162.941, -162.884, -162.795, -162.73, -162.697, -162.708, -162.83, -162.884, -162.896],
   y = [54.5006, 54.4959, 54.5146, 54.5217, 54.5006, 54.456, 54.4372, 54.4184, 54.4818, 54.5006] };
  WCL[687] = struct {
   x = [-160.732, -160.81, -160.798, -160.798, -160.786, -160.744, -160.721, -160.676, -160.643, -160.643, -160.643, -160.622, -160.61, -160.61, -160.631, -160.744, -160.819, -160.864, -160.887, -160.908,
        -160.864, -160.843, -160.786, -160.732],
   y = [55.3689, 55.4393, 55.4088, 55.3759, 55.3689, 55.3689, 55.4135, 55.4463, 55.4017, 55.3571, 55.3126, 55.275, 55.2703, 55.2515, 55.2069, 55.2375, 55.214, 55.214, 55.2891, 55.4135, 55.4322, 55.3689,
        55.3525, 55.3689] };
  WCL[688] = struct {
   x = [-160.854, -160.831, -160.854, -160.897, -160.986, -161.084, -161.162, -161.185, -161.152, -161.052, -160.941, -160.876, -160.854],
   y = [58.8469, 58.8586, 58.7788, 58.7108, 58.6192, 58.5958, 58.6122, 58.699, 58.7389, 58.7577, 58.8093, 58.8422, 58.8469] };
  WCL[689] = struct {
   x = [-165.689, -165.745, -165.898, -165.954, -166.064, -166.097, -166.186, -166.229, -166.416, -166.616, -166.75, -166.881, -166.991, -167.113, -167.212],
   y = [60.0039, 59.9757, 59.9875, 59.9429, 59.9311, 59.91, 59.8725, 59.8044, 59.849, 59.8725, 59.8983, 59.9429, 59.9828, 59.9757, 59.9992] };
  WCL[690] = struct {
   x = [-163.436, -163.349, -163.293, -163.206, -163.15, -163.084, -163.006, -162.917, -162.908, -162.83, -162.676, -162.664, -162.676, -162.652, -162.598, -162.497, -162.399, -162.309, -162.199, -162.047,
        -161.901, -161.77, -161.638, -161.537, -161.472, -161.439, -161.382, -161.296, -161.218, -161.152, -161.03, -160.965, -160.974, -161.03, -160.953, -160.854, -160.777, -160.765, -160.744,
        -160.631, -160.512, -160.401, -160.345, -160.357, -160.357, -160.401, -160.545, -160.643, -160.622, -160.599, -160.589, -160.545, -160.423, -160.279, -160.214, -160.003],
   y = [55.1882, 55.1952, 55.2187, 55.2445, 55.282, 55.2891, 55.2515, 55.2891, 55.3454, 55.3759, 55.4135, 55.4651, 55.5027, 55.5261, 55.5519, 55.6505, 55.7021, 55.7514, 55.8077, 55.8617, 55.911, 55.9603,
        55.9931, 56.0119, 56.0166, 56.0424, 56.0589, 56.0729, 56.0424, 56.0354, 56.0542, 56.0589, 56.0119, 55.9603, 55.8805, 55.857, 55.8875, 55.9368, 55.9298, 55.911, 55.8617, 55.85, 55.85, 55.85,
        55.8617, 55.9063, 55.9368, 55.9603, 55.9744, 56.0589, 56.0964, 56.1715, 56.2443, 56.3921, 56.4226, 56.5376] };
  WCL[691] = struct {
   x = [-159.925, -160.003, -160.103, -160.19, -160.258, -160.312, -160.369, -160.39, -160.378, -160.324, -160.324, -160.357, -160.434, -160.521, -160.61, -160.744, -160.854, -160.965, -161.052, -161.195,
        -161.328, -161.46, -161.505, -161.516, -161.615, -161.692, -161.847, -162.035, -162.199, -162.068, -161.913, -161.737, -161.659, -161.725, -161.814, -161.835, -161.737, -161.648, -161.749,
        -161.892, -161.979, -162.079, -162.056, -161.99, -161.969, -161.826, -161.781, -161.781, -161.88, -161.936, -161.979, -162.056, -162.19],
   y = [58.8704, 58.8258, 58.8539, 58.9056, 58.8938, 58.9337, 58.9572, 59.0065, 59.0745, 59.1215, 59.2224, 59.3022, 59.2294, 59.1379, 59.0581, 58.99, 58.9384, 58.929, 58.9384, 58.8774, 58.8258, 58.7976,
        58.7741, 58.746, 58.6873, 58.6638, 58.6239, 58.6357, 58.6544, 58.7178, 58.7225, 58.8258, 58.8305, 58.8774, 58.8657, 59.0182, 59.0863, 59.1144, 59.1496, 59.1496, 59.1097, 59.2459, 59.3186,
        59.3303, 59.3702, 59.4641, 59.504, 59.5533, 59.6331, 59.7152, 59.7927, 59.8983, 59.9828] };
  WCL[692] = struct {
   x = [-162.575, -162.643, -162.73, -162.84, -162.941, -163.06, -163.105, -163.227, -163.37, -163.492, -163.614, -163.689, -163.79, -163.901, -163.978, -164.121, -164.21, -164.077, -163.856, -163.802,
        -163.933, -164.067],
   y = [60.0813, 59.9921, 59.9217, 59.8537, 59.7833, 59.727, 59.6706, 59.6706, 59.7058, 59.6988, 59.727, 59.7152, 59.7387, 59.7387, 59.7762, 59.8373, 59.9053, 59.9429, 59.9546, 59.9828, 59.9828, 59.9757] };
  WCL[693] = struct {
   x = [-159.993, -160.092, -160.181, -160.247, -160.333, -160.423, -160.5, -160.512, -160.61, -160.667, -160.732, -160.765, -160.819, -160.798, -160.843, -160.965, -161.052, -161.195, -161.317, -161.451,
        -161.57, -161.627, -161.493, -161.516, -161.648, -161.77, -161.826, -161.868, -161.892, -161.913, -162.035, -162.089, -162.157, -162.223, -162.277, -162.378, -162.443, -162.521, -162.554,
        -162.542, -162.542, -162.575, -162.619, -162.697, -162.708, -162.685, -162.676, -162.652, -162.762, -162.852, -162.962, -163.06, -163.138, -163.227, -163.206, -163.182, -163.227, -163.337,
        -163.415, -163.415, -163.391, -163.427, -163.459, -163.537, -163.546, -163.436],
   y = [55.8242, 55.8242, 55.7937, 55.7373, 55.7256, 55.7186, 55.6881, 55.6505, 55.5707, 55.5942, 55.6388, 55.5942, 55.5778, 55.5144, 55.5073, 55.5261, 55.5402, 55.4768, 55.4463, 55.4393, 55.4135, 55.4651,
        55.5566, 55.6693, 55.6646, 55.6083, 55.5402, 55.4698, 55.4017, 55.3384, 55.2633, 55.1811, 55.1506, 55.1764, 55.1319, 55.099, 55.1248, 55.1178, 55.1248, 55.1506, 55.1882, 55.2375, 55.2703,
        55.3079, 55.2703, 55.214, 55.1506, 55.0802, 55.0497, 55.0309, 55.0051, 55.0544, 55.1436, 55.1506, 55.0873, 55.0169, 54.9488, 54.9089, 54.9089, 54.9605, 55.0309, 55.0873, 55.1365, 55.1319,
        55.1506, 55.1882] };
  WCL[694] = struct {
   x = [-169.873, -169.894, -169.894, -169.894, -169.862, -169.784, -169.751, -169.697, -169.707, -169.775, -169.829, -169.873, -169.775, -169.829, -169.873],
   y = [56.6479, 56.6362, 56.6174, 56.6174, 56.6104, 56.5916, 56.6033, 56.6174, 56.5611, 56.5493, 56.6033, 56.6479, 56.5493, 56.6033, 56.6479] };
  WCL[695] = struct {
   x = [-178.43, -178.451, -178.439, -178.406, -178.32, -178.23, -178.141, -178.076, -177.989, -177.843, -177.876, -177.965, -177.977, -178.054, -178.153, -178.242, -178.219, -178.186, -178.296, -178.418,
        -178.43],
   y = [51.9167, 51.9378, 51.9519, 51.9519, 51.9519, 51.9519, 51.9237, 51.8698, 51.8627, 51.8557, 51.8158, 51.7876, 51.6867, 51.6304, 51.6445, 51.6797, 51.7196, 51.7735, 51.8439, 51.8768, 51.9167] };
  WCL[696] = struct {
   x = [-177.822, -177.789, -177.7, -177.611, -177.501, -177.435, -177.39, -177.336, -177.28, -177.238, -177.247, -177.303, -177.357, -177.524, -177.59, -177.733, -177.822],
   y = [51.6867, 51.7196, 51.7477, 51.7548, 51.7806, 51.8017, 51.8557, 51.9308, 51.9167, 51.905, 51.8557, 51.7876, 51.7336, 51.7055, 51.7055, 51.6867, 51.6867] };
  WCL[697] = struct {
   x = [-177.038, -177.026, -176.994, -176.928, -176.928, -176.961, -176.895, -176.806, -176.752, -176.752, -176.773, -176.696, -176.585, -176.531, -176.508, -176.531, -176.618, -176.707, -176.806, -176.862,
        -176.94, -176.972, -177.015, -177.026, -177.038],
   y = [51.8087, 51.8087, 51.8087, 51.8299, 51.8838, 51.9448, 51.973, 52.0059, 51.9801, 51.905, 51.8557, 51.8439, 51.8486, 51.8557, 51.7735, 51.7688, 51.7407, 51.7266, 51.6914, 51.6515, 51.6374, 51.6984,
        51.7055, 51.7876, 51.8087] };
  WCL[698] = struct {
   x = [-174.454, -174.499, -174.443, -174.368, -174.245, -174.189, -174.189, -174.299, -174.323, -174.356, -174.433, -174.52, -174.576, -174.72, -174.818, -174.818, -174.741, -174.666, -174.597, -174.487,
        -174.454, -174.454, -174.454],
   y = [52.3861, 52.3861, 52.4048, 52.426, 52.3861, 52.3579, 52.2711, 52.2499, 52.1819, 52.142, 52.095, 52.1138, 52.0669, 52.0598, 52.095, 52.1138, 52.1561, 52.1819, 52.2171, 52.2499, 52.2969, 52.3438,
        52.3861] };
  WCL[699] = struct {
   x = [-174.156, -174.213, -174.189, -174.123, -174.058, -174.001, -173.882, -173.771, -173.682, -173.548, -173.438, -173.351, -173.241, -173.229, -173.34, -173.516, -173.649, -173.771, -173.903, -174.001,
        -174.102, -174.156],
   y = [52.1561, 52.149, 52.1561, 52.1561, 52.149, 52.149, 52.149, 52.1561, 52.1279, 52.1138, 52.1279, 52.1279, 52.1138, 52.081, 52.0528, 52.0528, 52.0411, 52.0481, 52.0669, 52.0669, 52.1138, 52.1561] };
  WCL[700] = struct {
   x = [-172.699, -172.678, -172.654, -172.589, -172.49, -172.478, -172.621, -172.755, -172.755, -172.699, -172.755, -172.699],
   y = [52.4189, 52.4119, 52.3978, 52.3861, 52.3861, 52.3391, 52.2828, 52.2828, 52.3649, 52.4189, 52.3649, 52.4189] };
  WCL[701] = struct {
   x = [-170.547, -170.559, -170.523, -170.448, -170.404, -170.458, -170.535, -170.547, -170.547],
   y = [57.2018, 57.2252, 57.237, 57.2252, 57.1783, 57.183, 57.1783, 57.2018, 57.2018] };
  WCL[702] = struct {
   x = [172.298, 172.241, 172.241, 172.286, 172.363, 172.429, 172.474, 172.539, 172.661, 172.751, 172.849, 172.938, 172.95, 173.049, 173.07, 173.135, 173.093, 173.016, 172.861, 172.751, 172.706, 172.64,
        172.64, 172.572, 172.506, 172.408, 172.298, 172.298],
   y = [52.9282, 52.9282, 52.9352, 52.9399, 52.961, 52.9751, 53.015, 53.008, 52.9868, 52.9939, 52.9751, 52.9751, 52.9282, 52.9352, 52.9071, 52.8601, 52.8343, 52.8085, 52.7874, 52.7475, 52.7944, 52.8202,
        52.8953, 52.8672, 52.8742, 52.8883, 52.9141, 52.9282] };
  WCL[703] = struct {
   x = [177.442, 177.486, 177.498, 177.498, 177.508, 177.508, 177.498, 177.465, 177.486, 177.397, 177.275, 177.144, 177.111, 177.254, 177.364, 177.442],
   y = [52.142, 52.196, 52.196, 52.1819, 52.1138, 52.0739, 52.034, 51.9918, 51.966, 51.9448, 51.9237, 51.8299, 51.912, 51.9801, 52.0481, 52.142] };
  WCL[704] = struct {
   x = [178.512, 178.524, 178.547, 178.622, 178.768, 178.911, 179.042, 179.143, 179.242, 179.275, 179.153, 179.042, 178.878, 178.744, 178.613, 178.524, 178.512],
   y = [51.6656, 51.6867, 51.6914, 51.6867, 51.6304, 51.5412, 51.4802, 51.452, 51.4379, 51.3769, 51.3699, 51.391, 51.4802, 51.5482, 51.5975, 51.6374, 51.6656] };
  WCL[705] = struct {
   x = [179.418, 179.462, 179.484, 179.552, 179.561, 179.516, 179.397, 179.331, 179.34, 179.373, 179.418],
   y = [52.0411, 52.034, 52.0199, 51.9918, 51.9519, 51.905, 51.8979, 51.9519, 51.973, 51.9988, 52.0411] };
  WCL[706] = struct {
   x = [170.376, 170.366, 170.31, 170.223, 170.089],
   y = [60.0485, 59.971, 59.9382, 59.9429, 59.971] };
  WCL[707] = struct {
   x = [164.999, 164.978, 164.811, 164.736],
   y = [59.8044, 59.7997, 59.8208, 59.9429] };
  WCL[708] = struct {
   x = [164.558, 164.328, 164.173, 164.105, 163.962, 163.72, 163.532, 163.422, 163.398, 163.267, 163.246, 163.312, 163.246, 163.091, 163.091, 163.133, 162.948, 162.88, 162.903, 162.781, 162.528, 162.284,
        162.162, 162.051, 161.986, 161.986, 162.051, 162.173, 162.307, 162.361, 162.427, 162.65, 162.814, 163.023, 163.201, 163.222, 163.145, 163.023, 162.802, 162.725, 162.715, 162.725, 162.748,
        162.748, 162.748, 162.802, 162.903, 163.079, 163.19, 163.211, 163.246, 163.267, 163.234, 163.112, 163.046, 163.002, 162.969, 162.826, 162.692, 162.638, 162.65, 162.704, 162.847, 162.969,
        162.924, 162.814, 162.704, 162.516, 162.23, 162.239, 162.295, 162.317, 162.34, 162.361, 162.173, 162.051, 161.908, 161.831, 161.72, 161.655, 161.643, 161.676, 161.777, 161.887, 162.018,
        162.129, 162.075, 161.932, 161.786, 161.566, 161.345, 161.235, 160.937, 160.606, 160.331, 160.099],
   y = [60.0367, 59.9593, 59.849, 59.9358, 59.9757, 59.8865, 59.8607, 59.7927, 59.6941, 59.5814, 59.4524, 59.3303, 59.274, 59.2271, 59.1543, 59.0863, 59.1191, 59.0698, 58.9736, 58.8539, 58.699, 58.516,
        58.3775, 58.2086, 58.0865, 57.9105, 57.8002, 57.7462, 57.7228, 57.765, 57.8988, 57.934, 57.8823, 57.8049, 57.758, 57.6641, 57.5749, 57.474, 57.3684, 57.2722, 57.1525, 57.0633, 56.9366, 56.8216,
        56.7136, 56.6831, 56.6831, 56.7019, 56.6151, 56.4954, 56.3663, 56.256, 56.1457, 56.0401, 55.9861, 55.9791, 55.9791, 56.0283, 56.1574, 56.2255, 56.2372, 56.3053, 56.3592, 56.4766, 56.5188,
        56.4461, 56.4461, 56.4203, 56.3592, 56.3053, 56.2677, 56.2443, 56.2325, 56.1762, 56.1152, 56.0659, 55.9298, 55.8007, 55.6693, 55.4956, 55.3642, 55.2375, 55.0802, 54.9981, 54.8901, 54.7892,
        54.693, 54.6038, 54.5264, 54.4888, 54.4888, 54.585, 54.5451, 54.4301, 54.2706, 54.1462] };
  WCL[709] = struct {
   x = [159.998, 160.01],
   y = [53.1535, 53.1464] };
  WCL[710] = struct {
   x = [159.998, 160.066, 160.164, 160.34, 160.463, 160.65, 160.927, 161.157, 161.368],
   y = [59.1942, 59.2341, 59.3069, 59.3585, 59.5251, 59.5861, 59.6542, 59.7997, 59.9546] };
  WCL[711] = struct {
   x = [164.37, 164.403, 164.271, 164.018, 163.741, 163.642, 163.786, 163.696, 163.577, 163.476, 163.366, 163.488, 163.765, 164.072, 164.438, 164.614, 164.581, 164.513, 164.37],
   y = [59.1708, 59.1778, 59.1027, 59.0581, 59.0135, 58.9548, 58.9337, 58.7906, 58.6591, 58.5676, 58.4526, 58.4972, 58.6404, 58.7624, 58.8422, 58.9502, 59.1027, 59.1825, 59.1708] };
  WCL[712] = struct {
   x = [166.48, 166.391, 166.313, 166.236, 166.158, 166.093, 166.081, 166.137, 166.158],
   y = [60.0086, 59.971, 59.9217, 59.8537, 59.8044, 59.8114, 59.8537, 59.8983, 59.9546] };
  WCL[713] = struct {
   x = [165.088, 165.1, 165.088, 165.044, 165.011],
   y = [60.0696, 59.9992, 59.9546, 59.8772, 59.8607] };
  WCL[714] = struct {
   x = [165.982, 166.039, 166.182, 166.203, 166.182, 166.182, 166.182, 166.215, 166.269, 166.28, 166.358, 166.489, 166.557, 166.59, 166.611, 166.611, 166.611, 166.602, 166.48, 166.337, 166.215, 166.116,
        166.048, 165.971, 165.86, 165.827, 165.827, 165.872, 165.938, 165.982, 165.982, 165.982],
   y = [55.3196, 55.3008, 55.2891, 55.282, 55.2703, 55.2445, 55.1952, 55.1365, 55.099, 55.0356, 54.9535, 54.8972, 54.808, 54.7446, 54.693, 54.7, 54.693, 54.7376, 54.7, 54.7963, 54.8455, 54.9488, 55.0309,
        55.1319, 55.2445, 55.2633, 55.3079, 55.3079, 55.3126, 55.3196, 55.3126, 55.3196] };
  WCL[715] = struct {
   x = [167.308, 167.318, 167.362, 167.428, 167.538, 167.627, 167.649, 167.738, 167.836, 167.881, 167.827, 167.782, 167.672, 167.55, 167.44, 167.35, 167.308],
   y = [54.8714, 54.8784, 54.8526, 54.7892, 54.6813, 54.7071, 54.686, 54.6484, 54.5663, 54.5334, 54.5146, 54.5522, 54.6367, 54.6813, 54.6813, 54.8197, 54.8714] };
  WCL[716] = struct {
   x = [160.099, 159.899, 159.878, 159.878, 159.888, 159.888, 159.911, 159.923, 159.988, 159.998],
   y = [54.1462, 54.0054, 53.8294, 53.6721, 53.5548, 53.4351, 53.3905, 53.2919, 53.2192, 53.1535] };
  WCL[717] = struct {
   x = [160.01, 159.866, 159.7, 159.491, 159.282, 158.984, 158.806, 158.719, 158.609, 158.519, 158.609, 158.585, 158.552, 158.519, 158.475, 158.486, 158.508, 158.475, 158.409, 158.254, 158.078, 157.858,
        157.658, 157.503, 157.315, 157.074, 156.874, 156.83, 156.82, 156.787, 156.719, 156.621, 156.6, 156.6, 156.588, 156.487, 156.323, 156.224, 156.168, 156.147, 156.046, 155.968, 155.926, 155.881,
        155.858, 155.804, 155.759, 155.694, 155.682, 155.628, 155.605, 155.605, 155.628, 155.738, 155.858, 155.935, 156.057, 156.29, 156.532, 156.654, 156.808, 156.907, 157.041, 156.984, 156.898,
        157.074, 157.449, 157.637, 157.747, 158.078, 158.421, 158.674, 158.939, 159.205, 159.413, 159.625, 159.777, 159.81, 159.944, 159.998],
   y = [53.1464, 53.1722, 53.2192, 53.1722, 53.1253, 52.9939, 52.8812, 52.9329, 52.9399, 52.846, 52.8061, 52.7334, 52.6395, 52.6207, 52.5597, 52.4518, 52.3579, 52.2147, 52.0997, 51.8768, 51.7336, 51.5811,
        51.4919, 51.3347, 51.1821, 51.0648, 50.9193, 50.8982, 50.9334, 51.0507, 51.168, 51.3206, 51.5553, 51.8158, 52.0598, 52.318, 52.5456, 52.8132, 53.0596, 53.3506, 53.5736, 53.7965, 53.9584,
        53.9983, 54.057, 54.2565, 54.4888, 54.6484, 54.9535, 55.1811, 55.4322, 55.6435, 55.9298, 56.1762, 56.4461, 56.6831, 56.8098, 56.8826, 57.0023, 57.0328, 57.0915, 57.2651, 57.4928, 57.6805,
        57.8002, 57.8166, 57.765, 57.8706, 57.9692, 57.9692, 58.063, 58.2367, 58.3869, 58.5089, 58.6709, 58.8023, 58.8985, 59.0698, 59.1496, 59.1942] };
  WCL[718] = struct {
   x = [154.666, 154.546, 154.424, 154.38, 154.269, 154.192, 154.281, 154.358, 154.236, 154.225, 154.412, 154.6, 154.776, 154.987, 155.107, 155.163, 155.142, 154.954, 154.886, 154.776, 154.633, 154.502,
        154.445, 154.225, 154.06, 153.915, 153.64, 153.42, 153.387, 153.143, 153.032, 152.889, 152.669, 152.535, 152.359, 152.237, 151.995, 151.742, 151.453, 151.354, 151.188, 151.167, 151.376,
        151.587, 151.786, 152.073, 152.281, 152.216, 152.094, 151.861, 151.786, 151.608, 151.444, 151.188, 151.024, 150.934, 150.758, 150.747, 150.66, 150.505, 150.317, 150.174],
   y = [60.086, 59.9358, 59.8161, 59.8161, 59.8607, 59.8161, 59.7105, 59.6049, 59.5416, 59.4688, 59.4805, 59.504, 59.4594, 59.4195, 59.3233, 59.2341, 59.1543, 59.1543, 59.1097, 59.1496, 59.1543, 59.159,
        59.0816, 59.0417, 59.0534, 59.1426, 59.1942, 59.1309, 59.0698, 59.0417, 58.9502, 58.8938, 58.9384, 59.0135, 58.9666, 58.8704, 58.8305, 58.8187, 58.8352, 58.9056, 59.0182, 59.0745, 59.1262,
        59.1262, 59.1191, 59.1379, 59.1825, 59.2505, 59.2623, 59.2904, 59.3867, 59.4805, 59.5416, 59.5697, 59.5486, 59.4313, 59.4594, 59.497, 59.5204, 59.5861, 59.5814, 59.6213] };
  WCL[719] = struct {
   x = [156.543, 156.543, 156.478, 156.367, 156.278, 156.302, 156.421, 156.51, 156.543],
   y = [50.8278, 50.877, 50.8348, 50.7714, 50.6963, 50.647, 50.6541, 50.7503, 50.8278] };
  WCL[720] = struct {
   x = [156.09, 156.057, 155.968, 155.87, 155.781, 155.682, 155.551, 155.45, 155.363, 155.295, 155.285, 155.295, 155.372, 155.494, 155.682, 155.881, 156.025, 156.09, 156.201, 156.201, 156.156, 156.09],
   y = [50.7175, 50.6893, 50.633, 50.5274, 50.4218, 50.3654, 50.3373, 50.2739, 50.2739, 50.2176, 50.0908, 50.0486, 50.0181, 50.0768, 50.1753, 50.2105, 50.3302, 50.4218, 50.5133, 50.6189, 50.7104, 50.7104] };
  WCL[721] = struct {
   x = [150.174, 149.909, 149.733, 149.477, 149.268, 149.059, 149.036, 149.137, 149.146, 148.914, 148.749, 148.839, 148.914, 148.904, 148.749, 148.66, 148.562, 148.496, 148.374, 148.01, 147.844, 147.844,
        147.722, 147.534, 147.391, 147.182, 146.895, 146.696, 146.496, 146.398, 146.374, 146.32, 146.121, 146.001, 145.912, 145.912, 145.879, 145.778, 145.548, 145.349, 145.194, 144.842, 144.455,
        144.178, 143.802, 143.427, 143.173, 142.976, 142.866, 142.634, 142.324, 141.993, 141.805, 141.585, 141.221, 140.866, 140.601, 140.547, 140.392, 140.181],
   y = [59.6213, 59.6589, 59.6988, 59.7199, 59.7105, 59.6378, 59.5814, 59.5134, 59.4594, 59.4524, 59.4078, 59.3303, 59.3303, 59.2388, 59.2153, 59.2107, 59.2153, 59.2904, 59.3679, 59.3749, 59.3303, 59.2505,
        59.2341, 59.2341, 59.2623, 59.3116, 59.3515, 59.4078, 59.4406, 59.4078, 59.2224, 59.1778, 59.1426, 59.159, 59.2388, 59.3585, 59.3679, 59.3585, 59.3914, 59.3867, 59.4031, 59.3749, 59.3679,
        59.3749, 59.3515, 59.3303, 59.3303, 59.2904, 59.2505, 59.1942, 59.0698, 58.9149, 58.7436, 58.584, 58.4292, 58.2719, 58.0865, 57.941, 57.7744, 57.711] };
  WCL[722] = struct {
   x = [139.862, 140.029, 140.193, 140.315, 140.327, 140.369, 140.568, 140.789, 141.021, 141.275, 141.409, 141.409, 141.418, 141.34, 141.275, 141.197, 141.176, 141.021, 140.855, 140.69, 140.524, 140.303,
        140.05],
   y = [54.1462, 54.1086, 54.0312, 53.9467, 53.8106, 53.7449, 53.6533, 53.5219, 53.4234, 53.3177, 53.2966, 53.2262, 53.1793, 53.1065, 53.0526, 53.0197, 53.0197, 53.0807, 53.1464, 53.1323, 53.1535, 53.252,
        53.2591] };
  WCL[723] = struct {
   x = [139.939, 140.094, 140.336, 140.58, 140.777, 140.965, 141.153, 141.23, 141.275, 141.23, 141.143, 141.275, 141.409, 141.495, 141.43, 141.397, 141.385, 141.275, 141.186, 141.054, 140.923, 140.822,
        140.855, 140.777, 140.768, 140.69, 140.667, 140.667, 140.613, 140.503, 140.503, 140.503, 140.557, 140.613, 140.69, 140.702],
   y = [53.1863, 53.1981, 53.1535, 53.0596, 53.0737, 52.961, 52.8859, 52.7545, 52.6067, 52.5198, 52.4048, 52.311, 52.2359, 52.1678, 52.0598, 52.0317, 51.9167, 51.8275, 51.7055, 51.6585, 51.5342, 51.4309,
        51.3628, 51.3135, 51.2525, 51.2385, 51.107, 51.0014, 50.8418, 50.7245, 50.6259, 50.4781, 50.3443, 50.1613, 50.119, 50.0322] };
  WCL[724] = struct {
   x = [142.645, 142.711, 142.831, 142.953, 142.986, 143.009, 143.087, 143.173, 143.284, 143.385, 143.394, 143.328, 143.197, 143.218, 143.295, 143.317, 143.307, 143.385, 143.495, 143.549, 143.582, 143.683,
        143.793, 143.924, 144.068],
   y = [54.341, 54.3668, 54.3269, 54.1344, 53.9467, 53.7308, 53.5219, 53.3506, 53.1464, 52.8132, 52.5668, 52.3297, 52.2758, 52.0739, 51.8698, 51.6984, 51.5342, 51.5482, 51.3699, 51.1821, 51.1399, 50.8348,
        50.5485, 50.2739, 50.0111] };
  WCL[725] = struct {
   x = [142.267, 142.225, 142.235, 142.214, 142.148, 142.136, 142.202, 142.246, 142.246, 142.157, 142.07, 141.904, 141.772, 141.805, 141.749, 141.707, 141.695, 141.772, 141.871, 141.894, 141.915, 141.904,
        141.859, 141.948, 142.136, 142.246, 142.38, 142.577, 142.622, 142.577, 142.622, 142.72, 142.777, 142.678, 142.645, 142.622, 142.589, 142.523, 142.413, 142.39, 142.434, 142.556, 142.645],
   y = [49.9759, 50.0486, 50.1824, 50.3584, 50.4922, 50.7034, 50.9193, 51.0859, 51.2314, 51.3699, 51.499, 51.6233, 51.6914, 51.7665, 51.8909, 52.1209, 52.318, 52.4729, 52.6395, 52.8413, 53.0127, 53.1981,
        53.3318, 53.4633, 53.5219, 53.5008, 53.4304, 53.4421, 53.5008, 53.5665, 53.658, 53.6533, 53.6862, 53.8223, 53.9209, 53.9256, 54.0101, 54.0875, 54.2049, 54.2753, 54.3011, 54.3011, 54.341] };
  WCL[726] = struct {
   x = [140.181, 139.961, 139.895, 139.895, 139.85, 139.773, 139.576, 139.421, 139.254, 139.078, 138.968, 138.869, 138.682, 138.604, 138.571, 138.461, 138.405, 138.273, 138.151, 138.139, 138.062, 137.898,
        137.797, 137.764, 137.731, 137.609, 137.454, 137.29, 137.114, 136.98, 136.849, 136.67, 136.56, 136.44, 136.297, 136.119, 135.955, 135.922, 135.811, 135.523, 135.326, 135.248, 135.236, 135.347,
        135.567, 135.755, 135.976, 136.184, 136.494, 136.605, 136.694, 136.783, 136.825, 136.816, 136.804, 136.837, 136.816, 136.804, 136.804, 136.792, 136.748, 136.759, 136.816, 136.893, 137.058,
        137.212, 137.266, 137.156, 137.224, 137.388, 137.686, 137.665, 137.499, 137.379, 137.489, 137.665, 137.698, 137.576, 137.454, 137.29, 137.311, 137.576, 137.919, 138.184, 138.383, 138.515,
        138.592, 138.437, 138.383, 138.581, 138.714, 138.759, 138.813, 138.759, 138.747, 138.759, 138.935, 139.144, 139.343, 139.519, 139.62, 139.763, 139.773, 139.806, 139.817, 139.862],
   y = [57.711, 57.6054, 57.6218, 57.6171, 57.5561, 57.4693, 57.3684, 57.2651, 57.2792, 57.1478, 57.0563, 57.0093, 56.9178, 56.8216, 56.7864, 56.7981, 56.7136, 56.6151, 56.5564, 56.4766, 56.378, 56.317,
        56.2184, 56.1832, 56.1386, 56.0659, 55.9861, 55.911, 55.8054, 55.7444, 55.6693, 55.5895, 55.5754, 55.5144, 55.3571, 55.275, 55.1929, 55.1929, 55.1624, 55.0169, 54.9347, 54.7822, 54.7235,
        54.686, 54.6343, 54.5404, 54.5451, 54.5522, 54.5968, 54.6226, 54.6155, 54.5968, 54.5006, 54.4231, 54.3597, 54.3152, 54.2565, 54.2189, 54.1861, 54.0945, 53.9725, 53.7707, 53.7449, 53.8294,
        53.8294, 53.9326, 54.0101, 54.1086, 54.2049, 54.2706, 54.2823, 54.2236, 54.1133, 54.0945, 53.9913, 53.9326, 53.8552, 53.7378, 53.658, 53.6205, 53.5665, 53.5477, 53.6064, 53.6792, 53.8153,
        53.9209, 53.8153, 53.6533, 53.5008, 53.5665, 53.7895, 53.9584, 54.057, 54.0617, 54.1978, 54.2823, 54.2307, 54.1861, 54.1861, 54.2307, 54.2565, 54.2565, 54.172, 54.2565, 54.2049, 54.1462] };
  WCL[727] = struct {
   x = [140.05, 139.939],
   y = [53.2591, 53.1863] };
  WCL[728] = struct {
   x = [137.356, 137.499, 137.323, 137.302, 137.388, 137.51, 137.632, 137.764, 137.919, 138.008, 138.074, 138.172, 137.985, 137.82, 137.654, 137.588, 137.51, 137.356],
   y = [54.8338, 54.9347, 54.8009, 54.7188, 54.8009, 54.7446, 54.6484, 54.6226, 54.7118, 54.7939, 54.9465, 55.0239, 55.1248, 55.1624, 55.1929, 55.092, 54.9418, 54.8338] };
  WCL[729] = struct {
   x = [109.59, 109.623, 109.646, 109.689, 109.811, 109.844, 109.811, 109.712, 109.7, 109.656, 109.569, 109.48, 109.468, 109.381, 109.149, 109.039, 109.018, 108.795, 108.586, 108.663, 108.863, 108.872,
        108.619, 108.443, 108.243, 107.99, 107.703, 107.459, 107.14, 106.908, 106.741, 106.478, 106.255, 106.046, 105.704, 105.218, 104.699, 104.401, 104.038, 103.826, 103.838, 104.192, 104.512,
        104.678, 104.911, 105.209, 105.483, 105.671, 105.882, 106.103, 106.41, 106.699, 106.896, 107.107, 107.361, 107.635, 107.746, 107.581, 107.218, 106.941, 107.095, 107.405, 107.602, 107.858,
        108.1, 108.243, 108.464, 108.63, 108.762, 108.917, 109.072, 109.194, 109.226, 109.304, 109.402, 109.503, 109.59],
   y = [55.6881, 55.6881, 55.6881, 55.6763, 55.6388, 55.5332, 55.3689, 55.1483, 54.9465, 54.7751, 54.571, 54.3668, 54.1344, 53.8927, 53.658, 53.5477, 53.7496, 53.6792, 53.4492, 53.4304, 53.4633, 53.2966,
        53.1981, 53.0854, 52.8601, 52.7076, 52.6536, 52.6137, 52.5269, 52.3438, 52.2898, 52.2499, 52.0317, 51.7665, 51.5694, 51.452, 51.3816, 51.4027, 51.452, 51.5623, 51.6304, 51.7055, 51.7524,
        51.7407, 51.8017, 51.8275, 51.9519, 52.1279, 52.3509, 52.4846, 52.6325, 52.7803, 52.9071, 52.9868, 53.0526, 53.1793, 53.2708, 53.2051, 53.0455, 52.9868, 53.2051, 53.3365, 53.4891, 53.6979,
        53.8622, 54.057, 54.2494, 54.4231, 54.6554, 54.8643, 55.0614, 55.1999, 55.3384, 55.5332, 55.6317, 55.681, 55.6881] };
  WCL[730] = struct {
   x = [29.9386, 30.0935, 30.2155],
   y = [59.9241, 59.9241, 59.964] };
  WCL[731] = struct {
   x = [31.4499, 31.4288, 31.3608, 31.1308, 30.9876],
   y = [60.0837, 59.9804, 59.9358, 59.9124, 59.9405] };
  WCL[732] = struct {
   x = [23.5787, 23.4449, 23.3463],
   y = [60.0062, 59.964, 59.9804] };
  WCL[733] = struct {
   x = [23.3135, 23.2243, 23.114, 22.9262, 22.8394, 23.0154, 23.0694],
   y = [60.0907, 59.9968, 59.9077, 59.8584, 59.8701, 59.9358, 59.9898] };
  WCL[734] = struct {
   x = [19.9176, 20.0044, 20.1382, 20.2813, 20.3916, 20.5465, 20.6662, 20.7765, 20.8657, 20.9525, 21.0089, 21.0417, 21.0535, 21.0746, 21.0628, 21.0417, 20.9971, 20.9197, 20.8422, 20.7319, 20.5888, 20.6568,
        20.7765, 20.9314, 21.1403, 21.1731, 21.1731, 21.2412, 21.2506, 21.2177, 21.1731, 21.1966, 21.1849, 21.1309, 21.098, 21.0417, 21.03, 21.0628, 21.03, 21.03, 20.9971, 20.976, 20.9197, 20.9197,
        20.9525, 20.9643, 20.9877, 21.0089, 21.0746, 21.1403, 21.2834, 21.3609, 21.3937, 21.5158, 21.6918, 21.8584, 22.079, 22.2879, 22.485, 22.5296, 22.6305, 22.7948, 22.9826, 23.0271, 23.0811,
        23.1046, 23.114, 23.1586, 23.2032, 23.3135, 23.4895, 23.6115, 23.7101, 23.8744, 23.9964, 24.0081, 24.0856, 24.1184, 24.2287, 24.2944, 24.3062, 24.3062, 24.3062, 24.3062, 24.3062, 24.3601,
        24.393, 24.4165, 24.4822, 24.4704, 24.339, 24.2053, 24.1513, 24.0081, 23.8978, 23.7875, 23.6561, 23.5787, 23.4684, 23.4449, 23.5106, 23.588, 23.7101, 23.6772, 23.5223, 23.4238, 23.4238,
        23.4449, 23.4895, 23.4566, 23.3792, 23.4003, 23.4238, 23.4449, 23.5223, 23.6209, 23.6444, 23.7429, 23.8978, 24.0175, 24.0292, 24.095, 24.163, 24.2498, 24.3719, 24.4822, 24.5362, 24.6371,
        24.7356, 24.7802, 24.8459, 24.968, 25.1205, 25.1886, 25.3411, 25.442, 25.4186, 25.5078, 25.5946, 25.5852, 25.6274, 25.7706, 25.8269, 25.8926, 25.9255, 26.0593, 26.2892, 26.5099, 26.6436,
        26.7305, 26.9511, 27.2374, 27.4814, 27.7443, 27.9555, 27.9977, 27.9766, 27.9555, 28.0095, 28.1432, 28.2535, 28.3638, 28.4389, 28.5492, 28.6948, 28.7816, 28.9459, 29.0022],
   y = [54.9347, 54.991, 54.984, 54.9981, 55.0028, 55.0661, 55.1365, 55.2187, 55.3008, 55.3877, 55.4886, 55.5754, 55.6693, 55.7256, 55.6622, 55.5684, 55.4369, 55.3501, 55.2492, 55.1483, 55.0614, 54.977,
        54.991, 54.9465, 54.9465, 55.0356, 55.2046, 55.2562, 55.3008, 55.3806, 55.3806, 55.4628, 55.5449, 55.6693, 55.7749, 55.8547, 55.9532, 55.7115, 55.7796, 55.9227, 56.0283, 56.1457, 56.256,
        56.3053, 56.3898, 56.5235, 56.6338, 56.7652, 56.8568, 56.9295, 57.0445, 57.1689, 57.3238, 57.467, 57.5866, 57.65, 57.6735, 57.7274, 57.7627, 57.7274, 57.6265, 57.5209, 57.42, 57.3121, 57.2463,
        57.3003, 57.3543, 57.2769, 57.1877, 57.1102, 57.0563, 57.0187, 57.0187, 57.0609, 57.0915, 57.0304, 56.9718, 57.122, 57.1877, 57.2581, 57.413, 57.5608, 57.6617, 57.7979, 57.88, 57.9973, 58.1147,
        58.2414, 58.347, 58.3986, 58.4151, 58.4151, 58.34, 58.293, 58.3752, 58.3705, 58.3916, 58.5019, 58.5817, 58.6803, 58.7718, 58.7718, 58.7882, 58.8234, 58.8352, 58.8398, 58.868, 58.9666, 58.9947,
        59.0393, 59.0464, 59.1144, 59.1637, 59.2365, 59.2764, 59.2881, 59.2928, 59.3045, 59.3397, 59.3491, 59.389, 59.4242, 59.4336, 59.4453, 59.5017, 59.5017, 59.5017, 59.497, 59.497, 59.5861, 59.558,
        59.5298, 59.5416, 59.5462, 59.5462, 59.5861, 59.6472, 59.6472, 59.6307, 59.6472, 59.6964, 59.619, 59.6354, 59.6307, 59.6472, 59.6354, 59.5861, 59.5744, 59.558, 59.5017, 59.4618, 59.4688,
        59.4688, 59.4571, 59.497, 59.5627, 59.6636, 59.7082, 59.7903, 59.7575, 59.6964, 59.7316, 59.8255, 59.8466, 59.8021, 59.8255, 59.8795, 59.9476] };
  WCL[735] = struct {
   x = [29.2556, 29.4434, 29.6194, 29.7743, 29.9386],
   y = [60.0062, 59.9898, 59.9569, 59.9405, 59.9241] };
  WCL[736] = struct {
   x = [22.938, 22.8394, 22.5414, 22.2996, 22.1893, 22.0884, 21.9898, 21.8256, 21.8256, 21.835, 21.8256, 21.8256, 21.9453, 22.0462, 22.0556, 21.957, 21.9453, 22.1001, 22.2222, 22.255, 22.2879, 22.3865,
        22.4968, 22.5859, 22.6728, 22.7502, 22.8159, 22.938, 23.0271, 23.1797, 23.1586, 23.06, 22.9943, 22.938, 22.938, 22.938],
   y = [58.6404, 58.6521, 58.6286, 58.6333, 58.5301, 58.5418, 58.5089, 58.5418, 58.455, 58.3986, 58.3705, 58.3165, 58.2414, 58.2132, 58.1428, 58.056, 57.9856, 58.0912, 58.2132, 58.2837, 58.3235, 58.3118,
        58.2837, 58.2837, 58.3001, 58.3705, 58.3869, 58.4151, 58.4737, 58.4855, 58.5254, 58.5653, 58.6122, 58.6239, 58.6521, 58.6404] };
  WCL[737] = struct {
   x = [22.2996, 22.7174, 22.7174, 22.6517, 22.5202, 22.485, 22.4099, 22.3536, 22.309, 22.1987, 22.0556, 21.9898, 22.0556, 22.2222, 22.309, 22.3653, 22.4968, 22.5742, 22.7408, 22.9051, 22.9051, 22.8065,
        22.6962, 22.5953, 22.4968, 22.4428, 22.3982, 22.3325, 22.2996],
   y = [58.9783, 59.0464, 59.0464, 59.0628, 59.1309, 59.0957, 59.0464, 58.983, 58.9783, 58.9783, 58.9666, 58.9314, 58.9079, 58.9079, 58.8352, 58.7436, 58.7483, 58.8352, 58.8751, 58.8915, 58.983, 59.0346,
        59.0558, 59.1027, 59.1144, 59.0745, 59.0159, 58.9877, 58.9783] };
  WCL[738] = struct {
   x = [27.458, 27.5143, 27.3265, 27.0942, 26.9299, 26.9182, 27.0614, 27.1505, 27.2045, 27.3265, 27.4134, 27.5589, 27.5589, 27.5589, 27.6575, 27.7678, 27.756, 27.702, 27.6129, 27.458],
   y = [59.0229, 59.0346, 59.0276, 58.9947, 58.9267, 58.8398, 58.7155, 58.5887, 58.4268, 58.3048, 58.1499, 58.1029, 58.2414, 58.3705, 58.4972, 58.6404, 58.807, 58.9431, 59.0065, 59.0229] };
  WCL[739] = struct {
   x = [11.1076, 11.0864, 11.063, 10.9644, 10.8541, 10.7884, 10.657, 10.6335, 10.6007, 10.5889, 10.5561, 10.5138, 10.5232, 10.657, 10.657, 10.5561, 10.4786, 10.4458, 10.4035, 10.3683, 10.3683, 10.3918,
        10.4786, 10.4458, 10.3355, 10.3355, 10.258, 10.1829, 10.204, 10.2252, 10.204, 10.1501, 10.2252, 10.3472, 10.3589, 10.3026, 10.2158, 10.204, 10.1829, 10.1383, 10.0046],
   y = [59.1238, 59.1708, 59.1708, 59.1637, 59.22, 59.2435, 59.2271, 59.2834, 59.3327, 59.4289, 59.6026, 59.7082, 59.7645, 59.7974, 59.8631, 59.8912, 59.8748, 59.8584, 59.8631, 59.8349, 59.7692, 59.6918,
        59.6143, 59.5509, 59.5345, 59.6073, 59.68, 59.7129, 59.6683, 59.6026, 59.558, 59.5416, 59.4618, 59.3961, 59.2928, 59.2599, 59.2318, 59.159, 59.1074, 59.0511, 59.0159] };
  WCL[740] = struct {
   x = [18.6268, 18.7043, 18.7371, 18.6925, 18.6714, 18.5822, 18.4719, 18.3851, 18.2302, 17.9228, 17.8007, 17.6576, 17.5801, 17.5238, 17.4698, 17.3267, 17.2046, 17.1953, 17.0732, 16.8972, 16.8737, 16.6531,
        16.51, 16.3903, 16.3129, 16.2119, 16.3129, 16.5663, 16.7963, 17.0075, 17.1178, 17.1601, 17.2492, 17.4698, 17.7773, 18.1434, 18.3288, 18.4297, 18.4508, 18.3405, 18.3405, 18.2185, 18.1316,
        18.1528, 18.1856, 17.9228, 17.789, 17.6247, 17.5684, 17.5684, 17.4253, 17.4253, 17.3267, 17.2281, 17.1178, 16.984, 16.8737, 16.8409, 16.7095, 16.4771, 16.0688, 16.0359, 16.0688, 16.1251,
        16.3011, 16.3997, 16.5428, 16.5992, 16.7189, 16.6766, 16.5006, 16.334, 16.2894, 16.4654, 16.6086, 16.632, 16.6086, 16.5874, 16.5992, 16.5428, 16.5335, 16.5874, 16.5006, 16.3903, 16.3668,
        16.3903, 16.4771, 16.51, 16.3997, 16.4771, 16.4889, 16.3903, 16.3129, 16.334, 16.2565, 16.2354, 16.0688, 15.9045, 15.8153, 15.6065, 15.3859, 15.1864, 15.0221, 14.7569, 14.5692, 14.4917,
        14.3157, 14.2054, 14.1186, 14.1725, 14.0951, 14.0083, 13.8205, 13.5107, 13.1798, 12.971, 12.7386, 12.7621, 12.8161, 12.8395, 12.8395, 12.8935, 12.8278, 12.6846, 12.5626, 12.4969, 12.3983,
        12.4101, 12.5298, 12.6729, 12.6401, 12.5955, 12.5415, 12.5743, 12.7058, 12.8161, 12.8161, 12.6846, 12.4758, 12.3749, 12.1777, 12.0674, 12.0228, 11.936, 11.8586, 11.7811, 11.8022, 11.7694,
        11.7483, 11.8679, 11.8257, 11.6262, 11.6591, 11.6825, 11.6262, 11.4713, 11.3399, 11.4713, 11.5605, 11.5159, 11.4385, 11.4502, 11.4948, 11.4174, 11.3845, 11.3282, 11.2742, 11.2953, 11.2836,
        11.2507, 11.1639, 11.131, 11.1193, 11.1076, 11.0747, 11.063, 11.0301, 11.0301, 11.1076, 11.1076],
   y = [60.0555, 59.9734, 59.9194, 59.8419, 59.7645, 59.7199, 59.6307, 59.5509, 59.5064, 59.4289, 59.3961, 59.4172, 59.5181, 59.5462, 59.4899, 59.5181, 59.5697, 59.5345, 59.5017, 59.5908, 59.5791, 59.5955,
        59.5744, 59.5908, 59.5298, 59.5017, 59.4735, 59.497, 59.4406, 59.3961, 59.3444, 59.2834, 59.3163, 59.2928, 59.3233, 59.3608, 59.3843, 59.4007, 59.3562, 59.3116, 59.2552, 59.2552, 59.2998,
        59.2482, 59.1801, 59.0675, 58.9478, 58.9314, 58.9361, 59.0511, 59.0745, 58.9595, 58.8915, 58.8281, 58.7835, 58.7882, 58.7765, 58.692, 58.6568, 58.6967, 58.6756, 58.6451, 58.6451, 58.6404,
        58.6333, 58.6239, 58.6052, 58.5817, 58.5254, 58.4784, 58.4667, 58.4972, 58.4855, 58.4385, 58.3752, 58.3282, 58.2297, 58.1663, 58.1194, 58.0607, 58.0208, 57.9551, 57.9504, 57.9786, 57.9387,
        57.8847, 57.8213, 57.7321, 57.6336, 57.5679, 57.4623, 57.3613, 57.2417, 57.1032, 56.9718, 56.7652, 56.5963, 56.3663, 56.1692, 56.1692, 56.2067, 56.1762, 56.2114, 56.2184, 56.1504, 56.0471,
        56.0776, 55.9603, 55.8171, 55.5989, 55.4557, 55.4135, 55.4628, 55.4557, 55.3877, 55.4252, 55.4182, 55.4698, 55.4815, 55.5379, 55.5989, 55.6552, 55.7561, 55.8735, 56.0213, 56.1199, 56.2255,
        56.2912, 56.256, 56.2489, 56.2865, 56.3475, 56.4132, 56.4508, 56.439, 56.4742, 56.554, 56.6573, 56.7371, 56.8521, 56.953, 57.1337, 57.2534, 57.3848, 57.3965, 57.3895, 57.5022, 57.5679, 57.65,
        57.7321, 57.8518, 57.8284, 57.9152, 57.9856, 58.0795, 58.1076, 58.1381, 58.2367, 58.2719, 58.293, 58.2883, 58.3048, 58.3799, 58.4221, 58.3517, 58.3165, 58.2837, 58.3165, 58.3916, 58.4338,
        58.4456, 58.4972, 58.5817, 58.6239, 58.7272, 58.7953, 58.8868, 58.983, 59.0745, 59.1074] };
  WCL[741] = struct {
   x = [13.8088, 13.8205, 13.8534, 13.9637, 13.8862, 13.898, 13.898, 13.7219, 13.621, 13.5671, 13.4333, 13.3136, 13.215, 13.1141, 13.0273, 12.9827, 12.9944, 13.0601, 13.0813, 13.1047, 12.9264, 12.7832,
        12.6846, 12.5861, 12.5532, 12.464, 12.4429, 12.5086, 12.4195, 12.288, 12.2434, 12.3537, 12.4523, 12.5532, 12.7386, 12.9052, 12.9827, 13.1258, 13.1916, 13.4779, 13.6328, 13.6985, 13.7431,
        13.7877],
   y = [58.8516, 58.9267, 59.0229, 59.0957, 59.1426, 59.2599, 59.3397, 59.3726, 59.3961, 59.4125, 59.4125, 59.3773, 59.3843, 59.3679, 59.3444, 59.267, 59.1637, 59.0464, 58.9877, 58.9267, 59.0628, 59.1074,
        59.0957, 59.0112, 58.9431, 58.8234, 58.7366, 58.6756, 58.5254, 58.4503, 58.4104, 58.455, 58.4033, 58.462, 58.5488, 58.6239, 58.5371, 58.577, 58.6521, 58.6967, 58.7202, 58.7718, 58.8164,
        58.8398] };
  WCL[742] = struct {
   x = [16.2448, 16.28, 16.3903, 16.4654, 16.5335, 16.6438, 16.7306, 16.808, 16.8644, 16.9183, 16.9301, 16.8409, 16.7752, 16.7306, 16.5992, 16.5006, 16.3457, 16.2894, 16.2683, 16.2448, 16.2448],
   y = [56.2795, 56.3053, 56.378, 56.493, 56.669, 56.8521, 56.9835, 57.1525, 57.2886, 57.3355, 57.3778, 57.3308, 57.2581, 57.1689, 56.9342, 56.8521, 56.669, 56.5235, 56.4273, 56.3287, 56.2795] };
  WCL[743] = struct {
   x = [18.0213, 18.1434, 18.2302, 18.2959, 18.4626, 18.5729, 18.5729, 18.6714, 18.77, 18.7277, 18.716, 18.7606, 18.7935, 18.8709, 18.8709, 18.8803, 18.7817, 18.6714, 18.5283, 18.4062, 18.1645, 18.0659,
        18.0331, 18.0425, 17.9979, 17.9979, 18.0542, 18.0425, 18.0753, 18.1434, 18.1199, 18.0542, 18.0213],
   y = [56.9647, 57.007, 57.0844, 57.1642, 57.2417, 57.3074, 57.3473, 57.3895, 57.4317, 57.5209, 57.7274, 57.7673, 57.7673, 57.8025, 57.8565, 57.9222, 57.9434, 57.8847, 57.9222, 57.8448, 57.6852, 57.5796,
        57.4787, 57.3965, 57.3003, 57.2346, 57.1877, 57.1102, 57.1149, 57.1149, 57.0563, 57.0023, 56.9647] };
  WCL[744] = struct {
   x = [14.9775, 14.9775, 14.9235, 14.7686, 14.6372, 14.602, 14.5809, 14.6701, 14.7898, 14.9118, 14.9775],
   y = [55.0215, 55.0474, 55.1811, 55.2562, 55.2867, 55.1858, 55.1107, 55.0544, 55.0286, 55.0286, 55.0215] };
  WCL[745] = struct {
   x = [9.8614, 10.0726, 10.258, 10.3801, 10.4129, 10.3589, 10.3261, 10.3683, 10.4246, 10.4246, 10.3026, 10.2369, 10.204, 10.1712, 10.1829, 10.1829, 10.1594, 10.1712, 10.1501, 10.0046, 10.0046, 10.1712,
        10.204, 10.082, 10.0726, 10.1149, 10.1923, 10.3143, 10.3918, 10.5232, 10.6992, 10.7884, 10.8213, 10.8095, 10.6335, 10.5467, 10.5232, 10.4129, 10.3472, 10.3472, 10.2486, 10.1712, 10.2158,
        10.1266],
   y = [57.5608, 57.6148, 57.6688, 57.7204, 57.6969, 57.6336, 57.5374, 57.4623, 57.3426, 57.2581, 57.1525, 57.0727, 57.0023, 56.9342, 56.8638, 56.8145, 56.8028, 56.8145, 56.7183, 56.6878, 56.6385, 56.669,
        56.6456, 56.5728, 56.5118, 56.4625, 56.5306, 56.5306, 56.4695, 56.5001, 56.5118, 56.4883, 56.4132, 56.3053, 56.2067, 56.1692, 56.2067, 56.1269, 56.195, 56.2419, 56.2255, 56.1011, 55.972,
        55.904] };
  WCL[746] = struct {
   x = [9.99517, 10.0937, 10.1712, 10.3918, 10.4692, 10.5889, 10.7767, 10.887, 10.9644, 10.9644, 10.8658, 10.7321, 10.6664, 10.6992],
   y = [54.3856, 54.3856, 54.4489, 54.3973, 54.3269, 54.3269, 54.3597, 54.3903, 54.3527, 54.2565, 54.1861, 54.1274, 54.0288, 53.9772] };
  WCL[747] = struct {
   x = [10.6992, 10.7438, 10.833, 10.9973, 11.1076, 11.2179, 11.3845, 11.3282, 11.3728, 11.4385, 11.6051, 11.8022, 11.9571, 11.9571, 12.0346, 12.112, 12.2317, 12.3655, 12.518, 12.6401, 12.464, 12.3655,
        12.3209, 12.4429, 12.5861, 12.7386, 12.8724, 12.971, 13.1141, 13.1587, 13.215, 13.3253, 13.4568, 13.5013, 13.5882, 13.6539, 13.6774, 13.6985, 13.8322, 13.9637, 14.1068, 14.196, 14.2828,
        14.4049, 14.4589, 14.4589, 14.3063, 14.2171],
   y = [53.9772, 53.9138, 53.9584, 54.0101, 53.9842, 53.9725, 53.9725, 54.0288, 54.0429, 54.0828, 54.165, 54.1978, 54.2166, 54.1462, 54.179, 54.2823, 54.3597, 54.456, 54.4747, 54.4607, 54.3973, 54.3386,
        54.2823, 54.3339, 54.3644, 54.3903, 54.3973, 54.3269, 54.2753, 54.1978, 54.1908, 54.1344, 54.1391, 54.1462, 54.0758, 53.9983, 53.9256, 53.8552, 53.8223, 53.7965, 53.7754, 53.7707, 53.7308,
        53.6909, 53.7566, 53.8411, 53.881, 53.888] };
  WCL[748] = struct {
   x = [13.5107, 13.5107, 13.5225, 13.5788, 13.5671, 13.4779, 13.3676, 13.3136, 13.1587, 13.0484, 13.093, 13.1047, 13.0601, 13.1376, 13.269, 13.3676, 13.3347, 13.2244, 13.1587, 13.2033, 13.3019, 13.3253,
        13.4685, 13.5553, 13.5342, 13.5013],
   y = [54.4677, 54.4489, 54.4043, 54.3457, 54.3011, 54.3386, 54.3198, 54.2893, 54.3081, 54.3644, 54.4114, 54.4607, 54.5451, 54.5569, 54.5123, 54.4935, 54.5451, 54.5897, 54.6414, 54.6789, 54.6672, 54.6155,
        54.6085, 54.5827, 54.5193, 54.4747] };
  WCL[749] = struct {
   x = [14.2171, 14.372, 14.548, 14.5692, 14.6701, 14.7804, 15.0221, 15.353, 15.5736, 15.7379, 15.9374, 16.1462, 16.3457, 16.5335, 16.7869, 17.0075, 17.2046, 17.491, 17.7444, 17.9885, 18.2419, 18.4508,
        18.6057, 18.7371, 18.716, 18.5822, 18.4837, 18.3851, 18.4062, 18.4954, 18.5822, 18.6386, 18.7606, 18.7817, 18.8592, 19.1244, 19.3778, 19.5538],
   y = [53.888, 53.9514, 53.9913, 53.9514, 54.0101, 54.0828, 54.1603, 54.1978, 54.2377, 54.2682, 54.3011, 54.3973, 54.5076, 54.5968, 54.6461, 54.7235, 54.7751, 54.8056, 54.8455, 54.8714, 54.8714, 54.8385,
        54.7939, 54.7118, 54.6414, 54.686, 54.7305, 54.7611, 54.693, 54.5522, 54.4372, 54.3457, 54.3128, 54.3527, 54.3903, 54.3973, 54.4301, 54.4865] };
  WCL[750] = struct {
   x = [19.5538, 19.5867, 19.6853, 19.7627, 19.819, 19.8401, 19.8613, 19.9059, 19.9176],
   y = [54.4865, 54.5193, 54.5639, 54.6015, 54.6531, 54.6977, 54.7165, 54.8056, 54.9347] };
  WCL[751] = struct {
   x = [-0.037549, 0.171319, 0.380187, 0.699357, 0.919959, 1.05373, 1.07485, 1.29545, 1.47147, 1.51605, 1.47147, 1.21801, 1.02087, 0.800271, 0.788537, 0.699357, 0.767415, 0.887104, 0.919959, 1.00914,
        0.964549, 0.865982, 0.976283, 1.09597, 1.18515, 1.32831, 1.36116, 1.34004, 1.32831, 1.41749, 1.54891, 1.69207, 1.79063, 1.8141, 1.75778, 1.5935, 1.41749, 1.27433, 1.03026, 0.800271, 0.633645,
        0.535078, 0.445899, 0.368453, 0.23703, 0.138463, 0.114995, 0.114995, 0.183053, 0.28162, 0.32621, 0.403656, 0.445899, 0.445899, 0.380187, 0.269886, 0.159585, 0.105608, 0.004694],
   y = [50.8019, 50.7691, 50.7972, 50.8583, 50.8935, 50.9287, 50.9709, 51.0601, 51.1516, 51.276, 51.3229, 51.33, 51.3159, 51.33, 51.4332, 51.5036, 51.5154, 51.5435, 51.6046, 51.6679, 51.7008, 51.7289,
        51.7759, 51.804, 51.7688, 51.797, 51.8651, 51.905, 51.9683, 51.9613, 52.0012, 52.1091, 52.3462, 52.5151, 52.6958, 52.8366, 52.9094, 52.9423, 52.9493, 52.9164, 52.9024, 52.8085, 52.7498,
        52.7968, 52.8155, 52.8226, 52.8226, 52.8296, 52.9282, 52.9892, 53.022, 53.0878, 53.1558, 53.2614, 53.3271, 53.4046, 53.4656, 53.5243, 53.5571] };
  WCL[752] = struct {
   x = [-0.061018, 0.028162, 0.037549, 0.061018, 0.147851, 0.258152, 0.269886, 0.225296, 0.093873],
   y = [53.6557, 53.6158, 53.6158, 53.6158, 53.6017, 53.5947, 53.6017, 53.773, 53.8364] };
  WCL[753] = struct {
   x = [5.01049, 5.0762, 5.0762, 4.98937, 4.83448, 4.7805, 4.92366, 4.98937, 5.01049],
   y = [53.2145, 53.2544, 53.2403, 53.1417, 53.0291, 52.9352, 52.9962, 53.1018, 53.2145] };
  WCL[754] = struct {
   x = [3.53433, 3.57892, 3.71035, 3.84177, 3.89809, 4.06237, 4.26185, 4.3393, 4.28297, 4.15155, 3.97554, 3.87462, 3.94268, 4.06237, 4.31583, 4.50358, 4.82275, 4.93305, 4.8556, 4.71244, 4.635, 4.47072,
        4.32756, 4.21726, 4.21726, 4.229, 4.23838, 4.25012, 4.28297, 4.29471, 4.34868, 4.4144, 4.49184, 4.635, 4.69132, 4.76877, 4.87907, 5.04335, 5.09967, 5.00111],
   y = [51.3511, 51.3722, 51.3793, 51.3159, 51.3018, 51.3581, 51.3722, 51.3652, 51.398, 51.4121, 51.398, 51.4192, 51.5083, 51.5224, 51.5647, 51.6257, 51.675, 51.7149, 51.7548, 51.7149, 51.7008, 51.7078,
        51.743, 51.804, 51.905, 51.9472, 51.9683, 51.9683, 51.9941, 51.9941, 52.0411, 52.0833, 52.1443, 52.2593, 52.4471, 52.6747, 52.8625, 52.9024, 52.9164, 52.9469] };
  WCL[755] = struct {
   x = [4.06237, 4.01778, 4.02952, 4.15155, 4.23838, 4.31583, 4.3393, 4.30644, 4.18441, 4.06237],
   y = [51.7618, 51.79, 51.79, 51.797, 51.743, 51.7008, 51.6679, 51.6679, 51.7149, 51.7618] };
  WCL[756] = struct {
   x = [3.83238, 3.86524, 3.93095, 4.00839, 4.09523, 4.10696, 4.04125, 3.91921, 3.83238],
   y = [51.6937, 51.7149, 51.7219, 51.7008, 51.675, 51.6398, 51.6468, 51.6867, 51.6937] };
  WCL[757] = struct {
   x = [3.66576, 3.64464, 3.64464, 3.69861, 3.75494, 3.75494, 3.66576],
   y = [51.5365, 51.4614, 51.4332, 51.4332, 51.4684, 51.5083, 51.5365] };
  WCL[758] = struct {
   x = [2.80681, 2.80681, 3.00395, 3.22455, 3.40056, 3.51086, 3.53433],
   y = [51.0953, 51.1375, 51.2056, 51.2877, 51.337, 51.3511, 51.3511] };
  WCL[759] = struct {
   x = [1.23913, 1.4832, 1.71554, 1.73666, 1.65921, 1.69207, 1.72492, 1.84696, 2.0793, 2.32102, 2.50877, 2.66366, 2.71764, 2.80681],
   y = [49.943, 50.058, 50.2058, 50.2833, 50.3326, 50.5086, 50.6846, 50.9005, 50.9639, 51.0108, 51.0249, 51.0531, 51.0601, 51.0953] };
  WCL[760] = struct {
   x = [10.0046, 9.7511, 9.52111, 9.56335, 9.39907, 9.26765, 9.22306, 9.10103, 9.04705, 8.82645, 8.65982, 8.46269, 8.09893, 7.96516, 7.87833, 7.65772, 7.43712, 7.13907, 7.2705, 6.99592, 6.9302, 6.75419,
        6.64389, 6.68613, 6.68613, 6.55471, 6.33411, 6.08065, 5.69577, 5.46343, 5.54088, 5.59485, 5.66057, 5.72863, 5.73801, 5.81546, 5.91402, 5.91402, 5.92576, 5.90464, 6.01494, 6.05718, 6.16748,
        6.16748, 6.16748, 6.05718, 6.12524, 6.31299, 6.23554, 6.13463, 5.87178, 5.73801, 5.76148, 5.72863, 5.59485, 5.55026, 5.48455, 5.32027, 5.29681, 5.2311, 5.2311, 5.30854, 5.48455, 5.58547,
        5.6723, 5.84831, 5.87178, 5.97035, 6.12524, 6.33411, 6.30125, 6.20268, 6.12524, 5.84831, 5.71689, 5.7826, 5.94923],
   y = [59.0159, 59.0159, 59.0839, 59.0065, 58.9079, 58.8398, 58.7835, 58.7319, 58.6286, 58.57, 58.462, 58.3118, 58.1827, 58.2883, 58.1616, 58.0607, 58.0138, 58.0724, 58.1546, 58.1663, 58.0795, 58.1311,
        58.0842, 58.2062, 58.3752, 58.2531, 58.2719, 58.3587, 58.5371, 58.6967, 58.8751, 58.9994, 58.9994, 58.9314, 58.9314, 58.9478, 58.9431, 58.9666, 58.983, 59.0511, 59.1074, 59.1708, 59.2552,
        59.2881, 59.3491, 59.3444, 59.389, 59.5064, 59.5416, 59.4805, 59.4054, 59.4453, 59.389, 59.3397, 59.3327, 59.4172, 59.3327, 59.3163, 59.3562, 59.3843, 59.4618, 59.5462, 59.6589, 59.6307,
        59.6589, 59.6519, 59.6918, 59.7481, 59.7809, 59.8255, 59.8584, 59.8138, 59.8185, 59.8091, 59.8513, 59.903, 59.9898] };
  WCL[761] = struct {
   x = [5.79434, 5.68403, 5.49629, 5.40711],
   y = [60.0837, 59.9804, 59.9194, 59.9569] };
  WCL[762] = struct {
   x = [5.43996, 5.4517, 5.4517, 5.4517, 5.40711, 5.36486, 5.29681, 5.28742, 5.27569, 5.28742, 5.35313, 5.39772],
   y = [59.741, 59.7809, 59.8419, 59.8912, 59.9124, 59.9522, 59.9476, 59.8865, 59.8349, 59.7974, 59.7692, 59.7528] };
  WCL[763] = struct {
   x = [5.26395, 5.26395, 5.26395, 5.25456, 5.22171, 5.17712, 5.17712, 5.16538, 5.16538, 5.17712, 5.19824],
   y = [59.159, 59.2153, 59.2764, 59.328, 59.3562, 59.328, 59.267, 59.1754, 59.1637, 59.159, 59.1637] };
  WCL[764] = struct {
   x = [8.65982, 8.62697, 8.62697, 8.62697, 8.62697, 8.54013, 8.4181, 8.24208, 8.17637, 8.08719, 8.09893, 8.13178, 8.13178, 8.06607, 8.0426, 8.02148, 8.03087, 8.13178, 8.15291, 8.29606, 8.38524, 8.54952,
        8.8593, 9.12449, 9.32163, 9.52111, 9.57509, 9.71824, 9.8614],
   y = [54.9089, 55.0098, 55.1225, 55.2304, 55.3501, 55.451, 55.5261, 55.5871, 55.5637, 55.5637, 55.6176, 55.7373, 55.8805, 55.9791, 56.1199, 56.2677, 56.3968, 56.5423, 56.6643, 56.9178, 57.0187, 57.1149,
        57.1525, 57.1689, 57.1994, 57.2839, 57.3848, 57.4787, 57.5608] };
  WCL[765] = struct {
   x = [10.1266, 9.89426, 9.81916, 9.90599, 9.88487, 9.65253, 9.54223, 9.66427, 9.67365, 9.57509, 9.50938, 9.55397, 9.5305, 9.50938, 9.52111, 9.56335, 9.56335, 9.47652, 9.38734, 9.38734, 9.39907, 9.52111,
        9.59856, 9.6408, 9.6408, 9.55397, 9.52111, 9.47652, 9.45305, 9.46479],
   y = [55.904, 55.9157, 55.8805, 55.7984, 55.7303, 55.7491, 55.7256, 55.6552, 55.6176, 55.5637, 55.512, 55.4745, 55.444, 55.3618, 55.275, 55.275, 55.2375, 55.2046, 55.1671, 55.0849, 55.0286, 55.0286,
        54.991, 54.9465, 54.9019, 54.9089, 54.9347, 54.9277, 54.9089, 54.8948] };
  WCL[766] = struct {
   x = [8.65982, 8.65982, 8.65982, 8.65043, 8.71615, 8.80298, 8.91328, 8.95787, 8.84757, 8.71615, 8.70441, 8.6387, 8.749, 8.82645, 8.79359, 8.84757, 8.93675, 8.91328, 8.83583, 8.82645, 8.94614, 9.06817,
        9.03532, 8.88042, 8.73727, 8.62697, 8.56125, 8.5284, 8.51666, 8.51666, 8.58238, 8.62697, 8.5284, 8.42983, 8.28668, 8.27494, 8.15291, 8.12005, 8.02148, 7.73517, 7.42539, 7.23764, 7.18366,
        7.10622, 7.04989, 7.06163, 7.13907, 7.26111, 7.34794, 7.43712, 7.48171, 7.44651, 7.35968, 7.08275],
   y = [54.9089, 54.9347, 54.8573, 54.8009, 54.7423, 54.6461, 54.578, 54.5076, 54.4607, 54.4231, 54.3856, 54.3339, 54.3269, 54.294, 54.2307, 54.165, 54.1204, 54.0758, 54.0687, 53.9842, 53.9326, 53.888,
        53.8599, 53.8552, 53.8481, 53.8552, 53.7824, 53.7308, 53.6205, 53.5407, 53.4093, 53.3177, 53.3952, 53.5337, 53.5595, 53.4492, 53.4421, 53.5736, 53.6792, 53.7097, 53.6839, 53.6533, 53.5665,
        53.5219, 53.4421, 53.4163, 53.3952, 53.3835, 53.3436, 53.2919, 53.191, 53.1981, 53.2966, 53.2567] };
  WCL[767] = struct {
   x = [9.46479, 9.38734, 9.37795, 9.37795, 9.50938, 9.66427, 9.79569, 9.85201, 9.89426, 9.88487, 9.77457, 9.84028, 9.98344, 9.99517],
   y = [54.8948, 54.8948, 54.8573, 54.8455, 54.8502, 54.8127, 54.7939, 54.7305, 54.6789, 54.5827, 54.5006, 54.4607, 54.4607, 54.3856] };
  WCL[768] = struct {
   x = [5.00111, 5.02223, 5.09967, 5.22171, 5.18651, 5.06682, 5.02223, 5.03396, 5.01049, 5.09967, 5.2311, 5.39772, 5.562, 5.71689, 5.83893, 5.84831, 5.92576, 5.93749, 5.8929, 5.80372, 5.71689, 5.57373,
        5.41884, 5.40711, 5.39772, 5.40711, 5.47517, 5.68403, 5.84831, 6.04779, 6.23554, 6.35523, 6.41155, 6.60869, 6.76358, 6.86449, 6.99592, 7.08275, 7.08275, 7.08275],
   y = [52.9469, 52.9258, 52.8273, 52.7522, 52.7052, 52.6794, 52.5926, 52.4846, 52.4119, 52.3438, 52.3156, 52.2969, 52.3368, 52.4048, 52.4987, 52.5855, 52.6536, 52.7005, 52.7334, 52.7803, 52.8601, 52.8789,
        52.8789, 52.9469, 53.0455, 53.1582, 53.2309, 53.3295, 53.3835, 53.4022, 53.3623, 53.3905, 53.428, 53.4492, 53.428, 53.3905, 53.3248, 53.2779, 53.245, 53.2567] };
  WCL[769] = struct {
   x = [-7.08275, -7.07336, -7.15315, -7.29396, -7.416, -7.45824, -7.40661, -7.28458, -7.16254, -7.08275],
   y = [62.0644, 62.0644, 62.0902, 62.0996, 62.116, 62.0902, 62.0597, 62.0433, 62.0433, 62.0644] };
  WCL[770] = struct {
   x = [-6.77531, -6.75185, -6.75185, -6.76123, -6.82694, -6.91612, -6.99592, -7.11795, -7.18366, -7.22591, -7.25876, -7.24937, -7.16019, -7.04755, -6.96306, -6.89735, -6.89735, -6.89735, -6.87388, -6.81756,
        -6.77531],
   y = [61.9611, 61.9564, 61.9752, 62.0339, 62.1113, 62.1724, 62.2193, 62.2709, 62.2803, 62.224, 62.1888, 62.1512, 62.1254, 62.1066, 62.0644, 62.0386, 62.0128, 61.9705, 61.9494, 61.9611, 61.9611] };
  WCL[771] = struct {
   x = [-6.66501, -6.62277, -6.58991, -6.5993, -6.64624, -6.72134, -6.75419, -6.84337, -6.8997, -6.94194, -7.00765, -6.96541, -6.86684, -6.79643, -6.75419, -6.72134, -6.66501],
   y = [62.055, 62.0644, 62.0691, 62.0996, 62.1372, 62.1841, 62.2193, 62.2404, 62.285, 62.3014, 62.2897, 62.2193, 62.1724, 62.1207, 62.0691, 62.0597, 62.055] };
  WCL[772] = struct {
   x = [-6.45614, -6.42329, -6.489, -6.5641, -6.66266, -6.67675, -6.62981, -6.55471, -6.489, -6.47492, -6.47492, -6.45614],
   y = [62.2662, 62.2756, 62.2803, 62.2897, 62.2967, 62.2545, 62.2029, 62.1724, 62.1841, 62.2029, 62.2404, 62.2662] };
  WCL[773] = struct {
   x = [-6.60869, -6.5993, -6.62277, -6.75419, -6.82929, -6.82929, -6.76358, -6.76358, -6.73072, -6.68848, -6.63216, -6.60869],
   y = [61.757, 61.8203, 61.8673, 61.8837, 61.8626, 61.825, 61.8109, 61.8109, 61.7781, 61.7523, 61.7429, 61.757] };
  WCL[774] = struct {
   x = [-6.89735, -6.91612, -6.91612, -6.93959, -6.9302, -6.90674, -6.85041, -6.79409, -6.75185, -6.74246, -6.76123, -6.82694, -6.89735],
   y = [61.6162, 61.649, 61.642, 61.595, 61.5481, 61.4589, 61.4073, 61.3956, 61.4167, 61.4707, 61.5223, 61.581, 61.6162] };
  WCL[775] = struct {
   x = [-1.23913, -1.2626, -1.2626, -1.28372, -1.31658, -1.42688, -1.47147, -1.42688, -1.40575, -1.34943, -1.41749, -1.54891, -1.50432, -1.40575, -1.28372, -1.23913, -1.23913, -1.27433],
   y = [60.6375, 60.6469, 60.6422, 60.6258, 60.5929, 60.5507, 60.5131, 60.4756, 60.4028, 60.3442, 60.3066, 60.2503, 60.1963, 60.1682, 60.2245, 60.1799, 60.0931, 60.0039] };
  WCL[776] = struct {
   x = [-1.16403, -1.08658, -1.08658, -1.07485, -1.05373, -1.06312, -1.07485, -1.05373, -1.05373, -1.05373, -1.03026, -1.06312, -1.16403, -1.21801, -1.21801, -1.23913, -1.23913],
   y = [59.9828, 60.0203, 60.0978, 60.147, 60.1799, 60.2245, 60.2738, 60.2996, 60.3277, 60.377, 60.4193, 60.4357, 60.4521, 60.4756, 60.5225, 60.5929, 60.6375] };
  WCL[777] = struct {
   x = [-19.9856, -19.9856],
   y = [63.5453, 63.5453] };
  WCL[778] = struct {
   x = [-20.0725, -19.8847, -19.7956, -19.697, -19.6407, -19.5186, -19.4201, -19.3778, -19.3872, -19.4341, -19.3637, -19.2229, -19.1103, -19.0023, -18.9366, -18.7395, -18.7395, -18.7019, -18.6503, -18.5283,
        -18.4813, -18.3311, -18.2279, -18.1762, -18.1199, -18.0777, -18.0777, -18.1105, -18.1856, -18.2607, -18.2748, -18.1856, -17.9885, -17.8007, -17.6552, -17.491, -17.3924, -17.3032, -17.3032,
        -17.0591, -17.0263, -16.8198, -16.6508, -16.5194, -16.463, -16.4537, -16.463, -16.51, -16.5663, -16.463, -16.3645, -16.2331, -16.0782, -16.0125, -15.7825, -15.759, -15.8012, -15.6604, -15.5055,
        -15.407, -15.3647, -15.3835, -15.0972, -15.0549, -14.9986, -14.8766, -14.6466, -14.5668, -14.8343, -14.9658, -15.0737, -15.1535, -15.0972, -14.9986, -14.7358, -14.7029, -14.6654, -14.7686,
        -14.7686, -14.4682, -14.3274, -14.3274, -14.3603, -14.3274, -14.0411, -14.0083, -13.9191, -13.8393, -13.7173, -13.7501, -13.8628, -13.7407, -13.6093, -13.553, -13.5436, -13.6093, -13.7642,
        -13.7642, -13.8393, -13.8862, -13.9613, -14.1725, -14.3368, -14.3368, -14.5011, -14.4589, -14.5574, -14.426, -14.2289, -14.426, -14.5903, -14.8343, -14.9094, -15.008, -15.13, -15.13, -15.2427,
        -15.3647, -15.3976, -15.637, -15.7355, -16.0359, -16.3457, -16.5757, -16.6743, -16.8292, -17.0075, -17.0591, -17.1718, -17.2375, -17.2469, -17.5238, -17.721, -18.0542, -18.3077, -18.4954,
        -18.6362, -18.9038, -19.0352, -19.2464, -19.4529, -19.6407, -19.7862, -19.927, -19.9833],
   y = [66.0587, 65.9766, 65.9015, 65.8334, 65.7442, 65.7442, 65.7701, 65.8874, 65.9766, 66.0493, 66.0658, 66.0845, 66.0587, 66.1432, 66.1479, 66.1995, 66.1291, 66.1338, 66.0845, 66.0704, 65.9719, 65.9109,
        65.8287, 65.7114, 65.6621, 65.7747, 65.878, 65.9367, 66.0094, 66.0986, 66.169, 66.1878, 66.169, 66.1033, 66.0094, 66.0212, 66.0587, 66.1291, 66.1784, 66.1925, 66.1244, 66.115, 66.1549, 66.1643,
        66.169, 66.2488, 66.3286, 66.4225, 66.4835, 66.4905, 66.4788, 66.5398, 66.5304, 66.4389, 66.3943, 66.3122, 66.2676, 66.2347, 66.1784, 66.1737, 66.2136, 66.2676, 66.2723, 66.3192, 66.3638,
        66.3896, 66.399, 66.3638, 66.3145, 66.223, 66.1549, 66.1033, 66.0399, 66.0047, 66.0704, 66.0212, 65.9273, 65.8076, 65.7255, 65.7489, 65.7536, 65.7349, 65.6856, 65.6222, 65.58, 65.6903, 65.6433,
        65.5518, 65.4438, 65.3523, 65.3007, 65.1904, 65.1951, 65.242, 65.1223, 65.0238, 64.9956, 64.9064, 64.89, 64.8477, 64.8102, 64.7773, 64.768, 64.7022, 64.7069, 64.6788, 64.6318, 64.6225, 64.6788,
        64.4981, 64.3948, 64.3573, 64.261, 64.2798, 64.3221, 64.376, 64.4324, 64.4371, 64.3268, 64.247, 64.1836, 64.0592, 63.9536, 63.8949, 63.9043, 63.909, 63.8762, 63.9231, 63.8902, 63.8456, 63.7776,
        63.74, 63.6415, 63.496, 63.4561, 63.4303, 63.3763, 63.4256, 63.388, 63.4561, 63.4749, 63.5147, 63.5499, 63.5406, 63.5453] };
  WCL[779] = struct {
   x = [-22.1893, -22.3301, -22.4193, -22.532, -22.6211, -22.7526, -22.8746, -22.9732, -23.1374, -23.1468, -23.2689, -23.2032, -23.0952, -23.0154, -23.1797, -23.3252, -23.358, -23.4331, -23.6209, -23.7007,
        -23.9119, -23.8086, -23.7101, -23.7993, -23.7242, -23.7993, -23.9541, -24.0527, -24.1325, -24.2968, -24.3954, -24.2405, -24.1513, -24.2311, -24.2968, -24.2311, -24.1747, -24.3156, -24.4282,
        -24.3954, -24.4517, -24.5033, -24.6488, -24.6582, -24.86, -24.8929, -24.7896, -24.8694, -24.9914, -25.1463, -25.3529, -25.2871, -25.1463, -25.0994, -25.0994, -25.1229, -25.09, -25.2543,
        -25.3669, -25.3435, -25.3998, -25.5406, -25.672, -25.672, -25.6204, -25.4984, -25.672, -25.7284, -25.7518, -25.841, -25.9067, -25.9161, -25.9724, -26.0945, -26.3244, -26.3573, -26.3808,
        -26.3479, -26.47, -26.6108, -26.6577, -26.7234, -26.7891, -26.944, -27.0097, -27.052, -27.1505, -27.296, -27.3383, -27.3524, -27.4368, -27.5401, -27.6481, -27.7607, -27.7372, -27.8264,
        -27.9015, -28.047, -28.0658, -28.0799, -28.0893, -28.2019, -28.3662, -28.3662, -28.507, -28.6056, -28.7088, -28.8731, -29.0468, -29.1923, -29.2251, -29.3237, -29.3894, -29.4223, -29.488,
        -29.5443, -29.6898, -29.7649, -29.8306, -29.8869, -29.9292],
   y = [70.0507, 69.9826, 69.931, 69.9756, 69.9333, 69.9263, 69.9521, 69.9028, 69.9263, 69.8747, 69.8629, 69.8277, 69.816, 69.7831, 69.8066, 69.8442, 69.7831, 69.7714, 69.7597, 69.7597, 69.7221, 69.6728,
        69.6376, 69.6259, 69.593, 69.5297, 69.5602, 69.6071, 69.5719, 69.5837, 69.5532, 69.5273, 69.464, 69.4757, 69.4804, 69.4499, 69.4076, 69.4311, 69.4499, 69.403, 69.3865, 69.3842, 69.3724,
        69.3067, 69.2833, 69.2528, 69.2434, 69.2199, 69.2598, 69.2786, 69.2199, 69.2058, 69.2011, 69.1894, 69.1542, 69.126, 69.0838, 69.1143, 69.1002, 69.0251, 69.0439, 69.0556, 69.0838, 69.0275,
        68.9852, 68.9571, 68.9617, 68.9101, 68.8655, 68.8655, 68.8655, 68.835, 68.7951, 68.774, 68.767, 68.7435, 68.72, 68.6825, 68.6942, 68.6872, 68.652, 68.6989, 68.6919, 68.6754, 68.652, 68.5675,
        68.6003, 68.5956, 68.5745, 68.5276, 68.5112, 68.5557, 68.5511, 68.5276, 68.4713, 68.4619, 68.5229, 68.5511, 68.5018, 68.4783, 68.4736, 68.4548, 68.49, 68.4173, 68.4501, 68.422, 68.3844,
        68.3844, 68.3774, 68.361, 68.2835, 68.3164, 68.3281, 68.2812, 68.2483, 68.2436, 68.2952, 68.3211, 68.3539, 68.4009, 68.4009] };
  WCL[780] = struct {
   x = [-19.9856, -20.0279, -20.1734, -20.3377, -20.502, -20.6146, -20.8681, -21.2201, -21.4548, -21.6613, -21.7927, -21.9054, -22.079, -22.3888, -22.5343, -22.6892, -22.6751, -22.6751, -22.7408, -22.478,
        -22.3794, -22.079, -21.8162, -21.8256, -21.8162, -22.1588, -21.9805, -21.9476, -21.9711, -21.8068, -21.9945, -22.1166, -22.2151, -22.248, -22.3325, -22.356, -22.37, -22.4123, -22.4451,
        -22.5766, -22.7221, -22.8206, -22.9098, -23.0741, -23.2947, -23.4824, -23.6561, -23.8673, -23.9565, -23.9471, -23.933, -23.6796, -23.351, -23.2149, -22.9943, -22.8112, -22.6986, -22.5202,
        -22.3325, -21.9288, -21.8725, -21.7599, -21.896, -21.9945, -22.1353, -22.3466, -22.5343, -22.5202, -22.5015, -22.4029, -22.2668, -22.1682, -22.0274, -21.849, -21.8584, -22.0462, -22.1494,
        -22.2809, -22.2902, -22.37, -22.3794, -22.4451, -22.5108, -22.6423, -22.7549, -22.9192, -23.0506, -23.2384, -23.3933, -23.459, -23.5247, -23.6796, -23.7922, -23.9236, -23.9893, -24.1114,
        -24.2757, -24.5385, -24.4869, -24.3977, -24.285, -23.9236, -23.7687, -23.98, -23.9142, -23.8908, -24.0785, -24.1208, -24.1771, -23.9987, -23.7453, -23.581, -23.2618, -23.4167, -23.3041,
        -23.2853, -23.6232, -23.8016, -23.7922, -23.5904, -23.7359, -23.7781, -23.8673, -23.8579, -23.6045, -23.5575, -23.689, -23.4918, -23.6373, -23.4918, -23.1398, -23.1304, -23.0835, -22.952,
        -22.8441, -22.6986, -22.6423, -22.5766, -22.478, -22.37, -22.5015, -22.478, -22.6235, -22.7878, -22.9427, -22.83, -22.6751, -22.5672, -22.5015, -22.6, -22.6657, -22.708, -22.83, -23.0084,
        -23.1633, -23.1633, -23.0976, -23.1304, -23.0412, -22.952, -22.9614, -22.7878, -22.6751, -22.5437, -22.4686, -22.2902, -22.0133, -21.9147, -21.8068, -21.7176, -21.619, -21.5299, -21.3984,
        -21.5721, -21.4876, -21.3656, -21.3421, -21.4313, -21.4548, -21.5299, -21.7599, -21.6941, -21.4642, -21.4735, -21.3656, -21.3187, -21.4172, -21.2529, -21.2342, -21.2342, -21.098, -21.0887,
        -20.8681, -20.8774, -20.793, -20.6381, -20.5911, -20.4926, -20.3846, -20.3048, -20.2813, -20.3283, -20.3846, -20.394, -20.2813, -20.1734, -20.0748],
   y = [63.5453, 63.5288, 63.5781, 63.6039, 63.6955, 63.7307, 63.8081, 63.8668, 63.8269, 63.8574, 63.8668, 63.841, 63.8363, 63.8527, 63.8034, 63.8175, 63.9184, 63.9583, 64.0686, 63.9912, 64.0545, 64.1085,
        64.1883, 64.308, 64.3455, 64.308, 64.3854, 64.4324, 64.4934, 64.5779, 64.5685, 64.484, 64.5262, 64.5779, 64.5732, 64.6084, 64.6835, 64.768, 64.8196, 64.8337, 64.8149, 64.7867, 64.8243, 64.8431,
        64.8384, 64.8149, 64.7633, 64.7445, 64.7867, 64.8431, 64.8853, 64.8947, 64.9346, 64.9815, 64.9815, 65.0566, 65.005, 65.066, 65.0331, 65.0285, 65.0754, 65.2092, 65.1904, 65.1082, 65.1364,
        65.1481, 65.1622, 65.2232, 65.2373, 65.2913, 65.3054, 65.3523, 65.3969, 65.4485, 65.465, 65.5213, 65.472, 65.472, 65.5166, 65.5495, 65.5307, 65.594, 65.6034, 65.6034, 65.6128, 65.6222, 65.594,
        65.5025, 65.5025, 65.4884, 65.472, 65.4438, 65.4392, 65.4744, 65.4884, 65.5166, 65.5025, 65.465, 65.5518, 65.6292, 65.594, 65.526, 65.5753, 65.6128, 65.6175, 65.6762, 65.7161, 65.7607, 65.824,
        65.7841, 65.7349, 65.6433, 65.6715, 65.7114, 65.7395, 65.7982, 65.7747, 65.8616, 65.8827, 65.9015, 65.9648, 65.9907, 66.0587, 66.0399, 65.9954, 66.0165, 66.108, 66.0704, 66.1643, 66.1878,
        66.1103, 66, 65.9813, 65.9672, 66.0259, 65.9508, 65.8522, 65.9367, 65.8475, 65.9508, 66.054, 66.0986, 66.1197, 66.1596, 66.2136, 66.2981, 66.2394, 66.2394, 66.2629, 66.3075, 66.3239, 66.3591,
        66.3591, 66.3145, 66.338, 66.3544, 66.4037, 66.4694, 66.453, 66.4553, 66.4905, 66.4178, 66.4436, 66.4084, 66.4248, 66.2934, 66.2676, 66.1878, 66.1338, 66.0939, 66.0892, 66.0704, 66.0165,
        65.9601, 65.9367, 65.9062, 65.7982, 65.7701, 65.7302, 65.6973, 65.7536, 65.6621, 65.6222, 65.5448, 65.5987, 65.4931, 65.4579, 65.4251, 65.3664, 65.2326, 65.3288, 65.4063, 65.4157, 65.5893,
        65.6715, 65.6668, 65.594, 65.5166, 65.5893, 65.7067, 65.824, 65.9109, 66.0047, 66.1033, 66.1244, 66.108, 66.0587] };
  WCL[781] = struct {
   x = [-29.9315, -30.0536, -30.0864, -30.0348, -30.2178, -30.2742, -30.1334, -30.2648, -30.4197, -30.4291, -30.6591, -30.7811, -30.7811, -30.5746, -30.5417, -30.7154, -30.9172, -31.058, -31.0674, -31.1566,
        -31.1472, -31.1801, -31.0909, -31.0345, -31.1707, -31.2223, -31.3349, -31.5086, -31.6119, -31.687, -31.6776, -31.7761, -31.8747, -31.9076, -31.8419, -31.7527, -31.6964, -31.8419, -32.0625,
        -32.0859, -32.1516, -32.1845, -32.1845, -32.2033, -32.2596, -32.3722, -32.4051, -32.4145, -32.4145, -32.5037, -32.5928, -32.5459, -32.6022, -32.6914, -32.7806, -32.7571, -32.6022, -32.5037,
        -32.4051, -32.3488, -32.4802, -32.2596, -32.1282, -32.039, -32.1939, -32.1704, -32.2267, -32.2596, -32.3159, -32.438, -32.5271, -32.6586, -32.6445, -32.7571, -32.8651, -33.0763, -33.264,
        -33.2218, -33.2312, -33.3861, -33.4189, -33.4612, -33.4518, -33.6067, -33.541, -33.5175, -33.5504, -33.5832, -33.6067, -33.771, -33.8367, -33.8836, -34.015, -34.0338, -33.9681, -33.9493,
        -34.0807, -34.0573, -34.1465, -34.1371, -34.2028, -34.245, -34.2873, -34.3671, -34.4328, -34.5219, -34.5548, -34.5783, -34.6534, -34.6628, -34.7425, -34.8646, -34.9725, -35.0617, -35.104,
        -35.2495, -35.118, -35.2495, -35.4795, -35.6343, -35.8456, -35.8784, -35.8221, -35.7658, -35.6813, -35.7564, -35.7564, -35.7564, -35.9207, -35.977, -36.0192, -36.0427, -36.1741, -36.2304,
        -36.3196, -36.4088, -36.3619, -36.4933, -36.5637, -36.5825, -36.6153, -36.6716, -36.7467, -36.9345, -37.0565, -37.1222, -37.2208, -37.2443, -37.2349, -37.3992, -37.2443, -37.2771, -37.3757,
        -37.479, -37.6104, -37.7183, -37.7653, -37.7653, -37.8169, -37.6855, -37.6292, -37.6104, -37.479, -37.3428, -37.1692, -37.4086, -37.5212, -37.6432, -37.7418, -37.8498, -37.9295, -38.0844,
        -38.0516, -37.8169, -37.9295, -38.0187, -37.9718, -38.0516, -38.1924, -38.0187, -38.061, -38.1595, -38.2487, -38.2581, -38.3614, -38.4459, -38.4928, -38.3614, -38.2722, -38.1924, -38.2393,
        -38.1924, -38.2816, -38.3144, -38.3708, -38.4599, -38.535, -38.6007, -38.6993, -38.8026, -38.6665, -38.812, -38.9528, -39.1077, -39.1546, -39.3283, -39.3752, -39.3846, -39.4268, -39.5723,
        -39.6146, -39.7507, -39.8164, -39.8821],
   y = [68.4009, 68.3774, 68.3211, 68.2272, 68.2647, 68.2225, 68.1615, 68.0911, 68.1615, 68.2319, 68.2436, 68.2671, 68.1826, 68.1451, 68.0723, 68.0676, 68.1216, 68.1568, 68.1826, 68.1779, 68.1451, 68.1216,
        68.0746, 68.0418, 68.0723, 68.0676, 68.0911, 68.0911, 68.0582, 68.0418, 68.0887, 68.1169, 68.1286, 68.2108, 68.2061, 68.1896, 68.2483, 68.2554, 68.2835, 68.3328, 68.368, 68.3938, 68.4009,
        68.4619, 68.4853, 68.4009, 68.4713, 68.5299, 68.5581, 68.591, 68.6238, 68.5464, 68.5464, 68.5276, 68.5276, 68.49, 68.4056, 68.361, 68.2647, 68.2225, 68.2014, 68.1615, 68.1286, 68.0629, 68.0418,
        67.9526, 67.9315, 67.8728, 67.8986, 67.8611, 67.9198, 67.8611, 67.8188, 67.7813, 67.7273, 67.678, 67.7038, 67.6358, 67.5607, 67.5466, 67.4973, 67.4269, 67.3612, 67.3706, 67.333, 67.2486,
        67.2298, 67.1359, 67.0632, 67.0303, 67.0092, 66.9646, 66.9317, 66.9012, 66.8449, 66.805, 66.8496, 66.7933, 66.798, 66.7675, 66.7111, 66.6266, 66.7135, 66.6454, 66.7534, 66.6501, 66.575,
        66.5234, 66.4694, 66.399, 66.3591, 66.3591, 66.3286, 66.3474, 66.4436, 66.3474, 66.2582, 66.2629, 66.3028, 66.3286, 66.4248, 66.3544, 66.3028, 66.2723, 66.2277, 66.1502, 66.1385, 66.0704,
        66.1033, 66.054, 65.9954, 65.9109, 65.9648, 65.9367, 65.9648, 65.9367, 66.0493, 66.0446, 65.9766, 65.9555, 65.9015, 65.8287, 65.9062, 65.8968, 66.0259, 66.054, 66.0986, 65.9954, 65.9109,
        65.9461, 65.878, 65.824, 65.8475, 65.8874, 65.9156, 65.9156, 65.9367, 65.9648, 66.0094, 66.0587, 66.0798, 66.169, 66.2042, 66.2582, 66.3192, 66.3075, 66.3286, 66.3028, 66.3638, 66.4225,
        66.3239, 66.3638, 66.3145, 66.2347, 66.2394, 66.1995, 66.1502, 66.1244, 66.108, 66.0446, 65.9461, 65.9461, 65.9719, 65.9109, 65.9461, 66, 65.8827, 65.8193, 65.824, 65.7982, 65.7747, 65.702,
        65.702, 65.6339, 65.6527, 65.6527, 65.6762, 65.6715, 65.6715, 65.6762, 65.6034, 65.5893, 65.594, 65.5706, 65.6246, 65.6856, 65.6246, 65.6128, 65.5659, 65.5753, 65.6386, 65.6668, 65.6081,
        65.5213] };
  WCL[782] = struct {
   x = [-40.0886, -39.9572, -39.9244, -39.9478, -39.8493, -39.9009],
   y = [65.4579, 65.4345, 65.3781, 65.3335, 65.2702, 65.2373] };
  WCL[783] = struct {
   x = [-39.8798, -40.0018, -40.0769, -40.2224, -40.0863],
   y = [65.5213, 65.5518, 65.5213, 65.5025, 65.4579] };
  WCL[784] = struct {
   x = [-39.9009, -40.0042, -40.0793, -40.0886, -40.2013, -40.3327, -40.4219, -40.5298, -40.6753, -40.6847, -40.8068, -41.0039, -41.1259, -41.2151, -41.3043, -41.1072, -41.0743, -40.9523, -40.8537, -40.6988,
        -40.6425, -40.497, -40.3327, -40.3984, -40.5111, -40.5862, -40.9053, -41.1072, -41.3372, -41.5014, -41.4451, -41.1823, -40.9194, -40.7411, -40.6425, -40.6847, -40.7739, -40.6847, -40.6425,
        -40.5627, -40.8396, -41.14, -41.4029, -41.4123, -41.0743, -40.8068, -40.7411, -40.7504, -40.8537, -40.8725, -40.9053, -40.9523, -41.0508, -41.1823, -41.1729, -41.1588, -41.0837, -41.248,
        -41.2057, -41.3935, -41.5906, -41.6141, -41.5014, -41.5484, -41.6563, -41.6141, -41.3794, -41.3794, -41.5671, -41.7455, -41.8535, -41.9896, -42.1069, -41.9192, -41.7455, -41.6657],
   y = [65.2373, 65.1951, 65.1481, 65.0848, 65.0238, 65.0848, 65.0425, 65.0942, 65.1176, 65.0285, 65.066, 65.0566, 65.1036, 65.005, 64.9158, 64.9393, 64.8571, 64.8196, 64.7069, 64.6412, 64.4793, 64.4183,
        64.3948, 64.3573, 64.3854, 64.3995, 64.4042, 64.362, 64.2939, 64.308, 64.1977, 64.1883, 64.1836, 64.1648, 64.0874, 64.0498, 64.017, 63.909, 63.841, 63.7588, 63.74, 63.7776, 63.8128, 63.7588,
        63.726, 63.6955, 63.6509, 63.5593, 63.5499, 63.4561, 63.4256, 63.4256, 63.4021, 63.4303, 63.4068, 63.3763, 63.327, 63.3176, 63.2378, 63.2589, 63.327, 63.2988, 63.2496, 63.2144, 63.205, 63.1346,
        63.1205, 63.0642, 63.0595, 63.0759, 63.0947, 63.1533, 63.1533, 63.0759, 63.029, 63.0008] };
  WCL[785] = struct {
   x = [-41.6657, -41.4827, -41.4827, -41.4827, -41.4827, -41.5061, -41.6375, -41.6704, -41.7033, -41.8018, -41.891, -41.9567, -41.9567, -41.9708, -42.0694, -42.2102, -42.2759, -42.412, -42.4636, -42.4777,
        -42.52, -42.6185, -42.7734, -42.8063, -42.6326, -42.4777, -42.4214, -42.3791, -42.412, -42.29, -42.2665, -42.2571, -42.243, -42.1257, -42.0881, -42.1585, -42.1022, -42.2008, -42.2665, -42.3557,
        -42.4214, -42.4777, -42.3791, -42.4636, -42.6303, -42.6725, -42.6185, -42.6303, -42.6396, -42.6631, -42.6842, -42.696, -42.696, -42.7288, -42.7617, -42.8063, -42.8274, -42.8274, -42.8391,
        -42.8603, -42.8603, -42.8391, -42.8391, -42.9283, -43.0269, -43.0809, -43.1489, -43.2921, -43.4352, -43.6441, -43.6112, -43.4352, -43.3015, -43.2146, -43.1372, -43.1043, -43.1254, -43.1043,
        -43.0715, -43.0715, -43.0715, -43.1818, -43.224, -43.4118, -43.5995, -43.8764, -44.0079, -44.0524, -44.0524, -44.151, -44.2285, -44.4045, -44.6462, -44.8011, -44.9349, -44.9888, -45.01,
        -45.0663, -45.1203, -45.132, -45.1555, -45.1555, -45.0334, -44.9442, -44.9231, -44.9114, -44.9114, -44.8903, -44.8903, -45.0452, -45.2306, -45.2751, -45.2423, -45.1977, -45.1883, -45.2658,
        -45.3526, -45.4418, -45.5192, -45.5521, -45.6178, -45.6506, -45.6835, -45.6952, -45.7163, -45.7281, -45.7163, -45.6389, -45.3972, -45.3409, -45.4089, -45.5286, -45.7163, -45.8712, -46.0261,
        -46.0918, -46.1904, -46.059, -45.9581, -46.1036, -45.8384, -45.6624, -45.5943, -45.883, -45.9933, -45.9698, -45.9158, -45.9815, -46.1036, -46.2139, -46.2561, -46.3782, -46.4885, -46.6199,
        -46.9085, -47.0517, -47.1268, -47.1949, -47.3591, -47.3803, -47.392, -47.4366, -47.4483, -47.4248, -47.3803, -47.5351, -47.69, -47.7464, -47.8121, -47.8989, -47.9435, -47.9764, -48.1312,
        -48.0538, -47.8778, -47.7464, -47.6243, -47.5915, -47.568, -47.6361, -47.7018, -47.8778, -47.9998, -48.0209, -48.0421, -48.0773, -48.1195, -48.1852, -48.2861, -48.3964, -48.4739, -48.5936,
        -48.5724, -48.5067, -48.5584, -48.6945, -48.779, -48.8353, -48.901, -48.9667, -48.6804, -48.2955, -48.1735, -48.3612, -48.4833, -48.793, -48.9573, -49.023, -49.1779, -49.2343, -48.9808,
        -48.8916, -49.023, -49.0137, -49.1779, -49.3, -49.3328, -49.1873, -48.9151, -48.9339, -49.0137, -49.1873, -49.3422, -49.422, -49.4971, -49.5769, -49.5957, -49.6614, -49.5065, -49.3751,
        -49.5534, -49.6191, -49.4549, -49.5628, -49.6097, -49.8069, -49.8069, -49.7506, -49.8961],
   y = [63.0008, 63.0055, 63.0102, 62.975, 62.9186, 62.9093, 62.914, 62.8741, 62.8506, 62.8342, 62.8295, 62.8248, 62.8201, 62.8201, 62.813, 62.813, 62.8342, 62.8248, 62.799, 62.7426, 62.7145, 62.6934,
        62.6887, 62.6722, 62.6417, 62.6183, 62.5619, 62.4892, 62.454, 62.4188, 62.3108, 62.1583, 62.055, 62.0128, 61.9447, 61.8508, 61.8039, 61.7523, 61.6795, 61.6068, 61.5598, 61.48, 61.48, 61.3909,
        61.3228, 61.2853, 61.2641, 61.2102, 61.189, 61.1632, 61.067, 61.0459, 61.0412, 60.9825, 60.9543, 60.9215, 60.8581, 60.8323, 60.7995, 60.7455, 60.7032, 60.6258, 60.5929, 60.5507, 60.546,
        60.5554, 60.5178, 60.5131, 60.5343, 60.5507, 60.5296, 60.4897, 60.4357, 60.431, 60.3489, 60.2738, 60.2245, 60.201, 60.1635, 60.1423, 60.147, 60.147, 60.1423, 60.1353, 60.1423, 60.1517, 60.1588,
        60.2503, 60.2902, 60.201, 60.1635, 60.0696, 60.0367, 60.1259, 60.1142, 60.1142, 60.1142, 60.1306, 60.1916, 60.1916, 60.1916, 60.2456, 60.3324, 60.37, 60.4099, 60.4521, 60.4686, 60.5178,
        60.5671, 60.485, 60.4474, 60.4803, 60.5507, 60.6047, 60.6704, 60.6704, 60.6375, 60.5343, 60.4803, 60.5343, 60.5835, 60.6586, 60.6798, 60.6798, 60.6962, 60.7032, 60.7173, 60.7666, 60.8464,
        60.9379, 60.9661, 60.8699, 60.7948, 60.7995, 60.7455, 60.7408, 60.7572, 60.7948, 60.8323, 60.8323, 60.9004, 60.9497, 60.9919, 60.945, 60.9661, 61.0201, 61.0764, 61.0623, 61.0459, 61.0201,
        61.0553, 61.0294, 60.9919, 60.945, 60.959, 60.9661, 60.9543, 60.945, 60.9708, 60.9755, 60.9755, 60.9708, 60.945, 60.9004, 60.8581, 60.8417, 60.8417, 60.8417, 60.8417, 60.8464, 60.8581, 60.8534,
        60.8323, 60.8839, 60.9285, 60.9825, 61.0201, 61.0294, 61.0553, 61.0834, 61.0717, 61.121, 61.1257, 61.1304, 61.1351, 61.1468, 61.2148, 61.2524, 61.2383, 61.2102, 61.2219, 61.2688, 61.3439,
        61.3791, 61.3956, 61.4002, 61.3956, 61.3909, 61.3744, 61.4002, 61.4542, 61.5059, 61.5317, 61.581, 61.581, 61.4965, 61.4707, 61.48, 61.527, 61.5645, 61.6162, 61.6842, 61.7006, 61.7429, 61.7898,
        61.8673, 61.9189, 61.9752, 62.048, 62.1113, 62.1372, 62.0855, 62.0222, 62.0222, 62.0034, 62.0034, 62.055, 62.0949, 62.116, 62.1583, 62.1418, 62.2029, 62.2498, 62.2498, 62.2498, 62.285, 62.3625,
        62.4282, 62.4751] };
  WCL[786] = struct {
   x = [-43.1489, -43.0926, -43.1043, -43.2029, -43.3672, -43.4681, -43.5666, -43.7098, -43.8647, -43.9515, -43.9187],
   y = [60.0649, 60.0696, 60.0696, 60.0485, 60.0696, 60.086, 60.0978, 60.1095, 60.1095, 60.0813, 60.0203] };
  WCL[787] = struct {
   x = [-43.5127, -43.5009, -43.4563, -43.3789, -43.2029, -43.1489],
   y = [59.971, 60.0086, 60.0485, 60.0485, 60.0367, 60.0649] };
  WCL[788] = struct {
   x = [-50.6869, -50.7198, -50.7198, -50.809, -50.9216, -51.0624, -51.1188, -51.2173, -51.2737, -51.452, -51.2737, -51.4614, -51.4145, -51.4943, -51.5177, -51.4708, -51.1939, -50.978, -50.9216, -50.6869,
        -50.4804, -50.18, -50.5133, -50.7574, -50.7433, -50.5696, -50.3021, -50.0251, -50.1472, -50.349, -50.6869, -50.8324, -51.0765, -51.2173, -51.4849, -51.6726, -51.9355, -52.0434, -52.1091,
        -52.1091, -52.1091, -52.0763, -52.0904, -52.0434, -51.8228, -51.9355, -51.9355, -52.0575, -52.1091, -52.1326, -52.2875, -52.2781, -52.2452, -52.1795, -52.2218, -52.4095, -52.4846, -52.3767,
        -52.1795, -52.3861, -52.4001, -52.5081, -52.6724, -52.6864, -52.7052, -52.8601, -53.1136, -52.9962, -53.0385, -52.9258, -52.8179, -53.1136, -53.137, -53.3905, -53.3811, -53.2262, -53.0807,
        -53.0713, -52.9164, -53.1136, -53.3483, -53.4891, -53.2919, -53.5219, -53.5548, -53.5876, -53.4891, -53.137, -53.2497, -53.3905, -53.1793, -53.0479, -53.0385, -53.015, -53.3342, -53.6017,
        -53.5548, -53.7331, -53.888, -53.8739, -53.8739, -53.7895, -53.6768, -53.7331, -53.3483, -53.4562, -53.644, -53.4562, -53.5031, -53.4797, -53.5125, -53.3905, -53.2497, -53.015, -53.1934,
        -52.893, -52.785, -53.0713, -53.2356, -53.3905, -53.2497, -52.8507, -52.8507, -52.6067, -52.3767, -52.3532, -52.1795, -52.0012, -51.8557, -51.56, -51.2502, -51.0531, -50.9404, -51.2502,
        -51.3159, -51.0765, -51.3159, -51.5036, -51.2314, -51.1845, -51.1422, -51.1939, -51.5506, -51.4614, -51.3722, -51.6163, -52.0106, -52.3532, -52.4658, -52.5738, -52.8085, -52.8601, -52.8273,
        -52.893, -52.6067, -52.5409, -52.1232, -51.8698, -51.56, -51.161, -51.0531, -50.9545, -51.1939, -51.2643, -51.0437, -51.1751, -51.1751, -51.161, -51.1751, -51.1422, -51.0108, -50.7667,
        -50.6447, -50.4006, -50.5696, -50.2598, -50.3912, -50.7104, -50.978, -50.9639, -50.9639, -50.7104, -50.7667, -50.8324, -50.5884, -50.6541, -50.3349, -50.4335],
   y = [63.0008, 63.036, 63.0688, 63.0806, 63.1698, 63.2777, 63.4467, 63.4256, 63.496, 63.6274, 63.7494, 63.7213, 63.9231, 64.0357, 64.0874, 64.1648, 64.1695, 64.2329, 64.3573, 64.3807, 64.4136, 64.4699,
        64.4699, 64.5497, 64.6178, 64.6882, 64.6553, 64.6365, 64.7539, 64.7351, 64.8149, 64.6882, 64.6553, 64.6788, 64.4605, 64.2704, 64.2751, 64.2423, 64.3573, 64.423, 64.4558, 64.545, 64.6506,
        64.7163, 64.8196, 64.8431, 64.8712, 64.8055, 64.9064, 65.0003, 65.0331, 65.0895, 65.1528, 65.1857, 65.3007, 65.1904, 65.296, 65.3335, 65.4251, 65.4251, 65.4978, 65.4884, 65.4931, 65.6222,
        65.7255, 65.6856, 65.6574, 65.7442, 65.7935, 65.8663, 65.9015, 65.8381, 65.8968, 65.9367, 66.0094, 66.0611, 66.1244, 66.1197, 66.1784, 66.1479, 66.1502, 66.1784, 66.2746, 66.2746, 66.3427,
        66.4436, 66.5304, 66.5046, 66.5562, 66.6149, 66.7135, 66.7487, 66.8261, 66.8965, 66.9059, 66.8871, 66.9787, 67.035, 67.0772, 67.1875, 67.2509, 67.319, 67.3542, 67.4269, 67.5396, 67.5091,
        67.5677, 67.6663, 67.6733, 67.7813, 67.8775, 67.9104, 67.9432, 67.9479, 68.0254, 67.9643, 67.9925, 68.1122, 68.1075, 68.1122, 68.1544, 68.1075, 68.1404, 68.1732, 68.1333, 68.1333, 68.0911,
        68.0582, 68.0394, 68.0512, 68.0958, 68.0582, 68.0958, 68.1568, 68.2155, 68.1943, 68.2342, 68.2765, 68.3375, 68.3657, 68.4149, 68.4337, 68.4431, 68.3844, 68.3258, 68.2952, 68.2647, 68.2155,
        68.2507, 68.2718, 68.2272, 68.3046, 68.3117, 68.3821, 68.49, 68.5511, 68.5863, 68.5792, 68.5346, 68.5792, 68.6989, 68.7552, 68.7435, 68.774, 68.8632, 68.9172, 68.9735, 68.9735, 69.0134,
        69.1119, 69.1589, 69.1354, 69.0509, 69.0603, 69.1307, 69.2011, 69.302, 69.2176, 69.2903, 69.349, 69.4499, 69.4921, 69.5226, 69.6189, 69.6611, 69.7714, 69.809, 69.9075] };
  WCL[789] = struct {
   x = [-50.3889, -50.9615],
   y = [70.0061, 69.9897] };
  WCL[790] = struct {
   x = [-53.191, -52.8578, -52.5527, -52.1537, -51.9096, -51.9096, -51.9096, -51.9425, -52.2194, -52.5621, -52.9141, -53.2896, -53.6088, -53.7871, -53.9091, -54.1485, -54.1391, -53.9091, -53.6745, -53.4773,
        -53.4679, -53.6181, -53.651, -53.8622, -53.8716, -54.0593, -54.4677, -54.8338, -54.8667, -54.6789, -54.3597, -54.5897, -54.8901, -54.8009, -54.5663, -54.2612, -54.524],
   y = [70.1774, 69.9897, 69.8348, 69.7949, 69.7151, 69.6681, 69.6071, 69.5414, 69.4311, 69.3842, 69.349, 69.3067, 69.2833, 69.2903, 69.3185, 69.3842, 69.4382, 69.4522, 69.4311, 69.4569, 69.5062, 69.464,
        69.5156, 69.5179, 69.5954, 69.5461, 69.5461, 69.5954, 69.6846, 69.7034, 69.6681, 69.7597, 69.8395, 69.8841, 69.9005, 69.8958, 69.9709] };
  WCL[791] = struct {
   x = [-49.9289, -50.1167, -50.1824, -50.1683, -50.0838, -50.0463, -50.1167, -50.1354, -50.2152, -50.2575, -50.2903, -50.3467, -50.3795, -50.3795, -50.3983, -50.5813, -50.6846],
   y = [62.4751, 62.4798, 62.5408, 62.6535, 62.7638, 62.799, 62.7896, 62.7849, 62.7426, 62.7239, 62.7943, 62.8342, 62.8647, 62.9351, 62.975, 62.9961, 63.0008] };
  WCL[792] = struct {
   x = [-68.8069, -69.0275, -68.8397, -68.7083],
   y = [70.0319, 69.938, 69.931, 69.931] };
  WCL[793] = struct {
   x = [-67.4504, -67.1875, -67.0749, -67.2392, -67.6053, -67.8916, -68.0324, -68.2765, -68.5299, -68.652, -68.8491, -68.6942, -68.4642, -68.3187, -68.1873, -68.1028, -67.793, -67.4175, -67.042, -66.7463,
        -66.4882, -66.4882, -66.5774, -66.8543, -67.15, -67.4269, -67.6945, -68.0465, -68.023, -68.1685, -68.2671, -68.422, -68.5393, -68.5205, -68.253, -67.9479, -67.7461, -67.793, -68.1216, -68.1544,
        -68.0136, -67.8353, -67.8587, -67.6945, -68.0136, -68.023, -68.3328, -68.2342, -67.9995, -67.6381, -67.3612, -67.1641, -66.8637, -66.69, -66.7557, -66.8778, -66.7792, -66.5351, -66.1925,
        -66.047, -66.2253, -66.2582, -66.0611, -65.9484, -65.709, -65.3195, -64.9674, -64.9815, -65.066, -65.1129, -65.0331, -64.836, -64.6248, -64.4605, -64.362, -64.2164, -64.1179, -63.9207,
        -63.9207, -64.1507, -64.2399, -64.0287, -64.1836, -63.9301, -63.5922, -63.3575, -63.2355, -63.029, -63.0055, -63.1932, -63.2496, -63.3012, -63.0806, -62.8928, -62.5971, -62.4, -62.1794,
        -62.0339, -61.8368, -61.7053, -61.4941, -61.3627, -61.4847, -61.6584, -61.7241, -61.7804, -61.5927, -61.6913, -61.9353, -62.3671, -62.5971, -62.3014, -62.3437, -62.5643, -62.6065, -62.2216,
        -62.0808, -62.2545, -62.3859, -62.3859, -62.4, -62.6065, -62.6863, -62.7192, -62.9961, -63.2355, -63.4232, -63.4373, -63.3481, -63.2824, -63.2261, -63.381, -63.4467, -63.4702, -63.5124,
        -63.5781, -63.7236, -63.8222, -63.8879, -64.1085, -64.2399, -64.4183, -64.5497, -64.7468, -64.606, -64.6577, -64.9346, -65.066, -65.1552, -65.2772, -65.3429, -65.4227, -65.3429, -65.5729,
        -65.8827, -65.9484, -66.169, -66.3896, -66.6994, -66.9763, -66.8778, -66.8965, -67.1171, -67.3284, -67.4598, -67.1969, -67.3612, -67.2063, -67.3377, -67.5818, -67.7461, -67.9151, -67.8353,
        -67.9338, -68.1544, -68.0324, -67.901, -67.549, -67.272, -67.15, -67.3612, -67.5067, -67.6616, -67.7696, -67.9151, -67.8587, -67.9338, -67.9151, -67.5583, -67.272, -67.3612, -67.3284, -67.1641,
        -67.272, -67.1641, -67.2392, -67.2533, -66.9998, -66.9435, -66.9294, -66.7886, -66.6994, -66.6665, -66.6337, -66.4882, -66.2676, -66.1174, -66.338, -66.1925, -66.0704, -65.9954, -65.8968,
        -65.6762, -65.6433, -65.6292, -65.4884, -65.3523, -65.2772, -65.235, -65.2444, -65.0144, -64.9581, -65.0331, -65.2538, -65.235, -65.1129, -64.8689, -64.6248, -64.5919, -64.5591, -64.5403,
        -64.4371, -64.484, -64.4934, -64.5732, -64.5168, -64.5262, -64.7468, -64.9581, -64.9674, -65.0003, -64.9017, -64.8266, -64.7938, -64.606],
   y = [70.0319, 69.9122, 69.816, 69.7104, 69.7714, 69.7479, 69.6728, 69.6259, 69.6447, 69.6376, 69.5297, 69.5766, 69.5226, 69.4687, 69.41, 69.4921, 69.4311, 69.4405, 69.4217, 69.349, 69.2645, 69.2011,
        69.1471, 69.1847, 69.1378, 69.1354, 69.1824, 69.2715, 69.1894, 69.1706, 69.1964, 69.1894, 69.1706, 69.1143, 69.1237, 69.0509, 68.9805, 68.9336, 68.9453, 68.8773, 68.8256, 68.7599, 68.7224,
        68.6754, 68.6684, 68.6074, 68.6003, 68.5417, 68.5065, 68.4736, 68.4384, 68.4384, 68.4102, 68.368, 68.2929, 68.2155, 68.1709, 68.1732, 68.2108, 68.1005, 68.0559, 67.9878, 67.9597, 67.8822,
        67.9808, 67.9925, 68.0301, 67.9315, 67.8892, 67.8188, 67.7743, 67.786, 67.8071, 67.7696, 67.732, 67.6452, 67.5677, 67.3964, 67.3284, 67.3166, 67.2744, 67.2509, 67.1969, 67.1875, 67.2087,
        67.2274, 67.2744, 67.265, 67.1922, 67.0984, 66.9646, 66.8754, 66.913, 66.9622, 66.974, 66.9482, 66.9646, 66.9059, 66.8707, 66.8097, 66.7229, 66.6055, 66.5187, 66.4952, 66.4178, 66.3826,
        66.2981, 66.2324, 66.2347, 66.3427, 66.2981, 66.2629, 66.1549, 66.169, 66.0704, 66.0939, 66.0047, 65.9719, 65.9601, 65.8475, 65.7747, 65.7395, 65.6574, 65.5706, 65.5847, 65.6715, 65.7442,
        65.6433, 65.526, 65.4298, 65.2913, 65.2819, 65.1716, 65.0754, 65.0003, 64.8994, 64.9393, 65.0285, 65.1223, 65.0519, 65.2373, 65.1857, 65.1622, 65.2866, 65.4063, 65.4837, 65.4204, 65.5072,
        65.5565, 65.6222, 65.6856, 65.8569, 65.9813, 65.9414, 65.9461, 66.0986, 66.1385, 66.2042, 66.2981, 66.3192, 66.3896, 66.514, 66.5257, 66.5961, 66.5257, 66.4342, 66.4389, 66.2981, 66.2981,
        66.399, 66.46, 66.453, 66.3732, 66.3732, 66.3826, 66.2488, 66.2089, 66.1737, 66.0704, 65.9461, 65.9367, 65.9015, 65.8733, 65.8968, 65.8733, 65.7395, 65.6128, 65.5753, 65.648, 65.6081, 65.5401,
        65.472, 65.3969, 65.3758, 65.3148, 65.2608, 65.2185, 65.1857, 65.1036, 65.0519, 64.9487, 64.8008, 64.8431, 64.9346, 64.9534, 64.9252, 64.9346, 64.8055, 64.7633, 64.7961, 64.8337, 64.8665,
        64.8618, 64.768, 64.5966, 64.6647, 64.7116, 64.6835, 64.6225, 64.5544, 64.5826, 64.4652, 64.3268, 64.2986, 64.1507, 64.0968, 64.0357, 64.0264, 63.9278, 63.8081, 63.794, 63.7048, 63.6227,
        63.4655, 63.4373, 63.3223, 63.2824, 63.3927, 63.4655, 63.327, 63.2636, 63.1698, 63.0759, 63.0149, 63.0243] };
  WCL[794] = struct {
   x = [-67.6616, -67.6616, -67.8447, -67.9338, -67.9995, -68.0653, -68.1357, -68.253, -68.3563, -68.5205, -68.6942, -68.896, -68.6614, -68.5205, -68.3563, -68.2436, -67.9995, -67.7461, -67.8118, -67.6475,
        -67.4598, -67.2392, -67.0514, -66.8449, -66.7229],
   y = [63.0102, 63.1087, 63.09, 63.1932, 63.2237, 63.205, 63.2284, 63.3082, 63.3575, 63.4749, 63.5593, 63.7307, 63.7635, 63.7048, 63.6626, 63.5922, 63.4749, 63.4561, 63.5499, 63.5734, 63.442, 63.2824,
        63.2284, 63.2237, 63.0407] };
  WCL[795] = struct {
   x = [-70.1258, -69.9944, -69.8301, -69.7526, -69.7738, -69.6517, -69.5414, -69.4429, -69.4311, -69.4546, -69.5414, -69.6963, -69.6963, -69.7292, -69.6963, -69.708, -69.6306, -69.5977, -69.586, -69.4874,
        -69.4217, -69.532, -69.6752, -69.6963, -69.762, -69.9826],
   y = [60.8746, 60.9121, 60.9379, 60.959, 61.0248, 61.0623, 60.9708, 60.9074, 60.8534, 60.8088, 60.7455, 60.6868, 60.6164, 60.5296, 60.4521, 60.3981, 60.3864, 60.3231, 60.2503, 60.1635, 60.0931, 60.0766,
        60.0766, 60.0367, 60.0039, 60.0039] };
  WCL[796] = struct {
   x = [-65.3312, -65.2209, -65.0777, -64.9557, -64.9346, -64.944, -64.9111, -64.8571, -64.7797, -64.6905, -64.5262, -64.4699, -64.5474, -64.4042, -64.4159, -64.3713],
   y = [59.9875, 60.0696, 60.1095, 60.1752, 60.1916, 60.2081, 60.2292, 60.3066, 60.316, 60.316, 60.2996, 60.2456, 60.1963, 60.1799, 60.1024, 60.0367] };
  WCL[797] = struct {
   x = [-68.2202, -68.2741, -68.3305, -68.2741, -68.0981, -67.955, -67.8329, -67.8658, -68.0653, -68.1638, -68.2202],
   y = [60.485, 60.5178, 60.316, 60.2832, 60.3559, 60.377, 60.4686, 60.5929, 60.5929, 60.5554, 60.485] };
  WCL[798] = struct {
   x = [-65.0777, -65.0543, -65.0449, -64.9557, -64.89, -64.782, -64.7351, -64.7586, -64.782, -64.9041, -65.1106, -65.1435, -65.2326, -65.3641, -65.5283, -65.5096, -65.3312, -65.1904, -65.0777],
   y = [61.7053, 61.7265, 61.71, 61.6842, 61.6162, 61.6162, 61.5903, 61.5129, 61.3862, 61.3533, 61.3956, 61.4753, 61.4707, 61.5692, 61.5528, 61.642, 61.7006, 61.71, 61.7053] };
  WCL[799] = struct {
   x = [-64.606, -64.5262, -64.5262, -64.4464, -64.4136, -64.4699, -64.6577, -64.7797, -64.9346, -65.0097, -64.9768, -64.9111, -64.8454, -64.8219, -64.7797, -64.7562, -64.8876, -64.9768, -65.0895, -65.188,
        -65.174, -65.1082, -65.1082, -65.2631, -65.3054, -65.3758, -65.3946, -65.4603, -65.5401, -65.5729, -65.7278, -65.8498, -65.9015, -65.9156, -66.1549, -66.3896, -66.7229],
   y = [63.0243, 63.0055, 62.975, 62.9609, 62.9093, 62.8436, 62.86, 62.9351, 62.9492, 62.8905, 62.8295, 62.7591, 62.6887, 62.6276, 62.5971, 62.5666, 62.5455, 62.6276, 62.6582, 62.6769, 62.7239, 62.799,
        62.8506, 62.813, 62.8436, 62.9046, 62.9609, 62.9797, 62.9351, 62.9046, 62.9351, 62.9797, 62.9186, 62.8999, 63.036, 63.0383, 63.0407] };
  WCL[800] = struct {
   x = [-67.5349, -67.4128, -67.3706, -67.3377, -67.1922, -67.028, -66.8731, -66.7651, -66.6196, -66.4553, -66.3333, -66.1362, -66.0798, -66.1033, -66.1784, -66.1784, -66.0564, -66.0235, -66.3662, -66.4553,
        -66.6196, -66.8402, -67.0608, -67.3143, -67.4128, -67.6006, -67.6569, -67.8306, -68.0512, -68.2202, -68.361, -68.5581, -68.6379, -68.7599, -68.7599, -68.8444, -69.0462, -69.2199, -69.234,
        -69.3091, -69.4311, -69.4874, -69.464, -69.4405, -69.5532, -69.7268, -69.7972, -69.9474, -69.9146, -69.9474],
   y = [63.0008, 62.9398, 62.8788, 62.8436, 62.7943, 62.7192, 62.6934, 62.6183, 62.5666, 62.5056, 62.4329, 62.3272, 62.2498, 62.1465, 62.0949, 62.0386, 61.9705, 61.8978, 61.8767, 61.9306, 62.0034, 62.1207,
        62.048, 62.0433, 62.0597, 62.1372, 62.2029, 62.2076, 62.2076, 62.2709, 62.2498, 62.2897, 62.3531, 62.3789, 62.4845, 62.3836, 62.4141, 62.5267, 62.6112, 62.6464, 62.6464, 62.7332, 62.7685,
        62.8201, 62.7896, 62.8083, 62.7426, 62.7778, 62.8905, 62.9304] };
  WCL[801] = struct {
   x = [-76.3848, -76.3191, -76.2862, -76.1642, -75.9858, -75.7652, -75.6807, -75.5258, -75.3146, -75.1738, -75.1504, -75.1738, -75.1081, -75.1832, -75.4601, -75.7324, -76.0891, -76.4083, -76.6382, -76.7509,
        -77.0372, -77.2343, -77.3235, -77.3423, -77.3094, -77.2015, -77.1029, -76.948, -76.7603, -76.5397, -76.3848],
   y = [68.2929, 68.3211, 68.2882, 68.2436, 68.2999, 68.2882, 68.2061, 68.1943, 68.2061, 68.0958, 67.9855, 67.8188, 67.6522, 67.502, 67.3659, 67.2744, 67.2486, 67.2227, 67.2063, 67.2345, 67.265, 67.4175,
        67.5677, 67.7038, 67.8353, 67.9033, 68.0089, 68.1333, 68.2342, 68.2554, 68.2929] };
  WCL[802] = struct {
   x = [-74.8077, -74.7185, -74.5636, -74.3665, -74.3102, -74.0708, -73.9253, -73.7282, -73.5639, -73.761, -74.1365, -74.4792, -74.7655, -74.8077],
   y = [67.9925, 68.0559, 68.0676, 68.1451, 68.0582, 68.023, 67.9855, 67.969, 67.7954, 67.7273, 67.7789, 67.7696, 67.8728, 67.9925] };
  WCL[803] = struct {
   x = [-78.5791, -78.5791, -78.5979, -78.687, -78.7668, -78.8513, -79.0391, -79.2409, -79.4615, -79.4474, -79.3817, -79.2597, -79.0391, -78.9217, -78.7997, -78.6542, -78.5228, -78.4101, -78.3585, -78.3585,
        -78.3585, -78.443, -78.5462, -78.6448, -78.7105, -78.7199, -78.7199, -78.5791],
   y = [69.2082, 69.1894, 69.1542, 69.1073, 69.0486, 68.9617, 68.9054, 68.882, 68.8773, 68.9547, 68.9923, 69.0721, 69.1143, 69.1777, 69.2481, 69.2316, 69.2058, 69.2058, 69.241, 69.295, 69.3419, 69.3607,
        69.3654, 69.3537, 69.3255, 69.2786, 69.2598, 69.2082] };
  WCL[804] = struct {
   x = [-80.067, -79.9684, -79.837, -79.7243, -79.5601, -79.438, -79.4803, -79.7009, -79.8886],
   y = [69.7221, 69.7151, 69.7667, 69.7831, 69.7667, 69.7221, 69.6376, 69.6118, 69.5766] };
  WCL[805] = struct {
   x = [-78.5016, -78.5673, -78.5251, -78.4359, -78.2482, -78.0182, -77.9196, -77.7741, -77.6333, -77.5676, -77.7178, -77.8633, -78.1168, -78.3608, -78.5016],
   y = [63.3974, 63.3974, 63.4209, 63.4467, 63.4702, 63.4561, 63.4209, 63.4162, 63.3434, 63.2144, 63.1393, 63.0853, 63.2003, 63.3176, 63.3974] };
  WCL[806] = struct {
   x = [-77.4549, -77.2907, -77.056, -76.8166, -76.7368, -76.7274, -76.704, -76.9011, -77.1217, -77.3423, -77.3986, -77.4549],
   y = [63.6955, 63.6579, 63.6321, 63.5687, 63.5194, 63.4749, 63.4068, 63.4068, 63.4608, 63.5546, 63.6274, 63.6955] };
  WCL[807] = struct {
   x = [-71.2781, -71.3015, -71.4001, -71.4987, -71.6864, -71.8319, -71.8742, -71.9868, -72.0854, -71.8742, -71.5972, -71.3438, -71.555, -71.6113, -71.7334, -71.9399, -72.0525, -72.0619, -72.2074, -72.306,
        -72.306, -72.3952, -72.5266, -72.6252, -72.7003, -72.7566, -72.8458, -72.9772, -73.01, -73.0523, -73.0335, -73.0758, -73.1743, -73.287, -73.3949, -73.4606, -73.4841, -73.4278, -73.5263,
        -73.6625, -73.7516, -73.8033, -73.9722, -74.038, -74.2116, -74.3102, -74.4651, -74.4979, -74.6341, -74.6669, -74.6669, -74.7655, -74.972, -74.9626, -74.7185, -74.5543, -74.5871, -74.7185,
        -74.6998, -74.8734, -74.9626, -75.3287, -75.3287, -75.3803, -75.4789, -75.6667, -75.7324, -75.9671, -75.8779, -75.9013, -76.0234, -76.3519, -76.2956, -76.5631, -76.704, -76.7603, -76.8589,
        -77.2343, -77.3798, -77.6098, -77.7741, -77.9947, -78.1731, -78.1825, -78.1496, -77.9947, -77.6521, -77.4315, -77.4784, -77.2907, -77.4315, -77.225, -77.0888, -76.8682, -76.6289, -76.244,
        -76.0093, -75.8544, -75.723, -75.3944, -75.2255, -75.1926, -75.0612, -74.742, -74.6106, -74.4557, -74.3008, -74.2445, -73.991, -73.9816, -73.639, -73.639, -73.8033, -73.8831, -74.2022,
        -74.4322, -74.4792, -74.3337, -74.1131, -73.869, -73.6484, -73.5639, -73.3527, -73.24, -73.0898, -73.0194, -73.0194, -73.01, -73.01, -72.8551, -72.6111, -72.4609, -72.3717, -72.4937, -72.5688,
        -72.5782, -72.7143, -72.7566, -72.888, -73.0007, -73.0429, -73.1978, -73.3198, -73.5404, -73.7047, -73.8502, -73.9582, -73.9488, -73.9253, -74.2914, -74.3008, -74.1788, -74.2773, -74.5308,
        -74.7655, -74.9532, -74.8312, -74.9298, -74.62, -74.343, -74.7091, -75.0706, -75.3803, -75.6338, -75.7558, -76.3097, -76.6711, -76.6382, -76.7368, -76.474, -76.0985, -75.7324, -75.8356,
        -76.2064, -76.5068, -76.7274, -76.9386, -77.1592, -76.9386, -76.9011, -76.9903, -77.2109, -77.3892, -77.4643, -77.53, -77.6849, -77.7084],
   y = [63.0008, 63.029, 63.0642, 63.1252, 63.1533, 63.2683, 63.3317, 63.4209, 63.5499, 63.4373, 63.4795, 63.6133, 63.6133, 63.6814, 63.7095, 63.8456, 63.7494, 63.6767, 63.7001, 63.7729, 63.8363, 63.8574,
        63.8081, 63.8269, 63.8949, 63.9583, 63.9818, 63.9583, 64.0357, 64.0639, 64.1273, 64.1977, 64.2704, 64.3268, 64.3033, 64.3948, 64.4887, 64.5356, 64.6084, 64.6271, 64.5966, 64.5779, 64.6835,
        64.6929, 64.6694, 64.6318, 64.6929, 64.7633, 64.6976, 64.7726, 64.8665, 64.8196, 64.7914, 64.6835, 64.6459, 64.5075, 64.5497, 64.4934, 64.4089, 64.4371, 64.4417, 64.4511, 64.4558, 64.5262,
        64.5966, 64.6013, 64.5966, 64.5591, 64.4746, 64.3948, 64.3713, 64.3573, 64.2892, 64.261, 64.2376, 64.2071, 64.2517, 64.2986, 64.2798, 64.3995, 64.3948, 64.4887, 64.6506, 64.768, 64.9041,
        65.0191, 65.1364, 65.1669, 65.3523, 65.3476, 65.4345, 65.4438, 65.3922, 65.3664, 65.3664, 65.296, 65.3195, 65.3242, 65.3242, 65.2467, 65.2725, 65.3781, 65.3828, 65.3758, 65.411, 65.4532,
        65.472, 65.5307, 65.4744, 65.5166, 65.4485, 65.6292, 65.6574, 65.7841, 65.9109, 66.0704, 66.1549, 66.223, 66.2981, 66.4436, 66.5093, 66.5351, 66.6619, 66.6783, 66.6877, 66.7135, 66.7534,
        66.8449, 66.974, 67.042, 67.0866, 67.1664, 67.265, 67.2908, 67.4081, 67.5396, 67.6264, 67.7954, 67.8564, 67.9714, 68.1333, 68.2225, 68.2882, 68.2389, 68.2765, 68.3211, 68.4267, 68.5511,
        68.6707, 68.7106, 68.6426, 68.5581, 68.5182, 68.6144, 68.7271, 68.8116, 68.9031, 68.9383, 68.9664, 68.997, 69.0251, 68.9923, 68.8937, 68.9336, 68.8303, 68.6942, 68.652, 68.8655, 69.0322,
        69.0368, 68.997, 69.1143, 69.2598, 69.3795, 69.4757, 69.5696, 69.6142, 69.6259, 69.6846, 69.7667, 69.809, 69.8066, 69.8559, 69.7855, 69.7526, 69.809, 69.9427] };
  WCL[808] = struct {
   x = [-78.8865, -78.7879, -78.9241, -79.3277, -79.6844, -79.8252, -79.9238],
   y = [70.1117, 69.9498, 69.877, 69.87, 69.8254, 69.8137, 69.9005] };
  WCL[809] = struct {
   x = [-77.746, -77.746, -77.645, -77.5347, -77.5019, -77.6122, -77.645, -77.6802, -77.7553, -77.8117, -77.7553, -77.7225, -77.8328, -77.8445, -77.7342, -77.5019, -77.5136, -77.6896, -77.7905, -77.9337,
        -78.0651, -78.1965, -78.1965, -78.0111, -77.922, -77.7999, -77.8117, -77.7952, -77.7788, -77.7295, -77.7248, -77.6568, -77.6474, -77.6403, -77.6802, -77.7577, -77.8187, -77.8891, -77.9712,
        -77.9877, -78.0323, -78.044, -78.0393, -78.0276, -77.9783, -77.929, -77.8938, -77.8844, -77.8234, -77.7413, -77.6849, -77.6802, -77.5911, -77.5207, -77.4362, -77.3611, -77.2836, -77.2179,
        -77.1405, -77.0795, -77.0349, -77.0231, -76.941, -76.8518, -76.7814, -76.7204, -76.7485, -76.7251, -76.6336, -76.5608, -76.4834, -76.4458, -76.4716, -76.4223, -76.3238, -76.2956, -76.312,
        -76.2628, -76.1595, -76.0586, -75.96, -75.8873, -76.1032, -76.1079, -76.0492, -75.9154, -75.845, -75.7723, -75.6338, -75.5399, -75.4648, -75.4038, -75.3311, -75.2442, -75.2161, -75.209,
        -75.2372, -75.2865, -75.3193, -75.2865, -75.2208, -75.1926, -75.1926, -75.1386, -75.0565, -75.0565, -75.0565, -74.9908, -74.9673, -75.0447, -75.0283, -74.9391, -74.9133, -74.8406, -74.7631,
        -74.7091, -74.7467, -74.796, -74.8523, -74.8453, -74.8523, -74.796, -74.7068, -74.5519, -74.5965, -74.5848, -74.4322, -74.2539, -74.0685, -74.0121, -73.9136, -73.8572, -73.8033, -73.7024,
        -73.6366, -73.517, -73.3949, -73.2846, -73.1954, -73.1626, -73.1743, -73.0851],
   y = [59.9828, 60.0649, 60.0414, 60.0766, 60.1588, 60.2409, 60.2456, 60.2832, 60.3489, 60.4099, 60.5131, 60.5835, 60.6047, 60.6704, 60.6798, 60.7173, 60.7337, 60.7337, 60.7713, 60.7337, 60.7502, 60.8253,
        60.9332, 61.0717, 61.2313, 61.2806, 61.3909, 61.4073, 61.4073, 61.4753, 61.5598, 61.5739, 61.649, 61.7382, 61.6748, 61.649, 61.6959, 61.7171, 61.7945, 61.8978, 61.9823, 62.1066, 62.2146,
        62.2545, 62.2803, 62.3108, 62.3625, 62.393, 62.4798, 62.5619, 62.5877, 62.5197, 62.5361, 62.5361, 62.5267, 62.5267, 62.515, 62.4892, 62.4657, 62.4587, 62.4235, 62.3977, 62.3625, 62.3484,
        62.3413, 62.3155, 62.2404, 62.224, 62.2803, 62.2897, 62.285, 62.2334, 62.1724, 62.1771, 62.1677, 62.1771, 62.224, 62.2709, 62.3108, 62.3366, 62.3718, 62.4235, 62.3226, 62.3272, 62.3484,
        62.3977, 62.4751, 62.4493, 62.3625, 62.3272, 62.2193, 62.1982, 62.1677, 62.1583, 62.1113, 62.0386, 61.987, 61.8837, 61.8109, 61.8415, 61.8203, 61.8579, 61.9494, 61.9564, 61.9236, 61.8579,
        61.7945, 61.7265, 61.6631, 61.6162, 61.5434, 61.5739, 61.642, 61.6584, 61.642, 61.5903, 61.5223, 61.5059, 61.4589, 61.3909, 61.3697, 61.3228, 61.2641, 61.243, 61.1843, 61.1515, 61.1726,
        61.1796, 61.1632, 61.0717, 60.9966, 61.0294, 61.0623, 61.0459, 61.0412, 61.0201, 61.0459, 61.0717, 61.013, 60.9543, 60.8839, 60.8746] };
  WCL[810] = struct {
   x = [-69.9826, -70.1375, -70.1821, -70.3041, -70.4802, -70.7665, -70.9753, -71.1631, -71.2616, -71.1631, -70.9096],
   y = [60.0039, 60.025, 60.025, 60.025, 60.0156, 60.0367, 60.0978, 60.086, 60.0532, 60.0485, 60.032] };
  WCL[811] = struct {
   x = [-80.0834, -79.9942, -79.8722, -79.7642, -79.7736, -79.6985, -79.5859, -79.4685, -79.3981, -79.3887, -79.553, -79.8628],
   y = [61.6701, 61.5645, 61.5528, 61.649, 61.7429, 61.7992, 61.9142, 62.0738, 62.2076, 62.2803, 62.3413, 62.3625] };
  WCL[812] = struct {
   x = [-69.9498, -70.1047, -70.2173, -70.2267, -70.2361, -70.2924, -70.4051, -70.513, -70.6257, -70.6585, -70.8134, -71.0011, -71.2781],
   y = [62.9304, 62.8248, 62.7943, 62.7896, 62.7731, 62.7685, 62.813, 62.8436, 62.8436, 62.8999, 62.9304, 62.9703, 63.0008] };
  WCL[813] = struct {
   x = [-71.0317, -70.9613, -70.8298, -70.6749, -70.5905, -70.5107, -70.4027, -70.445, -70.6421, -70.8298, -70.9753, -71.0645, -71.0317],
   y = [62.7849, 62.7896, 62.7544, 62.7426, 62.7286, 62.698, 62.6464, 62.5971, 62.5666, 62.6112, 62.684, 62.7591, 62.7849] };
  WCL[814] = struct {
   x = [-74.052, -74.0755, -74.0286, -74.0098, -74.1084, -74.2867, -74.4181, -74.6575, -74.7796, -74.6716, -74.5073, -74.329, -74.1506, -74.052],
   y = [62.6934, 62.7426, 62.7192, 62.6769, 62.6417, 62.6276, 62.637, 62.6582, 62.6934, 62.7286, 62.698, 62.7239, 62.7332, 62.6934] };
  WCL[815] = struct {
   x = [-90.0152, -89.9964, -89.9589, -89.8509, -89.8275, -89.8603, -89.8603, -89.6303, -89.4989, -89.2877, -89.0577, -89.0248, -88.9591, -88.7479, -88.4381, -88.2316, -88.1284, -88.0861, -88.0298, -87.8984,
        -87.8749, -88.0439, -88.2081, -88.2645, -88.2645, -88.349, -88.4381, -88.3865, -88.4522, -88.4522, -88.2504, -88.1518, -87.9969, -87.889, -87.7106, -87.5886, -87.4806, -87.3586, -87.1615,
        -87.1474, -87.0957, -86.9409, -86.8517, -86.786, -86.739, -86.6311, -86.5513, -86.5654, -86.5419, -86.4856, -86.3776, -86.157, -86.0584, -86.0115, -85.9458, -85.9458, -85.9458, -85.9364,
        -85.8238, -85.805, -85.7721, -85.7909, -85.6923, -85.4717, -85.4154, -85.3403, -85.2652, -85.1526, -84.9226, -84.7348, -84.824, -84.9085, -85.0634, -85.1526, -85.0775, -85.1197, -84.9554,
        -84.9742, -84.8897, -84.6785, -84.7207, -84.8569, -84.9977, -85.0869, -85.2417, -85.3826, -85.4623, -85.453, -85.4154, -85.4389, -85.4858, -85.453, -85.5187, -85.4858, -85.453, -85.453,
        -85.3497, -85.1103, -84.9554, -84.7771, -84.5565, -84.2138, -84.0918, -84.0402, -83.8196, -83.5849, -83.3784, -83.1578, -83.0123, -82.8151, -82.6696, -82.5851, -82.571, -82.5851, -82.5523,
        -82.7166, -82.571, -82.3833, -82.2425, -82.3176, -82.2753, -82.0313, -81.745, -81.5807, -81.4023, -81.3601, -81.4493, -81.5478, -81.6699, -81.8435, -81.9562, -81.6699, -81.5666, -81.3695,
        -81.2803, -81.2944, -81.3929, -81.6135, -81.7356, -81.8201, -81.8435, -81.8999, -82.0219, -82.0876, -82.2096, -82.2753, -82.5382, -82.5617, -82.6274, -82.4725, -82.3176, -82.4068, -82.4068,
        -82.2847, -82.2847, -82.3176, -82.1205, -81.975, -82.0782, -82.1205, -82.097, -81.867, -81.637, -81.4258, -81.3038, -81.2944, -81.3038, -81.3929, -81.3929, -81.3131, -81.4258, -81.4258,
        -81.5337, -81.6886, -81.8013, -82.0078, -82.0313, -82.1111, -82.1298, -82.2519, -82.4631, -82.6368, -82.7917, -82.97, -83.0357, -83.1437, -83.2329, -83.3643, -83.4441, -83.5004, -83.6881,
        -83.7867, -83.8853, -83.951, -84.0261, -84.1153, -84.1387, -84.1481, -84.1293, -83.9087, -83.8853, -83.7539, -83.7632, -83.7867, -83.9932, -84.073, -84.2373, -84.3922, -84.425, -84.5142,
        -84.5142, -84.5001, -84.6879, -84.8005, -84.932, -85.1197, -85.162, -85.2183, -85.3075, -85.3497, -85.5046, -85.6689, -85.7909, -85.8378, -85.8566, -85.9927, -86.1805, -86.3682, -86.4996,
        -86.5841, -86.6311, -86.6874, -86.6639, -86.739, -86.6874, -86.4339, -86.265, -86.1007, -86.0021, -86.035, -86.0444, -86.1007, -86.1007, -86.1007, -86.1664, -86.2978, -86.3682, -86.4339,
        -86.5419, -86.6733, -86.7062, -86.7625, -86.9268, -86.9831, -87.0488, -87.1286, -87.3163, -87.5229, -87.7576, -87.9875, -88.0439, -88.1518, -88.2504, -88.5508, -88.7057, -88.7151, -88.8465,
        -88.9685, -89.0248, -89.114, -89.3346, -89.4989, -89.6397, -89.7852],
   y = [68.6355, 68.6707, 68.6942, 68.6942, 68.8139, 68.9336, 69.0368, 69.1307, 69.2434, 69.2833, 69.2715, 69.2176, 69.1495, 69.0368, 68.9406, 68.8702, 68.7881, 68.6707, 68.5417, 68.4102, 68.2929, 68.2483,
        68.2647, 68.2952, 68.4102, 68.3445, 68.2554, 68.1333, 68.0301, 67.955, 67.8118, 67.7109, 67.6264, 67.5396, 67.4128, 67.3237, 67.2955, 67.2298, 67.2697, 67.3237, 67.4081, 67.4387, 67.4081,
        67.448, 67.4175, 67.4058, 67.556, 67.6945, 67.7414, 67.7954, 67.8775, 67.9714, 68.0676, 68.1333, 68.1943, 68.2507, 68.2812, 68.2929, 68.3375, 68.4454, 68.6027, 68.6543, 68.6989, 68.7083,
        68.7552, 68.7388, 68.7599, 68.774, 68.7482, 68.7623, 68.7951, 68.7998, 68.8116, 68.8538, 68.8937, 68.9289, 68.9289, 68.9805, 68.997, 68.9782, 69.0392, 69.0439, 69.1026, 69.119, 69.126, 69.1964,
        69.2833, 69.349, 69.3795, 69.4217, 69.4851, 69.5508, 69.6423, 69.6681, 69.7362, 69.8019, 69.8019, 69.7761, 69.8254, 69.8277, 69.8348, 69.8019, 69.755, 69.7339, 69.6846, 69.6541, 69.6564,
        69.6752, 69.6447, 69.6564, 69.6916, 69.6423, 69.6118, 69.5344, 69.4804, 69.4522, 69.4499, 69.3537, 69.2645, 69.2199, 69.2082, 69.2246, 69.2246, 69.1777, 69.1612, 69.1002, 69.0509, 69.004,
        68.9735, 68.9265, 68.8468, 68.835, 68.8303, 68.8139, 68.7318, 68.6308, 68.5792, 68.5182, 68.5159, 68.4501, 68.4454, 68.3891, 68.4337, 68.5276, 68.5112, 68.4666, 68.5159, 68.4736, 68.4619,
        68.3774, 68.3164, 68.2952, 68.2671, 68.26, 68.2014, 68.1497, 68.1896, 68.2155, 68.1052, 67.9878, 67.8939, 67.7531, 67.6358, 67.5607, 67.4762, 67.4293, 67.4222, 67.3917, 67.2861, 67.2181,
        67.1617, 67.1031, 67.0819, 67.0561, 67.0162, 66.9224, 66.8613, 66.8191, 66.7933, 66.7323, 66.6712, 66.5891, 66.5844, 66.5562, 66.4788, 66.4389, 66.3873, 66.3896, 66.4225, 66.4788, 66.5539,
        66.5891, 66.6196, 66.6712, 66.7064, 66.6665, 66.6619, 66.6501, 66.5797, 66.453, 66.3544, 66.223, 66.1784, 66.223, 66.2324, 66.2746, 66.338, 66.3685, 66.3896, 66.3896, 66.2535, 66.2136, 66.2488,
        66.2629, 66.2488, 66.3075, 66.2723, 66.3544, 66.4999, 66.5797, 66.5914, 66.5398, 66.4999, 66.4905, 66.514, 66.5046, 66.5398, 66.5609, 66.575, 66.5492, 66.5234, 66.5046, 66.46, 66.4436, 66.3732,
        66.3286, 66.2887, 66.2793, 66.2183, 66.1502, 66.0751, 66.0493, 66.0446, 66.0094, 65.9648, 65.9156, 65.8569, 65.7841, 65.7161, 65.6668, 65.6128, 65.6034, 65.5706, 65.4931, 65.4345, 65.3922,
        65.3758, 65.3711, 65.3711, 65.3523, 65.4157, 65.4837, 65.5119, 65.6339, 65.6856, 65.702, 65.7302, 65.7536, 65.756, 65.7654, 65.8475, 65.8921, 65.9062, 65.8733] };
  WCL[816] = struct {
   x = [-90.1795, -89.8275, -89.574, -89.3534, -89.2314, -89.0436, -88.8465, -88.5273, -88.2081, -88.0955, -87.9218, -87.7998, -87.4806, -87.2694, -87.0394, -87.0066, -86.908, -86.9737, -87.0488, -87.0488,
        -87.2366, -87.3163, -87.4478, -87.5229, -87.5792, -87.7106, -87.7998, -87.8233, -87.8233, -87.8233, -87.9077, -88.0298, -88.0955, -88.1518, -88.2316, -88.349, -88.4945, -88.6728, -88.8606,
        -89.0436, -89.114, -89.2548, -89.4332, -89.4848, -89.5083, -89.4989, -89.5552, -89.6069, -89.6867, -89.8181, -89.9073, -89.8509, -89.8509, -89.8509, -89.8603, -89.8415, -89.8087, -89.8181,
        -89.9073],
   y = [65.8193, 65.7395, 65.6175, 65.472, 65.4157, 65.3523, 65.3429, 65.3242, 65.2772, 65.2514, 65.2819, 65.2702, 65.2725, 65.2608, 65.2514, 65.2092, 65.2139, 65.1223, 65.0519, 64.9815, 64.89, 64.7445,
        64.7163, 64.6225, 64.5638, 64.5356, 64.5215, 64.5168, 64.484, 64.3995, 64.3455, 64.2235, 64.1742, 64.1883, 64.1507, 64.0827, 64.0498, 64.0123, 64.0357, 64.0498, 63.9912, 64.0451, 64.0827,
        64.1085, 64.078, 64.0357, 64.078, 64.1273, 64.1085, 64.1601, 64.1414, 64.1015, 64.0686, 64.0498, 64.031, 63.9958, 63.9583, 63.9442, 63.9442] };
  WCL[817] = struct {
   x = [-90.0692, -89.9706],
   y = [63.8902, 63.8316] };
  WCL[818] = struct {
   x = [-86.6522, -86.6287, -86.5818, -86.5489, -86.5395, -86.5161, -86.5959, -86.7273, -86.9808, -86.9808, -87.0042, -86.9808, -86.9573, -86.8165, -86.6944, -86.6616, -86.6522],
   y = [68.26, 68.2647, 68.1943, 68.1075, 67.9925, 67.8681, 67.7602, 67.7696, 67.8681, 67.9643, 68.0582, 68.1404, 68.2272, 68.3117, 68.2718, 68.26, 68.26] };
  WCL[819] = struct {
   x = [-85.8144, -85.7487, -85.7393, -85.6689, -85.5703, -85.5281, -85.3075, -85.2183, -85.0634, -85.1526, -85.0869, -85.0446, -84.8897, -84.8005, -84.6456, -84.5142, -84.303, -84.1481, -84.1481, -84.1716,
        -83.9745, -83.8055, -83.6881, -83.5192, -83.4535, -83.4206, -83.3549, -83.1906, -82.97, -82.7917, -82.571, -82.4068, -82.1862, -82.0219, -81.9984, -81.9421, -81.8341, -81.8201, -81.8013,
        -81.745, -81.6464, -81.7121, -81.5995, -81.4164, -81.2803, -81.2287, -81.1254, -80.994, -80.8297, -80.7405, -80.6185, -80.6091, -80.5997, -80.6326, -80.5669, -80.365, -80.2336, -80.2101,
        -80.3791, -80.534, -80.6748, -80.8203, -80.9189, -80.9846, -81.1254, -81.2709, -81.4023, -81.5009, -81.5009, -81.5244, -81.6464, -81.7215, -81.8905, -82.0313, -82.1205, -82.219, -82.4396,
        -82.4819, -82.3974, -82.4396, -82.571, -82.6368, -82.7917, -82.9372, -82.9559, -82.97, -82.9794, -83.1343, -83.0451, -82.9372, -82.97, -83.1437, -83.3549, -83.5426, -83.6318, -83.6177,
        -83.6506, -83.6741, -83.6318, -83.6506, -83.6975, -83.7867, -83.8524, -83.9275, -83.9838, -84.0261, -84.1293, -84.2795, -84.4157, -84.4814, -84.5565, -84.5471, -84.6034, -84.7348, -84.9414,
        -85.1197, -85.2746, -85.3309, -85.4717, -85.6172, -85.6595, -85.636, -85.6923, -85.7064, -85.7393, -85.805, -85.8144, -85.8378, -85.9693, -86.0913, -86.2321, -86.2978, -86.3776, -86.387,
        -86.4527, -86.5654, -86.6874, -86.8047, -86.9268, -87.0817, -87.1708, -87.2272, -87.1708, -87.0957, -87.016, -86.8751, -86.6405, -86.4762, -86.2884, -86.2978, -86.3682, -86.3776, -86.387,
        -86.4199, -86.387, -86.3776, -86.279, -86.2227, -86.2321, -86.1899, -86.1805, -86.1476, -86.1664, -86.0913, -86.0115, -85.9129, -85.8144],
   y = [65.7888, 65.8193, 65.8287, 65.8827, 65.8968, 65.81, 65.8193, 65.6926, 65.6081, 65.5354, 65.4345, 65.2866, 65.2608, 65.3711, 65.4626, 65.4532, 65.3711, 65.3523, 65.3054, 65.2232, 65.1669, 65.1481,
        65.1528, 65.1435, 65.1364, 65.0754, 65.0144, 64.9252, 64.89, 64.7961, 64.7492, 64.7633, 64.6929, 64.6647, 64.6506, 64.5779, 64.4981, 64.4464, 64.4042, 64.2517, 64.1789, 64.078, 64.0451,
        64.0733, 64.0686, 64.0357, 64.0874, 64.1507, 64.1085, 64.0357, 63.9958, 63.9489, 63.909, 63.8809, 63.8715, 63.8574, 63.8316, 63.7776, 63.7635, 63.6908, 63.6274, 63.5875, 63.5406, 63.5007,
        63.4749, 63.5007, 63.5288, 63.5781, 63.5687, 63.6039, 63.5734, 63.6462, 63.6814, 63.6861, 63.726, 63.6861, 63.7001, 63.7541, 63.8456, 63.9348, 63.9724, 63.9724, 63.9818, 64.0052, 64.0005,
        63.9912, 63.963, 64.0005, 64.1015, 64.1601, 64.1789, 64.1648, 64.146, 64.1015, 64.0921, 64.0404, 63.9395, 63.9043, 63.8621, 63.7682, 63.7682, 63.7635, 63.7095, 63.6955, 63.6908, 63.6368,
        63.5875, 63.5922, 63.5054, 63.4303, 63.442, 63.4209, 63.3528, 63.273, 63.2331, 63.2003, 63.1604, 63.1838, 63.1604, 63.3035, 63.4749, 63.6415, 63.6908, 63.8128, 63.8855, 63.8456, 63.7588,
        63.7166, 63.7095, 63.6673, 63.672, 63.6673, 63.672, 63.6814, 63.6908, 63.6814, 63.6415, 63.5992, 63.5734, 63.5593, 63.6086, 63.6908, 63.7588, 63.841, 63.9043, 63.9677, 64.0005, 64.078, 64.1742,
        64.2704, 64.3127, 64.3995, 64.5028, 64.5638, 64.6318, 64.6882, 64.8196, 64.9393, 64.9674, 65.0519, 65.1998, 65.3476, 65.4485, 65.526, 65.6715, 65.7442, 65.7888] };
  WCL[820] = struct {
   x = [-83.8947, -83.8712, -83.8055, -83.7398, -83.721, -83.6506, -83.552, -83.3643, -83.3455, -83.552, -83.7632, -83.8618, -84.0918, -84.1387, -84.2795, -84.4673, -84.4814, -84.3453, -84.1481, -83.951,
        -83.8947],
   y = [65.9648, 65.9461, 65.9273, 65.9226, 65.8569, 65.7654, 65.7114, 65.6809, 65.6222, 65.6246, 65.6433, 65.7161, 65.7982, 65.8616, 65.9766, 66.0704, 66.115, 66.115, 66.0798, 65.9907, 65.9648] };
  WCL[821] = struct {
   x = [-80.2547, -80.189, -80.1092, -80.067],
   y = [69.7855, 69.7479, 69.7292, 69.7221] };
  WCL[822] = struct {
   x = [-79.891, -80.0693, -80.0693, -80.0693, -80.2571, -80.3885, -80.4777, -80.534, -80.4307, -80.2571],
   y = [69.5766, 69.6306, 69.5414, 69.464, 69.5414, 69.6071, 69.7034, 69.7667, 69.7667, 69.7855] };
  WCL[823] = struct {
   x = [-79.9238, -80.0365, -80.1679],
   y = [69.9005, 69.9709, 69.9685] };
  WCL[824] = struct {
   x = [-80.2758, -80.5199],
   y = [70.0014, 69.9685] };
  WCL[825] = struct {
   x = [-81.7778, -81.4023, -81.0597, -80.9611, -81.1489, -81.3789, -81.6464, -81.8764, -82.0219, -82.1533, -82.1298, -82.1956, -82.219, -82.2753, -82.5523, -82.8574, -83.1906, -83.4675, -83.599, -83.6177,
        -83.6506, -83.9181, -84.3593, -84.8569],
   y = [70.0272, 69.9803, 69.8277, 69.7409, 69.7456, 69.8817, 69.938, 69.8958, 69.7972, 69.8395, 69.8559, 69.8207, 69.8066, 69.8348, 69.8465, 69.9193, 69.9897, 69.9451, 69.9216, 69.9216, 69.8817, 69.931,
        69.9568, 69.9803] };
  WCL[826] = struct {
   x = [-85.6266, -85.3638, -85.5844],
   y = [70.0272, 69.9709, 69.9709] };
  WCL[827] = struct {
   x = [-82.5405, -82.564, -82.8269, -83.1038, -83.3009, -83.4558, -83.5638, -83.7186, -83.8078, -83.8501, -83.7186, -83.5215, -83.4464, -83.2258, -83.0803, -82.9489, -82.8362, -82.8503, -82.7377, -82.5734,
        -82.4607, -82.2871, -82.1557, -82.1087, -82.043, -82.0571, -82.1885, -82.3434, -82.5405],
   y = [62.9304, 62.914, 62.8741, 62.813, 62.7332, 62.6183, 62.4751, 62.3413, 62.2545, 62.116, 62.0902, 62.1583, 62.1841, 62.1888, 62.2709, 62.2967, 62.3226, 62.4141, 62.5056, 62.5314, 62.5666, 62.637,
        62.6887, 62.7849, 62.8436, 62.8952, 62.9257, 62.9609, 62.9304] };
  WCL[828] = struct {
   x = [-80.0505, -80.0505, -80.074, -80.1397, -80.1632, -80.2477, -80.2946, -80.3275, -80.3603, -80.3603, -80.2054, -80.0834],
   y = [62.3578, 62.3413, 62.3413, 62.3319, 62.285, 62.2592, 62.1771, 62.055, 61.8978, 61.7851, 61.7429, 61.6701] };
  WCL[829] = struct {
   x = [-79.8628, -80.0505],
   y = [62.3625, 62.3578] };
  WCL[830] = struct {
   x = [-96.936, -96.9548, -96.9877, -97.1003, -97.1989, -97.3538, -97.3068, -97.3538, -97.5274, -97.6072, -97.7715, -97.9358, -97.9592, -98.1564, -98.2456, -98.3113, -98.2033, -98.2127, -98.3019, -98.4802,
        -98.5553, -98.5319, -98.546, -98.5131, -98.4568, -98.5882, -98.6445, -98.7666, -98.8839, -99.0529, -99.1749, -99.3063, -99.4612, -99.5129, -99.5035, -99.4471, -99.3392, -99.128, -99.1374,
        -99.0529, -98.9214, -98.7525, -98.7008, -98.5647, -98.5131, -98.4098, -98.3676, -98.2127, -98.2033, -98.0907, -97.8701, -97.8184, -97.6823, -97.5087, -97.3303, -97.166, -97.1003, -97.0674,
        -97.0111, -96.8891, -96.7342, -96.5793, -96.4714, -96.293, -96.185, -96.0959, -95.9738, -95.8753, -95.711, -95.5655, -95.5561, -95.4247, -95.2557, -95.2041, -95.2885, -95.4998, -95.5889,
        -95.7298, -95.8189, -95.8189, -95.9738, -96.0301, -96.0959, -96.2507, -96.3399, -96.4948, -96.5934, -96.7483, -96.8797, -96.936],
   y = [69.5344, 69.5508, 69.5813, 69.6447, 69.6728, 69.6681, 69.7292, 69.7667, 69.7902, 69.8512, 69.8841, 69.8817, 69.8137, 69.7479, 69.6611, 69.6001, 69.5508, 69.5179, 69.5578, 69.5578, 69.5156, 69.4874,
        69.41, 69.356, 69.302, 69.2551, 69.1894, 69.1847, 69.1589, 69.1307, 69.1307, 69.119, 69.0838, 69.004, 68.9054, 68.8913, 68.835, 68.8022, 68.8632, 68.8984, 68.9054, 68.8655, 68.8069, 68.7834,
        68.7552, 68.7787, 68.8139, 68.774, 68.7153, 68.7036, 68.7153, 68.6754, 68.6121, 68.5745, 68.5557, 68.5464, 68.5229, 68.5745, 68.5722, 68.5346, 68.4994, 68.4783, 68.4548, 68.483, 68.5628,
        68.6191, 68.6262, 68.659, 68.7271, 68.7318, 68.6801, 68.7036, 68.7623, 68.8256, 68.8397, 68.7951, 68.8397, 68.8937, 68.9735, 69.0368, 69.1307, 69.1824, 69.2011, 69.241, 69.3091, 69.3607,
        69.403, 69.4616, 69.4992, 69.5344] };
  WCL[831] = struct {
   x = [-100.027, -99.6936, -99.5246, -99.3556, -99.2711, -99.2148, -99.0787, -99.0787, -98.8534, -98.6844, -98.492, -98.4075, -98.5765, -98.5765, -98.7126, -98.7689, -98.6844, -98.5765, -98.5201, -98.4357,
        -98.323, -98.323, -98.1822, -98.0179, -97.9334, -97.68, -97.5673, -97.3467, -97.1777, -97.1777, -97.2059, -97.2904, -97.3186, -97.4547, -97.4547, -97.5392, -97.9334, -98.0179, -97.9898,
        -97.9616, -98.0977, -98.154, -98.2104, -98.3793, -98.4638, -98.323, -98.323, -98.323, -98.5201, -98.5201, -98.6563, -98.7126, -98.5765, -98.4638, -98.3793, -98.3793, -98.2385, -98.1822,
        -98.1259, -97.9334, -97.7363, -97.7363, -97.7363, -98.0414, -97.9053, -97.68, -97.6518, -97.4547, -97.4547, -97.5392, -97.2904, -97.1777, -97.0369, -97.0088, -96.8961, -97.0088, -96.868,
        -96.7037, -96.5629, -96.5629, -96.6474, -96.6755, -96.7882, -96.7037, -96.5629, -96.4784, -96.4784, -96.0607, -96.0325, -96.0607, -96.0325, -96.0607, -96.1733, -96.2015, -96.0607, -96.2813,
        -96.4221, -96.3376, -96.0607, -96.0607, -96.0607, -95.9762, -95.779, -95.6945, -95.5303, -95.5303, -95.6382, -95.5584, -95.3613, -95.305, -95.2768, -95.3613, -95.3331, -95.4176, -95.6664,
        -95.6664, -95.5584, -95.5303, -95.5866, -95.6382, -95.5303, -95.4458, -95.3613, -95.3331, -95.2487, -94.9436, -94.9436, -95.0609, -94.9107, -94.8262, -94.7277, -94.6244, -94.5399, -94.4179,
        -94.3615, -94.2958, -94.2864, -94.1738, -93.9532, -93.7091, -93.5777, -93.7748, -93.7655, -93.742, -93.7655, -93.8218, -93.9438, -94.0518, -94.0658, -94.1175, -94.1316, -94.0095, -93.9297,
        -93.9954, -94.1503, -94.2958, -94.5071, -94.6244, -94.737, -94.6807, -94.6807, -94.6713, -94.6572, -94.5728, -94.4366, -94.3709, -94.2958, -94.2958, -94.4038, -94.4038, -94.3709, -94.2301,
        -94.0658, -93.8781, -93.8218, -93.9438, -93.9532, -93.8969, -93.7091, -93.6434, -93.6763, -93.7326, -93.9297, -94.0518, -94.1832, -94.4273, -94.5071, -94.6807, -94.7793, -94.7699, -94.8591,
        -94.9999, -95.0187, -95.0187, -95.0703, -95.2064, -95.2909, -95.3472, -95.3378, -95.4458, -95.657, -95.8213, -96.0747, -96.1404, -96.0747, -96.1733, -96.3939],
   y = [67.8611, 67.8282, 67.8494, 67.786, 67.7555, 67.7766, 67.786, 67.7437, 67.7766, 67.786, 67.786, 67.786, 67.8705, 67.9033, 67.9643, 68.0394, 68.0911, 68.0911, 68.07, 67.9972, 67.955, 67.8705, 67.786,
        67.7649, 67.7015, 67.6686, 67.6264, 67.6381, 67.7132, 67.786, 67.955, 67.955, 67.9033, 67.9033, 67.955, 67.9972, 67.9855, 67.9244, 67.9033, 67.84, 67.8611, 67.9244, 67.9972, 68.0817, 68.1427,
        68.1122, 68.1732, 68.2061, 68.2671, 68.3093, 68.3703, 68.4126, 68.4126, 68.3492, 68.3821, 68.4126, 68.4009, 68.4126, 68.3703, 68.3703, 68.3915, 68.4126, 68.4126, 68.5464, 68.5651, 68.5464,
        68.4947, 68.4126, 68.4126, 68.5159, 68.5041, 68.4126, 68.3821, 68.3915, 68.361, 68.3093, 68.2788, 68.3093, 68.3305, 68.2788, 68.2061, 68.1333, 68.0911, 68.0488, 68.0817, 68.0277, 68.1638,
        68.2671, 68.2577, 68.1943, 68.1122, 67.9972, 67.9033, 67.7766, 67.6686, 67.6264, 67.5842, 67.4668, 67.4785, 67.3589, 67.3049, 67.3377, 67.4128, 67.3589, 67.4246, 67.3589, 67.3166, 67.2509,
        67.2415, 67.3166, 67.3589, 67.4785, 67.5842, 67.6804, 67.7437, 67.786, 67.9456, 67.9972, 68.0277, 68.1005, 68.1005, 68.1216, 68.1005, 68.07, 68.0911, 68.1005, 68.1005, 68.0958, 68.0676,
        68.0512, 68.0723, 68.1544, 68.1943, 68.2108, 68.2389, 68.2483, 68.3093, 68.4173, 68.483, 68.5393, 68.6121, 68.6308, 68.6825, 68.7787, 68.8655, 68.9664, 68.9805, 69.0134, 68.9664, 68.9101,
        68.8468, 68.8749, 68.9172, 68.8632, 68.7599, 68.7318, 68.7153, 68.7224, 68.7341, 68.7998, 68.8468, 68.8655, 68.9406, 68.9664, 68.9547, 69.0134, 69.0908, 69.1542, 69.126, 69.1471, 69.2316,
        69.3419, 69.3678, 69.3865, 69.4147, 69.3185, 69.2199, 69.173, 69.2668, 69.3724, 69.4264, 69.5039, 69.5039, 69.464, 69.4264, 69.4382, 69.4921, 69.6118, 69.6447, 69.5884, 69.5578, 69.6001,
        69.6306, 69.6306, 69.6329, 69.6728, 69.6799, 69.7292, 69.7597, 69.7644, 69.7456, 69.7456, 69.8066, 69.8559, 69.8958, 69.9193, 69.9521] };
  WCL[832] = struct {
   x = [-92.1555, -92.0147, -92.0335, -92.0898, -92.2212, -92.3433, -92.3855, -92.4653, -92.5733, -92.5967, -92.7516, -92.6953, -92.7281, -92.5075, -92.3761, -92.2776, -92.179, -91.9443, -91.8457, -91.7049,
        -91.6392, -91.5172, -91.3717, -91.3623, -91.4609, -91.3388, -91.2825, -91.2309, -91.1511, -91.0619, -90.9868, -90.9117, -90.8976, -90.8742, -90.8742, -90.9211, -90.9962, -91.1652, -91.2731,
        -91.4186, -91.198, -90.9774, -90.8084, -90.7897, -90.6113, -90.6348, -90.4705, -90.5456, -90.5456, -90.5878, -90.5033, -90.4142, -90.3813, -90.3578, -90.217, -90.1044, -90.0152, -90.0152,
        -90.0152],
   y = [70.0507, 69.9685, 69.9427, 69.9122, 69.8559, 69.8559, 69.8348, 69.816, 69.816, 69.7784, 69.7362, 69.6916, 69.6681, 69.6728, 69.6752, 69.6259, 69.5391, 69.4992, 69.5109, 69.5179, 69.5461, 69.6118,
        69.6189, 69.5813, 69.5226, 69.5391, 69.5062, 69.4734, 69.4968, 69.4804, 69.4522, 69.5109, 69.5039, 69.4405, 69.4147, 69.3419, 69.2973, 69.356, 69.3325, 69.349, 69.2481, 69.126, 69.1073,
        69.0439, 68.9406, 68.9148, 68.8069, 68.6989, 68.5839, 68.5229, 68.4173, 68.3657, 68.3164, 68.3046, 68.3492, 68.4337, 68.5112, 68.5722, 68.6355] };
  WCL[833] = struct {
   x = [-89.7829, -90.0786, -90.1771],
   y = [65.8733, 65.9109, 65.8193] };
  WCL[834] = struct {
   x = [-89.9049, -90.0363, -90.0598, -90.0692],
   y = [63.9442, 63.9818, 63.8949, 63.8902] };
  WCL[835] = struct {
   x = [-89.9706, -90.0269, -90.102, -90.1912, -90.3226, -90.3649, -90.4541, -90.5667, -90.6418, -90.7404, -90.853, -90.9516, -91.0408, -91.1816, -91.2942, -91.4022, -91.482, -91.5336, -91.6697, -91.7777,
        -91.8575, -91.9326, -91.9983, -92.0875, -92.1203, -92.1766, -92.3738, -92.5709, -92.6366, -92.8244, -92.9699, -93.2656, -93.4674, -93.519, -93.5519, -93.6223, -93.5754, -93.4111, -93.289,
        -93.3125, -93.2984, -93.1576, -92.8572, -92.6366, -92.585, -92.585, -92.4723, -92.3644, -92.1203, -92.0875, -92.1626, -92.233, -92.2846, -92.186, -92.1203, -91.9748, -91.7918, -91.6228,
        -91.5712, -91.4257, -91.3599, -91.1159, -91.0736, -90.9094, -90.8718, -90.8953, -90.8296, -90.7404, -90.7075, -90.6747, -90.6559, -90.6418, -90.6183, -90.6183, -90.6324, -90.6324, -90.6324],
   y = [63.8316, 63.7893, 63.6814, 63.618, 63.6227, 63.564, 63.5499, 63.5687, 63.496, 63.5593, 63.5828, 63.6321, 63.6415, 63.672, 63.7166, 63.7541, 63.7307, 63.8363, 63.7682, 63.7635, 63.7588, 63.7987,
        63.7682, 63.7635, 63.7635, 63.7682, 63.8128, 63.841, 63.8855, 63.9231, 63.9489, 64.0217, 64.0686, 64.0264, 63.9865, 63.9442, 63.8855, 63.8363, 63.8715, 63.9043, 63.9536, 63.8855, 63.8527,
        63.8034, 63.7541, 63.7166, 63.7776, 63.7588, 63.7307, 63.6814, 63.672, 63.6415, 63.5828, 63.5734, 63.6415, 63.6861, 63.7095, 63.6814, 63.618, 63.564, 63.5335, 63.5054, 63.4608, 63.4866,
        63.4514, 63.4068, 63.3669, 63.327, 63.2777, 63.219, 63.1651, 63.1299, 63.09, 63.0548, 63.029, 63.0196, 63.0008] };
  WCL[836] = struct {
   x = [-90.6324, -90.6747, -90.7638, -90.8953, -91.0642, -91.1863, -91.3036, -91.4491, -91.6369, -91.834, -92.0781, -92.233, -92.4395, -92.5052, -92.3409, -92.2189, -92.0452, -91.9983, -92.233, -92.3972,
        -92.6085, -92.6413, -92.6836, -92.6836, -92.6413, -92.5944, -92.707, -92.8948, -93.0121, -93.0027, -93.1576, -93.1482, -93.2233, -93.2327, -93.2468, -93.3782, -93.519, -93.4674, -93.3782,
        -93.4439, -93.5566, -93.7068, -93.9415, -94.0635, -94.2606, -94.5141, -94.6033, -94.4014, -94.2043, -94.1926, -94.2489, -94.2818, -94.3357, -94.5235, -94.5117, -94.5117, -94.5117, -94.5798,
        -94.723, -94.7558, -94.7441, -94.6995, -94.723, -94.7323],
   y = [63.0008, 62.9445, 62.8952, 62.8999, 62.8905, 62.8905, 62.8342, 62.8037, 62.7943, 62.813, 62.813, 62.7896, 62.8248, 62.7638, 62.6323, 62.6323, 62.6464, 62.5877, 62.5572, 62.5807, 62.5713, 62.5009,
        62.4094, 62.3108, 62.1841, 62.0597, 62.1325, 62.224, 62.1982, 62.1372, 62.0855, 62.0292, 61.9564, 61.9189, 61.9189, 61.9189, 61.9189, 61.825, 61.7898, 61.7171, 61.6279, 61.5692, 61.4918,
        61.4214, 61.4167, 61.412, 61.3697, 61.3017, 61.2266, 61.1679, 61.067, 60.945, 60.8839, 60.8534, 60.7783, 60.6962, 60.6258, 60.539, 60.5225, 60.4803, 60.4193, 60.3324, 60.2409, 60.0931] };
  WCL[837] = struct {
   x = [-100.862, -100.951, -100.919, -100.905, -100.984, -101.116, -101.271, -101.379, -101.435, -101.435, -101.435, -101.501, -101.557, -101.566, -101.59, -101.646, -101.702, -101.81, -101.867, -101.975,
        -102.04, -102.04, -102.097, -102.209, -102.219, -102.284, -102.341, -102.406, -102.529, -102.618, -102.627, -102.552, -102.529, -102.529, -102.585, -102.66, -102.773, -102.89, -103.012,
        -103.092, -103.143, -103.157, -103.059, -103.035, -103.059, -103.059, -103.068, -103.059, -102.97, -102.871, -102.627, -102.561, -102.364, -102.284, -102.144, -102.097, -101.998, -101.932,
        -101.998, -102.031, -102.13, -102.12, -102.209, -102.242, -102.177, -102.064, -102.022, -101.89, -101.801, -101.59, -101.548, -101.58, -101.688, -101.735, -101.768, -101.867, -101.867,
        -102.008, -102.162, -102.364, -102.406, -102.651, -102.759, -102.791, -102.89, -102.946, -103.092, -103.256, -103.355, -103.51, -103.674, -103.838, -103.918, -104.059, -104.237, -104.336,
        -104.49, -104.57, -104.711, -105.03, -105.251, -105.274, -105.11, -104.988, -105.077, -105.143, -105.241, -105.396, -105.537, -105.57, -105.683, -105.823, -105.927, -106.068, -106.199,
        -106.368, -106.387, -106.321, -106.368, -106.443, -106.64, -106.697, -106.795, -107.049, -106.992, -106.95, -106.959, -107.128, -107.283, -107.415, -107.499, -107.842, -107.907, -108.086,
        -108.405, -108.583, -108.649, -108.551, -108.738, -109.025, -109.32, -109.564, -109.916],
   y = [70.0131, 69.9333, 69.8512, 69.7761, 69.6846, 69.6681, 69.6916, 69.7644, 69.8019, 69.8582, 69.9451, 69.877, 69.809, 69.7409, 69.6869, 69.6611, 69.7245, 69.7456, 69.7034, 69.7644, 69.8465, 69.931,
        69.8817, 69.8747, 69.8958, 69.8442, 69.8066, 69.8066, 69.7761, 69.755, 69.7151, 69.7104, 69.6329, 69.5578, 69.5532, 69.5508, 69.5766, 69.5813, 69.5837, 69.5837, 69.5719, 69.5461, 69.464,
        69.4217, 69.3842, 69.3537, 69.3138, 69.2715, 69.3443, 69.3983, 69.4522, 69.5109, 69.5109, 69.5179, 69.5156, 69.4804, 69.4734, 69.4616, 69.4217, 69.3771, 69.3325, 69.3067, 69.3138, 69.2715,
        69.2082, 69.2011, 69.2739, 69.2481, 69.2293, 69.2199, 69.1542, 69.1002, 69.0908, 69.1073, 69.0603, 69.0134, 68.9852, 68.9899, 68.9617, 68.9453, 68.882, 68.8773, 68.9054, 68.8421, 68.8233,
        68.7881, 68.8256, 68.8022, 68.7834, 68.7951, 68.8139, 68.8256, 68.882, 68.9054, 68.9148, 68.9547, 68.95, 68.8937, 68.9101, 68.9219, 68.9547, 68.997, 69.0509, 69.1119, 69.1471, 69.126, 69.1237,
        69.126, 69.1307, 69.173, 69.1894, 69.2011, 69.1847, 69.2011, 69.2011, 69.2082, 69.2715, 69.3208, 69.4264, 69.4921, 69.5179, 69.464, 69.4147, 69.3771, 69.3185, 69.2645, 69.2129, 69.1612,
        69.0744, 69.004, 68.9899, 68.9735, 68.9664, 68.9688, 68.9617, 68.9664, 68.9617, 68.9289, 68.8186, 68.7388, 68.7036, 68.6473, 68.6262] };
  WCL[838] = struct {
   x = [-110.029, -109.977],
   y = [67.9972, 67.9761] };
  WCL[839] = struct {
   x = [-110.029, -109.921, -109.865, -109.808, -109.752, -109.752, -109.808, -109.696, -109.47, -109.499, -109.165, -109.109, -108.996, -108.996, -109.025, -108.996, -109.025, -108.94, -108.996, -108.884,
        -108.884, -108.856, -108.771, -108.748, -108.663, -108.663, -108.579, -108.635, -108.635, -108.494, -108.41, -108.269, -108.105, -108.02, -107.964, -107.907, -108.048, -108.184, -108.353,
        -108.438, -108.382, -108.161, -107.992, -107.851, -107.823, -107.739, -107.574, -107.433, -107.264, -107.377, -107.462, -107.293, -107.321, -107.462, -107.574, -107.654, -107.795, -107.71,
        -107.71, -107.795, -107.795, -108.048, -108.133, -107.992, -107.879, -107.823, -107.71, -107.851, -107.795, -107.739, -107.49, -107.264, -107.236, -107.236, -106.819, -106.847, -106.678,
        -106.481, -106.509, -106.424, -106.373, -106.26, -106.176, -106.204, -105.95, -106.063, -105.922, -105.922, -106.063, -106.091, -106.373, -106.678, -106.537, -106.678, -106.847, -106.706,
        -106.678, -106.678, -106.678, -106.734, -106.79, -106.762, -106.847, -106.875, -106.983, -107.067, -107.152, -107.236, -107.264, -107.321, -107.462, -107.574, -107.626, -107.795, -107.879,
        -107.992, -107.936, -107.795, -107.795, -107.71, -107.907, -108.048, -108.133, -108.02, -108.184, -108.353, -108.353, -108.522, -108.466, -108.635, -108.579, -108.827, -108.884, -108.856,
        -108.663, -108.663, -108.522, -108.382, -108.105, -107.936, -107.795, -107.598, -107.377, -107.095, -106.983, -106.706, -106.678, -106.424, -106.26, -106.091, -106.035, -105.894, -105.838,
        -105.645, -105.589, -105.73, -105.701, -105.589, -105.561, -105.392, -105.392, -105.561, -105.673, -105.673, -105.561, -105.392, -105.279, -105.251, -105.087, -104.946, -104.833, -104.749,
        -104.777, -104.749, -104.749, -104.58, -104.444, -104.19, -104.021, -103.913, -103.857, -103.857, -103.716, -103.688, -103.632, -103.632, -103.326, -103.326, -103.045, -102.904, -102.768,
        -102.627, -102.345, -102.205, -102.205, -102.069, -102.012, -101.984, -101.871, -101.815, -101.566, -101.369, -101.228, -100.951, -100.839, -100.529, -100.365, -100.196, -100.027],
   y = [67.9244, 67.9033, 67.9338, 67.8705, 67.84, 67.786, 67.786, 67.7766, 67.7437, 67.786, 67.786, 67.7766, 67.786, 67.7649, 67.7226, 67.6804, 67.6475, 67.5842, 67.4997, 67.4457, 67.4879, 67.4551, 67.5419,
        67.6686, 67.6593, 67.5842, 67.5842, 67.5208, 67.4457, 67.4246, 67.4785, 67.3589, 67.3589, 67.2955, 67.2181, 67.1641, 67.1007, 67.1218, 67.0772, 67.1007, 67.035, 66.9247, 66.8144, 66.7815,
        66.9247, 66.9364, 66.8918, 66.9693, 66.9576, 67.035, 67.089, 67.143, 67.1641, 67.1641, 67.2509, 67.3589, 67.434, 67.4668, 67.5842, 67.617, 67.6686, 67.7226, 67.786, 67.9033, 67.8916, 67.9456,
        67.9643, 67.9855, 68.0277, 68.0911, 68.0911, 68.1427, 68.1732, 68.1333, 68.1544, 68.2366, 68.2577, 68.246, 68.361, 68.3821, 68.2976, 68.4126, 68.4525, 68.5159, 68.5557, 68.5769, 68.6262,
        68.6872, 68.6684, 68.6872, 68.6567, 68.6074, 68.5346, 68.5252, 68.5252, 68.4126, 68.4642, 68.4009, 68.3398, 68.3093, 68.3492, 68.3821, 68.3703, 68.4126, 68.4126, 68.4126, 68.3915, 68.3492,
        68.3187, 68.361, 68.3703, 68.3703, 68.4009, 68.3703, 68.3703, 68.3305, 68.2788, 68.2671, 68.2366, 68.2155, 68.2061, 68.1638, 68.1732, 68.2061, 68.2061, 68.1849, 68.2272, 68.2061, 68.2577,
        68.2577, 68.3093, 68.2788, 68.3093, 68.3821, 68.483, 68.5252, 68.6262, 68.6684, 68.6684, 68.6872, 68.6872, 68.72, 68.7693, 68.7998, 68.8608, 68.8913, 68.95, 68.9312, 68.9923, 68.95, 68.9219,
        68.9101, 68.8209, 68.7693, 68.7599, 68.72, 68.72, 68.6567, 68.5956, 68.5651, 68.5252, 68.5252, 68.5041, 68.4736, 68.483, 68.4126, 68.3915, 68.361, 68.2882, 68.2788, 68.2577, 68.2577, 68.246,
        68.1849, 68.1216, 68.07, 68.0817, 68.0394, 68.0817, 68.0488, 68.0488, 68.1122, 68.0817, 68.0911, 68.1333, 68.2061, 68.1544, 68.0394, 67.9338, 67.9033, 67.786, 67.786, 67.7015, 67.7132, 67.7766,
        67.7649, 67.786, 67.7766, 67.7437, 67.7649, 67.7132, 67.7555, 67.786, 67.7649, 67.7437, 67.786, 67.8705, 67.8705, 67.8611] };
  WCL[840] = struct {
   x = [-110.024, -109.874, -109.705, -109.55, -109.428, -109.33, -109.222, -109.133, -109.1, -109.142, -109.278, -109.409, -109.452, -109.442, -109.475, -109.607, -109.794, -109.916],
   y = [62.8905, 62.8694, 62.8553, 62.8389, 62.8342, 62.8201, 62.7896, 62.7332, 62.6887, 62.6582, 62.6183, 62.5572, 62.5455, 62.6018, 62.6183, 62.6582, 62.698, 62.7286] };
  WCL[841] = struct {
   x = [-110.146, -109.949, -109.851, -109.884],
   y = [62.5525, 62.5525, 62.5314, 62.461] };
  WCL[842] = struct {
   x = [-120.017, -119.942, -119.778, -119.599, -119.379, -119.215, -119.027, -118.881, -118.717, -118.618, -118.496, -118.365, -118.121, -117.9, -117.745, -117.661, -117.483, -117.318, -117.182, -117.051,
        -116.985, -117.018],
   y = [69.356, 69.356, 69.349, 69.3067, 69.3067, 69.2856, 69.2528, 69.2551, 69.2316, 69.1659, 69.1307, 69.0955, 69.0674, 69.0157, 68.9735, 68.9571, 68.9571, 68.9547, 68.9219, 68.9101, 68.9054, 68.9312] };
  WCL[843] = struct {
   x = [-119.963, -119.963],
   y = [66.9904, 66.9904] };
  WCL[844] = struct {
   x = [-120.062, -119.94, -119.973, -119.86, -119.663, -119.489, -119.456, -119.541, -119.71, -119.874, -119.94, -119.93, -119.827, -119.686, -119.686, -119.696, -119.564, -119.376, -119.212, -119.081,
        -118.945, -118.804, -118.672, -118.649, -118.583, -118.518, -118.452, -118.316, -118.217, -118.029, -117.987, -117.889, -117.865, -117.978, -118.029, -118.119, -118.217, -118.363, -118.461,
        -118.438, -118.25, -118.133, -118.062, -118.142, -118.133, -117.997, -117.931, -117.865, -117.879, -117.865, -117.842, -117.799, -117.701, -117.635, -117.612, -117.659, -117.743, -117.776,
        -117.691, -117.644, -117.635, -117.546, -117.48, -117.457, -117.438, -117.457, -117.537, -117.635, -117.71, -117.734, -117.691, -117.691, -117.579, -117.555, -117.579, -117.579, -117.659,
        -117.799, -117.912, -118.076, -118.25, -118.25, -118.151, -118.283, -118.461, -118.56, -118.672, -118.794, -119.048, -119.123, -119.269, -119.4, -119.531, -119.677, -119.639, -119.653,
        -119.743, -119.851, -119.996],
   y = [65.1176, 65.1528, 65.2232, 65.2819, 65.2702, 65.2561, 65.296, 65.3664, 65.4016, 65.4345, 65.4837, 65.526, 65.5706, 65.6128, 65.6574, 65.6903, 65.7255, 65.7114, 65.7302, 65.7489, 65.7161, 65.6668,
        65.6527, 65.5987, 65.5706, 65.5753, 65.5706, 65.5893, 65.6034, 65.6292, 65.6433, 65.648, 65.6973, 65.702, 65.7067, 65.7067, 65.6715, 65.6386, 65.6433, 65.6973, 65.7701, 65.8193, 65.8475,
        65.8663, 65.9015, 65.932, 65.9461, 65.9719, 65.9907, 66.0259, 66.0493, 66.0798, 66.1432, 66.1737, 66.1995, 66.2042, 66.2042, 66.2441, 66.2746, 66.2887, 66.3145, 66.3591, 66.3896, 66.453,
        66.4741, 66.4788, 66.4248, 66.4037, 66.3943, 66.453, 66.4835, 66.5257, 66.5562, 66.5703, 66.5844, 66.6196, 66.622, 66.6008, 66.5914, 66.5304, 66.453, 66.4248, 66.4295, 66.3591, 66.3122,
        66.2981, 66.2981, 66.2887, 66.2535, 66.2629, 66.2981, 66.2934, 66.2746, 66.2723, 66.3333, 66.338, 66.3192, 66.3427, 66.3544] };
  WCL[845] = struct {
   x = [-120.073, -119.933, -119.778, -119.689, -119.557, -119.445, -119.304, -119.29, -119.28, -119.158, -119.06, -119.092, -119.182, -119.247, -119.322, -119.379, -119.501, -119.524, -119.599, -119.665,
        -119.698, -119.712, -119.764, -119.876, -119.951, -119.966, -119.966],
   y = [66.5961, 66.622, 66.6454, 66.6525, 66.6924, 66.7229, 66.7675, 66.7792, 66.8097, 66.8097, 66.8355, 66.8566, 66.866, 66.8871, 66.8918, 66.8918, 66.9224, 66.9482, 66.9482, 66.9435, 66.9224, 66.9059,
        66.9059, 66.927, 66.9576, 66.974, 66.9904] };
  WCL[846] = struct {
   x = [-114.889, -114.932, -114.978, -115.03, -115.03, -115.11, -115.241, -115.363, -115.518, -115.725, -115.913, -116.067, -116.213, -116.452, -116.574, -116.574, -116.776, -116.908, -116.917, -116.964,
        -117.161, -117.236, -117.292, -117.368],
   y = [69.2856, 69.2786, 69.2481, 69.2481, 69.2551, 69.2528, 69.2551, 69.2481, 69.2856, 69.2856, 69.2973, 69.3255, 69.3654, 69.4264, 69.4616, 69.5297, 69.5508, 69.593, 69.6447, 69.7174, 69.7761, 69.8254,
        69.8888, 69.9521] };
  WCL[847] = struct {
   x = [-109.916, -110.048, -110.203, -110.292, -110.414, -110.569, -110.7, -110.789, -110.954, -111.052, -111.198, -111.329, -111.371, -111.24, -111.249, -111.493, -111.714, -111.911, -112.024, -112.146,
        -112.352, -112.784, -112.981, -113.258, -113.225, -113.028, -113.183, -113.371, -113.512, -113.577, -113.676, -113.676, -113.61, -113.61, -113.765, -113.775, -113.854, -114.019, -114.028,
        -114.183, -114.338, -114.507, -114.77, -114.892],
   y = [68.6262, 68.6543, 68.6355, 68.6121, 68.6238, 68.6402, 68.6027, 68.6473, 68.6238, 68.6027, 68.6003, 68.591, 68.5722, 68.5675, 68.5393, 68.5393, 68.5229, 68.4994, 68.5229, 68.5229, 68.5229, 68.5159,
        68.5112, 68.49, 68.5159, 68.5675, 68.5956, 68.6426, 68.7341, 68.8233, 68.8937, 68.9923, 69.065, 69.0885, 69.1237, 69.1495, 69.1894, 69.2293, 69.2668, 69.2833, 69.2973, 69.2973, 69.2903,
        69.2856] };
  WCL[848] = struct {
   x = [-117.016, -116.762, -116.457, -116.232, -116.035, -115.978, -116.147, -116.288, -116.204, -116.063, -115.955, -115.758, -115.758, -115.673, -115.561, -115.255, -115.363, -115.086, -115.058, -114.889,
        -114.889, -114.805, -114.697, -114.556, -114.472, -114.331, -114.218, -114.218, -114.082, -114.082, -114.138, -114.054, -114.218, -114.274, -114.387, -114.443, -114.472, -114.725, -115.058,
        -114.833, -114.974, -115.34, -115.34, -115.255, -115.34, -115.34, -115.34, -115.617, -115.504, -115.42, -115.284, -115.392, -115.171, -114.749, -114.669, -114.528, -114.443, -114.303, -114.218,
        -113.998, -113.631, -113.383, -113.045, -113.073, -112.852, -112.402, -112.35, -112.125, -112.012, -112.068, -111.871, -111.984, -111.843, -111.787, -111.594, -111.566, -111.425, -111.341,
        -111.148, -111.148, -111.172, -110.979, -110.839, -110.613, -110.421, -110.196, -110.111, -110.027],
   y = [68.9312, 68.9101, 68.8913, 68.9007, 68.8209, 68.8796, 68.9312, 68.9805, 68.9923, 69.0016, 68.9711, 68.9617, 69.0016, 68.9923, 68.9923, 68.9617, 68.9312, 68.8702, 68.8209, 68.8514, 68.7904, 68.7693,
        68.7599, 68.72, 68.6684, 68.5956, 68.5557, 68.4126, 68.483, 68.4009, 68.3398, 68.2976, 68.3093, 68.2577, 68.246, 68.2882, 68.3093, 68.2788, 68.2976, 68.2577, 68.1943, 68.2155, 68.2061, 68.1732,
        68.1333, 68.1005, 68.0394, 67.9761, 67.9338, 67.9244, 67.9456, 67.9033, 67.8705, 67.786, 67.84, 67.8188, 67.7766, 67.7766, 67.786, 67.7555, 67.7649, 67.7344, 67.7649, 67.7437, 67.7226, 67.7344,
        67.786, 67.786, 67.786, 67.7344, 67.7344, 67.7649, 67.8188, 67.7649, 67.7555, 67.786, 67.7766, 67.8916, 67.9127, 67.8705, 67.786, 67.786, 67.8188, 67.9456, 67.9972, 67.9972, 68.0606, 67.9972] };
  WCL[849] = struct {
   x = [-109.975, -110.027],
   y = [67.9761, 67.9244] };
  WCL[850] = struct {
   x = [-115.788, -115.737, -115.737, -115.746, -115.755, -115.769, -115.755, -115.713, -115.624, -115.502, -115.394, -115.258, -115.117, -115.018, -114.906, -114.784, -114.629, -114.53, -114.422, -114.333,
        -114.319, -114.343, -114.375, -114.455, -114.441, -114.333, -114.267, -114.211, -114.155, -114.155, -114.178, -114.164, -114.178, -114.164, -114.113, -114.023, -113.911, -113.793, -113.723,
        -113.624, -113.582, -113.535, -113.502, -113.394, -113.385, -113.427, -113.469, -113.451, -113.305, -113.16, -112.986, -112.831, -112.653, -112.489, -112.277, -112.113, -111.949, -111.883,
        -111.836, -111.817, -111.705, -111.573, -111.418, -111.263, -111.076, -110.954, -110.832, -110.7, -110.545, -110.4, -110.301, -110.17, -110.024],
   y = [62.9961, 63.0313, 63.0196, 62.975, 62.9398, 62.8905, 62.8389, 62.799, 62.7403, 62.6887, 62.6582, 62.5971, 62.5666, 62.5103, 62.4493, 62.4141, 62.393, 62.4141, 62.4962, 62.5267, 62.576, 62.6934,
        62.7497, 62.8295, 62.8201, 62.7685, 62.7192, 62.6675, 62.6136, 62.5831, 62.5455, 62.5103, 62.461, 62.4188, 62.3578, 62.2967, 62.2146, 62.1583, 62.1465, 62.1677, 62.2099, 62.285, 62.3578,
        62.3319, 62.2545, 62.2099, 62.1418, 62.0855, 62.0503, 62.0386, 62.0503, 62.0761, 62.0902, 62.116, 62.2029, 62.3108, 62.3671, 62.4399, 62.522, 62.5971, 62.6582, 62.7192, 62.7497, 62.7896,
        62.8248, 62.8506, 62.8647, 62.8905, 62.8999, 62.9163, 62.9093, 62.8999, 62.8905] };
  WCL[851] = struct {
   x = [-109.914, -110.092, -110.247, -110.397, -110.543, -110.552, -110.468, -110.346, -110.191, -110.078, -110.111, -110.266, -110.388, -110.355, -110.233, -110.144],
   y = [62.7286, 62.7403, 62.745, 62.7403, 62.7332, 62.7239, 62.7145, 62.698, 62.6793, 62.6722, 62.637, 62.637, 62.637, 62.5924, 62.5877, 62.5525] };
  WCL[852] = struct {
   x = [-109.881, -110.06, -110.29, -110.421, -110.534, -110.674, -110.843, -110.961, -111.064, -111.031, -110.984, -110.975, -110.975, -111.172, -111.36, -111.439, -111.547, -111.679, -111.745, -111.768,
        -111.759, -111.777, -111.801, -111.932, -112.003, -112.087, -112.233, -112.308, -112.444, -112.632, -112.838, -112.904, -113.082, -113.261, -113.303, -113.413, -113.556, -113.711, -113.744,
        -113.765, -113.798, -113.875, -114.03, -114.185, -114.328, -114.516, -114.659, -114.838, -115.058, -115.211, -115.345, -115.488, -115.654, -115.786, -115.92, -116.084, -116.239, -116.337,
        -116.415, -116.504, -116.614, -116.725, -116.835, -116.924, -117.044, -117.166, -117.222, -117.288, -117.353, -117.464, -117.53, -117.619, -117.684, -117.741, -117.806, -117.938, -117.994,
        -117.994, -118.102, -118.224, -118.212, -118.112, -118.112, -118.201, -118.29, -118.4, -118.487, -118.609, -118.665, -118.708, -118.628, -118.562, -118.473, -118.398, -118.299, -118.154,
        -118.022, -117.844, -117.703, -117.591, -117.515, -117.403, -117.304, -117.248, -117.105, -116.905, -116.741, -116.696, -116.631, -116.485, -116.387, -116.265, -116.11, -115.955, -115.821,
        -115.812, -115.711, -115.711, -115.856, -115.8, -115.645, -115.561, -115.49, -115.293, -115.302, -115.392, -115.316, -115.26, -115.27, -115.194, -115.284, -115.237, -115.162, -115.072,
        -114.964, -114.861, -114.73, -114.608, -114.598, -114.622, -114.655, -114.795, -114.786, -114.664, -114.598, -114.556, -114.5, -114.556, -114.72, -114.875, -115.082, -115.04, -114.95, -114.917,
        -114.941, -115.007, -115.129, -115.227, -115.26, -115.359, -115.415, -115.561, -115.758, -115.758, -115.8, -115.955, -116.11, -116.067, -116.119, -116.265, -116.344, -116.429, -116.363,
        -116.222, -116.053, -115.945, -115.847, -115.8],
   y = [62.461, 62.4845, 62.4798, 62.4305, 62.4141, 62.4446, 62.4446, 62.4751, 62.4962, 62.4399, 62.3836, 62.3319, 62.3108, 62.292, 62.2357, 62.1888, 62.1465, 62.0949, 62.1066, 62.0691, 62.0128, 61.9564,
        61.8978, 61.8203, 61.7781, 61.7429, 61.6748, 61.6326, 61.5645, 61.5059, 61.5012, 61.4589, 61.4542, 61.48, 61.4002, 61.3275, 61.3181, 61.2806, 61.243, 61.1632, 61.0881, 61.013, 60.9755, 61.0036,
        61.0506, 61.0294, 60.9919, 60.9332, 60.9004, 60.8746, 60.8581, 60.8323, 60.8323, 60.8581, 60.8417, 60.8088, 60.7995, 60.8417, 60.9332, 60.9614, 61.0083, 61.0294, 61.0506, 61.0787, 61.1092,
        61.121, 61.1515, 61.1679, 61.1843, 61.2008, 61.2219, 61.2524, 61.2806, 61.3275, 61.3744, 61.4002, 61.4167, 61.3862, 61.3956, 61.365, 61.3111, 61.2853, 61.2735, 61.2806, 61.2806, 61.2806,
        61.2806, 61.2899, 61.3111, 61.3533, 61.3956, 61.4378, 61.4965, 61.5551, 61.5856, 61.581, 61.5551, 61.5012, 61.4707, 61.4331, 61.4073, 61.365, 61.3322, 61.3181, 61.3111, 61.2806, 61.2853,
        61.3275, 61.4167, 61.4073, 61.3533, 61.2899, 61.2641, 61.2102, 61.2055, 61.2853, 61.3181, 61.365, 61.3956, 61.4495, 61.4378, 61.4542, 61.4167, 61.48, 61.5059, 61.5763, 61.5903, 61.649, 61.7053,
        61.7523, 61.7992, 61.8415, 61.8532, 61.8415, 61.825, 61.8156, 61.8109, 61.8203, 61.8579, 61.8884, 61.9306, 61.9259, 61.9705, 61.9611, 61.9705, 61.9823, 62.0128, 62.0644, 62.0644, 62.0855,
        62.102, 62.1418, 62.1771, 62.2287, 62.2498, 62.2404, 62.2287, 62.285, 62.3671, 62.4446, 62.4798, 62.4962, 62.515, 62.5572, 62.637, 62.698, 62.7544, 62.8201, 62.8811, 62.9093, 62.9398, 62.9914,
        63.0102, 62.9703, 62.9562, 62.9797, 62.9961, 62.9961] };
  WCL[853] = struct {
   x = [-113.061, -113.061, -113.019, -112.93, -112.808, -112.653, -112.498, -112.357, -112.226, -112.169, -112.169, -112.259, -112.277, -112.212, -112.235, -112.39, -112.446, -112.564, -112.611, -112.7,
        -112.784, -112.831, -112.93, -113.019, -113.061],
   y = [61.8415, 61.8461, 61.872, 61.879, 61.879, 61.8978, 61.9189, 61.9518, 61.9916, 61.9916, 61.9447, 61.8978, 61.8626, 61.825, 61.7851, 61.7734, 61.7382, 61.7171, 61.7476, 61.7781, 61.7687, 61.7945,
        61.8203, 61.825, 61.8415] };
  WCL[854] = struct {
   x = [-116.441, -116.518, -116.429, -116.363, -116.286, -116.175, -116.164, -116.164, -116.241, -116.351, -116.483, -116.649, -116.617, -116.572, -116.518, -116.45, -116.441],
   y = [61.2219, 61.2172, 61.2008, 61.1632, 61.1304, 61.0928, 61.0341, 60.9989, 61.0036, 61.0459, 61.0787, 61.1092, 61.1421, 61.1961, 61.2266, 61.2219, 61.2219] };
  WCL[855] = struct {
   x = [-111.261, -111.205, -111.172, -111.106, -111.008, -110.876, -110.806, -110.731, -110.665, -110.665, -110.764, -110.895, -111.031, -111.106, -111.195, -111.303, -111.35, -111.425, -111.557, -111.679,
        -111.702, -111.58, -111.482, -111.369, -111.261],
   y = [62.6628, 62.6582, 62.6628, 62.6582, 62.6675, 62.698, 62.6934, 62.6887, 62.6793, 62.6136, 62.6065, 62.6018, 62.5619, 62.5572, 62.576, 62.5713, 62.5455, 62.5361, 62.4751, 62.4493, 62.4845, 62.5666,
        62.6229, 62.6276, 62.6628] };
  WCL[856] = struct {
   x = [-129.613, -129.688, -129.853, -129.998],
   y = [70.0202, 69.9803, 69.9333, 69.8958] };
  WCL[857] = struct {
   x = [-130.165, -129.954, -129.78, -129.592, -129.447, -129.273, -129.085, -128.954, -128.954, -128.987, -129.062, -129.174, -129.207, -129.085, -128.907, -128.766, -128.621, -128.424],
   y = [69.7034, 69.7362, 69.8019, 69.8277, 69.8395, 69.8582, 69.9146, 69.9685, 69.9263, 69.8465, 69.8512, 69.7667, 69.7034, 69.7339, 69.7644, 69.7972, 69.877, 69.9498] };
  WCL[858] = struct {
   x = [-127.163, -127.018, -126.943, -126.877, -126.797, -126.68, -126.501, -126.436, -126.403, -126.389, -126.356, -126.225, -126.06, -125.873, -125.694, -125.474, -125.488, -125.422, -125.234, -125.234,
        -125.375, -125.553, -125.694, -125.643, -125.497, -125.356, -125.211, -125.3, -125.408, -125.497, -125.389, -125.145, -125.145, -125.286, -125.286, -125.234, -125.079, -124.924, -124.859],
   y = [70.0366, 69.9568, 69.8629, 69.7831, 69.7034, 69.6236, 69.5837, 69.5226, 69.5039, 69.4874, 69.4921, 69.4499, 69.3912, 69.3654, 69.3654, 69.3607, 69.3842, 69.3983, 69.403, 69.4335, 69.4616, 69.4452,
        69.4264, 69.4968, 69.4992, 69.5039, 69.5297, 69.5344, 69.5813, 69.6306, 69.6681, 69.7034, 69.7526, 69.7644, 69.8442, 69.7855, 69.8629, 69.9216, 69.9897] };
  WCL[859] = struct {
   x = [-124.915, -125.046, -125.122, -125.187],
   y = [70.0014, 69.9756, 69.9216, 69.9427] };
  WCL[860] = struct {
   x = [-124.76, -124.704],
   y = [70.0272, 69.9615] };
  WCL[861] = struct {
   x = [-124.429, -124.453, -124.462, -124.462, -124.476, -124.518, -124.495, -124.396, -124.265, -124.166, -124.068, -124.101, -124.143, -124.265, -124.41, -124.443, -124.486, -124.518, -124.561, -124.551,
        -124.429, -124.307, -124.176, -124.166, -124.11, -124.021, -123.899, -123.735, -123.636, -123.481, -123.439, -123.359, -123.26, -123.237, -123.138, -123.138, -123.12, -122.965, -122.876,
        -122.697, -122.5, -122.345, -122.214, -122.204, -122.115, -122.026, -121.927, -121.805, -121.651, -121.543, -121.397, -121.242, -121.101, -120.956, -120.834, -120.726, -120.571, -120.482,
        -120.318, -120.172, -120.153, -120.017],
   y = [70.0014, 69.9615, 69.9146, 69.87, 69.809, 69.7667, 69.7174, 69.6916, 69.7245, 69.7221, 69.7057, 69.6869, 69.6447, 69.5696, 69.4921, 69.4452, 69.403, 69.464, 69.403, 69.3607, 69.3654, 69.3443,
        69.3419, 69.3419, 69.349, 69.3912, 69.3912, 69.3771, 69.356, 69.3983, 69.4616, 69.4968, 69.5156, 69.5884, 69.6846, 69.7292, 69.8019, 69.8019, 69.8207, 69.8395, 69.8066, 69.8137, 69.8066,
        69.7902, 69.7902, 69.8066, 69.8019, 69.8019, 69.7784, 69.7761, 69.7597, 69.7104, 69.6728, 69.6329, 69.6071, 69.5461, 69.4851, 69.4335, 69.403, 69.3771, 69.3678, 69.356] };
  WCL[862] = struct {
   x = [-119.963, -120.005, -120.081, -120.184, -120.249, -120.259, -120.372, -120.456, -120.545, -120.625, -120.709, -120.813, -120.921, -121.019, -121.076, -121.132, -121.23, -121.287, -121.306, -121.404,
        -121.517, -121.606, -121.658, -121.761, -121.812, -121.892, -121.967, -122.08, -122.202, -122.287, -122.409, -122.498, -122.61, -122.676, -122.775, -122.84, -122.883, -122.93, -123.005,
        -123.094, -123.084, -123.103, -123.192, -123.272, -123.38, -123.559, -123.709, -123.845, -123.986, -124.051, -124.131, -124.253, -124.338, -124.361, -124.417, -124.474, -124.525, -124.624,
        -124.737, -124.826, -124.901, -125.014, -125.089, -125.122, -125.145, -125.154, -125.122, -125.112, -125.023, -124.948, -124.892, -124.868, -124.845, -124.746, -124.694, -124.638, -124.638,
        -124.558, -124.46, -124.319, -124.239, -124.22, -124.187, -124.108, -124.009, -123.92, -123.821, -123.676, -123.512, -123.39, -123.291, -123.225, -123.202, -123.183, -123.084, -123.052,
        -123.052, -123.052, -123.094, -122.972, -122.864, -122.751, -122.587, -122.456, -122.343, -122.202, -122.033, -121.949, -121.845, -121.747, -121.69, -121.672, -121.615, -121.54, -121.418,
        -121.32, -121.273, -121.338, -121.437, -121.559, -121.761, -121.911, -122.08, -122.221, -122.287, -122.366, -122.465, -122.54, -122.563, -122.563, -122.554, -122.432, -122.333, -122.254,
        -122.188, -122.211, -122.244, -122.254, -122.423, -122.54, -122.686, -122.775, -122.84, -122.84, -122.793, -122.793, -122.817, -122.85, -122.939, -123.014, -123.117, -123.202, -123.305,
        -123.436, -123.479, -123.469, -123.39, -123.371, -123.235, -123.15, -123.052, -122.972, -122.775, -122.578, -122.423, -122.319, -122.211, -122.047, -121.892, -121.761, -121.615, -121.615,
        -121.507, -121.507, -121.46, -121.273, -121.151, -120.954, -120.789, -120.625, -120.522, -120.456, -120.456, -120.423, -120.503, -120.569, -120.677, -120.789, -120.911, -120.977, -120.977,
        -121.118, -121.254, -121.362, -121.47, -121.493, -121.428, -121.32, -121.174, -121.052, -120.93, -120.813, -120.555, -120.414, -120.259, -120.193, -120.062],
   y = [66.9904, 66.9787, 66.9646, 66.9435, 66.9012, 66.866, 66.8355, 66.8003, 66.8097, 66.8097, 66.8003, 66.7792, 66.7628, 66.7534, 66.7182, 66.6971, 66.6877, 66.6525, 66.6055, 66.5914, 66.5844, 66.5844,
        66.5961, 66.5609, 66.5351, 66.514, 66.5187, 66.4905, 66.4835, 66.5234, 66.4999, 66.4882, 66.4788, 66.4788, 66.4788, 66.4741, 66.46, 66.4295, 66.4178, 66.4295, 66.4882, 66.5046, 66.4835,
        66.4436, 66.4084, 66.3826, 66.3427, 66.2887, 66.2535, 66.2488, 66.2723, 66.3239, 66.3286, 66.3122, 66.284, 66.2676, 66.2488, 66.2488, 66.2488, 66.2347, 66.2441, 66.2723, 66.2488, 66.1878,
        66.1432, 66.0798, 66.0259, 65.986, 65.9367, 65.9015, 65.8874, 65.9273, 65.9766, 65.9813, 65.9508, 65.9907, 66.0798, 66.1103, 66.1244, 66.1479, 66.1502, 66.1244, 66.0798, 66.0493, 66.054,
        66.0587, 66.0493, 66.0798, 66.108, 66.1385, 66.115, 66.0892, 66.0446, 66.0141, 66.0259, 66.0704, 66.115, 66.1549, 66.1925, 66.1831, 66.1737, 66.1549, 66.1643, 66.1596, 66.1643, 66.1784,
        66.1643, 66.1549, 66.1432, 66.115, 66.0986, 66.0446, 66.0094, 66.0094, 66.0165, 66.0094, 65.9648, 65.8921, 65.8686, 65.8686, 65.8616, 65.8733, 65.8921, 65.9273, 65.9601, 65.9648, 65.9508,
        65.9508, 65.9226, 65.878, 65.8475, 65.8287, 65.8193, 65.8029, 65.756, 65.6973, 65.648, 65.5847, 65.5307, 65.526, 65.5166, 65.4837, 65.4485, 65.4063, 65.3711, 65.3148, 65.2702, 65.2279, 65.1904,
        65.1716, 65.1575, 65.1528, 65.1481, 65.1435, 65.1223, 65.0942, 65.0895, 65.0331, 65.0285, 65.0003, 64.9627, 64.9393, 64.9299, 64.9252, 64.9299, 64.9393, 64.89, 64.8994, 64.9487, 65.005,
        65.0519, 65.1622, 65.2232, 65.2819, 65.3195, 65.3828, 65.4251, 65.5025, 65.5354, 65.5354, 65.5354, 65.4837, 65.4251, 65.3476, 65.2655, 65.1951, 65.1411, 65.1036, 65.0989, 65.0425, 64.9815,
        64.9581, 64.9064, 64.9064, 64.8618, 64.8102, 64.7492, 64.7163, 64.7586, 64.8149, 64.8477, 64.8806, 64.9158, 65.0191, 65.0801, 65.0942, 65.1176] };
  WCL[863] = struct {
   x = [-119.996, -120.081, -120.235, -120.357, -120.437, -120.372, -120.249, -120.259, -120.315, -120.217, -120.104, -120.071],
   y = [66.3544, 66.3333, 66.3239, 66.3591, 66.4037, 66.4178, 66.4248, 66.4741, 66.5187, 66.5304, 66.5492, 66.5961] };
  WCL[864] = struct {
   x = [-140.033, -139.911, -139.747, -139.691, -139.536, -139.362, -139.226, -139.038, -138.907, -138.752, -138.635, -138.489, -138.433, -138.381, -138.311, -138.123, -137.94, -137.771, -137.64, -137.541,
        -137.419, -137.278, -137.133, -137.025, -136.889, -136.79, -136.734, -136.724, -136.635, -136.537, -136.537, -136.593, -136.429, -136.208, -136.119, -135.987, -135.786, -135.588, -135.401,
        -135.293, -135.269, -135.171, -135.227, -135.326, -135.424, -135.443, -135.48, -135.556, -135.434, -135.302, -135.157, -134.992, -134.828, -134.697, -134.664, -134.575, -134.41, -134.288,
        -134.241, -134.166, -134.19, -134.274, -134.33, -134.41, -134.509, -134.598, -134.65, -134.65, -134.565, -134.462, -134.298, -134.157, -134.021, -133.913, -133.8, -133.734, -133.561, -133.326,
        -133.176, -133.03, -133.007, -133.007, -132.875, -132.833, -132.744, -132.697, -132.589, -132.476, -132.467, -132.369, -132.204, -132.092, -132.059, -132.035, -132.002, -131.904, -131.749,
        -131.519, -131.51, -131.486, -131.322, -131.153, -131.021],
   y = [69.6564, 69.6564, 69.6447, 69.5954, 69.593, 69.5884, 69.5344, 69.4734, 69.3983, 69.3067, 69.2481, 69.2481, 69.2199, 69.2598, 69.2011, 69.1307, 69.0674, 69.0368, 68.9923, 68.95, 68.9571, 68.9453,
        68.9336, 68.9148, 68.8867, 68.8773, 68.8773, 68.8867, 68.8937, 68.8702, 68.9054, 68.95, 68.9172, 68.8937, 68.8655, 68.8585, 68.8022, 68.7341, 68.659, 68.6684, 68.6707, 68.6637, 68.7106,
        68.7599, 68.7904, 68.8303, 68.8632, 68.8937, 68.9148, 68.8913, 68.8702, 68.8867, 68.8749, 68.8397, 68.7482, 68.6919, 68.6426, 68.6003, 68.5557, 68.6308, 68.7224, 68.7599, 68.7998, 68.8468,
        68.9101, 68.9688, 69.0275, 69.065, 69.0838, 69.1143, 69.1964, 69.2316, 69.2645, 69.3302, 69.3255, 69.3654, 69.3795, 69.3912, 69.4194, 69.4757, 69.5508, 69.6118, 69.6541, 69.6306, 69.6329,
        69.6846, 69.7339, 69.7339, 69.7034, 69.7104, 69.7034, 69.7456, 69.7761, 69.7761, 69.7855, 69.7855, 69.8348, 69.8817, 69.9146, 69.9498, 69.938, 69.9075, 69.9709] };
  WCL[865] = struct {
   x = [-129.998, -130.13, -130.261, -130.407, -130.571, -130.735, -130.914, -131.036, -131.167, -131.308, -131.463, -131.73, -131.927, -132.124, -132.19, -132.059, -132.181, -132.345, -132.401, -132.491,
        -132.589, -132.645, -132.664, -132.753, -132.974, -132.988, -132.843, -132.678, -132.566, -132.444, -132.312, -132.214, -132.049, -131.838, -131.815, -131.885, -131.829, -131.683, -131.594,
        -131.608, -131.575, -131.477, -131.444, -131.411, -131.355, -131.289, -131.242, -131.12, -131.021, -131.036, -131.087, -131.087, -130.956, -130.815, -130.749, -130.561, -130.439, -130.407,
        -130.163],
   y = [69.8958, 69.8747, 69.8465, 69.8066, 69.7221, 69.6846, 69.6189, 69.5602, 69.5884, 69.6001, 69.5813, 69.5578, 69.5226, 69.4616, 69.403, 69.41, 69.3678, 69.3255, 69.2715, 69.2856, 69.2246, 69.1612,
        69.2176, 69.2316, 69.1119, 69.0509, 69.0603, 69.119, 69.126, 69.1589, 69.1706, 69.1824, 69.2363, 69.3302, 69.3537, 69.3842, 69.403, 69.4335, 69.4217, 69.3325, 69.356, 69.4147, 69.3443, 69.3607,
        69.4217, 69.4569, 69.4522, 69.4804, 69.4382, 69.3443, 69.2316, 69.2011, 69.2973, 69.3842, 69.4405, 69.5461, 69.6236, 69.6869, 69.7034] };
  WCL[866] = struct {
   x = [-135.06, -135.042, -135.018, -134.952, -134.905, -134.952, -135.028, -134.929, -134.83, -134.732, -134.652, -134.544, -134.497, -134.577, -134.652, -134.577, -134.521, -134.478, -134.553, -134.577,
        -134.497, -134.488, -134.497, -134.413, -134.314, -134.323, -134.258, -134.159, -134.056, -133.971, -133.882, -133.892, -133.939, -134.079, -134.201, -134.314, -134.455, -134.6, -134.718,
        -134.83, -134.938, -134.962, -134.952, -135.06, -135.042, -134.873, -134.741, -134.666, -134.751, -134.896, -134.995, -135.126, -135.215, -135.239, -135.328, -135.361, -135.314, -135.248,
        -135.159, -135.107, -135.074, -135.06],
   y = [69.3842, 69.3771, 69.3724, 69.3654, 69.3959, 69.4311, 69.4616, 69.4616, 69.4264, 69.4522, 69.4217, 69.4616, 69.5156, 69.4921, 69.5156, 69.5461, 69.5884, 69.6376, 69.6799, 69.7362, 69.7526, 69.7339,
        69.6916, 69.6681, 69.6447, 69.5884, 69.5532, 69.5297, 69.5696, 69.6024, 69.5602, 69.4921, 69.4382, 69.3795, 69.3091, 69.2363, 69.1542, 69.1073, 69.0486, 69.0509, 69.0721, 69.0509, 69.0251,
        69.0721, 69.1073, 69.0908, 69.1378, 69.2082, 69.1824, 69.1354, 69.1589, 69.2011, 69.2903, 69.3654, 69.3912, 69.4264, 69.4616, 69.4616, 69.4311, 69.3912, 69.3842, 69.3842] };
  WCL[867] = struct {
   x = [-135.955, -135.987, -135.987, -135.997, -135.997, -135.987, -135.987, -135.931, -135.832, -135.786, -135.776, -135.687, -135.579, -135.48, -135.424, -135.326, -135.246, -135.204, -135.124, -135.039,
        -135.016, -135.058, -135.124, -135.171, -135.246, -135.293, -135.344, -135.401, -135.443, -135.434, -135.401, -135.434, -135.513, -135.598, -135.645, -135.687, -135.734, -135.786, -135.842,
        -135.884, -135.931, -135.955],
   y = [69.2668, 69.2668, 69.2645, 69.2316, 69.2082, 69.1824, 69.1471, 69.126, 69.1073, 69.0603, 69.0439, 69.0509, 69.0368, 68.997, 68.9735, 68.9664, 68.9617, 68.9289, 68.9148, 68.9219, 68.9571, 68.9852,
        69.0204, 69.0368, 69.0439, 69.0721, 69.0838, 69.0908, 69.1026, 69.119, 69.1471, 69.1589, 69.1706, 69.173, 69.1894, 69.2363, 69.2668, 69.2856, 69.2903, 69.2856, 69.2715, 69.2668] };
  WCL[868] = struct {
   x = [-139.435, -139.557, -139.634, -139.712],
   y = [59.9546, 60.025, 60.086, 60.0579] };
  WCL[869] = struct {
   x = [-145.328, -145.314, -145.084, -144.952],
   y = [70.0084, 69.9897, 69.9897, 69.9826] };
  WCL[870] = struct {
   x = [-144.919, -144.797, -144.61],
   y = [70.0249, 69.9826, 69.9826] };
  WCL[871] = struct {
   x = [-142.561, -142.453, -142.462, -142.307, -142.11, -142.077, -141.974, -141.8, -141.688, -141.58, -141.5, -141.448, -141.392, -141.317, -141.27, -141.228, -141.228, -141.214, -141.162, -141.082,
        -141.026, -141.007],
   y = [70.0061, 69.9568, 69.9333, 69.9146, 69.8747, 69.8348, 69.8207, 69.8066, 69.7902, 69.7667, 69.7292, 69.6916, 69.6564, 69.6376, 69.6728, 69.6987, 69.7292, 69.7409, 69.7456, 69.7456, 69.7245, 69.7221] };
  WCL[872] = struct {
   x = [-141.005, -140.991, -140.949, -140.827, -140.639, -140.376, -140.198, -140.033],
   y = [69.7221, 69.7034, 69.6987, 69.6728, 69.6541, 69.6447, 69.6494, 69.6564] };
  WCL[873] = struct {
   x = [-141.402, -141.512, -141.643],
   y = [59.9921, 60.032, 60.0086] };
  WCL[874] = struct {
   x = [-142.084, -142.218, -142.34, -142.493, -142.615, -142.758, -142.835, -142.913, -143.023, -143.089, -143.145, -143.288, -143.467, -143.631, -143.753, -143.873, -143.974, -144.072, -144.138, -144.215,
        -144.183, -144.194, -144.293, -144.403, -144.492, -144.581, -144.635, -144.659, -144.725, -144.802, -144.912, -144.978, -144.978, -145.023, -145.077, -145.1],
   y = [59.9875, 60.0039, 60.025, 60.0367, 60.086, 60.0978, 60.1095, 60.0696, 60.1095, 60.1142, 60.0532, 60.0485, 60.0414, 60.0485, 60.0414, 60.025, 60.0039, 60.0039, 60.0156, 60.032, 60.0649, 60.0978,
        60.1682, 60.201, 60.1963, 60.2245, 60.2339, 60.1963, 60.2174, 60.2738, 60.2996, 60.3231, 60.3981, 60.4592, 60.485, 60.492] };
  WCL[875] = struct {
   x = [-147.815, -147.792, -147.771, -147.682, -147.55, -147.473, -147.395, -147.318, -147.285, -147.264, -147.252, -147.196, -147.208, -147.086, -147.064, -147.219, -147.341, -147.473, -147.485],
   y = [60.0156, 60.0485, 60.0532, 60.086, 60.1306, 60.1963, 60.2339, 60.2832, 60.3606, 60.3935, 60.377, 60.3489, 60.3113, 60.3277, 60.2949, 60.2081, 60.1306, 60.0579, 60.0039] };
  WCL[876] = struct {
   x = [-147.837, -147.827, -147.815],
   y = [59.9757, 60.0039, 60.0156] };
  WCL[877] = struct {
   x = [-146.811, -146.856, -146.832, -146.823, -146.778, -146.689, -146.611, -146.558, -146.501, -146.379, -146.313, -146.259, -146.281, -146.48, -146.611, -146.668, -146.656, -146.722, -146.778, -146.811],
   y = [60.3864, 60.3606, 60.3817, 60.4474, 60.492, 60.5131, 60.485, 60.4967, 60.4427, 60.4427, 60.4474, 60.4263, 60.3864, 60.3489, 60.3113, 60.3066, 60.3559, 60.3723, 60.3606, 60.3864] };
  WCL[878] = struct {
   x = [-145.154, -145.199, -145.255, -145.255, -145.288, -145.288, -145.431, -145.541, -145.628, -145.729, -145.816, -145.884, -145.917, -145.95, -145.926, -145.86, -145.816, -145.851, -145.926, -146.015,
        -146.06, -146.114, -146.269, -146.424, -146.346, -146.17, -146.248, -146.424, -146.567, -146.71, -146.832, -146.787, -146.623, -146.501, -146.346, -146.358, -146.48, -146.668, -146.787,
        -146.787, -146.778, -146.745, -146.734, -146.457, -146.435, -146.656, -146.877, -146.999, -147.163, -147.285, -147.384, -147.461, -147.571, -147.717, -147.717, -147.792, -147.827, -147.827,
        -147.893, -148.048, -148.146, -148.09, -147.914, -147.914, -148.102, -148.203, -148.268, -148.289, -148.301, -148.322, -148.4, -148.566, -148.433, -148.379, -148.477, -148.698, -148.766,
        -148.686, -148.698, -148.51, -148.51, -148.4, -148.379, -148.388, -148.555, -148.742, -148.71, -148.51, -148.4, -148.235, -148.146, -148.158, -148.191, -148.278, -148.322, -148.423, -148.51,
        -148.468, -148.388, -148.355, -148.289, -148.367, -148.489, -148.566, -148.611, -148.644, -148.742],
   y = [60.492, 60.485, 60.4592, 60.4146, 60.3817, 60.3489, 60.3395, 60.3723, 60.4193, 60.4756, 60.5178, 60.5296, 60.5343, 60.546, 60.5507, 60.6047, 60.654, 60.6751, 60.6704, 60.6704, 60.6915, 60.7032,
        60.6798, 60.6586, 60.7079, 60.8159, 60.8323, 60.783, 60.7126, 60.7173, 60.7291, 60.7455, 60.7666, 60.8088, 60.8534, 60.9004, 60.891, 60.8464, 60.8628, 60.9332, 60.9708, 61.013, 61.0787,
        61.1045, 61.1421, 61.1421, 61.0928, 61.0036, 60.9825, 60.9989, 60.9825, 60.9825, 60.9755, 60.9825, 61.0506, 61.1726, 61.0787, 60.9379, 60.8957, 60.891, 60.9379, 61.0294, 61.1796, 61.2688,
        61.1726, 61.067, 61.0623, 61.0999, 61.121, 61.1304, 61.1092, 61.0083, 61.0036, 60.9919, 60.8957, 60.8628, 60.7995, 60.7877, 60.7291, 60.7455, 60.6962, 60.7408, 60.7408, 60.654, 60.5882,
        60.5624, 60.485, 60.5178, 60.5178, 60.6047, 60.539, 60.492, 60.4474, 60.4427, 60.4028, 60.3606, 60.2667, 60.2667, 60.2785, 60.2503, 60.2409, 60.1682, 60.1423, 60.1517, 60.0696, 60.0203,
        60.0156] };
  WCL[879] = struct {
   x = [-149.195, -149.294, -149.395],
   y = [59.9757, 60.0086, 60.0086] };
  WCL[880] = struct {
   x = [-149.472, -149.538, -149.547],
   y = [59.9875, 60.1024, 60.0156] };
  WCL[881] = struct {
   x = [-150.045, -149.913, -149.714, -149.482, -149.24, -149.13, -149.195, -149.395, -149.515, -149.714, -149.881, -149.979],
   y = [60.9379, 60.9614, 60.945, 60.9168, 60.8957, 60.8957, 60.9238, 60.9543, 60.9919, 61.0248, 61.0412, 61.0459] };
  WCL[882] = struct {
   x = [-150.033, -149.967, -149.869, -149.702, -149.63, -149.559, -149.641, -149.78],
   y = [61.2266, 61.2524, 61.3369, 61.4073, 61.4378, 61.5223, 61.5176, 61.4495] };
  WCL[883] = struct {
   x = [-160.073, -159.876, -159.679, -159.524, -159.524, -159.632, -159.876],
   y = [66.4553, 66.4647, 66.4694, 66.5093, 66.5656, 66.5891, 66.5562] };
  WCL[884] = struct {
   x = [-160.106, -159.918, -159.801, -159.679, -159.589, -159.744, -159.876],
   y = [66.575, 66.6149, 66.6196, 66.6407, 66.683, 66.6783, 66.6924] };
  WCL[885] = struct {
   x = [-151.8, -151.779, -151.678, -151.535, -151.458, -151.413, -151.425, -151.481, -151.404, -151.359, -151.228, -151.049, -150.885, -150.73, -150.596, -150.477, -150.355, -150.221, -150.045],
   y = [59.9476, 60.0156, 60.1189, 60.201, 60.2832, 60.4357, 60.5718, 60.6915, 60.7619, 60.7666, 60.7995, 60.8417, 60.8863, 60.9614, 61.0036, 61.0201, 60.9121, 60.8957, 60.9379] };
  WCL[886] = struct {
   x = [-149.979, -150.122, -150.134, -150.033],
   y = [61.0459, 61.1092, 61.2008, 61.2266] };
  WCL[887] = struct {
   x = [-149.857, -150.021, -150.209, -150.373, -150.585, -150.683, -150.66, -150.603, -150.495, -150.383, -150.284, -150.261, -150.219, -150.195, -150.219, -150.308, -150.35, -150.35, -150.383, -150.472,
        -150.538, -150.627, -150.735, -150.801, -150.838, -150.801, -150.812, -150.988, -151.122, -151.244, -151.465, -151.629, -151.861, -151.894, -152.005, -152.115, -152.204, -152.249, -152.368,
        -152.446, -152.591, -152.69, -152.777, -152.922, -153.042, -153.164, -153.032, -152.889, -152.788, -152.744, -152.755],
   y = [61.4495, 61.365, 61.2641, 61.2477, 61.2477, 61.3017, 61.4002, 61.5223, 61.6232, 61.7171, 61.8321, 61.9611, 62.0949, 62.1888, 62.2662, 62.285, 62.1888, 62.1113, 62.0339, 61.9189, 61.8156, 61.7124,
        61.6232, 61.5434, 61.4378, 61.358, 61.297, 61.2008, 61.1468, 61.0787, 61.0294, 60.9825, 60.9238, 60.7995, 60.7337, 60.6704, 60.6094, 60.5882, 60.4099, 60.3606, 60.2996, 60.2456, 60.2456,
        60.2503, 60.316, 60.2996, 60.2409, 60.1916, 60.1423, 60.086, 60.0203] };
  WCL[888] = struct {
   x = [-164.551, -164.706, -164.706, -164.663, -164.584, -164.485, -164.386, -164.33],
   y = [63.0501, 63.0102, 63.0243, 63.029, 63.0454, 63.0454, 63.029, 63.0055] };
  WCL[889] = struct {
   x = [-163.757, -163.734, -163.668, -163.57, -163.481, -163.382, -163.185, -163.03, -162.917, -162.776, -162.612, -162.457, -162.359, -162.279, -162.105, -162.082, -162.18, -162.27, -162.204, -162.04,
        -161.871, -161.697, -161.575, -161.364, -161.223, -161.077, -160.946, -160.847, -160.791, -160.791, -160.768, -160.791, -160.824, -160.946, -160.913, -160.899, -160.955, -161.087, -161.176,
        -161.34, -161.443, -161.486, -161.476, -161.321, -161.167, -161.068, -160.937, -160.824, -160.749, -160.749, -160.833, -160.955, -161.045, -161.152, -161.265, -161.378, -161.476, -161.598,
        -161.763, -161.73, -161.894, -162.115, -162.223, -162.443, -162.598, -162.711, -162.776, -162.833, -162.852, -163.021, -163.129, -163.241, -163.391, -163.34, -163.218, -163.053, -163.072,
        -163.119, -163.185, -163.34, -163.57, -163.635, -163.692, -163.814, -163.87, -163.936, -164.091, -164.307, -164.494, -164.649, -164.87, -165.058, -165.227, -165.414, -165.611, -165.865,
        -166.039, -166.175, -166.217, -166.24, -166.33, -166.447, -166.513, -166.47, -166.48, -166.682, -166.79, -166.912, -166.855, -166.804, -166.822, -166.649, -166.503, -166.348, -166.193,
        -166.447, -166.682, -166.869, -166.78, -166.837, -167.123, -167.362, -167.583, -167.827, -168.024, -168.137, -168.137, -168.024, -167.916, -167.958, -168.038, -167.94, -167.827, -167.705,
        -167.517, -167.55, -167.55, -167.484, -167.311, -167.207, -167.024, -166.935, -166.912, -166.837, -166.733, -166.527, -166.306, -166.142, -166.118, -166.039, -165.832, -165.644, -165.555,
        -165.677, -165.841, -165.654, -165.499, -165.424, -165.344, -165.227, -165.072, -164.992, -164.828, -164.706, -164.565, -164.41, -164.208, -164.002, -163.8, -163.659, -163.701, -163.945,
        -164.002, -163.823, -163.781, -163.823, -163.8, -163.889, -164.1, -164.091, -163.889, -163.823, -163.823, -163.659, -163.438, -163.227, -163.021, -162.931, -162.866, -162.72, -162.744,
        -162.664, -162.622, -162.523, -162.392, -162.293, -162.26, -162.091, -161.993, -161.95, -161.828, -161.763, -161.763, -161.673, -161.584, -161.509, -161.443, -161.331, -161.209, -161.143,
        -161.101, -161.12, -160.955, -160.946, -161.068, -161.298, -161.551, -161.772, -161.861, -161.819, -161.96, -161.96, -161.871, -161.969, -162.082, -162.016, -161.903, -161.73, -161.561,
        -161.411, -161.275, -161.199, -161.167, -161.045, -160.988, -160.866, -160.782, -160.66, -160.425, -160.294, -160.218, -160.073],
   y = [63.0008, 63.0055, 63.029, 63.0595, 63.0994, 63.0642, 63.036, 63.0642, 63.1346, 63.1885, 63.2636, 63.3622, 63.4608, 63.4913, 63.496, 63.4749, 63.4209, 63.388, 63.388, 63.4209, 63.4373, 63.4209,
        63.4608, 63.4561, 63.4795, 63.5406, 63.6227, 63.7001, 63.7588, 63.8081, 63.8668, 63.9231, 63.9865, 64.0264, 64.0592, 64.1273, 64.1742, 64.2845, 64.3854, 64.4183, 64.376, 64.4324, 64.5168,
        64.5168, 64.4793, 64.5028, 64.5122, 64.5826, 64.6647, 64.7726, 64.7914, 64.8618, 64.8994, 64.8665, 64.8384, 64.7773, 64.768, 64.7961, 64.8149, 64.7586, 64.7351, 64.6882, 64.6178, 64.5262,
        64.4089, 64.2845, 64.2845, 64.423, 64.5075, 64.5591, 64.6318, 64.6271, 64.6225, 64.5779, 64.5309, 64.5122, 64.4371, 64.4042, 64.4417, 64.5122, 64.5591, 64.5732, 64.5732, 64.5591, 64.5544,
        64.5356, 64.5403, 64.5638, 64.5262, 64.4934, 64.4417, 64.4887, 64.5122, 64.5075, 64.5262, 64.5403, 64.5638, 64.5685, 64.5872, 64.6084, 64.6225, 64.6835, 64.7304, 64.829, 64.9158, 65.005,
        65.0942, 65.1951, 65.2655, 65.2326, 65.1669, 65.1411, 65.242, 65.2772, 65.296, 65.3195, 65.3617, 65.3828, 65.3335, 65.3195, 65.3711, 65.3922, 65.4626, 65.5354, 65.58, 65.6246, 65.6926, 65.7302,
        65.7442, 65.702, 65.648, 65.6527, 65.6856, 65.7161, 65.7536, 65.7654, 65.8029, 65.8522, 65.9062, 65.8569, 65.8968, 65.9601, 66.0047, 65.9813, 66.054, 66.1103, 66.1643, 66.1878, 66.1432,
        66.1103, 66.1244, 66.1244, 66.1432, 66.2183, 66.2676, 66.3544, 66.3943, 66.4225, 66.4084, 66.4389, 66.4436, 66.4553, 66.5351, 66.5914, 66.5961, 66.5914, 66.6008, 66.6266, 66.6266, 66.6149,
        66.6008, 66.6008, 66.5844, 66.5398, 66.4741, 66.3826, 66.3028, 66.2676, 66.2089, 66.1995, 66.2042, 66.1925, 66.1244, 66.0704, 66.0892, 66.0845, 66.0704, 66.0493, 66.0939, 66.1033, 66.0587,
        66.0845, 66.054, 66.0611, 66.0704, 66.0399, 66.0611, 66.0751, 66.0306, 65.9813, 65.9648, 66, 66.054, 66.1103, 66.2089, 66.2793, 66.2723, 66.2488, 66.2277, 66.223, 66.1737, 66.1338, 66.2136,
        66.2746, 66.3333, 66.3685, 66.399, 66.4084, 66.3286, 66.2934, 66.3145, 66.3943, 66.4741, 66.5492, 66.6759, 66.636, 66.5398, 66.4788, 66.4647, 66.4741, 66.5234, 66.5304, 66.4882, 66.4788,
        66.4483, 66.4248, 66.3732, 66.3638, 66.3685, 66.3826, 66.4248, 66.4553] };
  WCL[890] = struct {
   x = [-159.873, -160.052, -160.094, -160.193, -160.103],
   y = [66.5562, 66.5351, 66.4952, 66.514, 66.575] };
  WCL[891] = struct {
   x = [-159.873, -160.005, -160.169, -160.225, -160.291, -160.338, -160.47, -160.568, -160.69, -160.854, -161.019, -161.15, -161.197, -161.263, -161.239, -161.305, -161.305, -161.352, -161.516, -161.648,
        -161.793, -161.882, -161.826, -161.77, -161.596, -161.526, -161.671, -161.803, -161.948, -162.023, -162.122, -162.309, -162.497, -162.577, -162.497, -162.422, -162.356, -162.422, -162.53,
        -162.53, -162.497, -162.511, -162.643, -162.774, -162.807, -162.882, -162.83, -163.051, -163.258, -163.412, -163.68, -163.755, -163.779, -163.746, -163.821, -163.877, -163.943, -164.107,
        -164.107, -164.187, -164.206, -164.375, -164.628, -164.816, -164.99, -165.168, -165.365, -165.586, -165.806, -165.985, -166.093, -166.28, -166.459, -166.614, -166.745, -166.886, -166.844,
        -166.698, -166.581, -166.51, -166.581, -166.698, -166.712, -166.599, -166.524, -166.426, -166.28, -166.172, -166.257, -166.337, -166.313, -166.304, -166.304, -166.28, -166.257, -166.271],
   y = [66.6924, 66.6525, 66.6712, 66.6454, 66.6149, 66.6008, 66.5891, 66.575, 66.6149, 66.6525, 66.6454, 66.6454, 66.6055, 66.6196, 66.6924, 66.6619, 66.5797, 66.5398, 66.5797, 66.622, 66.7111, 66.7628,
        66.8238, 66.8965, 66.9224, 66.9576, 67.0115, 67.035, 67.042, 67.0373, 67.0303, 67.0092, 66.9622, 66.9904, 67.0514, 67.1078, 67.1617, 67.1664, 67.15, 67.1359, 67.1078, 67.0819, 67.089, 67.0937,
        67.0373, 67.0045, 66.9693, 66.9998, 67.0373, 67.0819, 67.1078, 67.2181, 67.3025, 67.3166, 67.3542, 67.448, 67.5067, 67.5607, 67.5935, 67.6194, 67.6311, 67.6992, 67.7578, 67.8118, 67.8822,
        67.9385, 68.0066, 68.0301, 68.0793, 68.1216, 68.1943, 68.2671, 68.3046, 68.3258, 68.3398, 68.3657, 68.4056, 68.4431, 68.4736, 68.4853, 68.4267, 68.3938, 68.3844, 68.3821, 68.4267, 68.4501,
        68.4713, 68.5112, 68.4947, 68.5065, 68.5581, 68.6262, 68.7341, 68.8116, 68.8468, 68.8773] };
  WCL[892] = struct {
   x = [-166.271, -166.271, -166.205, -166.06, -165.872, -165.863, -166.017, -165.816, -165.595, -165.323, -165.088, -164.914, -164.694, -164.483, -164.286, -164.018, -163.952, -163.811, -163.567, -163.469,
        -163.337, -163.29, -163.225, -163.215, -163.159, -163.07, -163.018, -163.037, -162.882, -162.685, -162.619, -162.83, -163.037, -163.018, -162.896, -162.694, -162.399, -162.333, -162.455,
        -162.577, -162.751, -162.929, -162.798, -162.694, -162.53],
   y = [68.8773, 68.8913, 68.8913, 68.8913, 68.9031, 68.882, 68.8538, 68.8468, 68.8749, 68.9054, 68.9148, 68.9336, 68.9336, 68.9547, 68.9735, 68.9923, 69.0157, 69.0674, 69.1471, 69.2293, 69.3067, 69.3724,
        69.464, 69.356, 69.3865, 69.4992, 69.5837, 69.6447, 69.5954, 69.5508, 69.5766, 69.6189, 69.7104, 69.7339, 69.7479, 69.7479, 69.7479, 69.7644, 69.7714, 69.7761, 69.7761, 69.8019, 69.8582,
        69.9122, 69.9709] };
  WCL[893] = struct {
   x = [-162.291, -162.211, -162.07, -161.868, -161.695, -161.704, -161.901],
   y = [70.0178, 69.9803, 69.9521, 69.9216, 69.8888, 69.9263, 69.9568] };
  WCL[894] = struct {
   x = [-162.467, -162.401, -162.368, -162.326, -162.27, -162.223, -162.148, -162.058, -162.016, -162.049, -162.223, -162.345, -162.424, -162.514, -162.532, -162.523, -162.523, -162.467],
   y = [66.9317, 66.9083, 66.9317, 66.9576, 66.9576, 66.8965, 66.8402, 66.7933, 66.7534, 66.7229, 66.737, 66.7416, 66.7698, 66.8308, 66.866, 66.8965, 66.9317, 66.9317] };
  WCL[895] = struct {
   x = [-169.833, -169.833, -169.833, -169.885],
   y = [63.036, 63.036, 63.0806, 63.1158] };
  WCL[896] = struct {
   x = [-170.075, -169.977, -169.845, -169.737, -169.648, -169.601, -169.568, -169.46, -169.371, -169.296, -169.24, -169.174, -169.019, -168.92, -168.798, -168.775, -168.798, -168.798, -168.841, -168.986,
        -169.118, -169.249, -169.371, -169.437, -169.493, -169.549, -169.568, -169.714],
   y = [63.4913, 63.4655, 63.4514, 63.4209, 63.3974, 63.3528, 63.3434, 63.3387, 63.3223, 63.3035, 63.3082, 63.3317, 63.3434, 63.3129, 63.3082, 63.2683, 63.2378, 63.2144, 63.1791, 63.1533, 63.1698, 63.1885,
        63.1838, 63.1346, 63.1087, 63.0759, 63.0454, 63.0454] };
  WCL[897] = struct {
   x = [-167.566, -167.543, -167.4, -167.257, -167.069, -166.881, -166.738, -166.616, -166.517, -166.395, -166.285, -166.163, -165.987, -165.856, -165.766, -165.722, -165.733, -165.745, -165.668, -165.656,
        -165.689],
   y = [60.262, 60.2667, 60.262, 60.2503, 60.2667, 60.2738, 60.3277, 60.3606, 60.3935, 60.4028, 60.3935, 60.3489, 60.3395, 60.3324, 60.3231, 60.3066, 60.2174, 60.086, 60.0978, 60.0414, 60.0039] };
  WCL[898] = struct {
   x = [-167.212, -167.322, -167.411, -167.477, -167.489, -167.566, -167.566],
   y = [59.9992, 60.0367, 60.0813, 60.1189, 60.1846, 60.2245, 60.262] };
  WCL[899] = struct {
   x = [-165.379, -165.37, -165.292, -165.191, -165.048, -165.004, -164.86, -164.663, -164.408, -164.321, -164.321, -164.44, -164.628, -164.738, -164.849, -164.938, -165.048, -165.191, -165.292, -165.379],
   y = [60.6704, 60.7032, 60.7291, 60.7572, 60.8042, 60.8581, 60.837, 60.8088, 60.7619, 60.6798, 60.6211, 60.539, 60.4357, 60.3817, 60.4639, 60.5953, 60.6211, 60.654, 60.654, 60.6704] };
  WCL[900] = struct {
   x = [-164.33, -164.232, -164.297, -164.321, -164.368, -164.476, -164.598, -164.738, -164.842, -164.907, -164.964, -164.973, -164.973, -164.964, -164.551],
   y = [63.0055, 62.975, 62.8999, 62.8295, 62.7802, 62.7802, 62.799, 62.7943, 62.7403, 62.7027, 62.7332, 62.7802, 62.8295, 62.8811, 63.0501] };
  WCL[901] = struct {
   x = [-162.19, -162.277, -162.288, -162.345, -162.443, -162.476, -162.443, -162.333, -162.157, -161.979, -161.803, -161.671, -161.561, -161.561, -161.57, -161.638, -161.803, -161.913, -162.047, -162.244,
        -162.345, -162.431, -162.575, -162.652, -162.586, -162.455, -162.431, -162.542, -162.575],
   y = [59.9828, 60.0649, 60.147, 60.2409, 60.2785, 60.316, 60.3653, 60.4474, 60.5718, 60.654, 60.7408, 60.8042, 60.8464, 60.8534, 60.8581, 60.8464, 60.8417, 60.7995, 60.7126, 60.654, 60.6094, 60.5296,
        60.4099, 60.3559, 60.2573, 60.2127, 60.1635, 60.1423, 60.0813] };
  WCL[902] = struct {
   x = [-164.067, -164.077, -164.011, -164.056, -164.199, -164.33, -164.419, -164.497, -164.586, -164.586, -164.541, -164.44, -164.342, -164.232, -164.166, -164.121, -163.966, -163.811, -163.725, -163.779,
        -163.945, -164.056, -164.154, -164.297, -164.408, -164.551, -164.696, -164.774, -164.673, -164.607, -164.738, -164.872, -164.938, -164.961, -165.027, -165.126, -165.203, -165.292, -165.37,
        -165.391, -165.447, -165.468, -165.501, -165.468, -165.391, -165.44, -165.496, -165.616, -165.726, -165.87, -165.865, -165.804, -165.726, -165.626, -165.532, -165.576, -165.515, -165.527,
        -165.621, -165.698, -165.748, -165.813, -165.881, -165.952, -165.996, -166.029, -166.025, -166.001, -165.952, -165.902, -165.874, -165.902, -165.952, -165.975, -165.975, -165.947, -165.891,
        -165.841, -165.804, -165.78, -165.787, -165.792, -165.792, -165.78, -165.764, -165.748, -165.719, -165.698, -165.687, -165.654, -165.626, -165.621, -165.588, -165.548, -165.506, -165.466,
        -165.405, -165.356, -165.339, -165.323, -165.269, -165.212, -165.151, -165.114, -165.086, -165.053, -165.004, -164.952, -164.919, -164.943, -164.964, -164.952, -164.919, -164.76],
   y = [59.9757, 60.0039, 60.032, 60.0766, 60.0532, 60.0086, 60.025, 60.0931, 60.1963, 60.2738, 60.3489, 60.431, 60.485, 60.5554, 60.6258, 60.654, 60.6915, 60.7079, 60.7713, 60.8628, 60.8581, 60.8253,
        60.8159, 60.8206, 60.8417, 60.8534, 60.8863, 60.9004, 60.9755, 61.0341, 61.0201, 61.0201, 61.0341, 61.1139, 61.0553, 60.9919, 60.9755, 61.0036, 61.0553, 61.1257, 61.1679, 61.2102, 61.2688,
        61.3369, 61.3956, 61.4073, 61.3697, 61.3228, 61.3486, 61.4002, 61.4495, 61.5176, 61.5481, 61.5551, 61.5763, 61.6068, 61.6279, 61.6654, 61.6584, 61.6373, 61.5903, 61.5645, 61.5434, 61.5481,
        61.5387, 61.5129, 61.5763, 61.6373, 61.6537, 61.6584, 61.6959, 61.7218, 61.7429, 61.7687, 61.8156, 61.8368, 61.825, 61.8415, 61.8461, 61.8884, 61.9259, 61.9776, 62.0339, 62.0855, 62.1278,
        62.1677, 62.2287, 62.2709, 62.3226, 62.3625, 62.3742, 62.3836, 62.4235, 62.4704, 62.5009, 62.5525, 62.5666, 62.5408, 62.5877, 62.6183, 62.6675, 62.7286, 62.7544, 62.7544, 62.7497, 62.7403,
        62.7286, 62.7239, 62.7638, 62.8342, 62.8647, 62.8999, 62.9445, 63.0008] };
  WCL[903] = struct {
   x = [-170.099, -169.944, -169.869, -169.69, -169.779, -169.958],
   y = [66.1596, 66.1103, 66.1338, 66.115, 66.0306, 66.0493] };
  WCL[904] = struct {
   x = [-169.887, -170.024, -170.197, -170.343, -170.474, -170.582, -170.685, -170.817, -170.958, -171.127, -171.244, -171.38, -171.511, -171.666, -171.788, -171.854, -171.906, -171.939, -171.906, -171.854,
        -171.821, -171.831, -171.741, -171.652, -171.432, -171.192, -171.094, -171.014, -170.892, -170.817, -170.695, -170.606, -170.516, -170.441, -170.329, -170.254, -170.141, -170.075, -170.066,
        -170.075],
   y = [63.1158, 63.1393, 63.1604, 63.1932, 63.2496, 63.2895, 63.327, 63.3575, 63.3763, 63.3833, 63.3763, 63.327, 63.3082, 63.3387, 63.3833, 63.4303, 63.4608, 63.5054, 63.5546, 63.6227, 63.726, 63.7846,
        63.7447, 63.6955, 63.6814, 63.6415, 63.6133, 63.5922, 63.6086, 63.6415, 63.6509, 63.6908, 63.7001, 63.6908, 63.672, 63.6368, 63.6227, 63.564, 63.5194, 63.4913] };
  WCL[905] = struct {
   x = [-172.931, -172.91, -172.91, -172.92, -172.866, -172.788, -172.699, -172.6, -172.544, -172.556, -172.687, -172.776, -172.866, -172.976, -172.985, -172.997, -172.952, -172.931],
   y = [60.546, 60.5554, 60.5507, 60.5014, 60.4756, 60.4427, 60.4099, 60.4028, 60.377, 60.3723, 60.3653, 60.3981, 60.4357, 60.5014, 60.5225, 60.5225, 60.546, 60.546] };
  WCL[906] = struct {
   x = [179.944, 180],
   y = [68.9219, 68.9078] };
  WCL[907] = struct {
   x = [-180, -179.85, -179.681, -179.615, -179.564, -179.474, -179.32, -179.188, -179.043, -178.921, -178.822, -178.78, -178.766, -178.681, -178.601, -178.493, -178.423, -178.282, -178.24, -178.371,
        -178.404, -178.193, -178.104, -178.085, -178.127, -178.104, -177.949, -177.752, -177.564, -177.545, -177.531, -177.39, -177.301, -177.245, -177.254, -177.278, -177.137, -176.982, -176.935,
        -176.883, -176.982, -177.071, -176.869, -176.738, -176.541, -176.339, -176.175, -176.156, -176.198, -176.208, -176.386, -176.485, -176.395, -176.297, -176.133, -175.846, -175.715, -175.579,
        -175.438, -175.457, -175.546, -175.626, -175.734, -175.813, -175.888, -175.935, -175.954, -175.879, -175.766, -175.626, -175.471, -175.405, -175.391, -175.316, -175.161, -174.983, -174.865,
        -174.809, -174.842, -174.875, -174.997, -175.039, -174.964, -174.917, -174.865, -174.832, -174.696, -174.687, -174.729, -174.645, -174.49, -174.41, -174.433, -174.391, -174.335, -174.222,
        -174.114, -173.983, -173.983, -173.992, -174.025, -174.034, -173.959, -173.828, -173.795, -173.795, -173.894, -173.969, -174.034, -174.171, -174.147, -174.081, -174.081, -174.067, -174.114,
        -174.189, -174.335, -174.203, -174.222, -174.443, -174.49, -174.565, -174.377, -174.171, -174.049, -173.936, -173.894, -173.828, -173.781, -173.748, -173.772, -173.626, -173.584, -173.485,
        -173.396, -173.274, -173.429, -173.485, -173.387, -173.321, -173.241, -173.175, -173.218, -173.241, -173.199, -173.068, -172.899, -172.692, -172.504, -172.425, -172.626, -172.823, -172.659,
        -172.448, -172.251, -172.049, -171.932, -171.763, -171.73, -171.664, -171.467, -171.458, -171.336, -171.049, -170.946, -170.725, -170.519, -170.528, -170.65, -170.561, -170.519, -170.341,
        -170.186, -170.233, -170.101],
   y = [68.9077, 68.8702, 68.8749, 68.8749, 68.8585, 68.8139, 68.7951, 68.7623, 68.7435, 68.7435, 68.659, 68.5792, 68.5276, 68.49, 68.4666, 68.4666, 68.4454, 68.422, 68.4173, 68.4947, 68.5112, 68.4384,
        68.3774, 68.3375, 68.2812, 68.2436, 68.2718, 68.2952, 68.2483, 68.2436, 68.2108, 68.2272, 68.1849, 68.1544, 68.1497, 68.1075, 68.1075, 68.1075, 68.0418, 68.0066, 67.9925, 67.9432, 67.9198,
        67.8775, 67.8822, 67.8658, 67.8517, 67.8494, 67.8118, 67.7437, 67.7226, 67.6851, 67.6733, 67.6405, 67.6264, 67.5466, 67.5091, 67.4645, 67.4387, 67.502, 67.556, 67.5513, 67.6029, 67.6522,
        67.6851, 67.7226, 67.7578, 67.7743, 67.7531, 67.7273, 67.7062, 67.6851, 67.6194, 67.5302, 67.4809, 67.448, 67.4011, 67.3401, 67.3072, 67.3847, 67.4387, 67.3847, 67.3237, 67.2439, 67.1664,
        67.0561, 66.9435, 66.866, 66.7745, 66.7229, 66.683, 66.6619, 66.6196, 66.5445, 66.5398, 66.5609, 66.5445, 66.5234, 66.4952, 66.4131, 66.3826, 66.3286, 66.3826, 66.3873, 66.46, 66.5304, 66.6196,
        66.6454, 66.6102, 66.6266, 66.6924, 66.8003, 66.9083, 66.9787, 66.9951, 67.0092, 67.0209, 67.042, 67.0725, 67.0514, 66.9998, 67.0561, 67.089, 67.0984, 67.0819, 67.0679, 67.0256, 67.0092,
        67.0514, 67.089, 67.1406, 67.1242, 67.0725, 67.0819, 67.0632, 67.0373, 67.0115, 66.9482, 66.8918, 66.8238, 66.8238, 66.8825, 66.9388, 66.9881, 66.9881, 66.9482, 66.9059, 66.8825, 66.8918,
        66.9482, 66.9646, 66.9904, 67.0115, 67.0092, 66.9881, 66.9787, 66.9529, 66.9482, 66.8707, 66.8097, 66.7698, 66.7111, 66.6266, 66.5398, 66.453, 66.399, 66.3192, 66.284, 66.2488, 66.2042,
        66.2324, 66.2934, 66.2277, 66.1737, 66.1596] };
  WCL[908] = struct {
   x = [-169.955, -170.129, -170.293, -170.415, -170.528, -170.561, -170.594, -170.561, -170.617, -170.847, -170.979, -171.166, -171.354, -171.429, -171.378, -171.101, -171.035, -171.11, -171.331, -171.631,
        -171.885, -171.993, -172.157, -172.227, -172.18, -172.391, -172.654, -172.457, -172.269, -172.157, -172.058, -172.237, -172.481, -172.645, -172.645, -172.636, -172.711, -172.941, -173.086,
        -172.842, -172.734, -172.776, -172.973, -172.964, -172.856, -172.842, -172.687, -172.621, -172.514, -172.514, -172.391, -172.345, -172.579, -172.734, -172.898, -173.063, -173.039, -172.875,
        -172.866, -172.997, -173.11, -173.25, -173.396, -173.26, -173.227, -173.33, -173.34, -173.415, -173.57, -173.771, -173.734, -173.814, -174.034, -174.034, -174.245, -174.189, -174.297, -174.485,
        -174.663, -174.785, -174.64, -174.607, -174.64, -174.705],
   y = [66.0493, 66.0399, 65.9648, 65.9367, 65.9226, 65.8381, 65.756, 65.6856, 65.6081, 65.6246, 65.6973, 65.7442, 65.8428, 65.7841, 65.6668, 65.594, 65.526, 65.479, 65.4837, 65.526, 65.5072, 65.479,
        65.4204, 65.3335, 65.2514, 65.242, 65.242, 65.2045, 65.2139, 65.1129, 65.0566, 64.9956, 64.9064, 64.9581, 64.9299, 64.8853, 64.8524, 64.8571, 64.7961, 64.8243, 64.7726, 64.7257, 64.721,
        64.6788, 64.6553, 64.6318, 64.6506, 64.6553, 64.6013, 64.5779, 64.484, 64.4324, 64.3995, 64.4793, 64.5122, 64.4934, 64.4558, 64.4464, 64.3174, 64.2563, 64.2423, 64.261, 64.3221, 64.4699,
        64.5309, 64.5826, 64.4746, 64.3854, 64.3455, 64.376, 64.4887, 64.4136, 64.4136, 64.4981, 64.5732, 64.6225, 64.6506, 64.6506, 64.7022, 64.7539, 64.8337, 64.9534, 64.9909, 64.9346] };
  WCL[909] = struct {
   x = [-174.675, -174.708, -174.698, -174.774, -174.896, -174.942, -175.05, -174.975, -174.919, -175.008, -175.196, -175.384, -175.402, -175.492, -175.703, -175.909, -175.867, -175.858, -175.956, -175.956,
        -175.933, -176.121, -176.308, -176.398, -176.52, -176.562, -176.707, -176.848, -177.078, -177.266, -177.364, -177.566, -177.721, -177.876, -178.064, -178.303, -178.491, -178.491, -178.524,
        -178.58, -178.711, -178.791, -178.89, -178.909, -178.979, -178.843, -178.721, -178.571, -178.58, -178.514, -178.547, -178.655, -178.711, -178.81, -178.941, -179.143, -179.186, -179.232,
        -179.383, -179.373, -179.35, -179.561, -179.674, -179.782, -179.937, -179.88, -179.871, -179.692, -179.538, -179.331, -179.308, -179.462, -179.495, -179.65, -179.791, -179.969],
   y = [64.9674, 64.8665, 64.8102, 64.8243, 64.8243, 64.8712, 64.7867, 64.7586, 64.7398, 64.7304, 64.768, 64.7773, 64.829, 64.8853, 64.9627, 65.0425, 65.127, 65.2373, 65.3195, 65.3664, 65.4532, 65.4884,
        65.4931, 65.5518, 65.5987, 65.594, 65.5847, 65.6034, 65.6128, 65.5987, 65.5354, 65.4978, 65.4884, 65.5072, 65.526, 65.526, 65.5659, 65.6926, 65.7442, 65.8029, 65.8076, 65.8921, 65.9648,
        66.0047, 66.0658, 66.0399, 66.0446, 66.0939, 66.1925, 66.338, 66.3544, 66.3474, 66.2723, 66.1831, 66.1596, 66.2629, 66.3427, 66.2723, 66.3286, 66.2136, 66.1479, 66.0986, 66.1549, 66.1244,
        66.1244, 66.0212, 65.8733, 65.7794, 65.7161, 65.6527, 65.5354, 65.4532, 65.3429, 65.2514, 65.1669, 65.0848] };
  WCL[910] = struct {
   x = [169.885, 170.073, 170.246, 170.448, 170.622, 170.758, 170.866, 170.997, 170.965, 170.899, 170.899, 170.791, 170.721, 170.669, 170.589, 170.336, 170.059, 170.139, 170.336, 170.467, 170.481, 170.481,
        170.415],
   y = [68.7552, 68.7599, 68.7951, 68.8468, 68.9219, 68.9805, 69.0392, 69.0908, 69.1425, 69.2316, 69.295, 69.3372, 69.4076, 69.4804, 69.5649, 69.593, 69.6118, 69.6869, 69.7174, 69.7292, 69.8137, 69.87,
        69.931] };
  WCL[911] = struct {
   x = [171.331, 171.584, 171.856, 172.068, 172.33, 172.476, 172.687, 172.95, 173.227, 173.448, 173.678, 173.856, 174.119, 174.386, 174.56, 174.738, 174.827, 174.827, 174.86, 174.912, 175.067, 175.222,
        175.409, 175.63, 175.86, 176.071, 176.259, 176.437, 176.62, 176.841, 177.099, 177.231, 177.306, 177.362, 177.639, 177.878, 178.066, 178.212, 178.301, 178.385, 178.423, 178.507, 178.719,
        178.939, 179.127, 179.301, 179.479, 179.657, 179.812, 179.944],
   y = [70.0178, 69.9991, 69.9826, 69.9568, 69.938, 69.9521, 69.9193, 69.8629, 69.8395, 69.8841, 69.8817, 69.8653, 69.8888, 69.8629, 69.8207, 69.8324, 69.8395, 69.8465, 69.8582, 69.8747, 69.87, 69.87,
        69.8653, 69.8629, 69.87, 69.8559, 69.7902, 69.7339, 69.6752, 69.6635, 69.6259, 69.6071, 69.6118, 69.6142, 69.5156, 69.4734, 69.4499, 69.4147, 69.3983, 69.4194, 69.3865, 69.3537, 69.3255,
        69.2786, 69.2316, 69.1706, 69.0908, 69.0251, 68.9688, 68.9219] };
  WCL[912] = struct {
   x = [-179.969, -180],
   y = [65.0848, 65.0738] };
  WCL[913] = struct {
   x = [180, 179.833, 179.632, 179.458, 179.19, 178.937, 178.782, 178.576, 178.552, 178.618, 178.477, 178.299, 178.209, 178.224, 178.177, 178.055, 177.937, 177.726, 177.548, 177.383, 177.393, 177.327,
        177.139, 176.886, 176.698, 176.581, 176.402, 176.393, 176.346, 176.257, 176.238, 176.435, 176.623, 176.844, 176.966, 177.153, 177.205, 177.163, 176.9, 176.712, 176.445, 176.271, 176.083,
        175.961, 175.905, 175.961, 176.191, 176.257, 176.257, 176.158, 176.036, 175.985, 176.14, 176.069, 175.905, 175.731, 175.576, 175.511, 175.665, 175.928, 176.116, 176.29, 176.346, 176.215,
        176.205, 176.28, 176.445, 176.59, 176.764, 176.952, 177.139, 177.294, 177.407, 177.374, 177.374, 177.407, 177.562, 177.759, 177.89, 178.087, 178.2, 178.322, 178.388, 178.463, 178.519, 178.519,
        178.454, 178.289, 178.299, 178.454, 178.576, 178.651, 178.684, 178.73, 178.773, 178.838, 178.984, 179.139, 179.27, 179.402, 179.303, 179.106, 178.993],
   y = [65.0738, 65.0144, 64.9064, 64.8149, 64.7867, 64.7351, 64.6459, 64.5779, 64.6225, 64.6835, 64.6647, 64.66, 64.7163, 64.7445, 64.7257, 64.6835, 64.6976, 64.7257, 64.7867, 64.8431, 64.9111, 64.9487,
        65.0144, 65.066, 65.0566, 65.066, 65.0707, 65.1364, 65.1669, 65.0801, 65.0191, 64.9768, 64.9534, 65.0097, 64.944, 64.8994, 64.8102, 64.7726, 64.7961, 64.8524, 64.8477, 64.8947, 64.9581,
        64.9815, 64.9158, 64.8571, 64.8243, 64.721, 64.6976, 64.6365, 64.5215, 64.4042, 64.3502, 64.1977, 64.1742, 64.1507, 64.0498, 64.0052, 64.0592, 64.1414, 64.1507, 64.2704, 64.3807, 64.4464,
        64.5591, 64.6318, 64.6553, 64.6013, 64.6225, 64.6741, 64.7257, 64.7257, 64.7257, 64.6365, 64.5591, 64.4605, 64.3713, 64.3408, 64.2704, 64.2282, 64.261, 64.3573, 64.261, 64.1789, 64.0733,
        64.0639, 64.0874, 64.0639, 64.017, 63.9184, 63.9442, 63.9677, 63.794, 63.6509, 63.5922, 63.4256, 63.2988, 63.205, 63.2003, 63.0688, 63.0407, 63.0501, 63.029] };
  WCL[914] = struct {
   x = [178.995, 178.944, 178.953, 178.963, 179.01, 179.108, 179.263, 179.441, 179.507, 179.451, 179.319, 179.23, 179.197, 179.075, 178.995, 178.92, 178.798, 178.643, 178.512, 178.381, 178.212, 177.958,
        177.784, 177.672, 177.606, 177.498, 177.564, 177.573, 177.475, 177.386, 177.329, 177.329, 177.273, 177.207, 177.156, 177.066, 176.987, 176.935, 176.968, 177.043, 177.099, 177.188, 177.221,
        177.156, 177.001, 176.822, 176.658, 176.513, 176.348, 176.217, 176.015, 175.893, 175.785, 175.776, 175.752, 175.663, 175.522, 175.367, 175.231, 175.165, 175.212, 175.222, 175.165, 175.067,
        175.076, 175.057, 174.978, 174.837, 174.715, 174.569, 174.429, 174.306, 174.175, 174.086, 174.02, 173.954, 173.842, 173.743, 173.654, 173.579, 173.466, 173.41, 173.391, 173.344, 173.288,
        173.288, 173.222, 173.133, 173.025, 172.936, 172.828, 172.762, 172.781, 172.847, 172.814, 172.76, 172.682, 172.57, 172.427, 172.361, 172.406, 172.384, 172.373, 172.284, 172.173, 172.119,
        172.03, 172.009, 172.096, 172.108, 172.086, 171.985, 171.908, 171.831, 171.753, 171.699, 171.589, 171.589, 171.5, 171.434, 171.368, 171.258, 171.157, 171.058, 171.004, 170.892, 170.782,
        170.671, 170.584, 170.528, 170.561, 170.561, 170.528, 170.474, 170.418, 170.373, 170.373],
   y = [63.029, 63.0266, 63.0055, 62.975, 62.9445, 62.9304, 62.8999, 62.8342, 62.7497, 62.6323, 62.5361, 62.4845, 62.4493, 62.4446, 62.3319, 62.3108, 62.3366, 62.3625, 62.4141, 62.4446, 62.515, 62.5525,
        62.5572, 62.6183, 62.684, 62.6793, 62.6276, 62.5666, 62.5666, 62.5713, 62.5971, 62.6441, 62.6488, 62.6934, 62.7239, 62.8201, 62.8694, 62.8389, 62.7638, 62.698, 62.6276, 62.6018, 62.5713,
        62.5408, 62.515, 62.5009, 62.4657, 62.4188, 62.3671, 62.3108, 62.2404, 62.2193, 62.1888, 62.1771, 62.1888, 62.1771, 62.1465, 62.1113, 62.0761, 62.0761, 62.1278, 62.1677, 62.163, 62.1536,
        62.0855, 62.0339, 61.9705, 61.9353, 61.9095, 61.8415, 61.8156, 61.8062, 61.7945, 61.7734, 61.7476, 61.7171, 61.6584, 61.6373, 61.6537, 61.7265, 61.7382, 61.6748, 61.5692, 61.5223, 61.5223,
        61.5176, 61.4542, 61.4073, 61.3744, 61.3744, 61.4167, 61.3909, 61.3533, 61.3017, 61.2594, 61.2594, 61.2266, 61.189, 61.1961, 61.1961, 61.1421, 61.0717, 61.0294, 60.9989, 61.013, 61.0341,
        61.0834, 61.0341, 60.9989, 60.9497, 60.9121, 60.9004, 60.8628, 60.8628, 60.8534, 60.8253, 60.7877, 60.7455, 60.7337, 60.6915, 60.6586, 60.5718, 60.5718, 60.6164, 60.546, 60.5343, 60.4967,
        60.4592, 60.4474, 60.431, 60.4099, 60.3559, 60.2996, 60.2503, 60.1799, 60.1189, 60.0485] };
  WCL[915] = struct {
   x = [170.089, 170.068],
   y = [59.971, 60.0203] };
  WCL[916] = struct {
   x = [159.988, 160.078, 160.209, 160.477, 160.683, 160.838, 160.993, 161.082, 161.246, 161.303, 161.279, 161.425, 161.612, 161.72, 162.021, 162.077, 162.218, 162.284, 162.382, 162.504, 162.659, 162.781,
        162.913, 163.11, 163.312, 163.476, 163.664, 163.828, 164.039, 164.095, 164.194, 164.293, 164.415, 164.569, 164.753, 164.931, 165.02],
   y = [69.7784, 69.7221, 69.7221, 69.7221, 69.6799, 69.6283, 69.5532, 69.5367, 69.5555, 69.4921, 69.3818, 69.3631, 69.4358, 69.5179, 69.5109, 69.6212, 69.6142, 69.6681, 69.694, 69.6987, 69.6822, 69.6564,
        69.6822, 69.7174, 69.7221, 69.701, 69.6987, 69.7362, 69.7784, 69.7409, 69.6799, 69.6142, 69.5907, 69.5555, 69.5672, 69.5649, 69.5602] };
  WCL[917] = struct {
   x = [164.999, 165.041, 165.088, 165.22, 165.318, 165.464, 165.604, 165.759, 165.881, 165.971, 166.144, 166.29, 166.431, 166.553, 166.698, 166.82, 166.961, 167.083, 167.172, 167.28, 167.383, 167.501,
        167.623, 167.754, 167.89, 167.989, 168.008, 167.876, 167.843, 167.998, 168.177, 168.228, 168.219, 168.219, 168.195, 168.209, 168.252, 168.374, 168.519, 168.669, 168.806, 168.979, 169.003,
        169.125, 169.289, 169.388, 169.453, 169.51, 169.664, 169.885],
   y = [69.5696, 69.5532, 69.5719, 69.593, 69.5954, 69.5813, 69.5719, 69.5719, 69.5719, 69.5508, 69.5179, 69.4921, 69.4921, 69.4921, 69.4968, 69.4757, 69.4874, 69.5297, 69.5766, 69.6306, 69.6728, 69.7174,
        69.755, 69.7714, 69.7479, 69.7174, 69.6916, 69.6681, 69.6118, 69.593, 69.5719, 69.5039, 69.4405, 69.3771, 69.3067, 69.2551, 69.2082, 69.173, 69.173, 69.1706, 69.1589, 69.119, 69.0838, 69.0603,
        69.0157, 68.9406, 68.8538, 68.7998, 68.767, 68.7552] };
  WCL[918] = struct {
   x = [168.099, 168.099],
   y = [69.9944, 69.9944] };
  WCL[919] = struct {
   x = [168.475, 168.686, 168.883, 169.127, 169.315, 169.38, 169.366, 169.258, 169.268, 169.258, 169.216, 169.047, 168.808, 168.54, 168.4, 168.231, 168.132, 167.925, 167.836, 167.893, 168.099],
   y = [70.0202, 69.9615, 69.931, 69.8841, 69.8442, 69.7784, 69.7292, 69.7104, 69.6635, 69.6118, 69.5532, 69.5391, 69.5649, 69.5954, 69.6681, 69.7104, 69.7761, 69.816, 69.8888, 69.9568, 69.9944] };
  WCL[920] = struct {
   x = [164.736, 164.558],
   y = [59.9429, 60.0367] };
  WCL[921] = struct {
   x = [161.368, 161.61, 161.821, 161.908, 161.908, 161.964, 162.173, 162.45, 162.692, 162.781, 162.913, 163.133, 163.333, 163.532, 163.708, 163.708, 163.565, 163.577, 163.61, 163.72, 163.863, 163.875,
        163.854, 163.847, 163.858, 163.896, 163.952, 163.957, 163.957, 163.969, 163.997, 164.009, 164.025, 164.107, 164.213, 164.278, 164.382],
   y = [59.9546, 60.0978, 60.1893, 60.2996, 60.4146, 60.4521, 60.4967, 60.5601, 60.5671, 60.6422, 60.7244, 60.7619, 60.7455, 60.7995, 60.8792, 60.9379, 60.9966, 61.0975, 61.1515, 61.1843, 61.2641, 61.3744,
        61.4261, 61.5223, 61.5856, 61.6279, 61.6795, 61.8109, 61.9142, 62.0222, 62.0902, 62.1418, 62.224, 62.3108, 62.3977, 62.4399, 62.4141] };
  WCL[922] = struct {
   x = [164.889, 164.992, 164.701, 164.438, 164.185, 163.884, 163.621, 163.358, 163.288, 163.302, 163.114, 163.124, 163.082, 163.016, 163.091, 163.269, 163.19, 163.114, 162.969, 162.927, 162.95, 162.828,
        162.772, 162.659, 162.396, 162.209, 162.021, 161.81, 161.655, 161.589, 161.413, 161.235, 161.014, 160.859, 160.749, 160.573, 160.34, 160.197, 160.176, 160.242, 160.34, 160.441, 160.221],
   y = [62.407, 62.5197, 62.6464, 62.6816, 62.6464, 62.5924, 62.5455, 62.4329, 62.346, 62.2756, 62.1325, 62.0597, 61.9236, 61.7992, 61.7171, 61.6889, 61.6068, 61.5387, 61.527, 61.595, 61.6748, 61.7171,
        61.642, 61.6115, 61.5997, 61.4965, 61.3791, 61.3064, 61.2266, 61.1843, 61.1186, 61.013, 60.8839, 60.7924, 60.7173, 60.7126, 60.6047, 60.5671, 60.6586, 60.7995, 60.9168, 61.013, 60.9919] };
  WCL[923] = struct {
   x = [159.988, 160.021],
   y = [61.067, 61.1351] };
  WCL[924] = struct {
   x = [159.911, 160.021, 160.061, 160.132, 160.155, 160.176, 160.209, 160.139, 160.028],
   y = [61.3158, 61.3956, 61.4847, 61.5528, 61.6678, 61.7992, 61.9564, 61.8931, 61.7945] };
  WCL[925] = struct {
   x = [170.068, 169.967, 169.88, 169.812, 169.693, 169.615, 169.472, 169.371, 169.261, 169.162, 169.028, 168.897, 168.808, 168.709, 168.611, 168.468, 168.346, 168.235, 168.134, 167.991, 167.893, 167.782,
        167.693, 167.639, 167.473, 167.395, 167.329, 167.273, 167.219, 167.174, 167.12, 167.043, 166.975, 166.933, 166.954, 166.909, 166.832, 166.712, 166.623, 166.557, 166.48],
   y = [60.0203, 60.0978, 60.1588, 60.2127, 60.2949, 60.4028, 60.4474, 60.4967, 60.5131, 60.546, 60.5554, 60.5718, 60.5718, 60.5789, 60.5718, 60.5882, 60.5953, 60.5835, 60.6211, 60.5507, 60.539, 60.539,
        60.5014, 60.4967, 60.4263, 60.4099, 60.4146, 60.4099, 60.377, 60.3723, 60.3395, 60.3723, 60.3981, 60.3723, 60.2785, 60.2292, 60.1846, 60.1259, 60.086, 60.0579, 60.0086] };
  WCL[926] = struct {
   x = [166.158, 166.158, 166.203, 166.247, 166.226, 166.182, 166.226, 166.269, 166.236, 166.182, 166.149, 166.081, 166.027, 165.917, 165.851, 165.708, 165.562, 165.485, 165.374, 165.264, 165.166, 165.088],
   y = [59.9546, 60.032, 60.1189, 60.1752, 60.2503, 60.3442, 60.3817, 60.4146, 60.4756, 60.4967, 60.5131, 60.485, 60.4521, 60.3606, 60.3231, 60.2738, 60.2573, 60.2339, 60.2174, 60.1799, 60.1095, 60.0696] };
  WCL[927] = struct {
   x = [159.923, 159.834, 159.749, 159.683, 159.693, 159.88, 159.988],
   y = [70.0953, 69.9967, 69.9216, 69.8512, 69.8019, 69.7855, 69.7784] };
  WCL[928] = struct {
   x = [160.221, 159.956, 159.866, 159.899, 159.944, 159.988],
   y = [60.9919, 60.9285, 60.9661, 61.0248, 61.0459, 61.067] };
  WCL[929] = struct {
   x = [160.021, 159.911, 159.911],
   y = [61.1351, 61.2055, 61.3158] };
  WCL[930] = struct {
   x = [160.033, 159.855, 159.7, 159.578, 159.512, 159.399, 159.146, 159.029, 158.958, 158.85, 158.695, 158.484, 158.32, 158.109, 157.855, 157.611, 157.447, 157.25, 157.038, 156.916, 156.729, 156.672,
        156.63, 156.452, 156.243, 156.043, 155.966, 155.957, 155.802, 155.614, 155.48, 155.316, 155.084, 154.919, 154.851, 154.797, 154.663],
   y = [61.7945, 61.6889, 61.6889, 61.6795, 61.7781, 61.8461, 61.9189, 61.9283, 61.872, 61.8368, 61.8297, 61.8039, 61.7945, 61.764, 61.7828, 61.7828, 61.7171, 61.6795, 61.5739, 61.5176, 61.4472, 61.3158,
        61.2219, 61.1468, 61.0341, 60.9168, 60.837, 60.7549, 60.6798, 60.5882, 60.5343, 60.4521, 60.3981, 60.316, 60.2878, 60.201, 60.086] };
  WCL[931] = struct {
   x = [70.3206, 70.4895, 70.6773, 70.865, 71.0622, 71.203, 71.4377, 71.494, 71.5785, 71.5926, 71.3485, 71.1936, 70.9824, 70.7852, 70.6304, 70.4661, 70.4097, 70.3206],
   y = [66.6407, 66.6501, 66.6712, 66.7416, 66.805, 66.8191, 66.8519, 66.8191, 66.6924, 66.6572, 66.6712, 66.6055, 66.5445, 66.4999, 66.5046, 66.5562, 66.5844, 66.6407] };
  WCL[932] = struct {
   x = [72.5266, 72.5501, 72.5829, 72.5923, 72.5923, 72.6721, 72.6721, 72.6158, 72.5594, 72.6064, 72.6392, 72.5923, 72.5172, 72.5594, 72.5407, 72.4609, 72.4609, 72.5594, 72.7284, 72.8692, 73.0804, 73.2447,
        73.409, 73.4325, 73.4325, 73.2353, 73.1555, 73.0804, 73.0664, 73.1039, 73.1227, 73.0241, 73.0898, 72.949, 72.7378, 72.6299, 72.5829, 72.4844, 72.475, 72.3388, 72.1746, 72.1511, 72.1511,
        72.1089, 71.8976, 71.724, 71.5128, 71.3813, 71.3579, 71.494, 71.6113, 71.7662, 71.9305, 72.0431, 71.9305, 71.7662, 71.6677, 71.494, 71.4048, 71.3485, 71.095, 70.9401, 70.851, 70.7759, 70.5975,
        70.4895, 70.2549, 70.1891],
   y = [70.0061, 69.9451, 69.9146, 69.9146, 69.8841, 69.8442, 69.7151, 69.6728, 69.6236, 69.5414, 69.4687, 69.4264, 69.3865, 69.2973, 69.2058, 69.119, 69.004, 68.9406, 68.8632, 68.7834, 68.7083, 68.6543,
        68.6027, 68.4947, 68.4102, 68.3093, 68.2507, 68.138, 68.0582, 67.955, 67.8235, 67.7531, 67.7038, 67.6264, 67.5982, 67.5724, 67.502, 67.4175, 67.3495, 67.2814, 67.2814, 67.2227, 67.1594,
        67.1336, 67.1148, 67.1031, 67.1547, 67.1875, 67.0937, 67.0514, 67.0819, 67.0866, 67.0819, 67.0209, 66.9646, 66.9435, 66.9364, 66.9482, 66.927, 66.8801, 66.8355, 66.8871, 66.8144, 66.7675,
        66.7675, 66.7017, 66.6924, 66.7534] };
  WCL[933] = struct {
   x = [69.7808, 70.0343, 70.0577, 70.0812, 70.0249, 70.0249, 70.1, 70.2877, 70.4426, 70.5412, 70.743, 70.9824, 71.3485, 71.6348, 71.8319, 72.0431, 72.2872, 72.4093, 72.2731, 72.3529, 72.6158, 72.8692,
        73.0335, 73.3104, 73.4982, 73.6437, 73.7845, 73.8502, 73.8643, 73.8971, 74.052, 74.16, 74.2492, 74.4369, 74.6904, 74.7138, 74.7326, 74.7655, 74.7655, 74.7467, 74.6904, 74.5777, 74.6012,
        74.4698, 74.4134, 74.3149, 74.2726, 74.2914, 74.3571, 74.4228, 74.39, 74.39, 74.559, 74.911, 75.1081, 75.2536, 75.3944, 75.6291, 75.8169, 75.9577, 76.1032, 76.3895, 76.5866, 76.6429, 76.7087,
        76.6758, 76.7509, 76.9199, 77.0044, 77.1733, 77.2907, 77.3235, 77.2907, 77.3047, 77.2719, 77.239, 77.2062, 77.225, 77.347, 77.4127, 77.3235, 77.2813, 77.2578, 77.2719, 77.239, 77.1592, 77.1921,
        77.3798, 77.5676, 77.6474, 77.6896, 77.8774, 78.1308, 78.2951, 78.3749, 78.4265, 78.6284, 78.8255, 78.8818, 78.7598, 78.5157, 78.3937, 78.1872, 78.0651, 77.9994, 77.8351, 77.7882, 77.699,
        77.6568, 77.6145, 77.5441, 77.5113, 77.5113, 77.5254, 77.6004, 77.7882, 78.0557, 78.1637, 78.1402, 78.0651, 77.9853, 77.8351, 77.7319, 77.7553, 77.746, 77.6662, 77.577, 77.3047, 76.995,
        76.8401, 76.6523, 76.4974, 76.2205, 76.0468, 75.8028, 75.6291, 75.4085, 75.1973, 75.141, 74.8547, 74.6247, 74.4134, 74.1177, 73.9065, 73.7516, 73.7657, 73.8831, 73.93, 73.7845, 73.5545,
        73.5968, 73.6437],
   y = [66.6971, 66.6783, 66.5914, 66.4999, 66.46, 66.3896, 66.3638, 66.3427, 66.338, 66.3333, 66.338, 66.3685, 66.3544, 66.2934, 66.2582, 66.223, 66.2582, 66.4037, 66.5257, 66.5797, 66.6196, 66.6572,
        66.744, 66.805, 66.8707, 66.9646, 67.0514, 67.1922, 67.2955, 67.3401, 67.4058, 67.4645, 67.5091, 67.5889, 67.6733, 67.7062, 67.7954, 67.8564, 67.9198, 67.9855, 68.0676, 68.1169, 68.1849,
        68.2319, 68.2999, 68.3211, 68.3938, 68.4102, 68.4548, 68.5393, 68.6238, 68.6825, 68.7271, 68.774, 68.8186, 68.8585, 68.8585, 68.8773, 68.8937, 68.9219, 68.9336, 68.9383, 68.9571, 68.9031,
        68.8421, 68.7599, 68.6989, 68.659, 68.6027, 68.5722, 68.5112, 68.4713, 68.4173, 68.3657, 68.3492, 68.2999, 68.26, 68.1943, 68.2108, 68.1826, 68.1075, 68.0676, 67.969, 67.8892, 67.8235, 67.7907,
        67.7156, 67.678, 67.6405, 67.5865, 67.5278, 67.5349, 67.5302, 67.5138, 67.5067, 67.5184, 67.5349, 67.5443, 67.5677, 67.6029, 67.6522, 67.6452, 67.6475, 67.6616, 67.678, 67.7156, 67.7109,
        67.7109, 67.7038, 67.6992, 67.7602, 67.8892, 67.9597, 68.0559, 68.138, 68.1896, 68.2061, 68.26, 68.3398, 68.3445, 68.4149, 68.5182, 68.6121, 68.6825, 68.7904, 68.8538, 68.8702, 68.9148,
        68.9805, 69.0204, 69.0838, 69.1119, 69.1777, 69.2058, 69.2176, 69.2082, 69.1706, 69.1143, 69.0556, 69.0721, 69.1073, 69.1073, 69.0744, 69.0885, 69.1941, 69.2528, 69.3607, 69.4569, 69.5837,
        69.7362, 69.8629, 69.9427] };
  WCL[934] = struct {
   x = [60.0203, 60.1752, 60.3723, 60.5178, 60.5695, 60.5178, 60.4146, 60.2832, 60.1517, 60.0626],
   y = [70.0507, 69.9709, 69.8934, 69.816, 69.755, 69.7292, 69.6635, 69.6987, 69.7034, 69.6752] };
  WCL[935] = struct {
   x = [59.8889, 60.1283, 60.4474, 60.5929, 60.7713, 60.8042, 60.8464, 60.8792, 60.9121, 60.959, 60.8558, 60.8699, 60.7572, 60.6915, 60.5929, 60.4615, 60.3395, 60.3301, 60.4287, 60.6023, 60.6915, 60.7901,
        60.8042, 60.9027, 61.1468, 61.3205, 61.4519, 61.5411, 61.7288, 61.9259, 62.1043, 62.2686, 62.4892, 62.7661, 62.9304, 63.1181, 63.273, 63.5265, 63.7893, 63.9583, 64.1226, 64.3526, 64.5638,
        64.7046, 64.9627, 65.2256, 65.3664, 65.5682, 65.6903, 65.7747, 65.6339, 65.6105, 65.7888, 65.8405, 65.9296, 66.0986, 66.2394, 66.5492, 66.859, 67.0655, 67.2767, 67.5396, 67.6851, 67.8587,
        68.0277, 68.2014, 68.1826, 68.2812, 68.5346, 68.5769, 68.7318, 68.8444, 68.9101, 69.1166, 69.0181, 69.0087, 68.7975, 68.5018, 68.3563, 68.2248, 68.1169, 68.07, 68.1263, 68.0136, 67.8963,
        67.6616, 67.4316, 67.2439, 67.0984, 66.9904, 66.9904, 66.8684, 66.859, 66.859],
   y = [68.6919, 68.6942, 68.7106, 68.7787, 68.8538, 68.8585, 68.8937, 68.9617, 69.0368, 69.119, 69.1307, 69.0838, 69.2129, 69.302, 69.3537, 69.4382, 69.5461, 69.6071, 69.6306, 69.6564, 69.6869, 69.7409,
        69.8066, 69.8277, 69.8254, 69.8066, 69.7831, 69.7456, 69.7479, 69.7292, 69.7339, 69.7221, 69.7174, 69.7245, 69.6846, 69.6423, 69.6423, 69.6071, 69.5532, 69.5297, 69.5179, 69.4405, 69.3771,
        69.3255, 69.2903, 69.2199, 69.1824, 69.1589, 69.1542, 69.1307, 69.1307, 69.0721, 69.0509, 69.0439, 69.0322, 69.0087, 68.9571, 68.8773, 68.8538, 68.7998, 68.7106, 68.5722, 68.5112, 68.483,
        68.4384, 68.3539, 68.26, 68.2061, 68.2765, 68.361, 68.5112, 68.6402, 68.7341, 68.8421, 68.8867, 68.9265, 68.9031, 68.9782, 69.1119, 69.2011, 69.302, 69.3607, 69.4992, 69.5039, 69.4804, 69.5578,
        69.6001, 69.6541, 69.6329, 69.6142, 69.5226, 69.6189, 69.7644, 69.8888] };
  WCL[936] = struct {
   x = [66.913, 67.0256, 67.2134],
   y = [70.0366, 69.9756, 69.9756] };
  WCL[937] = struct {
   x = [70.1891, 69.9685, 69.7808, 69.7808],
   y = [66.7534, 66.8097, 66.7792, 66.6971] };
  WCL[938] = struct {
   x = [49.8421, 50.0533, 50.2082, 50.3302, 50.4288, 50.6353, 50.8231, 50.9451, 50.9545, 51.0014, 51.2549, 51.452, 51.6867, 51.6398, 51.9073, 52.1373, 52.2687, 52.3579, 52.4565, 52.6114, 52.6442, 52.6348,
        52.5785, 52.4142, 52.6207, 52.8226, 52.954, 52.8554, 52.9305, 53.076, 53.2966, 53.2825, 53.2168, 53.2732, 53.428, 53.4046, 53.5689, 53.8035, 53.9444, 54.0664, 54.1978, 54.287, 54.3856, 54.4513,
        54.517],
   y = [68.0089, 68.0676, 68.1169, 68.138, 68.1826, 68.2483, 68.3093, 68.3281, 68.3281, 68.3117, 68.361, 68.3891, 68.4267, 68.4548, 68.483, 68.5065, 68.429, 68.3375, 68.3492, 68.4009, 68.4666, 68.5229,
        68.5839, 68.5675, 68.6637, 68.7341, 68.7341, 68.6801, 68.6191, 68.5956, 68.6262, 68.6801, 68.767, 68.7951, 68.8139, 68.8655, 68.8937, 68.9336, 68.9453, 68.9453, 68.9336, 68.9453, 68.9617,
        68.9664, 68.9735] };
  WCL[939] = struct {
   x = [49.8961, 50.0181, 50.0932, 50.0932, 50.1824, 50.3138, 50.2809, 50.2387, 50.1589],
   y = [69.2598, 69.1706, 69.1612, 69.2786, 69.1847, 69.0838, 68.997, 69.0392, 69.0908] };
  WCL[940] = struct {
   x = [58.807, 58.7648],
   y = [69.9638, 69.9685] };
  WCL[941] = struct {
   x = [60.0649, 59.8678, 59.6378, 59.6894, 59.5157, 59.328, 59.1637, 59.0088, 59.0933, 59.2482, 59.206, 59.0417, 58.8868, 58.807],
   y = [69.6752, 69.6728, 69.7339, 69.7902, 69.8348, 69.8442, 69.877, 69.9146, 69.8442, 69.7714, 69.7292, 69.8465, 69.938, 69.9638] };
  WCL[942] = struct {
   x = [54.517, 54.6296, 54.5968, 54.4982, 54.3762, 54.2213, 54.057, 53.9256, 53.7801, 53.7707, 53.8364, 53.8458, 53.935, 53.9913, 54.057, 54.0241, 53.9772, 53.9021, 53.9021, 54.0007, 54.0101, 54.0007,
        54.1978, 54.3433, 54.4184, 54.5639, 54.7188, 54.8831, 54.9253, 54.9394, 54.9488, 55.0239, 55.1788, 55.3149, 55.4557, 55.6669, 55.8735, 56.0941, 56.2489, 56.4038, 56.5822, 56.7465, 57.0234,
        57.1314, 57.366, 57.4975, 57.596, 57.7274, 57.995, 58.0372, 58.178, 58.4456, 58.7225, 58.8962, 59.0839, 59.0511, 59.1966, 59.3702, 59.4594, 59.3139, 59.1966, 59.1168, 59.1168, 59.3139, 59.558,
        59.7786, 59.9898, 59.9663, 59.8772, 59.8912],
   y = [68.9735, 68.9852, 68.9688, 68.95, 68.9265, 68.9172, 68.9219, 68.9172, 68.882, 68.8139, 68.8186, 68.8749, 68.8538, 68.8233, 68.7787, 68.6801, 68.6402, 68.5792, 68.5276, 68.49, 68.3774, 68.3117,
        68.2812, 68.2483, 68.2671, 68.1896, 68.1451, 68.1451, 68.2061, 68.2812, 68.3445, 68.4102, 68.4454, 68.4947, 68.5417, 68.5393, 68.5956, 68.6355, 68.6121, 68.5722, 68.5792, 68.6027, 68.5792,
        68.5229, 68.5299, 68.5956, 68.6707, 68.7341, 68.774, 68.8116, 68.882, 68.9265, 68.9735, 68.997, 68.9852, 68.9265, 68.8421, 68.7505, 68.72, 68.6825, 68.6262, 68.5557, 68.4384, 68.3657, 68.3211,
        68.361, 68.4666, 68.5299, 68.5956, 68.6919] };
  WCL[943] = struct {
   x = [39.915, 40.0182, 40.0839, 40.206, 40.3045, 40.4031, 40.3937, 40.5345, 40.6472, 40.8537, 41.0415, 41.0508, 41.0649, 41.14, 41.1541, 41.1635, 41.2198, 41.3841, 41.3278, 41.1729, 41.0086, 40.7786,
        40.6143, 40.436, 40.2717, 40.2248, 40.0933],
   y = [68.0793, 68.0465, 67.9878, 67.9385, 67.8869, 67.8564, 67.7954, 67.7367, 67.7696, 67.7203, 67.6241, 67.5231, 67.4715, 67.3659, 67.319, 67.2087, 67.1805, 67.1453, 66.9787, 66.8003, 66.6971, 66.5539,
        66.4882, 66.399, 66.3544, 66.3333, 66.2746] };
  WCL[944] = struct {
   x = [39.9502, 40.1567, 40.3444, 40.4759, 40.5744, 40.7293, 40.8091, 40.9077, 40.9077, 40.7528, 40.7199, 40.565, 40.4665, 40.3351, 40.1332, 40.0347],
   y = [64.6084, 64.5732, 64.545, 64.5262, 64.5403, 64.5028, 64.4746, 64.4511, 64.4981, 64.6013, 64.6694, 64.8102, 64.9158, 65.0285, 65.1411, 65.2139] };
  WCL[945] = struct {
   x = [39.861, 40.0816, 40.3022, 40.5087, 40.6308, 40.8514, 41.0156, 41.2503, 41.4568, 41.6446, 41.8229, 41.9544, 42.1093, 42.175, 42.3299, 42.405, 42.527, 42.6256, 42.7054, 42.9823, 43.1466, 43.2874,
        43.4094, 43.5972, 43.5784, 43.4423, 43.5314, 43.6769, 43.8318, 43.9633, 44.0055, 44.1275, 44.1933, 44.2026, 44.2167, 44.2496, 44.1839, 44.1369, 44.151, 44.1698, 44.259, 44.3247, 44.4467,
        44.5124, 44.4608, 44.5359, 44.5453, 44.4373, 44.381, 44.2167, 43.982, 43.8976, 43.8647, 43.8741, 43.9492, 44.029, 44.1369, 44.1182, 44.1182, 44.1182, 44.151, 44.2026, 44.1698, 44.1698, 44.1698,
        43.9961, 43.7849, 43.508, 43.3108, 43.278, 43.3672, 43.6206, 43.8412, 44.029, 44.2026, 44.4045, 44.6438, 44.8644, 45.0522, 45.2869, 45.4934, 45.7469, 45.9487, 45.9581, 46.0801, 46.235, 46.4227,
        46.4885, 46.5213, 46.6762, 46.5307, 46.1787, 45.8595, 45.6718, 45.4089, 45.3291, 45.1977, 45.0006, 45.01, 45.0991, 45.2963, 45.5075, 45.6811, 45.7609, 45.8595, 45.9675, 46.1224, 46.3007,
        46.5307, 46.5636, 46.6527, 46.7419, 46.9297, 47.1174, 47.3145, 47.4929, 47.6149, 47.7464, 47.7229, 47.6994, 47.7792, 47.8543, 47.92, 48.0327, 48.1406, 48.319, 48.5631, 48.7039, 48.8822,
        48.8024, 48.8259, 48.9808, 49.1685, 49.4079, 49.5957, 49.8397],
   y = [65.6175, 65.648, 65.7208, 65.8029, 65.9156, 65.9766, 65.9766, 66.0212, 66.054, 66.1385, 66.2441, 66.2981, 66.3685, 66.4483, 66.5234, 66.4905, 66.453, 66.399, 66.3826, 66.3826, 66.3826, 66.3544,
        66.3075, 66.2488, 66.1643, 66.0493, 66.0892, 66.1338, 66.1432, 66.0704, 65.986, 65.9156, 65.8733, 65.9601, 66.0352, 66.1432, 66.2347, 66.2723, 66.2934, 66.338, 66.4483, 66.4952, 66.5656,
        66.6971, 66.7792, 66.8707, 66.9576, 67.0467, 67.0937, 67.15, 67.15, 67.1805, 67.2978, 67.3964, 67.4879, 67.5935, 67.7414, 67.8447, 67.8494, 67.8564, 67.9033, 67.9925, 68.0911, 68.2155, 68.3211,
        68.3891, 68.4454, 68.5276, 68.6003, 68.6637, 68.6473, 68.6003, 68.5675, 68.5229, 68.5065, 68.5276, 68.5229, 68.5112, 68.5159, 68.5182, 68.483, 68.4619, 68.4056, 68.2952, 68.2647, 68.138,
        68.1732, 68.0512, 67.9221, 67.8118, 67.8071, 67.7602, 67.7484, 67.732, 67.6945, 67.6029, 67.5349, 67.5278, 67.4222, 67.3166, 67.2509, 67.1617, 67.0772, 66.9881, 66.9083, 66.8449, 66.8003,
        66.8003, 66.8566, 66.7534, 66.6501, 66.798, 66.8191, 66.8613, 66.8496, 66.8918, 66.913, 66.9998, 67.1031, 67.1805, 67.265, 67.3753, 67.4856, 67.5771, 67.5607, 67.61, 67.6194, 67.6358, 67.6851,
        67.7038, 67.8118, 67.8024, 67.8564, 67.9198, 67.9597, 68.0089] };
  WCL[946] = struct {
   x = [48.6616, 48.7133, 48.7133, 48.793, 48.8822, 48.9573, 48.9902, 49.0794, 49.1685, 49.3234, 49.4877, 49.7177, 49.8961],
   y = [69.4452, 69.4616, 69.4734, 69.4757, 69.4499, 69.4616, 69.4804, 69.4992, 69.4992, 69.4522, 69.3912, 69.3255, 69.2598] };
  WCL[947] = struct {
   x = [50.1613, 49.9547, 49.7529, 49.5886, 49.3352, 49.1803, 49.016, 48.8283, 48.7391, 48.6827, 48.5419, 48.4105, 48.2979, 48.2744, 48.3307, 48.2228, 48.1664, 48.2228, 48.2322, 48.2415, 48.3119, 48.4199,
        48.5419, 48.664],
   y = [69.0908, 69.0087, 68.8937, 68.8186, 68.767, 68.767, 68.6942, 68.6872, 68.7341, 68.6801, 68.7318, 68.8139, 68.8233, 68.7318, 68.6872, 68.7552, 68.8256, 68.9101, 69.0322, 69.1378, 69.2363, 69.3138,
        69.3912, 69.4452] };
  WCL[948] = struct {
   x = [29.9785, 30.0019, 30.0677, 30.1005, 30.1005, 30.1662, 30.246, 30.3446, 30.3352, 30.4338, 30.4666, 30.476, 30.5886, 30.6872, 30.8421, 30.8937],
   y = [69.7034, 69.6846, 69.6423, 69.6564, 69.7362, 69.694, 69.6846, 69.6752, 69.6541, 69.6541, 69.7292, 69.7855, 69.7831, 69.7597, 69.755, 69.755] };
  WCL[949] = struct {
   x = [30.8937, 30.9078, 30.9313, 31.105, 31.2927, 31.3913, 31.4804, 31.6588, 31.7339, 31.701, 31.7339, 31.809, 31.9639, 31.978, 31.8794, 31.978, 32.1845, 32.452, 32.4708, 32.6163, 32.7477, 32.8604, 32.912,
        33.0247, 33.081, 33.0575, 32.9683, 32.8932, 32.7477, 32.5412, 32.4286, 32.33, 32.208, 32.1423, 32.0672, 32.0202, 32.1751, 32.3206, 32.3722, 32.3957, 32.3394, 32.3957, 32.4943, 32.6163, 32.7618,
        32.8134, 32.7806, 32.8463, 32.9683, 32.9261, 32.8932, 33.0669, 33.2453, 33.2875, 33.2687, 33.2453, 33.311, 33.3204, 33.3344, 33.3673, 33.3861, 33.5222, 33.6865, 33.8273, 33.9071, 34.1699,
        34.3811, 34.5924, 34.7989, 34.9631, 35.118, 35.2166, 35.287, 35.2307, 35.4419, 35.6813, 35.8596, 35.9582, 36.0802, 36.3337, 36.5778, 36.728, 36.85, 36.972, 37.1926, 37.3241, 37.493, 37.6338,
        37.7981, 38, 38.0985, 38.1971, 38.3285, 38.3614, 38.4928, 38.5163, 38.6805, 38.8918, 38.7462, 38.8354, 38.9809, 39.0466, 39.1124, 39.1452, 39.3095, 39.4409, 39.6099, 39.7178, 39.8633, 39.7835,
        39.915],
   y = [69.755, 69.755, 69.7644, 69.7292, 69.7057, 69.6752, 69.6306, 69.6329, 69.6752, 69.7456, 69.7949, 69.7644, 69.7714, 69.87, 69.9193, 69.9005, 69.8817, 69.8254, 69.7949, 69.7526, 69.7667, 69.7479,
        69.7339, 69.7362, 69.7221, 69.6564, 69.6071, 69.5578, 69.5766, 69.5719, 69.6189, 69.6142, 69.6423, 69.7221, 69.6189, 69.5391, 69.5414, 69.5461, 69.5039, 69.4616, 69.4147, 69.4147, 69.4405,
        69.4569, 69.4569, 69.4452, 69.3771, 69.403, 69.4217, 69.349, 69.295, 69.3419, 69.3959, 69.3771, 69.3325, 69.295, 69.2739, 69.2715, 69.2598, 69.241, 69.2082, 69.2481, 69.2481, 69.2598, 69.3067,
        69.2598, 69.2434, 69.2199, 69.2082, 69.1824, 69.1495, 69.1495, 69.1824, 69.2176, 69.1894, 69.173, 69.1659, 69.1307, 69.1073, 69.0603, 69.0087, 68.9383, 68.9054, 68.8702, 68.8233, 68.7623,
        68.7106, 68.6754, 68.6074, 68.5417, 68.4994, 68.4713, 68.4009, 68.3492, 68.2765, 68.2812, 68.2483, 68.2342, 68.3539, 68.3046, 68.2507, 68.1662, 68.1286, 68.1849, 68.1169, 68.0746, 68.0418,
        68.0066, 68.0254, 68.1169, 68.0793] };
  WCL[950] = struct {
   x = [40.0933, 39.807, 39.5442, 39.2907, 39.0701, 38.8354, 38.7134, 38.4834, 38.2628, 37.9671, 37.831, 37.6338, 37.7887, 37.723, 37.601, 37.4461, 37.249, 36.9814, 36.7092, 36.4651, 36.2023, 35.9019,
        35.6062, 35.4607, 35.2307, 35.118, 34.8881, 34.7895, 34.5126, 34.4375, 34.3905, 34.2826, 34.1934, 34.1042, 33.9963, 33.8273, 33.6865, 33.6208, 33.555, 33.4236, 33.2124, 33.0481, 32.8604,
        32.8369, 32.6726, 32.4849, 32.3957, 32.2174, 32.1094, 31.9874, 32.2737, 32.4051, 32.4849, 32.7618, 32.6492, 32.8041, 32.9918, 33.0669, 33.2547, 33.5644, 33.6067, 33.5738, 33.5738, 33.8742,
        34.0714, 34.1605, 34.2685, 34.4375, 34.5219, 34.6581, 34.7895, 34.8223, 34.7801, 34.7332, 34.7097, 34.4562, 34.3811, 34.5595, 34.7097, 34.7003, 34.7801, 34.8881, 34.8552, 34.8974, 34.7238,
        34.6768, 34.9209, 35.0336, 35.1838, 35.3527, 35.5405, 35.6484, 35.8268, 35.9582, 36.1131, 36.2351, 36.2868, 36.5965, 36.9063, 37.1692, 37.3475, 37.601, 37.601, 37.6667, 37.7981, 37.9342,
        38.1642, 38.2628, 38.1642, 38.1079, 38.0516, 38.0187, 37.9108, 37.7136, 37.5775, 37.3569, 37.202, 37.0612, 36.9486, 36.8641, 36.7092, 36.5965, 36.5543, 36.6106, 36.728, 36.8969, 36.9392,
        37.1363, 37.479, 37.7794, 37.8545, 38.1408, 38.3754, 38.2628, 38.0844, 38.1314, 38.1642, 38.2957, 38.474, 38.5726, 38.7369, 39.0138, 39.3423, 39.4785, 39.6991, 39.9291, 39.9713, 39.9525],
   y = [66.2746, 66.2277, 66.1784, 66.1291, 66.115, 66.0704, 66.0751, 66.0845, 66.0892, 66.0939, 66.108, 66.1291, 66.108, 66.115, 66.1244, 66.1948, 66.223, 66.2723, 66.2887, 66.284, 66.3427, 66.3732,
        66.4037, 66.4741, 66.4741, 66.5445, 66.575, 66.5844, 66.5398, 66.6266, 66.7229, 66.737, 66.6877, 66.7111, 66.6619, 66.7135, 66.8191, 66.7534, 66.737, 66.7745, 66.8308, 66.8707, 66.974, 67.0561,
        67.0937, 67.1406, 67.1289, 67.1289, 67.1148, 67.0561, 67.0209, 66.9364, 66.8238, 66.7135, 66.6759, 66.6313, 66.5609, 66.5187, 66.4553, 66.4389, 66.4037, 66.338, 66.2887, 66.2394, 66.2347,
        66.1385, 66.1338, 66.1197, 66.0798, 65.9273, 65.8686, 65.7888, 65.594, 65.5072, 65.4251, 65.3922, 65.3335, 65.1951, 65.0519, 64.9393, 64.9205, 64.8712, 64.7398, 64.5826, 64.5075, 64.4417,
        64.4981, 64.3995, 64.3408, 64.2892, 64.2986, 64.3455, 64.3033, 64.2188, 64.1648, 64.078, 64.017, 63.9818, 63.9043, 63.8456, 63.8316, 63.7682, 63.7987, 63.8574, 63.9043, 63.9395, 63.9043,
        63.8809, 63.9442, 64.0874, 64.2024, 64.3174, 64.4464, 64.4136, 64.3854, 64.3408, 64.4183, 64.4746, 64.5732, 64.66, 64.6929, 64.8055, 64.89, 64.89, 64.9158, 65.005, 65.1622, 65.1435, 65.0754,
        64.9909, 64.8853, 64.8618, 64.8196, 64.8055, 64.7539, 64.6553, 64.5919, 64.6929, 64.7492, 64.7961, 64.7492, 64.7304, 64.6225, 64.5919, 64.5544, 64.5356, 64.5638, 64.6084] };
  WCL[951] = struct {
   x = [40.037, 39.915, 39.8821, 39.8633, 39.8633],
   y = [65.2139, 65.3195, 65.4532, 65.5213, 65.6175] };
  WCL[952] = struct {
   x = [30.2155, 30.1826, 30.0489],
   y = [59.964, 60.0015, 60.0133] };
  WCL[953] = struct {
   x = [30.9313, 30.9548, 31.0205, 31.1425, 31.2317, 31.3396, 31.4711, 31.5837, 31.6917, 31.8137, 32.0132, 32.2103, 32.3417, 32.452, 32.5084, 32.553, 32.6069, 32.7407, 32.7407, 32.8064, 32.7618, 32.6304,
        32.6187, 32.553, 32.3769, 32.222, 32.1117, 31.9803, 31.7808, 31.572, 31.4499],
   y = [61.6819, 61.6936, 61.6772, 61.6725, 61.6467, 61.6138, 61.5317, 61.4824, 61.3885, 61.3298, 61.3087, 61.2712, 61.2078, 61.1351, 61.0693, 60.9731, 60.898, 60.8675, 60.7689, 60.6892, 60.5601, 60.5319,
        60.3958, 60.3301, 60.1893, 60.1775, 60.2221, 60.2433, 60.2315, 60.1822, 60.0837] };
  WCL[954] = struct {
   x = [30.9876, 30.9641, 30.9759, 30.9102, 30.8867, 30.7881, 30.6332, 30.4807, 30.4244, 30.4361, 30.5018, 30.469, 30.3587, 30.3141, 30.2929, 30.2929, 30.2484, 30.1381, 30.0606, 30.0278, 30.0935, 30.1146,
        30.1263, 30.192, 30.2413, 30.2812, 30.3188, 30.3751, 30.4079, 30.4502, 30.5018, 30.5112],
   y = [59.9405, 60.0555, 60.1822, 60.3043, 60.424, 60.553, 60.6469, 60.722, 60.7596, 60.7689, 60.8394, 60.8722, 60.959, 60.9802, 61.0318, 61.0435, 61.0482, 61.1022, 61.06, 61.3298, 61.3791, 61.4261, 61.473,
        61.5317, 61.5411, 61.6044, 61.6302, 61.6725, 61.6725, 61.6678, 61.6889, 61.7358] };
  WCL[955] = struct {
   x = [19.988, 20.0959, 20.1851, 20.2931, 20.34, 20.3072, 20.2931, 20.2602, 20.1523, 20.0725, 20.218, 20.34, 20.4621, 20.6263, 20.5935, 20.5606, 20.6592, 20.7249, 20.8563, 21.0018, 20.922, 20.9361, 21.0112,
        21.0769, 21.1661, 21.2224, 21.321],
   y = [69.8066, 69.8888, 69.8958, 69.9146, 69.7855, 69.6728, 69.5226, 69.4499, 69.3771, 69.3185, 69.3771, 69.5039, 69.5532, 69.5954, 69.6329, 69.7339, 69.7479, 69.8277, 69.8888, 69.9146, 69.8465, 69.8066,
        69.7831, 69.8277, 69.8254, 69.8934, 69.8888] };
  WCL[956] = struct {
   x = [21.2553, 21.3867, 21.5416, 21.7059, 21.8842, 22.0063, 22.0063, 21.9265, 21.9265, 21.9734],
   y = [70.0131, 69.9709, 69.9146, 69.8395, 69.7362, 69.6987, 69.7902, 69.8582, 69.931, 69.9897] };
  WCL[957] = struct {
   x = [22.9872, 22.9872, 23.0858, 23.2266],
   y = [70.0483, 69.9897, 69.9521, 69.9568] };
  WCL[958] = struct {
   x = [29.1735, 29.4035, 29.5161, 29.4504, 29.8259, 29.9104, 29.9433, 29.8916, 29.7555, 29.6241, 29.549, 29.7461, 29.8916, 29.9339, 29.9761],
   y = [70.0014, 69.9568, 69.9193, 69.8841, 69.8958, 69.8559, 69.816, 69.7784, 69.7667, 69.7292, 69.6494, 69.6916, 69.694, 69.7034, 69.7034] };
  WCL[959] = struct {
   x = [19.866, 20.0068, 20.1429, 20.218, 20.34, 20.4386, 20.4808, 20.5372, 20.6592, 20.7155, 20.8047, 20.8469, 20.9361, 21.0582, 21.1098, 21.1567, 21.2224, 21.2881, 21.3632, 21.3867, 21.3632, 21.4336,
        21.5322, 21.5651, 21.4665, 21.4524, 21.3867, 21.2553, 21.1896, 21.1567, 21.199, 21.1755, 21.091, 21.1004, 21.1755, 21.199, 21.3116, 21.3632, 21.3961, 21.4336, 21.5087, 21.5416, 21.5651,
        21.5087, 21.4524, 21.3773, 21.3961, 21.5322, 21.5979, 21.6073, 21.6871, 21.7528, 21.8373, 21.8936, 21.9171, 21.8608, 21.8842, 22.0391, 22.0579, 21.9734, 21.8608, 21.8608, 21.9593, 22.072,
        22.1612, 22.2363, 22.2363, 22.2128, 22.2269, 22.2926, 22.3114, 22.3254, 22.3912, 22.424, 22.4663, 22.4663, 22.532, 22.6118, 22.654, 22.7009, 22.7995, 22.8324, 22.9075, 22.9732, 23.0764,
        23.1609, 23.3158, 23.5036, 23.635, 23.8039, 23.9119, 23.9917, 24.0668, 24.0246],
   y = [63.6509, 63.7048, 63.6814, 63.7213, 63.7447, 63.7776, 63.8128, 63.8456, 63.8762, 63.9231, 64.0052, 64.078, 64.132, 64.1836, 64.2329, 64.2892, 64.3361, 64.3854, 64.4277, 64.3854, 64.3033, 64.3502,
        64.4089, 64.4464, 64.4887, 64.5309, 64.5638, 64.6178, 64.6506, 64.7022, 64.7163, 64.7726, 64.8149, 64.8431, 64.8759, 64.9252, 64.9487, 64.9581, 65.0003, 65.0425, 65.0942, 65.1435, 65.1951,
        65.2326, 65.2819, 65.3335, 65.3523, 65.3828, 65.4063, 65.4626, 65.472, 65.4532, 65.4438, 65.4532, 65.5025, 65.5401, 65.5495, 65.5495, 65.5893, 65.648, 65.7067, 65.7349, 65.702, 65.6715,
        65.6128, 65.6128, 65.6527, 65.7349, 65.7208, 65.7349, 65.8193, 65.8921, 65.9015, 65.8921, 65.9015, 65.8616, 65.8193, 65.8076, 65.8686, 65.9062, 65.8733, 65.8381, 65.7982, 65.7982, 65.7607,
        65.8522, 65.8569, 65.8569, 65.8663, 65.824, 65.7654, 65.824, 65.8475, 65.9367] };
  WCL[960] = struct {
   x = [24.0246, 24.0903, 24.0997, 24.1325, 24.1654, 24.2545, 24.3109, 24.4329, 24.5315, 24.5737, 24.6535, 24.8178, 24.9398, 25.0806, 25.1463, 25.259, 25.3247, 25.3482, 25.3247, 25.3153, 25.3153, 25.3012,
        25.3669, 25.489, 25.3482, 25.2825, 25.3341, 25.1463, 24.9586, 24.7849, 24.6394, 24.5643, 24.4658, 24.2311, 24.0433, 23.8697, 23.6162, 23.4285, 23.2642, 23.1281, 22.9966, 22.8324, 22.7197,
        22.546, 22.424, 22.3348, 22.2128, 22.0391, 21.8514, 21.6965, 21.6167, 21.5745, 21.7293, 21.6496],
   y = [65.9367, 65.9109, 65.9109, 65.8921, 65.878, 65.8663, 65.8146, 65.7888, 65.8287, 65.7841, 65.702, 65.648, 65.6222, 65.5518, 65.5166, 65.4438, 65.3781, 65.3382, 65.2725, 65.2139, 65.1528, 65.1129,
        65.0472, 64.9627, 64.9721, 64.9111, 64.8384, 64.8853, 64.8759, 64.8384, 64.8149, 64.7022, 64.5403, 64.4371, 64.3268, 64.2235, 64.0921, 64.0052, 63.9043, 63.8175, 63.7001, 63.5781, 63.5007,
        63.4467, 63.3575, 63.3223, 63.2449, 63.2449, 63.2237, 63.2542, 63.1791, 63.1346, 63.0595, 63.0407] };
  WCL[961] = struct {
   x = [27.6246, 27.6129, 27.5472, 27.458, 27.3383, 27.1271, 26.9745, 26.7539, 26.5756, 26.3995, 26.247, 26.125, 26.0264, 25.8715, 25.8269, 25.8809, 25.7823, 25.6603, 25.5406, 25.4632, 25.3857, 25.32,
        25.1534, 25.022, 24.8671, 24.7474, 24.6042, 24.4587, 24.3836, 24.2827, 24.2287, 24.1513, 24.0175, 23.8744, 23.5787],
   y = [60.5859, 60.6023, 60.553, 60.5225, 60.5366, 60.5437, 60.5929, 60.5319, 60.4991, 60.4944, 60.4897, 60.4404, 60.4334, 60.4615, 60.4897, 60.4005, 60.3747, 60.3536, 60.3371, 60.424, 60.3465, 60.2972,
        60.2878, 60.255, 60.1987, 60.2268, 60.2104, 60.1611, 60.1071, 60.0626, 60.0672, 60.1071, 60.079, 60.079, 60.0062] };
  WCL[962] = struct {
   x = [23.3463, 23.358, 23.3135],
   y = [59.9804, 60.0672, 60.0907] };
  WCL[963] = struct {
   x = [23.0694, 23.0154, 22.938, 22.9497, 22.9051, 22.8723, 22.7831, 22.7948, 22.7737, 22.4968, 22.431, 22.3536, 22.3982, 22.3536, 22.0344, 21.835, 21.7364, 21.6378, 21.5369, 21.2834, 21.2623, 21.3609,
        21.3844, 21.328, 21.2506, 21.2177, 21.2506, 21.2412, 21.1731, 21.2295, 21.2834, 21.3187, 21.3726, 21.3844, 21.3844, 21.3515, 21.3632, 21.3562, 21.3844, 21.4336, 21.4782, 21.4735, 21.4243,
        21.3844, 21.3679, 21.3515, 21.3468, 21.2741, 21.2365, 21.2529, 21.2694, 21.2365, 21.1966, 21.1755, 21.1638, 21.152, 21.1802, 21.2037, 21.2365, 21.2623, 21.3233, 21.3515, 21.3515, 21.3468,
        21.5017],
   y = [59.9898, 60.0015, 60.025, 60.0837, 60.1236, 60.1564, 60.2151, 60.3254, 60.3747, 60.2925, 60.3207, 60.3418, 60.4075, 60.4498, 60.4897, 60.5366, 60.5976, 60.6352, 60.6633, 60.668, 60.7314, 60.7384,
        60.7807, 60.7854, 60.8088, 60.8558, 60.898, 60.952, 61.0013, 61.0905, 61.0905, 61.1351, 61.2148, 61.2923, 61.3416, 61.4143, 61.473, 61.5246, 61.5622, 61.5669, 61.5199, 61.5458, 61.595, 61.6678,
        61.7499, 61.7969, 61.8907, 61.9682, 62.0574, 62.1817, 62.278, 62.3812, 62.4329, 62.4939, 62.5737, 62.6652, 62.7567, 62.8224, 62.8834, 62.9022, 62.9069, 62.9773, 63.0172, 63.0219, 63.0407] };
  WCL[964] = struct {
   x = [19.9833, 20.049],
   y = [60.025, 60.0907] };
  WCL[965] = struct {
   x = [19.9833, 20.0396, 20.049, 20.0162],
   y = [60.2221, 60.2808, 60.3254, 60.3465] };
  WCL[966] = struct {
   x = [21.7692, 21.7247, 21.6824, 21.6144, 21.5815, 21.6589, 21.7575, 21.8021],
   y = [60.2104, 60.255, 60.248, 60.2433, 60.1822, 60.14, 60.14, 60.1564] };
  WCL[967] = struct {
   x = [22.5531, 22.6188, 22.6728, 22.6962, 22.7056, 22.6399, 22.5414, 22.3865, 22.3747, 22.2996, 22.255, 22.309, 22.4193, 22.485, 22.5531],
   y = [60.0508, 60.0837, 60.1729, 60.2644, 60.3043, 60.2714, 60.255, 60.2151, 60.1447, 60.1165, 60.0672, 60.0743, 60.0837, 60.0626, 60.0508] };
  WCL[968] = struct {
   x = [29.0022, 29.1008, 29.2556],
   y = [59.9476, 60.0015, 60.0062] };
  WCL[969] = struct {
   x = [30.0489, 29.9386, 29.9503, 29.9503, 29.7743, 29.5865, 29.4223, 29.2439, 29.0468, 28.8708, 28.7276, 28.6173, 28.5164, 28.4296, 28.4296, 28.507, 28.5845, 28.5845, 28.5845, 28.5938, 28.6267, 28.6056,
        28.4953, 28.385, 28.2629, 28.1526, 28.0658, 27.8663, 27.7678, 27.6246],
   y = [60.0133, 60.0461, 60.0907, 60.133, 60.1987, 60.2315, 60.2151, 60.1893, 60.1893, 60.255, 60.3207, 60.3536, 60.3583, 60.4334, 60.5695, 60.553, 60.4897, 60.4944, 60.5648, 60.6469, 60.6938, 60.7549,
        60.7314, 60.7009, 60.6633, 60.6023, 60.553, 60.5601, 60.5695, 60.5859] };
  WCL[970] = struct {
   x = [9.94119, 10.1055, 10.1853, 10.1383, 10.1947, 10.3824, 10.5795, 10.7016, 10.7156, 10.6687, 10.6124, 10.7673, 10.9691, 11.1005, 11.2085, 11.3446, 11.4103, 11.3305, 11.232, 11.0677, 10.8377, 10.6499,
        10.4716, 10.3402, 10.1618, 10.0398],
   y = [63.4115, 63.3974, 63.4115, 63.4655, 63.496, 63.496, 63.5194, 63.5499, 63.5781, 63.6039, 63.6368, 63.7048, 63.7635, 63.7682, 63.7987, 63.8316, 63.8809, 63.9043, 63.8456, 63.8081, 63.7682, 63.7048,
        63.6509, 63.6039, 63.5546, 63.5781] };
  WCL[971] = struct {
   x = [9.96232, 10.0046],
   y = [63.7541, 63.7987] };
  WCL[972] = struct {
   x = [9.92946, 10.1172, 10.1172, 10.0515, 10.1031, 10.1829, 10.2486, 10.3237, 10.4129, 10.5349, 10.5584, 10.6335, 10.7462, 10.7978, 10.9104, 10.9668, 11.0888, 11.2061, 11.408, 11.4831, 11.4267, 11.4408,
        11.6286, 11.7365, 11.6708, 11.5065, 11.3282, 11.361, 11.408, 11.5629, 11.6943, 11.7131, 11.8492, 11.9008, 11.9665, 12.0698, 12.1683, 12.2669, 12.3889, 12.3561, 12.3561, 12.3749, 12.342,
        12.2669, 12.1543, 12.0698, 12.1214, 12.2434, 12.3092, 12.2904, 12.2904, 12.3092, 12.3983, 12.5298, 12.6659, 12.6518, 12.6189, 12.5767, 12.511, 12.4547, 12.464, 12.5298, 12.6424, 12.7081,
        12.6659, 12.7081, 12.8161, 12.7504, 12.7738, 12.9616, 13.1399, 13.0179, 13.0179, 12.8161, 12.6753, 12.5204, 12.6189, 12.7738, 12.9287, 12.9522, 12.9522, 13.0273, 13.215, 13.4028, 13.445,
        13.3699, 13.2807, 13.1728, 13.1493, 13.1587, 13.2714, 13.4262, 13.4779, 13.2479, 13.2714, 13.3605, 13.3136, 13.5248, 13.5248, 13.7219, 13.7877, 13.6797, 13.7219, 13.7642, 13.8346, 13.9519,
        14.0646, 14.196, 14.2289, 14.2758, 14.426, 14.4495, 14.4729, 14.3509, 14.3603, 14.5386, 14.6372, 14.8343, 14.9798, 15.0033, 15.0784, 15.2239, 15.4117, 15.529, 15.6323, 15.7731, 15.5853,
        15.4211, 15.2427, 15.2239, 15.299, 15.3225, 15.3882, 15.4774, 15.5759, 15.4868, 15.6182, 15.4633, 15.4445, 15.4962, 15.6182, 15.7731, 15.82, 15.6323, 15.82, 15.8857, 16.0172, 16.05, 16.0829,
        16.1251, 16.2565, 16.2565, 16.402, 16.3926, 16.2941, 16.4114, 16.402, 16.3129, 16.2237, 16.3035, 16.4771, 16.5475, 16.7681, 16.8104, 17.0075, 17.1389, 17.2844, 17.3173, 17.4065, 17.5942,
        17.5379, 17.4722, 17.261, 17.0967, 16.9747, 16.8761, 16.6977, 16.5147, 16.5335, 16.6461, 16.7353, 17.0216, 17.1624, 17.2093, 17.4065, 17.505, 17.3736, 17.4628, 17.6928, 17.6834, 17.6271,
        17.8383, 17.7256, 17.5708, 17.5144, 17.5473, 17.5942, 17.7022, 17.8477, 17.9697, 18.134, 18.0448, 18.0213, 18.0448, 18.1011, 18.1434, 18.1997, 18.2983, 18.2419, 18.256, 18.3968, 18.4438,
        18.5189, 18.6644, 18.7395, 18.838, 18.6972, 18.5987, 18.533, 18.4766, 18.5752, 18.6972, 18.838, 19.007, 19.2042, 19.2605, 19.3262, 19.467, 19.4905, 19.359, 19.2042, 19.0258, 19.0164, 18.9929,
        19.007, 19.0821, 19.1713, 19.3356, 19.5233, 19.5796, 19.6313, 19.6454, 19.6313, 19.6547, 19.6547, 19.7439, 19.8096, 19.988],
   y = [63.9231, 63.9771, 64.0052, 64.0874, 64.1367, 64.1836, 64.2376, 64.2751, 64.3361, 64.3854, 64.4324, 64.4558, 64.4887, 64.5403, 64.6084, 64.6084, 64.5591, 64.5122, 64.5403, 64.5966, 64.6835, 64.7633,
        64.8102, 64.8196, 64.8477, 64.8008, 64.7773, 64.8431, 64.9346, 64.9581, 65.005, 65.0754, 65.1528, 65.1481, 65.127, 65.1857, 65.1716, 65.1435, 65.1951, 65.2514, 65.3054, 65.3382, 65.3711,
        65.3054, 65.2819, 65.3054, 65.3781, 65.4016, 65.4438, 65.5119, 65.5706, 65.5495, 65.4744, 65.4579, 65.3922, 65.465, 65.5401, 65.5495, 65.5753, 65.6527, 65.6527, 65.6973, 65.6715, 65.6903,
        65.756, 65.8029, 65.8428, 65.8686, 65.9367, 65.9109, 65.8827, 65.9813, 66.054, 66.0165, 65.9555, 65.9461, 66.0493, 66.0704, 66.1197, 66.1784, 66.2277, 66.2582, 66.2723, 66.2793, 66.3474,
        66.3826, 66.3286, 66.3896, 66.4952, 66.5093, 66.4788, 66.514, 66.5562, 66.575, 66.6407, 66.6665, 66.7064, 66.7182, 66.7886, 66.8003, 66.8355, 66.9059, 66.9388, 66.9529, 66.9881, 66.9904,
        67.0162, 67.0045, 67.0632, 67.1406, 67.1617, 67.1969, 67.2556, 67.2556, 67.319, 67.3753, 67.4926, 67.5607, 67.5677, 67.5184, 67.4879, 67.4856, 67.4598, 67.4879, 67.5349, 67.5935, 67.5607,
        67.5466, 67.5724, 67.5982, 67.5935, 67.5935, 67.6147, 67.6945, 67.7156, 67.7531, 67.7789, 67.8235, 67.8939, 67.9526, 67.9878, 68.0019, 68.0301, 68.0582, 68.0723, 68.1615, 68.1732, 68.1052,
        67.9925, 67.9151, 67.9268, 68.0019, 67.9808, 68.0559, 68.1544, 68.1497, 68.2014, 68.2671, 68.3117, 68.3657, 68.3985, 68.422, 68.4009, 68.3657, 68.361, 68.3375, 68.3046, 68.3985, 68.4454,
        68.4454, 68.5018, 68.5229, 68.4666, 68.5065, 68.5346, 68.4947, 68.4713, 68.5182, 68.5417, 68.5722, 68.6308, 68.6707, 68.6543, 68.7036, 68.6707, 68.7083, 68.7623, 68.774, 68.7482, 68.7951,
        68.8585, 68.8538, 68.9101, 68.8913, 68.9383, 69.0087, 69.0603, 69.0955, 69.126, 69.1425, 69.1777, 69.2246, 69.2833, 69.3255, 69.3842, 69.4382, 69.4734, 69.4335, 69.3443, 69.3138, 69.349,
        69.2668, 69.3067, 69.2598, 69.2833, 69.3208, 69.3654, 69.403, 69.464, 69.5109, 69.5179, 69.5273, 69.4687, 69.3654, 69.3208, 69.2715, 69.2176, 69.2082, 69.2739, 69.3607, 69.3842, 69.4734,
        69.5391, 69.6024, 69.6329, 69.694, 69.7479, 69.755, 69.7597, 69.7667, 69.7714, 69.7057, 69.6236, 69.5297, 69.4382, 69.5649, 69.7151, 69.8066] };
  WCL[973] = struct {
   x = [18.2349, 18.3147, 18.3898, 18.3804, 18.5212, 18.5353, 18.6433, 18.6526, 18.7747, 18.9624, 19.0281, 19.1736, 19.2394, 19.3942, 19.4928, 19.6242, 19.7134, 19.812, 19.8683],
   y = [63.0008, 63.0196, 63.0501, 63.0853, 63.1205, 63.2003, 63.219, 63.3176, 63.327, 63.3129, 63.3387, 63.388, 63.4749, 63.5241, 63.5687, 63.5147, 63.5335, 63.5734, 63.6509] };
  WCL[974] = struct {
   x = [15.6323, 15.6839, 15.6839, 15.7731, 15.8294, 15.8529, 15.8059, 15.8529, 15.9045, 15.8529, 15.7496, 15.7965, 15.928, 16.0031, 16.172, 16.1392, 16.2049, 16.3786, 16.5241, 16.5475, 16.5006, 16.402,
        16.2143, 16.0735, 15.9937, 15.9045, 15.8716, 15.7731, 15.6839, 15.6839, 15.5994, 15.5196, 15.4868, 15.5196, 15.3553, 15.177, 14.9798, 14.825, 14.8343, 14.8484, 14.7358, 14.717, 14.6044,
        14.4495, 14.2946, 14.2617, 14.2758, 14.2758, 14.3744, 14.4401, 14.5621, 14.717, 14.8015, 14.8907, 14.9892, 15.1019, 15.2427, 15.2098, 15.2662, 15.3647, 15.4445, 15.4445, 15.5196, 15.5196,
        15.5431, 15.5525, 15.5994, 15.6323],
   y = [68.9265, 68.9383, 68.9406, 68.9453, 68.8937, 68.8397, 68.7787, 68.7951, 68.7341, 68.659, 68.5839, 68.6121, 68.6402, 68.7341, 68.7505, 68.7998, 68.8655, 68.8397, 68.7717, 68.6825, 68.6238, 68.5722,
        68.5511, 68.4994, 68.4267, 68.49, 68.4713, 68.4431, 68.3938, 68.3492, 68.3093, 68.361, 68.429, 68.4666, 68.3727, 68.3046, 68.2436, 68.2108, 68.2554, 68.2765, 68.2812, 68.2812, 68.2178, 68.2014,
        68.1732, 68.2061, 68.2554, 68.2999, 68.3164, 68.3657, 68.361, 68.4009, 68.3938, 68.4384, 68.4454, 68.4572, 68.49, 68.5346, 68.5557, 68.6003, 68.6262, 68.6989, 68.72, 68.774, 68.8186, 68.8585,
        68.9101, 68.9265] };
  WCL[975] = struct {
   x = [14.4518, 14.5081, 14.4987, 14.4847, 14.4987, 14.5738, 14.6161, 14.7287, 14.7851, 14.8273, 14.9165, 15.0573, 15.0714, 15.0573, 15.0901, 15.1699, 15.2263, 15.2779, 15.2779, 15.3014, 15.3108, 15.1699,
        15.0573, 14.9165, 14.8602, 14.9822, 15.0479, 14.9259, 14.8273, 14.7194, 14.6396, 14.4847, 14.419, 14.4283, 14.4518],
   y = [68.6801, 68.6942, 68.7083, 68.7271, 68.7623, 68.7341, 68.7787, 68.7599, 68.7998, 68.8773, 68.8139, 68.8233, 68.882, 68.9571, 68.95, 68.8937, 68.8585, 68.8186, 68.7552, 68.7106, 68.6402, 68.6308,
        68.591, 68.5745, 68.6308, 68.6754, 68.7271, 68.6707, 68.6754, 68.652, 68.6308, 68.5863, 68.6262, 68.6637, 68.6801] };
  WCL[976] = struct {
   x = [15.7754, 15.8552, 15.8552, 15.8881, 15.9209, 16.0195, 16.1087, 16.1838, 16.1275, 16.0946, 16.043, 15.9866, 15.9068, 15.7989, 15.6769, 15.5642, 15.522, 15.5548, 15.6769, 15.766, 15.7754],
   y = [69.1941, 69.1964, 69.2082, 69.2528, 69.302, 69.3067, 69.2856, 69.2715, 69.2246, 69.1471, 69.1073, 69.0509, 69.0251, 69.0016, 68.9805, 68.9852, 69.0251, 69.1119, 69.1307, 69.1542, 69.1941] };
  WCL[977] = struct {
   x = [14.9259, 14.893, 14.8602, 14.8273, 14.9822, 15.1042, 15.2028, 15.2779, 15.3248, 15.292, 15.1606, 15.0714, 14.9071, 14.8038, 14.7194, 14.771, 14.8273, 14.94, 14.9259],
   y = [67.7954, 67.7977, 67.8282, 67.8681, 67.8447, 67.8494, 67.8775, 67.9104, 67.8611, 67.8071, 67.786, 67.7273, 67.6827, 67.6733, 67.6992, 67.732, 67.7649, 67.7743, 67.7954] };
  WCL[978] = struct {
   x = [16.9442, 16.8784, 16.8221, 16.9442, 16.9911, 17.0099, 16.8878, 17.0005, 16.9348, 16.977, 17.0756, 17.2117, 17.3196, 17.2211, 17.2445, 17.3196, 17.4651, 17.5309, 17.4511, 17.4745, 17.5731, 17.5966,
        17.6529, 17.728, 17.8031, 17.8406, 17.9157, 17.9251, 17.9721, 17.8735, 17.8594, 17.8735, 17.728, 17.5637, 17.5496, 17.4182, 17.3103, 17.1413, 17.0005, 16.8691, 16.7893, 16.8127, 16.8878,
        16.977, 17.0333, 16.9911, 16.9442],
   y = [69.1941, 69.2082, 69.2293, 69.2199, 69.2481, 69.2856, 69.302, 69.3185, 69.349, 69.3678, 69.3959, 69.3654, 69.4147, 69.4616, 69.464, 69.4804, 69.4522, 69.4851, 69.5696, 69.5649, 69.5039, 69.5602,
        69.5719, 69.5649, 69.5039, 69.4804, 69.4452, 69.4264, 69.3607, 69.2903, 69.2363, 69.1612, 69.1471, 69.1471, 69.1894, 69.173, 69.0838, 69.0791, 69.0674, 69.0368, 69.0838, 69.0955, 69.1119,
        69.1354, 69.1706, 69.173, 69.1941] };
  WCL[979] = struct {
   x = [18.2443, 18.3476, 18.3804, 18.4649, 18.5118, 18.5447, 18.6339, 18.6855, 18.7418, 18.7418, 18.9202, 18.9859, 18.9202, 18.831, 18.8216, 18.6996, 18.5212, 18.3663, 18.2443, 18.08, 18.0941, 18.2255,
        18.2678, 18.2584, 18.2443],
   y = [69.694, 69.6869, 69.6916, 69.6916, 69.7104, 69.7831, 69.7292, 69.7292, 69.7972, 69.8559, 69.816, 69.7667, 69.7221, 69.6564, 69.6024, 69.5766, 69.5578, 69.5414, 69.5508, 69.5696, 69.6001, 69.5954,
        69.6259, 69.6541, 69.694] };
  WCL[980] = struct {
   x = [18.8639, 18.7888],
   y = [70.0202, 69.9897] };
  WCL[981] = struct {
   x = [19.1502, 19.1502, 19.1267, 19.1971, 19.2159],
   y = [70.0319, 69.9944, 69.9521, 69.9521, 69.9873] };
  WCL[982] = struct {
   x = [19.3473, 19.4365, 19.5679, 19.5257, 19.4271, 19.3473, 19.2159, 19.1267, 18.9953, 18.8404, 18.7981, 18.8732, 18.8873],
   y = [70.0061, 69.9897, 69.9756, 69.938, 69.8817, 69.8277, 69.7972, 69.8207, 69.8653, 69.9005, 69.9521, 69.931, 69.9756] };
  WCL[983] = struct {
   x = [18.0871, 18.0871, 18.1058, 18.2279, 18.2936, 18.2513, 18.2185, 18.195, 18.073, 17.9744, 17.9087, 17.8524, 17.7444, 17.7116, 17.7022, 17.6881, 17.7679, 17.7538, 17.721, 17.7444, 17.735, 17.6224,
        17.5473, 17.5332, 17.3595, 17.2375, 17.2281, 17.2797, 17.4112, 17.3924, 17.3126, 17.2704, 17.2375, 17.2704, 17.3126, 17.3032, 17.2046, 17.0263, 17.0498, 16.9934, 16.9747, 17.0075, 17.0075,
        17.0404, 17.0263, 17.0146, 17.0474, 17.0709, 17.1366, 17.1249, 17.0591, 17.214, 17.214, 17.1577, 16.9934, 16.8503, 16.6743, 16.5076, 16.287, 16.1321, 16.2659, 16.4865, 16.6508, 16.7728,
        16.9277, 17.1037, 17.2469, 17.3009, 17.3689, 17.3689, 17.4346, 17.4675, 17.5543, 17.7092, 17.8641, 17.9416, 17.9627, 18.1293, 18.3593, 18.371, 18.2279, 18.2936, 18.3945, 18.6245],
   y = [63.0125, 63.0078, 63.0078, 62.9891, 62.9632, 62.9374, 62.9186, 62.8529, 62.8834, 62.8013, 62.7919, 62.8177, 62.8529, 62.8928, 62.9327, 62.8623, 62.7708, 62.7004, 62.6347, 62.6089, 62.5854, 62.508,
        62.4892, 62.4634, 62.5126, 62.5385, 62.4164, 62.346, 62.2639, 62.2169, 62.1137, 62.0057, 61.933, 61.8297, 61.7452, 61.6725, 61.6983, 61.7452, 61.6936, 61.6255, 61.5622, 61.4894, 61.4355,
        61.3674, 61.2923, 61.2078, 61.0764, 60.9637, 60.8839, 60.7924, 60.722, 60.7056, 60.5929, 60.5061, 60.4334, 60.3629, 60.3254, 60.2433, 60.2221, 60.1658, 60.1447, 60.1775, 60.1775, 60.2221,
        60.2878, 60.3536, 60.4498, 60.5319, 60.6188, 60.6563, 60.6516, 60.607, 60.553, 60.5765, 60.5765, 60.4897, 60.4005, 60.3536, 60.309, 60.2761, 60.2714, 60.2315, 60.14, 60.0555] };
  WCL[984] = struct {
   x = [19.9622, 19.8284, 19.7744, 19.6641, 19.7298, 19.7181, 19.6078, 19.5867, 19.6078, 19.7181, 19.8519, 19.9833],
   y = [60.3911, 60.4075, 60.4334, 60.3911, 60.3465, 60.2878, 60.3043, 60.2221, 60.14, 60.1283, 60.1001, 60.025] };
  WCL[985] = struct {
   x = [20.049, 19.9293, 19.8847, 19.9833],
   y = [60.0907, 60.1236, 60.1987, 60.2221] };
  WCL[986] = struct {
   x = [20.0162, 19.9833, 19.9622],
   y = [60.3465, 60.37, 60.3911] };
  WCL[987] = struct {
   x = [7.54742, 7.54742, 7.5615, 7.59436, 7.69293, 7.73517, 7.79149, 7.8572, 7.92291, 7.96985, 8.02148, 8.04495, 8.03556, 8.07781, 8.1576, 8.19046, 8.17637, 8.26555, 8.34065, 8.3782, 8.34065, 8.38759,
        8.46269, 8.51901, 8.54248, 8.60819, 8.70676, 8.80532, 8.86165, 8.9696, 9.10572, 9.13858, 9.13858, 9.21368, 9.42489, 9.39203, 9.29347, 9.29347, 9.30285, 9.35918, 9.42489, 9.55631, 9.61264,
        9.7112, 9.7863, 9.86609, 9.94119, 9.95058, 9.94119],
   y = [63.0102, 63.0243, 63.0688, 63.1252, 63.1439, 63.1604, 63.1651, 63.1791, 63.1838, 63.1745, 63.1791, 63.2003, 63.2378, 63.273, 63.2824, 63.2449, 63.2144, 63.2097, 63.219, 63.2496, 63.3082, 63.3082,
        63.3035, 63.3528, 63.3927, 63.4303, 63.4514, 63.4702, 63.4702, 63.5054, 63.5241, 63.4749, 63.4068, 63.3528, 63.4068, 63.4467, 63.4866, 63.5054, 63.5453, 63.564, 63.5992, 63.6086, 63.6462,
        63.6509, 63.6133, 63.5499, 63.4913, 63.4467, 63.4115] };
  WCL[988] = struct {
   x = [10.0398, 9.91773, 9.96466],
   y = [63.5781, 63.6767, 63.7541] };
  WCL[989] = struct {
   x = [10.0069, 9.80038, 9.69712, 9.7112, 9.93181],
   y = [63.7987, 63.7588, 63.7987, 63.8809, 63.9231] };
  WCL[990] = struct {
   x = [8.45095, 8.52605, 8.54013, 8.64809, 8.7138, 8.73727, 8.84991, 8.95787, 9.12215, 9.08929, 9.16909, 9.20194, 9.02358, 8.80298, 8.59176, 8.4932, 8.45095],
   y = [63.5687, 63.5734, 63.5992, 63.5922, 63.6227, 63.6579, 63.6814, 63.7001, 63.7001, 63.6673, 63.6368, 63.5828, 63.5593, 63.5335, 63.4749, 63.5241, 63.5687] };
  WCL[991] = struct {
   x = [4.93305, 4.94478, 4.90019, 4.84621, 4.82275, 4.7805, 4.7805, 4.76877, 4.75704, 4.76877, 4.80162, 4.84621, 4.8908, 4.90019, 4.93305, 4.93305],
   y = [61.1304, 61.1515, 61.121, 61.0999, 61.0881, 61.0623, 61.0412, 61.0294, 61.0036, 60.9872, 60.9872, 60.9919, 61.0083, 61.0341, 61.0928, 61.1304] };
  WCL[992] = struct {
   x = [5.00111, 5.06682, 5.04335, 4.98937, 4.9659, 4.9659, 4.98937, 4.98937, 5.00111],
   y = [60.6704, 60.7783, 60.7783, 60.7783, 60.7502, 60.7126, 60.6751, 60.6751, 60.6704] };
  WCL[993] = struct {
   x = [5.94923, 6.08065, 6.27778, 6.2684, 6.1581, 6.33411, 6.49839, 6.58757, 6.62042, 6.66501, 6.69787, 6.98418, 6.8199, 6.5993, 6.58757, 6.49839, 6.32237, 6.13463, 5.95861, 5.80372, 5.79434],
   y = [59.9898, 60.0837, 60.14, 60.1658, 60.2104, 60.309, 60.4404, 60.4287, 60.3254, 60.3864, 60.4779, 60.5483, 60.5648, 60.5319, 60.5695, 60.6141, 60.5437, 60.3629, 60.248, 60.194, 60.0837] };
  WCL[994] = struct {
   x = [5.40711, 5.40711, 5.562, 5.70516, 5.65118, 5.58547, 5.55026, 5.55026, 5.43996, 5.32027, 5.27569, 5.18651, 5.15365, 5.24283, 5.38599, 5.43996, 5.37425, 5.4517, 5.69577, 5.54088, 5.29681, 5.17712,
        5.00111, 5.00111, 5.00111, 5.03396, 5.05508, 5.16538, 5.32027, 5.32027, 5.30854, 5.24283, 5.06682, 5.08794, 5.13253, 5.18651, 5.36486, 5.52914, 5.68403, 5.86005, 6.03606, 6.23554, 6.47727,
        6.66501, 6.84102, 6.89735, 6.89735, 7.10622, 7.15081, 7.02877, 7.06163, 7.3808, 7.64599, 7.54742, 7.42539, 7.31508, 7.34794, 7.3808, 7.45824, 7.44651, 7.33151, 7.31508, 7.27988, 7.0804,
        7.01469, 6.98184, 6.8598, 6.65093, 6.55236, 6.60634, 6.57348, 6.41859, 6.35288, 6.14401, 6.02198, 5.86709, 5.6371, 5.47047, 5.29446, 5.18416, 5.18416, 5.10671, 5.11845, 5.16304, 5.11845,
        5.17947, 5.22875, 5.22405, 5.22405, 5.23579, 5.17477, 5.09733, 5.10906, 5.13018, 5.19589, 5.20293, 5.30619, 5.36252, 5.4165, 5.48925, 5.55965, 5.69812, 5.78495, 5.76383, 5.68169, 5.57608,
        5.54792, 5.62536, 5.69342, 5.64179, 5.57139, 5.5221, 5.46108, 5.40476, 5.35078, 5.26865, 5.20763, 5.1912, 5.22875, 5.20763, 5.16773, 5.20763, 5.27334, 5.35078, 5.33905, 5.34609, 5.4165,
        5.46578, 5.53853, 5.57139, 5.60424, 5.65353, 5.64179, 5.59251, 5.60894, 5.62536, 5.67465, 5.75913, 5.80372, 5.84127, 5.87413, 5.8577, 5.83658, 5.82954, 5.86239, 5.89525, 5.87413, 5.84597,
        5.78026, 5.77556, 5.82954, 5.82954, 5.77556, 5.76383, 5.82015, 5.87882, 5.92341, 5.98443, 6.06657, 6.12289, 6.20034, 6.27778, 6.34819, 6.32003, 6.34349, 6.38105, 6.33176, 6.32003, 6.29421,
        6.1933, 6.10647, 6.15575, 6.11585, 6.06657, 6.083, 6.16748, 6.25432, 6.32707, 6.36462],
   y = [59.9569, 60.0015, 60.0626, 60.0743, 60.1283, 60.1729, 60.194, 60.2761, 60.2104, 60.1658, 60.2268, 60.2761, 60.3465, 60.4005, 60.4404, 60.4568, 60.4944, 60.5601, 60.6774, 60.6633, 60.5812, 60.5929,
        60.7596, 60.7854, 60.7854, 60.7854, 60.8088, 60.7596, 60.7267, 60.8018, 60.8558, 60.8394, 60.8511, 60.9426, 60.9849, 61.0482, 61.0224, 61.0107, 61.0811, 61.0693, 61.0858, 61.0647, 61.0975,
        61.0975, 61.0318, 60.9684, 60.8722, 60.9309, 60.9473, 61.0107, 61.0975, 61.1444, 61.2454, 61.2712, 61.236, 61.2242, 61.2665, 61.3674, 61.5035, 61.5411, 61.4777, 61.358, 61.2782, 61.1703,
        61.182, 61.2031, 61.1773, 61.1773, 61.2031, 61.3205, 61.4049, 61.2712, 61.1656, 61.1656, 61.2078, 61.1867, 61.128, 61.1022, 61.1069, 61.128, 61.1914, 61.2242, 61.2876, 61.3251, 61.3838,
        61.4636, 61.4988, 61.5246, 61.5575, 61.6138, 61.6255, 61.6561, 61.7194, 61.7358, 61.7312, 61.7781, 61.8555, 61.8555, 61.818, 61.818, 61.818, 61.8344, 61.886, 61.9072, 61.8813, 61.8767, 61.9025,
        61.9424, 61.9752, 61.9682, 61.9494, 61.933, 61.9166, 61.9166, 61.9236, 61.9166, 61.9166, 61.9424, 61.9893, 62.0527, 62.1184, 62.1489, 62.1043, 62.0409, 62.0879, 62.1395, 62.1348, 62.1348,
        62.1606, 62.0926, 62.0621, 62.0409, 62.0973, 62.1442, 62.1747, 62.2263, 62.2944, 62.346, 62.2944, 62.1747, 62.1864, 62.2521, 62.3202, 62.3812, 62.4164, 62.4681, 62.4681, 62.4422, 62.4375,
        62.4821, 62.4939, 62.5033, 62.5244, 62.5432, 62.5854, 62.5737, 62.6089, 62.6159, 62.5995, 62.6042, 62.5854, 62.5901, 62.5549, 62.6089, 62.6464, 62.6769, 62.7004, 62.7215, 62.7567, 62.7262,
        62.7473, 62.7966, 62.8271, 62.8529, 62.9069, 62.9679, 62.9585, 62.9186, 63.0102] };
  WCL[995] = struct {
   x = [5.8929, 5.90229, 5.98208, 6.01494, 5.98208, 5.86005, 5.81311, 5.81311, 5.81311, 5.82719, 5.84597, 5.8929],
   y = [62.2381, 62.2521, 62.3085, 62.3507, 62.3812, 62.3953, 62.3648, 62.3085, 62.2686, 62.2381, 62.2334, 62.2381] };
  WCL[996] = struct {
   x = [5.00111, 5.03396, 5.13253, 5.14192, 5.03396, 5.00111, 5.00111, 5.00111, 5.00111],
   y = [61.7617, 61.7969, 61.825, 61.8508, 61.8649, 61.8767, 61.8813, 61.8297, 61.7969] };
  WCL[997] = struct {
   x = [-8.38289, -8.51432, -8.749, -8.95552, -9.10103, -9.10103, -8.93675, -8.85696, -8.46269, -8.18576, -8.02148, -7.99801, -7.96516, -8.03087, -8.20923, -8.38289],
   y = [71.0762, 71.0152, 70.9636, 70.9167, 70.8533, 70.8228, 70.8486, 70.9002, 70.9613, 70.9847, 71.0692, 71.1255, 71.1607, 71.1631, 71.1584, 71.0762] };
  WCL[998] = struct {
   x = [-17.7468, -17.6435, -17.5872, -17.5215, -17.5402, -17.6435, -17.606, -17.6623, -17.8125, -17.8782, -17.9627, -18.0002, -17.9815, -17.8782, -17.7468],
   y = [79.201, 79.2104, 79.1775, 79.1353, 79.1095, 79.0907, 79.0555, 79.0273, 79.0414, 78.9992, 78.9992, 79.0578, 79.1142, 79.1587, 79.201] };
  WCL[999] = struct {
   x = [-17.735, -17.7069, -17.6505, -17.5379, -17.5754, -17.6318, -17.7726, -18.0073, -18.0918, -18.0918, -17.9509, -17.8665, -17.735],
   y = [77.9008, 77.8962, 77.8938, 77.8445, 77.7905, 77.7131, 77.6638, 77.638, 77.6427, 77.7272, 77.8046, 77.8774, 77.9008] };
  WCL[1000] = struct {
   x = [-20.0725, -19.9129],
   y = [79.7713, 79.7947] };
  WCL[1001] = struct {
   x = [-20.0068, -19.7815],
   y = [79.9238, 79.9801] };
  WCL[1002] = struct {
   x = [-17.5144, -17.5895, -17.6459, -17.8336, -17.8618, -18.0307, -18.1152, -18.5846, -18.7723, -19.0821, -19.2135, -19.2793, -19.3731, -19.5609, -19.5984, -19.6547, -19.6547, -19.7392, -19.6266, -19.6641,
        -19.6923, -19.8613, -19.8519, -19.6641, -19.5984, -19.3074, -19.1009, -19.2135, -19.2886, -19.2793, -19.467, -19.5984, -19.589, -19.6641, -19.6829, -19.88, -19.9458, -19.9458, -19.8988, -19.88],
   y = [80.013, 79.9848, 79.9379, 79.8745, 79.8276, 79.7689, 79.7502, 79.7243, 79.7431, 79.7642, 79.7619, 79.7478, 79.7384, 79.6657, 79.6023, 79.5249, 79.5014, 79.438, 79.4099, 79.3606, 79.3113, 79.2221,
        79.1423, 79.2315, 79.3089, 79.3089, 79.2972, 79.2127, 79.1846, 79.2503, 79.2385, 79.1846, 79.1071, 79.1001, 79.1259, 79.093, 79.1095, 79.0156, 78.9522, 78.8889] };
  WCL[1003] = struct {
   x = [-20.0959, -19.6266, -19.2417, -19.054, -19.0258, -19.2886, -19.5421, -19.8519],
   y = [77.814, 77.7765, 77.7342, 77.6826, 77.5981, 77.5652, 77.6122, 77.6896] };
  WCL[1004] = struct {
   x = [-20.0068, -19.7721, -19.3966, -19.3778, -19.1431, -19.0305, -19.0493, -18.8709, -18.7207, -18.4672, -18.3358, -18.2889, -18.2795, -18.2419, -18.2044, -18.195, -18.2044, -18.2044, -18.2044],
   y = [77.3658, 77.3681, 77.3094, 77.2343, 77.2226, 77.232, 77.3141, 77.3587, 77.3259, 77.293, 77.3071, 77.2977, 77.2719, 77.1851, 77.1029, 77.0443, 77.0044, 76.9997, 76.9856] };
  WCL[1005] = struct {
   x = [-18.1974, -18.1974, -18.2067, -18.2067, -18.2537, -18.2537, -18.3851, -18.4696, -18.6104, -18.892, -19.2394, -19.4553, -19.7463],
   y = [76.9856, 76.9856, 76.9809, 76.9386, 76.8894, 76.8589, 76.8119, 76.7767, 76.7767, 76.8143, 76.8377, 76.8612, 76.9011] };
  WCL[1006] = struct {
   x = [-20.2156, -19.9434, -19.8589, -19.8026, -19.812],
   y = [76.2158, 76.1994, 76.1384, 76.0844, 76.0351] };
  WCL[1007] = struct {
   x = [-20.0842, -19.8401, -19.6242, -19.5773, -19.5726, -19.4834, -19.4834, -19.6148, -19.6618, -19.7697],
   y = [75.8873, 75.8192, 75.73, 75.7066, 75.5751, 75.4179, 75.2747, 75.1316, 75.1198, 75.1128] };
  WCL[1008] = struct {
   x = [-20.2063, -19.9434, -19.7228, -19.521, -19.3896, -19.1924, -19.1361, -19.3004, -19.6195],
   y = [74.6317, 74.5801, 74.6082, 74.6434, 74.5449, 74.4792, 74.4064, 74.3313, 74.2163] };
  WCL[1009] = struct {
   x = [-18.7489, -18.7489, -18.7489, -18.7395, -18.7301, -18.6644, -18.6644, -18.7207, -18.7676, -18.7864, -18.8709, -18.8897, -19.0305, -19.068, -19.1244, -19.1431, -18.9366, -18.8521, -19.0023, -19.0493,
        -19.0305, -18.8521, -18.7489, -18.7489],
   y = [76.5021, 76.5233, 76.3942, 76.2674, 76.1454, 76.0328, 75.9295, 75.9717, 76.0445, 76.1313, 76.2393, 76.3285, 76.3965, 76.427, 76.4904, 76.5397, 76.5585, 76.5796, 76.5866, 76.6852, 76.7227, 76.6476,
        76.5772, 76.5021] };
  WCL[1010] = struct {
   x = [-17.9768, -17.9674, -18.1551, -18.188, -18.1457, -18.0894, -17.958, -17.7562, -17.6482, -17.4933, -17.4698, -17.5262, -17.6482, -17.8359, -17.9908, -18.2208, -18.3851, -18.4978, -18.6714, -18.8826,
        -18.939, -18.9249, -18.9155, -18.892, -18.9483, -18.9061, -18.7184, -18.5869, -18.4649, -18.2537, -18.188, -18.0659, -18.0096, -17.9768],
   y = [75.3006, 75.2677, 75.2348, 75.1949, 75.155, 75.1363, 75.0729, 75.0729, 75.1222, 75.1339, 75.0518, 74.9861, 74.9251, 74.9767, 75.0049, 74.9767, 74.9696, 74.9579, 74.9837, 74.9767, 75.0682, 75.1222,
        75.1785, 75.2395, 75.2818, 75.3193, 75.3381, 75.3428, 75.2935, 75.3146, 75.3803, 75.3874, 75.3217, 75.3006] };
  WCL[1011] = struct {
   x = [-20.5254, -20.4315, -20.4409, -20.6099, -20.4315, -20.3564, -20.2532, -20.0654],
   y = [80.013, 79.9778, 79.9402, 79.8487, 79.837, 79.7971, 79.7502, 79.7713] };
  WCL[1012] = struct {
   x = [-19.9082, -20.1335, -20.1335, -20.0021],
   y = [79.7947, 79.8018, 79.8534, 79.9238] };
  WCL[1013] = struct {
   x = [-19.8847, -20.2133, -20.5512, -20.7296, -20.908, -21.0675, -21.1333, -21.0769, -21.0675, -21.0957, -21.3304, -21.2553, -20.9549, -21.0582, -21.1333, -21.2928, -21.3304, -21.4149, -21.6496, -21.3304,
        -21.3773, -21.612, -21.8091, -21.6683, -21.7059, -21.8936, -21.734, -21.49, -21.3492, -21.199, -21.0018, -20.8047, -20.6263, -20.5982, -20.3072, -20.1006],
   y = [78.8889, 78.8208, 78.8466, 78.8208, 78.7856, 78.7997, 78.7715, 78.741, 78.6847, 78.6424, 78.6354, 78.6049, 78.6096, 78.5603, 78.4946, 78.389, 78.2646, 78.1496, 78.1144, 78.098, 78.0581, 77.9924,
        77.8821, 77.929, 77.8351, 77.7319, 77.6591, 77.6638, 77.753, 77.8304, 77.9314, 77.9689, 77.9384, 77.8868, 77.8422, 77.814] };
  WCL[1014] = struct {
   x = [-19.8519, -20.3119, -20.6029, -20.6592, -20.7812, -20.7155, -20.4151, -20.218, -20.509, -20.5278, -20.509, -20.6967, -20.2743, -20.2274, -20.6029, -20.2837, -20.0021],
   y = [77.6896, 77.692, 77.7201, 77.6896, 77.6521, 77.6216, 77.5746, 77.5371, 77.5394, 77.5207, 77.4948, 77.4878, 77.4432, 77.4033, 77.3869, 77.3705, 77.3658] };
  WCL[1015] = struct {
   x = [-19.7533, -20.1757, -20.4949, -20.908, -21.1614, -21.49, -21.7153, -21.4243, -21.1426, -20.7578, -20.7014, -20.8892, -21.199, -21.5651, -21.6777, -21.2928, -21.0394, -20.8516, -21.1239, -21.2741,
        -21.4336, -21.6496, -21.7528, -21.9218, -22.1377, -22.4475, -22.4569, -22.7197, -22.6446, -22.532, -22.532, -22.3724, -22.2503, -22.1471, -21.8655, -22.0063, -21.8279, -22.2503, -22.4663,
        -22.6165, -22.3161, -21.9499, -21.7059, -21.6683, -21.6496, -21.6871, -21.6871, -21.7059, -21.5838, -21.4806, -21.4243, -21.1426, -21.2647, -21.0957, -20.8704, -20.6545, -20.8516, -21.0582,
        -20.861, -20.5794, -20.2227],
   y = [76.9011, 76.9199, 76.941, 76.9457, 76.9433, 76.948, 76.9504, 76.9246, 76.9175, 76.8964, 76.88, 76.8471, 76.8659, 76.8682, 76.8424, 76.8143, 76.7978, 76.7931, 76.7697, 76.7579, 76.6969, 76.6476,
        76.6382, 76.7204, 76.7908, 76.772, 76.7251, 76.6828, 76.6265, 76.5772, 76.5162, 76.5749, 76.5678, 76.5139, 76.4927, 76.4505, 76.42, 76.4388, 76.4482, 76.4223, 76.3871, 76.3848, 76.3965,
        76.3707, 76.3472, 76.3285, 76.2674, 76.2205, 76.2182, 76.2158, 76.1759, 76.1548, 76.2416, 76.2135, 76.1431, 76.1337, 76.2135, 76.2792, 76.2346, 76.1923, 76.2158] };
  WCL[1016] = struct {
   x = [-19.819, -20.0725, -20.2978, -20.4386, -20.6169, -21.0675, -21.0769, -20.7296, -20.5043, -20.3259, -20.2227, -20.0913],
   y = [76.0351, 76.0281, 75.9905, 75.9882, 75.9741, 75.9788, 75.9436, 75.9295, 75.8873, 75.8755, 75.8779, 75.8873] };
  WCL[1017] = struct {
   x = [-19.7744, -20.0185, -20.0748, -20.2156, -21.9288, -21.9617, -21.9147, -21.6613, -21.5533, -21.8819, -22.1682, -21.9382, -21.5205, -21.267, -20.9807, -20.7038, -20.6568, -20.9009, -21.375, -21.6284,
        -21.3984, -21.1215, -20.9009, -20.7695, -20.7789, -20.8258, -20.8023, -20.8587, -21.0558, -21.0652, -21.0558, -20.9432, -20.7226, -20.5254, -20.2063],
   y = [75.1128, 75.148, 75.2489, 75.3146, 75.5446, 75.554, 75.547, 75.4906, 75.4507, 75.4601, 75.4625, 75.4132, 75.4038, 75.3193, 75.2442, 75.1903, 75.1128, 75.1222, 75.0706, 74.9767, 74.9955, 75.0494,
        75.0377, 75.0002, 74.9321, 74.8382, 74.7584, 74.6833, 74.6599, 74.6599, 74.6411, 74.6341, 74.6552, 74.6458, 74.6317] };
  WCL[1018] = struct {
   x = [-19.6195, -20.0983, -20.2626, -20.3846, -20.6146, -20.9009, -21.267, -21.7411, -21.9054, -22.1166, -22.0274, -21.9945, -22.2011, -22.5672, -22.1917, -22.1682, -22.2809, -22.37, -22.3325, -22.4451,
        -22.4686, -22.1353, -21.7082, -21.5064, -21.3187, -21.2342, -21.2107, -20.9338, -20.624, -20.4034, -20.4175, -20.5724, -20.5911, -20.5911, -20.5395, -20.7038, -21.1215, -21.4735, -21.6941,
        -21.8068, -22.1166, -22.2151, -22.4686, -22.7878, -23.0835, -23.4167, -23.7453, -24.0457, -24.0973, -24.2663, -24.3883, -24.4399, -24.4728, -24.5057, -24.6934, -24.9609, -25.036, -24.9609,
        -24.8952, -25.2003, -25.4115, -25.543, -25.7072, -25.9184, -26.2705, -26.5474, -26.7445, -26.9557, -26.5239, -26.6929, -27.0215, -26.8666, -26.6037, -26.1719, -25.8621, -25.6556, -25.2566,
        -25.0595, -25.1815, -25.2895, -25.6227, -25.8762, -26.2188, -26.5239, -26.7586, -26.5896, -26.7258, -26.8666, -26.4723, -26.4723, -26.4254, -26.1719, -25.8105, -25.5242, -25.2238, -24.8811,
        -24.8295, -24.8295, -25.1158, -25.5993, -25.6227, -25.4115, -25.4772, -25.5664, -25.5664, -25.4115, -25.3364, -25.2003, -24.9046, -24.7169, -24.5385, -24.3085, -24.1208, -24.0457, -23.98,
        -23.8673, -23.8345, -23.5716, -23.3041, -23.2384, -23.0835, -22.7878, -22.6329, -22.7737, -22.83, -23.0506, -23.2055, -23.1163, -23.1069, -23.0835, -23.0412, -22.9098, -22.5343, -22.5437,
        -22.6892, -22.5766, -22.4123, -22.2809, -22.1025, -22.1917, -22.3325, -22.6, -22.6094, -22.4686, -22.3325, -22.2011, -21.9476, -21.8068, -21.8256, -21.8256, -21.9147, -21.8584, -21.8725,
        -21.896, -22.1353, -21.9147, -21.8162, -21.8584, -21.8068, -21.9054, -21.8396, -21.7739, -21.7833, -21.7927, -21.6378, -21.6613, -21.9288, -22.0602, -22.356, -22.4451, -22.478, -22.6, -22.6563,
        -22.6657, -22.6657, -23.0412, -23.4918, -23.8673, -24.1208, -24.2428, -24.2757, -24.5714, -24.6512, -24.7826, -25.1346, -25.435, -25.7964, -26.1484, -26.4019, -26.6366, -26.9464, -27.2421,
        -27.6833, -27.6504, -27.5612, -27.8382, -27.5096, -27.3078, -27.3547, -27.0308, -26.6694, -26.369, -25.9091, -25.6415, -25.6087, -25.895, -25.7776, -26.0311, -26.3033, -26.6788, -27.0872,
        -27.4392, -27.7255, -27.7631, -27.9461, -28.2794, -28.3357, -28.0916, -28.2465, -28.4108, -28.6314, -28.9834, -28.6877, -28.4765, -27.9602, -27.4627, -26.998, -26.7023, -26.5709, -26.8243,
        -27.1764, -27.4063, -27.7161, -28.0024, -28.3122, -28.6549, -28.4249, -28.0024, -27.6927, -27.3735, -27.1998, -26.7774, -26.4254, -26.1625, -25.8762, -25.4772, -25.3693, -25.1815, -24.9938,
        -25.0032, -24.8811, -24.4963, -24.0973, -23.811, -23.5575, -23.1633, -22.8769, -22.5906, -22.3137, -22.1494],
   y = [74.2163, 74.2445, 74.3196, 74.4181, 74.451, 74.4416, 74.4651, 74.451, 74.505, 74.5261, 74.4181, 74.343, 74.3078, 74.275, 74.2633, 74.2022, 74.1717, 74.1412, 74.0779, 74.0685, 74.0286, 73.9957,
        74.0262, 73.9652, 73.9464, 73.9277, 73.923, 73.9042, 73.8737, 73.8314, 73.7892, 73.7047, 73.6179, 73.5498, 73.4911, 73.4466, 73.4442, 73.4794, 73.4113, 73.3715, 73.3339, 73.2518, 73.2447,
        73.2893, 73.3245, 73.409, 73.5029, 73.6179, 73.7563, 73.7657, 73.7516, 73.7399, 73.6672, 73.5498, 73.5615, 73.6179, 73.6343, 73.5263, 73.4747, 73.4372, 73.3973, 73.3316, 73.2799, 73.2259,
        73.2283, 73.3128, 73.3034, 73.287, 73.233, 73.1767, 73.1368, 73.0898, 73.1462, 73.1767, 73.1344, 73.0664, 73.0546, 73.0194, 72.9537, 72.8974, 72.8669, 72.8129, 72.7801, 72.7965, 72.7988,
        72.7472, 72.7284, 72.6815, 72.7073, 72.604, 72.6064, 72.6909, 72.773, 72.7918, 72.7308, 72.6721, 72.5594, 72.4773, 72.4374, 72.4186, 72.374, 72.3553, 72.252, 72.1628, 72.083, 72.1699, 72.2849,
        72.3623, 72.4186, 72.4046, 72.3553, 72.3036, 72.2614, 72.2379, 72.2145, 72.1746, 72.1159, 72.0971, 72.0549, 71.9798, 71.968, 71.9328, 71.9023, 71.8296, 71.7733, 71.6935, 71.5714, 71.5832,
        71.5855, 71.5902, 71.6066, 71.6817, 71.7662, 71.6794, 71.5691, 71.5926, 71.6723, 71.7099, 71.7099, 71.6231, 71.562, 71.4424, 71.271, 71.2851, 71.433, 71.4189, 71.4987, 71.4776, 71.3649,
        71.3133, 71.271, 71.1959, 71.1772, 71.0974, 71.0504, 71.0575, 70.9941, 70.9495, 70.8416, 70.8064, 70.7524, 70.6961, 70.6562, 70.5905, 70.574, 70.459, 70.4238, 70.5036, 70.4825, 70.6233,
        70.8017, 70.8087, 70.6961, 70.5693, 70.4403, 70.4144, 70.4778, 70.5811, 70.7336, 70.9261, 71.0622, 71.1725, 71.2804, 71.3227, 71.3133, 71.4283, 71.5362, 71.562, 71.609, 71.5573, 71.5761,
        71.7099, 71.8061, 71.7216, 71.6418, 71.616, 71.5785, 71.5362, 71.4635, 71.5128, 71.4541, 71.4776, 71.4611, 71.3626, 71.21, 71.271, 71.1678, 71.0762, 71.0152, 70.9495, 70.9401, 70.9472, 71.0434,
        70.9214, 70.9566, 70.9425, 70.8932, 70.7852, 70.6233, 70.5482, 70.5083, 70.459, 70.452, 70.3863, 70.3745, 70.398, 70.4262, 70.4121, 70.3346, 70.2901, 70.2783, 70.2079, 70.1328, 70.0882,
        70.1375, 70.0389, 70.0272, 70.0061, 70.0084, 70.0014, 70.1633, 70.2267, 70.1774, 70.2478, 70.2924, 70.3675, 70.2854, 70.33, 70.3487, 70.2713, 70.2901, 70.215, 70.1633, 70.1023, 70.1023,
        70.0695, 70.0624, 70.0695, 70.114, 70.1399] };
  WCL[1019] = struct {
   x = [-22.1471, -22.1001, -22.133, -22.1893],
   y = [70.1399, 70.1399, 70.1023, 70.0507] };
  WCL[1020] = struct {
   x = [-21.7599, -21.9711, -22.0039, -21.849, -21.4172, -21.2107, -20.9244, -20.6146, -20.3189, -20.2063, -20.4691, -20.4832, -20.5817, -20.9807, -21.2999, -21.5627, -21.7599],
   y = [74.3618, 74.2773, 74.1952, 74.1553, 74.1224, 74.0825, 74.1037, 74.1224, 74.1342, 74.1975, 74.2445, 74.3337, 74.3947, 74.4134, 74.4322, 74.4088, 74.3618] };
  WCL[1021] = struct {
   x = [-25.6415, -25.6885, -25.3364, -24.9375, -24.6183, -24.3414, -24.0973, -23.8908, -23.6373, -23.2712, -23.0412, -22.9943, -23.2477, -23.6232, -23.9659, -24.2991, -24.6512, -24.8952, -24.7403, -24.3648,
        -24.1771, -23.9002, -23.6373, -23.3839, -23.3839, -23.5247, -23.7781, -23.9659, -24.285, -24.7263, -24.9281, -25.083, -25.2707, -25.4209, -25.543, -25.6415],
   y = [73.2072, 73.1203, 73.0922, 73.0711, 73.0241, 73.0053, 73.0147, 73.0499, 73.064, 73.0711, 73.0992, 73.1509, 73.1696, 73.1696, 73.1884, 73.2236, 73.2612, 73.287, 73.2823, 73.2893, 73.2588, 73.2189,
        73.2166, 73.1978, 73.2447, 73.2776, 73.3081, 73.3433, 73.3785, 73.3879, 73.3996, 73.409, 73.3527, 73.2893, 73.233, 73.2072] };
  WCL[1022] = struct {
   x = [-22.0368, -22.2245, -22.248, -22.1494, -22.0274, -22.0931, -22.2668, -22.4357, -22.5766, -22.6986, -22.708, -22.8206, -22.9427, -23.0412, -23.2712, -23.4355, -23.6702, -23.8673, -24.1536, -24.2991,
        -24.5714, -24.6418, -24.4306, -24.1771, -23.811, -23.5716, -23.3182, -23.1163, -22.8957, -22.6751, -22.5015, -22.3231, -22.2011, -22.0462, -22.0368],
   y = [72.8716, 72.8317, 72.7965, 72.7472, 72.712, 72.7143, 72.6909, 72.712, 72.705, 72.7143, 72.7448, 72.7824, 72.8246, 72.8481, 72.8598, 72.8645, 72.8598, 72.881, 72.8974, 72.8997, 72.9091, 72.9608,
        72.9748, 72.9842, 73.0053, 73.0147, 73.0359, 73.0053, 73.01, 72.9748, 72.9772, 72.942, 72.9279, 72.8997, 72.8716] };
  WCL[1023] = struct {
   x = [-23.0623, -22.8277, -22.4099, -22.2316, -22.0673, -22.0016, -22.1987, -22.3301, -22.6305, -22.7948, -22.5507, -22.4099, -22.2222, -22.2316, -22.2316, -22.3536, -22.5507, -22.5648, -22.7948, -22.9732,
        -23.1938, -23.5693, -23.7899, -23.9964, -24.1325, -24.2639, -24.386, -24.4047, -24.4939, -24.4939, -24.5362, -24.569, -24.4517, -24.3062, -24.1419, -23.9213, -23.6678, -23.5129, -23.3674,
        -23.2032, -23.0623],
   y = [72.7214, 72.6228, 72.5735, 72.4937, 72.5031, 72.3858, 72.3952, 72.4233, 72.428, 72.367, 72.3224, 72.2567, 72.2473, 72.2004, 72.1136, 72.0995, 72.1441, 72.2051, 72.1675, 72.245, 72.3318, 72.3975,
        72.4186, 72.4585, 72.5336, 72.5548, 72.5923, 72.6557, 72.6885, 72.7284, 72.8082, 72.8317, 72.8551, 72.8669, 72.8598, 72.8458, 72.8246, 72.8458, 72.8223, 72.7918, 72.7214] };
  WCL[1024] = struct {
   x = [-25.4538, -25.4444, -25.4444, -25.3881, -25.4538, -25.6227, -25.8199, -25.9513, -26.017, -26.1391, -26.369, -26.6694, -26.9464, -27.2186, -27.4955, -27.6504, -27.8945, -28.0681, -28.0588, -28.0494,
        -28.0024, -27.6269, -27.2327, -27.1529, -26.89, -26.416, -26.1625, -25.9091, -25.8293, -25.6885, -25.5242, -25.4538],
   y = [70.8627, 70.7688, 70.7453, 70.635, 70.6092, 70.5811, 70.574, 70.5881, 70.5341, 70.5412, 70.567, 70.5224, 70.5107, 70.4778, 70.4567, 70.4567, 70.4332, 70.445, 70.4966, 70.5506, 70.6233, 70.7054,
        70.8087, 70.8486, 70.8744, 70.9143, 70.9941, 71.0434, 71.0011, 70.9542, 70.9167, 70.8627] };
  WCL[1025] = struct {
   x = [-50.4335, -50.3912],
   y = [69.9075, 70.0061] };
  WCL[1026] = struct {
   x = [-50.9639, -51.0624, -51.3394, -51.4943, -51.8134, -52.3861, -52.8179, -53.414, -53.9115, -54.1837, -54.4935, -54.2635, -53.9631, -53.6346, -53.2168, -52.8179, -52.4189, -52.034, -51.6726, -51.1845,
        -50.6541, -50.6118, -50.6869, -51.0437, -51.2079, -51.2643, -50.9639, -51.0296, -50.809, -50.809, -51.0953, -50.701, -50.809, -51.0531, -51.3159, -51.6257, -51.9026, -51.3957, -51.3394,
        -51.4708, -51.3957, -51.8228, -52.1889, -51.9355, -51.6726, -52.0012, -52.5316, -52.3532, -52.264, -51.9026, -52.4987, -52.9962, -52.7756, -52.1091, -51.8698, -52.5973, -52.7756, -52.7615,
        -52.8742, -53.1277, -53.1042, -52.9399, -53.2356, -53.4234, -53.5454, -53.5689, -53.8739, -53.8317, -53.9537, -54.0945, -54.1743, -53.9866, -54.0101, -54.5358, -54.9019, -55.1788, -55.3665,
        -55.6622, -55.7514, -55.5731, -55.2633, -55.0474, -54.8127, -55.2304, -55.5402, -55.2867, -55.0239, -54.9347, -54.869, -54.747, -54.8268, -54.6484, -54.6906, -54.7141, -54.6906, -54.5827,
        -54.5592, -54.7047, -55.0474, -55.3196, -55.6388, -55.4745, -55.3853, -55.1976, -55.3994, -55.3525, -55.2774, -55.1319, -55.451, -55.5965, -55.5637, -55.498, -55.5214, -55.5637, -55.6622,
        -55.8594, -55.9955, -56.0049, -55.972, -55.7937, -55.8171, -55.7608, -55.8922, -56.1128, -55.9626, -56.08, -56.3006, -56.2489, -56.2489, -56.1832, -56.2255, -56.4132, -56.6432, -56.2255,
        -56.2818, -56.4883, -56.8216, -56.3804, -56.2489, -56.4461, -56.5681, -56.7652, -57.0187, -57.1079, -57.0422, -56.8873, -56.9201, -57.1501, -57.3191, -57.6383, -57.9105, -58.0654, -58.1217,
        -58.3189, -58.3001, -58.3423, -58.5301, -58.4644, -58.286, -58.2672, -58.3423, -58.3658, -58.3986, -58.4972, -58.5629, -58.7084, -58.7835, -58.8305, -58.8915, -59.0135, -59.1356, -59.1121,
        -59.1402, -59.2247],
   y = [69.9897, 70.0272, 70.0436, 70.0014, 70.0249, 70.0695, 70.2455, 70.3346, 70.3863, 70.5247, 70.6796, 70.7852, 70.7993, 70.7759, 70.7547, 70.7453, 70.6609, 70.5482, 70.4332, 70.3675, 70.33, 70.4051,
        70.4731, 70.4262, 70.4919, 70.5881, 70.5482, 70.6069, 70.6163, 70.6679, 70.7289, 70.682, 70.7735, 70.8697, 70.8955, 70.9495, 71.0504, 70.9753, 71.0223, 71.0692, 71.1467, 71.1255, 71.1138,
        71.2288, 71.3274, 71.264, 71.1537, 71.2523, 71.3649, 71.44, 71.3766, 71.4001, 71.494, 71.5644, 71.6653, 71.6231, 71.6301, 71.6489, 71.6653, 71.6935, 71.7733, 71.8953, 71.7991, 71.9, 71.9939,
        71.8202, 71.7545, 71.6958, 71.6489, 71.6888, 71.6207, 71.5503, 71.4494, 71.3555, 71.3931, 71.4846, 71.4048, 71.5832, 71.6747, 71.7615, 71.8061, 71.8953, 71.9704, 71.9235, 72.0244, 72.0619,
        72.1699, 72.2215, 72.3107, 72.3553, 72.4421, 72.4656, 72.5501, 72.6463, 72.7214, 72.7495, 72.8505, 72.942, 72.996, 73.0194, 73.0452, 73.0851, 73.1555, 73.1907, 73.2447, 73.3057, 73.3527,
        73.3973, 73.3433, 73.3363, 73.3691, 73.4466, 73.5029, 73.5334, 73.5428, 73.5897, 73.6226, 73.6648, 73.7211, 73.6953, 73.7329, 73.8361, 73.8831, 73.9417, 74.0028, 74.045, 74.0309, 74.1506,
        74.1576, 74.2492, 74.275, 74.2961, 74.3313, 74.3313, 74.4088, 74.397, 74.4228, 74.4557, 74.4815, 74.5496, 74.6082, 74.6693, 74.688, 74.7467, 74.7842, 74.8195, 74.857, 74.8922, 74.9368, 74.9884,
        75.0424, 75.1081, 75.2043, 75.2935, 75.324, 75.3897, 75.3311, 75.3968, 75.4789, 75.5282, 75.547, 75.6033, 75.6221, 75.6643, 75.723, 75.6972, 75.7418, 75.7465, 75.8474, 75.777, 75.7676, 75.8779,
        75.9225, 75.9131] };
  WCL[1027] = struct {
   x = [-54.348, -54.2166, -53.9772, -53.644, -53.1934],
   y = [70.3276, 70.3112, 70.2642, 70.2478, 70.1774] };
  WCL[1028] = struct {
   x = [-54.5264, -54.7564, -54.747, -54.517, -54.348],
   y = [69.9709, 70.1187, 70.2384, 70.2901, 70.3276] };
  WCL[1029] = struct {
   x = [-70.0976, -69.9427],
   y = [70.8064, 70.8228] };
  WCL[1030] = struct {
   x = [-70.1422, -69.9216, -69.8137, -69.5461, -69.3396, -69.1284, -69.0861, -68.9312, -68.72, -68.49, -68.476, -68.5229, -68.5651, -68.7106, -68.8186, -68.8514, -69.1284, -69.4053, -69.6681, -69.579,
        -69.3584, -69.2035, -68.9406, -68.6966, -68.6966, -68.8092],
   y = [70.5905, 70.6961, 70.7148, 70.7383, 70.7125, 70.6562, 70.6538, 70.635, 70.6092, 70.567, 70.4403, 70.3558, 70.3675, 70.4379, 70.3464, 70.3018, 70.2713, 70.2267, 70.1516, 70.1023, 70.1563, 70.1563,
        70.168, 70.1704, 70.0953, 70.0319] };
  WCL[1031] = struct {
   x = [-68.7083, -68.5628, -68.2765, -68.3093, -68.1779, -67.8259, -67.4504],
   y = [69.931, 70.0178, 70.0953, 70.1845, 70.2783, 70.222, 70.0319] };
  WCL[1032] = struct {
   x = [-68.95, -68.9782, -68.9594, -68.8561, -68.6308, -68.5088, -68.3962, -68.1615, -67.9174, -67.6827, -67.4387, -67.2415, -67.035, -66.7815, -66.5844, -66.4624, -66.1808, -65.9367, -65.9085, -66.1714,
        -66.4342, -66.6877, -67.0538, -67.3729, -67.7484, -68.1145, -68.4149, -68.6496, -68.95],
   y = [77.2649, 77.2766, 77.2789, 77.3001, 77.3188, 77.347, 77.3845, 77.3845, 77.3916, 77.3963, 77.3963, 77.4057, 77.3963, 77.3775, 77.347, 77.3024, 77.2789, 77.2531, 77.225, 77.1874, 77.1663, 77.1592,
        77.1874, 77.2132, 77.232, 77.2109, 77.2156, 77.2414, 77.2649] };
  WCL[1033] = struct {
   x = [-70.2642, -69.9638, -69.6447, -69.2316, -68.8374, -68.6121, -68.5557, -68.5088, -68.3117, -68.2272, -68.1427, -67.955, -67.6639, -67.3166, -66.9693, -66.5938, -66.3497, -66.1808, -66.115, -65.7959,
        -65.6832, -65.6551, -65.7114, -66.0118, -66.2746, -66.3779, -66.4248, -66.2746, -65.9742, -65.824, -65.7583, -65.824, -65.9742, -66.0118, -66.0963, -66.3591, -66.4905, -66.622, -66.7252,
        -66.4624, -66.4248, -66.6126, -66.9224, -67.1758, -67.4762, -67.6639, -67.8611, -68.2272, -68.3117, -68.5557, -68.7341, -68.6027, -68.6215, -69.0251, -69.3349, -69.6447],
   y = [77.2273, 77.2297, 77.2203, 77.2179, 77.1921, 77.1874, 77.1851, 77.178, 77.1686, 77.1639, 77.1686, 77.1874, 77.1991, 77.178, 77.1522, 77.1381, 77.1076, 77.1123, 77.1569, 77.1874, 77.2179, 77.2484,
        77.2719, 77.2977, 77.3353, 77.3798, 77.4268, 77.4667, 77.4596, 77.4972, 77.5676, 77.5887, 77.6075, 77.6615, 77.7014, 77.6849, 77.6357, 77.6263, 77.6568, 77.6873, 77.7342, 77.7272, 77.6638,
        77.5676, 77.5535, 77.5441, 77.5465, 77.5418, 77.6075, 77.6802, 77.6685, 77.5793, 77.5207, 77.4784, 77.4855, 77.5207] };
  WCL[1034] = struct {
   x = [-70.1187, -69.9498, -69.6212, -69.6306, -69.9967],
   y = [77.6122, 77.631, 77.692, 77.7272, 77.6826] };
  WCL[1035] = struct {
   x = [-70.1094, -69.7714, -69.5743, -69.4804, -69.4335, -69.2739, -69.0861, -68.9829, -68.9547, -69.2363, -69.3396, -69.1988, -69.0204, -68.72, -68.4666, -68.1286, -67.9127, -67.9033, -67.678, -67.5842,
        -67.4621, -67.2462, -67.1148, -67.0491, -66.9646, -66.8332, -66.6642, -66.5609, -66.4952, -66.2793, -66.2042, -66.0446, -65.9789, -65.932, -65.7818, -65.6128, -65.5096, -65.3218, -65.2561,
        -65.1716, -65.0214, -64.8243, -64.8431, -64.8431, -64.9182, -65.0684, -65.2373, -65.1904, -65.1247, -64.8431, -64.5708, -64.4018, -64.5051, -64.6271, -64.8337, -65.0684, -64.8806],
   y = [78.802, 78.8114, 78.8091, 78.8278, 78.8607, 78.8748, 78.856, 78.8537, 78.8983, 78.9147, 78.9522, 78.9874, 79.0156, 79.0555, 79.0743, 79.0813, 79.0602, 79.0954, 79.1095, 79.1259, 79.1235, 79.147,
        79.1142, 79.0813, 79.0883, 79.1118, 79.1048, 79.0977, 79.0813, 79.0837, 79.0625, 79.0883, 79.1189, 79.1611, 79.2127, 79.2667, 79.3066, 79.3371, 79.3723, 79.4075, 79.4521, 79.492, 79.5225,
        79.5788, 79.6258, 79.6751, 79.7032, 79.7548, 79.8018, 79.7854, 79.7971, 79.8417, 79.8745, 79.9168, 79.9097, 79.9402, 79.9825] };
  WCL[1036] = struct {
   x = [-65.4978, -65.7137, -65.986, -66.1643],
   y = [80.0224, 79.9848, 79.9543, 79.9895] };
  WCL[1037] = struct {
   x = [-59.8865, -60.2057, -60.6281, -60.6657, -60.6563, -60.6657, -60.6845, -60.8065, -60.8534, -60.9098, -60.9285, -61.0693, -61.4355, -61.7265, -61.9705, -62.102, -62.3085, -62.4305, -62.6464, -62.8154,
        -63.1815, -63.4068, -63.7447, -63.9043, -63.9513, -63.9606, -64.2141, -64.43, -64.5051, -64.6459, -64.89, -65.0777, -65.2749, -65.5847, -65.5471, -65.4251, -65.7067, -65.9601, -66.1385,
        -66.2042, -66.392, -66.6548, -66.7111, -66.927, -67.2087, -66.9083, -66.6079, -66.2981, -66.7862, -67.2274, -67.2744, -67.5842, -68.1568, -68.5135, -68.8139, -69.011, -69.302, -69.5273,
        -69.2739, -68.8608, -68.4853, -68.1662, -68.138, -68.5604, -68.6449, -68.9265, -69.2082, -69.4804, -69.8841],
   y = [75.9131, 75.9553, 76.0492, 76.0492, 76.0163, 75.9553, 75.9929, 76.021, 76.0633, 76.1032, 76.1079, 76.1243, 76.1337, 76.1642, 76.1806, 76.1923, 76.1994, 76.1783, 76.2017, 76.2933, 76.312, 76.1877,
        76.1126, 76.1665, 76.1923, 76.2792, 76.2416, 76.183, 76.1149, 76.1313, 76.1032, 76.0093, 75.9717, 76.0468, 76.1595, 76.2205, 76.2416, 76.2628, 76.1783, 76.0891, 76.0961, 76.1994, 76.2158,
        76.2158, 76.1219, 76.075, 75.9647, 75.8497, 75.9037, 75.9272, 75.9295, 75.9788, 76.014, 76.0891, 76.1525, 76.2322, 76.2956, 76.3824, 76.4575, 76.5514, 76.5514, 76.6101, 76.6711, 76.6476,
        76.6453, 76.6664, 76.6688, 76.6922, 76.7345] };
  WCL[1038] = struct {
   x = [-70.0084, -69.8301, -69.6423, -69.7644, -69.924],
   y = [76.833, 76.9152, 76.9879, 77.0067, 76.9293] };
  WCL[1039] = struct {
   x = [-79.384, -79.1587, -79.1306, -79.0837, -79.1306, -79.2714, -79.5249, -79.844],
   y = [76.9692, 76.9997, 77.0161, 77.0536, 77.0912, 77.1522, 77.1639, 77.1616] };
  WCL[1040] = struct {
   x = [-80.2547, -79.9074, -79.5882, -79.3535, -79.147, -79.0156, -78.8654, -78.8936, -78.8842, -78.6119, -78.3491, -78.1989, -78.0956, -78.0675, -78.0862, -77.7765, -77.7577, -77.983, -78.1144, -78.1332,
        -78.0862, -78.1989, -78.3303, -78.2928, -78.2928, -78.3491, -78.3491, -78.1144, -77.8328, -77.6075, -77.4479, -77.2789, -77.1569, -76.9504, -76.7908, -76.6875, -76.5561, -76.4247, -76.1337,
        -76.0774, -76.0023, -75.8802, -75.8051, -75.8145, -75.7676, -75.6831, -75.7957, -76.0304, -76.2088, -76.312, -76.5092, -76.65, -76.8847, -77.0818, -77.0349, -76.8659, -76.7626, -76.5373,
        -76.3871, -76.2463, -75.9553, -75.7488, -75.6831, -75.6455, -75.608, -75.5141, -75.3733, -75.1762, -75.1386, -75.3076, -75.5423, -75.7676, -75.9835, -76.2088, -76.4528, -76.3027, -76.0023,
        -75.7582, -75.5423, -75.3264, -75.1105, -75.0072, -74.857, -74.7256, -74.857, -74.857, -74.8101, -74.8007, -74.7913, -74.8007, -74.8007, -74.8007, -74.8288, -74.857, -75.0541, -75.3639,
        -75.5704, -75.8145, -76.0867, -76.312, -76.3027, -76.021, -75.8802, -75.8802, -75.9553, -76.0116, -76.1243, -76.3214, -76.5373, -76.7814, -77.0818, -77.3165, -77.5606, -77.7295, -77.8609,
        -77.7201, -77.6638, -77.8797, -77.9173, -77.645, -77.4761, -77.2038, -76.9316, -76.6781, -76.359, -76.19, -76.19, -76.359, -76.6124, -76.9034, -77.232, -77.4667, -77.7765, -77.936, -77.6826,
        -77.3822, -77.1851, -76.8471, -76.6406, -76.3308, -76.021, -75.9178, -75.9365, -75.899, -75.8427, -75.6643, -75.4484, -75.2325, -74.9509, -74.735, -74.5942, -74.5566, -74.5754, -74.688,
        -74.8288, -74.7444, -74.6223, -74.7537, -75.0541, -75.4296, -75.6174, -75.777, -75.9647, -76.068, -76.2369, -76.5373, -76.8659, -77.1569, -77.3728, -77.523, -77.6357, -77.4573, -77.4385,
        -77.6544, -77.8328, -77.9173, -77.983, -77.8234, -77.692, -77.5136, -77.2226, -77.2602, -77.3259, -77.4197, -77.2883, -77.2038, -77.1193, -76.9504, -76.7251, -76.4622, -76.3027, -76.068,
        -76.0023, -76.1994, -76.2463, -76.4528, -76.7532, -76.9598, -77.1475, -77.2789, -77.1475, -77.0912, -77.0067, -76.941, -76.7251, -76.5092, -76.2839, -76.0961, -75.9553, -75.7488, -75.486,
        -75.2043, -75.0729, -75.1292, -74.9978, -74.9509, -74.9227, -74.6786, -74.3876, -74.2374, -74.2562, -74.1717, -74.0872, -73.9746, -73.693, -73.6179, -73.6366, -73.7493, -73.7868, -74.1154,
        -74.444, -74.857, -74.7913, -74.4533, -74.2562, -74.0591, -73.7587, -73.5428, -73.5897, -73.5428, -73.3832, -73.2799, -73.0828, -73.0171, -73.0171, -72.8481, -72.5947, -72.36, -72.1628],
   y = [77.225, 77.2297, 77.232, 77.2297, 77.2367, 77.2907, 77.2766, 77.2508, 77.2414, 77.2789, 77.3071, 77.3705, 77.4244, 77.4878, 77.516, 77.5418, 77.5699, 77.6263, 77.6474, 77.6849, 77.7201, 77.7835,
        77.8046, 77.8093, 77.8539, 77.8962, 77.922, 77.8962, 77.8915, 77.8821, 77.8868, 77.8915, 77.8563, 77.8633, 77.8844, 77.922, 77.9501, 77.9712, 77.9431, 77.9126, 77.9431, 77.9783, 78.0065,
        78.0205, 78.0463, 78.0769, 78.0816, 78.0886, 78.1027, 78.098, 78.1144, 78.1144, 78.1379, 78.1496, 78.166, 78.1919, 78.2036, 78.1965, 78.1965, 78.1825, 78.166, 78.1402, 78.159, 78.1825, 78.2059,
        78.2458, 78.2623, 78.281, 78.3115, 78.3702, 78.4031, 78.4148, 78.4242, 78.4594, 78.4688, 78.4899, 78.4782, 78.4805, 78.4852, 78.4922, 78.5087, 78.5392, 78.5486, 78.5767, 78.6049, 78.6448,
        78.673, 78.673, 78.68, 78.7035, 78.7222, 78.741, 78.7739, 78.8114, 78.8466, 78.8748, 78.8818, 78.8724, 78.8677, 78.863, 78.8912, 78.91, 78.9358, 78.9546, 78.9733, 78.9898, 78.9851, 78.978,
        78.9851, 78.9874, 78.9757, 78.9733, 78.9475, 78.9029, 78.8983, 78.9522, 78.9968, 79.0015, 79.032, 79.025, 79.0109, 79.0179, 79.0438, 79.0555, 79.0484, 79.0484, 79.0766, 79.0977, 79.1095,
        79.1189, 79.1212, 79.1212, 79.1189, 79.14, 79.1564, 79.1611, 79.1494, 79.1611, 79.1869, 79.1775, 79.1611, 79.1306, 79.086, 79.0625, 79.0414, 79.0203, 79.0203, 79.0086, 79.0086, 78.9945,
        79.0086, 79.0438, 79.0907, 79.1118, 79.1517, 79.1799, 79.2104, 79.2338, 79.2268, 79.2268, 79.2245, 79.2198, 79.2221, 79.2503, 79.2456, 79.2385, 79.2362, 79.2268, 79.2127, 79.1963, 79.194,
        79.2221, 79.2479, 79.269, 79.269, 79.2831, 79.2972, 79.3136, 79.316, 79.3136, 79.3066, 79.3441, 79.3887, 79.4099, 79.4169, 79.3981, 79.3535, 79.3324, 79.3277, 79.3254, 79.323, 79.3183, 79.3441,
        79.384, 79.4099, 79.4169, 79.4286, 79.4474, 79.4451, 79.485, 79.492, 79.4873, 79.492, 79.4897, 79.4709, 79.4709, 79.4521, 79.4122, 79.4099, 79.3817, 79.3723, 79.3653, 79.3934, 79.4474, 79.4357,
        79.431, 79.4239, 79.4239, 79.4169, 79.4615, 79.4967, 79.4897, 79.5225, 79.4803, 79.492, 79.546, 79.5929, 79.661, 79.7056, 79.7103, 79.7361, 79.7431, 79.7877, 79.8018, 79.7877, 79.7619, 79.7478,
        79.7455, 79.7079, 79.6657, 79.6445, 79.6375, 79.6422, 79.6445, 79.6258, 79.6352, 79.6492, 79.6492, 79.6469] };
  WCL[1041] = struct {
   x = [-72.1605, -71.9915, -71.7756, -71.6254, -71.5034, -71.3719, -71.5315, -71.8319, -71.7944, -71.6066, -71.2311, -71.1654],
   y = [79.6469, 79.6751, 79.6821, 79.7196, 79.7548, 79.8088, 79.8229, 79.8159, 79.8558, 79.905, 79.9309, 79.959] };
  WCL[1042] = struct {
   x = [-72.2473, -72.7073],
   y = [80.006, 79.9989] };
  WCL[1043] = struct {
   x = [-80.189, -79.9543, -79.7478, -79.7478, -79.7666, -79.7666, -79.79, -79.79, -79.9215, -79.9872, -79.9449],
   y = [75.1738, 75.2043, 75.2841, 75.3193, 75.3827, 75.4578, 75.4648, 75.4719, 75.4672, 75.5047, 75.5352] };
  WCL[1044] = struct {
   x = [-79.2292, -79.2104, -79.2385, -79.2949, -79.3606, -79.1822, -79.079, -79.1447, -79.323, -79.492, -79.492, -79.5765, -79.6891, -79.8112, -79.905, -79.8487, -79.7924, -79.7924, -79.6891, -79.5953,
        -79.5859, -79.5295, -79.5295, -79.4357, -79.3043, -79.2292],
   y = [76.0468, 76.0398, 75.9905, 75.9624, 75.9107, 75.8685, 75.8145, 75.8122, 75.8215, 75.8098, 75.7746, 75.7605, 75.7676, 75.7981, 75.8051, 75.838, 75.8685, 75.9037, 75.9506, 75.9976, 76.0351, 76.0609,
        76.0844, 76.0938, 76.075, 76.0468] };
  WCL[1045] = struct {
   x = [-79.7032, -79.5624, -79.4498, -79.6046, -79.661, -79.661, -79.7924],
   y = [74.8476, 74.864, 74.9063, 74.9438, 75.0025, 75.033, 75.0283] };
  WCL[1046] = struct {
   x = [-80.0459, -79.9473, -79.8581, -79.7924, -79.7032],
   y = [74.8171, 74.8171, 74.8382, 74.8453, 74.8476] };
  WCL[1047] = struct {
   x = [-77.7084, -77.699, -77.7178, -77.9525, -78.1825, -78.2716, -78.4922, -78.7035, -78.8865, -79.2855, -79.384, -79.5953, -79.431, -79.1541, -78.8865],
   y = [69.9427, 70.0624, 70.2079, 70.1845, 70.2408, 70.1751, 70.2478, 70.3276, 70.3863, 70.4332, 70.4332, 70.3628, 70.3182, 70.3229, 70.1117] };
  WCL[1048] = struct {
   x = [-80.1444, -79.9379, -79.8722, -79.8816, -79.8252, -79.7032, -79.4638, -79.262, -79.1306, -78.9335, -78.6237, -78.4031, -78.6894, -78.6659, -78.3937, -78.2388, -78.0604, -77.699, -77.2015, -76.7931,
        -76.3191, -76.1313, -75.7558, -75.371, -75.4367, -75.2724, -75.0846, -75.0283, -75.2395, -75.1269, -74.7514, -74.3665, -74.3102, -74.4322, -74.742, -74.8875, -75.0706, -74.7749, -74.7983,
        -74.6528, -74.3665, -74.0896, -73.5827, -73.5498, -73.7047, -73.8596, -74.0145, -74.1037, -73.8361, -73.6484, -73.6953, -73.5498, -73.5263, -73.2964, -73.287, -73.2541, -73.3621, -73.1321,
        -73.0429, -72.8786, -72.658, -72.5923, -72.5454, -72.428, -72.2074, -71.8413, -71.4424, -71.1889, -71.0903, -71.1889, -71.325, -71.6113, -72.1276, -71.8742, -71.5785, -71.433, -71.1467,
        -70.8603, -70.6491, -70.5834, -70.7383, -70.8838, -70.8932, -70.5599, -70.3065, -70.1],
   y = [72.3576, 72.4984, 72.4046, 72.3154, 72.3107, 72.2872, 72.3647, 72.2872, 72.2239, 72.3036, 72.3036, 72.3342, 72.4139, 72.5313, 72.6205, 72.658, 72.6815, 72.7308, 72.7214, 72.6486, 72.5923, 72.5594,
        72.5571, 72.5008, 72.4468, 72.4186, 72.2849, 72.1933, 72.0995, 72.0619, 72.0666, 72.0291, 71.9023, 71.8272, 71.7615, 71.7169, 71.6559, 71.7075, 71.5972, 71.6278, 71.7474, 71.7404, 71.7826,
        71.7474, 71.6888, 71.6066, 71.5081, 71.4424, 71.5573, 71.5433, 71.4118, 71.3579, 71.3227, 71.3414, 71.2922, 71.3485, 71.5268, 71.548, 71.5221, 71.5081, 71.5926, 71.6888, 71.6935, 71.6489,
        71.5902, 71.5409, 71.4752, 71.3367, 71.2077, 71.1349, 71.0645, 71.0716, 71.0317, 71.0082, 71.0011, 70.9284, 70.9542, 71.0317, 71.0223, 70.9073, 70.7665, 70.689, 70.6233, 70.7125, 70.7453,
        70.8064] };
  WCL[1049] = struct {
   x = [-69.9427, -70.0835, -70.3276, -70.4403, -70.4403, -70.2525, -70.1399],
   y = [70.8228, 70.7008, 70.621, 70.5247, 70.4895, 70.567, 70.5905] };
  WCL[1050] = struct {
   x = [-80.3204, -79.8135, -79.485, -79.2784, -79.0907, -78.9874, -78.5838, -78.5181, -78.152, -78.0956, -78.0487, -78.1144, -78.2834, -78.6307, -78.8466, -78.9874, -79.1846, -79.1658, -79.3254, -79.5507,
        -79.5131, -79.3911],
   y = [76.2205, 76.2604, 76.3472, 76.4341, 76.5256, 76.3965, 76.4646, 76.5303, 76.5678, 76.6312, 76.7415, 76.8612, 76.9433, 76.9152, 76.8542, 76.7931, 76.8049, 76.8729, 76.8894, 76.9199, 76.948, 76.9692] };
  WCL[1051] = struct {
   x = [-78.964, -78.7434, -78.6307, -78.2458, -77.9501, -77.6403, -77.3306, -77.1991, -77.1663, -77.0114, -76.833, -76.6688, -76.5373, -76.3402, -76.3731, -76.3167, -76.1525, -76.3073, -76.6359, -77.0771,
        -77.5746, -77.936, -78.2693, -78.6542, -78.9968, -79.4286, -79.79],
   y = [73.6038, 73.6155, 73.6226, 73.6155, 73.5686, 73.5052, 73.4653, 73.3996, 73.3456, 73.2799, 73.2447, 73.1462, 73.0851, 73.0241, 72.9842, 72.9256, 72.8669, 72.7965, 72.7894, 72.8035, 72.8598, 72.8669,
        72.8669, 72.7988, 72.7378, 72.6979, 72.7542] };
  WCL[1052] = struct {
   x = [-80.0904, -79.6492, -79.1189, -78.964],
   y = [73.6648, 73.6085, 73.578, 73.6038] };
  WCL[1053] = struct {
   x = [-71.2992, -71.2992, -71.271, -71.1584, -71.1396, -70.9988, -70.7547, -70.7829, -70.4919, -70.2666],
   y = [76.9997, 77.0044, 77.0255, 77.0466, 77.0771, 77.124, 77.1311, 77.178, 77.2109, 77.2273] };
  WCL[1054] = struct {
   x = [-69.6447, -70.0296, -70.1234],
   y = [77.5207, 77.5488, 77.6122] };
  WCL[1055] = struct {
   x = [-69.9991, -70.2056, -70.5811, -70.4684, -70.1868, -70.0929, -70.4966, -70.8251, -71.2288, -71.3227, -71.257, -71.5198, -71.7357, -71.8765, -72.0267, -72.2051, -72.2614, -72.2614, -72.2426, -72.3271,
        -72.5806, -72.7026, -72.8434, -72.9279, -72.7402, -72.6744, -72.6744, -72.4867, -72.4491, -72.7965, -72.8246, -72.7026, -72.6369, -72.6463, -72.59, -72.3459, -72.0737, -71.8859, -71.7263,
        -71.5292, -71.3321, -71.2476, -71.21, -71.0974, -71.0035, -70.9284, -70.8345, -70.5717, -70.3652, -70.2056, -70.1117],
   y = [77.6826, 77.6779, 77.6826, 77.7506, 77.8281, 77.8586, 77.8304, 77.7929, 77.7812, 77.8304, 77.8797, 77.8821, 77.9196, 77.9102, 77.9314, 77.9431, 77.9666, 77.9712, 78.0135, 78.0557, 78.0816, 78.1144,
        78.1379, 78.1731, 78.1919, 78.2482, 78.274, 78.2763, 78.2881, 78.3068, 78.3444, 78.3866, 78.4195, 78.4383, 78.4922, 78.5345, 78.5462, 78.5673, 78.619, 78.6331, 78.6448, 78.6354, 78.6354,
        78.619, 78.6096, 78.6753, 78.7269, 78.7434, 78.7574, 78.7833, 78.802] };
  WCL[1056] = struct {
   x = [-69.8888, -70.0108],
   y = [76.7345, 76.833] };
  WCL[1057] = struct {
   x = [-69.9216, -70.2032, -70.5599, -70.851, -70.8322, -70.8885, -71.095, -71.2171, -71.3203, -71.433, -71.2922],
   y = [76.9293, 76.7931, 76.772, 76.8236, 76.8776, 76.9128, 76.9339, 76.9551, 76.9668, 76.9598, 77.0044] };
  WCL[1058] = struct {
   x = [-87.2459, -87.1708, -87.1615, -87.1145, -87.1708, -87.3117, -87.3023, -87.0582, -87.03, -87.03, -87.0957, -87.1896, -87.2741, -87.3117, -87.5088, -87.6214, -87.6214, -87.5463, -87.4337, -87.3492,
        -87.2366, -87.1615, -87.0488, -86.9174, -86.8329, -86.7109, -86.6076, -86.4856, -86.4199, -86.279, -86.2603, -86.2697, -86.2227, -86.3354, -86.326, -86.2415, -86.1382, -86.0819, -86.0913,
        -86.0819, -86.0162, -85.9974, -85.9599, -85.866, -85.8566, -85.7815, -85.6689, -85.4905, -85.3497, -85.3309, -85.284, -85.1901, -85.1244, -85.0775, -85.0962, -85.0962, -85.1995, -85.3309,
        -85.4717, -85.406, -85.5093, -85.7815, -86.0913, -86.3729, -86.5794, -86.739, -86.7672, -86.7109, -86.7766, -86.9831, -86.9925, -87.0206, -87.03, -87.0957, -87.1239, -87.2084, -87.2178,
        -87.2084, -87.1802, -87.1708, -87.4243, -87.537, -87.6778, -87.781, -87.9312, -88.0908, -88.1284, -88.1377, -88.1377, -88.1941, -88.2222, -88.2035, -88.1941, -88.1847, -88.1377, -88.1377,
        -88.1941, -88.2316, -88.2879, -88.3912, -88.3912, -88.4287, -88.4194, -88.3255, -88.3536, -88.2879, -88.1565, -88.1002, -88.119, -88.1377, -88.1377, -88.2504, -88.3912, -88.4757, -88.6353,
        -88.7291, -88.7573, -88.823, -88.8606, -88.8887, -88.87, -88.7479, -88.6353, -88.5602, -88.5977, -88.6353, -88.6916, -88.7573, -88.8887, -89.0765, -89.1703, -89.2736, -89.3581, -89.3863,
        -89.4801, -89.4895, -89.5177, -89.6491, -89.7711, -89.8932, -89.9683],
   y = [80.0505, 79.9895, 79.9778, 79.9473, 79.9379, 79.9074, 79.8933, 79.9121, 79.9144, 79.898, 79.8628, 79.8065, 79.7408, 79.6891, 79.6704, 79.6281, 79.5859, 79.5906, 79.6234, 79.6539, 79.6539, 79.6539,
        79.6797, 79.7032, 79.6985, 79.6751, 79.6657, 79.6234, 79.6023, 79.5788, 79.5413, 79.5131, 79.4732, 79.4521, 79.4263, 79.4474, 79.4615, 79.492, 79.5108, 79.5295, 79.5624, 79.5859, 79.6, 79.5976,
        79.6023, 79.5741, 79.5225, 79.4756, 79.4474, 79.4286, 79.3958, 79.3747, 79.3606, 79.3324, 79.2761, 79.2432, 79.2221, 79.194, 79.1752, 79.1893, 79.1587, 79.1165, 79.0907, 79.0649, 79.0414,
        79.0297, 78.978, 78.9687, 78.9405, 78.9029, 78.9452, 78.9733, 79.0015, 79.0179, 79.025, 79.0344, 78.9921, 78.9452, 78.9006, 78.8466, 78.7692, 78.7011, 78.6354, 78.6237, 78.6612, 78.7199,
        78.7856, 78.849, 78.8842, 78.9147, 78.9311, 78.9522, 78.9663, 79.0062, 79.0438, 79.0719, 79.0555, 79.0062, 78.9616, 78.9006, 78.8372, 78.795, 78.7363, 78.6636, 78.6237, 78.6213, 78.5908,
        78.5275, 78.4782, 78.4617, 78.4265, 78.4289, 78.4265, 78.4782, 78.5275, 78.5533, 78.5908, 78.5932, 78.5556, 78.5228, 78.4617, 78.4125, 78.3679, 78.3514, 78.2904, 78.2271, 78.159, 78.1144,
        78.1191, 78.166, 78.2106, 78.2623, 78.2975, 78.3374, 78.3467, 78.3467, 78.3538, 78.389, 78.4007, 78.4406, 78.4993] };
  WCL[1059] = struct {
   x = [-90.0809, -89.9589, -89.9026, -89.7054, -89.5177, -89.3956, -89.4332, -89.5177, -89.7524, -89.9026],
   y = [78.4265, 78.3749, 78.335, 78.281, 78.1801, 78.1144, 78.0933, 78.1402, 78.1637, 78.2271] };
  WCL[1060] = struct {
   x = [-88.6376, -88.6376, -88.5437, -88.4874, -88.4499, -88.4405, -88.4311, -88.356, -88.2809, -88.187, -88.1307, -88.0931, -88.0087, -87.9523, -87.9148, -87.8303, -87.7928, -87.7834, -87.6801, -87.5393,
        -87.5299, -87.6144, -87.5956, -87.5393, -87.4736, -87.4079, -87.3046, -87.2295, -87.2107, -87.0981, -87.0042, -86.9948, -86.9854, -86.901, -86.9291, -87.0981, -87.2295, -87.314, -87.3422,
        -87.3422, -87.2765, -87.1544, -87.1356, -87.2858, -87.3609, -87.3703, -87.2858, -87.1732, -87.0418, -87.0512, -87.1826, -87.2952, -87.4173, -87.4924, -87.6801, -87.8021, -87.8209, -87.8866,
        -87.8585, -87.8491, -87.8679, -87.9054, -87.9805, -88.065, -88.0744, -88.1964, -88.356, -88.3935, -88.3842, -88.3748, -88.3654, -88.4217, -88.3935, -88.4217, -88.4217, -88.3091, -88.187,
        -88.065, -87.9242, -87.7928, -87.7177, -87.6238, -87.5017, -87.4266, -87.2952, -87.1075, -86.9385, -86.7789, -86.6006, -86.4785, -86.3659, -86.3002, -86.272, -86.2438, -86.1218, -86.0467,
        -85.9247, -85.9153, -85.8684, -85.859, -85.8026, -85.8214, -85.7463, -85.5773, -85.4178, -85.2676, -85.108, -85.0141, -84.9015, -84.8545, -84.7231, -84.6386, -84.5072, -84.4696, -84.4978,
        -84.5729, -84.6104, -84.4227, -84.2631, -84.0941, -83.9721, -83.8219, -83.6811, -83.5685, -83.5685, -83.6342, -83.7468, -83.775, -83.7844, -83.7844, -83.6811, -83.5685, -83.4652, -83.4558,
        -83.3807, -83.2587, -83.2117, -83.1179, -83.024, -82.9958, -82.9395, -82.8175, -82.7705, -82.7142, -82.686, -82.6954, -82.6954, -82.6109, -82.5265, -82.4607, -82.5077, -82.686, -82.8081,
        -82.8738, -82.8644, -82.9207, -82.9958, -83.1366, -83.2681, -83.3807, -83.4276, -83.6154, -83.7844, -83.9439, -84.0378, -84.188, -84.3382, -84.4602, -84.5447, -84.6104, -84.6386, -84.7325,
        -84.8921, -84.9015, -84.939, -84.9015, -84.8357, -84.7794, -84.7325, -84.7606, -84.77, -84.817, -84.8357, -84.9015, -84.9859, -85.0329, -85.1268, -85.2206, -85.2769, -85.3427, -85.3614,
        -85.4084, -85.4084, -85.3802, -85.3333, -85.2112, -85.1925, -85.2957, -85.352, -85.3614, -85.4647, -85.5398, -85.5492, -85.5398, -85.4929, -85.352, -85.1737, -85.108, -85.1268, -85.0517,
        -84.9578, -84.8451, -84.8639, -84.9484, -85.0517, -85.1737, -85.3333, -85.4835, -85.5961, -85.6994, -85.7839, -85.8684, -85.8871, -85.7463, -85.6994, -85.5492, -85.4459, -85.3333, -85.1737,
        -85.0423, -85.0047, -85.0892, -85.2582, -85.2394, -85.2018, -85.2018, -85.1268, -85.0892, -84.9766, -84.8357, -84.7606, -84.7606, -84.6574, -84.4509, -84.235, -84.1505, -84.1411, -84.2725,
        -84.3101, -84.3194, -84.4509, -84.6292, -84.7794, -84.9202, -84.9578, -84.9578, -84.9484, -84.8921, -84.7794, -84.6855, -84.648, -84.6668, -84.7606, -84.8451, -84.8921, -84.8639, -84.8639,
        -84.8545, -84.7888, -84.7606, -84.7137, -84.7137, -84.7794, -84.8357, -84.8827, -84.8921, -84.9766, -85.0047, -85.0329, -85.1268, -85.1737, -85.2863, -85.4271, -85.5492, -85.5867, -85.6524,
        -85.7839, -85.9341, -85.8965, -85.9998, -86.1218, -86.2438, -86.2814, -86.272, -86.3002, -86.2157, -86.103, -86.0373, -85.9904, -85.9341, -85.9341, -85.9904, -86.103, -86.1124, -86.1687,
        -86.1781, -86.2438, -86.3283, -86.4128, -86.4691, -86.5442, -86.5818, -86.6569, -86.7414, -86.8165, -86.9197, -87.0605, -87.2107, -87.2952, -87.3516, -87.436, -87.5393, -87.5581, -87.5111,
        -87.3703, -87.1826, -87.1263, -87.2107, -87.2858, -87.3985, -87.5017, -87.5581, -87.5675, -87.5581, -87.5487, -87.5768, -87.5956, -87.5956, -87.6707, -87.6989, -87.6707, -87.605, -87.5393,
        -87.4079, -87.2858, -87.0605, -86.9948, -87.0887, -87.192, -87.1638, -87.1263, -87.0605, -87.023, -86.9103, -86.8165, -86.6757, -86.5255, -86.3283, -86.1218, -85.8871, -85.5867, -85.5116,
        -85.4459, -85.4178, -85.3051, -85.2394, -85.1361, -85.0329, -85.0047, -84.7606, -84.5166, -84.3382, -84.1411, -83.9721, -83.7656, -83.5966, -83.4746, -83.3807, -83.193, -82.9864, -82.8738,
        -82.7987, -82.7518, -82.5734, -82.4044, -82.3669, -82.4889, -82.6109, -82.5077, -82.3575, -82.3012, -82.3575, -82.5265, -82.686, -82.7987, -82.8738, -82.9395, -83.0521, -83.1554, -83.2681,
        -83.3432, -83.315, -83.2023, -82.9958, -82.8456, -82.6297, -82.4232, -82.3105, -82.2448, -82.1791, -82.0853, -82.0008, -81.9444, -81.9351, -81.8881, -81.8224, -81.7661, -81.813, -81.8036,
        -81.7661, -81.7661, -81.7098, -81.6159, -81.6159, -81.7849, -81.8881, -81.982, -82.0477, -82.151, -82.1885, -82.2542, -82.3669, -82.4795, -82.564, -82.5828, -82.6485, -82.8081, -82.9583,
        -83.1085, -83.2399, -83.3432, -83.437, -83.6154, -83.7844, -83.9533, -84.0941, -84.1974, -84.2913, -84.4509, -84.5635, -84.6292, -84.7794, -84.77, -84.8357, -84.8921, -84.9015, -84.8357,
        -84.7606, -84.648, -84.5353, -84.2913, -84.188, -84.0754, -83.8782, -83.7093, -83.5309, -83.4089, -83.3995, -83.5121, -83.6529, -83.7562, -83.775, -83.8125, -83.9439, -84.0378, -84.1505,
        -84.066, -84.0378, -84.0284, -84.1411, -84.2913, -84.4039, -84.4227, -84.4602, -84.5353, -84.5072, -84.5072, -84.5729, -84.5635, -84.5917, -84.7325, -84.8827, -85.0235, -85.108, -85.1361,
        -85.1455, -85.1455, -85.1737, -85.2394, -85.3614, -85.5116, -85.6149, -85.7088, -85.7463, -85.9528, -86.1875, -86.394, -86.5912, -86.6099, -86.6099, -86.6099, -86.6006, -86.4973, -86.3283,
        -86.1687, -85.9247, -85.6806, -85.521, -85.4647, -85.4741, -85.6618, -85.8308, -86.0467, -86.2157, -86.3565],
   y = [76.9997, 77.009, 77.0114, 77.0137, 77.0208, 77.0372, 77.0677, 77.0771, 77.0771, 77.0583, 77.0677, 77.0583, 77.0419, 77.0536, 77.0724, 77.0771, 77.0771, 77.0513, 77.056, 77.0607, 77.0677, 77.0888,
        77.1076, 77.1217, 77.124, 77.124, 77.1381, 77.1405, 77.117, 77.1029, 77.0888, 77.1029, 77.117, 77.1381, 77.1452, 77.1475, 77.1616, 77.1475, 77.1616, 77.1944, 77.2038, 77.2156, 77.232, 77.2367,
        77.2367, 77.2531, 77.2766, 77.2977, 77.3094, 77.3329, 77.3353, 77.3306, 77.3071, 77.293, 77.2977, 77.3188, 77.3517, 77.3845, 77.4268, 77.4573, 77.469, 77.4831, 77.5113, 77.5254, 77.5465,
        77.5699, 77.5864, 77.6028, 77.638, 77.6709, 77.6967, 77.7154, 77.7272, 77.7436, 77.7694, 77.7694, 77.807, 77.7999, 77.8164, 77.8281, 77.8469, 77.8609, 77.8656, 77.8633, 77.868, 77.8469,
        77.8516, 77.8351, 77.8257, 77.7929, 77.7671, 77.7436, 77.7319, 77.6943, 77.6615, 77.6075, 77.5629, 77.5324, 77.4925, 77.4784, 77.4549, 77.4197, 77.4291, 77.401, 77.3798, 77.3822, 77.3681,
        77.3681, 77.3564, 77.3329, 77.3165, 77.2907, 77.293, 77.3188, 77.3423, 77.3752, 77.3869, 77.3869, 77.3892, 77.4057, 77.3798, 77.3752, 77.3564, 77.3658, 77.3916, 77.3963, 77.4033, 77.4197,
        77.4409, 77.4503, 77.462, 77.4901, 77.5136, 77.5324, 77.5559, 77.5864, 77.6075, 77.6474, 77.6873, 77.7225, 77.7483, 77.7929, 77.8211, 77.8703, 77.9008, 77.9267, 77.9595, 77.9783, 77.9971,
        78.0158, 78.0229, 78.0205, 77.9924, 77.9548, 77.9102, 77.8539, 77.8046, 77.7577, 77.692, 77.6239, 77.584, 77.5277, 77.5042, 77.4901, 77.5019, 77.4995, 77.5183, 77.5207, 77.5136, 77.5042,
        77.5066, 77.5066, 77.5136, 77.5465, 77.5746, 77.584, 77.6239, 77.645, 77.6779, 77.7154, 77.7342, 77.7154, 77.6662, 77.6357, 77.6028, 77.5934, 77.6122, 77.6286, 77.6357, 77.638, 77.6685,
        77.6967, 77.7131, 77.7389, 77.76, 77.7905, 77.807, 77.7929, 77.7905, 77.7905, 77.7905, 77.7858, 77.807, 77.8281, 77.8398, 77.8469, 77.8516, 77.8445, 77.8304, 77.8492, 77.8703, 77.8727, 77.8891,
        77.9032, 77.8797, 77.875, 77.868, 77.875, 77.8469, 77.8422, 77.8774, 77.8821, 77.9008, 77.936, 77.9595, 77.9853, 77.9971, 77.9853, 77.9783, 77.9947, 78.0018, 78.0158, 78.0205, 78.0323, 78.0581,
        78.0769, 78.0886, 78.1097, 78.1097, 78.1097, 78.1214, 78.1261, 78.1285, 78.1144, 78.0933, 78.1144, 78.1449, 78.1261, 78.1449, 78.152, 78.1637, 78.1731, 78.1684, 78.152, 78.1707, 78.2012,
        78.2364, 78.267, 78.2834, 78.3068, 78.335, 78.335, 78.3139, 78.3209, 78.3467, 78.3843, 78.4242, 78.4547, 78.4899, 78.5251, 78.5556, 78.5603, 78.5321, 78.5016, 78.4782, 78.4383, 78.389, 78.3327,
        78.2904, 78.2482, 78.1989, 78.152, 78.0956, 78.0698, 78.0839, 78.0886, 78.0769, 78.0581, 78.0698, 78.0604, 78.037, 78.0581, 78.0839, 78.1097, 78.1308, 78.1566, 78.2012, 78.2294, 78.2693,
        78.3115, 78.3374, 78.3303, 78.3068, 78.2951, 78.2834, 78.2458, 78.1989, 78.1801, 78.1801, 78.1872, 78.1426, 78.1144, 78.105, 78.105, 78.0839, 78.0886, 78.0698, 78.0769, 78.0792, 78.0886,
        78.0933, 78.0933, 78.1027, 78.1261, 78.1566, 78.159, 78.1684, 78.1707, 78.1825, 78.1942, 78.1801, 78.1731, 78.2012, 78.2271, 78.2576, 78.2763, 78.3092, 78.3303, 78.3702, 78.3796, 78.3984,
        78.4312, 78.4594, 78.4969, 78.5087, 78.5228, 78.5415, 78.5486, 78.5556, 78.5908, 78.6354, 78.6776, 78.7152, 78.7527, 78.7692, 78.7715, 78.7762, 78.7786, 78.7809, 78.795, 78.8114, 78.8302,
        78.8584, 78.8865, 78.8889, 78.8912, 78.8818, 78.8771, 78.8724, 78.8466, 78.8396, 78.8232, 78.8208, 78.8208, 78.7739, 78.7504, 78.7175, 78.7035, 78.6542, 78.6471, 78.6143, 78.5885, 78.5533,
        78.5298, 78.5157, 78.5368, 78.5814, 78.6471, 78.6471, 78.6424, 78.6706, 78.6753, 78.6776, 78.6823, 78.6753, 78.6988, 78.7011, 78.7035, 78.7269, 78.7387, 78.7715, 78.7739, 78.7997, 78.7856,
        78.7856, 78.7809, 78.7762, 78.795, 78.7997, 78.7762, 78.7598, 78.7645, 78.7903, 78.7903, 78.7856, 78.8044, 78.8208, 78.8325, 78.8607, 78.8889, 78.8912, 78.9123, 78.9452, 78.9593, 78.9499,
        78.9288, 78.9006, 78.8818, 78.8795, 78.8537, 78.8466, 78.8419, 78.8396, 78.8419, 78.8419, 78.8419, 78.8419, 78.863, 78.8795, 78.8889, 78.8889, 78.8889, 78.8912, 78.8936, 78.9006, 78.9053,
        78.9053, 78.91, 78.9311, 78.9522, 78.9757, 78.9945, 78.9945, 78.9945, 79.0273, 79.0508, 79.0719, 79.1001, 79.1095, 79.0766, 79.0696, 79.0344, 78.9992, 78.9827, 78.9757, 78.978, 79.0039,
        79.0156, 79.0039, 79.0062, 79.0391, 79.0602, 79.0883, 79.0813, 79.1118, 79.1423, 79.1752, 79.1986, 79.2104, 79.1775, 79.1611, 79.1634, 79.194, 79.2174, 79.2503, 79.2597, 79.2949, 79.3324,
        79.3723, 79.4075, 79.4427, 79.4826, 79.5084, 79.5413, 79.5694, 79.5765, 79.5812, 79.6117, 79.6445, 79.661, 79.6797, 79.7032, 79.7173, 79.7173, 79.7243, 79.7525, 79.7736, 79.8182, 79.8417,
        79.8745, 79.9262, 79.9637, 79.9731, 79.9708, 79.9402, 79.9309, 79.905, 79.891, 79.9074, 79.9238, 79.9543, 79.9614, 79.9801, 79.9942, 79.9989] };
  WCL[1061] = struct {
   x = [-83.2094, -82.9935, -82.8621, -82.6837, -82.4866, -82.3833, -82.2988, -82.2988, -82.1956, -82.1674, -82.1111, -81.989, -81.9796, -81.9703, -81.9703, -81.9703, -81.8858, -81.7919, -81.5572, -81.3695,
        -81.2193, -81.0503, -80.9377, -80.9095, -80.947, -80.947, -80.9001, -80.8062, -80.7405, -80.6091, -80.4589, -80.2993, -80.1116, -80.0646, -80.0928, -80.2242, -80.2899, -80.3462, -80.4777,
        -80.6185, -80.7875, -80.9283, -81.0597, -81.1817, -81.2568, -81.3789, -81.5103, -81.6323, -81.7074, -81.7074, -81.8013, -81.8107, -81.8013, -81.8201, -81.8482, -81.8858, -81.8952, -81.7825,
        -81.6792, -81.6605, -81.698, -81.7637, -81.8201, -81.9233, -81.989, -82.0454, -82.1392, -82.3082, -82.4959],
   y = [80.006, 79.9496, 79.9191, 79.8745, 79.8299, 79.8088, 79.783, 79.7337, 79.7173, 79.6727, 79.6445, 79.6445, 79.6164, 79.5812, 79.5624, 79.5507, 79.5554, 79.5788, 79.5788, 79.5577, 79.5319, 79.5084,
        79.4967, 79.5014, 79.5155, 79.5366, 79.5601, 79.5694, 79.5718, 79.5765, 79.5929, 79.6093, 79.6093, 79.6352, 79.6539, 79.6539, 79.6539, 79.6492, 79.6352, 79.6164, 79.614, 79.6023, 79.6164,
        79.6281, 79.6492, 79.6539, 79.6657, 79.6704, 79.7009, 79.7173, 79.7173, 79.7431, 79.7689, 79.8065, 79.8487, 79.8628, 79.8698, 79.8816, 79.8933, 79.9074, 79.9191, 79.9191, 79.9309, 79.9262,
        79.9379, 79.9402, 79.9543, 79.9848, 79.9966] };
  WCL[1062] = struct {
   x = [-79.8464, -80.0623, -80.3439, -80.4378, -80.2219, -80.2406, -80.4096, -80.588, -80.7945, -81.048, -81.4328, -81.6863, -81.9116, -81.9773, -82.0148, -82.1275, -81.8553, -81.4892, -81.3108, -81.3108,
        -81.63, -81.6957, -81.6863, -81.3859, -81.1606, -80.8978, -80.6067, -80.25],
   y = [77.1616, 77.117, 77.0372, 77.0396, 77.117, 77.1663, 77.1639, 77.171, 77.1968, 77.1991, 77.178, 77.1123, 77.0443, 77.1029, 77.178, 77.2179, 77.2062, 77.2297, 77.2555, 77.2813, 77.3282, 77.3939,
        77.4338, 77.3517, 77.3047, 77.2602, 77.2297, 77.225] };
  WCL[1063] = struct {
   x = [-90.1255, -89.7735, -89.6374, -89.4168, -89.4308, -89.4966, -89.628, -89.5951, -89.628, -89.4966, -89.3511, -89.3839, -89.2431, -89.1305, -89.0413, -88.9568, -88.9427, -88.8582, -88.7784, -88.6892,
        -88.6892, -88.7456, -88.755, -88.769, -88.755, -88.6141, -88.5015, -88.3466, -88.1636, -88.0181, -87.9523, -87.7646, -87.5862, -87.5768, -87.4454, -87.4642, -87.2811, -87.0934, -86.9057,
        -86.7836, -86.8165, -86.915, -86.8493, -86.5818, -86.441, -86.3847, -86.2532, -86.1781, -86.1218, -85.8214, -85.69, -85.69, -85.582, -85.4506, -85.338, -85.2816, -85.1174, -85.1596, -85.1737,
        -85.1502, -85.1268, -85.0282, -84.8217, -84.5776, -84.1927, -83.8923, -83.6952, -83.4511, -83.376, -83.5731, -83.6717, -83.9721, -83.8923, -83.7374, -83.5403, -83.4089, -83.207, -83.1742,
        -83.1883, -83.0099, -82.8691, -82.6485, -82.5922, -82.4795, -82.4044, -82.2167, -81.9632, -81.752, -81.5314, -81.3437, -81.1935, -81.17, -81.0057, -80.818, -80.8039, -80.7288, -80.5316,
        -80.2876, -80.2876, -80.2876, -80.2876, -80.3627, -80.3204, -80.419, -80.6725, -80.6631, -80.649, -80.4753, -80.2876, -80.189],
   y = [74.5026, 74.5026, 74.5496, 74.6153, 74.6364, 74.6599, 74.6904, 74.7514, 74.7467, 74.7584, 74.7397, 74.8101, 74.6951, 74.6481, 74.6833, 74.742, 74.8077, 74.8007, 74.7678, 74.7561, 74.7045, 74.5942,
        74.5214, 74.4792, 74.4792, 74.4651, 74.4533, 74.4792, 74.4463, 74.4862, 74.4369, 74.458, 74.4627, 74.5284, 74.5284, 74.4627, 74.4463, 74.4533, 74.444, 74.4674, 74.5097, 74.5683, 74.5097,
        74.4627, 74.505, 74.5707, 74.5097, 74.444, 74.451, 74.4627, 74.5261, 74.6528, 74.5848, 74.4979, 74.4792, 74.5214, 74.5683, 74.6364, 74.6411, 74.688, 74.6176, 74.505, 74.5097, 74.505, 74.5261,
        74.5378, 74.5566, 74.688, 74.735, 74.7889, 74.8664, 74.8945, 74.8922, 74.8969, 74.8711, 74.8007, 74.803, 74.7303, 74.6035, 74.5355, 74.4909, 74.505, 74.5519, 74.5026, 74.4885, 74.4651, 74.451,
        74.5097, 74.5331, 74.5519, 74.559, 74.5566, 74.5519, 74.559, 74.5613, 74.5519, 74.5848, 74.5871, 74.6951, 74.7256, 74.5566, 74.7561, 74.7889, 74.8406, 74.9133, 74.9837, 75.0424, 75.0635,
        75.1081, 75.1738] };
  WCL[1064] = struct {
   x = [-79.9449, -80.1233, -80.3955, -80.5645, -80.3533, -80.3768, -80.419, -80.5504, -80.6161, -80.7288, -80.9494, -81.17, -81.2404, -81.2263, -81.2592, -81.3061, -81.3108, -81.4563, -81.5689, -81.6487,
        -81.7379, -81.813, -81.8881, -82.0055, -82.1181, -82.1932, -82.2542, -82.334, -82.3669, -82.4232, -82.442, -82.503, -82.5265, -82.5969, -82.6579, -82.7283, -82.8409, -82.9301, -83.0287,
        -83.0146, -82.9583, -83.0803, -83.3244, -83.5027, -83.6107, -83.6529, -83.6436, -83.6201, -83.6342, -83.6529, -83.8078, -83.8735, -83.9299, -84.127, -84.3054, -84.418, -84.5682, -84.5917,
        -84.5823, -84.3852, -84.0425, -83.9956, -84.2068, -84.3852, -84.6151, -84.7794, -84.9015, -84.9249, -84.9578, -85.1314, -85.2206, -85.0329, -84.9108, -85.0235, -85.2441, -85.4412, -85.5069,
        -85.6196, -85.751, -85.9153, -86.0608, -86.0373, -86.0936, -86.1265, -86.1781, -86.2814, -86.4551, -86.502, -86.6006, -86.6897, -86.7648, -86.6897, -86.7226, -86.8775, -86.9197, -86.9291,
        -87.0089, -87.1403, -87.1967, -87.2952, -87.3375, -87.2952, -87.2483, -87.3703, -87.4595, -87.5487, -87.6707, -87.8256, -87.8913, -87.9007, -87.957, -88.0462, -88.0556, -88.0134, -88.1636,
        -88.1448, -87.957, -87.7787, -87.7364, -87.6473, -87.6942, -87.7317, -87.7834, -87.8585, -87.9101],
   y = [75.5352, 75.5305, 75.5141, 75.5305, 75.5611, 75.6033, 75.6244, 75.6549, 75.6573, 75.6784, 75.6596, 75.6784, 75.6878, 75.7676, 75.8028, 75.791, 75.7957, 75.8192, 75.8333, 75.838, 75.8497, 75.8544,
        75.8356, 75.8239, 75.7817, 75.7488, 75.7652, 75.7629, 75.7981, 75.7934, 75.7863, 75.8122, 75.8192, 75.8145, 75.7676, 75.7582, 75.7136, 75.6995, 75.6737, 75.6408, 75.6244, 75.6315, 75.6526,
        75.6361, 75.5798, 75.5963, 75.5822, 75.5282, 75.5047, 75.5282, 75.5352, 75.5517, 75.5916, 75.5352, 75.5376, 75.5141, 75.4977, 75.4977, 75.4625, 75.4249, 75.3991, 75.3592, 75.385, 75.3803,
        75.3874, 75.3686, 75.3451, 75.3404, 75.3311, 75.3193, 75.3428, 75.3663, 75.3874, 75.4249, 75.4367, 75.4625, 75.5, 75.5399, 75.5634, 75.5775, 75.5235, 75.4625, 75.4249, 75.4531, 75.5235,
        75.5517, 75.4977, 75.4625, 75.4296, 75.4273, 75.432, 75.4883, 75.5071, 75.554, 75.5822, 75.5822, 75.5963, 75.6315, 75.6291, 75.5963, 75.5446, 75.5, 75.432, 75.371, 75.4132, 75.5165, 75.5399,
        75.5517, 75.5258, 75.5258, 75.5071, 75.5118, 75.5258, 75.5517, 75.5352, 75.5728, 75.5751, 75.5892, 75.6315, 75.7324, 75.7441, 75.7981, 75.784, 75.7676, 75.7652] };
  WCL[1065] = struct {
   x = [-90.0598, -89.9941, -89.9378, -89.9565],
   y = [75.784, 75.8028, 75.8333, 75.8497] };
  WCL[1066] = struct {
   x = [-90.2687, -89.9401, -89.6867, -89.5177, -89.4238, -89.3299, -89.3863, -89.4989, -89.696, -89.7148, -89.7993, -89.9589],
   y = [76.1079, 76.0891, 76.0938, 76.1219, 76.1548, 76.2252, 76.2839, 76.2792, 76.2674, 76.2721, 76.2815, 76.2839] };
  WCL[1067] = struct {
   x = [-79.7924, -80.0365, -80.2101, -80.2242, -80.3791, -80.4871, -80.4964, -80.4213, -80.3462, -80.1914, -80.0459],
   y = [75.0283, 75.0142, 74.9884, 74.9626, 74.9696, 74.9391, 74.9321, 74.8899, 74.8523, 74.8077, 74.8171] };
  WCL[1068] = struct {
   x = [-80.1655, -80.2735],
   y = [69.9685, 70.0014] };
  WCL[1069] = struct {
   x = [-80.5199, -80.764, -81.0597, -81.4352, -81.7215, -81.7778],
   y = [69.9685, 70.0084, 70.0389, 70.0882, 70.1117, 70.0272] };
  WCL[1070] = struct {
   x = [-84.8545, -85.3614, -85.9106, -85.6243],
   y = [69.9803, 70.0483, 70.0366, 70.0272] };
  WCL[1071] = struct {
   x = [-85.582, -86.0561, -86.3847, -86.5959, -86.6522, -86.7836, -87.1356, -87.6191, -87.8725, -87.9195, -88.2621, -88.0415, -88.3466, -88.6705, -89.0788, -89.3088, -89.4966, -89.3651, -89.5294, -89.229,
        -88.8582, -88.5813, -88.248, -87.7881, -87.4548, -87.2342, -87.5534, -87.8866, -88.0931, -88.6376, -89.0413, -89.5388, -89.9565],
   y = [69.9709, 70.0178, 70.1258, 70.2572, 70.4379, 70.3088, 70.2736, 70.3112, 70.3018, 70.2666, 70.3112, 70.3276, 70.4051, 70.4637, 70.6233, 70.7876, 70.8815, 70.9894, 71.0504, 71.0223, 71.0152, 70.9542,
        70.9002, 70.9073, 70.9566, 70.9777, 71.0317, 71.1326, 71.2171, 71.2382, 71.2429, 71.2851, 71.3086] };
  WCL[1072] = struct {
   x = [-90.1255, -89.9471],
   y = [71.6512, 71.7615] };
  WCL[1073] = struct {
   x = [-90.0692, -89.8814, -89.6937, -89.9706],
   y = [72.0549, 72.13, 72.1675, 72.1699] };
  WCL[1074] = struct {
   x = [-90.0481, -89.9166, -89.9073, -89.7383, -89.696, -89.6632, -89.3205, -89.5083, -89.4848, -89.2548, -89.3863, -89.2877, -89.1797, -89.0671, -88.9451, -88.8371, -88.7151, -88.4381, -88.1096, -87.8327,
        -87.5229, -87.1708, -86.7062, -86.4199, -86.1664, -85.9599, -85.7252, -85.5046, -85.2746, -85.1103],
   y = [72.367, 72.4374, 72.5806, 72.6064, 72.6697, 72.7472, 72.7472, 72.7801, 72.9232, 72.9678, 73.0476, 73.1203, 73.1814, 73.2072, 73.2541, 73.2987, 73.4254, 73.5123, 73.6038, 73.6648, 73.7164, 73.761,
        73.8103, 73.815, 73.8103, 73.7915, 73.7845, 73.7915, 73.7822, 73.7352] };
  WCL[1075] = struct {
   x = [-85.108, -84.9953, -84.9953, -85.1737, -85.4506, -85.69, -85.8871, -86.1218, -86.3283, -86.4081, -86.6053, -86.7695, -86.8024, -86.5724, -86.4175, -86.5818, -86.6053, -86.441, -86.1781, -86.0092,
        -85.7369, -85.5257, -85.2065, -85.1174, -85.061, -85.061, -85.3145, -85.5022, -85.5914, -85.7698, -86.042, -86.2626, -86.6053, -86.7179, -86.4081, -86.2438, -86.042, -85.7792, -85.5492,
        -85.2957, -85.2488, -85.1268, -85.061, -85.1174, -85.1596, -85.061, -84.9202, -84.9061, -84.8733, -84.9061, -84.8217, -84.8217, -84.7513, -84.8076, -85.0751, -85.46, -85.7041, -85.7135,
        -85.9998, -86.0655, -85.8777, -85.6665, -85.6337, -85.5492, -85.3943, -85.2159, -85.0282, -84.9296, -84.9953, -85.108, -85.2816, -85.6243, -85.7463, -85.6665, -85.8214, -85.8355, -85.7463,
        -85.5257, -85.0845, -84.8639, -84.6433, -84.357, -84.1364, -84.4978, -84.831, -85.1408, -85.4271, -85.6149, -85.4929, -85.3708, -85.338, -84.8874, -84.5447, -84.2772, -83.958, -83.9252,
        -84.5213, -84.9953, -85.23, -85.1174, -84.8874, -84.6668, -84.4649, -84.6339, -84.587, -84.2584, -83.9815, -83.8501, -83.7844, -83.7186, -84.005, -83.6482, -83.2868, -83.0568, -82.8691,
        -82.5687, -82.2495, -81.9069, -81.6112, -81.522, -81.3671, -81.3343, -81.3014, -81.2686, -81.3906, -81.522, -81.1371, -80.8508, -80.7053, -80.7288, -80.9024, -80.8837, -80.7288, -80.5739,
        -80.4425, -80.5645, -80.649, -80.7288, -80.8508, -80.9353, -80.9729, -80.8602, -80.7053, -80.7523, -81.0057, -81.0902, -80.9729, -81.0057, -80.7053, -80.4753, -80.3955, -80.2313, -80.1421,
        -80.1421],
   y = [73.7352, 73.7141, 73.6953, 73.6249, 73.5428, 73.4794, 73.3715, 73.2917, 73.1767, 73.0124, 72.8857, 72.7824, 72.6134, 72.4867, 72.3623, 72.191, 72.0314, 71.8789, 71.7521, 71.6817, 71.5691, 71.4682,
        71.4424, 71.3907, 71.3133, 71.2663, 71.2851, 71.2523, 71.2171, 71.2006, 71.1865, 71.1044, 71.0246, 71.0011, 70.9847, 71.0364, 71.0856, 71.1279, 71.149, 71.142, 71.1396, 71.1607, 71.1279,
        71.0716, 70.9636, 70.9261, 71.0504, 71.1631, 71.2452, 71.3414, 71.386, 71.4189, 71.548, 71.6278, 71.6278, 71.67, 71.7662, 71.8624, 71.9235, 71.9774, 72.0173, 72.0854, 72.1511, 72.2285, 72.2285,
        72.2098, 72.2684, 72.3717, 72.3975, 72.367, 72.3623, 72.428, 72.5172, 72.5688, 72.7143, 72.8763, 72.9279, 72.9326, 72.8997, 72.8575, 72.8129, 72.7378, 72.7495, 72.8481, 72.8927, 72.9842,
        73.003, 73.0499, 73.0711, 73.0546, 73.1016, 73.0664, 73.0241, 73.003, 72.9866, 73.0241, 73.0828, 73.1438, 73.2283, 73.3175, 73.3409, 73.24, 73.2072, 73.3409, 73.402, 73.4466, 73.409, 73.3433,
        73.287, 73.3879, 73.47, 73.5662, 73.632, 73.6484, 73.6953, 73.6859, 73.7, 73.7047, 73.7071, 73.646, 73.578, 73.4747, 73.3808, 73.3128, 73.2635, 73.2259, 73.1884, 73.1203, 73.0546, 72.9889,
        72.9514, 72.8997, 72.7965, 72.7566, 72.712, 72.6064, 72.4844, 72.367, 72.2849, 72.184, 72.1534, 72.1558, 72.0901, 72.0619, 72.0854, 72.0244, 71.9634, 71.9023, 71.9751, 72.1042, 72.2004,
        72.1863, 72.3177, 72.3576] };
  WCL[1076] = struct {
   x = [-88.5531, -88.7878, -88.7878, -88.8629, -88.9849, -89.229, -89.5013, -89.6139, -89.5764, -89.5857, -89.6327, -89.689, -89.6233, -89.4825, -89.3417, -89.2102, -88.9756, -88.9286, -88.8817, -88.8347,
        -88.7033, -88.7878, -88.8535, -88.7221, -88.6658, -88.6752, -88.6752, -88.6282, -88.6001, -88.3935, -88.2527, -88.1589, -88.0181, -87.8679, -87.7834, -87.8115, -87.7552, -87.6989, -87.4266,
        -87.3234, -87.0887, -87.0042, -86.7977, -86.7883, -86.6944, -86.6757, -86.7695, -86.7695, -86.4598, -86.2157, -86.7977, -86.7883, -86.6944, -86.6757, -86.7695, -86.7695, -86.4598, -86.2157,
        -86.0185, -85.8308, -85.5867, -85.5304, -85.2863, -84.9202, -84.6855, -84.8264, -85.2018, -85.2394, -85.2769, -85.0235, -84.77, -84.4602, -84.4602, -84.3288, -84.0378, -83.6529, -83.4089,
        -83.3901, -83.5966, -83.6248, -83.4934, -83.3432, -83.2117, -83.1554, -82.8926, -82.7424, -82.395, -82.273, -82.4232, -82.6673, -82.4983, -82.1791, -82.151, -82.1416, -82.0383, -81.6628,
        -81.5314, -81.3718, -81.1841, -81.0339, -81.0151, -81.1184, -81.231, -81.1277, -80.9212, -80.3204],
   y = [76.9997, 76.9762, 76.9715, 76.9386, 76.9246, 76.88, 76.8283, 76.7415, 76.6946, 76.6289, 76.5702, 76.535, 76.5092, 76.4458, 76.3942, 76.3848, 76.4599, 76.535, 76.4505, 76.4083, 76.5162, 76.6289,
        76.7204, 76.7697, 76.6922, 76.6054, 76.5491, 76.4411, 76.3824, 76.3731, 76.3707, 76.3214, 76.3332, 76.3472, 76.427, 76.5186, 76.4834, 76.4364, 76.3895, 76.3449, 76.3285, 76.4106, 76.4599,
        76.5678, 76.5186, 76.4482, 76.3637, 76.3379, 76.3214, 76.3027, 76.4599, 76.5678, 76.5186, 76.4482, 76.3637, 76.3379, 76.3214, 76.3027, 76.2674, 76.2721, 76.2557, 76.2135, 76.2322, 76.2393,
        76.2463, 76.2933, 76.4153, 76.4341, 76.5256, 76.3965, 76.3942, 76.4669, 76.603, 76.5561, 76.4364, 76.3824, 76.4176, 76.5279, 76.6758, 76.7227, 76.6688, 76.5608, 76.4482, 76.3918, 76.3425,
        76.3637, 76.3731, 76.481, 76.5585, 76.65, 76.6148, 76.6007, 76.5725, 76.4998, 76.4482, 76.4294, 76.4505, 76.4622, 76.4505, 76.3942, 76.3308, 76.2276, 76.1806, 76.1219, 76.1454, 76.2205] };
  WCL[1077] = struct {
   x = [-79.7924, -80.0928, -80.2101, -80.243, -80.2571, -80.4777, -80.7734, -80.9846, -80.9752, -80.8719, -80.8626, -80.717, -80.4777, -80.2336, -80.0928],
   y = [72.7542, 72.8293, 72.9443, 73.0711, 73.1767, 73.1884, 73.2236, 73.3222, 73.4325, 73.47, 73.6038, 73.693, 73.7024, 73.6531, 73.6648] };
  WCL[1078] = struct {
   x = [-99.9447, -99.9869],
   y = [71.8718, 71.8812] };
  WCL[1079] = struct {
   x = [-100.001, -99.9353, -99.8883],
   y = [73.1203, 73.1063, 73.1297] };
  WCL[1080] = struct {
   x = [-100.09, -99.9869, -99.9212, -99.8132, -99.6584, -99.5598, -99.4378, -99.3157, -99.2406, -99.02, -98.9308, -98.7102, -98.4896, -98.3019, -98.1798, -98.1047, -98.0578, -97.917, -97.8044, -97.7293,
        -97.5744, -97.4429, -97.4101, -97.274, -97.1754, -97.0674, -97.0768, -97.1003, -97.2083, -97.274, -97.3209, -97.518, -97.5838, -97.518, -97.565, -97.6495, -97.6307, -97.518, -97.4523, -97.3068,
        -97.274, -97.274, -97.288, -97.3538, -97.5509, -97.7152, -97.795, -97.8701, -97.9686, -98.0484, -98.1892, -98.2784, -98.4098, -98.4239, -98.4333, -98.4333, -98.4239, -98.2596, -98.025,
        -97.7621, -97.5978, -97.518, -97.4289, -97.3444, -97.3444, -97.3444, -97.3538, -97.2974, -97.2646, -97.2552, -97.166, -97.1003, -97.0768, -97.1566, -97.1566, -97.1238, -96.8985, -96.7154,
        -96.7013, -96.6779, -96.6122, -96.5699, -96.5793, -96.5277, -96.4714, -96.3916, -96.3916, -96.4385, -96.5605, -96.6262, -96.7013, -96.6685, -96.5934, -96.5464, -96.5277, -96.5605, -96.6779,
        -96.692, -96.5464, -96.5042, -96.4807, -96.4948, -96.5699, -96.6356, -96.8328, -96.9126, -97.1003, -97.3303, -97.5744, -97.7621, -97.9029, -98.0907, -98.1376, -97.9921, -97.917, -97.917,
        -98.0719, -98.2221, -98.2596, -98.4098, -98.5319, -98.5647, -98.6117, -98.7008, -98.8088, -98.8839, -99.0388, -99.0857, -99.1374, -99.1843, -99.2078, -99.3157, -99.3955, -99.5598, -99.7006,
        -99.8132, -99.879, -99.9447],
   y = [73.8854, 73.8619, 73.8244, 73.7822, 73.7751, 73.761, 73.7235, 73.7071, 73.7, 73.7117, 73.6953, 73.7211, 73.7329, 73.7634, 73.8033, 73.815, 73.8197, 73.8432, 73.8526, 73.8502, 73.8619, 73.8807,
        73.8432, 73.8033, 73.754, 73.7423, 73.6836, 73.6249, 73.585, 73.5475, 73.5475, 73.5381, 73.5052, 73.4958, 73.4888, 73.4841, 73.463, 73.4583, 73.4442, 73.4466, 73.3996, 73.355, 73.3128, 73.3316,
        73.3081, 73.2705, 73.2424, 73.2142, 73.1649, 73.118, 73.0828, 73.0312, 72.9889, 72.9584, 72.9256, 72.8927, 72.8669, 72.9185, 72.9936, 73.0147, 72.9866, 72.9514, 72.9279, 72.8997, 72.8763,
        72.8481, 72.834, 72.8176, 72.7894, 72.7636, 72.7566, 72.7237, 72.6838, 72.6369, 72.5829, 72.5853, 72.6486, 72.658, 72.6956, 72.7214, 72.6956, 72.6697, 72.6134, 72.5853, 72.5336, 72.4421,
        72.3905, 72.3576, 72.3154, 72.299, 72.2919, 72.2708, 72.2239, 72.1746, 72.0854, 72.0478, 72.0079, 71.9915, 71.9986, 71.961, 71.8859, 71.8132, 71.7967, 71.8202, 71.7733, 71.6723, 71.6231,
        71.5691, 71.6019, 71.6019, 71.5855, 71.6019, 71.5691, 71.5081, 71.4752, 71.4189, 71.379, 71.3438, 71.2992, 71.2523, 71.2593, 71.2499, 71.2992, 71.3626, 71.3555, 71.3086, 71.3086, 71.3555,
        71.4189, 71.5081, 71.5503, 71.5996, 71.6935, 71.7615, 71.792, 71.8202, 71.8577, 71.8718] };
  WCL[1081] = struct {
   x = [-96.6802, -96.6708, -96.6943, -96.6708, -96.5957, -96.638, -96.7178, -96.7929, -96.8914, -97.0698, -97.2341, -97.2012, -97.1449, -97.0886, -97.0229, -96.8257, -96.7272, -96.7037, -96.6802],
   y = [72.9748, 72.9256, 72.942, 72.9866, 73.0382, 73.0969, 73.118, 73.1438, 73.1602, 73.1203, 73.0711, 73.0124, 72.9701, 72.9349, 72.881, 72.8974, 72.9349, 72.949, 72.9748] };
  WCL[1082] = struct {
   x = [-98.6656, -98.5812, -98.5671, -98.5014, -98.3465, -98.2151, -98.0836, -97.9381, -97.863, -97.8724, -98.093, -98.3606, -98.4591, -98.6469, -98.8675, -99.0317, -99.074, -99.2289, -99.4401, -99.3838,
        -99.3274, -99.2852, -99.1209, -99.0646, -98.9895, -98.9332, -98.8675, -98.8111, -98.7548, -98.6656],
   y = [73.8009, 73.8009, 73.8103, 73.8244, 73.8338, 73.8713, 73.9136, 73.9722, 74.0286, 74.0732, 74.0779, 74.045, 74.0286, 73.9981, 73.9887, 73.9417, 73.9065, 73.8925, 73.8502, 73.8291, 73.8314, 73.8244,
        73.8103, 73.8314, 73.8314, 73.7751, 73.8009, 73.8127, 73.7939, 73.8009] };
  WCL[1083] = struct {
   x = [-96.3939, -96.4502, -96.6051, -96.6286, -96.638, -96.638, -96.6145, -96.5066, -96.4408, -96.3751, -96.3188, -96.2531, -96.1968, -96.1545, -96.0888, -96.0653, -96.1404, -96.2202, -96.3188, -96.4737,
        -96.4737, -96.4408, -96.3611, -96.3188, -96.3751, -96.3611, -96.3282, -96.3423, -96.3282, -96.2625, -96.1733, -96.1404, -96.1311, -95.9996, -95.8213, -95.7227, -95.657, -95.5913, -95.5678,
        -95.5256, -95.5678, -95.6805, -95.7884, -95.8213, -95.7227, -95.6664, -95.5021, -95.4035, -95.3378, -95.3144, -95.2064, -95.1172, -95.0374],
   y = [69.9521, 70.0084, 70.0695, 70.1563, 70.2455, 70.33, 70.3863, 70.4191, 70.4191, 70.4778, 70.5341, 70.574, 70.5693, 70.5365, 70.567, 70.6562, 70.6233, 70.6796, 70.7665, 70.8275, 70.8862, 70.9683,
        71.0223, 71.0504, 71.0528, 71.1185, 71.1631, 71.21, 71.2523, 71.2804, 71.2734, 71.2945, 71.3438, 71.379, 71.3156, 71.264, 71.2358, 71.2452, 71.2875, 71.3414, 71.4611, 71.4682, 71.5034, 71.5691,
        71.616, 71.6583, 71.7075, 71.7545, 71.8061, 71.8624, 71.9375, 71.9493, 71.954] };
  WCL[1084] = struct {
   x = [-96.4103, -96.4432, -96.5183, -96.5652, -96.6075, -96.5981, -96.6309, -96.7624, -96.753, -96.6966, -96.7717, -96.8422, -96.9501, -97.0252, -97.0158, -97.105, -97.1379, -97.0815, -96.9501, -96.8281,
        -96.7389, -96.6403, -96.5089, -96.4667, -96.4526, -96.4009, -96.3634, -96.2789, -96.1757, -96.1569, -96.124, -96.1897, -96.2977, -96.354, -96.4854, -96.5089, -96.5089],
   y = [75.6338, 75.6408, 75.6127, 75.5845, 75.554, 75.5329, 75.4977, 75.4813, 75.4648, 75.4507, 75.4249, 75.4249, 75.4132, 75.4343, 75.4766, 75.5, 75.5493, 75.5681, 75.5892, 75.6221, 75.6127, 75.6385,
        75.6479, 75.6854, 75.7206, 75.7089, 75.6807, 75.6573, 75.6502, 75.6127, 75.5798, 75.5493, 75.5141, 75.5399, 75.5634, 75.6009, 75.6338] };
  WCL[1085] = struct {
   x = [-97.6894, -97.6612, -97.6894, -97.6894, -97.6612, -97.6894, -97.7738, -97.9334, -97.9616, -97.8771, -97.7832, -97.6894, -97.6236, -97.5204, -97.4922, -97.5204, -97.5204, -97.5439, -97.5345, -97.4875,
        -97.4875, -97.5439, -97.6002, -97.6424, -97.7316, -97.8208, -97.8865, -98.0179, -98.0742, -98.0085, -97.9287, -97.9053, -97.9287, -97.9851, -98.0742, -98.1165, -98.1165, -98.1587, -98.1165,
        -98.0179, -98.1071, -98.1259, -98.0836, -98.1165, -98.2151, -98.262, -98.3042, -98.2714, -98.1822, -98.14, -98.0602, -98.0085, -97.8724, -97.8302, -97.9053, -98.0179, -98.1165, -98.1963,
        -98.2948, -98.2948, -98.262, -98.262, -98.1728, -98.1587, -98.3371, -98.5014, -98.6328, -98.6563, -98.7454, -98.844, -98.966, -99.0317, -99.0083, -99.1209, -99.1538, -99.2524, -99.3415,
        -99.3415, -99.3979, -99.5387, -99.6044, -99.5527, -99.5856, -99.7029, -99.7733],
   y = [76.0023, 75.9905, 75.9788, 75.9553, 75.9248, 75.8802, 75.8614, 75.8286, 75.7957, 75.8004, 75.7746, 75.7723, 75.7676, 75.777, 75.7206, 75.6549, 75.5869, 75.5212, 75.4906, 75.4507, 75.4367, 75.4719,
        75.5212, 75.5775, 75.6268, 75.6455, 75.6268, 75.6338, 75.6009, 75.5986, 75.5798, 75.5376, 75.5212, 75.5564, 75.5493, 75.5352, 75.5235, 75.5071, 75.4953, 75.4695, 75.4648, 75.439, 75.4085,
        75.4038, 75.4343, 75.4085, 75.3616, 75.3287, 75.3428, 75.3193, 75.2794, 75.2536, 75.1926, 75.1574, 75.1433, 75.1597, 75.2114, 75.2278, 75.2208, 75.1879, 75.1574, 75.1339, 75.0987, 75.0494,
        75.0424, 75.0307, 75.0307, 75.0706, 75.0518, 75.0236, 75.0236, 75.0447, 75.0893, 75.0753, 75.026, 75.033, 75.0377, 75.087, 75.0518, 75.0307, 75.0729, 75.1222, 75.1058, 75.0377, 74.9978] };
  WCL[1086] = struct {
   x = [-100.062, -99.8602, -99.8508],
   y = [75.5188, 75.5188, 75.5493] };
  WCL[1087] = struct {
   x = [-100.05, -99.9752, -99.7452, -99.7546, -99.8438, -99.825, -99.7311, -99.6513, -99.5574, -99.4776, -99.4119, -99.426, -99.4823, -99.5668, -99.6607, -99.7452, -99.7593, -99.7733, -99.8438],
   y = [75.5939, 75.6127, 75.6361, 75.6573, 75.6807, 75.7136, 75.7253, 75.7371, 75.7418, 75.7418, 75.7652, 75.7863, 75.7652, 75.7699, 75.7582, 75.7511, 75.7582, 75.7652, 75.7347] };
  WCL[1088] = struct {
   x = [-100.083, -99.9893, -99.886, -99.7358, -99.717, -99.8954, -99.9705],
   y = [75.9788, 75.9741, 75.9717, 76.0046, 76.0187, 76.0069, 76.0422] };
  WCL[1089] = struct {
   x = [-100.095, -99.9353],
   y = [76.2041, 76.2041] };
  WCL[1090] = struct {
   x = [-100.029, -99.9353, -99.8602, -99.7663, -99.7194, -99.6818, -99.5598, -99.4941, -99.4096, -99.4002, -99.3908, -99.2594, -99.1467, -99.1092, -99.1561, -99.1186, -99.0904, -99.0247, -98.8839, -98.8182,
        -98.9214, -98.959, -98.8933, -98.7431, -98.6117, -98.5366, -98.6117, -98.7243, -98.837, -98.7149, -98.5084, -98.3958, -98.2831, -98.1705, -98.2925, -98.4051, -98.3394, -98.2174, -98.1235,
        -97.9827, -97.8513, -97.7762, -97.7668, -97.8231, -97.8231, -97.8231, -97.7574, -97.6823, -97.6448, -97.7105, -97.7011, -97.7199, -97.7105, -97.6917],
   y = [76.6453, 76.6547, 76.6617, 76.6617, 76.6429, 76.6242, 76.6007, 76.5725, 76.5631, 76.5631, 76.5655, 76.542, 76.542, 76.5397, 76.4998, 76.4716, 76.4998, 76.5021, 76.481, 76.5233, 76.5796, 76.6007,
        76.6265, 76.6218, 76.6148, 76.6641, 76.6711, 76.6758, 76.6899, 76.6969, 76.6875, 76.6429, 76.6124, 76.5937, 76.6101, 76.603, 76.5796, 76.5631, 76.5655, 76.5514, 76.5139, 76.4857, 76.4317,
        76.3824, 76.3472, 76.3073, 76.2768, 76.2322, 76.1665, 76.1173, 76.0867, 76.0539, 76.0163, 76.0023] };
  WCL[1091] = struct {
   x = [-95.0656, -95.075, -95.1219, -95.2909, -95.4223, -95.5725, -95.7509, -95.9574, -96.1921, -96.2953, -96.4361, -96.5019, -96.5957, -96.7178, -96.6614, -96.6145, -96.6145, -96.6333, -96.7365, -96.8116,
        -96.9431, -97.1777, -97.2247, -97.1777, -97.0651, -97.1214, -97.2528, -97.403, -97.4781, -97.6377, -97.7504, -97.6941, -97.6659, -97.4688, -97.2059, -96.9431, -96.7553, -96.868, -96.821,
        -97.0276, -97.2059, -97.3843, -97.5439, -97.7504, -97.8724, -97.91, -97.9193, -97.9381, -97.8724, -97.7692, -97.7973, -97.9851, -98.0226, -98.0414, -98.0132, -98.1446, -98.2855, -98.3606,
        -98.3136, -98.2291, -98.1259, -98.0695, -98.0977, -98.2197, -98.2761, -98.2948, -98.2948, -98.2667, -98.2761, -98.154, -97.9569, -97.7785, -97.6096, -97.403, -97.1965, -97.0276, -96.9055,
        -96.7741, -96.6427, -96.4455, -96.2953, -96.1733, -96.0607, -95.9856, -96.0419, -96.0419, -96.07, -95.9668, -95.8166, -95.6852, -95.5256, -95.3566, -95.1876, -95.0844, -95.0844, -94.9999,
        -94.9529, -94.8779, -94.8121, -94.784, -94.7934, -94.8779, -94.9905, -94.9999, -94.9999, -94.9999, -94.9905, -94.9811, -94.8591, -94.784, -94.8028, -94.8028, -94.8685, -94.9248, -94.9436,
        -94.9999, -95.0656],
   y = [77.9337, 77.936, 77.9431, 77.9431, 77.922, 77.9149, 77.8938, 77.868, 77.8609, 77.8774, 77.9149, 77.8962, 77.875, 77.8492, 77.8563, 77.8422, 77.8234, 77.8164, 77.7929, 77.76, 77.7694, 77.7929,
        77.8187, 77.8445, 77.8985, 77.9173, 77.9431, 77.9877, 78.0111, 78.0299, 78.0628, 78.0933, 78.1121, 78.1168, 78.1003, 78.1027, 78.1402, 78.1449, 78.1872, 78.1989, 78.2294, 78.2482, 78.2341,
        78.2435, 78.2576, 78.2787, 78.2787, 78.281, 78.2834, 78.2998, 78.3327, 78.3421, 78.3843, 78.4265, 78.4711, 78.4782, 78.5345, 78.572, 78.5838, 78.5885, 78.6025, 78.6307, 78.6753, 78.6917,
        78.7387, 78.7645, 78.8278, 78.8396, 78.8654, 78.8936, 78.9194, 78.9147, 78.9053, 78.8983, 78.863, 78.8396, 78.8232, 78.7903, 78.7645, 78.7574, 78.734, 78.6988, 78.6917, 78.6636, 78.6166,
        78.5791, 78.5439, 78.5392, 78.5345, 78.5533, 78.558, 78.5251, 78.5063, 78.4946, 78.4946, 78.45, 78.4383, 78.4195, 78.389, 78.3608, 78.3327, 78.3233, 78.3068, 78.2975, 78.1285, 78.1285, 78.105,
        78.1168, 78.098, 78.0745, 78.044, 78.0135, 77.983, 77.9548, 77.9267, 77.9173, 77.9337] };
  WCL[1092] = struct {
   x = [-100.029, -99.9634, -99.8977, -99.8414, -99.7757, -99.7006, -99.6818, -99.7663, -99.7757, -99.8696, -99.9728, -99.9541, -99.9071, -99.9259, -99.9259, -99.8039, -99.6724, -99.6067, -99.4941, -99.3533,
        -99.2312, -99.1186, -99.0341, -99.0904, -99.2031, -99.1655, -99.2594, -99.372, -99.5504, -99.7945, -99.9353],
   y = [78.7246, 78.6988, 78.673, 78.6988, 78.6894, 78.6683, 78.6378, 78.6143, 78.5673, 78.5321, 78.4969, 78.4664, 78.4336, 78.4031, 78.3702, 78.3655, 78.3491, 78.281, 78.2529, 78.1919, 78.1543, 78.1121,
        78.0722, 78.0323, 77.9806, 77.9314, 77.8727, 77.868, 77.8492, 77.8539, 77.8351] };
  WCL[1093] = struct {
   x = [-98.3418, -98.3324, -98.2761, -98.2385, -98.2667, -98.2667, -98.2385, -98.2667, -98.3136, -98.3606, -98.4263, -98.5483, -98.6328, -98.7736, -98.9144, -98.8862, -98.905, -99.0364, -99.1679, -99.365,
        -99.534, -99.534, -99.5434, -99.534, -99.534],
   y = [80.0271, 79.9825, 79.952, 79.9191, 79.8769, 79.8323, 79.7877, 79.7431, 79.6938, 79.6539, 79.6187, 79.6328, 79.668, 79.7079, 79.7455, 79.7689, 79.8088, 79.8346, 79.8346, 79.8276, 79.8393, 79.8628,
        79.9003, 79.9356, 79.959] };
  WCL[1094] = struct {
   x = [-89.9706, -90.0551, -90.1208, -90.1959, -90.2147, -90.0833],
   y = [78.4993, 78.5439, 78.5603, 78.5533, 78.4805, 78.4265] };
  WCL[1095] = struct {
   x = [-89.9049, -90.1302, -90.271, -90.4118, -90.4306, -90.4681, -90.6183, -90.5996, -90.4963, -90.3179, -90.1959, -90.1302, -90.1771, -90.3367, -90.5245, -90.8061, -91.0126, -91.1722, -91.2661, -91.3506,
        -91.5101, -91.6603, -91.8387, -92.0077, -92.0922, -92.1766, -92.2424, -92.2142, -92.1391, -92.1485, -92.2048, -92.2236, -92.2987, -92.355, -92.4583, -92.4958, -92.524, -92.5803, -92.7305,
        -92.8338, -92.8338, -92.8431, -92.768, -92.646, -92.4583, -92.4019, -92.2893, -92.1015, -91.9232, -91.7542, -91.6885, -91.651, -91.6885, -91.7918, -91.9607, -92.0922, -92.2424, -92.3456,
        -92.4207, -92.5709, -92.6648, -92.7493, -92.8338, -92.8995, -92.9652, -93.0872, -93.106, -93.0591, -93.1905, -93.2937, -93.3313, -93.2374, -93.106, -92.8995, -92.7117, -92.7399, -92.9089,
        -93.0215, -93.1154, -93.2656, -93.4439, -93.519, -93.5566, -93.7256, -93.7725, -93.5566, -93.3594, -93.275, -93.1905, -93.0966, -92.9182, -92.7023, -92.5897, -92.4583, -92.3456, -92.0358,
        -91.8481, -91.6603, -91.3787, -91.0689, -90.7873, -90.5245, -90.3837, -90.3273, -90.271, -90.1865, -90.2147, -90.4024, -90.6653, -90.9375, -91.2285, -91.3599, -91.5383, -91.6603, -91.7918,
        -91.9513, -92.0358, -92.0546, -92.1954, -92.2799, -92.2705, -92.0828, -92.0358, -91.895, -91.5759, -91.3412, -91.1065, -91.022, -91.0408, -91.2848, -91.4632, -91.651, -91.895, -92.0828,
        -92.1579, -92.1109, -92.0922, -92.2799, -92.4207, -92.5521, -92.5803, -92.5897, -92.7023, -92.7493, -92.9089, -92.984, -92.8901, -92.8525, -92.8901, -92.9652, -93.0215, -93.0966, -93.1905,
        -93.228, -93.2468, -93.4252, -93.5566, -93.6692, -93.7443, -93.7349, -93.7256, -93.6786, -93.5284, -93.4721, -93.519, -93.5941, -93.6129, -93.6598, -93.7819, -93.857, -93.9696, -94.101,
        -94.0917, -94.0729, -94.1104, -94.1668, -94.2982, -94.4672, -94.5516, -94.6267, -94.6831, -94.7018, -94.6831, -94.7675, -94.8051, -94.8239, -94.8426, -94.9084, -94.9928, -95.0022, -95.0492,
        -95.0679, -95.1149, -95.1994, -95.3026, -95.3777, -95.359, -95.3965, -95.3683, -95.2463, -95.2463, -95.0022, -94.9741, -94.9271, -94.8145, -94.6361, -94.608, -94.3733, -94.1949, -93.9978,
        -93.9039, -93.8851, -93.8382, -93.8194, -93.7913, -93.8851, -94.0447, -94.1198, -94.1668, -94.3263, -94.4765, -94.5986, -94.7488, -94.9271, -95.0022, -95.0773, -95.0773, -95.1618, -95.3026,
        -95.4528, -95.5185, -95.65, -95.7438, -95.8377, -95.9222, -95.8659, -96.0067, -96.0724, -95.9879, -95.6781, -95.6218, -95.6687, -95.8283, -96.0161, -96.0818],
   y = [78.2271, 78.2458, 78.2364, 78.2317, 78.2341, 78.2411, 78.2411, 78.2247, 78.1965, 78.1848, 78.166, 78.1003, 78.0722, 78.0769, 78.0651, 78.0698, 78.0886, 78.1027, 78.1027, 78.1097, 78.1097, 78.1473,
        78.152, 78.152, 78.2012, 78.22, 78.2341, 78.2482, 78.2623, 78.2857, 78.2951, 78.2623, 78.2599, 78.2552, 78.267, 78.2928, 78.2951, 78.3139, 78.3374, 78.3866, 78.4242, 78.4406, 78.4383, 78.4406,
        78.4594, 78.457, 78.4594, 78.4617, 78.4688, 78.4758, 78.4805, 78.4969, 78.5087, 78.5157, 78.5063, 78.5275, 78.5462, 78.5533, 78.5486, 78.5791, 78.5697, 78.5627, 78.5603, 78.5533, 78.5556,
        78.5908, 78.6143, 78.6354, 78.6354, 78.68, 78.7293, 78.7434, 78.7222, 78.7199, 78.7152, 78.7527, 78.7645, 78.795, 78.8044, 78.8044, 78.7973, 78.8255, 78.863, 78.9123, 78.9593, 79.0062, 79.0156,
        79.0273, 79.0484, 79.0883, 79.1259, 79.1212, 79.1212, 79.1329, 79.1423, 79.1376, 79.1423, 79.147, 79.1681, 79.1846, 79.1916, 79.1986, 79.201, 79.208, 79.2409, 79.2714, 79.269, 79.2432, 79.2385,
        79.2292, 79.2221, 79.1986, 79.2057, 79.2057, 79.1986, 79.194, 79.2033, 79.2057, 79.2174, 79.2385, 79.2644, 79.2667, 79.2878, 79.2761, 79.2831, 79.2925, 79.3089, 79.323, 79.3371, 79.3324,
        79.323, 79.3207, 79.316, 79.3277, 79.3465, 79.377, 79.4028, 79.3981, 79.3887, 79.3817, 79.3723, 79.3301, 79.3277, 79.3207, 79.3089, 79.3207, 79.3747, 79.4122, 79.4427, 79.4099, 79.3582,
        79.3395, 79.3324, 79.2925, 79.2761, 79.2292, 79.2268, 79.2597, 79.2597, 79.2831, 79.3136, 79.3371, 79.3535, 79.3723, 79.377, 79.37, 79.37, 79.3794, 79.3934, 79.3958, 79.4005, 79.37, 79.3535,
        79.3418, 79.3183, 79.3183, 79.3019, 79.2667, 79.2292, 79.262, 79.2878, 79.2996, 79.316, 79.3465, 79.3794, 79.3794, 79.384, 79.3629, 79.3629, 79.3653, 79.3864, 79.384, 79.3817, 79.3794, 79.4028,
        79.4169, 79.4544, 79.4756, 79.5037, 79.5272, 79.5272, 79.5014, 79.4943, 79.4943, 79.5108, 79.5178, 79.5342, 79.5389, 79.5389, 79.5648, 79.5835, 79.6046, 79.6234, 79.6492, 79.6633, 79.6422,
        79.6164, 79.6023, 79.5882, 79.5976, 79.6187, 79.6164, 79.6187, 79.6046, 79.5976, 79.6211, 79.6211, 79.6234, 79.6422, 79.6657, 79.7032, 79.7337, 79.7807, 79.8041, 79.8182, 79.8323, 79.844,
        79.8675, 79.8769, 79.8745, 79.898, 79.9168, 79.9191, 79.9332, 79.9731] };
  WCL[1096] = struct {
   x = [-94.8239, -94.73],
   y = [80.0036, 79.9895] };
  WCL[1097] = struct {
   x = [-94.3498, -94.2278, -94.1621, -94.0494, -94.0682],
   y = [80.0036, 79.9942, 79.9731, 79.9496, 79.9708] };
  WCL[1098] = struct {
   x = [-95.0656, -94.9999, -94.9905, -94.9248, -94.8497, -94.7558, -94.6526, -94.5493, -94.5493, -94.4836, -94.3146, -94.2395, -94.1644, -94.0893, -93.9861, -93.9297, -93.8359, -93.7326, -93.6669, -93.6481,
        -93.6012, -93.5167, -93.3759, -93.3008, -93.2726, -93.2069, -93.2351, -93.3477, -93.4416, -93.4885, -93.4416, -93.4228, -93.4604, -93.4698, -93.4322, -93.4885, -93.5261, -93.6199, -93.6669,
        -93.742, -93.8452, -93.8828, -93.9673, -94.1081, -94.2583, -94.3052, -94.3709, -94.4742, -94.5962, -94.6995, -94.7277, -94.8028, -94.8685, -94.9717, -94.9999, -95.0187, -95.0187, -95.0562,
        -95.2064, -95.3378, -95.5725, -95.7415, -95.8823, -96.0607, -96.1358, -96.1358, -96.1921, -96.1451, -95.9574, -95.7978, -95.6288, -95.5819, -95.5537, -95.4411, -95.3754, -95.2909, -95.197,
        -95.0656],
   y = [77.7389, 77.7272, 77.7201, 77.7201, 77.7436, 77.7389, 77.7553, 77.753, 77.7342, 77.7342, 77.7201, 77.7342, 77.7178, 77.7154, 77.7131, 77.6943, 77.7014, 77.7014, 77.7201, 77.7342, 77.7225, 77.6967,
        77.6943, 77.6802, 77.6568, 77.6216, 77.6028, 77.5934, 77.5911, 77.5699, 77.5512, 77.516, 77.4878, 77.462, 77.4409, 77.4197, 77.3963, 77.3916, 77.4033, 77.3986, 77.3892, 77.3892, 77.401,
        77.3986, 77.4057, 77.4197, 77.4409, 77.4315, 77.4291, 77.4291, 77.4456, 77.4432, 77.4197, 77.415, 77.4127, 77.4315, 77.4315, 77.4338, 77.4503, 77.4479, 77.4526, 77.4948, 77.4948, 77.5488,
        77.5958, 77.6216, 77.6333, 77.6802, 77.7154, 77.7319, 77.7389, 77.7624, 77.7929, 77.7929, 77.7389, 77.746, 77.7272, 77.7389] };
  WCL[1099] = struct {
   x = [-94.9999, -94.9905, -94.9811, -94.8591, -94.6807, -94.446, -94.2864, -94.1081, -93.9391, -93.8171, -93.9579, -93.9391, -93.8734, -93.8077, -93.7889, -93.6481, -93.5261, -93.4979, -93.4228, -93.2445,
        -93.1694, -93.0755, -92.9159, -92.8314, -92.8408, -92.7469, -92.7375, -92.531, -92.362, -92.3433, -92.2869, -92.1555, -92.3855, -92.6531, -92.5545, -92.5873, -92.639, -92.6624, -92.5545,
        -92.2869, -92.1884, -92.2541, -92.0147, -91.9443, -92.2447, -92.3527, -92.2869, -92.3104, -92.2682, -92.2353, -92.0804, -91.8927, -91.7566, -91.8786, -91.8598, -91.7566, -91.5735, -91.4186,
        -91.4045, -91.3154, -91.1417, -91.0103, -91.0854, -91.198, -91.1511, -91.076, -90.9539, -90.9117, -90.9117, -90.724, -90.5221, -90.5878, -90.4001, -90.3344, -90.2827, -90.1279],
   y = [76.2111, 76.2628, 76.2862, 76.244, 76.244, 76.2393, 76.2111, 76.2205, 76.2064, 76.2158, 76.2346, 76.2698, 76.2792, 76.2792, 76.2651, 76.2886, 76.312, 76.3355, 76.3566, 76.3285, 76.2416, 76.1806,
        76.1219, 76.0234, 75.9506, 75.9342, 75.9342, 75.8638, 75.784, 75.6901, 75.6056, 75.5704, 75.5047, 75.3733, 75.3545, 75.3123, 75.2372, 75.1809, 75.1058, 75.1316, 75.1269, 75.0706, 75.0987,
        75.0776, 75.0518, 74.9743, 74.8547, 74.8007, 74.7842, 74.7725, 74.7256, 74.7185, 74.6951, 74.6833, 74.6434, 74.6153, 74.6223, 74.6317, 74.7162, 74.7303, 74.796, 74.8101, 74.7702, 74.7091,
        74.6904, 74.6693, 74.6176, 74.5988, 74.7303, 74.6082, 74.6059, 74.5566, 74.5449, 74.5848, 74.5144, 74.5026] };
  WCL[1100] = struct {
   x = [-89.9378, -90.0598],
   y = [75.7652, 75.784] };
  WCL[1101] = struct {
   x = [-89.9589, -90.0621, -90.0903, -90.1936, -90.2781, -90.3438, -90.4752, -90.616, -90.6911, -90.6817, -90.8225, -90.9446, -90.9539, -91.0572, -91.2731, -91.4139, -91.3764, -91.2262, -91.198, -91.3107,
        -91.1417, -90.8788, -90.6535, -90.5503, -90.7286, -90.9633, -91.076, -91.0197, -91.0103, -91.0854, -91.1605, -91.3951, -91.6298, -91.7143, -91.5078, -91.6298, -91.6298, -91.6017, -91.3294,
        -91.0666, -90.8788, -90.6442, -90.5127, -90.6817, -90.4658, -90.2687],
   y = [75.8497, 75.8802, 75.9107, 75.9295, 75.8849, 75.8732, 75.9037, 75.8685, 75.8591, 75.9178, 75.9013, 75.9037, 75.9154, 75.8591, 75.7863, 75.7558, 75.8497, 75.8732, 75.8943, 75.9412, 75.9553, 75.96,
        75.9694, 75.9788, 76.0093, 75.9929, 76.0257, 76.0351, 76.0539, 76.0539, 76.0844, 76.1173, 76.1126, 76.1243, 76.1759, 76.1877, 76.2017, 76.2229, 76.1478, 76.1266, 76.1337, 76.0727, 76.0727,
        76.1102, 76.1149, 76.1079] };
  WCL[1102] = struct {
   x = [-89.9589, -90.0809, -90.2217, -90.325, -90.4564, -90.5972, -90.6817, -90.7474, -90.8507, -91.0009, -91.0854, -91.1605, -91.2637, -91.4233, -91.5829, -91.6298, -91.6298, -91.4233, -91.3107, -91.1699,
        -91.1041, -91.0197, -90.8695, -90.7099, -90.5597, -90.5972, -90.616, -90.7099, -90.8976, -91.0197, -91.1886, -91.3576, -91.4702, -91.4984, -91.5172, -91.6862, -91.7143, -91.8551, -91.949,
        -92.0898, -92.2588, -92.4935, -92.6061, -92.7751, -92.9534, -93.0567, -93.1787, -93.2726, -93.3477, -93.3947, -93.4791, -93.5918, -93.6481, -93.7326, -93.7044, -93.7044, -93.7138, -93.5824,
        -93.4979, -93.5261, -93.4698, -93.4134, -93.4791, -93.4979, -93.6106, -93.742, -93.8077, -93.9767, -94.0705, -94.155, -94.2489, -94.3146, -94.3991, -94.5117, -94.6526, -94.7464, -94.8779,
        -94.9999, -94.9999, -95.075, -95.1689, -95.2533, -95.3378, -95.4599, -95.6007, -95.7133, -96.0607, -96.1076, -96.1921, -96.2578, -96.3235, -96.3798, -96.4831, -96.5676, -96.5863, -96.7553,
        -96.8961, -97.0088, -96.9994, -96.9618, -96.8961, -96.7553, -96.6239, -96.4361, -96.3235, -96.3611, -96.5676, -96.6145, -96.4549, -96.5676, -96.6802, -96.868, -96.9055, -96.8398, -96.8961,
        -96.7835, -96.6333, -96.5206, -96.5112, -96.3047, -96.0888, -95.9198, -95.7509, -95.5725, -95.4505, -95.2909, -95.1689, -94.9999],
   y = [76.2839, 76.2956, 76.3167, 76.3355, 76.3566, 76.3637, 76.3425, 76.3707, 76.3871, 76.4036, 76.4294, 76.4341, 76.4294, 76.4341, 76.4341, 76.4622, 76.4834, 76.4787, 76.4669, 76.4669, 76.4951, 76.4622,
        76.4364, 76.427, 76.4411, 76.4693, 76.5162, 76.5514, 76.5749, 76.6054, 76.6312, 76.6429, 76.6688, 76.6711, 76.6922, 76.7016, 76.6828, 76.6758, 76.6664, 76.65, 76.596, 76.5796, 76.6148, 76.5796,
        76.5866, 76.596, 76.5937, 76.5514, 76.5186, 76.4927, 76.4575, 76.427, 76.42, 76.4083, 76.4505, 76.4716, 76.481, 76.5397, 76.5796, 76.6289, 76.6711, 76.7392, 76.7532, 76.8025, 76.8354, 76.9034,
        76.9457, 76.9316, 76.9692, 76.9316, 76.9152, 76.9152, 76.9081, 76.9551, 76.9692, 76.995, 77.0067, 77.0302, 76.2111, 76.2674, 76.3425, 76.3543, 76.3801, 76.3824, 76.3918, 76.3942, 76.5937,
        76.5866, 76.5866, 76.6171, 76.6547, 76.6453, 76.6641, 76.6781, 76.711, 76.7251, 76.7227, 76.7485, 76.765, 76.8166, 76.8307, 76.7931, 76.7814, 76.772, 76.7767, 76.7931, 76.8002, 76.819, 76.819,
        76.8448, 76.8823, 76.8964, 76.9339, 76.9598, 76.9926, 76.9903, 76.9809, 77.009, 77.0466, 77.0654, 77.0865, 77.0935, 77.0865, 77.0818, 77.0536, 77.0489, 77.0255, 77.0302] };
  WCL[1103] = struct {
   x = [-94.9999, -95.0327, -94.9858, -94.8732, -94.6338, -94.446, -94.2818, -94.2911, -94.1691, -94.0799, -94.1456, -93.9485, -93.8265, -93.8405, -93.6857, -93.5871, -93.6528, -93.6622, -93.5965, -93.6059,
        -93.6199, -93.6528, -93.5871, -93.5402, -93.5636, -93.5871, -93.6387, -93.7936, -94.0377, -94.2677, -94.4883, -94.7323, -94.9201, -94.9529, -94.9529, -94.9858, -95.0515, -95.1078, -95.1266,
        -95.1501, -95.2158, -95.2956, -95.3284, -95.4035, -95.4599, -95.427, -95.5021, -95.5678, -95.5819, -95.5913, -95.6148, -95.6664, -95.7556, -95.8025, -95.9011, -95.9433, -96.056, -96.056,
        -95.9762, -96.0419, -96.0888, -96.1451, -96.1874, -96.2766, -96.3282, -96.4174, -96.5066, -96.53, -96.5159, -96.408, -96.3282, -96.3845, -96.2953, -96.1076, -95.887, -95.9574, -96.0982,
        -95.8776, -95.9011, -96.0653, -96.056, -95.9574, -95.9433, -95.8025, -95.8025, -95.9668, -96.1076, -96.1217, -96.009, -95.8541, -95.7227, -95.7227, -95.7884, -95.7227, -95.549, -95.4129,
        -95.3378, -95.1407, -95.0844],
   y = [75.6573, 75.6901, 75.6361, 75.6197, 75.5892, 75.5352, 75.5118, 75.4766, 75.4742, 75.4202, 75.3428, 75.3968, 75.3404, 75.3146, 75.2794, 75.2348, 75.2067, 75.1903, 75.1269, 75.1011, 75.0682, 75.0236,
        74.9344, 74.8453, 74.7631, 74.7091, 74.6716, 74.6364, 74.6411, 74.6176, 74.6364, 74.6223, 74.6669, 74.7045, 74.7045, 74.7045, 74.7115, 74.7256, 74.7444, 74.7725, 74.803, 74.8265, 74.8195,
        74.8288, 74.8077, 74.7819, 74.7749, 74.796, 74.8124, 74.8523, 74.8617, 74.8734, 74.8734, 74.8406, 74.8617, 74.8992, 74.9532, 74.9908, 75.033, 75.0166, 75.0565, 74.9931, 74.9626, 74.9579,
        75.0189, 75.0377, 75.1269, 75.1738, 75.2161, 75.2161, 75.2348, 75.2677, 75.2794, 75.2771, 75.2888, 75.3052, 75.3123, 75.3287, 75.3498, 75.3569, 75.3827, 75.4296, 75.3991, 75.4109, 75.4578,
        75.4648, 75.4202, 75.4578, 75.486, 75.4813, 75.4695, 75.4953, 75.5657, 75.5822, 75.615, 75.6455, 75.6667, 75.6807, 75.6901] };
  WCL[1104] = struct {
   x = [-95.0703, -94.9999, -94.9905, -94.8356, -94.7277, -94.5071, -94.3381, -94.1738, -93.634, -93.4885, -93.2257, -92.9628, -92.8971, -92.883, -92.7939, -92.7422, -92.5404, -92.3996, -92.3104, -92.179,
        -92.1321, -92.0476, -92.0476, -91.9443, -91.7237, -91.5923, -91.3717, -91.2168, -91.029, -90.7333, -90.5691, -90.4236, -90.3344, -90.3485, -90.4376, -90.4705, -90.5362, -90.6582, -90.7568,
        -90.8742, -90.9962, -91.0948, -91.1417, -91.0197, -91.029, -91.0948, -91.1323, -91.2074, -91.2309, -91.4702, -91.6064, -91.428, -91.5172, -91.6251, -91.7237, -91.8598, -91.9021, -91.9021,
        -92.0663, -92.1884, -92.3667, -92.409, -92.639, -92.8408, -93.0802, -93.2491, -93.2914, -93.4557, -93.634, -93.9297, -94.263, -94.3944, -94.2489, -94.216, -93.9767, -93.8969, -93.7748,
        -93.6246, -93.5214, -93.634, -93.6669, -93.6997, -93.7655, -93.864, -93.9626, -93.9954, -94.0424, -94.1503, -94.1738, -94.1503, -94.1738, -94.263, -94.3944, -94.4273, -94.5821, -94.6385,
        -94.7277, -94.8356, -94.9107, -94.9905, -94.9999, -95.0046, -95.0046, -95.0374, -95.1266, -95.2252, -95.2158, -95.1172, -94.9529, -94.9952, -95.1266, -95.2158, -95.2252, -95.2252, -95.2252,
        -95.1829, -95.028, -94.9389, -95.1407, -95.2487, -95.2815, -95.3378, -95.4035, -95.5256, -95.6148, -95.6476, -95.5913, -95.657, -95.6805, -95.6899, -95.6805, -95.6805, -95.6805, -95.6805,
        -95.6664, -95.6007, -95.6476, -95.6241, -95.6241, -95.6241, -95.6335, -95.657, -95.6899, -95.6992, -95.6664, -95.6664, -95.6899, -95.7133, -95.6805, -95.5913, -95.5115, -95.4599, -95.3801,
        -95.2909, -95.2064, -95.1501, -95.1501, -94.9999, -94.9248, -94.8591, -94.7699, -94.6244, -94.7136, -94.8779, -94.9905, -94.9905, -94.9999, -95.0374, -95.0374, -95.0609, -95.1501, -95.2721,
        -95.3284, -95.4035, -95.4035, -95.4035, -95.3942, -95.2815, -95.1736, -95.0703],
   y = [74.0732, 74.0473, 74.0779, 74.1107, 74.0685, 74.0802, 74.0919, 74.1107, 74.1717, 74.1717, 74.1436, 74.1224, 74.1107, 74.0708, 74.0403, 74.0943, 74.0779, 74.0051, 73.9558, 73.9863, 74.0168, 74.0192,
        74.0239, 74.0192, 74.0192, 74.0168, 74.0098, 73.9934, 73.9769, 73.9464, 73.9159, 73.9065, 73.9018, 73.8713, 73.8455, 73.7751, 73.693, 73.6413, 73.5803, 73.5686, 73.5615, 73.5874, 73.5404,
        73.5193, 73.463, 73.416, 73.3808, 73.3245, 73.2729, 73.2447, 73.2166, 73.1978, 73.0969, 73.0452, 72.9514, 72.8716, 72.8645, 72.8575, 72.7707, 72.712, 72.705, 72.7073, 72.7589, 72.7237, 72.7472,
        72.7894, 72.7965, 72.7918, 72.7707, 72.7495, 72.7472, 72.7331, 72.7143, 72.6909, 72.6909, 72.6721, 72.59, 72.5407, 72.4609, 72.428, 72.4069, 72.3576, 72.3107, 72.3013, 72.2567, 72.2051,
        72.1675, 72.1464, 72.13, 72.1159, 72.0783, 72.0525, 72.0408, 72.0079, 72.022, 72.015, 72.0056, 72.0056, 72.0056, 71.9868, 72.0056, 72.015, 72.015, 72.0103, 72.0009, 72.0291, 72.076, 72.1206,
        72.1441, 72.1675, 72.1534, 72.184, 72.2637, 72.3318, 72.4069, 72.4679, 72.5102, 72.5594, 72.536, 72.5266, 72.5571, 72.5993, 72.6322, 72.6533, 72.6909, 72.7308, 72.7542, 72.8059, 72.8551,
        72.9115, 72.9631, 73.0053, 73.0476, 73.0922, 73.118, 73.1344, 73.172, 73.1907, 73.2353, 73.287, 73.3409, 73.4067, 73.463, 73.5498, 73.5897, 73.6226, 73.6648, 73.7, 73.7352, 73.7634, 73.7939,
        73.7751, 73.7704, 73.7423, 73.7305, 73.7071, 73.7071, 73.6484, 73.6578, 73.6249, 73.6085, 73.6273, 73.6672, 73.6765, 73.7024, 73.7305, 73.7352, 73.8408, 73.8408, 73.8408, 73.8408, 73.8666,
        73.8854, 73.9042, 73.9558, 74.0028, 74.0309, 74.045, 74.0661, 74.0732] };
  WCL[1105] = struct {
   x = [-94.9999, -95.0374, -94.9999, -94.8779, -94.7042, -94.615, -94.5071, -94.5399, -94.5915, -94.6244, -94.5915, -94.4507, -94.4507, -94.4366, -94.2724, -94.216, -94.0424, -93.8875, -93.7326, -93.742,
        -93.7983, -93.7091, -93.5214, -93.2914, -93.16, -93.0473, -92.9957, -92.9394, -92.9394, -92.93, -92.9488, -93.0145, -93.0051, -92.7939, -92.7422, -92.6765, -92.4747, -92.3433, -92.3339,
        -92.3339, -92.3198, -92.301, -92.1461, -92.0898, -92.0663, -92.0663, -92.0147, -91.9349, -91.9115, -91.7941, -91.7706, -91.7472, -91.6251, -91.5594, -91.5735, -91.658, -91.78, -91.8927,
        -92.0476, -92.2118, -92.3339, -92.4559, -92.4324, -92.4418, -92.5075, -92.4982, -92.1555],
   y = [71.9375, 71.954, 71.9446, 71.9399, 71.9563, 71.9328, 71.8953, 71.8507, 71.8272, 71.7756, 71.7381, 71.7756, 71.7146, 71.6723, 71.7216, 71.7733, 71.7029, 71.724, 71.7474, 71.6864, 71.5996, 71.5292,
        71.5198, 71.4072, 71.3367, 71.2922, 71.1959, 71.1044, 71.0364, 70.9824, 70.8486, 70.8556, 70.7993, 70.7852, 70.6609, 70.6726, 70.6679, 70.6139, 70.6092, 70.5881, 70.5224, 70.4848, 70.4004,
        70.337, 70.2924, 70.2736, 70.2736, 70.3112, 70.3417, 70.3041, 70.2314, 70.1938, 70.1774, 70.1305, 70.1211, 70.1117, 70.107, 70.1023, 70.1258, 70.1586, 70.1751, 70.1633, 70.1446, 70.107,
        70.0835, 70.0695, 70.0507] };
  WCL[1106] = struct {
   x = [-89.9589, -90.0387, -90.1279, -90.1138, -90.1279],
   y = [71.3086, 71.3438, 71.4189, 71.5292, 71.6512] };
  WCL[1107] = struct {
   x = [-89.9495, -90.0715, -90.1936, -90.0715],
   y = [71.7615, 71.8366, 71.9634, 72.0549] };
  WCL[1108] = struct {
   x = [-89.9706, -90.0457],
   y = [72.1699, 72.367] };
  WCL[1109] = struct {
   x = [-110.069, -109.98],
   y = [72.5407, 72.5641] };
  WCL[1110] = struct {
   x = [-110.003, -109.914],
   y = [72.6533, 72.6956] };
  WCL[1111] = struct {
   x = [-110.069, -109.881, -109.848, -109.825, -109.703, -109.618, -109.628, -109.571, -109.482, -109.341, -109.243, -109.121, -109.121, -108.966, -108.989, -109.041, -109.008, -108.844, -108.712, -108.712,
        -108.736, -108.712, -108.67, -108.581, -108.567, -108.548, -108.501, -108.384, -108.281, -108.417, -108.403, -108.403, -108.403, -108.304, -108.196, -108.126, -108.027, -107.994, -107.976,
        -107.886, -107.741, -107.6, -107.577, -107.422, -107.422, -107.534, -107.642, -107.699, -107.685, -107.774, -107.83, -107.854, -107.863, -107.797, -107.83, -107.821, -107.896, -107.905,
        -107.905, -107.943, -107.961, -107.976, -108.051, -108.126, -108.173, -108.229, -108.149, -108.126, -108.206, -108.271, -108.196, -108.018, -107.976, -108.173, -108.163, -108.083, -108.093,
        -107.976, -107.83, -107.708, -107.577, -107.553, -107.365, -107.225, -107.004, -106.948, -106.915, -106.891, -106.99, -107.023, -107.023, -106.938, -106.826, -106.736, -106.694, -106.596,
        -106.549, -106.506, -106.474, -106.384, -106.197, -106.098, -105.943, -105.934, -105.868, -105.704, -105.657, -105.558, -105.525, -105.394, -105.371, -105.446, -105.493, -105.38, -105.338,
        -105.281, -105.281, -105.305, -105.272, -105.239, -105.216, -105.216, -105.192, -105.159, -105.108, -105.094, -105.094, -105.075, -105.075, -105.019, -104.986, -104.953, -104.929, -104.92,
        -104.634, -104.512, -104.446, -104.413, -104.446, -104.39, -104.38, -104.38, -104.413, -104.544, -104.544, -104.521, -104.521, -104.601, -104.53, -104.413, -104.268, -104.211, -104.07,
        -103.981, -103.939, -103.836, -103.639, -103.582, -103.498, -103.395, -103.287, -103.197, -103.089, -103.01, -103.01, -102.944, -102.747, -102.634, -102.681, -102.583, -102.348, -102.184,
        -101.986, -101.808, -101.587, -101.498, -101.531, -101.512, -101.343, -101.226, -101.179, -101.024, -100.94, -100.902, -100.86],
   y = [72.9748, 72.9514, 72.9068, 72.8904, 72.8997, 72.8974, 72.8598, 72.8246, 72.7801, 72.7472, 72.766, 72.7308, 72.7073, 72.6744, 72.6416, 72.6252, 72.6228, 72.6064, 72.5242, 72.4374, 72.3553, 72.2943,
        72.245, 72.1933, 72.1065, 72.0314, 72.0103, 71.9845, 71.9469, 71.8648, 71.7967, 71.7662, 71.7334, 71.7287, 71.6817, 71.6489, 71.6559, 71.6723, 71.7029, 71.7216, 71.7381, 71.7967, 71.8554,
        71.8789, 71.9117, 71.961, 72.0079, 72.0737, 72.1323, 72.1628, 72.2567, 72.3271, 72.3459, 72.3506, 72.3975, 72.4585, 72.4984, 72.5641, 72.6064, 72.6744, 72.712, 72.7237, 72.8035, 72.9115,
        72.9889, 73.0241, 73.0241, 73.0898, 73.0922, 73.1509, 73.1767, 73.179, 73.2095, 73.2635, 73.3128, 73.3081, 73.3409, 73.3409, 73.3222, 73.3175, 73.3034, 73.2705, 73.2142, 73.1767, 73.1602,
        73.1462, 73.1555, 73.1907, 73.2001, 73.2424, 73.2705, 73.2799, 73.2729, 73.2541, 73.2236, 73.2142, 73.2001, 73.1509, 73.1156, 73.0758, 73.0452, 73.0218, 72.996, 72.9608, 72.9514, 72.9185,
        72.8974, 72.8716, 72.8411, 72.8059, 72.773, 72.7378, 72.6979, 72.6979, 72.6557, 72.604, 72.543, 72.5102, 72.4844, 72.4609, 72.4116, 72.4022, 72.3717, 72.3271, 72.2708, 72.2145, 72.1511,
        72.0666, 72.015, 71.9939, 71.9446, 71.8859, 71.8624, 71.8577, 71.7451, 71.6747, 71.6278, 71.5855, 71.5409, 71.5034, 71.4424, 71.4001, 71.4072, 71.3766, 71.3086, 71.264, 71.2241, 71.1537,
        71.0762, 71.0387, 71.0152, 70.9613, 70.8885, 70.8392, 70.8017, 70.7852, 70.7688, 70.7195, 70.6562, 70.6327, 70.628, 70.6233, 70.5834, 70.6092, 70.6609, 70.6796, 70.6491, 70.6163, 70.567,
        70.5294, 70.4708, 70.4051, 70.3417, 70.3088, 70.3088, 70.2666, 70.215, 70.1586, 70.1469, 70.1774, 70.1516, 70.1704, 70.1586, 70.0929, 70.0131] };
  WCL[1112] = struct {
   x = [-104.974, -104.974, -104.908, -104.735, -104.645, -104.603, -104.589, -104.547, -104.622, -104.753, -104.866, -105.077, -105.251, -105.448, -105.626, -105.805, -105.978, -106.18, -106.434, -106.631,
        -106.729, -106.819, -107.025, -107.063, -106.983, -106.804, -106.706, -106.584, -106.476, -106.354, -106.246, -106.19, -106.077, -105.96, -105.856, -105.772, -105.739, -105.626, -105.462,
        -105.364, -105.317, -105.317, -105.218, -105.152, -105.054, -104.988, -104.974],
   y = [73.0804, 73.1203, 73.1767, 73.2612, 73.3644, 73.4536, 73.5193, 73.5569, 73.6038, 73.6085, 73.6765, 73.7141, 73.7423, 73.754, 73.7423, 73.7258, 73.7352, 73.7211, 73.7235, 73.693, 73.646, 73.5897,
        73.5099, 73.4583, 73.4419, 73.4325, 73.3996, 73.3879, 73.3409, 73.3034, 73.2799, 73.2447, 73.1884, 73.1602, 73.1016, 73.0664, 73.0147, 72.9748, 72.9185, 72.8833, 72.9021, 72.9256, 72.9279,
        72.9608, 72.996, 73.0476, 73.0804] };
  WCL[1113] = struct {
   x = [-99.9893, -100.036, -100.111, -100.191, -100.2, -100.313, -100.445, -100.543, -100.642, -100.717, -100.853, -100.975, -101.017, -101.181, -101.449, -101.501, -101.515, -101.688, -101.712, -101.843,
        -102.054, -102.12, -102.261, -102.416, -102.43, -102.439, -102.406, -102.341, -102.177, -102.022, -101.942, -101.9, -101.787, -101.656, -101.637, -101.679, -101.58, -101.515, -101.369, -101.28,
        -101.271, -101.196, -101.139, -100.806, -100.684, -100.642, -100.552, -100.52, -100.388, -100.29, -100.177, -100.276, -100.313, -100.243, -100.111, -100.036, -100.022, -100.102, -100.177,
        -100.276, -100.313, -100.412, -100.412, -100.43, -100.398, -100.233, -100.102, -100.003],
   y = [71.8812, 71.8953, 71.954, 72.0079, 72.0525, 72.0924, 72.1065, 72.0854, 72.0666, 72.137, 72.2004, 72.2004, 72.1534, 72.1699, 72.1746, 72.2285, 72.2544, 72.3083, 72.3459, 72.3787, 72.4304, 72.4773,
        72.5242, 72.5923, 72.6416, 72.6979, 72.8129, 72.8997, 72.9443, 72.9443, 72.9537, 72.9326, 72.8997, 72.8763, 72.8669, 72.8293, 72.7965, 72.7894, 72.7801, 72.7472, 72.6885, 72.6533, 72.604,
        72.597, 72.6228, 72.6486, 72.6486, 72.6369, 72.6533, 72.6744, 72.7167, 72.773, 72.8645, 72.8645, 72.8059, 72.8223, 72.8974, 72.9349, 72.942, 72.9373, 72.9983, 73.01, 73.0828, 73.1462, 73.1649,
        73.1767, 73.1462, 73.1203] };
  WCL[1114] = struct {
   x = [-99.8907, -100.036, -100.158, -100.29, -100.355, -100.388, -100.552, -100.754, -100.886, -101.041, -101.181, -101.327, -101.369, -101.346, -101.163, -100.984, -100.853, -100.806, -100.632, -100.585,
        -100.665, -100.651, -100.773, -100.895, -100.994, -101.149, -101.181, -101.228, -101.163, -101.041, -100.886, -100.707, -100.609, -100.487, -100.365, -100.266, -100.191, -100.144, -100.144,
        -100.299, -100.379, -100.454, -100.276, -100.093],
   y = [73.1297, 73.1884, 73.2447, 73.3034, 73.2729, 73.2259, 73.2072, 73.1978, 73.2283, 73.2518, 73.3363, 73.3456, 73.4278, 73.4818, 73.4935, 73.4841, 73.4512, 73.4113, 73.3808, 73.416, 73.4794, 73.5263,
        73.5334, 73.5404, 73.585, 73.5874, 73.6226, 73.6531, 73.6906, 73.7258, 73.761, 73.7798, 73.7446, 73.7352, 73.7258, 73.7117, 73.6906, 73.7235, 73.761, 73.7728, 73.7892, 73.8244, 73.8643,
        73.8854] };
  WCL[1115] = struct {
   x = [-99.7757, -100.006, -100.217, -100.447, -100.588, -100.588, -100.588, -100.602, -100.611, -100.667, -100.644, -100.522, -100.391, -100.236, -100.137, -100.292, -100.423, -100.367, -100.348, -100.489,
        -100.658, -100.724, -100.667, -100.503, -100.555, -100.635, -100.757, -100.621, -100.423, -100.25, -100.179, -100.259, -100.062],
   y = [74.9978, 75.0025, 74.9955, 75.026, 75.0706, 75.1058, 75.1128, 75.1457, 75.1856, 75.2255, 75.2442, 75.256, 75.2419, 75.2466, 75.2677, 75.2794, 75.2794, 75.3358, 75.3498, 75.3451, 75.3217, 75.3733,
        75.3827, 75.4085, 75.4226, 75.4437, 75.4648, 75.4672, 75.4766, 75.4672, 75.4953, 75.5071, 75.5188] };
  WCL[1116] = struct {
   x = [-99.8508, -100.062, -100.259, -100.203, -100.006, -100.048],
   y = [75.5493, 75.5446, 75.5305, 75.5493, 75.5728, 75.5939] };
  WCL[1117] = struct {
   x = [-99.8414, -100.02, -100.189, -100.311, -100.358, -100.48, -100.611, -100.733, -100.841, -101.01, -101.151, -101.296, -101.503, -101.606, -101.724, -101.789, -101.902, -102.033, -102.212, -102.353,
        -102.432, -102.507, -102.653, -102.751, -102.827, -102.784, -102.672, -102.7, -102.653, -102.512, -102.39, -102.268, -102.127, -102.174, -102.277, -102.221, -102.343, -102.324, -102.071,
        -101.949, -101.808, -101.705, -101.611, -101.545, -101.367, -101.179, -101.02, -101.151, -101.32, -101.329, -101.367, -101.508, -101.648, -101.583, -101.433, -101.433, -101.648, -101.874,
        -101.902, -102.024, -102.024, -101.958, -102.08, -102.221, -102.212, -102.24, -102.137, -102.024, -102.062, -101.996, -101.874, -101.771, -101.639, -101.545, -101.498, -101.423, -101.376,
        -101.367, -101.376, -101.367, -101.292, -101.198, -101.085, -101.142, -101.207, -101.104, -100.926, -100.813, -100.822, -100.757, -100.625, -100.513, -100.597, -100.625, -100.409, -100.231,
        -100.137, -100.1],
   y = [75.7347, 75.7394, 75.7253, 75.7066, 75.7019, 75.7089, 75.6948, 75.6784, 75.662, 75.6502, 75.6385, 75.6385, 75.6479, 75.6103, 75.6338, 75.6268, 75.5892, 75.5775, 75.5775, 75.5775, 75.5657, 75.5376,
        75.554, 75.5728, 75.6361, 75.6455, 75.6455, 75.6831, 75.723, 75.7347, 75.723, 75.7136, 75.7019, 75.7394, 75.7558, 75.7957, 75.7957, 75.8568, 75.8966, 75.899, 75.8661, 75.8427, 75.8286, 75.7863,
        75.7817, 75.7957, 75.838, 75.8262, 75.8028, 75.8474, 75.8802, 75.8966, 75.9319, 75.9741, 76.0163, 76.0422, 76.0116, 75.9976, 76.0515, 76.1055, 76.1689, 76.2182, 76.2276, 76.2229, 76.2839,
        76.3285, 76.3754, 76.4059, 76.4153, 76.4552, 76.4763, 76.4622, 76.4622, 76.4458, 76.4435, 76.4505, 76.4294, 76.427, 76.4059, 76.3754, 76.3824, 76.3637, 76.3425, 76.3285, 76.2886, 76.2581,
        76.2041, 76.1642, 76.1478, 76.1219, 76.1384, 76.1266, 76.1126, 76.0914, 76.0562, 76.0116, 75.9905, 75.9788] };
  WCL[1118] = struct {
   x = [-99.9728, -100.048, -100.114, -100.236, -100.095],
   y = [76.0422, 76.075, 76.1196, 76.1923, 76.2041] };
  WCL[1119] = struct {
   x = [-99.94, -100.034, -100.193, -100.447, -100.616, -100.541, -100.334, -100.287, -100.128, -100.259, -100.409, -100.55, -100.719, -100.888, -101.038, -101.076, -101.057, -100.869, -100.794, -100.606,
        -100.475, -100.297, -100.156, -100.034],
   y = [76.2041, 76.2416, 76.2276, 76.2276, 76.2698, 76.2933, 76.2886, 76.3214, 76.3261, 76.3425, 76.3848, 76.3777, 76.3848, 76.42, 76.4834, 76.4998, 76.5279, 76.535, 76.6007, 76.6382, 76.657, 76.657,
        76.6523, 76.6453] };
  WCL[1120] = struct {
   x = [-100.05, -100.069, -100.032],
   y = [78.7363, 78.7316, 78.7246] };
  WCL[1121] = struct {
   x = [-99.94, -100.024, -100.156, -100.259, -100.381, -100.56, -100.682, -100.691, -100.813, -100.869, -100.879, -100.879, -100.869, -100.973, -101.066, -101.076, -101.132, -101.198, -101.376, -101.479,
        -101.489, -101.555, -101.705, -101.827, -102.043, -102.268, -102.465, -102.644, -102.841, -102.878, -102.859, -103.01, -103.216, -103.432, -103.592, -103.77, -103.817, -103.723, -103.854,
        -104.07, -104.296, -104.446, -104.605, -104.718, -104.765, -104.831, -104.953, -104.962, -104.981, -104.99, -104.99, -104.896, -104.774, -104.568, -104.446, -104.23, -104.061, -103.901,
        -103.77, -103.592, -103.535, -103.432, -103.563, -103.817, -104.033, -103.995, -103.808, -103.62, -103.526, -103.507, -103.582, -103.667, -103.742, -103.854, -103.883, -103.901, -103.977,
        -104.07, -104.23, -104.268, -104.23, -104.192, -104.164, -104.127, -104.061, -104.023, -104.07, -104.155, -104.286, -104.38, -104.53, -104.634, -104.671, -104.737, -104.878, -105.009, -105.056,
        -105.103, -104.972, -104.943, -104.896, -104.84, -104.803, -104.793, -104.887, -105.028, -105.216, -105.432, -105.638, -105.657, -105.676, -105.704, -105.713, -105.741, -105.713, -105.647,
        -105.591, -105.544, -105.328, -105.047, -104.803, -104.624, -104.465, -104.221, -104.023, -103.817, -103.779, -103.854, -103.808, -103.629, -103.376, -103.31, -103.329, -103.338, -103.376,
        -103.385, -103.348, -103.263, -103.141, -103.085, -103.066, -103.066, -102.991, -102.831, -102.719, -102.531, -102.343, -102.24, -102.08, -101.921, -101.883, -101.893, -101.733, -101.695,
        -101.536, -101.508, -101.602, -101.611, -101.442, -101.451, -101.414, -101.282, -101.142, -101.001, -100.907, -100.785, -100.7, -100.616, -100.494, -100.419, -100.315, -100.25, -100.222,
        -100.297, -100.287, -100.222, -100.128, -100.071, -100.053],
   y = [77.8351, 77.7999, 77.8164, 77.8492, 77.8445, 77.8915, 77.9149, 77.9548, 78.0088, 78.0393, 78.0698, 78.105, 78.1261, 78.1543, 78.1848, 78.2224, 78.2317, 78.2083, 78.2036, 78.2106, 78.2411, 78.2716,
        78.2716, 78.2881, 78.2975, 78.2998, 78.2716, 78.2716, 78.3045, 78.3538, 78.3843, 78.3632, 78.3514, 78.3256, 78.328, 78.2951, 78.2787, 78.2763, 78.2505, 78.2458, 78.2599, 78.2857, 78.3068,
        78.328, 78.3397, 78.3632, 78.4054, 78.4265, 78.4617, 78.4969, 78.5251, 78.5509, 78.5627, 78.565, 78.5533, 78.5275, 78.5087, 78.5181, 78.5087, 78.5016, 78.5486, 78.5955, 78.6096, 78.6049,
        78.6166, 78.6495, 78.6565, 78.6847, 78.7222, 78.7645, 78.7574, 78.7457, 78.7574, 78.7387, 78.7809, 78.7833, 78.7856, 78.7434, 78.7574, 78.8161, 78.8302, 78.8325, 78.8325, 78.8466, 78.8701,
        78.9006, 78.9452, 78.971, 78.9757, 78.9475, 78.9358, 78.8889, 78.8466, 78.8161, 78.8067, 78.7926, 78.8067, 78.8302, 78.8513, 78.8748, 78.91, 78.9335, 78.9569, 79.0132, 79.0109, 79.0226,
        78.9992, 78.9851, 78.9898, 79.0015, 79.0414, 79.0766, 79.1189, 79.1587, 79.1869, 79.2385, 79.2808, 79.2949, 79.2737, 79.2878, 79.2972, 79.3043, 79.3277, 79.3488, 79.3371, 79.3113, 79.2949,
        79.2573, 79.2644, 79.2784, 79.2737, 79.2338, 79.1893, 79.1423, 79.0743, 79.0203, 79.0179, 78.9874, 78.9546, 78.9898, 79.0273, 79.0883, 79.1282, 79.1705, 79.1986, 79.1986, 79.1846, 79.1658,
        79.1142, 79.0907, 79.0602, 79.0414, 79.0602, 79.0954, 79.0602, 79.0015, 78.9381, 78.9123, 78.8936, 78.8701, 78.8607, 78.8983, 78.9147, 78.917, 78.9428, 78.9264, 78.91, 78.8936, 78.8936,
        78.8724, 78.8584, 78.8302, 78.802, 78.7574, 78.7387, 78.734, 78.7457, 78.7457, 78.7363] };
  WCL[1122] = struct {
   x = [-101.991, -101.982, -101.888, -101.766, -101.034, -100.921, -101.09, -101.306, -101.55, -101.719, -101.869, -102.113, -102.32, -102.498, -102.564, -102.611, -102.442, -102.263, -102.113, -101.991],
   y = [77.9478, 77.9501, 77.9548, 77.9666, 77.8422, 77.7929, 77.7905, 77.8023, 77.7905, 77.76, 77.7506, 77.7553, 77.7952, 77.8445, 77.868, 77.9126, 77.9314, 77.9501, 77.9501, 77.9478] };
  WCL[1123] = struct {
   x = [-105.023, -104.967, -104.92, -104.967, -104.939, -104.892, -104.77, -104.666, -104.666, -104.573, -104.479, -104.404, -104.357, -104.432, -104.582, -104.451, -104.291, -104.272, -104.272, -104.375,
        -104.498, -104.582, -104.789, -104.958, -105.136, -105.173, -105.258, -105.352, -105.464, -105.521, -105.605, -105.68, -105.652, -105.699, -105.755, -105.793, -105.84, -105.934, -106.009,
        -105.971, -105.896, -105.755, -105.577, -105.408, -105.324, -105.173, -105.108, -105.08, -105.023],
   y = [77.5723, 77.5535, 77.516, 77.462, 77.4409, 77.4291, 77.4268, 77.4174, 77.3939, 77.3658, 77.34, 77.3376, 77.3235, 77.286, 77.2813, 77.2719, 77.2461, 77.2203, 77.1991, 77.1569, 77.1452, 77.11, 77.1381,
        77.178, 77.171, 77.225, 77.2555, 77.3001, 77.3752, 77.3892, 77.4432, 77.4925, 77.516, 77.5441, 77.5864, 77.6216, 77.6685, 77.7131, 77.7319, 77.7647, 77.746, 77.7413, 77.7389, 77.7014, 77.6591,
        77.6333, 77.6098, 77.5746, 77.5723] };
  WCL[1124] = struct {
   x = [-104.009, -104, -103.944, -103.897, -103.803, -103.625, -103.587, -103.343, -103.277, -103.174, -103.08, -103.127, -103.202, -103.324, -103.456, -103.662, -103.831, -103.944, -104.131, -104.31,
        -104.282, -104.103, -104.122, -104.253, -104.375, -104.498, -104.441, -104.413, -104.413, -104.516, -104.573, -104.573, -104.62, -104.695, -104.695, -104.62, -104.544, -104.375, -104.216, -104,
        -103.953, -103.991, -104, -104.009],
   y = [76.603, 76.596, 76.5984, 76.6289, 76.5984, 76.5702, 76.5233, 76.4974, 76.4763, 76.4505, 76.4106, 76.3801, 76.3566, 76.3472, 76.3332, 76.3167, 76.3238, 76.3285, 76.3191, 76.3308, 76.3472, 76.3519,
        76.359, 76.3754, 76.3942, 76.42, 76.4458, 76.4998, 76.5279, 76.4951, 76.5045, 76.5373, 76.5397, 76.5561, 76.5866, 76.6124, 76.6289, 76.657, 76.6758, 76.6735, 76.6336, 76.6171, 76.6148, 76.603] };
  WCL[1125] = struct {
   x = [-101.022, -100.947, -100.881, -100.75, -100.675, -100.552, -100.675, -100.853, -101.088, -101.322, -101.491, -101.632, -101.688, -101.679, -101.463, -101.294, -101.163, -101.031, -101.022],
   y = [76.7392, 76.7415, 76.7485, 76.765, 76.7439, 76.7157, 76.6875, 76.6382, 76.603, 76.5631, 76.5631, 76.5678, 76.5819, 76.6148, 76.6429, 76.6805, 76.7251, 76.7368, 76.7392] };
  WCL[1126] = struct {
   x = [-103.028, -103, -103, -102.925, -102.766, -102.7, -102.578, -102.606, -102.681, -102.859, -103.113, -103.338, -103.517, -103.704, -103.93, -103.995, -103.995, -104.061, -104.146, -104.277, -104.418,
        -104.465, -104.512, -104.446, -104.418, -104.258, -104.023, -103.723, -103.554, -103.366, -103.226, -103.132, -103.028],
   y = [76.3238, 76.3214, 76.3238, 76.3238, 76.3238, 76.2839, 76.2487, 76.183, 76.1454, 76.1149, 76.0844, 76.0633, 76.0609, 76.0633, 76.0562, 76.0586, 76.082, 76.0867, 76.0586, 76.1008, 76.1032, 76.1407,
        76.1783, 76.2017, 76.2393, 76.2416, 76.2487, 76.2721, 76.298, 76.3073, 76.3214, 76.3402, 76.3238] };
  WCL[1127] = struct {
   x = [-103.05, -102.965, -102.918, -102.777, -102.674, -102.524, -102.392, -102.421, -102.552, -102.73, -103.003, -103.256, -103.5, -103.707, -103.876, -103.932, -103.932, -103.923, -103.754, -103.566,
        -103.359, -103.2, -103.05],
   y = [76.0703, 76.075, 76.075, 76.1008, 76.1055, 76.0914, 76.0797, 76.0328, 76.0023, 75.9835, 75.9624, 75.9365, 75.9225, 75.9084, 75.9037, 75.906, 75.9319, 75.9553, 75.9764, 76.0023, 76.0304, 76.0515,
        76.0703] };
  WCL[1128] = struct {
   x = [-103.061, -102.996, -102.958, -102.864, -102.733, -102.526, -102.423, -102.263, -102.198, -102.123, -102.235, -102.414, -102.526, -102.583, -102.639, -102.742, -102.855, -102.996, -103.136, -103.324,
        -103.427, -103.399, -103.277, -103.221, -103.146, -103.061],
   y = [75.9178, 75.9319, 75.9436, 75.9553, 75.9647, 75.9764, 75.9952, 76.0116, 75.9905, 75.9506, 75.9295, 75.8849, 75.8427, 75.8122, 75.7746, 75.784, 75.7887, 75.7652, 75.7488, 75.7535, 75.7746, 75.8051,
        75.845, 75.8802, 75.9084, 75.9178] };
  WCL[1129] = struct {
   x = [-104.087, -103.965, -103.965, -103.955, -103.932, -103.833, -103.777, -103.735, -103.636, -103.791, -103.988, -104.199, -104.472, -104.674, -104.861, -104.904, -104.772, -104.692, -104.641, -104.584,
        -104.42, -104.232, -104.087],
   y = [75.4461, 75.4273, 75.4226, 75.4202, 75.3803, 75.3475, 75.2841, 75.2395, 75.1738, 75.1034, 75.0729, 75.0494, 75.0753, 75.0846, 75.1269, 75.1809, 75.256, 75.3052, 75.3733, 75.4296, 75.4437, 75.4531,
        75.4461] };
  WCL[1130] = struct {
   x = [-110.043, -109.893, -109.818, -109.705, -109.705, -109.639, -109.564, -109.489, -109.452, -109.358, -109.198, -108.982, -108.795, -108.748, -108.635, -108.532, -108.541, -108.607, -108.729, -108.673,
        -108.56, -108.597, -108.569, -108.56, -108.513, -108.372, -108.278, -108.128, -108.109, -108.128, -108.137, -108.1, -108.241, -108.325, -108.41, -108.532, -108.579, -108.579, -108.466,
        -108.363, -108.25, -108.166, -108.044, -107.931, -107.79, -107.696, -107.602, -107.631, -107.696, -107.856, -107.931, -108.025, -107.987, -107.884, -107.8, -107.668, -107.48, -107.33, -107.152,
        -107.114, -107.086, -107.002, -106.927, -106.87, -106.814, -106.776, -106.682, -106.767, -106.795, -106.851, -106.739, -106.664, -106.476, -106.298, -106.269, -106.091, -105.866, -105.678,
        -105.565, -105.462, -105.481, -105.443, -105.439, -105.495, -105.673, -105.692, -105.739, -105.781, -105.87, -105.913, -105.978, -106.068, -106.124, -106.288, -106.42, -106.476, -106.565,
        -106.762, -106.997, -107.218, -107.335, -107.48, -107.645, -107.743, -107.724, -107.776, -107.889, -108.053, -108.119, -108.396, -108.419, -108.682, -108.757, -108.715, -108.583, -108.738,
        -108.738, -108.781, -108.879, -108.945, -109.09, -109.311, -109.419, -109.508, -109.729, -109.949],
   y = [76.5068, 76.5115, 76.5162, 76.5561, 76.6101, 76.6382, 76.6828, 76.7345, 76.772, 76.8166, 76.8377, 76.8307, 76.8542, 76.826, 76.7838, 76.7439, 76.7016, 76.6758, 76.65, 76.5984, 76.5631, 76.5326,
        76.4881, 76.4552, 76.4317, 76.427, 76.3895, 76.3472, 76.3097, 76.2839, 76.2628, 76.2416, 76.2041, 76.1642, 76.1196, 76.0985, 76.0985, 76.0938, 76.0914, 76.0727, 76.0797, 76.0774, 76.082,
        76.0938, 76.0703, 76.0328, 76.014, 75.9858, 75.9365, 75.9342, 75.892, 75.8474, 75.8051, 75.8427, 75.8966, 75.9154, 75.9319, 75.9201, 75.9178, 75.899, 75.8614, 75.7934, 75.7347, 75.7042,
        75.7699, 75.8122, 75.8145, 75.8403, 75.906, 75.9694, 76.0116, 76.0656, 76.0539, 76.0539, 76.0492, 76.0328, 75.9788, 75.9319, 75.8497, 75.7957, 75.7394, 75.7066, 75.6361, 75.5704, 75.5305,
        75.4766, 75.4062, 75.3311, 75.2865, 75.2255, 75.155, 75.1058, 75.0659, 75.0729, 75.0612, 75.0471, 75.0377, 74.9673, 74.9462, 74.9438, 74.9626, 74.972, 75.0095, 75.0447, 75.1081, 75.0987,
        75.0612, 75.0072, 74.972, 74.9673, 74.9673, 74.979, 74.9908, 74.9955, 75.0354, 75.094, 75.1339, 75.1222, 75.1128, 75.0635, 75.0095, 74.9908, 74.9579, 74.9251, 74.9063, 74.8875] };
  WCL[1131] = struct {
   x = [-110.064, -109.797, -109.468, -109.149, -108.98, -108.947, -108.914, -108.905, -108.914, -108.975, -109.041, -109.074, -109.093, -109.21, -109.266, -109.294, -109.374, -109.431, -109.341, -109.215,
        -109.201, -109.102, -109.121, -109.149, -109.205, -109.294, -109.313, -109.407, -109.449],
   y = [75.5775, 75.5704, 75.5564, 75.5352, 75.5493, 75.5704, 75.6221, 75.6714, 75.723, 75.7676, 75.791, 75.8286, 75.8497, 75.8403, 75.8638, 75.906, 75.8755, 75.9154, 75.9365, 75.9788, 76.0304, 76.0398,
        76.1008, 76.1618, 76.2064, 76.2229, 76.2487, 76.2534, 76.2628] };
  WCL[1132] = struct {
   x = [-110.076, -109.888, -109.832, -109.869, -109.963],
   y = [78.0933, 78.0745, 78.0299, 77.9642, 77.929] };
  WCL[1133] = struct {
   x = [-110.02, -110.001, -109.86, -109.738, -109.71, -109.56, -109.55, -109.513, -109.503, -109.503, -109.56, -109.7, -109.935],
   y = [78.6049, 78.5814, 78.5673, 78.5744, 78.5439, 78.511, 78.4782, 78.4594, 78.4312, 78.3773, 78.3186, 78.3022, 78.2904] };
  WCL[1134] = struct {
   x = [-120.017, -119.9, -119.853, -119.665, -119.477, -119.379, -119.355, -119.322, -119.29, -119.238, -119.224, -119.069, -118.848, -118.726, -118.576, -118.398, -118.234, -117.989, -117.727, -117.572,
        -117.492, -117.337, -117.206, -117.065, -116.943, -116.764, -116.6, -116.356, -116.168, -115.948, -115.849, -115.793, -115.741, -115.652, -115.521, -115.389, -115.389, -115.399, -115.53,
        -115.661, -115.76, -115.915, -116.06, -116.234, -116.412, -116.656, -116.755, -116.877, -116.952, -117.107, -117.271, -117.417, -117.558, -117.778, -117.989, -118.154, -118.332, -118.562,
        -118.863, -118.97, -119.027, -118.895, -118.994, -119.027, -119.224, -119.28, -119.158, -119.168, -119.158, -119.191, -119.337, -119.459, -119.557, -119.557, -119.698, -119.754, -119.797,
        -119.843, -119.853, -119.764, -119.689, -119.778, -119.909],
   y = [72.2309, 72.2567, 72.2778, 72.2943, 72.3271, 72.3506, 72.4069, 72.4679, 72.5149, 72.5641, 72.6369, 72.6909, 72.7402, 72.7589, 72.773, 72.8387, 72.8645, 72.9115, 72.9866, 73.0218, 73.0288, 73.0546,
        73.1016, 73.1344, 73.1626, 73.2072, 73.2494, 73.2799, 73.3034, 73.3316, 73.355, 73.3715, 73.3879, 73.402, 73.4348, 73.4724, 73.5217, 73.5803, 73.6061, 73.6554, 73.7024, 73.7211, 73.7704,
        73.8455, 73.8925, 73.9558, 74.0074, 74.045, 74.0732, 74.1295, 74.1576, 74.2187, 74.2562, 74.2773, 74.2844, 74.2985, 74.3055, 74.2773, 74.2398, 74.2116, 74.1717, 74.1224, 74.0802, 74.0309,
        73.9981, 74.0591, 74.099, 74.1342, 74.1764, 74.2163, 74.2187, 74.228, 74.1952, 74.1717, 74.1436, 74.0802, 74.0473, 74.0919, 74.153, 74.1858, 74.2187, 74.2421, 74.2656] };
  WCL[1135] = struct {
   x = [-117.37, -117.417, -117.253, -117.041, -116.577, -116.304, -115.995, -115.793, -115.474, -115.253, -115.089, -114.835, -114.624, -114.46, -114.239, -113.976, -113.634, -113.268, -113.014, -112.883,
        -112.742, -112.554, -112.409, -112.352, -112.418, -112.409, -112.254, -112.099, -112.132, -112.244, -112.366, -112.531, -112.742, -112.808, -113.014, -113.15, -113.291, -113.39, -113.554,
        -113.667, -113.798, -113.897, -113.944, -114.174, -114.352, -114.436, -114.549, -114.727, -114.859, -115, -115.201, -115.389, -115.474, -115.507, -115.652, -115.793, -115.938, -116.126,
        -116.201, -116.422, -116.61, -116.91, -117.116, -117.238, -117.384, -117.558, -117.591, -117.581, -117.661, -117.835, -117.891, -118.032, -118.112, -118.234, -118.309, -118.144, -117.957,
        -117.792, -117.558, -117.37, -117.182, -116.929, -116.675, -116.488, -116.29, -116.135, -116.159, -116.07, -115.816, -115.708, -115.774, -115.896, -116.084, -116.028, -115.905, -115.751,
        -115.676, -115.563, -115.539, -115.718, -115.905, -116.192, -116.323, -116.478, -116.656, -116.844, -117.065, -117.163, -117.318, -117.403, -117.426, -117.572, -117.647, -117.558, -117.572,
        -117.67, -117.759, -117.835, -117.914, -118.144, -118.252, -118.276, -118.219, -118.112, -117.933, -117.825, -117.835, -117.858, -117.858, -117.957, -118.112, -118.342, -118.454, -118.464,
        -118.562, -118.717, -118.83, -118.947, -118.97, -119.05, -119.069, -118.97, -118.773, -118.661, -118.618, -118.487, -118.276, -118.121, -118.112, -118.144, -118.201, -118.356, -118.431,
        -118.496, -118.506, -118.431, -118.234, -117.966, -117.736, -117.515, -117.295, -117.098, -116.844, -116.722, -116.502, -116.389, -116.117, -115.915, -115.793, -115.539, -115.131, -114.892,
        -114.727, -114.615, -114.483, -114.427, -114.319, -114.296, -114.174, -114.117, -114.108, -114.098, -114.108, -114.117, -114.206, -114.338, -114.427, -114.45, -114.436, -114.483, -114.591,
        -114.727, -114.746, -114.638, -114.516, -114.418, -114.319, -114.183, -114.066, -114.042, -114.019, -113.845, -113.69, -113.61, -113.709, -113.709, -113.601, -113.455, -113.437, -113.38,
        -113.347, -113.249, -113.071, -112.906, -112.718, -112.596, -112.456, -112.244, -112, -111.86, -111.681, -111.55, -111.503, -111.395, -111.339, -111.306, -111.418, -111.493, -111.639, -111.86,
        -111.991, -111.991, -111.869, -111.794, -111.723, -111.681, -111.559, -111.493, -111.395, -111.306, -111.315, -111.362, -111.418, -111.418, -111.329, -111.263, -111.198, -111.141, -111.118,
        -111.062, -111.052, -110.93, -110.808, -110.733, -110.677, -110.62, -110.602, -110.48, -110.414, -110.39, -110.367, -110.292, -110.226, -110.17, -110.226, -110.268, -110.217, -110.16, -110.071],
   y = [69.9521, 70.0249, 70.0483, 70.0929, 70.1446, 70.1821, 70.2267, 70.2455, 70.2666, 70.2854, 70.2901, 70.2971, 70.3276, 70.3464, 70.3534, 70.3112, 70.3018, 70.2901, 70.2713, 70.2478, 70.2478, 70.2525,
        70.2971, 70.3276, 70.3675, 70.398, 70.3605, 70.3745, 70.4074, 70.4661, 70.4919, 70.4989, 70.5154, 70.5576, 70.5553, 70.574, 70.5998, 70.6397, 70.6468, 70.6562, 70.6421, 70.6749, 70.6867,
        70.6679, 70.6538, 70.6538, 70.6538, 70.621, 70.6092, 70.6092, 70.5764, 70.5553, 70.5412, 70.5365, 70.5693, 70.574, 70.5693, 70.5998, 70.6421, 70.6233, 70.6092, 70.5951, 70.635, 70.6609,
        70.6538, 70.6397, 70.6749, 70.7008, 70.7148, 70.7336, 70.7618, 70.8157, 70.8697, 70.9284, 71.0223, 71.0786, 71.149, 71.1725, 71.203, 71.2077, 71.2171, 71.2593, 71.3086, 71.3438, 71.3579,
        71.379, 71.3508, 71.3414, 71.3508, 71.3766, 71.4001, 71.4142, 71.4142, 71.4353, 71.4682, 71.4823, 71.4823, 71.5081, 71.5362, 71.5362, 71.5268, 71.5034, 71.4869, 71.4846, 71.4752, 71.447,
        71.4424, 71.4142, 71.4142, 71.4283, 71.4564, 71.4635, 71.447, 71.433, 71.3978, 71.3978, 71.3931, 71.3978, 71.3696, 71.3626, 71.3907, 71.44, 71.4916, 71.5198, 71.5221, 71.5268, 71.5691, 71.616,
        71.6512, 71.6512, 71.6348, 71.5832, 71.5996, 71.6442, 71.6583, 71.6559, 71.6348, 71.6301, 71.6864, 71.7615, 71.8648, 71.961, 72.0408, 72.083, 72.1136, 72.1769, 72.2098, 72.2239, 72.2614,
        72.3036, 72.3342, 72.3318, 72.3342, 72.3717, 72.4656, 72.543, 72.6252, 72.7073, 72.7965, 72.8575, 72.8997, 72.942, 72.9936, 73.0405, 73.057, 73.1156, 73.1274, 73.1696, 73.1907, 73.233, 73.287,
        73.3269, 73.3527, 73.3527, 73.3363, 73.2987, 73.2799, 73.2424, 73.1907, 73.1203, 73.0664, 72.9748, 72.8857, 72.8505, 72.8387, 72.8129, 72.766, 72.7308, 72.7073, 72.6744, 72.658, 72.6416,
        72.6158, 72.5735, 72.6134, 72.6205, 72.658, 72.6815, 72.6815, 72.6486, 72.6252, 72.6486, 72.6744, 72.7472, 72.7707, 72.834, 72.9185, 72.9514, 72.9866, 73.0053, 73.0124, 73.0218, 73.0218,
        72.9889, 72.9537, 72.9279, 72.9185, 72.8833, 72.8716, 72.8129, 72.7918, 72.7495, 72.7472, 72.6956, 72.6697, 72.5806, 72.5266, 72.4867, 72.4304, 72.3647, 72.3342, 72.3154, 72.3248, 72.3318,
        72.3342, 72.3412, 72.3881, 72.4233, 72.4515, 72.4585, 72.421, 72.3952, 72.3553, 72.3459, 72.3388, 72.3342, 72.3224, 72.3482, 72.3905, 72.4515, 72.4914, 72.5031, 72.5641, 72.5688, 72.5641,
        72.5594, 72.5501, 72.5665, 72.5548, 72.5407, 72.5266, 72.5031, 72.4773, 72.5031, 72.5594, 72.5759, 72.5829, 72.5594, 72.5407] };
  WCL[1136] = struct {
   x = [-109.982, -110.015, -110.146, -110.236, -110.25, -110.217, -110.114, -110.006],
   y = [72.5641, 72.6158, 72.6416, 72.6651, 72.6909, 72.705, 72.6815, 72.6533] };
  WCL[1137] = struct {
   x = [-109.916, -110.015, -110.071, -110.146, -110.179, -110.325, -110.447, -110.522, -110.536, -110.39, -110.236, -110.071],
   y = [72.6956, 72.7331, 72.7472, 72.7754, 72.7988, 72.8176, 72.8763, 72.9443, 72.9889, 72.996, 72.9795, 72.9748] };
  WCL[1138] = struct {
   x = [-110.052, -110.043, -110.043],
   y = [76.5021, 76.5186, 76.5068] };
  WCL[1139] = struct {
   x = [-109.954, -110.165, -110.353, -110.428, -110.541, -110.681, -110.649, -110.714, -110.827, -110.836, -110.85, -110.869, -110.991, -111.188, -111.357, -111.376, -111.653, -111.784, -112.061, -112.273,
        -112.493, -112.657, -112.911, -113.155, -113.507, -113.704, -113.817, -114.023, -114.202, -114.343, -114.399, -114.488, -114.488, -114.399, -114.235, -114.089, -113.85, -113.629, -113.54,
        -113.427, -113.296, -113.296, -113.089, -112.822, -112.493, -112.183, -111.972, -111.916, -111.742, -111.653, -111.611, -111.432, -111.324, -111.09, -111.015, -111.024, -111.221, -111.456,
        -111.676, -111.831, -112.029, -112.357, -112.437, -112.381, -112.503, -112.634, -112.657, -112.747, -112.78, -112.92, -113.141, -113.207, -113.362, -113.629, -113.934, -114.047, -114.037,
        -113.934, -113.784, -113.859, -113.892, -113.868, -113.596, -113.441, -113.563, -113.728, -113.915, -114.07, -114.155, -114.225, -114.31, -114.446, -114.596, -114.61, -114.465, -114.479,
        -114.587, -114.61, -114.732, -114.873, -115.108, -115.206, -115.206, -115.305, -115.446, -115.633, -115.614, -115.624, -115.854, -116.074, -116.253, -116.164, -116.309, -116.295, -116.539,
        -116.647, -116.891, -117.004, -117.243, -117.52, -117.708, -117.675, -117.609, -117.497, -117.29, -117.046, -116.793, -116.638, -116.384, -116.121, -116.009, -115.802, -115.46, -115.338,
        -115.248, -115.206, -115.535, -115.92, -116.173, -116.337, -116.647, -117.013, -117.342, -117.342, -117.238, -117.093, -116.915, -116.844, -116.722, -116.605, -116.464, -116.356, -116.304,
        -116.239, -116.112, -116.103, -116.215, -116.319, -116.426, -116.539, -116.661, -116.797, -116.947, -117.037, -117.121, -117.112, -117.065, -116.994, -117.018, -117.084, -116.985, -116.844,
        -116.703, -116.605, -116.52, -116.464, -116.347, -116.258, -116.121, -116.079, -116.164, -116.239, -116.37, -116.506, -116.633, -116.699, -116.685, -116.624, -116.506, -116.384, -116.286,
        -116.206, -116.135, -116.028, -115.905, -115.807, -115.812, -115.807, -115.844, -115.821, -115.798, -115.685, -115.549, -115.469, -115.417, -115.333, -115.197, -115.187, -115.089, -114.99,
        -114.962, -114.925, -114.887, -114.793, -114.695, -114.619, -114.685, -114.798, -114.812, -114.713, -114.634, -114.488, -114.488, -114.45, -114.427, -114.488, -114.441, -114.399, -114.422,
        -114.436, -114.314, -114.15, -114.061, -114.019, -113.929, -113.798, -113.629, -113.498, -113.399, -113.047],
   y = [74.8875, 74.8875, 74.85, 74.8406, 74.8382, 74.8148, 74.7725, 74.7279, 74.7115, 74.7045, 74.7045, 74.6857, 74.6669, 74.6341, 74.6247, 74.5988, 74.5895, 74.5425, 74.5003, 74.4721, 74.458, 74.4463,
        74.4533, 74.458, 74.4627, 74.4885, 74.5214, 74.5472, 74.5942, 74.6341, 74.6505, 74.6904, 74.7162, 74.7749, 74.8218, 74.8265, 74.8711, 74.8805, 74.8593, 74.864, 74.8899, 74.9415, 75.0189,
        75.0236, 75.0424, 75.0494, 75.0447, 74.9931, 74.9931, 75.0447, 75.0706, 75.1034, 75.1644, 75.2161, 75.263, 75.3052, 75.3052, 75.202, 75.2161, 75.1996, 75.1832, 75.1644, 75.2184, 75.27, 75.2442,
        75.2184, 75.2794, 75.2935, 75.1996, 75.1339, 75.1269, 75.1809, 75.1152, 75.0987, 75.0494, 75.1034, 75.1339, 75.1738, 75.2114, 75.2161, 75.2841, 75.3663, 75.4109, 75.4648, 75.4578, 75.4179,
        75.4343, 75.439, 75.3358, 75.256, 75.2419, 75.3217, 75.3358, 75.256, 75.202, 75.1152, 75.1011, 75.087, 75.0588, 75.0095, 75.0189, 75.1034, 75.1856, 75.1762, 75.141, 75.1457, 75.0588, 74.9931,
        75.0049, 75.0447, 75.0893, 75.141, 75.1691, 75.2301, 75.2067, 75.1433, 75.1597, 75.1715, 75.1809, 75.2184, 75.2771, 75.317, 75.3475, 75.4343, 75.5024, 75.5141, 75.5188, 75.5071, 75.5071,
        75.5047, 75.5212, 75.5892, 75.6338, 75.6549, 75.6995, 75.7347, 75.6714, 75.6549, 75.6009, 75.5869, 75.5657, 75.5728, 75.5798, 75.6854, 75.7957, 75.8075, 75.8122, 75.8145, 75.8403, 75.8614,
        75.8661, 75.8708, 75.8755, 75.8802, 75.9389, 75.9412, 75.9225, 75.9013, 75.899, 75.8966, 75.8873, 75.8708, 75.8708, 75.8591, 75.892, 75.9506, 75.9999, 76.0116, 76.0492, 76.0867, 76.1806,
        76.2135, 76.2041, 76.2017, 76.2088, 76.2111, 76.19, 76.1877, 76.1712, 76.1947, 76.1994, 76.2416, 76.2557, 76.2581, 76.2628, 76.3073, 76.3449, 76.42, 76.481, 76.481, 76.4834, 76.5233, 76.5139,
        76.5256, 76.5021, 76.4458, 76.3989, 76.359, 76.3308, 76.2768, 76.2299, 76.2182, 76.2299, 76.251, 76.2792, 76.2721, 76.2815, 76.2252, 76.2158, 76.2041, 76.1478, 76.0867, 76.0633, 75.9999,
        75.9999, 75.9553, 75.9248, 75.8943, 75.845, 75.8333, 75.8521, 75.8755, 75.8568, 75.8239, 75.7723, 75.7418, 75.7465, 75.6784, 75.6291, 75.6103, 75.5657, 75.554, 75.5634, 75.5611, 75.5869,
        75.5775, 75.5986, 75.5986, 75.5634, 75.5775] };
  WCL[1140] = struct {
   x = [-109.987, -110.193, -110.372, -110.4, -110.4, -110.484, -110.381, -110.193, -110.09, -110.052],
   y = [76.2628, 76.3167, 76.3472, 76.3895, 76.4294, 76.4599, 76.4599, 76.4763, 76.4951, 76.5021] };
  WCL[1141] = struct {
   x = [-118.06, -118.201, -118.163, -118.041, -117.938, -117.75, -117.581, -117.591, -117.703, -117.844, -117.872, -117.966, -118.032, -118.135, -118.219, -118.478, -118.764, -118.881, -119.083, -119.313,
        -119.473, -119.407, -119.144, -119.031, -118.834, -118.628, -118.365, -118.29, -118.14],
   y = [76.0492, 76.014, 76.0351, 76.0492, 76.1008, 76.1243, 76.1126, 76.0445, 75.9741, 75.8708, 75.8403, 75.8051, 75.7206, 75.6714, 75.6127, 75.5235, 75.5141, 75.5634, 75.5845, 75.6174, 75.6901, 75.7324,
        75.7887, 75.8333, 75.8966, 75.9624, 75.9905, 76.0328, 76.0492] };
  WCL[1142] = struct {
   x = [-120.045, -119.904, -119.829, -119.689, -119.557, -119.52, -119.407, -119.341, -119.322, -119.21, -119.078, -118.9, -118.853, -118.703, -118.581, -118.515, -118.262, -118.187, -118.027, -118.065,
        -118.008, -117.914, -117.961, -117.896, -117.727, -117.492, -117.342, -117.389, -117.285, -117.098, -116.901, -116.929, -117.051, -117.21, -117.323, -117.267, -117.135, -117.032, -116.891,
        -116.675, -116.581, -116.459, -116.337, -116.272, -116.262, -116.028, -115.962, -115.802, -115.643, -115.633, -115.68, -115.755, -115.915, -115.981, -116.168, -116.262, -116.431, -116.497,
        -116.497, -116.553, -116.572, -116.497, -116.422, -116.253, -116.131, -116.065, -116.056, -116.197, -116.356, -116.497, -116.45, -116.3, -116.243, -116.187, -116.093, -116.243, -116.356,
        -116.534, -116.816, -117.013, -117.126, -117.182, -117.163, -117.135, -117.21, -117.351, -117.454, -117.68, -117.774, -117.933, -118.083, -118.093, -118.027, -118.036, -117.896, -117.877,
        -118.008, -118.158, -118.449, -118.468, -118.459, -118.6, -118.478, -118.468, -118.459, -118.647, -118.825, -118.975, -118.975, -119.069, -118.966, -118.853, -118.778, -118.731, -118.665,
        -118.722, -118.975, -119.013, -119.078, -119.116, -119.238, -119.351, -119.491, -119.548, -119.642, -119.773, -119.782, -119.745, -119.67, -119.689, -119.773, -119.642, -119.792, -119.961,
        -119.764, -119.642, -119.613, -119.764, -119.811, -119.97],
   y = [77.0419, 77.0654, 77.1123, 77.1452, 77.1522, 77.1921, 77.2297, 77.2484, 77.3047, 77.3141, 77.3235, 77.3235, 77.3587, 77.354, 77.3212, 77.3446, 77.347, 77.3681, 77.3728, 77.3118, 77.2461, 77.2625,
        77.3188, 77.3587, 77.3165, 77.286, 77.286, 77.3212, 77.3212, 77.3071, 77.3564, 77.3775, 77.3728, 77.401, 77.4549, 77.4503, 77.469, 77.516, 77.5254, 77.5488, 77.5277, 77.5042, 77.4878, 77.4737,
        77.4503, 77.4197, 77.4057, 77.3564, 77.3235, 77.2836, 77.2602, 77.2555, 77.2437, 77.2085, 77.1968, 77.1874, 77.1898, 77.1757, 77.1569, 77.1499, 77.1193, 77.0724, 77.0466, 76.9973, 76.9738,
        76.9504, 76.9011, 76.9034, 76.9386, 76.9058, 76.8495, 76.8424, 76.7955, 76.7791, 76.6969, 76.6476, 76.6242, 76.5819, 76.5725, 76.5796, 76.5467, 76.4904, 76.427, 76.359, 76.3003, 76.2886,
        76.2721, 76.2815, 76.3214, 76.3777, 76.4458, 76.5162, 76.5843, 76.6547, 76.7556, 76.8096, 76.8119, 76.772, 76.7814, 76.7368, 76.711, 76.6805, 76.6547, 76.6007, 76.5467, 76.5045, 76.5608,
        76.5678, 76.5397, 76.4998, 76.4693, 76.4622, 76.4528, 76.4036, 76.3496, 76.3027, 76.2628, 76.1736, 76.1266, 76.1384, 76.1219, 76.2064, 76.2768, 76.3684, 76.3332, 76.3566, 76.305, 76.2369,
        76.2088, 76.1877, 76.1642, 76.1501, 76.1384, 76.1079, 76.0797, 76.0116, 75.9788, 75.9811, 75.9037, 75.8802] };
  WCL[1143] = struct {
   x = [-114.155, -114.08, -113.976, -113.864, -113.864, -113.77, -113.648, -113.657, -113.714, -113.845, -114.005, -114.155, -114.277, -114.474, -114.587, -114.756, -114.971, -114.906, -114.521, -114.314,
        -114.155],
   y = [76.8917, 76.8941, 76.887, 76.8659, 76.8659, 76.833, 76.8072, 76.7767, 76.7298, 76.7298, 76.7251, 76.7157, 76.7298, 76.7415, 76.7392, 76.7485, 76.8119, 76.8448, 76.8706, 76.8847, 76.8917] };
  WCL[1144] = struct {
   x = [-112.094, -112.066, -111.972, -111.972, -111.775, -111.606, -111.503, -111.334, -111.193, -111.155, -110.996, -110.93, -110.836, -110.62, -110.348, -110.245, -110.076],
   y = [78.0346, 78.0299, 78.0182, 78.0323, 78.0534, 78.0628, 78.0745, 78.0769, 78.0816, 78.0534, 78.044, 78.0698, 78.0909, 78.1003, 78.105, 78.0956, 78.0933] };
  WCL[1145] = struct {
   x = [-109.966, -110.125, -110.247, -110.341, -110.51, -110.623, -110.782, -111.026, -111.064, -111.148, -111.195, -111.148, -111.064, -111.139, -111.327, -111.477, -111.599, -111.543, -111.543, -111.543,
        -111.411, -111.383, -111.477, -111.402, -111.261, -111.148, -111.139, -111.064, -111.036, -110.951, -110.876, -110.688, -110.51, -110.341, -110.313, -110.304, -110.266, -110.304, -110.304,
        -110.332, -110.435, -110.557, -110.726, -110.904, -111.045, -111.205, -111.355, -111.515, -111.674, -111.899, -112.097, -112.331, -112.557, -112.669, -112.688, -112.81, -113.045, -113.242,
        -113.373, -113.383, -113.336, -113.383, -113.383, -113.439, -113.495, -113.477, -113.43, -113.364, -113.251, -113.063, -112.913, -112.829, -112.669, -112.5, -112.463, -112.359, -112.275,
        -112.2, -112.144, -112.097],
   y = [77.929, 77.9196, 77.9196, 77.9032, 77.9008, 77.8868, 77.8774, 77.8774, 77.8563, 77.8516, 77.8398, 77.8257, 77.8164, 77.8023, 77.8023, 77.8023, 77.7952, 77.7835, 77.7506, 77.7366, 77.7718, 77.7436,
        77.7084, 77.7201, 77.7577, 77.7577, 77.7084, 77.7248, 77.7765, 77.7952, 77.7671, 77.753, 77.7741, 77.7741, 77.7366, 77.6967, 77.6638, 77.6169, 77.577, 77.5254, 77.4972, 77.4737, 77.4596,
        77.4456, 77.4338, 77.4244, 77.4409, 77.4221, 77.3939, 77.3705, 77.347, 77.347, 77.3681, 77.415, 77.4526, 77.4667, 77.4737, 77.4948, 77.5347, 77.5864, 77.6122, 77.6544, 77.7154, 77.7413,
        77.7905, 77.8234, 77.8727, 77.8915, 77.8868, 77.9173, 77.9173, 77.9548, 77.9783, 77.9666, 77.9947, 78.0182, 78.0252, 78.0158, 78.0252, 78.0346] };
  WCL[1146] = struct {
   x = [-115.07, -115.042, -115.004, -114.957, -114.873, -114.882, -114.817, -114.657, -114.544, -114.488, -114.507, -114.479, -114.45, -114.291, -114.188, -114.103, -113.962, -113.84, -113.803, -113.756,
        -113.878, -114.047, -114.197, -114.357, -114.479, -114.629, -114.751, -114.863, -114.986, -115.032, -115.061, -115.183, -115.192, -115.126, -115.07],
   y = [77.9407, 77.9431, 77.9525, 77.9548, 77.9384, 77.9759, 78.0065, 78.0323, 78.0417, 78.0205, 77.9971, 77.9806, 77.9572, 77.9384, 77.9102, 77.8844, 77.868, 77.8398, 77.814, 77.7858, 77.7389, 77.7178,
        77.7037, 77.6943, 77.7225, 77.7553, 77.8023, 77.8469, 77.8445, 77.8727, 77.9055, 77.9314, 77.936, 77.9431, 77.9407] };
  WCL[1147] = struct {
   x = [-110.13, -110.092, -110.074, -110.017],
   y = [78.6307, 78.6354, 78.6354, 78.6049] };
  WCL[1148] = struct {
   x = [-109.933, -110.074, -110.233, -110.412, -110.59, -110.787, -111.012, -111.125, -111.294, -111.397, -111.5, -111.547, -111.538, -111.463, -111.585, -111.538, -111.669, -111.791, -111.914, -112.007,
        -112.082, -112.233, -112.392, -112.458, -112.599, -112.571, -112.58, -112.646, -112.711, -112.824, -112.927, -112.909, -112.937, -112.974, -113.059, -113.218, -113.35, -113.378, -113.491,
        -113.434, -113.35, -113.265, -113.078, -112.993, -112.852, -112.843, -112.74, -112.683, -112.533, -112.42, -112.28, -112.101, -111.951, -111.754, -111.594, -111.482, -111.472, -111.369,
        -111.294, -111.209, -111.059, -110.937, -110.834, -110.703, -110.599, -110.515, -110.505, -110.43, -110.383, -110.252, -110.13, -110.13, -110.13],
   y = [78.2904, 78.3092, 78.3022, 78.2857, 78.2716, 78.2834, 78.3256, 78.3632, 78.3726, 78.4054, 78.396, 78.4195, 78.3655, 78.3209, 78.3139, 78.2716, 78.2716, 78.267, 78.2599, 78.2951, 78.3421, 78.3467,
        78.3561, 78.3866, 78.3819, 78.3585, 78.3467, 78.3538, 78.3561, 78.3421, 78.3303, 78.2951, 78.267, 78.2881, 78.2834, 78.2623, 78.2716, 78.3115, 78.3115, 78.3514, 78.4054, 78.4289, 78.4617,
        78.443, 78.45, 78.4711, 78.4782, 78.4876, 78.5228, 78.5298, 78.5439, 78.5415, 78.5486, 78.5697, 78.5838, 78.6307, 78.6401, 78.68, 78.6542, 78.68, 78.6988, 78.7363, 78.7316, 78.7316, 78.741,
        78.7269, 78.7035, 78.6988, 78.6706, 78.673, 78.6542, 78.6378, 78.6307] };
  WCL[1149] = struct {
   x = [-130.141, -129.977, -129.879, -129.855, -129.714, -129.667, -129.559, -129.494, -129.559, -129.616],
   y = [70.0742, 70.1, 70.1633, 70.2079, 70.2408, 70.1938, 70.1469, 70.0648, 70.0507, 70.0202] };
  WCL[1150] = struct {
   x = [-128.421, -128.332, -128.388, -128.299, -128.111, -127.956, -127.816, -127.703, -127.834, -127.98, -128.102, -128.069, -128.055, -128.102, -128.276, -128.266, -128.177, -128.088, -127.881, -127.68,
        -127.515, -127.342, -127.22, -127.187, -127.163],
   y = [69.9498, 70.0319, 70.0812, 70.1187, 70.1469, 70.168, 70.1845, 70.2267, 70.2384, 70.2572, 70.2971, 70.3276, 70.3816, 70.4191, 70.398, 70.4473, 70.5083, 70.5412, 70.5036, 70.4191, 70.3558, 70.2924,
        70.2079, 70.1187, 70.0366] };
  WCL[1151] = struct {
   x = [-124.859, -124.915],
   y = [69.9897, 70.0014] };
  WCL[1152] = struct {
   x = [-125.19, -125.204, -125.147, -125.091, -125.025, -124.847, -124.762],
   y = [69.9427, 70.0014, 70.0507, 70.0483, 70.0272, 70.0483, 70.0272] };
  WCL[1153] = struct {
   x = [-124.704, -124.624, -124.493, -124.507, -124.647, -124.704, -124.737, -124.591, -124.474, -124.46, -124.408, -124.427],
   y = [69.9615, 70.0178, 70.0249, 70.0648, 70.0765, 70.0953, 70.114, 70.1211, 70.1258, 70.0953, 70.0554, 70.0014] };
  WCL[1154] = struct {
   x = [-124.286, -124.15, -124.117, -124.009, -123.92, -124.009, -124.075, -124.15, -124.22, -124.286, -124.45, -124.45, -124.507, -124.591, -124.662, -124.662, -124.746, -124.812, -124.769, -124.826,
        -124.892, -124.99, -124.948, -124.835, -124.779, -124.662, -124.647, -124.671, -124.793, -124.934, -125.056, -125.201, -125.22, -125.187, -125.145, -125.145, -125.112, -125.122, -125.187,
        -125.089, -125.065, -125.187, -125.3, -125.168, -125.253, -125.389, -125.431, -125.497, -125.506, -125.506, -125.619, -125.586, -125.506, -125.441, -125.577, -125.628, -125.661, -125.765,
        -125.685, -125.661, -125.741, -125.807, -125.783, -125.708, -125.75, -125.84, -125.985, -125.929, -125.765, -125.596, -125.431, -125.253, -125.103, -125.056, -125.22, -125.22, -125.065,
        -124.892, -124.793, -124.779, -124.769, -124.624, -124.624, -124.615, -124.516, -124.361, -124.173, -124.042, -123.929, -123.854, -123.732, -123.643, -123.559, -123.469, -123.347, -123.235,
        -123.014, -122.84, -122.695, -122.498, -122.441, -122.254, -122.066, -121.911, -121.845, -121.803, -121.648, -121.615, -121.648, -121.484, -121.437, -121.273, -121.198, -121.174, -121.085,
        -121.01, -120.897, -120.691, -120.555, -120.545, -120.503, -120.545, -120.545, -120.536, -120.578, -120.512, -120.372, -120.315, -120.268, -120.259, -120.217, -120.137, -120.015],
   y = [74.0168, 73.9535, 73.9535, 73.9159, 73.8643, 73.822, 73.7939, 73.7141, 73.646, 73.6038, 73.5569, 73.5287, 73.4653, 73.4442, 73.3738, 73.3269, 73.2823, 73.2635, 73.2189, 73.2048, 73.1555, 73.1344,
        73.0804, 73.0851, 73.0452, 73.0218, 72.9936, 72.9608, 72.942, 72.9349, 72.9091, 72.9279, 72.8927, 72.8481, 72.8293, 72.766, 72.7566, 72.7167, 72.6392, 72.6557, 72.5993, 72.5829, 72.536,
        72.5031, 72.482, 72.5008, 72.4609, 72.4304, 72.4022, 72.3623, 72.3388, 72.3036, 72.2919, 72.2637, 72.2614, 72.252, 72.245, 72.2309, 72.2168, 72.1746, 72.1816, 72.137, 72.0924, 72.0783, 72.0525,
        71.9845, 71.9751, 71.9258, 71.954, 71.9634, 71.9774, 71.9704, 71.9563, 71.9258, 71.9446, 71.9164, 71.8883, 71.8484, 71.8108, 71.8507, 71.8742, 71.8624, 71.8272, 71.7991, 71.792, 71.7733,
        71.7451, 71.6958, 71.6348, 71.5433, 71.4869, 71.386, 71.2945, 71.2147, 71.142, 71.0974, 71.1115, 71.1772, 71.2382, 71.257, 71.264, 71.3227, 71.386, 71.4494, 71.5034, 71.5198, 71.4776, 71.4494,
        71.4072, 71.4001, 71.3978, 71.4072, 71.4564, 71.5081, 71.4611, 71.4564, 71.4752, 71.5433, 71.6278, 71.6958, 71.7381, 71.7662, 71.8272, 71.9164, 71.968, 72.0173, 72.0854, 72.1229, 72.1933,
        72.2637, 72.2778, 72.2544, 72.2309] };
  WCL[1155] = struct {
   x = [-119.907, -120.104, -120.259, -120.423, -120.545, -120.677, -120.813, -120.911, -120.954, -121.076, -121.184, -121.329, -121.526, -121.658, -121.77, -121.836, -121.892, -121.991, -122.099, -122.254,
        -122.39, -122.498, -122.643, -122.751, -122.883, -123.028, -123.117, -123.136, -123.117, -123.094, -123.094, -123.117, -123.16, -123.235, -123.258, -123.324, -123.413, -123.469, -123.512,
        -123.666, -123.789, -123.789, -123.92, -124.075, -124.187, -124.187, -124.263, -124.46, -124.704, -124.615, -124.582, -124.525, -124.427, -124.286, -124.45, -124.46, -124.408, -124.352,
        -124.286],
   y = [74.2656, 74.2844, 74.3102, 74.343, 74.3947, 74.4322, 74.4627, 74.4792, 74.5331, 74.5519, 74.5566, 74.5707, 74.5918, 74.5848, 74.5801, 74.5472, 74.5284, 74.5472, 74.5496, 74.5214, 74.5144, 74.5214,
        74.5097, 74.5097, 74.5097, 74.5097, 74.505, 74.4815, 74.4698, 74.4228, 74.3736, 74.39, 74.4228, 74.4416, 74.4909, 74.5026, 74.4533, 74.4627, 74.4932, 74.4909, 74.4792, 74.4627, 74.4698,
        74.4792, 74.4745, 74.4416, 74.4369, 74.4322, 74.4017, 74.39, 74.343, 74.2891, 74.2609, 74.2374, 74.2116, 74.1553, 74.1013, 74.0567, 74.0168] };
  WCL[1156] = struct {
   x = [-120.132, -120.095, -120.085, -120.048],
   y = [77.0114, 76.9997, 77.0184, 77.0419] };
  WCL[1157] = struct {
   x = [-119.968, -120.118, -120.165, -120.278, -120.306, -120.381, -120.522, -120.54, -120.54, -120.681, -120.841, -120.775, -120.794, -120.841, -121.047, -121.094, -121.094, -121.132, -121.263, -121.423,
        -121.629, -121.789, -121.883, -121.911, -121.864, -121.864, -121.873, -121.996, -122.118, -122.249, -122.39, -122.399, -122.54, -122.7, -122.69, -122.568, -122.484, -122.624, -122.887,
        -122.934, -122.915, -122.831, -122.662, -122.634, -122.427, -122.38, -122.221, -122.061, -121.845, -121.761, -121.639, -121.601, -121.489, -121.423, -121.291, -121.263, -121.282, -121.226,
        -121.104, -120.991, -120.916, -120.85, -120.719, -120.606, -120.475, -120.24, -120.409, -120.296, -120.24, -120.212, -120.127],
   y = [75.8802, 75.8849, 75.9131, 75.9412, 75.8708, 75.8239, 75.9037, 75.9624, 76.0163, 76.0187, 76.0914, 76.1431, 76.19, 76.2276, 76.1853, 76.1008, 76.0539, 76.0163, 75.9506, 76.0046, 76.0257, 76.0328,
        76.0539, 76.082, 76.1079, 76.1407, 76.1665, 76.0633, 76.0422, 75.9717, 75.9506, 75.9459, 75.9459, 75.9929, 76.0163, 76.1055, 76.1689, 76.1877, 76.1126, 76.1947, 76.2158, 76.2393, 76.2792,
        76.3754, 76.3989, 76.427, 76.4552, 76.4552, 76.4247, 76.4458, 76.4317, 76.4599, 76.5326, 76.5678, 76.603, 76.6336, 76.7016, 76.6711, 76.6781, 76.7016, 76.711, 76.7439, 76.7462, 76.7767, 76.826,
        76.8471, 76.8589, 76.9269, 76.948, 76.9879, 77.0114] };
  WCL[1158] = struct {
   x = [-131.024, -130.991, -130.916, -130.752, -130.672, -130.564, -130.451, -130.362, -130.221, -130.141],
   y = [69.9709, 70.0319, 70.0812, 70.1187, 70.1211, 70.1586, 70.1399, 70.107, 70.1023, 70.0742] };
  WCL[1159] = struct {
   x = [-150.158, -149.871, -149.716, -149.571, -149.317, -149.111, -148.923, -148.67, -148.58, -148.557, -148.383, -148.247, -148.106, -147.952, -147.853, -147.797, -147.731, -147.576, -147.355, -147.158,
        -147.003, -146.839, -146.651, -146.431, -146.196, -146.041, -145.844, -145.67, -145.567, -145.417, -145.295, -145.328],
   y = [70.4238, 70.4473, 70.4895, 70.4966, 70.4919, 70.4637, 70.4262, 70.4191, 70.4051, 70.3464, 70.3487, 70.3417, 70.3159, 70.3088, 70.2666, 70.2525, 70.2572, 70.2408, 70.2032, 70.1821, 70.1446, 70.1586,
        70.1751, 70.1774, 70.1774, 70.1704, 70.1305, 70.0624, 70.0507, 70.0507, 70.0483, 70.0084] };
  WCL[1160] = struct {
   x = [-144.952, -144.919],
   y = [69.9826, 70.0249] };
  WCL[1161] = struct {
   x = [-144.612, -144.49, -144.391, -144.382, -144.368, -144.359, -144.251, -144.072, -143.875, -143.786, -143.673, -143.542, -143.486, -143.509, -143.42, -143.312, -143.232, -143.044, -142.871, -142.758,
        -142.561],
   y = [69.9826, 70.0202, 70.0319, 70.0507, 70.0554, 70.0507, 70.0554, 70.0695, 70.0812, 70.1023, 70.0812, 70.0131, 70.0507, 70.1187, 70.1, 70.1023, 70.1258, 70.1023, 70.0648, 70.0436, 70.0061] };
  WCL[1162] = struct {
   x = [-160.061, -159.972, -159.883, -159.873, -159.93, -159.873, -159.728, -159.564, -159.489, -159.662, -159.808, -159.897],
   y = [70.2901, 70.2478, 70.2924, 70.3675, 70.4144, 70.4778, 70.4731, 70.4895, 70.5107, 70.4966, 70.5247, 70.6022] };
  WCL[1163] = struct {
   x = [-160.106, -159.965, -159.965, -159.885, -159.655, -159.411, -159.181, -159.026, -159.092, -159.336, -159.359, -159.378, -159.444, -159.533, -159.411, -159.223, -159.27, -159.214, -159.101, -159.181,
        -158.993, -158.815, -158.529, -158.341, -158.374, -158.186, -157.965, -157.834, -157.679, -157.581, -157.482, -157.44, -157.36, -157.294, -157.252, -157.219, -157.163, -157.05, -156.984,
        -157.031, -156.975, -156.966, -156.952, -156.9, -156.844, -156.811, -156.712, -156.679, -156.647, -156.557, -156.412, -156.28, -156.126, -156.102, -156.083, -155.896, -155.731, -155.651,
        -155.576, -155.553, -155.576, -155.651, -155.764, -155.881, -156.027, -156.102, -156.215, -156.257, -156.346, -156.304, -156.158, -156.05, -155.947, -155.816, -155.774, -155.684, -155.586,
        -155.473, -155.454, -155.276, -155.309, -155.299, -155.464, -155.431, -155.309, -155.177, -155.201, -155.069, -155.079, -154.99, -154.76, -154.638, -154.558, -154.647, -154.516, -154.328,
        -154.22, -154.074, -153.952, -153.821, -153.746, -153.6, -153.6, -153.568, -153.413, -153.234, -153.07, -152.962, -152.915, -152.906, -152.774, -152.718, -152.76, -152.662, -152.474, -152.277,
        -152.136, -152.145, -152.267, -152.22, -152.056, -151.958, -152.136, -152.357, -152.235, -151.99, -151.779, -151.559, -151.582, -151.737, -151.582, -151.197, -151.066, -150.855, -150.7,
        -150.512, -150.334, -150.226, -150.183, -150.324, -150.357, -150.16],
   y = [70.6491, 70.6679, 70.6937, 70.7453, 70.8134, 70.8603, 70.8932, 70.9073, 70.8744, 70.8416, 70.8204, 70.7923, 70.7594, 70.7336, 70.7125, 70.6679, 70.7407, 70.7453, 70.7735, 70.8017, 70.7923, 70.7735,
        70.7876, 70.8017, 70.8322, 70.8345, 70.8345, 70.8697, 70.9167, 70.9472, 70.9824, 71.0035, 71.0575, 71.0856, 71.0997, 71.1326, 71.142, 71.1208, 71.142, 71.2077, 71.2006, 71.2147, 71.2382,
        71.2593, 71.2804, 71.3156, 71.3414, 71.3227, 71.3156, 71.2851, 71.257, 71.2734, 71.2429, 71.1772, 71.1631, 71.1865, 71.1865, 71.1725, 71.142, 71.0927, 71.0528, 71.0246, 71.0011, 70.9613,
        70.9472, 70.9261, 70.9495, 70.9096, 70.8392, 70.8556, 70.8486, 70.8157, 70.7993, 70.7688, 70.8486, 70.7759, 70.8275, 70.7852, 70.7289, 70.7266, 70.7993, 70.8416, 70.9143, 71.0011, 71.0387,
        71.0317, 71.0997, 71.0575, 71.1115, 71.1044, 71.0716, 70.9777, 70.9261, 70.8791, 70.8486, 70.8416, 70.8064, 70.8064, 70.8697, 70.9002, 70.8463, 70.8697, 70.8791, 70.8791, 70.9214, 70.9354,
        70.9331, 70.8627, 70.8087, 70.7759, 70.8064, 70.7759, 70.8697, 70.8674, 70.8674, 70.8463, 70.8228, 70.7524, 70.6609, 70.6233, 70.621, 70.5811, 70.5811, 70.574, 70.5482, 70.5553, 70.5553,
        70.5482, 70.5083, 70.4567, 70.4473, 70.4379, 70.4473, 70.459, 70.4778, 70.5036, 70.4966, 70.4825, 70.4403, 70.4262, 70.398, 70.4238] };
  WCL[1164] = struct {
   x = [-162.532, -162.457, -162.293],
   y = [69.9709, 70.0319, 70.0178] };
  WCL[1165] = struct {
   x = [-161.903, -162.04, -162.237, -162.204, -162.082, -161.993, -161.894, -161.805, -161.871, -161.903, -161.903, -161.918, -161.983, -161.861, -161.796, -161.617, -161.462, -161.242, -161.068, -160.88,
        -160.692, -160.58, -160.528, -160.749, -160.735, -160.528, -160.373, -160.237, -160.153, -160.064, -160.031, -160.139, -160.12, -160.228, -160.073, -160.064],
   y = [69.9568, 70.0084, 70.0624, 70.1375, 70.222, 70.2901, 70.2854, 70.2736, 70.2572, 70.2455, 70.2197, 70.1891, 70.1328, 70.1751, 70.215, 70.2197, 70.2337, 70.2924, 70.3159, 70.3605, 70.4051, 70.4403,
        70.4708, 70.4309, 70.4567, 70.5107, 70.5506, 70.5834, 70.5764, 70.5576, 70.5083, 70.4661, 70.398, 70.33, 70.3675, 70.2901] };
  WCL[1166] = struct {
   x = [-159.897, -160.028, -160.103],
   y = [70.6022, 70.6163, 70.6491] };
  WCL[1167] = struct {
   x = [-179.946, -179.927, -179.805, -179.707, -179.707, -179.692, -179.641, -179.519, -179.462, -179.35, -179.242, -179.129, -179.078, -179.078, -178.979, -178.876, -178.711, -178.482, -178.369, -178.195,
        -178.115, -178.172, -178.162, -178.073, -177.951, -177.829, -177.655, -177.599, -177.665, -177.853, -178.026, -178.228, -178.383, -178.458, -178.604, -178.791, -178.941, -179.129, -179.284,
        -179.486, -179.594, -179.725, -179.904],
   y = [71.5409, 71.5268, 71.5503, 71.5503, 71.555, 71.548, 71.5292, 71.5268, 71.5057, 71.5339, 71.555, 71.5409, 71.5644, 71.555, 71.548, 71.5292, 71.4987, 71.4846, 71.4541, 71.447, 71.4212, 71.4001,
        71.3766, 71.3555, 71.3133, 71.2851, 71.2382, 71.1678, 71.1115, 71.0622, 71.0293, 71.0011, 70.9824, 70.9777, 70.9706, 70.9425, 70.9354, 70.9002, 70.9073, 70.9214, 70.9542, 70.9613, 70.9683] };
  WCL[1168] = struct {
   x = [179.822, 180],
   y = [71.5292, 71.5383] };
  WCL[1169] = struct {
   x = [-180, -179.948],
   y = [71.5383, 71.5409] };
  WCL[1170] = struct {
   x = [170.415, 170.448, 170.613, 170.833, 171.096, 171.331],
   y = [69.931, 70.0084, 70.0483, 70.0554, 70.0436, 70.0178] };
  WCL[1171] = struct {
   x = [-179.901, -180],
   y = [70.9683, 70.9765] };
  WCL[1172] = struct {
   x = [180, 179.902, 179.747, 179.634, 179.592, 179.484, 179.296, 179.183, 179.127, 179.052, 178.906, 178.784, 178.752, 178.667, 178.634, 178.554, 178.479, 178.554, 178.719, 178.784, 178.972, 179.127,
        179.24, 179.348, 179.503, 179.625, 179.7, 179.822],
   y = [70.9765, 70.9847, 70.9566, 70.9143, 70.8627, 70.8556, 70.8556, 70.8463, 70.8697, 70.8533, 70.8533, 70.8275, 70.8744, 70.9167, 70.9354, 70.9706, 71.0293, 71.1068, 71.142, 71.21, 71.2593, 71.3133,
        71.3696, 71.3978, 71.447, 71.4494, 71.5057, 71.5292] };
  WCL[1173] = struct {
   x = [159.923, 160.003, 160.003, 160.045, 160.045],
   y = [70.452, 70.4074, 70.3675, 70.2994, 70.2267] };
  WCL[1174] = struct {
   x = [168.099, 168.165, 168.263, 168.475],
   y = [69.9944, 70.0061, 70.0319, 70.0202] };
  WCL[1175] = struct {
   x = [149.965, 150.031, 150.05, 150.04],
   y = [72.0173, 71.9563, 71.9023, 71.8437] };
  WCL[1176] = struct {
   x = [149.921, 150.099, 150.24, 150.329, 150.474, 150.662, 150.662, 150.728, 150.901, 151.211, 151.422, 151.62, 151.652, 151.732, 151.882, 152.005, 152.216, 152.371, 152.338, 152.206, 151.972, 151.784,
        151.897, 152.127, 152.305, 152.479, 152.591, 152.732, 152.943, 153.131, 153.187, 153.23, 153.309, 153.384, 153.53, 153.661, 153.859, 154.046, 154.211, 154.342, 154.455, 154.563, 154.619,
        154.741, 154.971, 155.224, 155.426, 155.558, 155.764, 155.952, 156.055, 156.285, 156.529, 156.75, 156.947, 157.158, 157.365, 157.651, 157.886, 158.092, 158.36, 158.623, 158.843, 158.998,
        159.162, 159.284, 159.35, 159.472, 159.627, 159.726, 159.815, 159.815, 159.923],
   y = [71.6536, 71.6254, 71.6184, 71.5456, 71.5081, 71.4846, 71.4377, 71.386, 71.3813, 71.4212, 71.4306, 71.4377, 71.3931, 71.3461, 71.2898, 71.2546, 71.203, 71.156, 71.0786, 71.0129, 70.9824, 70.9706,
        70.9425, 70.9378, 70.8885, 70.8416, 70.7899, 70.8345, 70.8298, 70.8744, 70.8815, 70.8697, 70.8885, 70.9073, 70.8908, 70.9096, 70.9167, 70.9167, 70.8908, 70.8979, 70.9143, 70.9378, 70.9636,
        70.9753, 70.9495, 70.9354, 70.9378, 70.9777, 70.9777, 70.9636, 70.9894, 71.0058, 71.0105, 70.9988, 71.0129, 71.0246, 71.0105, 71.0129, 70.9824, 70.9777, 70.9683, 70.9636, 70.9378, 70.8744,
        70.8416, 70.7805, 70.7407, 70.7078, 70.628, 70.6022, 70.5623, 70.5318, 70.452] };
  WCL[1177] = struct {
   x = [160.045, 159.979, 159.923],
   y = [70.2267, 70.1657, 70.0953] };
  WCL[1178] = struct {
   x = [149.888, 150.01, 150.221, 150.352, 150.629, 150.85, 151.089, 150.901, 150.901, 150.925, 150.695, 150.305, 150.155, 150.033],
   y = [75.202, 75.2114, 75.1926, 75.1386, 75.1128, 75.1081, 75.0964, 75.0565, 75.0119, 74.9274, 74.8406, 74.7749, 74.7537, 74.7326] };
  WCL[1179] = struct {
   x = [139.984, 140.04, 140.031],
   y = [71.4424, 71.494, 71.5644] };
  WCL[1180] = struct {
   x = [139.907, 140.005, 140.146, 140.367, 140.512, 140.643, 140.69, 140.723, 140.822, 141.01, 141.164, 141.118, 141.01, 140.996, 140.808, 140.831, 140.977, 141.118, 141.282, 141.352, 141.47, 141.549,
        141.704, 141.859, 141.976, 142.023, 142.066, 142.197, 142.474, 142.695, 142.896, 143.103, 143.291, 143.535, 143.765, 144.009, 144.248, 144.525, 144.671, 144.826, 144.966, 145.243, 145.52,
        145.708, 145.816, 145.905, 146.093, 146.215, 146.346, 146.501, 146.623, 146.745, 146.609, 146.389, 146.135, 145.961, 145.807, 145.694, 145.422, 145.131, 144.826, 144.591, 144.549, 144.417,
        144.351, 144.427, 144.459, 144.45, 144.351, 144.248, 144.183, 144.286, 144.624, 144.891, 145.145, 145.422, 145.605, 145.807, 146.004, 146.168, 146.356, 146.567, 146.764, 146.82, 146.623,
        146.435, 146.201, 146.069, 145.914, 145.882, 145.896, 145.896, 145.863, 145.905, 145.98, 146.069, 146.159, 146.149, 145.994, 145.849, 145.727, 145.727, 145.694, 145.661, 145.553, 145.44,
        145.342, 145.187, 144.999, 144.966, 145.145, 145.267, 145.398, 145.351, 145.234, 145.131, 145.013, 144.877, 144.924, 144.957, 144.999, 145.163, 145.253, 145.253, 145.3, 145.422, 145.586,
        145.75, 145.914, 146.004, 146.182, 146.356, 146.511, 146.576, 146.633, 146.745, 146.877, 147.017, 147.041, 147.172, 147.36, 147.538, 147.67, 147.867, 148.031, 148.111, 148.12, 148.135, 148.21,
        148.374, 148.487, 148.641, 148.848, 148.961, 149.05, 149.233, 149.444, 149.608, 149.787, 149.852, 149.965],
   y = [72.0948, 72.1253, 72.1699, 72.184, 72.2403, 72.313, 72.3975, 72.4491, 72.5008, 72.5571, 72.5759, 72.6064, 72.6416, 72.6909, 72.7402, 72.7801, 72.8317, 72.827, 72.8199, 72.8106, 72.7894, 72.7495,
        72.7167, 72.7167, 72.7096, 72.7143, 72.7096, 72.7003, 72.6979, 72.6838, 72.6932, 72.6909, 72.6909, 72.6909, 72.6651, 72.658, 72.6604, 72.651, 72.6181, 72.59, 72.5735, 72.5665, 72.5524, 72.543,
        72.5524, 72.5196, 72.4797, 72.4374, 72.4304, 72.4374, 72.4139, 72.374, 72.3647, 72.3647, 72.3717, 72.3881, 72.421, 72.421, 72.428, 72.4468, 72.4679, 72.5008, 72.489, 72.4867, 72.4445, 72.3928,
        72.3271, 72.2731, 72.2379, 72.1699, 72.137, 72.1253, 72.1253, 72.1417, 72.1534, 72.1722, 72.2309, 72.2567, 72.2708, 72.3036, 72.3083, 72.3177, 72.3248, 72.2802, 72.2074, 72.1065, 72.0103,
        71.9211, 71.9188, 71.9892, 72.0455, 72.1089, 72.1675, 72.2074, 72.1488, 72.1136, 72.1628, 72.2192, 72.2379, 72.2145, 72.1605, 72.1065, 72.0032, 71.9634, 71.9563, 71.9188, 71.9493, 71.9211,
        71.9399, 71.8883, 71.8671, 71.8742, 71.846, 71.8343, 71.8202, 71.7639, 71.792, 71.7498, 71.7099, 71.6723, 71.6231, 71.6442, 71.6512, 71.5855, 71.6137, 71.6442, 71.6864, 71.7263, 71.7615,
        71.8343, 71.8953, 71.9399, 71.9727, 72.0408, 72.1182, 72.1933, 72.2379, 72.2637, 72.3013, 72.3412, 72.367, 72.367, 72.3647, 72.3717, 72.36, 72.3271, 72.3412, 72.3435, 72.3506, 72.3576, 72.3647,
        72.3365, 72.3107, 72.2778, 72.2731, 72.2614, 72.2239, 72.191, 72.1534, 72.0854, 72.0173] };
  WCL[1181] = struct {
   x = [150.043, 149.935, 149.798, 149.691, 149.578, 149.47, 149.348, 149.235, 149.118, 149.005, 149.005, 149.104, 149.235, 149.381, 149.479, 149.667, 149.766, 149.921],
   y = [71.8437, 71.8272, 71.7967, 71.846, 71.8789, 71.8976, 71.8742, 71.8132, 71.7498, 71.7099, 71.6723, 71.6888, 71.6747, 71.6677, 71.6395, 71.6465, 71.6888, 71.6536] };
  WCL[1182] = struct {
   x = [139.961, 140.045, 140.167, 140.336, 140.308, 140.336, 140.27, 140.214, 140.252, 140.233, 140.266, 140.411, 140.51, 140.608, 140.697, 140.787, 140.838, 140.829, 140.829, 140.913, 140.895, 140.782,
        140.782, 140.754, 140.773, 140.829, 140.942, 141.082, 141.233, 141.364, 141.542, 141.664, 141.796, 141.908, 141.946, 142.012, 142.096, 142.265, 142.406, 142.411, 142.336, 142.303, 142.223,
        142.124, 142.073, 142.073, 142.106, 142.204, 142.303, 142.458, 142.556, 142.753, 142.941, 143.129, 143.26, 143.317, 143.293, 143.096, 142.922, 142.678, 142.458, 142.411, 142.401, 142.214,
        142.124, 141.904, 141.805, 141.918, 141.937, 141.763, 141.617, 141.495, 141.209, 141.003, 140.768, 140.514, 140.228, 139.998],
   y = [75.7934, 75.8239, 75.8122, 75.8028, 75.7652, 75.7441, 75.7206, 75.6831, 75.6784, 75.6502, 75.6174, 75.6268, 75.5986, 75.6244, 75.608, 75.608, 75.6221, 75.6502, 75.7324, 75.7605, 75.777, 75.8122,
        75.8755, 75.953, 76.0093, 76.0328, 76.0656, 76.0797, 76.0562, 76.0328, 76.0069, 75.9647, 75.9178, 75.8732, 75.8239, 75.7863, 75.7347, 75.7324, 75.7089, 75.6479, 75.6009, 75.5446, 75.493,
        75.4554, 75.4038, 75.3616, 75.3146, 75.2255, 75.1597, 75.1081, 75.0682, 75.0565, 75.0471, 75.0142, 74.9814, 74.9485, 74.9251, 74.9133, 74.8969, 74.9251, 74.9556, 74.9133, 74.8711, 74.8453,
        74.8805, 74.9462, 74.9767, 74.979, 75.0307, 75.0729, 75.033, 74.9814, 74.9626, 74.9485, 74.9133, 74.8476, 74.8312, 74.796] };
  WCL[1183] = struct {
   x = [141.329, 141.376, 141.47, 141.62, 141.751, 141.92, 142.108, 142.23, 142.343, 142.502, 142.699, 142.943, 143.122, 143.272, 143.45, 143.638, 143.845, 143.967, 144.089, 144.22, 144.398, 144.52, 144.652,
        144.84, 144.985, 145.135, 145.257, 145.379, 145.412, 145.347, 145.206, 145.084, 144.915, 144.849, 144.642, 144.694, 144.661, 144.661, 144.765, 144.675, 144.619, 144.375, 144.3, 144.023,
        143.849, 143.671, 143.338, 143, 142.732, 142.577, 142.436, 142.324, 142.357, 142.436, 142.502, 142.591, 142.676, 142.695, 142.544, 142.479, 142.357, 142.235, 142.122, 142.019, 141.878, 141.775,
        141.671, 141.521, 141.418, 141.399, 141.39, 141.371],
   y = [76.1431, 76.1431, 76.1313, 76.1055, 76.0586, 76.0093, 75.9741, 75.9201, 75.8802, 75.8521, 75.8521, 75.8262, 75.845, 75.8826, 75.8943, 75.8732, 75.8591, 75.7934, 75.7441, 75.7183, 75.7112, 75.6784,
        75.6549, 75.6291, 75.608, 75.5751, 75.5305, 75.5305, 75.4977, 75.4953, 75.5, 75.5071, 75.493, 75.4531, 75.4414, 75.4226, 75.385, 75.3217, 75.27, 75.2395, 75.1597, 75.1245, 75.0893, 75.0541,
        75.0166, 75.0541, 75.0612, 75.0846, 75.1386, 75.2372, 75.3217, 75.4038, 75.4813, 75.5352, 75.5986, 75.662, 75.7324, 75.7699, 75.7699, 75.8169, 75.8427, 75.8779, 75.9201, 75.9553, 75.9976,
        76.0304, 76.0586, 76.0867, 76.1266, 76.1384, 76.1431, 76.1431] };
  WCL[1184] = struct {
   x = [146.292, 146.306, 146.405, 146.569, 146.668, 146.724, 146.888, 147.165, 147.442, 147.574, 147.761, 147.991, 148.221, 148.367, 148.555, 148.597, 148.456, 148.433, 148.221, 148.343, 148.63, 148.916,
        149.052, 149.202, 149.381, 149.559, 149.667, 149.888],
   y = [75.5587, 75.5517, 75.5282, 75.5141, 75.486, 75.4484, 75.4226, 75.4015, 75.3991, 75.4155, 75.4038, 75.3874, 75.3757, 75.385, 75.371, 75.3311, 75.3287, 75.3006, 75.2513, 75.2536, 75.263, 75.256,
        75.2489, 75.2747, 75.2747, 75.2489, 75.1949, 75.202] };
  WCL[1185] = struct {
   x = [150.031, 149.843, 149.721, 149.622, 149.477, 149.388, 149.346, 149.148, 148.862, 148.66, 148.374, 148.12, 147.956, 147.778, 147.614, 147.44, 147.271, 147.064, 147.064, 147.041, 146.811, 146.722,
        146.511, 146.323, 146.257, 146.248, 146.257, 146.29],
   y = [74.7326, 74.7209, 74.7608, 74.7326, 74.7444, 74.7936, 74.7373, 74.7373, 74.7678, 74.8312, 74.8288, 74.85, 74.9086, 74.9321, 74.9908, 74.9978, 75.0119, 75.0494, 75.1175, 75.1809, 75.2536, 75.3428,
        75.3991, 75.4648, 75.5235, 75.5446, 75.5517, 75.5587] };
  WCL[1186] = struct {
   x = [140.953, 140.977, 140.808, 140.521, 140.249, 140.104, 140.08, 140.094, 140.127, 140.146, 140.146, 140.282, 140.479, 140.545, 140.437, 140.39, 140.489, 140.667, 140.7, 140.798, 140.864, 140.897,
        140.967, 141.047, 140.916, 140.953],
   y = [74.1389, 74.2234, 74.2703, 74.3078, 74.2656, 74.1975, 74.16, 74.106, 74.0544, 74.0028, 73.9464, 73.9159, 73.8925, 73.8925, 73.9159, 73.9464, 73.9558, 73.9535, 73.9887, 74.0028, 74.0168, 74.0333,
        74.0661, 74.0825, 74.099, 74.1389] };
  WCL[1187] = struct {
   x = [141.329, 141.338, 141.404, 141.517, 141.639, 141.77, 141.944, 142.131, 142.333, 142.497, 142.727, 142.873, 143.061, 143.202, 143.347, 143.422, 143.422, 143.502, 143.511, 143.324, 143.183, 142.939,
        142.695, 142.573, 142.399, 142.188, 141.878, 141.746, 141.502, 141.207, 140.888, 140.536, 140.282, 140.029],
   y = [73.8925, 73.8948, 73.8948, 73.8995, 73.9253, 73.9277, 73.923, 73.8971, 73.8455, 73.8385, 73.7423, 73.6977, 73.6672, 73.5921, 73.5263, 73.4536, 73.348, 73.2752, 73.2259, 73.2447, 73.2259, 73.233,
        73.2729, 73.2658, 73.2799, 73.2658, 73.3081, 73.3527, 73.3386, 73.3621, 73.4184, 73.4395, 73.3996, 73.3621] };
  WCL[1188] = struct {
   x = [139.785, 140.047, 140.315, 140.399, 140.536, 140.634, 140.742, 140.756, 140.841, 140.953, 141.118, 141.249, 141.329],
   y = [73.4724, 73.4677, 73.4841, 73.5217, 73.6108, 73.6883, 73.7681, 73.8267, 73.876, 73.8878, 73.876, 73.8878, 73.8925] };
  WCL[1189] = struct {
   x = [139.972, 140.005, 140.061, 140.179, 140.367, 140.536, 140.676, 140.709, 140.69, 140.587, 140.554, 140.47, 140.446, 140.446, 140.446, 140.348, 140.282, 140.137],
   y = [72.4726, 72.4773, 72.4632, 72.4609, 72.4632, 72.4797, 72.489, 72.4773, 72.4327, 72.36, 72.3083, 72.2614, 72.2943, 72.3506, 72.3647, 72.2708, 72.2309, 72.2379] };
  WCL[1190] = struct {
   x = [129.947, 130.055, 130.144, 130.233, 130.233, 130.322, 130.463, 130.599, 130.707, 130.838, 130.862, 130.937, 130.993, 131.05, 131.115, 131.214, 131.336, 131.402, 131.58, 131.632, 131.735, 131.899,
        132.031, 132.021, 132.007, 132.054, 132.078, 132.218, 132.35, 132.439, 132.514, 132.627, 132.725, 132.758, 132.782, 132.838, 132.904, 133.012, 133.091, 133.157, 133.209, 133.265, 133.42,
        133.565, 133.673, 133.838, 134.049, 134.227, 134.359, 134.546, 134.636, 134.72, 134.823, 134.833, 134.776, 134.823, 134.908, 135.02, 135.185, 135.363, 135.584, 135.79, 135.959, 136.058,
        136.264, 136.368, 136.508, 136.598, 136.621, 136.729, 136.86, 137.114, 137.245, 137.414, 137.48, 137.635, 137.635, 137.719, 137.855, 137.963, 138.062, 138.273, 138.33, 138.494, 138.583,
        138.714, 138.738, 138.757, 138.846, 138.935, 138.977, 139.066, 139.212, 139.334, 139.343, 139.433, 139.451, 139.639, 139.817, 139.982],
   y = [71.0387, 71.0786, 71.0387, 71.0129, 70.9706, 70.9237, 70.8955, 70.9026, 70.8533, 70.8627, 70.9378, 70.9214, 70.8486, 70.8134, 70.7289, 70.7078, 70.6937, 70.7805, 70.8556, 70.9683, 71.0058, 71.1396,
        71.1678, 71.2452, 71.3673, 71.4424, 71.5386, 71.6254, 71.731, 71.8132, 71.8742, 71.9469, 71.9704, 71.9493, 71.8718, 71.8085, 71.7474, 71.7169, 71.6888, 71.6465, 71.6066, 71.5573, 71.5221,
        71.4658, 71.4095, 71.4001, 71.4236, 71.4072, 71.4095, 71.4001, 71.4236, 71.379, 71.3367, 71.3743, 71.4517, 71.5081, 71.5667, 71.6184, 71.6254, 71.6324, 71.6395, 71.5949, 71.5926, 71.5433,
        71.5292, 71.5081, 71.5221, 71.5151, 71.5128, 71.4869, 71.4588, 71.4001, 71.3954, 71.3649, 71.2969, 71.2828, 71.2264, 71.1959, 71.1842, 71.142, 71.1819, 71.1772, 71.2405, 71.2828, 71.3321,
        71.3719, 71.379, 71.3813, 71.4001, 71.4236, 71.4001, 71.4377, 71.4236, 71.4025, 71.4236, 71.4869, 71.5128, 71.4729, 71.4424, 71.4424] };
  WCL[1191] = struct {
   x = [140.029, 139.982, 139.817, 139.728, 139.719, 139.705, 139.883, 139.86, 139.794, 139.752, 139.775, 139.907],
   y = [71.5644, 71.5996, 71.6583, 71.7381, 71.8085, 71.8507, 71.8531, 71.9047, 71.954, 71.9845, 72.0197, 72.0948] };
  WCL[1192] = struct {
   x = [136.013, 136.013, 136.07, 136.182, 136.257, 136.224, 136.215, 136.079, 135.971, 135.75, 135.671, 135.497, 135.417, 135.431, 135.539, 135.685, 135.849, 135.98, 136.013],
   y = [74.0544, 74.0426, 74.0192, 73.9981, 73.9488, 73.8878, 73.876, 73.8854, 73.9605, 74.0286, 74.0943, 74.1459, 74.1694, 74.2398, 74.2351, 74.1483, 74.0919, 74.0497, 74.0544] };
  WCL[1193] = struct {
   x = [135.849, 135.816, 135.793, 135.727, 135.783, 135.807, 135.717, 135.628, 135.487, 135.417, 135.464, 135.408, 135.464, 135.497, 135.464, 135.478, 135.483, 135.483, 135.487, 135.511, 135.553, 135.624,
        135.764, 135.886, 135.854, 135.797, 135.755],
   y = [75.5493, 75.5141, 75.4883, 75.4672, 75.4202, 75.3827, 75.3451, 75.3545, 75.3757, 75.4273, 75.4695, 75.5, 75.5446, 75.6244, 75.6925, 75.7441, 75.8004, 75.845, 75.8215, 75.7887, 75.7558, 75.6737,
        75.6268, 75.6174, 75.5986, 75.5681, 75.5493] };
  WCL[1194] = struct {
   x = [139.918, 139.956, 139.965],
   y = [75.838, 75.8239, 75.7934] };
  WCL[1195] = struct {
   x = [140.005, 139.85, 139.672, 139.817, 139.949, 139.775, 139.639, 139.639, 139.63, 139.564, 139.386, 139.376, 139.367, 139.334, 139.498, 139.409, 139.334, 139.221, 139.043, 138.879, 138.738, 138.569,
        138.306, 138.128, 137.94, 137.799, 137.766, 137.719, 137.654, 137.433, 137.283, 137.194, 137.095, 136.95, 136.86, 136.771, 136.705, 136.673, 136.86, 137.039, 137.259, 137.17, 136.992, 136.95,
        136.992, 137.147, 137.128, 136.917, 136.818, 136.795, 136.828, 136.917, 136.992, 137.048, 137.062, 137.081, 136.982, 136.982, 137.058, 137.184, 137.222, 137.203, 137.175, 137.133, 137.1,
        137.058, 137.081, 137.151, 137.203, 137.278, 137.363, 137.442, 137.48, 137.532, 137.56, 137.602, 137.658, 137.771, 137.823, 137.912, 137.945, 137.898, 137.912, 137.973, 138.067, 138.156,
        138.203, 138.287, 138.306, 138.376, 138.32, 138.362, 138.409, 138.447],
   y = [74.796, 74.7983, 74.8429, 74.8335, 74.8687, 74.8969, 74.9603, 75.0189, 75.0565, 75.0494, 75.0189, 74.979, 74.9556, 74.9063, 74.9204, 74.8593, 74.7537, 74.6998, 74.688, 74.7185, 74.735, 74.742,
        74.7537, 74.7866, 74.8171, 74.9016, 74.9415, 74.9767, 75.0354, 75.0283, 75.0541, 75.0846, 75.0893, 75.1269, 75.2114, 75.2653, 75.3217, 75.3381, 75.3686, 75.3381, 75.3639, 75.4179, 75.4062,
        75.4296, 75.4461, 75.4461, 75.5024, 75.5423, 75.5775, 75.5916, 75.5963, 75.5751, 75.547, 75.5775, 75.6009, 75.6315, 75.676, 75.7465, 75.777, 75.7629, 75.7934, 75.8122, 75.8497, 75.8262,
        75.8568, 75.9201, 75.9459, 75.9905, 75.9741, 75.9905, 76.0023, 76.0281, 76.0774, 76.1384, 76.1196, 76.136, 76.1665, 76.1994, 76.2041, 76.2182, 76.2111, 76.1595, 76.1126, 76.0797, 76.0304,
        75.9788, 75.9671, 75.9788, 75.9459, 75.9154, 75.8732, 75.8286, 75.8262, 75.838] };
  WCL[1196] = struct {
   x = [140.029, 139.841, 139.705, 139.653, 139.785],
   y = [73.3621, 73.3668, 73.416, 73.4372, 73.4724] };
  WCL[1197] = struct {
   x = [137.039, 137.048, 137.081, 137.194, 137.302, 137.367, 137.513, 137.733, 137.907, 137.977, 137.855, 137.776, 137.621, 137.489, 137.349, 137.203, 136.959, 136.771, 136.738, 136.785, 136.907, 136.95,
        137.015, 137.039],
   y = [71.6137, 71.6137, 71.6019, 71.5667, 71.5644, 71.5879, 71.5855, 71.5785, 71.5714, 71.5339, 71.4799, 71.4353, 71.4564, 71.4353, 71.4494, 71.4729, 71.5081, 71.5362, 71.5597, 71.5714, 71.5738, 71.5996,
        71.609, 71.6137] };
  WCL[1198] = struct {
   x = [138.043, 138.062, 138.161, 138.25, 138.348, 138.461, 138.517, 138.527, 138.635, 138.804, 138.935, 138.958, 138.958, 138.944, 138.902, 138.822, 138.724, 138.559, 138.437, 138.315, 138.24, 138.109,
        137.996, 137.874, 137.865, 137.987, 138.053, 138.029, 138.029, 138.029, 138.043],
   y = [71.5268, 71.5386, 71.5573, 71.5785, 71.5785, 71.5433, 71.5785, 71.6395, 71.6465, 71.6184, 71.5714, 71.5198, 71.494, 71.4635, 71.4377, 71.4142, 71.379, 71.3367, 71.3508, 71.3508, 71.3438, 71.325,
        71.3579, 71.3508, 71.4001, 71.4377, 71.4729, 71.5081, 71.5151, 71.5362, 71.5268] };
  WCL[1199] = struct {
   x = [139.972, 139.972],
   y = [72.4726, 72.4726] };
  WCL[1200] = struct {
   x = [140.139, 139.961, 139.81, 139.599, 139.533, 139.486, 139.388, 139.266, 139.158, 139.233, 139.369, 139.336, 139.379, 139.51, 139.51, 139.468, 139.609, 139.778, 139.895, 139.961, 139.975],
   y = [72.2379, 72.2802, 72.2239, 72.2332, 72.2262, 72.1957, 72.1769, 72.2098, 72.2544, 72.2943, 72.3083, 72.3529, 72.3881, 72.3975, 72.4374, 72.4867, 72.5078, 72.5031, 72.5031, 72.4937, 72.4726] };
  WCL[1201] = struct {
   x = [119.977, 120.142, 120.296, 120.461, 120.672, 120.893, 121.047, 121.08, 121.123, 121.245, 121.432, 121.564, 121.719, 121.916, 122.028, 122.094, 122.207, 122.291, 122.348, 122.46, 122.578, 122.732,
        123, 123.141, 123.239, 123.385, 123.451, 123.352, 123.221, 123.164, 123.174, 123.23, 123.141, 123.155, 123.206, 123.31, 123.319, 123.352, 123.441, 123.526, 123.638, 123.803, 123.859, 123.826,
        123.826, 123.859, 123.967, 124.07, 124.187, 124.399, 124.544, 124.643, 124.708, 124.816, 124.938, 125.028, 125.093, 125.173, 125.215, 125.225, 125.281, 125.394, 125.459, 125.558, 125.647,
        125.713, 125.788, 125.919, 125.999, 126.088, 126.206, 126.328, 126.408, 126.539, 126.562, 126.614, 126.68, 126.75, 126.802, 126.938, 126.971, 127.046, 127.177, 127.267, 127.276, 127.299,
        127.356, 127.454, 127.543, 127.642, 127.764, 127.83, 127.961, 128.14, 128.191, 128.262, 128.191, 128.28, 128.402, 128.492, 128.524, 128.501, 128.646, 128.82, 128.933, 128.933, 129.088, 129.177,
        129.341, 129.449, 129.463, 129.515, 129.473, 129.463, 129.43, 129.308, 129.341, 129.318, 129.341, 129.416, 129.36, 129.374, 129.374, 129.252, 129.219, 129.308, 129.397, 129.351, 129.252,
        129.228, 129.383, 129.538, 129.627, 129.548, 129.463, 129.463, 129.505, 129.407, 129.416, 129.529, 129.416, 129.219, 129.196, 129.261, 129.177, 129.074, 129.041, 128.966, 128.9, 128.834,
        128.712, 128.736, 128.768, 128.844, 128.886, 128.844, 128.942, 129.031, 129.121, 129.139, 129.177, 129.177, 129.196, 129.261, 129.261, 129.318, 129.374, 129.496, 129.505, 129.562, 129.693,
        129.749, 129.815, 129.904, 129.947],
   y = [73.0218, 73.0171, 73.003, 72.996, 72.9819, 72.9514, 72.9209, 72.9302, 72.942, 72.9467, 72.9608, 72.9701, 72.9701, 72.942, 72.9443, 72.9537, 72.9631, 72.942, 72.9725, 72.9467, 72.9772, 73.0124,
        73.0241, 73.0405, 73.0335, 73.0828, 73.125, 73.2001, 73.2964, 73.3738, 73.3996, 73.4301, 73.4442, 73.4724, 73.5475, 73.6648, 73.7516, 73.7775, 73.7446, 73.7024, 73.6859, 73.6883, 73.7094,
        73.7657, 73.7986, 73.8361, 73.8338, 73.815, 73.7845, 73.7751, 73.7681, 73.7446, 73.7376, 73.754, 73.7094, 73.6789, 73.7071, 73.6883, 73.646, 73.6085, 73.5827, 73.5686, 73.6085, 73.578, 73.5968,
        73.5803, 73.5498, 73.5615, 73.5498, 73.5334, 73.5639, 73.5263, 73.4864, 73.4536, 73.4207, 73.402, 73.4301, 73.4395, 73.4653, 73.4677, 73.5123, 73.5334, 73.5193, 73.5123, 73.5052, 73.5029,
        73.4911, 73.4935, 73.4864, 73.4818, 73.4935, 73.5005, 73.4771, 73.4583, 73.4372, 73.416, 73.3973, 73.3738, 73.3715, 73.3668, 73.3269, 73.2893, 73.2565, 73.2259, 73.2189, 73.1837, 73.1368,
        73.0781, 73.0476, 73.0218, 72.9936, 72.9725, 72.9443, 72.8997, 72.8833, 72.8763, 72.827, 72.7918, 72.7519, 72.7425, 72.7308, 72.6768, 72.658, 72.6345, 72.5853, 72.5594, 72.5407, 72.5289,
        72.4937, 72.4632, 72.4398, 72.4069, 72.3764, 72.3201, 72.2731, 72.2192, 72.2027, 72.1722, 72.1534, 72.1605, 72.1229, 72.0924, 72.0549, 72.0338, 72.0267, 72.0525, 72.0244, 72.0032, 71.9704,
        71.9352, 71.8883, 71.7967, 71.7568, 71.7381, 71.6888, 71.6395, 71.5949, 71.6137, 71.6324, 71.6536, 71.6254, 71.6066, 71.5855, 71.5456, 71.4799, 71.4588, 71.379, 71.2875, 71.2476, 71.2687,
        71.2405, 71.1607, 71.1021, 71.0739, 71.0387] };
  WCL[1202] = struct {
   x = [109.942, 110.064, 110.177, 110.252, 110.383, 110.515, 110.627, 110.778, 110.956, 111.078, 111.078, 111.069, 111.134, 111.228, 111.397, 111.472, 111.407, 111.454, 111.5, 111.613, 111.773, 111.857,
        111.923, 112.045, 112.073, 112.073, 112.205, 112.355, 112.477, 112.618, 112.674, 112.636, 112.608, 112.552, 112.477, 112.58, 112.665, 112.721, 112.749, 112.796, 112.89, 112.909, 113.078,
        113.021, 112.965, 113.059, 113.143, 113.247, 113.34, 113.425, 113.406, 113.425, 113.472, 113.491, 113.631, 113.753, 113.81, 113.81, 113.735, 113.669, 113.603, 113.491, 113.369, 113.425,
        113.491, 113.35, 112.993, 112.871, 112.74, 112.618, 112.43, 112.289, 112.242, 112.364, 112.505, 112.608, 112.542, 112.632, 112.632, 112.73, 112.796, 112.829, 112.97, 113.026, 113.04, 113.19,
        113.279, 113.425, 113.523, 113.547, 113.556, 113.392, 113.401, 113.458, 113.5, 113.5, 113.5, 113.401, 113.336, 113.228, 113.124, 113.124, 113.073, 112.937, 112.862, 112.763, 112.608, 112.608,
        112.707, 112.599, 112.477, 112.345, 112.388, 112.312, 112.125, 111.989, 111.937, 112.054, 112.2, 112.068, 111.89, 111.782, 111.627, 111.637, 111.726, 111.637, 111.613, 111.515, 111.397,
        111.294, 111.177, 111.031, 110.918, 110.876, 110.853, 110.665, 110.599, 110.491, 110.548, 110.557, 110.426, 110.224, 110.172, 110.191, 110.074, 109.905, 109.853],
   y = [76.6617, 76.6781, 76.6946, 76.718, 76.7204, 76.7087, 76.6993, 76.6969, 76.7063, 76.7133, 76.6922, 76.6641, 76.6382, 76.6617, 76.65, 76.6195, 76.5937, 76.5913, 76.5749, 76.5608, 76.5444, 76.5092,
        76.4951, 76.4482, 76.4223, 76.3989, 76.413, 76.3754, 76.3379, 76.312, 76.2956, 76.2674, 76.2252, 76.2111, 76.2064, 76.1736, 76.1525, 76.1173, 76.0867, 76.0656, 76.068, 76.0985, 76.1384,
        76.1665, 76.197, 76.2111, 76.2182, 76.1736, 76.1384, 76.0867, 76.0398, 75.9929, 75.9483, 75.9131, 75.899, 75.9131, 75.8614, 75.7934, 75.7253, 75.6737, 75.6221, 75.5751, 75.5634, 75.5892,
        75.6408, 75.6667, 75.6901, 75.7112, 75.73, 75.7699, 75.8215, 75.8286, 75.8028, 75.7535, 75.7277, 75.7089, 75.6807, 75.6244, 75.6056, 75.5634, 75.5845, 75.6268, 75.6385, 75.6338, 75.5681,
        75.5329, 75.5118, 75.5071, 75.5118, 75.4531, 75.4226, 75.439, 75.4179, 75.3757, 75.371, 75.3264, 75.2841, 75.256, 75.2137, 75.1785, 75.155, 75.1363, 75.0823, 75.026, 75.0283, 75.0283, 74.9884,
        74.9814, 74.965, 74.9251, 74.9227, 74.8734, 74.9227, 74.9204, 74.8969, 74.8734, 74.8429, 74.8359, 74.8218, 74.8077, 74.7983, 74.7819, 74.7678, 74.7561, 74.7232, 74.6857, 74.6387, 74.6575,
        74.6833, 74.6927, 74.6622, 74.62, 74.6223, 74.5777, 74.5543, 74.5378, 74.5519, 74.5425, 74.5402, 74.5284, 74.4956, 74.4839, 74.4463, 74.4252, 74.3994, 74.3548, 74.3337] };
  WCL[1203] = struct {
   x = [109.987, 110.085, 110.175, 110.231, 110.297, 110.428, 110.625, 110.846, 110.78, 110.714, 110.559, 110.475, 110.287, 110.151],
   y = [73.4911, 73.524, 73.5123, 73.5498, 73.5803, 73.6014, 73.6249, 73.6554, 73.7071, 73.7446, 73.7681, 73.7258, 73.7024, 73.6531] };
  WCL[1204] = struct {
   x = [109.787, 110.017, 110.294, 110.491, 110.703, 110.792, 110.867, 110.9, 111.031, 111.266, 111.496, 111.472, 111.308, 111.266, 111.177, 111.2, 111.275, 111.472, 111.641, 111.684, 111.862, 112.115,
        112.411, 112.711, 112.829, 112.876, 112.819, 112.777, 112.876, 112.984, 113.139, 113.171, 113.27, 113.359, 113.369, 113.251, 113.186, 113.27, 113.218, 113.326, 113.505, 113.439, 113.186,
        113.129, 113.406, 113.735, 113.866, 114.101, 114.251, 114.251, 114.462, 114.626, 114.486, 114.289, 114.044, 113.824, 113.645, 113.547, 113.777, 114.044, 114.185, 114.396, 114.594, 114.673,
        114.706, 114.861, 115.035, 115.279, 115.321, 115.321, 115.476, 115.664, 115.819, 116.03, 116.218, 116.358, 116.603, 116.823, 117.02, 117.255, 117.508, 117.706, 117.846, 118.034, 118.236, 118.4,
        118.597, 118.729, 118.865, 118.916],
   y = [73.9746, 74.0004, 74.0121, 73.9652, 73.9464, 73.8925, 73.9136, 73.9535, 74.0004, 74.0286, 74.0333, 74.0028, 73.9347, 73.9793, 73.9277, 73.8572, 73.7939, 73.7446, 73.7071, 73.6789, 73.693, 73.6789,
        73.6789, 73.7047, 73.7751, 73.8643, 73.9347, 74.0028, 73.937, 73.876, 73.8056, 73.7164, 73.693, 73.6202, 73.5498, 73.4653, 73.4207, 73.3761, 73.3269, 73.3104, 73.3057, 73.2518, 73.2635,
        73.2283, 73.1978, 73.2635, 73.3057, 73.2752, 73.2001, 73.3057, 73.1438, 73.1485, 73.2189, 73.3057, 73.3644, 73.3715, 73.4442, 73.4958, 73.5193, 73.5545, 73.5686, 73.5709, 73.5522, 73.5545,
        73.5639, 73.5639, 73.5827, 73.6108, 73.6179, 73.6601, 73.6765, 73.6883, 73.6859, 73.6695, 73.646, 73.6273, 73.6273, 73.5968, 73.5991, 73.5592, 73.5475, 73.5381, 73.5522, 73.5615, 73.531,
        73.5263, 73.5381, 73.5334, 73.5052, 73.4653] };
  WCL[1205] = struct {
   x = [113.019, 113.085, 113.15, 113.305, 113.352, 113.352, 113.329, 113.263, 113.207, 113.052, 112.911, 112.78, 112.578, 112.39, 112.193, 111.982, 111.808, 111.611, 111.475, 111.489, 111.554, 111.686,
        111.817, 111.93, 112.005, 111.982, 111.93, 112.005, 112.118, 112.291, 112.446, 112.578, 112.7, 112.812, 112.944, 113.019],
   y = [74.512, 74.4909, 74.4627, 74.4627, 74.4346, 74.3876, 74.3125, 74.2562, 74.2093, 74.1811, 74.1365, 74.0966, 74.1177, 74.1365, 74.1365, 74.2116, 74.2327, 74.2515, 74.3078, 74.3337, 74.3736, 74.3407,
        74.336, 74.3642, 74.4111, 74.4557, 74.5026, 74.5472, 74.5308, 74.5191, 74.505, 74.5003, 74.4956, 74.4932, 74.5026, 74.512] };
  WCL[1206] = struct {
   x = [118.942, 118.942, 118.877, 118.811, 118.778, 118.689, 118.581, 118.459, 118.426, 118.459, 118.468, 118.459, 118.435, 118.435, 118.435, 118.445, 118.534, 118.689, 118.867, 119.064, 119.229, 119.393,
        119.529, 119.637, 119.759, 119.937, 119.989, 119.98],
   y = [73.4677, 73.4559, 73.463, 73.4653, 73.4301, 73.4207, 73.4348, 73.416, 73.3738, 73.3386, 73.3057, 73.2893, 73.2752, 73.2729, 73.2447, 73.2095, 73.1673, 73.1532, 73.111, 73.0851, 73.0687, 73.0593,
        73.0288, 73.003, 72.9819, 72.9561, 72.9819, 73.0218] };
  WCL[1207] = struct {
   x = [99.8649, 100.015, 100.118, 100.231],
   y = [79.6657, 79.6774, 79.6727, 79.6539] };
  WCL[1208] = struct {
   x = [99.9752, 100.032, 100.088, 100.088, 100.154, 100.276, 100.229, 100.388, 100.548, 100.604, 100.736, 100.867, 101.064, 100.989, 100.792, 100.792, 100.782, 100.811, 100.895, 100.998, 101.055, 101.149,
        101.036, 101.027, 101.149, 101.336, 101.224, 101.233, 101.421, 101.59, 101.618, 101.59, 101.712, 101.815, 101.862, 101.965, 102.153, 102.191, 102.181, 102.153, 102.228, 102.36, 102.444,
        102.594, 102.763, 102.866, 102.876, 102.866, 102.904, 102.989, 103.12, 103.12, 102.998, 102.904, 102.923, 102.866, 102.876, 103.054, 103.139, 103.308, 103.308, 103.411, 103.542, 103.599,
        103.627, 103.674, 103.796, 103.852, 103.871, 104.04, 103.965, 104.04, 104.115, 104.068, 104.077, 104.199, 104.303, 104.275, 104.331, 104.425, 104.462, 104.5, 104.566, 104.659, 104.866, 105.054,
        105.195, 105.27, 105.41, 105.439, 105.457, 105.457, 105.401, 105.457, 105.457, 105.288, 105.129, 104.932, 104.819, 104.65, 104.434, 104.19, 103.993, 103.749, 103.448, 103.261, 103.064, 102.989,
        102.951, 102.81, 102.716, 102.538, 102.406, 102.247, 102.04, 101.918, 101.74, 101.468, 101.111, 101.045, 100.998, 100.905, 100.745, 100.529, 100.294, 100.069],
   y = [78.4312, 78.4359, 78.457, 78.4876, 78.5345, 78.6002, 78.6354, 78.7058, 78.7856, 78.7739, 78.7457, 78.7269, 78.7293, 78.7574, 78.7973, 78.8607, 78.9123, 78.9757, 78.9968, 79.0179, 79.0156, 79.025,
        79.032, 79.0766, 79.1071, 79.1142, 79.1259, 79.1775, 79.2245, 79.2456, 79.2784, 79.3254, 79.3441, 79.2949, 79.2338, 79.2221, 79.2151, 79.262, 79.3207, 79.3723, 79.4005, 79.4286, 79.4169,
        79.4075, 79.4028, 79.384, 79.3371, 79.2996, 79.3277, 79.3488, 79.323, 79.2831, 79.2338, 79.1775, 79.1048, 79.0555, 79.0109, 79.0555, 79.0766, 79.14, 79.2033, 79.208, 79.1752, 79.1587, 79.1048,
        79.0883, 79.1259, 79.1658, 79.1799, 79.1634, 79.1329, 79.1095, 79.1001, 79.0508, 79.0226, 78.978, 78.978, 79.025, 79.0344, 79.0508, 78.9851, 78.9569, 78.917, 78.8677, 78.8419, 78.8232, 78.7997,
        78.7903, 78.7527, 78.7199, 78.6753, 78.6143, 78.572, 78.5251, 78.4594, 78.3984, 78.3303, 78.2904, 78.3092, 78.2975, 78.2975, 78.2693, 78.2388, 78.213, 78.22, 78.2012, 78.213, 78.2036, 78.1707,
        78.152, 78.1731, 78.2153, 78.2036, 78.2036, 78.1801, 78.2059, 78.1825, 78.1825, 78.1872, 78.152, 78.1168, 78.1121, 78.0581, 78.0417, 78.0135, 77.9689] };
  WCL[1209] = struct {
   x = [99.9869, 100.034],
   y = [78.3397, 78.3655] };
  WCL[1210] = struct {
   x = [107.023, 107.041, 107.107, 107.22, 107.351, 107.455, 107.605, 107.642, 107.774, 107.802, 107.642, 107.586, 107.473, 107.276, 107.023, 106.91, 106.694, 106.535, 106.6, 106.76, 106.844, 106.966,
        107.023],
   y = [78.1754, 78.1637, 78.1566, 78.1355, 78.1496, 78.1731, 78.1731, 78.152, 78.152, 78.1191, 78.0698, 78.0276, 78.051, 78.0534, 78.0769, 78.0886, 78.105, 78.1097, 78.1308, 78.1285, 78.1613, 78.1684,
        78.1754] };
  WCL[1211] = struct {
   x = [99.9517, 100.093, 100.299, 100.44, 100.59, 100.712, 100.862, 101.012, 101.116, 101.135, 101.069, 101.069, 101.069, 101.069, 101.116, 101.228, 101.322, 101.341, 101.228, 101.191, 101.181, 101.059,
        100.928, 100.853, 100.853, 100.815, 100.9, 100.984, 100.994, 101.106, 101.172, 101.163, 101.191, 101.294, 101.388, 101.444, 101.51, 101.557, 101.679, 101.707, 101.82, 101.867, 101.989, 102.083,
        102.101, 102.111, 102.13, 102.186, 102.289, 102.374, 102.477, 102.543, 102.618, 102.683, 102.777, 102.927, 102.946, 102.993, 103.143, 103.312, 103.425, 103.538, 103.632, 103.641, 103.716,
        103.838, 103.904, 104.007, 104.101, 104.157, 104.26, 104.26, 104.279, 104.392, 104.448, 104.523, 104.636, 104.702, 104.824, 104.974, 105.115, 105.293, 105.443, 105.528, 105.594, 105.697,
        105.744, 105.8, 105.885, 105.978, 106.044, 105.885, 105.744, 105.612, 105.612, 105.528, 105.453, 105.274, 105.049, 104.946, 104.824, 104.711, 104.608, 104.476, 104.326, 104.26, 104.373,
        104.542, 104.72, 104.899, 105.04, 105.152, 105.274, 105.396, 105.518, 105.631, 105.678, 105.734, 105.744, 105.669, 105.669, 105.781, 105.866, 105.969, 106.025, 106.1, 106.176, 106.241, 106.401,
        106.467, 106.56, 106.598, 106.664, 106.739, 106.748, 106.804, 106.898, 107.02, 107.133, 107.227, 107.208, 107.218, 107.311, 107.311, 107.18, 107.152, 107.208, 107.058, 106.983, 106.992,
        107.086, 107.18, 107.283, 107.18, 107.095, 106.927, 106.842, 106.776, 106.645, 106.56, 106.476, 106.401, 106.307, 106.176, 106.082, 106.119, 106.119, 106.166, 106.307, 106.485, 106.645,
        106.748, 106.908, 107.058, 107.189, 107.311, 107.368, 107.415, 107.443, 107.631, 107.734, 107.743, 107.809, 107.865, 107.987, 107.95, 107.8, 107.856, 107.884, 107.94, 108.091, 108.119, 108.128,
        108.241, 108.269, 108.194, 108.166, 108.26, 108.344, 108.372, 108.447, 108.607, 108.757, 108.954, 108.992, 109.086, 109.236, 109.311, 109.367, 109.489, 109.593, 109.677, 109.771, 109.808,
        109.846, 109.94],
   y = [76.4505, 76.4482, 76.4435, 76.4505, 76.4622, 76.474, 76.4998, 76.5092, 76.5303, 76.5373, 76.5514, 76.5819, 76.6312, 76.6547, 76.6242, 76.6218, 76.6148, 76.6336, 76.6523, 76.6781, 76.711, 76.7227,
        76.7345, 76.765, 76.8025, 76.8119, 76.8542, 76.8894, 76.9222, 76.9222, 76.9574, 76.9832, 77.0255, 77.0701, 77.0982, 77.1193, 77.1499, 77.1639, 77.1944, 77.2226, 77.2367, 77.2602, 77.2742,
        77.2954, 77.2977, 77.3024, 77.3188, 77.347, 77.3822, 77.4057, 77.4409, 77.4714, 77.4972, 77.5136, 77.5347, 77.5582, 77.5887, 77.6098, 77.6075, 77.6004, 77.5958, 77.6051, 77.6286, 77.6427,
        77.6568, 77.6779, 77.6896, 77.6849, 77.6943, 77.692, 77.6755, 77.6521, 77.638, 77.6333, 77.6568, 77.6709, 77.6427, 77.6028, 77.5652, 77.5418, 77.5207, 77.5089, 77.5136, 77.5441, 77.53, 77.5277,
        77.4878, 77.4244, 77.3775, 77.347, 77.3212, 77.347, 77.3423, 77.34, 77.3118, 77.2742, 77.239, 77.2367, 77.1944, 77.1522, 77.1053, 77.0701, 77.0349, 77.0231, 76.9668, 76.9386, 76.9598, 76.9762,
        77.002, 77.0044, 77.0231, 77.0466, 77.0583, 77.0583, 77.0607, 77.0513, 77.0771, 77.1287, 77.0701, 77.0184, 77.0044, 77.0208, 77.0231, 77.0443, 77.0255, 77.0044, 77.0114, 77.0349, 77.0349,
        77.0231, 77.0255, 77.0044, 76.9856, 76.9973, 76.9997, 76.9903, 76.9762, 76.9715, 76.9715, 76.9621, 76.948, 76.9175, 76.9011, 76.8776, 76.88, 76.8659, 76.8401, 76.8049, 76.7744, 76.7556,
        76.7838, 76.8143, 76.8283, 76.7884, 76.7368, 76.7204, 76.718, 76.6735, 76.6312, 76.5702, 76.5092, 76.5209, 76.5068, 76.4904, 76.4763, 76.4716, 76.4482, 76.4458, 76.4482, 76.4482, 76.4341,
        76.4247, 76.4482, 76.4435, 76.4435, 76.4458, 76.4904, 76.4927, 76.4763, 76.4998, 76.5444, 76.5913, 76.5866, 76.5913, 76.6359, 76.6382, 76.6289, 76.6547, 76.7016, 76.711, 76.711, 76.7298,
        76.7415, 76.7368, 76.718, 76.6969, 76.6664, 76.6781, 76.7016, 76.6875, 76.6735, 76.6781, 76.6852, 76.7016, 76.6969, 76.6946, 76.7087, 76.7345, 76.7227, 76.704, 76.6875, 76.6641, 76.6735,
        76.6735, 76.6735, 76.6617] };
  WCL[1212] = struct {
   x = [110.031, 109.994, 109.994, 109.984, 109.928, 109.886, 109.731, 109.51, 109.224, 108.938, 108.708, 108.637, 108.506, 108.365, 108.252, 108.145, 108.032, 107.858, 107.605, 107.361, 107.117, 107.107,
        107.196, 107.018, 107.065, 107.027, 106.91, 106.807, 106.765, 106.586, 106.333, 106.168, 106.037, 105.981, 105.971, 105.882, 105.793, 105.718, 105.652, 105.596, 105.45, 105.333, 105.197,
        105.112, 105, 104.779, 104.713, 104.549, 104.427, 104.305, 104.141, 104.164, 104.282, 104.329, 104.559, 104.638, 104.559, 104.657, 104.826, 105.023, 105.131, 105.263, 105.464, 105.652, 105.76,
        105.882, 106.014, 106.183, 106.422, 106.61, 106.788, 106.919, 106.887, 106.708, 106.511, 106.291, 106.168, 106.145, 106.248, 106.347, 106.657, 106.854, 106.985, 107.262, 107.516, 107.591,
        107.746, 107.957, 108.112, 108.299, 108.243, 108.276, 108.487, 108.661, 108.881, 109.135, 109.191, 109.168, 109.266, 109.337, 109.224, 109.102, 109.116, 109.182, 109.29, 109.435, 109.642,
        109.811, 109.984],
   y = [74.3337, 74.329, 74.3243, 74.3196, 74.2656, 74.2257, 74.1576, 74.1013, 74.0685, 73.991, 73.9277, 73.8432, 73.7915, 73.7071, 73.6531, 73.5874, 73.6273, 73.6085, 73.6061, 73.6014, 73.6014, 73.5733,
        73.5615, 73.5404, 73.5146, 73.4864, 73.4489, 73.3973, 73.3198, 73.2964, 73.2752, 73.287, 73.2072, 73.1579, 73.1274, 73.0945, 73.0405, 72.9889, 72.9443, 72.9091, 72.8481, 72.8199, 72.7847,
        72.7894, 72.8012, 72.7566, 72.7214, 72.7003, 72.658, 72.6252, 72.5923, 72.5501, 72.5407, 72.5759, 72.5947, 72.5829, 72.5524, 72.5853, 72.6557, 72.705, 72.7261, 72.7566, 72.7589, 72.7988,
        72.8598, 72.8857, 72.942, 72.9044, 72.8857, 72.8575, 72.8411, 72.8575, 72.8669, 72.888, 72.9185, 72.9514, 73.0429, 73.1133, 73.1297, 73.1649, 73.1297, 73.125, 73.1626, 73.1532, 73.1485,
        73.1344, 73.1649, 73.233, 73.2236, 73.2166, 73.2518, 73.2964, 73.2917, 73.3175, 73.3198, 73.348, 73.3855, 73.4278, 73.3973, 73.4067, 73.4559, 73.524, 73.5263, 73.5005, 73.4653, 73.416, 73.416,
        73.4653, 73.4911] };
  WCL[1213] = struct {
   x = [110.151, 109.977, 109.799, 109.691, 109.677, 109.776, 109.668, 109.522, 109.438, 109.447, 109.644, 109.743, 109.79],
   y = [73.6531, 73.6578, 73.6484, 73.6789, 73.639, 73.6249, 73.6014, 73.6672, 73.7423, 73.7915, 73.8572, 73.9042, 73.9746] };
  WCL[1214] = struct {
   x = [92.3667, 92.5733, 92.761, 92.9488, 93.1083, 93.2961, 93.4651, 93.634, 93.6622, 93.5402, 93.4087, 93.3149, 93.2585, 93.099, 92.8924, 92.761, 92.6671, 92.5075, 92.301, 92.1227, 91.9537, 91.7753,
        91.6251, 91.4937, 91.3247, 91.3341, 91.4468, 91.6158, 91.6908, 91.8129, 91.9255, 91.9912, 92.1321, 92.2165, 92.2541, 92.1321, 91.9631, 91.8035, 91.6158, 91.4562, 91.3154, 91.3341, 91.4468,
        91.428, 91.3998, 91.2966, 91.1464, 91.1276, 91.1933, 91.259, 91.2684],
   y = [80.0106, 79.9895, 79.9825, 79.9872, 79.9778, 79.9496, 79.9262, 79.9074, 79.8816, 79.8299, 79.8182, 79.7971, 79.7783, 79.7455, 79.7032, 79.7032, 79.6844, 79.6704, 79.6891, 79.6797, 79.6586, 79.6422,
        79.661, 79.6915, 79.7149, 79.729, 79.7361, 79.7267, 79.7079, 79.6891, 79.7079, 79.722, 79.7126, 79.7126, 79.7361, 79.7619, 79.783, 79.7947, 79.8088, 79.8159, 79.8299, 79.8417, 79.8487, 79.8745,
        79.9074, 79.9215, 79.9285, 79.9543, 79.9708, 79.9848, 79.9989] };
  WCL[1215] = struct {
   x = [94.0306, 93.8335],
   y = [80.0036, 79.9966] };
  WCL[1216] = struct {
   x = [98.0578, 98.1329, 98.1329, 98.1141, 97.9264, 97.6729, 97.6729, 97.6542, 97.4946, 97.3725, 97.1754, 96.9877, 96.8938, 96.9877, 97.1379, 97.3256, 97.4476, 97.5603, 97.7293, 97.8325, 97.8982, 97.9921,
        98.1517, 98.3394, 98.4051, 98.4239, 98.4333, 98.4521],
   y = [80.02, 79.9895, 79.9426, 79.8957, 79.8722, 79.8534, 79.8299, 79.783, 79.7877, 79.7689, 79.7478, 79.7267, 79.7032, 79.7032, 79.6962, 79.7243, 79.7502, 79.7478, 79.7572, 79.7783, 79.8065, 79.837,
        79.8628, 79.8863, 79.898, 79.9215, 79.9543, 79.9895] };
  WCL[1217] = struct {
   x = [98.8628, 98.9379, 99.0975],
   y = [80.0013, 79.9895, 79.9919] };
  WCL[1218] = struct {
   x = [99.2805, 99.5152, 99.534, 99.656, 99.7311, 99.8813, 99.8438, 99.9752, 99.9564, 99.8531, 99.8625],
   y = [80.0106, 79.9825, 79.9426, 79.9027, 79.8698, 79.8299, 79.79, 79.8065, 79.7666, 79.7103, 79.6657] };
  WCL[1219] = struct {
   x = [100.229, 99.9189, 99.8813, 99.8156, 99.7311, 99.7123, 99.6654, 99.6654, 99.6936, 99.6748, 99.6748, 99.5997, 99.4213, 99.2336, 99.074, 98.9707, 99.0083, 99.1209, 99.3274, 99.5058, 99.4776, 99.5903,
        99.7029, 99.7123, 99.6654, 99.6936, 99.4776, 99.2336, 99.2242, 98.9801, 98.783, 98.5577, 98.3418, 98.0695, 97.8818, 97.6565, 97.5814, 97.4594, 97.2622, 97.0839, 96.99, 96.6427, 96.4268,
        96.2202, 96.2296, 96.0419, 95.8541, 95.779, 95.7603, 95.7509, 95.6945, 95.5913, 95.4599, 95.1782, 95.028, 94.9248, 94.8309, 94.7089, 94.6056, 94.3522, 94.2301, 94.2113, 94.2113, 94.2207,
        94.202, 94.1363, 94.1363, 93.9861, 93.9016, 93.7795, 93.742, 93.6669, 93.5167, 93.4416, 93.3853, 93.2163, 93.1975, 93.2069, 93.1412, 93.0755, 92.9441, 92.822, 92.9534, 93.1318, 93.2726, 93.404,
        93.5448, 93.695, 93.7326, 93.8265, 93.8828, 94.0518, 94.2864, 94.4742, 94.4836, 94.324, 94.2395, 94.2958, 94.3522, 94.493, 94.5775],
   y = [79.6539, 79.6352, 79.6023, 79.5483, 79.5202, 79.4779, 79.4709, 79.4263, 79.377, 79.3324, 79.2878, 79.2385, 79.2503, 79.2761, 79.3043, 79.3066, 79.2338, 79.1963, 79.1447, 79.1329, 79.0883, 79.0649,
        79.0203, 78.9569, 78.9288, 78.8701, 78.8349, 78.8232, 78.7715, 78.7739, 78.7481, 78.7621, 78.7668, 78.7715, 78.7715, 78.7715, 78.8138, 78.8396, 78.8584, 78.8677, 78.91, 78.9335, 78.9499,
        78.9546, 78.9921, 79.0015, 78.9851, 79.0109, 79.0555, 79.1095, 79.1587, 79.1235, 79.079, 79.0156, 78.9945, 79.0414, 79.0508, 79.086, 79.1517, 79.1846, 79.2268, 79.2784, 79.3207, 79.3723,
        79.4122, 79.4498, 79.4779, 79.4826, 79.4709, 79.4544, 79.485, 79.499, 79.4826, 79.5108, 79.5389, 79.5249, 79.499, 79.4662, 79.4498, 79.5178, 79.5342, 79.553, 79.5601, 79.5436, 79.5788, 79.6093,
        79.6281, 79.6586, 79.7079, 79.7455, 79.7525, 79.7619, 79.7689, 79.7971, 79.8206, 79.844, 79.898, 79.9191, 79.9496, 79.9708, 79.9848] };
  WCL[1220] = struct {
   x = [99.9423, 99.9705],
   y = [78.4242, 78.4312] };
  WCL[1221] = struct {
   x = [100.064, 99.9423, 99.7921, 99.6325, 99.4448, 99.4636, 99.5574, 99.5293, 99.3321, 99.3791, 99.3979, 99.473, 99.595, 99.717, 99.886, 99.9423, 99.9799],
   y = [77.9689, 77.9267, 77.936, 77.9243, 77.9384, 77.9689, 78.0088, 78.0463, 78.0252, 78.0769, 78.1285, 78.1684, 78.2036, 78.2623, 78.3115, 78.3233, 78.3397] };
  WCL[1222] = struct {
   x = [100.032, 99.9282, 99.9282, 99.947],
   y = [78.3655, 78.3843, 78.4125, 78.4242] };
  WCL[1223] = struct {
   x = [89.9143, 90.0786, 90.0035],
   y = [75.4625, 75.4813, 75.4883] };
  WCL[1224] = struct {
   x = [89.98, 90.1114, 90.332, 90.4635, 90.7638, 90.7967, 90.9047, 91.0502, 91.1722, 91.3036, 91.435, 91.5242, 91.5242, 91.5242, 91.5477, 91.5805, 91.6228, 91.6885, 91.7448, 91.8528, 91.9138, 91.956,
        92.0405, 92.1015, 92.1626, 92.233, 92.3268, 92.4442, 92.4911, 92.5005, 92.4817, 92.4911, 92.4489, 92.416, 92.4207, 92.4113, 92.3738, 92.355, 92.2705, 92.1813, 92.1485, 92.1532, 92.1485, 92.186,
        92.2001, 92.2236, 92.2799, 92.3503, 92.4066, 92.3832, 92.4019, 92.524, 92.6319, 92.6366, 92.5662, 92.5803, 92.6976, 92.7446, 92.8291, 92.8713, 92.9089, 92.9511, 93.0027, 93.0074, 93.0309,
        93.1154, 93.1717, 93.2562, 93.2844, 93.3219, 93.3266, 93.3688, 93.4205, 93.4627, 93.5425, 93.6317, 93.7115, 93.7772, 93.7631, 93.7302, 93.6411, 93.5754, 93.5519, 93.5847, 93.6692, 93.7349,
        93.8241, 93.9039, 93.9462, 93.9602, 94.0541, 94.0447, 93.9884, 93.9602, 93.9884, 94.1104, 94.148, 94.1761, 94.2184, 94.2653, 94.3217, 94.331, 94.331, 94.331, 94.4014, 94.439, 94.4812, 94.5329,
        94.5939, 94.6408, 94.6361, 94.608, 94.5423, 94.5798, 94.6174, 94.6878, 94.7159, 94.7629, 94.7675, 94.7441, 94.7253, 94.7629, 94.838, 94.9037, 94.9928, 95.0633, 95.1196, 95.1149, 95.1759,
        95.2745, 95.2932, 95.2839, 95.2604, 95.2604, 95.3543, 95.4481, 95.4763, 95.4763, 95.5326, 95.5936, 95.6265, 95.6265, 95.6265, 95.6265, 95.6359, 95.6547, 95.664, 95.6547, 95.5936, 95.5045,
        95.4247, 95.4294, 95.3543, 95.2932, 95.2604, 95.2463, 95.19, 95.1947, 95.1665, 95.1618],
   y = [75.5517, 75.5845, 75.5751, 75.5822, 75.6268, 75.6408, 75.6221, 75.6127, 75.6291, 75.6502, 75.6455, 75.6432, 75.6268, 75.6807, 75.7183, 75.7066, 75.7089, 75.7112, 75.7277, 75.7277, 75.7418, 75.7535,
        75.7629, 75.777, 75.7981, 75.8028, 75.8145, 75.8403, 75.8638, 75.8873, 75.9154, 75.9365, 75.9365, 75.9295, 75.899, 75.8873, 75.9131, 75.9671, 75.9483, 75.9201, 75.9365, 75.9858, 76.0187,
        76.0351, 76.0069, 75.9858, 76.0187, 76.0468, 76.0515, 76.0163, 76.0023, 76.0257, 76.0281, 76.0468, 76.0656, 76.0844, 76.0914, 76.0867, 76.0914, 76.075, 76.0445, 76.0656, 76.0844, 76.1008,
        76.1149, 76.1079, 76.0867, 76.0914, 76.0844, 76.075, 76.1055, 76.1219, 76.1149, 76.1219, 76.129, 76.1266, 76.0938, 76.075, 76.0328, 75.9788, 75.9365, 75.906, 75.8826, 75.8568, 75.8732, 75.9365,
        75.9483, 75.96, 75.9905, 75.9647, 75.9577, 75.9225, 75.8896, 75.8638, 75.8568, 75.906, 75.8943, 75.906, 75.9365, 75.9717, 75.9905, 75.9741, 75.9483, 75.9131, 75.9483, 75.953, 75.9741, 75.9694,
        75.9553, 75.9671, 76.0023, 76.0234, 76.0257, 76.0727, 76.075, 76.0797, 76.0562, 76.0586, 76.0867, 76.1079, 76.1337, 76.1478, 76.1618, 76.19, 76.2205, 76.2487, 76.2463, 76.197, 76.2017, 76.2088,
        76.1877, 76.1618, 76.1431, 76.1266, 76.1102, 76.1126, 76.0891, 76.0375, 76.014, 76.0116, 76.0562, 76.0961, 76.1266, 76.1431, 76.1595, 76.183, 76.1994, 76.2369, 76.2557, 76.2487, 76.2557,
        76.2886, 76.3238, 76.3261, 76.3332, 76.3824, 76.4106, 76.4317, 76.4458, 76.4646] };
  WCL[1225] = struct {
   x = [98.8041, 98.8886, 98.9355, 99.0012, 99.1045, 99.2453, 99.3767, 99.5363, 99.7147, 99.8179, 99.8743, 99.8837, 99.9587],
   y = [76.4646, 76.4693, 76.4646, 76.4599, 76.4505, 76.4411, 76.4317, 76.427, 76.4223, 76.4388, 76.4411, 76.4411, 76.4505] };
  WCL[1226] = struct {
   x = [96.3212, 96.3212, 96.2836, 96.2836, 96.3305, 96.3118, 96.3212, 96.1897, 96.0583, 95.9269, 95.7579, 95.6453, 95.5045, 95.4481, 95.3261, 95.2885, 95.2698, 95.3543, 95.373, 95.4387, 95.4575, 95.5138,
        95.6077, 95.711, 95.7767, 95.9175, 96.0114, 96.0301, 96.1334, 96.2742, 96.4244, 96.5558, 96.584, 96.5746, 96.5558, 96.5089, 96.4901, 96.3963, 96.3212],
   y = [76.2745, 76.2721, 76.2487, 76.2158, 76.1853, 76.1595, 76.129, 76.1337, 76.136, 76.1384, 76.1478, 76.1712, 76.1759, 76.1947, 76.2111, 76.2393, 76.251, 76.2651, 76.2487, 76.244, 76.2158, 76.2487,
        76.2557, 76.2956, 76.2745, 76.2674, 76.251, 76.2369, 76.2651, 76.2886, 76.2768, 76.2487, 76.2252, 76.1853, 76.1571, 76.1853, 76.2369, 76.2628, 76.2745] };
  WCL[1227] = struct {
   x = [96.3, 96.3, 96.1968, 96.0841, 95.9996, 95.8682, 95.7368, 95.6711, 95.6711, 95.7556, 95.8776, 95.8682, 95.7368, 95.5772, 95.4364, 95.2393, 95.1736, 95.3331, 95.4833, 95.6335, 95.8682, 96.0278,
        96.1217, 96.2719, 96.3376, 96.3939, 96.3939, 96.3657, 96.3376, 96.3282, 96.3],
   y = [77.0044, 76.9879, 76.9645, 76.9316, 76.8987, 76.8894, 76.8612, 76.8682, 76.8941, 76.8964, 76.9339, 76.9527, 76.9269, 76.9105, 76.9128, 76.8964, 76.9152, 76.948, 76.9973, 77.0044, 77.0466, 77.0677,
        77.0959, 77.1405, 77.117, 77.0888, 77.0489, 77.009, 76.995, 76.9903, 77.0044] };
  WCL[1228] = struct {
   x = [80.3416, 80.4542, 80.5622, 80.5856, 80.5199, 80.4073, 80.3416, 80.4073, 80.4636, 80.5528, 80.5528, 80.6513, 80.7828, 80.8297, 80.7405, 80.7077, 80.6419, 80.5762, 80.5528, 80.6279, 80.7405, 80.8391,
        80.8719, 81.0362, 81.1489, 81.2474, 81.3038, 81.3038, 81.4023, 81.4915, 81.5995, 81.7121, 81.8435, 81.9656, 82.0547, 82.1721, 82.2941, 82.449, 82.6368, 82.7166, 82.801, 82.9325, 83.1765,
        83.3737, 83.6647, 83.9181, 84.0683, 84.1716, 84.3687, 84.5471, 84.7536, 84.8193, 84.8522, 84.9836, 85.1948, 85.4295, 85.4811, 85.6266, 85.758, 85.8472, 85.9693, 86.0772, 86.1993, 86.265,
        86.4856, 86.6639, 86.6733, 86.5747, 86.5419, 86.6311, 86.5747, 86.5841, 86.6968, 86.7953, 86.8376, 86.8939, 86.786, 86.7156, 86.6827, 86.6405, 86.5184, 86.3541, 86.1899, 86.0772, 85.9693,
        85.8566, 85.8472, 85.8895, 86.0021, 86.0772, 86.0678, 86.0209, 86.1429, 86.2415, 86.2744, 86.3072, 86.3776, 86.4433, 86.5419, 86.5841, 86.5184, 86.5184, 86.509, 86.5841, 86.509, 86.4621,
        86.3448, 86.2321, 86.0209, 85.8331, 85.6923, 85.6689, 85.7487, 85.7017, 85.8144, 85.8895, 85.9693, 85.9129, 85.8707, 85.8895, 85.9552, 86.0021, 86.0866, 86.1899, 86.2321, 86.3307, 86.4762,
        86.5184, 86.5841, 86.6498, 86.7719, 86.6639, 86.7156, 86.8282, 86.8845, 86.9174, 86.9502, 87.0488, 87.0911, 87.1568, 87.1145, 87.1708, 87.26, 87.2929, 87.4008, 87.5886, 87.6543, 87.5229,
        87.3586, 87.26, 87.1145, 86.9502, 86.8282, 86.9174, 87.0723, 87.2459, 87.49, 87.6449, 87.7012, 87.7341, 87.7763, 87.8984, 87.9875, 88.0392, 88.0204, 88.0392, 88.241, 88.3255, 88.3959, 88.4381,
        88.5038, 88.6165, 88.6493, 88.7902, 88.9685, 88.9216, 88.823, 88.8465, 88.9685, 89.0671, 89.1891, 89.2783, 89.3769, 89.4754, 89.6069, 89.7289, 89.9166],
   y = [73.1649, 73.1837, 73.2095, 73.2612, 73.2729, 73.2705, 73.3151, 73.3527, 73.3456, 73.3292, 73.3644, 73.3855, 73.4184, 73.4583, 73.4653, 73.4372, 73.4301, 73.4771, 73.5263, 73.5404, 73.5498, 73.5475,
        73.5686, 73.5921, 73.5686, 73.5615, 73.5733, 73.5968, 73.6061, 73.6108, 73.6202, 73.6296, 73.6413, 73.6578, 73.6578, 73.6554, 73.6413, 73.6531, 73.6601, 73.6484, 73.6343, 73.639, 73.646,
        73.6484, 73.6601, 73.6765, 73.7024, 73.7047, 73.7141, 73.7235, 73.754, 73.7516, 73.7164, 73.7047, 73.6953, 73.7516, 73.7939, 73.8127, 73.815, 73.8056, 73.8385, 73.8526, 73.8572, 73.8526,
        73.8619, 73.8854, 73.9183, 73.9347, 73.9582, 73.9746, 74.0098, 74.0286, 74.0215, 74.0074, 74.0215, 74.038, 74.0591, 74.0966, 74.1483, 74.2257, 74.2421, 74.2539, 74.2562, 74.2679, 74.2844,
        74.3102, 74.3595, 74.3806, 74.3853, 74.3853, 74.3994, 74.4158, 74.4346, 74.4393, 74.4557, 74.4627, 74.4674, 74.458, 74.4487, 74.4698, 74.4839, 74.5026, 74.5238, 74.5402, 74.5472, 74.5613,
        74.573, 74.6059, 74.6082, 74.6106, 74.6247, 74.6505, 74.6622, 74.7068, 74.7232, 74.7303, 74.749, 74.7866, 74.8124, 74.8195, 74.8171, 74.8054, 74.796, 74.7819, 74.7326, 74.7162, 74.6974,
        74.6716, 74.6434, 74.62, 74.6481, 74.688, 74.7373, 74.7373, 74.7678, 74.7866, 74.8171, 74.8547, 74.8805, 74.9227, 74.9485, 74.9673, 74.9743, 74.9485, 74.9344, 74.9767, 75.0072, 75.0119,
        75.0213, 75.0166, 75.0424, 75.0541, 75.0799, 75.1011, 75.1269, 75.1245, 75.1152, 75.1081, 75.0823, 75.0682, 75.0729, 75.094, 75.0964, 75.1152, 75.1457, 75.1668, 75.2067, 75.2372, 75.2677,
        75.2935, 75.3029, 75.3076, 75.3428, 75.3475, 75.3686, 75.3944, 75.3944, 75.4155, 75.4179, 75.4273, 75.4695, 75.4414, 75.4179, 75.4038, 75.4367, 75.4625, 75.4625] };
  WCL[1229] = struct {
   x = [90.0035, 89.8908, 89.8345, 89.98],
   y = [75.4883, 75.5024, 75.5235, 75.5517] };
  WCL[1230] = struct {
   x = [89.4966, 89.4684, 89.4684, 89.4402, 89.5059, 89.5153, 89.5904, 89.5717, 89.4872, 89.337, 89.2806, 89.1962, 89.1398, 89.1962, 89.2619, 89.3651, 89.4496, 89.4966],
   y = [77.3047, 77.3024, 77.286, 77.2602, 77.2437, 77.2085, 77.1991, 77.1592, 77.1311, 77.1522, 77.1264, 77.1358, 77.2085, 77.2508, 77.2719, 77.293, 77.3047, 77.3047] };
  WCL[1231] = struct {
   x = [82.1885, 82.2542, 82.273, 82.2308, 82.1087, 82.043, 82.0524, 82.043, 81.9444, 81.921, 81.8318, 81.7238, 81.6112, 81.6018, 81.6581, 81.691, 81.6253, 81.5455, 81.4938, 81.5032, 81.6581, 81.6816,
        81.7661, 81.8553, 81.8881, 81.9867, 81.8224, 81.8553, 82.0759, 82.2401, 82.1557, 82.1885],
   y = [75.4132, 75.4015, 75.3428, 75.3404, 75.3569, 75.2982, 75.2325, 75.1926, 75.2489, 75.3522, 75.3287, 75.3686, 75.3827, 75.3217, 75.2513, 75.202, 75.2442, 75.3193, 75.371, 75.4273, 75.4648, 75.4249,
        75.4038, 75.4343, 75.4601, 75.4977, 75.4836, 75.5094, 75.5165, 75.4883, 75.4578, 75.4132] };
  WCL[1232] = struct {
   x = [82.3199, 82.3199, 82.3528, 82.442, 82.564, 82.6391, 82.7142, 82.7705, 82.8175, 82.7471, 82.6485, 82.5171, 82.4279, 82.3856, 82.3528, 82.3199, 82.3199],
   y = [70.398, 70.4238, 70.4708, 70.5154, 70.5177, 70.452, 70.3792, 70.3088, 70.2408, 70.215, 70.1586, 70.1891, 70.2314, 70.2736, 70.3159, 70.3605, 70.398] };
  WCL[1233] = struct {
   x = [79.9614, 80.182, 80.3791, 80.5903, 80.8438, 80.8438, 80.9189, 81.1536, 81.2944, 81.3272, 81.4164, 81.5384, 81.6699, 81.8341, 81.989, 82.1674, 82.3551, 82.5757, 82.7166, 82.8855, 82.9841, 83.1061,
        83.3033, 83.2798, 83.2704, 83.1155, 83.0733, 83.0592, 83.0169, 82.6743, 82.4302, 82.2988, 82.3317, 82.3551, 82.2988, 82.3129, 82.3223, 82.388, 82.2988, 82.2425, 82.2331, 82.2331, 82.2425,
        82.2425, 82.2894, 82.388, 82.4772, 82.6321, 82.7963, 82.9512, 83.092, 83.2469, 83.3033, 83.2375, 83.1484, 83.1812, 83.3126, 83.5239, 83.599, 83.721, 83.8008, 83.8102, 83.7679, 83.6788, 83.6224,
        83.5896, 83.4488, 83.3033, 83.3361, 83.2704, 83.4018, 83.4253, 83.613, 83.6881, 83.7022, 83.6553, 83.5473, 83.4816, 83.4253, 83.2704, 83.0827, 82.8949, 82.7166, 82.5851, 82.4772, 82.2753,
        82.3458, 82.4302, 82.3223, 82.2002, 82.0547, 81.9139, 81.7168, 81.4821, 81.2287, 81.1066, 80.9752, 80.9001, 80.886, 80.9001, 80.8109, 80.7875, 80.8344, 80.8672, 80.9423, 80.9517, 81.0221,
        80.9189, 80.8672, 80.8438, 80.8203, 80.7546, 80.6466, 80.6654, 80.5809, 80.5105, 80.4589, 80.4448, 80.3416],
   y = [72.198, 72.1441, 72.0995, 72.069, 72.0314, 71.9845, 71.9375, 71.8507, 71.7662, 71.7803, 71.7451, 71.7146, 71.7029, 71.67, 71.6794, 71.6817, 71.6864, 71.7029, 71.724, 71.7451, 71.7216, 71.6958,
        71.6559, 71.5902, 71.5573, 71.5433, 71.4752, 71.4072, 71.3649, 71.3062, 71.2499, 71.1748, 71.1138, 71.0152, 70.9331, 70.8463, 70.7594, 70.7195, 70.6233, 70.5435, 70.4567, 70.3863, 70.33,
        70.2572, 70.1821, 70.1375, 70.114, 70.1117, 70.1023, 70.0812, 70.0648, 70.0483, 70.1, 70.1516, 70.2267, 70.2736, 70.2525, 70.2736, 70.2971, 70.3605, 70.445, 70.5365, 70.6468, 70.7383, 70.8463,
        70.9284, 70.9941, 71.0622, 71.1349, 71.2288, 71.3133, 71.4118, 71.4682, 71.5832, 71.6371, 71.6747, 71.7451, 71.8296, 71.7897, 71.8296, 71.8484, 71.8648, 71.9, 71.9915, 72.0291, 72.0478,
        72.0737, 72.1394, 72.2309, 72.2309, 72.2544, 72.2943, 72.3083, 72.3083, 72.3342, 72.3811, 72.4233, 72.482, 72.5336, 72.5806, 72.6205, 72.7003, 72.7636, 72.7918, 72.8246, 72.8645, 72.8927,
        72.9279, 72.9795, 73.0241, 73.0452, 73.0382, 73.064, 73.0898, 73.111, 73.1274, 73.1368, 73.1368, 73.1649] };
  WCL[1234] = struct {
   x = [70.2431, 70.1446, 70.1774, 70.276, 70.2854, 70.2337, 70.3417, 70.4731, 70.8064, 71.0927, 71.2147, 71.1584, 71.0833, 71.1819, 71.3461, 71.5245, 71.5996, 71.3367, 71.1819, 70.8627, 70.6609, 70.52,
        70.4543, 70.3745, 70.3417, 70.3182, 70.2431],
   y = [73.0147, 73.0218, 73.0546, 73.1344, 73.2048, 73.2799, 73.3433, 73.4348, 73.463, 73.4325, 73.3785, 73.3316, 73.2494, 73.2424, 73.2424, 73.2353, 73.1767, 73.1509, 73.0851, 73.0992, 73.0617, 73.0405,
        73.1203, 73.0711, 73.0405, 73.0147, 73.0147] };
  WCL[1235] = struct {
   x = [74.2022, 74.2586, 74.4134, 74.6341, 74.8547, 74.8781, 74.7326, 74.6575, 74.6575, 74.4698, 74.2257, 74.2022],
   y = [73.0194, 73.0664, 73.0851, 73.064, 73.0382, 73.0194, 73.0053, 72.9068, 72.8411, 72.9021, 72.9678, 73.0194] };
  WCL[1236] = struct {
   x = [79.2761, 79.2667, 79.3794, 79.4967, 79.5295, 79.5295, 79.4451, 79.323, 79.0555, 78.8349, 78.7035, 78.8161, 78.971, 79.079, 79.1775, 79.2432, 79.3089, 79.2761],
   y = [73.0124, 72.9608, 72.9279, 72.8927, 72.7801, 72.7003, 72.6721, 72.7003, 72.705, 72.7566, 72.8129, 72.8904, 72.9443, 73.0053, 73.0476, 73.0405, 73.0218, 73.0124] };
  WCL[1237] = struct {
   x = [77.3681, 77.4667, 77.4901, 77.5887, 77.7201, 77.9314, 78.1848, 78.3162, 78.2693, 78.0862, 77.8844, 77.5887, 77.3212, 77.2038, 77.0255, 76.9692, 77.0818, 77.2226, 77.3681],
   y = [72.5008, 72.5149, 72.5242, 72.5759, 72.5688, 72.543, 72.5266, 72.475, 72.4022, 72.3647, 72.2872, 72.2614, 72.2544, 72.2332, 72.2332, 72.2919, 72.3787, 72.4445, 72.5008] };
  WCL[1238] = struct {
   x = [69.9474, 70.1446, 70.3652, 70.4731, 70.7923, 70.9706, 71.2898, 71.5667, 71.6653, 71.8296, 72.1065, 72.3834, 72.5055, 72.59],
   y = [72.8716, 72.8716, 72.8716, 72.8763, 72.8598, 72.8763, 72.8833, 72.8927, 72.8739, 72.8246, 72.7824, 72.7495, 72.7402, 72.7237] };
  WCL[1239] = struct {
   x = [72.5923, 72.6299, 72.6299, 72.5501, 72.6299, 72.6299, 72.5923, 72.5923, 72.6158, 72.6815, 72.6815, 72.6392, 72.5266, 72.5078, 72.4093, 72.2544, 72.2731, 72.2637, 72.1182, 72.0854, 71.9868, 71.7991,
        71.7897, 71.7662, 71.7568, 71.8789, 72.0009, 72.1886, 72.4093, 72.5594, 72.5407, 72.5172, 72.6486, 72.7707, 72.7707, 72.6815, 72.6956, 72.7284, 72.7143, 72.5735, 72.4844, 72.4093, 72.4421,
        72.5594, 72.5735, 72.5266],
   y = [72.7237, 72.7237, 72.7167, 72.6838, 72.6533, 72.6205, 72.5665, 72.5266, 72.428, 72.3459, 72.3083, 72.2004, 72.1206, 72.0737, 71.9563, 71.8624, 71.8061, 71.7615, 71.7099, 71.6489, 71.6301, 71.5926,
        71.5832, 71.5761, 71.4987, 71.4283, 71.3555, 71.2804, 71.2358, 71.1889, 71.1279, 71.0504, 70.9566, 70.8533, 70.7336, 70.6397, 70.5365, 70.4379, 70.398, 70.3605, 70.33, 70.2572, 70.1774,
        70.1328, 70.0953, 70.0061] };
  WCL[1240] = struct {
   x = [73.6413, 73.7634, 73.7493, 73.7962, 74.0591, 74.0825, 74.2374, 74.3031, 74.3031, 74.1342, 73.8948, 73.8385, 73.7962, 73.6273, 73.4958, 73.3879, 73.1861, 72.9983, 72.9983, 73.1203, 73.2846, 73.355,
        73.3644, 73.3409, 73.5287, 73.6507, 73.8807, 74.1013, 74.2234, 74.2468, 74.2797, 74.5895, 74.8429, 74.9837, 74.9978, 74.9837, 74.8992, 74.7631, 74.7537, 75.0166, 75.1058, 75.3264, 75.3921,
        75.3733, 75.4155, 75.5704, 75.5939, 75.439, 75.4813, 75.5376, 75.6127, 75.5611, 75.439, 75.3592, 75.2043, 75.3592, 75.3029, 75.2513, 75.2607, 75.3592, 75.4719, 75.4813, 75.4578, 75.2184,
        75.1715, 75.2513, 75.3592, 75.5376, 75.7676, 75.9788, 76.1102, 76.2886, 76.4857, 76.7298, 76.9175, 77.171, 77.3353, 77.401, 77.5746, 77.753, 77.8093, 77.9501, 78.1285, 78.4148, 78.4148,
        78.1848, 78.1613, 78.152, 77.9407, 77.8656, 77.9079, 77.7295, 77.6122, 77.5559, 77.5418, 77.4432, 77.2883, 77.1147, 76.8612, 76.6171, 76.4388, 76.3402, 76.298, 76.1994, 76.1196, 76.0023,
        76.1759, 76.3731, 76.5514, 76.8377, 76.8377, 77.2038, 77.4901, 77.8328, 78.0299, 78.105, 77.8656, 77.6544, 77.3681, 77.4338, 77.5981, 77.8328, 77.9407, 78.0956, 78.3819, 78.7246, 79.0109,
        79.3207, 79.5741, 79.8135, 79.959],
   y = [69.9427, 70.0061, 70.114, 70.1962, 70.3088, 70.3558, 70.4825, 70.6022, 70.6609, 70.7876, 70.9214, 71.0105, 71.1138, 71.2288, 71.3086, 71.3579, 71.4259, 71.4869, 71.5409, 71.5785, 71.6371, 71.7029,
        71.7521, 71.8296, 71.8789, 71.9094, 71.9563, 71.9939, 72.0056, 72.0079, 72.0291, 72.0995, 72.184, 72.2684, 72.3787, 72.5196, 72.6322, 72.7402, 72.8035, 72.8458, 72.7918, 72.7331, 72.7003,
        72.6557, 72.5993, 72.5407, 72.5242, 72.4937, 72.3975, 72.3318, 72.2544, 72.2004, 72.1065, 72.0385, 71.9915, 71.9305, 71.8577, 71.7897, 71.7287, 71.6583, 71.6019, 71.4846, 71.4541, 71.44,
        71.4189, 71.379, 71.3344, 71.2875, 71.2523, 71.2358, 71.2171, 71.21, 71.203, 71.1959, 71.1631, 71.1607, 71.1584, 71.1584, 71.142, 71.0903, 71.0293, 70.9824, 70.9495, 70.9401, 70.9965, 71.0856,
        71.1772, 71.264, 71.2382, 71.2804, 71.3696, 71.3086, 71.2922, 71.2804, 71.2781, 71.3156, 71.3227, 71.3931, 71.4283, 71.4869, 71.5409, 71.5503, 71.6278, 71.7146, 71.8296, 71.8718, 71.9469,
        71.9563, 71.8624, 71.7686, 71.6817, 71.7169, 71.785, 71.8249, 71.8953, 71.9751, 72.0385, 72.0525, 72.0455, 72.1534, 72.1769, 72.2074, 72.2684, 72.3248, 72.3412, 72.3388, 72.3342, 72.3318,
        72.3013, 72.2567, 72.198] };
  WCL[1241] = struct {
   x = [59.9335, 60.018, 60.1118, 60.2151, 60.3559, 60.4498, 60.6, 60.6563, 60.5343, 60.4404, 60.6094, 60.7032, 60.8816, 61.0318, 61.1163, 60.9849, 60.9191, 60.9191, 60.9191, 61.0412, 61.1069, 61.0224,
        61.3979, 61.4261, 61.6232, 61.7828, 61.9987, 62.2428, 62.2521, 62.4399, 62.5901, 62.7497, 62.8342, 63.0501, 63.2378, 63.4068, 63.5007, 63.6321, 63.8292, 63.8762, 63.9325, 64.1015, 64.1859,
        64.308, 64.3831, 64.5427, 64.6647, 64.89, 65.059, 65.2373, 65.3124, 65.4063, 65.5753, 65.7161, 65.8381, 65.8193, 65.7724, 65.8475, 65.9132, 66.0634, 66.2418, 66.2793, 66.2512, 66.392, 66.4671,
        66.5891, 66.6642, 66.7956, 66.8989, 67.1711, 67.3495, 67.5654, 67.7719, 67.9315, 68.0723, 68.2694, 68.476, 68.6262, 68.7106, 68.8608, 68.8421, 68.8796, 68.9735, 69.0392, 69.0298, 68.8702,
        68.7951, 68.8984, 69.0486, 69.0016, 68.7951, 68.6825, 68.598, 68.5511, 68.4666, 68.4196, 68.3445, 68.2413, 68.2976, 68.3727, 68.2319, 68.0723, 67.9127, 67.7437, 67.5842, 67.387, 67.2556,
        67.0303, 66.8801, 66.6548, 66.4671, 66.2699, 66.0822, 65.8663, 65.6598, 65.5002, 65.3687, 65.3218, 65.1247, 65.0308, 64.8712, 64.5145, 64.1531, 63.8574, 63.7447, 63.6696, 63.5288, 63.4162,
        63.449, 63.3927, 63.3411, 63.1064, 62.8107, 62.6558, 62.3695, 62.3132, 62.224, 62.1489, 62.0362, 61.9142, 61.7828, 61.6514, 61.4542, 61.2993, 61.1351, 60.9567, 60.8159, 60.7267, 60.6938,
        60.4826, 60.4075, 60.2949, 60.4263, 60.5718, 60.6281, 60.4498, 60.262, 60.1635, 60.2057, 60.1306, 59.9992],
   y = [75.9999, 76.0445, 76.0633, 76.0727, 76.0938, 76.1102, 76.1008, 76.0891, 76.0398, 75.9647, 75.953, 76.0069, 76.0328, 76.0398, 76.0891, 76.1149, 76.1642, 76.2111, 76.2252, 76.2205, 76.2581, 76.2815,
        76.305, 76.2651, 76.2628, 76.251, 76.2557, 76.2463, 76.1571, 76.1384, 76.19, 76.1689, 76.1665, 76.2017, 76.2346, 76.251, 76.2815, 76.298, 76.298, 76.2674, 76.2581, 76.2674, 76.2933, 76.2886,
        76.3214, 76.3285, 76.3637, 76.427, 76.4153, 76.4223, 76.4505, 76.5186, 76.4881, 76.5021, 76.5397, 76.5866, 76.657, 76.7087, 76.7204, 76.7321, 76.7814, 76.7767, 76.772, 76.7814, 76.8049,
        76.8354, 76.8823, 76.8753, 76.8612, 76.9034, 76.9457, 76.9527, 76.9457, 76.9433, 76.9316, 76.9105, 76.8941, 76.8612, 76.8236, 76.7955, 76.7485, 76.718, 76.6688, 76.6312, 76.5796, 76.5538,
        76.481, 76.4787, 76.474, 76.4458, 76.4106, 76.3777, 76.3519, 76.3285, 76.305, 76.2628, 76.2346, 76.197, 76.1712, 76.1454, 76.1642, 76.1806, 76.1806, 76.1665, 76.1313, 76.082, 76.0468, 76.0046,
        75.9999, 75.9671, 75.9671, 75.9459, 75.9154, 75.9272, 75.9084, 75.8966, 75.8685, 75.8215, 75.777, 75.8286, 75.7582, 75.7066, 75.6526, 75.6009, 75.608, 75.6573, 75.6948, 75.676, 75.6268,
        75.6103, 75.5681, 75.5047, 75.4249, 75.4249, 75.3968, 75.3592, 75.3217, 75.3592, 75.3076, 75.2677, 75.2043, 75.155, 75.1011, 75.0494, 75.0307, 75.0189, 75.0189, 74.9908, 75.0283, 75.0236,
        75.0518, 74.9767, 74.9438, 74.9344, 74.8852, 74.8429, 74.796, 74.7584, 74.7138, 74.6646, 74.6599] };
  WCL[1242] = struct {
   x = [59.7668, 60.0203],
   y = [70.1469, 70.0507] };
  WCL[1243] = struct {
   x = [66.859, 66.9106],
   y = [69.8888, 70.0366] };
  WCL[1244] = struct {
   x = [67.211, 67.333, 67.2204, 67.1782, 67.319, 67.3096, 67.319, 67.4175, 67.3002, 67.0327, 66.8918, 66.7792, 66.6384, 66.5257, 66.6149, 66.8027, 66.9435, 67.0327, 67.0327, 67.0327, 66.9763, 66.859,
        66.7792, 66.69, 66.69, 66.737, 66.812, 66.9341, 67.3096, 67.6522, 68.0136, 68.2014, 68.3703, 68.4454, 68.5112, 68.5675, 68.666, 68.7646, 68.8209, 68.8867, 69.0181, 69.173, 69.2621, 69.2621,
        69.3513, 69.3607, 69.5156, 69.7268, 69.8582, 69.7362, 69.9474],
   y = [69.9756, 70.0742, 70.1516, 70.2854, 70.3605, 70.4848, 70.6163, 70.6961, 70.7923, 70.8017, 70.7547, 70.7407, 70.7219, 70.7993, 70.9026, 70.9354, 70.9777, 71.0364, 71.0762, 71.1255, 71.142, 71.0716,
        71.0364, 71.0434, 71.0762, 71.149, 71.2663, 71.2945, 71.3438, 71.4494, 71.5832, 71.6888, 71.785, 71.8883, 72.0314, 72.1816, 72.2473, 72.367, 72.4844, 72.59, 72.6838, 72.7636, 72.8387, 72.8763,
        72.9256, 72.9279, 72.9326, 72.9514, 72.9443, 72.8997, 72.8716] };
  WCL[1245] = struct {
   x = [58.7976, 58.685, 58.5911, 58.5817, 58.5629, 58.5535, 58.5629, 58.638, 58.6756, 58.7507, 58.7413, 58.7788, 58.807, 58.8727, 58.8915, 58.9572, 59.0323, 59.0229, 59.0511, 59.1168, 59.2107, 59.2764,
        59.3515, 59.3796, 59.4735, 59.5204, 59.4923, 59.4735, 59.558, 59.5955, 59.6988, 59.7457, 59.6894, 59.6894, 59.6519, 59.6519],
   y = [80.0013, 79.9966, 79.9919, 79.9731, 79.9684, 79.952, 79.9426, 79.9426, 79.9309, 79.9215, 79.9003, 79.8863, 79.8792, 79.9027, 79.9097, 79.9003, 79.8886, 79.8769, 79.8628, 79.8487, 79.8534, 79.8534,
        79.8558, 79.8698, 79.8675, 79.8745, 79.8957, 79.9003, 79.9027, 79.8933, 79.898, 79.9121, 79.9309, 79.9567, 79.9801, 79.9966] };
  WCL[1246] = struct {
   x = [53.7449, 53.8012, 53.8998, 54.0547, 54.2518, 54.4067, 54.6367, 54.8714, 55.1013, 55.2985, 55.4628, 55.7303, 55.918, 56.1386, 56.3592, 56.5001, 56.547, 56.4015, 56.2607, 56.0495, 55.918, 56.0917,
        56.2935, 56.425, 56.3452, 56.1386, 55.904, 55.7162, 55.9368, 56.148, 56.3029, 56.3264, 56.2466, 56.0917, 55.918, 55.6974, 55.5754, 55.5519, 55.5191, 55.8195, 55.8852, 55.8289, 55.7538, 55.6974,
        55.5848, 55.5191, 55.5848, 55.5754, 55.4534, 55.4628, 55.444, 55.4628, 55.5519, 55.5425, 55.4205, 55.444, 55.5191, 55.5519, 55.5848, 55.6646, 55.7162, 55.8711, 56.026, 56.0823, 56.1058,
        56.3264, 56.5235, 56.7207, 56.8662, 57.0398, 57.1759, 57.3402, 57.4388, 57.6359, 57.5186, 57.5186, 57.4059, 57.3402, 57.1759, 56.9647, 56.8662, 56.9741, 57.1619, 57.2839, 57.1431, 57.1102,
        57.1102, 56.9553, 56.8005, 56.7019, 56.5986, 56.6549, 56.4672, 56.3029, 56.2137, 56.2466, 56.3592, 56.4484, 56.4813, 56.3264, 56.2278, 56.0401, 55.9368, 55.9931, 55.904, 55.7866, 55.7303,
        55.6317, 55.4768, 55.4205, 55.3454, 55.2985, 55.2891, 55.1764, 54.9793, 54.7916, 54.7258, 54.9371, 54.923, 54.6273, 54.5287, 54.3973, 54.2846, 54.2189, 54.0875, 53.9983, 53.9232, 53.7871,
        53.7543, 53.7449, 53.7683, 53.7543, 53.6792, 53.7449, 53.9561, 53.9983, 53.9326, 54.0875, 54.3081, 54.3269, 54.1063, 54.1955, 54.0734, 53.9326, 53.8434, 53.7355, 53.5665, 53.5665, 53.6322,
        53.6557, 53.7543, 53.9889, 53.8998, 53.7449, 53.5665, 53.4586, 53.4586, 53.4116, 53.4257, 53.4586, 53.3037, 53.2708, 53.238, 53.1065, 53.0408, 52.9188, 52.8531, 52.7968, 52.7968, 52.6512,
        52.5527, 52.4964, 52.4776, 52.4306, 52.3884, 52.2992, 52.2758, 52.0551, 51.9472, 51.8815, 51.7594, 51.6609, 51.6139, 51.5295, 51.4966, 51.4825, 51.5154, 51.5295, 51.5952, 51.7688, 51.858,
        51.9707, 52.0786, 52.2429, 52.2335, 52.3321, 52.3978, 52.4306, 52.5104, 52.5198, 52.5198, 52.6419, 52.6512, 52.5996, 52.6512, 52.7404, 52.7404, 52.8202, 52.8953, 53.0033, 53.0408, 52.839,
        52.7639, 52.8202, 52.8625, 52.6982, 52.5621, 52.4119, 52.4306, 52.4776, 52.609, 52.7827, 53.0033, 53.1394, 53.36, 53.191, 53.1065, 53.1722, 53.2943, 53.2473, 53.1488, 53.1253, 53.2239, 53.4586,
        53.6698, 53.7449],
   y = [73.2612, 73.2776, 73.2705, 73.2518, 73.2518, 73.3034, 73.3339, 73.3691, 73.3691, 73.3151, 73.2917, 73.287, 73.2518, 73.1954, 73.1907, 73.1274, 73.0804, 73.0617, 73.0476, 73.0476, 73.0288, 73.0194,
        73.0053, 73.0053, 72.9701, 72.9772, 72.9443, 72.9115, 72.9021, 72.9232, 72.9349, 72.8716, 72.8223, 72.7801, 72.7636, 72.7801, 72.7754, 72.7566, 72.7237, 72.7331, 72.6721, 72.6463, 72.604,
        72.5172, 72.536, 72.5078, 72.4844, 72.4233, 72.4186, 72.3553, 72.2849, 72.2098, 72.1511, 72.0901, 71.9774, 71.9164, 71.8812, 71.7474, 71.6559, 71.548, 71.4916, 71.4212, 71.3015, 71.2523,
        71.1819, 71.0786, 70.9941, 70.9331, 70.9096, 70.8416, 70.8087, 70.7665, 70.7735, 70.7125, 70.6961, 70.6397, 70.621, 70.6092, 70.6233, 70.6327, 70.5998, 70.5881, 70.5811, 70.5623, 70.5506,
        70.5107, 70.4966, 70.5341, 70.5905, 70.635, 70.6562, 70.7289, 70.6961, 70.6961, 70.6609, 70.621, 70.6069, 70.6233, 70.5623, 70.567, 70.5506, 70.6069, 70.6022, 70.5553, 70.5834, 70.6609,
        70.6562, 70.6562, 70.689, 70.6233, 70.6421, 70.6163, 70.5506, 70.5506, 70.6069, 70.6679, 70.7195, 70.7195, 70.7477, 70.7594, 70.7407, 70.7289, 70.7266, 70.7125, 70.7266, 70.7852, 70.7547,
        70.7759, 70.8064, 70.8556, 70.9214, 70.9613, 71.0035, 71.0246, 70.9847, 71.0105, 71.0645, 71.0387, 71.0504, 71.0716, 71.0903, 71.1208, 71.1467, 71.1607, 71.149, 71.1631, 71.2171, 71.264,
        71.257, 71.3133, 71.3438, 71.4494, 71.44, 71.3978, 71.3015, 71.264, 71.3508, 71.4001, 71.4682, 71.5198, 71.4776, 71.4564, 71.447, 71.4823, 71.4259, 71.4494, 71.4189, 71.494, 71.5785, 71.6207,
        71.562, 71.5292, 71.4611, 71.4916, 71.548, 71.5151, 71.4564, 71.4564, 71.4611, 71.4916, 71.4987, 71.5503, 71.6066, 71.6747, 71.7592, 71.8413, 71.9047, 71.9798, 72.0314, 72.083, 72.0995,
        72.0549, 72.0924, 72.0995, 72.0455, 72.015, 72.0619, 72.1206, 72.1323, 72.2098, 72.2544, 72.2567, 72.299, 72.3553, 72.3553, 72.3623, 72.4351, 72.4468, 72.5008, 72.5336, 72.5665, 72.5172,
        72.5548, 72.59, 72.6228, 72.6134, 72.6416, 72.6956, 72.7284, 72.7801, 72.8223, 72.8575, 72.8857, 72.8645, 72.9091, 72.9091, 72.9443, 72.9748, 72.9842, 73.0194, 73.0476, 73.0922, 73.1297,
        73.2072, 73.2353, 73.2612] };
  WCL[1247] = struct {
   x = [53.6698, 53.6698, 53.7543, 53.9749, 54.0969, 54.2424, 54.4818, 54.7587, 54.6695, 54.7681, 54.9465, 55.0685, 55.2797, 55.5332, 55.7866, 56.1058, 55.9274, 55.6834, 55.3783, 55.0779, 55.2234, 55.5519,
        55.6317, 55.3548, 55.1905, 55.2234, 55.5097, 55.8523, 56.0072, 55.8711, 55.6646, 55.5519, 55.4628, 55.4534, 55.5097, 55.6083, 55.7725, 56.0166, 56.1809, 56.4156, 56.6456, 56.4343, 56.3123,
        56.2044, 56.0072, 55.8523, 55.6411, 55.5848, 55.5989, 55.7303, 55.8523, 55.904, 55.8054, 55.796, 55.7162, 55.5519, 55.5425, 55.4299, 55.5519, 55.5003, 55.5989, 55.7303, 55.918, 55.9274,
        56.0729, 56.2701, 56.4672, 56.6362, 56.7019, 56.8662, 57.0211, 57.1525, 57.2604, 57.3496, 57.481, 57.6265, 57.8471, 57.8706, 57.7908, 57.8565, 57.8706, 57.9786, 57.9363, 58.2132, 58.2883,
        58.3963, 58.4526, 58.5653, 58.4996, 58.5418, 58.5747, 58.6404, 58.7249, 58.8046, 58.861, 58.9455, 59.0534, 59.1473, 59.213, 59.2365],
   y = [73.7634, 73.7892, 73.7657, 73.7822, 73.8314, 73.8831, 73.9159, 73.9347, 73.9629, 74.0473, 74.0802, 74.1224, 74.0872, 74.0732, 74.0262, 74.0028, 74.0732, 74.1224, 74.1717, 74.2093, 74.1952, 74.2093,
        74.2609, 74.3055, 74.3313, 74.3642, 74.3782, 74.3782, 74.3994, 74.4416, 74.4557, 74.4932, 74.5449, 74.6059, 74.6223, 74.5871, 74.6223, 74.6481, 74.6317, 74.6247, 74.6646, 74.6528, 74.6716,
        74.7326, 74.7467, 74.7303, 74.7279, 74.7678, 74.7725, 74.8077, 74.8805, 74.9251, 74.9227, 74.9462, 74.9861, 74.9227, 74.9908, 75.0166, 75.0682, 75.0917, 75.1457, 75.1386, 75.141, 75.0659,
        75.0424, 75.0776, 75.0987, 75.1691, 75.2489, 75.263, 75.2912, 75.2818, 75.2841, 75.2771, 75.2653, 75.3052, 75.3639, 75.4179, 75.4672, 75.4883, 75.4836, 75.5, 75.554, 75.6174, 75.608, 75.6033,
        75.6174, 75.6291, 75.6573, 75.6901, 75.7511, 75.791, 75.8145, 75.8356, 75.8591, 75.8779, 75.8873, 75.9131, 75.9483, 75.9999] };
  WCL[1248] = struct {
   x = [60.0766, 59.8748, 59.7105, 59.4477, 59.4899, 59.7105, 59.8232, 59.8889, 59.856, 59.6777, 59.4336, 59.213, 59.0393, 59.0581, 59.0393, 59.1285, 59.213, 59.0159, 58.983, 59.0816, 59.091, 59.0581,
        59.0252, 59.0159, 58.9736, 58.861, 58.7859, 58.5653, 58.2555, 58.2226, 58.4667, 58.5981, 58.6873, 58.7389, 58.753, 58.6638, 58.6169, 58.5981, 58.4855, 58.4432, 58.4526, 58.4996, 58.4198,
        58.2555, 58.2132, 58.2461, 58.2649, 58.2555, 58.1241, 58.0349, 57.9035, 57.8378, 57.7157, 57.5608, 57.5045, 57.4482, 57.4951, 57.5045, 57.4294, 57.2745, 57.3637, 57.5702, 57.6922, 57.7016,
        57.88, 57.7814, 57.8237, 57.8378, 57.8049, 57.7908, 57.6171, 57.528, 57.4059, 57.2417, 56.9413, 56.777, 56.9413, 57.2088, 57.3731, 57.4623, 57.5186, 57.4623, 57.3637, 57.3308, 57.129, 57.0398,
        56.8427, 56.9319, 57.0211, 57.0633, 57.1525, 57.1853, 57.1525, 57.0398, 57.007, 56.9553, 56.8427, 56.8333, 56.8005, 56.7113, 56.5658, 56.4343, 56.3592, 56.2372, 56.1058, 55.918, 55.7631,
        55.566, 55.444, 55.2985, 55.1577, 54.923, 54.7822, 54.585, 54.4818, 54.3738, 54.2846, 54.1063, 54.1438, 54.3175, 54.3973, 54.2753, 54.2846, 54.3738, 54.4959, 54.7024, 54.8385, 54.9793, 55.092,
        55.0779, 54.923, 54.8385, 54.6367, 54.4959, 54.4161, 54.2424, 54.0077, 53.9749, 53.8669, 53.7543, 53.7026, 53.6792, 53.6698],
   y = [74.6599, 74.6458, 74.6951, 74.7209, 74.688, 74.6176, 74.5496, 74.5097, 74.5097, 74.5449, 74.6059, 74.6364, 74.6176, 74.5965, 74.5683, 74.5026, 74.4815, 74.5144, 74.4862, 74.4322, 74.4111, 74.3994,
        74.3829, 74.3782, 74.3736, 74.3736, 74.4228, 74.4698, 74.5331, 74.4862, 74.4205, 74.3524, 74.2985, 74.2609, 74.2069, 74.2116, 74.1764, 74.1295, 74.167, 74.1647, 74.1037, 74.0825, 74.0591,
        74.1037, 74.0473, 74.0145, 73.9746, 73.9769, 73.9629, 74.0051, 74.0145, 74.0802, 74.1506, 74.2187, 74.1647, 74.1224, 74.0661, 74.0591, 74.045, 74.0309, 74.0192, 73.9981, 73.9676, 73.9136,
        73.8901, 73.8666, 73.8338, 73.7915, 73.7469, 73.7235, 73.7211, 73.7939, 73.8103, 73.8197, 73.8314, 73.8197, 73.7822, 73.7728, 73.7071, 73.6625, 73.6038, 73.5686, 73.5569, 73.5263, 73.5662,
        73.6273, 73.646, 73.5592, 73.5005, 73.4841, 73.4583, 73.416, 73.4113, 73.3715, 73.3409, 73.3057, 73.2987, 73.2799, 73.2635, 73.2259, 73.233, 73.2682, 73.2588, 73.2588, 73.2588, 73.3034,
        73.3222, 73.3175, 73.3316, 73.3738, 73.409, 73.402, 73.3926, 73.3691, 73.3433, 73.3222, 73.3222, 73.3316, 73.3503, 73.3808, 73.4254, 73.4278, 73.4888, 73.5287, 73.5592, 73.6061, 73.5991,
        73.632, 73.6554, 73.6718, 73.7047, 73.6672, 73.6554, 73.6132, 73.5874, 73.5897, 73.5991, 73.646, 73.6765, 73.693, 73.7258, 73.754, 73.7634] };
  WCL[1249] = struct {
   x = [52.311, 52.311, 52.3907, 52.5785, 52.7522, 52.9211, 52.9634, 53.1277, 53.184, 53.1934, 53.1746, 53.1277, 53.1511, 53.2074, 53.1417, 53.0056, 52.8742, 52.7662, 52.6207, 52.3814, 52.311],
   y = [71.2875, 71.3086, 71.3156, 71.3555, 71.3227, 71.264, 71.2992, 71.2147, 71.1819, 71.1208, 71.0434, 71.0223, 70.9566, 70.8932, 70.8791, 71.0082, 71.0856, 71.2006, 71.2077, 71.2499, 71.2875] };
  WCL[1250] = struct {
   x = [58.7648, 58.7084, 58.6662, 58.6662, 58.577, 58.5348, 58.699, 58.8774, 58.7882, 58.6333, 58.699, 58.7976, 58.8633, 59.0276, 59.1966, 59.3608, 59.3937, 59.4688, 59.7692],
   y = [69.9685, 70.0272, 70.0835, 70.1258, 70.1446, 70.2079, 70.1821, 70.1962, 70.2197, 70.2736, 70.2854, 70.337, 70.4121, 70.4309, 70.33, 70.2901, 70.2478, 70.2009, 70.1469] };
  WCL[1251] = struct {
   x = [47.6243, 47.6431, 47.6525, 47.7276, 47.7933, 47.8308, 47.9153],
   y = [80.006, 79.9872, 79.9543, 79.9614, 79.9684, 79.9801, 79.9895] };
  WCL[1252] = struct {
   x = [29.9574, 30.0559, 30.2296, 30.3328, 30.1639, 30.3188, 30.6051, 30.6849, 30.76, 30.9712, 31.0134, 30.882, 30.6614, 30.5394, 30.3751, 30.1967, 29.9996],
   y = [70.7054, 70.6421, 70.621, 70.5905, 70.5553, 70.5294, 70.5247, 70.4848, 70.4403, 70.4051, 70.3229, 70.2783, 70.2478, 70.215, 70.1586, 70.1305, 70.0695] };
  WCL[1253] = struct {
   x = [19.988, 20.2321, 20.3072, 20.3541, 20.4855, 20.6263, 20.861, 21.0675, 21.2177, 21.3585, 21.4243, 21.4149, 21.4243, 21.443, 21.4055, 21.3773, 21.3492, 21.2834, 21.1614, 21.0675, 20.9361, 20.739,
        20.6545, 20.5231, 20.4574, 20.2602, 20.1664, 20.1194, 20.0255],
   y = [78.9687, 78.9569, 78.9311, 78.91, 78.9006, 78.8842, 78.8748, 78.8607, 78.8677, 78.8584, 78.849, 78.8161, 78.7856, 78.7574, 78.7293, 78.7011, 78.68, 78.6612, 78.6589, 78.6542, 78.673, 78.68, 78.673,
        78.6495, 78.6354, 78.6448, 78.6354, 78.6143, 78.5885] };
  WCL[1254] = struct {
   x = [21.3891, 21.436, 21.4548, 21.3984, 21.3046, 21.3984, 21.6144, 21.8021, 21.9147, 22.1213, 22.1307, 22.2151, 22.3372, 22.478, 22.3841, 22.2151, 22.187, 22.1588, 22.1213, 22.1307, 22.0931, 21.9335,
        21.7833, 21.6613, 21.5486, 21.3891, 21.1825, 21.145, 21.0323, 20.8915, 20.7883, 20.7601, 20.6568, 20.6475, 20.6099, 20.6381, 20.6381, 20.4034, 20.3189, 20.1781, 20.272, 20.4315, 20.6756,
        20.7695, 20.8915, 21.0136, 21.0511, 21.1074, 21.2764, 21.3891],
   y = [78.6049, 78.6166, 78.6143, 78.5814, 78.5697, 78.5673, 78.5744, 78.5861, 78.5908, 78.5885, 78.565, 78.5345, 78.504, 78.4782, 78.457, 78.4312, 78.3984, 78.3608, 78.3162, 78.2975, 78.2599, 78.2529,
        78.2529, 78.2576, 78.2294, 78.2083, 78.2106, 78.1825, 78.1872, 78.1989, 78.2247, 78.2576, 78.2951, 78.3068, 78.3491, 78.3796, 78.4171, 78.4265, 78.4477, 78.4711, 78.4829, 78.5016, 78.5087,
        78.5368, 78.5204, 78.5415, 78.5814, 78.6002, 78.6143, 78.6049] };
  WCL[1255] = struct {
   x = [22.8887, 22.9356, 22.9356, 22.9262, 22.9262, 22.9356, 22.9826, 23.0013, 23.0201, 23.0764, 23.1328, 23.1703, 23.1703, 23.2266, 23.236, 23.2548, 23.3205, 23.358, 23.358, 23.3017, 23.2923, 23.2923,
        23.2923, 23.3017, 23.3111, 23.2454, 23.1328, 23.1046, 23.0764, 23.0483, 23.0483, 23.0389, 23.0858, 23.114, 23.1703, 23.236, 23.2736, 23.358, 23.3862, 23.4989, 23.574, 23.6303, 23.7148, 23.8086,
        23.9307, 24.0339, 24.109, 24.1372, 24.2029, 24.2498, 24.3249, 24.4282, 24.4658, 24.5127, 24.5221, 24.5127, 24.5315, 24.616, 24.6441, 24.6441, 24.6911, 24.7474, 24.8225, 24.8882, 24.907,
        24.8976, 24.8319, 24.7662, 24.7098, 24.6911, 24.6441, 24.5784, 24.4939, 24.4094, 24.3156, 24.2029, 24.1466, 24.1278, 24.1278, 24.0715, 24.0246, 24.0058, 23.9401, 23.9213, 23.865, 23.8556,
        23.865, 23.818, 23.7711, 23.7523, 23.7148, 23.6303, 23.5927, 23.6115, 23.6115, 23.5176, 23.4613, 23.4519, 23.3956, 23.3674, 23.358, 23.3017, 23.2736, 23.1985, 23.1891, 23.1609, 23.1328,
        23.1328, 23.0577, 22.9732, 22.9262, 22.8793, 22.8511, 22.8511, 22.8699, 22.8699, 22.7948, 22.7573, 22.6915, 22.6634, 22.6634, 22.5977, 22.5789, 22.5507, 22.5038, 22.485, 22.4381, 22.3912,
        22.4663, 22.5038, 22.5977, 22.6165, 22.5789, 22.5695, 22.6165, 22.6728, 22.7291, 22.8042, 22.8042, 22.8324, 22.8699, 22.8324, 22.7291, 22.6822, 22.6634, 22.6352, 22.5977, 22.5789, 22.6071,
        22.6352, 22.6165, 22.5414, 22.4475, 22.3818, 22.2973, 22.2785, 22.2597, 22.194, 22.1377, 22.072, 22.0532, 21.9875, 21.8842, 21.7998, 21.7153, 21.6589, 21.5651, 21.5087, 21.4243, 21.3773,
        21.3116, 21.2459, 21.1896, 21.1614, 21.1145, 21.0582, 20.9924, 20.9267, 20.8516, 20.8516, 20.8704, 20.8516, 20.8422, 20.908, 20.9361, 20.9267, 20.9549, 21.0488, 21.152, 21.2177, 21.2365,
        21.2177, 21.2365, 21.2459, 21.2365, 21.2553, 21.3398, 21.3398, 21.3679, 21.4149, 21.4806, 21.4994, 21.5369, 21.5651, 21.5557, 21.4806, 21.3961, 21.3022, 21.2083, 21.1426, 21.152, 21.1426,
        21.0582, 20.9924, 20.9643, 20.8516, 20.8422, 20.8798, 20.8798, 20.8986, 20.9924, 21.0863, 21.1614, 21.2459, 21.3679, 21.4994, 21.6496, 21.734, 21.781, 21.8467, 21.9781, 22.0626, 22.1283,
        22.194, 22.3067, 22.3724, 22.4475, 22.5414, 22.5507, 22.5507, 22.5789, 22.6915, 22.7291, 22.7197, 22.7573, 22.7854, 22.8136, 22.8324, 22.8605, 22.8887],
   y = [78.281, 78.2623, 78.2482, 78.2482, 78.2388, 78.2294, 78.22, 78.2059, 78.1919, 78.2036, 78.213, 78.213, 78.2083, 78.1989, 78.1895, 78.1707, 78.1613, 78.1449, 78.1379, 78.1285, 78.1238, 78.1191,
        78.0816, 78.098, 78.0839, 78.0909, 78.0933, 78.0675, 78.0487, 78.0393, 78.0158, 77.9994, 77.9806, 77.9572, 77.9501, 77.9478, 77.9243, 77.9102, 77.8868, 77.875, 77.8445, 77.8398, 77.8351,
        77.8117, 77.8211, 77.8445, 77.8516, 77.8797, 77.8868, 77.8727, 77.8445, 77.8422, 77.8234, 77.807, 77.8023, 77.7812, 77.7765, 77.7812, 77.7624, 77.7506, 77.746, 77.7389, 77.7389, 77.7342,
        77.7248, 77.7061, 77.7014, 77.7108, 77.7108, 77.7037, 77.7037, 77.7061, 77.692, 77.6826, 77.6755, 77.6638, 77.6591, 77.6427, 77.6216, 77.6122, 77.6028, 77.5958, 77.5817, 77.5629, 77.5441,
        77.5207, 77.5113, 77.4972, 77.4878, 77.469, 77.4596, 77.462, 77.4526, 77.4221, 77.4174, 77.4197, 77.4057, 77.3822, 77.3822, 77.3775, 77.3752, 77.3705, 77.3611, 77.3681, 77.3845, 77.3939,
        77.3869, 77.3658, 77.3611, 77.3493, 77.3353, 77.3259, 77.3118, 77.2954, 77.2883, 77.2789, 77.2672, 77.2602, 77.2484, 77.2766, 77.2789, 77.2672, 77.2649, 77.2578, 77.2578, 77.2742, 77.2907,
        77.3071, 77.3259, 77.3446, 77.3634, 77.3798, 77.408, 77.4291, 77.4503, 77.4643, 77.4714, 77.4761, 77.4972, 77.5136, 77.5394, 77.5394, 77.5347, 77.5207, 77.4901, 77.4831, 77.4972, 77.5183,
        77.5254, 77.5441, 77.5535, 77.5582, 77.5606, 77.5488, 77.5277, 77.4995, 77.4878, 77.4808, 77.4878, 77.4808, 77.462, 77.462, 77.469, 77.469, 77.4667, 77.462, 77.462, 77.4573, 77.4573, 77.4456,
        77.4291, 77.4338, 77.4385, 77.4526, 77.4432, 77.4385, 77.4315, 77.4479, 77.4596, 77.4667, 77.4972, 77.516, 77.5254, 77.5347, 77.5512, 77.5699, 77.5699, 77.577, 77.5958, 77.6122, 77.631,
        77.6521, 77.6591, 77.6896, 77.7154, 77.7272, 77.7553, 77.7765, 77.7882, 77.7999, 77.8304, 77.8727, 77.8891, 77.9032, 77.9149, 77.9243, 77.9267, 77.9478, 77.9525, 77.9595, 77.99, 78.0018,
        78.0018, 78.0182, 78.037, 78.0463, 78.051, 78.0628, 78.0722, 78.0956, 78.105, 78.1191, 78.1285, 78.1355, 78.1637, 78.1707, 78.1778, 78.1825, 78.1942, 78.1872, 78.1801, 78.1731, 78.1754,
        78.1872, 78.1848, 78.1825, 78.1848, 78.2059, 78.2106, 78.22, 78.22, 78.22, 78.22, 78.2271, 78.2271, 78.2294, 78.2388, 78.2599, 78.281, 78.281] };
  WCL[1256] = struct {
   x = [26.4277, 26.4653, 26.4653, 26.4653, 26.5122, 26.5404, 26.5404, 26.5967, 26.6718, 26.7375, 26.822, 26.8689, 26.8971, 26.9346, 26.9346, 26.8877, 26.8032, 26.775, 26.6718, 26.6436, 26.6342, 26.6155,
        26.5685, 26.484, 26.4465, 26.4183, 26.4183, 26.3995, 26.3526, 26.362, 26.3808, 26.3995, 26.3995, 26.4089, 26.4183, 26.4277],
   y = [78.7035, 78.7058, 78.6988, 78.6941, 78.6894, 78.6589, 78.6542, 78.6542, 78.6284, 78.626, 78.6331, 78.6378, 78.6659, 78.68, 78.687, 78.687, 78.7152, 78.7293, 78.7363, 78.7574, 78.7692, 78.7739,
        78.7903, 78.8091, 78.8067, 78.7997, 78.7762, 78.7762, 78.7668, 78.7645, 78.7481, 78.734, 78.7316, 78.7222, 78.7082, 78.7035] };
  WCL[1257] = struct {
   x = [28.9764, 29.0045, 29.0045, 29.0233, 29.089, 29.1266, 29.2017, 29.2768, 29.2862, 29.3049, 29.4082, 29.4364, 29.5021, 29.549, 29.5584, 29.549, 29.5302, 29.4833, 29.4833, 29.4739, 29.38, 29.3049,
        29.2298, 29.1641, 29.0984, 29.0327, 28.9764, 28.967, 29.0045, 29.0233, 28.9764, 28.8731, 28.7886, 28.6854, 28.6572, 28.6197, 28.6009, 28.5539, 28.4788, 28.4413, 28.4413, 28.4601, 28.4601,
        28.4601, 28.4413, 28.3944, 28.3756, 28.338, 28.2911, 28.2348, 28.1878, 28.1503, 28.0189, 28.0001, 28.0001, 28.0001, 27.9907, 27.9813, 27.9156, 27.8499, 27.7936, 27.7936, 27.7936, 27.803,
        27.7936, 27.7936, 27.8499, 27.925, 28.0001, 28.0376, 28.0376, 28.1034, 28.1503, 28.1784, 28.2723, 28.338, 28.3474, 28.3568, 28.3474, 28.3474, 28.4037, 28.4788, 28.5539, 28.5446, 28.5352,
        28.5539, 28.6103, 28.7135, 28.8074, 28.8543, 28.9013, 28.9388, 28.9764],
   y = [78.8396, 78.8443, 78.8419, 78.8396, 78.8419, 78.8325, 78.8232, 78.8232, 78.8372, 78.8513, 78.8584, 78.863, 78.8654, 78.8701, 78.8748, 78.8936, 78.9029, 78.9006, 78.8936, 78.8865, 78.8748, 78.8748,
        78.8748, 78.8748, 78.8936, 78.8936, 78.8983, 78.9029, 78.91, 78.917, 78.9217, 78.9358, 78.9358, 78.9358, 78.9405, 78.9522, 78.9546, 78.9499, 78.9452, 78.9428, 78.9358, 78.9217, 78.9194,
        78.9147, 78.9147, 78.9123, 78.9053, 78.8936, 78.8936, 78.9006, 78.9029, 78.8959, 78.8724, 78.8724, 78.8748, 78.8701, 78.8349, 78.8372, 78.8513, 78.8584, 78.8584, 78.8513, 78.8302, 78.8255,
        78.8161, 78.8091, 78.8067, 78.8091, 78.8044, 78.8067, 78.8278, 78.8302, 78.8419, 78.8466, 78.8513, 78.863, 78.863, 78.8701, 78.863, 78.8677, 78.8654, 78.856, 78.8654, 78.8795, 78.8842, 78.8865,
        78.8912, 78.8842, 78.8889, 78.8748, 78.8584, 78.849, 78.8396] };
  WCL[1258] = struct {
   x = [27.113, 27.1036, 27.1224, 27.0379, 27.0473, 27.0285, 27.0754, 27.1411, 27.1787, 27.1693, 27.1224, 27.0942, 27.0379, 26.9628, 26.8501, 26.775, 26.7563, 26.6999, 26.5873, 26.5873, 26.6155, 26.5873,
        26.5122, 26.4653, 26.4465, 26.3902, 26.362, 26.3244, 26.2963, 26.2587, 26.2212, 26.1649, 26.1085, 26.0522, 26.0334, 25.9865, 25.9583, 25.8739, 25.8269, 25.8269, 25.8363, 25.9208, 25.949,
        25.902, 25.9114, 25.902, 25.8926, 25.8832, 25.8175, 25.78, 25.7518, 25.7612, 25.78, 25.7518, 25.6955, 25.7049, 25.7049, 25.6579, 25.5922, 25.5265, 25.442, 25.4045, 25.3763, 25.3388, 25.2918,
        25.2449, 25.2074, 25.1604, 25.1416, 25.1229, 25.0665, 25.0196, 25.0196, 24.9633, 24.8976, 24.8694, 24.8412, 24.8225, 24.7849, 24.7004, 24.6253, 24.5315, 24.4376, 24.3719, 24.2968, 24.2405,
        24.2405, 24.2405, 24.2217, 24.2123, 24.156, 24.1372, 24.0903, 24.0527, 24.0246, 23.9964, 23.9495, 23.9401, 23.9025, 23.8274, 23.7523, 23.696, 23.6866, 23.6866, 23.6303, 23.5927, 23.5552,
        23.4519, 23.3862, 23.3768, 23.3862, 23.3674, 23.3111, 23.2454, 23.236, 23.236, 23.2266, 23.1703, 23.1046, 23.0201, 22.9356, 22.9168, 22.945, 22.9544, 22.8887, 22.8324, 22.8042, 22.823, 22.8324,
        22.8136, 22.7854, 22.7009, 22.6728, 22.6446, 22.6165, 22.6071, 22.6352, 22.6915, 22.7009, 22.7197, 22.6822, 22.6446, 22.5977, 22.5507, 22.4944, 22.4099, 22.3067, 22.2316, 22.1659, 22.1189,
        22.0908, 22.0626, 21.9875, 21.9218, 21.8655, 21.781, 21.734, 21.6965, 21.6214, 21.5651, 21.5463, 21.5463, 21.5557, 21.5369, 21.4806, 21.4618, 21.4712, 21.443, 21.3773, 21.3398, 21.3022,
        21.2553, 21.1896, 21.1239, 21.0957, 21.0957, 21.0675, 21.03, 20.9924, 20.9267, 20.8798, 20.8329, 20.7859, 20.7296, 20.7296, 20.739, 20.7202, 20.6827, 20.6545, 20.5982, 20.5137, 20.4292,
        20.3447, 20.3259, 20.2321, 20.1664, 20.1194, 20.0913],
   y = [80.006, 79.9895, 79.9684, 79.9356, 79.9191, 79.9097, 79.9074, 79.8957, 79.8863, 79.8605, 79.8393, 79.8206, 79.8182, 79.8159, 79.8159, 79.8088, 79.7877, 79.7807, 79.7502, 79.7337, 79.7243, 79.7243,
        79.7196, 79.6985, 79.6821, 79.6751, 79.6915, 79.6844, 79.6751, 79.661, 79.6422, 79.6398, 79.6563, 79.6516, 79.6328, 79.6328, 79.6305, 79.6093, 79.5976, 79.5671, 79.5436, 79.5413, 79.5178,
        79.5131, 79.4967, 79.4873, 79.4803, 79.4709, 79.4709, 79.4732, 79.4638, 79.4474, 79.431, 79.4146, 79.4099, 79.3934, 79.3864, 79.377, 79.3747, 79.37, 79.3512, 79.3418, 79.3348, 79.3324, 79.3207,
        79.3207, 79.3324, 79.3324, 79.3207, 79.3136, 79.3113, 79.3183, 79.3324, 79.3348, 79.3465, 79.3512, 79.3512, 79.3488, 79.3418, 79.3254, 79.3066, 79.3019, 79.2949, 79.2925, 79.2902, 79.2761,
        79.2573, 79.2526, 79.2503, 79.2268, 79.2104, 79.1869, 79.1752, 79.1728, 79.1728, 79.1775, 79.1752, 79.1634, 79.1587, 79.1541, 79.147, 79.147, 79.147, 79.1587, 79.1517, 79.147, 79.147, 79.147,
        79.147, 79.1517, 79.1634, 79.1541, 79.1447, 79.147, 79.1541, 79.1634, 79.1658, 79.1705, 79.1611, 79.1658, 79.1775, 79.1893, 79.1963, 79.1986, 79.1986, 79.2104, 79.2268, 79.2432, 79.2503,
        79.2385, 79.2385, 79.2479, 79.255, 79.269, 79.2972, 79.323, 79.3301, 79.3559, 79.3653, 79.3864, 79.3934, 79.3723, 79.3653, 79.3606, 79.3676, 79.37, 79.3723, 79.384, 79.3934, 79.3934, 79.3864,
        79.377, 79.3723, 79.3582, 79.3488, 79.3582, 79.37, 79.3747, 79.37, 79.377, 79.384, 79.3887, 79.3958, 79.4028, 79.4028, 79.3864, 79.3723, 79.3606, 79.3747, 79.3911, 79.3817, 79.3723, 79.37,
        79.3723, 79.37, 79.3535, 79.3488, 79.3512, 79.3441, 79.3418, 79.3395, 79.3418, 79.3582, 79.37, 79.3817, 79.4028, 79.4169, 79.4333, 79.438, 79.4333, 79.4286, 79.431, 79.431, 79.4333, 79.4427,
        79.4544, 79.4709, 79.485] };
  WCL[1259] = struct {
   x = [19.9317, 20.0443, 20.1382, 20.1851, 20.1851, 20.279, 20.3259, 20.4386, 20.5888, 20.6827, 20.7765, 20.8516, 20.8986, 20.9267, 21.03, 21.0957, 21.0675, 21.03, 21.0394, 21.0206, 20.9267, 20.8141,
        20.8047, 20.7671, 20.6827, 20.5982, 20.5512, 20.6076, 20.6263, 20.6169, 20.5794, 20.6169, 20.6733, 20.7671, 20.8704, 20.8704, 20.9361, 20.9737, 21.0206, 21.1145, 21.1708, 21.2553, 21.3116,
        21.3116, 21.3492, 21.5087, 21.6308, 21.7528, 21.8091, 21.8373, 21.8091, 21.8185, 21.8842, 21.9218, 21.9406, 21.9781, 21.9781, 21.9969, 21.9687, 21.8467, 21.7998, 21.8373, 21.9218, 21.9312,
        21.9312, 21.8842, 21.8467, 21.781, 21.7247, 21.6496, 21.612, 21.5557, 21.443, 21.3398, 21.3022, 21.2177, 21.1145, 21.0018, 20.8704, 20.8422, 20.8141, 20.7859, 20.739, 20.5982, 20.4667, 20.3729,
        20.3635, 20.3541, 20.2415, 20.157, 20.1382, 20.11, 20.0255],
   y = [79.614, 79.607, 79.6046, 79.5976, 79.5788, 79.5788, 79.553, 79.5577, 79.5577, 79.553, 79.5648, 79.5601, 79.553, 79.5366, 79.5413, 79.5483, 79.553, 79.5671, 79.5788, 79.5929, 79.5835, 79.5882,
        79.6046, 79.6, 79.6023, 79.614, 79.6187, 79.6281, 79.6398, 79.6445, 79.6586, 79.661, 79.6774, 79.6797, 79.6844, 79.6633, 79.6586, 79.6797, 79.6844, 79.6797, 79.6868, 79.6915, 79.6938, 79.6844,
        79.6774, 79.6797, 79.6751, 79.6797, 79.6821, 79.6915, 79.6962, 79.7079, 79.7103, 79.7126, 79.7103, 79.7173, 79.7431, 79.7525, 79.7572, 79.7502, 79.7595, 79.7689, 79.7736, 79.7947, 79.8088,
        79.7994, 79.7877, 79.7877, 79.7994, 79.7994, 79.7854, 79.776, 79.776, 79.7783, 79.7736, 79.7736, 79.7783, 79.7736, 79.776, 79.776, 79.7783, 79.7807, 79.7595, 79.7642, 79.7642, 79.7642, 79.7595,
        79.7502, 79.7525, 79.7525, 79.7502, 79.7478, 79.722] };
  WCL[1260] = struct {
   x = [22.1283, 22.1846, 22.2973, 22.3254, 22.3536, 22.4381],
   y = [80.0106, 79.9872, 79.9684, 79.9825, 79.9942, 79.9919] };
  WCL[1261] = struct {
   x = [21.321, 21.2553],
   y = [69.8888, 70.0131] };
  WCL[1262] = struct {
   x = [21.9734, 21.9734, 21.8514, 21.6965, 21.5416, 21.5416, 21.5416, 21.6073, 21.6871, 21.7951, 21.8044, 21.8936, 21.9406, 22.0579, 22.0908, 22.1612, 22.2128, 22.2128, 22.3583, 22.4475, 22.3912, 22.5648,
        22.7854, 22.9215, 22.9403, 22.9638, 22.8746, 22.8417, 23.0858, 22.9872],
   y = [69.9897, 70.0624, 70.0624, 70.0882, 70.1211, 70.168, 70.283, 70.2596, 70.1938, 70.1375, 70.2408, 70.2736, 70.2455, 70.2736, 70.2126, 70.2314, 70.1704, 70.1258, 70.107, 70.1187, 70.1962, 70.1938,
        70.1891, 70.1751, 70.1704, 70.1586, 70.1211, 70.0953, 70.0695, 70.0483] };
  WCL[1263] = struct {
   x = [23.229, 23.2524, 23.2994, 23.2524, 23.2759, 23.3651, 23.5059, 23.5951, 23.8251, 24.1114, 24.2804, 24.2146, 24.3461, 24.5995, 24.5667, 24.4212, 24.3132, 24.3789, 24.5667, 24.5854, 24.5995, 24.7309,
        24.7873, 24.9609, 24.9844, 25.0501, 25.1722, 25.1956, 25.3599, 25.5054, 25.6227, 25.7025, 25.8011, 25.6697, 25.5899, 25.4256, 25.2707, 25.2144, 25.1487, 25.1722, 25.1299, 25.1158, 25.1299,
        25.0267, 24.9281, 24.9093, 24.9281, 25.083, 25.2144, 25.3364, 25.4819, 25.327, 25.435, 25.604, 25.7119, 25.8105, 25.834, 26.0123, 26.1766, 26.3315, 26.4394, 26.5615, 26.6413, 26.6741, 26.6741,
        26.66, 26.4958, 26.5286, 26.6741, 26.5849, 26.5615, 26.6929, 26.8243, 27.0355, 27.0262, 27.0919, 27.2233, 27.1998, 27.2468, 27.3688, 27.4533, 27.5002, 27.5565, 27.5894, 27.6739, 27.7865,
        28.0165, 28.1292, 28.2606, 28.2277, 28.4483, 28.4249, 28.2934, 27.9743, 27.8616, 28.0165, 27.7537, 27.918, 28.2043, 28.2841, 28.1151, 27.918, 27.9743, 28.0963, 28.1714, 28.2841, 28.3498,
        28.5704, 28.5234, 28.5375, 28.5704, 28.7346, 28.8802, 28.9975, 29.1665, 29.2087, 29.2838, 29.2979, 29.2744, 29.2838, 29.3167, 29.4387, 29.6734, 29.7156, 29.7156, 29.8377, 29.9597],
   y = [69.9568, 70.0507, 70.1187, 70.168, 70.2032, 70.2337, 70.2971, 70.3628, 70.4473, 70.4473, 70.4473, 70.5294, 70.6022, 70.6468, 70.6749, 70.7219, 70.8064, 70.8017, 70.7735, 70.8345, 70.9214, 70.9214,
        70.8862, 70.9613, 70.9167, 70.8791, 70.8533, 70.7759, 70.7876, 70.8204, 70.8603, 70.8463, 70.8228, 70.7336, 70.7054, 70.635, 70.5553, 70.4966, 70.4473, 70.4051, 70.3276, 70.2924, 70.2525,
        70.2337, 70.1821, 70.1117, 70.0319, 70.0577, 70.1258, 70.1962, 70.283, 70.2267, 70.3417, 70.4121, 70.4966, 70.5365, 70.5506, 70.6538, 70.7407, 70.8204, 70.8932, 70.9167, 70.9073, 70.8345,
        70.7735, 70.7078, 70.6538, 70.635, 70.567, 70.459, 70.3605, 70.4051, 70.452, 70.452, 70.5435, 70.6022, 70.6092, 70.6749, 70.7266, 70.7805, 70.8603, 70.9261, 70.9706, 71.0762, 71.0434, 71.0011,
        71.0575, 71.0575, 71.0223, 70.9636, 70.9636, 70.8674, 70.7852, 70.7993, 70.7477, 70.682, 70.6092, 70.6092, 70.6749, 70.6163, 70.5294, 70.4708, 70.445, 70.4637, 70.459, 70.4262, 70.4191,
        70.4567, 70.5435, 70.6609, 70.7407, 70.8345, 70.8486, 70.8228, 70.8463, 70.8064, 70.7852, 70.7266, 70.6961, 70.6726, 70.6562, 70.6749, 70.7054, 70.6726, 70.628, 70.6468, 70.7054] };
  WCL[1264] = struct {
   x = [30.0019, 29.7391, 29.5185, 29.373, 29.2322, 29.0961, 28.9224, 28.6783, 28.6783, 28.6455, 28.7253, 28.9787, 29.0632, 29.1759],
   y = [70.0695, 70.0765, 70.0929, 70.1023, 70.1258, 70.1211, 70.1516, 70.1751, 70.1469, 70.1258, 70.0929, 70.0695, 70.0436, 70.0014] };
  WCL[1265] = struct {
   x = [19.9997, 20.0654],
   y = [70.0953, 70.0742] };
  WCL[1266] = struct {
   x = [20.6827, 20.7343, 20.7484, 20.7906, 20.7578, 20.7484, 20.7014, 20.6263, 20.6263, 20.5606, 20.4714, 20.4714, 20.5278, 20.6827],
   y = [70.215, 70.2197, 70.1962, 70.1751, 70.1328, 70.0436, 70.0272, 70.0389, 70.1, 70.0624, 70.0882, 70.1469, 70.1962, 70.215] };
  WCL[1267] = struct {
   x = [22.7338, 22.7009, 22.6024, 22.4663, 22.4475, 22.4991, 22.6024, 22.6869, 22.7526, 22.823, 22.9215, 22.9309, 22.8887, 22.7666, 22.7338],
   y = [70.2267, 70.2267, 70.2572, 70.2854, 70.3229, 70.3558, 70.3534, 70.3417, 70.337, 70.3276, 70.2901, 70.2384, 70.2314, 70.2267, 70.2267] };
  WCL[1268] = struct {
   x = [22.3583, 22.4146, 22.4146, 22.4334, 22.4897, 22.5226, 22.5554, 22.5554, 22.5554, 22.5648, 22.6446, 22.7666, 22.776, 22.7666, 22.8746, 22.9215, 23.0436, 23.0952, 23.0858, 22.9872, 22.9638, 22.8887,
        22.7526, 22.6681, 22.5648, 22.4991, 22.3583, 22.302, 22.2457, 22.194, 22.1799, 22.0485, 22.0814, 22.1706, 22.302, 22.3583],
   y = [70.6468, 70.6562, 70.682, 70.6749, 70.6749, 70.7008, 70.689, 70.6796, 70.6679, 70.6233, 70.6397, 70.6468, 70.682, 70.7125, 70.7125, 70.6726, 70.6562, 70.6233, 70.5834, 70.5834, 70.5365, 70.5083,
        70.5365, 70.4919, 70.4848, 70.5154, 70.4825, 70.5224, 70.452, 70.5576, 70.574, 70.5881, 70.6069, 70.6327, 70.621, 70.6468] };
  WCL[1269] = struct {
   x = [23.3839, 23.3322, 23.3322, 23.2431, 23.1633, 23.0647, 22.9896, 22.9661, 22.9896, 22.9661, 22.9333, 23.0084, 23.1116, 23.1633, 23.2431, 23.3322, 23.4167, 23.4965, 23.5622, 23.5857, 23.5059, 23.3839],
   y = [70.3276, 70.337, 70.3088, 70.283, 70.2854, 70.2924, 70.3346, 70.3675, 70.4004, 70.4332, 70.4473, 70.4895, 70.4661, 70.5294, 70.5154, 70.5107, 70.5576, 70.5365, 70.4966, 70.4473, 70.3863, 70.3276] };
  WCL[1270] = struct {
   x = [23.9354, 23.8462, 23.8133, 23.7476, 23.6913, 23.7382, 23.7899, 23.879, 24.0011, 24.0339, 24.0574, 24.0433, 24.0433, 23.9354],
   y = [70.5177, 70.4919, 70.5294, 70.6022, 70.6961, 70.7289, 70.682, 70.689, 70.6538, 70.5905, 70.574, 70.5506, 70.5247, 70.5177] };
  WCL[1271] = struct {
   x = [25.8316, 25.8973, 25.9208, 25.9771, 25.9959, 25.9959, 25.9208, 25.8316, 25.7659, 25.7237, 25.6345, 25.5547, 25.5031, 25.4796, 25.4655, 25.4233, 25.3904, 25.5218, 25.5688, 25.3998, 25.2825, 25.3012,
        25.4561, 25.611, 25.6533, 25.6345, 25.7424, 25.8316],
   y = [71.0622, 71.0246, 71.0082, 70.9894, 70.9683, 70.9425, 70.9542, 70.9613, 70.9331, 70.9167, 70.9354, 70.9261, 70.9331, 70.9331, 70.9354, 70.9284, 70.9566, 70.9824, 71.0176, 71.0223, 71.0082, 71.0434,
        71.0528, 71.0387, 71.0786, 71.1115, 71.0716, 71.0622] };
  WCL[1272] = struct {
   x = [14.7499, 14.9376, 14.9752, 15.0315, 15.0409, 15.0503, 15.116, 15.1629, 15.2005, 15.2286, 15.2568, 15.2756, 15.2849, 15.2943, 15.3225, 15.3413, 15.3694, 15.3694, 15.3694, 15.3694, 15.3694, 15.3882,
        15.4445, 15.4915, 15.4915, 15.4821, 15.4915, 15.4351, 15.3882, 15.4164, 15.4727, 15.4915, 15.5572, 15.6323, 15.7168, 15.8294, 15.8857, 15.9702, 16.0359, 16.0829, 16.1955, 16.3082, 16.4584,
        16.4865, 16.4114, 16.2894, 16.2237, 16.1204, 16.0641, 16.0078, 15.9984, 15.9796, 15.9139, 15.8857, 15.8763, 15.8482, 15.8294, 15.8294, 15.8294, 15.8294, 15.7919, 15.7825, 15.7543, 15.7543,
        15.7449, 15.6792, 15.6698, 15.6698, 15.7355, 15.7355, 15.7637, 15.8763, 16.0359, 15.9984, 15.9327, 15.9702, 15.989, 15.9984],
   y = [79.7408, 79.7196, 79.7196, 79.7196, 79.7079, 79.6915, 79.6821, 79.6633, 79.6422, 79.6164, 79.5953, 79.5648, 79.5342, 79.5108, 79.4756, 79.4451, 79.4216, 79.4169, 79.4169, 79.4052, 79.3887, 79.3582,
        79.3254, 79.2902, 79.2479, 79.2221, 79.1963, 79.1752, 79.1494, 79.1329, 79.1376, 79.1587, 79.1658, 79.1658, 79.1376, 79.093, 79.0578, 79.0344, 78.9992, 78.9616, 78.9311, 78.9123, 78.9006,
        78.917, 78.9569, 79.0086, 79.0484, 79.086, 79.1142, 79.1494, 79.1822, 79.2338, 79.262, 79.3066, 79.3606, 79.3864, 79.4075, 79.4591, 79.5014, 79.5507, 79.5741, 79.6117, 79.6469, 79.6915,
        79.6962, 79.6938, 79.7243, 79.7478, 79.7854, 79.8206, 79.8558, 79.8534, 79.8675, 79.8886, 79.9097, 79.9285, 79.959, 79.9989] };
  WCL[1273] = struct {
   x = [16.5569, 16.5569, 16.5757, 16.5757, 16.5757, 16.5757, 16.5757, 16.679, 16.7353, 16.8573, 16.8667, 16.8198, 16.9512, 17.0638, 17.1389, 17.2516, 17.4018, 17.5426, 17.6834, 17.7773, 17.9087, 17.8712,
        17.8618, 17.7773, 17.7491, 17.7303, 17.6928, 17.5708, 17.5238, 17.5801, 17.6459, 17.674, 17.7116, 17.7022, 17.6459, 17.5614, 17.5708, 17.6459, 17.7303, 17.7585, 17.8054, 17.9369, 18.0495,
        18.0589, 17.965, 17.9838, 18.0026, 18.0589, 18.1152, 18.1997, 18.2748, 18.4532, 18.5095, 18.594, 18.6503, 18.6878, 18.7254, 18.7817, 18.838, 18.8287, 18.7817, 18.7817, 18.716, 18.7254, 18.7629,
        18.9131, 19.0446, 19.1948, 19.3731, 19.5045, 19.636, 19.6454, 19.6735, 19.7017, 19.7674, 19.8425, 19.9833],
   y = [80.0106, 79.9895, 79.959, 79.9379, 79.9097, 79.891, 79.8675, 79.8722, 79.8558, 79.8487, 79.8769, 79.9097, 79.9191, 79.9285, 79.9097, 79.8957, 79.8675, 79.837, 79.8088, 79.7713, 79.7384, 79.7056,
        79.6751, 79.661, 79.6305, 79.5929, 79.5577, 79.5366, 79.5084, 79.5178, 79.499, 79.4709, 79.4451, 79.4216, 79.3887, 79.3653, 79.3324, 79.3606, 79.3794, 79.3864, 79.4146, 79.431, 79.4591,
        79.4638, 79.4897, 79.5272, 79.5507, 79.5741, 79.5906, 79.6046, 79.5741, 79.5718, 79.5507, 79.5131, 79.4756, 79.4427, 79.4216, 79.3934, 79.3723, 79.3348, 79.3136, 79.2902, 79.2432, 79.208,
        79.1728, 79.1752, 79.1658, 79.1705, 79.1611, 79.1282, 79.1189, 79.1142, 79.093, 79.0649, 79.032, 78.9921, 78.9687] };
  WCL[1274] = struct {
   x = [20.0208, 19.927, 19.8237, 19.758, 19.6923, 19.5139, 19.589, 19.5609, 19.5233, 19.3543, 19.2886, 19.1478, 19.0164, 19.0633, 19.0633, 19.0352, 19.0727, 19.007, 19.0164, 19.0446, 18.9695, 18.9695,
        18.9695, 18.9976, 18.9695, 18.9038, 18.8193, 18.6878, 18.5752, 18.425, 18.425, 18.4532, 18.4532, 18.4344, 18.4532, 18.4907, 18.4626, 18.4062, 18.4062, 18.3875, 18.4344, 18.425, 18.3875,
        18.3499, 18.3311, 18.3311, 18.3311, 18.2748, 18.2748, 18.2091, 18.0683, 17.9462, 17.7585, 17.5801, 17.6083, 17.7397, 17.7773, 17.8148, 17.7491, 17.6271, 17.4675, 17.4018, 17.3736, 17.383,
        17.3736, 17.3736, 17.3924, 17.4018, 17.3924, 17.3548, 17.3079, 17.3361, 17.383, 17.3173, 17.3736, 17.3548, 17.2797, 17.1765, 17.0826, 17.0451, 17.0732, 17.0638, 17.0169, 17.1014, 17.1202,
        17.0638, 17.0545, 17.0451, 16.9512, 16.8479, 16.679, 16.5569, 16.5569, 16.5381, 16.4912, 16.4255, 16.3316, 16.3504, 16.3222, 16.3316, 16.3035, 16.1439, 15.9655, 15.7308, 15.5994, 15.5994,
        15.6745, 15.759, 15.9186, 16.05, 16.1814, 16.219, 16.3692, 16.3692, 16.2565, 16.1345, 16.0594, 15.9092, 15.8247, 15.7308, 15.6088, 15.468, 15.299, 15.1958, 15.1958, 15.0549, 14.9705, 14.7827,
        14.6137, 14.4166, 14.3697, 14.304, 14.2289, 14.1725, 14.1444, 14.0881, 14.0223, 14.0505, 14.0974, 14.013, 14.1162, 14.2758, 14.4166, 14.5293, 14.548, 14.5856, 14.7264, 14.7733, 14.7827,
        14.9705, 15.0268, 15.1582, 15.2802, 15.3741, 15.3741, 15.4117, 15.5713, 15.7402, 15.9749, 15.9937, 15.881, 15.759, 15.5713, 15.3835, 15.2427, 15.0456, 14.8954, 14.8484, 14.8954, 14.9986,
        15.1113, 15.2427, 15.3178, 15.4868, 15.5994, 15.6933, 15.7965, 15.8904, 15.9092, 16.0688, 16.0782, 16.1439, 16.2941, 16.3035, 16.3316, 16.4349, 16.51, 16.5569, 16.7071, 16.8292, 16.9324,
        16.8761, 16.7916, 16.8761, 16.8573, 16.8104, 16.6696, 16.632, 16.5381, 16.4067, 16.2565, 16.0688, 15.928, 15.8529, 15.7965, 15.7121, 15.6088, 15.5431, 15.4117, 15.3366, 15.2427, 15.2427,
        15.177, 15.0925, 14.9986, 14.9611, 14.8954, 14.839, 14.7264, 14.6888, 14.6044, 14.6044, 14.5199, 14.4542, 14.3321, 14.2382, 14.1631, 14.0599, 13.8909, 13.7595, 13.7595, 13.797, 13.7407,
        13.7219, 13.7032, 13.6656, 13.6656, 13.6468, 13.6562, 13.6468, 13.7219, 13.8534, 13.9285, 14.0223, 14.0881, 14.2195, 14.2758, 14.1725, 14.2007, 14.3697, 14.548, 14.7546, 14.8672, 15.0456,
        15.0362, 15.1113, 15.1676, 15.2896, 15.3553, 15.4868, 15.6276, 15.6839, 15.5619, 15.6276, 15.7308, 15.8529, 15.928, 16.0594, 16.1345, 16.2565, 16.2847, 16.3504, 16.4255, 16.6132, 16.7634,
        16.801, 16.8855, 16.9887, 17.0545, 17.1108, 17.2797, 17.3642, 17.3079, 17.1483, 16.9512, 16.8949, 16.8949, 16.7728, 16.6602, 16.6039, 16.5475, 16.4255, 16.3786, 16.4067, 16.4724, 16.5006,
        16.632, 16.6977, 16.7916, 16.6883, 16.6696, 16.6132, 16.5381, 16.3786, 16.4067, 16.3786, 16.2659, 16.1063, 16.0688, 15.881, 15.7308, 15.5994, 15.4962, 15.3553, 15.299, 15.2896, 15.4023,
        15.4774, 15.5431, 15.5243, 15.4304, 15.2521, 15.2145, 15.2145, 15.1113, 15.0549, 14.9611, 14.8578, 14.8766, 14.8578, 14.7546, 14.6231, 14.5762, 14.5293, 14.5199, 14.5011, 14.4166, 14.5386,
        14.6419, 14.6888, 14.595, 14.4542, 14.3697, 14.2758, 14.1725, 14.1819, 14.1444, 14.0411, 13.9942, 13.9003, 13.8346, 13.7032, 13.5811, 13.553, 13.5154, 13.4497, 13.3934, 13.3558, 13.2526,
        13.2056, 13.0742, 13.0273, 12.9428, 13.0367, 13.0179, 12.9897, 12.8771, 12.7926, 12.6424, 12.6236, 12.5955, 12.4453, 12.4734, 12.6893, 12.8208, 12.971, 13.1587, 13.1963, 12.9897, 12.8489,
        12.6893, 12.5673, 12.3983, 12.22, 12.0322, 12.0041, 11.882, 11.9102, 11.8069, 11.6567, 11.6755, 11.7506, 11.8726, 11.9102, 11.9665, 11.9383, 11.7882, 11.591, 11.4596, 11.3469, 11.5629, 11.6849,
        11.9383, 12.1355, 12.3796, 12.4171, 12.3983, 12.3232, 12.1261, 11.9665, 11.9477, 11.8445, 11.6567, 11.7037, 11.8539, 11.8069, 11.8539, 12.0041, 11.9759, 12.0228, 12.0228, 11.882, 11.7788,
        11.7694, 11.6755, 11.469, 11.3657, 11.5065, 11.6286, 11.4502, 11.2718, 11.178, 11.1592, 11.1217, 11.1123, 10.9996, 10.9996, 10.9808, 10.9339, 10.887, 10.887, 10.8306, 10.8119, 10.8119, 10.8119,
        10.7743, 10.6898, 10.6711, 10.8119, 10.9527, 11.009, 10.8964, 10.8025, 10.7743, 10.7931, 10.9057, 10.9902, 11.0559, 11.2718, 11.3094, 11.2061, 11.131, 11.1404, 11.2061, 11.3, 11.4033, 11.4971,
        11.4596, 11.4596, 11.5159, 11.5347, 11.638, 11.7037, 11.7506, 11.8445, 11.9477, 12.0134, 12.0792, 12.1543, 12.2106, 12.2012, 12.1543, 12.1355, 12.1355, 12.2857, 12.3232, 12.3232, 12.464,
        12.4828, 12.6236, 12.7269, 12.8959, 12.971, 13.1493, 13.262, 13.4779, 13.5905, 13.7595, 13.797, 13.8628, 13.8534, 13.7689, 13.6656, 13.4591, 13.3371, 13.2807, 13.0179, 12.9616, 12.7269, 12.633,
        12.4359, 12.3702, 12.464, 12.5955, 12.7269, 12.9428, 13.1024, 13.262, 13.4497, 13.3934, 13.3558, 13.2056, 13.2714, 13.2338, 13.3183, 13.4215, 13.4591, 13.6562, 13.7407, 13.8721, 14.0974,
        14.2289, 14.1162, 14.0317, 14.013, 13.9472, 13.8721, 13.8721, 13.844, 13.9003, 14.0881, 14.0974, 14.2664, 14.304, 14.3415, 14.4542, 14.5199, 14.6419, 14.7452],
   y = [78.5885, 78.5673, 78.5744, 78.5791, 78.572, 78.565, 78.5345, 78.5063, 78.4594, 78.4617, 78.4312, 78.4336, 78.4148, 78.3749, 78.3186, 78.2763, 78.2505, 78.2458, 78.2271, 78.2059, 78.1731, 78.1402,
        78.098, 78.051, 78.0088, 77.9971, 78.0088, 78.0252, 78.0158, 78.0018, 77.9759, 77.9548, 77.9196, 77.8821, 77.8304, 77.807, 77.7765, 77.7319, 77.7084, 77.6709, 77.6732, 77.6427, 77.5981,
        77.5723, 77.5699, 77.5559, 77.53, 77.5042, 77.469, 77.462, 77.4831, 77.469, 77.4643, 77.4362, 77.401, 77.4104, 77.3892, 77.3634, 77.3564, 77.3564, 77.3118, 77.2766, 77.2414, 77.2156, 77.1898,
        77.1522, 77.1381, 77.117, 77.0935, 77.063, 77.0325, 77.0208, 76.995, 76.9645, 76.9551, 76.9339, 76.8964, 76.8706, 76.8518, 76.8213, 76.8002, 76.7884, 76.7556, 76.7087, 76.6758, 76.6641,
        76.6406, 76.603, 76.5796, 76.5561, 76.5538, 76.5866, 76.6148, 76.6101, 76.5913, 76.5608, 76.5561, 76.603, 76.6289, 76.657, 76.6899, 76.7298, 76.7556, 76.8049, 76.8377, 76.8823, 76.8987,
        76.9105, 76.9222, 76.9152, 76.9246, 76.8917, 76.9504, 76.9809, 76.9762, 77.0114, 77.0114, 77.0067, 76.9856, 76.9973, 76.9832, 76.9809, 76.9879, 77.002, 77.0536, 77.0959, 77.0912, 77.1147,
        77.1358, 77.1522, 77.1921, 77.225, 77.239, 77.2719, 77.3141, 77.3306, 77.3705, 77.3939, 77.4338, 77.4948, 77.5113, 77.5606, 77.5559, 77.5488, 77.5136, 77.4573, 77.4643, 77.4925, 77.5207,
        77.523, 77.5066, 77.5019, 77.5089, 77.516, 77.5113, 77.5019, 77.4948, 77.4878, 77.4714, 77.4995, 77.5441, 77.5441, 77.5699, 77.577, 77.577, 77.5864, 77.6169, 77.6474, 77.6568, 77.6474, 77.6638,
        77.6779, 77.6849, 77.6943, 77.7154, 77.7108, 77.7248, 77.7108, 77.7178, 77.7295, 77.753, 77.7506, 77.753, 77.7812, 77.7999, 77.7647, 77.7718, 77.7858, 77.7835, 77.7671, 77.7835, 77.8375,
        77.8656, 77.8609, 77.9008, 77.9032, 77.8891, 77.8445, 77.8422, 77.8164, 77.8281, 77.8211, 77.8023, 77.7999, 77.7999, 77.8211, 77.8281, 77.8257, 77.8187, 77.7999, 77.7788, 77.7741, 77.7718,
        77.7624, 77.76, 77.753, 77.7694, 77.7389, 77.7342, 77.7671, 77.7624, 77.7389, 77.7413, 77.7319, 77.7319, 77.7577, 77.7366, 77.7108, 77.7108, 77.7248, 77.7483, 77.7765, 77.807, 77.8257, 77.8609,
        77.875, 77.9032, 77.9478, 77.983, 78.0065, 78.0299, 78.0417, 78.044, 78.0135, 77.9642, 77.9501, 77.9806, 78.0229, 78.051, 78.0651, 78.0675, 78.0816, 78.0675, 78.0769, 78.1379, 78.1449, 78.1801,
        78.1919, 78.2271, 78.2247, 78.1989, 78.2294, 78.2857, 78.2928, 78.3139, 78.3421, 78.3374, 78.3421, 78.3491, 78.3514, 78.3397, 78.3327, 78.3209, 78.3186, 78.3186, 78.3421, 78.3444, 78.3561,
        78.3819, 78.3773, 78.3866, 78.4101, 78.4336, 78.4148, 78.3937, 78.3866, 78.3843, 78.396, 78.4336, 78.457, 78.4195, 78.4312, 78.4594, 78.4969, 78.5181, 78.558, 78.5838, 78.6166, 78.6518,
        78.6589, 78.7105, 78.7199, 78.687, 78.6659, 78.619, 78.565, 78.5251, 78.5134, 78.4735, 78.4594, 78.4477, 78.4265, 78.4641, 78.4876, 78.565, 78.5744, 78.6284, 78.673, 78.7152, 78.7739, 78.8232,
        78.7621, 78.6964, 78.6542, 78.626, 78.5955, 78.5955, 78.6331, 78.673, 78.7152, 78.7222, 78.7011, 78.6659, 78.5979, 78.5627, 78.4922, 78.4453, 78.4383, 78.4125, 78.3679, 78.3514, 78.3702,
        78.3632, 78.3655, 78.3045, 78.2834, 78.2294, 78.2271, 78.2247, 78.2271, 78.1848, 78.1895, 78.2059, 78.1895, 78.1872, 78.22, 78.2036, 78.1919, 78.213, 78.1965, 78.22, 78.2083, 78.2576, 78.281,
        78.3068, 78.3514, 78.3374, 78.3514, 78.3561, 78.3749, 78.3984, 78.4453, 78.4735, 78.4735, 78.4876, 78.4782, 78.5016, 78.5533, 78.5251, 78.5275, 78.5275, 78.5392, 78.5509, 78.5791, 78.5979,
        78.6237, 78.6518, 78.68, 78.7035, 78.7199, 78.7387, 78.795, 78.7973, 78.802, 78.8255, 78.8278, 78.8396, 78.8889, 78.91, 78.9499, 78.9522, 78.9147, 78.91, 78.8936, 78.8795, 78.917, 78.9335,
        78.9663, 78.9428, 78.9569, 78.9992, 79.0062, 79.0344, 79.0766, 79.0954, 79.1306, 79.1681, 79.194, 79.2198, 79.255, 79.269, 79.262, 79.2338, 79.2151, 79.2479, 79.2644, 79.2644, 79.2268, 79.1963,
        79.1095, 79.086, 79.1048, 79.1611, 79.1846, 79.2245, 79.2784, 79.2268, 79.2878, 79.323, 79.3043, 79.2996, 79.2949, 79.3418, 79.3864, 79.4286, 79.4685, 79.485, 79.5178, 79.5272, 79.5084,
        79.5436, 79.5835, 79.6187, 79.6563, 79.6844, 79.6821, 79.6469, 79.614, 79.607, 79.6281, 79.6774, 79.7103, 79.7431, 79.7548, 79.7431, 79.7243, 79.7548, 79.776, 79.7877, 79.7971, 79.8041,
        79.8159, 79.79, 79.7595, 79.7243, 79.6962, 79.6657, 79.6868, 79.6633, 79.668, 79.7056, 79.7689, 79.8018, 79.8206, 79.8276, 79.7994, 79.7642, 79.7525, 79.7361, 79.7384, 79.7502, 79.7619, 79.776,
        79.7924, 79.8159, 79.8229, 79.8346, 79.8276, 79.7877, 79.7595, 79.7243, 79.6938, 79.6727, 79.6633, 79.6938, 79.6727, 79.6704, 79.6422, 79.5859, 79.5671, 79.5648, 79.5342, 79.5202, 79.5178,
        79.5389, 79.553, 79.5624, 79.5648, 79.553, 79.5342, 79.5084, 79.485, 79.4474, 79.4052, 79.3723, 79.4169, 79.4568, 79.4028, 79.3277, 79.2714, 79.2104, 79.201, 79.2385, 79.2949, 79.323, 79.3864,
        79.4451, 79.4873, 79.5272, 79.5741, 79.6117, 79.6352, 79.6704, 79.722, 79.7408, 79.7807, 79.7736, 79.7595, 79.7408] };
  WCL[1275] = struct {
   x = [10.5138, 10.5138, 10.6265, 10.7579, 10.8612, 10.8705, 10.9926, 11.1428, 11.1615, 11.1615, 11.1991, 11.2179, 11.2836, 11.3399, 11.3024, 11.3681, 11.4338, 11.4432, 11.6309, 11.7436, 11.8374, 11.875,
        11.8562, 11.8844, 11.9407, 12.044, 12.0909, 12.0721, 12.1003, 12.1378, 12.0252, 11.9125, 11.8562, 11.8187, 11.8093, 11.7717, 11.6497, 11.6028, 11.5934, 11.4432, 11.4056, 11.3399, 11.2648,
        11.1615, 11.0771, 11.0207, 10.8893, 10.8799, 10.833, 10.7579, 10.6734, 10.5701, 10.5044, 10.5138, 10.5138, 10.5138],
   y = [78.8748, 78.8842, 78.8654, 78.8513, 78.8278, 78.7973, 78.7621, 78.7269, 78.7011, 78.6659, 78.6166, 78.5814, 78.572, 78.5603, 78.5228, 78.5251, 78.5087, 78.4876, 78.4547, 78.4195, 78.4171, 78.3984,
        78.3866, 78.3538, 78.3115, 78.2951, 78.2716, 78.2505, 78.22, 78.1965, 78.2012, 78.2036, 78.22, 78.2576, 78.2904, 78.3139, 78.3374, 78.3444, 78.3679, 78.3984, 78.4148, 78.4359, 78.4312, 78.443,
        78.4641, 78.4641, 78.5298, 78.5627, 78.6166, 78.6706, 78.7246, 78.7316, 78.7574, 78.802, 78.8466, 78.8748] };
  WCL[1276] = struct {
   x = [18.885, 18.8944, 18.8944, 18.8944, 18.9038, 18.9601, 19.0258, 19.0164, 18.9742, 19.0164, 19.0727, 19.1384, 19.2042, 19.2135, 19.2135, 19.237, 19.2699, 19.2699, 19.2276, 19.1713, 19.0915, 19.0493,
        19.0399, 19.0258, 18.9601, 18.9178, 18.8709, 18.8709, 18.885],
   y = [74.4533, 74.4533, 74.4487, 74.4369, 74.4228, 74.4041, 74.3947, 74.3736, 74.3477, 74.3454, 74.3384, 74.3595, 74.3736, 74.3947, 74.4228, 74.4299, 74.4604, 74.4745, 74.4768, 74.4979, 74.5191, 74.5191,
        74.4979, 74.4862, 74.4862, 74.4932, 74.4839, 74.4698, 74.4533] };
  WCL[1277] = struct {
   x = [20.0866, 19.9833, 19.9082, 19.8894, 19.8519, 19.7956, 19.7298, 19.6923, 19.6923, 19.7017, 19.7768, 19.8519, 19.9176, 19.927],
   y = [79.485, 79.4873, 79.5014, 79.5155, 79.5295, 79.5389, 79.553, 79.5765, 79.5835, 79.5882, 79.5953, 79.5929, 79.607, 79.614] };
  WCL[1278] = struct {
   x = [20.0208, 19.927, 19.8331, 19.7392, 19.7111, 19.7111, 19.6641, 19.5703, 19.5139, 19.4388, 19.4388, 19.3543, 19.298, 19.2323, 19.1666, 19.1291, 19.0727, 18.9976, 18.9413, 18.8756, 18.8474, 18.8099,
        18.7911, 18.7629, 18.6785, 18.7066, 18.7442, 18.6785, 18.6503, 18.6597, 18.6221, 18.5564, 18.5283, 18.4907, 18.425, 18.425, 18.425, 18.3687, 18.303, 18.3124, 18.3124, 18.3124, 18.2748, 18.1809,
        18.1528, 18.1903, 18.2748, 18.3781, 18.4532, 18.5564, 18.5658, 18.5001, 18.4719, 18.5095, 18.5658, 18.6972, 18.7254, 18.7066, 18.7254],
   y = [79.722, 79.7196, 79.7196, 79.7267, 79.7267, 79.7126, 79.7079, 79.6985, 79.6844, 79.6891, 79.7079, 79.7173, 79.7103, 79.7173, 79.7126, 79.7103, 79.7103, 79.7126, 79.7173, 79.7103, 79.7079, 79.7103,
        79.7103, 79.7196, 79.729, 79.7384, 79.7525, 79.7572, 79.7713, 79.783, 79.7807, 79.7666, 79.7455, 79.7689, 79.783, 79.7877, 79.8041, 79.8112, 79.8276, 79.8323, 79.8487, 79.8675, 79.8675,
        79.8722, 79.8886, 79.8933, 79.9097, 79.9309, 79.9285, 79.9332, 79.9426, 79.9496, 79.9614, 79.9661, 79.9614, 79.9449, 79.9614, 79.9801, 79.9895] };
  WCL[1279] = struct {
   x = [18.8639, 18.8639],
   y = [70.0202, 70.0202] };
  WCL[1280] = struct {
   x = [18.7864, 18.7301, 18.6832, 18.7395, 18.8521, 19.0399, 19.115, 19.1478],
   y = [69.9897, 70.0061, 70.0554, 70.0812, 70.0624, 70.0624, 70.0554, 70.0319] };
  WCL[1281] = struct {
   x = [19.2159, 19.3473],
   y = [69.9873, 70.0061] };
  WCL[1282] = struct {
   x = [18.885, 18.9178, 18.8615],
   y = [69.9756, 70.0178, 70.0202] };
  WCL[1283] = struct {
   x = [19.8214, 19.8542, 19.8777, 19.9434, 19.9997],
   y = [70.1586, 70.168, 70.1399, 70.1375, 70.0953] };
  WCL[1284] = struct {
   x = [20.0654, 19.9903, 19.8448, 19.7228, 19.704, 19.6242, 19.5491, 19.6008, 19.6899, 19.7134, 19.7463, 19.8214],
   y = [70.0742, 70.0202, 70.0389, 70.0624, 70.0929, 70.1751, 70.2314, 70.2384, 70.2197, 70.1938, 70.1399, 70.1586] };
  WCL[1285] = struct {
   x = [-14.2054, -14.25, -14.25, -14.2289, -14.2054, -14.2054],
   y = [-7.91822, -7.87363, -7.92761, -7.96281, -7.95108, -7.91822] };
  WCL[1286] = struct {
   x = [-36.1577, -36.0357, -35.9371, -35.8925, -35.7493, -35.5616, -35.3856, -35.1884, -35.0547, -35.0101, -34.9444, -34.8669, -34.8341, -34.7801, -34.8341, -34.8787, -34.989, -35.0993, -35.1884, -35.409,
        -35.6508, -35.9042, -36.2469, -36.4229, -36.6435, -36.8735, -37.073, -37.2701, -37.6573, -37.899, -38.1525, -38.1642, -38.2534, -38.5397, -38.9035, -39.2227, -39.3869, -39.7413, -39.983],
   y = [-10.1501, -9.9764, -9.88722, -9.85671, -9.72528, -9.51642, -9.29816, -8.94848, -8.70676, -8.46503, -8.2679, -8.06138, -7.81966, -7.45589, -7.10387, -6.81756, -6.41155, -6.01494, -5.58547, -5.30854,
        -5.15365, -5.12079, -5.15365, -5.14192, -5.17712, -4.99876, -5.02223, -4.76877, -4.56929, -4.35807, -4.12574, -4.12574, -4.01544, -3.76198, -3.49678, -3.28557, -3.16353, -2.9664, -2.88661] };
  WCL[1287] = struct {
   x = [-50.0486, -49.9172],
   y = [0.037549, -0.018775] };
  WCL[1288] = struct {
   x = [-39.983, -40.2482, -40.4125, -40.8326, -41.1635, -41.5812, -42.0013, -42.4519, -42.7077, -42.9048, -43.1489, -43.3015, -43.3132, -43.3249, -43.5009, -43.5455, -43.5784, -43.7661, -43.8647, -43.9844,
        -44.1721, -44.2496, -44.3059, -44.3505, -44.3716, -44.4256, -44.4936, -44.4936, -44.3505, -44.2824, -44.327, -44.4162, -44.503, -44.6039, -44.6908, -44.5594, -44.4936, -44.6579, -44.7142,
        -44.7471, -44.8668, -44.8668, -45.0334, -45.1766, -45.3409, -45.3409, -45.3643, -45.5732, -45.7821, -45.8478, -45.8924, -45.9933, -46.1247, -46.4439, -46.6316, -46.8851, -47.0939, -47.1385,
        -47.2606, -47.2817, -47.4037, -47.6243, -47.7346, -47.9435, -47.967, -47.9881, -47.9998, -48.0209, -48.1101, -48.143, -48.197, -48.2861, -48.3401, -48.3518, -48.4504, -48.5279, -48.6053,
        -48.7273, -48.9808, -49.0137, -49.0582, -49.1568, -49.2014, -49.2343, -49.2788, -49.3117, -49.3657, -49.3657, -49.3328, -49.2671, -49.2225, -49.2225, -49.2788, -49.3328, -49.2014, -48.9573,
        -48.8048, -48.6499, -48.6053, -48.4621, -48.4504, -48.3847, -48.373, -48.6827, -48.9479, -49.0911, -49.1122, -49.124, -49.2225, -49.4994, -49.7858, -49.9383, -49.95],
   y = [-2.88661, -2.88661, -2.88661, -2.91008, -2.9664, -2.9312, -2.8209, -2.73172, -2.66601, -2.58856, -2.46652, -2.47826, -2.48999, -2.47826, -2.45479, -2.40081, -2.3351, -2.53224, -2.60029, -2.65427,
        -2.74345, -2.75518, -2.76692, -2.77631, -2.78804, -2.80916, -2.80916, -2.60029, -2.43367, -2.32337, -2.25766, -2.17786, -2.17786, -2.20133, -2.11215, -2.09103, -1.99012, -1.75778, -1.58177,
        -1.51371, -1.51371, -1.51371, -1.54656, -1.62401, -1.72492, -1.49259, -1.3377, -1.22505, -1.20393, -1.1476, -1.09363, -1.12648, -1.00445, -1.02557, -0.938734, -0.83782, -0.74864, -0.694663,
        -0.605483, -0.605483, -0.661808, -0.638339, -0.650073, -0.694663, -0.715785, -0.715785, -0.79323, -0.861288, -0.870676, -1.08189, -1.08189, -1.08189, -1.26025, -1.448, -1.57003, -1.54656,
        -1.53718, -1.59115, -1.73666, -1.8141, -1.93614, -2.17786, -2.26704, -2.34449, -2.51111, -2.5557, -2.41255, -2.24592, -2.10042, -1.9244, -1.84696, -1.8141, -1.80237, -1.72492, -1.65921,
        -1.52544, -1.42453, -1.24852, -1.11475, -0.83782, -0.528038, -0.328557, -0.295701, -0.218256, -0.1854, -0.218256, -0.251111, -0.251111, -0.307435, -0.283967, -0.206521, -0.206521, -0.206521] };
  WCL[1289] = struct {
   x = [-49.95, -50.0064, -50.0157, -50.0721, -50.1941, -50.2809, -50.3021, -50.3373, -50.3373, -50.3912, -50.5227, -50.6658, -50.8324, -51.0202, -51.1962, -51.3182, -51.3394, -51.2619, -51.2408, -51.3511,
        -51.3628, -51.2525, -51.0413, -50.931],
   y = [-0.206521, -0.239377, -0.239377, -0.328557, -0.295701, -0.373147, -0.605483, -0.826086, -0.938734, -1.02557, -1.06077, -0.983324, -1.02557, -1.12648, -1.19219, -1.21566, -1.07016, -0.894144,
        -0.706397, -0.617218, -0.438858, -0.262846, -0.105608, -0.007041] };
  WCL[1290] = struct {
   x = [-80.0177, -79.9285, -79.8957, -79.9074, -79.9074, -79.9074, -79.8628, -79.7854, -79.7619, -79.7525, -79.7971, -79.8182, -79.8628, -79.8957, -79.9285, -79.9825],
   y = [-2.29051, -2.1896, -2.17786, -2.25766, -2.35622, -2.52285, -2.5557, -2.48999, -2.47826, -2.5557, -2.66601, -2.85375, -2.9664, -3.10956, -3.15415, -3.24098] };
  WCL[1291] = struct {
   x = [-80.0388, -79.8182, -79.7079, -79.6187, -79.5765, -79.4873, -79.323, -79.1447, -78.9687, -78.9147, -78.8044, -78.8044, -78.7481, -78.6612, -78.5932, -78.5392, -78.5275, -78.4171, -78.3749, -78.3186,
        -78.274, -78.2294],
   y = [-6.80817, -7.07101, -7.16958, -7.36906, -7.53334, -7.70935, -7.93934, -8.12709, -8.23504, -8.32422, -8.45565, -8.52136, -8.73023, -8.90389, -9.07991, -9.155, -9.16674, -9.4507, -9.58213, -9.73467,
        -9.84497, -9.98578] };
  WCL[1292] = struct {
   x = [-89.3557, -89.3323, -89.3323, -89.3557, -89.466, -89.5529, -89.6092, -89.6092, -89.5435, -89.466, -89.4097, -89.3557],
   y = [-0.74864, -0.727519, -0.804964, -0.870676, -0.97159, -0.992711, -0.959856, -0.88241, -0.83782, -0.783843, -0.760375, -0.74864] };
  WCL[1293] = struct {
   x = [-80.0599, -80.1374, -80.2923, -80.4026, -80.4237, -80.4026, -80.4683, -80.5903, -80.7452, -80.8649, -80.8203, -80.8203, -80.8649, -80.8203, -80.7992, -80.832, -80.8555, -80.8884, -80.8438, -80.7452,
        -80.6349, -80.5129, -80.4026, -80.3369, -80.3134, -80.304, -80.2359, -80.1608, -80.1256, -80.0834, -80.0177],
   y = [0.171319, -0.018775, -0.218256, -0.361413, -0.516304, -0.682929, -0.715785, -0.905878, -0.983324, -1.08189, -1.38229, -1.59115, -1.69207, -1.95726, -2.0793, -2.1896, -2.20133, -2.24592, -2.3351,
        -2.41255, -2.46652, -2.54397, -2.64488, -2.67774, -2.67774, -2.65427, -2.57682, -2.51111, -2.46652, -2.36796, -2.29051] };
  WCL[1294] = struct {
   x = [-79.9825, -80.0599, -80.1608, -80.2594, -80.2711],
   y = [-3.24098, -3.30904, -3.36301, -3.39587, -3.39587] };
  WCL[1295] = struct {
   x = [-80.2711, -80.2923, -80.3462, -80.5246, -80.6772, -80.7992, -80.8978, -81.0198, -81.1747, -81.2521, -81.2732, -81.285, -81.2075, -81.1512, -81.1301, -81.1418, -81.0972, -80.9869, -80.8884, -80.8649,
        -81.0409, -81.1418, -80.9987, -80.8884, -80.6232, -80.4354, -80.2266, -80.0388],
   y = [-3.39587, -3.41934, -3.50852, -3.59535, -3.72912, -3.85116, -3.94972, -4.08349, -4.20318, -4.31583, -4.51296, -4.69132, -4.90019, -4.97764, -5.08794, -5.15365, -5.32966, -5.4517, -5.59485, -5.72863,
        -5.84831, -5.97035, -6.11351, -6.21207, -6.36696, -6.489, -6.63216, -6.80817] };
  WCL[1296] = struct {
   x = [-91.4609, -91.4397, -91.428, -91.428, -91.5383, -91.6158, -91.6486, -91.6369, -91.5712, -91.4726, -91.4609],
   y = [-0.328557, -0.384881, -0.495182, -0.528038, -0.528038, -0.495182, -0.429471, -0.361413, -0.349678, -0.340291, -0.328557] };
  WCL[1297] = struct {
   x = [-91.3717, -91.2614, -91.2074, -91.0971, -91.0314, -91.0197, -90.9422, -90.9094, -90.9211, -91.0408, -91.1511, -91.3177, -91.428, -91.4937, -91.4609, -91.3717, -91.252, -91.1863, -91.2191, -91.2848,
        -91.3506, -91.3951, -91.4397, -91.4726, -91.5266],
   y = [0.103261, -0.161932, -0.307435, -0.417737, -0.506916, -0.605483, -0.682929, -0.826086, -0.938734, -1.01618, -1.06077, -1.09363, -1.08189, -0.97159, -0.83782, -0.772109, -0.694663, -0.605483,
        -0.506916, -0.384881, -0.272233, -0.1854, -0.14081, -0.129076, -0.05163] };
  WCL[1298] = struct {
   x = [-90.7122, -90.8319, -90.8765, -90.8765, -90.8108, -90.7333, -90.6676, -90.623, -90.623, -90.7122],
   y = [-0.206521, -0.218256, -0.295701, -0.373147, -0.384881, -0.417737, -0.417737, -0.349678, -0.272233, -0.206521] };
  WCL[1299] = struct {
   x = [-90.3461, -90.325, -90.2593, -90.5245, -90.4142, -90.3461, -90.5245, -90.4142, -90.3461],
   y = [-0.528038, -0.549159, -0.626605, -0.572628, -0.549159, -0.528038, -0.572628, -0.549159, -0.528038] };
  WCL[1300] = struct {
   x = [-140.031, -139.989, -139.956, -139.965],
   y = [-8.72788, -8.76074, -8.84991, -8.90389] };
  WCL[1301] = struct {
   x = [-138.961, -138.961, -138.907, -138.872, -138.818, -138.74, -138.752, -138.818, -138.872, -138.949, -138.961],
   y = [-9.71355, -9.69243, -9.64784, -9.69243, -9.70182, -9.72528, -9.74641, -9.74641, -9.76753, -9.76753, -9.71355] };
  WCL[1302] = struct {
   x = [-140.087, -140.054, -140.031],
   y = [-8.72788, -8.71849, -8.72788] };
  WCL[1303] = struct {
   x = [-139.965, -140.01, -140.076, -140.12, -140.12, -140.108, -140.087],
   y = [-8.90389, -8.95787, -8.93675, -8.89216, -8.80532, -8.76074, -8.72788] };
  WCL[1304] = struct {
   x = [-157.907, -157.895, -157.85, -157.785, -157.74, -157.752, -157.806, -157.85, -157.907],
   y = [-9.00246, -8.99072, -8.9696, -9.01419, -9.07991, -9.10103, -9.08929, -9.03532, -9.00246] };
  WCL[1305] = struct {
   x = [-173.063, -173.086, -173.074, -173.063],
   y = [-2.77631, -2.75284, -2.78569, -2.77631] };
  WCL[1306] = struct {
   x = [-170.669, -170.69, -170.69, -170.657, -170.657, -170.669],
   y = [-3.70565, -3.71504, -3.6728, -3.69392, -3.73851, -3.70565] };
  WCL[1307] = struct {
   x = [-171.805, -171.793, -171.727, -171.751, -171.784, -171.805],
   y = [-9.14562, -9.13388, -9.18786, -9.23245, -9.22072, -9.14562] };
  WCL[1308] = struct {
   x = [174.318, 174.339, 174.372, 174.384, 174.372, 174.339, 174.318],
   y = [-0.635992, -0.680582, -0.769762, -0.814352, -0.758028, -0.692316, -0.635992] };
  WCL[1309] = struct {
   x = [179.153, 179.164, 179.143, 179.099, 179.099, 179.099, 179.153],
   y = [-8.50962, -8.59646, -8.65278, -8.62931, -8.55421, -8.52136, -8.50962] };
  WCL[1310] = struct {
   x = [160.817, 160.817, 160.85, 160.969, 161.047, 161.091, 161.091, 161.213, 161.324, 161.39, 161.368, 161.378, 161.479, 161.566, 161.655, 161.676, 161.666, 161.634, 161.446, 161.324, 161.148, 161.014,
        160.904, 160.85, 160.838, 160.817],
   y = [-8.38994, -8.33361, -8.32422, -8.42279, -8.57534, -8.68564, -8.77247, -8.87104, -9.03532, -9.155, -9.23245, -9.29816, -9.31928, -9.31928, -9.42958, -9.57039, -9.61498, -9.52815, -9.48356, -9.38499,
        -9.19959, -8.95787, -8.73961, -8.54248, -8.42279, -8.38994] };
  WCL[1311] = struct {
   x = [160.263, 160.254, 160.34, 160.441, 160.519, 160.519, 160.418, 160.319, 160.263],
   y = [-8.99072, -8.94848, -8.92501, -8.9696, -9.05644, -9.13388, -9.13388, -9.04705, -8.99072] };
  WCL[1312] = struct {
   x = [159.932, 160.099, 160.153, 160.209, 160.242, 160.263, 160.352, 160.43, 160.54, 160.671, 160.793, 160.892, 160.969, 160.96, 160.859, 160.695, 160.54, 160.364, 160.221, 160.075],
   y = [-9.33102, -9.35214, -9.36387, -9.36387, -9.35214, -9.35214, -9.37561, -9.36387, -9.46244, -9.53754, -9.60325, -9.6807, -9.70182, -9.83324, -9.92007, -9.88722, -9.84497, -9.82385, -9.83324, -9.83324] };
  WCL[1313] = struct {
   x = [169.538, 169.526, 169.526, 169.559, 169.592, 169.592, 169.549, 169.526, 169.526, 169.538],
   y = [-0.835473, -0.814352, -0.802618, -0.802618, -0.868329, -0.924653, -0.924653, -0.880063, -0.847207, -0.835473] };
  WCL[1314] = struct {
   x = [150.129, 150.141, 150.261, 150.371, 150.481, 150.549, 150.549, 150.449, 150.329, 150.207, 150.129, 150.129],
   y = [-2.47591, -2.43132, -2.4102, -2.4102, -2.42193, -2.56509, -2.64254, -2.66366, -2.68713, -2.57682, -2.50877, -2.47591] };
  WCL[1315] = struct {
   x = [150.958, 150.967, 151.024, 151.146, 151.3, 151.411, 151.563, 151.742, 151.929, 152.127, 152.314, 152.425, 152.558, 152.669, 152.767, 152.889, 153.089, 153.154, 153.22, 153.298, 153.298, 153.22,
        153.131, 152.967, 152.91, 152.845, 152.791, 152.713, 152.612, 152.514, 152.403, 152.305, 152.127, 151.939, 151.786, 151.664, 151.521, 151.411, 151.256, 151.101, 150.958, 150.958, 150.958],
   y = [-2.64254, -2.58621, -2.62141, -2.69886, -2.78569, -2.88661, -2.96405, -3.08609, -3.21751, -3.24098, -3.38414, -3.49444, -3.58362, -3.66106, -3.80422, -3.90513, -3.93799, -4.01544, -4.19145, -4.31348,
        -4.41205, -4.54582, -4.76642, -4.77816, -4.57867, -4.31348, -4.12574, -3.90513, -3.72677, -3.57188, -3.48505, -3.46158, -3.44046, -3.31842, -3.1518, -3.14006, -3.02976, -2.97579, -2.87487,
        -2.76457, -2.75284, -2.68713, -2.64254] };
  WCL[1316] = struct {
   x = [149.909, 150.052, 150.052, 150.085, 150.151, 150.24, 150.294, 150.207, 150.195, 150.294, 150.449, 150.615, 150.747, 150.901, 151.035, 151.188, 151.354, 151.53, 151.641, 151.796, 151.819, 151.786,
        151.709, 151.718, 151.873, 151.995, 152.105, 152.15, 152.27, 152.326, 152.403, 152.502, 152.514, 152.481, 152.371, 152.26, 152.204, 152.27, 152.15, 152.007, 151.786, 151.709, 151.697, 151.498,
        151.256, 151.035, 150.857, 150.66, 150.472, 150.305, 150.106],
   y = [-5.51741, -5.43996, -5.31793, -5.24048, -5.10906, -5.05274, -5.08559, -5.25222, -5.42823, -5.49629, -5.46108, -5.4517, -5.51741, -5.50568, -5.4517, -5.27334, -5.05274, -4.94243, -4.95417, -4.86499,
        -4.67724, -4.51296, -4.34634, -4.25716, -4.2243, -4.29002, -4.28063, -4.23604, -4.20318, -4.23604, -4.25716, -4.43552, -4.67724, -4.84387, -4.95417, -4.99876, -5.13018, -5.32966, -5.55026,
        -5.58312, -5.55026, -5.63945, -5.77087, -5.97035, -6.09004, -6.19095, -6.24493, -6.27778, -6.27778, -6.27778, -6.32237] };
  WCL[1317] = struct {
   x = [154.678, 154.656, 154.699, 154.743, 154.767, 154.767, 154.743, 154.689, 154.689, 154.678],
   y = [-4.95417, -4.9659, -4.93305, -5.0762, -5.2311, -5.32966, -5.29681, -5.15365, -5.03162, -4.95417] };
  WCL[1318] = struct {
   x = [154.832, 154.886, 154.943, 155.098, 155.208, 155.351, 155.494, 155.628, 155.816, 156.013, 156.057, 156.069, 156.025, 155.903, 155.748, 155.593, 155.407, 155.33, 155.318, 155.163, 154.987, 154.865,
        154.832, 154.844, 154.832],
   y = [-5.41884, -5.42823, -5.46108, -5.48455, -5.64883, -5.82719, -5.97035, -6.11351, -6.27778, -6.43267, -6.58757, -6.68613, -6.74011, -6.80582, -6.82929, -6.79643, -6.65328, -6.49839, -6.24493, -6.14636,
        -5.97035, -5.77087, -5.60659, -5.50568, -5.41884] };
  WCL[1319] = struct {
   x = [156.588, 156.654, 156.776, 156.874, 157.062, 157.118, 157.196, 157.372, 157.449, 157.569, 157.604, 157.449, 157.25, 157.041, 156.797, 156.675, 156.588],
   y = [-6.59695, -6.55471, -6.68613, -6.75185, -6.7847, -6.88327, -6.97245, -7.03816, -7.12499, -7.15785, -7.27988, -7.301, -7.20244, -7.13673, -6.895, -6.71899, -6.59695] };
  WCL[1320] = struct {
   x = [156.654, 156.743, 156.82, 156.862, 156.841, 156.82, 156.752, 156.665, 156.621, 156.654],
   y = [-7.50049, -7.53334, -7.62017, -7.70935, -7.81731, -7.88302, -7.77506, -7.66476, -7.54273, -7.50049] };
  WCL[1321] = struct {
   x = [157.205, 157.261, 157.261, 157.172, 157.095, 157.085, 157.151, 157.205],
   y = [-7.85016, -7.91822, -8.04964, -8.0825, -7.95108, -7.78445, -7.79619, -7.85016] };
  WCL[1322] = struct {
   x = [157.372, 157.36, 157.383, 157.494, 157.604, 157.658, 157.703, 157.879, 157.947, 157.834, 157.703, 157.581, 157.449, 157.372],
   y = [-8.18107, -8.11536, -8.00506, -7.96047, -7.96047, -8.0825, -8.14821, -8.24678, -8.4885, -8.4885, -8.33361, -8.22331, -8.24678, -8.18107] };
  WCL[1323] = struct {
   x = [157.383, 157.348, 157.383, 157.449, 157.47, 157.482, 157.459, 157.383],
   y = [-8.55421, -8.47677, -8.44391, -8.42279, -8.47677, -8.5636, -8.60819, -8.55421] };
  WCL[1324] = struct {
   x = [158.045, 158.001, 158.034, 158.09, 158.156, 158.188, 158.177, 158.111, 158.045],
   y = [-8.60819, -8.58707, -8.53075, -8.50962, -8.50962, -8.57534, -8.64105, -8.6739, -8.60819] };
  WCL[1325] = struct {
   x = [158.63, 158.662, 158.796, 158.916, 159.061, 159.214, 159.327, 159.47, 159.601, 159.7, 159.801, 159.911, 159.932, 159.944, 159.932, 159.822, 159.667, 159.512, 159.369, 159.237, 159.083, 158.916,
        158.763, 158.674, 158.63],
   y = [-7.50987, -7.48875, -7.5662, -7.65303, -7.74221, -7.82904, -7.90649, -7.96047, -8.01679, -8.10362, -8.13648, -8.22331, -8.32422, -8.42279, -8.45565, -8.39932, -8.31249, -8.20219, -8.10362, -8.03791,
        -7.95108, -7.87363, -7.74221, -7.59905, -7.50987] };
  WCL[1326] = struct {
   x = [159.679, 159.768, 159.801, 159.932],
   y = [-9.27704, -9.22072, -9.25357, -9.33102] };
  WCL[1327] = struct {
   x = [160.075, 159.911, 159.768, 159.667, 159.658, 159.679],
   y = [-9.83324, -9.75814, -9.62672, -9.50468, -9.39673, -9.27704] };
  WCL[1328] = struct {
   x = [152.723, 152.845, 152.946, 153.065, 153.077, 152.934, 152.812, 152.746, 152.723],
   y = [-8.94848, -9.04705, -9.10103, -9.16674, -9.27704, -9.27704, -9.14562, -8.98134, -8.94848] };
  WCL[1329] = struct {
   x = [150.648, 150.693, 150.747, 150.89, 151.035, 151.056, 151.078, 150.913, 150.693, 150.66, 150.636, 150.648],
   y = [-9.34275, -9.34275, -9.36387, -9.39673, -9.4953, -9.60325, -9.70182, -9.6807, -9.62672, -9.52815, -9.40846, -9.34275] };
  WCL[1330] = struct {
   x = [150.967, 151.035, 151.134, 151.188, 151.289, 151.376],
   y = [-9.80038, -9.84497, -9.94354, -9.95293, -9.93181, -9.96466] };
  WCL[1331] = struct {
   x = [151.122, 151.045, 150.991, 150.967],
   y = [-10.0186, -9.93181, -9.83324, -9.80038] };
  WCL[1332] = struct {
   x = [149.953, 150.151, 150.019],
   y = [-9.59386, -9.69243, -9.76753] };
  WCL[1333] = struct {
   x = [146.696, 146.729, 146.795, 146.851, 147.015, 147.203, 147.116, 147.015, 146.938, 146.895, 146.773, 146.696, 146.729, 146.708, 146.696],
   y = [-1.9995, -1.97838, -1.97838, -2.03236, -2.04409, -2.01124, -2.21072, -2.16613, -2.1544, -2.19898, -2.25531, -2.1544, -2.06756, -2.05583, -1.9995] };
  WCL[1334] = struct {
   x = [149.665, 149.721, 149.775, 149.852, 149.798, 149.676, 149.644, 149.665],
   y = [-1.33535, -1.31423, -1.40106, -1.51136, -1.57942, -1.49024, -1.37994, -1.33535] };
  WCL[1335] = struct {
   x = [147.182, 147.247, 147.304, 147.358, 147.292, 147.215, 147.17, 147.182],
   y = [-5.25222, -5.25222, -5.29681, -5.39537, -5.47282, -5.4517, -5.3414, -5.25222] };
  WCL[1336] = struct {
   x = [147.921, 147.944, 147.987, 148.031, 148.109, 148.186, 148.207, 148.207, 148.097, 148.02, 147.977, 147.933, 147.921],
   y = [-5.53853, -5.51741, -5.52914, -5.52914, -5.562, -5.61598, -5.73801, -5.80372, -5.80372, -5.72628, -5.63945, -5.58312, -5.53853] };
  WCL[1337] = struct {
   x = [148.451, 148.484, 148.541, 148.651, 148.771, 148.937, 149.235, 149.357, 149.477, 149.599, 149.733, 149.909],
   y = [-5.51741, -5.47282, -5.43996, -5.49629, -5.50568, -5.51741, -5.53853, -5.562, -5.562, -5.53853, -5.53853, -5.51741] };
  WCL[1338] = struct {
   x = [150.106, 149.852, 149.644, 149.51, 149.313, 149.113, 148.925, 148.672, 148.505, 148.451, 148.451],
   y = [-6.32237, -6.33411, -6.13463, -6.13463, -6.19095, -6.04779, -5.93749, -5.86944, -5.72628, -5.52914, -5.51741] };
  WCL[1339] = struct {
   x = [141.12, 141.197, 141.373, 141.495, 141.65, 141.85, 142.026, 142.225, 142.422, 142.577, 142.777, 142.943, 143.129, 143.361, 143.528, 143.704, 143.892, 144.124, 144.323, 144.509, 144.61, 144.765,
        144.985, 145.194, 145.438, 145.668, 145.868, 145.957, 145.966, 145.957, 146.022, 146.276, 146.485, 146.663, 146.827, 146.994, 147.215, 147.423, 147.59, 147.778, 147.933, 148.02, 147.998,
        147.844, 147.712, 147.501, 147.269, 147.182, 147.269, 147.325, 147.435, 147.59, 147.799, 147.954, 148.076, 148.12, 148.142],
   y = [-2.57682, -2.62141, -2.69886, -2.73172, -2.75284, -2.87487, -2.91946, -2.97579, -3.09548, -3.17292, -3.21751, -3.29496, -3.31842, -3.36301, -3.42873, -3.50617, -3.59535, -3.69392, -3.75963, -3.75963,
        -3.86054, -4.08115, -4.21492, -4.34634, -4.46837, -4.62326, -4.8556, -5.0762, -5.25222, -5.41884, -5.50568, -5.57139, -5.61598, -5.72628, -5.82719, -5.8929, -5.94688, -5.99147, -5.97974,
        -6.09004, -6.26605, -6.4538, -6.71899, -6.71899, -6.7847, -6.80582, -6.80582, -6.90674, -7.04989, -7.24703, -7.47702, -7.54273, -7.77506, -7.89475, -7.9722, -8.01679, -8.04964] };
  WCL[1340] = struct {
   x = [148.13, 148.186, 148.264, 148.308, 148.329, 148.341, 148.451, 148.583, 148.728, 148.871, 149.003, 149.146, 149.313, 149.4, 149.39, 149.357, 149.51, 149.721, 149.953],
   y = [-8.04964, -8.04964, -8.07077, -8.11536, -8.22331, -8.46503, -8.60819, -8.70676, -8.99072, -9.07991, -9.05644, -9.04705, -9.02358, -9.10103, -9.27704, -9.44132, -9.56101, -9.56101, -9.59386] };
  WCL[1341] = struct {
   x = [150.019, 149.953],
   y = [-9.76753, -9.89895] };
  WCL[1342] = struct {
   x = [147.9, 147.766, 147.623, 147.48, 147.348, 147.226, 147.203, 147.203, 147.104, 147.083, 147.125, 146.982, 146.773, 146.663, 146.553, 146.41, 146.365, 146.154, 145.966, 145.79, 145.602, 145.438,
        145.337, 145.283, 145.26, 145.161, 145.03, 144.941, 144.929, 144.797, 144.687, 144.544, 144.443, 144.377, 144.255, 144.035, 143.903, 143.957, 144.014, 143.87, 143.65, 143.561, 143.605, 143.483,
        143.317, 143.129, 143.241, 143.349, 143.427, 143.462, 143.373, 143.263, 143.173, 143.063, 142.908, 142.798, 142.678, 142.544, 142.512, 142.422, 142.291, 142.124, 141.96, 141.805, 141.65,
        141.573, 141.519, 141.34],
   y = [-10.0515, -9.88722, -9.71355, -9.57039, -9.47417, -9.39673, -9.39673, -9.38499, -9.27704, -9.22072, -9.07991, -9.06817, -9.03532, -8.80532, -8.65278, -8.47677, -8.24678, -8.10362, -8.04964, -7.98393,
        -7.9722, -7.8619, -7.85016, -7.81731, -7.81731, -7.76333, -7.73048, -7.66476, -7.59905, -7.55446, -7.62017, -7.50049, -7.5662, -7.64364, -7.69762, -7.64364, -7.55446, -7.81731, -7.9722,
        -8.02618, -8.00506, -8.10362, -8.19046, -8.2679, -8.32422, -8.36647, -8.46503, -8.60819, -8.79359, -8.90389, -9.00246, -9.03532, -9.07991, -9.155, -9.25357, -9.28643, -9.25357, -9.21133,
        -9.18786, -9.17847, -9.17847, -9.18786, -9.23245, -9.22072, -9.17847, -9.12215, -9.13388, -9.14562] };
  WCL[1343] = struct {
   x = [141.34, 141.331, 141.12, 140.977, 140.667, 140.503, 140.315, 140.148, 140.115, 140.029],
   y = [-9.14562, -9.12215, -9.05644, -8.94848, -8.68564, -8.4885, -8.30076, -8.20219, -8.13648, -8.05903] };
  WCL[1344] = struct {
   x = [139.928, 140.106, 140.193, 140.336, 140.437, 140.514, 140.613, 140.702, 140.801, 140.834, 140.834, 140.866, 140.965, 141.01, 141.075, 141.12],
   y = [-2.34449, -2.33275, -2.28816, -2.35388, -2.33275, -2.34449, -2.39847, -2.45479, -2.46418, -2.49938, -2.55336, -2.58621, -2.58621, -2.56509, -2.57682, -2.57682] };
  WCL[1345] = struct {
   x = [140.029, 139.796, 139.62, 139.421, 139.266, 139.123, 139.024, 139.001, 139.012, 138.956, 138.825, 138.736, 138.538, 138.36, 138.205, 138.008, 137.865, 137.853, 137.975, 138.085, 138.24, 138.383,
        138.604, 138.759, 138.935, 139.066, 139.066, 138.879, 138.902, 139.034, 138.836, 138.625, 138.726, 138.768, 138.801, 138.649, 138.503, 138.461, 138.437, 138.449, 138.372, 138.229, 138.184,
        138.074, 137.952, 137.865, 137.841, 137.797, 137.752, 137.654, 137.555, 137.4, 137.224, 137.046, 136.893, 136.67, 136.429, 136.253, 136.065, 135.898, 135.701, 135.534, 135.391, 135.257,
        135.147, 135.114, 135.06, 134.882, 134.783, 134.751, 134.596, 134.42, 134.396, 134.33, 134.154, 134.065, 134.023, 133.845, 133.812, 133.824, 133.901, 133.878, 133.835, 133.8, 133.746, 133.669,
        133.603, 133.57, 133.493, 133.371, 133.195, 133.028, 132.941, 132.897, 132.885, 132.829, 132.852, 132.84, 132.718, 132.575, 132.387, 132.146, 131.991, 132.024, 132.19, 132.444, 132.697,
        132.906, 133.061, 133.239, 133.425, 133.645, 133.812, 133.824, 133.889, 133.91, 133.734, 133.382, 132.974, 132.685, 132.521, 132.366, 132.101, 132.045, 131.958, 131.913, 131.803, 131.627,
        131.449, 131.317, 131.273, 131.073, 130.986, 130.953, 130.963, 131.007, 131.097, 131.195, 131.251, 131.317, 131.505, 131.648, 131.836, 132.002, 132.167, 132.399, 132.664, 132.906, 133.138,
        133.26, 133.415, 133.612, 133.812, 134.023, 134.089, 134.208, 134.232, 134.154, 134.133, 134.187, 134.244, 134.387, 134.474, 134.507, 134.518, 134.706, 134.795, 134.849, 134.938, 135.093,
        135.248, 135.38, 135.624, 135.722, 135.854, 135.976, 136.008, 136.098, 136.22, 136.363, 136.473, 136.637, 136.771, 136.882, 137.102, 137.257, 137.245, 137.245, 137.323, 137.555, 137.621,
        137.898, 138.017, 138.085, 138.205, 138.437, 138.649, 138.768, 138.846, 139.045, 139.278, 139.409, 139.564, 139.707, 139.817, 139.928],
   y = [-8.05903, -8.11536, -8.18107, -8.14821, -8.12474, -8.25617, -8.16933, -8.29137, -8.33361, -8.35708, -8.36647, -8.29137, -8.35708, -8.39932, -8.41106, -8.43218, -8.42279, -8.25617, -8.05903, -7.78445,
        -7.59905, -7.45589, -7.37845, -7.39018, -7.50987, -7.55446, -7.42304, -7.25876, -7.20244, -7.16958, -7.10387, -6.96071, -6.91612, -6.73072, -6.60869, -6.49839, -6.25666, -6.05718, -5.88117,
        -5.80372, -5.73801, -5.53853, -5.4517, -5.4517, -5.40711, -5.29681, -5.29681, -5.24048, -5.17477, -5.15365, -5.13018, -5.04335, -4.9659, -4.97529, -4.89785, -4.77816, -4.7101, -4.68898,
        -4.5998, -4.54582, -4.53409, -4.51296, -4.48011, -4.46837, -4.45664, -4.39093, -4.37919, -4.29002, -4.15859, -4.0037, -4.01544, -3.98258, -4.10227, -3.99197, -3.87228, -3.83708, -3.72677,
        -3.64933, -3.52729, -3.32781, -3.11894, -2.97579, -3.09548, -3.28557, -3.42873, -3.58362, -3.69392, -3.83708, -3.95911, -4.09288, -4.114, -4.13747, -4.09288, -4.02482, -3.86054, -3.64933,
        -3.5625, -3.36301, -3.36301, -3.11894, -2.99691, -3.00864, -2.95232, -2.78569, -2.70825, -2.69886, -2.75284, -2.6308, -2.55336, -2.48765, -2.55336, -2.56509, -2.49938, -2.30929, -2.18725,
        -2.06756, -2.14501, -2.24357, -2.33275, -2.32102, -2.27643, -2.32102, -2.17786, -2.06756, -1.90094, -1.70145, -1.60054, -1.51136, -1.53483, -1.45738, -1.57942, -1.49024, -1.45738, -1.43392,
        -1.4128, -1.31423, -1.20158, -1.01383, -0.868329, -0.826086, -0.781496, -0.781496, -0.725172, -0.570281, -0.427124, -0.359066, -0.359066, -0.41539, -0.570281, -0.704051, -0.790883, -0.781496,
        -0.781496, -0.790883, -1.07954, -1.26964, -1.55595, -1.78829, -2.02297, -2.24357, -2.45479, -2.65427, -2.79743, -2.69886, -2.54162, -2.60968, -2.90773, -2.98517, -3.18465, -3.28557, -3.3724,
        -3.38414, -3.36301, -3.14006, -3.08609, -2.99691, -2.85375, -2.77631, -2.67539, -2.56509, -2.2647, -2.22245, -2.22245, -2.19898, -2.14501, -2.03236, -1.8892, -1.85635, -1.7789, -1.61228,
        -1.57942, -1.4128, -1.46912, -1.56769, -1.60054, -1.68972, -1.78829, -1.81176, -1.8892, -1.95491, -2.02297, -2.12154, -2.22245, -2.28816, -2.32102, -2.34449] };
  WCL[1346] = struct {
   x = [135.49, 135.436, 135.513, 135.666, 135.734, 135.865, 136.02, 136.163, 136.253, 136.33, 136.253, 136.041, 135.943, 135.91, 135.767, 135.588, 135.49],
   y = [-0.74864, -0.704051, -0.692316, -0.704051, -0.736906, -0.74864, -0.826086, -0.948121, -1.04669, -1.1476, -1.18046, -1.20158, -1.13587, -1.02557, -0.912919, -0.826086, -0.74864] };
  WCL[1347] = struct {
   x = [135.612, 135.612, 135.666, 135.755, 135.898, 136.032, 136.208, 136.384, 136.527, 136.748, 136.837, 136.87, 136.661, 136.45, 136.262, 136.074, 135.898, 135.734, 135.633, 135.612],
   y = [-1.6334, -1.60054, -1.60054, -1.60054, -1.62401, -1.64513, -1.68972, -1.68972, -1.71084, -1.72258, -1.76717, -1.87747, -1.91032, -1.8892, -1.86808, -1.81176, -1.75543, -1.70145, -1.67799, -1.6334] };
  WCL[1348] = struct {
   x = [130.423, 130.378, 130.423, 130.5, 130.578, 130.7, 131.261, 131.141, 130.975, 130.852, 130.721, 130.578, 130.456, 130.367, 130.367, 130.423],
   y = [-0.114995, -0.114995, -0.105608, -0.070405, -0.070405, -0.049284, -0.337944, -0.349678, -0.382534, -0.436511, -0.359066, -0.305089, -0.28162, -0.215909, -0.171319, -0.114995] };
  WCL[1349] = struct {
   x = [130.7, 130.721, 130.82, 130.909, 130.998, 131.007, 131.007, 130.897, 130.81, 130.733, 130.7],
   y = [-0.990364, -0.990364, -0.936387, -0.912919, -0.969243, -1.1124, -1.26964, -1.29076, -1.24617, -1.09128, -0.990364] };
  WCL[1350] = struct {
   x = [129.958, 130.069, 130.17, 130.301, 130.39, 130.367, 130.39, 130.235, 130.048],
   y = [-1.80002, -1.76717, -1.71084, -1.71084, -1.83288, -1.91032, -1.9995, -2.06756, -2.08868] };
  WCL[1351] = struct {
   x = [129.893, 130.036, 130.202, 130.346, 130.512, 130.587, 130.632, 130.7, 130.843, 130.864, 130.787, 130.655, 130.512, 130.357, 130.202, 130.036],
   y = [-3.01803, -3.0415, -3.0415, -3.00864, -3.06262, -3.24098, -3.36301, -3.44985, -3.49444, -3.72677, -3.90513, -3.80422, -3.73851, -3.64933, -3.57188, -3.47332] };
  WCL[1352] = struct {
   x = [133.073, 133.084, 133.138, 133.195, 133.162, 133.084, 133.028, 133.049, 133.073],
   y = [-5.39537, -5.38364, -5.3414, -5.36252, -5.49629, -5.6723, -5.68169, -5.51741, -5.39537] };
  WCL[1353] = struct {
   x = [134.298, 134.354, 134.441, 134.507, 134.652, 134.706, 134.762, 134.762, 134.762, 134.739, 134.739, 134.772, 134.706, 134.607, 134.497, 134.408, 134.354, 134.375, 134.431, 134.375, 134.298],
   y = [-5.80372, -5.72628, -5.57139, -5.4517, -5.50568, -5.59485, -5.72628, -5.79199, -5.88117, -6.0032, -6.05718, -6.25666, -6.37635, -6.36696, -6.32237, -6.21207, -6.06892, -5.97035, -5.83658, -5.86944,
        -5.80372] };
  WCL[1354] = struct {
   x = [134.232, 134.22, 134.199, 134.199, 134.232, 134.277, 134.309, 134.363, 134.408, 134.518, 134.563, 134.575, 134.596, 134.584, 134.518, 134.453, 134.387, 134.298, 134.253, 134.232],
   y = [-6.21207, -6.16748, -6.13463, -6.10177, -6.08065, -6.12289, -6.17922, -6.24493, -6.26605, -6.31064, -6.35523, -6.42094, -6.49839, -6.53124, -6.51951, -6.43267, -6.35523, -6.29891, -6.24493, -6.21207] };
  WCL[1355] = struct {
   x = [134.176, 134.244, 134.309, 134.375, 134.431, 134.485, 134.485, 134.453, 134.396, 134.265, 134.187, 134.122, 134.098, 134.133, 134.199, 134.244, 134.265, 134.232, 134.176, 134.166, 134.176],
   y = [-6.23319, -6.26605, -6.36696, -6.43267, -6.49839, -6.5641, -6.6744, -6.79643, -6.90674, -6.93959, -6.90674, -6.79643, -6.65328, -6.5641, -6.73072, -6.69552, -6.46553, -6.47727, -6.35523, -6.23319,
        -6.23319] };
  WCL[1356] = struct {
   x = [131.183, 131.207, 131.261, 131.24, 131.228, 131.317, 131.46, 131.538, 131.603, 131.714, 131.714, 131.704, 131.594, 131.493, 131.395, 131.317, 131.183, 131.183],
   y = [-7.78445, -7.76333, -7.7516, -7.65303, -7.52161, -7.45589, -7.34559, -7.15785, -7.12499, -7.20244, -7.4113, -7.58732, -7.76333, -7.89475, -8.02618, -7.9722, -7.87363, -7.78445] };
  WCL[1357] = struct {
   x = [129.782, 129.881, 129.958],
   y = [-1.85635, -1.85635, -1.80002] };
  WCL[1358] = struct {
   x = [130.048, 129.925, 129.815, 129.782],
   y = [-2.08868, -2.01124, -1.93379, -1.85635] };
  WCL[1359] = struct {
   x = [127.982, 127.938, 127.917, 127.961, 128.048, 128.137, 128.238, 128.391, 128.578, 128.7, 128.909, 129.088, 129.198, 129.374, 129.541, 129.728, 129.893],
   y = [-3.52729, -3.48505, -3.24098, -3.13068, -3.07435, -2.95232, -2.86314, -2.87487, -2.89834, -2.90773, -2.88661, -2.84202, -2.90773, -2.86314, -2.8209, -2.89834, -3.01803] };
  WCL[1360] = struct {
   x = [130.036, 129.893, 129.705, 129.541, 129.541, 129.44, 129.243, 129.055, 128.966, 128.9, 128.745, 128.689, 128.536, 128.358, 128.247, 128.161, 128.116, 128.071, 128.027, 127.982],
   y = [-3.47332, -3.39587, -3.36301, -3.36301, -3.44046, -3.46158, -3.44985, -3.42873, -3.35128, -3.28557, -3.35128, -3.44985, -3.48505, -3.40526, -3.28557, -3.19639, -3.25037, -3.33955, -3.46158, -3.52729] };
  WCL[1361] = struct {
   x = [129.651, 129.639, 129.749, 129.839, 129.848, 129.771, 129.717, 129.651],
   y = [-7.90649, -7.8619, -7.8619, -7.85016, -7.93934, -8.03791, -8.02618, -7.90649] };
  WCL[1362] = struct {
   x = [128.59, 128.546, 128.635, 128.689, 128.689, 128.635, 128.59],
   y = [-7.18131, -7.15785, -7.12499, -7.14846, -7.22356, -7.26815, -7.18131] };
  WCL[1363] = struct {
   x = [125.908, 125.83, 125.884, 125.962, 126.095, 126.26, 126.394, 126.483, 126.614, 126.68, 126.814, 126.856, 126.659, 126.516, 126.328, 126.161, 126.03, 125.908],
   y = [-8.02618, -7.92761, -7.78445, -7.69762, -7.70935, -7.70935, -7.66476, -7.59905, -7.53334, -7.60844, -7.68588, -7.73048, -7.85016, -7.9722, -7.99332, -7.98393, -8.03791, -8.02618] };
  WCL[1364] = struct {
   x = [124.781, 124.814, 124.892, 124.99, 125.068, 125.124, 125.279, 125.422, 125.598, 125.797, 125.908, 126.072, 126.217, 126.405, 126.581, 126.781, 126.933, 127.044, 127.198, 127.231, 127.133, 126.835,
        126.635, 126.394, 126.128, 125.896, 125.687, 125.511, 125.356, 125.213, 125.112],
   y = [-9.06817, -9.05644, -8.99072, -8.90389, -8.7842, -8.66217, -8.65278, -8.62931, -8.59646, -8.5636, -8.49789, -8.53075, -8.52136, -8.52136, -8.49789, -8.43218, -8.41106, -8.35708, -8.41106, -8.47677,
        -8.70676, -8.79359, -8.90389, -9.02358, -9.08929, -9.21133, -9.25357, -9.28643, -9.39673, -9.48356, -9.41785] };
  WCL[1365] = struct {
   x = [125.112, 125.124, 125.103, 125.068, 125.025, 124.915, 124.826],
   y = [-9.41785, -9.48356, -9.50468, -9.60325, -9.69243, -9.791, -9.88722] };
  WCL[1366] = struct {
   x = [123.723, 123.678, 123.688, 123.711, 123.833, 123.976, 124.054, 124.141, 124.218, 124.307, 124.439, 124.54, 124.617, 124.692, 124.781],
   y = [-10.0093, -9.88722, -9.76753, -9.62672, -9.50468, -9.38499, -9.35214, -9.33102, -9.27704, -9.23245, -9.18786, -9.13388, -9.13388, -9.12215, -9.06817] };
  WCL[1367] = struct {
   x = [124.495, 124.54, 124.605, 124.671, 124.769, 125.025, 125.112, 125.035, 124.915, 124.748, 124.582, 124.471, 124.528, 124.495],
   y = [-8.19046, -8.16933, -8.18107, -8.19046, -8.18107, -8.1576, -8.2679, -8.3782, -8.43218, -8.45565, -8.45565, -8.39932, -8.30076, -8.19046] };
  WCL[1368] = struct {
   x = [124.021, 123.997, 124.063, 124.164, 124.274, 124.274, 124.173, 124.108, 124.021],
   y = [-8.46503, -8.36647, -8.32422, -8.2679, -8.23504, -8.36647, -8.55421, -8.4885, -8.46503] };
  WCL[1369] = struct {
   x = [123.357, 123.38, 123.401, 123.467, 123.523, 123.577, 123.666, 123.765, 123.821, 123.723, 123.643, 123.535, 123.413, 123.357],
   y = [-8.52136, -8.45565, -8.30076, -8.22331, -8.30076, -8.33361, -8.20219, -8.18107, -8.2679, -8.43218, -8.50962, -8.58707, -8.57534, -8.52136] };
  WCL[1370] = struct {
   x = [120.001, 120.057, 120.144, 120.254, 120.388, 120.498, 120.663, 120.763, 120.951, 121.139, 121.238, 121.36, 121.491, 121.625, 121.768, 121.956, 122.054, 122.111, 122.23, 122.43, 122.528, 122.639,
        122.707, 122.817, 122.915, 122.915, 122.862, 122.871, 122.972, 123.014, 122.96, 122.915, 122.784, 122.596, 122.43, 122.221, 121.977, 121.712, 121.524, 121.336, 121.127, 120.939, 120.773,
        120.597, 120.355, 120.188, 120.034],
   y = [-8.44391, -8.44391, -8.36647, -8.32422, -8.2679, -8.23504, -8.21392, -8.29137, -8.30076, -8.34535, -8.45565, -8.52136, -8.57534, -8.50962, -8.4885, -8.44391, -8.43218, -8.50962, -8.5636, -8.59646,
        -8.54248, -8.41106, -8.36647, -8.33361, -8.25617, -8.29137, -8.20219, -8.03791, -8.11536, -8.32422, -8.39932, -8.4885, -8.60819, -8.70676, -8.75135, -8.75135, -8.84991, -8.81706, -8.81706,
        -8.90389, -8.94848, -8.8593, -8.83818, -8.79359, -8.80532, -8.80532, -8.81706] };
  WCL[1371] = struct {
   x = [119.879, 120.001],
   y = [-8.4885, -8.44391] };
  WCL[1372] = struct {
   x = [120.001, 120.057, 120.134, 120.245, 120.299, 120.421, 120.609, 120.796],
   y = [-9.27704, -9.33102, -9.4507, -9.46244, -9.52815, -9.61498, -9.71355, -9.93181] };
  WCL[1373] = struct {
   x = [119.902, 120.001],
   y = [-9.35214, -9.27704] };
  WCL[1374] = struct {
   x = [120.188, 120.144, 120.144, 120.156, 120.278, 120.475, 120.63, 120.752, 120.928, 121.139, 121.184, 121.327, 121.47, 121.667, 121.834, 121.899, 122.054, 122.143, 122.287, 122.43, 122.674, 122.838,
        122.862, 122.883, 123.038, 123.169, 123.368, 123.479, 123.458, 123.345, 123.115, 122.894, 122.683, 122.519, 122.352, 122.078, 121.944, 121.744, 121.646, 121.414, 121.437, 121.69, 121.899,
        122.054, 122.209, 122.298, 122.376, 122.486, 122.418, 122.287, 122.263, 122.287, 122.507, 122.606, 122.606, 122.716, 122.85, 122.894, 122.838, 122.761, 122.495, 122.23, 122.12, 121.932,
        121.658, 121.568, 121.547, 121.625, 121.524, 121.315, 121.106, 120.984, 121.127, 121.127, 121.094, 121.05, 120.829, 120.585, 120.388, 120.299, 120.433, 120.442, 120.433, 120.421, 120.421,
        120.487, 120.364, 120.388, 120.465, 120.409, 120.156],
   y = [0.028162, -0.215909, -0.481101, -0.758028, -0.912919, -0.980977, -1.20158, -1.40106, -1.39167, -1.37994, -1.26964, -1.13587, -0.957509, -0.868329, -0.924653, -0.924653, -0.969243, -0.826086,
        -0.802618, -0.790883, -0.790883, -0.790883, -0.713438, -0.635992, -0.582015, -0.537425, -0.549159, -0.692316, -0.903532, -1.02557, -0.903532, -0.957509, -1.20158, -1.4128, -1.56769, -1.70145,
        -1.73431, -1.85635, -1.87747, -1.80002, -1.98777, -2.14501, -2.42193, -2.66366, -2.86314, -2.89834, -3.08609, -3.21751, -3.32781, -3.3724, -3.52729, -3.63759, -3.79249, -3.91452, -4.08115,
        -4.15859, -4.12574, -4.28063, -4.35807, -4.44725, -4.43552, -4.50123, -4.7453, -4.86499, -4.84387, -4.635, -4.37919, -4.15859, -4.01544, -3.86054, -3.66106, -3.39587, -3.11894, -2.90773,
        -2.75284, -2.64254, -2.62141, -2.75284, -2.91946, -3.1518, -3.39587, -3.64933, -3.90513, -4.25716, -4.57867, -4.7453, -4.99876, -5.28507, -5.55026, -5.59485, -5.61598] };
  WCL[1375] = struct {
   x = [121.866, 121.878, 121.944, 122.033, 122.111, 122.099, 122.01, 121.911, 121.866, 121.866],
   y = [-5.19824, -5.15365, -5.12079, -5.13018, -5.24048, -5.40711, -5.43996, -5.3414, -5.21936, -5.19824] };
  WCL[1376] = struct {
   x = [122.376, 122.397, 122.418, 122.451, 122.507, 122.617, 122.695, 122.707, 122.739, 122.728, 122.674, 122.639, 122.596, 122.474, 122.331, 122.331, 122.341, 122.364, 122.376],
   y = [-4.72183, -4.73357, -4.73357, -4.73357, -4.72183, -4.64439, -4.66786, -4.79928, -4.95417, -5.05274, -5.19824, -5.38364, -5.31793, -5.38364, -5.3414, -5.12079, -4.9659, -4.79928, -4.72183] };
  WCL[1377] = struct {
   x = [122.805, 122.85, 122.862, 122.862, 122.871, 122.927, 122.981, 123.047, 123.148, 123.202, 123.16, 123.047, 123.014, 123.026, 123.115, 123.169, 123.169, 123.047, 122.948, 122.793, 122.674, 122.674,
        122.707, 122.761, 122.805],
   y = [-4.99876, -4.95417, -4.84387, -4.67724, -4.52235, -4.46837, -4.35807, -4.42378, -4.57867, -4.75469, -4.73357, -4.86499, -4.9659, -5.05274, -5.10906, -5.24048, -5.37425, -5.43996, -5.53853, -5.6723,
        -5.562, -5.43996, -5.28507, -5.17477, -4.99876] };
  WCL[1378] = struct {
   x = [123.038, 123.082, 123.082, 123.115, 123.202, 123.246, 123.214, 123.136, 123.047, 123.026, 123.038],
   y = [-4.01544, -3.95911, -3.95911, -3.99197, -4.02482, -4.114, -4.2243, -4.23604, -4.17971, -4.03656, -4.01544] };
  WCL[1379] = struct {
   x = [123.07, 123.092, 123.181, 123.202, 123.169, 123.103, 123.014, 122.862, 122.826, 122.862, 122.927, 123.005, 123.07],
   y = [-1.15699, -1.15699, -1.15699, -1.21331, -1.31423, -1.4128, -1.55595, -1.54656, -1.37994, -1.29076, -1.22505, -1.19219, -1.15699] };
  WCL[1380] = struct {
   x = [123.225, 123.312, 123.39, 123.512, 123.535, 123.523, 123.458, 123.324, 123.279, 123.214, 123.202, 123.225],
   y = [-1.37994, -1.26964, -1.24617, -1.26964, -1.31423, -1.44565, -1.55595, -1.49024, -1.59115, -1.51136, -1.4128, -1.37994] };
  WCL[1381] = struct {
   x = [124.417, 124.439, 124.495, 124.617, 124.769, 124.892, 124.957, 125.058, 125.136, 125.213, 125.246, 125.267, 125.213, 125.079, 124.936, 124.769, 124.605, 124.495, 124.45, 124.406, 124.394, 124.417],
   y = [-1.6686, -1.67799, -1.6686, -1.6686, -1.67799, -1.71084, -1.74605, -1.73431, -1.73431, -1.73431, -1.75543, -1.81176, -1.82349, -1.90094, -1.91032, -1.91032, -2.02297, -1.9995, -1.91032, -1.78829,
        -1.73431, -1.6686] };
  WCL[1382] = struct {
   x = [125.422, 125.401, 125.488, 125.631, 125.765, 125.884, 125.997, 126.173, 126.283, 126.304, 126.239, 126.063, 125.908, 125.699, 125.544, 125.466, 125.422],
   y = [-1.82349, -1.78829, -1.81176, -1.85635, -1.82349, -1.80002, -1.80002, -1.78829, -1.7789, -1.78829, -1.84461, -1.8892, -1.91032, -1.93379, -1.95491, -1.87747, -1.82349] };
  WCL[1383] = struct {
   x = [125.908, 125.952, 125.973, 125.973, 126.006, 126.039, 125.997, 125.896, 125.896, 125.908],
   y = [-2.01124, -1.96665, -1.97838, -2.17786, -2.34449, -2.44306, -2.37734, -2.1544, -2.06756, -2.01124] };
  WCL[1384] = struct {
   x = [126.063, 126.095, 126.128, 126.239, 126.328, 126.492, 126.736, 126.945, 127.088, 127.088, 127.189, 127.231, 127.166, 126.966, 126.802, 126.548, 126.337, 126.206, 126.095, 126.039, 126.039, 126.063],
   y = [-3.08609, -3.08609, -3.17292, -3.14006, -3.10721, -3.06262, -3.01803, -3.14006, -3.24098, -3.32781, -3.31842, -3.5625, -3.69392, -3.77136, -3.86054, -3.80422, -3.70565, -3.53903, -3.3724, -3.20812,
        -3.11894, -3.08609] };
  WCL[1385] = struct {
   x = [127.928, 127.961, 128.048, 128.17, 128.304, 128.37, 128.28, 128.093, 127.928, 127.75],
   y = [0.061018, -0.269886, -0.45998, -0.647726, -0.802618, -0.868329, -0.891797, -0.704051, -0.481101, -0.269886] };
  WCL[1386] = struct {
   x = [127.353, 127.431, 127.52, 127.597, 127.619, 127.696, 127.839, 127.795, 127.675, 127.52, 127.487, 127.431, 127.365, 127.353],
   y = [-0.359066, -0.349678, -0.359066, -0.45998, -0.603137, -0.692316, -0.781496, -0.868329, -0.781496, -0.802618, -0.668848, -0.582015, -0.427124, -0.359066] };
  WCL[1387] = struct {
   x = [127.41, 127.475, 127.619, 127.795, 127.949, 128.116, 128.161, 128.116, 127.928, 127.708, 127.541, 127.419, 127.398, 127.386, 127.41],
   y = [-1.44565, -1.43392, -1.35647, -1.39167, -1.45738, -1.53483, -1.60054, -1.68972, -1.68972, -1.70145, -1.73431, -1.6686, -1.59115, -1.51136, -1.44565] };
  WCL[1388] = struct {
   x = [128.06, 128.006, 128.006, 128.006, 128.048, 128.161, 128.238, 128.313, 128.304, 128.247, 128.182, 128.06, 128.006],
   y = [-3.81595, -3.7831, -3.7831, -3.70565, -3.62821, -3.58362, -3.51791, -3.52729, -3.6728, -3.75024, -3.75024, -3.81595, -3.7831] };
  WCL[1389] = struct {
   x = [112.151, 112.075, 111.963, 111.843, 111.754, 111.676, 111.554, 111.435, 111.334, 111.268, 111.224, 111.048, 110.86, 110.717, 110.583, 110.419, 110.297, 110.165, 110.01],
   y = [-8.25617, -8.2679, -8.25617, -8.32422, -8.35708, -8.32422, -8.29137, -8.2679, -8.22331, -8.1576, -8.21392, -8.21392, -8.18107, -8.13648, -8.09189, -8.00506, -7.90649, -7.82904, -7.76333] };
  WCL[1390] = struct {
   x = [109.945, 110.088, 110.264, 110.419, 110.541, 110.639, 110.705, 110.738, 110.86, 111.003, 111.104, 111.235, 111.411, 111.566, 111.655, 111.754, 111.864, 111.986, 112.052, 112.052, 112.118, 112.118,
        112.162, 112.251, 112.427, 112.571, 112.66, 112.714, 112.848, 112.88, 112.892, 113.056, 113.268, 113.455, 113.641, 113.84, 114.084, 114.293, 114.436, 114.448, 114.415, 114.392, 114.436,
        114.504, 114.558, 114.469, 114.371, 114.084, 113.753, 113.465, 113.211, 112.979, 112.726, 112.493, 112.284, 112.207, 112.186, 112.151],
   y = [-6.83868, -6.87388, -6.88327, -6.93959, -6.88327, -6.70725, -6.60869, -6.47727, -6.37635, -6.38809, -6.55471, -6.66266, -6.66266, -6.57583, -6.64154, -6.66266, -6.73072, -6.76358, -6.79643, -6.79643,
        -6.83868, -6.83868, -6.83868, -6.82929, -6.82929, -6.90674, -7.04989, -7.16958, -7.26815, -7.44416, -7.5662, -7.65303, -7.70935, -7.69762, -7.66476, -7.6765, -7.63191, -7.66476, -7.77506,
        -8.01679, -8.21392, -8.38994, -8.50962, -8.60819, -8.79359, -8.75135, -8.62931, -8.55421, -8.47677, -8.35708, -8.27963, -8.35708, -8.42279, -8.3782, -8.29137, -8.2679, -8.2679, -8.25617] };
  WCL[1391] = struct {
   x = [117.583, 117.527, 117.518, 117.518, 117.384, 117.274, 117.142, 116.976, 116.898, 116.788, 116.633, 116.49, 116.326, 116.358, 116.457, 116.49, 116.49, 116.591, 116.689, 116.656, 116.591, 116.49,
        116.436, 116.412, 116.347, 116.258, 116.236, 116.248, 116.225, 116.171, 116.105, 116.082, 116.049, 115.983, 115.905, 115.828, 115.772, 115.762, 115.685, 115.575, 115.42, 115.253, 115.101,
        114.901, 114.746, 114.725, 114.702, 114.68, 114.657, 114.603, 114.547, 114.469, 114.392, 114.326, 114.195, 114.084, 113.929, 113.763, 113.721, 113.697, 113.521, 113.366, 113.256, 113.167,
        113.068, 113.024, 113.045, 112.892, 112.648, 112.517, 112.406, 112.317, 112.219, 112.186, 112.151, 112.04, 112.04, 112.007, 111.963, 111.921, 111.909, 111.909, 111.897, 111.876, 111.831,
        111.777, 111.742, 111.688, 111.655, 111.599, 111.545, 111.423, 111.334, 111.235, 111.158, 110.994, 110.893, 110.782, 110.717, 110.639, 110.529, 110.463, 110.329, 110.264, 110.264, 110.275,
        110.21, 110.165, 110.132, 110.121, 110.022, 110.01, 110.055, 110.076, 110.043],
   y = [0.007041, -0.159585, -0.3708, -0.481101, -0.647726, -0.912919, -1.18046, -1.31423, -1.26964, -1.40106, -1.51136, -1.64513, -1.76717, -1.82349, -1.85635, -2.02297, -2.1544, -2.19898, -2.25531,
        -2.39847, -2.59795, -2.53224, -2.58621, -2.80916, -2.96405, -2.97579, -3.06262, -3.1518, -3.21751, -3.24098, -3.31842, -3.42873, -3.53903, -3.57188, -3.63759, -3.69392, -3.70565, -3.70565,
        -3.75963, -3.81595, -3.88166, -3.95911, -4.03656, -4.12574, -4.17033, -4.09288, -3.97085, -3.82769, -3.71504, -3.55076, -3.50617, -3.48505, -3.41699, -3.36301, -3.36301, -3.36301, -3.41699,
        -3.44046, -3.35128, -3.19639, -3.20812, -3.19639, -3.13068, -3.06262, -3.02976, -3.08609, -3.18465, -3.28557, -3.38414, -3.39587, -3.3724, -3.36301, -3.39587, -3.41699, -3.42873, -3.44046,
        -3.44046, -3.46158, -3.48505, -3.48505, -3.47332, -3.3724, -3.27384, -3.1518, -2.99691, -2.85375, -2.77631, -2.86314, -2.90773, -2.95232, -2.9312, -2.89834, -2.91946, -2.99691, -3.02976,
        -3.0415, -3.00864, -3.01803, -3.02976, -2.96405, -2.88661, -2.94058, -2.95232, -2.69886, -2.54162, -2.5205, -2.43132, -2.28816, -2.14501, -2.06756, -1.85635, -1.76717, -1.71084, -1.55595,
        -1.34708] };
  WCL[1392] = struct {
   x = [116.269, 116.293, 116.302, 116.335, 116.347, 116.358, 116.358, 116.368, 116.391, 116.368, 116.281, 116.192, 116.159, 116.138, 116.126, 116.126, 116.147, 116.171, 116.215, 116.258, 116.269],
   y = [-3.28557, -3.27384, -3.32781, -3.44046, -3.50617, -3.62821, -3.75024, -3.80422, -3.88166, -3.91452, -4.01544, -4.02482, -3.91452, -3.77136, -3.60474, -3.52729, -3.41699, -3.33955, -3.29496, -3.2621,
        -3.28557] };
  WCL[1393] = struct {
   x = [120.034, 119.923, 119.879, 119.858, 119.879],
   y = [-8.81706, -8.80532, -8.75135, -8.59646, -8.4885] };
  WCL[1394] = struct {
   x = [120.134, 119.914, 119.682, 119.461, 119.24, 119.095, 119.095, 119.217, 119.372, 119.548, 119.67, 119.825, 119.902],
   y = [-10.0632, -9.94354, -9.77926, -9.791, -9.76753, -9.59386, -9.4507, -9.39673, -9.36387, -9.35214, -9.34275, -9.36387, -9.35214] };
  WCL[1395] = struct {
   x = [117.109, 117.142, 117.285, 117.473, 117.628, 117.783, 117.849, 117.936, 118.025, 118.102, 118.245, 118.224, 118.036, 117.893, 117.849, 117.98, 118.147, 118.278, 118.412, 118.588, 118.71, 118.722,
        118.797, 118.931, 119.029, 119.118, 119.13, 119.205, 119.095, 118.942, 118.919, 118.931, 118.686, 118.51, 118.489, 118.224, 117.959, 117.706, 117.473, 117.285, 117.065, 116.91, 116.854,
        116.854, 116.832, 116.964, 117.065, 117.109],
   y = [-8.41106, -8.38994, -8.35708, -8.38994, -8.43218, -8.55421, -8.64105, -8.68564, -8.68564, -8.64105, -8.60819, -8.53075, -8.43218, -8.31249, -8.14821, -8.07077, -8.16933, -8.32422, -8.31249, -8.2679,
        -8.35708, -8.49789, -8.41106, -8.2679, -8.41106, -8.49789, -8.61993, -8.68564, -8.71849, -8.6739, -8.71849, -8.79359, -8.82645, -8.77247, -8.68564, -8.82645, -8.90389, -9.00246, -9.03532,
        -9.07991, -9.06817, -9.00246, -8.89216, -8.75135, -8.61993, -8.53075, -8.4885, -8.41106] };
  WCL[1396] = struct {
   x = [116.105, 116.159, 116.236, 116.368, 116.567, 116.722, 116.743, 116.666, 116.656, 116.534, 116.335, 116.105, 115.959, 115.959, 116.082, 116.105, 116.105],
   y = [-8.45565, -8.38994, -8.34535, -8.22331, -8.21392, -8.31249, -8.52136, -8.71849, -8.90389, -8.93675, -8.93675, -8.93675, -8.87104, -8.73961, -8.68564, -8.57534, -8.45565] };
  WCL[1397] = struct {
   x = [115.044, 115.089, 115.22, 115.399, 115.575, 115.673, 115.685, 115.53, 115.375, 115.265, 115.232, 115.101, 114.955, 114.802, 114.669, 114.58, 114.558, 114.603, 114.779, 114.913, 114.967, 115.023,
        115.044],
   y = [-8.09189, -8.09189, -8.03791, -8.0825, -8.16933, -8.30076, -8.44391, -8.52136, -8.61993, -8.81706, -8.71849, -8.53075, -8.41106, -8.38994, -8.36647, -8.30076, -8.12474, -8.04964, -8.07077, -8.07077,
        -8.09189, -8.09189, -8.09189] };
  WCL[1398] = struct {
   x = [112.803, 112.869, 113.002, 113.146, 113.399, 113.599, 113.807, 113.974, 114.073, 113.974, 113.763, 113.509, 113.301, 113.124, 112.946, 112.836, 112.803, 112.803],
   y = [-6.92786, -6.93959, -6.86215, -6.85041, -6.86215, -6.86215, -6.87388, -6.85041, -7.04989, -7.09214, -7.13673, -7.22356, -7.19305, -7.18131, -7.15785, -7.10387, -7.01469, -6.92786] };
  WCL[1399] = struct {
   x = [115.298, 115.309, 115.542, 115.607, 115.619, 115.518, 115.399, 115.298, 115.298],
   y = [-6.7847, -6.77297, -6.77297, -6.81756, -6.895, -6.94898, -7.0053, -6.87388, -6.7847] };
  WCL[1400] = struct {
   x = [120.156, 119.923, 119.682, 119.527, 119.416, 119.461, 119.47, 119.548, 119.571, 119.625, 119.691, 119.637, 119.581, 119.548, 119.318, 119.107, 118.952, 118.919, 118.919, 118.931, 119.13, 119.25,
        119.384, 119.372, 119.372, 119.47, 119.581, 119.714, 119.804, 119.846, 119.89, 119.879, 119.747, 119.792],
   y = [-5.61598, -5.6723, -5.64883, -5.57139, -5.29681, -5.14192, -5.12079, -4.92131, -4.68898, -4.35807, -4.06003, -3.94738, -3.66106, -3.46158, -3.47332, -3.50617, -3.2621, -3.01803, -2.84202, -2.66366,
        -2.42193, -2.10981, -1.85635, -1.59115, -1.30249, -1.10301, -0.903532, -0.769762, -0.603137, -0.781496, -0.570281, -0.183053, -0.037549, -0.049284] };
  WCL[1401] = struct {
   x = [103.838, 103.74, 103.596, 103.65, 103.772, 103.728, 103.596, 103.528, 103.617, 103.805, 103.96, 104.103, 104.27, 104.458, 104.556, 104.601, 104.613, 104.655, 104.622, 104.732, 104.899, 105.009,
        104.932, 104.899, 105.042, 105.209, 105.373, 105.54, 105.727, 105.77, 105.892, 105.969, 106.103, 106.145, 106.103, 105.992, 105.969, 106.025, 106.046, 105.992, 105.969, 105.948, 105.969,
        105.969, 105.969, 105.969, 105.969, 105.948, 105.924, 105.882, 105.781, 105.694, 105.582, 105.429, 105.373, 105.34, 105.218, 105.03, 104.866, 104.732, 104.8, 104.866, 104.777, 104.577, 104.446,
        104.291, 104.148, 104.026, 103.894, 103.718, 103.606, 103.43, 103.265, 103.087, 102.967, 102.789, 102.601, 102.47, 102.404, 102.303, 102.094, 101.918, 101.785, 101.62, 101.498, 101.365,
        101.221, 101.09, 101.012, 101.001, 100.98, 100.846, 100.67, 100.571, 100.538, 100.372, 100.261, 100.139, 100.029],
   y = [0.072752, -0.037549, -0.171319, -0.227643, -0.248765, -0.382534, -0.492835, -0.668848, -0.802618, -0.936387, -1.0021, -0.990364, -1.04669, -1.05842, -1.2579, -1.42453, -1.60054, -1.74605, -1.87747,
        -1.8892, -1.96665, -2.07695, -2.2647, -2.34449, -2.30929, -2.35388, -2.34449, -2.36561, -2.43132, -2.55336, -2.69886, -2.85375, -2.97579, -3.14006, -3.30669, -3.57188, -3.72677, -3.77136,
        -3.88166, -4.09288, -4.2243, -4.41205, -4.56694, -4.73357, -4.90958, -5.09733, -5.27334, -5.46108, -5.60659, -5.7826, -5.79199, -5.72628, -5.61598, -5.50568, -5.562, -5.72628, -5.75913,
        -5.61598, -5.50568, -5.52914, -5.66057, -5.81546, -5.86005, -5.69342, -5.55026, -5.41884, -5.28507, -5.10906, -4.9659, -4.86499, -4.84387, -4.68898, -4.52235, -4.41205, -4.3346, -4.19145,
        -4.114, -3.93799, -3.70565, -3.57188, -3.44046, -3.25037, -3.08609, -2.89834, -2.7411, -2.59795, -2.46418, -2.32102, -2.19898, -2.07695, -1.87747, -1.60054, -1.26964, -1.09128, -0.868329,
        -0.626605, -0.537425, -0.382534, -0.305089] };
  WCL[1402] = struct {
   x = [104.446, 104.425, 104.479, 104.556, 104.678, 104.655, 104.655, 104.544, 104.49, 104.645, 104.688, 104.699, 104.446],
   y = [-0.481101, -0.41539, -0.3708, -0.349678, -0.513957, -0.558547, -0.591402, -0.603137, -0.570281, -0.359066, -0.427124, -0.471714, -0.481101] };
  WCL[1403] = struct {
   x = [105.263, 105.319, 105.396, 105.429, 105.295, 105.45, 105.516, 105.617, 105.727, 105.748, 105.77, 105.847, 105.87, 105.87, 105.915, 106.025, 106.091, 106.19, 106.246, 106.312, 106.401, 106.511,
        106.61, 106.699, 106.809, 106.786, 106.753, 106.741, 106.786, 106.753, 106.687, 106.565, 106.389, 106.222, 106.112, 106.046, 106.058, 105.981, 105.882, 105.76, 105.65, 105.528, 105.396,
        105.319, 105.295, 105.263],
   y = [-1.8892, -1.87747, -1.80002, -1.72258, -1.92206, -1.62401, -1.56769, -1.53483, -1.56769, -1.6686, -1.71084, -1.71084, -1.68972, -1.59115, -1.49024, -1.49024, -1.65687, -1.81176, -2.08868, -2.27643,
        -2.39847, -2.45479, -2.50877, -2.5205, -2.50877, -2.66366, -2.79743, -2.91946, -3.01803, -3.0415, -3.06262, -2.95232, -2.87487, -2.80916, -2.70825, -2.55336, -2.44306, -2.28816, -2.14501,
        -2.08868, -2.02297, -2.01124, -1.97838, -1.95491, -1.92206, -1.8892] };
  WCL[1404] = struct {
   x = [107.79, 107.823, 107.912, 108.034, 108.133, 108.243, 108.342, 108.353, 108.332, 108.231, 108.123, 108.044, 107.957, 107.802, 107.736, 107.692, 107.68, 107.692, 107.736, 107.757, 107.79],
   y = [-2.54162, -2.50877, -2.50877, -2.53224, -2.59795, -2.64254, -2.75284, -2.90773, -3.0415, -3.1518, -3.19639, -3.08609, -3.13068, -3.14006, -3.08609, -2.95232, -2.8209, -2.71998, -2.6308, -2.57682,
        -2.54162] };
  WCL[1405] = struct {
   x = [110.01, 109.876, 109.757, 109.635, 109.546, 109.426, 109.28, 109.194, 109.128, 109.039, 108.982, 108.806, 108.708, 108.619, 108.553, 108.443, 108.276, 108.067, 107.912, 107.692, 107.448, 107.218,
        107.063, 106.875, 106.765, 106.643, 106.553, 106.478, 106.52, 106.643, 106.621, 106.467, 106.246, 106.103, 105.924, 105.77, 105.65, 105.561, 105.45, 105.385, 105.34, 105.34, 105.34, 105.352,
        105.373, 105.417, 105.483, 105.594, 105.704, 105.76, 105.847, 105.948, 105.992, 106.068, 106.168, 106.3, 106.478, 106.732, 106.941, 107.084, 107.185, 107.394, 107.504, 107.56, 107.736, 107.891,
        108.145, 108.321, 108.443, 108.565, 108.708, 108.795, 108.905, 108.994, 109.104, 109.203, 109.325, 109.468, 109.59, 109.811, 109.945],
   y = [-7.76333, -7.74221, -7.70935, -7.69762, -7.70935, -7.6765, -7.65303, -7.6765, -7.71874, -7.70935, -7.64364, -7.64364, -7.74221, -7.78445, -7.78445, -7.76333, -7.73048, -7.68588, -7.64364, -7.53334,
        -7.48875, -7.48875, -7.47702, -7.42304, -7.35733, -7.31274, -7.29162, -7.22356, -7.15785, -7.0053, -6.98184, -6.97245, -6.88327, -6.82929, -6.82929, -6.81756, -6.77297, -6.76358, -6.75185,
        -6.6744, -6.64154, -6.62981, -6.62042, -6.53124, -6.59695, -6.60869, -6.66266, -6.66266, -6.59695, -6.46553, -6.46553, -6.32237, -6.16748, -5.99147, -5.94688, -5.91402, -5.94688, -5.97974,
        -6.02433, -5.94688, -5.90229, -5.90229, -6.02433, -6.16748, -6.19095, -6.16748, -6.24493, -6.26605, -6.38809, -6.49839, -6.76358, -6.77297, -6.79643, -6.7847, -6.7847, -6.79643, -6.79643,
        -6.80582, -6.7847, -6.82929, -6.83868] };
  WCL[1406] = struct {
   x = [110.043, 109.987, 109.9, 109.811, 109.766, 109.689, 109.623, 109.513, 109.414, 109.316, 109.238, 109.203, 109.203],
   y = [-1.34708, -1.20158, -1.03495, -0.891797, -0.781496, -0.692316, -0.659461, -0.626605, -0.513957, -0.525691, -0.45998, -0.248765, -0.082139] };
  WCL[1407] = struct {
   x = [100.029, 99.9634],
   y = [-0.305089, -0.147851] };
  WCL[1408] = struct {
   x = [98.837, 98.9027, 99.0247, 99.0693, 99.1022, 99.2359, 99.3345, 99.4119, 99.3791, 99.2453, 99.135, 99.0153, 98.9027, 98.8041, 98.7712, 98.7712, 98.8041, 98.837],
   y = [-0.936387, -0.912919, -0.891797, -0.948121, -1.12413, -1.40106, -1.5231, -1.62401, -1.71084, -1.75543, -1.71084, -1.62401, -1.46912, -1.2579, -1.12413, -1.07954, -1.0021, -0.936387] };
  WCL[1409] = struct {
   x = [55.275, 55.2867, 55.2867, 55.275],
   y = [-4.56929, -4.56929, -4.5247, -4.56929] };
  WCL[1410] = struct {
   x = [43.1372, 42.8603, 42.8063, 42.5293, 42.6068, 42.6185, 42.6396, 42.5411, 42.4871, 42.3087, 42.1233, 42.0013, 41.8699, 41.7924],
   y = [0.19244, -0.161932, -0.150197, -0.272233, -0.161932, -0.194787, -0.361413, -0.384881, -0.560894, -0.74864, -0.97159, -1.12648, -1.30484, -1.34708] };
  WCL[1411] = struct {
   x = [41.7924, 41.6821, 41.5061, 41.3512, 41.1963, 41.0743, 40.9875, 40.8866, 40.8772, 40.6777, 40.436, 40.2928, 40.2482, 40.0933],
   y = [-1.34708, -1.63574, -1.84696, -1.9244, -1.95726, -1.95726, -2.06756, -2.20133, -2.37734, -2.47826, -2.60029, -2.85375, -3.17527, -3.50852] };
  WCL[1412] = struct {
   x = [46.4439, 46.4885, 46.357, 46.3007, 46.4439, 46.5776, 46.4439],
   y = [-9.41785, -9.39673, -9.39673, -9.48356, -9.4953, -9.40846, -9.41785] };
  WCL[1413] = struct {
   x = [34.1511, 34.062, 34.062, 34.008, 34.062, 34.0408, 33.9423, 33.9188, 33.9094, 33.8085, 33.7991, 33.7428, 33.6982, 33.6771, 33.555, 33.5011, 33.555, 33.4682, 33.379, 33.3462, 33.4119, 33.5011, 33.6114,
        33.6982, 33.7663, 33.8085, 33.7874, 33.5996, 33.4447, 33.1467, 33.0716, 33.0153, 33.027, 32.905, 32.8721, 32.8604, 32.7829, 32.6726, 32.5623, 32.3863, 32.2995, 32.1986, 32.2314, 32.1235,
        32.0554, 32.0554, 32.0014, 31.9123, 31.8254, 31.8348, 31.8137, 31.7691, 31.748, 31.748, 31.7691, 31.8466, 31.8583, 31.87, 31.7808],
   y = [-0.938734, -1.02557, -1.02557, -1.1476, -1.26964, -1.32596, -1.30484, -1.40341, -1.46912, -1.49259, -1.58177, -1.63574, -1.7038, -1.78125, -1.78125, -1.83523, -1.89155, -1.95726, -2.01358, -2.06756,
        -2.09103, -2.14501, -2.14501, -2.13562, -2.12389, -2.20133, -2.24592, -2.40081, -2.45479, -2.46652, -2.48999, -2.66601, -2.79977, -2.89834, -2.69886, -2.53224, -2.52285, -2.58856, -2.51111,
        -2.31163, -2.25766, -2.37734, -2.53224, -2.54397, -2.66601, -2.67774, -2.69886, -2.75518, -2.58856, -2.45479, -2.35622, -2.20133, -2.04644, -1.86808, -1.65921, -1.41514, -1.23678, -1.04903,
        -0.894144] };
  WCL[1414] = struct {
   x = [29.8611, 30.0606, 30.2601, 30.4455, 30.5886, 30.6121, 30.6004, 30.6778, 30.7553, 30.8773, 30.9641, 31.0299, 31.1636],
   y = [-6.38809, -6.51012, -6.60869, -6.79643, -6.96071, -7.12734, -7.36906, -7.57793, -7.7868, -8.00506, -8.22566, -8.32422, -8.55421] };
  WCL[1415] = struct {
   x = [34.0408, 34.0526, 34.184, 34.4046, 34.5595],
   y = [-9.69243, -9.61498, -9.56101, -9.74641, -9.99752] };
  WCL[1416] = struct {
   x = [34.0408, 34.0737],
   y = [-9.69243, -9.88722] };
  WCL[1417] = struct {
   x = [29.9949, 30.2366, 30.4126, 30.4807, 30.5793],
   y = [-7.23764, -7.31274, -7.53334, -7.87363, -8.14821] };
  WCL[1418] = struct {
   x = [30.6332, 30.6121, 30.5886, 30.6121, 30.6778, 30.8093, 31.0087, 31.1636],
   y = [-8.23504, -8.27963, -8.39932, -8.55421, -8.58707, -8.60819, -8.77247, -8.55421] };
  WCL[1419] = struct {
   x = [34.0197, 34.008, 34.0737, 34.1394, 34.1957, 34.3154, 34.4375, 34.6252, 34.813, 34.813, 34.6041, 34.4492, 34.2497, 34.1066, 34.1511],
   y = [0.114995, -0.028162, -0.105608, -0.194787, -0.349678, -0.328557, -0.272233, -0.239377, -0.251111, -0.340291, -0.417737, -0.539772, -0.539772, -0.783843, -0.938734] };
  WCL[1420] = struct {
   x = [32.4192, 32.3323, 32.1563, 31.9569, 31.9451, 31.9803, 31.9029, 31.8137, 31.7691, 31.7597, 31.7808],
   y = [0.103261, -0.028162, -0.09622, -0.14081, -0.239377, -0.349678, -0.506916, -0.706397, -0.804964, -0.849554, -0.894144] };
  WCL[1421] = struct {
   x = [40.0933, 39.8727, 39.753, 39.5864, 39.333],
   y = [-3.50852, -3.8934, -4.22665, -4.5247, -4.64673] };
  WCL[1422] = struct {
   x = [39.9291, 39.9619, 39.9502, 39.9291, 39.9502, 39.8727, 39.7742, 39.7953, 39.7859, 39.8399, 39.9056, 39.9173, 39.9291],
   y = [-5.08794, -4.88846, -4.95652, -5.05508, -5.20997, -5.41884, -5.3414, -5.10906, -4.88846, -4.92131, -4.88846, -4.99876, -5.08794] };
  WCL[1423] = struct {
   x = [39.3869, 39.4433, 39.4315, 39.4315, 39.4644, 39.5207, 39.5747, 39.6075, 39.631, 39.5536, 39.455, 39.3869, 39.3447, 39.333, 39.3869],
   y = [-5.76148, -5.81546, -5.73801, -5.91637, -6.12524, -6.20268, -6.21207, -6.33411, -6.43267, -6.38809, -6.30125, -6.25666, -6.02667, -5.8929, -5.76148] };
  WCL[1424] = struct {
   x = [39.333, 39.333, 39.3001, 39.2672, 39.1898, 39.1335, 39.0678, 38.9481, 38.8918, 38.9129, 38.9481, 39.1124, 39.2344, 39.3447, 39.4972, 39.6193, 39.6193, 39.5653, 39.4761, 39.455, 39.4221, 39.4761,
        39.4761, 39.4972, 39.4761, 39.3658, 39.333, 39.3775, 39.4761, 39.5653, 39.5653, 39.4972, 39.5418, 39.5981, 39.6427, 39.7413, 39.753, 39.7296, 39.8281, 39.8727, 39.8727, 39.9291, 39.9291],
   y = [-4.64673, -4.73357, -4.8556, -4.99876, -5.28742, -5.39772, -5.63945, -5.94923, -6.08065, -6.28013, -6.38809, -6.51012, -6.58757, -6.74246, -6.87388, -7.09448, -7.16958, -7.20478, -7.33621, -7.55446,
        -7.6765, -7.7868, -7.87363, -7.98393, -8.15995, -8.16933, -8.30076, -8.47677, -8.70676, -8.7842, -8.82879, -8.88277, -8.93675, -9.03532, -9.09164, -9.24418, -9.34275, -9.47417, -9.56101,
        -9.66896, -9.791, -9.85671, -9.99752] };
  WCL[1425] = struct {
   x = [29.6969, 29.6969, 29.6851, 29.7086, 29.786, 29.8517, 29.84, 29.8846, 29.962, 29.9057, 29.7954, 29.8611],
   y = [-4.38154, -4.48011, -4.71244, -5.01049, -5.0762, -5.17712, -5.43996, -5.68403, -5.91637, -6.03606, -6.14636, -6.38809] };
  WCL[1426] = struct {
   x = [29.3331, 29.4105, 29.4434, 29.4645, 29.4645, 29.4974, 29.5865, 29.6969],
   y = [-3.38648, -3.40761, -3.54137, -3.72912, -3.94972, -4.11635, -4.27124, -4.38154] };
  WCL[1427] = struct {
   x = [29.3331, 29.2885, 29.2556, 29.1782, 29.1571, 29.1665, 29.2439, 29.2768, 29.1665, 29.1571, 29.2674, 29.3871, 29.3777, 29.3331, 29.3988, 29.4974, 29.6194, 29.7508, 29.9949],
   y = [-3.38648, -3.47332, -3.66106, -3.9286, -4.13747, -4.24777, -4.19379, -4.39093, -4.78989, -5.08794, -5.41884, -5.66057, -5.88117, -6.11351, -6.41155, -6.54298, -6.84102, -7.0053, -7.23764] };
  WCL[1428] = struct {
   x = [13.2901, 13.1916, 13.1258, 13.0601, 13.0601, 13.147, 13.2807, 13.2901, 13.2807, 13.215, 13.1587, 13.093, 13.0155, 12.9498, 12.8841, 12.8395, 12.7832, 12.6635, 12.5415, 12.4523, 12.342, 12.288,
        12.288],
   y = [-10.0186, -9.73467, -9.4507, -9.18786, -9.06817, -8.94848, -8.81706, -8.6739, -8.50962, -8.25851, -8.1928, -8.0825, -7.8619, -7.63191, -7.41365, -7.19305, -7.04989, -6.895, -6.73072, -6.56644,
        -6.41155, -6.30125, -6.24728] };
  WCL[1429] = struct {
   x = [12.288, 12.2669, 12.3092, 12.2998, 12.2223],
   y = [-6.24728, -6.16983, -6.05953, -5.91637, -5.83893] };
  WCL[1430] = struct {
   x = [12.2223, 12.1331, 12.1003, 12.112, 12.0674, 11.9689],
   y = [-5.83893, -5.72863, -5.59485, -5.41884, -5.16538, -5.06682] };
  WCL[1431] = struct {
   x = [11.9689, 11.9243, 11.8468, 11.7905, 11.6825, 11.537, 11.4174, 11.2953, 11.0747],
   y = [-5.06682, -4.9659, -4.80162, -4.64673, -4.50358, -4.36981, -4.24777, -4.07176, -3.97085] };
  WCL[1432] = struct {
   x = [11.0747, 11.0419, 10.9527, 10.7673, 10.5232, 10.3026, 10.0726],
   y = [-3.97085, -3.91687, -3.71739, -3.54137, -3.26445, -3.06497, -2.88661] };
  WCL[1433] = struct {
   x = [10.0726, 9.87313, 9.72998, 9.56335, 9.36622, 9.24418, 9.16674, 9.06817, 8.99072, 8.90389, 8.81471, 8.749, 8.82645, 8.95787, 9.06817, 9.19959, 9.23245],
   y = [-2.88661, -2.7106, -2.48999, -2.27878, -2.03471, -1.9244, -1.74605, -1.49259, -1.24852, -1.07016, -0.870676, -0.760375, -0.74864, -0.694663, -0.539772, -0.429471, -0.194787] };
  WCL[1434] = struct {
   x = [-40.0628, -39.9314, -39.8211, -39.7437, -39.7319, -39.7319, -39.7319, -39.7319, -39.6991, -39.624, -39.5137, -39.3588, -39.2602, -39.2696, -39.2931, -39.2696, -39.2602, -39.2696, -39.225, -39.1499,
        -39.1264, -39.1264, -39.0818, -39.0396, -39.0161, -39.0161, -39.0725, -39.1147, -39.1147, -39.1147, -39.1147, -39.1147, -39.1264, -39.1382, -39.1593, -39.1593, -39.1264, -39.1147, -39.1053,
        -39.0725, -39.0161, -39.0161, -39.0161, -39.0161, -38.873, -38.6852, -38.6195, -38.5867, -38.4552, -38.3097, -38.1783, -38.0469, -37.9366, -37.838, -37.7606, -37.6925, -37.594, -37.5728,
        -37.54, -37.4508, -37.2959],
   y = [-19.7791, -19.6641, -19.5374, -19.3614, -19.1736, -18.9742, -18.7629, -18.5236, -18.3241, -18.1223, -18.0073, -17.8594, -17.7022, -17.5966, -17.437, -17.2375, -17.0357, -16.9934, -16.855, -16.632,
        -16.4302, -16.2917, -16.165, -16.0899, -15.8975, -15.7707, -15.6651, -15.536, -15.4281, -15.2591, -15.1183, -14.9916, -14.9798, -14.8836, -14.7757, -14.595, -14.3368, -14.1233, -13.8862,
        -13.7243, -13.7126, -13.5412, -13.3699, -13.2714, -13.1212, -12.9264, -12.9264, -13.0461, -12.9803, -12.8724, -12.6987, -12.5274, -12.342, -12.1918, -12.0393, -11.9102, -11.7037, -11.6708,
        -11.584, -11.4643, -11.2155] };
  WCL[1435] = struct {
   x = [-37.2959, -37.1833, -37.1011, -37.0284, -37.0166, -36.9298, -36.6435, -36.5097, -36.3008, -36.1577],
   y = [-11.2155, -11.1334, -11.0677, -10.9996, -10.9785, -10.8377, -10.6617, -10.5866, -10.3683, -10.1501] };
  WCL[1436] = struct {
   x = [-40.2506, -40.1849, -40.152, -40.0628],
   y = [-20.1851, -19.9645, -19.8613, -19.7791] };
  WCL[1437] = struct {
   x = [-70.0976, -69.9756, -69.8981, -69.8207, -69.7104, -69.5672, -69.4569, -69.3701, -69.2363, -69.1706, -69.0063, -68.8069, -68.7529, -68.8632, -68.9054, -68.7951, -68.6754, -68.6848, -68.7857, -68.8514,
        -69.0063, -69.1378, -69.2152, -69.3466, -69.4804, -69.5461, -69.6447, -69.7667, -69.8653, -69.8864],
   y = [-15.407, -15.3436, -15.3342, -15.2051, -15.3342, -15.4398, -15.5149, -15.5994, -15.6651, -15.7919, -15.813, -15.8247, -15.8763, -15.9726, -16.0265, -16.111, -16.1955, -16.3786, -16.334, -16.4091,
        -16.4513, -16.3457, -16.2401, -16.1955, -16.111, -16.0054, -15.9092, -15.813, -15.7919, -15.813] };
  WCL[1438] = struct {
   x = [-77.7694, -77.7483, -77.7154, -77.7154, -77.7037, -77.7037, -77.638, -77.4831, -77.3634, -77.2625, -77.1757, -77.0419, -76.9433, -76.8659, -76.7462, -76.6453, -76.5678, -76.4364, -76.3472, -76.2815,
        -76.3144, -76.481, -76.4247, -76.2604, -76.1501, -76.0515, -75.9295, -75.7863, -75.6197, -75.4554, -75.3123, -75.1691, -74.9931, -74.7725, -74.5848, -74.4299, -74.1647, -74.0661, -74.0333,
        -73.9887, -73.8784, -73.7024, -73.5146, -73.294, -73.0851, -72.8411, -72.59, -72.3787, -72.1816, -71.9164, -71.7169, -71.5738, -71.4095, -71.3203, -71.1443, -70.8791, -70.7266, -70.5811,
        -70.5811],
   y = [-11.1263, -11.1592, -11.2695, -11.3235, -11.3657, -11.422, -11.5089, -11.584, -11.7459, -11.9853, -12.1918, -12.3209, -12.5039, -12.6236, -12.7973, -12.9592, -13.1, -13.2385, -13.4544, -13.6375,
        -13.8111, -13.9824, -14.1655, -14.3274, -14.4659, -14.6278, -14.7663, -14.8297, -14.9705, -15.1418, -15.3014, -15.4187, -15.5572, -15.6417, -15.7825, -15.8341, -15.9726, -16.0899, -16.1321,
        -16.1439, -16.1861, -16.2495, -16.3457, -16.4302, -16.5147, -16.632, -16.7071, -16.855, -17.0028, -17.1413, -17.2375, -17.4276, -17.5754, -17.7022, -17.8594, -18.0589, -18.1645, -18.249,
        -18.2607] };
  WCL[1439] = struct {
   x = [-70.5811, -70.5717, -70.5271, -70.4731, -70.4379, -70.4051, -70.3628, -70.3393, -70.3393, -70.3276, -70.2619],
   y = [-18.2607, -18.2607, -18.3546, -18.533, -18.7442, -18.9742, -19.1924, -19.4341, -19.5796, -19.6735, -19.8308] };
  WCL[1440] = struct {
   x = [-70.0413, -70.0742, -70.0742, -70.0976, -70.0976],
   y = [-15.5994, -15.59, -15.5689, -15.4516, -15.407] };
  WCL[1441] = struct {
   x = [-69.8864, -70.0319, -70.053, -70.0413],
   y = [-15.813, -15.8036, -15.6862, -15.5994] };
  WCL[1442] = struct {
   x = [-78.2294, -78.1965, -78.1426, -78.0651, -77.9759, -77.8563, -77.8328, -77.8093, -77.7835, -77.7694],
   y = [-9.98578, -10.1172, -10.2275, -10.3801, -10.5091, -10.6851, -10.901, -10.9996, -11.0724, -11.1263] };
  WCL[1443] = struct {
   x = [-149.454, -149.421, -149.444, -149.444, -149.388, -149.289, -149.355, -149.421, -149.576, -149.674, -149.674, -149.554, -149.454],
   y = [-17.5684, -17.5473, -17.6552, -17.6975, -17.728, -17.7914, -17.9181, -17.7702, -17.7702, -17.6435, -17.505, -17.5379, -17.5684] };
  WCL[1444] = struct {
   x = [-145.64, -145.628, -145.562, -145.541, -145.628, -145.673, -145.694, -145.64, -145.673, -145.694, -145.64],
   y = [-17.3267, -17.3361, -17.3994, -17.4957, -17.505, -17.39, -17.2938, -17.3267, -17.39, -17.2938, -17.3267] };
  WCL[1445] = struct {
   x = [-145.121, -145.121],
   y = [-16.0735, -16.043] };
  WCL[1446] = struct {
   x = [-146.433, -146.511, -146.522, -146.522, -146.532, -146.466, -146.367, -146.433],
   y = [-14.4283, -14.4495, -14.4495, -14.5034, -14.5785, -14.5574, -14.4401, -14.4283] };
  WCL[1447] = struct {
   x = [-146.114, -146.18, -146.201, -146.191, -146.114, -146.025, -146.037, -146.114],
   y = [-14.3744, -14.3955, -14.4401, -14.5034, -14.4917, -14.3955, -14.3532, -14.3744] };
  WCL[1448] = struct {
   x = [-140.336, -140.303, -140.303, -140.348, -140.357, -140.282, -140.249, -140.27, -140.336],
   y = [-15.989, -15.9256, -15.9467, -16.0007, -16.043, -16.0312, -15.935, -15.9139, -15.989] };
  WCL[1449] = struct {
   x = [-140.59, -140.568, -140.524, -140.536, -140.568, -140.578],
   y = [-19.6782, -19.6899, -19.6383, -19.5961, -19.6055, -19.6172] };
  WCL[1450] = struct {
   x = [-145.067, -145.121, -145.133, -145.166, -145.142, -145.077, -145.067],
   y = [-19.8143, -19.8049, -19.8143, -19.8777, -19.9293, -19.8988, -19.8143] };
  WCL[1451] = struct {
   x = [-150.005, -149.993, -149.939, -149.918, -149.96],
   y = [-17.5684, -17.5895, -17.559, -17.4957, -17.4839] };
  WCL[1452] = struct {
   x = [-151.627, -151.636, -151.57, -151.516, -151.484, -151.505, -151.582, -151.627],
   y = [-16.7447, -16.7329, -16.7235, -16.7541, -16.8292, -16.8808, -16.8385, -16.7447] };
  WCL[1453] = struct {
   x = [-150.005, -150.005],
   y = [-17.505, -17.5684] };
  WCL[1454] = struct {
   x = [-149.96, -150.005],
   y = [-17.4839, -17.505] };
  WCL[1455] = struct {
   x = [-160.908, -160.897, -160.854, -160.81, -160.81, -160.876, -160.908, -160.908],
   y = [-10.3683, -10.3237, -10.3472, -10.3894, -10.4223, -10.434, -10.4012, -10.3683] };
  WCL[1456] = struct {
   x = [-172.828, -172.673, -172.518, -172.452, -172.375, -172.387, -172.452, -172.518, -172.628, -172.762, -172.873, -172.959, -173.037, -173.093, -173.081, -173.049, -172.938, -172.828],
   y = [-13.4708, -13.492, -13.5154, -13.5999, -13.7079, -13.8698, -13.8369, -13.8064, -13.8276, -13.8158, -13.7407, -13.6116, -13.5577, -13.4708, -13.4614, -13.492, -13.4826, -13.4708] };
  WCL[1457] = struct {
   x = [-171.946, -171.88, -171.802, -171.758, -171.725, -171.626, -171.605, -171.68, -171.88, -171.978, -172.056, -172.133, -172.232, -172.276, -172.211, -172.112, -172.035, -171.946],
   y = [-13.8487, -13.8487, -13.8698, -13.8815, -13.9355, -13.9449, -14.0317, -14.0528, -14.0622, -14.0411, -14.0317, -14.0411, -13.9778, -13.8909, -13.8276, -13.8369, -13.8369, -13.8487] };
  WCL[1458] = struct {
   x = [-170.016, -170.016, -170.005, -170.005, -170.016, -170.094, -170.115, -170.094, -170.038, -170.016],
   y = [-18.9695, -18.885, -18.9789, -19.0845, -19.1267, -19.1056, -18.9577, -18.8944, -18.8732, -18.9695] };
  WCL[1459] = struct {
   x = [177.789, 177.81, 177.878, 177.965, 178.031, 178.066, 178.108, 178.176, 178.287, 178.352, 178.385, 178.395, 178.418, 178.451, 178.43, 178.32, 178.176, 178.066, 177.965, 177.867, 177.777, 177.679,
        177.58, 177.524, 177.437, 177.404, 177.249, 177.16, 177.127, 177.172, 177.205, 177.303, 177.437, 177.515, 177.59, 177.7, 177.789],
   y = [-17.4206, -17.4323, -17.4206, -17.3783, -17.3572, -17.4206, -17.4534, -17.5262, -17.5379, -17.613, -17.7069, -17.728, -17.8641, -18.0331, -18.1176, -18.1176, -18.1293, -18.2232, -18.2443, -18.2443,
        -18.2326, -18.2232, -18.2232, -18.1504, -18.1293, -18.1293, -18.1176, -18.0331, -17.9064, -17.7914, -17.6552, -17.5684, -17.4628, -17.4417, -17.4206, -17.4417, -17.4206] };
  WCL[1460] = struct {
   x = [178.749, 178.826, 178.826, 178.848, 178.97, 179.113, 179.277, 179.453, 179.599, 179.84, 179.784, 179.631, 179.509, 179.42, 179.387, 179.477, 179.608, 179.697, 179.697, 179.697, 179.73, 179.62,
        179.488, 179.301, 179.157, 179.134, 179.08, 178.97, 178.904, 178.815, 178.726, 178.639, 178.505, 178.463, 178.352, 178.352, 178.517, 178.639, 178.749],
   y = [-16.4677, -16.5217, -16.51, -16.4677, -16.4138, -16.3621, -16.3082, -16.2331, -16.2025, -16.1814, -16.2542, -16.3504, -16.4466, -16.5945, -16.6602, -16.5945, -16.51, -16.4677, -16.5734, -16.6484,
        -16.7658, -16.7541, -16.7541, -16.7658, -16.7869, -16.6907, -16.6602, -16.7541, -16.9136, -16.923, -16.9019, -16.9887, -16.8808, -16.8174, -16.7869, -16.6602, -16.6273, -16.5522, -16.4677] };
  WCL[1461] = struct {
   x = [159.967, 160.089, 160.209, 160.308, 160.319, 160.209, 160.089],
   y = [-11.7318, -11.7318, -11.7647, -11.8398, -11.9055, -11.8961, -11.8398] };
  WCL[1462] = struct {
   x = [166.445, 166.477, 166.51, 166.531, 166.531, 166.576, 166.599, 166.609, 166.632, 166.698, 166.775, 166.82, 166.853, 166.898, 166.94, 166.952, 166.963, 166.996, 167.008, 167.029, 167.041, 167.05,
        167.074, 167.029, 166.93, 166.862, 166.752, 166.665, 166.599, 166.531, 166.51, 166.477, 166.456, 166.445],
   y = [-14.7287, -14.7076, -14.7405, -14.8156, -14.8672, -14.9423, -15.0291, -15.1136, -15.1793, -15.2098, -15.1676, -15.0831, -15.008, -14.9634, -15.0174, -15.0831, -15.1254, -15.1347, -15.1793, -15.2216,
        -15.2756, -15.3812, -15.4774, -15.5196, -15.5314, -15.5314, -15.5947, -15.6065, -15.4351, -15.3061, -15.0831, -14.9329, -14.8038, -14.7287] };
  WCL[1463] = struct {
   x = [167.062, 167.083, 167.151, 167.205, 167.238, 167.271, 167.315, 167.336, 167.393, 167.437, 167.503, 167.569, 167.59, 167.613, 167.634, 167.602, 167.524, 167.414, 167.294, 167.271, 167.238, 167.226,
        167.172, 167.083, 167.062, 167.062],
   y = [-15.8716, -15.8623, -15.8716, -15.9139, -15.9796, -16.0312, -16.043, -16.0735, -16.1392, -16.2119, -16.2542, -16.3082, -16.341, -16.3715, -16.4372, -16.4889, -16.4889, -16.5006, -16.5006, -16.3199,
        -16.1486, -16.0735, -16.0641, -16.0101, -15.9256, -15.8716] };
  WCL[1464] = struct {
   x = [168.031, 168.055, 168.055, 168.064, 168.099, 168.132, 168.132, 168.132, 168.099, 168.055, 168.031, 168.022, 168.031],
   y = [-15.4774, -15.4563, -15.5102, -15.5947, -15.7027, -15.7449, -15.8083, -15.8928, -15.9139, -15.82, -15.6792, -15.5196, -15.4774] };
  WCL[1465] = struct {
   x = [168.022, 168.055, 168.055, 168.055, 168.064, 168.108, 168.141, 168.055, 167.965, 167.879, 167.834, 167.834, 167.911, 167.965, 168.022],
   y = [-16.0852, -16.0524, -16.0524, -16.1392, -16.2237, -16.287, -16.3293, -16.3504, -16.3504, -16.3082, -16.2237, -16.1908, -16.1697, -16.1063, -16.0852] };
  WCL[1466] = struct {
   x = [168.308, 168.329, 168.374, 168.395, 168.407, 168.329, 168.252, 168.153, 168.087, 168.099, 168.153, 168.242, 168.308],
   y = [-17.5473, -17.559, -17.5895, -17.6764, -17.7702, -17.8007, -17.8007, -17.7914, -17.728, -17.6646, -17.5895, -17.5379, -17.5473] };
  WCL[1467] = struct {
   x = [168.836, 168.836, 168.859, 168.946, 168.979, 168.97, 168.991, 169.014, 168.97, 168.869, 168.827, 168.836, 168.869, 168.827, 168.836],
   y = [-18.6433, -18.6221, -18.6221, -18.655, -18.7277, -18.7489, -18.8427, -18.9366, -18.9366, -18.8427, -18.7277, -18.6433, -18.8427, -18.7277, -18.6433] };
  WCL[1468] = struct {
   x = [169.101, 169.146, 169.158, 169.223, 169.256, 169.233, 169.212, 169.125, 169.08, 169.08, 169.101],
   y = [-19.3661, -19.345, -19.3755, -19.4388, -19.4811, -19.5538, -19.6055, -19.5656, -19.4905, -19.4083, -19.3661] };
  WCL[1469] = struct {
   x = [161.488, 161.467, 161.566, 161.676, 161.842, 162.009, 162.173, 162.239, 162.349, 162.385, 162.439, 162.349, 162.129, 161.932, 161.786, 161.643, 161.556, 161.488, 161.488],
   y = [-10.1829, -10.1712, -10.2486, -10.258, -10.3237, -10.3801, -10.4458, -10.5748, -10.6523, -10.7274, -10.772, -10.8471, -10.7814, -10.6945, -10.5209, -10.4458, -10.2815, -10.1947, -10.1829] };
  WCL[1470] = struct {
   x = [165.729, 165.741, 165.75, 165.795, 165.851, 165.961, 166.071, 166.081, 165.994, 165.938, 165.839, 165.827, 165.783, 165.729],
   y = [-10.8048, -10.8377, -10.7814, -10.718, -10.6523, -10.6523, -10.7063, -10.7931, -10.826, -10.8142, -10.8799, -10.826, -10.826, -10.8048] };
  WCL[1471] = struct {
   x = [159.979, 159.878, 159.911, 159.967],
   y = [-11.8398, -11.8069, -11.7764, -11.7318] };
  WCL[1472] = struct {
   x = [160.089, 159.979],
   y = [-11.8398, -11.8398] };
  WCL[1473] = struct {
   x = [151.376, 151.387, 151.244, 151.122],
   y = [-9.96466, -10.1055, -10.1172, -10.0186] };
  WCL[1474] = struct {
   x = [149.953, 150.174, 150.329, 150.517, 150.669, 150.815, 150.979, 151.012, 150.824, 150.615, 150.481, 150.615, 150.77, 150.681, 150.439, 150.174, 150.118, 150.106],
   y = [-9.89895, -10.0186, -10.0961, -10.1618, -10.1712, -10.1947, -10.2275, -10.3143, -10.3683, -10.3472, -10.3566, -10.3894, -10.4997, -10.6077, -10.6851, -10.6077, -10.488, -10.4223] };
  WCL[1475] = struct {
   x = [139.867, 140.045, 140.275, 140.475, 140.782, 140.859, 140.916, 140.937, 141.059, 141.169, 141.214, 141.279, 141.357, 141.444, 141.545, 141.587, 141.554, 141.455, 141.488, 141.467, 141.444, 141.545,
        141.61, 141.587, 141.587, 141.676, 141.697, 141.578, 141.554, 141.578, 141.655, 141.852, 141.93, 142, 142.225, 142.225, 142.225],
   y = [-17.7186, -17.782, -17.7303, -17.6459, -17.5496, -17.4346, -17.1694, -17.0638, -16.8409, -16.6391, -16.4584, -16.2049, -15.8411, -15.5431, -15.3295, -15.116, -14.8484, -14.5152, -14.2359, -14.0106,
        -13.6562, -13.4614, -13.3112, -13.1282, -13.0085, -12.8255, -12.6518, -12.6213, -12.6213, -12.4148, -12.0909, -12.0252, -11.8633, -11.5253, -11.2554, -11.0254, -10.9996] };
  WCL[1476] = struct {
   x = [142.723, 142.723, 142.746, 142.767, 142.779, 142.779, 142.833, 143.033, 143.077, 143.065, 143.042, 143.152, 143.23, 143.319, 143.319, 143.331, 143.417, 143.462, 143.539, 143.539, 143.528, 143.593,
        143.626, 143.704, 143.838, 144.002, 144.124, 144.234, 144.398, 144.553, 144.619, 144.675, 144.851, 144.995, 145.138, 145.304, 145.26, 145.314, 145.293, 145.281, 145.347, 145.358, 145.436,
        145.447, 145.447, 145.501, 145.623, 145.734, 145.877, 145.91, 145.987, 146.053, 146.098, 146.098, 146.107, 146.074, 146.065, 146.107, 146.229, 146.318, 146.295, 146.351, 146.351, 146.318,
        146.339, 146.339, 146.405, 146.649, 146.726, 146.902, 147.079, 147.255, 147.332, 147.431, 147.496, 147.564, 147.663, 147.761, 147.839, 147.982, 148.149],
   y = [-11.0254, -11.0489, -11.124, -11.2648, -11.5159, -11.753, -11.9266, -11.9266, -12.0463, -12.166, -12.2951, -12.3936, -12.5767, -12.6424, -12.8255, -12.8912, -12.9334, -13.2033, -13.4192, -13.56,
        -13.7853, -13.9355, -14.1092, -14.2899, -14.4307, -14.4941, -14.3861, -14.3133, -14.2476, -14.2805, -14.4096, -14.5598, -14.6349, -14.8062, -14.8813, -15.0291, -15.2122, -15.3178, -15.4891,
        -15.651, -15.82, -15.9702, -16.1298, -16.3105, -16.4161, -16.6179, -16.7658, -16.9254, -16.9254, -17.1483, -17.3079, -17.498, -17.6975, -17.8782, -17.9838, -18.1082, -18.2772, -18.4672,
        -18.5705, -18.5916, -18.6761, -18.6127, -18.655, -18.7301, -18.77, -18.8756, -19.054, -19.1267, -19.2323, -19.3051, -19.3356, -19.3896, -19.3778, -19.3051, -19.5984, -19.7439, -19.8167,
        -19.7956, -19.765, -19.8589, -19.8894] };
  WCL[1477] = struct {
   x = [142.225, 142.225, 142.246, 142.303, 142.401, 142.544, 142.622, 142.72, 142.72, 142.727, 142.744, 142.763, 142.727, 142.727],
   y = [-10.9996, -10.9785, -10.9574, -10.9128, -10.8917, -10.8048, -10.7063, -10.7509, -10.8471, -10.8588, -10.9245, -10.9902, -10.9996, -11.0254] };
  WCL[1478] = struct {
   x = [150.106, 149.852, 149.622, 149.378, 149.158, 148.97, 148.848, 148.541, 148.386, 148.174, 147.987, 147.9],
   y = [-10.4223, -10.3894, -10.3566, -10.3026, -10.2815, -10.2369, -10.204, -10.1947, -10.1618, -10.1501, -10.0515, -10.0515] };
  WCL[1479] = struct {
   x = [129.923, 130.111, 130.231, 130.209, 130.066, 130.099, 130.111, 130.188, 130.353, 130.508, 130.54, 130.684, 130.806, 130.883, 130.904, 130.794, 130.86, 130.958, 131.068, 131.202, 131.223, 131.28,
        131.435, 131.664, 131.841, 131.995, 132.16, 132.237, 132.347, 132.514, 132.502, 132.502, 132.392, 132.204, 132.061, 131.93, 131.841],
   y = [-13.5811, -13.5811, -13.4614, -13.2455, -13.2573, -13.1493, -12.978, -12.7926, -12.7504, -12.6846, -12.5438, -12.5767, -12.6518, -12.7832, -12.6518, -12.4805, -12.4476, -12.2951, -12.3279, -12.3819,
        -12.2857, -12.2857, -12.3279, -12.3068, -12.349, -12.349, -12.2951, -12.2739, -12.2317, -12.1331, -11.9383, -11.699, -11.591, -11.5699, -11.6028, -11.5042, -11.3962] };
  WCL[1480] = struct {
   x = [131.841, 131.676, 131.688, 131.742, 131.819, 131.852, 131.909, 131.951, 131.974, 132.007, 132.085, 132.085, 132.139, 132.171, 132.228, 132.315, 132.371, 132.448, 132.547, 132.601, 132.645, 132.735,
        132.812, 132.878, 133.009, 133.143, 133.274, 133.34, 133.418, 133.483, 133.605, 133.671, 133.671, 133.671, 133.77, 133.936, 134.014, 134.112, 134.279, 134.422, 134.586, 134.631, 134.729,
        134.929, 135.016, 135.105, 135.138, 135.204, 135.314, 135.391, 135.457, 135.502, 135.546, 135.645, 135.776, 135.832, 135.767, 135.722, 135.755, 135.8, 135.898, 135.922, 135.922, 135.898,
        135.898, 135.922, 135.976, 136.086, 136.196, 136.163, 136.175, 136.361, 136.417, 136.461, 136.483, 136.626, 136.736, 136.837, 136.781, 136.715, 136.616, 136.539, 136.461, 136.429, 136.471,
        136.494, 136.539, 136.471, 136.417, 136.339, 136.285, 136.196, 136.152, 136.074, 136.098, 136.131, 136.065, 135.987, 135.898, 135.865, 135.811, 135.811, 135.886, 135.987, 135.955, 135.886,
        135.886, 135.844, 135.71, 135.656, 135.534, 135.534, 135.645, 135.71, 135.854, 135.943, 135.955, 136.008, 136.119, 136.285, 136.328, 136.396, 136.396, 136.572, 136.769, 136.968, 137.21,
        137.288, 137.475, 137.675, 137.839, 137.949, 138.104, 138.247, 138.379, 138.677, 138.886, 139.062, 139.13, 139.207, 139.35, 139.515, 139.747, 139.78, 139.867],
   y = [-11.3962, -11.3305, -11.3094, -11.2977, -11.3305, -11.2554, -11.1991, -11.2648, -11.4174, -11.5253, -11.4056, -11.2554, -11.232, -11.232, -11.3094, -11.2977, -11.3962, -11.4831, -11.5699, -11.6028,
        -11.4713, -11.4502, -11.5042, -11.6239, -11.7647, -11.7858, -11.8304, -11.8187, -11.8398, -11.8961, -11.8069, -11.7858, -11.8069, -11.9055, -11.9501, -11.9595, -12.0252, -12.0909, -12.058,
        -12.1214, -12.0463, -12.0463, -12.166, -12.166, -12.22, -12.3279, -12.2951, -12.2317, -12.1331, -12.0909, -12.0674, -12.0792, -12.0792, -12.0792, -12.0252, -12.0041, -12.112, -12.1871,
        -12.2411, -12.2622, -12.1777, -12.2317, -12.2951, -12.403, -12.5344, -12.5978, -12.5673, -12.5344, -12.3397, -12.2951, -12.1331, -11.9712, -12.0252, -12.112, -12.3819, -12.2857, -12.2739,
        -12.3608, -12.4359, -12.6095, -12.7926, -12.8466, -12.8372, -12.8912, -12.9451, -12.9991, -13.0414, -13.1704, -13.2455, -13.2901, -13.2784, -13.2573, -13.1611, -13.269, -13.4309, -13.506,
        -13.4403, -13.3112, -13.2995, -13.5154, -13.6774, -13.7736, -13.7525, -13.7313, -13.8604, -14.0646, -14.1303, -14.2359, -14.4729, -14.6771, -14.8484, -15.0291, -15.1042, -15.2239, -15.1911,
        -15.2756, -15.2873, -15.3295, -15.3835, -15.4375, -15.6182, -15.7684, -16.0125, -15.8951, -15.9491, -15.9913, -15.9796, -16.1087, -16.226, -16.3105, -16.4067, -16.5428, -16.6719, -16.693,
        -16.7658, -16.8315, -16.8831, -17.0099, -17.2023, -17.3173, -17.4135, -17.5191, -17.6552, -17.6764, -17.7186] };
  WCL[1481] = struct {
   x = [130.055, 130.001],
   y = [-11.8398, -11.8515] };
  WCL[1482] = struct {
   x = [129.968, 130.01, 130.033, 130.111, 130.177, 130.221, 130.287, 130.364, 130.451, 130.409, 130.209, 130.055],
   y = [-11.699, -11.6779, -11.5253, -11.4174, -11.4291, -11.5699, -11.6779, -11.753, -11.8304, -11.8398, -11.8304, -11.8398] };
  WCL[1483] = struct {
   x = [130.254, 130.299, 130.397, 130.463, 130.552, 130.618, 130.749, 130.871, 130.958, 131.08, 131.179, 131.268, 131.357, 131.378, 131.39, 131.312, 131.247, 131.104, 130.97, 130.827, 130.705, 130.552,
        130.43, 130.364, 130.308, 130.254],
   y = [-11.2554, -11.3094, -11.3516, -11.4174, -11.4502, -11.4713, -11.4385, -11.4056, -11.4174, -11.3423, -11.2554, -11.2859, -11.3962, -11.4713, -11.5793, -11.6567, -11.753, -11.8304, -11.9055, -11.9055,
        -11.8069, -11.7107, -11.645, -11.5159, -11.3845, -11.2554] };
  WCL[1484] = struct {
   x = [136.429, 136.396, 136.396, 136.417, 136.527, 136.593, 136.593, 136.483, 136.429],
   y = [-11.5253, -11.5253, -11.4502, -11.2977, -11.1357, -11.1123, -11.2437, -11.3962, -11.5253] };
  WCL[1485] = struct {
   x = [136.405, 136.461, 136.483, 136.572, 136.626, 136.759, 136.781, 136.802, 136.87, 136.703, 136.494, 136.396, 136.405],
   y = [-13.8182, -13.8182, -13.8721, -13.7431, -13.8933, -13.797, -14.0857, -14.1937, -14.2594, -14.3227, -14.3133, -14.1843, -13.8182] };
  WCL[1486] = struct {
   x = [139.58, 139.67, 139.681, 139.78, 139.702, 139.536, 139.536, 139.47, 139.294, 139.207, 139.318, 139.404, 139.482, 139.58],
   y = [-16.4372, -16.4701, -16.4795, -16.5663, -16.5663, -16.6179, -16.6719, -16.7235, -16.7775, -16.6508, -16.5217, -16.5006, -16.4161, -16.4372] };
  WCL[1487] = struct {
   x = [119.879, 120.045, 120.221, 120.475, 120.663, 120.871, 121.015, 121.113, 121.268, 121.322, 121.432, 121.489, 121.51, 121.599, 121.653, 121.688, 121.697, 121.82, 121.974, 122.204, 122.261, 122.195,
        122.15, 122.094, 122.061, 122.085, 122.228, 122.305, 122.416, 122.47, 122.624, 122.678, 122.702, 122.678, 122.756, 122.8, 122.845, 122.923, 122.965, 123.033, 123.153, 123.263, 123.34, 123.418,
        123.495, 123.507, 123.549, 123.683, 123.65, 123.516, 123.418, 123.462, 123.451, 123.439, 123.528, 123.716, 123.871, 123.957, 124.091, 124.234, 124.387, 124.432, 124.267, 124.3, 124.378,
        124.488, 124.542, 124.509, 124.455, 124.542, 124.685, 124.884, 124.929, 124.873, 124.828, 124.84, 124.917, 125.072, 125.237, 125.237, 125.126, 125.117, 125.269, 125.359, 125.481, 125.523,
        125.645, 125.666, 125.767, 125.898, 125.987, 125.987, 125.887, 125.943, 126.03, 126.065, 126.175, 126.241, 126.286, 126.307, 126.417, 126.527, 126.626, 126.692, 126.649, 126.802, 126.858,
        126.968, 127.013, 127.189, 127.365, 127.508, 127.663, 127.806, 127.961, 128.06, 128.039, 127.994, 127.994, 127.961, 127.905, 127.818, 127.884, 127.971, 128.015, 128.048, 128.081, 128.114,
        128.104, 128.104, 128.125, 128.191, 128.247, 128.269, 128.391, 128.379, 128.445, 128.623, 128.853, 129.062, 129.184, 129.261, 129.383, 129.503, 129.592, 129.536, 129.67, 129.681, 129.559,
        129.437, 129.282, 129.372, 129.426, 129.515, 129.604, 129.658, 129.714, 129.714, 129.801, 129.923],
   y = [-19.9411, -19.9105, -19.8049, -19.765, -19.7228, -19.6289, -19.5351, -19.4412, -19.3051, -19.2628, -19.1689, -19.054, -18.8873, -18.7606, -18.6878, -18.6127, -18.4978, -18.5095, -18.3523, -18.141,
        -18.0448, -18.0026, -17.8148, -17.6247, -17.4769, -17.2657, -17.0427, -16.9887, -17.0005, -16.8737, -16.7447, -16.7447, -16.7447, -16.6508, -16.4584, -16.4372, -16.6086, -16.7142, -16.8315,
        -16.9887, -17.2117, -17.3384, -17.4863, -17.5708, -17.3807, -17.1178, -17.1272, -17.0943, -16.8831, -16.6813, -16.6086, -16.5217, -16.4795, -16.2354, -16.1603, -16.2471, -16.2143, -16.3434,
        -16.5006, -16.5123, -16.5428, -16.4161, -16.3434, -16.1298, -16.0453, -15.9491, -15.8411, -15.7355, -15.5431, -15.4375, -15.3929, -15.5008, -15.4586, -15.3506, -15.245, -15.1371, -15.0526,
        -15.0948, -15.1582, -15.0737, -14.8907, -14.6771, -14.6114, -14.602, -14.5598, -14.548, -14.548, -14.6982, -14.6443, -14.5386, -14.3345, -14.2359, -14.1092, -14.0012, -14.1092, -14.2594,
        -14.1843, -14.0763, -14.0974, -14.1843, -14.1725, -14.1092, -14.0552, -13.9895, -13.8604, -13.8393, -13.8815, -13.9261, -13.9472, -13.9472, -14.0552, -14.1725, -14.3767, -14.5692, -14.6771,
        -14.7522, -14.8578, -15.0526, -15.245, -15.414, -15.5548, -15.651, -15.6182, -15.4375, -15.2873, -15.3084, -15.2544, -15.1254, -15.0948, -15.0737, -15.0526, -14.9869, -14.9986, -14.9658,
        -14.9986, -14.8813, -14.8367, -14.9024, -14.8907, -14.8578, -14.9329, -15.0197, -15.1371, -15.2544, -15.2333, -14.9564, -14.8907, -14.8578, -14.7311, -14.602, -14.3861, -14.1092, -14.1397,
        -14.0974, -13.9801, -13.8815, -13.797, -13.668, -13.5483, -13.5811] };
  WCL[1488] = struct {
   x = [130.001, 129.968, 129.968],
   y = [-11.8515, -11.7436, -11.699] };
  WCL[1489] = struct {
   x = [124.826, 124.704, 124.572, 124.471, 124.295, 124.131, 123.965, 123.833, 123.666, 123.568, 123.699, 123.723],
   y = [-9.88722, -10.0515, -10.1055, -10.1712, -10.204, -10.3026, -10.3472, -10.3683, -10.3566, -10.258, -10.1055, -10.0093] };
  WCL[1490] = struct {
   x = [122.904, 122.915, 122.981, 123.082, 123.181, 123.279, 123.38, 123.413, 123.425, 123.425, 123.38, 123.258, 123.103, 122.981, 122.904],
   y = [-10.901, -10.7931, -10.772, -10.7602, -10.6734, -10.5866, -10.5091, -10.4997, -10.542, -10.6194, -10.718, -10.8142, -10.9245, -10.9339, -10.901] };
  WCL[1491] = struct {
   x = [120.796, 120.85, 120.74, 120.552, 120.322, 120.134],
   y = [-9.93181, -10.1172, -10.2158, -10.2909, -10.2815, -10.0632] };
  WCL[1492] = struct {
   x = [118.865, 119.008, 119.25],
   y = [-20.0772, -19.9528, -19.9411] };
  WCL[1493] = struct {
   x = [119.426, 119.646, 119.879],
   y = [-20.0349, -19.9411, -19.9411] };
  WCL[1494] = struct {
   x = [49.95, 50.0275, 50.0486, 50.1049, 50.227, 50.2598, 50.227, 50.1378],
   y = [-14.0528, -14.3321, -14.5903, -14.8789, -15.1465, -15.4234, -15.6909, -15.9045] };
  WCL[1495] = struct {
   x = [57.4928, 57.5585],
   y = [-20.0115, -19.981] };
  WCL[1496] = struct {
   x = [40.6777, 40.6331, 40.5838, 40.5322, 40.4665, 40.4008, 40.3374, 40.3585, 40.3585, 40.4125, 40.4242, 40.3139, 40.3468, 40.4031, 40.4031, 40.3796, 40.3914, 40.4125, 40.436, 40.4242, 40.3914, 40.4688,
        40.5345, 40.6002, 40.5017, 40.5791, 40.666, 40.666, 40.5791, 40.5134, 40.3468, 40.4899, 40.4806, 40.4454, 40.3022, 40.2928, 40.2811, 40.1168],
   y = [-10.9996, -11.1123, -11.1756, -11.2625, -11.4056, -11.5347, -11.7741, -12.0134, -12.2082, -12.5438, -12.802, -12.9006, -13.0719, -13.0719, -13.3746, -13.675, -13.8064, -13.966, -14.0411, -14.0951,
        -14.1913, -14.2453, -14.1913, -14.3204, -14.4495, -14.4612, -14.5152, -14.6748, -14.8461, -15.0385, -15.2098, -15.2521, -15.4023, -15.5314, -15.6065, -15.6065, -15.6158, -15.7965] };
  WCL[1497] = struct {
   x = [43.1137, 43.1137, 43.1254, 43.2029, 43.2592, 43.3015, 43.3249, 43.3789, 43.3789, 43.3249, 43.1818, 43.1583, 43.1372, 43.1137],
   y = [-11.3939, -11.3516, -11.3516, -11.361, -11.4479, -11.6004, -11.7623, -11.8492, -11.9477, -11.9266, -11.828, -11.6661, -11.4807, -11.3939] };
  WCL[1498] = struct {
   x = [44.3059, 44.3716, 44.3927, 44.4373, 44.5148, 44.8128, 44.8785, 44.9888, 45.0874, 45.1203, 45.1883, 45.2869, 45.3972, 45.4957, 45.5192, 45.6952, 45.7938, 45.9158, 46.0261, 46.113, 46.235, 46.3782,
        46.5002, 46.4556, 46.2679, 46.289, 46.4767, 46.6973, 46.8194, 46.8405, 46.8522, 46.9954, 47.1057, 47.0517, 46.9297, 47.0517, 47.2042, 47.162, 47.2934, 47.3474, 47.3052, 47.4248, 47.5915,
        47.7464, 47.8449, 47.8003, 47.6361, 47.5469, 47.5915, 47.7464, 47.8449, 47.8567, 47.8003, 47.7464, 47.8003, 47.9881, 48.0655, 48.1641, 48.2298, 48.2955, 48.3307, 48.4621, 48.6264, 48.7156,
        48.7367, 48.7367, 48.6264, 48.7156, 48.8916, 49.023, 49.2343, 49.3985, 49.5863, 49.5863, 49.72, 49.7858, 49.8069, 49.9054, 49.95],
   y = [-16.2659, -16.1697, -16.1603, -16.1697, -16.2025, -16.1814, -16.1063, -16.043, -15.989, -16.0852, -16.1181, -16.0007, -16.043, -15.9256, -15.8083, -15.766, -15.8294, -15.7238, -15.7543, -15.935,
        -16.0641, -16.0969, -16.0735, -15.9139, -15.7543, -15.5853, -15.3812, -15.2521, -15.3061, -15.4351, -15.5407, -15.5525, -15.4891, -15.3694, -15.1559, -15.0503, -14.8578, -14.8156, -14.6865,
        -14.8156, -15.0291, -14.8789, -14.6748, -14.6865, -14.6654, -14.5903, -14.5457, -14.4072, -14.2993, -14.0951, -14.203, -13.9778, -13.8487, -13.6328, -13.5577, -13.6234, -13.8158, -13.6656,
        -13.5694, -13.4826, -13.2878, -13.3863, -13.2784, -13.0179, -12.7691, -12.6072, -12.4453, -12.4781, -12.2622, -12.1754, -12.3584, -12.5978, -12.7809, -12.7691, -13.0085, -13.2549, -13.4614,
        -13.7618, -14.0528] };
  WCL[1499] = struct {
   x = [50.1378, 49.9946, 49.8843, 49.8069, 49.6426, 49.5745, 49.5745, 49.6966, 49.6966, 49.6191, 49.5088, 49.3657, 49.3211, 49.3211, 49.2671, 49.2108, 49.1451, 49.0465, 48.9479, 48.9127, 48.8916, 48.8259,
        48.7602, 48.6827],
   y = [-15.9045, -15.9256, -15.6909, -15.4351, -15.4445, -15.8505, -16.1275, -16.341, -16.6273, -16.8174, -16.9348, -17.1882, -17.4957, -17.7491, -18.0753, -18.371, -18.655, -18.9272, -19.1666, -19.1783,
        -19.3356, -19.5116, -19.7322, -19.9598] };
  WCL[1500] = struct {
   x = [44.3153, 44.4045, 44.3599, 44.4373, 44.3833, 44.2285, 44.1299, 44.0736, 44.0196, 43.9961, 43.9961, 43.8764, 43.8084, 43.8764, 44.0524, 44.2167, 44.3153, 44.3059, 44.3059],
   y = [-20.0232, -19.8355, -19.6594, -19.5656, -19.3966, -19.1455, -18.8122, -18.5705, -18.3288, -18.0542, -17.8007, -17.6341, -17.5262, -17.3994, -17.1366, -16.7869, -16.5522, -16.3926, -16.2659] };
  WCL[1501] = struct {
   x = [39.9291, 40.0065, 40.1262, 40.2271, 40.2928, 40.3914, 40.4899],
   y = [-9.99752, -10.0515, -10.1078, -10.2815, -10.2932, -10.3237, -10.3894] };
  WCL[1502] = struct {
   x = [40.4899, 40.5791, 40.6448, 40.6777, 40.6448, 40.6566, 40.6566, 40.6777],
   y = [-10.3894, -10.5209, -10.5537, -10.6523, -10.6945, -10.8142, -10.9456, -10.9996] };
  WCL[1503] = struct {
   x = [40.1168, 39.983, 39.9056, 39.6075, 39.6521, 39.7084, 39.631, 39.4433, 39.2227, 39.0584, 38.9481, 38.704, 38.5608, 38.3966, 38.1525, 37.8897, 37.5799, 37.3593, 37.0518, 36.8524, 36.8078, 36.7209,
        36.62, 36.566, 36.4886, 36.566, 36.6106, 36.5097, 36.39, 36.3008, 36.1788, 36.0802, 36.0263, 35.9582, 35.9042, 35.7493, 35.6625, 35.5616, 35.3973, 35.2213, 34.989, 34.8458, 34.6463],
   y = [-15.7965, -15.9561, -16.1181, -16.2776, -16.3082, -16.3621, -16.51, -16.6484, -16.7752, -16.8385, -17.031, -17.1577, -17.146, -17.1577, -17.2422, -17.3572, -17.4957, -17.6013, -17.8219, -17.9815,
        -18.1504, -18.2114, -18.2232, -18.2326, -18.2654, -18.317, -18.3593, -18.4766, -18.601, -18.7066, -18.8944, -18.8545, -18.6855, -18.8122, -18.9577, -19.0305, -19.1267, -19.2417, -19.3966,
        -19.575, -19.7205, -19.8871, -19.9387] };
  WCL[1504] = struct {
   x = [34.9444, 34.8998, 34.7684, 34.6792, 34.6135, 34.5689],
   y = [-11.6004, -11.5018, -11.4267, -11.3071, -11.2413, -11.0559] };
  WCL[1505] = struct {
   x = [34.3272, 34.3272, 34.1723, 34.1723, 34.1723, 34.13, 34.1394, 34.184, 34.2051, 34.2169, 34.1511, 34.0737, 34.0291, 34.008, 34.0197, 34.062, 34.1066, 34.1629, 34.1723, 34.1723, 34.2403, 34.3717,
        34.4703, 34.482, 34.482, 34.5266, 34.6252, 34.7144, 34.813, 34.9444, 35.0664, 35.2096, 35.1978, 35.165, 35.1204, 35.0875, 35.0007, 34.8787, 34.8458, 34.8247, 34.8247, 34.7895],
   y = [-10.9996, -11.0348, -11.0888, -11.1428, -11.1991, -11.2531, -11.3399, -11.469, -11.6333, -11.753, -11.882, -12.0134, -12.1308, -12.3045, -12.4453, -12.5978, -12.7691, -12.9217, -13.147, -13.3629,
        -13.5788, -13.5905, -13.6985, -13.8909, -14.0857, -14.2148, -14.2359, -14.257, -14.1913, -14.2148, -14.365, -14.4706, -14.4495, -14.3861, -14.2242, -14.0106, -13.8581, -13.7736, -13.6656,
        -13.6656, -13.6445, -13.6445] };
  WCL[1506] = struct {
   x = [34.9444, 34.8458, 34.8575, 34.8669, 34.8669, 34.8458, 34.7895, 34.7238, 34.6581, 34.6369, 34.6369, 34.6463, 34.6581, 34.6581, 34.6792, 34.7355, 34.7895, 34.8458],
   y = [-11.6004, -11.6004, -11.6661, -11.7741, -11.8492, -11.9689, -12.0885, -12.2294, -12.403, -12.5438, -12.7152, -12.8677, -13.0508, -13.2338, -13.3957, -13.5459, -13.6445, -13.6445] };
  WCL[1507] = struct {
   x = [34.5595, 34.6463, 34.7144, 34.644, 34.5689],
   y = [-9.99752, -10.2932, -10.6312, -10.7509, -10.9996] };
  WCL[1508] = struct {
   x = [34.0737, 34.1394, 34.1957, 34.2614, 34.3154, 34.3272, 34.3272, 34.3272],
   y = [-9.88722, -10.0515, -10.2486, -10.3801, -10.542, -10.6734, -10.7931, -10.9996] };
  WCL[1509] = struct {
   x = [11.6262, 11.6262, 11.6051, 11.5722, 11.5722, 11.6591, 11.6825, 11.6708, 11.6591, 11.6591, 11.6591, 11.6825, 11.8257, 11.8797, 11.9454, 12.0228, 12.0463, 12.1003, 12.1331, 12.1777, 12.2106, 12.288,
        12.3537, 12.3537, 12.4195, 12.4312, 12.4758, 12.5415, 12.7058, 12.8161, 12.8278, 12.8489, 13.0155, 13.215, 13.3347, 13.4568, 13.5436, 13.621, 13.6539, 13.6656, 13.6656, 13.6774, 13.6774,
        13.6774, 13.6774, 13.6985, 13.7219, 13.7219, 13.7102, 13.6985],
   y = [-17.3149, -17.2093, -17.0732, -16.8385, -16.6367, -16.6813, -16.7963, -16.6813, -16.4889, -16.3199, -16.1392, -15.9256, -15.7754, -15.6158, -15.3389, -15.1676, -14.9212, -14.7616, -14.5457, -14.4283,
        -14.2359, -14.0622, -13.912, -13.7736, -13.6867, -13.5694, -13.4497, -13.3089, -13.1798, -13.0625, -12.9428, -12.8231, -12.748, -12.6612, -12.5321, -12.3913, -12.2411, -12.044, -11.882,
        -11.7952, -11.7858, -11.7623, -11.6661, -11.5558, -11.4596, -11.3939, -11.3071, -11.2202, -11.1005, -11.0348] };
  WCL[1510] = struct {
   x = [12.8935, 12.8067, 12.7621, 12.6964, 12.6518, 12.6189, 12.5298, 12.464, 12.4312, 12.3749, 12.3092, 12.2317, 12.1214, 12.0463, 11.9689, 11.9571, 11.936, 11.8468, 11.7576, 11.6708, 11.6708, 11.6473,
        11.6145, 11.6051, 11.6145, 11.6145, 11.6262, 11.6262],
   y = [-20.0537, -19.9199, -19.7932, -19.6782, -19.6055, -19.4811, -19.3872, -19.2933, -19.1877, -19.1267, -19.0633, -18.8944, -18.7489, -18.6949, -18.6433, -18.6127, -18.601, -18.4766, -18.317, -18.1504,
        -17.9815, -17.9275, -17.8219, -17.7186, -17.5801, -17.4417, -17.3361, -17.3149] };
  WCL[1511] = struct {
   x = [13.6985, 13.7877, 13.7548, 13.7102, 13.4779, 13.391, 13.2901],
   y = [-11.0348, -10.901, -10.7814, -10.5537, -10.3683, -10.1501, -10.0186] };
  WCL[1512] = struct {
   x = [-50.0979, -49.9876, -49.8773, -49.7764, -49.6332, -49.5464, -49.4901, -49.4243, -49.3375, -49.2037, -49.0606, -48.8728, -48.84, -48.84, -48.7414, -48.664, -48.6194, -48.6076, -48.6076, -48.5748,
        -48.5983, -48.5748, -48.3988, -48.387, -48.5654, -48.6522, -48.6311, -48.664, -48.6968, -48.6968, -48.6757, -48.718, -48.7508, -48.6757, -48.6311, -48.6522, -48.5983, -48.4973, -48.387,
        -48.2556, -48.1664, -48.0139, -47.8801, -47.8379, -47.683, -47.5727, -47.4295, -47.2958, -47.1643, -47.054, -46.9226, -46.7349, -46.5354, -46.3922, -46.2819, -46.2162, -46.0402, -45.8313,
        -45.6107, -45.4887, -45.3573, -45.2564, -45.2235, -45.1578, -45.1695, -45.0921, -44.9043, -44.7049, -44.5946, -44.6274, -44.7283, -44.7495, -44.672, -44.5617, -44.442, -44.374, -44.3857,
        -44.32, -44.1534, -44.0102, -43.9445, -43.8225, -43.7028, -43.5596, -43.4164, -43.2498, -43.1395, -43.0738, -42.9095, -42.7218, -42.6326, -42.5223, -42.3909, -42.2031, -42.06, -41.9943,
        -42.0271, -42.0717, -42.06, -41.9708, -41.7502, -41.5531, -41.3653, -41.2104, -41.133, -41.1119, -41.1236, -41.1119, -41.079, -40.957, -40.8349, -40.7364, -40.6143, -40.4946, -40.3843,
        -40.2506],
   y = [-29.8588, -29.7039, -29.5302, -29.3777, -29.204, -29.136, -29.0867, -29.0186, -28.9036, -28.7769, -28.6901, -28.6525, -28.6525, -28.4953, -28.3685, -28.3498, -28.1245, -27.8616, -27.6363, -27.3805,
        -27.2726, -27.2233, -27.2327, -27.2233, -27.0754, -26.998, -26.7797, -26.7305, -26.6906, -26.6319, -26.5122, -26.3949, -26.2846, -26.1461, -25.9771, -25.8879, -25.7964, -25.6885, -25.5383,
        -25.4373, -25.3482, -25.2285, -25.1088, -25.0079, -24.8577, -24.7474, -24.6465, -24.5948, -24.5057, -24.393, -24.2733, -24.1419, -24.0316, -23.9495, -23.9893, -23.9401, -23.7781, -23.757,
        -23.7875, -23.7476, -23.5857, -23.5552, -23.4824, -23.4637, -23.412, -23.3721, -23.3111, -23.2806, -23.2689, -23.1985, -23.1468, -23.1468, -23.067, -23.006, -22.9028, -22.9333, -23.0154,
        -23.0365, -22.9849, -22.9333, -22.9028, -22.9333, -22.9849, -22.9943, -22.9638, -22.8723, -22.8629, -22.945, -22.9638, -22.9544, -22.945, -22.9145, -22.9145, -22.8816, -22.8511, -22.8019,
        -22.7197, -22.6892, -22.6282, -22.4733, -22.3301, -22.2292, -22.1565, -22.0743, -21.9711, -21.849, -21.673, -21.5087, -21.3327, -21.1473, -20.9713, -20.8681, -20.7765, -20.6193, -20.4433,
        -20.1851] };
  WCL[1513] = struct {
   x = [-50.1847, -50.1401, -50.0979],
   y = [-30.1169, -29.9644, -29.8588] };
  WCL[1514] = struct {
   x = [-70.2619, -70.229, -70.229, -70.2525, -70.2619, -70.2736, -70.2173, -70.1633, -70.1633, -70.2173, -70.229, -70.2948, -70.3182, -70.3182, -70.3393, -70.3957, -70.4731, -70.5599, -70.6257, -70.6257,
        -70.5811, -70.5154, -70.5271, -70.5717, -70.5717, -70.5717, -70.5717, -70.5717, -70.5482, -70.5388, -70.5388, -70.5388, -70.5482, -70.6702, -70.7477, -70.7923, -70.7923, -70.7805, -70.7477,
        -70.7266, -70.7266, -70.7594, -70.7688, -70.7923, -70.8697, -70.9448, -71.0011, -71.0129, -71.0129, -71.0129, -71.0011, -71.0551, -71.1443, -71.1889, -71.2218, -71.2546, -71.3086, -71.3532,
        -71.4752, -71.5409],
   y = [-19.8308, -20.0068, -20.2039, -20.3705, -20.631, -20.9197, -21.1168, -21.375, -21.5604, -21.7458, -21.9499, -22.1659, -22.3724, -22.6775, -22.8816, -23.006, -23.006, -23.1069, -23.3299, -23.4331,
        -23.4942, -23.5646, -23.7664, -23.9706, -24.1513, -24.3437, -24.5761, -24.7755, -24.9469, -25.0173, -25.0971, -25.2684, -25.4092, -25.4984, -25.6392, -25.7776, -25.8879, -25.9959, -26.1367,
        -26.2447, -26.3949, -26.5333, -26.6999, -26.8501, -27.0567, -27.1247, -27.2139, -27.3031, -27.4016, -27.4509, -27.5472, -27.6176, -27.7724, -27.9696, -28.2137, -28.3803, -28.5539, -28.7088,
        -28.8543, -28.9412] };
  WCL[1515] = struct {
   x = [-71.5409, -71.5972, -71.6066, -71.6066, -71.5644, -71.5198, -71.4869, -71.4424, -71.4424, -71.4306, -71.3978],
   y = [-28.9412, -29.0491, -29.1266, -29.204, -29.2721, -29.3988, -29.5044, -29.5912, -29.7062, -29.8025, -29.8893] };
  WCL[1516] = struct {
   x = [-109.499, -109.4, -109.377, -109.409, -109.454, -109.487, -109.531, -109.543],
   y = [-27.1693, -27.1599, -27.12, -27.0801, -27.0707, -27.052, -27.1294, -27.1693] };
  WCL[1517] = struct {
   x = [-124.946, -124.978, -124.922, -124.913, -124.955, -124.946],
   y = [-24.6418, -24.7122, -24.7122, -24.6512, -24.6723, -24.6418] };
  WCL[1518] = struct {
   x = [-135.685, -135.706, -135.76, -135.75, -135.649, -135.64, -135.685],
   y = [-21.4618, -21.4618, -21.5439, -21.5862, -21.5651, -21.4829, -21.4618] };
  WCL[1519] = struct {
   x = [-130.27, -130.282, -130.247, -130.238, -130.27],
   y = [-25.0431, -25.083, -25.0924, -25.0431, -25.0431] };
  WCL[1520] = struct {
   x = [-137.25, -137.283, -137.292, -137.292, -137.25, -137.238, -137.25],
   y = [-23.1116, -23.1022, -23.1633, -23.1844, -23.1938, -23.1539, -23.1116] };
  WCL[1521] = struct {
   x = [-140.798, -140.756, -140.789, -140.81, -140.756, -140.733, -140.744, -140.798],
   y = [-21.7505, -21.6589, -21.6683, -21.7716, -21.7622, -21.6777, -21.6683, -21.7505] };
  WCL[1522] = struct {
   x = [-175.219, -175.187, -175.187, -175.219, -175.287, -175.386, -175.496, -175.541, -175.475, -175.353, -175.219],
   y = [-21.1333, -21.1215, -21.2037, -21.2975, -21.2881, -21.2365, -21.1943, -21.1004, -21.1004, -21.1333, -21.1333] };
  WCL[1523] = struct {
   x = [165.011, 164.872, 164.642, 164.508, 164.41, 164.3, 164.213, 164.025, 164.025, 163.924, 164.034, 164.189, 164.278, 164.377, 164.52, 164.652, 164.675, 164.753, 164.851, 164.863, 164.999, 164.999,
        165.011, 165.022, 165.067, 165.109, 165.154, 165.187, 165.243, 165.255, 165.276, 165.297, 165.365, 165.419, 165.55, 165.619, 165.651, 165.717, 165.771, 165.926, 166.036, 166.158, 166.29, 166.4,
        166.489, 166.543, 166.653, 166.797, 166.841, 166.808, 166.731, 166.71, 166.698, 166.653, 166.621, 166.489, 166.445, 166.29, 166.191, 166.048, 166.003, 165.982, 165.893, 165.827, 165.738,
        165.705, 165.628, 165.583, 165.473, 165.342, 165.231, 165.133, 165.109, 165.044, 165.011],
   y = [-21.4313, -21.3703, -21.2482, -21.0816, -20.9478, -20.8657, -20.6592, -20.5137, -20.3682, -20.2532, -20.2227, -20.3048, -20.2837, -20.3471, -20.4503, -20.4926, -20.5864, -20.6592, -20.6686, -20.6686,
        -20.7179, -20.7179, -20.739, -20.7695, -20.8211, -20.8845, -20.9572, -20.9666, -21.0089, -21.0605, -21.1121, -21.1943, -21.2154, -21.2975, -21.3703, -21.443, -21.4829, -21.4947, -21.6073,
        -21.6167, -21.6589, -21.6895, -21.8115, -21.8936, -21.9781, -22.079, -22.0602, -22.1612, -22.3161, -22.356, -22.356, -22.3982, -22.4803, -22.4592, -22.3771, -22.3043, -22.255, -22.255,
        -22.1823, -22.1823, -22.1213, -22.0086, -22.0086, -21.957, -21.957, -21.8842, -21.8842, -21.8326, -21.8115, -21.6988, -21.6895, -21.6378, -21.5252, -21.4618, -21.4313] };
  WCL[1524] = struct {
   x = [166.996, 167.029, 167.106, 167.172, 167.172, 167.217, 167.304, 167.304, 167.282, 167.226, 167.128, 167.008, 166.984, 167.008, 166.996, 166.996],
   y = [-20.8023, -20.8117, -20.7906, -20.8211, -20.9267, -20.9877, -21.0183, -21.1215, -21.1731, -21.1943, -21.1333, -21.0605, -20.9784, -20.915, -20.8422, -20.8023] };
  WCL[1525] = struct {
   x = [167.712, 167.667, 167.712, 167.768, 167.81, 167.879, 167.944, 167.965, 167.977, 167.944, 167.81, 167.745, 167.712],
   y = [-21.4524, -21.4219, -21.4008, -21.4102, -21.3914, -21.443, -21.443, -21.4947, -21.5862, -21.6472, -21.6378, -21.5557, -21.4524] };
  WCL[1526] = struct {
   x = [149.923, 150.033, 150.089, 150.197, 150.397, 150.54, 150.561, 150.585, 150.662, 150.672, 150.716, 150.782, 150.782, 150.805, 150.871, 151.014, 151.124, 151.246, 151.39, 151.622, 151.753, 151.875,
        151.983, 152.019, 152.171, 152.359, 152.349, 152.448, 152.591, 152.558, 152.767, 152.824, 152.845, 152.856, 152.955, 153.009, 153.077, 153.098, 153.044, 153.077, 153.098, 153.119, 153.143,
        153.208, 153.131, 153.098, 153.131, 153.143, 153.208, 153.253, 153.298, 153.384, 153.429, 153.45, 153.506, 153.551, 153.551, 153.56, 153.56, 153.56, 153.518, 153.45, 153.363, 153.33],
   y = [-22.2034, -22.1213, -22.2855, -22.3677, -22.4592, -22.5742, -22.4592, -22.3161, -22.3794, -22.4921, -22.5414, -22.8089, -23.0623, -23.3064, -23.5411, -23.6514, -23.7852, -23.8861, -23.9964, -23.9964,
        -24.0879, -24.2287, -24.4118, -24.5127, -24.6418, -24.7028, -24.7521, -24.8929, -25.0642, -25.1651, -25.2449, -25.3857, -25.5242, -25.6838, -25.8433, -25.9231, -25.8246, -26.0733, -26.3221,
        -26.4207, -26.4887, -26.6882, -26.944, -27.0919, -27.052, -27.1717, -27.2303, -27.3383, -27.4063, -27.5941, -27.702, -27.8687, -27.986, -28.0728, -28.2113, -28.3263, -28.4436, -28.561,
        -28.5891, -28.7558, -29.0069, -29.1336, -29.3941, -29.7203] };
  WCL[1527] = struct {
   x = [153.098, 153.119, 153.176, 153.22, 153.274, 153.274, 153.176, 153.098, 153.065, 153.009, 152.976, 152.988, 152.976, 153.086, 153.154, 153.098],
   y = [-24.7943, -24.7333, -24.8037, -24.8835, -24.975, -25.2144, -25.4139, -25.5641, -25.6744, -25.6744, -25.4937, -25.3552, -25.2144, -25.0736, -24.9234, -24.7943] };
  WCL[1528] = struct {
   x = [153.45, 153.518, 153.539, 153.518, 153.474, 153.429, 153.417, 153.45, 153.417, 153.45],
   y = [-27.3477, -27.3383, -27.4063, -27.526, -27.6222, -27.5941, -27.4462, -27.3477, -27.4462, -27.3477] };
  WCL[1529] = struct {
   x = [148.149, 148.247, 148.358, 148.468, 148.555, 148.688, 148.775, 148.853, 148.897, 148.876, 148.766, 148.742, 148.853, 149.029, 149.073, 149.207, 149.24, 149.294, 149.327, 149.461, 149.482, 149.482,
        149.493, 149.538, 149.547, 149.547, 149.571, 149.648, 149.691, 149.747, 149.813, 149.923, 149.923],
   y = [-19.8894, -20.0349, -20.1382, -20.0654, -20.1382, -20.2109, -20.2649, -20.3353, -20.5325, -20.5231, -20.502, -20.7108, -20.8141, -20.8352, -20.9267, -21.0816, -21.2365, -21.3398, -21.4548, -21.4853,
        -21.5979, -21.8655, -22.0509, -22.133, -22.1424, -22.194, -22.2762, -22.3489, -22.4921, -22.5015, -22.4592, -22.4592, -22.2034] };
  WCL[1530] = struct {
   x = [113.913, 114.023, 114.023, 114.023, 114.023, 114.014, 114.002, 113.958, 113.946, 114.002, 113.99, 113.958, 114.047, 114.202, 114.277, 114.366, 114.455, 114.587, 114.73, 114.852, 114.983, 115.126,
        115.338, 115.424, 115.556, 115.69, 115.8, 115.92, 116.063, 116.173, 116.307, 116.495, 116.626, 116.692, 116.736, 116.891, 117.034, 117.189, 117.377, 117.53, 117.684, 117.872, 118.06, 118.236,
        118.435, 118.611, 118.865],
   y = [-21.8655, -21.7739, -21.7927, -21.8561, -21.9265, -22.0391, -22.0908, -22.1541, -22.2339, -22.2762, -22.3982, -22.4921, -22.5226, -22.3982, -22.1729, -22.0509, -21.896, -21.8443, -21.7622, -21.7317,
        -21.619, -21.5885, -21.558, -21.4125, -21.2881, -21.2177, -21.0417, -20.9784, -20.9267, -20.8446, -20.7836, -20.6475, -20.5747, -20.5536, -20.5958, -20.6169, -20.6169, -20.6686, -20.6897,
        -20.6592, -20.6169, -20.4621, -20.3682, -20.3259, -20.3166, -20.2837, -20.0772] };
  WCL[1531] = struct {
   x = [119.25, 119.426],
   y = [-19.9411, -20.0349] };
  WCL[1532] = struct {
   x = [114.962, 114.929, 114.929, 114.929, 114.929, 114.861, 114.807, 114.718, 114.565, 114.521, 114.345, 114.124, 114.101, 113.981, 113.871, 113.66, 113.495, 113.373, 113.329, 113.218, 113.186, 113.143,
        113.164, 113.197, 113.286, 113.352, 113.406, 113.43, 113.507, 113.549, 113.683, 113.782, 113.76, 113.594, 113.484, 113.385, 113.373, 113.451, 113.54, 113.594, 113.627, 113.782, 113.826,
        113.847, 113.981, 114.08, 114.124, 114.091, 113.958, 113.847, 113.803, 113.826, 113.76, 113.695, 113.528, 113.451, 113.296, 113.308, 113.296, 113.385, 113.528, 113.638, 113.671, 113.65,
        113.671, 113.671, 113.617, 113.54, 113.584, 113.627, 113.728, 113.782, 113.838, 113.913],
   y = [-30.1052, -29.9315, -29.7696, -29.6147, -29.4317, -29.2392, -29.143, -29.0069, -28.7933, -28.5704, -28.3474, -27.9954, -27.6222, -27.3782, -27.052, -26.9346, -26.7563, -26.6178, -26.47, -26.3902,
        -26.301, -26.2329, -26.1414, -26.2212, -26.2329, -26.2611, -26.3808, -26.5404, -26.5005, -26.599, -26.6178, -26.5286, -26.3808, -26.2329, -25.9842, -25.7542, -25.5852, -25.6439, -25.8246,
        -26.0241, -26.2118, -26.0522, -26.193, -26.3808, -26.3995, -26.2822, -26.0733, -25.7753, -25.6556, -25.4538, -25.5054, -25.3857, -25.2942, -25.1839, -24.853, -24.7239, -24.4517, -24.1982,
        -23.9471, -23.7852, -23.6514, -23.5505, -23.3886, -23.2454, -23.0318, -22.9098, -22.8183, -22.6962, -22.5226, -22.3161, -22.1729, -22.0509, -21.9171, -21.8655] };
  WCL[1533] = struct {
   x = [55.1225, 55.045, 55.1553, 55.275, 55.4299, 55.5942, 55.7045, 55.7045, 55.6834, 55.5073, 55.3407, 55.1882, 55.1225],
   y = [-20.9877, -20.9666, -20.9455, -20.915, -20.8939, -20.9784, -21.1426, -21.2881, -21.3914, -21.4102, -21.3187, -21.152, -20.9877] };
  WCL[1534] = struct {
   x = [57.3919, 57.4153, 57.4928],
   y = [-20.0748, -20.0443, -20.0115] };
  WCL[1535] = struct {
   x = [57.5585, 57.5679, 57.5679, 57.6359, 57.6125, 57.4928, 57.2933, 57.2604, 57.3261, 57.3919],
   y = [-19.981, -20.0654, -20.1265, -20.2837, -20.4175, -20.4902, -20.502, -20.3353, -20.1593, -20.0748] };
  WCL[1536] = struct {
   x = [48.6827, 48.5936, 48.5161, 48.3847, 48.2955, 48.2087, 48.0984, 47.967, 47.8449, 47.7675, 47.7112, 47.6572, 47.568, 47.514, 47.4366, 47.2934, 47.1949, 47.0611, 46.9743, 46.7982, 46.6316, 46.4767,
        46.1904, 45.8924, 45.5943, 45.308, 45.0545, 44.8574, 44.7142, 44.4491, 44.205, 43.9961, 43.853, 43.7309, 43.623, 43.5666, 43.5666, 43.6441, 43.5549, 43.4563, 43.2921, 43.17, 43.17, 43.1372,
        43.1372, 43.1912, 43.2686, 43.3343, 43.4024, 43.5221, 43.6652, 43.7215, 43.8318, 43.9961, 44.151, 44.3153],
   y = [-19.9598, -20.2508, -20.4292, -20.739, -21.1426, -21.4219, -21.7927, -22.0908, -22.4897, -22.7549, -22.9591, -23.2454, -23.52, -23.7922, -24.0762, -24.339, -24.5808, -24.792, -24.9328, -25.083,
        -25.1628, -25.1323, -25.1933, -25.2519, -25.4233, -25.5735, -25.6134, -25.4632, -25.374, -25.2121, -25.0924, -24.9328, -24.7309, -24.5714, -24.278, -23.9659, -23.7312, -23.5693, -23.2947,
        -23.06, -22.9286, -22.6634, -22.417, -22.2034, -21.8936, -21.8021, -21.6988, -21.5346, -21.4219, -21.2764, -21.2248, -21.0488, -20.7601, -20.4808, -20.2109, -20.0232] };
  WCL[1537] = struct {
   x = [32.8158, 32.7947, 32.7501, 32.6398, 32.6304, 32.5858, 32.4872, 32.4309, 32.3863, 32.222, 32.0132, 31.8583, 31.6917, 31.5391, 31.3842, 31.2622, 31.2293, 31.173, 31.1073, 31.0087, 30.9548],
   y = [-26.7351, -26.944, -27.12, -27.3265, -27.5425, -27.7584, -27.9649, -28.209, -28.453, -28.6666, -28.8121, -28.9271, -28.9764, -29.0726, -29.2463, -29.4012, -29.4598, -29.5255, -29.6898, -29.8048,
        -29.9292] };
  WCL[1538] = struct {
   x = [34.6463, 34.6369, 34.6135, 34.5924, 34.6252, 34.7355, 34.8669, 34.9444, 34.9561, 34.989, 35.0101, 35.0547, 35.1556, 35.2307, 35.2542, 35.287, 35.3856, 35.463, 35.463, 35.4513, 35.4302, 35.4184,
        35.3973, 35.3199, 35.2542, 35.165, 35.1978, 35.2659, 35.3973, 35.3973, 35.341, 35.2424, 35.1321, 34.9772, 34.7144, 34.3835, 34.1629, 33.9094, 33.5785, 33.3462, 33.2148, 33.027, 32.8604,
        32.7736, 32.6398, 32.4755, 32.4426, 32.5858, 32.7055, 32.8158, 32.8721, 32.851, 32.8393, 32.8393, 32.8275, 32.8158],
   y = [-19.9387, -20.0866, -20.2508, -20.4503, -20.6146, -20.6967, -20.7906, -20.8939, -21.0394, -21.0699, -21.2248, -21.3492, -21.6073, -21.8115, -22.14, -22.2222, -22.14, -22.2034, -22.3865, -22.5812,
        -22.776, -22.9802, -23.2149, -23.4378, -23.7124, -23.9964, -24.1771, -24.0246, -23.8744, -23.9354, -24.156, -24.3484, -24.4798, -24.6512, -24.8319, -24.9516, -24.9727, -25.0619, -25.1933,
        -25.2942, -25.3928, -25.4843, -25.5242, -25.604, -25.8222, -26.0123, -26.0123, -26.132, -26.2892, -26.2095, -26.1109, -26.2799, -26.4488, -26.5967, -26.7351, -26.7351] };
  WCL[1539] = struct {
   x = [16.4325, 16.3786, 16.3011, 16.1791, 16.0477, 15.9585, 15.881, 15.7825, 15.6722, 15.5853, 15.5173, 15.4422, 15.3741, 15.2638, 15.1652, 15.1441, 15.1113, 15.0432, 15.0221, 15.001, 14.9329, 14.9001,
        14.8907, 14.9235, 14.8672, 14.8015, 14.7569, 14.7475, 14.7358, 14.7569, 14.7686, 14.7475, 14.6912, 14.5809, 14.5269, 14.4917, 14.4049, 14.372, 14.3814, 14.3814, 14.372, 14.372, 14.372, 14.3603,
        14.372, 14.3931, 14.3814, 14.3485, 14.2946, 14.196, 14.074, 13.9519, 13.8416, 13.7642, 13.7548, 13.621, 13.5107, 13.4333, 13.3582, 13.2901, 13.2362, 13.1916, 13.1258, 13.093, 12.9827, 12.8935],
   y = [-28.6666, -28.6361, -28.6361, -28.5305, -28.3944, -28.3263, -28.2864, -28.209, -28.101, -27.9461, -27.8076, -27.7208, -27.6316, -27.5143, -27.4063, -27.2585, -27.0426, -26.8642, -26.707, -26.538,
        -26.3784, -26.3291, -26.3198, -26.1508, -26.0311, -25.9114, -25.7518, -25.5641, -25.374, -25.2426, -25.1323, -25.0126, -24.9023, -24.7122, -24.5901, -24.4798, -24.3179, -24.1372, -23.9541,
        -23.8345, -23.6913, -23.52, -23.3674, -23.2032, -23.1022, -22.9075, -22.7549, -22.5601, -22.4381, -22.2433, -22.1001, -21.957, -21.8326, -21.6777, -21.5768, -21.4219, -21.2553, -21.1121,
        -20.9971, -20.8634, -20.7507, -20.6052, -20.502, -20.3564, -20.1992, -20.0537] };
  WCL[1540] = struct {
   x = [17.0943, 17.0404, 16.9512, 16.8409, 16.7869, 16.686, 16.5217, 16.4654, 16.4771, 16.4325],
   y = [-30.1029, -29.9386, -29.671, -29.441, -29.3143, -29.0726, -28.9177, -28.8403, -28.7346, -28.6666] };
  WCL[1541] = struct {
   x = [-53.2403, -53.1183, -52.9516, -52.8202, -52.7099, -52.6231, -52.5339, -52.4776, -52.4682, -52.4447, -52.4119, -52.3579, -52.203, -51.9707, -51.7829, -51.5858, -51.4426, -51.3089, -51.1986, -51.1235,
        -50.9897, -50.8583, -50.7597, -50.6142, -50.5039, -50.4171, -50.3185, -50.2833, -50.2504, -50.1847],
   y = [-33.6114, -33.5386, -33.426, -33.2875, -33.1678, -32.9824, -32.8815, -32.7595, -32.7031, -32.6398, -32.4802, -32.3863, -32.2174, -32.0789, -31.9381, -31.8442, -31.7691, -31.6635, -31.5884, -31.4945,
        -31.3349, -31.1824, -31.0322, -30.8703, -30.6989, -30.5558, -30.4713, -30.3751, -30.2695, -30.1169] };
  WCL[1542] = struct {
   x = [-60.0203, -59.8537, -59.6659, -59.4359, -59.1942, -58.983, -58.7953, -58.7084, -58.685, -58.5418, -58.3118, -58.0677, -57.826, -57.6265, -57.5843, -57.5162, -57.3965, -57.1759, -57.0211, -56.8333,
        -56.7113, -56.7113, -56.7113, -56.7441, -56.9225, -57.1642, -57.3731, -57.4411, -57.352, -57.2417, -57.2417, -57.2534, -57.3731, -57.5397, -57.7697, -57.9574, -58.2015, -58.4104, -58.5089,
        -58.5535, -58.4972, -58.4644, -58.4761, -58.4761],
   y = [-38.8284, -38.8096, -38.8096, -38.7768, -38.7251, -38.6641, -38.6031, -38.5608, -38.5608, -38.5608, -38.4998, -38.4294, -38.3356, -38.1877, -38.0305, -37.838, -37.6808, -37.4883, -37.2959, -37.0941,
        -36.8899, -36.7045, -36.4722, -36.3314, -36.3478, -36.2516, -36.045, -35.8573, -35.7141, -35.4888, -35.4443, -35.3363, -35.2377, -35.0922, -34.9279, -34.82, -34.7027, -34.6111, -34.4187,
        -34.3107, -34.2732, -34.1465, -34.008, -33.9986] };
  WCL[1543] = struct {
   x = [-58.4761, -58.455, -58.3447, -58.2672, -58.1452, -58.0232, -57.9574, -57.8378, -57.6829, -57.4623, -57.2745, -57.0985, -56.8216, -56.5798, -56.3921, -56.2372, -56.1269, -55.9744, -55.7749, -55.5332,
        -55.3454, -55.1459, -54.9042, -54.6062, -54.3856, -54.1767, -54.0101, -53.935, -53.8786, -53.7355, -53.6134, -53.5689, -53.536, -53.4703, -53.4163, -53.36, -53.3506, -53.3271],
   y = [-33.9986, -33.9892, -34.0174, -34.0995, -34.2192, -34.3201, -34.4375, -34.4562, -34.4656, -34.4844, -34.4844, -34.5642, -34.6745, -34.7566, -34.7918, -34.8646, -34.874, -34.8575, -34.82, -34.82,
        -34.8575, -34.8834, -34.9467, -34.8575, -34.766, -34.6369, -34.4844, -34.4375, -34.3201, -34.191, -34.008, -33.9165, -33.8249, -33.7686, -33.7428, -33.7052, -33.6865, -33.6137] };
  WCL[1544] = struct {
   x = [-51.2666, -51.2549, -51.2431, -51.2431, -51.1986, -51.1328, -51.1235, -51.1235, -51.0343, -50.9568, -50.8465, -50.7034, -50.6142, -50.5485, -50.5485, -50.5931, -50.6588, -50.6823, -50.6823, -50.748,
        -50.7362, -50.7362, -50.7362, -50.8465, -50.924, -50.9897, -51.0225, -51.0789, -51.1774, -51.1986, -51.2314, -51.2549, -51.2666, -51.3652, -51.452, -51.5858, -51.7078, -51.8158, -51.9261,
        -52.027, -52.1044, -52.1138, -52.0927, -52.1044, -52.1819, -52.2241, -52.2687, -52.2804, -52.2687, -52.1913, -52.1701, -52.1256, -52.1044, -52.0927, -52.0481, -51.9261, -51.7947, -51.7289,
        -51.7289, -51.6186, -51.5623, -51.5201, -51.4755, -51.4426, -51.4192, -51.3863, -51.3652, -51.3206, -51.2995, -51.2995, -51.2995, -51.2995, -51.2666],
   y = [-30.0043, -30.0606, -30.1193, -30.1944, -30.2624, -30.3282, -30.3774, -30.4244, -30.4431, -30.3868, -30.368, -30.3282, -30.2812, -30.3375, -30.4619, -30.5206, -30.53, -30.4901, -30.4619, -30.4619,
        -30.5581, -30.652, -30.7764, -30.9008, -30.9571, -30.9759, -31.004, -31.0909, -31.1566, -31.2692, -31.3725, -31.4499, -31.5344, -31.572, -31.6283, -31.7128, -31.7973, -31.8536, -31.8911,
        -31.8724, -31.8724, -31.9475, -32.0507, -32.0507, -31.985, -31.9193, -31.863, -31.7879, -31.7597, -31.7316, -31.6283, -31.6283, -31.5438, -31.4312, -31.3631, -31.3537, -31.3443, -31.2692,
        -31.2411, -31.1941, -31.051, -30.9383, -30.8139, -30.7295, -30.7576, -30.7201, -30.6145, -30.53, -30.4525, -30.3868, -30.2812, -30.1381, -30.0043] };
  WCL[1545] = struct {
   x = [-62.3695, -62.2803, -62.2475, -62.1489, -62.1043, -62.1583, -62.3132, -62.346, -62.3578, -62.3132, -62.2146, -62.0714, -61.8062, -61.5528, -61.2993, -61.0553, -60.8347, -60.6586, -60.4287, -60.2738,
        -60.0203],
   y = [-40.0394, -39.8446, -39.8704, -39.7343, -39.4785, -39.3682, -39.2485, -38.9997, -38.8964, -38.758, -38.8354, -38.9316, -38.9575, -38.9903, -38.9997, -38.9903, -38.9575, -38.9481, -38.9223, -38.88,
        -38.8284] };
  WCL[1546] = struct {
   x = [-71.3978, -71.4541, -71.4869, -71.5855, -71.7075, -71.7287, -71.7498, -71.7498, -71.7287, -71.7169, -71.6841, -71.663, -71.6301, -71.6184, -71.6184, -71.5855, -71.5292, -71.5292, -71.5292, -71.5972,
        -71.7169, -71.7287, -71.7169, -71.7169, -71.8507, -71.9375, -71.9821, -72.0056, -72.015, -72.015, -72.0713, -72.137, -72.2145, -72.2356, -72.3459, -72.4351, -72.5336, -72.6111, -72.6439,
        -72.6768, -72.7754, -72.8199, -72.8528, -72.9185, -72.9748, -73.0405, -73.1063, -73.2166, -73.2166, -73.1954, -73.1954, -73.1954, -73.3386, -73.4935, -73.6366, -73.6366, -73.6366, -73.6578,
        -73.6578, -73.6249, -73.5592, -73.5146, -73.5592, -73.5803, -73.5921, -73.5592, -73.5029, -73.4489, -73.416, -73.3715, -73.3503, -73.3386, -73.3386, -73.4043, -73.4935, -73.5263, -73.6366,
        -73.7024],
   y = [-29.8893, -30.0418, -30.185, -30.2343, -30.3, -30.4995, -30.7013, -30.8515, -31.0416, -31.2035, -31.3725, -31.5626, -31.7503, -31.9475, -32.1634, -32.2948, -32.4825, -32.6492, -32.8721, -33.0317,
        -33.2054, -33.3908, -33.6231, -33.7592, -33.8883, -34.1465, -34.3295, -34.3459, -34.4375, -34.5384, -34.6463, -34.7918, -34.9913, -35.118, -35.2635, -35.3903, -35.5428, -35.7141, -35.8127,
        -35.9465, -36.0638, -36.2422, -36.3853, -36.5261, -36.6599, -36.6951, -36.6411, -36.6857, -36.8007, -36.9345, -37.0659, -37.1809, -37.2349, -37.1809, -37.3147, -37.4813, -37.594, -37.7348,
        -37.831, -37.9178, -38.0305, -38.1877, -38.5608, -38.5608, -38.5867, -38.6547, -38.7509, -38.8284, -38.9481, -39.103, -39.2579, -39.3846, -39.523, -39.6756, -39.7601, -39.8446, -39.8868,
        -39.9126] };
  WCL[1547] = struct {
   x = [172.816, 172.84, 172.84, 172.849, 172.905, 172.905, 172.917, 172.971, 173.016, 173.016, 173.027, 173.114, 173.203, 173.29, 173.314, 173.302, 173.412, 173.511, 173.612, 173.732, 173.832, 173.952,
        174.008, 173.976, 173.875, 174.008, 174.086, 174.119, 174.196, 174.206, 174.262, 174.339, 174.414, 174.45, 174.45, 174.492, 174.45, 174.295, 174.206, 174.229, 174.262, 174.304, 174.372,
        174.414, 174.536, 174.659, 174.757, 174.781, 174.68, 174.647, 174.67, 174.724, 174.724, 174.659, 174.593, 174.635, 174.691, 174.757, 174.781, 174.846, 174.9, 174.957, 175.076, 175.154, 175.177,
        175.219, 175.33, 175.407, 175.407, 175.398, 175.386, 175.32, 175.276, 175.353, 175.407, 175.541, 175.661, 175.661, 175.661, 175.705, 175.717, 175.75, 175.75, 175.738, 175.75, 175.762, 175.827,
        175.872, 175.881, 175.905, 176.036, 176.203, 176.4, 176.632, 176.897, 177.041, 177.139, 177.217, 177.303, 177.381, 177.458, 177.524, 177.601, 177.691, 177.756, 177.834, 177.888, 177.989,
        178.141, 178.263, 178.385, 178.385, 178.362, 178.329, 178.296, 178.275, 178.275, 178.275, 178.252, 178.263, 178.219, 178.054, 177.911, 177.888, 177.878, 177.846, 177.855, 177.911, 177.911,
        177.81, 177.789, 177.768, 177.679, 177.557, 177.425, 177.282, 177.172, 177.083, 176.951, 176.874, 176.775, 176.787, 176.829, 176.886, 176.972, 176.951, 176.862],
   y = [-34.468, -34.4492, -34.4586, -34.4586, -34.468, -34.5149, -34.6229, -34.7425, -34.7965, -34.8505, -34.8857, -34.9866, -34.9514, -34.8693, -34.9045, -34.9772, -34.9772, -34.9678, -35.0406, -35.0594,
        -35.05, -35.1321, -35.1767, -35.1955, -35.2307, -35.3034, -35.348, -35.2847, -35.3386, -35.4208, -35.51, -35.5827, -35.6367, -35.6907, -35.7705, -35.8878, -35.8526, -35.7892, -35.7892,
        -35.8972, -35.9136, -35.9066, -35.9512, -36.0662, -36.1647, -36.3337, -36.4229, -36.4581, -36.5473, -36.6552, -36.681, -36.7068, -36.8125, -36.8218, -36.8758, -36.911, -36.911, -36.9462,
        -36.972, -36.972, -36.972, -36.9462, -36.9368, -36.965, -37.087, -37.202, -37.2372, -37.1574, -36.965, -36.7773, -36.6975, -36.6177, -36.566, -36.5567, -36.6458, -36.681, -36.6975, -36.7866,
        -36.85, -36.8758, -37.0002, -37.08, -37.1574, -37.2114, -37.256, -37.3076, -37.3687, -37.4578, -37.5189, -37.6151, -37.6949, -37.6949, -37.8169, -37.8779, -37.9295, -38, -38.0164, -38.0164,
        -37.9295, -37.9295, -37.8333, -37.7723, -37.7371, -37.6667, -37.6409, -37.6057, -37.5634, -37.5353, -37.5634, -37.6151, -37.6855, -37.7559, -37.8779, -37.9648, -38.0352, -38.0774, -38.19,
        -38.3121, -38.3567, -38.4505, -38.582, -38.6665, -38.711, -38.7955, -38.9082, -38.9598, -39.0795, -39.1053, -39.1569, -39.2602, -39.1663, -39.0889, -39.0631, -39.0631, -39.0889, -39.0889,
        -39.1241, -39.1921, -39.3283, -39.3963, -39.4996, -39.5583, -39.5418, -39.6615, -39.7131, -39.8305, -39.9924] };
  WCL[1548] = struct {
   x = [175.001, 174.856, 174.635, 174.426, 174.271, 174.107, 173.898, 173.722, 173.743, 173.919, 174.095, 174.339, 174.492, 174.56, 174.569, 174.593, 174.647, 174.769, 174.802, 174.802, 174.802, 174.769,
        174.856, 174.813, 174.757, 174.79, 174.724, 174.56, 174.525, 174.635, 174.691, 174.712, 174.703, 174.614, 174.426, 174.339, 174.229, 174.161, 174.14, 174.161, 174.196, 174.25, 174.316, 174.339,
        174.262, 174.196, 174.238, 174.229, 174.238, 174.161, 174.119, 173.997, 173.976, 174.008, 174.008, 173.764, 173.612, 173.445, 173.347, 173.335, 173.412, 173.478, 173.424, 173.391, 173.269,
        173.138, 173.06, 173.07, 173.081, 173.07, 173.037, 172.95, 172.873, 172.795, 172.685, 172.652, 172.739, 172.816],
   y = [-40.0347, -39.915, -39.8915, -39.8047, -39.6944, -39.577, -39.5512, -39.3893, -39.218, -39.0795, -39.0021, -38.9434, -38.8143, -38.6078, -38.4412, -38.2957, -38.1384, -38.1478, -38.1032, -38.0516,
        -37.9648, -37.8427, -37.7723, -37.7207, -37.5282, -37.3616, -37.3874, -37.256, -37.1668, -37.1504, -37.0964, -37.0424, -36.9908, -36.972, -37.0072, -36.857, -36.6975, -36.5473, -36.5027,
        -36.5473, -36.6012, -36.6623, -36.6106, -36.4323, -36.4323, -36.3971, -36.3525, -36.2727, -36.1905, -36.1553, -36.2093, -36.1905, -36.2727, -36.3525, -36.3877, -36.0662, -35.9066, -35.7353,
        -35.5921, -35.5381, -35.456, -35.348, -35.3128, -35.3832, -35.4748, -35.348, -35.2025, -35.1392, -35.1133, -35.0688, -34.9678, -34.8599, -34.759, -34.6604, -34.5595, -34.4961, -34.4774,
        -34.468] };
  WCL[1549] = struct {
   x = [153.33, 153.265, 153.23, 153.197, 153.119, 153.054, 153.032, 153.032, 153.021, 153.054, 153.065, 153.009, 153],
   y = [-29.7203, -30.0465, -30.2178, -30.2577, -30.4291, -30.5628, -30.7623, -30.828, -30.8491, -30.9993, -31.1425, -31.2458, -31.3396] };
  WCL[1550] = struct {
   x = [153, 152.8, 152.788, 152.734, 152.645, 152.568, 152.457, 152.457, 152.493, 152.415, 152.293, 152.239, 152.105, 152.084, 152.171, 151.974, 151.852, 151.798, 151.688, 151.577, 151.444, 151.324,
        151.268, 151.246, 151.366, 151.357, 151.256, 151.214, 151.19, 151.035, 150.97, 150.892, 150.85, 150.794, 150.782, 150.74, 150.639, 150.585, 150.507, 150.409, 150.364, 150.298, 150.242, 150.221,
        150.176, 150.155, 150.209, 150.165, 150.143, 150.078, 150.021],
   y = [-31.3396, -31.7644, -31.8114, -31.9428, -32.0366, -32.1587, -32.2619, -32.3652, -32.4591, -32.5506, -32.6633, -32.7102, -32.7196, -32.7665, -32.8134, -32.858, -32.9144, -33.0434, -33.1749, -33.4049,
        -33.5621, -33.5809, -33.609, -33.6724, -33.7099, -33.9024, -34.076, -34.0854, -34.2239, -34.3882, -34.5783, -34.8153, -34.942, -35.0781, -35.2401, -35.2049, -35.1415, -35.2589, -35.4302,
        -35.5287, -35.6836, -35.7094, -35.808, -36.0239, -36.1671, -36.2821, -36.3971, -36.4604, -36.6388, -36.735, -36.85] };
  WCL[1551] = struct {
   x = [149.989, 150.012, 150.045, 150.054, 150.021],
   y = [-37.08, -37.141, -37.2208, -37.3804, -37.5024] };
  WCL[1552] = struct {
   x = [150.021, 150.021],
   y = [-37.5024, -37.5024] };
  WCL[1553] = struct {
   x = [150.021, 149.979, 149.989],
   y = [-36.85, -36.9908, -37.08] };
  WCL[1554] = struct {
   x = [150.021, 149.979, 149.834, 149.768, 149.735, 149.538, 149.371, 149.097, 148.843, 148.566, 148.379, 148.158, 147.949, 147.74, 147.574, 147.454, 147.332, 147.266, 147.201, 147.156, 146.935, 146.837,
        146.769, 146.682, 146.515, 146.384, 146.306, 146.295, 146.351, 146.428, 146.539, 146.626, 146.548, 146.527, 146.45, 146.306, 146.196, 146.119, 145.964, 145.877, 145.799, 145.677, 145.546,
        145.523, 145.579, 145.457, 145.325, 145.236, 145.06, 144.863, 144.786, 145.027, 145.072, 145.051, 144.884, 144.774, 144.586, 144.41, 144.488, 144.664, 144.675, 144.509, 144.312, 144.124,
        143.948, 143.814, 143.683, 143.626, 143.385, 143.307, 142.943, 142.612, 142.448, 142.183, 141.951, 141.808, 141.655, 141.533, 141.5, 141.488, 141.357, 141.026, 140.683, 140.418, 140.254,
        140.144],
   y = [-37.5024, -37.5986, -37.5376, -37.547, -37.7395, -37.7817, -37.8169, -37.8263, -37.8263, -37.8263, -37.8357, -37.8967, -37.9741, -38.061, -38.183, -38.2534, -38.3215, -38.3825, -38.4271, -38.5045,
        -38.6172, -38.704, -38.643, -38.6594, -38.704, -38.6946, -38.711, -38.8401, -38.9105, -38.8237, -38.7556, -38.7556, -38.9105, -39.056, -39.1335, -38.9786, -38.8753, -38.8753, -38.8237,
        -38.6852, -38.6782, -38.6852, -38.5726, -38.4881, -38.3309, -38.2346, -38.3144, -38.3754, -38.4271, -38.3661, -38.2886, -38.2886, -38.1314, -37.9389, -37.8709, -37.9225, -38.0445, -38.061,
        -38.0962, -38.183, -38.2276, -38.244, -38.3215, -38.4435, -38.65, -38.704, -38.7556, -38.7721, -38.8237, -38.7462, -38.6172, -38.4435, -38.3919, -38.3661, -38.305, -38.2628, -38.2886, -38.3661,
        -38.3754, -38.3661, -38.2018, -38.0539, -38.0281, -37.9037, -37.7488, -37.5541] };
  WCL[1555] = struct {
   x = [147.905, 147.97, 148.048, 148.137, 148.28],
   y = [-39.7976, -39.7483, -39.7906, -39.8657, -39.9596] };
  WCL[1556] = struct {
   x = [148.104, 147.994, 147.905, 147.86, 147.905],
   y = [-40.1215, -39.9596, -39.9079, -39.8657, -39.7976] };
  WCL[1557] = struct {
   x = [140.144, 139.956, 139.813, 139.724, 139.714, 139.801, 139.813, 139.724, 139.681, 139.637, 139.592, 139.515, 139.393, 139.261, 139.217, 139.261, 139.339, 139.318, 139.184, 139.052, 138.942, 138.886,
        138.775, 138.599, 138.48, 138.402, 138.292, 138.158, 138.292, 138.435, 138.501, 138.489, 138.379, 138.224, 138.092, 137.982, 137.872, 137.86, 137.794, 137.762, 137.729, 137.574, 137.377,
        137.156, 136.98, 136.858, 136.903, 136.945, 137.198, 137.398, 137.454, 137.475, 137.442, 137.431, 137.454, 137.454, 137.496, 137.564, 137.675, 137.785, 137.884, 137.884, 137.884, 137.895,
        137.994, 137.97, 137.928, 137.916, 137.916, 137.839, 137.839, 137.762, 137.708, 137.717, 137.717, 137.618, 137.487, 137.398, 137.276, 137.255, 137.165, 136.968, 136.891, 136.87, 136.616,
        136.515, 136.396, 136.285, 136.142, 136.098, 135.922, 135.865, 135.854, 135.955, 135.955, 135.854, 135.722, 135.556, 135.37, 135.269, 135.117, 135.105, 135.182, 135.281, 135.391, 135.424,
        135.37, 135.269, 135.227, 135.093, 134.929, 134.851, 134.774, 134.697, 134.598, 134.389, 134.223, 134.091, 134.091, 134.178, 134.223, 134.002, 133.838, 133.441, 133.34, 133.373, 133.373,
        133.209, 133.021, 132.624, 132.514, 132.425, 132.359, 132.315],
   y = [-37.5541, -37.4508, -37.195, -37.0096, -36.9298, -36.7796, -36.5848, -36.3971, -36.2727, -36.1671, -36.1131, -36.0333, -35.9066, -35.7446, -35.7001, -35.6109, -35.4583, -35.4208, -35.4302, -35.4396,
        -35.4748, -35.5123, -35.5569, -35.6203, -35.6555, -35.6555, -35.6742, -35.6297, -35.4748, -35.2401, -35.0054, -34.7613, -34.4961, -34.3154, -34.2239, -34.414, -34.5689, -34.7613, -34.942,
        -35.1227, -35.1697, -35.1321, -35.2143, -35.2143, -35.3128, -35.2143, -35.0875, -34.9326, -34.9514, -34.8622, -34.6416, -34.4703, -34.4422, -34.3882, -34.2966, -34.1582, -34.0854, -33.9751,
        -33.7381, -33.6818, -33.5245, -33.3955, -33.2195, -33.1467, -33.1655, -33.0528, -32.9519, -32.9425, -32.7947, -32.7571, -32.5882, -32.5154, -32.6445, -32.8041, -32.9519, -32.9707, -33.1561,
        -33.3227, -33.4987, -33.609, -33.7287, -33.7193, -33.7475, -33.8296, -33.9305, -34.0221, -34.0667, -34.2966, -34.4328, -34.4797, -34.597, -34.7707, -34.7801, -34.7707, -35.0336, -34.9514,
        -34.989, -34.8059, -34.7238, -34.6158, -34.6064, -34.4609, -34.4961, -34.5618, -34.6346, -34.4961, -34.2873, -34.1863, -33.9657, -33.7827, -33.6818, -33.4142, -33.3321, -33.203, -33.2288,
        -33.1561, -33.0622, -32.9237, -32.7665, -32.7759, -32.56, -32.506, -32.4779, -32.1399, -32.1024, -32.2056, -32.2244, -32.2056, -32.1024, -31.9428, -31.9615, -32.0179, -32.046, -32.0366] };
  WCL[1558] = struct {
   x = [132.315, 132.293, 132.195, 132.04, 131.688, 131.477, 131.268, 131.092, 130.892, 130.573, 130.287],
   y = [-32.0366, -32.0273, -31.9897, -31.8677, -31.6142, -31.5579, -31.5297, -31.5297, -31.5203, -31.5579, -31.5861] };
  WCL[1559] = struct {
   x = [136.548, 136.593, 136.802, 136.98, 137.165, 137.309, 137.442, 137.553, 137.574, 137.684, 137.839, 137.994, 138.104, 137.884, 137.64, 137.618, 137.487, 137.299, 137.079, 136.879, 136.649, 136.539,
        136.548],
   y = [-35.8174, -35.7728, -35.7001, -35.6649, -35.6203, -35.5921, -35.6297, -35.6367, -35.7446, -35.7728, -35.7986, -35.7986, -35.916, -35.8972, -35.916, -35.9958, -36.0779, -36.0333, -36.0403, -36.0403,
        -36.0239, -35.8808, -35.8174] };
  WCL[1560] = struct {
   x = [130.287, 129.956, 129.536, 129.184, 128.808, 128.468, 128.093, 127.905, 127.63, 127.332, 126.957, 126.692, 126.593, 126.307, 126.086, 125.821, 125.633, 125.391, 125.15, 124.84, 124.685, 124.542,
        124.333, 124.101, 124.002, 123.925, 123.838, 123.727, 123.659, 123.528, 123.307, 123.153, 122.977, 122.744, 122.592, 122.38, 122.183, 122.085, 121.996, 121.885, 121.599, 121.289, 121.069,
        120.773, 120.484, 120.21, 120.188, 120.177],
   y = [-31.5861, -31.5767, -31.5673, -31.5767, -31.7269, -31.8677, -32.0836, -32.0836, -32.1117, -32.2244, -32.2713, -32.2619, -32.2713, -32.2619, -32.215, -32.2713, -32.4121, -32.5224, -32.6445, -32.7102,
        -32.8228, -32.8768, -32.8956, -33.0082, -33.1936, -33.4236, -33.5081, -33.6536, -33.7827, -33.8202, -33.8648, -33.8648, -33.8109, -33.8296, -33.8484, -33.893, -33.9305, -33.8296, -33.7827,
        -33.8202, -33.8484, -33.7921, -33.7921, -33.9024, -33.9024, -33.893, -33.893, -33.893] };
  WCL[1561] = struct {
   x = [120.177, 119.989, 119.79, 119.613, 119.527, 119.56, 119.527, 119.405, 119.205, 118.886, 118.799, 118.621, 118.457, 118.367, 118.236, 118.081, 117.905, 117.884, 117.75, 117.598, 117.476, 117.398,
        117.133, 116.879, 116.593, 116.539, 116.471, 116.438, 116.307, 116.307, 116.119, 115.943, 115.844, 115.699, 115.502, 115.338, 115.093, 115.082, 115.072, 115.061, 115.072, 115.072],
   y = [-33.893, -33.8836, -33.9564, -34.0854, -34.1957, -34.2779, -34.2873, -34.3882, -34.4234, -34.4515, -34.5149, -34.6064, -34.7519, -34.8622, -34.9514, -34.996, -34.9256, -35.0781, -35.0336, -35.0875,
        -34.9702, -34.9796, -34.996, -35.0054, -34.9702, -34.9162, -34.8622, -34.8622, -34.8059, -34.8247, -34.8059, -34.6604, -34.5783, -34.4328, -34.2966, -34.2427, -34.2051, -34.0502, -33.893,
        -33.7545, -33.5715, -33.5527] };
  WCL[1562] = struct {
   x = [115.072, 115.082, 115.105, 115.26, 115.424, 115.612, 115.722, 115.722, 115.645, 115.645, 115.666, 115.734, 115.711, 115.699, 115.744, 115.699, 115.6, 115.502, 115.392, 115.347, 115.248, 115.126,
        115.007, 114.962],
   y = [-33.5527, -33.5433, -33.5621, -33.609, -33.5809, -33.3321, -33.1185, -32.9707, -32.7759, -32.5506, -32.6445, -32.4779, -32.2901, -32.1493, -31.8771, -31.6612, -31.4828, -31.3021, -31.0768, -30.8961,
        -30.7905, -30.6192, -30.3422, -30.1052] };
  WCL[1563] = struct {
   x = [30.9548, 30.7764, 30.6567, 30.5558, 30.3915, 30.2601, 30.0606],
   y = [-29.9292, -30.2085, -30.3986, -30.5699, -30.76, -30.9501, -31.1965] };
  WCL[1564] = struct {
   x = [30.0606, 29.8611, 29.6311, 29.4551, 29.2885, 29.1782, 29.0022, 28.7816, 28.5281, 28.2747, 28.1526, 27.9766, 27.8006, 27.634, 27.4814, 27.282, 27.1271, 26.9299, 26.7211, 26.5216, 26.3221, 26.1367,
        25.9161, 25.7612, 25.6603, 25.6274, 25.6955, 25.6838, 25.3974, 25.1651, 24.9445, 24.8131, 24.5596, 24.3273, 24.041, 23.8744, 23.6209, 23.3792, 23.2243, 23.0811, 22.884, 22.6517, 22.4968,
        22.3536, 22.1893, 22.1001, 21.9898, 21.8256, 21.7927, 21.6824, 21.5369, 21.2952, 21.2412, 21.0628, 20.8986, 20.7319, 20.6662, 20.5677, 20.4808, 20.3353, 20.2696, 20.1499],
   y = [-31.1965, -31.3584, -31.5368, -31.6494, -31.8207, -31.9522, -32.1775, -32.3722, -32.6163, -32.7736, -32.8674, -33.0434, -33.1819, -33.2734, -33.3673, -33.487, -33.5785, -33.6724, -33.7451, -33.7733,
        -33.7639, -33.7545, -33.7639, -33.7897, -33.8367, -33.8648, -34.0009, -34.1112, -34.076, -34.0197, -34.0385, -34.2309, -34.2309, -34.1676, -34.1019, -34.0667, -34.0291, -34.0479, -34.1582,
        -34.1676, -34.1112, -34.0573, -34.0479, -34.0667, -34.0854, -34.1676, -34.2309, -34.2849, -34.4046, -34.4328, -34.4234, -34.4046, -34.4586, -34.4328, -34.414, -34.4422, -34.4961, -34.4961,
        -34.5055, -34.5149, -34.5877, -34.705] };
  WCL[1565] = struct {
   x = [20.1499, 19.9833, 19.8613, 19.7744, 19.6853, 19.6195, 19.4647, 19.3543, 19.2886, 19.2558, 19.2675, 19.1901, 19.0234, 18.8474, 18.7606, 18.7817, 18.77, 18.6057, 18.4719, 18.4297, 18.4297, 18.4508,
        18.418, 18.3194, 18.3077, 18.3194, 18.3616, 18.3968, 18.3851, 18.2419, 18.1316, 18.0542, 17.9439, 18.0096, 17.965, 17.8453, 17.8007, 17.8125, 17.8665, 18.0753, 18.1645, 18.2513, 18.2302,
        18.2091, 18.0988, 17.9322, 17.8007, 17.7022, 17.6013, 17.5027, 17.3807, 17.2938, 17.1835, 17.0943],
   y = [-34.705, -34.7425, -34.7425, -34.7425, -34.7238, -34.6956, -34.705, -34.651, -34.6322, -34.6041, -34.4961, -34.4234, -34.4234, -34.4422, -34.4046, -34.3037, -34.1206, -34.1112, -34.13, -34.1863,
        -34.2943, -34.3952, -34.414, -34.2755, -34.1112, -33.9939, -33.9188, -33.8742, -33.7545, -33.6067, -33.4236, -33.3767, -33.2195, -33.2007, -33.0998, -33.0247, -33.0153, -32.8487, -32.7454,
        -32.8205, -32.7078, -32.5694, -32.2314, -31.9428, -31.7456, -31.5274, -31.3396, -31.227, -31.0651, -30.8186, -30.6755, -30.5042, -30.3047, -30.1029] };
  WCL[1566] = struct {
   x = [-68.6426, -68.6637, -68.6848, -68.6308],
   y = [-50.0181, -49.9336, -49.7834, -49.8702] };
  WCL[1567] = struct {
   x = [-67.9902, -67.8376, -67.7484, -67.6827, -67.617, -67.5935, -67.7484, -67.7038, -67.5396, -67.3072, -67.2626, -67.1758, -67.0092, -66.866, -66.6783, -66.5257, -66.392, -66.216, -66.0728, -65.9296,
        -65.9085, -65.763, -65.7865, -65.8639, -65.7982, -65.6644, -65.6644, -65.709, -65.8405, -66.0611, -66.2934, -66.5023, -66.7463, -66.9435, -67.143, -67.3072, -67.441, -67.5513, -67.5842,
        -67.5724, -67.5278, -67.4175, -67.319, -67.2626, -67.1641, -67.0866, -67.0092, -66.8566, -66.6689, -66.4811, -66.3591, -66.1714, -65.9836, -65.8522, -65.6551, -65.5659, -65.5096, -65.5213,
        -65.6199, -65.5002, -65.3664, -65.2796, -65.2021, -65.2021, -65.1787, -65.2467, -65.2021, -65.0472, -64.9815, -64.9041, -64.7492, -64.606, -64.5286, -64.43, -64.3526, -64.2869],
   y = [-50.0674, -49.9547, -49.7764, -49.598, -49.4126, -49.3399, -49.4197, -49.2741, -49.0371, -48.8987, -48.8846, -48.8118, -48.6874, -48.6358, -48.5208, -48.4481, -48.3589, -48.2415, -48.1758, -48.1758,
        -48.0421, -48.0045, -47.8872, -47.7393, -47.7018, -47.5539, -47.3521, -47.216, -47.1268, -47.1479, -47.1409, -47.0963, -47.0212, -46.871, -46.7114, -46.6199, -46.4767, -46.3007, -46.2092,
        -46.1411, -46.0027, -45.8712, -45.7633, -45.6553, -45.5474, -45.4864, -45.423, -45.2986, -45.2376, -45.2376, -45.1203, -45.0498, -45.0592, -45.0733, -45.0827, -45.0827, -45.0029, -44.9419,
        -44.7612, -44.6274, -44.5875, -44.5242, -44.3763, -44.1698, -43.9867, -43.7873, -43.6277, -43.4845, -43.3719, -43.2827, -43.2264, -43.17, -43.1395, -43.0973, -43.0574, -43.041] };
  WCL[1568] = struct {
   x = [-64.2869, -64.2751, -64.3737, -64.5192, -64.7163, -64.8712, -65.0261, -65.0261, -64.9369, -64.8384, -64.7163, -64.5614, -64.3643, -64.2657, -64.254, -64.1108, -63.9442, -63.7471, -63.6133, -63.5805,
        -63.625, -63.6368, -63.7799, -64.0005, -64.2986, -64.1766, -64.1202, -64.2751, -64.484, -64.5943, -64.5192, -64.7609, -65.0144, -65.0355, -65.0355, -65.0472, -65.1575, -65.1575, -64.9581,
        -64.8595, -64.7938, -64.5192, -64.3526, -64.1437, -63.9677, -63.8011, -63.5476, -63.3153, -63.1064, -62.8647, -62.6887, -62.5009, -62.3789, -62.3249, -62.3132, -62.3695, -62.4023, -62.3695],
   y = [-43.041, -43.0104, -42.9776, -42.9447, -42.9283, -42.872, -42.8157, -42.7335, -42.6702, -42.6608, -42.5716, -42.5317, -42.5716, -42.6537, -42.8157, -42.8649, -42.872, -42.8157, -42.7429, -42.5951,
        -42.3604, -42.2454, -42.0811, -42.1961, -42.2125, -42.2688, -42.3838, -42.4261, -42.4566, -42.3932, -42.2947, -42.2383, -42.0647, -41.8769, -41.6117, -41.431, -41.1729, -40.8725, -40.6894,
        -40.781, -40.8467, -40.9053, -40.964, -41.0321, -41.1236, -41.1635, -41.1729, -41.1729, -41.1823, -41.0813, -41.0391, -40.9734, -40.8725, -40.7551, -40.6542, -40.4547, -40.2506, -40.0394] };
  WCL[1569] = struct {
   x = [-73.7024, -73.7141, -73.7024, -73.7024, -73.7141, -73.7235, -73.8127, -73.8455, -73.8666, -73.8784, -73.8572, -73.8009, -73.7141, -73.6578, -73.6812, -73.5592, -73.4278, -73.2729, -73.1954, -73.1063,
        -73.0851, -72.9748, -72.8082, -72.6885, -72.5782, -72.4797, -72.3576, -72.313, -72.313, -72.4233, -72.5336, -72.7096, -72.766, -72.6111, -72.4562, -72.4445, -72.4233, -72.4233, -72.5125,
        -72.6557, -72.7331, -72.6322, -72.5336, -72.6651, -72.8082, -72.7754, -72.7542, -72.8199, -72.8528, -72.9514, -72.996, -73.0194, -73.0077, -72.9514, -72.8974, -72.9091, -72.9302, -72.8857,
        -72.9866, -73.0523, -73.0851, -73.1509, -73.2072, -73.2283, -73.2494],
   y = [-39.9126, -40.0394, -40.0652, -40.1849, -40.3937, -40.5627, -40.7387, -40.8983, -41.0649, -41.2714, -41.431, -41.5296, -41.5554, -41.6375, -41.7197, -41.769, -41.8018, -41.769, -41.7525, -41.6939,
        -41.5953, -41.546, -41.5554, -41.661, -41.7033, -41.7033, -41.661, -41.654, -41.6704, -41.7525, -41.7784, -41.8675, -41.999, -42.0482, -42.0482, -42.1562, -42.3346, -42.4894, -42.3017,
        -42.2782, -42.4097, -42.5387, -42.6115, -42.588, -42.7335, -42.879, -43.0175, -43.17, -43.2592, -43.332, -43.3883, -43.4775, -43.5643, -43.5972, -43.6277, -43.6535, -43.7168, -43.7873, -43.867,
        -43.9633, -44.0595, -44.0994, -44.1299, -44.1698, -44.1698] };
  WCL[1570] = struct {
   x = [-73.2494, -73.2494, -73.1954, -73.0617, -72.9514, -72.8528, -72.7542, -72.6885, -72.6885, -72.6885, -72.7214, -72.8317, -72.9091, -73.0617, -73.2494, -73.3386, -73.3386, -73.3715, -73.3715, -73.2072,
        -72.9514, -72.8317, -72.8199, -72.9302, -72.9302, -73.1063, -73.2612, -73.3386, -73.4372, -73.5029, -73.5263, -73.5263, -73.4489, -73.3057, -73.172, -73.172, -73.3057, -73.3926, -73.3926,
        -73.2729, -73.24, -73.2729, -73.3926, -73.47, -73.5029, -73.5475, -73.5592, -73.5709, -73.5921, -73.5709, -73.5146, -73.4606, -73.4278, -73.3597, -73.2729, -73.2823, -73.3715, -73.416, -73.416,
        -73.5263, -73.6578, -73.7235, -73.7469, -73.8338, -73.8338, -73.8009, -73.8127, -73.9112, -74.0004, -74.1224, -74.2093, -74.1647, -74.0544, -74.0661, -74.1553, -74.1647, -74.1318, -74.1975,
        -74.2327, -74.3407, -74.4745, -74.5519, -74.7045, -74.85, -75.026, -75.0142, -74.9696, -74.9039, -74.8054, -74.749, -74.749, -74.9157, -75.0588, -75.202, -75.3451, -75.4883, -75.5329, -75.5657,
        -75.5329, -75.3451, -75.3006, -75.3451, -75.4226, -75.3123, -75.1574, -75.0471, -74.9485, -74.7936, -74.6833, -74.5848, -74.4839, -74.3407, -74.221, -74.0872, -74.0333, -73.9769, -74.1553,
        -74.0872, -74.0215, -74.1764, -74.3642, -74.4627, -74.4088, -74.2867, -74.1764, -74.0215, -74.1553, -74.1647, -74.1318, -74.0779, -73.9676, -73.8455, -73.8009, -73.7469, -73.6906, -73.5709,
        -73.4278, -73.3386, -73.4606, -73.6366, -73.7469, -73.9112, -74.1318, -74.1975, -74.1764, -74.1764, -74.2421, -74.3407, -74.4839, -74.4956, -74.4839, -74.4088, -74.3642, -74.3078, -74.1975,
        -74.0333, -73.9347, -73.8572, -73.8127, -73.7563, -73.9441, -74.1224, -74.2327, -74.275, -74.1553, -74.2327, -74.2985, -74.3078, -74.3078, -74.2985, -74.2867, -74.275, -74.2421, -74.1436,
        -74.0544, -74.0215, -73.8572, -73.7798, -73.8572, -73.8901, -73.9347, -73.8901, -73.7798, -73.7798, -73.8572, -73.9676, -74.0779, -74.2421, -74.2421, -74.2421, -74.2867, -74.2985],
   y = [-44.1698, -44.1933, -44.2496, -44.2895, -44.3528, -44.4303, -44.4538, -44.55, -44.6509, -44.7847, -44.8621, -44.9325, -45.0029, -45.0358, -45.0264, -45.0663, -45.1531, -45.2681, -45.2845, -45.2986,
        -45.4089, -45.4089, -45.5004, -45.5169, -45.0663, -45.4394, -45.362, -45.3784, -45.4394, -45.5098, -45.5943, -45.6412, -45.6107, -45.5638, -45.5544, -45.6107, -45.6483, -45.7187, -45.7727,
        -45.7868, -45.8032, -45.8102, -45.8173, -45.8337, -45.8478, -45.9111, -46.0167, -46.1177, -46.2256, -46.1928, -46.1317, -46.0566, -45.9792, -45.9111, -45.8712, -45.9651, -46.1106, -46.2092,
        -46.2632, -46.3007, -46.3852, -46.4392, -46.5284, -46.5518, -46.4767, -46.3383, -46.2632, -46.3383, -46.4603, -46.4462, -46.3476, -46.2772, -46.1787, -46.0942, -46.1552, -46.0801, -45.9862,
        -45.9651, -45.9323, -45.8877, -45.9557, -45.9111, -45.8877, -45.8947, -45.9252, -46.0332, -46.0942, -46.1247, -46.1012, -46.1247, -46.2162, -46.2772, -46.2937, -46.4392, -46.5284, -46.5823,
        -46.7114, -46.8405, -46.9531, -46.9766, -46.8616, -46.7654, -46.7255, -46.6574, -46.6058, -46.6668, -46.749, -46.7935, -46.8405, -46.817, -46.7419, -46.7349, -46.7935, -46.8851, -46.9672,
        -47.0048, -47.1198, -47.1714, -47.2089, -47.2394, -47.3145, -47.4342, -47.4718, -47.5164, -47.561, -47.5891, -47.6642, -47.7229, -47.7534, -47.791, -47.805, -47.798, -47.7159, -47.791,
        -47.9529, -48.1242, -48.1829, -48.2181, -48.2275, -48.1453, -48.0796, -48.0491, -48.028, -48.0655, -48.1829, -48.2415, -48.1242, -48.0045, -47.9834, -48.1172, -48.2345, -48.3213, -48.4387,
        -48.4833, -48.4551, -48.4316, -48.4176, -48.3659, -48.5138, -48.6076, -48.5044, -48.5044, -48.5208, -48.6593, -48.7461, -48.7391, -48.8048, -48.9198, -49.0442, -49.1732, -49.253, -49.3469,
        -49.4337, -49.4689, -49.3539, -49.246, -49.1521, -49.131, -49.3047, -49.3821, -49.4689, -49.5417, -49.5769, -49.6121, -49.6121, -49.5839, -49.6262, -49.706, -49.7693, -49.7764, -49.8186,
        -49.9266] };
  WCL[1571] = struct {
   x = [-73.8784, -73.8338, -73.8338, -73.8244, -73.8127, -73.7469, -73.6812, -73.6038, -73.5709, -73.5263, -73.4818, -73.47, -73.4043, -73.3926, -73.5475, -73.6366, -73.7024, -73.7469, -73.7563, -73.6812,
        -73.5803, -73.5381, -73.7024, -73.6812, -73.5921, -73.5803, -73.7469, -73.7024, -73.7798, -73.9441, -74.2867, -74.397, -74.2985, -74.2421, -74.1764, -74.1647, -74.1882, -74.1764, -74.1107,
        -74.099, -74.0661, -73.9887, -73.9112, -73.8784],
   y = [-41.7525, -41.7267, -41.7455, -41.8347, -41.8769, -41.8675, -41.8675, -41.884, -41.9333, -42.0248, -42.1233, -42.1961, -42.3275, -42.4003, -42.4261, -42.5716, -42.5552, -42.6209, -42.6608, -42.7593,
        -42.7758, -42.8649, -42.879, -42.9612, -43.0902, -43.2123, -43.2029, -43.3648, -43.4446, -43.4352, -43.3648, -43.1958, -43.0574, -42.9048, -42.7335, -42.5786, -42.4261, -42.2782, -42.114,
        -41.9661, -41.884, -41.8605, -41.769, -41.7525] };
  WCL[1572] = struct {
   x = [-73.0077, -72.996, -72.9514, -72.9091, -72.8199, -72.7754, -72.766, -72.7988, -72.8082, -72.8082, -72.8082, -72.9866, -73.1297, -73.2612, -73.3503, -73.416, -73.4489, -73.3057, -73.1509, -73.0077],
   y = [-44.4068, -44.4467, -44.4538, -44.4772, -44.5007, -44.5641, -44.611, -44.6908, -44.6978, -44.7072, -44.8386, -44.9325, -44.956, -44.9325, -44.848, -44.7307, -44.6204, -44.5335, -44.4631, -44.4068] };
  WCL[1573] = struct {
   x = [-74.2985, -74.3196, -74.2985, -74.2093, -74.1107, -74.0333, -73.9676, -73.9558, -74.0779, -74.1975, -74.2327, -74.275, -74.2985],
   y = [-44.6509, -44.6603, -44.6438, -44.6438, -44.6438, -44.6603, -44.6814, -44.7142, -44.7612, -44.7847, -44.7448, -44.7072, -44.6509] };
  WCL[1574] = struct {
   x = [-74.2327, -74.3313, -74.3196, -74.275, -74.1553, -74.0004, -73.9676, -73.9769, -74.1107, -74.2327],
   y = [-45.0498, -45.0264, -44.9888, -44.9654, -44.9255, -44.9419, -44.9724, -44.9888, -45.0193, -45.0498] };
  WCL[1575] = struct {
   x = [-74.1647, -74.2327, -74.1882, -74.1107, -73.9558, -73.8127, -73.7469, -73.7469, -73.8666, -74.0004, -74.099, -74.1647],
   y = [-45.1296, -45.1296, -45.1062, -45.1062, -45.0897, -45.0827, -45.1907, -45.2916, -45.2446, -45.1601, -45.1531, -45.1296] };
  WCL[1576] = struct {
   x = [-74.1436, -74.1882, -74.1553, -74.0872, -73.9887, -73.9112, -73.8572, -73.8901, -73.9887, -74.0544, -74.1436],
   y = [-45.2376, -45.2212, -45.2306, -45.2446, -45.2845, -45.3385, -45.3925, -45.369, -45.3549, -45.3221, -45.2376] };
  WCL[1577] = struct {
   x = [-73.8666, -73.8244, -73.7681, -73.7469, -73.7563, -73.8572, -73.8784, -73.8784, -73.8666, -73.8666],
   y = [-45.7961, -45.7727, -45.8267, -45.9323, -45.9416, -45.9487, -45.8478, -45.7797, -45.8267, -45.7961] };
  WCL[1578] = struct {
   x = [-74.2985, -74.451, -74.6294, -74.6176, -74.5284, -74.4745, -74.397, -74.3078, -74.2656, -74.2985],
   y = [-47.7769, -47.7839, -47.7323, -47.6431, -47.6056, -47.561, -47.6056, -47.6642, -47.7604, -47.7769] };
  WCL[1579] = struct {
   x = [-74.9814, -74.9814, -75.0706, -75.2137, -75.3006, -75.2912, -75.256, -75.202, -75.0799, -74.9814],
   y = [-47.7018, -47.7534, -47.8355, -47.8496, -47.8121, -47.805, -47.7534, -47.7393, -47.7018, -47.7018] };
  WCL[1580] = struct {
   x = [-75.1363, -75.1363, -75.1034, -75.0025, -74.9157, -74.8148, -74.7819, -74.7936, -74.8382, -74.9368, -75.026, -75.1245, -75.1363],
   y = [-48.0937, -48.0796, -48.0867, -48.1242, -48.1312, -48.1594, -48.204, -48.3002, -48.3284, -48.3448, -48.2486, -48.1172, -48.0937] };
  WCL[1581] = struct {
   x = [-75.4437, -75.4554, -75.4437, -75.3991, -75.324, -75.3006, -75.2677, -75.2466, -75.202, -75.1574, -75.1034, -75.0706, -75.0706, -75.1128, -75.1691, -75.2137, -75.3006, -75.378, -75.4437],
   y = [-48.1688, -48.0726, -48.0726, -48.0655, -48.028, -48.0655, -48.1594, -48.2181, -48.3002, -48.3964, -48.4621, -48.5138, -48.5701, -48.6076, -48.5419, -48.4176, -48.3659, -48.3002, -48.1688] };
  WCL[1582] = struct {
   x = [-74.9157, -74.9157, -74.8382, -74.7725, -74.6951, -74.6387, -74.5284, -74.5073, -74.5942, -74.7397, -74.8711, -74.9368, -74.9485, -74.9157],
   y = [-48.4105, -48.4035, -48.4176, -48.4387, -48.4387, -48.5208, -48.5936, -48.6288, -48.6147, -48.5865, -48.5419, -48.4762, -48.4176, -48.4105] };
  WCL[1583] = struct {
   x = [-75.4672, -75.4883, -75.4883, -75.4883, -75.4437, -75.378, -75.324, -75.3006, -75.2912, -75.3569, -75.4343, -75.4766, -75.4672],
   y = [-48.6076, -48.5419, -48.5208, -48.4973, -48.4762, -48.4621, -48.5044, -48.5701, -48.6217, -48.6945, -48.7461, -48.6874, -48.6076] };
  WCL[1584] = struct {
   x = [-75.0142, -74.9814, -75.0025, -75.0142, -74.9931, -74.8922, -74.7936, -74.6716, -74.573, -74.5073, -74.4745, -74.4627, -74.4416, -74.4416, -74.4299, -74.4416],
   y = [-48.9855, -48.9714, -48.9127, -48.8705, -48.8048, -48.725, -48.6804, -48.7391, -48.8259, -48.9362, -49.0653, -49.2108, -49.3821, -49.6403, -49.7905, -49.9195] };
  WCL[1585] = struct {
   x = [-74.7279, -74.85, -74.85, -74.85, -74.85, -74.8054, -74.8148, -74.9039, -75.0799, -75.1903, -75.2348, -75.202, -75.1128, -74.9814, -74.9603, -75.0706, -75.1691, -75.1245, -75.0142],
   y = [-50.0463, -49.9336, -49.8045, -49.72, -49.6332, -49.5628, -49.5182, -49.483, -49.4337, -49.3399, -49.3047, -49.2601, -49.2741, -49.3328, -49.1967, -49.1873, -49.1169, -49.0442, -48.9855] };
  WCL[1586] = struct {
   x = [-75.5212, -75.5329, -75.5329, -75.5446, -75.5094, -75.4437, -75.3897, -75.324, -75.2231, -75.2677, -75.4226, -75.5212, -75.5212],
   y = [-49.706, -49.6989, -49.6919, -49.6473, -49.6332, -49.6332, -49.6403, -49.7482, -49.8468, -49.8632, -49.8562, -49.7905, -49.706] };
  WCL[1587] = struct {
   x = [-176.322, -176.313, -176.313, -176.346, -176.379, -176.379, -176.4, -176.51, -176.609, -176.686, -176.632, -176.653, -176.743, -176.862, -176.841, -176.677, -176.51, -176.4, -176.322],
   y = [-43.7919, -43.7755, -43.8225, -43.8717, -43.9421, -44.0219, -44.0454, -44.0689, -44.1182, -44.0689, -43.9351, -43.8858, -43.8788, -43.8389, -43.8154, -43.7755, -43.7661, -43.806, -43.7919] };
  WCL[1588] = struct {
   x = [172.068, 172.1, 172.112, 172.112, 172.122, 172.178, 172.265, 172.333, 172.41, 172.485, 172.518, 172.596, 172.628, 172.628, 172.628, 172.697, 172.807, 172.905, 172.983, 172.983, 172.983, 173.037,
        173.159, 173.236, 173.335, 173.358, 173.401, 173.502, 173.567, 173.588, 173.645, 173.699, 173.755, 173.842, 173.875, 173.854, 173.776, 173.797, 173.821, 173.788, 173.764, 173.832, 173.943,
        173.943, 173.898, 173.919, 173.952, 173.976, 174.018, 174.074, 174.119, 174.14, 174.161, 174.119, 174.062, 174.008, 174.062, 174.128, 174.086, 174.053, 174.095, 174.14, 174.184, 174.14,
        174.053, 173.976, 173.943, 173.832, 173.722, 173.699, 173.621, 173.502, 173.424, 173.347, 173.281, 173.236, 173.18, 173.114, 173.103, 173.093, 173.037, 172.95, 172.894, 172.882, 172.873,
        172.873, 172.84, 172.828, 172.816, 172.807, 172.795, 172.783, 172.772, 172.772, 172.762, 172.739, 172.729, 172.729, 172.697, 172.685, 172.739, 172.861, 172.992, 173.037, 173.016, 172.927,
        172.917, 172.718, 172.518, 172.399, 172.342, 172.342, 172.288, 172.199, 172.145, 172.112, 171.969, 171.835, 171.77, 171.671, 171.582, 171.549, 171.495, 171.46, 171.45, 171.46, 171.406, 171.284,
        171.239, 171.23, 171.218, 171.119, 171.119, 171.141, 171.054, 170.965, 170.953, 170.92, 170.92, 170.854, 170.81, 170.699, 170.667, 170.657, 170.678, 170.667, 170.601, 170.622, 170.699, 170.657,
        170.512, 170.359, 170.258, 170.171, 170.028],
   y = [-41.0508, -41.0602, -41.0602, -40.9851, -40.8678, -40.8255, -40.7927, -40.734, -40.6589, -40.6495, -40.6073, -40.5744, -40.6237, -40.6918, -40.7763, -40.8584, -40.8842, -40.8748, -40.9593, -41.0508,
        -41.1917, -41.2926, -41.3418, -41.2832, -41.2339, -41.2081, -41.1752, -41.1424, -41.0931, -41.0602, -41.0344, -41.0086, -40.9757, -40.9664, -40.9851, -41.0016, -41.0602, -41.0837, -41.1846,
        -41.2667, -41.316, -41.2761, -41.2339, -41.2269, -41.1682, -41.1518, -41.0767, -41.0508, -41.0344, -41.0344, -41.0673, -41.1095, -41.1353, -41.1917, -41.2433, -41.2832, -41.2926, -41.3747,
        -41.4498, -41.5249, -41.5977, -41.6892, -41.7713, -41.8628, -41.952, -42.0436, -42.1421, -42.2642, -42.2806, -42.3369, -42.412, -42.5106, -42.6467, -42.7617, -42.8673, -42.8908, -42.9471,
        -42.9729, -42.9729, -42.987, -43.0363, -43.0762, -43.109, -43.1161, -43.1161, -43.1254, -43.1325, -43.1419, -43.1489, -43.1489, -43.1653, -43.1912, -43.1912, -43.2052, -43.2616, -43.3578,
        -43.447, -43.5197, -43.6065, -43.6629, -43.6723, -43.7028, -43.7591, -43.8717, -43.9187, -43.8318, -43.9187, -43.8717, -43.8154, -43.7661, -43.7919, -43.8788, -43.9351, -43.8788, -43.8717,
        -43.9492, -44.0055, -44.0384, -44.0783, -44.1252, -44.1182, -44.0454, -44.029, -44.0384, -44.0923, -44.1557, -44.3153, -44.4819, -44.6298, -44.78, -44.8738, -44.9208, -44.9747, -45.078,
        -45.1555, -45.2259, -45.2939, -45.3948, -45.5051, -45.5896, -45.6366, -45.7117, -45.775, -45.7985, -45.829, -45.8736, -45.9135, -45.9205, -45.937, -46.0214, -46.0121, -46.0895, -46.174,
        -46.2585, -46.3101] };
  WCL[1589] = struct {
   x = [169.995, 170.115, 170.139, 170.225, 170.336, 170.437, 170.547, 170.732, 170.854, 170.932, 171.009, 171.108, 171.141, 171.164, 171.206, 171.23, 171.251, 171.274, 171.373, 171.483, 171.57, 171.704,
        171.814, 171.88, 171.99, 172.023, 172.056, 172.068],
   y = [-43.3507, -43.2944, -43.1818, -43.1325, -43.0527, -43.0034, -42.9964, -42.9072, -42.818, -42.7523, -42.6561, -42.5575, -42.4683, -42.4284, -42.3134, -42.2407, -42.1421, -42.0177, -41.8535, -41.8042,
        -41.7478, -41.6821, -41.6563, -41.5648, -41.4005, -41.2926, -41.1424, -41.0508] };
  WCL[1590] = struct {
   x = [176.862, 176.775, 176.71, 176.62, 176.588, 176.477, 176.367, 176.203, 176.081, 175.982, 175.771, 175.639, 175.452, 175.297, 175.177, 175.088, 174.712, 174.647, 174.647, 174.659, 174.745, 174.802,
        174.879, 175.001, 175.1, 175.133, 175.187, 175.187, 175.111, 175.001],
   y = [-39.9924, -40.1098, -40.2201, -40.3304, -40.4407, -40.5322, -40.6495, -40.8584, -41.018, -41.1518, -41.3747, -41.4404, -41.5319, -41.5812, -41.492, -41.3911, -41.4005, -41.2926, -41.1518, -41.2269,
        -41.1259, -41.0837, -40.9429, -40.842, -40.6824, -40.4993, -40.3632, -40.2459, -40.145, -40.0347] };
  WCL[1591] = struct {
   x = [167.879, 167.911, 167.911, 167.921, 167.989, 168.043, 168.064, 168.043, 167.954, 167.944, 168.022, 168.087, 168.099, 168.099, 168.01, 167.932, 167.834, 167.768, 167.691, 167.646, 167.602, 167.503,
        167.482, 167.503, 167.524, 167.569, 167.613, 167.658, 167.7, 167.7, 167.712, 167.778, 167.834, 167.879],
   y = [-46.7443, -46.7067, -46.7138, -46.7583, -46.8123, -46.8663, -46.9109, -46.9109, -46.9344, -46.9484, -46.9719, -46.986, -47.0165, -47.0846, -47.0986, -47.1667, -47.1972, -47.1503, -47.1667, -47.2183,
        -47.3098, -47.2793, -47.2254, -47.1503, -47.0916, -47.054, -47.0001, -46.9414, -46.8663, -46.7677, -46.7208, -46.7138, -46.7208, -46.7443] };
  WCL[1592] = struct {
   x = [170.028, 169.885, 169.817, 169.808, 169.718, 169.686, 169.674, 169.641, 169.587, 169.498, 169.366, 169.233, 169.158, 169.08, 168.958, 168.848, 168.815, 168.749, 168.639, 168.517, 168.451, 168.451,
        168.341, 168.308, 168.308, 168.275, 168.231, 168.12, 168.01, 167.9, 167.768, 167.712, 167.613, 167.447, 167.372, 167.315, 167.151, 166.973, 166.775, 166.665, 166.719, 166.82, 166.874, 166.808,
        166.677, 166.621, 166.653, 166.698, 166.686, 166.621, 166.489, 166.456, 166.51, 166.609, 166.743, 166.874, 166.829, 166.797, 166.874, 166.93, 166.841, 166.743, 166.731, 166.841, 166.898,
        166.841, 166.841, 166.93, 166.973, 167.029, 167.062, 167.128, 167.172, 167.16, 167.074, 167.008, 167.017, 167.041, 167.05, 167.074, 167.106, 167.205, 167.271, 167.315, 167.336, 167.372,
        167.426, 167.458, 167.426, 167.447, 167.503, 167.557, 167.712, 167.822, 167.843, 167.944, 168.064, 168.132, 168.209, 168.308, 168.43, 168.573, 168.705, 168.869, 169.014, 169.146, 169.268,
        169.388, 169.531, 169.664, 169.862, 169.995],
   y = [-46.3101, -46.3336, -46.3641, -46.4181, -46.4415, -46.4861, -46.5237, -46.5401, -46.5636, -46.5917, -46.6222, -46.6527, -46.6457, -46.6692, -46.6903, -46.6527, -46.5471, -46.5401, -46.5847, -46.6081,
        -46.5917, -46.5636, -46.5917, -46.5237, -46.4486, -46.4626, -46.4626, -46.4181, -46.4181, -46.411, -46.4016, -46.2491, -46.1881, -46.174, -46.2115, -46.2585, -46.2796, -46.2585, -46.242,
        -46.181, -46.0895, -46.0121, -45.9276, -45.9886, -46.0731, -46.0825, -45.998, -45.937, -45.9205, -45.9675, -45.9815, -45.9064, -45.8431, -45.8126, -45.7985, -45.721, -45.7117, -45.6741,
        -45.6366, -45.5497, -45.5591, -45.5356, -45.4582, -45.4582, -45.3948, -45.3643, -45.3103, -45.3103, -45.3502, -45.3338, -45.3268, -45.3338, -45.3103, -45.2798, -45.2704, -45.2165, -45.1461,
        -45.0991, -45.0686, -44.9912, -44.9607, -44.9043, -44.8903, -44.9372, -45.0146, -45.0381, -45.0381, -44.9912, -44.8668, -44.8105, -44.7964, -44.6697, -44.6392, -44.5594, -44.4256, -44.3317,
        -44.2848, -44.0923, -44.0853, -44.029, -44.029, -43.982, -43.982, -43.9492, -43.8858, -43.7919, -43.7192, -43.6629, -43.5995, -43.5103, -43.4071, -43.3507] };
  WCL[1593] = struct {
   x = [148.28, 148.336, 148.325, 148.137, 148.104],
   y = [-39.9596, -40.1379, -40.2224, -40.2905, -40.1215] };
  WCL[1594] = struct {
   x = [148.137, 148.092, 148.17, 148.28, 148.358, 148.468, 148.501, 148.369, 148.292, 148.137, 148.292, 148.137],
   y = [-40.4242, -40.3726, -40.3562, -40.3397, -40.3233, -40.3562, -40.45, -40.4665, -40.45, -40.4242, -40.45, -40.4242] };
  WCL[1595] = struct {
   x = [144.795, 144.807, 144.84, 144.983, 145.117, 145.227, 145.304, 145.436, 145.558, 145.788, 145.997, 146.217, 146.405, 146.583, 146.682, 146.759, 146.879, 147.067, 147.255, 147.463, 147.651, 147.761,
        147.905, 147.961, 148.125, 148.28, 148.325, 148.346, 148.369, 148.336, 148.336, 148.336, 148.358, 148.369, 148.423, 148.379, 148.379, 148.358, 148.301, 148.268, 148.17, 148.081, 148.092,
        148.006, 147.994, 147.982, 147.905, 147.785, 147.607, 147.564, 147.564, 147.496, 147.377, 147.332, 147.243, 147.144, 147.067, 147.034, 147.034, 146.902, 146.792, 146.682, 146.56, 146.473,
        146.274, 146.098, 146.032, 146.152, 146.229, 146.053, 145.877, 145.743, 145.656, 145.591, 145.513, 145.436, 145.379, 145.325, 145.281, 145.227, 145.215, 145.325, 145.447, 145.567, 145.436,
        145.314, 145.203, 145.105, 145.006, 144.941, 144.851, 144.774, 144.795, 144.786, 144.795],
   y = [-40.6753, -40.6753, -40.7176, -40.7598, -40.8279, -40.7763, -40.7763, -40.8185, -40.8607, -40.9851, -41.1095, -41.1353, -41.1518, -41.1682, -41.1682, -41.1189, -41.0602, -41.0016, -41.0109, -40.9851,
        -40.8279, -40.8678, -40.8936, -40.7692, -40.8185, -40.9687, -41.1682, -41.2761, -41.4169, -41.5671, -41.715, -41.8394, -42.0037, -42.0858, -42.1679, -42.3393, -42.2008, -42.0764, -41.9778,
        -42.0529, -42.175, -42.2994, -42.5035, -42.52, -42.6561, -42.7711, -42.8274, -42.8274, -42.7781, -42.8344, -42.9494, -43.0621, -43.1184, -43.2545, -43.1747, -43.1419, -43.2146, -43.3601,
        -43.4728, -43.5925, -43.6089, -43.5361, -43.5127, -43.5526, -43.4962, -43.4962, -43.4164, -43.3836, -43.2968, -43.2968, -43.2311, -43.0292, -42.9236, -42.8837, -42.8204, -42.7312, -42.5833,
        -42.4449, -42.3228, -42.1421, -42.1421, -42.29, -42.4378, -42.29, -42.2172, -42.0764, -41.9379, -41.7643, -41.6164, -41.5085, -41.3606, -41.2175, -41.018, -40.8443, -40.6753] };
  WCL[1596] = struct {
   x = [70.0272, 69.9991, 69.9991],
   y = [-49.1638, -49.1826, -49.2554] };
  WCL[1597] = struct {
   x = [68.9664, 69.9991],
   y = [-49.1099, -49.1451] };
  WCL[1598] = struct {
   x = [69.9709, 70.168, 70.3628, 70.2525, 70.3628, 70.2525, 70.1962, 70.1399, 69.9991, 69.9991, 70.3065, 70.4191, 70.5036, 70.5858, 70.6421, 70.5599, 70.3346, 70.3065, 70.2525, 70.1399, 70.0554],
   y = [-49.6567, -49.6919, -49.6567, -49.6027, -49.5464, -49.5112, -49.53, -49.5839, -49.5112, -49.3657, -49.3281, -49.4197, -49.3469, -49.2202, -49.1099, -49.0348, -49.0723, -49.1451, -49.1099, -49.1451,
        -49.1451] };
  WCL[1599] = struct {
   x = [69.9991, 69.9709, 69.9169, 69.8324, 69.7479, 69.6376, 69.6376, 69.5813, 69.4687, 69.6916, 69.7479, 69.8043, 69.7479, 69.7761, 69.6658, 69.5813, 69.3842, 69.1894, 69.1049, 69.1894, 69.1331, 69.2739,
        69.2739, 69.1331, 69.1612, 69.0767, 68.9664, 68.9664, 68.9101, 68.8538, 68.9101, 68.9383, 68.9101, 68.9664],
   y = [-49.2554, -49.2741, -49.2929, -49.3117, -49.3117, -49.2929, -49.2202, -49.1638, -49.1099, -49.1286, -49.1099, -49.0535, -49.0183, -48.9808, -48.9432, -48.9996, -49.0535, -49.0535, -49.0348, -48.9432,
        -48.908, -48.8517, -48.7414, -48.7954, -48.6663, -48.6475, -48.6311, -48.6851, -48.7414, -48.8142, -48.8893, -48.9432, -49.0348, -49.1099] };
  WCL[1600] = struct {
   x = [69.9991, 69.0509, 69.1612, 69.2457, 69.2457, 69.1612, 69.0767, 69.0228, 69.1894, 69.1049, 69.1894, 69.302, 69.4123, 69.5813, 69.5813, 69.7761, 69.7761, 69.8324, 69.9169, 69.9709],
   y = [-49.1451, -49.2929, -49.2929, -49.2014, -49.3281, -49.3281, -49.4384, -49.5464, -49.6755, -49.6379, -49.6191, -49.5839, -49.5112, -49.53, -49.6567, -49.6567, -49.53, -49.5112, -49.5652, -49.6567] };
  WCL[1601] = struct {
   x = [37.6127, 37.6456, 37.7019, 37.7113, 37.8122, 37.8216, 37.8005, 37.6691, 37.601, 37.6127],
   y = [-46.8804, -46.8733, -46.8569, -46.8733, -46.9109, -46.986, -47.0165, -47.0165, -46.925, -46.8804] };
  WCL[1602] = struct {
   x = [-26.3597, -26.3033, -26.2822, -26.2611, -26.2611, -26.3597, -26.4582, -26.4911, -26.4911, -26.4371, -26.3597],
   y = [-58.408, -58.4151, -58.4033, -58.4432, -58.4949, -58.4949, -58.4597, -58.4315, -58.4033, -58.4033, -58.408] };
  WCL[1603] = struct {
   x = [-38.1243, -38.1572, -38.1126, -38.0563, -37.9812, -37.8709, -37.7817, -37.7043, -37.5822, -37.4508, -37.4508, -37.3522, -37.2419, -37.2091, -37.1316, -37.0659, -36.9979, -36.911, -36.8453, -36.8125,
        -36.7561, -36.7115, -36.6341, -36.667, -36.7022, -36.6012, -36.5684, -36.5355, -36.437, -36.3713, -36.2821, -36.2164, -36.1506, -36.0943, -36.0732, -35.9629, -35.93, -35.9066, -35.9066,
        -35.8855, -35.8291, -35.8526, -35.9512, -35.9958, -35.984, -35.9512, -35.984, -36.0403, -36.1389, -36.2492, -36.3032, -36.3267, -36.3924, -36.4816, -36.5238, -36.5919, -36.7115, -36.7561,
        -36.789, -36.8993, -37.0424, -37.1527, -37.2748, -37.4062, -37.4062, -37.3616, -37.3851, -37.3968, -37.5071, -37.6503, -37.7488, -37.7488, -37.7371, -37.77, -37.77, -37.8028, -37.892, -37.9694,
        -38.0563, -38.1337, -38.1243],
   y = [-54.0288, -54.003, -54.0101, -54.003, -53.996, -53.9889, -53.9772, -54.003, -54.003, -53.9772, -54.0148, -54.0218, -54.0476, -54.0992, -54.0875, -54.0687, -54.0945, -54.0945, -54.1251, -54.1321,
        -54.1391, -54.1579, -54.1908, -54.2166, -54.2541, -54.2541, -54.294, -54.3269, -54.3527, -54.3011, -54.3269, -54.409, -54.4794, -54.5052, -54.5381, -54.5944, -54.5757, -54.5944, -54.6461,
        -54.7094, -54.7728, -54.7728, -54.7681, -54.7869, -54.8244, -54.869, -54.876, -54.8948, -54.869, -54.7728, -54.7352, -54.6719, -54.639, -54.5827, -54.5311, -54.4935, -54.4677, -54.4489,
        -54.409, -54.3644, -54.3269, -54.3011, -54.2753, -54.2682, -54.2424, -54.2025, -54.1767, -54.1391, -54.1509, -54.172, -54.165, -54.1321, -54.1133, -54.0945, -54.0734, -54.0617, -54.0547,
        -54.0359, -54.0547, -54.0148, -54.0288] };
  WCL[1604] = struct {
   x = [-59.3139, -59.2928, -59.2693, -59.1825, -59.159, -59.2036, -59.2482, -59.3468, -59.4453, -59.5556, -59.6331, -59.6777, -59.7551, -59.8537, -59.9311, -59.9757],
   y = [-51.4239, -51.3886, -51.4168, -51.4379, -51.4919, -51.5271, -51.6163, -51.6491, -51.7383, -51.8416, -51.8744, -51.9237, -51.9918, -52.0035, -51.9918, -51.9965] };
  WCL[1605] = struct {
   x = [-60.0203, -59.964],
   y = [-51.7313, -51.7242] };
  WCL[1606] = struct {
   x = [-60.0414, -59.9429, -59.8983, -59.7105, -59.5462, -59.4242, -59.3139, -59.3139],
   y = [-51.506, -51.4708, -51.4497, -51.4637, -51.4919, -51.4919, -51.4849, -51.4239] };
  WCL[1607] = struct {
   x = [-58.4644, -58.3869, -58.3447, -58.2766, -58.2438, -58.2015, -58.0795, -57.9128, -57.8378, -57.7932, -57.7157, -57.7157, -57.9035, -57.9481, -57.8378, -57.7157, -57.6946, -57.7814, -57.9246, -58.0466,
        -58.178, -58.3329, -58.5089, -58.753, -58.8516, -58.7624, -58.6427, -58.6075, -58.5653, -58.5653, -58.5981, -58.6427, -58.7624, -58.8516, -58.9502, -59.0487, -59.0159, -58.9736, -58.8844,
        -58.807, -58.807, -58.9502, -59.0933, -59.1379, -59.2271, -59.2928, -59.2811, -59.2599, -59.2599, -59.2036, -59.2928, -59.3914, -59.4805, -59.5908, -59.5791, -59.5017, -59.5017, -59.4571,
        -59.4359, -59.4125, -59.3585, -59.3139, -59.159, -59.105, -59.0065, -59.0159, -59.0159, -58.983, -58.983, -58.9173, -58.8633, -58.7741, -58.7859, -58.7741, -58.7084, -58.5864, -58.4644],
   y = [-51.3276, -51.3065, -51.3464, -51.4098, -51.4708, -51.4309, -51.4239, -51.4309, -51.4426, -51.4778, -51.5201, -51.574, -51.5952, -51.6233, -51.628, -51.6562, -51.7172, -51.7665, -51.7806, -51.8486,
        -51.9167, -51.9284, -51.9167, -51.8815, -51.8815, -51.9355, -51.9847, -52.0106, -52.0716, -52.1537, -52.1396, -52.0786, -52.0997, -52.1209, -52.0575, -52.0786, -52.1326, -52.2405, -52.2405,
        -52.2218, -52.2476, -52.2476, -52.2359, -52.2007, -52.1795, -52.2077, -52.2476, -52.2898, -52.3438, -52.3626, -52.3485, -52.3086, -52.2218, -52.1326, -52.1068, -52.0645, -51.9847, -51.9495,
        -51.9237, -51.8604, -51.8486, -51.8416, -51.7594, -51.7242, -51.6703, -51.5881, -51.56, -51.5201, -51.4708, -51.4637, -51.4309, -51.4309, -51.3534, -51.2995, -51.3276, -51.3464, -51.3276] };
  WCL[1608] = struct {
   x = [-70.0648, -69.8981, -69.7996, -69.6001, -69.5226, -69.4804, -69.3584, -69.2269, -69.0838, -69.0063, -68.8397, -68.6426, -68.4314, -68.3657, -68.3657],
   y = [-52.5785, -52.5175, -52.5104, -52.4846, -52.4095, -52.3227, -52.2758, -52.2476, -52.2546, -52.2898, -52.3438, -52.3438, -52.3767, -52.3626, -52.3227] };
  WCL[1609] = struct {
   x = [-68.3657, -68.3657, -68.3891, -68.4314, -68.5088, -68.6308, -68.7411, -68.8069, -68.8843, -69.0157, -69.126, -69.2269, -69.3912, -69.2809, -69.0603, -69.0063, -69.0603, -69.1166, -69.1941, -69.2927,
        -69.403, -69.2692, -69.1495, -69.0392, -68.8726, -68.6637, -68.4877, -68.4548, -68.4548, -68.4994, -68.6426],
   y = [-52.3227, -52.3086, -52.2945, -52.2758, -52.2147, -52.1209, -52.0387, -51.9495, -51.8486, -51.6984, -51.6773, -51.6914, -51.6093, -51.5811, -51.5881, -51.513, -51.3347, -51.2103, -51.0507, -51.0836,
        -51.0836, -50.9592, -50.7081, -50.471, -50.3725, -50.2809, -50.2105, -50.1894, -50.1307, -50.0674, -50.0181] };
  WCL[1610] = struct {
   x = [-68.6308, -68.4877, -68.3539, -68.2436, -68.1685, -67.9902],
   y = [-49.8702, -50.0322, -50.1167, -50.1237, -50.1237, -50.0674] };
  WCL[1611] = struct {
   x = [-68.652, -68.5534, -68.5323, -68.4548, -68.3985, -68.3117, -68.2225, -68.1568, -68.1451, -68.2225, -68.3445, -68.4548, -68.4877, -68.422, -68.2999, -68.1896, -68.1122, -68.0465, -67.9902, -67.9033,
        -67.7484, -67.5842, -67.5067, -67.4293, -67.319, -67.1758, -67.0866, -66.9669, -66.8895, -66.8003, -66.69, -66.6126, -66.4811, -66.3591, -66.2277, -66.0939, -65.9179, -65.8076, -65.587,
        -65.357, -65.1575, -65.1458, -65.1787, -65.2467, -65.3242, -65.4767, -65.6316, -65.7419, -65.8405, -65.9742, -66.1714, -66.3051, -66.3474, -66.514, -66.69, -66.812, -66.9669, -67.1101,
        -67.3072, -67.4832, -67.6381, -67.7813, -67.8376, -67.9362, -68.0676, -68.2108, -68.3774, -68.4994, -68.5863],
   y = [-52.6583, -52.7123, -52.7381, -52.8061, -52.8578, -52.9446, -53.0455, -53.1323, -53.1769, -53.1112, -53.0925, -53.1183, -53.191, -53.2825, -53.3107, -53.3764, -53.4679, -53.5853, -53.6698, -53.7308,
        -53.8012, -53.8599, -53.9045, -53.9889, -54.0617, -54.0805, -54.1251, -54.1509, -54.2166, -54.2424, -54.2612, -54.3198, -54.402, -54.4607, -54.524, -54.5757, -54.639, -54.6789, -54.6977,
        -54.6977, -54.6719, -54.7423, -54.862, -54.9136, -54.9253, -54.977, -54.9511, -54.9394, -54.9253, -54.9652, -54.9957, -55.0215, -55.0591, -55.0333, -54.9582, -54.9253, -54.9136, -54.8948,
        -54.869, -54.876, -54.876, -54.869, -54.8502, -54.8244, -54.8127, -54.8056, -54.8432, -54.8878, -54.8948] };
  WCL[1612] = struct {
   x = [-69.1589, -69.1495, -69.2481, -69.3701, -69.5015, -69.6329, -69.7339, -69.7761, -69.8113, -69.8442, -69.8442, -69.8442, -69.7996, -69.7667, -69.7339, -69.6447, -69.5133, -69.4687, -69.2809, -69.1941,
        -69.1589],
   y = [-54.9957, -55.0075, -55.0215, -55.0333, -55.0591, -55.0591, -55.0779, -55.0779, -55.0661, -55.0145, -54.977, -54.9324, -54.9206, -54.9136, -54.9253, -54.9136, -54.9066, -54.9206, -54.9394, -54.9582,
        -54.9957] };
  WCL[1613] = struct {
   x = [-69.9427, -69.9099, -69.8113, -69.6329, -69.4804, -69.3372, -69.2152, -69.1378, -69.0392, -68.95, -68.8186, -68.6754, -68.5323, -68.4314, -68.3445, -68.2882, -68.2554, -68.3445, -68.4642, -68.5863,
        -68.7083, -68.8514, -68.896, -68.8303, -68.6966, -68.5651, -68.4102, -68.2671, -68.1896, -68.1779, -68.1451, -68.3211, -68.4431, -68.5088, -68.4994, -68.3539, -68.2014, -68.0911, -68.0676,
        -67.9902, -67.969, -67.9151, -67.9151, -67.969, -68.0676, -68.1333, -68.1451, -68.2436, -68.3657, -68.4548, -68.598, -68.652, -68.7411, -68.7623, -68.774, -68.7857, -68.8726, -68.9054,
        -68.9054, -68.9946, -69.2152, -69.2927, -69.2363, -69.0721, -68.9946, -69.0603, -69.2809, -69.3466, -69.3912, -69.403, -69.4804, -69.6118, -69.6001, -69.4687, -69.3701, -69.3372, -69.3466,
        -69.4569, -69.5907, -69.7339, -69.8442, -69.9427, -69.9638, -69.9545, -69.9427],
   y = [-55.1741, -55.153, -55.1483, -55.1295, -55.1037, -55.0849, -55.0521, -55.0403, -55.0286, -55.0286, -55.0333, -55.0028, -54.9699, -54.977, -55.0333, -55.0591, -55.1225, -55.1412, -55.153, -55.1483,
        -55.1225, -55.092, -55.1037, -55.1412, -55.1858, -55.1976, -55.2116, -55.2116, -55.2422, -55.2867, -55.3126, -55.3173, -55.3126, -55.3313, -55.3548, -55.3806, -55.3924, -55.4299, -55.5003,
        -55.5238, -55.6106, -55.681, -55.7115, -55.7115, -55.6622, -55.6106, -55.5355, -55.5238, -55.4628, -55.4745, -55.4933, -55.512, -55.4674, -55.4557, -55.4182, -55.3736, -55.3478, -55.3126,
        -55.3055, -55.2727, -55.268, -55.3243, -55.3924, -55.4862, -55.5496, -55.5566, -55.5496, -55.5355, -55.4862, -55.444, -55.4111, -55.3618, -55.336, -55.2985, -55.2609, -55.2304, -55.2116,
        -55.2234, -55.2727, -55.2914, -55.2914, -55.2727, -55.2163, -55.1788, -55.1741] };
  WCL[1614] = struct {
   x = [-68.0911, -68.1122, -68.2108, -68.2554, -68.2788, -68.2225, -68.0676, -67.9479, -67.793, -67.6381, -67.5067, -67.3401, -67.1758, -67.1101, -67.0772, -67.042, -67.042, -67.0538, -67.1312, -67.2298,
        -67.3307, -67.3636, -67.3729, -67.3964, -67.495, -67.5724, -67.5724, -67.6827, -67.7273, -67.6945, -67.7813, -67.8916, -68.0136, -68.0465, -68.0676, -68.0793, -68.0911, -66.6008, -66.5586,
        -66.4811, -66.4248, -66.4366, -66.514, -66.5797, -66.6126, -66.6126],
   y = [-55.0145, -55.0333, -55.0215, -55.0075, -54.9582, -54.9511, -54.9511, -54.9324, -54.9206, -54.9394, -54.9582, -54.9652, -54.977, -55.0286, -55.1107, -55.1788, -55.2163, -55.2539, -55.3055, -55.3313,
        -55.336, -55.3173, -55.268, -55.2422, -55.2234, -55.2234, -55.268, -55.2727, -55.2609, -55.2351, -55.2304, -55.2234, -55.2046, -55.153, -55.1037, -55.0521, -55.0145, -55.2234, -55.2163,
        -55.2116, -55.2492, -55.2727, -55.3055, -55.2867, -55.2539, -55.2163] };
  WCL[1615] = struct {
   x = [-64.4089, -64.3972, -64.3197, -64.2305, -64.1766, -64.132, -64.0451, -63.956, -63.8011, -63.7142, -63.6814, -63.6908, -63.6908, -63.7565, -63.8456, -63.9442, -64.0663, -64.0874, -64.1554, -64.2657,
        -64.3737, -64.4089, -64.4957, -64.5732, -64.606, -64.5849, -64.5192, -64.5075, -64.4629, -64.4089],
   y = [-54.7986, -54.7939, -54.7939, -54.7728, -54.7611, -54.7728, -54.7728, -54.7728, -54.7423, -54.7352, -54.754, -54.7986, -54.8197, -54.8315, -54.8244, -54.8127, -54.8244, -54.8831, -54.869, -54.862,
        -54.8948, -54.9206, -54.9066, -54.9066, -54.8948, -54.8573, -54.8056, -54.7728, -54.7869, -54.7986] };
  WCL[1616] = struct {
   x = [-60.891, -60.8464, -60.8347, -60.8018, -60.7689, -60.7924, -60.8699, -60.9802, -61.067, -61.1562, -61.1984, -61.1984, -61.2102, -61.1444, -61.1327, -61.1773, -61.1562, -61.0999, -61.0881, -61.0459,
        -60.945, -60.891],
   y = [-51.8205, -51.7993, -51.8275, -51.8275, -51.8416, -51.9495, -51.9918, -51.9636, -51.9096, -51.8744, -51.8416, -51.7853, -51.7172, -51.6914, -51.7172, -51.7806, -51.8486, -51.8744, -51.8345, -51.7993,
        -51.7993, -51.8205] };
  WCL[1617] = struct {
   x = [-59.9757, -60.1071, -60.1306, -60.1752, -60.1635, -60.1635, -60.2738, -60.3277, -60.4169, -60.5272, -60.5601, -60.5812, -60.6915, -60.715, -60.6704, -60.7032, -60.8253, -60.891, -60.8581, -60.6915,
        -60.6915, -60.6915, -60.5812, -60.438, -60.3395, -60.4826, -60.4709, -60.3512, -60.3066, -60.2409, -60.1635, -60.0203],
   y = [-51.9965, -52.0458, -52.0106, -52.0176, -52.0528, -52.1209, -52.1748, -52.1678, -52.2077, -52.2476, -52.2077, -52.1467, -52.1326, -52.1467, -52.1936, -52.1795, -52.1068, -52.0645, -52.0458, -52.0458,
        -51.9965, -51.9777, -51.9965, -51.9965, -51.9566, -51.9495, -51.9237, -51.8416, -51.8557, -51.8604, -51.7454, -51.7313] };
  WCL[1618] = struct {
   x = [-59.964, -60.0297, -60.1071, -60.1846, -60.2292, -60.3606, -60.4052, -60.3395, -60.3277, -60.3723, -60.4826, -60.5061, -60.4287, -60.3184, -60.1846, -60.0414],
   y = [-51.7242, -51.6703, -51.6421, -51.5811, -51.5529, -51.56, -51.5342, -51.5271, -51.499, -51.4637, -51.4497, -51.3816, -51.4168, -51.4567, -51.4849, -51.506] };
  WCL[1619] = struct {
   x = [-68.7411, -68.659, -68.7435, -68.9101, -69.1049, -69.2739, -69.3584, -69.4123, -69.6376, -69.6916, -69.7479, -69.9451],
   y = [-52.6583, -52.6372, -52.5503, -52.6184, -52.6536, -52.6372, -52.5339, -52.4658, -52.5175, -52.6372, -52.7381, -52.7709] };
  WCL[1620] = struct {
   x = [-69.9991, -69.7761, -69.5532, -69.4123, -69.3302, -69.3842, -69.6376, -69.8606, -69.9169],
   y = [-53.3952, -53.3436, -53.3271, -53.3436, -53.428, -53.5618, -53.6111, -53.644, -53.6932] };
  WCL[1621] = struct {
   x = [-69.9991, -69.7761, -69.5813, -69.4123, -69.3584, -69.525, -69.5532, -69.6916, -69.8606, -69.8888],
   y = [-54.1556, -54.2049, -54.2377, -54.287, -54.3691, -54.3691, -54.4184, -54.3034, -54.287, -54.402] };
  WCL[1622] = struct {
   x = [-69.9991, -69.9709, -69.8043, -69.5532, -69.2739, -69.0228, -68.9101, -68.7435, -68.7975, -68.6027, -68.5863],
   y = [-54.8878, -54.8549, -54.8549, -54.8549, -54.8878, -55.0004, -54.9511, -54.9511, -54.8878, -54.8878, -54.8948] };
  WCL[1623] = struct {
   x = [-74.2985, -74.3078, -74.3853, -74.4627, -74.5519, -74.4627, -74.3407, -74.221, -74.099, -73.9887, -73.8901, -73.8572, -74.0661, -74.1224, -74.0333, -73.9347, -73.8338, -73.7798, -73.7141, -73.7024,
        -73.8784, -74.0872, -74.1318, -74.0779, -73.9441, -73.7469, -73.6366, -73.5709, -73.5029, -73.5263, -73.5381, -73.47, -73.3832, -73.2823, -73.2612, -73.2823, -73.3926, -73.3715, -73.2823,
        -73.2494, -73.1954, -73.1391, -73.0405, -72.9748, -73.0288, -73.0851, -73.0405, -72.8974, -72.7542, -72.6557, -72.6439, -72.8317, -73.0194, -73.2072, -73.118, -72.9091, -72.6439, -72.5782,
        -72.5008, -72.4797, -72.489, -72.6228, -72.6439, -72.6228, -72.5665, -72.4562, -72.4351, -72.5008, -72.6111, -72.7096, -72.7542, -72.7542, -72.6322, -72.489, -72.4562, -72.4562, -72.3459,
        -72.1042, -71.9047, -71.6958, -71.5081, -71.4095, -71.2992, -71.0997, -71.1232, -71.21, -71.3438, -71.5527, -71.785, -71.8953, -71.961, -72.1136, -72.0924, -72.137, -72.2919, -72.3013,
        -72.2145, -72.015, -71.8507, -71.6512, -71.4963, -71.3321, -71.2335, -71.1326, -71.0786, -71.0223, -70.9894, -70.9683, -70.9566, -70.912, -70.858, -70.8369, -70.8369, -70.8463, -70.8369,
        -70.8017, -70.6702, -70.5482, -70.4285, -70.2854, -70.1962, -70.0648],
   y = [-49.9266, -50.0111, -50.0603, -50.1307, -50.1824, -50.295, -50.3795, -50.4288, -50.4429, -50.4499, -50.5274, -50.5696, -50.5274, -50.5978, -50.7151, -50.7785, -50.7785, -50.701, -50.7292, -50.863,
        -50.9029, -50.9029, -50.9873, -51.1188, -51.215, -51.3464, -51.4239, -51.5342, -51.6421, -51.7524, -51.9167, -52.0458, -52.0786, -52.0857, -51.9495, -51.7923, -51.6233, -51.6562, -51.7806,
        -51.8815, -51.9707, -52.0246, -51.9777, -51.8744, -51.7993, -51.7102, -51.6984, -51.7853, -51.8275, -51.8064, -51.7242, -51.6351, -51.5529, -51.4637, -51.4708, -51.5388, -51.6914, -51.7383,
        -51.8064, -51.9026, -51.9707, -51.9965, -52.0035, -52.0176, -52.0645, -52.1467, -52.2828, -52.3907, -52.4236, -52.4635, -52.5104, -52.5785, -52.5644, -52.5456, -52.6043, -52.6794, -52.6184,
        -52.5574, -52.5574, -52.5644, -52.5973, -52.7193, -52.8132, -52.8859, -52.9657, -53.0925, -53.1652, -53.2426, -53.2638, -53.3295, -53.4022, -53.4492, -53.3811, -53.3037, -53.3483, -53.4609,
        -53.5665, -53.6768, -53.7238, -53.7824, -53.7895, -53.827, -53.8458, -53.8411, -53.8082, -53.7355, -53.651, -53.4679, -53.2825, -53.191, -53.0784, -53.0127, -52.9117, -52.7921, -52.7662,
        -52.7592, -52.7381, -52.6911, -52.7052, -52.6794, -52.6254, -52.5785] };
  WCL[1624] = struct {
   x = [-74.4416, -74.5519, -74.7279],
   y = [-49.9195, -50.0322, -50.0463] };
  WCL[1625] = struct {
   x = [-75.378, -75.3897, -75.3897, -75.3569, -75.2794, -75.1809, -75.0471, -74.8828, -74.85, -74.9039, -74.9368, -74.9603, -75.1363, -75.1903, -75.324, -75.378],
   y = [-50.3021, -50.2669, -50.2528, -50.1894, -50.2035, -50.2387, -50.1824, -50.1542, -50.2246, -50.288, -50.3443, -50.4851, -50.4288, -50.3584, -50.4147, -50.3021] };
  WCL[1626] = struct {
   x = [-74.4416, -74.573, -74.573, -74.573, -74.6294, -74.6951, -74.6951, -74.6294, -74.5284, -74.3642, -74.2539, -74.1882, -74.1553, -74.1882, -74.2985, -74.3736, -74.4416],
   y = [-50.6869, -50.6658, -50.6447, -50.5978, -50.5274, -50.4288, -50.3936, -50.3795, -50.471, -50.5555, -50.6377, -50.7433, -50.8278, -50.8348, -50.7644, -50.7221, -50.6869] };
  WCL[1627] = struct {
   x = [-74.4416, -74.4956, -74.4745, -74.6294, -74.749, -74.8265, -74.8382, -74.8265, -74.8265, -74.8148, -74.7045, -74.5848, -74.4839, -74.4416, -74.4088],
   y = [-51.0014, -51.0836, -51.1892, -51.154, -51.0155, -50.8958, -50.7221, -50.6588, -50.6588, -50.6799, -50.7081, -50.7362, -50.7714, -50.87, -50.9662] };
  WCL[1628] = struct {
   x = [-74.8828, -74.8711, -74.8593, -74.8265, -74.749, -74.6951, -74.6294, -74.5942, -74.573, -74.6294, -74.749, -74.8382, -74.8593, -74.8711, -74.8828],
   y = [-51.4309, -51.3957, -51.3347, -51.3065, -51.2643, -51.2103, -51.1821, -51.3276, -51.3886, -51.4168, -51.4426, -51.4239, -51.4168, -51.4168, -51.4309] };
  WCL[1629] = struct {
   x = [-74.8711, -74.9696, -74.9696, -74.9696, -74.9603, -74.9368, -74.8828, -74.85, -74.7936, -74.7819, -74.8148, -74.85, -74.8711],
   y = [-52.1396, -52.1326, -52.0645, -51.9707, -51.9026, -51.8345, -51.7923, -51.6914, -51.7383, -51.8885, -52.0387, -52.1396, -52.1396] };
  WCL[1630] = struct {
   x = [-74.0544, -73.9769, -73.923, -73.8666, -73.9676, -74.1107, -74.1647, -74.1647, -74.1647, -74.1647, -74.1224, -74.0544],
   y = [-51.6773, -51.7665, -51.8205, -51.8815, -51.8674, -51.7923, -51.7102, -51.6703, -51.6562, -51.6022, -51.567, -51.6773] };
  WCL[1631] = struct {
   x = [-73.7141, -73.7024, -73.6906, -73.6695, -73.6578, -73.6578, -73.7235, -73.8338, -73.8901, -73.8901, -73.8572, -73.8127, -73.7141],
   y = [-51.5342, -51.5342, -51.567, -51.6022, -51.6703, -51.7313, -51.6703, -51.5952, -51.4919, -51.4239, -51.4497, -51.5201, -51.5342] };
  WCL[1632] = struct {
   x = [-73.5381, -73.5146, -73.5146, -73.5029, -73.4606, -73.3715, -73.2729, -73.1954, -73.0851, -73.0077, -72.9866, -72.9514, -72.8528, -72.8199, -72.7871, -72.8411, -72.8645, -72.7988, -72.6768, -72.6322,
        -72.6111, -72.6885, -72.7988, -72.9185, -72.8411, -72.7542, -72.6557, -72.6228, -72.5548, -72.4445, -72.3694, -72.2356, -72.137, -72.0056, -71.7733, -71.6066, -71.5527, -71.4869, -71.4869,
        -71.6301, -71.7615, -71.9047, -72.0385, -72.1816, -72.2591, -72.2473, -72.2919, -72.3787, -72.489, -72.4562, -72.5219, -72.5219, -72.4679, -72.6228, -72.7542, -72.9091, -73.0851, -73.118,
        -72.9748, -72.7988, -72.6979, -72.6885, -72.7871, -72.8763, -72.9185, -72.996, -73.0851, -73.2729, -73.24, -73.24, -73.416, -73.4606, -73.4489, -73.4818, -73.5921, -73.5709, -73.5709, -73.5381,
        -73.5592, -73.5381, -73.5381],
   y = [-52.2218, -52.2147, -52.2077, -52.1795, -52.1678, -52.2077, -52.2218, -52.1678, -52.1396, -52.1795, -52.2828, -52.3368, -52.2898, -52.2147, -52.1537, -52.1068, -52.0246, -52.0387, -52.0927, -52.1936,
        -52.3297, -52.3907, -52.4705, -52.6442, -52.6653, -52.6911, -52.7592, -52.8249, -52.8718, -52.9188, -52.8578, -52.7921, -52.7123, -52.7052, -52.7334, -52.7193, -52.7263, -52.8061, -52.9,
        -52.9657, -53.0314, -53.1042, -53.1511, -53.1323, -53.1441, -53.2098, -53.238, -53.2051, -53.2497, -53.3154, -53.3483, -53.4421, -53.5078, -53.475, -53.4022, -53.3224, -53.2309, -53.191,
        -53.1769, -53.2309, -53.191, -53.1253, -53.0784, -52.9587, -53.0127, -53.0572, -53.0572, -53.0455, -52.9845, -52.9329, -52.893, -52.839, -52.7592, -52.7451, -52.7662, -52.6911, -52.5855,
        -52.4776, -52.3696, -52.2898, -52.2218] };
  WCL[1633] = struct {
   x = [-73.7235, -73.6578, -73.6695, -73.6812, -73.7024, -73.7681, -73.8127, -73.8572, -73.8784, -73.8784, -73.8572, -73.8127, -73.8127, -73.8127, -73.8572, -73.9018, -73.9441, -73.9441, -73.8666, -73.7681,
        -73.7235],
   y = [-52.4494, -52.4776, -52.5386, -52.5926, -52.6442, -52.6911, -52.7123, -52.7052, -52.6864, -52.6724, -52.6512, -52.6254, -52.5973, -52.5855, -52.6114, -52.6114, -52.5574, -52.4917, -52.4705, -52.4565,
        -52.4494] };
  WCL[1634] = struct {
   x = [-74.6716, -74.6951, -74.6716, -74.6059, -74.5402, -74.4745, -74.4181, -74.2985, -74.2093, -74.1318, -74.0544, -73.9347, -73.8338, -73.7235, -73.6038, -73.5381, -73.5029, -73.4606, -73.416, -73.3715,
        -73.294, -73.2166, -73.1297, -73.1063, -73.2072, -73.3057, -73.4043, -73.5029, -73.5709, -73.7235, -73.8127, -73.8338, -73.8127, -73.8666, -74.0121, -74.0779, -74.0872, -74.2093, -74.3407,
        -74.4745, -74.5613, -74.6176, -74.6505, -74.6716],
   y = [-52.7592, -52.7522, -52.7522, -52.778, -52.8179, -52.8718, -52.9117, -52.9446, -52.9775, -52.9986, -53.0197, -53.0526, -53.0713, -53.1042, -53.137, -53.1652, -53.2426, -53.2426, -53.2098, -53.2168,
        -53.2708, -53.3037, -53.3436, -53.3882, -53.3811, -53.3882, -53.3553, -53.3365, -53.2708, -53.184, -53.1183, -53.1652, -53.2051, -53.2168, -53.191, -53.1722, -53.1582, -53.1183, -53.0572,
        -53.0197, -52.9399, -52.8578, -52.7991, -52.7592] };
  WCL[1635] = struct {
   x = [-74.6294, -74.6294, -74.6176, -74.5942, -74.573, -74.573, -74.573, -74.573, -74.5848, -74.6294, -74.7045, -74.7397, -74.749, -74.749, -74.7397, -74.6951, -74.6622, -74.6622, -74.6833, -74.6622,
        -74.6294],
   y = [-51.7313, -51.7172, -51.7594, -51.7735, -51.7993, -51.8416, -51.9026, -51.9495, -52.0246, -52.0927, -52.1209, -52.0927, -52.0317, -51.9707, -51.9167, -51.8885, -51.8604, -51.7993, -51.7454, -51.7242,
        -51.7313] };
  WCL[1636] = struct {
   x = [-73.6132, -73.5921, -73.5263, -73.4372, -73.416, -73.3597, -73.24, -73.1626, -73.0851, -73.0851, -73.0523, -72.9748, -72.9302, -72.8645, -72.8317, -72.7754, -72.6885, -72.6439, -72.5993, -72.489,
        -72.4351, -72.4351, -72.3342, -72.2356, -72.2027, -72.2145, -72.2145, -72.3459, -72.3576, -72.3694, -72.4233, -72.5454, -72.5993, -72.6228, -72.5993, -72.6651, -72.8082, -72.9091, -72.9748,
        -72.8974, -72.8082, -72.9091, -73.0288, -73.1391, -73.24, -73.294, -73.2823, -73.294, -73.2729, -73.2729, -73.3503, -73.4489, -73.47, -73.4818, -73.5146, -73.5263, -73.6038, -73.6812, -73.6812,
        -73.6132],
   y = [-53.3952, -53.3882, -53.3811, -53.4022, -53.4609, -53.4679, -53.475, -53.5196, -53.5266, -53.4492, -53.4421, -53.4492, -53.475, -53.5266, -53.475, -53.5196, -53.6322, -53.6181, -53.5923, -53.6064,
        -53.644, -53.6839, -53.7238, -53.7566, -53.7942, -53.82, -53.8528, -53.8599, -53.9514, -54.0359, -54.0805, -54.0617, -54.0218, -54.0547, -54.1321, -54.1509, -54.1204, -54.1251, -54.0992,
        -54.0734, -54.0617, -54.0359, -54.0218, -54.0288, -54.0148, -53.9702, -53.9185, -53.8669, -53.82, -53.7942, -53.7895, -53.7613, -53.7167, -53.6768, -53.6252, -53.5923, -53.5337, -53.4539,
        -53.428, -53.3952] };
  WCL[1637] = struct {
   x = [-71.8836, -71.8601, -71.8507, -71.7944, -71.7287, -71.6841, -71.7075, -71.7075, -71.663, -71.6184, -71.5527, -71.4306, -71.386, -71.3086, -71.2663, -71.2663, -71.2218, -71.1654, -71.0903, -71.0223,
        -71.0011, -71.0997, -71.2218, -71.2663, -71.2757, -71.3321, -71.4963, -71.663, -71.7169, -71.6841, -71.7075, -71.8507, -71.8953, -71.9939, -72.0924, -72.1042, -72.1042, -72.137, -72.1581,
        -72.1816, -72.0924, -71.9939, -71.9282, -71.8836],
   y = [-53.8599, -53.8599, -53.8857, -53.9256, -53.9326, -53.9842, -54.0101, -54.0617, -54.0945, -54.0288, -53.9889, -53.996, -54.0429, -54.0288, -54.0945, -54.1321, -54.1978, -54.1579, -54.165, -54.2166,
        -54.287, -54.3574, -54.3574, -54.3316, -54.3058, -54.2612, -54.2424, -54.2283, -54.2753, -54.3011, -54.3128, -54.2799, -54.2354, -54.1908, -54.172, -54.1204, -54.1063, -54.0547, -54.003,
        -53.9514, -53.9584, -53.9256, -53.8669, -53.8599] };
  WCL[1638] = struct {
   x = [-70.5482, -70.5154, -70.5154, -70.4825, -70.4731, -70.4731, -70.4496, -70.4285, -70.3957, -70.3722, -70.3839, -70.4285, -70.4942, -70.5717, -70.6374, -70.6257, -70.5717, -70.506, -70.3839, -70.3182,
        -70.3511, -70.4614, -70.5482, -70.6491, -70.7477, -70.8251, -70.8463, -70.8369, -70.7923, -70.7266, -70.6491, -70.6163, -70.6257, -70.6702, -70.6585, -70.5811, -70.5482],
   y = [-53.5923, -53.5923, -53.5994, -53.6111, -53.6979, -53.7824, -53.8669, -53.9373, -53.9889, -54.0547, -54.0945, -54.0875, -54.0547, -54.003, -53.9842, -54.0875, -54.1767, -54.2166, -54.2236, -54.2799,
        -54.287, -54.2753, -54.2494, -54.1837, -54.1321, -54.0547, -53.9702, -53.9045, -53.8786, -53.8739, -53.8739, -53.8458, -53.7824, -53.7308, -53.7026, -53.644, -53.5923] };
  WCL[1639] = struct {
   x = [-71.0011, -71.0129, -71.0129, -70.9894, -70.9237, -70.8463, -70.8017, -70.7688, -70.7477, -70.7477, -70.7805, -70.8369, -70.8908, -70.9683, -71.0011],
   y = [-55.1107, -55.0966, -55.0779, -55.0333, -55.0028, -54.9887, -54.984, -55.0333, -55.0849, -55.1295, -55.153, -55.16, -55.1483, -55.1225, -55.1107] };
  WCL[1640] = struct {
   x = [-70.3628, -70.3957, -70.3957, -70.4496, -70.5388, -70.6045, -70.6163, -70.5717, -70.5154, -70.4379, -70.3628, -70.3276, -70.3393, -70.3628],
   y = [-55.2727, -55.2609, -55.268, -55.268, -55.2234, -55.1671, -55.1295, -55.1154, -55.1225, -55.1483, -55.1412, -55.1929, -55.2351, -55.2727] };
  WCL[1641] = struct {
   x = [-69.9451, -69.9991, -70.1399, -70.4191, -70.2243, -70.1962, -70.1117, -70.168, -70.2783, -70.3628, -70.4473, -70.4473, -70.4755, -70.4191, -70.3065, -70.168, -69.9991],
   y = [-52.7709, -52.8061, -52.7381, -52.7545, -52.7897, -52.8578, -52.9423, -53.0244, -53.0432, -52.9915, -53.008, -53.1441, -53.2614, -53.3436, -53.3788, -53.4445, -53.3952] };
  WCL[1642] = struct {
   x = [-69.9169, -70.0835, -70.1399, -70.1117, -70.0554, -69.9991],
   y = [-53.6932, -53.6932, -53.7777, -53.9091, -54.1063, -54.1556] };
  WCL[1643] = struct {
   x = [-69.8888, -70.0272, -70.1117, -70.2243, -70.1962, -70.0272, -70.1399, -70.1962, -70.3065, -70.5036, -70.6702, -70.9495, -70.9495, -70.7547, -70.6984, -70.7547, -70.8111, -70.8392, -70.9214, -71.0903,
        -71.0903, -71.3133, -71.3696, -71.2851, -71.2851, -71.3133, -71.1467, -71.0903, -70.9495, -70.8111, -70.9214, -70.9495, -70.8111, -70.8111, -70.8932, -70.7266, -70.6139, -70.4191, -70.2525,
        -69.9991],
   y = [-54.402, -54.3034, -54.2541, -54.3198, -54.3691, -54.402, -54.4184, -54.4654, -54.3856, -54.3034, -54.2541, -54.1063, -54.2049, -54.287, -54.3691, -54.3856, -54.3691, -54.3691, -54.4184, -54.4184,
        -54.4818, -54.4982, -54.5146, -54.5639, -54.6461, -54.6766, -54.6766, -54.6461, -54.6601, -54.6766, -54.7258, -54.7587, -54.7587, -54.7916, -54.8221, -54.8385, -54.8056, -54.8549, -54.8878,
        -54.8878] };
  WCL[1644] = struct {
   x = [166.168, 166.191, 166.179, 166.179, 166.179, 166.201, 166.224, 166.179, 166.147, 166.093, 166.069, 166.102, 166.102, 165.992, 165.914, 165.926, 165.947, 165.914, 165.947, 166.036, 166.102, 166.114,
        166.135, 166.168],
   y = [-50.6001, -50.64, -50.6963, -50.7386, -50.809, -50.8371, -50.8723, -50.8442, -50.8512, -50.8653, -50.8864, -50.9005, -50.9545, -50.9474, -50.9122, -50.9005, -50.8442, -50.816, -50.7527, -50.6823,
        -50.6001, -50.572, -50.572, -50.6001] };
  WCL[1645] = struct {
   x = [-1.50901, -1.41514, -1.41045, -1.46208, -1.0725, -0.908225, -0.880063, -0.988018, -0.988018, -0.959856, -1.11944, -1.2274, -1.02087, -0.776802, -0.420083, -0.32621, -0.213562, -0.138463, -0.147851],
   y = [-70.053, -69.985, -69.938, -69.877, -69.7456, -69.6963, -69.6564, -69.6376, -69.4921, -69.3701, -69.2715, -69.1542, -69.1119, -69.1425, -69.0814, -69.0674, -69.1213, -69.1237, -69.1307] };
  WCL[1646] = struct {
   x = [0.021122, -0.058671],
   y = [-69.0955, -69.2011] };
  WCL[1647] = struct {
   x = [-45.6976, -45.6201, -45.5004, -45.3995, -45.3127, -45.2352, -45.1695, -45.1132, -45.0592, -45.0592, -45.1461, -45.2352, -45.3338, -45.4324, -45.5755, -45.6647, -45.6858, -45.775, -45.8853, -45.9745,
        -45.9839, -45.9416, -45.6976],
   y = [-60.5483, -60.5483, -60.5483, -60.553, -60.5953, -60.6446, -60.6657, -60.7103, -60.7431, -60.7643, -60.7431, -60.7197, -60.6985, -60.6938, -60.7056, -60.6446, -60.6188, -60.6399, -60.6516, -60.6352,
        -60.6188, -60.5859, -60.5483] };
  WCL[1648] = struct {
   x = [-54.6461, -54.6508, -54.693, -54.876, -55.1459, -55.3407, -55.4721, -55.5613, -55.444, -55.1248, -54.8009, -54.6719, -54.6461],
   y = [-61.1444, -61.1703, -61.1914, -61.2031, -61.2547, -61.3228, -61.2946, -61.1632, -61.1186, -61.1022, -61.1022, -61.1233, -61.1444] };
  WCL[1649] = struct {
   x = [-58.9431, -58.9807, -58.9243, -58.8305, -58.7366, -58.6427, -58.5441, -58.4784, -58.3658, -58.1123, -57.8589, -57.6993, -57.6383, -57.8025, -57.9387, -58.1029, -58.2015, -58.347, -58.4221, -58.6005,
        -58.7037, -58.8023, -58.9431],
   y = [-62.2733, -62.2686, -62.224, -62.1841, -62.1536, -62.1066, -62.0902, -62.0691, -62.0456, -62.0104, -61.9846, -61.9963, -62.0761, -62.1043, -62.1231, -62.1512, -62.224, -62.2662, -62.2475, -62.3014,
        -62.3085, -62.2733, -62.2733] };
  WCL[1650] = struct {
   x = [-60.0273, -59.9194],
   y = [-62.6417, -62.7192] };
  WCL[1651] = struct {
   x = [-58.3165, -58.2977, -58.1945, -58.171, -58.1945, -58.2508, -58.4667, -58.4667, -58.4385, -58.3353, -58.2977, -58.0912, -57.88, -57.8941, -57.6829, -57.528, -57.3919, -57.2041, -57.2417, -57.4482,
        -57.4951, -57.8471, -58.1053, -58.2039, -58.3165],
   y = [-64.4488, -64.4488, -64.4089, -64.3643, -64.3009, -64.2446, -64.2352, -64.2352, -64.1414, -64.1108, -64.0076, -63.9536, -63.9137, -64.0217, -64.031, -64.0451, -64.1038, -64.1977, -64.2915, -64.3573,
        -64.4347, -64.4769, -64.5028, -64.491, -64.4488] };
  WCL[1652] = struct {
   x = [-57.4904, -57.5327, -57.3825, -57.1102, -56.9506, -57.0398, -57.1994],
   y = [-64.6013, -64.5872, -64.5262, -64.4816, -64.4535, -64.5098, -64.5966] };
  WCL[1653] = struct {
   x = [-60.0626, -59.7997, -59.7622, -59.5838, -59.4993, -59.3538, -59.2271, -59.0393, -58.8469, -58.4057, -58.1851, -57.9128, -57.6688, -57.3543, -57.1384, -57.0117, -56.9319, -57.0257, -57.1947, -57.2182,
        -57.3168, -57.467, -57.6641, -57.7345, -57.7486, -57.8941, -58.1288, -58.3822, -58.5324, -58.6779, -58.8046, -58.9079, -58.8422, -58.9079, -58.9689, -59.091, -59.2552, -59.4665, -59.4993,
        -59.565, -59.8232],
   y = [-63.9536, -63.8902, -63.8879, -63.9207, -63.8832, -63.8081, -63.7424, -63.74, -63.5875, -63.5171, -63.5101, -63.442, -63.3998, -63.3599, -63.3434, -63.3833, -63.4936, -63.6438, -63.6814, -63.6157,
        -63.557, -63.5523, -63.625, -63.6861, -63.7048, -63.7142, -63.7799, -63.8644, -63.9818, -64.1061, -64.247, -64.4535, -64.606, -64.7515, -64.9792, -65.1622, -65.3195, -65.465, -65.5659,
        -65.6785, -65.9508] };
  WCL[1654] = struct {
   x = [-57.2839, -57.4341, -57.4904],
   y = [-64.6365, -64.6389, -64.6013] };
  WCL[1655] = struct {
   x = [-57.7392, -57.6782, -57.3637, -57.2229, -57.2041, -57.3919, -57.4857, -57.6078, -57.7392],
   y = [-63.9372, -63.8996, -63.8762, -63.9067, -63.9489, -64.0029, -64.0099, -63.9771, -63.9372] };
  WCL[1656] = struct {
   x = [-60.7173, -60.7173, -60.7455, -60.8347, -60.8159, -60.8675, -60.7877, -60.6047, -60.5718, -60.3794, -60.2667, -60.3981, -60.6047],
   y = [-67.042, -67.1242, -67.3424, -67.6475, -67.9573, -68.2342, -68.5159, -68.828, -69.0204, -69.173, -69.3865, -69.5696, -69.8301] };
  WCL[1657] = struct {
   x = [-61.121, -61.1022, -61.0834, -61.1069, -61.0459, -60.9098, -60.8581, -60.7783, -60.6751, -60.4967, -60.2339, -60.0978, -60.0743, -60.0273],
   y = [-62.7638, -62.7778, -62.7708, -62.7473, -62.7098, -62.6887, -62.6699, -62.6253, -62.6112, -62.6089, -62.6042, -62.5807, -62.5948, -62.6417] };
  WCL[1658] = struct {
   x = [-59.917, -60.0954, -60.2315, -60.3629, -60.424, -60.5554, -60.7197, -60.8417, -61.0107, -61.1186, -61.1186],
   y = [-62.7192, -62.7872, -62.8107, -62.7919, -62.7614, -62.7286, -62.7286, -62.7426, -62.7661, -62.7802, -62.7638] };
  WCL[1659] = struct {
   x = [-62.691, -62.6488, -62.6441, -62.6582, -62.4798, -62.5549, -62.5455, -62.3484, -62.2545, -62.0949, -62.0714, -62.0949, -62.2263, -62.3296, -62.3578, -62.4892, -62.5784, -62.691],
   y = [-64.5661, -64.5826, -64.5075, -64.4253, -64.3455, -64.2563, -64.1836, -64.1367, -64.1414, -64.193, -64.2164, -64.2915, -64.3784, -64.4558, -64.491, -64.5122, -64.538, -64.5661] };
  WCL[1660] = struct {
   x = [-64.2962, -64.2024, -64.1554, -63.9865, -63.7846, -63.6391, -63.5593, -63.4279, -63.2777, -63.2496, -63.3716, -63.2589, -63.1651, -63.1228, -63.2402, -63.2167, -63.0853, -62.9351, -62.8553, -63.1369,
        -63.3153, -63.4842, -63.6908, -63.8597, -64.0334, -64.1132, -64.2352, -64.2962],
   y = [-64.7562, -64.7281, -64.6365, -64.5919, -64.5474, -64.4699, -64.369, -64.3408, -64.3361, -64.3925, -64.4723, -64.4676, -64.4652, -64.5051, -64.5826, -64.6389, -64.6201, -64.5826, -64.6131, -64.6623,
        -64.7421, -64.7633, -64.8008, -64.8454, -64.8477, -64.8266, -64.7867, -64.7562] };
  WCL[1661] = struct {
   x = [-66.2488, -66.2253, -66.3098, -66.1878, -66.1315, -66.0939, -65.9954, -65.9156, -65.6996, -65.6386, -65.5917, -65.7513, -65.8686, -65.9672, -66.108, -66.2253, -66.2488],
   y = [-65.9414, -65.9484, -65.8921, -65.8569, -65.8076, -65.7513, -65.6762, -65.6175, -65.5917, -65.6316, -65.7231, -65.7489, -65.8146, -65.8897, -65.9414, -65.9484, -65.9414] };
  WCL[1662] = struct {
   x = [-69.0767, -68.9923, -69.0955, -69.1378, -69.302, -69.3396, -69.288, -69.2316, -69.0721, -68.8702, -68.7106, -68.5698, -68.5041, -68.3539, -68.138, -67.9503, -67.8235, -67.9127, -68.0019, -68.0301,
        -67.9268, -67.9503, -68.1052, -68.3023, -68.199, -68.0254, -68.138, -68.2319, -68.429, -68.4947, -68.6308, -68.7904, -68.9688, -69.0767],
   y = [-67.8259, -67.8353, -67.8095, -67.7578, -67.6898, -67.6053, -67.5396, -67.448, -67.3589, -67.2626, -67.1406, -66.9998, -66.9505, -66.8871, -66.8731, -66.8238, -66.8167, -66.9153, -67.0233, -67.1312,
        -67.2392, -67.3049, -67.3542, -67.4527, -67.495, -67.5114, -67.5959, -67.6851, -67.6921, -67.6992, -67.7649, -67.8282, -67.8353, -67.8259] };
  WCL[1663] = struct {
   x = [-70.0648, -69.9333, -69.7221, -69.5578, -69.5532, -69.5109, -69.3748, -69.3513],
   y = [-69.295, -69.3631, -69.3678, -69.3959, -69.5297, -69.5813, -69.6329, -69.8348] };
  WCL[1664] = struct {
   x = [-68.8374, -68.6825, -68.5276, -68.659, -68.7388, -68.3821, -68.2741, -68.4337, -68.1943, -67.6593, -67.4903, -67.4856, -67.3213, -67.0913, -67.1101, -67.0632, -66.927, -67.1523, -66.9693, -66.7252,
        -67.1054, -67.1242, -67.035, -66.8379, -66.683, -66.8801, -66.8801, -66.8379, -66.7722, -66.7252, -66.5938, -66.5985, -66.7252, -66.9646, -67.0538, -67.2931, -67.5325, -67.5513, -67.5935,
        -67.678, -67.6123, -67.5466, -67.5419, -67.3917, -67.2603, -67.1101, -67.0491, -66.8801, -66.8379, -66.8566, -66.8144, -66.6877, -66.6407, -66.4483, -66.3216, -66.3263, -66.3309, -66.3309,
        -66.0352, -65.58, -65.4016, -65.5706, -65.763, -65.7255, -65.5893, -65.4673, -65.4251, -65.303, -65.2279, -65.0871, -65.0871, -65.073, -65.0918, -64.951, -64.8477, -64.6225, -64.3972, -64.3831,
        -64.5943, -64.599, -64.4206, -64.3033, -64.5333, -64.4769, -64.2751, -64.0733, -64.139, -64.0404, -63.8621, -63.6649, -63.7307, -63.9747, -64.031, -64.1202, -64.0545, -63.8058, -63.2988,
        -63.0923, -62.9046, -63.1486, -63.1346, -62.7872, -62.4023, -62.5244, -62.4728, -62.3554, -62.3554, -62.2381, -62.0691, -61.9471, -61.6701, -61.5059, -61.5059, -61.2899, -61.1022, -60.952,
        -60.8253, -60.4404, -60.0649],
   y = [-70.0014, -69.9028, -69.7714, -69.6259, -69.4663, -69.4522, -69.3372, -69.2363, -69.1964, -69.2082, -69.1166, -68.9829, -68.9242, -68.8327, -68.6496, -68.5675, -68.4666, -68.3586, -68.3328, -68.3164,
        -68.2248, -68.1497, -68.0911, -67.9995, -67.8752, -67.8587, -67.8329, -67.8001, -67.7743, -67.7578, -67.6992, -67.6334, -67.6264, -67.6593, -67.6499, -67.6334, -67.6311, -67.549, -67.4832,
        -67.448, -67.4246, -67.3659, -67.2744, -67.1852, -67.1007, -67.0608, -67.028, -67.028, -67.1523, -67.2673, -67.3424, -67.3589, -67.3589, -67.3847, -67.333, -67.1336, -66.9177, -66.8778,
        -66.8754, -66.7722, -66.6313, -66.5492, -66.4342, -66.3356, -66.2418, -66.2253, -66.216, -66.1737, -66.1244, -66.1057, -66.0634, -66.0141, -66.0141, -66.0047, -66.0446, -66.0728, -66.0704,
        -66.0047, -65.9156, -65.8757, -65.8545, -65.8287, -65.7255, -65.7067, -65.7208, -65.7325, -65.6598, -65.641, -65.6457, -65.6339, -65.5776, -65.5236, -65.411, -65.3124, -65.1951, -65.1481,
        -65.1364, -65.1974, -65.1505, -65.0848, -65.0496, -64.9158, -64.9041, -64.8665, -64.7163, -64.6459, -64.7374, -64.7492, -64.7187, -64.6647, -64.6131, -64.5661, -64.4746, -64.4253, -64.3995,
        -64.2376, -64.1414, -64.0146, -63.9536] };
  WCL[1665] = struct {
   x = [-59.8255, -60.2104, -60.4451, -60.6328, -60.722, -60.7549, -60.7173],
   y = [-65.9508, -66.169, -66.3615, -66.5961, -66.8074, -66.9928, -67.042] };
  WCL[1666] = struct {
   x = [-75.9107, -75.8215, -75.8309, -76.0422, -76.0375, -75.7324, -75.3897, -75.3006, -75.1644, -74.8171, -74.7279, -74.6012, -74.3994, -74.3947, -74.5683, -74.6434],
   y = [-70.0108, -69.9732, -69.8888, -69.8183, -69.7573, -69.7245, -69.7245, -69.6283, -69.593, -69.6283, -69.7151, -69.7644, -69.7996, -69.8582, -69.9263, -69.9897] };
  WCL[1667] = struct {
   x = [-74.573, -74.3618, -74.3524, -74.1741, -73.8361, -73.4935, -73.301, -73.0053, -72.705, -72.6392, -72.6486, -72.8035, -72.7988, -72.658, -72.5407, -72.4703, -72.2872, -71.9962, -71.785, -71.6817,
        -71.6958, -71.9023, -72.0338, -71.7803, -71.5738, -71.3485, -71.0575, -70.6116, -70.3534, -70.2455, -70.1469, -70.0389, -70.0389, -70.0906, -70.0671],
   y = [-70.0014, -69.8934, -69.87, -69.8512, -69.7996, -69.7972, -69.7949, -69.7902, -69.7503, -69.701, -69.6517, -69.6048, -69.579, -69.5086, -69.4452, -69.4382, -69.471, -69.5461, -69.5696, -69.532,
        -69.3889, -69.2645, -69.2269, -69.1589, -69.0721, -69.004, -68.9453, -68.9219, -68.9195, -68.9735, -69.058, -69.1354, -69.119, -69.1683, -69.295] };
  WCL[1668] = struct {
   x = [-90.4846, -90.4283, -90.3578, -90.4048, -90.4611, -90.6301, -90.8413, -90.9211, -90.8413, -90.7427, -90.6207, -90.4846],
   y = [-68.8045, -68.8608, -68.9265, -69.0275, -69.0744, -69.065, -69.0322, -68.9547, -68.8679, -68.8092, -68.7928, -68.8045] };
  WCL[1669] = struct {
   x = [159.754, 160.064, 160.265],
   y = [-69.6752, -69.8559, -69.9779] };
  WCL[1670] = struct {
   x = [164.133, 164.166, 164.278, 164.419, 164.518, 164.457, 164.335, 164.171, 164.128, 164.189],
   y = [-67.549, -67.5044, -67.4527, -67.4668, -67.5912, -67.6475, -67.6639, -67.6241, -67.5771, -67.5231] };
  WCL[1671] = struct {
   x = [149.864, 150.094, 150.587, 150.995, 151.521, 152.239, 152.943, 153.394, 154.201, 154.492, 154.403, 154.375, 154.398, 154.685, 154.91, 155.398, 156.384, 156.923, 156.83, 156.82, 157.125, 157.29,
        157.703, 158.153, 158.303, 158.721, 159.13, 159.754],
   y = [-68.5018, -68.4478, -68.3844, -68.4149, -68.4877, -68.5229, -68.483, -68.4079, -68.5628, -68.6567, -68.6848, -68.7505, -68.9101, -69.065, -68.9805, -69.0251, -69.0979, -69.1917, -69.0744, -68.95,
        -69.0885, -69.2199, -69.2269, -69.2692, -69.3232, -69.3513, -69.4804, -69.6752] };
  WCL[1672] = struct {
   x = [139.923, 140.636, 141.364, 141.641, 141.829, 142.115, 142.373, 142.763, 143.143, 143.528, 143.828, 144.161, 144.452, 144.973, 145.546, 145.828, 145.504, 145.147, 145.274, 145.56, 145.987, 146.302,
        146.438, 146.621, 147.025, 147.494, 148.02, 148.156, 148.057, 147.738, 147.841, 148.203, 148.63, 148.864, 148.982, 149.137, 149.31, 149.432, 149.512, 149.662, 149.737, 149.827, 149.864],
   y = [-66.69, -66.7675, -66.8144, -66.7933, -66.8942, -67.0115, -67.0233, -66.9552, -66.8989, -67.042, -67.0796, -67.089, -67.1453, -67.0021, -67.0491, -67.2345, -67.3917, -67.5466, -67.6733, -67.6452,
        -67.7297, -67.8188, -67.8658, -67.9104, -68.0207, -68.0418, -68.1075, -68.2671, -68.3328, -68.3938, -68.4501, -68.4361, -68.429, -68.3774, -68.368, -68.4595, -68.5065, -68.4525, -68.5346,
        -68.5252, -68.4947, -68.5112, -68.5018] };
  WCL[1673] = struct {
   x = [129.895, 130.256, 130.425, 130.932, 131.294, 131.871, 132.134, 132.514, 132.843, 133.533, 133.955, 134.457, 134.518, 134.509, 134.706, 134.908, 135.218, 135.415, 135.588, 135.551, 135.448, 135.208,
        134.79, 135.039, 135.227, 135.692, 135.893, 136.03, 136.583, 137.194, 137.616, 138.832, 138.912, 139.409, 139.921],
   y = [-66.0071, -65.9531, -66.0963, -66.1972, -66.1925, -66.2113, -66.1925, -66.1831, -66.1362, -66.1315, -66.1901, -66.0517, -65.8639, -65.6973, -65.4767, -65.3124, -65.1082, -65.0918, -65.2491, -65.4884,
        -65.7724, -65.9907, -66.0775, -66.0634, -66.0775, -66.169, -66.1901, -66.3028, -66.3591, -66.3615, -66.4765, -66.6337, -66.6454, -66.6689, -66.69] };
  WCL[1674] = struct {
   x = [119.604, 120.017, 120.44, 120.932, 121.294, 121.651, 121.036, 120.829, 120.853, 121.181, 121.369, 121.524, 121.824, 122.167, 122.411, 122.524, 122.81, 123.007, 123.303, 123.636, 123.866, 123.875,
        123.749, 123.542, 123.364, 123.279, 123.434, 123.828, 124.312, 124.861, 125.387, 125.856, 126.18, 126.462, 126.832, 126.926, 127.1, 127.377, 127.494, 127.813, 128.198, 128.705, 129.123,
        129.442, 129.569, 129.536, 129.719, 129.714, 129.818, 129.897],
   y = [-66.9693, -66.859, -66.7886, -66.805, -66.6337, -66.4624, -66.3708, -66.1244, -65.8311, -65.7208, -65.7419, -66.0094, -66.2582, -66.3967, -66.4811, -66.6008, -66.7252, -66.7862, -66.7909, -66.6736,
        -66.6243, -66.5985, -66.5985, -66.5727, -66.5516, -66.4553, -66.3356, -66.2394, -66.2019, -66.2582, -66.2981, -66.3497, -66.3615, -66.4459, -66.5656, -66.5821, -66.5562, -66.683, -66.8332,
        -67.0233, -67.035, -67.0068, -67.0373, -66.9669, -66.7534, -66.6196, -66.4342, -66.3403, -66.216, -66.0071] };
  WCL[1675] = struct {
   x = [109.959, 110.212, 110.386, 110.475, 110.71, 110.573, 110.522, 110.926, 111.465, 111.897, 112.122, 112.404, 112.869, 113.249, 113.573, 113.878, 114.249, 114.502, 114.488, 114.704, 114.849, 115.032,
        115.173, 115.314, 115.91, 116.361, 116.666, 116.699, 116.943, 117.332, 117.717, 117.741, 118.097, 118.51, 119.017, 119.604],
   y = [-66.6525, -66.6501, -66.5703, -66.5304, -66.4835, -66.399, -66.2911, -66.1643, -66.0235, -65.9367, -65.8968, -65.8405, -65.7841, -65.7536, -65.9343, -66.0493, -66.169, -66.3075, -66.4412, -66.5375,
        -66.5445, -66.4929, -66.4835, -66.4319, -66.3873, -66.5187, -66.6196, -66.7017, -66.9036, -66.9435, -66.974, -66.981, -66.9693, -66.9224, -66.92, -66.9693] };
  WCL[1676] = struct {
   x = [99.8226, 100.198, 100.428, 100.775, 101.043, 101.264, 101.573, 101.972, 102.273, 102.583, 102.686, 102.906, 102.874, 103.066, 103.31, 103.625, 104.047, 104.526, 104.803, 105.009, 105.305, 105,
        104.413, 104.141, 104.131, 104.038, 104.413, 104.845, 105.375, 105.831, 106.525, 107.009, 107.553, 107.877, 108.069, 108.22, 108.454, 108.703, 108.722, 109.05, 109.149, 109.28, 109.646,
        109.914, 109.956],
   y = [-65.8217, -65.7043, -65.5823, -65.4697, -65.4603, -65.5401, -65.6504, -65.7583, -65.7677, -65.7161, -65.5307, -65.3547, -65.1998, -65.1646, -65.2913, -65.4063, -65.4016, -65.4579, -65.6339, -65.8452,
        -66.054, -66.054, -65.9273, -65.885, -65.9179, -65.9954, -66.0845, -66.1362, -66.2089, -66.3333, -66.4506, -66.4976, -66.5774, -66.6149, -66.5985, -66.6619, -66.8519, -66.9669, -66.9763,
        -66.9364, -66.8566, -66.7111, -66.6454, -66.6501, -66.6525] };
  WCL[1677] = struct {
   x = [89.5458, 90.2076, 90.8929, 91.8504, 92.2353, 92.6531, 93.0473, 93.634, 94.0471, 94.3099],
   y = [-66.92, -66.8613, -66.7487, -66.6243, -66.6196, -66.6736, -66.6783, -66.6665, -66.6008, -66.6266] };
  WCL[1678] = struct {
   x = [94.3005, 94.4554, 94.737, 94.8356, 94.9999, 95.1407, 95.1736, 95.1548, 95.3003, 95.5209, 95.6148, 95.5584, 95.4974, 95.5631, 95.7556, 96.0935, 96.5629, 97.0135, 97.3186, 97.7128, 98.1165, 98.4403,
        98.8628, 99.3321, 99.5246, 99.825],
   y = [-66.636, -66.6126, -66.5375, -66.4178, -66.4319, -66.4929, -66.4811, -66.3474, -66.3239, -66.2652, -66.0939, -65.8616, -65.6199, -65.3687, -65.1951, -65.1528, -65.2115, -65.3265, -65.4509, -65.5987,
        -65.7372, -65.817, -65.8686, -65.9179, -65.9461, -65.8217] };
  WCL[1679] = struct {
   x = [79.7689, 80.3603, 81.0832, 81.2944, 81.5009, 81.4211, 81.7919, 82.1721, 82.2847, 82.3692, 81.8201, 81.3507, 81.3366, 81.7403, 82.0829, 82.1815, 82.4959, 82.8292, 83.2657, 83.552, 83.843, 84.3875,
        84.7114, 84.9226, 85.284, 85.7252, 86.1523, 86.265, 86.8845, 87.6496, 88.2739, 88.5179, 88.9216, 89.5458],
   y = [-68.0348, -67.9995, -67.9432, -67.8963, -67.6898, -67.4856, -67.3964, -67.2345, -67.1148, -66.9599, -66.7135, -66.3849, -66.0493, -65.8616, -65.8193, -66.0259, -66.1502, -66.345, -66.5046, -66.5703,
        -66.5023, -66.4811, -66.5234, -66.6619, -66.5516, -66.392, -66.3286, -66.3145, -66.1643, -66.1291, -66.1925, -66.3568, -66.7017, -66.92] };
  WCL[1680] = struct {
   x = [69.6822, 70.1, 70.2408, 70.4755, 70.5975, 70.7665, 71.3579, 71.616, 72.0385, 72.1417, 72.475, 73.2494, 73.4935, 73.4278, 73.301, 73.4559, 73.8831, 74.6294, 74.3477, 74.6341, 74.7889, 74.2257,
        73.9394, 73.808, 74.1928, 74.5308, 74.9908, 75.371, 75.7183, 75.4461, 75.5352, 75.845, 76.0422, 76.5444, 77.1405, 77.5817, 77.929, 77.7225, 77.8868, 78.2435, 78.7457, 78.9898, 79.7689],
   y = [-68.0535, -68.3258, -68.4408, -68.544, -68.6308, -68.6872, -68.7341, -68.7177, -68.5534, -68.4102, -68.3938, -68.5464, -68.6097, -68.6895, -68.9007, -68.9242, -68.7928, -68.9876, -69.0462, -69.2668,
        -69.4382, -69.5907, -69.6611, -69.7972, -69.8183, -69.7526, -69.5907, -69.3114, -69.3443, -69.5226, -69.579, -69.5884, -69.525, -69.4992, -69.3114, -69.2387, -69.0486, -68.781, -68.6402,
        -68.4971, -68.3305, -68.199, -68.0348] };
  WCL[1681] = struct {
   x = [59.68, 60.0884, 60.8534, 61.1585, 61.3322, 61.2759, 61.3228, 61.3275, 61.5012, 61.9236, 62.3132, 62.5807, 62.914, 63.2895, 63.6649, 64.2282, 64.8947, 65.3265, 65.6034, 66.4107, 67.1242, 67.908,
        68.5511, 69.1612, 69.6259, 69.6822],
   y = [-67.5419, -67.5278, -67.556, -67.5818, -67.6123, -67.6522, -67.6898, -67.678, -67.6686, -67.6686, -67.7062, -67.7625, -67.7461, -67.6569, -67.6334, -67.7015, -67.8001, -67.8095, -67.8611, -67.8822,
        -67.9597, -67.962, -67.9902, -67.901, -67.8846, -68.0535] };
  WCL[1682] = struct {
   x = [49.9923, 50.2739, 50.4851, 50.2504, 50.1049, 50.3912, 50.3865, 50.0392],
   y = [-67.3495, -67.3025, -67.2814, -67.1969, -67.1054, -67.0303, -66.9364, -66.8871] };
  WCL[1683] = struct {
   x = [49.9031, 50.0627, 50.4663, 50.7996, 51.1281, 51.5694, 52.1795, 52.5456, 52.8226, 53.3858, 53.8364, 54.2447, 54.6531, 55.0239, 55.2492, 55.6153, 55.8688, 56.0847, 56.1034, 56.2583, 56.4367, 56.8028,
        57.2018, 57.2018, 57.0703, 56.723, 56.4695, 56.4508, 56.5728, 56.6244, 56.8122, 57.0328, 57.1736, 57.413, 57.6007, 57.8119, 58.0185, 58.1546, 58.3235, 58.5629, 58.8727, 58.7225, 58.4456,
        58.6568, 58.9149, 59.1825, 59.2529, 59.4031, 59.4453, 59.4876, 59.68],
   y = [-66.6712, -66.5938, -66.4014, -66.3685, -66.3356, -66.1761, -66.0775, -66.0704, -66.0352, -66.0165, -66.0352, -66.0704, -66.0751, -66.0798, -66.1479, -66.2582, -66.4014, -66.4436, -66.4577, -66.5351,
        -66.568, -66.6384, -66.7534, -66.8426, -66.8871, -66.859, -66.9059, -66.9998, -67.1171, -67.2227, -67.1828, -67.2063, -67.1993, -67.157, -67.1453, -67.1922, -67.2063, -67.2392, -67.3143,
        -67.3307, -67.3988, -67.4175, -67.4128, -67.4973, -67.549, -67.5067, -67.4152, -67.4246, -67.4739, -67.5231, -67.5419] };
  WCL[1684] = struct {
   x = [39.9408, 40.2787, 40.5979, 41.0297, 41.3958, 41.8652, 42.1844, 42.3768, 42.6068, 42.8039, 43.3108, 43.9163, 44.2684, 44.8081, 45.024, 45.1273, 45.6201, 45.7891, 46.2491, 46.6903, 46.9156, 47.1878,
        47.0188, 46.9344, 47.0705, 47.1268, 47.3286, 47.4694, 47.7792, 48.0702, 48.38, 48.089, 48.6898, 49.0418, 48.9949, 48.7414, 48.6147, 48.3706, 48.1641, 48.2533, 48.4176, 48.6429, 48.9714,
        49.1685, 49.4924, 49.7599, 49.6614, 49.9899],
   y = [-68.9664, -68.8233, -68.7717, -68.6848, -68.5816, -68.49, -68.4478, -68.3187, -68.2108, -68.0958, -67.9831, -67.8986, -67.7578, -67.6874, -67.6898, -67.7461, -67.6968, -67.4809, -67.3448, -67.4433,
        -67.495, -67.5677, -67.5865, -67.6358, -67.7132, -67.8212, -67.8517, -67.8447, -67.8188, -67.7414, -67.6405, -67.4011, -67.5583, -67.5114, -67.4363, -67.4457, -67.4598, -67.3401, -67.2486,
        -67.1242, -67.0843, -67.0984, -67.0749, -67.1383, -67.204, -67.319, -67.3753, -67.3495] };
  WCL[1685] = struct {
   x = [50.0369, 49.9008],
   y = [-66.8871, -66.6712] };
  WCL[1686] = struct {
   x = [29.8588, 30.2624, 30.3657, 30.6614, 30.9055, 31.2199, 31.5907, 31.9052, 32.1493, 32.3652, 32.905, 33.0599, 33.0129, 33.1866, 32.9707, 32.8017, 33.1866, 33.6747, 33.9611, 34.2755, 34.5431, 35.0969,
        35.524, 35.5851, 35.8104, 36.3079, 36.5567, 36.613, 36.8195, 37.2654, 37.5752],
   y = [-69.3537, -69.3443, -69.2387, -69.1331, -69.1565, -69.2316, -69.2058, -69.3325, -69.4452, -69.5508, -69.6048, -69.5273, -69.3349, -69.2997, -69.2011, -69.0439, -68.767, -68.7177, -68.774, -68.9219,
        -69.0556, -69.2269, -69.4076, -69.471, -69.5907, -69.6705, -69.7456, -69.7573, -69.7315, -69.694, -69.8113] };
  WCL[1687] = struct {
   x = [38.1502, 38.535, 38.6289, 38.8636, 39.4785, 39.3423, 39.225, 39.3517, 39.3611, 39.3142, 39.4221, 39.3705, 39.5113, 39.6991, 39.7178, 39.9431],
   y = [-70.1328, -69.9404, -69.8371, -69.7526, -69.6752, -69.6024, -69.5532, -69.5578, -69.417, -69.3279, -69.2199, -69.1307, -69.0016, -68.9923, -68.9899, -68.9664] };
  WCL[1688] = struct {
   x = [19.5257, 20.042],
   y = [-69.9944, -69.9873] };
  WCL[1689] = struct {
   x = [27.5073, 27.634, 27.986, 28.5821, 28.8919, 28.9998, 29.0796, 29.3988, 29.8588],
   y = [-70.053, -69.9568, -69.9052, -69.8442, -69.809, -69.7644, -69.6259, -69.4898, -69.3537] };
  WCL[1690] = struct {
   x = [9.86844, 10.0515],
   y = [-69.9756, -69.9638] };
  WCL[1691] = struct {
   x = [12.6002, 12.863, 12.6659, 12.8865, 13.3136, 13.5013, 13.4779, 13.9519, 14.2899, 14.5715, 14.717, 14.7592, 14.839, 15.0925, 15.2802, 15.3037, 15.3272, 15.6182, 16.3457, 16.9747, 17.5614, 17.8195,
        17.9744, 18.3734],
   y = [-70.1047, -69.985, -69.931, -69.8418, -69.7831, -69.6517, -69.5485, -69.4734, -69.417, -69.3818, -69.3724, -69.3912, -69.3701, -69.4593, -69.6447, -69.7127, -69.8418, -69.8465, -69.7925, -69.809,
        -69.8137, -69.8559, -69.9028, -69.9897] };
  WCL[1692] = struct {
   x = [19.1478, 19.5233],
   y = [-70.0436, -69.9944] };
  WCL[1693] = struct {
   x = [-0.150197, 0.023468],
   y = [-69.1307, -69.0955] };
  WCL[1694] = struct {
   x = [-0.053977, 0.04459, 0.208868, 0.359066, 0.673542, 1.01618],
   y = [-69.2011, -69.3771, -69.5203, -69.6517, -69.8512, -69.9803] };
  WCL[1695] = struct {
   x = [1.34239, 1.7836],
   y = [-70.0624, -69.9333] };
  WCL[1696] = struct {
   x = [6.8246, 7.17193],
   y = [-70.0108, -69.992] };
  WCL[1697] = struct {
   x = [7.54507, 7.90649, 8.23974],
   y = [-70.0037, -69.9944, -69.9897] };
  WCL[1698] = struct {
   x = [9.23949, 9.10337, 9.29581, 9.86844],
   y = [-70.0436, -69.9897, -69.9615, -69.9756] };
  WCL[1699] = struct {
   x = [-10.1501, -9.88252, -9.65253, -9.53988, -9.28173, -9.02358, -8.70441, -8.37116, -8.29606, -7.84078, -7.58262, -7.36672, -7.01469, -6.75185, -6.49369, -6.2637, -6.01494, -5.57373, -5.18885, -4.86499,
        -4.55051, -4.31114, -3.97789, -3.6728, -3.3724, -3.04854, -2.80447, -2.59795, -2.21776, -1.8188, -1.50901],
   y = [-70.9143, -70.7805, -70.6609, -70.635, -70.5787, -70.5858, -70.5717, -70.6022, -70.7054, -70.7172, -70.6233, -70.52, -70.5013, -70.4848, -70.4848, -70.52, -70.5083, -70.4661, -70.459, -70.4708,
        -70.4496, -70.4074, -70.3534, -70.3581, -70.3605, -70.33, -70.3159, -70.2666, -70.1516, -70.1234, -70.053] };
  WCL[1700] = struct {
   x = [-20.1053, -19.7674, -19.5609, -19.4529, -19.237, -18.9695, -18.9554, -18.9319, -18.7348, -18.4954, -17.9838, -17.5004, -17.1014, -16.8338, -16.5428, -16.4208, -16.2612, -15.9139, -15.529, -15.1254,
        -14.5903, -14.1397, -13.7736, -13.5154, -13.3183, -12.9803, -12.9287, -13.0461, -12.9569, -12.7785, -12.5016, -12.2951, -12.3045, -12.1402, -11.8539, -11.5769, -11.3704, -11.009, -10.6711,
        -10.3754, -10.1501],
   y = [-73.4794, -73.3151, -73.1649, -73.0664, -72.9983, -72.9115, -72.8692, -72.7613, -72.6345, -72.6299, -72.6862, -72.6885, -72.6134, -72.5712, -72.5336, -72.4867, -72.428, -72.2872, -72.2098, -72.1347,
        -72.0619, -72.0197, -72.0103, -72.0596, -72.1042, -72.076, -72.0267, -71.9892, -71.8953, -71.8319, -71.7709, -71.6512, -71.5245, -71.4072, -71.3508, -71.2382, -71.142, -71.0598, -71.0317,
        -70.9941, -70.9143] };
  WCL[1701] = struct {
   x = [-30.2366, -29.8424, -29.2791, -28.791, -28.4249, -27.9555, -27.458, -27.0449, -26.7915, -26.6507, -26.4911, -26.8431, -26.6835, -26.1109, -25.4491, -25.0313, -24.7638, -24.8577, -24.4259, -24.0457,
        -23.9847, -23.642, -23.0225, -22.6376, -22.4404, -22.309, -21.9758, -21.558, -21.2342, -21.0652, -20.7272, -20.4362, -20.2391, -20.1077],
   y = [-76.718, -76.596, -76.4763, -76.3754, -76.3073, -76.2416, -76.1877, -76.1055, -76.021, -75.9154, -75.791, -75.676, -75.5657, -75.432, -75.3545, -75.2935, -75.256, -75.2653, -75.087, -74.857, -74.749,
        -74.5355, -74.3548, -74.2562, -74.214, -74.2022, -74.221, -74.2163, -74.1248, -74.0403, -73.9089, -73.7634, -73.5944, -73.4794] };
  WCL[1702] = struct {
   x = [-31.9756, -31.572, -31.2153, -31.0087, -30.7083, -30.239],
   y = [-77.117, -77.1264, -76.9738, -76.9105, -76.8283, -76.718] };
  WCL[1703] = struct {
   x = [-40.7997, -39.4949, -38.6876, -37.5142, -36.1905, -35.2612, -35.4302, -34.9702, -33.9751, -32.8487, -31.9756],
   y = [-77.8257, -77.9947, -77.9642, -77.9853, -77.9853, -77.929, -77.8703, -77.6075, -77.4221, -77.1968, -77.117] };
  WCL[1704] = struct {
   x = [-50.3232, -49.713, -49.0277, -47.967, -47.6572, -46.7466, -45.9205, -44.9067, -44.9442, -44.4279, -43.5737, -42.9072, -42.4284, -41.9215, -40.795],
   y = [-77.6591, -77.7366, -77.6896, -77.7225, -77.7553, -77.9126, -78.0581, -78.1238, -78.1261, -78.2505, -78.1848, -77.9666, -77.8187, -77.6849, -77.8257] };
  WCL[1705] = struct {
   x = [-45.998, -46.1106, -46.5049, -47.4718, -48.2603, -48.326, -47.622, -46.9273, -46.2233, -45.8853, -46.0918],
   y = [-77.6802, -77.6896, -77.7295, -77.7131, -77.645, -77.5512, -77.4714, -77.5347, -77.6263, -77.699, -77.7248] };
  WCL[1706] = struct {
   x = [-60.2386, -59.9616],
   y = [-72.1605, -72.3036] };
  WCL[1707] = struct {
   x = [-60.2104, -59.6988, -59.3186, -59.6565],
   y = [-72.8904, -73.0147, -73.3433, -73.3808] };
  WCL[1708] = struct {
   x = [-60.5906, -59.6378, -58.7648, -58.0936, -57.6336, -57.1689, -56.7652, -56.5118, -56.1316, -55.8265, -55.5402, -55.2257, -54.7094, -54.4513, -54.1321, -54.0852, -54.226, -54.4607, -54.4466, -54.2072],
   y = [-75.2818, -75.4484, -75.608, -75.7488, -76.0046, -76.2416, -76.4223, -76.4575, -76.6735, -76.7884, -76.8542, -76.8964, -77.0466, -77.0771, -77.1405, -77.2414, -77.2695, -77.3094, -77.4573, -77.6591] };
  WCL[1709] = struct {
   x = [-60.6023, -60.8135, -60.8699, -60.6915, -60.5929, -60.5929, -60.4193, -60.3442, -60.2972, -60.5554, -60.5413, -60.2268, -60.1752, -60.4991, -60.2362],
   y = [-69.8301, -70.0976, -70.3957, -70.5576, -70.6984, -70.9261, -71.0645, -71.3579, -71.6348, -71.8061, -71.8554, -71.8272, -71.961, -72.0408, -72.1605] };
  WCL[1710] = struct {
   x = [-59.9616, -60.2057, -60.4592, -60.2104],
   y = [-72.3036, -72.4726, -72.6932, -72.8904] };
  WCL[1711] = struct {
   x = [-59.6542, -60.0344, -60.546, -60.7197, -60.8558, -60.8558, -60.5789, -61.0154, -60.9778, -60.6774, -60.5695, -60.9966, -61.4988, -61.757, -61.4284, -60.5882],
   y = [-73.3808, -73.3949, -73.4536, -73.6202, -73.8197, -73.9816, -74.1436, -74.1412, -74.1553, -74.2844, -74.3923, -74.5144, -74.7232, -74.9086, -75.0987, -75.2818] };
  WCL[1712] = struct {
   x = [-69.3537, -69.2082, -69.0157, -68.8374],
   y = [-69.8348, -70.0131, -70.0577, -70.0014] };
  WCL[1713] = struct {
   x = [-74.9368, -75.148, -75.2466, -75.4484, -75.7441, -75.9084],
   y = [-70.0436, -70.0249, -70.0202, -70.0272, -70.0296, -70.0108] };
  WCL[1714] = struct {
   x = [-74.6434, -74.6669, -74.7936, -74.9391],
   y = [-69.9897, -70.0483, -70.0601, -70.0436] };
  WCL[1715] = struct {
   x = [-80.0365, -79.4826, -78.971, -78.6753, -78.3327, -77.9666, -77.99, -78.267, -78.6471, -78.7082, -78.2763, -77.76, -77.5441, -77.9712, -78.0135, -77.6755, -77.1311, -76.8682, -76.8025, -76.6336,
        -76.6289, -76.474, -76.3707, -76.0985, -76.1454, -75.8497, -75.7981, -75.7981, -75.432, -75.0799, -74.6294, -74.3384, -73.9347, -73.7047, -73.8408, -73.761, -73.4653, -73.1884, -72.9068,
        -72.658, -72.7613, -72.3154, -71.9305, -71.8038, -72.3764, -72.6439, -72.5594, -72.2168, -72.1323, -72.3248, -72.5219, -72.5454, -72.428, -72.4327, -72.719, -73.1133, -73.2447, -73.6155,
        -73.9018, -74.5402, -74.9251, -75.2959, -75.5399, -75.5118, -75.1973, -74.9063, -74.8687, -74.3384, -74.0426, -73.6625, -73.3198, -73.2588, -73.6859, -73.8408, -74.2445, -74.7702, -75.2724,
        -75.4085, -75.155, -75.0799, -74.6528, -74.5683, -74.39, -74.3337, -74.221, -74.3853, -74.6012, -74.6106, -74.6294, -74.573],
   y = [-73.0851, -73.0194, -73.0851, -73.2518, -73.4442, -73.4818, -73.3433, -73.1814, -73.0452, -72.8857, -72.8528, -72.8223, -72.7566, -72.6557, -72.5923, -72.5594, -72.5571, -72.5876, -72.5735, -72.5829,
        -72.6697, -72.7026, -72.6721, -72.6956, -72.8293, -73.0312, -73.0523, -72.9396, -72.9044, -72.895, -72.9232, -72.9772, -72.9842, -73.0687, -73.1743, -73.2705, -73.2635, -73.3151, -73.5029,
        -73.4841, -73.355, -73.3409, -73.3503, -73.2095, -73.0922, -73.0171, -72.9443, -72.8692, -72.6932, -72.5853, -72.4937, -72.4844, -72.4632, -72.3107, -72.1746, -72.1206, -72.1746, -72.2027,
        -72.1581, -72.0995, -71.9821, -71.8366, -71.6723, -71.494, -71.5081, -71.4541, -71.3719, -71.3696, -71.3555, -71.2593, -71.1889, -71.0903, -70.9706, -70.9073, -70.8815, -70.9073, -70.9284,
        -70.8862, -70.7876, -70.6726, -70.7524, -70.7313, -70.7313, -70.7242, -70.5693, -70.4004, -70.1962, -70.0929, -70.0765, -70.0014] };
  WCL[1716] = struct {
   x = [-90.3015, -89.9636, -89.4379, -88.8183, -88.1988, -87.7576, -87.3117, -86.5935, -86.1758, -85.8003, -85.2934, -84.4861, -83.8806, -82.9982, -82.2753, -81.7215, -81.2521, -80.4964, -80.0365],
   y = [-72.4961, -72.4937, -72.5055, -72.5172, -72.5289, -72.5618, -72.5407, -72.6369, -72.7894, -72.881, -72.8857, -72.8669, -72.8857, -72.9279, -72.9889, -73.0335, -73.0077, -73.0617, -73.0851] };
  WCL[1717] = struct {
   x = [-100.055, -99.8672],
   y = [-72.015, -72.0314] };
  WCL[1718] = struct {
   x = [-100.046, -99.7733, -99.5246, -99.4636, -99.2758, -99.1209, -98.9566, -98.6422, -98.4403, -98.2808, -98.1775, -98.2432, -98.4685, -98.4216, -98.1353, -97.9616, -98.0038, -97.802, -97.5626, -97.5392,
        -97.3749, -97.3279, -97.0933, -96.9571, -96.7976, -96.5441, -96.6286, -96.5817, -96.3376, -95.9292, -95.8917, -96.2062, -96.1264, -95.7603, -95.6945, -96.117, -95.9198, -95.5397, -95.305,
        -95.197, -95.1219, -95.0187, -94.906, -94.5258, -94.0799, -93.6763, -93.282, -92.8877, -92.409, -91.9537, -91.4702, -90.8601, -90.3015],
   y = [-72.0877, -72.1417, -72.1793, -72.1816, -72.1699, -72.2684, -72.2215, -72.2708, -72.3482, -72.3177, -72.2192, -72.137, -72.1558, -72.0361, -72.0197, -72.0924, -72.2121, -72.2403, -72.2825, -72.1206,
        -72.1018, -72.1769, -72.2192, -72.0995, -72.0478, -71.9962, -72.1089, -72.252, -72.2074, -72.2098, -72.2919, -72.374, -72.4374, -72.4116, -72.4797, -72.5336, -72.6345, -72.7261, -72.7566,
        -72.7143, -72.7073, -72.705, -72.7331, -72.7167, -72.6604, -72.705, -72.7331, -72.7448, -72.7378, -72.6909, -72.5993, -72.543, -72.4961] };
  WCL[1719] = struct {
   x = [-104.904, -104.88, -104.702, -104.458, -104.298, -104.162, -104.195, -104.331, -104.537, -104.796, -104.908, -104.904],
   y = [-73.3081, -73.2588, -73.2471, -73.2799, -73.3456, -73.4184, -73.4864, -73.5099, -73.4747, -73.4184, -73.3503, -73.3081] };
  WCL[1720] = struct {
   x = [-110.029, -109.682, -109.686, -109.194, -108.33, -107.602, -107.011, -107.208, -107.494, -107.598, -107.227, -106.612, -105.608, -104.796, -104.894, -105.251, -105.481, -105.223, -104.932, -104.631,
        -103.711, -102.843, -102.106, -101.505, -101.181, -101.35, -101.637, -101.89, -101.792, -101.928, -101.989, -101.928, -102.022, -102.134, -102.317, -101.998, -101.557, -101.144, -100.876,
        -100.975, -101.289, -101.594, -102.017, -102.331, -102.679, -103.012, -103.125, -103.275, -103.416, -103.495, -103.491, -103.228, -103.115, -102.899, -102.801, -102.838, -102.871, -102.998,
        -103.336, -103.589, -103.585, -103.153, -102.651, -102.28, -102.214, -101.787, -101.214, -100.712, -100.055],
   y = [-74.3501, -74.4487, -74.5871, -74.7185, -74.911, -75.0401, -75.0706, -74.8711, -74.6059, -74.3923, -74.2891, -74.2609, -74.275, -74.3524, -74.4393, -74.6575, -74.8148, -74.8781, -74.8101, -74.8054,
        -74.8945, -75.0166, -75.1433, -75.1152, -75.0095, -74.8335, -74.7209, -74.6528, -74.5261, -74.3149, -74.2187, -74.1389, -73.984, -73.8385, -73.7493, -73.7399, -73.754, -73.7516, -73.7,
        -73.5897, -73.4982, -73.4583, -73.4982, -73.5076, -73.5005, -73.4301, -73.3761, -73.2917, -73.2189, -73.1509, -73.0781, -73.0171, -72.9584, -72.942, -72.8904, -72.8598, -72.7589, -72.6697,
        -72.5641, -72.4961, -72.4233, -72.4046, -72.4116, -72.3248, -72.2262, -72.1417, -72.0502, -72.0619, -72.015] };
  WCL[1721] = struct {
   x = [-99.8649, -100.043],
   y = [-72.0314, -72.0877] };
  WCL[1722] = struct {
   x = [-120.142, -119.808, -119.522, -118.954, -118.306, -117.588, -117.18, -116.612, -115.988, -115.424, -114.734, -114.148, -113.979, -114.148, -113.749, -113.448, -113.195, -112.622, -112.026, -111.693,
        -111.472, -111.031, -110.604, -110.027],
   y = [-73.9159, -73.9511, -73.9863, -74.038, -74.0591, -74.1271, -74.1553, -74.1037, -74.1342, -74.1811, -74.1365, -74.0121, -74.0614, -74.1623, -74.1412, -74.1905, -74.2891, -74.3266, -74.3712, -74.3407,
        -74.2327, -74.2679, -74.3313, -74.3501] };
  WCL[1723] = struct {
   x = [-130.224, -129.721, -129.069, -128.412],
   y = [-74.1928, -74.1999, -74.1553, -74.0966] };
  WCL[1724] = struct {
   x = [-127.504, -127.025, -126.41, -125.955, -125.668, -125.753, -125.513, -124.889, -124.467, -124.03, -123.805, -123.5, -123.03, -122.528, -121.881, -121.214, -120.763, -120.393, -120.247, -120.144],
   y = [-73.9887, -73.9136, -73.8033, -73.7188, -73.6202, -73.517, -73.3691, -73.3363, -73.3691, -73.47, -73.5686, -73.6249, -73.7, -73.7258, -73.7657, -73.808, -73.8526, -73.8925, -73.9042, -73.9159] };
  WCL[1725] = struct {
   x = [-140.475, -139.212, -137.513, -135.931, -135.316],
   y = [-75.1269, -74.911, -74.7444, -74.6739, -74.7373] };
  WCL[1726] = struct {
   x = [-135.319, -134.84, -134.516, -134.201, -133.892, -133.521, -133.047, -132.784, -132.333, -131.948, -131.446, -131.132, -130.63, -130.226],
   y = [-74.7373, -74.8241, -74.8077, -74.7655, -74.7397, -74.7068, -74.6341, -74.6035, -74.5402, -74.4181, -74.3078, -74.2492, -74.1999, -74.1928] };
  WCL[1727] = struct {
   x = [-150.134, -149.515, -148.717, -148.144],
   y = [-76.6148, -76.4622, -76.3332, -76.2604] };
  WCL[1728] = struct {
   x = [-145.802, -145.75, -146.187, -146.623, -146.956, -147.224, -146.98, -146.675, -146.248, -145.924, -145.802],
   y = [-75.3358, -75.4883, -75.5963, -75.6667, -75.6901, -75.608, -75.554, -75.4977, -75.4249, -75.3498, -75.3358] };
  WCL[1729] = struct {
   x = [-147.461, -147.198, -146.626, -145.828, -145.621, -146.09, -146.588, -146.635, -146.044, -145.044, -144.016, -142.932, -142.279, -141.768, -141.453, -140.472],
   y = [-76.3472, -76.4505, -76.5303, -76.5467, -76.4411, -76.2651, -76.0891, -75.9788, -75.8309, -75.6268, -75.4836, -75.3475, -75.2372, -75.1668, -75.1715, -75.1269] };
  WCL[1730] = struct {
   x = [-160.465, -159.629, -158.634, -158.569, -158.625, -158.709, -158.465, -157.348, -156.203, -154.936, -154.354, -153.856, -153.284, -152.702, -151.894, -151.415, -150.937, -150.542, -150.158, -150.158,
        -150.129],
   y = [-77.9759, -77.9008, -77.8234, -77.6521, -77.4995, -77.286, -77.2038, -77.1733, -77.1381, -77.117, -77.1616, -77.1639, -77.2766, -77.2602, -77.2555, -77.1076, -76.8847, -76.772, -76.718, -76.711,
        -76.6148] };
  WCL[1731] = struct {
   x = [-170.491, -169.467, -168.416, -166.989, -166.144, -164.999, -164.454, -164.051, -163.478, -162.408, -161.601, -160.465],
   y = [-78.565, -78.5603, -78.5697, -78.5462, -78.5298, -78.565, -78.4078, -78.2482, -78.2012, -78.1731, -78.0346, -77.9759] };
  WCL[1732] = struct {
   x = [179.939, 180],
   y = [-78.3303, -78.3359] };
  WCL[1733] = struct {
   x = [-180, -178.982, -177.667, -176.287, -175.067, -173.95, -173.865, -173.621, -171.838, -170.486],
   y = [-78.3359, -78.4289, -78.5134, -78.5063, -78.5345, -78.5157, -78.5204, -78.5087, -78.5744, -78.565] };
  WCL[1734] = struct {
   x = [169.988, 170.2, 170.378, 170.486, 170.275],
   y = [-71.4118, -71.5855, -71.7639, -71.9141, -72.0314] };
  WCL[1735] = struct {
   x = [169.911, 170.005, 170.662, 172.258, 173.525, 174.351, 175.299, 176.482, 177.44, 178.5, 179.937],
   y = [-77.53, -77.5136, -77.5629, -77.6755, -77.753, -77.8117, -77.8516, -77.9525, -78.0487, -78.1895, -78.3303] };
  WCL[1736] = struct {
   x = [160.263, 160.282, 160.225, 160.671, 161.019, 161.202, 161.46, 161.859, 162.112, 162.272, 162.666, 162.84, 162.976, 163.591, 164.088, 164.248, 164.525, 165.304, 165.863, 166.276, 166.168, 166.299,
        166.571, 166.89, 166.989, 167.458, 167.599, 168.125, 168.571, 169.021, 169.294, 169.589, 169.772, 169.88, 169.777, 169.988],
   y = [-69.9779, -70.0202, -70.1258, -70.3229, -70.4051, -70.3487, -70.4168, -70.4567, -70.5506, -70.7195, -70.8533, -70.8251, -70.7313, -70.7923, -70.7641, -70.7524, -70.7594, -70.7172, -70.628, -70.6585,
        -70.7477, -70.7594, -70.7524, -70.851, -71.0199, -70.9941, -71.0598, -71.1701, -71.3321, -71.5104, -71.6395, -71.6817, -71.67, -71.5714, -71.4377, -71.4118] };
  WCL[1737] = struct {
   x = [170.275, 169.974, 169.679, 169.904, 169.913, 169.852, 169.566, 169.331, 169.073, 168.796, 168.383, 168.055, 167.613, 167.022, 166.872, 166.618, 166.224, 165.572, 164.952, 164.896, 165.126, 165.332,
        165.431, 164.741, 164.145, 163.84, 163.422, 163.502, 163.872, 164.361, 164.229, 163.431, 162.69, 162.708, 162.877, 162.446, 162.502, 162.643, 162.952, 162.802, 162.539, 162.615, 162.554,
        162.539, 162.61, 162.446, 162.37, 162.483, 162.774, 162.737, 162.849, 162.957, 162.943, 163.182, 163.422, 163.488, 163.732, 163.83, 164.215, 164.365, 164.276, 163.971, 164.149, 164.314,
        164.328, 164.417, 164.539, 164.675, 164.708, 164.825, 164.933, 165.295, 165.557, 165.811, 166.013, 166.177],
   y = [-72.0314, -72.0807, -72.1417, -72.299, -72.4327, -72.6604, -72.8669, -73.0147, -73.125, -73.2447, -73.3574, -73.4466, -73.4911, -73.7235, -73.923, -74.1037, -74.275, -74.3102, -74.2656, -74.2961,
        -74.4111, -74.5613, -74.6669, -74.6528, -74.7537, -74.918, -75.1011, -75.263, -75.4719, -75.5493, -75.6315, -75.5775, -75.5329, -75.5939, -75.662, -75.838, -75.9624, -76.1219, -76.1571,
        -76.2956, -76.3472, -76.3989, -76.5772, -76.7157, -76.8753, -76.9363, -76.9621, -77.0067, -77.117, -77.2695, -77.3564, -77.462, -77.6004, -77.7178, -77.8117, -77.9548, -77.8821, -77.8821,
        -77.8563, -77.7858, -77.7366, -77.638, -77.4831, -77.3329, -77.2109, -77.1452, -77.1428, -77.2179, -77.2414, -77.2883, -77.3658, -77.4244, -77.4409, -77.4761, -77.6051, -77.53] };
  WCL[1738] = struct {
   x = [169.603, 169.373, 169.495, 169.603, 169.341, 169.373, 169.495, 169.603],
   y = [-73.3386, -73.3151, -73.3081, -73.3386, -73.3902, -73.3151, -73.3081, -73.3386] };
  WCL[1739] = struct {
   x = [37.5775, 37.9624, 38.1502],
   y = [-69.8113, -70.0202, -70.1328] };
  WCL[1740] = struct {
   x = [20.042, 20.5489, 21.084, 21.4266, 21.328, 20.9291, 21.0229, 21.7927, 22.4592, 23.0084, 23.3322, 23.933, 24.3789, 24.5197, 24.6418, 24.792, 24.8671, 24.8905, 25.2848, 25.5664, 25.7307, 26.0921,
        26.2799, 26.1484, 26.6178, 27.0825, 27.5096],
   y = [-69.9873, -70.0554, -70.0953, -70.1586, -70.1563, -70.168, -70.2361, -70.2361, -70.2079, -70.2337, -70.3933, -70.4403, -70.3393, -70.2619, -70.2267, -70.2173, -70.2431, -70.2502, -70.2032, -70.283,
        -70.1891, -70.1328, -70.0788, -70.0061, -70.0225, -70.0225, -70.053] };
  WCL[1741] = struct {
   x = [10.0515, 10.0844, 10.5256, 10.7649, 11.0043, 11.422, 11.929, 12.3279, 12.6002],
   y = [-69.9638, -70.0272, -70.0014, -70.0202, -70.0366, -70.0577, -70.0436, -70.0953, -70.1047] };
  WCL[1742] = struct {
   x = [18.3734, 18.7583, 19.1478],
   y = [-69.9897, -70.0178, -70.0436] };
  WCL[1743] = struct {
   x = [1.01383, 1.34239],
   y = [-69.9803, -70.0624] };
  WCL[1744] = struct {
   x = [1.78125, 2.20837, 2.45714, 2.81386, 3.22221, 3.63055, 4.06237, 4.67724, 5.0809, 5.11845, 5.447, 5.61598, 5.56435, 5.77556, 5.85066, 5.97269, 6.01494, 6.27778, 6.39982, 6.25901, 6.49369, 6.82225],
   y = [-69.9333, -70.0155, -70.0742, -70.1117, -70.1, -70.107, -70.0695, -70.0249, -70.0366, -70.1281, -70.1375, -70.2079, -70.2901, -70.2596, -70.1798, -70.1586, -70.2243, -70.2525, -70.2032, -70.1281,
        -70.0389, -70.0108] };
  WCL[1745] = struct {
   x = [7.17193, 7.54742],
   y = [-69.992, -70.0037] };
  WCL[1746] = struct {
   x = [8.24208, 8.83818, 8.93675, 9.1949, 9.24184],
   y = [-69.9897, -70.0084, -70.0765, -70.107, -70.0436] };
  WCL[1747] = struct {
   x = [-19.4435, -19.2746, -19.0399, -18.9085, -18.7207, -18.6456, -18.5611, -18.5893, -18.777, -18.9742, -19.2839, -19.4435],
   y = [81.7684, 81.7919, 81.7825, 81.745, 81.7191, 81.6863, 81.6652, 81.644, 81.6487, 81.6863, 81.7332, 81.7684] };
  WCL[1748] = struct {
   x = [-19.5984, -19.5045, -19.2886, -19.0633, -19.0915, -19.3919, -19.7298, -19.927, -19.9364, -19.8988, -19.7674, -19.5984],
   y = [80.2313, 80.2031, 80.1867, 80.1538, 80.1233, 80.0998, 80.0529, 80.02, 80.0998, 80.1773, 80.2125, 80.2313] };
  WCL[1749] = struct {
   x = [-20.0185, -19.9903, -19.812, -19.5961, -19.3332, -19.183, -18.9483, -18.9577, -18.7043, -18.4133, -18.2255, -18.2161, -18.3006, -18.188, -17.9815, -18.0472, -17.8125, -17.9439, -18.0659, -17.9721,
        -18.0659, -18.1035, -18.0753, -17.7655, -17.5684, -17.6998, -17.6435, -17.606, -17.4558, -17.3807, -17.2211, -17.0897, -16.977, -16.7423, -16.4232, -16.1791, -15.8787, -15.6346, -15.4563,
        -15.2028, -14.7616, -14.5551, -14.3767, -14.1514, -13.9449, -13.6914, -13.4755, -13.2315, -12.9592, -12.7433, -12.5462, -12.349, -12.2927, -12.5743, -12.8372, -13.1751, -13.269, -13.3723,
        -13.6914, -13.7478, -13.851, -14.0482, -14.0388, -14.1514, -14.2641, -14.4518, -14.6959, -14.9587, -15.1183, -15.1183, -15.0714, -14.8555, -14.7428, -14.9212, -15.3061, -15.6909, -15.9632,
        -16.0101, -16.0101, -16.0477, -16.3293, -16.7142, -17.0709, -17.39, -17.7186, -17.9251, -18.0941, -18.1504, -18.3194, -18.4884, -18.5916, -18.8357, -19.1455, -19.4647, -19.7369],
   y = [81.4633, 81.4962, 81.5244, 81.522, 81.5244, 81.522, 81.4586, 81.5103, 81.4868, 81.4493, 81.454, 81.4962, 81.5431, 81.5596, 81.5291, 81.5736, 81.576, 81.6112, 81.63, 81.6628, 81.6933, 81.7098,
        81.7098, 81.7051, 81.6769, 81.7121, 81.7637, 81.8248, 81.8154, 81.8529, 81.8341, 81.8459, 81.874, 81.8693, 81.8717, 81.8905, 81.8858, 81.8787, 81.8928, 81.9045, 81.8858, 81.8529, 81.8318,
        81.8154, 81.8083, 81.8107, 81.7966, 81.7661, 81.7403, 81.6839, 81.6511, 81.6159, 81.5689, 81.4962, 81.4422, 81.3882, 81.3319, 81.2967, 81.2756, 81.2216, 81.1911, 81.1653, 81.1301, 81.109,
        81.109, 81.1113, 81.1113, 81.116, 81.0808, 81.048, 81.0127, 80.9729, 80.9071, 80.8532, 80.8461, 80.832, 80.8062, 80.7945, 80.7663, 80.7311, 80.703, 80.7053, 80.7241, 80.7311, 80.7429, 80.7828,
        80.7546, 80.703, 80.6889, 80.6584, 80.6232, 80.6138, 80.6373, 80.6466, 80.6654] };
  WCL[1750] = struct {
   x = [-20.0185, -19.7087, -19.5022, -19.2675, -18.9296, -18.6386, -18.2067, -17.8406, -17.7186, -17.2492, -16.9488, -16.6297, -16.2636, -16.2448, -16.3762, -16.5734, -16.7048, -16.8268, -16.7048, -16.8268,
        -17.0615, -17.4558, -17.7749, -18.0096, -18.3663, -18.6573, -19.0422, -19.3332, -19.5304, -19.8965],
   y = [80.6279, 80.6067, 80.5809, 80.5598, 80.5528, 80.5387, 80.5434, 80.5669, 80.6044, 80.595, 80.5645, 80.5528, 80.5363, 80.4964, 80.4354, 80.3861, 80.3908, 80.3838, 80.3298, 80.2782, 80.2266, 80.2172,
        80.2242, 80.2242, 80.2125, 80.2242, 80.2313, 80.25, 80.2688, 80.2711] };
  WCL[1751] = struct {
   x = [-19.7744, -19.6524, -19.4271, -19.1079, -18.9108, -18.6949, -18.4696, -18.2725, -18.019, -17.7092, -17.4745, -17.5121],
   y = [79.9801, 80.006, 80.0341, 80.0623, 80.0529, 80.067, 80.1022, 80.1116, 80.1256, 80.1139, 80.0717, 80.013] };
  WCL[1752] = struct {
   x = [-21.1215, -20.9995, -20.9056, -20.8493, -20.6709, -20.5395, -20.3142, -20.2297, -20.0514, -20.042, -20.2391, -20.5395, -20.793, -20.9995, -21.206, -21.375, -21.2999, -21.1215],
   y = [82.1533, 82.1392, 82.1087, 82.0806, 82.043, 82.0102, 81.9773, 81.9257, 81.8764, 81.8341, 81.8318, 81.8553, 81.921, 81.9796, 82.0148, 82.0712, 82.1345, 82.1533] };
  WCL[1753] = struct {
   x = [-22.8511, -22.7854, -22.6634, -22.5601, -22.532, -22.4944, -22.5132, -22.532, -22.5226, -22.5507, -22.6352, -22.7103, -22.7854, -22.8981, -23.0764, -23.4144, -23.7148, -23.5646, -23.4613, -23.2172,
        -23.358, -23.4613, -23.6303, -23.8086, -24.0433, -24.1466, -24.2498, -24.5972, -24.5127, -24.738, -24.907, -24.9633, -24.9257, -24.7098, -24.5784, -24.4845, -24.2686, -24.1466, -23.9025,
        -23.9213, -23.7805, -23.6866, -23.8274, -23.818, -23.6491, -23.4425, -23.1609, -23.0201, -22.8793, -22.6915, -22.363, -22.072, -21.9312, -21.7622, -21.7153, -21.5463, -21.2553, -20.9549,
        -20.7671, -20.5137, -20.3259, -20.5325, -20.5418, -20.4667, -20.4855, -20.4386, -20.157, -20.0255],
   y = [82.0336, 82.0172, 81.9562, 81.9116, 81.8717, 81.8154, 81.7614, 81.698, 81.6253, 81.5502, 81.4774, 81.4211, 81.3906, 81.339, 81.2944, 81.2779, 81.231, 81.2216, 81.2427, 81.2521, 81.1794, 81.1231,
        81.0362, 80.9893, 80.9236, 80.8391, 80.7452, 80.6654, 80.6396, 80.5692, 80.4824, 80.4331, 80.4307, 80.4824, 80.5246, 80.5575, 80.5903, 80.6255, 80.6889, 80.7264, 80.7734, 80.8273, 80.8297,
        80.8555, 80.8743, 80.9118, 80.9447, 80.9963, 81.048, 81.1207, 81.1888, 81.2451, 81.2944, 81.3483, 81.4023, 81.4469, 81.5173, 81.5736, 81.5948, 81.6229, 81.6487, 81.5971, 81.5431, 81.5079,
        81.4798, 81.4188, 81.454, 81.4633] };
  WCL[1754] = struct {
   x = [-19.7416, -20.0607, -20.5677, -20.7179, -21.497, -21.2529, -20.9901, -20.8868, -20.6146, -20.4269, -20.286, -20.0232],
   y = [80.6654, 80.6795, 80.6349, 80.5762, 80.5551, 80.5316, 80.534, 80.4871, 80.534, 80.5551, 80.6044, 80.6279] };
  WCL[1755] = struct {
   x = [-19.8965, -20.0842, -20.2063, -20.5536, -20.4034, -20.5254],
   y = [80.2711, 80.2031, 80.1397, 80.0881, 80.0787, 80.013] };
  WCL[1756] = struct {
   x = [-30.1756, -29.9879, -29.894],
   y = [83.5004, 83.5027, 83.484] };
  WCL[1757] = struct {
   x = [-30.0536, -29.7156, -29.5842, -29.3026, -29.4716, -29.6969, -29.5842, -29.3589, -28.9834, -28.4202, -27.7067, -27.2186, -26.937, -26.768, -26.8806, -27.181, -27.5378, -27.8382, -28.3451, -28.9271,
        -29.2087, -29.6781],
   y = [83.4065, 83.3995, 83.4042, 83.4206, 83.4229, 83.4441, 83.4511, 83.4605, 83.4652, 83.4511, 83.4183, 83.3737, 83.3432, 83.3033, 83.2516, 83.1953, 83.1765, 83.1507, 83.1296, 83.1366, 83.139, 83.146] };
  WCL[1758] = struct {
   x = [-30.3188, -29.9245, -29.5865, -29.5865, -29.6616, -29.3988, -29.2674, -28.798, -28.3662, -27.9719, -27.6903, -27.4275, -26.9956, -26.5826, -26.1508, -25.9255, -25.7753, -25.9443, -25.8879, -25.7565,
        -25.6814, -25.7565, -25.8879, -25.9255, -26.1883, -26.47, -26.2071, -26.3104, -26.4981, -26.8736, -27.2209, -27.0238, -26.6953, -26.2822, -26.1602, -25.8316, -25.0055, -24.6488, -24.386,
        -24.4235, -24.4329, -24.3578, -24.386, -24.339, -24.1043, -23.9072, -23.7382, -23.588, -23.4191, -23.297, -23.0436, -22.9309, -22.8277, -22.715, -22.5085, -22.2738, -21.9546, -21.72, -21.5698,
        -21.5885, -21.6449, -21.6542, -21.8232, -22.0391, -22.3301, -22.5836, -22.5367, -22.6869, -23.0436, -23.3534, -23.7382, -24.0105, -24.2827, -24.7896, -25.0525, -25.5218, -25.7659, -25.9161,
        -25.9255, -26.2071, -26.423, -26.6859, -27.0238, -27.3336, -27.6246, -27.9532, -28.385, -28.7886, -29.2486],
   y = [83.0897, 83.0803, 83.0615, 83.0287, 82.9935, 83.0216, 83.0521, 83.0475, 83.0404, 83.0357, 83.0428, 83.078, 83.1085, 83.1296, 83.1296, 83.1038, 83.0897, 83.0615, 83.0146, 82.9982, 82.9747, 82.9372,
        82.9066, 82.8973, 82.8644, 82.8433, 82.8456, 82.8081, 82.7823, 82.7588, 82.733, 82.7306, 82.7447, 82.7588, 82.7846, 82.8175, 82.8621, 82.855, 82.8738, 82.848, 82.8315, 82.801, 82.7682, 82.733,
        82.7752, 82.8104, 82.8198, 82.7917, 82.7682, 82.7541, 82.7494, 82.7729, 82.7823, 82.7471, 82.7377, 82.7095, 82.6649, 82.6368, 82.5945, 82.5593, 82.5358, 82.5335, 82.5265, 82.4842, 82.4514,
        82.4255, 82.381, 82.3317, 82.2965, 82.2707, 82.2566, 82.266, 82.2566, 82.219, 82.1721, 82.1369, 82.1251, 82.1251, 82.1251, 82.1251, 82.1228, 82.1228, 82.1345, 82.1345, 82.1557, 82.1557, 82.158,
        82.1627, 82.151] };
  WCL[1759] = struct {
   x = [-30.1639, -29.901, -29.4129, -29.1782, -28.5492, -28.1644, -27.6293, -27.1881, -26.6624, -26.5967, -26.1649, -26.1273, -26.3432, -26.9816, -27.2913, -27.6856, -28.1362, -28.2582, -28.1456, -27.9954,
        -27.7326, -27.1693, -26.4089, -25.7424, -25.7237, -25.3388, -24.8037, -24.616, -24.7755, -24.7849, -24.8037, -24.86, -24.8037, -24.5502, -24.3625, -23.8931, -23.3956, -22.9075, -22.8511],
   y = [81.9468, 81.9585, 81.975, 81.9796, 81.9796, 81.975, 81.9796, 81.9726, 81.9703, 81.9491, 81.9092, 81.7966, 81.7168, 81.6417, 81.6206, 81.5549, 81.5079, 81.4469, 81.4305, 81.407, 81.4047, 81.4915,
        81.576, 81.6276, 81.6652, 81.6957, 81.6957, 81.7332, 81.7637, 81.806, 81.8506, 81.9022, 81.9515, 81.9867, 82.0055, 82.0313, 82.036, 82.0454, 82.0336] };
  WCL[1760] = struct {
   x = [-40.0558, -39.8868],
   y = [83.1554, 83.1085] };
  WCL[1761] = struct {
   x = [-40.0793, -39.9009, -39.7225, -39.6474, -39.854, -39.976],
   y = [82.8362, 82.7705, 82.7541, 82.8081, 82.8081, 82.8503] };
  WCL[1762] = struct {
   x = [-40.0675, -39.7671, -39.279, -38.9598, -38.5843, -38.5843, -38.8284, -39.1476, -39.4855, -39.8422, -39.8798, -39.5606, -39.1851, -38.7721, -38.359, -38.3027, -38.5655, -38.8472, -38.9035, -39.0912,
        -39.0912, -39.3353, -39.7108, -39.8985, -39.9549],
   y = [82.9911, 82.9817, 82.9888, 83.0099, 82.9864, 82.9982, 83.0193, 83.0193, 83.0263, 83.0216, 83.0498, 83.0733, 83.0967, 83.1061, 83.1108, 83.1343, 83.1413, 83.1437, 83.1531, 83.1671, 83.1859, 83.1906,
        83.1859, 83.2, 83.2164] };
  WCL[1763] = struct {
   x = [-40.0229, -39.9478, -39.9478],
   y = [83.2516, 83.2774, 83.2986] };
  WCL[1764] = struct {
   x = [-40.0675, -39.8798],
   y = [83.3596, 83.3784] };
  WCL[1765] = struct {
   x = [-40.0018, -39.9455, -39.7577, -39.4949, -39.3259, -39.1194, -39.0067, -39.2508, -39.1757, -39.1382, -38.8941, -38.7439, -38.6876, -38.4998, -38.237, -38.0868, -37.8615, -37.5986, -37.5611, -37.5799,
        -37.8052, -37.8427, -37.7676, -37.8239, -38.0492, -38.0868, -37.9178, -37.5423, -37.1668, -37.0542, -36.8664, -36.6599, -36.5097, -36.322, -36.1905, -35.9277, -35.7587, -35.7587, -35.6085,
        -35.4208, -35.2143, -35.1767, -35.2706, -35.3269, -35.0641, -34.8388, -34.5384, -34.3131, -34.0502, -33.8249, -33.7686, -33.6184, -33.3368, -33.1115, -32.9237, -32.8674, -32.6421, -32.2666,
        -31.816, -31.4969, -31.1589, -30.8022, -30.6708, -30.6332, -30.3328, -30.4267, -30.1639],
   y = [83.4042, 83.43, 83.4206, 83.4065, 83.3948, 83.3666, 83.3737, 83.4206, 83.4488, 83.4793, 83.5051, 83.4957, 83.4699, 83.4605, 83.4464, 83.4112, 83.3666, 83.3455, 83.3549, 83.376, 83.3948, 83.4229,
        83.4347, 83.4605, 83.4863, 83.5145, 83.5379, 83.5544, 83.5567, 83.5426, 83.552, 83.5473, 83.5379, 83.5638, 83.5896, 83.5919, 83.5426, 83.5685, 83.5614, 83.5074, 83.4605, 83.4957, 83.5309,
        83.5614, 83.5896, 83.6037, 83.606, 83.6083, 83.6083, 83.606, 83.5896, 83.5685, 83.5825, 83.5966, 83.5708, 83.5661, 83.552, 83.5708, 83.5872, 83.5778, 83.5497, 83.5286, 83.5262, 83.4816,
        83.4722, 83.498, 83.5004] };
  WCL[1766] = struct {
   x = [-29.8869, -30.0559, -30.2812, -30.1873, -30.1498, -30.0559],
   y = [83.484, 83.4746, 83.4628, 83.4441, 83.4394, 83.4065] };
  WCL[1767] = struct {
   x = [-29.6781, -30.1099, -30.598, -31.1988, -31.5743, -31.8559, -32.2502, -32.5318, -32.8322, -33.264, -33.6771, -33.9024, -34.0714, -34.2028, -34.5219, -34.7097, -34.8036, -34.5219, -34.2216, -33.9212,
        -33.8648, -34.184, -34.5407, -34.935, -35.0664, -35.3293, -35.8362, -36.2117, -36.6059, -36.7561, -36.4933, -36.5872, -36.559, -36.437, -36.3243, -36.0427, -35.7423, -35.686, -35.4044,
        -35.1978, -35.1415, -35.1509, -34.9819, -34.9819, -34.9819, -34.9068, -34.6252, -34.3811, -34.062, -33.8179, -33.4236, -33.1608, -33.0669, -32.9355, -32.8604, -32.7853, -32.4849, -32.1282,
        -31.809, -31.5462, -31.1143, -30.7013, -30.2507],
   y = [83.146, 83.1625, 83.1671, 83.1625, 83.1578, 83.1531, 83.1179, 83.092, 83.0662, 83.0428, 83.0568, 83.0803, 83.0991, 83.1085, 83.1319, 83.1484, 83.1226, 83.0991, 83.0874, 83.0545, 83.0287, 82.9935,
        82.9747, 82.9841, 82.9606, 82.9325, 82.9231, 82.9137, 82.8996, 82.8879, 82.8808, 82.848, 82.8034, 82.8034, 82.8433, 82.8808, 82.8949, 82.8926, 82.8855, 82.8808, 82.8433, 82.7846, 82.8034,
        82.8433, 82.8761, 82.9113, 82.9278, 82.9325, 82.9301, 82.9348, 82.9489, 82.9536, 82.9113, 82.902, 82.9489, 82.9817, 83.0169, 83.0498, 83.0545, 83.0897, 83.1038, 83.1014, 83.0897] };
  WCL[1768] = struct {
   x = [-29.6663, -30.192, -30.6989, -31.1777, -31.3373, -31.5438, -31.7691, -31.9944, -32.3699, -32.905, -32.6515, -32.2197, -31.8536, -31.5907, -31.4218, -31.3842, -31.7785, -32.1446, -32.5576, -32.7829,
        -33.0082, -33.3743, -33.609, -34.0502, -34.1347, -34.1066, -34.0502, -33.7311, -33.4964, -33.1678, -32.736, -32.3323, -31.863, -31.4499, -31.1589, -30.9149, -30.9618, -30.6614, -30.5394,
        -30.1639],
   y = [82.151, 82.1251, 82.104, 82.104, 82.1181, 82.151, 82.1557, 82.1674, 82.1744, 82.158, 82.1416, 82.1463, 82.1369, 82.1111, 82.0923, 82.0454, 82.0148, 81.9843, 81.9585, 81.9562, 81.9304, 81.8999,
        81.8717, 81.8294, 81.7567, 81.7027, 81.637, 81.6558, 81.6816, 81.7473, 81.8013, 81.8294, 81.8529, 81.8764, 81.8975, 81.8999, 81.8482, 81.8905, 81.9257, 81.9468] };
  WCL[1769] = struct {
   x = [-48.4152, -48.3777, -48.0679, -47.852, -47.7581, -47.5891, -47.561, -47.5234, -47.3544, -47.1479, -46.9132, -46.7443, -46.5753, -46.3406, -46.1716, -46.0121, -45.8431, -45.7868, -45.6553, -45.4488,
        -45.3268, -45.2986, -45.5051, -45.599, -45.7868, -45.8619, -45.9276, -45.8712, -45.7586, -45.6459, -45.6647, -45.8243, -46.0121, -46.0965, -46.228, -46.4533, -46.6692, -46.9883, -47.2136,
        -47.3075, -47.4765, -47.6548, -47.8144, -47.9083, -48.0773, -48.2181, -48.3589, -48.4152],
   y = [82.6227, 82.625, 82.6344, 82.6508, 82.6438, 82.6321, 82.6485, 82.672, 82.672, 82.6696, 82.6508, 82.625, 82.5992, 82.5757, 82.5452, 82.5358, 82.503, 82.4654, 82.4302, 82.4021, 82.3645, 82.3246,
        82.2941, 82.2894, 82.2542, 82.2261, 82.1956, 82.1604, 82.1228, 82.0829, 82.0477, 82.0289, 82.0665, 82.1228, 82.1697, 82.2026, 82.2261, 82.2683, 82.2941, 82.3176, 82.3411, 82.3692, 82.4138,
        82.4772, 82.5335, 82.5781, 82.6086, 82.6227] };
  WCL[1770] = struct {
   x = [-47.6994, -47.6619, -47.5305, -47.3052, -46.9297, -46.9484, -47.1362, -47.2676, -47.4741, -47.5305, -47.399, -47.1362, -47.0048, -46.8358, -46.7982, -46.6293, -46.5729, -46.3289, -46.0473, -45.9909,
        -45.9346, -46.1036, -46.3664, -46.5542, -46.7044, -46.9297, -47.2113, -47.4366, -47.5868, -47.6994],
   y = [82.9442, 82.9512, 82.9372, 82.9418, 82.8973, 82.9254, 82.9653, 82.97, 82.9724, 82.9958, 83.0357, 83.0428, 83.0733, 83.0662, 83.0381, 83.0123, 82.9982, 82.9794, 82.9372, 82.8832, 82.8433, 82.8433,
        82.855, 82.855, 82.8409, 82.8456, 82.8503, 82.8973, 82.916, 82.9442] };
  WCL[1771] = struct {
   x = [-42.5387, -42.4261, -42.182, -42.013, -41.9004, -41.7126, -41.4874, -41.2621, -41.0931, -40.9429, -41.1682, -41.4686, -41.7314, -41.9379, -42.2383, -42.3885, -42.52, -42.5387],
   y = [83.1648, 83.1648, 83.1671, 83.1554, 83.1226, 83.0733, 83.0475, 83.024, 82.9935, 82.9606, 82.9536, 82.97, 83.0099, 83.0475, 83.0967, 83.1249, 83.1484, 83.1648] };
  WCL[1772] = struct {
   x = [-41.6329, -41.6329, -41.5202, -41.201, -41.0133, -40.8255, -40.5815, -40.4125, -40.45, -40.5251, -40.6378, -40.8443, -41.1447, -41.4639, -41.6329],
   y = [83.1343, 83.1671, 83.1648, 83.1484, 83.1249, 83.1038, 83.0827, 83.0498, 83.0169, 82.9864, 82.9559, 83.0005, 83.0545, 83.085, 83.1343] };
  WCL[1773] = struct {
   x = [-41.6117, -41.5366, -41.424, -41.1048, -40.842, -40.6542, -40.4853, -40.26, -40.0722],
   y = [83.3619, 83.3549, 83.3314, 83.2939, 83.2634, 83.2117, 83.1859, 83.1648, 83.1554] };
  WCL[1774] = struct {
   x = [-39.9009, -40.0135, -40.3139, -40.6519, -40.8396, -41.0086, -41.2339, -41.5155, -41.6657, -41.7408, -41.6469, -41.6094],
   y = [83.1085, 83.0944, 83.1202, 83.1437, 83.1859, 83.2329, 83.2446, 83.2516, 83.2657, 83.315, 83.3338, 83.3619] };
  WCL[1775] = struct {
   x = [-50.1354, -49.9289, -49.6567, -49.6097, -49.253, -48.8494, -48.4833, -48.4457, -48.1923, -47.7229, -47.338, -46.9719, -46.8029, -46.7372, -46.5589, -46.343, -46.0519, -45.7422, -45.4418, -45.3667,
        -45.1789, -45.0757, -45.0757, -45.1038, -45.3291, -45.5075, -45.5826, -45.6765, -45.7703, -45.6483, -45.5169, -45.4042, -44.9912, -44.794, -44.503, -44.6438, -44.9255, -45.1695, -45.254,
        -45.4512, -45.5732, -45.6671, -45.6014, -45.5263, -45.4042, -45.1977, -44.7659, -44.3341, -43.9022, -43.6488, -43.5174, -43.3953, -43.4892, -43.7427, -43.9116, -44.2965, -44.503, -44.7377,
        -44.6532, -44.6345, -44.7847, -44.9912, -45.2822, -45.4136, -45.5075, -45.6577, -45.8267, -45.9581, -46.2679, -46.4274, -46.5213, -46.4838, -46.2021, -45.8736, -45.5356, -45.2259, -44.8128,
        -44.3434, -43.8835, -43.4047, -43.1513, -42.9353, -42.926, -42.8978, -42.8415, -42.7007, -42.6819, -42.7007, -42.7101, -42.757, -42.7758, -42.8884, -42.7664, -42.4472, -42.2125, -41.9497,
        -41.7995, -41.6117, -41.3489, -41.1611, -41.0954, -40.9171, -40.9171, -40.842, -40.9171, -41.0579, -40.964, -40.9828, -40.964, -40.9734, -41.1236, -41.2456, -41.5366, -41.7713, -41.8934,
        -41.9872, -42.1187, -42.3158, -42.5505, -42.757, -43.0198, -43.3766, -43.5643, -43.7614, -43.8835, -44.1275, -44.5594, -44.9349, -45.1789, -45.2446, -45.1789, -45.0757, -44.8316, -44.6063,
        -44.4843, -44.5406, -44.6345, -44.7659, -44.8691, -44.9912, -45.0851, -45.1977, -45.3103, -45.423, -45.3761, -45.4512, -45.5169, -45.5638, -45.4887, -45.4512, -45.3761, -45.3291, -45.2916,
        -45.2165, -45.1414, -45.1508, -45.1883, -45.1414, -44.9818, -44.7189, -44.6157, -44.5687, -44.6157, -44.4279, -44.2777, -44.3716, -44.5406, -44.6814, -44.794, -44.6438, -44.4655, -44.3716,
        -44.4092, -44.4561, -44.4092, -44.2402, -44.1088, -43.9867, -43.8459, -43.7614, -43.752, -43.8271, -43.8553, -43.7427, -43.5925, -43.4423, -43.339, -43.2451, -43.1231, -42.9635, -42.879,
        -42.7382, -42.5599, -42.4848, -42.2595],
   y = [82.4913, 82.4701, 82.4232, 82.3645, 82.3176, 82.28, 82.2237, 82.1627, 82.1533, 82.1181, 82.0923, 82.0383, 81.9867, 81.9397, 81.9163, 81.867, 81.8177, 81.7661, 81.745, 81.7684, 81.7778, 81.7966,
        81.8388, 81.8553, 81.874, 81.9116, 81.9421, 81.9609, 82.0008, 82.0242, 82.0407, 82.0641, 82.0782, 82.0571, 82.0477, 82.0618, 82.0829, 82.1064, 82.1087, 82.104, 82.1298, 82.1674, 82.1979,
        82.2284, 82.266, 82.3012, 82.2941, 82.2566, 82.2308, 82.1909, 82.2049, 82.2073, 82.2355, 82.2589, 82.266, 82.28, 82.3012, 82.3293, 82.3856, 82.388, 82.4044, 82.4443, 82.4772, 82.5077, 82.5382,
        82.5734, 82.5945, 82.6391, 82.6743, 82.7142, 82.733, 82.7682, 82.7705, 82.7752, 82.7635, 82.7588, 82.7611, 82.7729, 82.7705, 82.7682, 82.7283, 82.6696, 82.6109, 82.5288, 82.4584, 82.4748,
        82.4983, 82.5452, 82.5875, 82.625, 82.6837, 82.7236, 82.7447, 82.7189, 82.672, 82.6391, 82.5945, 82.5546, 82.5405, 82.4842, 82.4326, 82.3927, 82.4326, 82.4819, 82.5077, 82.5476, 82.5828,
        82.6156, 82.6485, 82.6485, 82.6579, 82.6837, 82.7095, 82.7119, 82.733, 82.7564, 82.7705, 82.7752, 82.7917, 82.8175, 82.8269, 82.8433, 82.8433, 82.8433, 82.8409, 82.8339, 82.8315, 82.8362,
        82.8667, 82.902, 82.9465, 82.9325, 82.8973, 82.8996, 82.9043, 82.9184, 82.909, 82.916, 82.9113, 82.9325, 82.9536, 82.9817, 82.9911, 83.0099, 83.0381, 83.0545, 83.0521, 83.0944, 83.1296,
        83.1085, 83.1202, 83.139, 83.1132, 83.0944, 83.0733, 83.0967, 83.1319, 83.1531, 83.1484, 83.1085, 83.0686, 83.0803, 83.1179, 83.1085, 83.0686, 83.1108, 83.1578, 83.1531, 83.1859, 83.2047,
        83.1859, 83.2117, 83.2258, 83.2352, 83.2704, 83.2634, 83.2282, 83.2282, 83.2, 83.1695, 83.1554, 83.1296, 83.0803, 83.0874, 83.0944, 83.0615, 83.0287, 82.9911, 82.9442, 82.9372, 82.9231,
        82.9348, 82.9278, 82.8879, 82.8362] };
  WCL[1776] = struct {
   x = [-39.9783, -40.091, -40.2412, -40.2036, -40.0722],
   y = [82.8503, 82.8926, 82.9278, 82.97, 82.9911] };
  WCL[1777] = struct {
   x = [-39.9572, -40.0323, -40.0323],
   y = [83.2164, 83.2282, 83.2516] };
  WCL[1778] = struct {
   x = [-39.9572, -40.0323, -40.0699],
   y = [83.2986, 83.3291, 83.3596] };
  WCL[1779] = struct {
   x = [-39.8798, -40.0112],
   y = [83.3784, 83.4042] };
  WCL[1780] = struct {
   x = [-53.7214, -53.6745, -53.5618, -53.3553, -53.1863, -53.1863, -53.1206, -52.8953, -52.5949, -52.3978, -52.2194, -51.9566, -51.8158, -51.8345, -52.0786, -52.2382, -52.4166, -52.6512, -52.8296, -53.0174,
        -53.2145, -53.3928, -53.5618, -53.7402, -53.8153, -53.9279, -53.9655, -53.8998, -53.8528, -53.7214],
   y = [82.3176, 82.3082, 82.3012, 82.28, 82.2566, 82.219, 82.1791, 82.1533, 82.1228, 82.0806, 82.0383, 82.0008, 81.9726, 81.9585, 81.9444, 81.9491, 81.9491, 81.9703, 81.982, 82.0172, 82.05, 82.0547,
        82.0759, 82.1158, 82.165, 82.2073, 82.2472, 82.2683, 82.2965, 82.3176] };
  WCL[1781] = struct {
   x = [-60.194, -59.8936, -59.612, -59.4148, -59.3961, -59.3679, -59.3116, -59.1144, -58.8985, -58.7014, -58.4292, -58.2132, -58.0537, -57.9128, -57.7908, -57.7908, -57.6594, -57.5749, -57.6594, -57.7345,
        -58.0161, -58.2039, -58.3165, -58.4479, -58.4855, -58.6732, -58.8046, -58.9079, -58.9924, -59.1801, -59.4148, -59.612, -59.7997],
   y = [81.8693, 81.8576, 81.8412, 81.7896, 81.745, 81.7121, 81.6675, 81.6253, 81.5971, 81.5713, 81.5572, 81.5337, 81.4962, 81.4704, 81.4446, 81.4164, 81.3906, 81.4258, 81.4704, 81.5384, 81.5689, 81.6018,
        81.6112, 81.6206, 81.6605, 81.6464, 81.7145, 81.7684, 81.813, 81.8717, 81.9116, 81.9163, 81.9397] };
  WCL[1782] = struct {
   x = [-60.0297, -59.8701, -59.6448, -59.4289, -59.1473, -58.9689, -58.6498, -58.3869, -58.1241, -57.7486, -57.4763, -57.1853, -57.0257, -57.007, -56.8943, -56.8192, -56.4625, -56.378, -56.1715, -55.918,
        -56.0119, -56.1809, -56.1997, -55.9368, -55.7397, -55.4393, -55.1764, -54.9887, -54.8479, -54.6977, -54.5944, -54.4912, -54.3222, -54.2565, -54.1908, -54.0593, -54.0406, -54.0687, -54.1344,
        -54.2002, -54.2565, -54.2565, -54.2002, -54.05, -53.881, -53.9091, -53.9279, -53.6745, -53.4961, -53.4586, -53.4961, -53.5243, -53.2708, -53.0267, -52.7921, -52.6419, -52.6043, -52.4447,
        -52.1349, -51.8533, -51.5717, -51.3558, -51.0742, -50.8864, -50.6236, -50.6987, -50.9334, -51.1211, -51.2995, -51.4872, -51.6562, -51.2995, -50.877, -50.6424, -50.4358, -50.2105, -50.2856,
        -50.7081, -50.8864, -50.9991, -51.1868, -51.3558, -51.4778, -51.5248, -51.5248, -51.5342, -51.5717, -51.6468, -51.7594, -51.8627, -51.8345, -51.797, -51.7876],
   y = [81.9679, 82.0055, 82.0313, 82.0594, 82.0806, 82.0946, 82.1134, 82.1205, 82.1369, 82.151, 82.1791, 82.2002, 82.1979, 82.1627, 82.1932, 82.2331, 82.2542, 82.2378, 82.2261, 82.2378, 82.2472, 82.2448,
        82.266, 82.2941, 82.3176, 82.3528, 82.3458, 82.3293, 82.2636, 82.2378, 82.2049, 82.165, 82.1228, 82.0454, 81.9726, 81.9304, 81.8787, 81.813, 81.7543, 81.7051, 81.644, 81.5877, 81.5337, 81.5267,
        81.5666, 81.6276, 81.691, 81.7497, 81.8224, 81.8881, 81.9421, 82.0008, 81.975, 81.9491, 81.9327, 81.9163, 81.9139, 81.9069, 81.8975, 81.8482, 81.806, 81.759, 81.7309, 81.7027, 81.6863, 81.7309,
        81.7708, 81.7942, 81.8482, 81.8834, 81.9257, 81.9022, 81.8928, 81.86, 81.8623, 81.8952, 81.9444, 81.989, 82.036, 82.0853, 82.1298, 82.1674, 82.2073, 82.2519, 82.2941, 82.334, 82.3716, 82.3974,
        82.4373, 82.4795, 82.4983, 82.5077, 82.5077] };
  WCL[1783] = struct {
   x = [-51.79, -51.6679, -51.4802, -51.1047, -50.8137, -50.5696, -50.288, -50.1566, -50.1378],
   y = [82.5077, 82.5077, 82.51, 82.5147, 82.5053, 82.503, 82.4913, 82.4913, 82.4913] };
  WCL[1784] = struct {
   x = [-70.1962, -69.755, -69.6517, -69.5109, -69.3983, -69.3701, -69.2011, -69.0415, -68.7975, -68.6848, -68.4971, -68.1873, -68.1216, -68.0089, -67.8775, -67.7367, -67.5959, -67.6241, -67.5583, -67.3424,
        -67.089, -66.92, -66.7792, -66.6008, -66.5257, -66.3004, -66.1221, -66.0094, -65.7935, -65.4274, -65.1082, -64.9017, -64.7515, -64.6201, -64.5919, -64.5356, -64.5919, -64.7609, -64.9017,
        -65.0238, -65.3242, -65.6996, -66.0188, -66.4037, -66.6478, -66.8637, -67.1265, -67.4175, -67.6898, -68.0089, -68.2999, -68.5534, -68.8162, -69.0603, -69.2762, -69.5578, -69.7831, -69.9897],
   y = [80.2805, 80.3298, 80.3979, 80.4519, 80.5152, 80.5481, 80.5762, 80.6208, 80.6537, 80.7006, 80.7382, 80.7476, 80.7734, 80.8015, 80.8227, 80.8508, 80.8649, 80.879, 80.8931, 80.9142, 80.9447, 80.9963,
        81.0151, 81.048, 81.0949, 81.1465, 81.2028, 81.2028, 81.2169, 81.2732, 81.3202, 81.3483, 81.3835, 81.4188, 81.454, 81.4985, 81.5314, 81.5291, 81.5244, 81.5244, 81.5079, 81.4845, 81.4586,
        81.4164, 81.3835, 81.3554, 81.339, 81.3272, 81.3202, 81.2944, 81.2592, 81.231, 81.2099, 81.177, 81.1629, 81.17, 81.1324, 81.0902] };
  WCL[1785] = struct {
   x = [-70.0319, -69.9286, -69.9286, -69.9005, -69.7784, -69.5813, -69.3372, -69.0838, -68.8585, -68.605, -68.3516, -68.1169, -67.9198, -67.6663, -67.3565, -67.1218, -67.0843, -67.0561, -67.1688, -67.3565,
        -67.5724, -67.8916, -68.1263, -68.3046, -68.5581, -68.7458, -68.9617, -69.1495, -69.1777, -69.1683, -69.0462, -68.8116, -68.7552, -68.9148, -69.0838, -69.234, -69.3654, -69.4499, -69.2434,
        -69.0462, -68.8773, -68.6801, -68.5393, -68.2389, -68.1638, -68.3046, -68.2483, -68.0418, -67.9479, -67.7414, -67.3565, -67.1688, -66.9341, -66.7276, -66.4929, -66.2113, -66.0892, -65.9109,
        -65.6668, -65.479, -65.4978, -65.7701, -66.0329, -65.8639, -65.6762, -65.4697, -65.3007, -65.0472, -64.9627, -64.8501, -64.7093, -64.5497, -64.3526, -64.3526, -64.3807, -64.3244, -64.1554,
        -64.0803, -63.9395, -63.8363, -63.7799, -63.6297, -63.4138, -63.0665, -62.7661, -62.6253, -62.4563, -62.1465, -61.9963, -61.9306, -61.7429, -61.5458, -61.3486, -61.2829, -61.2454, -61.1984,
        -61.1609, -61.1703, -61.2547, -61.3956, -61.6115, -61.8931, -62.2498, -62.3906, -62.3718],
   y = [81.1723, 81.2099, 81.238, 81.2427, 81.231, 81.2122, 81.2216, 81.2451, 81.2732, 81.3108, 81.3507, 81.3789, 81.4047, 81.4281, 81.4586, 81.4821, 81.4962, 81.5032, 81.5197, 81.5291, 81.5291, 81.5103,
        81.5009, 81.4962, 81.5009, 81.5056, 81.4938, 81.4774, 81.5126, 81.5502, 81.5337, 81.5337, 81.5549, 81.583, 81.6159, 81.644, 81.6769, 81.7051, 81.6769, 81.6417, 81.6018, 81.5807, 81.5619,
        81.5689, 81.5854, 81.6159, 81.6135, 81.6065, 81.5948, 81.5924, 81.5877, 81.5948, 81.6159, 81.6253, 81.6042, 81.6206, 81.6253, 81.6159, 81.6417, 81.6605, 81.6652, 81.6605, 81.6628, 81.6886,
        81.7004, 81.7121, 81.7215, 81.7309, 81.7074, 81.698, 81.691, 81.7004, 81.7191, 81.7543, 81.7708, 81.7614, 81.7403, 81.745, 81.7708, 81.8013, 81.8201, 81.8388, 81.8482, 81.8811, 81.9163,
        81.9656, 81.9796, 82.0172, 82.0688, 82.104, 82.1228, 82.1533, 82.1768, 82.2049, 82.219, 82.2495, 82.2918, 82.3434, 82.3739, 82.4279, 82.4514, 82.4607, 82.4701, 82.4842, 82.5006] };
  WCL[1786] = struct {
   x = [-62.3671, -62.3859, -62.6394, -62.8177, -62.921, -63.1369, -63.3904, -63.2683, -63.2496, -63.1745, -63.0524, -63.2402, -63.5218, -63.5969, -63.6157, -63.8222, -63.9912, -63.7565, -63.7001, -63.8034,
        -63.9161, -64.0193, -64.132, -64.2822, -64.4042, -64.545, -64.7234, -64.7328, -64.7421, -64.7891, -64.7703, -64.7797, -64.9111, -64.9768, -64.9017, -64.9768, -65.2397, -65.174, -65.005,
        -65.1176, -65.2303, -65.4274, -65.3993, -65.5025, -66.047, -66.4506, -66.5633, -66.629, -66.8825, -66.92, -66.9576, -67.1359, -67.2861, -67.4926, -67.8587, -68.2248, -68.5346, -68.3375,
        -67.9526, -67.4645, -67.0514, -66.7041, -66.4412, -66.3662, -66.3286, -66.4412, -66.5727, -66.6572, -66.7416, -66.8543, -66.9388, -67.0233, -67.0327, -67.1078, -67.2298, -67.2955, -67.4645,
        -67.5583, -67.671, -67.7367, -67.7461, -67.8118],
   y = [82.5006, 82.503, 82.503, 82.4654, 82.4913, 82.4607, 82.4162, 82.4561, 82.4913, 82.5265, 82.5593, 82.5734, 82.6039, 82.6438, 82.672, 82.7095, 82.733, 82.7588, 82.801, 82.8574, 82.8433, 82.8527,
        82.8362, 82.8034, 82.7823, 82.8057, 82.8151, 82.848, 82.8738, 82.8761, 82.8949, 82.916, 82.909, 82.8855, 82.8667, 82.848, 82.8081, 82.7917, 82.7611, 82.733, 82.7799, 82.8081, 82.848, 82.8503,
        82.8034, 82.7541, 82.733, 82.7212, 82.7119, 82.7072, 82.7048, 82.6954, 82.672, 82.6508, 82.6461, 82.625, 82.6086, 82.6696, 82.6954, 82.7424, 82.7893, 82.8175, 82.8644, 82.9184, 82.9512,
        82.9536, 82.9583, 82.9653, 82.9653, 82.9536, 82.9747, 82.9888, 82.9911, 82.9794, 82.9724, 82.9958, 83.0169, 82.9958, 82.9935, 82.9817, 82.9583, 83.0193] };
  WCL[1787] = struct {
   x = [-64.8806, -64.6178, -64.43, -64.308, -64.2423, -63.9888, -63.7447, -64.0076, -64.1578, -64.1859, -64.1202, -64.2423, -64.2798, -64.3268, -64.4582, -64.6178, -64.8994, -65.1059, -65.2843, -65.3781,
        -65.4908],
   y = [79.9825, 80.0153, 80.0482, 80.0811, 80.1092, 80.1022, 80.121, 80.1303, 80.1749, 80.2266, 80.2524, 80.2618, 80.2195, 80.1608, 80.1022, 80.0717, 80.0435, 80.0412, 80.0552, 80.0811, 80.0224] };
  WCL[1788] = struct {
   x = [-66.1596, -66.2441, -66.46, -66.6008, -66.798, -67.0514, -67.1922, -67.1922, -67.2486, -67.3518, -67.5114, -67.6334, -67.6616, -67.6334, -67.6334, -67.6992, -67.671, -67.4363, -67.3049, -67.1078,
        -67.0233, -66.9576, -66.8355, -66.7135, -66.7323, -66.8167, -66.7041, -66.5163, -66.3943, -66.2441, -66.0376, -65.8405, -65.6809, -65.4931, -65.4274, -65.6433, -65.7278, -65.6809, -65.5307,
        -65.3429, -65.1458, -65.1552, -64.9862, -64.883, -64.9205, -64.7797, -64.5262, -64.3854, -64.3385, -64.3385, -64.2915, -64.1601, -64.0663, -63.9254, -63.8222, -63.5875, -63.3716, -63.1557,
        -62.8271, -62.6112, -62.4986, -62.339, -62.0855, -61.9447, -61.7851, -61.5504, -61.5598, -61.6443, -61.7288, -61.7288, -61.7664, -61.5504, -61.3909, -61.3439, -61.5504, -61.6255, -61.6349,
        -61.7288, -61.757, -61.8696, -61.8696, -61.879, -61.9447, -61.9259, -61.8696, -61.7288, -61.4847, -61.2407, -61.0248, -60.8746, -60.7713],
   y = [79.9895, 80.0318, 80.0787, 80.0552, 80.0388, 80.0224, 80.0505, 80.0529, 80.0787, 80.1069, 80.1186, 80.1491, 80.2054, 80.2148, 80.2688, 80.3087, 80.3251, 80.3509, 80.3768, 80.3908, 80.4073, 80.4143,
        80.4472, 80.4894, 80.5035, 80.5363, 80.5598, 80.5622, 80.5786, 80.6044, 80.6373, 80.6373, 80.6208, 80.5927, 80.6185, 80.6419, 80.6795, 80.7124, 80.7452, 80.7687, 80.8015, 80.8461, 80.8719,
        80.9118, 80.9541, 80.9916, 81.0104, 81.0409, 81.0667, 81.0667, 81.0949, 81.1348, 81.1442, 81.1348, 81.1583, 81.2028, 81.1981, 81.1888, 81.1794, 81.1583, 81.116, 81.0972, 81.0714, 81.0972,
        81.1113, 81.1559, 81.1958, 81.2427, 81.2826, 81.3061, 81.3554, 81.4, 81.4563, 81.5056, 81.5337, 81.5549, 81.5783, 81.6088, 81.63, 81.644, 81.6347, 81.6464, 81.6816, 81.7191, 81.7684, 81.813,
        81.8482, 81.8764, 81.8928, 81.9092, 81.9163] };
  WCL[1789] = struct {
   x = [-60.7689, -60.7408, -60.6281, -60.4404, -60.1869],
   y = [81.9163, 81.921, 81.9045, 81.8858, 81.8693] };
  WCL[1790] = struct {
   x = [-59.7974, -60.032],
   y = [81.9397, 81.9679] };
  WCL[1791] = struct {
   x = [-80.1773, -79.9708, -79.8018, -79.5483, -79.2385, -79.0508, -78.8349, -78.6565, -78.4782, -78.4031, -78.4125, -78.6096, -78.863, -79.0883, -79.2855, -79.5483, -79.7642, -79.8957],
   y = [80.4988, 80.5176, 80.527, 80.5363, 80.5363, 80.5481, 80.5598, 80.5622, 80.5645, 80.5762, 80.588, 80.5974, 80.595, 80.5903, 80.6114, 80.5974, 80.595, 80.5833] };
  WCL[1792] = struct {
   x = [-80.2101, -79.9567, -79.8346, -79.6375, -79.4779, -79.262, -79.1587, -79.0367, -78.9616, -78.8208, -78.5392, -78.2294, -78.0135, -77.8915, -77.8351, -77.76, -77.76, -77.76, -77.6568, -77.5066,
        -77.3376, -77.1592, -77.0278, -76.9058, -76.8213, -76.6805, -76.5585, -76.5585, -76.5866, -76.6523, -76.6242, -76.5209, -76.3895, -76.2299, -76.136, -76.136, -76.1642, -76.1266, -75.9483,
        -75.7042, -75.5634, -75.4977, -75.4977, -75.5822, -75.6854, -75.8732, -76.1266, -76.3238, -76.4834, -76.5866, -76.7368, -76.8964, -77.0184, -77.1123, -77.225, -77.4127, -77.5066, -77.5723,
        -77.6849, -77.7506, -77.8351, -77.8539, -77.9384, -78.0886, -78.2012, -78.3514, -78.4547, -78.5204, -78.6049, -78.7833, -78.9241, -79.0273, -79.093, -79.2151, -79.3653, -79.4216, -79.4873,
        -79.5061, -79.5155, -79.5436, -79.4498, -79.2338, -79.1306, -79.1306, -79.1306, -79.0461, -78.849, -78.8114, -78.7739, -78.7739, -78.7551, -78.6894, -78.5955, -78.4641, -78.3984, -78.2857,
        -78.1355, -78.1074, -78.0135, -77.929, -77.8445, -77.807, -77.6943, -77.638, -77.7131, -77.8445, -77.9478, -78.1355, -78.2857, -78.4829, -78.6049, -78.7645, -78.9428, -79.0367, -79.1775,
        -79.2808, -79.3653, -79.4122, -79.4216, -79.384, -79.3934, -79.4591, -79.5342, -79.6, -79.6751, -79.722, -79.7314, -79.7971, -79.891],
   y = [80.6232, 80.6373, 80.6607, 80.6889, 80.7006, 80.7194, 80.7382, 80.7405, 80.7499, 80.7593, 80.7804, 80.8015, 80.8109, 80.8062, 80.7828, 80.7828, 80.7945, 80.818, 80.8273, 80.8273, 80.8133, 80.8156,
        80.8203, 80.8227, 80.8039, 80.7875, 80.7921, 80.818, 80.8391, 80.8532, 80.8672, 80.8766, 80.8837, 80.8602, 80.8579, 80.8813, 80.8978, 80.9118, 80.9189, 80.9189, 80.9306, 80.9377, 80.9517,
        80.9541, 80.94, 80.933, 80.9283, 80.9165, 80.9048, 80.9001, 80.8907, 80.886, 80.8766, 80.8766, 80.8837, 80.8837, 80.8884, 80.9001, 80.9001, 80.9001, 80.8978, 80.9048, 80.9071, 80.8978, 80.9024,
        80.8954, 80.8907, 80.8884, 80.8766, 80.8743, 80.8579, 80.8532, 80.8602, 80.8438, 80.8367, 80.8485, 80.8579, 80.8813, 80.9118, 80.9541, 80.9869, 81.0268, 81.0597, 81.1043, 81.1277, 81.1629,
        81.1794, 81.2216, 81.238, 81.2427, 81.2545, 81.2709, 81.2873, 81.2991, 81.3225, 81.339, 81.3507, 81.3554, 81.3695, 81.3835, 81.3953, 81.4141, 81.4446, 81.4657, 81.4657, 81.4493, 81.4211, 81.4,
        81.3765, 81.3577, 81.3413, 81.3178, 81.285, 81.2474, 81.2169, 81.1841, 81.1395, 81.116, 81.1113, 81.1066, 81.0925, 81.0832, 81.0738, 81.0832, 81.0972, 81.1066, 81.1043, 81.1066, 81.1301] };
  WCL[1793] = struct {
   x = [-80.067, -79.9637, -79.8698, -79.7666, -79.8511, -79.9262, -79.8698, -79.8511, -79.7384, -79.7384, -79.8041, -79.9356, -79.9356],
   y = [81.1254, 81.1043, 81.0855, 81.0526, 81.0104, 80.9353, 80.9212, 80.947, 80.9377, 80.9118, 80.8813, 80.8532, 80.818] };
  WCL[1794] = struct {
   x = [-71.1678, -70.9049, -70.8862, -71.2523, -71.6653, -72.2473],
   y = [79.959, 80.0013, 80.0388, 80.0623, 80.0412, 80.006] };
  WCL[1795] = struct {
   x = [-72.7003, -72.4937, -72.4093, -72.2403, -71.8272, -71.5268, -71.095, -70.682, -70.391, -70.3346, -70.5036, -70.804, -70.8697, -70.804, -70.7101, -70.682, -70.682, -70.6257, -70.1938],
   y = [79.9989, 80.0271, 80.0787, 80.0599, 80.067, 80.1092, 80.1491, 80.1327, 80.1374, 80.1749, 80.2242, 80.2618, 80.3134, 80.3815, 80.4073, 80.3744, 80.3181, 80.2829, 80.2805] };
  WCL[1796] = struct {
   x = [-69.9873, -70.1469, -70.3159, -70.3253, -70.1751, -70.2126, -70.4567, -70.8791, -71.0293, -71.0575, -70.9073, -70.7571, -70.513, -70.2502, -70.0249],
   y = [81.0902, 81.0644, 81.048, 81.0667, 81.1137, 81.1137, 81.0972, 81.0972, 81.1254, 81.1418, 81.1395, 81.1348, 81.1324, 81.1395, 81.1723] };
  WCL[1797] = struct {
   x = [-69.7667, -70.0296, -70.3487, -70.5177, -70.6116, -70.6679, -70.9307, -71.2311, -72.1136, -72.1323, -72.1886, -72.2074, -72.489, -72.3952, -72.489, -72.9209, -73.24, -73.7094, -73.7282, -73.822,
        -73.8971, -73.9722, -74.1788, -74.3477, -74.2726, -74.2163, -74.038, -73.8033, -74.1882, -74.3665, -74.742, -74.8922, -74.9861, -75.1363, -75.3991, -75.8122, -76.2064, -76.3942, -76.6946,
        -77.0325, -77.2578, -77.2766, -77.2766, -77.5394, -77.7084, -77.9712, -78.2153, -78.4782, -78.4594, -78.3467, -78.1027, -77.8211, -77.577, -77.3986, -77.1639, -77.2203, -77.117, -76.7603,
        -76.4223, -76.1877, -76.3848, -76.704, -76.9668, -77.0044, -76.826, -76.9386, -77.2484, -77.577, -77.699, -77.9806, -78.2998, -78.6753, -79.1071, -79.2949, -79.2949, -79.4451, -79.4826,
        -79.6704, -79.7642, -79.8393, -80.0646, -80.2148, -80.3087],
   y = [83.0193, 83.031, 83.0146, 83.0521, 83.0874, 83.1085, 83.1061, 83.1038, 83.0521, 83.031, 83.0052, 82.9935, 83.0193, 83.0803, 83.0944, 83.1061, 83.085, 83.0874, 83.0874, 83.0662, 83.0568, 83.031,
        82.9911, 82.9325, 82.8926, 82.848, 82.7893, 82.7447, 82.7752, 82.8175, 82.8644, 82.8926, 82.9113, 82.9653, 83.0123, 83.0381, 83.0381, 83.0263, 83.0334, 83.0475, 83.0404, 83.0404, 83.0404,
        83.0263, 83.0076, 83.0076, 83.0146, 83.0029, 82.9583, 82.9418, 82.9113, 82.8574, 82.8151, 82.7729, 82.7705, 82.7447, 82.686, 82.6368, 82.5828, 82.5476, 82.5077, 82.449, 82.4162, 82.4701,
        82.5218, 82.5851, 82.6391, 82.6931, 82.7447, 82.8057, 82.8667, 82.902, 82.9325, 82.902, 82.8761, 82.8644, 82.8433, 82.8386, 82.8761, 82.9231, 82.9301, 82.8832, 82.8761] };
  WCL[1798] = struct {
   x = [-90.0363, -89.811, -89.6327, -89.6702, -89.5576, -89.4637, -89.3041, -89.229, -89.1727, -89.06, -88.9849, -88.9474, -88.938, -88.938, -89.0413, -89.0882, -89.0131, -88.9756, -88.9474, -88.9943,
        -88.9943, -88.9943, -88.9005, -88.816, -88.8911, -88.8535, -88.6939, -88.5719, -88.4217, -88.3748, -88.2527, -88.1401, -88.1307, -88.1307, -88.2246, -88.2058, -88.2903, -88.4405, -88.5625,
        -88.5719, -88.5719, -88.5625, -88.4217, -88.187, -88.0087, -87.8585, -87.6801, -87.5956, -87.605, -87.5487, -87.5299, -87.5299, -87.6332, -87.6707, -87.8115, -87.8585, -87.9711, -87.9899,
        -87.8679, -87.7364, -87.6144, -87.5393, -87.3797, -87.2483],
   y = [80.527, 80.4894, 80.4847, 80.5176, 80.5129, 80.5223, 80.5199, 80.5035, 80.4941, 80.4824, 80.4472, 80.4331, 80.4026, 80.3908, 80.3908, 80.3697, 80.3603, 80.3439, 80.3087, 80.2852, 80.2336, 80.1937,
        80.1585, 80.1397, 80.1608, 80.1468, 80.1116, 80.0717, 80.0646, 80.0904, 80.0881, 80.0904, 80.1022, 80.121, 80.1397, 80.1679, 80.182, 80.2148, 80.2524, 80.2688, 80.3157, 80.3462, 80.4026,
        80.4073, 80.3932, 80.3908, 80.3721, 80.3486, 80.2946, 80.2571, 80.2242, 80.1679, 80.1585, 80.1374, 80.121, 80.1022, 80.0975, 80.0764, 80.0599, 80.0646, 80.0811, 80.0764, 80.0646, 80.0505] };
  WCL[1799] = struct {
   x = [-86.3541, -86.5231, -86.6733, -86.739, -86.739, -86.7296, -86.7296, -86.6827, -86.6545, -86.617, -86.5231, -86.3729, -86.1852, -86.035, -85.8191, -85.5281, -85.284, -85.1432, -85.0681, -84.9179,
        -84.5424, -84.2044, -84.0261, -83.8665, -83.6318, -83.4253, -83.2094],
   y = [79.9989, 80.0083, 80.0224, 80.0717, 80.121, 80.1632, 80.1867, 80.2101, 80.243, 80.2665, 80.304, 80.3181, 80.3134, 80.3181, 80.3228, 80.2805, 80.2406, 80.2266, 80.2477, 80.2594, 80.2594, 80.2359,
        80.2148, 80.1679, 80.1069, 80.0693, 80.006] };
  WCL[1800] = struct {
   x = [-82.4936, -82.5593, -82.5593, -82.7001, -82.8597, -83.0193, -83.1319, -83.254, -83.376, -83.4323, -83.4229, -83.2352, -83.0568, -82.8973, -82.794, -82.625, -82.3903, -82.2308, -82.0618, -81.9491,
        -81.921, -81.7989, -81.5924, -81.4141, -81.17, -81.0292, -80.8414, -80.7382, -80.5974, -80.4941, -80.4378, -80.4659, -80.5974, -80.7476, -80.71, -80.588, -80.5035, -80.4002, -80.1749],
   y = [79.9966, 80.02, 80.0646, 80.1045, 80.1444, 80.1749, 80.2101, 80.2406, 80.2711, 80.2805, 80.311, 80.3228, 80.3275, 80.3275, 80.3369, 80.3509, 80.3721, 80.3603, 80.3627, 80.3556, 80.3556, 80.365,
        80.3908, 80.4073, 80.4143, 80.4167, 80.4143, 80.4307, 80.4167, 80.4167, 80.426, 80.4378, 80.4495, 80.4659, 80.4894, 80.5082, 80.5082, 80.5011, 80.4988] };
  WCL[1801] = struct {
   x = [-79.891, -80.0787, -80.1632, -80.2852, -80.4636, -80.4073, -80.3322, -80.2101],
   y = [80.5833, 80.5739, 80.5669, 80.5692, 80.5786, 80.5903, 80.6067, 80.6232] };
  WCL[1802] = struct {
   x = [-79.891, -79.9942, -80.0599, -80.1632, -80.1632, -80.0599],
   y = [81.1301, 81.1418, 81.177, 81.1676, 81.1465, 81.1254] };
  WCL[1803] = struct {
   x = [-79.9356, -80.0388, -80.1608, -80.311, -80.4425, -80.48, -80.5551, -80.7147, -80.8743, -81.0433, -81.2404, -81.4563, -81.6628, -81.7661, -81.9257, -82.0195, -82.1322, -82.2073, -82.3293, -82.4514,
        -82.6391, -82.7518, -82.8456, -82.9301, -83.1366, -83.3056, -83.4652, -83.5872, -83.5966, -83.5966, -83.4934, -83.4652, -83.315, -83.1179, -82.9301, -82.7518, -82.6767, -82.6673, -82.5734,
        -82.5077, -82.3387, -82.2261, -82.0853, -82.0102, -81.9914, -82.1322, -82.2636, -82.3012, -82.3105, -82.3012, -82.273, -82.2542, -82.3199, -82.2542, -82.2355, -82.3012, -82.4044, -82.5171,
        -82.6391, -82.6297, -82.4795, -82.4607, -82.5546, -82.6673, -82.8175, -82.9395, -83.0615, -83.2211, -83.3901, -83.4746, -83.5027, -83.5966, -83.6154, -83.6342, -83.7468, -83.8219, -83.8501,
        -83.9346, -83.9439, -83.9158, -83.8782, -83.7656, -83.7562, -83.7093, -83.6623, -83.728, -83.8125, -83.897, -84.0378, -84.1974, -84.2162, -84.2068, -84.1317, -84.1223, -84.188, -84.1974,
        -84.2537, -84.2631, -84.3382, -84.479, -84.6292, -84.7794, -84.9202, -85.0423, -85.1361, -85.2582, -85.352, -85.4178, -85.3802, -85.2863, -85.1549, -85.2018, -85.3427, -85.4178, -85.4929,
        -85.5586, -85.5867, -85.6337, -85.6618, -85.8026, -85.9528, -86.0561, -86.2345, -86.394, -86.4316, -86.4128, -86.3659, -86.3659, -86.272, -86.272, -86.2345, -86.2157, -86.1875, -86.103,
        -86.0467, -86.0467, -85.9998, -85.9153, -85.8214, -85.7933, -85.6806, -85.7088, -85.6524, -85.5867, -85.5304, -85.5304, -85.3614, -85.2863, -85.1455, -85.108, -84.9202, -84.7982, -84.7137,
        -84.5729, -84.4039, -84.188, -84.0284, -83.897, -83.7468, -83.7093, -83.6623, -83.5966, -83.4089, -83.2117, -83.0615, -82.8644, -82.6297, -82.4795, -82.2448, -82.1322, -82.1322, -82.3012,
        -82.4326, -82.5922, -82.8081, -82.9677, -83.1272, -83.2774, -83.4652, -83.5966, -83.7656, -83.8782, -84.019, -84.0566, -84.1505, -84.3194, -84.4509, -84.6104, -84.7043, -84.8264, -84.9672,
        -85.0892, -85.23, -85.3802, -85.4553, -85.5867, -85.6712, -85.8308, -85.8965, -86.1687, -86.3283, -86.4691, -86.5255, -86.6006, -86.7414, -86.8165, -86.8165, -86.8446, -86.9385, -87.0418,
        -87.1356, -87.2483, -87.3234, -87.4173, -87.5111, -87.6144, -87.6613, -87.7458, -87.9148, -88.0744, -88.1213, -88.1682, -88.2809, -88.4217, -88.6001, -88.7597, -88.9286, -89.0131, -89.1164,
        -89.229, -89.3041, -89.3417],
   y = [80.818, 80.7968, 80.7828, 80.7593, 80.7264, 80.7217, 80.7241, 80.703, 80.6795, 80.6466, 80.6232, 80.6021, 80.5903, 80.5715, 80.5528, 80.541, 80.527, 80.5246, 80.5176, 80.5082, 80.4964, 80.4918,
        80.4894, 80.48, 80.4659, 80.4519, 80.4636, 80.5082, 80.5293, 80.5528, 80.5692, 80.5903, 80.5974, 80.6255, 80.6419, 80.6419, 80.6419, 80.6654, 80.6795, 80.71, 80.7358, 80.7523, 80.7663, 80.7734,
        80.7804, 80.7757, 80.7663, 80.7828, 80.7875, 80.8039, 80.818, 80.8227, 80.8391, 80.8626, 80.886, 80.8743, 80.8461, 80.8438, 80.8367, 80.825, 80.8109, 80.7828, 80.764, 80.7264, 80.703, 80.6889,
        80.6654, 80.6466, 80.6349, 80.6185, 80.5974, 80.5974, 80.5974, 80.5974, 80.5927, 80.588, 80.5974, 80.6044, 80.6302, 80.649, 80.6772, 80.6936, 80.717, 80.7382, 80.7523, 80.7569, 80.717, 80.6865,
        80.6654, 80.6513, 80.6185, 80.5856, 80.5692, 80.5528, 80.5105, 80.5082, 80.4894, 80.4683, 80.4565, 80.4565, 80.4495, 80.4542, 80.4542, 80.4753, 80.48, 80.48, 80.4824, 80.4964, 80.5152, 80.5363,
        80.5598, 80.5762, 80.5598, 80.5246, 80.5223, 80.5152, 80.4964, 80.4894, 80.4894, 80.4988, 80.5011, 80.5152, 80.534, 80.5434, 80.5622, 80.5786, 80.5974, 80.6349, 80.6537, 80.6654, 80.6889,
        80.71, 80.7264, 80.7452, 80.7663, 80.7734, 80.7898, 80.8086, 80.825, 80.8438, 80.8579, 80.8719, 80.8837, 80.8978, 80.9118, 80.9236, 80.9236, 80.9377, 80.9517, 80.9635, 80.9705, 80.9893,
        80.9916, 80.9822, 80.9869, 80.9916, 80.9987, 81.0151, 81.0151, 81.0151, 81.0034, 81.0221, 81.0292, 81.048, 81.0526, 81.055, 81.055, 81.0785, 81.0878, 81.0996, 81.1043, 81.0972, 81.0972,
        81.0855, 81.0808, 81.0738, 81.0714, 81.0738, 81.062, 81.0573, 81.055, 81.048, 81.0503, 81.055, 81.055, 81.048, 81.0386, 81.0292, 81.0292, 81.0221, 81.0104, 81.0057, 81.0151, 81.0151, 81.0081,
        80.9916, 80.9705, 80.9541, 80.9236, 80.8837, 80.8391, 80.7898, 80.7476, 80.7147, 80.6772, 80.6349, 80.6091, 80.5833, 80.5622, 80.5528, 80.5598, 80.5598, 80.5598, 80.5692, 80.5856, 80.6138,
        80.6232, 80.6302, 80.6349, 80.6513, 80.6584, 80.6631, 80.6959, 80.7335, 80.7593, 80.7851, 80.7898, 80.7968, 80.8297, 80.8414, 80.8532, 80.8649] };
  WCL[1804] = struct {
   x = [-79.9684, -80.0247, -80.25, -80.5316, -80.8508, -81.1888, -81.2263, -81.0386, -80.7945, -80.7945, -80.8696, -80.7006, -80.4096, -80.2406, -80.5692, -80.8602, -80.8133, -80.8133, -81.0573, -81.3296,
        -81.6487, -82.0055, -82.2495, -82.3622, -82.1181, -82.0055, -81.8553, -81.6769, -81.5173, -81.4798, -81.3953, -81.292, -81.4798, -81.7896, -82.0242, -82.3059, -82.6813, -83.0005, -83.2258,
        -83.3385, -83.2634, -83.0005, -82.625, -82.4936, -82.7752, -82.963, -83.1226, -83.3478, -83.4323, -83.545, -83.5262, -83.4042, -83.1319, -82.7752, -82.503, -82.3434, -81.9867, -81.7051,
        -81.4516, -81.5361, -81.6018, -81.3765, -81.2545, -81.7145, -82.0899, -82.334, -82.6532, -82.9066, -83.1977, -83.4136, -83.6858, -83.8454, -83.8548, -83.7327, -83.6576, -83.5638, -83.2728,
        -83.1977, -83.4699, -83.7515, -83.8923, -83.8454, -83.9393, -84.1364, -84.2115, -84.3523, -84.3617, -84.4086, -84.6339, -84.8404, -85.0376, -85.1784, -85.4037, -85.4225, -85.291, -85.5445,
        -85.7135, -85.9106, -86.1171, -86.4175, -86.5865, -86.3706, -86.3049, -86.2955, -86.1547, -86.211, -86.2204, -86.4926, -86.7836, -87.0558, -87.2811, -87.3656, -87.3469, -87.0652, -86.7555,
        -86.5208, -86.1922, -85.8637, -85.5257, -85.2723, -85.4506, -85.4037, -85.7135, -86.042, -86.089, -85.9294, -86.1265, -86.4363, -86.793, -87.0652, -87.2905, -87.2811, -87.4313, -87.2624,
        -87.3656, -87.544, -87.5628, -87.6566, -87.666, -87.6754, -87.8162, -88.0134, -88.1448, -88.3138, -88.6141, -88.7456, -89.0178, -89.2431, -89.5717, -89.5623, -89.3839, -89.2619, -89.5247,
        -89.6749, -89.6749, -89.6468, -89.8063, -89.9565],
   y = [82.8761, 82.9418, 82.963, 82.9536, 82.9442, 82.8926, 82.8761, 82.8409, 82.8386, 82.8269, 82.7752, 82.7283, 82.7095, 82.686, 82.6954, 82.7001, 82.6602, 82.6438, 82.6954, 82.7635, 82.787, 82.794,
        82.7987, 82.7658, 82.7236, 82.6767, 82.6391, 82.5992, 82.5922, 82.557, 82.5241, 82.4725, 82.4584, 82.4983, 82.5452, 82.5898, 82.618, 82.625, 82.6086, 82.571, 82.5452, 82.503, 82.4913, 82.4701,
        82.4631, 82.4772, 82.4772, 82.4725, 82.4443, 82.4021, 82.3481, 82.2965, 82.266, 82.2214, 82.1885, 82.1674, 82.1369, 82.1087, 82.1087, 82.0782, 82.0454, 81.9914, 81.9585, 82.0031, 82.0383,
        82.0759, 82.1181, 82.1674, 82.2143, 82.2542, 82.266, 82.2542, 82.2167, 82.165, 82.1392, 82.0923, 82.0829, 82.0712, 82.0759, 82.0759, 82.0759, 82.1416, 82.1956, 82.2448, 82.2918, 82.3176,
        82.3176, 82.3223, 82.3364, 82.3411, 82.3152, 82.3481, 82.3739, 82.4021, 82.4302, 82.4326, 82.4373, 82.4443, 82.4373, 82.4232, 82.3974, 82.3739, 82.3481, 82.3199, 82.3012, 82.273, 82.2472,
        82.2425, 82.2167, 82.2261, 82.2261, 82.1956, 82.151, 82.1134, 82.1134, 82.0829, 82.0547, 82.0313, 82.0172, 82.0125, 81.9961, 81.9397, 81.9937, 81.9984, 81.9538, 81.8975, 81.9327, 82.0148,
        82.05, 82.0454, 82.0383, 81.9961, 81.9773, 81.921, 81.921, 81.9796, 82.0313, 82.0477, 82.05, 82.05, 82.0594, 82.0806, 82.0665, 82.097, 82.0946, 82.0688, 82.0571, 82.0031, 81.9656, 81.9397,
        81.8999, 81.8482, 81.9022, 81.9139, 81.86, 81.7849, 81.8271, 81.8529] };
  WCL[1805] = struct {
   x = [-90.0363, -89.8486, -89.7829, -89.4825, -89.398, -89.6139, -89.9706],
   y = [81.63, 81.5995, 81.5948, 81.6018, 81.5901, 81.5643, 81.5291] };
  WCL[1806] = struct {
   x = [-90.0598, -89.9002, -89.6374, -89.3839, -89.1492, -88.9709, -88.6329, -88.2574, -88.0978, -88.004, -87.7505, -87.4595, -87.3375, -87.0183, -86.6522, -86.4832, -86.624, -86.7742, -86.9808, -87.3656,
        -87.713, -87.9383, -88.126, -88.4827, -88.6329, -88.9427, -89.3933, -89.75, -89.75, -89.5529, -89.2619, -89.0366, -88.7831, -88.999, -89.3651, -89.7031, -89.8908, -89.9471, -89.919, -89.9284,
        -89.7688, -89.5153, -89.2056, -88.9803, -88.7643, -88.5015, -88.2856, -88.0978, -87.7505, -87.4595, -87.2342, -87.1122, -86.8681, -86.6522, -86.4269, -86.211, -85.9857, -85.8918, -85.6947,
        -85.3286, -85, -84.7747, -84.5025, -84.5588, -84.7653, -85.1408, -85.5069, -85.7792, -85.7698, -86.1828, -86.5395, -86.8493, -87.1216, -87.2905, -87.5346, -87.8444, -88.1636, -88.417, -88.5766,
        -88.8864, -89.1492, -89.2056, -89.3651, -89.337],
   y = [81.3577, 81.3789, 81.4117, 81.4281, 81.461, 81.4868, 81.515, 81.5361, 81.5596, 81.5384, 81.5361, 81.5267, 81.5032, 81.5079, 81.5197, 81.5056, 81.4892, 81.454, 81.4704, 81.4751, 81.4962, 81.5056,
        81.4985, 81.4774, 81.4469, 81.4047, 81.3319, 81.3108, 81.285, 81.2662, 81.2427, 81.231, 81.2357, 81.2005, 81.2005, 81.1935, 81.1817, 81.1395, 81.0949, 81.0573, 81.0127, 80.9916, 80.9822,
        81.0081, 81.0221, 81.0174, 81.0221, 81.0292, 81.0292, 81.0339, 81.0268, 81.0339, 81.048, 81.0691, 81.0808, 81.1231, 81.1583, 81.1864, 81.1888, 81.2122, 81.2427, 81.2592, 81.2662, 81.2357,
        81.1958, 81.1629, 81.116, 81.0667, 81.0409, 81.0127, 80.9705, 80.9541, 80.9447, 80.9377, 80.9423, 80.9635, 80.9635, 80.9611, 80.933, 80.9259, 80.9095, 80.8813, 80.8719, 80.8649] };
  WCL[1807] = struct {
   x = [-99.1233, -99.0857, -99.0576, -99.0106, -98.9543, -98.9074, -98.8323, -98.729, -98.6351, -98.4755, -98.3441],
   y = [80.0764, 80.0811, 80.074, 80.0552, 80.0412, 80.0365, 80.0459, 80.0599, 80.0576, 80.0576, 80.0271] };
  WCL[1808] = struct {
   x = [-99.5316, -99.4847, -99.4002, -99.3439, -99.2594, -99.1186],
   y = [79.959, 80.0036, 80.0247, 80.0529, 80.067, 80.0764] };
  WCL[1809] = struct {
   x = [-95.1876, -94.9999, -94.9529, -94.9248, -94.7652, -94.6338, -94.5117, -94.3709, -94.1644, -93.9861, -93.8359, -93.7326, -93.6012, -93.5824, -93.6481, -93.8077, -93.9579, -94.1081, -94.1363, -94.155,
        -94.2113, -94.2207, -94.2583, -94.1363, -93.9767, -93.8171, -93.7138, -93.4979, -93.5824, -93.6669, -93.8171, -93.8452, -93.8828, -94.0612, -94.1832, -94.3709, -94.4742, -94.5775, -94.615,
        -94.6526, -94.615, -94.446, -94.446, -94.6526, -94.8121, -94.9154, -94.9717, -94.9717, -94.8872, -94.8028, -94.7652, -94.906, -94.9529, -94.9529, -94.9436, -94.9717, -94.9248, -94.8403,
        -94.6995, -94.6056, -94.5962, -94.5399, -94.3709, -94.1738, -94.1738, -94.1269, -94.155, -94.0893, -94.033, -93.9954, -93.9203, -93.8171, -93.8077, -93.7983, -93.7326, -93.6012, -93.5824,
        -93.6669, -93.7889, -93.7889, -93.9203, -93.9954, -93.8734, -93.7138, -93.5918, -93.4416, -93.3665, -93.3383, -93.2163, -93.113, -93.0285, -93.0004, -92.9534, -92.8032, -92.7469, -92.7375,
        -92.7, -92.6531, -92.6718, -92.6437, -92.6812, -92.7751, -92.8502, -92.9253, -92.9253, -92.9628, -93.113, -93.2257, -93.3008, -93.4698, -93.6106, -93.7138, -93.8077, -93.8546, -93.9391,
        -94.0705, -94.2395, -94.3428, -94.3991, -94.446, -94.4366, -94.4366, -94.3991, -94.4085, -94.4648, -94.5305, -94.6056, -94.6056, -94.5117, -94.3709, -94.2489, -94.1269, -93.9391, -93.4979,
        -93.404, -93.4885, -93.5542, -93.404, -93.2351, -93.1036, -92.9534, -92.8971, -92.8596, -92.7469, -92.4278, -92.2306, -92.0804, -92.0147, -91.7894, -91.658, -91.5453, -91.5641, -91.5453,
        -91.6017, -91.5829, -91.5078, -91.4045, -91.3482, -91.3294, -91.245, -91.2074, -91.1229, -91.0666, -91.076, -91.029, -90.9633, -90.9633, -90.9727, -90.9352, -90.7662, -90.7099, -90.6348,
        -90.5691, -90.616, -90.6911, -90.7099, -90.7005, -90.6911, -90.6535, -90.616, -90.5127, -90.4752, -90.4376, -90.2405, -90.034],
   y = [80.5786, 80.5739, 80.5598, 80.5598, 80.5715, 80.5739, 80.5434, 80.5387, 80.541, 80.5434, 80.5363, 80.5293, 80.5152, 80.5504, 80.5786, 80.5762, 80.5809, 80.5856, 80.5856, 80.5856, 80.5903, 80.6161,
        80.649, 80.6584, 80.6513, 80.6513, 80.6701, 80.6748, 80.6912, 80.703, 80.6889, 80.7053, 80.717, 80.7124, 80.6959, 80.6959, 80.717, 80.717, 80.7311, 80.7429, 80.7616, 80.7757, 80.7875, 80.7734,
        80.7616, 80.7663, 80.7875, 80.8156, 80.8227, 80.8438, 80.8555, 80.8532, 80.8649, 80.8884, 80.9095, 80.9353, 80.9588, 80.9775, 80.9775, 80.9916, 81.0127, 80.9963, 80.9916, 80.9846, 80.9611,
        80.9517, 80.9377, 80.9306, 80.94, 80.9588, 80.9611, 80.9588, 80.9705, 80.9799, 80.9916, 80.9822, 80.9963, 81.0104, 81.0104, 81.0268, 81.0339, 81.0644, 81.0433, 81.0339, 81.0503, 81.048,
        81.0456, 81.0456, 81.0456, 81.0362, 81.0433, 81.0456, 81.055, 81.0691, 81.0738, 81.0902, 81.1066, 81.1231, 81.1348, 81.1559, 81.1629, 81.1442, 81.1465, 81.1606, 81.177, 81.1817, 81.1817,
        81.1888, 81.1676, 81.1559, 81.1465, 81.1442, 81.1465, 81.1629, 81.1465, 81.1559, 81.1653, 81.1747, 81.2005, 81.2169, 81.2357, 81.2474, 81.2615, 81.2709, 81.2709, 81.2709, 81.2779, 81.2944,
        81.3178, 81.3272, 81.3413, 81.3178, 81.3014, 81.2709, 81.2803, 81.3225, 81.3437, 81.339, 81.346, 81.3296, 81.3225, 81.3131, 81.3108, 81.2897, 81.2709, 81.2545, 81.2334, 81.1958, 81.17, 81.1254,
        81.1043, 81.0738, 81.0362, 81.0292, 81.0174, 81.001, 80.9658, 80.9259, 80.8954, 80.8672, 80.8508, 80.8438, 80.8109, 80.8015, 80.7968, 80.7616, 80.7476, 80.7124, 80.6936, 80.6865, 80.656,
        80.6161, 80.595, 80.5762, 80.5293, 80.5035, 80.4964, 80.4964, 80.5199, 80.5528, 80.5551, 80.5669, 80.5363, 80.527, 80.527] };
  WCL[1810] = struct {
   x = [-96.0794, -96.0419, -95.8166, -95.7039, -95.9011, -96.1076, -96.1921, -96.2109, -96.2202, -96.2015, -96.07, -95.948, -95.7978, -95.6382, -95.5068, -95.4505, -95.3191, -95.3003, -95.1125, -95.1125,
        -94.9999, -94.9248, -94.8497, -94.8215],
   y = [79.9731, 80.0013, 80.0224, 80.0388, 80.0459, 80.0341, 80.0787, 80.1233, 80.135, 80.1421, 80.1515, 80.1468, 80.1327, 80.1186, 80.1022, 80.0904, 80.0693, 80.0365, 80.0271, 80.0271, 80.013, 80.006,
        80.013, 80.0036] };
  WCL[1811] = struct {
   x = [-94.7253, -94.5939, -94.4718, -94.3498],
   y = [79.9895, 80.0083, 80.0177, 80.0036] };
  WCL[1812] = struct {
   x = [-94.0729, -94.1949, -94.2982, -94.2982, -94.1761, -94.0447, -93.8851, -93.7913, -93.6692, -93.5941, -93.6786, -93.81, -93.9696, -94.0541, -94.0917, -94.1761, -94.2888, -94.3827, -94.5516, -94.6267,
        -94.8239, -94.9553, -95.0679, -95.0022, -95.0679, -95.1055, -95.1994, -95.2463, -95.3308, -95.3214, -95.1994, -95.1149, -95.3214, -95.1994, -95.1149, -95.1149, -95.0022, -94.9835, -94.8426,
        -94.7488, -94.6643, -94.6174, -94.7206, -94.8051, -94.9271, -94.9741, -94.9928, -95.1337, -95.0022, -95.1337, -95.1618, -95.2369, -95.312, -95.481, -95.603, -95.603, -95.7063, -95.8753,
        -96.063, -96.0818, -96.063, -96.1287, -96.2789, -96.2977, -96.2038, -96.0912, -95.941, -95.8377, -95.6969, -95.4434, -95.2557, -95.0022, -94.9271, -94.8614, -94.8708, -94.9271, -94.9459,
        -94.9741, -94.9928, -95.4434, -95.0022, -95.0773, -95.1806, -95.3026, -95.4247, -95.5279, -95.6969, -95.8002, -95.7626, -95.6875, -95.6875, -95.6781, -95.6687, -95.7438, -95.6875, -95.5749,
        -95.4341, -95.3777, -95.3402, -95.2932, -95.19],
   y = [79.9708, 80.0083, 80.0341, 80.067, 80.0975, 80.1092, 80.1186, 80.121, 80.128, 80.1303, 80.1468, 80.1421, 80.1397, 80.1585, 80.1867, 80.189, 80.1515, 80.121, 80.1116, 80.0881, 80.0717, 80.0904,
        80.1069, 80.0951, 80.1069, 80.1116, 80.1139, 80.1327, 80.1444, 80.1538, 80.1796, 80.182, 80.1538, 80.1796, 80.182, 80.182, 80.1679, 80.1679, 80.1749, 80.1679, 80.196, 80.2266, 80.2172, 80.2148,
        80.2125, 80.2007, 80.2031, 80.2336, 80.2078, 80.2336, 80.243, 80.2359, 80.2007, 80.1843, 80.1796, 80.2172, 80.2219, 80.2406, 80.243, 80.2735, 80.297, 80.3134, 80.3251, 80.3392, 80.3416, 80.365,
        80.3486, 80.3744, 80.3744, 80.3674, 80.365, 80.3322, 80.3369, 80.3439, 80.3674, 80.365, 80.358, 80.3697, 80.365, 80.3674, 80.3627, 80.3838, 80.3885, 80.4002, 80.412, 80.4307, 80.4425, 80.4519,
        80.4753, 80.48, 80.5011, 80.5199, 80.5363, 80.5692, 80.5762, 80.5762, 80.5739, 80.5762, 80.5762, 80.5762, 80.5786] };
  WCL[1813] = struct {
   x = [-89.9495, -90.0246, -90.1936, -90.4283, -90.6535, -90.8413, -91.1041, -91.1699, -91.2919, -91.4515, -91.6298, -91.7706, -91.827, -91.9115, -91.9584, -92.0241, -91.8739, -91.6674, -91.5078, -91.3951,
        -91.3951, -91.3388, -91.0666, -90.8976, -90.8788, -90.8225, -90.6911, -90.616, -90.4095, -90.2123, -90.156, -90.034, -90.034],
   y = [81.8529, 81.9022, 81.9045, 81.8975, 81.8905, 81.8576, 81.8388, 81.7919, 81.7708, 81.7825, 81.7497, 81.7332, 81.6863, 81.6652, 81.644, 81.6159, 81.6159, 81.6042, 81.5854, 81.5525, 81.5314, 81.5173,
        81.5408, 81.5619, 81.5408, 81.5689, 81.5995, 81.63, 81.637, 81.6112, 81.6558, 81.6699, 81.63] };
  WCL[1814] = struct {
   x = [-89.9706, -90.393, -90.6559, -90.4494, -90.3179, -90.2898, -90.0645],
   y = [81.5291, 81.4586, 81.4094, 81.3835, 81.3624, 81.3343, 81.3577] };
  WCL[1815] = struct {
   x = [91.2966, 91.2966, 91.2121, 91.0901, 91.0243, 91.0619, 91.2309, 91.3154, 91.3998, 91.4562, 91.5782, 91.6533, 91.7659, 91.841, 91.9912, 92.1414, 92.3667],
   y = [80.0013, 80.0083, 80.0271, 80.0459, 80.0599, 80.0623, 80.0459, 80.0482, 80.0459, 80.0529, 80.0552, 80.0623, 80.0717, 80.0459, 80.0412, 80.0177, 80.0106] };
  WCL[1816] = struct {
   x = [91.2755, 91.3036],
   y = [79.9989, 80.0013] };
  WCL[1817] = struct {
   x = [90.9117, 90.9305, 91.0525, 91.2027, 91.4186, 91.5594, 91.6533, 91.6533, 91.5125, 91.3341, 91.1276, 90.9492, 90.7709, 90.6019, 90.3954, 90.2076, 90.0481],
   y = [81.224, 81.224, 81.2122, 81.2075, 81.1981, 81.1794, 81.1583, 81.1207, 81.0902, 81.0714, 81.055, 81.0409, 81.048, 81.048, 81.055, 81.0738, 81.0972] };
  WCL[1818] = struct {
   x = [89.973, 90.1232, 90.3203, 90.5456, 90.7427, 90.8742, 90.9117],
   y = [81.1676, 81.1888, 81.1888, 81.1981, 81.2028, 81.2122, 81.224] };
  WCL[1819] = struct {
   x = [94.8708, 94.9084, 95.021, 95.1055, 95.0679, 95.021, 95.2275, 95.4434, 95.5655, 95.6687, 95.8189, 95.9504, 96.0442, 96.1099, 96.2038, 96.3071, 96.3634, 96.4197, 96.5042, 96.5981, 96.6826, 96.8609,
        96.9642, 97.0581, 97.1895, 97.3021, 97.3772, 97.4617, 97.4523, 97.4523, 97.5931, 97.7058, 97.7809, 97.6495, 97.4336, 97.3397, 97.1238, 97.0487, 97.0581, 97.0393, 97.0205, 96.9642, 97.0487,
        97.2176, 97.2552, 97.2176, 97.2552, 97.3303, 97.3303, 97.2646, 97.105, 96.9266, 96.7107, 96.5418, 96.2977, 96.1006, 95.8283, 95.5749, 95.4247, 95.2088, 94.8802, 94.7675, 94.5329, 94.2419,
        94.0353],
   y = [81.1019, 81.109, 81.1301, 81.1536, 81.177, 81.2099, 81.224, 81.231, 81.2521, 81.224, 81.2028, 81.1747, 81.1348, 81.0878, 81.0667, 81.048, 81.0081, 80.9658, 80.9353, 80.9165, 80.8766, 80.8555,
        80.8367, 80.8227, 80.8015, 80.8015, 80.7663, 80.7358, 80.703, 80.6842, 80.6889, 80.6959, 80.6842, 80.6584, 80.6443, 80.6138, 80.6114, 80.5669, 80.5387, 80.4918, 80.4472, 80.4002, 80.3979,
        80.4143, 80.3838, 80.3462, 80.3392, 80.3322, 80.304, 80.2711, 80.2406, 80.25, 80.2524, 80.2406, 80.2453, 80.2406, 80.2242, 80.2054, 80.1702, 80.1632, 80.121, 80.0717, 80.0529, 80.0435, 80.0036] };
  WCL[1820] = struct {
   x = [93.8335, 93.6645, 93.4674, 93.3548, 93.1952, 93.045, 92.9699, 92.7446, 92.4817, 92.2846, 92.0781, 92.1626, 92.2752, 92.0311, 91.8622, 91.6463, 91.5618, 91.374, 91.3834, 91.4961, 91.5618, 91.7026,
        91.9373, 92.0124, 91.8246, 91.7777, 91.9091, 92.125, 92.3128, 92.4817, 92.6413, 92.6319, 92.754, 92.8103, 92.7446, 92.6507, 92.5193, 92.4536, 92.4723, 92.5662, 92.6883, 92.8103, 93.0168,
        93.1107, 93.1295, 93.2233, 93.3641, 93.5519, 93.7209, 93.918, 94.0213, 94.209, 94.3967, 94.5751, 94.7065, 94.838, 94.8661, 94.8286, 94.8567, 94.8661],
   y = [79.9966, 80.0177, 80.0177, 80.0459, 80.0623, 80.0811, 80.1303, 80.1468, 80.1397, 80.1608, 80.189, 80.2336, 80.2594, 80.2594, 80.3017, 80.2993, 80.2594, 80.2477, 80.2946, 80.3228, 80.3556, 80.358,
        80.3721, 80.3955, 80.3932, 80.426, 80.4401, 80.4448, 80.4636, 80.4941, 80.5035, 80.5363, 80.5739, 80.6021, 80.6208, 80.6419, 80.6373, 80.6678, 80.7077, 80.7593, 80.7804, 80.8086, 80.8367,
        80.8719, 80.9071, 80.9095, 80.9095, 80.9236, 80.9377, 80.9705, 80.9916, 80.9916, 80.9869, 80.9846, 80.9846, 81.0151, 81.0292, 81.0691, 81.0996, 81.1019] };
  WCL[1821] = struct {
   x = [94.8802, 94.9177, 94.9741, 94.9271, 94.9271, 95.0679, 95.2463, 95.3777, 95.3402, 95.4998, 95.6406, 95.8565, 95.941, 96.1569, 96.2883, 96.4385, 96.7577, 96.9548, 97.1707, 97.2176, 97.4336, 97.4711,
        97.5462, 97.5838, 97.6589, 97.7339, 97.9029, 98.0625],
   y = [80.0153, 80.0177, 80.0294, 80.0623, 80.0811, 80.0599, 80.0294, 80.0341, 80.0693, 80.0857, 80.1045, 80.1022, 80.1186, 80.1139, 80.1163, 80.1303, 80.1421, 80.1397, 80.1397, 80.1585, 80.1632, 80.1397,
        80.1256, 80.1116, 80.0928, 80.0717, 80.0646, 80.02] };
  WCL[1822] = struct {
   x = [98.4544, 98.3606, 98.3418, 98.3793, 98.4638, 98.5577, 98.6516, 98.7924, 98.905, 98.8675],
   y = [79.9895, 80.0083, 80.0271, 80.0412, 80.0505, 80.0388, 80.0318, 80.0388, 80.0318, 80.0013] };
  WCL[1823] = struct {
   x = [99.1045, 99.2829],
   y = [79.9919, 80.0106] };
  WCL[1824] = struct {
   x = [94.5728, 94.6572, 94.7793, 94.8544, 94.8732],
   y = [79.9848, 80.0083, 80.0177, 80.0153, 80.0153] };
  WCL[1825] = struct {
   x = [79.9825, 80.0482, 80.1608, 80.2923, 80.4049, 80.6396, 80.7053, 80.7992, 80.8743, 80.9212, 80.9306, 80.9306, 80.9682, 81.0339, 80.9869, 80.8461, 80.6772, 80.5176, 80.2735, 80.0764],
   y = [80.8626, 80.879, 80.8907, 80.8813, 80.8719, 80.8461, 80.8297, 80.8227, 80.8156, 80.8109, 80.8086, 80.8086, 80.8086, 80.7945, 80.7734, 80.7523, 80.7476, 80.7311, 80.7241, 80.7194] };
  WCL[1826] = struct {
   x = [90.0504, 89.9471, 89.9753],
   y = [81.0972, 81.1277, 81.1676] };
  WCL[1827] = struct {
   x = [79.9262, 79.9449, 79.9825],
   y = [80.8485, 80.8508, 80.8626] };
  WCL[1828] = struct {
   x = [80.0693, 79.8722, 79.722, 79.6281, 79.6281, 79.7502, 79.8346, 79.9003, 79.9191],
   y = [80.7194, 80.7217, 80.7476, 80.771, 80.8039, 80.8273, 80.8438, 80.8508, 80.8485] };
  WCL[1829] = struct {
   x = [60.5061, 60.4873, 60.4498, 60.3559, 60.2714, 60.2526, 60.1588, 60.0931],
   y = [80.7898, 80.7945, 80.7921, 80.7898, 80.7945, 80.8062, 80.8109, 80.8062] };
  WCL[1830] = struct {
   x = [59.9335, 60.018, 60.0743, 60.1494, 60.1963, 60.2339, 60.262, 60.3559, 60.4122, 60.4873, 60.5812, 60.5906, 60.5906, 60.6, 60.6938, 60.8159, 60.9191, 60.9849, 61.0693, 61.1538, 61.2759, 61.2853,
        61.351, 61.473, 61.5387, 61.5387, 61.5105, 61.4824, 61.4824, 61.4824, 61.473, 61.4167, 61.3979, 61.3604, 61.2665, 61.2008, 61.1632, 61.1726, 61.2008, 61.2102, 61.2571, 61.3604, 61.4073,
        61.5387, 61.6983, 61.8109, 61.9705, 62.102, 62.2146, 62.2991, 62.2897, 62.2803, 62.2897, 62.2803, 62.2052, 62.1864, 62.1771, 62.1677, 62.1583, 62.1677, 62.1489, 62.0926, 62.0832, 62.0832,
        62.0644, 61.9987, 61.933, 61.8391, 61.7265, 61.6983, 61.6138, 61.5199, 61.473, 61.3979, 61.3228, 61.2571, 61.2102, 61.1069, 61.0224, 60.9567, 60.8816, 60.844, 60.7971, 60.7126, 60.6375,
        60.6094, 60.5718, 60.5061],
   y = [80.3768, 80.3815, 80.3744, 80.365, 80.365, 80.3627, 80.3791, 80.3838, 80.3721, 80.3744, 80.3744, 80.3721, 80.3697, 80.3627, 80.3533, 80.3509, 80.3627, 80.3462, 80.3392, 80.3462, 80.3509, 80.3674,
        80.3674, 80.3697, 80.3768, 80.4026, 80.4307, 80.4612, 80.4824, 80.4964, 80.5129, 80.5316, 80.5457, 80.5622, 80.5739, 80.5833, 80.5997, 80.6232, 80.6279, 80.6419, 80.6513, 80.6607, 80.6701,
        80.6701, 80.6818, 80.6748, 80.6678, 80.6701, 80.6795, 80.6889, 80.6983, 80.717, 80.7358, 80.7476, 80.7569, 80.7687, 80.7898, 80.7945, 80.8015, 80.8039, 80.8109, 80.8203, 80.825, 80.8414,
        80.8626, 80.8672, 80.879, 80.8813, 80.8743, 80.8532, 80.8367, 80.825, 80.8086, 80.7945, 80.7898, 80.7968, 80.8062, 80.8039, 80.8039, 80.8109, 80.8086, 80.8015, 80.7875, 80.7781, 80.7781,
        80.7851, 80.7898, 80.7898] };
  WCL[1831] = struct {
   x = [60.9802, 60.9802, 60.9708, 60.9614, 60.9614, 60.9051, 60.783, 60.661, 60.5483, 60.492, 60.4263, 60.3794, 60.2198, 60.1729, 60.1165, 60.1916, 60.2573, 60.2949, 60.3981, 60.5202, 60.661, 60.7924,
        60.8863, 60.9896, 61.121, 61.2993, 61.4214, 61.4871, 61.5152, 61.534, 61.4683, 61.3275, 61.3275, 61.2993, 61.2993, 61.365, 61.4401, 61.4308, 61.2899, 61.1491, 61.0553, 61.0177, 60.9802],
   y = [81.0714, 81.0738, 81.0714, 81.055, 81.0362, 81.0409, 81.0339, 81.0268, 81.0174, 81.0268, 81.0221, 81.0057, 80.9916, 80.9658, 80.9377, 80.9283, 80.9142, 80.8954, 80.8954, 80.8766, 80.8719, 80.8766,
        80.8555, 80.8602, 80.8696, 80.8907, 80.9118, 80.9259, 80.94, 80.9588, 80.9729, 80.9799, 80.9916, 81.0034, 81.0104, 81.0034, 81.0127, 81.0268, 81.0315, 81.0433, 81.0644, 81.0738, 81.0714] };
  WCL[1832] = struct {
   x = [63.5288, 63.5101, 63.4725, 63.435, 63.3129, 63.2378, 63.1158, 63.0125, 62.9093, 62.8436, 62.7872, 62.7215, 62.6464, 62.6183, 62.6558, 62.7215, 62.7309, 62.7309, 62.7497, 62.8342, 62.8436, 62.7966,
        62.7778, 62.6652, 62.5619, 62.4774, 62.4587, 62.4962, 62.6183, 62.7966, 62.8436, 62.975, 63.0688, 63.0688, 63.1815, 63.3035, 63.3411, 63.3786, 63.4725, 63.557, 63.5758, 63.6039, 63.7072,
        63.7541, 63.7635, 63.8668, 63.9419, 63.9419, 64.017, 64.0639, 64.1202, 64.2047, 64.3268, 64.43, 64.552, 64.6271, 64.7492, 64.8524, 64.9275, 65.012, 65.0308, 65.059, 65.1059, 65.1998, 65.2561,
        65.3124, 65.3781, 65.4251, 65.4908, 65.5096, 65.4438, 65.3781, 65.3124, 65.1998, 65.1435, 65.1435, 65.1341, 65.1341, 65.0496, 64.9651, 64.9463, 64.8712, 64.8337, 64.8243, 64.8149, 64.7492,
        64.7773, 64.8149, 64.89, 64.9182, 64.9182, 64.8994, 64.8994, 64.8994, 64.9463, 64.9651, 64.9182, 64.8712, 64.8055, 64.7022, 64.6741, 64.6178, 64.5051, 64.3925, 64.308, 64.2329, 64.139, 64.0639,
        63.9137, 63.8011, 63.6884, 63.6227, 63.5664, 63.5288],
   y = [81.2873, 81.292, 81.2897, 81.2756, 81.2545, 81.238, 81.224, 81.1958, 81.1723, 81.1536, 81.1465, 81.1231, 81.1066, 81.0949, 81.1043, 81.1019, 81.0878, 81.0785, 81.0691, 81.055, 81.0386, 81.0245,
        80.9963, 80.9658, 80.9588, 80.9377, 80.9189, 80.9048, 80.8907, 80.879, 80.8532, 80.8414, 80.8273, 80.7992, 80.7828, 80.7687, 80.7523, 80.7358, 80.7546, 80.7757, 80.7945, 80.8109, 80.8227,
        80.8391, 80.8626, 80.8719, 80.8766, 80.8954, 80.9118, 80.9259, 80.933, 80.9541, 80.9564, 80.9588, 80.9611, 80.9588, 80.9729, 80.9752, 80.9729, 80.9705, 80.9611, 80.9494, 80.9541, 80.9635,
        80.9611, 80.9588, 80.9588, 80.9635, 80.9588, 80.9658, 80.9752, 80.9987, 81.0151, 81.0245, 81.0433, 81.062, 81.0808, 81.0996, 81.0996, 81.1019, 81.1231, 81.1301, 81.1465, 81.1489, 81.1583,
        81.1723, 81.1911, 81.2099, 81.224, 81.238, 81.2568, 81.2779, 81.292, 81.3038, 81.3225, 81.3366, 81.3483, 81.3624, 81.3695, 81.3624, 81.3789, 81.3859, 81.3859, 81.3789, 81.3789, 81.3859,
        81.3789, 81.3577, 81.3343, 81.3202, 81.3014, 81.2897, 81.285, 81.2873] };
  WCL[1833] = struct {
   x = [63.5382, 63.6133, 63.6321, 63.6696, 63.6884, 63.6321, 63.557, 63.4068, 63.2566, 63.1627, 63.1252, 62.975, 62.8529, 62.806, 62.7121, 62.5995, 62.5244, 62.4587, 62.4117, 62.4211, 62.4211, 62.393,
        62.4117, 62.4211, 62.4681, 62.5525, 62.5995, 62.6276, 62.6746, 62.7309, 62.7778, 62.8436, 62.9093, 62.9374, 62.975, 63.0595, 63.1627, 63.2003, 63.266, 63.3599, 63.4068, 63.4443, 63.5007,
        63.5382],
   y = [81.6933, 81.6957, 81.6816, 81.6652, 81.6511, 81.6323, 81.6159, 81.6253, 81.6112, 81.6159, 81.6347, 81.637, 81.6581, 81.6769, 81.6769, 81.6652, 81.6675, 81.6675, 81.6722, 81.6863, 81.691, 81.7004,
        81.7121, 81.7215, 81.7215, 81.7285, 81.7121, 81.7004, 81.6957, 81.7074, 81.7262, 81.7215, 81.7168, 81.7309, 81.745, 81.7543, 81.7497, 81.7356, 81.7356, 81.745, 81.7356, 81.7215, 81.7098,
        81.6933] };
  WCL[1834] = struct {
   x = [50.8442, 50.8348, 50.7691, 50.694, 50.3748, 50.2528, 50.2997, 50.3936, 50.3936, 50.3748, 50.3373, 50.2997, 50.2434, 50.2246, 50.1495, 50.065],
   y = [80.8719, 80.8719, 80.8649, 80.8391, 80.8227, 80.8203, 80.8438, 80.8626, 80.879, 80.8907, 80.8696, 80.8649, 80.8602, 80.8696, 80.8579, 80.8532] };
  WCL[1835] = struct {
   x = [49.997, 50.0064],
   y = [80.8532, 80.8649] };
  WCL[1836] = struct {
   x = [49.9289, 50.0134, 50.0697, 50.0979, 50.051, 50.0322],
   y = [80.8062, 80.7875, 80.771, 80.7593, 80.7523, 80.7311] };
  WCL[1837] = struct {
   x = [49.9641, 50.0392, 50.1143, 50.058, 50.0674, 50.1613, 50.2176, 50.2457, 50.227, 50.2551, 50.2927, 50.2833, 50.2833, 50.18, 50.0768],
   y = [80.3181, 80.3134, 80.3204, 80.3392, 80.3392, 80.3298, 80.3462, 80.358, 80.3744, 80.3791, 80.3885, 80.4073, 80.426, 80.4213, 80.4378] };
  WCL[1838] = struct {
   x = [49.9641, 50.0674, 50.0674, 50.1425, 50.2645, 50.2833, 50.2927, 50.3772, 50.4804, 50.5837, 50.6963, 50.7433, 50.8184, 50.8841, 50.9873, 51.0437, 51.0437, 51.1094, 51.2314, 51.3065, 51.3628, 51.3722,
        51.4004, 51.4943, 51.513, 51.513, 51.5694, 51.5787, 51.5787, 51.5787, 51.5318, 51.4661, 51.4192, 51.391, 51.3628, 51.2971, 51.2314, 51.1657, 51.0624, 50.9592, 50.8935, 50.8371, 50.7902,
        50.6963, 50.5649, 50.4523, 50.349, 50.2927, 50.2833, 50.3021, 50.3115, 50.349, 50.4241, 50.5555, 50.64, 50.7057, 50.8184, 50.8935, 50.8841, 50.8841, 50.931, 50.9592, 50.978, 50.9592, 50.9592,
        50.8841, 50.8653, 50.8371],
   y = [80.4847, 80.4964, 80.5058, 80.5082, 80.5105, 80.527, 80.5293, 80.527, 80.5223, 80.5129, 80.5058, 80.5176, 80.5129, 80.4988, 80.4964, 80.4964, 80.5152, 80.527, 80.5457, 80.5316, 80.5316, 80.5481,
        80.5622, 80.5762, 80.5997, 80.6161, 80.6302, 80.649, 80.6631, 80.6818, 80.6912, 80.6818, 80.6912, 80.7147, 80.7264, 80.7335, 80.7335, 80.7194, 80.7241, 80.7264, 80.717, 80.71, 80.6983, 80.6983,
        80.6842, 80.6748, 80.6748, 80.6865, 80.703, 80.7147, 80.7382, 80.7593, 80.7616, 80.764, 80.7875, 80.7898, 80.7851, 80.8039, 80.8227, 80.8367, 80.8532, 80.8672, 80.8743, 80.8837, 80.886, 80.886,
        80.8766, 80.8719] };
  WCL[1839] = struct {
   x = [53.0291, 53.0666, 53.0854, 53.1042, 53.1136, 53.2262, 53.3576, 53.3764, 53.4703, 53.5736, 53.5736, 53.6393, 53.7144, 53.6768, 53.658, 53.6111, 53.5548, 53.4609, 53.3858, 53.4609, 53.4609, 53.414,
        53.3576, 53.2732, 53.1699, 53.1323, 53.076, 53.0385, 52.9634, 52.9446, 52.954, 52.8977, 52.832, 52.7756, 52.6911, 52.6536, 52.6724, 52.6348, 52.5409, 52.5034, 52.4377, 52.3814, 52.372, 52.3438,
        52.3532, 52.4658, 52.5597, 52.6536, 52.7944, 52.8132, 52.8507, 52.8507, 52.785, 52.785, 52.832, 52.832, 52.8226, 52.8789, 52.9352, 52.9728, 52.9821, 53.0291],
   y = [80.3533, 80.358, 80.365, 80.3486, 80.3392, 80.3345, 80.3322, 80.3087, 80.2852, 80.2641, 80.2453, 80.2406, 80.2242, 80.196, 80.1655, 80.1655, 80.1726, 80.1702, 80.1608, 80.1421, 80.1139, 80.1163,
        80.1303, 80.128, 80.1139, 80.1116, 80.1163, 80.1327, 80.135, 80.1233, 80.1045, 80.1139, 80.121, 80.121, 80.1163, 80.1256, 80.1468, 80.1397, 80.1421, 80.1632, 80.1749, 80.1867, 80.1984, 80.2054,
        80.2125, 80.2078, 80.2289, 80.2359, 80.2453, 80.2618, 80.2876, 80.297, 80.297, 80.3064, 80.3181, 80.3298, 80.3392, 80.3416, 80.3322, 80.3369, 80.3509, 80.3533] };
  WCL[1840] = struct {
   x = [56.7676, 56.7582, 56.7394, 56.5798, 56.5986, 56.6362, 56.6268, 56.5329, 56.4766, 56.4109, 56.2982, 56.2044, 56.1574, 56.0729, 55.9884, 55.9227, 55.8664, 55.857, 55.9133, 55.9133, 55.9321, 55.9227,
        55.8852, 55.857, 55.8664, 55.8664, 55.8101, 55.7819, 55.8007, 55.8758, 55.9415, 55.9978, 56.0448, 56.1386, 56.195, 56.2607, 56.2982, 56.3264, 56.4203, 56.4954, 56.5517, 56.5798, 56.5986,
        56.6268, 56.7113, 56.7864, 56.8896, 56.9366, 56.9553, 56.9178, 56.8709, 56.8615, 56.8051, 56.8802, 56.9178, 56.9366, 56.9835, 57.0023, 56.9929, 57.0117, 57.068, 57.0586, 57.1055, 57.1055,
        57.0774, 57.0492, 56.9366, 56.8802, 56.8333, 56.8051, 56.7676],
   y = [80.3275, 80.3369, 80.3298, 80.3298, 80.3064, 80.2923, 80.2899, 80.2805, 80.2876, 80.2946, 80.2876, 80.2829, 80.2899, 80.2899, 80.2876, 80.2829, 80.2758, 80.2547, 80.2359, 80.2078, 80.196, 80.182,
        80.1749, 80.1608, 80.1233, 80.1116, 80.0928, 80.074, 80.0599, 80.0646, 80.067, 80.0717, 80.0576, 80.0529, 80.0623, 80.0717, 80.0646, 80.0459, 80.0388, 80.0388, 80.0505, 80.0552, 80.0552,
        80.0365, 80.0341, 80.0412, 80.0365, 80.0388, 80.0552, 80.0693, 80.0951, 80.1163, 80.128, 80.1327, 80.1468, 80.1749, 80.189, 80.2054, 80.2148, 80.2172, 80.2195, 80.2406, 80.2618, 80.2782,
        80.297, 80.304, 80.311, 80.3275, 80.3345, 80.3322, 80.3275] };
  WCL[1841] = struct {
   x = [58.1029, 58.0842, 58.0278, 57.9621, 57.9246, 57.8683, 57.8119, 57.765, 57.6711, 57.5585, 57.4458, 57.3707, 57.352, 57.352, 57.2862, 57.2205, 57.1736, 57.1079, 57.0985, 57.0985, 57.1267, 57.2018,
        57.2205, 57.2018, 57.2018, 57.2205, 57.2675, 57.3144, 57.3144, 57.305, 57.2581, 57.2581, 57.2299, 57.2205, 57.2581, 57.2393, 57.2205, 57.136, 57.1173, 57.1079, 57.1454, 57.2487, 57.2675,
        57.3238, 57.4083, 57.4083, 57.4834, 57.5679, 57.6711, 57.7368, 57.7932, 57.765, 57.7556, 57.7087, 57.6711, 57.6711, 57.6148, 57.6899, 57.8119, 57.9621, 58.0278, 58.1687, 58.2062, 58.2719,
        58.3752, 58.4597, 58.5535, 58.6099, 58.6943, 58.7788, 58.8727, 58.9103, 58.9478, 59.0041, 59.0792, 59.1168, 59.1168, 59.098, 58.976, 58.9103, 58.807, 58.6756, 58.5817, 58.4597, 58.3376,
        58.3189, 58.2438, 58.1499, 58.1217, 58.1029],
   y = [80.4565, 80.4659, 80.4425, 80.4378, 80.4495, 80.4542, 80.4331, 80.426, 80.4284, 80.4307, 80.4284, 80.426, 80.4331, 80.4495, 80.4519, 80.4495, 80.4542, 80.4495, 80.4354, 80.426, 80.4096, 80.3908,
        80.3791, 80.3721, 80.358, 80.3509, 80.3369, 80.3251, 80.304, 80.2782, 80.2618, 80.2383, 80.2336, 80.2219, 80.1984, 80.1796, 80.1702, 80.1702, 80.1515, 80.1374, 80.1327, 80.1233, 80.1022,
        80.0975, 80.0881, 80.0576, 80.0552, 80.0576, 80.0505, 80.0529, 80.0599, 80.074, 80.1022, 80.1092, 80.135, 80.1515, 80.182, 80.1867, 80.1914, 80.2078, 80.2242, 80.2359, 80.2547, 80.2735,
        80.2805, 80.2829, 80.2876, 80.2829, 80.2852, 80.2852, 80.2852, 80.2782, 80.2665, 80.2805, 80.2852, 80.297, 80.3087, 80.3228, 80.3345, 80.3486, 80.3486, 80.3533, 80.3744, 80.3838, 80.3955,
        80.4143, 80.4284, 80.4472, 80.4542, 80.4565] };
  WCL[1842] = struct {
   x = [59.2271, 59.2177, 59.1895, 59.1426, 59.0769, 58.9548, 58.8046],
   y = [80.0459, 80.0459, 80.0459, 80.0247, 80.0294, 80.02, 80.0013] };
  WCL[1843] = struct {
   x = [59.6565, 59.5908, 59.497, 59.4125, 59.3374, 59.2623, 59.2247],
   y = [79.9966, 80.006, 80.013, 80.0318, 80.0388, 80.0435, 80.0459] };
  WCL[1844] = struct {
   x = [60.0978, 60.0039, 59.9382, 59.8161, 59.7129, 59.6659, 59.6847, 59.6659, 59.6378, 59.5908, 59.5908, 59.6002, 59.5439, 59.4782, 59.3843, 59.3562, 59.2904, 59.2717, 59.2529, 59.3186, 59.3374, 59.3468,
        59.3562, 59.3843, 59.497, 59.5345, 59.5251, 59.5345, 59.5908, 59.6659, 59.7598, 59.8255, 59.8818, 59.9382],
   y = [80.8062, 80.8062, 80.818, 80.8039, 80.7898, 80.7757, 80.7616, 80.7523, 80.7405, 80.7288, 80.71, 80.6795, 80.6607, 80.6396, 80.6208, 80.5997, 80.5692, 80.534, 80.5223, 80.5058, 80.4683, 80.4589,
        80.4378, 80.4167, 80.4143, 80.4073, 80.3932, 80.3815, 80.3674, 80.358, 80.3627, 80.3603, 80.3744, 80.3768] };
  WCL[1845] = struct {
   x = [58.4221, 58.4221, 58.4033, 58.3658, 58.3095, 58.2531, 58.1687, 58.1123, 58.0091, 57.9246, 57.934, 57.934, 57.887, 57.934, 57.8964, 57.8776, 57.9809, 58.0278, 58.1499, 58.2719, 58.3846, 58.4691,
        58.5254, 58.6099, 58.6568, 58.7507, 58.8258, 58.8915, 58.9666, 58.9666, 58.9666, 58.9947, 58.9572, 58.8727, 58.7601, 58.7037, 58.6286, 58.516, 58.4221],
   y = [81.7872, 81.7919, 81.7919, 81.7825, 81.7825, 81.7989, 81.8036, 81.7919, 81.7872, 81.7708, 81.7543, 81.7403, 81.7238, 81.7145, 81.7004, 81.6816, 81.6722, 81.6464, 81.6323, 81.6417, 81.6534, 81.6511,
        81.644, 81.6417, 81.6558, 81.6558, 81.6605, 81.6839, 81.7027, 81.7168, 81.7356, 81.7497, 81.7684, 81.7989, 81.8013, 81.7849, 81.7825, 81.7872, 81.7872] };
  WCL[1846] = struct {
   x = [57.6711, 57.5866, 57.5585, 57.4552, 57.4083, 57.3801, 57.2393, 57.1267, 57.0985, 56.9859, 56.8732, 56.8544, 56.7981, 56.7042, 56.723, 56.7042, 56.6948, 56.7793, 56.9108, 56.9859, 57.0516, 57.1173,
        57.1454, 57.1924, 57.2487, 57.2956, 57.3707, 57.474, 57.5397, 57.5397, 57.5397, 57.6054, 57.6993, 57.7556, 57.765, 57.8776, 57.9809, 58.056, 58.1405, 58.1874, 58.2344, 58.2719, 58.3189,
        58.3658, 58.3846, 58.3376, 58.2062, 58.178, 58.1123, 58.0466, 57.934, 57.8776, 57.8119, 57.7932, 57.6711],
   y = [81.5314, 81.5314, 81.5291, 81.5361, 81.5244, 81.522, 81.5244, 81.515, 81.5009, 81.4985, 81.4868, 81.4727, 81.4633, 81.4586, 81.4399, 81.4258, 81.4141, 81.4117, 81.4164, 81.4117, 81.3929, 81.3859,
        81.3906, 81.4, 81.3953, 81.3835, 81.3789, 81.3765, 81.3789, 81.3718, 81.3601, 81.3624, 81.3577, 81.3483, 81.3366, 81.3507, 81.3601, 81.3601, 81.3695, 81.3859, 81.3882, 81.3953, 81.3976,
        81.4117, 81.4234, 81.4375, 81.4328, 81.4493, 81.4633, 81.4868, 81.4985, 81.515, 81.5244, 81.5361, 81.5314] };
  WCL[1847] = struct {
   x = [56.5353, 56.5353, 56.4226, 56.3663, 56.3663, 56.3569, 56.3569, 56.4226, 56.4508, 56.4132, 56.4038, 56.432, 56.5071, 56.4695, 56.4038, 56.4038, 56.4132, 56.4132, 56.3569, 56.3006, 56.2818, 56.2255,
        56.1785, 56.1316, 56.0659, 56.0096, 55.9814, 55.8969, 55.8594, 55.85, 55.85, 55.8124, 55.7467, 55.6716, 55.6153, 55.5496, 55.4276, 55.4088, 55.4182, 55.4369, 55.5214, 55.6059, 55.681, 55.7279,
        55.803, 55.9251, 56.0096, 56.0659, 56.1692, 56.1973, 56.1973, 56.2349, 56.3757, 56.4883, 56.5353, 56.5446, 56.4883, 56.5259, 56.6385, 56.7324, 56.8732, 56.9483, 56.9483, 57.0609, 57.1924,
        57.2956, 57.3801, 57.4552, 57.5022, 57.5679, 57.6336, 57.6711, 57.6805, 57.7274, 57.7932, 57.7932, 57.7744, 57.7274, 57.6242, 57.5585, 57.5209, 57.5022, 57.4271, 57.305, 57.2299, 57.2299,
        57.0985, 56.9483, 56.8732, 56.7981, 56.6855, 56.5353],
   y = [81.3695, 81.3742, 81.3765, 81.3671, 81.3483, 81.3413, 81.3319, 81.3249, 81.3131, 81.3038, 81.2826, 81.2826, 81.2732, 81.2662, 81.2662, 81.2474, 81.231, 81.2169, 81.2169, 81.2146, 81.2357, 81.2427,
        81.2334, 81.2263, 81.2357, 81.2427, 81.2357, 81.231, 81.2334, 81.238, 81.2498, 81.2545, 81.2686, 81.2709, 81.2686, 81.2592, 81.2592, 81.2451, 81.2263, 81.2075, 81.1935, 81.1841, 81.1794,
        81.1583, 81.1536, 81.1606, 81.1606, 81.17, 81.177, 81.1747, 81.1489, 81.1489, 81.1559, 81.1606, 81.1606, 81.1653, 81.177, 81.1911, 81.1958, 81.2005, 81.2099, 81.2193, 81.2075, 81.2099, 81.2146,
        81.2193, 81.2334, 81.238, 81.2568, 81.2592, 81.2709, 81.2662, 81.2592, 81.2615, 81.2639, 81.2756, 81.2944, 81.3061, 81.3108, 81.3178, 81.3038, 81.2944, 81.3038, 81.3014, 81.3108, 81.3249,
        81.3296, 81.3366, 81.3624, 81.3765, 81.3742, 81.3695] };
  WCL[1848] = struct {
   x = [58.7084, 58.7272, 58.7272, 58.746, 58.7084, 58.7741, 58.7741, 58.746, 58.6615, 58.5582, 58.4456, 58.3611, 58.3423, 58.2766, 58.2109, 58.117, 58.0607, 57.9856, 57.8636, 57.7697, 57.7603, 57.8542,
        57.9481, 58.0607, 58.0889, 58.1827, 58.2578, 58.286, 58.3235, 58.408, 58.4925, 58.5488, 58.6146, 58.6803, 58.7084],
   y = [80.8532, 80.8602, 80.8555, 80.8414, 80.825, 80.8086, 80.7828, 80.771, 80.7546, 80.7288, 80.7124, 80.71, 80.7264, 80.7311, 80.7335, 80.7264, 80.7405, 80.7523, 80.7546, 80.7569, 80.7687, 80.7828,
        80.7945, 80.8086, 80.8227, 80.825, 80.825, 80.8414, 80.8461, 80.8461, 80.8532, 80.8555, 80.8555, 80.8555, 80.8532] };
  WCL[1849] = struct {
   x = [56.1058, 56.0682, 56.0307, 55.9838, 55.9087, 55.796, 55.7303, 55.7115, 55.6083, 55.4768, 55.4017, 55.4299, 55.5332, 55.5895, 55.505, 55.4956, 55.5238, 55.505, 55.5519, 55.6458, 55.7209, 55.796,
        55.918, 56.0495, 56.1621, 56.2278, 56.2466, 56.3123, 56.425, 56.547, 56.6033, 56.6033, 56.6221, 56.7066, 56.8286, 56.8849, 56.9225, 56.9225, 56.9131, 56.8474, 56.7629, 56.7911, 56.7629, 56.669,
        56.5658, 56.5376, 56.4907, 56.3874, 56.3029, 56.1809, 56.1058],
   y = [80.7194, 80.7194, 80.717, 80.703, 80.7124, 80.717, 80.7053, 80.6889, 80.6818, 80.6772, 80.6701, 80.6631, 80.6631, 80.656, 80.6513, 80.6373, 80.6232, 80.6138, 80.5997, 80.5997, 80.5903, 80.5927,
        80.5997, 80.5974, 80.6091, 80.6114, 80.5997, 80.5927, 80.5927, 80.5927, 80.595, 80.5997, 80.6091, 80.6021, 80.6161, 80.6232, 80.6349, 80.6537, 80.6701, 80.6795, 80.6818, 80.6889, 80.6983,
        80.7053, 80.71, 80.7194, 80.7194, 80.7217, 80.7194, 80.7194, 80.7194] };
  WCL[1850] = struct {
   x = [54.6062, 54.6155, 54.5968, 54.5123, 54.3996, 54.2588, 54.1086, 54.0148, 54.0148, 54.0711, 54.0711, 54.0711, 54.0523, 54.0992, 54.1837, 54.3245, 54.3996, 54.4747, 54.5592, 54.6062, 54.6062, 54.578,
        54.6437, 54.7094, 54.7845, 54.869, 54.9535, 55.038, 55.16, 55.2445, 55.282, 55.3196, 55.4229, 55.5543, 55.6951, 55.7326, 55.7984, 55.8077, 55.6951, 55.6482, 55.5918, 55.5449, 55.4322, 55.3102,
        55.2539, 55.282, 55.3571, 55.3571, 55.3008, 55.2163, 55.0943, 54.9817, 54.9723, 54.9723, 55.0004, 54.991, 54.9347, 54.8596, 54.7657, 54.6813, 54.6062],
   y = [80.8602, 80.8626, 80.8626, 80.8532, 80.8532, 80.8414, 80.8438, 80.8461, 80.8344, 80.8156, 80.7968, 80.7781, 80.7663, 80.7663, 80.764, 80.7663, 80.7616, 80.7663, 80.7546, 80.7382, 80.7264, 80.7077,
        80.703, 80.71, 80.7006, 80.6959, 80.703, 80.717, 80.7124, 80.7006, 80.6959, 80.71, 80.7264, 80.7311, 80.7335, 80.7429, 80.7476, 80.7593, 80.7663, 80.7875, 80.7828, 80.7616, 80.7663, 80.7757,
        80.7828, 80.7921, 80.8015, 80.818, 80.8273, 80.8133, 80.8062, 80.8062, 80.8156, 80.8297, 80.8391, 80.8508, 80.8602, 80.8485, 80.8485, 80.8579, 80.8602] };
  WCL[1851] = struct {
   x = [55.8641, 55.8265, 55.8171, 55.7514, 55.6951, 55.6106, 55.4698, 55.3947, 55.282, 55.2163, 55.1506, 55.0192, 54.9347, 54.8408, 54.7376, 54.6531, 54.5592, 54.5311, 54.5311, 54.4841, 54.4654, 54.3715,
        54.4184, 54.5123, 54.5311, 54.6437, 54.7657, 54.7939, 54.7751, 54.8784, 55.0004, 55.0943, 55.2351, 55.3196, 55.3947, 55.3665, 55.3947, 55.4792, 55.5637, 55.6575, 55.6857, 55.742, 55.789,
        55.789, 55.9485, 56.0706, 56.1926, 56.3053, 56.4179, 56.5118, 56.5775, 56.662, 56.7089, 56.7089, 56.7089, 56.7371, 56.831, 56.8779, 56.8779, 56.8779, 56.8873, 56.953, 56.9999, 56.9999, 56.9812,
        57.0563, 57.1314, 57.2909, 57.4881, 57.5256, 57.5913, 57.6101, 57.6195, 57.6383, 57.5726, 57.4599, 57.413, 57.2909, 57.1783, 57.1032, 57.0187, 56.9154, 56.8497, 56.7652, 56.6901, 56.6338,
        56.6057, 56.5493, 56.4085, 56.3053, 56.2865, 56.324, 56.3428, 56.2489, 56.1832, 56.0518, 55.9298, 55.8641],
   y = [80.9705, 80.9752, 80.9775, 80.9869, 81.0057, 81.0104, 81.0104, 81.0245, 81.0245, 81.0245, 81.0386, 81.0503, 81.0456, 81.0503, 81.0409, 81.0433, 81.0362, 81.0245, 81.0104, 80.9916, 80.9775, 80.9682,
        80.9588, 80.9541, 80.9494, 80.9447, 80.9423, 80.933, 80.9142, 80.9165, 80.9189, 80.9236, 80.9189, 80.9189, 80.9071, 80.8837, 80.879, 80.8672, 80.8719, 80.8813, 80.8696, 80.8508, 80.8227,
        80.8109, 80.7968, 80.7875, 80.7898, 80.7828, 80.7757, 80.7546, 80.7593, 80.7687, 80.7476, 80.7335, 80.7194, 80.7147, 80.7241, 80.7194, 80.7053, 80.6959, 80.6889, 80.6889, 80.6748, 80.6607,
        80.6631, 80.6607, 80.6607, 80.6725, 80.6748, 80.6912, 80.7053, 80.717, 80.7358, 80.7546, 80.7593, 80.7757, 80.7968, 80.8086, 80.8297, 80.8532, 80.8649, 80.8719, 80.886, 80.8931, 80.9071,
        80.9118, 80.9118, 80.9095, 80.9189, 80.9259, 80.933, 80.9423, 80.9588, 80.9799, 80.9893, 80.9799, 80.9658, 80.9705] };
  WCL[1852] = struct {
   x = [56.8216, 56.8216, 56.6995, 56.5587, 56.5306, 56.4555, 56.371, 56.2865, 56.2302, 56.1832, 56.1081, 56.1175, 56.202, 56.3428, 56.4461, 56.5493, 56.7089, 56.8028, 56.8216, 56.8779, 56.9342, 56.9718,
        57.0656, 57.2252, 57.3379, 57.4881, 57.5538, 57.5538, 57.5538, 57.6852, 57.7697, 57.8166, 57.8354, 57.873, 57.9199, 57.9481, 57.9668, 58.0044, 58.0138, 58.0232, 58.0232, 57.9481, 57.8166,
        57.7134, 57.6758, 57.5726, 57.4505, 57.366, 57.2534, 57.1501, 57.0563, 56.9905, 56.8873, 56.8216],
   y = [81.0433, 81.0456, 81.0456, 81.0456, 81.0573, 81.0691, 81.055, 81.0573, 81.0667, 81.0761, 81.062, 81.048, 81.0268, 81.0174, 80.9916, 80.9729, 80.9541, 80.94, 80.933, 80.9283, 80.9283, 80.9118,
        80.8837, 80.8461, 80.818, 80.8156, 80.818, 80.8109, 80.7898, 80.7945, 80.7898, 80.8039, 80.8062, 80.825, 80.825, 80.8297, 80.8391, 80.8461, 80.8602, 80.8719, 80.879, 80.8813, 80.8837, 80.9071,
        80.9259, 80.9377, 80.9588, 80.9729, 80.9916, 81.0034, 81.0127, 81.0245, 81.0409, 81.0433] };
  WCL[1853] = struct {
   x = [46.5119, 46.3242, 46.2021, 46.0989, 46.0801, 45.9956, 45.8267, 45.714, 45.714, 45.6671, 45.5732, 45.4605, 45.4136, 45.3854, 45.3291, 45.2728, 45.2071, 45.1508, 45.0851, 45.0193, 44.9349, 44.8598,
        44.8316, 44.7659, 44.6908, 44.6157, 44.5218, 44.4655, 44.4373, 44.381, 44.3247, 44.3247, 44.3622, 44.3341, 44.259, 44.1933, 44.2496, 44.2777, 44.212, 44.212, 44.2965, 44.2777, 44.2308, 44.2496,
        44.3059, 44.2684, 44.212, 44.2496, 44.2965, 44.3716, 44.4655, 44.5218, 44.6063, 44.7471, 44.8879, 44.963, 45.0569, 45.1226, 45.1226, 45.0944, 45.1414, 45.2259, 45.2728, 45.301, 45.301, 45.3479,
        45.4136, 45.4887, 45.5169, 45.5544, 45.6483, 45.7516, 45.7797, 45.8079, 45.8454, 45.9487, 45.9768, 46.005, 46.0801, 46.1364, 46.1458, 46.1458, 46.2209, 46.2491, 46.1928, 46.1458, 46.0989,
        46.0895, 46.1834, 46.2021, 46.0989, 45.9675, 45.9205, 45.9111, 45.9581, 45.9581, 45.9675, 46.0613, 46.127, 46.1552, 46.2115, 46.2772, 46.2866, 46.2866, 46.2679, 46.3336, 46.3993, 46.4462,
        46.4744, 46.5401, 46.6246, 46.6903, 46.8405, 46.9813, 47.1033, 47.2629, 47.4131, 47.5915, 47.6384, 47.7229, 47.7792, 47.8731, 47.9106, 47.9482, 47.9951, 48.0514, 48.1453, 48.211, 48.2861,
        48.3706, 48.4082, 48.4551, 48.5865, 48.7086, 48.9057, 48.8963, 48.9057, 48.8118, 48.7086, 48.6053, 48.4927, 48.4645, 48.38, 48.2767, 48.1078, 47.9951, 47.9106, 47.8355, 47.7792, 47.6666,
        47.6196, 47.507, 47.3849, 47.3849, 47.3005, 47.2254, 47.169, 47.1315, 47.047, 46.9907, 46.9437, 46.925, 46.8968, 46.8217, 46.6527, 46.5776, 46.5119],
   y = [80.8602, 80.8508, 80.8414, 80.832, 80.818, 80.818, 80.8015, 80.7875, 80.7734, 80.7734, 80.7687, 80.7616, 80.7476, 80.7476, 80.7476, 80.7452, 80.7264, 80.717, 80.7241, 80.7241, 80.7194, 80.7053,
        80.6889, 80.71, 80.71, 80.6959, 80.6959, 80.6889, 80.6678, 80.6584, 80.656, 80.6513, 80.6373, 80.6255, 80.6232, 80.6067, 80.6091, 80.595, 80.5809, 80.5762, 80.5762, 80.5598, 80.5528, 80.5504,
        80.5457, 80.5316, 80.5246, 80.5223, 80.5035, 80.5011, 80.5152, 80.5199, 80.5223, 80.5152, 80.5152, 80.5129, 80.4918, 80.4894, 80.4753, 80.4636, 80.4472, 80.4542, 80.4612, 80.473, 80.5035,
        80.5035, 80.4871, 80.4871, 80.4871, 80.4988, 80.5035, 80.5011, 80.4918, 80.4753, 80.4589, 80.4612, 80.4448, 80.4143, 80.4096, 80.4213, 80.4425, 80.4495, 80.4425, 80.4495, 80.4659, 80.4824,
        80.4988, 80.5176, 80.5223, 80.541, 80.5528, 80.5622, 80.5762, 80.595, 80.6185, 80.6419, 80.6607, 80.6607, 80.6607, 80.6795, 80.6795, 80.6795, 80.6912, 80.71, 80.7264, 80.717, 80.717, 80.7311,
        80.7523, 80.7523, 80.7452, 80.7476, 80.7569, 80.7546, 80.7546, 80.7523, 80.7523, 80.7476, 80.7241, 80.7147, 80.6912, 80.6889, 80.6654, 80.6607, 80.6607, 80.6466, 80.6466, 80.6443, 80.649,
        80.6607, 80.649, 80.6466, 80.6513, 80.6631, 80.6818, 80.6818, 80.6959, 80.7077, 80.717, 80.7335, 80.7405, 80.7382, 80.7523, 80.7734, 80.7781, 80.7945, 80.8156, 80.8156, 80.818, 80.8109, 80.825,
        80.8227, 80.8297, 80.8461, 80.8485, 80.8532, 80.8391, 80.8414, 80.8461, 80.8555, 80.8461, 80.8344, 80.8344, 80.8438, 80.8438, 80.8532, 80.8602] };
  WCL[1854] = struct {
   x = [50.0627, 49.997],
   y = [80.8532, 80.8532] };
  WCL[1855] = struct {
   x = [50.0064, 49.9876, 49.9313, 49.8843, 49.7623, 49.6778, 49.6309, 49.6778, 49.7341, 49.6778, 49.6121, 49.6215, 49.6872, 49.7623, 49.828, 49.9313],
   y = [80.8649, 80.879, 80.8696, 80.8672, 80.8602, 80.8626, 80.8485, 80.8461, 80.8414, 80.8344, 80.8227, 80.818, 80.8109, 80.8015, 80.8015, 80.8062] };
  WCL[1856] = struct {
   x = [50.0298, 49.9641, 49.8702, 49.7951, 49.7388, 49.6731, 49.5699, 49.5041, 49.476, 49.4666, 49.4854, 49.4854, 49.4666, 49.4103, 49.3539, 49.3352, 49.2788, 49.2507, 49.2976, 49.2788, 49.2601, 49.2319,
        49.1944, 49.1756, 49.1662, 49.2319, 49.2882, 49.2788, 49.2131, 49.1756, 49.1474, 49.0442, 48.9878, 48.9127, 48.8564, 48.8564, 48.7813, 48.6687, 48.6405, 48.5936, 48.5279, 48.4152, 48.3119,
        48.3026, 48.2368, 48.1524, 48.0867, 48.0116, 47.9458, 47.8707, 47.8614, 47.852, 47.7863, 47.7581, 47.7018, 47.5891, 47.4671, 47.3826, 47.3169, 47.2324, 47.1761, 47.1104, 47.0353, 46.9508,
        46.8757, 46.8194, 46.7818, 46.6692, 46.6504, 46.6692, 46.6316, 46.5941, 46.5377, 46.5284, 46.5659, 46.6222, 46.7067, 46.8194, 46.9038, 46.9414, 46.9414, 46.9789, 47.1479, 47.2418, 47.3638,
        47.392, 47.3638, 47.3169, 47.2981, 47.3826, 47.4483, 47.514, 47.5891, 47.5797, 47.5516, 47.6454, 47.6548, 47.5797, 47.4859, 47.4577, 47.3544, 47.2793, 47.27, 47.2136, 47.1291, 47.0634, 46.9883,
        46.9883, 46.9883, 46.8945, 46.8569, 46.8757, 46.8194, 46.7912, 46.7349, 46.6692, 46.6973, 46.6973, 46.6222, 46.6035, 46.6504, 46.7255, 46.8194, 46.81, 46.8569, 46.9789, 46.9696, 46.8757,
        46.9132, 46.9602, 46.9226, 46.9789, 47.0447, 47.1385, 47.2042, 47.2887, 47.392, 47.4202, 47.4202, 47.4765, 47.5891, 47.683, 47.7205, 47.7205, 47.7112, 47.683, 47.7393, 47.805, 47.852, 47.8895,
        47.8895, 47.8614, 47.7863, 47.7018, 47.6454, 47.6454, 47.6173],
   y = [80.7311, 80.7311, 80.7358, 80.7335, 80.7264, 80.7358, 80.7311, 80.7241, 80.717, 80.7077, 80.6936, 80.6865, 80.6818, 80.6889, 80.6912, 80.6818, 80.6772, 80.6678, 80.6513, 80.649, 80.649, 80.6513,
        80.6537, 80.6537, 80.6396, 80.6631, 80.6537, 80.6443, 80.6466, 80.6513, 80.6419, 80.6302, 80.6232, 80.6232, 80.6161, 80.6067, 80.6067, 80.6021, 80.6091, 80.6114, 80.6044, 80.5927, 80.5856,
        80.5997, 80.6091, 80.6044, 80.6067, 80.6138, 80.6255, 80.6302, 80.6232, 80.5997, 80.6021, 80.6091, 80.6091, 80.6114, 80.6138, 80.6067, 80.6208, 80.6302, 80.6373, 80.6232, 80.6161, 80.6255,
        80.6302, 80.6279, 80.6114, 80.6021, 80.5903, 80.5833, 80.5622, 80.5481, 80.5316, 80.5223, 80.5035, 80.5058, 80.5035, 80.5129, 80.5199, 80.527, 80.5058, 80.5082, 80.5129, 80.4871, 80.4777,
        80.4612, 80.4519, 80.4425, 80.4284, 80.426, 80.4213, 80.4307, 80.4307, 80.419, 80.4073, 80.4002, 80.3697, 80.3744, 80.3674, 80.3838, 80.3885, 80.3955, 80.3791, 80.3674, 80.3721, 80.3721,
        80.365, 80.3533, 80.3392, 80.3416, 80.3228, 80.3064, 80.3134, 80.3134, 80.3017, 80.2805, 80.2594, 80.2336, 80.2336, 80.2172, 80.2078, 80.2078, 80.1914, 80.1702, 80.1702, 80.1796, 80.1632,
        80.1468, 80.1421, 80.1374, 80.1163, 80.1116, 80.121, 80.1233, 80.128, 80.1233, 80.1256, 80.135, 80.1444, 80.1585, 80.1702, 80.1773, 80.1679, 80.1491, 80.1327, 80.1233, 80.1256, 80.1397, 80.128,
        80.1022, 80.0881, 80.0787, 80.0764, 80.0646, 80.0459, 80.0318, 80.006] };
  WCL[1857] = struct {
   x = [47.9106, 47.9857, 48.0139, 47.9951, 48.0233, 48.089, 48.1453, 48.1453, 48.2204, 48.3049, 48.3143, 48.3612, 48.3518, 48.3049, 48.3143, 48.3049, 48.3331, 48.4176, 48.4363, 48.4739, 48.5678, 48.6522,
        48.6898, 48.6898, 48.6616, 48.6898, 48.7743, 48.8212, 48.8775, 48.9432, 48.962, 48.9245, 48.8588, 48.8963, 48.9151, 48.8681, 48.9151, 48.9714, 48.9902, 48.9245, 48.8775, 48.8588, 48.8212,
        48.7273, 48.6616, 48.6241, 48.6241, 48.6053, 48.6804, 48.7743, 48.8212, 48.9432, 49.0277, 49.0653, 49.1592, 49.1685, 49.2906, 49.3469, 49.3751, 49.4126, 49.4689, 49.5159, 49.5346, 49.5628,
        49.6191, 49.6848, 49.7318, 49.7599, 49.8069, 49.9101, 49.9665],
   y = [79.9895, 80.0106, 80.0224, 80.0388, 80.0505, 80.0599, 80.0435, 80.0153, 80.013, 80.02, 80.0435, 80.0459, 80.0576, 80.074, 80.0904, 80.1233, 80.1303, 80.1233, 80.1045, 80.0951, 80.1022, 80.0857,
        80.0951, 80.1116, 80.1327, 80.135, 80.1421, 80.121, 80.1233, 80.1421, 80.1608, 80.1702, 80.1796, 80.182, 80.2054, 80.2313, 80.2219, 80.2101, 80.2266, 80.2477, 80.2618, 80.2805, 80.2923,
        80.2946, 80.2993, 80.304, 80.3322, 80.3392, 80.3416, 80.3462, 80.3369, 80.3369, 80.3392, 80.3157, 80.3228, 80.3298, 80.3345, 80.3298, 80.3134, 80.3181, 80.3298, 80.3251, 80.3228, 80.3181,
        80.3134, 80.3181, 80.3322, 80.3275, 80.311, 80.3134, 80.3181] };
  WCL[1858] = struct {
   x = [50.0838, 49.9993, 49.9148, 49.8867, 49.8773, 49.8961, 49.9712],
   y = [80.4378, 80.4401, 80.4425, 80.4589, 80.4589, 80.4824, 80.4847] };
  WCL[1859] = struct {
   x = [33.4635, 33.4917, 33.4917, 33.4917, 33.5386, 33.5105, 33.4166, 33.3509, 33.3133, 33.2007, 33.1256, 33.0223, 32.9472, 32.9003, 32.8252, 32.7031, 32.5905, 32.5623, 32.4309, 32.3934, 32.3464, 32.2807,
        32.2619, 32.1962, 32.093, 32.0836, 32.0836, 32.0648, 32.0366, 32.0366, 32.0085, 31.9709, 31.9615, 31.9522, 31.8771, 31.8207, 31.6987, 31.6424, 31.6142, 31.5579, 31.5016, 31.4359, 31.4359,
        31.4359, 31.4359, 31.3701, 31.3232, 31.3701, 31.4452, 31.511, 31.5579, 31.6612, 31.7456, 31.8207, 31.9334, 32.0648, 32.093, 32.1775, 32.2619, 32.337, 32.3746, 32.4309, 32.4685, 32.5342,
        32.6374, 32.7031, 32.7595, 32.797, 32.8158, 32.8815, 32.9472, 32.9566, 32.9754, 33.0599, 33.1068, 33.1725, 33.2382, 33.3227, 33.4166, 33.4541, 33.4635],
   y = [80.2078, 80.2125, 80.2078, 80.196, 80.1796, 80.1796, 80.1773, 80.1796, 80.1726, 80.1679, 80.1468, 80.135, 80.1327, 80.1186, 80.1233, 80.121, 80.1139, 80.0975, 80.0928, 80.0811, 80.074, 80.0764,
        80.0717, 80.0717, 80.0787, 80.0811, 80.0764, 80.074, 80.067, 80.0529, 80.0505, 80.0482, 80.0412, 80.0388, 80.0388, 80.0505, 80.0294, 80.0435, 80.0388, 80.0365, 80.0247, 80.0318, 80.0459,
        80.0576, 80.0646, 80.0646, 80.0717, 80.0764, 80.0834, 80.0975, 80.0928, 80.0951, 80.1092, 80.1069, 80.1092, 80.1327, 80.1256, 80.128, 80.1421, 80.1515, 80.1562, 80.1585, 80.1632, 80.1679,
        80.1702, 80.1679, 80.1702, 80.1867, 80.189, 80.1914, 80.2054, 80.2219, 80.2242, 80.2172, 80.2172, 80.2242, 80.2172, 80.2148, 80.2172, 80.2172, 80.2078] };
  WCL[1860] = struct {
   x = [22.7784, 22.7596, 22.7596, 22.8065, 22.8253, 22.9098, 23.0318, 23.1069, 23.182, 23.2196, 23.229, 23.2008, 23.2008, 23.182, 23.0788, 23.0694, 23.1163, 23.1351, 23.1633, 23.2008, 23.2384, 23.2008,
        23.182, 23.1727, 23.1633, 23.0976, 23.0412, 23.0318, 23.0318, 23.0037, 22.9661, 22.9474, 22.9286, 22.9192, 22.938, 22.9755, 23.0131, 23.06, 23.0694, 23.0694, 23.1069, 23.1539, 23.1539, 23.1539,
        23.182, 23.2384, 23.2477, 23.2571, 23.2853, 23.2853, 23.2947, 23.3135, 23.3604, 23.3792, 23.3792, 23.4073, 23.4355, 23.473, 23.4824, 23.4824, 23.473, 23.5012, 23.5575, 23.6045, 23.5857,
        23.5763, 23.642, 23.6983, 23.7265, 23.7077, 23.6796, 23.6608, 23.6983, 23.7547, 23.811, 23.811, 23.7641, 23.7265, 23.7265, 23.7641, 23.8204, 23.8579, 23.8485, 23.7734, 23.7359, 23.7359,
        23.7734, 23.8204, 23.8767, 23.9518, 24.0175, 24.0269, 24.0269, 23.9893, 23.98, 23.9987, 24.0644, 24.1208, 24.1865, 24.2053, 24.2053, 24.1489, 24.1208, 24.1302, 24.1771, 24.2146, 24.2991,
        24.3742, 24.4212, 24.4681, 24.5244, 24.5714, 24.5901, 24.6465, 24.6934, 24.7309, 24.7685, 24.806, 24.8154, 24.8154, 24.8342, 24.8811, 24.8905, 24.8905, 24.8811, 24.8624, 24.8624, 24.8624,
        24.8624, 24.7779, 24.7309, 24.6559, 24.6183, 24.6183, 24.6183, 24.5901, 24.5526, 24.5244, 24.4869, 24.4869, 24.5244, 24.6089, 24.6934, 24.7309, 24.7591, 24.7967, 24.8342, 24.8999, 24.9281,
        24.9469, 25.0126, 25.0689, 25.0877, 25.1158, 25.1722, 25.2848, 25.3505, 25.3881, 25.3693, 25.3036, 25.2473, 25.2191, 25.2566, 25.313, 25.3411, 25.3505, 25.4256, 25.4444, 25.4444, 25.4538,
        25.5101, 25.5289, 25.5383, 25.5664, 25.6134, 25.6603, 25.6978, 25.726, 25.7636, 25.7823, 25.7917, 25.8199, 25.8105, 25.8105, 25.8105, 25.7823, 25.7823, 25.848, 25.895, 25.9325, 25.9701, 26.017,
        26.0264, 26.1672, 26.2705, 26.355, 26.4582, 26.5333, 26.599, 26.646, 26.6647, 26.7305, 26.7962, 26.8431, 26.8243, 26.8243, 26.8806, 26.9088, 26.9557, 26.9557, 26.9557, 26.9745, 27.0402,
        27.1059, 27.1153, 27.1341, 27.1341, 27.1153, 27.1059],
   y = [80.5129, 80.5152, 80.5129, 80.4988, 80.4824, 80.4683, 80.473, 80.4659, 80.4659, 80.4542, 80.4401, 80.4213, 80.4073, 80.3932, 80.3885, 80.3815, 80.3697, 80.3533, 80.3439, 80.3345, 80.3228, 80.3064,
        80.2876, 80.2735, 80.2547, 80.243, 80.2336, 80.2172, 80.2007, 80.189, 80.1749, 80.1632, 80.1538, 80.1397, 80.1256, 80.1092, 80.1022, 80.1022, 80.1022, 80.1022, 80.1045, 80.121, 80.1327,
        80.1421, 80.1397, 80.1421, 80.1608, 80.1679, 80.1632, 80.1491, 80.1327, 80.1327, 80.1327, 80.1585, 80.1773, 80.1773, 80.1632, 80.1421, 80.1327, 80.1256, 80.1139, 80.1022, 80.0998, 80.1163,
        80.1233, 80.1327, 80.1327, 80.1327, 80.1444, 80.1538, 80.1632, 80.1702, 80.1726, 80.1702, 80.182, 80.1914, 80.2054, 80.2172, 80.2336, 80.2313, 80.2289, 80.2336, 80.2453, 80.2524, 80.2571,
        80.2711, 80.2688, 80.2547, 80.2453, 80.2453, 80.2453, 80.2524, 80.2641, 80.2735, 80.2805, 80.2782, 80.2618, 80.2524, 80.243, 80.243, 80.2618, 80.2735, 80.2805, 80.2876, 80.2805, 80.2711,
        80.2641, 80.2571, 80.2406, 80.2406, 80.2453, 80.2453, 80.2641, 80.2665, 80.2594, 80.2665, 80.2946, 80.3157, 80.3298, 80.3322, 80.3345, 80.3369, 80.3228, 80.3017, 80.2899, 80.2735, 80.2571,
        80.2359, 80.2242, 80.2172, 80.2078, 80.2007, 80.1914, 80.1796, 80.1632, 80.1515, 80.1421, 80.1327, 80.1139, 80.1163, 80.128, 80.1374, 80.1538, 80.1702, 80.1867, 80.1984, 80.2007, 80.2078,
        80.2007, 80.196, 80.1937, 80.1984, 80.2172, 80.2289, 80.2359, 80.243, 80.243, 80.2383, 80.2336, 80.2336, 80.2195, 80.2078, 80.2007, 80.2007, 80.1867, 80.1702, 80.1702, 80.182, 80.1984, 80.2078,
        80.2031, 80.1867, 80.1749, 80.1632, 80.1655, 80.182, 80.196, 80.1914, 80.1843, 80.1796, 80.1796, 80.1726, 80.1726, 80.1702, 80.1515, 80.1444, 80.1397, 80.1397, 80.1421, 80.1468, 80.1608,
        80.1608, 80.1515, 80.1515, 80.1515, 80.1515, 80.1538, 80.1515, 80.1397, 80.1327, 80.1327, 80.1397, 80.1327, 80.1233, 80.1022, 80.0951, 80.0857, 80.074, 80.074, 80.0623, 80.0552, 80.0552,
        80.0599, 80.0623, 80.0552, 80.0552, 80.0388, 80.0177, 80.006] };
  WCL[1861] = struct {
   x = [19.9551, 20.0302, 20.1335, 20.218, 20.3025, 20.3025, 20.3212, 20.3588, 20.4339, 20.4527, 20.4245, 20.4339, 20.4621, 20.5559, 20.6498, 20.7343, 20.7437, 20.7343, 20.7343, 20.7812, 20.7812, 20.7812,
        20.7812, 20.7812, 20.8469, 20.9596, 21.0535, 21.1098, 21.2224, 21.2412, 21.2881, 21.3069, 21.3726, 21.4571, 21.4947, 21.5791, 21.6449, 21.7481, 21.7575, 21.7575, 21.7857, 21.8608, 21.9077,
        21.8983, 21.9171, 21.9171, 21.842, 21.7387, 21.7106, 21.5416, 21.5885, 21.6824, 21.7951, 21.842, 21.9077, 22.0016, 22.0391, 22.1236],
   y = [80.4026, 80.3885, 80.4096, 80.419, 80.4096, 80.3908, 80.3744, 80.3627, 80.3462, 80.3298, 80.3228, 80.2993, 80.2805, 80.2711, 80.2805, 80.2899, 80.2735, 80.2524, 80.2453, 80.2313, 80.2195, 80.2101,
        80.2125, 80.2078, 80.1984, 80.196, 80.196, 80.196, 80.2148, 80.2172, 80.2313, 80.2195, 80.2125, 80.2289, 80.2406, 80.2336, 80.2078, 80.2078, 80.2289, 80.2524, 80.2547, 80.2665, 80.25, 80.2219,
        80.1984, 80.196, 80.1726, 80.1515, 80.1303, 80.0951, 80.0998, 80.1163, 80.128, 80.1303, 80.1092, 80.0764, 80.0388, 80.0106] };
  WCL[1862] = struct {
   x = [22.4357, 22.4451, 22.37, 22.3231, 22.4076, 22.4264, 22.3794, 22.37, 22.3513, 22.3606, 22.3794, 22.4076, 22.4264, 22.4264, 22.4451, 22.4451, 22.4451, 22.417, 22.3231, 22.3231, 22.3513, 22.3794,
        22.3794, 22.3794, 22.4357, 22.5015, 22.539, 22.5484, 22.6141, 22.6329, 22.7174, 22.7925, 22.7925, 22.8019, 22.7831, 22.7925, 22.8206, 22.8112, 22.7925, 22.7925, 22.7643, 22.7643, 22.7549,
        22.7643, 22.7831],
   y = [79.9919, 80.0106, 80.0247, 80.0341, 80.0412, 80.0576, 80.074, 80.0928, 80.1233, 80.1491, 80.1984, 80.2172, 80.2477, 80.2618, 80.2805, 80.2923, 80.3064, 80.3157, 80.3181, 80.3392, 80.3556, 80.3815,
        80.4002, 80.4026, 80.4002, 80.4073, 80.4073, 80.3674, 80.3462, 80.3228, 80.3298, 80.3345, 80.3462, 80.358, 80.3627, 80.3979, 80.4213, 80.4284, 80.4589, 80.4753, 80.4941, 80.5058, 80.5058,
        80.5129, 80.5129] };
  WCL[1863] = struct {
   x = [15.996, 16.0993, 16.1838, 16.3058, 16.4091, 16.5123, 16.5593, 16.5593],
   y = [79.9989, 80.013, 80.0505, 80.0623, 80.0365, 80.0318, 80.0177, 80.0106] };
  WCL[1864] = struct {
   x = [18.7301, 18.8146, 18.8709, 18.8803, 18.824, 18.7207, 18.6456, 18.5611, 18.5048, 18.4954, 18.4579, 18.4297, 18.3921, 18.3828, 18.3452, 18.317, 18.2795, 18.2044, 18.1575, 18.0918, 18.0636, 18.0824,
        18.1387, 18.1387, 18.0636, 18.0354, 18.0167, 17.9509, 17.904, 17.8477, 17.8101, 17.782, 17.7632, 17.6505, 17.5473, 17.5473, 17.6224, 17.7538, 17.8289, 17.9134, 18.0354, 18.0918, 18.1011,
        18.1199, 18.2044, 18.317, 18.4109, 18.4766, 18.5611, 18.655, 18.7019, 18.7207, 18.7864, 18.8521, 18.9085, 18.9085, 18.9085, 18.9835, 19.0868, 19.0962, 19.1619, 19.1619, 19.2746, 19.359,
        19.4248, 19.4248, 19.406, 19.406, 19.359, 19.3872, 19.4529, 19.5092, 19.4623, 19.4529, 19.4248, 19.3872, 19.4154, 19.4905, 19.5656, 19.6031, 19.6031, 19.5656, 19.5092, 19.406, 19.3309, 19.2839,
        19.1713, 19.1337, 19.1244, 19.068, 19.0399, 19.0399, 19.0962, 19.0962, 19.0962, 19.0868, 19.0868, 19.0774, 19.0774, 19.1244, 19.1901, 19.2558, 19.3403, 19.3872, 19.4154, 19.4811, 19.528,
        19.575, 19.6594, 19.7251, 19.7815, 19.8284, 19.8378, 19.8566, 19.8566, 19.8566, 19.8284, 19.8284, 19.7627, 19.7345, 19.7345, 19.8002, 19.8002, 19.8284, 19.8002, 19.7721, 19.8284, 19.8472,
        19.8472, 19.7909, 19.7439, 19.7345, 19.7627, 19.8472, 19.8847, 19.866, 19.7815, 19.7251, 19.7158, 19.7158, 19.6782, 19.6407, 19.6125, 19.5843, 19.575, 19.5468, 19.5092, 19.4905, 19.4435,
        19.4435, 19.4905, 19.5186, 19.5656, 19.6031, 19.6688, 19.6782, 19.6688, 19.6688, 19.7064, 19.7439, 19.7627, 19.8002, 19.9598, 19.988, 19.9504, 19.9598],
   y = [79.9895, 80.0036, 80.0177, 80.0294, 80.0247, 80.0177, 80.013, 80.013, 80.013, 80.0247, 80.0271, 80.0341, 80.0341, 80.02, 80.0153, 80.0153, 80.0224, 80.0247, 80.0459, 80.0459, 80.067, 80.074, 80.0764,
        80.0904, 80.1022, 80.1233, 80.1327, 80.1421, 80.1327, 80.1374, 80.1397, 80.1397, 80.1421, 80.1327, 80.1256, 80.1421, 80.1608, 80.189, 80.1937, 80.1796, 80.182, 80.182, 80.1749, 80.1608,
        80.1608, 80.1679, 80.1608, 80.1702, 80.1444, 80.1421, 80.1702, 80.1867, 80.1796, 80.1726, 80.1632, 80.1397, 80.1233, 80.1186, 80.1069, 80.0857, 80.067, 80.0482, 80.0153, 80.0013, 80.013,
        80.0247, 80.0529, 80.0693, 80.0881, 80.0811, 80.074, 80.0834, 80.0951, 80.1045, 80.1116, 80.1327, 80.1327, 80.1256, 80.1163, 80.1233, 80.135, 80.1468, 80.1585, 80.1608, 80.1632, 80.1796,
        80.1726, 80.1984, 80.2054, 80.2219, 80.2289, 80.243, 80.2524, 80.2711, 80.2829, 80.2993, 80.3134, 80.3369, 80.3509, 80.3322, 80.3157, 80.2923, 80.2758, 80.2477, 80.2242, 80.2101, 80.2148,
        80.2148, 80.2078, 80.189, 80.1679, 80.1773, 80.189, 80.1984, 80.2007, 80.2148, 80.2242, 80.2219, 80.2289, 80.2477, 80.2524, 80.2594, 80.2805, 80.3017, 80.3064, 80.3275, 80.3251, 80.3251,
        80.3275, 80.3275, 80.3345, 80.3416, 80.3486, 80.3627, 80.3744, 80.3838, 80.3744, 80.3744, 80.3932, 80.4026, 80.4049, 80.3838, 80.365, 80.3674, 80.3838, 80.3885, 80.3861, 80.3932, 80.4073,
        80.426, 80.4542, 80.4683, 80.4753, 80.4519, 80.4425, 80.4659, 80.4871, 80.5011, 80.5152, 80.5152, 80.4918, 80.4659, 80.4659, 80.4495, 80.4167, 80.4026] };
  return WCL;
}

%
% Code from src/misc/atomic/xstar/XSTAR_read_pops.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define XSTAR_read_pops()
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{XSTAR_read_pops}
%\synopsis{reads the ionization balances from an XSTAR population file}
%\usage{Struct_Type XSTAR_read_pops(String_Type filename)}
%\qualifiers{
%\qualifier{verbose}{}
%\qualifier{Z}{array of Z values of elements to include}
%\qualifier{ions}{array of ions to include (overrides Z qualifier)}
%}
%!%-
{
  variable filename;
  switch(_NARGS)
  { case 1: filename = (); }
  { help(_function_name()); return; }

  variable nr_shells = fits_get_num_hdus(filename)-2;
  variable F = fits_open_file(filename, "r");
  ()=fits_movrel_hdu(F, 1);  % skip primary extension
  variable parameters = fits_read_table(F);
  variable density = parameters.value[wherefirst(parameters.parameter=="density")];
  variable luminosty = 1e38 * parameters.value[wherefirst(parameters.parameter=="rlrad38")];

  variable ions = qualifier("ions");
  if(ions==NULL)
  {
    variable Zs = qualifier("Z", [1,2,6,7,8,10,12,14,16,18,20,26]);
    ions = String_Type[int(sum(Zs))];
    variable Z, ion, i=0;
    foreach Z (Zs)
      _for ion (1, Z, 1)
      { ions[i] = strlow(atom_name(Z)+"_"+Roman(ion));
	i++;
      }
  }
  variable field, fields = ["r", "r_outer", "log_xi", "temp", "press", ions];
  variable pops = @Struct_Type(fields);
  foreach field (fields)
    set_struct_field(pops, field, Float_Type[nr_shells]);

  variable verbose = qualifier_exists("verbose");
  _for i (0, nr_shells-1, 1)
  {
    if(verbose)  vmessage("%d/%d", i, nr_shells);
    ()=fits_movrel_hdu(F, 1);
    (pops.r[i], pops.r_outer[i], pops.temp[i], pops.press[i]) = fits_read_key(F, "rinner", "router", "temperat", "pressure");
    variable shell = fits_read_table(F, ["population", "ion", "e_excitation"]);
    foreach ion (ions)
    {
      variable ind = where(shell.ion==ion and shell.e_excitation==0);
      if(length(ind))
	get_struct_field(pops, ion)[i] = shell.population[ind[0]];
    }
  }
  fits_close_file(F);

  pops.log_xi = log10(luminosty/density/pops.r^2);
  pops.temp *= 1e4;
  return pops;
}

%
% Code from src/slang/arrays/__push_array.sl
%
%%%%%%%%%%%%%%%%%%%
define __push_array()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{__push_array}
%\synopsis{pushes the values of an array onto the stack}
%\usage{(Any_Type v1, v2, ...) = __push_array(Any_Type a[]);}
%\description
%   NOTE that the Stack is limted to 2499 Elements, if the length
%   of a exceeds this value a Stack_Overflow Error is thrown!
%!%-
{
  switch(_NARGS)
  { case 1: variable a = (); }
  { help(_function_name()); return; }

  foreach (a)
    ; % left on stack
}

%
% Code from src/fitting/_fitfun_cache.sl
%
% *** IMPORTANT NOTE ***
% We need to have this script been copied into the isiscripts.sl
% before most of all remaining scripts! That is due to calls to
% fitfun_init_cache in these others scripts, which requires this
% function to be *fully* defined already (and not just its existence)!
% That's why this script is named starting with and underscore.

%!%+
%\function{fitfun_cache}
%\usage{see below}
%\description
%    The so-called caching extension is part of the ISISscripts and not
%    an ISIS internal feature.
%    
%    The caching extension provides an easy way to manage additional data for
%    the internal usage of user-defined fit-functions. The cache consists of
%    an slang structure for each defined dataset and fit-function using this
%    extension. This concept is similar to the metadata associated to each
%    dataset (see, e.g., 'get_dataset_metadata').
%
%    Inside of a user-defined fit-function, e.g., 'myfun' the cache for this
%    function and the current dataset (Isis_Active_Dataset) can be retrieved
%    by 'fitfun_get_cache':
%#v+
%    define myfun_fit(lo, hi, pars) {
%      ...
%      variable cache = fitfun_get_cache();
%      ...
%    }
%#v-
%    Here, the variable 'cache' would hold the corresponding slang structure.
%    Since slang structures are internally passed and returned via references,
%    any changes of the fields of 'cache' are permanent.
%
%    In order to define and/or initialize the fields of the structure, one
%    has to call 'fitfun_init_cache' and providing the slang structure.
%    For the example above this could look like
%#v+
%    fitfun_init_cache("myfun", &myfun_fit, struct { tempresult, lastpars });
%#v-
%    The caching extension can be used to, e.g., save temporary results from
%    calculations, which can then be retrieved once the model is evaluated
%    again. For this purpose it might be useful to also save the parameters
%    in a field like 'lastpars' into the cache, which can then be checked
%    against changes:
%#v+
%    define myfun_fit(lo, hi, pars) {
%      ...
%      variable cache = fitfun_get_cache();
%      if (any(cache.lastpars != pars)) {
%        ...
%      }  
%      cache.lastpars = pars;
%      ...
%    }
%#v-
%    Please read the documentation of 'fitfun_get_cache' and
%    'fitfun_init_cache' fot details and further examples.
%\seealso{fitfun_get_cache, fitfun_init_cache, fitfun_cache_enabled}
%
%!%-

private variable _fitfun_cache_enabled = 1;
private variable _fitfun_cache = Struct_Type[0];
private variable _fitfun_internal_cache = Assoc_Type[Struct_Type];
private variable _fitfun_name = Assoc_Type[String_Type];

%%%%%%%%%%%%%%%%%%%%%
define fitfun_cache_enabled()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fitfun_cache_enabled}
%\synopsis{returns whether the ISISscripts caching extension
%    for fit-functions is enabled (1) or disabled (0).}
%\usage{Integer_Type fitfun_cache_enabled();}
%\seealso{fitfun_cache, fitfun_enable_cache, fitfun_disable_cache}
%!%-
{
  return _fitfun_cache_enabled;
}

%%%%%%%%%%%%%%%%%%%%%
define fitfun_enable_cache()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fitfun_enable_cache}
%\synopsis{enables the ISISscripts caching extension}
%\usage{fitfun_enable_cache();}
%\seealso{fitfun_cache, fitfun_disable_cache, fitfun_cache_enabled}
%!%-
{
  _fitfun_cache_enabled = 1;
}

%%%%%%%%%%%%%%%%%%%%%
define fitfun_disable_cache()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fitfun_disable_cache}
%\synopsis{disables the ISISscripts caching extension}
%\usage{fitfun_disable_cache();}
%\seealso{fitfun_cache, fitfun_enable_cache, fitfun_cache_enabled}
%!%-
{
  _fitfun_cache_enabled = 0;
}

%%%%%%%%%%%%%%%%%%%%%
define fitfun_get_cache()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fitfun_get_cache}
%\synopsis{return the cache for the current dataset and the given fit-function}
%\usage{Struct_Type fitfun_get_cache([String_Type fit-function_name]);}
%\description
%    The cached slang structure for the currently evaluated
%    fit-function and the current dataset (Isis_Active_Dataset) is
%    returned. In case caching is disabled (by 'fitfun_disable_cache')
%    a new structure as defined using 'fitfun_init_cache' is returned.
%
%    The optional and only parameter specifies the name of a
%    fit-function, which stack is to be returned. This allows to share
%    and access the cache between different fit-functions. If this
%    particular fit-function has not been evaluated yet (or is not
%    used in the current model) NULL is returned.
%
%    Note: if the number of bins of the data grid of the current
%          dataset changes, the cache is initialized again with the
%          structure defined by 'fitfun_init_cache'. This might be
%          necessary as soon as, e.g., the fit-function is
%          'eval_fun'ed on a user-grid.
%\example
%    % use the cache inside a fit-function in order to save a temporary
%    % result, which depends on a subset of the input parameters only
%    define myfun_fit(lo, hi, pars) {
%      ...
%      variable cache = fitfun_get_cache();
%      if (any(cache.lastpars[[:1]] != pars[[:1]])) {
%        % assuming that 'expensive_calculation' depends on the first
%        two fit-parameters  
%        cache.tempresult = expensive_calculation(pars[0], pars[1]);
%      }  
%      cache.lastpars = pars;
%      ...
%      return cache.tempresult ...;
%    }
%
%    % retrieve the cache from another fit-function inside a fit-function
%    define mybetter_fit(lo, hi, pars) {
%      ...
%      variable ocache = fitfun_get_cache("myfun");
%      if (ocache != NULL) {
%        ocache.tempresult ...
%        ...
%      }
%      ...  
%    }
%
%    % investigate the temporary results after a fit from the command line
%    isis> Isis_Active_Dataset = 2; % get the cache for this dataset
%    isis> cache = fitfun_get_cache("myfun");
%    isis> print(cache.tempresult);
%\seealso{fitfun_cache, fitfun_init_cache}
%!%-
{
  % determine the cache's name from the frame stack
  variable fitfun, remote = 0, num_bins = NULL;
  if (_NARGS == 0) {
    variable call = _get_frame_info(_get_frame_depth()-1);
    % function was called from the known fit-function
    if (assoc_key_exists(_fitfun_name, call.function)) {
      fitfun = _fitfun_name[call.function];
      % get the number of bins of the grid the function was called with
      num_bins = length(_get_frame_variable(_get_frame_depth()-1, call.locals[0]));
    }
    % function was probably called from another function
    else {
      throw RunTimeError, sprintf("fitfun_get_cache called without an argument or from the unknown function '%s'", call.function);
    }
  }
  % the name is given as an argument -> remote access
  else {
    remote = 1;
    fitfun = ();
  }
  
  if (Isis_Active_Dataset < 1) {
    throw RunTimeError, "no cache for dataset 0 defined"$;
  }
  
  % return fresh structure if caching is disabled
  ifnot (fitfun_cache_enabled) {
    return remote ? NULL : struct_copy(_fitfun_internal_cache[fitfun].init);
  }

  % check the length of the cache-struct-array
  if (length(_fitfun_cache) < Isis_Active_Dataset) {
    % increase the array length
    loop (Isis_Active_Dataset-length(_fitfun_cache)) {
      _fitfun_cache = [_fitfun_cache, @Struct_Type(String_Type[0])];
    }
  }
  
  % check the struct-field on existence 
  ifnot (struct_field_exists(_fitfun_cache[Isis_Active_Dataset-1], fitfun)) {
    if (remote) { return NULL; }
    ifnot (assoc_key_exists(_fitfun_internal_cache, fitfun)) {
      throw RunTimeError, sprintf("no initializing struct for fit-function '%s' found"$, fitfun);
    }
    _fitfun_cache[Isis_Active_Dataset-1] = struct_combine(
      _fitfun_cache[Isis_Active_Dataset-1], @Struct_Type(fitfun)
    );
    set_struct_field(
      _fitfun_cache[Isis_Active_Dataset-1], fitfun,
      struct_copy(_fitfun_internal_cache[fitfun].init)
    );
  }

  % check the length of the internal number of bins
  if (length(_fitfun_internal_cache[fitfun].num_bins) < Isis_Active_Dataset) {
    % increase the array length
    loop (Isis_Active_Dataset-length(_fitfun_internal_cache[fitfun].num_bins)) {
      _fitfun_internal_cache[fitfun].num_bins = [_fitfun_internal_cache[fitfun].num_bins, num_bins];
    }
  }

  % reset cache in case the number of bins of the current grid changed (eval_fun!!!)
  if (num_bins != NULL && num_bins != _fitfun_internal_cache[fitfun].num_bins[Isis_Active_Dataset-1]) {
    set_struct_field(
      _fitfun_cache[Isis_Active_Dataset-1], fitfun,
      struct_copy(_fitfun_internal_cache[fitfun]).init
    );
    _fitfun_internal_cache[fitfun].num_bins[Isis_Active_Dataset-1] = num_bins;
  }
  
  % return the cache for the fit-function
  return get_struct_field(_fitfun_cache[Isis_Active_Dataset-1], fitfun);
}

%%%%%%%%%%%%%%%%%%%%%
public define fitfun_init_cache()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fitfun_init_cache}
%\synopsis{}
%\usage{fitfun_init_cache(String_Type fit-function_name, Ref_Type fit-function_handle, Struct_Type init_struct);}
%\description
%    Defines the structure, which is used to initialize the cache of the
%    the given fit-function. This should be called after a user-defined
%    fit-function, which uses caching, has been added to the list of
%    available models.
%
%    The first parameter 'fit-function_name' specifies the name of the
%    fit-function, 'fit-function_handle' is the reference to the actual
%    function calculating the model, and 'init_struct' is the structure
%    the cache will be initialized with for each dataset.
%\example
%    % Define a new fit-function, add it to the list of available
%    % models, and initialize its cache with a structure with empty
%    % fields. Inside the fit-function 'myfun' the value of lastpars
%    % can be checked on NULL or changed parameters in order to
%    % trigger an expensive calculation.
%
%    define myfun_fit(lo, hi, par) {
%      ...
%    }
%  
%    add_slang_function("myfun", ["par1", "par2", ...]);
%
%    fitfun_init_cache(
%      "myfun", &myfun_fit, struct { tempresult, lastpars = NULL }
%    );
%\seealso{fitfun_cache, fitfun_get_cache}
%!%-
{
  variable fitfun, handle, s;;
  switch (_NARGS)
    { case 3: (fitfun, handle, s) = (); }
    { help(_function_name); }

  if (typeof(handle) != Ref_Type) {
    vmessage("error (%s): handle has to be of Ref_Type", _function_name);
    return;
  }
  _fitfun_name[substr(string(handle), 2, -1)] = fitfun;

  % set the initializing structure
  _fitfun_internal_cache[fitfun] = struct {
    init = s,
    num_bins = Integer_Type[0]
  };
}

%
% Code from src/slang/arrays/_maximum.sl
%
define _maximum()
%!%+
%\function{_maximum}
%\synopsis{computes the maximum of >=2 values}
%\usage{Array_Type _maximum(Array_Type x1, x2, ...)}
%\description
%    \code{x1}, \code{x2}, ... have to be arrays of the same length, or scalars.
%    The \code{_maximum} function returns an array of the maximum values
%    by successively calling the \code{_max} function.
%    If all \code{x1}, \code{x2}, ... are scalars, \code{_maximum} returns a scalar value,
%    in this case, \code{_maximum(x1, x2, ...)} is equivalent to \code{max([x1, x2, ...])}.
%\seealso{_max, max, _minimum, _min, min}
%!%-
{
  if(_NARGS<2)
    return help(_function_name());
  variable m = _max((), ());
  loop(_NARGS-2)  m = _max(m, ());
  return m;
}

%
% Code from src/slang/arrays/_minimum.sl
%
%%%%%%%%%%%%%%%
define _minimum()
%%%%%%%%%%%%%%%
%!%+
%\function{_minimum}
%\synopsis{computes the minimum of >=2 values}
%\usage{Array_Type _minimum(Array_Type x1, x2, ...)}
%\description
%    \code{x1}, \code{x2}, ... have to be arrays of the same length, or scalars.
%    The \code{_minimum} function returns an array of the minimum values
%    by successively calling the \code{_min} function.
%    If all \code{x1}, \code{x2}, ... are scalars, \code{_minimum} returns a scalar value, too,
%    in this case, \code{_minimum(x1, x2, ...)} is equivalent to \code{min([x1, x2, ...])}.
%\seealso{_min, min, _maximum, _max, max}
%!%-
{
  if(_NARGS<2)
    return help(_function_name());
  variable m = _min((), ());
  loop(_NARGS-2)  m = _min(m, ());
  return m;
}

%
% Code from src/plot/_pgplot.sl
%
%!%+
%\function{_pgaxis}
%\synopsis{draw a labelled graph axis}
%\usage{_pgaxis(opt, x1, y1, x2, y2, v1, v2, step, nsub, dmajl dmajr, fmin, disp, orient);}
%\description
%    Besides \code{String_Type opt} and \code{Integer_Type nsub},
%    all parameters are of \code{Double_Type}.\n
%    \code{_pgaxis} draws an axis from world-coordinate position (\code{x1},\code{y1}) to (\code{x2},\code{y2}).
%
%    Normally, this routine draws a standard linear axis with equal subdivisions.
%    The quantity described by the axis runs from \code{v1} to \code{v2}.
%    If the '\code{L}' option is specified, the routine draws a logarithmic axis.
%    In this case, the quantity described by the axis runs from \code{10^v1} to \code{10^v2}.
%    A log. axis always has major, labeled, tick marks spaced by one or more decades.
%    If the major tick marks are spaced by one decade (as specified by the \code{step} argument),
%    then minor tick marks are placed at 2, 3, ..., 9 times each power of 10;
%    otherwise minor tick marks are spaced by one decade.  If the axis spans
%    less than two decades, numeric labels are placed at 1, 2, and 5 times each
%    power of ten.  If the axis spans less than one decade, or if it spans many decades,
%    it is preferable to use a linear axis labeled with log(quantity of interest).
%
%    Arguments:\n
%    \code{opt}    : a string containing single-letter codes for various options.\n
%             The options currently recognized are:\n
%             \code{L} : draw a logarithmic axis\n
%             \code{N} : write numeric labels\n
%             \code{1} : force decimal labelling, instead of automatic choice\n
%             \code{2} : force exponential labelling, instead of automatic.\n
%    \code{x1}, \code{y1} : world coordinates of one endpoint of the axis.\n
%    \code{x2}, \code{y2} : world coordinates of the other endpoint of the axis.\n
%    \code{v1}     : axis value at first endpoint.\n
%    \code{v2}     : axis value at second endpoint.\n
%    \code{step}   : major tick marks are drawn at axis value \code{0.0} plus or minus
%             integer multiples of step.  If \code{step==0.0}, a value is chosen automatically.\n
%    \code{nsub}   : minor tick marks are drawn to divide the major divisions into \code{nsub} equal
%             subdivisions (ignored if \code{step==0.0}).   If \code{nsub <= 1},
%             no minor tick marks are drawn. \code{nsub} is ignored for a logarithmic axis.\n
%    \code{dmajl}  : length of major tick marks drawn to left of axis
%             (as seen looking from first endpoint to second),
%             in units of the character height.\n
%    \code{dmajr}  : length of major tick marks drawn to right of axis,
%             in units of the character height.\n
%    \code{fmin}   : length of minor tick marks, as fraction of major.\n
%    \code{disp}   : displacement of baseline of tick labels to right of axis,
%             in units of the character height.\n
%    \code{orient} : orientation of label text, in degrees;
%             angle between baseline of text and direction of axis (0-360 deg).
%\seealso{http://www.astro.caltech.edu/~tjp/pgplot/subroutines.html#PGAXIS}
%!%-

%!%+
%\function{_pgbox}
%\synopsis{annotate the viewport with frame, axes, numeric labels, etc.}
%\usage{_pgbox(xopt, xtick, nxsub,  yopt, ytick, nysub);}
%\description
%    \code{x}/\code{yopt}: string of options for X (horizontal) / Y (vertical) axis of plot.
%            Options are single letters, and may be in any order:\n
%            "A": draw Axis (X axis is horizontal line Y=0, Y axis is vertical line X=0).
%            "B": draw bottom (X) or left (Y) edge of frame.
%            "C": draw top (X) or right (Y) edge of frame.
%            "G": draw Grid of vertical (X) or horizontal (Y) lines.
%            "I": Invert the tick marks; ie draw them outside the viewport instead of inside.
%            "L": label axis Logarithmically (see below).
%            "N": write Numeric labels in the conventional location
%                 -- below the viewport (X) or to the left of the viewport (Y).
%            "P": extend ("Project") major tick marks outside the box
%                 (ignored if option I is specified).
%            "M": write numeric labels in the unconventional location
%                 -- above the viewport (X) or to the right of the viewport (Y).
%            "T": draw major Tick marks at the major coordinate interval.
%            "S": draw minor tick marks (Subticks).
%            "V": orient numeric labels Vertically. This is only applicable to Y.
%                 The default is to write Y-labels parallel to the axis.
%            "1": force decimal labelling, instead of automatic choice (see PGNUMB).
%            "2": force exponential labelling, instead of automatic.
%
%    \code{x}/\code{ytick}: world coordinate interval between major tick marks on X/Y axis.
%             If \code{x}/\code{ytick==0}, the interval is chosen by _pgbox,
%             so that there will be at least 3 major tick marks along the axis.
%
%    \code{n}\{\code{x}/\code{y}\}\code{sub}: number of subintervals to divide the major coordinate interval into.
%               If \code{x}/\code{ytick==0} or \code{nx/ysub==0},the number is chosen by _pgbox.
%
%    To get a complete frame, specify BC in both \code{xopt} and \code{yopt}.
%    Tick marks, if requested, are drawn on the axes or frame or both,
%    depending which are requested. If none of ABC is specified,
%    tick marks will not be drawn.
%
%    For a logarithmic axis, the major tick interval is always 1.0.
%    The numeric label is 10^x where x is the world coordinate at the tick mark.
%    If subticks are requested, 8 subticks are drawn between each major tick at equal logarithmic intervals.
%\seealso{http://www.astro.caltech.edu/~tjp/pgplot/subroutines.html#PGBOX}
%!%-

%!%+
%\function{_pgsls (set line style)}
%\synopsis{set the line style attribute for subsequent plotting}
%\usage{_pgsls(Integer_Type linestyle);}
%\description
%    This attribute affects line primitives only;
%    it does not affect graph  markers, text, or area fill.
%    Five different line styles are available, with the following codes:
%    1 (full line), 2 (dashed), 3 (dot-dash-dot-dash), 4 (dotted),
%    5 (dash-dot-dot-dot). The default is 1 (normal full line).
%!%-

%
% Code from src/misc/aStar.sl
%
% estimates path from (x0,y0) to (x1,y1) by their distance
private define aStar_estimateCost(x0, y0, x1, y1) { return sqrt(floor(x0*x1 + y0*y1)); }

% returns length of best path starting at (x,y)
private define aStar_pathLength(cameFrom, x, y) {
  variable len = 1;
  while (cameFrom[x,y] != NULL) {
    (x,y) = (cameFrom[x,y].x,cameFrom[x,y].y);
    len++;
  }
  return len;
}

% returns best path starting at (x,y) and its cost
private define aStar_constructPath(cameFrom, costMap, x, y) {
  variable pathX = x, pathY = y, cost = costMap[x,y];
  % follow path backwards
  while (cameFrom[x,y] != NULL) {
    (x,y) = (cameFrom[x,y].x,cameFrom[x,y].y);
    pathX = [pathX,x]; pathY = [pathY,y];
    ifnot (qualifier_exists("noSum")) cost += costMap[x,y];
  }
  return struct { x = pathX, y = pathY, cost = cost };
}

%%%%%%%%%%%%%%%%%%%%%
define aStar() {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{aStar}
%\synopsis{pathfinding algorithm A*}
%\usage{Struct_Type aStar(Double_Type[][] graph, Integer_Type startX, startY, endX, endY);}
%\qualifiers{
%    \qualifier{max}{find the most expensive path instead of the cheapest one}
%    \qualifier{meanCost}{costs are normalized by the best path length to any point
%               in the graph, resulting in a mean cost on a path}
%    \qualifier{estimate}{reference to a function, which estimates the cost between
%               two nodes in the graph (parameters: x0, y0, x1, y1). By
%               default, the distance between both nodes is used}
%    \qualifier{plot}{plots the graph and the working progress at each step. Nodes
%               in the open list are shown in red, in the closed list in
%               blue, nodes of infinite cost in green and the best path,
%               finally, in black. The function sleeps 0.01 seconds after
%               each step or the value given with this qualifier}
%}
%\description
%    Pathfinding algorithm with high performance and accuracy by
%      Hart, Nilsson & Raphael, "A Formal Basis for the Heuristic
%      Determination of Minimum Cost Paths", IEEE Transactions on
%      Systems Science and Cybernetics SSC4 (2), 100–107, 1968
%                               
%    The A* algorithm (called A Star) finds the shortest (or in general
%    cheapest) path between to nodes in a graph. The algorithm looks for
%    the best way following the lowest known heuristic cost. This cost
%    has to be estimated at each new discovered node and is, by default,
%    the straight-line distance between both points. The runtime and
%    accuracy of the search strongly depends on the choice of this
%    estimation.
%
%    The 'graph' has to be given as 2d-array defining the cost at each
%    point (node). Infinite costs (_Inf) are allowed and corresponds to
%    insuperable borders. The starting node and goal node habe to be
%    given as array indices 'startX', 'startY' and 'endX', 'endY',
%    respectively.
%
%    The returned structure contains
%      x    - x-indices of the best path
%      y    - y-indices of the best path
%      cost - summed costs along the best path
%!%-
  variable map, x0, y0, x1, y1;
  switch (_NARGS)
    { case 5: (map,x0,y0,x1,y1) = (); }
    { help(_function_name); return; }

  % type checks
  ifnot (length(array_shape(map)) == 2) { vmessage("error (%s): given map has to be a 2d-array", _function_name); return; }
  ifnot (__is_numeric(map)) { vmessage("error (%s): given map has to contain numerical values", _function_name); return; }
  ifnot (typeof(x0) == Integer_Type && typeof(y0) == Integer_Type && typeof(x1) == Integer_Type && typeof(y1) == Integer_Type)
    { vmessage("error (%s): given nodes have to be integer numbers", _function_name); return; }
  % check if given nodes lie within the map
  ifnot (0 <= x0 < length(map[*,0])) { vmessage("error (%s): x0 lies out of given map", _function_name); return; }
  ifnot (0 <= y0 < length(map[0,*])) { vmessage("error (%s): y0 lies out of given map", _function_name); return; }
  ifnot (0 <= x1 < length(map[*,0])) { vmessage("error (%s): x1 lies out of given map", _function_name); return; }
  ifnot (0 <= y1 < length(map[0,*])) { vmessage("error (%s): y1 lies out of given map", _function_name); return; }
  % qualifiers
  variable estimateFun = qualifier("estimate", &aStar_estimateCost);
  if (typeof(estimateFun) != Ref_Type) { vmessage("error (%s): estimate qualifier has to be a reference", _function_name); return; }
  variable doplot = qualifier("plot", NULL);
  if (qualifier_exists("plot") && doplot == NULL) doplot = .01;
  
  % initialise open and closed lists
  variable openList_x = x0, openList_y = y0, openList_f = 0.;
  variable closedList_x = Integer_Type[0], closedList_y = Integer_Type[0];
  % initialise cost map
  variable costMap = Double_Type[length(map[*,0]),length(map[0,*])] + _Inf; % cost from starting node to current node
  costMap[x0,y0] = map[x0,y0];
  % initialise map of navigated nodes
  variable cameFrom = Struct_Type[length(map[*,0]),length(map[0,*])];
  % repeat until best way is found or no solution exists
  if (doplot != NULL) point_style(4);
  while (length(openList_x) > 0) {
    % get node corresponding to best f-value
    variable i = (qualifier_exists("max") ? where_max(openList_f)[0] : where_min(openList_f)[0]);
    variable node = struct { x = openList_x[i], y = openList_y[i], f = openList_f[i] };
    % if node is the destination return the best path
    if (node.x == x1 && node.y == y1) {
      variable best = aStar_constructPath(cameFrom, costMap, x1, y1; noSum);
      if (doplot != NULL) {
	color(1);
        oplot(best.y,best.x);
      }
      return best;
    }
    % move actual indices from open to closed list
    closedList_x = [closedList_x, node.x];
    closedList_y = [closedList_y, node.y];
    if (doplot != NULL) { color(4); oplot(node.y,node.x); }
    openList_x = array_remove(openList_x, i);
    openList_y = array_remove(openList_y, i);
    openList_f = array_remove(openList_f, i);
    % loop over neighbors of the actual node
    variable n;
    foreach n ({[0,1],[1,0],[0,-1],[-1,0]}) {
      variable neigh = struct { x = node.x + n[0], y = node.y + n[1] };
      if (0 <= neigh.x < length(map[*,0]) && 0 <= neigh.y < length(map[0,*])) {
        % proceed if neighbor is not in closed list
        if (wherefirst(closedList_x == neigh.x and closedList_y == neigh.y) == NULL)
        {
	  if (isinf(map[neigh.x,neigh.y])) { if (doplot != NULL) color(3); }	  
	  else {
	    if (doplot != NULL) color(2);
    	    % tentative cost to get to the neighbor node
	    variable cost;
	    if (qualifier_exists("meanCost")) {
    	      variable len = aStar_pathLength(cameFrom, node.x, node.y) + 1;
  	      cost = costMap[node.x,node.y]*(len-1)/len + map[neigh.x,neigh.y]/len;
	    } else cost = costMap[node.x,node.y] + map[neigh.x,neigh.y];
	    % add neighbor to open list if not already in it or costs are better than before
	    if (wherefirst(openList_x == neigh.x and openList_y == neigh.y) == NULL
		|| (qualifier_exists("max") ? cost > costMap[neigh.x,neigh.y] : cost < costMap[neigh.x,neigh.y])) {
	      openList_x = [openList_x, neigh.x]; openList_y = [openList_y, neigh.y]; openList_f = [openList_f, cost + @estimateFun(neigh.x, neigh.y, x1, y1)];
	      costMap[neigh.x,neigh.y] = cost;
	      cameFrom[neigh.x,neigh.y] = node;
	    }
	  }
	  if (doplot != NULL) {
	    oplot(neigh.y,neigh.x);
	    sleep(doplot);
	  }
        }
      }
    }
  };
  % no solution exists
  return aStar_constructPath(cameFrom, costMap, x1, y1; noSum);
}

%
% Code from src/FITS/ad_image/ad_create_colormap.sl
%

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

define ad_colormap_norm_col(array) {
   return int(array/max(array)*255.99999999999);
}

define ad_colormap_gauss(x, mu, Delta)
{ return exp(-((x-mu)/Delta)^2);
}

%%%%%%%%%%%%%%%%%%
% TD, 10/06/2010
%%%%%%%%%%%%%%%%%%
define ad_create_colormap(img,alp,gmax) {
%%%%%%%%%%%%%%%%%%
%!%+
%\function{ad_create_colormap}
%\synopsis{creates a redshift-colormap, where yellow is exactly at
%where(img==1)}
%\usage{g_br = ad_create_colormap(img,alp,gmax);}
%!%-
   variable R = [0:255]*0.;
   variable G = [0:255]*0.;
   variable B = [0:255]*0.;
   
   variable RI = Integer_Type[256];
   variable GI = Integer_Type[256];
   variable BI = Integer_Type[256];
   
   variable g_br;
   if (gmax <= 1) {
      g_br = 256;
   } else {
      % where is redshift == 1?
      g_br = int(1./gmax*255.999999999);
   }
   
   
   
      %% colors before g=1
   
   variable i1 = [0:g_br-1];
   R[i1] = i1^(1./2.2);
   RI[i1] = ad_colormap_norm_col(R[i1]);


   BI[i1] = 0;

   G = ad_colormap_gauss([0:255],g_br,g_br/2.5);

   %% color after g=1
   if (gmax > 1) {
      variable i2 = [g_br:255];
      variable i21= [g_br:g_br+(255-g_br)/2];
      variable i22 = [max(i21)+1:255];

      R[i21] = (sqrt(max(i21)-i21))^(2.);
      RI[i21] = ad_colormap_norm_col(R[i21]);
      R[i22] = ((i22-min(i22))/(1.*max(i22)-min(i22)))^(1.1);
      RI[i22] = ad_colormap_norm_col(R[i22])/3*2;

      B[i21] = ((i21-g_br)/(max(i21)*1.-g_br)*1.)^(1.);
      BI[i21] = ad_colormap_norm_col(B[i21]);

      B[i22] = ad_colormap_gauss(i22,min(i22),(255-g_br)/2.);
      BI[i22] = ad_colormap_norm_col(B[i22]);

      G[i21] = (sqrt(max(i21)-i21)/sqrt(255-g_br))^(2.);
      GI[i21] = ad_colormap_norm_col(G[i21]);
      G[i22] = 0.;%(sqrt(255-i2)/sqrt(255-g_br))^(2.);
   }
   GI[i1] = ad_colormap_norm_col(G[i1]);

     
   png_add_colormap("redshift", ((RI << 16) | (GI << 8) | BI ));
   
      
   %% PLOT COLOR-SCHEME ???
   if (qualifier_exists("plot")) {
      variable x = [0:255];
      color(2);
      plot(x,RI);oplot(x,GI);oplot(x,BI);
   }
   
   return g_br;
}


#endif

%
% Code from src/FITS/ad_image/ad_init_raw.sl
%
%%%%%%%%%%%%%%%%%%%%%
% TD, 2010-06-07
%%%%%%%%%%%%%%%%%%%%%
define ad_init_raw(fitstable) {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ad_init_raw}
%\synopsis{subroutine, which reads the data and the header of a
%      raw-redshift image (used by ad_make_image) }
%\usage{(data,nre,ng,a,theta0grad) = ad_init_raw(filename);}
%!%-

   variable d = fits_read_table(fitstable);
   
   % nre is not the length of the table, but 2*nre , 
   % because here k=1 & 2 are displayed in one table
   variable nre = fits_read_key(fitstable,"nre");
   variable ng = fits_read_key(fitstable,"ng");
   variable a = fits_read_key(fitstable,"a");
   variable theta0grad = fits_read_key(fitstable,"theta0");  
  return d,nre,ng,a,theta0grad;
}

%
% Code from src/FITS/ad_image/ad_init_raw_timing.sl
%
%%%%%%%%%%%%%%%%%%%%%
% TD & Fabi Brod, 2013-08-12
%%%%%%%%%%%%%%%%%%%%%
define ad_init_raw_timing(fitstable) {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ad_init_raw_timing}
%\synopsis{subroutine, which reads the data and the header of a
%      raw-redshift image FOR TIMING (used by ad_make_image) }
%\usage{(data,nre,ng,a,theta0grad) = ad_init_raw_timing(filename);}
%!%-

   variable d = fits_read_table(fitstable);

   variable dim;
   (dim, , ) = array_info (d.xdisc);
   
   variable nr = dim[0];
   variable nphi = dim[1];
   return d,nr,nphi;
}

%
% Code from src/FITS/ad_image/ad_interpol_image.sl
%
%%%%%%%%%%%%%%%%%%
% TD, 09/06/2010
%%%%%%%%%%%%%%%%%%
define ad_interpol_image(g,alp_lo,bet_lo,sz_max,rmin){
%%%%%%%%%%%%%%%%%%
%!%+
%\function{ad_interpol_image}
%\synopsis{interpolates the image maximal sz_max times beyond rmin}
%\usage{image = ad_interpol_image(image,alp_lo,bet_lo,sz_max,rmin);}
%\seealso{ad_make_image, ad_init_raw, ad_read_image}
%!%-

   
   
   variable size = [length(alp_lo),length(bet_lo)];
   % +1 is needed, as alp_lo is only 
   variable num = Integer_Type[size[1],size[0]];   
   num[*,*] = 1;
   num[where(g == 0)] = 0;

   variable i,j;
   variable sz = 1;
   
   for (j = 1; j < size[1]; j++) {
      for (i = 0; i < (size[0]); i++) {
	 if ( sqrt(alp_lo[i]^2+bet_lo[j]^2) > rmin &&  (num[j,i] == 0) &&  (num[j,i-1] != 0)) {
	    sz = 1;
	    while ( (sz  <= sz_max) && ( i+sz < size[0]) && (num[j,i+sz] ==  0) ) {
	       sz++;
	    }
	    if (sz < sz_max and i+sz < size[0]){
	       g[j,i] = 1./(sz+1)*g[j,i-1] + sz/(sz+1.)*g[j,i+sz];
	       num[j,i] = 1;
	    } else {
	       g[j,i] = g[j,i-1];
	    }
	 }
	 
      }
   }
   for (i = 0; i < size[0]; i++) { 
      for (j = 1; j < (size[1]); j++) { 
	 if ( sqrt(alp_lo[i]^2+bet_lo[j]^2) > rmin &&  (num[j,i] == 0) &&  (num[j-1,i] != 0)) {
	    sz = 1;
	 while ( (sz  <= sz_max) && ( j+sz < size[1]) && (num[j+sz,i] ==  0) ) {
	    sz++;
	 }
	    if (sz < sz_max and j+sz < size[1]){
	       g[j,i] = 1./(sz+1)*g[j-1,i] + sz/(sz+1.)*g[j+sz,i];
	       num[j,i] = 1;
	    } else {
	       g[j,i] = g[j-1,i];
	    }
      }
	 
      }
   }
   

   
   return g;
}


%
% Code from src/FITS/ad_image/ad_make_image.sl
%

%%%%%%%%%%%%%%%%%%
% TD, 2010-06-07
%%%%%%%%%%%%%%%%%%
define ad_make_image() {
%%%%%%%%%%%%%%%%%%
%!%+
%\function{ad_make_image}
%\synopsis{creates a redshift image from the raw data by sorting with
%respect to alpha and beta}
%\usage{(img, alp_lo, alp_hi, bet_lo, bet_hi) = ad_make_image(filename);}
%\qualifiers{
%\qualifier{sho_minmax}{show the extrem values of alpha and beta?}
%\qualifier{timing}{read a Brod-Timing-FITS-Table}
%\qualifier{field}{specify which values are plotted}
%\qualifier{disk}{kuse intrinsic x and y instead of alpha and beta (only working for timing yet)}
%}
%!%-

   variable fitstable,size;

   variable alp_extr = [0,0];
   variable bet_extr = [0,0];
   
   
   switch(_NARGS) 
   {case 2: (fitstable,size) = ();}
   {case 4: (fitstable,size, alp_extr,bet_extr) = (); }
   {help(_function_name()); return; }
   
   variable i,j,k,l,ind;
   variable d,nre,ng,nr,nphi;
   variable alp,bet;
   
   if (qualifier_exists("timing"))
   {
     (d,nr,nphi) = ad_init_raw_timing(fitstable);
      if (qualifier("disk"))
      {
	alp = d.xdisc;
	bet = d.ydisc;
      }
      else
      {
	alp = d.aproj;
	bet = d.bproj;	 
      }
   }
   else
   {
      (d,nre,ng,,) = ad_init_raw(fitstable);
      nr = 2*nre;
      nphi = ng;
      alp = d.alp;
      bet = d.bet;
   }

   % which value should be sorted
   variable val;   
   if (qualifier_exists("cosne"))
   {
      val = d.cosne;
   }
   else if (qualifier_exists("trff"))
   {
      val = d.trff;
   }
   else if (qualifier_exists("field"))
   {
      val = get_struct_field( d, qualifier("field") );
   }
   else
   {
      val = d.g;
   }
   

   
   
   if ((alp_extr[0] == alp_extr[1]) or (bet_extr[0] == bet_extr[1])){
      bet_extr = [min(bet),max(bet)];
      alp_extr = [min(alp),max(alp)];
   }
   
   if (qualifier_exists("sho_minmax")){
      message("\nalp = ["+sprintf("%.2f",alp_extr[0])+","+sprintf("%.2f",alp_extr[1])+
	      "] and bet = ["+sprintf("%.2f",bet_extr[0])+","+sprintf("%.2f",bet_extr[1])+"]");
   }
   
   variable bet_lo,bet_hi,alp_lo,alp_hi;
   (alp_lo,alp_hi) = linear_grid(alp_extr[0],alp_extr[1],size[0]);
   (bet_lo,bet_hi) = linear_grid(bet_extr[0],bet_extr[1],size[1]);
   
   variable g = Double_Type[size[1],size[0]];
   variable num = Integer_Type[size[1],size[0]];
   variable lum = Double_Type[size[1],size[0]];
   
   %  ++++++++++++++      SORT      +++++++++++++++++
   %  sort the arrays for ascending alpha in order to
   %  make reduce the final steps of sorting
   %  
   %  sort alpha in r-direction (I), by stepping through 
   %  the array in phi-direction (J)
   
   for (l=0; l < nphi;l++ ) {
      
      ind = array_sort(alp[[:-1],l]);
      
      alp[[:-1],l] = alp[ind,l];
      bet[[:-1],l] = bet[ind,l];
      val[[:-1],l] = val[ind,l];
   }
   
   %  ++++++++++++++   WRITE THE IMAGE     +++++++++++++++++
   
   %  ---- BEGIN: LOOP over the GRID -----
   %  remember: the array is already sorted
   %  
   %% define a starting value for each value of J
   variable start_I = Integer_Type[nphi]; 
   
   message("Calculating image:");
   for (i = 0; i < size[0]; i++) {             % ALPHA-LOOP  
      for(l = 0;l < nphi; l++) {                          % J-LOOP
	 for (k = start_I[l]; k < nr; k++) {             % I-LOOP
	    
	    % count only values that are in the correct bin
	    if (alp_lo[i] <= alp[k,l] < alp_hi[i]) {  
	       
	       for (j = 0; j < size[1]; j++) { 	            % BETA-LOOP
		  if (bet_lo[j] <= bet[k,l] < bet_hi[j]) {
		     % calculate the redshift for this grid-point
		     g[j,i] = g[j,i]*(num[j,i])/(num[j,i]+1.) 
		       + 1/(num[j,i]+1.)*val[k,l];	  
		     
		     %	    lum[j,i] += d.lum[k,l];
	     
		     num[j,i]++; 
		  }
	       }
	       
	       
	    } else if (alp_hi[i] <= alp[k,l]){  
	       start_I[l] = k; %set start_I to minimize efforts
	       break; % jump out of the loop if alpha is too large (sorted!)
	    }
	    
	 }
      } 
      if ((int((10000*(i+1)/size[0])) mod 500) == 0) {
	 message(string(int((100./size[0]*(i+1))))+"%");
      }
      
   }
   return g, alp_lo, alp_hi, bet_lo, bet_hi;
}

%
% Code from src/FITS/ad_image/ad_read_image.sl
%
%%%%%%%%%%%%%%%%%%
% TD, 2010-06-08
%%%%%%%%%%%%%%%%%%
define ad_read_image() {
%%%%%%%%%%%%%%%%%%
%!%+
%\function{ad_read_image}
%\synopsis{reads the image and the alpha-beta-grid from the FITS-file
%created by ad_make_image}
%\usage{ad_read_image(filename);}
%\seealso{ad_make_image, ad_init_raw}
%!%-
   
   variable f;
   
   switch(_NARGS)
   {case 1: f = ();}
   {help(_function_name()); return; }
   
%  #1)  read the image
   variable img = fits_read_img(f);
				  

   
%  #2)  get the minimum and maximum values of alp and bet
   variable alp_min = fits_read_key(f,"CRVAL1");
   variable alp_delta = fits_read_key(f,"CDELT1");
   variable alp_naxis = fits_read_key(f,"NAXIS1");
   alp_min -= 0.5*alp_delta;  % xfig needs the edge of the bin, but the fits-file uses the middle of th bin
   variable alp_max = alp_min + (alp_naxis+1)*alp_delta;

   variable bet_min = fits_read_key(f,"CRVAL2");
   variable bet_delta = fits_read_key(f,"CDELT2");
   variable bet_naxis = fits_read_key(f,"NAXIS2");
   bet_min -= 0.5*bet_delta;  % xfig needs the edge of the bin, but the fits-file uses the middle of th bin
   variable bet_max = bet_min + (bet_naxis+1)*bet_delta;

   
%  #3) reconstruct alpha and beta grid
   variable nalp_lo = [0:alp_naxis-1]*1./(alp_naxis);
   variable nalp_hi = [1:alp_naxis]*1./(alp_naxis);
   variable alp_lo = (alp_max-alp_min)*nalp_lo + alp_min;
   variable alp_hi = (alp_max-alp_min)*nalp_hi + alp_min;
   
   
   variable nbet_lo = [0:bet_naxis-1]*1./(bet_naxis);
   variable nbet_hi = [1:bet_naxis]*1./(bet_naxis);
   variable bet_lo = (bet_max-bet_min)*nbet_lo + bet_min;
   variable bet_hi = (bet_max-bet_min)*nbet_hi + bet_min;
      
   %  #4) return image and alp, bet grid
   return img,alp_lo,alp_hi,bet_lo,bet_hi;
}

%
% Code from src/FITS/ad_image/ad_write_image.sl
%
%%%%%%%%%%%%%%%%%%%%%
% TD, 2010-12-02
%%%%%%%%%%%%%%%%%%%%%
define ad_write_image() {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ad_write_image}
%\synopsis{subroutine, which reads the data and the header of a
%      raw-redshift image (ad_make_image) and writes it into a
%      fits-image with the filename "image_*"}
%\usage{(img,alp,bet) = ad_write_image(filename,size [alp_extr,bet_extr]);}
%!%-

   variable filename,size,alp_extr,bet_extr,filename0;
   variable g, alp_lo, alp_hi, bet_lo, bet_hi;
   variable output = "_image";
   variable suff = ".FITS";
   
   if (qualifier_exists("cosne")) {output = "_cosne"+output;}
   else if (qualifier_exists("trff")) {output = "_trff"+output;}
   
   switch (_NARGS)
   {case 2:
	(filename,size) = ();
      filename0 = filename;
      filename = substr(filename,1,strlen(filename)-5);
      (g, alp_lo, alp_hi, bet_lo, bet_hi) = ad_make_image(filename0,size;; __qualifiers());
      
   }
   {case 4:
	(filename,size,alp_extr,bet_extr) = ();
      filename0 = filename;
      filename = substr(filename,1,strlen(filename)-5);
      (g, alp_lo, alp_hi, bet_lo, bet_hi) = ad_make_image(filename0,size,alp_extr,bet_extr;; __qualifiers());
   }
   {
      help(_function_name()); return;
   }
   
   variable alp = 0.5*(alp_lo+alp_hi);
   variable bet = 0.5*(bet_lo+bet_hi);
   fits_write_image(filename+output+suff,"alp_bet",g,alp,bet,"alpha","beta");

  return g,alp_lo,alp_hi,bet_lo,bet_hi;
}

%
% Code from src/slang/structs/histograms/add_hist.sl
%

define add_hist(){
%!%+
%\function{add_hist}
%\synopsis{add histograms}
%\usage{Struct_Type specs = add_hist([hist1,hist2,...]);}
%\usage{Struct_Type specs = add_hist(hist1,hist2,...);}
%\description
%    Add histogram structures with fields bin_lo, bin_hi, value, and
%    err. The function just assumes that the histograms are on the
%    same grid. It copies the grid from the very first structure in
%    the array, then adds the values of the subsequent structures.
%    Uncertainties (err) are added in quadrature. 
%\seealso{scale_hist, shift_hist, stretch_hist}
%!%-   
   variable structs, ii;
   if(_NARGS==0)  { help(_function_name()); return; }
   else if(_NARGS==1){ structs = ();}
   else{ 
	  structs = __pop_list(_NARGS);
	  _for ii(0,_NARGS-1,1){
		 if( typeof( structs[ii] ) != Struct_Type ){ 
			help(_function_name()); return;
		 }
	  }
	  structs = list_to_array(structs);
   }
   	  
   
   variable nb = length(structs);
   variable ns = @structs[0];
   if(struct_field_exists(ns, "err")){
	  ns.err = ns.err^2;
   }else{
	  if(struct_field_exists(ns, "value")){
		 ns = create_struct_field(ns, "err", Double_Type[length(ns.value)] );
	  }else{
		 print("First struct missing 'value' field. Abort.");
		 return;
	  }
   }
   variable i;
   _for i (1,nb-1,1){
	  ifnot(struct_field_exists(structs[i], "value")){
		 print(sprintf("Struct #%d is missing the 'value' field. Skipping."));
		 continue;
	  }
	  ns.value += structs[i].value;
	  
	  if(struct_field_exists(structs[i],"err")){
		 ns.err += structs[i].err^2;
	  }
   }
   ns.err = sqrt(ns.err);
   
   return ns;
}

%
% Code from src/misc/satellites/Chandra/aglc.sl
%
% -*- mode: slang; mode: fold -*-


%; File: aglc.sl                "ACIS Grating Light Curve"
%; Author: David P. Huenemoerder <dph@space.mit.edu>
%; Original version: 2003.09.05
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% This file is part of aglc (Acis Grating Light Curve).
% Copyright (c) 2007-2008 Massachusetts Institute of Technology 
%
% This software was developed by the MIT Kavli Institute for
% Astrophysics and Space Research under contract SV3-73016 from the
% Smithsonian Institution.
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either  version 2 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
% General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
% 02110-1301, USA.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% versioning...
%
private variable _version = [1, 5, 3]   ; % major, minor, patch
variable aglc_version = sum( _version * [ 10000, 100, 1 ] ) ; 
variable aglc_version_string = sprintf("%d.%d.%d", 
_version[0],
_version[1],
_version[2] ) ; 
%
% PURPOSE: Compute light or phase curves from grating/ACIS event data.
%          Do chip by chip to apply proper GTI (implicitly from EXPNO)
%          then sum counts, average the exposure and compute rate.
%
%          Return a light curve structure.  Optionally write to file.
%
%          Parameters allow fairly general combinations of gratings
%          (HEG, MEG), orders, and wavelength regions.  E.g., lists of
%          wmin, wmax can be used to provide light curves for high-T
%          lines vs low-T lines.
%
%          Cross-dispersion limits can be set to select different
%          rectangular region in diffraction coordinates, such as for
%          binning backgrounds.
%
%  Unsupported:
%
%          Chip background (tg_part=99; outside Level 1.5 mask) is not
%          supported, since there is no wavelength assigned to the
%          non-grating order spatial region
%
%          Inter-order background (tg_m=99) is not supported, since
%          there are no wavelengths assigned to the unresolved events.
%
%          Zero-order is not supported, since there are no wavelengths
%          assigned in the zero order region. (use dmextract)
%
%          HRC: doesn't have complications of multi-chip GTI, so
%          dmextract and dm-filters will suffice.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INPUTS:
%
% evt1a: ACIS/grating events; must have grating coordinates.
%         Assumed "level 2": grade and status filtered (if not, a
%         filter can be specified).
% 
% stat1: Exposure reference; one EXPNO for every exposed frame.
%        stat1 is preferred, since it is smaller and will bin faster.
%        evt1 also has EXPNO, but could have dropped frames for very
%        bright sources.
%
% evt1: (optional) Unfiltered file for full exposure number set;
%       Can be given instead of stat1, but will be slower since will
%       contain many instances of each unique expno.
%
% tbin: time bin size, in seconds.
%
% wmin:  array of minima of wavelength ranges
%
% wmax:  array of maxima of wavelength ranges ... must be same number as in wmin 
%        These are paired 1-1
%        e.g.   [12.0], [12.2]                single line
%                12.0,  12.2                  same as above
%               [12.0, 18.9], [12.2, 19.1]    two regions, Ne10, O8
%
% tg: array of grating region types, one or both of ["HEG","MEG"], or "LEG"
%       Also accepted are scalars, lowercase, first character:
%                "h"
%                ["h", "MEG"]         etc
%                (but "LEG" should not occur w/ HEG or MEG).    
%
% orders:   array of integer orders;
%           e.g., [-1,1]; 1; [-3,-2,-1]; [[-3:-1],[1:3]];
%
% bkg: Optional flag;  any value means extract background curve
%      instead of source.
%
% Returned value is a  light curve structure:
% ( similar to that produced by dmextract opt=lc1 )
%
%     time_min = Double_Type[]
%     time_max = Double_Type[]
%     time = Double_Type[]
%     counts = Integer_Type[]
%     count_rate = Double_Type[]
%     stat_err = Double_Type[]
%     exposure = Double_Type[]
%
% additionally returned  in the structure are:
%     timezero  (== TSTART)
%     mjdref
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% GENERAL PROCEDURE: 
%
%     Form a filter from the parameters: wmin, wmax, tg, orders.  For
%     the source region, the filter is the default tg_d_min to
%     tg_d_max range.
%
%     If background, then the filter excludes the source region, and
%     backscale is computed as the ratio of bkg region width to source
%     region width, and the rate is scaled to the src region width.
%
%     The filter will specify one or more CCDs.  Each must be processed
%     independently to account for the proper exposure.    
%
%     The light curve will be made by binning the number of
%     occurrences of each exposure number (EXPNO) vs exposure number,
%     and converted to a rate vie each frame's exposure time of
%     TIMEDEL-0.04104 seconds.
%
%     The exposure per frame is done implicitly from the frames
%     present within the entire range of frames.  It is assumed that
%     the input events have already been filtered on the GTI.  NOTE:
%     this could easily be relaxed by referencing an appropriate set
%     of GTI tables and intersecting them with event times.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% REVISION HISTORY.
%
% 2011-02-08 - v-1.5.3 (Mh)
%  - fixed aglc_read_events().READ_GRP_PER_EXPNO for CC mode (=> 512)
%  - modified aglc_read_events().tstart such that times calculated
%    from expno's will better agree with original times
%  - _aglc2 adjusts the time bin size to a multiple of the frame time
%    s.timedel * s.READ_GRP_PER_EXPNO
%    instead of the effective frame exposure s.frame_exp
%  - remove xround in favour of S-Lang's intrinsic nint function
%
% 2010.09.16 - v-1.5.2
%  change the internal round() function to xround().
%  It is private, but one user (M.H.) reported a conflict with the
%  intrinsic round.  (I should use the intrinsic round() or nint()).
%  My round() fails on negative numbers, which don't matter here.
%  As a patch, rename it.
%
% 2010.08.24 - v-1.5.1
%    bug fix: aglc_write_curve() barfed due to NaNs in sum(c.exposure) for
%             FITS header keyword.  Now omit NaN in the sum.
%             
% 2010.08.10 - v-1.5.0:
%   fixed bug in aglc_phased() which caused a discontinuity in phase curves. 
%
%
% 2010.08.04 - 
%
%  Work the same changes as below into _aglc_phased2(), to handle the
%  rate and exposure properly.
%
% 2010.08.04 - 
%    Adopt changes made by Manfred Hanke which fixed an error with
%    differential frame dropouts on different CCDs.  Also compute a
%    self-consistent exposure, and add his rate error column to the
%    output. 
%
%   NOTE: this change has not yet been incorporated into the phase
%   computation; changes to _aglc_phased will be similar to those in
%   _aglc2 relative to  _aglc. 
%
% 2009.07.10 
%    v 1.3.9 - fix bug in status bit check.
%
% 2008.04.12 - cc is correct now (I think - see 2007.07.05 comments)
%    v 1.3.7: add BACKSCAL == 1.0 keywords to output file.
%             apply jed patches for external GTI tables.
%              Changes are back-compatible. New optional arguments
%              allow specification  of GTI files (like an flt1) in
%              aglc_read_events(), or via a qualifier, gti="flt1" in
%              aglc(). 
%
% 2007.07.03 - still bug in cc-mode; curve right shape, but time scaled wrong.
%              need to review both time and exposure. [fixed in 1.3.6]
%
% 2007.06.15 - bug in phase binning - used frame_exp, not timedel.
%              0.04104s /frame in time before conversion to phase.
%
%
% 2007.06.12  found bug in time grid - omitted the 0.04104 s/frame;
%             Found by getting wrong period in EX Hya.
%
%
% 2007.03.27  changed copyright to gpl version
%
% 2007.02.16  fix output counts data type; histogram() creates
%             UInteger_Type, but cfitsio requires Integer_Type.
%
% 2007.01.31  fix grid problem; time_min[1] was > time_max[0].
%
%
% 2006.10.14  revert to isis histogram to remove external dependency
%             on histogram module.  Should suffice to change "hist1d"
%             to "histogram" and in instances which use the reverse
%             indices, to change "&r" to ", &r".

% 2005.09.23  1.2.4 fixed aglc_write_curve - had 2 extname fields.
%
% 2005.06.10  1.2.3 fixed spelling error in usage message
%        .13        changed "static" to "private"
%                   appended EXPERIMENTAL utility, garf_gaps()
%
% 2005.05.23  1.2.2 modify for slang 2 array indexing
%
% 2005.04.05  1.2.1 modify conditional loads for isis vs ciao
%             Ciao 3.2 does not need namespace use and now supports
%             "require/provide".
%
% 2005.03.20    1.2.0  read stat1 file for exposure.  Stat1 has
%                   one record for each unique expno.  Much smaller
%                   file, bins very fast.
%
% 2004.07.02   Add output function.
%              Add mjdref to phase curve structure.

% 2004.07.01  Deem it 1.0.0 and ready for release.
%              Ran tests; updated html docs. Ran demos.

% 2004.06.30 v0.1r7 - modify ephemeris interface
%         - not stored/hidden, but explictly passed.
%         -  Added ephemeris to phase curve structure.

% 2004.05.04 v0.16 - allow cfitsio filters.
% Remove any cfitsio filters (in "[]" in filename).
% fixed small bug in aglc_filter, w/ scalar index "list"
% fixed small error in phase grid; truncate if max(phi) > 1.0

% 2004.05.03  v0.15 - make sherpa compatible.

% message(" version 0.14 has bug in exposure"); % traced to hist1d().---fixed there.

%% 2004.04.15 v0.14 - use histogram module; modify histogram calls.
%%                  ** need to test overflow bin for short phase ranges.
%%                  ** see notes on v0.11

%% 2004.04 v 0.13 add explicit time read; need for gti filtering.

%    2003.09.13 - punt dmextract and do it right.
%                 Use expno instead of times.
%                 Add phase option, w/ input of period and epoch.
%
%   2003.10.01  talked to gea re cc:
%        use READMODE to see if CONTINUOUS vs TIMED
%	or if TIMEDEL<0.2, frametime=510*TIMEDEL.
%        (TIMEDEL could be as small as 0.24104 in TE, if sub-array)
%
% version 0.12 - 2004.02.22 -
%                           
%      Unify naming scheme: agilc -> aglc
%                      change agpc_ to aglc_
%     agpc =>  Acis Grating Phase Curve
%     Make names phase-curve specific, but have aglc prefix for
%     uniformity w/in package.
%
%
% version 0.11 - 2003.12.24 - simplify phase gridding?
%                             Store phases in event structure? (not yet).
%                   To avoid confusion from the overflow bin, 
%                   make all grids explicit lo-hi.
%                   Had a problem w/ phase ranges like: [0.1:0.2:0.01],
%                     since the sum did not give the proper exposure unless
%                     ignoring the last bin. 
%                   All occurrences change w/ phase_grid.
%
% version 0.10 - 2003.12.12- - add reverse histogram option.
%                 define reverse-index variable in light curve struct
%                 return revidx from counts = histogram() statements (2)
%                   Careful to re-index into lev selection
%                   Also need to concatenate revidx over ccd loop.
%
%    NOTE: 2003.12.14: Also need revidx for exposure!!!
%          Needed to re-compute proper exposure after filtering on
%          selected light curve or phase curve bins.
%
%%% Supply reverse indices, based on selections of specific
%%% light curve or phase curve bins.  E.g., given a light curve, we may
%%% wish to select a high-state via where(count_rate>val), then bin a
%%% spectrum from the events those bins reference.  This is trivially done 
%%% externally for few simple time regions, but is more difficult if there
%%% are many.  It is most effective when considering phase, for which a bin
%%% could come from many event times.
%%%
%%% Such a selection effectively defines a new set of time intervals (GTI)
%%% which could be written to a FITS file filter.
%%%
%%% Alternatively, when the bins are selected in the light (or phase) curve,
%%% the exposure is also specified by the sum of the exposure over the bins.
%%%
%%% The reverse indices will be added to the light/phase curve structure.
%%% To apply (externally), the events will have to be read explicitly with
%%% aglc_read_events.  The indices will then point into .the evt1a data.
%
%%% Also add a helper function to reform an Array_Type of arrays into a single
%%%  list --- Integer_Type[] = cat_arrays( Array_Type[] );
%
% version 0.9 - 2003.11.21 - added extraction/bkg region support functions.
%                          - added 3 explicit regions:
%                                lower background
%                                source
%                                upper background
% version 0.8 - cc mode fixes
%
% version 0.7 - 2003.09.19
%               add time filtering; needed for phase support.
%
% version 0.6 - 2003.09.18
%               phase support revisions
%
% version 0.5 - 2003.09.17
%               begin adding phase support:
%
%                 t = expno_to_time(s);
%                    use timedel, timezero.
%                    Ignore differences between ccd's.
%
%                 p = time_to_phase(t, hjd_epoch, period);
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% ---------- NOTES ON TIME CONVERSION -------------------------------
%
%  cxo time to jd: (tricky,tricky - cxo uses MJD!!! )...
%
%   T_conj_HJD = 2452506.4328;   %(for example)
%   P_day =  0.27831149;         %(for example)
%   sec_day =  86400.0;
%   dmjd = 2400000.5;
%
%   mjdref = fits_read_key("evt1a","MJDREF"); % get from header.
%
%   % t is cxo evt time:
%   nrot = ((t / sec_day + mjdref + dmjd) - T_conj_HJD ) / P_day;
%   phi =  nrot mod 1.0;
% ------------------------------------------------------------
%
%         aglc_set_ephem( hjd_epoch, period );   % deprecate v0.17
%         (hjd_epoch, period) = aglc_get_ephem();
%         Set/get private variables          
%
%         clone aglc, _aglc into
%                  agpc(...);
%                 _agpc(...);
%         check that hjd_epoch and period variables are set.
%         still use tbin in time, but produce phase curve
%         mods in _aglc to convert expno_ to phase before histograms.
%  ----------------------------------------------------------------
%
% version 0.5 - 2003.09.17
%               Change argument parsing to get rid of _aglc as public
%               function.  use __pop_args to test first arg; if struct, 
%               then skip file-reading and call _aglc.

% version: 0.4 - 2003.09.16
%                Fix time.  The first frame is not at TSTART, but 
%                could be some time later.  Noticed 1200s difference from
%                dmextract output.  will have to read time column.
%
% version: 0.3 - rearrange functions so that the light curve can be computed
%                independently of data-read.
% version: 0.2
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                   END OF REVISION HISTORY
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

%

%% *** revise *** ciao 3.2 does not require this wrapper ...
%% ***        *** ciao 3.2 has require
%%

%%% You must have the histogram module installed.
%%% It is comprised of histogram.sl, and histogram-module.so, which 
%%% can be obtained from
%%% http://space.mit.edu/CXC/software/slang/modules/
%%% (it is not (yet) bundled with ciao).
%%% If your system does not have paths configured to automatically
%%% find it, you will need to  set appropriate environment variables, 
%%% or set paths in the slang interpreter via something like:
%%% 
%%%  set_import_module_path( "blah/lib/slang/modules:" + get_import_module_path ) ; 
%%%  set_slang_load_path(    "blah/slsh/local_packages:" + get_slang_load_path );
%%% 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% global data

%%% re-form source and background extraction region info into a structure.
%%%  NOTE: this still only supports rectangles, and not the "bow-tie"
%%%
private variable ext_regions =
struct { s_min,
s_max,
backscale,
src_backscale,
bkg_backscale,
b_lo_min,
b_lo_max,
b_hi_min,
b_hi_max
} ;
%%%%%

% Default grating source region, cross-dispersion:
%  Redefine limits so bkg region lower and upper can be discontinuous
%  from source region:


private variable DEF_s_min = -6.6e-4,
DEF_s_max =  6.6e-4,
DEF_b_lo_min = -6.0e-03,
DEF_b_lo_max = DEF_s_min, 
DEF_b_hi_min =  DEF_s_max,
DEF_b_hi_max =  6.0e-03;

% Phase out these variables.................use the struct instead ***
%
% static variable d_min = DEF_d_min,                 % ***
%                d_max = DEF_d_max ;                 % ***

% static variable max_bkg_tgd = DEF_b_up_max,        % ***
%                 min_bkg_tgd = DEF_b_low_min ;      % ***

%
%    BACKSCAL is the ratio of bkg region width to src region width, and the
%         divisor for the counts to scale to src region size.
%

% static variable backscale = 1.0;    % default for source light curve % ***
% static variable src_backscale = 1.0; % ***
% static variable bkg_backscale = (max_bkg_tgd - min_bkg_tgd) / (d_max - d_min) - 1.0 ; % ***

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% STATUS_FILTER=1 will cause the events to be filtered on status==0.
%  If this is not set, and status != 0 is detected, a warning is issued.
%
private variable FILTER_STATUS = 0 ; 
define aglc_set_status_filter( flag )  %{{{
%!%+
%\function{aglc_set_status_filter}
%\synopsis{Set a flag to specify if the grating events should be filtered on STATUS=0.}
%\usage{aglc_set_status_filter( flag );}
%\description
%    By default, the status filter is zero, meaning all good
%    events.  Different status bit fields qualify attributes of "bad"
%    pixels.  If, for some reason, the grating events have not been
%    filtered on status, but should be, then setting this flag causes
%    all events with non-zero status to be ignored.
%
%    If flag=1, then ignore events with non-zero status.
%    If flag=0, then accept all events.
%\seealso{aglc, aglc_phased}
%!%-
{
    FILTER_STATUS = flag and 1 ;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
private define calc_backscales()  %{{{
{
    %    print(ext_regions); %%%% debug ***
    ext_regions.src_backscale = 1.0 ;
    ext_regions.bkg_backscale =
    (
    ( ext_regions.b_hi_max - ext_regions.b_hi_min )
    + ( ext_regions.b_lo_max - ext_regions.b_lo_min)
    )
    / (ext_regions.s_max - ext_regions.s_min)  ;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
define aglc_set_regs( s_min, s_max, b_lo_min, b_lo_max, b_hi_min, b_hi_max )  %{{{
%!%+
%\function{aglc_set_regs}
%\synopsis{Set the spatial binning regions (cross-dispersion) for the source
%    and backgrounds on both sides, and compute the background scaling factors.}
%\usage{aglc_set_regs( s_min, s_max, b_lo_min, b_lo_max, b_hi_min, b_hi_max );}
%\description
%   Limits are given in cross-dispersion coordinates, tg_d, in units of degrees.\n
%   \code{s_min} = source minumum tg_d.\n
%   The center of the source region is defined as tg_d = 0.\n
%   \code{s_max} = source region maximum tg_d.\n
%   \code{b_lo_min} = background minimum, on the "low" side (tg_d < 0).\n
%   \code{b_lo_max} = background maxmimum, on the "low" side (tg_d < 0).\n
%   \code{b_hi_min} = background minimum, on the "high" side (tg_d < 0).\n
%   \code{b_hi_max} = background maxmimum, on the "high" side (tg_d < 0).
%
%   The limits follow the constraint that
%      \code{b_lo_min < b_lo_max <= s_min < s_max <= b_hi_min < b_hi_max}.
%
%   Default values are the same as the default spectral extraction
%   regions of tgextract:\n
%      \code{s_min    = -6.6e-04}  [ deg ]\n
%      \code{s_max    =  6.6e-04}  [ deg ]\n
%      \code{b_lo_min = -6.0e-03}  [ deg ]\n
%      \code{b_lo_max =  s_min}    [ deg ]\n
%      \code{b_hi_min =  s_max}    [ deg ]\n
%      \code{b_hi_max =  6.0e-03}  [ deg ]\n
%\seealso{aglc_reset_regs, aglc_get_regs}
%!%-
{
    if (     ( s_min >= s_max )
    or ( b_lo_min >= b_lo_max )
    or ( b_lo_max > s_min )
    or ( b_hi_min < s_max )
    or ( b_hi_min >= b_hi_max )
    )
    {
	error("Bkg/Src limits inconsistent. Order must be: b_lo_min < b_lo_max <= s_min < s_max <= b_hi_min < b_hi_max");
	return;
    }

    ext_regions.s_min = s_min ;
    ext_regions.s_max = s_max ;

    %    ext_regions.min_bkg_tgd = b_min ;     % ***
    %    ext_regions.max_bkg_tgd = b_max ;     % ***

    ext_regions.b_lo_min = b_lo_min;
    ext_regions.b_lo_max = b_lo_max;
    ext_regions.b_hi_min = b_hi_min;
    ext_regions.b_hi_max = b_hi_max;

    calc_backscales();
    
    %    d_min = s_min ;             % ***
    %    d_max = s_max ;             % ***

    %    min_bkg_tgd = b_min ;       % ***
    %    max_bkg_tgd = b_max ;       % ***
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


aglc_set_regs( DEF_s_min, DEF_s_max, DEF_b_lo_min, DEF_b_lo_max, DEF_b_hi_min, DEF_b_hi_max );

% ext_regions.d_min = d_min;                    % ***
% ext_regions.d_max = d_max;                    % ***
% ext_regions.max_bkg_tgd = max_bkg_tgd ;       % ***
% ext_regions.min_bkg_tgd = min_bkg_tgd ;       % ***

% calc_backscales() ;

%%%%

private variable tg_part_map = Assoc_Type[ Integer_Type ];

tg_part_map["h"] = 1;     % heg
tg_part_map["m"] = 2;     % meg
tg_part_map["l"] = 3;     % leg

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
define aglc_reset_regs()  %{{{
%!%+
%\functon{aglc_reset_regs}
%\synopsis{Reset the source and background binning regions (in cross-dispersion
%    direction) to their defaults.  Calculate and store background scale factors.}
%\usage{aglc_reset_regs;}
%\description
%    Reset to defaults.
%\example
%    print( aglc_get_regs );   % Are they as desired? if not...
%    aglc_reset_regs;
%\seealso{aglc_set_regs, aglc_get_regs}
%!%-
{
    aglc_set_regs( DEF_s_min, DEF_s_max, DEF_b_lo_min, DEF_b_lo_max, DEF_b_hi_min, DEF_b_hi_max );
    %    calc_backscales() ;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
define aglc_get_regs()  %{{{
%!%+
%\function{aglc_get_regs}
%\synopsis{Retrieve stored source and background regions
%    (cross-dispersion limits), and associated BACKSCAL values.}
%\usage{r = aglc_get_regs;}
%\description
%    Retrieves a structure containing the stored values used for binning
%    light curves for source or background.
%\seealso{aglc_set_regs, aglc_reset_regs}
%!%-
{
    return ext_regions;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%%% this function from j.davis %%%, w/ slight modification.
%%% returns array of same length as t, with flags where t is in GTI.
%%% Leaves where() to the application (mainly so one can " where( a or b or c)" over many ccdid).
%%
private define gti_flag (g, t)  %{{{
{
    % (from j.davis gtifuns.sl) %
    variable num = length (t);
    variable ok = Int_Type[num];
    variable start = g.start;
    variable stop = g.stop;
    variable n = 0;

    EXIT_BLOCK
    {
	return ok;
    }
    
    _for (0, length(start)-1, 1)
    {
	variable i = ();
	variable s = start[i];
	
	while (t[n] < s)
	{
	    n++;
	    if (n == num)
	    return;
	}
	s = stop[i];

	variable n0 = n;
	while (t[n] < s)
	{
	    n++;
	    if (n == num)
	    break;
	}
	if (n == n0)
	continue;
	
	ok[[n0:n-1]] = 1;
    }
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% apply the evt1a GTI filters to the exp_ref events ccd and expno columns,
%  Since structs are passed by references, changing structure fields will 
%   change in the input parameter.
%
private define aglc_filter_gti( s_evt )  %{{{
{
    %    variable c = where( hist1d( s_evt.ccd, [0:9] ) ); % count ccds, get ids.
    % 1.3.0:
    variable c = where( histogram( s_evt.ccd, [0:9] ) ); % count ccds, get ids.
    variable n = length( c );

    variable i ;
    variable flag = Integer_Type[ length( s_evt.expno ) ] ;

    for ( i=0; i<n; i++ )
    {
	()=printf(" %d", c[i] ); () = fflush(stdout); %%% DEBUG
%%%v-1.3.7	variable g = fits_read_table( sprintf( "%s[GTI%d]", s_evt.fevt_1a, c[i] ) ) ;
%%%v+1.3.7+
	variable g = fits_read_table( sprintf( "%s[GTI%d]", s_evt.fgti, c[i] ) ) ;
%%%v-1.3.7-
	variable lccd = where( s_evt.ccd == c[i] ) ;
	flag[ lccd ]  = flag[ lccd ] or gti_flag( g, s_evt.time[ lccd ] );
    }
    flag = where( flag );

    s_evt.expno = s_evt.expno[ flag ];  
    s_evt.ccd = s_evt.ccd[ flag ];
    s_evt.time = s_evt.time[ flag ];
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
define aglc_read_events ()  %{{{
%!%+
%\function{aglc_read_events}
%\synopsis{Read subset of events from grating file and from exposure file into a structure.}
%\usage{Struct_Type s = aglc_read_events(String_Type tgevt, expno_ref);}
%\description
%    \code{tgevt} should be a Level 1.5 (or 2) grating events file name.
%    \code{expno_ref} should be a Level 1 ``stat'' or events (unfiltered) file name.
%    The returned value is a structure.  The detailed contents of the
%    structure are given below in the example.  While the exposure can
%    be computed from the EXPNO column of the unfiltered event file,
%    use of the `stat1' file is more efficient.
%
%    Definitions for Event Structure Fields:\n
%    - \code{fevt_1a}: File name for grating events.\n
%    - \code{fevt_exp}: File name for exposure reference (unfiltered events or ``stat1'' file)\n
%    - \code{expno[]}: Exposure number column from exposure reference file.\n
%    - \code{ccd}: CCD_ID column from exposure reference file.\n
%    - \code{expno_1a }: Exposure number column from grating event file.\n
%    - \code{ccd_1a}: CCD_ID column from grating event file.\n
%    - \code{tgpart}: TG_PART column from grating event file. (value is 1 for HEG, 2 for MEG, 0 for zero order, 99 for background)\n
%    - \code{order}: TG_M column from grating events file, giving the diffraction order.\n
%    - \code{tgd}: TG_D column from grating events file. This is the cross-dispersion coordinate in degrees.\n
%    - \code{wave}: TG_LAM column from grating events file, giving the wavelength of each event.\n
%    - \code{timedel}: TIMEDEL keyword from grating event file, needed for scaling exposure numbers to time. (See the ahelp on time.)\n
%    - \code{timepixr}: TIMEPIXR keyword from event file. (See the ahelp on time.)\n
%    - \code{frame_exp}: the exposure time per frame (EXPNO). (See the ahelp on time.)\n
%    - \code{mjdref}: Modified Julian Day reference time for Chandra observations, from the event file header.\n
%    - \code{tstart}: TSTART, from the event file header, is the observaton start time in seconds since MJDREF.\n
%    - \code{status}: Event status column, from the grating events file. (See the definition of ACIS status bits.)\n
%    - \code{time}: The TIME column from the exposure reference event file.\n
%    - \code{cycle}: unused.\n
%\seealso{aglc_filter, aglc, aglc_phased}
%!%-
{
    % file_1a is "Level-1.5" or "-1a", with grating coordinates.
    %
    % file_1  is "Level-1" - could be same as file_1a, but it MUST be
    %         unfiltered (have bad status).
    %
    %   Often, the 1a or Level 2 files have bad status removed.
    %   If the 1a file has not been filtered, it can be used for both args.
    %   The "bad status" file is needed merely to occupy all frames exposed.
    %

%%%v-1.3.7    variable file_1a, file_1;
%%%v+1.3.7+
    variable file_1a, file_1, gtifile;
%%%v+1.3.7-

%%%v-1.3.7    if (_NARGS != 2)
%%%v-1.3.7    {
%%%v-1.3.7	message("USAGE:  s = aglc_read_events( tgevt, expno_ref )");
%%%v+1.3.7+	
    switch (_NARGS)
    {
	case 2:
	(file_1a, file_1) = ();
	gtifile = NULL;
    }
    {
	case 3:
	(file_1a, file_1, gtifile) = ();
    }
    {
	message("USAGE:  s = aglc_read_events( tgevt, expno_ref[, gtifile])");
%%%v+1.3.7-

	message("     tgevt should be a Level 1.5 (or 2) grating events file name.");
	message("     expno_ref should be a stat1 or Level 1 events (unfiltered) file name;");
	message("        (stat1 is more efficient)." ) ; 
	message("     The returned value is a structure.");
	message("See also: aglc(), aglc_phased().");
	return -1;
    }

%%%v-1.3.7    file_1 = ();
%%%v-1.3.7    file_1a = ();

    variable s = struct
    {
	fevt_1a,
	filt_1a,
%%%v+1.3.7+
        fgti,
%%%v+1.3.7-
	fevt_exp,
	filt_exp,
	expno,
	ccd,
	expno_1a,
	ccd_1a,
	tgpart,
	order,
	tgd,
	wave,
	timedel,
	timepixr,
	frame_exp,
	mjdref,
	tstart,
	status,
	time,
	cycle,
	READ_GRP_PER_EXPNO
    }
    ;

    %    variable expno, ccdid, tgpart, order, tgd, wave, estat, etime ;
    
    vmessage("%% Reading files %s, %s ...", file_1a, file_1 );

    (s.expno_1a, s.ccd_1a, s.tgpart, s.order, s.tgd, s.wave, s.status) =
    fits_read_col (file_1a,
    "expno",
    "ccd_id",
    "tg_part",
    "tg_m",
    "tg_d",
    "tg_lam",
    "status");

    vmessage("%% Read %d events.", length( s.expno_1a ) );

    variable tmp_names;

    tmp_names = strchop(file_1a, '[', 0 ) ;

    s.fevt_1a =  tmp_names[0] ; % truncate any cfitsio filter.
    if ( length(tmp_names) > 1 )   s.filt_1a =  "[" + tmp_names[1] ; % store any cfitsio filter.

%%%v+1.3.7+
    if (gtifile == NULL)
      gtifile = s.fevt_1a;
   
    s.fgti = gtifile;
%%%v+1.3.7-

    tmp_names = strchop(file_1, '[', 0 ) ;
    s.fevt_exp = tmp_names[0]  ;  % truncate any cfitsio filter.
    if ( length(tmp_names) > 1 ) s.filt_exp = "[" + tmp_names[1]  ;  % store any cfitsio filter.

    s.timedel  = fits_read_key (file_1a, "TIMEDEL");
    s.timepixr = fits_read_key (file_1a, "TIMEPIXR");

    %% v0.8 cc-mode fix:
    if ( fits_read_key (file_1a, "READMODE") == "TIMED")
    {
	s.READ_GRP_PER_EXPNO = 1 ; 
	s.frame_exp = s.timedel - 0.04104;
    }
    else
    {
	s.READ_GRP_PER_EXPNO = 512;
	s.frame_exp = s.timedel  * s.READ_GRP_PER_EXPNO ; 
    }
    
    s.mjdref = fits_read_key (file_1a, "MJDREF");

    %%% *** could define/set for 1a list:
    %%%    s.time = s.expno * s.frame_exp + s.frame_exp*s.timepixr  %%% ***
    %
    %% v 0.13 add explicit time read; need for gti filtering.

    if (file_1a == file_1)   % all in one file - don't read twice
    {
	s.expno = s.expno_1a;
	s.ccd = s.ccd_1a;
    }
    else    % have separate unfiltered file
    {
	(s.expno, s.ccd) = fits_read_col(file_1, "expno", "ccd_id");
	vmessage("%% Read %d exposure reference records.", length(s.expno) );
	%	s.expno = expno;
	%	s.ccd = ccdid;
	%	s.time = etime;
    }
    s.time = fits_read_col(file_1, "time");

    % Get a time-stamp from the minimum time present in the file and correct it for its exposure number
    % -- or even compute the mean for all times (though this will take way longer).
    % while there could be different min times for each ccd (and will be),
    % it can only differ by a few seconds.  So for >100 s bin light curves,
    % it will be moot. (CC grating timing will be a seperate story).

    % read times from the big file, keeping only the minimum:
    % (only read the first row; the file is time-sorted)

    %%% need to read either an EVENTS or EXPSTATS extension. 

  % s.tstart =      s.time[0] - s.expno[0] * s.timedel * s.READ_GRP_PER_EXPNO - s.frame_exp * s.timepixr ;  % only consider first time
    s.tstart = mean(s.time    - s.expno    * s.timedel * s.READ_GRP_PER_EXPNO - s.frame_exp * s.timepixr);

    % DEBUG:
    () = printf("%% Applying gti filters..."); () = fflush(stdout);
    % DEBUG:

    aglc_filter_gti( s ); % use evt1a GTI table to filter expref data.

    message("");

    return s;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
private define cons_event_flags( s, do_bkg, orders, tg, wmin, wmax )  %{{{
{
    %
    % construct event filter:

    variable i;
    variable n_evt = length( s.expno_1a ) ; 
    variable tmp_flags = Integer_Type[ n_evt ] ;
    variable evt_flags = Integer_Type[ n_evt ] ;    

    variable r = ext_regions;

    if (do_bkg)   % filter on events outside source tg_d region:
    {
        evt_flags = 
	( (s.tgd >= r.b_lo_min ) and (s.tgd < r.b_lo_max ) )
	or ( (s.tgd >= r.b_hi_min ) and (s.tgd < r.b_hi_max ) ) ;

    }
    else  % source events: filter on source tg_d region
    {
	evt_flags = (s.tgd > r.s_min) and (s.tgd <= r.s_max) ; 
    }

    tmp_flags *= 0 ; 
    for (i=0; i<length(orders); i++)
    {
	tmp_flags = tmp_flags or ( s.order == orders[i] ) ;
    }
    evt_flags = evt_flags and tmp_flags ; 

    tmp_flags *= 0 ; 
    for (i=0; i<length(tg); i++)
    {
	tmp_flags = tmp_flags or (s.tgpart == tg_part_map[tg[i]]) ; 
    }
    evt_flags = evt_flags and tmp_flags ; 

    tmp_flags *= 0 ; 
    for (i=0; i<length(wmin); i++)
    {
	tmp_flags = tmp_flags or ((s.wave>wmin[i] and s.wave<=wmax[i] ));
    }
    evt_flags = evt_flags and tmp_flags ;

    if (FILTER_STATUS == 1)
    {
	tmp_flags = (s.status == 0);
	evt_flags = evt_flags and tmp_flags ;
    }
    else
    {
	% ignore afterglow, but check for other bits...

% v1.3.9 - looks like a bug - too many bytes in the next line:
%	if ( max( (s.status & 0xFFFF0FFFF )>0) ) 
%
% According to http://space.mit.edu/ASC/docs/memo_event_status_bits_2.3.pdf
% only bit 16 is used to flag afterglow, while bits 17-19 are unused.
% Want to ignore  bits 16-19 altogether (counting from 0 at the LSb)
%
	if ( max( (s.status & 0xFFF0FFFF )>0) ) 
	{
	    vmessage("%% WARNING: non-zero event status detected.");
	    vmessage("%%   If this isn't desired, pre-filter events");
	    vmessage("%%   or, set via:  aglc_set_status_filter(1);");
	}
    }

    return evt_flags ;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
private define aglc_usage()  %{{{
{
    message("%% USAGE: lc = aglc(tgevt, expno_ref, tbin, wmin, wmax, tg, orders[, bkg]);");
    message("%% USAGE: lc = aglc(evt_struct, tbin, wmin, wmax, tg, orders[, bkg]);");
    message("%% USAGE:  2nd form: use evt_struct = aglc_read_events(tgevt, expno_ref)" );
    message("%%  tgevt grating event file (Level 1.5)");
    message("%%  expno_ref: Exposure reference, either stat1 file (most efficient) or unfiltered events.");
    message("%%  tbin = time bin length, in seconds");
    message("%%  wmin, wmax: arrays of wavelength region limits.");
    message("%%  tg: grating types array; [\"heg\",\"meg\"]; \"leg\"; \"meg\"");
    message("%%  orders: array of orders to bin");
    message("%%  bkg: if 1, then bin background region");
}
%}}}

private define aglc_phased_usage()  %{{{
{
    message("%% USAGE: pc = aglc_phased(tgevt, expno_ref, phase_info, ephem, wmin, wmax, tg, orders[, bkg]);");
    message("%% USAGE: pc = aglc_phased(evt_struct, phase_info, ephem, wmin, wmax, tg, orders[, bkg]);");
    message("%% USAGE:  2nd form: use evt_struct = aglc_read_events(tgevt, expno_ref)" );
    message("%%  tgevt: grating event file.");
    message("%%  expno_ref: stat1 or unfiltered event file.");
    message("%%  phase_info = [phase_min, phase_max, phase_bin]");
    message("%%  ephem = [hjd_phase_0, period_days]");
    message("%%  wmin, wmax: arrays of wavelength region limits.");
    message("%%  tg: grating types array; [\"heg\",\"meg\"]; \"leg\"; \"meg\"");
    message("%%  orders: array of orders to bin");
    message("%%  bkg: if 1, then bin background region");
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
private define aglc_par_check(tbin, wmin, wmax, tg, orders)  %{{{
{
    variable i;

    if ( length( where(orders==0 or orders==99) ) > 0 )
    {
        message("orders: must not include 0 or 99");
	return 0;
    }

    for (i=0; i<length(tg); i++) tg = strlow(substr(tg[i],1,1));
    if ( length(where(tg=="h" or tg=="m" or tg=="l")) == 0 )
    {
        message("tg: invalid grating part; must be \"h\", \"m\", or \"l\"");
	return 0;
    }

    % minimal check on wavelength regions
    if (
    ( min(wmin) <0 )  or
    ( min(wmax) <0 )  or
    ( min( (wmax-wmin) )<0 )  or
    ( length(wmin) != length(wmax) )
    )
    {
        message("wmin,wmax error: wavelengths<0, or wmax<wmin or different length arrays.");
	return 0;
    }

    if (tbin<0)
    {
	message("tbin: must be >0");
	return 0;
    }

    return 1;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
private define aglc_phased_par_check(phases, wmin, wmax, tg, orders)  %{{{
{
    variable i;

    if ( length( where(orders==0 or orders==99) ) > 0 )
    {
        message("orders: must not include 0 or 99");
	return 0;
    }

    for (i=0; i<length(tg); i++) tg = strlow(substr(tg[i],1,1));
    if ( length(where(tg=="h" or tg=="m" or tg=="l")) == 0 )
    {
        message("tg: invalid grating part; must be \"h\", \"m\", or \"l\"");
	return 0;
    }

    % minimal check on wavelength regions
    if (
    ( min(wmin) <0 )  or
    ( min(wmax) <0 )  or
    ( min( (wmax-wmin) )<0 )  or
    ( length(wmin) != length(wmax) )
    )
    {
        message("wmin,wmax error: wavelengths<0, or wmax<wmin or different length arrays.");
	return 0;
    }

    if (phases[0]<0.0 or
    phases[0]>1.0 or
    phases[1]<0.0 or
    phases[1]<phases[0] or
    phases[1]>1.0 or
    phases[2]<0.0 or
    phases[2]>1.0)
    {
	message("phases invalid:  phases -> phase_min, phase_max, phase_step");
	return 0;
    }

    return 1;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
private define _aglc( s, tbin, wmin, wmax, tg, orders, do_bkg )  %{{{
{
    variable i, j;

    variable evt_flags = cons_event_flags( s, do_bkg, orders, tg, wmin, wmax );

    % quick test for null selection:
    if (length(where(evt_flags)) == 0)
    {
	message("Null filter. Check selection and obs. params.");
	return 1;
    }
    
    % Determine which ccds have been selected.
    %
    %    variable ccd = where( hist1d( s.ccd_1a[ where(evt_flags)], [0:9] ) );
    % 1.3.0:
    variable ccd = where( histogram( s.ccd_1a[ where(evt_flags)], [0:9] ) );
    variable nccd = length(ccd);

    % To make sure each ccd's light curve is commensurate with the others,
    % we must define a common grid.  To avoid aliasing on the time grid,
    % the specified time bin (tbin) will be adjusted to the nearest integer
    % number of frames, of length TIMEDEL (but with effective exposure per
    % frame of TIMEDEL - 0.04104)..

    variable frames, all_frames, min_expno, max_expno, expno_grid, dexpno;

    min_expno = min( s.expno );
    max_expno = max( s.expno );
    all_frames = [ min_expno : max_expno ] ;
    dexpno = nint( tbin / s.frame_exp );
    expno_grid = [min_expno : max_expno : dexpno ] ;

    variable time_step = s.timedel * dexpno * s.READ_GRP_PER_EXPNO ;
    variable ontime_per_bin = dexpno * s.frame_exp  ;
    variable npts = length(expno_grid);

    vmessage("%% Frames per tbin = %d", dexpno );    
    vmessage("%% time_step = %0.3e", time_step);     
    vmessage("%% ontime per bin = %0.3e", ontime_per_bin );

    variable c = struct
    {
	time,
	time_min,
	time_max,
	counts,
	count_rate,
	stat_err,
	exposure,
	timezero,
	mjdref,
	revidx1a,   % reverse indices of counts histogram bins
	revidx      % reverse indices for exposure histogram bins
    }
    ;

    c.time     =   Double_Type[npts];
    c.time_min =   Double_Type[npts];
    c.time_max =   Double_Type[npts];
    c.counts   =   Integer_Type[npts];
    c.count_rate = Double_Type[npts];
    c.stat_err =   Double_Type[npts];
    c.exposure =   Double_Type[npts];
    c.timezero =   s.tstart;
    c.mjdref   =   s.mjdref;
    c.revidx1a   =   Array_Type[npts];  % one array for every counts bin
    c.revidx =       Array_Type[npts];  % one array for every exposure bin

    variable exposure, count_rate, counts ; 
    variable revidx1a, revidx ;

    % need to initialize each element of c.revidx1a to Integer_Type[0]
    %  so they can be accumulated via concatenation.    
    %
    c.revidx1a[*] = Integer_Type[0];
    c.revidx[*]   = Integer_Type[0];

    () = printf("%% Computing rates for ccd_id = ");

    for (i=0; i<nccd; i++)
    {
	() = printf("%d ", ccd[i] );
	() = fflush(stdout);

	variable lev = where( evt_flags and s.ccd_1a==ccd[i] );
	%	counts = hist1d( s.expno_1a[lev], expno_grid, &revidx1a ) ;
	% 1.3.0: use isis histogram to avoid external dependency:
	counts = histogram( s.expno_1a[lev], expno_grid,, &revidx1a ) ;
	c.counts += counts ;

	% Compute the exposure from the unfiltered (evt1) file, which has
	% bad grades, etc, and probably something in every frame:
	% To compute the exposed frames, we have to use the exposure number
	% filtered ONLY by the ccd_id:

	variable frame_count, frames_exposed, l_ccd, l_exp;
	
	l_ccd = where(s.ccd==ccd[i]) ;

	% 1.3.0:
	%	frame_count = hist1d( s.expno[ l_ccd ], all_frames ) ;
	frame_count = histogram( s.expno[ l_ccd ], all_frames ) ;

	l_exp = where( frame_count ) ;
	frames_exposed = all_frames[ l_exp ] ;

	%%DEBUG: ***  defined some globals just to save data...; kludgey.
	%LEXP=l_exp;
	%FEXP=frames_exposed;
	%FCOUNT=frame_count;
	%ALLFRAMES=all_frames;
	%LCCD=l_ccd;
	%%DEBUG: ***


	% 1.3.0:
	%	exposure = hist1d( frames_exposed, expno_grid ) * s.frame_exp;
	exposure = histogram( frames_exposed, expno_grid ) * s.frame_exp;
	c.exposure += exposure ;

	%% For exposure reverse indices, 
	%%  compute throw-away histogram of expno on desired grid.
	%%  (Most of the exposure computation is to determine *which*
	%%  frames were exposed, not how many event are in each frame.)

	% 1.3.0:
	%	() = hist1d( s.expno[ l_ccd ], expno_grid, &revidx );
	() = histogram( s.expno[ l_ccd ], expno_grid,, &revidx );

	% Accumulate the reverse indices for exposure.  Since we indexed into
	% s.expno_1a with lev, we need to un-index to get the "raw" indices
        % from the full evt1a list.  That means for each ccd, we want
	% lev[revidx[j]], where j is an index into expno_grid.

	% To concatenate over ccd index i, we can do something like
	% c.revidx1a[j] = [c.revidx1a[j], lev[revidx1a[j]]]  for all j.

	for (j=0; j<npts; j++)
	{
	    c.revidx1a[j] = [ c.revidx1a[j],  lev[ revidx1a[j] ] ] ;
	    c.revidx[j] =   [ c.revidx[j],  l_ccd[ revidx[j] ] ] ;
	}
    }

    () = printf("\n");

    c.exposure /= nccd ;         % mean exposure time
    j = where( c.exposure > 0 );
    c.count_rate[j] = c.counts[j] / c.exposure[j] ;
    c.stat_err = sqrt( c.counts );


    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % v 1.3.4 - bug fix in c.time_min:      ( dph 2007.06.12)
    % v 1.3.5 -- NOT FIXED FOR CC
    % v 1.3.6 --- FIX IT: (dph 2007.07.05); cc time is *510
    %
    %
    %    c.time_min = expno_grid * s.frame_exp + s.frame_exp*s.timepixr ;
    % c.time_min = expno_grid * s.timedel   + s.frame_exp*s.timepixr ;
    %
    % v1.3.6:
    c.time_min = expno_grid * s.timedel * s.READ_GRP_PER_EXPNO + s.frame_exp*s.timepixr ;
    %
    % Inputs:
    %    min_expno = min( s.expno );
    %    max_expno = max( s.expno );
    %    all_frames = [ min_expno : max_expno ] ;
    %    s.frame_exp = s.timedel - 0.04104;
    %    dexpno = nint( tbin / s.frame_exp );
    %    expno_grid = [min_expno : max_expno : dexpno ] ;
    %     time_step = s.timedel * dexpno ;
    %    ontime_per_bin = dexpno * s.frame_exp  ;
    %
    % EXPLANATION:
    % - dexpno is nearest integer number of frames per requested tbin
    % - tbin is input by user
    % - time_step is unused except for info
    % - s.timedel is from the evt header, and is the frame time plus
    %    0.04104. (flush time)
    %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %    c.time_max = c.time_min + time_step ;
    %    c.time = c.time_min + time_step / 2.0 ;

#iftrue
    if (npts > 1 ) c.time_max = make_hi_grid( c.time_min ) ;
    else
    {
% 1.3.8 experimental:    2009.02.23 --- seems to work.
%	message("%% aglc-1.3.8.sl    experimental .........");
	c.time_max = c.time_min + time_step ;
    }
%
#endif
    %c.time_max = make_hi_grid( c.time_min ) ;
    c.time = ( c.time_min + c.time_max ) / 2.0 ;

    vmessage("%% min time = %f sec", c.time_min[0]);
    vmessage("%% max time = %f sec", (c.time_max[[-1:-1]])[0]);

    if (do_bkg)
    {
	variable bkg_backscale = ext_regions.bkg_backscale ;      % ***
	vmessage("%% Scaling background by 1./%f", bkg_backscale );
	c.stat_err = sqrt( c.counts ) / bkg_backscale ; 
	c.counts /= bkg_backscale ;	
	c.count_rate /= bkg_backscale ;
    }

    return c ;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
private define _aglc2(s, tbin, wmin, wmax, tg, orders, do_bkg)  %{{{
{
  % While _aglc computes the mean count rate as
  %     ( sum_{ccd} counts[ccd] ) / [ ( sum_{ccd} exposure[ccd] )/N_{ccd} ].
  % I suggest to use
  %     sum_{ccd} ( counts[ccd] / exposure[ccd] ).
  %
  % Manfred Hanke, 2010-05-17

  variable evt_flags = cons_event_flags( s, do_bkg, orders, tg, wmin, wmax );

  % quick test for null selection:
  if (length(where(evt_flags)) == 0)
  {
    message("Null filter. Check selection and obs. params.");
    return 1;
  }

  % Determine which ccds have been selected.
  variable i = where(evt_flags);
  variable ccd = where([any(s.ccd_1a[i]==0), any(s.ccd_1a[i]==1), any(s.ccd_1a[i]==2), any(s.ccd_1a[i]==3), any(s.ccd_1a[i]==4),
                        any(s.ccd_1a[i]==5), any(s.ccd_1a[i]==6), any(s.ccd_1a[i]==7), any(s.ccd_1a[i]==8), any(s.ccd_1a[i]==9)]);
  variable nccd = length(ccd);

  % To make sure each ccd's light curve is commensurate with the others,
  % we must define a common grid.  To avoid aliasing on the time grid,
  % the specified time bin (tbin) will be adjusted to the nearest integer
  % number of frames, of length TIMEDEL (but with effective exposure per
  % frame of TIMEDEL - 0.04104).
  variable min_expno = min(s.expno);
  variable max_expno = max(s.expno);
  variable all_frames = [min_expno : max_expno];
  variable dexpno = nint(tbin/(s.timedel * s.READ_GRP_PER_EXPNO));  % dexpno is nearest integer number of frames per requested tbin
  variable expno_grid = [min_expno : max_expno : dexpno];
  variable time_step = s.timedel * dexpno * s.READ_GRP_PER_EXPNO;  % time_step is unused except for info
                     % s.timedel is from the evt header, and is the frame time plus 0.04104. (flush time)
  variable npts = length(expno_grid);
  vmessage("%% Frames per tbin = %d", dexpno );
  vmessage("%% time_step = %0.3e", time_step);
  vmessage("%% ontime per bin = %0.3e", dexpno * s.frame_exp);

  variable c = struct {
      time,
      time_min = expno_grid * s.timedel * s.READ_GRP_PER_EXPNO + s.frame_exp*s.timepixr,
      time_max,
      ccd_counts = Integer_Type[10,npts],
      ccd_exposure = Double_Type[10,npts],
      counts = 0,
      count_rate = 0,
      count_rate_err = 0,
      stat_err,
      exposure = 0,
      timezero = s.tstart,
      mjdref = s.mjdref,
    };
  c.time_max = (npts>1 ? make_hi_grid(c.time_min) : c.time_min + time_step);
  c.time = 0.5*(c.time_min + c.time_max);
  vmessage("%% min time = %f sec", c.time_min[0]);
  vmessage("%% max time = %f sec", (c.time_max[[-1:-1]])[0]);

  variable skip_rev = qualifier_exists("skiprev");
  ifnot(skip_rev)
  {
    c = struct_combine(c, struct {
        revidx1a    = Array_Type[npts], % reverse indices of counts histogram bins
        revidx      = Array_Type[npts]  % reverse indices for exposure histogram bins
      });
    % need to initialize each element of c.revidx*
    % so they can be accumulated via concatenation
    c.revidx1a[*] = Integer_Type[0];
    c.revidx[*]   = Integer_Type[0];
  }

  () = printf("%% Computing rates for ccd_id = ");
  foreach ccd (ccd)
  {
    () = printf("%d ", ccd);
    () = fflush(stdout);

    variable lev = where(evt_flags and s.ccd_1a==ccd);
    variable revidx1a, counts = histogram(s.expno_1a[lev], expno_grid,, &revidx1a);
    c.ccd_counts[ccd, *] += counts;
    c.counts += counts;  % will not be used

    % Compute the exposure from the unfiltered (evt1) file, which has
    % bad grades, etc, and probably something in every frame:
    % To compute the exposed frames, we have to use the exposure number
    % filtered ONLY by the ccd_id:
    variable l_ccd = where(s.ccd==ccd);
    variable frames_exposed = all_frames[ where( histogram(s.expno[ l_ccd ], all_frames) ) ];
    variable exposure = histogram(frames_exposed, expno_grid) * s.frame_exp;

    c.ccd_exposure[ccd, *] += exposure;
    c.count_rate           += counts/exposure;  % contribution to the total count rate
    c.count_rate_err       += counts/exposure^2;  % contribution to the square of (the total count rate error = sqrt(counts)/exposure)

    ifnot(skip_rev)
    {
      % For exposure reverse indices,
      % compute throw-away histogram of expno on desired grid.
      % (Most of the exposure computation is to determine *which*
      % frames were exposed, not how many event are in each frame.)
      variable revidx;
      () = histogram(s.expno[ l_ccd ], expno_grid,, &revidx);

      % Accumulate the reverse indices for exposure.  Since we indexed into
      % s.expno_1a with lev, we need to un-index to get the "raw" indices
      % from the full evt1a list.  That means for each ccd, we want
      % lev[revidx[j]], where j is an index into expno_grid.
      variable j;
      _for j (0, npts-1, 1)
      {
	c.revidx1a[j] = [ c.revidx1a[j], lev[ revidx1a[j] ] ];
	c.revidx[j] =   [ c.revidx[j], l_ccd[ revidx[j] ] ];
      }
    }
  }
  () = printf("\n");
  c.stat_err       = sqrt(c.counts);
  c.count_rate_err = sqrt(c.count_rate_err);

  % 2010.08.04 dph  v-1.4.0:
  % self-consistent exposure, defined by  counts / rate:
  c.exposure  =  c.counts / c.count_rate ; 

  if(do_bkg)
  {
    variable bkg_backscale = ext_regions.bkg_backscale;  % ***
    vmessage("%% Scaling background by 1./%f", bkg_backscale);
    c.stat_err   /= bkg_backscale;
    c.counts     /= bkg_backscale;
    c.count_rate /= bkg_backscale;
    c.count_rate_err /= bkg_backscale;
  }
  return c;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
define aglc() % ACIS Grating Light Curve  %{{{
%!%+
%\function{aglc}
%\synopsis{Bin a light curve from ACIS grating events.}
%\usage{lc = aglc(tgevt, expno_ref, tbin, wmin, wmax, tg, orders[, bkg]);}
%\altusage{lc = aglc(   evt_struct,    tbin, wmin, wmax, tg, orders[, bkg]);}
%
%\description
%    Given an event file name (\code{tgevt}) and an exposure reference file
%    (unfiltered events or ``stat1'') name (\code{expno_ref}), bin a light curve
%    for the specified time bin in seconds (\code{tbin}), wavelength ranges
%    specified by \code{wmin}, \code{wmax}, in Angstroms.  If \code{wmin} and \code{wmax} are arrays,
%    then they must be the same length and represent low-high pairs of
%    wavelength regions.  The argument, \code{tg}, is a scalar or array of
%    grating names, and should be one or more of \code{"HEG"} and \code{"MEG"}, or \code{"LEG"}
%    (case-insensitive); only the first character is necessary.  The
%    argument \code{orders} should be an array of integers specifying the
%    diffraction orders to bin (excluding zero).  If the last argument is
%    present, then events are binned from the background region instead of
%    the source region.
%
%    In the second form, an event structure as returned by
%    \code{aglc_read_events} is used instead of two file names.  This is
%    more efficient for multiple calls, since it avoids multiple file
%    reads.
%
%    The two files required in the first form are the grating coordinates
%    event file, probably filtered of bad events; i.e., typically the file
%    from which a spectrum would be binned ("Level 1.5" or "Level 2").
%    The second file, the exposure reference file, should be unfiltered
%    events or the ``stat1'' file.  It is used to count the exposed frames
%    to determine the exposure; any event - cosmic ray, bad pixel, photon
%    - will suffice to mark a frame.  Use of the ``stat1'' file is more
%    efficient, since it has only one entry for each unique frame.
%
%    Standard binning regions are applied for source and background
%    events.   They may be changed with \code{aglc_set_regs}.
%
%    The return value is a structure of the form\n
%       \code{time = Double_Type[]}        % event time, bin center, since timezero.\n
%       \code{time_min = Double_Type[]}    % lower edge of time bin, seconds since timezero.\n
%       \code{time_max = Double_Type[]}    % upper edge of time bin, seconds since timezero.\n
%       \code{counts = UInteger_Type[]}    % counts per time bin.\n
%       \code{count_rate = Double_Type[]}  % Counts per second ( counts / exposure ).\n
%       \code{stat_err = Double_Type[]}    % Statistical error ( sqrt(counts) ).\n
%       \code{exposure = Double_Type[]}    % Exposure per bin, averaged over all chips included, in seconds.\n
%       \code{timezero = Double_Type}      % reference time, in seconds since MJDREF.\n
%       \code{mjdref = Integer_Type}       % Modified Julian Day reference point of Chandra data (from event header)\n
%       \code{revidx1a = Array_Type[]}     % reverse index array for the grating events.\n
%       \code{revidx = Array_Type[]}       % reverse index array for the exposure reference events.\n
%    The times are in seconds since timezero.  Timezero is the number of
%    seconds since 1998.0, which is also the reference MJD.
%
%    The \code{count_rate} is per second, and is equivalent to counts/exposure.
%    The exposure is computed properly for event selections spanning
%    multiple CCDs with possibly different GTI tables.
%
%    The reverse indices, \code{revidx1a}, are returned by the \code{histogram}
%    function; for each bin of the light curve they point to
%    the items in the event list which have been binned.  This facilitates
%    subsequent operations on events in specific bins, especially when
%    selected on non-time criteria. (see \code{aglc_filter}).
%
%    \code{revidx} is like \code{revidx1a}, but for the exposure record event list.
%\example
%     Compute the light curve for the sum of Ne X 12A and O VIII 19A:\n
%     \code{wlo = [12.1, 18.8];}\n
%     \code{whi = [12.2, 19.1];}\n
%     \code{g = ["H", "M"];}\n
%     \code{o = [-1,1];}\n
%     \code{c = aglc( "evt2.fits", "evt1.fits", 1000, wlo, whi, g, o );}\n
%     \code{hplot(c.time_min, c.time_max, c.counts);}
%\seealso{aglc_read_events, aglc_filter, aglc_set_regs}
%!%-
{
    %   USAGE: lc = aglc(fevt1a, fevt1, tbin, wmin, wmax, tg, orders[, bkg]);");
    %   USAGE: lc = aglc(evt_struct, tbin, wmin, wmax, tg, orders[, bkg]);");

    % 1st form: 7 or 8 args.
    % 2nd form: 6 or 7 args.

    variable args = __pop_args(_NARGS) ;

    variable s, fevt1a, fevt1, tg, orders, wmin, wmax, tbin, do_bkg=0;
    variable do_read = 0;
    variable i;

    if ( (_NARGS < 6) or (_NARGS > 8) )
    {
	aglc_usage();
	return 1;
    }

    if ( typeof(args[0].value) == Struct_Type )
    do_read = 0;
    else
    do_read = 1;

    __push_args(args);

    if ( (_NARGS == 8) or (_NARGS==7 and not do_read) )
    {  
	do_bkg = ();
	if (do_bkg > 0) ext_regions.backscale = ext_regions.bkg_backscale;
    }
    
    orders = ();

    tg = ();

    % only the 1st char of "tg" matters, as "h", "l", "m"
    tg = [tg] ;   % force to array type, if not already.
    for (i=0; i<length(tg); i++)  tg[i] = strlow(substr(tg[i],1,1));

    wmax = ();
    wmin = ();
    tbin = ();

%%%v+1.3.7+
    variable gtifile;
%%%v+1.3.7-

    if (do_read)
    {
	fevt1 = ();
	fevt1a = ();
%%%v+1.3.7+
       gtifile = qualifier ("gti", fevt1a);
%%%v+1.3.7-
    }
    else
    {
	s = ();
    }

    !if (aglc_par_check(tbin, wmin, wmax, tg, orders) )
    {
	aglc_usage();
	return 1;
    }

    %  command args OK - read file:
    if (do_read)
    {
%%%v-1.3.7	s = aglc_read_events( fevt1a, fevt1 ) ;
%%%v+1.3.7+
	s = aglc_read_events( fevt1a, fevt1, gtifile ) ;
%%%v+1.3.7-
    }

    if(qualifier_exists("aglc1"))
      return _aglc(s, tbin, wmin, wmax, tg, orders, do_bkg);
    else
      return _aglc2(s, tbin, wmin, wmax, tg, orders, do_bkg;; __qualifiers);
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%....................................................................
%....................................................................
%%%               aglc_phased: Acis Gratings Phase Curve
%....................................................................
%....................................................................

%%%%  global information:

private variable  aglc_hjd_epoch = NULL,
aglc_days_period = NULL,
sec_day = 86400.0,
d_mjd = 2400000.5; % definition of MJD = JD - d_mjd

% sometimes want a time filter when doing phase curves:
% Use units of seconds since expno=0 (t=0)
%
private variable  aglc_trange = struct {tmin, tmax};
aglc_trange.tmin = 0.0; 
aglc_trange.tmax = NULL;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Elementary time conversion functions:
%
private define aglc_set_ephem()  %{{{
{
    variable hjd, pd;

    if (_NARGS != 2)
    {
	message("USAGE:  aglc_set_ephem( HJD_zero_phase, Period_in_days )");
    }
    else
    {
	aglc_days_period = ();
	aglc_hjd_epoch = ();
    }
} %}}}

define aglc_get_ephem()  %{{{
%!%+
%\function{aglc_get_ephem}
%\synopsis{Retrieve a stored ephemeris, the last one used.}
%\usage{(jd0, pd) = aglc_get_ephem;}
%\description
%    jd0 will be the Julian day of zero phase.
%    pd  will be the period in days.
%\seealso{aglc_pr_ephem, aglc_phased}
%!%-
{
    %
    % NOTE: tests broken: cannot test NULL against 0.
    %       Use switch statement instead?               *** fix ***
    %  
    if ( aglc_hjd_epoch == NULL or
    aglc_days_period == NULL or
    aglc_hjd_epoch <=0 or
    aglc_days_period <= 0 )
    error("ERROR: Invalid ephemeris. (Not set, or bad values.)");

    return  aglc_hjd_epoch, aglc_days_period;
} %}}}

define aglc_pr_ephem()  %{{{
%!%+
%\function{aglc_pr_ephem}
%\synopsis{Print the stored ephemeris used in the last phase-curve binning.}
%\usage{aglc_pr_ephem}
%\description
%    Retrieves the ephemeris stored by aglc_phased and prints to terminal.
%\seealso{aglc_get_ephem, aglc_phased}
%!%-
{
    vmessage("Epoch[JD] = %14.6f, Period[day] = %10.6f", aglc_get_ephem );
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
define aglc_set_trange()  %{{{
%!%+
%\function{aglc_set_trange}
%\synopsis{Set a simple time range filter to be applied to a light curve
%    before extracting a phase-binned curve. This is useful for
%    excluding flares, which seem to predominantly occur at the
%    beginning or end of an observation.}
%\usage{aglc_set_trange( tmin, tmax );}
%\description
%    \code{tmin, tmax} are relative times in seconds, since the start of the
%    observation.  The interval specifies the times of interest to be included
%    in phase binning. \code{tmax==NULL} means to the end of the observation.
%\example
%    \code{aglc_set_trange( 0, 2000 );   %} first 2ks will be included in phase binning.\n
%     \code{aglc_set_trange( 5000, NULL);  %} from 5ks to the end will be selected.
%\seealso{aglc_get_trange, aglc_pr_trange}
%!%-
{
    if (_NARGS != 2)
    {
	message("USAGE:  aglc_set_trange( tmin, tmax )");
	message("     tmin,tmax in relative (elapsed) time [sec].");
    }
    else
    {
	aglc_trange.tmax = ();
	aglc_trange.tmin = ();
    }
} %}}}

define aglc_get_trange()  %{{{
%!%+
%\function{aglc_get_trange}
%\synopsis{Retrieve the time filter which will be applied in phase-binning.}
%\usage{(tmin, tmax) = aglc_get_trange;}
%\description
%    tmin,tmax are relative times in seconds, since the start of the
%    observation.
%\seealso{aglc_set_trange, aglc_pr_trange;}
%!%-
{
    if (aglc_trange.tmax == NULL)
    {
	if ( aglc_trange.tmin <0)
	error("ERROR: Invalid tmin. NULL tmax (unfiltered)");
    }
    else
    {
	if ( (aglc_trange.tmin < 0) or
	(aglc_trange.tmin > aglc_trange.tmax) or
	(aglc_trange.tmax < 0) )
	error("ERROR: Invalid trange.");
    }
    return  aglc_trange.tmin, aglc_trange.tmax ;
} %}}}

define aglc_pr_trange()  %{{{
%!%+
%\function{aglc_pr_trange}
%\synopsis{Print to the terminal the current time filter.}
%\usage{aglc_pr_trange;}
%\description
%    Prints the min and max time filter used in binning phase curves.
%\seealso{aglc_set_trange, aglc_get_trange;}
%!%-
{
    vmessage("tmin = %S, tmax = %S [sec since timezero]", aglc_get_trange );
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define aglc_tobs_to_jd()  %{{{
%!%+
%\function{aglc_tobs_to_jd}
%\synopsis{Convert Chandra obervation times, given in seconds since MJDREF, to Julian day.}
%\usage{jd = aglc_tobs_to_jd( t_obs, mjdref );}
%\description
%    \code{t_obs} is a scalar or array of Chandra times in seconds since \code{mjdref}.
%    \code{mjdref} is the Modified Julian Day reference, found in Chandra
%    headers, or in the aglc event structure.
%\seealso{aglc_jd_to_rotnum, aglc_jd_to_phase, aglc_tobs_to_phase, aglc_read_events}
%!%-
{
    % Convert Chandra t_obs = seconds since mjdref to JD. (NOT MJD!!!)
    % Chandra uses Modified Julian Day reference.
    % MJDREF can be found in and event file header (or from aglc_read_events)

    variable t_obs, mjdref;

    if (_NARGS !=2 )
    {
	message("USAGE:  jd = aglc_tobs_to_jd( t_obs, mjdref );");
	return -1;
    }
    mjdref = ();
    t_obs = ();
    
    return  (t_obs / sec_day) + mjdref + d_mjd ;
} %}}}

%define agpc_expno_to_time( expno, timedel, tstart )   %% **** changed....
private define aglc_expno_to_time( expno, frametime, timepixr, tstart )  %{{{
{
    % convert (approximately) the exposure number to time in seconds
    %   since mjdref (chandra's definition of time).
    % This is approximate because only one tstart is used for all
    % CCDs.  EXPNO is simply scaled by the TIMEDEL value.
    % timezero is the time of the first frame.
    % NOTE: this is not the same as the exposure time! TIMEDEL is the 
    %       time between successive frame reads, nominally 3.24104 seconds,
    %       of which 0.04104 is for frame-shift.  But TIMEDEL could be shorter.
    % NOTE: aglc_read_events returns a structure containing
    %        timedel and timezero, as well as mjdref, timepixr

    %    return expno * timedel + tstart ;    %**
    % v0.11 - use frametime = s.frame_exp, not timedel (for cc)    
    return expno * frametime + frametime*timepixr + tstart ; 
} %}}}

%%% **** fix; isn't tstart needed in the inverse?      ***** FIX???
%define agpc_time_to_expno( t, timedel )
% v0.11 - not timedel, but frametime = s.frame_exp
private define aglc_time_to_expno( t, frametime, timepixr )  %{{{
{
    % convert (approximately) the RELATIVE time (elapsed) in sec  to
    % approx expno.

    %    return int( t / timedel ) ; 
    return int( (t-frametime*timepixr) / frametime ) ; 
} %}}}

%%% v0.11 change from timedel to frametime...
private define aglc_get_expno_filt_range( expno, frametime, timepixr )  %{{{
{
    % given the expno array and the timedel, retrieve the stored
    % time filter min,max, and compute equivalent expno limits.
    % If time is outside the expno range, use the expno limit.

    variable expno_filt_min, expno_filt_max;
    variable tmin, tmax ;
    (tmin, tmax) = aglc_get_trange;

    if (tmax == NULL)
    {
	expno_filt_max = max(expno);
    }
    else
    {
	expno_filt_max = min( [ [aglc_time_to_expno( tmax, frametime, timepixr )],
	[max(expno)] ] );
    }

    if (tmin <= 0.0 )
    {
	expno_filt_min = min(expno);
    }
    else
    {
	expno_filt_min = max( [ [aglc_time_to_expno( tmin, frametime,timepixr )],
	[min(expno)] ] ) ;
    }

    return expno_filt_min, expno_filt_max ;
} %}}}


define aglc_jd_to_rotnum()  %{{{
%!%+
%\function{aglc_jd_to_rotnum}
%\synopsis{Convert Julian days to rotation number.}
%\usage{rotation_number = aglc_jd_to_rotnum( t_jd, hjd_epoch, period_days );}
%\description
%     \code{rotation_number} = number of full cycles plus fractional phase, given an ephemeris.\n
%     \code{t_jd}        = times, in Julian day numbers.\n
%     \code{hjd_epoch}   = epoch of zero-phase, in HJD\n
%     \code{period_days} = period, in days.
%\seealso{aglc_tobs_to_jd, aglc_jd_to_phase, aglc_tobs_to_phase, aglc_phased, aglc}
%!%-
{
    %   t_jd is julian day
    %   nrot = ((t / sec_day + mjdref + dmjd) - T_conj_HJD ) / P_day;

    variable t_jd, hjd_epoch, days_period;

    if ( _NARGS != 3 )
    {
	message("USAGE:  rotation_number = aglc_jd_to_rotnum( t_jd, hjd_epoch, period_days );");
	return -1;
    }

    days_period=();
    hjd_epoch = ();
    t_jd = ();

    return ( ( t_jd - hjd_epoch ) / days_period ) ;
} %}}}


define aglc_jd_to_phase()  %{{{
%!%-
%\function{aglc_jd_to_phase}
%\synopsis{Convert a list of times to phase, given an ephemeris.}
%\usage{phase = aglc_jd_to_phase( t_jd, hjd_epoch, days_period );}
%\description
%    phase       = rotation_number modulo 1;
%                  fractional phase, given an ephemeris.
%
%    t_jd        = times, in Julian day numbers.
%    hjd_epoch   = epoch of zero-phase, in HJD
%    period_days = period, in days.
%
%    This can be useful for plotting multple cycles without binning on
%    phase.
%
%    This is functionally equivalent to  aglc_jd_to_rotnum() mod 1.
%\seealso{aglc_jd_to_rotnum, aglc_tobs_to_phase, aglc, aglc_phased}
%!%-
{
    variable  t_jd, hjd_epoch, days_period ;

    %   % t is cxo evt time:
    %   nrot = ((t / sec_day + mjdref + dmjd) - T_conj_HJD ) / P_day;
    %   phi =  nrot mod 1.0;

    if (_NARGS != 3 )
    {
	message("USAGE:  phase = aglc_jd_to_phase(t_jd, hjd_epoch, days_period);");
	return -1;
    }

    days_period = ();
    hjd_epoch = ();
    t_jd = ();

    return aglc_jd_to_rotnum( t_jd, hjd_epoch, days_period) mod 1.0 ; 
} %}}}

%define agpc_expno_to_phase( expno, timedel, tstart, mjdref )
% v0.11 *** FIX - not timedel, but frametime...
%
private define aglc_expno_to_phase( expno, frametime, timepixr, tstart, mjdref )  %{{{
{
    % put conversion functions together;
    %  (could use the structure,
    %   s = aglc_read_events(f1a, f1);
    % which has time reference info, but leave explicit here)

    variable t = aglc_expno_to_time( expno, frametime, timepixr, tstart );
    variable jd = aglc_tobs_to_jd( t, mjdref );
    variable jd0, pd;

    (jd0, pd) = aglc_get_ephem();

    variable phi = aglc_jd_to_phase( jd, jd0, pd ) ;
    if (min(phi)<0)
    {
	variable l = where(phi<0);
	phi[l] =  1.0 + phi[l];
    }
    
    return phi;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% v0.17 - added ephem to interface
%
define aglc_tobs_to_phase()  %{{{
%!%-
%\function{aglc_tobs_to_phase}
%\synopsis{Convert Chandra observation times to phase, given an ephemeris.}
%\usage{phase = aglc_tobs_to_phase( tobs, mjdref, ephem );}
%\description
%    phase  =  phase for given times and stored ephemeris.
%    tobs   =  observed times, in seconds since mjdref
%    mjdref =  Reference MJD (from Chandra header or aglc event  structure)
%    ephem  =  Two-element array specifying ephemeris.
%              ephem[0] = Epoch JD of phase 0.00.
%              ephem[1] = Period in days.
%\seealso{aglc_jd_to_rotnum, aglc_jd_to_phase, aglc_tobs_to_jd,  aglc_read_events, aglc, aglc_phased}
%!%-
{
    variable  tobs, mjdref, ephem ;

    if (_NARGS != 3 )
    {
	message("USAGE: phase = aglc_tobs_to_phase( tobs, mjdref, ephem[] );");
	return -1;
    }
    
    ephem  = () ;   % two element array, jd0, pd_days
    mjdref = ();
    tobs = ();
    
    variable jd = aglc_tobs_to_jd( tobs, mjdref );
    variable jd0, pd;
    %    (jd0, pd) = aglc_get_ephem();  % v0.17 (I should use cvs...)
    jd0 = ephem[0];               % v0.17
    pd  = ephem[1];               % v0.17
    aglc_set_ephem( jd0, pd ) ;   % v0.17 for future ref only.

    return aglc_jd_to_phase( jd, jd0, pd ) ; 
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Elementary time conversions can now be used in phase-curve binning
%  functions.  Copy aglc() and _aglc(), and convert expno-grid binning to
%  phase-grid binning.
%
private define _aglc_phased( s, phases, wmin, wmax, tg, orders, do_bkg )  %{{{
{
    variable i, j;
    variable evt_flags = cons_event_flags( s, do_bkg, orders, tg, wmin, wmax );

    variable phase_min = phases[0],
    phase_max = phases[1],
    dphase    = phases[2],
    tbin;

    % quick test for null selection:
    if (length(where(evt_flags)) == 0)
    {
	message("Null filter. Check selection and obs. params.");
	return 1;
    }
    
    % Determine which ccds have been selected.
    %
    % 1.3.0:
    %    variable ccd = where( hist1d( s.ccd_1a[ where(evt_flags)], [0:9] ) );
    variable ccd = where( histogram( s.ccd_1a[ where(evt_flags)],   [0:9] ) ); %%%***
    variable nccd = length(ccd);

    % To make sure each ccd's light curve is commensurate with the others,
    % we must define a common grid.  To avoid aliasing on the time grid,
    % the specified time bin (tbin) will be adjusted to the nearest integer
    % number of frames, of length TIMEDEL (but with effective exposure per
    % frame of TIMEDEL - 0.04104)..

    variable frames, all_frames, min_expno, max_expno, dexpno;
    variable phase_grid;
    variable jd0, period;
    variable tfilt_min, tfilt_max, min_expno_tfilt, max_expno_tfilt ;  % if a time-filter given

    (jd0, period) = aglc_get_ephem();

    %    (min_expno_tfilt, max_expno_tfilt) =
    %           agpc_get_expno_filt_range( s.expno, s.timedel );
    % v0.11 *** change..........
    (min_expno_tfilt, max_expno_tfilt) =
        aglc_get_expno_filt_range( s.expno, s.frame_exp, s.timepixr );

    %    min_expno = min( s.expno );
    %    max_expno = max( s.expno );

    % Constrain min,max by possible time filter:
    % (These are already constrained to the range of s.expno
    %   by agpc_get_expno_filt_range().)
    %
    min_expno = min_expno_tfilt;
    max_expno = max_expno_tfilt;

    all_frames = [ min_expno : max_expno ] ;


    %%% *** fix *** don't use dexpno. but should adjust dphase to integral expno.
    tbin = dphase * period * sec_day ;

%%%
%%% 2007.07.05 *** BROKEN FOR CC-mode if dexpno < 1
%%%

    dexpno = nint( tbin / s.frame_exp );  % integral # frames per tbin.

    % trial adjustment for cc-mode, short bins:  *** TEST 1.3.6
    if ( dexpno < 1 )     dexpno = ( tbin / s.frame_exp );

    % convert integral number back to dphase:
    dphase = dexpno * s.frame_exp / period / sec_day ; 

    phase_grid = [ phase_min : phase_max : dphase ] ;

    vmessage("%% Frames per phase bin = %S",    dexpno );    
    vmessage( "%% Phase bin adjusted to %10.4f", dphase );


    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% *** Slang-2 dependency on indexing?  ***

    variable phase_grid_hi = NULL ;

    if ( ( _slang_version / 10000) == 1 )
    {
	phase_grid_hi = [ phase_grid[[1:-1]], phase_grid[[-1:-1]]+dphase];%***v0.11
    }
    else
    {
	phase_grid_hi = [ phase_grid[[1:]], ( phase_grid[[-1:-1]]) + dphase ]; % v1.2.2/slang2
    }
    %%% ***

    variable npts = length( phase_grid );

    %%%+ 2004.05.22....
    if ( phase_grid_hi[npts-1] > 1.0 )   % truncate if max(phi) > 1
    {
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	%%% *** Slang-2 dependency on indexing?  ***
	if ( (_slang_version /10000) == 1 )
	{
	    phase_grid    = phase_grid[ [0:-2] ];
	    phase_grid_hi = phase_grid_hi[ [0:-2] ];
	}
	else
	{
	    phase_grid    = phase_grid[ [:-2] ];
	    phase_grid_hi = phase_grid_hi[ [:-2] ];
	}
	%%% ***
	npts -= 1;
    }
    %%%- 2004.05.22....
    
    %%% 2004.06.30 --- add ephemeris fields to struct:

    variable c = struct   %%% ***
    {
	phase,
	phase_min,
	phase_max,
	counts,
	count_rate,
	stat_err,
	exposure,
	mjdref, 
	revidx1a,        % for evt1a
	revidx,           % for all frames
	epoch,           % JD of phase=0.000
	period           % period in days.
    }
    ;

    c.phase      = Double_Type[npts];
    c.phase_min  = Double_Type[npts];
    c.phase_max  = Double_Type[npts];
    c.counts     = Integer_Type[npts];
    c.count_rate = Double_Type[npts];
    c.stat_err   = Double_Type[npts];
    c.exposure   = Double_Type[npts];
    c.mjdref     = s.mjdref ; 
    c.revidx1a   = Array_Type[npts];  % one array for every counts bin
    c.revidx     = Array_Type[npts];  % one array for every exposure bin
    c.epoch      = jd0;               
    c.period     = period;

    variable exposure, count_rate, counts ; 
    variable revidx1a, revidx;

    % need to initialize each element of c.revidx1a to Integer_Type[0]
    %  so they can be accumulated via concatenation.    
    %
    c.revidx1a[*] = Integer_Type[0] ;
    c.revidx[*]   = Integer_Type[0] ;

    () = printf("%% Computing rates for ccd_id = ");
    for (i=0; i<nccd; i++)
    {
	() = printf("%d ", ccd[i] );
	() = fflush(stdout);

	% constrain selection by any time-filter:
	%
	variable lev = where( evt_flags
	and (s.ccd_1a==ccd[i])
	and (s.expno_1a >= min_expno)
	and (s.expno_1a <= max_expno) );

	%---------------------------------------------------------------------
	if (length(lev)==0) %%% should do something else here ***** FIX
	{
	    vmessage("WARNING: Null selection on cdd=%d. Check the trange filter.", ccd[i]);
	    continue;
	}
	%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

	% dither is necessary to anti-alias the expno grid transformation to phase grid.
	%
	variable dither_frames = urand(length(lev)) ;

	%%% v0.11 *** changed call.......................
	%%% v0.14 **** changed back - need to test for case of filtered phase grid.
	% 1.3.0:
	% 	counts = hist1d(
	% 	   aglc_expno_to_phase( s.expno_1a[lev]+dither_frames,
	% 	                        s.frame_exp,
	% 	                        s.timepixr,
	% 	                        s.tstart,
	% 	                        s.mjdref
	% 	                      ),
	% 	                      phase_grid,
	% %	                      phase_grid_hi,   % *** v0.11
	% 	                      &revidx1a ) ; 

	% v 1.3.6: (dph 2007.07.05) FIX for CC; time is timedel*510

%%% ***
	counts = histogram(
            aglc_expno_to_phase( s.expno_1a[lev]+dither_frames,
	%	s.frame_exp,  % v 1.3.5: CHANGE TO TIMEDEL????
%		s.timedel,    %%% 1.3.5 IS THIS CORRECT??? what about elsewhere???
		s.timedel * s.READ_GRP_PER_EXPNO,    %%% 1.3.6 CHECK
		s.timepixr,
		s.tstart,
		s.mjdref
		),
		phase_grid,,
		%	                      phase_grid_hi,   % *** v0.11
		&revidx1a ) ; 
	c.counts += counts ; 

	% Compute the exposure from the unfiltered (evt1) file, which has
	% bad grades, etc, and probably something in every frame:
	% To compute the exposed frames, we have to use the exposure number
	% filtered ONLY by the ccd_id:

	variable frame_count, frames_exposed, lexp, l_ccd;

	l_ccd = where(s.ccd==ccd[i]) ;
	% 1.3.0:
	%	frame_count = hist1d( s.expno[ l_ccd ], all_frames ) ;
	frame_count = histogram( s.expno[ l_ccd ], all_frames ) ;

	lexp = where( frame_count );
	frames_exposed = all_frames[ lexp ] ;

	dither_frames = urand(length(frames_exposed)) ;

	%%% v0.11 *** changed call...............................
	%%% v0.14 *** changed back - need to check in detail...
	% 1.3.0:
	% 	exposure = hist1d(
	% 	  aglc_expno_to_phase( frames_exposed+dither_frames,
	% 	                       s.frame_exp,
	% 	                       s.timepixr,
	% 	                       s.tstart,
	% 	                       s.mjdref
	% 	                     ),
	% 	                      phase_grid,
	% %	                      phase_grid_hi         % v0.11 ***
	%                              ) * s.frame_exp;
%%% ***
	exposure = histogram(
	   aglc_expno_to_phase( frames_exposed+dither_frames,
	   s.frame_exp,
	   s.timepixr,
	   s.tstart,
	   s.mjdref
	   ),
	   phase_grid,,
	   %	                      phase_grid_hi         % v0.11 ***
	   ) * s.frame_exp;
	c.exposure += exposure ;

	%% For exposure reverse indices, 
	%%  compute throw-away histogram of expno on desired grid.
	%%  (Most of the exposure computation is to determine *which*
	%%  frames were exposed, not how many event are in each frame.)

	%%% v0.11 *** changed.....................................
	%%%%v0.14 *** changed back --- need to test.
	% 1.3.0:
	% 	() = hist1d(
	% 	     aglc_expno_to_phase( s.expno[ l_ccd ],
	% 	                          s.frame_exp,
	% 	                          s.timepixr,
	% 	                          s.tstart,
	% 	                          s.mjdref
	% 	                         ),
	% 	                         phase_grid,
	% %	                         phase_grid_hi,  % *** v0.11
	%                                  &revidx );
	() = histogram(	aglc_expno_to_phase( s.expno[ l_ccd ],
	s.frame_exp,
	s.timepixr,
	s.tstart,
	s.mjdref
	),
	phase_grid,,
	%	                         phase_grid_hi,  % *** v0.11
	&revidx );

	% Accumulate the reverse indices for counts 
	% Since we indexed into counts w/ lev,
	% un-index to get the "raw" indices
	% from the full lists.  That means for each ccd, we want
	% lev[revidx[j]], where j is an index into expno_grid.

	% To concatenate over ccd index i, we can do something like
	% c.revidx1a[j] = [c.revidx1a[j], lev[revidx1a[j]]]  for all j.

	for (j=0; j<npts; j++)
	{
	    c.revidx1a[j] = [ c.revidx1a[j],    lev[ revidx1a[j] ] ];
	    c.revidx[j] =   [   c.revidx[j],  l_ccd[ revidx[j] ] ];
	}
    }  
    %----------------- for (i=0; i<nccd; i++)
    () = printf("\n");
    
    c.exposure /= nccd ;         % mean exposure time   %%% ***
    j = where( c.exposure > 0 );
    c.count_rate[j] = c.counts[j] / c.exposure[j] ;     %%% ***
    c.stat_err = sqrt( c.counts );

%%% *** rate error

    c.phase_min = phase_grid ;
    c.phase_max = c.phase_min + dphase ;
    c.phase = c.phase_min + dphase / 2.0 ;

    if (do_bkg)
    {
	variable bkg_backscale = ext_regions.bkg_backscale ; % ***
	vmessage("%% Scaling background by 1./%f", bkg_backscale );
	c.stat_err = sqrt( c.counts ) / bkg_backscale ; 
	c.counts /= bkg_backscale ;	
	c.count_rate /= bkg_backscale ;
    }

    return c ;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%% ***** 1020.08.04  dph
% make changes to fix the rate and exposure computation for multiple
% chips w/ different dropped frames.  Follow Manfred's changes to
% _aglc() in _aglc2().
%
private define _aglc_phased2( s, phases, wmin, wmax, tg, orders, do_bkg )  %{{{
{
    variable j;
    variable evt_flags = cons_event_flags( s, do_bkg, orders, tg, wmin, wmax );

    variable
      phase_min = phases[0],
      phase_max = phases[1],
      dphase    = phases[2],
      tbin;

    % quick test for null selection:
    if (length(where(evt_flags)) == 0)
    {
	message("Null filter. Check selection and obs. params.");
	return 1;
    }
    
    % Determine which ccds have been selected.
    %
    % 1.3.0:
    %    variable ccd = where( hist1d( s.ccd_1a[ where(evt_flags)], [0:9] ) );
    variable i = where( evt_flags );
    variable ccd = where([any(s.ccd_1a[i]==0), any(s.ccd_1a[i]==1), any(s.ccd_1a[i]==2), any(s.ccd_1a[i]==3), any(s.ccd_1a[i]==4),
                        any(s.ccd_1a[i]==5), any(s.ccd_1a[i]==6), any(s.ccd_1a[i]==7), any(s.ccd_1a[i]==8), any(s.ccd_1a[i]==9)]);
    variable nccd = length( ccd );

    % To make sure each ccd's light curve is commensurate with the
    % others, we must define a common grid.  To avoid aliasing on the
    % time grid, the specified time bin (tbin) will be adjusted to the
    % nearest integer number of frames, of length TIMEDEL (but with
    % effective exposure per frame of TIMEDEL - 0.04104)..

    variable frames, all_frames, min_expno, max_expno, dexpno;
    variable phase_grid;
    variable jd0, period;
    variable tfilt_min, tfilt_max, min_expno_tfilt, max_expno_tfilt ;  % if a time-filter given

    (jd0, period) = aglc_get_ephem();

    %    (min_expno_tfilt, max_expno_tfilt) =
    %           agpc_get_expno_filt_range( s.expno, s.timedel );
    % v0.11 *** change..........
    (min_expno_tfilt, max_expno_tfilt) =
        aglc_get_expno_filt_range( s.expno, s.frame_exp, s.timepixr );

    %    min_expno = min( s.expno );
    %    max_expno = max( s.expno );

    % Constrain min,max by possible time filter:
    % (These are already constrained to the range of s.expno
    %   by agpc_get_expno_filt_range().)
    %
    min_expno = min_expno_tfilt;
    max_expno = max_expno_tfilt;
    all_frames = [ min_expno : max_expno ] ;

    %%% *** fix *** don't use dexpno. but should adjust dphase to integral expno.
    tbin = dphase * period * sec_day ;

%%%
%%% 2007.07.05 *** BROKEN FOR CC-mode if dexpno < 1
%%%

    dexpno = nint( tbin / s.frame_exp );  % integral # frames per tbin.

    % trial adjustment for cc-mode, short bins:  *** TEST 1.3.6
    if ( dexpno < 1 )     dexpno = ( tbin / s.frame_exp );

    % convert integral number back to dphase:
    dphase = dexpno * s.frame_exp / period / sec_day ; 

    phase_grid = [ phase_min : phase_max : dphase ] ;

    vmessage("%% Frames per phase bin = %S",    dexpno );    
    vmessage( "%% Phase bin adjusted to %10.4f", dphase );

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% *** Slang-2 dependency on indexing?  ***

    variable phase_grid_hi = NULL ;

    if ( ( _slang_version / 10000) == 1 )
    {
	phase_grid_hi = [ phase_grid[[1:-1]], phase_grid[[-1:-1]]+dphase];%***v0.11
    }
    else
    {
	phase_grid_hi = [ phase_grid[[1:]],  phase_grid[-1] + dphase ]; % v1.2.2/slang2
    }
    %%% ***

    variable npts = length( phase_grid );

    %%%+ 2004.05.22....
    if ( phase_grid_hi[npts-1] > 1.0 )   % truncate if max(phi) > 1
    {
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	%%% *** Slang-2 dependency on indexing?  ***
	if ( (_slang_version /10000) == 1 )
	{
	    phase_grid    = phase_grid[ [0:-2] ];
	    phase_grid_hi = phase_grid_hi[ [0:-2] ];
	}
	else
	{
	    phase_grid    = phase_grid[ [:-2] ];
	    phase_grid_hi = phase_grid_hi[ [:-2] ];
	}
	%%% ***
	npts -= 1;
    }
    %%%- 2004.05.22....
    
    %%% 2004.06.30 --- add ephemeris fields to struct:

    variable c = struct   %%% ***
    {
	phase          = Double_Type[npts], 
	phase_min      = Double_Type[npts],
	phase_max      = Double_Type[npts],
	ccd_counts     = Integer_Type[10,npts],
	ccd_exposure   = Double_Type[10,npts],
	counts         = Integer_Type[ npts ],
	count_rate     = Double_Type[npts],
	stat_err       = Double_Type[npts],
	count_rate_err = Double_Type[npts],
	exposure       = Double_Type[npts],
	mjdref         = s.mjdref, 
	revidx1a       = Array_Type[npts],    % for evt1a
	revidx         = Array_Type[npts],    % for all frames
	epoch          = jd0,                 % JD of phase = 0.000
	period         = period               % period in days.
    }
    ;

    variable exposure, count_rate, counts ; 
    variable revidx1a, revidx;

    % need to initialize each element of c.revidx1a to Integer_Type[0]
    %  so they can be accumulated via concatenation.    
    %
    c.revidx1a[*] = Integer_Type[0] ;
    c.revidx[*]   = Integer_Type[0] ;

    () = printf("%% Computing rates for ccd_id = ");
    for (i=0; i<nccd; i++)
    {
	() = printf("%d ", ccd[i] );
	() = fflush(stdout);

	% constrain selection by any time-filter:
	%
	variable lev = where( evt_flags
  	           and ( s.ccd_1a==ccd[i])
	           and ( s.expno_1a >= min_expno)
	           and ( s.expno_1a <= max_expno) );

	%---------------------------------------------------------------------
	if (length(lev)==0) %%% should do something else here ***** FIX
	{
	    vmessage("WARNING: Null selection on cdd=%d. Check the trange filter.", ccd[i]);
	    continue;
	}
	%^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

	% dither is necessary to anti-alias the expno grid transformation to phase grid.
	%
	variable dither_frames = urand( length( lev ) ) ;

	%%% v0.11 *** changed call.......................
	%%% v0.14 **** changed back - need to test for case of filtered phase grid.
	% 1.3.0:
	% 	counts = hist1d(
	% 	   aglc_expno_to_phase( s.expno_1a[lev]+dither_frames,
	% 	                        s.frame_exp,
	% 	                        s.timepixr,
	% 	                        s.tstart,
	% 	                        s.mjdref
	% 	                      ),
	% 	                      phase_grid,
	% %	                      phase_grid_hi,   % *** v0.11
	% 	                      &revidx1a ) ; 

	% v 1.3.6: (dph 2007.07.05) FIX for CC; time is timedel*510

%%% ***
	counts = histogram(
            aglc_expno_to_phase( s.expno_1a[ lev ] + dither_frames,
	%	s.frame_exp,  % v 1.3.5: CHANGE TO TIMEDEL????
%		s.timedel,    %%% 1.3.5 IS THIS CORRECT??? what about elsewhere???
		s.timedel * s.READ_GRP_PER_EXPNO,    %%% 1.3.6 CHECK
		s.timepixr,
		s.tstart,
		s.mjdref
		),
%		phase_grid,,
		phase_grid, phase_grid_hi,
		%	                      phase_grid_hi,   % *** v0.11
		&revidx1a ) ; 

	c.counts += counts ;
	c.ccd_counts[ ccd[i], * ] += counts ; 

	% Compute the exposure from the unfiltered (evt1) file, which has
	% bad grades, etc, and probably something in every frame:
	% To compute the exposed frames, we have to use the exposure number
	% filtered ONLY by the ccd_id:

	variable frame_count, frames_exposed, lexp, l_ccd;

	l_ccd = where(s.ccd==ccd[i]) ;
	% 1.3.0:
	%	frame_count = hist1d( s.expno[ l_ccd ], all_frames ) ;

	frame_count = histogram( s.expno[ l_ccd ], all_frames ) ;
	lexp = where( frame_count );
	frames_exposed = all_frames[ lexp ] ;
	dither_frames = urand(length(frames_exposed)) ;

	%%% v0.11 *** changed call...............................
	%%% v0.14 *** changed back - need to check in detail...
	% 1.3.0:
	% 	exposure = hist1d(
	% 	  aglc_expno_to_phase( frames_exposed+dither_frames,
	% 	                       s.frame_exp,
	% 	                       s.timepixr,
	% 	                       s.tstart,
	% 	                       s.mjdref
	% 	                     ),
	% 	                      phase_grid,
	% %	                      phase_grid_hi         % v0.11 ***
	%                              ) * s.frame_exp;
%%% ***
	exposure = histogram(
	   aglc_expno_to_phase( frames_exposed + dither_frames,
%	   s.frame_exp,
	   s.timedel * s.READ_GRP_PER_EXPNO,    %%% 1.5.0 *** CHECK
	   s.timepixr,
	   s.tstart,
	   s.mjdref
	   ),
%	   phase_grid,,
	   phase_grid, phase_grid_hi,
	   %	                      phase_grid_hi         % v0.11 ***
	   ) * s.frame_exp;

	c.exposure += exposure ;  %%% *** will be overwritten; as is, is WEIGHTED and may not make sense.
	c.ccd_exposure[ ccd[i], * ] += exposure ; 

	c.count_rate     += counts / exposure;   % contribution to the total count rate
	c.count_rate_err += counts / exposure^2; % contribution to the square of (the total count rate error = sqrt(counts)/exposure)

	%% For exposure reverse indices, 
	%%  compute throw-away histogram of expno on desired grid.
	%%  (Most of the exposure computation is to determine *which*
	%%  frames were exposed, not how many event are in each frame.)

	%%% v0.11 *** changed.....................................
	%%%%v0.14 *** changed back --- need to test.
	% 1.3.0:
	% 	() = hist1d(
	% 	     aglc_expno_to_phase( s.expno[ l_ccd ],
	% 	                          s.frame_exp,
	% 	                          s.timepixr,
	% 	                          s.tstart,
	% 	                          s.mjdref
	% 	                         ),
	% 	                         phase_grid,
	% %	                         phase_grid_hi,  % *** v0.11
	%                                  &revidx );
	() = histogram(	aglc_expno_to_phase( s.expno[ l_ccd ],
	s.frame_exp,
	s.timepixr,
	s.tstart,
	s.mjdref
	),
%	phase_grid,,
	phase_grid, phase_grid_hi,
	%	                         phase_grid_hi,  % *** v0.11
	&revidx );

	% Accumulate the reverse indices for counts 
	% Since we indexed into counts w/ lev,
	% un-index to get the "raw" indices
	% from the full lists.  That means for each ccd, we want
	% lev[revidx[j]], where j is an index into expno_grid.

	% To concatenate over ccd index i, we can do something like
	% c.revidx1a[j] = [c.revidx1a[j], lev[revidx1a[j]]]  for all j.

	for (j=0; j<npts; j++)
	{
	    c.revidx1a[j] = [ c.revidx1a[j],    lev[ revidx1a[j] ] ];
	    c.revidx[j] =   [   c.revidx[j],  l_ccd[ revidx[j] ] ];
	}
    }  
    %----------------- for (i=0; i<nccd; i++)
    () = printf("\n");
    
    c.exposure /= nccd ;         % mean exposure time  (w/ possibly bad weighting) %%% ***

    c.stat_err       = sqrt( c.counts ) ;
    c.count_rate_err = sqrt( c.count_rate_err ) ; 
    c.exposure       = c.counts  / c.count_rate ; 

%%% *** rate error

    c.phase_min = phase_grid ;
    c.phase_max = c.phase_min + dphase ;
    c.phase     = c.phase_min + dphase / 2.0 ;

    if (do_bkg)
    {
	variable bkg_backscale = ext_regions.bkg_backscale ; % ***
	vmessage("%% Scaling background by 1./%f", bkg_backscale );
	c.counts     /= bkg_backscale ;	
	c.stat_err   /= bkg_backscale ; 
	c.count_rate /= bkg_backscale ;
	c.count_rate_err /= bkg_backscale ; 
    }

    return c ;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% v0.17 explicitly add ephem to interface
% 
define aglc_phased()  % ACIS Grating Phase Curve  %{{{
%!%+
%\function{aglc_phased}
%\synopsis{Bin a phased light curve.}
%\usage{pc = aglc_phased(tgevt, expno_ref, phase_info, ephem, wmin, wmax, tg, orders[, bkg]);
%\altusage{pc = aglc_phased(   evt_struct   , phase_info, ephem, wmin, wmax, tg, orders[, bkg]);}
%}
%\description
%    Arguments are similar to those in \code{aglc}, with the exception of
%    \code{phase_info} and \code{ephem}.  For other arguments, see \code{aglc}.
%
%    \code{phase_info}  = a 3-element array giving the phase minimum, maximum,
%                  and bin size desired for the resultant phased light
%                  curve.
%    \code{ephem} = a two-element array giving the JD of zero phase (NOT MJD!)
%            and the period in days.
% 
%    Definitions for Phase Curve Structure Fields
%    The phase curve fields are the same as for the light curve, with the exception
%    of phase replacing time and the addition of the ephemeris. The new fields are:\n
%    1. \code{phase}: Value of phase at the bin center.\n
%    2. \code{phase_min}: Value of phase at the low edge of the bin.\n
%    3. \code{phase_max}: Value of phase at the high edge of the bin.\n
%    4. \code{epoch}: Epoch of the ephemeris, in Julian Days (NOT MJD!).\n
%    5. period: Period of the ephemeris, in days.
%\seealso{aglc, aglc_filter, aglc_read_events}
%!%-
{
    %   USAGE: ephem = [ jd0, period_days ];
    %   USAGE: bin_info = [phase_min, phase_max, step];
    %   USAGE: c = aglc_phased(fevt1a, fevt1, bin_info, ephem, wmin, wmax, tg, orders[, bkg]);");
    %   USAGE: c = aglc_phased(evt_struct,    bin_info, ephem, wmin, wmax, tg, orders[, bkg]);");

    % 1st form: 8 or 9 args.
    % 2nd form: 7 or 8 args.

    variable args = __pop_args(_NARGS) ;

    variable s, fevt1a, fevt1, tg, orders, wmin, wmax, phases, ephem, do_bkg=0;
    variable do_read = 0;
    variable i;

    if ( (_NARGS < 7) or (_NARGS > 9) )
    {
	aglc_phased_usage;
	return -1;
    }

    if ( typeof(args[0].value) == Struct_Type )
    do_read = 0;
    else
    do_read = 1;

    __push_args(args);


    if ( (_NARGS == 9) or (_NARGS==8 and not do_read) )
    {  
	do_bkg = ();
	%	if (do_bkg > 0) backscale = bkg_backscale;    % ***
	if (do_bkg > 0) ext_regions.backscale = ext_regions.bkg_backscale;
    }
    
    orders = ();

    tg = ();

    % only the 1st char of "tg" matters, as "h", "l", "m"
    tg = [tg] ;   % force to array type, if not already.
    for (i=0; i<length(tg); i++)  tg[i] = strlow(substr(tg[i],1,1));

    wmax = ();
    wmin = ();
    ephem = () ; 

    if (length(ephem) !=2)
    {
	usage;
	return -1;
    }

    aglc_set_ephem( ephem[0], ephem[1] ) ;  % store in static vars.

    phases = ();

    if (length(phases) !=3)
    {
	usage;
	return -1;
    }

    if (do_read)
    {
	fevt1 = ();
	fevt1a = ();
    }
    else
    {
	s = ();
    }

    !if (aglc_phased_par_check(phases, wmin, wmax, tg, orders) )
    {
	usage;
	return -1;
    }

    %  command args OK - read file:

    if (do_read)
    s = aglc_read_events( fevt1a, fevt1 ) ;

    if( qualifier_exists("aglc1") )
      return _aglc_phased( s, phases, wmin, wmax, tg, orders, do_bkg );
    else
      return _aglc_phased2( s, phases, wmin, wmax, tg, orders, do_bkg;; __qualifiers );
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%% take a 1D reverse index array, and create a single vector of indices.
%%   NOTE: this only works for the specific case of all the Array_Type[k]
%%   being the same type, namely, Integer_Type.
%%   
%
private define reform_revidx( rev )  %{{{
{
    % Example:   if c is the return value from aglc, then
    %            a = reform_revidx( c.revidx );
    %    will result in "a" being a 1D list of all the elements indexed
    %    by revidx.  This can be useful because then c.wave[a] will be the
    %    subset list of wavelengths of events binned into the light curve
    %    

    variable i;
    variable a = Integer_Type[0];

    for (i=0; i<length(rev); i++)	a = [a, rev[ i ] ] ;

    return a ;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Given an array of time bin indices, use the reverse indices in the 
%  light or phase curve structure to filter events in the event structure,
%  ev_struct.  
%  Returns a new event structure, which can then be processed w/ aglc or aglc_phased
%  to make new light or phase curves.
%
%  NOTE: One should only use the full curve data to filter, 
%        not already filtered structures, since all info is not in the 
%        filtered result.
%
define aglc_filter()  %{{{
%\function{aglc_filter}
%\synopsis{Apply a time or phase filter specified to the event structures.}
%\usage{s = aglc_filter( c_struct, ev_struct, c_bin_list );}
%\description
%    Given an index list, c_bin_list, derived from a light or phase
%    curve histogram, use the reverse histogram indices in the curve
%    structure, c_struct, and filter the events to include only the
%    events binned into the index list.  Return a new event structure,
%    s, suitable for input to aglc() or aglc_phased().
%
%    The input c_struct is as returned by aglc() or aglc_phased().  The reverse
%    indices are saved by the histogram functions and keep lists of all
%    indices contributing to a bin value.
%
%    The input ev_struct is an event structure as returned by
%    aglc_read_events().
%\example
%   s = aglc_read_events( fevt_1a, fevt );
%   ephem = [ 2452506.4328, 0.27831149 ] ;
%   p = aglc_phased( s, [0.0, 1.0, 0.02], ephem, 1.7, 25, ["H","M"], [-1,1] );
%   % Select events from a phase range and filter:
%   s_new = aglc_filter( p, s, where(p.phase_min > 0.5) );
%\seealso{aglc_read_events, aglc, aglc_phased}
%!%-
{
    if (_NARGS != 3 )
    {
	()=printf("\nUSAGE:  s = aglc_filter( c_struct, ev_struct, c_bin_list );");
	()=printf("\n   Apply a time or phase filter specified by the time or phase");
	()=printf("\n    bin list, c_bin_list, to the event structure, ev_struct, ");
	()=printf("\n    using reverse histogram indices in c_struct.");
	()=printf("\n   Return a new event structure, s, suitable for input to");
	()=printf("\n    aglc() or aglc_phased().");
	()=printf("\n\nNOTE: c_struct can be as returned by aglc() or aglc_phased().");
	()=printf("\n\nEXAMPLE: ");
	()=printf("\n  s = aglc_read_events( fevt_1a, fevt );");
	()=printf("\n  ephem = [ 2452506.4328,0.27831149 ] ;"); % v0.17
	()=printf("\n  p = aglc_phased( s, [0.0, 1.0, 0.02], ephem, 1.7, 25, [\"H\",\"M\"], [-1,1] );");
	()=printf("\n  s_new = aglc_filter( p, s, where(p.phase_min > 0.5) );\n");

	return -1;
    }

    variable time_bin_indices = ();
    variable ev_struct = ();
    variable lc_struct = ();

    variable l, l_1a, s ;

    if ( length( time_bin_indices ) == 1 )  % need to force to vector.
    time_bin_indices = [time_bin_indices];

    l_1a = reform_revidx( lc_struct.revidx1a[ time_bin_indices ] ) ;
    l    = reform_revidx( lc_struct.revidx[   time_bin_indices ] );

    s = @ev_struct;

    s.expno    = s.expno[ l ];
    s.expno_1a = s.expno_1a[ l_1a ] ;
    s.ccd      = s.ccd[ l ];
    s.ccd_1a   = s.ccd_1a[ l_1a ];
    s.tgpart   = s.tgpart[ l_1a ];
    s.order    = s.order[ l_1a ];
    s.tgd      = s.tgd[ l_1a ];
    s.wave     = s.wave[ l_1a ];
    s.status   = s.status[ l_1a ];

    return s;
} %}}}
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%
%%%%  provide functions to write rudimentary FITS files.
%%%%%   Partially CXC-compliant, in extension keywords.
%%%%%   NULL primary only has required header keywords.
%
% Function Courtesy of John Houck:
% (http://space.mit.edu/CXC/ISIS/examples/ex_write_bintable.sl)
%
% This function summarizes the basics of writing a FITS bintable with
% a single HDU
%
private define write_bintable (filename, extname, data_struct, keyword_struct, history)  %{{{
{
    variable fp = fits_open_file (filename, "c");
    if (fp == NULL)
    {
	vmessage ("Failed to create file %s", filename);
	return -1;
    }

    fits_write_binary_table (fp, extname, data_struct, keyword_struct);
    () = _fits_write_history (fp, history);

    fp = NULL;
    return 0;
} %}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
define aglc_write_curve()  % ( fout, f_ref_hdr, c[, history ] )  %{{{
%!%+
%\function{aglc_write_curve}
%\synopsis{Write a light or phase curve to a FITS bintable file}
%\usage{aglc_write_curve( outfile, hdr_ref_file, curve_struct [, history_string ] );" );}
%\description
%    \code{outfile} = name of output file ( \code{String_Type}).\n
%    \code{hdr_ref_file} =  Reference file for copying of header info ( \code{String_Type}).\n
%    \code{curve_struct} =  The light or phase curve structure, as created by \code{aglc} or  \code{aglc_phased}.\n
%    \code{history_string} = arbitrary note (\code{String_Type}).
%\seealso{aglc, aglc_phased}
%!%-
{
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % fout = String_Type, output file name
    %
    % f_hdr_ref = String_Type, grating event file name, for header info.
    %     (Can use event struct fevt_1a field:   s.fevt_1a )
    %
    % c = Struct_Type, and is the output of aglc() or aglc_phased().
    %
    % history = String_Type  arbitrary history, notes, comment string.
    %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if (_NARGS < 3 )
    {
	message("");
	message("%% Write a light or phase curve to a FITS bintable file.");
	message( "%% USAGE: aglc_write_curve( outfile, hdr_ref_file, curve_struct [, history_string ] );" );
	message("%%    outfile: String_Type; name of output file.");
	message("%%    hdr_ref_file: Reference file for copy of header info.");
	message("%%    curve_struct:  light curve structure, as created by aglc() or aglc_phased().");
	message("%%    history_string: string field to record arbitrary note.");
	message("");
	return ; 
    }

    variable fout, f_ref_hdr, c, history ;
    history = "aglc_write_curve";            % default note.
    if (_NARGS == 4 ) history = ();
    c = ();
    f_ref_hdr = ();
    fout = ();

    % define header fields:
    %
    variable hdr =  struct {
	date,
	creator,
	mission,
	telescop,
	instrume,
	detnam,
	grating,
	content,
	extname,
	hduname,
	hduclass,
	hduclas1,
	object,
	obs_id,
	exposure,
	mjdref,
	timezero,
	aglc_epoch,
	aglc_period,
	backscal
    }
    ;
    
    variable extname, ctmp ;

    % define output structure conditionally on type of curve:
    %
    if ( struct_field_exists( c, "phase" ) )
    {
	extname = "PHASECURVE" ;
	hdr.extname  = extname ; 
	hdr.aglc_epoch    = c.epoch ;
	hdr.aglc_period   = c.period ;
	hdr.timezero = 0 ;
	
	ctmp = struct
	{
	    phase,
	    phase_min,
	    phase_max,
	    counts,
	    count_rate,
	    stat_err,
	    error,
	    exposure
	}
	;
	
	ctmp.phase     = c.phase ;
	ctmp.phase_min = c.phase_min ;
	ctmp.phase_max = c.phase_max ;	
    }
    else
    {
	extname      = "LIGHTCURVE" ;
	hdr.extname  = extname ; 
	hdr.aglc_epoch    = -1 ;
	hdr.aglc_period   = -1 ;
	hdr.timezero = c.timezero ;

	ctmp = struct	{
	    time,
	    time_min,
	    time_max,
	    counts,
	    count_rate,
	    stat_err,
	    error,
	    exposure
	}
	;
	
	ctmp.time     = c.time ;
	ctmp.time_min = c.time_min ;
	ctmp.time_max = c.time_max ;
    }

    % Copy remaining structure fields to output structure: 

    %    ctmp.counts     = c.counts ;
    ctmp.counts     = int( c.counts ) ;  % 1.3.1, to please cfitsio.
    ctmp.count_rate = c.count_rate ;
    ctmp.error      = c.count_rate_err ; 
    ctmp.stat_err   = c.stat_err ;
    ctmp.exposure   = c.exposure ;

    % Fill in header fields from curve structure or reference file header:

    hdr.hduname  = extname ;
    hdr.content  = extname ; 
    hdr.hduclass = "OGIP" ;
    hdr.hduclas1 = extname ;

%    hdr.exposure = sum( c.exposure ) ;
    hdr.exposure = sum( c.exposure[where(not isnan(c.exposure))] ) ;   % *** 
    hdr.mjdref   = c.mjdref ;
    hdr.object   = fits_read_key ( f_ref_hdr, "object" );
    hdr.obs_id   = fits_read_key ( f_ref_hdr, "obs_id" );

    hdr.mission  = fits_read_key ( f_ref_hdr, "mission" );
    hdr.telescop = fits_read_key ( f_ref_hdr, "telescop" );
    hdr.instrume = fits_read_key ( f_ref_hdr, "instrume" );
    hdr.detnam   = fits_read_key ( f_ref_hdr, "detnam" );
    hdr.grating  = fits_read_key ( f_ref_hdr, "grating" );

    hdr.creator = sprintf("aglc-%s", aglc_version_string );

    % 2008.04.12
    hdr.backscal = 1.0 ;  % by definition. IF background, also scaled.

    % construct the DATE created string
    %
    variable t = localtime( _time ) ; 
    hdr.date = sprintf("%04d-%02d-%02dT%02d:%02d:%02d",
    t.tm_year + 1900,
    t.tm_mon + 1,
    t.tm_mday, 
    t.tm_hour,
    t.tm_min,
    t.tm_sec )
    ;

    % output:
    () = write_bintable (fout, extname, ctmp, hdr, history);
} %}}}




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
provide("aglc"); 

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%vmessage("%% aglc version %s loaded.", aglc_version_string);



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

%%%% EXPERIMENTAL CODE %%%%


define aglc_truncate_events( s, l )  %{{{
{
    variable k = get_struct_field_names( s ) ;
    variable n = length( s.wave ) ;

    foreach ( k )
    {
	variable f = () ; 
	variable x = get_struct_field ( s, f ) ;
	if ( length( x ) == n )  set_struct_field( s, f,  x[ l ] ) ;
    }
} %}}}

define aglc_truncate_wave( s, wlo, whi )  %{{{
{
    variable i ;
    variable n = length( wlo ) ;

    variable l = Integer_Type[ length( s.wave ) ] ; 
    l *= 0 ; 

    for ( i=0; i<n; i++ )
    l = l or ( (s.wave > wlo[i]) and (s.wave <= whi[i] ) );

    aglc_truncate_events( s, where(l) ) ; 
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%; File: garf_gaps.sl
%; Author: D. Huenemoerder
%; Original version: 2004.07.19
%;====================================================================
% version 1.1 - 2004.12.17 - rob found bug... 
%
% version 1  - 2004.12.17 mod for ciao 3.2
% version: 0.2
%
% purpose: return wavelength regions in acis chip gaps, given an arf and order.
%
% method: use the fracexpo column, which is <1 if in a gap or near edge.
%         use the TG_M keyword to scale to desired order, for given
%         grating, and sign of order from header.

private define garf_gaps_usage()  %{{{
{
    message("");
    message("%% Struct_Type = garf_gaps( f_arf[, order_desired] );" ) ;
    message("");
    message("%% Given a grating arf file name, f_arf, return arrays of low and high");
    message("%%  wavelengths marking regions dithered into gaps.");
    message("%%  If order_desired is not specified, order is taken from the arf header.");
    message("%%  Restrictions: Due to geometric constraints, order_desired is");
    message("%%    forced to have the same sign as the arf header's order (TG_M).");
    message("");
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define garf_gaps()  % (f_arf, tg_m_desired)  %{{{
{
    variable f_arf = NULL;
    variable tg_m_desired = NULL ;
    variable r = struct{ low, high, type } ; 

    variable EPS = 0.003;   % some numerical noise in FRACEXPO;

    switch (_NARGS)

    {
	case 0:
	garf_gaps_usage;
	return NULL ; 
    }
    {
	case 1:
	f_arf = () ;
    }
    {
	case 2:
	tg_m_desired = () ;
	f_arf = () ;
    }
    {
	garf_gaps_usage;
	return NULL ; 
    }

    variable fexp, wlo, whi, tgm, l ; 

    (fexp, wlo, whi) = fits_read_col( f_arf, "FRACEXPO", "BIN_LO", "BIN_HI" ) ;

    % use ascending order arrays:
    wlo = reverse(wlo);
    whi = reverse(whi);
    fexp = reverse(fexp);

    tgm =  fits_read_key( f_arf, "TG_M" ) ;
    if ( tg_m_desired == NULL )  tg_m_desired = tgm ;

    % scale wavelengths to desired order: 
    wlo = wlo * abs( tgm / tg_m_desired ) ; 
    whi = whi * abs( tgm / tg_m_desired ) ; 

    % find the min and max indices of the regions...
    l =  fexp < (1.0-EPS) ;    % array l = 1 in the gaps. Find starts and stops.

    % differential: dl = -1 in bin before gap starts. +1 in last of gap.
    %               Edge effect: never(?) starts in a gap.
    %          2004.12.17: WRONG. Rob has one. Is bug.
    %                            always(?) ends in a gap (chip edge).    
    %
    % Complication: bad pixels create regions where FRACEXPO<1.
    %               Need to find chunks where FRACEXPO<1 and min(region)<0.6.
    %
    %    For MEG, the chip gaps make high-low > 0.6
    %    For HEG,                             > 0.3
    %              (but so can badpix in either case)

    if (l[0] == 1 ) l[0] = 0 ; %%% hack edge effect - starting in gap.

    variable dl;
    dl = l - shift(l,1);
    r.low  = wlo[ where( dl == -1 ) + 1 ] ;
    r.high = whi[ where( dl == 1 ) ] ;

    %
    % try to classify regions based on depth:
    %
    r.type = Integer_Type[ length( r.low ) ] ;
    variable i ;

    for ( i=0; i<length(r.low); i++ )
    {
	variable ll ;
	ll = where ( (wlo >= r.low[i])  and  (whi <= r.high[i] ) ) ; 

	if ( min( fexp[ll] ) < 0.5 )
	{
	    r.type[i] = 1 ;
	}
	else
	{
	    r.type[i] = 0 ;
	}
    }
    
    return r ;
} %}}}


define pr_garf_gaps( r )  %{{{
{
    variable i ;
    variable typ = ["badpix", "gap"];

    () = printf("%2s %8s %8s %8s %s\n",
    "n", "wmin", "wmax", "dw", "type" );

    for (i=0; i<length(r.low); i++)
    {
	() = printf("");
	() = printf("%2d %8.3f %8.3f %8.3f %s\n",
	i, r.low[i], r.high[i], r.high[i]-r.low[i], typ[r.type[i]] ) ;
	() = printf("");
    }
} %}}}

define plt_garf_gaps( r, typ, yra, sty, col ) %{{{
{
    % overplot styled rectangle at gap and bpix ranges.
    % r is a structure as returned by garf_gaps
    % typ is the type of gap to draw: "gap" or "badpix"
    % yra is array giving ymin, ymax of box to plot
    % sty is a pgplot fill style (1-4);
    % col is a pgplot color index

    variable styp = ["badpix", "gap"] ;

    variable l = where( styp[ r.type ] == typ ) ;

    if ( length(l) == 0 ) error("%% plt_garf_gaps: Bad type.");

    _pgsfs( sty ) ;
    _pgsci( col ) ; 

    variable i ;
    for (i=0; i<length(l); i++)
    {
	variable n = l[ i ] ; 
	_pgrect( r.low[ n ], r.high[ n ], yra[0], yra[1] );
    }
} %}}}

provide("garf_gaps");
provide("pr_garf_gaps");


%use it via:
%
% s = aglc_read_events( evt1a, evt0 ) ; 
% g = garf_gaps ( garf ) ;  % external; finds chip gaps low, high wavelengths.
% aglc_truncate_events( s, g.low, g.high ) ; % remove gaps from events.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Code from src/slang/physics/alpha_ff.sl
%
define alpha_ff() {
%!%+
%\function{alpha_ff}
%\synopsis{Calculate the absorption coefficient for free-free absorption}
%\usage{Double_Type alpha = alpha_ff(nu,T);}
%
%\qualifiers{
%\qualifier{Z}{average nuclear charge (default=1)}
%\qualifier{ne}{electron particle density (cm^-3; default: 1e10)}
%\qualifier{np}{proton particle density (cm^-3; default: 1e10)}
%}
%\description
%    This function returns the absorption coefficient for free-free radiation
%    (bremsstrahlung). The frequency, nu, is measured in Hz and can be an array,
%    the temperature T is measured in K.
%
%    For the moment this function assumes the Gauntt factor equals unity.
%
%    Note: per Kirchhoff's law the total bremsstrahlung spectrum from
%    a slab of size R is
%    B_nu(nu,T)*(1.-exp(-R*alpha_ff(nu,T)))
%
%\seealso{j_ff}
%!%-

    variable nu,T;
    (nu,T)=();
    
    variable h=6.62606885e-27; % erg s
    % variable c=2.99792457e10;  % cm/s
    variable k=1.3806504e-16;  % erg/K
    % variable qe=4.80320451e-10;% esu 
    % variable me=9.10938e-28;   % g

    variable Z=qualifier("Z",1.);
    variable ne=qualifier("ne",1e10);
    variable ni=qualifier("ni",1e10);
    
    variable gaunt_ff=1.; % for simplicity and until I've programmed it
    
    %    variable prefac=4*qe^6/(3.*me*h*c) * sqrt(2*PI/3/k/me);
    variable prefac=3.6923492876535404e+08;
    
    variable alpha=prefac*Z*Z*ne*ni/sqrt(T)/nu/nu/nu *(1.-exp(-h*nu/k/T)) *gaunt_ff;

    return alpha;
}

%
% Code from src/slang/strings/angle2string.sl
%
define angle2string()
%!%+
%\function{angle2string}
%\synopsis{convert an angle to a string for pretty printing}
%\usage{string=angle2string(angle;qualifiers)}
%\qualifiers{
%   \qualifier{hoursign}{angle is a hour angle, always include sign}
%   \qualifier{declination}{angle is a declination, always include sign}
%   \qualifier{latitude}{angle is a latitude, always include sign}
%   \qualifier{hours}{display hours, not degrees}
%   \qualifier{deg}{input angle is in degrees (default: radians)}
%   \qualifier{separator}{string to insert between the degrees/hours,
%        minutes, and after the seconds. If a scalar: insert between
%        degrees and minutes only [e.g., ":"]. If an array: insert
%        the three array elements between the output numbers (e.g.,
%        ["h","m","s"]). Default is ["d","m","s"] and ["h","m","s"]
%        depending on whether the hours-qualifier is set or not.}
%   \qualifier{blank}{do not display leading zeros}
%   \qualifier{secfmt}{sprintf format for the seconds (default:
%           %04.1f, i.e., a precision of 0.1 sec)}
%   \qualifier{mas}{display to a precision of milliseconds
%        corresponds to secfmt="%06.3f"}
%   \qualifier{muas}{display to a precision of microseconds
%        corresponds to secfmt="%08.5f"}
%}
%\description
%  This function is used to produce well formatted strings out of
%  angular quantities which are given in radians or degrees.
%  (the default is radian).
%
%  Angles can be displayed either in degrees or in hours, 
%  various separators can be used to separate the hms-terms,
%  and the routine can be told to always display a sign (e.g.,
%  for declination- or latitude-like quantities), or not.
%
%  This routine is array safe.
%
%\seealso{hms2deg,dms2deg,generate_iauname}
%!%-
{
    variable angle=();

    if (Array_Type==typeof(angle)) {
	variable retarr=String_Type[length(angle)];
	variable i;
	_for i(0,length(angle)-1,1) {
	    retarr[i]=angle2string(angle[i];;__qualifiers());
	}
	return retarr;
    }


    
    variable sg=" ";
    if (qualifier_exists("hoursign") or
        qualifier_exists("declination") or
        qualifier_exists("latitude")) {
	sg="+";
    }

    if (angle<0.) {
	sg="-";
	angle=abs(angle);
    }

    if (not qualifier_exists("deg")) {
	angle=angle*180./PI;
    }

    variable sep;
    if (qualifier_exists("separator")) {
	sep=qualifier("separator");
	if (length(sep)==1) {
	    sep=[sep,sep,""];
	}
    } else {
	sep=["d","m","s"];
	if (qualifier_exists("hours")) {
	    sep[0]="h";
	}
    }
    
    variable d,m,s;
    if (qualifier_exists("hours")) {
	(d,m,s)=deg2dms(angle;hours);
    } else {
	(d,m,s)=deg2dms(angle);
    }

    variable fmt;
    variable zero="0";
    if (qualifier_exists("blank")) {
	zero=" ";
    }

    variable secfmt;
    if (qualifier_exists("mas")) {
	secfmt="%06.3f";
    } else {
	if (qualifier_exists("muas")) {
	    secfmt="%08.5f";
	} else {
	    secfmt=qualifier("secfmt","%04.1f");
	}
    }
    
    if (qualifier_exists("hours") or qualifier_exists("latitude") or qualifier_exists("declination")) {
	% fmt for degree/hour has only two digits
	fmt="%"+zero+"2u"+sep[0]+"%"+zero+"2u"+sep[1]+secfmt+sep[2];
    } else {
	% fmt for degree/hour has only two digits
	fmt="%"+zero+"3u"+sep[0]+"%"+zero+"2u"+sep[1]+secfmt+sep[2];
    }

    variable ret=sg+sprintf(fmt,d,m,s);
    return ret;
}

%
% Code from src/slang/math/angle_to_rad.sl
%
define angle_to_rad()
%!%+
%\function{angle_to_rad}
%\synopsis{converts an angle in degrees or h:m:s format into radian}
%\usage{Double_Type angle_to_rad(Double_Type x);}
%\qualifiers{
%\qualifier{unit}{[\code{="deg"}] unit of \code{x} (\code{"deg"}/\code{"hms"}/\code{"rad"})}
%}
%\description
%    \code{x} can be a scalar value or (unless \code{unit="rad"}) an array of the form\n
%    - \code{[deg, arcmin]} (respectively \code{[hour, min]} for \code{unit="hms"})\n
%    - \code{[deg, arcmin, arcsec]} (respectively \code{[hour, min, sec]} for \code{unit="hms"})\n
%    - \code{[sign, deg, arcmin, arcsec]} (respectively \code{[sign, hour, min, sec]}).\n
%    The latter representation is needed for \code{-1 < deg}/\code{hour < 0}.
%\seealso{hms2deg,dms2deg}
%!%-
{
  if(_NARGS==0)  { help(_function_name()); return; }
  variable unit = qualifier("unit", "deg");
  foreach (__pop_list(_NARGS))
  { variable x = ();
    variable s = 1;
    if(typeof(x)==Array_Type)
    { if(unit=="rad")
	vmessage(`error (%s): cannot convert array with unit="rad"`, _function_name());
      if(x[0]<0)  s = -1;
      switch(length(x))
      { case 2: x = s*(abs(x[0]) + x[1]/60.); }
      { case 3: x = s*(abs(x[0]) + x[1]/60. + x[2]/3600.); }
      { case 4: x = s*(x[1] + x[2]/60. + x[3]/3600.); }
    }
    switch(unit)
    { case "deg": x *= PI/180.; }
    { case "hms": x *= PI/12.; }
    { case "rad": ; }
    { vmessage(`error (%s): unknown unit="%s"`, _function_name(), unit); }
    x;  % left on stack
  }
}

%
% Code from src/slang/math/angular_separation.sl
%
define angular_separation()
%!%+
%\function{angular_separation}
%\usage{ sep=angular_separation(ra1,dec1,ra2,dec2);}
%\synopsis{calculates the angular distance between two points on a sphere}
%\qualifiers{
%\qualifier{deg}{if set, the input coordinates and output are in degrees (default: radian)}
%\qualifier{radian}{if set, the input coordinates and output are in radian (the default))}
%}
%\description
% This routine calculates the angular separation between points on the sky
% with coordinates ra1/dec1 and ra2/dec2.
%
% The function is equivalent to greatcircle_distance but is compatible in terms
% of its qualifiers with the other coordinate system routines. It also returns
% the angular separation in the units of the input parameters rather than
% always in rad and, by using the haversine formula, is less prone to round
% off errors for small angular separations.
%
% This function is array safe (for either ra1/dec1 or ra2/dec2).
%
%\seealso{hms2deg,dms2deg,angle2string,position_angle}
%!%-
{
    variable ra1in,dec1in,ra2in,dec2in;
    variable ra1,dec1,ra2,dec2;
    switch(_NARGS)
    { case 4: (ra1in,dec1in,ra2in,dec2in)=(); }
    { help(_function_name()); return; }

    if (qualifier_exists("deg")) {
	ra1=ra1in*PI/180.;
	dec1=dec1in*PI/180.;
	ra2=ra2in*PI/180.;
	dec2=dec2in*PI/180.;
    } else {
	ra1=@ra1in;
	dec1=@dec1in;
	ra2=@ra2in;
	dec2=@dec2in;
    }

    % haversine equation (per RW Sinnott, 1984, Sky and Telescope)
    variable a=sin((dec2-dec1)/2.)^2.+cos(dec1)*cos(dec2)*sin((ra2-ra1)/2.)^2.;
    variable dist=2*atan2(sqrt(a),sqrt(1.-a));
    
    if (qualifier_exists("deg")) {
	return dist*180./PI;
    }
    return dist;
}

%
% Code from src/slang/ansi_escape_code.sl
%
private variable _ansi_escape_codes = Assoc_Type[String_Type];
_ansi_escape_codes["reset"] = "0m";
% normal colors
_ansi_escape_codes["black"] = "30m";
_ansi_escape_codes["red"] = "31m";
_ansi_escape_codes["green"] = "32m";
_ansi_escape_codes["yellow"] = "33m";
_ansi_escape_codes["blue"] = "34m";
_ansi_escape_codes["magenta"] = "35m";
_ansi_escape_codes["cyan"] = "36m";
_ansi_escape_codes["white"] = "37m";
% bright colors
_ansi_escape_codes["bblack"] = "90m";
_ansi_escape_codes["bred"] = "91m";
_ansi_escape_codes["bgreen"] = "92m";
_ansi_escape_codes["byellow"] = "93m";
_ansi_escape_codes["bblue"] = "94m";
_ansi_escape_codes["bmagenta"] = "95m";
_ansi_escape_codes["bcyan"] = "96m";
_ansi_escape_codes["bwhite"] = "97m";
% text formats
_ansi_escape_codes["bold"] = "1m";
_ansi_escape_codes["italic"] = "3m";
_ansi_escape_codes["underline"] = "4m";
_ansi_escape_codes["blink"] = "5m";
% cursor movement
_ansi_escape_codes["up"] = "A";
_ansi_escape_codes["down"] = "B";
_ansi_escape_codes["forward"] = "C";
_ansi_escape_codes["back"] = "D";

%%%%%%%%%%%%%%%
define ansi_escape_code()
%%%%%%%%%%%%%%%
%!%+
%\function{ansi_escape_code}
%\synopsis{transforms a text format, e.g., a color into its corresponding ANSI code}
%\usage{String_Type ansi_escape_code(String_Type format[, String_Type text]);}
%\qualifiers{
%  \qualifier{list}{print a list of available keywords}
%}
%\description
%  Transforms the given (human readable) format string into
%  a sequence of ANSI escape codes. The format string consists
%  of keywords, e.g., colors separated by semicolons. A list
%  of supported keywords is printed setting the 'list'
%  qualifiers.
%
%  The cursor movement keywords 'up', 'down', 'forwards', and
%  'back' allow an optional preceding number, which specifies
%  the amount of the movement.
%
%  In order to reset a previous format you may use the 'reset'
%  keyword. In case an optional text is provided, this text
%  and the reset ANSI code is appended to the returned string
%  automatically.
%\example
%  ansi_escape_code("red;blink", "ALERT");
%  ansi_escape_code("3up"); % move three lines up
%  ansi_escape_code("blue"); % blue color *from now on*
%  ansi_escape_code("reset"); % turn off the format again
%!%-
{
  % list of keywords
  if (qualifier_exists("list")) {
    variable list = assoc_get_keys(_ansi_escape_codes);
    list = list[array_sort(list, &strcmp)];
    array_map(Void_Type, &message, list);
    return;
  }

  variable format, text = NULL;
  switch (_NARGS)
    { case 1: format = (); }
    { case 2: (format,text) = (); }
    { help(_function_name); return; }

  variable key, out = "", num, pos;

  % split the input format
  foreach key (strchop(format, ';', 0)) {
    % amount of cursor movement
    pos = array_map(Integer_Type, &is_substr, key, ["up", "down", "forward", "back"]);
    pos = pos[where(pos > 0)];
    if (length(pos) == 1) {
      num = atoi(substr(key, 1, pos[0]-1));
      key = substr(key, pos[0], -1);
    } else {
      num = 0;
    }
    % replace keyword
    if (assoc_key_exists(_ansi_escape_codes, key)) {
      out += sprintf("\033[%s%s", num > 0 ? sprintf("%d", num) : "", _ansi_escape_codes[key]);
    } else {
      vmessage("error (%s): unknown keyword '%s'", _function_name, key);
      return;
    }
  }

  % return
  if (text == NULL) { return out; }
  else { return sprintf("%s%s\033[0m", out, text); }
}

%
% Code from src/Mike/append_chain.sl
%
private define compare_header (t1, t2)
{
	variable ii;
	variable field_names1=get_struct_field_names(t1);
	variable field_names2=get_struct_field_names(t2);
	if (length(field_names1) != length(field_names2))
	{
		return -1;
	}
	_for ii(0, length(field_names1)-1, 1)
	{
		if (field_names1[ii] != field_names2[ii])
		{
			return -1;
		} else {
			if (get_struct_field(t1, field_names1[ii]) != get_struct_field(t2, field_names2[ii]) )
			{
				return -1;
			}
		}
	}
	return 0;
}

private define check_chain_headers (t1, t2)
{
	variable field_names1=get_struct_field_names(t1);
	variable field_names2=get_struct_field_names(t2);
	variable ii;
	if (length(field_names1) != length(field_names2))
	{
		return -1;
	}
	_for ii(0, length(field_names1)-1, 1)
	{
		if (field_names1[ii] != field_names2[ii])
		{
			return -1;
		}
	}

	_for ii(0, length(field_names1)-1, 1)
	{
		if ( (field_names1[ii] != "NAXIS2") && (field_names1[ii] != "NSTEPS") )
		{
			if ( get_struct_field (t1, field_names1[ii]) != get_struct_field(t2, field_names1[ii]) )
				return -1;
		}
	}
	return 0;
}

private define check_parameters_headers (t1, t2)
{
	variable field_names1=get_struct_field_names(t1);
	variable field_names2=get_struct_field_names(t2);
	variable ii;
	if (length(field_names1) != length(field_names2))
	{
		return -1;
	}
	_for ii(0, length(field_names1)-1, 1)
	{
		if (field_names1[ii] != field_names2[ii])
		{
			return -1;
		}
	}

	_for ii(0, length(field_names1)-1, 1)
	{
		if ( field_names1[ii] != "NAXIS2")
		{
			if ( get_struct_field (t1, field_names1[ii]) != get_struct_field(t2, field_names1[ii]) )
				return -1;
		}
	}
	return 0;
}

%%%%%%%%%%%%%%%%%%%%%%%%%
define append_chain  ()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{append_chain}
%\synopsis{Append a chain fits file to another chain fits file}
%\usage{append_chain(String_Type chainfile1, String_Type chainfile2, String_Type chainoutfile)}
%\qualifiers{
%\qualifier{verbose}{show progress}
%}
%\description
%       This function takes two chains stored in chainfile1 and chainfile2 written with the write_chain function
%       (or by emcee itself) and appends the second one to the first one, again writing a fits file.
%       This function does not make use of read_chain or write_chain, it's consisting only of
%       fits routines. The function itself does perform some sanity and safety checks (same data, same fit_fun)
%       but you have to make sure that you use the correct chains in the correct order.
%\seealso{combine_chain, read_chain, write_chain, emcee}
%!%-
{
	variable name1, name2, fname;
	switch(_NARGS)
	{ case 3: (name1, name2, fname) = (); }
	{ help(_function_name()); return; }
	if ( qualifier_exists("verbose") )
		vmessage("Analysing ...");
%TESTS
	%Starting for the first extension, PARAMETERS, which should basically be the same
	variable ii, jj;
	
	try {
		variable hp1 = fits_read_header(name1+"[1]");
	} catch AnyError: {
		sprintf("Error opening "+name1, stderr);
		throw OpenError;
	}
	try{
		variable hp2 = fits_read_header(name2+"[1]");
	} catch AnyError: {
		sprintf("Error opening "+name2, stderr);
		throw OpenError;
	}
	if ( compare_header(hp1, hp2) != 0 )
	{
		sprintf("*** Error: Header 1 (PARAMETERS) of "+name1+" and "+name2+" are not the same, apparently not the same model or data was used", stderr);
		return -1;
	} 

	%Second extension, MCMCCHAIN
	variable hm1 = fits_read_header(name1+"[2]");
	variable hm2 = fits_read_header(name2+"[2]");
	if (check_chain_headers(hm1, hm2) != 0)
	{
		sprintf("Error: Header 2 (MCMCCHAIN) of "+name1+" and "+name2+" are not the same, chains not from the same MCMC configuration", stderr);
		return -1;
	}
	variable nfree = hm1.NFREEPAR;
	variable nw = hm1.NWALKERS;
	%That's the offset


	%Third extension, CHAINSTATS
	variable hc1 = fits_read_header(name1+"[3]");
	variable hc2 = fits_read_header(name2+"[3]");
	if (check_parameters_headers (hc1, hc2) != 0)
	{
		sprintf("*** Error: Header 3 (CHAINSTATS) of "+name1+" and "+name2+" are not the same", stderr);
	}



%Reading first extensions
	variable sp1 = fits_read_table(name1+"[1]");
	variable sp2 = fits_read_table(name2+"[1]");
	_for ii(0, length(sp1.free_par)-1, 1)
	{
		if (sp1.free_par[ii] != sp2.free_par[ii])
		{
			sprintf("*** Error: Free parameters inconsistent", stderr);
			return -1;
		}
	}
	
	variable ff=fits_open_file(fname, "c");
%Writing first extensions
	if (qualifier_exists("verbose"))
		vmessage("***First Extension");
	variable extname = "PARAMETERS";
	variable colname = ["FREE_PAR"];
	variable nrows = nfree;
	variable ttype = ["J"];
	variable tunit = [" parameter indices"];
	fits_create_binary_table(ff, extname, nrows, colname, ttype, tunit);
	fits_update_key (ff, "PFILE", hp1.PFILE, "");
	fits_update_key (ff, "MODEL", hp1.MODEL, "");
	variable hist_comment =
	["This file was created by merging the results",
	"of two seperate emcee runs using the mcmc routine for isis, written by",
	"Mike Nowak, assuming that the first part was used as the init_chain",
	"qualifier for the second one. A variety of checks for sanity",
	"of the datasets are performed but the user has still to be careful.",
	"This extension contains information about the model",
	"PFILE are the parameter files that started the chain",
	"MODEL is the fit function that was applied to the data",
	"FREE_PAR are the parameter indices of the free parameters"];
	array_map(Void_Type, &fits_write_comment, ff, hist_comment);
	if (qualifier_exists("verbose"))
		vmessage("\tWriting parameters");
	if(_fits_write_col(ff,fits_get_colnum(ff,"FREE_PAR"),1,1,sp1.free_par))
		throw IOError;

%Reading second extensions
	if (qualifier_exists("verbose"))
		vmessage("***SecondExtension");
	extname = "MCMCCHAIN";
	colname = ["FITSTAT","UPDATE"], ttype = ["D","J"],
	tunit =  [" fit statistics"," update indicator"];
	nrows = hm1.NAXIS2 + hm2.NAXIS2;
	variable nsteps = hm1.NSTEPS + hm2.NSTEPS ;
	_for ii(0, nfree-1, 1)
	{
		colname = [colname,"CHAINS"+string(sp1.free_par[ii])];
		ttype = [ttype,"D"];
		tunit = [tunit, " parameter values"];
	}

	fits_create_binary_table(ff, extname, nrows, colname, ttype, tunit);
	
	fits_update_key(ff, "NWALKERS", nw, "");
	fits_update_key(ff, "NFREEPAR", nfree, "");
	fits_update_key(ff, "NSTEPS", nsteps, "");
	fits_update_key(ff, "CDF_SCL_LO", hm1.CDF_SCL_LO, "");
	fits_update_key(ff, "CDF_SCL_HI", hm1.CDF_SCL_HI, "");
	hist_comment =
	["A Markov chain generated by the emcee hammer subroutine",
	"NWALKERS is the number of walkers *per* free parameter",
	"NFREEPAR is the number of free parameters",
	"NSTEPS is the number of iterations for each walker",
	"CDF_SCL_LO: The step amplitude distribution goes as 1/sqrt(z),",
	"CDF_SCL_HI: bounded by z = 1/CDF_SCL_LO -> CDF_SCL_HI","",
	"The following file columns are the results unpacked",
	"as 1D vectors of length NWALKERS*NFREEPAR*NSTEPS:","",
	"FITSTAT contains the vector of walker fit statistics",
	"UPDATE is a yes/no answer as to whether a walker updated",
	"CHAINS# are the values for the free parameter given by",
	"parameter index #"
	];

	array_map(Void_Type, &fits_write_comment, ff, hist_comment);

	variable curr_col;
	variable fo = hm1.NAXIS2 + 1;
	variable COLS = String_Type [nfree+2];
	COLS[0] = "FITSTAT";
	COLS[1] = "UPDATE";
	_for ii (0, nfree-1) {
		COLS[ii+2] = "CHAINS"+string(string(sp1.free_par[ii]));
		}

	_for ii(0, length(COLS)-1, 1) {
		if ( qualifier_exists("verbose") )
			vmessage("\tReading "+COLS[ii]+", 1. chain");
		curr_col = fits_read_col (name1+"[2]", COLS[ii]);
		if ( qualifier_exists("verbose") )
			vmessage("\tWriting "+COLS[ii]+", 1. chain");
		if(_fits_write_col(ff,fits_get_colnum(ff,COLS[ii]),1,1,curr_col))
			throw IOError;

		if ( qualifier_exists("verbose") )
			vmessage("\tReading "+COLS[ii]+", 2. chain");
		curr_col = fits_read_col (name2+"[2]", COLS[ii]);
		if ( qualifier_exists("verbose") )
			vmessage("\tWriting "+COLS[ii]+", 2. chain");
		if(_fits_write_col(ff,fits_get_colnum(ff,COLS[ii]),fo,1,curr_col))
			throw IOError;
		}

%Writing third extension
	if (qualifier_exists("verbose"))
		vmessage("***Third Extension");
	COLS=["FRAC_UPDATE", "MIN_STAT", "MED_STAT", "MAX_STAT"];
	extname = "CHAINSTATS";
	colname = ["FRAC_UPDATE","MIN_STAT","MED_STAT","MAX_STAT"];
	nrows   = hc1.NAXIS2 + hc2.NAXIS2;
	fo = hc1.NAXIS2 + 1;
	ttype   = ["D","D","D","D"];
	tunit   = [" fraction"," chi2"," chi2"," chi2"];

	fits_create_binary_table(ff, extname, nrows, colname, ttype, tunit);
	hist_comment =
	["This extension contains some useful summary information",
	"for the individual chain steps",
	"FRAC_UPDATE is the fraction of walkers that updated",
	"MIN_STAT: minimum chi2 for a given step",
	"MED_STAT: median chi2 for a given step",
	"MAX_STAT: maximum chi2 for a given step"
	];
	array_map(Void_Type, &fits_write_comment, ff, hist_comment);

	_for ii(0, length(COLS)-1, 1) {
		if (qualifier_exists("verbose"))
			vmessage("\tReading "+COLS[ii]+", 1. chain");
		curr_col = fits_read_col(name1+"[3]", COLS[ii]);
		if (qualifier_exists("verbose"))
			vmessage("\tWriting "+COLS[ii]+", 1. chain");
		if (_fits_write_col(ff, fits_get_colnum(ff,COLS[ii]),1,1,curr_col))
			throw IOError;

		if (qualifier_exists("verbose"))
			vmessage("\tReading "+COLS[ii]+", 2. chain");
		curr_col = fits_read_col(name2+"[3]", COLS[ii]);
		if (qualifier_exists("verbose"))
			vmessage("\tWriting "+COLS[ii]+", 2. chain");
		if (_fits_write_col(ff, fits_get_colnum(ff, COLS[ii]),fo,1,curr_col))
			throw IOError;
	}

	fits_close_file(ff);
	return ;
}


define combine_chain ()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{combine_chain}
%\synopsis{Combine several mcmc chains generated with the same configuration}
%\usage{combine_chain(Array_Type chaininfiles, String_Type chainoutfile)}
%\qualifiers{
%\qualifier{verbose}{show progress}
%}
%\description
%       This function takes mcmc chains stored in files passed as an array of strings in chaininfiles,
%	written with the write_chain function (or by emcee itself) and combines them to one chain file,
%	again writing a fits file.
%       This function does not make use of read_chain or write_chain, it's consisting only of
%       fits routines. The function itself does perform some sanity and safety checks (same data, same fit_fun)
%       but you have to make sure that you use the correct chains in the correct order.
%\seealso{read_chain, write_chain, emcee, append_chain}
%!%-
{
	variable names, fname;
	switch(_NARGS)
	{ case 2: (names, fname) = (); }
	{ help(_function_name()); return; }
%What we need
	variable ii, jj;
	variable nf = length(names);
	variable hp = Struct_Type [nf];
	variable hm = @hp;
	variable hc = @hp;
%Reading all files headers of all extensions
	if (qualifier_exists("verbose"))
		vmessage("Reading headers ...");
	_for ii(0, nf-1, 1) {
		try {
			hp[ii] = fits_read_header(names[ii]+"[1]");
			hm[ii] = fits_read_header(names[ii]+"[2]");
			hc[ii] = fits_read_header(names[ii]+"[3]");
		} catch AnyError: {
			vmessage("*** Error opening "+names[ii]);
			throw OpenError;
		}
	}
%Testing for consistency
	if (qualifier_exists("verbose"))
		vmessage("Analysing ...");
	_for ii(1, nf-1, 1) {
		if (compare_header(hp[0], hp[ii]) != 0) {
			sprintf("*** Error: Header 1 (PARAMETERS) of "+names[ii]+" not the same as others, apparently not the same model or data was used", stderr);
			throw DataError;
		}
	}
	_for ii(1, nf-1, 1) {
		if (check_chain_headers(hm[0], hm[ii]) != 0) {
			sprintf("*** Error: Header 2 (MCMCCHAIN) of "+names[ii]+" not the same as others, apparently chains not from the same mcmc configuration", stderr);
			throw DataError;
		}
	}
	_for ii(1, nf-1, 1) {
		if (check_parameters_headers(hc[0], hc[ii]) != 0) {
			sprintf("*** Error: Header 3 (CHAINSTATS) of "+names[ii]+" not the same as others, apparently chains not from the same mcmc configuration", stderr);
			throw DataError;
		}
	}
	variable sp = Struct_Type[nf];
	_for ii(0, nf-1, 1) {
		try {
			sp[ii] = fits_read_table(names[ii]+"[1]");
		} catch AnyError: {
			sprintf("*** Error reading extension 1 (PARAMETERS) from "+names[ii], stderr);
			throw OpenError;
		}
	}
	variable nfree = hm[0].NFREEPAR;
	variable nw = hm[0].NWALKERS;
	_for ii(1, nf-1, 1) {
		_for jj(0, length(sp[0].free_par)-1, 1) {
			if (sp[0].free_par[jj] != sp[ii].free_par[jj]) {
				sprintf("*** Error: Free parameters of "+names[ii]+" differ from others", stderr);
				throw DataError;
			}
		}
	}
%Actually writing the file
%First extension
	variable ff=fits_open_file(fname, "c");
	if (qualifier_exists("verbose"))
		vmessage("***First Extension");
	variable extname = "PARAMETERS";
	variable colname = ["FREE_PAR"];
	variable nrows = nfree;
	variable ttype = ["J"];
	variable tunit = [" parameter indices"];
	fits_create_binary_table(ff, extname, nrows, colname, ttype, tunit);
	fits_update_key (ff, "PFILE", hp[0].PFILE, "");
	fits_update_key (ff, "MODEL", hp[0].MODEL, "");
	variable hist_comment =
	["This file was created by merging the results",
	"of several emcee runs using the mcmc routine for isis, written by",
	"Mike Nowak, assuming that the first part was used as the init_chain",
	"qualifier for the second one. A variety of checks for sanity",
	"of the datasets are performed but the user has still to be careful.",
	"This extension contains information about the model",
	"PFILE are the parameter files that started the chain",
	"MODEL is the fit function that was applied to the data",
	"FREE_PAR are the parameter indices of the free parameters"];
	array_map(Void_Type, &fits_write_comment, ff, hist_comment);
	if (qualifier_exists("verbose"))
		vmessage("\tWriting parameters");
	if(_fits_write_col(ff,fits_get_colnum(ff,"FREE_PAR"),1,1,sp[0].free_par))
		throw IOError;

%Second extension
	if (qualifier_exists("verbose"))
		vmessage("***Second extension");
	extname = "MCMCCHAIN";
	colname = ["FITSTAT","UPDATE"], ttype = ["D","J"],
	tunit =  [" fit statistics"," update indicator"];
	variable starts = Integer_Type [nf];
	variable nsteps = 0;
	nrows = 0;
	_for ii(0, nf-1, 1) {
		starts [ii] = nrows + 1;
		nrows += hm[ii].NAXIS2;
		nsteps += hm[ii].NSTEPS;
	}
	_for ii(0, nfree-1, 1)
	{
		colname = [colname,"CHAINS"+string(sp[0].free_par[ii])];
		ttype = [ttype,"D"];
		tunit = [tunit, " parameter values"];
	}
	fits_create_binary_table(ff, extname, nrows, colname, ttype, tunit);
	
	fits_update_key(ff, "NWALKERS", nw, "");
	fits_update_key(ff, "NFREEPAR", nfree, "");
	fits_update_key(ff, "NSTEPS", nsteps, "");
	fits_update_key(ff, "CDF_SCL_LO", hm[0].CDF_SCL_LO, "");
	fits_update_key(ff, "CDF_SCL_HI", hm[0].CDF_SCL_HI, "");
	hist_comment =
	["A Markov chain generated by the emcee hammer subroutine",
	"NWALKERS is the number of walkers *per* free parameter",
	"NFREEPAR is the number of free parameters",
	"NSTEPS is the number of iterations for each walker",
	"CDF_SCL_LO: The step amplitude distribution goes as 1/sqrt(z),",
	"CDF_SCL_HI: bounded by z = 1/CDF_SCL_LO -> CDF_SCL_HI","",
	"The following file columns are the results unpacked",
	"as 1D vectors of length NWALKERS*NFREEPAR*NSTEPS:","",
	"FITSTAT contains the vector of walker fit statistics",
	"UPDATE is a yes/no answer as to whether a walker updated",
	"CHAINS# are the values for the free parameter given by",
	"parameter index #"
	];
	
	array_map(Void_Type, &fits_write_comment, ff, hist_comment);

	variable curr_col;
	_for jj(0, nf-1, 1) {
		_for ii(0, length(colname)-1, 1) {
			if (qualifier_exists("verbose"))
				vmessage("\tReading "+colname[ii]+" "+string(jj+1)+".chain");
			curr_col = fits_read_col(names[jj]+"[2]", colname[ii]);
			if (qualifier_exists("verbose"))
				vmessage("\tWriting "+colname[ii]+" "+string(jj+1)+".chain");
			if(_fits_write_col(ff,fits_get_colnum(ff,colname[ii]),starts[jj],1,curr_col))
				throw IOError;
		}
	}

%Third extension
	if (qualifier_exists("verbose"))
		vmessage("***Third Extension");
	extname = "CHAINSTATS";
	colname = ["FRAC_UPDATE","MIN_STAT","MED_STAT","MAX_STAT"];
	nrows = 0;
	starts [*] = 0;
	_for ii (0, nf-1, 1) {
		starts[ii] = nrows+1;
		nrows += hc[ii].NAXIS2;
	}
	ttype   = ["D","D","D","D"];
	tunit   = [" fraction"," chi2"," chi2"," chi2"];

	fits_create_binary_table(ff, extname, nrows, colname, ttype, tunit);
	hist_comment =
	["This extension contains some useful summary information",
	"for the individual chain steps",
	"FRAC_UPDATE is the fraction of walkers that updated",
	"MIN_STAT: minimum chi2 for a given step",
	"MED_STAT: median chi2 for a given step",
	"MAX_STAT: maximum chi2 for a given step"
	];
	array_map(Void_Type, &fits_write_comment, ff, hist_comment);

	_for jj(0, nf-1, 1) {
		_for ii(0, length(colname)-1, 1) {
			if (qualifier_exists("verbose"))
				vmessage("\tReading "+colname[ii]+" "+string(jj+1)+".chain");
			curr_col = fits_read_col(names[jj]+"[3]", colname[ii]);
			if (qualifier_exists("verbose"))
				vmessage("\tWriting "+colname[ii]+" "+string(jj+1)+".chain");
			if(_fits_write_col(ff,fits_get_colnum(ff,colname[ii]),starts[jj],1,curr_col))
				throw IOError;
		}
	}

	fits_close_file(ff);
	return ;
}

%
% Code from src/slang/structs/tables/struct-arrays/append_struct_arrays.sl
%
define append_struct_arrays()
%!%+
%\function{append_struct_arrays}
%\synopsis{appends a structure's fields to another structures's fields}
%\usage{append_struct_arrays(Struct_Type &s, Struct_Type additional_s[]);}
%\description
%    \code{s} and \code{additional_s} have to be structures with the same fields.
%    \code{s}, which is changed by this function, has to be passed by reference.
%    For every \code{field},\n
%       \code{s.field = [s.field, additional_s.field];}\n
%    The following two statements are equivalent:\n
%       \code{append_struct_arrays(&s, additional_s);}\n
%       \code{s = merge_struct_arrays( [s, additional_s] );}
%\seealso{merge_struct_arrays}
%!%-
{
  variable s, additional_s;
  switch(_NARGS)
  { case 2: (s, additional_s) = (); }
  { help(_function_name()); return; }

  if(@s==NULL) { @s = additional_s; return; }
  @s = merge_struct_arrays( [@s, additional_s] );
}

%
% Code from src/slang/aproposer.sl
%

% define default public variables
#if (__get_reference("APROPOS_FORMAT_NAME") == NULL)
variable APROPOS_FORMAT_NAME = "cyan";
#endif
#if (__get_reference("APROPOS_FORMAT_MATCH") == NULL)
variable APROPOS_FORMAT_MATCH = "bred";
#endif
#if (__get_reference("APROPOS_LINE_WIDTH") == NULL)
variable APROPOS_LINE_WIDTH = isatty(stdin)
  ? string_matches(
    fgetslines(popen("stty size", "r"))[0],
    `[0-9]* \([0-9]*\)\n`
  )
  : NULL;
APROPOS_LINE_WIDTH = APROPOS_LINE_WIDTH == NULL ? 80 : atoi(APROPOS_LINE_WIDTH[1]);
#endif
#if (__get_reference("APROPOS_ENABLE_SYNOPSIS") == NULL)
variable APROPOS_ENABLE_SYNOPSIS = 1;
#endif

% Helper function for printing the result of 'apropos_slang'
% taking the terminal's size into account. Input parameters
% are the _formatted_, i.e., highlighted function names
% (String_Type[]) and their _original_ string length
% (Integer_Type[]).
private define _print_apropos(s, len) {
  % calculate number of matches fitting into one line
  variable maxlen = max(len);
  variable cols = APROPOS_LINE_WIDTH / (maxlen + 2);
  % fill input with empty elements such that 'length(s) mod cols == 0'
  variable acol = length(s) mod cols;
  if (acol > 0) { acol = cols - acol; }
  s = [s, array_map(String_Type, &sprintf, "", ones(acol))];
  len = [len, Integer_Type[acol]];
  % define sprintf format ("%s" times cols)
  variable format = strjoin(array_map(String_Type, &sprintf, "%%s", ones(cols)));
  % since escape codes are used in the strings,
  % their individual string lengths are larger than the actual printed ones
  % -> move cursor manually back and forth
  variable i;
  _for i (0, length(s)-1, 1) {
    s[i] = sprintf("%s%s%s", s[i],
      ansi_escape_code(sprintf("%dback", len[i])),
      ansi_escape_code(sprintf("%dforward", maxlen + 2))
    );
  }
  % loop and print
  _for i (0, length(s)-1, cols) {
    vmessage(format, __push_array(s[[i:i+cols-1]]));
  }
}

% Helper function for printing the result of the synopsis
% search, within 'aproposer', taking the terminal's size into
% account. Input paramter are _formatted_ output lines to
% print (String_Type[]), their _original_ string length
% (Integer_Type[]), and the string length of the function
% (Integer_Type[]).
private define _print_aproposer(s, len, funlen) {
  variable n = length(s);
  vmessage("\nFound %d synopsis match%s:", n, n > 0 ? "es" : "");
  variable tmp, cslen, k;
  _for n (0, length(s)-1, 1) {
    tmp = s[n];
    % trim string of fit into the terminal
    if (len[n] > APROPOS_LINE_WIDTH) {
      % split into words
      tmp = strchop(tmp, ' ', 0);
      % cumsum length
      cslen = int(cumsum(array_map(Integer_Type, &strlen, tmp) + [0,ones(length(tmp)-1)]));
      % split at terminal width
      k = where(diff(int(cslen) / APROPOS_LINE_WIDTH));
      if (length(k) > 0) { % add line break and indent
	k++; if (k[-1] >= length(tmp)) { k = k[[:-2]]; }
	tmp[k] = tmp[k] + "\n";
	k++; if (k[-1] >= length(tmp)) { k = k[[:-2]]; }
	tmp[k] = sprintf(sprintf("%%%ds", funlen[n]+1), "") + tmp[k];
	   % +1 = +2 (': ') -1 (already existing space due to strjoin below)
      }
      % merge
      tmp = strjoin(tmp, " ");
    }
    % print
    vmessage(tmp);
  }
}

% a modified clone of the intrinsic slang 'apropos' (from cmds.sl),
% which highlights the matching substrings
private define apropos_slang()
{
  if (_NARGS != 1) { help(_function_name); return; }

  % substring to search for
  variable what = ();
  what = str_delete_chars(what, ";\"");
  variable wlen = strlen(what);
  variable owhat = strlow(what); % remember original search keyword
  if (0 == is_substr (what, "\\")) { what = "\\C" + what; } % case insensitive search

  % define namespaces and masks
  variable namespaces = [_isis->Isis_Public_Namespace_Name, current_namespace()];
  variable masks = Assoc_Type[];
  masks["function"] = 1|2;
  masks["variable"] = 4|8;

  % prepare output structure
  variable s = struct_array(length(masks)*length(namespaces), struct {
    section = "", % section's title
    output = String_Type[0], % matches in each section
    length = Integer_Type[0] % string length of matches
  });
 
  % loop over types and namespaces
  variable ns, type, list, n, i, pos, k = 0;
  foreach type (["function", "variable"]) {
    foreach ns (namespaces) {
      % find variables or functions within the namespace
      list = _apropos (ns, what, masks[type]);
      n = length(list);
      if (n == 0) { continue; }
      s[k].output = String_Type[n];
      s[k].length = Integer_Type[n];
      % add matches to output
      s[k].section = sprintf("%sFound %d %s match%s in namespace %S:",
	length(s) > 0 ? "\n" : "",
        n, type, n>1 ? "es" : "", ns
      );
      list = list[array_sort(list, &strcmp)];
      _for i (0, n-1, 1) {
	pos = is_substr(strlow(list[i]), owhat); % need to use original keyword here!
	s[k].length[i] = strlen(list[i]);
	s[k].output[i] = sprintf("%s%s%s",
	  ansi_escape_code(APROPOS_FORMAT_NAME, substr(list[i], 1, pos-1)),
	  ansi_escape_code(APROPOS_FORMAT_MATCH, substr(list[i], pos, wlen)),
	  ansi_escape_code(APROPOS_FORMAT_NAME, substr(list[i], pos+wlen, -1))
        );
      }
      k++;
    }
  }

  %%% print matches
  _for i (0, length(s)-1, 1) {
    if (length(s[i].output) > 0) {
      vmessage(s[i].section);
      _print_apropos(s[i].output, s[i].length);
    }
  }
}

%%%%%%%%%%%%%%%
define aproposer()
%%%%%%%%%%%%%%%
%!%+
%\function{aproposer}
%\synopsis{recall object names and the documentation satisfying a regular expression}
%\usage{aproposer("s")}
%\description
%  This is an extended version of S-Lang's 'apropos' function,
%  which may be used to get a list of all defined objects
%  whose name matches the regular expression "s". In addition,
%  the SYNOPSIS of all functions described in the documentation
%  files is checked on matches as well. Finally, the output is
%  formatted such that the matching substring and the object's
%  name are printed in different colors (see below for format
%  options).
%
%  In order to use this function instead of the intrinsic 'apropos'
%  function, you can put
%    alias("aproposer", "apropos");
%  into, e.g., your .isisrc file.
%
%  Further variables can be defined within the .isisrc file for
%  more control options:
%    APROPOS_ENABLE_SYNOPSIS - 0 = turn off the synopsis search
%    APROPOS_FORMAT_NAME     - format string for the object's name
%    APROPOS_FORMAT_MATCH    - format string for the substring match
%    APROPOS_LINE_WIDTH      - number of columns of the terminal,
%        which is needed to format the output (default: output of
%        `stty size` or 80 if the first attempt fails)
%  Read the documentation of 'ansi_escape_code' for details about
%  the format string.
%\seealso{apropos, .apropos, help, who,
%    get_doc_files, ansi_escape_code}
%!%-
{
  if (_NARGS != 1) { help(_function_name); return; }

  % substring to search for
  variable what = ();
  what = strlow(str_delete_chars(what, ";\""));
  variable wlen = strlen(what);

  % call old apropos
  apropos_slang(what);

  ifnot (APROPOS_ENABLE_SYNOPSIS) { return; }
  
  % init data structure
  variable s = struct {
    function = String_Type[0], % function's name
    output   = String_Type[0], % output string (formatted)
    length   = Integer_Type[0] % original string length
  };
  % loop over all documentation files
  variable file, in_synopsis, in_function, i, content, pos, match, synops, funname;
  variable docfiles = get_doc_files();
  foreach file (docfiles) {
    % loop over the whole content of the file
    content = fgetslines(fopen(file, "r"));
    in_synopsis = 0; in_function = 0; match = "";
    _for i (0, length(content)-1, 1) {
      if (in_function) {
	% extract and check synopsis
        if (in_synopsis == 1) {
	  % check on end of synopsis (CAPITAL word at beginning)
	  if (string_match(content[i], "^ [A-Z]*\n"R)) {
            % check for a match in synopsis	    
            pos = is_substr(strlow(synops), what);
	    % match in synopsis found
	    if (pos) {
	      % add function's name (if not done below already) ...
	      if (match == "") {
		% ... to data structure
		s.function = [s.function, funname];
		% ... and to prepared output
		match = ansi_escape_code(
		  APROPOS_FORMAT_NAME, funname
		);
		% ... and add the string length
		s.length = [s.length, strlen(funname)];
	      }
	      % add and highlight synopsis
	      s.length[-1] = s.length[-1] + strlen(synops) + 2;
	      match = sprintf("%s: %s", match, sprintf("%s%s%s",
	        substr(synops, 1, pos-1),
	        ansi_escape_code(APROPOS_FORMAT_MATCH, substr(synops, pos, wlen)),
		substr(synops, pos+wlen, -1)
	      ));
	    }
	    % no match found, but match in function's name already
	    % -> add synopsis
	    else if (match != "") {
	      match = sprintf("%s: %s", match, synops);
	      s.length[-1] = s.length[-1] + strlen(synops) + 2;
	    }
	    % close synopsis section
	    in_synopsis = 2;
	  }
	  % add current line to synopsis string
	  else {
	    if (strlen(synops) > 0) { synops += " "; }
	    if (content[i] != "\n") {
              synops += str_delete_chars(strtrim_beg(content[i]), "\n");
	    }
	  }
	}
	else {
          % check for a match in function's name
	  if (in_function == 1) {
	    funname =  str_delete_chars(content[i-1], "\n");
	    synops = ""; % reset synopsis
            pos = is_substr(strlow(funname), what);
	    % match in name found
            if (pos) {
	      % add function's name and its length to data structure
	      s.function = [s.function, funname];
	      s.length = [s.length, strlen(funname)];
	      % prepare formatted output string, i.e., highlight match in function's name
	      match = sprintf("%s%s%s",
	        ansi_escape_code(APROPOS_FORMAT_NAME, substr(funname, 1, pos-1)),
	        ansi_escape_code(APROPOS_FORMAT_MATCH, substr(funname, pos, wlen)),
		ansi_escape_code(APROPOS_FORMAT_NAME, substr(funname, pos+wlen, -1))
	      );
	    }
	    in_function = 2;
	  }
	  % check on an opening synopsis section
	  if (in_synopsis == 0) {
            in_synopsis = string_match(content[i], "^ SYNOPSIS\n"R);
	  }
	  % check on end of function's documentation
	  if (string_match(content[i], "^-+\n"R) || i == length(content)-1) {
	    % synopsis section is not provided for this function
	    if (in_synopsis != 2) {
	      if (length(s.function) == length(s.output)+1) {
                s.function = s.function[[:-2]];
	        s.length = s.length[[:-2]];
	      }
	    }
            % * add prepared match to output
	    else if (match != "") {
	      s.output = [s.output, match]; match = "";
	    }
	    % reset
	    in_function = 0;
	    in_synopsis = 0;
	  }
	}
      }
      % check on beginning of a function's documentation
      else {
	in_function = string_match(content[i], "^[a-zA-Z].*\n"R);
      }
    }
  }

  %%% print matches
  if (length(s.function) > 0) {
    % remove multiple occurences of the same function and sort
    struct_filter(s, unique(s.function));
    struct_filter(s, array_sort(s.function, &strcmp));
    _print_aproposer(s.output, s.length, strlen(s.function));
  }
}

%
% Code from src/sixte/arfconstruct.sl
%

%%%%%%%%%%%%%%%%%%%%%%%
define ARFfromMirrorArea(arf, mirrorarea){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ARFfromMirrorArea}
%\synopsis{Sets the ARF values to the interpolated values of the given mirror area array.}
%\usage{ARFfromMirrorArea(arf, mirrorarea);}
%\seealso{ARFreadMirrorArea}
%!%-

  variable interpol_mirrorarea=interpol(0.5*(arf.energ_lo+arf.energ_hi), mirrorarea.energy, mirrorarea.area;extrapolate="none",null_value=0.0);
  interpol_mirrorarea[where(interpol_mirrorarea<0)] = 0.0;
  
  arf.specresp=interpol_mirrorarea;

}

%%%%%%%%%%%%%%%%%%%%%%%
private define ARFmultiplyFilter(arf, filter){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ARFmultiplyFilter}
%\synopsis{Multiplies the ARF values with the interpolated transmission of the given filter.}
%\usage{ARFmultiplyFilter(arf, filter);}
%\seealso{ARFfromMirrorArea, ARFmultiplySupport, ARFmultiplyFactor, ARFmultiplyDetQE}
%!%-

  variable interpol_filter=interpol(0.5*(arf.energ_lo+arf.energ_hi), filter.energy, filter.transmission;extrapolate="none",null_value=0.0);

  arf.specresp=arf.specresp*interpol_filter;

}

%%%%%%%%%%%%%%%%%%%%%%%
private define ARFmultiplySupport(arf, support, openFraction){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ARFmultiplySupport}
%\synopsis{Multiplies the ARF values with the interpolated transmission of the given support material, taking the open fraction into account.}
%\usage{ARFmultiplySupport(arf, support, openFraction);}
%\seealso{ARFfromMirrorArea, ARFmultiplyFilter, ARFmultiplyFactor, ARFmultiplyDetQE}
%!%-

  variable interpol_support=interpol(0.5*(arf.energ_lo+arf.energ_hi), support.energy, support.transmission;extrapolate="none",null_value=0.0);
  
arf.specresp=arf.specresp*(openFraction+(1.0-openFraction)*interpol_support);


}

%%%%%%%%%%%%%%%%%%%%%%%
private define ARFmultiplyFactor(arf, factor){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ARFmultiplyFactor}
%\synopsis{Multiplies the ARF values with the given factor.}
%\usage{ARFmultiplyFactor(arf, factor);}
%\seealso{ARFfromMirrorArea, ARFmultiplyFilter, ARFmultiplySupport, ARFmultiplyDetQE}
%!%-

  arf.specresp=arf.specresp*factor;

}

%%%%%%%%%%%%%%%%%%%%%%%
private define ARFmultiplyDetQE(arf, detectorLayers){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ARFmultiplyDetQE}
%\synopsis{Multiplies the ARF values with the total quantum efficiency, corresponding to the total transmission of the given detector layers.}
%\usage{ARFmultiplyDetQE(arf, detectorLayers);}
%\seealso{ARFfromMirrorArea, ARFmultiplyFilter, ARFmultiplySupport, ARFmultiplyFactor}
%!%-

  variable ii;
  variable total_transmission=Double_Type[length(arf.energ_hi)];
  total_transmission=1.;

  for(ii=0; ii<length(detectorLayers); ii++){
    variable layer_transmission=interpol(0.5*(arf.energ_lo+arf.energ_hi), 
					 detectorLayers[ii].energy, 
                                         detectorLayers[ii].transmission;
                                         extrapolate="none",null_value=0.0);
    total_transmission=total_transmission*layer_transmission;
  }
  
  variable qe=1.0-total_transmission;

   qe[where(qe<0)] = 0.0;
   
   arf.specresp=arf.specresp*qe;
}

%%%%%%%%%%%%%%%%%%%%%%%
private define ARFreadHenkeTransm(henkename){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ARFreadHenkeTransm}
%\synopsis{Reads a standard henke table ASCII-file and returns it as a structure. Energy (first column) must be given in eV and is converted into keV. Transmission (second column) between [0.;1.]. First 2 rows are assumed to be comments.}
%\usage{variable henketable=ARFreadHenkeTransm(henkename);}
%\seealso{ARFmultiplyFilter, ARFmultiplySupport, ARFmultiplyFactor, ARFmultiplyDetQE, ARFreadMirrorArea}
%!%-

   variable startline = qualifier("qestartline",3);
  variable henketable=ascii_read_table(henkename,
          [{"%F","energy"}, % [eV]
           {"%F","transmission"}];
            startline=startline);
            
  % convert ev to keV
   henketable.energy=henketable.energy/1000.;
  
  return henketable;
}



%%%%%%%%%%%%%%%%%%%%%%%
define ARFreadMirrorArea(mirrorname){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ARFreadMirrorArea}
%\synopsis{Reads the mirror area ASCII-file and returns it as a structure.}
%\usage{variable mirrorarea=ARFreadMirrorArea(mirrorname);}
%\description
%    
%    The mirror area file has to be in the format:
%
%    TEXT
%    energy[0] [kev] area[0] [cm^2]
%    energy[1] [kev] area[1] [cm^2]
%\seealso{ARFfromMirrorArea}
%!%-

  variable arfstartline=qualifier("arfstartline", 2);
  vmessage("ARF File: read from line %d", arfstartline);
  
  variable mirrorarea=ascii_read_table(mirrorname,
          [{"%F","energy"}, % [keV]
           {"%F","area"}]; % [cm^2]
            startline=arfstartline);
  
  return mirrorarea;
}

%%%%%%%%%%%%%%%%%%%%%%%
define ARFconstruct(e_lo, e_hi, mirrorareafile, filterfiles, supportfiles, openFraction, detectorlayerfiles, factor)
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ARFconstruct}
%\synopsis{Constructs an ARF in keV and cm^2-units from mirror area, filters, filter supports, detector material and additional factor information.}
%\usage{variable arf=ARFconstruct(e_lo, e_hi, mirrorareafile, filterfiles, supportfiles, open_fraction, detectorlayerfiles, factor);}
%\description
%    An ARF is built from a mirror area, filter-,
%    support grid-, detector layer transmission
%    information. The transmission files-parameters
%    have to be file names of transmission tables 
%    in the format:
%
%    TEXT
%    TEXT
%    energy[0] [ev] transmission[0]
%    energy[1] [ev] transmission[1]
%    ...             ...
%
%    These arrays are interpolated to the input energy 
%    grid (e_lo, e_hi).
%    The open_fraction is an array with the same lenght
%    as the number of filter support files and gives the
%    fraction of the support grid that is open for X-rays.
%    The mirror area file has to be in the format:
%
%    TEXT
%    energy[0] [kev] area[0] [cm^2]
%    energy[1] [kev] area[1] [cm^2]
%    ...             ...
%
%    The factor is an additional factor multiplied to the
%    ARF.
%    
%    The return value is a structure
%    
%    arf=struct{
%           energ_lo=e_lo,
%           energ_hi=e_hi,
%           specresp
%           };
%
%
%    To construct an ARF without mirrors included, it can be done by
%    setting mirrorareafile=NULL
%!%-
{
    variable arf=struct{
	energ_lo=e_lo,
	energ_hi=e_hi,
	specresp
    };

   variable mirrorarea;
   % allow to give no mirror area (makes sense for straylight
   % simulations)
   if (mirrorareafile==NULL){
      arf.specresp = e_lo*0 + 1;
   } else {
      % Get the mirror area
      mirrorarea=ARFreadMirrorArea(mirrorareafile ;; __qualifiers);
      ARFfromMirrorArea(arf, mirrorarea);
   }
      
  
    % Multiply with all filter transmissions
    variable ii;
    variable filter_transm;

    for(ii=0; ii<length(filterfiles); ii++){
	if(strcmp(filterfiles[ii], "NONE")){
	    filter_transm=ARFreadHenkeTransm(filterfiles[ii];;__qualifiers());
	    ARFmultiplyFilter(arf, filter_transm);
	}
    }

    % Multiply with all filter support transmissions
    variable support_transmission;
    for(ii=0; ii<length(supportfiles); ii++){
	if(strcmp(supportfiles[ii], "NONE")){
	    support_transmission=ARFreadHenkeTransm(supportfiles[ii];;__qualifiers());
	    ARFmultiplySupport(arf, support_transmission, openFraction[ii]);
	} else {
	    if (openFraction[ii]!=1.) {
		arf.specresp*=openFraction[ii];
	    }
	}
    }

    % Multiply with detector quantum efficiency
    variable n_layers=0;
    for(ii=0; ii<length(detectorlayerfiles); ii++){
	if(strcmp(detectorlayerfiles[ii], "NONE")){
	    n_layers++;
	}
    }
    if(n_layers>0){
	variable detector_layers=Struct_Type[n_layers];
	n_layers=0;
	for(ii=0; ii<length(detectorlayerfiles); ii++){
	    if(strcmp(detectorlayerfiles[ii], "NONE")){
		detector_layers[n_layers]=ARFreadHenkeTransm(detectorlayerfiles[ii];;__qualifiers() );
		n_layers++;
	    }
	}
	ARFmultiplyDetQE(arf, detector_layers);
    }
  
    % Multiply with additional factor
    ARFmultiplyFactor(arf, factor);
  
    % Return finished ARF
  
    return arf;

}

%
% Code from src/plot/array2image.sl
%
%%%%%%%%%%%%%%%%%%%%%
%
% array2image.sl
% A function to produce a 2D image from three 1D arrays (e.g, measurements at a
% set of x and y coordinates).
% 
% version 0.9
% Author: Paul Hemphill (pbh)
%
% Changelog:
% 2018-09-10: Added "help" qualifier
% 2018-09-08: Added some error handling, updated documentation and version number, added changelog (pbh)
% 2017-04-10: Added version number, updated documentation (pbh)
% 2017-04-03: Improved handling of user-defined pixel functions (pbh)
% 2017-03-31: Initial commit
%
%%%%%%%%%%%%%%%%%%%%%

% Version info
% Numerical version is <major>*1e5 + <minor>*1e3 + <patch>*1e1
% (so, e.g., version 1.2.3 would be 10203)
private variable _array2image_versionString = "0.9.0";
private variable _array2image_version = 900;

%%%%%%%%%%%%%%%%%%%%%
% Generic wrapper for user-defined functions for assigning pixel values
private define array2image_pixel(rev,z,quals)
{
  quals = struct_combine(@quals,"ndx");
  set_struct_field(quals,"ndx",rev);
  variable func = quals.func;
  return (@func)((@z)[rev];;quals);
}

%%%%%%%%%%%%%%%%%%%%%
define array2image()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{array2image}
%\synopsis{Converts three x, y, z arrays into a 2D array}
%\usage{image[,] = array2image(x,y,z)}
%\qualifiers{
%\qualifier{nbinx}{number of bins along x-axis}
%\qualifier{nbiny}{number of bins along y-axis}
%\qualifier{xgrid}{1D array containing grid of values for x-axis. Ignores nbinx.}
%\qualifier{ygrid}{1D array containing grid of values for y-axis. Ignores nbiny.}
%\qualifier{func}{reference to a single-parameter S-Lang function used to determine pixel values from z-values (see below).}
%\qualifier{func_quals}{struct containing additional qualifiers to be passed to the function defined in func}
%\qualifier{include_inf}{include points where x, y, or z are inf or -inf}
%\qualifier{include_nan}{Note that this means the returned 2D array is of the form
%   im[y,x]. By default this ignores any points with infinite or NaN values
%   (this can be turned off with the include_inf and include_nan switches).}
%}
%\description
%   The x-y grid is determined automatically from the ranges of the data, or
%   can be defined by the "xgrid" and "ygrid" qualifiers. If "xgrid" or "ygrid"
%   is a reference to a local variable, the automatically-determined grids will
%   be stored in these variables. This function uses histogram2d() to bin the
%   data and figure out which points fall into which bins, so see that
%   function's documentation for how the grids are handled.
%
%   Notes regarding coordinate grids: 
%   * histogram2d() uses the last bin as an "overflow" bin, so user-supplied
%     binning should have their last bin be SMALLER than the largest x- or
%     y-value (otherwise you will have a column or a row with no points
%     included).
%   * The grids define the EDGES of the bins used by histogram2d(). I think
%     (although I am not sure) that plot_image() and plot_contour() use any
%     supplied coordinate grid to define the CENTERS of the pixels. Check the
%     documentation of PGPLOT's PGIMAG routine to be sure.
%   * Finally, I have no idea what histogram2d() does when it gets a
%     pathologically-designed grid (e.g., bins with zero width, overlapping
%     bins, or bins with lower bounds larger than their upper bounds). Please
%     only use monotonically-increasing bins.
%
%   The value of each pixel is, by default, the average of all points which
%   fall into that bin. This can be changed by passing a reference to a
%   function via the "func" qualifier. 
%
%   The "func" qualifier must be a reference to an S-Lang function which takes
%   a single array as input and outputs a single value, e.g., mean(), sum(), or
%   length(). The input to this function is an array containing all z-values
%   that fall into the current pixel. However, this function also receives a
%   set of qualifiers. By default, this gives @func access to the data, the
%   coordinate grids, and the list of points in the current pixel via the
%   following qualifiers:
%   qualifier("x"), qualifier("y"), qualifier("z"): x, y, and z data coordinates
%   qualifier("xgrid"), qualifier("ygrid"): x and y coordinate grids
%   qualifier("ndx"): array of data indices in the current pixel
%
%   Extra qualifiers can be supplied via the func_quals qualifier, which
%   should be set to a structure containing any additional qualifiers you need.
%   This will be merged with the above set of qualifiers using struct_combine()
%   with user-defined qualifiers taking precedence over the defaults in cases
%   where the names are the same.
%
%   As an example, to set each pixel's value to the innner product of the
%   z-values and some other array of numbers (which we'll call "z2"), one could
%   define:
%   \code{
%   define arr2im_dotprod(zValues) {
%     variable z2 = qualifier("extra_array");
%     variable ndx = qualifier("ndx");
%     return inner_product(zValues,z2[ndx])[0];
%   }
%   }
%   
%   and then after reading in or defining x, y, z, and z2 arrays, do:
%   \code{
%   variable im = array2image(x,y,z;func=&arr2im_dotprod,func_quals=struct{extra_array=z2});
%   }
%
%   By default this function removes infinite and NaN values from the x, y, and
%   z arrays before doing anything; the include_inf and include_nan switches
%   disable this behavior. Note that including infinite and/or NaN values will
%   cause problems for the default grids and pixel value assignment, so
%   include_inf and include_nan are best used with user-defined functions and
%   grids.
%
%   This script is under development; please contact Paul Hemphill
%   (pbh@space.mit.edu) regarding bugs or missing features (please do not
%   report missing bugs; any bugs that are not present are missing
%   intentionally).
%
%   \seealso{histogram2d, plot_image}
% 
%!%-
{
  variable x,y,z;
  if(_NARGS == 0 || qualifier_exists("help")) {
    help(_function_name);
    return;
  } else if(_NARGS != 3) {
    usage("im[,] = array2image(x,y,z);");
  } else {
    (x,y,z) = ();
  }
  %%%%%%%%%%%%%%%%%%%%%
  % Arrays need to be equal lengths
  ifnot(length(x) == length(y) == length(z)) {
    error("Error: arrays must be equal length");
  }

  variable include_nan = qualifier("include_nan",0);
  variable include_inf = qualifier("include_inf",0);

  variable nbinx = qualifier("nbinx",32);
  variable nbiny = qualifier("nbiny",32);

  variable func = qualifier("func",&mean);
  variable is_simple = qualifier_exists("func") ? 0 : 1;

  %%%%%%%%%%%%%%%%%%%%%
  % If not specifically told not to, strip not-a-number and infinite values
  variable ndx;
  ifnot(include_nan) {
    ndx = where(not(isnan(x)) and not(isnan(y)) and not(isnan(z)));
    x = x[ndx];
    y = y[ndx];
    z = z[ndx];
  }
  ifnot(include_inf) {
    ndx = where(not(isinf(x)) and not(isinf(y)) and not(isinf(z)));
    x = x[ndx];
    y = y[ndx];
    z = z[ndx];
  }
  variable xgrid,ygrid,xgridRef,ygridRef;
  %%%%%%%%%%%%%%%%%%%%%
  % If there are no user-provided grids, figure out default grids for x and y
  % coordinates. If xgrid/ygrid are given as Ref_Type, we store the created
  % grid in the referenced variable.
  if(not(qualifier_exists("xgrid")) || typeof(qualifier("xgrid")) == Ref_Type) {
    %%%%%%%%%%%%%%%%%%%%%
    % histogram2d() uses the last bin as an "overflow" bin. This means if we
    % make the grid end at max(x), the final bin will always have zero elements
    % in it. So, instead, we place the last bin one bin-width lower than max(x).
    xgrid = [min(x):max(x)-(1.0/nbinx)*(max(x)-min(x)):#nbinx];
    if(qualifier_exists("xgrid")) {
      xgridRef = qualifier("xgrid");
      @xgridRef = xgrid;
    }
  } else {
    %%%%%%%%%%%%%%%%%%%%%
    % If xgrid is given and is NOT a Ref_Type, use it for the X grid.
    xgrid = qualifier("xgrid");
  }
  %%%%%%%%%%%%%%%%%%%%%
  % Repeat for Y.
  if(not(qualifier_exists("ygrid")) || typeof(qualifier("ygrid")) == Ref_Type) {
    %%%%%%%%%%%%%%%%%%%%%
    % See note for xgrid above
    ygrid = [min(y):max(y)-(1.0/nbiny)*(max(y)-min(y)):#nbiny];
    if(qualifier_exists("ygrid")) {
      ygridRef = qualifier("ygrid");
      @ygridRef = ygrid;
    }
  } else {
    ygrid = qualifier("ygrid");
  }
  nbinx = length(xgrid);
  nbiny = length(ygrid);

  %%%%%%%%%%%%%%%%%%%%%
  % histogram2d()'s reverse-index (fifth argument of histogram2d()) tells us
  % which points are in which bin
  variable rev;
  () = histogram2d(y,x,ygrid,xgrid,&rev);
  %%%%%%%%%%%%%%%%%%%%%
  % Turn rev into a 1D array so we can loop over it more easily. After this,
  % rev[i] contains the indices of all points falling into the i-th bin, so
  % z[rev[i]] gives us all z-values in that bin.
  reshape(rev,[nbinx*nbiny]);

  %%%%%%%%%%%%%%%%%%%%%
  % Now we just apply array2image_pixel() to each pixel. array2image_pixel() in
  % turn calls whatever function the user provided (by default this is mean()).
  variable defaultQuals = struct{x=x,y=y,z=z,xgrid=xgrid,ygrid=ygrid,func=func};
  variable userQuals = qualifier("func_quals");
  if(typeof(userQuals) != Struct_Type)
    throw InvalidParmError,"func_quals must be a Struct_Type",userQuals;
  variable quals = struct_combine(defaultQuals,userQuals);
  variable im = array_map(Double_Type,&array2image_pixel,rev,&z,&quals);
  %%%%%%%%%%%%%%%%%%%%%
  % The image is now stored in "im", but at the moment it's a 1D array, so the
  % last step is to reshape it into a 2D array
  reshape(im,[nbiny,nbinx]);

  return im;
}

%
% Code from src/slang/arrays/array2matrix.sl
%
private define indarray( s, e ){
  return [s:e];
}
private define __get( a, i ){
  return a[i[where(i<length(a))]];
}

define array2matrix()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{array2matrix}
%\synopsis{Transforms an Array_Type into a matrix}
%\usage{Any_Type[l,...] = array2matrix(Array_Type[l] array);}
%\qualifiers{
%\qualifier{filler[=-_NaN]:}{Filler for empty array entries}
%}
%\description
%    Similar to array_flatten, this function flattens the
%    array into a single one. In contrast to array_flatten,
%    the array_shape of the individual entries is conserved.
%    If the array_shape of the individual array entries is
%    not the same, the maximal dimension is determined and
%    empty entries are being filled with the value given
%    with the filler qualifier.
%\example
%    variable arr = Array_Type[3];
%    arr[0] = _reshape( [1:3*4*5], [3,4,5] );
%    arr[1] = _reshape( [1:3*4], [3,4] );
%    arr[2] = _reshape( [1:2*3*4], [2,3,4] );
%    
%    variable mat = array2matrix(arr);
%    print(mat);
%\seealso{array_flatten}
%!%-
{
  variable filler = qualifier("filler",-_NaN);
    
  variable A;
  switch(_NARGS)
  { case 1: A = (); }
  { help(_function_name); return; }

  variable len = length(A); % length of Array
  variable dim = array_map( Array_Type, &array_shape, A ); % dimensions of each Array entry
  variable udim = dim[unique(dim)];
  
  variable M;
  if( length(udim) == 1 ){
    M = array_flatten( A );
    reshape( M, [len,dim[0]] );
  }
  else{
    % Find the maximal length in each dimension
    variable mdlen= max(array_map(Integer_Type, &length, udim)); % #dims for each entry
    variable mdim = Integer_Type[mdlen];
    _for $1 ( 0, mdlen-1 ){ % in each dimension find the maximum
      mdim[$1] = max(array_flatten(array_map( Array_Type, &__get, udim, $1 )));
    }

    variable IND;
    M = (@filler)[Integer_Type[len,__push_array(mdim)]]; % Create matrix with max dimension
    _for $1 ( 0, len-1 ){
      IND = array_map( Array_Type, &indarray, 0, [ dim[$1]-1, Integer_Type[length(mdim)-length(dim[$1])] ] );
      M[ $1, __push_array(IND) ] = A[ $1 ];
    }
  }
  
  return M;
}

%
% Code from src/slang/arrays/array_density.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define array_density()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{array_density}
%\synopsis{calculates the 'intensity' of an array, i.e. the invers of
%the spacing}
%\usage{array_density(bin_lo, bin_hi);}
%!%-
{
   variable r_lo,r_hi;
   switch(_NARGS)
   { case 2: (r_lo, r_hi) = (); }
   {    help(_function_name()); return; }
   
   variable i;
   variable n = length(r_lo);
   variable dr = Double_Type[n];
   _for i(0,n-1,1)
   {
      dr[i] = (r_hi[i]-r_lo[i]);
   }
   
   variable intens = 1./dr;
   
   return intens;     
}

%
% Code from src/fitting/xyfit/array_fit_gauss.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define array_fit_gauss()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{array_fit_gauss}
%\synopsis{performs a gaussian fit on given x- or xy-data}
%\usage{Struct_Type array_fit_gauss(x [,y] [,dy] [,c] [,s] [,a] [,o]);}
%\qualifiers{
%\qualifier{frz}{boolean array determining which parameters (c,s,a,o) are frozen}
%\qualifier{keep}{keeps the data and fit function (be careful, needs to be deleted for the next fit}
%\qualifier{plot}{plot the given data and oplot the fit}
%\qualifier{oplot}{only oplot the fit}
%}
%\description
%   Tries to fit the given data by a gaussian and an offset. If the y-data is
%   omitted, the given x-data is interpreted as y-data. The x-values are then
%   generated as indices of the y-data.
%   The uncertainties of the y-data can be passed to the fit algorithm by the
%   `dy' parameter. If not given the errors are calculated assuming Poisson
%   statistics.
%   The four remaining parameters are the starting values:
%     c - center position
%     s - sigma
%     a - area
%     o - offset in y-direction
%   They may also be given by an array as parameter `c' in the order shown
%   above. The same order is used for the freeze status qualifier `frz'.
%   Performing the fit is done with the actual choosen fit method. The result
%   is returned as a structure of the form
%     Struct_Type { center, sigma, area, offset, chisqr }
%   where chisqr is the reduced chi-square value of the best fit.
%!%-
{
  variable x,y,dy,c,s,a,o;
  switch(_NARGS)
    { case 1: (y) = (); x = [0:length(y)-1]; }
    { case 2: (x,y) = (); }
    { case 3: (x,y,dy) = (); }
    { case 4: (x,y,dy,c) = (); }
    { case 5: (x,y,dy,c,s) = (); }
    { case 6: (x,y,dy,c,s,a) = (); }
    { case 7: (x,y,dy,c,s,a,o) = (); }
    { help(_function_name()); return; }
  if (typeof(y) == Null_Type) { y = x; x = [0:length(y)-1]; }
  if (length(c) > 1)
  {
    if (length(c) > 3) o=c[3];
    if (length(c) > 2) a=c[2];
    c=c[0];
    s=c[1];
  }
  else ifnot (length(c)==1) { throw RunTimeError, sprintf("error (%s): first starting value must be either a single integer or an array of length four",_function_name); }
  variable frz = qualifier("frz",[0,0,0,0]);
  ifnot (length(x)==length(y)) { throw RunTimeError, sprintf("error (%s): x- and y-arrays must be of equal length",_function_name); }
  ifnot (length(frz)==4) { vmessage("warning (%s): freeze qualifier must be an array of length four",_function_name); frz=[0,0,0,0]; }
  % starting values if not given
  variable dx = 0; if (min(x) < 1.) dx = 1.-min(x); % x values must be greater 1
  if ((not __is_initialized(&dy)) || typeof(dy) == Null_Type) dy = sqrt(y);
  else ifnot(length(y)==length(dy)) { throw RunTimeError, sprintf("error (%s): y- and dy-arrays must be of equal length",_function_name); }
  if ((not __is_initialized(&c)) || typeof(c) == Null_Type) c = x[where_max(x+dx)]+dx; % center
  else c = c+dx;
  if ((not __is_initialized(&s)) || typeof(s) == Null_Type) s = .5*(max(x)-min(x)); % sigma
  if ((not __is_initialized(&a)) || typeof(a) == Null_Type) a = (max(y)-min(y)); % area
  if ((not __is_initialized(&o)) || typeof(o) == Null_Type) o = 0.; % offset
  % store current fit specifications
  variable ff = get_fit_fun;
  variable fm = get_fit_method;
  variable freep = freeParameters;
  freeze(freep);
  % define counts
  variable id = define_counts(x+dx,make_hi_grid(x+dx),y,dy);
  variable mffC = sprintf("constant(%d)",id);
  variable mffG = sprintf("gauss(%d)",id);
  % define fit function and starting values
  fit_fun(mffC + " + " + mffG);
  set_par(mffC + ".factor",o,frz[3],0,0);
  set_par(mffG + ".center",c,frz[0],min(x+dx),max(x+dx));
  set_par(mffG + ".sigma",s,frz[1],0,DOUBLE_MAX);
  set_par(mffG + ".area",a,frz[2],0,0);
  % do the fit
  variable stat;
  ()=fit_counts(&stat;fit_verbose=-2);
  % return values
  variable ret = struct {
    center = get_par(mffG + ".center")-dx-.5*mean(make_hi_grid(x)-x),
    sigma = get_par(mffG + ".sigma"),
    area = get_par(mffG + ".area"),
    offset = get_par(mffC + ".factor"),
    chisqr = stat.statistic / (stat.num_bins - stat.num_variable_params)
  };
  % eventually plot
  if (qualifier_exists("plot")) plot(x,y);
  if (qualifier_exists("oplot") || qualifier_exists("plot"))
  {
    variable model = get_model_counts(id);
    oplot(model.bin_lo-dx,model.value);
  }

   if (not qualifier_exists("keep")){ 
      % delete data and restore previous options
      delete_data(id);
      fit_fun(ff);
      set_fit_method(fm);
      thaw(freep);
   }

  return ret;
}

%
% Code from src/slang/arrays/array_flatten.sl
%
define array_flatten(arr)
%!%+
%\function{array_flatten}
%\synopsis{flattens an array of Array_Type[] into a single array}
%\usage{Any_Type array_flatten(Array_Type[] array);}
%\description
%    The given 'array' itself contains arrays of any
%    data type. This function flattens the 'array'
%    such that all internal arrays are merged.
%
%    Also the array can have entries of any data type,
%    they must be of the same type! Otherwise the function
%    will throw an error.
%    To avoid this split the array, e.g., with
%
%       arr1 = arr[where(_typeof(arr)==Double_Type)];
%
%    or if there are NULL entries,
%
%       arr1 = arr[wherenot( arr == NULL )];
%
%   and use arr1 as argument for array_flatten!
%    
%\example
%    variable arr = Array_Type[3];
%    arr[0] = [1,2,3,4];
%    arr[1] = [5,6];
%    arr[2] = [7,8,9];
%    
%    variable arrF = array_flatten(arr);
%    % returns an Integer_Type[9] containing
%    % [1,2,3,4,5,6,7,8,9]
%!%-
{
  % In principle this function would look like:
  %    return [__push_array(arr[wherenot(arr))])];
  % However, the stack is limited to ~2000 elements!
  % Therefore a split is neccessary. Note that this method here is
  % MUCH MUCH faster then just using a loop over all entries, as e.g.
  % foreach e (arr) ret = [ret, e];

  
  variable len = length(arr);
  variable stacklimit = 2000; 
  variable af = [ __push_array( arr[[0:(len<stacklimit?len:stacklimit)-1]] ) ];
  
  variable left = len-stacklimit;
  variable i;
  _for i ( 1, len/stacklimit ){
    af = [ af, __push_array( arr[ i*stacklimit+ [0:(left<stacklimit?left:stacklimit)-1] ] ) ];
    left -= stacklimit;
  }
  
  return af; 
}

%
% Code from src/slang/arrays/array_insert.sl
%
define array_insert( A, IA, i )
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{array_insert}
%\synopsis{Inserts an array into another array}
%\usage{Any_Type[] = array_insert(Any_Type[] A, Any_Type[] IA, Integer_Type i);}
%\description
%    Inserts the array or single value IA into the array A at the index
%    position i. A and IA have to be of the same DataType_Type!
%    Basically it is:
%
%         [ A[[:i-1]], IA , A[[i:]] ]
%    
%\example
%    variable A  = [1:10];
%    variable IA = [2,1];
%    variable NA = array_insert( A, IA, 3 );
%    print(NA);
%!%-
{
  return [ A[[:i-1]], IA , A[[i:]] ]  ;
}

%
% Code from src/slang/arrays/array_map_index.sl
%
%{{{% array_map_index private functions
private define __aindex (start, stop, arg) {
  variable n = (typeof(start)==Ref_Type)?@start:[start:stop];
  return struct{value=(typeof(arg.value)==Array_Type)?arg.value[n]:arg.value};
}
private define __findex (start, stop, f, args_ref) {
  % single run of array_map_index
  variable args = array_map(Struct_Type, &__aindex, start, stop, @args_ref);
  return @f(__push_args(args));
}
%}}}%

define array_map_index ()
%{{{%
%!%+
%\function{array_map_index}
%\synopsis{Call a function on a subset of arrays}
%\usage{Array_Type array_map_index (Int_Type[] start, Int_Type[] stop,
%                              DataType_Type type, Ref_Type func, args, ...);}
%\altusage{Array_Type array_map_index (Int_Type[] index, DataType_Type type,
%                              Ref_Type func, args, ...);}
%\altusage{Array_Type array_map_index (Ref_Type start, Ref_Type stop,
%                              DataType_Type type, Ref_Type func, args, ...);}
%\altusage{Array_Type array_map_index (Ref_Type index, DataType_Type type,
%                              Ref_Type func, args, ...);}
%\description
%  The \code{array_map_index} function is somewhat the opposite of the
%  \code{array_map}. As \code{array_map} allows it to apply a function to every
%  array element individually, \code{array_map_index} can be used to apply a
%  function (taking an array) on one or more subsamples of the target array.
%
%  Case 1: If called with \code{start} and \code{stop} the function \code{func} is
%  applied to \code{args[start[i]:stop[i]]} for each \code{i}.
%
%  Case 2: If called with only one index array \code{index} the function is
%  applied to args[index]. This is the same as \code{func(args[index]);}.
%
%  IMPORTANT: array_map_index applys \code{func} to an array!
%
%  Case 3: If called with \code{&start} and \code{&stop} the function behaves the
%  same as for case 1, but this is usefull when you want to nest \code{array_map}
%  calls.
%
%  Case 4: If called with \code{index} the it behaves the same as for case 2 but,
%  again, this might be usefull for \code{array_map} nesting.
%
%  NOTE: This function can hide a lot of important operations so it should be
%  avoided for implementing algorithms. However, it is very useful for fast
%  scripting and interactive (shell) operations.
%
%\example
%  % Case 1: apply mean() to a subset of data
%  data = rand_gauss(0.1, 100) % generate 100 normal distributed random numbers
%  means = array_map_index([0,5,25], [99,94,74], &mean, data);
%  print(means);
%  % this outputs the same as
%  print(mean(data)); print(mean(data[[5:94]])); print(mean(data[[25:74]]));
%
%  % Case 2: apply mean() to specified subset
%  data = rand_gauss(0.1, 100) % 100 random numbers
%  idx = where(data < 0) % select only negative data
%  neg_mean = array_map_index(idx, Double_Type, &mean, data);
%  print(neg_mean);
%  % this results in the same as
%  print(mean(data[n]));
%
%  % Case 3: apply mean() to array of arrays
%  data = Array_Type[3];
%  data[0] = rand_gauss(0.1, 100);
%  data[1] = rand_gauss(0.01, 100);
%  data[2] = rand_gauss(1, 100);
%  start = [0,5,25];
%  stop = [99,94,74];
%  means = array_map(Array_Type, &array_map_index, &start, &stop,
%                    Double_Type, &mean, data);
%  % returns the mean as in example for case 1 but applied to all sub arrays
%  % note that you have to declare the start and stop arrays explicitly because
%  % &[5:94] is not a valid expression
%
%  % Case 4: Same as case 3 but with only one index array
%
%  % Use array_map_index to call a function n times
%  rn = array_map_index([1:5], Array_Type, &rand_gauss, 0.1, 100);
%  % gives 5 arrays of 100 random numbers each
%
%\seealso{array_map, array_map_qualifiers}
%!%-
{
  variable rt, start, stop, f;
  variable args = __pop_args(_NARGS);
  (start, stop, rt) = __push_args(args[[:2]]);
  if (typeof(stop) == DataType_Type) {
    if (typeof(start) == Ref_Type)
      return array_map(stop, &__findex, start, @start, rt, &(args[[3:]])); % stop is $
    else
      return array_map(stop, &__findex, &start, @start, rt, &(args[[3:]])); % stop is$
  } else {
    f = __push_args(args[3]);
    if (typeof(start) == Ref_Type)
      return array_map(rt, &__findex, @start, @stop, f, &(args[[4:]]));
    else
      return array_map(rt, &__findex, start, stop, f, &(args[[4:]]));
  }
}
%}}}%

%
% Code from src/slang/arrays/array_map_qualifiers.sl
%

private define call_func_with_qualifiers ()
{
  variable q = ();
  variable args = __pop_list (_NARGS-2);
  variable f = ();
  
  return (@f)(__push_list(args) ;; q);
}

define array_map_qualifiers()
%!%+
%\function{array_map_qualifiers}
%\synopsis{Apply a function to each element of an array (also if given as qualifier)}
%\usage{Array_Type array_map (DataType_Type type, Ref_Type func, args, ... );}
%\altusage{(Array_Type, ...) array_map (DataType_Type type, ..., Ref_Type func, args, ... );}
%\description
%   The 'array_map_qualifiers' function is an extension of the 'array_map'
%   function and supports all its features.
%   This function also works for qualifiers! First of all qualifiers given to
%   the 'array_map_qualifiers' are passed onto the called function 'func'.
%
%   Furthermore if the qualifier(s) itself is an array, it will be mapped too!
%   This is especially helpfull when used for calls of integration function
%   such as 'qromb' as these require parameters to be given as qualifiers!
%
%   This functionality can be turned off with the "noqualmap"
%   qualifier, which is necessary for, e.g., epochfolding routines (and much
%   faster!)
%
%   NOTE that this function works as 'array_map', i.e., the mapping will apply
%   for the first argument which is an array! As the qualifiers are passed after
%   the arguments, an argument array will supersede a qualifier array as long as
%   they do not have the same length.
%
%\example
%   define t(x){ return qualifier("a",0)*x;};
%
%   %% Mapping function using an argument array:
%     print(array_map_qualifiers( Double_Type, &t, [1,2,3,4] ; a=1 ));
%
%   %% Mapping function using a qualifier array:
%     print(array_map_qualifiers( Double_Type, &t, 1 ; a=[1,2,3,4] ));
%
%   %% Mapping function argument array superseding qualifier array:
%     print(array_map_qualifiers( Double_Type, &t, [3,4,5,6] ; a=[1,2] ));
%     
%   %% Mapping function using argument array AND qualifier array:
%     print(array_map_qualifiers( Double_Type, &t, [3,4,5,6] ; a=[1,2,3,4] ));
%
%   %% Map function using a qualifier array application on 'qromb' integrater
%     print(array_map_qualifiers( Double_Type, &qromb, &t, 0, 1 ; a=[0:2] ));
%
%\seealso{array_map, struct_of_arrays_2_struct_array}
%!%-
{
   variable q  ;
   if (qualifier_exists("noqualmap"))
     {
	q= __qualifiers ();
     }
   else
     {
	q= struct_of_arrays_2_struct_array(__qualifiers ());
     }

   variable args = __pop_list (_NARGS-1);
   variable ret = ();
  
   return array_map (ret, &call_func_with_qualifiers,
		    __push_list (args), q);
}

%
% Code from src/slang/arrays/array_permutation_matrix.sl
%
define array_permutation_matrix()
%!%+
%\function{array_permutation_matrix}
%\synopsis{returns all possible permutations of the given arrays}
%\usage{Array_Type array_permutation_matrix(Array_Type[] or List_Type arrays);}
%\description
%    Calculates the index arrays of all possible permutations of the
%    elements in the given arrays (see the below example). The input
%    has to be either an array of arrays (Array_Type) or a list of
%    arrays (List_Type). The returned matrix (J,K) is an array of
%    integer arrays. The element (J,K) is the index of the K^th input
%    array, which corresponds to the J^th permutation.    
%\example
%    variable arr = Array_Type[3];
%    arr[0] = [10,20,30,40];
%    arr[1] = [50,60];
%    arr[2] = [70,80,90];
%    
%    variable matrix = array_permutation_matrix(arr);
%    % Returns an Array_Type[24] corresponding to 4*2*3 = 24 possible
%    % permutations. Each element is an Integer_Type[3] containing the
%    % indices for each of the 3 input arrays:
%    % matrix[0]  = [0,0,0] -> [10,50,70]
%    % matrix[1]  = [1,0,0] -> [20,50,70]
%    % matrix[2]  = [2,0,0] -> [30,50,70]
%    % matrix[3]  = [3,0,0] -> [40,50,70]
%    % matrix[4]  = [0,1,0] -> [10,60,70]
%    % matrix[5]  = [1,1,0] -> [20,60,70]
%    % ...
%    % matrix[23] = [3,1,2] -> [40,60,90]
%!%-
{
  variable arr;
  switch (_NARGS)
    { case 1: (arr) = (); }
    { help(_function_name); }

  % sanity checks
  if (typeof(arr) == List_Type) {
    arr = list_to_array(arr);
  } else if (typeof(arr) != Array_Type) {
    vmessage("error (%s): input has to be of Array_Type or List_Type", _function_name);
  }
  
  % get array dimensions and number of permutations
  variable dim = array_map(Array_Type, &array_shape, arr);
  variable num = int(outer_product(__push_array(dim))[*][0]); 
  
  % initialize matrix
  variable matrix = Array_Type[num];
  matrix[0] = Integer_Type[length(dim)]; % start with first period indices

  % loop over permutations
  variable n, d;
  _for n (1, num-1, 1) {
    % start with previous permutation
    matrix[n] = @(matrix[n-1]);
    % increment
    d = 0;
    forever {
      matrix[n][d]++;
      if (matrix[n][d] == dim[d][0]) { matrix[n][d] = 0; d++; }
      else { break; }
    }
  }

  return matrix;
}

%
% Code from src/slang/arrays/array_permute.sl
%

% only activate if module rand could be loaded
#ifeval __get_reference("rand_flat")!=NULL

%%%%%%%%%%%%%%%%%%%%
define array_permute()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{array_permute}
%\synopsis{generates a random permutation of [0 : n-1]}
%\usage{Integer_Type[] random_array(Array_Type a)
%\altusage{Integer_Type[] random_array(Integer_Type n)}
%}
%\description
%    The return value is an array of indices.
%    To randomize an array \code{a}, use \code{a[array_permute(a)]}.
%\seealso{array_sort}
%!%-
{
  variable n;
  switch(_NARGS)
  { case 1: n = (); }
  { help(_function_name()); return; }

  if(typeof(n)==Array_Type)  n = length(n);
  variable numbers = { 0 };
  variable i;
  _for i (1, n-1, 1)  list_append(numbers, i);
  variable ind = Integer_Type[n];
  _for i (0, n-1, 1)
    ind[i] = list_pop( numbers, int(rand_flat(0, n-i)) );
  return ind;
}

#endif

%
% Code from src/slang/arrays/array_remove.sl
%
define array_remove()
%!%+
%\function{array_remove}
%\synopsis{removes one element from an array}
%\usage{Array_Type a_ = array_remove(Array_Type a, Integer_Type i);}
%\description
%    \code{a_} contains all elements of \code{a} excecpt of the one indexed by \code{i}.
%!%-
{
  variable a, i;
  switch(_NARGS)
  { case 2: (a, i) = (); }
  { help(_function_name()); return; }

  if(i==0)
    return a[[1:]];
  else
    return [ a[[:i-1]], a[[i+1:]] ];
}

%
% Code from src/slang/arrays/array_unique.sl
%
%%%%%%%%%%%%%%%%%%%
define array_unique()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{array_unique}
%\synopsis{takes elements of an array only once}
%\usage{Array_Type array_unique(Array_Type a)}
%\description
%   Duplicated array elements are removed from the array.
%   This function is basically
%      a[unique(a)]
%   where unique gives the indicies of the unique array
%   elements.
%\seealso{unique}
%!%-
{
  variable a;
  switch(_NARGS)
  { case 1: a = (); }
  { help(_function_name()); return; }

  return a[unique(a)];
}

%
% Code from src/slang/arrays/array_zip.sl
%
%%%%%%%%%%%%%%%%%%%%%
define array_zip()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{array_zip}
%\synopsis{"zip" two or more arrays together}
%\usage{Any_Type array_zip(Any_Type[] a, b, ...);}
%\description
%    Returns an array of tuples (or higher orders), where each tuple i
%    contains the i-th element from each of the given arrays. The
%    returned array is truncated to the length of the shortest
%    array. Thus, the number of tuples is equal to this smallest
%    length.
%\example
%    a = [1:10];
%    b = [11:20];
%    c = array_zip(a,b);
%    % c = [1,11,2,12,3,13,4,14,...,10,20];
%!%-
{
  % put all arrays into an Array_Type[] array
  variable arr = __pop_args(_NARGS);
  arr = array_struct_field(arr, "value");
  % get shortest length among all arrays
  variable l = min(array_map(Integer_Type, &length, arr));
  % defined index array
  variable idx = [:l-1];
  % truncate and merge all arrays
  variable merge = (_typeof(arr[0]))[_NARGS*l];
  variable i;
  _for i (0, _NARGS-1, 1) {
    % indices where to put the current array into the merged one
    idx = [i::_NARGS];
    % add array to merge
    merge[idx] = arr[i];
  }
  return merge;
}

%
% Code from src/data/pulsar/arrivaltimes.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL
% -*- mode: slang; mode: fold -*-


% There are also plot function which uses the
% xfig module to create eps-files


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_det()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_det}
%\synopsis{Determines the arrival times from a lightcurve using a pulse pattern}
%\usage{Struct_Type atime_det(Struct_Type lc, Struct_Type[] pattern, Double_Type t0, Struct_Type ephemeris[, Struct_Type orbit]);
% or Struct_Type atime_det(String_Type lc, Struct_Type[] pattern, Double_Type t0, Double_Type[] period[, Struct_Type orbit]);}
%\qualifiers{
%    \qualifier{time}{name of the time field in the FITS-file,
%             see fits_read_lc for details}
%    \qualifier{indiv}{determine individual pulses. If no value
%             is assigned, all arrival times are re-
%             turned. Otherwise a given integer sets
%             the number of pulses to average}
%    \qualifier{movem}{if 'indiv' is greater one, so it is averaged
%             over a number of arrival times, a
%             moving mean is used to get all individual
%             pulses. Note that the pulses are then not
%             statistically independent!}
%    \qualifier{ccfint}{interpolates the cross correlation of the
%             pulse pattern and the actual analyzed
%             pulse to increase the accuracy. Only works
%             well for clear signals! The number of
%             interpolated bins between each original
%             bin is set to the assigned number
%             (default = 4)}
%    \qualifier{varerr}{estimates the error by a local standard
%             deviation. Therefore the variation of the
%             given number of arrival times is used
%             (default = 10)}
%    \qualifier{mcerr}{estimates the error of each arrival time
%             by performing Monte Carlo simulations. If
%             an integer is assigned it sets the number
%             of runs (default = 10000). Has a higher
%             priority than 'varerr'}
%    \qualifier{mcgaus}{fit a gaussian to the Monte Carlo distri-
%             bution, if used for error estimation}
%    \qualifier{getpat}{if multiple patterns are given, the index
%             of the used one is stored in this variable,
%             which has to be given as a reference (@var)}
%    \qualifier{match}{reference to a variable (&var) where the
%             matching pulses are saved as structures
%             similar to the reference profile}
%    \qualifier{ccflim}{matches below the given cross-correlation
%             values are skipped (default = 0), allowed
%             range is -1 to 1.}
%    \qualifier{chatty}{boolean value for output messages
%             (default = 1). If set to 2, also echo
%             result of Monte Carlo error estimation}
%    \qualifier{debug}{plot the cross correlation and the matching
%             pulse, echo the found phase shift and
%             sleep the given seconds (default = 1) or,
%             if set to 'user' wait until a key is
%             pressed}
%}
%\description
%    Using one or more pulse pattern the arrival times
%    of pulses in a lightcurve are determined using
%    phase connection. The pattern must be given as a
%    structure with the fields
%      Double_Type bin_lo
%      Double Type bin_hi
%      Double_Type value
%      Double_Type error,
%    as, for example, returned by the 'epfold' function.
%    The lightcurve can be passed by a structure with
%    the fields 'time', 'rate', 'error' and 'fracexp',
%    or by the filename to a FITS-file.
%    To transform the phase shift to time, the pulse
%    period at the actual position in the lightcurve is
%    needed. For a first guess this may be a constant
%    value (in seconds) or an array of two elements
%    defining a range of periods (min/max values, in
%    seconds). In the latter case the pulse period is
%    searched by using the 'getVarPeriod' function and
%    additional qualifiers are passed. However, the
%    correct way is to pass the pulse ephemeris and
%    additional orbital parameters to calculate the
%    pulse poriod more precisely. The corresponding
%    structures are described in the 'check_pulseperiod_orbit_struct'
%    function. The values may be found by a fit to the
%    arrival times using a constant pulse period. This
%    leads to slightly different results, hence the
%    determination of the arrival times and their
%    analysis is an iterative procedure.
%    If the 'indiv' qualifier is omitted the given
%    lightcurve is folded and the returned arrival
%    time corresponds to the phase shift of the given
%    pattern to the resulting profile.
%    If the method for estimating the uncertainty of
%    the arrival times is not specified by qualifiers,
%    the default uncertainty is set to one phase bin.
%    The number of bins are derived from the given
%    pulse pattern.
%    The returned structure has the following fields:
%    arrtimes - array of determined arrival times (MJD)
%    error    - their uncertainties (days)
%    arrnum   - relative pulse number of each arrival
%               time to a specific pulse. Dramatically
%               increases the speed of a fit.
%    numref   - index of the reference pulse, should be
%               zero. Is updated if arrival times are
%               merged.
%    reft0    - the given reference time t0
%\seealso{atime_merge, save_atime, define_atime, arrtimes, pfold}
%!%-
{
 variable lcF, refprof, reft0, per, orb = NULL;
 switch(_NARGS)
   { case 4: (lcF,refprof,reft0,per) = (); }
   { case 5: (lcF,refprof,reft0,per,orb) = (); }
   { help(_function_name()); return; }

 variable t0found = 0;
 variable timefield = qualifier("time","time");
 variable chatty = qualifier("chatty",1);
 if (chatty == NULL) chatty = 1;
 variable indiv = qualifier("indiv", 1);
 if (indiv == NULL) indiv = 1;
 indiv = [qualifier_exists("indiv"), indiv, qualifier_exists("movem") && indiv > 1];
 variable ccflim = qualifier("ccflim", 0.);
 ifnot (-1 <= ccflim <= 1) { vmessage("warning (%s): given ccf limit is outside the allowed range, using default value", _function_name); ccflim = 0.; }
 variable ccfint = qualifier("ccfint", 1);
 if (ccfint == NULL) ccfint = 10;
 variable mcerr = qualifier("mcerr", 0);
 if (mcerr == NULL) mcerr = 10000;
 variable varerr = qualifier("varerr", 0);
 if (varerr == NULL) varerr = 10;
 if (varerr != 0 && varerr < 2) { vmessage("warning (%s): need at least 2 datapoints to determine error by local variance", _function_name); varerr = 2; }
 if (varerr > 0 && mcerr > 0) varerr = 0;
 variable matchret = qualifier("match", NULL);
 if (typeof(matchret) == Ref_Type) % initialise return structure for matches
   @matchret = Struct_Type[0];
 else ifnot (typeof(matchret) == Null_Type) { vmessage("warning (%s): 'match' has to be of Ref_Type", _function_name); matchret = NULL; }
 variable getpat = qualifier("getpat", NULL), usedpat;
 ifnot (typeof(getpat) == Ref_Type or typeof(getpat) == Null_Type) { vmessage("warning (%s): 'getpat' has to be of Ref_Type", _function_name); getpat = NULL; }
 
 variable debug = qualifier("debug", -1);
 if (debug == NULL) debug = 1;
  
 % load lightcurve if not already given
 variable lc, lco;
 if (typeof(lcF) == String_Type)
 {
   lc = fits_read_lc(lcF;time=timefield);
   % convert fracexp to totalexp
   lc.fracexp = fits_lc_exposure(lcF)*lc.fracexp/86400.;
 } else lc = @lcF; % lc already given
 lco = @lc;

 % sort lc by time
 lc = sort_struct_arrays(lc, "time");

 % check if the period has to be estimated by Epoch Folding
 if (typeof(per) == Array_Type and length(per) == 2)
 {
   if (chatty) message("Determining pulse period of lightcurve");  
   per = getVarPeriod(lc, per[0]/86400, per[1]/86400;; struct_combine(__qualifiers, struct { dt=lc.fracexp }));
   if (chatty) vmessage("  p_lc  = %f s", per*86400);
   per = check_pulseperiod_orbit_struct(per*86400);
 }
 else if (typeof(per) == Double_Type) { per = struct { p0 = per, t0 = reft0 }; }
 else if (typeof(per) == Struct_Type && check_pulseperiod_orbit_struct(per) == 0) {
   vmessage("error (%s): pulse period structure not consistent with its definition", _function_name);
   return;
 }
 else { vmessage("error (%s): pulse period must be a single double type, a two element array or a structure", _function_name); }
 ifnot (orb == NULL || check_pulseperiod_orbit_struct(orb)) { vmessage("error (%s): orbit structure not given properly", _function_name); return; }
   
 % renorm reference profile(s)
 variable p;
 _for p (0, length(refprof)-1, 1)
 {
   refprof[p].error = refprof[p].error/(max(refprof[p].value)-min(refprof[p].value));
   refprof[p].value = (refprof[p].value-min(refprof[p].value))/(max(refprof[p].value)-min(refprof[p].value));
 }

 variable nums;
 variable nbins = length(refprof[0].value);
 if (indiv[0]) % prepare section to find individual pulses
 {
   % number of pulses in lc to look for
   nums = ceil((max(lc.time)-min(lc.time)) / per.p0 * 86400);
   ifnot (indiv[2]) nums = nums / indiv[1]; % reduced number by no moving mean
   nums = int(nums);
   % rebin lightcurve similar to profile if ALL pulses have to be found
   if (indiv[1] == 1) lc = rebin_lc(lc, per.p0 / 86400 / nbins);
 }
 else nums = 1;
 % loop over number of pulses (except if using profile only)
 % to get the pulse arrival times
 if (chatty) message("Getting pulse arrival times");
 variable i, f, t0, ccf, dphi, ddphi, ndx, j=0, lasttime = NULL, lastnum, acttime, actnum, ir = NULL;
 variable atimes = Double_Type[0], datimes = Double_Type[0], npuls = Integer_Type[0], refpuls = Integer_Type[0];
 _for i (0, nums-1, 1)
 {
   % determine arrival times of (each) individual pulses
   if (indiv[0])
   {
     % find actual pulse part of lc (attend moving mean, indiv[2]==1)
     if (indiv[2]) ndx = where(lc.time[0] + per.p0/86400*i <= lc.time < lc.time[0] + per.p0/86400*(i+indiv[1]));
     else ndx = where(lc.time[0] + per.p0/86400*i*indiv[1] <= lc.time < lc.time[0] + per.p0/86400*(i+1)*indiv[1]);
     % only use if number of bins are correct and ALL pulses have to be found
     if (length(ndx) == nbins && indiv[1] == 1)
     {
       t0 = lc.time[ndx[0]];
       % extract part
       f = struct { value = lc.rate[ndx], error = lc.error[ndx], bin_lo = (lc.time[ndx]-lc.time[ndx[0]]) * 86400 / per.p0, bin_hi = NULL };
       f.bin_hi = make_hi_grid(f.bin_lo);
       j++;
     }
     % average over given number of pulses by folding the lightcurve
     else if (length(ndx) >= nbins && indiv[1] > 1)
     {
       t0 = lc.time[ndx[0]];
       f = pfold(lc.time[ndx], lc.rate[ndx], per.p0/86400, lc.error[ndx];; struct_combine(__qualifiers, struct { nbins = nbins, dt=lc.fracexp[ndx], t0=t0 }));
       if (sum(isnan(f.value)) > 0) f = NULL;
       else j++;
     }
     else f = NULL;
   }
   % arrival time of the pulse profile
   else
   {
     if (chatty) message("  -> pulse profile");
     t0 = lc.time[0];
     % create profile
     f = pfold(lc.time, lc.rate, per.p0/86400, lc.error;; struct_combine(__qualifiers, struct { nbins = nbins, dt=lc.fracexp, t0=t0 }));
     ndx = [0,length(lc.time)-1];
   }

   ifnot (f == NULL)
   {
     % renorm profile / lc part
     f.error = f.error/(max(f.value)-min(f.value));
     f.value = (f.value-min(f.value))/(max(f.value)-min(f.value));
     % loop over all given pattern
     variable ccfmax = -2.;
     _for p (0, length(refprof)-1, 1)
     {
       variable tempccf = CCF_1d(refprof[p].value, f.value);
       % assign cross-correlation in case of a better match
       if (max(tempccf) > ccfmax)
       {
	 ccf = @tempccf;
	 ccfmax = max(tempccf);
	 usedpat = p;
         if (typeof(getpat) == Ref_Type) @getpat = usedpat;
       }
     }
     if (chatty == 2)
     {
       vmessage("  -> cross-correlation is %.3f", max(ccf));
       if (length(refprof) > 1) vmessage("     with pulse pattern no. %d", usedpat);
     }
     % check on ccf limit
     if (max(ccf) < ccflim)
     {
       ccf[0] = _NaN;
       message("  -> below the given limit and SKIPPED");
     }
     ifnot (isnan(ccf[0]))
     {
       % calculate the actual pulse period
       variable pact = pulseperiod(t0, per, orb);

       % get reference pulse number, if not set
       % (is needed if pulse number 0 had less nbins than required)
       if (ir == NULL) ir = i;

       %%% PULSE ARRIVAL TIME DETERMINATION %%%
       variable tstart = lc.time[ndx[0]], tstop = lc.time[ndx[-1]];
       ndx = 0;
       % periodic borders
       ccf = [ccf, ccf[0]];
       if (qualifier_exists("debug"))
       {
	 xrange; xlabel("Phase Shift"); title("Result of Cross Correlation");
	 yrange(-1,1); ylabel("Correlation");
	 connect_points(0);
	 point_style(4);
	 color(1);
	 plot(1.*[0:nbins]/nbins, ccf);
       }
#ifexists gsl->interp_cspline
       if (ccfint > 1)% && t0 != reft0) % interpolate the ccf result to get the position more properly
       {
	 % interpolate
         ccf = gsl->interp_cspline_periodic([0:length(ccf)-1:1./ccfint], [0:length(ccf)-1], ccf);
	 if (qualifier_exists("debug"))
	 {
	   color(4); point_style(-1);
	   connect_points(1);
	   oplot(1.*[0:length(ccf)-1] / (length(ccf) - 1), ccf);
	 }
       }
#endif
       % calculate phase shift
       dphi  = 1.*where_max(ccf) / (length(ccf)-1);
       if (length(dphi) > 1) dphi = min(dphi);
       if (dphi == 1.) dphi = 0.;
       if (qualifier_exists("debug"))
       {
	 vmessage("  phase shift: %.3f", dphi);
	 color(2); connect_points(1); point_style(-1);
	 oplot([dphi,dphi],[-1,1]);
	 if (typeof(debug) == String_Type && debug == "user") ()=keyinput(;silent, nchr=1);
	 else sleep(debug);
	 color(1); xlabel("Pulse Phase"); xrange;
	 ylabel("Normalized Count Rate"); title("Aligned Pulse and Reference");
	 yrange(min([f.value-f.error,refprof[usedpat].value-refprof[usedpat].error]),
		max([f.value+f.error,refprof[usedpat].value+refprof[usedpat].error]));
         hplot_with_err(f.bin_lo, f.bin_hi, shift(f.value, -int(dphi*nbins)), shift(f.error, -int(dphi*nbins)));
	 ohplot_with_err(refprof[usedpat]);
	 if (typeof(debug) == String_Type && debug == "user") ()=keyinput(;silent, nchr=1);
	 else sleep(debug);
       }
       % set relative pulse number
       if (indiv[2]) actnum = i-ir;
       else actnum = (i-ir) * indiv[1];
       % add match to returning structure array (if set)
       if (typeof(matchret) == Ref_Type) 
         @matchret = [@matchret, struct { bin_lo = f.bin_lo, bin_hi = f.bin_hi, value = shift(f.value, -int(dphi*nbins)),
	                                  error = shift(f.error, -int(dphi*nbins)), ccf = max(ccf), tstart = tstart, tstop = tstop }];
       % transform phase shift into arrival time
       acttime = t0 + (1.-dphi)*pact/86400;
       % check if arrival times does not match the pulse number
       ifnot (lasttime == NULL)
       {
	 variable is = (acttime-lasttime)*86400/pact;
	 variable should = actnum-lastnum;
       	 if (abs(is - should) > qualifier("phidif", .2))
       	 {
       	   if (qualifier_exists("debug")) vmessage("  phidif is %.3f, should be %.3f -> shifting time", is, should);
       	   acttime = acttime - round(is-should) * pact/86400;
       	 }
	 else if (qualifier_exists("debug")) vmessage("  phidif is %.3f, should be %.3f", (acttime-lasttime)*86400/pact, actnum-lastnum);
       }
       lasttime = acttime;
       lastnum = actnum;
       atimes = [atimes, acttime];
       npuls = [npuls, actnum];
       refpuls = [refpuls, 0];

       %%% ERROR ESTIMATION %%%
       ddphi = NULL;
#ifexists gsl->interp_cspline
       if (mcerr > 1 && ccfint > 1) % estimate error using Monte Carlo techniques
       {
	 variable mc, lo, hi, dphis = Double_Type[mcerr], mphi, lphi = NULL;
	 variable fm = Double_Type[length(f.value)];
	 variable refm = Double_Type[length(refprof[usedpat].value)];
	 _for mc (0, mcerr-1, 1) % MC loop
	 {
	   % random noisy profile
	   fm = f.value + grand(length(fm)) * f.error;
	   refm = refprof[usedpat].value + grand(length(refm)) * refprof[usedpat].error;
           % cross correlate
           ccf = CCF_1d(refm,fm);
	   % periodic borders
	   ccf = [ccf, ccf[0]];
	   ccf = gsl->interp_cspline_periodic([0:length(ccf)-1:1./ccfint], [0:length(ccf)-1], ccf);
	   % get shift
	   mphi = 1.*where_max(ccf) / (length(ccf)-1);
           if (length(mphi) > 1) mphi = min(mphi);
           if (mphi == 1.) mphi = 0.;
	   % check on phase shifts greater 1
%	   ifnot (lphi == NULL) if (abs(lphi - mphi) > 2.*qualifier("phidif", .2)) mphi = mphi + 1.*sign(lphi-mphi);
	   lphi = mphi;
	   dphis[mc] = mphi;
	 }
	 % do a gaussian fit to the distribution
	 if (qualifier_exists("mcgaus"))
	 {
	   variable x_lo = [0.:2.-1./sqrt(mcerr):1./sqrt(mcerr)];
           variable x_hi = x_lo + x_lo[1];
           variable h = histogram(dphis, x_lo, x_hi);
   	   ddphi = array_fit_gauss(x_lo, h,, x_lo[where_max(h)[0]], moment(dphis).sdev, 1000; frz=[0,0,0,1]).sigma;
	   ddphi = ddphi * pact / 86400;
	 }
	 else ddphi = moment(dphis).sdev * pact / 86400;

	 if (chatty == 2) vmessage("  -> estimated error = %f phases", ddphi/pact*86400);
	 if (qualifier_exists("debug") && qualifier_exists("mcgaus"))
	 {
	   xrange(0,2); yrange(0,1.1*max(h)/sum(h));
	   xlabel("Phase Shift"); ylabel("Number Density");
	   color(1);
	   hplot(x_lo, x_hi, h/sum(h));
           if (typeof(debug) == String_Type && debug == "user") ()=keyinput(;silent, nchr=1);
	   else sleep(debug);
	 }
	 
         datimes = [datimes,ddphi];
       }
       else if (varerr > 0)
#else
       if (varerr > 0) % estimate error by a local mean
#endif
       {
	 % calculate local variance
	 ndx = [((length(atimes) - 1) / varerr) * varerr: length(atimes) - 1]; % indices of last #varerr-1 number of datapoints
	 ddphi = moment((atimes[ndx] - atimes[ndx-1]) / (npuls[ndx] - npuls[ndx-1])).sdev;
	 if (length(ndx) == varerr)
	 {
           if (chatty == 2) vmessage("  -> estimated error = %f phases", ddphi/pact*86400);
           datimes = [datimes, ddphi * ones(length(ndx))];
	 }
       }
       else datimes = [datimes, pact/86400/nbins]; % if no error was estimated: error = 1 phase bin
     }
   }
 }
 % set remaining errors for local variance
 if (varerr > 0 && length(datimes) < length(atimes))
 {
   if (length(ndx) == 1 || ddphi == 0.0) ddphi = pact/86400/nbins;
   else if (length(ndx) < varerr/2 && length(datimes) > 0) ddphi = datimes[length(datimes)-1];
   datimes = [datimes, ddphi * ones(length(atimes)-length(datimes))];
   if (chatty == 2) vmessage("  -> estimated error = %f phases", ddphi/pact*86400);
 }
 if (indiv[0] and chatty) vmessage("  -> individual pulses (%d/%d)", j, int(nums));

 return struct { arrtimes = atimes, error = datimes, arrnum = npuls, numref = refpuls, reft0 = reft0 };
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_merge()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_merge}
%\synopsis{merges two or more structures of arrival times into one}
%\usage{Struct_Type atime_merge(Struct_Type atime1, Struct_Type atime2);
% or Struct_Type atime_merge(Struct_Type[] atime);}
%\qualifiers{
%    \qualifier{nosort}{do not sort the merged arrival times}
%}
%\description
%    Either both given structures or an array of
%    structures containing arrival times are merged.
%    The fields of the structures must be equal to
%    the ones described in 'atime_det'.
%    The relative pulse numbers and their references
%    are updated to the indices of the new array of
%    arrival times. In addition the arrival times are
%    sorted by time unless the 'nosort' qualifier is
%    given.
%\seealso{atime_det}
%!%-
{
 variable a1, a2 = NULL;
 switch(_NARGS)
   { case 1: (a1) = (); }
   { case 2: (a1,a2) = (); }
   { help(_function_name()); return; }
 if (__is_initialized(&a1)) a1 = @a1;
 ifnot (a2 == NULL) a2 = @a2; 

 % check on correct parameters
 if (a2 == NULL && _typeof(a1) != Struct_Type)
   throw RunTimeError, sprintf("error (%s): need an array of structures", _function_name);
 if (a2 != NULL && (length(a1)+length(a2) > 2 || (typeof(a1) != Struct_Type && _typeof(a2) != Struct_Type)))
   throw RunTimeError, sprintf("error (%s): need two single structures", _function_name);
 
 % merge structures
 variable merg, i, reft0 = NULL;
 if (a2 == NULL) % array given
 {
   if (length(a1) > 1)
   {
     _for i (1, length(a1)-1, 1)
     {
       a1[i].numref += length(a1[i-1].numref); % shift reference numbers
       if (reft0 == NULL) reft0 = a1[i-1].reft0;
       else ifnot (reft0 == a1[i-1].reft0) vmessage("warning (%s): reference times of matching profile are not equal", _function_name);
     }
     merg = merge_struct_arrays(a1);
     merg.reft0 = reft0;
   }
   else merg = a1;
 }
 else % two structures given
 {
   ifnot (__is_initialized(&a1)) merg = a2;
   else ifnot (__is_initialized(&a2)) merg = a1;
   else
   {
     % a2.arrtimes = a2.arrtimes - shift; % correct arrival times
     a2.numref += length(a1.numref); % shift reference numbers
     if (a1.reft0 == a2.reft0) reft0 = a1.reft0;
     else vmessage("warning (%s): reference times of matching profile are not equal", _function_name);
     merg = merge_struct_arrays([a1,a2]);
     merg.reft0 = reft0;
   }
 }
 
 % sort by time
 ifnot (qualifier_exists("nosort"))
 {
   variable ndx = array_sort(merg.arrtimes), nr = @(merg.numref);
   merg.arrtimes = merg.arrtimes[ndx];
   merg.error = merg.error[ndx];
   merg.arrnum = merg.arrnum[ndx];
   _for i (0, length(nr)-1, 1) merg.numref[i] = wherefirst(ndx == nr[ndx[i]]); % update reference numbers
 }

 return merg;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define save_atime()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{save_atime}
%\synopsis{saves a structure of arrival times into a FITS-file}
%\usage{save_atime(String_Type filename, Struct_Type atime, String_Type extname[, String_Type[] comments]);}
%\qualifiers{
%    \qualifier{obj}{observed object, written into FITS header}
%    \qualifier{sat}{used satellite, written into extension header}
%    \qualifier{nfold}{number of arrival times, which were merged
%              during the determination. Corresponds to the
%              'indiv' qualifier for the 'atime_det' function.
%              Written into extension header}
%    \qualifier{hfits}{structure for additional FITS-header fields}
%    \qualifier{hext}{structure for additional extension-header fields}
%    \qualifier{newfile}{if the given filename exists a new file is
%              created instead of updating the existing one}
%    \qualifier{newext}{if the given extension already exists in the
%              FITS-file a further one with the same name is
%              added instead of updating the existing one}
%}
%\description
%    A structure of arrival times, e.g. as returned by the
%    'atime_det' function, is save into the given FITS-file
%    creating or updating the extension of the geiven name.
%    Most of the qualifiers allow to store addiotional
%    informations into the header of the FITS-file or the
%    extension. The optional fourth parameter can be used
%    to write additional comments into the FITS-file.
%\seealso{atime_det, atime_merge, load_atime}
%!%-
{
 variable file, atimes, ext, comm;
 switch(_NARGS)
   { case 3: (file,atimes,ext) = (); comm = NULL; }
   { case 4: (file,atimes,ext,comm) = (); }
   { help(_function_name()); return; }
 variable object = qualifier("obj", NULL);
 variable sat = qualifier("sat", NULL);
 variable nfold = qualifier("nfold", NULL);
 variable hfits = qualifier("hfits", NULL);
 variable htabl = qualifier("hext", NULL);
 variable doup = not qualifier_exists("newfile");
 variable doov = not qualifier_exists("newext");

 % store pulse index of matching profile (dphi != 0)
 variable reft0 = atimes.reft0;
 
 % create fits file, if it does not exist
 variable fp, i;
 if (access(file, F_OK)==0 && doup) fp = fits_open_file(file,"w");
 else { fp = fits_open_file(file,"c"); doup = 0; }
  
 % fits header
 ifnot (doup)
 {
   fits_create_image_hdu(fp, NULL, 0, 0);
   if (object != NULL) fits_update_key(fp, "OBJECT", object, "observed OBJECT");
   % additional fits header informations
   if (typeof(hfits) == Struct_Type)
     _for i (0, length(get_struct_field_names(hfits))-1, 1)
       fits_update_key(fp, get_struct_field_names(hfits)[i], get_struct_field(hfits, get_struct_field_names(hfits)[i]));
 }
  
 % eventually delete existing table of same extension
 if (doov && _fits_movnam_hdu(fp, -1, ext, 0)==0) ()=_fits_delete_hdu(fp);  
 % create table of arrival times
 fits_write_binary_table(fp, ext, reduce_struct(atimes,"reft0"));
 % table header
 if (sat != NULL) fits_update_key(fp, "SATELLIT", sat, "used SATELLITE(S)");
 fits_update_key(fp, "TUNIT1", "MJD", "physical unit of field 1");
 fits_update_key(fp, "TUNIT2", "MJD", "physical unit of field 2");
 fits_update_key(fp, "TUNIT3", "rel. pulse number", "physical unit of field 3");
 fits_update_key(fp, "TUNIT4", "reference index", "physical unit of field 4");
 fits_update_key(fp, "REFT0", reft0, "reference time of matching profile");
 if (nfold != NULL) fits_update_key(fp, "NFOLD", nfold, "number of pulses used for folding");
 % additional table header informations
 variable htn, htk;
 if (typeof(htabl) == Struct_Type)
 {
   htn = get_struct_field_names(htabl);
   _for i (0, length(htn)-1, 1)
   {
     htk = get_struct_field(htabl, htn[i]);
     if (length(htk)==1) fits_update_key(fp, htn[i], htk);
     else fits_update_key(fp, htn[i], htk[0], htk[1]);
   }
 }
 % comments
 if (typeof(comm) != Array_Type) comm = [comm];
 if (_typeof(comm) == String_Type)
   _for i (0, length(comm)-1, 1) fits_write_comment(fp, comm[i]);
 fits_close_file(fp);
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define load_atime()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{load_atime}
%\synopsis{loads a structure containing arrival times from a FITS-file}
%\usage{Struct_Type load_atime(String_Type filename, String_Type extname);}
%\description
%    The arrival times, which were stored previously by
%    'save_atime', are loaded and returned as a structure.
%    This structure has the fields described in the
%    'atime_det' function. The extension table containing
%    the arrival times must be specified as well as the
%    FITS-file itself.
%\seealso{save_atime, atime_det}
%!%-
{
 variable file, ext;
 switch(_NARGS)
   { case 2: (file,ext) = (); }
   { help(_function_name()); return; }

 variable fp = fits_open_file(file, "r");
 variable atimes, reft0;
 % move to extension and load table
 if (_fits_movnam_hdu(fp, _FITS_BINARY_TBL, ext, 0)==0) atimes = fits_read_table(fp); else atimes = NULL;
 ifnot (atimes == NULL)
 {
   reft0 = fits_read_key(fp, "REFT0");
   atimes = struct_combine(atimes, struct { reft0 = reft0 });
 }
 else vmessage("error (%s): extension '%s' not found in binary table", _function_name, ext);

 fits_close_file(fp);
 return atimes;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_metavalid(ind)
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_metavalid}
%\synopsis{checks if a given dataset of arrival times contains valid metadata}
%\usage{Integer_Type[] atime_metavalid(Integer_Type index);}
%\description
%    If a dataset of arrival times is defined using the
%    'define_atime' function, there are also a lot of
%    addiotional needed informations defined in the
%    metadata. These are used by fitting the data by the
%    'arrtimes' fit function. This function can be used
%    to find all defined data containing arrival times,
%    which is implemented in 'atime_dataind'.
%\seealso{define_atime, arrtimes, get_dataset_metadata, atime_dataind}
%!%-
{
  ifnot (typeof(ind) == Array_Type) ind = [ind];
  variable i, ok = 0;
  foreach i (ind)
  {
    variable metadata = struct_copy(get_dataset_metadata(i));
    if (typeof(metadata) == Struct_Type)
    {
      if (
           struct_field_exists(metadata, "datapnum")
        && struct_field_exists(metadata, "refpnum")
        && struct_field_exists(metadata, "reft0")
        && struct_field_exists(metadata, "modelpnum")
        && struct_field_exists(metadata, "notice")
        && struct_field_exists(metadata, "id")
        && struct_field_exists(metadata, "useref")
        && struct_field_exists(metadata, "shiftref")
        && struct_field_exists(metadata, "freeparams")
      ) ok += 1;
    }
  }
  if (length(ind) == ok) return 1;
  else return 0;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_dataind()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_dataind}
%\synopsis{returns the datasets, which contain arrival times}
%\usage{Integer_Type[] atime_dataind();}
%\description
%    Loops over all defined datasets and checks, which
%    are containing arrival times. Therefore the function
%    'atime_metavalid' is used. If no dataset is found,
%    the function returns NULL.
%\seealso{atime_metavalid, define_atime}
%!%-
{
 variable i, ind = Integer_Type[0];

 foreach i (all_data)
   if (atime_metavalid(i)) ind = [ind, i];
 
 return length(ind) == 0 ? NULL : ind;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define xnotice_atime()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{xnotice_atime}
%\synopsis{notices a range of defined arrival times}
%\usage{xnotice_atime(Integer_Type index[, Double_Type time_lo, Double_Type time_hi]);
% or xnotice_atime(Integer_Type index, Double_Type time);
% or xnotice_atime(Double_Type time);}
%\description
%    There are three different ways to notice the time
%    range of arrival times defined in a dataset:
%    1. Like 'xnotice' set the time range of dataset
%       'index' explicitly from 'time_lo' to 'time_hi'.
%    2. Find the border of the time range of dataset
%       'index' nearest to the given time. Then set
%       the found border to this time. The border of
%       a possible adjacent dataset is changed respect-
%       ively.
%    3. Find that dataset, which time range has the
%       nearest border to the given time and proceed
%       like option two.
%    If the dataset only is given, then the full time
%    range is used.
%\seealso{define_atime, xnotice}
%!%-
{
 variable ind = NULL, cut = NULL, mn = NULL, mx;
 switch(_NARGS)
   { case 1: (cut) = (); }
   { case 2: (ind,cut) = (); }
   { case 3: (ind,mn,mx) = (); }
   { help(_function_name()); return; }

 variable meta, dc, nind = atime_dataind;
 if (typeof(nind) == Null_Type) { vmessage("warning (%s): no datasets containing arrival times defined", _function_name); return; }
 % check if the full range of a specific dataset should be used
 if (typeof(cut) == Integer_Type && ind == NULL)
 {
   ifnot (wherefirst(nind == cut) == NULL) % dataset found
   {
     dc = get_data_counts(cut);
     meta = struct_copy(get_dataset_metadata(cut));
     meta.notice = [dc.value[0],dc.value[length(dc.value)-1]];
     set_dataset_metadata(cut, meta);
     xnotice(cut, dc.bin_lo[0], dc.bin_hi[length(dc.bin_hi)-1]);
     return;
   }
   else { vmessage("error (%s): dataset %d not found or does not contain arrival times", _function_name, cut); return; }
 }
 % if only one dataset exists skip ind parameter (for option 2)
 if (length(nind) == 1 && ind != NULL && cut != NULL) { mn = ind; mx = cut; ind = nind[0]; }
 % cut adjacent segments at given time (only 'cut' given)
 % or cut given segment and adjacent one
 if (ind == NULL || (ind != NULL && mn == NULL))
 {
   variable noti, dif, ndx, i;
   dif = Double_Type[length(nind),2]; % [distance to given time, left/right border]
   _for i (0, length(nind)-1, 1)
   {
     noti = struct_copy(get_dataset_metadata(nind[i])).notice;
     dif[i,*] = [min(abs(noti - cut)), where_min(abs(noti - cut))];
   }
  
   if (ind == NULL) % find the two datasets closest to given time
   {   
     if (length(nind) > 1) ndx = array_sort(dif[*,0])[[0,1]]; % select two closests one
     else ndx = 0; % only one dataset given
     dif = dif[ndx,*];
     nind = nind[ndx];
   }
   else % find adjacent dateset to given one
   {
     ndx = array_sort(dif[*,0]);
     if (length(ndx)>2) { dif[1,*] = dif[ndx[2],*]; nind[1] = nind[ndx[2]]; } % select third closest one (-> adjacent dataset)
     else nind = Integer_Type[1];
     % select border of given dataset
     dif[0,*] = [0.0, where_max(abs(noti - cut))];
     nind[0] = ind;
     % check if given time is inside interval of given dataset
     noti = struct_copy(get_dataset_metadata(ind)).notice;
     ifnot (noti[0] < cut < noti[1]) { vmessage("warning (%s): given time outside of dataset", _function_name); return; }
   }
   % set new borders and notice
   % first check if given time is enclosed by dataset (left border smaller cut)
   meta = struct_copy(get_dataset_metadata(nind[0]));
   if (meta.notice[0] < cut)
   {
     meta.notice[int(dif[0,1])] = double(cut);
     xnotice(nind[0], meta.notice[0], meta.notice[1]);
     set_dataset_metadata(nind[0], meta);
     dc = get_data_counts(nind[0]);
     ndx = where(meta.notice[0] <= dc.value <= meta.notice[1]);
     xnotice(nind[0], dc.bin_lo[ndx[0]], dc.bin_hi[ndx[length(ndx)-1]]);
   }
   if (length(nind) > 1)
   {
     meta = struct_copy(get_dataset_metadata(nind[1]));
     meta.notice[int(dif[1,1])] = double(cut);
     set_dataset_metadata(nind[1], meta);
     dc = get_data_counts(nind[1]);
     ndx = where(meta.notice[0] <= dc.value <= meta.notice[1]);
     if (length(ndx) > 0) xnotice(nind[1], dc.bin_lo[ndx[0]], dc.bin_hi[ndx[length(ndx)-1]]);
     else xnotice(nind[1], 0, 0);
   }
 }
  
 % notice by given borders
 if (mn != NULL)
 {
   if (typeof(ind) != Array_Type) ind = [ind];
   _for i (0, length(ind)-1, 1)
   {
     meta = struct_copy(get_dataset_metadata(ind[i]));
     meta.notice = double([mn, mx]);
     set_dataset_metadata(ind[i], meta);
     dc = get_data_counts(ind[i]);
     ndx = where(mn <= dc.value <= mx);
     if (length(ndx) > 0) xnotice(ind[i], dc.bin_lo[ndx[0]], dc.bin_hi[ndx[length(ndx)-1]]);
     else xnotice(nind[i], 0, 0);
   }
 }
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_useref()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_useref}
%\synopsis{switches the use of a reference pulse number}
%\usage{atime_useref(Integer_Type[] dataid[, Integer_Type boolean]);}
%\description
%    While fitting pulse arrival times the pulse number
%    of each pulse must be determined. By using the
%    atime_det function to determine the arrival times
%    each pulse gets a pulse number relative to the
%    first pulse found in the input lightcurve. Once
%    the pulse number of this reference pulse is deter-
%    mined during a fit, the numbers of all other pulses
%    are also set. Using this function the use of rela-
%    tive pulse numbers during a fit can be turned off
%    (boolean=0) or on (boolean=1, default).
%\seealso{arrtimes, atime_det, atime_shiftref}
%!%-
{
 variable i, id, st, meta;
 switch(_NARGS)
   { case 1: (id) = (); st = NULL; }
   { case 2: (id,st) = (); }
   { help(_function_name()); return; }

 if (typeof(id) != Array_Type) id = [id];
 % loop over ids
 _for i (0, length(id)-1, 1)
 {
   if (atime_metavalid(id[i]))
   {
     meta = struct_copy(get_dataset_metadata(id[i]));
     if (st == NULL) meta.useref = 0;
     else meta.useref = (st == 1);
     set_dataset_metadata(id[i], meta);
   }
 }
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_shiftref()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_shiftref}
%\synopsis{switches the shift of the reference pulse number}
%\usage{atime_shiftref(Integer_Type[] dataid[, Integer_Type boolean]);}
%\description
%    While fitting pulse arrival times a reference pulse
%    number may be used. The reference pulse is the
%    first one of the dataset, but while fitting it is
%    by default shifted to the pulse nearest to the
%    fixed reference time 'tpuls0' of the pulse ephe-
%    meris. With this function this shift can be turned
%    off (boolean=0) or on (boolean=1, default).
%\seealso{arrtimes, atime_useref, atime_det}
%!%-
{
 variable i, id, st, meta;
 switch(_NARGS)
   { case 1: (id) = (); st = NULL; }
   { case 2: (id,st) = (); }
   { help(_function_name()); return; }

 if (typeof(id) != Array_Type) id = [id];
 % loop over ids
 _for i (0, length(id)-1, 1)
 {
   if (atime_metavalid(id[i]))
   {
     meta = struct_copy(get_dataset_metadata(id[i]));
     if (st == NULL) meta.shiftref = 1;
     else meta.shiftref = (st == 1);
     set_dataset_metadata(id[i], meta);
   }
 }
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_xinclude()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_xinclude}
%\synopsis{Specifies the datasets, which should be taken into account in the fit}
%\usage{atime_xinclude(Integer_Type[] index);}
%\description
%    During the fit only the given datasets containing
%    arrival times are taken into account. The remain-
%    ing ones are noticed such that no bins are used.
%    Hence the fit function and parameters are not
%    affected by changed dataset indices.
%\seealso{xnotice_atime, define_atime}
%!%-
{
 variable id, nind = atime_dataind;
 switch(_NARGS)
   { case 0: id = nind; }
   { case 1: (id) = (); }
   { help(_function_name()); return; }
 variable i, meta, fp, fr;
 ifnot (typeof(id) == Array_Type) id = [id];

 % include dataset
 _for i (0, length(id)-1, 1)
 { 
   if (wherefirst(nind == id[i]) == NULL) vmessage("warning (%s): given dataset does not exist or does not contain arrival times",_function_name);
   else
   {
     meta = struct_copy(get_dataset_metadata(id[i]));
     xnotice_atime(id[i],meta.notice[0],meta.notice[1]);
     thaw(meta.freeparams);
   }
 }
 % loop over datasets and exclude
 _for i (0, length(nind)-1, 1)
 {
   if (wherefirst(id == nind[i]) == NULL)
   {
     meta = struct_copy(get_dataset_metadata(nind[i]));
     xnotice(nind[i],0,0);
     % store free parameters
     fp = get_params(sprintf("arrtimes(%d).*", nind[i]));
     fr = wherenot(array_struct_field(fp, "freeze"));
     ifnot (fr == NULL)
     {
       fp = array_struct_field(fp, "index")[wherenot(array_struct_field(fp, "freeze"))];
       meta.freeparams = fp;
       set_dataset_metadata(nind[i], meta);
     }
     % freeze free parameters
     freeze(fp);
   }
 }
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define define_atime()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{define_atime}
%\synopsis{defines a dataset of arrival times}
%\usage{Integer_Type define_atime(Struct_Type atime[, Integer_Type divide]);}
%\qualifiers{
%    \qualifier{modnum}{a reference to a variable, where to store
%             the pulse numbers found by the model}
%    \qualifier{noff}{do not add 'arrtimes' to the actual fit-
%             function correspondig to the dataset}
%}
%\description
%    Takes the given arrival times structure to define
%    an ISIS dataset. This structure may be created by
%    'atime_det' or loaded by 'load_atime'. Accordingly
%    all usual ISIS routines, which operate on datasets,
%    e.g. fitting, can be used. The fit function
%    'arrtimes' handle arrival times including orbital
%    motion and pulse ephemeris. If not disabled by the
%    'noff' qualifier the actual fit function is auto-
%    matically extended by the resulting dataset using
%    the 'arrtimes' model. It also takes the qualifier
%    'modnum' into account to return the pulse numbers
%    determined by the model. The reference to the
%    pulse numbers is stored into the metadata of the
%    dataset.
%    If the second parameter 'divide' is given the
%    dataset is defined several times accordingly to
%    the given integer. The datasets are then noticed
%    automatically such that the whole data is divided
%    into the given number of parts of equal length.
%    Each part may be then fitted individually.
%    The returned integer corresponds to ISIS dataset
%    index.
%\seealso{arrtimes, atimes_det, load_atime, atime_metavalid, xnotice_atime}
%!%-
{
 variable s, cut;
 switch(_NARGS)
   { case 1: (s) = (); cut = Double_Type[0]; }
   { case 2: (s,cut) = (); }
   { help(_function_name()); return; }

 message("\nNOTE: there was a change in the definition of the ephemeris structure!");
 message("      If you encounter any problems or your previous fit doesn't work anymore");
 message("      please write an email to matthias.kuehnel@sternwarte.uni-erlangen.de\n");
 variable modnum = qualifier("modnum",NULL);
 if (typeof(modnum) != Ref_Type && modnum != NULL) { vmessage("warning (%s): qualifier modnum must be a reference",_function_name); modnum = NULL; }

 % if no cutting times are given, but number of segments,
 % create segments with equal length in time
 variable arrtimeshig = make_hi_grid(s.arrtimes);
 if (typeof(cut) == Integer_Type)
 {
   if (cut == 1) cut = Double_Type[0];
   else cut = 1.*min(s.arrtimes) + 1.*(max(arrtimeshig)-min(s.arrtimes)) * [1:cut-1] / cut;
 }
 % insert border of segments
 cut = [min(s.arrtimes),cut,max(arrtimeshig)];
 
 % check if smallest error allowed is small enough
 if (Minimum_Stat_Err < 0 or Minimum_Stat_Err > min(s.error))
 {
   Minimum_Stat_Err = 0.9*min(s.error);
   vmessage("warning (%s): Minimum_Stat_Err too large, changed to describe errors",_function_name);
 }

 variable seg = length(cut) - 1; % number of segments
 variable tmin, tmax;
 variable i, ind = Integer_Type[0];
 if (modnum != NULL) @modnum = Integer_Type[seg,length(s.arrnum)];
 % loop over segments
 _for i (0, seg-1, 1)
 {
   % define data
   ind = [ind, define_counts(s.arrtimes,arrtimeshig,s.arrtimes,s.error)];
   % store relative pulse numbers and used time interval (via metadata)
   set_dataset_metadata(ind[i], struct { datapnum = s.arrnum,
                                         refpnum = s.numref,
                                         reft0 = s.reft0,
                                         notice = [cut[i],cut[i+1]],
                                         modelpnum = modnum,
                                         id = i,
                                         useref = 1,
                                         shiftref = 1,
                                         freeparams = Integer_Type[0] });
   % xnotice
   xnotice_atime(ind[i],cut[i],cut[i+1]);
 }

 % define fit function, except if qualifier set
 ifnot (qualifier_exists("noff")) fit_fun("arrtimes(Isis_Active_Dataset)");
 
 return ind;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_get_t0()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_get_t0}
%\synopsis{returns the reference time of the arrival times}
%\usage{Double_Type atime_get_t0(Integer_Type[] dataid);}
%\description
%    This function returns the emitting time of the pulse
%    of number 0 as used in the polynomial to calculate
%    pulse arrival times.
%\seealso{arrtimes, pulse_time}
%!%-
{
 variable id, i;
 switch(_NARGS)
   { case 1: (id) = (); }
   { help(_function_name()); return; }
 if (typeof(id) == Integer_Type) id = [id];
 variable t0 = Double_Type[length(id)];
 _for i (0, length(id)-1, 1) t0[i] = struct_copy(get_dataset_metadata(id[i])).reft0;

 if (length(t0) == 1) t0 = t0[0];
 return t0;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define pulse_time()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{pulse_time}
%\synopsis{returns the pulse arrival time of the given pulse number with
%    respect to the pulse ephemeris and orbit}
%\usage{Double_Type pulse_time(Integer_Type[] number, Struct_Type ephemeris[, Struct_Type orbit]);
% or Double_Type pulse_time(Integer_Type[] number, Double_Type pulseperiod[, Struct_Type orbit]);}
%\qualifiers{
%    \qualifier{MJD}{the values of the pulse ephemeris are given
%           in days (default: seconds)}
%    \qualifier{eph}{may be set to the pulse ephemeris structure}
%    \qualifier{orb}{may be set to the orbital structure}
%    \qualifier{dphi}{an additive constant phase shift}
%}
%\description
%    Calculates the expected pulse arrival time of the
%    given pulse number, which may be a single value
%    or an array. The structures containing the pulse
%    ephemeris and the orbital parameters must follow
%    the conditions decribed in 'check_pulseperiod_orbit_struct'.
%    Instead of the pulse ephemeris the pulse period
%    may be given only. These structures may be also
%    passed to the function by qualifiers, hence the
%    structure parameters can be omitted.
%    The equation to calculate the arrival time is
%    similar to Hilditch eq. 3.53, including the terms
%    of the fourth order (p3dot). The numerical calcu-
%    lation is optimized using the Horner schema. For
%    details see 'arrtimes'.
%\seealso{check_pulseperiod_orbit_struct, pulseperiod, arrtimes}
%!%-
{
 variable n, eph, orb = NULL;
 switch(_NARGS)
   { case 1: (n) = (); eph = qualifier("eph", NULL); orb = qualifier("orb", NULL); }
   { case 2: (n,eph) = (); }
   { case 3: (n,eph,orb) = ();}
   { help(_function_name()); return; }

 variable dphi = qualifier("dphi",0);

 % backwards compatibility
 if (orb != NULL) {
   ifnot (struct_field_exists(orb, "torb0")) { orb = struct_combine(orb, struct { torb0 = orb.tau }); }
   ifnot (struct_field_exists(orb, "tau")) { orb = struct_combine(orb, struct { tau = orb.torb0 }); }
 }
 if (not struct_field_exists(eph, "tpuls0")) { eph = struct_combine(eph, struct { tpuls0 = eph.t0, ppuls = eph.p0 }); }
 if (not struct_field_exists(eph, "t0")) { eph = struct_combine(eph, struct { t0 = eph.tpuls0, p0 = eph.ppuls }); }
 ifnot (check_pulseperiod_orbit_struct(eph)) { vmessage("error (%s): pulse period structure not given properly", _function_name); return; }
 if (orb != NULL && check_pulseperiod_orbit_struct(orb) == 0) { vmessage("error (%s): orbit structure not given properly", _function_name); return; }

 variable t0 = eph.t0;
 variable p0 = eph.p0;
 variable pdot = eph.pdot;
 variable p2dot = eph.p2dot;
 variable p3dot = eph.p3dot;
 % transform into days if values are given in seconds (standard)
 ifnot (qualifier_exists("MJD"))
 {
   p0 /= 86400.;
   p2dot *= 86400.;
   p3dot *= 86400. * 86400.;
 }
 % pulse emitted time
 variable tn = ((((((pdot*pdot + 4.*p2dot*p0)*pdot + p0*p0*p3dot) * p0 * n / 24. + ((p0*p2dot + pdot*pdot)*p0)) * n / 6. + p0 * pdot / 2.) * n) + p0) * n + t0;
 % phase shift of pattern
 if (-1 < dphi < 1) tn += dphi * pulseperiod(tn, eph) / 86400;
 % take orbit into account
 variable z = 0.;
 if (orb != NULL && orb.asini > 0. && orb.porb > 0) z = BinaryPos(tn; porb=orb.porb, eccentricity=orb.ecc, omega=orb.omega,
					     asini=orb.asini, t0=orb.torb0, pporb=orb.pporb);
 return tn + z / 86400.;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define arrtimes_fit(lo,hi,pars)
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{arrtimes}
%\synopsis{fit function for modelling arrival times}
%\usage{arrtimes(id)}
%\description
%    This model can bes used in arrival time fits. The
%    pulse arrival times are calculated using the
%    'pulse_time' function, including orbital motion
%    and up to the third derivative of the pulse
%    period. The observed pulse arrival time t_obs(n)
%    is calculated by
%
%      t_obs(n) = t0 + A*n + B*n^2 + C*n^3 + D*n^4
%                 + z(t_emit(n))/c
%                 
%    where A to D are coefficients depending on the
%    pulse period and its higher derivatives, z/c is
%    the Doppler Shift due to orbital motion (see
%    'BinaryPos' for details) and t_emit(n) is the
%    time, when the nth pulse is emitted in the bary-
%    centre of the binary. This time is found solving
%    the above equation:
%    
%      t_emit(n) = t_obs(n) - z/c
%      
%    However, this can not be solved since the
%    observed arrival times include the Doppler
%    Shift. Hence the emission time must be found
%    iteratively, starting at t_emit(n) = t_obs(n).
%    The found pulse numbers are stored in a reference
%    variable, if set by 'define_atime'.
%    The model parameters are
%      ppuls - pulse period (s)
%      pdot  - first derivative (s/s)
%      p2dot - second derivative (s/s^2)
%      p3dot - third derivative (s/s^3)
%      dphi  - constant additive phase shift
%      porb  - orbital period (d)
%      torb0 - time of periastron passage (MJD)
%      asini - projected semi major axis (lts)
%      ecc   - eccentricity
%      omega - angle of periastron (degrees)
%      pporb - change of orbital period (d/d)
%    The reference time of the pulse ephemeris,
%    called tpuls0 in 'atime_get_ephemeris' for
%    example, is fix and set by 'define_atime'.
%    If you know what you are doing you can
%    change this time using 'atime_set_ephemeris'.
%\seealso{pulse_time, atime_set_ephemeris, define_atime, BinaryPos}
%!%-
{
 variable p0 = pars[0]/86400.;    % pulse period in reference frame of the binary [d]
 variable pdot = pars[1];         % first derivative of the pulse period [d/d]
 variable p2dot = pars[2]*86400.; % second derivative of the pulse period [d/d^2]
 variable p3dot = pars[3]*86400.*86400.; % third derivative of the pulse period [d/d^2]
 variable dphi = pars[4];         % phase shift of search pattern
 variable porb = pars[5];         % orbial period [days]
 variable ecc = pars[6];          % eccentricity of the orbit
 variable omega = pars[7];        % angle of periastron [degrees]
 variable asini = pars[8];        % projected major axis [lts]
 variable t0 = pars[9];           % time of periastron
 variable pporb = pars[10];       % first derivative of the orbital period [s/s]
 variable maxiter = 50;
 variable limit = 1e-12;
 variable chatty = 0;
 
 % check on valid metadata
 ifnot (atime_metavalid(Isis_Active_Dataset)) throw RunTimeError, "given dataset does not contain arrival times";
 % relative puls numbers as determined by getArrivalTimes
 variable meta = struct_copy(get_dataset_metadata(Isis_Active_Dataset));
 variable noti = @ (get_data_info(Isis_Active_Dataset).notice_list); % indices of used data
 variable pnum, z, tobs, temit, i, m, temit_new, dt;
 pnum = Double_Type[length(lo)];
 tobs = Double_Type[length(lo)]; tobs[*] = lo[*];
 temit = Double_Type[length(lo)];
 temit_new = Double_Type[length(lo)]; temit_new[*] = lo[*];
 
 % correct doppler shift of reference time to get pulse emission time for n=0
 variable tn0;
 variable tn0o = meta.reft0;
 variable tn0_new = tn0o;
 % correct this time for orbit (if orbit given)
 m=0;
 do
 {
   tn0 = tn0_new;
   if (asini>0 && porb>0)
   {
     z = BinaryPos(tn0; porb=porb, eccentricity=ecc, omega=omega, asini=asini, t0=t0, pporb=pporb);
     z = z / 86400.;
   } else z = 0.;
   tn0_new = tn0o - z;
   m++;
 } while (abs(tn0_new - tn0) > limit && m<maxiter);
 tn0 = tn0_new;
 % because tn0 != given reference time for pulse period, the pulse
 % ephemeris at tn0 must be calculated now
 variable neph = pulseperiod_transform(tn0, struct { p0 = p0*86400, pdot = pdot, p2dot = p2dot/86400, p3dot = p3dot/86400/86400, t0 = tn0o});
 p0 = neph.p0/86400; pdot = neph.pdot; p2dot = neph.p2dot*86400; p3dot = neph.p3dot*86400*86400;
 % shift reference pulse number if not disabled
 if (meta.shiftref)
 {
   variable ni = where_min(abs(lo-tn0));
   variable nr = meta.refpnum[noti[ni]];
   variable sndx = where(meta.refpnum == nr);
   meta.datapnum[sndx] -= meta.datapnum[noti[ni]];
   meta.refpnum[sndx] = noti[ni];
 }
 % get indices of all pulse numbers to determine
 variable ndx_det, ndx_ref;
 if (meta.useref) ndx_det = where(meta.datapnum[noti] == 0 | meta.refpnum[noti] < noti[0] | meta.refpnum[noti] > noti[length(noti)-1], &ndx_ref);
 else { ndx_det = [0:length(noti)-1]; ndx_ref = Integer_Type[0]; }

 dt = max(lo)-min(lo);
 % determine pulse numbers by orbit correct the arrival time and solve t(n)
 % this must be done iteratively
 variable n, refi;
 _for i (0, length(ndx_det)-1, 1)
 {
   m = 0; n = ndx_det[i];
   do
   {
     temit[n] = temit_new[n];
     % get z component of neutron star at emitted time (in lts) (if orbit given)
     if (porb>0 && asini>0)
     {
       z = BinaryPos(temit[n]; porb=porb, eccentricity=ecc, omega=omega, asini=asini, t0=t0, pporb=pporb);
       z = z / 86400.;
     } else z = 0;
     % find pulse number, solve Hilditch 3.53 for given t
     if (pdot == 0.0 && p2dot == 0.0 && p3dot == 0.0) pnum[n] = (lo[n] - z - tn0) / p0 - dphi;
     else pnum[n] = find_function_value(&pulse_time, lo[n]-z, (min(lo)-.5*dt-tn0)/p0, (max(lo)+.5*dt-tn0)/p0;
                    eps=1e-12, qualifiers = struct { dphi = dphi, eph = struct { tpuls0=tn0, ppuls=p0, pdot=pdot, p2dot=p2dot, p3dot=p3dot }, MJD }); % MJD qualifier important!
     % pulse number must be an integer
     pnum[n]=round(pnum[n]);
     % calculate new pulse arrival time
     temit_new[n] = pulse_time(int(pnum[n]), struct { tpuls0=tn0, ppuls=p0, pdot=pdot, p2dot=p2dot, p3dot=p3dot }; dphi = dphi, MJD);
     m++;
   } while (abs(temit_new[n] - temit[n]) > limit && m < maxiter);
   temit[n] = temit_new[n];
   if (porb>0 && asini>0)
   {
     z = BinaryPos(temit[n]; porb=porb, eccentricity=ecc, omega=omega, asini=asini, t0=t0, pporb=pporb);
     z = z / 86400.;
   } else z = 0;
   tobs[n] = temit[n] + z;
 }
 % set reference pulse numbers and calculate arrival times
 _for i (0, length(ndx_ref)-1, 1)
 {
   n = ndx_ref[i];
   refi = where(noti == meta.refpnum[noti[n]])[0];
   pnum[n] = pnum[refi] + meta.datapnum[noti[n]];
   % calculate arrival time
   temit[n] = pulse_time(int(pnum[n]), struct { tpuls0=tn0, ppuls=p0, pdot=pdot, p2dot=p2dot, p3dot=p3dot }; dphi = dphi, MJD);
   if (porb>0 && asini>0)
   {
     z = BinaryPos(temit[n]; porb=porb, eccentricity=ecc, omega=omega, asini=asini, t0=t0, pporb=pporb);
     z = z / 86400.;
   } else z = 0;
   tobs[n] = temit[n] + z;
 }
 % store derived pulse numbers if reference to variable is given
 if (meta.modelpnum != NULL) (@meta.modelpnum)[meta.id,noti] = int(pnum);
 return tobs;
} %}}}

% set default values and limits
add_slang_function("arrtimes",["ppuls [s]","pdot [s/s]","p2dot [s/s^2]","p3dot [s/s^3]","dphi","porb [days]","ecc","omega [degrees]","asini [lts]","torb0 [MJD]","pporb [s/s]"]);
define arrtimes_default_hook(i)
{
  variable defs = [
    struct { value = 1.0, freeze = 0, hard_min = 0.0, hard_max = 1e10, min = 0, max = 0, step = 1e-3, relstep = 1e-5 },
    struct { value = 0.0, freeze = 0, hard_min = -1.0, hard_max = 1.0, min = -1e-6, max = 1e-6, step = 1e-11, relstep = 1e-13 },
    struct { value = 0.0, freeze = 0, hard_min = -1.0, hard_max = 1.0, min = -1e-11, max = 1e-11, step = 1e-13, relstep = 1e-15 },
    struct { value = 0.0, freeze = 0, hard_min = -1.0, hard_max = 1.0, min = -1e-18, max = 1e-18, step = 1e-20, relstep = 1e-22 },
    struct { value = 0.0, freeze = 1, hard_min = -1.0, hard_max = 1.0, min = -1.0, max = 1.0, step = 1e-2, relstep = 1e-4 },
    struct { value = 0.0, freeze = 0, hard_min = 0.0, hard_max = 1000, min = 0, max = 1000, step = 1e-2, relstep = 1e-4 },
    struct { value = 0.0, freeze = 0, hard_min = 0.0, hard_max = 1.0, min = 0.0, max = 1.0, step = 1e-2, relstep = 1e-4 },
    struct { value = 0.0, freeze = 0, hard_min = -360, hard_max = 360.0, min = 0, max = 0, step = 1e-2, relstep = 1e-4 },
    struct { value = 0.0, freeze = 0, hard_min = 0.0, hard_max = 1e10, min = 0, max = 0, step = 1e-2, relstep = 1e-4 },
    struct { value = 0.0, freeze = 0, hard_min = 0.0, hard_max = 1e10, min = 0, max = 0, step = 1e-2, relstep = 1e-4 },
    struct { value = 0.0, freeze = 1, hard_min = -1.0, hard_max = 1.0, min = -1e-4, max = 1e-4, step = 1e-6, relstep = 1e-8 }
  ];
  return defs[i];
}
set_param_default_hook("arrtimes", &arrtimes_default_hook);


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_set_ephemeris()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_set_ephemeris}
%\synopsis{sets the actual model parameters of a given
%    dataset of pulse arrival times}
%\usage{atime_set_ephemeris(Integer_Type[] idx, Struct_Type eph[, Struct_Type orb]);}
%\altusage{atime_set_ephemeris(Integer_Type[] idx, Double_Type eph[, Struct_Type orb]);}
%\qualifiers{
%    \qualifier{sett0}{also set the reference time of the pulse
%               ephemeris, which is used internally}
%}
%\description
%    The given structure(s) containing the pulse ephemeris
%    and the orbital parameters are used to set the model
%    parameters of dataset 'idx'. Instead of the pulse
%    ephemeris the pulse period may be given only. Other-
%    wise the structure has to contain:
%      ppuls  - pulse period (s)
%      pdot   - first derivative (s/s)
%      p2dot  - second derivative (s/s^2)
%      p3dot  - third derivative (s/s^3)
%      tpuls0 - reference time (MJD)
%    The orbital parameters are stored in the second
%    structure:
%      porb   - orbital period (d)
%      torb0  - time of periastron passage (MJD)
%      asini  - projected semi major axis (lts)
%      ecc    - eccentricity
%      omega  - angle of periastron (degrees)
%    By default the reference time of the pulse ephemeris
%    is NOT set, because this should be fixed and set to
%    the same time as used for the pulse pattern. This
%    time therefore defines the arrival time of the pulse
%    of number n=0. Only change this time if you know what
%    effects will result (see arrtimes function)!    
%\seealso{arrtimes, atime_get_ephemeris}
%!%-
{
  variable id, eph, orb = NULL;
  switch(_NARGS)
    { case 2: (id,eph) = (); }
    { case 3: (id,eph,orb) = (); }
    { help(_function_name()); return; }
  ifnot (typeof(id) == Array_Type) id = [id];
  % check on valid metadata
  ifnot (atime_metavalid(id)) throw RunTimeError, "given dataset does not contain arrival times";
  % check given structures
  ifnot (check_pulseperiod_orbit_struct(eph)) { vmessage("error (%s): pulse period structure not given properly", _function_name); return; }
  ifnot (orb != NULL && check_pulseperiod_orbit_struct(orb)) { vmessage("error (%s): orbit structure not given properly", _function_name); return; }
  % backwards compatibility
  if (orb != NULL && not struct_field_exists(orb, "torb0")) {
    orb = struct_combine(orb, struct { torb0 = orb.tau });
  }
  if (not struct_field_exists(eph, "tpuls0")) {
    eph = struct_combine(eph, struct { tpuls0 = eph.t0, ppuls = eph.p0 });
  }
  % set model values
  variable f, meta, i;
  foreach i (id)
  {
    foreach f (get_struct_field_names(eph))
    {
      if (f == "tpuls0" && qualifier_exists("sett0"))
      {
        meta = struct_copy(get_dataset_metadata(i));
        meta.reft0 = eph.tpuls0;
        set_dataset_metadata(i, meta);
      }
      else ifnot (f == "tpuls0") set_par(sprintf("arrtimes(%d).%s",i,f), @get_struct_field(eph,f));
    }
    if (orb != NULL && orb.asini > 0 && orb.porb > 0)
      foreach f (get_struct_field_names(orb)) set_par(sprintf("arrtimes(%d).%s",i,f), @get_struct_field(orb,f));
    else {
      set_par(sprintf("arrtimes(%d).porb",i), 0, 1, 0, 0);
      set_par(sprintf("arrtimes(%d).ecc",i), 0, 1, 0, 0);
      set_par(sprintf("arrtimes(%d).omega",i), 0, 1, 0, 0);
      set_par(sprintf("arrtimes(%d).asini",i), 0, 1, 0, 0);
      set_par(sprintf("arrtimes(%d).torb0",i), 0, 1, 0, 0);
      set_par(sprintf("arrtimes(%d).pporb",i), 0, 1, 0, 0);
    }
  }
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_get_ephemeris()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_get_ephemeris}
%\synopsis{retrieves the actual model parameters of a given
%    dataset of pulse arrival times}
%\usage{(Struct_Type, Struct_Type) atime_get_ephemeris(Integer_Type idx);}
%\description
%    Returnes two structures containing the model parameters
%    of a dataset, which contains pulse arrival times. The
%    first structure describes the pulse ephemeris:
%      ppuls  - pulse period (s)
%      pdot   - first derivative (s/s)
%      p2dot  - second derivative (s/s^2)
%      p3dot  - third derivative (s/s^3)
%      tpuls0 - reference time (MJD)
%    The orbital parameters are stored in the second
%    structure:
%      porb   - orbital period (d)
%      torb0  - time of periastron passage (MJD)
%      asini  - projected semi major axis (lts)
%      ecc    - eccentricity
%      omega  - angle of periastron (degrees)
%    If porb or asini is zero, the returned orbital
%    structure will be set to NULL.
%\seealso{arrtimes, check_pulseperiod_orbit_struct}
%!%-
{
  variable id;
  switch(_NARGS)
    { case 1: (id) = (); }
    { help(_function_name()); return; }
  % check on valid metadata
  ifnot (atime_metavalid(Isis_Active_Dataset)) throw RunTimeError, "given dataset does not contain arrival times";
  % build structures
  variable eph = struct {
    p0  = get_par(sprintf("arrtimes(%d).ppuls",id)),
    pdot   = get_par(sprintf("arrtimes(%d).pdot",id)),
    p2dot  = get_par(sprintf("arrtimes(%d).p2dot",id)),
    p3dot  = get_par(sprintf("arrtimes(%d).p3dot",id)),
    t0 = struct_copy(get_dataset_metadata(id)).reft0
  };
  if (get_par(sprintf("arrtimes(%d).porb",id)) > 0 && get_par(sprintf("arrtimes(%d).asini",id)) > 0)
    return (eph, struct {
      porb   = get_par(sprintf("arrtimes(%d).porb",id)),
      tau  = get_par(sprintf("arrtimes(%d).torb0",id)),
      asini  = get_par(sprintf("arrtimes(%d).asini",id)),
      ecc    = get_par(sprintf("arrtimes(%d).ecc",id)),
      omega  = get_par(sprintf("arrtimes(%d).omega",id)),
      pporb  = get_par(sprintf("arrtimes(%d).pporb",id)) });
  else return (eph, NULL);
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_save_ephemeris()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_save_ephemeris}
%\synopsis{saves pulse ephemeris and orbital parameters into as an S-lang script}
%\usage{atime_save_ephemeris(Integer_Type idx, String_Type filename);
% or atime_save_ephemeris(Struct_Type eph[, Struct_Type orb], String_Type filename)}
%\description
%    The fitted pulse ephemeris and orbital paramters
%    of the given dataset or the structures themself
%    are saved into a file. This file is an S-lang
%    script and can be called directly or loaded by
%    using 'atime_load_ephemeris'.
%\seealso{atime_load_ephemeris, atime_get_ephemeris}
%!%-
{
 % process arguments
 variable id, filename, eph, orb = NULL;
 switch(_NARGS)
   { case 2:
     (id, filename) = ();
     if (typeof(id) == Struct_Type) eph = id;
     else
     {
       if (length(id) > 1) { vmessage("error (%s): only one dataset is allowed", _function_name); return; }
       (eph,orb) = atime_get_ephemeris(id);
     }
   }
   { case 3: (eph, orb, filename) = (); }
   { help(_function_name()); return; }

 % write S-lang file
 variable f, fp = fopen(filename, "w+");
 ()=fprintf(fp, "(\n  struct {\n");
 foreach f (get_struct_field_names(eph)) ()=fprintf(fp, "    %s = %S,\n", f, get_struct_field(eph, f));
 ()=fprintf(fp, "  },\n");
 if (orb != NULL && orb.asini > 0 && orb.porb > 0)
 {
   ()=fprintf(fp, "  struct {\n");
   foreach f (get_struct_field_names(orb)) ()=fprintf(fp, "    %s = %S,\n", f, get_struct_field(orb, f));
   ()=fprintf(fp, "  }\n");
 } else ()=fprintf(fp, "  NULL\n");
 ()=fprintf(fp, ")\n");
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_load_ephemeris()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_load_ephemeris}
%\synopsis{returns previously ssaved pulse ephemeris and orbital parameters}
%\usage{atime_load_ephemeris(Integer_Type idx, String_Type filename);
% or (Struct_Type, Struct_Type) atime_load_ephemeris(String_Type filename);}
%\description
%    The pulse ephemeris and orbital parameters, which
%    have been previously saved by using the function
%    'atime_save_ephemeris', are either assigned to the
%    given dataset 'idx' or returned as structures.
%    The first one defines the pulse ephemeris and the
%    second one the orbit.
%    The following commands are equal:
%      (eph, orb) = atime_load_ephemeris(filename);
%      (eph, orb, ) = evalfile(filename);
%\seealso{atime_save_ephemeris, atime_get_ephemeris, evalfile}
%!%-
{
 variable eph, orb, idx, filename;
 switch(_NARGS)
   { case 1: filename = (); (eph, orb, ) = evalfile(filename); return (eph, orb); }
   { case 2: (idx, filename) = (); (eph, orb, ) = evalfile(filename); atime_set_ephemeris(idx, eph, orb); }
   { help(_function_name()); return; }
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define plot_atime()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{plot_atime}
%\synopsis{plots the residuals of a pulse arrival times fit}
%\usage{plot_atime([Integer_Type[] index]);}
%\qualifiers{
%    \qualifier{t0}{reference time of xticks}
%    \qualifier{col}{array of colors for datasets}
%    \qualifier{noerrbars}{do not plot errorbars}
%    \qualifier{connect_points}{see 'connect_points'}
%    \qualifier{style}{see 'point_style'}
%    \qualifier{xunit}{used time unit (d or s)}
%    \qualifier{yunit}{used residual unit (phi or s)}
%    \qualifier{xlunit}{label of xunit}
%    \qualifier{ylunit}{label of yunit}
%    \qualifier{xrng}{time range}
%    \qualifier{yrng}{residual range (default -0.5 to 0.5)}
%    \qualifier{ploteph}{overplot the pulse ephemeris}
%    \qualifier{plotorb}{overplot the orbital motion}
%    \qualifier{plotpnum}{overplot the modelled pulse numbers
%                     ('modnum' reference must be set,
%                     see 'define_atime')}
%    \qualifier{plotdif}{overplot the meassured pulse period
%                     ('modnum' reference must be set,
%                     see 'define_atime')}
%    \qualifier{y2rng}{yrange of overplot}
%    \qualifier{y2rel}{if pulse period tics are displayed
%                     wrong, you should switch to relative
%                     values and milliseconds}
%    \qualifier{extcol}{color of overplot}
%    \qualifier{extcolY}{yaxis color of overplot}
%    \qualifier{oplot}{do not erase plot window}
%    \qualifier{lshift}{if 'peph' and 'porb' set, shift the
%                     labels printed at the ephemeris by
%                     the given value in y-direction. May
%                     be a two value array [eph,orb].
%                     Values are interpreted as relative
%                     coordinates.}
%}
%\description
%    By default plots the residuals of all datasets
%    containing arrival times. The residuals are
%    calculated by
%      residuals = data - model
%    Using the optional
%    first parameter the indices of the datasets to
%    be plotted can be specified. Also works if
%    some datasets are excluded from the fit (see
%    'atime_xinclude').
%    The residuals in units of pulse phase or seconds
%    are plotted against the meassured pulse arrival
%    times in days or seconds. The units can be set
%    by using the accordant qualifiers. Further
%    informations can be overplotted, like the
%    pulse period in the binary barycentre.
%\seealso{arrtimes, atime_dataind, atime_xinclude, plot_data}
%!%-
{
 variable fn;
 switch(_NARGS)
   { case 0: fn = atime_dataind; }
   { case 1: fn = ();}
   { help(_function_name()); return; }

 % check if datasets containing arrival times exist
 if (typeof(fn) == Null_Type) { vmessage("warning (%s): no datasets containing arrival times defined",_function_name); return; }
 
 if (typeof(fn) != Array_Type) fn=[fn];
 % loop over datasets to get arrival times
 variable i, atimes = Double_Type[0], datimes = Double_Type[0], mtimes = Double_Type[0], ndx=0, noti;
 _for i (0, length(fn)-1, 1)
 {
  noti = @ (get_data_info(fn[i]).notice_list);
  % apply notice
  atimes = [atimes, get_data_counts(fn[i]).value[noti]];
  datimes = [datimes, get_data_counts(fn[i]).err[noti]];
  mtimes = [mtimes, get_model_counts(fn[i]).value[noti]];
  ndx = [ndx,ndx[length(ndx)-1]+length(noti)]; % indices belonging to different datasets
 }
 variable po = get_plot_options.yopt;
 variable p0, res, ndxu;

 variable eph, orb;

 variable nomod = not sum(mtimes); % model evaluated?

 % apply units
 variable xunit = qualifier("xunit","d");
 variable yunit = qualifier("yunit","phi");
 variable xlunit = "MJD";
 variable ylunit = "Pulse Phase";
 if (xunit == "d") xunit = 1.;
 else if (xunit == "s") { xunit = 86400; xlunit = "s"; }
 else { vmessage("warning (%s): x-axis unit '%s' not known, assuming days", _function_name, xunit); xunit = 1.; }
 if (yunit == "phi") yunit = 1.;
 else if (yunit == "s") ylunit = "s";
 else { vmessage("warning (%s): y-axis unit '%s' not known, assuming pulse phase", _function_name, yunit); yunit = 1.; }

 variable dnum = max(atimes)-min(atimes);
 variable t0 = qualifier("t0",round(mean(atimes)));
 if (typeof(t0) == Integer_Type) t0 = 1.*t0;
  
 variable xrng = qualifier("xrng",[min(atimes)-t0-0.1*dnum,max(atimes)-t0+0.1*dnum]);
 variable yrng = qualifier("yrng",NULL);
 if (yrng == NULL) % check on |residuals|>0.5
 {
   yrng = [-0.5,0.5];
   _for i (0, length(fn)-1, 1)
   {
     % calculate the actual pulse period
     (eph,orb) = atime_get_ephemeris(fn[i]);
     p0 = pulseperiod(atimes, eph, orb);
     if (ylunit == "s") yunit = max(p0);
     res = (atimes-mtimes)*86400./p0;
     res = res + sign(res)*datimes*86400./p0;
     if (min(res) < -0.5 && yrng[0] > min(res)) yrng[0] = min(res);
     if (max(res) > +0.5 && yrng[1] < max(res)) yrng[1] = max(res);
   }
 }
 if (nomod) { yrng = xrng; }
 % apply ranges
 yrng = yrng * yunit;
 xrng = xrng * xunit;
 if (qualifier("oplot",0) == 0)
 {
   yrange(yrng[0],yrng[1]);
   xrange(xrng[0],xrng[1]);
 }

 variable pper = qualifier_exists("ploteph");
 variable porb = qualifier_exists("plotorb");
 variable pnum = qualifier_exists("plotpnum");
 variable pdif = qualifier_exists("plotdiff");
 variable y2rng = qualifier("y2rng",NULL);
 % check on multiple plot qualifiers
 if (sum([pper || porb,pnum,pdif]) > 1)
 {
   vmessage("warning (%s): only one plot qualifier allowed, will ignore all plot qualifiers",_function_name);
   pper = 0; pnum = 0; pdif = 0; porb = 0;
 }
 % deactivate right border -> second axis
 else if ((pper || porb || pnum || pdif) && not nomod) ()=change_plot_options(;yopt="BNST");

 variable colData = qualifier("col",[1:length(fn)]+1);
 if (length(colData)!=length(fn)) colData=[colData,[max(colData):max(colData)+length(fn)-length(colData)-1]+1];
  
 if (qualifier("oplot",0) == 0)
 {
   xlabel(sprintf("Time (" + xlunit + " - %.2f MJD)",t0));
   if (nomod) ylabel(sprintf("Time (" + xlunit + " - %.2f MJD)",t0));
   else ylabel("Residuals (" + ylunit + ")");
   % "erase" plot window
   color(0); plot(xrng,[0,0]);
 }
 ifnot (nomod)
 {
   color(1);
   line_style(2);
   connect_points(1);
   point_style(-1);
   oplot(xrng,[0,0]);
   line_style(1);
   _pgsls(1);
 }

 variable colP = qualifier("extcol",15);
 variable colY2 = qualifier("extcolY",colP);
 variable lshift = qualifier("lshift",[0,0]);
 if (length(lshift) == 1) lshift = [lshift, lshift];
 % eventually plot pulse ephemeris
 if (pper && not nomod)
 {
  variable pulsper = Double_Type[0], pulst = Double_Type[0], pdot, p2dot, p3dot, t0p, nt, ex;
  ndxu = 0;
  _for i (0, length(fn)-1, 1)
  {
    ifnot (ndx[i] == ndx[i+1])
    {
      p0 = get_par(sprintf("arrtimes(%d).ppuls",fn[i])) / 86400;
      pdot = get_par(sprintf("arrtimes(%d).pdot",fn[i]));
      p2dot = get_par(sprintf("arrtimes(%d).p2dot",fn[i])) * 86400;
      p3dot = get_par(sprintf("arrtimes(%d).p3dot",fn[i])) * 86400 * 86400;
      t0p = struct_copy(get_dataset_metadata(fn[i])).reft0;
      nt = [mtimes[ndx[i]], mtimes[ndx[i+1]-1]];
      nt = [nt[0]:nt[1]:(nt[1]-nt[0])/100];
      pulst = [pulst,nt];
      pulsper = [pulsper,p0 + pdot * (nt-t0p) + 1./2. * p2dot * sqr(nt-t0p) + 1./6. * p3dot * sqr(nt-t0p)*(nt-t0p)];
      ndxu = [ndxu,ndxu[length(ndxu)-1]+length(nt)];
    } else ndxu = [ndxu, ndxu[length(ndxu)-1]];
  }
  pulsper = pulsper*86400;
  dnum = max(pulsper)-min(pulsper);
  if (dnum == 0.0) ex=1.0; else ex=0.0;
  line_style(1);
  if (y2rng == NULL) y2rng=[min(pulsper)-0.1*dnum-ex,max(pulsper)+0.1*dnum+ex];
  if (qualifier_exists("y2rel")) % _pgaxis has problems with small numbers
  { % so this qualifier switches to a relative pulse period and milliseconds
    variable sw = y2rng[0] > 0 ? 1. : 1e3;
    variable rel = round(floor(y2rng[0]*sw))/sw;
    y2label(sprintf("Pulse Period (ms - %d %s)", int(rel*sw), sw == 1 ? "s" : "ms"); color=colP, f=0.07);
    y2axis((y2rng[0]-rel)*1e3, (y2rng[1]-rel)*1e3; color=colY2);
  }
  else
  {
    y2label("Pulse Period (s)"; color=colP, f=0.07);
    y2axis(y2rng[0], y2rng[1]; color=colY2);
  }
  _for i (0, length(fn)-1, 1)
  {
    ifnot (ndx[i] == ndx[i+1])
    {
      nt = [ndxu[i]:ndxu[i+1]-1];
      color(colP);
      if (dnum == 0.0) oplot(xrng, [0,0]);
      else oplot((pulst[nt]-t0)*xunit,(pulsper[nt]-y2rng[0])/(y2rng[1]-y2rng[0])*sum(abs(yrng))/yunit + yrng[0]/yunit);
    }
  }
  if (pper && porb)
  {
    color(colP);
    ex = where_min(pulst);
    xylabel_in_box((pulst[ex]-xrng[0]-t0)/(xrng[1]-xrng[0]), (pulsper[ex]-y2rng[0])/(y2rng[1]-y2rng[0]) + lshift[0], "P\\deph\\u(t)");
  }
  ()=change_plot_options(;yopt=po); % restore plot options
 }

 % eventually plot orbital motion
 if (porb && not nomod)
 {
  pulsper = Double_Type[0]; pulst = Double_Type[0];
  variable asini, ecc, omega, torb0, period;
  ndxu = 0;
  _for i (0, length(fn)-1, 1)
  {
    ifnot (ndx[i] == ndx[i+1])
    {
      p0 = get_par(sprintf("arrtimes(%d).ppuls",fn[i]));
      asini = get_par(sprintf("arrtimes(%d).asini",fn[i]));
      ecc = get_par(sprintf("arrtimes(%d).ecc",fn[i]));
      omega = get_par(sprintf("arrtimes(%d).omega",fn[i]));
      torb0 = get_par(sprintf("arrtimes(%d).torb0",fn[i]));
      period = get_par(sprintf("arrtimes(%d).porb",fn[i]));
      nt = [mtimes[ndx[i]], mtimes[ndx[i+1]-1]];
      nt = [nt[0]:nt[1]:(nt[1]-nt[0])/100];
      pulst = [pulst,nt];
      pulsper = [pulsper,pulseperiod(nt, p0, struct { asini = asini, omega = omega, ecc = ecc, tau = torb0, porb = period })];
      ndxu = [ndxu,ndxu[length(ndxu)-1]+length(nt)];
    } else ndxu = [ndxu, ndxu[length(ndxu)-1]];
  }
  ifnot (pper)
  {
    dnum = max(pulsper)-min(pulsper);
    if (dnum == 0.0) ex=1.0; else ex=0.0;
    line_style(1);
    if (y2rng == NULL) y2rng=[min(pulsper)-0.1*dnum-ex,max(pulsper)+0.1*dnum+ex];
    y2axis(y2rng[0],y2rng[1];color=colY2);
    y2label("Pulse Period (s)";color=colP,f=0.07);
  }
  _for i (0, length(fn)-1, 1)
  {
    ifnot (ndx[i] == ndx[i+1])
    {
      nt = [ndxu[i]:ndxu[i+1]-1];
      color(colP);
      if (pper && dnum == 0.0) oplot(xrng, [0,0]);
      else oplot((pulst[nt]-t0)*xunit,(pulsper[nt]-y2rng[0])/(y2rng[1]-y2rng[0])*sum(abs(yrng))/yunit + yrng[0]/yunit);
    }
  }
  if (pper && porb)
  {
    color(colP);
    ex = where_min(pulst);
    xylabel_in_box((pulst[ex]-xrng[0]-t0)/(xrng[1]-xrng[0]), (pulsper[ex]-y2rng[0])/(y2rng[1]-y2rng[0]) + lshift[1], "P\\dorb\\u(t)");
  }
  ()=change_plot_options(;yopt=po); % restore plot options
 }

 % eventually plot model pulse numbers
 if (pnum && not nomod)
 {
  pnum = qualifier("plotpnum");
  if (typeof(pnum) == Ref_Type && length(@pnum[0,*]) == length(mtimes))
  {
   pnum = @pnum;
   dnum = max(pnum)-min(pnum);
   line_style(1);
   if (y2rng == NULL) y2rng=[min(pnum)-0.1*dnum,max(pnum)+0.1*dnum];
   y2axis(y2rng[0],y2rng[1];color=colY2);
   y2label("Pulse Number";color=colP,f=0.07);
   color(colP);
   connect_points(0);
   point_style(5);
   oplot((mtimes-t0)*xunit,(pnum-y2rng[0])/(y2rng[1]-y2rng[0])*sum(abs(yrng))/yunit + yrng[0]/yunit);
   connect_points(1);
   point_style(-1);
  }
  else vmessage("warning (%s): qualifier 'plotpnum' must be a reference to the pulse number array",_function_name);
  ()=change_plot_options(;yopt=po); % restore plot options
 }

 % eventually plot 'measured' pulse period
 if (pdif && not nomod)
 {
  pdif = qualifier("plotdiff");
  if (typeof(pdif) == Ref_Type)
  {
   pdif = @pdif;
   variable mn, mx, pdife, ftimes;
   ndxu = wherenot(isinf(pdif));
   pdif = pdif[ndxu];
   pdife = pdife[ndxu];
   ftimes = ftimes[ndxu];
   dnum = max(pdif)-min(pdif);
   line_style(1);
   if (y2rng == NULL) y2rng=[min(pdif)-0.1*dnum,max(pdif)+0.1*dnum];
   y2axis(y2rng[0],y2rng[1];color=colY2);
   y2label("Arrival Times Difference (s)";color=colP,f=0.07);
   color(colP);
   connect_points(0);
   point_style(5);
   _for i (0, length(fn)-1, 1)
   {
     ifnot (ndx[i] == ndx[i+1])
     {
       mx = [ndx[i]+1:ndx[i+1]];
       mn = [ndx[i]:ndx[i+1]-1];
       pdife = sqrt(sqr(datimes[mx])+sqr(datimes[mn])) / (pdif[mx] - pdif[mn]) * 86400;
       pdif = (atimes[mx] - atimes[mn]) / (pdif[mx] - pdif[mn]) * 86400;
       ftimes = (mtimes[mx] + mtimes[mn]) / 2.;
       oplot_with_err((ftimes-t0)*xunit,(pdif-y2rng[0])/(y2rng[1]-y2rng[0])*sum(abs(yrng))/yunit + yrng[0]/yunit);
     }
   }
   connect_points(1);
   point_style(-1);
  }
  else vmessage("warning (%s): qualifier 'plotdiff' must be a reference to the pulse number array",_function_name);
  ()=change_plot_options(;yopt=po); % restore plot options
 }

 % style
 variable style = qualifier("style", 4);
 point_style(style);
 if (qualifier_exists("noerrbars")) connect_points(qualifier_exists("connect_points"));
 % plot arrival times
 _for i (0, length(fn)-1, 1)
 {
   (eph,orb) = atime_get_ephemeris(fn[i]);
   if (ndx[i]<ndx[i+1]) % check on ignored datasets
   {
     color(colData[i]);
     ndxu = [ndx[i]:ndx[i+1]-1]; % indices to use
     % calculate residuals (with respect to the actual pulse period)
     res = (atimes[ndxu] - mtimes[ndxu])*86400.;
     ifnot (ylunit == "s") p0 = pulseperiod(atimes[ndxu], eph, orb);
     else p0 = 1.;
     % if no model exists plot arrtimes also in y-direction
     if (nomod) { res = atimes[ndxu]-t0; p0 = 86400.; }
     if (qualifier_exists("noerrbars")) oplot((atimes[ndxu]-t0)*xunit,res/p0);
     else oplot_with_err((atimes[ndxu]-t0)*xunit,res/p0,datimes[ndxu]*86400./p0;;__qualifiers);
   }
 }
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_sim()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_sim}
%\synopsis{simulates pulse arrival times}
%\usage{Struct_Type atime_sim([Integer_Type n,] Double_Type tmin, tmax, Struct_Type eph[, orb]);}
%\qualifiers{
%    \qualifier{scramble}{1 sigma noise of the simulated arrival
%               times in seconds (default = 0)}
%    \qualifier{dn}{the increment of the arrival times if
%               they are not randomly generated
%               (default = 1)}
%}
%\description
%    Simulates the orbit and the pulsation of a pulsar
%    and returns the pulse arrival times in the time
%    range given by 'tmin' and 'tmax'. If the number
%    of arrival times 'n' is given, then n randomly
%    selected arrival times are returned. The returned
%    structure is equal to the atime_det function.
%    The uncertainties of the simulated arrival times
%    are calculated using the 'scramble' value. If no
%    value is given, the uncertainties are set to 0.1%
%    of the pulse period to avoid ISIS from treating
%    the arrival times as counts, resulting in poisson
%    errors.
%    The structures describing the pulse ephemeris and
%    the orbit must fullfil the conditions described
%    in the function check_pulseperiod_orbit_struct.
%
% NOTE
%    Due to speed reasons the returned arrival times
%    might be beyond the given time interval or not
%    all pulses in this interval are returned.
%\seealso{atime_det, check_pulseperiod_orbit_struct, arrtimes}
%!%-
{
 variable n = NULL, tmin, tmax, eph, orb = NULL;
 switch(_NARGS)
   { case 3: (tmin,tmax,eph) = (); }
   { case 4: (n,tmin,tmax,eph) = (); if (typeof(n) != Integer_Type) { orb = @eph; eph = @tmax; tmax = tmin; tmin = n; n = NULL; }}
   { case 5: (n,tmin,tmax,eph,orb) = (); }
   { help(_function_name()); return; }

 ifnot (check_pulseperiod_orbit_struct(eph)) { vmessage("error (%s): pulse period structure not given properly", _function_name); return; }
 ifnot (orb != NULL && check_pulseperiod_orbit_struct(orb)) { vmessage("error (%s): orbit structure not given properly", _function_name); return; }

 % backwards compatibility
 if (orb != NULL && not struct_field_exists(orb, "torb0")) {
   orb = struct_combine(orb, struct { torb0 = orb.tau });
 }
 if (not struct_field_exists(eph, "tpuls0")) {
   eph = struct_combine(eph, struct { tpuls0 = eph.t0, ppuls = eph.p0 });
 }

 variable scramble = qualifier("scramble", 0.0);
 variable dn = qualifier("dn", 1);

 variable atimes, nums;

 % correct the reference time for orbit (if orbit given)
 variable maxiter = 50;
 variable limit = 1e-12;
 variable tn0, tn0_new = eph.tpuls0, tn0o;
 variable m=0, z;
 tn0o = tn0_new;
 do
 {
   tn0 = tn0_new;
   if (orb != NULL && orb.asini>0 && orb.porb>0)
   {
     z = BinaryPos(eph.tpuls0; porb=orb.porb, eccentricity=orb.ecc, omega=orb.omega, asini=orb.asini, t0=orb.tau, pporb=orb.pporb);
     z = z / 86400.;
   } else z = 0.;
   tn0_new = tn0o - z;
   m++;
 } while (abs(tn0_new - tn0) > limit && m<maxiter);
 tn0_new = tn0;
 % because tn0 != given reference time for pulse period, the pulse
 % ephemeris at tn0 must be calculated now
 eph = pulseperiod_transform(tn0, eph);
  
 % calculate the pulse numbers roughly
 variable ndx;
 if (n == NULL)
 {
   nums = round([(tmin-eph.tpuls0) / (eph.p0/86400.):(tmax-eph.t0) / (eph.p0/86400.):dn]); % all pulse numbers
   ndx = [0:length(nums)-1];
 }
 else
 {
   nums = round((urand(n)*(tmax-tmin)+(tmin-eph.tpuls0)) / (eph.p0/86400.)); % random pulse numbers
   nums = nums[array_sort(nums)]; % sort pulse numbers
   % check on multiple equal pulse numbers
   ndx = wherenot(nums[[1:length(nums)-1]] - nums[[0:length(nums)-2]] == 0); % difference to next pulse number = 0 -> equal
   % only use not equal numbers
   ndx = [ndx,length(nums)-1]; % also include last number
   nums = nums[ndx];
 }
 % apply ephemeris and orbit
 atimes = pulse_time(nums, struct_combine(eph, struct { ppuls = eph.p0, tpuls0 = eph.t0 }), orb);

 % scramble arrival times and calculate the error
 atimes += grand(length(ndx))*scramble/86400.;
 variable atimes_err;
 if (scramble == 0) atimes_err = 1e-3 * eph.p0 * ones(length(atimes)) / 86400.;
 else atimes_err = ones(length(atimes))*scramble/86400.;

 return struct { arrtimes = atimes, error = atimes_err, arrnum = nums - nums[0], numref = ones(length(atimes))*0, reft0 = tn0o };
} %}}}


private define atime_sim_residuals_maxdif(x) %{{{
% Determine the maximum of the residuals of the
% parameter given by the necessary qualifier 'par'
% and its value 'x'. Use in the atime_sim_residuals
% to find the maximum variation of a parameter to
% produce nice plots.
{
  variable par = qualifier("par", NULL);
  variable id = qualifier("id", 1);
  if (par == NULL) return NULL;
  % set parameter and evaluate
  set_par(par,x);
  ()=eval_counts(;fit_verbose=-2);
  % return maximum residuals
  variable res = max(abs(get_data_counts(id).value-get_model_counts(id).value))*86400/get_par(sprintf("arrtimes(%d).ppuls",id));
  return res;
} %}}}

%%%%%%%%%%%%%%%%%%%%%%%%
define atime_sim_residuals()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_sim_residuals}
%\synopsis{simulates arrival times to calculate how the
%  residuals depend on parameter variations}
%\usage{atime_sim_residuals(Double_Type tmin, tmax, Struct_Type eph, orb, String_Type psfile);}
%\qualifiers{
%    \qualifier{n}{number of variations for each parameter
%           and sign (-> 2*n variatons, default: 3)}
%    \qualifier{tpd}{number of simulated arrival times per day
%           (default: 10)}
%    \qualifier{mres}{maximum allowed residuals (=yrange) in
%           units of pulse phase (default: 0.45),
%           must be in the range of 0.0-0.45}
%    \qualifier{minr}{smallest allowed relative variation of a
%           parameter, used to determine the overall
%           variation of this parameter to produce
%           nice plots (default: 1e-10)}
%    \qualifier{pars}{only simulate this given parameters
%           (default: all except tpuls0 and pporb)}
%}
%\description
%    This function determines the dependency of the
%    residuals on the given orbit and pulse ephemeris.
%    Therefore pulse arrival times are simulated and
%    the fit parameters are variied in the way, that
%    the shape of the resulting residuals can be seen
%    well. The resulting plots are stored in a Post-
%    Script file specified by the filename 'psfile'.
%    The labels show the absolute variation of the
%    the parameter.
%    The simulated times will be in the range of
%    'tmin' to 'tmax' (in MJD). The orbital parameter
%    and the pulse ephemeris structure must be in the
%    same form described in the function
%    check_pulseperiod_orbit_struct.
%\seealso{atime_sim, check_pulseperiod_orbit_struct, arrtimes}
%!%-
{
 variable tmin, tmax, eph, orb, psfile;
 switch(_NARGS)
   { case 5: (tmin,tmax,eph,orb,psfile) = (); }
   { help(_function_name()); return; }

 ifnot (check_pulseperiod_orbit_struct(eph)) { vmessage("error (%s): pulse period structure not given properly", _function_name); return; }
 if (orb == NULL) { vmessage("error (%s): orbit structure not given!", _function_name); return; }
 ifnot (check_pulseperiod_orbit_struct(orb)) { vmessage("error (%s): orbit structure not given properly", _function_name); return; }
 
 % backwards compatibility
 if (orb != NULL && not struct_field_exists(orb, "torb0")) {
   orb = struct_combine(orb, struct { torb0 = orb.tau });
 }
 if (not struct_field_exists(eph, "tpuls0")) {
   eph = struct_combine(eph, struct { tpuls0 = eph.t0, ppuls = eph.p0 });
 }

 variable n = qualifier("n",3);
 variable mres = qualifier("mres",0.45);
 variable minr = qualifier("minr",1e-10);
 variable tpd = qualifier("tpd",10);
#ifexists xfig_plot_new
 variable uxfig = qualifier_exists("xfig");
#else
 variable uxfig = 0;
#endif
 variable simpars = qualifier("pars",["ppuls","pdot","p2dot","p3dot","porb","ecc","omega","asini","torb0"]);
 ifnot (0.0 <= mres <= 0.45) { vmessage("warning (%s): given residual range must be between 0.0 and 0.45, using the maximum range", _function_name); mres=0.45; }

 % simulate arrival times
 variable s = atime_sim(tmin-1, tmax+1, eph, orb; dn = int(round((tmax-tmin+2)*tpd)));

 % define arrival times and set parameters
 variable id = define_atime(s)[0];

 variable tpl = int(round(0.5*(tmax + tmin)));
 ifnot (uxfig) ()=open_plot(sprintf("%s/cps",psfile));
 variable v;

 % loop over all model parameters
 variable pars = struct_combine(eph, orb);
 foreach (simpars)
 {
   if (uxfig)
   {
#ifexists xfig_plot_new
     variable pl = xfig_plot_new(10,7);
     pl.world(tmin-tpl, tmax-tpl, -mres, mres);
     pl.xlabel(sprintf("Time (MJD - %.2f)", tpl));
     pl.ylabel("Residuals (Pulse Phase)");
     pl.plot([tmin-tpl, tmax-tpl], [0,0]; line=2);
#else
     variable pl = NULL;
     vmessage("%s: xfig module not loaded", _function_name); return;
#endif
   }
   variable par = ();
   ifnot (struct_field_exists(pars,par)) { vmessage("warning (%s): unknown parameter %s", _function_name, par); return; }
   variable value = get_struct_field(pars, par);
   variable parstr = sprintf("arrtimes(%d).%s",id,par);
   % find maximum variation to get the given phase shift 'mres'
   atime_set_ephemeris(id, eph, orb);
   variable maxvar = NULL, minvar = NULL, mm = minr;
   while (maxvar == NULL && minvar == NULL)
   {
     variable nvmn = value*(1.-mm);
     variable nvmx = value*(1.+mm);
     % check on hard limits
     while (get_par_info(parstr).hard_min > nvmn)
     {
%       sprintf("warning (%s): residual range for parameter %s not reached due to its hard limit",_function_name,par);
       minvar = value*(1.-mm/10.);
       nvmn = value*(1.-mm/10.);
     }
     while (get_par_info(parstr).hard_max < nvmx)
     {
%       sprintf("warning (%s): residual range for parameter %s not reached due to its hard limit",_function_name,par);
       maxvar = value*(1.+mm/10.);
       nvmx = value*(1.+mm/10.);
     }
     % check on soft limits
     variable vmin = get_par_info(parstr).min; if (vmin > nvmn) vmin = nvmn - 1e-26;
     variable vmax = get_par_info(parstr).max; if (vmax < nvmx) vmax = nvmx + 1e-26;
     set_par(parstr, value, 0, vmin, vmax);
     % find values
     if (maxvar == NULL) maxvar = find_function_value(&atime_sim_residuals_maxdif, mres, value, nvmx; qualifiers = struct { par = parstr, id = id }, quiet, eps=1e-3);
     if (minvar == NULL) minvar = find_function_value(&atime_sim_residuals_maxdif, mres, nvmn, value; qualifiers = struct { par = parstr, id = id }, quiet, eps=1e-3);
     mm *= 2.0;
   }
   ifnot (maxvar == NULL || minvar == NULL) mm = min(([maxvar,minvar] / value - 1.) * [1., -1.]);
   else if (minvar == NULL) mm = maxvar / value - 1.;
   else mm = 1. + minvar / value;
   atime_set_ephemeris(id,eph,orb);
   ifnot (uxfig) title(sprintf("%s = %e %s",par,value,get_par_info(parstr).units));
   % loop over variations
   _for v (-n, n, 1)
   {
     ifnot (v==0)
     {
       % define color
       variable cs = .19 + abs(v)*(.8/n);
       ifnot (uxfig)
       {
         if (v>0) _pgscr(16+v+n,0,0,cs);
         else _pgscr(16+v+n,cs,0,0);
       }
       % set variation and evaluate the model
       variable newv = value*(1. + mm*v/n);
       vmin = get_par_info(parstr).min; if (vmin > newv) vmin = newv - 1e-26;
       vmax = get_par_info(parstr).max; if (vmax < newv) vmax = newv + 1e-26;
       set_par(parstr, newv, 0, vmin, vmax);
       ()=eval_counts(;fit_verbose=-2);
       % define the label
       variable data = get_data_counts(id);
       variable model = get_model_counts(id);
       variable teph, torb;
       (teph, torb) = atime_get_ephemeris(id);
       variable me = round(mean(data.value));
       variable in = where(model.bin_lo > me + .95*(tmax-tmin) - abs(tmin - tpl))[0];
       variable lx = data.bin_lo[in] - me;
       variable ly = (data.value[in] - model.value[in])*86400 / pulseperiod(data.value[in], teph, torb);
       if (ly<-mres) ly = -0.9*mres;
       if (ly>mres) ly = 0.9*mres;
       % plot residuals
       ifnot (uxfig)
       {
         plot_atime(;style=-1,col=16+v+n,oplot=(v>-n),xrng=[tmin,tmax]-tpl,yrng=[-mres,mres],connect_points,noerrbars);
         color(16+v+n); xylabel(lx,ly,sprintf("%.2e",value*mm*v/n),0.,0.5);
       }
       else
       {
	 variable res = (data.value - model.value)*86400 / pulseperiod(data.value, teph, torb);
	 pl.plot(data.value - tpl, res; color = "#"+rgb2hex(v<0 ? cs : 0, 0, v<0 ? 0 : cs; string));
	 pl.xylabel(0, .8*mres*v/n, sprintf("%.2e",value*mm*v/n), 0, 0; color = rgb2hex(v<0 ? cs : 0, 0, v<0 ? 0 : cs; str));
       }
     }
   }
   if (uxfig) pl.render(sprintf("%s_%s.eps", psfile, par));
 }

 ifnot (uxfig) close_plot;
 delete_data(id);
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_sim_orbitimpact()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_sim_orbitimpact}
%\synopsis{determines the impact of uncertainties of orbital
%  parameters on the pulse ephemeris}
%\usage{atime_sim_orbitimpact(Double_Type tmin, tmax, Struct_Type eph, orb, dorb);}
%\qualifiers{
%    \qualifier{tpd}{number of simulated arrival times per day
%            (default: 10)}
%    \qualifier{pars}{parameters of the pulse ephemeris which
%            are used to fit the simulated data
%            (default: ["ppuls", "pdot", "p2dot",
%                       "p3dot"])}
%    \qualifier{range}{allowed fitting ranges of the parameters
%            given by the 'pars' qualifier in accordant
%            units. The ranges have to be given as an
%            array of alternate min/max values for each
%            parameter in 'pars'}
%}
%\description
%    An uncertainty of an orbital parameter may lead
%    to a wrong pulse ephemeris after a successful
%    fit of pulse arrival times. This function deter-
%    mines the impact of uncertainties on the pulse
%    ephemeris by simulating arrival times in the
%    range 'tmin' and 'tmax' and fitting them within
%    the given orbital uncertainties. The returned
%    structure itself contains structures for each
%    orbital parameter, which contain the fitted
%    pulse ephemeris and therefore the impact of this
%    parameter. The highest phase shift left in the
%    residuals is also returned for each orbital
%    parameter.
%    The given orbital and pulse ephemeris structure
%    must fullfil the conditions described in the
%    check_pulseperiod_orbit_struct function. The uncertainties of
%    the parameters are itself given by a structure,
%    which may either contain double types for the
%    errors or arrays with two elements for the lower
%    and upper errors.
%\seealso{atime_sim, check_pulseperiod_orbit_struct, arrtimes}
%!%-
{
 variable tmin, tmax, eph, orb, dorb;
 switch(_NARGS)
   { case 5: (tmin,tmax,eph,orb,dorb) = (); }
   { help(_function_name()); return; }

 % backwards compatibility
 if (orb != NULL && not struct_field_exists(orb, "torb0")) {
   orb = struct_combine(orb, struct { torb0 = orb.tau });
 }
 if (not struct_field_exists(eph, "tpuls0")) {
   eph = struct_combine(eph, struct { tpuls0 = eph.t0, ppuls = eph.p0 });
 }

 ifnot (check_pulseperiod_orbit_struct(eph)) { vmessage("error (%s): pulse period structure not given properly", _function_name); return; }
 if (orb == NULL) { vmessage("error (%s): orbit structure not given!", _function_name); return; }
 ifnot (check_pulseperiod_orbit_struct(orb)) { vmessage("error (%s): orbit structure not given properly", _function_name); return; }
  
 % check uncertainty structure
 ifnot (typeof(dorb) == Struct_Type) throw RunTimeError, "given uncertainties must be a structure";
 else ifnot (struct_field_exists(dorb, "porb")
          && struct_field_exists(dorb, "torb0")
          && struct_field_exists(dorb, "asini")
          && struct_field_exists(dorb, "ecc")
          && struct_field_exists(dorb, "omega")) throw RunTimeError, "uncertainties are not defined properly";
 ifnot (struct_field_exists(dorb, "pporb")) dorb = struct_combine(dorb, struct { pporb = 0. });

 variable tpd = qualifier("tpd", 10);
 variable fpars = qualifier("pars", ["ppuls", "pdot", "p2dot","p3dot"]);
 variable range = qualifier("range", NULL);

 % simulate arrival times
 variable s = atime_sim(tmin-1, tmax+1, eph, orb; dn = int(round((tmax-tmin+2)*tpd)));
 % define arrival times and set parameters
 variable id = define_atime(s)[0];
 ifnot (range == NULL)
 {
   ifnot (2*length(fpars) == length(range)) throw RunTimeError, "'range' array must be of double length as 'pars' array";
   variable i;
   _for i (0, length(fpars)-1, 1)
     set_par(sprintf("arrtimes(%d).%s", id, fpars[i]), get_struct_field(eph, fpars[i]), 0, range[2*i], range[2*i+1]);
 }

 % loop over all model parameters
 variable impact = struct { porb, torb0, asini, ecc, omega, pporb };
 foreach (get_struct_field_names(orb))
 {
   variable p, stat, par = ();
   variable value = get_struct_field(orb, par);
   variable error = get_struct_field(dorb, par);
   ifnot (typeof(error) == Array_Type) error = [error,error];
   if (length(error) > 2) { vmessage("warning (%s): length of array of uncertainty %s wrong, trimming...", _function_name, par); error = error[[0,1]]; }

   % prepare resulting structure
   set_struct_field(impact, par, struct { ppuls, pdot, p2dot, p3dot, maxshift });
   % freeze all parameters
   freeze(freeParameters);
   _for i (-1, 1, 2)
   {
     atime_set_ephemeris(id, eph, orb);
     if (error[(i+1)/2] > 0.)
     {
       set_par(sprintf("arrtimes(%d).%s", id, par), value + sign(i)*error[(i+1)/2]); % set uncertain parameter
       % thaw pulse ephemeris and fit
       foreach (fpars)
       {
         p = ();
         thaw(sprintf("arrtimes(%d).%s", id, p));
         ()=fit_counts(&stat; fit_verbose=-2);
       }
     }
     % store fitted parameters
     foreach (fpars)
     {
       p = ();
       set_struct_field(get_struct_field(impact, par), p, get_par(sprintf("arrtimes(%d).%s",id,p)));
     }
     % store highest phase shift left
     get_struct_field(impact, par).maxshift = max(abs(get_data_counts(id).value-get_model_counts(id).value))*86400/get_par(sprintf("arrtimes(%d).ppuls", id));
   }
 }

 delete_data(id);
 return impact;
} %}}}

#endif

#endif

%
% Code from src/data/pulsar/arrivaltimes_beta.sl
%

% -*- mode: slang; mode: fold -*-

%%%%%%%%%%%%%%%%%%%%%%%%
define atime_det_beta()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_det_beta}
%\synopsis{Determines the arrival times from a lightcurve using a pulse pattern}
%\usage{Struct_Type atime_det_beta(Struct_Type lc, Struct_Type[] pattern, Double_Type t0[, Double_Type period]);
% or Struct_Type atime_det_beta(String_Type lc, Struct_Type[] pattern, Double_Type t0[, Double_Type period]);}
%\qualifiers{
%    \qualifier{time}{name of the time field in the FITS-file,
%             see fits_read_lc for details}
%    \qualifier{indiv}{determine individual pulses. If no value
%             is assigned, all arrival times are re-
%             turned. Otherwise a given integer sets
%             the number of pulses to average}
%    \qualifier{movem}{if 'indiv' is greater one, so it is aver-
%             aged over a number of arrival times, a
%             moving mean is used to get all individual
%             pulses. Note that the pulses are then not
%             statistically independent!}
%    \qualifier{ccfint}{interpolates the cross correlation of the
%             pulse pattern and the actual analyzed
%             pulse to increase the accuracy. Only works
%             well for clear signals! The number of
%             interpolated bins between each original
%             bin is set to the assigned number
%             (default = 4)}
%    \qualifier{varerr}{estimates the error by a local standard
%             deviation. Therefore the variation of the
%             given number of arrival times is used
%             (default = 10)}
%    \qualifier{mcerr}{estimates the error of each arrival time
%             by performing Monte Carlo simulations. If
%             an integer is assigned it sets the number
%             of runs (default = 10000). Has a higher
%             priority than 'varerr'}
%    \qualifier{mcgaus}{fit a gaussian to the Monte Carlo distribution, if
%             used for error estimation}
%    \qualifier{ccflim}{matches below the given cross-correlation
%             values are skipped (default = 0), allowed
%             range is -1 to 1.}
%    \qualifier{numdif}{maximum allowed deviation from expected
%             pulse number (default = 0.5 phases)}
%    \qualifier{match}{reference to a variable (&var) where the
%             matching pulses are saved as structures
%             similar to the reference profile}
%    \qualifier{slope}{reference to a variable (&var) where the
%             slope of the lightcurve is saved, which
%             is calculated by an interpolation of
%             the mean count rate in the single pulses}
%    \qualifier{skipsl}{do not take the slope of the lightcurve
%             into account}
%    \qualifier{mmnorm}{by default the pulse pattern and the
%             pulses are renormalized by
%               f = (f - mean(f)) / sdev(f)
%             If this qualifier is set, the min/max-
%             normalizaton is used instead
%               f = (f - max(f)) / (max(f) - min(f))}
%    \qualifier{chatty}{boolean value for output messages
%             (default = 1). If set to 2, also echo
%             result of Monte Carlo error estimation}
%    \qualifier{debug}{plot the cross correlation and the matching
%             pulse, echo the found phase shift and
%             sleep the given seconds (default = 1) or,
%             if set to 'user' wait until a key is
%             pressed}
%}
%\description
%    Using one or more pulse pattern the arrival times
%    of pulses in a lightcurve are determined using
%    phase connection. The patterns must be given as a
%    structure with the fields
%      Double_Type bin_lo
%      Double Type bin_hi
%      Double_Type value
%      Double_Type error,
%    as, for example, returned by the 'epfold' function.
%    If multiple patterns are given, the best matching
%    one is used to determine the arrival times.
%    The lightcurve can be passed by a structure with
%    the fields 'time', 'rate', 'error' and 'fracexp',
%    or by the filename to a FITS-file.
%    To determine individual pulses as enabled by the
%    'indiv' qualifier an approximate pulse period must
%    be given (in seconds) to cut the lightcurve into
%    segments were the pulses are searches. Instead of
%    the pulse period an array of two elements defining
%    a range of periods (min/max values, in seconds)
%    may also be given. In that case the 'getVarPeriod'
%    function is used to determine the pulse period and
%    additional qualifiers are passed.
%    If the 'indiv' qualifier is omitted the given
%    lightcurve is folded and the returned phase shift
%    corresponds to the phase shift of the used pattern
%    to the resulting profile.
%    If the method for estimating the uncertainty of
%    the phase shifts is not specified by qualifiers,
%    the default uncertainty is set to one phase bin.
%    The number of bins are derived from the given
%    pulse pattern.
%    The returned structure has the following fields:
%    t0       - array of first time bins of the pulses
%               determined from the lightcurve (MJD)
%    phi      - array of phase shifts of the pulses
%               with respect to the used pattern
%    error    - uncertainties of 'phi'
%    arrnum   - relative pulse number of each arrival
%               time to a specific pulse. Dramatically
%               increases the speed of a fit.
%    numref   - index of the reference pulse, should be
%               zero. Is updated if arrival times are
%               merged.
%    reft0    - the given reference time t0
%    p0       - the given pulse period
%    usedpat  - index of the used pulse pattern in case
%               of multiple ones given
%    The arrival times can be calculated by
%      t = t0 + phi * p0
%    as a first guess, since p0 is a function of time.
%    This fact is handled properly by the fit function
%    'arrtimes'. The function 'atime_calc' calculates
%    the arrival times according to the above equation
%    with respect to a given pulse ephemeris and
%    orbital  parameters.
%\seealso{atime_calc, atime_merge, save_atime, define_atime, arrtimes, pfold}
%!%-
{
 variable lcF, refprof, reft0, per;
 switch(_NARGS)
   { case 4: (lcF,refprof,reft0,per) = (); }
   { help(_function_name()); return; }

 % qualifiers and initialization of variables
 variable t0found = 0;
 variable timefield = qualifier("time","time");
 variable chatty = qualifier("chatty",1);
 if (chatty == NULL) chatty = 1;
 variable indiv = qualifier("indiv", 1);
 if (indiv == NULL) indiv = 1;
 indiv = [qualifier_exists("indiv"), indiv, qualifier_exists("movem") && indiv > 1];
 variable ccflim = qualifier("ccflim", 0.);
 ifnot (-1 <= ccflim <= 1) { vmessage("warning (%s): given ccf limit is outside the allowed range, using default value", _function_name); ccflim = 0.; }
 variable ccfint = qualifier("ccfint", 1);
 if (ccfint == NULL) ccfint = 10;
 variable mcerr = qualifier("mcerr", 0);
 if (mcerr == NULL) mcerr = 10000;
 variable varerr = qualifier("varerr", 0);
 if (varerr == NULL) varerr = 10;
 if (varerr != 0 && varerr < 2) { vmessage("warning (%s): need at least 2 datapoints to determine error by local variance", _function_name); varerr = 2; }
 if (varerr > 0 && mcerr > 0) varerr = 0;
 variable matchret = qualifier("match", NULL);
 if (typeof(matchret) == Ref_Type) % initialise return structure for matches
   @matchret = Struct_Type[0];
 else ifnot (typeof(matchret) == Null_Type) { vmessage("warning (%s): 'match' has to be of Ref_Type", _function_name); matchret = NULL; }
 variable dummy, slope = qualifier("slope", &dummy);
 ifnot (typeof(slope) == Ref_Type || typeof(slope) == Null_Type) { vmessage("warning (%s): 'slope' has to be of Ref_Type", _function_name); slope = &dummy; }
  
 variable debug = qualifier("debug", -1);
 if (debug == NULL) debug = 1;
  
 % load lightcurve if not already given
 variable lc, lco;
 if (typeof(lcF) == String_Type)
 {
   lc = fits_read_lc(lcF;time=timefield);
   % convert fracexp to totalexp
   lc.fracexp = fits_lc_exposure(lcF)*lc.fracexp/86400.;
 } else lc = @lcF; % lc already given
 lco = @lc;

 % sort lc by time
 lc = sort_struct_arrays(lc, "time");

 % check if the period has to be estimated by Epoch Folding
 if (typeof(per) == Array_Type and length(per) == 2)
 {
   if (chatty) message("Determining pulse period of lightcurve");  
   per = getVarPeriod(lc, per[0]/86400, per[1]/86400;; struct_combine(__qualifiers, struct { dt=lc.fracexp }));
   if (chatty) vmessage("  p_lc  = %f s", per*86400);
 }
 else ifnot (typeof(per) == Double_Type || per > 0)
   throw RunTimeError, "pulse period must be a single positive double type, a two element array or a structure";

 % renorm reference profile(s)
 variable p;
 _for p (0, length(refprof)-1, 1)
 {
   if (qualifier_exists("mmnorm"))
   {
     refprof[p].error = refprof[p].error/(max(refprof[p].value)-min(refprof[p].value));
     refprof[p].value = (refprof[p].value-min(refprof[p].value))/(max(refprof[p].value)-min(refprof[p].value));
   }
   else
   {
     variable mom = moment(refprof[p].value);
     refprof[p].error = refprof[p].error / mom.sdev;
     refprof[p].value = (refprof[p].value - mom.ave) / mom.sdev;
   }
 }

 % substract the slope from the lightcurve
 ifnot (qualifier_exists("skipsl"))
 {
   @slope = pulse2pulse_flux_lc(lc, per);
   lc.rate -= (@slope).rate;
 }
  
 variable nums;
 variable nbins = length(refprof[0].value);
 if (indiv[0]) % prepare section to find individual pulses
 {
   % number of pulses in lc to look for
   nums = ceil((max(lc.time)-min(lc.time)) / per * 86400);
   ifnot (indiv[2]) nums = nums / indiv[1]; % reduced number by no moving mean
   nums = int(nums);
   % rebin lightcurve similar to profile if ALL pulses have to be found
   if (indiv[1] == 1) lc = rebin_lc(lc, per / 86400 / nbins);
 }
 else nums = 1;

 % loop over number of pulses (except if using profile only)
 % to get the pulse arrival times
 if (chatty) message(nums == 1 ? "Matching profiles" : "Looking for matches in lightcurve");
 variable i, f, ccf, ndx, j=0, lastphi = NULL, lastnum, ir = NULL;
 variable t0 = Double_Type[0], phi = Double_Type[0], dphi = Double_Type[0];
 variable npuls = Integer_Type[0], refpuls = Integer_Type[0];
 variable usedpat = Integer_Type[0], errdx;
 _for i (0, nums-1, 1)
 {
   % determine arrival times of (each) individual pulses
   if (indiv[0])
   {
     % find actual pulse part of lc (attend moving mean, indiv[2]==1)
     if (indiv[2]) ndx = where(lc.time[0] + per/86400*i <= lc.time < lc.time[0] + per/86400*(i+indiv[1]));
     else ndx = where(lc.time[0] + per/86400*i*indiv[1] <= lc.time < lc.time[0] + per/86400*(i+1)*indiv[1]);
     % only use if number of bins are correct and ALL pulses have to be found
     if (length(ndx) == nbins && indiv[1] == 1)
     {
       t0 = [t0, lc.time[ndx[0]]];
       % extract part
       f = struct { value = lc.rate[ndx], error = lc.error[ndx], bin_lo = lc.time[ndx]-lc.time[ndx[0]], bin_hi = NULL };
       f.bin_hi = make_hi_grid(f.bin_lo);
       j++;
     }
     % average over given number of pulses by folding the lightcurve
     else if (length(ndx) >= nbins && indiv[1] > 1)
     {
       t0 = [t0, lc.time[ndx[0]]];
       f = pfold(lc.time[ndx], lc.rate[ndx], per/86400, lc.error[ndx];; struct_combine(__qualifiers, struct { nbins = nbins, dt=lc.fracexp[ndx], t0=t0[-1] }));
       if (sum(isnan(f.value)) > 0) f = NULL;
       else j++;
     }
     else f = NULL;
   }
   % arrival time of the pulse profile
   else
   {
     t0 = [t0, lc.time[0]];
     % create profile
     f = pfold(lc.time, lc.rate, per/86400, lc.error;; struct_combine(__qualifiers, struct { nbins = nbins, dt=lc.fracexp, t0=t0[-1] }));
   }

   ifnot (f == NULL)
   {
     % renorm profile / lc part
     if (qualifier_exists("mmnorm"))
     {
       f.error = f.error/(max(f.value)-min(f.value));
       f.value = (f.value-min(f.value))/(max(f.value)-min(f.value));
     }
     else
     {
       mom = moment(f.value);
       f.error = f.error / mom.sdev;
       f.value = (f.value - mom.ave) / mom.sdev;
     }
     % loop over all given pattern
     variable ccfmax = -2., newusedpat;
     _for p (0, length(refprof)-1, 1)
     {
       variable tempccf = CCF_1d(refprof[p].value, f.value);
       % assign cross-correlation in case of a better match
       if (max(tempccf) > ccfmax)
       {
	 ccf = @tempccf;
	 ccfmax = max(tempccf);
	 newusedpat = p;
       }
     }
     usedpat = [usedpat, newusedpat];
     if (chatty == 2)
     {
       vmessage("  -> cross-correlation is %.3f", max(ccf));
       if (length(refprof) > 1) vmessage("     with pulse pattern no. %d", usedpat[-1]);
     }
     % check on ccf limit
     if (max(ccf) < ccflim)
     {
       ccf[0] = _NaN;
       message("  -> below the given limit and SKIPPED");
     }
     ifnot (isnan(ccf[0]))
     {
       % get reference pulse number, if not set
       % (is needed if pulse number 0 had less nbins than required)
       if (ir == NULL) ir = i;

       %%% PHASE SHIFT DETERMINATION %%%
       ndx = 0;
       % periodic borders
       ccf = [ccf, ccf[0]];
       if (qualifier_exists("debug"))
       {
	 xrange; xlabel("Phase Shift"); title("Result of Cross Correlation");
	 yrange(-1,1); ylabel("Correlation");
	 connect_points(0);
	 point_style(4);
	 color(1);
	 plot(1.*[0:nbins]/nbins, ccf);
       }
#ifexists gsl->interp_cspline
       if (ccfint > 1)% && t0 != reft0) % interpolate the ccf result to get the position more properly
       {
	 % interpolate
         ccf = gsl->interp_cspline_periodic([0:length(ccf)-1:1./ccfint], [0:length(ccf)-1], ccf);
	 if (qualifier_exists("debug"))
	 {
	   color(4); point_style(-1);
	   connect_points(1);
	   oplot(1.*[0:length(ccf)-1] / (length(ccf) - 1), ccf);
	 }
       }
#endif
       % calculate phase shift
       variable actphi = 1.*where_max(ccf) / (length(ccf)-1);
       if (length(actphi) > 1) actphi = min(actphi);
       if (actphi == 1.) actphi = 0.;
       if (qualifier_exists("debug"))
       {
	 vmessage("  phase shift: %.3f", actphi);
	 color(2); connect_points(1); point_style(-1);
	 oplot([actphi,actphi],[-1,1]);
	 if (typeof(debug) == String_Type && debug == "user") ()=keyinput(;silent, nchr=1);
	 else sleep(debug);
	 color(1); xlabel("Pulse Phase"); xrange;
	 ylabel("Normalized Count Rate"); title("Aligned Pulse and Reference");
	 yrange(min([f.value-f.error,refprof[usedpat[-1]].value-refprof[usedpat[-1]].error]),
		max([f.value+f.error,refprof[usedpat[-1]].value+refprof[usedpat[-1]].error]));
         hplot_with_err(f.bin_lo, f.bin_hi, shift(f.value, -int(actphi*nbins)), f.error);
	 ohplot_with_err(refprof[usedpat[-1]]);
	 if (typeof(debug) == String_Type && debug == "user") ()=keyinput(;silent, nchr=1);
	 else sleep(debug);
       }
       % set relative pulse number
       variable actnum;
       if (indiv[2]) actnum = i-ir;
       else actnum = (i-ir) * indiv[1];
       % add match to returning structure array (if set)
       if (typeof(matchret) == Ref_Type)
         @matchret = [@matchret, struct { bin_lo = f.bin_lo, bin_hi = f.bin_hi, value = shift(f.value, -int(actphi*nbins)),
                                          error = shift(f.error, -int(actphi*nbins)), ccf = max(ccf), tstart = lc.time[ndx[0]], tstop = lc.time[ndx[-1]] }];

       % check if phase shift does not match the pulse number
       ifnot (lastphi == NULL)
       {
	 variable is = actphi - lastphi;
	 variable should = actnum - lastnum;
       	 if (abs(is - should) > qualifier("numdif", .5))
       	 {
       	   if (qualifier_exists("debug")) vmessage("  numdif is %.3f, should be %.3f -> shifting phase", is, should);
       	   actphi = actphi - round(is-should);
       	 }
	 else if (qualifier_exists("debug")) vmessage("  numdif is %.3f, should be %.3f", actphi - lastphi, actnum - lastnum);
       }
       lastphi = actphi;
       lastnum = actnum;
       phi = [phi, actphi];
       npuls = [npuls, actnum];
       refpuls = [refpuls, 0];

       %%% ERROR ESTIMATION %%%
#ifexists gsl->interp_cspline
       if (mcerr > 1 && ccfint > 1) % estimate error using Monte Carlo techniques
       {
	 variable mc, dphis = Double_Type[mcerr];
	 _for mc (0, mcerr-1, 1) % MC loop
	 {
	   % random noisy profile
	   variable fm = f.value + grand(length(f.value)) * f.error;
	   variable refm = refprof[usedpat[-1]].value + grand(length(refprof[usedpat[-1]].value)) * refprof[usedpat[-1]].error;
           % cross correlate
           ccf = CCF_1d(refm,fm);
	   % periodic borders
	   ccf = [ccf, ccf[0]];
	   ccf = gsl->interp_cspline_periodic([0:length(ccf)-1:1./ccfint], [0:length(ccf)-1], ccf);
	   % get shift
	   variable mphi = 1.*where_max(ccf) / (length(ccf)-1);
           if (length(mphi) > 1) mphi = min(mphi);
           if (mphi == 1.) mphi = 0.;
	   dphis[mc] = mphi;
	 }
	 % do a gaussian fit to the distribution
	 if (qualifier_exists("mcgaus"))
	 {
	   variable x_lo = [0.:2.-1./sqrt(mcerr):1./sqrt(mcerr)];
           variable x_hi = x_lo + x_lo[1];
           variable h = histogram(dphis, x_lo, x_hi);
   	   dphi = [dphi, array_fit_gauss(x_lo, h,, x_lo[where_max(h)[0]], moment(dphis).sdev, 1000; frz=[0,0,0,1]).sigma];
	 }
	 else dphi = [dphi, moment(dphis).sdev];

	 if (chatty == 2) vmessage("  -> estimated error = %f phases", dphi[-1]);
	 if (qualifier_exists("debug") && qualifier_exists("mcgaus"))
	 {
	   xrange(0,2); yrange(0,1.1*max(h)/sum(h));
	   xlabel("Phase Shift"); ylabel("Number Density");
	   color(1);
	   hplot(x_lo, x_hi, h/sum(h));
           if (typeof(debug) == String_Type && debug == "user") ()=keyinput(;silent, nchr=1);
	   else sleep(debug);
	 }
       }
       else if (varerr > 0)
#else
       if (varerr > 0) % estimate error by a local mean
#endif
       {
	 % calculate local variance
	 errdx = [((length(phi) - 1) / varerr) * varerr: length(phi) - 1]; % indices of last #varerr-1 number of datapoints
	 if (length(errdx) == varerr)
	 {
 	   dphi = [dphi, moment((phi[errdx] - phi[errdx-1]) / (npuls[errdx] - npuls[errdx-1])).sdev * ones(length(errdx))];
           if (chatty == 2) vmessage("  -> estimated error = %f phases", dphi[-1]);
	 }
       }
       % if no error was estimated: error = 1 phase bin       
       else dphi = [dphi, 1./nbins];
     }
   }
 }
 % set remaining errors for local variance
 if (varerr > 0 && length(dphi) < length(phi))
 {
   variable adphi;
   if (length(errdx) == 1 || length(dphi) == 0) adphi = 1./nbins;
   else if (length(errdx) < varerr/2 && length(dphi) > 0) adphi = dphi[-1];
   dphi = [dphi, adphi * ones(length(phi)-length(dphi))];
   if (chatty == 2) vmessage("  -> estimated error = %f phases", adphi);
 }
 if (indiv[0] and chatty) vmessage("  -> individual pulses (%d/%d)", j, int(nums));

 return struct { t0 = t0, phi = phi, error = dphi, arrnum = npuls, numref = refpuls, reft0 = reft0, p0 = per, usedpat = usedpat };
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_merge_beta()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_merge_beta}
%\synopsis{merges two or more structures of arrival times into one}
%\usage{Struct_Type atime_merge(Struct_Type atime1, Struct_Type atime2);
% or Struct_Type atime_merge(Struct_Type[] atime);}
%\qualifiers{
%    \qualifier{nosort}{do not sort the merged arrival times}
%}
%\description
%    Either both given structures or an array of
%    structures containing arrival times are merged.
%    The fields of the structures must be equal to
%    the ones described in 'atime_det'.
%    The relative pulse numbers and their references
%    are updated to the indices of the new array of
%    arrival times. In addition the arrival times are
%    sorted by time unless the 'nosort' qualifier is
%    given.
%\seealso{atime_det}
%!%-
{
 variable a1, a2 = NULL;
 switch(_NARGS)
   { case 1: (a1) = (); }
   { case 2: (a1,a2) = (); }
   { help(_function_name()); return; }
 if (__is_initialized(&a1)) a1 = @a1;
 ifnot (a2 == NULL) a2 = @a2; 

 % check on correct parameters
 if (a2 == NULL && _typeof(a1) != Struct_Type)
   throw RunTimeError, sprintf("error (%s): need an array of structures", _function_name);
 if (a2 != NULL && (length(a1)+length(a2) > 2 || (typeof(a1) != Struct_Type && _typeof(a2) != Struct_Type)))
   throw RunTimeError, sprintf("error (%s): need two single structures", _function_name);
 
 % merge structures
 variable merg, i, reft0 = NULL;
 if (a2 == NULL) % array given
 {
   if (length(a1) > 1)
   {
     _for i (1, length(a1)-1, 1)
     {
       a1[i].numref += length(a1[i-1].numref); % shift reference numbers
       if (reft0 == NULL) reft0 = a1[i-1].reft0;
       else ifnot (reft0 == a1[i-1].reft0) { vmessage("error (%s): reference times of matching profile are not equal", _function_name); return; }
     }
     merg = merge_struct_arrays(a1);
     merg.reft0 = reft0;
   }
   else merg = a1;
 }
 else % two structures given
 {
   ifnot (__is_initialized(&a1)) merg = a2;
   else ifnot (__is_initialized(&a2)) merg = a1;
   else
   {
     % a2.arrtimes = a2.arrtimes - shift; % correct arrival times
     a2.numref += length(a1.numref); % shift reference numbers
     if (a1.reft0 == a2.reft0) reft0 = a1.reft0;
     else vmessage("warning (%s): reference times of matching profile are not equal", _function_name);
     merg = merge_struct_arrays([a1,a2]);
     merg.reft0 = reft0;
   }
 }
 
 % sort by time
 ifnot (qualifier_exists("nosort"))
 {
   variable ndx = array_sort(merg.t0), nr = @(merg.numref);
   merg.t0 = merg.t0[ndx];
   merg.phi = merg.phi[ndx];
   merg.error = merg.error[ndx];
   merg.arrnum = merg.arrnum[ndx];
   merg.usedpat = merg.usedpat[ndx];
   _for i (0, length(nr)-1, 1) merg.numref[i] = wherefirst(ndx == nr[ndx[i]]); % update reference numbers
 }

 return merg;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define atime_calc_beta()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{atime_calc_beta}
%\synopsis{converts a pulse phaseshift into arrival times}
%\usage{(Double_Type[] arrtimes, errors) = atime_calc(Struct_Type atime, eph[, orb]);
% or (Double_Type[] arrtimes, errors) = atime_calc(Double_Type[] t0, phi, error, Struct_Type eph[, orb]);}
%\description
%    Takes the structure determined by 'atime_det'
%    or an array of times, phase shifts and their
%    errors and returns the pulse arrival times
%      t = t0 + phi * p(t0) [+ z(t0)/c]
%    To calculate the pulse period at each time
%    a pulse ephemeris structure is needed (see
%    'check_pulseperiod_orbit_struct'). If an
%    optional structure containing orbital
%    parameters is given, the Doppler shift delay
%    is included as well.
%\seealso{atime_det, check_pulseperiod_orbit_struct, pulseperiod}
%!%-
{
 variable atime, t0, phi, error, eph, orb = NULL;
 switch(_NARGS)
   { case 2: (atime,eph) = (); }
   { case 3: (atime,eph,orb) = (); }
   { case 4: (t0,phi,error,eph) = (); }
   { case 5: (t0,phi,error,eph,orb) = (); }
   { help(_function_name()); return; }
 if (typeof(atime) == Struct_Type) { t0 = atime.t0; phi = atime.phi; error = atime.error; }
 ifnot (length(t0) == length(phi) == length(error)) { vmessage("%s: t0, phi and error must be of same length", _function_name); return; }

 % backwards compatibility
 if (not struct_field_exists(eph, "p0")) {
   eph = struct_combine(eph, struct { p0 = eph.ppuls, t0 = eph.tpuls0 });
 }
 if (orb != NULL && not struct_field_exists(orb, "tau")) {
   orb = struct_combine(orb, struct { tau = orb.torb0 });
 }
 if (not check_pulseperiod_orbit_struct(eph)) {
   vmessage("error (%s): pulse period is not given properly", _function_name);
 }
 if (orb != NULL && not check_pulseperiod_orbit_struct(orb)) {
   vmessage("error (%s): pulse period is not given properly", _function_name);
 }

 return t0 + phi * pulseperiod(t0, eph, orb)/86400;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define save_atime_beta()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{save_atime_beta}
%\synopsis{saves a structure of arrival times into a FITS-file}
%\usage{save_atime(String_Type filename, Struct_Type atime, String_Type extname[, String_Type[] comments]);}
%\qualifiers{
%    \qualifier{obj}{observed object, written into FITS header}
%    \qualifier{sat}{used satellite, written into extension header}
%    \qualifier{nfold}{number of arrival times, which were merged
%              during the determination. Corresponds to the
%              'indiv' qualifier for the 'atime_det' function.
%              Written into extension header}
%    \qualifier{hfits}{structure for additional FITS-header fields}
%    \qualifier{hext}{structure for additional extension-header fields}
%    \qualifier{newfile}{if the given filename exists a new file is
%              created instead of updating the existing one}
%    \qualifier{newext}{if the given extension already exists in the
%              FITS-file a further one with the same name is
%              added instead of updating the existing one}
%}
%\description
%    A structure of arrival times, e.g. as returned by the
%    'atime_det' function, is save into the given FITS-file
%    creating or updating the extension of the geiven name.
%    Most of the qualifiers allow to store addiotional
%    informations into the header of the FITS-file or the
%    extension. The optional fourth parameter can be used
%    to write additional comments into the FITS-file.
%\seealso{atime_det, atime_merge, load_atime}
%!%-
{
 variable file, atimes, ext, comm;
 switch(_NARGS)
   { case 3: (file,atimes,ext) = (); comm = NULL; }
   { case 4: (file,atimes,ext,comm) = (); }
   { help(_function_name()); return; }
 variable object = qualifier("obj", NULL);
 variable sat = qualifier("sat", NULL);
 variable nfold = qualifier("nfold", NULL);
 variable hfits = qualifier("hfits", NULL);
 variable htabl = qualifier("hext", NULL);

 message("not implemented yet");
 return 0;
}

%
% Code from src/FITS/ascii_read_table.sl
%
define ascii_read_table()
%!%+
%\function{ascii_read_table}
%\synopsis{reads an ASCII table from a file into a structure}
%\usage{Struct_Type table = ascii_read_table(String_Type filename, String_Type formats[]);
%\altusage{Struct_Type table = ascii_read_table(String_Type filename, List_Type infos[]); with  infos[i] = { formats[i], columns[i] };}
%\altusage{(table, keys) = ascii_read_table(String_Type filename, List_Type infos[]);\n
%    % with  infos[i] = { formats[i], columns[i], units[i] };}
%}
%\description
%    The data format of the columns has to be specified as for sscanf; i.e.,
%    %s for strings, %d for decimal integers, %F for double precision floats, ...
%    Lines starting with a comment string ("#" by default; see below) are ignored.
%    The return value is a structure containing the table.
%    Therefore, the column names have to respect the conventions
%    for struct-field names (no special characters as "-"...).
%    If the column name is "", this column is skipped.
%    If a unit is given, the function ascii_read_table has a second return value,
%    namely a keys-structure which can be used for \code{fits_write_binary_table}.
%\qualifiers{
%\qualifier{comment}{string which indicates comments [default = "#"]}
%\qualifier{startline}{Only lines after this number are considered.}
%\qualifier{endline}{The file is not read after this line number.}
%\qualifier{verbose}{}
%}
%\examples
%    \code{variable tab1 = ascii_read_table(filename, [{"%s"}, {"%F"}, {"%F"}]);}\n
%    \code{%} reads \code{String_Type tab1.col1}, \code{Double_Type tab1.col2}, \code{Double_Type tab1.col3} from \code{filename}
%
%    \code{variable tab2 = ascii_read_table(filename, [{"%s","A"}, {"%F","B"}, {"%F","C"}]);}\n
%    \code{%} reads \code{String_Type tab2.A}, \code{Double_Type tab2.B}, \code{Double_Type tab2.C}
%
%    \code{variable tab3 = ascii_read_table(filename, [{"%s","A"}, {"%F",""}, {"%F","C"}]);}\n
%    \code{%} reads \code{String_Type tab3.A}, skips one \code{Double_Type} column and reads \code{Double_Type tab3.C}
%
%    \code{variable tab4, keys;}\n
%    \code{(tab4, keys) = ascii_read_table(filename, [{"%s","A"}, {"%F"}, {"%F","C","u"}]);}\n
%    \code{%} reads \code{String_Type tab4.A} and \code{Double_Type tab4.C} as before. With\n
%    \code{fits_write_binary_table(FITSfilename, "tab4", tab4, keys);}\n
%    \code{%} the unit "u" is assigned to the (second) column C.
%\seealso{sscanf, fits_read_table, fits_write_binary_table, readcol}
%!%-
{
  variable filename, infos;
  switch(_NARGS)
  { case 2: (filename, infos) = (); }
  { help(_function_name()); return; }

  infos = [infos];
  variable comment = qualifier("comment", "#");
  variable startline = qualifier("startline", 1);
  variable endline = qualifier("endline", -1);
  variable verbose = qualifier_exists("verbose");

  variable F = fopen(filename, "r");
  if(F==NULL)  { vmessage("error (%s): file %s cannot be read", _function_name(), filename); return; }

  variable formats, columns, units=NULL;
  variable i, n=length(infos);
  if(typeof(infos[0])==List_Type)
  { variable needunits = 0;
    _for i (0, n-1, 1)
      if(length(infos[i])>2)  { needunits = 1; break; }
    formats = String_Type[n];
    columns = String_Type[n];
    if(needunits)  units = String_Type[n];
    _for i (0, n-1, 1)
    { formats[i] = infos[i][0];
      variable length_infos_i = length(infos[i]);
      if(length_infos_i > 1)  columns[i] = infos[i][1];  else  columns[i] = "col"+string(i+1);
      if(needunits)
        if(length_infos_i > 2)  units[i] = infos[i][2];  else  units[i] = "";
    }
  }
  else
  { formats = infos;
    columns = "col" + array_map(String_Type, &string, [1:n]);
  }

  variable cols_to_use = where(columns != "");
  variable keys;
  if(units!=NULL)
  { variable key_fields = "TUNIT" + array_map(String_Type, &string, [1:length(cols_to_use)]);
    keys = @Struct_Type(key_fields);
    variable j = 1;
    _for i (0, n-1, 1)
      if(columns[i]!="")
      { set_struct_field(keys, key_fields[j-1], units[i]);
	j++;
      }
  }

  variable table = @Struct_Type(columns[cols_to_use]);

  variable line;
  variable line_nr = 0;
  while(fgets(&line, F)!=-1 && (endline<0 || line_nr<endline))
  {
    line_nr++;
    if(line_nr>=startline)
      if(string_match(line, "^"+comment, 1))
      { if(verbose)
	  vmessage("(%s): skipping line # %d starting with %s\n", _function_name(), line_nr, comment);
      }
      else
      { variable buffer = line;
        _for i (0, n-1, 1)
        { variable value=NULL;
	  variable fmt = formats[i]+"%[^\n]";
          if(sscanf(buffer, fmt, &value, &buffer)<2)
          {
	    variable fmtstr = strreplace(strreplace(fmt,
						    "\t", `\t`),
					 "\n", `\n`);
	    variable complete_fmtstr = strreplace(strreplace(strjoin(formats, ""),
							     "\t", `\t`),
						  "\n", `\n`);
	    vmessage(`error (%s): cannot read column %d in line %d,
"%s".
The complete format string is "%s". (comment="%s")
%s was scanning the remaining buffer "%s" for the format "%s".
`,
		     _function_name(), i+1, line_nr, line[[:-2]], complete_fmtstr, comment,
		     _function_name(), buffer, fmtstr);
            return;
          }
	  if(columns[i]!="")
	  {
	    variable a = get_struct_field(table, columns[i]);
	    if(a==NULL)
	      set_struct_field(table, columns[i], {value});
	    else
	      list_append(a, value);
	  }
        }
      }
  }
  ()=fclose(F);

  _for i (0, n-1, 1)
    if(columns[i]!="" && get_struct_field(table, columns[i])!=NULL)
      set_struct_field(table, columns[i], list_to_array(get_struct_field(table, columns[i])));

  return units==NULL ? table : (table, keys);
}

%
% Code from src/slang/structs/assoc_struct_combine.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define assoc_struct_combine()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{assoc_struct_combine}
%\synopsis{associate structure fields, combine structures and fill entries}
%\usage{Struct_Type \code{comb} = assoc_struct_combine(Struct_Type \code{s1}, Struct_Type \code{s2}, String_Type \code{f1}, String_Type \code{f2});}
%\qualifiers{
%\qualifier{double_fill [=_NaN]}{: filler for missing entries in Double_Type fields of \code{s2}}
%\qualifier{float_fill [=_NaN]}{:  filler for missing entries in Float_Type fields of \code{s2}}
%\qualifier{integer_fill [=_0]}{:  filler for missing entries in Integer_Type fields of \code{s2}}
%\qualifier{string_fill [=""]}{:   filler for missing entries in String_Type fields of \code{s2}}
%\qualifier{flag_field_name  [="both_struct_flag"]}{: additional field name for \code{comb}}
%}
%\description
%    This function combines two structures. It uses the field with name \code{f1} of the
%    structure \code{s1} and that with name \code{f2} of \code{s2} to associate the entries with the function
%    get_intersection. The fields of the returned structure \code{comb} have the length of the
%    field \code{f1}. The entries in these fields are filled according to the qualifiers, if
%    there is no association found for this entry.
%    Fields of a type for which no fill qualifier is given are filled with NULL and
%    returned as a list. Fillers can be defined via qualifiers for all variable types.
%    The suffix _fill is required, i.e., for example the qualifier to define a filler
%    for Complex_Type is \code{complex_fill}.
%    WARNING: This function fails for more dimensional fields. If there are fields with
%    the same name in both structures, the field of \code{s2} is used, as it is done by
%    struct_combine. The field \code{f2} is not included.
%    The function adds a field (name can be indentified via qualifier \code{flag_field_name}),
%    which indicates if the entry was found in both structures (1) or if it exists
%    only in \code{f1} (0).
%    
%\examples
%    a = struct{source = ["1", "2", "3"], flux=[0.1, 2.5, 0.7]};
%    b = struct{source = ["3", "2"], redshift=[0.65, 0.03], opt_id = ["Quasar","BL Lac"]};
%    c = assoc_struct_combine (a,b,"source", "source");
%    print_struct(c);
%
%    % without filling fields, but using just common fields this corresponds to:
%    (i1,i2) = get_intersection(a.source,b.source);
%    struct_filter(a,i1); 
%    struct_filter(b,i2);
%    d = struct_combine(a,b);
%    print_struct(d);
%\seealso{get_intersection, struct_combine, struct_filter}
%!%-
{
  variable s1, s2, f1, f2;
  switch(_NARGS)
  { case 4: (s1, s2, f1, f2) = (); }
  { help(_function_name()); return; }

  variable field = NULL;
  field = get_struct_field (s1,f1);
  if (field == NULL) return NULL;

  variable f, i1, i2;
  variable l     = length(field);
  variable flag  = qualifier ("flag_field_name",   "both_struct_flag");
  variable fill   = Assoc_Type[];
  fill["double"]  = _NaN;
  fill["float"]   = _NaN;
  fill["integer"] =    0;
  fill["string"]  =   "";
  variable qlfr = __qualifiers();
  if (qlfr != NULL) {
    foreach f (get_struct_field_names(qlfr))
    { fill[strlow(string(f)[[:-6]])] = get_struct_field(qlfr, f); }  }
  variable fn2 = get_struct_field_names (s2);
  fn2 = fn2[where(fn2 != f2)];
  variable r   = struct_combine (s1, [fn2, flag]);
  (i1, i2) = get_intersection (field, get_struct_field(s2,f2));
  variable t = Integer_Type[l];
  t[i1] = 1;
  set_struct_field(r, flag, t);
  foreach f (fn2)
  {
    variable type = _typeof( get_struct_field(s2, f) );
    variable type_string = strlow(string(type)[[:-6]]);
    if ( all( type_string != assoc_get_keys (fill)) )
    { t = {NULL}; t = t [ Integer_Type[l] ] ; }
    else
    { t = type[l]; t[*] = fill[type_string]; }
    t[i1] = get_struct_field(s2,f)[i2];
    set_struct_field(r, f, t);
  }
  return r;
}


%
% Code from src/slang/arrays/associate_arrays.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define associate_arrays(array1, array2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{associate_arrays}
%\synopsis{computes the smallest pairwise difference between elements of two arrays and returns the indices}
%\usage{Struct_Type result = associate_arrays(Double_Type[n1] A, Double_Type[n2] B);}
%\description
%    For each element \code{A[i]} this function searches the element \code{B[j]}
%    with the smallest difference to \code{A[i]}. It returns the indices \code{j}
%    as \code{result.index[i]} and the corresponding difference as \code{result.diff[i]}.
%\qualifiers{
%\qualifier{nr_closest}{ [=1]: set to N, to obtain the N closest elements.
%                       In this case additionally the index of the second, ..., N-th closest element is
%                       returned as \code{result.index2[*,0], ..., result.index2[*,N-2]},
%                       the same holds for \code{result.diff2}.}
%}
%\seealso{array_sort}
%!%-
{
  variable nc = qualifier("nr_closest", 1);

  variable result = struct {index, diff};
  variable i,j,d,k;
  result.index =  Integer_Type [length(array1), nc];
  result.diff  =  Double_Type  [length(array1), nc];
  result.diff[*,*] = DOUBLE_MAX;

  for (i=0; i<length(array1); i++)
  {
    for (j=0; j<length(array2); j++)
    {
      d = array1[i] - array2[j];

      k = nc-1;
      while ( k >= 0 && abs(d) < abs(result.diff[i,k]) )
        {
          if (k < nc-1)
          {
            result.diff[i,k+1] = result.diff[i,k];
            result.index[i,k+1] = result.index[i,k];
          }
          result.diff[i,k] = d;
          result.index[i,k] = j;
          k--;
        }
    }
  }
  if (nc > 1)
  {
    result = struct_combine(result,"index2","diff2");
    result.index2 = result.index[*,[1:]];
    result.diff2 = result.diff[*,[1:]];
  }
  result.index = result.index[*,0];
  result.diff = result.diff[*,0];
  return result;
}

%
% Code from src/slang/arrays/associate_coords.sl
%
define associate_coords(ra1, dec1, ra2, dec2)
%!%+
%\function{associate_coords}
%\synopsis{computes the smallest pairwise angular separation between arrays of positions and returns the indices}
%\usage{Struct_Type result = associate_coords(Double_Type[n1] RA1, Double_Type[n1] DEC1),Double_Type[n2] RA2, Double_Type[n2] DEC2);}
%\qualifiers{
%\qualifier{nr_closest}{[=1]      set to N, to obtain the N closest elements.
%                            In this case additionally the index of the second, ..., N-th closest element is
%                            returned as \code{result.index2[*,0], ..., result.index2[*,N-2]},
%                            the same holds for \code{result.sep2}.}
%\qualifier{unit}{[\code{="deg"}]        unit of the angular coordinates}
%\qualifier{alpha1_unit}{[\code{="deg"}] unit of the RA1}
%\qualifier{alpha2_unit}{[\code{="deg"}] unit of the RA2}
%\qualifier{delta1_unit}{[\code{="deg"}] unit of the DEC1}
%\qualifier{delta2_unit}{[\code{="deg"}] unit of the DEC2}
%}
%\description
%    For each position \code{(RA1[i], DEC1[i])} this function searches the element \code{(RA2[j],DEC2[j])}
%    with the smallest difference to \code{(RA1[i],DEC[i])}. It returns the indices \code{j}
%    as \code{result.index[i]} and the corresponding separation in rad as \code{result.sep[i]}.
%    The units of each coordinate can be set with the qualifier \code{unit}, but also independently,
%    as for the function \code{greatcircle_distance}.
%\seealso{greatcircle_distance, associate_arrays}
%!%-
{
  variable q = __qualifiers();
  variable nc = qualifier("nr_closest", 1);

  variable result = struct {index, sep};
  variable i,j,d,k;
  result.index =  Integer_Type [length(ra1), nc];
  result.sep  =  Double_Type  [length(ra1), nc];
  result.sep[*,*] = DOUBLE_MAX;

  for (i=0; i<length(ra1); i++)
  {
    for (j=0; j<length(ra2); j++)
    {
      d = greatcircle_distance(ra1[i], dec1[i], ra2[j], dec2[j];; q);

      k = nc-1;
      while ( k >= 0 && abs(d) < abs(result.sep[i,k]) )
        {
          if (k < nc-1)
          {
            result.sep[i,k+1] = result.sep[i,k];
            result.index[i,k+1] = result.index[i,k];
          }
          result.sep[i,k] = d;
          result.index[i,k] = j;
          k--;
        }
    }
  }
  if (nc > 1)
  {
    result = struct_combine(result,"index2","sep2");
    result.index2 = result.index[*,[1:]];
    result.sep2 = result.sep[*,[1:]];
  }
  result.index = result.index[*,0];
  result.sep = result.sep[*,0];
  return result;
}

%
% Code from src/misc/atomic/atom_name.sl
%
define atom_name()
%!%+
%\function{atom_name}
%\synopsis{DEPRECATED}
%\usage{String_Type atom_name(Integer_Type Z)}
%\description
% This function returns the symbol for atoms with proton number Z.
% This function is DEPRECATED, please use element_symbol.
%\seealso{element_symbol}
%!%-
{
    return element_symbol(();;__qualifiers);
}

%
% Code from src/misc/atomic/atomdb_v3.sl
%
%% needs to be set due to many more lines in versions greater than 3.0.0
%% http://www.atomdb.org/isis_and_atomdb3.pdf
%% (variable is already defined inside isis)
EM_Hash_Table_Size_Hint=500000;
%
% Code from src/sixte/attitude_lissajous_pattern.sl
%
define attitude_lissajous_pattern(prefix,expos,ra,dec){ %{{{
%!%+
%\function{attitude_lissajous_pattern}
%\synopsis{creates an Attitude File for a Lissajous pattern}
%\usage{attitude_lissajous_pattern(prefix,expos,ra,dec)}
%!%-

   %% attitude in RA, Dec given in degrees
   %% choose an amplitude of 15arcsec by default
   variable ampl = (qualifier("amplitude",15)*2/3600.);
   
   variable li = lissajous_pattern(;amplitude=ampl,tstop=expos,
				   x0=atof(ra),y0=atof(dec));
   
   variable att = struct{time=li.t*1.0001,ra=li.x,dec=li.y};
   variable skeys = struct{
      MJDREF=qualifier("MJDREF",52000.00),
      TSTART=min(att.time),
      TSTOP =max(att.time),
      ALIGNMEN=qualifier("alignment","NORTH"),
      origin="ECAP"
   };
   
   variable file = prefix+".att";
   
   fits_write_binary_table(file,"ATT",att,skeys);
   
   vmessage(" writing Attitude file %s",file);

   return file;
}
%}}}


%
% Code from src/slang/trees/avltree.sl
%
%
% Implementation of binary search and AVL trees
%

% Basic type definitions
private variable TREE_LEFT=0;
private variable TREE_RIGHT=1;

%
% basic key compare function
% (can be overridden if the tree contains more complex data structures)
%
private define tree_key_compare(k1,k2) {
    return (k1 < k2) ? -1 : ( (k1==k2) ? 0 : +1);
}

%
% Binary tree specific functions and structures
%
private define new_BinarySearchTreeNode() {
    variable T;
    variable data=NULL;
    if (_NARGS==1) {
        T=();
    }
    if (_NARGS==2) {
        (T,data)=();
    }
    
    return struct {     % node of a binary search tree
        child={NULL,NULL}, % value is NULL if child is empty
        data=data
    };
}

private define BinarySearchTreeInsert_helper(T,key) {
    % insert a node containing key into the binary tree T
    % this could also be programmed recursively, but iteratively is
    % faster. This returns the newly inserted key.
    
    % use new_node function such that this will also work on
    % binary search trees derived from the BinarySearchTree
    variable N=T.new_node(key);

    % empty tree: insert at root and be done with it
    if (typeof(T.root)!=Struct_Type) {
        T.root=N;
        return N;
    }

    % existing tree - iterative traversal to correct position
    variable no=T.root;
    while (1) {
        variable cmp=T.key_compare(key,no.data);
        % key < no.data ?
        if (cmp<0) {
            % empty child? - if yes, then insert
            if (no.child[TREE_LEFT]==NULL) {
                no.child[TREE_LEFT]=N;
                return N;
            }
            no=no.child[TREE_LEFT];
        } else {
            % key > no.data ?
            if (cmp>0) {
                % empty child? - if yes, then insert
                if (no.child[TREE_RIGHT]==NULL) {
                    no.child[TREE_RIGHT]=N;
                    return N;
                }
                no=no.child[TREE_RIGHT];
            } else {
                % key already exists, throw an error
                % FIXME: add graceful exit option
                throw UsageError,sprintf("insert: Key %S already exists!\n",key);
            }
        }
    }

    % should never reach this point
    throw UsageError,sprintf("This should never happen\n");
}

private define BinarySearchTreeInsert(T,key) {
    % insert a node containing key into the binary tree T
    ()=BinarySearchTreeInsert_helper(T,key);
}

private define BinarySearchTreeMinimum(N) {
    % return node with the smallest key in a binary search tree
    % if qualifier "node" is set returns the node, otherwise the key
    %
    % N can be a node or a tree

    % case of tree
    if (struct_field_exists(N,"root")) {
        N=N.root;
    }

    while (N.child[TREE_LEFT]!=NULL) {
        N=N.child[TREE_LEFT];
    }
    if (qualifier_exists("node")) {
        return N;
    }
    return N.data;
}

private define BinarySearchTreeMaximum(N) {
    % return the largest key in a binary search tree
    % if qualifier "node" is set returns the node, otherwise the key

    % case of tree
    if (struct_field_exists(N,"root")) {
        N=N.root;
    }

    while (N.child[TREE_RIGHT]!=NULL) {
        N=N.child[TREE_RIGHT];
    }
    if (qualifier_exists("node")) {
        return N;
    }
    return N.data;
}

% iterative deletion of a key in a binary search tree
private define BinarySearchTreeDeleteKey_helper_iterative(T,key,N);
private define BinarySearchTreeDeleteKey_helper_iterative(T,key,N) {
    %
    % deletion of node with key K from binary tree (version 1)
    %
    if (N==NULL) {
        return NULL;
    }

    %
    % traverse tree
    %
    variable cmp=T.key_compare(key,N.data);
    variable parent;
    while (N!=NULL && cmp!=0) {
        parent=N;
        if (cmp<0) {
            N=N.child[TREE_LEFT];
        } else {
            N=N.child[TREE_RIGHT];
        }

        % not found
        if (N==NULL) {
            return N;
        }
        cmp=T.key_compare(key,N.data);        
    }

    % N is now the key to be deleted
    
    if (N.child[TREE_LEFT]==NULL) {
        if (N.child[TREE_RIGHT]==NULL) {
            N=NULL; % kill data
            return N;
        }
        N=N.child[TREE_RIGHT];
        return N;
    }

    if (N.child[TREE_RIGHT]==NULL) {
        N=N.child[TREE_LEFT];
        return N;
    }

    % node has two children: replace current node with successor data
    % and delete successor
    
    % find successor node
    variable succ=BinarySearchTreeMinimum(N.child[TREE_RIGHT];node);
    % copy successor data
    N.data=succ.data;

    % delete inorder successor
    N.child[TREE_RIGHT]=BinarySearchTreeDeleteKey_helper_iterative(T,N.data,N.child[TREE_RIGHT]);
    
    return N;
}

%
% full recursive tree deletion (needed for the AVL tree)
%
private define BinarySearchTreeDeleteKey_helper_recursive(T,key,N);

private define BinarySearchTreeDeleteKey_helper_recursive(T,key,N) {
    % no node at current position, end recursion
    if (N==NULL) {
        return NULL;
    }

    % recursive deletion
    variable cmp=T.key_compare(key,N.data);
    if (cmp<0) {
        % key < N.data
        N.child[TREE_LEFT]=BinarySearchTreeDeleteKey_helper_recursive(T,key,N.child[TREE_LEFT]);
    } else {
        if (cmp>0) {
            % key > N.data
            N.child[TREE_RIGHT]=BinarySearchTreeDeleteKey_helper_recursive(T,key,N.child[TREE_RIGHT]);
        } else {
            % all ifs below: found it!
            %
            
            if (N.child[TREE_LEFT]==NULL && N.child[TREE_RIGHT]==NULL) {
                % no child present, just delete the thing
                N=NULL;
            } else {
                if (N.child[TREE_LEFT]==NULL) {
                    N=N.child[TREE_RIGHT];
                } else {
                    if (N.child[TREE_RIGHT]==NULL) {
                        N=N.child[TREE_LEFT];
                    } else {
                        % delete node with two children

                        % find inorder successor of current node
                        variable succ=BinarySearchTreeMinimum(N.child[TREE_RIGHT];node);
                        % save the data of the successor
                        N.data=succ.data;
                        % delete successor
                        N.child[TREE_RIGHT]=BinarySearchTreeDeleteKey_helper_recursive(T,N.data,N.child[TREE_RIGHT]);
                    }
                }
            }
        }
    }
    return N;
}

private define BinarySearchTreeDeleteKey(T,key) {
    ()=BinarySearchTreeDeleteKey_helper_iterative(T,key,T.root);
}

private define BinarySearchTreeDeleteNode(T,N) {
    ()=BinarySearchTreeDeleteKey_helper_iterative(T,N.key,T.root);
}

% find node for which data==key
% returns the Node of Null, if the key does not exist
private define BinarySearchTreeSearch(T,key) {
    variable N=T.root;
    while (N!=NULL && N.data != key ) {
        if (key < N.data) {
            N=N.child[TREE_LEFT];
        } else {
            N=N.child[TREE_RIGHT];
        }
    }

    return N;
}

% return true if key is contained in T
private define BinarySearchTreeKeyExists(T,key) {
    variable N=BinarySearchTreeSearch(T,key);
    return (N!=NULL);
}

private define BinarySearchTreeNodeSuccessor(T,N)
% find successor node of N in tree T
% (T isn't really needed here, but this helps the tree object)
{
    if (N.child[TREE_RIGHT]!=NULL) {
        return BinarySearchTreeMinimum(N.child[TREE_RIGHT];node);
    }
    variable y=N.parent;
    while (y!=NULL && N==y.child[TREE_RIGHT]) {
        N=y;
        y=y.parent;
    }
    return y;
}

private define BinarySearchTreeNodePredecessor(T,N)
% find predecessor node of N in tree T
% (T isn't really needed here, but this helps the tree object)
{
    if (N.child[TREE_LEFT]!=NULL) {
        return BinarySearchTreeMinimum(N.child[TREE_LEFT];node);
    }
    variable y=N.parent;
    while (y!=NULL && N==y.child[TREE_LEFT]) {
        N=y;
        y=y.parent;
    }
    return y;
}

private define BinarySearchPrintNode(N) {
    ()=printf("%S\n",N.data);
}

private define BinarySearchTraversal();

private define BinarySearchTraversal() {
    variable N,func;

    if (_NARGS==1) {
        N=();
        func=&BinarySearchPrintNode;
    }
    
    if (_NARGS==2) {
        (N,func)=();
    }

    if (N==NULL) {
        return;
    }

    if (struct_field_exists(N,"root")) {
        N=N.root;
        if (N==NULL) {
            return;
        }
    }
    
    if (qualifier_exists("preorder")) {
        (@func)(N;;__qualifiers());
        BinarySearchTraversal(N.child[TREE_LEFT],func;;__qualifiers());
        BinarySearchTraversal(N.child[TREE_RIGHT],func;;__qualifiers());
        return;
    }

    if (qualifier_exists("postorder")) {
        BinarySearchTraversal(N.child[TREE_LEFT],func;;__qualifiers());
        BinarySearchTraversal(N.child[TREE_RIGHT],func;;__qualifiers());
        (@func)(N;;__qualifiers());
        return;
    }

    % default is inorder traversal
    BinarySearchTraversal(N.child[TREE_LEFT],func;;__qualifiers());
    (@func)(N;;__qualifiers());
    BinarySearchTraversal(N.child[TREE_RIGHT],func;;__qualifiers());
    return;

}

private define BinarySearchTreeNodePrint();

private define BinarySearchTreeNodePrint() {
    variable N,sp;

    (N,sp)=();

    if (N==NULL) {
        return;
    }

    BinarySearchTreeNodePrint(N.child[TREE_RIGHT],sp+"    ");
    if (struct_field_exists(N,"height")) {
        ()=printf("%s%S-%i\n",sp,N.data,N.height);
    } else {
        ()=printf("%s%S\n",sp,N.data);
    }
    BinarySearchTreeNodePrint(N.child[TREE_LEFT],sp+"    ");
}

private define BinarySearchTreePrint(T) {
    BinarySearchTreeNodePrint(T.root,"");
}

private variable BinarySearchTree_Struct=struct {
    root=NULL,
    insert=&BinarySearchTreeInsert,
    print=&BinarySearchTreePrint,
    new_node=&new_BinarySearchTreeNode,
    key_compare=&tree_key_compare,
    search=&BinarySearchTreeSearch,
    delete=&BinarySearchTreeDeleteKey,
    delete_node=&BinarySearchTreeDeleteNode,
    exists=&BinarySearchTreeKeyExists,
    min=&BinarySearchTreeMinimum,
    max=&BinarySearchTreeMaximum,
    successor=&BinarySearchTreeNodeSuccessor,
    predecessor=&BinarySearchTreeNodePredecessor,
    traversal=&BinarySearchTraversal
};

define BinarySearchTree()
%!%+
%\function{BinarySearchTree}
%\synopsis{Creates a new binary search tree}
%\usage{Struct_Type=BinarySearchTree();}
%\altusage{Struct_Type=BinarySearchTree(cmpfunction);}
%\description
%A binary search tree is an object that permits fast
%searches and ordered operations on ordinal data types (i.e.,
%data that can be sorted). 
%
%The tree consists of nodes which are structs containing
%the data (or "key") in tag "data" and references to nodes with
%keys with values less than data in child[0] and to keys with
%values larger than data in child[1]. In general, users of the
%data structure should not worry about the structure of the
%nodes and use the functions described below for all tree
%operations.
%
%Searches in a binary search tree can be fast (O(log N)) if the
%search tree is balanced, that is, if there is a similar number
%of nodes below each node (if that makes sense...). This is typically
%the case if random (unsorted) data are inserted into the tree, and
%very much not the case if ordered data are inserted. In this case
%the search degenerates to O(N). The isisscripts provide a special
%binary search tree called AVLTree that has the same accessor functions
%as BinarySearchTree but ensures that the tree is height-balanced (at some
%small additional cost for key insertions and deletions).
%
%The BinarySearchTree and the AVLTree provide the following functions:
%  insert(key) - insert a data element key into the binary search tree
%  delete(key) - remove a data element key from the binary search tree
%  exists(key) - return true if the search tree contains key
%  search(key) - return the node containing the key (or NULL if
%                they key does not exist)
%  min         - return the minimum key of the search tree (or a subtree), or
%                the node containing the minimum key if the node qualifier
%                is set.
%  max         - return the maximum key of the search tree (or a subtree), or
%                the node containing the maximum key if the node qualifier
%                is set.
%  successor   - return the successor node to a node (i.e., the node
%                containing the next largest data value in the tree).
%  predecessor - return the predecessor node to a node (i.e., the node
%                containing the next smaller data value in the tree).
%  print       - print the structure of the tree
%  traversal(&func;quals) - traverse the tree, i.e., execute func for all
%                keys of the tree. The function func is called with each
%                node N (i.e., the data are found in N.data) and all
%                qualifiers given to traversal. The order in which the
%                nodes are traversed is defined by the following qualifiers:
%                preorder: operate on the node, then on the left children,
%                          then on the right children
%                postorder: operate first on the left children, then on the
%                          right children, then on the node
%                inorder (the default): operate on the left children, then
%                          on the node, then on the right children
%                Note that it is permitted that func modifies the tree (e.g.,
%                insert new elements, delete tree elements and so on; it is
%                best practice to use a qualifier that specifies the tree to
%                let func know about it). Depending on the traversal order
%                the modified elements may or may not be operated on as part
%                of the traversal.
%
%
%The default setup of the binary search tree works on all data types
%where the operators <, ==, and > have been defined. For other data
%types, initialize the tree with a reference to a comparison function
%of the style define key_compare(k1,k2) where k1 and k2 are of the
%data type contained in the search tree and which returns a negative
%number if k1<k2, 0 if k1==k2, and a positive number if k1>k2 (this
%is the same definition as the one used by s-lang's sort function or
%by the strcmp function).
%
%\example
%   % sort some numbers
%   % (obviously a better way would be array_sort...)
%
%   define arrapp(N) {
%     % append key of N to list
%     variable list=qualifier("list");
%     list_append(list,N.data);
%   }
%    
%   variable arr=[1,4,6,3,5,7,8,2,9];
%   variable t=BinarySearchTree();
%   variable i;
%   foreach i (arr) {
%     t.insert(i);
%   }
%    
%   % print the tree
%   t.print();
%
%   % delete element with key 8
%   t.delete(8);
%
%   variable ll={};
%   t.traversal(&arrapp;list=ll);
%   foreach i (ll) {
%     print(i);
%   }
%   % print the minimum and maximum values
%   ()=printf("min: %S - max: %S\n",t.min(),t.max());
%
%\example
%   % setup a tree for strings
%   variable arr=["A","X","D","B","C"];
%   variable t=AVLTree(&strcmp);
%   variable i;
%   foreach i (arr) {
%      t.insert(i);
%   }
%   t.print();
%   
%
%\seealso{AVLTree,sort}
%!%-
{
    variable T=@BinarySearchTree_Struct;

    if (_NARGS==1) {
        variable cmpf=();
        T.key_compare=cmpf;
    }
    return T;
}


%
% AVL node specific functions and structures
%
private define new_AVLNode() {
    variable T;
    variable data=NULL;
    if (_NARGS==1) {
        T=();
    }
    if (_NARGS==2) {
        (T,data)=();
    }

    variable N=new_BinarySearchTreeNode(T,data);
    return struct_combine(N,struct{height=0});
}

private define AVLNodeHeight(N) {
    % get the node height
    if (N==NULL) {
        return -1;
    }
    return N.height;
}

private define AVLNodeUpdateHeight(N) {
    variable lh=AVLNodeHeight(N.child[TREE_LEFT]);
    variable rh=AVLNodeHeight(N.child[TREE_RIGHT]);
    if (lh>rh) {
        N.height=lh+1;
    } else {
        N.height=rh+1;
    }
}

private define AVLNodeBalanceFactor(N) {
    return AVLNodeHeight(N.child[TREE_RIGHT])-AVLNodeHeight(N.child[TREE_LEFT]);
}

private define AVLTreeRotate(N,dir) {
    % rotate tree below N by dir (where dir==TREE_LEFT or TREE_RIGHT)
    variable sav=N.child[1-dir];

    N.child[1-dir]=sav.child[dir];
    sav.child[dir]=N;

    AVLNodeUpdateHeight(N);
    AVLNodeUpdateHeight(sav);
    
    return sav;
}

private define AVLTreeRebalance(N) {
    variable bf=AVLNodeBalanceFactor(N);

    % Left heavy?
    if (bf < -1 ) {
        if (AVLNodeBalanceFactor(N.child[TREE_LEFT])<=0) {
            N=AVLTreeRotate(N,TREE_RIGHT);
        } else {
            N.child[TREE_LEFT]=AVLTreeRotate(N.child[TREE_LEFT],TREE_LEFT);
            N=AVLTreeRotate(N,TREE_RIGHT);
        }
        
        % if (AVLNodeBalanceFactor(N.child[TREE_LEFT])>0) {
        %     N.child[TREE_LEFT]=AVLTreeRotate(N.child[TREE_LEFT],TREE_LEFT);
        % }
        % N=AVLTreeRotate(N,TREE_RIGHT);
    }

    % right heavy?
    if (bf>1) {
        if (AVLNodeBalanceFactor(N.child[TREE_RIGHT])>=0) {
            N=AVLTreeRotate(N,TREE_LEFT);
        } else {
            N.child[TREE_RIGHT]=AVLTreeRotate(N.child[TREE_RIGHT],TREE_RIGHT);
            N=AVLTreeRotate(N,TREE_LEFT);
        }
        % if (AVLNodeBalanceFactor(N.child[TREE_RIGHT])<0){
        %     N.child[TREE_RIGHT]=AVLTreeRotate(N.child[TREE_RIGHT],TREE_RIGHT);
        % }
        % N=AVLTreeRotate(N,TREE_LEFT);
    }

    return N;
    
}

private define AVLTreeInsert_helper(T,key,no);

private define AVLTreeInsert_helper(T,key,no) {
    % insert node N into binary tree no in tree T
    if (no==NULL) {
        return T.new_node(key);
    }
    
    variable cmp=T.key_compare(key,no.data);
    if (cmp<0) {
        % key < no.data
        no.child[TREE_LEFT]=AVLTreeInsert_helper(T,key,no.child[TREE_LEFT]);
    } else {
        if (cmp>0) {
            no.child[TREE_RIGHT]=AVLTreeInsert_helper(T,key,no.child[TREE_RIGHT]);
        } else {
            % equal - not permitted, so throw an error
            throw UsageError,sprintf("insert: Key %S already exists!\n",key);
        }
    }
    
    AVLNodeUpdateHeight(no);

    return AVLTreeRebalance(no);
}

private define AVLTreeInsert (T,key) {
    % insert a node containing key into the binary tree T
    if (T.root==NULL) {
        T.root=T.new_node(key);
        return;
    }
    T.root=AVLTreeInsert_helper(T,key,T.root);
}


private define AVLTreeDeleteKey_helper(T,key,root) {

    variable N=BinarySearchTreeDeleteKey_helper_recursive(T,key,root);
    if (N==NULL) {
        return root;
    }
    AVLNodeUpdateHeight(N);
    AVLTreeRebalance(N);
}

private define AVLTreeDeleteKey(T,key) {
    T.root=AVLTreeDeleteKey_helper(T,key,T.root);
}

private define AVLTreeDeleteNode(T,N) {
    T.root=AVLTreeDeleteKey_helper(T,N.key,T.root);
}

private variable AVLTree_Struct=struct {
    root=NULL,
    insert=&AVLTreeInsert,
    print=&BinarySearchTreePrint,
    new_node=&new_AVLNode,
    key_compare=&tree_key_compare,
    search=&BinarySearchTreeSearch,
    delete=&AVLTreeDeleteKey,
    delete_node=&AVLTreeDeleteNode,
    exists=&BinarySearchTreeKeyExists,
    min=&BinarySearchTreeMinimum,
    max=&BinarySearchTreeMaximum,
    successor=&BinarySearchTreeNodeSuccessor,
    predecessor=&BinarySearchTreeNodePredecessor,
    traversal=&BinarySearchTraversal
};

define AVLTree()
%!%+
%\function{AVLTree}
%\synopsis{Creates a new balanced binary search tree}
%\usage{Struct_Type=AVLTree();}
%\altusage{Struct_Type=AVLTree(cmpfunction);}
%\description
%A binary search tree is an object that permits fast
%searches and ordered operations on ordinal data types (i.e.,
%data that can be sorted).
%
%An AVLTree is a balanced binary search tree, that is, a search
%tree with a structure that ensures that searches will always
%be close to O(log(N)), where N is the number of elements in the
%tree.
%
%The interface of AVLTree is identical to that of BinarySearchTree,
%see there for a description of the member functions.
%
%\seealso{BinarySearchTree}
%!%-
{
    variable T=@AVLTree_Struct;

    if (_NARGS==1) {
        variable cmpf=();
        T.key_compare=cmpf;
    }
    return T;
}


% %
% % Red Black Tree specific functions
% %

% private define RBRotateDirRoot(T,P,dir) {
%     % T  : red-black tree
%     % P  : root of subtree (may be the root of T)
%     % dir: { TREE_LEFT, TREE_RIGHT }
%     variable G=P.parent;
%     variable S=P.child[1-dir];

%     if (typeof(S)!=Struct_Type) {
%         throw UsageError,"Need a true node!";
%     }

%     variable C = S.child[dir];
%     P.child[1-dir]=C;
    
%     if (typeof(C)==Struct_Type) {
%         C.parent=P;
%     }
    
%     S.child[dir]=P;
%     P.parent=S;
%     S.parent=G;

%     if (typeof(G)==Struct_Type) {
%         G.child[ (P == G.child[TREE_RIGHT]) ? TREE_RIGHT : TREE_LEFT ] = S;
%     } else {
%         T.root=S;
%     }
%   return S; % new root of subtree
% }

% define RBinsert1(T,N,P,dir) {
%     % T: red-black tree
%     % N: node to be inserted
%     % P: parent node of N (may be NULL)
%     % dir: side ( TREE_LEFT or TREE_RIGHT ) of P where to insert N

%     variable G; % parent node of P
%     variable U; % uncle of N

%     N.color = RED;
%     N.child[TREE_LEFT]  = NULL;
%     N.child[TREE_RIGHT] = NULL;
%     N.parent = P;
    
%     if (typeof(P) != Struct_Type ) {   % There is no parent
%         T.root = N;    % N is the new root of the tree T.
%         return; % insertion complete
%     }

%     P.child[dir]=N;% insert N as dir-child of P
%     % start of the (do while)-loop:
%     do {
%         if (P.color == BLACK) {
%             % Case_I1 (P black):
%             return; % insertion complete
%         }
%         % From now on P is red.
%         G=P.parent;
%         if (typeof(G) != Struct_Type) {
%             % P red and root
%             P.color = BLACK;
%             return; % insertion complete
%         }
%         % else: P red and G!=NULL.
%         dir = childDir(P); % the side of parent G on which node P is located
%         U = G.child[1-dir]; % uncle
%         if (typeof(U) != Struct_Type || U.color == BLACK) { % considered black
%             if (N == P.child[1-dir]) {
%                 % Case_I5 (P red && U black && N inner grandchild of G):
%                 RBRotateDirRoot(T,P,dir); % P is never the root
%                 N = P; % new current node
%                 P = G.child[dir]; % new parent of N
%                 % fall through to Case_I6
%             }
%             %Case_I6 (P red && U black && N outer grandchild of G):
%             RBRotateDirRoot(T,G,1-dir); % G may be the root
%             P.color = BLACK;
%             G.color = RED;
%             return; % insertion complete
%         }
         
%         % Case_I2 (P+U red):
%         P.color = BLACK;
%         U.color = BLACK;
%         G.color = RED;
%         N = G; % new current node
%         % iterate 1 black level higher
%         %   (= 2 tree levels)
%         P=N.parent;
%     } while (typeof(P) == Struct_Type);
%     % end of the (do while)-loop
%     % Leaving the (do while)-loop (after having fallen through from Case_I2).
        
%     % Case_I3: N is the root and red.
%     return; % insertion complete

% } % end of RBinsert1

% define RBdelete(T,N) {
%     %
%     % delete node N from RB tree T
%     %

%     %
%     % simple root?
%     %
%     if (N==T.root) {
%         if (typeof(N.child[TREE_LEFT])!=Struct_Type && typeof(N.child[TREE_RIGHT])!=Struct_Type) {
%             % tree consists of one node only
%             T.root=NULL;
%             return;
%         }

%         % now at least one child exists
% %%%        if (typeof(XXX

%     }

% }

% define RBdelete2(T,N) {
%     %
%     % treatment of complex delete cases
%     %
%     % T red-black tree
%     % N node to be deleted

%     variable P = N.parent;  % -> parent node of N
%     variable S;  % sibling of N
%     variable C; %  close   nephew
%     variable D; % distant nephew

%     % P != NULL, since N is not the root.
%     variable dir = childDir(N); % side of parent P on which the node N is located
%     % Replace N at its parent P by NULL
%     P.child[dir] = NULL;

%     % start of the (do while)-loop:
%     do {
%         dir = childDir(N);   % side of parent P on which node N is located
%         S = P.child[1-dir]; % sibling of N (has black height >= 1)
%         D = S.child[1-dir]; % distant nephew
%         C = S.child[  dir]; % close   nephew
%         if (S.color == RED) {
%             % Case_D3: S red && P+C+D black:
%             RBRotateDirRoot(T,P,dir); % P may be the root
%             P.color = RED;
%             S.color = BLACK;
%             S = C; % != NIL
%             % now: P red && S black
%             D = S.child[1-dir]; % distant nephew
%             if (typeof(D) == Struct_Type && D.color == RED) {
%                 % Case_D6: % D red && S black:
%                 RBRotateDirRoot(T,P,dir); % P may be the root
%                 S.color = P.color;
%                 P.color = BLACK;
%                 D.color = BLACK;
%                 return; 
%             }
%             C = S.child[  dir]; % close   nephew
%             if (typeof(C) == Struct_Type && C.color == RED) {
%                 % Case_D5: % C red && S+D black:
%                 RBRotateDirRoot(T,S,1-dir); % S is never the root
%                 S.color = RED;
%                 C.color = BLACK;
%                 D = S;
%                 S = C;
%                 RBRotateDirRoot(T,P,dir); % P may be the root
%                 S.color = P.color;
%                 P.color = BLACK;
%                 D.color = BLACK;
%                 return; 
%             }
%             % Otherwise C+D considered black.
%             S.color = RED;
%             P.color = BLACK;
%             return; 
%         }
%         % S is black:
%         if (typeof(D) == Struct_Type && D.color == RED) { % not considered black
%             % Case_D6: % D red && S black:
%             RBRotateDirRoot(T,P,dir); % P may be the root
%             S.color = P.color;
%             P.color = BLACK;
%             D.color = BLACK;
%             return; % deletion complete
%         }
%         if (typeof(C) == Struct_Type && C.color == RED) {% not considered black
%             % C red && S+D black:
%             RBRotateDirRoot(T,S,1-dir); % S is never the root
%             S.color = RED;
%             C.color = BLACK;
%             D = S;
%             S = C;
%             RBRotateDirRoot(T,P,dir); % P may be the root
%             S.color = P.color;
%             P.color = BLACK;
%             D.color = BLACK;
%             return;
%         }
%         % Here both nephews are == NULL (first iteration) or black (later).
%         if (P.color == RED) {
%             S.color = RED;
%             P.color = BLACK;
%             return; 
%         }
%         % Case_D1 (P+C+S+D black):
%         S.color = RED;
%         N = P; % new current node (maybe the root)
%         % iterate 1 black level
%         %   (= 1 tree level) higher
%         P=N.parent;
%   } while (typeof(P) == Struct_Type);
%   % end of the (do while)-loop
%   return; 

% } % end of RBdelete

% private define RedBlackTreeInsert(T,key) {
%     % insert a node containing  key into RB tree T

%     variable N=new_RBNode(data);

%     if (typeof(T.root)!=Struct_Type) {
%         T.root=N;
%         T.root.color=BLACK;
%         return;
%     }
    
%     variable x=T.root;

%     % find the future parent of our node
%     while (typeof(x) == Struct_Type ) {
%         variable y=x; %
%         if (N.data < x.data) {
%             x=x.child[TREE_LEFT];
%         } else {
%             x=x.child[TREE_RIGHT];
%         }
%     }
%     N.parent=y;

%     % where to insert
%     variable dir=(T.key_compare(N.data,y.data)==-1) ? TREE_LEFT : TREE_RIGHT;

%     RBinsert1(T,N,N.parent,dir);
% }



% variable RedBlackTree_Struct=struct {
%     root=NULL,
%     new_node=&new_RedBlackTreeNode,
%     insert=&RedBlackTreeInsert,
%     print=&BinarySearchTreePrint,
%     key_compare=&tree_key_compare,
%     search=&BinarySearchTreeSearch,
%     delete=&RBdelete,
%     exists=&BinarySearchTreeExists,
%     min=&BinarySearchTreeMinimum,
%     max=&BinarySearchTreeMaximum,
%     successor=&BinaryNodeSuccessor,
%     predecessor=&BinaryNodePredecessor
% };



% private define new_RBTree() {
%     return @RedBlackTree_Struct;
% }



% % function joinRightRB(TL, k, TR):
% %     if (TL.color=black) and (TL.blackHeight=TR.blackHeight):
% % return Node(TL, {k,red},TR)
% % T'=Node(TL.left, {TL.key,TL.color},joinRightRB(TL.right,k,TR))
% %     if (TL.color=black) and (T'.right.color=T'.right.right.color=red):
% %         T'.right.right.color=black;
% %         return rotateLeft(T',N,T')
% %     return T' /* T''[recte T'] */

% % function joinLeftRB(TL, k, TR):
% %   /* symmetric to joinRightRB */

% % function join(TL, k, TR):
% %     if TL.blackHeight>TR.blackHeight:
% %         T'=joinRightRB(TL,k,TR)
% %         if (T'.color=red) and (T'.right.color=red):
% %             T'.color=black
% %         return T'
% %     if TR.blackHeight>TL.blackHeight:
% %         /* symmetric */
% %     if (TL.color=black) and (TR.color=black):
% % return Node(TL,{k,red},TR)
% % return Node(TL,{k,blac},TR)

% % function split(T, k):
% %     if (T = nil) return (nil, false, nil)
% %     if (k = T.key) return (T.left, true, T.right)
% %     if (k < T.key):
% %         (L',b,R') = split(T.left, k)
% %         return (L',b,join(R',T.key,T.right))
% %     (L',b,R') = split(T.right, k)
% % return (join(T.left,T.key,L'),b,T.right)

% % function union(t1, t2):
% %     if t1 = nil return t2
% %     if t2 = nil return t1
% %     (L1,b,R1)=split(t1,t2.key)
% %     proc1=start:
% %         TL=union(L1,t2.left)
% %     proc2=start:
% %         TR=union(R1,t2.right)
% %     wait all proc1,proc2
% % return join(TL, t2.key, TR)

%
% Code from src/slang/astronomy/barycen.sl
%


define barycen();

define barycen() {
%!%+
%\function{barycen}
%\synopsis{barycenter an event list, FITS extension, or FITS file}
%\usage{barytime=barycen(eventlist);}
%\altusage{barycen(String_Type fitsfile);}
%\altusage{barycen(FITS_File_Type fptr);}
%\qualifiers{
%\qualifier{debug}{if set, print debugging information}
%\qualifier{ephemeris}{pointer to or name of a FITS ephemeris file (default: DE430)}
%\qualifier{orbit}{mandatory, structure containing the orbit information, in
%                    geocentric ICRS coordinates and velocities in tags
%                    x,y,z,vx,vy,vz in units of m and m/s, and time in tag mjd}
%\qualifier{deltat}{if set and the routine barycenters an event list or times, return the
%                     difference between the barycentered time and the normal time}
%\qualifier{ra}{right ascension (ICRS) of the source [rad]}
%\qualifier{dec}{declination (ICRS) of the source [rad]}
%\qualifier{deg}{if set, ra/dec are in degrees}
%}
%\description
% This routine barycenters an event list, a FITS extension that is compatible
% with this routine, or all FITS extensions in a FITS file that are compatible
% with this routine.
%
% Caveat emptor: Right now the function assumes that all times are in TT and
% coordinates are in the ICRS system, i.e., it does not know that certain
% JPL ephemerides are NOT in this coordinate system.
%
%\seealso{jpl_eph}
%
%!%-
    variable arg1;

    arg1=();

    variable debug=qualifier("debug",0);
    %
    variable i,nrow,rr,dd,ra,de,oldcolnum,newcolnum;
    variable writefits=-1;

    %
    % initialize the ephemeris
    %
    variable eph, ephkey;
    if (qualifier_exists("ephemeris")) {
        eph=qualifier("ephemeris");
      if (typeof(eph) == Struct_Type){ % user may manually initialize orbit 
        ephkey="USER DEFINED";
      } else {
	ephkey=eph;
	eph=jpl_initeph(getenv("LHEA_DATA")+"/"+eph;nocache);
	
      }
    } else {
      ephkey="JPLEPH.430";
      eph=jpl_initeph(getenv("LHEA_DATA")+"/JPLEPH.430";nocache);
    }

    %
    % initialize the orbit
    % TO DO: also allow Earth coordinates
    %
    variable orbit=qualifier("orbit");

    variable status;
    variable hdu;

    if (typeof(arg1)==String_Type) {
        %
        % 1st argument is a string: file name
        % barycenter all extensions
        %
        variable qual=__qualifiers();
        if (not qualifier_exists("ephemeris")) {
            qual=struct_combine(qual,struct { ephemeris=eph} );
        }

        variable fptr=fits_open_file(arg1,"w");
        variable numhdu=fits_get_num_hdus(fptr);

        _for hdu(1,numhdu,1) {
            if (debug>0) {
                ()=printf("Extension number %i: ",hdu);
            }
            fits_movabs_hdu(fptr,hdu);

            variable extname, timeref;
            status=_fits_read_key(fptr,"EXTNAME",&extname,NULL);
            status=_fits_read_key(fptr,"TIMEREF",&timeref,NULL);
	  
            if (typeof(extname)==String_Type && (extname=="EVENTS" || extname=="RATE") ) {
	         if (typeof(timeref)==String_Type && (timeref!="SOLARSYSTEM")){
		  if (debug>0) {
                    ()=printf("%s - barycentering\n",extname);
		  }
		  barycen(fptr;; qual);
		 } else {
		   ()=printf("%s already barycentered! - skipping!\n",extname);
		 }
            } else {
                if (debug>0) {
                    if (typeof(extname)==String_Type) {
                        ()=printf("%s - skipping\n",extname);
                    } else {
                        ()=printf("skipping\n");
                    }
                }
            }
        }
        fits_close_file(fptr);
        return;
    }

    if (typeof(arg1)==Fits_File_Type) {
        %
        % we are working on a Fits file extension
        %
        writefits=1;
        fptr=arg1;
    
        variable evthdr=fits_read_header(fptr);

        variable colnames=qualifier("colnames",["TIME","TSTART","TSTOP"]);
        variable newnames=qualifier("newnames",["TBARY","TSTART","TSTOP"]);

        if (length(colnames)!=length(newnames)) {
            throw UsageError,sprintf("%s: Usage error: colnames and newnames must have same length\n",_function_name());
        }
        
        oldcolnum={};
        newcolnum={};
    
        % check existence of possible time columns,
        % generate barycentric columns if required
        _for i(0,length(colnames)-1,1) {
            if (fits_binary_table_column_exists(fptr,colnames[i])) {
                variable timecol=fits_get_colnum(fptr,colnames[i]);
                list_append(oldcolnum,timecol);
                if (colnames[i]!=newnames[i]) {
                    variable barycol=timecol+1;
                    list_append(newcolnum,barycol);
                    status=_fits_insert_cols(fptr,barycol,[newnames[i]],["D"]);
                }
            }
        }
    
        % number of events
        nrow=evthdr.NAXIS2;

        % source position
        % rr,dd will be in deg, ra,de in rad
        rr=qualifier("ra",evthdr.RA_OBJ);
        dd=qualifier("dec",evthdr.DEC_OBJ);

        if (qualifier_exists("ra")) {
            if (qualifier_exists("deg")) {
                ra=rr*PI/180.;
            } else {
                ra=rr;
                rr*=180./PI; % convert to degrees for FITS header
            }
        } else {
            ra=rr*PI/180.;
        }

        if (qualifier_exists("dec")) {
            if (qualifier_exists("deg")) {
                de=dd*PI/180.;
            } else {
                de=dd;
                dd*=180./PI;
            }
        } else {
            de=dd*PI/180.;
        }
    }

    if (typeof(arg1[0])==Double_Type) {
        %
        % direct barycentering a time value or array
        %
        writefits=0;
        
        if (qualifier_exists("ra")) {
            ra=qualifier("ra");
        } else {
            throw UsageError,sprintf("%s: Usage error: need ra to barycenter a time array\n",_function_name());
        }

        if (qualifier_exists("dec")) {
            de=qualifier("dec");
        } else {
            throw UsageError,sprintf("%s: Usage error: need dec to barycenter a time array\n",_function_name());
        }

        if (qualifier_exists("deg")) {
            ra*=PI/180.;
            de*=PI/180.;
        }

        % ra, de are in rad
        % rr, dd are in deg        
        rr=ra*180./PI; 
        dd=de*180./PI;

        nrow=length(arg1);
        oldcolnum=0;

        variable tnew=Double_Type(length(arg1));
    }

    if (writefits==-1) {
        throw UsageError,sprintf("%s: first argument must be Double array or FITS file pointer\n",_function_name());
    }
    
    % unit vector towards the source
    variable srcdir=struct {x=cos(de)*cos(ra),y=cos(de)*sin(ra),z=sin(de)};

    %
    % now loop over all times
    %

    % speed of light
    variable c=299792458.; % m/s
    variable GMsun=132712440041.939400e9; % m^3/s^2 (per DE430)

    variable row,col;
    _for row (1,nrow,1) {
        if (debug>0) {
            if (row mod 10000==1) {
                ()=printf("Processing row %7i of %7i (%5.1f%% done)\n",row,nrow,(100.*row)/nrow);
            }
        }

        variable evt_tt;

        % loop over the columns we need to barycenter
        % note: if we barycenter an array, we set oldcolnum to 0 above, i.e.,
        % the loop is only executed once
        _for col (0,length(oldcolnum)-1,1) {
            if (writefits==1) {
                variable evt_time;
                status=_fits_read_col(fptr,timecol,row,1,&evt_time);
                evt_time=evt_time[0];
        
                % convert event time to mjd (in TT)
                evt_tt=(evthdr.TIMEZERO+evt_time)/86400. + evthdr.MJDREF;
            } else {
                evt_tt=arg1[row-1];
            }

            variable earthbary=jpl_eph_vec(evt_tt,"Earth";mjd,equatorial,mks,barycentric,ephemeris=eph);
            variable sunbary=jpl_eph_vec(evt_tt,"Sun";mjd,equatorial,mks,barycentric,ephemeris=eph);

            %
            % satellite position wrt center of Earth 
            variable satearth=struct {
                x=interpol(evt_tt,orbit.mjd,orbit.x),
                y=interpol(evt_tt,orbit.mjd,orbit.y),
                z=interpol(evt_tt,orbit.mjd,orbit.z),
                vx=interpol(evt_tt,orbit.mjd,orbit.vx),
                vy=interpol(evt_tt,orbit.mjd,orbit.vy),
                vz=interpol(evt_tt,orbit.mjd,orbit.vz)
            };

            % note: in principle it does not make sense to measure the SAT time in TT,
            % it can well be that in this step we are doing double counting. This might
	    % introduce an error at the few mus(? - CHECK) level

            % barycentric position of the satellite
            variable satbary=struct {
               x=satearth.x+earthbary.x,
               y=satearth.y+earthbary.y,
               z=satearth.z+earthbary.z
	    };
 
            % Roemer delay (s)
            variable troemer=-(satbary.x*srcdir.x + satbary.y*srcdir.y + satbary.z*srcdir.z)/c;
            
            % Shapiro delay due to Sun (s), see Lyne & Graham-Smith, Pulsar Astronomy, Eq. 5.12
	    % vector sun->satellite
	    variable Rsun=struct {
	        x=satbary.x-sunbary.x,
	        y=satbary.y-sunbary.y,
	        z=satbary.z-sunbary.z
	    };
	    variable dsun=sqrt(Rsun.x*Rsun.x+Rsun.y*Rsun.y+Rsun.z*Rsun.z);
	    variable rang=Rsun.x*srcdir.x+Rsun.y*srcdir.y+Rsun.z*srcdir.z;
	    variable tshapiro=-2.*GMsun*log(1+rang/dsun)/c/c/c;

            variable t2b=-tt2tdb(evt_tt;mjd,deltat); % tt-tdb, change sign because tt2tdb returns TDB-TT

            % time dilatation (Einstein delay) needed for conversion TCG to TCB and accounts for
	    % time shift between surface of Earth and geocenter
	    % See Petit & Luzum (2010), IERS Technical Note No 36, Eq. 10.5

	    variable tdil=-(satearth.x*earthbary.vx+satearth.y*earthbary.vy+satearth.z*earthbary.vz)/c/c;

            if (debug>0) {
	        if (row mod 10000==1) {
	            ()=printf("   roemer: %12.3f ms shapiro: %7.3g ms\n",troemer*1e3,tshapiro*1e3);
	            ()=printf("   t2b   : %12.3f ms tdil   : %7.3f ms\n",t2b*1e3,tdil*1e3);
	        }
            }
	    variable deltat=t2b+tdil+troemer+tshapiro;
        
            % now store barycentric value
            if (writefits==1) {
                status=_fits_write_col(fptr,newcolnum[col],row,1,evt_time-deltat);
            } else {
                if (qualifier_exists("deltat")) {
                    % return tt-tbary
                    tnew[row-1]=deltat;
                } else {
                    tnew[row-1]=evt_time-deltat;
                }
            }
        }
    } % end of event loop
        
    if (debug>0) {
        ()=printf("Processing row %7i of %7i (%5.1f%% done)\n",nrow,nrow,100.);
    }

    if (writefits==1) {
        status=_fits_update_key(fptr,"TIMEREF","SOLARSYSTEM","reference system");
        status=_fits_update_key(fptr,"PLEPHEM",ephkey,"ephemeris used");
        status=_fits_update_key(fptr,"RA_TDB",rr,"[deg] RA of TDB correction");
        status=_fits_update_key(fptr,"DEC_TDB",dd,"[deg] DEC of TDB correction");
        _for i(0,length(newcolnum)-1,1) {
            if (oldcolnum[i]!=newcolnum[i]) {
                status=_fits_update_key(fptr,sprintf("TUNIT%i",newcolnum[i]),"s",sprintf("Unit of %s column",newnames[i]));
            }
        }
        status=_fits_write_chksum(fptr);
    }

    if (writefits==0) {
        return tnew;
    }
}


%
% Code from src/misc/astronomy/bayesian_blocks.sl
%

%%%%%%%%%%%%%%%%%%%%%
define bayesian_blocks() {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{bayesian_blocks}
%\synopsis{find bayesian blocks in the given data}
%\usage{Struct_Type bayesian_blocks(Struct_Type[] data);}
%\qualifiers{
%    \qualifier{fp_rate}{value of the false positive rate, which is
%                the probability that a found change point
%                with the current value of ncp_prior is
%                actually not significant (default: .01)
%                WARNING: This has no influence on data of     
%                type 3, where the default ncp_prior, fp_rate
%                is 0.05, and it is not possible to change this
%                currently.}
%    \qualifier{ncp_prior}{controls the tolerance level used to find
%                the change points. The value has to be an
%                array of estimates for the order of number
%                of blocks for each given dataset. The
%                probability P that the final number of
%                blocks N_blocks is then given by
%                  log(P) ~ ncp_prior * N_blocks
%                Thus, ncp_prior has to be decreased to
%                increase the final number of blocks.
%                The default value is determined by a
%                formula derived from simulations, given by
%                  ncp_prior = 4 - log(73.53*fp_rate*N^-0.478) 
%                for data modes 1 and 2 and 
%                  ncp_prior = 1.32 + 0.577*log(N) 
%                for data mode 3, where N is the number of
%                data points.}
%    \qualifier{do_iter}{number of iterations on ncp_prior
%                (default: 0)}
%    \qualifier{dt_min}{events within the given time range are
%                considered to be simultaneous and a new
%                combined event at the average time is added
%                (default: 0)}
%    \qualifier{num_skip}{maximal number of consecutive simultaneous
%                events (time difference < dt_min)}
%    \qualifier{gti}{structure containing the GTIs for the given
%                events (data mode 1) with the fields 'start'
%                and 'stop'. The blocks are searched for each
%                GTI individually, thus the function returns
%                an array structures! If GTIs are given only
%                ONE input dataset is allowed (at the moment).}
%    \qualifier{gti_mindt}{all GTIs shorter than the given duration are
%                ignored (i.e. no blocks are searched)
%                (default: 0)}
%}
%\description
%    This function is an S-lang implementation of the
%    'find_blocks_mult' MatLab-program described in
%      Scargle et al., 2013, ApJ 764, 167
%    which source code is available on the arXiv.
%
%    WARNING: The full functionality has not been tested
%    yet, thus please use with caution!
%
%    Depending on the fields of the input 'data' structure,
%    the best Bayesian block representation is determined.
%    The available data modes are:
%    1: Event data
%      Double_Type[] tt        - time tags of events
%    2: Binned data (data without specific uncertainties,
%       e.g., counts vs. time)
%      Double_Type[] tt        - time tags
%      Double_Type[] nn_vec    - binned data at tt
%WARNING! Mode 2 possibly has bugs! Use with extreme caution!
%    3: Point measurements (data with known uncertainties,
%       e.g., radio flux vs. time)
%      Double_Type[] tt        - time tags
%      List_Type[]   cell_data - cell data at tt defined as
%                     { Double_Type[] data, uncertainties }
%
%    The output structure contains
%      change_points - time INDICES of the change points
%      tt_all        - combined time tags of all datasets
%      ncp_prior_vec - ncp_prior value of each dataset
%      data_matrix   - see Fig. 2 of Scargle et al. 2013
%      last          - index of the last change point
%                      over tt_all
%      best          - fitness (=statistics) over tt_all
%      index_vec     - dataset index over tt_all
%      data_mode_vec - data mode of each dataset
%
%   The algorithm can handle multiple datasets, which are
%   combined internally to one dataset (tt_all). Zero
%   blocks are handled as well.
%
%   NOTE: in case of events (data mode 1) it is important
%         to provide the GTIs via the corresponing qualifier
%         as well. Otherwise the block statistics are
%         probably wrong and thus the total block
%         represenation!
%
%   To finally retrieve the block represenatation of the
%   input data, the 'get_blocks_data' function can be used.
%\seealso{get_blocks_data}
%!%-
  variable data;
  switch (_NARGS)
   { case 1: data = (); }
   { help(_function_name); return; }

  % check on type of given data
  if (typeof(data) != Array_Type)
  {
    data = [data];
  }
  if (_typeof(data) != Struct_Type)
  {
    vmessage("error (%s): input data has to be a structure", _function_name);
    return;
  }

  % check if GTIs are given -> determine blocks for each GTI separately
  if (qualifier_exists("gti"))
  {
    % only one dataset is allowed
    if (length(data) > 1){
      vmessage("error (%s): in case of GTIs only one dataset is allowed", _function_name);
      return;
    }
    data = data[0];
    % only datamode 1 is allowed
    if (struct_field_exists(data, "cell_data") or struct_field_exists(data, "nn_vec"))
    {
      vmessage("error (%s): in case of GTIs only event data is allowed", _function_name);
      return;
    }
    % check GTI structure
    variable gti = qualifier("gti");
    if (typeof(gti) != Struct_Type)
    {
      vmessage("error (%s): GTIs have to be defined via a structure", _function_name);
      return;
    }
    ifnot (struct_field_exists(gti, "start") and struct_field_exists(gti, "stop"))
    {
      vmessage("error (%s): GTI structure is missing fields", _function_name);
      return;
    }
    
    % recursively call the function itself
    variable gtidt = qualifier("gti_mindt", 0.);
    struct_filter(gti, where(gti.stop - gti.start >= gtidt));
    vmessage("considering %d GTIs", length(gti.start));
    variable out = Struct_Type[length(gti.start)];
    variable qual = reduce_struct(__qualifiers, ["gti", "gti_mindt"]);        
    variable g, n;
    _for g (0, length(out)-1, 1) {
      n = where(gti.start[g] <= data.tt <= gti.stop[g]);
      if (length(n) > 1)
      {
	out[g] = bayesian_blocks(struct { tt = data.tt[n] } ;; qual);
      }
      else {
	vmessage("warning: GTI number %d has less than two events, ignoring...", g+1);
      }
    }
    return out;  
  }

  % copy data structure since it gets modified below
  data = COPY(data);
  
  % preferences and variables
  variable num_series = length(data); % number of time series
  variable do_iter = qualifier("do_iter", 0); % number of iterations on ncp_prior
  variable fp_rate = qualifier("fp_rate", .01); % value of false positive rate
  
  % set up data matrix
  variable data_mode_vec = Integer_Type[num_series]; % all zeros
  variable ncp_prior_vec = Double_Type[num_series];
  variable tt_start_vec = @ncp_prior_vec;
  variable tt_stop_vec = @ncp_prior_vec;
  variable ii_start_vec = @data_mode_vec;
  variable tt = Double_Type[0]; % initialize master time array
  variable ii_start = 1-1; % initialize marker for parsing the data
  variable row_count = 1; % first row is array containing all times

  variable d;
  _for d (0, num_series-1, 1) {
    % identify data mode (3=point measurements, 2=binned data, 1=event_data)
    if (struct_field_exists(data[d], "cell_data"))
    {
     data_mode_vec[d] = 3;
    } else if (struct_field_exists(data[d], "nn_vec"))
    {
      data_mode_vec[d] = 2; vmessage("WARNING! Entering mode 2.. there are possibly bugs in this mode! Use with caution!");
    } else
    {
      data_mode_vec[d] = 1;
    }

    row_count = row_count + 2; % these modes all take two rows

    % process time markers (all data modes must have one)
    variable tt_this = @(data[d].tt);
    variable dt_this = diff(tt_this);
    if (any(dt_this < 0))
    {
      vmessage("error (%s): data must be time ordered", _function_name);
      return;
    }

    if (data_mode_vec[d] == 1)
    {
      % combine any duplicate time_tags:
      variable dt_min = qualifier("dt_min", 0); % set to small value >0 to combine nearly equal times
      variable num_skip = qualifier("num_skip", 3);
      variable nn_vec = ones(length(tt_this));
      variable ii_dupe = where(dt_this <= dt_min); % indices of small intervals
      while (length(ii_dupe) > 0)
      {
        variable iu = ii_dupe[[0:length(ii_dupe)-1:num_skip]];
        % replace with average of the two identical (or close) times
        tt_this[iu] = .5*(tt_this[iu] + tt_this[iu+1]);
        % replace with sum of the two corresponding cell populations
        nn_vec[iu] = nn_vec[iu] + nn_vec[iu+1];

        % remove second member of these pairs
	tt_this = tt_this[complement([0:length(tt_this)-1], iu+1)];
	nn_vec = nn_vec[complement([0:length(nn_vec)-1], iu+1)];

        % any more duplicates?  If so, go again; if not, you are done.
        dt_this = diff(tt_this);
        ii_dupe = where(dt_this <= dt_min);
      }

      % store adjusted data back into master data structure
      data[d] = struct_combine(data[d], struct { tt = tt_this, nn_vec = nn_vec });
    }

    variable num_points_this = length(tt_this);
    tt = [tt, tt_this]; % concatenate all times

    % this array keeps track of index ranges of entries for each series:
    ii_start_vec[d] = ii_start;
    ii_start = ii_start + num_points_this; % update to start of next series

    tt_start_vec[d] = struct_field_exists(data[d], "tt_start") ? data[d].tt_start : tt[0] - 0.5 * median(diff(tt)); % default start time
    tt_stop_vec[d] = struct_field_exists(data[d], "tt_stop") ? data[d].tt_stop : tt[-1] + 0.5 * median(diff(tt)); % default stop time

    % store ncp_prior if present; if not, use default
%    ncp_prior_vec[d] = struct_field_exists(data[d], "ncp_prior") ? data[d].ncp_prior : 4 - log(fp_rate/(0.0136*num_points_this^.478));
    if (struct_field_exists(data[d], "cell_data"))
    {
      ncp_prior_vec[d] = (qualifier_exists("ncp_prior") ? qualifier("ncp_prior")[d] : 1.32 + 0.577*log(num_points_this));
    }
    else
    {
      ncp_prior_vec[d] = (qualifier_exists("ncp_prior") ? qualifier("ncp_prior")[d] : 4 - log(fp_rate/(0.0136*num_points_this^.478)));
    }
  } % for num_series

  % construct data matrix  (Figure 2: Top Panel)
  variable num_rows = row_count;
  variable num_data = length( tt ); % total number of data points
  variable data_matrix = Double_Type[num_data, num_rows];
  variable index_vec = Integer_Type[num_data];

  row_count = 0; % reset row counter for staging of data
  data_matrix[*,0] = @tt; % first row contains all times (unordered)

  variable ii_stop, tt_stop, tt_start, dt_start, dt_stop;
  _for d (0, num_series-1, 1)
  {
    % get index range for data for this series
    ii_start = ii_start_vec[d];
    ii_stop = d == num_series-1 ? num_data-1 : ii_start_vec[d+1] - 1;
    index_vec[[ii_start:ii_stop]] = d; % keep track of series

    % compute mode-dependent fitness data
    if (data_mode_vec[d] == 1 || data_mode_vec[d] == 2)
    {
      nn_vec  = data[d].nn_vec;
      tt_this = data[d].tt;
      tt_start = tt_start_vec[d];
      tt_stop = tt_stop_vec[d];

      dt_start = .5*(tt_this[1] + tt_this[0]) - tt_start;
      dt_stop = tt_stop - .5*(tt_this[-2] + tt_this[-1]);
      variable delt_tt = [dt_start, .5*(tt_this[[2:length(tt_this)-1]] - tt_this[[0:length(tt_this)-3]]), dt_stop];

      row_count = row_count + 1;
      data_matrix[[ii_start:ii_stop], row_count] = @delt_tt;

      row_count = row_count + 1;
      data_matrix[[ii_start:ii_stop], row_count] = @nn_vec;
    }
    else if (data_mode_vec[d] == 3)
    {
      variable cd = data[d].cell_data;
      row_count = row_count + 1;
      data_matrix[[ii_start:ii_stop], row_count] = cd[0] / cd[1]^2; % (x/sig^2 )

      row_count = row_count + 1;
      data_matrix[*, row_count] = 1; % non-zero denominator
      data_matrix[[ii_start:ii_stop], row_count] = 1. / cd[1]^2; %  (1/sig^2)
    }
  } % for num_series

  % redistribute data according to time order (Figure 2: Bottom Panel):
  variable num_points = length(tt);
  variable dm = COPY(data_matrix);
  variable ii_sort = array_sort(data_matrix[*,0]); % time order index
  tt = tt[ii_sort];
  data_matrix[*,*] = data_matrix[ii_sort,*]; % re-order everything
  index_vec = index_vec[ii_sort];

  
  % now apply the basic dynamic programming algorithm
  tt_start = min(tt_start_vec);
  tt_stop  = max(tt_stop_vec); % min( tt_stop      );
  % make array of lengths of the "last blocks"
  variable block_length = tt_stop - [tt_start, .5*(tt[[1:length(tt)-1]] + tt[[0:length(tt)-2]]), tt_stop];

  variable iter_count = 0;

  variable best, last, R, fit_vec, nn_cum_vec, arg_log_vec, fit_vec_this, sum_x_1, sum_x_0;
  variable num_cp, err_this, cpt_old = NULL;
  while (1)
  { % if iterating, continue until maximum number is reached
    best = Integer_Type[0];
    last = Integer_Type[0];

    _for R (0, num_points-1, 1)
    {
      fit_vec = Double_Type[R+1]; % initialize last-block fitness array
      row_count = 1; % initialize
      
      _for d (0, num_series-1, 1)
      {
%	ncp_prior = ncp_prior_vec[d];

	if (data_mode_vec[d] == 1 || data_mode_vec[d] == 2)
	{
	  row_count = row_count + 1;
	  delt_tt = data_matrix[[0:R], row_count-1];

	  row_count = row_count + 1;
	  nn_vec = data_matrix[[0:R], row_count-1];

	  nn_cum_vec = reverse(cumsum(reverse(nn_vec)));
	  arg_log_vec = reverse(cumsum(reverse(delt_tt)));
	  arg_log_vec[where(arg_log_vec <= 0)] = _Inf;

	  fit_vec_this = nn_cum_vec * (log(nn_cum_vec) - log(arg_log_vec));
	}
	else if (data_mode_vec[d] == 3)
	{ % measurements, normal errors
          row_count = row_count + 1;
          cd = data_matrix[*,row_count-1];
          sum_x_1 = cumsum(cd[[R:0:-1]]); % sum( x / sig^2 )
    
          row_count = row_count + 1;
          cd = data_matrix[*,row_count-1];
          sum_x_0 = cumsum(cd[[R:0:-1]]); % sum( 1 / sig^2 )
                
          fit_vec_this = (sum_x_1[[R:0:-1]] ^ 2 / (4 * sum_x_0[[R:0:-1]]) );
	}

        fit_vec_this[where(isnan(fit_vec_this))] = 0;
        fit_vec += fit_vec_this - ncp_prior_vec[d];
      }

      last = [last, where_max([0, best] + fit_vec)];
      best = [best, ([0, best] + fit_vec)[last[-1]]];
    }

    % now find changepoints by iteratively peeling off the last block
    variable index = last[num_points-1];
    variable change_points = Integer_Type[0];

    while (index > 0)
    {
      change_points = [index, change_points];
      index = last[index - 1];
    }

    if (do_iter == 0)
    {
      break;
    } % done; not iterating on ncp_prior
    else
    {
      iter_count++;
      num_cp = length(change_points);

      if (num_cp < 1)
      {
	num_cp = 1;
      }

      if (cpt_old != NULL)
      {
	if (num_cp == length(cpt_old))
	{ % compare with previous iteration
	  err_this = sum(abs(change_points - cpt_old));
	}
	else
	{
	  err_this = _Inf;
	}

	if (err_this == 0)
	{
	  vmessage("converged at %d", iter_count);
	  break;
	}

	if (iter_count > do_iter)
	{
	  vmessage("did not converge at %d", iter_count);
	  break;
	}
      }

      fp_rate = 1 - (1-fp_rate)^(1./num_cp);
%      ncp_prior_old = @(ncp_prior_vec[d]);
      ncp_prior_vec[d] = 4 - log(fp_rate / (.0136 * num_points^.478));
      cpt_old = @change_points;
    }
  }
  return struct {
    change_points = change_points,
    tt_all = tt,
    ncp_prior_vec = ncp_prior_vec,
    data_matrix = data_matrix,
    last = last,
    best = best,
    index_vec = index_vec,
    data_mode_vec = data_mode_vec
  };
}


%%%%%%%%%%%%%%%%%%%%%
define get_blocks_data() {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_blocks_data}
%\synopsis{returnes the data of a Bayesian block representation}
%\usage{Struct_Type blockdata = get_blocks_data(Struct_Type blocks[, Integer_Type dataindex]);}
%\altusage{blockdata= get_blocks_data(Struct_Type[] blocks[, ...]);}
%\qualifiers{
%    \qualifier{datafun}{reference to a function to be called to
%              derive the block data (default: see text)}
%    \qualifier{errfun}{same as 'datafun', but for the block
%              uncertainties}
%    \qualifier{bydt}{divides the output data and uncertainties
%              by the length of the corresponding block,
%              can be assigned to a number used to divide
%              the data to for further normalization}
%}
%\description
%    Returnes the data of all block over time (e.g., a
%    lightcurve) with variable binsize from a previous
%    run of 'bayesian_blocks'. The default output
%    structure has the fields:
%      time  - lower time bin of a block
%      dt    - length of each block
%      data  - the data in the blocks and
%      error - and the corresponding uncertainties
%    The data and their uncertainties are calculated
%    from the original input data by functions (if
%    multiple dataset are present use the optional
%    'dataindex' parameter), which will be called for
%    each block and can be specified via qualifiers.
%    Their calling sequence is
%      Double_Type function(time, dt, data[, error])
%    where error is given in case of data mode 3
%    (point measurements) only. The default functions
%    depend on the data mode used previously:
%    1: Event data
%      datafun = sum(events)
%      errfun  = sqrt(sum(events))
%    2: Binned data
%      datafun = sum(nn_vec)
%      errfun  = sqrt(sum(nn_vec))
%    3: Point measurements
%      datafun = weighted_mean(cell_data[0],
%                              cell_data[1]; err)
%      errfun  = sqrt(sumsq(datafun - cell_data[0])
%                     / (length(cell_data[0]) - 1)
%                     / length(cell_data[0]))
%\example
%    % let 'evts' be a list of time-tagged events,
%    % find bayesian blocks
%    blocks = bayesian_blocks(struct { tt = evts });
%
%    % create a lightcurve of the block representation,
%    % 'bydt' will convert counts to rate
%    lc = get_blocks_data(blocks; bydt);
%
%    % plot the blocks
%    hplot(lc.time, lc.time+lc.dt, lc.data);
%\seealso{bayesian_blocks}
%!%-
  variable blocks, dinx = 0;
  switch (_NARGS)
    { case 1: blocks = (); }
    { case 2: (blocks, dinx) = (); }
    { help(_function_name); return; }

  if (typeof(blocks) != Array_Type) { blocks = [blocks]; }
  
  variable nb = int(sum(array_map(Integer_Type, &length, array_struct_field(blocks, "change_points")))) + length(blocks);
  variable out = struct {
    time = Double_Type[nb], dt = Double_Type[nb], data = Double_Type[nb], error = Double_Type[nb] 
  };

  variable g, no = 0;
  _for g (0, length(blocks)-1, 1) {
    variable b, ii_start = 0, ii_end, ftime, fdt, data, err;
    nb = length(blocks[g].change_points)+1;
    _for b (0, nb-1, 1) {
      % data indices of the block
      ii_end = (b < nb-1 ? blocks[g].change_points[b]-1 : length(blocks[g].tt_all)-1);
      % time information
      ftime = blocks[g].tt_all[[ii_start:ii_end]];
      fdt = blocks[g].tt_all[[ii_start+1:ii_end]] - blocks[g].tt_all[[ii_start:ii_end-1]];
      out.time[no+b] = blocks[g].tt_all[ii_start];
      out.dt[no+b] = blocks[g].tt_all[ii_end + (b < nb-1 ? 1 : 0)] - blocks[g].tt_all[ii_start];
      % input data within a block
      if (blocks[g].data_mode_vec[dinx] < 3) {
        data = blocks[g].data_matrix[[ii_start:ii_end],2+2*dinx];
        err = NULL;
      } else {
        data = blocks[g].data_matrix[[ii_start:ii_end],1+2*dinx] / blocks[g].data_matrix[[ii_start:ii_end],2+2*dinx];
        err = 1./sqrt(blocks[g].data_matrix[[ii_start:ii_end],2+2*dinx]);
      }
      % calculate output data
      if (qualifier_exists("datafun")) {
        if (blocks[g].data_mode_vec[dinx] < 3) { out.data[no+b] = @(qualifier("datafun"))(ftime, fdt, data); }
        else { out.data[no+b] = @(qualifier("datafun"))(ftime, fdt, data, err); }
      }
      else {
        if (blocks[g].data_mode_vec[dinx] < 3) { out.data[no+b] = sum(data); }
        else { out.data[no+b] = weighted_mean(data; err = err); }
      }
      % calculate output uncertainties
      if (qualifier_exists("errfun")) {
      if (blocks[g].data_mode_vec[dinx] < 3) { out.error[no+b] = @(qualifier("errfun"))(ftime, fdt, data); }
        else { out.error[no+b] = @(qualifier("errfun"))(ftime, fdt, data, err); }
      }
      else {
        if (blocks[g].data_mode_vec[dinx] < 3) { out.error[no+b] = sqrt(out.data[no+b]); }
       	else { out.error[no+b] = sqrt(sumsq(out.data[no+b] - data) / (length(data) - 1) / length(data)); }
      }
      % eventually divide by dt
      if (qualifier_exists("bydt")) {
        variable bydt = qualifier("bydt");
        if (bydt == NULL) { bydt = 1.; }
        out.data[no+b] /= (out.dt[no+b] * bydt);
        out.error[no+b] /= (out.dt[no+b] * bydt);
      }

      ii_start = ii_end + 1;
    }

    no += nb;
  }

  return out;
}

%
% Code from src/misc/astronomy/bayesian_blocks_cperror.sl
%
%%%%%%%%%%%%%%%%%%%%%
define bayesian_blocks_cperror() {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{bayesian_blocks_cperror}
%\synopsis{calculates the probability distribution
%    of the change point positions}
%\usage{Struct_Type[] bayesian_blocks(Struct_Type blocks);}
%\description
%    After
%      Scargle et al., 2013, ApJ 764, 167
%    the uncertainty of the position of a change point in
%    a bayesian block representation can be estimated by
%    calculating the fitness of a block with variing
%    change point position, while all other change points
%    are kept fix. Although inter-change-point dependencies
%    are neglected in this approach, it can be used to
%    derive proper uncertainties as long as the resulting
%    distributions do not overlap.
%
%    This function is an implementation of this procedure
%    based on the MatLab-programs 'figure_cp_error' and
%    'cp_prob' by the authors above.
%
%    The input 'blocks' have to be calculated previously
%    by the 'bayesian_blocks' function. The returned
%    structure array contains the probability distribution
%    'prob' (integral = 1.) over the time-tags 'tt' for
%    each change point (i.e., the array index corresponds
%    to the change point index). From each individual
%    distribution an uncertainty for the location of the
%    change point can be estimated via, e.g., its width.
%\seealso{bayesian_blocks}
%!%-
  variable blocks;
  switch (_NARGS)
    { case 1: blocks = (); }
    { help(_function_name); return; }

  variable num_cp = length(blocks.change_points);
  variable id_cp, tt_1, cp_1, tt_2, cp_2, cp_center, tt_center;
  variable iu, tt_use, tt_use_min, tt_use_max;
  variable num, dt, dt_median, tt_range, block_length;
  variable id_left, t1, t2, n1, n2, log_prob_1, log_prob_2;
  variable log_prob_vec, log_prob_max;
  variable prob = struct_array(num_cp, struct { tt, prob });
  % loop over all change points
  _for id_cp (0, num_cp-1, 1) {
    %%% get position and times of adjacent and current change points
    % previous
    if (id_cp == 0) { tt_1 = blocks.tt_all[0]; }
    else {
      cp_1 = blocks.change_points[id_cp-1];
      tt_1 = blocks.tt_all[cp_1];
    }
    % current
    cp_center = blocks.change_points[id_cp];
    tt_center =  blocks.tt_all[cp_center];
    % following
    if (id_cp == num_cp-1) { tt_2 = blocks.tt_all[-1]; }
    else {
      cp_2 = blocks.change_points[id_cp+1];
      tt_2 = blocks.tt_all[cp_2];
    }

    % indices of data in current block
    iu = where(blocks.tt_all >= tt_1 and blocks.tt_all <= tt_2);
    tt_use = blocks.tt_all[iu];
    tt_use_min = tt_use[0];
    tt_use_max = tt_use[-1];

    % calculate fitness of current block with variing change point position
    num = length(tt_use);
    dt = diff(tt_use);
    if (min(dt) < 0) { vmessage("error (%s): points must be ordered", _function_name); return; }
    dt_median = median(dt);

    tt_range = tt_use_max - tt_use_min;
    block_length = tt_use_max - [tt_use_min, .5*(tt_use[[1:]] + tt_use[[:-2]]), tt_use_max];
    log_prob_vec = Double_Type[num];

    _for id_left (0, num-1, 1) {
      t2 = block_length[id_left];
      t1 = tt_range - t2;

      n1 = id_left;
      n2 = num - n1;

      log_prob_1 = id_left == 0 ? 0. : n1 * log(1. * n1 / t1);
      log_prob_2 = n2 * log(1. * n2 / t2);

      log_prob_vec[id_left] = log_prob_1 + log_prob_2;
    }
    
    log_prob_max = max(log_prob_vec);
    prob[id_cp].prob = exp(log_prob_vec - log_prob_max);
    prob[id_cp].prob /= sum(prob[id_cp].prob);
    prob[id_cp].tt = @(tt_use);
  }

  return prob;
}

%
% Code from src/misc/astronomy/bbootstrap.sl
%
%%%%%%%%%%%%%%%%%%%%%
define bbootstrap() {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{bbootstrap}
%\synopsis{1-parameter bootstrap calculation of function 'funct'}
%\usage{Array_Type bbootstrap(Array_Type data, Integer_Type num_samp
%                             [, funct = &mean, datatype = Double_Type]);}
%\qualifiers{
%    \qualifier{slow}{the balanced bootstrap is performed. This is
%            more accurate, especially for long-tailed
%            distributions, but is also much slower and
%            requires more memory.}
%}
%\description
%    This function has been ported from the IDL-routine
%    bbootstrap.pro by H.T. Freudenreich, HSTX, 2/95
%
%    This program randomly selects values to fill sets of
%    the same size as 'data'. Then calculates the 'funct'
%    of each, which is an optional function reference with
%    the &mean as default. It does this 'num_samp' times.
%    The 'datatype' returned by 'funct' has to be provided
%    to initialize the array returned by this function. 
%    If the slow-qualifier is set, the balanced bootstrap
%    method is used to obtain the samples, hence the name
%    'bbootstrap'. This method requires, however, more
%    virtual memory, and patience.
%
%    The user should choose 'num_samp' large enough to get
%    a good distribution. The sigma of that distribution
%    is then the standard deviation of the mean of the
%    returned 'funct'-vector.
%    For example, if input 'data' is normally distributed
%    with a standard deviation of 1.0, the standard
%    deviation of the returned 'funct'-vector (by default
%    the mean) will be ~1.0/sqrt(N-1), where N is the
%    number of values in 'data'.
%
%    WARNING: at least 5 points must be input.
%             The more, the better.
%!%-
  variable data, num_samp, funct = &mean, datatype = Double_Type;
  switch(_NARGS)
    { case 2: (data, num_samp) = (); }
    { case 4: (data, num_samp, funct, datatype) = (); }
    { help(_function_name); return; }

  variable anser = datatype[num_samp];
  variable n = length(data);
  variable i;
  
  if (qualifier_exists("slow")) {
    % concatenate everything into one long vector
    variable m = num_samp*n;
    variable biggy = Double_Type[m];
    variable k = n-1;
    variable i1 = 0;
    _for i (0, num_samp-1, 1) {
      biggy[[i1:i1+k]] = data;
      i1 += n;
    }

    % now scramble it!
    % select m numbers at random, repeating none
    biggy = biggy[array_permute(m)];

    % now divide it into num_samp units and perform the funct on each.
    i1 = 0;
    _for i (0, num_samp-1, 1) {
      anser[i] = @funct(biggy[[i1:i1+k]]);
      i1 += n;
    }
  }

  else {
    seed_random(_time*2+1);
    variable seeds = int(urand(num_samp)*1e6+1);
    _for i (0, num_samp-1, 1) {
      % update the random number seed and get uniform random numbers between 0 and n-1
      seed_random(seeds[i]);
      variable r = int(urand(n)*n);
      variable v = data[r];
      anser[i] = @funct(v);
    }
  }

  return anser;
}

%
% Code from src/slang/math/bezier.sl
%

% only activate if module vector could be loaded
#ifeval __get_reference("vector")!=NULL

define bezNp ()
{
%!%+
%\function{bezNp}
%\synopsis{calulates a Bezier curve between vectors}
%\usage{Vector_Type bez = bezNp(Vector_Type vectors);
%}
%\description
%   The Bezier curve between each pair of the arrays in 'vectors'
%   will be calculated. The 'vectors' arrays needs to be at least two
%   entries.
%   Returned value 'bez' will be one entry shorter than 'vectors'.
%   Main working horse for the bezier function.
%   
%\seealso{bezier}
%!%-

  switch(_NARGS)
    { case 1:
	variable bez = () ;
    }
    { return help(_function_name()); }

  if (length(bez)<2) {message("ERROR (bezNP): vectors array needs to be at least 2 entries"); return ;}
  
  variable wfac = [1:0:#length(bez[0].x)] ;
  variable v ;
  variable bez2 = Vector_Type[length(bez)-1] ;
  _for v (0, length(bez)-2, 1)
    {
      bez2[v] = vector(wfac*bez[v].x + (1-wfac)*bez[v+1].x, wfac*bez[v].y + (1-wfac)*bez[v+1].y, wfac*bez[v].z + (1-wfac)*bez[v+1].z) ;
    }
  return bez2 ;
}
  
define make_conlines ()
{
  variable points, nbins ;
  switch(_NARGS)
    { case 2:
	(points,nbins) = () ;
    }
    { return help(_function_name()); }

  variable p ;
  variable conlines = Vector_Type[length(points)-1] ;

  _for p (0, length(points)-2, 1)
    {
      conlines[p] = vector([points[p][0]:points[p+1][0]:#nbins],
			   [points[p][1]:points[p+1][1]:#nbins],
			   0) ;
    }
  return conlines ;
}

define bezier () {
%!%+
%\function{bezier}
%\synopsis{calulates a Bezier curve from given supporting points}
%\usage{Vector_Type bez = bezier(List_Type points, Integer_Type nbins);
%}
%\description
%   Bezier calculates the Bezier curve between the first and last
%   [x,y] entry of 'points', using the other entries of 'points' as
%   supporting points. 'points' must be a List_Type, with each
%   element containing the [x,y] values of the corresponding point.
%   'nbins' gives the numbers of points used to calculate each part
%   of the curve, as well as the length of the returned curve.
%   
%\qualifiers{
%\qualifier{allbez}{if set, all intermediate Bezier curves are
%          returned as a List, containing all the Vector_Type Bezier curves.}
%\qualifier{conline}{if set, the initial lines connecting the
%          points are also returned, i.e., the return value becomes (Vector_Type
%          conlines, Vector_Type bez).}
%}
%   
%\seealso{xfig_plot_allbezier, bezNp}
%!%-

  variable points, nbins ;
  switch(_NARGS)
    { case 2:
	(points,nbins) = () ;
    }
    { return help(_function_name()); }

  variable conlines = make_conlines( points, nbins) ;
  variable bez = bezNp(conlines) ;
  variable bezarr = {bez} ;

    while (length(bez) > 1)
      {
	bez = bezNp(bez) ;
	if (qualifier_exists("allbez"))
	  { list_append(bezarr,bez) ;}
      }
  if (qualifier_exists("conlines")) {
	if (qualifier_exists("allbez")) {
	  return conlines, bezarr ;
	}
    return conlines, bez[0] ;
  }
  
  else {
    if (qualifier_exists("allbez")) {
      return bezarr ;
    }
   return bez[0] ;
  }
}

define xfig_plot_allbezier () {
%!%+
%\function{xfig_plot_allbezier}
%\synopsis{adds the plot of all Bezier curves from the bezier function to a xfig plot object}
%\usage{Xfig_Struct pl_new = xfig_plot_allbezier(Xfig_struct pl, bez [, conlines ]);
%}
%\description
%   Adds the plot of all Bezier curves to a previously defined
%   xfig-object 'pl'. 'conlines' and 'bez' are the values returned from
%   the bezier-function called with both qualifiers "allbez" and
%   "conlines".
%   Colors of each Bezier line collection is changed automatically.
%   If conlines are plotted, they will be plotted in "gray".
%   
%\seealso{bezier}
%!%-

  variable pl, bez ;

  switch(_NARGS)
    { case 2:
	(pl,bez) = () ;
    }
    { case 3:
	variable conlines ;
	(pl,bez,conlines) = () ;
    }
    { return help(_function_name()); }

  variable b,b1,v ;

  if (__is_initialized(&conlines)) {
  foreach v (conlines)
    {
      pl.plot(v.x, v.y ; color="gray", forward_arrow) ;
    }
  }
  variable i =  1;

  foreach b1 (bez)
    {
      foreach b (b1)
	{
	  pl.plot(b.x, b.y ; color= i, depth = 20-i) ;
	}
      i++;
    }
  return pl ;
}

#endif

%
% Code from src/misc/satellites/RXTE/bgsubHEXTElc.sl
%
%%%%%%%%%%%%%%%%%%%
define bgsubHEXTElc()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{bgsubHEXTElc}
%\synopsis{calculates a background subtracted HEXTE lightcurve}
%\usage{Struct_Type lc = bgsubHEXTElc(Struct_Type srclc, Struct_Type bglc);
%\altusage{Struct_Type lc = bgsubHEXTElc(Struct_Type srclc, Struct_Type bglc1, Struct_Type bglc2);}
%}
%\description
%    HEXTE rocking -> Background is not measured simultaneously with source data.
%!%-
{
  variable srclc, bglc, bglc1, bglc2;
  switch(_NARGS)
  { case 2: (srclc, bglc) = (); }
  { case 3: (srclc, bglc1, bglc2) = (); bglc = sort_struct_arrays(merge_struct_arrays([bglc1, bglc2]), "time"); }
  { help(_function_name()); return; }

  variable av_bglc = struct { time=Double_Type[0], rate=Double_Type[0], error=Double_Type[0] };
  foreach ( split_lc_at_gaps( struct_filter(bglc, where(bglc.rate>0); copy), 30) )
  { variable bglc_part = ();
    av_bglc.time  = [av_bglc.time,  mean(bglc_part.time)];
    av_bglc.rate  = [av_bglc.rate,  mean(bglc_part.rate)];
    av_bglc.error = [av_bglc.error, mean(bglc_part.error)];
  }

  return struct {
    time  = srclc.time,
    rate  = srclc.rate - interpol(srclc.time, av_bglc.time, av_bglc.rate),
    error = sqrt( srclc.error^2 + interpol(srclc.time, av_bglc.time, av_bglc.error)^2 )
  };
}

%
% Code from src/histogram/bin_average.sl
%
define bin_average()
%!%+
%\function{bin_average}
%\synopsis{computes averages in histogram bins}
%\usage{Struct_Type bin_average(phi, rate, phi_lo[, phi_hi])}
%\qualifiers{
%\qualifier{err}{error on rate}
%\qualifier{quantiles}{array of quantiles to calculate for the distribution}
%\qualifier{quartiles}{calculate the .25, .5 and .75 quantiles}
%}
%\description
%    The fields of the returned structure are:\n
%    - \code{bin_lo} and \code{bin_hi}, defining the bins\n
%    - \code{value}: the average\n
%    - \code{err}: the standard error of the mean\n
%    - \code{n}: the number of points in each bin\n
%    and, if the \code{err} qualifier specifies an error array:\n
%    - \code{weighted_average}: an error-weighted average
%!%-
{
  variable phi, rate, phi_lo, phi_hi;
  switch(_NARGS)
  { case 3: (phi, rate, phi_lo) = (); phi_hi = make_hi_grid(phi_lo); }
  { case 4: (phi, rate, phi_lo, phi_hi) = (); }
  { help(_function_name()); return; }

  variable indices = wherenot(isnan(rate) or isinf(rate));
  phi = phi[indices];
  rate = rate[indices];

  variable N = length(phi_lo);
  variable rev;
  variable info = struct {
    bin_lo = phi_lo,
    bin_hi = phi_hi,
    value  = Double_Type[N],
    err    = Double_Type[N],
    n      = histogram(phi, phi_lo, phi_hi, &rev),
  };
  variable i;
  _for i (0, N-1, 1)
  { variable m = moment(rate[ rev[i] ]);
    info.value[i] = m.ave;
    info.err[i] = m.sdom;
  }
  variable err = qualifier("err");
  if(err!=NULL)
  {
    info = struct_combine(info, struct { weighted_average = Double_Type[N] });
    _for i (0, N-1, 1)
      info.weighted_average[i] = weighted_mean(rate[ rev[i] ], 1/err[ rev[i] ]^2);
  }
  variable quantiles = qualifier("quantiles");
  if(qualifier_exists("quartiles"))
  {
    if(quantiles==NULL)
      quantiles = [.25, .5, .75];
    else
    {
      quantiles = [quantiles, .25, .5, .75];
      quantiles = quantiles[array_sort(quantiles)];
    }
  }
  if(quantiles!=NULL)
  {
    variable fields = "quantile"+array_map(String_Type, &strtrans, array_map(String_Type, &string, quantiles), ".", "_");
    info = struct_combine(info, @Struct_Type(fields));
    _for i (0, N-1, 1)
    {
      variable r = rate[ rev[i] ];
      variable f;
      _for f (0, length(fields)-1, 1)
      {
	if(i==0)
	  set_struct_field(info, fields[f], Double_Type[N]);
	get_struct_field(info, fields[f])[i] = quantile(quantiles[f], r);
      }
    }
  }
  return info;
}

%
% Code from src/data/bin_corr_factor.sl
%
public variable unfold_corr_factor = Assoc_Type[Array_Type];
%%%%%%%%%%%%%%%%%%%%%%%%%%
define set_bin_corr_factor()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{set_bin_corr_factor}
%\synopsis{sets a simple bin correction factor (used in plot_data/unfold)}
%\usage{set_bin_corr_factor(Integer_Type data_id, Double_Type[nbins] corr_factor);}
%\seealso{load_fermi,get_bin_corr_factor,plot_unfold}
%!%-
{
   variable id,corr_factor;
   switch(_NARGS)
   { case 2: (id,corr_factor) = (); }
   { help(_function_name()); return; }
   unfold_corr_factor[string(id)]=@corr_factor;
   return;
}
define get_bin_corr_factor()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_bin_corr_factor}
%\synopsis{reads the bin correction factor set by set_bin_corr_factor}
%\usage{Double_Type[nbins] corr_factor = set_bin_corr_factor(Integer_Type data_id)}
%\seealso{load_fermi,set_bin_corr_factor}
%!%-
{
   variable id;
   switch(_NARGS)
   { case 1: (id) = (); }
   { help(_function_name()); return; }
   if(assoc_key_exists(unfold_corr_factor, string(id)) )
     return @unfold_corr_factor[string(id)];
   else
     return NULL;   
}

%
% Code from src/slang/arrays/binary_search.sl
%
define binary_search()
%!%+
%\function{binary_search}
%\synopsis{searches a value in a sorted list}
%\usage{Integer_Type ndx=binary_search(Array_Type arr, Data_Type val);}
%\qualifiers{
%\qualifier{interval}{do an interval search, return
%   ndx such that arr[ndx]<=val<add[ndx+1], or 0 if val<=arr[0] or
%   length(arr) if add[-1]<=val, see description. }
%}
%\description
%  The function performs a binary search for the value val in the array
%  arr. All data types where the comparison operators work are allowed
%  (i.e., mainly all numerical types).
%
%  Values in arr MUST be unique and sorted. This is not checked.
%  The binary search is much faster than, e.g., a call to
%  where, since the where function is O(n), while binary_search
%  scales is O(lb(N)) where lb is the logarithm to base 2.
%
%  If no qualifiers are set, the function returns the index of the array
%  element if val is found. The function returns -1 if the value is not found.
%
%  If the interval qualifier is set, then the function will do an interval
%  search such that arr[ndx]<=value<arr[ndx]. The function will return
%  0 if value<=arr[0] and length(arr) if value>arr[-1]. This can be
%  used to search the index where one would insert a number into
%  an array.
%!%-
{
    variable arr;
    variable value;

    (arr,value)=();

    variable left,mid,right;
    
    
    if (qualifier_exists("interval")) {
        % returns ndx such that
        % arr[ndx]<=val<arr[ndx+1]
        variable ll=length(arr);

        left=0;
        right=ll;

        while (left<right) {
            mid=(left+right)/2;
            if (arr[mid]==value) {
                return mid;
            } else {
                if (arr[mid]<value) {
                    left=mid+1;
                } else {
                    right=mid-1;
                }
            }
        }
        if (left==ll) {
            return ll;
        }
        if (arr[left]>value) {
            return left;
        }
        return left+1;
    }
    
    % returns ndx such that arr[ndx]==val
    left=0;
    right=length(arr)-1;

    while (left<=right) {
        mid=(left+right)/2;
        if (arr[mid]<value) {
            left=mid+1;
        } else {
            if (arr[mid]>value) {
                right=mid-1;
            } else {
                return mid;
            }
        }
    }
    return -1;
}

%
% Code from src/slang/math/binarydigits.sl
%
%%%%%%%%%%%%%%%%%%%
define binarydigits(x)
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{binarydigits}
%\synopsis{retrieves a number's dual representaion}
%\usage{String_Type binarydigits(Integer_Type x)}
%\qualifiers{
%\qualifier{n}{[\code{=16}] number of bits}
%}
%!%-
{
  vmessage("warning: %s is deprecated, use  sprintf(\"%.<n>B\", ...)  instead", _function_name());

  variable s="";
  loop( qualifier("n", 16) )
  { s = string( x & 1 ) + s;
    x = x >> 1;
  }
  return s;
}

%
% Code from src/fitting/xyfit/fit-functions/bknpowerlaw_xyfit.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define bknpowerlaw_xyfit()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{bknpowerlaw_xyfit}
%\synopsis{linear xy fit function to be used with xyfit_fun}
%\usage{xyfit_fun ("bknpowerlaw");}
%\description
%    This function is not meant to be called directly!
%    
%    Calling \code{xyfit_fun ("bknpowerlaw");} sets up a powerlaw fit
%    function for xy-data. It has the form \code{y = norm*x^{-index}}
%\seealso{xyfit_fun, define_xydata, plot_xyfit, linear_regression}
%!%-
{
   variable xref, yref, par;
   switch(_NARGS)
   { case 0: return ["norm", "index1", "index2", "ebr"]; } % fit parameters
   { case 3: (xref, yref, par)=(); }
   { return help(_function_name); }
   
   variable ind  = where(@xref <= par[3]);
   variable ind2 = where(@xref > par[3]);

   (@yref)  = par[0] * ((@xref) /par[3])^(-par[1]);
   if (length(ind2)>0){
      (@yref)[ind2] = par[0] * ((@xref)[ind2]/par[3])^(-par[2]);
   }
}

define bknpowerlaw_xyfit_default(i)
{
  switch(i)
  { case 0: return (1, 0, -1e5, 1e5); }
  { case 1: return (1, 0, -1, 4); }
  { case 2: return (1, 0, -1, 4); }
  { case 3: return (1, 0, -1e5, 1e5); }
}

%
% Code from src/slang/arrays/blocks_between_gaps.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define blocks_between_gaps()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{blocks_between_gaps}
%\synopsis{retrieves blocks in an array whose elements may contain gaps}
%\usage{Integer_Type blocks[] = blocks_between_gaps(Double_Type a[], Double_Type gap);}
%\description
%    The array a has to contain monotonically increasing values.
%    The return value will be an integer array of the same length,
%    whose elements start with 0 and increase by 1 whenever the difference
%    of sequential elements of the array a is larger than gap.
%\seealso{split_struct, split_lc_at_gaps}
%!%-
{
  variable a, gap;
  switch(_NARGS)
  { case 2: (a, gap) = (); }
  { help(_function_name); return; }

  variable i, blocks = Integer_Type[length(a)];
  blocks[0] = 0;
  _for i (1, length(a)-1, 1)
  { blocks[i] = blocks[i-1];
    if(a[i]-a[i-1]>gap)  blocks[i]++;
  }
  return blocks;
}

%
% Code from src/misc/play/brew.sl
%
define brew (bev) 
{
  if (bev == "coffee") {
    ()=printf(`                      
                                   {
                                }   }   {
                               {   {  }  }
                                }   }{  {
                               {  }{  }  }
                              ( }{ }{  { )
                            .- { { }  { }} -.
                           (  ( } { } { } }  )
                           |-..________ ..-'|
                           |                 |
                           |                 ;--.
                           |                (__  \
                           |                 | )  )
                           |                 |/  /
                           |                 (  /
                           |                 y'
                           |                 |
                            -.._________..-'

`R) ;
  }
  
else if (bev == "beer") {
  ()=printf(`
	     @@@@@@@@@@@
	     @@@@@@@@@@@@@@@@@~
	    @@@@@@@@@@@@@@@@@@@@
	     @@@|===============|  ___
	      @@|	   	| /HBBb 
	       @|  °	  	|/B   )D 
	       @|	  	|/     )D 
		|	   ° 	|      ||
		|  °		|      //
		| SCHLENKERLA	|     //
		|  °		|    //
		|  	   ° 	|   //
		|  °	 	| //
		|  °	 	|B/
		|  °	   °	|)
		|  °	   	|
		|  °	   °	|
               aHHHHHHHHHHHHHHHHH

`R) ;
  } 

else if (bev == "martini") {
  ()=printf(`
             
    __________/____
    =_       /   _=
      =_    /  _=
        =_(/)_=
          =_=
           |
           |
           |
           |
    _______|______

`R) ;
  }

else if (bev == "wine") {
  ()=printf(`
             
     _____________
     |           |
     |           |
     '.         .'
       '-.___.-'
           |
           |
           |
           |
    _______|______

`R) ;
  }
 	
          
  else {()=printf("This beverage is not served here.... sorry!\n But maybe on the moon....? \n\n") ;
 ()=printf(`
  .            /\         .       .       
        .     /  \      .          .      
             /    \   .        .    *     
            /      \              *        
            | (__) |   .    .   **         
     .     /| (oo) |\           **         
          / | /\/\ | \   .     . *         
      .  /  |=|==|=|  \     .      *       
     . /    | |  | |   \  .               
      / USA | || | NASA \     .            
     |______|    |_______|       .         
    .       (__||__)     .   .             
       .    /_\  /_\  .     .    .         
            !!!  !!!                       
                                           
`); 
  }
}

%
% Code from src/slang/math/cart2cyl.sl
%
define cart2cyl()
%!%+
%\function{cart2cyl}
%\synopsis{Convert Cartesian to cylindrical coordinates}
%\usage{cart2cyl(Double_Types x[], y[], z[], vx[], vy[], vz[]);}
%\description
%    Convert Cartesian (x,y,z,vx,vy,vz) to cylindrical (r,phi,z,vr,vphi,vz)
%    coordinates with phi rotating counter-clockwise from the positive x-axis.
%\example
%    (r,phi,z,vr,vphi,vz) = cart2cyl(1,1,1,2,3,4);
%    (r,phi,z,vr,vphi,vz) = cart2cyl([1,0],[1,0],[1,0],[2,0],[3,0],[4,0]);
%    (r,phi,z,vr,vphi,vz) = cart2cyl( cyl2cart(1,PI/2,1,2,3,4) );
%\seealso{cart2sphere, cyl2cart}
%!%-
{
  if(_NARGS!=6)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable x, y, z, vx, vy, vz;
    (x ,y, z, vx, vy, vz) = ();
    x = [1.*x]; y = [1.*y]; z = [1.*z]; vx = [1.*vx]; vy = [1.*vy]; vz = [1.*vz];
    variable len = length(x);
    variable r, ind1, ind2;
    % radius:
    r = sqrt(x^2 + y^2); % radius
    ind1 = where( r!=0, &ind2 ); % ind1 -> r!=0; ind2 -> r==0
    variable vr = Double_Type[len]; % initialize vr = radial velocity
    vr[ind1] = ( x[ind1] * vx[ind1] + y[ind1] * vy[ind1] ) / r[ind1];
    vr[ind2] = sqrt( (vx[ind2])^2 + (vy[ind2])^2 );
    % polar angle phi:
    variable phi = Double_Type[len]; % initialize phi
    variable vphi = Double_Type[len]; % initialize vphi
    vphi[ind1] = (vy[ind1] * x[ind1] - vx[ind1] * y[ind1] ) / (r[ind1])^2;
    vphi[ind2] = 0.*vphi[ind2]; % in principle, this line is not necessary as already done in initialization
    variable ind = where(x==0 and y==0); % to determine phi for points at the origin use velocities instead of coordinates
    x[ind] = vx[ind];
    y[ind] = vy[ind];
    variable i;
    _for i(0, len-1, 1)
    {
      if (x[i]==0)
      {
	if (y[i]==0) phi[i] = 0;
	else
	{
	  if (y[i]>0) phi[i] = PI/2.;
	  else phi[i] = 3./2*PI;
	}
      }
      else
      {
	if (x[i]>0)
	{
	  if (y[i]==0) phi[i] = 0;
	  else
	  {
	    if (y[i]>0) phi[i] = atan(y[i]/x[i]);
	    else phi[i] = atan(y[i]/x[i])+2.*PI;
	  }
	}
	else phi[i] = atan(y[i]/x[i])+PI;
      }
    }

    if(length(x)==1) {return (r[0], phi[0], z[0], vr[0], vphi[0], vz[0]);}
    else return (r, phi, z, vr, vphi, vz);
  }
}

define cyl2cart()
%!%+
%\function{cyl2cart}
%\synopsis{Convert cylindrical to Cartesian coordinates}
%\usage{cyl2cart(Double_Types r[], phi[], z[], vr[], vphi[], vz[]);}
%\description
%    Convert cylindrical (r,phi,z,vr,vphi,vz) to Cartesian (x,y,z,vx,vy,vz)
%    coordinates with phi rotating counter-clockwise from the positive x-axis.
%\example
%    (x,y,z,vx,vy,vz) = cyl2cart(1,PI/2,1,2,3,4);
%    (x,y,z,vx,vy,vz) = cyl2cart([1,0],[PI/2,0],[1,0],[2,0],[3,0],[4,0]);
%    (x,y,z,vx,vy,vz) = cyl2cart( cart2cyl(1,1,1,2,3,4) );
%\seealso{cart2cyl}
%!%-
{
  if(_NARGS!=6)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable r, phi, z, vr, vphi, vz;
    (r, phi, z, vr, vphi, vz) = ();
    r = 1.*r; phi = 1.*phi; z = 1.*z; vr = 1.*vr; vphi = 1.*vphi; vz = 1.*vz;
    variable x, y, vx, vy;
    x = r*cos(phi);
    y = r*sin(phi);
    vx = vr*cos(phi) - r*sin(phi)*vphi;
    vy = vr*sin(phi) + r*cos(phi)*vphi;
    return (x, y, z, vx, vy, vz);
  }
}

%
% Code from src/slang/math/cart2sphere.sl
%
define cart2sphere()
%!%+
%\function{cart2sphere}
%\synopsis{Convert Cartesian to spherical coordinates}
%\usage{cart2sphere(Double_Types x[], y[], z[], vx[], vy[], vz[]);}
%\description
%    Convert Cartesian (x,y,z,vx,vy,vz) to spherical (r,phi,theta,vr,vphi,vtheta)
%    coordinates with polar angle phi rotating counter-clockwise from the positive
%    x-axis and azimuth angle theta being zero for the positive z-axis.
%\example
%    (r,phi,theta,vr,vphi,vtheta) = cart2sphere(0,0,0,1,2,3);
%    (r,phi,theta,vr,vphi,vtheta) = cart2sphere(0,0,1,1,2,3);
%    (r,phi,theta,vr,vphi,vtheta) = cart2sphere(1,1,1,3,3,3);
%    (r,phi,theta,vr,vphi,vtheta) = cart2sphere(1,0,0,0,2,1);
%    (r,phi,theta,vr,vphi,vtheta) = cart2sphere([0,1],[0,0],[0,0],[1,0],[2,2],[3,1]);
%\seealso{cart2cyl}
%!%-
{
  if(_NARGS!=6)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable x, y, z, vx, vy, vz;
    (x ,y, z, vx, vy, vz) = ();
    x = [1.*x]; y = [1.*y]; z = [1.*z]; vx = [1.*vx]; vy = [1.*vy]; vz = [1.*vz];
    variable len = length(x);
    % radius:
    variable r, ind1, ind2;
    r = sqrt(x^2 + y^2 + z^2); % radius
    ind1 = where( r!=0, &ind2 ); % ind1 -> r!=0; ind2 -> r==0
    variable vr = Double_Type[len]; % initialize vr = radial velocity
    vr[ind1] = ( x[ind1] * vx[ind1] + y[ind1] * vy[ind1] + z[ind1] * vz[ind1]) / r[ind1];
    vr[ind2] = sqrt( (vx[ind2])^2 + (vy[ind2])^2 + (vz[ind2])^2 );
    %
    variable ind3, ind4;
    ind3 = where( x^2+y^2!=0, &ind4 ); % ind3 -> x^2+y^2!=0; ind4 -> x^2+y^2==0
    % azimuth angle theta:
    variable theta = Double_Type[len]; % initialize theta
    theta[ind1] = acos( z[ind1]/r[ind1] );
    theta[ind2] = 0.*theta[ind2]; % in principle, this line is not necessary as already done in initialization
    variable vtheta = Double_Type[len]; % initialize vtheta
    vtheta[ind3] = ( z[ind3]*vr[ind3]/r[ind3] - vz[ind3] ) / sqrt( (x[ind3])^2 + (y[ind3])^2 );
    vtheta[ind4] = sqrt( (vx[ind4])^2 + (vy[ind4])^2 ) / z[ind4]; % on the z-axis: vr=vz, vphi=0 -> vtheta*z=sqrt(vx^2+vy^2)
    vtheta[ind2] = 0.*theta[ind2]; % in origin, vtheta is zero as the entire velocity is assigned to vr; NOTE: 0.*theta[ind2] and not 0.*vtheta[ind2] as vtheta[ind2] might be nan due to previous line
    % polar angle phi:
    variable vphi = Double_Type[len]; % initialize vphi
    vphi[ind3] = (vy[ind3] * x[ind3] - vx[ind3] * y[ind3] ) / ( (x[ind3])^2 + (y[ind3])^2 );
    vphi[ind4] = 0.*vphi[ind4]; % in principle, this line is not necessary as already done in initialization
    variable phi = Double_Type[len]; % initialize phi
    variable ind = where(x==0 and y==0); % to determine phi for points on the z-axis use velocities instead of coordinates
    x[ind] = vx[ind];
    y[ind] = vy[ind];
    variable i;
    _for i(0, len-1, 1)
    {
      if (x[i]==0)
      {
	if (y[i]==0) phi[i] = 0;
	else
	{
	  if (y[i]>0) phi[i] = PI/2.;
	  else phi[i] = 3./2*PI;
	}
      }
      else
      {
	if (x[i]>0)
	{
	  if (y[i]==0) phi[i] = 0;
	  else
	  {
	    if (y[i]>0) phi[i] = atan(y[i]/x[i]);
	    else phi[i] = atan(y[i]/x[i])+2.*PI;
	  }
	}
	else phi[i] = atan(y[i]/x[i])+PI;
      }
    }
    %
    if(length(x)==1) {return (r[0], phi[0], theta[0], vr[0], vphi[0], vtheta[0]);}
    else return (r, phi, theta, vr, vphi, vtheta);
  }
}

%
% Code from src/plot/cb_color_scheme.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

private define to_rgb (r, g, b){
          return (r << 16) | (g << 8) | b;
}

xfig_new_color("cb_orange",to_rgb(230,159,0));
xfig_new_color("cb_orange2",to_rgb(242,180,0));
xfig_new_color("cb_skyblue",to_rgb(86,180,233));
xfig_new_color("cb_bluishgreen",to_rgb(0,158,115));
xfig_new_color("cb_yellow",to_rgb(240,228,66));
xfig_new_color("cb_blue",to_rgb(0,114,178));
xfig_new_color("cb_vermillion",to_rgb(213,94,0));
xfig_new_color("cb_vermillion2",to_rgb(230,69,25));
xfig_new_color("cb_reddishpurple",to_rgb(204,121,167));


variable CB_COLOR_SCHEME = ["black","cb_vermillion","cb_blue","cb_orange","cb_bluishgreen",
		"cb_reddishpurple","cb_skyblue","cb_yellow"];
variable CB_COLOR_SCHEME_NB = [CB_COLOR_SCHEME[[1:]],"black"];

variable CB_COLOR_SCHEME_ALT = ["black","cb_vermillion2","cb_blue",
				"cb_orange2","cb_bluishgreen",
		"cb_reddishpurple","cb_skyblue","cb_yellow"];
variable CB_COLOR_SCHEME_ALT_NB = [CB_COLOR_SCHEME_ALT[[1:]],"black"];

variable CB_COLOR_SCHEME2 = ["black","cb_blue","cb_vermillion",
		 "cb_reddishpurple","cb_bluishgreen",
		 "cb_orange",
		  "cb_skyblue","cb_yellow"];
variable CB_COLOR_SCHEME2_NB = [CB_COLOR_SCHEME2[[1:]],"black"];



#endif

%
% Code from src/slang/astronomy/cel2gal.sl
%
define cel2gal() %{{{
%!%+
%\function{cel2gal}
%\synopsis{Transform celestial to Cartesian Galactic coordinates}
%\usage{cel2gal(Double_Types ah[], am[], as[], dd[], dm[], ds[], dist[], vrad[], pma_cos_d[], pmd[]; qualifiers)}
%\description
%    Transform celestial coordinates (right ascension [h, m, s], declination [deg, arcmin, arcsec],
%    distance [kpc], radial velocity [km/s], proper motion in right ascension times cosine of declination
%    [mas/yr], proper motion in declination [mas/yr]) to right-handed, Cartesian Galactic coordinates
%    (x [kpc], y [kpc], z [kpc], vx [km/s], vy [km/s], vz [km/s]) with the Galactic center at the origin,
%    the Sun on the negative x-axis and the z-axis pointing to the north Galactic pole implying clockwise
%    Galactic rotation when seen from the half space with positive z. See function 'RD2rad' for detailed
%    information on how to properly enter right ascension and declination.
%\qualifiers{
%\qualifier{parallax}{Set this qualifier to use parallaxes [mas] instead of distances [kpc].}
%\qualifier{SunGCDist}{Sun-Galactic center distance [kpc];
%      default: 8.4 (see Model I in Irrgang et al., 2013, A&A, 549, A137)}
%\qualifier{vxs}{Sun's x-velocity component [km/s] relative to the local standard of rest;
%      default: 11.1 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vys}{Sun's y-velocity component [km/s] relative to the local standard of rest;
%      default: 12.24 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vzs}{Sun's z-velocity component [km/s] relative to the local standard of rest;
%      default: 7.25 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vlsr}{Local standard of rest velocity [km/s];
%      default: 242 (see Model I in Irrgang et al., 2013, A&A, 549, A137)}
%\qualifier{GC_NGP}{Celestial coordinates of the Galactic center and the north Galactic pole in the
%      format [ah_GC, am_GC, as_GC, dd_GC, dm_GC, ds_GC, ah_NGP, am_NGP, as_NGP, dd_NGP, dm_NGP, ds_NGP];
%      default: [17, 45, 37.224, -28, 56, 10.23, 12, 51, 26.282, 27, 07, 42.01] (J2000, Reid & Brunthaler, 2004, ApJ, 616, 872)}
%}
%\example
%    (x, y, z, vx, vy, vz) = cel2gal(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
%    % -> Sun's position and velocity
%    (x, y, z, vx, vy, vz) = cel2gal(17, 45, 37.224, -28, 56, 10.23, 8.4, -11.1, -3.171, -5.544);
%    % -> Galactic center's position and velocity
%    (x, y, z, vx, vy, vz) = cel2gal([0,17], [0,45], [0,37.224], [0,-28], [0,56], [0,10.23], [0,8.4], [0,-11.1], [0,-3.171], [0,-5.544]);
%    % -> position and velocity of Sun and Galactic center stored in the same array
%    (x, y, z, vx, vy, vz) = cel2gal(17, 45, 37.224, -28, 56, 10.23, 8.4, -11.1, -3.171, -5.544;
%    GC_NGP=[17, 45, 37.1991, -28, 56, 10.221, 12, 51, 26.2755, 27, 7, 41.704]);
%    % -> use non-standard coordinates for Galactic center and north Galactic pole
%    (x, y, z, vx, vy, vz) = cel2gal( gal2cel(0, 0, 0, 0, 0, 0) );
%    % compute vr and vphi:
%    vr = (x*vx+y*vy)/sqrt(x^2+y^2);
%    vphi = (y*vx-x*vy)/sqrt(x^2+y^2);
%\seealso{RD2rad, gal2cel}
%!%-
{
  if(_NARGS!=10)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable ah, am, as, dd, dm, ds, dist, rv, pma, pmd;
    (ah, am, as, dd, dm, ds, dist, rv, pma, pmd)=();
    ah = [1.*ah]; am = [1.*am]; as = [1.*as]; dd = [1.*dd]; dm = [1.*dm]; ds = [1.*ds]; dist = [1.*dist]; rv = [1.*rv]; pma = [1.*pma]; pmd = [1.*pmd];
    if(qualifier_exists("parallax"))
      dist = 1./dist; % conversion from parallax in units of mas to distance in kpc
    variable ind = where(dist<0);
    if(length(ind)>0)
    {
      vmessage("Warning in '%s': Replaced %d negative distance(s) by _NaN(s).", _function_name(), length(ind));
      dist[ind] = _NaN;
    }
    variable SunGCDist = qualifier("SunGCDist", 8.40); % Model I in Irrgang et al., 2013, A&A, 549, A137
    variable vxs = qualifier("vxs", 11.1);  % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vys = qualifier("vys", 12.24); % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vzs = qualifier("vzs", 7.25);  % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vlsr = qualifier("vlsr", 242.0); % Model I in Irrgang et al., 2013, A&A, 549, A137
    % convert right ascension and declination to radians:
    variable ra, decl;
    (ra, decl) = RD2rad(ah, am, as, dd, dm, ds);
    % convert polar coordinates to Cartesian celestial coordinates:
    variable cartCC = Array_Type[3]; % [x, y, z]
    cartCC[0] = dist*cos(ra)*cos(decl);
    cartCC[1] = dist*sin(ra)*cos(decl);
    cartCC[2] = dist*sin(decl);
    % rotation matrix from celestial coordinates to coordinates that are aligned to Galactic enter (x-axis) and north Galactic pole (z-axis)
    % see below ("if(qualifier_exists("GC_NGP")==1)") for a derivation of the matrix
    variable rotMat = Double_Type[3,3];
    % initialize rotMat with matrix derived from coordinates for Galactic center (17 45 37.224, -28 56 10.23) and north Galactic pole (12 51 26.282, 27 07 42.01)
    % as given in Reid & Brunthaler, 2004, ApJ, 616, 872
    rotMat[0,0] = -0.05487395617553902;
    rotMat[0,1] = -0.8734371822248346;
    rotMat[0,2] = -0.48383503143198114;
    rotMat[1,0] =  0.4941107750655301;
    rotMat[1,1] = -0.44482861498054843;
    rotMat[1,2] =  0.7469819577865507;
    rotMat[2,0] = -0.8676654832928334;
    rotMat[2,1] = -0.1980782471692138;
    rotMat[2,2] =  0.45598422900423397;
    if(qualifier_exists("GC_NGP")==1) % use other coordinates for Galactic center and north Galactic pole -> re-calculate rotMat
    {
      variable a_GC, d_GC, a_NGP, d_NGP; % right ascension (a) and declination (d) of Galactic center (GC, x-axis) and north Galactic pole (NGP, z-axis)
      (a_GC, d_GC) = RD2rad( qualifier("GC_NGP")[0],  qualifier("GC_NGP")[1],  qualifier("GC_NGP")[2],
			     qualifier("GC_NGP")[3],  qualifier("GC_NGP")[4],  qualifier("GC_NGP")[5] );
      (a_NGP, d_NGP) = RD2rad( qualifier("GC_NGP")[6],  qualifier("GC_NGP")[7],  qualifier("GC_NGP")[8],
			       qualifier("GC_NGP")[9],  qualifier("GC_NGP")[10],  qualifier("GC_NGP")[11] );
      variable v_GC = [cos(a_GC)*cos(d_GC), sin(a_GC)*cos(d_GC), sin(d_GC)]; % x-axis
      variable v_NGP = [cos(a_NGP)*cos(d_NGP), sin(a_NGP)*cos(d_NGP), sin(d_NGP)]; % z-axis
      % v_GC and v_NGP are not totally orthogonal -> subtract projection of v_NGP on v_GC from v_NGP to make them orthogonal effectively changing the coordinates of
      % the north Galactic pole; however, effects ( (12 51 26.282, 27 07 42.01) -> (12 51 26.305713809790063, 27 07 41.80494829525571) ) are totally negligible to
      % uncertainty being approximately 0.1 degree (see Reid & Brunthaler, 2004, ApJ, 616, 872)
      v_NGP = v_NGP - sum(v_GC*v_NGP)*v_GC;
      v_NGP = v_NGP/sqrt(sum(v_NGP^2)); % normalization
      variable v_y = -[ v_GC[1]*v_NGP[2]-v_GC[2]*v_NGP[1], v_GC[2]*v_NGP[0]-v_GC[0]*v_NGP[2], v_GC[0]*v_NGP[1]-v_GC[1]*v_NGP[0]]; % y-axis
      rotMat[0,0] = v_GC[0];
      rotMat[0,1] = v_GC[1];
      rotMat[0,2] = v_GC[2];
      rotMat[1,0] = v_y[0];
      rotMat[1,1] = v_y[1];
      rotMat[1,2] = v_y[2];
      rotMat[2,0] = v_NGP[0];
      rotMat[2,1] = v_NGP[1];
      rotMat[2,2] = v_NGP[2];
    }
    % multiply the Celestial Cartesian data with the rotation matrix
    % to get the coordinates in Galactic system and add shift for the
    % sun-gc distance:
    variable x, y, z;
    x = rotMat[0,0]*cartCC[0] + rotMat[0,1]*cartCC[1] + rotMat[0,2]*cartCC[2] - SunGCDist;
    y = rotMat[1,0]*cartCC[0] + rotMat[1,1]*cartCC[1] + rotMat[1,2]*cartCC[2];
    z = rotMat[2,0]*cartCC[0] + rotMat[2,1]*cartCC[1] + rotMat[2,2]*cartCC[2];
    % velocity calculations:
    % convert dist from kpc to km, and SunGCDist from kpc to km:
    variable kpc_to_km = 3.08567802e+16; % conversion factor from kpc to km
    dist *= kpc_to_km;
    SunGCDist *= kpc_to_km;
    % convert pmd and pma from milliarcseconds/year to radians/s
    variable mas_yr_to_rad_s = 1.536282790e-16; % conversion factor from mas/yr to radians/s = 2.*PI/(360.*60.*60.)/1000./(365.24977652625375*24.*60.*60.);
    pmd *= mas_yr_to_rad_s;
    pma *= mas_yr_to_rad_s;
    % convert radial velocity to a Cartesian vector in Celestial coordinates
    variable radvCC = Array_Type[3]; % [vx, vy, vz]
    radvCC[0] = rv*cos(ra)*cos(decl);
    radvCC[1] = rv*sin(ra)*cos(decl);
    radvCC[2] = rv*sin(decl);
    % convert angular declination velocity to a Cartesian vector in Celestial coordinates
    variable dvCC = Array_Type[3]; % [vx, vy, vz]
    dvCC[0] = -pmd*dist*cos(ra)*sin(decl);
    dvCC[1] = -pmd*dist*sin(ra)*sin(decl);
    dvCC[2] =  pmd*dist*cos(decl);
    % convert angular right ascention velocity to a Cartesian vector in Celestial coordinates
    variable avCC = Array_Type[3]; %[vx, vy, vz]
    avCC[0] = -pma*dist*sin(ra); % factor cos(decl) already in pma
    avCC[1] =  pma*dist*cos(ra); % factor cos(decl) already in pma
    avCC[2] =  dist*0.;
    % sum the three velocity components:
    variable totalvCC = Array_Type[3]; % [vx, vy, vz]
    totalvCC[0] = radvCC[0] + dvCC[0] + avCC[0];
    totalvCC[1] = radvCC[1] + dvCC[1] + avCC[1];
    totalvCC[2] = radvCC[2] + dvCC[2] + avCC[2];
    % rotate the vector by the rotation matrix, and add the motions of the Sun within the local standard of rest (vxs, vys, vzs),
    % and the motion of the local standard of rest itself (vlsr):
    variable vx, vy, vz;
    vx = rotMat[0,0]*totalvCC[0] + rotMat[0,1]*totalvCC[1] + rotMat[0,2]*totalvCC[2] + vxs;
    vy = rotMat[1,0]*totalvCC[0] + rotMat[1,1]*totalvCC[1] + rotMat[1,2]*totalvCC[2] + vys + vlsr;
    vz = rotMat[2,0]*totalvCC[0] + rotMat[2,1]*totalvCC[1] + rotMat[2,2]*totalvCC[2] + vzs;
    if(length(x)==1) {return (x[0], y[0], z[0], vx[0], vy[0], vz[0]);}
    else return (x, y, z, vx, vy, vz);
  }
}%}}}

define gal2cel() %{{{
%!%+
%\function{gal2cel}
%\synopsis{Transform Cartesian Galactic to celestial coordinates}
%\usage{gal2cel(Double_Types x[], y[], z[], vx[], vy[], vz[]; qualifiers)}
%\description
%    Transform right-handed, Cartesian Galactic coordinates (x [kpc], y [kpc], z [kpc], vx [km/s],
%    vy [km/s], vz [km/s]) with the Galactic center at the origin, the Sun on the negative x-axis
%    and the z-axis pointing to the north Galactic pole implying clockwise Galactic rotation when
%    seen from the half space with positive z to celestial coordinates (right ascension [h, m, s],
%    declination [deg, arcmin, arcsec], distance [kpc], radial velocity [km/s], proper motion in
%    right ascension times cosine of declination [mas/yr], proper motion in declination [mas/yr]).
%\qualifiers{
%\qualifier{SunGCDist}{Sun-Galactic center distance [kpc];
%      default: 8.4 (see Model I in Irrgang et al., 2013, A&A, 549, A137)}
%\qualifier{vxs}{Sun's x-velocity component [km/s] relative to the local standard of rest;
%      default: 11.1 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vys}{Sun's y-velocity component [km/s] relative to the local standard of rest;
%      default: 12.24 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vzs}{Sun's z-velocity component [km/s] relative to the local standard of rest;
%      default: 7.25 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vlsr}{Local standard of rest velocity [km/s];
%      default: 242 (see Model I in Irrgang et al., 2013, A&A, 549, A137)}
%\qualifier{GC_NGP}{Celestial coordinates of the Galactic center and the north Galactic pole in the
%      format [ah_GC, am_GC, as_GC, dd_GC, dm_GC, ds_GC, ah_NGP, am_NGP, as_NGP, dd_NGP, dm_NGP, ds_NGP];
%      default: [17, 45, 37.224, -28, 56, 10.23, 12, 51, 26.282, 27, 07, 42.01] (Reid & Brunthaler, 2004, ApJ, 616, 872)}
%}
%\example
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel(-8.4, 0, 0, 11.1, 254.24, 7.25);
%	% -> Sun's position and velocity
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel(0, 0, 0, 0, 0, 0);
%	% -> Galactic center's position and velocity
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel([-8.4,0], [0,0], [0,0], [11.1,0], [254.24,0], [7.25,0]);
%	% -> position and velocity of Sun and Galactic center stored in the same array
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel(-8.4, 0, 0, 11.1, 254.24, 7.25;
%	GC_NGP=[17, 45, 37.1991, -28, 56, 10.221, 12, 51, 26.2755, 27, 7, 41.704]);
%	% -> use non-standard coordinates for Galactic center and north Galactic pole
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel( cel2gal(17, 45, 37.224, -28, 56, 10.23, 8.33, -11.1, -2.91, -5.12) );
%\seealso{cel2gal, rad2RD}
%!%-
{
  if(_NARGS!=6)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable x, y, z, vx, vy, vz;
    (x, y, z, vx, vy, vz) = ();
    x = [1.*x]; y = [1.*y]; z = [1.*z]; vx = [1.*vx]; vy = [1.*vy]; vz = [1.*vz];
    variable SunGCDist = qualifier("SunGCDist", 8.40); % see Model I in Irrgang et al., 2013, A&A, 549, A137
    variable vxs = qualifier("vxs", 11.1);  % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vys = qualifier("vys", 12.24); % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vzs = qualifier("vzs", 7.25);  % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vlsr = qualifier("vlsr", 242.0);  % see Model I in Irrgang et al., 2013, A&A, 549, A137
    % shift Cartesian coordinate system from Galactic center to Sun:
    x += SunGCDist;
    vx -= vxs;
    vy -= (vys + vlsr);
    vz -= vzs;
    % rotation matrix from Galactically aligned to celestially aligned Cartesian coordinates
    % see below ("if(qualifier_exists("GC_NGP")==1)") for a derivation
    variable rotMat = Double_Type[3,3];
    % initialize rotMat with matrix derived from coordinates for Galactic center (17 45 37.224, -28 56 10.23) and north Galactic pole (12 51 26.282, 27 07 42.01)
    % as given in Reid & Brunthaler, 2004, ApJ, 616, 872
    rotMat[0,0] = -0.05487395617553902;
    rotMat[1,0] = -0.8734371822248346;
    rotMat[2,0] = -0.48383503143198114;
    rotMat[0,1] =  0.4941107750655301;
    rotMat[1,1] = -0.44482861498054843;
    rotMat[2,1] =  0.7469819577865507;
    rotMat[0,2] = -0.8676654832928334;
    rotMat[1,2] = -0.1980782471692138;
    rotMat[2,2] =  0.45598422900423397;
    if(qualifier_exists("GC_NGP")==1) % use other coordinates for Galactic center and north Galactic pole -> re-calculate rotMat
    {
      variable a_GC, d_GC, a_NGP, d_NGP; % right ascension (a) and declination (d) of Galactic center (GC, x-axis) and north Galactic pole (NGP, z-axis)
      (a_GC, d_GC) = RD2rad( qualifier("GC_NGP")[0],  qualifier("GC_NGP")[1],  qualifier("GC_NGP")[2],
			     qualifier("GC_NGP")[3],  qualifier("GC_NGP")[4],  qualifier("GC_NGP")[5] );
      (a_NGP, d_NGP) = RD2rad( qualifier("GC_NGP")[6],  qualifier("GC_NGP")[7],  qualifier("GC_NGP")[8],
			       qualifier("GC_NGP")[9],  qualifier("GC_NGP")[10],  qualifier("GC_NGP")[11] );
      variable v_GC = [cos(a_GC)*cos(d_GC), sin(a_GC)*cos(d_GC), sin(d_GC)]; % x-axis
      variable v_NGP = [cos(a_NGP)*cos(d_NGP), sin(a_NGP)*cos(d_NGP), sin(d_NGP)]; % z-axis
      % v_GC and v_NGP are not totally orthogonal -> subtract projection of v_NGP on v_GC from v_NGP to make them orthogonal effectively changing the coordinates of
      % the north Galactic pole; however, effects ( (12 51 26.282, 27 07 42.01) -> (12 51 26.305713809790063, 27 07 41.80494829525571) ) are totally negligible to
      % uncertainty being approximately 0.1 degree (see Reid & Brunthaler, 2004, ApJ, 616, 872)
      v_NGP = v_NGP - sum(v_GC*v_NGP)*v_GC;
      v_NGP = v_NGP/sqrt(sum(v_NGP^2)); % normalization
      variable v_y = -[ v_GC[1]*v_NGP[2]-v_GC[2]*v_NGP[1], v_GC[2]*v_NGP[0]-v_GC[0]*v_NGP[2], v_GC[0]*v_NGP[1]-v_GC[1]*v_NGP[0]]; % y-axis
      rotMat[0,0] = v_GC[0];
      rotMat[1,0] = v_GC[1];
      rotMat[2,0] = v_GC[2];
      rotMat[0,1] = v_y[0];
      rotMat[1,1] = v_y[1];
      rotMat[2,1] = v_y[2];
      rotMat[0,2] = v_NGP[0];
      rotMat[1,2] = v_NGP[1];
      rotMat[2,2] = v_NGP[2];
    }
    % rotate Cartesian coordiante system so that it is aligned to right ascension and declination:
    variable cartCC = Array_Type[3]; % Cartesian celestial coordinates
    cartCC[0] = rotMat[0,0]*x + rotMat[0,1]*y + rotMat[0,2]*z;
    cartCC[1] = rotMat[1,0]*x + rotMat[1,1]*y + rotMat[1,2]*z;
    cartCC[2] = rotMat[2,0]*x + rotMat[2,1]*y + rotMat[2,2]*z;
    variable cartCV = Array_Type[3]; % Cartesian celestial velocities
    cartCV[0] = rotMat[0,0]*vx + rotMat[0,1]*vy + rotMat[0,2]*vz;
    cartCV[1] = rotMat[1,0]*vx + rotMat[1,1]*vy + rotMat[1,2]*vz;
    cartCV[2] = rotMat[2,0]*vx + rotMat[2,1]*vy + rotMat[2,2]*vz;
    variable kpc_to_km = 3.08567802e+16; % conversion factor from kpc to km
    cartCC[0] *= kpc_to_km;
    cartCC[1] *= kpc_to_km;
    cartCC[2] *= kpc_to_km;
    variable dis, ra, theta, vrad, pma, pmd;
    % transform Cartesian to spherical coordinates:
    (dis, ra, theta, vrad, pma, pmd) = cart2sphere(cartCC[0], cartCC[1], cartCC[2], cartCV[0], cartCV[1], cartCV[2]);
    variable decl = -theta + PI/2.; % due to different definitions of azimuth angle theta and declination delta
    pmd *= -1.; % due to different definitions of azimuth angle theta and declination delta
    variable ah, am, as, dd, dm, ds;
    (ah, am, as, dd, dm, ds) = rad2RD(ra,decl);
    variable mas_yr_to_rad_s = 1.536282790e-16; % conversion factor from mas/yr to radians/s = 2.*PI/(360.*60.*60.)/1000./(365.24977652625375*24.*60.*60.);
    pmd /= mas_yr_to_rad_s;
    pma /= mas_yr_to_rad_s;
    pma *= cos(decl); % we use pma*cos(decl) instead of pma alone
    dis /= kpc_to_km;
    % return output:
    if(length(dis)==1) {return (ah[0], am[0], as[0], dd[0], dm[0], ds[0], dis[0], vrad[0], pma[0], pmd[0]);}
    else return (ah, am, as, dd, dm, ds, dis, vrad, pma, pmd);
  }
}%}}}

%
% Code from src/plot/change_plot_options.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define change_plot_options()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{change_plot_options}
%\synopsis{changes the currently used plot options}
%\usage{Struct_Type change_plot_options([plot_opt]; [default,] opt=value, opt2=value2, ...);}
%\description
%    The current plot options will be used as a starting point unless
%    an explicit argument \code{plot_opt} is specified, which is used in this case.
%
%    These plot options are then modified by the \code{opt} qualifiers (\code{opt} can be
%    any fieldname of the structure returned by \code{get_plot_options}, i.e.,
%    \code{xmin}, \code{xmax}, \code{ymin}, \code{ymax}, \code{xlabel}, \code{ylabel}, \code{tlabel}, \code{xopt}, \code{yopt}, \code{logx}, \code{logy},
%    \code{color}, \code{start_color}, \code{x_unit}, \code{line_style}, \code{start_line_style}, \code{line_width},
%    \code{frame_line_width}, \code{point_style}, \code{connect_points}, \code{char_height}, \code{point_size},
%    \code{ebar_term_length}, \code{use_errorbars}, \code{use_bin_density},
%    \code{ovp_xmin}, \code{ovp_xmax}, \code{ovp_ymin}, \code{ovp_ymax}),
%    unless the qualifier \code{default} is specified. In this latter case,
%    default plot options are set regardless of the starting plot options
%    and the other qualifiers.
%
%    The return value is the \code{get_plot_options} structure before \code{change_plot_options}
%    was called. It can be used to reset the plot options after a change was applied.
%\example
%    \code{variable plot_options = change_plot_options(; line_style=2);}\n
%    \code{plot(lineX, lineY);}\n
%    \code{set_plot_options(plot_options);}\n
%\seealso{get_plot_options, set_plot_options}
%!%-
{
  if(_NARGS>1)  { help(_function_name()); return; }

  variable initial_plot_options = get_plot_options();

  if(qualifier_exists("default"))
  {
    xrange;
    yrange;
    label("", "", "");
    xlin;
    ylin;
    color(1);
    % plot_unit("A or keV?");
    line_style(1);
    set_line_width(1);
    pointstyle(-1);
    connect_points(1);
    charsize(1);
    point_size(1);
    errorbars(0);
    plot_bin_integral;
    set_outer_viewport( struct { xmin=0.075765, xmax=0.924235, ymin=0.1, ymax=0.9 } );
    return initial_plot_options;
  }

  variable plot_options = (_NARGS ? () : initial_plot_options);
  plot_options = @plot_options;  % make a copy in order not to change the initial structure

  variable option;
  foreach option (get_struct_field_names(plot_options))
    if(qualifier_exists(option))
      set_struct_field(plot_options, option, qualifier(option));
  set_plot_options(plot_options);

  return initial_plot_options;
}

%
% Code from src/slang/math/chebyshev.sl
%
% predeclaration for a recursive function
define chebyshev_poly();

define chebyshev_poly(x,n) 
%!%+
%\function{chebyshev_poly}
%\synopsis{Returns the Chebyshev polynomial T(n,x) of the first kind of order n}
%\usage{Double_Type[] chebyshev_poly(Double_Type[] x, n)}
%\description
%    Evaluates the Chebyshev Polynomial of the first kind of order n
%    via the recurrence relation
%    
%    T(0,x) = 1
%    T(1,x) = x
%    T(n,x) = 2*x*T(n-1,x) - T(n-2,x)
%\example
%    x = [-1.:1.:#100];
%    chebyshev_poly(x,n); % returns 5th order Chebyshev polynomial evaluated
%                         % on [-1,1)
%\seealso{cheby_sum}
%!%-
{
  if (n==0) return 1;
  if (n==1) return x;
  return 2*x*chebyshev_poly(x,n-1) - chebyshev_poly(x, n-2);
}


define cheby_sum(coeffs, x)
%!%+
%\function{cheby_sum}
%\synopsis{Returns a weighted sum of Chebyshev polynomials up to a given order}
%\usage{Double_Type[] cheby_sum(Double_Type[] [a0,a1,...aN], Double_Type[] x)}
%\description
%    Return the value of the expression
%
%    a0*T(0,x) + a1*T(1,x) + a2*T(2,x) + ... + aN*T(N,x)
%
%    Where T are Chebyshev polynomials of the first kind of order 1...N.
%
%    This is array-safe in x and implemented to be much more efficient than
%    explicitly using multiple calls of chebyshev_poly()!
%\example
%    x = [-1.:1.:#100];
%    cheby_sum(x,[0, 1., 2.]); % returns chebyshev_poly(x,1)
%                              %    + 2.*chebyshev_poly(x,2)
%                         % on [-1,1)
%\seealso{chebyshev_poly}
%!%-
{
  % get the maximum order from coeffs
  variable ord = length(coeffs)-1;

  % trivial return
  if (ord == 0) return coeffs[0]*ones(length(x));

  % build a vector of the evaluated chebyshev polynomials without
  % the coefficients
  variable evals = Double_Type[length(x),ord+1];

  evals[*,0] = 1;
  evals[*,1] = x;

  variable ii;
  for (ii=2; ii<ord+1; ii++) {
    % use the recursion relation
    evals[*,ii] = 2*x*evals[*,ii-1] - evals[*,ii-2];
  }

  % multiply with the coefficients (did not do this earlier so we could
  % take advantage of recursion)
  for (ii=0; ii<ord+1; ii++) {
    evals[*,ii] *= coeffs[ii];
  }

  % return the sum over the 2nd dimension
  return sum(evals, 1);
}

%
% Code from src/fitting/parameters/check_parameter_hit.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define check_parameter_hit()
%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable thres = qualifier("thres", 1e-3);

  variable info;
  foreach info (get_params())
    if(   (info.value-info.min)/(info.max-info.min) < thres
       || (info.max-info.value)/(info.max-info.min) < thres )
      vmessage("%-25s: value = %-10g, min = %-10g, max = %-10g", info.name, info.value, info.min, info.max);
}

%
% Code from src/misc/simulation/checkrwlc.sl
%
define checkrwlc()
%!%+
%\function{checkrwlc [ff]}
%\usage{Struct_Type checkrwlc(time, value, error, bin_lo, bin_hi)
%\altusage{Struct_Type checkrwlc (time, value, bin_lo, bin_hi) % in which case resu.erro will be zero}
%}
%\description
%    This script implements the alogrithm proposed by M. de Kool et al. (1993)
%    to check for random walk in a pulse period evolution.
%    In principle, everything containing time and value information
%    can be checked for random walk with this script (i.e. a lightcurve).
%    The output of the script is a structure containing the fields\n
%      - bin_lo = lower delta t bin\n
%      - bin_hi = upper delta t bin\n
%      - value  = delta omega value\n
%      - erro = uncertainties calculated from the given uncertainties of the values\n
%      - dist  = not normalized delta omega value\n
%      - len = normalization facotr for delta omega (value = dist/len)
%
%    Care should be taken that, when using for period evolution, the period and the
%    time array should be given in the same units.
%
%    The code might not be perfectly optimized
%    and is getting slow for large arrays of time and value (>500).
%
%    To estimate the true errors of the result, a Monte Carlo simulation should be used.
%\qualifiers{
%\qualifier{grp_r}{[Double]: grouping parameter as defined in Eq. 5, de Kool (1993) (default 0.1)}
%\qualifier{verbose}{[Boolean]: print progress in form of current working timebin (default: false)}
%}
%\seealso{M. de Kool and U. Anzer, 1993, MNRAS, 262, 726}
%!%-
{
  variable a, b, er, lo, hi;
  switch(_NARGS)
  { case 4: (a, b, lo, hi) = (); er = 0*a ; }
  { case 5: (a, b, er, lo, hi) = (); }
  { help(_function_name()); return; }

  variable R = qualifier("grp_r", 0.1);  % group definition parameter (Eq. 5)
  variable chatty = qualifier_exists("verbose"); % print out progress or not?
  
  variable length_a = length(a);
  variable length_lo = length(lo);
  variable lenarr = Double_Type[length_lo];
  variable erroarr = Double_Type[length_lo];
  variable distarr = Double_Type[length_lo];
  variable i, j;
  _for j (0, length_lo-1, 1)
  {
    % j = 2 ;
    variable tmpres = Struct_Type[length_a];
    variable groupe = Double_Type[length_a] + 1;  % at least one point belongs to each group
    variable len = 0 ;
    _for i (1, length_a-1, 1)
    {
      % message("-=-=-=-=-=-=-=-=-=-=-=-=-=-=-") ;
      variable dt = a - shift(a,-i);  % what is the difference between two times in the distance i?
      % print(dt) ;
      % message("--") ;
      % print(lo[j]) ;
      % print(hi[j]) ;
      variable pairs = where(lo[j] <= dt <= hi[j]);  % check which differences in time lay in the regarded timebin
      % print(pairs) ;
      groupe[ where(abs(dt) <= R*lo[j]) ]++ ; % count how many other points belong to the point's group
% print(where(abs(dt) <= R*lo[j]) );
	% print((dt)) ;
      % message("----") ;
      
      tmpres[i-1] = struct {
	dist = abs(b[pairs-i]-b[pairs]),
	erro = sqrt((er[pairs-i])^2+(er[pairs])^2),
	pairs = pairs,
	shift = i };
      
      % dist += sum(abs(b[pairs-i]-b[pairs]));  % calculate the distance for all the pairs having the correct difference in time
      len += length(pairs);  % save the number of values going into dist to calc the average
      % print(tmpres[i-1].erro) ;
      % message("_------------------------_") ;
    }
    if(chatty)  vmessage("Iteration %d, dt = %f - %f", j, lo[j], hi[j]);
    % print(tmpres[1]) ;
    % message("-----" ) ;
    % print(groupe) ;
    % message("-=--=" );

    variable tmpdist = 0.0;
    variable tmperro = 0.0;
    variable reno = 0.0;
    variable nopairs = 0;

      % print(groupe[tmpres[0].pairs]) ;

    _for i (0, length_a-2, 1)
    { % applying the weight of the indivdual pairs to the respective pair in the sum
      variable w = 1. / groupe[tmpres[i].pairs] / groupe[tmpres[i].pairs-tmpres[i].shift];
      tmpdist += sum( w * tmpres[i].dist );
      tmperro += sqrt(sum( w * tmpres[i].erro^2 ));
	% print(w) ;
	% print(tmperro) ;
	% message("-------") ;
      reno += sum(w);
      nopairs += length(tmpres[i].pairs);
    }
    % message(",.,.,.,.,.,.,.,..,.,.") ;

    % renormalizing the weighted average
    % print(reno) ;
    % message("*****") ;
    % print(nopairs) ;
    % message("===========" );
    lenarr[j] = len;  % save the mean for each (lo,hi) bin in an array
    distarr[j] = tmpdist/reno*nopairs;
    erroarr[j] = tmperro/reno*nopairs;
  }

  return struct {
    bin_lo = lo,
    bin_hi = hi,
    value = distarr/lenarr,
    err = erroarr/lenarr,
    dist = distarr,
    len = lenarr
  };
}

%
% Code from src/slang/math/cholesky_decomposition.sl
%
define cholesky_decomposition() %{{{
%!%+
%\function{cholesky_decomposition}
%\synopsis{Decompose a matrix into the product of a lower triangular matrix and its transpose}
%\usage{Double_Type cd[n,n] = cholesky_decomposition(Double_Type m[n,n])}
%\description
%    The Cholesky decomposition is a decomposition of a symmetric, positive-definite
%    matrix into the (matrix) product of a lower triangular matrix and its transpose:
%      m = cd # transpose(cd).
%    The Cholesky–Banachiewicz algorithm is used here to carry out the decomposition.
%\notes
%    The Cholesky decomposition can be used to generate correlated variables that obey
%    a given covariance matrix. See the example section for details.
%\example
%    m = Double_Type[3,3]; m[0,*] = [ 9,-3,-6]; m[1,*] = [-3,10, 5]; m[2,*] = [-6, 5, 6];
%    print(m);                       % symmetric, positive-definite matrix
%    cd = cholesky_decomposition(m);
%    print(cd);                      % lower triangular matrix
%    print(cd#transpose(cd));        % matrix product of 'cd' and its transpose
%
%    % From three uncorrelated normal variables 'g', generate three correlated normal
%    % variables 'x' that obey the covariance matrix 'm':
%
%    g = Double_Type[3,100000];
%    g[0,*] = grand(100000); g[1,*] = grand(100000); g[2,*] = grand(100000);
%    (cov_mat, cor_mat) = covariance_correlation_matrix(g);
%    print(cov_mat); % the covariance matrix of 'g' shows that there are no correlations
%
%    x = cd#g;
%    (cov_mat, cor_mat) = covariance_correlation_matrix(x);
%    print(cov_mat); % the covariance matrix of 'x' is indeed (almost) 'm'
%\seealso{covariance_correlation_matrix}
%!%-
{
  if(_NARGS!=1)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable m = ();
  m *= 1.; % cast to Double_Type
  variable info = struct{ dims, ndims, type };
  (info.dims, info.ndims, info.type) = array_info(m);
  if(info.ndims!=2 || info.dims[0]!=info.dims[1] || info.type!=Double_Type)
    throw UsageError, sprintf("Usage error in '%s': The argument has to be a square matrix composed of Double_Types.", _function_name());
  variable cd = 0.*Double_Type[info.dims[0],info.dims[1]]; % Cholesky decomposed matrix initialized with zeros
  variable i, j;
  _for i(0, info.dims[0]-1, 1)
  {
    _for j(0, i, 1)
    {
      variable ind = [0:j-1:1];
      variable temp = m[i,j] - sum(cd[i,ind]*cd[j,ind]);
      if(i>j)
	cd[i,j] = temp/cd[j,j];
      else if(temp>0)
	cd[i,j] = sqrt(temp);
      else
	throw UsageError, sprintf("Usage error in '%s': The argument has to be a symmetric positive-definite matrix.", _function_name());
    }
  }
  return cd;
}%}}}

%
% Code from src/misc/math/circle_from_points.sl
%
define circle_from_points(x,y) 
%!%+
%\function{circle_from_points}
%\synopsis{find the best fitting circle for a set of points}
%\usage{pars=circle_from_points(Array_Type x,Array_Type y);
%}
%\description
%    Given arrays of x/y coordinates of points which lie approximately on
%    a circle, this function returns the center coordinate and radius of
%    the circle. This is a numerically unstable procedure, but the
%    routine generally returns parameters which are good enough to start
%    a more advanced fitting routine.
%    The function is an implementation of ideas described by L. Maisonobe
%    in a document entitled "Finding the circle that best fits a set of
%    points" (see http://www.spaceroots.org/downloads.html): For each
%    tuple of three points of  coordinates it finds the center
%    and radius, which is possible in an exact manner, assuming that
%    the points are not on a straight line. This is done for all or a
%    subset of the points, and the resulting radius and center coordinates
%    are then averaged.
%    The function returns a structure with the tags xc and yc (x- and y-
%    coordinate of the center of the circle) and radius (radius of the circle),
%    or NULL if no solution could be found (points are on a straight line).
%
% \qualifiers{
%    \qualifier{eps}{if the determinant of the three point solution is less than
%                      than this number, the points lie on a straight line}    
%    \qualifier{maxiter}{average at most maxiter tuples of three points}
%}
%!%-
{
    if (length(x)<3 or length(y)<3) {
	throw UsageError, sprintf("Usage error in %s: need at least three points to estimate circle",_function_name());
    }

    if (length(x)!=length(y)) {
	throw UsageError, sprintf("Usage error in %s: need the same number of x- and y-coordinates", _function_name());
    }

    variable eps=qualifier("eps",1e-7);
    variable maxiter=qualifier("maxiter",1000);
    
    variable npt=length(x);

    % helpers
    variable xy2=x*x+y*y;
    variable sx=0.;
    variable sy=0.;
    variable nn=0;
    variable i,j,k,xc,yc,del;
    variable totnum=(npt-2.)*(npt-1.)*npt;

    if (totnum<maxiter) {
	% go through all triple combinations, determine
	% the circle position, and average all positions
	_for i (0,npt-3,1) {
	    _for j (i+1,npt-2,1) {
		_for k (j+1,npt-1,1) {
		    del=2*((x[k]-x[j])*(y[j]-y[i])-(x[j]-x[i])*(y[k]-y[j]));
		    if (del>=eps) {
			xc= ((y[k]-y[j])*xy2[i]+(y[i]-y[k])*xy2[j]+(y[j]-y[i])*xy2[k])/del;
			yc=-((x[k]-x[j])*xy2[i]+(x[i]-x[k])*xy2[j]+(x[j]-x[i])*xy2[k])/del;
			sx+=xc;
			sy+=yc;
			nn++;
		    }
		}
	    }
	}
    } else {
	% too many points: go through a random subset of points
	% (note: tupels where two coordinates are the same will
	% automatically be rejected)
	variable ii;
	_for ii(0,maxiter-1,1) {
	    i=int(npt*urand());
	    j=int(npt*urand());
	    k=int(npt*urand());
	    del=2*((x[k]-x[j])*(y[j]-y[i])-(x[j]-x[i])*(y[k]-y[j]));
	    if (del>=eps) {
		xc= ((y[k]-y[j])*xy2[i]+(y[i]-y[k])*xy2[j]+(y[j]-y[i])*xy2[k])/del;
		yc=-((x[k]-x[j])*xy2[i]+(x[i]-x[k])*xy2[j]+(x[j]-x[i])*xy2[k])/del;
		sx+=xc;
		sy+=yc;
		nn++;
	    }
	}
    }
    % no radius found: points lie on a line
    if (nn==0) {
	return NULL;
    }
    % center coordinates
    xc=sx/nn;
    yc=sy/nn;
    variable ra=mean(sqrt((x-xc)^2.+(y-yc)^2.));

    return struct {xc=xc,yc=yc,radius=ra};
}

%
% Code from src/fitting/cl_save.sl
%
%%%%%%%%%%%%%%
define cl_save()
%%%%%%%%%%%%%%
%!%+
%\function{cl_save}
%\synopsis{computs single-parameter confidence lmits for several parameters, using multiple cores on one machine}
%\usage{Struct_Type results = cl_save([Integer_Type pars[]]);}
%\qualifiers{
%\qualifier{strict}{[=1]: restarts the calculation if a new best fit was found}
%\qualifier{saveoutput}{[=1]}
%\qualifier{basefilename}{[=<date_time>]}
%\qualifier{level}{[=1]: specifies the confidence level. Values of 0, 1, or 2
%                 indicate 68%, 90%, or 99% confidence levels respectively.
%                 By default, 90% confidence limits are computed.}
%\qualifier{tolerance}{convergence criterion for the calculation of the confidence
%      limits (see help for the conf command). Default: 1e-3}
%\qualifier{cleanup}{will remove all temporary files ending in *.[0-9][0-9]* from the basefilename directory}
%}
%\description
%    This function is a direct copy of fit_pars, however, using
%    "conf_loop" to allow multi-core support on one machine. The
%    number of slaves is determined by the global variable
%    Isis_Slaves.num_slaves.
%    The return value \code{results = struct { index, name, value, min, max, conf_min, conf_max, buf_below, buf_above, tex }}
%    is a table with the following information for each parameter:\n
%    \code{min} and \code{max} are the minimum/maximum values allowed.
%    \code{conf_min} and \code{conf_max} are the confidence limits.
%    \code{buf_below} (\code{buf_above}) is the fraction of the allowed range \code{[min:max]}
%    which separates the lower (upper) confidence limit from \code{min} (\code{max}).
%    If one of these buffers is 0, your confidence interval has bounced.
%\seealso{fit_pars, pvm_fit_pars, conf}
%!%-

{
   variable pars ;
   switch(_NARGS)
     { case 0: pars = thawedParameters() ; }
     { case 1: pars = () ;}
     { help(_function_name()); return; }

   
   variable lvls = qualifier("levels", 1);
   variable tol = qualifier("tolerance", 1e-3) ;

   variable saveoutput = qualifier("saveoutput", 1);
   variable basefilename = qualifier("basefilename", strftime("%Y-%m-%d_%H:%M:%S", localtime(_time())))+ "_";

  variable n = length(pars);
  variable results = struct {
    index     = pars,
    name      = String_Type[n],
    value     = Double_Type[n],
    min       = Double_Type[n],
    max       = Double_Type[n],
    conf_min  = Double_Type[n],
    conf_max  = Double_Type[n],
    buf_below = Double_Type[n],
    buf_above = Double_Type[n],
    tex       = String_Type[n],
  };

     variable i, info;
     variable clquals ;
   if (saveoutput) 
     {
	clquals = struct_combine(__qualifiers, struct{save=1,prefix=basefilename});
     }
   else clquals = __qualifiers ;
   
   print(clquals) ;
   
   (results.conf_min, results.conf_max ) = conf_loop(pars,lvls,tol;; clquals) ;

   if(saveoutput)
        try { save_par(basefilename+"best.par"); }
        catch AnyError: { vmessage("Could not write to %sbest.par.", basefilename); }

   _for i (0, n-1, 1)
       { 
	  info = get_par_info(pars[i]);
	  results.name[i] = info.name;
	  results.min[i] = info.min;
	  results.max[i] = info.max;
  	  results.value[i] = info.value;
  	  results.buf_below[i] = (results.conf_min[i] - info.min)/(info.max-info.min);
	  results.buf_above[i] = (info.max - results.conf_max[i])/(info.max-info.min);
	  results.tex[i] = TeX_value_pm_error(info.value, results.conf_min[i], results.conf_max[i]);
	  vmessage("(%10g <=) %10g < %25s = %10g < %10g (<= %10g)   [%4.1f%% (conf)%5.1f%%]   %s",
	       results.min[i], results.conf_min[i], results.name[i], results.value[i], results.conf_max[i], results.max[i],
  	       100*results.buf_below[i], 100*results.buf_above[i], results.tex[i]);
       }

   variable stat;
  ()=eval_counts(&stat);

  if(saveoutput)
    try
    {
      fits_write_binary_table(basefilename+"conf.fits", "CL", results, struct {
    chi2 = stat.statistic,
    num_bins = stat.num_bins,
    n_var_pars = stat.num_variable_params,
    dof = stat.num_bins-stat.num_variable_params,
    chi2red = stat.statistic/(stat.num_bins-stat.num_variable_params)
  });
      variable F = fopen(basefilename+"conf.txt", "w");
      ()=fprintf(F, "%s\n", get_fit_fun());
      ()=fprintf(F, "(%10s <=) %10s < %38s < %10s (<= %10s)\n", "min.all.", "conf.min.", "parameter value", "conf.max.", "max.all.");
      for($1=0; $1<length(results.name); $1++)
        ()=fprintf(F, "(%10g <=) %10g < %25s = %10g < %10g (<= %10g)   [%4.1f%% (conf)%5.1f%%]   %s\n",
		      results.min[$1], results.conf_min[$1], results.name[$1], results.value[$1], results.conf_max[$1], results.max[$1],
		      100*results.buf_below[$1], 100*results.buf_above[$1], results.tex[$1]);
      ()=fprintf(F, "\n%s/dof = %f/%d = %f\n", Fit_Statistic, stat.statistic, stat.num_bins-stat.num_variable_params, stat.statistic/(stat.num_bins-stat.num_variable_params));
      ()=fclose(F);
    }
    catch AnyError: vmessage("Could not write to %sconf.fits and/or %sconf.txt", basefilename, basefilename);

  vmessage("\n");
  vmessage("(%10s <=) %10s < %38s < %10s (<= %10s)", "min.all.", "conf.min.", "parameter value", "conf.max.", "max.all.");
  _for i (0, n-1, 1)
    vmessage("(%10g <=) %10g < %25s = %10g < %10g (<= %10g)   [%4.1f%% (conf)%5.1f%%]   %s",
	      results.min[i], results.conf_min[i], results.name[i], results.value[i], results.conf_max[i], results.max[i],
	      100*results.buf_below[i], 100*results.buf_above[i], results.tex[i]);

  if (qualifier_exists("cleanup"))
     {
	()=system(sprintf("rm %s*.[0-9][0-9]*", basefilename));
     }
   
   
  return results;
}

%
% Code from src/slang/geometry/clip_points_polygon.sl
%
define clip_points_polygon() 
%!%+
%\function{clip_points_polygon}
%\synopsis{Clip points against a polygon}
%\usage{(xc,yc)=clip_points_polygon(x,y,xp,yp)}
%\altusage{clipped=cohen_sutherland(points,poly)}
%\qualifiers{
% \qualifier{evenodd}{use the even-odd method to determine }
% \qualifier{crossing}{use the crossing number method}
% \qualifier{winding}{use the winding number method (the default)}
%}
%
%\description
% This function clips points defined by (x,y), where x and y can be arrays,
% against a closed polygon defined by the arrays (xp, yp) and returns the
% clipped points (xc,yc) where xc, yc are arrays (which can be empty
% if all points are outside of the polygon). Here, a closed polygon
% means that xp[0]==xp[-1] and yp[0]==yp[-1].
%
% Alternatively, the points and polygon can be defined as structs,
% where points=struct{ x=[], y=[] } and where the polygon is defined
% as poly=struct{x=xp,y=yp}
%
% The qualifiers define what to consider the "inside" of the polygon.
%
%\seealso{clip_points_rectangle, greiner_hormann, point_in_polygon}
%
%!%-
{
    variable src,clp;
    variable xx,yy,vx,vy;

    variable stru=0;
    if (_NARGS == 2 ) {
        (src,clp)=();
        stru=1;
        xx=src.x;
        yy=src.y;
        vx=clp.x;
        vy=clp.y;
    } else {
        if (_NARGS == 4 ) {
            (xx,yy,vx,vy)=();
        } else {
            throw UsageError,"clip_points_polygon: Use either 2 or 4 arguments\n";
        }
    }

    variable rx={};
    variable ry={};
    variable i;
    _for i(0,length(xx)-1,1) {
        if (point_in_polygon(xx[i],yy[i],vx,vy;;__qualifiers())) {
            list_append(rx,xx[i]);
            list_append(ry,yy[i]);
        }
    }

    if (length(rx)==0) {
        if (stru) {
            return struct {x=Double_Type[0],y=Double_Type[0]};
        }
        return (Double_Type[0],Double_Type[0]);
    }
    
    list_to_array(xx);
    list_to_array(yy);

    if (stru) {
        return struct{x=xx,y=yy};
    }

    return (xx,yy);
}

%
% Code from src/slang/geometry/clip_polyline_polygon.sl
%
define clip_polyline_polygon(polyline,polygon)
%!%+
%\function{clip_polyline_polygon}
%\usage{clipped=clip_polyline_polygon(line,polygon)}
%\qualifiers{
%\qualifier{evenodd}{use the crossing number method}
%\qualifier{crossing}{use the crossing number method}
%\qualifier{winding}{use the winding number method (the default)}
%}
%\description
% This function clips a polyline defined by line=struct{x=[],y=[]},
% where x and y are the points connected by the polyline, against
% a closed polygon=struct{x=[],y=[]}, where again x and y are the points
% connected by the polygon and where x[0]==x[-1] and y[0]==y[-1].
%
% The function returns a list of polylines that contain the segments
% of the line that are inside of the polygon. This list can be empty if
% there is no overlap.
%
% For complex polygons with intersecting segments, the qualifier defines
% what constitues the inside of the polygon.
%
% If you want to clip against a simple rectangle, use clip_polyline_rectangle
% for a faster algorithm.
%
%\seealso{point_in_polygon,clip_polyline_rectangle,clip_points_polygon}
%!%-
{
    variable evenodd=(qualifier_exists("crossing") || qualifier_exists("evenodd"));


    %
    % intersect each segment of poly with all segments of
    % the polygon
    % FIXME: this is ok for polygons and polylines with a small number
    % of segments. For polygons with a large number of segments we still
    % need to implement a version of the function that is based on
    % a binary search tree.
    %
    variable i,j;

    variable outpolyx={};
    variable outpolyy={};
    variable state={};

    variable crossing=2;
    variable oldpt=1;

    variable npolygon=length(polygon.x);
    variable nline=length(polyline.x);
    %
    % search for intersections of the segments of the polyline with
    % all segments of the polygon (so this can be expensive)
    %
    % we generate a new temporary polyline, outpoly, that contains
    % the points of the polyline and all crossing points,
    % remembering whether we're dealing with a new crossing point or
    % not
    %
    % FIXME: this can be memory intensive, since we're temporarily
    % duplicating polyline. A better approach might be to remember
    % the alphalists for each polyline segment (see below) and then
    % use these in the polyline splitting step that follows. This is a
    % bit more complicated to implement and up to now I have not
    % encountered real memory problems that have forced me to takle
    % this
    %
    _for i(0,nline-2) {
        list_append(outpolyx,polyline.x[i]);
        list_append(outpolyy,polyline.y[i]);
        list_append(state,oldpt);

        variable alphalist={};
        _for j(0,npolygon-2) {
            variable alphap,alphaq;
            variable res=line_intersect(polyline.x[i],polyline.y[i],polyline.x[i+1],polyline.y[i+1],
            polygon.x[j],polygon.y[j],polygon.x[j+1],polygon.y[j+1],&alphap,&alphaq);
            if (res==1) {
                list_append(alphalist,alphap);
            }
        }

        % if there are intersections: add them in ordered form (i.e., in the
        % direction of the polyline) to the outpoly
        if (length(alphalist)>0) {
            alphalist=list_to_array(alphalist);
            variable ndx=array_sort(alphalist);
            foreach alphap ( alphalist[ndx] ) {
                variable xx=polyline.x[i]+alphap*(polyline.x[i+1]-polyline.x[i]);
                variable yy=polyline.y[i]+alphap*(polyline.y[i+1]-polyline.y[i]);
                list_append(outpolyx,xx);
                list_append(outpolyy,yy);
                list_append(state,crossing);
            }
        }
    }

    % don't forget the last point
    list_append(outpolyx,polyline.x[-1]);
    list_append(outpolyy,polyline.y[-1]);
    list_append(state,oldpt);

    outpolyx=list_to_array(outpolyx);
    outpolyy=list_to_array(outpolyy);
    
    %
    % second step: walk along the polyline and cut it apart.
    %
    % This is easy in the case of the even odd rule, but a bit
    % more complex in the case of the winding number/crossing method
    %

    % determine whether the first point is inside or outside
    % (__qualifiers given to ensure that evenodd is taken into account)
    % this gives the "state" of the current segment
    variable inside=point_in_polygon(outpolyx[0],outpolyy[0],polygon.x,polygon.y;;__qualifiers());

    % this is where the individual segments will end up
    variable polypieces={};

    % coordinates of the current segment
    variable segx={};
    variable segy={};

    _for i(0,length(outpolyx)-2) {
        if (inside) {
            list_append(segx,outpolyx[i]);
            list_append(segy,outpolyy[i]);
        }
    
        variable newseg=0; % 1 if we start a new segment
        if (state[i]==crossing) {
            if (evenodd) {
                % even odd rule: the state of the segment changes with each crossing
                newseg=1;
            } else {
                % winding number rule: it is possible to either remain inside or go outside.
                % Test this by looking at the location of the middle point between point i and i+1,
                % to avoid round (the point in the middle is guaranteed not to be on a crossing,
                % after all)
                %
                variable xm=(outpolyx[i]+outpolyx[i+1])/2.;
                variable ym=(outpolyy[i]+outpolyy[i+1])/2.;
                % no need for qualifiers: the default of point_in_polygin is the
                % winding number/crossing number
                variable segpt=point_in_polygon(xm,ym,polygon.x,polygon.y);
                % a new segment starts if the state changes
                % (remember that inside=0 is outside)
                newseg=(segpt!=inside);
            }

            %
            % if we're starting a new segment, save the polygon or start a new one
            %
            if (newseg) {
                if (inside) {
                    % we're leaving the polygon, so save the current
                    % segment and start a new one
                    list_append(polypieces,struct{x=list_to_array(segx),y=list_to_array(segy)});
                    inside=0;
                    segx={};
                    segy={};
                } else {
                    % we're entering the polygon
                    inside=1;
                    list_append(segx,outpolyx[i]);
                    list_append(segy,outpolyy[i]);
                }
            }
        }
    }
        
    % the last point is never a crossing point, so the state can't change
    if (inside) {
        list_append(segx,outpolyx[-1]);
        list_append(segy,outpolyy[-1]);
        list_append(polypieces,struct{x=list_to_array(segx),y=list_to_array(segy)});
    }

    % and we're done
    return polypieces;
}


%
% Code from src/plot/close_aa1col_xlabel.sl
%
define close_aa1col_xlabel(x_min,x_max,y_min,y_max,x_label)
{
   color(1);
   xylabel((x_max-x_min)/2+x_min, y_min-(y_max-y_min)/7, x_label, 0, 0.5);
   charsize(1);
   _pgscf(1);
   close_plot;
}

%
% Code from src/plot/close_aa2col_xlabel.sl
%
define close_aa2col_xlabel(x_min,x_max,y_min,y_max,x_label)
{
  color(1);
  xlabel(label);
  %xylabel((x_max-x_min)/2+x_min,y_min-(y_max-y_min)/7,x_label,0,0.5);
  charsize(1);
  _pgscf(1);
  close_plot;
}

%
% Code from src/slang/arrays/close_match.sl
%
define close_match (a, b)
%!%+
%!%\function{close_match}
%!%\synopsis{Return index array for the larger array where
%!%smaller array has closest matching value.}
%!%\usage{Int_Type[] n = close_match(Array_Type a, Array_Type b);
%!%}
%!%\description
%!%    Takes two numerical arrays and returns the index array such that
%!%    the larger array is approximated with the smaller one by
%!%    \code{a[n]} or \code{b[n]}.
%!%    
%!%\note
%!%    If the arrays have only equal values, then a direct comparison
%!%    is much faster. close_match is allwas O(N^2).
%!%-
{
  variable k,j;
  variable idxo;
  variable lo,li,o,i;
  if (length(a)<length(b)) {
    lo = length(b);
    li = length(a);
    o = &b;
    i = &a;
  } else {
    lo = length(a);
    li = length(b);
    o = &a;
    i = &b;
  }
  idxo = @Int_Type[lo];
  %idxi = @Int_Type[li];

  variable diff, mindiff=_Inf;
  variable minidxi;
  _for k (0, lo-1, 1) {
    _for j (0, li-1, 1) {
      diff = abs((@o)[k]-(@i)[j]);
      if (diff < mindiff) {
	minidxi = j;
	mindiff = diff;
      }
    }
    idxo[k] = minidxi;
    mindiff = _Inf;
  }
  return idxo;
}
%
% Code from src/slang/astronomy/coco.sl
%

% only activate if module vector could be loaded
#ifeval __get_reference("vector")!=NULL
%
% Astronomical coordinate transforms
% (no relation to the old starlink code!)
%
% J. Wilms, 2017
%


private define apply_transform_vector() {
    %
    % apply a coordinate transform to a vector or an array
    % of vectors
    %

    variable mat,v;
    (mat,v)=();
    
    if (Array_Type==typeof(v)) {
	variable res=Array_Type[length(v)];
	variable i;
	_for i(0,length(v)-1,1) {
	    res[i]=mat*v[i];
	}
	return res;
    }
    return mat*v;
}

private define apply_transform_coord() {
    %
    % apply the transformation matrix trans to the coordinates alpha and beta
    %

    variable trans,alpha,delta;
    (trans,alpha,delta)=();
    
    variable i;
    variable r,lambda,beta;
    if (Array_Type==typeof(alpha)) {
	variable ex=Vector_Type[length(alpha)];
	if (qualifier_exists("deg")) {
	    _for i(0,length(alpha)-1,1) {
		ex[i]=trans*vector_astro(alpha[i],delta[i];deg);
	    }
	} else {
	    _for i(0,length(alpha)-1,1) {
		ex[i]=trans*vector_astro(alpha[i],delta[i]);
	    }
	}	    
	(r,lambda,beta)=array_map(Double_Type,Double_Type,Double_Type,&vector_to_spherical,ex);
	if (qualifier_exists("deg")) {
	    lambda *=180./PI;
	    beta *= 180./PI;
	}
    } else {
	if (qualifier_exists("deg")) {
	    (r,lambda,beta)=vector_to_spherical(trans*vector_astro(alpha,delta;deg);deg);
	} else {
	    (r,lambda,beta)=vector_to_spherical(trans*vector_astro(alpha,delta));
	}
    }
    return (lambda,beta);
}

define gcrs2j2000_matrix()
%!%+
%\function{gcrs2j2000_matrix}
%\synopsis{return the frame bias matrix to convert coordinates from the GCRS to J2000.0}
%\usage{Matrix33_Type mat=gcrs2j2000_matrix();}
%\description
%  This function returns the frame bias matrix needed to convert
%  a vector in the Geocentric Celestial Reference System (GCRS)
%  into the J2000.0 dynamical reference system (see
%  astronomical almanac, section B)
%
%\seealso{precess,precession_matrix}
%!%-
{
    return matrix33_new(+0.9999999999999942, -0.000000071,        +0.00000008056,
                        +0.000000071       , +0.9999999999999969, +0.00000003306,
                        -0.00000008056     , -0.00000003306,      +0.9999999999999962);
}

define precession_matrix(JD) 
%!%+
%\function{precession_matrix}
%\synopsis{return the precession matrix for the transform from J2000.0 to a given JD}
%\usage{Matrix33_Type mat=precession_matrix(JD)}
%
%\description
% This function calculates the precession matrix for the calculation
% of the precession FROM J2000.0 TO JD using the IAU 2000.0A theory
% (Capitaine et al., 2003, A&A 412, 567, Sect. 7)
%
% To get the matrix for the conversion from JD to J2000.0, transpose the
% matrix returned by this function.
%
% In most cases users will want to use the precess-function to
% precess astronomical coordinates.
%
% Note: This is mainly an internal function, it therefore does NOT
% have the usual qualifiers such as mjd or any safety checks, however,
% for some applications it will be useful to have direct access to the
% precession matrix.
%
%\seealso{precess,precession_matrix}
%!%-
%
{
    % number of Julian centuries since J2000.0
    variable T=(JD-2451545.0)/36525.;

    % obliquity at J2000.0 (arcsec)
    variable eps0=84381.406;

    % precession angles (all in arcsec)
    variable psi=((((-0.0000000951*T+0.000132851)*T-0.00114045)*T-1.0790069)*T+5038.481507)*T;
    variable ome=((((+0.0000003337*T-0.000000467)*T-0.00772503)*T+0.0512623)*T-0.025754)*T+eps0;
    variable chi=((((-0.0000000560*T+0.000170663)*T-0.00121197)*T-2.3814292)*T+10.556403)*T;

    % convert to rad
    variable as2rad=PI/648000.;
    psi*=as2rad;
    ome*=as2rad;
    chi*=as2rad;
    eps0*=as2rad;

    % build up the precession matrix
    return matrix33_rot(3,chi)*matrix33_rot(1,-ome)*matrix33_rot(3,-psi)*matrix33_rot(1,eps0);
}

define precession_matrix2()
%!%+
%\function{precession_matrix2}
%\synopsis{return the precession matrix for the transform between two equinoxes}
%\usage{Matrix33_Type mat=precession_matrix2(fromJD,toJD)}
%
%\description
% This function calculates the precession matrix for the calculation
% of the precession from fromJD TO toJD using the IAU 2000.0A theory
% (Capitaine et al., 2003, A&A 412, 567, Sect. 7)
%
% In most cases users will want to use the precess-function to
% precess astronomical coordinates.
%
% Note: This is mainly an internal function, it therefore does NOT
% have the usual qualifiers such as mjd or any safety checks, however,
% for some applications it will be useful to have direct access to the
% precession matrix.
%
%\seealso{precess,precession_matrix}
%!%-
%
{
    variable fromequinox,toequinox;
    (fromequinox,toequinox)=();

    %
    % build the precession matrix: precess from fromequinox to J2000 and
    % then from J2000 to toequinox.
    % J2000 = JD2451545
    %
    variable mtoj2000,mfromj2000;

    if (fromequinox==toequinox) {
	return matrix33_identity();
    }
    
    if (fromequinox != 2451545.0) {
	% from different from J2000.0?
	mtoj2000=(precession_matrix(fromequinox)).transpose(); % transform epoch->J2000
	if (toequinox == 2451545.0) {
	    return mtoj2000;
	}
    } 

    if (toequinox != 2451545.0) {
	% to different from J2000.0?
	mfromj2000=precession_matrix(toequinox); % transform J2000->epoch
	if (fromequinox == 2451545.0) {
	    return mfromj2000;
	}
    }
    % general case: product of both transforms
    return mfromj2000*mtoj2000;
}

define precess()
%!%+
%\function{precess}
%\synopsis{precess equatorial coordinates between two equinoxes}
%\usage{(alpha,delta)=precess(alpha,delta;qualifiers)}
%\altusage{AstroVector3_Type ecl=precess(eqp;qualifiers)}
%\qualifiers{
%\qualifier{fromequinox}{starting equinox of the transformation.
%                      Float: JD, string: epoch designation (e.g.,
%                      "J2000.0" or "B1950.0";
%                      default: J2000.0)}
%\qualifier{toequinox}{end equinox of the transformation (same interpretation)
%                    as fromequinox; required!}
%\qualifier{deg}{interpret angular arguments in degrees (default is radians!)
%                applies also to the return value.}
%\qualifier{mjd}{interpret equinoxes as a MJD (default: JD)}
%}
%\description
%
% Precess a coordinate from one equinox to another using
% the IAU 2000A model.
%
% The default equinox is J2000.0, formally the equinoxes are in TDB.
%
%\seealso{Vector_Type,JDofEpoch,dms2deg,hms2deg,precession_matrix,gcrs2j2000_matrix}
%!%-
{
    variable alpha,delta;
    switch(_NARGS)
    { case 1: alpha=();}
    { case 2: (alpha,delta)=();}
    { help(_function_name()); return; }

    %
    % set the epochs
    %
    variable fromequinox,toequinox;
    fromequinox=JDofEpoch(qualifier("fromequinox","J2000.0");;__qualifiers());
    if (qualifier_exists("toequinox")) {
	toequinox=JDofEpoch(qualifier("toequinox");;__qualifiers()); % take into account mjd qualifier
    } else {
	throw UsageError, sprintf("%s: toequinox qualifier is mandatory\n",_function_name());
    }

    % final transformation matrix
    variable transform=precession_matrix2(fromequinox,toequinox);

    %
    % Apply the coordinate transform
    %
    if (_NARGS==1) {
	%  argument is a vector: transform the vector
	return apply_transform_vector(transform,alpha);
    } 

    % arguments are coordinates
    if (qualifier_exists("deg")) {
	return apply_transform_coord(transform,alpha,delta;deg);
    }
    return apply_transform_coord(transform,alpha,delta);    
}

define nutation_matrix()
%!%+
%\function{nutation_matrix}
%\synopsis{Calculate the nutation matrix for the mean equinox JD}
%\usage{Matrix33_Type N=nutation_matrix(JD)}
%\qualifiers{
%   \qualifier{theory}{nutation theory to use, see help for
%                      function nutation_angles for details.
%                      Default: NU2000K}
%}
%\description
% This function returns the nutation matrix to convert coordinates
% for the mean equinox of JD to the true equinox of JD.
% 
% Note: This is mainly an internal function, it therefore does NOT
% have the usual qualifiers such as mjd, however, for some applications
% it is useful to have direct access to the nutation matrix
%!%-
{
    variable JD=();
    variable theory=qualifier("theory","NU2000K");

    variable dpsi,deps;
    (dpsi,deps)=nutation_angles(JD;theory=theory);

    % obliquity
    variable eps=obliquity(JD); % rad!
    % true obliquity
    variable epstrue=eps+deps;

    return matrix33_rot(1,-epstrue)*matrix33_rot(3,-dpsi)*matrix33_rot(1,eps);
}

define equatorial2ecliptical()
%!%+
%\function{equatorial2ecliptical}
%\synopsis{convert equatorial (alpha,delta) to ecliptical (lambda,beta) coordinates}
%\usage{(lambda,beta)=equatorial2ecliptical(alpha,delta;qualifiers)}
%\altusage{Vector_Type ecl=equatorial2ecliptical(eqp;qualifiers)}
%\qualifiers{
%\qualifier{equinox}{equinox of the transformation. Float: JD, string: equinox
%                  designation (e.g., "J2000.0" or "B1950.0";
%                  default: J2000.0)}
%\qualifier{deg}{interpret angular arguments in degrees (default is radians!)
%                applies also to the return value.}
%\qualifier{mjd}{interpret equinox as a MJD (default: JD)}
%\qualifier{inverse}{perform the conversion from ecliptical to equatorial
%                    coordinates (see ecliptical2equatorial() for an
%                    equivalent interface)}
%\qualifier{true}{perform calculation for apparent coordinates, i.e., include nutation terms}
%}
%\description
%  The routine takes coordinates in the equatorial (right ascension, 
%  declination) system and converts them into ecliptical coordinates 
%  using the obliquity of the ecliptic for the equinox and vice versa..
%
%  Alternatively, the routine also accepts an Vector_Type and then
%  returns a vector in ecliptical coordinates (or vice versa if the inverse
%  qualifier is given).
%
%  The default equinox is J2000.0.
%
%\seealso{Vector_Type, ecliptical2equatorial, JDofEpoch,dms2deg,hms2deg}
%!%-
{
    variable alpha,delta;
    switch(_NARGS)
    { case 1: alpha=();}
    { case 2: (alpha,delta)=();}
    { help(_function_name()); return; }

    %
    % get the obliquity
    % because __qualifiers() would also transfer the deg keyword,
    % do NOT just transfer everything 
    %
    variable epsilon;
    if (qualifier_exists("true")) {
	if (qualifier_exists("equinox")) {
	    if (qualifier_exists("mjd")) {
		epsilon=obliquity(JDofEpoch(qualifier("equinox"));mjd,true);
	    } else {
		epsilon=obliquity(JDofEpoch(qualifier("equinox"));true);
	    }
	} else {
	    epsilon=obliquity(JDofEpoch("J2000.0");true);
	}
    } else {
	if (qualifier_exists("equinox")) {
	    if (qualifier_exists("mjd")) {
		epsilon=obliquity(JDofEpoch(qualifier("equinox"));mjd);
	    } else {
		epsilon=obliquity(JDofEpoch(qualifier("equinox")));
	    }
	} else {
	    epsilon=obliquity(JDofEpoch("J2000.0"));
	}
    }
    %
    % rotation matrix
    %
    variable transform;
    if (qualifier_exists("inverse")) {
	transform=matrix33_rot(1,-epsilon);
    } else {
	transform=matrix33_rot(1,epsilon);
    }

    %
    % Apply the coordinate transform
    %
    if (_NARGS==1) {
	%  argument is a vector: transform the vector
	return apply_transform_vector(transform,alpha);
    } 

    % arguments are coordinates
    if (qualifier_exists("deg")) {
	return apply_transform_coord(transform,alpha,delta;deg);
    }
    return apply_transform_coord(transform,alpha,delta);    
}

define ecliptical2equatorial()
%!%+
%\function{ecliptical2equatorial}
%\synopsis{convert ecliptical (lambda,beta) to equatorial coordinates}
%\usage{(alpha,delta)=ecliptical2equatorial(lambda,beta;qualifiers)}
%\altusage{Vector_Type eqp=equatorial2ecliptical(ecl;qualifiers)}
%\qualifiers{
%\qualifier{equinox}{equinox of the transformation. Float: JD, string: equinox
%                  designation (e.g., "J2000.0" or "B1950.0";
%                  default: J2000.0)}
%\qualifier{deg}{interpret angular arguments in degrees (default is radians!)
%                applies also to the return value.}
%\qualifier{mjd}{interpret equinox as a MJD (default: JD)}
%\qualifier{true}{perform calculation for apparent coordinates, i.e., include nutation terms}
%}
%\description
%  The routine takes coordinates in the ecliptical (lambda, beta)
%  system and converts them into equatorial coordinates (right ascension,
%  declination), using the obliquity of the ecliptic for the equinox
%  and vice versa.
%
%  Alternatively, the routine also accepts a Vector_Type and then
%  returns an Vector_Type vector in equatorial coordinates (or
%  vice versa if the inverse qualifier is given).
%
%  The default equinox is J2000.0.
%
%  This routine just calls equatorial2ecliptical with all arguments and the
%  inverse qualifier, but is a simpler interface.
%
%\seealso{Vector_Type, equatorial2ecliptical, JDofEpoch,dms2deg,hms2deg}
%!%-
{
    if (_NARGS<1 or _NARGS>2) {
	help(_function_name()); return;
    }

    % sanity check, otherwise havoc ensues!
    if (qualifier_exists("inverse")) {
	throw UsageError, sprintf("%s: inverse qualifier is NOT allowed!\n",_function_name());
    }
    variable qual=__qualifiers();
    qual=struct { @qual, "inverse" };

    % I am sure there is a more elegant way to do this...
    variable alpha,delta;
    if (_NARGS==1) {
	alpha=();
	return equatorial2ecliptical(alpha ;;qual);
    }
    (alpha,delta)=();
    return equatorial2ecliptical(alpha,delta ;;qual);
}

define equatorial2horizon()
%!%+
%\function{equatorial2horizon}
%\synopsis{convert equatorial (alpha,delta) to horizon (elevation, azimuth) coordinates}
%\usage{(azi,ele)=equatorial2horizon(alpha,delta;qualifiers)}
%\altusage{Vector_Type hor=equatorial2horizon(eqp;qualifiers)}
%\qualifiers{
% \qualifier{JD}{JD for which the calculation is to be performed (in UT1/UTC). Mandatory.}
% \qualifier{lon}{geographic longitude of the observer, positive towards the east. Mandatory.}
% \qualifier{lat}{geographic latitude of the observer, positive towards the north. Mandatory.}
% \qualifier{deg}{interpret all angular arguments in degrees (default is radians!)
%                 applies also to the return value.}
% \qualifier{mjd}{interpret JD argument as a MJD (default: JD)}
% \qualifier{inverse}{perform the conversion from horizontal to equatorial
%                     coordinates (see horizontal2equatorial() for an
%                     equivalent interface)}
%}
%\description
%  The routine takes coordinates in the equatorial (right ascension, 
%  declination) system for the equinox and ecliptic of the date
%  (i.e., "apparent coordinates"), and converts them into the elevation
%  and azimuth for a given observer's position and time. Note that
%  the JD, lon, and lat qualifiers are mandatory!
%
%  The JD argument is to be measured in UT1, for most cases it is sufficient to
%  approximate this as UT. Note that ephemerides work in TDB or TT, which has
%  an offset of several tens of seconds. For most applications this should not
%  matter, especially because of the uncertainty of refraction, but please take
%  this into account if highest precision is needed.
%
%  Formally the coordinates have to be topocentric coordinates in
%  order to include parallax effects. For most applications this is not 
%  important as long as a precision of a few arcsec or worse is needed.
%  Eventually a routine applying all of these effects will be provided.
%
%  Alternatively, the routine also accepts an Vector_Type and then
%  returns a vector in horizon coordinates (or vice versa if the inverse
%  qualifier is given).
%
%\seealso{Vector_Type, horizontal2equatorial, JDofEpoch,dms2deg,hms2deg,tai2tt,utc2tai,tt2tdb}
%!%-
{
    variable alpha,delta,JD;
    switch(_NARGS)
    { case 1: (alpha)=();}
    { case 2: (alpha,delta)=();}
    { help(_function_name()); return; }

    variable lon,lat;

    if (qualifier_exists("lon")) {
	lon=qualifier("lon");
    } else {
	throw UsageError, sprintf("%s: lon qualifier is mandatory\n",_function_name());
    }

    if (qualifier_exists("lat")) {
	lat=qualifier("lat");
    } else {
	throw UsageError, sprintf("%s: lat qualifier is mandatory\n",_function_name());
    }

    if (qualifier_exists("deg")) {
	lon*=PI/180.;
	lat*=PI/180.;
    }
    
    if (qualifier_exists("JD")) {
	JD=qualifier("JD");
    } else {
	throw UsageError, sprintf("%s: JD qualifier is mandatory\n",_function_name());
    }
    if (qualifier_exists("mjd")) {
	JD=mjd2jd(JD);
    }
    
    %
    % local siderial time
    %
    variable lst=GMST(JD;lst,lon=lon,lat=lat);
    
    % s->rad
    lst=2.*PI*lst/86400.;

    %
    % rotation matrix
    %
    variable transform;
    if (qualifier_exists("inverse")) {
	% transform from H to HA system
	variable transform_h2ha=matrix33_rot(2,lat-PI/2.)*matrix33_rot(3,PI);
	% transform from HA to RA system
	variable transform_ha2ra=matrix33_rot(3,-lst)*matrix33_reflect(2);
	transform=transform_ha2ra*transform_h2ha;
    } else {
	% transform from RA to HA system
	variable transform_ra2ha=matrix33_reflect(2)*matrix33_rot(3,lst);
	% transform from HA to H system
	variable transform_ha2h=matrix33_rot(3,PI)*matrix33_rot(2,PI/2.-lat);
	transform=transform_ha2h*transform_ra2ha;

	variable t1=matrix33_new(cos(lst),sin(lst),0.,
 	                         sin(lst),-cos(lst), 0.,
	                         0.,0.,1.);
	variable t2=matrix33_new(sin(lat),0,-cos(lat),
	                         0,1,0,
	                         cos(lat),0,sin(lat));
	transform=t2*t1;
%	variable tt=t2*t1;
%	print(tt.as_array());
%	print("XXXX");
%	print(transform.as_array());
%	exit;

	
    }

    %
    % Apply the coordinate transform
    %
    if (_NARGS==1) {
	%  argument is a vector: transform the vector
	return apply_transform_vector(transform,alpha);
    }

    % arguments are coordinates
    if (qualifier_exists("deg")) {
	return apply_transform_coord(transform,alpha,delta;deg);
    }
    return apply_transform_coord(transform,alpha,delta);    
}

define horizon2equatorial()
%!%+
%\function{horizon2equatorial}
%\synopsis{convert horizon (azi,ele) coordinates to equatorial (alpha,delta)}
%\usage{(alpha,delta)=horizon2equatorial(azi,ele;qualifiers)}
%\altusage{Vector_Type eqp=horizon2equatorial(hor;qualifiers)}
%\qualifiers{
% \qualifier{JD}{JD for which the calculation is to be performed. Mandatory.}
% \qualifier{lon}{geographic longitude of the observer, positive towards the east. Mandatory.}
% \qualifier{lat}{geographic latitude of the observer, positive towards the north. Mandatory.}
% \qualifier{deg}{interpret all angular arguments in degrees (default is radians!)
%                 applies also to the return value.}
% \qualifier{mjd}{interpret date as a MJD (default: JD)}
%}
%
%\description
% The function  takes coordinates in the horizon (azimuth,elevation) system
% and converts them into the equatorial system for the ecliptic and
% elevation of the date. Alternatively the routine also accepts a
% Vector_Type and then returns avector in equatorial coordinates.
%
% See the description of the function equatorial2horizon for
% references and further information. This routine is just a
% wrapper around that function, with the qualifier "inverse" set.
%
%\seealso{Vector_Type, equatorial2horizon,dms2deg,hms2deg}
%!%-
{
    if (_NARGS<1 or _NARGS>2) {
	help(_function_name()); return;
    }

    % sanity check, otherwise havoc ensues!
    if (qualifier_exists("inverse")) {
	throw UsageError, sprintf("%s: inverse qualifier is NOT allowed!\n",_function_name());
    }
    variable qual=__qualifiers();
    qual=struct { @qual, "inverse" };

    % I am sure there is a more elegant way to do this...
    variable azi,ele;
    if (_NARGS==1) {
	azi=();
	return equatorial2horizon(azi ;;qual);
    }
    (azi,ele)=();
    return equatorial2horizon(azi,ele ;;qual);
}

define equatorial2galactic()
%!%+
%\function{equatorial2galactic}
%\synopsis{convert J2000.0 equatorial (alpha,delta) to galactic (l,b) coordinates}
%\usage{(l,b)=equatorial2galactic(alpha,delta;qualifiers)}
%\altusage{Vector_Type gal=equatorial2galactic(eqp;qualifiers)}
%\qualifiers{
%\qualifier{deg}{interpret angular arguments in degrees (default is radians!)
%                applies also to the return value.}
%\qualifier{inverse}{perform the conversion from galactic to equatorial
%                    coordinates (see galactic2equatorial() for an
%                    equivalent interface)}
%\qualifier{blaauw}{use the original Blaauw definition of the IAU galactic
%              coordinate system; see below for caveats}
%}
%\description
%  The function takes coordinates in the equatorial (right ascension, 
%  declination) system and converts them into galactic
%  coordinates (or vice versa, if the inverse qualifier is given) for
%  the IAU 1958 Galactic coordinate system (Blaauw et al., 1960,
%  MNRAS 121, 123).
%
%  Alternatively, the routine also accepts a Vector_Type and then
%  returns a vector in galactic coordinates (or vice versa if the 
%  inverse qualifier is given).
%
%  The function is for J2000.0 coordinates only, use the
%  precess function to convert to that equinox if needed.
%
%  See Lane (1979, PASP 91, 405) for a discussion of the subtleties of
%  this conversion and Johnson & Soderblom (1987, AJ 93, 864) for a
%  non-standard derivation of the conversion matrix (the one used here
%  is basically identical, but uses proper Euler angles). By default, 
%  the conversion is done using the precessed pole and inclination for 
%  the FK5 system (J2000.0) given by Liu et al (2011, A&A 526, A16), 
%  which is a thorough derivation of values that are essentially 
%  identical to those also given in the appendix of Reid & Brunthaler
%  (2004, ApJ 616, 872).
%
%  If the qualifier Blaauw is given, then the coordinates are first 
%  precessed to B1950.0 and then transformed. Note that this approach
%  is not good for the highest precision, since no conversion from,
%  e.g., FK5 to FK4 coordinates is performed as this would result in a
%  non-orthogonal coordinate system. Please use this qualifier only
%  if you know what you are doing and read Liu et al. (2011) before
%  using the qualifier.
%
%\seealso{Vector_Type, galactic2equatorial,precess,dms2deg,hms2deg}
%!%-
{
    variable alpha,delta;
    switch(_NARGS)
    { case 1: alpha=();}
    { case 2: (alpha,delta)=();}
    { help(_function_name()); return; }

    variable alpha_ngp,delta_ngp,theta0;
    if (qualifier_exists("blaauw")) {
	% build conversion matrix to galactic coordinates
	% (positions in B1950 by Blaauw)
	alpha_ngp=192.25*PI/180.;
	delta_ngp=27.4*PI/180.;
	theta0=123.*PI/180.;
    } else {
	% values for J2000.0 per Liu et al. (2011, eq.10)
	alpha_ngp=hms2deg(12,51,26.27549;radian);
	delta_ngp=dms2deg(27,07,41.7043;radian);
	theta0=122.93191857*PI/180.;
    }

    variable transform;

    if (qualifier_exists("inverse")) {
	transform= matrix33_rot(3,3.*PI/2.-alpha_ngp)*matrix33_rot(1,-(PI/2.-delta_ngp))*matrix33_rot(3,theta0-PI/2.);
    } else {
	%
	% transformation equatorial->galactic using Euler angles
	% (reproduces Johnson & Soderblom, section III, or
	% Liu 2011, but uses Euler angles and not the
	% non-rotation matrices that are used by Johnson & Soderblom )
	transform=matrix33_rot(3,PI/2.-theta0)*matrix33_rot(1,PI/2.-delta_ngp)*matrix33_rot(3,-(3.*PI/2.-alpha_ngp));

	% Johnson & Soderblom, section III (for B1950.0)
	%variable trans=matrix33_new(-0.06699,-0.87276,-0.48354,
	%                            +0.49273,-0.45035,+0.74458,
	%                            -0.86760,-0.18837,+0.46020);
    }

    if (qualifier_exists("blaauw")) {
	%
	% go to B1950.0 and use original Blaauw coordinates
	%
	
	% precess input coordinates to B1950.0
	variable prectrans=precession_matrix(JDofEpoch("B1950.0"));
	if (qualifier_exists("inverse")) {
	    transform=prectrans*transform;
	} else {
	    transform=transform*prectrans;
	}
    }
    
    %
    % Apply the coordinate transform
    %
    if (_NARGS==1) {
	%  argument is a vector: transform the vector
	return apply_transform_vector(transform,alpha);
    } 

    % arguments are coordinates
    if (qualifier_exists("deg")) {
	return apply_transform_coord(transform,alpha,delta;deg);
    }
    return apply_transform_coord(transform,alpha,delta);    
}

define galactic2equatorial()
%!%+
%\function{galactic2equatorial}
%\synopsis{convert galactic (l,b) coordinates to J2000.0 equatorial (alpha,delta)}
%\usage{(alpha,delta)=galactic2equatorial(l,b;qualifiers)}
%\altusage{Vector_Type eqp=equatorial2galactic(gal;qualifiers)}
%\qualifiers{
%\qualifier{deg}{interpret angular arguments in degrees (default is radians!)
%                applies also to the return value.}
%}
%
%\description
% The function  takes coordinates in the galactic (l,b) system
% and converts them into the equatorial J2000.0 system. Alternatively
% the routine also accepts a Vector_Type and then returns a
% vector in equatorial coordinates for J2000.0.
%
% The function is for J2000.0 coordinates only, use the
% precess function to convert the result to another equinox
% if needed.
%
% See the description of the function equatorial2galactic for
% references and further information. This routine is just a
% wrapper around that function, with the qualifier "inverse" set.
%
%\seealso{Vector_Type, equatorial2galactic,precess,dms2deg,hms2deg}
%!%-
{
    if (_NARGS<1 or _NARGS>2) {
	help(_function_name()); return;
    }

    % sanity check, otherwise havoc ensues!
    if (qualifier_exists("inverse")) {
	throw UsageError, sprintf("%s: inverse qualifier is NOT allowed!\n",_function_name());
    }
    variable qual=__qualifiers();
    qual=struct { @qual, "inverse" };

    % I am sure there is a more elegant way to do this...
    variable ell,b;
    if (_NARGS==1) {
	ell=();
	return equatorial2galactic(ell ;;qual);
    }
    (ell,b)=();
    return equatorial2galactic(ell,b ;;qual);
}


#endif

%
% Code from src/slang/geometry/cohen_sutherland.sl
%

private variable CS_inside=0; % 0000
private variable CS_left=1;   % 0001
private variable CS_right=2;  % 0010
private variable CS_bottom=4; % 0100
private variable CS_top=8;    % 1000

private define cs_outcode(x,y,xmin,ymin,xmax,ymax) {
    %
    % return location of x,y against box xmin,ymin,xmax,ymax
    variable code=CS_inside;
    if (x<xmin) {
        code |= CS_left;
    } else {
        if (x>xmax) {
            code |=CS_right;
        }
    }
    if (y<ymin) {
        code |= CS_bottom;
    } else {
        if (y>ymax) {
            code |=CS_top;
        }
    }
    return code;
}


define cohen_sutherland()
%!%+
%\function{cohen_sutherland}
%\synopsis{Clip a line against a rectangle}
%\usage{(xc0,yc0,xc1,yc1)=cohen_sutherland(x0,y0,x1,y1,xmin,ymin,xmax,ymax)}
%\altusage{clipped=cohen_sutherland(line,box)}
%\description
% This function clips a line defined by the points (x0,y0) and (x1,y1) against
% a box defined by the corner points (xmin,ymin) and (xmax,ymax) and returns
% the clipped line (xc0,yc0) -- (xc1,yc1). The clipped coordinates are set
% to _NaN if the line misses the box.
%
% The line segment and clipping box can be either defined directly by giving the
% coordinates or as structs, In the latter case, the line segment
% is defined as line= struct{ x=[x0,x1], y=[y0,y1] } and the
% rectangular box is defined either as box=struct{x=[xmin,xmax],y=[ymin,ymax]}
% or as box=struct{xmin=xmin,ymin=ymin,xmax=xmax,ymax=ymax}.
%
%\seealso{clip_points_rectangle,clip_points_polygon, clip_polyline_rectangle,greiner_hormann}
%
%!%-
{
    variable x0,y0,x1,y1,xmin,ymin,xmax,ymax;

    variable struret=0;
    
    if ( _NARGS == 2 ) {
        variable p,box;
        (p,box)=();
        x0=p.x[0]; y0=p.y[0];
        x1=p.x[1]; y1=p.y[1];

        if (struct_field_exists(box,"x")) {
            xmin=box.x[0]; xmax=box.x[1];
            ymin=box.y[0]; ymax=box.y[1];
        } else {
            % shp convention for bounding boxes
            xmin=box.xmin; ymin=box.ymin;
            xmax=box.xmax; ymax=box.ymax;
        }

        struret=1;
        
    } else {
        if (_NARGS==8) {
            (x0,y0,x1,y1,xmin,ymin,xmax,ymax)=();
        } else {
            throw UsageError,"cohen_sutherland: Use either 2 or 8 arguments\n";
        }
    }
    
    variable outcode0=cs_outcode(x0,y0,xmin,ymin,xmax,ymax);
    variable outcode1=cs_outcode(x1,y1,xmin,ymin,xmax,ymax);

    variable accept=0;

    while(1) {
        if ( not (outcode0 | outcode1)) {
            % bitwise or is 0: both points are inside the window
            accept=1;
            break;
        } else {
            % bitwise and is not 0: both points are outside the window
            if (outcode0 & outcode1) {
                break;
            } else {
                % both tests failed: calculate the line segment
                % to clip from an outside point to an intersection
                % with the clip edge;

                variable x,y;
                variable outcodeOut = outcode1 > outcode0 ? outcode1 : outcode0;

                % find the intersection point
                if (outcodeOut & CS_top) { %point is above clip window
                    x=x0+(x1-x0)*(ymax-y0)/(y1-y0);
                    y=ymax;
                } else {
                    if (outcodeOut & CS_bottom) { % point is below clip window
                        x=x0+(x1-x0)*(ymin-y0)/(y1-y0);
                        y=ymin;
                    } else {
                        if (outcodeOut & CS_right) { % right of clip window
                            x=xmax;
                            y=y0+(y1-y0)*(xmax-x0)/(x1-x0);
                        } else {
                            if (outcodeOut & CS_left) { % left of clip window
                                x=xmin;
                                y=y0+(y1-y0)*(xmin-x0)/(x1-x0);
                            }
                        }
                    }
                }

                % move outside point to intersection point
                % and move to next pass
                if (outcodeOut==outcode0) {
                    x0=x;
                    y0=y;
                    outcode0=cs_outcode(x0,y0,xmin,ymin,xmax,ymax);
                } else {
                    x1=x;
                    y1=y;
                    outcode1=cs_outcode(x1,y1,xmin,ymin,xmax,ymax);
                }
            }
        }
    }

    if (accept) {
        if (struret) {
            return struct{x=[x0,x1],y=[y0,y1]};
        }
        return (x0,y0,x1,y1);
    }

    % segment is not in the clip window
    if (struret) {
        return struct{x=Double_Type[0],y=Double_Type[0]};
    }
    return (_NaN,_NaN,_NaN,_NaN);
    
}

define clip_polyline_rectangle(src,box) {
%!%+
%\function{clip_polyline_rectangle}
%\synopsis{Clip a polyline against a rectangle}
%\usage{clipped=clip_polyline_rectangle(poly,box)}
%\description
% This function clips the polyline poly=struct{x=[],y=[]}, where x and y are arrays containing
% the points of the polyline, against a rectangle defined by the corner points (xmin,ymin) and
% (xmax,ymax). It returns a list of structs{x,y}, where each list element contains the points of
% a segment of the polygon that is inside the box.
%
%\seealso{clip_points_polygon, clip_polyline_rectangle, cohen_sutherland, greiner_hormann}
%
%!%-

    variable i;

    variable polyli={};
    variable xx={};
    variable yy={};
    
    _for i(0,length(src.x)-2,1) {
        variable seg=cohen_sutherland(struct{x=[src.x[i],src.x[i+1]],
                                             y=[src.y[i],src.y[i+1]]},box);
        % is the segment within the box?
        if (length(seg.x)>0) {
            % append?
            if (length(xx)>0) {
                % start of new segment different from old end point?
                if ((xx[-1]!=seg.x[0]) || yy[-1]!=seg.y[0]) {
                    xx=list_to_array(xx);
                    yy=list_to_array(yy);
                    list_append(polyli,struct{x=xx,y=yy});
                    xx={};
                    yy={};
                }
            }
            list_append(xx,seg.x[0]); list_append(xx,seg.x[1]);
            list_append(yy,seg.y[0]); list_append(yy,seg.y[1]);
        }
    }

    if (length(xx)>0) {
        xx=list_to_array(xx);
        yy=list_to_array(yy);
        list_append(polyli,struct{x=xx,y=yy});
    }

    return polyli;
}

define clip_points_rectangle()
%!%+
%\function{clip_points_rectangle}
%\synopsis{Clip points against a rectangle}
%\usage{(xc,yc)=clip_points_rectangle(x,y,xmin,ymin,xmax,ymax)}
%\altusage{clipped=clip_points_rectangle(points,box)}
%\description
%This function clips the points defined by (x,y), where x and y can be arrays,
%against a rectangle defined by the corner points (xmin,ymin) and (xmax,ymax) and 
%returns the clipped points (xc,yc) where xc, yc are arrays (which can be empty
%if all points are outside of the rectangle).
%
%Alternatively, the points and clipping rectangle can be defined as structs,
%where points=struct{ x=[], y=[] } and where the clipping rectangle is defined
%either as box=struct{x=[xmin,xmax],y=[ymin,ymax]}
%or as box=struct{xmin=xmin,ymin=ymin,xmax=xmax,ymax=ymax}.
%
%
%\seealso{clip_points_polygon, clip_polyline_rectangle, cohen_sutherland, greiner_hormann}
%
%!%-
{
    variable xmin,xmax,ymin,ymax;
    variable struret=0;
    variable xi,yi;
    
    if ( _NARGS == 2 ) {
        variable src,box;
        (src,box)=();
        xi=src.x;
        yi=src.y;

        if (struct_field_exists(box,"x")) {
            xmin=box.x[0]; xmax=box.x[1];
            ymin=box.y[0]; ymax=box.y[1];
        } else {
            % shp convention for bounding boxes
            xmin=box.xmin; ymin=box.ymin;
            xmax=box.xmax; ymax=box.ymax;
        }

        struret=1;
    } else {
        if (_NARGS==6) {
            (xi,yi,xmin,ymin,xmax,ymax)=();
        } else {
            throw UsageError,"clip_points: Use either 2 or 6 arguments\n";
        }
    }

    variable ndx=where(xmin<=xi<xmax && ymin<=yi<ymax);

    if (struret) {
        return struct {x=xi[ndx],y=yi[ndx]};
    }
    return (xi[ndx],yi[ndx]);
}


%
% Code from src/data/timing/colacal.sl
%
%%%%%%%%%%%%%%
define colacal(freq, cpd, noicpd, lopsd, hipsd, noilopsd, noihipsd, siglopsd, sighipsd, alln)
%%%%%%%%%%%%%%
%!%+
%\function{colacal}
%\synopsis{Timing Tools: Coherence and Lag Calculation}
%\usage{ Struct_Type cola = colacal ([10 required inputs]);}
%\description
%    Input:
%      freq     - Fourier Frequency Array
%      cpd      - CPD Array
%      noicpd   - Poisson Noise Contribution to CPD
%      lopsd    - Non-noise-corrected PSD (low channel)
%      hipsd    - Non-noise-corrected PSD (high channel)
%      noilopsd - Noise-Contribution to low PSD
%      noihipsd - Noise-Contribution to high PSD
%      siglopsd - Noise-corrected PSD (low channel)
%      sighipsd - Noise-corrected PSD (high channel)
%      alln     - number of averaged segments in each frequency bin
%    Output: Struct containing
%      rawcof   - non-noise-corrected coherence function [Vaughan & Nowak, 1997, ApJ, 474, L43 (Eqn. 2)]
%      cof      - noise-corrected coherence function [Vaughan & Nowak, 1997, ApJ, 474, L43 (Eqn. 8, Part 1)]
%      errcof   - one-sigma uncertainty of cof [Vaughan & Nowak, 1997, ApJ, 474, L43 (Eqn. 8, Part 2)]
%      lag      - time lag [Nowak et al., 1999, ApJ, 510, 874 (Sect. 4)]
%      errlag   - one-sigma uncertainty of lag [Nowak et al., 1999, ApJ, 510, 874 (Eqn. 16)]
%      sigcpd   - noise-corrected cross-power-density [sigcpd = cpd - noicpd]
%!%-
{
  variable sigcpd = sqr(abs(cpd)) - noicpd;
  variable cof = sigcpd / (siglopsd * sighipsd);
  variable rawcof = (sigcpd + noicpd) / (lopsd * hipsd);
  variable dcof = (1.-cof) / sqrt(abs(cof)) * sqrt(2./alln);
  variable errcof = sqrt(  sqr(dcof/cof) + 2.*sqr(noicpd/sigcpd)
			 + (sqr(noilopsd/siglopsd) + sqr(noihipsd/sighipsd))/alln
			);
  variable lag = atan2(Imag(cpd), Real(cpd)) / (2.*PI*freq);
  variable errlag = sqrt( (1.-rawcof)/(2.*rawcof)/alln ) / (2.*PI*freq);
  return struct{rawcof=rawcof, cof=cof, errcof=errcof,
                lag=lag, errlag=errlag, sigcpd=sigcpd};
}

%
% Code from src/slang/color/color_blackbody.sl
%
define color_blackbody (T)
%define blackbody_to_rgb (T)
%!%+
%\function{color_blackbody}
%\synopsis{Computes RGB colors of hot blackbody radiation.}
%\usage{Integer_Type color = color_blackbody(Double_Type T);}
%\description
%    \code{T} is the temperature of the blackbody in Kelvin.
%    The return value is its 24 bit RGB value.
%
%    It is computed after the code "RGB VALUES FOR HOT OBJECTS"
%    by Dan Bruton (astro@sfasu.edu), which can be found at
%       http://www.physics.sfasu.edu/astro/color.html.
%!%-
{
  variable wavelength = [380:780:5];  % wavelength in nm
  % integration by trapezoid method
  variable weight = wavelength*0 + 1.;  weight[[0,-1]] = 0.5;  % properly weight end points
  % generate a black body spectrum
  variable dis, x, y, z;

  variable i;

  x = Double_Type[length(T)];
  y = Double_Type[length(T)];
  z = Double_Type[length(T)];

  _for i (0, length(T)-1)
  {
    dis = 3.74183e-16/wavelength^5/(exp(1240./8.617e-5/(wavelength*T[i]))-1);
                               % CIE Color Matching Functions (x_bar,y_bar,z_bar)
                               % for wavelenghts in 5 nm increments from 380 nm to 780 nm.
    % simple integration over bands
    x[i] = sum( weight*dis*[0.0014, 0.0022, 0.0042, 0.0076, 0.0143, 0.0232, 0.0435, 0.0776, 0.1344,
			    0.2148, 0.2839, 0.3285, 0.3483, 0.3481, 0.3362, 0.3187, 0.2908, 0.2511,
			    0.1954, 0.1421, 0.0956, 0.0580, 0.0320, 0.0147, 0.0049, 0.0024, 0.0093,
			    0.0291, 0.0633, 0.1096, 0.1655, 0.2257, 0.2904, 0.3597, 0.4334, 0.5121,
			    0.5945, 0.6784, 0.7621, 0.8425, 0.9163, 0.9786, 1.0263, 1.0567, 1.0622,
			    1.0456, 1.0026, 0.9384, 0.8544, 0.7514, 0.6424, 0.5419, 0.4479, 0.3608,
			    0.2835, 0.2187, 0.1649, 0.1212, 0.0874, 0.0636, 0.0468, 0.0329, 0.0227,
			    0.0158, 0.0114, 0.0081, 0.0058, 0.0041, 0.0029, 0.0020, 0.0014, 0.0010,
			    0.0007, 0.0005, 0.0003, 0.0002, 0.0002, 0.0001, 0.0001, 0.0001, 0.0000] );
    y[i] = sum( weight*dis*[0.0000, 0.0001, 0.0001, 0.0002, 0.0004, 0.0006, 0.0012, 0.0022, 0.0040,
			    0.0073, 0.0116, 0.0168, 0.0230, 0.0298, 0.0380, 0.0480, 0.0600, 0.0739,
			    0.0910, 0.1126, 0.1390, 0.1693, 0.2080, 0.2586, 0.3230, 0.4073, 0.5030,
			    0.6082, 0.7100, 0.7932, 0.8620, 0.9149, 0.9540, 0.9803, 0.9950, 1.0000,
			    0.9950, 0.9786, 0.9520, 0.9154, 0.8700, 0.8163, 0.7570, 0.6949, 0.6310,
			    0.5668, 0.5030, 0.4412, 0.3810, 0.3210, 0.2650, 0.2170, 0.1750, 0.1382,
			    0.1070, 0.0816, 0.0610, 0.0446, 0.0320, 0.0232, 0.0170, 0.0119, 0.0082,
			    0.0057, 0.0041, 0.0029, 0.0021, 0.0015, 0.0010, 0.0007, 0.0005, 0.0004,
			    0.0002, 0.0002, 0.0001, 0.0001, 0.0001, 0.0000, 0.0000, 0.0000, 0.0000] );
    z[i] = sum( weight*dis*[0.0065, 0.0105, 0.0201, 0.0362, 0.0679, 0.1102, 0.2074, 0.3713, 0.6456,
			    1.0391, 1.3856, 1.6230, 1.7471, 1.7826, 1.7721, 1.7441, 1.6692, 1.5281,
			    1.2876, 1.0419, 0.8130, 0.6162, 0.4652, 0.3533, 0.2720, 0.2123, 0.1582,
			    0.1117, 0.0782, 0.0573, 0.0422, 0.0298, 0.0203, 0.0134, 0.0087, 0.0057,
			    0.0039, 0.0027, 0.0021, 0.0018, 0.0017, 0.0014, 0.0011, 0.0010, 0.0008,
			    0.0006, 0.0003, 0.0002, 0.0002, 0.0001, 0.0000, 0.0000, 0.0000, 0.0000,
			    0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,
			    0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,
			    0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000] );
    % re-normalize the color scale
    variable mx = max([x[i], y[i], z[i]]);
    x[i] /= mx;
    y[i] /= mx;
    z[i] /= mx;
  }

  % Chromaticity Coordinates for Red, Green, Blue and White
  variable xr=0.64  ,  yr=0.33  ,  zr=1-xr-yr;
  variable xg=0.29  ,  yg=0.60  ,  zg=1-xg-yg;
  variable xb=0.15  ,  yb=0.06  ,  zb=1-xb-yb;
  variable xw=0.3127,  yw=0.3291,  zw=1-xw-yw;
  % convert to RGB
  variable         denominator = (xr*yg-xg*yr)*zb + (xb*yr-xr*yb)*zg + (xg*yb-xb*yg)*zr;
  variable r = _max(0, _min(1, ( (x *yg-xg*y )*zb + (xg*yb-xb*yg)*z  + (xb*y -x *yb)*zg )/denominator ));
  variable g = _max(0, _min(1, ( (xr*y -x *yr)*zb + (xb*yr-xr*yb)*z  + (x *yb-xb*y )*zr )/denominator ));
  variable b = _max(0, _min(1, ( (xr*yg-xg*yr)*z  + (x *yr-xr*y )*zg + (xg*y -x *yg)*zr )/denominator ));

  variable gamma = 0.8;
  variable rangeMax = _max([1e-10, r, g, b]);
  r = (r/rangeMax)^gamma;
  g = (g/rangeMax)^gamma;
  b = (b/rangeMax)^gamma;

  return rgb2hex(r,g,b);
}

%
% Code from src/data/lightcurve/color_color_data.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define color_color_data()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{color_color_data}
%\synopsis{extracts light curves in 3 consecutive bands and the corresponding colors from an event list}
%\usage{Struct_Type dat = color_color_data(Struct_Type evts, Double_Type E0, E1, E2, E3);}
%\qualifiers{
%\qualifier{dt}{time resolution [default: 100]}
%\qualifier{field}{field used for selection of the bands [default: "pi"]}
%}
%\description
%    \code{dat.a} = lightcurve in "E0 <= field < E1" band\n
%    \code{dat.b} = lightcurve in "E1 <= field < E2" band\n
%    \code{dat.c} = lightcurve in "E2 <= field < E3" band\n
%    \code{dat.}x\code{_}y = x/y color for x, y = a | b | c
%\seealso{histogram}
%!%-
{
  variable evts, E0, E1, E2, E3; (evts, E0, E1, E2, E3) = ();

  variable t_lo = [min(evts.time) : max(evts.time) : qualifier("dt", 100)];
  variable t_hi = make_hi_grid(t_lo);
  variable field = get_struct_field(evts, qualifier("field", "pi"));
  variable dat = struct {
    time_lo = t_lo,
    time_hi = t_hi,
    A = histogram(evts.time[where(E0 <= field < E1)], t_lo, t_hi),
    B = histogram(evts.time[where(E1 <= field < E2)], t_lo, t_hi),
    C = histogram(evts.time[where(E2 <= field < E3)], t_lo, t_hi),
    A_B, A_C, B_A, B_C, C_A, C_B
  };
  print(dat);
  struct_filter(dat, where(dat.A>0 and dat.B>0 and dat.C>0));
  dat.A_B = 1.0*dat.A / dat.B;
  dat.A_C = 1.0*dat.A / dat.C;
  dat.B_A = 1.0*dat.B / dat.A;
  dat.B_C = 1.0*dat.B / dat.C;
  dat.C_A = 1.0*dat.C / dat.A;
  dat.C_B = 1.0*dat.C / dat.B;

  return dat;
}

%
% Code from src/slang/color/color_complex.sl
%
define color_complex (z)
%!%+
%\function{color_complex}
%\synopsis{Converts a complex number to a RGB color}
%\usage{Integer_Type color = complex2rgb(Complex_Type z);}
%\description
%    arg(\code{z}) determines the hue,
%    |\code{z}| determines the lightness:
%    0<=|\code{z}|<=1 is mapped from black to color,
%    1<=|\code{z}|<inf is mapped from color to white.
%    The return value is a 24-bit RGB value.
%\qualifiers{
%\qualifier{s[=1.0]}{saturation (from 0 to 1) of the color}
%}
%\seealso{hsl2rgb, png_write}
%!%-
{
  variable s = _max(0.0, _min(1.0, qualifier("s", 1.0)));
  variable h = (atan2(Imag(z), Real(z))/2/PI + 1.0) mod 1.0;
  variable l = abs(z)*0.5;
  variable w = where(l>0.5);
  l[w] = 1.0-0.25/l[w];

  return rgb2hex(hsluv2rgb(h,s,l));
}

%
% Code from src/slang/color/color_desaturate.sl
%
define color_desaturate (color)
%!%+
%\function{color_desaturate}
%\synopsis{Weighted color to gray conversion}
%\usage{UInt_Type gray = color_desaturate(UInt_Type[] color);}
%\description
%    Converts a RGB color, hex encoded, into a weighted gray scale using:
%
%    \code{gray} = (0.3 * R) + (0.59 * G) + (0.11 * B)
%    
%    The 8 bit gray value(s) are return as UInt_Type[] in the interval [0,255].
%
%\seealso{rgb2hsl, hex2rgb}
%!%-
{ 
  variable r, g, b;
  (r,g,b) = hex2rgb(color);

  variable gray = (0.3 * r/255.) + (0.59 * g/255.) + (0.11 * b/255.);

  return rgb2hex(r, g, b);
}

%
% Code from src/slang/color/color_mix.sl
%
define color_mix (c1,c2,fraction)
%!%+
%\function{color_mix}
%\synopsis{Mix two rgb colors}
%\usage{UInt_Type color = color_mix (UInt_Type c1, UInt_Type c2, Double_Type fraction);}
%\description
%  This function mixes two colors in rgb space. Given the rgb values in the
%  usual encoding in one 24bit integer, the color is mixed according to 
%  new color  = c*fraction + c2*(1-fraction)
%  The operations are analoguous to the color mixing performed by the xcolor
%  package of LaTeX (the operation is similar to the color1!fraction!color2
%  syntax). 
%  The function returns the integer specifying the new rgb values. \code{fraction}
%  can be an array. The colors can also be a hex string.
%!%-
{
  fraction = _max(0., _min(1., fraction));

  variable r1,g1,b1,r2,g2,b2;

  (r1,g1,b1) = hex2rgb(c1);
  (r2,g2,b2) = hex2rgb(c2);

  return rgb2hex((r1*fraction+(1.0-fraction)*r2)/255.,
		 (g1*fraction+(1.0-fraction)*g2)/255.,
		 (b1*fraction+(1.0-fraction)*b2)/255.);
}

%
% Code from src/slang/color/color_wavelength.sl
%
define color_wavelength (lambda)
%!%+
%\function{color_wavelength}
%\synopsis{Computes the color values of visual light}
%\usage{UInt_Type[] color_wavelength (Double_Type[] lambda);}
%\description
%    \code{380 <= lambda <= 780} is the wavelength in nm.
%    The return value is its 24bit RGB color value.
%
%    The conversion is performed after the code by Dan Bruton,
%    see http://www.physics.sfasu.edu/astro/color.html.
%!%-
{
  variable r = Double_Type[length(lambda)];
  variable g = @r;
  variable b = @r;
  variable w, spec;

  w = where(380 <= lambda < 420);
  spec = 0.3 + 0.7*(lambda[w]-380.)/(420-380);
  r[w] = (440.-lambda[w])/(440-380)*spec;
  % g[w] = 0;
  b[w] = spec;

  w = where(420 <= lambda < 440);
  r[w] = (440.-lambda[w])/(440-380);
  % g[w] = 0;
  b[w] = 255;

  w = where(440 <= lambda < 490);
  % r[w] = 0;
  g[w] = (lambda[w]-440.)/(490-440);
  b[w] = 1.0;

  w = where(490 <= lambda < 510);
  % r[w] = 0;
  g[w] = 1.0;
  b[w] = (510.-lambda[w])/(510-490);

  w = where(510 <= lambda < 580);
  r[w] = (lambda[w]-510.)/(580-510);
  g[w] = 1.0;
  % b[w] = 0;

  w = where(580 <= lambda < 645);
  r[w] = 1.0;
  g[w] = (645.-lambda[w])/(645-580);
  % b[w] = 0;

  w= where(645 <= lambda < 700);
  r[w] = 1.0;
  % b[w] = 0;
  % g[w] = 0;

  w = where(700 <= lambda < 780);
  r[w] = 0.3 + 0.7*(780.-lambda[w])/(780-700);
  % g[w] = 0;
  % b[w] = 0;

  return rgb2hex(r, g, b);
}

%
% Code from src/misc/radio/compare_modelfits.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

define compare_modelfits ()
%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{compare_modelfits}
%\synopsis{to compare modelfit parameters of different epochs}
%\usage{compare_modelfits([array of \code{epoch fitsfiles}]);}
%\qualifiers{
%\qualifier{plottype}{[="overlay"] layout of the comparison plot,
%                               by default it shows the overlay of all component positions,
%                               alternatives are "distancevsmjd","fluxvsdistance","TBvsdistance"}
%\qualifier{fileextent}{[=.eps] output is given by default as .eps-file in working directory}
%\qualifier{outputfile}{give outputfile name and directory by hand}
%\qualifier{sourcename}{[=default] set the name of the source, by default the name is read from the .fits file,\n
%                               set to NULL for not plotting a source name}
%\qualifier{ra_mas}{[=[20,-20]] ra range (for "overlay")}
%\qualifier{dec_mas}{[=[-20,20]] dec range (for "overlay")}
%\qualifier{mjd_min}{[=54101.0] lower limit of time axis (default: 1/1/2007)}
%\qualifier{mjd_max}{[=55927.0] lower limit of time axis (default: 1/1/2012)}
%\qualifier{distance_min}{[=0] lower limit of distance axis in mas}
%\qualifier{distance_max}{[=50] upper limit of distance axis in mas}
%\qualifier{flux_min}{[=1e-5] lower limit of flux axis in Jy}
%\qualifier{flux_max}{[=2] upper limit of flux axis in Jy}
%\qualifier{TB_min}{[=10^5] lower limit of brightness temperature axis in K}
%\qualifier{TB_max}{[=10^15] upper limit of brightness temperature axis in K}
%\qualifier{linestyle}{[=default] set to 0 if the flux values should not be conntected (or to another value for
%                               another line style}
%\qualifier{ex_comps}{[=0] set to 1 if components should be excluded and define
%                               the corresponding quadrant of the plot via the ex_comps-coordinates}
%\qualifier{ex_comps_ra}{[=NULL] quadrant coordinates in mas where components should be excluded}
%\qualifier{ex_comps_dec}{[=NULL] define quadrant coordinates in mas where components should be excluded}
%}
%\description
%    This function creates an overlay of a VLBI-clean image and the corresponding\n
%    model of Gaussian components which are over-plotted as ellipses.\n
%    The required input format are fits-file for both the clean and the modelfit images.
%    The format of the output file depends on the suffix of the given\n
%    \code{filename}. Possible formats of the output file are PDF, EPS,\n
%    PNG, GIF, etc.
%    If labelling = 1 is set, the components are labeled with J0,J1,J2,... depending on
%    their distance to [0,0].
%    The counterjet components (and core) can be excluded by defining the corresponding
%    quadrant of the the plot via ex_comps_ra and _dec.
%!%-
{
variable epochs;
  switch(_NARGS)
  { case 1: (epochs) = (); }
  { help(_function_name()); return; }
%%%%%%%%%%%%%%%
variable plottype       = qualifier("plottype", "overlay");
variable sourcename     = qualifier("sourcename", fits_read_key (epochs[0], "OBJECT"));
variable ra_mas         = qualifier("ra_mas", [20,-20]);
variable dec_mas        = qualifier("dec_mas", [-20,20]);
variable fileextent     = qualifier("fileextent", "eps");
variable mjd_min        = qualifier("mjd_min", 54101.0); %1.1.2007
variable mjd_max        = qualifier("mjd_max", 55927.0); % 1.1.2012
variable distance_min   = qualifier("distance_min", 0);
variable distance_max   = qualifier("distance_max", 50);
variable flux_min       = qualifier("flux_min",1e-3); % in Jy
variable flux_max       = qualifier("flux_max", 2);
variable ex_comps       = qualifier("ex_comps",0);
variable ex_comps_ra    = qualifier("ex_comps_ra", NULL);
variable ex_comps_dec   = qualifier("ex_comps_dec", NULL);
variable TB_min         = qualifier("TB_min", 10^5);
variable TB_max         = qualifier("TB_max", 10^15);
variable outputfile     = qualifier("outputfile", NULL);
%%%%%%%%%%%%%%%%%%%%
variable i,k;
variable factor = 3.6e6;
%%%%
variable pl = xfig_plot_new(15., 15.);
%%%%%
_for i (0, length(epochs)-1,1){
variable model = fits_read_table(epochs[i]);
variable RA = model.deltax*factor;
variable Decl = model.deltay*factor;
%%%%
variable excl;
if (ex_comps==1){
        %1.Quadrant:
        if (max(ex_comps_ra)<=0 and (min(ex_comps_dec)>=0)){
%       theta = +0-90
        excl = wherenot(RA<=0.0 and Decl>=0);}
        %2.Quadrant:
        if (min(ex_comps_ra)>=0 and (min(ex_comps_dec)>=0)){
%       theta = -0-90%
        excl = wherenot(RA>=0.0 and Decl>=0);}
        %3.Quadrant:
        if (min(ex_comps_ra)>=0 and (max(ex_comps_dec)<=0)){
%       theta -90-180
        excl = wherenot(RA>=0.0 and Decl<=0);}
        %4.Quadrant:
        if (max(ex_comps_ra)<=0 and (max(ex_comps_dec)<=0)){
% %     theta = +90-180
        excl = wherenot(RA<=0.0 and Decl<=0);}
model = struct_filter (model,excl;copy);
}
variable date_ = fits_read_key(epochs[i],"DATE-OBS");
variable mjd = fits_read_key(epochs[i],"DATE-OBS");
variable Y, m, d; ()=sscanf(mjd, "%4d-%2d-%2d, %2d:%2d", &Y, &m, &d);
mjd = MJDofDate(Y, m, d);
variable distance =  sqrt((model.deltax*factor)^2+ (model.deltay*factor)^2);
variable frequency = fits_read_key(epochs[i],"CRVAL3")/1.e9; %in GHz
variable TB = 1.22*1e+12*model.flux/(frequency^2*(model.major_ax*factor)*(model.minor_ax*factor)); % see Condon et al. 1982
%%%%%% overlay
if (plottype=="overlay"){
pl.world(max(ra_mas), min(ra_mas), min(dec_mas), max(dec_mas));
pl.xlabel("relative RA [mas]");
pl.ylabel("relative DEC [mas]");
pl.xylabel(0.85,0.85,sourcename; world00);
variable major_ticks = [-100:100:5];
variable minor_ticks= [-100:100:1];
pl.x1axis(; major=major_ticks, minor= minor_ticks);
pl.y1axis(; major=major_ticks, minor= minor_ticks);

pl.xylabel(0.8,0.85-(i+1)/20.,string(date_)+" ("+string(mjd)+")"; color=i+1, world00);
pl.plot(0.63,0.85-(i+1)/20.; sym="+", symcolor=i+1, world00);
pl.plot(model.deltax*factor,model.deltay*factor; sym="+", symcolor=i+1);


variable phi=[0:2*PI:#200]; 
variable ex, ey;
variable smaj = factor*model.major_ax;
variable smin = factor*model.minor_ax;
variable posangle = (model.posangle/180.) * PI;
variable ii;
_for ii (0, length(model.deltax)-1,1){
(ex, ey) = ellipse(smaj[ii]/2.0,smin[ii]/2.0,PI/2.0-posangle[ii],phi);
pl.plot(model.deltax[ii]*factor+ex,model.deltay[ii]*factor+ey; color=i+1) ;
}
}
%%%
if (plottype=="distancevsmjd"){
pl.world( mjd_min, mjd_max,distance_min,distance_max);
pl.xlabel("Epoch [MJD]");
pl.ylabel("Distance [mas]");
major_ticks = [-100:100:5];
minor_ticks= [-100:100:1];
pl.y1axis(; major=major_ticks, minor= minor_ticks);
_for k (0, length(model.deltax)-1,1){
pl.plot(mjd, distance[k]; sym=i+1, symcolor=i+1);}
pl.xylabel(0.8,0.85-(i+1)/20.,string(date_)+" ("+string(mjd)+")"; color=i+1, world00);
pl.plot(0.63,0.85-(i+1)/20.; sym=i+1, symcolor=i+1, world00);
}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if (plottype=="fluxvsdistance"){
model = struct_combine (model, struct{distance=Double_Type[length(model.flux)]});
variable a;
for (a=0; a<length(model.flux); a++) {model.distance[a]=distance[a];};
variable model_sort = struct_filter(model, array_sort(model.distance);copy);
pl.world( distance_min,distance_max,flux_min, flux_max; ylog);
pl.ylabel("Flux [Jy]");pl.xlabel("Distance [mas]");
pl.plot(model_sort.distance,model_sort.flux;  sym=i+1, symcolor=i+1, line= qualifier("linestyle",2), color = i+1);
pl.xylabel(0.8,0.85-(i+1)/20.,string(date_)+" ("+string(mjd)+")"; color=i+1, world00);
pl.plot(0.63,0.85-(i+1)/20.; sym=i+1, symcolor=i+1, world00);}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if (plottype=="TBvsdistance"){
model = struct_combine (model, struct{TB=Double_Type[length(model.flux)]});
variable b;for (b=0; b<length(model.flux); b++) {model.TB[b]=TB[b];};
%
model = struct_combine (model, struct{distance=Double_Type[length(model.flux)]});
for (a=0; a<length(model.flux); a++) {model.distance[a]=distance[a];};
model_sort = struct_filter(model, array_sort(model.distance);copy);
pl.world( distance_min,distance_max,TB_min, TB_max; ylog);
pl.ylabel("$T_B$ [K]"R);pl.xlabel("Distance [mas]");
pl.plot(model_sort.distance,model_sort.TB;  sym=i+1, symcolor=i+1, line= qualifier("linestyle",2), color = i+1);
pl.xylabel(0.8,0.85-(i+1)/20.,string(date_)+" ("+string(mjd)+")"; color=i+1, world00);
pl.plot(0.63,0.85-(i+1)/20.; sym=i+1, symcolor=i+1, world00);
}
%%%%%
}% for-loop reading in all files
if (outputfile!=NULL){xfig_new_hbox_compound (pl, 0.4).render(outputfile);}
else{
xfig_new_hbox_compound (pl, 0.4).render(sourcename+"_"+plottype+"."+fileextent);}
}


#endif

%
% Code from src/fitting/parameters/compare_par.sl
%
%%%%%%%%%%%%%%%%%%
define compare_par()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{compare_par}
%\synopsis{compares models saved in parameter files}
%\usage{compare_par([String_Type pattern]);
%\altusage{Struct_Type info = compare_par([String_Type pattern]; get_list);}
%}
%\qualifiers{
%\qualifier{fit_fun}{include fit-function}
%\qualifier{fit_fun=fitfun}{only include models using the fit-function \code{fitfun}}
%\qualifier{get_list}{the information is not printed, but returned}
%\qualifier{load_best}{load the best fit parameters}
%\qualifier{verbose}{}
%}
%\description
%    It is assumed that the corresponding data sets are already initialized.
%\seealso{load_par, eval_counts}
%!%-
{
  variable pat;
  switch(_NARGS)
  { case 0: pat = "*.par"; }
  { case 1: pat = (); }
  { help(_function_name()); return; }

  variable include_fit_fun = qualifier_exists("fit_fun");
  variable fit_fun_to_look_for = qualifier("fit_fun");
  if(fit_fun_to_look_for != NULL)  include_fit_fun = 0;
  variable verbose = qualifier_exists("verbose");

  variable FV = Fit_Verbose;
  variable old_fit_fun = get_fit_fun();
  variable params = get_params();

  variable info = struct {
    filename = String_Type[0],
    mtime = String_Type[0],
    numPar = Integer_Type[0],
    numFreePar = Integer_Type[0],
    statistic = Double_Type[0],
    red_stat = Double_Type[0],
    best
  };
  if(include_fit_fun)
    info = struct {
      filename = String_Type[0],
      mtime = String_Type[0],
      fit_fun = String_Type[0],
      numPar = Integer_Type[0],
      numFreePar = Integer_Type[0],
      statistic = Double_Type[0],
      red_stat = Double_Type[0],
      best
    };
  variable parfile, e;
  foreach parfile (glob(pat))
  {
    try (e)  load_par(parfile);
    catch AnyError:  { if(verbose)  vmessage("%% %s could not be loaded due to %s exception (%s)", parfile, e.descr, e.message);
                       continue; }
    if(verbose)
      vmessage("%% %s", parfile);
    if(fit_fun_to_look_for==NULL || fit_fun_to_look_for==get_fit_fun())
    {
      variable stat;
      Fit_Verbose = -1;
        ()=eval_counts(&stat);
      Fit_Verbose = FV;
      info.filename  = [info.filename, parfile];
      info.mtime     = [info.mtime, strftime("%Y-%m-%d %H:%M", localtime(stat_file(parfile).st_mtime))];
      if(include_fit_fun)
        info.fit_fun    = [info.fit_fun, get_fit_fun()];
      info.numPar    = [info.numPar, length(get_params)];
      info.numFreePar= [info.numFreePar, stat.num_variable_params];
      info.statistic = [info.statistic, stat.statistic];
      info.red_stat  = [info.red_stat, stat.statistic/(stat.num_bins-stat.num_variable_params)];
    }
  }
  if(old_fit_fun != "")
  {
    fit_fun(old_fit_fun);
    set_params(params);
  }

  info.best = Char_Type[length(info.filename)];  % info.best[*] = 0;
  info.best[where_max(info.statistic)] = -1;
  info.best[where_min(info.statistic)] = +1;
  struct_filter(info, reverse(array_sort(info.statistic)));

  if(qualifier_exists("get_list"))
    info;  % left on stack
  else
    print_struct(info);

  if(qualifier_exists("load_best"))
    if(any(info.best==1))
    { message("loading parameters from "+info.filename[-1]);
      load_par(info.filename[-1]);
    }
    else
      message("no best fit parameters available");
}

%
% Code from src/slang/arrays/complementary_array.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define complementary_array(B, A)
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{complementary_array}
%\synopsis{returns the "difference" of arrays}
%\usage{Array_Type C = complementary_array(Array_Type B, A)}
%\description
%    \code{C = B} \\ \code{A}
%!%-
{
  return B[complement(B,A)];
}

%
% Code from src/slang/arrays/compressedArrayStr.sl
%
define compressedArrayStr(a)
{
  variable i=1, i0=0;
  variable last = length(a)-1;
  variable allLines = "",  str = "[";
  _for i (1, last, 1)
  {
    if(a[i]!=a[i0])
    {
      if(i0>0)
        if(strlen(str)>72)
        {
          allLines = allLines + str + ",\n";
          str = " ";
        }
        else
          str += ", ";
      if(i-i0>1)
      {
        str += sprintf("0*[1:%d]", i-i0);
        if(a[i0]!=0)  str += sprintf("%+d", a[i0]);
      }
      else
        str += sprintf("%d", a[i0]);
      i0 = i;
    }
  }
  if(i0>0)  str += ", ";
  if(last+1-i0>1)
  {
    str += sprintf("0*[1:%d]", last+1-i0);
    if(a[i0]!=0)  str += sprintf("%+d", a[i0]);
  }
  else
    str += sprintf("%d", a[i0]);

  return allLines + str + "]";
}

%
% Code from src/slang/math/continued_fraction.sl
%
define continued_fraction()
%!%+
%\function{continued_fraction}
%\synopsis{computes a continued fraction}
%\usage{Double_Type continued_fraction(UInteger_Type a[])}
%\description
%    \code{                                  1 |      1 |}\n
%    \code{continued_fraction(a) = a[0] + ------ + ------ + ...}\n
%    \code{                               | a[1]   | a[2]}
%!%-
{
  variable a;
  switch(_NARGS)
  { case 1: a = (); }
  { help(_function_name()); return; }

  variable n_ = 0,  n = 1,  d_ = 1,  d = 0;
  variable i=0;
  _for i (0, length(a)-1, 1)
    (n_, d_, n, d) = (n, d, n_+a[i]*n, d_+a[i]*d);
  return 1.*n/d;
}

%
% Code from src/slang/math/continued_fraction_expansion.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define continued_fraction_expansion()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{continued_fraction_expansion}
%\synopsis{expands a floating point number as a continued fraction}
%\usage{UInteger_Type[] continued_fraction_expansion(Double_Type x)}
%\qualifiers{
%\qualifier{verbose}{}
%\qualifier{maxlen}{[\code{=64}]}
%}
%!%-
{
  variable x0;
  switch(_NARGS)
  { case 1: x0 = (); }
  { help(_function_name()); return; }

  if(x0<0)  x0 = -x0;
  variable verbose = qualifier_exists("verbose");
  if(verbose)  vmessage("continued fraction expansion of %S:", x0);
  variable N = qualifier("maxlen", 64);
  variable a = UInteger_Type[N], n = ULLong_Type[N+2], d = ULLong_Type[N+2];
  n[0] = 0; n[1] = 1;  d[0] = 1; d[1] = 0;
  variable i = 0, x = x0, cf = "[", fmt = "%d";
  do
  { variable a_i = int(x);
    a[i] = a_i;
    cf += sprintf(fmt, a_i);
    if(i<2)  { if(i==0)  fmt=";%d";  else  fmt=",%d"; }
    x = x-a_i;
    if(x>0) x = 1./x;
    n[i+2] = n[i] + a_i * n[i+1];
    d[i+2] = d[i] + a_i * d[i+1];
    variable frac = 1.*n[i+2] / d[i+2];
    variable diff = frac - x0;
    if(verbose)  vmessage("%s] = %lld/%lld = %S  (delta = %S)", cf, n[i+2], d[i+2], frac, diff);
    i++;
  } while(i<N && diff!=0.);
  return a[[:i-1]];
}

%
% Code from src/plot/contour_borders.sl
%

% only activate if module gcontour could be loaded
#ifeval __get_reference("gcontour_compute")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define contour_borders()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{contour_borders}
%\synopsis{Returns outer borders of a contour}
%\usage{Double_Type[] = contour_borders(fits);}
%\qualifiers{
%\qualifier{conf_level [2]}{: set confidence level for which you need the borders}
%}
%\description
%   This function takes as input a contour in fits-format saved with
%   save_conf and returns the array [par1_min,par1_max,par2_min,par2_max].
%   Select the confidence level with the conf_level quelifier (1,2,3).
%\example
%   contour_borders("contour.fits";);
%\seealso{function_name2, function_name3}   
%!%-
{
  variable cont;
  switch(_NARGS)
  { case 1: cont = (); }
  { help(_function_name()); return; }

  variable conf_level = qualifier("conf_level",2);
  variable delta_chi;
  switch(conf_level)
  { case 1: delta_chi=2.3; }
  { case 2: delta_chi=4.61; }
  { case 3: delta_chi=9.21; }  
  
  variable c, line, cont_min, cont_max, p1_min, p1_max, p2_min, p2_max, stepx, stepy;
  
  c = load_conf (cont);
  line = gcontour_compute(c.chisqr,delta_chi);
  
  stepx = double(c.px.max - c.px.min)/c.px.num;
  stepy = double(c.py.max - c.py.min)/c.py.num;
  
  p1_min = c.px.min + (min(line[0].x_list[0])) * stepx;
  p1_max = c.px.min + (max(line[0].x_list[0])) * stepx;
  
  p2_min = c.py.min + (min(line[0].y_list[0])) * stepy;
  p2_max = c.py.min + (max(line[0].y_list[0])) * stepy;
  
  return [p1_min,p1_max,p2_min,p2_max];
}

#endif

%
% Code from src/fitting/contour_trq.sl
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%    
define contour_trq(){
%%%%%%%%%%%%%%%%%%%%%%%%%%    
%!%+
%\function{contour_trq}
%\synopsis{creates and (with qualifier trq) sends bins^2 jobs to torque (at Remeis cluster) and
%        calculates chi_2 values for each point for chosen 2 parameters.}
%\usage{contour_trq(String_Type inputPar1, inputPar2, Int_Type bins,
%        Double_Type lowLimPar1, upLimPar1,lowLimPar2, upLimPar2,
%        String_Type startUpFile [, parFile], outDir, time);}  
%\qualifiers{
%\qualifier{trq}{Sends the jobs to Remeis torque}
%}
%
%\description
%    - \code{inputPar1}     input parameter 1 written as in .par file
%    - \code{inputPar2}     input parameter 2 written as in .par file
%    - \code{bins}          number of chosen bins
%    - \code{lowLimPar1}    lower parameter 1 value
%    - \code{upLimPar1}     upper parameter 1 value
%    - \code{lowLimPar2}    lower parameter 2 value
%    - \code{upLimPar2}     upper parameter 2 value
%    - \code{startUpFile}   upload file (with data, binning,
%                                 noticing etc.)
%    - \code{parFile}       best fit parameter file name (optional) 
%    - \code{outDir}        output directory for torque files and
%                                 calculated output files
%    - \code{time}          torque wall time
%
%    In case where values for all bins^2 points are not calculated (for
%    example when the torque walltime is shorter than needed), one can run
%    function missing_contour_trq in order to get all bins^2 values. 
%    
%    To plot the result use the function plot_contour_trq.
%
%
% EXAMPLE
%
%   contour_trq(`relconv(1).Incl`,`reflionx(1).Xi`,32,25,35,1000,4000,"data.sl",
%   "results/bestFit.par","outputInclXi","02:00:00";trq);
%
%   will calculate 32 by 32 values for two parameters and write the
%   output files to "outputInclXi" subdirectory. It will do that by
%   sending the 32x32 jobs to Remeis torque via qualifier "trq", and
%   giving it a walltime of 02:00:00 hours. 
%   
%\seealso{missing_contour_trq, plot_contour_trq}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%%%    

   variable i=Int_Type;
   variable j=Int_Type;
   variable par1=String_Type;
   variable par2=String_Type;
   variable bins=Int_Type;
   variable par1Low=Double_Type;
   variable par1High=Double_Type;
   variable par2Low=Double_Type;
   variable par2High=Double_Type;
   variable loadData=String_Type;
   variable loadPar=String_Type;
   variable outDir=String_Type;
   variable walltime=String_Type;
   
   switch(_NARGS)
     {case 10: (par1,par2,bins,par1Low,par1High,par2Low,par2High,loadData,outDir,walltime)=(); }    
     {case 11: (par1,par2,bins,par1Low,par1High,par2Low,par2High,loadData,loadPar,outDir,walltime)=(); }
     {help (_function_name());return;}
   

   variable xbins=[par1Low:par1High:#bins];
   variable ybins=[par2Low:par2High:#bins];

   ()=system("mkdir $outDir"$);
   variable files=outDir+"/Files";
   ()=system("mkdir $files"$);

   variable par1_out=strchop(par1,'.',0);
   variable par2_out=strchop(par2,'.',0);
   variable outFile=par1_out[1]+"_"+par2_out[1]+".trqFile";
   variable trqFile=fopen(outFile,"w");


   _for i(0,length(xbins)-1,1){
      _for j(0,length(ybins)-1,1){
	 variable str_x=sprintf("%012.8f",xbins[i]);
	 variable str_y=sprintf("%012.8f",ybins[j]);
	 
	 variable out=fopen(files+"/"+str_x+"_"+str_y+"_trq","w");
	
	 ()=fprintf(trqFile,"isis "+files+"/"+str_x+"_"+str_y+"_trq\n");
	 
	 ()=fprintf(out,"require(\"isisscripts\");\n");
	 ()=fprintf(out,"evalfile(\"$loadData\");\n"$);
	 if (is_defined("loadPar")==0)  message("No parameter file given! loadData must load the parameters instead!");
	 else ()=fprintf(out,"load_par(\"$loadPar\");\n"$);
	 ()=fprintf(out,"set_par(\"$par1\",%.8f,1);\n"$,xbins[i]);
	 ()=fprintf(out,"set_par(\"$par2\",%.8f,1);\n"$,ybins[j]);
	 ()=fprintf(out,"fit_counts;\n");
	 ()=fprintf(out,"save_statistics(\"%s/%012.8f_%012.8f_trq\");\n \n",outDir,xbins[i],ybins[j]);
	 ()=fprintf(out,"exit;\n");
	 ()=fclose(out);
      }
   }
   ()=fclose(trqFile);
   if(qualifier_exists("trq")){
      ()=system("qsub_array $outFile --walltime=$walltime --arch=x86_64"$);
   }
}


%
% Code from src/plot/coordX_in_box.sl
%
define coordX_in_box()
%!%+
%\function{coordX_in_box}
%\synopsis{calculates a world x-coordinate from the relative x-coordinate in the plot box}
%\usage{Double_Type coordY_in_box(Double_Type x_rel)}
%\description
%    The left boundary of the plot box has \code{x_rel==0},
%    and the right boundary has \code{x_rel==1}.
%    Logarithmic world-coordinates are properly taken into account.
%    Note that the xrange has to be set in advance
%    in order to calculate the world x-coordinate with \code{coordX_in_box}.
%\seealso{coordY_in_box, coords_in_box}
%!%-
{
  variable x_rel;
  switch(_NARGS)
  { case 1: x_rel = (); }
  { help(_function_name()); return; }

  variable plot_opt = get_plot_options();
  if(plot_opt.logx)
    return plot_opt.xmin * (plot_opt.xmax / plot_opt.xmin)^x_rel;
  else
    return plot_opt.xmin + x_rel * (plot_opt.xmax - plot_opt.xmin);
}

%
% Code from src/plot/coordY_in_box.sl
%
define coordY_in_box()
%!%+
%\function{coordY_in_box}
%\synopsis{calculates a world y-coordinate from the relative y-coordinate in the plot box}
%\usage{Double_Type coordY_in_box(Double_Type y_rel)}
%\description
%    The lower boundary of the plot box has \code{y_rel==0},
%    and the upper boundary has \code{y_rel==1}.
%    Logarithmic world-coordinates are properly taken into account.
%    Note that the yrange has to be set in advance
%    in order to calculate the world y-coordinate with \code{coordY_in_box}.
%\seealso{coordX_in_box, coords_in_box}
%!%-
{
  variable y_rel;
  switch(_NARGS)
  { case 1: y_rel = (); }
  { help(_function_name()); return; }

  variable plot_opt = get_plot_options();
  if(plot_opt.logy)
    return plot_opt.ymin * (plot_opt.ymax / plot_opt.ymin)^y_rel;
  else
    return plot_opt.ymin + y_rel * (plot_opt.ymax - plot_opt.ymin);
}

%
% Code from src/plot/coords_in_box.sl
%
define coords_in_box()
%!%+
%\function{coords_in_box}
%\synopsis{calculates world coordinates from the relative coordinates in the plot box}
%\usage{(Double_Type x, y) = coordY_in_box(Double_Type x_rel, y_rel);}
%\description
%    Note that the x- and yrange has to be set in advance
%    in order to calculate the world coordinates with \code{coords_in_box}.
%\seealso{coordX_in_box, coordY_in_box}
%!%-
{
  variable relX, relY;
  switch(_NARGS)
  { case 2: (relX, relY) = (); }
  { help(_function_name()); return; }

  return (coordX_in_box(relX), coordY_in_box(relY));
}

%
% Code from src/slang/math/correlation_coefficient.sl
%
define correlation_coefficient()
%!%+
%\function{correlation_coefficient}
%\synopsis{calculates the [weighted] linear correlation coefficient between two arrays}
%\usage{Double_Type rho = correlation_coefficient(Double_Type x[], Double_Type y[] [ , Double_Type w[] ]);}
%\description
%
%   The function of the unweighted correlation coefficient reads:
%
%   \code{                     n * sum(x*y)  -  sum(x) * sum(y)          }
%   \code{rho  =  -------------------------------------------------------}
%   \code{         sqrt(n*sum(x^2)-sum(x)^2) * sqrt(n*sum(y^2)-sum(y)^2) }
%   
%   w defines the weight of each point. By default it is set to w = 1.
%   
%\seealso{find_correlations}
%!%-
{
   variable x, y, w;
   switch(_NARGS)
     
   { case 2: 
       (x, y) = (); 
     variable n = length(x);
     variable sumX = sum(x);
     variable sumY = sum(y);
     return (n*sum(x*y)-sumX*sumY)/sqrt(n*sum(x^2)-sumX^2)/sqrt(n*sum(y^2)-sumY^2);
  }
   
   { case 3: 
	(x, y, w) = ();
      
      variable mx = sum(w*x)/sum(w);
      variable my = sum(w*y)/sum(w);
      
      variable covxy = sum( w* (x-mx) * (y-my))   / sum(w);
      variable covxx = sum( w* (x-mx) * (x-mx))   / sum(w);
      variable covyy = sum( w* (y-my) * (y-my))   / sum(w);
      
      return covxy/sqrt(covxx*covyy);
      }
      
   { help(_function_name()); return; }
   
}

%
% Code from src/slang/astronomy/cosmo_param.sl
%
define cosmo_param() %,omega_m, omega_lambda, omega_k, q0
%!%+
%\function{cosmo_param}
%\synopsis{Derive full set of cosmological density parameters from a partial set}
%\usage{Struct_Type P = cosmo_param();}
%\qualifiers{
%\qualifier{omega_m}{value of the normalized matter density (Omega_M)}
%\qualifier{omega_lambda}{value of the normalized cosmological constant (Omega_Lambda)}
%\qualifier{omega_k}{value of the the normalized curvature term (Omega_k)}
%\qualifier{q0}{value of the deceleration parameter (q0)}
%}
%\description
%    This procedure is called by lumdist to allow the user a choice
%    in defining any two of four cosmological density parameters.
%    Given any two of the four input parameters this  program will derive
%    the remaining two.
%    Omega_M      - Normalized matter energy density, non-negative numeric scalar
%    Omega_Lambda - Normalized cosmological constant, numeric scalar
%    Omega_k      - Normalized curvature parameter, numeric scalar.
%                   This is zero for a flat universe
%    q0           - Deceleration parameter, numeric scalar = -R*(R'')/(R')^2
%                   = 0.5*Omega_m - Omega_lambda
%    Here "normalized" means divided by the closure density so that
%    Omega_m + Omega_lambda + Omega_k = 1.    For a more
%    precise definition see Carroll, Press, & Turner (1992, ArAA, 30, 499).
%
%    If less than two parameters are defined, this procedure sets default
%    values of Omega_k=0 (flat space), Omega_lambda = 0.7, Omega_m = 0.3
%    and q0 = -0.55
%
%    If more than two parameters are defined upon input (overspecification), 
%    then the first two defined parameters in the ordered list Omega_m, 
%    Omega_lambda, Omega_k, q0 are used to define the cosmology.
%    
%    (This procedure is translated from the IDL COSMO_PARAM function.)
%
%\example
%    variable default_set = cosmo_param();
%    print (default_set);
%
%    variable my_set = cosmo_param(; omega_k = 0.1, omega_lambda = 0.65);
%    print (my_set);
%\seealso{lumdist}
%!%-
{
  variable omega_m      = qualifier("omega_m"      ,NULL);
  variable omega_lambda = qualifier("omega_lambda" ,NULL);
  variable omega_k      = qualifier("omega_k"      ,NULL);
  variable q0           = qualifier("q0"           ,NULL);
  
  % check if parameters are set:
  variable Nk      = (omega_k      != NULL);
  variable Nlambda = (omega_lambda != NULL);
  variable Nomega  = (omega_m      != NULL);
  variable Nq0     = (q0           != NULL);
  
%  if ( Nk + Nlambda + Nomega + Nq0 < 2) % less than 2 params are set
%  {
%    message("Two parameters have to be set.");
%    return;
%  }
  if ( Nk + Nlambda + Nomega + Nq0 > 2)
    vmessage("WARNING -- %s -- More than 2 parameters are set!", _function_name);
  
  % Check which two parameters are defined, and then determine the other two
  if (Nlambda and Nq0)
  {
    if (     Nk == 0)      omega_k = 1. - (2.*q0 + 3.*omega_lambda);
    if ( Nomega == 0)      omega_m = 1.- (omega_lambda + omega_k);
  }
  if (Nk and Nq0)
  {
    if ( Nomega == 0)      omega_m = (1 + q0 - omega_k)*2./3.;
    if (Nlambda == 0) omega_lambda = 1. - (omega_m + omega_k);
  }
  if (Nomega and Nq0)
  {
    if (     Nk == 0)      omega_k = 1. + q0 - 3.*omega_m/2.;
    if (Nlambda == 0) omega_lambda = 1. - (omega_m + omega_k);
  }
  if (Nlambda and Nk)
  {
    if ( Nomega == 0)      omega_m = 1.- (omega_lambda + omega_k);
    if (    Nq0 == 0)           q0 = (1 - (omega_k + 3.*omega_lambda))/2.;
  }
  if (Nomega and Nk)
  {
    if (Nlambda == 0) omega_lambda = 1. - (omega_m + omega_k);
    if (    Nq0 == 0)           q0 = -1. + omega_k + 3*omega_m/2;
  }
 if (Nomega and Nlambda)
  {
    if (     Nk == 0)      omega_k = 1. - (omega_m + omega_lambda);
    if (    Nq0 == 0)           q0 = omega_m/2.  - omega_lambda;
  }
  
  %Set default values
  if  (omega_k      == NULL) omega_k      = 0.;       %Default is flat space
  if  (omega_lambda == NULL) omega_lambda = 0.7;
  if  (omega_m      == NULL) omega_m      = 1 - omega_lambda;
  if  (q0           == NULL) q0           = (1 - (omega_k + 3*omega_lambda))*0.5;
  
  return struct{
    omega_k      = omega_k,
    omega_lambda = omega_lambda,
    omega_m      = omega_m,
    q0           = q0
  };
}

%
% Code from src/slang/math/cov_matrix.sl
%
define cov_matrix()
%!%+
%\function{cov_matrix}
%\synopsis{computes a covariance matrix}
%\usage{COV = cov_matrix(Struct_Type s);
%\altusage{COV = cov_matrix(Double_Type X[][]);}
%\altusage{COV = cov_matrix(Double_Type x0[], x1[], ...);}
%}
%\description
%    \code{COV[i,j] =} cov(x_i, x_j) = < (x_i - <x_i>) * (x_j - <x_j>) >
%!%-
{
  variable X;
  variable i, j, j1, j2, n, p, mu, cov;
  switch(_NARGS)
  { case 0: help(_function_name()); return; }
  { case 1: X = ();
    if(typeof(X) == Struct_Type)
    { variable tab = X;
      variable fieldnames = get_struct_field_names(tab);
      variable n_comp = length(fieldnames);
      variable n_data = length(get_struct_field(tab, fieldnames[0]));
      X = Double_Type[n_data, n_comp];
      variable mom = Struct_Type[n_comp];
      variable comp;
      _for comp (0, n_comp-1, 1)
        X[*, comp] = get_struct_field(tab, fieldnames[comp]);
    }
  }
  { % else:
    variable x0 = ();
    X = Double_Type[length(x0), _NARGS];
    X[*, _NARGS-1] = x0;
    _for i (_NARGS-2, 0, -1)
      X[*, i] = ();
  }

  variable dims = array_shape(X);

  if(length(dims)==2)  % X[i,j] = jth component of ith vector
  {
    n = dims[0];  % = length(X[*,0]);
    p = dims[1];

    mu = Double_Type[p];
    _for i (0, n-1, 1)
      _for j (0, p-1, 1)
        mu[j] += X[i,j]/n;

    cov = Double_Type[p,p];
    _for i (0, n-1, 1)
      _for j1 (0, p-1, 1)
        _for j2 (0, p-1, 1)
	  cov[j1, j2] += (X[i,j1]-mu[j1])*(X[i,j2]-mu[j2])/n;
    return cov;
  }
  vmessage("error (%s): %d-dimensional arrays are not supported", _function_name(), length(dims));
}

%
% Code from src/slang/math/covariance_correlation_matrix.sl
%
define covariance_correlation_matrix() %{{{
%!%+
%\function{covariance_correlation_matrix}
%\synopsis{Compute the covariance and correlation matrices of a set of variables}
%\usage{(Double_Types cov_mat[n,n], cor_mat[n,n]) = covariance_correlation_matrix(Double_Type a[n,N])}
%\description
%    This function estimates the covariance and correlation matrices of a set of
%    random variables. Consider a sample 'a' of 'N' column vectors of length 'n':
%    a = Double_Type[n,N]. Each of the 'n' rows of 'a' may represent the realization
%    of a random variable. The 'n' times 'n' covariance matrix 'cov_mat' between
%    these 'n' variables is
%      cov_mat[i,j] = 1/(N-1) * sum( (a[i,*]-mean(a[i,*]))*(a[j,*]-mean(a[j,*])).
%    The 'n' times 'n' correlation matrix 'cor_mat' is
%      cor_mat[i,j] = cov_mat[i,j]/sigma_i/sigma_j
%    where 'sigma_i' is the square root of the variance of 'a_i':
%      sigma_i = sqrt( 1/(N-1) * sum( (a[i,*]-mean(a[i,*]))^2 ) ).
%\notes
%    The Cholesky decomposition of the covariance matrix can be used to generate
%    correlated variables that obey a given covariance matrix. See the example
%    section for details.
%\example
%    % From three uncorrelated normal variables 'g', generate three correlated normal
%    % variables 'x' that obey the covariance matrix 'm':
%
%    g = Double_Type[3,100000];
%    g[0,*] = grand(100000); g[1,*] = grand(100000); g[2,*] = grand(100000);
%    (cov_mat, cor_mat) = covariance_correlation_matrix(g);
%    print(cov_mat); % the covariance matrix of 'g' shows that there are no correlations
%    print(cor_mat);
%
%    m = Double_Type[3,3]; m[0,*] = [ 9,-3,-6]; m[1,*] = [-3,10, 5]; m[2,*] = [-6, 5, 6];
%    cd = cholesky_decomposition(m);
%    x = cd#g; % matrix product of 'cd' and 'g'
%    (cov_mat, cor_mat) = covariance_correlation_matrix(x);
%    print(cov_mat); % the covariance matrix of 'x' is indeed (almost) 'm'
%    connect_points(0); plot(x[0,*],x[2,*]);
%\seealso{cholesky_decomposition}
%!%-
{
  if(_NARGS!=1)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable a = ();
  a *= 1.; % cast to Double_Type
  variable info = struct{ dims, ndims, type };
  (info.dims, info.ndims, info.type) = array_info(a);
  if(info.ndims!=2 || info.dims[1]<2 || info.type!=Double_Type)
    throw UsageError, sprintf("Usage error in '%s': The argument has to be a Double_Type[n,N] with N>1.", _function_name());
  variable i, j;
  variable len = info.dims[0];
  variable means = Double_Type[len], sigmassqr = Double_Type[len], sigmas = Double_Type[len];
  _for i(0, len-1, 1)
  {
    means[i] = mean(a[i,*]);
    sigmassqr[i] = sum((a[i,*]-means[i])^2);
    sigmas[i] = sqrt(sigmassqr[i]);
  }
  variable cov_mat = Double_Type[len,len]; % covariance matrix
  variable cor_mat = Double_Type[len,len]; % correlation matrix
  _for i(0, len-1, 1)
  {
    _for j(0, i, 1)
    {
      if(i==j)
	cov_mat[i,j] = sigmassqr[i];
      else
	cov_mat[i,j] = sum((a[i,*]-means[i])*(a[j,*]-means[j]));
      cov_mat[j,i] = cov_mat[i,j];
      cor_mat[i,j] = cov_mat[i,j]/sigmas[i]/sigmas[j];
      cor_mat[j,i] = cor_mat[i,j];
    }
  }
  return (1./(info.dims[1]-1.)*cov_mat, cor_mat);
}%}}}

%
% Code from src/slang/structs/create_struct_field.sl
%
define create_struct_field()
%!%+    
%\function{create_struct_field}
%\synopsis{create and set the value associated with a structure field}
%\usage{Struct_Type create_struct_field(s, field_name, field_value);}
%\qualifiers{
%    \qualifier{skip}{do not update an existing field}
%}
%\description
%    Does the same as the `set_struct_field' function except
%    that the given field is created first if it does not
%    exist.
%\seealso{set_struct_field, struct_combine}
%!%-
{
  variable s, field, value;
  switch (_NARGS)
  { case 3: (s, field, value) = (); }
  { help(_function_name); }

  if (qualifier_exists("skip") and struct_field_exists(s, field)) { return s; }
  s = struct_combine(s, field);
  set_struct_field(s, field, value);
  return s;
}

%
% Code from src/data/timing/cross_power_density.sl
%
private define cross_power_density(rate1, rate2, numseg, dimseg){
%!%+
%\function{cross_power_density}
%\synopsis{Timing Tools: Cross Power Density (aka. Cross-Spectrum)}
%\usage{(cpd, errcpd) = cross_power_density(rate1, rate2, numseg, dimseg)}
%\description
%    Calculates the CPD, averaged over all segments, as
%    Conj(DFT(rate1))*DFT(rate2)
%    using the dorDFT function. The return values are of Complex_Type.
%    The error is the standard error on the mean of the averaged CPD.
%\seealso{dorDFT}
%!%-
  variable startbin = 0;
  variable endbin = dimseg-1;
  
  variable cpd = Complex_Type[dimseg/2];  % = 0 + 0i
  variable cpds = Complex_Type[numseg, dimseg/2];  % Record of CPDs of all segments for error calculation

  variable seg=0;
  loop(numseg)
  {
    variable cpd_of_segment = Conj(dorDFT(rate1[[startbin:endbin]])) * dorDFT(rate2[[startbin:endbin]]);

    cpd += cpd_of_segment;
    startbin += dimseg;
    endbin += dimseg;

    % Divide by dimseg because dorDFT has normalization sqrt(N) and is
    % applied twice: once for DFT(rate1), once for DFT(rate2).
    cpds[seg,*] = cpd_of_segment / dimseg;
    seg++;
  }
  
  cpd /= (numseg * dimseg);  % Average over all segments
  
  % ====== Error calculation ========================================
  variable errcpd = Complex_Type[dimseg/2];
  variable freq;
  _for freq (0, dimseg/2-1, 1){
    errcpd[freq] = moment(Real(cpds[*,freq])).sdom + moment(Imag(cpds[*,freq])).sdom*1i;
  }

  return cpd, errcpd;
}

%
% Code from src/fitting/cstat_unc.sl
%
define kaastra_cstat_goodness() 
%!%+
%\function{kaastra_cstat_goodness}
%\synopsis{computes the theoretical Cash-statistics for a Poisson process}
%\usage{(Array_Type ce, Array_Type cn) = kaastra_cstat_goodness(Array_Type mu);}
%\description
% Given an array of predicted counts, mu, this function uses the
% expressions given by Kaastra (2017, A&A 605, A51) to compute the expected
% value of the contribution to the C-statistic for this bin, ce, and its
% variance, cn.
%
% This can be used to evaluate the expected value of the C-statistic of
% a fit (see function cstat_goodness()), or to show the expected model
% contribution in a plot of the cash statistics residuals.
%
% The function uses Kaastra's approximation to the exact formulae (his
% Eqs. 8-22), which are better than a few times 1e-4. Use the exact
% qualifier if you cannot live with this (it is unclear why you would
% want to do this).
% 
%\qualifiers{
%\qualifier{exact}{use Kaastra's exact equations 4-6 (slow; use only
%                    for testing, not necessary for practical work)}
%}
% 
%\seealso{cstat_goodness}
%!%-
{
    variable mu=();

    variable ce, cn, sn;
    variable k,pmu,argu,ndx;

    if (qualifier_exists("exact")) {
        %
        % Implementation of Kaastra's exact equations
        % Slow, but - well - exact...
        %

        
        % stop the loop once the relative change of ce and sn is
        % less than relmin
        % (typically still needs up to 800 passages through the loop...
        variable relmin=qualifier("relmin",1e-4);
        % k=0
        pmu=exp(-mu);
        argu=mu;
        ce=pmu*argu;
        sn=ce*argu;

        variable marel=1.;
        k=1;

        variable susu={};

        while (marel>relmin) {
            pmu*=mu/k;
            argu=mu+k*(log(k/mu)-1.);
            variable f1=pmu*argu;
            variable f2=f1*argu;
            ce+=f1;
            sn+=f2;
            marel=max([max(f1/ce),max(f2/sn)]);
            k++;
        }

        ce*=2.;
        sn*=4.;
        cn=sn-ce*ce;

        return(ce,cn);
    }

    %
    % Kaastra approximations (Eq. 8--22)
    %

    ce=Double_Type[length(mu)];
    cn=Double_Type[length(mu)];
    
    ndx=where(mu==0);
    if (length(ndx)>0) {
        ce[ndx]=0.;
    }
    ndx=where(0<mu<=0.5);
    if (length(ndx)>0) {
        ce[ndx]=((-0.25*mu[ndx]+1.38)*mu[ndx]-2*log(mu[ndx]))*mu[ndx];
    }
    ndx=where(0.5<mu<=2.);
    if (length(ndx)>0) {
        ce[ndx]=((((-0.00335*mu[ndx]+0.04259)*mu[ndx]-0.27331)*mu[ndx]+1.381)*mu[ndx]-2*log(mu[ndx]))*mu[ndx];
    }
    ndx=where(2.<mu<=5.);
    if (length(ndx)>0) {
        ce[ndx]=1.019275+0.1345*mu[ndx]^(0.461-0.9*log(mu[ndx]));
    }
    ndx=where(5.<=mu<10.);
    if (length(ndx)>0) {
        ce[ndx]=1.00624+0.604*mu[ndx]^(-1.68);
    }
    ndx=where(10.<mu);
    if (length(ndx)>0) {
        ce[ndx]=(0.226/mu[ndx]+0.1649)/mu[ndx]+1;
    }

    ndx=where(0.<=mu<=0.1);
    if (length(ndx)>0) {
        variable j;
        pmu=exp(-mu[ndx]);
        sn=pmu*mu[ndx]*mu[ndx];
        _for k(1,4) {
            pmu*=mu[ndx]/k;
            argu=mu[ndx]+k*(log(k/mu[ndx])-1.);
            sn+=pmu*argu*argu;
        }
        cn[ndx]=4.*sn-ce[ndx]^2.;
    }
    ndx=where(0.1<mu<=0.2);
    if (length(ndx)>0) {
        cn[ndx]=(((-262.*mu[ndx]+195.)*mu[ndx]-51.24)*mu[ndx]+4.34)*mu[ndx]+0.77005;
    }
    ndx=where(0.2<mu<=0.3);
    if (length(ndx)>0) {
        cn[ndx]=(4.23*mu[ndx]-2.8254)*mu[ndx]+1.12522;
    }
    ndx=where(0.3<mu<=0.5);
    if (length(ndx)>0) {
        cn[ndx]=((-3.7*mu[ndx]+7.328)*mu[ndx]-3.6926)*mu[ndx]+1.20641;
    }
    ndx=where(0.5<mu<1.);
    if (length(ndx)>0) {
        cn[ndx]=(((1.28*mu[ndx]-5.191)*mu[ndx]+7.666)*mu[ndx]-3.5446)*mu[ndx]+1.15431;
    }
    ndx=where(1.<mu<=2.);
    if (length(ndx)>0) {
        cn[ndx]=(((0.1125*mu[ndx]-0.641)*mu[ndx]+0.859)*mu[ndx]+1.0914)*mu[ndx]-0.05748;
    }
    ndx=where(2.<mu<=3.);
    if (length(ndx)>0) {
        cn[ndx]=((0.089*mu[ndx]-0.872)*mu[ndx]+2.8422)*mu[ndx]-0.67539;
    }
    ndx=where(3.<mu<=5.);
    if (length(ndx)>0) {
        cn[ndx]=2.12336+0.012202*mu[ndx]^(5.717-2.6*log(mu[ndx]));
    }
    ndx=where(5.<mu<=10.);
    if (length(ndx)>0) {
        cn[ndx]=2.05159+0.331*mu[ndx]^(1.343-log(mu[ndx]));
    }

    ndx=where(mu>10.);
    if (length(ndx)>0) {
        variable mu1=1./mu[ndx];
        cn[ndx]=((12.*mu1+0.79)*mu1+0.6747)*mu1+2.;
    }

    return (ce,cn);
}

define cstat_goodness() 
%!%+
%\function{cstat_goodness}
%\synopsis{computes the theoretical Cash-statistics and its
%uncertainty for the current data and model}
%\usage{Struct_Type ctheo = cstat_goodness();}
%\description
% This function computes the theoretical Cash statistics for the
% currently set data, using the expressions given by Kaastra
% (2017, A&A 605, A51). These values are returned in the tags
% cstat_theory and c_variance. The tag n_bins contains the number
% of bins entering these values, it should be the same as that
% obtained with eval_stat_counts().
%
% The values can be used to compute the goodness of the fit by
% comparing its best fit statistics (obtained with eval_stat_counts())
% with cstat_theory and the variance. The "f-sigma" probability for
% the best fit cash statistic is the range cstat_theory - f *c_variance
% to cstat_theory + f * c_variance.
%
%\qualifiers{
%\qualifier{quiet}{if set, the function will not complain if the fit
%                    statistics is not set to Cash or if there are bins without counts.}
%\qualifier{data}{if set, uses the measured counts in the computation
%                   (the default is the expected counts in the bin)}
%}
% 
%\seealso{kaastra_cstat_goodness,eval_stat_counts,set_fit_statistic}
%
%!%-
{
    variable ids=all_data(1); % all data sets with noticed bins
    if (typeof(ids)==Null_Type) {
        throw UsageError,sprintf("%s: No data sets with noticed bins have been defined.",_function_name());
    }

    variable st=strsplit(get_fit_statistic(),";");
    if (st[0]!="cash") {
	if(not qualifier_exists("quiet")) {
	    ()=fprintf(stderr,"%s: Warning: fit statistic is %s and not cash!\n",_function_name(),st[0]);
	}
    }
    
    variable ii;
    variable csum=0.;
    variable cvar=0.;
    variable nnoti=0;


    _for ii(0,length(ids)-1,1) {
	variable mu;
	if (qualifier_exists("data")) {
	    mu=(get_data_counts(ids[ii])).value;
	} else {
	    mu=(get_model_counts(ids[ii])).value;
	}
	
        variable noti=(get_data_info(ids[ii])).notice;
        variable cn,sn;

	mu=mu[where(noti==1)]; % noticed data only

	variable ndx=where(mu==0);
	if (length(ndx)!=0) {
	    if(not qualifier_exists("quiet")) {
		()=fprintf(stderr,"%s: Warning: Dataset %i contains valid bins with 0 counts.\n",_function_name(),ids[ii]);
	    }
	}

	
        % we're guaranteed that there's at least one noticed bin here
        (cn,sn)=kaastra_cstat_goodness(mu);
	
        csum+=sum(cn);
        cvar+=sum(sn);
	nnoti+=int(sum(length(where(noti))));
    }

    return struct {cstat_theory=csum,c_variance=sqrt(cvar),num_bins=nnoti};
}

%
% comparison plot for  Kaastra

% variable mumin=1e-7;
% variable mumax=500;
% variable npts=1000;

% variable mu=exp(log(mumin)+[0:(npts)-1]*log(mumax/mumin)/npts);

% variable ce,cn;
% (ce,cn)=kaastra_cstat_goodness(mu;exact,relmin=1e-3);


% variable p=xfig_plot_new(15,10);

% p.world(1e-7,100,0,1.6;xlog);

% p.plot(mu,ce;color="red",depth=100);
% p.plot(mu,sqrt(cn);color="blue",depth=150);

% variable ce1,cn1;
% (ce1,cn1)=kaastra_cstat_goodness(mu);

% p.plot(mu,ce1;color="black",line=1,depth=99);
% p.plot(mu,sqrt(cn1);color="black",line=1,depth=149);

% p.render("cstat.pdf");

%
% Code from src/fitting/fit-functions/cutoffpl2.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%%%%%%%%%%%%%%%%%%%%
define cutoffpl2_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{cutoffpl2 (fit-function)}
%\synopsis{describes a power law with exponential cutoff in wavelength space}
%\description
%    S(l) = \code{norm} * l^(-\code{index}) * exp(-l/\code{cutoff})\n
%    The bin-integrated cutoffpl2 fit-function is computed
%    by the incomplete Gamma-function from the GSL library.
%    (The original cutoffpl didn't converge for low energies.)
%\seealso{cutoffpl, gamma_inc}
%!%-
{
  variable norm = par[0];
  variable index = par[1];
  variable cutoff = par[2];
  return norm * cutoff^(1-index) * ( gsl->gamma_inc(1-index, bin_lo/cutoff) - gsl->gamma_inc(1-index, bin_hi/cutoff) );
}

add_slang_function("cutoffpl2", ["norm", "index", "cutoff"]);


%%%%%%%%%%%%%%%%%%%%%%%%%
define cutoffpl2_defaults(i)
%%%%%%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
  { case 0: return ( 1, 0,  0, 1e10); }
  { case 1: return ( 1, 0, -2,    9); }
  { case 2: return (15, 0,  1,   50); }
}

set_param_default_hook("cutoffpl2", "cutoffpl2_defaults");


  % test code
% variable E_lo, E_hi; (E_lo, E_hi) = log_grid(0.001, 1e3, 10000);
% variable pars = [1, 1, 250];
% fit_fun("cutoffpl(1)");
% set_par(1, pars[0], 0, pars[0], pars[0]);
% set_par(2, pars[1], 0, pars[1], pars[1]);
% set_par(3, pars[2], 0, pars[2], pars[2]);
% variable y1 = eval_fun_keV(E_lo, E_hi);
% variable y2 = eval_fun2("cutoffpl2", E_lo, E_hi, pars);
% multiplot([1,1]); xlog; ylog;
% yrange;
%  hplot(E_lo, E_hi, y1);
% ohplot(E_lo, E_hi, y2);
% yrange(0.5,);
% hplot(E_lo, E_hi, y1/y2);

#endif

%
% Code from src/data/data_response.sl
%
private define responseEvalSetup (id)
{
  % to restore user settings
  variable data = struct {
    includes = all_data(1),
    par = get_params("delta(1).*"), % in case this is in use
  };

  variable info = get_data_info(id);
  rebin_data(id, 0); % need all 
  variable fakeId = define_counts(get_data(id));
  rebin_data(id, info.rebin);
  notice_list(id, info.notice_list);
  assign_arf(info.arfs, fakeId);
  assign_rmf(info.rmfs, fakeId);

  assign_model(fakeId, "delta(1)");
  set_par("delta(1).*"; min=0, max=0);
  exclude(all_data());

  variable rmfId = get_data_info(fakeId).rmfs[0]; % get the first rmf, this defines the data grid
  variable grid = get_rmf_data_grid(rmfId); % might be NULL if default RMF is used
  if (NULL == grid) grid = get_data(id); % use the data grid instead

  return fakeId, grid, data;
}

private define responseEval (fakeId, grid, channel, fit)
% expects that everything is set up!
{
  ifnot (0 <= channel < length(grid.bin_lo))
    return Double_Type[length(grid.bin_lo)];

  if (NULL == fit)
    fit = open_fit();

  if (qualifier_exists("energy")) {
    channel = length(grid.bin_lo)-channel-1;
    % delta on nominal bin wavelength
    () = fit.eval_statistic([1, (grid.bin_hi[channel]+grid.bin_lo[channel])*.5]);
    return reverse(get_model(fakeId).value);
  } else {
    % delta on nominal bin wavelength
    () = fit.eval_statistic([1, (grid.bin_hi[channel]+grid.bin_lo[channel])*.5]);
    return get_model(fakeId).value;
  }
}

private define responseEvalCleanup (fakeId, data)
{
  assign_model(fakeId, NULL);
  exclude(all_data);
  include(data.includes);
  if (NULL != data.par && data.par[0] != NULL)
    set_params(data.par);
  delete_data(fakeId);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_dataset_response ()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_dataset_response}
%\usage{Response = get_dataset_response(id, [channel]);}
%\synopsis{Compute response of dataset associated to 'id' to delta peak}
%\qualifiers{
%\qualifier{energy}{If given, return response on energy grid (default: wavelength)}
%}
%\description
%    This function evaluates a delta peak function set at each nominal
%    grid value folded through the response of dataset \code{id}. If a
%    channel is given, only the response for the corresponding channel
%    is calculated.
%
%    The returned value is either an array of arrays for each channel in the
%    detector, or the array of the response in channel \code{channel}.
%    The delta function is placed at the nominal energy of the channel. If
%    \code{channel} is outside of the detector channels, a zero array is
%    returned.
%
%    Per default the response is returned as 'seen' by ISIS, that is, on
%    a wavelength grid. Use the 'energy' qualifier to revers the
%    response.
%
%\seealso{get_rmf_data_grid, assign_rmf, assign_rsp}
%!%-
{
  variable id, channel = NULL;
  switch (_NARGS)
  { case 1: id = (); }
  { case 2: (id, channel) = (); }
  { help(_function_name); return; }

  % check if dataset exists
  if (all(id != all_data()))
    throw UsageError, sprintf("Dataset %d not defined", id);

  variable fakeId, grid, data;
  (fakeId, grid, data) = responseEvalSetup(id);

  variable response, fit = NULL;
  include(fakeId);

  if (NULL == channel) {
    response = Array_Type[length(grid.bin_lo)];
    fit = open_fit();
    _for channel (0, length(grid.bin_lo)-1)
      response[channel] = responseEval(fakeId, grid, channel, fit;; __qualifiers);
  } else {
    response = responseEval(fakeId, grid, channel, fit;; __qualifiers);
  }

  if (NULL != fit)
    fit.close(); % done

  responseEvalCleanup(fakeId, data);

  return response;
}

%
% Code from src/slang/date/dateOfMJD.sl
%
define dateOfMJD()
%!%+
%\function{dateOfMJD}
%\synopsis{calculates the date from its MJD value}
%\usage{Struct_Type date = dateOfMJD(MJD);}
%\description
%    The return value is a structure of the following form:\n
%       \code{date = struct { year, month, day, hour, minute, second }
%    and always in the Gregorian Calendar. See DateOfJD if you want more
%    functionality. This function is array safe.}
%\seealso{DateOfJD, MJDofDate}
%!%-
{
  variable MJD;
  switch(_NARGS)
  { case 1: MJD = (); }
  { help(_function_name()); return; }

  return DateOfJD(mjd2jd(MJD));
}

define DateOfMJD()
%!%+
%\function{DateOfMJD}
%\synopsis{calculates the date from its MJD value}
%\usage{Struct_Type date = dateOfMJD(MJD);}
%\description
%    The return value is a structure of the following form:\n
%       \code{date = struct { year, month, day, hour, minute, second }
%    and always in the Gregorian Calendar. See DateOfJD if you want more
%    functionality. This function is array safe.}
%\seealso{DateOfJD, MJDofDate}
%!%-
{
  variable MJD;
  switch(_NARGS)
  { case 1: MJD = (); }
  { help(_function_name()); return; }

  return DateOfJD(mjd2jd(MJD));
}

%
% Code from src/slang/date/dayOfWeek.sl
%
define dayOfWeek()
%!%+
%\function{dayOfWeek}
%\synopsis{returns the day of the week of the given date}
%\usage{Integer_Type dayOfWeek(Integer_Type year, month, day);}
%\description
%    This function returns the day of the week from 1 (Monday) to
%    7 (Sunday) of the given date in the Gregorian (default) or the
%    Julian calendar.
%
%\qualifiers{
%\qualifier{julian_calendar}{: the date is in the Julian calendar
%                              (default: Gregorian calendar)}
%}
%\example
%    dayOfWeek(1837, 9, 9);
%!%-
{
  variable y, m, d;
  switch(_NARGS)
    { case 3: (y,m,d) = (); }
    { help(_function_name()); return;}

  % sanity checks
  if (typeof(y) != Integer_Type) { vmessage("error (%s): year must be an integer", _function_name); return; }
  if (typeof(m) != Integer_Type) { vmessage("error (%s): month must be an integer", _function_name); return; }
  if (typeof(d) != Integer_Type) { vmessage("error (%s): day must be an integer", _function_name); return; }

  %jw do NOT check on month or day validity, since
  %astronomers sometimes use the 0th or the -1st of the
  %day and similar weird things...
  
  % return day of week
  % NB calculating the JD for 13h avoids potential
  % round off
  if (qualifier_exists("julian_calendar")) {
      return 1 + (int(JDofDate(y,m,d,13,0,0;julian_calendar)) mod 7);
  }
  
  return 1 + (int(JDofDate(y,m,d,13,0,0)) mod 7);
}

%
% Code from src/slang/date/daysPerMonth.sl
%
%%%%%%%%%%%%%%%%%%%
define daysPerMonth()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{daysPerMonth}
%\synopsis{returns the number of days of a month}
%\usage{Char_Type daysPerMonth(m[, y])}
%\description
%    \code{m} = 1, 2, ... 12 specifies the month Jan, Feb, ..., Dec.\n
%    If \code{y} is not specified, a non-leapyear is assumed.
%\seealso{leapyear}
%!%-
{
  variable m, y, leapdays=0;
  switch(_NARGS)
  { case 1: m = (); }
  { case 2: (m, y) = (); leapdays = leapyear(y); }
  { help(_function_name()); return; }

  switch(m)
  { case  1: return 31; }
  { case  2: return 28 + leapdays; }
  { case  3: return 31; }
  { case  4: return 30; }
  { case  5: return 31; }
  { case  6: return 30; }
  { case  7: return 31; }
  { case  8: return 31; }
  { case  9: return 30; }
  { case 10: return 31; }
  { case 11: return 30; }
  { case 12: return 31; }
  return 0;
}

%
% Code from src/data/timing/dcf.sl
%
%%%%%%%%%%%%%
define dcf ()
%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{dcf}
%\synopsis{compute the Edelson & Krolik discrete correlation function and returns a structure with the correlation}
%\usage{dcf(time_a, val_a, tima_be, val_b);}
%\qualifiers{
%\qualifier{err_a [=0]}{:  array containing the measurement errors of
%             val_a. If given, this is taken into
%             account by using eq. (3) of Edelson & Krolik.
%             NOTE: There are lots of problems with interpreting the
%             DCF computed this way instead of using Edelson & Krolik,
%             eq. (2), and it is NOT recommended to give erra and errb
%             (see, e.g., White & Peterson, 1994, PASP, 106, 879)}
%\qualifier{err_b [=0]}{:	as err_a except for val_b}
%\qualifier{minlag [=mininum time difference between time_a and time_b]}{:	minumum lag to consider}
%\qualifier{maxlag [=maximum time difference between time_a and time_b]}{:	maximum lag to consider}
%\qualifier{numf [=int(min([length(val_a) ,length(val_b) ])/10)]}{:	number of lag bins for which to compute the DCF}
%\qualifier{minpt [=10]}{:	minimum number of data points for a DCF value to be}
%}
%\description
%    This is the imported version of the IDL subroutine dcf.pro \n
%    It will compute the Edelson & Krolik discrete correlation function. \n
%!%-
{
  variable dcf_struct= struct{lag,cor,sdev,numpt};
  variable time_a, val_a, time_b, val_b;
  switch(_NARGS)
  { case  4: (time_a, val_a, time_b, val_b) = (); }
  { help(_function_name()); return; }

  variable err_a = mean(qualifier("err_a",0));
  variable err_b = mean(qualifier("err_b",0));
  variable minpt = qualifier("minpt",10);
  if (minpt < 0) {
    print("minpt must be positive");
    return -1;
  }

  if(length(val_a) != length(time_a) || length(val_b) != length(time_b) ) {
    print("arraysize of value and time is different");
    return -1;
  }

  variable numf = qualifier("numf",int(min([length(val_a) ,length(val_b) ])/10));

  variable a = moment(val_a);
  variable b = moment(val_b);

  
  variable udcf = Double_Type [length(val_b), length(val_a)];
  variable dt = Double_Type [length(time_b), length(time_a)];
  variable ii;
  _for ii (0, length(val_b) - 1, 1) {
    udcf[ii,*] = (val_a - a.ave) * (val_b[ii] - b.ave);
    dt[ii,*] = time_a - time_b[ii];
  }
  udcf /= sqrt( (sqr(a.sdev) - sqr(err_a)) * (sqr(b.sdev) - sqr(err_b)));

  udcf = _reshape(udcf,length(val_b)*length(val_a));
  dt = _reshape(dt,length(val_b)*length(val_a));

  variable minlag = qualifier("minlag",min(dt));
  variable maxlag = qualifier("maxlag",max(dt));

  dcf_struct.lag = minlag + (maxlag - minlag) * [0:numf] / double(numf);
  
  variable index;
  index = array_sort(dt);
  udcf = udcf[index];
  dt = dt[index];
  dcf_struct.cor = Double_Type[numf];
  dcf_struct.cor[*] = _NaN;
  dcf_struct.sdev = Double_Type[numf];
  dcf_struct.numpt = Integer_Type[numf];
  variable npt1 = length(dt) - 1;

% go to the first lag bin
  variable start = 0;
  while(start < npt1 && dt[start] < dcf_struct.lag[0]) {
    start++;
  }
  variable stop = start;
  variable mom;

  _for ii (0, length(dcf_struct.lag)-2, 1) {
    while(stop < npt1 && dt[stop] < dcf_struct.lag[ii+1]) {
    stop++;
    }
    if(start < stop) {
      if(stop == npt1 && dt[stop] < dcf_struct.lag[ii+1]) {
        mom = moment(udcf[[start:stop]]);
        dcf_struct.cor[ii] = mom.ave;
        dcf_struct.sdev[ii] = mom.sdev;
        dcf_struct.numpt[ii] = stop-start;
      } else {
        mom = moment(udcf[[start:stop-1]]);
        dcf_struct.cor[ii] = mom.ave;
        dcf_struct.sdev[ii] = mom.sdev;
        dcf_struct.numpt[ii] = stop-start-1;
      }
      start=stop;
    }
  }
  dcf_struct.lag=(dcf_struct.lag[[0:numf-1]]+dcf_struct.lag[[1:numf]])/2.;
  variable ndx = where(dcf_struct.numpt < minpt);
  if(length(ndx) > 0) {
    dcf_struct.cor[ndx] = _NaN;
    dcf_struct.sdev[ndx] = _NaN;
  }
 
  return dcf_struct;
}

%
% Code from src/fitting/fit_2d/define_counts_2d.sl
%
define define_counts_2d()
%!%+
%\function{define_counts_2d}
%\synopsis{defines a pseudo-spectrum from two-dimensional data}
%\usage{Integer_Type define_counts_2d(value[, err][, X, Y])}
%\description
%    \code{value} and \code{err} are 2d arrays of Double_Type.
%    ISIS usually deals with 1d spectra. For fitting 2d data, \code{define_counts_2d}
%    can be used to define a pseudo 1d spectrum by reshaping arrays,
%    to which ISIS' internal fit routines can be applied.\n
%    User defined fit-functions should not use the bin_lo, bin_hi arguments,
%    but the actual data-grid which can be set / obtained with \code{set}/\code{get_2d_data_grid},
%    if the user doesn't prefer to take care of the data grid on his own.
%    If the optional 1d Double_Type arrays \code{X} and \code{Y} are specified,
%    \code{set_2d_data_grid} is already called by \code{define_counts_2d}.
%\seealso{gauss_2d}
%!%-
{
  variable value, err=NULL, X=NULL, Y;
  switch(_NARGS)
  { case 1:  value             = (); }
  { case 2: (value, err)       = (); }
  { case 3: (value,      X, Y) = (); }
  { case 4: (value, err, X, Y) = (); }
  { help(_function_name()); return; }

  if(err==NULL)  err = 0*value+1;
  if(X!=NULL)  set_2d_data_grid(X, Y);

  variable value_1d = @value;
  variable err_1d = @err;
  reshape(value_1d, [length(value)]);
  reshape(err_1d, [length(err)]);

  variable bin_lo = _A([1:length(value)]);
  return define_counts(bin_lo, make_hi_grid(bin_lo), value_1d, err_1d;;__qualifiers());
}

%
% Code from src/slang/color/defined-cmaps.sl
%

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL
% colormaps given at https://www.kennethmoreland.com/color-advice/


png_add_colormap("smooth-cool-warm", [
0x3b4cc0,0x3c4ec2,0x3d50c3,0x3e51c5,0x3f53c7,0x4055c8,
0x4257ca,0x4358cb,0x445acc,0x455cce,0x465ecf,0x485fd1,
0x4961d2,0x4a63d4,0x4b64d5,0x4c66d6,0x4e68d8,0x4f6ad9,
0x506bda,0x516ddb,0x536fdd,0x5470de,0x5572df,0x5674e0,
0x5875e2,0x5977e3,0x5a78e4,0x5b7ae5,0x5d7ce6,0x5e7de7,
0x5f7fe8,0x6181e9,0x6282ea,0x6384eb,0x6585ec,0x6687ed,
0x6788ee,0x698aef,0x6a8cf0,0x6b8df0,0x6d8ff1,0x6e90f2,
0x6f92f3,0x7193f4,0x7295f4,0x7396f5,0x7598f6,0x7699f6,
0x779af7,0x799cf8,0x7a9df8,0x7b9ff9,0x7da0f9,0x7ea2fa,
0x80a3fa,0x81a4fb,0x82a6fb,0x84a7fc,0x85a8fc,0x86aafc,
0x88abfd,0x89acfd,0x8baefd,0x8caffe,0x8db0fe,0x8fb1fe,
0x90b2fe,0x92b4fe,0x93b5ff,0x94b6ff,0x96b7ff,0x97b8ff,
0x99baff,0x9abbff,0x9bbcff,0x9dbdff,0x9ebeff,0x9fbfff,
0xa1c0ff,0xa2c1ff,0xa3c2fe,0xa5c3fe,0xa6c4fe,0xa8c5fe,
0xa9c6fe,0xaac7fd,0xacc8fd,0xadc9fd,0xaec9fc,0xb0cafc,
0xb1cbfc,0xb2ccfb,0xb4cdfb,0xb5cefa,0xb6cefa,0xb7cff9,
0xb9d0f9,0xbad1f8,0xbbd1f8,0xbdd2f7,0xbed3f6,0xbfd3f6,
0xc0d4f5,0xc1d4f4,0xc3d5f4,0xc4d6f3,0xc5d6f2,0xc6d7f1,
0xc8d7f1,0xc9d8f0,0xcad8ef,0xcbd8ee,0xccd9ed,0xcdd9ec,
0xcedaeb,0xd0daea,0xd1dae9,0xd2dbe8,0xd3dbe7,0xd4dbe6,
0xd5dbe5,0xd6dce4,0xd7dce3,0xd8dce2,0xd9dce1,0xdadce0,
0xdbddde,0xdcdddd,0xdddcdc,0xdedcdb,0xdfdcd9,0xe1dbd8,
0xe2dad6,0xe3dad5,0xe4d9d3,0xe5d9d2,0xe5d8d1,0xe6d8cf,
0xe7d7ce,0xe8d6cc,0xe9d6cb,0xead5c9,0xebd4c8,0xebd3c6,
0xecd3c5,0xedd2c3,0xeed1c2,0xeed0c0,0xefcfbf,0xefcebd,
0xf0cebb,0xf1cdba,0xf1ccb8,0xf2cbb7,0xf2cab5,0xf3c9b4,
0xf3c8b2,0xf4c7b1,0xf4c6af,0xf4c5ad,0xf5c4ac,0xf5c3aa,
0xf5c1a9,0xf6c0a7,0xf6bfa6,0xf6bea4,0xf6bda2,0xf7bca1,
0xf7ba9f,0xf7b99e,0xf7b89c,0xf7b79b,0xf7b599,0xf7b497,
0xf7b396,0xf7b194,0xf7b093,0xf7af91,0xf7ad90,0xf7ac8e,
0xf7ab8c,0xf7a98b,0xf7a889,0xf7a688,0xf6a586,0xf6a385,
0xf6a283,0xf6a081,0xf59f80,0xf59d7e,0xf59c7d,0xf49a7b,
0xf4997a,0xf49778,0xf39577,0xf39475,0xf29274,0xf29072,
0xf18f71,0xf18d6f,0xf08b6e,0xf08a6c,0xef886b,0xee8669,
0xee8568,0xed8366,0xed8165,0xec7f63,0xeb7d62,0xea7c60,
0xea7a5f,0xe9785d,0xe8765c,0xe7745b,0xe67259,0xe57058,
0xe56f56,0xe46d55,0xe36b54,0xe26952,0xe16751,0xe0654f,
0xdf634e,0xde614d,0xdd5f4b,0xdc5d4a,0xdb5b49,0xda5947,
0xd85646,0xd75445,0xd65244,0xd55042,0xd44e41,0xd34c40,
0xd1493e,0xd0473d,0xcf453c,0xce433b,0xcc4039,0xcb3e38,
0xca3b37,0xc83936,0xc73635,0xc63434,0xc43132,0xc32e31,
0xc12b30,0xc0282f,0xbf252e,0xbd222d,0xbc1e2c,0xba1a2b,
0xb91629,0xb71128,0xb60b27,0xb40426]);

png_add_colormap("bent-cool-warm", [
0x3b4cc0,0x3c4ec1,0x3d4fc2,0x3e50c2,0x3f52c3,0x4053c4,
0x4155c4,0x4256c5,0x4357c5,0x4459c6,0x455ac7,0x465bc7,
0x475dc8,0x485ec8,0x495fc9,0x4a61ca,0x4b62ca,0x4c64cb,
0x4d65cb,0x4e66cc,0x4f68cc,0x5169cd,0x526ace,0x536cce,
0x546dcf,0x556ecf,0x5670d0,0x5771d0,0x5972d1,0x5a74d1,
0x5b75d2,0x5c76d2,0x5d78d3,0x5f79d3,0x607ad4,0x617cd4,
0x627dd5,0x647ed5,0x6580d6,0x6681d6,0x6782d6,0x6983d7,
0x6a85d7,0x6b86d8,0x6d87d8,0x6e89d9,0x6f8ad9,0x718bda,
0x728dda,0x738eda,0x758fdb,0x7691db,0x7792dc,0x7993dc,
0x7a95dc,0x7b96dd,0x7d97dd,0x7e98de,0x809ade,0x819bde,
0x829cdf,0x849edf,0x859fdf,0x87a0e0,0x88a2e0,0x8aa3e1,
0x8ba4e1,0x8ca5e1,0x8ea7e2,0x8fa8e2,0x91a9e2,0x92abe3,
0x94ace3,0x95ade3,0x97afe4,0x98b0e4,0x9ab1e4,0x9bb2e5,
0x9db4e5,0x9fb5e5,0xa0b6e6,0xa2b8e6,0xa3b9e6,0xa5bae6,
0xa6bbe7,0xa8bde7,0xaabee7,0xabbfe8,0xadc1e8,0xaec2e8,
0xb0c3e8,0xb2c4e9,0xb3c6e9,0xb5c7e9,0xb7c8ea,0xb8caea,
0xbacbea,0xbcccea,0xbdcdeb,0xbfcfeb,0xc1d0eb,0xc2d1ec,
0xc4d2ec,0xc6d4ec,0xc8d5ec,0xc9d6ed,0xcbd7ed,0xcdd9ed,
0xcfdaed,0xd0dbee,0xd2dcee,0xd4deee,0xd6dfee,0xd7e0ef,
0xd9e1ef,0xdbe3ef,0xdde4ef,0xdfe5f0,0xe1e6f0,0xe2e8f0,
0xe4e9f0,0xe6eaf1,0xe8ebf1,0xeaedf1,0xeceef1,0xeeeff2,
0xeff0f2,0xf1f2f2,0xf2f1f1,0xf2f0ef,0xf1eeed,0xf1edeb,
0xf1ebe8,0xf1eae6,0xf0e8e4,0xf0e7e2,0xf0e5e0,0xefe4de,
0xefe2db,0xefe1d9,0xeedfd7,0xeeded5,0xeedcd3,0xeddbd1,
0xedd9cf,0xedd8cd,0xecd6cb,0xecd5c9,0xecd3c7,0xebd2c4,
0xebd0c2,0xebcfc0,0xeacdbe,0xeaccbc,0xe9caba,0xe9c9b8,
0xe9c7b6,0xe8c5b4,0xe8c4b3,0xe8c2b1,0xe7c1af,0xe7bfad,
0xe6beab,0xe6bca9,0xe6bba7,0xe5b9a5,0xe5b8a3,0xe4b6a1,
0xe4b59f,0xe4b39e,0xe3b19c,0xe3b09a,0xe2ae98,0xe2ad96,
0xe2ab94,0xe1aa93,0xe1a891,0xe0a78f,0xe0a58d,0xdfa38c,
0xdfa28a,0xdfa088,0xde9f86,0xde9d85,0xdd9c83,0xdd9a81,
0xdc9880,0xdc977e,0xdb957c,0xdb947b,0xda9279,0xda9077,
0xd98f76,0xd98d74,0xd98c72,0xd88a71,0xd8886f,0xd7876e,
0xd7856c,0xd6846b,0xd68269,0xd58068,0xd47f66,0xd47d65,
0xd37b63,0xd37a62,0xd27860,0xd2775f,0xd1755d,0xd1735c,
0xd0725a,0xd07059,0xcf6e58,0xcf6c56,0xce6b55,0xcd6953,
0xcd6752,0xcc6651,0xcc644f,0xcb624e,0xcb604d,0xca5f4b,
0xc95d4a,0xc95b49,0xc85948,0xc85746,0xc75645,0xc65444,
0xc65243,0xc55041,0xc54e40,0xc44c3f,0xc34a3e,0xc3483d,
0xc2463c,0xc1443a,0xc14239,0xc04038,0xc03e37,0xbf3c36,
0xbe3a35,0xbe3734,0xbd3533,0xbc3232,0xbc3031,0xbb2d30,
0xba2b2f,0xba282e,0xb9252d,0xb8222c,0xb81e2b,0xb71b2a,
0xb61629,0xb51228,0xb50c27,0xb40426]);

png_add_colormap("viridis", [
0x440154,0x440256,0x450457,0x450559,0x46075a,0x46085c,
0x460a5d,0x460b5e,0x470d60,0x470e61,0x471063,0x471164,
0x471365,0x481467,0x481668,0x481769,0x48186a,0x481a6c,
0x481b6d,0x481c6e,0x481d6f,0x481f70,0x482071,0x482173,
0x482374,0x482475,0x482576,0x482677,0x482878,0x482979,
0x472a7a,0x472c7a,0x472d7b,0x472e7c,0x472f7d,0x46307e,
0x46327e,0x46337f,0x463480,0x453581,0x453781,0x453882,
0x443983,0x443a83,0x443b84,0x433d84,0x433e85,0x423f85,
0x424086,0x424186,0x414287,0x414487,0x404588,0x404688,
0x3f4788,0x3f4889,0x3e4989,0x3e4a89,0x3e4c8a,0x3d4d8a,
0x3d4e8a,0x3c4f8a,0x3c508b,0x3b518b,0x3b528b,0x3a538b,
0x3a548c,0x39558c,0x39568c,0x38588c,0x38598c,0x375a8c,
0x375b8d,0x365c8d,0x365d8d,0x355e8d,0x355f8d,0x34608d,
0x34618d,0x33628d,0x33638d,0x32648e,0x32658e,0x31668e,
0x31678e,0x31688e,0x30698e,0x306a8e,0x2f6b8e,0x2f6c8e,
0x2e6d8e,0x2e6e8e,0x2e6f8e,0x2d708e,0x2d718e,0x2c718e,
0x2c728e,0x2c738e,0x2b748e,0x2b758e,0x2a768e,0x2a778e,
0x2a788e,0x29798e,0x297a8e,0x297b8e,0x287c8e,0x287d8e,
0x277e8e,0x277f8e,0x27808e,0x26818e,0x26828e,0x26828e,
0x25838e,0x25848e,0x25858e,0x24868e,0x24878e,0x23888e,
0x23898e,0x238a8d,0x228b8d,0x228c8d,0x228d8d,0x218e8d,
0x218f8d,0x21908d,0x21918c,0x20928c,0x20928c,0x20938c,
0x1f948c,0x1f958b,0x1f968b,0x1f978b,0x1f988b,0x1f998a,
0x1f9a8a,0x1e9b8a,0x1e9c89,0x1e9d89,0x1f9e89,0x1f9f88,
0x1fa088,0x1fa188,0x1fa187,0x1fa287,0x20a386,0x20a486,
0x21a585,0x21a685,0x22a785,0x22a884,0x23a983,0x24aa83,
0x25ab82,0x25ac82,0x26ad81,0x27ad81,0x28ae80,0x29af7f,
0x2ab07f,0x2cb17e,0x2db27d,0x2eb37c,0x2fb47c,0x31b57b,
0x32b67a,0x34b679,0x35b779,0x37b878,0x38b977,0x3aba76,
0x3bbb75,0x3dbc74,0x3fbc73,0x40bd72,0x42be71,0x44bf70,
0x46c06f,0x48c16e,0x4ac16d,0x4cc26c,0x4ec36b,0x50c46a,
0x52c569,0x54c568,0x56c667,0x58c765,0x5ac864,0x5cc863,
0x5ec962,0x60ca60,0x63cb5f,0x65cb5e,0x67cc5c,0x69cd5b,
0x6ccd5a,0x6ece58,0x70cf57,0x73d056,0x75d054,0x77d153,
0x7ad151,0x7cd250,0x7fd34e,0x81d34d,0x84d44b,0x86d549,
0x89d548,0x8bd646,0x8ed645,0x90d743,0x93d741,0x95d840,
0x98d83e,0x9bd93c,0x9dd93b,0xa0da39,0xa2da37,0xa5db36,
0xa8db34,0xaadc32,0xaddc30,0xb0dd2f,0xb2dd2d,0xb5de2b,
0xb8de29,0xbade28,0xbddf26,0xc0df25,0xc2df23,0xc5e021,
0xc8e020,0xcae11f,0xcde11d,0xd0e11c,0xd2e21b,0xd5e21a,
0xd8e219,0xdae319,0xdde318,0xdfe318,0xe2e418,0xe5e419,
0xe7e419,0xeae51a,0xece51b,0xefe51c,0xf1e51d,0xf4e61e,
0xf6e620,0xf8e621,0xfbe723,0xfde725]);

png_add_colormap("plasma", [
0x0d0887,0x100788,0x130789,0x16078a,0x19068c,0x1b068d,
0x1d068e,0x20068f,0x220690,0x240691,0x260591,0x280592,
0x2a0593,0x2c0594,0x2e0595,0x2f0596,0x310597,0x330597,
0x350498,0x370499,0x38049a,0x3a049a,0x3c049b,0x3e049c,
0x3f049c,0x41049d,0x43039e,0x44039e,0x46039f,0x48039f,
0x4903a0,0x4b03a1,0x4c02a1,0x4e02a2,0x5002a2,0x5102a3,
0x5302a3,0x5502a4,0x5601a4,0x5801a4,0x5901a5,0x5b01a5,
0x5c01a6,0x5e01a6,0x6001a6,0x6100a7,0x6300a7,0x6400a7,
0x6600a7,0x6700a8,0x6900a8,0x6a00a8,0x6c00a8,0x6e00a8,
0x6f00a8,0x7100a8,0x7201a8,0x7401a8,0x7501a8,0x7701a8,
0x7801a8,0x7a02a8,0x7b02a8,0x7d03a8,0x7e03a8,0x8004a8,
0x8104a7,0x8305a7,0x8405a7,0x8606a6,0x8707a6,0x8808a6,
0x8a09a5,0x8b0aa5,0x8d0ba5,0x8e0ca4,0x8f0da4,0x910ea3,
0x920fa3,0x9410a2,0x9511a1,0x9613a1,0x9814a0,0x99159f,
0x9a169f,0x9c179e,0x9d189d,0x9e199d,0xa01a9c,0xa11b9b,
0xa21d9a,0xa31e9a,0xa51f99,0xa62098,0xa72197,0xa82296,
0xaa2395,0xab2494,0xac2694,0xad2793,0xae2892,0xb02991,
0xb12a90,0xb22b8f,0xb32c8e,0xb42e8d,0xb52f8c,0xb6308b,
0xb7318a,0xb83289,0xba3388,0xbb3488,0xbc3587,0xbd3786,
0xbe3885,0xbf3984,0xc03a83,0xc13b82,0xc23c81,0xc33d80,
0xc43e7f,0xc5407e,0xc6417d,0xc7427c,0xc8437b,0xc9447a,
0xca457a,0xcb4679,0xcc4778,0xcc4977,0xcd4a76,0xce4b75,
0xcf4c74,0xd04d73,0xd14e72,0xd24f71,0xd35171,0xd45270,
0xd5536f,0xd5546e,0xd6556d,0xd7566c,0xd8576b,0xd9586a,
0xda5a6a,0xda5b69,0xdb5c68,0xdc5d67,0xdd5e66,0xde5f65,
0xde6164,0xdf6263,0xe06363,0xe16462,0xe26561,0xe26660,
0xe3685f,0xe4695e,0xe56a5d,0xe56b5d,0xe66c5c,0xe76e5b,
0xe76f5a,0xe87059,0xe97158,0xe97257,0xea7457,0xeb7556,
0xeb7655,0xec7754,0xed7953,0xed7a52,0xee7b51,0xef7c51,
0xef7e50,0xf07f4f,0xf0804e,0xf1814d,0xf1834c,0xf2844b,
0xf3854b,0xf3874a,0xf48849,0xf48948,0xf58b47,0xf58c46,
0xf68d45,0xf68f44,0xf79044,0xf79143,0xf79342,0xf89441,
0xf89540,0xf9973f,0xf9983e,0xf99a3e,0xfa9b3d,0xfa9c3c,
0xfa9e3b,0xfb9f3a,0xfba139,0xfba238,0xfca338,0xfca537,
0xfca636,0xfca835,0xfca934,0xfdab33,0xfdac33,0xfdae32,
0xfdaf31,0xfdb130,0xfdb22f,0xfdb42f,0xfdb52e,0xfeb72d,
0xfeb82c,0xfeba2c,0xfebb2b,0xfebd2a,0xfebe2a,0xfec029,
0xfdc229,0xfdc328,0xfdc527,0xfdc627,0xfdc827,0xfdca26,
0xfdcb26,0xfccd25,0xfcce25,0xfcd025,0xfcd225,0xfbd324,
0xfbd524,0xfbd724,0xfad824,0xfada24,0xf9dc24,0xf9dd25,
0xf8df25,0xf8e125,0xf7e225,0xf7e425,0xf6e626,0xf6e826,
0xf5e926,0xf5eb27,0xf4ed27,0xf3ee27,0xf3f027,0xf2f227,
0xf1f426,0xf1f525,0xf0f724,0xf0f921]);

png_add_colormap("blackbody", [
0x000000,0x030101,0x070201,0x0a0302,0x0d0402,0x100503,
0x120603,0x140704,0x160804,0x180905,0x1a0a05,0x1b0b06,
0x1d0b06,0x1e0c07,0x200d08,0x210e08,0x220f09,0x240f09,
0x25100a,0x26100a,0x28110b,0x29110b,0x2b120c,0x2c120c,
0x2e120d,0x2f130d,0x31130e,0x32130e,0x34140f,0x36140f,
0x37140f,0x391510,0x3a1510,0x3c1510,0x3e1611,0x3f1611,
0x411611,0x421712,0x441712,0x461712,0x471813,0x491813,
0x4b1813,0x4c1914,0x4e1914,0x501914,0x511914,0x531a15,
0x551a15,0x561a15,0x581a15,0x5a1b16,0x5b1b16,0x5d1b16,
0x5f1b16,0x611c17,0x621c17,0x641c17,0x661c17,0x681d18,
0x691d18,0x6b1d18,0x6d1d18,0x6f1d19,0x701e19,0x721e19,
0x741e19,0x761e1a,0x771e1a,0x791f1a,0x7b1f1a,0x7d1f1b,
0x7f1f1b,0x801f1b,0x821f1b,0x84201c,0x86201c,0x88201c,
0x89201c,0x8b201d,0x8d201d,0x8f201d,0x91211d,0x93211e,
0x94211e,0x96211e,0x98211f,0x9a211f,0x9c211f,0x9e211f,
0xa02120,0xa12220,0xa32220,0xa52220,0xa72221,0xa92221,
0xab2221,0xad2221,0xaf2222,0xb12222,0xb22222,0xb32422,
0xb42622,0xb52821,0xb62a21,0xb72c21,0xb82d21,0xb92f20,
0xba3120,0xbb3220,0xbc341f,0xbd351f,0xbe371f,0xbf381f,
0xc03a1e,0xc13b1e,0xc23d1e,0xc33e1d,0xc4401d,0xc5411c,
0xc6421c,0xc7441c,0xc8451b,0xc9471b,0xca481a,0xcb491a,
0xcc4b19,0xcd4c19,0xce4d18,0xcf4f18,0xd05017,0xd15217,
0xd25316,0xd35415,0xd45515,0xd55714,0xd65813,0xd75913,
0xd85b12,0xd95c11,0xda5d10,0xdb5f0f,0xdc600e,0xdd610d,
0xde620c,0xdf640b,0xe06509,0xe16608,0xe26807,0xe36905,
0xe36b05,0xe36d06,0xe46e07,0xe47007,0xe47208,0xe47408,
0xe57609,0xe5770a,0xe5790a,0xe57b0b,0xe57c0c,0xe67e0c,
0xe6800d,0xe6820e,0xe6830e,0xe6850f,0xe6870f,0xe78810,
0xe78a11,0xe78c11,0xe78d12,0xe78f13,0xe79113,0xe79214,
0xe89415,0xe89615,0xe89716,0xe89916,0xe89a17,0xe89c18,
0xe89e18,0xe89f19,0xe8a11a,0xe8a21a,0xe9a41b,0xe9a61b,
0xe9a71c,0xe9a91d,0xe9aa1d,0xe9ac1e,0xe9ae1e,0xe9af1f,
0xe9b120,0xe9b220,0xe9b421,0xe9b522,0xe9b722,0xe9b923,
0xe9ba23,0xe9bc24,0xe9bd25,0xe9bf25,0xe9c026,0xe9c226,
0xe9c327,0xe9c528,0xe9c728,0xe9c829,0xe8ca2a,0xe8cb2a,
0xe8cd2b,0xe8ce2b,0xe8d02c,0xe8d12d,0xe8d32d,0xe8d52e,
0xe8d62f,0xe8d82f,0xe7d930,0xe7db30,0xe7dc31,0xe7de32,
0xe7df32,0xe7e133,0xe6e234,0xe6e434,0xe6e535,0xe7e73c,
0xe9e745,0xeae84e,0xece957,0xedea5e,0xeeeb66,0xf0ec6d,
0xf1ec75,0xf2ed7c,0xf3ee83,0xf5ef89,0xf6f090,0xf7f197,
0xf8f19e,0xf9f2a4,0xf9f3ab,0xfaf4b1,0xfbf5b8,0xfcf6be,
0xfcf7c5,0xfdf8cb,0xfdf9d2,0xfef9d8,0xfefadf,0xfefbe5,
0xfffcec,0xfffdf2,0xfffef9,0xffffff]);

png_add_colormap("inferno", [
0x000004,0x010005,0x010106,0x010108,0x02010a,0x02020c,
0x02020e,0x030210,0x040312,0x040314,0x050417,0x060419,
0x07051b,0x08051d,0x09061f,0x0a0722,0x0b0724,0x0c0826,
0x0d0829,0x0e092b,0x10092d,0x110a30,0x120a32,0x140b34,
0x150b37,0x160b39,0x180c3c,0x190c3e,0x1b0c41,0x1c0c43,
0x1e0c45,0x1f0c48,0x210c4a,0x230c4c,0x240c4f,0x260c51,
0x280b53,0x290b55,0x2b0b57,0x2d0b59,0x2f0a5b,0x310a5c,
0x320a5e,0x340a5f,0x360961,0x380962,0x390963,0x3b0964,
0x3d0965,0x3e0966,0x400a67,0x420a68,0x440a68,0x450a69,
0x470b6a,0x490b6a,0x4a0c6b,0x4c0c6b,0x4d0d6c,0x4f0d6c,
0x510e6c,0x520e6d,0x540f6d,0x550f6d,0x57106e,0x59106e,
0x5a116e,0x5c126e,0x5d126e,0x5f136e,0x61136e,0x62146e,
0x64156e,0x65156e,0x67166e,0x69166e,0x6a176e,0x6c186e,
0x6d186e,0x6f196e,0x71196e,0x721a6e,0x741a6e,0x751b6e,
0x771c6d,0x781c6d,0x7a1d6d,0x7c1d6d,0x7d1e6d,0x7f1e6c,
0x801f6c,0x82206c,0x84206b,0x85216b,0x87216b,0x88226a,
0x8a226a,0x8c2369,0x8d2369,0x8f2469,0x902568,0x922568,
0x932667,0x952667,0x972766,0x982766,0x9a2865,0x9b2964,
0x9d2964,0x9f2a63,0xa02a63,0xa22b62,0xa32c61,0xa52c60,
0xa62d60,0xa82e5f,0xa92e5e,0xab2f5e,0xad305d,0xae305c,
0xb0315b,0xb1325a,0xb3325a,0xb43359,0xb63458,0xb73557,
0xb93556,0xba3655,0xbc3754,0xbd3853,0xbf3952,0xc03a51,
0xc13a50,0xc33b4f,0xc43c4e,0xc63d4d,0xc73e4c,0xc83f4b,
0xca404a,0xcb4149,0xcc4248,0xce4347,0xcf4446,0xd04545,
0xd24644,0xd34743,0xd44842,0xd54a41,0xd74b3f,0xd84c3e,
0xd94d3d,0xda4e3c,0xdb503b,0xdd513a,0xde5238,0xdf5337,
0xe05536,0xe15635,0xe25734,0xe35933,0xe45a31,0xe55c30,
0xe65d2f,0xe75e2e,0xe8602d,0xe9612b,0xea632a,0xeb6429,
0xeb6628,0xec6726,0xed6925,0xee6a24,0xef6c23,0xef6e21,
0xf06f20,0xf1711f,0xf1731d,0xf2741c,0xf3761b,0xf37819,
0xf47918,0xf57b17,0xf57d15,0xf67e14,0xf68013,0xf78212,
0xf78410,0xf8850f,0xf8870e,0xf8890c,0xf98b0b,0xf98c0a,
0xf98e09,0xfa9008,0xfa9207,0xfa9407,0xfb9606,0xfb9706,
0xfb9906,0xfb9b06,0xfb9d07,0xfc9f07,0xfca108,0xfca309,
0xfca50a,0xfca60c,0xfca80d,0xfcaa0f,0xfcac11,0xfcae12,
0xfcb014,0xfcb216,0xfcb418,0xfbb61a,0xfbb81d,0xfbba1f,
0xfbbc21,0xfbbe23,0xfac026,0xfac228,0xfac42a,0xfac62d,
0xf9c72f,0xf9c932,0xf9cb35,0xf8cd37,0xf8cf3a,0xf7d13d,
0xf7d340,0xf6d543,0xf6d746,0xf5d949,0xf5db4c,0xf4dd4f,
0xf4df53,0xf4e156,0xf3e35a,0xf3e55d,0xf2e661,0xf2e865,
0xf2ea69,0xf1ec6d,0xf1ed71,0xf1ef75,0xf1f179,0xf2f27d,
0xf2f482,0xf3f586,0xf3f68a,0xf4f88e,0xf5f992,0xf6fa96,
0xf8fb9a,0xf9fc9d,0xfafda1,0xfcffa4]);

png_add_colormap("kindlman", [
0x000000,0x050004,0x090008,0x0d010d,0x110110,0x140114,
0x160117,0x19011a,0x1b011d,0x1d0220,0x1e0223,0x1f0226,
0x20022a,0x21022d,0x220230,0x230233,0x240336,0x250339,
0x25033c,0x26033f,0x260342,0x260344,0x270347,0x27044a,
0x27044d,0x270450,0x270453,0x270456,0x270459,0x27045d,
0x270560,0x270563,0x260566,0x26056a,0x25056d,0x250570,
0x240674,0x230677,0x22067b,0x21067e,0x200681,0x200684,
0x1f0688,0x1e078b,0x1d078e,0x1c0791,0x1b0794,0x1a0797,
0x19079a,0x19079d,0x1808a0,0x1808a3,0x1408a6,0x0f08aa,
0x0809ae,0x080caf,0x080faf,0x0813af,0x0816af,0x0819af,
0x081cae,0x0820ad,0x0823ac,0x0826aa,0x0829a8,0x082ba6,
0x082ea5,0x0831a3,0x0833a0,0x08359e,0x08389c,0x073a9a,
0x073c98,0x073e95,0x074093,0x074291,0x07448f,0x07468d,
0x07478b,0x074989,0x074b87,0x064c85,0x064e84,0x065082,
0x065180,0x06537f,0x06547d,0x06567b,0x06577a,0x065878,
0x065a77,0x065b76,0x065d74,0x065e73,0x055f72,0x066071,
0x056270,0x05636e,0x05646d,0x05666c,0x05676b,0x05686a,
0x056969,0x056b68,0x056c67,0x056d66,0x056e65,0x057064,
0x057163,0x057262,0x067360,0x06755f,0x06765e,0x06775c,
0x06785b,0x067a59,0x067b58,0x067c56,0x067d54,0x067f53,
0x068051,0x06814f,0x06824d,0x06844b,0x06854a,0x078648,
0x068746,0x078943,0x078a41,0x078b3f,0x078c3d,0x078e3b,
0x078f38,0x079036,0x079134,0x079331,0x07942f,0x07952c,
0x07962a,0x079727,0x079925,0x079a22,0x089b1f,0x089c1d,
0x089d1a,0x089f17,0x08a014,0x08a112,0x08a20f,0x08a30c,
0x08a50a,0x08a608,0x0ca708,0x0fa808,0x11a908,0x12aa08,
0x14ab08,0x16ad08,0x18ae08,0x1aaf08,0x1db008,0x20b109,
0x23b209,0x26b309,0x29b409,0x2db509,0x30b609,0x34b709,
0x38b809,0x3bb909,0x3fba09,0x43bb09,0x47bc09,0x4bbd09,
0x4fbe09,0x53be09,0x57bf09,0x5bc009,0x5fc109,0x63c109,
0x67c209,0x6bc309,0x6fc409,0x74c409,0x78c509,0x7cc60a,
0x80c60a,0x85c70a,0x89c70a,0x8dc80a,0x91c80a,0x96c90a,
0x9ac90a,0x9eca0a,0xa3ca0a,0xa7ca0a,0xabcb0a,0xafcb0a,
0xb4cb0a,0xb8cc0a,0xbccc0a,0xc1cc0a,0xc5cd0a,0xc9cd0a,
0xcdcd0a,0xd1cd0a,0xd6cd0a,0xdacd0b,0xdfcd0b,0xe4cd0b,
0xe9cd0b,0xedcd0b,0xf3cd0c,0xf6cc39,0xf7cd56,0xf8cd69,
0xf9ce77,0xf9cf83,0xfacf8d,0xfad095,0xfad19d,0xfbd2a3,
0xfbd3a9,0xfbd4ae,0xfbd6b3,0xfcd7b8,0xfcd8bc,0xfcd9c0,
0xfcdac3,0xfcdcc7,0xfcddca,0xfddecd,0xfde0d0,0xfde1d3,
0xfde2d5,0xfde3d8,0xfde5da,0xfde6dd,0xfde8df,0xfee9e1,
0xfeeae3,0xfeece5,0xfeede7,0xfeeee9,0xfef0eb,0xfef1ed,
0xfef2ef,0xfef4f1,0xfef5f3,0xfef7f5,0xfff8f6,0xfff9f8,
0xfffbfa,0xfffcfc,0xfffefd,0xffffff]);

png_add_colormap("kindlman-extended", [
0x000000,0x050004,0x090009,0x0d010d,0x100111,0x130115,
0x160118,0x18011b,0x1a011e,0x1b0222,0x1c0226,0x1d022a,
0x1d022e,0x1e0232,0x1e0335,0x1e0339,0x1e033d,0x1d0341,
0x1d0344,0x1c0348,0x1b044b,0x1b044f,0x1a0452,0x190455,
0x180458,0x17045c,0x16055f,0x150562,0x140565,0x130567,
0x12056a,0x12056d,0x11056f,0x0e0573,0x080677,0x060878,
0x060b78,0x060f77,0x061276,0x061674,0x051972,0x051c70,
0x051f6d,0x05216b,0x052468,0x052665,0x052863,0x052a60,
0x052c5e,0x042e5b,0x043059,0x043157,0x043355,0x043453,
0x043651,0x04374f,0x04394d,0x043a4c,0x043b4a,0x033d49,
0x033e47,0x033f46,0x034145,0x034243,0x034342,0x034441,
0x034540,0x03473f,0x03483d,0x04493c,0x044a3a,0x044b38,
0x044d37,0x044e35,0x044f33,0x045031,0x04512f,0x04522d,
0x04542b,0x045529,0x045627,0x045724,0x045822,0x04591f,
0x045b1d,0x045c1a,0x055d18,0x055e15,0x055f12,0x05600f,
0x05610d,0x05620a,0x056408,0x056506,0x066605,0x086705,
0x0a6805,0x0b6905,0x0d6a05,0x0f6b05,0x116c05,0x146d05,
0x166e05,0x1a6f05,0x1d7005,0x207005,0x247105,0x287205,
0x2b7306,0x2f7406,0x337406,0x377506,0x3b7606,0x3f7606,
0x437706,0x477706,0x4c7806,0x507806,0x547906,0x587906,
0x5c7a06,0x617a06,0x657a06,0x697b06,0x6d7b06,0x717b06,
0x767b06,0x7a7b06,0x7e7b06,0x827b06,0x877b07,0x8b7b07,
0x907b07,0x957a07,0x9a7a07,0xa07908,0xa57808,0xab7708,
0xb17608,0xb77509,0xbd7309,0xc47109,0xca6f0a,0xd16c0a,
0xd8690a,0xde660b,0xe5620b,0xec5e0b,0xf35a0c,0xf45b1b,
0xf55c25,0xf55e2e,0xf56034,0xf6623a,0xf6633f,0xf66543,
0xf66747,0xf6694a,0xf66b4d,0xf76d4f,0xf76f53,0xf77057,
0xf7725b,0xf77360,0xf87565,0xf8766a,0xf87870,0xf87976,
0xf97a7b,0xf97b81,0xf97d87,0xf97e8d,0xf97f93,0xf98199,
0xf9829e,0xf983a4,0xf984a9,0xf985af,0xf986b4,0xf987ba,
0xf989bf,0xf98ac4,0xf98bc9,0xfa8cce,0xfa8dd3,0xfa8ed8,
0xfa8fdd,0xfa90e1,0xfa91e6,0xfa92eb,0xfa93ef,0xfa94f3,
0xfa95f8,0xf898fa,0xf59bfa,0xf29ffa,0xefa2fb,0xeca5fb,
0xeaa8fb,0xe8abfb,0xe6adfb,0xe5b0fb,0xe3b2fb,0xe2b4fb,
0xe1b6fb,0xe0b8fc,0xe0bafc,0xdfbcfc,0xdfbefc,0xdebffc,
0xdec1fc,0xdec3fc,0xdec4fc,0xdfc6fc,0xdfc7fc,0xdfc9fc,
0xe0cafc,0xe0ccfd,0xe1cdfd,0xe2cffd,0xe2d0fd,0xe3d1fd,
0xe4d3fd,0xe5d4fd,0xe5d5fd,0xe6d7fd,0xe7d8fd,0xe7dafd,
0xe7dbfd,0xe8ddfd,0xe8defd,0xe8e0fd,0xe8e1fe,0xe9e3fe,
0xe9e4fe,0xe9e6fe,0xe9e7fe,0xe9e9fe,0xe9eafe,0xeaecfe,
0xeaedfe,0xeaeffe,0xebf0fe,0xebf2fe,0xecf3fe,0xedf5fe,
0xedf6fe,0xeef7fe,0xeff9fe,0xf0fafe,0xf2fbfe,0xf3fcfe,
0xf5fdff,0xf8feff,0xfbffff,0xffffff]);

png_add_colormap("cividis", [
0x00204c,0x00204e,0x002150,0x002251,0x002353,0x002355,
0x002456,0x002558,0x00265a,0x00265b,0x00275d,0x00285f,
0x002861,0x002963,0x002a64,0x002a66,0x002b68,0x002c6a,
0x002d6c,0x002d6d,0x002e6e,0x002e6f,0x002f6f,0x002f6f,
0x00306f,0x00316f,0x00316f,0x00326e,0x00336e,0x00346e,
0x00346e,0x01356e,0x06366e,0x0a376d,0x0e376d,0x12386d,
0x15396d,0x17396d,0x1a3a6c,0x1c3b6c,0x1e3c6c,0x203c6c,
0x223d6c,0x243e6c,0x263e6c,0x273f6c,0x29406b,0x2b416b,
0x2c416b,0x2e426b,0x2f436b,0x31446b,0x32446b,0x33456b,
0x35466b,0x36466b,0x37476b,0x38486b,0x3a496b,0x3b496b,
0x3c4a6b,0x3d4b6b,0x3e4b6b,0x404c6b,0x414d6b,0x424e6b,
0x434e6b,0x444f6b,0x45506b,0x46506b,0x47516b,0x48526b,
0x49536b,0x4a536b,0x4b546b,0x4c556b,0x4d556b,0x4e566b,
0x4f576c,0x50586c,0x51586c,0x52596c,0x535a6c,0x545a6c,
0x555b6c,0x565c6c,0x575d6d,0x585d6d,0x595e6d,0x5a5f6d,
0x5b5f6d,0x5c606d,0x5d616e,0x5e626e,0x5f626e,0x5f636e,
0x60646e,0x61656f,0x62656f,0x63666f,0x64676f,0x65676f,
0x666870,0x676970,0x686a70,0x686a70,0x696b71,0x6a6c71,
0x6b6d71,0x6c6d72,0x6d6e72,0x6e6f72,0x6f6f72,0x6f7073,
0x707173,0x717273,0x727274,0x737374,0x747475,0x757575,
0x757575,0x767676,0x777776,0x787876,0x797877,0x7a7977,
0x7b7a77,0x7b7b78,0x7c7b78,0x7d7c78,0x7e7d78,0x7f7e78,
0x807e78,0x817f78,0x828078,0x838178,0x848178,0x858278,
0x868378,0x878478,0x888578,0x898578,0x8a8678,0x8b8778,
0x8c8878,0x8d8878,0x8e8978,0x8f8a78,0x908b78,0x918c78,
0x928c78,0x938d78,0x948e78,0x958f78,0x968f77,0x979077,
0x989177,0x999277,0x9a9377,0x9b9377,0x9c9477,0x9d9577,
0x9e9676,0x9f9776,0xa09876,0xa19876,0xa29976,0xa39a75,
0xa49b75,0xa59c75,0xa69c75,0xa79d75,0xa89e74,0xa99f74,
0xaaa074,0xaba174,0xaca173,0xada273,0xaea373,0xafa473,
0xb0a572,0xb1a672,0xb2a672,0xb4a771,0xb5a871,0xb6a971,
0xb7aa70,0xb8ab70,0xb9ab70,0xbaac6f,0xbbad6f,0xbcae6e,
0xbdaf6e,0xbeb06e,0xbfb16d,0xc0b16d,0xc1b26c,0xc2b36c,
0xc3b46c,0xc5b56b,0xc6b66b,0xc7b76a,0xc8b86a,0xc9b869,
0xcab969,0xcbba68,0xccbb68,0xcdbc67,0xcebd67,0xd0be66,
0xd1bf66,0xd2c065,0xd3c065,0xd4c164,0xd5c263,0xd6c363,
0xd7c462,0xd8c561,0xd9c661,0xdbc760,0xdcc860,0xddc95f,
0xdeca5e,0xdfcb5d,0xe0cb5d,0xe1cc5c,0xe3cd5b,0xe4ce5b,
0xe5cf5a,0xe6d059,0xe7d158,0xe8d257,0xe9d356,0xebd456,
0xecd555,0xedd654,0xeed753,0xefd852,0xf0d951,0xf1da50,
0xf3db4f,0xf4dc4e,0xf5dd4d,0xf6de4c,0xf7df4b,0xf9e049,
0xfae048,0xfbe147,0xfce246,0xfde345,0xffe443,0xffe542,
0xffe642,0xffe743,0xffe844,0xffe945]);

png_add_colormap("turbo", [
0x30123b,0x311542,0x32184a,0x341b51,0x351e58,0x36215f,
0x372365,0x38266c,0x392972,0x3a2c79,0x3b2f7f,0x3c3285,
0x3c358b,0x3d3791,0x3e3a96,0x3f3d9c,0x4040a1,0x4043a6,
0x4145ab,0x4148b0,0x424bb5,0x434eba,0x4350be,0x4353c2,
0x4456c7,0x4458cb,0x455bce,0x455ed2,0x4560d6,0x4563d9,
0x4666dd,0x4668e0,0x466be3,0x466de6,0x4670e8,0x4673eb,
0x4675ed,0x4678f0,0x467af2,0x467df4,0x467ff6,0x4682f8,
0x4584f9,0x4587fb,0x4589fc,0x448cfd,0x438efd,0x4291fe,
0x4193fe,0x4096fe,0x3f98fe,0x3e9bfe,0x3c9dfd,0x3ba0fc,
0x39a2fc,0x38a5fb,0x36a8f9,0x34aaf8,0x33acf6,0x31aff5,
0x2fb1f3,0x2db4f1,0x2bb6ef,0x2ab9ed,0x28bbeb,0x26bde9,
0x25c0e6,0x23c2e4,0x21c4e1,0x20c6df,0x1ec9dc,0x1dcbda,
0x1ccdd7,0x1bcfd4,0x1ad1d2,0x19d3cf,0x18d5cc,0x18d7ca,
0x17d9c7,0x17dac4,0x17dcc2,0x17debf,0x18e0bd,0x18e1ba,
0x19e3b8,0x1ae4b6,0x1be5b4,0x1de7b1,0x1ee8af,0x20e9ac,
0x22eba9,0x24eca6,0x27eda3,0x29eea0,0x2cef9d,0x2ff09a,
0x32f197,0x35f394,0x38f491,0x3bf48d,0x3ff58a,0x42f687,
0x46f783,0x4af880,0x4df97c,0x51f979,0x55fa76,0x59fb72,
0x5dfb6f,0x61fc6c,0x65fc68,0x69fd65,0x6dfd62,0x71fd5f,
0x74fe5c,0x78fe59,0x7cfe56,0x80fe53,0x84fe50,0x87fe4d,
0x8bfe4b,0x8efe48,0x92fe46,0x95fe44,0x98fe42,0x9bfd40,
0x9efd3e,0xa1fc3d,0xa4fc3b,0xa6fb3a,0xa9fb39,0xacfa37,
0xaef937,0xb1f836,0xb3f835,0xb6f735,0xb9f534,0xbbf434,
0xbef334,0xc0f233,0xc3f133,0xc5ef33,0xc8ee33,0xcaed33,
0xcdeb34,0xcfea34,0xd1e834,0xd4e735,0xd6e535,0xd8e335,
0xdae236,0xdde036,0xdfde36,0xe1dc37,0xe3da37,0xe5d838,
0xe7d738,0xe8d538,0xead339,0xecd139,0xedcf39,0xefcd39,
0xf0cb3a,0xf2c83a,0xf3c63a,0xf4c43a,0xf6c23a,0xf7c039,
0xf8be39,0xf9bc39,0xf9ba38,0xfab737,0xfbb537,0xfbb336,
0xfcb035,0xfcae34,0xfdab33,0xfda932,0xfda631,0xfda330,
0xfea12f,0xfe9e2e,0xfe9b2d,0xfe982c,0xfd952b,0xfd9229,
0xfd8f28,0xfd8c27,0xfc8926,0xfc8624,0xfb8323,0xfb8022,
0xfa7d20,0xfa7a1f,0xf9771e,0xf8741c,0xf7711b,0xf76e1a,
0xf66b18,0xf56817,0xf46516,0xf36315,0xf26014,0xf15d13,
0xef5a11,0xee5810,0xed550f,0xec520e,0xea500d,0xe94d0d,
0xe84b0c,0xe6490b,0xe5460a,0xe3440a,0xe24209,0xe04008,
0xde3e08,0xdd3c07,0xdb3a07,0xd93806,0xd73606,0xd63405,
0xd43205,0xd23005,0xd02f04,0xce2d04,0xcb2b03,0xc92903,
0xc72803,0xc52602,0xc32402,0xc02302,0xbe2102,0xbb1f01,
0xb91e01,0xb61c01,0xb41b01,0xb11901,0xae1801,0xac1601,
0xa91501,0xa61401,0xa31201,0xa01101,0x9d1001,0x9a0e01,
0x970d01,0x940c01,0x910b01,0x8e0a01,0x8b0901,0x870801,
0x840701,0x810602,0x7d0502,0x7a0402]);

png_add_colormap("crest", [
0xa5cd90,0xa4cd90,0xa3cc91,0xa2cb91,0xa1cb91,0x9fca91,
0x9eca91,0x9dc991,0x9cc991,0x9bc891,0x9ac891,0x99c791,
0x98c791,0x97c691,0x95c691,0x94c591,0x93c491,0x92c491,
0x91c391,0x90c391,0x8fc291,0x8ec291,0x8dc191,0x8bc191,
0x8ac091,0x89c091,0x88bf91,0x87bf91,0x86be91,0x85be91,
0x84bd91,0x82bd91,0x81bc91,0x80bc91,0x7fbb91,0x7ebb91,
0x7dba91,0x7cb991,0x7bb991,0x79b891,0x78b891,0x77b791,
0x76b791,0x75b691,0x74b691,0x73b591,0x72b591,0x71b490,
0x70b490,0x6fb390,0x6eb390,0x6db290,0x6cb190,0x6bb190,
0x6ab090,0x69b090,0x68af90,0x67af90,0x66ae90,0x65ae90,
0x64ad90,0x63ac90,0x62ac90,0x61ab90,0x60ab90,0x60aa90,
0x5faa90,0x5ea990,0x5da890,0x5ca890,0x5ba790,0x5aa790,
0x5aa690,0x59a590,0x58a590,0x57a490,0x56a490,0x56a390,
0x55a390,0x54a290,0x53a190,0x53a190,0x52a090,0x51a090,
0x509f90,0x509e90,0x4f9e90,0x4e9d90,0x4d9d90,0x4d9c90,
0x4c9b90,0x4b9b90,0x4b9a90,0x4a9a8f,0x49998f,0x49988f,
0x48988f,0x47978f,0x46978f,0x46968f,0x45958f,0x44958f,
0x44948f,0x43948f,0x43938f,0x42928f,0x41928f,0x41918f,
0x40918f,0x3f908f,0x3f908e,0x3e8f8e,0x3d8e8e,0x3d8e8e,
0x3c8d8e,0x3b8d8e,0x3b8c8e,0x3a8b8e,0x398b8e,0x398a8e,
0x388a8e,0x38898e,0x37888e,0x36888e,0x36878d,0x35878d,
0x34868d,0x34858d,0x33858d,0x32848d,0x32848d,0x31838d,
0x30828d,0x30828d,0x2f818d,0x2e818d,0x2e808d,0x2d808d,
0x2c7f8c,0x2c7e8c,0x2b7e8c,0x2b7d8c,0x2a7d8c,0x297c8c,
0x297b8c,0x287b8c,0x277a8c,0x277a8c,0x26798c,0x26788c,
0x25788c,0x25778c,0x24778b,0x23768b,0x23758b,0x22758b,
0x22748b,0x21748b,0x21738b,0x20728b,0x20728b,0x20718b,
0x1f708b,0x1f708b,0x1e6f8b,0x1e6f8a,0x1e6e8a,0x1d6d8a,
0x1d6d8a,0x1d6c8a,0x1d6b8a,0x1d6b8a,0x1c6a8a,0x1c6a8a,
0x1c6989,0x1c6889,0x1c6889,0x1c6789,0x1c6689,0x1c6689,
0x1c6589,0x1c6488,0x1c6488,0x1c6388,0x1c6288,0x1c6288,
0x1c6188,0x1c6087,0x1d6087,0x1d5f87,0x1d5e87,0x1d5e87,
0x1d5d86,0x1e5c86,0x1e5c86,0x1e5b86,0x1e5a86,0x1e5a85,
0x1f5985,0x1f5885,0x1f5885,0x1f5784,0x205684,0x205684,
0x205584,0x215483,0x215483,0x215383,0x215283,0x225282,
0x225182,0x225082,0x235081,0x234f81,0x234e81,0x244d80,
0x244d80,0x244c80,0x254b7f,0x254b7f,0x254a7f,0x25497e,
0x26497e,0x26487e,0x26477d,0x27467d,0x27467c,0x27457c,
0x27447c,0x28447b,0x28437b,0x28427a,0x28417a,0x29417a,
0x294079,0x293f79,0x293f78,0x293e78,0x2a3d78,0x2a3c77,
0x2a3c77,0x2a3b76,0x2a3a76,0x2a3976,0x2b3975,0x2b3875,
0x2b3775,0x2b3774,0x2b3674,0x2b3574,0x2b3473,0x2b3373,
0x2c3373,0x2c3272,0x2c3172,0x2c3072]);

png_add_colormap("flare", [
0xedb081,0xedaf80,0xedae7f,0xedad7f,0xedac7e,0xedab7e,
0xedaa7d,0xeda97c,0xeda87c,0xeda77b,0xeda67b,0xeca57a,
0xeca479,0xeca379,0xeca278,0xeca178,0xeca077,0xec9f76,
0xec9e76,0xec9d75,0xec9c75,0xec9c74,0xeb9b73,0xeb9a73,
0xeb9972,0xeb9872,0xeb9771,0xeb9670,0xeb9570,0xeb946f,
0xeb936f,0xea926e,0xea916e,0xea906d,0xea8f6c,0xea8e6c,
0xea8d6b,0xea8c6b,0xea8b6a,0xea8a69,0xe98969,0xe98868,
0xe98768,0xe98667,0xe98567,0xe98466,0xe98366,0xe88265,
0xe88165,0xe88064,0xe87f64,0xe87e63,0xe87d63,0xe77c62,
0xe77b62,0xe77a62,0xe77961,0xe77861,0xe77760,0xe67660,
0xe67560,0xe6745f,0xe6735f,0xe5725f,0xe5715e,0xe5705e,
0xe56f5e,0xe56e5d,0xe46d5d,0xe46c5d,0xe46b5d,0xe36a5d,
0xe3695c,0xe3685c,0xe3675c,0xe2665c,0xe2655c,0xe2645c,
0xe1635c,0xe1625c,0xe1615c,0xe0605c,0xe05f5c,0xe05e5c,
0xdf5d5c,0xdf5d5c,0xde5c5c,0xde5b5c,0xdd5a5c,0xdd595c,
0xdd585c,0xdc575c,0xdc565c,0xdb555d,0xdb555d,0xda545d,
0xd9535d,0xd9525d,0xd8515e,0xd8515e,0xd7505e,0xd74f5f,
0xd64e5f,0xd54e5f,0xd54d5f,0xd44c60,0xd34b60,0xd34b60,
0xd24a61,0xd14a61,0xd14961,0xd04862,0xcf4862,0xce4762,
0xce4763,0xcd4663,0xcc4664,0xcb4564,0xca4564,0xca4465,
0xc94465,0xc84365,0xc74366,0xc64266,0xc54266,0xc44267,
0xc44167,0xc34167,0xc24068,0xc14068,0xc04068,0xbf3f69,
0xbe3f69,0xbd3f69,0xbc3f69,0xbb3e6a,0xba3e6a,0xb93e6a,
0xb83d6a,0xb73d6b,0xb63d6b,0xb53d6b,0xb53c6b,0xb43c6c,
0xb33c6c,0xb23c6c,0xb13b6c,0xb03b6c,0xaf3b6d,0xae3b6d,
0xad3a6d,0xac3a6d,0xab3a6d,0xaa3a6e,0xa9396e,0xa8396e,
0xa7396e,0xa6396e,0xa5386e,0xa4386f,0xa3386f,0xa3386f,
0xa2376f,0xa1376f,0xa0376f,0x9f376f,0x9e366f,0x9d3670,
0x9c3670,0x9b3670,0x9a3570,0x993570,0x983570,0x973570,
0x963470,0x953470,0x953470,0x943470,0x933371,0x923371,
0x913371,0x903371,0x8f3271,0x8e3271,0x8d3271,0x8c3271,
0x8b3171,0x8a3171,0x893171,0x883171,0x873071,0x873071,
0x863071,0x853071,0x843070,0x832f70,0x822f70,0x812f70,
0x802f70,0x7f2f70,0x7e2e70,0x7d2e70,0x7c2e70,0x7b2e70,
0x7a2e6f,0x792d6f,0x782d6f,0x772d6f,0x762d6f,0x752d6f,
0x742c6e,0x742c6e,0x732c6e,0x722c6e,0x712c6e,0x702c6d,
0x6f2b6d,0x6e2b6d,0x6d2b6d,0x6c2b6d,0x6b2b6c,0x6a2a6c,
0x692a6c,0x682a6b,0x672a6b,0x662a6b,0x652a6b,0x64296a,
0x63296a,0x62296a,0x622969,0x612969,0x602869,0x5f2869,
0x5e2868,0x5d2868,0x5c2868,0x5b2767,0x5a2767,0x592767,
0x582766,0x572666,0x562666,0x562665,0x552665,0x542565,
0x532565,0x522564,0x512564,0x502464,0x4f2463,0x4e2463,
0x4d2363,0x4d2362,0x4c2362,0x4b2362]);

png_add_colormap("iceandfire", [
0xbde7db,0xbae5da,0xb7e3da,0xb5e1d9,0xb2dfd8,0xafddd8,
0xacdbd7,0xaad9d6,0xa7d7d6,0xa4d5d5,0xa1d3d4,0x9ed1d4,
0x9bd0d3,0x98ced3,0x96ccd2,0x93cad1,0x90c8d1,0x8dc6d1,
0x8ac4d0,0x87c3d0,0x84c1cf,0x81bfcf,0x7ebdcf,0x7bbbce,
0x78bace,0x75b8ce,0x72b6ce,0x6eb4ce,0x6bb2ce,0x68b1ce,
0x65afcd,0x62adcd,0x60abcd,0x5da9cd,0x5aa7cd,0x57a6cd,
0x55a4cd,0x52a2cd,0x50a0ce,0x4e9ece,0x4b9cce,0x499ace,
0x4798ce,0x4596ce,0x4394ce,0x4192cf,0x3f90cf,0x3d8ecf,
0x3c8ccf,0x3a8ad0,0x3987d0,0x3885d0,0x3783d0,0x3681d0,
0x367fd0,0x367cd0,0x377ad0,0x3778d0,0x3875cf,0x3a73cf,
0x3b70ce,0x3c6ecc,0x3e6ccb,0x3f69c9,0x4167c8,0x4265c6,
0x4362c3,0x4460c1,0x455ebf,0x465cbc,0x475ab9,0x4858b7,
0x4956b3,0x4954b0,0x4952ad,0x4950a9,0x494fa5,0x494da2,
0x494b9e,0x484a9a,0x484996,0x474792,0x46468e,0x45448a,
0x454386,0x444282,0x43417f,0x413f7b,0x403e77,0x3f3d74,
0x3e3c70,0x3d3b6d,0x3c3969,0x3a3866,0x393762,0x38365f,
0x37355c,0x353359,0x343256,0x333153,0x323050,0x312f4d,
0x2f2e4a,0x2e2d47,0x2d2c45,0x2c2b42,0x2b2a40,0x2a293d,
0x29283b,0x282738,0x272636,0x262534,0x252432,0x242330,
0x23232e,0x22222c,0x22212b,0x212129,0x212027,0x202026,
0x201f25,0x1f1f23,0x1f1e22,0x1f1e21,0x1f1e20,0x1f1e1f,
0x1f1e1f,0x1f1e1e,0x1f1e1e,0x201e1e,0x211d1e,0x221d1e,
0x231d1e,0x241d1e,0x251d1f,0x271d1f,0x281e1f,0x2a1e20,
0x2b1e20,0x2d1e21,0x2f1f22,0x311f22,0x321f23,0x342024,
0x362025,0x382125,0x3a2126,0x3c2127,0x3f2228,0x412229,
0x43232a,0x45242b,0x48242c,0x4a252d,0x4d252e,0x4f262f,
0x522630,0x542732,0x572733,0x592834,0x5c2835,0x5f2936,
0x622937,0x642a38,0x672a39,0x6a2a3a,0x6d2b3b,0x702b3c,
0x722b3d,0x752c3e,0x782c3f,0x7b2c3f,0x7e2d40,0x812d41,
0x842d41,0x872d42,0x8a2d43,0x8d2d43,0x902e44,0x932e44,
0x962e44,0x992e44,0x9c2e44,0xa02e44,0xa32f44,0xa52f44,
0xa82f43,0xab3043,0xae3142,0xb13242,0xb43241,0xb63340,
0xb93440,0xbb363f,0xbe373e,0xc0383d,0xc33a3c,0xc53b3b,
0xc73d3a,0xca3f3a,0xcc4139,0xce4338,0xd04537,0xd24736,
0xd44936,0xd64b35,0xd74e35,0xd95034,0xdb5234,0xdc5533,
0xde5733,0xe05a33,0xe15c33,0xe25f33,0xe46133,0xe56433,
0xe66733,0xe86a34,0xe96c35,0xea6f36,0xeb7238,0xeb7539,
0xec783b,0xed7b3d,0xee7e40,0xee8142,0xef8445,0xf08748,
0xf0894b,0xf18c4e,0xf28f51,0xf29254,0xf39558,0xf3985b,
0xf49a5f,0xf49d62,0xf5a066,0xf6a36a,0xf6a56d,0xf7a871,
0xf7ab75,0xf8ae79,0xf8b07c,0xf9b380,0xfab684,0xfab987,
0xfbbb8b,0xfbbe8f,0xfcc192,0xfcc496,0xfdc69a,0xfdc99e,
0xfecca1,0xfecfa5,0xffd1a9,0xffd4ad]);

png_add_colormap("mako", [
0x0b0305,0x0c0406,0x0d0407,0x0e0509,0x0f060a,0x10070c,
0x11070d,0x12080e,0x130910,0x150a11,0x160b13,0x170b14,
0x180c16,0x190d17,0x1a0e19,0x1b0f1a,0x1c0f1c,0x1d101d,
0x1e111f,0x1f1220,0x201222,0x211323,0x221425,0x231526,
0x241528,0x241629,0x25172b,0x26182c,0x27182e,0x28192f,
0x291a31,0x2a1a33,0x2b1b34,0x2c1c36,0x2c1d37,0x2d1d39,
0x2e1e3b,0x2f1f3c,0x30203e,0x302040,0x312141,0x322243,
0x332345,0x332346,0x342448,0x35254a,0x35264c,0x36264d,
0x36274f,0x372851,0x382952,0x382a54,0x392a56,0x392b58,
0x3a2c59,0x3a2d5b,0x3b2e5d,0x3b2f5f,0x3c2f60,0x3c3062,
0x3d3164,0x3d3266,0x3d3367,0x3e3469,0x3e346b,0x3e356d,
0x3f366f,0x3f3770,0x3f3872,0x403974,0x403a76,0x403b78,
0x403b79,0x403c7b,0x403d7d,0x403e7f,0x413f80,0x414082,
0x414184,0x404285,0x404387,0x404488,0x40458a,0x40468b,
0x40478d,0x3f498e,0x3f4a8f,0x3f4b91,0x3e4c92,0x3e4d93,
0x3e4e94,0x3d5095,0x3d5196,0x3c5296,0x3c5397,0x3c5498,
0x3b5699,0x3b5799,0x3a589a,0x3a599a,0x3a5a9b,0x395c9b,
0x395d9c,0x395e9c,0x385f9c,0x38609d,0x38629d,0x37639d,
0x37649e,0x37659e,0x37669e,0x36689f,0x36699f,0x366a9f,
0x366b9f,0x366ca0,0x366da0,0x356ea0,0x3570a0,0x3571a1,
0x3572a1,0x3573a1,0x3574a1,0x3575a2,0x3576a2,0x3578a2,
0x3579a2,0x357aa3,0x347ba3,0x347ca3,0x347da3,0x347ea4,
0x347fa4,0x3481a4,0x3482a4,0x3483a5,0x3484a5,0x3485a5,
0x3486a5,0x3487a6,0x3488a6,0x3489a6,0x348ba7,0x348ca7,
0x348da7,0x348ea7,0x348fa8,0x3490a8,0x3491a8,0x3492a8,
0x3494a9,0x3495a9,0x3496a9,0x3497a9,0x3498a9,0x3499aa,
0x349aaa,0x349baa,0x349daa,0x359eaa,0x359fab,0x35a0ab,
0x35a1ab,0x35a2ab,0x36a3ab,0x36a4ac,0x36a5ac,0x37a7ac,
0x37a8ac,0x38a9ac,0x38aaac,0x38abac,0x39acac,0x3aadad,
0x3aaead,0x3bb0ad,0x3bb1ad,0x3cb2ad,0x3db3ad,0x3eb4ad,
0x3eb5ad,0x3fb6ad,0x40b7ad,0x41b8ad,0x42baad,0x43bbad,
0x44bcad,0x45bdad,0x46bead,0x47bfad,0x48c0ad,0x49c1ad,
0x4ac2ad,0x4cc4ad,0x4dc5ad,0x4ec6ad,0x50c7ad,0x52c8ad,
0x53c9ad,0x55caad,0x57cbad,0x59ccad,0x5bcdad,0x5dcead,
0x60cfad,0x62d0ad,0x65d1ad,0x68d1ad,0x6ad2ad,0x6dd3ad,
0x70d4ad,0x73d5ae,0x76d5ae,0x79d6ae,0x7cd7af,0x7fd7b0,
0x82d8b0,0x85d9b1,0x88dab2,0x8bdab2,0x8edbb3,0x91dcb4,
0x94dcb5,0x96ddb6,0x99deb7,0x9cdeb8,0x9fdfb9,0xa1e0ba,
0xa4e0bb,0xa6e1bc,0xa9e2bd,0xabe3bf,0xaee3c0,0xb0e4c1,
0xb2e5c3,0xb5e5c4,0xb7e6c6,0xb9e7c7,0xbce8c9,0xbee8ca,
0xc0e9cc,0xc2eace,0xc4ebcf,0xc6ebd1,0xc8ecd3,0xcaedd4,
0xcdeed6,0xcfefd8,0xd1efd9,0xd3f0db,0xd5f1dd,0xd7f2de,
0xd9f3e0,0xdbf3e2,0xddf4e4,0xdff5e5]);

png_add_colormap("rocket", [
0x020419,0x03051a,0x04051b,0x05061c,0x07071d,0x08071e,
0x09081f,0x0b0920,0x0c0a21,0x0e0a22,0x0f0b23,0x110c24,
0x120c25,0x140d26,0x150e27,0x170e28,0x180f29,0x19102a,
0x1b102b,0x1c112c,0x1e112d,0x1f122e,0x21122f,0x221330,
0x241331,0x251432,0x261433,0x281534,0x291535,0x2b1637,
0x2c1638,0x2e1739,0x2f173a,0x31173b,0x32183c,0x34183d,
0x35183e,0x37193f,0x38193f,0x3a1940,0x3b1a41,0x3d1a42,
0x3e1a43,0x401b44,0x411b45,0x431b46,0x441b46,0x461c47,
0x481c48,0x491c49,0x4b1c4a,0x4c1d4a,0x4e1d4b,0x4f1d4c,
0x511d4d,0x531d4d,0x541d4e,0x561e4f,0x571e4f,0x591e50,
0x5b1e50,0x5c1e51,0x5e1e52,0x5f1e52,0x611e53,0x631e53,
0x641f54,0x661f54,0x681f55,0x691f55,0x6b1f56,0x6c1f56,
0x6e1f56,0x701f57,0x711f57,0x731f58,0x751f58,0x761f58,
0x781e59,0x7a1e59,0x7b1e59,0x7d1e59,0x7f1e5a,0x811e5a,
0x821e5a,0x841e5a,0x861e5a,0x871d5b,0x891d5b,0x8b1d5b,
0x8c1d5b,0x8e1d5b,0x901c5b,0x921c5b,0x931c5b,0x951b5b,
0x971b5b,0x991b5b,0x9a1b5b,0x9c1a5b,0x9e1a5b,0x9f1a5b,
0xa1195b,0xa3195b,0xa5195a,0xa6185a,0xa8185a,0xaa185a,
0xab175a,0xad1759,0xaf1759,0xb11759,0xb21658,0xb41658,
0xb61657,0xb71657,0xb91656,0xba1656,0xbc1655,0xbe1655,
0xbf1654,0xc11654,0xc21753,0xc41752,0xc61852,0xc71851,
0xc81950,0xca1a50,0xcb1a4f,0xcd1b4e,0xce1c4e,0xd01d4d,
0xd11e4c,0xd2204b,0xd4214b,0xd5224a,0xd62349,0xd72548,
0xd92648,0xda2847,0xdb2946,0xdc2b46,0xdd2c45,0xde2e44,
0xe02f43,0xe13143,0xe23242,0xe33441,0xe43641,0xe53740,
0xe63940,0xe63b3f,0xe73d3f,0xe83e3e,0xe9403e,0xea423e,
0xea443e,0xeb463e,0xec483e,0xec4a3e,0xed4c3e,0xed4e3e,
0xee503e,0xee523e,0xef543f,0xef563f,0xf05840,0xf05a41,
0xf05c41,0xf15e42,0xf16043,0xf16244,0xf16345,0xf26546,
0xf26747,0xf26948,0xf26b49,0xf36d4a,0xf36f4c,0xf3714d,
0xf3734e,0xf4744f,0xf47651,0xf47852,0xf47a54,0xf47c55,
0xf47d57,0xf47f58,0xf58159,0xf5835b,0xf5845c,0xf5865e,
0xf5885f,0xf58a61,0xf58b63,0xf58d64,0xf58f66,0xf69167,
0xf69269,0xf6946b,0xf6966c,0xf6976e,0xf69970,0xf69b71,
0xf69c73,0xf69e75,0xf6a077,0xf6a178,0xf6a37a,0xf6a57c,
0xf6a67e,0xf6a880,0xf6a981,0xf6ab83,0xf6ad85,0xf6ae87,
0xf6b089,0xf7b28b,0xf7b38d,0xf7b58f,0xf7b691,0xf7b893,
0xf7b995,0xf7bb97,0xf7bd99,0xf7be9b,0xf7c09d,0xf7c1a0,
0xf7c3a2,0xf7c4a4,0xf7c6a6,0xf7c7a8,0xf7c9aa,0xf7caad,
0xf8ccaf,0xf8cdb1,0xf8cfb3,0xf8d0b6,0xf8d2b8,0xf8d3ba,
0xf8d5bc,0xf8d6be,0xf9d8c1,0xf9d9c3,0xf9dbc5,0xf9dcc7,
0xf9dec9,0xf9dfcc,0xf9e1ce,0xfae2d0,0xfae4d2,0xfae5d4,
0xfae7d6,0xfae8d9,0xfbeadb,0xfbebdd]);

png_add_colormap("vlag", [
0x2369bd,0x266abd,0x296cbd,0x2c6dbd,0x2f6ebc,0x316fbc,
0x3470bc,0x3671bc,0x3872bc,0x3b73bc,0x3d74bc,0x3f75bc,
0x4176bc,0x4377bc,0x4578bc,0x4779bc,0x497abc,0x4b7cbc,
0x4d7dbc,0x4f7ebc,0x517fbc,0x5380bc,0x5581bc,0x5782bd,
0x5983bd,0x5b84bd,0x5c85bd,0x5e86bd,0x6087bd,0x6288be,
0x6389be,0x658abe,0x678bbe,0x698cbf,0x6a8dbf,0x6c8fbf,
0x6e90bf,0x6f91c0,0x7192c0,0x7393c0,0x7494c1,0x7695c1,
0x7896c1,0x7997c2,0x7b98c2,0x7d99c2,0x7e9ac3,0x809bc3,
0x829cc3,0x839ec4,0x859fc4,0x87a0c5,0x88a1c5,0x8aa2c6,
0x8ba3c6,0x8da4c6,0x8fa5c7,0x90a6c7,0x92a7c8,0x93a9c8,
0x95aac9,0x97abc9,0x98acca,0x9aadca,0x9baecb,0x9dafcb,
0x9fb0cc,0xa0b2cc,0xa2b3cd,0xa3b4ce,0xa5b5ce,0xa7b6cf,
0xa8b7cf,0xaab8d0,0xacbad1,0xadbbd1,0xafbcd2,0xb0bdd2,
0xb2bed3,0xb4bfd4,0xb5c1d4,0xb7c2d5,0xb8c3d6,0xbac4d6,
0xbcc5d7,0xbdc7d8,0xbfc8d8,0xc0c9d9,0xc2cada,0xc4ccdb,
0xc5cddb,0xc7cedc,0xc9cfdd,0xcad1de,0xccd2de,0xcdd3df,
0xcfd4e0,0xd1d6e1,0xd2d7e2,0xd4d8e2,0xd6d9e3,0xd7dbe4,
0xd9dce5,0xdbdde6,0xdcdfe7,0xdee0e8,0xe0e1e8,0xe1e3e9,
0xe3e4ea,0xe4e5eb,0xe6e7ec,0xe8e8ed,0xe9e9ee,0xebeaef,
0xedecf0,0xeeedf0,0xf0eef1,0xf1eff2,0xf3f0f3,0xf4f2f3,
0xf5f2f4,0xf6f3f5,0xf7f4f5,0xf8f5f5,0xf9f5f5,0xfaf6f5,
0xfaf6f5,0xfaf6f5,0xfbf6f5,0xfbf5f4,0xfbf5f4,0xfbf4f3,
0xfbf3f2,0xfaf2f1,0xfaf1f0,0xfaf0ef,0xf9eeed,0xf9edec,
0xf8ecea,0xf7eae9,0xf7e9e7,0xf6e7e6,0xf5e5e4,0xf4e4e2,
0xf4e2e1,0xf3e1df,0xf2dfdd,0xf1dddc,0xf1dcda,0xf0dad8,
0xefd8d7,0xeed7d5,0xeed5d4,0xedd4d2,0xecd2d0,0xebd0cf,
0xebcfcd,0xeacdcb,0xe9ccca,0xe9cac8,0xe8c8c6,0xe7c7c5,
0xe7c5c3,0xe6c4c2,0xe5c2c0,0xe4c1be,0xe4bfbd,0xe3bdbb,
0xe2bcba,0xe2bab8,0xe1b9b6,0xe1b7b5,0xe0b6b3,0xdfb4b2,
0xdfb3b0,0xdeb1af,0xddafad,0xddaeab,0xdcacaa,0xdbaba8,
0xdba9a7,0xdaa8a5,0xdaa6a4,0xd9a5a2,0xd8a3a1,0xd8a29f,
0xd7a09e,0xd69f9c,0xd69d9a,0xd59c99,0xd59a97,0xd49996,
0xd39794,0xd39693,0xd29491,0xd19390,0xd1918e,0xd0908d,
0xd08e8b,0xcf8d8a,0xce8b88,0xce8a87,0xcd8885,0xcd8784,
0xcc8582,0xcb8481,0xcb827f,0xca817e,0xc97f7d,0xc97e7b,
0xc87c7a,0xc87b78,0xc77977,0xc67875,0xc67674,0xc57572,
0xc47371,0xc4726f,0xc3706e,0xc26f6d,0xc26d6b,0xc16c6a,
0xc06a68,0xc06967,0xbf6765,0xbe6664,0xbe6463,0xbd6361,
0xbc6160,0xbc605e,0xbb5e5d,0xba5d5b,0xba5b5a,0xb95a59,
0xb85857,0xb85756,0xb75554,0xb65353,0xb65252,0xb55050,
0xb44f4f,0xb34d4e,0xb34c4c,0xb24a4b,0xb14949,0xb04748,
0xb04547,0xaf4445,0xae4244,0xad4143,0xad3f41,0xac3d40,
0xab3c3f,0xaa3a3d,0xa9383c,0xa9373b]);

#endif

%
% Code from src/slang/astronomy/deg2dms.sl
%
define deg2dms(deg) {
%!%+
%\function{deg2dms}
%\synopsis{Convert a floating point angle in degrees to degrees, minutes, seconds}
%\usage{(d,m,s) = deg2dms(degree)}
%\qualifiers{
%\qualifier{hours}{If set, return the coordinate in hours, minutes, sec (RA)}
%\qualifier{radian}{If set, the angle is in radians, not degrees}
%}
%\description
% This is a convenience routine to convert astronomical coordinates given
% as a floating point number into a number in degrees, minutes, seconds.
%
% For negative angles, the sign of the coordinate is given to the highest 
% non-zero integer.
%
% Use angle2string to generate strings from angles, and generate_iauname
% if you want to build strings to use in source names.
%
%\seealso{dms2deg, hms2deg, angle2string, generate_iauname}
%!%-
    variable sgn;
    sgn=sign(deg);
    variable angle=abs(deg);

    if (qualifier_exists("radian")) {
	angle*=180./PI;
    }

    if (qualifier_exists("hours")) {
	angle=angle/15.;
    }

    variable d=int(angle);
    angle=(angle-d)*60.;
    variable m=int(angle);
    variable s=(angle-m)*60.;

    if (d==0) {
	if (m==0) {
	    s=sgn*s;
	} else {
	    m=sgn*m;
	}
    } else {
	d=sgn*d;
    }

    return (d,m,s);
}

%
% Code from src/fitting/fit-functions/detconst.sl
%
%!%+
%\function{detconst}
%\synopsis{fit-function providing detector calibration constants}
%\usage{detconst(id)}
%\description
%    Simply returns a constant depending on the detector of
%    the current dataset the model is evaluated on. This can
%    be used to account for flux calibration differences
%    between several detectors.
%
%    To use this fit-function, the detectors has to be
%    defined first using 'detconst_init'.
%\seealso{detconst_init}
%!%-


%%%%%%%%%%%%%%%%%%%%%
define detconst_init() {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{detconst_init}
%\synopsis{initializes the "dectonst" fit-function}
%\usage{detconst_init(String_Type[] detectors);
% or detconst_init(List_Type detectors);}
%\description
%    Before any detector calibration constants can be fitted
%    by using the 'detconst' fit-function, this function has
%    to be called. It initalizes and defines the fit-function
%    based on the given detector names (and if statements).
%
%    In general, the detector names are given as an array of
%    strings. Each entry has to represent the 'instrument'
%    field defined in the FITS-header of any used dataset.
%    Internally, the 'detconst' matches this field against
%    the given names and returns the corresponding parameter.
%
%    If the instrument field is not enough to determine the
%    parameter (e.g., if multiple instruments are build within
%    the same detector), a list of strings can be provided
%    instead of an array. If a list item itself is an array
%    with exactly two strings, the first one is the detector
%    name and the second specifies the if statement to be
%    evaluated to determine the corresponding parameter.
%    Within the if statement, information about the dataset
%    can be accessed via the 'info' struct (see get_data_info).
%
%    If the 'debug' qualifier is given, the S-Lang code of
%    the fit-function is printed out as well.
%\example
%    % RXTE consisted of two instruments: PCA and HEXTE
%    detconst_init(["PCA", "HEXTE"]);
%    
%    % SUZAKU consists of six instruments: XIS0-3, PIN,
%    % and GSO. PIN and GSO are, however, within one detector
%    % called HXD (as in the FITS-header). Thus, identify
%    % these two instruments via the PHA-filename
%    detconst_init({"XIS0", "XIS1", "XIS2", "XIS3",
%      ["PIN", "is_substr(info.file, \\"hxd_pin\\")"],
%      ["GSO", "is_substr(info.file, \\"hxd_gso\\")"]
%    }; debug);
%\seealso{detconst, get_data_info, add_slang_function}
%!%-
  variable dtctrs = list_new();
  variable d;
  
  if (_NARGS == 1) {
    variable in = ();
    % convert input into list
    if (typeof(in) == String_Type) list_append(dtctrs, in);
    else if (typeof(in) == Array_Type && _typeof(in) == String_Type) foreach d (in) list_append(dtctrs, d);
    else if (typeof(in) == List_Type) dtctrs = in;
    else { vmessage("error (%s): input list of detectors is of wrong type", _function_name); return; }
  }
  else { help(_function_name); return; }

  % define fit function
  variable ffstrng = "", nonspecial = 0;
  variable dtctrnms = String_Type[0], special = Integer_Type[0];
  ffstrng += "define detconst_fit(lo, hi, pars) {\n  variable dtctrs =";
  ffstrng += " [";
  _for d (0, length(dtctrs)-1, 1) { % get detector names and entries with special if cases
    if (typeof(dtctrs[d]) == String_Type) {
      ffstrng += sprintf("\"%s\",", dtctrs[d]);
      dtctrnms = [dtctrnms, dtctrs[d]]; nonspecial++;
    }
    else {
      ffstrng += sprintf("\"%s\",", dtctrs[d][0]);
      dtctrnms = [dtctrnms, dtctrs[d][0]], special = [special, d];
    }
  }
  ffstrng = substr(ffstrng, 1, strlen(ffstrng)-1);
  if (nonspecial > 0) { ffstrng += "];\n"; }
  else { ffstrng += "String_Type[0];\n"; }
  ffstrng += "  variable info = get_data_info(Isis_Active_Dataset);\n";
  % search and return detector calibration constant without special if case
  ffstrng += "  variable n = wherefirst(dtctrs == info.instrument);\n";
  ffstrng += "  if (n != NULL) { return pars[n[0]]; }\n";
  % write statements for detectors with special if case
  foreach d (special) {
    ffstrng += sprintf("  if (%s) { return pars[%d]; }\n", dtctrs[d][1], d);
  }
  ffstrng += "  vmessage(\"warning (detconst): instrument '%s' of dataset %d does not have a known fit parameter\", info.instrument, Isis_Active_Dataset);\n";
  ffstrng += "  return 0;\n}";

  if (qualifier_exists("debug")) message(ffstrng);
  eval(ffstrng, "isis");

  % defaults function
  eval("private define detconst_defaults(i) { return struct { value = 1, freeze = i == 0 ? 1 : 0, min = .5, max = 2, hard_min = 0, hard_max = 10, step = 1e-3, relstep = 1e-4 }; }", "isis");

  % add function to fit-functions
  eval(sprintf("add_slang_function(\"detconst\", [\"%s\"]);", strjoin(dtctrnms, "\",\"")), "isis");
  eval("set_param_default_hook(\"detconst\", &detconst_defaults);", "isis");
}

%
% Code from src/slang/arrays/diff.sl
%
%%%%%%%%%%%%%%%%%%%%%
define diff(a)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{diff}
%\synopsis{returns the differences between adjacent elements in an array}
%\usage{Double_Type diff(Double_Type[] array);}
%\description
%    a = sqr([1:4]); % [1,4,9,16]
%    b = diff(a);    % [3,5,7]
%!%-
{
  return a[[1:length(a)-1]] - a[[0:length(a)-2]];
}

%
% Code from src/misc/radio/difmap_restore.sl
%
define difmap_restore()
%!%+
%\function{difmap_restore}
%\synopsis{DIFMAP is used to restore a radio image with a new beam}
%\usage{String_Type outname = difmap_restore(String_Type \code{fitsfile}, Double_Type \code{smajor}, \code{sminor}, \code{pos_angle})}
%\qualifiers{
%\qualifier{chan}{[=i] choose channel (i,q,u)}
%\qualifier{xsize}{[= key "NAXIS1"] number of RA pixels}
%\qualifier{xsetp}{[= key "CDELT1"] step size of each RA pixel in mas}
%\qualifier{ysize}{[= key "NAXIS2"] number of DEC pixels}
%\qualifier{ysetp}{[= key "CDELT2"] step size of each DEC pixel in mas}
%\qualifier{uvtaper}{[= " "] pply a uvtaper before restoring}
%\qualifier{outname}{[= "xxxxMHz_restore.fits"] filename of the output, be default the
%                                frequency is read from the input file}
%\qualifier{overwrite}{overwrite file with outname if it already exists}
%}
%\description
%    This function creates a radio image with a given beam using DIFMAP. The
%    new beam has to be defined by its semimajor and semiminor axis and its
%    position angle. The new images is saved in the current working directory.
%    It is assumed that the UVF and MOD files have the same basename as the
%    provided file name if the FITS image.
%\seealso{make_spix}
%!%-
{
  message("running restore");
  variable path, major, minor, posang;
  switch(_NARGS)
  { case 4: (path, major, minor, posang) = (); }
  { help(_function_name()); return NULL; }

  if (system("which difmap  > /dev/null 2>&1") != 0)
  {
    vmessage("ERROR %s: program DIFMAP not found",_function_name());
    return NULL;
  };
  variable chan = qualifier("chan", "i");
  variable xsize  = qualifier("xsize", fits_read_key (path, "NAXIS1"));
  variable xstep  = qualifier("xstep", fits_read_key (path, "CDELT1")*(3.6e+6));
  variable ysize  = qualifier("ysize", fits_read_key (path, "NAXIS2"));
  variable ystep  = qualifier("ystep", fits_read_key (path, "CDELT2")*(3.6e+6));
  variable uvtaper = qualifier("uvtaper", " ");
  variable freqMHz= fits_read_key (path, "CRVAL3");
  freqMHz = freqMHz*1e-6;
  variable stem   = string_matches(path, "\\(.*\\).fits")[1];

  if ( stat_file (stem+".uvf")==NULL or stat_file (stem+".mod")==NULL)
  {
    vmessage("ERROR %s: required files for restoring the beam not found:",_function_name());
    vmessage("\t%s",stem+".uvf");
    vmessage("\t%s",stem+".mod");
    return NULL;
  };

  variable orig_major  = fits_read_key(path, "BMAJ")*(3.6e+6)/2.0; % clean x beam major axis in mas (semi major!)
  variable orig_minor  = fits_read_key(path, "BMIN")*(3.6e+6)/2.0; % clean x beam minor axis in mas (semi minor!)
  variable orig_posang = fits_read_key(path, "BPA");

  variable outname = qualifier("outname", sprintf("%dMHz_restore.fits", nint(freqMHz)));

  variable write_it = 1;
  if (stat_file(outname) != NULL) 
  {
    write_it = 0;
    vmessage("WARNING %s: outfile %s already exists",_function_name(),outname);
  }
  if (write_it == 0 and qualifier_exists("overwrite"))
  {
    write_it = 1; vmessage("overwriting %s",outname);
  }
  if (write_it == 0) vmessage("Using existing file. For a new file use the qualifier \"overwrite\".");
  
  if (write_it == 1)
  {
    %% DIFMAP uses full axis, not semi axis,
    %% DIFMAP seems to set the xsize and ysize to the half, when using wmap (saved image is smaller by 0.5 in both dimensions)
    variable difscript = `device /NULL
obs `+stem+`.uvf
rmod `+stem+`.mod
maps `+sprintf(`%f, %f, %f, %f`,2*xsize, xstep, 2*ysize, ystep)+`
uvw 0,-1
`+sprintf(`select %s`, chan)+`
uvtaper `+sprintf(`%s`, uvtaper)+`
restore `+sprintf(`%f,%f,%f`,2*major, 2*minor, posang)+`
wmap `+outname+`
quit`;
    ()=system("echo \'"+difscript+"\' | difmap > /dev/null 2>&1");
    try { ()=fits_read_key (outname, "NAXIS1"); }
    catch AnyError: { vmessage("DIFMAP failed to create an [%d,%d] image with a resolution of %f x %f mas",
			      xsize,ysize,xstep,ystep); return NULL;}
    ; % ; to terminate try block
  }
  return outname;
}

%
% Code from src/plot/disk_map.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

private define x_axis_up(dist, theta, phi)
{
   xfig_set_eye(dist, theta, phi, computed_roll_angle(dist, theta, phi, (qualifier_exists("flip") ? -1 : 1)*vector(1,0,0), 90)); %set vector to the axis that should be up side!
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define disk_map()
%!%+
%\function{disk_map}
%\synopsis{Plots the megamaser disk using the output of Mark
%   Reid's Bayesian disk fitting routine.}
%\usage{disk_map(infile);}
%\qualifiers{
%\qualifier{radius: }{size of the plotted disk radius, in mas, default: reference radius from file}
%\qualifier{no_axes: }{flag to turn off the axes}
%\qualifier{no_los: }{flag to turn off the line-of-sight bar}
%\qualifier{no_disk: }{flag to turn off the wireframe disk model}
%\qualifier{no_data: }{flag to turn off the data points}
%\qualifier{no_grid: }{flag to turn off the grid}
%\qualifier{no_color: }{flag to plot everything in greyscale}
%\qualifier{scale: }{flag to scale the size of the data points by SNR}
%\qualifier{reverse_x: }{flag to reverse the x-axis orientation; not used here}
%\qualifier{side: }{see the maser disk along the line-of-sight in the x-y-plane}
%\qualifier{above: }{see the maser disk from above in the x-z-plane}
%\qualifier{projection: }{projects the maser spots in addition to the 3d view to the 2d planes}
%\qualifier{scalen: }{half scale length of the axes; default 1.1*radius of the disk, axes are [-scalen,scalen]}
%\qualifier{scalen_frac: }{distance from ticlabels to axes; default 1.05}
%\qualifier{label_dist: }{distance from labels to axes; default 1.3*radius}
%\qualifier{tics_length: }{length of the tics; default 0.2}
%}
%\description
%    Plots the megamaser disk using the output of Mark
%   Reid's Bayesian disk fitting routine. The input file for
%   this script is simply a text file containing the printout
%   of the Bayesian program. For instance, if the Bayesian
%   executable had the file name "fit_disk_v20*", then the Unix command
%        
%   ./fit_disk_v20* > input.prt
%           
%   would generate the appropriate input file for this script.
%   This script is matched to version 20 of Mark's program.
%   The default view on the disk is at theta=120° and phi=-130°,
%   with the x-axis showing upwards.
%
%   This function is rewritten from the IDL program disk_map.pro
%   from Dom Pesce in the version from July, 22, 2014.
%\example
%   For a 2D plot from above:
%   
%   isis> bla=disk_map("/userdata/data/litzinger/Radio/NGC1194/accel/disk_fitting/fit_disk_v20/jan2615/outjan2715";above);
%   isis> bla.render("disk_map_above.pdf");
%
%   or for a 3D plot with projection:
%
%   isis> bla=disk_map("/userdata/data/litzinger/Radio/NGC1194/accel/disk_fitting/fit_disk_v20/jan2615/outjan2715";projection);
%   isis> bla.render("disk_map.pdf");
%
%   Be careful!: If you plotted the 3d view and would like to plot
%   in 2d again you have to restart isis, as it still would plot in
%   3d view. To set the labels and tics correctly first plot the disk
%   with the default values and then change the values.
%   
%\seealso{xfig_3d_orbit_on_cube}
%!%-
{
   % ;---------------------------------------------------
   % ;----Reading in and organizing the data-------------
   % ;---------------------------------------------------
   variable infile;
   switch(_NARGS)
     { case 1: infile = ();}
     { help(_function_name()); return; }

%   variable a = fopen("/userdata/data/litzinger/Radio/NGC1194/accel/disk_fitting/fit_disk_v20/jan2615/out2715","r"); %open the file created by gfortran with ./model.out > v160714
   variable fp = fopen(infile,"r"); %open the file created by gfortran with ./model.out > v160714
   variable line = fgetslines(fp); %get the lines of the file
   variable x1,x2,x3,x4,x5,x6,x7,x8,x9;
   variable x_0 = 0.0;
   variable y_0 = 0.0;
   variable i_0 = 0.0;
   variable i_1 = 0.0;
   variable i_2 = 0.0;
   variable w_0 = 0.0;
   variable w_1 = 0.0;
   variable w_2 = 0.0;
   variable e = 0.0;
   variable peri_0 = 0.0;
   variable peri_1 = 0.0;
   variable r_ref = 0.0;
     
   %% Get the line after which the parameters are listed
   variable dl = is_substr(line[*],"Best global parameter values"); %search for the line which contains this string
   variable parameter_ref = where(dl>0); %get the line number
   
   %% Get the line where the number of unflagged maser spots is written
   variable nr = is_substr(line[*],"Number of unflagged"); %search for the line which contains this string
   variable nr1 = where(nr>0); %get the line number
   variable d = sscanf(line[nr1[0]],"%s %s %s %s %s %i\n",&x1, &x2, &x3, &x4, &x5, &x6);
   if (d==6)   {      print("all parameter ok");   } else {      print("one parameter missing");   };
   variable number = x6;
   
   %% Get the line where the reference radius is written
   variable ref_rad = is_substr(line[*],"Reference radius for"); %search for the line which contains this string
   variable ref1 = where(ref_rad>0); %get the line number
   d = sscanf(line[ref1[0]],"%s %s %s %s %s %s %f\n",&x1, &x2, &x3, &x4, &x5, &x6, &x7);
   if (d==7){   print("all parameter ok");} else {   print("one parameter missing");};
   r_ref = x7;
   
   variable vel = Double_Type[0];
   variable rad = Double_Type[0];
   variable phi = Double_Type[0];
   variable RA = Double_Type[0];
   variable DEC = Double_Type[0];
   variable y_err = Double_Type[0];
   
   variable values1 = line[parameter_ref+1];
   d = sscanf(values1[0],"%f %f %f %f %f %f\n",&x1, &x2, &x3, &x4, &x5, &x6);
   if (d==6){   print("all parameter ok");} else {   print("one parameter missing");};
   x_0 = x4;
   y_0 = x5;
   i_0 = x6*(PI/180.);
   
   variable values2 = line[parameter_ref+2];
   d = sscanf(values2[0],"%f %f %f %f %f %f\n",&x1, &x2, &x3, &x4, &x5, &x6);
   if (d==6) {  print("all parameter ok");} else {   print("one parameter missing");};
   i_1 = x1*(PI/180.);
   i_2 = x2*(PI/180.);
   w_0 = x3*(PI/180.);
   w_1 = x4*(PI/180.);
   w_2 = x5*(PI/180.);
   e = x6;
   
   variable values3 = line[parameter_ref+3];
   d = sscanf(values3[0],"%f %f %f\n",&x1, &x2, &x3);
   if (d==3){   print("all parameter ok");} else {   print("one parameter missing");};
   peri_0 = x1*(PI/180.);
   peri_1 = x2*(PI/180.);
   
   variable error_floors = is_substr(line[*],"Error floors (x,y,Vsys,Vhv,A)"); %search for the line which contains this string
   variable err1 = where(error_floors>0); %get the line number
   variable k;
   _for k(0,number-1,1)
   {
      variable erro = line[err1+2+k];
      d = sscanf(erro[0],"%f %f %f %f %f %f %f %f %f\n",&x1, &x2, &x3, &x4, &x5, &x6, &x7, &x8, &x9);
      if (d==9){print("all parameter ok");} else { print("one parameter missing");};
      vel = [vel,x2];
      RA = [RA,x4];
      DEC = [DEC,x6];
      y_err = [y_err,x7];
   };
   
   variable rad_phi = is_substr(line[*],"n_f  r(mas) phi(deg)"); %search for the line which contains this string
   variable rad1 = where(rad_phi>0); %get the line number
   _for k(0,number-1,1)
   {
      variable radphi = line[rad1+1+k];
      d = sscanf(radphi[0],"%f %f %f %f %f %f %f\n",&x1, &x2, &x3, &x4, &x5, &x6, &x7);
      if (d==7){print("all parameter ok");} else { print("one parameter missing");};
      rad = [rad,x2];
      phi = [phi,x3*(PI/180.)];
   };

   if (x_0 == 0){ print("No value found for parameter: BH Xo");};
   if (y_0 == 0){ print("No value found for parameter: BH Yo");};
   if (i_0 == 0){ print("No value found for parameter: i");};
   if (i_1 == 0){ print("No value found for parameter: di/dr");};
   if (i_2 == 0){ print("No value found for parameter: d2i/dr2");};
   if (w_0 == 0){ print("No value found for parameter: PA");};
   if (w_1 == 0){ print("No value found for parameter: dPA/dr");};
   if (w_2 == 0){ print("No value found for parameter: d2PA/dr2");};
   if (e == 0){ print("No value found for parameter: eccentricity");};
   if (peri_0 == 0){ print("No value found for parameter: peri azimuth");};
   if (peri_1 == 0){ print("No value found for parameter: dP-az/dr");};
   if (r_ref == 0){ print("No value found for parameter: reference radius");};

   variable index = where(vel > 0);
   vel = vel[index];
   rad = rad[index];
   phi = phi[index];
   RA = RA[index];
   DEC = DEC[index];
   y_err = y_err[index];

   variable i_data = i_0 + (i_1*(rad-r_ref)) + (i_2*((rad-r_ref)^2.));
   variable w_data = w_0 + (w_1*(rad-r_ref)) + (w_2*((rad-r_ref)^2.));

   variable x_data = RA - x_0;
   variable y_data = DEC - y_0;
   variable z_data = rad*cos(phi)*sin(i_data);

   index = array_sort(vel);
   vel = vel[index];
      
   variable type = String_Type[length(x_data)];
   type[0] = "b";
   variable str = ["b","s","r"];
   variable count = 0;

   variable n;
   _for n(1,length(vel)-1,1)
   {
      if (abs(vel[n]-vel[n-1]) > 100.)
        {
           count = count+1;
           if (count > 2)
             {
                count = 2;
             }
        }
           type[n] = str[count];
   };
   
   x_data = x_data[index];
   y_data = y_data[index];
   z_data = z_data[index];

   %---------------------------------------------------
   %------------Creating the warped disk---------------
   %---------------------------------------------------
   number = 50;
   if (qualifier_exists("radius"))
       { variable radius = qualifier("radius");   }
   else { radius = 1.0*max(abs(rad));       };       
   
   variable r = ([1:8:1])*(radius/8.);
   variable ecc_anom = [0:number+1:1]*(2.*PI/number);
   
   variable psi = peri_0 + (peri_1*r);
   
   variable X = Double_Type[length(r),length(ecc_anom)];
   variable Y = Double_Type[length(r),length(ecc_anom)];
   variable Z = Double_Type[length(r),length(ecc_anom)];
   
   %variable k;
   _for n(0,length(r)-1,1)
   {
      variable a = r[n]/(1. - e);
   
      variable i_now = i_0 + (i_1*(r[n]-r_ref)) + (i_2*((r[n]-r_ref)^2.));
      variable omega_now = w_0 + (w_1*(r[n]-r_ref)) + (w_2*((r[n]-r_ref)^2.));
   
      _for k(0,(length(ecc_anom)/2)-1,1) {
   
         variable rad_now = a*(1. - (e*cos(ecc_anom[k])));
   
         variable cos_phi = (cos(ecc_anom[k]) - e)/(1. - e*cos(ecc_anom[k]));
         variable sin_phi = sqrt(1. - (cos_phi^2.));
   
         variable cosp = (sin_phi*sin(psi[n])) + (cos_phi*cos(psi[n]));
         variable sinp = (sin_phi*cos(psi[n])) - (cos_phi*sin(psi[n]));
   
         X[n,k] = rad_now * ( (sin(omega_now)*cosp) - (cos(omega_now)*cos(i_now)*sinp) );
         Y[n,k] = rad_now * ( (cos(omega_now)*cosp) + (sin(omega_now)*cos(i_now)*sinp) );
         Z[n,k] = -rad_now * ( sin(i_now)*sinp );
      }
        
      _for k((length(ecc_anom)/2),length(ecc_anom)-1,1){
   
         rad_now = a*(1. - (e*cos(ecc_anom[k])));
   
         cos_phi = (cos(ecc_anom[k]) - e)/(1. - e*cos(ecc_anom[k]));
         sin_phi = -sqrt(1. - (cos_phi^2.));
   
         cosp = (sin_phi*sin(psi[n])) + (cos_phi*cos(psi[n]));
         sinp = (sin_phi*cos(psi[n])) - (cos_phi*sin(psi[n]));
   
         X[n,k] = rad_now * ( (sin(omega_now)*cosp) - (cos(omega_now)*cos(i_now)*sinp) );
         Y[n,k] = rad_now * ( (cos(omega_now)*cosp) + (sin(omega_now)*cos(i_now)*sinp) );
         Z[n,k] = -rad_now * ( sin(i_now)*sinp );
      };
   };


   %---------------------------------------------------
   %-------------------Plotting------------------------
   %---------------------------------------------------
   variable W=10, H=10; %effects only the 2d plots side and above
   variable xfig = xfig_plot_new(W,H);
   variable i,j;
   variable xlabel = ("East offset (mas)");
   variable ylabel = ("North offset (mas)");
   variable zlabel = ("Z offset (mas)");
   variable scalen = qualifier("scalen",nint(ceil(1.1*radius))); %scale length for the axis
   variable scalen_frac = qualifier("scalen_frac",1.15/1.1); %distance from ticlabels to axis
   variable label_dist = qualifier("label_dist",1.3*radius); %distance from labels to axis
   variable tics_length = qualifier("tics_length",0.2); %length of the tics
   
   if (qualifier_exists("side")) %plots the maser spots in the x-y-plane 
     {
	xfig = xfig_plot_new(W,H);
	xfig.world(scalen,-scalen,-scalen,scalen);
	xfig.xlabel("x offset (mas)");
	xfig.ylabel("y offset (mas)");
	_for j(0,length(x_data)-1,1)
	  {
	     variable theta = [1:21:1]*(2.*PI/20.);
	     phi = [1:21:1]*(PI/20.);
	     rad = 0.01*radius;

	     _for n(0,length(theta)-1,1)
	       {
		  variable x_sym = rad*sin(phi[n])*cos(theta[n]);
                  variable y_sym = rad*sin(phi[n])*sin(theta[n]);
 	       }
	     if (type[j] == "b")
	       {variable colo=sprintf("#%02X%02X%02X",0,0,255);   }
	     if (type[j] == "s")
	       {        colo=sprintf("#%02X%02X%02X",0,150,50);   }
	     if (type[j] == "r")
	       {        colo=sprintf("#%02X%02X%02X",255,0,0);     }

	     variable new_phi=w_0+i_0; %rotate the disk so that it is at 0 north offset and tilt it with the inclination angle
             xfig.plot((x_sym+x_data[j])*cos(new_phi)-(y_sym+y_data[j])*sin(new_phi),(x_sym+x_data[j])*sin(new_phi)+(y_sym+y_data[j])*cos(new_phi);sym="circle",size=0.7,color="black",fill=20,fillcolor=colo);

	     _for n(0,length(r)-1,1)
	       {
		  xfig.plot(X[n,*]*cos(new_phi)-Y[n,*]*sin(new_phi),X[n,*]*sin(new_phi)+Y[n,*]*cos(new_phi));
	       }
	     variable dring = 2;
	     _for n(0,int(number/dring)-1,1)
	       {
		  xfig.plot(X[*,n*dring]*cos(new_phi)-Y[*,n*dring]*sin(new_phi),X[*,n*dring]*sin(new_phi)+Y[*,n*dring]*cos(new_phi);size=0.3);
               }
          }
     }
   else if (qualifier_exists("above")) %plots the maser spots as seen from above
     {
	xfig = xfig_plot_new(W,H);
	xfig.world(scalen,-scalen,scalen,-scalen);
	xfig.xlabel("x offset (mas)");
	xfig.ylabel("z offset (mas)");
	xfig.plot([-radius,radius],[0,0];line=0,width=3); %midline

	_for j(0,length(x_data)-1,1)
	  {
	     theta = [1:21:1]*(2.*PI/20.);
	     phi = [1:21:1]*(PI/20.);
	     rad = 0.01*radius;

	     _for n(0,length(theta)-1,1)
	       {
		  x_sym = rad*sin(phi[n])*cos(theta[n]);
                  y_sym = rad*sin(phi[n])*sin(theta[n]);
                  variable z_sym = rad*cos(phi[n]);
	       }
	     if (type[j] == "b")
	       { colo=sprintf("#%02X%02X%02X",0,0,255);     }
	     if (type[j] == "s")
	       { colo=sprintf("#%02X%02X%02X",0,150,50);    }
	     if (type[j] == "r")
	       { colo=sprintf("#%02X%02X%02X",255,0,0);     }

	     new_phi=w_0+i_0; %rotate the disk so that it is at 0 north offset and tilt it with the inclination angle
	     xfig.plot((x_sym+x_data[j])*cos(new_phi)-(y_sym+y_data[j])*sin(new_phi),(z_sym+z_data[j]);sym="circle",size=0.7,color="black",fill=20,fillcolor=colo);
	     xfig.plot(x_0*cos(new_phi)-y_0*sin(new_phi),0;sym="circle",color="black",fill=20,fillcolor="black",size=0.7);

	     _for n(0,length(r)-1,length(r)-1)
	       { xfig.plot(X[n,*]*cos(new_phi)-Y[n,*]*sin(new_phi),Z[n,*];line=5);     }
	     }
	}
   
   else  %plots the maser spots in 3d
     {
   	xfig_set_focus(vector(0,0,0));
        x_axis_up(1e5, 120, -130 );% theta is rotating along the y-axis
   
      if (qualifier_exists("no_axes"))
     { }
   else
     {
	xfig.add_object(xfig_new_polyline([-scalen,scalen],[scalen,scalen],[-scalen,-scalen]));%xaxis
	xfig.add_object(xfig_new_polyline([-scalen,-scalen],[-scalen,scalen],[scalen,scalen])); %yaxis
        xfig.add_object(xfig_new_polyline([-scalen,-scalen],[scalen,scalen],[-scalen,scalen])); %zaxis 
        xfig.add_object( xfig_new_text(xlabel; x0=0, y0=label_dist, z0=-label_dist, just=[0,0], rotate=-90)); % x-label
        xfig.add_object( xfig_new_text(ylabel; x0=-label_dist, y0=0, z0=label_dist, just=[0,0], rotate=-37)); % y-label
        xfig.add_object( xfig_new_text(zlabel; x0=-label_dist, y0=label_dist, z0=0, just=[0,0], rotate=25)); % z-label
	   if (qualifier_exists("no_grid"))
     {
	_for i(-scalen,scalen,1) %xaxis tics
	  { xfig.add_object(xfig_new_polyline([i,i],[scalen,scalen+tics_length],[-scalen,-scalen]));
	    xfig.add_object( xfig_new_text(sprintf("%i",i); x0=i, y0=scalen_frac*scalen, z0=-scalen_frac*scalen, just=[0.5,0], rotate=0 )); %xaxis
	  }
	_for j(-scalen,scalen,1) %yaxis tics
	  {  xfig.add_object(xfig_new_polyline([-scalen,-scalen],[j,j],[scalen,scalen+tics_length]));
	     xfig.add_object( xfig_new_text(sprintf("%i",j); x0=-scalen_frac*scalen, y0=j, z0=scalen_frac*scalen, just=[0.5,0.5], rotate=0 )); %yaxis
	  }
	_for k(-scalen,scalen,1)%zaxis tics
	  {  xfig.add_object(xfig_new_polyline([-scalen,-scalen],[scalen,scalen+tics_length],[k,k]));
	     xfig.add_object( xfig_new_text(sprintf("%i",k); x0=-scalen_frac*scalen, y0=scalen_frac*scalen, z0=k, just=[0,0], rotate=0 ));  %zaxis
	  }
     }
   else 
     {
	_for i(-scalen,scalen,1)
	  {
	     xfig.add_object(xfig_new_polyline([i,i],[-scalen,scalen],[-scalen,-scalen];color="lightgrey"));%xaxis grid
	     xfig.add_object(xfig_new_polyline([i,i],[-scalen,-scalen],[-scalen,scalen];color="lightgrey"));%xaxis grid
	     xfig.add_object( xfig_new_text(sprintf("%i",i); x0=i, y0=scalen_frac*scalen, z0=-scalen_frac*scalen, just=[0.5,0], rotate=0 )); %xaxis
	     xfig.add_object(xfig_new_polyline([i,i],[scalen,scalen+tics_length],[-scalen,-scalen]));%xaxis tics
	  }
	_for j(-scalen,scalen,1)
	  {
	     xfig.add_object(xfig_new_polyline([-scalen,-scalen],[j,j],[-scalen,scalen];color="lightgrey"));%yaxis grid
             xfig.add_object(xfig_new_polyline([-scalen,scalen],[j,j],[-scalen,-scalen];color="lightgrey"));%yaxis grid
             xfig.add_object( xfig_new_text(sprintf("%i",j); x0=-scalen_frac*scalen, y0=j, z0=scalen_frac*scalen, just=[0.5,0.5], rotate=0 )); %yaxis
             xfig.add_object(xfig_new_polyline([-scalen,-scalen],[j,j],[scalen,scalen+tics_length]));%yaxis tics
	  }
	_for k(-scalen,scalen,1)
	  {
             xfig.add_object(xfig_new_polyline([-scalen,-scalen],[-scalen,scalen],[k,k];color="lightgrey"));%zaxis grid
             xfig.add_object(xfig_new_polyline([-scalen,scalen],[-scalen,-scalen],[k,k];color="lightgrey"));%zaxis grid
             xfig.add_object( xfig_new_text(sprintf("%i",k); x0=-scalen_frac*scalen, y0=scalen_frac*scalen, z0=k, just=[0,0], rotate=0 ));  %zaxis
             xfig.add_object(xfig_new_polyline([-scalen,-scalen],[scalen,scalen+tics_length],[k,k]));%zaxis tics
	  }
     }
     };
   
   if (qualifier_exists("no_los"))
     {}
   else { xfig.add_object(xfig_new_polyline([0.,0.],[0,0],[0.,1.1*radius];color="blue",width=3)); }%los

   variable scaling = min(y_err)/y_err;

   if (qualifier_exists("no_data"))
     {}
   else 
     {
     	_for j(0,length(x_data)-1,1)
        {
           theta = [1:21:1]*(2.*PI/20.);
           phi = [1:21:1]*(PI/20.);
	   if (qualifier_exists("scale"))
	     { rad = 0.015*radius*scaling[j]; }
	   else{ rad = 0.01*radius;}
        
           _for n(0,length(theta)-1,1)
             {
		x_sym = rad*sin(phi[n])*cos(theta[n]);
                y_sym = rad*sin(phi[n])*sin(theta[n]);
                z_sym = rad*cos(phi[n]);
             }
           if (type[j] == "b")
             {
                colo=sprintf("#%02X%02X%02X",0,0,255);
             }
           if (type[j] == "s")
             {
                colo=sprintf("#%02X%02X%02X",0,150,50);
             }
           if (type[j] == "r")
             {
                colo=sprintf("#%02X%02X%02X",255,0,0);
             }
          xfig.add_object(xfig_new_text(`\LARGE .`;z0=z_sym+z_data[j],x0=(x_sym+x_data[j]),y0=y_sym+y_data[j],color=colo,depth=1));
        
           _for n(0,length(r)-1,1)
             {
                xfig.add_object(xfig_new_polyline(X[n,*],Y[n,*],Z[n,*];depth=2));
             }
           dring = 2;
           _for n(0,int(number/dring)-1,1)
             {
                xfig.add_object(xfig_new_polyline(X[*,n*dring],Y[*,n*dring],Z[*,n*dring];depth=2));
             }
	   if (qualifier_exists("projection"))
	     {
%		xfig.add_object(xfig_new_text(`\LARGE .`;z0=z_sym+z_data[j],x0=(x_sym+x_data[j]),y0=-scalen,color=colo,depth=1)); %z-east-plane
		xfig.add_object(xfig_new_text(`\LARGE .`;z0=z_sym+z_data[j],x0=-scalen,y0=y_sym+y_data[j],color=colo,depth=1));   %z-north-plane
		xfig.add_object(xfig_new_polyline([-scalen,-scalen],[-radius,radius],[0,0];line=0,width=1)); %midline
        	     _for n(0,length(r)-1,length(r)-1)
	       { xfig.add_object(xfig_new_polyline(-scalen,Y[n,*],Z[n,*];line=5)); %projected inner and outer Kepler curves on z-north-plane
	       }
		xfig.add_object(xfig_new_text(`\LARGE .`;z0=-scalen,x0=(x_sym+x_data[j]),y0=y_sym+y_data[j],color=colo,depth=1)); %north-east-plane
	     }
        }
     }
     }
   
   return xfig; 
}

#endif

%
% Code from src/fitting/fit-functions/diskline2.sl
%

% only activate if module xspec could be loaded
#ifeval __get_reference("load_xspec_local_models")!=NULL


%%%%%%%%%%%%%%%%%%%%
define diskline2_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{diskline2}
%\synopsis{describes a line emission from a relativistic accretion disk}
%\description
%    norm    = photons/cm**2/s in the spectrum\n
%    LineE   = line energy\n
%    Betor10 = power law dependence of emissivity.
%              If this parameter is 10 or greater then the accretion
%              disk emissivity law (1-sqrt(6/R))/R**3 is used.
%              Otherwise the emissivity scales as R**par2.\n
%    Rin     = inner radius (units of GM/c**2)\n
%    Rout    = outer radius (units of GM/c**2)\n
%    Incl    = inclination (degrees)\n
%    (The original diskline model uses Rin(M) and Rout(M).)
%\seealso{diskline / Fabian et al., MNRAS 238, 729.}
%!%-
{
  return diskline_fit(bin_lo, bin_hi, par);
}

add_slang_function("diskline2", ["norm", "LineE [keV]", "Betor10", "Rin", "Rout", "Incl [deg]"]);

#endif

%
% Code from src/plot/distribution_matrix.sl
%

% only activate if module gcontour could be loaded
#ifeval __get_reference("gcontour_compute")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL
% -*- mode: slang; mode: fold; -*- %


define hist1d_confidence (lo, hi, h)
%!%+
%\function{hist1d_confidence}
%#c%{{{
%\synopsis{Find the confidence interval for histogram}
%\usage{(Min,Max) = hist1d_confidence(lo, hi, hist);}
%\qualifiers{
%\qualifier{conf}{[=0.9] confidence limit (between 0 and 1)}
%\qualifier{index}{if given, return the indices of lo and hi bin}
%}
%\description
%    Find the bin boundary values such that they enclose the smallest volume
%    equal to \code{conf}. The result is guaranteed to have at least \code{conf}
%    integral. Works only for unimodal dsitributions.
%!%-
{
  variable conf = _min(1.0, _max(0.68, qualifier("conf", 0.9)));
  variable i,j,ii,jj,s,t;

  s = [0,cumsum(h*1./sum(h))];

  i = wherefirst(s>conf);
  j=0;
  ii=i;
  jj=j;

  t = s[i]-s[j];

  while (i<(length(s)-1))
  {
    i++;
    j++;

    if ((s[i]-s[j])>=t)
    {
      ii=i;
      while (j<i)
      {
	if ((s[i]-s[j])<conf)
	  break;
	j++;
      }
      jj=j;
      t = s[i]-s[j];
    }
  }
  if (jj>0) jj--;
  if (ii>=length(hi)) ii = length(hi)-1;

  if (qualifier_exists("index"))
    return jj,ii;
  return lo[jj], hi[ii];
}
%}}}

define hist2d_confidence (x, y, img)
%!%+
%\function{hist2d_confidence}
%#c%{{{
%!%-
{
#iffalse
  variable conf = _min(1.0, _max(0.0, qualifier("conf", 0.9)));

  variable dim = array_shape(img);
  variable wmax = wherefirstmax(img);

  variable l,r,t,b;

  l = wmax, r = wmax, t = wmax, b = wmax;

  variable i;
  variable r = gcontour_compute(img+imin, [s])[0];
  variable x0 = x[0], xm = (x[-1]-x0)/dim[0];
  variable y0 = y[0], ym = (y[-1]-y0)/dim[1];

  _for i (0, length(r.x_list)-1)
  {
    r.x_list[i] = r.x_list[i]*xm+x0;
    r.y_list[i] = r.y_list[i]*ym+y0;
  }

  return r;
#endif
}
%}}}

private define distM_array (dist) %{{{
{
  variable i,j;

  variable rnames = {"pairs", "hist2d", "hist1d", "grids"};
  if (qualifier_exists("conf"))
  {
    list_append(rnames, "conf1d");
    list_append(rnames, "conf2d");
  }
  variable r = @Struct_Type(list_to_array(rnames));
  variable conf = qualifier("conf", 0.9);
  if (conf == NULL)
    conf = 0.9;

  r.grids = Array_Type[length(dist)];
  _for i (0, length(dist)-1)
    r.grids[i] = qualifier(sprintf("grid%d", i),
	[qualifier(sprintf("grid%dmin", i), min(dist[i])):qualifier(sprintf("grid%dmax",i), max(dist[i])):#qualifier("n", 50)]);

  r.hist1d = Array_Type[length(dist)];
  if (qualifier_exists("conf"))
    r.conf1d = Double_Type[length(dist), 2];

  _for i (0, length(dist)-1)
  {
    r.hist1d[i] = histogram(dist[i], r.grids[i][[:-2]], r.grids[i][[1:]]);
    r.hist1d[i] *= 1./length(dist[i]);
    if (qualifier_exists("conf"))
    {
      r.conf1d[i, *] = [hist1d_confidence(r.grids[i][[:-2]], r.grids[i][[1:]], r.hist1d[i]; conf=conf)];
    }
  }

  r.hist2d = Array_Type[length(dist)-1];
  r.pairs = Array_Type[length(dist)-1];
  if (qualifier_exists("conf"))
    r.conf2d = Array_Type[length(dist)-1];

  _for i (0, length(dist)-2)
  {
    variable x = 2*i; ifnot (x<length(dist)) x = 2*(length(dist)-i-1)+1;
    r.pairs[i] = Array_Type[length(dist)-1-i];
    r.hist2d[i] = Array_Type[length(dist)-1-i];

    if (qualifier_exists("conf"))
      r.conf2d[i] = Struct_Type[length(dist)-1-i];

    _for j (0, length(dist)-2-i)
    {
      variable y = 2*j+1; ifnot (y < length(dist)) y = 2*(length(dist)-1-j);
      r.pairs[i][j] = [x,y];
      r.hist2d[i][j] = histogram2d(dist[x], dist[y], r.grids[x][[:-2]], r.grids[y][[1:]]);
      r.hist2d[i][j] *= 1./length(dist[i]);

      if (qualifier_exists("conf"))
	r.conf2d[i][j] = hist2d_confidence(r.grids[x], r.grids[y], r.hist2d[i][j]; conf=conf);
    }
  }

  return r;
}
%}}}

private define distM_struct (dist) %{{{
{
  variable fields = qualifier("fields", get_struct_field_names(dist));

  variable arr = Array_Type[length(fields)];
  variable i;
  _for i (0, length(arr)-1)
    arr[i] = get_struct_field(dist, fields[i]);

  return distM_array(arr;; __qualifiers());
}
%}}}%

define distribution_matrix (distribution_points)
%!%+
%\function{distribution_triangle}
%#c%{{{
%\synopsis{Calculate 1D and 2D probability distributions}
%\usage{Struct_Type dm = distribution_matrix(Array_Type/List_Type v);
%\altusage{dm = distribution_matrix(Struct_Type v);}}
%\qualifiers{
%\qualifier{gridX}{histogram grid for dimension X where X starts counting at 0}
%\qualifier{gridXmin}{[=min(vX)] bottom boundary for dimension X grid}
%\qualifier{gridXmax}{[=max(vX)] top boundary for dimension X grid}
%\qualifier{n}{number of grid points (only relevant for min/max)}
%\qualifier{fields}{Use only struct fields given in this name array (only
%  relevant if input is a struct)}
%}
%\description
%    Calculates the probability distributions from a given high dimensional
%    set of state vectors (think MCMC results). The resulting structure
%    can be used with \code{xfig_plot_distribution_matrix} to plot the distributions
%    in a projection like triangle matrix.
%
%\seealso{xfig_plot_distribution_matrix}
%!%-
{
  if (Struct_Type == typeof(distribution_points))
    return distM_struct(distribution_points;; __qualifiers());
  else
    return distM_array(distribution_points;; __qualifiers());
}
%}}}

private define xfig_prepare_dm_xpp (dm, tile_size) %{{{
{
  variable xpp = {};
  variable i,j,x,y;

  _for i (0, length(dm.pairs)-1)
  {
    list_append(xpp, {});
    _for j (0, length(dm.pairs[i])-1)
    {
      list_append(xpp[-1], xfig_plot_new(tile_size,tile_size));
      x = dm.pairs[i][j][0];
      y = dm.pairs[i][j][1];
      xpp[-1][-1].xaxis(;; qualifier_exists("axis") ? qualifier("axis")[x] : qualifier(sprintf("axis%d", x)));
      xpp[-1][-1].yaxis(;; qualifier_exists("axis") ? qualifier("axis")[y] : qualifier(sprintf("axis%d", y)));
      xpp[-1][-1].axis(; off);
      xpp[-1][-1].world(min(dm.grids[x]), max(dm.grids[x]), min(dm.grids[y]), max(dm.grids[y]));
      xpp[-1][-1].plot_png(-transpose(dm.hist2d[i][j]); cmap=qualifier("cmap", "gray"));
      if (qualifier_exists("mark"))
	xpp[-1][-1].plot(qualifier("mark")[x], qualifier("mark")[y]; sym="x", color=qualifier("color", "red"));
    }
  }

  return xpp;
}
%}}}

private define xfig_prepare_dm_ypp (dm, tile_size) %{{{
{
  variable i,x,lo,hi;

  variable ypp = {};
  variable ymax = 0.0;
  _for i (0, length(dm.pairs)-1)
  {
    x = dm.pairs[i][0][0];
    ymax = _max(ymax, max(dm.hist1d[x]));
  }

  _for i (0, length(dm.pairs)-1)
  {
    list_append(ypp, xfig_plot_new(tile_size,tile_size));
    x = dm.pairs[i][0][0];
    ypp[-1].x2axis(;; qualifier_exists("axis") ? qualifier("axis")[x] : qualifier(sprintf("axis%d", x)));
    ypp[-1].xaxis(;off);
    ypp[-1].yaxis(; off);
    ypp[-1].world(min(dm.grids[x]), max(dm.grids[x]),
		  qualifier("y1min", qualifier("ymin", 0)),
		  qualifier("y1max", qualifier("ymax", ymax)));
    ypp[-1].hplot(dm.grids[x], dm.hist1d[x]);
    if (qualifier_exists("mark"))
      ypp[-1].plot(qualifier("mark")[x]+[0,0], [0,1]; world10, color=qualifier("color", "red"));
    if (qualifier_exists("conf"))
    {
      ifnot (struct_field_exists(dm, "conf1d"))
	(lo,hi) = hist1d_confidence(dm.grids[x][[:-2]], dm.grids[x][[1:]], dm.hist1d[x]);
      else
	(lo, hi) = (dm.conf1d[x][0],dm.conf1d[x][1]);
      ypp[-1].plot([lo, lo], [0,1]; world10, line=1);
      ypp[-1].plot([hi, hi], [0,1]; world10, line=1);
    }
  }

  return ypp;
}
%}}}

private define xfig_prepare_dm_zpp (dm, tile_size) %{{{
{
  variable zpp = {};
  variable zmax = 0.0;
  variable j,y,lo,hi;
  _for j (0, length(dm.pairs[0])-1)
  {
    y = dm.pairs[0][j][1];
    zmax = _max(zmax, max(dm.hist1d[y]));
  }

  _for j (0, length(dm.pairs[0])-1)
  {
    list_append(zpp, xfig_plot_new(tile_size,tile_size));
    y = dm.pairs[0][j][1];
    zpp[-1].yaxis(;; qualifier_exists("axis") ? qualifier("axis")[y] : qualifier(sprintf("axis%d", y)));
    zpp[-1].yaxis(; off);
    zpp[-1].xaxis(; off);
    zpp[-1].world(qualifier("y2max", qualifier("ymax", zmax)),
		  qualifier("y2min", qualifier("ymin", 0)),
		  min(dm.grids[y]), max(dm.grids[y]));
    zpp[-1].plot(dm.hist1d[y][[0:2*length(dm.hist1d[y])-1]/2],
		 dm.grids[y][[1:2*(length(dm.grids[y])-1)]/2]);
    if (qualifier_exists("mark"))
      zpp[-1].plot([0,1], qualifier("mark")[y]+[0,0]; world01, color=qualifier("color", "red"));
    if (qualifier_exists("conf"))
    {
      ifnot (struct_field_exists(dm, "conf1d"))
	(lo,hi) = hist1d_confidence(dm.grids[y][[:-2]], dm.grids[y][[1:]], dm.hist1d[y]);
      else
	(lo,hi) = (dm.conf1d[y][0], dm.conf1d[y][1]);
      zpp[-1].plot([0,1], [lo, lo]; world01, line=1);
      zpp[-1].plot([0,1], [hi, hi]; world01, line=1);
    }
  }

  return zpp;
}
%}}}

private define xfig_prepare_dm (dm) %{{{
{
  variable tile_size = qualifier("tile_size", qualifier("size", 20)*1./(length(dm.grids)-1));

  variable xpp = xfig_prepare_dm_xpp(dm, tile_size;; __qualifiers());
  variable ypp = xfig_prepare_dm_ypp(dm, tile_size;; __qualifiers());
  variable zpp = xfig_prepare_dm_zpp(dm, tile_size;; __qualifiers());

  variable i;

  variable blank = xfig_plot_new(tile_size,tile_size);
  blank.axis(; off);
  variable Y = xfig_new_hbox_compound(blank, __push_list(ypp));
  variable Z = xfig_new_vbox_compound(__push_list(zpp));
  _for i (0, length(dm.pairs)-1)
    Z = xfig_new_hbox_compound(Z, xfig_new_vbox_compound(__push_list(xpp[i])));

  variable X = xfig_new_vbox_compound(Y, Z);

  return X, xpp, ypp, zpp;
}
%}}}

define xfig_plot_distribution_matrix ()
%!%+
%\function{xfig_plot_distribution_matrix}
%#c%{{{
%\synopsis{Plot distribution matrix}
%\usage{xfig_plot_distribution_matrix(Struct_Type dm[, String_Type file]);}
%\qualifiers{
%\qualifier{names}{string array containting the names for all dimensions}
%\qualifier{nameX}{add label for dimension X}
%}
%\description
%    Plot the distribution matrix to file \code{file} using xfig. If 
%    \code{file} is not given or is \code{NULL} return instead the xfig
%    object.
%
%\example
%    variable dm = distribution_matrix({v0, v1, v2}); % vX are the state values in the dimensions
%    variable names = ["dim0", "dim1", "dim2"];
%    xfig_plot_distribution_matrix(dm, "distribution.pdf"; names=names, name1="other label");
%
%\seealso{distribution_matrix}
%!%-
{ 
  variable dm, file = NULL;

  switch (_NARGS)
  { case 1: dm = (); }
  { (dm, file) = (); }
  variable X, xpp, ypp, zpp, i, j;
  (X, xpp, ypp, zpp) = xfig_prepare_dm(dm;; __qualifiers());
  variable y1label_x = _Inf;

  _for i (0, length(dm.pairs)-1)
  {
    ypp[i].x2axis(; on);
    if (qualifier_exists(sprintf("labels%d", dm.pairs[i][0][0])) || qualifier_exists("labels"))
      ypp[i].x2label(qualifier(sprintf("labels%d", dm.pairs[i][0][0]), qualifier("labels")[dm.pairs[i][0][0]]));
  }

  _for j (0, length(dm.pairs[0])-1)
  {
    zpp[j].y1axis(; on);
    if (qualifier_exists(sprintf("labels%d", dm.pairs[0][j][1])) || qualifier_exists("labels"))
      zpp[j].ylabel(qualifier(sprintf("labels%d", dm.pairs[0][j][1]), qualifier("labels")[dm.pairs[0][j][1]]));
    if (zpp[j].plot_data.y1axis.axis_label != NULL)
      y1label_x = _min(y1label_x, zpp[j].plot_data.y1axis.axis_label.X.x);
  }

  _for j (0, length(zpp)-1)
  {
    if (zpp[j].plot_data.y1axis.axis_label != NULL)
      zpp[j].plot_data.y1axis.axis_label.X.x = y1label_x;
  }

  if (NULL != file)
    X.render(file);
  else
    return X;
}
%}}}

#endif

#endif

%
% Code from src/slang/astronomy/dms2deg.sl
%
define dms2deg(d,m,s) {
%!%+
%\function{dms2deg}
%\synopsis{Convert angle in degree, minute, seconds to floating points degrees or radian}
%\usage{degree = dms2deg(d,m,s);}
%\qualifiers{
%\qualifier{hours}{If set, the coordinate is given in hour, minutes, sec (RA)}
%\qualifier{radian}{If set, return angle in radians, not degrees}
%}
%\description
%This is a convenience routine to convert astronomical coordinates given
%in degrees, minutes, and seconds into a floating point number.
%The sign of the highest non-zero argument decides the sign of the
%returned floating point number. In other words: The routine tries to 
%be intelligent for negative declinations in the 0>=dec>-1 strip. 
%For example, if the declination is -0d 14' 30.2", call this routine
%as dms2deg(0,-14,30.2).
%
%This routine is array safe (as long as d, m, s are arrays of equal length).
%\example
% % Convert the coordinates 
% % alpha=19h 49m 35.49s and dec=-30d 12' 31.8"
% variable ra=hms2deg(19,49,35.49);
% variable dec=dms2deg(-30,12,31.8);
%
%\seealso{hms2deg}
%!%-
   variable sgn;
   if (typeof(d)==Array_Type || typeof(m)==Array_Type || typeof(s)==Array_Type) {
       if (length(d)!=length(m) || length(m)!=length(s)) {
           throw UsageError("%s: d,m,s arguments must be of equal length\n",_function_name());
       }
       sgn=Double_Type[length(d)];
       variable ndx,cmplndx;
       ndx=where(d!=0.,&cmplndx);
       if (length(ndx)>0) {
           % everything not close to the equator
           sgn[ndx]=sign(d[ndx]);
       }
       if (length(cmplndx)>0) {
           % equator
           variable ndx2,cmplndx2;
           ndx2=where(m[cmplndx]!=0.,&cmplndx2);
           if (length(ndx2)>0) {
               sgn[cmplndx[ndx2]]=sign(m[cmplndx[ndx2]]);
           }
           if (length(cmplndx2)>0) {
               % d, m are 0
               variable ndx3,cmplndx3;
               ndx3=where(s[cmplndx[cmplndx2]]<0.,&cmplndx3);
               if (length(ndx3)!=0) {
                   sgn[cmplndx[cmplndx2[ndx3]]]=-1.;
               }
               if (length(cmplndx3)>0) {
                   sgn[cmplndx[cmplndx2[cmplndx3]]]=+1.;
               }
           }
       }
   } else {
       % case that d, m, s are not arrays
       if (d!=0.) {
           sgn=sign(d);
       } else {
           if (m!=0.) {
               sgn=sign(m);
           } else {
               sgn=sign(s);
           }
       }
   }

   variable ret = sgn*(abs(d)+(abs(m)+abs(s)/60.)/60.);

   if (qualifier_exists("hours")) {
       ret*=15.;
   }

   if (qualifier_exists("radian")) {
       ret*=PI/180.;
   }

   return ret;
}

define hms2deg(h,m,s) {
%!%+
%\function{hms2deg}
%\synopsis{Convert angle in hour, minute, seconds to degrees or radian}
%\usage{degree = hms2deg(h,m,s);}
%\qualifiers{
%\qualifier{radian}{If set, return angle in radians, not degrees}
%}
%\description
%This is a convenience routine to convert astronomical coordinates given
%in hours, minutes, and seconds into a floating point number.
%
%The routine is equivalent to calling dms2deg with the hours qualifier.
%
%This routine is array safe (as long as h, m, s are arrays of equal length).
%
%\seealso{dms2deg}
%!%-
    if (qualifier_exists("radian")) {
	return dms2deg(h,m,s;hours,radian);
    } else {
	return dms2deg(h,m,s;hours);
    }
}

%
% Code from src/data/dof.sl
%
%%%%%%%%%%%%%%%%
define dof()
%%%%%%%%%%%%%%%%
%!%+
%\function{dof}
%
%\synopsis{Number of degrees of freedom}
%\usage{Double_Type = dof (hist_index);}
%\description
%       Use this function to retrieve number of
%       degrees of freedom.
%\example
%       isis>xray = load_data("data.pha");
%       isis>variable num = dof(1);
%
%\seealso{num_bin}
%!%-
{
  variable dset;
  
  switch(_NARGS)
  { case 1: dset  = (); }
  { help(_function_name()); return; }

  return int(num_bin(dset; dof));
}
%
% Code from src/fitting/fit-functions/pulsar/dopplerorbit.sl
%
%%%%%%%%%%%%%%%%%%%%%
private define dopplerorbit_fit(lo, hi, par)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{dopplerorbit}
%\synopsis{fit-function for the orbital Doppler shift factor}
%\usage{fit_fun("dopplerorbit(1; qualifiers) * ...");}
%\qualifiers{
%    \qualifier{t90}{switch to the time of mean longitude 90 degrees (see below)}
%    \qualifier{K}{switch to velocity semi amplitude (see below)}
%    \qualifier{metacor}{string array of structure-field-paths inside the meta-
%             data of the current dataset, which will be assumed to
%             be time in MJD and corrected for binary motion (default:
%             ["time"]). Set to NULL to disable any correction.}
%}
%\description
%    Calculates the radial velocoity, v_rad, of a star in a binary and
%    returns the corresponding Doppler factor, beta, given by
%      beta = (1 + V_rad/Const_c)
%    assuming that V_rad/Const_c << 1. The grid of the data (lo,hi)
%    the model is fitted to needs has to be the time in MJD.
%
%    The parameters of the fit-function are:
%      asini - semi major axis (lt-s)
%           or velocity semi amplitude (km/s)
%      porb  - orbital period (days)
%      tau   - time of periastron passage (MJD)
%           or time of mean longitude 90 degrees (MJD)
%      ecc   - eccentricity
%      omega - longitude of periastron (degrees)
%      v0    - systemic velocity (km/s)
%
%    In case the 't90'-qualifier is set, the parameter 'tau' has the
%    meaning of the time of mean longitude of 90 degrees. In case the
%    'K'-qualifiers is set, the parameter 'asini' has the meaning of
%    the velocity semi amplitude.
%
%    The influence of the binary motion is removed from the input
%    time-array 'lo' using 'BinaryCor'. The corrected times are saved
%    as 'binarytime' into the fitfun_cache of the fit-function. This
%    can be used by other fit-functions, like 'pulsartaylor'.
%
%    The fit-function also corrects time arrays inside the metadata of
%    the current dataset (Isis_Active_Dataset). The name of the fields
%    are set by the 'metacor'-qualifier, which is an array of strings
%    giving the "path" to the field. By default, the 'time'-field
%    inside the metadata is corrected in order to simplify the usage
%    of, e.g., 'pulsartorque'. The corrected times are save into the
%    'metacor'-struct inside the cache using the same "path".
%\seealso{radial_velocity_binary, BinaryCor, KeplerEquation}
%!%-
{
  % input parameters
  variable asini = par[0], porb = par[1], tau = par[2],
           ecc = par[3], omega = par[4], v0 = par[5];

  % sanity checks
  if (porb <= 0 || asini <= 0) {
    throw InvalidParmError, "porb and asini have to be greater zero"$;
  }

  % build qualifier struct for 'radial_velocity_binary'
  % and 'BinaryCor'
  variable rvb = struct {
    P = porb, omega = omega, e = ecc, degrees, v0 = v0
  };
  variable bc = struct {
    porb = porb, eccentricity = ecc, omega = omega   
  };
  if (qualifier_exists("t90")) {
    rvb = struct_combine(rvb, struct { T90 = tau });
    bc = struct_combine(bc, struct { t90 = tau });
  } else {
    rvb = struct_combine(rvb, struct { T0 = tau });
    bc = struct_combine(bc, struct { t0 = tau });
  }
  if (qualifier_exists("K")) {
    rvb = struct_combine(rvb, struct { K = asini });
    bc = struct_combine(bc, struct { % need to convert K to asini
      asini = asini / (2*PI) * (porb*86400. * sqrt((1.-ecc)*(1.+ecc))) / (Const_c*1e-5)
    });
  } else {
    rvb = struct_combine(rvb, struct { asini = asini*Const_c*1e-5 });
    bc = struct_combine(bc, struct { asini = asini });
  }

  % transform times into binary's barycenter
  variable cache = fitfun_get_cache();
  variable dometacor = 0;
  if (cache.binarytime == NULL || any(cache.lastpar[[:4]] != par[[:4]])) {
    cache.binarytime = BinaryCor(lo;; bc);
    cache.lastpar = par;
    dometacor = 1;
  }
  
  % purge binary corrected metadata
  variable metafields = qualifier("metacor", ["time"]);
  variable metacor = length(get_struct_field_names(cache.metacor));
  if (metafields == NULL) {
    set_struct_field(cache, "metacor", @Struct_Type(String_Type[0]));
  }
  else if (length(metafields) != metacor) { dometacor = 1; }
  % binary correct metadata
  if (dometacor) {
    % transform metadata fields
    if (metafields != NULL) {
      variable f, tfield, cfield, cname;
      % loop over fields to correct
      _for f (0, length(metafields)-1, 1) {
	tfield = get_dataset_metadata(Isis_Active_Dataset);
	cname = "metacor"; cfield = cache;
	if (tfield == NULL) { continue; }
	% walk along the path and recursively set 'tfield' and 'cfield'
	variable path;
	foreach path (strchop(metafields[f], '.', 0)) {
	  ifnot (struct_field_exists(tfield, path)) { tfield = NULL; break; }
	  % create path inside the cache if necessary
	  ifnot (struct_field_exists(get_struct_field(cfield, cname), path)) {
	    set_struct_field(cfield, cname, struct_combine(
	      get_struct_field(cfield, cname), path
	    ));
	  }
	  % set 'tfield' and 'cfield'
	  tfield = get_struct_field(tfield, path);
	  cfield = get_struct_field(cfield, cname);
	  cname = path;
	  % create empty structure in 'cfield'
	  if (typeof(tfield) == Struct_Type) {
	    if (get_struct_field(cfield, path) == NULL) {
	      set_struct_field(cfield, path, @Struct_Type(String_Type[0]));
	    }
	  }
	}
	if (tfield == NULL) { continue; }
	% correct the times
	set_struct_field(cfield, cname, BinaryCor(tfield;; bc));
      }
    }
  }

  % compute radial velocity and corresponding Doppler factor
  return (1. + radial_velocity_binary(cache.binarytime;; rvb) / (Const_c*1e-5));
}


% add fit-function
add_slang_function("dopplerorbit", [
  "asini [lt-s]", "porb [days]", "tau [MJD]", "ecc", "omega [degrees]", "v0 [km/s]"
]);
set_function_category("dopplerorbit", ISIS_FUN_ADDMUL);


% define parameter defaults
private define dopplerorbit_defaults(i) {
  return [
    struct { value = 100.0, freeze = 0, hard_min = 1, hard_max = 10000, min = 10, max = 1000, step = 1e-2, relstep = 1e-4 },
    struct { value = 10.0, freeze = 0, hard_min = 0.007, hard_max = 10000, min = 1, max = 1000, step = 1e-2, relstep = 1e-4 },
    struct { value = 56000.0, freeze = 0, hard_min = 15000, hard_max = 100000, min = 15000, max = int(ceil(UNIXtime2MJD(_time))), step = 1e-1, relstep = 1e-3 },
    struct { value = 0.0, freeze = 0, hard_min = 0.0, hard_max = 1.0, min = 0, max = 1.0, step = 1e-2, relstep = 1e-4 },
    struct { value = 0.0, freeze = 0, hard_min = -360, hard_max = 360, min = -180, max = 180, step = 1e-1, relstep = 1e-3 },
    struct { value = 0.0, freeze = 1, hard_min = -1000.0, hard_max = 1000.0, min = -200, max = 200, step = 1e-2, relstep = 1e-4 }
  ][i];
}
set_param_default_hook("dopplerorbit", &dopplerorbit_defaults);


% setup the fit-function's cache
fitfun_init_cache("dopplerorbit", &dopplerorbit_fit, struct {
  binarytime = NULL, lastpar, metacor = @Struct_Type(String_Type[0])
});

%
% Code from src/data/timing/dorDFT.sl
%
private define dorDFT(rate)
%!%+
%\function{dorDFT}
%\synopsis{Timing Tools: Discrete Fourier Transform}
%\usage{dft = dorDFT(rate);}
%\description
%    Performs a Fast Fourier Transform on a given rate array with the
%    same properties as IDL FFT:
%    * Normalization = sqrt(length(rate))
%    * renormalize rate around 0 by subtracting mean such that
%      variability is calculated wrt. mean rate
%    * return only meaningful bins for PSD calculations, i.e. first
%      bin up to Nyquist frequency
%!%-
{
  variable l = length(rate);
  return fft(rate-mean(rate), 1)[[1:l/2]]*l; %%% IDL Correction Factor;
}

%
% Code from src/plot/draw_plot_commands.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%
define draw_plot_commands()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{draw_plot_commands}
%\synopsis{creates plot commands for drawing lines with the mouse}
%\usage{draw_plot_commands();}
%\qualifiers{
%\qualifier{init}{[\code{=1}] initializes plot window at the beginning}
%}
%!%-
{
  xlin;
  ylin;
  xrange(0, 1);
  yrange(0, 1);

  title("left mouse button = set point;   middle mouse button = finish line;  right mouse button = exit");
  if(qualifier("init", 1))
    plot(0, 0);
  variable XX = {}, YY = {}, X = Double_Type[0], Y = @X, i;
  do
  {
    variable x, y, ch;
    cursor(&x,  &y, &ch);
    if(ch=="A")
      X = [X, x],
      Y = [Y, y];
    if(ch=="D" && length(X)>0)
    { list_append(XX, X);
      list_append(YY, Y);
      X = Double_Type[0];
      Y = Double_Type[0];
    }
    if(length(X)>0)  plot(X, Y);  else  plot(0, 0);
    _for i (0, length(XX)-1, 1)
    { color(1);
      oplot(XX[i], YY[i]);
    }
  } while(ch!="X");
  if(length(XX[i])>0)
  { list_append(XX, X);
    list_append(YY, Y);
  }

    vmessage("xrange(0, 1);\nyrange(0, 1);\n%s],\n%s]);",
             strjoin_wrap(array_map(String_Type, &sprintf, "%.3f", XX[0]), ", "; initial=" plot([", newline="       ", maxlen=80),
   	     strjoin_wrap(array_map(String_Type, &sprintf, "%.3f", YY[0]), ", "; initial="      [", newline="       ", maxlen=80));
  _for i (1, length(XX)-1, 1)
    vmessage("color(1);\n%s],\n%s]);",
 	     strjoin_wrap(array_map(String_Type, &sprintf, "%.3f", XX[i]), ", "; initial="oplot([", newline="       ", maxlen=80),
	     strjoin_wrap(array_map(String_Type, &sprintf, "%.3f", YY[i]), ", "; initial="      [", newline="       ", maxlen=80));
}

%
% Code from src/fitting/fit-functions/dummy.sl
%
%%%%%%%%%%%%%%%%
define dummy_fit(lo, hi, par)
%%%%%%%%%%%%%%%%
{
  return 0*lo + 1.;
}

%%%%%%%%%%%%
define dummy_default(i)
%%%%%%%%%%%%
{
  switch(i)
  {    case 0: return (0.0, 0, 0.0, 1e+4);}
}

add_slang_function ("dummy", ["const"]);
set_param_default_hook("dummy", &dummy_default);

%
% Code from src/data/ebv.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define e_bv()
%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{e_bv}
%\synopsis{Calculates the E(B_V) color excess after Predehl & Schmitt (1995)}
%\usage{Double_Type = e_bv(Double_Type N_H);}
%\qualifiers{
%    \qualifier{R_V}{Scalar specifying the ratio of total to selective extinction
%               R(V) = A(V) / E(B - V). If not specified, then R = 3.1
%               Extreme values of R(V) range from 2.3 to 5.3}
%}
%\description
%     From a given hydrogen absorption column density N_H in the diredtory
%     of the object, the color excess E(B-V) is calculated after
%     Predehl & Schmitt (1995): E(B-V) = N_H/(1.79e21*R_V).
%
% EXAMPLE
%     Calculate the color excess for Cen A (RA 13h25m27.6s DEC -43d01m09s) for
%     N_H = 8.09e20 and an "average" reddening of for the diffuse interstellar
%     medium (R(V) = 3.1).
%
%     isis> N_H = 8.09e20;
%     isis> ebv = e_bv(N_H);
%     isis> print(ebv);
%
%\seealso{fm_unred;}         
%!%-
{
   variable N_H;
   switch(_NARGS)
     { case 1: (N_H) = (); }
     { help(_function_name()); return; }

   variable R_V = qualifier("R_V", 3.1);
	
   variable ebv = N_H/(1.79e21*R_V);

   return ebv;
} %}}}

%
% Code from src/slang/edit_var.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define assoc_has_default_value(hash, defaultvalref)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable min_char =  32;  % ' '
  variable max_char = 126;  % '~'

  % find a string that is no key of hash
  variable N = 0;
  variable key = "";
  while(assoc_key_exists(hash, key))
  {
    N++;
    variable a = UChar_Type[N];
    a[*] = min_char;
    key = "";
    loop(N)  key += char(min_char);
    while(assoc_key_exists(hash, key))
    { variable i = -1;
      do i++, a[i] = min_char+((a[i]-min_char+1) mod (max_char-min_char+1));
      while(a[i]==min_char && i<N-1);
      if(a[i]==min_char && i==N-1)
	break;
      key = "";
      _for i (N-1, 0, -1)
        key += char(a[i]);
    }
  }

  try @defaultvalref = hash[key];
  catch AnyError: return 0;
  return 1;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define datatype_to_string();
private define datatype_to_string(val)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable i, last, type = typeof(val);
  variable indent = qualifier("indent", 0);
  variable sp = "";  loop(indent)  sp += "  ";
  variable fold = (indent>0);  % qualifier_exists("fold");

  indent++;
  if(type == String_Type || type == BString_Type)
      sprintf(`"%S"%s`, val, (type==BString_Type ? "B" : ""));  % left on stack
  else if(type == Array_Type || type == List_Type)
  {
      (type==Array_Type ? "[" : "{") + sprintf(" %% %S", val) + (fold ? " %{{{" : "") + "\n";  % left on stack
      last = length(val)-1;
      _for i (0, last, 1)
        () + sp+"  "
           + datatype_to_string(val[i]; indent=indent)
	   + (i<last ? "," : "") + "\n";  % left on stack
      () + (fold ? "%}}}\n" : "") + sp+ (type==Array_Type ? "]" : "}");  % left on stack
  }
  else if(is_struct_type(val))  % also works with typedef'ed struct's
  {
      sprintf("struct { %% %S", val) + (fold ? " %{{{" : "") + "\n";  % left on stack
      variable field = get_struct_field_names(val);
      last = length(field)-1;
      _for i (0, last, 1)
        () + sp+"  " + field[i] + " = "
           + datatype_to_string(get_struct_field(val, field[i]); indent=indent)
	   + (i<last ? "," : "") + "\n";  % left on stack
      () + (fold ? "%}}}\n" : "") + sp + "}";  % left on stack
  }
  else if(type == Assoc_Type)
  {
      variable defval;
      "($0=Assoc_Type[" + string(_typeof(assoc_get_values(val)))
                        + (assoc_has_default_value(val, &defval) ? ", "+datatype_to_string(defval; indent=indent) : "")
	                + "]," + (fold ? " %{{{" : "") + "\n";  % left on stack
      variable key = assoc_get_keys(val);
      last = length(key)-1;
      _for i (0, last, 1)
        () + sp+`  $0["`+key[i]+`"] = `
           + datatype_to_string(val[key[i]]; indent=indent)
           + ",\n";  % left on stack
      () + (fold ? "%}}}\n" : "") + sp + "$0)";  % left on stack
  }
  else if(type == Char_Type)
      sprintf("typecast(%d,Char_Type)", val);  % left on stack
  else if(type == UChar_Type)
      sprintf(`'\x%02X'`, val);  % left on stack
  else if(type == Short_Type)  % == Int16_Type
      sprintf("%dH", val);  % left on stack
  else if(type == UShort_Type)  % == UInt16_Type
      sprintf("%dUH", val);  % left on stack
  else if(type == Long_Type)  % == Int32_Type
      sprintf("%dL", val);  % left on stack
  else if(type == ULong_Type)  % == UInt32_Type
      sprintf("%ldUL", val);  % left on stack
  else if(type == LLong_Type)  % == Int64_Type
      sprintf("%lldLL", val);  % left on stack
  else if(type == ULLong_Type)  % == UInt64_Type
      sprintf("%lldULL", val);  % left on stack
  else if(type == Float_Type)
      sprintf("%ff", val);  % left on stack
  else % (Void_Type, Null_Type, Double_Type, Complex_Type, DataType_Type)
      sprintf("%S", val);  % left on stack
}


%%%%%%%%%%%%%%%
define edit_var()
%%%%%%%%%%%%%%%
%!%+
%\function{edit_var}
%\synopsis{allows to edit S-Lang variables in an editor}
%\usage{edit_var(&x);
%\altusage{Any_Type y = edit_var(Any_Type x);}
%}
%\qualifiers{
%\qualifier{tmpfile}{temporary file [default: /tmp/edit_var_$UID_$PID]}
%}
%\description
%    edit_var supports the following data types:
%    - Undefined_Type (=Void_Type), Null_Type,
%    - Integer_Type, Double_Type, Complex_Type
%    - String_Type, BString_Type
%    as well as
%    - Array_Type
%    - Assoc_Type
%    - Struct_Type
%    - List_Type
%    in a recursive way. (Circular linked list
%    are currently not supported.)
%
%    S-Lang code defining the variable x is shown
%    in the editor specified by the EDITOR environ-
%    ment variable or jed, if EDITOR is undefined.
%    edit_var uses jed's folding mode (one should
%    run 'Buffers' => 'Folding' => 'Enable Folding')
%    for nested data structures, which can hence be
%    very easily investigated.
%
%    The S-Lang code for x can be modified. After
%    saving the temporary file and closing the editor,
%    the file is evaluated, which should return an
%    S-Lang object that is either stored in x
%    (if passed by reference) or returned.
%\example
%    \code{i = edit_var(struct { uc='A', s=1H, us=1UH, l=1L, ul=1UL, ll=1LL, ull=1ULL });}
%!%-
{
  variable x;
  switch(_NARGS)
  { case 1: x = (); }
  { help(_function_name()); return; }

  variable is_ref = (typeof(x)==Ref_Type);
  variable xx = ( is_ref ? @x : x );
  variable t = typeof(xx);

  variable tmpfile = qualifier("tmpfile", sprintf("/tmp/edit_var_%d_%d", getuid(), getpid()));
  variable fp = fopen(tmpfile, "w");
  ()=fputs("% -*- mode: SLang; mode: fold -*-\n\n"
	   +(t==List_Type ? "($0=" : "")
	   +datatype_to_string(xx)
	   +(t==List_Type ? ", $0)" : ""),
	   fp);
  ()=fclose(fp);

  variable e = getenv("EDITOR");  if(e==NULL)  e = "jed";
  if(system(e+" "+tmpfile))
    if(is_ref)  return;  else  return x;

  try(e)  ()=evalfile(tmpfile);  % S-Lang code produces return value that is left on stack
  catch AnyError:
  {
    vmessage(`"%s" exception:
%s
%s:%d`, e.descr, e.message, e.file, e.line);
    if(is_ref)  return;  else  return NULL;
  }
  finally ()=remove(tmpfile);

  if(is_ref)  @x = ();
}

%
% Code from src/misc/atomic/element_name.sl
%

private variable _element_sym = ["Bare","H","He","Li","Be","B","C","N","O","F","Ne",
				  "Na","Mg","Al","Si","P","S","Cl","Ar","K","Ca","Sc",
				  "Ti","V","Cr","Mn","Fe","Co","Ni","Cu","Zn","Ga","Ge",
				  "As","Se","Br","Kr", "Rb","Sr","Y","Zr","Nb","Mo","Tc",
				  "Ru","Rh","Pd","Ag","Cd","In","Sn","Sb","Te","I","Xe",
				  "Cs","Ba","La","Ce","Pr","Nd","Pm","Sm","Eu","Gd","Tb",
				  "Dy","Ho","Er","Tm","Yb","Lu","Hf","Ta","W","Re","Os",
				  "Ir","Pt","Au","Hg","Tl","Pb","Bi","Po","At","Rn",
				  "Rr","Ra","Ac","Th","Pa","U","Np","Pu","Am","Cm","Bk",
				  "Cf","Es","Fm","Md","No","Lr","Rf","Db","Sg","Bh","Hs",
				  "Mt","Ds","Rg","Cn","Nh","Fl","Mc","Lv","Ts","Og"];

private variable _element_name = ["Bare","Hydrogen","Helium","Lithium","Beryllium","Boron",
				  "Carbon","Nitrogen","Oxygen","Fluorine","Neon",
				  "Sodium","Magnesium","Aluminum","Silicon","Phosphorus",
				  "Sulfur","Chlorine","Argon","Potassium","Calcium","Scandium",
				  "Titanium","Vanadium","Chromium","Manganese","Iron","Cobalt",
				  "Nickel","Copper","Zinc","Gallium","Germanium",
				  "Arsenic","Selenium","Bromine","Krypton","Rubidium","Strontium",
				  "Yttrium","Zirconium","Niobium","Molybdenum","Technetium",
				  "Ruthenium","Rhodium","Palladium","Silver","Cadmium","Indium",
				  "Tin","Antimony","Tellurium","Iodine","Xenon",
				  "Cesium","Barium","Lanthanum","Cerium","Praseodymium",
				  "Neodymium","Promethium","Samarium","Europium","Gadolinium",
				  "Terbium",
				  "Dysprosium","Holmium","Erbium","Thulium","Ytterbium",
				  "Lutetium","Hafnium","Tantalum","Tungsten","Rhenium",
				  "Osmium",
				  "Oridium","Platinum","Gold","Mercury","Thallium","Lead",
				  "Bismuth","Polonium","Astatine","Radon",
				  "Francium","Radium","Actinum","Thorium","Protactinium",
				  "Uranium","Neptunium","Plutonium","Americium","Curium",
				  "Berkelium",
				  "Californium","Einsteinium","Fermium","Mendelevium","Nobelium",
				  "Lawrencium","Rutherfordium","Dubnium","Seaborgium",
				  "Bohrium","Hassium",
				  "Meitnerium","Darmstadtium","Roentgenium","Copernicium",
				  "Nihonium","Flerovium","Moscovium","Livermorium",
				  "Tennessine","Oganesson"];

private variable _element2Z=Assoc_Type[Int_Type];

define element_name() {
%!%+
%\function{element_name}
%\synopsis{returns the name of the element with proton number Z}
%\usage{String_Type element_name(Integer_Type Z)}
%\qualifiers{
%\qualifier{lc}{return symbol or name in lower case (default: Capitalized)}
%}
%\description
% This function returns the name of the element with
% nuclear charge Z for all named elements.
%
% This function is array safe.
%\seealso{element_symbol,element2Z}
%!%-
  variable Z;
  switch(_NARGS)
  { case 1: Z = (); }
  { help(_function_name()); return; }

  if (min(Z)<0 or max(Z)>=length(_element_name)) {
      throw UsageError,sprintf("%s: Element Z outside of 0<Z<%i",_function_name(),length(_element_name));
  }

  if (qualifier_exists("lc")) {
      return strlow(_element_name[Z]);
  } 
  return _element_name[Z];
}

define element_symbol() {
%!%+
%\function{element_symbol}
%\synopsis{returns the symbol of the element with proton number Z}
%\usage{String_Type element_symbol(Integer_Type Z)}
%\qualifiers{
%\qualifier{full}{return full element name rather than symbol}
%\qualifier{lc}{return symbol or name in lower case (default: Capitalized)}
%}
%\description
% This function returns the symbol or the name of the element with
% nuclear charge Z for all named elements. 
% Z=0 returns 'Bare'.
%
% This function is array safe.
%\seealso{element_name, element2Z}
%!%-
  variable Z;
  switch(_NARGS)
  { case 1: Z = (); }
  { help(_function_name()); return; }

  if (min(Z)<0 or max(Z)>=length(_element_sym)) {
      throw UsageError,sprintf("%s: Element Z outside of 0<Z<%i",_function_name(),length(_element_sym));
  }

  if (qualifier_exists("full")) {
      if (qualifier_exists("lc")) {
	  return strlow(_element_name[Z]);
      } 
      return _element_name[Z];
  }
  
  if (qualifier_exists("lc")) {
      return strlow(_element_sym[Z]);
  } 
  
  return _element_sym[Z];
}

define element2Z() {
%!%+
%\function{element2Z}
%\synopsis{returns Z for an element symbol or name}
%\usage{z=element2Z(String_Type name)}
%\description
% This function returns the nuclear charge for the given element symbol
% symbol or element name.
%
% This function is array safe.
%\seealso{element_name, element_symbol}
%!%-
   variable names;
   switch(_NARGS)
   { case 1: names=(); }
   { help(_function_name()); return; }

   % are we initialized?
   variable i;
   if (length(_element2Z)==0) {
       _for i(0,length(_element_sym)-1,1) {
	   _element2Z[strup(_element_sym[i])]=i;
	   _element2Z[strup(_element_name[i])]=i;
       }
   }

   if (typeof(names)==String_Type) {
       return _element2Z[strup(names)];
   }

   variable ret=Integer_Type[length(names)];
   
   _for i(0,length(names)-1,1) {
       ret[i]=_element2Z[strup(names[i])];
   }
   return ret;
}

%
% Code from src/slang/math/ellipse.sl
%
define ellipse ()
%!%+
%\function{ellipse}
%\synopsis{calculates points on an ellipse centered at the origin}
%\usage{(Double_Type X,Y) = ellipse(Double_Type smaj, smin, posang, phi)
%}
%\qualifiers{
%\qualifier{x}{[\code{=1}] compression/streching factor along the x-axis}
%\qualifier{y}{[\code{=1}] compression/streching factor along the y-axis}
%}
%\description
%    This function calculates the coordinates \code{X}, \code{Y} of an ellipse with
%    semimajor axis \code{smaj}, semiminor axis \code{smin}, and position angle \code{posang}
%    (measured in rad with respect to the x-axis) which is parameterized with \code{phi}.
%    The complete ellipse is covered when \code{phi} covers the values from 0 to 2*PI.
%\examples
%    variable phi=[0:2*PI:#200];
%    plot ( ellipse(5,3,0.2*PI,phi) );
%    oplot( ellipse(5,3,0.2*PI,phi ; x=0.5) );
%\seealso{enclosing_ellipse}
%!%-
{
  variable a,b,p,t;
  switch(_NARGS)
  { case 4: (a,b,p,t) = (); }
  { help(_function_name()); return; }
  
  variable x=qualifier("x",1.);
  variable y=qualifier("y",1.);
  return (
	  x*a*cos(t)*cos(p)-x*b*sin(t)*sin(p),
	  y*a*cos(t)*sin(p)+y*b*sin(t)*cos(p)
	 );
  
}

%
% Code from src/fitting/xyfit/fit-functions/ellipse_xy_xyfit.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define ellipse_xy_xyfit()
%%%%%%%%%%%%%%%%%%%%%%
{
  variable xref, yref, par;
  switch(_NARGS)
  { case 0: return ["x0 [x center]", "y0 [y center]", "semi_major", "axis_ratio [minor/major]", "pos_angle [deg]"]; } % fit parameters
  { case 3: (xref, yref, par) = (); }
  { return help(_function_name); }

  variable phi = qualifier("curve_parameter");
  if(phi == NULL)
   throw UsageError, "error in ellipse xyfit function: please set_xyfit_qualifier(; curve_parameter=[0:2*PI:#steps]);";

  variable x, y;
  (x, y) = ellipse(par[2], par[2]*par[3], par[4]*PI/180, phi);
  (@xref, @yref) = (x+par[0], y+par[1]);
}

define ellipse_xy_xyfit_default(i)
{
  switch(i)
  { case 0: return (0   , 0   , -1e5 , 1e5 ); } % x-center 
  { case 1: return (0   , 0   , -1e5 , 1e5 ); } % y-center 
  { case 2: return (1   , 0   , 0    , 1e5 ); } % semi-major axis
  { case 3: return (1   , 0   , 1e-9 ,   1 ); } % axis ratio minor/major
  { case 4: return (1   , 0   , -360 , 360 ); } % position angle [deg]
}

%
% Code from src/fitting/ensemble-samplers/emcee.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

% only activate if module rand could be loaded
#ifeval __get_reference("rand_flat")!=NULL

% only activate if module socket could be loaded
#ifeval __get_reference("socket")!=NULL
% -*- mode: slang; mode: fold; -*- %

%
% Implementation of the emcee hammer (Foreman-Mackey 2013) with
% the principle idea that multiple nodes (engines) are responsible
% for a part of the walkers. For efficiency the walkers are
% distributed equally to each engine. To keep the statistical
% properties the walkers are seperated in to two groups (see ref)
% where the next step of group one depends on the current position
% of group two and the next step of group two depends on the new
% position of group one. For most efficiency we try to reduce the
% required computations to the minimum possible such that the model
% evaluation plus the necessary communication is everything that
% happens in the main loop.
% 
% To prevent any side effects from the PRNG we let the master
% calculate enough for each step and distribute them to the
% slaves.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private variable EmceeInitRegister = Assoc_Type[Ref_Type, &NULL];
private variable EmceeStepRegister = Assoc_Type[Ref_Type, &NULL];
private variable EmceeFileRegister = Assoc_Type[Ref_Type, &NULL];
private variable EmceeShipRegister = Assoc_Type[Ref_Type, &NULL];
private variable EmceeProgressRegister = Assoc_Type[Ref_Type, &NULL];

% Engine and Leader %{{{

private variable EmceeEngine = struct {
  % engine numbers
  id, % engine id (master is 0)
  numberEngines, % total number of engines
  numberSteps, % total number of steps

  % arrays and length
  walkers, % walker array for this engine (master has all)
  pivots, % pivot array for this engine (master has all)
  rolls, % random numbers required (master has all)
  update, % update indicator array (master has all)
  stat, % statistic array (master has all)
  totalNumberWalkers, % total number of walkers
  totalNumberSet1, % total number of walkers in set 1
  totalNumberSet2, % total number of walkers in set 2

  % per engine walkers
  numberWalkers, % number of walkers this engine handles
  numberWalkersSet1, % number of walkers this engine handles in set 1
  numberWalkersSet2, % number of walkers this engine handles in set 2

  % selected set
  setOffset, % current set offset (for access in walker array)
  setLength, % current set length

  % fit
  fit, % fit object
  numberParameters, % number of (free) parameters

  gears = NULL, % step, random generators
  leader = NULL, % write buffers etc.
};

private variable EmceeLeader = struct {
  walkersPerSet1, % number of walkers for each engine id in set 1
  walkersPerSet2, % number of walkers for each engine id in set 2
  walkersPerSet, % selected set walkers
  totalOffset, % start of set in walker array
  writeBuffer, % total write buffer array
  inFile, % input file handle
  outFile, % output file handle
  progress, % progress handle
};

private variable EmceeGears = struct {
  upick, % function
  urand, % function

  step, % step interface
};

private define emceeDrawSet (engine, set) %{{{
{
  variable urand = engine.gears.urand;
  variable upick = engine.gears.upick;

  variable totalNumberSet,
    totalNumberComplement,
    totalOffset;

  if (1 == set) {
    totalNumberSet = engine.totalNumberSet1;
    totalNumberComplement = engine.totalNumberSet2;
    engine.setOffset = 0;
    engine.setLength = engine.numberWalkersSet1;
    totalOffset = 0;
    if (0 == engine.id) {
      engine.leader.totalOffset = totalOffset;
      engine.leader.walkersPerSet = engine.leader.walkersPerSet1;
    }
  } else if (2 == set) {
    totalNumberSet = engine.totalNumberSet2;
    totalNumberComplement = engine.totalNumberSet1;
    engine.setOffset = engine.numberWalkersSet1;
    engine.setLength = engine.numberWalkersSet2;
    totalOffset = engine.totalNumberSet1;
    if (0 == engine.id) {
      engine.leader.totalOffset = totalOffset;
      engine.leader.walkersPerSet = engine.leader.walkersPerSet2;
      engine.setOffset = totalOffset;
    }
  }

  % master picks new pivots & randoms
  if (0 == engine.id) {
    variable pick = @upick(0, totalNumberComplement-1, totalNumberSet)+totalOffset;
    variable i;
    _for i (0, totalNumberSet-1)
      engine.pivots[i+totalOffset] = @(engine.walkers[pick[i]]);

    variable numberRandoms = engine.gears.step.numberRandoms;
    engine.rolls[[0:totalNumberSet*numberRandoms-1]+totalOffset*numberRandoms]
      = @urand(totalNumberSet*numberRandoms);
  }
}
%}}}

private define emceeSetupGears (engine, urand, upick, step) %{{{
{
  variable gears = struct { @EmceeGears };
  gears.urand = urand;
  gears.upick = upick;
  gears.step = step;

  % here we can set the rolls
  engine.rolls = Double_Type[length(engine.walkers)*step.numberRandoms];

  engine.gears = gears;
}
%}}}

private define emceeSetupWriteBuffer (leader, numberWalkers, numberSteps) %{{{
{
  % buffer size should ideally be as large as the write routine wants
  % but is limited by the maximum array size and must be at least
  % as large as one iteration requires
  variable size = min([[leader.outFile.cycle, numberSteps]*numberWalkers, INT_MAX-(INT_MAX mod numberWalkers)]);
  variable writeBuffer = struct {
    size = size,
    cycle = size/numberWalkers,
    walkers = Array_Type[size],
    update = Double_Type[size],
    stat = Double_Type[size],
  };

  variable i;
  _for i (0, size-1)
    writeBuffer.walkers[i] = Double_Type[num_free_params()];

  leader.writeBuffer = writeBuffer;
}
%}}}

private define emceeSetupLeader (engine, inFile, outFile, progress) %{{{
{
  if (0 == engine.id) {
    variable nEngines = engine.numberEngines;
    variable id;
    variable leader = @EmceeLeader;
    leader.walkersPerSet1 = Int_Type[nEngines];
    leader.walkersPerSet2 = Int_Type[nEngines];
    leader.inFile = inFile;
    leader.outFile = outFile;
    leader.progress = progress;

    emceeSetupWriteBuffer(leader, engine.totalNumberWalkers, engine.numberSteps);

    variable set1 = engine.totalNumberSet1;
    variable set2 = engine.totalNumberSet2;

    _for id (0, engine.numberEngines-1) {
      leader.walkersPerSet1[id] = set1/nEngines + ((set1 mod nEngines) > (nEngines-id-1));
      leader.walkersPerSet2[id] = set2/nEngines + ((set2 mod nEngines) > (nEngines-id-1));
    }

    engine.leader = leader;
  }
}
%}}}

private define emceeSetupEngine (ship, totalNumberWalkers, totalSteps) %{{{
{
  variable engine = @EmceeEngine;
  ship.engine = engine;

  variable set1 = totalNumberWalkers/2;
  variable set2 = totalNumberWalkers - set1;
  engine.totalNumberSet1 = set1;
  engine.totalNumberSet2 = set2;
  engine.totalNumberWalkers = totalNumberWalkers;
  engine.numberSteps = totalSteps;

  % set sail (get number of engines and set id)
  ship.setSail();

  % divide walkers evenly (remainders are given to highest ids)
  engine.numberWalkersSet1 = set1/engine.numberEngines
    + ((set1 mod engine.numberEngines) > (engine.numberEngines-engine.id-1));
  engine.numberWalkersSet2 = set2/engine.numberEngines
    + ((set2 mod engine.numberEngines) > (engine.numberEngines-engine.id-1));
  engine.numberWalkers = engine.numberWalkersSet1 + engine.numberWalkersSet2;
  engine.numberParameters = num_free_params();
  engine.fit = open_fit();

  % master stores all values
  variable size = engine.id ? engine.numberWalkers : totalNumberWalkers;

  engine.walkers = Array_Type[size];
  engine.pivots  = Array_Type[size];
  % rolls can only be set up after step is known
  % engine.rolls   = Double_Type[size*engine.gears.step.numberRandoms];
  engine.update  = Int_Type[size];
  engine.stat    = Double_Type[size] + DOUBLE_MAX;

  variable j;
  _for j (0, size-1) {
    engine.walkers[j] = Double_Type[engine.numberParameters];
    engine.pivots[j] = Double_Type[engine.numberParameters];
  }
}
%}}}
%}}}

%{{{ File interface:
%!%+
%\function{emcee--file}
%\synopsis{Set emcee file input and output methods}
%\usage{input="method;options"
%  \altusage{output="method;options"}}
%\description
%  The file input/output methods can be set with the function string
%    "method;parameter"
%
%  Available methods:
%  fits : Fits file interface to write the chain as fits table extension
%    ; filename  : [emcee-<date>.fits] The input/output file name.
%    ; parameter : If given, on read we draw new starting positions from the
%                  parameter settings stored in the file instead of reading
%                  the last iterations.
%
%  mike : Fits file interface (compatible to previous emcee routine)
%    ; filename  : [emcee-<date>.fits] The input/output file name.
%    ; parameter : if given, on read we draw new starting position from the
%                  header
%    ; cycle     : [=50] the number of steps to calculate before writing to
%                  file
%
%  par  : Parameter file interface to draw initial walkers from parameter
%         files.
%    ; filename : [emcee-<data>.fits] Multiple parameter files can be separated
%                 by a semi colon with an optional additional multiplier
%                 (separated by a colon). A string of the form
%                 'file1.par:2;file2.par' means that 2/3 of all walkers are
%                 drawn from file1.par and 1/3 is drawn from file2.par.
%!%-
%      1: create - open new file pointer and write necessary intial values
%      2: open - open existing file for read/write
%      3: read - open file and return n walkers and how many walkers were used
%      4: write - write cycle steps to the file (n)
%      5: close - close open file at end
private variable EMCEE_FILE_READ = 0x1, EMCEE_FILE_WRITE=0x2, EMCEE_FILE_RANGE = 0x4;
private variable EmceeFile = struct {
  create, % function
  open, % function
  read, % function
  write, % function
  close, % function

  mode = 0, % 1 read, 2 write, 4 range bit (read parameter range instead of position)
  has = 0,  % same as mode, but lists all available (if FILE_RANGE is given it means it is the prevered method)
  handle, % file handle
  filename, % full file name
  cycle, % number of steps before file gets written

  % additional private data
};

%{{{ Fits file functions

% Create function %{{{
private define __emceeFitsWriteT1(handle, engine) %{{{
{
  variable dataInfo;
  list_data(&dataInfo);

  variable par = __parameters(engine.fit.object);
  variable params = get_params();
  variable numberTotalParams = length(params);

  %variable parNames = array_map(String_Type, &get_struct_field, get_params(), "name")[par.index-1];

%  fits_create_binary_table(handle, "PARAMETERS", num_free_params(),
%			   ["FREE_PAR", "FREE_PAR_NAME"],
%			   ["J", sprintf("%dA", max(array_map(Int_Type, &strlen, parNames)))],
%			   [" parameter indices", " parameter names"]);
  variable paramsTable = struct {
    name=String_Type[numberTotalParams],
    index=Int_Type[numberTotalParams],
    value=Double_Type[numberTotalParams],
    min=Double_Type[numberTotalParams],
    max=Double_Type[numberTotalParams],
    hard_min=Double_Type[numberTotalParams],
    hard_max=Double_Type[numberTotalParams],
    freeze=Int_Type[numberTotalParams],
    tie=String_Type[numberTotalParams],
    units=String_Type[numberTotalParams],
    fun=String_Type[numberTotalParams],
    free=Int_Type[numberTotalParams], % combines freeze, fun and tie
  };
  variable j;
  _for j (0, numberTotalParams-1) {
    paramsTable.name[j] = params[j].name;
    paramsTable.index[j] = params[j].index;
    paramsTable.value[j] = params[j].value;
    paramsTable.min[j] = params[j].min;
    paramsTable.max[j] = params[j].max;
    paramsTable.hard_min[j] = params[j].hard_min;
    paramsTable.hard_max[j] = params[j].hard_max;
    paramsTable.freeze[j] = params[j].freeze;
    paramsTable.tie[j] = (params[j].tie == NULL) ? "" : params[j].tie;
    paramsTable.units[j] = params[j].units;
    paramsTable.fun[j] = (params[j].fun == NULL) ? "" : params[j].fun;
    paramsTable.free[j] = (not params[j].freeze) and (params[j].fun == NULL) and (params[j].tie == NULL);
  }
  fits_write_binary_table(handle, "PARAMETERS", paramsTable);

  fits_update_key(handle, "MODEL", get_fit_fun(), "model function");
  fits_update_key(handle, "STATISTIC", get_fit_statistic(), " latest fit statistic");
  fits_update_key(handle, "SLOPPY", 0, " sloppy level");

  array_map(&fits_write_comment, handle, strchop(dataInfo, '\n', 0));

  % sort to index order here
%  if (_fits_write_col(handle, fits_get_colnum(handle, "FREE_PAR"), 1, 1, par.index[array_sort(par.index)])
%      && _fits_write_col(handle, fits_get_colnum(handle, "FREE_PAR_NAME"), 1, 1, parNames))
%    throw IOError;
}
%}}}

private define __emceeFitsWriteT2(handle, engine) %{{{
{
  variable par = __parameters(engine.fit.object);

  fits_create_binary_table(handle, "MCMCCHAIN", 0,
			   ["FITSTAT", "UPDATE", array_map(String_Type, &sprintf, "CHAINS%d", par.index)],
			   ["D", "J", ["D"][par.index*0]],
			   [" fit statistics", " update indicator", [" parameter values"][par.index*0]]);
  fits_update_key(handle, "NWALKERS", engine.totalNumberWalkers/engine.numberParameters, " Number of walkers per free parameter");
  fits_update_key(handle, "NFREEPAR", engine.numberParameters, " Number of free parameters");
  fits_update_key(handle, "NSTEPS", engine.numberSteps, " Numer of iteration steps done");
}
%}}}

private define __emceeFitsWriteT3(handle, engine) %{{{
{
  fits_create_binary_table(handle, "CHAINSTATS", 0,
			   ["FRAC_UPDATE", "MIN_STAT", "MED_STAT", "MAX_STAT"], ["D", "D", "D", "D"],
			   [" fraction", [sprintf(" %s", get_fit_statistic)][[0:2]*0]]);
  fits_update_key(handle, "STATISTIC", get_fit_statistic(), " fit statistic");
}
%}}}

private define emceeFileFitsCreate (file, engine) %{{{
{
  file.mode |= EMCEE_FILE_WRITE;
  file.mode &= ~EMCEE_FILE_READ;

  % Create fits file and write headers
  file.handle = fits_open_file(file.filename, "c");

  % write first table
  __emceeFitsWriteT1(file.handle, engine);

  % write second table
  __emceeFitsWriteT2(file.handle, engine);

  % write third table
  %__emceeFitsWriteT3(file.handle, engine);

  % move back to chain table
  () = _fits_movnam_hdu(file.handle, _FITS_BINARY_TBL, "MCMCCHAIN", 0);

  % set write cycle
  () = _fits_get_rowsize(file.handle, &(file.cycle));
  file.cycle = file.cycle/engine.totalNumberWalkers;
  if (file.cycle < 1)
    file.cycle = 1;

  % fits routine customs
  file.numberSteps = 0;
  file.sloppy = 0;
}
%}}}
%}}}

% Open function %{{{

private define __emceeFitsReadChecks (file, engine) %{{{
{
  variable handle = file.handle;
  if (_fits_movnam_hdu(handle, _FITS_BINARY_TBL, "PARAMETERS", 0)) {
    fits_close_file(handle);
    handle = NULL;
    throw IOError, "Not a emcee chain file";
  }

  if (fits_read_key(handle, "MODEL") != get_fit_fun()) {
    fits_close_file(handle);
    handle = NULL;
    throw IsisError, "Current model and chain model do not match";
  }

  variable tab = fits_read_table(handle);
  ifnot (struct_field_exists(tab, "free")
	|| struct_field_exists(tab, "value")) {
    fits_close_file(handle);
    handle = NULL;
    throw IOError, "Not a emcee chain file";
  }

  variable par = __parameters(engine.fit.object);
  if ((length(where(tab.free)) != num_free_params())
      || any(tab.index[where(tab.free)] != par.index[array_sort(par.index)])) {
    fits_close_file(handle);
    handle = NULL;
    throw UsageError, "Free parameters and chain parameters differ";
  }
}
%}}}

private define __emceeFitsWriteChecks (file, engine) %{{{
{
  variable handle = file.handle;
  if (_fits_movnam_hdu(handle, _FITS_BINARY_TBL, "PARAMETERS", 0)) {
    fits_close_file(handle);
    handle = NULL;
    throw IOError, "Not a emcee chain file";
  }

  if ((fits_read_key(handle, "STATISTIC") != get_fit_statistic()) && (file.sloppy<2)) {
    fits_close_file(handle);
    handle = NULL;
    throw UsageError, "Current fit statistic and chain fit statistic differ, increase sloppy level (at least 2) to continue anyway";
  }
  fits_update_key(handle, "STATISTIC", get_fit_statistic());

  if (_fits_movnam_hdu(handle, _FITS_BINARY_TBL, "MCMCCHAIN", 0)) {
    fits_close_file(handle);
    handle = NULL;
    throw IOError, "Not a emcee chain file";
  }

  if ((fits_read_key(handle, "NWALKERS")*fits_read_key(handle, "NFREEPAR")) != length(engine.walkers)
      && (file.sloppy<1)) {
    fits_close_file(handle);
    handle = NULL;
    throw IOError, "Number of walkers differs from number used in chain file, increase sloppy level to continue";
  }
}
%}}}

private define emceeFileFitsOpen (file, engine) %{{{
{
  file.mode |= EMCEE_FILE_READ | EMCEE_FILE_WRITE;

  file.handle = fits_open_file(file.filename, "w");

  __emceeFitsReadChecks(file, engine);
  __emceeFitsWriteChecks(file, engine);

  if (_fits_movnam_hdu(file.handle, _FITS_BINARY_TBL, "MCMCCHAIN", 0)) {
    fits_close_file(file.handle);
    file.handle = NULL;
    throw IOError, "Not a emcee chain file";
  }

  () = _fits_get_rowsize(file.handle, &(file.cycle));
  file.cycle = file.cycle/length(engine.walkers);
  if (file.cycle < 1)
    file.cycle = 1;

  file.numberSteps = fits_get_num_rows(file.handle);
}
%}}}
%}}}

% Read function %{{{
private define emceeFileFitsRead (file, engine, numberWalkers) %{{{
{
  file.mode |= EMCEE_FILE_READ;
  file.mode &= ~EMCEE_FILE_WRITE;

  file.handle = fits_open_file(file.filename, "r");

  __emceeFitsReadChecks(file, engine);

  if (file.mode & EMCEE_FILE_RANGE) {
    if (_fits_movnam_hdu(file.handle, _FITS_BINARY_TBL, "PARAMETERS", 0)) {
      fits_close_file(file.handle);
      file.handle = NULL;
      throw IOError, "Not a emcee chain file";
    }

    variable params = fits_read_table(file.handle);
    return [struct {
      weight = 1,
      name = params.name,
      value = params.value,
      min = params.min,
      max = params.max
    }], numberWalkers;
  } else {
    if (_fits_movnam_hdu(file.handle, _FITS_BINARY_TBL, "MCMCCHAIN", 0)) {
      fits_close_file(file.handle);
      file.handle = NULL;
      throw IOError, "Not a emcee chain file";
    }

    variable totalNumberWalkers = fits_read_key(file.handle, "NWALKERS")
      *fits_read_key(file.handle, "NFREEPAR");
    variable numberParameters = fits_get_num_cols(file.handle);
    variable totalNumberRecords = fits_get_num_rows(file.handle);
    variable walkerDistribution; % this has to be an array of arrays with the parameter distribution in each
    () = _fits_read_cols(file.handle,
			 [3:numberParameters],
			 max([0, totalNumberRecords-numberWalkers]),
			 numberWalkers,
			 &walkerDistribution);

    fits_close_file(file.handle);

    return walkerDistribution, totalNumberWalkers;
  }
}
%}}}
%}}}

% Write function %{{{
private define emceeFileFitsWrite (file, engine, numberWalkersSteps) %{{{
{
  if (numberWalkersSteps > engine.leader.writeBuffer.size)
    throw InternalError, "Trying to write more than accessible";

  variable par = __parameters(engine.fit.object);
  variable npar = engine.numberParameters;

  %variable walkersPerCycle = engine.leader.writeBuffer.size; % total_walkers*steps_per_cycle
  variable i,j;
  variable firstIndex = fits_get_num_rows(file.handle)+1; % first index of this cycle
  variable parCycle = Double_Type[numberWalkersSteps];

  _for j (0, npar-1, 1) {
    _for i (0, numberWalkersSteps-1, 1)
      parCycle[i] = engine.leader.writeBuffer.walkers[i][j];
    () = _fits_write_col(file.handle,
			 fits_get_colnum(file.handle, sprintf("CHAINS%d", par.index[j])),
			 firstIndex,
			 1,
			 parCycle);
  }
  () = _fits_write_col(file.handle,
		       fits_get_colnum(file.handle, "FITSTAT"),
		       firstIndex,
		       1,
		       engine.leader.writeBuffer.stat[[:numberWalkersSteps-1]]);
  () = _fits_write_col(file.handle,
		       fits_get_colnum(file.handle, "UPDATE"),
		       firstIndex,
		       1,
		       engine.leader.writeBuffer.update[[:numberWalkersSteps-1]]);
}
%}}}
%}}}

% Close function %{{{
private define emceeFileFitsClose (file, engine) %{{{
{
  % todo: write fitstat table
  variable nHDUs = fits_get_num_hdus(file.handle);
  variable i;

  if (file.mode & EMCEE_FILE_WRITE) {
    _for i (1, nHDUs) {
      () = _fits_movabs_hdu(file.handle, i);
      fits_write_chksum(file.handle);
    }
  }

  fits_close_file(file.handle);
}
%}}}
%}}}
%}}}
private define emceeFileFits () %{{{
{
  variable file = struct { @EmceeFile, numberSteps, sloppy };
  file.create = &emceeFileFitsCreate;
  file.open = &emceeFileFitsOpen;
  file.read = &emceeFileFitsRead;
  file.write = &emceeFileFitsWrite;
  file.close = &emceeFileFitsClose;

  file.filename = qualifier("filename", strftime("emcee-%Y%m%d-%H%M%S.fits"));
  file.mode |= qualifier_exists("parameter") ? EMCEE_FILE_RANGE : 0;
  file.has |= EMCEE_FILE_READ | EMCEE_FILE_WRITE;
  file.cycle = 1;

  file.numberSteps = 0;
  file.sloppy = 0;

  return file;
}
%}}}
EmceeFileRegister["fits"] = &emceeFileFits;

%{{{ Mikes file format (similar to fits, but contains a third table with some stats)

% Create function %{{{
private define emceeFileMikeCreate (file, engine) %{{{
{
  file.mode |= EMCEE_FILE_WRITE;
  file.mode &= ~EMCEE_FILE_READ;

  % Create fits file and write headers
  file.handle = fits_open_file(file.filename, "c");

  % write first table
  __emceeFitsWriteT1(file.handle, engine);

  % write second table
  __emceeFitsWriteT2(file.handle, engine);

  % write third table
  __emceeFitsWriteT3(file.handle, engine);

  % move back to chain table
  () = _fits_movnam_hdu(file.handle, _FITS_BINARY_TBL, "MCMCCHAIN", 0);

  % set write cycle
  () = _fits_get_rowsize(file.handle, &(file.cycle));
  file.cycle = file.cycle/engine.totalNumberWalkers;
  if (file.cycle < 1)
    file.cycle = 1;

  % fits routine customs
  file.numberSteps = 0;
  file.sloppy = 0;
}
%}}}
%}}}

% Open function %{{{
private define __emceeMikeWriteChecks (file, engine) %{{{
{
  variable handle = file.handle;
  if (_fits_movnam_hdu(handle, _FITS_BINARY_TBL, "PARAMETERS", 0)) {
    fits_close_file(handle);
    handle = NULL;
    throw IOError, "Not a emcee chain file";
  }

  if ((fits_read_key(handle, "STATISTIC") != get_fit_statistic())) {
    fits_close_file(handle);
    handle = NULL;
    throw UsageError, sprintf("Current fit statistic (%s) and chain fit statistic (%s) differ.", fits_read_key(handle, "STATISTIC"), get_fit_statistic());
  }

  if (_fits_movnam_hdu(handle, _FITS_BINARY_TBL, "CHAINSTATS", 0)) {
    fits_close_file(handle);
    handle = NULL;
    throw IOError, "Not a emcee chain file";
  }

  if (_fits_movnam_hdu(handle, _FITS_BINARY_TBL, "MCMCCHAIN", 0)) {
    fits_close_file(handle);
    handle = NULL;
    throw IOError, "Not a emcee chain file";
  }

  if ((fits_read_key(handle, "NWALKERS")*fits_read_key(handle, "NFREEPAR")) != length(engine.walkers)) {
    fits_close_file(handle);
    handle = NULL;
    throw IOError, "Number of walkers differs from number used in chain file";
  }
}
%}}}

private define emceeFileMikeOpen (file, engine) %{{{
{
  file.mode |= EMCEE_FILE_READ | EMCEE_FILE_WRITE;

  file.handle = fits_open_file(file.filename, "w");

  if (_fits_movnam_hdu(file.handle, _FITS_BINARY_TBL, "MCMCCHAIN", 0)) {
    fits_close_file(file.handle);
    file.handle = NULL;
    throw IOError, "Not a emcee chain file";
  }

  __emceeFitsReadChecks(file, engine);
  __emceeMikeWriteChecks(file, engine);

  if (file.cycle < 0)
    file.cycle = 50; % default to 50 cycles per write

  file.numberSteps = fits_get_num_rows(file.handle);
}
%}}}
%}}}

% Read function %{{{
% use the fits read function
%}}}

% Write function %{{{
private define emceeFileMikeWrite (file, engine, numberWalkersSteps) %{{{
{
  if (numberWalkersSteps > engine.leader.writeBuffer.size)
    throw InternalError, "Trying to write more than accessible";

  variable par = __parameters(engine.fit.object);
  variable npar = engine.numberParameters;

  variable i,j;
  variable firstIndex = fits_get_num_rows(file.handle)+1; % first index of this cycle
  variable parCycle = Double_Type[numberWalkersSteps];
  variable stat_min, stat_med, stat_max, frac_update;
  stat_min = Double_Type[numberWalkersSteps/engine.totalNumberWalkers];
  stat_med = Double_Type[numberWalkersSteps/engine.totalNumberWalkers];
  stat_max = Double_Type[numberWalkersSteps/engine.totalNumberWalkers];
  frac_update = Double_Type[numberWalkersSteps/engine.totalNumberWalkers];

  _for j (0, npar-1, 1) {
    _for i (0, numberWalkersSteps-1, 1)
      parCycle[i] = engine.leader.writeBuffer.walkers[i][j];
    () = _fits_write_col(file.handle,
			 fits_get_colnum(file.handle, sprintf("CHAINS%d", par.index[j])),
			 firstIndex,
			 1,
			 parCycle);
  }
  () = _fits_write_col(file.handle,
		       fits_get_colnum(file.handle, "FITSTAT"),
		       firstIndex,
		       1,
		       engine.leader.writeBuffer.stat[[:numberWalkersSteps-1]]);
  () = _fits_write_col(file.handle,
		       fits_get_colnum(file.handle, "UPDATE"),
		       firstIndex,
		       1,
		       engine.leader.writeBuffer.update[[:numberWalkersSteps-1]]);

  () = _fits_movnam_hdu(file.handle, "CHAINSTATS", _FITS_BINARY_TBL, 0);
  firstIndex = fits_get_num_rows(file.handle)+1;
  _for j (0, numberWalkersSteps/engine.totalNumberWalkers-1) {
    stat_min[j] = min(engine.leader.writeBuffer.stat[[0:engine.totalNumberWalkers-1]+j*engine.totalNumberWalkers]);
    stat_med[j] = median(engine.leader.writeBuffer.stat[[0:engine.totalNumberWalkers-1]+j*engine.totalNumberWalkers]);
    stat_max[j] = max(engine.leader.writeBuffer.stat[[0:engine.totalNumberWalkers-1]+j*engine.totalNumberWalkers]);
    frac_update[j] = sum(engine.leader.writeBuffer.update[[0:engine.totalNumberWalkers-1]+j*engine.totalNumberWalkers])/engine.totalNumberWalkers;
  }
  () = _fits_write_col(file.handle, fits_get_colnum(file.handle, "FRAC_UPDATE"), firstIndex, 1, frac_update);
  () = _fits_write_col(file.handle, fits_get_colnum(file.handle, "MIN_STAT"), firstIndex, 1, stat_min);
  () = _fits_write_col(file.handle, fits_get_colnum(file.handle, "MED_STAT"), firstIndex, 1, stat_med);
  () = _fits_write_col(file.handle, fits_get_colnum(file.handle, "MAX_STAT"), firstIndex, 1, stat_max);

  () = _fits_movnam_hdu(file.handle, "MCMCCHAIN", _FITS_BINARY_TBL, 0);
}
%}}}
%}}}

% Close function %{{{
% use fits close
%}}}
%}}}
private define emceeFileMike () %{{{
{
  variable file = struct { @EmceeFile, numberSteps };
  file.create = &emceeFileMikeCreate;
  file.open = &emceeFileMikeOpen;
  file.read = &emceeFileFitsRead;
  file.write = &emceeFileMikeWrite;
  file.close = &emceeFileFitsClose;

  file.filename = qualifier("filename", strftime("emcee-%Y%m%d-%H%M%S.fits"));
  file.mode |= qualifier_exists("parameter") ? EMCEE_FILE_RANGE : 0;
  file.has |= EMCEE_FILE_READ | EMCEE_FILE_WRITE;
  file.cycle = qualifier("cycle", 50);

  file.numberSteps = 0;

  return file;
}
%}}}
EmceeFileRegister["mike"] = &emceeFileMike;

%{{{ Par file functions
%{{{ create function
%}}}

%{{{ open function
%}}}

%{{{ read function
private define emceeFileParRead (file, engine, numberWalkers) %{{{
{
  % we read a number of par files (eventually with multipliers)
  variable file_list = strchop(file.filename, ';', 0);
  variable params = Struct_Type[length(file_list)];  
  variable weight;
  variable i, j, s, p;

  _for i (0, length(file_list)-1) {
    s = strchop(file_list[i], ':', 0);
    if (length(s)>1) weight = atoi(s[1]); % zero, if error, effectively disabling that file
    else weight = 1;

    p = read_par(s[0]);
    if (weight>0) {
      params[i] = struct {
	weight = weight,
	name = String_Type[length(p)],
	value = Double_Type[length(p)],
	min = Double_Type[length(p)],
	max = Double_Type[length(p)],
      };
      _for j (0, length(p)-1) {
	params[i].name = p[j].name;
	params[i].value = p[j].value;
	params[i].min = p[j].min;
	params[i].max = p[j].max;
      }
    }
  }

  return params, numberWalkers;
}
%}}}
%}}}

%{{{ write function
%}}}

%{{{ close function
%}}}
%}}}
private define emceeFilePar () %{{{
{
  variable file = struct { @EmceeFile };
  file.create = &NULL;
  file.open = &NULL;
  file.read = &emceeFileParRead;
  file.write = &NULL;
  file.close = &NULL;

  file.filename = qualifier("filename", strftime("emcee-%Y%m%d-%H%M%S.par"));
  % there is no other mode
  file.mode = EMCEE_FILE_READ | EMCEE_FILE_RANGE;
  file.has = EMCEE_FILE_READ | EMCEE_FILE_RANGE;
  file.cycle = 1;

  return file;
}
%}}}
EmceeFileRegister["par"] = &emceeFilePar;
%}}}

%{{{ Init interface
%!%+
%\function{emcee--init}
%\synopsis{Set emcee initialization function}
%\usage{init="method;parameters";}
%\description
%  The initialization method can be set with the function string
%    "method;parameter"
%  Initialization methods that read from file use the defined input
%  method (default: fits).
%
%  Available methods:
%  uniform : Draw initial walker positions from a uniform distribution
%            within the parameter ranges.
%    ; width : [=1.0] Sub range used for initialization (must be <= 1).
%                     The initiale cube center coincides as much as
%                     possible with the current parameter values.
%
%  gauss   : Draw initial walker positions from a gaussian distribution
%            within parameter ranges.
%    ; sigma : [=0.1] Sigma of the gauss function in terms of the
%                     parameter range. I.e, sigma=1 is the full parameter
%                     range. Allowed range between 1e-3 and 2 (outside will
%                     be cliped).
%
%  file    : Load initial walkers from a valid chain file created by the
%            emcee method. This is used with 'continue'. The parameter ranges
%            can be different compared to a previous run. This is also true
%            for the number of walkers. In order to truely continue a chain
%            those have to be set equal to the previous run.
%
%  chain   : Draw initial walkers from an approximated CDF of an existing
%            chain file.
%    ; steps : [=10] The number of steps to concider for constructing the CDF
%                 (from the end of the chain)
%    ; rng   : [=&rand_uniform] uniform random number generator
%
%!%-
%      1: pick - get walkers from parameters and distribution or file
private variable EmceeInit = struct {
  pick, % function

  % private data
};

private define __emcee_params (par, feed, total) %{{{
{
  variable params = Assoc_Type[UInt_Type, 0]; % need full information to match names == index
  variable n = UInt_Type[length(feed)];
  variable eparams = Struct_Type[length(feed)];
  variable c = 0, cc = 0;
  variable i,j,w,p;

  _for i (0, length(feed)-1) {
    if (feed[i] == NULL) continue;
    ifnot (struct_field_exists(feed[i], "name")) % current set, weight = 1
      c++;
    else
      c += (feed[i].weight>0) ? feed[i].weight : 0;
  }

  foreach p (get_params)
    params[p.name] = p.index;

  _for i (0, length(feed)-1) {
    if (feed[i] == NULL) continue;
    eparams[i] = @Struct_Type(get_struct_field_names(par)); % as prototype
    eparams[i].value = Double_Type[length(par.index)];
    eparams[i].min = Double_Type[length(par.index)];
    eparams[i].max = Double_Type[length(par.index)];

    % use current set of parameters
    ifnot (struct_field_exists(feed[i], "name")) {
      eparams[i].value = par.value;
      eparams[i].min = par.min;
      eparams[i].max = par.max;
      n[i] = int(round(1./c*total));
    } else {
      _for j (0, length(par.index)-1) {
	w = where(par.index[j] == params[feed[i].name]);
	if (length(w) != 1) % also errors if for whatever reason the parameter is twice in feed
	  throw UsageError, sprintf("Unspecified parameter %d", par.index[j]);
	w = w[0];
	if ((par.min[j] > feed[i].min[w]) || (par.max[j] < feed[i].max[w]))
	  throw UsageError, sprintf("Parameter ranges for paramter %d not enclosed", par.index[j]);
	if ((feed[i].min[w[0]] == -DOUBLE_MAX) || (feed[i].max[w[0]] == DOUBLE_MAX))
	  throw UsageError, sprintf("Unspecified ranges for parameter %d", par.index[j]);

	eparams[i].value[w] = feed[i].value[j];
	eparams[i].min[w] = feed[i].min[j];
	eparams[i].max[w] = feed[i].max[j];
      }

      n[i] = int(round(1.*feed[i].weight/c*total));
    }

    cc += n[i];
  }

  n[0] += cc-total;
  return eparams, n;
}
%}}}

%{{{ Uniform initialization function
% pick random parameter values within the boundaries
private define emceeInitUniformPick (init, engine) %{{{
{
  variable i, j, k, feed;
  variable c, p;
  variable file = engine.leader.inFile;
  variable par = __parameters(engine.fit.object);
  variable numParameter = length(par.value);
  variable minb, maxb;

  if (file.mode & EMCEE_FILE_RANGE)
    (feed, ) = file.read(engine, engine.totalNumberWalkers);
  else
    feed = @par;

  minb = _max(par.min, par.value-init.width*0.5*(par.max-par.min));
  maxb = _min(par.max, par.value+init.width*0.5*(par.max-par.min));

  (p,c) = __emcee_params(par, feed, engine.totalNumberWalkers);

  k = 0;
  _for i (0, length(c)-1) {
    _for j (0, c[i]-1)
      engine.walkers[j+k] = _max(p[i].min, _min(p[i].max, rand_uniform(numParameter)*(maxb-minb)+minb));
    k += c[i];
  }
}
%}}}
%}}}
private define emceeInitUniform () %{{{
{
  variable init = struct { @EmceeInit, width };
  init.pick = &emceeInitUniformPick;
  init.width = _min(_max(qualifier("width", 1.0), 1e-3), 1.0);

  return init;
}
%}}}
EmceeInitRegister["uniform"] =  &emceeInitUniform;

%{{{ Gauss initialization function
private define rand_gauss_cut (sigma, v, bmin, bmax) %{{{
{
  variable mu = (v-bmin)/(bmax-bmin);
  variable alpha = -mu/sigma/sqrt(2);
  variable beta = (1.0-mu)/sigma/sqrt(2);
  variable phia = 0.5*(1+gsl->erf(alpha));
  variable phib = 0.5*(1+gsl->erf(beta));
  variable k = phia+rand_uniform(length(v))*(phib-phia);

  return _min(bmax,_max(bmin,gsl->cdf_gaussian_Pinv(k, sigma)*sigma*(bmax-bmin)+v));
}
%}}}

private define emceeInitGaussPick (init, engine) %{{{
{
  variable i,j,k,feed;
  variable c,p;
  variable file = engine.leader.inFile;
  variable par = __parameters(engine.fit.object);
  variable numParameter = length(par.value);

  if (file.mode & EMCEE_FILE_RANGE)
    (feed, ) = file.read(engine, engine.totalNumberWalkers);
  else
    feed = @par;

  (p,c) = __emcee_params(par, feed, engine.totalNumberWalkers);

  k = 0;
  _for i (0, length(c)-1) {
    _for j (0, c[i]-1) {
      engine.walkers[j+k] = rand_gauss_cut(init.sigma, p[i].value, p[i].min, p[i].max);
    }
    k += c[i];
  }
}
%}}}
%}}}
private define emceeInitGauss () %{{{
{
  variable init = struct { @EmceeInit, sigma };
  init.pick = &emceeInitGaussPick;
  init.sigma = _max(_min(qualifier("sigma", 0.1), 2.0), 1e-3);

  return init;
}
%}}}
EmceeInitRegister["gauss"] = &emceeInitGauss;

%{{{ File initialization function
private define fisher_yates (have, want) %{{{
% perform a Fisher Yates shuffle of the
% available walkers. Repeat pattern if more
% are requested.
{
  vmessage("Shuffling and drawing %d from %d", want, have);
  variable p = [0:have-1];
  variable i,j;
  _for j (0, have-2) {
    i = rand_int(j, have-1);
    if (i != j)
      array_swap(p, j, i);
  }

  return p[[0:want-1] mod have];
}
%}}}

private define emceeInitFilePick (init, engine) %{{{
{
  variable file = engine.leader.inFile;
  variable walkerDistribution, readNumber;

  (walkerDistribution, readNumber) = file.read(engine, engine.totalNumberWalkers);

  variable i,j;
  variable par = __parameters(engine.fit.object);
  variable parV = Double_Type[engine.numberParameters];
  variable valid = Int_Type[length(walkerDistribution[0])];
  % filter to walkers in par range
  _for i (0, length(valid)-1)
  {
    _for j (0, length(parV)-1)
    {
      ifnot (par.min[j]<=walkerDistribution[j][i]<=par.max[j])
	break;
      else if (j==(length(parV)-1))
	valid[i] = 1;
    }
  }
  valid = where(valid);
  if (length(valid)<2*engine.numberParameters)
    throw DataError, "Not enough walkers in current parameter range";

  % randomize (and bootstrap if necessary)
  variable randomize = fisher_yates(length(valid), length(engine.walkers));
  _for i (0, length(engine.walkers)-1) {
    _for j (0, length(parV)-1)
      parV[j] = walkerDistribution[j][valid[randomize[i]]];
    engine.walkers[i] = @parV;
  }
}
%}}}
%}}}
private define emceeInitFile () %{{{
{
  variable init = struct { @EmceeInit };
  init.pick = &emceeInitFilePick;

  return init;
}
%}}}
EmceeInitRegister["file"] = &emceeInitFile;

%{{{ Chain initialization function
define empiric_cdf_inverse (p, a, amin, amax) %{{{
{
  variable s = array_sort(p);

  if (p[s][0]<0 || p[s][-1]>=1)
    throw DomainError, "not in range 0<=p<1";

  a = a[array_sort(a)];
  a = a[where(amin<=a<=amax)]; % restrict to cdf in range

  if (length(a)==0) % nothing in par range, fallback to uniform
    a = [0.5*(amin+amax)];

  variable u = unique(a);
  variable ecdf = [u/1./length(a), 1.];
  variable lo = [amin, a[u]];
  variable hi = [a[u], amax];

  variable r = Double_Type[length(p)];
  variable k, i = 0;
  variable m = .5*([(ecdf[[1:]]-ecdf[[:-2]])/(lo[[1:]]-lo[[:-2]]), 0.]
		   +[0., (ecdf[[1:]]-ecdf[[:-2]])/(hi[[1:]]-hi[[:-2]])]);

  variable hitsmin = (amin == a[0]); % gives NaN if true and p == 0
  _for k (0, length(p)-1) {
    while (p[s[k]] > ecdf[i+1]) i++;
    if (hitsmin && p[s[k]]==0)
      r[s[k]] = amin;
    else
      r[s[k]] = (p[s[k]]-ecdf[i])/m[i]+lo[i];
    if (r[s[k]]>amax)
      r[s[k]] = amax;
  }

  return r;
}
%}}}

private define emceeInitChainPick (init, engine) %{{{
{
  variable file = engine.leader.inFile;
  variable walkerDistribution, numberSteps;
  (walkerDistribution, numberSteps) = file.read(engine, init.steps);
  variable par = __parameters(engine.fit.object);;

  variable parRand;
  variable i,j;
  _for i (0, engine.numberParameters-1) {
    parRand = empiric_cdf_inverse(@(init.rng)(engine.totalNumberWalkers),
				  walkerDistribution[i],
				 par.min[i],
				 par.max[i]);
    _for j (0, engine.totalNumberWalkers-1)
      engine.walkers[j][i] = parRand[j];
  }
}
%}}}
%}}}
private define emceeInitChain () %{{{
{
  variable init = struct { @EmceeInit, rng, steps };
  init.pick = &emceeInitChainPick;
  init.rng = qualifier("rng", &rand_uniform);
  init.steps = qualifier("steps", 10);

  return init;
}
%}}}
EmceeInitRegister["chain"] = &emceeInitChain;
%}}}

%{{{ Ship interface:
%!%+
%\function{emcee--driver}
%\synopsis{Set emcee parallel computation method}
%\usage{driver="method;options"}
%\description
%  The driver method can be set with the function string
%    "method;parameter"
%
%  Available methods:
%  serial : The serial driver. No parallelization at all
%
%  fork : The fork (& socket) parallel driver. Per default uses
%          _num_cpus many tasks.
%    ; tasks : [=_num_cpus] Number of total processes used
%
%  mpi : The mpi parallel driver using as many nodes as registered
%        in an mpi environment
%!%-
%      1: setSail - set id for engines and how many there are
%      2: leader_send - leader sends to members
%      3: member_send - members send to leader
%      4: leader_receive - leader receives members
%      5: member_receive - member receive leader
%      6: enterHarbor - cleanup if necessary
%      7: abort - cleanup handler if SIGTERM is received
private variable EmceeShip = struct {
  setSail, % function
  leaderSend, % function
  memberSend, % function
  leaderReceive, % function
  memberReceive, % function
  enterHarbor, % function
  abort = &NULL, % function, has a default

  engine, % the working horse

  % private data
};

%{{{ serial ship functions
private define void () %{{{
{
  variable args = __pop_list(_NARGS);
}
%}}}

private define emceeSerialInit (ship) %{{{
{
  ship.engine.numberEngines = 1;
  ship.engine.id = 0;
}
%}}}
%}}}
private define emceeShipSerial () %{{{
{
  variable ship = struct { @EmceeShip };
  ship.setSail = &emceeSerialInit;
  ship.leaderSend = &void;
  ship.memberSend = &void;
  ship.leaderReceive = &void;
  ship.memberReceive = &void;
  ship.enterHarbor = &void;

  return ship;
}
%}}}
EmceeShipRegister["serial"] = &emceeShipSerial;

%{{{ fork ship functions
private define elementType (t) %{{{
{
  switch (t)
  { case Char_Type:    "c"; }
  { case UChar_Type:   "C"; }
  { case Short_Type:   "h"; }
  { case UShort_Type:  "H"; }
  { case Int_Type:     "i"; }
  { case UInt_Type:    "I"; }
  { case Long_Type:    "l"; }
  { case ULong_Type:   "L"; }
  { case LLong_Type:   "m"; }
  { case ULLong_Type:  "M"; }
  { case Int16_Type:   "j"; }
  { case UInt16_Type:  "J"; }
  { case Int32_Type:   "k"; }
  { case UInt32_Type:  "K"; }
  { case Int64_Type:   "q"; }
  { case UInt64_Type:  "Q"; }
  { case Float_Type:   "f"; }
  { case Double_Type:  "d"; }
  { case Float32_Type: "F"; }
  { case Float64_Type: "D"; }
  { case String_Type:  "s"; }
  { case Null_Type:    "x"; }
}
%}}}

private define writeArray (fp, array) %{{{
{
  variable bytes, msg, fmt;
  fmt = sprintf("%s%d", elementType(_typeof(array)), length(array));
  msg = pack(fmt, array);
  bytes = write(fp, msg);
  return bstrlen(msg)-bytes;
}
%}}}

private define readArray (fp, array) %{{{
{
  variable i, bytes=0, msg=""B, fmt;
  fmt = sprintf("%s%d", elementType(_typeof(array)), length(array));
  variable missing = sizeof_pack(fmt);
  variable total = ""B;
  if (length(array)) {
    while (missing) {
      bytes = read(fp, &msg, missing);
      if (-1 == bytes)
	throw InternalError;
      total += msg;
      missing -= bytes;
    }
    array[*] = unpack(fmt, total);
  }

  return bstrlen(msg);
}
%}}}

private define reap_childs (sig) {
  while (NULL != waitpid(-1, WNOHANG));
}

private define term_process_list (plist) %{{{
{
  variable pid;
  foreach pid (plist) {
    if (pid > 0)
      () = kill(pid, SIGTERM);
  }
}
%}}}

private define emceeForkSetSail (ship) %{{{
{
  variable sockRead, sockWrite;
  variable pid=-1, cid=0, t;
  variable flags;
  variable ppid = getpid();
  variable pids = Int_Type[ship.tasks];
  variable sock = FD_Type[ship.tasks];
  signal(SIGCHLD, &reap_childs);

  _for t (1, ship.tasks-1) {
    (sockRead, sockWrite) = socketpair(AF_UNIX, SOCK_STREAM, 0);
    cid++;
    pid = fork();
    if (pid ==  -1) {
      % need to reap all processes thus far
      term_process_list(pids);
      throw InternalError, sprintf("Unable to fork engine %d", cid);
    } else if (pid == 0) {
      () = close(sockWrite);
      ship.socket = sockRead;
      signal(SIGTERM, SIG_DFL); % make sure term is not disguised
      ship.pids = ppid;
      break;
    } else { % set master pipes
%      flags = fcntl_getfd(sockWrite);
%      fcntl_setfd(sockWrite, flags | O_NONBLOCK);
      sock[cid] = sockWrite;
      () = close(sockRead);
      pids[t] = pid;
    }
  }

  if (0 == pid) {
    ship.engine.id = cid;
  } else {
    ship.pids = pids;
    ship.socket = sock;
    ship.engine.id = 0;
  }

  ship.engine.numberEngines = ship.tasks;
}
%}}}

private define emceeForkLeaderSend (ship) %{{{
{
  variable walkers,
    pivots,
    rolls;
  variable engine = ship.engine;
  variable totalOffset = engine.leader.totalOffset;
  variable walkersPerSet = engine.leader.walkersPerSet;

  variable i,j;
  variable firstIndex = walkersPerSet[0];
  _for i (1, engine.numberEngines-1) {
    % set the walkers for node i
    walkers = engine.walkers[[0:walkersPerSet[i]-1]+firstIndex+totalOffset];
    % pick the pivots for node i
    pivots  = engine.pivots[[0:walkersPerSet[i]-1]+firstIndex+totalOffset];
    % set the randoms for node i
    rolls = engine.rolls[[0:walkersPerSet[i]*engine.gears.step.numberRandoms-1]
			 +(firstIndex+totalOffset)*engine.gears.step.numberRandoms];

    _for j (0, walkersPerSet[i]-1, 1) {
      () = writeArray(ship.socket[i], walkers[j]);
      () = writeArray(ship.socket[i], pivots[j]);
    }
    () = writeArray(ship.socket[i], rolls);
    firstIndex += walkersPerSet[i];
  }
}
%}}}

private define emceeForkMemberSend (ship) %{{{
{
  variable engine = ship.engine;
  variable setOffset = engine.setOffset;
  variable setLength = engine.setLength;

  variable i;
  _for i (0, setLength-1, 1)
    () = writeArray(ship.socket, engine.walkers[i+setOffset]);

  () = writeArray(ship.socket, engine.update[[0:setLength-1]+setOffset]);
  () = writeArray(ship.socket, engine.stat[[0:setLength-1]+setOffset]);
}
%}}}

private define emceeForkLeaderReceive (ship) %{{{
{
  variable walker,
    stat,
    update;

  variable engine = ship.engine;
  variable totalOffset = engine.leader.totalOffset;
  variable walkersPerSet = engine.leader.walkersPerSet;

  variable i,j;
  variable firstIndex = walkersPerSet[0]; % skip master walkers

  walker = Double_Type[engine.numberParameters];
  _for i (1, engine.numberEngines-1) {
    stat   = Double_Type[walkersPerSet[i]];
    update = Int_Type[walkersPerSet[i]];

    _for j (0, walkersPerSet[i]-1, 1) {
      () = readArray(ship.socket[i], walker);
      engine.walkers[j+firstIndex+totalOffset][*] = walker;
    }

    () = readArray(ship.socket[i], update);
    () = readArray(ship.socket[i], stat);

    engine.update[[0:walkersPerSet[i]-1]+firstIndex+totalOffset] = update;
    engine.stat[[0:walkersPerSet[i]-1]+firstIndex+totalOffset]   = stat;

    firstIndex += walkersPerSet[i];
  }
}
%}}}

private define emceeForkMemberReceive (ship) %{{{
{
  variable rolls,
    param;

  variable engine = ship.engine;
  variable setOffset = engine.setOffset;
  variable setLength = engine.setLength; 
  variable nRolls = engine.gears.step.numberRandoms;

  variable j;
  rolls = Double_Type[setLength*nRolls];
  param = Double_Type[engine.numberParameters];
  _for j (0, setLength-1, 1) {
    () = readArray(ship.socket, param);
    engine.walkers[j+setOffset][*] = param;
    () = readArray(ship.socket, param);
    engine.pivots[j+setOffset][*] = param;
  }
  () = readArray(ship.socket, rolls);

  engine.rolls[[0:setLength*nRolls-1]+setOffset*nRolls] = @rolls;
}
%}}}

private define emceeForkEnterHarbor (ship) %{{{
{
  variable id;
  if (ship.engine.id == 0) {
    _for id (1, ship.engine.numberEngines-1)
      () = close(ship.socket[id]);
    term_process_list(ship.pids);
    reap_childs(0);
  } else {
    () = close(ship.socket);
    _exit(0); % no handlers
  }
}
%}}}

private define emceeForkAbort (ship) %{{{
{
  term_process_list(ship.pids);
  reap_childs(0);
}
%}}}
%}}}
private define emceeShipFork () %{{{
{
  variable ship = struct { @EmceeShip, socket, tasks, pids };
  ship.setSail = &emceeForkSetSail;
  ship.leaderSend = &emceeForkLeaderSend;
  ship.memberSend = &emceeForkMemberSend;
  ship.leaderReceive = &emceeForkLeaderReceive;
  ship.memberReceive = &emceeForkMemberReceive;
  ship.enterHarbor = &emceeForkEnterHarbor;
  ship.abort = &emceeForkAbort;
  ship.tasks = qualifier("tasks", _num_cpus());

  return ship;
}
%}}}
EmceeShipRegister["fork"] = &emceeShipFork;

%{{{ MPI Ship functions
#ifexists rcl_mpi_init
private define emceeMPISetSail (ship) %{{{
{
  variable engine = ship.engine;
  engine.id = rcl_mpi_init();
  engine.numberEngines = rcl_mpi_numtasks();
  rcl_init_mpi_request(engine.numberEngines);
}
%}}}

private define emceeMPILeaderSend (ship) %{{{
{
  variable walkers,
    pivots,
    rolls;
  variable engine = ship.engine;
  variable totalOffset = engine.leader.totalOffset;
  variable walkersPerSet = engine.leader.walkersPerSet;

  variable i,j;
  variable firstIndex = walkersPerSet[0]; % skip master walkers

  _for i (1, engine.numberEngines-1) { % loop over the slave nodes and send relevant data
    % set the walkers for node i
    walkers = engine.walkers[[0:walkersPerSet[i]-1]+firstIndex+totalOffset];
    % pick the pivots for node i
    pivots  = engine.pivots[[0:walkersPerSet[i]-1]+firstIndex+totalOffset];
    % set the randoms for node i
    rolls = engine.rolls[[0:walkersPerSet[i]*engine.gears.step.numberRandoms-1]
			 +(firstIndex+totalOffset)*engine.gears.step.numberRandoms];

    _for j (0, walkersPerSet[i]-1, 1) {
      () = rcl_mpi_org_isend_double(walkers[j], length(walkers[j]), i, 0); % send current walkers with tag 0
      () = rcl_mpi_org_isend_double(pivots[j], length(pivots[j]), i, 1); % send pivots from other set with tag 1
    }

    () = rcl_mpi_org_isend_double(rolls, length(rolls), i, 2); % send random numbers with tag 2
    firstIndex += walkersPerSet[i];
  }
}
%}}}

private define emceeMPILeaderReceive (ship) %{{{
{
  variable walker,
    stat,
    update;

  variable engine = ship.engine;
  variable totalOffset = engine.leader.totalOffset;
  variable walkersPerSet = engine.leader.walkersPerSet;

  variable i,j;
  variable firstIndex = walkersPerSet[0]; % skip master walkers

  walker = Double_Type[engine.numberParameters];
  _for i (1, engine.numberEngines-1) {
    stat   = Double_Type[walkersPerSet[i]];
    update = Int_Type[walkersPerSet[i]];

    _for j (0, walkersPerSet[i]-1, 1) {
      () = rcl_mpi_org_recv_double(walker, length(walker), i, i);
      engine.walkers[j+firstIndex+totalOffset][*] = walker;
    }

    () = rcl_mpi_org_recv_int(update, length(update), i, i);
    () = rcl_mpi_org_recv_double(stat, length(stat), i, i);

    engine.update[[0:walkersPerSet[i]-1]+firstIndex+totalOffset] = update;
    engine.stat[[0:walkersPerSet[i]-1]+firstIndex+totalOffset]   = stat;

    firstIndex += walkersPerSet[i];
  }
}
%}}}

private define emceeMPIMemberSend (ship) %{{{
{
  variable engine = ship.engine;
  variable setOffset = engine.setOffset;
  variable setLength = engine.setLength;

  variable i;
  _for i (0, setLength-1, 1) {
    () = rcl_mpi_org_isend_double(engine.walkers[i+setOffset],
				  length(engine.walkers[0]), 0, engine.id);
  }

  () = rcl_mpi_org_isend_int(engine.update[[0:setLength-1]+setOffset], setLength, 0, engine.id);
  () = rcl_mpi_org_isend_double(engine.stat[[0:setLength-1]+setOffset], setLength, 0, engine.id);
}
%}}}

private define emceeMPIMemberReceive (ship) %{{{
{
  variable rolls;
  variable engine = ship.engine;
  variable setOffset = engine.setOffset;
  variable setLength = engine.setLength; 
  variable nRolls = engine.gears.step.numberRandoms;

  variable j;
  rolls = Double_Type[setLength*nRolls];
  _for j (0, setLength-1, 1) {
    () = rcl_mpi_org_recv_double(engine.walkers[j+setOffset],
				 length(engine.walkers[j+setOffset]), 0, 0); % receive walkers (tag 0)
    () = rcl_mpi_org_recv_double(engine.pivots[j+setOffset],
				 length(engine.pivots[j+setOffset]), 0, 1); % receive pivot points (tag 1)
  }
  () = rcl_mpi_org_recv_double(rolls,
			       setLength*engine.gears.step.numberRandoms, 0, 2); % receive random numbers (tag 2)

  engine.rolls[[0:setLength*nRolls-1]+setOffset*nRolls] = @rolls;
}
%}}}

private define emceeMPIEnterHarbor (ship) %{{{
{
  
}
%}}}

private define emceeMPIAbort (ship) %{{{
{
#ifexists rcl_mpi_abort
  rcl_mpi_abort(1); % hopefully 0 is allway success
#endif
  throw InternalError;
}
%}}}
%}}}
private define emceeShipMPI () %{{{
{
  variable ship = struct { @EmceeShip };
  ship.setSail = &emceeMPISetSail;
  ship.leaderSend = &emceeMPILeaderSend;
  ship.memberSend = &emceeMPIMemberSend;
  ship.leaderReceive = &emceeMPILeaderReceive;
  ship.memberReceive = &emceeMPIMemberReceive;
  ship.enterHarbor = &emceeMPIEnterHarbor;
  ship.abort = &emceeMPIAbort;

  return ship;
}
%}}}
EmceeShipRegister["mpi"] = &emceeShipMPI;
#endif
%}}}

%{{{ Step interface:
%!%+
%\function{emcee--step}
%\synopsis{Set emcee step algorithm}
%\usage{step="method;options"}
%\description
%  The step algorithm can be set with the function string
%    "method;parameter"
%
%  Available algorithms:
%  stretch : The stretch move as described in Goodman & Weare 2010
%    ; scale : [=2] Scale for the range of possible moves
%!%-
%      1: move - loop over walkers and update
private variable EmceeStep = struct {
  move, % function

  numberRandoms, % random number required per step

  % private data
};

%{{{ Stretch move functions (Foreman & Mackey)
% define inverse cumulative distribution function for generating
% random numbers following 1/z^2 when z in [1/a, a]
% TODO: should make this an adjustable thing
private define stretchInverseCDF (u, a) %{{{
{
  return (u*(a-1.)+1.)^2./a;
}
%}}}

% stretch move as of Goodman & Weare 2010
% Move must evaluate the fit function
private define emceeStepStretchMove (step, engine) %{{{
{
  variable j;
  variable z;
  variable proposed;
  variable newStat;
  variable startIndex = engine.setOffset;
  variable setLength = engine.setLength;

  _for j (startIndex, startIndex+setLength-1) {
    z = stretchInverseCDF(engine.rolls[j*step.numberRandoms], step.scale);
    proposed = engine.pivots[j] + z*(engine.walkers[j]-engine.pivots[j]);

    engine.update[j] = 0;

    try {
      newStat = engine.fit.eval_statistic(proposed; nocopy);

      % accept or reject dimansionally normalized. Assuming statistic is -2 log likelihood
      if (log(engine.rolls[j*step.numberRandoms+1])
	  <= (log(z)*(engine.fit.num_vary-1)+(engine.stat[j]-newStat)/2.)) {
	engine.stat[j] = newStat;
	engine.walkers[j][*] = proposed;
	engine.update[j] = 1;
      }
    } catch IsisError;
  }
}
%}}}
%}}}
private define emceeStepStretch () %{{{
{
  variable step = struct { @EmceeStep, scale };
  step.move = &emceeStepStretchMove;
  step.numberRandoms = 2;
  step.scale = qualifier("scale", 2);

  return step;
}
%}}}
EmceeStepRegister["stretch"] = &emceeStepStretch;
%}}}

%{{{ Progress interface
%!%+
%\function{emcee--progress}
%\synopsis{Set emcee progress report}
%\usage{progress="method;options"}
%\description
%  To get a progress report for the running emcee
%  algorithm use one of the available options.
%
%  Available report methods:
%  none   : Do not report
%  report : Report the number of steps done every n steps
%    ; n : [=50] Report for every n steps.
%    ; overwrite : if given, overwrite last status (useful for
%       interactive sessions).
%    ; format : [="Status: %D/%T (%%P)"] The report format
%       where %D is the current step, %T total steps and %P
%       the percentage.
%!%-
private variable EmceeProgress = struct {
  reporter, % function
  finish, % function

  every = 50, % when to change report

  % private data
};

%{{{ None
private define emceeProgressNone () %{{{
{
  variable progress = struct { @EmceeProgress };
  progress.reporter = &NULL;
  progress.finish = &NULL;
  progress.every = -1;

  return progress;
}
%}}}
%}}}
EmceeProgressRegister["none"] = &emceeProgressNone;

%{{{ Report
private define emceeProgressReportReporter (progress, engine, step)
{
  variable str = strreplace(progress.format, "%D", sprintf("%d", step));
  str = strreplace(str, "%T", sprintf("%d", engine.numberSteps));
  str = strreplace(str, "%P", sprintf("%.02lf", step*100./engine.numberSteps));
  if (engine.numberSteps == step) {
    progress.last_out = printf("%c%s\n", (progress.overwrite ? '\r' : '\0'), str);
    progress.last = 1;
  } else {
    progress.last_out = printf("%c%s%c", (progress.overwrite ? '\r' : '\0'), str, (progress.overwrite ? '\0' : '\n'));
    () = fflush(stdout);
  }
}
private define emceeProgressReportFinish (progress, engine, step)
{
  emceeProgressReportReporter(progress, engine, step);

  () = printf("-- DONE --\n");
}
%}}}
private define emceeProgressReport () %{{{
{
  variable progress = struct { @EmceeProgress, last_out = 0, overwrite, format, last = 0 };
  progress.reporter = &emceeProgressReportReporter;
  progress.finish = &emceeProgressReportFinish;
  progress.every = qualifier("n", 50);
  progress.overwrite = qualifier_exists("overwrite");
  progress.format = qualifier("format", "Status: %D/%T (%%P)");

  return progress;
}
%}}}
EmceeProgressRegister["report"] = &emceeProgressReport;
%}}}

%%% emcee call
private define emceeOption (str) %{{{
{
  variable s = strchop(str, ';', 0);
  return strtrim(s[0]), length(s)>1 ? eval(sprintf("struct {%s}", s[1])) : NULL;
}
%}}}

private variable EMCEE_ABORT = 0;
private variable EMCEE_OLD_HANDLE = SIG_DFL;
private define emceeLoop (ship, step, output) %{{{
{
  variable engine = ship.engine;
  variable s=0, j, set, cycle = 0, leader, size, offset, timer;

  offset = 0;
  tic; % start timer

  if (0 == engine.id) {
    leader = engine.leader;
    if (leader.progress.every > 0)
      leader.progress.reporter(engine, 0);
  }

  try {
    _for s (0, engine.numberSteps-1) {
      if (EMCEE_ABORT) break;
      _for set (1, 2) {
	if (EMCEE_ABORT) break;
	emceeDrawSet(engine, set);

	if (0 == engine.id)
	  ship.leaderSend();
	else
	  ship.memberReceive();

	step.move(engine);

	if (0 == engine.id)
	  ship.leaderReceive();
	else
	  ship.memberSend();
      }
      timer = toc; % get elapsed time

      if (0 == engine.id) {
	size = leader.writeBuffer.size;
	cycle = (s-offset) mod leader.writeBuffer.cycle;

	% write to buffer
	_for j (0, engine.totalNumberWalkers-1) {
	  leader.writeBuffer.walkers[j+cycle*engine.totalNumberWalkers][*] = @(engine.walkers[j]);
	  leader.writeBuffer.stat[j+cycle*engine.totalNumberWalkers] = engine.stat[j];
	  leader.writeBuffer.update[j+cycle*engine.totalNumberWalkers] = engine.update[j];
	}

	if ((leader.progress.every > 0) && not (s mod leader.progress.every))
	  leader.progress.reporter(engine, s+1);

	if (cycle == (leader.writeBuffer.cycle-1)) {
	  output.write(engine, size);
	  output.close(engine); % flush output
	  output.open(engine);
	  offset = 0; % we write full buffer, so no offset at all
	  tic; % restart timer
	} else if (timer >= 6e2) { % write every 10 min
	  output.write(engine, (cycle+1)*engine.totalNumberWalkers);
	  output.close(engine); % flush output
	  output.open(engine);
	  offset = cycle mod leader.writeBuffer.cycle;
	  tic; % restart timer
	}
      }
    }
  }
  catch UserBreakError: % make sure we stop gracefully
  {
    EMCEE_ABORT = 1;
  }

  % write remaining steps
  if (0 == engine.id) {
    if (cycle < (leader.writeBuffer.cycle-1)) {
      output.write(engine, (cycle+1)*engine.totalNumberWalkers);
      if (leader.progress.every > 0)
	leader.progress.reporter(engine, engine.numberSteps);
    }
    output.close(engine);
  }

  if (EMCEE_ABORT && (&NULL != ship.abort))
    ship.abort();
  EMCEE_ABORT=0;

  % set signal handler back
  signal(SIGTERM, EMCEE_OLD_HANDLE);

  if (0 == engine.id) {
    if (leader.progress.every > 0)
      leader.progress.finish(engine, engine.numberSteps);
  }
}
%}}}

% catch SIGTERM and set abort flag
private define emceeSignalHandler (sig) %{{{
{
  EMCEE_ABORT = 1;
  signal(SIGTERM, EMCEE_OLD_HANDLE);
}
%}}}

private define file_exists (filename) %{{{
{
  variable s = stat_file(filename);
  if (s == NULL) return 0;
  return not stat_is("dir", s.st_mode);
}
%}}}

private define emceeSetup (ship, steps, options) %{{{
{
  variable leader, size, engine;
  variable j, set, err;
  
  engine = ship.engine;
  if (0 == engine.id) {
    emceeSetupLeader(engine, options.input, options.output, options.progress);
    options.init.pick(engine);
  }

  % set walkers and eval once
  _for set (1, 2) {
    emceeDrawSet(engine, set);

    if (0 == engine.id)
      ship.leaderSend();
    else
      ship.memberReceive();
  }

  _for j (0, length(engine.walkers)-1)
    engine.stat[j] = engine.fit.eval_statistic(engine.walkers[j]; nocopy);

    % set handler
  signal(SIGTERM, &emceeSignalHandler, &EMCEE_OLD_HANDLE);

  if (0 == engine.id) {
    if (get_struct_field(options, "continue"))
    {
      ifnot (file_exists(options.output.filename))
      {
	if (ship.abort != NULL)
	{
	  vmessage("File '%s' does not exist. Aborting", options.output.filename);
	  ship.abort();
	}

	throw ReadError, sprintf("File '%s' does not exist", options.output.filename);
      }
      try (err) {
	options.output.open(engine);
      }
      catch AnyError:
      {
	if (ship.abort != NULL)
	  ship.abort();
	throw err.error, err.message; % forward error, make sure we aborted
      }
    }
    else
    {
      if (file_exists(options.output.filename) && not (options.clobber))
      {
	if (ship.abort != NULL)
	{
	  vmessage("File '%s' exists. Refusing to overwrite.", options.output.filename);
	  ship.abort();
	}

	throw WriteError, sprintf("File '%s' exists.", options.output.filename);
      }
      try (err) {
	options.output.create(engine);
      }
      catch AnyError:
      {
	if (ship.abort != NULL)
	  ship.abort();
	throw err.error, err.message; % forward error
      }
    }

    leader = engine.leader;

    % write initial walkers to buffer
    _for j (0, length(engine.walkers)-1) {
      leader.writeBuffer.walkers[j] = @(engine.walkers[j]);
      leader.writeBuffer.stat[j] = engine.stat[j];
      leader.writeBuffer.update[j] = 1;
    }

    % if we create new file write initial walkers to it
    ifnot (get_struct_field(options, "continue"))
      options.output.write(engine, engine.totalNumberWalkers);
  }
}
%}}}

private define emceeDefaultFile (file) %{{{
{
  variable default = "fits";
  if (NULL != file) {
    variable ext = path_extname(file);
    if (strlen(ext))
      return ext[[1:]];
  }

  return default;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%
define emcee_hammer (steps)
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{emcee_hammer}
%\synopsis{Explore parameter space with MCMC method}
%\usage{emcee_hammer (Int_Type);}
%#c%{{{
%\qualifiers{
%  \qualifier{Basic Qualifiers}{}
%  \qualifier{walkers}{[=10]: Number of walkers per parameter}
%  \qualifier{continue}{If given (and possible set to a file) continue chain from this file
%                         (using init="file", file="fits" per default)}
%  \qualifier{cont}{same as 'continue'}
%  \qualifier{infile}{Set the input file name for reading and continuing}
%  \qualifier{outfile}{Set the output file name}
%  \qualifier{clobber}{Ovwerwrite output file if exists}
%  \qualifier{Advanced Qualifiers}{}
%  \qualifier{init}{[="uniform" or "file"] The walker initialization method}
%  \qualifier{driver}{[="mpi"] The parallelization method}
%  \qualifier{step}{[="stretch"] The walker step algorithm}
%  \qualifier{input}{[="fits"] The file reading method}
%  \qualifier{output}{[="fits"] The file writing method}
%  \qualifier{progress}{="none"] Show progres}
%  \qualifier{urand}{[=&rand_uniform] PRNG for uniform numbers (Double_Type[] = urand(Int_Type))}
%  \qualifier{upick}{[=&rand_int] PRNG to chose complement walker (Int_Type[] = upick(Int_Type, Int_Type, Int_Type))}
%}
%
%\description
%  The MCMC parameter space exploration algorithm as described by
%  Foreman-Mackey et al. The function expects that data and a model is loaded.
%  The only input parameter gives the number of iterations the algorithm
%  performs. The resulting walker positions are written to a file which can
%  be set with the "outfile" qualifier.
%
%  The function allows to choose other algorithms for the step proposition,
%  the read and write routines and how the walker ensamble is initialized.
%  To get more information about the methods read 'help emcee_<method>'.
%
%  Per default a new chain is started when the function is called. To continue
%  a chain use the "continue" qualifier. If the intention is to continue a
%  previous chain, make sure the paremter ranges are set to the same values as
%  for the previous run. \code{emcee_hammer} always uses the current settings.
%
%  To set a prior for parameters use the \code{set_fit_constraint} interface.
%  This ensures that the stored fit values are according to the posterior
%  and not just the likelihood.
%
%\seealso{emcee--init, emcee--step, emcee--driver, emcee--input, emcee--file, emcee--progress}
%!%-
{
  % options
  variable oContinue = qualifier("continue", qualifier("cont"));
  variable oInfile = qualifier("infile", oContinue);
  variable oOutfile = qualifier("outfile", oContinue);
  variable oInread = emceeDefaultFile(oInfile);
  variable oOutwrite = emceeDefaultFile(oOutfile);

  % advanced options
  variable shipHandle, shipOption;
  % special case driver: If we are in an MPI environment use the MPI driver per default
#ifexists rcl_mpi_init
  (shipHandle, shipOption) = emceeOption(qualifier("driver", "mpi"));
#else
  (shipHandle, shipOption) = emceeOption(qualifier("driver", "serial"));
#endif

  variable stepHandle, stepOption;
  (stepHandle, stepOption) = emceeOption(qualifier("step", "stretch"));

  variable inputHandle, inputOption;
  (inputHandle, inputOption) = emceeOption(qualifier("input", oInread));
  if (NULL != oInfile) inputOption = struct { @inputOption, filename=oInfile };

  variable progressHandle, progressOption;
  (progressHandle, progressOption) = emceeOption(qualifier("progress", "none"));

  variable outputHandle, outputOption;
  (outputHandle, outputOption) = emceeOption(qualifier("output", oOutwrite));
  if (NULL != oOutfile) outputOption = struct { @outputOption, filename=oOutfile };

  variable totalNumberWalkers = qualifier("walkers", 10)*num_free_params();
  variable ship = @(EmceeShipRegister[shipHandle])(;;shipOption);
  emceeSetupEngine(ship, totalNumberWalkers, steps);

  variable Input = NULL, Output = NULL, Progress = NULL;
  variable Initfile = "uniform";
  if (0 == ship.engine.id) {
    Input = @(EmceeFileRegister[inputHandle])(;;inputOption);
    Output = @(EmceeFileRegister[outputHandle])(;;outputOption);
    Progress = @(EmceeProgressRegister[progressHandle])(;;progressOption);

    if (NULL == Input)
      throw UsageError, sprintf("Input handler '%s' is unknown'", inputHandle);
    if (NULL == Output)
      throw UsageError, sprintf("Output handler '%s' is unknown'", outputHandle);
    if (NULL == Progress)
      throw UsageError, sprintf("Progress handler '%s' is unknown'", progressHandle);

    if ( not (Input.has & EMCEE_FILE_READ)
	 || ( Input.read == &NULL ) )
      throw UsageError, sprintf("File handle '%s' can not be used for input", inputHandle);
    if ( not (Output.has & EMCEE_FILE_WRITE)
	 || ( Output.write == &NULL ) )
      throw UsageError, sprintf("File handle '%s' can not be used for output", outputHandle);

    ifnot (Input.has & EMCEE_FILE_RANGE)
      Initfile = "file";
  }

  variable initHandle, initOption;
  if (NULL != oInfile)
    (initHandle, initOption) = emceeOption(qualifier("init", Initfile));
  else
    (initHandle, initOption) = emceeOption(qualifier("init", "uniform"));

  variable options = @Struct_Type(["init", "step", "output", "input", "urand", "upick", "progress", "continue", "clobber"]);
  
  options.init = @(EmceeInitRegister[initHandle])(;;initOption);
  options.step = @(EmceeStepRegister[stepHandle])(;;stepOption);
  options.output = Output;
  options.input = Input;
  options.urand = qualifier("urand", &rand_uniform);
  options.upick = qualifier("upick", &rand_int);
  options.progress = Progress;
  set_struct_field(options, "continue", qualifier_exists("continue") || qualifier_exists("cont"););
  options.clobber = qualifier_exists("clobber") || qualifier_exists("overwrite");

  emceeSetupGears(ship.engine, options.urand, options.upick, options.step);

  emceeSetup(ship, steps, options);

  emceeLoop(ship, options.step, options.output);

  ship.enterHarbor();
}
%}}}

#endif

#endif

#endif

%
% Code from src/slang/structs/empty_struct.sl
%
define empty_struct()
%!%+
%\function{empty_struct}
%\synopsis{Returns a struct with 0 fields}
%\usage{ Struct_Type = empty_struct();}
%\description
%  This function returns a struct with 0 fields,
%  which is usable with other funtions, e.g.,
%  struct_field_exists.
%!%-
{
  return @Struct_Type(String_Type[0]);
}

%
% Code from src/slang/math/enclosing_ellipse.sl
%
private define r_e (a,b,p,t) % distance from the origin to the point on ellipse at phase t
{
  variable y=qualifier("y",1.);
  return sqrt(cos(t)^2*a^2*(cos(p)^2+y^2*sin(p)^2)+
	      sin(t)^2*b^2*(y^2*cos(p)^2+sin(p)^2)+
	      sin(t)*cos(t)*2*a*b*(y^2-1)*cos(p)*sin(p));
}

private define t_m (a,b,p,y) % phase parameter of ellipse where r_e is maximal (minimal?)
{
  return 0.5*atan( 2*a*b*cos(p)*sin(p)*(y^2-1.)/
		   (a^2*(cos(p)^2+y^2*sin(p)^2)-b^2*(y^2*cos(p)^2+sin(p)^2)));
}

%%%%%%%%%%%%%%%%%%%%%%%%%%
define enclosing_ellipse()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{enclosing_ellipse}
%\synopsis{calculates the smallest ellipse enclosing two ellipses centered at the origin}
%\usage{(Double_Type smaj,smin,posang) = enclosing_ellipse(Double_Type smaj1, smin1, posang1, smaj2, smin2, posang2)
%}
%\description
%    This function calculates the semimajor axis \code{smaj}, the semiminor axis \code{smin},
%    and the position angle \code{posang} of the smallest ellipse enclosing
%    two ellipses centered at the origin.
%\examples
%
%    smaj1 = 5.; smin1 = 2.; posang1 = 0.3;
%    smaj2 = 4.; smin2 = 1.; posang2 = 0.7*PI;
%    variable phi=[0:2*PI:#200];
%    
%    plot ( ellipse( enclosing_ellipse(smaj1, smin1, posang1, smaj2, smin2, posang2) ,phi) );
%    oplot( ellipse(smaj1, smin1, posang1 ,phi) );
%    oplot( ellipse(smaj2, smin2, posang2 ,phi) );
%\seealso{ellipse}
%!%-
{
  variable a1,b1,p1,a2,b2,p2;
  switch(_NARGS)
  { case 6: (a1,b1,p1,a2,b2,p2) = (); }
  { help(_function_name()); return; }
  
  (a1, b1,   a2, b2) = (abs(a1), abs(b1),    abs(a2),abs(b2)); % using only positive axes
  if (b1 > a1) (a1,b1,p1) = (b1,a1,p1+0.5*PI); % make sure that a1 is the semimajor axis
  if (b2 > a2) (a2,b2,p2) = (b2,a2,p2+0.5*PI); % make sure that a2 is the semimajor axis

  if ( a1 <= b2 ) return (a2,b2,p2);
  if ( a2 <= b1 ) return (a1,b1,p1);
  if ( a1==a2 && b1==b2 && ( (p1-p2) mod PI) == 0 ) return (a1,b1,p1);
  
   if (a2 < a1)  % make a2 the larger semimajor, check if necessary!
  {
    variable tmp = a2; a2 = a1; a1=tmp;
    tmp = b2; b2 = b1; b1 = tmp;
    tmp = p2; p2 = p1; p1 = tmp;
  }
  variable dp = p1;
  p1 -=dp;
  p2 -=dp;
  
  variable Y = a1/b1;
  
  variable A = r_e(a2,b2,p2,t_m(a2,b2,p2,Y);y=Y); % or: max(a1,r_e(a2,b2,p2,t_m(a2,b2,p2,Y);y=Y));
  variable B = max([r_e(a2,b2,p2,t_m(a2,b2,p2,Y)+0.5*PI;y=Y),a1]);
  
  variable mx,my;
  (mx,my)= ellipse (a2,b2,p2,t_m(a2,b2,p2,Y);y=Y);
  variable P = atan2(my,mx);
  
  variable a = r_e(A,B,P,t_m(A,B,P,1./Y);y=1./Y);
  variable b = r_e(A,B,P,t_m(A,B,P,1./Y)+0.5*PI;y=1./Y);
  (mx,my)= ellipse (A,B,P,t_m(A,B,P,1./Y);y=1./Y);
  variable p = atan2 (my,mx);
  p +=dp;
  if (b > a) (a,b,p) = (b,a, p+0.5*PI); % make sure that 'a' is the semimajor axis
  return (a,b,p mod PI);
}

%
% Code from src/data/spectrum.datastructures/energyflux.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define energyflux()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{energyflux}
%\synopsis{evaluates the energy flux of the current fit model in a given energy range}
%\usage{Double_Type energyflux(Double_Type Emin, Emax)}
%\qualifiers{
%\qualifier{factor}{[=\code{1.001}] step size of the logarithmic energy grid}
%\qualifier{Emin}{minimum energy of (extended) grid}
%\qualifier{Emax}{maximum energy of (extended) grid}
%\qualifier{cgs}{return flux in erg/cm^2/s}
%}
%\description
%    This function calculates the energy flux of the current best fit model
%    in keV/cm^2/s by integrating over the model on a logarithmic energy grid. 
%    The logarithmic step size of the model is given by the \code{factor}
%    qualifier.
%    
%    In other words, energyflux returns\n
%      \code{int_{Emin}^{Emax} E * S_E(E) dE}  (in keV/s/cm^2)
%    where \code{S_E(E)} (in 1/s/cm^2/keV) is defined by \code{fit_fun}.
%    
%    If the fit-function is a convolution model, e.g., Compton reflection,
%    it may be necessary to use an extended grid defined by the \code{Emin} 
%    and \code{Emax} qualifiers.
%
%    Note: While physicists generally use the term ``flux density'' for 
%    the quantity returned by this function, with the exception of radio
%    astronomy astronomers generally call the returned value the ``flux''.
%    The change in function name is based on the experience that most
%    users of isisscripts did not recognize that this function is what
%    they required.
%
%\seealso{eval_fun, eval_fun_keV, luminosity}
%!%-
{
  variable Emin, Emax;
  switch(_NARGS)
  { case 2:  (Emin, Emax) = (); }
  { help(_function_name()); return; }

  variable Elo = qualifier("Emin", Emin);
  variable Ehi = qualifier("Emax", Emax);
  if(Elo > Emin) { 
    vmessage("warning (%s): extending grid from Emin = %S keV to Emin = %S keV", _function_name(), Elo, Emin);
    Elo = Emin;
  }
  if(Ehi < Emax) { 
    vmessage("warning (%s): extending grid from Emax = %S keV to Emax = %S keV", _function_name(), Ehi, Emax);
    Ehi = Emax;
  }
  variable default_factor = 1.001;
  variable factor = qualifier("factor", default_factor);
  if(factor<=1) { 
    vmessage("warning (%s): factor = %S <= 1 is not allowed, using factor = %S", _function_name(), factor, default_factor);
    factor = default_factor;
  }
  variable n = int( log(double(Ehi)/Elo)/log(factor) +1 );
  (Elo, Ehi) = log_grid(Elo, Ehi, n);
  variable photons_s_cm2 = eval_fun_keV(Elo, Ehi);
  variable keV_s_cm2 = photons_s_cm2 * .5*(Elo+Ehi);
  variable totflux=sum( keV_s_cm2[where(Emin <= Elo and Ehi <= Emax)] );

  if (qualifier_exists("cgs")) {
      totflux=totflux*1.60217646e-9;   % keV -> erg
  }
  return totflux;
}

%
% Code from src/data/spectrum.datastructures/energyfluxdensity.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define energyfluxdensity()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{energyfluxdensity}
%\synopsis{deprecated function}
%\usage{do not use}
%\description
%    This function had the wrong name. Please use the energyflux function
%    instead.
%\seealso{energyflux}
%!%-
{
  vmessage("The function energyfluxdensity has been deprecated because of\n");
  vmessage("its incorrect name. Please use the energyflux function instead.");
}

%
% Code from src/fitting/fit-functions/enflux.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define enflux_fit(lo, hi, par, y)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{enflux (fit-function)}
%\synopsis{fits the photon flux in a given energy range}
%
%\qualifiers{
%    \qualifier{norm}{If given a reference to a variable, store the
%                     normalization factor in there.}
%}
%\description
%    This function can be used as a convolution model to determine
%    the energy flux [keV/s/cm^2] of the model in the energy range
%    given by \code{E_min} and \code{E_max}. Only bins within(!) this energy range
%    are considered for the calculation of the flux. The energy
%    flux is calculated by multiplying the photon flux in each bin
%    by the mean energy of the bin. For that reason the model should
%    be evaluated on a fine grid. It is strongly recommended to use
%    a very fine user grid for this purpose.
%    As this function fits the normalization of the total convolved
%    model, the normalizations of its components are not defined
%    absolutely, but only relativ to each others. For that reason it
%    is meaningful to freeze the normalization of one component, at
%    best the one of the continuum to avoid ambiguities during
%    fitting.
%
%    IMPORTANT:
%    1) E_min and E_max have to be completely covered by the data grid
%    2) the function requires a fine grid for evaluation
%    3) absolute normalizations of functions convolved with enflux are useless
%    
%    If the 1) and 2) are not fulfilled by the data (RMF grid), a user
%    grid has to be used (see example). In case 1) the command:
%    \code{set_kernel (data_id, "std;eval=all");} is required in order to
%    evaluate the model on bins outside of the range of data set \code{data_id}
%    (see set_eval_grid_method).
%
%\examples
%    % data definition:
%    \code{variable lo = _A([1:10]);}
%    \code{variable hi = make_hi_grid(lo);}
%    \code{variable my_data = define_counts(lo,hi,lo,sqrt(lo));}\n
%    \code{variable my_emin = 2.5;}
%    \code{variable my_emax = 6.5;}
%    % defining a fine user grid
%    \code{define fine_grid(id, s)}
%    \code{\{}
%    \code{   (s.bin_lo,s.bin_hi) = _A(log_grid(1,10,1000));}
%    \code{   return s;}
%    \code{\}}
%    
%    \code{set_eval_grid_method (USER_GRID, my_data, &fine_grid);}\n
%    \code{fit_fun("enflux(1,powerlaw(1))");}
%
%    \code{set_par("enflux(1).E_min",  my_emin,  1); % keV}
%    \code{set_par("enflux(1).E_max",  my_emax,  1); % keV}
%    \code{freeze("powerlaw(1).norm");}
%    \code{()=fit_counts();}
%    \code{list_par;}
%
%
%    % It is also possible to determine only the flux of certain
%    % model components, e.g., the unabsorbed flux:
%    \code{fit_fun(phabs(1)*enflux(1, powerlaw(1)));}
%
%\seealso{phflux, set_eval_grid_method}
%!%-
{
  variable flux = par[0];
  variable l1 = _A(par[2]);
  variable l2 = _A(par[1]);

  variable norm_value;
  variable store_norm = qualifier("norm");

  if (Fit_Verbose >= 0 && (l1 < lo[0] or l2 > hi[-1]))
    vmessage("warning: enflux energy range (%.2e-%.2e keV) not covered by data grid (%.2e-%.2e keV)",
	     par[1],par[2],_A(hi[-1]),_A(lo[0]));
  variable i = where(l1 <= lo and hi <= l2);
  variable mean_e = _A(1.)/(0.5*(hi+lo)[i]);
  norm_value = sum( mean_e*y[i] );

  if (typeof(store_norm) == Ref_Type)
    @store_norm = norm_value;
    
  return y * flux / norm_value;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define enflux_default(i)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{ switch(i)
  { case 0: return struct{value=0.1, freeze=0, min=0, max=1e20, hard_min=0, hard_max=1e100, step=1e-5, relstep=1e-3};}
  { case 1: return (3.0 , 1, 0, 1e20 ); }
  { case 2: return (10. , 1, 0, 1e20 ); }
}

add_slang_function("enflux", ["enflux [keV/s/cm^2]", "E_min [keV]", "E_max [keV]"], [0]);
set_param_default_hook("enflux", "enflux_default");
set_function_category("enflux", ISIS_FUN_OPERATOR);

%
% Code from src/misc/image/enlarge_image.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%%%%%%%%%%%%%%%%%%%%
define enlarge_image()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{enlarge_image}
%\synopsis{enlarge an image to subpixel resolution by 2d interpolation}
%\usage{Double_Type IMG = enlarge_image(img, Integer_Type n);
%\altusage{Double_Type IMG = enlarge_image(img, Integer_Type nx, ny);}
%}
%\qualifiers{
%\qualifier{interp}{reference to interpolation function (see below)}
%\qualifier{y_first}{interpolate first in y-, then in x-direction}
%}
%\description
%    The pixels of \code{img} are mapped to every \code{nx}-th pixel
%    in x-direction and every \code{ny}-th pixel in y-direction
%    (in the first usage, nx = ny = n) of \code{IMG}:
%#v+
%       IMG[ [::ny], [::nx] ] = img;
%#v-
%    As no extrapolation is performed at the boundary,
%    width and height of \code{IMG} are smaller than \code{nx*w} and \code{ny*h},
%    where \code{w} and \code{h} are width and height of \code{img}.
%
%    Intermediate pixels are interpolated using the function
%    specified by the \code{interp} qualifier, which defaults to
%    \code{gsl->interpol_cspline} if the gsl module is available,
%    and otherwise to ISIS' \code{interpol} function. In general,
%    it can be a reference to any function of the form
%#v+
%       newy[] = interpol(newx[], oldx[], oldy[]);
%#v-
%    All qualifiers of \code{enlarge_image} are passed to \code{@interp}.
%
%    The interpolation is first performed in x-direction
%    and then in y-direction -- unless the \code{y_first} qualifier
%    is specified. For linear and cubic spline interpolation,
%    the final result is independent of the order.
%!%-
{
  variable img, nx, ny;
  switch(_NARGS)
  { case 2: (img, nx) = (); ny = nx; }
  { case 3: (img, nx, ny) = (); }
  { return help(_function_name()); }

  % dimensions of original image img
  variable dims = array_shape(img);
  variable y = [0 : dims[0]-1];
  variable x = [0 : dims[1]-1];

  % dimensions of final image IMG
  nx = nint(nx);
  ny = nint(ny);
  variable DIMS = [(dims[0]-1)*ny+1, (dims[1]-1)*nx+1];
  variable Y = [0 : DIMS[0]-1], Yval = Y * 1./ny;
  variable X = [0 : DIMS[1]-1], Xval = X * 1./nx;

  % interpolation routine
  variable interp = qualifier("interp",
#ifexists gsl->interp_cspline
			      &gsl->interp_cspline
#else
			      &interpol
#endif
			     );
  variable ix, iy, iX, iY;
  variable Img, IMG = Double_Type[DIMS[0], DIMS[1]];

  if(qualifier_exists("y_first"))
  {
    % message("interpolate first in y (1st index)");
    Img = Double_Type[DIMS[0], dims[1]];
    foreach ix (x)
      Img[*,ix] = (@interp)(Yval, y, img[*,ix];; __qualifiers);
    foreach iY (Y)
      IMG[iY,*] = (@interp)(Xval, x, Img[iY,*];; __qualifiers);
  }
  else
  {
    % message("% interpolate first in x (2nd index)");
    Img = Double_Type[dims[0], DIMS[1]];
    foreach iy (y)
      Img[iy,*] = (@interp)(Xval, x, img[iy,*];; __qualifiers);
    foreach iX (X)
      IMG[*,iX] = (@interp)(Yval, y, Img[*,iX];; __qualifiers);
  }

  return IMG;
}

#endif

%
% Code from src/plot/epatplot.sl
%
%%%%%%%%%%%%%%%
define epatplot()
%%%%%%%%%%%%%%%
%!%+
%\function{epatplot}
%\usage{epatplot(Struct_Type events);}
%!%-
{
  variable e;
  switch(_NARGS)
  { case 1: e = (); }
  { help(_function_name()); return; }

  variable po = get_plot_options;

  variable N = 75;
  variable Elo, Ehi; (Elo, Ehi) = log_grid(0.2, 12, N);
  variable inv_dE = 1./(Ehi-Elo);
  variable tot = histogram(e.pi/1e3, Elo, Ehi) * inv_dE;
  tot[where(tot==0)] = 1;
  variable s = where(   e.pattern== 0);  if(length(s)>0)  s = histogram(e.pi[s]/1e3, Elo, Ehi) * inv_dE;  else s = Integer_Type[N];
  variable d = where(1<=e.pattern<= 4);  if(length(d)>0)  d = histogram(e.pi[d]/1e3, Elo, Ehi) * inv_dE;  else d = Integer_Type[N];
  variable t = where(5<=e.pattern<= 8);  if(length(t)>0)  t = histogram(e.pi[t]/1e3, Elo, Ehi) * inv_dE;  else t = Integer_Type[N];
  variable q = where(9<=e.pattern<=12);  if(length(q)>0)  d = histogram(e.pi[q]/1e3, Elo, Ehi) * inv_dE;  else q = Integer_Type[N];

  multiplot([2,1,2]);
  xlabel("Energy [keV]");
  xrange(0.15, 13);
  xlog;

  ylabel("counts / keV");
  $1 = [s, d, t, q];
  yrange(min_max($1[where($1>0)]));
  ylog;
  color(2);  hplot(Elo, Ehi, s);
  color(4); ohplot(Elo, Ehi, d);
  color(3); ohplot(Elo, Ehi, t);
  color(5); ohplot(Elo, Ehi, q);
  color(2); xylabel_in_box(0.95, 0.85, "s", 0, 1);
  color(4); xylabel_in_box(0.95, 0.7 , "d", 0, 1);
  color(3); xylabel_in_box(0.95, 0.55, "t", 0, 1);
  color(5); xylabel_in_box(0.95, 0.4 , "q", 0, 1);

  ylabel("counts / keV");
  $1 = [s, 4*d, 3*t, 4*q];
  yrange(min_max($1[where($1>0)]));
  color(2);  hplot(Elo  , Ehi  ,    s);
  color(4); ohplot(Elo/2, Ehi/2,  4*d);
  color(3); ohplot(Elo/3, Ehi/3,  9*t);
  color(5); ohplot(Elo/4, Ehi/4, 16*q);
  color(2); xylabel_in_box(0.95, 0.85, "1s @ E/1", 0, 1);
  color(4); xylabel_in_box(0.95, 0.65, "2d @ E/2", 0, 1);
  color(3); xylabel_in_box(0.95, 0.45, "3t @ E/3", 0, 1);
  color(5); xylabel_in_box(0.95, 0.25, "4q @ E/4", 0, 1);

  ylabel("fraction");
  ylin;
  yrange(-0.05, 1.05);
  color(6);  hplot(Elo, Ehi, (s+d)/tot);
  color(2); ohplot(Elo, Ehi,  s   /tot);
  color(4); ohplot(Elo, Ehi,    d /tot);
  color(3); ohplot(Elo, Ehi,   t  /tot);
  color(5); ohplot(Elo, Ehi,   q  /tot);
  color(6); xylabel_in_box(0.95, 0.85, "s+d", 0, 1);
  color(2); xylabel_in_box(0.95, 0.7 , "s", 0, 1);
  color(4); xylabel_in_box(0.95, 0.55, "d", 0, 1);
  color(3); xylabel_in_box(0.95, 0.4 , "t", 0, 1);
  color(5); xylabel_in_box(0.95, 0.25, "q", 0, 1);

  set_plot_options(po);
}

%
% Code from src/data/timing/epferror.sl
%
% used as default for the 'pget' qualifier: determine the maximum in
% the chi^2 landscape and return the corrsponding period
private define _epferror_pget(p, stat) {
  variable imax = where_max(stat);
  if (length(imax) != 1) { return -1; }
  return p[imax[0]];
}

%%%%%%%%%%%%%%%%%%%%%%%%
define epferror ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{epferror}
%\synopsis{Estimate epoch folding uncertainty with Monte-Carlo simulation
%    approach.}
%\usage{(Double_Type err) = epferror(Double_Type time, rate, period[, rate_error]);}
%\qualifiers{
%\qualifier{pstart}{start period for epoch folding period search.
%      (default: 0.5*period)}
%\qualifier{pstop}{stop period for epoch folding period search.
%      (default: 1.5*period)}
%\qualifier{ntrial}{number of MC iterations. (default: 20)}
%\qualifier{pget}{function reference to determine the period from the
%            chi^2 landscape. Arguments passed are
%              Double_Type[] p, stat
%            as returned by epfold. Has to return the period and -1
%            if the period could not be determined. (default: find
%            maximum, see below)}
%\qualifier{pdist}{set to a variable reference in order to retrieve
%             the simulated period distribution.}
%\qualifier{chatty}{chattiness of this function: If >0 print some debug
%              messages. If >1 plot the chi^2 landscape at each MC
%              iteration. (default: 0)}
%\qualifier{fchatty}{chattiness piped to epfold.}
%}
%\description
%    Note: All qualifiers are passed to pfold (for pulse profile
%    calculation) and epfold (for period search)! 
% 
%    This routine tries to estimate the uncertainty of a previously
%    received period using the epoche folding approach (see epfold).
%    It is adopted from the IDL script with the same name, but does
%    not yet allow for GTI or Poisson statistics. It implements the
%    following strategy:
%    1.) calculate a mean profile with given period. 
%    2.) compute the intensity for all times applying the
%        period multiplied profile.
%    3.) simulate an uncertainty for all times (assuming normal
%        distribution with sigma = error, or, if not given
%        sqrt(rate)).
%    4.) perform epoch folding for that simulated lightcurve.
%    5.) determine the maximum of the epoch folding chi^2 landscape
%        and remember the corresponding period. Use the 'pget'
%        qualifier in order to implement a user-defined approach!
%    6.) go to step 2.) Ntrial times, which results in a distribution
%        of determined periods.
%    7.) compute the standard deviation of the period distribution
%        obtained and return this as the uncertainty of the intial
%        epoch folding
% 
%    NOTE: The processing may last a long time (be prepared to wait  
%          hours to weeks)!
%
%    NOTE: Is is _important_ to use the same qualifiers used for
%          epfold previously for the actual data! That is epferror
%          has to search for the period in the same way as for the
%          data! This might further require to provide a function
%          for the determination of the best period from an epoch
%          folding result (see 'pget' qualifier).
%
%    References:
%       Davies, S.R., 1990, MNRAS 244, 93
%       Larsson, S., 1996, A&AS 117, 197
%       Leahy, D.A., Darbro, W., Elsner, R.F., et al., 1983,
%          ApJ 266, 160-170
%       Schwarzenberg-Czerny, A., 1989, MNRAS 241, 153
%
%!%-
{
  variable intime, inrate, period, rateerr = Double_Type[0] ;
  
  switch(_NARGS)
     {case 3: (intime,inrate,period) = ();}
     {case 4: (intime, inrate, period, rateerr) = (); }
     {help(_function_name()); return; }

 
   % qualifiers for epfold
   variable pstop = qualifier("pstop", 1.5*period) ;  
   variable pstart = qualifier("pstart", 0.5*period) ;
   variable pget = qualifier("pget", &_epferror_pget) ;
   if (pget != NULL && typeof(pget) != Ref_Type) {
     vmessage("error (%s): pget has to be a function reference!",
              _function_name);
   }

   %runs of the Monte Carlo loop
   variable ntrial = qualifier("ntrial", 20) ;
   
   %chattiness of the script: chatty != print progress
   %chatty > 1 : overplot chi^2 distribution
   variable chatty = qualifier("chatty", 0) ;
   variable fchatty = qualifier("fchatty", 0) ;
 
   if (chatty) {variable begin = _time ; message("Started MC loop..."); } 
  
   if (length(rateerr) == 0) {rateerr= sqrt(inrate);} 
  
   variable profile = pfold(intime, inrate, period ;; __qualifiers  ) ;

   variable simbaserate = interpol_points((intime mod period)/period, (profile.bin_lo+profile.bin_hi)/2., profile.value) ;

    variable simrate, epf, maxchierg = Double_Type[ntrial], i;

  variable bad = Integer_Type[ntrial] ;
  
  _for i (0, ntrial-1, 1) {
   
   % seed_random(qualifier("seed", _time)) ;

   % print(qualifier("seed", _time)) ;
    
   simrate = rateerr* grand(length(simbaserate)) + simbaserate ;

       % print(intime) ;
    
    % epf = epfold(intime, simrate, pstart, pstop; @__qualifiers(), chatty=fchatty ) ;  % requires S-Lang >= pre-2.2.3-39
    epf = epfold(intime, simrate, pstart, pstop;; struct_combine(__qualifiers(), struct { chatty=fchatty }));

    variable psim = @pget(epf.p, epf.stat);
    if (psim > -1) {
      maxchierg[i] = psim;
    } else {
      message("no unique maxium found, ignoring!");
      bad[i] = 1;
    }

    if (chatty) {
     ()=printf("Processed: %.2f%%  ---  Period %f\n", (100.*i)/ntrial, maxchierg[i]) ;
   }
  
   if (chatty > 1){ 
     color(2) ; oplot(epf.p, epf.stat) ;
   }
    
  }
  
  variable  err = moment(maxchierg[wherenot(bad)]).sdev ;
  
  if (chatty){
    ()=printf("Total runtime was %.2f sec.\n", _time-begin) ;
    ()=printf("%d failed period determinations.\n", length(where(bad)));
    ()=printf("Estimated uncertainty is %f.\n", err);
  }
  % return(struct{t = intime, r = simrate, pstart=pstart,p stop=pstop}) ;

  if (qualifier_exists("pdist")) {
    variable pdist = qualifier("pdist");
    if (typeof(pdist) != Ref_Type) {
      vmessage("error (%s): pdist has to be a variable reference!",
               _function_name);
      @pdist = NULL;
    } else {
      @pdist = maxchierg;
    }
  }
 return err ;
}

%
% Code from src/data/timing/epfold.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define epfold ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{epfold}
%\synopsis{peforms epoch folding on a lightcurve in a given period
%interval}
%\usage{(Struct_Type res) = epfold(Double_Type t, r, pstart, pstop);
%or (Struct_Type res) = epfold(Double_Type t, pstart, pstop) ; (event data)}
%
%\qualifiers{
%\qualifier{nbins}{number of bins for the pulse profile (default=20)}
%\qualifier{exact}{calculate the pulse profile in a more exact way, see description of pfold (not recommended as it takes a very long time!).}
%\qualifier{dt}{exposure of every lightcurve time bin, should be given to ensure correct results.}
%\qualifier{sampling}{how many periods per peak to use (default=10)}
%\qualifier{nsrch}{how many periods to search in a linear grid (default not set)}
%\qualifier{dp}{delta period of linear period grid (default  not set)}
%\qualifier{lstat}{use L-statistics instead of chi^2 statistics}
%\qualifier{chatty}{set the verbosity, chatty-1 is piped to pfold (default=0)}
%\qualifier{gti}{GTIs for event data, given as struct{start=Double_Type, stop=Double_Type}} 
%}
%
%
%\description
%   Performs epoch folding on a given lightcurve between the periods
%   pstart and pstop. The function was adopted from
%   the IDL routine of the same name. GTI correction only implemented
%   for event-data yet.
%
%   By default, periods are sampled according to the triangular rule
%   for estimating the period error, using "sampling" periods per peak.
%   If a linear grid is to be used, either "dp" for a given distance
%   between two consecutive periods or "nsrch" for a given number of
%   periods can be given. These qualifiers are mutually exclusive. 
%
%   The returned structure "res" contains four fields: "p" for the
%   evaluated period and "stat" for the value of the statistic used.
%   Additionally the field "nbins" contains the number of phase bins
%   used and "badp" contains indices for res.p marking periods
%   where the pulse profile showd empty phase bins. Values of
%   res.stat[res.badp] should be taken with great care!
%        
%   Compared to the similar function sitar_epfold_rate, epfold.sl
%   uses the chi^2 statistic as a default and is based on pfold.sl,
%   which can take errors of the lightcurve into account.
%   If the qualifier "lstat" is given, the statistic is switched to
%   the l-stat statistic as in sitar_epfold_rate, but errors of the
%   lightcurve are no longer taken into account. lstat is not
%   available for event-data.
%
%   If the "exact" qualifier is given, the function takes the exposure
%   time of every time bin into account in the sense that, that a
%   given time bin may overlap over two phase bins. The corresponding
%   exposure time in every phase bin is reduced accordingly.
%
%   NOTE: the "exact" qualifiers slows the script considerably down,
%   depending on the length of the lightcurve and the number of bins
%   for up to a factor of >100!
%   
%   If "exact" is not given, the script is ~10% slower than
%   sitar_epfold_rate.
%   
%   The script is still in the develepment phase, please report any
%   bugs or missing features to Felix
%   (felix.fuerst@sternwarte.uni-erlangen.de).
%   
%   NOTE: Please read  
%          Davies, S.R., 1990, MNRAS 244, 93 (L-statistics!)
%          Larsson, S., 1996, A&AS 117, 197
%          Leahy, D.A., Darbro, W., Elsner, R.F., et al.,1983, ApJ 266, 160-170
%          Schwarzenberg-Czerny, A., 1989, MNRAS 241, 153
%   
%!%-
{
     variable t,r,pstart, pstop ;

  variable eventdata = 0;
  % disable chatty if not set
   variable chatty = qualifier("chatty",0) ;
  
    switch(_NARGS)
    {case 3: (t,  pstart, pstop) = ();
      eventdata = 1 ;
      if (chatty>0) {print("using event data"); } 
    }
    {case 4: (t,r,pstart, pstop) = (); }
    {help(_function_name()); return; }
  
   variable t0 = qualifier ("t0", t[0]);
   variable nbins = qualifier ("nbins", 20) ;
   variable sampl = qualifier("sampling", 10)  ; %periods per peak
   variable dp = qualifier("dp", -10) ; %delta p for linear period sampling
   variable nsrch = qualifier("nsrch", -10) ; %number of periods in linear period grid

   nsrch = int(nsrch) ;

  ifnot(qualifier_exists("dt") or eventdata==1){
    message("WARNING (epfold): no dt qualifier given, creating differential grid! Might lead to unexpected results!") ;
  }
   variable dt = qualifier("dt" , make_hi_grid(t)-t) ; %exposure times

  ifnot (qualifier_exists("gti") or eventdata==0){
    message("WARNING (epfold): using eventdata without gti qualifier, taking time of first and last event. This is most likely NOT what you want!") ;
  }
  variable gti = qualifier("gti", struct{start=min(t), stop=max(t)}) ;

   variable tges = sum(dt) ; %effective observation time
  
   if (tges < 0) {message("ERROR (epfold): Please provide lightcurve with continously increasing time (tges was < 0)."); return;}
  
   if (chatty>0) {message(">> Starting data analysis <<") ;}
       
   if (chatty>0) {
      ()=printf("Determining number of trial periods between %f and %f.\n", pstart, pstop) ;
   }

  if (eventdata and qualifier_exists("lstat")) {
    message("Warning (epfold): using eventdata, lstat qualifier not sensible! Reverting to use of chi^2 statistic.") ;
  }

   variable ergdim=1 ;
   variable p;
   variable i = 1 ;
   
   if (dp > 0 && nsrch == -10)
     {	
	if (chatty>0) {()=printf("Making linear period grid with delta P = %f.\n", dp) ;}
	p = [pstart:pstop:dp] ;
	ergdim = length(p) ;
     }
   else if (dp == -10  && nsrch > 0)
     {
	if (chatty>0) {()=printf("Making linear period grid with %d entries.\n", nsrch) ;}
	p = [pstart:pstop:#nsrch] ;
	ergdim = nsrch ;
     }
   else if (dp > 0 && nsrch > 0)
     {
        ()=printf("Please use either dp or nsrch!\n") ;
	return NULL ;
     }
   else { 
      if (chatty>0) ()=printf("Making intelligent grid!\n") ;
      variable p_bin_hi = pstart;
      variable jj = 1;
      while (p_bin_hi  <= pstop){ % This may look pedestrian, but I determine first how much periods the grid will contain, because doing the period calculation twice will still be WAY faster than copying the arrays
        p_bin_hi = p_bin_hi + sqr(p_bin_hi)/(tges*sampl);
	jj++;
      };
      p  = Double_Type[jj]; % initialize array long enough for all test periods
      p[0] = pstart;
      while (p[i-1] <= pstop) 
	{  ergdim++ ;
	   p[i]= p[i-1]+p[i-1]*p[i-1]/(tges*sampl);
	   i++ ;
   	}
     p = p[[0:i-1]]; % cut the empty array entries
   }
 
   if (chatty>0) {()=printf("number of periods = %d\n....done\n",ergdim);}
   
   if (chatty>0) {variable tstart = _time(); variable tlast = _time() ; }
   
   variable ptest = pstart ;
   variable pp, ndx  ;
   
   variable parr = Double_Type[ergdim] ;
   variable chierg = Double_Type[ergdim] ;
   
   variable perc, titer, remain ;

   variable factor  =  (1.* (length(t)-nbins)) / (nbins -1.) ; % (N-M)/(M-1) as in Davies Eq. 11

   variable finitepp, faulty = Integer_Type[0] ;

 
   %subtract mean from rate, as we need only the
   %difference for the statistics and it does not change
   %the generality
  ifnot(eventdata){
  variable rm = moment(r) ;
   r -= rm.ave ;
      
   if (chatty>0)  {
      ()=printf("mean = %f \n", rm.ave) ;
      ()=printf("effective observation time: %f\n", tges  );
   }
  }
  
   _for i (0, length(parr)-1, 1)
     {
	ptest = p[i] ;
        if(eventdata) {pp = pfold(t,ptest ; nbins=nbins,chatty=chatty-1, t0=t0,gti=gti);}
	else if(qualifier_exists("exact")) { pp = pfold(t,r,ptest ; nbins = nbins, dt = dt, chatty=chatty-1, t0=t0, exact) ; }
	else { pp = pfold(t,r,ptest  ; nbins = nbins, dt = dt, chatty=chatty-1, t0=t0 ) ; }

	finitepp = wherenot(isnan(pp.value) or isinf(pp.value));
	
	if (length(finitepp) < nbins) {
	   ()=printf("Warning: period %f has faulty phase bins! Ignoring those phase bins, but please check!\n", ptest);
	   faulty = [faulty, i] ;
	   struct_filter(pp, finitepp) ;
	} ;

	    if (eventdata) {
		chierg[i] =  sum((pp.value[*]-mean(pp.value))^2. / (mean(pp.value)/pp.ttot[*])) ;
	    } else {
		if (qualifier_exists("lstat")) 
		{
		    % see Davies Eq. 11 
		    chierg[i] = factor * ( sum(pp.npts*pp.value^2))/(sum (pp.error*(pp.npts-1.))) ;
		}
		else 
		{
		    % see Davies Eq. 4 or Larsson Eq. 1
		    chierg[i] =  sum(pp.npts*pp.value^2.)/rm.var ;
		}
	    }
	
     if(chatty>0) 
	  {
	     if (_time()-tlast > 10)
	       {
		  tlast=_time() ;
		  titer = (tlast-tstart)/double(i) ;
		  remain=(ergdim-i)*titer ;
		  
		  ()=printf("%.2f%% done\n", i*100./ergdim ) ;
		  ()=printf("its taking %.2fsec per iteration\n", titer) ;
		  
		  if (remain <= 600)  ()=printf("%.2f sec remaining...\n", remain) ;
		  else if (remain <= 7200) ()=printf("%.2f min remaining...\n", remain/60.) ;
		  else ()=printf("%.2f h remaining...\n", remain/3600.) ;
	       }
	  }
     }

  %structure containing searched periods, statistics value,
  %phasebins used and indices of periods with gaps in the pulse profile
   return struct{p=p, stat=chierg, nbins=nbins, badp = faulty } ;
   
}

%
% Code from src/data/timing/epfoldpdot.sl
%

%%%%%%%%%%%%%%%%%%%%%%%%
define epfoldpdot ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{epfoldpdot}
%\synopsis{peforms epoch folding on a lightcurve in a given period
%interval}
%\usage{(Struct_Type res) = epfoldpdot(Double_Type t, r, pstart, pstop);
%or (Struct_Type res) = epfoldpdot(Double_Type t, pstart, pstop) ; (event data)}
%
%\qualifiers{
%\qualifier{nbins}{number of bins for the pulse profile}
%\qualifier{exact}{calculate the pulse profile in a more exact way, see description of pfold (not recommed as it takes a very long time!).}
%\qualifier{dt}{exposure of every lightcurve time bin, should be given to ensure correct results.}
%\qualifier{sampling}{how many periods per peak to use (default=10)}
%\qualifier{nsrch}{how many periods to search in a linear grid (default not set)}
%\qualifier{dp}{delta period of linear period grid (default  not set)}
%\qualifier{lstat}{use L-statistics instead of chi^2 statistics}
%\qualifier{chatty}{set the verbosity, chatty-1 is piped to pfold (default=0)}
%\qualifier{gti}{GTIs for event data, given as struct{start=Double_Type, stop=Double_Type}}
%\qualifier{pdstart}{start p-dot for grid search (default=0)}
%\qualifier{pdstop}{stop p-dot for grid search (default=0)}
%\qualifier{pdnsrch}{search point for p-dot grid, (default = nsrch
%(if defined, otherwise 10))}
%}
%
%
%\description
%   Performs epoch folding on a given lightcurve between the periods
%   pstart and pstop and period derivatives (p-dot) pdstart and pdstop.
%   The function was adopted from
%   the IDL routine of the same name. GTI correction only implemented
%   for event-data yet.
%
%   By default, periods are sampled according to the triangular rule
%   for estimating the period error, using "sampling" periods per peak.
%   If a linear grid is to be used, either "dp" for a given distance
%   between two consecutive periods or "nsrch" for a given number of
%   periods can be given. These qualifiers are mutually exclusive.
%
%   P-dot is always sampled on a linear grid with  pdnsrch points.
%
%   The routine uses the S-lang function 'parallel_map' to use
%   Isis_Slaves.num_slaves cores on your machine for parallel
%   computing. On a MacBook Pro with quadcorse i7 this gives a speed
%   improvement of a factor ~>2 over single core calculations. 
%
%   The returned structure "res" contains five fields: "p" for the
%   evaluated period, pd for the evaluated period derviates,
%   and "stat" for the value of the statistic used. "stat" is a 2D
%   array of dimnesion (np, npd). 
%   Additionally the field "nbins" contains the number of phase bins
%   used and "badp" contains indices for res.p marking periods
%   where the pulse profile showd empty phase bins. Values of
%   res.stat[res.badp] should be taken with great care!
%        
%   Compared to the similar function sitar_epfold_rate, epfold.sl
%   uses the chi^2 statistic as a default and is based on pfold.sl,
%   which can take errors of the lightcurve into account.
%   If the qualifier "lstat" is given, the statistic is switched to
%   the l-stat statistic as in sitar_epfold_rate, but errors of the
%   lightcurve are no longer taken into account. lstat is not
%   available for event-data.
%
%   If the "exact" qualifier is given, the function takes the exposure
%   time of every time bin into account in the sense that, that a
%   given time bin may overlap over two phase bins. The corresponding
%   exposure time in every phase bin is reduced accordingly.
%
%   NOTE: the "exact" qualifiers slows the script considerably down,
%   depending on the length of the lightcurve and the number of bins
%   for up to a factor of >100!
%
%   NOTE: "exact" qualifier is currenlty untested and may lead to
%   erronoeus results!
%   
%   If "exact" is not given, the script is ~10% slower than
%   sitar_epfold_rate.
%   
%   The script is still in the develepment phase, please report any
%   bugs or missing features to Felix
%   (felix.fuerst@fau.de).
%   
%   NOTE: Please read  
%          Davies, S.R., 1990, MNRAS 244, 93 (L-statistics!)
%          Larsson, S., 1996, A&AS 117, 197
%          Leahy, D.A., Darbro, W., Elsner, R.F., et al.,1983, ApJ 266, 160-170
%          Schwarzenberg-Czerny, A., 1989, MNRAS 241, 153
%   
%!%-
{
     variable t,r,pstart, pstop ;

  variable eventdata = 0;
  % disable chatty if not set
   variable chatty = qualifier("chatty",0) ;
  
    switch(_NARGS)
    {case 3: (t,  pstart, pstop) = ();
      eventdata = 1 ;
      if (chatty) {print("using event data"); } 
    }
    {case 4: (t,r,pstart, pstop) = (); }
    {help(_function_name()); return; }
  
   variable t0 = qualifier ("t0", t[0]);
   variable nbins = qualifier ("nbins", 20) ;
   variable sampl = qualifier("sampling", 10)  ; %periods per peak
   variable dp = qualifier("dp", -10) ; %delta p for linear period sampling
   variable nsrch = qualifier("nsrch", -10) ; %number of periods in linear period grid
  
   nsrch = int(nsrch) ;

   variable pdstart=0;
   variable pdstop ;
   
   variable pd=0;
   
   variable pdnsrch=1 ;
   
   if (qualifier_exists("pdstart") and qualifier_exists("pdstop"))
     {
	pdnsrch= qualifier("pdnsrch", qualifier("nsrch", 10)) ;
	pdstart = qualifier("pdstart") ;
	pdstop = qualifier("pdstop") ;
	pd = [pdstart:pdstop:#pdnsrch];
     }

  ifnot(qualifier_exists("dt") or eventdata==1){
    message("WARNING (epfoldpdot): no dt qualifier given, creating differential grid! Might lead to unexpected results!") ;
  }
   variable dt = qualifier("dt" , make_hi_grid(t)-t) ; %exposure times

  ifnot (qualifier_exists("gti") or eventdata==0){
    message("WARNING (epfoldpdot): using eventdata without gti qualifier, taking time of first and last event. This is most likely NOT what you want!") ;
  }
  variable gti = qualifier("gti", struct{start=min(t), stop=max(t)}) ;

   variable tges = sum(dt) ; %effective observation time
  
   if (tges < 0) {message("ERROR (epfoldpdot): Please provide lightcurve with continously increasing time (tges was < 0)."); return;}
  
   if (chatty) {message(">> Starting data analysis <<") ;}
       
   if (chatty) {
      ()=printf("Determining number of trial periods between %f and %f.\n", pstart, pstop) ;
   }

  if (eventdata and qualifier_exists("lstat")) {
    message("Warning (epfoldpdot): using eventdata, lstat qualifier not sensible! Reverting to use of chi^2 statistic.") ;
  }

   variable ergdim=1 ;
   variable p = pstart ;
   variable j, i = 1 ;
   
   if (dp > 0 && nsrch == -10)
     {	
	if (chatty) {()=printf("Making linear period grid with delta P = %f.\n", dp) ;}
	p = [pstart:pstop:dp] ;
	ergdim = length(p) ;
     }
   else if (dp == -10  && nsrch > 0)
     {
	if (chatty) {()=printf("Making linear period grid with %dx%d entries.\n", nsrch, pdnsrch) ;}
	p = [pstart:pstop:#nsrch] ;
	ergdim = nsrch ;
     }
   else if (dp > 0 && nsrch > 0)
     {
	()=printf("Please use either dp or nsrch!\n") ;
	return NULL ;
     }
   else { 
      if (chatty) ()=printf("Making intelligent grid!\n") ; 
      while (p[-1] <= pstop) 
	{
	   ergdim++ ;
	   p=[p, p[i-1]+p[i-1]*p[i-1]/(tges*sampl) ];
	   i++ ;
   	}
   }
 
   if (chatty) {()=printf("number of periods = %d\n....done\n",ergdim*pdnsrch);}
   
   if (chatty) {variable tstart = _time(); variable tlast = _time() ; }
   
   variable ptest = pstart ;
   variable pdtest = pdstart ;
   variable pp,pparr, ndx  ;
   
   % variable parr = Double_Type[ergdim] ;
   variable chierg = Double_Type[ergdim,pdnsrch] ;
   
   variable perc, titer, remain ;

   variable factor  =  (1.* (length(t)-nbins)) / (nbins -1.) ; % (N-M)/(M-1) as in Davies Eq. 11

   variable finitepp, faulty = Integer_Type[0] ;

 
   %susbstract mean from rate, as we need only the
   %difference for the statistics and it does not change
   %the generality
  ifnot(eventdata){
  variable rm = moment(r) ;
   r -= rm.ave ;
      
   if (chatty)  {
      ()=printf("mean = %f \n", rm.ave) ;
      ()=printf("effective observation time: %f\n", tges  );
   }
  }
   _for j (0, pdnsrch-1, 1)
     {
	pdtest = pd[j] ;
	
	if (chatty==2) ()=printf("Entry %d,%d: P=%.4f, Pdot=%.4e\n", i,j, ptest, pdtest) ;
	     
	if(eventdata) {
	   pparr = parallel_map(Struct_Type, &pfold, p ;
					t=t, pdot=pdtest, nbins=nbins,chatty=chatty-1, t0=t0,gti=gti, noqualmap);
	}

	else if(qualifier_exists("exact")) {
	   pp = parallel_map(Struct_Type, &pfold, p ;
				     t=t, p=p, nbins = nbins, dt = dt, chatty=chatty-1, t0=t0, exact, noqualmap) ;
	}
	else {
	   pparr = parallel_map(Struct_Type, &pfold, p  ;
					t=t,r=r, pdot=pdtest, nbins = nbins, dt = dt, chatty=chatty-1, t0=t0,
					       noqualmap) ;
	}

	_for i (0, ergdim-1, 1)
	  {
	     pp = pparr[i] ;
	     ptest = p[i] ;

	     finitepp = wherenot(isnan(pp.value) or isinf(pp.value));
	
	     if (length(finitepp) < nbins) {
		()=printf("Warning: period %f has faulty phase bins! Ignoring those phase bins, but please check!\n", ptest);
		faulty = [faulty, i] ;
		struct_filter(pp, finitepp) ;
	     } ;

	     if (eventdata) chierg[i,j] =  sum((pp.value-mean(pp.value))^2. / mean(pp.value)) ;
	     else if (qualifier_exists("lstat")) 
	       {
		  % see Davies Eq. 11 
		  chierg[i,j] = factor * ( sum(pp.npts*pp.value^2))/(sum (pp.error*(pp.npts-1.))) ;
	       }
	     else 
	       {
		  % see Davies Eq. 4 or Larsson Eq. 1
		  chierg[i,j] =  sum(pp.npts*pp.value^2.)/rm.var ;
	       }
	
	     if(chatty) 
	       {
		  if (_time()-tlast > 10)
		    {
		       tlast=_time() ;
		       titer = (tlast-tstart)/double(j) ;
		       remain=(ergdim-j)*titer ;
		       
		       ()=printf("%.2f%% done\n", j*100./ergdim ) ;
		       ()=printf("its taking %.2fsec per iteration\n", titer) ;
		       
		       if (remain <= 600)  ()=printf("%.2f sec remaining...\n", remain) ;
		       else if (remain <= 7200) ()=printf("%.2f min remaining...\n", remain/60.) ;
		       else ()=printf("%.2f h remaining...\n", remain/3600.) ;
		    }
	       }
	  }
	
     }
   

  %structure containing searched periods, statistics value,
  %phasebins used and indices of periods with gaps in the pulse profile
   return struct{p=p,pd=pd, stat=chierg, nbins=nbins, badp = faulty } ;
   
}

%
% Code from src/slang/date/equation_equinoxes.sl
%
define equation_equinoxes() {
%!%+
%\function{equation_equinoxes}
%\usage { egam = equation_equinoxes(JD;qualifiers)}
%\qualifiers{
%   \qualifier{mjd}{the time argument is in MJD, not in JD}
%   \qualifier{deg}{return equation of equinoxes in degrees}
%   \qualifier{arcsec}{return equation of equinoxes in arcseconds}
%   \qualifier{mas}{return equation of equinoxes in milliarcseconds}
%   \qualifier{seconds}{return equation of equinoxes in seconds}
%}
%
%\description
% This function calculates the equation of equinoxes, i.e., the 
% difference between the Greenwich Mean Siderial Time and the
% Greenwich Apparent Siderial Time caused by the motion of the
% equinox due to nutation. The implementation uses a truncated
% form of a longer series due to the IERS (2003), given by
% Eq. 2.14 of Kaplan (2009, USNO circular 181). The default
% returns the equation of equinoxes in radian (more commonly
% the values are tabulated in units of seconds).
%
% This function is array safe.
%
%\seealso{gmst,nutation_angles}
%!%-
variable JD;
    switch(_NARGS)
    { case 1 : JD=(); }
    { help(_function_name()); return;}

    variable T;
    if (qualifier_exists("mjd")) {
	T=((JD-51544.5)+qualifier("JD_frac",0.))/36525.0;
    } else {
	T=((JD-2451545.)+qualifier("JD_frac",0.))/36525.0;
    }

    % conversion parameters
    variable asec360=3600.*360.;
    variable asec2rad=PI/648000.;

    % Fundamental arguments from the Simon et al. (1994) lunar theory
    % see nutation_angles.sl
    
    
    variable Om, F, D;
    
    % Mean argument of the latitude of the Moon
    F = ((335779.526232 +
               T * (1739527262.8478 +
               T * (      - 12.7512 +
               T * (      -  0.001037 +
               T * (         0.00000417))))) mod asec360) * asec2rad;

    % Mean elongation of the moon from the Sun
    D = ((1072260.70369 +
               T * (1602961601.2090 +
               T * (       - 6.3706 +
               T * (         0.006593 +
               T * (       - 0.00003169))))) mod asec360) * asec2rad;

    % Mean longitude of the Moon's ascending node
    Om= ((450160.398036 +
               T * ( - 6962890.5431 +
               T * (         7.4722 +
               T * (         0.007702 +
               T * (       - 0.00005939))))) mod asec360) * asec2rad;

    variable eps,dpsi,deps;
    if (qualifier_exists("mjd")) {
        eps=obliquity(JD;mjd); % in rad
        (dpsi,deps)=nutation_angles(JD;mjd,arcsec); % in arcsec!
    } else {
        eps=obliquity(JD); % in rad
        (dpsi,deps)=nutation_angles(JD;arcsec); % in arcsec!
    }
        
    variable egam=dpsi*cos(eps)
       + 0.00264096 * sin(Om)
       + 0.00006352 * sin(2.*Om)
       + 0.00001175 * sin(2.*F-2.*D+3.*Om)
       + 0.00001121 * sin(2.*F-2.*D+Om)
       - 0.00000455 * sin(2.*F-2.*D+2.*Om)
       + 0.00000202 * sin(2.*F+3.*Om)
       + 0.00000198 * sin(2.*F+Om)
       - 0.00000172 * sin(3.*Om)
       - 0.00000087 * T * sin(Om);

    if (qualifier_exists("arcsec")) {
        return egam;
    } 
    if (qualifier_exists("deg")) {
        return egam/3600.;
    }
    if (qualifier_exists("mas")) {
        return egam*1000.;
    }
    if (qualifier_exists("seconds")) {
        return egam/15.;
    }
    return egam*asec2rad;
}

%
% Code from src/fitting/fit-functions/eqwFit.sl
%
%%%%%%%%%%%%%%%%%%%%%
define eqwFit_fit(lo, hi, pars)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{eqwFit}
%\synopsis{fit function for replacing a feature's norm by its equivalent width}
%\usage{eqwFit(id, continuum, feature1, ..., featureN)}
%\description
%    The so-called equivalent width of a feature in a spectrum
%    is a measurement for its flux F_{feature} compared to the
%    underlying continuum F_{continuum}. It is defined as the
%    width 'eqw' of a rectangle centered at the features
%    maximum or minimum 'E_0' and with a height equal to
%    the continuum at E_0:
%
%      $\\int F_{feature}(E) dE = F_{continuum}(E_0) * eqw$
%
%    As a result, the equivalent width stays constant if the
%    flux of continuum and the feature are correlated.
%
%    The equivalent width of a feature replaces its norm. Due
%    to that, the norm has to be at a fixed value, e.g, 1.
%    The fit function defined here then scales the feature
%    to the equivalent width, given as a fit parameter.
%    The function provies several fit parameters:
%      E_min  - lower energy boundary for the feature's flux
%      E_max  - higher energy boundary
%      widthN - the equivalent width of the Nth feature
%      multiN - if 1, the feature is considered to be
%               multiplied with the continuum, otherwise
%               its additive
%    The returned value of the fit function is the continuum
%    with all the given features applied!
%
%    *IMPORTANT*
%    The number of features the fit function can handle has
%    to be set beforehand via 'eqwFit_init'. Once set, the
%    number CAN NOT be changed afterwards. The number of
%    given features has to be fulfilled EVERY TIME the fit
%    function is used! If the function should be evaluated
%    for less features, the remaining features have to be
%    set to ZERO!
%
%    *NOTE*
%    Some multiplicative models do not provide a proper
%    normalization, such that the norm N is defined like
%
%      model = N * ...
%
%    For example, although the gaussian absorption 'gabs'
%    has a depth 'tau', due to its definition
%
%      gabs = exp(-tau * exp(...))
%
%    it CAN NOT be used with eqwFit properly! Instead,
%    use additive models to mimic a multiplicative one:
%
%      mygabs = 1 - egauss
%\example
%    % init the fit function to handle two features
%    eqwFit_init(2);
%
%    % equivalent width of a single gaussian,
%    % the resulting model is equal to:
%    % powerlaw(1) + egauss(1)
%    fit_fun("eqwFit(1, powerlaw(1), egauss(1), 0)");%
%    set_par("eqwFit(1).width1", 300); % eqw = 300eV
%    set_par("eqwFit(1).multi1", 0); % additive
%    set_par("egauss(1).area", 1, 1); % freeze area
%
%    % equivalent width of an additive and
%    % multiplicative gaussian,
%    % the resulting model is equal to:
%    % powerlaw(1)*(1 - egauss(2)) + egauss(1)
%    fit_fun("eqwFit(1, powerlaw(1), egauss(1), 1 - egauss(2))");
%    set_par("eqwFit(1).width2", 200); % eqw = 200eV
%    set_par("eqwFit(1).multi2", 1); % multiplicative
%    set_par("egauss(2).area", 1, 1); % freeze area
%\seealso{eqwFit_init, eqw}
%!%-
{
  % get all additional parameters (continuum, feature1, feature2, ...)
  if (_stkdepth < 2) { vmessage("error (%s): need a continuum model and at least one feature model", _function_name); exit; }
  variable feature = Array_Type[(length(pars)-2)/2];
  variable i;
  _for i (length(feature)-1, 0, -1) {
    variable newf = ();
    if (typeof(newf) == Array_Type && length(newf) == length(lo)) feature[i] = newf;
  }
  variable continuum = ();
  variable n = length(feature);
  % useful names for parameters
  variable emin = pars[0], emax = pars[1], eqw = pars[[2:1+n]], mult = pars[[2+n:length(pars)-1]], f;
  % select energy range for eqw calculations
  variable iw = (emin == 0 && emax == 0 ? [0:length(lo)-1] : where(lo >= _A(emax) and hi <= _A(emin)));
  % flux density of continuum
  variable fdcon = reverse(reverse(continuum) / (_A(lo) - _A(hi)));
  % loop over all features
  variable resflux = @continuum;
  _for f (0, n-1, 1) {
    % proceed if feature is non-zero
    if (feature[f] != NULL) {
      % flux and flux density within each feature
      variable featflux = (mult[f] == 0 ? feature[f] : continuum*(feature[f] - 1));
      variable fdfea = reverse(reverse(featflux) / (_A(lo) - _A(hi)));
      % get energy index where each feature has its maximum in flux density  
      variable iwe = min(where_max(abs(fdfea[iw])));
      % flux density of continuum at feature's maximum
      variable fdconmax = fdcon[iw[iwe]];
      % calculate scaling factor of feature, such that given eqw fits
      variable fluxfea = sum(featflux[iw]); % total flux in feature (should be a fixed value!)
      variable c = eqw[f] * fdconmax / fluxfea / 1000; % factor 1000 from keV to eV
      % add feature's flux to the resulting flux
      ifnot (isnan(c) || isinf(c)) resflux += c*featflux;
    }
  }
  return resflux;
}

%%%%%%%%%%%%%%%%%%%%%
define eqwFit_init() {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{eqwFit_init}
%\synopsis{initializes the fit function 'eqwFit'}
%\usage{eqwFit_init(Integer_Type N);}
%\description
%    Defines and initializes the equivalent width fit
%    function 'eqwFit' to handle 'N' number of features.
%    Note, that the number has to be set before one can
%    use the fit function and that it can only be set once!
%\seealso{eqwFit, eqw}
%!%-
  if (is_defined("eqwFit")) { vmessage("error(%s): fit function has been initialized already", _function_name); return; }
  variable numfeat;
  switch (_NARGS)
    { case 0: numfeat = 1; }
    { case 1: numfeat = (); }
    { help(_function_name); return; }
  % width and multiplicative parameter names
  variable widthpars = array_map(String_Type, &sprintf, "width%d [eV]", [1:numfeat]);
  variable multipars = array_map(String_Type, &sprintf, "multi%d", [1:numfeat]);
  % define fit functions
  add_slang_function("eqwFit", ["E_min", "E_max", widthpars, multipars]);
  % define fit defaults
  variable eqwFit_defaults = "[\
    struct { value = 0, freeze = 1, hard_min = 0, hard_max = 1000, min = 0.0, max = 1000.0, step = 1e-3, relstep = 1e-5 },\
    struct { value = 0, freeze = 1, hard_min = 0, hard_max = 1000, min = 0.0, max = 1000.0, step = 1e-3, relstep = 1e-5 },\
    struct { value = 1000, freeze = 0, hard_min = -1e6, hard_max = 1e6, min = -10000, max = 10000, step = 1e-3, relstep = 1e-5 },\
    struct { value = 0, freeze = 1, hard_min = 0, hard_max = 1, min = 0, max = 1, step = 1e-3, relstep = 1e-5 }\
  ];";
  % define default hook
  eval(sprintf("define eqwFit_default_hook(i) {\
    variable var = %s;\
    return (i < 2 ? var[i] : (i-2) < %d ? var[2] : var[3]);\
  }\
  set_param_default_hook(\"eqwFit\", &eqwFit_default_hook);", eqwFit_defaults, numfeat));
}

%
% Code from src/slang/math/erfinv.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

private define erfinv_kernel(z,eps) {
    %
    % newton method for inverse erf
    %
    variable start=z;
    variable stp;
    variable del=10.;
    while(abs(del)>eps) {
        variable dfdx=2.*exp(-start^2)/sqrt(PI);
        del=(gsl->erf(start)-z)/dfdx;
        start=start-del;
    }
    return start;
}

define erfinv(z)
%!%+
%\function{erfinv}
%\synopsis{Computes the inverse error function in abs(z)<1}
%\usage{Double_Type erfinv(Double_Type z);}
%\qualifiers{
%   \qualifier{eps}{See description. Typically not needed.}
%}    
%\description
%    This function computes the inverse error function, i.e.,
%    erf(erfinv(x))=x. This can be used to calculate a confidence level by
%    \code{sqrt(2)*erfinv(fraction)}
%
%    By default the function uses two fast polynomial approximations
%    by Mike Giles which have a relative accuracy of better than 1.2e-7
%    over the whole interval between 0 and 1.
%
%    If the eps qualifier is given, the method switches to a Newton-Raphson
%    method that terminates once the relative error of the function is
%    smaller than eps. This is significantly slower and only needed if
%    an extremely high precision is needed (pretty much never). On Joern's
%    laptop, this method needs on average 0.08ms per function evaluation
%    compared to 0.03 mus for the polynomial approximation. 
%
%    The function is array safe.
%
%\seealso{erf [in gsl],cerf,cerfc}
%!%-
{

    if (max(z)>=1. || min(z)<=-1.) { % z can be array
      vmessage("Valid range only abs(z)<1");
      return 0;
    }

    if (qualifier_exists("eps")) {
        variable eps=qualifier("eps",1e-10);

        if (typeof(z)==Array_Type) {
            return parallel_map(Double_Type,&erfinv_kernel,z,eps);
        }
        return erfinv_kernel(z,eps);
    }

    %
    % inverse erf based on GPU code by Mike Giles,
    % https://people.maths.ox.ac.uk/gilesm/files/gems_erfinv.pdf
    % 
    % The code is implemented as two Horner schemes. We use them
    % twice, once for arrays and once for scalars, avoiding a loop
    % for speed reasons.
    %
    
    variable w = -log((1.0-z)*(1.0+z));
    variable p;
    if (typeof(z)==Array_Type) {
        variable ndx,ndx2;
        p=Double_Type[length(w)];
        ndx=where(w<5.,&ndx2);
        if ( length(ndx)>0 ) {
            w[ndx] = w[ndx] - 2.500000;
            p[ndx] = 2.81022636e-08;
            p[ndx] = 3.43273939e-07 + p[ndx]*w[ndx];
            p[ndx] = -3.5233877e-06 + p[ndx]*w[ndx];
            p[ndx] = -4.39150654e-06 + p[ndx]*w[ndx];
            p[ndx] = 0.00021858087 + p[ndx]*w[ndx];
            p[ndx] = -0.00125372503 + p[ndx]*w[ndx];
            p[ndx] = -0.00417768164 + p[ndx]*w[ndx];
            p[ndx] = 0.246640727 + p[ndx]*w[ndx];
            p[ndx] = 1.50140941 + p[ndx]*w[ndx];
        }
        if (length(ndx2)>0) {
            w[ndx2] = sqrt(w[ndx2]) - 3.000000;
            p[ndx2] = -0.000200214257;
            p[ndx2] = 0.000100950558 + p[ndx2]*w[ndx2];
            p[ndx2] = 0.00134934322 + p[ndx2]*w[ndx2];
            p[ndx2] = -0.00367342844 + p[ndx2]*w[ndx2];
            p[ndx2] = 0.00573950773 + p[ndx2]*w[ndx2];
            p[ndx2] = -0.0076224613 + p[ndx2]*w[ndx2];
            p[ndx2] = 0.00943887047 + p[ndx2]*w[ndx2];
            p[ndx2] = 1.00167406 + p[ndx2]*w[ndx2];
            p[ndx2] = 2.83297682 + p[ndx2]*w[ndx2];
        }
        return p*z;
    } 

    if ( w < 5. ) {
        w = w - 2.500000;
        p = 2.81022636e-08;
        p = 3.43273939e-07 + p*w;
        p = -3.5233877e-06 + p*w;
        p = -4.39150654e-06 + p*w;
        p = 0.00021858087 + p*w;
        p = -0.00125372503 + p*w;
        p = -0.00417768164 + p*w;
        p = 0.246640727 + p*w;
        p = 1.50140941 + p*w;
    } else {
        w = sqrt(w) - 3.000000;
        p = -0.000200214257;
        p = 0.000100950558 + p*w;
        p = 0.00134934322 + p*w;
        p = -0.00367342844 + p*w;
        p = 0.00573950773 + p*w;
        p = -0.0076224613 + p*w;
        p = 0.00943887047 + p*w;
        p = 1.00167406 + p*w;
        p = 2.83297682 + p*w;
    }

    return p*z;
}

#endif

%
% Code from src/plot/erg2keV.sl
%
define erg2keV(value)
%!%+
%\function{erg2keV}
%\synopsis{converts energy from erg to keV}
%\usage{Double_Type new_value = erg2keV(Double_Type old_value)}
%\qualifiers{
%\qualifier{y_fac}{: divide the value by 10^{y_fac} }
%}
%\seealso{plot_unfold}
%!%- 
{
   variable y_fac = qualifier("y_fac",0);

   variable fac0 = 624150647.996;

   return value*fac0/(10^y_fac);
}

%
% Code from src/plot/err_map_gaussian.sl
%
define err_map_gaussian ()
%!%+
%\function{err_map_gaussian}
%\synopsis{plots data points with their errorbars}
%\usage{err_map_gaussian(x, [xErr,] y, yErr);
%\altusage{err_map_gaussian(Struct_Type s);}
%}
%\qualifiers{
%\qualifier{xerr}{               change 3-argument-syntax to \code{err_map_gaussian(x, xErr, y);}}
%\qualifier{xminmax}{            changes the meaning of \code{xErr} -- and \code{x}, if \code{xErr} is not a list}
%\qualifier{yminmax}{            changes the meaning of \code{yErr} -- and \code{y}, if \code{yErr} is not a list}
%\qualifier{minmax}{             equivalent to both \code{x}- and \code{yminmax}}
%\qualifier{x_pixel [=400]}{     number of x-axis bins of the image}
%\qualifier{y_pixel [=400]}{     number of y-axis bins of the image}
%\qualifier{i}{                  index-array of subset of data points to be plotted}
%\qualifier{xmin [=min(x data)]}{minimal value of x-axis}
%\qualifier{xmax [=max(x data)]}{maximal value of x-axis}
%\qualifier{ymin [=min(y data)]}{minimal value of y-axis}
%\qualifier{ymax [=max(y data)]}{maximal value of y-axis}
%\qualifier{min_xerr [=1e-10]}{  minimal error for x-values}
%\qualifier{min_yerr [=1e-10]}{  minimal error for y-values}
%\qualifier{xlog}{               switch to logarithmic x-axis}
%\qualifier{ylog}{               switch to logarithmic y-axis}
%}
%\description
%    Plots a 2D gaussian for each data point. The given errors are used as the
%    width(1 sigma) of the Gaussian profiles. If asymmetric errors are used, the
%    profile is the combination of two Gaussians. The volume of each profile is
%    normalized.
%    
%    In order to use asymetric errors for x and/or y,
%    the correspondig \code{Err} argument has to be a list \code{{ Err1, Err2 }}.
%    If one of the \code{minmax} qualifiers is used,
%    the corresponding \code{Err} list contains directly minimum and maximum values.
%
%    If one of the \code{minmax} qualifiers is used, but \code{Err} is not a list,
%    the value and \code{Err} arguments actually mean minimum and maximum values.
%    The actual value is infered to be the mean of minimum and maximum.
%\examples
%    % examples with symmetrical errorbars:\n
%     plot_image( err_map_gaussian([1,2], [0.1,0.3], [1,1], [0.5,0.3];xmin=0,xmax=3,ymin=0,ymax=2) );\n
%    \n
%    % examples with asymmetrical errorbars:
%     plot_image( err_map_gaussian([1], {[0.1],[0.3]}, [1], {[0.1],[0.1]};xmin=0,xmax=2,ymin=0,ymax=2) );\n
%    \n
%\seealso{plot_with_err}
%!%-
{
  variable s, x, xErr=NULL, y, yErr=NULL;
  switch(_NARGS)
  { case 1:
      s = ();
      x    = (s.bin_lo + s.bin_hi)/2.;
      xErr = (s.bin_hi - s.bin_lo)/2.;
      y    =  s.value;
      if(struct_field_exists(s, "err"))    yErr = s.err;  % <- standard
      if(struct_field_exists(s, "error"))  yErr = s.error;  % <- convenient (?)
  }
  { case 2:
      vmessage("warning (%s): only two arguments specified");
      help(_function_name()); return;
  }
  { case 3:
      if(qualifier_exists("xerr"))
        (x, xErr, y      ) = ();
      else
        (x,       y, yErr) = ();
  }
  { case 4: (x, xErr, y, yErr) = (); }
  { help(_function_name()); return; }
  
  variable xminmax = _NARGS>1 && (qualifier_exists("minmax") || qualifier_exists("xminmax"));
  variable yminmax = _NARGS>1 && (qualifier_exists("minmax") || qualifier_exists("yminmax"));
  variable xMin, xMax, yMin, yMax;
  if(xErr==NULL)
    (xMin, xMax) = (0, 0);%(x, x);
  else
    if(typeof(xErr)!=List_Type)
      if(xminmax)
       (xMin, xMax, x) = ((x+xErr)/2.-x, xErr-(x+xErr)/2., (x+xErr)/2.);%(x, xErr, (x+xErr)/2.);
      else
       (xMin, xMax) = (xErr, xErr);%(x-xErr, x+xErr);
    else
      if(xminmax)
        (xMin, xMax) = (x-xErr[0], xErr[1]-x);%(xErr[0], xErr[1]);
      else
        (xMin, xMax) = (xErr[0], xErr[1]);%(x-xErr[0], x+xErr[1]);

  if(yErr==NULL)
    (yMin, yMax) = (y, y);
  else
    if(typeof(yErr)!=List_Type)
      if(yminmax)
        (yMin, yMax, y) = ((y+yErr)/2.-y, yErr-(y+yErr)/2., (y+yErr)/2.);%(y, yErr, (y+yErr)/2.);
      else
        (yMin, yMax) = (yErr, yErr);%(y-yErr, y+yErr);
    else
      if(yminmax)
        (yMin, yMax) = (y-yErr[0], yErr[1]-y);%(yErr[0], yErr[1]);
      else
        (yMin, yMax) = (yErr[0], yErr[1]);%(y-yErr[0], y+yErr[1]);

  variable i = qualifier("i");
  if(i!=NULL)
    (x,    xMin,    xMax,    y,    yMin,    yMax   ) =
    (x[i], xMin[i], xMax[i], y[i], yMin[i], yMax[i]);
  
  
  
  variable min_xerr  = qualifier("xerr_min",1e-10);
  variable min_yerr  = qualifier("yerr_min",1e-10);
  variable X = qualifier("x_pixel",400);
  variable Y = qualifier("y_pixel",400);
  variable A = Double_Type[Y,X];
  variable j,k;

  variable y_min = qualifier("ymin",min(y+yMin));
  variable y_max = qualifier("ymax",max(y+yMax));
  variable ylo,yhi;
  if (qualifier_exists("ylog")) (ylo, yhi) = log_grid (y_min, y_max, Y);
  else(ylo, yhi) = linear_grid (y_min, y_max, Y);

  variable x_min = qualifier("xmin",min(x+xMin));
  variable x_max = qualifier("xmax",max(x+xMax));
  variable xlo,xhi;
  if (qualifier_exists("xlog")) (xlo, xhi) = log_grid (x_min, x_max, X);
  else (xlo, xhi) = linear_grid (x_min, x_max, X);
  
  A[*,*]=0.;
  variable B = @A;

  variable fl1,fl2;
  variable ff= get_fit_fun;
  fit_fun( "egauss");
  set_par( "egauss(1).area",          1,  1,             0,          5);
  set_par( "egauss(1).center",    y_min,  0,   -DOUBLE_MAX, DOUBLE_MAX);
  set_par( "egauss(1).sigma",  min_xerr,  0, 1./DOUBLE_MAX, DOUBLE_MAX);

  _for j(0,length(y)-1,1)
  {
    B[*,*]=0.;
    set_par( "egauss(1).center", y[j]);
    set_par( "egauss(1).sigma", max( [yMin[j], min_yerr] ) );
    fl1 = eval_fun_keV(ylo,yhi);
    if(typeof(yErr)==List_Type) % yMax!=yMin
    {
      set_par( "egauss(1).sigma", max( [yMax[j], min_yerr] ) );
      fl2 = eval_fun_keV(ylo,yhi)*(max( [yMax[j], min_yerr] )/max( [yMin[j], min_yerr] ));
      fl2[where(ylo <= y[j])]=0.;      fl1[where(ylo > y[j])]=0.;
      fl1 = (fl1+fl2)*( 2./(1+ (max( [yMax[j], min_yerr] )/max( [yMin[j], min_yerr] ))) );      
    }
    _for k(0,X-1,1) B[*,k]=fl1; 
    set_par( "egauss(1).center", x[j]);
    set_par( "egauss(1).sigma", max( [xMin[j],min_xerr] ) );
    fl1 = eval_fun_keV(xlo,xhi);
    if(typeof(xErr)==List_Type) % xMax!=xMin
    {
      set_par( "egauss(1).sigma", max( [xMax[j], min_xerr] ) );
      fl2 = eval_fun_keV(xlo,xhi)*(max( [xMax[j], min_xerr] )/max( [xMin[j], min_xerr] ));
      fl2[where(xlo <= x[j])]=0.;      fl1[where(xlo > x[j])]=0.;
      fl1 = (fl1+fl2)*( 2./(1+ (max( [xMax[j], min_xerr] )/max( [xMin[j], min_xerr] ))) );      
    }
    _for k(0,Y-1,1) B[k,*] *= fl1;
    A+=B;
  }

  fit_fun(ff);
  return A;
}

%
% Code from src/slang/strings/escapeStr.sl
%
define escapeStr(s)
{
  message("escapeStr is deprecated. Use S-Lang's make_printable_string instead.");
  return make_printable_string(s);
}

%
% Code from src/fitting/parameters/escapedParameterName.sl
%
define escapedParameterName(par)
{
  return strreplace(
          strreplace(
           strreplace(
            strreplace(
             strreplace(
              par, "(", ""
             ), ")", ""
            ), ".", "_"
           ), "/", "_"
	  ), ">", "_"
         );
}

%
% Code from src/slang/arrays/eval_array.sl
%
define eval_array()
%!%+
%\function{eval_array}
%\synopsis{evaluate an expression for an array of values}
%\usage{Array_Type eval_array(Type_Type type, String_Type expression, Array_Type values)
%\altusage{eval_array(String_Type expression, Array_Type values);}
%}
%\qualifiers{
%\qualifier{marker}{[\code{="*"}] place holder in the \code{expression} string,
%                     where the \code{values} are to be inserted}
%\qualifier{n_markers}{[=all] number of markers in \code{expression} to be replaced;
%                        if positive, the first \code{n_markers} are replaced,
%                        if negative, the last \code{|n_markers|} are replaced.}
%}
%\description
%    The \code{marker} string in \code{expression} is sequentially replaced with
%    each value of the array \code{values} (note that the values are converted
%    through the \code{string} function), and the resulting string is evaluated.
%    Unless \code{type==Void_Type}, each evaluation is expected to result in one value
%    of this \code{type}, and \code{eval_array} returns an array of all these values.
%    If \code{type==Void_Type} (which can be omitted), no return value is expected.
%\examples
%    \code{%} set fields in an array of structures\n
%    \code{public variable s = Struct_Type[8];  _for $1 (0, 7, 1)  s[$1] = struct { a, b };}\n
%    \code{eval_array("s[*].a = 1", [0:3]);  % =>  s[[0:3]].a = 1;}\n
%    \code{eval_array("s[*].a = 2", [4:7]);  % =>  s[[4:7]].a = 2;}\n
%
%    \code{%} using the array several times\n
%    \code{eval_array("s[*].b = *",     [0:3]);               % =>  s[[0:3]].b =  [0:3];}\n
%    \code{eval_array("s[*].b = * * 2", [4:7]; n_markers=2);  % =>  s[[4:7]].b =  [4:7] * 2;}\n
%
%    \code{%} using a different marker to clarify the last example\n
%    \code{eval_array("s[#].b = 2 * #", [4:7]; marker="#");   % =>  s[[4:7]].b =  2 * [4:7];}\n
%\seealso{array_map, array_struct_field, eval}
%!%-
{
  variable type=Void_Type, expression, values;
  switch(_NARGS)
  { case 2: (expression, values) = (); }
  { case 3: (type, expression, values) = (); }
  { help(_function_name()); return; }

  variable i, n = length(values), value;
  variable marker = qualifier("marker", "*");
  variable n_markers = qualifier("n_markers", strlen(expression));
  if(type != Void_Type)
  {
    variable a = type[n];
    _for i (0, n-1, 1)
    {
      strreplace(expression, marker, string(values[i]), n_markers);  % (replaced_string, n_replacements) left on stack
      _pop_n(1);                                                     % remove n_replacements from stack
      a[i] = eval;                                                   % evaluate replaced_string
    }
    return a;
  }
  else
    foreach value (values)
    {
      strreplace(expression, marker, string(value), n_markers);  % (replaced_string, n_replacements) left on stack
      _pop_n(1);                                                 % remove n_replacements from stack
      eval;                                                      % evaluate replaced_string
    }
}

%
% Code from src/data/spectrum.datastructures/eval_fun2_keV.sl
%
%%%%%%%%%%%%%%%%%%%%
define eval_fun2_keV()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{eval_fun2_keV}
%\synopsis{evaluate a fit-function on a user-defined energy-grid}
%\usage{flux = eval_fun2_keV(handle, E_bin_lo, E_bin_hi[, params[, args]]);}
%\description
%    The fit-function given by \code{handle} (\code{S_E(E)}) is evaluated
%    on an arbitrary grid  defined by \code{E_bin_lo} and \code{E_bin_hi} (in keV):\n
%       \code{flux = int_{E_bin_lo}^{E_bin_hi} S_E(E) dE}\n
%    The unit of \code{flux} is ph/s/cm^2/bin.
%\seealso{eval_fun2, eval_fun_keV}
%!%-
{
  variable handle, keV_bin_lo, keV_bin_hi, params, args;
  switch(_NARGS)
  { case 3: (handle, keV_bin_lo, keV_bin_hi) = (); }
  { case 4: (handle, keV_bin_lo, keV_bin_hi, params) = (); }
  { case 5: (handle, keV_bin_lo, keV_bin_hi, params, args) = (); }
  { help(_function_name()); return; }

  variable A_bin_lo = _A(keV_bin_hi);
  variable A_bin_hi = _A(keV_bin_lo);
  variable flux;
  switch(_NARGS)
  { case 3: flux = eval_fun2(handle, A_bin_lo, A_bin_hi); }
  { case 4: flux = eval_fun2(handle, A_bin_lo, A_bin_hi, params); }
  { case 5: flux = eval_fun2(handle, A_bin_lo, A_bin_hi, params, args); }
  return reverse(flux);
}

%
% Code from src/data/spectrum.datastructures/eval_fun_keV.sl
%
define eval_fun_keV()
%!%+
%\function{eval_fun_keV}
%\synopsis{evaluate the fit-function on a user-defined energy-grid}
%\usage{Double_Type flux[] = eval_fun_keV(Double_Type E_bin_lo[], E_bin_hi[]);}
%\description
%    The currently defined fit-function \code{S_E(E)} is evaluated
%    on an arbitrary grid  defined by \code{E_bin_lo} and \code{E_bin_hi} (in keV):\n
%       \code{flux = int_{E_bin_lo}^{E_bin_hi} S_E(E) dE}\n
%   Note that flux is bin integrated, i.e., the unit of \code{flux} is
%   ph/s/cm^2. To plot the flux density (ph/s/cm^2/keV), you need to
%   divide this quantity by the bin  width de=E_bin_hi - E_bin_lo. 
%\seealso{eval_fun}
%!%-
{
  variable keV_bin_lo, keV_bin_hi;
  switch(_NARGS)
  { case 2: (keV_bin_lo, keV_bin_hi) = (); }
  { help(_function_name()); return; }

  variable A_bin_lo = _A(keV_bin_hi);
  variable A_bin_hi = _A(keV_bin_lo);
  variable flux = eval_fun(A_bin_lo, A_bin_hi);
  return reverse(flux);
}

%
% Code from src/fitting/xyfit/fit-functions/exponential_xyfit.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define exponential_xyfit()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{exponential_xyfit}
%\synopsis{linear xy fit function to be used with xyfit_fun}
%\usage{xyfit_fun ("exponential");}
%\description
%    This function is not meant to be called directly!
%    
%    Calling \code{xyfit_fun ("exponential");} sets up a powerlaw fit
%    function for xy-data. It has the form \code{y = norm*x^{-index}}
%\seealso{xyfit_fun, define_xydata, plot_xyfit, linear_regression}
%!%-
{
  variable xref, yref, par;
  switch(_NARGS)
  { case 0: return ["norm", "y0", "roll"]; } % fit parameters
  { case 3: (xref, yref, par)=(); }
  { return help(_function_name); }

  @yref = par[0] * (exp((@xref/par[2])) - 1 ) + par[1] ;
}

define exponential_xyfit_default(i)
{
  switch(i)
  { case 0: return (1, 0, -1e5, 1e5); }
  { case 1: return (0, 0, -1e5, 1e5); }
  { case 2: return (1, 0, -1e5, 1e5); }
}

%
% Code from src/misc/atomic/ext_info_string.sl
%
define ext_info_string()
%!%+
%\function{ext_info_string}
%\synopsis{converts the ext_line_info into a string}
%\usage{String_Type ext_info_string(Struct_Type info[, Integer_Type format])}
%\description
%    format=0 => string [default] \n
%    format=1 => PGPLOT \n
%    format=2 => TeX
%\seealso{ext_line_info, ext_info_string}
%!%-
{
  variable info, format=0;
  switch(_NARGS)
  { case 1:  info = (); }
  { case 2:  (info, format) = (); }
  { print("usage: ext_info_string(info[, format]); format = 0->string, 1->PGPLOT, 2->TeX"); return; }

  if(info != NULL)
  {
    variable type = ext_line_type(info.nr, format);
    if(type != NULL)
    {
      if(info.ion>0)
      { return atom_name(info.Z) + " " + Roman(info.ion) + " " + type; }
      else
      { return atom_name(info.Z) + type; }
    }
  }
  return NULL;
}

%
% Code from src/misc/atomic/ext_info_string_PGPLOT.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define ext_info_string_PGPLOT()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ext_info_string_PGPLOT}
%\synopsis{converts the ext_line_info into a PGPLOT string}
%\usage{String_Type ext_info_string_PGPLOT(Struct_Type info)}
%\seealso{ext_line_info, ext_info_string}
%!%-
{
  variable info;
  switch(_NARGS)
  { case 1:  info = (); }
  { help(_function_name()); return; }

  return ext_info_string(info, 1);
}

%
% Code from src/misc/atomic/ext_line_info.sl
%
define ext_line_info()
%!%+
%\function{ext_line_info}
%\usage{Struct_Type info = ext_line_info(Integer_Type id);
%\altusage{Struct_Type info = ext_line_info(Integer_Type Z, Integer_Type ion, Integer_Type nr);}
%}
%\seealso{line_info}
%!%-
{
  variable id, Z, ion, nr, l = NULL, A = NULL, g_up = NULL, g_low = NULL, gf = NULL;

  switch(_NARGS)
  { case 1:
      id = ();
      if(id<99000000 or id>99999999)  { require_atoms(); return struct_combine(line_info(id), struct{ nr=-id }); }
      Z   = (id - 99000000)/10000;
      ion = (id - 99000000 - 10000*Z)/100;
      nr  =  id - 99000000 - 10000*Z - 100*ion;
  }
  { case 3: (Z, ion, nr) = (); }
  { help(_function_name()); return NULL; }

  % # K alpha lines  (House, 1969) #
  % Aluminium
  if(Z==13 && ion== 2 && nr== 1)  l = 8.339;
  if(Z==13 && ion== 3 && nr== 1)  l = 8.336;
  if(Z==13 && ion== 4 && nr== 1)  l = 8.332;
  if(Z==13 && ion== 5 && nr== 1)  l = 8.328;
  if(Z==13 && ion== 6 && nr== 1)  l = 8.269;
  if(Z==13 && ion== 7 && nr== 1)  l = 8.203;
  if(Z==13 && ion== 8 && nr== 1)  l = 8.129;
  if(Z==13 && ion== 9 && nr== 1)  l = 8.050;
  if(Z==13 && ion==10 && nr== 1)  l = 7.964;
  if(Z==13 && ion==11 && nr== 1)  l = 7.885;

  % Silicon
  if(Z==14 && ion== 2 && nr== 1)  l = 7.126;
  if(Z==14 && ion== 3 && nr== 1)  l = 7.124;
  if(Z==14 && ion== 4 && nr== 1)  l = 7.121;
  if(Z==14 && ion== 5 && nr== 1)  l = 7.117;
  if(Z==14 && ion== 6 && nr== 1)  l = 7.112;
  if(Z==14 && ion== 7 && nr== 1)  l = 7.063;
  if(Z==14 && ion== 8 && nr== 1)  l = 7.007;
  if(Z==14 && ion== 9 && nr== 1)  l = 6.947;
  if(Z==14 && ion==10 && nr== 1)  l = 6.882;
  if(Z==14 && ion==11 && nr== 1)  l = 6.813;
  if(Z==14 && ion==12 && nr== 1)  l = 6.750;

  % Argon
  if(Z==18 && ion== 2 && nr== 1)  l = 4.193;
  if(Z==18 && ion== 3 && nr== 1)  l = 4.192;
  if(Z==18 && ion== 4 && nr== 1)  l = 4.190;
  if(Z==18 && ion== 5 && nr== 1)  l = 4.189;
  if(Z==18 && ion== 6 && nr== 1)  l = 4.186;
  if(Z==18 && ion== 7 && nr== 1)  l = 4.184;
  if(Z==18 && ion== 8 && nr== 1)  l = 4.180;
  if(Z==18 && ion== 9 && nr== 1)  l = 4.178;
  if(Z==18 && ion==10 && nr== 1)  l = 4.174;
  if(Z==18 && ion==11 && nr== 1)  l = 4.147;
  if(Z==18 && ion==12 && nr== 1)  l = 4.119;
  if(Z==18 && ion==13 && nr== 1)  l = 4.089;
  if(Z==18 && ion==14 && nr== 1)  l = 4.057;
  if(Z==18 && ion==15 && nr== 1)  l = 4.025;
  if(Z==18 && ion==16 && nr== 1)  l = 3.995;

  % Iron K alpha line  (House, 1969) #
  if(Z==26 && ion== 1 && nr== 1)  l = 1.937;
  if(Z==26 && ion== 2 && nr== 1)  l = 1.937;
  if(Z==26 && ion== 3 && nr== 1)  l = 1.937;
  if(Z==26 && ion== 4 && nr== 1)  l = 1.937;
  if(Z==26 && ion== 5 && nr== 1)  l = 1.937;
  if(Z==26 && ion== 6 && nr== 1)  l = 1.937;
  if(Z==26 && ion== 7 && nr== 1)  l = 1.937;
  if(Z==26 && ion== 8 && nr== 1)  l = 1.938;
  if(Z==26 && ion== 9 && nr== 1)  l = 1.938;
  if(Z==26 && ion==10 && nr== 1)  l = 1.938;
  if(Z==26 && ion==11 && nr== 1)  l = 1.937;
  if(Z==26 && ion==12 && nr== 1)  l = 1.936;
  if(Z==26 && ion==13 && nr== 1)  l = 1.934;
  if(Z==26 && ion==14 && nr== 1)  l = 1.933;
  if(Z==26 && ion==15 && nr== 1)  l = 1.931;
  if(Z==26 && ion==16 && nr== 1)  l = 1.930;
  if(Z==26 && ion==17 && nr== 1)  l = 1.928;
  if(Z==26 && ion==18 && nr== 1)  l = 1.927;
  if(Z==26 && ion==19 && nr== 1)  l = 1.917;
  if(Z==26 && ion==20 && nr== 1)  l = 1.907;
  if(Z==26 && ion==21 && nr== 1)  l = 1.897;
  if(Z==26 && ion==22 && nr== 1)  l = 1.886;
  if(Z==26 && ion==23 && nr== 1)  l = 1.875;
  if(Z==26 && ion==24 && nr== 1)  l = 1.865;
  if(Z==26 && ion==25 && nr== 1)  l = 1.855;
  if(Z==26 && ion==26 && nr== 1)  l = ext_line_info(26,26,21).lambda;

  % RRCs  (Pollock, priv. comm.)
  if(Z== 6 && ion== 6 && nr==11)  l = 25.3035;
  if(Z== 7 && ion== 7 && nr==11)  l = 18.5872;
  if(Z== 8 && ion== 8 && nr==11)  l = 14.2281;
  if(Z==10 && ion==10 && nr==11)  l =  9.1019;
  if(Z==12 && ion==12 && nr==11)  l =  6.3172;
  if(Z==13 && ion==13 && nr==11)  l =  5.3810;
  if(Z==14 && ion==14 && nr==11)  l =  4.6381;
  if(Z==16 && ion==16 && nr==11)  l =  3.5483;
  if(Z==18 && ion==18 && nr==11)  l =  2.8012;
  if(Z==26 && ion==26 && nr==11)  l =  1.3364;

  if(Z== 6 && ion== 5 && nr==11)  l = 31.6219;
  if(Z== 7 && ion== 6 && nr==11)  l = 22.4582;
  if(Z== 8 && ion== 7 && nr==11)  l = 16.7709;
  if(Z==10 && ion== 9 && nr==11)  l = 10.3682;
  if(Z==12 && ion==11 && nr==11)  l =  7.0374;
  if(Z==13 && ion==12 && nr==11)  l =  5.9438;
  if(Z==14 && ion==13 && nr==11)  l =  5.0863;
  if(Z==16 && ion==15 && nr==11)  l =  3.8460;
  if(Z==18 && ion==17 && nr==11)  l =  3.0089;
  if(Z==26 && ion==25 && nr==11)  l =  1.4045;

  variable atomdbids = Integer_Type[0];

  % H-like ions: Lyman series
  if( Z==ion && nr>=21 && nr<50 )
  {
    if(Z==1 or Z==2 or Z==6 or Z==7 or Z==8 or Z==10 or Z==12 or Z==13 or Z==14 or Z==16 or Z==18 or Z==20 or Z==26 or Z==28)
    { require_atoms();
      % Ly alpha
      if(nr==21)  atomdbids = where(trans(Z, ion, [ 4,  3], 1));   if(nr==31)  atomdbids = where(trans(Z, ion, [ 3], 1));   if(nr==41)  atomdbids = where(trans(Z, ion, [ 4], 1));
      % Ly beta
      if(nr==22)  atomdbids = where(trans(Z, ion, [ 7,  6], 1));   if(nr==32)  atomdbids = where(trans(Z, ion, [ 6], 1));   if(nr==42)  atomdbids = where(trans(Z, ion, [ 7], 1));
      % Ly gamma
      if(nr==23)  atomdbids = where(trans(Z, ion, [12, 11], 1));   if(nr==33)  atomdbids = where(trans(Z, ion, [11], 1));   if(nr==43)  atomdbids = where(trans(Z, ion, [12], 1));
      % Ly delta
      if(nr==24)  atomdbids = where(trans(Z, ion, [19, 18], 1));   if(nr==34)  atomdbids = where(trans(Z, ion, [18], 1));   if(nr==44)  atomdbids = where(trans(Z, ion, [19], 1));
    }

    if(Z==8) % O VIII
    { if(nr==35)  l = 14.6344;   if(nr==45)  l = 14.6343;
      if(nr==36)  l = 14.5243;   if(nr==46)  l = 14.5242;
      if(nr==37)  l = 14.4538;   if(nr==47)  l = 14.4537;
      if(nr==38)  l = 14.4058;   if(nr==48)  l = 14.4057;
      if(nr==39)  l = 14.3717;   if(nr==49)  l = 14.3716;
    }
    if(Z==10) % Ne X
    { if(nr==35)  l =  9.3617;   if(nr==45)  l =  9.3616;
      if(nr==36)  l =  9.2913;   if(nr==46)  l =  9.2912;
      if(nr==37)  l =  9.2462;   if(nr==47)  l =  9.2461;
      if(nr==38)  l =  9.2155;   if(nr==48)  l =  9.2154;
      if(nr==39)  l =  9.1936;   if(nr==49)  l =  9.1936;
    }
    if(Z==11) % Na XI
    { if(nr==31)  l = 10.0286;   if(nr==41)  l = 10.0232;
      if(nr==32)  l =  8.4603;   if(nr==42)  l =  8.4591;
      if(nr==33)  l =  8.0214;   if(nr==43)  l =  8.0209;
      if(nr==34)  l =  7.8333;   if(nr==44)  l =  7.8331;
      if(nr==35)  l =  7.7349;   if(nr==45)  l =  7.7347;
      if(nr==36)  l =  7.6767;   if(nr==46)  l =  7.6766;
      if(nr==37)  l =  7.6394;   if(nr==47)  l =  7.6393;
      if(nr==38)  l =  7.6140;   if(nr==48)  l =  7.6140;
      if(nr==39)  l =  7.5960;   if(nr==49)  l =  7.5960;
    }
    if(Z==12) % Mg XII
    { if(nr==35)  l =  6.4975;   if(nr==45)  l =  6.4974;
      if(nr==36)  l =  6.4486;   if(nr==46)  l =  6.4486;
      if(nr==37)  l =  6.4173;   if(nr==47)  l =  6.4173;
      if(nr==38)  l =  6.3960;   if(nr==48)  l =  6.3960;
      if(nr==39)  l =  6.3809;   if(nr==49)  l =  6.3809;
    }
    if(Z==13) % Al XIII
    { if(nr==35)  l =  5.5346;   if(nr==45)  l =  5.5344;
      if(nr==36)  l =  5.4929;   if(nr==46)  l =  5.4929;
      if(nr==37)  l =  5.4663;   if(nr==47)  l =  5.4662;
      if(nr==38)  l =  5.4481;   if(nr==48)  l =  5.4481;
      if(nr==39)  l =  5.4352;   if(nr==49)  l =  5.4352;
    }
    if(Z==14) % Si XIV
    { if(nr==35)  l =  4.7705;   if(nr==45)  l =  4.7704;
      if(nr==36)  l =  4.7346;   if(nr==46)  l =  4.7345;
      if(nr==37)  l =  4.7116;   if(nr==47)  l =  4.7116;
      if(nr==38)  l =  4.6960;   if(nr==48)  l =  4.6960;
      if(nr==39)  l =  4.6849;   if(nr==49)  l =  4.6848;
    }
    if(Z==16) % S XVI
    { if(nr==35)  l =  3.6496;   if(nr==45)  l =  3.6496;
      if(nr==36)  l =  3.6221;   if(nr==46)  l =  3.6221;
      if(nr==37)  l =  3.6045;   if(nr==47)  l =  3.6045;
      if(nr==38)  l =  3.5926;   if(nr==48)  l =  3.5926;
      if(nr==39)  l =  3.5841;   if(nr==49)  l =  3.5841;
    }
    if(Z==18) % Ar XVIII
    { if(nr==35)  l =  2.8810;   if(nr==45)  l =  2.8810;
      if(nr==36)  l =  2.8594;   if(nr==46)  l =  2.8594;
      if(nr==37)  l =  2.8455;   if(nr==47)  l =  2.8455;
      if(nr==38)  l =  2.8361;   if(nr==48)  l =  2.8361;
      if(nr==39)  l =  2.8294;   if(nr==49)  l =  2.8294;
    }
    if(Z==20) % Ca XX
    { if(nr==35)  l =  2.3313;   if(nr==45)  l =  2.3313;
      if(nr==36)  l =  2.3138;   if(nr==46)  l =  2.3138;
      if(nr==37)  l =  2.3026;   if(nr==47)  l =  2.3026;
      if(nr==38)  l =  2.2949;   if(nr==48)  l =  2.2949;
      if(nr==39)  l =  2.2895;   if(nr==49)  l =  2.2895;
    }
    if(Z==26) % Fe XXVI
    { if(nr==35)  l =  1.3744;   if(nr==45)  l =  1.3744;
      if(nr==36)  l =  1.3641;   if(nr==46)  l =  1.3641;
      if(nr==37)  l =  1.3575;   if(nr==47)  l =  1.3575;
      if(nr==38)  l =  1.3530;   if(nr==48)  l =  1.3530;
      if(nr==39)  l =  1.3498;   if(nr==49)  l =  1.3498;
    }

    if(nr<30)
      l = ( 0.5*ext_line_info(Z, ion, nr+10).lambda
           +1.0*ext_line_info(Z, ion, nr+20).lambda
          )/1.5;

    g_low = 2;
    if(nr>20 && nr<30)  g_up = [4, 2];
    if(nr>30 && nr<40)  g_up = 2;
    if(nr>40 && nr<50)  g_up = 4;
  }

  % He-like ions: triplet and resonance absorption line series
  if( Z==ion+1 && nr>=19 && nr<50 )
  { if(Z==1 or Z==2 or Z==6 or Z==7 or Z==8 or Z==10 or Z==12 or Z==13 or Z==14 or Z==16 or Z==18 or Z==20 or Z==26 or Z==28)
    { require_atoms();
      % He f
      if(nr==19)  atomdbids = where(trans(Z, ion, [2], 1));
      % He i
      if(nr==20)  atomdbids = where(trans(Z, ion, [5], 1));  % should be a mixture of 5 and 6
      if(nr==30)  atomdbids = where(trans(Z, ion, [6], 1));
      if(nr==40)  atomdbids = where(trans(Z, ion, [5], 1));
      % He alpha
      if(nr==21)  atomdbids = where(trans(Z, ion, [ 7], 1));
      % He beta
      if(nr==22)  atomdbids = where(trans(Z, ion, [13], 1));
      % He gamma
      if(nr==23)  atomdbids = where(trans(Z, ion, [23], 1));
      % He delta
      if(nr==24)  atomdbids = where(trans(Z, ion, [37], 1));
    }
    if(Z==8) % O VII
    { if(nr==25)  l = 17.2000;
      if(nr==26)  l = 17.0860;
      if(nr==27)  l = 17.0092;
      if(nr==28)  l = 16.9584;
      if(nr==29)  l = 16.9223;
    }
    if(Z==10) % Ne IX
    { if(nr==25)  l = 10.6426;
      if(nr==26)  l = 10.5650;
      if(nr==27)  l = 10.5130;
    }
    if(Z==11) % Na X
    { if(nr==19)  l = 11.190 ;  % Mewe et al. (1985, Table 1, p. [14])
      if(nr==20)  l = 11.080 ;   % Mewe et al. (1985, Table 1, p. [14])
      if(nr==21)  l = 11.0027;
      if(nr==22)  l =  9.4330;
      if(nr==23)  l =  8.9828;
      if(nr==24)  l =  8.7884;
      if(nr==25)  l =  8.6862;
      if(nr==26)  l =  8.6257;
      if(nr==27)  l =  8.5869;
      if(nr==28)  l =  8.5605;
      if(nr==29)  l =  8.5417;
    }
    if(Z==12) % Mg XI
    { if(nr==25)  l =  7.2247;
      if(nr==26)  l =  7.1741;
      if(nr==27)  l =  7.1415;
      if(nr==28)  l =  7.1194;
      if(nr==29)  l =  7.1037;
    }
    if(Z==13) % Al XII
    { if(nr==25)  l =  6.1028;
      if(nr==26)  l =  6.0598;
      if(nr==27)  l =  6.0322;
      if(nr==28)  l =  6.0134;
      if(nr==29)  l =  6.0000;
    }
    if(Z==14) % Si XIII
    { if(nr==25)  l =  5.2231;
      if(nr==26)  l =  5.1861;
      if(nr==27)  l =  5.1623;
      if(nr==28)  l =  5.1462;
      if(nr==29)  l =  5.1347;
    }
    if(Z==16) % S XV
    { if(nr==25)  l =  3.9501;
      if(nr==26)  l =  3.9219;
      if(nr==27)  l =  3.9039;
      if(nr==28)  l =  3.8916;
      if(nr==29)  l =  3.8828;
    }
    if(Z==18) % Ar XVII
    { if(nr==25)  l =  3.0950;
    }

    g_low = 1; % 1s^2 ~ ^1S_0
    if(nr==19) g_up = 3;  % He f  ->  1s2s ~ ^3S_1
    if(nr==20) g_up = 3;  % He i  ->  1s2p ~ ^3P_1
    if(nr>20 && nr<30)  g_up = 3;  % He r  ->  1s2p ~ ^1P_1
  }

  if(l==NULL && length(atomdbids)>0) { require_atoms(); l = line_info(atomdbids[0]).lambda; }
  if(A==NULL && length(atomdbids)>0) { require_atoms(); A = array_struct_field( array_map(Struct_Type, &line_info, atomdbids), "A"); }

  % return info structure
  if(l==NULL)  return NULL;
  variable info = struct { Z=Z, ion=ion, nr=nr, lambda=l, A=A, g_up=[g_up], g_low=g_low, gf };
  if(A!=NULL)  info.gf = sum(1.4992e-16*l^2 * g_up * A);

  return info;
}

%
% Code from src/misc/atomic/ext_line_type.sl
%
define ext_line_type()
{
  variable nr, format=0;
  switch(_NARGS)
  { case 1:  nr = (); }
  { case 2:  (nr, format) = (); }
  { print("usage: ext_line_type(nr[, format]); format = 0->string, 1->PGPLOT, 2->TeX"); return; }

  variable ret = [NULL, NULL, NULL];
  switch(nr)
  { case  1:  ret = ["Kalpha", `K\ga`, "K$\alpha$"]; }

  { case 11:  ret = "RRC" + ["", "", ""]; }

  { case 19:  ret = "f" + ["", "", ""]; }

  { case 20:  ret = "i" + ["", "", ""]; }
  { case 30:  ret = "i" + ["", "", ""]; }
  { case 40:  ret = "i" + ["", "", ""]; }

  { case 21:  ret = ["alpha", `\ga`, `$\alpha$`]; }
  { case 31:  ret = ["alpha", `\ga`, `$\alpha$`]; }
  { case 41:  ret = ["alpha", `\ga`, `$\alpha$`]; }

  { case 22:  ret = ["beta", `\gb`, `$\beta$`]; }
  { case 32:  ret = ["beta", `\gb`, `$\beta$`]; }
  { case 42:  ret = ["beta", `\gb`, `$\beta$`]; }

  { case 23:  ret = ["gamma", `\gg`, `$\gamma$`]; }
  { case 33:  ret = ["gamma", `\gg`, `$\gamma$`]; }
  { case 43:  ret = ["gamma", `\gg`, `$\gamma$`]; }

  { case 24:  ret = ["delta", `\gd`, `$\delta$`]; }
  { case 34:  ret = ["delta", `\gd`, `$\delta$`]; }
  { case 44:  ret = ["delta", `\gd`, `$\delta$`]; }

  { case 25:  ret = ["epsilon", `\ge`, `$\epsilon$`]; }
  { case 35:  ret = ["epsilon", `\ge`, `$\epsilon$`]; }
  { case 45:  ret = ["epsilon", `\ge`, `$\epsilon$`]; }

  { case 26:  ret = ["zeta", `\gz`, `$\zeta$`]; }
  { case 36:  ret = ["zeta", `\gz`, `$\zeta$`]; }
  { case 46:  ret = ["zeta", `\gz`, `$\zeta$`]; }

  return ret[format];
}

%
% Code from src/slang/math/factorial.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

define factorial()
%!%+
%\function{factorial}
%\synopsis{calculates n!}
%\usage{Double_Type factorial(Integer_Type n);}
%\description
%    \code{n}!  \code{=  n * (n-1) * ... * 2 * 1}
%
%    Note that \code{n} is always converted to an integer (without rounding);
%    for fractional \code{n} use, e.g., the GSL module's Gamma function.
%\seealso{gsl->gamma}
%!%-
{
  variable n;
  switch(_NARGS)
  { case 1: n = int( () ); }
  { help(_function_name()); return; }

#ifexists gsl->gamma
  return gsl->gamma(n+1);
#else
  return prod([2:n]);
#endif

}

#endif

%
% Code from src/data/spectrum.loaddata/factorized_arf_rmf.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%
define factorized_arf_rmf(arf, rmf)
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{factorized_arf_rmf}
%\synopsis{changes an ARF/RMF pair into factorized ARF/normalized RMF}
%\usage{(newARF, newRMF) = factorized_arf_rmf(ARF, RMF);}
%\seealso{factor_rsp}
%!%-
{
  variable arf_factor = factor_rsp(rmf);
%  ()=printf("# max arf_factor = %g\n", max(get_arf(arf_factor).value));
  if(arf==NULL)
  { arf = arf_factor; }
  else
  { variable arfData = get_arf(arf);
    arfData.value *= get_arf(arf_factor).value;
    delete_arf(arf_factor);
    put_arf(arf, arfData);
  }
  return (arf, rmf);
}

%
% Code from src/slang/math/factorset.sl
%
define factorset(n)
  %%%%%%%%%%%%%%%%%%%
  %!%+
  %\function{factorset}
  %\synopsis{factorizes an integer number into all possible integer factors}
  %\usage{Integer_Type factors[] = factorset(Integer_Type x);}
  %\description
  %    The array of factors will be ordered:
  %    \code{factor[i] <= factor[j]  %} for \code{i<j}.
  %!%-
{
  if( _typeof(n) != Integer_Type )
    return n;

  variable I = [1:n];
  
  return I[where( n mod I == 0 )];
}

%
% Code from src/data/pulsar/fake_pulsar_lightcurve.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define fake_pulsar_lightcurve()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fake_pulsar_lightcurve}
%\synopsis{creates a synthetic lightcurve of a pulsar}
%\usage{Struct_Type synthetic_pulsar_lightcurve(
%      Struct_Type lightcurve or Double_Type[] time,
%      Double_Type or Struct_Type period[, Struct_Type orbit
%      [, Struct_Type profile[, Struct_Type or Double_type fluxevolution
%      [, Ref_Type noise_fun]]]]
%    );}
%\qualifiers{
%    \qualifier{lcdt}{time resolution of the input lightcurve in days
%           (default: difference of first two time bins)}
%    \qualifier{interpol}{function reference used to to time-grid interpolations
%               (default: &interpol_points)}
%    \qualifier{pfold}{structure of qualifiers to be passed to 'pfold'
%            (default: struct { nbins = 32, dt = ..., t0 = ..., pdot = ... })}
%    \qualifier{fluxlc}{structure of qualifiers to be passed to 'pulse2pulse_flux_lc'
%             (default: NULL)}
%    \qualifier{tophase}{structure of qualifiers to be passed to 'pulseperiod2phase'
%              (default: struct { t0 = ... })}
%    \qualifier{chatty}{show or hide output messages (default: 1)}
%}
%\description
%    This function fakes a pulsar's lightcurve including the following aspects:
%    - longterm flux evolution (on timescales larger than the pulse period)
%    - lightcurve modulation by pulse profile
%    - pulse period change including orbital motion
%    - gaussian or user-defined observation noise
%
%    Two modi are possible:
%    a) providing an observed initial lightcurve from which all needed aspects
%       are derived. A pulse period or its evolution is mandatory. Certain
%       aspects can be overwritten by user input. The resulting faked and
%       the input lightcurve have the same time-grid.
%    b) providing the output time-grid. The to be included aspects have to be
%       given explicitely.
%\seealso{check_pulseperiod_orbit_struct,pulse2pulse_flux_lc,pulseperiod2phase,pfold}
%!%-
{
  variable in, period, orbit = NULL, profile = NULL, flux = NULL, noise = 1;
  switch (_NARGS)
    { case 2: (in, period) = (); }
    { case 3: (in, period, orbit) = (); }
    { case 4: (in, period, orbit, profile) = (); }
    { case 5: (in, period, orbit, profile, flux) = (); }
    { case 6: (in, period, orbit, profile, flux, noise) = (); }
    { help(_function_name); }

  % qualifiers
  variable chatty = qualifier("chatty", 1);
  variable lcdt = qualifier("lcdt", NULL);
  variable intpol = qualifier("interpol", &interpol_points);

  % determine mode: lightcurve or time-grid given
  variable mode;
  switch (typeof(in))
    { case Struct_Type:
      ifnot (struct_field_exists(in, "time") && struct_field_exists(in, "rate")) {
	vmessage("error (%s): input lightcurve needs 'time' and 'rate'", _function_name);
	return;
      }
      if (chatty) { message("mode 1: lightcurve given"); }
      if (lcdt == NULL) {
	lcdt = in.time[1]-in.time[0];
	if (chatty) { vmessage("  - assuming %f s time resolution", lcdt*86400); }
      }
      if (any(in.fracexp < 1) > 0) {
	message("warning: input lightcurve has bins with fracexp < 1");
	message("         results might be wrong");
      }
      mode = 1;
    }
    { case Array_Type:
      ifnot (_typeof(in) == Double_Type) {
	vmessage("error (%s): input time-grid is not of Double_Type", _function_name);
	return;
      }
      mode = 2;
      if (chatty) { message("mode 2: time-grid given"); }
      if ((profile == NULL && period != NULL) || typeof(flux) != Struct_Type) {
	vmessage("error (%s): a profile and flux evolution is mandatory in this mode", _function_name);
	return;
      }
    }

  % init fake lightcurve
  variable fakelc = struct {
    time = mode == 1 ? in.time : in, rate, error
  };

  % check period evolution
  variable periodtype;
  if (period != NULL) { % if period==NULL: switch pulsations off
    periodtype = check_pulseperiod_orbit_struct(period);
    if (typeof(periodtype) == Integer_Type) {
      ifnot (0 < periodtype < 3) {
        vmessage("error (%s): period is not given properly", _function_name);
        return;
      }
    } else { period = periodtype; periodtype = 2; }
  }

  % check orbit structure
  if (orbit != NULL && check_pulseperiod_orbit_struct(orbit) != 3) {
    vmessage("error (%s): orbit is not given properly", _function_name);
  }
  
  % calculate flux evolution (evolution always given in mode 2 -> assuming mode 1)
  if (flux == NULL && period != NULL) {
    if (chatty) {
      message("calculating flux evolution");
      message("  - from pulse to pulse variations");
    }
    flux = pulse2pulse_flux_lc(in, period, orbit;; qualifier("fluxlc"));
  }
  % if flux is a number -> average over larger (given) time intervals
  if (typeof(flux) == Double_Type) {
    if (chatty) {
      message("calculating flux evolution");
      message("  - from lightcurve flux evolution");
    }
    flux = pulse2pulse_flux_lc(in, flux;; qualifier("fluxlc", NULL));
  }
  % interpol flux evolution
  if (typeof(flux) == Struct_Type) {
    flux = @intpol(fakelc.time, flux.time, flux.rate);
  }

  % calculate profile (subtract the flux evolution)
  if (period != NULL && profile == NULL) { % if mode=2: profile always given
    variable tmp = struct { p0, t0 = in.time[0], pdot = 0, pddot = 0 };
    if (chatty) { message("calculating profile"); }
    if (periodtype == 2) {
      tmp.p0 = period.p0;
      if (struct_field_exists(period, "pdot")) { tmp.pdot = period.pdot; }
      if (struct_field_exists(period, "p2dot")) { tmp.pddot = period.p2dot; }
    } else {
      tmp.p0 = @intpol(tmp.t0, period.time, period.period);
    }
    if (chatty) { vmessage("  - period %f s", tmp.p0); }
    profile = pfold(in.time, in.rate-flux, tmp.p0/86400;; struct_combine(struct {
      nbins = 32, pdot = tmp.pdot, pddot = tmp.pddot*86400,
      dt = in.fracexp*lcdt, t0 = tmp.t0
    }, qualifier("pfold", NULL)));
    hplot(profile);
  }

  % calculate pulse phase
  variable inphi;
  if (period != NULL) {
    if (chatty) { message("calculating pulse phase"); }
    inphi = pulseperiod2phase(mode == 1 ? in.time : in, struct_combine(
      period, struct { t0 = mode == 1 ? in.time[0] : in[0] }, qualifier("tophase", NULL))
    );
  }
  
  %%% now start to fake the lightcurve
  if (chatty) { vmessage("creating synthetic lightcurve"); }
  fakelc.rate = Double_Type[length(fakelc.time)];
  fakelc.error = @(fakelc.rate);
  % add flux evolution
  if (flux != NULL) {
    if (chatty) vmessage("  - flux evolution");
    fakelc.rate += flux;
  }
  % add pulse profile (here the pulse phase is important! -> orbit, spin-change)
  if (period != NULL && profile != NULL) {
    if (chatty) vmessage("  - pulse profile and period changes");
    fakelc.rate += @intpol(
      inphi mod 1, .5*(profile.bin_lo + profile.bin_hi), profile.value-mean(profile.value)
    );
  }

  % if input lightcurve given match the mean value
  if (mode == 1) {
    if (chatty) { message("  - aligning means of input and fake"); }
    fakelc.rate +=  mean(in.rate) - mean(fakelc.rate);
  }

  % add noise
  if (typeof(noise) == Ref_Type) {
    if (chatty) { message("  - calling noise function"); }
    (fakelc.rate, fakelc.error) = mode == 1
                                ? (@noise)(fakelc.rate, in.rate)
	                        : (@noise)(fakelc.rate);
  } else if (noise != NULL) {
    if (mode == 1) {
      variable sdev = moment(in.rate - fakelc.rate).sdev; % just get the noise, no features
      if (chatty) {
	vmessage("  - adding gaussian noise\n    - input has sigma = %f", sdev);
      }
      fakelc.error = sdev * ones(length(fakelc.time));
    } else {
      if (chatty) { message("  - adding gaussian noise\n    - assuming cts/s"); }
      fakelc.error = sqrt(fakelc.rate/(make_hi_grid(in)-in)/86400);
    }
    fakelc.rate += grand(length(fakelc.rate)) * fakelc.error;
  }

  return fakelc;
}

%
% Code from src/slang/date/fermi2MJD.sl
%
%%%%%%%%%%%%%%%%%%
define fermi2MJD()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{fermi2MJD}
%\synopsis{calculate MJD from Fermi seconds}
%\usage{Double_Type = fermi2MJD (Double_Type);}
%\description
%	Calculates MJD(UTC) for a given Fermi Mission
%	Elapsed Time (MET) in seconds.
%\example
%	variable m = 239557420.0;
%	variable fermi_s = fermi2MJD(m);
%\seealso{MJD2fermi, MJDref_satellite}
%!%-
{ 
  variable fermi_in;
  
  switch(_NARGS)
  { case 1: fermi_in  = (); }
  { help(_function_name()); return; }

  variable mjd_ref=51910.; % MJDref_satellite ("fermi");
  variable diff1=utc2tai(mjd_ref;mjd,deltat);
  variable diff2=utc2tai(mjd_ref+fermi_in/86400.;mjd,deltat);
  return mjd_ref + (fermi_in-(diff2-diff1))/86400.;
}

%
% Code from src/data/GTI/filter_gti.sl
%
define filter_gti()
%!%+
%\function{filter_gti}
%\usage{Integer_Type ind[] = filter_gti(Double_Type time[], Struct_Type gti);
% or                      filter_gti(Double_Type time_lo[], time_hi[], Struct_Type gti);}
%\qualifiers{
%    \qualifier{minfracexp}{minimum fractional exposure a time bin has to have,
%                otherwise it is considered bad (lightcurves only, default: 1e-4)}
%    \qualifier{fracexp}{if set to a reference, returns the fractional exposure
%                of each time bin (lightcurves only)}
%    \qualifier{exposure}{if set to a reference, returns the livetime
%                in each time bin (lightcurves only)}
%    \qualifier{indarray}{return an array of index arrays instead (in case
%                of events only)}
%}
%\description
%    For a lightcurve given by \code{time_lo} and \code{time_hi}, this function
%    returns all indices \code{ind} to the time bins, for which the fractional
%    exposure time as defined by the good time intervals defined by \code{gti}
%    is at least \code{fracexp}.
%
%    For a list of events measured at times \code{time}, return a list of indices
%    to all events that were measured during the given set of good time intervals.
%    The good time intervals are defined by a \code{struct {start,stop}} where
%    \code{start} and \code{stop} define the start and stop times.
%
%    The code assumes that
%    - time and time_hi are in ascending order
%    - the gti does not contain any overlapping intervals
%    - gti, time, and time_hi have the same unit
%
%    Warning: the function applies a time sorting to the gti if necessary. This
%    	will MODIFY THE INPUT since structures are passed as references in S-Lang!
%!%-
{
    variable time, gti, time_hi = NULL;
    switch(_NARGS)
    { case 2: (time, gti) = (); }
    { case 3: (time, time_hi, gti) = (); }
    { help(_function_name()); return; }

    % consistency checks
    if (length(gti.start) != length(gti.stop)) {
	vmessage("error (%s): gti.start and gti.stop have to be of equal length", _function_name);
	return;
    }
    if (any(gti.start > gti.stop)) {
	vmessage("error (%s): gti.start > gti.stop detected", _function_name);
	return;
    }
    if (time_hi != NULL) {
	if (length(time) != length(time_hi)) {
	    vmessage("error (%s): time_lo and time_hi have to be of equal length", _function_name);
	    return;
	}
	if (any(time > time_hi)) {
	    vmessage("error (%s): time_lo > time_hi detected", _function_name);
	    return;
	}
    }

    variable i;

    %
    % ensure that gti is time sorted
    %
    if (any(gti.start[[1:]] - gti.start[[:-2]] < 0)) {
    	struct_filter(gti, array_sort(gti.start));
    	vmessage("warning (%s): GTI struct has been ordered by time!", _function_name);
    } 

    variable indarray = qualifier_exists("indarray");
    variable ind = indarray ? Array_Type[length(gti.start)] : Integer_Type[0];
      
    % abort if no GTI overlaps with time
    if (gti.stop[-1] < time[0] || gti.start[0] > (time_hi == NULL ? time[-1] : time_hi[-1])) {
    	return ind;
    } 

    %
    % filtering of events
    %
    if (time_hi == NULL ) {
	_for i (0,length(gti.start)-1,1) {
	    if (indarray) { ind[i] = where(gti.start[i]<=time<=gti.stop[i]); }
	    else { ind = [ind,where(gti.start[i]<=time<=gti.stop[i])]; }
	}
	return ind;
    }

    %
    % filtering of lightcurves
    %

    % calculate exposure of all lightcurve bins
    %
    variable npt=length(time);
    variable ngi=length(gti.stop);
    variable expo=Double_Type[npt];
    i=0; % counter for the light curve
    variable j=0; % counter for the GTI
    forever {
	% skip over all gti intervals that stop before the current lc bin
        % and
	% skip over all lc intervals that stop before the current gti bin
	while (j<ngi && i<npt) {
	  if (gti.stop[j]<time[i]) { j++; continue; }
	  if (time_hi[i]<gti.start[j]) { i++; }
	  else { break; }
	}
    	if (j == ngi || i == npt) { break; }
      
        if (gti.start[j]<=time[i]) {
	    % gti starts before the current time bin
            if (gti.stop[j]>=time_hi[i]) {
	        % gti ends after time bin -> time bin is fully exposed
                expo[i]+=time_hi[i]-time[i];
                i++; % go to next time bin
            } else {
		% gti overlaps at start of bin
                expo[i]+=gti.stop[j]-time[i];
                j++; % go to next gti bin
            }
        } else {
	    % gti starts later than time_lo
            if (gti.stop[j]<=time_hi[i]) {
		% gti is located completely in the time bin
                expo[i]+=gti.stop[j]-gti.start[j];
                j++; % go to the next gti bin
            } else {
		% gti extends beyond the end of the time bin
                expo[i]+=time_hi[i]-gti.start[j];
                i++; % go to next time bin
            }
        }
    }
	
    variable fracexp=expo[*]/(time_hi[*]-time[*]);

    % apply the fracexp limit resp. the gap finder
    ind = where(fracexp >= qualifier("minfracexp",1e-4));
    
    if (qualifier_exists("fracexp")) {
	fracexp = fracexp[ind];
	@(qualifier("fracexp")) = fracexp;
    }
    
    if (qualifier_exists("exposure")) {
	expo=expo[ind];
	@(qualifier("exposure")) = expo;
    }

    return ind;
}

%
% Code from src/slang/math/find_correlations.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define find_correlations()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{find_correlations}
%\synopsis{calculates all correlations between columns of a table}
%\usage{Struct_Type info = find_correlatins(Struct_Type s);}
%\description
%    \code{info.corr[i] = correlation_coefficient(s.<info.x[i]>, s.<info.y[i]>);}
%\seealso{correlation_coefficient}
%!%-
{
  variable s;
  switch(_NARGS)
  { case 1: s = (); }
  { help(_function_name()); return; }

  variable fields = get_struct_field_names(s);
  variable n_fields = length(fields);

  variable i, j, info = struct { x=String_Type[0], y=String_Type[0], corr=Double_Type[0] };
  _for i (0, n_fields-2, 1)
    _for j (i+1, n_fields-1, 1)
    { info.x = [info.x, fields[i]];
      info.y = [info.y, fields[j]];
      info.corr = [info.corr,
		   correlation_coefficient(get_struct_field(s, fields[i]),
					   get_struct_field(s, fields[j])
				          )
		  ];
    }
  struct_filter(info, array_sort(1-abs(info.corr)));
  return info;
}

%
% Code from src/slang/misc/find_directories.sl
%
define find_directories()
{
  variable path = ".";
  switch(_NARGS)
  { case 0: path = "."; }
  { case 1: path = (); }
  { return; }

  variable f, dirs = String_Type[0];
  foreach f (glob(path+"/*"))
    if(stat_is("dir", stat_file(f).st_mode))
      dirs = [dirs, f, find_directories(f)];
  return dirs;
}

%
% Code from src/slang/misc/find_files.sl
%
define find_files(path)
{
  variable files = String_Type[0];
  foreach (glob(path+"/*"))
  { variable f = ();
    variable stmode = stat_file(f).st_mode;
    if(stat_is("reg", stmode))  files = [files, f];
    if(stat_is("dir", stmode) && qualifier_exists("recursive"))  files = [files, find_files(f; recursive)];
  }
  return files;
}

%
% Code from src/slang/math/find_function_maximum.sl
%
define find_function_maximum()
%!%+
%\function{find_function_maximum}
%\synopsis{looks for the position of a function's maximum value}
%\usage{Double_Type x0 = find_function_maximum(Ref_Type &f, Double_Type x1, Double_Type x2);
%\altusage{Double_Type x0 = find_function_maximum(Ref_Type &f, Double_Type x1, Double_Type x2, &f_x0);}
%}
%\qualifiers{
%\qualifier{qualifiers}{structure of qualifiers to be passed to f}
%\qualifier{eps}{[=1e-12]}
%}
%\description
%    \code{f} has to be a real function with one argument.
%    A binary search is performed to find \code{x0} such that\n
%       \code{f(x0)  =  max( f([x1:x2]) )}.\n
%    If \code{f} is not convex in \code{[x1:x2]}, the algorithm does not need
%    to succeed. Otherwise, the accuracy of \code{x0} is \code{(x2-x1)*eps}.
%\seealso{find_function_value}
%!%-
{
  variable f, x1, x2, valref=NULL;
  switch(_NARGS)
  { case 3: (f, x1, x2) = (); }
  { case 4: (f, x1, x2, valref) = (); }
  { help(_function_name()); return; }

  if(x2<x1)  (x1, x2) = (x2, x1);
  variable qualifiers = qualifier("qualifiers", NULL);
  variable eps = (x2-x1)*qualifier("eps", 1e-12);
  while(x2-x1>eps)
  { variable xn1 = (2*x1+x2)/3., xn2 = (x1+2*x2)/3.;
    if(@f(xn1;; qualifiers) > @f(xn2;; qualifiers))
      x2 = xn2;
    else
      x1 = xn1;
  }
  variable x = (x1+x2)/2.;
  if(valref!=NULL)  @valref = @f(x;; qualifiers);
  return x;
}

%
% Code from src/slang/math/find_function_value.sl
%
define find_function_value()
%!%+
%\function{find_function_value}
%\synopsis{computes an inverse function}
%\usage{Double_Type x0 = find_function_value(Ref_Type &f, Double_Type val, x1, x2);}
%\qualifiers{
%\qualifier{qualifiers}{structure of qualifiers to be passed to f}
%\qualifier{eps}{[=1e-12]}
%\qualifier{quiet}{do not show error message}
%}
%\description
%    \code{f} has to be a real function with one argument.
%    A binary search is performed to find \code{x0} such that\n
%       \code{f(x0)  =  val} .\n
%    If \code{f} is not strictly monotonic in \code{[x1:x2]}, the algorithm does not
%    need to succeed. Otherwise, the accuracy of \code{x0} is \code{(x2-x1)*eps}.
%\seealso{find_multiargumentfunction_value, find_function_maximum}
%!%-
{
  variable f, val, x1, x2;
  switch(_NARGS)
  { case 4: (f, val, x1, x2) = (); }
  { help(_function_name()); return; }

  if(x2<x1)  (x1, x2) = (x2, x1);
  variable qualifiers =  qualifier("qualifiers", NULL);
  variable f_x1 = @f(x1;; qualifiers);
  variable f_x2 = @f(x2;; qualifiers);
  ifnot( (f_x1 <= val <= f_x2) or (f_x1 >= val >= f_x2) )
  { ifnot(qualifier_exists("quiet"))
      vmessage("error (" + _function_name() + "): value %g is not enclosed by f(x1=%g)=%g and f(x2=%g)=%g.", val, x1, f_x1, x2, f_x2);
    return NULL;
  }
  variable eps = (x2-x1)*qualifier("eps", 1e-12);
  while(x2-x1>eps)
  { variable x = (x1+x2)/2.;
    variable f_x = @f(x;; qualifiers);
    if( (val-f_x1)*(val-f_x) > 0)
      (x1, f_x1) = (x, f_x);
    else
      (x2, f_x2) = (x, f_x);
  }
  return (x1+x2)/2.;
}

%
% Code from src/slang/math/find_multiargumentfunction_value.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define find_multiargumentfunction_value()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{find_multiargumentfunction_value}
%\synopsis{computes an inverse function}
%\usage{Double_Type xi0 = find_multiargumentfunction_value(Ref_Type &f, Double_Type val, x1, x2, ..., xn);}
%\qualifiers{
%\qualifier{qualifiers}{structure of qualifiers to be passed to f}
%\qualifier{eps}{[=1e-12]}
%}
%\description
%    \code{f} has to be a real function with n arguments.
%    While \code{xi = [xi_min, xi_max]} is an array, \code{xj} (for \code{j!=i}) are constants.
%    A binary search is performed to find \code{xi0} such that
%    \code{f(x1, x2, ...,  xn)  =  val} for \code{xi = xi0}.\n
%    If \code{f} is not strictly monotonic for \code{xi_min < xi < xi_max}, the algorithm does not
%    need to succeed. Otherwise, the accuracy of \code{xi0} is \code{(xi_max-xi_min)/1e12}.
%\seealso{find_function_value}
%!%-
{
  variable arg = __pop_args(_NARGS-2);
  variable val = ();
  variable f = ();

  variable i;
  _for i (0, length(arg)-1, 1)
    if(typeof(arg[i].value)==Array_Type)
      break;
  if(typeof(arg[i].value)!=Array_Type)
  { message("error (" + _function_name() + "): one argument has to be an array");
    return NULL;
  }
  variable x1 = min(arg[i].value);
  variable x2 = max(arg[i].value);
  variable qualifiers = qualifier("qualifiers", NULL);
  arg[i].value = x1;  variable f_x1 = @f( __push_args(arg) ;; qualifiers);
  arg[i].value = x2;  variable f_x2 = @f( __push_args(arg) ;; qualifiers);
  ifnot(f_x1 <= val <= f_x2 || f_x1 >= val >= f_x2)
    return vmessage("error (%s): value %g is not enclosed by f(x1=%g)=%g and f(x2=%g)=%g", _function_name(), val, x1, f_x1, x2, f_x2);
  variable eps = (x2-x1)*qualifier("eps", 1e-12);
  while(x2-x1>eps)
  { arg[i].value = (x1+x2)/2.;
    variable f_x = @f( __push_args(arg) ;; qualifiers);
    if( (val-f_x1)*(val-f_x) > 0)
      (x1, f_x1) = (arg[i].value, f_x);
    else
      (x2, f_x2) = (arg[i].value, f_x);
  }
  return (x1+x2)/2.;
}

%
% Code from src/data/timing/find_peak.sl
%
private define gliding_average ( p, stat, region ) { %{{{
   % region in promille of total length of p
   variable r = nint( length(p) * region / 1000. );
   variable out = Double_Type [ length(stat) ];
   variable i;
   _for i ( 0, length(p) - 1, 1 ) {
      variable x = p [[max([0,i-r]):min([length(p)-1,i+r])]];
      variable y = stat [[max([0,i-r]):min([length(p)-1,i+r])]];
      variable model = linear_regression ( x, y );
      out[i] = model.a + model.b * x[where_min(abs(x-p[i]))];
      }
   return out;
}
%}}}

private define quality_function ( mean1, sigma, x ) { %{{{
   if (abs(x-mean1)<sigma) {
      return 1. - 0.5*((x-mean1)/sigma)^6.;
   } else {
      return 0.5/((x-mean1)/sigma)^2.;
   }
}
%}}}

private define error_estimate ( p1, stat1, period, error2_lo, error2_hi ) { %{{{
   variable lo = error2_lo;
   variable hi = error2_hi;
   variable index_min = where_min(abs(p1 - period + lo));
   variable index_max = where_min(abs(p1 - period - hi));
   variable p = p1[[index_min:index_max]];
   variable stat = stat1[[index_min:index_max]];
   variable area = Double_Type [ length(p) ];
   area[0] = stat[0];
   variable i;
   _for i ( 1, length(area)-1, 1) {
      area[i] = area[i-1] + stat[i]; }
   variable areas = area - mean(area);
   areas = sign(areas) * log1p( abs(areas) );

   variable prime = Double_Type [ length(areas) ];
   _for i ( 0, length(prime)-1, 1 ) {
      variable grid = [ max( [ 0, i-1 ] ): min ( [ length(prime)-1, i+1 ] ) ];
      prime[i] = linear_regression ( p[grid], areas[grid] ).b;
      }
   variable where_peak = where_min( abs( period - p ) );
   variable mi = where_min ( abs( prime[[:where_peak]] - 0.75 * mean(prime) ) );
   variable ma = where_min ( abs( prime[[where_peak:]] - 0.75 * mean(prime) ) ) + where_peak;
   variable error_lo = p[where_peak] - p[mi];
   variable error_hi = p[ma] - p[where_peak];

   return mean( [ error_lo, error_hi ] );
}
%}}}

%private define sqrsinc_xyfit() { %{{{
%   variable xref, yref, par;
%   switch (_NARGS)
%   { case 0: return ["center [center]", "norm [norm]", "width [width]", "offset [offset]", "asymmetry [asymmetry]" ]; }
%   { case 3: (xref, yref, par) = (); }
%   { return; }
%   variable x = @xref - par[0];
%   variable xx = x*(1+(1+sign(x))*par[4])/par[2];
%   @yref = par[3] + par[1] * sqr( sin(xx)/(xx) );
%}
%}}}

private define get_qualifier ( qualifier_structure, qualifier_name, default_value ) { %{{{
   try {
      return get_struct_field ( qualifier_structure, qualifier_name );
   } catch InvalidParmError: {
      return default_value;
   }
}
%}}}

private define print_info(arg) {%{{{
   if (arg == "flags" ) {
      () = printf(`
      ; end_reached: If the distance from the peak to one of the ends of
                     the period search range is small, this flag is
                     either -1 or 1 for the lower resp. upper end.
      ; no_blocks: This flag counts the smoothing iterations necessary
                   until the block analysis succeeds (max 4). Indicator
                   for a noisy statistics.
      ; bad_profile: Indicates that the folded profile is quite noisy.
      ; bad_blocks: Indicates that all blocks where maxima are found are
                    quite broad relative to dp, usually an indicator
                    that the statistics is not well behaved.
      ; difference: An integer valued measure for the difference between
                    the peak found and the initial guess in terms of sigma.
      ; exact: Set if epfold was done with exact qualifier.
      ; fit_succeeded: If 0, the peak found from the fit differed
                       considerably from the first estimate. In this case
                       the fitted peak is not used.
      ; dp: Gives the dp used.
      ; [ t ]: The timespan of the lightcurve. If short relative to p, this
           can cause problems.

      `R);
   } else if (arg == "error") {
      () = printf(`
      Two errors are given: 'err_theory' and 'err_area'. Normally, both
      should be approximately equal.

      'err_theory' stems from the timing background of this function and is
      calculated using an empirical estimate given by Leahy 1987. Since it
      is sensitive to the height of the peak, it should NOT be used if you
      want to find the peak of some other curve than one stemming from the
      epochfolding statistics. In the context of timing, this error might
      be slightly underestimated. Furthermore, it is to be taken with
      great care if you used the 'exact' qualifier for epfold, because in
      this case the statistics seem to be wrong.

      'err_area' is a measure for the approximate width of the plateau on
      top of the peak. It basically finds the interval for which the
      curve's fluctuations are more significant than any underlying slope.
      Normally, it is slightly larger than 'err_theory'.

      The field named 'error' is just the combination of both errors added
      in quadrature. This should be a reasonable and perhaps slightly
      too conservative estimate of the error.

      `R);
   }
}
%}}}

%%%%%%%%%%%%%%%%%%%%%
define find_peak ( ) %{{{
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{find_peak}
%\synopsis{Written for epoch folding purposes, it finds peaks of certain characteristics}
%\usage{Struct_Type find_peak(Struct_Type input)}
%\description
%
% The input should be a Struct_Type of the form:
%     { p, stat, [ lc ], [ nbins_epfold ] , [ expectation ] , [ sigma ] }
% The individual fields are:
%     * p, stat: arrays of doubles, the peak is looked for in stat
%     * lc: struct { time, rate }, the original lightcurve
%     * nbins_epfold: necessary in order to obtain an error estimate
%     * expectation: approximate location of the peak in p
%     * sigma: approximate error the expectation value has
% lc and nbins_epfold reflect the original application of this function,
% but it can be used without these arguments to find peaks in other
% curves.
% The output is a Struct_Type of the form:
%     { period, error, [ profile ], err_area, err_theory, width_lo,
%     width_hi, statvsp, flags, [ bayes ], [ fit ], badness }
% flags is a structure of indicators regarding the quality of the
% output, badness is a coarse estimate how well the process worked.
% A value of 0 is desirable, 1-2 could be acceptable, 3-4 is usually
% an indication of bad input data and for larger values something has
% seriously gone wrong. Fields marked [] will not be output if the
% qualifiers chosen require it.
% In the default configuration, this function works the following way:
%
%     (1) A bayesian block analysis finds possible maxima
%         * if the block analysis fails, a smoothing algorithm is used
%           to remove noise in the curve, since the block analysis is
%           quite susceptible to noise
%     (2) The maxima are ranked considering:
%         * the width (broad is an advantage, but very broad maxima in
%           comparison to dp = p^2 / T resp sigma are discarded)
%         * the distance to the expected value, where anything within
%           sigma is ranked approximately equal
%     (3) Maxima with a quality exceeding the tolerance level are
%         taken and the accurate peak is found:
%         * a first estimate is a weighted mean of points within the
%           maximum block
%         * the accurate value is determined by fitting a given
%           xy-function to the peak. The proportion of dp used for
%           choosing the range for the fit is controlled by the
%           qualifier fit_part. If you do not expect considerable
%           secondary maxima, this variable can be set as big as 1.
%           However, if secondary peaks are expected to occur, a value
%           this large can make the fit useless.
%     (4) For each possible peak the lightcurve is folded and the
%         resulting pulse profile is examined. The peak corresponding to
%         the best overall result is taken.
%
% If the fit_fct qualifier is used, the function has to have the following
% parameters (in this specific order):
%     (1) peak
%     (2) measure for the height
%     (3) measure for the broadness of the peak
%     (4) absolute offset
%     (5) measure for the asymmetry of the peak
% All qualifiers can also be passed using a structure named
% 'find_peak_qualifiers'. If this qualifier is given, its content will
% overwrite all other qualifiers. If using this structure, all
% qualifiers are expected to have values (i.e. only_max = 1). This is
% helpful if you want to make things easier to read in scripts.
%
%\qualifiers{
%\qualifier{only_max}{neither block analysis nor fit are performed}
%\qualifier{blocks_but_only_max}{block analysis is performed but only the highest bin in the best block taken}
%\qualifier{weighting}{weights for properties of a peak [area, sharpness, distance to expectation, profile quality] (default: [1,1,1,1])}
%\qualifier{no_fit}{nlock analysis is performed but only the first estimate taken}
%\qualifier{fit_fct}{an xyfit function for the fit, passed as a string (default: "sqrsinc")}
%\qualifier{fit_part}{measure for the part of the peak used for the fit (default: .3)}
%\qualifier{tolerance}{determines how many peaks are passed to step (4), in [0,1) (default: .99)}
%\qualifier{smoothing}{how strong the smoothing algorithm works (default: 10.)}
%\qualifier{ncp_prior}{argument for block analysis. If a string is given, the default block analysis routine will be used (default: 100)}
%\qualifier{fp_rate}{argument for block analysis, see its help}
%\qualifier{no_profile}{No pfold will be executed. Corresponds to tolerance -> 1 and less computation}
%\qualifier{nbins_pfold}{argument for pfold, see its help}
%\qualifier{pfold_not_exact}{changes how pfold is executed. Recommended as long as the exact-bug is not fixed}
%\qualifier{exact}{passes the information that the statistic was calculated using the exact qualifier}
%\qualifier{flag_info}{detailed information regarding the content of the flags structure is given, NOTHING else is done}
%\qualifier{error_info}{detailed information about the calculation of errors is given, NOTHING else is done}
%\qualifier{chatty}{boolean value (default: 1)};
%}
%\seealso{pulseperiod_search, epfold, pulseperiod_epfold, split_and_epfold_lc, bayesian_blocks, pfold}
%!%-
{
   variable qu = qualifier("find_peak_qualifiers", empty_struct());
   if (qualifier_exists("flag_info") or struct_field_exists(qu, "flag_info")) {
      print_info("flags");
      return;
   }
   if (qualifier_exists("error_info") or struct_field_exists(qu, "error_info")) {
      print_info("error");
      return;
   }

   variable lc = 0;
   variable p, stat, nbins_epfold;
   variable expectation = 0.;
   variable sigma_expectation = DOUBLE_MAX;
   % Qualifiers as usual %{{{
   % These qualifiers are given as usual. If a structure named 'find_peak_qualifiers' is given, it will *overwrite* any input here!
   variable find_peak_qualifiers_default = empty_struct ();
   variable aa;
   if (qualifier_exists("weighting")) { aa=struct{weighting=qualifier("weighting")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("ncp_prior")) { aa=struct{ncp_prior=qualifier("ncp_prior")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("fp_rate")) { aa=struct{fp_rate=qualifier("fp_rate")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("smoothing")) { aa=struct{smoothing=qualifier("smoothing")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("fit_fct")) { aa=struct{fit_fct=qualifier("fit_fct")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("fit_part")) { aa=struct{fit_part=qualifier("fit_part")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("tolerance")) { aa=struct{tolerance=qualifier("tolerance")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("dp")) { aa=struct{dp=qualifier("dp")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("nbins_pfold")) { aa=struct{nbins_pfold=qualifier("nbins_pfold")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("chatty")) { aa=struct{chatty=qualifier("chatty")};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("blocks_but_only_max")) { aa=struct{blocks_but_only_max=1};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("no_profile")) { aa=struct{no_profile=1};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("exact")) { aa=struct{exact=1};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("pfold_not_exact")) { aa=struct{pfold_not_exact=1};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("only_max")) { aa=struct{only_max=1};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   if (qualifier_exists("no_fit")) { aa=struct{no_fit=1};find_peak_qualifiers_default=struct_combine(find_peak_qualifiers_default,aa); }
   %}}}

   if (length(get_struct_field_names(find_peak_qualifiers_default)) != 0 and qualifier_exists("find_peak_qualifiers") ) {
      vmessage("Warning(%s): You have provided qualifiers both as a structure and as individual arguments, I will only consider the structure", _function_name);
   }

   variable find_peak_qualifiers = qualifier ( "find_peak_qualifiers", find_peak_qualifiers_default );
   % Qualifiers as structure, Sanity checks %{{{
   % If present, the structure 'find_peak_qualifiers' overwrites all other qualifiers.
   variable weighting = get_qualifier ( find_peak_qualifiers, "weighting", [1,1,1,1] );%{{{
   ifnot (typeof(weighting) == Array_Type and (_typeof(weighting) == Double_Type or _typeof(weighting) == Integer_Type)) {
      vmessage("Warning(%s): weighting has to be an array of numbers, setting it to default", _function_name);
      weighting = [1,1,1,1];
   }
   ifnot (length(weighting) == 4) {
      vmessage("Warning(%s): weighting has to be an array of lenght 4, setting it to default", _function_name);
      weighting = [1,1,1,1];
   } 
   weighting = double(weighting); %}}}
   variable ncp_prior = get_qualifier ( find_peak_qualifiers, "ncp_prior", 100. );%{{{
   ifnot (typeof(ncp_prior) == Double_Type or typeof(ncp_prior) == Integer_Type) {
      ifnot (ncp_prior == "default") {
         vmessage("Warning(%s): ncp_prior has to be a number or 'default', setting it to 100", _function_name);
         ncp_prior = 100.;
      }
   }
   ifnot (typeof(ncp_prior) == String_Type) {
      ncp_prior = double(ncp_prior);
   } %}}}
   variable fp_rate = get_qualifier ( find_peak_qualifiers, "fp_rate", 0.3 );%{{{
   ifnot (typeof(fp_rate) == Double_Type and fp_rate <= 1.) {
      vmessage("Warning(%s): fp_rate has to be a number <= 1.0, setting it to 0.3", _function_name);
      fp_rate = 0.3;
   } %}}}
   variable smoothing = get_qualifier ( find_peak_qualifiers, "smoothing", 10. );%{{{
   ifnot (typeof(smoothing) == Double_Type or typeof(smoothing) == Integer_Type) {
      vmessage("Warning(%s): smoothing has to be a number, setting it to 10", _function_name);
      smoothing = 10.;
   }
   smoothing = double(smoothing);%}}}
   variable fit_fct = get_qualifier ( find_peak_qualifiers, "fit_fct", "sqrsinc" );%{{{
   ifnot (typeof(fit_fct) == String_Type) {
      vmessage("Warning(%s): fit_fct has to be String_Type, setting it to sqrsinc", _function_name);
      fit_fct = "sqrsinc";
   }%}}}
   variable fit_part = get_qualifier ( find_peak_qualifiers, "fit_part", 0.3 );%{{{
   ifnot (typeof(fit_part) == Double_Type) {
      vmessage("Warning(%s): fit_part has to be Double_Type, setting it to 0.3", _function_name);
      fit_part = 0.3;
   }%}}}
   variable tolerance = get_qualifier ( find_peak_qualifiers, "tolerance", 0.99 );%{{{
   ifnot (typeof(tolerance) == Double_Type or typeof(tolerance) == Integer_Type) {
      vmessage("Warning(%s): tolerance has to be a number, setting it to 0.99", _function_name);
      tolerance = 0.99;
   }
   tolerance = double(tolerance);%}}}
   variable dp = get_qualifier ( find_peak_qualifiers, "dp", -1 );%{{{
   ifnot (typeof(dp) == Double_Type or typeof(dp) == Integer_Type) {
      vmessage("Warning(%s): dp must be a number, using default routine", _function_name);
      dp = -1;
   }%}}}
   variable nbins_pfold = get_qualifier ( find_peak_qualifiers, "nbins_pfold", 100 );%{{{
   ifnot (typeof(nbins_pfold) == Integer_Type or typeof(nbins_pfold) == Double_Type) {
      vmessage("Warning(%s): nbins_pfold has to be a number, setting it to 100", _function_name);
      nbins_pfold = 100.;
   }
   nbins_pfold = nint(nbins_pfold);%}}}
   variable chatty = get_qualifier ( find_peak_qualifiers, "chatty", 1 );%{{{
   try {
      ifnot (chatty == 1 or chatty == 0) {
         vmessage("Warning(%s): chatty must be either 1 or 0, setting it to 1", _function_name);
         chatty = 1;
      }
   } catch AnyError: {
      vmessage("Warning(%s): chatty must be either 1 or 0, setting it to 1", _function_name);
      chatty = 1;
   }%}}}
   variable blocks_but_only_max = get_qualifier ( find_peak_qualifiers, "blocks_but_only_max", 0);%{{{
   ifnot (typeof(blocks_but_only_max) == Integer_Type) {
      vmessage("Warning(%s): blocks_but_only_max can only be 1 or 0, setting it to default", _function_name);
      blocks_but_only_max = 0;
   }
   ifnot (blocks_but_only_max == 0 or blocks_but_only_max == 1) {
      vmessage("Warning(%s): blocks_but_only_max can only be 1 or 0, setting it to default", _function_name);
      blocks_but_only_max = 0;
   }%}}}
   variable no_profile = get_qualifier ( find_peak_qualifiers, "no_profile", 0);%{{{
   ifnot (typeof(no_profile) == Integer_Type) {
      vmessage("Warning(%s): no_profile can only be 1 or 0, setting it to default", _function_name);
      no_profile = 0;
   }
   ifnot (no_profile == 0 or no_profile == 1) {
      vmessage("Warning(%s): no_profile can only be 1 or 0, setting it to default", _function_name);
      no_profile = 0;
   }%}}}
   variable exact = get_qualifier ( find_peak_qualifiers, "exact", 0);%{{{
   ifnot (typeof(exact) == Integer_Type) {
      vmessage("Warning(%s): exact can only be 1 or 0, setting it to default", _function_name);
      exact = 0;
   }
   ifnot (exact == 0 or exact == 1) {
      vmessage("Warning(%s): exact can only be 1 or 0, setting it to default", _function_name);
      exact = 0;
   }%}}}
   variable pfold_not_exact = get_qualifier ( find_peak_qualifiers, "pfold_not_exact", 0);%{{{
   ifnot (typeof(pfold_not_exact) == Integer_Type) {
      vmessage("Warning(%s): pfold_not_exact can only be 1 or 0, setting it to default", _function_name);
      pfold_not_exact = 0;
   }
   ifnot (pfold_not_exact == 0 or pfold_not_exact == 1) {
      vmessage("Warning(%s): pfold_not_exact can only be 1 or 0, setting it to default", _function_name);
      pfold_not_exact = 0;
   }%}}}
   variable only_max = get_qualifier ( find_peak_qualifiers, "only_max", 0);%{{{
   ifnot (typeof(only_max) == Integer_Type) {
      vmessage("Warning(%s): only_max can only be 1 or 0, setting it to default", _function_name);
      only_max = 0;
   }
   ifnot (only_max == 0 or only_max == 1) {
      vmessage("Warning(%s): only_max can only be 1 or 0, setting it to default", _function_name);
      only_max = 0;
   }%}}}
   variable no_fit = get_qualifier ( find_peak_qualifiers, "no_fit", 0);%{{{
   ifnot (typeof(no_fit) == Integer_Type) {
      vmessage("Warning(%s): no_fit can only be 1 or 0, setting it to default", _function_name);
      no_fit = 0;
   }
   ifnot (no_fit == 0 or no_fit == 1) {
      vmessage("Warning(%s): no_fit can only be 1 or 0, setting it to default", _function_name);
      no_fit = 0;
   }%}}}
   variable dp_given = (struct_field_exists(find_peak_qualifiers, "dp")) ? 1 : 0;
   %}}}

   switch (_NARGS)
   { case 1:
      variable input_struct = (); }
   { help(_function_name); return; }
   % Input processing, sanity checks %{{{
   variable no_epfold = 0.;
   try {
      p = input_struct.p;
      stat = input_struct.stat;
      ifnot (typeof(p) == Array_Type and typeof(stat) == Array_Type and _typeof(p) == Double_Type and _typeof(stat) == Double_Type) {
         vmessage("Error(%s): p and stat must be arrays of doubles", _function_name);
         return;
      }
      stat = stat [ array_sort(p) ];
      p = p [ array_sort(p) ];
   } catch AnyError: {
      vmessage("Error(%s): Input structure does not contain fields named p and stat",_function_name);
      return;
   }
   variable lc_provided = 1;
   try {
      lc = input_struct.lc;
      try {
         variable ttest = lc.time;
         variable rtest = lc.rate;
      } catch AnyError: {
         vmessage ("Warning(%s): Lightcurve does not have fields named time and rate, doing no_profile routine", _function_name);
         no_profile = 1;
         lc_provided = 0;
      }
      ifnot (typeof(ttest) == Array_Type and typeof(rtest) == Array_Type) {
         vmessage ("Warning(%s): Lightcurve fields are no arrays, doing no_profile routine", _function_name);
         no_profile = 1;
         lc_provided = 0;
      }
   } catch AnyError: {
      vmessage("Warning(%s): No lightcurve (field named lc) provided", _function_name);
      no_profile = 1;
      lc_provided = 0;
   }
   try {
      nbins_epfold = input_struct.nbins_epfold;
      ifnot (typeof(nbins_epfold) == Integer_Type) {
         vmessage("Warning(%s): nbins_epfold has to be Integer_Type, setting it to 32. Will make errors useless", _function_name);
         nbins_epfold = 32;
         no_epfold = 1;
      }
   } catch AnyError: {
      vmessage("Warning(%s): No field named nbins_epfold provided, setting to 32 (will make err_theory useless)", _function_name);
      nbins_epfold = 32;
      no_epfold = 1.; }
   try {
      expectation = input_struct.expectation;
      ifnot (typeof(expectation) == Double_Type) {
         vmessage("Warning(%s): expectation has to be Double_Type, setting it to the average of p", _function_name);
         expectation = mean(p);
      }
   } catch AnyError: {
      vmessage("Warning(%s): No field named expectation provided, setting it to the average of p", _function_name);
      expectation = mean(p); }
   try {
      sigma_expectation = input_struct.sigma;
      ifnot (typeof(sigma_expectation) == Double_Type) {
         vmessage("Warning(%s): sigma must be Double_Type, setting it to DOUBLE_MAX", _function_name);
         sigma_expectation = DOUBLE_MAX;
      }
   } catch AnyError: {
      vmessage("Warning(%s): No field named sigma provided, setting it to DOUBLE_MAX", _function_name);
      sigma_expectation = DOUBLE_MAX; }

   variable dp_default;
   
   if (lc_provided) {
      if (dp_given == 0) {
         if(chatty){vmessage("(%s): Setting dp to p^2 / T  -- theoretical prediction", _function_name);}
      }
      dp_default = sqr(expectation) / ( max(lc.time) - min(lc.time) );
   } else {
      if (sigma_expectation < DOUBLE_MAX - 1.) {
         if (dp_given == 0) {
            vmessage("Warning(%s): Setting dp to sigma -- because no profile given", _function_name);
         }
         dp_default = sigma_expectation;
      } else {
         if (dp_given == 0) {
            vmessage("Warning(%s): Setting dp to 0.1 * period span -- because no sigma and no profile given", _function_name);
         }
         dp_default = 0.1 * (max(p)-min(p));
      }
   }
   ifnot (typeof(dp) == Double_Type or typeof(dp) == Integer_Type) {
      vmessage("Warning(%s): dp has to be a number, setting it to %lf", _function_name, dp_default);
      dp = dp_default;
   }
   if (dp < 0) { dp = dp_default; }
   dp = double(dp);
   %}}}
   

   variable flags = struct { %{{{  % contains warnings and is used to calculate badness of epochfolding
      end_reached = 0,  % -1 for beginning, 1 for end reached
      no_blocks = 0, % either 0 or 1
      bad_profile = 0,
      bad_blocks = 0, % if no sufficiently small block has been found ( less than ~ 1.5 dp width );
      difference = 0, % measures the difference between the peak found and the prediction - a high value indicates problems
      exact = exact, % 1 if epfold was done with exact qualifier -- currently (as of Sept 16) this is NOT recommended!
      fit_succeeded = 0, % 1 if the fit seems to have been successful
      dp = dp, % measure for the width of the peak
      t = (no_profile==0) ? max(lc.time) - min(lc.time) : 0., % timespan -- short timespans w.r.t. period can pose a problem
   };
   if (no_profile) { flags = reduce_struct(flags, "t"); }
   %}}}

   variable where_noise = where ( stat - median(stat) < 3. * moment(stat - median(stat) ).sdev );
   variable sdev_noise = moment ( stat[where_noise] ).sdev;                                        % finds standard deviation of these values
   variable avg_noise = mean ( stat[where_noise] );                                                % finds mean value of noise

   variable blocks, where_maxima;
   if ( only_max ) {
      if(chatty){vmessage("(%s): Skipping bayesian block analysis as you have chosen", _function_name);}
      blocks = [0];
      where_maxima = [0];
   } else {
% bayesian block analysis, find local maxima
%{{{
   if(chatty){vmessage("(%s): Doing bayesian block analysis", _function_name);}
   variable stat_noisy = @stat; % the original statistics is later used in order to find the peak, we don't want to loose this information by smoothing things out

   variable smoothing_iterations = -1;

   variable bayes_input = struct { tt = p , nn_vec = stat };
   variable bayes_red; 
   do { % while the statistics curve is too noisy for a reasonable block analysis, the curve is smoothed
      if (smoothing_iterations > -0.5) { stat = gliding_average ( p, stat, smoothing ); }
      if (typeof(ncp_prior) == String_Type) {
         bayes_red = bayesian_blocks ( bayes_input; fp_rate=fp_rate ).change_points; % the curve is split into bayesian blocks ( N ~ 10^1 )
      } else {
         bayes_red = bayesian_blocks ( bayes_input; ncp_prior = ncp_prior,fp_rate=fp_rate ).change_points; % the curve is split into bayesian blocks ( N ~ 10^1 )
      }
      variable bayes = Integer_Type[length(bayes_red)+2];
      bayes[0]=0;
      bayes[[1:length(bayes)-2]]=bayes_red;
      bayes[-1]=length(p)-1;                                                                          % 'bayes' contains the boundaries between blocks and the start & end
      blocks = Double_Type[length(bayes)-1];
      variable i;
      _for i (0, length(blocks)-1, 1) {
         blocks[i] = mean ( stat[[bayes[i]:bayes[i+1]]] ); }                                          % 'blocks' contains the avg signal for each bayesian block
      where_maxima = Integer_Type[length(blocks)];
      variable where_minima = Integer_Type[length(blocks)];
      _for i (1, length(where_maxima)-2, 1) {
         if ( blocks[i-1] < blocks[i] and blocks[i+1] < blocks[i] ) { where_maxima[i] = 1; }          % contains indices of blocks that are local maxima
         else { where_maxima[i] = 0;
            if ( blocks[i-1] < blocks[i] and blocks[i+1] < blocks[i] ) { where_minima[i] = 1; }
            else { where_minima[i] = 0; }
              }
      }
      where_maxima[0]=0;
      where_maxima[length(where_maxima)-1]=0;
      smoothing_iterations += 1;
      bayes_input.nn_vec = stat;
   } while ( (length(bayes_red) < 2. or sum(where_maxima)<1.) and smoothing_iterations < 4 );
   flags.no_blocks = smoothing_iterations;

%}}}   
   }

   blocks = blocks * typecast(where_maxima, Double_Type); % now 'blocks' is zero everywhere except for the maxima
   variable conversion = length(p) / ( max(p) - min(p) ); % conversion factor between binning and period grid

   variable bayes_struct;
   if ( only_max ) {
      bayes_struct = empty_struct();
   } else { %{{{
      try {
         bayes_struct = struct {                                                                                                                           
            bin_lo = p[ bayes[[0:length(bayes)-2]] ],                                                                                               
            bin_hi = p[ bayes[[1:length(bayes)-1]] ],                                                                                               
            midpoints = 0.5 * p[ bayes[[0:length(bayes)-2]] ] + 0.5 * p[ bayes[[1:length(bayes)-1]] ],                                              
            blocks = blocks };
      } catch AnyError: {
         bayes_struct = empty_struct();
      }
   } %}}}

% determine quality of individual blocks
%{{{
   if (length(blocks) < 3 or sum(blocks) == 0 ) {
      variable pmin0 = where_max(stat)[0] - nint( conversion * dp );
      variable pmax0 = where_max(stat)[0] + nint( conversion * dp );
      variable first_p0 = where_max(stat)[0];
         if ( pmax0 >= length(p) ) {
               flags.end_reached = 1;
               pmax0 = length(p) - 1;
               pmin0 = nint ( 2*first_p0 - pmax0 ); }
         else {
            if ( pmin0 < 0 ) {
               flags.end_reached = -1;
               pmin0 = 0;
               pmax0 = nint ( 2*first_p0 ); }
              }
   } else {
      variable reached_ends = 0*ones(length(blocks));
      variable imps = Double_Type[length(blocks)];  % 'imps' contains the importance of a peak
      variable pmin = Integer_Type[length(blocks)];
      variable pmax = Integer_Type[length(blocks)];
      _for i (0, length(blocks)-1, 1) {
         if ( blocks[i]< 1e-5 ) { imps[i] = -DOUBLE_MAX; }
         else {
            variable midpoint = mean([bayes[i],bayes[i+1]]);
            pmin[i] = nint( midpoint - dp*conversion ); % a neighbourhood of +- p^2/T is defined
            pmax[i] = nint( midpoint + dp*conversion );
            if ( pmax[i] >= length(p) ) { % a flag is set if the possible peak is too close to a boundary of the period interval
                  reached_ends[i] = 1;                  
                  pmax[i] = length(p) - 1;
                  pmin[i] = max ( [ 0, nint ( 2*mean([bayes[i],bayes[i+1]]) - pmax[i] ) ] ); }
            else {
               if ( pmin[i] < 0 ) {
                  reached_ends[i] = -1;
                  pmin[i] = 0;
                  pmax[i] = min ( [ length(p)-1, nint ( 2*mean([bayes[i],bayes[i+1]]) ) ] ); }
                 }
   
            imps[i] = (sum(stat[[nint(0.5*pmin[i]+0.5*midpoint):nint(0.5*pmax[i]+0.5*midpoint)]]))^weighting[0] * (log(log1p(3.*dp/(p[bayes[i+1]] - p[bayes[i]]))))^weighting[1];
            
            if ( expectation>1e-10 and sigma_expectation>1e-10 ) {
               imps[i] *= (quality_function(expectation, sigma_expectation, p[nint(midpoint)]))^weighting[2]; }
         }
            % sharp features are slightly advantageous, but very broad features are highly disadvantageous
            % but a big area in the +- p^2/T interval is favoured
         }
%}}}

   variable possible_blocks = where(imps > (tolerance * max(imps) ) ); % usually only the true peak should fall into this category
   pmin0 = pmin[ possible_blocks ]; % but sometimes multiple similar maxima are found
   pmax0 = pmax[ possible_blocks ];
   if (length(possible_blocks)==0) { % this is needed if all imps are negative -- should not happen for sufficiently good data
      flags.bad_blocks = 1;
      possible_blocks = where_max(imps);
      pmin0 = pmin[ possible_blocks ];
      pmax0 = pmax[ possible_blocks ];
      }
   }

% determine the period for each possible peak -- normally just one iteration
%{{{
   variable v;
   variable smoothness = Double_Type [ length(pmin0) ]; % used to distinguish between various possible true peaks -- it measures the quality of the folded pulse profile
   variable periods = Double_Type [ length(pmin0) ];
   variable folded_profiles = {}; % contains the posstible profiles -- the best one is passed to output
   variable cum_areas = {}; % contains the cumulated area under a peak dependent on period
   variable fitsuccess = Integer_Type [ length(pmin0) ];
   variable fit_outcomes = Struct_Type [ length(pmin0) ];
   variable error_areas = Double_Type [ length(pmin0) ];

   if ( length(pmin0) > 1 ) {
      if(chatty){vmessage("(%s): Found %d possible peaks", _function_name, length(pmin0) );}
   }
   _for v ( 0, length(pmin0) - 1, 1 ) { 
      %preliminaries %{{{
      if (length(blocks)>2 and sum(blocks)!=0) { first_p0 = where_max ( stat[[pmin0[v]:pmax0[v]]] )[0] + pmin0[v]; }
      else { first_p0 = first_p0; }
      variable p_half_lo = ( where_min ( abs(stat[[pmin0[v]:first_p0]] - avg_noise - 0.5 * ( max(stat[[pmin0[v]:pmax0[v]]]) - avg_noise ) ) ) + pmin0[v] )[0]; % index
      variable p_half_hi = ( where_min ( abs(stat[[first_p0:pmax0[v]]] - avg_noise - 0.5 * ( max(stat[[pmin0[v]:pmax0[v]]]) - avg_noise ) ) ) + first_p0 )[0]; % index
      variable grid = p[[pmin0[v]:pmax0[v]]]; % the period grid for the individual possible peak within +- p^2/T
      variable peak_data = stat[[pmin0[v]:pmax0[v]]]; % chi^2 on the grid
      variable cum_area = Double_Type[length(grid)];
   
      cum_area[0] = ( grid[0] - p[pmin0[v]-1] ) * peak_data[0];
      _for i (1, length(cum_area)-1, 1) {
         cum_area[i] = cum_area[i-1] + ( grid[i] - grid[i-1] ) * peak_data[i]; }
      list_append ( cum_areas, cum_area );

      variable period = p[ where_min( abs(cum_area[[p_half_lo-pmin0[v]:p_half_hi-pmin0[v]]] - 0.5 * cum_area[p_half_hi-pmin0[v]] ) )[0] + p_half_lo ];
      % this approach is currently not used
      variable period_max = p[ where_max( stat[[pmin0[v]:pmax0[v]]] )[0] + pmin0[v] ];
      %}}}

      if (length(blocks) < 3 or sum(blocks) == 0) {
         period = period_max; % just the maximum in the statistic is taken
         fitsuccess [ v ] = 0; 
         fit_outcomes [ v ] = empty_struct();
      }
      else { 
      % find first period estimate %{{{
         variable peak_slice = [bayes[possible_blocks[v]]:bayes[possible_blocks[v] + 1]]; % index interval between the bayesian block boundaries of the possible peak
         variable n_important;
         n_important = max ( [ min( [ 3, length(peak_slice) ] ), nint(length(peak_slice) / 10.) ] ); % the number of points in 'peak slice' that is used
         variable important_indices = [peak_slice [ array_sort(stat_noisy[peak_slice])[[-n_important:]] ] ]; % corresponding periods, sorted by their statistic
         variable important_periods = p [ important_indices ];
         variable u;
         period = 0;
         variable u_sum = 0;
         _for u ( 0, n_important - 1, 1) {
            period += important_periods[u]/(u+1)^0.5; % a weighted average of the 'important_periods' is taken
            % only the index and not the statistic is taken as weight in order to reduce the impact of unexpectedly high statistics
            u_sum += 1/(u+1)^0.5;
         }
         period /= u_sum;
      %}}}
         ifnot (no_fit or blocks_but_only_max) {
         % find fit area %{{{
         variable period_index = where_min (abs(period-p[peak_slice]) ) + min(peak_slice);
         error_areas [v] = error_estimate ( p, stat, period, 0.25*dp, 0.25*dp );
         variable lo_boundary = where_min ( abs(p - period + fit_part*dp) );
         variable hi_boundary = where_min ( abs(p - period - fit_part*dp) );
         %}}}

         % perform fit %{{{
         variable xdata = p [[lo_boundary:hi_boundary]];
         variable ydata = stat_noisy [[lo_boundary:hi_boundary]]; % although the smoothed curve may look nicer, we don't want to loose information for the fit.
         variable w_dist = (abs(xdata - period))^0.25;
         variable id = define_xydata ( xdata, ydata, 0.1*(max(ydata)-min(ydata))*w_dist/(max(w_dist)) );
         try{
            xyfit_fun ( fit_fct );
            set_par ( 1, period );
            set_par ( 2, max(ydata) );
            set_par ( 3, 0.5 * (max(xdata) - min(xdata)) );
            set_par ( 4, min(ydata) );
            set_par ( 5, 0 );
            if (chatty) { vmessage ("(%s): Performing fit of possible peaks, using %s", _function_name, fit_fct); }
         } catch AnyError: {
            vmessage("Warning(%s): Fit function not suitable, using sqrsinc instead",_function_name);
            xyfit_fun ( "sqrsinc" );
            set_par ( 1, period );
            set_par ( 2, max(ydata) );
            set_par ( 3, 0.5 * (max(xdata) - min(xdata)) );
            set_par ( 4, min(ydata) );
            set_par ( 5, 0 );
         }
         variable a = fit_counts(; fit_verbose=-1);
         variable model;
         ( , model ) = get_xymodel ( id );
         fit_outcomes [v] = struct { xdata = xdata, ydata = ydata, ymodel = model };
         variable fit_period = get_par(1);
         delete_data (all_data);
         if (abs(fit_period - period) < 1.*min( [error_areas[v], sigma_expectation] ) and fit_period>min(p) and fit_period<max(p)) {
            period = @fit_period;
            fitsuccess[v] = 1;
         } else {
            fitsuccess[v] = 0;
         }
         %}}}
         } else {
            fit_outcomes[v] = empty_struct();
            if (v==0) { if(chatty){vmessage("(%s): No fit is performed as you have chosen", _function_name); }}
         }
      }

      ifnot (no_profile) {
         variable dt = lc.time - shift(lc.time, -1);
         dt[0] = median( dt[[1:length(dt)-1]] );
                                                                                                                                                           
         variable ix = 0;
         variable factor = 4.;
         variable folded_profile;
         if (nbins_pfold == 100 ) {
         while (ix==0) { %{{{
            if ( pfold_not_exact ) {
               folded_profile = pfold ( lc.time, lc.rate, period;nbins=min([100,nint(factor*period/0.125)]),dt=dt );
            } else {
               folded_profile = pfold ( lc.time, lc.rate, period;nbins=min([100,nint(factor*period/0.125)]),dt=dt,exact );
            }
            if (abs(mean(folded_profile.value)) > 0) { % only possible if no 'nan' is in the profile
               ix = 1;
            } else {
               factor *= 0.90;
               ix = 0;
            }
         } %}}}
         } else {
         while (ix==0) { %{{{
            if ( pfold_not_exact ) {
               folded_profile = pfold ( lc.time, lc.rate, period;nbins=nbins_pfold,dt=dt );
            } else {
               folded_profile = pfold ( lc.time, lc.rate, period;nbins=nbins_pfold,dt=dt,exact );
            }
            if (abs(mean(folded_profile.value)) > 0) { % only possible if no 'nan' is in the profile
               ix = 1;
            } else {
               vmessage("error(%s): the nbins_pfold you provided oversampled the pulse profile too much", _function_name);
               return;
            }
         } %}}}
         }
         list_append ( folded_profiles, folded_profile );
                                                                                                                                                           
         % estimate quality of profile %{{{
         variable j;
         variable jump = Double_Type [length(folded_profile.value)]; % contains the absolute differences between the flux in neighbouring phase bins
         variable random_jump = Double_Type [length(folded_profile.value)]; % contains the abs diff between the flux in two random phase bins
         _for j (0, length(jump)-1, 1) {
            jump[j] = abs(shift(folded_profile.value,j)[0] - shift(folded_profile.value,j+1)[0]);
            variable jj;
            variable all_jumps = Double_Type [length(random_jump)];
            _for jj (0, length(jump)-1, 1) {
               all_jumps[jj] = abs(shift(folded_profile.value,j)[0] - shift(folded_profile.value,jj)[0]); }
               random_jump[j] = mean(all_jumps);
            }
         if ( abs(mean(jump)) > 1e-5 ) { smoothness[v] = mean(random_jump) / mean(jump); } % the higher this ratio the less noise is in the folded profile
         else { smoothness[v] = mean(random_jump) * 1e10; } 
         %}}}
      }
      periods[v] = period;
   }
%}}}

% choose true peak and determine errors, set flags
%{{{
   if (length(blocks) < 3 or sum(blocks)==0) {
      imps = ones(1);
      possible_blocks = 0; }
   
   if (no_profile) { smoothness = 1.; }

   variable here_peak =  where_max( smoothness^weighting[3] * imps[possible_blocks] )[0]; % the index of the most probable maximum in the previously generated lists
   period = periods [ here_peak ];
   flags.fit_succeeded = fitsuccess[ here_peak ];
   variable fit_outcome = fit_outcomes[ here_peak ];
   variable out_profile;
   if (no_profile) {
      out_profile = empty_struct();
   } else {
      out_profile = folded_profiles [ here_peak ];
   }
   pmin0 = pmin0 [ here_peak ];
   pmax0 = pmax0 [ here_peak ];  
   variable where_period = where_min( abs(p - period) )[0];
   variable cum_a = cum_areas [ here_peak ];
   variable error68_lo = period - p[ where_min( abs(cum_a[[:where_period-pmin0]] - 0.32 * cum_a[where_period-pmin0] ) )[0] + pmin0];
   variable error68_hi = p[ where_min( abs(cum_a[[where_period-pmin0:]] - 0.32 * cum_a[where_period-pmin0] - 0.68 * cum_a[-1] ) )[0] + where_period] - period;
   variable error_area = error_estimate ( p, stat, period, error68_lo, error68_hi );
   variable correction_offset = max ( [ 0., avg_noise - nbins_epfold + 1 ] ); % the difference between the expected noise of (nbins-1) and the noise found in the statvsp
   variable error_theory;
   %% Leahy 1987: sigma(p) / delta(P) = 0.71 * (chired^2 - 1 )^-0.63 ( Monte Carlo simulation ) where delta(P)=P^2 / 2T
   %% dp = p^2/T is the half width of the peak
   %% chired^2 = chi^2 / (n-1) -- chi^2 of peak {i.e. max(stat)}, n is number of bins in folded lightcurve (nbins qualifier of epfold)
   error_theory = 0.5*dp*0.71*( (max(stat[[pmin0:pmax0]] - correction_offset)/(nbins_epfold - 1)) - 1 )^(-0.63); % the error estimate proposed by Leahy 1987
   ifnot ( error_theory < 1. or error_theory > 1. or error_theory == 1. ) { % if it is nan
      error_theory = error_area;
   }
   flags.bad_profile = (no_profile==0) ? nint ( log(length(out_profile.value))/(log(10)*max( smoothness )) ) : 0; % a flag set if even the best profile seems to be pure noise
   
   if ( expectation>1e-10 and sigma_expectation>1e-10 ) {
      flags.difference = nint(abs(period-expectation)/(3.*sigma_expectation));
   }
%}}}

   variable badness = nint( (1-flags.fit_succeeded) + abs(flags.end_reached) + flags.no_blocks + flags.difference + flags.bad_profile + flags.bad_blocks);
   if (no_fit or only_max or blocks_but_only_max) { badness += -1; }
   % combines all flags

variable out;
out = struct { %{{{
   period = period,                                                                                                                           
   error = sqrt ( sqr(error_area) + sqr(error_theory) ),                                                                                      
   profile = out_profile,                                                                                                                     
   err_area = error_area,                                                                                                                     
   err_theory = error_theory,                                                                                                                 
   % it has been estimated for sinusoidal signals, but its application in this case shows that the width of the noisy peak is nearly always   
   % excellently described by this error estimate -- hence we conclude it to be a viable way of giving limits to our period                   
   %width_lo = error68_lo,                                                                                                                    
   %width_hi = error68_hi,                                                                                                                    
   statvsp = struct {                                                                                                                         
      p= p,                                                                                                                                   
      stat = stat_noisy },                                                                                                                          
   flags = flags,                                                                                                                             
   bayes = bayes_struct,
   fit = fit_outcome,                                                                                                                         
   badness = badness }; %}}}                                                                                                                  
if ( length(get_struct_field_names(out.profile)) == 0 ) { out = reduce_struct(out, "profile"); }
if ( length(get_struct_field_names(out.fit)) == 0 ) { out = reduce_struct(out, "fit"); }
if ( length(get_struct_field_names(out.bayes)) == 0 ) { out = reduce_struct(out, "bayes"); }

   return out;
}
%}}}

%
% Code from src/slang/math/first_valid_digit.sl
%
define first_valid_digit( x )
%!%+
%\function{first_valid_digit}
%\synopsis{gives the position of the first valid digit}
%\usage{Integer_Type fvd = first_valid_digit( Integer/Double_Type x )};
%!%-
{
  variable absx = abs(x);

  if( absx == 0 ){
    return 0;
  }
  else if( absx < 1 ){
    return int(floor(log10(absx)));
  }
  else if( absx == 1 ){
    return 1;
  }
  else{
    return int(ceil(log10(absx)*(1+DOUBLE_EPSILON)));
  }
}

%
% Code from src/fitting/fit_brute_force.sl
%
%%%%%%%%%%%%%%%%%%%%%
define fit_brute_force()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fit_brute_force}
%\synopsis{Performs a fit by stepping the given parameters and does a
%    usual fit for the remaining ones}
%
%\usage{Struct_Type fit_brute_force(Integer_Type[] par [, Double_Type[] stepsize])
% or Struct_Type fit_brute_force(String_Type[] par [, Double_Type[] stepsize])}
%\qualifiers{
%\qualifier{nofit}{instead of fitting the remaining parameters the
%             model is just evaluated}
%\qualifier{nomap}{the chi-square map will not be created}
%\qualifier{chatty}{boolean value to show the remaining time
%              (default=1)}
%}
%\description
%    For the given parameters the best fit of the actual model
%    is found by going through the complete parameter range.
%    The parameters can be passed as an array containing their
%    indices or names. The stepsize for each parameter can
%    either be given by the second `stepsize' parameter or the
%    intrinsic stepsizes are used, which can be set using a
%    qualifier of the `set_par' function. The latter can also
%    be used to determine the minimum and maximum value of each
%    parameter. While the given parameters are stepped the other
%    parameters of the model are fitted by the actual fitting
%    method, that means the function `fit_counts' is called.
%    If the qualifier `nofit' is given `eval_counts' is used
%    instead, evaluating the model without performing a chi-
%    square minimization.
%    This function is very similar to the `get_confmap' function.
%    But this function steps n given parameters in a very small
%    grid. In contrast it is possible to perform a fit for all
%    parameters of the model. However BE CAREFUL as the runtime
%    rises exponentially with the number of parameters!! For
%    this reason the function is chatty by default to show the
%    estimated remaining time during the fit.
%    The returned structure contains an array `bestpar' of the
%    best values found for each given parameter, the corres-
%    ponding reduced chi-square value `bestchi' and a structure
%    `chimap', which holds the chi-square values for each tested
%    parameter combination up to two given parameters. Because
%    this map may be very big its creation can be suppressed by
%    the `nomap' qualifier. In the structure the one or two
%    dimensional array (depending on the number of parameters)
%    `chisqr' contains the reduced chi-square values. In case of
%    two dimensions the first one specifies the second parameter
%    to directly pass it to image function like `plot_image'. The
%    corresponding parameter values are stored in the x-field for
%    the first parameters and y-field for the second one.
%\seealso{get_confmap, fit_counts, eval_counts, set_par, plot_image}
%!%-
{
 variable par, inc = NULL;
 switch(_NARGS)
   { case 1: (par) = (); }
   { case 2: (par, inc) = (); }
   { help(_function_name()); return; }
 if (typeof(par)!=Array_Type) par=[par];

 variable i, j, st, dst, ok, stat, chi, tact, tstart, freep, bestfreep;
 variable n = length(par);
 ifnot (inc == NULL || length(inc) == n) throw RunTimeError, sprintf("error (%s): length of step size parameter must be equal number of parameters",_function_name);
 if (inc == NULL) inc = 0.*ones(n);
 variable mn = Double_Type[n];
 variable mx = Double_Type[n];
 variable fr = Double_Type[n];
 variable bestchi = DOUBLE_MAX, bestpar = Double_Type[n];

 % get minimum, maximum and step size of each parameter
 _for i (0, n-1, 1)
 {
  mn[i] = get_par_info(par[i]).min;
  mx[i] = get_par_info(par[i]).max;
  if (inc[i] == 0.0) inc[i] = get_par_info(par[i]).step;
  if (inc[i] == 0.0)
  {
    vmessage("warning (%s): step size of parameter %d is zero, using 100 steps", _function_name, par[i]);
    inc[i] = 1.0/99.*(mx[i]-mn[i]);
  }
  fr[i] = get_par_info(par[i]).freeze;
 }
 variable chatty = qualifier("chatty",1);
 % number of steps for each parameter
 variable num = int(round((mx - mn) / inc)) + 1;
 variable tnum = int(prod(num)); % total number of steps
 % initialise confidence map, if number of parameters smaller 3
 variable chimap;
 if (n>2 or qualifier_exists("nomap")) chimap = NULL;
 else if (n==2) chimap = struct { x = Double_Type[num[0]], y = Double_Type[num[1]], chisqr = Double_Type[num[1],num[0]] };
 else chimap = struct { x = Double_Type[num[0]], chisqr = Double_Type[num[0]] };
 if (chatty)
 {
  vmessage("Number of steps: %d", tnum);
  message("Estimating remaining time roughly,");
  message("which will take at least one minute...");
 }
 st = 0; % total step number
 % freeze parameters
 freeze(par);
 % set initial parameters
 _for i (0, n-1, 1) set_par(par[i], mn[i]);
 % get initial values of free parameters
 freep = Double_Type[length(freeParameters),2];
 freep[*,0] = freeParameters;
 bestfreep = Double_Type[length(freeParameters)];
 _for i (0, length(freep[*,0])-1, 1) freep[i,1] = get_par(int(freep[i,0]));
 variable nofit = (qualifier_exists("nofit") || length(freep[*,0]) == 0);

 % main loop until number of steps is reached
 i = ones(n)*0; % actual step number of each parameter
 tstart = _time; dst=1; % initialise timing
 while (st < tnum)
 {
  % set free parameters
  _for j (0, length(freep[*,0])-1, 1) set_par(int(freep[j,0]),freep[j,1]);
  % fit free parameters
  if (nofit) ()=eval_counts(&stat; fit_verbose=-2);
  else ()=fit_counts(&stat; fit_verbose=-2);
  % compare actual chi square value with best value
  chi = stat.statistic / (stat.num_bins - stat.num_variable_params);
  if (chi<bestchi)
  {
   bestchi = chi;
   _for j (0, n-1, 1) bestpar[j] = get_par_info(par[j]).value;
   _for j (0, length(bestfreep)-1, 1) bestfreep[j] = get_par_info(int(freep[j,0])).value;
  }
  % store chi square value
  if (typeof(chimap) != Null_Type)
  {
   if (n==2) { chimap.chisqr[i[1],i[0]] = chi; chimap.x[i[0]] = get_par(par[0]); chimap.y[i[1]] = get_par(par[1]); }
   else { chimap.chisqr[i[0]] = chi; chimap.x[i[0]] = get_par(par[0]); }
  }

  % increase step number recursivly and set new parameter values
  j = 0; ok = 1;
  do
  {
   if (ok==0) j++;
   i[j]++;
   if (i[j]==num[j] and j<n-1) { i[j] = 0; ok = 0; }
   else ok = 1;
   if (i[j]==num[j]-1) set_par(par[j], mx[j]);
   else if (i[j]<num[j]-1) set_par(par[j], mn[j] + i[j]*inc[j]);
  }
  while (ok==0);

  st++;
  % determine remaining time
  if (chatty)
  {
   tact = _time; % actual time
   if (dst > 0)
   {
    if (tact-tstart>60 && st mod dst==0)
    {
      tact = (tact - tstart)*(tnum-st)/st;
      vmessage("  ...remaining time: %02d:%02d:%02d", tact/3600, tact mod 3600 / 60, tact mod 60);
      if (dst==1) dst = (tnum-st)/9;
    }
   }
  }
 }

 if (chatty)
 {
  tact = _time - tstart;
  vmessage("Total runtime: %02d:%02d:%02d", tact/3600, tact mod 3600 / 60, tact mod 60);
  vmessage("Best chi^2 = %.2f", bestchi);
 }
 % set best values
 _for i (0, n-1, 1) set_par(par[i], bestpar[i]);
 _for i (0, length(bestfreep)-1, 1) set_par(int(freep[i,0]), bestfreep[i]);
 % evaluate
 ()=eval_counts(; fit_verbose=-2);
 % restore freeze status of parameters
 _for i (0, n-1, 1) if (fr[i]==0) thaw(par[i]);

 return struct { bestpar = bestpar, bestchi = bestchi, chimap = chimap };
}

%
% Code from src/misc/image/fit_gauss_to_img_noise.sl
%
define fit_gauss_to_img_noise ()
%!%+
%\function{fit_gauss_to_img_noise}
%\synopsis{fits a gaussian profile to the distribution of pixel values}
%\usage{mu,sigma = fit_gauss_to_img_noise(\code{img})}
%\qualifiers{
%\qualifier{grid_scale [="log"]}{change between fitting on "lin" or "log" grid}
%\qualifier{cut_nsig [=NULL]}{set to \code{N} in order to ignore values above
%                           \code{N} sigma after the first iteration}
%\qualifier{keep_data}{do not delete loaded data after fitting}
%}
%\description
%    This function fits a Gaussian profile to the distribution of
%    pixel values in an image (arrays of different dimensions can be
%    given) and returns the center \code{mu} and the width \code{sigma} of the
%    best fit profile.
%    If the majority of pixels in the image include only noise values,
%    the obtained \code{mu} and \code{sigma} values characterize the mean value
%    (base level) of the noise and its amplitude.
%
%    NOTE: If the data is kept and further fitting is performed, the following
%    relation has to be considered \code{mu=get_par("gauss(1).center")+min(img)-1e-5}
%    (necessary to provide a proper grid for fitting).
%\example
%    img = grand(500*500);    % image with random numbers around 0 with sigma=1
%    reshape(img,[500,500]);
%    (mu,sigma) = fit_gauss_to_img_noise (img);
%\seealso{plot_vlbi_map}
%!%-
{
  switch(_NARGS)
  { case 1:   variable map = (); }
  { help(_function_name()); return; }

  variable ad = all_data;
  variable ff = get_fit_fun();
  variable fs = get_fit_statistic();
  set_fit_statistic("chisqr");
  exclude(all_data);
  variable lo,hi;
  variable map_min = min(map);
  variable map_max = max(map);
  variable len = length(map);
  variable grid_len   = qualifier("grid_len",nint(sqrt(len)));
  variable grid_scale = qualifier("grid_scale","log");
  if (grid_scale == "lin")
    (lo,hi) = linear_grid( 1e-5 , map_max+(1e-5-map_min) , grid_len);
  else
    (lo,hi) = log_grid( 1e-5 , map_max+(1e-5-map_min) , grid_len);
  variable nr = histogram ( map-map_min + 1e-5 , lo,hi);
  variable px_dat = define_counts(lo , hi , nr , (nr)^(0.5)+1);
  fit_fun("gauss");
  set_par("gauss(1).center", median( map[where(isnan(map)==0)]-map_min + 1e-5) , 0 , -hi[-1], 5*hi[-1] ); % assuming mu is not far from value range
  set_par("gauss(1).area"  , 0.99*len                , 0 , 1   , 2*len);
  set_par("gauss(1).sigma" , quantile (0.84, _reshape(map[where(isnan(map)==0)]-map_min+1e-5,length(dup)))
	  - get_par("gauss(1).center") , 0 , 1e-10, hi[-1] ); % assuming sigma is smaller than value range
  set_fit_method("subplex");
  ()=fit_counts;
  set_fit_method("mpfit");
  variable cut_nsig = qualifier ("cut_nsig",NULL);
  if (cut_nsig != NULL ) {ignore (px_dat, get_par("gauss(1).center")+cut_nsig*get_par("gauss(1).sigma"));}
  ()=fit_counts;
  variable mu  = get_par("gauss(1).center") + map_min - 1e-5 ;
  variable sig = get_par("gauss(1).sigma");
  ifnot (qualifier_exists("keep_data"))
  {
    delete_data(px_dat);
    if (ad != NULL) include(ad);
    if (ff != NULL) fit_fun(ff);
    set_fit_statistic (fs);
  }
  return mu , sig;
}

%
% Code from src/fitting/fit_interactive.sl
%
private define get_params_and_stepsize()
{
  variable params = get_params();
  if(params==NULL)
    return Struct_Type[0], Integer_Type[0];

  variable stepsize = array_struct_field(params, "step");
  stepsize[where(stepsize==0)] = 1e-3;
  return params, nint(log10(stepsize));
}

%%%%%%%%%%%%%%%%%%%%%
define fit_interactive()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fit_interactive}
%\synopsis{Interactivly change parameters and evaluate the model}
%
%\usage{fit_interactive(Ref_Type plotfunction[, Any_Type arg0, arg1, ...]}
%\description
%    The Ref_Type 'plotfunction' is a reference to
%    the plot function, which has to be used (see
%    example below). Any additional arguments are
%    passed to this function as well as given
%    qualifiers.
%    The user may changes the parameters interactivly
%    by the following keys:
%      LEFT/RIGHT: increase/decrease value of active
%                  parameter by one stepsize
%      UP/DOWN   : change active parameter
%      PGUP/PGDN : increase/decrease stepsize by one
%                  order
%      t         : freeze/thaw active parameter
%      r         : change range of active parameter
%      v         : set value of active parameter
%      z         : set the internal step size used by
%                  fit_counts
%      m         : modify fit function
%      f         : perform a fit by running fit_counts
%      s         : show/hide frozen parameter
%      p/P       : load/save parameters in 'parfile'
%      q         : quit interactive mode
%    Every time the value of a parameter is changed,
%    the model is evaluated automatically and the
%    given plot function is called.
%    Actual parameter values and the reduced chi-
%    square are printed out to the plot window.
%    The active parameter is shown in red, thawed
%    ones in black and frozen ones in gray. If the
%    parameter turns blue, one of the borders of the
%    range is reached.
%\qualifiers{
%    \qualifier{parfile}{filename, which is used to save or load
%              the fit parameters using 'save_par' or
%              'load_par', respectively. The user will
%              be asked for if 'parfile' is not set.}
%    \qualifier{plotscript}{filename, used to load in a user-defined 
%                 plotting script. If not set, function
%                 reverts to Ref_Type 'plotfunction'.}
%}
%\example
%    variable id = load_data("example.pha");
%    fit_fun("cutoffpl");
%    fit_interactive(&plot_data, id; res=1);
%
%    The appearing plot window then is internaly
%    called by
%    plot_data(id; res=1);
%\seealso{eval_counts, open_plot, keyinput}
%!%-
{
 if (_NARGS < 1) { help(_function_name()); return; }

 variable args = __pop_list(_NARGS-1); % get given parameters
 variable plotfun = (); % get given plot function
 variable params, stepsize; (params, stepsize) = get_params_and_stepsize();
 variable actpar = 0, stat, i, j, showf = 1, isfrz, update = 1, istart = 0, iend = min([4, length(params)-1]), y0;
 variable parfile = qualifier("parfile", "");
 variable plotscript = qualifier("plotscript", "");

 % prepare plot window
 (@plotfun)(__push_list(args) ;; __qualifiers);
 variable vs_w, vs_a;
 (,vs_w,,vs_a) = _pgqvsz(1); vs_a /= vs_w; % current window dimensions (view surface), width (inch) and aspect
 variable pwin = open_plot;
 variable nvs_w = 6., nvs_a = .31; % new window dimensions
 variable chz = 2.8; % default char size
 _pgscr(16, 0.8, 0.8, 0.8);

 % MAIN LOOP
 forever
 {
   % eval model
   if (update == 1)
   {
     % check on min/max values
     _for i (0, length(params)-1, 1)
     {
       params[i].value = _max(params[i].min, params[i].value);
       params[i].value = _min(params[i].max, params[i].value);
     }
     set_params(params);
     ()=eval_counts(&stat; fit_verbose=-1);
   }

   % parameter plot
   window(pwin);
   xlin; ylin;
   _pgpap(nvs_w, nvs_a); % set view surface dimensions
   color(1);
   xrange(0,1); yrange(0,1); label("", "", ""); erase();
   charsize(1.4*chz); xylabel(0.5,1.02,"Interactive Fit",0,0.5);
   charsize(.8*chz); xylabel(1.02,-0.1,"press 'h' for help",0,1);
   charsize(chz); xylabel(-.02, .88, sprintf(`\gx\u2\d\d\bred\u = %.1f / %d = %.2f`,
					     stat.statistic, stat.num_bins - stat.num_variable_params,
					     stat.statistic/(stat.num_bins - stat.num_variable_params)
					    ));
   xylabel(-.02, .76, sprintf("%s %d to %d:", (showf ? "Parameter" : "Free parameter"), istart+1, iend+1));
   j = 0;
   _for i (istart, iend, 1)
   {
       isfrz = (params[i].freeze || params[i].tie != NULL || params[i].fun != NULL);  % frozen parameter?
       % set color
       variable fcol = (i == actpar) ? 2 : (isfrz ? 16 : 1);
       % print parameter
       if (isfrz == 0 || showf)
       {
	 y0 = .64 - .12*(i-j-istart);
	 color(fcol);
         xylabel(.0, y0, params[i].name);
	 if (params[i].value == params[i].max || params[i].value == params[i].min) color(4);
         xylabel(.3, y0, sprintf("= %S", params[i].value));
	 color(fcol);
         xylabel(.8, y0, sprintf(`stepsize = 10\u%d\d`, stepsize[i]));
       }
       else j++;
   }
   charsize(1);
   % model plot
   if (update)
   {
     window(1);
     _pgpap(vs_w, vs_a); % set view surface dimensions
    if (plotscript == "")
    {
      (@plotfun)(__push_list(args) ;; __qualifiers);
    }
    else
    {
      if (access(plotscript, F_OK) == 0)
      {
        evalfile(plotscript);
        print("Plotting script successfully loaded...");
      }
      else
      {
        vmessage("error: file '%s' does not exist", plotscript);
        (@plotfun)(__push_list(args) ;; __qualifiers);
      }
    } 
      update = 0;
   }

   % user input
   variable newpar, f;
   switch (keyinput(;nchr=1, silent))
     % increase parameter value
     { case "RIGHT_ARROW": if (length(params) > 0) params[actpar].value += 10.^stepsize[actpar]; update = 1; }
     % decrease parameter value
     { case "LEFT_ARROW": if (length(params) > 0) params[actpar].value -= 10.^stepsize[actpar]; update = 1; }
     % select parameter above
     { case "UP_ARROW":
       if (length(params) > 0 && actpar > 0)
       {
	 newpar = actpar;
	 do % check on freeze state
	   newpar--;
         while (showf == 0 && newpar > 0 && (params[newpar].freeze || params[newpar].tie != NULL || params[newpar].fun != NULL));
         if (showf || (params[newpar].freeze == 0 && params[newpar].tie == NULL && params[newpar].fun == NULL)) actpar = newpar;
       }
       if (actpar < istart) % update index of shown parameter
       {
	 istart = actpar; iend = istart; f = [0,0];
	 if (showf) iend += 4;
	 else
	 {
	   do
	   {
	     iend++;
             ifnot(   iend >= length(params)-1
		   || params[iend].freeze || params[iend].tie != NULL || params[iend].fun != NULL)
	       f = [f[0]+1, iend];
	   } while (iend < length(params)-1 && f[0] < 4);
           if (f[1] == 0) f[1] = istart;
           if (f[0] < 4) iend = f[1];
	 }
	 if (iend > length(params)-1) iend = length(params)-1;
       }
     }
     % select parameter below
     { case "DOWN_ARROW":
       if (length(params) > 0 && actpar < length(params)-1)
       {
	 newpar = actpar;
	 do % check on freeze state
	   newpar++;
	 while (showf == 0 && newpar < length(params)-1 && (params[newpar].freeze || params[newpar].tie != NULL || params[newpar].fun != NULL));
         if (showf || (params[newpar].freeze == 0 && params[newpar].tie == NULL && params[newpar].fun == NULL)) actpar = newpar;
       }
       if (actpar > iend) % update index of shown parameter
       {
	 iend = actpar; istart = iend; j = 0;
	 if (showf) istart -= 4;
	 else
	   do
	   {
	     istart--;
	     ifnot (   istart <= 0
		    || params[istart].freeze || params[istart].tie != NULL || params[istart].fun != NULL)
	       j++;
	   } while (istart > 0 && j < 4);
         if (istart < 0) istart = 0;
       }
     }
     % increase parameter stepsize
     { case "PAGE_UP": if (length(params)) stepsize[actpar]++; }
     % decrease parameter stepsize
     { case "PAGE_DOWN": if (length(params)) stepsize[actpar]--; }
     % switch freeze state of parameter
     { case "t": params[actpar].freeze = not params[actpar].freeze; set_params(params); }
     % change range of parameter
     { case "r": window(pwin);
       vmessage("parameter range of %s:", params[actpar].name);
       erase; charsize(1.4*chz); color(1); xylabel(0.5, 0.5, "enter new parameter range into console", 0, 0.5);
       variable newmin = atof(keyinput(; prompt = "  min = ", default = sprintf("%S",params[actpar].min)));
       variable newmax = atof(keyinput(; prompt = "  max = ", default = sprintf("%S",params[actpar].max)));
       if (newmin > newmax) message("error: minimum value larger than maximum value");
       else if (newmin < params[actpar].value < newmax) { params[actpar].min = newmin; params[actpar].max = newmax; set_params(params); }
       else vmessage("error: actual parameter value %S not enclosed by given range", params[actpar].value);
     }
     % set parameter value explicitely
     { case "v": window(pwin);
       vmessage("value of %s:", params[actpar].name);
       erase; charsize(1.4*chz); color(1); xylabel(0.5, 0.5, "enter new parameter value into console", 0, 0.5);
       variable newval = atof(keyinput(; prompt = " value = ", default = sprintf("%S",params[actpar].value)));
       if (params[actpar].min <= newval <= params[actpar].max) { params[actpar].value = newval; set_params(params); update = 1; }
       else vmessage("error: given parameter value not enclosed by given range (%S to %S)", params[actpar].min, params[actpar].max);
     }
     % set fitting stepsize
     { case "z": window(pwin);
       vmessage("fitting stepsize of %s:", params[actpar].name);
       erase; charsize(1.4*chz); color(1); xylabel(0.5, 0.5, "enter new fitting stepsize into console", 0, 0.5);
       variable newstep = atof(keyinput(; prompt = " stepsize = ", default = sprintf("%S",params[actpar].relstep)));
       params[actpar].relstep = newstep; set_params(params);
     }
     % toggle printing of frozen parameter
     { case "s": showf = not showf;
       % set actual parameter index
       newpar = actpar;
       do
       {
	 isfrz = (params[newpar].freeze || params[newpar].tie != NULL || params[newpar].fun != NULL);
	 if (showf == 0 && isfrz) newpar++;
	 if (newpar > length(params)-1) newpar = 0;
       } while (isfrz && actpar != newpar);
       actpar = newpar;
       % update first and last shown parameter
       istart = actpar; iend = istart; f = [0,0];
       if (showf) iend += 4;
       else
       {
	 do
         {
           iend++;
	   ifnot (   iend >= length(params)-1
		  || params[iend].freeze || params[iend].tie != NULL || params[iend].fun != NULL)
	     f = [f[0]+1, iend];
         } while (iend < length(params)-1 && f[0] < 4);
	 if (f[1] == 0) f[1] = istart;
         if (f[0] < 4) iend = f[1];
       }
       if (iend > length(params)-1) iend = length(params)-1;
     }
     % modify fit function
     { case "m": window(pwin);
       erase; charsize(1.4*chz); color(1); xylabel(0.5, 0.5, "enter fit function into console", 0, 0.5); charsize(1);
       fit_fun(keyinput(; prompt = "fit function = ", default = get_fit_fun));
       (params, stepsize) = get_params_and_stepsize();
       update = 1;
     }
     % perform a fit
     { case "f": window(pwin);
       erase; charsize(1.4*chz); color(1); xylabel(0.5, 0.5, "performing a fit...", 0, 0.5); charsize(1);
       ()=fit_counts(&stat; fit_verbose=-1);
       (params, ) = get_params_and_stepsize();  % get_params would give NULL for empty fit_fun
       update = 2;
     }
     % load parameter
     { case "p": if (parfile == "") {
         window(pwin);
         erase; charsize(1.4*chz); color(1); xylabel(0.5, 0.5, "enter parameter filename into console", 0, 0.5);
         parfile = keyinput(; prompt = "parameter filename: ");
       }
       if (access(parfile, F_OK) == 0) { load_par(parfile); params = get_params(); update = 1; message("parameters loaded"); }
       else { vmessage("error: file '%s' does not exist", parfile); parfile = ""; }
     }
     % save parameter
     { case "P": if (parfile == "") {
         window(pwin);
         erase; charsize(1.4*chz); color(1); xylabel(0.5, 0.5, "enter parameter filename into console", 0, 0.5);
         parfile = keyinput(; prompt = "parameter filename: ");
       }
       save_par(parfile);
       message("parameters saved"); 
     }
     % quit
     { case "q": close_plot(pwin); window(1); return; }
     % help
     { case "h": window(pwin);
       erase; charsize(chz); color(1);
       xylabel(-0.02,1.05,"keyboard commands:");
       y0 = .95;
       variable txt;
       foreach txt ([
		     {"Up/Down",     "- change active parameter"},
		     {"Left/Right",  "- decrease/increase active parameter value by stepsize"},
		     {"PgUp/PgDown", "- increase/decrease stepsize of active parameter"},
		     {"t",           "- freeze/thaw active parameter"},
		     {"r",           "- change range of active parameter"},
		     {"v",           "- set value of active parameter"},
		     {"z",           "- set the internal step size used by fit_counts"},
		     {"m",           "- modify fit function"},
		     {"f",           "- perform a fit by running fit_counts"},
		     {"s",           "- show/hide frozen parameters"},
		     {"p/P",         "- load/save fit parameters"},
		     {"q",           "- quit"}
		    ])
       {
	 xylabel(.09, y0, txt[0], 0, 0.5);
	 xylabel(.2,  y0, txt[1]);
	 y0 -= .095;
       }
       ()=keyinput(;nchr=1, silent);
     }
  } % main loop
}

%
% Code from src/fitting/fit_pars.sl
%
variable fit_pars_last_results = NULL;

%%%%%%%%%%%%%%
define fit_pars()
%%%%%%%%%%%%%%
%!%+
%\function{fit_pars}
%\synopsis{computes single-parameter confidence limits for several parameters}
%\usage{Struct_Type results = fit_pars([Integer_Type pars[]]);}
%\qualifiers{
%\qualifier{strict}{[=1]: restarts the calculation if a new best fit was found}
%\qualifier{saveoutput}{[=1]}
%\qualifier{basefilename}{[=<date_time>]}
%\qualifier{level}{[=1]: specifies the confidence level. Values of 0, 1, or 2
%                 indicate 68%, 90%, or 99% confidence levels respectively.
%                 By default, 90% confidence limits are computed.}
%\qualifier{chi2diff}{if given, fit_pars invokes fconf instead of
%                    conf with the value given to caluculate confidence in chi2
%                    range. Defaults to 1.0.}
%\qualifier{tolerance}{convergence criterion for the calculation of the confidence
%      limits (see help for the conf command). Default: 1e-3}
%}
%\qualifier{quiet}{If set, don't print any information on stdout}
%\description
%    The return value \code{results = struct { index, name, value, min, max, conf_min, conf_max, buf_below, buf_above, tex }}
%    is a table with the following information for each parameter:\n
%    \code{min} and \code{max} are the minimum/maximum values allowed.
%    \code{conf_min} and \code{conf_max} are the confidence limits.
%    \code{buf_below} (\code{buf_above}) is the fraction of the allowed range \code{[min:max]}
%    which separates the lower (upper) confidence limit from \code{min} (\code{max}).
%    If one of these buffers is 0, your confidence interval has bounced.
%\seealso{pvm_fit_pars, conf, fconf}
%!%-
{
  variable pars;
  switch(_NARGS)
  { case 0: pars = thawedParameters(); }
  { case 1: pars = (); }
  { help(_function_name()); return; }

  variable strict = qualifier("strict", 1);
  variable saveoutput = qualifier("saveoutput", 1);
  variable basefilename = qualifier("basefilename", strftime("%Y-%m-%d_%H:%M:%S", localtime(_time())))+ "_";
  variable level = qualifier("level",1);
  variable tolerance = qualifier("tolerance",1e-3);
  variable chi2diff = qualifier("chi2diff", 1.);
  variable quiet = qualifier_exists("quiet");
  if (not(level==0 or level==1 or level==2)) {
	 message("requested confidence level not available. Default to level =1: 90%\n");
	 level=1;
  }
   
  variable n = length(pars);
  variable results = struct {
    index     = pars,
    name      = String_Type[n],
    value     = Double_Type[n],
    min       = Double_Type[n],
    max       = Double_Type[n],
    conf_min  = Double_Type[n],
    conf_max  = Double_Type[n],
    buf_below = Double_Type[n],
    buf_above = Double_Type[n],
    tex       = String_Type[n],
  };

  variable i, info;
  _for i (0, n-1, 1)
  { info = get_par_info(pars[i]);
    results.name[i] = info.name;
    results.min[i] = info.min;
    results.max[i] = info.max;
  }

  thaw(pars);
  variable ok;
  do
  { ok = 1;
    for(i=0; i<n && ok; i++)
    { do
      { 
	if (qualifier_exists("chi2diff"))
	  (results.conf_min[i], results.conf_max[i]) = fconf(pars[i],chi2diff,tolerance);
	else
	  (results.conf_min[i], results.conf_max[i]) = conf(pars[i],int(level),tolerance);
        if(results.conf_min[i]==results.conf_max[i] && strict)
          ok = 0;
      } while(results.conf_min[i]==results.conf_max[i]);
      if(saveoutput)
        try { save_par(basefilename+"best.par"); }
	catch AnyError: { vmessage("Could not write to %sbest.par.", basefilename); }

      info = get_par_info(pars[i]);
      results.value[i] = info.value;
      results.buf_below[i] = (results.conf_min[i] - info.min)/(info.max-info.min);
      results.buf_above[i] = (info.max - results.conf_max[i])/(info.max-info.min);
      results.tex[i] = TeX_value_pm_error(info.value, results.conf_min[i], results.conf_max[i]);
      ifnot (quiet) {
        vmessage("(%10g <=) %10g < %25s = %10g < %10g (<= %10g)   [%4.1f%% (conf)%5.1f%%]   %s",
	    results.min[i], results.conf_min[i], results.name[i], results.value[i], results.conf_max[i], results.max[i],
	    100*results.buf_below[i], 100*results.buf_above[i], results.tex[i]);
      }
    }
  } while(not ok);

  variable stat;
  ()=eval_counts(&stat);

  if(saveoutput)
    try
    {
      fits_write_binary_table(basefilename+"conf.fits", "fit_pars-results", results, struct {
    chi2 = stat.statistic,
    num_bins = stat.num_bins,
    n_var_pars = stat.num_variable_params,
    dof = stat.num_bins-stat.num_variable_params,
    chi2red = stat.statistic/(stat.num_bins-stat.num_variable_params)
  });
      variable F = fopen(basefilename+"conf.txt", "w");
      ()=fprintf(F, "%s\n", get_fit_fun());
      ()=fprintf(F, "(%10s <=) %10s < %38s < %10s (<= %10s)\n", "min.all.", "conf.min.", "parameter value", "conf.max.", "max.all.");
      for($1=0; $1<length(results.name); $1++)
        ()=fprintf(F, "(%10g <=) %10g < %25s = %10g < %10g (<= %10g)   [%4.1f%% (conf)%5.1f%%]   %s\n",
		      results.min[$1], results.conf_min[$1], results.name[$1], results.value[$1], results.conf_max[$1], results.max[$1],
		      100*results.buf_below[$1], 100*results.buf_above[$1], results.tex[$1]);
      ()=fprintf(F, "\n%s/dof = %f/%d = %f\n", Fit_Statistic, stat.statistic, stat.num_bins-stat.num_variable_params, stat.statistic/(stat.num_bins-stat.num_variable_params));
      ()=fclose(F);
    }
    catch AnyError: vmessage("Could not write to %sconf.fits and/or %sconf.txt", basefilename, basefilename);

  ifnot (quiet) {
    vmessage("\n");
    vmessage("(%10s <=) %10s < %38s < %10s (<= %10s)", "min.all.", "conf.min.", "parameter value", "conf.max.", "max.all.");
    _for i (0, n-1, 1)
    vmessage("(%10g <=) %10g < %25s = %10g < %10g (<= %10g)   [%4.1f%% (conf)%5.1f%%]   %s",
	      results.min[i], results.conf_min[i], results.name[i], results.value[i], results.conf_max[i], results.max[i],
	      100*results.buf_below[i], 100*results.buf_above[i], results.tex[i]);
    }

  fit_pars_last_results = results;
  return results;
}

%
% Code from src/fitting/fit_pars_read_stdoutlogfile.sl
%
define fit_pars_read_stdoutlogfile()
%!%+
%\function{fit_pars_read_stdoutlogfile}
%\synopsis{reads chi2-improvements from a stdout-logfile created by (pvm_)fit_pars}
%\usage{(t, chi2) = fit_pars_read_stdoutlogfile(String_Type filename);}
%\description
%    \code{t} is the number of seconds since the start of (\code{pvm_})\code{fit_pars}.
%\seealso{fit_pars, pvm_fit_pars}
%!%-
{
  variable filename;
  switch(_NARGS)
  { case 1: filename = (); }
  { help(_function_name()); return; }
  
  variable F = fopen(filename, "r");
  variable T0, t=Double_Type[0], chi2 = Double_Type[0];
  variable m = string_matches(fgetslines(F, 1)[0], `^\(started on\) \(\d\d\d\d\)-\(\d\d\)-\(\d\d\)_\(\d\d\):\(\d\d\):\(\d\d\)`, 1);
  T0 = mktime(struct { tm_sec=atoi(m[7]), tm_min=atoi(m[6]), tm_hour=atoi(m[5]), tm_mday=atoi(m[4]), tm_mon=atoi(m[3])-1, tm_year=atoi(m[2])-1900, tm_wday=0, tm_yday=0, tm_isdst=0 });
  foreach ( fgetslines(F) )
  { % line = first argument of string_matches: left on stack
    m = string_matches(`^found improved fit (chisqr=\([^)]*\)) .* \[\(\d\d\d\d\)-\(\d\d\)-\(\d\d\) \(\d\d\):\(\d\d\):\(\d\d\)\]`, 1);  % second and third argument
    if(m!=NULL)
    { chi2 = [chi2, atof(m[1])];
      t = [t, mktime(struct { tm_sec=atoi(m[7]), tm_min=atoi(m[6]), tm_hour=atoi(m[5]), tm_mday=atoi(m[4]), tm_mon=atoi(m[3])-1, tm_year=atoi(m[2])-1900, tm_wday=0, tm_yday=0, tm_isdst=0 })];
    }
  }
  ()=fclose(F);
  return  t-T0, chi2;
}

%
% Code from src/fitting/fit_steppar.sl
%
private define append_interpol_table(interpol_table, info, i1, i2, i, min_improvement)
{
  variable improvement = info.stat[i] - interpol(info.par[i], info.par[[i1,i2]], info.stat[[i1,i2]]);
  if(improvement > min_improvement)
  { interpol_table.i1 = [interpol_table.i1, i1];
    interpol_table.i2 = [interpol_table.i2, i2];
    interpol_table.i  = [interpol_table.i , i ];
    interpol_table.improvement = [interpol_table.improvement, improvement];
  }
}


private define improvement_by_interpolation(info, i1, i2, i, parname, parfiles)
{
  variable verbose = qualifier("verbose", 1);
  variable doplots = qualifier_exists("plot");
  variable fit = qualifier("fit", 1);
  if(verbose)
    vmessage("interpolation %s = %10g < %10g < %10g  (expected improvement: %f)",
  	     parname, info.par[i1], info.par[i], info.par[i2],
	     info.stat[i] - interpol(info.par[i], info.par[[i1,i2]], info.stat[[i1,i2]])
	    );
  set_params( interpol_params(info.params[i1], info.params[i2], parname, info.par[i]) );
  variable s, stat0 = info.stat[i];
  ()=eval_counts(&s);
  if(s.statistic < info.stat[i])
  { if(verbose)  vmessage( "  improvement: statistic = %g < %g", s.statistic, info.stat[i]);
    save_par(parfiles[i]);
     info.stat[i] = s.statistic;
     info.params[i] = get_params();
    if(doplots)  oplot(info.par, info.stat);
    if(fit)
    {
      if(verbose)  { ()=printf("   fitting..."); ()=fflush(stdout); }
      ()=fit_counts(&s);
      if(verbose)  vmessage(": statistic = %g", s.statistic);
      save_par(parfiles[i]);
       info.stat[i] = s.statistic;
       info.params[i] = get_params();
    }
    if(verbose)  message("");
    return stat0 - info.stat[i];
  }
  return 0;
}


%%%%%%%%%%%%%%%%%%
define fit_steppar()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{fit_steppar}
%\synopsis{tries to obtain better fits for a stepped parameter}
%\usage{Struct_Type result = fit_steppar(String_Type parname, parfiles[]);}
%\qualifiers{
%\qualifier{method}{"optimize", "interpol_guess", or "eval_all"}
%\qualifier{extrapol}{allows for extrapolation as well}
%\qualifier{fit [=1]}{}
%\qualifier{min_improvement}{}
%\qualifier{plot}{}
%\qualifier{verbose [=1]}{}
%}
%\description
%!%-
{
  variable parname, parfiles;
  switch(_NARGS)
  { case 2: (parname, parfiles) = (); }
  { help(_function_name()); return; }

  % qualifiers
  variable method = qualifier("method", "optimize");
  if(all(method!=["optimize", "interpol_guess", "eval_all"]))
  { vmessage(`error (%s): method="%s" unknown`, _function_name(), method);
    return;
  }
  variable extrapol = qualifier_exists("extrapol");
  variable min_improvement = qualifier("min_improvement", 0);
  variable doplots = qualifier_exists("plot");
  variable verbose = qualifier("verbose", 1);

  variable FitVerbose = Fit_Verbose;
  Fit_Verbose = -1;
  variable n = length(parfiles);
  variable info = struct { par = Double_Type[n],
                           stat0 = Double_Type[n],
                           stat,
                           params= Array_Type[n]
                         };
  parfiles = parfiles[array_sort(parfiles)];
  variable i, s;
  _for i (0, n-1, 1)
  { load_par(parfiles[i]);
    info.params[i] = get_params();
    info.par[i] = get_par(parname);
    ()=eval_counts(&s);
    info.stat0[i] = s.statistic;
    if(verbose)  vmessage("%20s: %10g %10g", parfiles[i], info.par[i], info.stat0[i]);
  }
  if(verbose)  message("");

  i = array_sort(info.par);
  parfiles = parfiles[i];
  struct_filter(info, i);
  info.stat = @info.stat0;

  if(doplots)
  { pointstyle(2);
    plot(info.par, info.stat);
  }

  variable di, i1, i2, improved;
  % i1 < i < i2 = i1+di   (unless extrapolation is allowed)

  switch(method)
  { case "optimize":
    %%%%%%%%%%%%%%%
    variable interpol_table = struct
    { i1 = Integer_Type[0],
      i2 = Integer_Type[0],
      i  = Integer_Type[0],
      improvement = Double_Type[0]
    };
    % find improvement due to all interpolation-combinations:
    % loop( (n^3 - 3n^2 + 2n)/6 )
    if(extrapol)
      _for i (0, n-1, 1)  % 0 <= i < n
        _for i1 (0, n-2, 1)  % i1 < i2 < n
          _for i2 (i1+1, n-1, 1)  % i1 < i2 < n
  	    append_interpol_table(interpol_table,  info, i1, i2, i, min_improvement);
    else
      _for i (1, n-2, 1)  % 0 <= i1 < i < i2 < n
        _for i1 (0, i-1, 1)  % 0 <= i1 < i
          _for i2 (i+1, n-1, 1)  % i < i2 < n
  	    append_interpol_table(interpol_table,  info, i1, i2, i, min_improvement);

    while(length(interpol_table.i)>0);
    { % find optimal improvement:
      variable opt = array_sort(interpol_table.improvement)[-1];
      variable iopt = interpol_table.i[opt];
      i1 = interpol_table.i1[opt];
      i2 = interpol_table.i2[opt];
      improved = improvement_by_interpolation(info, i1, i2, iopt, parname, parfiles;; __qualifiers());
      if(improved>0)
      { % update table:
        interpol_table.improvement[where(interpol_table.i ==iopt)] -= improved;  % has just been improved
        interpol_table.improvement[where(interpol_table.i1==iopt or interpol_table.i2==iopt)] = 0;  % previously calculated improvement may no longer be valid
	% recalculate improvement of interpolation from iopt (as i1 or i2):
	if(extrapol)
	{
  	  i2 = iopt;
          _for i (0, n-1, 1)  % 0 <= i < n
           _for i1 (0, i2-1, 1)  % 0 <= i1 < i2
             append_interpol_table(interpol_table,  info, i1, i2, i, min_improvement);
	  i1 = iopt;
          _for i (0, n-1, 1)  % 0 <= i < n
           _for i2 (i1+1, n-1, 1)  % i1 < i2 < n
             append_interpol_table(interpol_table,  info, i1, i2, i, min_improvement);
	}
	else
	{
  	  i2 = iopt;
          _for i (1, iopt-1, 1)  % 0 <= i1 < i < i2
           _for i1 (0, i-1, 1)  % 0 <= i1 < i
             append_interpol_table(interpol_table,  info, i1, i2, i, min_improvement);
	  i1 = iopt;
          _for i (iopt+1, n-2, 1)  % i1 < i < i2 < n
           _for i2 (i+1, n-1, 1)  % i < i2 < n
             append_interpol_table(interpol_table,  info, i1, i2, i, min_improvement);
	}
        struct_filter(interpol_table, where(interpol_table.improvement > min_improvement));
	if(doplots)  oplot(info.par, info.stat);
      }
      else
        struct_filter(interpol_table, [[0:opt-1],[opt+1:length(interpol_table.i)-1]]);
    }
  }
  { case "interpol_guess":
    %%%%%%%%%%%%%%%%%%%%%
    do
    { improved = 0;

      _for i (1, n-2, 1)
        _for di (2, n-1, 1)
          _for i1 (int(_max(0, i+1-di)), int(_min(i-1, n-1-di)), 1)   % i1 >= 0  &&  i1 + di > i   &&   i1 < i  &&  i1 + di < n
          { i2 = i1+di;
  	    if(   interpol(info.par[i], info.par[[i1,i2]], info.stat[[i1,i2]]) < info.stat[i]
	       && improvement_by_interpolation(info, i1, i2, i, parname, parfiles;; __qualifiers()) > 0 )
	    { improved = 1;
    	      if(doplots)  oplot(info.par, info.stat);
	    }
	  }
    } while(improved);
  }
  { case "eval_all":
    %%%%%%%%%%%%%%%
    do
    { improved = 0;
      _for i (1, n-2, 1)   % i1 < i < i2
        _for di (2, n-1, 1)  % di = i2 - i1
          _for i1 (int(_max(0, i+1-di)), int(_min(i-1, n-1-di)), 1)   % i1 >= 0  &&  i1 + di > i   &&   i1 < i  &&  i1 + di < n
	    if( improvement_by_interpolation(info, i1, i2, i, parname, parfiles;; __qualifiers()) > 0 )
	    { improved = 1;
    	      if(doplots)  oplot(info.par, info.stat);
	    }
    } while(improved);
  }

  Fit_Verbose = FitVerbose;
  return info;
}

%
% Code from src/FITS/fits_save_fit/fits_add_fit.sl
%
define fits_add_fit()
%!%+                                                                                                      
%\function{fits_add_fit}
%\synopsis{adds different saved models and observation info togethter in one FITS table}
%\usage{Struct_Type str = fits_add_fit(String_Type filename, String_Type save1
%                         [,String_Type save2] [, ...);
% or Struct_Type str = fits_add_fit(String_Type filename, Struct_Type save1
%                      [, Struct_Type save2] [, ...);}
% or Struct_Type str = fits_add_fit(String_Type filename, Array_Type);
%\description
%
%   This function is based on fits_save_fit and fits_load_fit_struct.
%   If called with String_Type filenames, these files are loaded with
%   fits_load_fit_struct, merged, and saved to the (new) file
%   "filename".
%   If called with Struct_Type, these structure should have been
%   created with fits_save_fit_struct, which will then be merged and
%   saved to "filename".
%   If called with Array_Type, the entries of the array should either
%   be strings loadable with fits_load_fit_struct or structures
%   created with fits_save_fit_struct.
%   It returns the merged structure.
%                                                                                                         
%\seealso{fits_load_fit_struct,fits_save_fit_struct,fits_write_fits_struct,fits_save_fit,merge_struct_arrays}
%!%-
{
  variable fil ;
  %return help only when no or only one argument (the file name) is given
  switch (_NARGS) 
  { case 0 or case 1 : help(_function_name()); return; }
  { case 2 : fil = () ;} %could be already an array, so we don't need to create one
  { 
    
    %last argument determines the type of all arguments: String (files)
    %or Structure?
    
    variable i, fil0 = () ;
    fil = [fil0][Integer_Type[_NARGS-1]] ;
    
    _for i (1, _NARGS-2, 1)
      fil[i] = () ;
    
    %reverse entries so that the order in the FITS file corresponds to
    %the order given as arguments
    fil = reverse(fil) ;
  }
  variable name = () ;

  variable nfits = length(fil);
  variable stru = Struct_Type[nfits] ;
  
  %if files, read them and merge them
  if (typeof(fil[0]) == String_Type)
  {
    _for i (0, length(fil)-1, 1)
    {
      if(qualifier_exists("verbose")){
	()=printf("Loading saved fit %s...\n", fil[i]) ; 
      }
      stru[i] = fits_load_fit_struct(fil[i]) ;
    }
      }
  %if already structures only merge them
  else if (typeof(fil[0]) == Struct_Type)
  {
    stru = fil;
    %	stru = merge_struct_arrays(fil ; reshape = 0) ;
    %	stru = merge_struct_arrays(fil) ;
  }
  %if any other type, return help message ;
  else
  {
    vmessage(`error (%s): arguments have to be either Struct_Type or String_Type`, _function_name() ); return() ;
  }


  %% ---- initialize structure to save data in ----
  
  %%loop over all elements of stru
  variable len = 0;
  _for i (0, length(stru)-1, 1){    
    %%len = maximum number of data set loaded so far
    len = max([len,length(stru[i].exposure)]);    
  }    

  %% --- create info fields ---
  variable lenfields = ["tstop","tstart_mjd","tstop_mjd","exposure","mjdref","exposure",
			"tfirst_mjd","tlast_mjd"];

  variable dummy_nfits_len = Double_Type[nfits,len];
  dummy_nfits_len[*,*] = _NaN;
  variable allinone = struct {fit_fun=String_Type[nfits],
    tstart=@dummy_nfits_len,
    tstop=@dummy_nfits_len,
    tstart_mjd=@dummy_nfits_len,
    tstop_mjd=@dummy_nfits_len,
    exposure=@dummy_nfits_len,
    data=String_Type[nfits],
    bkg=String_Type[nfits],
    instrument=String_Type[nfits],
    target=String_Type[nfits],
    telescope=String_Type[nfits],
    filter=String_Type[nfits],
    date_obs=String_Type[nfits],
    date_end=String_Type[nfits],
    obs_id=String_Type[nfits],
    rebin=String_Type[nfits],
    notice_list=String_Type[nfits],
    chi_red=Double_Type[nfits],
    chi=Double_Type[nfits],
    dof=UInteger_Type[nfits],
    fit_statistic=String_Type[nfits],
    cwd=String_Type[nfits],
    mjdref=@dummy_nfits_len,
    sys_err=String_Type[nfits],
    tfirst_mjd=@dummy_nfits_len,
    tlast_mjd=@dummy_nfits_len};

  %% --- set other fields & fill up special fields with values ---
  
  variable diff,j,temprow,names,row,n;
  variable dims, num_dims, data_type;

  if(qualifier_exists("verbose")){
    ()=printf("Creating template for combined structure") ;
  }
  
  %% ----------- start loop over all elements of stru
  _for i (0, length(stru)-1, 1){

    names = get_struct_field_names(stru[i]);
    diff = complementary_array(names,get_struct_field_names(allinone));

    %% if field names not yet in allinone exist, create such a field
    %% in the allinone structure and intialize it to defauls values
    if (length(diff) != 0) {

      %%create new fields in allinone
      allinone = struct_combine(allinone, diff);
      
      %%set all new entries in the new fields to defauls values
      _for j (0,length(diff)-1,1) {
	(dims, num_dims, data_type) = array_info(get_struct_field(stru[i],diff[j]));       
	%%depending of number of dimensions (1D or 2D) of the array
	%%constiting a row of the structre ...
	switch (num_dims)
	%% -- 1D array --
	{ num_dims == 1 :
	    set_struct_field(allinone,diff[j],data_type[nfits]);
	  %% -- different data types for 1D arrays --
	  switch (data_type)
	  {data_type == Double_Type :
	      temprow = get_struct_field(allinone,diff[j]); temprow[*] = _NaN;
	    set_struct_field(allinone,diff[j],temprow);
	  }
	  {data_type == String_Type :
	      temprow = get_struct_field(allinone,diff[j]); temprow[*] = "";
	    set_struct_field(allinone,diff[j],temprow);}
	  {data_type == Integer_Type :
	      temprow = get_struct_field(allinone,diff[j]); temprow[*] = INT_MIN;
	    set_struct_field(allinone,diff[j],temprow);}
	  {vmessage(`error (%s): Structure fiels of other types than Double_Type, String_Type \nand Integer_Type not implemented`, _function_name() ); return();}
	}	
	%% -- 2D array --
	{ num_dims == 2 :
	    set_struct_field(allinone,diff[j],data_type[nfits,dims[1]]);
	  %% -- different data types for 2D arrays --
	  switch (data_type)
	  {data_type == Double_Type :
	      temprow = get_struct_field(allinone,diff[j]); temprow[*,*] = _NaN;
	    set_struct_field(allinone,diff[j],temprow);
	  }
	  {data_type == String_Type :
	      temprow = get_struct_field(allinone,diff[j]); temprow[*,*] = "";
	    set_struct_field(allinone,diff[j],temprow);}
	  {data_type == Integer_Type :
	      temprow = get_struct_field(allinone,diff[j]); temprow[*,*] = INT_MIN;
	    set_struct_field(allinone,diff[j],temprow);}
	  {vmessage(`error (%s): Structure fiels of other types than Double_Type, String_Type \nand Integer_Type not implemented`, _function_name() ); return();}
	}
	%%error message for arrays with other number of dimensions
	{vmessage(`error (%s): Structure fiels with more than two dimensions not implemented`, _function_name() ); return();};
      }    
    }
  }
  %% ----------- end loop over all elements of stru

  
  %% ----------- start loop over all fields of allinone
  variable temparr,allnames;
  allnames = get_struct_field_names(allinone);
  foreach row (allnames) {
    
    if(qualifier_exists("verbose")){
      ()=printf("Working on row %s ...\n", row) ; 
    }
    
    temparr = get_struct_field(allinone,row);
    
    %% -- now loop over all elements of stru
    _for i (0, length(stru)-1, 1){
      
      names = get_struct_field_names(stru[i]);
      if(length(where(names == row)) == 0) {
	%%if row does not exist in this fits, skip it
	continue;
	
      } else {

	%%if row exists, take care of whether it's 1D or 2D
	(dims, num_dims, data_type) = array_info(get_struct_field(stru[i],row));
	
	%%everything that is not 1 or 2 should already have been
	%%caught in the last step
	switch (num_dims)
	{num_dims == 1 :
	    temparr[i] = get_struct_field(stru[i],row)[0];
	}
	{num_dims == 2 :
	    
	    %%we expect a field to be an array of type
	    %%Some_Type[1,n], i.e. dims = [1,n], therefore:
	    _for n (0,dims[1]-1,1) {
	      temparr[i,n] = get_struct_field(stru[i],row)[0,n];
	    }	 	  
	};	
      }            
    }    
    set_struct_field(allinone,row,temparr);    
  }    

  fits_save_fit_write(name, allinone) ;
  
  %%% TO DO
  % * instruments should be aphabetically sorted; sort all data in
  % respective places
    
  return allinone ;
}

%
% Code from src/FITS/fits_append_binary_table.sl
%
define fits_append_binary_table()
%!%+
%\function{fits_append_binary_table}
%\synopsis{appends a binary table to the end of a FITS file}
%\usage{fits_append_binary_table(filename, [extname], data[, keys[, hist]]);}
%\description
%    This function employs functions from ISIS' cfitsio module
%    in order to open/create a FITS file,
%#c    find the number of HDUs,
%#c    move to the end of the file,
%    and append a binary table extension.
%#c\seealso{fits_open_file, _fits_get_num_hdus, _fits_movabs_hdu, fits_write_binary_table}
%\seealso{fits_open_file, fits_write_binary_table}
%!%-
{
  variable filename, extname=NULL, data, keys=NULL, hist=NULL;
  switch(_NARGS)
  { case 2: (filename,          data) = (); }
  { case 3: (filename, extname, data) = (); }
  { case 4: (filename, extname, data, keys) = (); }
  { case 5: (filename, extname, data, keys, hist) = (); }
  { help(_function_name()); return; }

  variable F = typeof(filename)==String_Type
               ? fits_open_file(filename, stat_file(filename)==NULL ? "c" : "w")
               : filename;
%  variable num_hdus; ()=_fits_get_num_hdus(F, &num_hdus);
%  ()=_fits_movabs_hdu(F, num_hdus);
  if(extname==NULL)  extname = "";
  fits_write_binary_table(F, extname, data, keys, hist);
  fits_close_file(F);
}

%
% Code from src/FITS/fits_append_extension.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_append_extension()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_append_extension}
%\synopsis{appends a FITS extension to another FITS file}
%\usage{fits_append_extension(String_Type infiles[], String_Type outfile);}
%\description
%    As the external FTOOL fappend is used for this task,
%    \code{infiles} may contain extension numbers according to the FTOOLS conventions.
%    The extensions are appended at the end of \code{outfile}.
%
%    This function should usually not be used!
%    ISIS' cfitsio module allows to write several extensions
%    into a file after opening it with fits_open_file.
%\seealso{fappend [FTOOLS], fits_append_tmp_extension}
%!%-
{
  variable infiles, outfile;
  switch(_NARGS)
  { case 2: (infiles, outfile) = (); }
  { help(_function_name()); return; }

  variable infile;
  foreach infile ([infiles])
  {
    variable cmd = `fappend "$infile" "$outfile"`$;
    if(qualifier_exists("verbose"))  message(cmd);
    ()=system(cmd);
  }
}

%
% Code from src/FITS/fits_append_tmp_extension.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_append_tmp_extension()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_append_tmp_extension}
%\synopsis{appends a temporary FITS extension to another FITS file before deleting it}
%\usage{fits_append_tmp_extension(String_Type infiles[], String_Type outfile);}
%\description
%    As the external FTOOL fappend is used for this task,
%    \code{infiles} may contain extension numbers according to the FTOOLS conventions.
%    The extensions are appended at the end of \code{outfile}.
%    Afterwards, all \code{infiles} are deleted.
%
%    This function should usually not be used!
%    ISIS' cfitsio module allows to write several extensions
%    into a file after opening it with fits_open_file.
%\seealso{fappend [FTOOLS], fits_append_extension}
%!%-
{
  variable infiles, outfile;
  switch(_NARGS)
  { case 2: (infiles, outfile) = (); }
  { help(_function_name()); return; }

  fits_append_extension(infiles, outfile;; __qualifiers());

  variable infile;
  foreach infile ([infiles])
  {
    variable m = string_match(infile, `\(.*\)\[`, 1);
    if(m!=NULL)  infile = m[1];
    if(qualifier_exists("verbose"))  vmessage("removing %s", infile);
    ()=remove(infile);
  }
}

%
% Code from src/FITS/fits_column_unit_struct.sl
%
define fits_column_unit_struct()
%!%+
%\function{fits_column_unit_struct}
%\synopsis{creates a structure of FITS header keywords containing the units of columns}
%\usage{Struct_Type fits_column_unit_struct(Struct_Type data; field1=unit1, field2=unit2)}
%\example
%    \code{variable data = struct { time, rate };}\n
%    \code{fits_write_binary_table("lc.fits", "lightcurve", data, fits_column_unit_struct(data; time="MJD", rate="counts/s/PCU") );}
%\seealso{fits_write_binary_table}
%!%-
{
  variable data;
  switch(_NARGS)
  { case 1: data = (); }
  { help(_function_name()); return; }

  variable fields = get_struct_field_names(data);

  variable keywords = String_Type[0];
  data = String_Type[0];
  variable q, qs = get_struct_field_names(__qualifiers());
  foreach q (qs)
  { variable i = where(fields==q);
    if(length(i)>0)
    { keywords = [keywords, "TUNIT"+string(i[0]+1)];
      data = [data, qualifier(q)];
    }
    else
      vmessage(`warning (%s): column "%s" does not exist in the data`, _function_name(), q);
  }
  variable keys = @Struct_Type(keywords);
  set_struct_fields(keys, __push_array(data));
  return keys;
}

%
% Code from src/FITS/fits_get_hdu_names.sl
%
define fits_get_hdu_names()
%!%+
%\function{fits_get_hdu_names}
%\synopsis{returns the names of all extensions within a FITS-file}
%\usage{String_Type[] fits_get_hdu_names(Fits_File_Type fp);}
%\description
%  Moves to the first extension of a FITS-file using
%  `_fits_movabs_hdu' and then iterates over all extensions
%  using `_fits_movrel_hdu' to read the 'EXTNAME' keyword.
%  The string array of all extension names is returned.
%  
%  Note that the file-pointer is located at the last
%  extension in the end. Furthermore the first extension
%  has the index 1 (at least for `_fits_movabs_hdu'), which
%  has to be taken into account if the indices of the name-
%  array are used to find specific extensions.
%\seealso{fits_open_file, fits_read_key}
%!%-
{
  variable fp;
  switch (_NARGS)
    { case 1: fp = (); }
    { help(_function_name); return; }

  % move to first extension
  if (_fits_movabs_hdu(fp, 1) > 0) {
    vmessage("error (%s): cannot move to first extension, is the file opened?", _function_name);
    return;
  }

  % read extension names
  variable names = String_Type[0];
  do {
    names = [names, fits_read_key(fp, "extname")];
  } while (_fits_movrel_hdu(fp, 1) == 0);

  return names;
}

%
% Code from src/FITS/fits_lc_exposure.sl
%
define fits_lc_exposure(file)
%!%+
%\function{fits_lc_exposure}
%\synopsis{returns the exposure time of a lightcurve,
%    given by a FITS-file, in seconds}
%\usage{Double_Type fits_lc_exposure(String_Type file)}
%!%-
{
  variable dt, u;
  (dt,u) = fits_read_key(file,"TIMEDEL","TIMEUNIT");
  u = strlow(u);
  switch(u)
    { case "s": return dt; }
    { case "d": return dt*86400.; }
    { vmessage("warning (%s): lightcurve fits file contains unknown TIMEUNIT=\"%S\"",  _function_name(), u); }
}

%
% Code from src/fitting/fits_list_fit_pars.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_list_fit_pars(){
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_list_fit_pars}
%\synopsis{Lists parameter values and confidence intervals for a file saved with fits_save_fit.}
%\usage{fits_list_fit_pars(String_Type fit.fits)}
%\description
%    When loading a previously saved fit with fits_load_fit, \code{list_par}
%    does not print the borders of the parameter confidence intervals saved
%    in appropriate fields of fit.fits.
%    Instead the parameter limits are printed.
%    As a work-around use this function to get a similar list as for
%    \code{list_par} including the confidence intervals if already determined.
%\example
%    isis> fits_list_fit_pars("fit.fits");
%\seealso{fits_save_fit,fits_load_fit}
%!%-

  variable fit;
  switch(_NARGS)
  { case 1: (fit) = ();}
  { help(_function_name()); return; }
  
  variable fit_struct, fields, i, cnf_lo_frmt, cnf_hi_frmt, val_frmt, lim_lo_frmt, lim_hi_frmt;
  
  variable val = Double_Type[0];
  variable cnf_lo = Double_Type[0];
  variable cnf_hi = Double_Type[0];
  variable lim_lo = Double_Type[0];
  variable lim_hi = Double_Type[0];
  variable tie = String_Type[0];
  variable frz = Integer_Type[0];
  variable name = String_Type[0];
  
  fit_struct=fits_load_fit_struct(fit);
  fields=get_struct_field_names(fit_struct);
  
  vmessage(fit_struct.fit_fun[0]);
  vmessage("%-30s%-10s%-10s%-15s%-15s%-15s%-15s%-15s","param","tie-to","freeze","lim_lo","conf_min","value","conf_max","lim_hi");
  vmessage("----------------------------------------------------------------------------------------------------------------------------------------");
  for (i=0;i<length(fields);i++) {
    if (is_substr(fields[i],"value")>0) val=[val,get_struct_field(fit_struct,fields[i])[0]];
    if (is_substr(fields[i],"conf")>0) {
      cnf_lo=[cnf_lo,get_struct_field(fit_struct,fields[i])[0,0]];
      cnf_hi=[cnf_hi,get_struct_field(fit_struct,fields[i])[0,1]];
      name=[name,fields[i]];
    }
    if (is_substr(fields[i],"lim")>0) {
      lim_lo=[lim_lo,get_struct_field(fit_struct,fields[i])[0,0]];
      lim_hi=[lim_hi,get_struct_field(fit_struct,fields[i])[0,1]];
    }
    if (is_substr(fields[i],"tie")>0) {
      if (typeof(get_struct_field(fit_struct,fields[i])[0]) == String_Type) {
	tie=[tie,get_struct_field(fit_struct,fields[i])[0]];
      }
      else tie=[tie,string(get_struct_field(fit_struct,fields[i])[0])];
    }
    if (is_substr(fields[i],"freeze")>0) frz=[frz,get_struct_field(fit_struct,fields[i])[0]];
  }
  for (i=0;i<length(name);i++) {

#ifdef    
    if ((floor(cnf_lo[i]*10000)==0. || floor(cnf_lo[i]/1000.)>0) && cnf_lo[i]!=0.) cnf_lo_frmt = "%-15.2e";
    else if ((floor(cnf_lo[i]*10000)>0. || floor(cnf_lo[i]/1000.)==0) && cnf_lo[i]!=0.) cnf_lo_frmt = "%-15.3f";
    else if (cnf_lo[i]==0.) cnf_lo_frmt = "%-15.0f";
    if ((floor(cnf_hi[i]*10000)==0. || floor(cnf_hi[i]/1000.)>0) && cnf_hi[i]!=0.) cnf_hi_frmt = "%-15.2e";
    else if ((floor(cnf_hi[i]*10000)>0. || floor(cnf_hi[i]/1000.)==0) && cnf_hi[i]!=0.) cnf_hi_frmt = "%-15.3f";
    else if (cnf_hi[i]==0.) cnf_hi_frmt = "%-15.0f";
    
    if ((floor(lim_lo[i]*10000)==0. || floor(lim_lo[i]/1000.)>0) && lim_lo[i]!=0.) lim_lo_frmt = "%-15.2e";
    else if ((floor(lim_lo[i]*10000)>0. || floor(lim_lo[i]/1000.)==0) && lim_lo[i]!=0.) lim_lo_frmt = "%-15.3f";
    else if (lim_lo[i]==0.) lim_lo_frmt = "%-15.0f";
    if ((floor(lim_hi[i]*10000)==0. || floor(lim_hi[i]/1000.)>0) && lim_hi[i]!=0.) lim_hi_frmt = "%-15.2e";
    else if ((floor(lim_hi[i]*10000)>0. || floor(lim_hi[i]/1000.)==0) && lim_hi[i]!=0.) lim_hi_frmt = "%-15.3f";
    else if (lim_hi[i]==0.) lim_hi_frmt = "%-15.0f";
#endif    

    if ((floor(cnf_lo[i]*1000)==0.) && cnf_lo[i]!=0.) cnf_lo_frmt = "%-15.3e";
    else if ((floor(cnf_lo[i]*1000)>0.)) cnf_lo_frmt = "%-15.4f";
    if ((floor(cnf_lo[i])>1000.)) cnf_lo_frmt = "%-15.3e";    
    else if (cnf_lo[i]==0.) cnf_lo_frmt = "%-15.0f";
    if ((floor(cnf_hi[i]*1000)==0.) && cnf_hi[i]!=0.) cnf_hi_frmt = "%-15.3e";
    else if ((floor(cnf_hi[i]*1000)>0.)) cnf_hi_frmt = "%-15.4f";
    if ((floor(cnf_hi[i])>1000.)) cnf_hi_frmt = "%-15.3e";    
    else if (cnf_hi[i]==0.) cnf_hi_frmt = "%-15.0f";
    
    if ((floor(lim_lo[i]*1000)==0.) && lim_lo[i]!=0.) lim_lo_frmt = "%-15.3e";
    else if ((floor(lim_lo[i]*1000)>0.)) lim_lo_frmt = "%-15.4f";
    if ((floor(lim_lo[i])>1000.)) lim_lo_frmt = "%-15.3e";    
    else if (lim_lo[i]==0.) lim_lo_frmt = "%-15.0f";
    if ((floor(lim_hi[i]*1000)==0.) && lim_hi[i]!=0.) lim_hi_frmt = "%-15.3e";
    else if ((floor(lim_hi[i]*1000)>0.)) lim_hi_frmt = "%-15.4f";
    if ((floor(lim_hi[i])>1000.)) lim_hi_frmt = "%-15.3e";    
    else if (lim_hi[i]==0.) lim_hi_frmt = "%-15.0f";
    
    if ((floor(val[i]*1000)==0.) && val[i]!=0.) val_frmt = "%-15.3e";
    else if ((floor(val[i]*1000)>0.)) val_frmt = "%-15.4f";
    if ((floor(val[i])>1000.)) val_frmt = "%-15.3e";    
    else if (val[i]==0.) val_frmt = "%-15.0f";

    variable formatstring = "%-30s%-10s%-10u"+sprintf("%s%s%s%s%s",lim_lo_frmt,cnf_lo_frmt,val_frmt,cnf_hi_frmt,lim_hi_frmt);
    vmessage(formatstring,strreplace(name[i],"_conf",""),tie[i],frz[i],lim_lo[i],cnf_lo[i],val[i],cnf_hi[i],lim_hi[i]);        
  }
}
%
% Code from src/FITS/fits_save_fit/fits_load_fit.sl
%
define fits_load_fit()
%!%+
%\function{fits_load_fit}
%\synopsis{defines the data and model of a FITS file written by 'fits_save_fit'}
%\usage{Integer_Type = fits_load_fit(String_Type filename[, Integer_Type index = 0]);}
%\qualifiers{
%  \qualifier{loadfun}{data load function (defualt = &load_data)
%            The following format is neccessary: 
%            Argument: String_Type Filename
%            Return:   Dataset ID}
%  \qualifier{nodata}{do not load the data}
%  \qualifier{norebin}{do not notice and rebin the data}
%  \qualifier{nomodel}{do not define the model}
%  \qualifier{noff}{do not set the fit function. Instead
%            only values of existing parameters
%            of the current model are loaded and
%            set. Maybe more useful in combination
%            with 'nodata' to just restore actual
%            fit parameters without overwritting
%            any additional (not saved) components}
%  \qualifier{noeval}{do not evaluate the model}
%  \qualifier{noerr}{do not load the systematic error fraction}
%  \qualifier{toeval}{structure of qualifiers
%            passed to 'eval_counts'}
%  \qualifier{strct}{reference to a variable to return the
%            structure loaded by fits_load_fit_struct}
%  \qualifier{ROC}{array of values to set
%            Rmf_OGIP_Compliance for each
%            spectrum before loading
%            (default values = 2)}
%}
%\description
%  This function restores the fit saved previously
%  by 'fits_save_fit'. That includes the loaded
%  data and the used model, which is evaluated at
%  the end.
%\seealso{fits_save_fit, fits_load_fit_struct, fits_list_fit_pars}
%!%-
{
  variable fitsfile, ind;
  switch(_NARGS)
    { case 1: (fitsfile) = (); ind = 0; }
    { case 2: (fitsfile, ind) = (); }
    { help(_function_name()); return; }

  % load fits structure
  variable fit = fits_load_fit_struct(fitsfile);

  % extract spectrum filenames and load the data
  variable dataind = Integer_Type[0];
  variable i;
  variable loadfun = qualifier("loadfun",&load_data);
  ifnot (qualifier_exists("nodata"))
  {
    % note: use same delimiter as in 'fits_save_fit'!
    variable specfiles = strchop(fit.data[ind], '\n', 0);
    variable backfiles = strchop(fit.bkg[ind], '\n', 0);
    variable binlists = strchop(fit.rebin[ind], '\n', 0);
    variable notlists = strchop(fit.notice_list[ind], '\n', 0);
    % loop over instruments
    variable roc = qualifier("ROC", ones(length(specfiles))*2);
    _for i (0, length(specfiles)-1, 1)
    {
      Rmf_OGIP_Compliance = roc[i];
      if (specfiles[i] == " "){
	vmessage("WARNING [fits_load_fit]: Data set no. %d has no spectrum file, ignoring", i) ;
	continue ;
      }
      dataind = [dataind, (@loadfun)( specfiles[i] ;; __qualifiers) ];

       if (backfiles[i] == " "){
	  vmessage("WARNING [fits_load_fit]: Data set no. %d has no background", i) ;	  
       } else {
	  () = define_back(dataind[i], backfiles[i]);
       }

       Rmf_OGIP_Compliance = 2;
      % notice and rebin the data
      ifnot (qualifier_exists("norebin"))
      {
	rebin_data(dataind[-1], rebin_human2isis(binlists[i]));
	ignore(dataind[-1]);
	notice_list(dataind[-1], notice_human2isis(notlists[i]));
      }
    }
  }

  % define the fit function and model
  ifnot (qualifier_exists("nomodel") || typeof(fit.fit_fun) == Null_Type)
  {
    ifnot (qualifier_exists("noff")) fit_fun(fit.fit_fun[ind]);
    % get all model parameters
    variable params = get_params;
    % loop over parameter structure
    % and restore values accordingly
    % note: use same field names than 'fits_save_fit'!
    _for i (0, length(params)-1, 1)
    {
      variable par = fits_conv_to_legal_char(params[i].name);
      if (struct_field_exists(fit, par + "_value"))
      {
        if (struct_field_exists(fit, par + "_lim"))
        {
          params[i].min = get_struct_field(fit, par + "_lim")[ind,0];
          params[i].max = get_struct_field(fit, par + "_lim")[ind,1];
	}
        params[i].value = get_struct_field(fit, par + "_value")[ind];
        params[i].fun = get_struct_field(fit, par + "_fun")[ind];
        if (strlen(params[i].fun) < 2) params[i].fun = NULL;
        params[i].tie = get_struct_field(fit, par + "_tie")[ind];
        params[i].freeze = get_struct_field(fit, par + "_freeze")[ind];
        params[i].units = get_struct_field(fit, par + "_units")[ind];
      }
    }
    set_params(params);
  }

  %set the systematic error fraction
  ifnot (qualifier_exists("noerr"))
  {
    % Extract sys_err values
    variable sys_err = fits_read_table(fitsfile + "[1]").sys_err;
    sys_err = strchop(sys_err[0], '\n', 0);

    % set sys_err values for each dataset
    variable j=0;
    _for j (0, length(sys_err)-1, 1)
    {
      set_sys_err_frac(all_data[-1-j], string2sys_err_array(sys_err[-1-j]));
    }
  }

  % evaluate the model
  ifnot (qualifier_exists("nodata") || qualifier_exists("noeval"))
  {
    ()=eval_counts(;; qualifier("toeval"));
  }

  % eventually return the fit-structure
  if (qualifier_exists("strct")) {
    variable strct = qualifier("strct");
    @strct = fit;
  }
  
  return dataind;
}

%
% Code from src/FITS/fits_save_fit/fits_load_fit_struct.sl
%
define fits_load_fit_struct(name)
%!%+
%\function{fits_load_fit_struct}
%\synopsis{loads a FITS file written by 'fits_save_fit'}
%\usage{Struct_Type str = fits_load_fit_struct(String_Type filename);}
%\seealso{fits_save_fit}
%!%-
{
  variable l = fits_get_num_hdus(name);

   if (l <= 1)
   {
      message(" *** Warning: No extension contained in the FITS file!");return "";
   }
   
  variable strs = Struct_Type[l-1];
  
  variable i;
  variable str = fits_read_table(name+"[1]";casesen);
  _for i(1,l-2,1)
  {
    str = struct_combine(str, fits_read_table(name+"["+string(i+1)+"]";casesen));
  }
  
  return str;
}

%
% Code from src/FITS/fits_modify_header.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_modify_header()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_modify_header}
%\synopsis{modifies the header of a FITS file}
%\usage{fits_modify_header(String_Type filename, keyword, value[, comment]);}
%\description
%    fits_modify_header uses the external FTOOL \code{fmodhead} and is therefore deprecated.
%    Use \code{fits_update_key} from ISIS' cfitsio module instead.
%\seealso{fits_update_key}
%!%-
{
  message("fits_modify_header is deprecated. Use fits_update_key instead.");

  variable filename, keyword, value, comment="";
  switch(_NARGS)
  { case 3: (filename, keyword, value) = (); }
  { case 4: (filename, keyword, value, comment) = (); }
  { help(_function_name()); return; }

  variable tmpfile = sprintf(".tmp.fits_modify_header.%d.%d", getuid(), getpid());
  variable F = fopen(tmpfile, "w");
  ()=fprintf(F, "%s %s / %s\n", keyword, value, comment);
  ()=fclose(F);
  ()=system(`fmodhead "$filename" $tmpfile`$);
  ()=remove(tmpfile);
}

%
% Code from src/FITS/fits_nr_extensions.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_nr_extensions()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_nr_extensions}
%\synopsis{counts the extensions of a FITS file}
%\usage{Integer_Type fits_nr_extensions(String_Type filename)}
%\description
%    This function counts the number of extensions
%    in addition to the primary extension, i.e.,
%    returns \code{fits_num_hdus(filename)-1}.
%\seealso{_fits_get_num_hdus, fits_num_hdus}
%!%-
{
  variable filename;
  switch(_NARGS)
  { case 1: filename = (); }
  { help(_function_name()); return; }

  variable fp;
  if(_fits_open_file (&fp, filename, "r"))
    return -1;  % error
  variable nr = 0;
  while(_fits_movrel_hdu(fp, 1)==0)  nr++;
  fits_close_file(fp);

  return nr;
}

%
% Code from src/FITS/fits_num_hdus.sl
%
%%%%%%%%%%%%%%%%%%%%
define fits_num_hdus()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_num_hdus}
%\usage{Integer_Type fits_num_hdus(String_Type filename)}
%\description
%    This function is just a wrapper around the
%    \code{_fits_get_num_hdus} function from ISIS' cfitsio module.
%\seealso{_fits_get_num_hdus}
%!%-
{
  variable filename;
  switch(_NARGS)
  { case 1: filename = (); }
  { help(_function_name()); return; }

  variable fp = fits_open_file(filename, "r");
  variable num; ()=_fits_get_num_hdus(fp, &num);
  fits_close_file(fp);
  return num;
}

%
% Code from src/FITS/fits_plot_rmf.sl
%
%%%%%%%%%%%%%%%%%%%%
define fits_plot_rmf()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_plot_rmf}
%\synopsis{plots a redistribution matrix from a compressed RMF file}
%\usage{fits_plot_rmf(String_Type RMFfile);
%\altusage{(interpol_matrix_density, Ebounds, energ) = fits_plot_rmf(RMFfile; getvalues)}
%}
%\qualifiers{
%\qualifier{nx}{number of pixels in x-direction [default=400]}
%\qualifier{ny}{number of pixels in x-direction [default=300]}
%\qualifier{getvalues}{retrieves the interpolated matrix density}
%\qualifier{noplot}{skip plotting, but retrieves the interpolated matrix density}
%}
%\seealso{fits_read_rmf}
%!%-
{
  variable RMFfile;
  switch(_NARGS)
  { case 1: RMFfile = (); }
  { help(_function_name()); return; }

  variable rmf = fits_read_rmf(RMFfile);

  variable EMIN, EMAX, ELO, EHI;
  (EMIN, ELO) = get_grid(rmf.ebounds.bin_lo, rmf.energy.bin_lo);
  (EMAX, EHI) = get_grid(rmf.ebounds.bin_hi, rmf.energy.bin_hi);

  % bin density
  variable r_density = rmf.matrix / ((EMAX-EMIN) * (EHI-ELO));

  % new grid
  variable x = [ rmf.ebounds.bin_hi[0] : rmf.ebounds.bin_lo[-1] : #qualifier("nx", 400)];
  variable y = [ rmf.energy.bin_hi[0]  : rmf.energy.bin_lo[-1]  : #qualifier("ny", 300) ];
  variable r2_density = interpol2D(x, y, rmf.ebounds.center, rmf.energy.center, r_density);

  % bin integrated
  variable r2 =  r2_density * (x[1]-x[0]) * (y[1]-y[0]);

  ifnot(qualifier_exists("noplot"))
  { xlabel("EBOUNDS [keV]");
    ylabel("energy [keV]");
    plot_image(log(1+max(r2))-log(1+r2), , x, y);
  }

  if(qualifier_exists("getvalues") || qualifier_exists("noplot"))
    return r2, x, y;
}

%
% Code from src/FITS/fits_read_header.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define fits_read_header()
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Read a FITS header}
%\usage{Struct_Type fits_read_header(String_Type file or Fits_File_Type fp);}
%\description
%    This function reads the header of the fits file given by the
%    `file' argument and returns it as a structure.  If `file' is
%    a string, then the file will be opened, read-out, and closed
%    automatically. Otherwise, `file' should represent an already
%    opened FITS file (which will remain opened).
%\qualifiers{
%   \qualifier{lowercase}{return structure tags in lower case
%                           (default: upper case)}
%}
%\seealso{fits_read_records, fits_open_file}
%!%-
{
  variable fp;
  switch (_NARGS)
    { case 1: fp = (); }
    { help(_function_name); return; }

  % read the header of the FITS-file into an array of strings
  variable rec;
  if (typeof(fp) == String_Type) {
    if (string_matches(fp, "\[[-_A-Z0-9]*\]$"R, 1) == NULL) { fp += "[1]"; }
    fp = fits_open_file(fp, "r");
    rec = fits_read_records(fp);
    fits_close_file(fp);
  } else {
    rec = fits_read_records(fp);
  }

  % loop over all records
  variable head = NULL, n, key, value, extr, sub;
  _for n (0, length(rec)-1, 1) {
    sub = substr(rec[n], 1, 8);
    % check on continued string field
    if (sub == "CONTINUE") {
      value = get_struct_field(head, key);
      set_struct_field(head, key, sprintf("%s%s",
	substr(value, 1, strlen(value)-1), % remove '&'
	string_matches(rec[n], "'\(.*\)'"R)[1]
      ));
    }
    % check on comment or history
    else ifnot (any(sub == ["COMMENT ", "HISTORY "])) {
      % split the string at the first = into key and value
      key = strchop(rec[n], '=', 0);
      if (length(key) > 1) {
        value = strjoin(key[[1:]], "=");
        key = key[0];
        % check on HIERARCH
        if (sub == "HIERARCH") { key = substr(key, 9, strlen(key)); }
	% replace illegal characters
	key = strreplace(key, "-", "_");
	key = strreplace(key, " ", "");
	if (qualifier_exists("lowercase")) {
	    key=strlow(key);
	}
	% distinguish between numbers and strings
        extr = string_matches(value, "'\(.*\)'"R);
	if (extr == NULL)  { % number
	  value = strtrim(strchop(value, '/', 0)[0]);
	  % empty value?
	  if (strlen(value) == 0) { value = NULL; }
	  % logical 'true'
	  else if (value == "T") { value = '\1'; }
	  % logical 'false'
	  else if (value == "F") { value = '\0'; }
	  % float or integer?
	  else {
            value = is_substr(value, ".") ? atof(value) : atoi(value);
	  }
	} else { % string
	  value = strtrim_end(extr[1]);
	}
	% add to output header structure
	head = head == NULL ? struct_combine(key) : struct_combine(head, key);
        set_struct_field(head, key, value);
      }
    }
  }

  return head;
}

%
% Code from src/FITS/fits_read_key_int_frac.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_read_key_int_frac()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_read_key_int_frac}
%\synopsis{reads a keyword from a FITS file, which may be split in integer and fractional part}
%\usage{fits_read_key_int_frac(String_Type filename, key);}
%\seealso{fits_read_key}
%!%-
{
  variable filename, key;
  switch(_NARGS)
  { case 2: (filename, key) = (); }
  { help(_function_name()); return; }

  variable value;
  if(fits_key_exists(filename, key))
  {
    value = fits_read_key(filename, key);
    if(typeof(value)==String_Type)  value = atof(value);
    return value;
  }
% else
  if(fits_key_exists(filename, key+"I"))
  {
    value = fits_read_key(filename, key+"I");
    if(typeof(value)==String_Type)  value = atof(value);
    if(fits_key_exists(filename, key+"F"))
    {
      variable value2 = fits_read_key(filename, key+"F");
      if(typeof(value2)==String_Type)  value2 = atof(value2);
      value += value2;
    }
    return value;
  }
  return NULL;
}

%
% Code from src/FITS/fits_read_lc.sl
%
%%%%%%%%%%%%%%%%%%%
define fits_read_lc()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_read_lc}
%\synopsis{reads a light curve file in FITS format}
%\usage{Struct_Type lc = fits_read_lc(String_Type filename[]);}
%\description
%     Reads light curves from \code{filename}, which can be
%     a globbing expression, an array of filenames or both.
%     If there are several light curves, they will be merged
%     into one data structure with ascending times.
%
%     The function assures that the returned structure contains
%     the fields \code{time}, \code{rate} and \code{error}.
%     The time field is always converted into Modified Julian Date
%     according to the MJDREF[{I,F}], TIMEUNIT and TIMEZERO keywords.
%\qualifiers{
%\qualifier{verbose}{show the filename of the light curves when reading more than one}
%\qualifier{cut}{cut all fields but time, rate, error}
%\qualifier{rate_per_PCU}{divide count rate for RXTE-PCA light curves
%                     by number of PCUs determined from the given filterfile
%                     (*.xfl, see 'RXTE_nr_PCUs_from_filterfile').}
%\qualifier{time}{[=\code{"time"}]: name of the time field, e.g., \code{"barytime"};
%                      if \code{!= "time"}, this field is renamed \code{"time"},
%                      overwriting any previously existing \code{time} field}
%\qualifier{time_in_s}{add structure fields for time in sec, MJDref
%                     and T0 to the output structure}
%\qualifier{filename}{add filename structure field}
%\qualifier{extension}{add number of extension to be read}
%}
%\seealso{fits_read_table, fits_read_key, fits_read_key_int_frac, RXTE_nr_PCUs_from_filename}
%!%-
{
  variable filename;
  switch(_NARGS)
  { case 1: filename = (); }
  { help(_function_name()); return; }

  variable lc;
  variable verbose = qualifier_exists("verbose");

  variable filenames = glob(filename);
  if(length(filenames)>1)
  { variable lcs = Struct_Type[0];
    foreach filename (filenames)
      lcs = [lcs, fits_read_lc(filename;; __qualifiers())];
    lc = merge_struct_arrays(lcs);
    struct_filter(lc, array_sort(lc.time));
    return lc;
  }

  if(length(filenames)==0)
  {
    vmessage("warning (%s): file(s) %s not found", _function_name(), filename);
    return struct { time=Double_Type[0], rate=Double_Type[0], error=Double_Type[0] };
  }
  else
    filename = filenames[0];

  % if qualifier for extension set, use it
  if(qualifier_exists("extension")) filename = filename+"+"+string(qualifier("extension"));

  % reading light curve
  if(verbose)  vmessage("reading %s", filename);
  lc = fits_read_table(filename);
  variable fieldnames = get_struct_field_names(lc);

  % maybe use another field for the times
  variable time_field_name = qualifier("time", "time");
  if(time_field_name!="time")
    if(all(fieldnames!=time_field_name))
      vmessage(`warning (%s): field name "%s" does not exist`, _function_name(), time_field_name);
    else {
      variable otime = all(fieldnames != "time") ? NULL : @lc.time; % keep original time array for #PCU determination
      lc = struct_combine(reduce_struct(lc, time_field_name),  % remove the old name of the time field we actually want to use
			  struct { time = get_struct_field(lc, time_field_name) }  % overwrite the field called 'time'
			 );
    }
  
  variable MJDref = fits_read_key_int_frac(filename, "MJDREF");
  if(MJDref==NULL)
  { MJDref = 0;
    if(verbose)
      vmessage("warning (%s): %s does not contain MJDREF[I,F] keyword[s].", _function_name(), filename);
  }

  variable T0 = fits_read_key_int_frac(filename, "TIMEZERO");
  if(T0==NULL)  T0 = 0;
  if(qualifier_exists("time_in_s"))
    % create extra structure fields if qualifier time_in_sec is set
    lc = struct_combine(lc, struct {
      time_in_s = lc.time,
      MJDref    = [MJDref:MJDref:#length(lc.time)],
      T0        = [  T0  :  T0  :#length(lc.time)]
    });

  if(qualifier_exists("filename"))
    lc = struct_combine(lc, struct {
      filename = [filename][Integer_Type[length(lc.time)]]  % Integer_Type[n] = [0, 0, ..., 0]  is here used as index-array
    });

  variable TIMEUNIT = fits_read_key(filename, "TIMEUNIT");
  if(TIMEUNIT==NULL && any(fieldnames=="time"))
    TIMEUNIT = fits_read_key(filename, "TUNIT"+string(wherefirst(fieldnames=="time")+1));
  if(TIMEUNIT!=NULL)
    TIMEUNIT = strlow(TIMEUNIT);
  switch( TIMEUNIT )
  { case "s": lc.time = MJDref + (T0+lc.time)/86400.; }
  { case "d":
      lc.time = MJDref +  T0+lc.time;
      if(qualifier_exists("time_in_s"))  lc.time_in_s *= 86400.;
  }
  { % else:
      vmessage(`warning (%s): %s contains unknown TIMEUNIT="%S", not converting time`, _function_name(), filename, TIMEUNIT);
  }

  variable TELESCOP = "???";
  if(fits_key_exists(filename, "TELESCOP")) TELESCOP = fits_read_key(filename, "TELESCOP");
  if(fits_key_exists(filename, "TELSCOP"))  TELESCOP = fits_read_key(filename, "TELSCOP");
  switch( strup(TELESCOP) )
  { case "XTE":  % lc = struct { time, rate, error, fracexp, deadc }
      if(qualifier_exists("rate_per_PCU"))
      {
	if (access(qualifier("rate_per_PCU"), F_OK) == 0) {
  	  lc = struct_combine(lc, struct { nr_PCUs = RXTE_nr_PCUs_from_filterfile(qualifier("rate_per_PCU"), otime) });
	  if(verbose)  vmessage(" -> %d-%d PCUs", min(lc.nr_PCUs), max(lc.nr_PCUs));
	  lc.rate  /= lc.nr_PCUs;
	  lc.error /= lc.nr_PCUs;
	} else { vmessage("warning (%s): given filterfile does not exist, number of PCUs not determined", _function_name); }
      }
  }
  { case "CHANDRA":  % lc = struct { time_bin, time_min, time, time_max, counts, stat_err, count_rate, count_rate_err, exposure }
      fieldnames[where(fieldnames=="count_rate")] = "rate";
      fieldnames[where(fieldnames=="count_rate_err")] = "error";
      lc = struct_combine(rename_struct_fields(lc, fieldnames), struct { fracexp = lc.exposure/(lc.time_max-lc.time_min) });
  }
  { case "SWIFT":  % lc = struct { time, rate, error, (fracexp) }
      ;
  }
  { case "INTEGRAL":  % lc = struct { time, tot_counts, backv, backe, rate, error, fracexp, barytime }
      ;
  }
  { case "XMM":
      if(all(fieldnames!="rate") && any(fieldnames=="counts"))
      { variable inv_dt = 1./(86400*(make_hi_grid(lc.time)-lc.time));
        lc = struct { time = lc.time,
  	              rate = lc.counts * inv_dt,
	              error = sqrt(lc.counts) * inv_dt
                    };
      }
  }
  { % else:
      if( not (any(fieldnames=="time") && any(fieldnames=="rate") && any(fieldnames=="error")) )
      { vmessage("error (%s): file %s (TELESCOP=%S) does not contain the columns time, rate and error", _function_name(), filename, TELESCOP);
        return NULL;
      }
  }
  if(qualifier_exists("cut"))
     return struct { time=lc.time, rate=lc.rate, error=lc.error };
  else
     return lc;
}

%
% Code from src/FITS/fits_read_rmf.sl
%
private define dimi(arr, dim, i, j) {
  switch (dim)
  { case 1: return arr[i][j]; }
  { case 2: return arr[i,j]; }
  { message("error (fits_read_rmf): array has other dimensions than expected"); return; }
}

%%%%%%%%%%%%%%%%%%%%
define fits_read_rmf()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_read_rmf}
%\synopsis{retrieves a matrix from a compressed RMF file}
%\usage{Struct_Type rmf = fits_read_rmf(String_Type RMFfile);}
%\description
%     \code{rmf.matrix[j,i]} describes \code{rmf.ebounds.}*\code{[i]} and \code{rmf.energy.}*\code{[j]}.
%\qualifiers{
%\qualifier{check}{checks rmf-normalization: \code{rmf.matrixsum_ebounds} is the sum over all ebounds, which should be 1.}
%\qualifier{spec}{\code{rmf.whitespectrum} is the sum over all ebounds, which should be 1.}
%\qualifier{float}{use for larger RMFs to be able to be loaded in isis.}
%}
%\seealso{fits_plot_rmf}
%!%-
{
  variable RMFfile;
  switch(_NARGS)
  { case 1: RMFfile = (); }
  { help(_function_name()); return; }

  variable fields = ["matrix", "ebounds", "energy"];
  if(qualifier_exists("check"))  fields = [fields, "matrixsum_ebounds"];
  if(qualifier_exists("monospec"))  fields = [fields, "matrixsum_energy"];
  variable result = @Struct_Type(fields);

  variable ebounds = fits_read_table(RMFfile+"[EBOUNDS]");
  variable rmf;
  try { rmf = fits_read_table(RMFfile+"[MATRIX]"); } catch AnyError: { rmf = fits_read_table(RMFfile+"[SPECRESP MATRIX]"); };
  variable n = length(rmf.energ_lo);

  result.matrix = (qualifier_exists("float") ? Float_Type : Double_Type)[n, length(ebounds.channel)];

  result.matrix[*] = 0;
  variable dims = length(array_shape(rmf.n_chan));
  variable dimm = length(array_shape(rmf.matrix));

  variable tlmin = fits_read_key(RMFfile, "TLMIN4");
  if (tlmin == NULL) { % default value
    tlmin = 1;
  }

  variable group_index, iMatrix = 0;
  variable c, i;
  _for c (0, n-1, 1) {
    iMatrix = 0;
    _for group_index (0, length(dims == 1 ? rmf.n_chan[c] : rmf.n_chan[c,*])-1, 1) {
      if (qualifier_exists("verbose")) {
  	vmessage(
	  "channel %d, part %d: first = %d, n = %d\n", c, group_index,
          dimi(rmf.f_chan, dims, c, group_index),
          dimi(rmf.n_chan, dims, c, group_index)
        );
      }
      _for i (0, dimi(rmf.n_chan, dims, c, group_index)-1, 1) {
        result.matrix[c, dimi(rmf.f_chan, dims, c, group_index)+i-tlmin] += dimi(rmf.matrix, dimm, c, iMatrix);
        iMatrix++;
      }
    }
  }
  
  result.ebounds = struct { bin_lo=ebounds.e_min, bin_hi=ebounds.e_max, center=(ebounds.e_min+ebounds.e_max)/2. };
  result.energy = struct { bin_lo=rmf.energ_lo, bin_hi=rmf.energ_hi, center=(rmf.energ_lo+rmf.energ_hi)/2. };

  if (qualifier_exists("check")) {
    n = length(result.matrix[*,0]);
    result.matrixsum_ebounds = Double_Type[n];
    _for i (0, n-1, 1) {
      result.matrixsum_ebounds[i] = sum(result.matrix[i,*]);
    }
  }

  return result;
}

%
% Code from src/FITS/fits_read_unsigned_img.sl
%
define fits_read_unsigned_img()
%!%+
%\function{fits_read_unsigned_img}
%\synopsis{reads an image of unsigned integers}
%\usage{Integer_Type img[] = fits_read_unsigned_img(String_Type filename);}
%!%-
{
  variable filename;
  switch(_NARGS)
  { case 1: filename = (); }
  { help(_function_name()); return; }

  variable short_image = fits_read_img(filename);
  variable image = Integer_Type[length(short_image[*,0]),length(short_image[0,*])];
  variable pos, neg = where(short_image<0, &pos);
  image[pos] = short_image[pos];
  image[neg] = short_image[neg] + 65536;
  return image;
}

%
% Code from src/data/spectrum.datastructures/fits_save_data_model_flux.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_save_data_model_flux()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_save_data_model_flux}
%\synopsis{saves ISIS spectral data into a FITS file}
%\usage{fits_save_data_model_flux([filename[, ids]]);}
%!%-
{
  variable ids=all_data, filename=NULL;
  switch(_NARGS)
  { case 0: ; }
  { case 1: filename = (); }
  { case 2: (filename, ids) = (); }
  { help(_function_name()); return; }
  if(filename==NULL)  filename = "data_model_flux.fits";

  variable F = fits_open_file(filename, "c");
  variable id;
  foreach id ([ids])
  { variable d = get_data_counts(id);
    variable f = get_data_flux(id);
    fits_write_binary_table(F, "spectrum$id"$, struct {
      bin_lo   = d.bin_lo,
      bin_hi   = d.bin_hi,
      value    = d.value,
      err      = d.err,
      model    = get_model_counts(id).value,
      flux     = f.value/(f.bin_hi-f.bin_lo),
      flux_err = f.err / (f.bin_hi-f.bin_lo)
    }, struct {
      TUNIT1 = "A",
      TUNIT2 = "A",
      TUNIT3 = "counts/bin",
      TUNIT4 = "counts/bin",
      TUNIT5 = "counts/bin",
      TUNIT6 = "photons/s/cm^2/A",
      TUNIT7 = "photons/s/cm^2/A",
    });
  }
  fits_close_file(F);
}

%
% Code from src/FITS/fits_save_fit/fits_save_fit.sl
%
% ################################################################# %
% ####################   FITS_SAVE_FIT  ########################### %
% ################################################################# %


% =================== %
% NULL value for the FITS table
private variable null_var = " ";
private variable empty_var = "";

% field for each parameter
private variable fits_struct = ["value","lim","freeze","tie","units","fun","conf"];
% field seperator for 2d String in fits table
private variable sep_var = "\n";
% NULL value for CONF
private variable null_conf = 0;
% =================== %
public variable silent = -1;

% =================== %
define fits_conv_to_legal_char(aba_str)
% =================== %
%!%+
%\function{fits_conv_to_legal_char}
%\synopsis{converts a string into legal characters to be used as a FITS column name}
%\usage{String_Type legal_str = fits_conv_to_legal_char(String_Type str);}
%
%\seealso{fits_save_fit}
%!%-
  %
  % ********* WARNING: 'fits_save_fit' relies heavily on that function: Keep that
  %                    in mind when changing it!
  %
{
  aba_str = strtrans (aba_str,".", "");
  aba_str = strtrans (aba_str,"()", "_");
  return aba_str;
}

% =================== %
private define _fits_save_fit_extname(i,n)
  % =================== %
{
  % same extname for all files:
  return "fits_save_fit_"+string(i+1)+"-"+string(n);
}

% =================== %
define fits_save_fit_write()
% =================== %
%!%+
%\function{fits_save_fit_write}
%\synopsis{writes a FITS table with info on model and observation}
%\usage{fits_save_fit_write([, Struct/String_Type conf]);}
%\description
%
%   This function saves information about the model and the observation
%   in a FITS table. It uses the structure created by
%   fits_save_fit_struct. Using self-created structures might lead to
%   strange results. See "help fits_savs_fit" for more information.
%
%\seealso{fits_save_fit,fits_load_fit_struct}
%!%-
{
  variable name,str;
  switch (_NARGS)
  { case 2: (name, str) = (); }
  { help(_function_name()); return;  }
  % number of columns allowed by FITS
  variable _n_up = 998;
  % split if length of str exceeds this value
  variable strs = [split_struct_cols(str,_n_up)];
  variable n = length(strs);

  fits_write_binary_table(name, _fits_save_fit_extname(0,n), strs[0]);

  variable i;
  _for i(0, n-2,1)
  {
    fits_append_binary_table(name, _fits_save_fit_extname(i+1,n), strs[i+1]);
  }
  return;
}

% =================== %
private define _fits_save_fit_make_struct(p)
  % =================== %
{
  % convert the names to "FITS conform style"
  variable name = fits_conv_to_legal_char(p.name);

  % write limits into
  variable lim = Double_Type[1,2];
  lim[0,0] = p.min;
  lim[0,1] = p.max;
  % write limits into
  variable l_conf = Double_Type[1,2];
  l_conf[0,0] = null_conf;
  l_conf[0,1] = null_conf;
  % l_conf=null_conf;

  variable np = @Struct_Type(fits_struct);
  np.value = p.value;
  np.lim = lim;
  np.conf = l_conf;
  np.freeze = p.freeze;
  np.tie = [p.tie];
  np.units = [p.units]; % assures that the complete string and not only the first Char is written
  np.fun = [p.fun]; % same as above

  % write hard limits if desired
  if (qualifier_exists("hard_limits"))
  {
    variable hard_lim = Double_Type[length(p.min),2];
    hard_lim[*,0] = p.hard_min;
    hard_lim[*,1] = p.hard_max;
    np = struct_combine (np, struct{limhard=hard_lim} );
  }

  variable names = get_struct_field_names(np);
  variable n = length(names);

  % get the name of the parameter
  variable str_names = name + "_" + names;

  variable str = @Struct_Type([str_names]);

  variable i;
  _for i(0, n-1, 1)
  {
    set_struct_field(str, str_names[i], get_struct_field(np, names[i]));
  }

  return str;
}

% =================== %
private define _fits_save_fit_check_values(p)
  % =================== %
{
  p.tie = ( p.tie == NULL) ? 0 : p.tie;
  p.fun = ( p.fun == NULL) ? null_var : p.fun;
  p.hard_min = ( p.hard_min <= -DOUBLE_MAX ) ? -DOUBLE_MAX : p.hard_min;
  p.hard_max = ( p.hard_max >= DOUBLE_MAX ) ? DOUBLE_MAX : p.hard_max;
  p.units = ( p.units == "") ? null_var : p.units;
  return p;
}

% =================== %
define _fsf_add_string(str,na,info)
  % =================== %
{

  variable s = get_struct_field(str,na);
  variable t;
  variable datset = qualifier("datset",0) ;

  variable key = qualifier("key",NULL);
  ifnot (key == NULL)
  { % === %
    variable n = fits_get_num_hdus(info);

    variable i;
    _for i(0,n-1,1)
    {
      if(key=="MJDREF") { t = fits_read_key_int_frac(info+"["+string(i)+"]",key); }
      else { t = fits_read_key(info+"["+string(i)+"]",key); }
      if (t != NULL) break;
      }
    % check if the key word exists
    if (t == NULL)
    {
      if( silent == -1) vmessage(" *** Warning: key %s not contained in %s",key, info);
      t = null_var;
      if (qualifier_exists("double")) t = 0.;

      }
    % write the string
    info = t;
  }
  % === %

  % INTEGRAL speciality!!!! -> TFIRST and TLAST are set
  if ( (key == "TSTART" or key == "TSTOP") and ( typeof(info) == String_Type))  info = 0.;

  if( typeof(info) == String_Type) {
    s = s + (s==empty_var?empty_var:sep_var) + info;
    %   set_struct_field(str,na,s);
   }
  else {
    s[0,datset] = info ;
    %    set_struct_field(str,na,s);
   }
  set_struct_field(str,na,s);
  return str;
}

% =================== %
define fits_save_fit_struct()
% =================== %
%!%+
%\function{fits_save_fit_struct}
%\synopsis{saves the model and info of the observation to a structure}
%\usage{Struct_Type str = fits_save_fit_struct([, Struct/String_Type conf]);}
%\description
%
%   This function saves information about the model and the observation
%   in a structure table. Using fits_save_fit_write it can be written to a
%   fits file. See "help fits_save_fit" for more information.
%
%\seealso{fits_save_fit,fits_load_fit_write}
%!%-
{

  % get input parameters
  variable str = Struct_Type[0];
  variable conf_dat = NULL;
  variable info = qualifier("info",Struct_Type[0]);
  silent = qualifier_exists("silent") ? 1 : -1;
  switch (_NARGS)
  { case 1: (conf_dat) = (); }
  { _NARGS > 1: help(_function_name()); return;  }

  if (_typeof(conf_dat) == String_Type) conf_dat = fits_read_table(conf_dat);

  % =============  PARAMS ==================
  variable params = get_params();
  if (params == NULL)
  {
    if (silent != 1)
    {
      message(" *** Warning: No model has been loaded!");
      message(" *** No parameter values and confidence limits will be included.");
     }
   }
  else
  {

    variable param;
    foreach param(params)
    {
      % write NULL etc. correctly
      param = _fits_save_fit_check_values(param);
      % get the structures
      str = is_struct_type(str)==1?
	struct_combine(str,_fits_save_fit_make_struct(param;;__qualifiers)):
	_fits_save_fit_make_struct(param;;__qualifiers);
      }

    % =============  CONF ==================

    if (_typeof(conf_dat) == Struct_Type)
    {
      variable c,i,conf_name;
      _for i(0,length(conf_dat.name)-1,1)
      {
	conf_name=fits_conv_to_legal_char(conf_dat.name[i]);

	if ( length( where (get_struct_field_names(str) == conf_name+"_value") ) > 0)
	{

	  variable single_c = Double_Type[1,2];
	  single_c[0,0] = conf_dat.conf_min[i];
	  single_c[0,1] = conf_dat.conf_max[i];

	  set_struct_field(str,conf_name+"_value",conf_dat.value[i]);
	  set_struct_field(str,conf_name+"_conf",single_c);

	    }
	else
	{
	  vmessage(" *** Warning: confidence levels of %s could not be assigned",
		   conf_dat.name[i]);
	    }
	 }
      }

   }

  % =============  INFO ==================

  variable mi = struct{
    fit_fun=empty_var,
    tstart=0.,
    tstop=0.,
    tstart_mjd=0.,
    tstop_mjd=0.,
    exposure=0.,
    data=empty_var,
    bkg=empty_var,
    instrument=empty_var,
    target=empty_var,
    telescope=empty_var,
    filter=empty_var,
    date_obs=empty_var,
    date_end=empty_var,
    obs_id=empty_var,
    rebin=empty_var,
    notice_list=empty_var,
    chi_red=0.,
    chi=0.,
    dof=0,
    fit_statistic=empty_var,
    cwd=[getcwd],
    mjdref=0.,
    sys_err=empty_var
  };

  % write info of the data
  variable dummy_fpt;
  if (all_data == NULL){
     
     mi.exposure = Double_Type[1,1];
    mi.tstart = Double_Type[1,1]; mi.tstop = Double_Type[1,1]; mi.mjdref = Double_Type[1,1] ;
    mi.tstart_mjd = Double_Type[1,1]; mi.tstop_mjd = Double_Type[1,1];
    mi.tstart[0,0]=0.;
    mi.tstop[0,0]=0.;
    mi.tstart_mjd[0,0]=0.;
    mi.tstop_mjd[0,0]=0.;
    mi.exposure[0,0]=0.;
     
  } else {
    variable len = length(all_data);

    % ==== INTEGRAL  ====
    % check if one of the intruments is INTEGRAL
    variable integral = 0;
    variable all_dat_info = Struct_Type[len];
    _for i(0,len-1,1)
    {
      all_dat_info[i] = get_data_info(all_data[i]);
      if ( (all_dat_info[i].file != "") &&
	   (_fits_open_file(&dummy_fpt,all_dat_info[i].file,"r") == 0 ) )% make sure data files exist and is fits
      {
	if (fits_read_key(all_dat_info[i].file,"TELESCOP") == "INTEGRAL")
	  integral = 1;
       }
    }

    % merged integral does not have useful TSTART and TSTOP, TFIRST
    % and TLAST:
    % -> add more fields
    if (integral == 1)
    {
      variable mi_integ = @Struct_Type(["tfirst_mjd","tlast_mjd"]);
      set_struct_fields(mi_integ,0.,0.);
      variable tfirst_mjd = Double_Type[1,len];
      variable tlast_mjd = Double_Type[1,len];
      mi_integ.tfirst_mjd = Double_Type[1,len]; mi_integ.tlast_mjd = Double_Type[1,len] ;
     }

    % ===================

    variable exposure = Double_Type[1,len];
    mi.tstart = Double_Type[1,len]; mi.tstop = Double_Type[1,len]; mi.mjdref = Double_Type[1,len] ;
    mi.tstart_mjd = Double_Type[1,len]; mi.tstop_mjd = Double_Type[1,len];

    _for i(0,len-1,1)
    {
      variable dat_info = all_dat_info[i];
      % make sure data files exist and
      if ( access( dat_info.file, F_OK ) == 0 )
      {
	mi = _fsf_add_string(mi,"data",dat_info.file);
	mi = _fsf_add_string(mi,"bkg",dat_info.bgd_file);
	mi = _fsf_add_string(mi,"instrument",dat_info.instrument);

	% make sure data files is fits
	if( _fits_open_file(&dummy_fpt,dat_info.file,"r") == 0 )
	{
	  mi = _fsf_add_string(mi,"telescope",dat_info.file;key="TELESCOP");
	  mi = _fsf_add_string(mi,"sys_err",sys_err_array2string(get_sys_err_frac(all_data[i])));
	  mi = _fsf_add_string(mi,"date_obs",dat_info.file;key="DATE-OBS");
	  mi = _fsf_add_string(mi,"date_end",dat_info.file;key="DATE-END");
	  mi = _fsf_add_string(mi,"obs_id",dat_info.file;key="OBS_ID");
	  mi = (integral==1)
	    ? _fsf_add_string(mi,"target",dat_info.file;key="NAME")
	      : _fsf_add_string(mi,"target",dat_info.file;key="OBJECT");
	  mi = _fsf_add_string(mi,"filter",dat_info.file;key="FILTER");
	  mi = _fsf_add_string(mi,"tstart",dat_info.file;key="TSTART",datset=i,double);
	  mi = _fsf_add_string(mi,"tstop",dat_info.file;key="TSTOP",datset=i,double);
	  mi = _fsf_add_string(mi,"mjdref",dat_info.file;key="MJDREF",datset=i,double);

	  mi.tstart_mjd[0,i] = mi.tstart[0,i]/86400. + mi.mjdref[0,i];
	  mi.tstop_mjd[0,i] = mi.tstop[0,i]/86400. + mi.mjdref[0,i];
	  
	  if (integral == 1)  % ====  INTEGRAL  =====  %
	  {
	    mi_integ = _fsf_add_string(mi_integ,"tfirst_mjd",dat_info.file;key="TFIRST",datset=i,double);
	    mi_integ = _fsf_add_string(mi_integ,"tlast_mjd",dat_info.file;key="TLAST",datset=i,double);
	    if (mi_integ.tfirst_mjd[0,i] != mi_integ.tlast_mjd[0,i])
	    {
	      mi_integ.tfirst_mjd[0,i] += mi.mjdref[0,i];
	      mi_integ.tlast_mjd[0,i] += mi.mjdref[0,i];
	    }
	  }
	}
      }
      
      exposure[0,i] = get_data_exposure(all_data[i]);
      
      variable reb = null_var, notice = null_var;
      if (length(dat_info.notice_list) > 0)
      {
	reb = rebin_isis2human(dat_info.rebin);
	notice = notice_isis2human(dat_info.notice_list);
      }
      mi.rebin = mi.rebin+(i>0?sep_var:"")+reb;
      mi.notice_list = mi.notice_list+(i>0?sep_var:"")+notice;

     }
    variable target = dat_info.target;

    if (integral == 1) mi = struct_combine(mi,mi_integ);

    mi.exposure = exposure;

    mi.fit_statistic=get_fit_statistic;
    if((strtok(mi.fit_statistic,";")[0] == "chisqr") || (strtok(mi.fit_statistic,";")[0] == "cash") )
    {
      variable tmp_stat, tmp_verb;
      tmp_verb = Fit_Verbose;Fit_Verbose=-1;
      () = eval_counts(&tmp_stat);
      Fit_Verbose=tmp_verb;

      mi.chi = tmp_stat.statistic;
      mi.dof = tmp_stat.num_bins-tmp_stat.num_variable_params;
      mi.chi_red = mi.chi/mi.dof;
     }
  }
  % make sure that all Strings are written complete
  mi.data = [mi.data];mi.bkg = [mi.bkg];mi.instrument = [mi.instrument]; mi.telescope=[mi.telescope];
  mi.date_obs=[mi.date_obs];mi.date_end=[mi.date_end];mi.target=[mi.target];mi.filter=[mi.filter];
  mi.rebin=[mi.rebin];mi.notice_list=[mi.notice_list];mi.obs_id=[mi.obs_id];mi.fit_statistic=[mi.fit_statistic];
  mi.fit_fun=[mi.fit_fun];mi.sys_err=[mi.sys_err];

  % write fit function
  if (get_fit_fun()!=NULL) mi.fit_fun = [get_fit_fun()];

  if (is_struct_type(info) == 1)
  {
    mi = struct_combine(mi,info);
  }

  % write the fits file
  if (is_struct_type(str) == 1) str = struct_combine(mi,str);
  else str = mi;

  return str;
}

% =================== %
define fits_save_fit()
% =================== %
%!%+
%\function{fits_save_fit}
%\synopsis{saves the model and info of the observation to a FITS table}
%\usage{fits_save_fit(String_Type filename[, Struct/String_Type conf]);}
%\description
%
%   This function saves information about the model and the observation
%   in a FITS table. This routine combines the calls of
%   \code{fits_save_fit_struct} and \code{fits_save_fit_write}.
%
%   Additionally, confidence intervals can be given in form of a
%   structure, containig the fields
%   \code{name}:      name of the parameter like given in fit_fun(...)
%   \code{value}:     best fit value of the parameter (might have changed
%                     during error calculation)
%   \code{conf_min}:  lower confidence limit
%   \code{conf_max}:  upper confidence limit
%
%   Therefore, the conf-Structure would, e.g., look like
%
%   variable conf = struct {
%                 name = ["powerlaw(1).index","powerlaw(1).norm"],
%                 value = [2,1e-4],
%                 conf_min = [1.8,1e-5],
%                 conf_max = [2.2,2e-4]
%                 };
%
%   Alternatively, the filename of a FITS table created by
%   pvm_fit_pars or the structure returned by pvm_fit_pars can be given.
%
%
%   The values of the model are overwritten, as the error calculation
%   should only yield values equal or better than the original ones.
%
%\qualifiers{
%\qualifier{info=Struct_Type}{appends the given structure to the table}
%\qualifier{hard_limits}{also saves all hard limits of the parameters}
%\qualifier{silent}{no warnings are printed to STDOUT}
%}
%\seealso{fits_load_fit_struct,fits_write_TeX_table,save_par,pvm_fit_pars,fits_save_fit_struct,fits_write_fits_struct,fits_list_fit_pars}
%!%-
{
  % get input parameters
  variable name;
  variable str;
  variable conf_dat = NULL;
  variable info = qualifier("info",Struct_Type[0]);

  switch (_NARGS)
  { case 1: (name) = (); str = fits_save_fit_struct(;;__qualifiers);}
  { case 2: (name, conf_dat) = (); str = fits_save_fit_struct(conf_dat;;__qualifiers);}
  { help(_function_name()); return; }

  fits_save_fit_write(name,str);

  return;
}

%
% Code from src/FITS/fits_wcs_struct.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define fits_wcs_struct()
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_wcs_struct}
%\synopsis{creates a structure with a WCS that can be written to FITS file}
%\usage{Struct_Type fits_wcs_struct(String_Type filename)
%\altusage{Struct_Type fits_wcs_struct(Double_Type X[], Y[] [, String_Type xtype, ytype[, xunit, yunit]])}
%}
%\qualifiers{
%\qualifier{arrays}{return a struct { ctype=[ctype1, ctype2], ... } with arrays
%              instead of struct { ctype1=ctype1, ctype2=ctype2, ... }.
%              This form is, e.g., required by ds9_put_wcs_struct.}
%}
%\description
%    The (linear) World Coordinate System (WCS) can be read from a FITS file,
%    or can be defined from an array of \code{X} and \code{Y} values.
%!%-
{
  variable X, Y, xtype="", ytype="", xunit="", yunit="";
  variable filename=NULL, xcol="1", ycol="2";
  switch(_NARGS)
  { case 2: (X, Y) = (); }
  { case 4: (X, Y, xtype, ytype) = (); }
  { case 6: (X, Y, xtype, ytype, xunit, yunit) = (); }
  { case 1: filename = (); }
  { case 3: (filename, xcol, ycol) = (); }
  { help(_function_name()); return; }

  variable Ikeys = ["ctype", "cunit", "crval", "cdelt", "crpix"], keys = Ikeys;
  variable Tkeys = ["tctyp", "tunit", "tcrvl", "tcdlt", "tcrpx"];
  variable           ctype1,  cunit1,  crval1,  cdelt1,  crpix1,
                     ctype2,  cunit2,  crval2,  cdelt2,  crpix2;
  if(filename==NULL)
  {
    ctype1 = xtype;  cunit1 = xunit;  crval1 = X[0];  cdelt1 = X[1]-X[0];  crpix1 = 1;
    ctype2 = ytype;  cunit2 = yunit;  crval2 = Y[0];  cdelt2 = Y[1]-Y[0];  crpix2 = 1;
  }
  else
  {
    ifnot(fits_key_exists(filename, keys[0]+xcol))
    {
      keys = Tkeys;
      ifnot(fits_key_exists(filename, keys[0]+xcol))
        return NULL;
    }
    (ctype1, cunit1, crval1, cdelt1, crpix1) = fits_read_key(filename, keys[0]+xcol, keys[1]+xcol, keys[2]+xcol, keys[3]+xcol, keys[4]+xcol);
    (ctype2, cunit2, crval2, cdelt2, crpix2) = fits_read_key(filename, keys[0]+ycol, keys[1]+ycol, keys[2]+ycol, keys[3]+ycol, keys[4]+ycol);
  }

  variable s;
  if( qualifier_exists("arrays") )
  {
    s = @Struct_Type(Ikeys);
    set_struct_fields(s, [ctype1, ctype2], [cunit1, cunit2], [crval1, crval2], [cdelt1, cdelt2], [crpix1, crpix2]);
  }
  else
  {
    s = @Struct_Type([Ikeys+"1", Ikeys+"2"]);
    set_struct_fields(s, ctype1, cunit1, crval1, cdelt1, crpix1,  ctype2, cunit2, crval2, cdelt2, crpix2);
  }
  return s;
}

%
% Code from src/FITS/fits_write_arf.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_write_arf()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_write_arf}
%\synopsis{Write a FITS ancilliary response matrix file}
%\usage{fits_write_arf(arfname,arf;qualifiers);}
% \qualifiers{
% \qualifier{telescope}{telescope for the ARF}
% \qualifier{instrument}{instrument for the ARF}
% \qualifier{filter}{filter of the instrument}
% \qualifier{detnam}{name of the detector}
% \qualifier{exposure}{exposure time associated with the ARF}
% \qualifier{chantype}{see FITS ARF specification (default: PI)}
% \qualifier{origin}{who write this file (default: ECAP)}
%}
%\description
% Write a FITS compliant ancilliary response matrix
%    arfname: name of the file to be written
%    arf: structure containing the following fields:
%       bin_lo, bin_hi: energy bounds (keV)
%       area: array of length(ebounds.bin_lo) containing the effective
%              area in cm^2
%    note: the energies given MUST be the same as the input energies
%          of the corresponding response matrix!
%\seealso{fits_write_rmf}
%!%-
{
    variable arfname;
    variable arf;

    switch(_NARGS)
    { case 2: (arfname,arf)=(); }
    { help(_function_name()); return;}
    
    % add some more sanity checking here eventually
    % (e.g., energy ordering...)
    if (length(arf.bin_lo) != length(arf.bin_hi) ) {
	throw RunTimeError,"Energy bins have different length.";
    }
    if (length(arf.bin_lo) != length(arf.area) ) {
	throw RunTimeError,"Area and Energy bins have different length.";
    }

    variable arf_data=struct{ 
	ENERG_LO=arf.bin_lo, 
	ENERG_HI=arf.bin_hi, 
	SPECRESP=arf.area 
    };

    variable arf_keys=struct{
	EXTNAME ="SPECRESP",
	TELESCOP=qualifier("telescope","dummy"),
	INSTRUME=qualifier("instrument","dummy"),
	FILTER  =qualifier("filter","none"),
	DETNAM=qualifier("detnam","none"),
	HDUCLASS="OGIP",
	HDUCLAS1="RESPONSE",
	HDUCLAS2="SPECRESP",
	HDUVERS ="1.1.0",
	ORIGIN  =qualifier("origin","ECAP"),
	VERSION ="1.1.1",
	TUNIT1  ="keV",
	TUNIT2  ="keV",
	TUNIT3  ="cm**2"
    };

    if (qualifier_exists("exposure")) {
	arf_keys=struct_combine(arf_keys,
  	         struct{EXPOSURE=qualifier("exposure")});
    }

    variable hist=struct{
	history=["file written by fits_write_arf.sl"]
    };

    variable fitsfile=fits_open_file(arfname, "c");
    fits_write_binary_table(fitsfile, "SPECRESP", arf_data, arf_keys, hist);
    fits_write_chksum(fitsfile);
    fits_close_file(fitsfile);
}

%
% Code from src/FITS/fits_write_arf_diff.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_write_arf_diff()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_write_arf_diff}
%\synopsis{writes the difference of two ARFs in a corresponding FITS file}
%\usage{fits_write_arf_diff(String_Type arffile0, arffile1, arffile2);}
%\description
%    arf0  =  arf1 - arf2\n
%    The important header keywords are copied from arf1,
%    assuming that they are the are the same in arf2.
%!%-
{
  variable arffile0, arffile1, arffile2;
  switch(_NARGS)
  { case 3: (arffile0, arffile1, arffile2) = (); }
  { help(_function_name()); return; }

  if(qualifier_exists("verbose"))
    vmessage("ARF(%s) = ARF(%s) - ARF(%s)", arffile0, arffile1, arffile2);

  variable a = fits_read_table(arffile1);
  a.specresp -= fits_read_col(arffile2, "SPECRESP");
  variable keys = fits_read_key_struct(arffile1,
	"TUNIT1", "TUNIT2", "TUNIT3", "TELESCOP", "INSTRUME", "FILTER", "ARFVERSN",
	"HDUCLASS", "HDUCLAS1", "HDUCLAS2", "HDUVERS1", "HDUVERS2", "HDUVERS");

  fits_write_binary_table(arffile0, "SPECRESP", a, keys);
}

%
% Code from src/FITS/fits_write_binary_table_extensions.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_write_binary_table_extensions()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_write_binary_table_extensions}
%\synopsis{writes a binary FITS table with several extensions}
%\usage{fits_write_binary_table_extensions(filename, data1, data2, ...);}
%\description
%    \code{data1}, \code{data2}, ... are either just the data structures
%    or a list of the arguments 2, 3[, 4[, 5]] of \code{fits_write_binary_table}:
%    \code{{ extname, data[, keys[, hist]] }}
%
%    This function should usually not be used!
%    ISIS' cfitsio module allows to write several extensions
%    into a file after opening it with fits_open_file.
%\examples
%    \code{fits_write_binary_table_extensions("data.fits", struct { a1, b1 }, struct { a2, b2 });}\n
%    \code{fits_write_binary_table_extensions("data.fits",}
%    \code{                                   { "first", struct { a1, b1 } },}
%    \code{                                   { "second", struct { a2, b2 } });}\n
%    \code{fits_write_binary_table_extensions("data.fits",}
%    \code{                                   { "first", struct { a1, b1 }, struct { key11="value11"; key12="value12" } });}\n
%!%-
{
  if(_NARGS==0)  return help(_function_name());

  variable arg = __pop_list(_NARGS);
  variable FITSfilename = list_pop(arg);
  variable F = fits_open_file(FITSfilename, "c");
  variable i=1;
  foreach arg (arg)
  { switch(length(arg))
    { case 1: fits_write_binary_table(F, sprintf("ext%d", i), arg[0]); }
    { case 2: fits_write_binary_table(F, arg[0], arg[1]); }
    { case 3: fits_write_binary_table(F, arg[0], arg[1], arg[2]); }
    { case 4: fits_write_binary_table(F, arg[0], arg[1], arg[2], arg[3]); }
    i++;
  }
  fits_close_file(F);
}

%
% Code from src/FITS/fits_write_gti.sl
%
%%%%%%%%%%%%%%%%%%%%%
define fits_write_gti()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_write_gti}
%\synopsis{creates a FITS file with good time intervals}
%\usage{fits_write_gti(String_Type filename, Struct_Type gti, Double_Type MJDref);
%\altusage{fits_write_gti(String_Type filename, Double_Type start[], stop[], MJDref);}
%}
%\qualifiers{
%\qualifier{creator}{[\code{="isisscripts:fits_write_gti"}] (XMM SAS needs an arbitrary value)}
%\qualifier{date}{[\code{="1998-JAN-01"}] (XMM SAS needs an arbitrary value)}
%\qualifier{combineGTIs}{[\code{=1}]: combine intervals that adjoin each other}
%\qualifier{verbose}{[\code{=1}]: tell when intervals are combined}
%}
%\description
%     Good time intervals are organized as \code{gti} structures
%     containing arrays \code{START} and \code{STOP} of the corresponding times,
%     usually measured in s since the reference date \code{MJDref}.
%     (If \code{MJDref} is a string, the date is read from this file.)
%     Intervals that adjoin each other are combined.
%
%     The header keywords \code{creator} and \code{date} (according to the qualifiers)
%     are written to the primary header of the created FITS file.
%\seealso{fits_write_binary_table}
%!%-
{
  variable filename, gti, start, stop, MJDref;
  switch(_NARGS)
  { case 3: (filename, gti, MJDref) = (); }
  { case 4: (filename, start, stop, MJDref) = ();
            gti = struct { START=[start], STOP=[stop] };
  }
  { help(_function_name()); return; }

  variable fields = get_struct_field_names(gti);
  if(fields[0]=="start" && fields[1]=="stop")
    gti = struct { START=gti.start, STOP=gti.stop };  % for XMM SAS
  if(typeof(MJDref)==String_Type)
    MJDref = fits_read_key_int_frac(MJDref, "MJDREF");

  % sort GTIs by time
  struct_filter(gti, array_sort(gti.START));

  variable n = length(gti.START);
  if(qualifier("combineGTIs", 1) && n>0)
  { % combine intervals that adjoin each other
    start = { gti.START[0] };  % start with
    stop  = { gti.STOP[0] };   % first interval
    variable i;
    _for i (1, length(gti.START)-1, 1)
      if(gti.STOP[i-1] == gti.START[i])
        stop[-1] = gti.STOP[i];  % enlarge last interval
      else
        list_append(start, gti.START[i]),  % add new
        list_append(stop, gti.STOP[i]);    % interval
    if(qualifier("verbose", 1))
      vmessage("%% %s: %d intervals could be combined to %d intervals", _function_name(), n, length(start));
    gti = struct { START=list_to_array(start), STOP=list_to_array(stop) };
  }

  fits_write_binary_table(filename, "STDGTI", gti, struct { MJDref = MJDref });
  fits_update_key(filename+"[0]", "CREATOR", qualifier("creator", "isisscripts:fits_write_gti"));  % for XMM SAS
  fits_update_key(filename+"[0]", "DATE", qualifier("date", "1998-JAN-01"));  % dummy date to make XMM-arfgen happy
}

%
% Code from src/FITS/fits_write_image.sl
%
define fits_write_image()
%!%+
%\function{fits_write_image}
%\synopsis{writes an image to a FITS file}
%\usage{fits_write_image(FITSfile[, extname], image);
%\altusage{fits_write_image(FITSfile, extname, image, [xvalues, yvalues[, xlabel, ylabel]][, comments]);}
%}
%\qualifiers{
%\qualifier{WCS}{[=\code{""}]: world coordinate system to use, e.g. \code{"P"}}
%}
%\description
%    It is assumed that \code{x}/\code{yvalues} (if provided) are linear arrays,
%    such that \code{CRVAL = values[0]} and \code{CDELT = (values[-1]-values[0])/(length(values)-1)}.
%    \code{x}/\code{ylabel} can be a "label [unit]" string.
%\seealso{fits_write_image_hdu}
%!%-
{
  variable FITSfile, extname="image", image, xvalues=NULL, yvalues=NULL, xlabel="x", ylabel="y", comments=NULL;
  switch(_NARGS)
  { case 2: (FITSfile,          image                                            ) = (); }
  { case 3: (FITSfile, extname, image                                            ) = (); }
  { case 4: (FITSfile, extname, image,                                   comments) = (); }
  { case 5: (FITSfile, extname, image, xvalues, yvalues                          ) = (); }
  { case 6: (FITSfile, extname, image, xvalues, yvalues,                 comments) = (); }
  { case 7: (FITSfile, extname, image, xvalues, yvalues, xlabel, ylabel          ) = (); }
  { case 8: (FITSfile, extname, image, xvalues, yvalues, xlabel, ylabel, comments) = (); }
  { help(_function_name()); return; }

  variable xunit="", yunit="";
  variable m = string_matches(xlabel, `\(.*\) \[\(.*\)\]`, 1);
  if(m!=NULL)  (xlabel, xunit) = (m[1], m[2]);
  m = string_matches(ylabel, `\(.*\) \[\(.*\)\]`, 1);
  if(m!=NULL)  (ylabel, yunit) = (m[1], m[2]);

  fits_write_image_hdu(FITSfile, extname, image, NULL, struct { history=NULL, comment=comments });
  variable axis;
  variable WCS = qualifier("WCS", "");
  if(xvalues!=NULL && yvalues!=NULL)
  {
    foreach axis ([{"1", xlabel, xunit, xvalues}, {"2", ylabel, yunit, yvalues}])
    {
      variable i=axis[0], val = axis[3],  delt = (val[-1]-val[0])/(length(val)-1.);
      % world coordinate system used for, e.g., region files
      fits_update_key(FITSfile, "CTYPE"+i+WCS, axis[1], "i.e., a 'LINEAR' WCS");
      fits_update_key(FITSfile, "CUNIT"+i+WCS, axis[2], "Unit of axis $i"$);
      fits_update_key(FITSfile, "CRPIX"+i+WCS, 1,  "WCS reference pixel of ax.$i"$);
      fits_update_key(FITSfile, "CRVAL"+i+WCS, val[0],  "WCS $i-coord. at ref. pixel"$);
      fits_update_key(FITSfile, "CDELT"+i+WCS, delt,  "WCS pixel size of axis $i"$);
      % physical coordinates shown in ds9
      fits_update_key(FITSfile, "LTV$i"$, 1-val[0]/delt, "IRAF image coordinate Value");
      fits_update_key(FITSfile, "LTM${i}_$i"$, 1/delt, "IRAF image transform Matrix");
    }
    fits_write_comment(FITSfile, "image coordinate[x] = LTM1_1 * x + LTM1_2 * y + LTV1");
    fits_write_comment(FITSfile, "image coordinate[y] = LTM2_1 * x + LTM2_2 * y + LTV2");
  }
}

%
% Code from src/FITS/fits_write_pha_file.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define fits_write_pha_file()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_write_pha_file}
%\synopsis{writes a spectrum to an OGIP PHA file inserting the required header keywords}
%\usage{fits_write_pha_file(String_Type filename, Integer_Type data)}
%\altusage{fits_write_pha_file(String_Type filename, Array_Type data[, Array_Type stat_err])
%}
%
%\description
%    The 'filename' argument is the name of the output FITS file.
%
%    The 'data' argument can be either the index of a data set, in 
%    which case the isis data set is written to a file, 
%    an array containing the spectrum. 
%      If the array is of IntegerType, the spectrum contains total counts,
%      If the array is of FloatType/DoubleType, the spectrum is given in counts/s.
%    (DoubleType arrays will be type-casted to FloatType.) 
%
%    If 'data' is an array, then the statistical uncertainty is assumed
%    to be Poisson, unless the errors are given in the 3rd (optional) 
%    argument of the function, 'stat_err'. Note that this argument is
%    mandatory for the case of count rate spectra.
%    If 'data' is an index to a data set, then the errors are taken 
%    directly from the data set.
%
%\qualifiers{
%\qualifier{TELESCOP}{the "telescope" (mission/satellite name) ["unknown"]}
%\qualifier{INSTRUME}{the instrument/detector ["unknown"]}
%\qualifier{FILTER}{the instrument filter in use (if any) ["none"]}
%\qualifier{EXPOSURE}{the integration time (in seconds) for the PHA data
%                 (assumed to be corrected for deadtime, data drop-outs etc. ) [1.0]}
%\qualifier{AREASCAL}{nominal effective area [1.0]}
%\qualifier{BACKFILE}{the name of the corresponding background file (if any) ["none"]}
%\qualifier{BACKSCAL}{background scale factor [1.0]}
%\qualifier{CORRFILE}{the name of the corresponding correction file (if any) ["none"]}
%\qualifier{CORRSCAL}{the correction scaling factor [1.0]}
%\qualifier{RESPFILE}{the name of the corresponding (default) redistribution matrix file ["none"]}
%\qualifier{ANCRFILE}{the name of the corresponding (default) ancillary response file ["none"]}
%\qualifier{HDUCLASS}{should contain the string "OGIP" to indicate that this is an OGIP style file ["OGIP"]}
%\qualifier{HDUCLAS1}{should contain the string "SPECTRUM" to indicate this is a spectrum ["SPECTRUM"]}
%\qualifier{HDUCLAS2}{indicating the type of data stored: "TOTAL", "NET", "BKG" ["TOTAL"]}
%\qualifier{HDUVERS1}{the version number of the format ["1.2.1"]}
%\qualifier{CHANTYPE}{whether the channels used in the file have been corrected in anyway,
%                 values: "PHA" or "PI" (see also CAL/GEN/92-002, George et al. 1992, Section 7)
%                 ["PHA"]}
%\qualifier{start_channel}{start value of 'channel' column [1]}
%}
%
%   \seealso{Definition of PHA FITS format: OGIP/92-007 and OGIP/92-007a}
%!%-
{
    % Get the start value for the channel column. 
    % If this qualifier is not set the channel start at unity.
    variable start_channel = qualifier("START_CHANNEL", 1);
    
    % Set the required header keywords getting the values from the function qualifiers.
    variable keys = struct {
	% mandatory:
	TELESCOP = qualifier("TELESCOP"  , "unknown"),
	INSTRUME = qualifier("INSTRUME"  , "unknown"),
	FILTER   = qualifier("FILTER"    , "none"),
	EXPOSURE = qualifier("EXPOSURE"  , 1.0),
	AREASCAL = qualifier("AREASCAL"  , 1.0),
	BACKFILE = qualifier("BACKFILE"  , "none"),
	BACKSCAL = qualifier("BACKSCAL"  , 1.0),
	CORRFILE = qualifier("CORRFILE"  , "none"),
	CORRSCAL = qualifier("CORRSCAL"  , 1.0),
	RESPFILE = qualifier("RESPFILE"  , "none"),
	ANCRFILE = qualifier("ANCRFILE"  , "none"),
	HDUCLASS = qualifier("HDUCLASS"  , "OGIP"),
	HDUCLAS1 = qualifier("HDUCLAS1"  , "SPECTRUM"),
	HDUVERS  = qualifier("HDUVERS"   , "1.2.1"),
	CHANTYPE = qualifier("CHANTYPE"  , "PHA"),
	DETCHANS,
	% optional:
	CREATOR  = "{fits_write_pha_file} v{1.0.0}",
	HDUCLAS2 = qualifier("HDUCLAS2"  , "TOTAL"),
	HDUCLAS3 = "COUNTS",
	GROUPING = 0,
	% self-defined, leave only in for backwards compatibility
	AREAFILE = qualifier("AREAFILE"  , "none")
    };

    % Get the values of the function parameters:
    variable filename, data, stat_err,poisserr,staterr;
    switch(_NARGS) { 
       case 3:  % 'stat_err' is specified
	  (filename, data, stat_err) = ();
	  if (typeof(data)==Integer_Type) {
	      ()=fprintf(stderr,"stat_err argument is only allowed if data is an array.\n");
	      throw SyntaxError;
	  }
	  poisserr=0;  % false
	  staterr = 1;  % true
      }
      { 
	  case 2:  % 'stat_err' is not specified.
	  (filename, data) = ();
	  if (typeof(data)==Array_Type) {
	      poisserr=1; % true
	      staterr = 0;
	  }
      }

      % Wrong number of parameters: show help information
      { help(_function_name()); return; }

    % TODO: insert optional PHA header keywords


    % Define the data structure for the output to the FITS table:
    variable SPECTRUM = struct {CHANNEL, COUNTS};

    % If 'data' is just an integer number, obtain spectral
    % data from the loaded spectrum
    if (typeof(data) == Integer_Type)  {
	poisserr = 0;  % false
	staterr = 1;  % true

	variable indx=data;
	data = get_data_counts(indx);
	      
	if(qualifier_exists("Angstrom"))  {
	    keys = struct_combine(keys, struct { TUNIT3 = "Angstrom", TUNIT4 = "Angstrom" });
	} else {
	    data = _A(data);
	    keys = struct_combine(keys, struct { TUNIT3 = "keV", TUNIT4 = "keV" });
	}

	stat_err=data.err;
	
	SPECTRUM = struct_combine(SPECTRUM, struct { BIN_LO = data.bin_lo, BIN_HI = data.bin_hi });
	data = data.value;
	if(max(data mod 1)==0) data = typecast(data, Integer_Type);

	keys.EXPOSURE=qualifier("EXPOSURE",get_data_exposure(indx));
	keys.BACKSCAL=qualifier("BACKSCAL",get_data_backscale(indx));

	variable info=get_data_info(indx);

	% get instrument from arf since it is not set after a fakeit
	if (info.arfs[0]>0) {
	    keys.ANCRFILE=qualifier("ANCRFILE",get_arf_info(info.arfs[0]).file);
	    keys.INSTRUME=qualifier("INSTRUME",get_arf_info(info.arfs[0]).instrument);
	}
	if (info.rmfs[0]>0) {
	    keys.RESPFILE=qualifier("RESPFILE",get_rmf_info(info.rmfs[0]).arg_string);
	}
	% as an improvement we should add a qualifier that allows us to
	% also write the background assigned to a data set in the same step
	keys.BACKFILE=qualifier("BACKFILE",info.bgd_file);

    } else if (typeof(data) == Array_Type) {

	% If data is a floating point array, then we have
	% a rate spectrum and a spectral error column is
	% mandatory
	if ((typeof(data[0]) == Float_Type) || (typeof(data[0]) == Double_Type))  {
	    data = typecast(data, Float_Type);
	    keys = struct_combine(keys, struct {TTYPE2 = "RATE"});  
	    keys.HDUCLAS3 = "RATE";
	    if(poisserr) {
		()=fprintf(stderr,"Error: rate spectra cannot have Poisson errors\n");
		throw SyntaxError;
	    }
	} else if (typeof(data[0]) != Integer_Type) {
	    ()=fprintf(stderr, "Error: 'data' has unsupported data type (must be an array of either Float, Double, or Integer)!\n");
	    throw SyntaxError;
	}
    } else {
	()=fprintf(stderr, "Error:' data' must be either an array or an integer type!\n");
	throw SyntaxError;
    }
    SPECTRUM.COUNTS = data;

    if (staterr == 1) {
	if ((typeof(stat_err[0]) == Float_Type) || (typeof(stat_err[0]) == Double_Type)) {
	    stat_err = typecast(stat_err, Float_Type);
	    SPECTRUM=struct_combine(SPECTRUM,struct{STAT_ERR=stat_err});
	} else {
	    ()=fprintf(stderr, "Error: 'stat_err' has unsupported data type (must be an array of either Float, Double)!\n");
	    throw SyntaxError;
	}
    }

    % Set the first column of the spectrum table (PHA channels):
    keys.DETCHANS = length(data);   % number of channels
    SPECTRUM.CHANNEL = start_channel + [0:keys.DETCHANS-1];
    % Append TLMINnnn and TLMAXnnn keywords if channel does NOT start with unity:
    if (start_channel != 1) {
	keys = struct_combine(keys, 
	struct { TLMIN001 = start_channel, 
	TLMAX001 = start_channel+keys.DETCHANS-1 }
	);
    }

    % Write the spectrum to the FITS file:
    fits_write_binary_table(filename, "SPECTRUM", SPECTRUM, keys);

    % update logicals (I do not know an easier way)
    variable fd=fits_open_file(filename,"w");
    fits_movabs_hdu(fd,2); % 1st extension is the spectrum
    fits_update_logical(fd,"POISSERR",poisserr,"Poisson errors appropriate?");
    fits_update_logical(fd,"STAT_ERR",staterr,"Assume statistical errors?");
    fits_update_key(fd,"SYS_ERR",0.0,"no global systematic error");
    fits_close_file(fd);
}

%
% Code from src/FITS/fits_write_rmf.sl
%
define fits_write_rmf()
%!%+
%\function{fits_write_rmf}
%\synopsis{Write a FITS response matrix file}
%\usage{fits_write_rmf(rmfname,rmf;qualifiers);}
% \qualifiers{
% \qualifier{telescope}{telescope for the response}
% \qualifier{instrument}{instrument for the response}
% \qualifier{filter}{filter of the instrument}
% \qualifier{detnam}{name of the detector}
% \qualifier{effarea}{effective area of the detector (default: 1cm**2)}
% \qualifier{lo_thresh}{lower threshold of the matrix, i.e., values below
%       this number have been set to zero. Default: 0.}
% \qualifier{chantype}{see FITS RMF specification (default: PI)}
% \qualifier{channel}{array of channel numbers for the ebounds extension 
%     (default: channels are assumed to start at 1)}
% \qualifier{origin}{who write this file (default: ECAP)}
% \qualifier{constantwidth}{write data with width constantwidth around 
%       maximum of the matrix.}
%}
%\description
% Write a FITS compliant response matrix
%    rmfname: name of the file to be written
%    rmf: structure containing the following fields:
%       ebounds (bin_lo, bin_hi): channel energies of the response matrix
%       energy (bin_lo, bin_hi): input energies of the response matrix
% The response matrix can be given in two ways. For most decent resolution
% instruments, define the matrix as
%       matrix[energy,channel]: with dimensions matrix[length(energy),length(ebounds)]
% If this does not work (this is the case once the total array size exceeds
% the limits imposed by s-lang), then define matrix as an Array of Arrays:
%       matrix=Array_Type[length(energy)];
% and assign an Array_Type[length(ebounds)] to each element of matrix
% 
% Note: High resolution matrices can be VERY large. This code does not yet 
% write variable length arrays due to time reasons. As a work around it
% is possible just to write information around the diagonal using the
% constantwidth qualifier.
%\seealso{fits_write_arf,fits_read_rmf}
%!%-
{
    variable rmfname;
    variable rmf;

    switch(_NARGS)
    { case 2: (rmfname,rmf)=(); }
    { help(_function_name()); return;}
    
    % add some sanity checking here eventually...

    variable nch=length(rmf.ebounds.bin_lo);
    variable nin=length(rmf.energy.bin_lo);

    variable channel=qualifier("channel",[1:nch]);
    
    if (length(channel)!=nch) {
	throw RunTimeError,"Channel must have same length as number of channels.";
    }

    variable masiz=array_shape(rmf.matrix);
    variable is2d=(length(masiz)==2);
    if (is2d) {
	% 2D matrix
	if (masiz[0] !=nin) {
	    throw RunTimeError,"Matrix: 1st dimension has wrong length";
	}
	if (masiz[1] != nch) {
	    throw RunTimeError,"Matrix: 2nd dimension has wrong length";
	}
    } else {
	% rmf.matrix is an array of arrays
	if (masiz[0] != nin) {
	    throw RunTimeError,"Matrix: Number of input channels is wrong";
	}
	% checking of the number of channels is deferred to below
    }
    
    %
    % build the EBOUNDS extension
    %

    variable eb_data=struct{
	CHANNEL=channel,
	E_MIN=rmf.ebounds.bin_lo,
	E_MAX=rmf.ebounds.bin_hi
    };

    variable eb_keys=struct{
	EXTNAME ="EBOUNDS",
	TELESCOP=qualifier("telescope","dummy"),
	INSTRUME=qualifier("instrument","dummy"),
	FILTER  =qualifier("filter","none"),
	CHANTYPE=qualifier("chantype","PI"),
	DETCHANS=nch,
	HDUCLASS="OGIP",
	HDUCLAS1="RESPONSE",
	HDUCLAS2="EBOUNDS",
	HDUVERS ="1.2.0",
	ORIGIN  =qualifier("origin","ECAP"),
	VERSION ="1.1.1",
	TUNIT2  ="keV",
	TUNIT3  ="keV"
    };
	
    %
    % build the RMF extension
    %

    variable rmf_keys=struct{
	EXTNAME ="MATRIX",
	TELESCOP=qualifier("telescope","dummy"),
	INSTRUME=qualifier("instrument","dummy"),
	FILTER  =qualifier("filter","none"),
	CHANTYPE=qualifier("chantype","PI"),
	DETCHANS=nch,
	LO_THRES=qualifier("lo_thresh",0.),
	DETNAM=qualifier("detnam","none"),
	EFFAREA=qualifier("effarea","1."),
	NUMGRP=-1,
	NUMELT=-1,
	HDUCLASS="OGIP",
	HDUCLAS1="RESPONSE",
	HDUCLAS2="RSP_MATRIX",
	HDUCLAS3="REDIST",
	HDUVERS ="1.3.0",
	ORIGIN  =qualifier("origin","ECAP"),
	VERSION ="1.1.1",
	TLMIN4  =min(channel),
	TLMAX4  =max(channel),
	TUNIT1  ="keV",
	TUNIT2  ="keV"
    };

	
    variable rmf_data;

    % if maxwidth exists, then 
    if (qualifier_exists("constantwidth")) {
	% constant width response matrix

	if (qualifier("constantwidth")<=0) {
	    throw RunTimeError,"constantwidth must be positive";
	}

	% make sure that width does not exceed the full matrix
        variable maxwidth=min([qualifier("constantwidth"),nch]);

	% set the optimized matrix
	rmf_data=struct{
	    ENERG_LO=rmf.energy.bin_lo,
	    ENERG_HI=rmf.energy.bin_hi,
	    N_GRP   =Integer_Type[nin],
	    F_CHAN  =Integer_Type[nin],
	    N_CHAN  =Integer_Type[nin],
	    MATRIX  =Double_Type[nin,maxwidth]
	};

	% cut out data around the maximum of the matrix
	variable ii;
	for(ii=0;ii<nin;ii++) {
	    variable row;
	    if (is2d) {
		row=rmf.matrix[ii,*];
	    } else {
		row=rmf.matrix[ii][*];
	    }
	    variable ndx=wherefirstmax(row);
	    variable istart=max([ndx[0]-maxwidth/2,0]);
	    variable iend=istart+maxwidth-1;
	    if (iend>=nch) {
		iend=nch-1;
		istart=iend-maxwidth+1;
	    }
	    rmf_data.N_GRP[ii]=1;
	    rmf_data.F_CHAN[ii]=channel[istart];
	    rmf_data.N_CHAN[ii]=maxwidth;
	    rmf_data.MATRIX[ii,*]=row[[istart:iend]];
	}

    } else {
	% write the unoptimized matrix
	if (not is2d) {
	    % this will probably fail - in the case we use an array of arrays
	    % we really should write the matrix row by row...
	    % but then, people who write unoptimized matrices should be punished
	    % anyway...
	    variable mat=Double_Type[nin,nch];
	    for (ii=0; ii<nin;ii++) {
		mat[ii,*]=rmf.matrix[ii][*];
	    }
	    rmf_data=struct{
		ENERG_LO=rmf.energy.bin_lo,
		ENERG_HI=rmf.energy.bin_hi,
		N_GRP   =Integer_Type[nin]+1,
		F_CHAN  =Integer_Type[nin]+channel[0],
		N_CHAN  =Integer_Type[nin]+nch,
		MATRIX  =rmf.matrix
	    };
	    
	} else {
	    rmf_data=struct{
		ENERG_LO=rmf.energy.bin_lo,
		ENERG_HI=rmf.energy.bin_hi,
		N_GRP   =Integer_Type[nin]+1,
		F_CHAN  =Integer_Type[nin]+channel[0],
		N_CHAN  =Integer_Type[nin]+nch,
		MATRIX  =rmf.matrix
	    };
	}
    }

    variable hist=struct{
	history=["file written by fits_write_rmf.sl"]
    };

    rmf_keys.NUMGRP=sum(rmf_data.N_GRP);
    rmf_keys.NUMELT=sum(rmf_data.N_CHAN);

    % Store the RMF in a FITS file.
    variable fitsfile=fits_open_file(rmfname, "c");
    fits_write_binary_table(fitsfile, "EBOUNDS", eb_data, eb_keys, hist);
    fits_write_chksum(fitsfile);
    fits_write_binary_table(fitsfile, "MATRIX", rmf_data, rmf_keys, hist);
    fits_write_chksum(fitsfile);
    fits_close_file(fitsfile);
}

%
% Code from src/sixte/fits_write_sixte_vignetting.sl
%

private define get_vignet_keys(nenergies, noffsetangles, nrot){

   variable version = qualifier("version",strftime("%Y%m%d"));
   
   variable keys = struct {
      HDUCLASS = "OGIP",
      HDUCLAS1 = "RESPONSE",
      HDUVERS1 = "1.0.0",
      HDUCLAS2 = "VIGNET",
      HDUVERS2 = "1.1.0",
      VERSION  = version,
      
      MISSION  = qualifier("mission","none"),
      TELESCOP = qualifier("telescope","none"),
      DETNAM   = qualifier("detnam","none"),
      INSTRUME = qualifier("instrument","none"),
      
      TUNIT1   = "keV",
      TUNIT2   = "keV",
      TUNIT3   = "keV",
      TUNIT4   = "degree",
      TUNIT5   = "degree",
      TDIM6    = sprintf("(%d,%d,%d)", nenergies, noffsetangles, nrot)
     };
   
     return keys;

}

private define get_vignet_data(nenergies, noffsetangles, nrot){

   %      VIGNET   = Float_Type[1, length(energies)*length(offaxangles)*nPhi] 
   return struct {
      ENERG_LO = Float_Type[1, nenergies],
      ENERG_HI = Float_Type[1, nenergies],
      ENERGY   = Float_Type[1, nenergies],
      THETA    = Float_Type[1, noffsetangles],
      PHI      = Float_Type[1, nrot],
      VIGNET   = Float_Type[1, nrot, noffsetangles, nenergies] %% definition is not compliant with OGIP, but that's how we read it in sixte ...
   };
}

%%%%%%%%%%%%%%%%%%%%%%%
define fits_write_sixte_vignetting(){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{fits_write_sixte_vignetting}
%\synopsis{write a vignetting file, which can be used in a Sixte instrument configuration XML file.}
%\usage{fits_write_sixte_vignetting((string) filename, energies, offaxangles [, phi = 0]);}
%\description
%  energies given in keV, offaxis angles given in arcmin
%!%-

   variable fname, energies, offaxangles, vign;
   switch(_NARGS)
   { case 4: (fname, energies, offaxangles, vign) = (); }
   { help(_function_name()); return; }
   
   variable nener = length(energies);
   variable nthet = length(offaxangles);
   variable nphi = 1;  %% currently not supported by sixte
   
   variable vignetting_data = get_vignet_data(nener,nthet,nphi);   
   variable keys=get_vignet_keys(nener, nthet, nphi);
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % Fill the data structure:
  
   vignetting_data.THETA[0,*] = offaxangles*1./60.;
   vignetting_data.PHI[0, 0]  = 0.;
   vignetting_data.ENERG_LO[0,  0] = 0.0;
   vignetting_data.ENERG_HI[0, -1] = 20.0;
   
   
   variable ii;
   for (ii=0; ii<length(energies); ii++) {
      vignetting_data.ENERG_HI[0, ii-1] = 0.5*(energies[ii-1]+energies[ii]);
      vignetting_data.ENERG_LO[0,   ii] = 0.5*(energies[ii-1]+energies[ii]);
      vignetting_data.ENERGY[0,ii] = energies[ii];
      
      
      vignetting_data.VIGNET[0,0,*,ii]=vign[ii,*]/vign[ii,0]; %%.area[jj]/vig.area[0];
      print(vignetting_data.VIGNET[0,0,*,ii]);
      
      % for(jj=0; jj<length(offaxangles); jj++){
      %   vignetting_data.VIGNET[0,ii+jj*length(energies)]=vign[ii][*]/vign[ii][0]; %%.area[jj]/vig.area[0];
      % }
   }
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%					 
  % Store the data in the FITS file:
  variable fitsfile=fits_open_file(fname, "c"); % c => create file (delete, if already exists)
  fits_write_binary_table(fitsfile, "VIGNET", vignetting_data, keys);
  fits_close_file(fitsfile);

   
}
%
% Code from src/FITS/fits_write_tex_table.sl
%
define fits_write_tex_table()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
%!%+
%\function{fits_write_tex_table}
%\synopsis{creates a TeX table ready for input in your document.tex
%          file.}
%
%\usage{fits_write_tex_table(String_Type inputFile);}
%
%\qualifiers{
%\qualifier{pars}{model names given as an array}
%\qualifier{exclude}{parameters to exclude. One can give whole
%                  parameter name as is written in the window output, or just a main
%                  part of it (see examples below).}
%\qualifier{extraInfo}{extra information such as target name,
%                  exposure, etc., the input name is given in
%                  the fits header. This qualifier can lead to
%                  nasty look of your table. }
%\qualifier{texMulti}{tries to fix long extraInfo output in your table.}                  
%\qualifier{everyPar}{parameter name as given can be used as a
%                  qualifier itself for changing:
%                  name, digits, factor and sciMode (see
%                  TeX_value_pm_error on how to use them).
%                  Example:
%                  powerlaw_1_PhoIndex_value={"name","$\\Gamma$","digits",3}
%                  }
%\qualifier{sci}{change scientific output in one step, for
%                  an easier look of the value and its errors in the table.
%                  sci=0   will give you maximally correct output.
%                  sci=1   will also give you nice output, but may not
%                  care about last significant digit.}
%\qualifier{colNames}{give new names for your output columns. Applicable
%                  only when all names change.}  
%\qualifier{fullStat}{prints also chi^2_red values.}
%\qualifier{silent}{No output generated.}
%\qualifier{flip}{flips the table.}
%\qualifier{pdf}{produces pdfout.pdf file for a quick look at
%                  your table.}
%\qualifier{output}{TeX output name, default is "default.tex"}
%}                         
%\description
%    - \code{inFile}   input fits file produced by fits_save_fit.
%                  In the case of several files(in other words
%                  fits for the same model) produced by fits_save_fit,
%                  first add the files with fits_add_fit and use it in 
%                  fits_write_tex_table.
%
%                  Parameter names, digits, factor output have
%                  all default values, but that can be changed
%                  with the last mentioned qualifer in the list
%                  above. 
%
%      NOTE: the function is still under development. Hence, it may
%      not be applicable to all available Xspec or local models.
%      If issues emerge, contact refiz.duro@sternwarte.uni-erlangen.de.   
%
%                           
%\example
%
%   fits_write_tex_table("input.fits";
%                        pars=["cutoffpl","reflionx","diskbb","constant"]
%                       ,exclude=["diskbb_1_tin_value","norm"]
%                       ,constant_1_factor_value=["name","$c_\\mathrm{PCA}$"]
%                       ,output="my_out.tex"
%                       .extraInfo=["target","instrument"]
%                       ,target={"name","Source"}
%                       ,flip
%                       ,pdf);
%
%      will:
%         1.  use input fits_save_fit file "input.fits"
%         2.  look for components "cutoffpl" ,"reflionx", "diskbb" and
%             "constant" in your main model
%         3.  exclude parameter "diskbb_1_tin_value" and all parameters
%             with "norm" in the parameter name
%         4.  change name of now qualifier (actually a parameter)
%             "constant_1_factor_value" to $c_\\mathrm{PCA}$
%         5.  write TeX output to my_out.tex
%         6.  include target and instrument information
%         7.  change name of "target" to "Source"
%         8.  flip your table
%         9.  produce a pdfout.pdf
%
%\seealso{fits_save_fit, fits_add_fit,  TeX_value_pm_error}%
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
variable inFile;
   
   switch(_NARGS)
     {case 1 : inFile=();}
     {help (_function_name());return;}
	


   %ASSOCIATIVE ARRAY
   variable assoc=Assoc_Type[Struct_Type];
   %PHABS
   assoc["phabs_nh"]=struct{name=`$N_{\mathrm{H}}$`,unit=`[$10^{22}\mathrm{cm}^{-2}$]`,digits=1};
   %WABS
   assoc["wabs_nh"]=struct{name=`$N_{\mathrm{H}}$`,unit=`[$10^{22}\mathrm{cm}^{-2}$]`,digits=1};
   %PCFABS
   assoc["pcfabs_nh"]=struct{name=`$N_{\mathrm{H}}$`,unit=`[$10^{22}\mathrm{cm}^{-2}$]`,digits=1};
   assoc["pcfabs_cvrfract"]=struct{name=`$Abs_{\mathrm{frac}}$`,digits=1};
   %APPEC
   assoc["apec_norm"]=struct{name=`$A_{\mathrm{apec}}$`,digits=1};
   assoc["apec_kt"]=struct{name=`$kT_\mathrm{apec}$`,unit=`[keV]`,digits=1};
   assoc["apec_abundanc"]=struct{name=`$\mathrm{A}/\mathrm{A}_\odot$`,digits=1};
   assoc["apec_redshift"]=struct{name=`$z$`,digits=1};
   %BAPEC
   assoc["bapec_norm"]=struct{name=`$A_\mathrm{bapec}$`,digits=1};
   assoc["bapec_kt"]=struct{name=`$kT_\mathrm{bapec}$`,unit=`[keV]`,digits=1};
   assoc["bapec_abundanc"]=struct{name=`$\mathrm{A}/\mathrm{A}_\odot$`,digits=1};
   assoc["bapec_redshift"]=struct{name=`$z$`,digits=1};
   assoc["bapec_velocity"]=struct{name=`$v$`,digits=1};
   %TBNEW
   assoc["tbnew_nh"]=struct{name=`$N_{\mathrm{H}}$`,unit=`[$10^{22}\mathrm{cm}^{-2}$]`,digits=3};
   assoc["tbnew_he"]=struct{name=`${\mathrm{He}}$`,digits=2};
   assoc["tbnew_c"]=struct{name=`${\mathrm{C}}$`,digits=2};
   assoc["tbnew_n"]=struct{name=`${\mathrm{N}}$`,digits=2};
   assoc["tbnew_o"]=struct{name=`${\mathrm{O}}$`,digits=2};
   assoc["tbnew_ne"]=struct{name=`${\mathrm{Ne}}$`,digits=2};
   assoc["tbnew_na"]=struct{name=`${\mathrm{Na}}$`,digits=2};
   assoc["tbnew_mg"]=struct{name=`${\mathrm{Mg}}$`,digits=2};
   assoc["tbnew_al"]=struct{name=`${\mathrm{Al}}$`,digits=2};
   assoc["tbnew_si"]=struct{name=`${\mathrm{Si}}$`,digits=2};
   assoc["tbnew_s"]=struct{name=`${\mathrm{S}}$`,digits=2};
   assoc["tbnew_cl"]=struct{name=`${\mathrm{Cl}}$`,digits=2};
   assoc["tbnew_ar"]=struct{name=`${\mathrm{Ar}}$`,digits=2};
   assoc["tbnew_ca"]=struct{name=`${\mathrm{Ca}}$`,digits=2};
   assoc["tbnew_cr"]=struct{name=`${\mathrm{Cr}}$`,digits=2};
   assoc["tbnew_fe"]=struct{name=`${\mathrm{Fe}}$`,digits=2};
   assoc["tbnew_co"]=struct{name=`${\mathrm{Co}}$`,digits=2};
   assoc["tbnew_ni"]=struct{name=`${\mathrm{Ni}}$`,digits=2};
   assoc["tbnew_h2"]=struct{name=`${\mathrm{H2}}$`,digits=2};
   assoc["tbnew_rho"]=struct{name=`${\mathrm{rho}}$`,digits=2};%??
   assoc["tbnew_amin"]=struct{name=`${\mathrm{amin}}$`,digits=2};%??
   assoc["tbnew_amax"]=struct{name=`${\mathrm{amax}}$`,digits=2};%??
   assoc["tbnew_pl"]=struct{name=`${\mathrm{PL}}$`,digits=2};
   assoc["tbnew_h_dep"]=struct{name=`${\mathrm{N_\mathrm{H\_dep}}}$`,digits=2};
   % assoc["tbnew_he_dep"]=struct{name=`${\mathrm{He\_dep}}$`,digits=2};
   % assoc["tbnew_c_dep"]=struct{name=`${\mathrm{C\_dep}}$`,digits=2};
   % assoc["tbnew_n_dep"]=struct{name=`${\mathrm{N\_dep}}$`,digits=2};
   % assoc["tbnew_o_dep"]=struct{name=`${\mathrm{O\_dep}}$`,digits=2};
   % assoc["tbnew_ne_dep"]=struct{name=`${\mathrm{Ne\_dep}}$`,digits=2};
   % assoc["tbnew_na_dep"]=struct{name=`${\mathrm{Na\_dep}}$`,digits=2};
   % assoc["tbnew_mg_dep"]=struct{name=`${\mathrm{Mg\_dep}}$`,digits=2};
   % assoc["tbnew_al_dep"]=struct{name=`${\mathrm{Al\_dep}}$`,digits=2};
   % assoc["tbnew_si_dep"]=struct{name=`${\mathrm{Si\_dep}}$`,digits=2};
   % assoc["tbnew_s_dep"]=struct{name=`${\mathrm{S\_dep}}$`,digits=2};
   % assoc["tbnew_cl_dep"]=struct{name=`${\mathrm{Cl\_dep}}$`,digits=2};
   % assoc["tbnew_ar_dep"]=struct{name=`${\mathrm{Ar\_dep}}$`,digits=2};
   % assoc["tbnew_ca_dep"]=struct{name=`${\mathrm{Ca\_dep}}$`,digits=2};
   % assoc["tbnew_cr_dep"]=struct{name=`${\mathrm{Cr\_dep}}$`,digits=2};
   % assoc["tbnew_fe_dep"]=struct{name=`${\mathrm{Fe\_dep}}$`,digits=2};
   % assoc["tbnew_co_dep"]=struct{name=`${\mathrm{Co\_dep}}$`,digits=2};
   % assoc["tbnew_ni_dep"]=struct{name=`${\mathrm{Ni\_dep}}$`,digits=2};
   assoc["tbnew_redshift"]=struct{name=`${z}$`,digits=2};
   %CUTOFF POWER
   assoc["cutoffpl_phoindex"]=struct{name=`$\Gamma_\mathrm{pl}$`,digits=2};
   assoc["cutoffpl_norm"]=struct{name=`$A_\mathrm{pl}$`,digits=2};
   assoc["cutoffpl_highecut"]=struct{name=`$E_\mathrm{{fold}}$`,digits=1,unit=`[keV]`,sciMode=2};
   %POWER LAW
   assoc["powerlaw_phoindex"]=struct{name=`$\Gamma_\mathrm{pow}$`,digits=2};
   assoc["powerlaw_norm"]=struct{name=`$A_\mathrm{pow}$`,factor=1e4,digits=2};
   %BKN POWER LAW
   assoc["bknpower_phoindx1"]=struct{name=`$\Gamma_\mathrm{1}$`,digits=2};
   assoc["bknpower_norm"]=struct{name=`$A_\mathrm{bkn}$`,digits=2};
   assoc["bknpower_breake"]=struct{name=`$E_\mathrm{pow}$`,unit=`[keV]`,digits=1};
   assoc["bknpower_phoindx2"]=struct{name=`$\Gamma_\mathrm{2}$`,digits=2};
   %BKN2POW
   assoc["bkn2pow_norm"]=struct{name=`$A_\mathrm{bkn}$`,digits=2};
   assoc["bkn2pow_phoindx1"]=struct{name=`$\Gamma_\mathrm{1}$`,digits=2};
   assoc["bkn2pow_breake1"]=struct{name=`$E_\mathrm{1}$`,digits=2};
   assoc["bkn2pow_phoindx2"]=struct{name=`$\Gamma_\mathrm{2}$`,digits=2};
   assoc["bkn2pow_breake2"]=struct{name=`$E_\mathrm{2}$`,digits=2};
   assoc["bkn2pow_phoindx3"]=struct{name=`$\Gamma_\mathrm{3}$`,digits=2};
   %EGAUSS
   assoc["egauss_area"]=struct{name=`$F_\mathrm{6.4\,keV}$`,unit=`[cgs]`,digits=1};%,factor=1e5,,`$[\mathrm{phot}\,\mathrm{s}^{-1}\, \mathrm{cm}^{-1}]$`
   assoc["egauss_center"]=struct{name=`$E_\mathrm{gauss}$`,unit=`[keV]`,digits=2};
   assoc["egauss_sigma"]=struct{name=`$E_\mathrm{\sigma}$`,unit=`[keV]`,digits=2,factor=1e6};
   %GAUSSIAN
   assoc["gaussian_norm"]=struct{name=`$A_\mathrm{gauss}$`,digits=1};%,factor=1e5,,`$[\mathrm{phot}\,\mathrm{s}^{-1}\, \mathrm{cm}^{-1}]$`
   assoc["gaussian_linee"]=struct{name=`$E_\mathrm{gauss}$`,unit=`[keV]`,digits=2};
   assoc["gaussian_sigma"]=struct{name=`$E_\mathrm{\sigma}$`,unit=`[keV]`,digits=2,factor=1e6};
   %GABS
   assoc["gabs_linee"]=struct{name=`$E_\mathrm{gabs}$`,unit=`[keV]`,digits=3};
   assoc["gabs_sigma"]=struct{name=`$E_\mathrm{\sigma}$`,unit=`[keV]`,digits=3};
   assoc["gabs_tau"]=struct{name=`$\tau$`,digits=3,factor=1e3};
   %BREMSS
   assoc["bremss_norm"]=struct{name=`$A_\mathrm{bremss}$`,digits=2};
   assoc["bremss_kt"]=struct{name=`$kT$`,unit=`[keV]`,digits=2};
   %DISK
   assoc["disk_norm"]=struct{name=`$A_\mathrm{disk}$`,digits=2};
   assoc["disk_accrate"]=struct{name=`$Acc$`,digits=2};
   assoc["disk_cenmass"]=struct{name=`$m$`,unit=`$[\mathrm{M}_{\odot}$]`,digits=2};
   assoc["disk_rinn"]=struct{name=`$r_\mathrm{in}$`,digits=2};
   %DISKBB
   assoc["diskbb_tin"]=struct{name=`$kT_\mathrm{BB}$`,digits=2,unit=`[keV]`};
   assoc["diskbb_norm"]=struct{name=`$A_\mathrm{BB}$`,digits=1,sciMode=4};
   %BBODY
   assoc["bbody_kt"]=struct{name=`$kT_\mathrm{bbody}$`,digits=2,unit=`[keV]`};
   assoc["bbody_norm"]=struct{name=`$A_\mathrm{bbody}$`,digits=1,sciMode=4};
   %REFLIONX
   assoc["reflionx_norm"]=struct{name=`$A_\mathrm{ref}$`,digits=3,factor=1e5};
   assoc["reflionx_abund"]=struct{name=`$\mathrm{Fe}/\mathrm{Fe}_\odot$`,digits=2};
   assoc["reflionx_gamma"]=struct{name=`$\Gamma$`,digits=2};
   assoc["reflionx_xi"]=struct{name=`$\xi$`,digits=1,sciMode=4,unit=`$[\mathrm{erg}\,\mathrm{cm}\,\mathrm{s}^{-1}]$`};
   assoc["reflionx_redshift"]=struct{name=`$z$`,digits=1};
   %REFLECT
   assoc["reflect_rel"]=struct{name=`$A_\mathrm{rel\_ref}$`,digits=3};
   assoc["reflect_redshift"]=struct{name=`$z$`,digits=1};
   assoc["reflect_abund"]=struct{name=`$A_\mathrm{abund}$`,digits=1};
   assoc["reflect_fe_abund"]=struct{name=`$\mathrm{Fe}/\mathrm{Fe}_\odot$`,digits=2};
   assoc["reflect_cosincl"]=struct{name=`$\mathrm{cos}\,\theta$`,digits=1};
   %RELCONV 
   assoc["relconv_index1"]=struct{name=`$\epsilon_{1}$`,digits=1};
   assoc["relconv_index2"]=struct{name=`$\epsilon_{2}$`,digits=1};
   assoc["relconv_rbr"]=struct{name=`$r_\mathrm{br}$`,unit=`$[\mathrm{GM\,c}^{-2}]$`,digits=1,sciMode=3};
   assoc["relconv_rin"]=struct{name=`$r_\mathrm{in}$`,digits=1};
   assoc["relconv_rout"]=struct{name=`$r_\mathrm{out}$`,unit=`$[\mathrm{GM\,c}^{-2}]$`,digits=1,sciMode=3};
   assoc["relconv_a"]=struct{name=`a`,digits=3};
   assoc["relconv_incl"]=struct{name=`$i$`,digits=1,unit=`$[\mathrm{deg}]$`};
   assoc["relconv_limb"]=struct{name=`$\mathrm{limb}$`};
   %RELLINE
   assoc["relline_norm"]=struct{name=`$A_\mathrm{relline}$`,digits=3};
   assoc["relline_linee"]=struct{name=`$E_\mathrm{relline}$`,unit=`[keV]`,digits=2};
   assoc["relline_index1"]=struct{name=`$\epsilon_{1}$`,digits=1};
   assoc["relline_index2"]=struct{name=`$\epsilon_{2}$`,digits=1};
   assoc["relline_rbr"]=struct{name=`$r_\mathrm{br}$`,unit=`$[\mathrm{GM\,c}^{-2}]$`,digits=1,sciMode=3};
   assoc["relline_rin"]=struct{name=`$r_\mathrm{in}$`,digits=1};
   assoc["relline_rout"]=struct{name=`$r_\mathrm{out}$`,unit=`$[\mathrm{GM\,c}^{-2}]$`,digits=1,sciMode=3};
   assoc["relline_a"]=struct{name=`a`,digits=3};
   assoc["relline_incl"]=struct{name=`$i$`,digits=1,unit=`$[\mathrm{deg}]$`};
   assoc["relline_limb"]=struct{name=`$\mathrm{limb}$`};
   assoc["relline_z"]=struct{name=`$z$`,digits=1};
   %COMPTT
   assoc["comptt_norm"]=struct{name=`$A_\mathrm{compTT}$`,digits=3};
   assoc["comptt_redshift"]=struct{name=`$z$`,digits=1};
   assoc["comptt_t0"]=struct{name=`$T_{0}$`,digits=1,unit=`[keV]`};
   assoc["comptt_kt"]=struct{name=`$kT$`,digits=1,unit=`[keV]`};
   assoc["comptt_taup"]=struct{name=`$\tau$`,digits=1};
   assoc["comptt_approx"]=struct{name=`$A_\mathrm{approx}$`,digits=1};
  %COMPBB
   assoc["compbb_norm"]=struct{name=`$A_\mathrm{compbb}$`,digits=1};
   assoc["compbb_kt"]=struct{name=`$kT$`,unit=`[keV]`,digits=1};
   assoc["compbb_kte"]=struct{name=`$kTe$`,unit=`[keV]`,digits=1};
   assoc["compbb_tau"]=struct{name=`$\tau$`,digits=1};
   %COMPST
   assoc["compst_norm"]=struct{name=`$A_\mathrm{compst}$`,digits=1};
   assoc["compst_kt"]=struct{name=`$kT$`,unit=`[keV]`,digits=1};
   assoc["compst_tau"]=struct{name=`$\tau$`,digits=1};
   %BBODYRAD
   assoc["bbodyrad_norm"]=struct{name=`$A_\mathrm{bbody}$`,digits=1};
   assoc["bbodyrad_kt"]=struct{name=`$kT$`,unit=`[keV]`,digits=1};
   %BMC
   assoc["bmc_norm"]=struct{name=`$A_\mathrm{bmc}$`,digits=1};
   assoc["bmc_kt"]=struct{name=`$kT$`,unit=`[keV]`,digits=1};
   assoc["bmc_alpha"]=struct{name=`$\alpha$`,digits=1};
   assoc["bmc_log"]=struct{name=`$A_\mathrm{log}$`,digits=1};
   %CONSTANT
   assoc["constant_factor"]=struct{name=`$c$`};
   %GAINSHIFT
   assoc["gainshift_intercept"]=struct{name=`$s_\mathrm{intercept}$`,unit=`[keV]`,digits=2};
   assoc["gainshift_slope"]=struct{name=`$s_\mathrm{gainshift}$`,digits=3};

   variable assoc_keys=assoc_get_keys(assoc);
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


   %Get all info from table
   variable inputStruct=fits_load_fit_struct(inFile);
   variable names=get_struct_field_names(inputStruct);
   variable rows=fits_read_key(inFile,"NAXIS2");
   variable forTexTable;
   variable i,j,k,z,h;
   variable tempFile="tempFile.txt";
   variable texOut=fopen(tempFile,"w");
   %QUALIFIER INPUT
   variable inputModel=qualifier("pars",NULL);
   variable inputExtraInfo=qualifier("extraInfo",NULL);
   variable inputExclude=qualifier("exclude",NULL);
   variable colNames=qualifier("colNames",NULL);
   variable outputFile=qualifier("output","default.tex");
   variable sciOut=qualifier("sci",NULL);


   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %GET MODEL PARAMETER VALUES
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   variable oldName="";
   _for i(0,length(names)-1,1){
      _for j(0,length(inputModel)-1,1){   
	 if(is_substr(names[i],inputModel[j]) and is_substr(names[i],"value")){
	    variable parVal=get_struct_field(inputStruct,names[i]); %get values
	    variable name=names[i];
	 }
	 if(is_substr(names[i],inputModel[j]) and is_substr(names[i],"conf")){
	    variable parConf=get_struct_field(inputStruct,names[i]); %get conf values
	    variable texOutput;
	    %excluded:
	    variable exclude=0;
	    if(inputExclude!=NULL){
	       _for z(0,length(inputExclude)-1,1){
		  if(is_substr(name,inputExclude[z])){
		     exclude=1;
		  }
	       }
	    }
	    %end excluded

	    ifnot(exclude){
	       if (not qualifier_exists("silent")) ()=fprintf(stdout,"%s  \n",name);
	       variable holdName=name;
	       variable holdNameParts=strchop(holdName,'_',0);
	       if(oldName!=holdNameParts[0]){
		  ()=fprintf(texOut,"\\hline \n");
	       }
	       oldName=holdNameParts[0];
	       name=strlow(strreplace(name,"_","\\_"));

	       %ASSOCIATIVE ARRAYS INFO INPUT:
	       variable units="";
	       variable digits=1;
	       variable fac=1;
	       variable sciMode=2;
	       variable startRun;
	       _for z(0,length(assoc_keys)-1,1){
		  variable assocKeysParts=strchop(assoc_keys[z],'_',0);
		  if(is_substr(name,assocKeysParts[0]) and strcmp(strup(assocKeysParts[1]),strup(holdNameParts[2]))==0){
		     name=assoc[assoc_keys[z]].name;
		     variable origName=name;
		     if(struct_field_exists(assoc[assoc_keys[z]],"unit")!=0){
			units=assoc[assoc_keys[z]].unit;
			name=name+"  "+string(units);
		     }
		     if(struct_field_exists(assoc[assoc_keys[z]],"digits")!=0){
			digits=assoc[assoc_keys[z]].digits;
		     }
		     if(struct_field_exists(assoc[assoc_keys[z]],"sciMode")!=0){
			sciMode=assoc[assoc_keys[z]].sciMode;
		     }
		     if(struct_field_exists(assoc[assoc_keys[z]],"factor")!=0){
			fac=assoc[assoc_keys[z]].factor;
			variable inputFactor=sprintf("%e",fac);
			inputFactor=strchop(inputFactor,'e',0);
			inputFactor=inputFactor[1];
			ifnot(is_substr(inputFactor,"00")){
			   if(is_substr(inputFactor,"+")){
			      inputFactor=strtrans(inputFactor,"+","-");
			   }else{
			      inputFactor=strtrans(inputFactor,"-","");
			   }
			   inputFactor=sprintf("[$10^{%s}$]",inputFactor);
			   name=origName+"  "+ inputFactor+" "+string(units);
			}
			
		     }
		  }
	       }
	       %END ASSOCIATIVE ARRAYS INFO INPUT

	       %NEW LOOK INPUT
	       variable factorOut=1;
	       if(qualifier_exists(holdName)){
		  variable newQual=qualifier(holdName);
		  _for z(0,length(newQual)-1,1){
		     variable str=string(newQual[z]);
		     if(is_substr("name",str)){
			name=newQual[z+1];
			origName=name;
		     }
		     if(is_substr("units",str)){
			units=newQual[z+1];name=origName+" "+units;
		     }		     
		     if(is_substr("digits",str))digits=newQual[z+1];
		     if(is_substr("sciMode",str))sciMode=newQual[z+1];
		     if(is_substr("factor",str)){
			fac=newQual[z+1];
			factorOut=fac;
			inputFactor=sprintf("%e",fac);
			inputFactor=strchop(inputFactor,'e',0);
			inputFactor=inputFactor[1];
			ifnot(is_substr(inputFactor,"00")){
			   if(is_substr(inputFactor,"+")){
			      inputFactor=strtrans(inputFactor,"+","-");
			   }else{
			      inputFactor=strtrans(inputFactor,"-","");
			   }
			   inputFactor=sprintf("[$10^{%s}$]",inputFactor);
			   ifnot (__is_initialized(&origName)){
			     name=oldName+"  "+inputFactor+" "+string(units); 
			   }else{
			      name=origName+"  "+inputFactor+" "+string(units);
			   }				
			}
		     }
		  }
	       }
		 
	       ()=fprintf(texOut,"%s  ",name);
	       
	       _for k(0,rows-1,1){
		  ifnot(parConf[k,0]==0 and parConf[k,1]==0){
		     if(sciOut==0){%
			sciMode=10;
			texOutput=TeX_value_pm_error(parVal[k],parConf[k,0],parConf[k,1];factor=fac,sci=sciMode);
		     }else if(sciOut==1){%
			sciMode=10;
			texOutput=TeX_value_pm_error(parVal[k],parConf[k,0],parConf[k,1];factor=fac,sci=sciMode);
			if(is_substr(texOutput,"times10^{1}")){%
			   variable newV, newV_hi,newV_low;
			   ifnot(is_substr(texOutput,"pm")){%
			      %get value, upper and lower error
			      newV=strchop(texOutput,'(',1);
			      newV=strchop(newV[1],'^',1); % newV[0]
			      newV_hi=strchop(newV[1],'+',1);
			      newV_hi=strchop(newV_hi[1],'}',1); %newV_hi[0]
			      newV_low=strchop(newV[1],'-',1);
			      newV_low=strchop(newV_low[1],'}',1); %newV_low[0]%
			      texOutput=sprintf("$%d^{+%d}_{-%d}$",integer(newV[0])*10,integer(newV_hi[0])*10,integer(newV_low[0])*10);
			   }else{
			      newV=strchop(texOutput,'m',1); %newV[0]
			      variable newV_hilow=strchop(newV[1],'\\',1);%newV_hilow[0]
			      newV=strchop(newV[0],'(',1);
			      newV=strchop(newV[1],'\\',1);
			      texOutput=sprintf("$%d\\pm%d$",integer(newV[0])*10,integer(newV_hilow[0])*10);
			   }
			}
		     }else{%
			texOutput=TeX_value_pm_error(parVal[k],parConf[k,0],parConf[k,1];factor=fac,sci=sciMode); 
		     }
		  }else{

		     %FIND CORRECT DECIMAL (for writing non-zero output)
		     variable decim=string(parVal[k]);
		     variable p;
		     if(is_substr(decim,".")){
			_for p(0,strlen(decim)-1,1){
			   if(decim[p]==46){%ascii for .
			      break;
			   }
			}
			variable newDigits=0;
			variable onlyZeros=0;
			p++;
			_for h(p,strlen(decim)-1,1){
			   if(decim[h]!=48){
			      onlyZeros=1;
			      break;
			   }else{
			      newDigits++;
			   }
			}
		     }
		     ifnot(onlyZeros){
			digits=1;
		     }else{
			newDigits=newDigits+2;
			if(newDigits>digits){
			   digits=newDigits;
			}
		     }
		     
		     if(factorOut!=1){
			texOutput=sprintf("%.${digits}f"$,parVal[k]*factorOut);
		     }else{
			texOutput=sprintf("%.${digits}f"$,parVal[k]);
		     }
		  }
		  ()=fprintf(texOut," \&   %s  ",string(texOutput));
	       }
	       ()=fprintf(texOut," \n");
	    }
	 }
      }	       
   }
   
   %END GET MODEL PARAMETER VALUES (with exclusion)
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %GET EXTRA INFO
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
   
   if(qualifier_exists("extraInfo")){
      ()=fprintf(texOut,"\\hline \n");
      _for j(0,length(inputExtraInfo)-1,1){
	 _for i(0,length(names)-1,1){
	    if(is_substr(names[i],inputExtraInfo[j])){
	       variable extraInfoVal=get_struct_field(inputStruct,names[i]); %get values
	       variable nameExtraInfo=names[i];
	    }
	 }
	 digits=1;
	 
	 %NEW LOOK INPUT
	 if(qualifier_exists(inputExtraInfo[j])){
	    newQual=qualifier(inputExtraInfo[j]);
	    origName=inputExtraInfo[j];
	    name=origName;
	    _for z(0,length(newQual)-1,1){
	       str=string(newQual[z]);
	       if(is_substr("name",str)){
		  name=newQual[z+1];
		  origName=name;
	       }
	       if(is_substr("units",str)){
		  units=newQual[z+1];
		  name=origName+" "+units;
	       }		     
	       if(is_substr("digits",str)){
		  digits=newQual[z+1];
	       }
	    }
	    nameExtraInfo=name;
	 }
	 ()=fprintf(stdout,"%s  \n",string(nameExtraInfo));
	 nameExtraInfo=strreplace(nameExtraInfo,"_","\\_");
	 ()=fprintf(texOut,"%s  ",string(nameExtraInfo));

	 %Find array info
	 variable arrDim,arrNum,arrTyp;
	 (arrDim,arrNum,arrTyp)=array_info(extraInfoVal);
	 _for k(0,rows-1,1){
	    if(arrNum==1){
	       ()=fprintf(texOut," \& ");
	       variable extraInfoVal_clean=strtrans(string(extraInfoVal[k]),"\n"," ");
	       extraInfoVal_clean=strreplace(extraInfoVal_clean,"_","\\_");
	       variable te=strtok(extraInfoVal_clean," ");

	       _for z(0,length(te)-2,1){
		  ()=fprintf(texOut,"  %s \\par ",te[z]);
	       }
	       if(length(te)!=0){
		  ()=fprintf(texOut,"  %s ",te[length(te)-1]);
	       }
	    }else{
	       ()=fprintf(texOut," \& ");
	       variable arrayShape=array_shape(extraInfoVal);

	       if(arrayShape[1]==1){%
		  extraInfoVal_clean=strtrans(string(extraInfoVal[k,0]),"\n"," ");
		  variable te_2=strtok(extraInfoVal_clean," ");
		  ()=fprintf(texOut,"  %.${digits}f"$,atof(te_2[0]));
		  ()=fprintf(texOut,"  \\par");
	       }else{
		  _for z(0,rows-1,1){
		     extraInfoVal_clean=strtrans(string(extraInfoVal[z,0]),"\n"," ");
		     variable te_1=strtok(extraInfoVal_clean," ");
		     ()=fprintf(texOut,"  %.${digits}f"$,atof(te_1[0]));
		     ()=fprintf(texOut,"  \\par");
		  }
	       }
	    }
	 }
	 ()=fprintf(texOut,"\n");
      }
   }
   %END GET EXTRA INFO
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %GET STATISTICS
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   _for i(0,length(names)-1,1){
      if(string_matches(names[i],"chi")!=NULL){
	 variable chi=get_struct_field(inputStruct,names[i]); %get values
      }
      if(string_matches(names[i],"dof")!=NULL){
	 variable dof=get_struct_field(inputStruct,names[i]); %get values
      }
      if(string_matches(names[i],"chi_red")!=NULL){
	 variable chi_red=get_struct_field(inputStruct,names[i]); %get values
      }
   }
   ()=fprintf(texOut,"\\hline \n");
   ()=fprintf(texOut,"$\\chi^{2}/\\mathrm{dof}$ ");
   _for j(0,rows-1,1){
      ()=fprintf(texOut,"\&  %.1f/%d",chi[j],dof[j]);
   }
   ()=fprintf(texOut,"  \n");
   if(qualifier_exists("fullStat")){
      ()=fprintf(texOut,"$\\chi^{2}_\\mathrm{red}$ ",);
      _for j(0,rows-1,1){
	 ()=fprintf(texOut,"\&  %.2f",chi_red[j]);
      }
      ()=fprintf(texOut,"  \n");
   }
   %END GET STATISTICS
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   ()=fclose(texOut);



   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %WRITE TEX FILE
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   variable lineBreak="";
   variable inTex=fopen(tempFile,"r");
   variable lines=fgetslines(inTex);
   variable outTex=fopen(outputFile,"w");
   %set up tex tabular code
   variable columns="l";
   ()=fprintf(outTex,"\\begin{tabular}{");
   if(qualifier_exists("texMulti")){
      columns="p{4cm}";
   }
   _for k(0,rows,1){
      ()=fprintf(outTex,columns);
   }
   ()=fprintf(outTex,"} \n \\hline \n");

   ()=fprintf(outTex,"Parameter ");
   if(colNames!=NULL and length(colNames)==rows){
      _for k(0,rows-1,1){
	 ()=fprintf(outTex," & %s ",colNames[k]);
      }
   }else{
      _for k(1,rows,1){
	 ()=fprintf(outTex," &  ");
      }
   }
   ()=fprintf(outTex," \\\\ \n");
   
   _for k(0,length(lines)-1,1){
      variable chop=strtrans(lines[k],"\n","");
      
      ifnot(is_substr(lines[k],"\hline")){
	 ()=fprintf(outTex,"%s \\\\ \n",chop);
      }else{
	 ()=fprintf(outTex,"%s \n",chop);
      }
   }
   ()=fprintf(outTex,"\\hline \n");
   ()=fprintf(outTex,"\\end{tabular}");
   ()=fclose(inTex);
   ()=fclose(outTex);
   %END WRITE TEX FILE
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %FLIP
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   if(qualifier_exists("flip")){
      variable regTex=fopen(outputFile,"r");
      variable outputTexFile="flip.tex";
      variable flipTex=fopen(outputTexFile,"w");
      variable inLines=fgetslines(regTex);
      %GET NUMBER OF \HLINES
      variable hlines=0;
      _for k(0,length(inLines)-1,1){
	 if(is_substr(inLines[k],"\hline")){
	    hlines++;
	 }
      }
          
      variable arr=Array_Type[length(inLines)-hlines-2];      
      j=0;
      _for i(0,length(inLines)-1,1){
	 ifnot(is_substr(inLines[i],"\hline")){
	    variable lineParts=strchop(inLines[i],'&',0);
	    if(is_substr(inLines[i],"&")){
	       arr[j]=lineParts;
	       j++;
	    }
	 }
      }
      %write flip.tex file
      ()=fprintf(flipTex,"\\begin{tabular}{");
      if(qualifier_exists("texMulti")){
	 _for k(0,length(inLines)-5,1){
	    ()=fprintf(flipTex,`p{4cm}`);
	 }
      }else{
	 _for k(0,length(inLines)-5,1){
	    ()=fprintf(flipTex,`l`);
	 }
      }
      ()=fprintf(flipTex,"} \n \\hline \n");
      
      h=0;
      _for k(0,rows,1){
	 _for j(1,length(arr)-1,1){
	    variable clean=strtrans(arr[j][k],"\n","");
	    (clean,z)=strreplace(clean,`\\`,"",1);
	    if(j==1){
	       ()=fprintf(flipTex," %s ",clean);
	    }else{
	       ()=fprintf(flipTex,"& %s ",clean);	 
	    }
	 }
	 ()=fprintf(flipTex,"\\\\ \n");
	 ifnot(h){
	    ()=fprintf(flipTex,"\\hline \n");
	    h=1;
	 }
      }
      ()=fprintf(flipTex,"\\hline \n");
      ()=fprintf(flipTex,"\\end{tabular}");
      
      ()=fclose(regTex);
      ()=fclose(flipTex);
      ()=system("cp $outputTexFile $outputFile"$);
      ()=system("rm $outputTexFile"$);
   }
   %END FLIP
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %PDF OUTPUT
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   if(qualifier_exists("pdf")){
      variable pdf=fopen("pdfout.tex","w");
      ()=fprintf(pdf,"\\documentclass[]{article} \\usepackage[margin=0.5in]{geometry} \\begin{document} \\begin{table}[!ht]");
      ()=fprintf(pdf,"\\input{$outputFile}"$);
      ()=fprintf(pdf,"\\end{table} \\end{document}");
      ()=fclose(pdf);
      ()=system("xelatex pdfout.tex >/dev/null 2>&1");
   }
   %END PDF OUTPUT
   ()= remove("tempFile.txt");
}

%
% Code from src/data/flux2lum.sl
%
%%%%%%%%%%%%%%%%
define flux2lum()
%%%%%%%%%%%%%%%%
%!%+
%\function{flux2lum}
%
%\synopsis{Calculates the source luminosity}
%\usage{Double_Type lum = flux2lum (Double_Type flux, Double_Type z);}
%\qualifiers{
%\qualifier{silent}{         If set, the program will not display adopted
%                       cosmological parameters at the terminal.}
%\qualifier{h0}{[=70] Hubble parameter in km/s/Mpc}
%\qualifier{omega_m}{[=0.3] Matter density, normalized to the closure density,
%                       default is 0.3. Must be non-negative.}
%\qualifier{omega_lambda}{[=0.7] Cosmological constant, normalized to the
%                       critical density.}
%\qualifier{omega_k}{[=0] curvature constant, normalized to the critical density.
%                       Default is 0, indicating a flat universe.}
%\qualifier{q0}{[=-0.55] Deceleration parameter, numeric scalar = -R*(R'')/(R')^2}
%}
%\description
%    This function calculates the luminosity of a source
%    given a \code{flux} in erg/s/cm^2 as well as a redshift \code{z} using
%    the cosmological parameters specified by the qualifiers
%    (which are passed to the function \code{cosmo_param}). The
%    distance is calculated with the function \code{lumdist}.
%    \code{Flux} and \code{z} can be scalars or vectors. Use
%    energyflux to calculate the flux.
%\example
%    variable flux=3e-13;
%    variable z=0.0705;
%    variable l = flux2lum(flux,z);
%\seealso{energyflux, cosmo_param, lumdist}
%!%-
{ 
  variable fl1 = NULL;
  variable redz = NULL;
  
  switch(_NARGS)
  { case 2: (fl1, redz) = ();}
  { help(_function_name()); return; }
  
  % 1. Get distance from Z
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  variable dist = lumdist (redz ;; __qualifiers); % in Mpc
  variable Mpc_in_cm = 3.08567758e24;
  dist *= Mpc_in_cm;
  
  % 2. Calculate Luminosity
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  variable lumi = fl1*4*PI*dist^2;
  return lumi;  
}
%
% Code from src/data/fm_unredd.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define fm_unred()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{fm_unred}
%\synopsis{Deredden a flux vector using the Fitzpatrick (1999) parameterization}
%\usage{Double_Type = fm_unred(Double_Type[] wave, Double_Type[] flux, Double_Type ebv);}
%\qualifiers{
%    \qualifier{N_H}{Hydrogen absorption column}
%    \qualifier{R_V}{Scalar specifying the ratio of total to selective extinction
%               R(V) = A(V) / E(B - V). If not specified, then R = 3.1
%               Extreme values of R(V) range from 2.3 to 5.3}
%    \qualifier{LMC2}{If set, then the fit parameters are set to the values determined
%               for the LMC2 field (including 30 Dor) by Misselt et al.
%               Note that neither 'AVGLMC' or 'LMC2' will alter the default
%               value of 'R_V' which is poorly known for the LMC.}
%    \qualifier{AVGLMC}{If set, then the default fit parameters c1,c2,c3,c4,gamma,x0
%               are set to the average values determined for reddening in the
%               general Large Magellanic Cloud (LMC) field by Misselt et al.
%               (1999, ApJ, 515, 128)}
%    \qualifier{extcurve}{If set to a variable as Ref_Type, the E(wave-V)/E(B-V)
%               extinction curve is returned, interpolated onto the input
%               wavelength vector}
%    \qualifier{gamma}{Width of 2200 A bump in microns (default = 0.99)}
%    \qualifier{x0}{Centroid of 2200 A bump in microns (default = 4.596)}
%    \qualifier{c1}{Intercept of the linear UV extinction component
%               (default = 2.030 - 3.007 * c2)}
%    \qualifier{c2}{Slope of the linear UV extinction component
%               (default = -0.824 + 4.717 / R_V)}
%    \qualifier{c3}{Strength of the 2200 A bump (default = 3.23)}
%    \qualifier{c4}{FUV curvature (default = 0.41)}
%    \qualifier{wilm}{If N_H value has been determined using wilm abundances and Verner 
%               cross-sections use this updated correlation (Nowak et al., 2012)}
%    \qualifier{ngc3227}{Use the reddening curve for NGC3227 after Crenshaw, D.~M.,
%               Kraemer, S.~B., Bruhweiler, F.~C., & Ruiz, J.~R. 2001, ApJ, 555, 633
%               provided locally in /home/beuchert/work/ngc3227/scripts/reddening/}
%
%}
%\description
%     The unreddened flux vector of the input 'flux' is calculated on the
%     wavelength vector 'wave' using the Fitzpatrick (1999) parameterization.
%     The scalar 'ebv' is the color excess E(B-V). If a negative EBV is supplied,
%     then fluxes will be reddened rather than dereddened. If the scalar is not
%     known the hydrogen absorption column in the directory of the object must be
%     given as a qualifier and E(B-V) is calculated by
%       E(B-V) = N_H/(1.79e21*R_V).
%     
%     The R-dependent Galactic extinction curve is that of Fitzpatrick & Massa 
%     (Fitzpatrick, 1999, PASP, 111, 63; astro-ph/9809387 ).    
%     Parameterization is valid from the IR to the far-UV (3.5 microns to 0.1 
%     microns). UV extinction curve is extrapolated down to 912 Angstroms.
%     This Function is adopted from the IDL-Function fm_unred.
%         
%     The five input qualifiers 'gamma', 'x0', 'c1', 'c2', 'c3' and 'c4' allow the
%     user to customize the adopted extinction curve. For example, see Clayton et al.
%     (2003, ApJ, 588, 871) for examples of these parameters in different interstellar
%     environments.           
%
% EXAMPLE
%     Determine how a flat spectrum (in wavelength) between 1200 A and 6500 A
%     is altered by a reddening of E(B-V) = 0.5. Assume an "average"
%     reddening for the diffuse interstellar medium (R(V) = 3.1)
%
%     isis> wave = [1200:6500:#500];                               %Create a wavelength vector from 1200 to 6500 Angstrom with 500 steps between
%     isis> flux = 1.*ones(500);                                   %Create a "flat" flux vector
%     isis> ebv = 0.5;                                             %Value for E(B-V)
%     isis> variable var;                                          %Referenz for extcurve 
%     isis> funred = fm_unred(wave, flux, ebv; extcurve=&var);     %Redden flux vector
%     isis> plot(wave,var);                                        %Plots the extinctioncurve versus the wavelength
%
%\seealso{e_bv;}
%!%-
{
 variable wave, flux, ebv = NULL;
 switch (_NARGS)
     { case 2: (wave, flux)      = (); }
     { case 3: (wave, flux, ebv) = (); }
     { help(_function_name()); return; }
	 
 variable R_V = qualifier("R_V", 3.1);
 variable N_H = qualifier("N_H", 1.e20);
 variable LMC2 = qualifier("LMC2");
 variable AVGLMC = qualifier("AVGLMC");
 variable x0 = qualifier("x0");
 variable gamma = qualifier("gamma");
 variable c1 = qualifier("c1");
 variable c2 = qualifier("c2");
 variable c3 = qualifier("c3");
 variable c4 = qualifier("c4");
 variable wilm = qualifier("wilm");
 variable ngc3227 = qualifier("ngc3227");

  
 variable x = 10000./ wave;                % Convert to inverse microns 
 variable curve = x*0.;

% Set default values of c1,c2,c3,c4,gamma and x0 parameters

 if (qualifier_exists("LMC2"))   
     {
	 ifnot (qualifier_exists("x0"))  {x0    =  4.626;};
         ifnot (qualifier_exists("gamma"))  {gamma =  1.05;};	
         ifnot (qualifier_exists("c4"))  {c4   =  0.42;};
         ifnot (qualifier_exists("c3"))  {c3    =  1.92;};	
         ifnot (qualifier_exists("c2"))  {c2    = 1.31;};
         ifnot (qualifier_exists("c1"))  {c1    =  -2.16;};
     }
 else if (qualifier_exists("AVGLMC")) 
     {
	 ifnot (qualifier_exists("x0"))  {x0 = 4.596;};
         ifnot (qualifier_exists("gamma"))  {gamma = 0.91;};
         ifnot (qualifier_exists("c4"))  {c4   =  0.64;};
         ifnot (qualifier_exists("c3"))  {c3    =  2.73;};
         ifnot (qualifier_exists("c2"))  {c2    = 1.11;};
         ifnot (qualifier_exists("c1"))  {c1    =  -1.28;};
     }
   
 else 
     {
	 ifnot (qualifier_exists("x0"))  {x0    =  4.596;};
         ifnot (qualifier_exists("gamma"))  {gamma =  0.99;};
         ifnot (qualifier_exists("c3"))  {c3    =  3.23;};
         ifnot (qualifier_exists("c4"))  {c4   =  0.41;};
         ifnot (qualifier_exists("c2"))  {c2    = -0.824 + 4.717/R_V;};
         ifnot (qualifier_exists("c1"))  {c1    =  2.030 - 3.007*c2;};
     }
   
 if (ebv == NULL)
 {
   if (qualifier_exists("N_H")) 
     { 
	 if (qualifier_exists("wilm")){
	     ebv = N_H/(2.69e21*R_V);
	 }
	 else {
	     ebv = N_H/(1.79e21*R_V);
	 }
     }
     else 
     {
	 message("Either 'ebv' or 'N_H' must be given!"); 
	 return; 
     }	  
     
 }

% Compute UV portion of A(lambda)/E(B-V) curve using FM fitting function and 
% R-dependent coefficients
 
 variable xcutuv = 10000.0/2700.0;
 variable xspluv = 10000.0/[2700.0,2600.0];
 variable iopir;
 variable iuv = where(x >= xcutuv, &iopir);
 variable N_UV = length(iuv);
 variable Nopir = length(iopir);
 if (length(N_UV) > 0)  {variable xuv = [xspluv,x[iuv]];}
   else  {xuv = xspluv;}

 variable yuv = c1  + c2*xuv;
    yuv += c3*xuv^2./((xuv^2.-x0^2.)^2. +(xuv*gamma)^2.);
    yuv += c4*(0.5392*((xuv-5.9)*(xuv>5.9))^2.+0.05644*((xuv-5.9)*(xuv>5.9))^3.);
    yuv += R_V;
  variable yspluv = yuv[[0:1]];                  % save spline points

 if (length(N_UV) > 0)  { curve[iuv] = yuv[[2:length(yuv)-1]];}      % remove spline points
 
% Compute optical portion of A(lambda)/E(B-V) curve
% using cubic_spline anchored in UV, optical, and IR

#ifexists gsl->interp_cspline
 variable xsplopir = [0.0,10000.0/[26500.0,12200.0,6000.0,5470.0,4670.0,4110.0]];
 variable ysplir   = [0.0,0.26469,0.82925]*R_V/3.1 ;
 variable ysplop   = [polynom([-4.22809e-01, 1.00270, 2.13572e-04], R_V ),
		      polynom([-5.13540e-02, 1.00216, -7.35778e-05], R_V ),
		      polynom([ 7.00127e-01, 1.00184, -3.32598e-05], R_V ),
		      polynom([ 1.19456, 1.01707, -5.46959e-03, 7.97809e-04,-4.45636e-05], R_V ) ];
  
 variable ysplopir = [ysplir,ysplop];

 if (Nopir > 0)  {
          curve[iopir] = gsl->interp_cspline(x[iopir],[xsplopir,xspluv],[ysplopir,yspluv]);
 }
#endif

  variable curve_cpy = @curve;
  
 % special case: reddening curve for NGC3227 after Crenshaw, D.~M.,
 % Kraemer, S.~B., Bruhweiler, F.~C., & Ruiz, J.~R. 2001, ApJ, 555, 633

  variable fname = "/home/beuchert/work/ngc3227/scripts/reddening/f3_1_edit_curve.dat";
  if (qualifier_exists("ngc3227") && (stat_file(fname)!=NULL)) {
    variable x_ngc3227,y_ngc3227;
    (x_ngc3227,y_ngc3227)=readcol(fname,1,2);
    variable unity_array = wave*0.+1.;
    variable id = where(x_ngc3227[0] < wave < x_ngc3227[-1]);
    curve = [unity_array[where(wave<x_ngc3227[0])],interpol(wave[id],x_ngc3227,y_ngc3227),curve_cpy[where(wave>x_ngc3227[-1])]];
  }

 % Now apply extinction correction to input flux vector

 curve *= ebv; % = A_lambda, before: curve == A_lambda / E(B-V)
   
 variable funred = flux * 10.^(0.4*curve);
    	   
%       variable funred = flux * 10.^(0.4*curve);       %Derive unreddened flux
 variable extcurve = qualifier("extcurve", NULL);
 ifnot (typeof(extcurve) == Ref_Type || typeof(extcurve) == Null_Type) { message("'extcurve' must be a reference!"); return; }
 else if (typeof(extcurve) == Ref_Type) @extcurve = 1.*curve/ebv - R_V; % == E(lambda-V)/E(B-V)

 return funred; 
}  %}}}

#endif

%
% Code from src/data/timing/foucalc.sl
%
%%%%%%%%%%%%%%
define foucalc()
%%%%%%%%%%%%%%
%!%+
%\function{foucalc}
%\synopsis{calculates power-, cross power-, coherence- and timelag-spectra for timing analysis}
%\usage{Struct_Type foucalc(Struct_Type lc, Integer_Type dimseg)}
%\qualifiers{
%\qualifier{verbose}{}
%\qualifier{normtype}{normalization type of PSD data, can be \code{"Miyamoto"} [default], \code{"Leahy"}, or \code{"Schlittgen"}}
%\qualifier{normindiv}{}
%\qualifier{avgbkg}{array of average background rates for each energy band}
%\qualifier{numinst}{[\code{=1}] number of activated PCUs on XTE, required for noise correction}
%\qualifier{deadtime}{[\code{=1e-5}] detector deadtime in seconds}
%\qualifier{nonparalyzable}{Set this qualifier if the deadtime is non-paralyzable}
%\qualifier{fmin}{minimum frequency used for RMS calculation}
%\qualifier{fmax}{maximum frequency used for RMS calculation}
%\qualifier{RMS}{DEPRECATED, use rms qualifier instead.}
%\qualifier{rms}{reference to a variable to store the signal, noise RMS and error
%           of each light curve in the [fmin, fmax] band. Can only be used with
%           normtype="Miyamoto". The error is calculated using Vaughan
%           et al., MNRAS 345, 1271, 2003 Eq. 11.}
%\qualifier{avgrate}{reference to a variable to store the average rate of each light curve}
%\qualifier{compact}{compact the output structure, only keep the most important quantities}
%\qualifier{noCPD}{do not calculate cross power densities and derived quantities}
%}
%\description
%    \code{lc} contains (properly segmented) light curves in several energy bands.
%    The best performance is achieved with a common structure of arrays,
%    \code{lc = struct { time=[t_1, t_2, ...], rate1=[r1_1, r1_2, ...], rate2=... };}.
%    [However, one can also use an array of structures (with an enormous overhead),
%    \code{lc = [ struct { time=t_1, rate1=r1_1, rate2=...}, struct { time=t_2, rate1=r1_2, rate2=... }, ... ];}.]
%    Also specifying "rate" instead of "rate1" is possible if no CPD
%    should be computed.
%
%    \code{dimseg} is the segment size used for the FFTs, which should therefore be a power of 2.
%
%    The returned structure contains the following fields:\n
%    - \code{freq}:
%        the frequency grid\n
%    - \code{numavgall}:
%        The bin \code{i} in all power spectra has been averaged over \code{numavgall[i]} original bins.
%        Here, \code{numavgall[i] = numseg}, as no frequency rebinning has been performed.\n
%    - Power spectra for each energy band:\n
%      * \code{rawpsd}, \code{errpsd = rawpsd/sqrt(numseg)}, \code{noipsd}:
%          the raw power spectrum (from \code{makepsd}),
%          its error from the average over segments,
%          and the noise level (from \code{psdcorr_zhang})\n
%      * \code{sigpsd = rawpsd - noipsd}:
%          the signal power spectrum\n
%      * \code{rawnormpsd}, \code{errnormpsd}, \code{noinormpsd}, \code{effnoinormpsd = noinormpsd / sqrt(numseg)}:
%          the normalized power spectrum,
%          its error, the noise level,
%          and the effective noise level in the normalized power spectrum\n
%      * \code{signormpsd = rawnormpsd - noinormpsd}:
%          the signal in the normalized power spectrum\n
%    - Cross power spectra, coherence and time lag functions for each pair of energy bands:\n
%      * \code{realcpd}, \code{imagcpd}:
%          real and imaginary part of the cross power density\n
%      * \code{errrealcpd}, \code{errimagcpd}:
%          standard error on the mean of the averaged cross power density\n
%      * \code{noicpd = ( sigpsd_lo * noipsd_hi + sigpsd_hi * noipsd_lo + noipsd_lo * noipsd_hi ) / numseg}\n
%      * \code{rawcof}, \code{cof}, \code{errcof}:
%          non-noise-corrected (raw) and noise-corrected coherence function and its one-sigma uncertainty\n
%      * \code{lag}, \code{errlag}:
%          time lag and its one-sigma uncertainty\n
%\seealso{makepsd, cross_power_density, colacal, psdcorr_zhang}
%!%-
{
  variable lc, dimseg;
  switch(_NARGS)
  { case 2: (lc, dimseg) = (); }
  { help(_function_name()); return; }

  variable verbose = qualifier_exists("verbose");

  % Lengths, times and channels
  variable t = array_struct_field(lc, "time");;
  variable bintime = t[1]-t[0];
  variable dimlc = length(t);
  variable numseg = int(dimlc/dimseg);
  variable tseg = bintime * dimseg;
  variable timelc = bintime * dimlc;
  variable str_fields = get_struct_field_names(lc[0]);  % lc[0]=lc, if lc is no array but struct
  variable numchan = length(str_fields[where(is_substr(str_fields, "rate"))]);  

  % Lightcurve parameters
  variable avgrate = Double_Type[numchan];
  variable rms_deprecated = Double_Type[numchan];  % used for deprecated RMS qualifier
  
  variable fieldnames = ["freq", "numavgall"];  % field names of output structure
  variable rms_fieldnames = String_Type[0];
  
  variable i, j;
  _for i (1, numchan, 1){
    fieldnames = [fieldnames, ["rawpsd", "errpsd", "normpsd", "errnormpsd", "noipsd", "sigpsd", "noinormpsd", "signormpsd", "effnoinormpsd"]+string(i)];
    rms_fieldnames = [rms_fieldnames, ["sigrms", "errrms", "noirms"]+string(i)];
  }
  
  _for i (1, numchan, 1)
    _for j (i+1, numchan, 1)
      fieldnames = [fieldnames, ["normrealcpd", "normimagcpd", "errnormrealcpd", "errnormimagcpd",
				 "realcpd", "imagcpd", "errrealcpd", "errimagcpd", "noicpd", 
				 "rawcof", "cof", "errcof", "lag", "errlag"]+string(i)+string(j)];  % assuming i < j < 100 ...
  variable fouquant = @Struct_Type(fieldnames);  % create structure
  variable rmsquant = @Struct_Type(rms_fieldnames);

  % Calculating Fourier Frequency Array
  fouquant.numavgall = ones(dimseg/2) * numseg;
  fouquant.freq = foufreq(t[[0:dimseg-1]]);
  if(verbose)
  {
    variable tgap = t[-1]-t[0] - (dimlc-1) * bintime;
    vmessage("  light curves: %g s with %d bins of %g s  (%g s gaps)", t[-1]-t[0], dimlc, bintime, tgap);
    if(tgap != 0.)
      message("POSSIBLE DETECTION OF GAPS IN THE LIGHTCURVE!\nIf the light curves contain gaps, one cannot rely on the Discrete Fourier Transformation.");
    vmessage("  segmentation: %d segments of %d bins", numseg, dimseg);
    vmessage("=> frequencies: %g-%g Hz.", fouquant.freq[0], fouquant.freq[-1]);
  }

  % Normalization Qualifiers
  variable normtype = qualifier("normtype", "Miyamoto");
  variable normindiv = qualifier_exists("normindiv");
  variable avgbkg = qualifier("avgbkg", Double_Type[numchan]);
  if(verbose)  vmessage("Power spectra will be calculated in %s-normalization.", normtype);
  
  if (not qualifier_exists("deadtime")){
    vmessage("warning (foucalc): Deadtime keyword not set. Using default for RXTE: 1e-05 s");
  }

  % RMS Qualifiers (see, e.g., Uttley et al., A&A Review 22, 72, 2014, Sect. 2.2.3)
  variable fmin = qualifier("fmin", min(fouquant.freq));
  variable fmax = qualifier("fmax", max(fouquant.freq));

  _for i (1, numchan, 1)
  {
    variable istr = string(i);
    if (verbose) ()=printf("Calculating power spectrum for rate%d.  ", i);

    % If "rate1" is not available in struct, use "rate"
    variable rate = struct_field_exists(lc, "rate"+istr) ? array_struct_field(lc, "rate"+istr) : array_struct_field(lc, "rate");
    avgrate[i-1] = mean(rate);

    % Calculating PSDs
    variable psd, nipsd;
    (psd, nipsd) = makepsd(rate, tseg, dimseg;; struct_combine(__qualifiers, struct { avgbkg=avgbkg[i-1] }));
    set_struct_field(fouquant, "rawpsd"+istr, psd);
    if(normindiv)
    { fouquant = struct_combine(fouquant, "normindivpsd"+istr);
      set_struct_field(fouquant, "normindivpsd"+istr, nipsd);
    }
    % No frequency rebinning at this point!

    % Calculating PSD errors! (errpsd)
    variable errpsd = psd/sqrt(numseg);
    set_struct_field(fouquant, "errpsd"+istr, errpsd);

    % Calculating normalized PSDs and their errors! (normpsd, errnormpsd)
    variable tempnormpsd, temperrnormpsd;
    if(normindiv)
    { tempnormpsd = nipsd;
      temperrnormpsd = tempnormpsd/sqrt(numseg);
    }
    else
    { tempnormpsd = normpsd(psd, normtype, avgrate[i-1], avgbkg[i-1], tseg, dimseg);
      temperrnormpsd = errpsd / psd[0] * tempnormpsd[0];
    }
    set_struct_field(fouquant, "normpsd"+istr, tempnormpsd);
    set_struct_field(fouquant, "errnormpsd"+istr, temperrnormpsd);
    temperrnormpsd = NULL;

    % Calculating observational noise with deadtime influence (noipsd, noinormpsd)
    variable noipsd, noinormpsd;
    (, noipsd, noinormpsd) = psdcorr_zhang(avgrate[i-1], tseg, dimseg;; struct_combine(__qualifiers(), struct { avgbkg=avgbkg[i-1] }));
    set_struct_field(fouquant, "noipsd"+istr, noipsd);
    set_struct_field(fouquant, "noinormpsd"+istr, noinormpsd);

    % effective noise level
    set_struct_field(fouquant, "effnoinormpsd"+istr, noinormpsd/sqrt(numseg));
    % calculate normalized signal psd
    variable signormpsd = tempnormpsd - noinormpsd;
    set_struct_field(fouquant, "signormpsd"+istr, signormpsd);

    if(normtype=="Miyamoto")
    {
      variable idx = where(fmin <= fouquant.freq <= fmax);
      variable idxmin = min(idx);
      variable idxmax = max(idx);
      variable df = fouquant.freq[[idxmin+1:idxmax]] - fouquant.freq[[idxmin:idxmax-1]];
      rms_deprecated[i-1] = sqrt( sum(signormpsd[[idxmin:idxmax-1]] * df) );
      
      variable sigrms = sqrt( sum(signormpsd[[idxmin:idxmax-1]] * df) );
      variable noirms = sqrt( sum(noinormpsd[[idxmin:idxmax-1]] * df) );
      variable errrms = sqrt((4.*sigrms^2*noirms^2+2.*noirms^4)/(numseg*length(idx)*sigrms^2));  % Vaughan+03 Eq. 11
      set_struct_field(rmsquant, "sigrms"+istr, sigrms);
      set_struct_field(rmsquant, "errrms"+istr, errrms);
      set_struct_field(rmsquant, "noirms"+istr, noirms);
      if(verbose)  vmessage("RMS(%g-%g Hz) = (%g +/- %g) %%", fmin, fmax, sigrms*100, errrms*100);
    }
    else
      if(verbose)  message("");  % to finish the earlier printf without "\n"

    % correct PSD for observational noise and deadtime
    set_struct_field(fouquant, "sigpsd"+istr, psd - noipsd);
    noipsd = NULL; noinormpsd = NULL; signormpsd = NULL;
  }

  if(qualifier_exists("noCPD"))
    vmessage("warning (%s): noCPD qualifier set => skipping calculation of cross power densities", _function_name());
  else
  {
  % CPD calculations
  variable lo, hi;
  _for lo (1, numchan-1, 1)
    _for hi (lo+1, numchan, 1)
    {
      variable lostr = string(lo);
      variable histr = string(hi);

      if (verbose) vmessage("Calculating cross power spectrum (=> coherence and timelags) for rate%d and rate%d.", lo, hi);

      variable cpd, errcpd;
      (cpd, errcpd) = cross_power_density(array_struct_field(lc, "rate"+lostr),
					  array_struct_field(lc, "rate"+histr),
					  numseg, dimseg);
      set_struct_field(fouquant, "realcpd"+lostr+histr, Real(cpd));
      set_struct_field(fouquant, "imagcpd"+lostr+histr, Imag(cpd));
      set_struct_field(fouquant, "errrealcpd"+lostr+histr, Real(errcpd));
      set_struct_field(fouquant, "errimagcpd"+lostr+histr, Imag(errcpd));

      variable tempnormcpd = normpsd(cpd, normtype, sqrt(avgrate[lo-1]*avgrate[hi-1]), sqrt(avgbkg[lo-1]*avgbkg[hi-1]), tseg, dimseg);
      variable temperrnormcpd = errcpd / cpd[0] * tempnormcpd[0];
      set_struct_field(fouquant, "normrealcpd"+lostr+histr, Real(tempnormcpd));
      set_struct_field(fouquant, "errnormrealcpd"+lostr+histr, Real(temperrnormcpd));
      set_struct_field(fouquant, "normimagcpd"+lostr+histr, Imag(tempnormcpd));
      set_struct_field(fouquant, "errnormimagcpd"+lostr+histr, Imag(temperrnormcpd));
      
      % use SIGPSD to calculate the cross power density noise
      variable sigpsd_lo = get_struct_field(fouquant, "sigpsd"+lostr);
      variable sigpsd_hi = get_struct_field(fouquant, "sigpsd"+histr);
      variable noipsd_lo = get_struct_field(fouquant, "noipsd"+lostr);
      variable noipsd_hi = get_struct_field(fouquant, "noipsd"+histr);
      variable noicpd = ( sigpsd_lo * noipsd_hi + sigpsd_hi * noipsd_lo + noipsd_lo * noipsd_hi ) / numseg;  % fouquant.numavgall;
      set_struct_field(fouquant, "noicpd"+lostr+histr, noicpd);

      variable psd_lo = get_struct_field(fouquant,"rawpsd"+lostr);
      variable psd_hi = get_struct_field(fouquant,"rawpsd"+histr);

      variable cola = colacal(fouquant.freq, cpd, noicpd, psd_lo, psd_hi, noipsd_lo, noipsd_hi, sigpsd_lo, sigpsd_hi, numseg);  % fouquant.numavgall
      set_struct_field(fouquant, "rawcof"+lostr+histr, cola.rawcof);
      set_struct_field(fouquant, "cof"   +lostr+histr, cola.cof);
      set_struct_field(fouquant, "errcof"+lostr+histr, cola.errcof);
      set_struct_field(fouquant, "lag"   +lostr+histr, cola.lag);
      set_struct_field(fouquant, "errlag"+lostr+histr, cola.errlag);
      cola= NULL;
      cpd = NULL; errcpd=NULL;
    }
  }

  variable rmsref = qualifier("rms");
  if(normtype=="Miyamoto" and rmsref!=NULL)
  {
    if(typeof(rmsref) == Ref_Type)
      (@rmsref) = rmsquant;
    else
      vmessage("warning (%s): rms qualifier has to be a reference", _function_name());
  }
  
  % Deprecated RMS qualifier (kept for backwards-compatibility)
  variable RMSref = qualifier("RMS");
  if(normtype=="Miyamoto" and RMSref!=NULL)
  {
    if(typeof(RMSref)==Ref_Type){
      vmessage("warning (%s): RMS qualifier is deprecated, please change to rms.", _function_name());
      (@RMSref) = rms_deprecated;
    } else {
      vmessage("warning (%s): RMS qualifier has to be a reference", _function_name());
    }
  }

  variable avgrateref = qualifier("avgrate");
  if(avgrateref!=NULL)
  {
    if(typeof(avgrateref)==Ref_Type)
      (@avgrateref) = avgrate;
    else
      vmessage("warning (%s): avgrate qualifier has to be a reference", _function_name());
  }

  if(qualifier_exists("compact"))
  {
    fieldnames = ["freq", "numavgall"];  % field names of output structure
    _for i (1, numchan, 1)
      fieldnames = [fieldnames, "signormpsd"+string(i)];
    _for i (1, numchan, 1)
      _for j (i+1, numchan, 1)
        fieldnames = [fieldnames, ["rawcof", "cof", "errcof", "lag", "errlag"]+string(i)+string(j)];
    variable compact_fouquant = @Struct_Type(fieldnames);  % create structure
    variable fieldname;
    foreach fieldname (fieldnames)
      set_struct_field(compact_fouquant, fieldname, get_struct_field(fouquant, fieldname));
    fouquant = compact_fouquant;
  }

  return fouquant;
}

%
% Code from src/data/timing/foufreq.sl
%
%%%%%%%%%%%%%%
define foufreq(time)
%%%%%%%%%%%%%%
%!%+
%\function{foufreq}
%\synopsis{Timing Tools: Calculation of the Fourier Frequency Array}
%\usage{Double_Type freq[] = foufreq(Double_Type time[]);}
%\description
%     Calculates the Fourier frequency array corresponding
%     to a given equally-binned time array.
%!%-
{
  variable dimseg = length(time);
  return [1:dimseg/2] * 1. / ( (time[1]-time[0]) * dimseg );
}

%
% Code from src/plot/fplot.sl
%

define fplot() {
%!%+
%\function{fplot}
%\synopsis{plots data and residuals from the fits-files created with 'save_plot'}
%\usage{fplot(filename, [color, [dataset]]);}
%\qualifiers{
%\qualifier{flux}{plot flux [photons/s/cm^2/keV] and not [counts/bin]}
%\qualifier{overplot}{Overplot previous plot}
%\qualifier{type}{defines what should be plotted. It can have the values 'model', 'data', 'res' (residuals),
%      'ratio', 'diff' (difference between model an data. By default model and data is plotted.)}
%\qualifier{nolabel}{The routine uses the labels given in 'plot_options' and does not create its own. }
%\qualifier{space}{If at the qualifer 'type' data, model or diff is chosen, one can specifiy if it is
% plotted in counts/bin (by default), counts/keV/s ('density') or photons/keV/s/cm^2 ('flux').} 
%\qualifier{auto}{Set the ranges automatically.}
%\qualifier{xauto}{Set the x-range automatically.}
%\qualifier{yauto}{Set the y-range automatically.}
%\qualifier{unit}{Specifies the unit of the x-axis. By default the unit from the fitstable is taken. 
%       Possible values are 'keV' or Angstrom 'A'.}
%}
%\description
%   This function is designed to plot the data which was saved with
%   'save_plot' in a fits-file. Using the qualifiers you can choose
%   how the data should be displayed. By default the model and the
%   data is plotted in counts/bin.
%   
%   The data, model and diff qualifier can be combined with the
%   qualifier 'flux' in order to plot in photons/s/cm^2/keV instead of
%   counts/bin.
%   
%   To define the colors of the plot, you have to provide an array of
%   colors. The length of the array is equal to the number of loaded
%   spectra, as each color is associated with the spectra in the same order
%   given in the fits-file.
%   
%   If you provide the array 'dataset', only these spectra are
%   plotted. The numbers are assinged according to the order of the
%   single spectra in the extension of the fits-file. Using for
%   example 'fv <filename>', you can easily look up these numbers, as the
%   instrument which created the desired spectrum is also named in the
%   header of each extension.
%\examples
%   % save one dataset in a fits-file
%   save_plot("my_data",1);
%   % plot data and model in flux and the residuals in an additional panel
%   % -the data should be in red, color(2) and the data in black, color(1)
%   % -the ranges are set automatically
%   multiplot([3,1]);
%   fplot("my_data",2;type="data",space="flux",auto);
%   ofplot("my_data",1;type="model",space="flux",auto);
%   ofplot("my_data",2;type="res",auto);
%\seealso{save_plot}
%!%-
   
   
  variable rng = [0.,0.];
  variable xrng = [0.,0.];
  variable col;
  variable filename = "save_plot.fits";
  variable dataset;
  variable typ = qualifier("type","default");
  variable space = qualifier("space","counts");
    
   %% Plot Flux or Counts??
  variable ym = "model";
  variable yd = "value";
  variable yr = "err";
  if (space == "density") {
    ym = "model_density";
    yd = "density";
    yr = "density_err";
  } else if (space == "flux"){
    ym = "model_flux";
    yd = "flux";
    yr = "flux_err";    
  }
  
   variable a,b,c;
   switch(_NARGS)
   {case 1:filename = ();}
   {case 2: (filename, col) = (); }
   {case 3: (filename, col, dataset) = (); }   
   {help(_function_name()); return; }
   
   %% get number of datasets
   if (is_substr(filename,".fits") == 0 &  is_substr(filename,".FITS") == 0){
      filename = filename+".fits";      
   }
   
   variable n = fits_read_key(filename,"NUMDAT");
   ifnot (__is_initialized(&dataset)) {
      dataset = [1:n];
   }
   %% check if the required datasets are present
   variable ind = [where(dataset > n), where(dataset < 1)];
   if (length(ind) > 0){
      message("The required dataset '"+string(dataset[ind][0])
	      +"' is not present in "+filename);
      return;
   }
   n = length(dataset);
   
   ifnot (__is_initialized(&col)) {
      col = [1:length(dataset)];
   } 
   %% check if number of colors match the number of spectra
   else if (length(col) != length(dataset)){
    message("The number of colors and spectra have to be equal!");  
    return;
   }
  
  
  
  
   %% ++++++++  read data  +++++++++
   variable i,fits_ext;
   variable dat = Struct_Type[n];
   variable inf = Struct_Type[n];
   variable diff = Array_Type[n];
   variable res = Array_Type[n];
   variable ratio = Array_Type[n];
   variable ratio_err = Array_Type[n];
   variable tmp,j;
   variable plo = get_plot_options;

  for (i = 0; i < n; i++){
     %% get data
     fits_ext = filename+"+"+string(dataset[i]);
     
       %% write additional information
     inf[i] = struct{
       fitfun = fits_read_key(fits_ext,"fitfun"),
       model = fits_read_key(fits_ext,"model"),
       exposure = fits_read_key(fits_ext,"exposure")
     };

     %% get the real data
     dat[i] = fits_read_table(fits_ext);
     
     %% write also the data in bin-density
     dat[i] = struct_combine(dat[i], "density", "density_err","model_density");
     dat[i].density = dat[i].value/(dat[i].bin_hi-dat[i].bin_lo)/inf[i].exposure;
     dat[i].density_err = dat[i].err/(dat[i].bin_hi-dat[i].bin_lo)/inf[i].exposure;
     dat[i].model_density = dat[i].model/(dat[i].bin_hi-dat[i].bin_lo)/inf[i].exposure;
     
     %% convert units if desired
     variable unit = fits_read_key(fits_ext,"tunit1");
     if (qualifier("unit",unit) != unit) {
       unit = qualifier("unit",unit);
       dat[i] = _A(dat[i]);
       dat[i].model = reverse(dat[i].model);
       dat[i].model_flux = reverse(dat[i].model_flux);
       dat[i].flux = reverse(dat[i].flux);
       dat[i].flux_err = reverse(dat[i].flux_err);
     }
      diff[i] = (get_struct_field(dat[i],yd) - get_struct_field(dat[i],ym));
      res[i] = (dat[i].value - dat[i].model)/dat[i].err;
      ratio[i] = dat[i].value/dat[i].model;
      ratio_err[i] = dat[i].err/dat[i].model;
   }
   
   %% ++++++++ set labels ++++++++
   if (not qualifier_exists("nolabel")) {
     xlabel("Energy [keV]");
     if (unit == "A")  xlabel(`Wavelength [\A]`);
     if (space == "flux") {
       ylabel(`photons s\u-1\d cm\u-2\d keV\u-1\d`);
     }
     else if (space == "density") {
       ylabel("counts/keV/s");
     }
     else {
       ylabel("counts/bin");
     }
	   
     if (typ == "ratio")  ylabel("Ratio");
     if (typ == "res")  ylabel(`\gx`);
   }
   
   %% ++++++++  set ranges +++++++
    
   if (qualifier_exists("auto") or qualifier_exists("xauto")){
      xrng[0] = min_struct_field(dat,"bin_lo");
      xrng[1] = max_struct_field(dat,"bin_hi");
      xrange(xrng[0],xrng[1]);
   }
   
   
   if (qualifier_exists("auto") or qualifier_exists("yauto")){
      if (typ == "ratio") {
	 tmp = array_map(Double_Type, &min,ratio);
	 j = where(tmp == min(tmp))[0];
	 rng[0] = min(tmp)- ratio_err[j][where(ratio[j] == min(tmp))[0]];
	 
	 tmp = array_map(Double_Type, &max,ratio);
	 j = where(tmp == max(tmp))[0];
	 rng[1] = max(tmp) + ratio_err[j][where(ratio[j] == max(tmp))[0]];
      }
      else if (typ == "res") {
	 rng[0] = min(array_map(Double_Type, &min,res))-1;
	 rng[1] = max(array_map(Double_Type, &max,res))+1;
      }
      else if (typ == "diff") {
	 tmp = array_map(Double_Type, &min,diff);
	 j = where(tmp == min(tmp))[0];
	 rng[0] = min(tmp) - get_struct_field(dat[j],yr)[where(diff[j] == min(tmp))[0]];
	 
	 tmp = array_map(Double_Type, &max,diff);
	 j = where(tmp == max(tmp))[0];
	 rng[1] = max(tmp) + get_struct_field(dat[j],yr)[where(diff[j] == max(tmp))[0]];
      } else{
	 
	 tmp = Double_Type[n];
	 for (j = 0; j < n; j++ ) {
	    tmp[j] = min(get_struct_field(dat[j],ym));
	 }
	 j = where(tmp == min(tmp))[0];
	 rng[0] = min(tmp) - 
	   get_struct_field(dat[j],yr)[where(get_struct_field(dat[j],ym) == min(tmp))[0]];
	 
	 % make sure that ymin is not negative
	 if (rng[0] < 0) rng[0]=0;
	 
	 for (j = 0; j < n; j++ ) {
	    tmp[j] = max(get_struct_field(dat[j],ym));
	 }
	 j = where(tmp == max(tmp))[0];
	 rng[1] = max(tmp) +
	   get_struct_field(dat[j],yr)[where(get_struct_field(dat[j],ym) == max(tmp))[0]];
      }
      yrange(rng[0],rng[1]);   
   }
   
   %% +++++++  PLOT +++++++++++++
   % options
   
   % initialize plot, if a new plot should be created
  plo = get_plot_options;
  if ( (not qualifier_exists("overplot")) ){ %&& qualifier_exists("auto")){
    if (typ == "ratio" || typ == "res" || typ == "diff"){
      ylin;
    }
    init_plot;
   }
  if (typ == "model"){
    for (i =0;i<n;i++){
      color(col[i]);
      connect_points(1);
      ohplot(dat[i].bin_lo,dat[i].bin_hi,
	     get_struct_field(dat[i],ym));
    }
  } else if (typ == "data"){
    for (i =0;i<n;i++){
      color(col[i]);
      connect_points(0);
      oplot_with_err(dat[i].bin_lo,dat[i].bin_hi,
		     get_struct_field(dat[i],yd),
		     get_struct_field(dat[i],yr);xminmax);
    }
  } else if (typ == "ratio"){
    for (i =0;i<n;i++){
      %% line at 1
      connect_points(1);
      color(1);
      ohplot(dat[i].bin_lo,dat[i].bin_hi,ratio[i]*0+1);
      
      color(col[i]);
      connect_points(0);
      oplot_with_err(dat[i].bin_lo,dat[i].bin_hi,
		     ratio[i],
		     ratio_err[i];xminmax);
    }
   } else if (typ == "res"){
      for (i =0;i<n;i++){
	 %% line at 0
	 connect_points(1);
	 color(1);
	 ohplot(dat[i].bin_lo,dat[i].bin_hi,res[i]*0);
	 color(col[i]);
	 connect_points(0);
	 oplot_with_err(dat[i].bin_lo,dat[i].bin_hi,
			 res[i],
			 res[i]*0+1.;xminmax);
      }
   } else if (typ == "diff"){
      for (i =0;i<n;i++){
	 %% line at 0
	 connect_points(1);
	color(1);
	ohplot(dat[i].bin_lo,dat[i].bin_hi,diff[i]*0);
	
	 color(col[i]);
	connect_points(0);
	oplot_with_err(dat[i].bin_lo,dat[i].bin_hi,
			diff[i],
			get_struct_field(dat[i],yr);xminmax);
      }
   } else {
      %% else: plot DATA and MODEL
     for (i =0;i<n;i++){
       color(col[i]+2);
       connect_points(1);

       ohplot(dat[i].bin_lo,dat[i].bin_hi,
	      get_struct_field(dat[i],ym));
       
       
       color(col[i]);
       oplot_with_err(dat[i].bin_lo,dat[i].bin_hi,
		      get_struct_field(dat[i],yd),
		      get_struct_field(dat[i],yr);xminmax);
       
     }
   }
   set_plot_options(plo);  
}

%
% Code from src/slang/fread_struct.sl
%
%%%%%%%%%%%%%%%
define fread_struct()
%%%%%%%%%%%%%%%
%!%+
%\function{fread_struct}
%\synopsis{Read binary data from a file into a pre-defined structure}
%\usage{fread_struct(Struct_Type s, File_Type fp);}
%\qualifiers{
%    \qualifier{char_to_string}{convert fields of an array of chars (Char_Type[])
%                     with length greater one into strings}
%    \qualifier{chatty}{be verbose}
%}
%\description
%    This function uses `fread' to reads binary data from a file into
%    the fields of a structure. All fields have to have a defined
%    data type! That is at least each field value has to be set to a
%    certain `DataType_Type'. In case field values are arrays, the
%    corresponding amount of objects of the array's data type are read
%    from the file. The fields have to be defined in the same order as
%    their corresponding objects should be read from the file. Note
%    that the function does not return anything, but the fields of the
%    given structure are updated instead! Finally, make sure to use
%    the correct number of bits to be read for each field. In doubt
%    always use, e.g., Int32_Type instead of Integer_Type as the
%    latter might depend on how S-lang was compiled.
%\example
%    % define the structure to be read
%    variable s = struct {
%      count = Int16_Type, % first, read one 16-bit integer
%      list = Float32_Type[10], % secondly, read 10x 32-bit floats
%      msg = Char_Type[64] % finally, read 64x characters
%    };
%    % read the file
%    variable fp = fopen("mybinaryformat.file", "r");  
%    fread_struct(s, fp; char_to_string);
%    ()=fclose(fp);
%    % print the message, which will be a string due to the
%    % char_to_string-qualifier
%    message(s.msg);  
%\seealso{fread}
%!%-
{
  variable s, fp;
  switch (_NARGS)
    { case 2: (s, fp) = (); }
    { help(_function_name); return; }

  % define variables and loop over all fields of the structure
  variable fieldnames = get_struct_field_names(s);
  variable fieldname;
  variable chatty = qualifier_exists("chatty");
  foreach fieldname (fieldnames) {
    % determine data type and size of the field
    variable field = get_struct_field(s, fieldname);
    variable type = typeof(field) == DataType_Type ? field : typeof(field);
    variable size = 1;
    if (type == Array_Type) {
      type = _typeof(field);
      size = length(field);
    }
    % print infos
    if (chatty) {
      ()=fprintf(stdout, "reading %d objects of %s into '%s'... ", size, string(type), fieldname);
      ()=fflush(stdout);
    }
    % read the data!
    variable num = fread(&field, type, size, fp);
    % print infos
    if (chatty) {
      if (num == -1) {
	()=fprintf(stdout, "failed");
	()=fflush(stdout);
	throw ReadError, "fread failed";
      } else {
        ()=fprintf(stdout, "got %d", num);
      }
    }
    % convert character array into string
    if (type == Char_Type && size > 1 && qualifier_exists("char_to_string")) {
      if (chatty) { ()=fprintf(stdout, " -> to string"); }
      field =  pack(sprintf("c%d", size), field);
    }
    if (chatty) { ()=fprintf(stdout, "\n"); ()=fflush(stdout); }
    % set the field to the read data
    set_struct_field(s, fieldname, field);
  }
}

%
% Code from src/fitting/parameters/freeParameters.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define freeParameters()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{freeParameters}
%\synopsis{find all free parameters of the current fit-function}
%\usage{Integer_Type[] freeParameters()}
%\description
%    Free parameters are not frozen, tied to another one,
%    or derived as functions of other parameters.
%\seealso{thawedParameters}
%!%-
{
  variable ind = {};
  variable par;
  foreach par ( get_params() )
    if(par.freeze==0 && par.tie==NULL && par.fun==NULL)
      list_append(ind, par.index);
  return list_to_array(ind, Integer_Type);
}

%
% Code from src/slang/astronomy/frequency2velocity.sl
%
define frequency2velocity()
%!%+
%\function{frequency2velocity}
%\synopsis{Calculate the velocity to a given frequency.}
%\usage{frequency2velocity(freq,restfreq);}
%\qualifiers{
%    \qualifier{veldef}{The velocity definition which must be one of
%                   OPTICAL, RADIO, or TRUE.  Defaults to RADIO.}
%}
%\description
%    Convert frequency to velocity (m/s) using the given rest
%    frequency and velocity definition. The units (Hz, MHz, GHz, etc)
%    of the frequencies to convert must match that of the rest frequency argument.
%    
%    Adopted from the gbtidl script freqtovelo.pro
%!%-
{
   variable freq, restfreq;
   switch(_NARGS)
     { case 2: (freq,restfreq) = (); }
     { help(_function_name()); return; }

   variable veldef = qualifier("veldef","RADIO");
   switch(veldef)
     {
      case "RADIO": variable vel = Const_c*1e-2*(1. - freq / restfreq);
     }
     {
      case "OPTICAL": vel = Const_c*1e-2*(restfreq / freq -1.);
     }
     {
      case "TRUE": variable g = (freq / restfreq)^2;
	vel = Const_c*1e-2*(1. - g) / (1. + g);
     }
     {
       message( "unrecognized velocity definition");
     } 

    return vel;
}

%
% Code from src/slang/math/ftest_xspec.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%%%%%%%%%%%%%%%%%%
define ftest_xspec()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{ftest_xspec}
%\synopsis{calculates the F-test probability as in xspec}
%\usage{Double_Type ftest_xspec(chisq2, dof2, chisq1, dof1)}
%\description
%    The new chi-square and DOF, chisq2 and dof2,
%    should come from adding an extra model component
%    to (or thawing a frozen parameter of)
%    the model which gave chisq1 and dof1.
%    If the F-test probability is low then it is
%    reasonable to add the extra model component.
%
%    WARNING: It is not correct to use the F-test statistic
%    to test for the presence of a line
%    (see Protassov et al astro-ph/0201457).
%!%-
{
  variable chi2, dof2, chi1, dof1;
  switch(_NARGS)
  { case 4: (chi2, dof2, chi1, dof1) = (); }
  { return help(_function_name()); }

  if (chi2 > chi1) {
    vmessage("warning %s: Fit did not improve! chi2 should be smaller than chi1!",_function_name());
    return -_NaN;}
  if (dof2 > dof1) {
    vmessage("warning %s: DOF increased! dof2 should be less than dof1!",_function_name());
    return -_NaN;}
  return gsl->beta_inc(dof2/2., (dof1-dof2)/2., chi2*1./chi1);
}


#endif

%
% Code from src/slang/astronomy/g_earth.sl
%
define g_earth()
%!%+
%\function{g_earth}
%\synopsis{Calculates the acceleration at the Earth's surface as a function of latitude}
%\usage{Double_Type g = g_earth(Double_Type phi);}
%
%\qualifiers{
%\qualifier{deg}{phi is in degrees, not in radian.}
%\qualifier{altitude}{altitude of the observer above the geoid [m]}
%\qualifier{mks}{return g in m/s^2, not in cm/s^2}
%\qualifier{allen}{use the equation of Allen (1973, Astrophys. Quantities)}
%\qualifier{almanac}{Urban & Seidelman (2013; Expl. Suppl. Astron. Almanac)}
%}
%\description
%    This code calculates the surface acceleration on the earth as
%    a function of the latitude (in rad). The equations used are 
%    from a compilation of different approximations to g by
%    Mangum & Wallace (2015), PASP 127, 74.
%
%    The default returned is for the WGS84 geoid.
%
%    This routine is array safe in phi or altitude.
%\seealso{dms2deg}
%!%-
{
    variable phi;
    switch(_NARGS)
    { case 1: phi = (); }
    { help(_function_name()); return; }

    if (qualifier_exists("deg")) {
	phi*=PI/180.;
    }


    variable h0=qualifier("altitude",0.)/1000.; % m->km

    variable g,sin2phi,sinphi,cosphi,cos2phi;
    if (qualifier_exists("allen")) {
	sinphi=sin(phi);
	sin2phi=sin(2.*phi);
	g=9.757823*(1.+0.005301*sinphi*sinphi-0.0000059*sin2phi*sin2phi-0.000284*h0);
    } else {
	if (qualifier_exists("almanac")) {
	    cos2phi=cos(2.*phi);
	    g=9.784*(1.-0.0026*cos2phi-0.00028*h0);
	} else {
	    % WGS84
	    sinphi=sin(phi);
	    g=9.7803267714*((1.+0.00193185138639*sinphi*sinphi)/sqrt(1.-0.00669437999013*sinphi*sinphi))-0.02253-0.0027774*h0;

	}
    }
    if (qualifier_exists("mks")){
	return g;
    }
    return g*100.;
}

%
% Code from src/slang/astronomy/galLB_from_RAdec.sl
%
define galLB_from_RAdec()
%!%+
%\function{galLB_from_RAdec}
%\synopsis{convert equatorial (RA, dec) coordinates to galactic (l, b) coordinates}
%\usage{(l, b) = galLB_from_RAdec(RA, dec);}
%\qualifiers{
%\qualifier{ra_unit }{ [\code{="deg"}]: set the unit of the right ascension \code{RA}
%                         \code{l_unit="rad"} \code{=>}  \code{RA} in rad
%                         \code{l_unit="hms"} \code{=>}  \code{RA} in hours as a scalar
%                         or an array of the form \code{[H, M]} or \code{[H, M, S]} }
%\qualifier{dec_unit}{ [\code{="deg"}]: set the unit of the declination \code{dec}}
%}
%\description
%    This function is deprecated. Please use equatorial2galactic instead.
%\seealso{equatorial2galactic,galactic2equatorial}
%!%-
{
    variable RA, dec;
    switch(_NARGS)
  { case 2: (RA, dec) = (); }
  { help(_function_name()); return; }
  
    variable  ra_unit = qualifier( "ra_unit", "deg");
    variable dec_unit = qualifier("dec_unit", "deg");
    RA  = angle_to_rad(RA ; unit =  ra_unit);
    dec = angle_to_rad(dec; unit = dec_unit);

  variable ell,b;

  (ell,b)=equatorial2galactic(RA,dec; blaauw);

  return ell*180./PI, b*180./PI;

}

%
% Code from src/misc/astronomy/galactic_rotation_velocity.sl
%
define galactic_rotation_velocity(r)
%!%+
%\function{galactic_rotation_velocity}
%\usage{Double_Type galactic_rotation_velocity(Double_Type r; model=m)}
%\qualifiers{
%\qualifier{model}{\code{m=1} selects Model I (B&T, Fig. 2.20), \code{m=2} selects Model II (B&T, Fig. 2.22)
%             If neither \code{m=1} nor \code{m=2} is specified, the constant 220 is returned.}
%}
%\description
%    \code{r} is the distance from the Galactic Center in kpc.
%    (\code{0.18 <= r <= 11.9} is covered by a lookup table.)
%    The return value is the Galactic rotation velocity in km/s.
%\seealso{J. Binney & S. Tremaine: Galactic Dynamics (2nd ed.), Sect. 2.7 (pp. 110ff)}
%!%-
{
  switch(qualifier("model"))
  { case 1: return interpol(r,
 [0.18, 0.20, 0.23, 0.25, 0.27, 0.29, 0.31, 0.33, 0.35, 0.38, 0.40, 0.42, 0.44, 0.46, 0.48, 0.51, 0.53, 0.55, 0.57, 0.59, 0.61,
  0.63, 0.66, 0.68, 0.70, 0.72, 0.74, 0.76, 0.78, 0.81, 0.83, 0.85, 0.87, 0.89, 0.91, 0.94, 0.96, 0.98, 1.00, 1.02, 1.04, 1.06, 1.09, 1.11,
  1.13, 1.15, 1.17, 1.19, 1.22, 1.24, 1.26, 1.28, 1.30, 1.32, 1.34, 1.37, 1.39, 1.41, 1.43, 1.45, 1.47, 1.49, 1.52, 1.54, 1.56, 1.58, 1.60,
  1.62, 1.65, 1.67, 1.69, 1.71, 1.73, 1.75, 1.77, 1.80, 1.82, 1.84, 1.86, 1.88, 1.90, 1.92, 1.95, 1.97, 1.99, 2.01, 2.03, 2.05, 2.08, 2.10,
  2.12, 2.14, 2.16, 2.18, 2.20, 2.23, 2.25, 2.27, 2.29, 2.31, 2.33, 2.35, 2.38, 2.40, 2.42, 2.44, 2.46, 2.48, 2.51, 2.53, 2.55, 2.57, 2.59,
  2.61, 2.63, 2.66, 2.68, 2.70, 2.72, 2.74, 2.76, 2.78, 2.81, 2.83, 2.85, 2.87, 2.89, 2.91, 2.94, 2.96, 2.98, 3.00, 3.02, 3.04, 3.06, 3.09,
  3.11, 3.13, 3.15, 3.17, 3.19, 3.22, 3.24, 3.26, 3.28, 3.30, 3.32, 3.34, 3.37, 3.39, 3.41, 3.43, 3.45, 3.47, 3.49, 3.52, 3.54, 3.56, 3.58,
  3.60, 3.62, 3.65, 3.67, 3.69, 3.71, 3.73, 3.75, 3.77, 3.80, 3.82, 3.84, 3.86, 3.88, 3.90, 3.92, 3.95, 3.97, 3.99, 4.01, 4.03, 4.05, 4.08,
  4.10, 4.12, 4.14, 4.16, 4.18, 4.20, 4.23, 4.25, 4.27, 4.29, 4.31, 4.33, 4.35, 4.38, 4.40, 4.42, 4.44, 4.46, 4.48, 4.51, 4.53, 4.55, 4.57,
  4.59, 4.61, 4.63, 4.66, 4.68, 4.70, 4.72, 4.74, 4.76, 4.78, 4.81, 4.83, 4.85, 4.87, 4.89, 4.91, 4.94, 4.96, 4.98, 5.00, 5.02, 5.04, 5.06,
  5.09, 5.11, 5.13, 5.15, 5.17, 5.19, 5.22, 5.24, 5.26, 5.28, 5.30, 5.32, 5.34, 5.37, 5.39, 5.41, 5.43, 5.45, 5.47, 5.49, 5.52, 5.54, 5.56,
  5.58, 5.60, 5.62, 5.65, 5.67, 5.69, 5.71, 5.73, 5.75, 5.77, 5.80, 5.82, 5.84, 5.86, 5.88, 5.90, 5.92, 5.95, 5.97, 5.99, 6.01, 6.03, 6.05,
  6.08, 6.10, 6.12, 6.14, 6.16, 6.18, 6.20, 6.23, 6.25, 6.27, 6.29, 6.31, 6.33, 6.35, 6.38, 6.40, 6.42, 6.44, 6.46, 6.48, 6.51, 6.53, 6.55,
  6.57, 6.59, 6.61, 6.63, 6.66, 6.68, 6.70, 6.72, 6.74, 6.76, 6.78, 6.81, 6.83, 6.85, 6.87, 6.89, 6.91, 6.94, 6.96, 6.98, 7.00, 7.02, 7.04,
  7.06, 7.09, 7.11, 7.13, 7.15, 7.17, 7.19, 7.22, 7.24, 7.26, 7.28, 7.30, 7.32, 7.34, 7.37, 7.39, 7.41, 7.43, 7.45, 7.47, 7.49, 7.52, 7.54,
  7.56, 7.58, 7.60, 7.62, 7.65, 7.67, 7.69, 7.71, 7.73, 7.75, 7.77, 7.80, 7.82, 7.84, 7.86, 7.88, 7.90, 7.92, 7.95, 7.97, 7.99, 8.01, 8.03,
  8.05, 8.08, 8.10, 8.12, 8.14, 8.16, 8.18, 8.20, 8.23, 8.25, 8.27, 8.29, 8.31, 8.33, 8.35, 8.38, 8.40, 8.42, 8.44, 8.46, 8.48, 8.51, 8.53,
  8.55, 8.57, 8.59, 8.61, 8.63, 8.66, 8.68, 8.70, 8.72, 8.74, 8.76, 8.78, 8.81, 8.83, 8.85, 8.87, 8.89, 8.91, 8.94, 8.96, 8.98, 9.00, 9.02,
  9.04, 9.06, 9.09, 9.11, 9.13, 9.15, 9.17, 9.19, 9.22, 9.24, 9.26, 9.28, 9.30, 9.32, 9.34, 9.37, 9.39, 9.41, 9.43, 9.45, 9.47, 9.49, 9.52,
  9.54, 9.56, 9.58, 9.60, 9.62, 9.65, 9.67, 9.69, 9.71, 9.73, 9.75, 9.77, 9.80, 9.82, 9.84, 9.86, 9.88, 9.90, 9.92, 9.95, 9.97, 9.99, 10.01,
  10.03, 10.05, 10.08, 10.10, 10.12, 10.14, 10.16, 10.18, 10.20, 10.23, 10.25, 10.27, 10.29, 10.31, 10.33, 10.35, 10.38, 10.40, 10.42, 10.44,
  10.46, 10.48, 10.51, 10.53, 10.55, 10.57, 10.59, 10.61, 10.63, 10.66, 10.68, 10.70, 10.72, 10.74, 10.76, 10.78, 10.81, 10.83, 10.85, 10.87,
  10.89, 10.91, 10.94, 10.96, 10.98, 11.00, 11.02, 11.04, 11.06, 11.09, 11.11, 11.13, 11.15, 11.17, 11.19, 11.22, 11.24, 11.26, 11.28, 11.30,
  11.32, 11.34, 11.37, 11.39, 11.41, 11.43, 11.45, 11.47, 11.49, 11.52, 11.54, 11.56, 11.58, 11.60, 11.62, 11.65, 11.67, 11.69, 11.71, 11.73,
  11.75, 11.77, 11.80, 11.82, 11.84, 11.86, 11.88, 11.90, 11.901],
 [101.84, 102.64, 103.63, 105.30, 106.65, 108.33, 109.78, 111.46, 112.69, 114.48, 115.94, 117.39, 118.62, 119.76, 121.08, 122.31,
  123.32, 124.55, 125.90, 127.13, 128.47, 129.70, 131.04, 132.39, 133.62, 134.85, 136.08, 137.09, 138.43, 139.30, 140.22, 141.23, 142.23,
  143.24, 144.25, 145.14, 146.15, 147.05, 148.05, 148.95, 149.96, 150.85, 151.86, 152.86, 153.87, 154.88, 155.89, 156.78, 157.68, 158.68,
  159.47, 160.25, 161.03, 162.04, 162.82, 163.72, 164.62, 165.40, 165.96, 166.63, 167.30, 167.97, 168.53, 169.09, 169.76, 170.32, 170.99,
  171.78, 172.34, 173.12, 173.90, 174.69, 175.58, 176.48, 177.37, 178.16, 178.72, 179.27, 179.83, 180.28, 180.84, 181.18, 181.74, 182.30,
  182.74, 183.30, 183.75, 184.20, 184.53, 184.98, 185.32, 185.88, 186.32, 186.77, 187.33, 187.78, 188.34, 188.90, 189.57, 189.91, 190.58,
  191.14, 191.92, 192.37, 192.93, 193.38, 193.71, 194.05, 194.49, 194.72, 195.17, 195.61, 195.95, 196.28, 196.62, 196.96, 197.29, 197.40,
  197.85, 197.96, 198.30, 198.52, 198.86, 199.31, 199.64, 200.09, 200.54, 200.76, 201.21, 201.66, 202.10, 202.55, 202.89, 203.11, 203.33,
  203.78, 204.01, 204.23, 204.34, 204.57, 204.90, 205.13, 205.35, 205.57, 205.80, 206.02, 206.24, 206.47, 206.92, 207.14, 207.36, 207.59,
  207.92, 208.26, 208.48, 208.71, 208.93, 209.38, 209.60, 209.71, 210.05, 210.16, 210.27, 210.50, 210.72, 210.94, 211.06, 211.17, 211.39,
  211.62, 211.84, 212.06, 212.29, 212.40, 212.62, 212.74, 212.96, 213.18, 213.41, 213.63, 213.85, 214.08, 214.19, 214.30, 214.53, 214.75,
  214.97, 214.97, 215.20, 215.31, 215.42, 215.53, 215.53, 215.53, 215.64, 215.64, 215.76, 215.87, 216.09, 216.20, 216.32, 216.54, 216.76,
  216.99, 217.21, 217.44, 217.66, 217.77, 217.88, 217.88, 218.11, 218.11, 218.22, 218.33, 218.44, 218.55, 218.67, 218.67, 218.67, 218.78,
  218.78, 218.89, 218.89, 218.89, 219.00, 219.11, 219.11, 219.23, 219.45, 219.45, 219.56, 219.90, 220.01, 220.01, 220.12, 220.34, 220.34,
  220.34, 220.46, 220.46, 220.46, 220.46, 220.57, 220.57, 220.57, 220.68, 220.68, 220.68, 220.79, 220.79, 220.79, 220.79, 220.90, 220.90,
  220.90, 221.02, 221.13, 221.24, 221.24, 221.35, 221.46, 221.58, 221.69, 221.80, 221.80, 221.91, 221.91, 221.91, 221.91, 221.91, 221.91,
  221.80, 221.80, 221.80, 221.91, 221.91, 221.91, 221.91, 221.91, 221.91, 221.91, 221.91, 221.91, 222.02, 222.02, 222.02, 222.14, 222.14,
  222.25, 222.25, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36, 222.36,
  222.36, 222.25, 222.36, 222.36, 222.14, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.36, 222.36,
  222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25, 222.25,
  222.25, 222.25, 222.25, 222.25, 222.25, 222.14, 222.14, 222.02, 222.02, 222.02, 222.02, 222.02, 222.02, 222.02, 222.02, 222.02, 222.02,
  221.91, 221.91, 221.91, 221.91, 221.91, 221.91, 221.91, 221.91, 221.91, 221.80, 221.80, 221.80, 221.80, 221.80, 221.80, 221.80, 221.80,
  221.80, 221.80, 221.69, 221.69, 221.69, 221.69, 221.69, 221.69, 221.69, 221.69, 221.46, 221.35, 221.35, 221.35, 221.35, 221.35, 221.24,
  221.24, 221.24, 221.24, 221.24, 221.24, 221.13, 221.13, 221.13, 221.13, 221.13, 221.02, 221.02, 221.02, 221.02, 220.90, 220.90, 220.90,
  220.90, 220.90, 220.90, 220.90, 220.79, 220.79, 220.79, 220.79, 220.79, 220.79, 220.79, 220.57, 220.57, 220.57, 220.46, 220.46, 220.46,
  220.46, 220.46, 220.34, 220.34, 220.34, 220.34, 220.23, 220.23, 220.23, 220.23, 220.23, 220.23, 220.23, 220.12, 220.12, 220.12, 220.12,
  220.12, 220.01, 220.01, 220.01, 220.01, 220.01, 219.90, 219.90, 219.90, 219.90, 219.90, 219.79, 219.79, 219.79, 219.79, 219.79, 219.67,
  219.45, 219.45, 219.45, 219.45, 219.45, 219.45, 219.34, 219.34, 219.34, 219.34, 219.34, 219.34, 219.34, 219.23, 219.23, 219.23, 219.23,
  219.23, 219.23, 219.11, 219.11, 219.11, 219.11, 219.11, 219.00, 219.00, 219.00, 219.00, 219.00, 219.00, 219.00, 218.89, 218.89, 218.89,
  218.89, 218.89, 218.89, 218.78, 218.67, 218.55, 218.55, 218.55, 218.44, 218.44, 218.44, 218.44, 218.44, 218.33, 218.33, 218.44, 218.33,
  218.33, 218.33, 218.33, 218.33, 218.33, 218.22, 218.22, 218.22, 218.22, 218.22, 218.22, 218.22, 218.22, 218.11, 218.11, 218.11, 218.11,
  218.11, 218.11, 218.11, 218.11, 217.99, 217.88, 218.11, 217.88, 217.88, 217.77, 217.77, 217.77, 217.66, 217.66, 217.66, 217.66, 217.66,
  217.66, 217.66, 217.55, 217.55]); }
  { case 2: return interpol(r,
 [0.18, 0.20, 0.23, 0.25, 0.27, 0.29, 0.31, 0.33, 0.36, 0.38, 0.40, 0.42, 0.44, 0.46, 0.48, 0.51, 0.53, 0.55, 0.57, 0.59, 0.61,
  0.64, 0.66, 0.68, 0.70, 0.72, 0.74, 0.76, 0.79, 0.81, 0.83, 0.85, 0.87, 0.89, 0.92, 0.94, 0.96, 0.98, 1.00, 1.02, 1.04, 1.07, 1.09, 1.11,
  1.13, 1.15, 1.17, 1.20, 1.22, 1.24, 1.26, 1.28, 1.30, 1.32, 1.35, 1.37, 1.39, 1.41, 1.43, 1.45, 1.48, 1.50, 1.52, 1.54, 1.56, 1.58, 1.61,
  1.63, 1.65, 1.67, 1.69, 1.71, 1.73, 1.76, 1.78, 1.80, 1.82, 1.84, 1.86, 1.89, 1.91, 1.93, 1.95, 1.97, 1.99, 2.01, 2.04, 2.06, 2.08, 2.10,
  2.12, 2.14, 2.17, 2.19, 2.21, 2.23, 2.25, 2.27, 2.29, 2.32, 2.34, 2.36, 2.38, 2.40, 2.42, 2.45, 2.47, 2.49, 2.51, 2.53, 2.55, 2.57, 2.60,
  2.62, 2.64, 2.66, 2.68, 2.70, 2.73, 2.75, 2.77, 2.79, 2.81, 2.83, 2.85, 2.88, 2.90, 2.92, 2.94, 2.96, 2.98, 3.01, 3.03, 3.05, 3.07, 3.09,
  3.11, 3.13, 3.16, 3.18, 3.20, 3.22, 3.24, 3.26, 3.29, 3.31, 3.33, 3.35, 3.37, 3.39, 3.41, 3.44, 3.46, 3.48, 3.50, 3.52, 3.54, 3.57, 3.59,
  3.61, 3.63, 3.65, 3.67, 3.69, 3.72, 3.74, 3.76, 3.78, 3.80, 3.82, 3.85, 3.87, 3.89, 3.91, 3.93, 3.95, 3.97, 4.00, 4.02, 4.04, 4.06, 4.08,
  4.10, 4.13, 4.15, 4.17, 4.19, 4.21, 4.23, 4.25, 4.28, 4.30, 4.32, 4.34, 4.36, 4.38, 4.41, 4.43, 4.45, 4.47, 4.49, 4.51, 4.54, 4.56, 4.58,
  4.60, 4.62, 4.64, 4.66, 4.69, 4.71, 4.73, 4.75, 4.77, 4.79, 4.82, 4.84, 4.86, 4.88, 4.90, 4.92, 4.94, 4.97, 4.99, 5.01, 5.03, 5.05, 5.07,
  5.10, 5.12, 5.14, 5.16, 5.18, 5.20, 5.22, 5.25, 5.27, 5.29, 5.31, 5.33, 5.35, 5.38, 5.40, 5.42, 5.44, 5.46, 5.48, 5.50, 5.53, 5.55, 5.57,
  5.59, 5.61, 5.63, 5.66, 5.68, 5.70, 5.72, 5.74, 5.76, 5.78, 5.81, 5.83, 5.85, 5.87, 5.89, 5.91, 5.94, 5.96, 5.98, 6.00, 6.02, 6.04, 6.06,
  6.09, 6.11, 6.13, 6.15, 6.17, 6.19, 6.22, 6.24, 6.26, 6.28, 6.30, 6.32, 6.34, 6.37, 6.39, 6.41, 6.43, 6.45, 6.47, 6.50, 6.52, 6.54, 6.56,
  6.58, 6.60, 6.62, 6.65, 6.67, 6.69, 6.71, 6.73, 6.75, 6.78, 6.80, 6.82, 6.84, 6.86, 6.88, 6.90, 6.93, 6.95, 6.97, 6.99, 7.01, 7.03, 7.06,
  7.08, 7.10, 7.12, 7.14, 7.16, 7.18, 7.21, 7.23, 7.25, 7.27, 7.29, 7.31, 7.34, 7.36, 7.38, 7.40, 7.42, 7.44, 7.46, 7.49, 7.51, 7.53, 7.55,
  7.57, 7.59, 7.62, 7.64, 7.66, 7.68, 7.70, 7.72, 7.75, 7.77, 7.79, 7.81, 7.83, 7.85, 7.87, 7.90, 7.92, 7.94, 7.96, 7.98, 8.00, 8.03, 8.05,
  8.07, 8.09, 8.11, 8.13, 8.15, 8.18, 8.20, 8.22, 8.24, 8.26, 8.28, 8.31, 8.33, 8.35, 8.37, 8.39, 8.41, 8.43, 8.46, 8.48, 8.50, 8.52, 8.54,
  8.56, 8.59, 8.61, 8.63, 8.65, 8.67, 8.69, 8.71, 8.74, 8.76, 8.78, 8.80, 8.82, 8.84, 8.87, 8.89, 8.91, 8.93, 8.95, 8.97, 8.99, 9.02, 9.04,
  9.06, 9.08, 9.10, 9.12, 9.15, 9.17, 9.19, 9.21, 9.23, 9.25, 9.27, 9.30, 9.32, 9.34, 9.36, 9.38, 9.40, 9.43, 9.45, 9.47, 9.49, 9.51, 9.53,
  9.55, 9.58, 9.60, 9.62, 9.64, 9.66, 9.68, 9.71, 9.73, 9.75, 9.77, 9.79, 9.81, 9.83, 9.86, 9.88, 9.90, 9.92, 9.94, 9.96, 9.99, 10.01, 10.03,
  10.05, 10.07, 10.09, 10.11, 10.14, 10.16, 10.18, 10.20, 10.22, 10.24, 10.27, 10.29, 10.31, 10.33, 10.35, 10.37, 10.39, 10.42, 10.44, 10.46,
  10.48, 10.50, 10.52, 10.55, 10.57, 10.59, 10.61, 10.63, 10.65, 10.68, 10.70, 10.72, 10.74, 10.76, 10.78, 10.80, 10.83, 10.85, 10.87, 10.89,
  10.91, 10.93, 10.96, 10.98, 11.00, 11.02, 11.04, 11.06, 11.08, 11.11, 11.13, 11.15, 11.17, 11.19, 11.21, 11.24, 11.26, 11.28, 11.30, 11.32,
  11.34, 11.36, 11.39, 11.41, 11.43, 11.45, 11.47, 11.49, 11.52, 11.54, 11.56, 11.58, 11.60, 11.62, 11.64, 11.67, 11.69, 11.71, 11.73, 11.75,
  11.77, 11.80, 11.82, 11.84, 11.86, 11.88, 11.90, 11.92, 11.921],
 [144.15, 144.71, 145.95, 147.85, 149.31, 151.10, 153.11, 154.79, 156.70, 158.49, 160.28, 161.63, 162.86, 164.00, 164.99, 165.99,
  167.11, 168.12, 168.79, 169.91, 170.81, 171.59, 172.38, 173.28, 174.17, 174.96, 175.63, 176.41, 176.97, 177.76, 178.43, 178.99, 179.66,
  180.22, 180.89, 181.45, 182.01, 182.46, 183.13, 183.47, 183.92, 184.48, 184.81, 185.26, 185.71, 186.16, 186.60, 187.05, 187.50, 187.95,
  188.40, 188.73, 189.07, 189.52, 189.96, 190.19, 190.64, 191.08, 191.31, 191.64, 191.98, 192.32, 192.54, 192.88, 193.10, 193.44, 193.77,
  194.22, 194.44, 194.67, 195.00, 195.23, 195.56, 195.79, 196.12, 196.35, 196.57, 196.80, 197.24, 197.47, 197.69, 197.92, 198.14, 198.36,
  198.59, 198.81, 199.04, 199.26, 199.48, 199.71, 199.82, 200.04, 200.38, 200.38, 200.60, 200.83, 201.05, 201.28, 201.39, 201.61, 201.72,
  201.84, 201.95, 202.17, 202.40, 202.62, 202.73, 202.84, 203.07, 203.29, 203.52, 203.52, 203.74, 203.97, 204.08, 204.19, 204.41, 204.64,
  204.64, 204.86, 204.97, 205.09, 205.31, 205.42, 205.53, 205.76, 205.76, 205.98, 206.21, 206.21, 206.43, 206.65, 206.65, 206.88, 206.99,
  207.10, 207.33, 207.33, 207.55, 207.55, 207.77, 207.77, 208.00, 208.22, 208.11, 208.45, 208.45, 208.56, 208.67, 208.78, 208.89, 209.01,
  209.12, 209.23, 209.34, 209.57, 209.57, 209.79, 209.79, 210.01, 210.01, 210.24, 210.24, 210.46, 210.46, 210.69, 210.69, 210.91, 210.91,
  211.13, 211.13, 211.36, 211.36, 211.58, 211.58, 211.81, 211.81, 212.03, 212.03, 212.14, 212.25, 212.25, 212.48, 212.48, 212.70, 212.70,
  212.93, 212.93, 213.04, 213.04, 213.15, 213.15, 213.15, 213.26, 213.37, 213.49, 213.60, 213.71, 213.82, 213.82, 214.05, 214.05, 214.27,
  214.27, 214.38, 214.49, 214.61, 214.72, 214.72, 214.94, 214.94, 214.94, 215.17, 215.17, 215.17, 215.39, 215.39, 215.50, 215.61, 215.61,
  215.61, 215.61, 215.61, 215.84, 215.84, 215.84, 215.84, 216.06, 216.06, 216.17, 216.29, 216.29, 216.40, 216.40, 216.62, 216.51, 216.73,
  216.73, 216.73, 216.85, 216.96, 216.96, 216.96, 217.07, 217.18, 217.18, 217.18, 217.18, 217.18, 217.18, 217.29, 217.29, 217.29, 217.41,
  217.41, 217.41, 217.63, 217.63, 217.85, 217.85, 217.97, 218.08, 218.08, 218.30, 218.30, 218.30, 218.30, 218.30, 218.41, 218.53, 218.53,
  218.53, 218.53, 218.53, 218.53, 218.53, 218.64, 218.64, 218.64, 218.53, 218.64, 218.64, 218.75, 218.75, 218.75, 218.75, 218.75, 218.86,
  218.97, 218.97, 218.97, 218.97, 218.97, 218.97, 218.97, 218.97, 219.09, 219.20, 219.20, 219.20, 219.20, 219.20, 219.20, 219.31, 219.31,
  219.42, 219.42, 219.42, 219.42, 219.42, 219.42, 219.42, 219.53, 219.53, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65,
  219.53, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.76, 219.76, 219.76, 219.87, 219.87, 219.87,
  219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87, 219.87,
  219.87, 219.87, 219.87, 219.87, 219.98, 219.98, 219.98, 219.98, 219.98, 219.87, 219.87, 219.76, 219.76, 219.76, 219.76, 219.87, 219.87,
  219.76, 219.76, 219.76, 219.76, 219.76, 219.76, 219.76, 219.76, 219.76, 219.76, 219.76, 219.87, 219.76, 219.76, 219.76, 219.76, 219.76,
  219.76, 219.76, 219.76, 219.76, 219.76, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65, 219.65,
  219.65, 219.65, 219.53, 219.53, 219.42, 219.42, 219.42, 219.31, 219.20, 219.20, 219.20, 219.20, 219.20, 219.20, 219.20, 219.20, 219.20,
  219.20, 219.20, 219.20, 219.20, 219.20, 219.20, 219.09, 218.97, 218.97, 218.97, 218.97, 218.97, 218.97, 218.97, 218.97, 218.97, 218.97,
  218.97, 218.97, 218.97, 218.97, 218.86, 218.75, 218.75, 218.75, 218.75, 218.75, 218.75, 218.75, 218.75, 218.53, 218.53, 218.30, 218.30,
  218.30, 218.30, 218.30, 218.30, 218.30, 218.30, 218.30, 218.30, 218.19, 218.08, 218.08, 218.08, 218.08, 218.08, 218.08, 218.08, 218.08,
  218.08, 217.97, 217.97, 217.97, 217.85, 217.85, 217.85, 217.85, 217.85, 217.74, 217.85, 217.74, 217.63, 217.63, 217.63, 217.63, 217.63,
  217.63, 217.52, 217.52, 217.52, 217.52, 217.29, 217.29, 217.18, 217.18, 217.07, 216.96, 216.96, 216.96, 216.96, 216.96, 216.96, 216.96,
  216.85, 216.85, 216.85, 216.85, 216.73, 216.73, 216.73, 216.73, 216.62, 216.62, 216.51, 216.51, 216.51, 216.51, 216.51, 216.51, 216.40,
  216.40, 216.40, 216.29, 216.29, 216.29, 216.29, 216.29, 216.17, 216.17, 216.06, 216.06, 216.06, 216.06, 216.06, 216.06, 215.95, 215.84,
  215.84, 215.61, 215.73, 215.73]); }
  { return r*0 + 220; }

% % numbers obtained from:
% variable X1, Y1, X2, Y2;
% (X1, Y1) =  png_read_curve("model1_.png", 271,0, 1387,12,  1140,0, 23,250);
% message(strjoin_wrap(array_map(String_Type, &sprintf, "%.2f", X1[[::2]]); maxlen=144, initial="variable r =\n [", final="];", newline="  "));
% message(strjoin_wrap(array_map(String_Type, &sprintf, "%.2f", Y1[[::2]]); maxlen=144, initial="variable v =\n [", final="];", newline="  "));
% (X2, Y2) =  png_read_curve("model2_.png", 270,0, 1384,12,  1139,0, 23,250);
% message(strjoin_wrap(array_map(String_Type, &sprintf, "%.2f", X2[[::2]]); maxlen=144, initial="variable r =\n [", final="];"));
% message(strjoin_wrap(array_map(String_Type, &sprintf, "%.2f", Y2[[::2]]); maxlen=144, initial="variable v =\n [", final="];"));
}

%
% Code from src/fitting/fit-functions/galaxy_elliptical.sl
%

public variable is_table_loaded = 0;
public variable galaxy_elliptical_table;

define galaxy_elliptical_fit(lo,hi,par)
%!%+
%\function{galaxy_elliptical (fit-function)}
%\synopsis{fits a host galaxy in the optical/UV}
%\description
%	This function fits an elliptical host galaxy to the data, taking into
%       account the redshift z.
%  
%
%\examples
%    % data definition:
%    load_data("optical.pha");
%    fit_fun("galaxy_elliptical(1)+powerlaw(1)");
%    
%
%!%-
{

  if ( is_table_loaded == 0)
  {
    galaxy_elliptical_table =  ascii_read_table("/userdata/data/krauss/tanami/host_galaxy/all.dat",
						[{"%f", "x"}, {"%f", "E"}, {"%f", "S0"}, {"%f", "Sa"}, {"%f", "Sb"}, {"%f", "Sc"}]);
                                                % loading float (%f instead %F) given the precision
    % provide the table in some other form (function as a modul?) such that it works outside of remeis
    % or define at least a loading function, where the path can be specified (and provide table online)?
    is_table_loaded = 1;
  }
  variable a = galaxy_elliptical_table;
  
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5
  % GET LO-HI GRID in ANGSTROM
  
  variable b_lo=Double_Type[length(a.x)];
  variable b_hi=Double_Type[length(a.x)];
  
  variable x2=a.x*10^4;
  b_lo = x2;
  b_hi = make_hi_grid(b_lo);
  
  % Angstrom rest frame -> Angstrom observed wavelength
  %%%%%%%%%%%%%%%%%%%%%
  variable z0 = par[1];
  variable f0 = par[0];
  
  variable blo = b_lo*(z0+1);
  variable bhi = b_hi*(z0+1);
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % GET OPTICAL GRID
  
  %%%%%%%%%%%%%
  
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % REBIN DATA TO FIT DATA GRID
  
  variable new = rebin_mean (lo, hi, blo, bhi, a.E);
  variable f1 = new*f0;
  
  return f1;
}

add_slang_function("galaxy_elliptical", ["f [normalization factor]","z [ ]"],[0]);
% calculate and provide a proper normalization unit?

private define galaxy_elliptical_defaults(i)
{
  switch(i)
  {case 0 : return (1, 0, 1e-7, 3000); } % Factor
  {case 1 : return (0.01, 1, 0, 10);  } %REDSHIFT
}

set_param_default_hook("galaxy_elliptical", &galaxy_elliptical_defaults);

%
% Code from src/fitting/fit-functions/galridge.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
private define galridge_fit(lo, hi, par)
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{galridge (fit-function)}
%\synopsis{model for Galactic Ridge Emission}
%\description
%    Model for Galactic Ridge emission based on measurements
%    and analysis by:
%    Ebisawa, K., et al., 2008, PASJ, 60, S223-S230
%    Obst, M., 2011, Bsc-Thesis, Remeis observatory
%
%    The model consists of two bremsstrahlung continua and
%    three gaussian emission lines, which originate from
%    neutral, hydrogen- and helium-like Fe K_alpha lines
%    at 6.40, 6.67 and 7.00 keV:
%      2*bremss + 3*egauss
%    The equivlant widths (areas) of those lines are set to
%    free by default, but scaled initially by
%      85:458:129
%    as found by Ebisawa. To fix those ratios or different
%    ones, one might use 'set_par_fun', e.g.
%    
%    set_par_fun("galridge(1).Fe67","galridge(1).Fe64/85*458");
%    set_par_fun("galridge(1).Fe70","galridge(1).Fe64/85*129");
%
%    The line widths are all freezed to 0.05 keV.
%    The two continua are each described by a normalization
%    factor (norm1/norm2) and a plasma temperature (kT1/kT2).
%    By default, the second bremsstrahlung continuum is
%    disabled by setting its norm to zero.
%\seealso{bremss,egauss,set_par_fun}
%!%-
{
  % parameters of the Bremsstrahlung models [norm, kT]
  variable bremss1Par = par[[0,2]];
  variable bremss2Par = par[[1,3]];
  % parameters of the iron lines [area, center, sigma]
  variable egauss1Par = [par[4], 6.40, 0.05];
  variable egauss2Par = [par[5], 6.67, 0.05];
  variable egauss3Par = [par[6], 7.00, 0.05];
  % 2*bremss + 3*egauss
  return eval_fun2("bremss", lo, hi, bremss1Par)
       + eval_fun2("bremss", lo, hi, bremss2Par)
       + eval_fun2("egauss", lo, hi, egauss1Par)
       + eval_fun2("egauss", lo, hi, egauss2Par)
       + eval_fun2("egauss", lo, hi, egauss3Par);        
}

private define galridge_default(i)
{
 variable defs = Struct_Type[7], w = 1e-5, s1 = 458./85, s2 = 129./85;
 defs[0] = struct { value = 0.01, freeze = 0, min = 0, max = 0.1,  hard_min = 0, hard_max = 1  , step = 1e-4, relstep = 1e-5 };
 defs[1] = struct { value = 0,    freeze = 1, min = 0, max = 0.2,  hard_min = 0, hard_max = 1  , step = 1e-4, relstep = 1e-5 };
 defs[2] = struct { value = 4,    freeze = 0, min = 0, max = 30,   hard_min = 0, hard_max = 100, step = 1e-2, relstep = 1e-5 };
 defs[3] = struct { value = 2,    freeze = 1, min = 0, max = 15,   hard_min = 0, hard_max = 100, step = 1e-2, relstep = 1e-5 };
 defs[4] = struct { value = w,    freeze = 0, min = 0, max = 0.01, hard_min = 0, hard_max = 1  , step = 1e-7, relstep = 1e-5 };
 defs[5] = struct { value = w*s1, freeze = 0, min = 0, max = 0.01, hard_min = 0, hard_max = 1  , step = 1e-7, relstep = 1e-5 };
 defs[6] = struct { value = w*s2, freeze = 0, min = 0, max = 0.01, hard_min = 0, hard_max = 1  , step = 1e-7, relstep = 1e-5 };
 return defs[i];
}

add_slang_function("galridge", ["norm1","norm2","kT1 [keV]","kT2 [keV]","Fe64 [photons/s/cm^2]","Fe67 [photons/s/cm^2]","Fe70 [photons/s/cm^2]"]);
set_param_default_hook("galridge", "galridge_default");

%
% Code from src/fitting/fit_2d/gauss_2d.sl
%
define gauss_2d_fit(bin_lo, bin_hi, par)
{
  variable x0 = par[0];
  variable y0 = par[1];
  variable sx = par[2];
  variable sy = par[3];
  variable A  = par[4];

  variable XX, YY;
  (XX, YY) = get_2d_data_grid();

  return A * exp( -(XX-x0)^2/(2*sx^2) -(YY-y0)^2/(2*sy^2) ) / (2*PI*sx*sy);
}

define gauss_2d_defaults(i)
{
  switch(i)
  { case 0: return 0, 0, 0, 0; }
  { case 1: return 0, 0, 0, 0; }
  { case 2: return 1, 0, 1e-3, 1e3; }
  { case 3: return 1, 0, 1e-3, 1e3; }
  { case 4: return 1, 0, 0, 0; }
}

add_slang_function("gauss_2d", ["x0", "y0", "sx", "sy", "A"]);
set_param_default_hook("gauss_2d", "gauss_2d_defaults");

%
% Code from src/fitting/fit_2d/gauss_2d_integrated.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%!%+
%\function{gauss_2d_integrated}
%\synopsis{fit a two dimensional Gaussian profile to a bin integrated image}
%\description
%    For fitting 2d data, define_counts_2d can be used to define
%    a pseudo 1d spectrum by reshaping arrays.
%    Contrary to \code{gauss_2d} this function allows to fit bin
%    integrated values. Before using the fit function \code{gauss_2d_integrated}
%    the "counts" have to be defined with \code{define_counts_2d} and the
%    data grid has to be defined with \code{set_2d_data_grid}.
%    To fit integrated values (X_lo,X_hi, Y_lo, Y_hi) have to be specified
%    in \code{set_2d_data_grid}.
%\examples
%    variable delt = 0.1;
%    variable X = [-2:2:delt];
%    variable Y = @X;%
%    variable XX, YY;  (XX, YY) = get_grid(X, Y);
%    variable img = cos(XX)^2*cos(YY)^2*100;
%
%    variable id = define_counts_2d (img);
%    set_2d_data_grid (X-0.5*delt, X+0.5*delt,Y-0.5*delt, Y+0.5*delt); % valid value in pixel center
%    fit_fun("gauss_2d_integrated(1)");
%    set_fit_statistic ("chisqr;sigma=gehrels");
%    set_par("gauss_2d_integrated(1).x0",0);
%    set_par("gauss_2d_integrated(1).A", sum(img));
%    ()=fit_counts;
%    plot_image(get_2d_model(id));
%\seealso{gauss_2d, define_counts_2d, set_2d_data_grid}
%!%-

%%%%%%%%%%%%%%%%%%%
define gauss_2d_integrated_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%%%%
{
  variable x0 = par[0];
  variable y0 = par[1];
  variable sx = par[2];
  variable sy = par[3];
  variable A  = par[4];

  variable X,X_hi, Y,Y_hi;
  (X,X_hi, Y,Y_hi) = get_2d_data_grid(;bins);
  
  variable valx = 0.5*( gsl->erf((X_hi-x0)/(sqrt(2)*sx)) - gsl->erf((X-x0)/(sqrt(2)*sx)));
  variable valy = 0.5*( gsl->erf((Y_hi-y0)/(sqrt(2)*sy)) - gsl->erf((Y-y0)/(sqrt(2)*sy)));

  variable valXX, valYY;  (valXX, valYY) = get_grid(valx, valy);
  return A * valXX * valYY;
}

%%%%%%%%%%%%%%%%%%%%%%%%
define gauss_2d_integrated_defaults(i)
%%%%%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
  { case 0: return 0, 0, -1e10, 1e10; }
  { case 1: return 0, 0, -1e10, 1e10; }
  { case 2: return 1, 0, 1e-10, 1e10; }
  { case 3: return 1, 0, 1e-10, 1e10; }
  { case 4: return 1, 0, 0, 1e10; }
}

add_slang_function("gauss_2d_integrated", ["x0", "y0", "sx", "sy", "A"]);
set_param_default_hook("gauss_2d_integrated", "gauss_2d_integrated_defaults");

#endif

%
% Code from src/fitting/xyfit/fit-functions/gauss_xyfit.sl
%
%%%%%%%%%%%%%%%%%%
define gauss_xyfit()
%%%%%%%%%%%%%%%%%%
{
  variable xref, yref, par;
  switch(_NARGS)
  { case 0: return ["center [center]", "area [area]", "sigma [sigma]"];} % fit parameters
  { case 3: (xref, yref, par) = (); }
  { return help(_function_name); }

  @yref = par[1] / (2*PI*sqrt(par[2])) * exp ( -0.5* ( (@xref - par[0])/par[2] )^2 );
}

define gauss_xyfit_default(i)
{
  switch(i)
  { case 0: return (1   , 0, -1e5  , 1e5); }
  { case 1: return (1   , 0, -1e5  , 1e5); }
  { case 2: return (0.01, 0,  1e-10, 1e5); }
}

%
% Code from src/data/gehrels_confidence_limits.sl
%
% These are approximations for the lower and upper confidence bounds derived
% for a Poisson distribution with n counts and a confidence limit S as
% calculated by Gehrels 1986
private define gehrels_error_upper_7(n, S) {
  return n + S * sqrt(n + 3./4.) + (S^2 + 3.)/4.;
}

private define gehrels_error_upper_9(n, S) {
  return (n + 1.) * ( 1. - 1. / (9. * (n+1.)) + S / ( 3. * sqrt(n+1.)) ) ^3;
}

private define gehrels_error_upper_10(n, S) {
  return n + S * sqrt(n+1.) + (S^2 + 2.)/3.;
}

private define gehrels_error_lower_11(n, S) {
  return n - S * sqrt(n - 1./4.) + (S^2 - 1.)/4.;
}

private define gehrels_error_lower_12(n, S) {
  return n * ( 1. - 1./(9. * n) - S/(3 * sqrt(n)))^3;
}

private define gehrels_error_lower_13(n, S) {
  return n - S * sqrt(n) + (S^2 - 1.)/3.;
}

private define gehrels_error_lower_14(n, S, beta, gamma) {
  return n * ( 1 - 1./(9*n) - S/(3 * sqrt(n)) + beta * n^gamma)^3;
}

private define __calc_gehrels_beta(CL) {
  return -3.3002e-04 / (0.23377524 * (CL-1.00474780)) + 0.00198147;
}

private define __calc_gehrels_gamma(CL) {
  return -0.72681571 / (5.81655865*(CL-0.85566045)) -5.7704e-05 / (0.10719164*(CL-1.00160824)) -1.18709242;
}

private define __CL_to_S(CL) {
  return sqrt(2) * erfinv(2 * CL - 1);
}


%!%+
%\function{gehrels_upper_confidence}
%\synopsis{Calculate upper confidence bounds according to Gehrels, 1986}
%\usage{Double_Type gehrels_upper_confidence(Double_Type n, Double_Type CL);}
%\description
%    Estimate the upper bound of the confidence interval given by the level
%    CL and the number of counts n.
%    It uses equation 9 in Gehrels, 1986.
%    Please note that the confidence level must be in [0.8413,0.9995].
%\seealso{gehrels_lower_confidence, gehrels_confidence}
%!%-
define gehrels_upper_confidence(n, CL) {
  if ( CL < 0.8413 || CL > 0.9995 ) {
    throw UsageError, "CL must be in [0.8413,0.9995]";
  }
  variable S = __CL_to_S(CL);
  return gehrels_error_upper_9(n, S);
}


%!%+
%\function{gehrels_lower_confidence}
%\synopsis{Calculate lower confidence bounds according to Gehrels, 1986}
%\usage{Double_Type gehrels_lower_confidence(Double_Type n, Double_Type CL);}
%\description
%    Estimate the lower bound of the confidence interval given by the level
%    CL and the number of counts n.
%    It uses equation 14 in Gehrels, 1986.
%    Please note that the confidence level must be in [0.8413,0.9995].
%    The values for beta and gamma necessary for the calculation are
%    interpolated using a single parametric powerlaw for beta and a double
%    parametric powerlaw for gamma which was fitted to the data given in
%    table 3 of Gehrels, 1986.
%    Reading the source code is encounraged.
%\seealso{gehrels_lower_confidence, gehrels_confidence}
%!%-
define gehrels_lower_confidence(n, CL) {
  if ( CL < 0.8413 || CL > 0.9995 ) {
    throw UsageError, "CL must be in [0.8413,0.9995]";
  }
  variable S = __CL_to_S(CL);
  variable beta, gamma;
  if ( S == 1.0 ) {
    beta = 0.0;
    gamma = 0.0;
  } else {
    beta = __calc_gehrels_beta(CL);
    gamma = __calc_gehrels_gamma(CL);
  }
  if ( CL <= 0.85566045 ) {
    beta = 0.0;
  }
  return gehrels_error_lower_14(n, S, beta, gamma);
}

%!%+
%\function{gehrels_confidence}
%\synopsis{Calculate lower and upper confidence bounds according to Gehrels, 1986}
%\usage{Double_Type[2] conf = gehrels_confidence(Integer_Type n, Double_Type CL);}
%\description
%    Estimate the lower and upper bounds of the confidence interval given by
%    the level CL and the number of counts n using
%    \code{gehrels_lower_confidence} and \code{gehrles_upper_confidence}.
%    Please also read the documentation of these two functions.
%    The first entry in the retured array is the lower bound, the second one
%    the upper bound.
%\seealso{gehrels_lower_confidence, gehrles_upper_confidence}
%!%-
define gehrels_confidence(n, CL) {
  variable ret = Double_Type[2];
  ret[0] = gehrels_lower_confidence(n, CL);
  ret[1] = gehrels_upper_confidence(n, CL);
  return ret;
}

%!%+
%\function{gehrels_error}
%\synopsis{Calculate lower and upper error according to Gehrels, 1986}
%\usage{Double_Type[2] err = gehrels_error(Integer_Type n, Double_Type CL);}
%\description
%    Same as \code{gehrels_confidence} but returns the errors instead.
%\seealso{gehrels_lower_confidence, gehrles_upper_confidence}
%!%-
define gehrels_error(n, CL) {
  variable ret = gehrels_confidence(n, CL);
  ret[0] = n - ret[0];
  ret[1] = ret[1] - n;
  return ret;
}

%
% Code from src/slang/strings/generate_iauname.sl
%
define generate_iauname() {
%!%+
%\function{generate_iauname}
%\synopsis{for a given position, generate a coordinate string obeying the IAU convention}
%\usage{string=generate_iauname(ra,dec)}
%\qualifiers{
% \qualifier{prefix}{prefix for the string (e.g., "XMMU_")}
% \qualifier{radian}{if set, the angles are in radians, not degrees}
%}
%\description
% This is a convenience routine to produce strings of the type
% XMMU Jhhmmss.s+ddmmss
% from J2000.0 positions. The routine obeys the IAU convention of
% truncating (not rounding!) the coordinate to the digits shown.
%
% The default-string is "Jhhmmss.s+ddmmss", use the prefix qualifier
% to prepend the mission name.
%
% Use angle2string to format coordinates with appropriate rounding.
%
%\seealso{dms2deg, hms2deg, angle2string, deg2dms}
%!%-
    variable ra,dec;
    
    switch(_NARGS) 
	{case 2:  (ra,dec)=(); }
	{return help(_function_name());
    }

    variable hh,mm,ss;
    variable dd,dm,ds;


    if (qualifier_exists("radian")) {
	ra*=180./PI;
	dec*=180./PI;
    }

    variable prefix=qualifier("prefix","");
	
    variable sgn= (dec<0.) ? "-" : "+";
	
    (hh,mm,ss)=deg2dms(ra;hours);
    (dd,dm,ds)=deg2dms(abs(dec));

    % chop off digits per IAU convention
    ss=int(ss*10.)/10.;
	
    return sprintf("%sJ%02i%02i%04.1f%1s%02i%02i%02i",
          prefix,int(hh),int(mm),ss,sgn,int(dd),int(dm),int(ds));
}

%
% Code from src/slang/astronomy/geographic2vector.sl
%

% only activate if module vector could be loaded
#ifeval __get_reference("vector")!=NULL

define geographic2vector()
%!%+
%\function{geographic2vector}
%\synopsis{Calculate a geocentric vector from geographic or geodetic coordinates.}
%\usage{Vector_Type=geographic2vector(lambda,phi,hei;qualifiers)}
%\qualifiers{
%   \qualifier{deg}{if set, then lambda and phi are measured in degrees,
%                     (default: radian).}
%   \qualifier{datum}{a string defining the geodetic datum. The default
%                     is WGS 84, which corresponds to GPS coordinates. Other
%                     geodetic data are IERS 2010, GRS 80, IAU 1976, GRS 67,
%                     IAU 1964, Hayford (1924), Clarke (1866), and Airy (1830),
%                     per the parameters given in the Astronomical Almanac.}
%   \qualifier{geocentric}{The coordinates are not geodetic, but geocentric. 
%                     In this case hei is interpreted as the distance from the
%                     geocenter (otherwise it is the height with respect to
%                     the reference ellipsoid).
%                     Note: it is very rare for coordinates to be geocentric!
%                     GPS is geodetic.}
%}
%\description
%
% For a given set of geodetic coordinates, i.e., geographic longitude (positive
% towards the east!), latitude (positive on the Northern hemisphere), and a
% height  above the reference spheroid (close to height above mean sea level),
% calculate the x,y,z-position in a geocentric coordinate system. Here, the
% xy-plane is the Earth's equator, the xz-plane the ITRS meridian (close to
% Greenwich), and the z-axis points towards the Earth's north pole. The xyz-
% values are in meters.
%
% Note that the longitude and latitude are assumed to be in rad, unless the
% deg qualifier is given. The height is in meters.
%
% The routine returns a Vector_Type. If called with arrays, then an array
% of Vector_Type is returned.
%
%\seealso{dms2deg}
%!%-
{
    variable lam,phi,hei;

    switch(_NARGS)
    { case 3: (lam,phi,hei)=(); }
    { throw UsageError,sprintf("%s: Need longitude, latitude, and height!\n",_function_name());}

    variable la,ph,h;
    if (qualifier_exists("deg")) {
	la=lam*PI/180.;
	ph=phi*PI/180.;
	h=hei;
    } else {
	if (Vector_Type==typeof(lam)) {
	    la=lam[*];
	    ph=phi[*];
	    h=hei[*];
	} else {
	    la=lam;
	    ph=phi;
	    h=hei;
	}
    }

    variable vecarr,i;
    if (qualifier_exists("geocentric")) {
	if (Vector_Type==typeof(la)) {
	    vecarr=Vector_Type[length(la)];
	    _for i (0,length(la),1) {
		vecarr[i]=vector_astro(h[i],la[i],ph[i];astro);
	    }
	    return vecarr;
	} 
	return vector_astro(h,la,ph;astro);
    }

    %
    % geoid parameters for different datums
    % (see Astronomical Almanac, section K)
    %
    variable a,f;
    variable datum=qualifier("datum","WGS 84");
    switch(datum) 
    {case "WGS 84"          : a=6378137.0  ; f=1./298.257223563; }
    {case "IERS 2010"       : a=6378136.6  ; f=1./298.25642;     }
    {case "GRS 80"          : a=6378137.0  ; f=1./298.257222;    }
    {case "IAU 1976"        : a=6378140.0  ; f=1./298.257;       }
    {case "GRS 67"          : a=6378160.0  ; f=1./298.247167;    }
    {case "IAU 1964"        : a=6378160.0  ; f=1./298.25;        }
    {case "Hayford (1924)"  : a=6378388.0  ; f=1./297.;          }
    {case "Clarke (1866)"   : a=6378206.4  ; f=1./294.978698;    }
    {case "Airy (1830)"     : a=6377563.396; f=1./299.324964;    }
    {throw UsageError,sprintf("%s: Geographic datum %s not known\n",_function_name(), datum);}

    %
    % convert to x,y,z
    %
    variable sp,cp,sl,cl;
    (sp,cp)=sincos(ph);
    (sl,cl)=sincos(la);

    variable C=1./sqrt(cp*cp+(1.-f)*(1.-f)*sp*sp);
    variable S=C*(1.-f)*(1.-f);

    if (Vector_Type==typeof(la)) {
	vecarr=Vector_Type[length(la)];
	_for i(0,length(la)-1,1) {
	    vecarr[i]= vector((a*C[i]+h[i])*cp[i]*cl[i], (a*C[i]+h[i])*cp[i]*sl[i], (a*S[i]+h[i])*sp[i]);
	}
	return vecarr;
    }
    return vector((a*C+h)*cp*cl, (a*C+h)*cp*sl, (a*S+h)*sp);
}

#endif

%
% Code from src/slang/strings/get1match.sl
%
define get1match()
%!%+
%\function{get1match}
%\synopsis{returns one (the first) matching substring of a regular expression matching}
%\usage{String_Type match_str = get1match(String_Type str, String_Type regexp);}
%\description
%   get1match is deprecated. Use S-Lang's string_matches instead.
%\seealso{string_matches}
%!%-
{
  message("get1match is deprecated. Use S-Lang's string_matches instead.");

  variable str, regexp;
  switch(_NARGS)
  { case 2: (str, regexp) = (); }
  { help(_function_name()); return; }

  if(string_match(str, regexp, 1)) {
      variable pos, len;
      (pos, len) = string_match_nth(1);
      return substr(str, pos+1, len);
  }

  return NULL;
}

%
% Code from src/slang/strings/get2matches.sl
%
define get2matches(str, regexp)
{
  if(string_match(str, regexp, 1))
  {
    variable pos1, len1, pos2, len2;
    (pos1, len1) = string_match_nth(1);
    (pos2, len2) = string_match_nth(2);
    return (substr(str, pos1+1, len1), substr(str, pos2+1, len2));
  }

  return NULL;
}

%
% Code from src/data/GTI/getCommonGTIs.sl
%
%%%%%%%%%%%%%%%%%%%%
define getCommonGTIs()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{getCommonGTIs}
%\synopsis{returns a new set of GTIs which are both contained in gti1 and gti2}
%\usage{Struct_Type gti = getCommonGTIs(Struct_Type gti1, Struct_Type gti2);}
%\description
%    Good time intervals (GTIs) are stored in \code{{ start, stop }}-structs.
%!%-
{
  variable gti1, gti2;
  switch(_NARGS)
  { case 2: (gti1, gti2) = (); }
  { help(_function_name()); return; }

  % patch to make sure gti1 has the lower starting time.
  if(gti2.start[0] < gti1.start[0]){
    variable temp = gti2;
    gti2 = gti1;
    gti1 = temp;
  }

  variable gti = struct { start=Double_Type[0], stop=Double_Type[0] };
  variable len1 = length(gti1.start);
  variable len2 = length(gti2.start);
  variable i1=0, i2=0;
  while(i1<len1 && i2<len2)
  {
    variable t = _max(gti1.start[i1], gti2.start[i2]);
    while(i1<len1 && gti1.stop[i1]<t)  i1++;
    while(i2<len2 && gti2.stop[i2]<t)  i2++;

    if(i1<len1 && i2<len2 && gti1.start[i1]<= t && gti2.start[i2]<=t)
    { gti.start = [gti.start, t];
      gti.stop  = [gti.stop,  _min(gti1.stop[i1], gti2.stop[i2])];
      if(gti1.stop[i1] <= gti2.stop[i2])  i1++;
      if(gti1.stop[i1] >= gti2.stop[i2])  i2++;
    }
  }
  return gti;
}

%
% Code from src/data/timing/getVarPeriod.sl
%
%%%%%%%%%%%%%%%%%%%%%
define getVarPeriod(lc, minp, maxp)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{getVarPeriod}
%\synopsis{Determines the period of a variation in a lightcurve}
%\usage{Double_Type getVarPeriod(Struct_Type lc, Double_Type minperiod, Double_Type maxperiod)}
%\qualifiers{
%    \qualifier{plot}{plot chi-square distribution and modelled
%              gaussians. Pause the assigned number in
%              seconds between each step (default 0.3)}
%    \qualifier{nogauss}{skip the gaussian fit}
%
% \qualifier{PIPED TO EPFOLD}{:}
%    \qualifier{dt}{exposure of every lightcurve time bin, should
%               be given to ensure correct results.}
%    \qualifier{sampling}{how many periods per peak to use (default=10)}
%    \qualifier{nsrch}{how many periods to search in a linear grid (default not set)}
%    \qualifier{dp}{delta period of linear period grid (default not set)}
%    \qualifier{lstat}{use L-statistics instead of chi^2 statistics}
%    \qualifier{nbins}{number of bins for the pulse profile}
%    \qualifier{exact}{calculate the pulse profile in a more exact
%               way, see description of pfold (not recommed
%               as it takes a very long time!).}}
%\description
%    ****
%    This function is deprecated!! It will be removed in
%    2017 Mai. If you heavily depend on this function,
%    please write an email to
%    matthias.kuehnel@sternwarte.uni-erlangen.de
%    ****
%    Performs an Epoch Folding of the lightcurve in the
%    given period range (same time unit as used in the
%    lightcurve) and returns the period corresponding
%    to the maximum.
%    If not skipped by the 'nogauss' qualifier, a gaussian
%    it fitted to the maximum to get its position more
%    accurate. The gaussian parameters and used data-
%    points are applied automatically and improved step
%    by step. The steps can be monitored using the 'plot'
%    qualifier.
%\seealso{epfold}
%!%-
{
 message("****");
 message("The function getVarPeriod is deprecated! It will be removed in");
 message("2017 Mai. If you heavily depend on this function, please write");
 message("an to matthias.kuehnel@sternwarte.uni-erlangen.de");
 message("****");

 variable pldt = qualifier("plot",0.3); if (pldt == NULL) pldt = 0.3;
 variable epf = epfold(lc.time, lc.rate, minp, maxp;; __qualifiers);
 variable period = epf.p[where_max(epf.stat)];
 if (length(period)>1) period = period[0];

 % eventually do a gaussian fit
 ifnot (qualifier_exists("nogauss"))
 {
  delete_data(all_data);
  % select only high "chisquared" data
  variable ndx = moment(epf.stat);
  ndx = where(epf.stat > ndx.ave + ndx.sdev);
  if (length(ndx)<5) { vmessage("warning (%s): no significant gaussian region in chi^2 distribution found ndx", _function_name); ndx = [0:length(epf.stat)-1]; }
  % define data
  variable dp0 = max(epf.p[ndx]) - min(epf.p[ndx]);
  period = min(epf.p[ndx]) + dp0/2.;
  variable grid = make_hi_grid(epf.p[ndx]);
  ()=define_counts(epf.p[ndx], grid, epf.stat[ndx], epf.stat[ndx]*0.05); % error is not correct!
  set_data_exposure(1,1);
  fit_fun("constant(1)+gauss(1)");
  set_par("constant(1).factor",min(epf.stat[ndx]),0,0.0,max(epf.stat[ndx]));
  set_par("gauss(1).sigma",dp0/2,0,0.0,1.5*dp0);
  set_par("gauss(1).center",period,0,min(epf.p[ndx]),max(epf.p[ndx]));
  set_par("gauss(1).area",length(ndx)*max(epf.stat[ndx]),0,0.0,3.0*length(ndx)*max(epf.stat[ndx]));

  variable model = struct { value = 0 };
  variable fm = get_fit_method;

  xrange; yrange;
  set_fit_method("subplex");
  if (qualifier_exists("plot"))
  {
    ()=eval_counts(;fit_verbose=-1);
    xlabel("Period (s)"); ylabel("\\gx\\u2\\d-value");
    model=get_model_counts(1);
    color(15); plot(epf.p*86400,epf.stat);
    color(1); oplot(epf.p[ndx]*86400,epf.stat[ndx]);
    color(2); oplot(epf.p[ndx]*86400,model.value);
    sleep(pldt);
  }
  % fit until maximum position is near the global one and the
  % shrinking of the used range has stopped
  ()=fit_counts(;fit_verbose=-1); % fit
  model = get_model_counts(1);
  if (qualifier_exists("plot"))
  {
    color(15); plot(epf.p*86400,epf.stat);
    color(1); oplot(epf.p[ndx]*86400,epf.stat[ndx]);
    color(2); oplot(epf.p[ndx]*86400,model.value);
  }
  if (qualifier_exists("plot")) sleep(pldt*5);
 }
 delete_data(all_data);
 set_fit_method(fm);
 return period;
}

%
% Code from src/fitting/fit_2d/get_2d_data_grid.sl
%
private variable fit_2d_data_grid = struct { X, Y, X_hi, Y_hi, nx, ny, XX, YY, XX_1d, YY_1d };

%%%%%%%%%%%%%%%%%%%%%%%
define set_2d_data_grid()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{set_2d_data_grid}
%\usage{set_2d_data_grid(Double_Type X[], Double_Type Y[]);}
%\altusage{set_2d_data_grid(Double_Type X_lo[], Double_Type X_hi[], Double_Type Y_lo[], Double_Type Y_hi[]);}
%\synopsis{define a two dimensional data grid required for 2D fits}
%\description
%    For fitting 2d data the corresponding grid is set with this function.
%    If \code{set_2d_data_grid} is not called, the function \code{define_counts_2d}
%    uses the indices of the image as grid.
%    For binned fit functions bin_lo and bin_hi have to be provided.
%    If only single X and Y arrays are provided, only fit functions
%    which are evaluated on these grid points can be used.
%\seealso{define_counts_2d, gauss_2d_integrated, gauss_2d}
%!%-
{
  variable X, Y, X_hi, Y_hi;
  switch(_NARGS)
  { case 2: (X, Y) = (); }
  { case 4: (X, X_hi, Y, Y_hi) = ();
            fit_2d_data_grid.X_hi  = X_hi; fit_2d_data_grid.Y_hi  = Y_hi; }

  variable nx = length(X);
  variable ny = length(Y);
  variable XX, YY;  (XX, YY) = get_grid(X, Y);
  fit_2d_data_grid.X  = X;      fit_2d_data_grid.Y  = Y;
  fit_2d_data_grid.nx = nx;     fit_2d_data_grid.ny = ny;
  fit_2d_data_grid.XX = @XX;    fit_2d_data_grid.YY = @YY;
  reshape(XX, length(XX));      reshape(YY, length(YY));
  fit_2d_data_grid.XX_1d = XX;  fit_2d_data_grid.YY_1d = YY;
}

%%%%%%%%%%%%%%%%%%%%%%%
define get_2d_data_grid()
%%%%%%%%%%%%%%%%%%%%%%%
{
  if(qualifier_exists("bins"))
    return (fit_2d_data_grid.X, fit_2d_data_grid.X_hi, fit_2d_data_grid.Y, fit_2d_data_grid.Y_hi);

  if(qualifier_exists("oned"))
    return (fit_2d_data_grid.XX_1d, fit_2d_data_grid.YY_1d);

  if(qualifier_exists("dim"))
    return (fit_2d_data_grid.nx, fit_2d_data_grid.ny);

  return (fit_2d_data_grid.XX, fit_2d_data_grid.YY);
}

%
% Code from src/fitting/fit_2d/get_2d_model.sl
%
%%%%%%%%%%%%%%%%%%%
define get_2d_model()
%%%%%%%%%%%%%%%%%%%
{
  variable id;
  switch(_NARGS)
  { case 0: id = 1; }
  { case 1: id = (); }
  { help(_function_name()); return; }

  variable model = get_model_counts(id);
  variable nx, ny; (nx, ny) = get_2d_data_grid(; dim);
  reshape(model.value, [ny, nx]);
  return model.value;
}

%
% Code from src/data/get_all_data.sl
%
%%%%%%%%%%%%%%%%
define get_all_data()
%%%%%%%%%%%%%%%%
%!%+
%\function{get_all_data}
%
%\synopsis{Get a list of all data-set indices}
%\usage{List_Type = get_all_data;}
%\description
%
%       This function returns a list of data indices, which have been obtained via all_data. 
%       The information is saved in a List, not an Array, which allows plotting of all datasets. 
%
%\example
%	isis>variable d = get_all_data;
%	isis>plot_data(d);
%
%\seealso{all_data}
%!%-

 { 
variable ad = all_data;
variable nl = {};

variable k;

_for k (0, length(ad)-1,1) {list_append(nl, ad[k]);}

if (nl[0]==NULL) {return NULL;}
else{

 return nl; }
     
}



%
% Code from src/slang/get_arg_struct.sl
%
define get_arg_struct() %{{{
%!%+
%\function{get_arg_struct}
%\synopsis{obtains the arguments/options the current isis instances was called with.}
%\usage{Struct_Type = get_arg_struct( );}
%\altusage{Struct_Type = get_arg_struct( String_Type[] name, type);}
%\qualifiers{
%\qualifier{delim}{[="="] Delimeter between argument name and value (e.g., also, "=.,").}
%\qualifier{prefix}{[="--"] Argument prefix. Arguments missing this prefix are ignored.}
%}
%\description
%   This function returns a Struct_Type with fields corresponding to
%   the names of those arguments/options the current isis instance
%   was called with, which have the given 'prefix'. Thereby the 'name'
%   is defined as the string enclosed by the 'prefix' and the 'delimeter'.
%
%   Expected SYNTAX: <prefix>"argname"<delimeter>"value"
%
%   It is possible to give several 'delimeter' as a string-chain (e.g.,
%   "=,.-"). If an argument/option contains several delimeters, only the
%   string between the 1st and 2nd delimeter is taken as value for the
%   corresponding name.
%
%   This function can be given an string array 'name' and 'type', assigning
%   a Data_Type to the argument 'name'. Thereby 'type' can be:
%
%         "d" : Integer_Type
%         "f" : Double_Type
%
%   Otherwise, if used without arguments ('name' & 'type'), the
%   values are returned as String_Type.
%
%   If an argument has no value given, i.e. does not contain a delimeter
%   the corresponding field of the returned structure is equal NULL.
%\example
%   /> isis -g --arg --arg1=1e-3 --arg2=1.01 --arg3=abcd --arg4=1=bla
%
%   isis> print( get_arg_struct );
%   {arg=NULL,
%    arg1="1e-3",
%    arg2="1.01",
%    arg3="abcd",
%    arg4="1"}
%
%   isis> print( get_arg_struct("arg1","f") );
%   {arg=NULL,
%    arg1=0.001,
%    arg2="1.01",
%    arg3="abcd",
%    arg4="1"}
%
%   isis> print( get_arg_struct(["arg1","arg2","arg4"],["f","f","d"]) );
%   {arg=NULL,
%    arg1=0.001,
%    arg2=1.01,
%    arg3="abcd",
%    arg4=1}
%
%\seealso{atof, atoi, strreplace, strtok, set_struct_field, array_map, is_substr}
%!%-
{
  variable delim  = qualifier("delim","=");
  variable prefix = qualifier("prefix","--");

  variable name = NULL, type;
  switch(_NARGS)
  { case 2 : (name,type)=(); name = [name]; type=[type]; }


  % Get arguments with prefix 'prefix'
  variable opt = __argv[ where( is_substr( __argv, prefix ) == 1) ];
  % remove prefix
  opt = array_map( String_Type, &strreplace, opt, prefix, "" );
  % split these arguments at 'delimeter'
  opt = array_map( Array_Type, &strtok, opt, delim );

  variable nopt = length(opt);
  if( nopt == 0 )
  { vmessage("WARNING: <%s>: No argument(s) with prefix=%s found!",  _function_name, prefix ); }

  variable args = @Struct_Type(String_Type[0]);  % Ideally, one would first collect all fields and then construct the struct...
  variable i;
  _for i (0, nopt-1, 1 ){
    variable field = opt[i][0];
    args = struct_combine(args, field);

    variable val = NULL;
    if( length(opt[i]) > 1 ) {
      val = opt[i][1];

      variable j = wherefirst(name == field);
      if( j != NULL ){
	if( type[j]=="d" ){
	  val = atoi(val);
	}
	else if( type[j]=="f" ){
	  val = atof(val);
	}
      }
    }
    set_struct_field(args, field, val);
  }
  return args;
}
%}}}

%
% Code from src/data/spectrum.datastructures/get_combined_data_model_residuals.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_combined_data_model_residuals()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_combined_data_model_residuals}
%\synopsis{returns these 3 structures for a combination of data sets}
%\usage{(d, m, r) = get_combined_data_model_residuals(Integer_Type id[]);}
%\description
%   All \code{d}, \code{m} and \code{r} are \code{bin_lo, bin_hi, value, err} structures.
%   \code{d} and \code{m} are the sum of data and model counts of all data sets,
%   rebinned to the grid of the first data set id[0].
%   \code{d.err} is calculated from quadratic error propagation.
%   \code{r.value = (d.value-m.value)/d.err} contains the residuals
%   unless the \code{ratio} qualifier is set (see below).
%\qualifiers{
%\qualifier{ratio}{use ratio-residuals \code{r.value = d.value/m.value}}
%}
%\seealso{get_data_counts, get_model_counts, rebin}
%!%-
{
  if(_NARGS!=1)
  { help(_function_name()); return;
  }
  variable ids = (); ids = [ids];
  variable data = get_data_counts(ids[0]);
  variable model = get_model_counts(ids[0]);
  variable i;
  _for i (1, length(ids)-1, 1)
  { variable d = get_data_counts(ids[i]);
    variable m = get_model_counts(ids[i]);
    data.value += rebin(data.bin_lo, data.bin_hi,  d.bin_lo, d.bin_hi, d.value);
    data.err = sqrt( data.err^2 + rebin(data.bin_lo, data.bin_hi,  d.bin_lo, d.bin_hi, d.err^2) );
    model.value += rebin(model.bin_lo, model.bin_hi,  m.bin_lo, m.bin_hi, m.value);
  }
  variable res = @data;
  if(qualifier_exists("ratio"))
    res.value = data.value/model.value,
    res.err = data.err/model.value;
  else
    res.value = (data.value-model.value)/data.err,
    res.err = res.value*0+1;
  return (data, model, res);
}

%
% Code from src/misc/radio/get_component.sl
%

define get_component()
%!%+
%\function{get_component}
%\synopsis{to extract the component from the epoch files}
%\usage{get_component([array of position number in the epoch file], [array of epoch fits files]);}
%\qualifiers{
%\qualifier{quadrant}{[=1] Quadrant which is defined as a positive distance ( RA, DEC > 0 == 1 || RA < 0, DEC > 0 == 2 ||
%  RA < 0, DEC < 0 == 3 || RA > 0, DEC < 0 == 4)}
%\qualifier{zero}{shift all distances, so that the last component is at 0,0 ("core" component) }
%}
%\description
%   This function returns a structure of a component.
%   The required input is the position number starting by 1 in each epoch, and a corresponding list of epoch fits files.
%   If a component is not existent in an epoch, the position number of the component must be 0.
%!%-
{
  variable components, ep;
  switch(_NARGS)
  { case 2: (components, ep) = (); }
  { return help(_function_name()); }

  variable q = qualifier("quadrant", 1);
  variable Epoch = qualifier_exists ("zero")
                 ? read_ep(ep; quadrant=q, zero)
                 : read_ep(ep; quadrant=q);

  variable true_comp = where(components);
  variable ntrue_comp = length(true_comp);
  variable comp = struct {
    distance = Double_Type [ntrue_comp],
    flux     = Double_Type [ntrue_comp],
    mjd      = Double_Type [ntrue_comp],
    date     = Double_Type [ntrue_comp],
    tb       = Double_Type [ntrue_comp],
    comp_org = Double_Type [ntrue_comp],
    posangle = Double_Type [ntrue_comp],
    size     = Double_Type [ntrue_comp],
    epochnr  = Int_Type    [ntrue_comp],
    deltax   = Double_Type [ntrue_comp],
    deltay   = Double_Type [ntrue_comp],
    major_ax = Double_Type [ntrue_comp],
    minor_ax = Double_Type [ntrue_comp]
  };

  % to account for the 0 not in the list;
  variable cp = components[true_comp] - 1;
  variable ii;
  _for ii (0, ntrue_comp-1, 1)
  {
    variable jj = true_comp[ii];
    variable kk = cp[ii];
    comp.distance[ii] = Epoch[jj].distance[kk];
    comp.deltax  [ii] = Epoch[jj].deltax[kk];
    comp.deltay  [ii] = Epoch[jj].deltay[kk];
    comp.flux    [ii] = Epoch[jj].flux[kk];
    comp.mjd     [ii] = Epoch[jj].mjd;
    comp.date    [ii] = yearOfMJD(Epoch[jj].mjd);
    comp.tb      [ii] = Epoch[jj].tb[kk];
    comp.comp_org[ii] = components[jj];
    comp.posangle[ii] = Epoch[jj].posangle[kk];
    comp.major_ax[ii] = Epoch[jj].major_ax[kk];
    comp.minor_ax[ii] = Epoch[jj].minor_ax[kk];
    comp.size    [ii] = PI * Epoch[jj].major_ax[kk] * Epoch[jj].minor_ax[kk];
    comp.epochnr [ii] = jj;
  }

  return comp;
}

%
% Code from src/fitting/get_confmap.sl
%

private variable confmap_pars, confmap_save_basefilename;

%%%%%%%%%%%%%%%%%%%%%%%%
private define confmap_save_hook(p)
%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable fp = fopen(sprintf("%s_%d.dat", confmap_save_basefilename, getpid()), "a");
  variable par;
  foreach par (get_params(confmap_pars))
    ()=fprintf(fp, "%S\t", par.value);
  ()=fprintf(fp, "%S\n", p.statistic);
  ()=fclose(fp);
}

%%%%%%%%%%%%%%%%%%
define get_confmap()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_confmap}
%\synopsis{computes a 2d confidence map and possibly stores the fit-parameters in a file}
%\usage{conf_map = get_confmap(par1, min1, max1[, n1], par2, min2, max2[, n2]);}
%\qualifiers{
%\qualifier{save}{[=\code{"confmap"}]: save all parameters to \code{save}+".fits"}
%\qualifier{fail}{[=\code{NULL}]: failure recovery hook, see \code{conf_map_counts}}
%\qualifier{mask}{[=\code{NULL}]: region mask out hook, see \code{conf_map_counts}}
%}
%\description
%    (All qualifiers are also passed to the \code{conf_map_counts} function.)
%
%    \code{par1} is stepped from \code{min1} to \code{max1} in \code{n1} (default=8) steps;
%    \code{par2} is stepped from \code{min2} to \code{max2} in \code{n2} (default=8) steps.
%    A save hook is used to write each step's parameter values and chi^2 to files
%    named \code{sav+"*.dat}, which are finally collected by \code{get_confmap_collect_results}
%    and converted to a table.
%    Parameters from this table can be set with \code{set_par_from_confmap_table}.
%\seealso{conf_grid, conf_map_counts, load_conf, plot_conf, get_confmap_collect_results, set_par_from_confmap_table}
%!%-
{
  variable par1, min1, max1, n1=8,
           par2, min2, max2, n2=8;
  switch(_NARGS)
  { case  6: (par1, min1, max1,     par2, min2, max2) = (); }
  { case  8: (par1, min1, max1, n1, par2, min2, max2, n2) = (); }
  { help(_function_name()); return; }

   variable grid1, grid2;

#ifeval _isis_version_string  < "1.6.1-37"
   grid1 = conf_grid(par1, min1, (n1*max1-min1)/(n1-1.), n1);
   grid2 = conf_grid(par2, min2, (n2*max2-min2)/(n2-1.), n2);
#else
   grid1 = conf_grid(par1, min1, max1, n1);
   grid2 = conf_grid(par2, min2, max2, n2);
#endif   
   
  variable info = struct { fail=qualifier("fail"), save, mask=qualifier("mask") };
  thaw(par1, par2);
  confmap_pars = freeParameters();
  confmap_save_basefilename = qualifier("save", "confmap");
  variable fit_info, FV=Fit_Verbose;
  Fit_Verbose = -1;
  ()=eval_counts(&fit_info);
  Fit_Verbose = FV;
  if(length( glob(confmap_save_basefilename+"_[0-9]*.dat") )>0)
    vmessage("warning (%s): There are already %s_*.dat files.\n        confmap data will not be saved.", _function_name(), confmap_save_basefilename);
  else
  { info.save = &confmap_save_hook;
    variable fp = fopen(confmap_save_basefilename+".info", "w");
    ()=fprintf(fp, "%S\t%S\t%S\t%S\t%d\t%S\n", get_par_info(par1).name, get_par(par1), min1, max1, n1, grid1.max);
    ()=fprintf(fp, "%S\t%S\t%S\t%S\t%d\t%S\n", get_par_info(par2).name, get_par(par2), min2, max2, n2, grid2.max);
    ()=fprintf(fp, "%s = %S\n", Fit_Statistic, fit_info.statistic);
    ()=fprintf(fp, "%S\n", get_fit_fun());
    variable par;
    foreach par (get_params(confmap_pars))
      ()=fprintf(fp, "%s\t", escapedParameterName(par.name));
    ()=fprintf(fp, "%s\n", Fit_Statistic);
    ()=fclose(fp);
  }

  variable confmap = conf_map_counts(grid1, grid2, info;; __qualifiers);
  confmap;  % return value, left on stack

  try { ()=save_conf(confmap, confmap_save_basefilename+".fits"); }
  catch AnyError: { vmessage("warning (%s) could not save confmap to %s", _function_name(), confmap_save_basefilename);
                    return;
                  }
  if(info.save != NULL)
    get_confmap_collect_results(confmap_save_basefilename
				%; remove_files, use_file_from_save_conf, @__qualifiers  % requires S-Lang >= pre-2.2.3-39
				;; struct_combine(struct { remove_files, use_file_from_save_conf },
						  __qualifiers)
			       );
}

%
% Code from src/fitting/get_confmap_collect_results.sl
%

private variable confmap_data_x, confmap_data_y, confmap_data_absdiff_x, confmap_data_absdiff_y;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define compare_confmap_data_indices(i, j)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  if(confmap_data_y[i] < confmap_data_y[j] - confmap_data_absdiff_y)
    return -1;
  if(confmap_data_y[i] > confmap_data_y[j] + confmap_data_absdiff_y)
    return +1;

  if(confmap_data_x[i] < confmap_data_x[j] - confmap_data_absdiff_x)
    return -1;
  if(confmap_data_x[i] > confmap_data_x[j] + confmap_data_absdiff_x)
    return +1;
  return 0;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_confmap_collect_results()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_confmap_collect_results}
%\synopsis{collect results produced by get_confmap}
%\usage{get_confmap_collect_results(String_Type save_basefilename);}
%\description
%    \code{get_confmap_collect_results} is used internally by \code{get_confmap},
%    but it can also collect the results of an unfinsihed calculation.
%\qualifiers{
%\qualifier{remove_files}{: if set, \code{save_basefilename+"_*.dat"} files will be deleted when read}
%\qualifier{use_file_from_save_conf}{: if set, results are appended to \code{save_basefilename+".fits"}}
%}
%\seealso{get_confmap}
%!%-
{
  if(_NARGS!=1) return help(_function_name());
  variable confmap_save_basefilename = ();
  variable remove_files = qualifier("remove_files");
  if(remove_files==NULL)  remove_files = qualifier_exists("remove_files");
  variable use_file_from_save_conf = qualifier_exists("use_file_from_save_conf");

  variable infofile = confmap_save_basefilename+".info";
  variable fp = fopen(infofile, "r");

  % tokens of line #1: 0=parameter name, 1=best fit value, 2=min[grid], 3=max[grid], 4=n[grid], 5=max[grid,ISIS]
  variable tok = strtok(fgetslines(fp,1)[0], "\t");
  variable origpar1 = tok[0],  best_par1 = atof(tok[1]), n1 = atoi(tok[4]), max1=atof(tok[5]);
  variable par1 = escapedParameterName(origpar1);
  variable Xvalues = [atof(tok[2]) : atof(tok[3]) : #n1];
  confmap_data_absdiff_x = (Xvalues[1]-Xvalues[0])/3.;

  % tokens of line #2: 0=parameter name, 1=best fit value, 2=min[grid], 3=max[grid], 4=n[grid], 5=max[grid,ISIS]
           tok = strtok(fgetslines(fp,1)[0], "\t");
  variable origpar2 = tok[0],  best_par2 = atof(tok[1]), n2 = atoi(tok[4]), max2=atof(tok[5]);
  variable par2 = escapedParameterName(origpar2);
  variable Yvalues = [atof(tok[2]) : atof(tok[3]) : #n2];
  confmap_data_absdiff_y = (Yvalues[1]-Yvalues[0])/3.;

  % tokens of line #3: 0=fit_statistic, 1=best_statistic
           tok = strtok(fgetslines(fp,1)[0], "= ");
  variable fit_statistic = tok[0], best_statistic = atof(tok[1]);

  % line #4: the fit-function
  variable fitfun = fgetslines(fp,1)[0];

  % tokens of line #5: parameter names and chi2
  variable fieldnames = strtok(fgetslines(fp,1)[0], "\t\n");
  variable i, n = length(fieldnames);
  variable confmap_data = @Struct_Type(fieldnames);

  ()=fclose(fp);

  variable file, datafiles = glob(confmap_save_basefilename+"_[0-9]*.dat");
  foreach file (datafiles)
    _for i (0, n-1, 1)
    {
      variable col = readcol(file, 1+i);
      variable COL = get_struct_field(confmap_data, fieldnames[i]);
      set_struct_field(confmap_data, fieldnames[i], COL==NULL ? col : [COL, col]);
    }
  col = NULL; COL = NULL;

  confmap_data_x = get_struct_field(confmap_data, par1);
  confmap_data_y = get_struct_field(confmap_data, par2);

  variable n_data = length(get_struct_field(confmap_data, fit_statistic));
  if(n_data < n1*n2)
  {
    vmessage("warning (%s): expecting %d parameter combinations, found %d",
	     _function_name(), n1*n2, n_data);
    variable x, y, Xmiss={}, Ymiss={};
    foreach x (Xvalues)
      foreach y (Yvalues)
	ifnot(any(    feqs(confmap_data_x, x, 0, confmap_data_absdiff_x)
		  and feqs(confmap_data_y, y, 0, confmap_data_absdiff_y)))
	  list_append(Xmiss, x),
	  list_append(Ymiss, y);
    confmap_data_x = [confmap_data_x, list_to_array(Xmiss)];
    confmap_data_y = [confmap_data_y, list_to_array(Ymiss)];
    set_struct_field(confmap_data, par1, confmap_data_x);
    set_struct_field(confmap_data, par2, confmap_data_y);
    variable n_miss = length(Xmiss);
    vmessage("adding %d NaNs => %d %s= %dx%d parameter combinations in total", n_miss, n_data+n_miss, (n_data+n_miss-n1*n2 ? "!" : ""), n1, n2);
    variable NaNs = Double_Type[n_miss] + _NaN;
    variable field;
    foreach field (fieldnames)
      if(field!=par1 && field!=par2)
	set_struct_field(confmap_data, field,
			 [get_struct_field(confmap_data, field), NaNs]);
  }
  n_data = length(get_struct_field(confmap_data, fit_statistic));
  if(n_data != n1*n2)
    vmessage("warning (%s): something is wrong!", _function_name());

  % sort parameter table
  struct_filter(confmap_data, array_sort([0:n_data-1], &compare_confmap_data_indices));

  _for i (0, n-1, 1)
    set_struct_field(confmap_data, fieldnames[i],
		     typecast(get_struct_field(confmap_data, fieldnames[i]), Float_Type)
		    );
  fp = fits_open_file(confmap_save_basefilename+".fits", use_file_from_save_conf ? "w" : "c");
  ifnot(use_file_from_save_conf)
  {
    % fits_write_image(fp, "chi2", _reshape(confmap_data.chi2, [n2, n1]),
    %                  Xvalues, Yvalues, par1, par2; WCS="P");

    % after <ISIS-source-dir>/share/fits_module_dep.sl:
    ()=_fits_create_img(fp, -32, [n2, n1]);
    ()=_fits_write_img(fp, _reshape(get_struct_field(confmap_data, fit_statistic)-best_statistic, [n2, n1]));
    ()=_fits_update_key(fp, "BESTSTAT", best_statistic, "Best-fit $fit_statistic value"$);
    ()=_fits_update_key(fp, "BEST_X",  best_par1, "X-coordinate, best-fit $fit_statistic value"$);
    ()=_fits_update_key(fp, "BEST_Y",  best_par2, "Y-coordinate, best-fit $fit_statistic value"$);
    ()=_fits_update_key(fp, "PXNAME", origpar1, "");
    ()=_fits_update_key(fp, "PXMIN", Xvalues[0], "");
    ()=_fits_update_key(fp, "PXMAX", max1, "");
    ()=_fits_update_key(fp, "PXNUM", n1, "");
    ()=_fits_update_key(fp, "PYNAME", origpar2, "");
    ()=_fits_update_key(fp, "PYMIN", Yvalues[0], "");
    ()=_fits_update_key(fp, "PYMAX", max2, "");
    ()=_fits_update_key(fp, "PYNUM", n2, "");
    ()=_fits_write_comment (fp, `fit_fun("$fitfun")`$);
    variable dx = (max1-Xvalues[0])/double(n1);
    variable dy = (max2-Yvalues[0])/double(n2);
    ()=_fits_update_key(fp, "CTYPE1P", par1, "");
    ()=_fits_update_key(fp, "CRVAL1P", Xvalues[0], "");
    ()=_fits_update_key(fp, "CRPIX1P", 1.0, "");
    ()=_fits_update_key(fp, "CDELT1P", dx, "");
    ()=_fits_update_key(fp, "WCSTY1P", "PHYSICAL", "");
    ()=_fits_update_key(fp, "CUNIT1P", "", "");
    ()=_fits_update_key(fp, "CTYPE2P", par2, "");
    ()=_fits_update_key(fp, "CRVAL2P", Yvalues[0], "");
    ()=_fits_update_key(fp, "CRPIX2P", 1.0, "");
    ()=_fits_update_key(fp, "CDELT2P", dy, "");
    ()=_fits_update_key(fp, "WCSTY2P", "PHYSICAL", "");
    ()=_fits_update_key(fp, "CUNIT2P", "", "");
    if(dx != 0.0 && dy != 0.0)
    {
      ()=_fits_update_key(fp, "LTV1",   1.0 - Xvalues[0]/dx, "");
      ()=_fits_update_key(fp, "LTM1_1", 1.0/dx, "");
      ()=_fits_update_key(fp, "LTV2",   1.0 -Yvalues[0]/dy, "");
      ()=_fits_update_key(fp, "LTM2_2", 1.0/dy, "");
    }
  }

  fits_write_binary_table(fp, "confmap data", confmap_data);
  variable par;
  foreach par (fieldnames[[:-2]])
    ifnot(par==par1 || par==par2)
      fits_write_image(fp, par,
		       _reshape(get_struct_field(confmap_data, par), [n2, n1]),
		       Xvalues, Yvalues, par1, par2);
  fits_close_file(fp);

  % If no error has occured so far:
  if(remove_files)
  {
    foreach file (datafiles)
      ()=remove(file);
    ()=remove(infofile);
  }
}

%
% Code from src/plot/get_contour_lines.sl
%
define get_contour_lines()
%!%+
%\function{get_contour_lines}
%\synopsis{finds a set of contour lines for a 2d-array of values}
%\usage{Struct_Type l[] = get_contour_lines(Double_Type f[], Double_Type f0);}
%\qualifiers{
%\qualifier{save}{filename of a FITS file to save contours}
%}
%\description
%    \code{f} has to be a two-dimensional array (an image).
%    The return value is an array of \code{struct { x, y }},
%    whose fields \code{x} and \code{y} contain the indices
%    for the contour line \code{f[y,x] = f0}.
%!%-
{
  variable f, f0;
  switch(_NARGS)
  { case 2: (f, f0) = (); }
  { help(_function_name()); return; }

  variable save = qualifier("save");
  variable F, n, list, i;
  if(save!=NULL && stat_file(save)!=NULL)
  {
    F = fits_open_file(save, "r");
    ()=_fits_get_num_hdus(F, &n);
    fits_move_to_interesting_hdu(F);
    n--;
    list = Struct_Type[n];
    _for i (0, n-1, 1)
    { list[i] = fits_read_table(F);
      ()=_fits_movrel_hdu(F, +1);
    }
    fits_close_file(F);
    return list;
  }

  variable w = length(f[0,*]);
  variable h = length(f[*,0]);
  variable p = Integer_Type[h, w];
  p[*] = 0;
  n=0;
  variable x, y;
  foreach (where(f>=f0))
  { i = ();
    y = i/w;
    x = i mod w;
    if(   any(f[[int(_max(0, y-1)):int(_min(h-1, y+1))], x]<f0)
       or any(f[y, [int(_max(0, x-1)):int(_min(w-1, x+1))]]<f0))
    { n++; p[y,x] = n; }
    else
    { p[y,x] = -1; }
  }
%  _for $1 (h-1, 0, -1) { print(strjoin(array_map(String_Type, &sprintf, "%3d", p[$1,*]), " ")); }

  i = where(p>0);
  x = i mod w;
  y = i / w;
  list = Struct_Type[0];
  while(length(x)>0)
  { variable X, Y;
    for(X=Integer_Type[0], Y=Integer_Type[0], n=0;  length(x)>0;  )
    { X = [X, x[n]];
      Y = [Y, y[n]];
      x = array_remove(x, n);
      y = array_remove(y, n);
      if(length(x)>0)
      { n = where_min( (x-X[-1])^2 + (y-Y[-1])^2 )[0];
        if( (x[n]-X[-1])^2 + (y[n]-Y[-1])^2 > 2)
          break;
      }
    }
    list = [list, struct {x=X, y=Y}];
  }

  variable j, something_changed;
  do
  { something_changed = 0;
    for(i=0; i<length(list)-1 and not something_changed; i++)
      for(j=i+1; j<length(list) and not something_changed; j++)
      {
	if( (list[i].x[ 0]-list[j].x[ 0])^2 + (list[i].y[ 0]-list[j].y[ 0])^2 <= 2 )
	{ % list[i] = *---|, list[j] = *---|  (* can be connected)
	  list[i].x = [reverse(list[i].x), list[j].x];
	  list[i].y = [reverse(list[i].y), list[j].y];
	  list = array_remove(list, j);
	  something_changed = 1;
	  break;
	}
	if( (list[i].x[ 0]-list[j].x[-1])^2 + (list[i].y[ 0]-list[j].y[-1])^2 <= 2 )
	{ % list[i] = *---|, list[j] = |---*  (* can be connected)
	  list[i].x = [list[j].x, list[i].x];
	  list[i].y = [list[j].y, list[i].y];
	  list = array_remove(list, j);
	  something_changed = 1;
	  break;
	}
	if( (list[i].x[-1]-list[j].x[ 0])^2 + (list[i].y[-1]-list[j].y[ 0])^2 <= 2 )
	{ % list[i] = |---*, list[j] = *---|  (* can be connected)
	  list[i].x = [list[i].x, list[j].x];
	  list[i].y = [list[i].y, list[j].y];
	  list = array_remove(list, j);
	  something_changed = 1;
	  break;
	}
	if( (list[i].x[-1]-list[j].x[-1])^2 + (list[i].y[-1]-list[j].y[-1])^2 <= 2 )
	{ % list[i] = |---*, list[j] = |---*  (* can be connected)
	  list[i].x = [list[i].x, reverse(list[j].x)];
	  list[i].y = [list[i].y, reverse(list[j].y)];
	  list = array_remove(list, j);
          something_changed = 1;
	  break;
	}
      }
  } while(something_changed);

  if(save!=NULL)
  { F = fits_open_file(save, "c");
    _for i (0, length(list)-1, 1)
      fits_write_binary_table(F, string(i), list[i]);
    fits_close_file(F);
  }
  return list;
}

%
% Code from src/slang/arrays/2d/get_coordinatearrays_of_image.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_coordinatearrays_of_image()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_coordinatearrays_of_image}
%\synopsis{returns two 2d arrays of image coordinates}
%\usage{(XX, YY) = get_coordinatearrays_of_image(Any_Type img[,]);}
%\description
%    \code{XX[y, x] = x}  and  \code{YY[y, x] = y} for every coordinate (\code{x,y}) of \code{img}.
%!%-
{
  variable img;
  switch(_NARGS)
  { case 1: img = (); }
  { help(_function_name()); return; }

  variable nx = length(img[0,*]);
  variable ny = length(img[*,0]);

  variable XX = Integer_Type[ny, nx];
  variable YY = Integer_Type[ny, nx];
  variable x, y;
  _for x (0, nx-1, 1)  XX[*,x] = x;
  _for y (0, ny-1, 1)  YY[y,*] = y;

  return (XX, YY);
}

%
% Code from src/data/get_count_rate.sl
%
%%%%%%%%%%%%%%%%
define cut_dataset_range()
%%%%%%%%%%%%%%%%
%!%+
%\function{cut_dataset_range}
%
%\synopsis{Get get_data_counts-struct clipped to energy range}
%\usage{Struct_Type = cut_dataset_range (hist_index , E_min , E_max);}
%\description
%       This function will return a structre just as get_data_counts
%       but clipped to the energy range from E_min up to E_max (energies in keV).
%\example
%       isis>id = load_data("data.fits");
%       isis>variable cut_spectrum = cut_dataset_range (id , 0.2 , 2.3);
%\seealso{get_data_counts}
%!%-
{
    variable setid , e_min , e_max;
    switch(_NARGS)
    { case 3: (setid , e_min , e_max) = (); }
    { return help(_function_name()); }

    variable data = _A(get_data_counts(setid));
    variable ind = where(data.bin_lo >= e_min and data.bin_hi < e_max);
    variable cnt_kev_flt = struct_filter(data , ind ; copy);
    return cnt_kev_flt;

}

private define __counts_from_dataset(id) {
  variable dat = _A(get_data_counts(id));
  variable E_range = qualifier("E_range", NULL);
  if ( E_range != NULL ) {
    variable ind = where(dat.bin_lo >= E_range[0] and dat.bin_hi <= E_range[1]);
    struct_filter(dat, ind);
  }
  return sum(dat.value);
}

private define __get_background_counts(id) {
  variable back_id = -1;
  try {
    back_id = load_data(get_data_info(id).bgd_file);
    if ( back_id == -1 ) {
      throw RunTimeError, sprintf("Something went wrong loading %s", get_data_info(id).bgd_file);
    }
    return __counts_from_dataset(back_id; E_range = qualifier("E_range", NULL));
  } finally {
    if ( back_id != -1 ) {
      delete_data(back_id);
    }
  }
}


%%%%%%%%%%%%%%%%
define get_count_rate(dset)
%%%%%%%%%%%%%%%%
%!%+
%\function{get_count_rate}
%
%\synopsis{Get different variants of countrates and counts}
%\usage{Double_Type CR = get_count_rate (hist_index);
%or
%Double_Type (CR , CR_err) = get_count_rate (hist_index ; err);}
%\description
%
%       Use this function to retrive background subtracted
%       countrates (default), countrates, background rates, or counts
%       and respective uncertainties for spectra that have been loaded.
%       Errors are estimated using the methode described by Gehrels 1986,
%       see \code{gehrels_error} for details. If the err_asym qualifier is not
%       specified the average of both errors is returned. If it is, an array
%       is returned which contains the lower error in the first entry and the
%       upper error as the second entry. The default confidence level for the
%       error is 0.9, but can be changed using the err_conf qualifier within
%       the possibilities offered by \code{gehrels_error}.
%
%\qualifiers{
%    \qualifier{tex}{additional TeXoutput of CRs, choose this if you need
%         values and uncertainties in LaTeX-format (only printed in terminal)}
%    \qualifier{bkg}{0 (default: background subtracted CR), 1 (count rates),
%         2 (background count rate)}
%    \qualifier{fake}{Set this qualifier if you have a faked spectrum and a
%          faked background (which has not been deleted!);
%          Assumption: hist_index(bkg) = hist_index(dset) +1 !!}
%    \qualifier{fake_expo}{set exposure time of fake spectrum, in case
%               set_data_exposure was not used when faking}
%    \qualifier{bkg_id}{set hist_index(bkg) if background is loaded or faked, if set to 0, no background
%            is subtracted. Will overrule the id-setting made by fake-qualifier}
%    \qualifier{E_range}{Set this qualifier if you want the countrate within
%             a certain energy range. Use E_range = [E_min , E_max].
%             If not given full energy range is used. (Energy in keV!)}
%    \qualifier{counts}{if present function will only return counts instead
%             of countrates}
%    \qualifier{err}{if present function will return CR and
%             respective uncertainty}
%    \qualifier{err_conf}{Confidence limit for the error estimation}
%    \qualifier{err_asym}{Return asymmetric errorbars as an array}
%}
%
%\example
%       isis> xray = load_data("data.pha");
%       isis> variable countrate = get_count_rate (1 ;tex, bkg=1 , E_range = [0.2 , 2.3]);
%       isis> variable counts , counts_err;
%       isis> (counts , counts_err) = get_count_rate (1 ;tex, bkg=1 , E_range = [0.2 , 2.3] , counts , err);
%
%\seealso{get_data_counts; TeX_value_pm_error; cut_dataset_range;}
%!%-
{
  variable E_range = qualifier("E_range", NULL);
  variable bkg = qualifier("bkg", 0);
  variable counts_data, back_file, counts_back, counts_err,
	   counts_back_err, counts_sub, counts_sub_err, expos;
  variable bkg_id = qualifier("bkg_id", NULL);
  variable err = qualifier_exists("err");
  variable err_conf = qualifier("err_conf", 0.9);
  variable err_asym = qualifier_exists("err_asym");

  if ( qualifier_exists("fake") || qualifier_exists("bkg_id") ) {
    back_file = qualifier("bkg_id", dset + 1);
    if (back_file == 0) {
      counts_back = 0.0;
    } else {
      counts_back = __counts_from_dataset(back_file ; E_range = E_range);
    }
  } else {
    counts_back = __get_background_counts(dset; E_range = E_range);
  }

  counts_back *= get_data_backscale(dset) / get_back_backscale(dset);
  counts_data = __counts_from_dataset(dset ; E_range = E_range);
  counts_sub = counts_data - counts_back;

  if ( err ) {
    counts_back_err = gehrels_error(counts_back, err_conf);
    counts_err = gehrels_error(counts_data, err_conf);
    counts_sub_err = sqrt(counts_err^2 + counts_back_err^2);
  }

  ifnot ( qualifier_exists("counts") ) {
    expos = qualifier("fake_expo", get_data_exposure(dset));
    counts_data /= expos;
    counts_back /= expos;
    counts_sub /= expos;
    if ( err ) {
      counts_err /= expos;
      counts_back_err /= expos;
      counts_sub_err /= expos;
    }
  }

  variable ret, ret_err;
  if ( bkg == 0 ) {
    ret = counts_sub;
    ret_err = counts_sub_err;
  } else if ( bkg == 1 ) {
    ret = counts_data;
    ret_err = counts_err;
  } else if ( bkg == 2 ) {
    ret = counts_back;
    ret_err = counts_back_err;
  } else {
    throw UsageError, "Wrong Input for Bkg qualifier!";
  }

  if ( qualifier_exists("tex") ) {
    vmessage(TeX_value_pm_error(ret, ret-ret_err, ret+ret_err));
  }
  if ( err ) {
    ifnot ( err_asym ) {
      ret_err = 0.5 * sum(ret_err);
    }
    return ret, ret_err;
  }
  return ret;
}

%
% Code from src/data/spectrum.datastructures/get_data_counts_with_tot_err.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_data_counts_with_tot_err()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_data_counts_with_tot_err}
%\synopsis{returns spectral data, taking systematic errors into account}
%\usage{Struct_Type data = get_data_counts_with_tot_err(Integer_Type id);}
%\description
%    \code{data.err = sqrt( stat_err^2 + [sys_err_frac * data.value]^2 );}
%\seealso{get_data_counts, get_sys_err_frac}
%!%-
{
  variable id;
  switch(_NARGS)
  { case 1: id = (); }
  { help(_function_name()); return; }

  variable dc = get_data_counts(id);
  variable sys_err_frac = get_sys_err_frac(id);
  if(length(sys_err_frac)>0)
    dc.err = sqrt( dc.err^2 + (sys_err_frac*dc.value)^2 );
  return dc;
}

%
% Code from src/slang/date/get_ephemeris.sl
%
private define print_ephem_data(ephem_data)
{
  variable ephem, field;
  foreach ephem (ephem_data)
  {
    ()=printf("%s :", ephem.name);
    variable comma = "";
    variable fields = get_struct_field_names(ephem);
    foreach field (fields[where(fields!="name")])
    {
      variable value =  get_struct_field(ephem, field);
      if(typeof(value)==String_Type)  value = make_printable_string(value);
      ()=printf("%s %s=%S", comma, field, value);
      comma = ",";
    }
    ()=printf("\n");
  }
}


define get_ephemeris()
%!%+
%\function{get_ephemeris}
%\synopsis{returns the epoch and period of an ephemeris}
%\usage{Struct_Type get_ephemeris(String_Type ephemeris)}
%\description
%    The list of known ephemerides is stored internally
%    and is shown if an unknown ephemeris is requested.
%    Ephemerides are returned as a structure, containing
%    at least the internal \code{name} of the ephemeris (which
%    usually contains the source name and a reference),
%    the epoch \code{T0}, and the period \code{P}.
%    More information may be stored in further reasonably named
%    fields, e.g., \code{Pdot} for the change of the period.
%\seealso{orbitalphase}
%!%-
{
  if(_NARGS!=1)  return help(_function_name());
  variable ephem_name = ();
  
  variable ephem_data = [
    struct{ name="CygX-1_Gies2008",         T0=52872.288,    P=  5.599829, T0_err=0.009 },
    struct{ name="CygX-1_Gies2003",         T0=51729.949,    P=  5.599829, T0_err=0.008 },
    struct{ name="CygX-1_Brocksopp1999",    T0=41874.207,    P=  5.599829, T0_err=0.009,  P_err=0.000016 },
    struct{ name="CygX-1_Lachowicz2006",    T0=50514.59,     P=151.43, note="superorbital modulation" },
    struct{ name="LMCX-1_Orosz2008",        T0=53390.8436,   P=  3.90917,  T0_err=0.0120, P_err=0.00005, reference="http://adsabs.harvard.edu/abs/2009ApJ...697..573O, Table 3, adopted value" },
	struct{ name="LMCX-4_Falanga2015",      T0=53013.5910,   P= 1.4083790,  T0_err = 0.0008, P_err = 7e-7},
	struct{ name="VelaX-1_Kreykenbohm2008", T0=52974.227,    P=  8.964386, T0_err=0.007,  P_err=0.000015 },
	struct{ name="VelaX-1_Falanga2015",     T0=42611.349,    P= 8.964427,   T0_err = 0.013,  P_err = 1.2e-5},
    struct{ name="CygX-2_Casares2010",      T0=51386.648,    P=  9.8445,   T0_err=0.018,  P_err=0.00019 },
    struct{ name="CygX-2_Casares1998",      T0=49339.00,     P=  9.8444,   T0_err=0.03,   P_err=0.0003 },
    struct{ name="CenX-3_Raichur2010",      T0=50506.788423, P=  2.08713936, Pdot=1.799e-6/365.25, Pddot=0 },  % be aware of \dot Porb
	struct{ name="CenX-3_Falanga2015",      T0=50506.788423, P= 2.08704106, T0_err = 7e-6, P_err = 3e-8},
	struct{ name="4U1907_Inam2009",         T0=54467.6,      P=  8.3753, asini= 83.0, eccentricity=0.28, omega=300 },
    struct{ name="4U0115_Mueller2010",      T0=49279.2677,   P= 24.31617 },
    struct{ name="IGR_J19140_Corbet2004",   T0=51593.4,      P= 13.558, note="orbit measured only from ASM data" },
    struct{ name="GX301-2_Koh1997",         T0=48802.79,     P= 41.498, asini=368.3, eccentricity=0.462, omega=310.4},
    struct{ name="4U1909_Levine2004",       T0=52631.383,    P= 4.4007, asini=47.83, eccentricity=0.021},
    struct{ name="4U1700_Falanga2015",      T0=53785.850,    P= 3.411581,   T0_err = 0.007,  P_err = 7e-6},
	struct{ name="4U1538_Falanga2015",      T0=52855.061,   P= 3.7284140,  T0_err = 0.013,  P_err = 7.6e-6},
	struct{ name="SMCX-1_Falanga2015",      T0=52846.688810, P= 3.891923160,T0_err = 2.4e-5, P_err = 6.6e-8},
	struct{ name="SAXJ1802_Falanga2015",    T0=52168.245,    P= 4.5697,     T0_err = 0.034,  P_err = 0.0001},
	struct{ name="XTEJ1855_Falanga2015",    T0=52704.009,    P= 6.07415,    T0_err = 0.017,  P_err = 8e-5},
	struct{ name="EXO1722_Falanga2015",     T0=53761.695,    P= 9.74079,    T0_err = 0.019,  P_err = 8e-5},
	struct{ name="OAO1657_Falanga2015",     T0=52674.119,    P=10.447355,   T0_err = 0.0017, P_err = 9.2e-5}
   ];

  variable j, i = {};  % ephem_name=="" shall be equivalent to an unknown ephemeris
  if(ephem_name != "")
    _for j (0, length(ephem_data)-1, 1)
      if(string_match(ephem_data[j].name, ephem_name))
	list_append(i, j);

  if(length(i))
  {
    if(length(i)>1)
    {
      vmessage(`warning (%s): The following ephemerides match the name "%s":`, _function_name(), ephem_name);
      print_ephem_data(ephem_data[list_to_array(i, Integer_Type)]);
      message("");
    }
    i = i[0];  % pick the first one
    if(ephem_data[i].name != ephem_name)
      vmessage(`warning (%s): using ephemeris "%s" instead of "%s"`, _function_name(), ephem_data[i].name, ephem_name);
    return ephem_data[i];
  }
  else
  {
    vmessage(`error (%s): ephemeris "%s" is not among list of known ephemerides:`, _function_name(), ephem_name);
    print_ephem_data(ephem_data);
  }
}

%
% Code from src/data/spectrum.datastructures/get_flux_corrected_convolved_model_flux.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_flux_corrected_convolved_model_flux(id)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_flux_corrected_convolved_model_flux}
%\synopsis{computes the model flux which can be compared with a flux-corrected spectrum}
%\usage{Struct_Type get_flux_corrected_convolved_model_flux(Integer_Type hist_index)}
%\description
%    The flux-corrected convolved model flux is computed as\n
%      \code{            int K(R(h,E), A(E), S(E)) dE  }\n
%      \code{ F(h)  =  --------------------------------} ,\n
%      \code{            int K(R(h,E), A(E),  1  ) dE  }\n
%    where \code{S(E)} is the flux model, \code{A(E)} is the effective area (the ARF),
%    \code{R(h,E)} is the redistribution function (the RMF),
%    and the kernel \code{K} defaults to \code{K(R,A,S) = R*A*S}.
%
%    The flux-corrected convolved model fluxes is defined in such a way
%    that a folded model is "unfolded" in the same way as the data are
%    by virtue of \code{flux_corr}. It can thus be compared with flux-corrected data.
%\seealso{flux_corr, get_model_counts, get_model_flux, get_convolved_model_flux}
%!%-
{
  variable data_info = get_data_info(id);
  variable FitVerbose = Fit_Verbose;

  rebin_data(id, 0);
  Fit_Verbose = -1;
  ()=eval_counts;
  variable m = get_model_counts(id);
  m.err = sqrt(m.value);
  variable new_id = define_counts(m);
  set_data_exposure(new_id, get_data_exposure(id));
  if(length(data_info.arfs))  assign_arf(data_info.arfs[0], new_id);  % How to treat multiple ARFs?
  if(length(data_info.rmfs))  assign_rmf(data_info.rmfs[0], new_id);  % How to treat multiple RMFs???
  % TODO: worry about background
  % TODO: probably insert  rebin_data(new_id, data_info.rebin);
  flux_corr(new_id);
  m = get_data_flux(new_id);
  delete_data(new_id);

  rebin_data(id, data_info.rebin);
  ignore(id); notice_list(id, data_info.notice_list);
  Fit_Verbose = FitVerbose;
  return m;
}

%
% Code from src/slang/arrays/2d/get_grid.sl
%
%%%%%%%%%%%%%%%
define get_grid()
%%%%%%%%%%%%%%%
%!%+
%\function{get_grid}
%\usage{(XX, YY) = get_grid(Double_Type X[], Double_Type Y[]);}
%\description
%    \code{XX} and \code{YY} are two-dimensional arrays
%    which span the grid defined by the arrays \code{X} and \code{Y}.
%\example
%    variable XX, YY;  (XX, YY) = get_grid([-2:2:0.01], [-2:2:0.01]);
%    plot_image( atan2(YY, XX) );
%!%-
{
  variable X, Y;
  switch(_NARGS)
  { case 2: (X, Y) = (); X = [X]; Y = [Y]; }
  { help(_function_name()); return; }

  variable iX, nX = length(X);
  variable iY, nY = length(Y);
  variable XX = Double_Type[nY, nX];
  variable YY = Double_Type[nY, nX];
  _for iX (0, nX-1, 1)
    XX[*, iX] = X[iX];
  _for iY (0, nY-1, 1)
    YY[iY, *] = Y[iY];
  return (XX, YY);
}

%
% Code from src/slang/arrays/get_intersection.sl
%
define get_intersection()
%!%+
%\function{get_intersection}
%\synopsis{finds elements occuring in two arrays}
%\usage{(Integer_Type i1[], i2[]) = get_intersection(array1[], array2[]);}
%\description
%     \code{array1[ i1 ]  ==  array2[ i2 ]}
%\qualifiers{
%\qualifier{ordered}{assumes that \code{array1} and \code{array2} are ordered increasingly}
%}
%!%-
{
  variable array1, array2;
  switch(_NARGS)
  { case 2: (array1, array2) = (); }
  { help(_function_name()); return; }

  variable l1 = length(array1);
  variable l2 = length(array2);
  variable indx1 = Integer_Type[0];
  variable indx2 = Integer_Type[0];
  variable i1=0, i2=0;
  if(qualifier_exists("ordered"))
    while(i1<l1 && i2<l2)
    { variable diff = array1[i1] - array2[i2];
      if(diff!=0)
	if(diff<0)  % array1[i1] < array2[i2]
	  i1++;
        else        % array1[i1] > array2[i2]
	  i2++;
      else	   % array1[i1] == array2[i2]
      { indx1 = [indx1, i1];
        indx2 = [indx2, i2];
        i1++;
        i2++;
      }
    }
  else
    _for i1 (0, l1-1, 1)
    { i2 = where(array2 == array1[i1]);
      switch(length(i2))
      { case 0: ; }
      { case 1: indx1 = [indx1, i1];  indx2 = [indx2, i2]; }
      { vmessage("error (%s): intersection not unique", _function_name()); return NULL, NULL; }
    }
  return indx1, indx2;
}

%
% Code from src/FITS/get_map_value_at_position.sl
%

% only activate if module fitswcs could be loaded
#ifeval __get_reference("fitwcs_get_img_wcs")!=NULL

%!%+
%\function{get_map_value_at_position}
%\synopsis{Extract map value from ra/dec position}
%\usage{Double_Type value = get_map_value_at_position(map,ra,dec);}
%\qualifiers{
%\qualifier{verbose}{Output detector coordinates on display. Default: 0}
%}
%\description
%   This function returns the value of a map at a certain ra/dec position
%   using the wcs of that map. Ra/Dec coordinates must be given as
%   decimal numbers.
%\seealso{read_difmap_fits, fitswcs_get_img_wcs, wcsfuns_project}
%!%-
define get_map_value_at_position() {
  variable map,ra,dec;
  switch(_NARGS)
  { case 3: (map,ra,dec) = (); }
  { help(_function_name);return;}
  variable verbose = qualifier("verbose",0);
  variable s = read_difmap_fits(map);
  variable wcs = fitswcs_get_img_wcs (map);
  variable o_map = @(s.img);
  variable pos_x,pos_y;
  (pos_x,pos_y) = wcsfuns_project (wcs, ra,dec);
  if (verbose == 1) vmessage("detector coordinates: x=%.2f y=%.2f",pos_x,pos_y);
  
  return o_map[int(pos_x),int(pos_y)];
}

#endif

%
% Code from src/data/get_modelcomponents.sl
%
define get_modelcomponents()
%!%+
%\function{get_modelcomponents}
%\synopsis{load binned model component values}
%\usage{Struct_Type[] mc = get_modelcomponent( Integer_Type hist_index);}
%\altusage{Struct_Type[] mc = get_modelcomponent( );}
%\qualifiers{
%\qualifier{fct [=NULL]:}{ NULL:    Using 'get_model' (default)
%                  "counts": Using 'get_model_counts'
%                  "flux":   Using 'get_model_flux'
% }
%}
%\description
%     For each model component (see 'get_fun_components') within the given
%     dataset 'hist_index' this funtion returns a stucture with three array
%     fields, bin_lo [Angstrom], bin_hi [Angstrom] and value, containging
%     only the contribution of this component to the overall model.
%     If 'hist_index' isn ot given all included datasets are used!
%
%     The model contribution of each component is obtained by setting all
%     the norm parameters (get_par_info: is_a_norm=1) of all other components
%     to zero and evaluating the model. NOTE THAT for each component an
%     'eval_counts' is called!
%
%     The 'fct' qualifier determines which function should be used to
%     obtain the model components.
%     
%                                              
%\seealso{get_fun_components, get_model, get_model_counts, get_model_flux}
%!%-
{
  %%% Grep excluded/included datasets
  variable exc_data = merge_struct_arrays(get_data_info(all_data)).exclude;
  variable inc_data = wherenot(exc_data)+1;
  exc_data = where(exc_data)+1;
  
  variable id;
  switch(_NARGS)
  { case 0 : id = inc_data; }
  { case 1 : id = ();
    %% exclude all other datasets but 'id'
    include (all_data);
    exclude (all_data[complement( all_data, id )]);
  }
  { help(_function_name); return;}

  %% Set get_model method according qualifier
  variable fct = qualifier("fct",NULL);
  variable get_model_fct = &get_model;
  ifnot( _isnull(fct) ){
    if( fct == "counts" ){
      get_model_fct = &get_model_counts;
    }
    else if( fct == "flux" ){
      get_model_fct = &get_model_flux;
    }
    else{ help(_function_name); return; }
  }
  
  %% Store the original parameters
  variable PARAMETERS = get_params;
  
  %% get fit fun component names
  variable cnames = get_fun_components;
  variable ncomp  = length(cnames);
  variable normid = Array_Type[ncomp];
    
  variable ii, p;
  _for ii ( 0, ncomp-1 ){ % loop over all components of the current dataset
    p = get_params( cnames[ii]+"*" );
    % find the indices of all norm parameter of the current component
    normid[ii] = where( array_map( Integer_Type, &get_struct_field, p, "is_a_norm" ) );
    normid[ii] = array_map( Integer_Type, &get_struct_field, p[normid[ii]], "index" );
  }

  %% unset 'set_par_fun' of norm parameter, otherwise it is not possible to set those Zero later
  array_map( Void_Type, &set_par_fun, array_flatten(normid), NULL );

  %% make sure min/max range of norm parameters includes 0!
  p = get_params( array_flatten(normid) );
  array_map( Void_Type, &set_struct_field, p, "min", 0 );
  set_params(p);
  
  %% number of norms per component
  variable nnorm = array_map( Integer_Type , &length, normid ); 

  %% Prepare structure containing model component data
  variable MC = struct_array( length(where(nnorm)),
			      struct{
				name,
				nnorm,
				bin_lo,
				bin_hi,
				value,
				err = NULL				
				}
			    );

  variable nind = where(nnorm);
  variable f;
  _for ii ( 0, length(MC)-1 ){
    MC[ii].name = cnames[nind[ii]];
    MC[ii].nnorm = nnorm[nind[ii]];

    %% store parameter
    p = get_params;
   
    %% set all norms to ZERO but that/those of the current component
    array_map( Void_Type, &set_par, array_flatten( array_remove(normid,nind[ii]) ), 0 );

    %% obtain model component data
    () = eval_counts();
    f = @get_model_fct(id);
    MC[ii] = struct_combine( MC[ii], f );
    
    %% reset parameter values
    set_params(p);
  }
  

  % reverse all changes
  set_params( PARAMETERS ); % IMPORTANT execute with the same included/excluded datasets

  include( all_data );
  exclude( exc_data );

  return MC;
}

%
% Code from src/slang/arrays/get_par_combinations.sl
%
define get_par_combinations()
%!%+
%\function{get_par_combinations}
%\synopsis{calculates all possible parameter combinations}
%\usage{Struct_Type[] parcomb = get_par_combinations( Struct_Type par )}
%\description
%   All possible parameter combinations are calculated for the
%   given parameter structure. That means if there are multiple
%   parameters, each with there own grid all possible combinations
%   are returned.
%
%   'par' must be a Struct_Type, but there are no restriction
%   otherwise. Each field is representing a parameter and its
%   value its grid, which can be any Data_Type (e.g., Double_Type
%   or String_Type)!
%
%   NOTE that the combination is done, s.t. the last parameter
%   in the struct is varied first and the first parameter last,
%   i.e., the first parameter in the parameter combination
%   will stay constant longest!
%
%\examples
%   variable par = struct{ a = [0:1:#3],
%                          b = ["hello","world"],
%                          c = [ 2001, 2012]
%                        };
%   variable parcomb = get_par_combinations( par );
%   print(parcomb);
%!%-
{
  variable par;
  switch(_NARGS)
  { case 1 : par = (); }
  { help(_function_name()); return; }
  
  %% get the names of the parameters
  variable parnames = get_struct_field_names( par );
  
  % get the length of each individual 'par' range
  variable Ni = array_map( Integer_Type, &length, array_map( Array_Type, &get_struct_field, par, parnames ));
  % number of all possible combinations
  variable N = nint(prod(Ni));
  
  % initialize parcom struct array:
  variable parcomb = @Struct_Type(parnames);
  % NOTE: It was parcomb = struct_array( N, @Struct_Type(parnames) )
  %       which is a memery eater !!!!
  
  variable ii;
  _for ii ( 0, length(Ni)-1 ){ % loop all parameters and set their values
    set_struct_field(parcomb, parnames[ii], _typeof( get_struct_field(par,parnames[ii]) )[N] );
  }
  
  variable nn, ind;
  _for nn ( 0, N-1 ){ % loog all possible combinations
    _for ii ( 0, length(Ni)-1 ){ % loop all parameters and set their values
      ind = ( nn / nint(prod( Ni[[ii+1:length(Ni)-1]] )) ) mod Ni[ii];
      get_struct_field( parcomb, parnames[ii] )[nn] = [get_struct_field( par, parnames[ii] )][ind];
    }
  }
  
  return parcomb;
}

%
% Code from src/sixte/get_param_from_filename.sl
%
define get_param_from_filename(){
%!%+
%\function{get_param_from_filename}
%\synopsis{returns a double parameter value from a string / filename}
%\usage{Double_Type val = get_param_from_filename(filename,key,regexp);}
%\description
%    For a given filename = data/f4l_flux0001000muCrabpattern.fits, the
%    value can be extracted by specifying the key="flux"; and a regexp
%    of regexp="flux\\([0-9.]+\\)muCrab"R;
%    Note that there is a strict naming convention:
%    1) only one "." to mark the file type
%    2) parameters are separated by "_"
%    3) the "key" parameter always has to be part of the parameter string
%       (i.e. fluxXXXXXXXmuCrab or fluxX.XXXXXXcgs without an underscore)
%    
%    If onle one argument is given, the above example for the flux is
%    automatically used as key and regexp.
%\seealso{simput_athenacrab}
%!%-

   variable s,key = "flux";
   variable regexp = "flux\([0-9.]+\)muCrab"R;
   switch(_NARGS)
   {case 1: (s) = (); }
   {case 3: (s,key,regexp) = (); }
   { help(_function_name()); return; }
   
   variable stmp = strchop(s,'/',0)[-1];

   s=stmp; % second part encodes the filetype
   stmp = strchop(s,'_',0);

   s = stmp[where(is_substr(stmp,key) == 1)];
   
   if (length(s) != 1){
      vmessage("Can not read convert parameter from file %s due to wrong format!",stmp);
      return -1.0;
   }
   s = s[0];
   stmp = string_matches (s,regexp)[1];

   return atof(stmp);
}

%
% Code from src/fitting/parameters/get_params_from_file.sl
%
%<DEPRECATED>
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_params_from_file()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_params_from_file}
%\synopsis{retrieves fit-parameter information from a file}
%\usage{Struct_Type params[] = get_params_from_file(String_Type params);}
%\seealso{load_par, get_params}
%!%-
{
  vmessage("*** Warning: Deprecated function, use 'read_par' instead");
  read_par();
#iffalse
  variable filename;
  switch(_NARGS)
  { case 1: filename = (); }
  { help(_function_name()); return; }

  variable fit_fun_before = get_fit_fun();
  variable params_before = get_params();

  load_par(filename);
  variable params = get_params();

  if(fit_fun_before != NULL)
  { fit_fun(fit_fun_before);
    set_params(params_before);
  }

  return params;
#endif
}
%</DEPRECATED>

% This must be in sync with the isis source, otherwise we can not
% read par files.

%%%%%%%%%%%%%%%
define read_par ()
%%%%%%%%%%%%%%%
%!%+
%\function{read_par}
%\synopsis{Read function parameter from file}
%\usage{Struct_Type[] params = read_par(filename[, funp]);}
%\description
%  This function process a parameter file according to the ISIS
%  file convention (safed, e.g., with save_par) and returns the
%  paramter settings as an array of structs. This array is
%  compatible to the array retireved by get_params and can
%  therefore be used with set_params.
%
%  If an additional parameter is given, it is expected to be a
%  pointer to a variable in which the function string gets stored.
%
%\seealso{get_params, set_params, save_par}
%!%-
{
  variable file, funp;
  switch (_NARGS)
  { case 1: file = (); funp = NULL; }
  { case 2: (file, funp) = (); }
  { help(_function_name()); return; }

  variable fp = fopen(file, "r");
  variable line, status, nvalues, line_num = 0;
  variable pars = {}, par;
  variable par_default = struct {
    name,
    index,
    value,
    min,
    max,
    freeze,
    tie,
    fun = NULL, % defaults to null
  };

  % skip first lines if they are comments (staring with '#')
  do {
    status = fgets(&line, fp);
    line_num++;
  } while ((status > 0) && (line[0] == '#'));

  if (status < 0)
    throw IOError, sprintf("Invalid parameter file: '%s'", file);

  % copy function string
  if (NULL != funp)
    @funp = strtrim(line);

  while (fgets(&line, fp) > 0) {
    line_num++;

    if (line[0] == '#') {
      if (line[[0:2]] == "#=>")
	par.fun = strtrim(line[[3:]]);
    } else {
      % everything else must be a parameter
      par = @(par_default);
      % this match is easier compared to isis internals
      nvalues = sscanf(line, "%u %s %d %d %le %le %le",
		       &(par.index),
		       &(par.name),
		       &(par.tie),
		       &(par.freeze),
		       &(par.value),
		       &(par.min),
		       &(par.max));
      if (nvalues == 0)
	continue;
      if (nvalues != 7)
	throw IOError, sprintf("Unable to parse %s:%d", file, line_num);
      list_append(pars, par);
    }
  }

  variable i;
  foreach par (pars) {
    if (par.tie != 0) {
      % not elegant, but do we expect to have thousands of parameters?
      _for i (0, length(pars)-1) {
	if (typeof(par.tie) != String_Type && pars[i].index == par.tie)
	  par.tie = pars[i].name;
      }

      if (String_Type != typeof(par.tie))
	par.tie = NULL;
    }
  }

  return list_to_array(pars);
}

%
% Code from src/fitting/parameters/get_params_table_from_files.sl
%
define get_params_table_from_files()
%!%+
%\function{get_params_table_from_files}
%\synopsis{retrieves fit-parameter information from files}
%\usage{Struct_Type get_params_table_from_files(String_Type filenames[]);}
%\qualifiers{
%\qualifier{free}{only include free parameters}
%\qualifier{par}{array of parameters to be included}
%\qualifier{filename}{include filenames into the structure}
%\qualifier{nominmax}{do not include min max values}
%\qualifier{eval_counts}{evaluates the model and returns the statistic as well.
%                   It is assumed that the appropriate data is already loaded.}
%\qualifier{verbose}{show name of parameter files while being processed}
%}
%\description
%    \code{filename} may be a globbing expression.
%    It is assumed that all parameter files rely on the same model
%    and that this model can be loaded with the current data set.
%!%-
{
  variable parfiles;
  switch(_NARGS)
  { case 1: parfiles = (); }
  { help(_function_name()); return; }

  parfiles = glob(parfiles);
  parfiles = parfiles[array_sort(parfiles)];
  variable i, n = length(parfiles);

  variable fieldname, fieldnames = String_Type[0];  % field names of the returned structure

  variable store_filenames = qualifier_exists("filename");
  if(store_filenames)  fieldnames = [fieldnames, "filename"];

  load_par(parfiles[0]);  % use parameter names from first file
  variable par = get_params();
  variable parnames = array_struct_field(par, "name");
  if(qualifier_exists("free"))
    parnames = parnames[where(    array_struct_field(par, "freeze")==0
%			      and array_struct_field(par, "tie")==NULL  % array_struct_field cannot work with NULLs and integers
%			      and array_struct_field(par, "fun")==NULL  % same here
			     )];
  parnames = qualifier("par", parnames);
  variable nominmax = qualifier_exists("nominmax");
  foreach par ( par )  % (get_params())
    if(any(par.name == parnames))
      fieldnames = [fieldnames, escapedParameterName(par.name)+(nominmax ? "" : ["", "_min", "_max"])];
  if(qualifier_exists("eval_counts"))
    fieldnames = [fieldnames, "statistic", "num_variable_params", "red_statistic"];
  variable params_table = @Struct_Type(fieldnames);
  foreach fieldname (fieldnames)  set_struct_field(params_table, fieldname, Double_Type[n]);
   if(store_filenames)  params_table.filename = parfiles;

  variable verbose = qualifier_exists("verbose") || qualifier_exists("eval_counts");
  _for i (0, n-1, 1)
  { if(verbose)  vmessage("processing %s", parfiles[i]);
    load_par(parfiles[i]);
    foreach par (get_params())
      if(any(par.name == parnames))
      { variable escaped_par_name = escapedParameterName(par.name);
        get_struct_field(params_table, escaped_par_name       )[i] = par.value;
	ifnot(nominmax)
	{ get_struct_field(params_table, escaped_par_name+"_min")[i] = par.min;
          get_struct_field(params_table, escaped_par_name+"_max")[i] = par.max;
	}
      }
    if(qualifier_exists("eval_counts"))
    { variable s;
      ()=eval_counts(&s);
      params_table.statistic          [i] = s.statistic;
      params_table.num_variable_params[i] = s.num_variable_params;
      params_table.red_statistic      [i] = s.statistic/(s.num_bins-s.num_variable_params);
    }
  }
  return params_table;
}

%
% Code from src/plot/get_rainbow_col.sl
%

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module rand could be loaded
#ifeval __get_reference("rand_flat")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

private define get_col_map()
{
   variable R = ([0:255])/255.;
   variable G = [0:255]/255.;
   variable B = [0:255]/255.;
   R = ad_colormap_gauss(0.6,R,0.3) +ad_colormap_gauss(1,R,0.3)*1.3 + 
     ad_colormap_gauss(0.47,R,0.13)/3;
   B = ad_colormap_gauss(0.2,B,0.2)+ad_colormap_gauss(0.08,B,0.07)/3;
   G = ad_colormap_gauss(0.95,G,0.3) + ad_colormap_gauss(0.21,G,0.12)/3;
   
   
   
   variable RI,GI,BI;
   RI = ad_colormap_norm_col(R);
   GI = ad_colormap_norm_col(G)*19/20;
   BI = ad_colormap_norm_col(B)*3/4;
   png_add_colormap("td_2d_rainbow", ((RI << 16) | (GI << 8) | BI ));
   
%   variable col_map = png_get_colormap("td_2d_rainbow");   
   
   return "td_2d_rainbow";
   
}

%%%%%%%%%%%%%
define get_rainbow_col(i,num)
%%%%%%%%%%%%%
%!%+
%\function{get_rainbow_col}
%\synopsis{get a rainbow color}
%\usage{color = get_rainbow_col(Integer_Type value, Integer_Type num_colors);}
%\description
%  The value has to be within [0:num_colors-1], to get the correct
%  color here.
%\seealso{xfig_new_color}
%!%-
{

   if (i > (num-1)) return "black";
   
   variable col_map = png_get_colormap(qualifier("col_map",get_col_map()));
    variable fnum,fname;
    if (qualifier_exists("linear_numbers")) {
	fnum=i;
    } else {
	fnum = rand();
    }
   fname="td_rainbow"+string(fnum);
   xfig_new_color(fname,col_map[nint(((i*1.)*255)/(num-1))];);
   
  return fname;
}

#endif

#endif

#endif

%
% Code from src/data/spectrum.datastructures/get_ratio.sl
%
%%%%%%%%%%%%%%%%
define get_ratio()
%%%%%%%%%%%%%%%%
%!%+
%\function{get_ratio}
%\synopsis{calculates data/model ratios}
%\usage{Struct_Type rat[] = get_ratio(Integer_Type id[]);}
%\description
%    Each \code{rat[i]} is a \code{{ bin_lo, bin_hi, value, err }} structure
%    where \code{rat[i].value} contains the (data counts)/(model counts)
%    ratios obtained for the data set \code{id[i]}.\n
%    If only a scalar \code{id} is given, only a single structure is returned.
%\seealso{get_data_counts, get_model_counts, get_residuals}
%!%-
{
  variable bin_lo, bin_hi, data, data_err, model;
  variable rat = struct { bin_lo, bin_hi, value, err };
  switch(_NARGS)
  { case 1:
      variable ids = (), id, rats = Struct_Type[0];
      foreach id ([ids]);
      { data = get_data_counts(id);
        model = get_model_counts(id);
        data.value = data.value/model.value;
        data.err = data.err/model.value;
        rats = [rats, data];
      }
      if(length(rats)==1) { return rats[0]; } else { return rats; }
  }
  { case 5:
      (bin_lo, bin_hi, data, data_err, model) = ();
      rat.bin_lo = bin_lo;
      rat.bin_hi = bin_hi;
      rat.value = data/model;
      rat.err = data_err/model;
      return rat;
  }
  { case 8:
      variable data1, data1_err, model1, data2, data2_err, model2;
      (bin_lo, bin_hi, data1, data1_err, model1, data2, data2_err, model2) = ();
      rat.bin_lo = bin_lo;
      rat.bin_hi = bin_hi;
      rat.value = (data1+data2)/(model1+model2);
      rat.err = sqrt(data1_err^2+data2_err^2)/(model1+model2);
      return rat;
  }
  {%else:
      message("usage: rats = get_data_model_ratio(ids); % ids may be an array");
      message("   or: rats = get_data_model_ratio(bin_lo, bin_hi, data, data_err, model");
      return NULL;
  }
}

%
% Code from src/data/spectrum.datastructures/get_residuals.sl
%
%%%%%%%%%%%%%%%%%%%%
define get_residuals()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_residuals}
%\synopsis{calculates (data-model)/error residuals}
%\usage{Struct_Type res = get_residuals(Integer_Type id[]);}
%\description
%    Each \code{res[i]} is a \code{{ bin_lo, bin_hi, value, err }} structure
%    where \code{res[i].value} contains the (data counts - model counts)/error
%    residuals obtained for the data set \code{id[i]}.\n
%    If only a scalar \code{id} is given, only a single structure is returned.
%\qualifiers{
%\qualifier{noticed}{restrict to noticed bins}
%\qualifier{keV}{convert Angstrom-bins to keV-bins}
%}
%\seealso{get_data_counts, get_model_counts, get_ratio}
%!%-
{
  variable ids;

  switch(_NARGS)
  { case 1: ids = (); }
  { help(_function_name()); return; }

  variable id, res = Struct_Type[0];
  foreach id (ids)
  { variable data = get_data_counts(id);
    variable model = get_model_counts(id);
    data.value = (data.value-model.value)/data.err;
    data.err = data.err*0+1;
    if(qualifier_exists("noticed"))
      struct_filter(data, get_data_info(id).notice_list);
    if(qualifier_exists("keV"))
      data = _A(data);
    res = [res, data];
  }
  if(length(res)==1) { return res[0]; } else { return res; }
}

%
% Code from src/data/spectrum.datastructures/get_selected_data_flux_en.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_selected_data_flux_en()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_selected_data_flux_en}
%\synopsis{returns the flux of a data set in energy units}
%\usage{flux = get_selected_data_flux_en(id[, Emin, Emax[, alpha]]);}
%\description
%    \code{flux} is a \code{{ bin_lo, bin_hi, value, err}} structure
%    containing energy bins of data set \code{id}
%    and the spectral photon flux density (in 1/s/cm^2/keV, unless \code{alpha!=0}).\n
%    If \code{Emin} and \code{Emax} are used, only values in this energy range are considered.\n
%    If \code{alpha}!=0, the flux values are multiplied with E^\code{alpha}.
%\example
%    \code{hplot( get_selected_data_flux_en(1, 4, 20, 1) );}\n
%    % plots the spectral energy flux density of data set 1 in the 4--20 keV range
%!%-
{
  variable id, E1=0, E2=1e38, alpha=0;
  switch(_NARGS)
  { case 1 :  id = (); }
  { case 3 : (id, E1, E2) = (); }
  { case 4: (id, E1, E2, alpha) = (); }

  variable f = _A( get_data_flux(id) );
  struct_filter(f, where(E1<=f.bin_lo and f.bin_hi<=E2));
  f.value /= (f.bin_hi - f.bin_lo);
  f.err /= (f.bin_hi - f.bin_lo);

  if(alpha!=0)
  { variable E = (f.bin_lo + f.bin_hi)/2;
    f.value *= E^alpha;
    f.err *= E^alpha;
  }

  return f;
}

%
% Code from src/fitting/fit-functions/simple_gpile/get_simple_gpile_info.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_simple_gpile_info()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_simple_gpile_info}
%\synopsis{retrieves pileup information within a simple_gpile2 model}
%\usage{Struct_Type info = get_simple_gpile_info(Integer_Type id);}
%!%-
{
  variable id;
  switch(_NARGS)
  { case 1: id = (); }
  { help(_function_name()); return; }

  variable info = struct { bin_lo, bin_hi, flux, arf, countrate, arf2, countrate2, arf3, countrate3, pileup, index_max_pileup };
  info.bin_lo = get_data_counts(id).bin_lo;
  info.bin_hi = get_data_counts(id).bin_hi;
  variable piled_fit_fun = get_fit_fun();
  fit_fun( get_unpiled_fit_fun() );
  info.flux = eval_fun(info.bin_lo, info.bin_hi);
  fit_fun( piled_fit_fun );

  variable model = "simple_gpile2("+string(id)+").";
  variable beta = get_par(model+"beta");
  variable indx = typecast(get_par(model+"data_indx"), Integer_Type);
  if(indx == 0 or beta == 0.)  return info;

  variable arf_indx = get_data_info(indx).arfs[0];
  variable arf = get_arf( arf_indx );
  variable fracexpo = get_arf_info(arf_indx[0]).fracexpo;
  if(length(fracexpo)>1)  fracexpo[where(fracexpo==0)] = 1.;  else  { if(fracexpo==0)  fracexpo = 1.; }

  info.arf = rebin(info.bin_lo, info.bin_hi, arf.bin_lo, arf.bin_hi, arf.value/fracexpo * (arf.bin_hi-arf.bin_lo) ) / (info.bin_hi-info.bin_lo);
  info.countrate = info.flux * info.arf / (info.bin_hi-info.bin_lo);

  variable mod_ord;
  variable arf2_indx = typecast(get_par(model+"arf2_indx"), Integer_Type);
  if(arf2_indx > 0)
  {
    arf = get_arf(arf2_indx);
    fracexpo = get_arf_info(arf2_indx).fracexpo;
    if(length(fracexpo)>1) { fracexpo[where(fracexpo==0)] = 1.; } else { if(fracexpo==0) { fracexpo = 1.; } }
    info.arf2 = rebin(info.bin_lo, info.bin_hi, 2*arf.bin_lo, 2*arf.bin_hi, arf.value/fracexpo * (arf.bin_hi-arf.bin_lo) ) / (info.bin_hi-info.bin_lo);
    info.countrate2 = rebin(info.bin_lo, info.bin_hi, 2*arf.bin_lo, 2*arf.bin_hi, arf.value/fracexpo * rebin(arf.bin_lo, arf.bin_hi, info.bin_lo, info.bin_hi, info.flux) )/(info.bin_hi-info.bin_lo);
  }
  else
    info.countrate2 = 0*info.countrate;

  variable arf3_indx = typecast(get_par(model+"arf3_indx"), Integer_Type);
  if(arf3_indx > 0)
  {
    arf = get_arf(arf3_indx);
    fracexpo = get_arf_info(arf3_indx).fracexpo;
    if(length(fracexpo)>1) { fracexpo[where(fracexpo==0)] = 1.; } else { if(fracexpo==0) { fracexpo = 1.; } }
    info.arf3 = rebin(info.bin_lo, info.bin_hi, 3*arf.bin_lo, 3*arf.bin_hi, arf.value/fracexpo * (arf.bin_hi-arf.bin_lo) ) / (info.bin_hi-info.bin_lo);
    info.countrate3 = rebin(info.bin_lo, info.bin_hi, 3*arf.bin_lo, 3*arf.bin_hi, arf.value/fracexpo * rebin(arf.bin_lo, arf.bin_hi, info.bin_lo, info.bin_hi, info.flux) )/(info.bin_hi-info.bin_lo);
  }
  else
    info.countrate3 = 0*info.countrate;

  info.pileup = 1 - exp(-beta*(info.countrate+info.countrate2+info.countrate3));

  info.index_max_pileup = where(info.pileup == min(info.pileup))[0];

  return info;
}

%
% Code from src/sixte/get_sixte_eventfile_statistic.sl
%
private define calc_pattern_fractions_of_grid(evt, lo, hi, field){
  variable grid = array_struct_field(evt, field);
  variable ret = struct{
    nsings = Double_Type[length(lo)],
    ndoubs = Double_Type[length(lo)],
    ntrips = Double_Type[length(lo)],
    nquads = Double_Type[length(lo)],
    nvalids = Double_Type[length(lo)],
    sfracs = Double_Type[length(lo)],
    dfracs = Double_Type[length(lo)],
    tfracs = Double_Type[length(lo)],
    qfracs = Double_Type[length(lo)],
    sfrac_errs = Double_Type[length(lo)],
    dfrac_errs = Double_Type[length(lo)],
    tfrac_errs = Double_Type[length(lo)],
    qfrac_errs = Double_Type[length(lo)]
  };

  variable ii;
  _for ii (0, length(lo)-1, 1){
    variable evt_filtered = struct_filter(evt, where(lo[ii] <= grid < hi[ii]) ; copy);
    % Note that currently the pile-up events and non-pileup events
    % are NOT distinguished for the energy/channel sorted patterns.
    ret.nsings[ii] = length(where(evt_filtered.type == 0));
    ret.ndoubs[ii] = length(where(evt_filtered.type == 1)) + length(where(evt_filtered.type == 2)) + length(where(evt_filtered.type == 3)) + length(where(evt_filtered.type == 4));
    ret.ntrips[ii] = length(where(evt_filtered.type == 5)) + length(where(evt_filtered.type == 6)) + length(where(evt_filtered.type == 7)) + length(where(evt_filtered.type == 8));
    ret.nquads[ii] = length(where(evt_filtered.type == 9)) + length(where(evt_filtered.type == 10)) + length(where(evt_filtered.type == 11)) + length(where(evt_filtered.type == 12));
    ret.nvalids[ii] = length(where(evt_filtered.type >= 0));
    
    if (ret.nsings[ii]>0) (ret.sfracs[ii], ret.sfrac_errs[ii]) = ratio_error_prop(1.*ret.nsings[ii], 1.*ret.nvalids[ii] ; Poisson);
    if (ret.ndoubs[ii]>0) (ret.dfracs[ii], ret.dfrac_errs[ii]) = ratio_error_prop(1.*ret.ndoubs[ii], 1.*ret.nvalids[ii] ; Poisson);
    if (ret.ntrips[ii]>0) (ret.tfracs[ii], ret.tfrac_errs[ii]) = ratio_error_prop(1.*ret.ntrips[ii], 1.*ret.nvalids[ii] ; Poisson);
    if (ret.nquads[ii]>0) (ret.qfracs[ii], ret.qfrac_errs[ii]) = ratio_error_prop(1.*ret.nquads[ii], 1.*ret.nvalids[ii] ; Poisson);
  }
  
  return ret;
}

%%%%%%%%%%%%%%%%%%%%%%%
define get_sixte_eventfile_statistic(){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_sixte_eventfile_statistic}
%\synopsis{Get the top-level information of a SIXTE event file}
%\usage{Struct_Type ret = get_sixte_eventfile_statistic(String_Type filename[, Double_Type flux]);}
%\description
%       This function returns the valid, invalid, fractional and
%       piled-up pattern numbers of the event file, together with the 
%       exposure and a possible flux entry (the argument is passed
%       directly into the struct). This information is based on the
%       event file keywords NVALID/NGRAD1 etc., written by SIXTE. 
%       
%       The pattern fraction errors are calculated assuming Poisson error.
%       Furthermore, it is possible to calculate (energy-)resolved
%       pattern fractions by giving the lo, hi, and field qualifiers.
%       Hereby, currently no distinction between non-piled up and
%       piled-up patterns is made (nvalids = TYPE>=0).
%       
%       If the given event file is an already loaded FITS file
%       (Struct_Type), the reading of the relevant header keywords is
%       skipped and only the energy/channel resolved patterns are
%       returned (requires lo, hi, and field qualifiers).
%\qualifiers{
%\qualifier{lo}{low grid for calculation of pattern fractions (in
%      unit of fields}
%\qualifier{hi}{high grid}
%\qualifier{field}{Name of event file column for filtering (for
%      example "signal" or "pha")}
%\qualifier{nphot}{Adds an additional struct field with number of
%      photons in the SIXTE event file}
%}
%\seealso{ratio_error_prop}
%!%-
  variable file, flux;
  switch(_NARGS)
  { case 2: (file, flux) = (); }
  { case 1: (file) = (); }
  { return help(_function_name()); }

  if (typeof(file) == Struct_Type){
    return calc_pattern_fractions_of_grid(file, qualifier("lo"), qualifier("hi"), qualifier("field"));
  }
  
  variable sr = fits_read_key_struct(file+"[1]", %{{{
				     "EXPOSURE",
				     "NVALID",
				     "NPVALID",
				     "NINVALID",
				     "NPINVALI",
				     "NGRAD0",
				     "NPGRA0",
				     "NGRAD1",
				     "NPGRA1",
				     "NGRAD2",
				     "NPGRA2",
				     "NGRAD3",
				     "NPGRA3",
				     "NGRAD4",
				     "NPGRA4",
				     "NGRAD5",
				     "NPGRA5",
				     "NGRAD6",
				     "NPGRA6",
				     "NGRAD7",
				     "NPGRA7",
				     "NGRAD8",
				     "NPGRA8",
				     "NGRAD9",
				     "NPGRA9",
				     "NGRAD10",
				     "NPGRA10",
				     "NGRAD11",
				     "NPGRA11",
				     "NGRAD12", 
				     "NPGRA12"
				    ); %}}}

  variable ret = struct {
    flux           = flux,
    exposure       = sr.exposure,
    nvalid         = sr.nvalid,
    nvalid_piled   = sr.npvalid,
    ninvalid       = sr.ninvalid,
    ninvalid_piled = sr.npinvali,
    nsing          = sr.ngrad0,
    nsing_piled    = sr.npgra0,
    ndoub          = sr.ngrad1+sr.ngrad2+sr.ngrad3+sr.ngrad4,
    ndoub_piled    = sr.npgra1+sr.npgra2+sr.npgra3+sr.npgra4,
    ntrip          = sr.ngrad5+sr.ngrad6+sr.ngrad7+sr.ngrad8,
    ntrip_piled    = sr.npgra5+sr.npgra6+sr.npgra7+sr.npgra8,
    nquad          = sr.ngrad9+sr.ngrad10+sr.ngrad11+sr.ngrad12,
    nquad_piled    = sr.npgra9+sr.npgra10+sr.npgra11+sr.npgra12,
    sfrac, sfrac_err, dfrac, dfrac_err, tfrac, tfrac_err, qfrac, qfrac_err
  };

  if (qualifier_exists("nphot")){
    ret = struct_combine(ret, struct{npho_sim=max(fits_read_table(file+"[1]").ph_id)});      
  }
  
  if (qualifier_exists("lo") and qualifier_exists("hi") and qualifier_exists("field")){
    % Calculate pattern fractions on a given (energy or channel) grid
    variable evt = fits_read_table(file);
    variable patfrac_ret = calc_pattern_fractions_of_grid(evt, qualifier("lo"), 
							  qualifier("hi"), qualifier("field"));
    ret = struct_combine(ret, patfrac_ret);
  }
  
  (ret.sfrac, ret.sfrac_err) = ratio_error_prop(1.*ret.nsing, 1.*ret.nvalid ; Poisson);
  (ret.dfrac, ret.dfrac_err) = ratio_error_prop(1.*ret.ndoub, 1.*ret.nvalid ; Poisson);
  (ret.tfrac, ret.tfrac_err) = ratio_error_prop(1.*ret.ntrip, 1.*ret.nvalid ; Poisson);
  (ret.qfrac, ret.qfrac_err) = ratio_error_prop(1.*ret.nquad, 1.*ret.nvalid ; Poisson);

  return ret;
}

%
% Code from src/sixte/get_sixte_xml_data.sl
%
% -*- mode: slang; mode: fold -*-

private define file2string( filename ){ %{{{
    % this function reads a textfile into a string and returns it
    variable fp = fopen(filename,"r");
    variable output = "";
    variable buffer;
    while(fgets(&buffer,fp) != -1){
          output += buffer;
    }
    () = fclose(fp);
    return(output);
}
%}}}
private define searchkey( text, element, attribute ){ %{{{
  % read an xml key, given the name of the element and attribute
  % this is still seriously limited - only do this if you know what you're doing
  % returns an empty string if nothing is found
  variable i_start,i_stop,pos,len;
  variable key = element+".*\n?[^>]*"R+attribute;
  variable expr = sprintf("\(%s=\"[^\"$]*\"\)"R,key);
  % what this expression does: first find the line where the element appears
  % then find whether the attribute appears in any of the following lines, but
  % still inside the tag (hence the [^/]*)
  % the value is then behind the attribute after an = sign and in quotes
  % with no further quotes or line-breaks inbetween
  if (string_match(text,expr)){
    (pos,len) = string_match_nth(1);
  } else {
    return(""); % nothing has been found
  }
  %i_stop is the last index before the closing quote character
  i_stop = pos + len -2;
  % count backwards until you find the opening quote character
  i_start = i_stop+1;
  do {
    i_start -= 1;
  } while (strncmp(text[[i_start-1]],"\"",1));
  %i_start = pos + strlen(key)+2;

  variable ret = text[[i_start:i_stop]];

  if( strlen(ret) == 0 ){
    vmessage(" ERROR(%s): XML has no attribute '%s' in element '%s'! Aborting ...",
	     _function_name, attribute, element
	    );
    exit;
  }
  return ret;
}
%}}}

define get_sixte_xml_data( xmlfile ){ %{{{
%!%+
%\function{get_sixte_xml_data}
%\synopsis{Loads a SIXTE XML file and outputs their attributes and values}
%\usage{Struct_Type xmlinfo = get_sixte_xml_data(String_Type filename);}
%\notes
%    This function currently only supports XMLs of the Athena WFI and
%    eROSITA missions.
%!%-
  
  %% check xml access
  if( access( xmlfile, F_OK )!= 0 ){
    vmessage(" ERROR(%s): XML '%s' is not accessible! Aborting ...",
	     _function_name, xmlfile
	    );
    exit;
  }
  variable xmldir  = path_dirname(xmlfile)+"/";
  variable xmltext = file2string( xmlfile );

  variable telescop = searchkey( xmltext, "instrument", "telescop" );
  variable instrume = searchkey( xmltext, "instrument", "instrume" );
  variable arffile  = searchkey( xmltext, "arf", "filename" );
  variable rmffile  = searchkey( xmltext, "rmf", "filename" );
  variable psffile  = searchkey( xmltext, "psf", "filename" );
  variable pirmffile= searchkey( xmltext, "pirmf", "filename" );
  variable pha2pifile= searchkey( xmltext, "pha2pi", "filename" );
  variable vigfile  = searchkey( xmltext, "vignetting", "filename" );
  variable enthresh = atof(searchkey( xmltext, "threshold_readout_lo_keV", "value" ));
  variable xwidth   = atoi(searchkey( xmltext, "dimensions", "xwidth" ));
  variable ywidth   = atoi(searchkey( xmltext, "dimensions", "ywidth" ));
  variable xdelt    = atof(searchkey( xmltext, "wcs", "xdelt" ));
  variable ydelt    = atof(searchkey( xmltext, "wcs", "ydelt" ));
  variable xrpix    = atof(searchkey( xmltext, "wcs", "xrpix" ));
  variable yrpix    = atof(searchkey( xmltext, "wcs", "yrpix" ));
  variable xrval    = atof(searchkey( xmltext, "wcs", "xrval" ));
  variable yrval    = atof(searchkey( xmltext, "wcs", "yrval" ));
  variable rota     = atof(searchkey( xmltext, "wcs", "rota" ));
  variable focallen = atof(searchkey( xmltext, "focallength", "value" ));
  variable fov      = atof(searchkey( xmltext, "fov", "diameter" ));

  
  variable xmldata = struct{
    xmldir  = xmldir,
    xmlfile = xmlfile,
    mission = telescop,
    instrument = instrume,
    arffile     = xmldir+"/"+arffile, % Filename of ARF (string)
    psffile     = xmldir+"/"+psffile, % Filename of PSF (string)
    vigfile     = xmldir+"/"+vigfile, % Filename of VIGNETTING (string)
    rmffile     = xmldir+"/"+rmffile, % Filename of nominal RMF (string)
    eboundsfile = xmldir+"/"+rmffile, % Filename of nominal RMF (string)
    pirmffile   = xmldir+"/"+pirmffile, % Filename of PI RMF (string)
    pha2pifile  = xmldir+"/"+pha2pifile, % Filename of Pha2Pi correction (string)
    loenthresh = enthresh, % Lower Energy Threshold [keV]
    xnumpix = xwidth,
    ynumpix = ywidth,
    xdelt   = xdelt,
    ydelt   = ydelt,
    xrpix   = xrpix,
    yrpix   = yrpix,
    xrval   = xrval,
    yrval   = yrval,
    rota    = rota,
    focallen=focallen,
    fov     = fov,
    
    %% derived quantities( ATTENTION X/Y WIDTH is changed to mm!!!)
    xwidth  = xwidth * xdelt * 1e3, % Detector width in x direction [mm]
    ywidth  = ywidth * ydelt * 1e3, % Detector width in y direction [mm]
    x0      = xwidth * xdelt * 1e3 * (-0.5), % Origin shift in x direction [mm]
    y0      = ywidth * ydelt * 1e3 * (-0.5), % Origin shift in y direction [mm]
  };
  
  return xmldata;
}
%}}}


%
% Code from src/misc/astronomy/get_source_counts.sl
%
define get_source_counts()
%!%+
%\function{get_source_counts}
%\synopsis{calculates source counts, background counts and background-subtracted source counts from given spectrum, with errors }
%\usage{Double_Type get_source_counts(Integer_Type id, Integer_Type backid, Double_Type emin, Double_Type emax);}
%\description
%    - spectrum and background have to be already loaded (id, backid)
%    - arf and rmf have to be already loaded and assigned
%\seealso{load_data,assign_rmf,assign_arf}
%!%-
{
  variable id,backid,emin,emax;
  switch(_NARGS)
  { case 4: (id,backid,emin,emax) = (); }
  { help(_function_name()); return; }

    variable cdat,cback,ilo,ihi,q,specArea,backArea,roc_val;

    variable totcts =Double_Type;
    variable speccts=Double_Type;
    variable backcts=Double_Type;
    variable totcts_err =Double_Type;
    variable speccts_err=Double_Type;
    variable backcts_err=Double_Type;
        
    cback   = _A(get_data_counts(backid)); % background data counts
    cdat    = _A(get_data_counts(id));   % spectrum data counts

    ilo     = where( cdat.bin_lo<emin ); % get counts within emin < cts < emax
    ihi     = where( cdat.bin_hi>emax );

    totcts       = sum( cdat.value[[ ilo[length(ilo)-1]:ihi[0] ]] ); % total counts
    totcts_err   = sqrt( sum((cdat.err[[ilo[length(ilo)-1]:ihi[0]]])^2) );
    backcts      = sum( cback.value[[ ilo[length(ilo)-1]:ihi[0]]] ); % background counts
    backcts_err  = sqrt( sum((cback.err[[ ilo[length(ilo)-1]:ihi[0]]])^2) );

    specArea     = get_data_backscale(id); % correct for different region sizes
    backArea     = get_data_backscale(backid);
    speccts      = totcts-(specArea/backArea)*backcts; % back. corr. spec. counts
    speccts_err  = sqrt(backcts_err^2+totcts_err^2);

    return (totcts,totcts_err,speccts,speccts_err,backcts,backcts_err);
}

%
% Code from src/slang/structs/get_struct_fields.sl
%
define get_struct_fields()
%!%+
%\function{get_struct_fields}
%\synopsis{returns several fields of a structure}
%\usage{(Any_Type val1, val2, ...) = get_struct_fields(Struct_Type s, String_Type fieldname1, fieldname2, ...);}
%\qualifiers{
%\qualifier{i}{indices used for filtering array fields}
%}
%\description
%    Each value corresponds to the according field of the structure.
%    If an \code{i} qualifier is given, array-typed field values
%    are filtered with these indices, i.e.,
%#v+
%       val = get_struct_feld(s, fieldname)[i];
%#v-
%    It also possible to create lists of structure field values,
%    by just passing lists (or, equivalently, arrays) as arguments.
%\examples
%#v+
%    variable table = struct { x=[1:5], y=[1:5]^2, err1=[1:5], err2=[1:5]*2 };
%
%    plot_with_err( get_struct_fields(table, "x", "y", "err1"; i=[1,0,4,2,3]) ; connect_points);
%    % equivalent to:
%    % plot_with_err( table.x[1,0,4,2,3], table.y[1,0,4,2,3], table.err1[1,0,4,2,3] ; connect_points);
%
%    plot_with_err( get_struct_fields(table, "x", "y", {"err1", "err2"}) );
%    % or (even less redundant):
%    plot_with_err( get_struct_fields(table, "x", "y", "err"+["1", "2"]) );
%    % equivalent to:
%    % plot_with_err( table.x, table.y, {table.err1, table.err2} );
%#v-
%\seealso{get_struct_field}
%!%-
{
  variable arg, args = __pop_list(_NARGS);
  variable s = list_pop(args);
  variable ind = qualifier("i");
  foreach arg (args)
  {
    variable field = NULL;
    switch(typeof(arg))
    { case List_Type:
	field = { get_struct_fields(s, __push_list(arg);; __qualifiers) };
    }
    { case Array_Type:
	field = { get_struct_fields(s, __push_array(arg);; __qualifiers) };
    }
    { case String_Type:
	field = get_struct_field(s, arg);
        if(ind!=NULL && typeof(field)==Array_Type)
	  field = field[ind];
    }
    { % else:
        vmessage("warning (%s): cannot process %S argument", _function_name(), typeof(arg));
    }
    field;  % left on stack
  }
}

%
% Code from src/fitting/fit-functions/simple_gpile/get_unpiled_data_flux.sl
%

define get_unpiled_data_flux()
{
  % process arguments
  variable id;
  switch(_NARGS)
  { case 1:  id = (); }
  { message("usage: get_unpiled_data_flux(id);"); return; }

  variable f = get_data_flux(id);
  Isis_Active_Dataset = id;
  variable pm = eval_fun(f.bin_lo, f.bin_hi);
  variable fitFun = get_fit_fun();
  fit_fun(get_unpiled_fit_fun());
  Isis_Active_Dataset = id;
  variable  m = eval_fun(f.bin_lo, f.bin_hi);
  fit_fun(fitFun);
  f.value *= m/pm;
  f.err *= m/pm;
  return f;
}

%
% Code from src/fitting/fit-functions/simple_gpile/get_unpiled_fit_fun.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_unpiled_fit_fun()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_unpiled_fit_fun}
%\synopsis{return the currently defined fit function without simple_gpile*}
%\usage{String_Type get_unpiled_fit_fun()}
%\description
%    \code{simple_gpile*(Isis_Active_Dataset, } is replaced by \code{(}.
%\seealso{get_fit_fun}
%!%-
{
  variable fitFun = get_fit_fun();
  variable m = string_matches(fitFun, `^\(.*\)simple_gpile\d* *( *Isis_Active_Dataset, *\(.*\)$`);
  if(m!=NULL)  fitFun = m[1] + "(" + m[2];
  return fitFun;
}

%
% Code from src/slang/get_variable_name.sl
%
%%%%%%%%%%%%%%%
define get_variable_name()
%%%%%%%%%%%%%%%
%!%+
%\function{get_variable_name}
%\synopsis{returns the namespace and name of a given reference}
%\usage{(namespace, name) = get_variable_name(&var);}
%\description
%    The name of the variable or function the given
%    reference points to will be returned as a string.
%    The namespace where it is defined is returned as
%    well. In case of a private namespace, however,
%    NULL will be returned instead.
%\seealso{current_namespace}
%!%-
{
  variable ref;
  switch (_NARGS)
    { case 1: ref = (); }
    { help(_function_name); return; }

  if (typeof(ref) != Ref_Type) { vmessage("error: given argument has to be a reference"); return; }

  % convert reference to variable name (eventually including the namespace)
  variable name = string_matches(string(ref), "^&\(.*\)$"R);
  if (name == NULL) { return (NULL, NULL); }
  else { name = name[1]; }
  % extract the namespace (may be undefined -> private)
  variable match = string_matches(name, "^\(.*\)->\(.*\)$"R);
  
  if (match == NULL) { return (NULL, name); }
  else { return (match[1], match[2]); } 
}

%
% Code from src/data/spectrum.datastructures/get_warmabs_model.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define get_warmabs_model()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_warmabs_model}
%\synopsis{computes the contribution of separate elements in a warmabs-model}
%\usage{Struct_Type data = get_warmabs_model();}
%!%-
{
  variable fitFun = get_fit_fun;
  fit_fun("warmabs(1)");

  variable atoms = ["c", "n", "o", "ne", "mg", "si", "s", "ar", "ca", "fe"];
  variable data = @Struct_Type(["bin_lo", "bin_hi", "warmabs", "warmabs_" + atoms]);
  data.bin_lo = [1:20:0.002];
  data.bin_hi = make_hi_grid(data.bin_lo);
  data.warmabs = eval_fun(data.bin_lo, data.bin_hi);

  variable abund = @Struct_Type(atoms);
  variable atom;
  foreach atom (atoms)
  { set_struct_field(abund, atom, get_par("warmabs*." + atom + "abund")[0]);
  }

  foreach atom (atoms)
  { set_par("warmabs*.*abund", 0);
    set_par("warmabs*."+atom+"abund", get_struct_field(abund, atom));
    set_struct_field(data, "warmabs_"+atom, eval_fun(data.bin_lo, data.bin_hi));
  }

  foreach atom (atoms)
  { set_par("warmabs*."+atom+"abund", get_struct_field(abund, atom));
  }

  fit_fun(fitFun);
  return data;
}

%
% Code from src/slang/physics/gravitational_radius.sl
%
%!%+
%\function{gravitational_radius}
%\synopsis{calculates the gravitational radius in Meters defined as r_g=GM/c^,
% for a mass given in units of M_sol.}
%\usage{gravitational_radius(mass_in_solar)}
%!%-
define gravitational_radius() {
   variable mass_in_solar;
   
   switch(_NARGS)
   { case 1: mass_in_solar = ();}
   { help(_function_name()); return; }

   variable G = 6.673e-11;
   variable c = 299792458.0;
   variable Msol = 1.988e30;  % in kg
   
   return    G*Msol*mass_in_solar / (c^2);
}

%
% Code from src/slang/math/greatcircle_coordinates.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define greatcircle_coordinates()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{greatcircle_coordinates}
%\synopsis{calculates the coordinates of the greatcircle between two points on a sphere}
%\usage{(Double_Type lambda[], phi[]) = greatcircle_coordinates(lambda1, phi1, lambda2, phi2);}
%\qualifiers{
%\qualifier{unit}{[\code{="deg"}] unit of the angular coordinates}
%\qualifier{delta}{[\code{=0.5}] angular step in degrees}
%}
%\description
%    (\code{lambda}i, \code{phi}i) are the spherical coordinates of point i.\n
%    \code{unit="deg"}: \code{lambda}i and \code{phi}i are in degrees.\n
%                They can be scalar values or arrays of the form
%                \code{[deg, arcmin]}, \code{[deg, arcmin, arcsec]} or \code{[sign, deg, arcmin, arcsec]}.
%    \code{unit="rad"}: \code{lambda}i and \code{phi}i are scalars in radian.\n
%    \code{unit="hms"}: code{lambda}i are scalars hour angles (24h = 360deg)
%                or arrays in h:m:s format, i.e., \code{[h, m]} or \code{[h, m, s]}.
%                The \code{phi}i are nevertheless in degrees as above.
%\seealso{greatcircle_distance}
%!%-
{
  variable l1, p1, l2, p2;
  switch(_NARGS)
  { case 4: (l1, p1, l2, p2) = (); }
  { help(_function_name()); return; }
  
  variable q = __qualifiers();
  (l1, l2) = angle_to_rad(l1, l2;; q);
  if(qualifier("unit", "deg")=="hms")  q.unit = "deg";  % if RA is in HMS, delta is still in deg
  (p1, p2) = angle_to_rad(p1, p2;; q);

  variable x = cos(l2)*cos(p2);
  variable y = sin(l2)*cos(p2);
  variable z = sin(p2);
  variable cos_l1 = cos(l1);
  variable sin_l1 = sin(l1);
  (x, y) = (cos_l1*x+sin_l1*y, cos_l1*y-sin_l1*x);  % Rz(-l1)
  variable cos_p1 = cos(p1);
  variable sin_p1 = sin(p1);
  (x, z) = (cos_p1*x+sin_p1*z, cos_p1*z-sin_p1*x);  % Ry(-p1)
  variable inv_sqrt = 1./sqrt(y^2+z^2);
  variable cos_d = z*inv_sqrt;
  variable sin_d = y*inv_sqrt;
  (   z) = (  cos_d*z+sin_d*y);  % Rx(d)
  variable phi = atan2(z, x);
  phi = [0:phi:#(1+int(phi/(qualifier("delta", 0.5)*PI/180)))];
  (x, y, z) = (cos(phi), 0*phi, sin(phi));
  (y, z) = (cos_d*y+sin_d*z, cos_d*z-sin_d*y);  % Rx(-d)
  (x, z) = (cos_p1*x-sin_p1*z, cos_p1*z+sin_p1*x);  % Ry(p1)
  (x, y) = (cos_l1*x-sin_l1*y, cos_l1*y+sin_l1*x);  % Rz(l1)
  variable l = atan2(y, x);
  variable b = asin(z);
  
  switch(qualifier("unit", "deg"))
  { case "deg":  l *= 180/PI;  b *= 180/PI; }
  { case "HMS":  l *= 15/PI;   b *= 180/PI; }
  return (l, b);
}

%
% Code from src/slang/math/greatcircle_distance.sl
%

define greatcircle_distance()
%!%+
%\function{greatcircle_distance}
%\synopsis{calculates the angular distance between two points on a sphere in radians}
%\usage{Double_Type greatcircle_distance(alpha1, delta1, alpha2, delta2)}
%\qualifiers{
%\qualifier{unit}{[\code{="deg"}] unit of the input angular coordinates}
%\qualifier{alpha1_unit}{[\code{="deg"}] unit of the alpha1}
%\qualifier{alpha2_unit}{[\code{="deg"}] unit of the alpha2}
%\qualifier{delta1_unit}{[\code{="deg"}] unit of the delta1}
%\qualifier{delta2_unit}{[\code{="deg"}] unit of the delta2}
%}
%\description
%
%    DEPRECATED - please use angular_separation instead
%    (\code{alpha}i, \code{delta}i) are the spherical coordinates of point i.\n
%    \code{unit="deg"}: \code{alpha}i and \code{delta}i are in degrees.\n
%                They can be scalar values or arrays of the form
%                \code{[deg, arcmin]}, \code{[deg, arcmin, arcsec]} or \code{[sign, deg, arcmin, arcsec]}.
%    \code{unit="rad"}: \code{alpha}i and \code{delta}i are scalars in radian.\n
%    \code{unit="hms"}: \code{alpha}i are scalars hour angles (24h = 360deg)
%                or arrays in h:m:s format, i.e., \code{[h, m]} or \code{[h, m, s]}.
%                The \code{delta}i are in degrees as above.
%                The units of each coordinate can be set independently.
%    Note that independent of the unit setting, the returned great circle
%    distance will always be in radian.
%\seealso{greatcircle_coordinates}
%!%-
{
  variable x1, y1, x2, y2;
  switch(_NARGS)
  { case 4: (x1, y1, x2, y2) = (); }
  { help(_function_name()); return; }

  variable q = __qualifiers();
  variable new_q = struct{ unit = qualifier("alpha1_unit",qualifier("unit","deg")) };
  x1 = angle_to_rad(x1;; new_q);
  new_q.unit = qualifier("alpha2_unit",qualifier("unit","deg"));
  x2 = angle_to_rad(x2;; new_q);
  if(qualifier("unit", "deg")=="hms")  q.unit = "deg";  % if RA is in HMS, delta is still in deg
  new_q.unit = qualifier("delta1_unit",qualifier("unit","deg"));
  y1 = angle_to_rad(y1;; new_q);
  new_q.unit = qualifier("delta2_unit",qualifier("unit","deg"));
  y2 = angle_to_rad(y2;; new_q);
  
  variable Dx = x2 - x1;
  % return acos( cos(Dx)*cos(y1)*cos(y2) + sin(y1)*sin(y2) );
  % This arccosine formula above can have large rounding errors for the common case where the distance is small
  variable c = cos(Dx);
  variable s1,c1,s2,c2;
  (s1,c1)=sincos(y1);
  (s2,c2)=sincos(y2);
  return atan2(sqrt((c2*sin(Dx))^2 + (c1*s2 - s1*c2*c)^2), s1*s2 + c1*c2*c);
}


%
% Code from src/slang/geometry/greiner_hormann.sl
%

define line_intersect() 
%!%+
%\function{line_intersect}
%\synopsis{Calculate the intersection of two lines defined by
%          two points each}
%\usage{res=line_intersect(P0,P1,P2,P3,&alphaP,&alphaQ);}
%\altusage{res=line_intersect(P0x,P0y,P1x,P1y,P2x,P2y,P3x,P3y,&alphaP,&alphaQ);}
%\description
% This function implements a fast line intersection algorithm
% based on 
% https://stackoverflow.com/questions/563198/how-do-you-detect-where-two-line-segments-intersect
% between the line segments P=P0--P1 and Q=P2--P3, where the points Pi 
% are defined as structures of the type struct {x,y} or through 
% coordinates P0x,P0y, etc.
%
% The function returns 1 if an intersection exists and 0 if not.
%
% If an intersection exists, the function returns the relative position
% of the intersection point on the segments in variables alphaP and alphaQ.
%!%-
{
    variable P0,P1,P2,P3,alphaP,alphaQ;

    if (_NARGS==6) {
	(P0,P1,P2,P3,alphaP,alphaQ)=();
    } else {
	if ( _NARGS==10 ) {
	    variable P0x,P0y,P1x,P1y,P2x,P2y,P3x,P3y;
	    (P0x,P0y,P1x,P1y,P2x,P2y,P3x,P3y,alphaP,alphaQ)=();
	    P0=struct{x=P0x,y=P0y};
	    P1=struct{x=P1x,y=P1y};
	    P2=struct{x=P2x,y=P2y};
	    P3=struct{x=P3x,y=P3y};
	} else {
	    help(_function_name());
	}
    }
    
    variable S10x=P1.x-P0.x;
    variable S10y=P1.y-P0.y;
    variable S32x=P3.x-P2.x;
    variable S32y=P3.y-P2.y;

    variable denom=S10x*S32y-S32x*S10y;
    if (denom==0.) {
	return 0; % colinear - we define segments to be not intersecting in this case
    }

    variable denomPositive=(denom>0.);

    variable S02x=P0.x-P2.x;
    variable S02y=P0.y-P2.y;

    variable snumer=S10x*S02y-S10y*S02x;

    if ((snumer<0)==denomPositive) {
	return 0; % no intersection
    }

    variable tnumer=S32x*S02y-S32y*S02x;
    if ((tnumer<0.)==denomPositive) {
	return 0; % no intersection
    }

    if (((snumer>denom)==denomPositive) or ((tnumer>denom)==denomPositive)) {
	return 0; % no intersection
    }

    @alphaP=tnumer/denom;
    @alphaQ=snumer/denom;
    return 1;
}

private define new_greiner_vertex() {
    %
    % private function: define a new vertex point for the
    % Greiner and Hormann algorithm
    %
    variable x,y,id,alpha,inter;
    inter=0;
    alpha=0.;
    if (_NARGS==3) {
	(x,y,id)=();
    }
    if (_NARGS==4) {
	(x,y,id,alpha)=();
	inter=1;
    }
    
    return struct {
        x=x,       % x-coordinate
	y=y,       % y-coordinate
        id=id,     % identifier
	next=NULL, % pointer to next vertex point
	prev=NULL, % pointer to current vertex point
	intersect=inter, % 0 not an intersect, 1 intersect
	neighbor=NULL, % pointer to this point in other array
	alpha=alpha, % storage for distance from last non-intersection point
	entry=-1,  % 1 if entry point, 0 if not, -1 undefined
	processed=0 % true if vertex has been taken care of
    };
}

private define new_greiner_vertexlist(P) {
    % build up a vertex list for the Greiner-Hormann algorithm
    variable i;
    variable sstart=new_greiner_vertex(P.x[0],P.y[0], P.id[0]);
    variable ssold=sstart;
    _for i(1,length(P.x)-1,1) {
	variable ssnew=new_greiner_vertex(P.x[i],P.y[i], P.id[i]);
	ssnew.prev=ssold;
	ssold.next=ssnew;
	ssold=ssnew;
    }
    sstart.prev=ssold;
    ssnew.next=sstart;
    return sstart;
}

private define greiner_insert_between(P,Q,ins) {
    % insert vertex np between non-intersecting points P and Q
    % nb: does not check that P, Q are non intersection points
    % for speed reasons

    % np must be an intersection vertex with alpha set
    
    % find correct insertion segment
    variable pp=P;
    while (pp.next != Q && pp.next.alpha<ins.alpha) {
	pp=pp.next;
    }
    
    ins.prev=pp;
    ins.next=pp.next;
    ins.next.prev=ins;
    pp.next=ins;
}

define greiner_hormann()
%!%+
%\function{greiner_hormann}
%\synopsis{Clipping and logical intersection of two polygons}
%\usage{List_Type polylist= greiner_hormann(src, clp);}
%\altusage{List_Type polylist= greiner_hormann(sx,sy,cx,cy);}
%\qualifiers{
%\qualifier{intersection}{return the intersection of src and clp
%    (i.e., clip src against clp), the default}
%\qualifier{union}{return the union of src and clp}
%\qualifier{without}{remove clp from src}
%\qualifier{perturb}{slightly perturb src to reduce probability of
%    failure of the algorithm (see description below)}
%}
%\description
% This function implements the Greiner-Hormann algorithm for polygon
% intersections (Greiner & Hormann, 1998, ACM Trans. Graph. 17(2), 71-83).
% The polygons are given as structures struct {x=[], y=[]}, other structure
% tags are ignored. The function returns a list of closed polygons of this
% type (which may be empty!). The returned polygons also include a tag
% id which is 0 if the point originates in src, 1 if the point originates in
% clp, and -1 if this is a newly inserted intersection point.
%
% Alternatively, the x- and y-coordinates of the polygon points can be
% given. In this case the return will still be a list of structs.
%
% The polygons must be closed, i.e., src.x[0]==src.x[-1], src.y[0]=src.y[-1],
% and the same for clp. The polygons can self-intersect, there is almost no
% limitation on their shape. In the case of intersection, the determination 
% that a point is inside a polygon is done using the winding number
% (see help for function point_in_polygon).
%
% The Greiner-Horman algorithm has an issue for polygons that have colinear
% overlapping sides. If the qualifier "perturb" is set, the coordinates
% in src are randomly perturbed at a level of 1e-8 to reduce the probability
% of this happening.
% 
%\example
%
% variable p=xfig_plot_new(15,15);
% p.world(-1.,1.,-1.,1.);
%
% variable src,clp;
%
% % a complex polygon
% src=struct {x=[-0.25,0.00,0.25,0.3,0.8,0.5,-0.25],
%             y=[ 0.80,-0.40,0.8,-0.4,0.0,0.8,0.80]};
%
% % a square
% clp=struct {x=[-0.5,0.5,0.5,-0.5,-0.5]+0.2,
%              y=[-0.5,-0.5,0.5,0.5,-0.5]+0.2};
%
% p.plot(src.x,src.y;color="blue",depth=150);
% p.plot(clp.x,clp.y;color="green",depth=150);
%
% variable res=greiner_hormann(src,clp;intersection);
%
% variable i;
% _for i(0,length(res)-1,1) {
%     p.plot(res[i].x,res[i].y;color="red",depth=50);
% }
%
% p.render("polygons.pdf");
%\seealso{point_in_polygon}
%!%-
{
    if (_NARGS != 2 or _NARGS!=4) {
	help(_function_name());
    }

    variable uni=qualifier_exists("union");
    variable inter=qualifier_exists("intersection");
    variable without=qualifier_exists("without");
    variable sm=uni+inter+without;
    
    if (sm>1) {
	throw UsageError,sprintf("Usage error in %s: only one qualifier of union, intersection, and without can be set!\n", _function_name());
    }

    % default: clip S with C (i.e., get the intersection of the polygons)
    if (sm==0) {
	inter=1;
    }

    % build up vertex list of the subject and the clip polygons
    variable S,C;
    if (_NARGS==2) {
        variable Stmp,Ctmp;
	(Stmp,Ctmp)=();
        S=struct { x=@Stmp.x, y=@Stmp.y, id=Char_Type[length(Stmp.x)]};
        C=struct { x=@Ctmp.x, y=@Ctmp.y, id=1+Char_Type[length(Ctmp.x)]};
    } else {
	variable sx,sy,cx,cy;
	(sx,sy,cx,cy)=();
	S=struct { x=sx[*], y=sy[*], id=Char_Type[length(sx)] };
	C=struct { x=cx[*], y=cy[*], id=1+Char_Type[length(cx)] };
    }

    if (S.x[0]!=S.x[-1] || S.y[0]!=S.y[-1] ) {
	throw UsageError,sprintf("Usage error in %s: Source polygon is not closed!\n", _function_name());
    }

    if (C.x[0]!=C.x[-1] || C.y[0]!=C.y[-1] ) {
	throw UsageError,sprintf("Usage error in %s: Clip polygon is not closed!\n", _function_name());
    }

    if (length(S.x)!=length(S.y)) {
        throw UsageError,sprintf("Usage error in %s: Source x and y arrays must be of equal length\n",_function_name());
    }

    if (length(C.x)!=length(C.y)) {
        throw UsageError,sprintf("Usage error in %s: Clip x and y arrays must be of equal length\n",_function_name());
    }

    if (qualifier_exists("perturb")) {
        S.x+=1e-8*urand(length(S.x));
        S.y+=1e-8*urand(length(S.x));

        S.x[-1]=S.x[0];
        S.y[-1]=S.y[0];
    }

    
    variable src=new_greiner_vertexlist(S);
    variable clp=new_greiner_vertexlist(C);

    % we'll be returning a list of polygons
    variable polylist={};

    %
    % Step 1: find all intersection points between the polygons and
    % insert them into the polygons
    %
    variable numinter=0;
    variable ss=src;
    do {
	% find next vortex in src that is not an intersection
	variable snxt=ss;
	do {
	    snxt=snxt.next;
	} while (snxt.intersect==1);

	variable cc=clp;
        do {
	    % find next vortex that is not an intersection
	    variable cnxt=cc;
	    do {
		cnxt=cnxt.next;
	    } while (cnxt.intersect==1);

	    % Is there an intersection?
	    variable alphaP,alphaQ;
	    if (line_intersect(ss,snxt,cc,cnxt,&alphaP,&alphaQ)==1) {
		numinter++;
		% insert new intersection point into both polygons
		variable newx=ss.x+alphaP*(snxt.x-ss.x);
		variable newy=ss.y+alphaP*(snxt.y-ss.y);

		variable np1=new_greiner_vertex(newx,newy,-1,alphaP);
		variable np2=new_greiner_vertex(newx,newy,-1,alphaQ);
		np1.neighbor=np2;
		np2.neighbor=np1;
		greiner_insert_between(ss,snxt,np1);
		greiner_insert_between(cc,cnxt,np2);
	    }
	    cc=cnxt;
	} while (cc!=clp);
	ss=snxt;

    } while (ss!=src);
        
    %
    % Deal with the case that no intersections were found
    %
    if (numinter==0) {
	% is S inside C or outside?
	variable insi=point_in_polygon(S.x[0],S.y[0],C.x,C.y);

	if (insi) {
	    % S is inside C
	    if (uni) {
		list_append(polylist,struct{x=C.x[*],y=C.y[*],id=C.id[*]});
	    }

	    % S without C gives empty list
	    if (inter) {
		list_append(polylist,struct{x=S.x[*],y=S.y[*],id=C.id[*]});
	    }
	} else {
	    % S is outside C

	    % union and without C 
	    if (uni or without) {
		list_append(polylist,struct{x=S.x[*],y=S.y[*],id=S.id[*]});
	    }

	    % intersection gives empty list
	}
	return polylist;
    }
		
    %
    % Phase 2: identify the entry and exit points;
    %   depending on the final operation desired, the
    %   initialization is done in a slightly different way
    %	

    % these different starting values will result in our three
    % cases
    variable sentry,centry;
    if (uni) {
	sentry=0;
	centry=0;
    }
    if (inter) {
	sentry=1;
	centry=1;
    }
    if (without) {
	sentry=0;
	centry=1;
    }

    % is starting point for ss inside of C?
    ss=src;
    variable entry=sentry xor point_in_polygon(ss.x,ss.y,C.x,C.y);	
    while (ss.next!=src) {
	if (ss.intersect) {
	    ss.entry=entry;
	    entry=not entry;
	}
	ss=ss.next;
    }

    % is starting point for cc inside of S?
    cc=clp;
    entry=centry xor point_in_polygon(cc.x,cc.y,S.x,S.y);	
    while (cc.next!=clp) {
	if (cc.intersect) {
	    cc.entry=entry;
	    entry=not entry;
	}
	cc=cc.next;
    }
    
    %
    % Step 3: Go through polygons and build up the new structure
    %
    variable unprocessed=1;
    do {
	
	%
	% go to the first unchecked intersection in the list
	%
	variable xx={};
	variable yy={};
        variable id={};
	
	cc=src;
	while ( (cc.intersect==0 || cc.processed==1) && cc.next!=src) {
	    cc=cc.next;
	}
	do {
	    cc.processed=1;
	    if (cc.entry==1) {
		do {
		    list_append(xx,cc.x);
		    list_append(yy,cc.y);
                    list_append(id,cc.id);
		    cc.processed=1;
		    cc=cc.next;
		} while (cc.intersect==0);
	    } else {
		do {
		    list_append(xx,cc.x);
		    list_append(yy,cc.y);
                    list_append(id,cc.id);
		    cc.processed=1;
		    cc=cc.prev;
		} while (cc.intersect==0);
	    }
	    cc.processed=1;
	    cc=cc.neighbor;
	} while (cc.processed==0);

        % save the current polygon on the output list
        xx=list_to_array(xx);
        yy=list_to_array(yy);
        id=list_to_array(id);    
	list_append(polylist,struct{x=[xx,xx[0]],y=[yy,yy[0]],id=[id,id[0]]});

	cc=src;
	unprocessed=0;
	while (cc.next!=src) {
	    if (cc.intersect==1 && cc.processed==0) {
		unprocessed++;
	    }
	    cc=cc.next;
	}
	    
	% is there still an unprocessed intersection in the list?
	cc=src;
	unprocessed=0;
	while (unprocessed==0 && cc.next!=src) {
	    if (cc.intersect==1 && cc.processed==0) {
		unprocessed=1;
	    }
	    cc=cc.next;
	}
    } while (unprocessed==1);
    
    return polylist;
}

%
% Code from src/slang/math/gridmapping.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%
public define gridmapping()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{gridmapping}
%\synopsis{computes a table of a 2d mapping}
%\usage{Struct_Type data = gridmapping(&getxy, X, Xfine, Y, Yfine);}
%!%-
{
  variable getxy, X, Xfine, Y, Yfine;
  switch(_NARGS)
  { case 3: (getxy, X, Y) = (); Xfine = X; Yfine = Y; }
  { case 5: (getxy, X, Xfine, Y, Yfine) = (); }
  { help(_function_name()); return; }

  variable n = length(X) * length(Yfine) + length(Y) * length(Xfine);
  variable dat = struct { x=Double_Type[n], y=Double_Type[n], x_=Double_Type[n], y_=Double_Type[n] };
  variable i = 0, x, y;
  foreach x (X)
    foreach y (Yfine)
    {
      (dat.x[i], dat.y[i], dat.x_[i], dat.y_[i]) = (x, y, @getxy(x, y));
      i++;
    }
  foreach y (Y)
    foreach x (Xfine)
    {
      (dat.x[i], dat.y[i], dat.x_[i], dat.y_[i]) = (x, y, @getxy(x, y));
      i++;
    }
  return dat;
}

%
% Code from src/data/spectrum.datastructures/group_data_keV.sl
%
%%%%%%%%%%%%%%%%%%%%%
define group_data_keV(id, deltaE)
%%%%%%%%%%%%%%%%%%%%%
{
  group_data(id, 1);
  variable d = get_data_counts(id);
  variable E = _A(1) / ((d.bin_lo+d.bin_hi)/2);
  rebin_data(id,  int( (-1)^int(E/deltaE) ) );
}

%
% Code from src/data/spectrum.datastructures/group_noticed_data.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%
define group_noticed_data()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{group_noticed_data}
%\synopsis{groups previously noticed spectral bins by an integer factor}
%\usage{group_noticed_data(Integer_Type id, Integer_Type factor);}
%\seealso{group_data}
%!%-
{
  variable ids, factor;
  switch(_NARGS)
  { case 2: (ids, factor) = (); }
  { help(_function_name()); }

  variable id;
  foreach id ([ids])
  {
    variable notice = get_data_info(id).notice;
    variable len = length(notice);
    variable ind = @notice;
    variable new_ignore_list = Integer_Type[0];
    variable i, new_i=0, nr=0, sgn=+1;
    _for i (0, len-1, 1)
    {
%    ()=printf("notice(%d) = %2d", i, notice[i]);
      if(notice[i]==0)
      { if(i>0 && notice[i-1]!=0) { sgn *= -1; new_i++; new_ignore_list = [new_ignore_list, new_i]; }
        if(i==0)  new_ignore_list = [new_ignore_list, new_i];
        ind[i] = sgn;
%      ()=printf(" => %2d [%d]\n", ind[i], new_i);
      }
      else
      { if(i>0 && notice[i-1]==0) { nr=0; sgn *= -1; new_i++; }
        nr++;
        ind[i] = sgn;
%      ()=printf(" => %2d [%d]\n", ind[i], new_i);
        if(nr==factor && i+1<len && notice[i+1]!=0) { nr=0; sgn *= -1; new_i++; }
      }
    }
%  print(new_ignore_list);

    rebin_data(id, ind);
    ignore_list(id, new_ignore_list);
  }
}

%
% Code from src/xspec/group_pha.sl
%
define group_pha (id)
%!%+
%\function{group_pha}
%\synopsis{Apply quality and grouping information from pha file}
%\usage{group_pha(indx)}
%\seealso{load_pha}
%!%-
{
  variable file = get_data_info(id).file;

  if (NULL == file)
    throw IsisError, sprintf("Warning: data set %d not found", id);

  variable fp = fits_open_file(file, "r");

  if (NULL == fp || _fits_movnam_hdu(fp, _FITS_BINARY_TBL, "SPECTRUM", 0))
    throw ReadError, sprintf("Failed opening file for reading: %s", file);

  variable qual, grp, colnum;
  
  ifnot (_fits_get_colnum(fp, "QUALITY", &colnum))
    qual = not reverse(fits_read_col(fp, colnum));
  else
    qual = 1;

  ifnot (_fits_get_colnum(fp, "GROUPING", &colnum))
    grp = x2i_group(fits_read_col(fp, colnum));
  else
    grp = 0;

  rebin_data(id, grp*qual);
  fits_close_file(fp);
} 

define load_pha (file)
%!%+
%\function{load_pha}
%\synopsis{Load pha file with grouping and quality information}
%\usage{load_pha("pha_filename")}
%\seealso{group_pha}
%!%-
{
  variable id = load_data(file);

  if (id < 1) return id; % error in load data

  try {
    group_pha(id);
  } catch IsisError, ReadError, sprintf("Failed opening file for reading: %s", file);

  return id;
}

%
% Code from src/fitting/fit_2d/haPSFcos.sl
%
%%%%%%%%%%%%%%%%%%%
define haPSFcos_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%%%%
{
  variable x0 = par[0];
  variable y0 = par[1];
  variable prefactor = par[2];
  variable c  = par[3];
  variable phase = par[4];

  variable XX, YY;
  (XX, YY) = get_2d_data_grid();

  variable phi = atan2(YY-y0, XX-x0);
  return prefactor * cos(c*phi + phase);
}

add_slang_function("haPSFcos", ["x0", "y0", "prefactor", "c", "phase"]);
% set_param_default_hook("haPSFcos", "haPSFcos_defaults");

%
% Code from src/misc/astronomy/hardnessratio.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL
% don't ask me why but it wont work otherwise. The machine knows! - M. Scott
%
%
%
define behr()
%!%+
%\function{behr}
%\synopsis{calculates fractional difference hardness ratio thorugh bayesian estimation}
%\usage{Double_Type (HR , HR_err_max , HR_err_min) =  behr(Double_Type soft_count, Double_Type hard_count);}
%\qualifiers{
%  \qualifier{scale_s}{Scaling factor for soft_count;  soft_count = scale_s*soft_count  (default = 1)}
%  \qualifier{scale_h}{Scaling factor for hard_count;  hard_count = scale_h*hard_count  (default = 1)}
%  \qualifier{back_s}{background counts in the soft band}
%  \qualifier{back_h}{background counts in the hard band}
%  \qualifier{bkg_scale_s}{Scaling factor for source background to observed background counts in the soft band.
%             Expected background in source region is ~ back_s/bkg_scale_s.}
%  \qualifier{bkg_scale_h}{Scaling factor for source background to observed background counts in the hard band.
%             Expected background in source region is ~ back_h/bkg_scale_h.}
%  \qualifier{aprox_n}{= run aproximation which keeps powers of bkg_scale_h^i*bkg_scale_s^j, with i+j <= n}
%  \qualifier{significance}{significance of calculated uncertainties in percent (default = 68 per cent)}
%  \qualifier{fulloutput}{if present function will additionally return the Grid used (2000 color values ranging from -0.999 to 0.999), and the
%             normalized probability distribution}
%}
%\description
%  This function gives a bayesian estimation for the so-called fractional difference hardness ratio according to (H-S)/(H+S),
%  assuming that soft_count and hard_count are the *intrinsic* counts.
%  Optionally, scaling factors for the indiviudal channels can be set.
%  If given, background counts are taken into account presuming a separate measurement of a Poisson background.
%  If the background counts are given a scaling factor for the source background has to be provided as well.
%  An approximative approach can be used as well, by only keeping up to the n-th power in the expansion in terms of
%  the scaling factor of the source background.
%
%  The hardness ratio estimation and the respecitive uncertainties are returned by default. 
%  Use qualifier if also the grid used (2000 color values ranging from -0.999 to 0.999), and the
%  normalized probability distribution are desired.
%
%  Based on Park et al., 2006, ApJ, 652, 610. Based on original isis-code by Mike Nowak.
%
%\seealso{hardnessratio;}
%!%-
{   
    variable A , B;
    switch(_NARGS)
    { case 2: (B , A) = (); }
    { help(_function_name()); return; }

    if (typeof(A) == Double_Type){
        A = [A];
                                 }
    if (typeof(B) == Double_Type){
        B = [B];
                                 }                             
 
    variable setlength = length(A);                             

    variable _EA = qualifier("scale_h" , 1.);
    variable _EB = qualifier("scale_s" , 1.);
    variable EA = Double_Type[setlength];
    variable EB = Double_Type[setlength];
    if (typeof(_EA) == Double_Type){
        EA[*] = _EA;
                                  }
    else{EA = _EA;}                              
    if (typeof(_EB) == Double_Type){
        EB[*] = _EB;
                                   }
    else{EB = _EB;}                               
              
    variable _DA = qualifier("back_h" , 0.0);
    variable _DB = qualifier("back_s" , 0.0);
    variable DA = Double_Type[setlength];
    variable DB = Double_Type[setlength];
    if (typeof(_DA) == Double_Type){
        DA[*] = _DA;
                                   }
    else{DA = _DA;}                               
    if (typeof(_DB) == Double_Type){
        DB[*] = _DB;
                                   }
    else{DB = _DB;}                               

    variable _GA = qualifier("bkg_scale_h" , 0.0);
    variable _GB = qualifier("bkg_scale_s" , 0.0);
    variable GA = Double_Type[setlength];
    variable GB = Double_Type[setlength];
    if (typeof(_GA) == Double_Type){
        GA[*] = _GA;
                                   }
    else{GA = _GA;}                                  
    if (typeof(_GB) == Double_Type){
        GB[*] = _GB;
                                   }
    else{GB = _GB;}

    variable n = qualifier("aprox" , 0.0);
    n = int(abs(n));

    variable pval = qualifier("significance" , 68.);
    pval = pval/100.;

    variable xx;

    variable grids = {};
    variable dists = {};

    variable fin_HRs = Double_Type[setlength];
    variable fin_HRs_up = Double_Type[setlength];
    variable fin_HRs_down = Double_Type[setlength];

    _for xx(0 , length(A) - 1 , 1){ 

        variable a = A[xx];
        variable b = B[xx];
        variable ea = EA[xx];
        variable eb = EB[xx];
        if (ea <= 0){ea = 1;}
        if (eb <= 0){eb = 1;}
        variable da = DA[xx];
        variable db = DB[xx];
        variable ga = GA[xx];
        variable gb = GB[xx];

        variable ngrid = int(2000*30);
        variable bhr_lo, bhr_hi;
        (bhr_lo,bhr_hi) = linear_grid(-1,1,ngrid);
        variable bhr = (bhr_lo+bhr_hi)/2.;
        variable dbhr = (bhr_hi-bhr_lo);
    
        variable pjk, j, k, psum;
    
        if (qualifier_exists("back_h") and qualifier_exists("back_s") and qualifier_exists("bkg_scale_h") and qualifier_exists("bkg_scale_s")){
            psum = 0;
            variable max_j, max_k;
            max_j = int(a);
            if (qualifier_exists("aprox")){
                max_j = min([int(a) , n]);
                                          }
            max_k = int(b);
            _for j (0,max_j,1){
                if (qualifier_exists("aprox")){
                    max_k = min([int(b) , n - j]);
                                              }
                _for k (0,max_k,1){
                    pjk = gsl->lngamma(da + j + 1) + gsl->lngamma(db + k + 1) - gsl->lngamma(da + 1) - gsl->lngamma(db + 1)
                          - j*log(ga + 1.) - k*log(gb + 1) + gsl->lngamma(a + b + 2 - j - k)
                          - gsl->lngamma(a - j + 1) - gsl->lngamma(b - k + 1) - gsl->lngamma(j + 1) - gsl->lngamma(k + 1)
                          + (a - j)*log(1 + bhr) + (b - k)*log(1 - bhr)
                          - (a + b + 2 - j - k)*log(ea + eb + (ea - eb)*bhr);
                    psum += exp(__tmp(pjk));
                                  }
                              }
            psum = psum/sum(psum*dbhr);
                                                                                                                                              }
        else{
            variable ph = gsl->lngamma(a + b + 2) + log(2.) - gsl->lngamma(a + 1) - gsl->lngamma(b + 1)
                          + (a + 1)*log(ea) + (b + 1)*log(eb) + a*log(1 + bhr)
                          + b*log(1 - bhr) - (a + b + 2)*log(ea + eb + (ea - eb)*bhr);
            psum = exp(__tmp(ph));
            }
    
        variable bhr_max = bhr[ where(psum == max(psum)) ];
    
        variable phsum = cumsum(psum*dbhr);
        variable iwa = where(phsum <= pval/2.);
        variable iwb = where(phsum >= 1-pval/2.);
        variable HR_fin, err_up, err_do;
        if ((length(bhr[iwb]) == 0) or (length(bhr[iwa]) == 0)){
            print(sprintf("%s: Ranges for uncertainties are empty.",_function_name()));
            print(sprintf("%s: Will only return best value and _NaN for uncertanties.",_function_name()));
            HR_fin = bhr_max[0]; 
            err_up = _NaN; 
            err_do = _NaN;
                                                               }
        else{                                                       
            variable bhrmin = max(bhr[iwa]);
            variable bhrmax = min(bhr[iwb]);
            variable result_struct;
    
            if ((bhr_max[0] > 0.0) and (bhr_max[0] >= bhrmax[0])){
                result_struct = round_conf(bhrmin[0] , bhr_max[0] , 1.0);
                                                                 }
            else if ((bhr_max[0] < 0.0) and (bhr_max[0] <= bhrmin[0])){
                result_struct = round_conf(-1.0 , bhr_max[0] , bhrmax[0]);
                                                                      }
            else{
                result_struct = round_conf(bhrmin[0] , bhr_max[0] , bhrmax[0]);
                }
            
            HR_fin = result_struct.value[0];
            err_up = result_struct.err_hi[0];
            err_do = result_struct.err_lo[0];
            }

        fin_HRs[xx] = HR_fin;
        fin_HRs_up[xx] = err_up;
        fin_HRs_down[xx] = err_do;

        list_append(grids , bhr);
        list_append(dists , psum);

        }

    if (qualifier_exists("fulloutput")){
        return (fin_HRs , fin_HRs_up , fin_HRs_down , grids , dists);
                                       }
    else{
        return (fin_HRs , fin_HRs_up , fin_HRs_down);
        }    
}
%
%
%
define hardnessratio()
%!%+
%\function{hardnessratio}
%\synopsis{calculates various X-ray hardness ratio from given count rates}
%\usage{Double_Type (HR , HR_err) = hardnessratio(Double_Type soft_count, Double_Type hard_count);}
%\altusage{Double_Type (HR , HR_err_up , HR_err_down) = hardnessratio(Double_Type soft_count, Double_Type hard_count ; bayesian);}
%\qualifiers{
%  \qualifier{color}{calculate the hardness ratio according to color=log10(S/H)}
%  \qualifier{hardness}{calculate the hardness ratio according to hardness=(H-S)/(H+S)}
%  \qualifier{ratio}{calculate the hardness ratio according to ratio=S/H (the default)}
%  \qualifier{bayesian}{returns hardness calculated using the bayesian estimation, calling 
%                         the function behr()}
%  \qualifier{back_s}{background counts in the soft band}
%  \qualifier{back_h}{background counts in the hard band}
%  \qualifier{backscale}{ratio between the extraction regions for the source and the background}
%  \qualifier{exposure}{Exposure per bin in seconds. If given, interpret soft_count, hard_count and t
%                         he backgrounds as rates. Multiply the rates with exposure to get the counts.}
%  \qualifier{backexposure}{if given, the background exposure. Only taken into
%               account if exposure is also set. If not given, it is assumed that
%               the source and background exposures are identical.}
%  \qualifier{ratio_type}{(DEPRECATED) Integer_Type, for hr=s/h choose 1,
%             for hr=(h-s)/(h+s) choose 2, for hr=log(s/h) choose 3, default = 1}
%  \qualifier{err_s}{Array containing the uncertainties of the soft light curve}
%  \qualifier{err_h}{Array containing the uncertainties of the hard light curve}
%      additional qualifiers are passed to 'behr' for bayesian estimation
%}
%\description
%  This function calculates the so-called hardness ratio or color according to 
%  the three common prescriptions used in X-ray astronomy: S/H, (H-S)/(H+S), and log10(S/H).
%  If given, background counts are subtracted before the calculation, taking into account
%  different source and background extraction regions and different source and background exposure
%  times if necessary (if counts are given but the background exposure was different, set the
%  source exposure to 1 and the background exposure to the ratio of the source and background exposure
%  times. If rates are given, give the exposure times.
%
%  Error bars are calculated using Gaussian error propagation. Contrary to earlier versions of
%  this routine they are ALWAYS calculated!
%
%  The bayesian approach or the function behr() should be used in low count regime.
%  If they qualifier "bayesian" is present the hardness is calcualted using the bayesian estimation 
%  implemented in the function behr(). If the background should be taken into account the function also requires 
%  the background scaling factors used in behr().
%  Requires gsl, make sure module is loaded!
%
%\seealso{behr;}
%!%-
{
    variable s,h;

    switch(_NARGS)
    { case 2: (s,h) = (); }
    { help(_function_name()); return; }

    if (__is_numeric(s)!=2) {
	s=typecast(s,Double_Type);
    }
    if (__is_numeric(h)!=2) {
	h=typecast(h,Double_Type);
    }

    %
    % backwards compatibility
    %
    variable type=0;
    if (qualifier_exists("ratio_type")) {
	if (qualifier_exists("color") or qualifier_exists("ratio") or qualifier_exists("hardness")) {
	    throw UsageError,sprintf("%s: ratio_type cannot be used with the color, ratio, or hardness qualifiers.",_function_name());

	}
    }
    if ((qualifier_exists("bayesian") and qualifier_exists("ratio")) or (qualifier_exists("bayesian") and qualifier_exists("color"))){
        throw UsageError,sprintf("%s: The bayesian approach is only available for the calculation of the hardness,",_function_name());
                                                                                                                                     }
    
    variable bs=qualifier("back_s",0.);
    variable bh=qualifier("back_h",0.);

    variable r=qualifier("backscale",1.);


    % Gehrels (1986) uncertainties
    variable sig_s = qualifier("err_s", sqrt(s+0.75)+1.);
    variable sig_h = qualifier("err_h", sqrt(h+0.75)+1.);

  
    %
    % exposure is given: convert rates to counts
    %
    if (qualifier_exists("exposure")) {
	variable expo=qualifier("exposure");
	s=s*expo;
	h=h*expo;

	variable backexpo=qualifier("backexposure",expo);
	bs=bs*backexpo;
	bh=bh*backexpo;

	r=r*expo/backexpo;

      % convert uncertainties from rates to counts
        if (qualifier_exists("err_s")){
	  sig_s = sig_s * expo;
        } else {
	  sig_s = sqrt(s+0.75)+1.;
          }

        if (qualifier_exists("err_h")){
          sig_h = sig_h * expo;
        } else {
	  sig_h = sqrt(h+0.75)+1.;
          }
     }
    
    % Gehrels (1986) uncertainties part 2
    variable sig_bs = (bs>0.) ? sqrt(bs+0.75)+1. : 0.;
    variable sig_bh = (bh>0.) ? sqrt(bh+0.75)+1. : 0.;

  
  
    variable dsbs=s-bs/r;
    variable dhbh=h-bh/r;

    variable sigs_sum=sqrt( sig_s^2.+(sig_bs/r)^2.);
    variable sigh_sum=sqrt( sig_h^2.+(sig_bh/r)^2.);

    % color: log10(S/H)
    if (qualifier_exists("color") || type==3) {
	variable color=log10(dsbs/dhbh);
	variable sig_color=sqrt( (sigs_sum/dsbs)^2. + (sigh_sum/dhbh)^2. )/log(10.);

	return (color,sig_color);
    }
    
    if (qualifier_exists("hardness") || type==2) {
	% hardness (H-S)/(H+S)
        variable hard;
        if (qualifier_exists("bayesian")){
            variable hard_u , hard_d;
            (hard , hard_u , hard_d) = behr(s , h ;; __qualifiers);
            return (hard , hard_u , hard_d);
                                         }
        else{                                 
	        hard=(h-s)/(h+s);
    	    variable sig_hard=2.*sqrt((dhbh*sigs_sum)^2.+(dsbs*sigh_sum)^2.)/(dhbh+dsbs)^2.;
	        return (hard,sig_hard);
            }
    }

    % default: S/H
    variable ratio=dsbs/dhbh;
    variable sig_ratio=ratio*sqrt( (sigs_sum/dsbs)^2. + (sigh_sum/dhbh)^2. );
    return (ratio,sig_ratio);

}




#endif

%
% Code from src/slang/math/hardnessratio_error_prop.sl
%
define hardnessratio_error_prop()
%!%+
%\function{hardnessratio_error_prop}
%\synopsis{calculates a ratio and error propagation}
%\usage{(hr, err) = hardnessratio_error_prop(h, h_err, s, s_err);}
%\description
%    \code{hr  = (h-s) / (h+s)}\n
%    \code{err = sqrt[ (2s/(h+s)^2 * h_err)^2  + (2h/(h+s)^2 * s_err)^2 ]}
%\seealso{ratio_error_prop}
%!%-
{
  variable h, h_err, s, s_err;
  switch(_NARGS)
  { case 4: (h, h_err, s, s_err) = (); }
  { help(_function_name()); return; }

  return 1.*(h-s)/(h+s),  2.*sqrt( 1.*(s*h_err)^2  + (h*s_err)^2 ) / (h+s)^2;
}

%
% Code from src/misc/astronomy/hardnessratio_from_dataset.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define hardnessratio_from_dataset()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{hardnessratio_from_dataset}
%\synopsis{calculates hardnessratios of the current model}
%\usage{Struct_Type hardnessratio_from_dataset(
%                  Integer_Type data-id, Integer_Type[2] soft_ch, hard_ch
%                );}
%\qualifiers{
%\qualifier{soft_en}{Double_Type[2]}
%\qualifier{hard_en}{Double_Type[2]}
%\qualifier{get_counts}{set to '&get_model_counts' if needed, default: get_data_counts}
%\qualifier{subtract_background}{subtract background from data. Only possible if get_counts is not set.}
%    additional qualifiers are passed to 'hardnessratio'
%}
%\description
%    - at least one dataset has to be defined (load_data)
%    - if energy bands are given, channels have to be set to [0,0]
%    - returns a structure containing: sc: soft counts, hc: hard counts, ratio: hardnessratio, err: error
%\seealso{hardnessratio}
%!%-
{
  variable id, soft_ch, hard_ch;
  switch (_NARGS)
    { case 3: (id,soft_ch,hard_ch) = (); }
    { help(_function_name()); return; }

  % check input
  if (length(soft_ch) != 2 || length(hard_ch) != 2) {
    vmessage("error (%s): channels must be arrays with 2 elements", _function_name);
    return;
  }
  if (length(id) != 1 || typeof(id) != Integer_Type) {
    vmessage("error (%s): 'data-id' must be a single integer", _function_name);
    return;
  }

  % get data or model counts
  variable getcounts = qualifier("get_counts", &get_data_counts);
  variable cts = _A(@getcounts(id));

  if (qualifier("subtract_background",1)) {
      if (qualifier_exists("get_counts")) {
	  vmessage("Cannot use subtract_background and get_counts qualifiers simultaneously", _function_name);
      }
      variable scale=get_back_data_scale_factor(id);
      variable backdata=get_back_data(id);

      cts.value-=scale*backdata;
  }


  % get the channels if energies are given
  % definition of no channels given: all array entries = 0
  if (all(soft_ch == 0) && all(hard_ch == 0)){
    % qualifiers have to be given in this case!
    variable soft_en = qualifier("soft_en", NULL);
    %variable mid_en = qualifier("mid_en", NULL);
    variable hard_en = qualifier("hard_en", NULL);
    if (soft_en == NULL || hard_en == NULL){
      vmessage("error (%s): energy-qualifiers have to be provided for each band", _function_name);
      return;
    }
    if (length(soft_en) != 2 || length(hard_en) != 2){
      vmessage("error (%s): energy-qualifiers must be arrays with 2 elements", _function_name);
      return;
    }
    % get corresponding channels
    soft_ch[0] = wherefirst(cts.bin_lo >= soft_en[0]);
    soft_ch[1] = wherelast(cts.bin_hi <= soft_en[1]);
    hard_ch[0] = wherefirst(cts.bin_lo >= hard_en[0]);
    hard_ch[1] = wherelast(cts.bin_hi <= hard_en[1]);
  }

  % sum counts up  
  variable sc = sum(cts.value[[soft_ch[0]:soft_ch[1]]]);
  variable hc = sum(cts.value[[hard_ch[0]:hard_ch[1]]]);
  variable ratio,err;
  (ratio,err) = hardnessratio(sc, hc;; __qualifiers);

  % return hardness,sc,hc
  variable hr=struct{sc=sc,hc=hc,ratio=ratio,err=err};
  %return (hardnessratio(sc, hc;; __qualifiers));%,sc,hc);
  return hr;
}

%
% Code from src/FITS/fits_save_fit/hardnessratio_from_spec.sl
%
%%%%%%%%%%%%%%%%%%%%%
define hardnessratio_from_spec()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{hardnessratio_from_spec}
%\synopsis{calculates hardness ratio from given spectrum}
%\usage{Struct_Type H = hardnessratio_from_spec(String_Type fits, freeze_model_comp)}
%\description
%    Calculates hardness ratio from a given spectrum using the model 'enflux'.
%    First argument is a fits file with data and model (from fits_save_fit).
%    Second argument is the model component of the continuum that has to be frozen (see enflux).
%    It first determines the soft and hard energy flux densities of two variable energy bands
%    and then derives the hardness ratio plus error.
%    For the hardness ratio two different definitions can be chosen (either h/s or
%    (h-s)/(h+s)).
%\qualifiers{
%\qualifier{hard_band}{Double_Type[2], hard energy band, default = [7.,10.]}
%\qualifier{soft_band}{Double_Type[2], soft energy band, default = [2.,4.]}
%\qualifier{hr_def}{Integer_Type, for hr=h/s choose 1, for hr=(h-s)/(h+s) choose 2, default = 1}
%\qualifier{roc}{Integer_Type, RMF OGIP compliance, default = 2}
%}
%\seealso{hardnessratio_error_prop,enflux,fits_save_fit}
%!%-
{
  variable fits,freeze_model_comp;
  switch(_NARGS)
  { case 2: (fits,freeze_model_comp) = (); }
  { help(_function_name()); return; }


  variable hard = qualifier("hard_band",[7.,10]);
  variable soft = qualifier("soft_band",[2.,4.]);
  variable roc = qualifier("roc",2);

  variable hr_def = qualifier("hr_def",1);

  ()=fits_load_fit(fits;ROC=roc);
  variable fun = get_fit_fun;
  fit_fun(sprintf("enflux(1,%s)",fun));
  set_par(freeze_model_comp,get_par(freeze_model_comp),1,0,1); % freeze continuum norm
  set_par("enflux(1).enflux",get_par("enflux(1).enflux"),0,0,1); % constrain enflux

  % hard energy range
  set_par("enflux(1).E_min",hard[0],1,0,100);
  set_par("enflux(1).E_max",hard[1],1,0,100);
  ()=fit_counts;
  variable h=fit_pars([get_par_info("enflux(1).enflux").index[0]]);

  % soft energy range
  set_par("enflux(1).E_min",soft[0],1,0,100);
  set_par("enflux(1).E_max",soft[1],1,0,100);
  ()=fit_counts;
  variable s=fit_pars([get_par_info("enflux(1).enflux").index[0]]);

  variable s_err = s.conf_max[0]-s.conf_min[0];
  variable h_err = h.conf_max[0]-h.conf_min[0];

  variable hr, err;
  (hr, err) = (hr_def==1)
            ? (h.value[0]/s.value[0], sqrt((1/s.value[0])^2*h_err^2+(h.value[0]/(s.value[0]^2))*s_err^2))
            : hardnessratio_error_prop(h.value[0],h_err,s.value[0],s_err);

  return struct{soft_flux=s.value[0],soft_flux_confmin=s.conf_min[0],soft_flux_confmax=s.conf_max[0],
    hard_flux=h.value[0],hard_flux_confmin=h.conf_min[0],hard_flux_conf_max=h.conf_max[0],hr=hr,hr_err=err};
}

%
% Code from src/misc/astronomy/hardnessratio_simulate_grid.sl
%
define hardnessratio_simulate_grid()
%!%+
%\function{hardnessratio_simulate_grid}
%\synopsis{calculates hardnessratios of the current model}
%\usage{Struct_Type[] hardnessratio_simulate_grid(String_Type/Integer_Type par1name, par2name,
%		     Integer_Type[2] soft_ch1, hard_ch1, soft_ch2, hard_ch2);
% or Struct_Type[] hardnessratio_simulate_grid(
%      String_Type/Integer_Type par1name, Double_Type par1min, par1max, step1,
%      String_Type/Integer_Type par2name, Double_Type par2min, par2max, step2,
%      Integer_Type soft_ch1, hard_ch1, soft_ch2, hard_ch2
%    );}
%\qualifiers{
%\qualifier{dataindex}{Integer_Type, dataset index to use, default = 1}
%\qualifier{grid1scale}{0 = linear (default), 1 = logarithmic}
%\qualifier{grid2scale}{0 = linear (default), 1 = logarithmic}
%\qualifier{par1grid}{Double_Type[], override value grid of paramter 1}
%\qualifier{par2grid}{Double_Type[], override value grid of paramter 2}
%\qualifier{sample1}{Integer_Type, sampling-factor along each track, default 10}
%\qualifier{sample2}{Integer_Type, sampling-factor along each track, default 10}
%\qualifier{exposure}{Double_Type, default: 1e4}
%\qualifier{arf}{String_Type, file path of arf}
%\qualifier{rmf}{String_Type, file path of rmf}
%\qualifier{rsp}{String_Type, file path of rsp (arf and rmf combined)}
%\qualifier{soft_en1}{Double_Type[2], additional qualifier, soft band for all tracks in one direction}
%\qualifier{hard_en1}{Double_Type[2], additional qualifier, hard band for all tracks in one direction}
%\qualifier{soft_en2}{Double_Type[2], additional qualifier, soft band for all tracks in other direction}
%\qualifier{hard_en2}{Double_Type[2], additional qualifier, hard band for all tracks in other direction}
%    additional qualifiers are passed to 'hardnessratio_from_dataset' and 'hardnessratio'
%}
%\description
%    - at least one dataset has to be defined (load_data)
%    - a model must exist (fit_fun)
%    - soft_ch1, hard_ch1, soft_ch2, hard_ch2 are passed to 'hardnessratio_from_dataset'; if energy-bands
%      (soft_en1...)are given, the channels have to be set to [0,0] each.
%    - if logarithmic gridscale is chosen, be careful 'parmin' is not <= zero (be aware of the default
%      values of your model)
%    - the output consists of an array of structures of tracks, where each track contains the values of par1, par2 and the 
%     corresponding hardnessratios hr1 and hr2, and the errors hr1err and hr2err e.g. within one struct par1 is kept constant
%    - either give arf AND rmf; or ONLY give rsp
%\seealso{hardnessratio, hardnessratio_from_dataset, xfigplot_hardnessratio_grid}
%!%-
{
 variable par1name, par2name, par1min, par2min, par1max, par2max, step1 = 10, step2 = 10;
 variable soft_ch1, hard_ch1, soft_ch2, hard_ch2;
 switch(_NARGS)
 { case 6: (par1name, par2name, soft_ch1, hard_ch1, soft_ch2, hard_ch2) = (); }
 { case 12: (par1name, par1min, par1max, step1, par2name, par2min, par2max, step2, soft_ch1, hard_ch1, soft_ch2, hard_ch2) = (); }
 { help(_function_name()); return; }

 % get default values for parameter ranges (if not already set by user)
 variable finipar = merge_struct_arrays(get_params);
 variable par1type = typeof(par1name) == String_Type ? 1 : 0;
 variable par2type = typeof(par2name) == String_Type ? 1 : 0;
 variable par1ind = wherefirst((par1type ? finipar.name : finipar.index) == par1name);
 variable par2ind = wherefirst((par2type ? finipar.name : finipar.index) == par2name);
 if (__is_numeric(par1min) == 0) { par1min = finipar.min[par1ind]; }
 if (__is_numeric(par1max) == 0) { par1max = finipar.max[par1ind]; }
 if (__is_numeric(par2min) == 0) { par2min = finipar.min[par2ind]; }
 if (__is_numeric(par2max) == 0) { par2max = finipar.max[par2ind]; }
 variable sample1 = qualifier("sample1", 10);
 variable sample2 = qualifier("sample2", 10);
  
 % build parameter arrays
 variable par1grid=qualifier("par1grid",NULL);
 variable par2grid=qualifier("par2grid",NULL);
 variable par1arr, par2arr;

 if(par1grid == NULL){
   par1arr =  qualifier("grid1scale", 0) == 0
   ? 1. * par1min + (par1max-par1min)*[0:step1-1]/(step1-1.)
   : exp(1. * log(par1min) + (log(par1max)-log(par1min))*[0:step1-1]/(step1-1.));
 }else{
   par1arr = par1grid;
 }

 if(par2grid == NULL){
   par2arr = qualifier("grid2scale", 0) == 0
   ? 1. * par2min + (par2max-par2min)*[0:step2-1]/(step2-1.)
   : 1. * exp(log(par2min) + (log(par2max)-log(par2min))*[0:step2-1]/(step2-1.));
 }else{
   par2arr = par2grid;
 }

 % dataset index to get the model/data counts
 variable idat = qualifier("dataindex", 1);

 % set data exposure, ARF, RMF, or data-exposure?
 variable expt = qualifier("exposure", 1e4);
 variable arf = qualifier("arf", NULL);
 if (arf != NULL) {
   vmessage("assigning ARF");
   if (typeof(arf) == String_Type) { arf = load_arf(arf); }
   assign_arf(arf, idat);
   set_arf_exposure(arf, expt);
 }
 variable rsp = qualifier("rsp", NULL);
 if (rsp != NULL) {
   vmessage("loading RSP");
   if (typeof(rsp) == String_Type) { rsp = load_rmf(rsp); }
   arf=factor_rsp(rsp);
   vmessage("assigning RMF");
   assign_rmf(rsp, idat);
   vmessage("assigning ARF");
   assign_rmf(arf, idat);
 }
 if (all(all_data != idat)) {
   vmessage("creating dataset and setting exposure");
   set_data_exposure(idat, expt);
 }
 variable rmf = qualifier("rmf", NULL);
 if (rmf != NULL) {
   vmessage("assigning RMF");
   if (typeof(rmf) == String_Type) { rmf = load_rmf(rmf); }
   assign_rmf(rmf, idat);
 }

 % set qualifier-structures to be passed to hardnessratio_from_dataset
 variable qual1 = COPY(__qualifiers), qual2 = COPY(__qualifiers);
 if (qualifier_exists("soft_en1")) { qual1 = struct_combine(qual1, struct { soft_en = qualifier("soft_en1") }); }
 if (qualifier_exists("hard_en1")) { qual1 = struct_combine(qual1, struct { hard_en = qualifier("hard_en1") }); }
 if (qualifier_exists("soft_en2")) { qual2 = struct_combine(qual2, struct { soft_en = qualifier("soft_en2") }); }
 if (qualifier_exists("hard_en2")) { qual2 = struct_combine(qual2, struct { hard_en = qualifier("hard_en2") }); }
 qual1 = struct_combine(qual1, struct { subtract_background=0, get_counts = &get_model_counts });
 qual2 = struct_combine(qual2, struct { subtract_background=0, get_counts = &get_model_counts });

 % prepare output struct-array
 variable tracks = Struct_Type[0];
 variable x, y, len, track, ratio1, ratio2;
 % loop over parameter 1
 _for x (0, length(par1arr)-1, 1) {
   % prepare track structure
   len = length(par2arr)*sample2;
   track = struct {
     par1 = par1arr[x] * ones(len), par2 = [par2arr[0]:par2arr[-1]:#len],
     hr1 = Double_Type[len], hr2 = Double_Type[len],
     hr1err = Double_Type[len], hr2err = Double_Type[len],
     linetype = -1
   };
   % set parameters
   set_par(par1name, par1arr[x]);
   _for y (0, len-1, 1) {
     set_par(par2name, track.par2[y]);
     % evaluate model (fold with detector response)
     ()=eval_counts(; fit_verbose = -1);
     % calculate hardnesses
     ratio1=hardnessratio_from_dataset(idat, soft_ch1, hard_ch1;; qual1);
     ratio2=hardnessratio_from_dataset(idat, soft_ch2, hard_ch2;; qual2);
     track.hr1[y]=ratio1.ratio;
     track.hr2[y]=ratio2.ratio;
     track.hr1err[y]=ratio1.err;
     track.hr2err[y]=ratio2.err;
     %track.hr1[y] = hardnessratio_from_dataset(idat, soft_ch1, hard_ch1;; qual1);
     %track.hr2[y] = hardnessratio_from_dataset(idat, soft_ch2, hard_ch2;; qual2);
   }

   tracks = [tracks, track];
 }
 % loop over parameter 2
 _for x (0, length(par2arr)-1, 1) {
   % prepare track structure
   len = length(par1arr)*sample1;
   track = struct {
     par1 = [par1arr[0]:par1arr[-1]:#len], par2 = par2arr[x] * ones(len),
     hr1 = Double_Type[len], hr2 = Double_Type[len],
     hr1err = Double_Type[len], hr2err = Double_Type[len],
     linetype = 1
   };
   % set parameters
   set_par(par2name, par2arr[x]);
   _for y (0, len-1, 1) {
     set_par(par1name, track.par1[y]);
     % evaluate model (fold with detector response)
     ()=eval_counts(; fit_verbose = -1);
     % calculate hardnesses
     ratio1=hardnessratio_from_dataset(idat, soft_ch1, hard_ch1;; qual1);
     ratio2=hardnessratio_from_dataset(idat, soft_ch2, hard_ch2;; qual2);
     track.hr1[y]=ratio1.ratio;
     track.hr2[y]=ratio2.ratio;
     track.hr1err[y]=ratio1.err;
     track.hr2err[y]=ratio2.err;
     %track.hr1[y] = hardnessratio_from_dataset(idat, soft_ch1, hard_ch1;; qual1);
     %track.hr2[y] = hardnessratio_from_dataset(idat, soft_ch2, hard_ch2;; qual2);
   }

   tracks = [tracks, track];
 }

 return tracks;
}

%
% Code from src/misc/astronomy/hardnessratio_simulate_grid_save.sl
%
define hardnessratio_simulate_grid_save (tracks,filename)
%!%+
%\function{hardnessratio_simulate_grid_save}
%\synopsis{saves the output of hardnessratio_simulate_grid into a fits
%    file}
%\usage{hardnessratio_simulate_grid_save(Struct_Type tracks,
%    String_Type filename);}
%\description
%    - tracks is the output strcutute of hardnessratio_simulate_grid
%\seealso{hardnessratio, hardnessratio_from_dataset,
%    xfigplot_hardnessratio_grid, xfigplot_hardnessratio_grid_load}
%!%-
{
  variable nn = length(tracks);  
  variable i;
  _for i (0,nn-1,1){
    
    variable temp = struct{par1=tracks[i].par1,
      par2=tracks[i].par2,
      hr1=tracks[i].hr1,
      hr2=tracks[i].hr2};
    
    if (i==0) {      
      fits_write_binary_table(filename,string(i+1),temp,
			      struct{linetype=tracks[0].linetype});
    }  else {      
      fits_append_binary_table(filename,string(i+1),temp,
			     struct{linetype=tracks[i].linetype});
    }
    
  }
  
}

define hardnessratio_simulate_grid_load (filename)
%!%+
%\function{hardnessratio_simulate_grid_load}
%\synopsis{load tracks saved with hardnessratio_simulate_grid_save}
%\usage{Struct_Type = hardnessratio_simulate_grid_save(String_Type filename);}
%\seealso{hardnessratio, hardnessratio_from_dataset,
%    xfigplot_hardnessratio_grid, xfigplot_hardnessratio_grid_save}
%!%-
{
  variable nn = fits_nr_extensions(filename);  
  variable tracks = Struct_Type[0];
  
  variable i;
  _for i (0,nn-1,1) {        
    tracks = [tracks,[struct_combine(fits_read_table(filename+"\["+string(i+1)+"\]"),
				     fits_read_key_struct(filename+"\["+string(i+1)+"\]","linetype"))]];       
  }
  
  return tracks;
  
}
%
% Code from src/slang/help.sl
%
#ifnexists help
define help(topic)
{
  variable doc = get_doc_string_from_file (topic);
  if (doc == NULL)
    vmessage ("No help found for %S\n", topic);
  else
    message (topic);
}
#endif

%
% Code from src/histogram/hist_fwhm.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define hist_fwhm_index ()
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{hist_fwhm_index}
%\usage{(lo,hi) = hist_fwhm_index(hist [, max_index]);}
%\synopsis{Compute index bounds spaning FWHM of histogram}
%\description
%    When given a single array hist_fwhm_index tries to find the FWHM
%    range from the maximum. This works much better if the second argument
%    is given which has to be the index of the maximum (or at least close to
%    it). By interative stepping of the boundaries the function will stop
%    when the lo and hi enclose the true fwhm or has hit the array boundaries.
%
%!%-
{
  variable Histogram; % input histogram
  variable maxIndex; % input index
  variable iplus; % runs to lower indices
  variable iminus; % runs to higher indices
  variable histLen; % hist length
  variable fullMax; % maximum value of hist
  variable newMax; % if new maximum found
  variable loBounced; % if low hit bound
  variable hiBounced; % if high hit bound
  variable interp = qualifier_exists("interp"); % linear interpolate to get "true" fwhm
  variable loSlope; % slope for interpolation
  variable hiSlope;
  variable iplusinterp;
  variable iminusinterp;

  switch (_NARGS)
  { case 1: Histogram = ();
    maxIndex = wherefirstmax(Histogram);
  }
  { case 2: (Histogram, maxIndex) = ();
  }
  { help(_function_name);
    return;
  }

  fullMax = Histogram[maxIndex];
  iplus = maxIndex;
  iminus = maxIndex;
  histLen = length(Histogram);

  if (histLen <= maxIndex || 0 > maxIndex)
    throw UsageError, "given index not in range of array";

  newMax = 0;
  loBounced = 0;
  hiBounced = 0;
  while (not loBounced || not hiBounced) {
    if (not loBounced && iminus>0)
      iminus--;

    if (not hiBounced && iplus < histLen-1)
      iplus++;

    if (iminus == 0 || Histogram[iminus] <= .5*fullMax)
      loBounced = 1;

    if ((iplus == (histLen-1)) || Histogram[iplus] <= .5*fullMax)
      hiBounced = 1;

    if (Histogram[iminus] > fullMax) {
      fullMax = Histogram[iminus];
      maxIndex = iminus;
      newMax = 1;
    }

    if (Histogram[iplus] > fullMax) {
      fullMax = Histogram[iplus];
      maxIndex = iplus;
      newMax = 1;
    }

    if (newMax) {
      hiBounced = 0;
      loBounced = 0;

      while (Histogram[iplus] < .5*fullMax)
	iplus--;

      while (Histogram[iminus] < .5*fullMax)
	iminus++;

      newMax = 0;
    }
  }

  if (interp) {
    loSlope = Histogram[iminus+1]-Histogram[iminus];
    hiSlope = Histogram[iplus]-Histogram[iplus-1];
    iminusinterp = 1.*iminus + ((iminus == 0)         ? 0. : .5*fullMax/loSlope - Histogram[iminus]/loSlope);
    iplusinterp = 1.*iplus  + ((iplus == histLen-1) ? 0. : .5*fullMax/hiSlope - Histogram[iplus]/hiSlope);
    iminus = (loSlope == 0) ? iminus : iminusinterp;
    iplus =  (hiSlope == 0) ? iplus  : iplusinterp;
  }

  return (iminus, iplus);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_instrument_resolution_from_data (id)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_instrument_resolution_from_data}
%\usage{R = get_instrument_resolution_from_data(id)}
%\synopsis{Compute energy resolution from data set 'id'}
%\qualifiers{
%\qualifier{energy}{If given will return the grid as energy grid
%               instead of wavelength}
%}
%\description
%   This function can be used to estimate the instrument resolution
%   from the data set specified with id. The return value is a struct
%   { value, bin_lo, bin_hi } where bin_lo and bin_hi is the data
%   grid (usually in Angstrom) and value is the resolution in keV.
%
%   Note that this function removes any binning from the data set.
%
%\seealso{hist_fwhm_index}
%!%-
{
  rebin_data(id, 0);
  variable response = get_dataset_response(id);
  variable cmin, cmax;
  variable resolution = Double_Type[length(response)];
  variable i;
  variable grid = get_rmf_data_grid(get_data_info(id).rmfs[0]);
  variable fmin, fmax;
  variable energy = qualifier_exists("energy");

  _for i (0, length(response)-1, 1) {
    (cmin,cmax) = hist_fwhm_index(response[i], i);
    fmin = (response[i][cmin+1] == response[i][cmin]) ? 0.0 :
      (max(response[i][[cmin:cmax]])*.5-response[i][cmin])/
      (response[i][cmin+1]-response[i][cmin]);
    fmax = (response[i][cmax] == response[i][cmax-1]) ? 0.0 :
      (max(response[i][[cmin:cmax]])*.5-response[i][cmax])/
      (response[i][cmax]-response[i][cmax-1]);
    resolution[i] = _A(grid.bin_lo[cmin] + (grid.bin_lo[cmin+1]-grid.bin_lo[cmin])*fmin)
      -_A(grid.bin_hi[cmax] - (grid.bin_hi[cmax]-grid.bin_hi[cmax-1])*fmax);
  }

  if (energy)
    return struct { value = resolution, bin_lo = _A(grid.bin_hi), bin_hi = _A(grid.bin_lo) };
  else
    return struct { value = resolution, @grid };
}

%
% Code from src/histogram/histogram2d_min_max.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define histogram2d_min_max()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{histogram2d_min_max}
%\synopsis{computes a 2d histogram between minimum and maximum data values}
%\usage{h2 = histogram2d_min_max(Double_Type Y, X);}
%\qualifiers{
%\qualifier{xmin}{[\code{=min(X)}]: first value of \code{Xlo}-grid}
%\qualifier{xmax}{[\code{=max(X)}]: last value of \code{Xhi}-grid}
%\qualifier{ymin}{[\code{=min(Y)}]: first value of \code{Ylo}-grid}
%\qualifier{ymax}{[\code{=max(Y)}]: last value of \code{Yhi}-grid}
%\qualifier{Nx}{[=50]: number of bins of (linear) \code{Xlo}-grid}
%\qualifier{Ny}{[=50]: number of bins of (linear) \code{Ylo}-grid}
%\qualifier{Xlo}{reference to a variable to store the \code{Xlo} array}
%\qualifier{Ylo}{reference to a variable to store the \code{Ylo} array}
%}
%\description
%    For 2d arrays, the order of the indices matters.
%    Almost all ISIS and related functions use \code{h2[iy, ix]}
%    with the first index corresponding to y, and the second to x.
%    For this reason, \code{histogram2d_min_max} -- just as
%    \code{histogram2d} (though not explicitly documented) --
%    needs to get the array \code{Y} of y-coordinates as first argument
%    and the array \code{X} of x-coordinates only as second argument,
%    if the resulting 2d array \code{h2} shall be used with functions
%    like \code{plot_image}, \code{png_write}, \code{ds9_view}, etc..
%
%    The x-grid \code{Xlo} starts at \code{xmin}, but ends before xmax,
%    such that \code{Xhi = make_hi_grid(Xlo)} would end at \code{xmax}.
%    The same is true for the y-grid \code{Ylo} with \code{ymin} and \code{ymax}.
%
%    Unlike \code{histogram2d(Y, X, Ylo, Xlo)}, \code{h2} will not contain
%    overflow bins in the last column and the last row, i.e.,
%    \code{h2[iy,ix]} corresponds to the number of pairs (\code{X}, \code{Y})
%    where \code{Xlo[ix] <= X < Xhi[ix]} and \code{Ylo[iy] <= Y < Xhi[iy]}.
%\example
%    variable n=10000; x=2*grand(n), y=grand(n), Xlo, Ylo;
%    variable h2 = histogram2d_min_max(y, x; Xlo=&Xlo, Ylo=&Ylo);
%    plot_image(h2, 0, Xlo+(Xlo[1]-Xlo[0])/2., Ylo+(Ylo[1]-Ylo[0])/2.);
%\seealso{histogram2d}
%!%-
{
  variable X, Y;
  switch(_NARGS)
  { case 2: (Y, X) = (); }
  { help(_function_name()); return; }

  variable Xlo = [qualifier("xmin", min(X)) : qualifier("xmax", max(X)) : #(qualifier("Nx", 50)+1)];
  variable Ylo = [qualifier("ymin", min(Y)) : qualifier("ymax", max(Y)) : #(qualifier("Ny", 50)+1)];
  variable XloRef = qualifier("Xlo");  if(typeof(XloRef)==Ref_Type)  @XloRef = Xlo[[:-2]];
  variable YloRef = qualifier("Ylo");  if(typeof(YloRef)==Ref_Type)  @YloRef = Ylo[[:-2]];
  return histogram2d(Y, X, Ylo, Xlo)[[:-2],[:-2]];
}

%
% Code from src/histogram/histogram3d.sl
%
define histogram3d()
%!%+
%\function{histogram3d}
%\synopsis{bins scatter data into a 3d histogram}
%\usage{Integer_Type[,,] histogram3d(Double_Type x[], y[], z[], Xgrid[], Ygrid[], Zgrid[])}
%\description
%    \code{histogram3d} computes the number \code{N[i,j,k]} of points \code{(x[m], y[m], z[m])}
%    that fall into the 3d-cell with \code{Xgrid[i] <= x < Xgrid[i+1]},
%    \code{Ygrid[j] <= y < Ygrid[j+1]} and \code{Zgrid[k] <= z < Xgrid[k+1]}.
%    The last bin in each dimension is an overflow bin,
%    such that its upper limit is at infinity.
%\seealso{histogram, histogram2d}
%!%-
{
  variable x, y, z, Xgrid, Ygrid, Zgrid;
  switch(_NARGS)
  { case 6: (x, y, z, Xgrid, Ygrid, Zgrid) = (); }
  { return help(_function_name()); }

  if(length(x)!=length(y) || length(y)!=length(z))
    return vmessage("error (%s): arrays x, y, z have to have the same length", _function_name());

  variable h3 = Integer_Type[length(Xgrid),length(Ygrid),length(Zgrid)];
  variable i, iz;
  _for iz (0, length(Zgrid)-2, 1)
  {
    i = where(Zgrid[iz] <= z < Zgrid[iz+1]);
    if(length(i))
      h3[*,*,iz] = histogram2d(x[i], y[i], Xgrid, Ygrid);
  }
  i = where(z >= Zgrid[-1]);  % overflow bin
  if(length(i))
    h3[*,*,-1] = histogram2d(x[i], y[i], Xgrid, Ygrid);

  return h3;
}

%
% Code from src/histogram/histogram_gaussian_probability.sl
%
define histogram_gaussian_probability()
%!%+
%\function{histogram_gaussian_probability}
%\usage{Double_Type[] histogram_gaussian_probability(Double_Type[] x, sigma, lo[, hi])}
%\description
%    While \code{histogram(x, lo[, hi])} increases the histogram bin \code{j}
%    where \code{lo[j] <= x[i] < hi[j]} by one for every \code{x[i]},
%    \code{histogram_gaussian_probability(x, lo[, hi])} adds the
%    the Gaussian proability of events with mean \code{x[i]}
%    and standard deviation \code{sigma[i]} to all bins.
%
%    Note that this function acts like a convolution
%    and therefore introduces an additional broadening.
%\example
%    variable n=10000, x=grand(n), sigma=ones(n);
%    variable lo=[-3:3:0.05],  hi=make_hi_grid(lo);
%    hplot(lo, hi, histogram(x, lo, hi));
%    ohplot(lo, hi, histogram_gaussian_probability(x, sigma, lo, hi));
%    % The first distribution follows N(0, 1),
%    % but the second one follows N(0, sqrt(1^2 + sigma^2)) = N(0, sqrt(2)).
%\seealso{histogram}
%!%-
{
  variable x, sigma, lo, hi;
  switch(_NARGS)
  { case 3: (x, sigma, lo) = (); hi = make_hi_grid(lo); }
  { case 4: (x, sigma, lo, hi) = (); }
  { return help(_function_name()); }

  variable handle_gauss = fitfun_handle("gauss");

  variable offset = 1-min(lo);
  lo += offset;  % in order to call eval_fun2
  hi += offset;  % with valid wavelength grid

  variable i, h=0*lo;
  _for i (0, length(x)-1, 1)
    h += eval_fun2(handle_gauss, lo, hi, [1, x[i]+offset, sigma[i]]);
  return h;
}


%
% Code from src/histogram/histogram_min_max.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define histogram_min_max()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{histogram_min_max}
%\synopsis{computes a histogram between minimum and maximum value}
%\usage{Struct_Type h = histogram_min_max(Double_Type X[, Double_Type dx]);}
%\qualifiers{
%\qualifier{log}{use a logarithmic grid with the following number of bins:}
%\qualifier{N}{[=100] number of bins of logarithmic or linear grid}
%}
%\description
%    The return value is a \code{{ bin_lo, bin_hi, value, err }} structure which
%    can directly be used with, e.g., \code{hplot_with_err}, \code{define_counts}, etc.
%\seealso{histogram}
%!%-
{
  variable X, dx=NULL;
  switch(_NARGS)
  { case 1: X = (); }
  { case 2: (X, dx) = (); }
  { return help(_function_name()); }

  variable h = struct { bin_lo, bin_hi, value, err };
  variable mn, mx; (mn, mx) = min_max( X[ where(not isnan(X) and not isinf(X)) ] );
  if(qualifier_exists("log"))
  {
    if(mn<=0)  mn = min( X[ where(not isnan(X) and not isinf(X) and X>0) ] );
    (h.bin_lo, h.bin_hi) = log_grid(mn, mx, qualifier("N", 100));
  }
  else
  { if(dx==NULL)  dx = (mx-mn)/qualifier("N", 100);
    h.bin_lo = [mn : mx : dx];
    h.bin_hi = make_hi_grid(h.bin_lo);
  }
  h.value  = histogram(X, h.bin_lo, h.bin_hi);
  h.err = sqrt(h.value);  % Poisson statistics (in the Gaussian limit)
  return h;
}

%
% Code from src/misc/ISISshell/history.sl
%
define history()
%!%+
%\function{history}
%\synopsis{shows the history of commands on the interactive ISIS-shell}
%\usage{history();}
%\seealso{save_input}
%!%-
{
  variable tmpFile = sprintf("/tmp/ISIS.history.%d.%d", getuid(), getpid());
  save_input(tmpFile);

  variable fp = fopen(tmpFile, "r"), line;
  foreach line (fp)
    message(line);
  ()=fclose(fp);
  ()=remove(tmpFile);
}

%
% Code from src/plot/hplot_filled.sl
%
define hplot_filled(lo,hi,val)
%!%+
%\function{hplot_filled}
%\synopsis{plot a filled histogram defined by slang arrays}
%\usage{hplot_filled(Array_Type bin_lo, Array_Type bin_hi, Array_Type values)
%}
%\qualifiers{
%\qualifier{fill_style}{[\code{=1}] set the fill style:\n
%                       \code{FS = 1} \code{=>} solid (default)\n
%                       \code{FS = 2} \code{=>} outline\n
%                       \code{FS = 3} \code{=>} hatched (cannot be used with ylog;)\n
%                       \code{FS = 4} \code{=>} cross-hatched (cannot be used with ylog;)\n
%}
%\qualifier{ticks}{[\code{=""}] set to \code{"I"} or \code{"P"} to invert or project the ticks
%                            of both axis (additional to initial plot options)}
%\qualifier{xopt}{[\code{=get_plot_options.xopt}] set the plot options of the x-axis directly}
%\qualifier{yopt}{[\code{=get_plot_options.yopt}] set the plot options of the y-axis directly}
%\qualifier{ymin}{[\code{=min(values)}] set the lower y-value to which the areas are filled}
%\qualifier{angle}{[\code{=degrees}] sets the angle of the hatched lines for FS=3 [default = 45] }
%}
%\description
%    This function plots a histogram described by three 1-D S-Lang arrays of size N.
%    The area below the histogram is filled. The fill style can be selected with a qualifier.
%\examples
%    \code{hplot_filled([1:5],[2:6],[1:5]);}\n
%
%    \code{%} invert the ticks\n
%    \code{hplot_filled([1:5],[2:6],[1:5]; ticks="I");}\n
%        
%    \code{%} change the fill style\n
%    \code{hplot_filled([1:5],[2:6],[1:5] ; fill_style=4);}\n
%\seealso{ohplot_filled, hplot}
%!%-
{
  variable popt = get_plot_options;
  variable init_xopt = popt.xopt;
  variable init_yopt = popt.yopt;
  popt.xopt =  qualifier("xopt", popt.xopt + qualifier("ticks", ""));
  popt.yopt =  qualifier("yopt", popt.yopt + qualifier("ticks", ""));
  set_plot_options(popt);
  
  hplot(lo,hi,val);
  variable fs = qualifier("fill_style",1);
  
  %get hatching style and save it in order to reset it in the end
  variable hangle, hsep, hphase ;
  (hangle, hsep, hphase) = _pgqhs ;
  
  _pgsfs(fs);  
  if (fs == 3) { _pgshs(qualifier("angle",45), 1, 0);}

  variable i;
  if (get_plot_options.logx)
  {
    lo = log10(lo);
    hi = log10(hi);
  }
  if (get_plot_options.logy) val = log10(val);
  variable y_min = qualifier("ymin",max([min(val),get_plot_options.ymin ]));
  if (get_plot_options.logy) y_min = log10(y_min);
  if (get_plot_options.logy && (fs > 2)) 
  {
    printf("warning: fill_style = 3,4 not allowed with ylog;\nfill_style was set to 1\n");
    _pgsfs(1);
  }
  _for i (0, length(lo)-1, 1) 
  {
    ()=_pgpoly(4,[lo[i],lo[i],hi[i],hi[i]],[y_min,val[i],val[i],y_min]);
  }
  
  popt = get_plot_options;
  popt.xopt = init_xopt;
  popt.yopt = init_yopt;
  set_plot_options(popt);
  
  %reset hatching style to default values:
  _pgshs(hangle, hsep, hphase);
}

%
% Code from src/plot/hplot_with_err.sl
%
%%%%%%%%%%%%%%%%%%%%%
define hplot_with_err()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{hplot_with_err}
%\synopsis{plots histogram data points with errorbars}
%\description
%    This function passes all its arguments and qualifiers to the
%    \code{plot_with_err} function, but adds the \code{histogram} and \code{xminmax} qualifiers.
%\seealso{[o][h]plot_with_err, [o][h]plot}
%!%-
{
  if(_NARGS)
  { variable args = __pop_list(_NARGS);
    plot_with_err(__push_list(args);; struct_combine(__qualifiers(), "histogram", "xminmax"));
  }
  else
    help(_function_name());
}

%
% Code from src/histogram/hquantile.sl
%
define hquantile()
{
  variable p, hist;
  switch(_NARGS)
  { case 2: (p, hist) = (); }
  { help(_function_name()); return; }

  variable i, total_sum = sum(hist), partial_sum = 0;
  _for i (0, length(hist)-1, 1)
  { partial_sum += hist[i];
    if(partial_sum >= p*total_sum) { return i; }
  }
  return length(hist)-1;
}

%
% Code from src/slang/color/hsl2rgb.sl
%
private define qphcol (q, p, h6)
{
  if (h6 < 1.0) return p + (q-p)*h6;
  if (h6 < 3.0) return q;
  if (h6 < 4.0) return p + (q-p)*(4.0-h6);
  return p;
}

private define map_hsl(h, s, l)
{
  if (s == 0)
    return l,l,l;

  variable q = l*(1.0+s);
  if (l > 0.5) q = l+s - l*s;

  variable p = 2.0*l - q;
  
  return qphcol(q, p, (h*6.0+2.0) mod 6.0), qphcol(q, p, h*6.0), qphcol(q, p, (h*6.0+4.0) mod 6.0);
}

define hsl2rgb(h, s, l)
%!%+
%\function{hsl2rgb}
%\synopsis{converts a (hue, saturation, lightness) color to (red, green, blue)}
%\usage{Integer_Type rgb = hsl2rgb(Double_Type h, s, l);}
%\description
%    The hue \code{h} is an angle in color-space. Here, \code{h} is between 0 and 1.
%    \code{h} of red is 0, \code{h} of green is 1/3, and \code{h} of blue is 2/3.
%    The lightness \code{l} of black is 0, and \code{l} of white is 1.
%    Like the hue, the saturation \code{0<=s<=1} matters only for \code{0<l<1}.
%
%    The return value is a(n array of) 24 bit RGB value(s).
%\seealso{rgb2hsl}
%!%-
{
  variable r, g, b;

  (r, g, b) = array_map(Double_Type, Double_Type, Double_Type, &map_hsl, h, s, l);

  return length(r) == 1 ? (int(r*255)[0], int(g*255)[0], int(b*255)[0]) :
    (int(r*255), int(g*255), int(b*255));
}

private define map_rhsl (ma, mi, l, r, g, b)
{
  if (ma == mi)
    return 0.0, 0.0, l;

  variable d = ma - mi;
  variable s = l > 0.5 ? d/(2 - ma - mi) : d/(ma+mi);
  variable h;

  switch (ma)
  { case r: h = (g - b)/d + (g < b ? 6.0 : 0.0); }
  { case g: h = (b - r)/d + 2.0; }
  { case b: h = (r - g)/d + 4.0; }

  return h, s, l;
}

define rgb2hsl(r, g, b)
%!%+
%\function{rgb2hsl}
%\synopsis{converts (red, green, blue) values to (hue, saturation, lightness)}
%\usage{(Double_Type h, s, l) = rgb2hsl(Integer_Type r, g, b);}
%\description
%    Converts 8 bit RGB values (0-255) to HSL values.
%
%    The return values are always normalized to [0, 1].
%    See \code{hsl2rgb} for a definition of hue, saturation, lightness.
%\seealso{hsl2rgb}
%!%-
{
  r = r/255.;
  g = g/255.;
  b = b/255.;

  variable ma = _max(r, g, b);
  variable mi = _min(r, g, b);
  variable h, s, l;

  l = (ma+mi)/2.0;

  (h,s,l) = array_map(Double_Type, Double_Type, Double_Type, &map_rhsl, ma, mi, l, r, g, b);

  return length(h) == 1 ? (h[0],s[0],l[0]) : (h, s, l);
}

%
% Code from src/slang/color/hsluv2rgb.sl
%
% Stolen from the HSLUV project, C implementation

private variable Triplet_m = {
  [3.24096994190452134377, -1.53738317757009345794, -0.49861076029300328366],
  [-0.96924363628087982613,  1.87596750150772066772,  0.04155505740717561247],
  [0.05563007969699360846, -0.20397695888897656435,  1.05697151424287856072]
};

private variable Triplet_minv = {
  [0.41239079926595948129,  0.35758433938387796373,  0.18048078840183428751],
  [0.21263900587151035754,  0.71516867876775592746,  0.07219231536073371500],
  [0.01933081871559185069,  0.11919477979462598791,  0.95053215224966058086]
};

private variable Ref_U = 0.19783000664283680764;
private variable Ref_V = 0.46831999493879100370;

private variable Kappa = 903.29629629629629629630;
private variable Epsilon = 0.00885645167903563082;

private define get_bounds (l)
{
  variable tl = l+16.0;
  variable sub1 = (tl*tl*tl)/1560896.0;
  variable sub2 = (sub1 > Epsilon ? sub1 : (l / Kappa));

  variable bounds = Double_Type[6,2];

  variable i,j;
  _for i (0, 2)
  {
    variable m1 = Triplet_m[i][0];
    variable m2 = Triplet_m[i][1];
    variable m3 = Triplet_m[i][2];

    _for j (0, 1)
    {
      variable top1 = (284517.0 * m1 - 94839.0 * m3) * sub2;
      variable top2 = (838422.0 * m3 + 769860.0 * m2 + 731718.0 * m1) * l * sub2 -  769860.0 * j * l;
      variable bottom = (632260.0 * m3 - 126452.0 * m2) * sub2 + 126452.0 * j;

      bounds[i*2 + j, 0] = top1/bottom;
      bounds[i*2 + j, 1] = top2/bottom;
    }
  }

  return bounds;
}

private define intersect_line_line (line1, line2)
{
  return (line1[1] - line2[1])/(line2[0] - line1[0]);
}

private define ray_length_until_intersect (theta, line)
{
  return line[1]/(sin(theta)-line[0]*cos(theta));
}

private define max_safe_chroma_for_l (l)
{
  variable min_len_sqr = DOUBLE_MAX;
  variable bounds = get_bounds(l);

  variable i;
  _for i (0, 5)
  {
    variable m1 = bounds[i, 0];
    variable b1 = bounds[i, 1];

    variable line2 = [-1.0/m1, 0.0];
    variable x = intersect_line_line (bounds[i, *], line2);
    variable dist = sumsq([x, b1+x*m1]);

    if (dist < min_len_sqr)
      min_len_sqr = dist;
  }

  return sqrt(min_len_sqr);
}

private define max_chroma_for_lh (l, h)
{
  variable min_len = DOUBLE_MAX;
  variable hrad = h * 0.01745329251994329577; % (2 * pi / 360)

  variable bounds = get_bounds(l);

  variable i;
  _for i (0, 5)
  {
    variable len = ray_length_until_intersect(hrad, bounds[i,*]);

    if (len >= 0 && len < min_len)
      min_len = len;
  }

  return min_len;
}

private define from_linear (c)
{
  if (c <= 0.0031308)
    return 12.92 * c;
  else
    return 1.055 * c^(1.0/2.4) - 0.055;
}

private define to_linear (c)
{
  if (c > 0.04045)
    return ((c + 0.055)/1.055)^2.4;
  else
    return c / 12.92;
}

private define xyz2rgb (x, y, z)
{
  variable r = from_linear(sum(Triplet_m[0]*[x,y,z]));
  variable g = from_linear(sum(Triplet_m[1]*[x,y,z]));
  variable b = from_linear(sum(Triplet_m[2]*[x,y,z]));

  return r,g,b;
}

private define rgb2xyz (r, g, b)
{
  variable rgbl = array_map(Double_Type, &to_linear, [r,g,b]);
  variable x, y, z;
  x = sum(Triplet_minv[0]*rgbl);
  y = sum(Triplet_minv[1]*rgbl);
  z = sum(Triplet_minv[2]*rgbl);

  return x, y, z;
}

private define y2l (y)
{
  if (y <= Epsilon)
    return y*Kappa;
  else
    return 116.0*y^(1./3) - 16.0;
}

private define l2y (l)
{
  if (l <= 8.0)
    return l/Kappa;
  else
    return ((l+16.0)/116.0)^3;
}

private define xyz2luv (x, y, z)
{
  variable var_u = (4.0 * x) / (x + (15.0 * y) + (3.0 * z));
  variable var_v = (9.0 * y) / (x + (15.0 * y) + (3.0 * z));

  variable l = y2l(y);
  variable u = 13.0*l*(var_u - Ref_U);
  variable v = 13.0*l*(var_v - Ref_V);

  if (l < 0.00000001)
    return l, 0.0, 0.0;
  else
    return l, u, v;
}

private define luv2xyz (l, u, v)
{
  if (l <= 0.00000001)
    return 0.0, 0.0, 0.0;

  variable var_u = u/(13.0*l) + Ref_U;
  variable var_v = v/(13.0*l) + Ref_V;
  variable y = l2y(l);
  variable x = -(9.0*y*var_u)/((var_u-4.0)*var_v-var_u*var_v);
  variable z = (9.0*y-(15.0*var_v*y)-(var_v*x))/(3.0*var_v);

  return x, y, z;
}

private define luv2lch (l, u, v)
{
  variable h, c = sqrt(u*u+v*v);

  if (c < 0.00000001)
    h = 0.0;
  else
  {
    h = atan2(v, u) * 57.29577951308232; % 180. / PI
    if (h < 0)
      h += 360.;
  }

  return l, c, h;
}

private define lch2luv (l, c, h)
{
  variable hrad = h * 0.017453292519943295; % PI / 180.
  variable u = cos(hrad)*c;
  variable v = sin(hrad)*c;

  return l, u, v;
}

private define hsluv2lch (h, s, l)
{
  variable c;

  if (l > 99.9999999 || l < 0.00000001)
    c = 0.0;
  else
    c = max_chroma_for_lh(l,h)/100.0*s;

  if (s<0.00000001)
    h = 0.0;

  return l, c, h;
}

private define lch2hsluv (l, c, h)
{
  variable s;

  if (l > 99.9999999 || l < 0.00000001)
    s = 0.0;
  else
    s = c / max_chroma_for_lh(l, h) * 100.0;

  if (c < 0.00000001)
    h = 0.0;

  return h, s, l;
}

private define hpluv2lch (h, s, l)
{
  variable c;

  if(l > 99.9999999 || l < 0.00000001)
    c = 0.0;
  else
    c = max_safe_chroma_for_l(l) / 100.0 * s;

  if (s < 0.00000001)
    h = 0.0;

  return l, c, h;
}

private define lch2hpluv (l, c, h)
{
  variable s;

  if (l > 99.9999999 || l < 0.00000001)
    s = 0.0;
  else
    s = c / max_safe_chroma_for_l(l) * 100.0;

  if (c < 0.00000001)
    h = 0.0;

  return h, s, l;
}

define hsluv2rgb (h, s, l)
%!%+
%\function{hsluv2rgb}
%\usage{Int_Type r, g, b = hsluv2rgb(Double_Type h, s, l);}
%\synopsis{Calculate RGB triplet from HSLuv space}
%\description
%  This function returns the RGB colors (as 8 bit values) of
%  the corresponding HSLuv color (see https://www.hsluv.org/).
%  The inputs range from 0 to 1.
%
%  This colorpsace is useful to generate custom color maps and
%  palettes as it attemps to give the same perceived lightness (l)
%  and saturation (s) for different hues (h). It does not contain
%  the full RGB space, however.
%
%\seealso{rgb2hsluv}
%!%-
{
  variable A, B, C;
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &hsluv2lch, h*360.,s*100.,l*100.);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &lch2luv, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &luv2xyz, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &xyz2rgb, A, B, C);

  (A, B, C) = (int(_max(0, _min(A*255, 255))), int(_max(0, _min(B*255, 255))), int(_max(0, _min(C*255, 255))));

  return length(A) == 1 ? (A[0], B[0], C[0]) : (A, B, C);
}

define hpluv2rgb (h, s, l)
%!%+
%\function{hpluv2rgb}
%\usage{Int_Type r, g, b = hpluv2rgb(Double_Type h, s, l);}
%\synopsis{Calculate RGB triplet from HPLuv space}
%\description
%  This function returns the RGB colors (as 8 bit values) of
%  the corresponding HPLuv color (see https://www.hsluv.org/).
%  The inputs range from 0 to 1.
%
%  This colorpsace is useful to generate custom color maps and
%  palettes as it attemps to give the same perceived lightness (l)
%  and saturation (s) for different hues (h). Compared to HSLuv
%  it tries to maximize the used color range.
%
%\seealso{rgb2hpluv, hsluv2rgb}
%!%-
{
  variable A, B, C;
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &hpluv2lch, h*360.,s*100.,l*100.);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &lch2luv, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &luv2xyz, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &xyz2rgb, A, B, C);

  (A, B, C) = (int(_max(0, _min(A*255, 255))), int(_max(0, _min(B*255, 255))), int(_max(0, _min(C*255, 255))));

  return length(A) == 1 ? (A[0], B[0], C[0]) : (A, B, C);
}

define rgb2hsluv (r, g, b)
%!%+
%\function{rgb2hsluv}
%\usage{Double_Type h, s, l = rgb2hsluv(Int_Type r, g, b);}
%\synopsis{Calculate HSLuv triplet from RGB space}
%\description
%  Inverse function of \code{hsluv2rgb}.
%
%\seealso{hsluv2rgb}
%!%-
{
  variable A, B, C;
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &rgb2xyz, r/255., g/255., b/255.);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &xyz2luv, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &luv2lch, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &lch2hsluv, A, B, C);

  (A, B, C) = (_max(0, _min(A/360., 1)), _max(0, _min(B/100., 1)), _max(0, _min(C/100., 1)));

  return length(A) == 1 ? (A[0], B[0], C[0]) : (A, B, C);
}

define  rgb2hpluv (r, g, b)
%!%+
%\function{rgb2hpluv}
%\usage{Double_Type h, s, l = rgb2hpluv(Int_Type r, g, b);}
%\synopsis{Calculate HSPuv triplet from RGB space}
%\description
%  Inverse function of \code{hpluv2rgb}.
%
%\seealso{hpluv2rgb}
%!%-
{
  variable A, B, C;
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &rgb2xyz, r/255., g/255., b/255.);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &xyz2luv, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &luv2lch, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &lch2hpluv, A, B, C);

  (A, B, C) = (_max(0, _min(A/360., 1)), _max(0, _min(B/100., 1)), _max(0, _min(C/100., 1)));

  return length(A) == 1 ? (A[0], B[0], C[0]) : (A, B, C);
}

%
% Code from src/slang/color/hsv2rgb.sl
%
private define map_hue (i,v,q,p,t)
{
  switch (i mod 6)
  { case 0: return v, t, p; }
  { case 1: return q, v, p; }
  { case 2: return p, v, t; }
  { case 3: return p, q, v; }
  { case 4: return t, p, v; }
  { case 5: return v, p, q; }
}

define hsv2rgb (h, s, v)
%!%+
%\function{hsv2rgb}
%\synopsis{Convert HSV color to RGB}
%\usage{(Int_Type r,g,b) = hsv2rgb(Double_Type h,s,v);}
%\description
%   HSV (Hue, Saturation, Value) as doubles (0-1) are
%   converted to RGB values (defined as [0-255] 8 bit
%   values). Works for array values two (lengths must
%   match).
%!%-
{
  variable i = int(floor(h*6));
  variable f = h*6-i;
  variable p = v*(1-s);
  variable q = v*(1-f*s);
  variable t = v*(1 - (1 - f)*s);

  variable r,g,b;
  (r,g,b) = array_map(Double_Type,Double_Type,Double_Type,&map_hue,i,v,q,p,t);

  return length(r) == 1 ? (int(r*255)[0], int(g*255)[0], int(b*255)[0]) :
    (int(r*255), int(g*255), int(b*255));
}

private define map_max (ma, mi, r, g, b, d)
{
  variable h;
  if (ma == mi)
    return 0.0;
  else
  {
    switch (ma)
    { case r: h = (g-b)/d + (g < b ? 6.0 : 0.0); }
    { case g: h = (b-r)/d + 2.0; }
    { case b: h = (r-g)/d + 4.0; }
  }

  return h/6.0;
}

define rgb2hsv (r, g, b)
%!%+
%\function{rgb2hsv}
%\synopsis{Convert RGB color to HSV}
%\usage{(Double_Type h,s,v) = rgb2hsv(Int_Type r,g,b);}
%\description
%   RGB colores encoded as 8 bit (0-255) values get
%   converted to HSV (Hue, Saturation, Value) as doubles
%   (0-1) are. Works for array values two (lengths must
%   match).
%!%-
{
  r = r/255.;
  g = g/255.;
  b = b/255.;

  variable ma = _max(r, g, b);
  variable mi = _min(r, g, b);
  variable h, s, v;

  v = ma;

  variable d = ma - mi;

  s = ma == 0.0 ? 0 : d/ma;

  return length(r) == 1 ? (array_map(Double_Type, &map_max, ma, mi, r, g, b, d)[0], s[0], v[0]) :
    (array_map(Double_Type, &map_max, ma, mi, r, g, b, d), s, v);
}
 
%
% Code from src/data/spectrum.datastructures/ignore_large_bins.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define ignore_large_bins()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ignore_large_bins}
%\synopsis{ignores those bins of a spectral dataset exceeding a maximal size}
%\usage{ignore_large_bis(Integer_Type id[], Double_Type maxsize);}
%\qualifiers{
%\qualifier{verbose}{}
%\qualifier{unit}{[=\code{"A"}]: maxsize may be in A or keV}
%}
%!%-
{
  variable id, maxsize;
  switch(_NARGS)
  { case 2: (id, maxsize) = (); }
  { help(_function_name()); }

  variable verbose = qualifier_exists("verbose");
  variable unit = qualifier("unit", "A");
  variable strlow_unit = strlow(unit);
  ifnot(any(strlow_unit==["a", "kev"]))
    return vmessage(`error (%s): unit="%s" is not allowed, only "A" or "keV"`, _function_name(), unit);
  unit = (strlow_unit=="kev");

  foreach id ([id])
  {
    variable d = get_data_counts(id);
    variable i, n = length(d.value);
    if(unit)
      (d.bin_lo, d.bin_hi) = (_A(1)/d.bin_hi, _A(1)/d.bin_lo);  % do not reverse arrays

    foreach i (get_data_info(id).notice_list)
      if(d.bin_hi[i]-d.bin_lo[i] > maxsize)
      {
	ignore_list(id, i);
	if(verbose)
	  vmessage("ignoring (wavelength-bin %4d) = (energy-bin %4d) @ [%6.2f, %6.2f] %s",
		   i, n-1-i, d.bin_lo[i], d.bin_hi[i], unit ? "keV" : "A"
		  );
      }
  }
}

%
% Code from src/plot/image2rgb.sl
%
define image2rgb()
%!%+
%\function{image2rgb}
%\synopsis{converts an image (2d array) to a 24bit RGB image}
%\usage{Integer_Type rgb[] = image2rgb(Double_Type img[]);
%\altusage{Integer_Type rgb[] = image2rgb(Double_Type R[], G[], B[]);}
%}
%\description
%    \code{min(img)} will be mapped to black, \code{max(img)} to white,
%    and other values to their linear gray scale.\n
%    This function can be used for \code{png_write}(\code{_flipped}).
%\seealso{png_write}
%!%-
{
  variable R, G, B;

  switch(_NARGS)
  { case 1: G = (); }
  { case 3: (R, G, B) = (); }
  { help(_function_name()); return; }

  variable G_0_255 = int( 255.9999999999999 * (G - min(G)) / (max(G)-min(G)) );
  if(_NARGS==1)
    return typecast(G_0_255, Char_Type);
  else
  { variable R_0_255 = int( 255.9999999999999 * (R - min(R)) / (max(R)-min(R)) );
    variable B_0_255 = int( 255.9999999999999 * (B - min(B)) / (max(B)-min(B)) );
    return (R_0_255 << 16) | (G_0_255 << 8) | B_0_255;
  }
}

%
% Code from src/slang/arrays/2d/indices_of_2Darray.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%
define indices_of_2Darray(field, id)
%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable w = length(field[0,*]);
  return (id/w, id mod w);
}

%
% Code from src/fitting/fit-functions/init_apec.sl
%
#ifeval getenv("ATOMDB")!=NULL
define init_Apec ()
{
   plasma(aped);
   create_aped_fun("Apec",default_plasma_state);
}
#endif
%
% Code from src/slang/structs/histograms/init_histo.sl
%
define init_histo(){
%!%+
%\function{init_histo}
%\synopsis{initialize a struct with histogram fields}
%\usage{init_histo([Integer_Type len]);}
%\description
%     Return a new struct with the typical fields of a histogram:
%     bin_lo, bin_hi, value, err. If the optional argument is present,
%     the fields are initialized with a Double_Type array of that
%     length. Else the fields are empty. 
%\seealso{read_histo, init_histo, add_hist, shift_hist, 
%           scale_hist, stretch_hist}
%!%-   
   variable s = struct{bin_lo,bin_hi,value,err};
   if(_NARGS==0) { return @s; }   
   else if(_NARGS==1){ 
	  variable len = ();
	  s.bin_lo = Double_Type[len];
	  s.bin_hi = Double_Type[len];
	  s.value = Double_Type[len];
	  s.err = Double_Type[len];
	  return @s;
   }
   else  { help(_function_name()); return; }
}

%
% Code from src/misc/radio/init_jet_fit.sl
%
define init_jet_fit()
%!%+
%\function{init_jet_fit}
%\synopsis{inititialize the fit function for the speed of jet components}
%\usage{init_jet_fit (Ref_Type jet_component_structure);}
%\qualifiers{
%\qualifier{pa}{[=90] position angle of the jet in degrees (required)}
%\qualifier{recalc_distance}{set to 1 in order to recalculate the distance
%                 of each component based on deltax and deltay}
%}
%\description
%    This function initializes the fit function \code{jet_speed}, which
%    can fit a linear function of the form:
%        dist (t)  = (t - t_0) * v
%    to each component, where t_0 is the ejection time of the component in
%    MJD and v is its speed in mas/year. For each epoch an offset (in mas)
%    can be fitted, allowing to correct for shifts of the image (along the
%    line in which the distance is measured).
%    For the initialization a structure with the name "jet_speed_struct"
%    is REQUIRED, which has to contain the fields \code{time} (date of epoch in MJD),
%    \code{distance} (in mas),  \code{derr} (uncertainty of the distance), and
%    \code{component} (integer values identifying the components).
%    The function extends this structure, and sets the initialized fit function.
%\example
%    variable jet_speed_struct = struct {
%      mjd       = [55500, 55550, 55600, 55650, 55500, 55600, 55650],
%      deltax    = [  0.2,   0.4,   0.5,   0.7,   0.6,   1.2,   1.5],
%      deltay    = [  0.0,   0.0,   0.0,   0.0,   0.0,   0.0,   0.0],
%      derr      = [ 0.01,  0.05,  0.01,   0.1,  0.02,   0.1,   0.1],
%      component = [    1,     1,     1,     1,     2,     2,     2],
%      };
%    init_jet_fit(&jet_speed_struct; pa = 90);
%    ()=fit_counts;
%    list_par;
%    plot_jet_speed ( jet_speed_struct );
%\seealso{plot_jet_speed}
%!%-
%
{
  % use a pointer to the jet_speed_struct, this allows the function to
  % change the structure permanently (add fields)
  variable jet_speed_struct_pointer;
  variable pa = qualifier("pa",90.);
  variable recalc_dist = qualifier("recalc_distance",0);
  
  switch(_NARGS)
  { case 1: jet_speed_struct_pointer = ();}
  { help(_function_name()); return; }

  % check if required fields are available and add further required ones
  variable req_fields = ["mjd", "derr", "comp", "deltax", "deltay"];
  variable fn         = get_struct_field_names ( @jet_speed_struct_pointer );
  variable f;
  foreach f (req_fields) if (all ( f != fn)) {
    vmessage("error: %s\tprovided structure does not include field: %s", _function_name, f);
    return;
  }
  variable length_sp_str = length((@jet_speed_struct_pointer).mjd);
  variable used_fields   = ["xshift","yshift","t0", "speed"];
  foreach f (used_fields) if (all ( f != fn)) {
    @jet_speed_struct_pointer = struct_combine (@jet_speed_struct_pointer, f);
    set_struct_field (@jet_speed_struct_pointer, f, Double_Type[length_sp_str]);
%    vmessage("%s: Adding field %s to jet_speed_struct",_function_name(),f);
  }
  if (all ( "distance" != fn) or recalc_dist) { @jet_speed_struct_pointer =
      struct_combine (@jet_speed_struct_pointer, struct{distance = hypot((@jet_speed_struct_pointer).deltax, (@jet_speed_struct_pointer).deltay)}); }
  % if (all ( "x" != fn)) { @jet_speed_struct_pointer =
  %     struct_combine (@jet_speed_struct_pointer, struct{x =get_struct_field (@jet_speed_struct_pointer,"distance")*sin(pa*PI/180) }); }
  % if (all ( "y" != fn)) { @jet_speed_struct_pointer =
  %     struct_combine (@jet_speed_struct_pointer, struct{y =get_struct_field (@jet_speed_struct_pointer,"distance")*cos(pa*PI/180) }); }

  % check if required structure is readable in the namespace for the function
  % MB: this could probably improved. If the variable is defined within a script
  %     then jss_status=-2 and it is seen by the function, if defined in the
  %     interactive shell jss_status=0 and it works in namespace "isis".
  %     Global variables???
  variable jss_status = is_defined("jet_speed_struct");
  if (jss_status == 0){
    try { eval("if(typeof(jet_speed_struct)==Struct_Type and Fit_Verbose == 1) message(\"jet_speed_struct initialized correctly\");","isis");}
    catch AnyError: { message("error: init_jet_fit!\n\trequired structure with name jet_speed_struct not initilized"); return; }
    ; % to terminate the try-block
  }

  % initialize the fit function
  variable def_str = `define jet_speed_fit (bin_lo, bin_hi, par) % "private define" causes namespace problem?!
{
  variable pa = par[0];
  variable sin_pa = sin(pa*PI/180);
  variable cos_pa = cos(pa*PI/180);
  variable t_indx   = unique( jet_speed_struct.epoch );
  variable n_epochs = length (t_indx);
  variable used_cmp = where(jet_speed_struct.comp > 0);
  variable cmpindx  = used_cmp[unique(jet_speed_struct.comp[used_cmp])];
  variable n_comps = length ( cmpindx );
  variable i,l;
  _for i(0, n_epochs-1, 1) {jet_speed_struct.xshift[where(jet_speed_struct.epoch == jet_speed_struct.epoch[t_indx[i]])] =
                            sin_pa*par[2*i+1] + cos_pa*par[2*i+2];}
  _for i(0, n_epochs-1, 1) {jet_speed_struct.yshift[where(jet_speed_struct.epoch == jet_speed_struct.epoch[t_indx[i]])] =
                            cos_pa*par[2*i+1] - sin_pa*par[2*i+2];}
  _for i(0, n_comps-1, 1)
    {
      l = where(jet_speed_struct.comp == jet_speed_struct.comp[cmpindx[i]]);
      jet_speed_struct.speed[l] = par[1+2*n_epochs+ 2*i +1];
      jet_speed_struct.t0   [l] = par[1+2*n_epochs+ 2*i ];
    }
  variable dist = (jet_speed_struct.mjd - jet_speed_struct.t0)*0.002737803091986241*jet_speed_struct.speed ;
  variable ret = hypot( sin_pa*dist + jet_speed_struct.xshift, cos_pa*dist + jet_speed_struct.yshift)*
         sign(atan2(sin_pa*dist + jet_speed_struct.xshift, cos_pa*dist + jet_speed_struct.yshift));
  return ret[used_cmp];
  % 0.002737803091986241 is one day in units of sidereal years
}

private define jet_speed_default(i)
{ switch(i)
`;
  variable sin_pa = sin(pa*PI/180);
  variable cos_pa = cos(pa*PI/180);
  variable n_epochs = length (unique( (@jet_speed_struct_pointer).epoch));
  variable epindx  = unique((@jet_speed_struct_pointer).epoch);
  epindx = epindx[array_sort ( (@jet_speed_struct_pointer).epoch[epindx] )];
  variable used_cmp = where((@jet_speed_struct_pointer).comp > 0);
  variable cmpindx  = used_cmp[unique((@jet_speed_struct_pointer).comp[used_cmp])];
  variable n_comps = length ( cmpindx );
  variable j;
  def_str = def_str + sprintf(" { case 0: return (%.6f , 1, -180,  180 ); }\n",pa);
  _for j(0,n_epochs-1,1) def_str = def_str + sprintf(" { case %d: return (%.4f , 1, -10,  10 ); }\n { case %d: return (%.4f , 1, -10,  10 ); }\n",
						     2*j+1, sin_pa*(@jet_speed_struct_pointer).xshift[epindx[j]] + cos_pa*(@jet_speed_struct_pointer).yshift[epindx[j]],
						     2*j+2, -cos_pa*(@jet_speed_struct_pointer).xshift[epindx[j]] + sin_pa*(@jet_speed_struct_pointer).yshift[epindx[j]]);
  _for j(0,n_comps-1,1) def_str = def_str +
    sprintf(" { case %d: return (5e4 , 0, 4e4, 7e4 ); }\n { case %d: return (0.0 , 0, -20,  20 ); }\n",j*2 + 2*n_epochs+1,j*2 + 2*n_epochs+2);
  def_str = def_str+
    `}

add_slang_function("jet_speed", ["pos_angle [deg]",
`;
  _for j(0,n_epochs-1,1) def_str = def_str + sprintf(" \"shft_ep_%s [mas]\",\n \"shftorth_ep_%s [mas]\",\n",
						     string((@jet_speed_struct_pointer).epoch[epindx[j]]),dup);
  _for j(0,n_comps-1,1) def_str = def_str +
    sprintf(" \"t0_%s [MJD]\",\n \"v_%s [mas/yr]\",\n",string((@jet_speed_struct_pointer).comp[cmpindx[j]]),dup);
  def_str = def_str+
    `  ]);
set_param_default_hook("jet_speed", "jet_speed_default");
`;

  if (jss_status == 0) eval(def_str, "isis");
  else eval(def_str);

  % define the "counts"
  variable jss_id =define_counts ([1:length((@jet_speed_struct_pointer).comp)], dup+1,
				  (@jet_speed_struct_pointer).distance, (@jet_speed_struct_pointer).derr);
  ignore(jss_id);
  notice_list(jss_id , used_cmp);

  fit_fun("jet_speed(1)");
}

%
% Code from src/plot/init_plot.sl
%
%%%%%%%%%%%%%%%%
define init_plot()
%%%%%%%%%%%%%%%%
%!%+
%\function{init_plot}
%\synopsis{starts a new plot, such that other plots can be overplotted}
%\usage{init_plot();}
%\description
%    plot is used to plot a single point in the lower left corner.
%    Before init_plot is called, x- and yrange have to be set.
%\seealso{plot, get_plot_options}
%!%-
{
  variable plot_options = get_plot_options();
  if(plot_options.xmin<-3e38 || plot_options.xmax>3e38 || plot_options.ymin<-3e38 || plot_options.ymax>3e38)
     return vmessage(`error (%s): The ranges are not set, see  help("%s");`,
		     _function_name(), dup);

  connect_points(-1);
  plot(plot_options.xmin, plot_options.ymin);
  connect_points(plot_options.connect_points);
}

%
% Code from src/slang/date/init_time_structure.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define init_time_structure()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{init_time_structure}
%\synopsis{creates a time structure}
%\usage{Struct_Type init_time_structure(Integer_Type Y[, m[, d[, H[, M[, S]]]]])}
%\description
%    The \code{tm_year} field of the returned structure
%    is \code{Y-1900}, and \code{tm_mon} is \code{m-1}.
%\seealso{localtime, mktime, strftime}
%!%-
{
  variable Y, m=1, d=1, H=0, M=0, S=0;
  switch(_NARGS)
  { case 1: Y = (); }
  { case 2: (Y, m) = (); }
  { case 3: (Y, m, d) = (); }
  { case 4: (Y, m, d, H) = (); }
  { case 5: (Y, m, d, H, M) = (); }
  { case 6: (Y, m, d, H, M, S) = (); }
  { help(_function_name()); return; }

  return localtime(mktime(struct {
    tm_sec   = S,
    tm_min   = M,
    tm_hour  = H,
    tm_mday  = d,
    tm_mon   = m-1,
    tm_year  = Y-1900,
    tm_wday  = 0,      % will be calculated by localtime
    tm_yday  = 0,      % will be calculated by localtime
    tm_isdst = -1      % will be calculated by localtime
  }));
}

%
% Code from src/slang/arrays/inner_product.sl
%
%%%%%%%%%%%%%%%%%
define inner_product( A, B )
%%%%%%%%%%%%%%%%%
%!%+
%\function{inner_product}
%\synopsis{}
%\usage{ M[n,...,j] = inner_product( A[n,...,m], B[m,...,j] );}
%\description
%    In ISIS the operator # calculates the inner product of two arrays
%    A, B, i.e., contracts the last and first dimension, respectively.
%    A and B can have an arbitrary number of dimensions, it is just required
%    that the last dimension of A and first dimension of B is of the same
%    size!
%\example
%    A = _reshape( [1:2*3], [2,3] );
%    B = _reshape( [1:3*4], [3,4] );
%    M = inner_product( A, B );
%    N = A # B;
%    vmessage("M\n"); print(M);
%    vmessage("N\n"); print(N);
%\seealso{operator #}
%!%-
{
  return A # B;
}

%%%%%%%%%%%%%%%%%
define inner_products()
%%%%%%%%%%%%%%%%%
%!%+
%\function{inner_products}
%\synopsis{}
%\usage{ M[n,...,l] = inner_products( A[n,...,m], B[m,...,j] [, ..., Z[k,...,l] ] );}
%\description
%    In ISIS the operator # calculates the inner product of two or more arrays.
%    This function is basically nesting the inner_product function!
%\seealso{inner_product, operator #}
%!%-
{
  ifnot( _NARGS ){ help(_function_name); return; }
  
  variable args = __pop_args(_NARGS);
  
  variable IP = __push_args( args[0] );
  variable i;
  _for i ( 1, _NARGS-1 ){
    IP = inner_product( IP, __push_args( args[i] ) );
  }
  return IP;
}

%
% Code from src/misc/satellites/INTEGRAL/integral_cat2ds9.sl
%
define integral_cat2ds9(inputfile,outputfile)
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{integral_cat2ds9}
%\synopsis{a cat2ds9 like function for INTEGRAL catalogs}
%\usage{ ()=integral_cat2ds9(String_Type inputfile,
%                      String_Type outputfile)}
%\qualifiers{                     
%\qualifier{source_name}{name of the column with source names
%                        (default: NAME)}
%\qualifier{ra_name}{name of the column with right ascension of the
%                      sources (default: RA_OBJ)}
%\qualifier{dec_name}{:name of the column with declination of the
%                      sources (default: DEC_OBJ)}
%\qualifier{color}{color to be used for the ds9 region (default:
%                    black)}
%\qualifier{symbol}{form of the ds9 region (default:box)}
%\qualifier{noname}{only boxes, no text with source names}
%}
%!%-
{
  %open fits file
  variable cf = fits_open_file(inputfile+"+1","r");

  %qualifiers and default values when no qualifier set
  variable ra_name     = qualifier("ra_name","RA_OBJ");
  variable dec_name    = qualifier("dec_name","DEC_OBJ");
  variable source_name = qualifier("source_name","NAME");
  variable color       = qualifier("color","black");
  variable symbol      = qualifier("symbol","box");
  
  %read variable from file
  variable source=fits_read_col(cf,source_name);
  variable ra = fits_read_col(cf,ra_name);
  variable dec = fits_read_col(cf,dec_name);
  %close fits file
  fits_close_file(cf);
  
  %open output file for writing, write region data
  variable pf=fopen(outputfile,"w");
  ()=fprintf(pf,"global move=0 \n");
  ()=fprintf(pf,"global color=%s \n",color);
  variable i;
  for (i=0;i<length(source)-1;i++){
    if (qualifier_exists("noname")){
      ()=fprintf(pf,"fk5;point(%f,%f) # point=%s \n",ra[i],dec[i],symbol);
    }
    else{
      ()=fprintf(pf,"fk5;point(%f,%f) # point=%s text={%s} \n",ra[i],dec[i],symbol,source[i]);
    }
  }
  ()=fclose(pf);
  

  return(0);
}


%
% Code from src/slang/math/integrate2d.sl
%
% -*- mode: slang; mode: fold -*-

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% integrate2d calculates 2-dimensional integrals of the form:
%%%
%%% I = int_x1^x2 int_y1(x)^y2(x) func(x,y) dx dy
%%%
%%% The solution is adopted from 'Numerical Recipes in C' Cap. 4.6
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Global variables:
private variable xsav;
private variable nrfunc;

%%% Inner function call: f2(y) = func(xsav,y)
private define f2(y){ %{{{
  return (@nrfunc)( xsav, y ;; __qualifiers );
}
%}}}

%%% Outer function call: f1(x) = func(x,&y(x));
private define f1(x){ %{{{
  xsav = x;
  return (@__qualifiers.intfun)( &f2,
				 (@__qualifiers.y1)(x;;__qualifiers),
				 (@__qualifiers.y2)(x;;__qualifiers)
				 ;;
				 __qualifiers
			       );
}
%}}}


define integrate2d( func, y1, y2, x1, x2 ) %{{{
%!%+
%\function{integrate2d}
%\synopsis{numerical computation of an 2-dim. integral}
%\usage{Double_Type I = integrate2d( Ref_Type func, y1, y2,
%                                    Double_Type x1, x2
%                                  );
%}
%\qualifiers{
%\qualifier{intfun[&qromb]:}{Reference to the integrator}
%\qualifier{NOTE}{Qualifiers are passed to all sub-functions!}
%}
%\description
%    This function numerically computes the 2-dimensional integral 'I' over the
%    integrant-function 'func' of the form:
%
%       I = int_x1^x2 int_y1(x)^y2(x) func(x,y) dx dy
%
%   The given limits 'x1' and 'x2' are the lower and upper integration limits
%   for the 'x' variable, respectively. 'y1' and 'y2', on the other hand, are
%   references to functions, which calculates the lower and upper limit y1(x)
%   and y2(x) for the 'y' variable depending on 'x'.
%
%   The implemented solution is adopted from the 'Numerical Recipes in C'
%   Chap. 4.6. !
%   
%   The integrator can be changed by the qualifier 'intfun', which is set to
%   'qromb' as default (see 'help qromb). Qualifiers given to 'integrate2d'
%   are passed to the 'intfun' function!
%
%\example
%   % Calculation of the area of an rectangle:
%   %    I = int_0^1 dx int_0^2 dy
%   %      = [ x ]_0^1 * [ y ]_0^2
%   %      = 2
%   
%   variable x1 = 0., x2 = 1.;
%   define y1(x){ return 0.; };
%   define y2(x){ return 2.; };
%   define func( x, y ){ return 1.; };
%   variable I = integrate2d( &func, &y1, &y2, x1, x2 );
%   vmessage("Area of rectangle is A = %g",I);
%
%   % Calculation of the area of the unit circle in cartesian coordinates:
%   %   I = int_-1^1 int_-sqrt(1-sqr(x))^sqrt(1-sqr(x)) dx dy
%   %     = int_-1^1 2 * sqrt(1-sqr(x)) dx
%   
%   define tfun( x, y ){ return 1; }
%   define y1(x){ return -sqrt(1-sqr(x)); }
%   define y2(x){ return  sqrt(1-sqr(x)); }
%   variable I = integrate2d( &tfun, &y1, &y2, -1., 1. ; qromb_max_iter=24, qromb_eps=1e-3 );
%   vmessage("Area of unite circle is A = %g, abs(A-PI)/PI = %g", I, abs(I-PI)/PI );
%   
%%\seealso{qromb, qsimp, integrate2d_test}
%!%-
{
  variable intfun = qualifier("intfun",&qromb);
  variable qualis = struct_combine(__qualifiers,struct{intfun=intfun,y1=y1,y2=y2} );
  nrfunc = func;
  return (@intfun)( &f1, x1, x2 ;; qualis );
}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FUNCTION TEST:
%%%  %{{{
private define tfun( x, y ){
  return 1.;
}
private define y1(x){
  return -sqrt(1-sqr(x));
}
private define y2(x){
  return sqrt(1-sqr(x));
}
define integrate2d_test()

{
  variable lim = qualifier("lim",1e-3);
  variable Acirc = integrate2d( &tfun, &y1, &y2, -1., 1. ; intfun=&qromb, qromb_max_iter=24, qromb_eps=lim*1e-1 );
  variable reldif = abs( Acirc - PI )/PI;
  if( reldif < lim ){
    return 1;
  }
  else{
    return 0;
  }
}

%}}}

%
% Code from src/slang/math/integrateAB5.sl
%

define integrateAB5()
%!%+
%\function{integrateAB5}
%\synopsis{integrates an ordinary differential equation with the 5th order Adams-Bashforth algorithm}
%\usage{x = integrateAB5(&f, t1, t2, dt[, x0]);}
%\description
%    The integral of the ordinary differential equation\n
%      dx/dt = f(x(t), t)  with  x(t0) = x0\n
%    reads\n
%      x(t) = x0 + int_{t0}^{t} f(x(t'), t') dt'
%
%    \code{&f} is a reference to a function with two arguments:
%       \code{define f(x, t)}\n
%       \code{\{}\n
%       \code{  return ...;}\n
%       \code{\}}
%
%    \code{x} may be be a scalar as well as an array.
%\seealso{integrateRK4}
%!%-
{
  variable f, t1, t2, dt, x0=0.;
  switch(_NARGS)
  { case 4: (f, t1, t2, dt    ) = (); }
  { case 5: (f, t1, t2, dt, x0) = (); }
  { help(_function_name()); return; }

  variable i, n = nint((t2-t1)*1./dt);
  if(n<5)  n = 5;
  dt = (t2-t1)*1./n;  % adjusting dt to have exactly n steps from t1 to t2

  if(__is_numeric(x0)==1)  x0 *= 1.;   % Integer_Type => Double_Type
  variable x = typeof(x0)[5]; x[0] = x0;
  _for i (0, 3, 1)
    x[i+1] = integrateRK4(f, t1+i*dt, t1+(i+1)*dt, dt, x[i];; __qualifiers);
  variable i0=0, i1=1, i2=2, i3=3, i4=4;

  _for i (0, n-5, 1)
  {
    x[i0] = x[i4]                                                     % i0 = (i+5) mod 5
            + dt*( 1901/720. * @f(x[i4], t1+(i+4)*dt;; __qualifiers)  % i4 = (i+4) mod 5
                  -1387/360. * @f(x[i3], t1+(i+3)*dt;; __qualifiers)  % i3 = (i+3) mod 5
                  + 109/ 30. * @f(x[i2], t1+(i+2)*dt;; __qualifiers)  % i2 = (i+2) mod 5
                  - 637/360. * @f(x[i1], t1+(i+1)*dt;; __qualifiers)  % i1 = (i+1) mod 5
                  + 251/720. * @f(x[i0], t1+ i   *dt;; __qualifiers)  % i0 =  i    mod 5
	         );
    (i0, i1, i2, i3, i4) = (i1, i2, i3, i4, i0);                      % i++
  }
  return x[i4];  % == x[(i+5) mod 5]  % (The i's have already been rotated.)
}

%
% Code from src/slang/math/integrateGK.sl
%
% -*- mode: slang; mode: fold -*- %

% Gauss Kronrod G7K15 nodes and weights
private variable GK = 
{
  % Nodes
  [0.000000000000000000000000000000000e+00, 2.077849550078984676006894037732449e-01, 4.058451513773971669066064120769615e-01, 5.860872354676911302941448382587296e-01,
   7.415311855993944398638647732807884e-01, 8.648644233597690727897127886409262e-01, 9.491079123427585245261896840478513e-01, 9.914553711208126392068546975263285e-01],
  % Kronrod weights
  [2.094821410847278280129991748917143e-01, 2.044329400752988924141619992346491e-01, 1.903505780647854099132564024210137e-01, 1.690047266392679028265834265985503e-01,
   1.406532597155259187451895905102379e-01, 1.047900103222501838398763225415180e-01, 6.309209262997855329070066318920429e-02, 2.293532201052922496373200805896959e-02],
  % Gauss weights
  [4.179591836734693877551020408163265e-01, 3.818300505051189449503697754889751e-01, 2.797053914892766679014677714237796e-01, 1.294849661688696932706114326790820e-01]
};

define integrateGK ()
%!%+
%\function{integrateGK}
%\synopsis{Integrate function with Gauss Kronrod method (G7K15)}
%\usage{Double_Type integrateGK(Ref_Typ fun, Double_Type a, b, ...);}
%\qualifiers{
%  \qualifier{max_intervals}{[=5]: Maximum number of intervals to use}
%}
%\description
%  Calculate the integral of a function \code{fun} from \code{a}
%  to \code{b}. It is expected that the first argument of \code{fun}
%  is the integration parameter. Additional arguments can be passed
%  after the integration boundaries.
%
%\example
%  variable v = integrateGK(&sin, 0, PI);
%\seealso{integrateRK4}
%!%-
{
  variable args = __pop_list(_NARGS-3);
  variable fun, a, b;
  (fun, a, b) = ();
  variable maxIntervals = qualifier("max_intervals", 5);
  variable leftBorder = Double_Type[maxIntervals];
  variable rightBorder = Double_Type[maxIntervals];
  variable errorValue = Double_Type[maxIntervals];
  variable integralValue = Double_Type[maxIntervals];
  variable integralCalculated = Int_Type[maxIntervals];
  variable totalIntegral, totalError;
  variable currentI = 0;
  variable newI = 1;
  variable lengthI = 1;
  variable max_inter = 0;
  variable m,d,x,y,intG,intK;

  leftBorder[0] = a;
  rightBorder[0] = b;
  integralCalculated[0] = 0;

  if (abs(a-b)<1e-12)
    return 1e-12*@fun(0.5*(b-a), __push_list(args));

  while (lengthI<maxIntervals)
  {
    ifnot (integralCalculated[currentI])
    {
      a = leftBorder[currentI];
      b = rightBorder[currentI];
      m = 0.5*(a+b);
      d = 0.5*(b-a);
      x = [GK[0][[::2]]*d+m, -GK[0][[2::2]]*d+m];
      y = @fun(x, __push_list(args);; __qualifiers());

      % Gauss nodes
      intK = sum(y*[GK[1][[::2]], GK[1][[2::2]]]);
      intG = sum(y*[GK[2], GK[2][[1:]]]);

      % Kronrod nodes
      x = [GK[0][[1::2]]*d+m, -GK[0][[1::2]]*d+m];
      y = @fun(x, __push_list(args);; __qualifiers());
      intK += sum(y*[GK[1][[1::2]],GK[1][[1::2]]]);

      integralValue[currentI] = intK*d;
      errorValue[currentI] = abs(intG-intK)*d;
    }

    if (currentI < newI)
    {
      totalIntegral = sum(integralValue[[currentI:newI-1]]);
      totalError = sumsq(errorValue[[currentI:newI-1]]);
    }
    else
    {
      totalIntegral = sum(integralValue[[[currentI:maxIntervals-1], [0:newI-1]]]);
      totalError = sumsq(errorValue[[[currentI:maxIntervals-1], [0:newI-1]]]);
    }

    if (max_inter<lengthI) { max_inter = lengthI; }
    if (abs(sqrt(totalError)/totalIntegral)<1e-8
	|| (abs(totalIntegral)<1e-14 && abs(totalError)<1e-16))
      return totalIntegral;

    if ( (not integralCalculated[currentI]) &&
	 ((abs(errorValue[currentI]/integralValue[currentI])>1e-8)
	  || (abs(integralValue[currentI])<1e-14 && abs(errorValue[currentI])<1e-16)))
    {
      m = 0.5*(leftBorder[currentI]+rightBorder[currentI]);
      leftBorder[newI] = leftBorder[currentI];
      rightBorder[newI] = m;
      errorValue[newI] = _Inf;

      newI++;
      if (newI==maxIntervals) newI = 0;

      leftBorder[newI] = m;
      rightBorder[newI] = rightBorder[currentI];
      errorValue[newI] = _Inf;

      newI++;
      if (newI==maxIntervals) newI = 0;

      lengthI++;
    }
    else
    {
      integralValue[newI] = integralValue[currentI];
      errorValue[newI] = errorValue[currentI];
      integralCalculated[newI] = 1;
      integralCalculated[currentI] = 0;
      newI++;

      if (newI==maxIntervals)
	newI = 0;
    }

    currentI++;
    if (currentI == maxIntervals)
      currentI = 0;
  }

  return totalIntegral;
}

%
% Code from src/slang/math/integrateRK4.sl
%
define integrateRK4()
%!%+
%\function{integrateRK4}
%\synopsis{integrates an ordinary differential equation with the 4th order Runge-Kutta algorithm}
%\usage{x = integrateRK4(&f, t1, t2, dt[, x0]);}
%\description
%    The integral of the ordinary differential equation\n
%      dx/dt = f(x(t), t)  with  x(t0) = x0\n
%    reads\n
%      x(t) = x0 + int_{t0}^{t} f(x(t'), t') dt'
%
%    \code{&f} is a reference to a function with two arguments:
%       \code{define f(x, t)}\n
%       \code{\{}\n
%       \code{  return ...;}\n
%       \code{\}}
%
%    \code{x} may be be a scalar as well as an array.
%\seealso{integrateAB5}
%!%-
{
  variable f, t1, t2, dt, x=0;
  switch(_NARGS)
  { case 4: (f, t1, t2, dt) = (); }
  { case 5: (f, t1, t2, dt, x) = (); }
  { help(_function_name()); return; }

  variable i, n = nint((t2-t1)*1./dt);
  dt = (t2-t1)*1./n;  % adjusting dt to have exactly n steps from t1 to t2
  _for i (0, n-1, 1)
  {
    variable t = t1 + i*dt;
    variable f1 = @f(x,           t       ;; __qualifiers);
    variable f2 = @f(x+0.5*dt*f1, t+0.5*dt;; __qualifiers);
    variable f3 = @f(x+0.5*dt*f2, t+0.5*dt;; __qualifiers);
    variable f4 = @f(x+    dt*f3, t+    dt;; __qualifiers);
    x += dt/6 * ( f1 + 2*f2 + 2*f3 + f4 );
  }
  return x;
}

%
% Code from src/slang/math/integrateRK45_adaptive.sl
%
define integrateRK45_adaptive()
%!%+
%\function{integrateRK45_adaptive}
%\synopsis{Integrate an ODE with an adaptive 4th/5th order Runge-Kutta algorithm}
%\usage{(t, x(t)) = integrateRK45_adaptive(&f, t0, t, [, x0]);}
%\description
%    Given an ordinary differential equation of the form\n
%      dx/dt = f(t, x(t))
%    (with \code{x} either a scalar or an array)\n
%    this function numerically computes the solution\n
%      x(t) = int_{t0}^{t} f(t', x(t')) dt' [ + x0 ]\n
%    by means of a 4th/5th order Runge-Kutta (RK) algorithm.
%
%    \code{&f} is a reference to a function with two arguments:\n
%       \code{define f(t, x)}\n
%       \code{\{}\n
%       \code{  return ...;}\n
%       \code{\}}
%
%    If x is an array, the adaptive stepsize is chosen according\n
%    to the needs of the "worst-offender" equation implying that\n
%    the desired accuracy is reached by each individual component.
%
%    Note: All qualifiers are also passed to the function f.
%\qualifiers{
%\qualifier{eps}{[\code{=1e-12}] absolut error control tolerance; lower limit: 1e-15}
%\qualifier{method}{[\code{="RKCK"}] choose among three different RK45 methods:
%      "RKF": RK-Fehlberg, "RKCK": RK-Cash-Karp, "RKDP": RK-Dormand-Prince}
%\qualifier{path}{return the entire path and not only the final result}
%\qualifier{verbose}{show intermediate "times" \code{t}}
%\qualifier{plus all qualifiers of the function f}{}
%}
%\example
%    % d/dt x(t) = -sin (t^2)*2t; -> analytical solution: x(t) = cos(t^2) for x(0)=1
%    define f(t,x)
%    {
%      return -sin(t^2)*2*t;
%    };
%    (t, x) = integrateRK45_adaptive(&f, 0, 10, 1; path);
%    plot(t,x-cos(t^2));
%
%    % d/dt [x1(t),x2(t)] = [x2(t),-x1(t)]; -> analytical solution: x1(t) = cos(t) for x1(0)=1, x2(0)=0
%    define f(t,x)
%    {
%      [x[1],-x[0]];
%    }
%    (t, x) = integrateRK45_adaptive(&f, 0, PI, [1,0]; path);
%    plot(t,x[*,0]-cos(t));
%    
%\seealso{integrateRKF45, integrateRK4, integrateAB5}
%!%-
{
  variable f, t0, t, x=0;
  switch(_NARGS)
  { case 3: (f, t0, t) = (); }
  { case 4: (f, t0, t, x) = (); }
  { help(_function_name()); return; }

  variable eps = _max( qualifier("eps", 1e-12), 1e-15 );
  variable method = qualifier("method", "RKCK");
  variable verbose = qualifier_exists("verbose");
  variable path = qualifier_exists("path");

  variable k1, k2, k3, k4, k5, k6, k7, dx_c, dx_e; % dx_c = dx continue, dx_e = dx error estimation
  variable coef = Double_Type[7,7]; % modified Butcher tableau
  if(method == "RKF")
  {
    coef[0,0]= 1./4;   coef[0,1]=  1./4;
    coef[1,0]= 3./8;   coef[1,1]=  3./32;      coef[1,2]=  9./32;
    coef[2,0]= 12./13; coef[2,1]=  1932./2197; coef[2,2]= -7200./2197;  coef[2,3]=  7296./2197;
    coef[3,0]= 1.;     coef[3,1]=  439./216;   coef[3,2]= -8.;          coef[3,3]=  3680./513;    coef[3,4]= -845./4104;
    coef[4,0]= 1./2;   coef[4,1]= -8./27;      coef[4,2]=  2.;          coef[4,3]= -3544./2565;   coef[4,4]=  1859./4104;   coef[4,5]= -11./40;
    coef[5,0]= 0;      coef[5,1]=  25./216;    coef[5,2]=  1408./2565;  coef[5,3]=  2197./4104;   coef[5,4]= -1./5;         coef[5,5]= 0; % fourth order solution
    coef[6,0]= 0;      coef[6,1]=  16./135;    coef[6,2]=  6656./12825; coef[6,3]=  28561./56430; coef[6,4]= -9./50;        coef[6,5]= 2./55; % fifth order solution
  }
  else if(method == "RKDP")
  {
    coef[0,0]= 1./5;  coef[0,1]= 1./5; 
    coef[1,0]= 3./10; coef[1,1]= 3./40;       coef[1,2]=  9./40;
    coef[2,0]= 4./5;  coef[2,1]= 44./45;      coef[2,2]= -56./15;      coef[2,3]= 32./9;
    coef[3,0]= 8./9;  coef[3,1]= 19372./6561; coef[3,2]= -25360./2187; coef[3,3]= 64448./6561; coef[3,4]= -212./729;
    coef[4,0]= 1.;    coef[4,1]= 9017./3168;  coef[4,2]= -355./33;     coef[4,3]= 46732./5247; coef[4,4]=  49./176;       coef[4,5]= -5103./18656;
    coef[5,0]= 1.;    coef[5,1]= 35./384;     coef[5,2]= 500./1113;    coef[5,3]=  125./192;   coef[5,4]= -2187./6784;    coef[5,5]= 11./84; % fourth order solution
    coef[6,0]= 0;     coef[6,1]= 5179./57600; coef[6,2]= 7571./16695;  coef[6,3]=  393./640;   coef[6,4]= -92097./339200; coef[6,5]= 187./2100;    coef[6,6]= 1./40; % fitfh order solution
  }
  else % use RKCK if method is none of the two above methods
  {
    coef[0,0]= 1./5;  coef[0,1]= 1./5;
    coef[1,0]= 3./10; coef[1,1]= 3./40;       coef[1,2]= 9./40;
    coef[2,0]= 3./5;  coef[2,1]= 3./10;       coef[2,2]= -9./10;       coef[2,3]= 6./5;
    coef[3,0]= 1.;    coef[3,1]= -11./54;     coef[3,2]= 5./2;         coef[3,3]= -70./27;       coef[3,4]= 35./27;
    coef[4,0]= 7./8;  coef[4,1]= 1631./55296; coef[4,2]= 175./512;     coef[4,3]= 575./13824;    coef[4,4]= 44275./110592; coef[4,5]= 253./4096;
    coef[5,0]= 0;     coef[5,1]= 2825./27648; coef[5,2]= 18575./48384; coef[5,3]= 13525./55296;  coef[5,4]= 277./14336;    coef[5,5]= 1./4; % fourth order solution
    coef[6,0]= 0;     coef[6,1]= 37./378;     coef[6,2]= 250./621;     coef[6,3]= 125./594;      coef[6,4]= 0;             coef[6,5]= 512./1771; % fitfh order solution    
  }
  % solving the ODEs:
  variable s = 1.*sign(t-t0); % for t<t0 make transformation: t->-t, t0->-t0, dt->-dt
  if(s==0) s=1.; % to account for t-t0==0
  t *= s; t0 *= s;
  variable dt = 1e-12*(t-t0); % intialize stepsize with arbitrary value as it is adjusted anyway
  if(path)
  {
    variable trajectory = struct{ t, x };
    trajectory.t = { s*t0 };
    trajectory.x = { x };
  }
  while(t0 < t0+dt <= t)
  {
    k1 = s * dt * @f(s*t0, x;; __qualifiers);
    k2 = s * dt * @f(s*(t0 + coef[0,0]*dt), x + coef[0,1]*k1;; __qualifiers);
    k3 = s * dt * @f(s*(t0 + coef[1,0]*dt), x + coef[1,1]*k1 + coef[1,2]*k2;; __qualifiers);
    k4 = s * dt * @f(s*(t0 + coef[2,0]*dt), x + coef[2,1]*k1 + coef[2,2]*k2 + coef[2,3]*k3;; __qualifiers);
    k5 = s * dt * @f(s*(t0 + coef[3,0]*dt), x + coef[3,1]*k1 + coef[3,2]*k2 + coef[3,3]*k3 + coef[3,4]*k4;; __qualifiers);
    k6 = s * dt * @f(s*(t0 + coef[4,0]*dt), x + coef[4,1]*k1 + coef[4,2]*k2 + coef[4,3]*k3 + coef[4,4]*k4 + coef[4,5]*k5;; __qualifiers);
    if(method == "RKDP") % for RKPD use fifth order solution to continue integration
    {
      dx_e = coef[5,1]*k1 + coef[5,2]*k3 + coef[5,3]*k4 + coef[5,4]*k5 + coef[5,5]*k6; % fourth order solution
      k7 = s * dt * @f(s*(t0 + coef[5,0]*dt), x + dx_e;; __qualifiers);
      dx_c = coef[6,1]*k1 + coef[6,2]*k3 + coef[6,3]*k4 + coef[6,4]*k5 + coef[6,5]*k6 + coef[6,6]*k7; % fifth order solution
    }
    else % for RKF, RKCK use fourth order solution to continue integration
    {
      dx_c = coef[5,1]*k1 + coef[5,2]*k3 + coef[5,3]*k4 + coef[5,4]*k5 + coef[5,5]*k6; % fourth order solution
      dx_e = coef[6,1]*k1 + coef[6,2]*k3 + coef[6,3]*k4 + coef[6,4]*k5 + coef[6,5]*k6; % fifth order solution
    }
    variable scale = 0.9*(eps/max(abs(dx_c-dx_e)))^(0.2); % 0.9 is safety factor; derivation see Numerical Recipes, Section "Adaptive Stepsize Control for Runge-Kutta"
    % variable scale = ( 0.5*eps*dt/max(abs(dx_c-dx_e)) )^0.25;  % "derivation of this formula can be found in advanced books on numerical analysis", statement in Numerical Methods Using Matlab, 4th Edition, 2004
    if(scale>1 or dt<=1e-15*t) % scale>1: accept stepsize only if correction factor is larger than 1
    {                          % dt<=abs(1e-15*t) in order to avoid infinite loops due to Double_Type precision limit at dt=1e-16*t
      x += dx_c;
      t0 += dt;
      if(path)
      {
	list_append(trajectory.t, s*t0);
	list_append(trajectory.x, x);
      }
      if(verbose) vmessage("t=%g", s*t0);
    }
    dt = _max(dt*scale, t*1e-15); % abs(1e-15*t) in order to avoid infinite loops due to Double_Type precision limit at dt=1e-16*t
    if(t0+dt > t) dt = t-t0;
  }
  if(path)
  {
    trajectory.t = list_to_array(trajectory.t);
    t0 = trajectory.t;
    variable len1 = length(x);
    if(len1==1)
    {
      trajectory.x = list_to_array(trajectory.x);
      x = trajectory.x;
    }
    else
    {
      variable i;
      variable len2 = length(trajectory.t);
      x = Double_Type[len2, len1];
      _for i(0,len2-1,1)
      {
	x[i,*] = trajectory.x[i];
      }
    }
  }
  return t0, x;
}

%
% Code from src/slang/math/integrateRKF45.sl
%
define integrateRKF45()
%!%+
%\function{integrateRKF45}
%\synopsis{integrates an ODE with the adaptive 4th/5th order Runge-Kutta-Fehlberg algorithm}
%\usage{x = integrateRKF45(&f, t1, t2, dt[, x0]);}
%\qualifiers{
%\qualifier{eps}{[\code{=1e-12}] error control tolerance}
%\qualifier{verbose}{show intermediate "times" \code{t}}
%}
%\description
%    This implementation of the adaptive Runge-Kutta-Fehlberg algorithm
%    is still considered EXPERIMENTAL and MAY REQUIRE FURTHER TESTING!
%
%    The integral of the ordinary differential equation\n
%      dx/dt = f(x(t), t)  with  x(t0) = x0\n
%    reads\n
%      x(t) = x0 + int_{t0}^{t} f(x(t'), t') dt'
%
%    \code{&f} is a reference to a function with two arguments:
%       \code{define f(x, t)}\n
%       \code{\{}\n
%       \code{  return ...;}\n
%       \code{\}}
%
%    \code{x} may be be a scalar as well as an array.
%\seealso{integrateRK4, integrateAB5}
%!%-
{
  variable f, t, t2, dt, x=0;
  switch(_NARGS)
  { case 4: (f, t, t2, dt) = (); }
  { case 5: (f, t, t2, dt, x) = (); }
  { help(_function_name()); return; }

  variable eps = qualifier("eps", 1e-12);
  variable verbose = qualifier_exists("verbose");

  % http://www.springerlink.com/content/f69256402h71w8m8/
  % http://math.fullerton.edu/mathews/n2003/RungeKuttaFehlbergMod.html
  % http://en.wikipedia.org/wiki/Runge%E2%80%93Kutta%E2%80%93Fehlberg_method

  while(t < t+dt <= t2)
  {
    if(verbose)  vmessage("t=%g", t);
    variable k1 = dt * @f(x,  t;; __qualifiers);
    variable k2 = dt * @f(x + 0.25*k1,  t + 0.25*dt;; __qualifiers);
    variable k3 = dt * @f(x + 0.09375*k1 + 0.28125*k2,  t + 0.375*dt;; __qualifiers);
    variable k4 = dt * @f(x + 1932./2197*k1 - 7200./2197*k2 + 7296./2197*k3,  t + 12./13*dt;; __qualifiers);
    variable k5 = dt * @f(x + 439./216*k1 - 8*k2 + 3680./513*k3 - 845./4104*k4,  t + dt;; __qualifiers);
    variable k6 = dt * @f(x - 8./27*k1 + 2*k2 - 3544./2565*k3 + 1859./4104*k4 - 0.275*k5,  t + 0.5*dt;; __qualifiers);
    variable dx4 = 25./216*k1 + 1408./2565*k3 + 2197./4104*k4 - 0.2*k5;
%   variable dx5 = 16./135*k1 + 6656./12825*k3 + 28561./56430*k4 - 0.18*k5 + 2./55*k6;
    variable dx4_dx5 = 1./360*k1 - 128./4275*k3 - 2197./75240*k4 + 0.02*k5 + 2./55*k6;  % dx4 - dx5
%    variable s = (eps*dt/2/abs(       dx4_dx5)  )^0.25;  % doesn't work for array-valued x
    variable s = ( eps*dt/2/sqrt(sumsq(dx4_dx5)) )^0.25;  % just a hack to make it work for array-valued x;
                                                          % remains to be seen whether it makes sense or not
    if(s>0.1)  % <- This condition was invented by Manfred and should probably be replaced by something more clever.
    {
      x += dx4;
      t += dt;
    }
    dt *= _max(0.1, _min(s, 4));
    if(t+dt>t2)  dt = t2-t;
  }
  if(verbose)  vmessage("t=%g", t);
  return x;
}

%
% Code from src/slang/math/integrate_trapez.sl
%
define integrate_trapez()
%!%+
%\function{integrate_trapez}
%\synopsis{numerical integration with the trapezoidal rule}
%\usage{Double_Type int = integrate_trapez(Double_Type x[], Double_Type y[]);
%    alternative:
%    Double_Type int = integrate_trapez(Ref_Type function, Double_Type min, Double_Type max);}
%\description
%    If two arguments (\code{x[]} and \code{y[] = function(x[])}) are given the integral is calculated via:
%    \code{int = sum_i (y[i-1] + y[i])/2 * (x[i]-x[i-1])}
%    An alternative usage with three arguments requires a reference to
%    the function and the integration limits. The function is evaluated
%    on an equidistant grid. The number of steps can be set with a qualifier.
%    The second case allows a forth argument specifying a previous result
%    with a smaller number of steps, which is meant for iterative usage by
%    the function \code{qsimp}.
%\qualifiers{
%\qualifier{steps [=100]}{:    number of equidistant grid points,
%                       on which the function is evaluated}
%}
%\example
%    % Integration of the sin function from 0 to PI/3
%    % The result should be 0.5
%    variable x = [0:PI/3:#100];
%
%    % The first method: providing grid and corresponding function values
%    integrate_trapez (x, sin(x));
%
%    % Alternative method using reference to the function and integration limits
%    integrate_trapez (&sin, 0, PI/3);
%    integrate_trapez (&sin, 0, PI/3; steps = 5000);    
%\seealso{integrateRK4, qsimp}
%!%-
{
  variable x, y;
  variable func, A, B;
  variable fac   = 1.0;
  variable R     = NULL;
  variable steps = nint(qualifier("steps", 100));
  
  switch(_NARGS)
  { case 2: (x, y) = ();
    if(length(x) != length(y) or length(x)<2)
      message("error ("+_function_name()+"): x and y have to be arrays of same length >1");
    else
      return sum( 0.5*(y[[:-2]]+y[[1:]]) * (x[[1:]]-x[[:-2]]) );
  }
  { case 3: (func, A, B)    = (); }
  { case 4: (func, A, B, R) = (); if (R != NULL) {fac = 0.5;} }
  { help(_function_name()); return;     }
  if (R == NULL) R = 0.;
  variable tnm = double( steps );
  variable del = ( B - A ) / tnm;                     %Spacing of the points to add
  variable xgrid = A + 0.5*del + [0:steps-1] * del;  %Grid of points @ compute function
  variable val = array_map_qualifiers( Double_Type, func, xgrid ;; __qualifiers );
  R = fac * ( R + (double(B)-A) * sum(val)/tnm );
  return R;
}

%
% Code from src/slang/arrays/2d/interpol2D.sl
%
%%%%%%%%%%%%%%%%%
define interpol2D()
%%%%%%%%%%%%%%%%%
%!%+
%\function{interpol2D}
%\synopsis{}
%\usage{img2 = interpol2D(X2, Y2,  X, Y, img);}
%\description
%    \code{img = Double_Type[ny, nx];}\n
%    \code{X  = Double_Type[nx];}\n
%    \code{Y  = Double_Type[ny];}\n
%    \code{min(X) <= X2 <= max(X)}\n
%    \code{min(Y) <= Y2 <= max(Y)}\n
%!%-
{
  variable X2, Y2, X, Y, img;
  switch(_NARGS)
  { case 5: (X2, Y2, X, Y, img) = (); }
  { help(_function_name()); return; }

  if(any(X2 < min(X)) or any(X2 > max(X)))
  { vmessage("error (%s): min(X) <= X2 <= max(X) is not fulfilled.", _function_name());
    return;
  }
  if(any(Y2 < min(Y)) or any(Y2 > max(Y)))
  { vmessage("error (%s): min(Y) <= Y2 <= max(Y) is not fulfilled.", _function_name());
    return;
  }

  variable img2 = Double_Type[ length(Y2), length(X2) ];

  _for $1 (0, length(Y2)-1 ){
    variable iy = wherelast( Y[[:-2]] <= Y2[$1] );    % such that  Y[iy] <= Y2[$1] < Y[iy+1]  holds
    
    variable dy  = Y2[$1] - Y[iy];
    variable dy_ = Y[iy+1] - Y2[$1];

    _for $2 (0, length(X2)-1 ){
      variable ix = wherelast( X[[:-2]] <= X2[$2] );   % such that  X[ix] <= X2[$2] < X[ix+1]  holds

      variable dx  = X2[$2] - X[ix];
      variable dx_ = X[ix+1] - X2[$2];

      variable w00 = dy_ * dx_;
      variable w10 = dy  * dx_;
      variable w01 = dy_ * dx;
      variable w11 = dy  * dx;

      img2[$1, $2] = (w00*img[iy,ix] + w10*img[iy+1,ix] + w01*img[iy,ix+1] + w11*img[iy+1,ix+1])
	            /(w00            + w10              + w01              + w11               );
    }
  }

  return img2;
}

%
% Code from src/slang/arrays/2d/interpol_2d_table.sl
%
define interpol_2d_table()
%!%+
%\function{interpol_2d_table}
%\usage{Double_Type interpol_2d_table(Double_Type x, y, Struct_Type table)}
%!%-
{
  variable x, y, tab, X, Y, Z;
  switch(_NARGS)
  { case 3: (x, y, tab) = (); }
  { case 5: (x, y, X, Y, Z) = (); tab = struct { x=X, y=Y, z=Z }; }
  { help(_function_name()); }

  if( min(tab.x) <= x <= max(tab.x)  &&  min(tab.y) <= y <= max(tab.y) )
  {
    variable ix, iy;
    for(ix=0; ix<length(tab.x)-2 && (tab.x[ix]-x)*(tab.x[ix+1]-x)>0; )
      ix++;
    for(iy=0; iy<length(tab.y)-2 && (tab.y[iy]-y)*(tab.y[iy+1]-y)>0; )
      iy++;
    if(qualifier_exists("verbose"))
    { vmessage("x[%d] = %f  <=  x = %f  <=  x[%d] = %f", ix, tab.x[ix],  x,  ix+1, tab.x[ix+1]);
      vmessage("y[%d] = %f  <=  y = %f  <=  y[%d] = %f", iy, tab.y[iy],  y,  iy+1, tab.y[iy+1]);
    }
    variable w00 = (tab.x[ix+1]-x) * (tab.y[iy+1]-y);
    variable w01 = (tab.x[ix+1]-x) * (y-tab.y[iy]);
    variable w10 = (x-tab.x[ix]) * (tab.y[iy+1]-y);
    variable w11 = (x-tab.x[ix]) * (y-tab.y[iy]);
    return (w00*tab.z[iy,ix] + w10*tab.z[iy+1,ix] + w01*tab.z[iy,ix+1] + w11*tab.z[iy+1,ix+1])
          /(w00              + w10                + w01                + w11                 );
  }
  vmessage("error (%s): extrapolation not allowed", _function_name());
  return log(-1.);
}

%
% Code from src/fitting/parameters/interpol_params.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define interpol_params()
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{set_params_interpol}
%\synopsis{interpolates between two parameter sets}
%\usage{Struct_Type params[] = interpol_params(Struct_Type p1[], Struct_Type p2[]);
%\altusage{Struct_Type params[] = interpol_params(Struct_Type p1[], Struct_Type p2[], Double_Type frac);}
%\altusage{Struct_Type params[] = interpol_params(Struct_Type p1[], Struct_Type p2[], par, Double_Type value);}
%}
%\description
%   \code{p1} and \code{p2} are parameter lists for the same fit-function
%   as obtained with \code{get_params}. Changing \code{frac} from 0 to 1,
%   the parameter set is interpolated from \code{p1} to \code{p2}.
%   If \code{frac} is not specified, \code{frac=0.5} is assumed.
%   \code{frac} can also be obtained by interpolating
%   the parameter \code{par} to the value \code{value}.
%!%-
{
  variable p1, p2, frac=0.5, par, value;
  variable i;
  switch(_NARGS)
  { case 2: (p1, p2) = (); }
  { case 3: (p1, p2, frac) = (); }
  { case 4: (p1, p2, par, value) = ();
      if(length(p1) != length(p2))
      { vmessage("error (%s): parameter sets from different fit-functions", _function_name());
	return;
      }
      variable v = NULL;
      if(typeof(par)==Integer_Type)
      { if(0 < par<=length(p1))
  	  v = [ p1[par-1].value, p2[par-1].value ];
      }
      else
        _for i (0, length(p1)-1, 1)
  	  if(p1[i].name == par)
	    v = [ p1[i].value, p2[i].value ];
      if(v!=NULL)
      { if(v[1]-v[0] != 0)
  	  frac = (value-v[0])/(v[1]-v[0]);
	else
	{ vmessage("error (%s): cannot interpolate parameter %S, which is %g in both sets.", _function_name(), par, v[0]);
	  return;
	}
      }
      else
      { vmessage("error (%s): parameter %S not found.", _function_name(), par);
	return;
      }
  }
  { help(_function_name()); return; }

  if(   length(p1) != length(p2)
     || any(array_struct_field(p1, "name") != array_struct_field(p2, "name")) )
  { vmessage("error (%s): cannot interpolate between parameter sets of different fit-functions.", _function_name());
    return;
  }

  if(frac<0 || frac>1)
    vmessage("warning (%s): frac=%f -- extrapolation", _function_name(), frac);

  variable p = Struct_Type[0];
  _for i (0, length(p1)-1, 1)
  {
    variable par_info = @(p1[i]);
    par_info.value = (1-frac) * p1[i].value + frac * p2[i].value;
    p = [p, par_info];
  }
  return p;
}

%
% Code from src/slang/math/interpol_polynomial.sl
%
define interpol_polynomial()
%!%+
%\function{interpol_polynomial}
%\synopsis{interpolates a polynomial function between data points}
%\usage{Double_Type y = interpol_polynomial(Double_Type x, Double_Type X[], Double_Type Y[]);}
%\description
%    \code{X} and \code{Y} are arrays of same length \code{n}.
%    \code{y} is the Lagrange polynomial (of degree \code{n-1}) interpolating
%    the data points (\code{X[i]}, \code{Y[i]}), evaluated at \code{x}:\n
%      \code{     n-1           n-1      x - X[j]  }\n
%      \code{y =  sum  Y[i] * product  ----------- }\n
%      \code{     i=0        j=0,j!=i   X[i]-X[j]  }
%
%    If \code{x} is an array (which doesn't need to be ordered), \code{y} will also be an array.
%\seealso{interpol_points}
%!%-
{
  variable x, X, Y;
  switch(_NARGS)
  { case 3: (x, X, Y) = (); }
  { help(_function_name()); return; }

  variable n = length(X);
  if(length(Y) != n)
  { vmessage("error (%s): the arrays X and Y must have the same size."); return; }

  variable y = Double_Type[0];
  if(length(x)>1)
  { variable x1;
    foreach x1 (x) { y = [y, interpol_polynomial(x1, X, Y)]; }
  }
  else
  { variable i, j;
    for(y=0,i=0; i<n; i++)
    { variable prod = Y[i];
      for(j=0; j<n; j++)
      { if(i!=j) { prod *= (x-X[j])/(X[i]-X[j]); }
      }
      y += prod;
    }
  }
  return y;
}

%
% Code from src/slang/math/inversemapping.sl
%
define inversemapping()
%!%+
%\function{inversemapping}
%\synopsis{computes the inverse of a 2d mapping numerically}
%\usage{(x, y) = inversemapping(&getxy, cache, x_, y_);}
%\description
%   (cache.x_, cache.y_) = getxy(cache.x, cache.y);
%   (x_, y_) = getxy(x, y);
%!%-
{
  variable getxy, cache, x_, y_;
  switch(_NARGS)
  { case 4: (getxy, cache, x_, y_) = (); }
  { help(_function_name()); return; }

  variable verbose = qualifier_exists("verbose");
  variable F = qualifier("stderr", stderr);

  if(verbose)  ()=fprintf(F, "# trying to invert x_=%f, y_=%f\n", x_, y_);

  loop(10)
%  forever
  {
    variable d2 = (x_-cache.x_)^2 + (y_-cache.y_)^2;
    variable all_i = array_sort(d2);
    variable i = all_i[[0:2]];

    variable ptr=1;
    while(   cache.x[all_i[ptr]] == cache.x[i[0]]
	  && cache.y[all_i[ptr]] == cache.y[i[0]])
      ptr++;
    i[1] = all_i[ptr];
    ptr++;
    while(   cache.x[all_i[ptr]] == cache.x[i[1]]
	  && cache.y[all_i[ptr]] == cache.y[i[1]])
      ptr++;
    i[2] = all_i[ptr];

    if( (cache.x[i[0]] == cache.x[i[1]] && cache.x[i[1]] == cache.x[i[2]]) )
    { ptr=3;
      while(cache.x[all_i[ptr]] == cache.x[i[0]])  ptr++;
      i[2] = all_i[ptr];
    }
    if( (cache.y[i[0]] == cache.y[i[1]] && cache.y[i[1]] == cache.y[i[2]]) )
    { ptr=3;
      while(cache.y[all_i[ptr]] == cache.y[i[0]])  ptr++;
      i[2] = all_i[ptr];
    }

    variable cachex  = cache.x [i];
    variable cachey  = cache.y [i];
    variable cachex_ = cache.x_[i];
    variable cachey_ = cache.y_[i];

    if(verbose)
    { variable tmpcache = struct_filter(cache, i; copy);
      print_struct(F, struct_combine(tmpcache, struct{ d2 = d2[i] }));
      % connect_points(0); pointstyle(-4);
      % plot([cachex_, x_], [cachey_, y_]);
      % xylabel(cachex[0], cachey[0], string(d2[i[0]]));
      % xylabel(cachex[1], cachey[1], string(d2[i[1]]));
      % xylabel(cachex[2], cachey[2], string(d2[i[2]]));
      % oplot([x_], [y_]);
    }

    variable dx_dx, dx_dy,  dy_dx, dy_dy,  x, y;
    %  x'1  =  x'0  +  (x1-x0) * dx'/dx  +  (y1-y0) * dx'/dy
    %  x'2  =  x'0  +  (x2-x0) * dx'/dx  +  (y2-y0) * dx'/dy
    (dx_dx, dx_dy) = solve_2d_system_of_equations( cachex[1]-cachex[0],  cachey[1]-cachey[0],  cachex_[1]-cachex_[0],
                                                   cachex[2]-cachex[0],  cachey[2]-cachey[0],  cachex_[2]-cachex_[0]
  					         );
    if(dx_dx==NULL)
    { vmessage("error (%s): cannot solve equations for dx'/dx, dx'/dy", _function_name());
      return NULL, NULL;
    }
    else
    { %  y'1  =  y'0  +  (x1-x0) * dy'/dx  +  (y1-y0) * dy'/dy
      %  y'2  =  y'0  +  (x2-x0) * dy'/dx  +  (y2-y0) * dy'/dy
      (dy_dx, dy_dy) = solve_2d_system_of_equations( cachex[1]-cachex[0],  cachey[1]-cachey[0],  cachey_[1]-cachey_[0],
                                                     cachex[2]-cachex[0],  cachey[2]-cachey[0],  cachey_[2]-cachey_[0]
    					           );
      if(dy_dx==NULL)
      { vmessage("error (%s): cannot solve equations for dy'/dx, dy'/dy", _function_name());
	return NULL, NULL;
      }

      %  x'  =  x'0  +  (x-x0) * dx'/dx  +  (y-y0) * dx'/dy
      %  y'  =  y'0  +  (x-x0) * dy'/dx  +  (y-y0) * dy'/dy
      (x, y) = solve_2d_system_of_equations( dx_dx,  dx_dy,  x_ - cachex_[0] + dx_dx * cachex[0] + dx_dy * cachey[0],
    					     dy_dx,  dy_dy,  y_ - cachey_[0] + dy_dx * cachex[0] + dy_dy * cachey[0]
				           );
      if(x==NULL)
      { vmessage("error (%s): cannot solve equations for x, y", _function_name());
	return NULL, NULL;
      }

      if(verbose)  ()=fprintf(F, "extrapolation => (x=%f, y=%f)\n", x, y);

      variable maxd2 = (max(cachex)-min(cachex))^2 + (max(cachey)-min(cachey))^2;
      variable meancachex = mean(cachex);
      variable meancachey = mean(cachey);
      variable d2_mean = (x-meancachex)^2+(y-meancachey)^2;

      variable x__, y__, d2_=0;
      do
      {
	if((isnan(d2_) || d2_>=d2[i[2]]) && d2_mean > maxd2)
	{ if(verbose)  ()=fprintf(F, "replacing (x=%f, y=%f) ", x, y);
	  x = (x+meancachex)/2;
	  y = (y+meancachey)/2;
	  if(verbose)  ()=fprintf(F, "by (x=%f, y=%f)\n", x, y);
	}
	% oplot(x, y);
        (x__, y__) = @getxy(x, y);
        d2_ = (x_-x__)^2 + (y_-y__)^2;

	if(verbose)  ()=fprintf(F, "(x=%f, y=%f) => (x_=%f, y_=%f), d2=%f\n", x, y, x__, y__, d2_);

        ifnot(isnan(d2_))
        {
          cache.x  = [cache.x , x  ];
          cache.y  = [cache.y , y  ];
          cache.x_ = [cache.x_, x__];
          cache.y_ = [cache.y_, y__];

          if(d2_ < qualifier("eps", 1e-3)^2)
          { if(verbose)  ()=fprintf(F, "success: (x=%f, y=%f) => (x_=%f, y_=%f)\n", x, y, x__, y__);
            return x, y;
          }
	}
      } while((isnan(d2_) || d2_>=d2[i[2]]) && d2_mean>maxd2);

#iffalse
	if(verbose)  ()=printf("warning (%s): extrapolation from (mean{x}=%f, mean{y}=%f) to (x=%f, y=%f) to far, rescaling to ", _function_name(), meancachex, meancachey, x, y);
	x = meancachex + (x-meancachex)*sqrt(10.*maxd2/d2_mean);
	y = meancachey + (y-meancachey)*sqrt(10.*maxd2/d2_mean);
	if(verbose)  ()=printf("(x=%f, y=%f)\n", x, y);

	(x__, y__) = @getxy(x, y);
	d2_ = (x_-x__)^2 + (y_-y__)^2;
        if(isnan(d2_))
      	  vmessage("warning (%s): function evaluation for (x=%f, y=%f) returned nan", _function_name(), x, y);
        else
        {
          cache.x  = [cache.x , x  ];
          cache.y  = [cache.y , y  ];
          cache.x_ = [cache.x_, x__];
          cache.y_ = [cache.y_, y__];

          if(d2_ < qualifier("eps", 1e-3)^2)
          { if(verbose)  vmessage("success (%s): (x=%f, y=%f) => (x_=%f, y_=%f)", _function_name(), x, y, x__, y__);
	    return x, y;
	  }
        }
	  d2_mean = (x-meancachex)^2+(y-meancachey)^2;
	  ()=printf("d2_mean=%f\n", d2_mean);
	} while(d2_mean > 100. * maxd2);
      }

%      if(verbose)
%      {
%	()=printf("x=%f, y=%f => x_=%f, y_=%f\n", x, y, x__, y__);
%        ()=printf("old distances^2: %f %f %f\n", d2[i[0]], d2[i[1]], d2[i[2]]);
%        ()=printf("new distances^2: %f\n", d2_);
%      }
#endif

      if(   isnan(d2_)
	 || (d2_ >= d2[i[2]])
	 || (cache.x[i[0]] == cache.x[i[1]] && cache.x[i[1]] == x) || (cache.y[i[0]] == cache.y[i[1]] && cache.y[i[1]] == y)
	)
      {
        if(verbose)
	{ if(isnan(d2_))
	    ()=fprintf(F, "# nan\n");
	  else
	    if( (d2_ >= d2[i[2]]) )
	      ()=fprintf(F, "# no improvement => finer grid required\n");
  	    else
	      ()=fprintf(F, "# degeneracy => finer grid required\n");
	  ()=fprintf(F, "dx'/dx = %f,  dx'/dy = %f,  dy'/dx = %f,  dy'/dy = %f\n", dx_dx, dx_dy, dy_dx, dy_dy);
	}

	variable X = cache.x[i];
	variable Y = cache.x[i];
	ifnot(isnan(d2_) || d2_ >= 2*d2[i[2]])
	  (X, Y) = ([x,X], [y,Y]);

	variable x1, y1;
	variable n = 4, improved=0;
	while(not improved and n<=8)
	{ if(verbose)  ()=fprintf(F, "%d x %d grid\n", n, n);
  	  foreach x1 ([1.5*min(X)-0.5*max(X) : 1.5*max(X)-0.5*min(X) : #n])
	    foreach y1 ([1.5*min(Y)-0.5*max(Y) : 1.5*max(Y)-0.5*min(Y) : #n])
	    { (x__, y__) = @getxy(x1, y1);
              cache.x  = [cache.x , x1 ];
              cache.y  = [cache.y , y1 ];
              cache.x_ = [cache.x_, x__];
              cache.y_ = [cache.y_, y__];
              if( (x_-x__)^2 + (y_-y__)^2 < d2[i[2]]
		   and (x1!=cache.x[i[0]] ||  y1!=cache.y[i[0]])
		   and (x1!=cache.x[i[1]] ||  y1!=cache.y[i[1]])
		)
		improved = 1;
	      if(verbose)  ()=fprintf(F, "(x=%f, y=%f) => (x_=%f, y_=%f), d2=%f (%d)", x1, y1, x__, y__, (x_-x__)^2 + (y_-y__)^2, improved);
	    }
	  n *= 2;
	}
	ifnot(improved)
	{ if(verbose)  ()=fprintf(F, "8 x 8 grid search did not improve result");
	  return NULL, NULL;
	}
      }
    }
  }

  return NULL, NULL;
}

%
% Code from src/fitting/fit-functions/ionabs.sl
%
%%%%%%%%%%%%%%%%%
define ionabs_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%%
%!%+
%\function{ionabs (fit-function)}
%\synopsis{multiplicative fit-function for ionized photoabsorption}
%\description
%    Currently, the cross section is only evaluated at the center of each bin!
%
%    The first fit parameter, \code{ionabs.N_H}, is a dummy parameter,
%    which is not taken into account for the model. One can, however, use it
%    to relate the ionized column densities to an equivalent hydrogen column density.
%\seealso{ionized_phabs_sigma}
%!%-
{
  %
  % USE A FINER GRID
  %
  variable energy = _A(1) / ((bin_lo+bin_hi)/2);
  variable tau = Double_Type[length(energy)];
  variable i=1, Z, nEl;
  for(Z=1, nEl=1; Z<=28; i++)
  {
    if(par[i]!=0)
      tau += ionized_phabs_sigma(energy; Z=Z, Nel=nEl) * par[i];
    if(nEl>1) nEl--; else { Z++; nEl = Z; }
  }
  return exp(-tau);
}
% a = ["\"N_H[cm^-2 (dummy)]\""]; for(Z=1,nEl=1; Z<=28; nEl--) { a = [a, sprintf("\"N_%s%d[cm^-2]\"", atom_name(Z), 1+Z-nEl)]; if(nEl==1) { Z++; nEl=Z+1; } }; strjoin_wrap(a, ","; maxlen=128, initial="add_slang_function(\"ionabs\", [", final="]);");
add_slang_function("ionabs", ["N_H[cm^-2 (dummy)]","N_H1[cm^-2]","N_He1[cm^-2]","N_He2[cm^-2]","N_Li1[cm^-2]","N_Li2[cm^-2]",
 "N_Li3[cm^-2]","N_Be1[cm^-2]","N_Be2[cm^-2]","N_Be3[cm^-2]","N_Be4[cm^-2]","N_B1[cm^-2]","N_B2[cm^-2]","N_B3[cm^-2]",
 "N_B4[cm^-2]","N_B5[cm^-2]","N_C1[cm^-2]","N_C2[cm^-2]","N_C3[cm^-2]","N_C4[cm^-2]","N_C5[cm^-2]","N_C6[cm^-2]","N_N1[cm^-2]",
 "N_N2[cm^-2]","N_N3[cm^-2]","N_N4[cm^-2]","N_N5[cm^-2]","N_N6[cm^-2]","N_N7[cm^-2]","N_O1[cm^-2]","N_O2[cm^-2]","N_O3[cm^-2]",
 "N_O4[cm^-2]","N_O5[cm^-2]","N_O6[cm^-2]","N_O7[cm^-2]","N_O8[cm^-2]","N_F1[cm^-2]","N_F2[cm^-2]","N_F3[cm^-2]","N_F4[cm^-2]",
 "N_F5[cm^-2]","N_F6[cm^-2]","N_F7[cm^-2]","N_F8[cm^-2]","N_F9[cm^-2]","N_Ne1[cm^-2]","N_Ne2[cm^-2]","N_Ne3[cm^-2]",
 "N_Ne4[cm^-2]","N_Ne5[cm^-2]","N_Ne6[cm^-2]","N_Ne7[cm^-2]","N_Ne8[cm^-2]","N_Ne9[cm^-2]","N_Ne10[cm^-2]","N_Na1[cm^-2]",
 "N_Na2[cm^-2]","N_Na3[cm^-2]","N_Na4[cm^-2]","N_Na5[cm^-2]","N_Na6[cm^-2]","N_Na7[cm^-2]","N_Na8[cm^-2]","N_Na9[cm^-2]",
 "N_Na10[cm^-2]","N_Na11[cm^-2]","N_Mg1[cm^-2]","N_Mg2[cm^-2]","N_Mg3[cm^-2]","N_Mg4[cm^-2]","N_Mg5[cm^-2]","N_Mg6[cm^-2]",
 "N_Mg7[cm^-2]","N_Mg8[cm^-2]","N_Mg9[cm^-2]","N_Mg10[cm^-2]","N_Mg11[cm^-2]","N_Mg12[cm^-2]","N_Al1[cm^-2]","N_Al2[cm^-2]",
 "N_Al3[cm^-2]","N_Al4[cm^-2]","N_Al5[cm^-2]","N_Al6[cm^-2]","N_Al7[cm^-2]","N_Al8[cm^-2]","N_Al9[cm^-2]","N_Al10[cm^-2]",
 "N_Al11[cm^-2]","N_Al12[cm^-2]","N_Al13[cm^-2]","N_Si1[cm^-2]","N_Si2[cm^-2]","N_Si3[cm^-2]","N_Si4[cm^-2]","N_Si5[cm^-2]",
 "N_Si6[cm^-2]","N_Si7[cm^-2]","N_Si8[cm^-2]","N_Si9[cm^-2]","N_Si10[cm^-2]","N_Si11[cm^-2]","N_Si12[cm^-2]","N_Si13[cm^-2]",
 "N_Si14[cm^-2]","N_P1[cm^-2]","N_P2[cm^-2]","N_P3[cm^-2]","N_P4[cm^-2]","N_P5[cm^-2]","N_P6[cm^-2]","N_P7[cm^-2]",
 "N_P8[cm^-2]","N_P9[cm^-2]","N_P10[cm^-2]","N_P11[cm^-2]","N_P12[cm^-2]","N_P13[cm^-2]","N_P14[cm^-2]","N_P15[cm^-2]",
 "N_S1[cm^-2]","N_S2[cm^-2]","N_S3[cm^-2]","N_S4[cm^-2]","N_S5[cm^-2]","N_S6[cm^-2]","N_S7[cm^-2]","N_S8[cm^-2]","N_S9[cm^-2]",
 "N_S10[cm^-2]","N_S11[cm^-2]","N_S12[cm^-2]","N_S13[cm^-2]","N_S14[cm^-2]","N_S15[cm^-2]","N_S16[cm^-2]","N_Cl1[cm^-2]",
 "N_Cl2[cm^-2]","N_Cl3[cm^-2]","N_Cl4[cm^-2]","N_Cl5[cm^-2]","N_Cl6[cm^-2]","N_Cl7[cm^-2]","N_Cl8[cm^-2]","N_Cl9[cm^-2]",
 "N_Cl10[cm^-2]","N_Cl11[cm^-2]","N_Cl12[cm^-2]","N_Cl13[cm^-2]","N_Cl14[cm^-2]","N_Cl15[cm^-2]","N_Cl16[cm^-2]",
 "N_Cl17[cm^-2]","N_Ar1[cm^-2]","N_Ar2[cm^-2]","N_Ar3[cm^-2]","N_Ar4[cm^-2]","N_Ar5[cm^-2]","N_Ar6[cm^-2]","N_Ar7[cm^-2]",
 "N_Ar8[cm^-2]","N_Ar9[cm^-2]","N_Ar10[cm^-2]","N_Ar11[cm^-2]","N_Ar12[cm^-2]","N_Ar13[cm^-2]","N_Ar14[cm^-2]","N_Ar15[cm^-2]",
 "N_Ar16[cm^-2]","N_Ar17[cm^-2]","N_Ar18[cm^-2]","N_K1[cm^-2]","N_K2[cm^-2]","N_K3[cm^-2]","N_K4[cm^-2]","N_K5[cm^-2]",
 "N_K6[cm^-2]","N_K7[cm^-2]","N_K8[cm^-2]","N_K9[cm^-2]","N_K10[cm^-2]","N_K11[cm^-2]","N_K12[cm^-2]","N_K13[cm^-2]",
 "N_K14[cm^-2]","N_K15[cm^-2]","N_K16[cm^-2]","N_K17[cm^-2]","N_K18[cm^-2]","N_K19[cm^-2]","N_Ca1[cm^-2]","N_Ca2[cm^-2]",
 "N_Ca3[cm^-2]","N_Ca4[cm^-2]","N_Ca5[cm^-2]","N_Ca6[cm^-2]","N_Ca7[cm^-2]","N_Ca8[cm^-2]","N_Ca9[cm^-2]","N_Ca10[cm^-2]",
 "N_Ca11[cm^-2]","N_Ca12[cm^-2]","N_Ca13[cm^-2]","N_Ca14[cm^-2]","N_Ca15[cm^-2]","N_Ca16[cm^-2]","N_Ca17[cm^-2]",
 "N_Ca18[cm^-2]","N_Ca19[cm^-2]","N_Ca20[cm^-2]","N_Sc1[cm^-2]","N_Sc2[cm^-2]","N_Sc3[cm^-2]","N_Sc4[cm^-2]","N_Sc5[cm^-2]",
 "N_Sc6[cm^-2]","N_Sc7[cm^-2]","N_Sc8[cm^-2]","N_Sc9[cm^-2]","N_Sc10[cm^-2]","N_Sc11[cm^-2]","N_Sc12[cm^-2]","N_Sc13[cm^-2]",
 "N_Sc14[cm^-2]","N_Sc15[cm^-2]","N_Sc16[cm^-2]","N_Sc17[cm^-2]","N_Sc18[cm^-2]","N_Sc19[cm^-2]","N_Sc20[cm^-2]",
 "N_Sc21[cm^-2]","N_Ti1[cm^-2]","N_Ti2[cm^-2]","N_Ti3[cm^-2]","N_Ti4[cm^-2]","N_Ti5[cm^-2]","N_Ti6[cm^-2]","N_Ti7[cm^-2]",
 "N_Ti8[cm^-2]","N_Ti9[cm^-2]","N_Ti10[cm^-2]","N_Ti11[cm^-2]","N_Ti12[cm^-2]","N_Ti13[cm^-2]","N_Ti14[cm^-2]","N_Ti15[cm^-2]",
 "N_Ti16[cm^-2]","N_Ti17[cm^-2]","N_Ti18[cm^-2]","N_Ti19[cm^-2]","N_Ti20[cm^-2]","N_Ti21[cm^-2]","N_Ti22[cm^-2]","N_V1[cm^-2]",
 "N_V2[cm^-2]","N_V3[cm^-2]","N_V4[cm^-2]","N_V5[cm^-2]","N_V6[cm^-2]","N_V7[cm^-2]","N_V8[cm^-2]","N_V9[cm^-2]","N_V10[cm^-2]",
 "N_V11[cm^-2]","N_V12[cm^-2]","N_V13[cm^-2]","N_V14[cm^-2]","N_V15[cm^-2]","N_V16[cm^-2]","N_V17[cm^-2]","N_V18[cm^-2]",
 "N_V19[cm^-2]","N_V20[cm^-2]","N_V21[cm^-2]","N_V22[cm^-2]","N_V23[cm^-2]","N_Cr1[cm^-2]","N_Cr2[cm^-2]","N_Cr3[cm^-2]",
 "N_Cr4[cm^-2]","N_Cr5[cm^-2]","N_Cr6[cm^-2]","N_Cr7[cm^-2]","N_Cr8[cm^-2]","N_Cr9[cm^-2]","N_Cr10[cm^-2]","N_Cr11[cm^-2]",
 "N_Cr12[cm^-2]","N_Cr13[cm^-2]","N_Cr14[cm^-2]","N_Cr15[cm^-2]","N_Cr16[cm^-2]","N_Cr17[cm^-2]","N_Cr18[cm^-2]",
 "N_Cr19[cm^-2]","N_Cr20[cm^-2]","N_Cr21[cm^-2]","N_Cr22[cm^-2]","N_Cr23[cm^-2]","N_Cr24[cm^-2]","N_Mn1[cm^-2]","N_Mn2[cm^-2]",
 "N_Mn3[cm^-2]","N_Mn4[cm^-2]","N_Mn5[cm^-2]","N_Mn6[cm^-2]","N_Mn7[cm^-2]","N_Mn8[cm^-2]","N_Mn9[cm^-2]","N_Mn10[cm^-2]",
 "N_Mn11[cm^-2]","N_Mn12[cm^-2]","N_Mn13[cm^-2]","N_Mn14[cm^-2]","N_Mn15[cm^-2]","N_Mn16[cm^-2]","N_Mn17[cm^-2]",
 "N_Mn18[cm^-2]","N_Mn19[cm^-2]","N_Mn20[cm^-2]","N_Mn21[cm^-2]","N_Mn22[cm^-2]","N_Mn23[cm^-2]","N_Mn24[cm^-2]",
 "N_Mn25[cm^-2]","N_Fe1[cm^-2]","N_Fe2[cm^-2]","N_Fe3[cm^-2]","N_Fe4[cm^-2]","N_Fe5[cm^-2]","N_Fe6[cm^-2]","N_Fe7[cm^-2]",
 "N_Fe8[cm^-2]","N_Fe9[cm^-2]","N_Fe10[cm^-2]","N_Fe11[cm^-2]","N_Fe12[cm^-2]","N_Fe13[cm^-2]","N_Fe14[cm^-2]","N_Fe15[cm^-2]",
 "N_Fe16[cm^-2]","N_Fe17[cm^-2]","N_Fe18[cm^-2]","N_Fe19[cm^-2]","N_Fe20[cm^-2]","N_Fe21[cm^-2]","N_Fe22[cm^-2]",
 "N_Fe23[cm^-2]","N_Fe24[cm^-2]","N_Fe25[cm^-2]","N_Fe26[cm^-2]","N_Co1[cm^-2]","N_Co2[cm^-2]","N_Co3[cm^-2]","N_Co4[cm^-2]",
 "N_Co5[cm^-2]","N_Co6[cm^-2]","N_Co7[cm^-2]","N_Co8[cm^-2]","N_Co9[cm^-2]","N_Co10[cm^-2]","N_Co11[cm^-2]","N_Co12[cm^-2]",
 "N_Co13[cm^-2]","N_Co14[cm^-2]","N_Co15[cm^-2]","N_Co16[cm^-2]","N_Co17[cm^-2]","N_Co18[cm^-2]","N_Co19[cm^-2]",
 "N_Co20[cm^-2]","N_Co21[cm^-2]","N_Co22[cm^-2]","N_Co23[cm^-2]","N_Co24[cm^-2]","N_Co25[cm^-2]","N_Co26[cm^-2]",
 "N_Co27[cm^-2]","N_Ni1[cm^-2]","N_Ni2[cm^-2]","N_Ni3[cm^-2]","N_Ni4[cm^-2]","N_Ni5[cm^-2]","N_Ni6[cm^-2]","N_Ni7[cm^-2]",
 "N_Ni8[cm^-2]","N_Ni9[cm^-2]","N_Ni10[cm^-2]","N_Ni11[cm^-2]","N_Ni12[cm^-2]","N_Ni13[cm^-2]","N_Ni14[cm^-2]","N_Ni15[cm^-2]",
 "N_Ni16[cm^-2]","N_Ni17[cm^-2]","N_Ni18[cm^-2]","N_Ni19[cm^-2]","N_Ni20[cm^-2]","N_Ni21[cm^-2]","N_Ni22[cm^-2]",
 "N_Ni23[cm^-2]","N_Ni24[cm^-2]","N_Ni25[cm^-2]","N_Ni26[cm^-2]","N_Ni27[cm^-2]","N_Ni28[cm^-2]"]);

%%%%%%%%%%%%%%%%%%%%%
define ionabs_default(i)
%%%%%%%%%%%%%%%%%%%%%
{
  return (0, 1, 0, 1e24);
}
set_param_default_hook("ionabs", &ionabs_default);

%
% Code from src/misc/atomic/ionized_phabs_sigma.sl
%
variable VernerYakovlev1995_Table1 = struct {
Z= [1,2,2,3,3,3,3,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,
 10,10,10,10,10,10,10,10,10,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,12,12,12,12,12,
 12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,
 13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,
 14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,
 15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
 16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,17,17,17,17,17,
 17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,
 17,17,17,17,17,17,17,17,17,17,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
 18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,19,19,19,19,19,19,19,19,19,19,19,19,
 19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,
 19,19,19,19,19,19,19,19,19,19,19,19,19,19,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,
 20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,
 20,20,20,20,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,
 21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,
 21,21,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,
 22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,
 22,22,22,22,22,22,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,
 23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,
 23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,
 24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,
 24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,25,25,25,25,25,25,25,25,25,25,25,25,
 25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,
 25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,
 25,25,25,25,25,25,25,25,25,25,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,
 26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,
 26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,27,27,27,27,
 27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,
 27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,
 27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,28,28,28,28,28,28,28,28,28,28,28,28,28,
 28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,
 28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,
 28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,
 29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,
 29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,
 29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,30,30,30,30,30,30,30,30,30,30,30,30,30,
 30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,
 30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,
 30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30,30],
Nel = [1,2,1,3,3,2,1,4,4,3,3,2,1,5,5,5,4,4,3,3,2,1,6,6,6,5,5,5,4,4,3,3,2,1,7,7,7,6,6,6,5,5,5,4,4,3,3,
 2,1,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,
 3,2,1,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,
 5,5,5,4,4,3,3,2,1,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,14,14,14,14,14,13,
 13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,15,15,15,15,15,14,14,14,14,14,13,13,13,
 13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,
 13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,17,17,17,17,17,16,16,16,16,16,15,15,
 15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,18,18,18,18,
 18,17,17,17,17,17,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,
 7,6,6,6,5,5,5,4,4,3,3,2,1,19,19,19,19,19,19,18,18,18,18,18,17,17,17,17,17,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,
 13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,20,20,20,20,20,20,19,19,19,19,19,19,18,18,
 18,18,18,17,17,17,17,17,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,
 8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,21,21,21,21,21,21,21,20,20,20,20,20,20,20,19,19,19,19,19,19,18,18,18,18,18,17,17,17,17,17,16,
 16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,
 3,2,1,22,22,22,22,22,22,22,21,21,21,21,21,21,21,20,20,20,20,20,20,19,19,19,19,19,19,18,18,18,18,18,17,17,17,17,17,16,16,16,16,
 16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,23,
 23,23,23,23,23,23,22,22,22,22,22,22,21,21,21,21,21,21,20,20,20,20,20,20,19,19,19,19,19,19,18,18,18,18,18,17,17,17,17,17,16,16,
 16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,
 1,24,24,24,24,24,24,24,23,23,23,23,23,23,22,22,22,22,22,22,21,21,21,21,21,21,20,20,20,20,20,20,19,19,19,19,19,19,18,18,18,18,
 18,17,17,17,17,17,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,
 7,6,6,6,5,5,5,4,4,3,3,2,1,25,25,25,25,25,25,25,24,24,24,24,24,24,24,23,23,23,23,23,23,22,22,22,22,22,22,21,21,21,21,21,21,20,
 20,20,20,20,20,19,19,19,19,19,19,18,18,18,18,18,17,17,17,17,17,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,
 12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,26,26,26,26,26,26,26,25,25,25,25,25,25,25,24,24,24,24,
 24,24,23,23,23,23,23,23,22,22,22,22,22,22,21,21,21,21,21,21,20,20,20,20,20,20,19,19,19,19,19,19,18,18,18,18,18,17,17,17,17,17,
 16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,
 3,3,2,1,27,27,27,27,27,27,27,26,26,26,26,26,26,25,25,25,25,25,25,24,24,24,24,24,24,23,23,23,23,23,23,22,22,22,22,22,22,21,21,
 21,21,21,21,20,20,20,20,20,20,19,19,19,19,19,19,18,18,18,18,18,17,17,17,17,17,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,
 13,13,13,13,12,12,12,12,11,11,11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,28,28,28,28,28,28,28,27,27,27,27,27,27,
 26,26,26,26,26,26,25,25,25,25,25,25,24,24,24,24,24,24,23,23,23,23,23,23,22,22,22,22,22,22,21,21,21,21,21,21,20,20,20,20,20,20,
 19,19,19,19,19,19,18,18,18,18,18,17,17,17,17,17,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,
 11,11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,29,29,29,29,29,29,29,28,28,28,28,28,28,27,27,27,27,27,27,26,26,26,26,
 26,26,25,25,25,25,25,25,24,24,24,24,24,24,23,23,23,23,23,23,22,22,22,22,22,22,21,21,21,21,21,21,20,20,20,20,20,20,19,19,19,19,
 19,19,18,18,18,18,18,17,17,17,17,17,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,11,10,10,
 10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1,30,30,30,30,30,30,30,29,29,29,29,29,29,29,28,28,28,28,28,28,27,27,27,27,27,27,26,
 26,26,26,26,26,25,25,25,25,25,25,24,24,24,24,24,24,23,23,23,23,23,23,22,22,22,22,22,22,21,21,21,21,21,21,20,20,20,20,20,20,19,
 19,19,19,19,19,18,18,18,18,18,17,17,17,17,17,16,16,16,16,16,15,15,15,15,15,14,14,14,14,14,13,13,13,13,13,12,12,12,12,11,11,11,
 11,10,10,10,9,9,9,8,8,8,7,7,7,6,6,6,5,5,5,4,4,3,3,2,1],
n = [1,1,1,2,1,1,1,2,1,2,1,1,1,2,2,1,2,1,2,1,1,1,2,2,1,2,2,1,2,1,2,1,1,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,
 1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,
 1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,3,3,
 2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,
 2,1,2,2,1,2,2,1,2,1,2,1,1,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,3,3,
 2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,3,3,2,2,1,3,3,2,2,1,3,3,2,
 2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,
 1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,4,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,
 1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,4,3,3,2,2,1,4,3,3,2,2,1,3,3,2,
 2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,4,3,3,3,
 2,2,1,4,3,3,3,2,2,1,3,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,
 2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,4,3,3,3,2,2,1,4,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,
 2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,4,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,
 2,1,3,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,
 1,2,1,2,1,1,1,4,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,
 1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,4,3,3,3,2,2,1,4,3,3,3,2,2,1,3,3,3,2,2,1,
 3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,
 2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,4,3,3,3,2,2,1,4,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,
 3,2,2,1,3,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,
 2,2,1,2,1,2,1,1,1,4,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,
 1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,
 4,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,
 2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,4,3,3,
 3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,
 1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,1,2,1,1,1,
 4,3,3,3,2,2,1,4,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,3,3,2,2,1,3,
 3,3,2,2,1,3,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,3,2,2,1,3,2,2,1,3,2,2,1,2,2,1,2,2,1,2,2,1,2,2,1,2,2,
 1,2,2,1,2,1,2,1,1,1],
l = [0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,
 0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,
 0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,
 1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,
 0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,
 1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,
 0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,
 0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,
 0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,0,1,0,1,
 0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,2,1,0,
 1,0,0,0,2,1,0,1,0,0,2,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,
 1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,2,1,0,1,0,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,
 1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,
 0,0,2,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,
 0,0,0,0,0,0,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,
 0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,2,1,0,1,0,0,0,2,1,0,1,0,0,2,1,0,1,0,0,
 2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,
 0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,2,1,0,1,0,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,
 0,1,0,0,2,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,
 1,0,0,0,0,0,0,0,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,
 0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,
 0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,1,0,
 1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,2,1,
 0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,
 0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,
 0,2,1,0,1,0,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,1,0,1,0,0,2,
 1,0,1,0,0,2,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,
 0,1,0,0,0,0,0,0,0,0],
E_th = [13.6,24.59,54.42,5.392,64.39,75.64,122.5,9.323,119.3,18.21,129.9,153.9,217.7,8.298,14.05,194,
 25.16,209.8,37.93,227.4,259.4,340.2,11.26,19.39,291,24.38,30.47,307.6,47.89,328.9,64.49,352.2,392.1,490,14.53,25.41,404.8,29.6,
 37.96,423.6,47.45,55.45,447.3,77.47,475.3,97.89,504.3,552.1,667.1,13.62,28.48,538,35.12,45.99,558.1,54.94,65.51,584,77.41,
 87.37,614.4,113.9,649.1,138.1,683.7,739.3,871.4,17.42,37.86,694,34.97,54.59,712.2,62.71,76.1,739.2,87.14,99.57,770.9,114.2,
 126.2,809.1,157.2,850.2,185.2,890.5,953.9,1103,21.56,48.47,870.1,40.96,63.74,883.1,63.46,87.21,913.1,97.12,113.2,948,126.2,
 141.5,987.3,157.9,171.9,1031,207.3,1078,239.1,1125,1196,1362,5.139,38.14,70.84,1079,47.29,73.47,1074,71.62,99.45,1110,98.92,
 126.9,1143,138.4,156.7,1185,172.2,189.9,1230,208.5,224.4,1281,264.2,1335,299.9,1386,1465,1649,7.646,54.9,94,1311,15.04,65.69,
 98.81,1304,80.14,111.1,1317,109.3,141.1,1356,141.3,173.5,1400,186.5,207.6,1449,224.9,244.4,1501,266,283.9,1558,328.2,1618,
 367.5,1675,1762,1963,5.986,11.33,80.4,125.6,1567,18.83,89.97,128.1,1559,28.45,102.6,140.7,1572,120,155.8,1588,153.8,190.3,1634,
 190.5,226.8,1688,241.4,266.2,1739,284.6,306.5,1799,330.1,350.2,1862,399.4,1929,442.1,1992,2086,2304,8.152,15.17,106,156,1846,
 16.35,22.4,118.6,161.9,1839,33.49,131.1,174.4,1852,45.14,146.6,189.9,1868,166.8,207.6,1887,205.1,246.8,1946,246.5,287.2,2001,
 303.2,331,2058,351.1,375.6,2125,401.4,423.4,2194,476.1,2268,523.5,2336,2438,2673,10.49,20.17,140,194,2154,19.73,27.09,149.5,
 198.7,2138,30.2,38.28,163.9,212.4,2157,51.44,179.5,228,2173,65.03,197.7,246.1,2192,220.4,266.3,2214,263.2,309.8,2280,309.4,
 355.2,2337,371.7,401.8,2407,424.5,452.2,2477,479.6,503.5,2553,560.4,2633,611.9,2707,2817,3070,10.36,21.3,170,235,2477,23.33,
 31.9,184.6,238.7,2469,34.83,44.15,199.5,253.6,2484,47.31,57.5,216.4,270.3,2502,72.68,235,288.8,2522,88.05,255.7,309.4,2544,
 280.9,332.1,2569,328.2,379.7,2641,379.1,429.6,2705,447.1,480.4,2782,504.8,534.6,2859,564.7,590.6,2941,651.7,3029,707.2,3107,
 3224,3494,12.97,25.31,209,278,2830,23.81,36.86,223.6,283.7,2818,39.61,50.19,238.2,297.9,2836,53.47,64.7,255.2,315.1,2851,67.82,
 79.97,276,335.4,2875,97.03,297.4,356.6,2898,114.2,320.7,379.7,2923,348.3,404.8,2951,400.1,456.5,3030,455.6,510.9,3100,529.3,
 566,3184,592,624.9,3266,656.7,684.6,3356,749.8,3448,809.4,3534,3659,3946,15.76,28.92,249.2,326,3203,27.63,41.98,266.2,331.7,
 3195,40.74,56.37,280.1,345.5,3211,59.81,71.74,298.7,364.2,3228,75.02,88.28,320,385.2,3253,91.01,105.6,342.6,407.6,3277,124.3,
 366.7,431.4,3303,143.5,392.5,457,3331,422.5,484.5,3361,478.7,540.3,3446,539,599.2,3523,618.3,658.4,3613,686.1,721.7,3702,755.8,
 785.6,3798,854.8,3898,918,3988,4121,4426,4.341,24.66,40.8,301.4,384.3,3614,31.63,47.28,306.7,378.3,3590,45.81,62.69,327.9,399,
 3611,60.91,79.25,345.2,416.5,3630,82.66,96.57,366.5,437.4,3651,99.44,115.2,390.9,461.8,3679,117.6,134.4,416.2,486.8,3706,154.7,
 443,513.3,3735,175.8,471.3,541.3,3766,503.8,571.2,3799,564.7,631,3890,629.5,694.5,3974,714.7,757.8,4070,786.7,825.5,4166,861.1,
 893.5,4269,968,4375,1035,4471,4611,4934,6.113,34.43,48.3,352.3,442.5,4043,11.87,40.9,60.37,363.8,441.5,4022,50.91,69.2,373.1,
 450.7,4032,67.27,86.8,394.4,471.9,4054,84.51,105.4,417.5,494.8,4078,108.8,124.9,442.3,519.3,4105,127.2,145.2,468.7,545.5,4133,
 147.2,166.4,496.7,573.2,4163,188.3,527,601.8,4198,211.3,556.9,632.6,4229,591.9,664.9,4265,657.2,728.7,4362,726.7,796.8,4453,
 817.7,864.2,4555,894.6,935.7,4659,974.5,1008,4767,1087,4880,1157,4982,5129,5470,7.342,8.01,33.6,56.4,405.4,503.2,4494,12.8,
 14.44,46.78,68.48,419,502,4474,24.76,55.91,77.62,428.5,511.5,4483,73.49,94.51,447.6,530.6,4502,91.87,114.1,473,555.2,4528,
 110.7,134.8,497.7,580.3,4554,138,155.7,523.6,605.2,4582,158.1,178.4,554.4,636.2,4615,180,201.5,585.2,666.6,4649,225.1,617.3,
 698.2,4684,249.8,650.5,730.9,4720,687.4,765.7,4759,756.7,833.3,4861,830.8,906.2,4960,927.5,977.5,5067,1009,1054,5178,1094,1130,
 5294,1213,5413,1288,5520,5675,6034,6.82,9.94,40,65,464,569,4972,13.58,16.13,52.53,76.54,477.2,565.4,4949,27.49,62.01,86.03,487,
 575.2,4959,43.27,79.17,103.5,506.6,594.9,4978,99.3,123.1,529.6,617.8,4999,119.5,144.5,556.3,644.1,5029,140.8,167.3,585.3,672.9,
 5059,170.4,190,612.9,699.9,5089,192.1,214.8,647.3,734.2,5126,215.9,239.7,680.7,767,5162,265,715.6,801.3,5200,291.5,751.1,836.3,
 5239,787.8,873.4,5281,863.1,944.8,5387,941.9,1022,5495,1044,1098,5607,1131,1179,5726,1221,1260,5848,1346,5974,1425,6087,6249,
 6626,6.74,12,47,77,527,638,5475,14.66,53.23,79.42,529.2,625.9,5443,29.31,68.13,94.52,548.7,642.1,5458,46.71,85.55,111.8,571.5,
 665.4,5482,65.28,105.9,132.9,592.3,685.9,5500,128.1,154.9,618.8,712.3,5526,150.6,178.1,646.8,740.1,5557,173.5,203.1,680,772.7,
 5591,205.8,227.4,709.3,801.7,5624,230.5,254.3,747.4,839.1,5664,255.7,281,783.4,874.6,5704,308.1,820.9,911.4,5745,336.3,858.9,
 948.8,5787,896,988.3,5831,975.8,1063,5941,1060,1146,6058,1168,1225,6174,1260,1311,6302,1355,1396,6431,1487,6563,1570,6682,6852,
 7246,6.767,8.66,49,79,585,703,5996,16.5,58.79,87.45,597,695.6,5966,30.96,74.36,103.2,616.4,716.1,5988,49.16,92.75,121.9,634.4,
 733.1,6002,69.46,113.4,142.7,660.2,759.8,6029,90.64,135.9,165.6,685.5,784.2,6052,160.2,190,715.3,814,6080,184.7,214.8,744.4,
 843.2,6114,209.3,241.9,781.9,879.7,6152,244.4,268,813,910.8,6187,270.8,296.9,854.6,951.2,6231,298.1,325.5,893.3,989.3,6274,
 354.8,933.4,1029,6318,384.2,973.8,1068,6362,1011,1110,6409,1097,1189,6523,1185,1276,6650,1299,1359,6769,1396,1450,6907,1497,
 1539,7042,1634,7181,1721,7306,7482,7895,7.434,14.3,59.4,94.6,655.4,781.6,6550,15.64,20.58,70.09,101.5,671.4,775.3,6520,33.67,
 80.62,112,682.2,786,6531,51.2,98.79,130,706.3,811.8,6556,72.4,121,153,728,831.9,6575,95.75,144.4,176.8,756,861.4,6603,119.3,
 169.1,201.6,786,890,6631,194.5,228.2,819.1,923,6664,221.8,254.7,849.2,953.6,6698,248.3,284,891,993.9,6741,286,311.8,923.8,1027,
 6778,314.4,342.7,969,1070,6826,343.6,373.1,1010,1111,6872,403,1053,1153,6919,435.2,1096,1195,6966,1133,1239,7016,1224,1321,
 7132,1317,1414,7270,1437,1500,7391,1539,1596,7539,1644,1689,7682,1788,7827,1880,7957,8141,8572,7.902,14.7,66,104,724,857,7124,
 16.19,21.93,76.17,110.2,734.1,852.1,7093,30.65,87.05,121.1,745.1,863.1,7104,54.8,106.7,141.1,766.9,884.9,7125,75.01,128.8,
 163.3,792,910.1,7149,99.06,152.7,187.6,820.2,938.3,7177,125,178.3,213.5,851.2,969.3,7207,151.1,205.5,240.9,884.9,1003,7240,
 233.6,269.6,921.1,1039,7275,262.1,299,959,1076,7316,290.2,329.2,998.3,1115,7359,330.8,360,1039,1155,7403,361,391.6,1081,1197,
 7450,392.2,423.8,1125,1240,7499,457,1181,1287,7553,489.3,1216,1329,7599,1262,1375,7651,1358,1460,7769,1456,1559,7918,1582,1648,
 8041,1689,1745,8184,1799,1847,8350,1950,8484,2046,8638,8829,9278,7.864,15.8,73,115,800,940,7725,17.08,76.21,112.9,799.7,925.2,
 7684,33.5,93.62,130.5,829.5,943.7,7702,51.27,112.8,149.5,855.6,973.1,7729,79.5,136.6,174,877.5,991.8,7749,102,161.9,199.9,
 907.4,1025,7776,129,187.6,225.5,937.9,1052,7807,157.8,215.8,254.4,969.2,1086,7836,186.1,245,283.3,1009,1123,7877,275.4,314.2,
 1048,1163,7915,305.3,343.9,1080,1196,7951,336,377.5,1131,1244,8005,379,408.8,1167,1281,8045,411,443.6,1219,1331,8102,444,477.7,
 1266,1376,8154,512,1314,1423,8207,546.6,1362,1470,8260,1397,1519,8315,1505,1606,8433,1603,1711,8595,1735,1803,8718,1846,1910,
 8890,1961,2012,9046,2119,9205,2219,9347,9545,10010,7.637,17,82,125,876,1024,8348,18.17,82.32,121.9,875.3,1009,8306,35.32,100.3,
 140.1,908.4,1028,8325,54.9,120.1,159.7,935.8,1059,8353,76.1,144.7,184.9,957.6,1077,8372,108,170.9,211.9,988.6,1112,8400,133,
 197.1,237.8,1019,1139,8432,162,226.6,268.2,1051,1174,8460,193,256.1,297.2,1092,1211,8503,224.6,287.7,329,1132,1251,8542,321,
 362,1174,1293,8584,352.1,393.3,1207,1328,8620,384,429.1,1262,1379,8680,430.2,462,1299,1419,8720,463.7,498.8,1356,1471,8783,
 498.4,534.7,1405,1519,8838,571.3,1456,1569,8894,607.1,1506,1618,8950,1541,1669,9007,1648,1760,9125,1756,1870,9300,1894,1965,
 9423,2011,2077,9609,2131,2184,9771,2295,9937,2399,10080,10290,10780,7.726,10.64,83,128.8,947,1106,8988,20.29,88.61,131.2,971.5,
 1096,8953,36.84,107.2,149.9,991.1,1116,8972,57.38,127.7,170.3,1020,1149,9001,79.9,152.8,196,1041,1166,9021,103,180.1,224.3,
 1073,1203,9048,139,206.7,250.4,1104,1229,9083,167,237.5,282.4,1137,1267,9108,199,267.2,311.3,1178,1302,9155,232,299.9,344.7,
 1211,1339,9181,266.1,333.5,377.8,1262,1386,9237,368.8,412.9,1307,1431,9282,401,445.8,1340,1467,9317,435,483.8,1400,1522,9384,
 484,518.4,1439,1563,9423,520,557.2,1499,1619,9493,557,594.9,1551,1670,9550,633,1604,1722,9610,670.6,1657,1773,9668,1690,1827,
 9729,1793,1920,9844,1905,2037,10040,2045,2133,10160,2173,2253,10360,2298,2363,10530,2459,10700,2585,10850,11060,11570,9.394,
 17.3,97,145,1037,1203,9667,17.96,26.94,102.1,147.9,1065,1195,9633,39.72,114.3,160.1,1077,1207,9645,59.4,136.5,182.6,1101,1231,
 9669,82.6,161.2,207.5,1129,1259,9696,108,189.6,236.9,1162,1298,9721,136,216.5,263.3,1193,1323,9758,175,248.7,296.9,1227,1362,
 9781,203,278.5,325.7,1267,1397,9831,238,312.8,360.9,1302,1436,9854,274,346.7,393.9,1353,1482,9915,310.8,382.8,429.9,1399,1528,
 9960,419.7,467.1,1448,1576,10010,454,501.5,1481,1613,10040,490,541.6,1546,1673,10120,542,578,1586,1716,10150,579,618.7,1650,
 1775,10230,619,658.3,1704,1828,10290,698,1760,1882,10350,737.4,1815,1936,10420,1846,1992,10480,1953,2087,10590,2070,2210,10800,
 2216,2309,10920,2363,2435,11130,2479,2550,11310,2647,11490,2780,11650,11870,12390],
E_0 = [0.4298,5.996,1.72,3.466,27.4,20.06,3.871,3.427,40.93,6.685,47.59,17.6,6.879,6.658,6.495,61.55,
 2.403,59.84,10.54,65.92,33.36,10.75,9.435,10.26,86.55,10.94,2.991,91.13,3.942,83.7,7.843,84.12,46.24,15.48,11.64,14.82,127,
 18.27,10.94,124.2,19.25,5.853,122,6.225,107,18.62,106,69.43,21.08,13.91,19.94,177.4,17.45,17.59,169,20.67,15.94,162,19.37,
 7.942,159.3,8.467,137.7,9.141,135.4,87.09,27.54,16.58,25.68,239,26.58,28.82,166,27.44,23.07,205.5,22.92,26.29,160.3,25.28,
 18.89,192.5,11.46,173.7,6.423,171.1,113.1,34.85,20,32.04,314.4,34.28,31.87,245.2,34.85,31.92,238.7,34.94,30.92,236,33.16,26.15,
 234.3,27.4,24.96,230.8,28.22,188.6,28.59,219.3,158.6,43.04,5.968,36.55,45.37,421.6,39.11,39.39,326.6,39.57,41.3,242.6,43.14,
 38.07,284.1,35.23,38.68,236,42.12,33.03,280.6,35.52,29.47,277.7,20.97,246.4,17.58,240.6,226.8,52.11,9.393,49.37,45.87,271.1,
 8.139,49.4,51.42,370.9,53.19,48.04,360.7,51.37,48.1,334.5,51.14,47.1,325.9,41.55,45,277.7,40.16,43.41,274.2,24.61,39.67,308.2,
 39.81,267.1,40.41,285.8,204.2,62.03,18.6,12.04,64.45,55.94,367,10.12,61.54,51.71,304.9,10.27,60.16,56.29,320,60.33,57.35,348.2,
 60.49,56.51,343.2,50.32,53.4,326.8,57.46,57.26,339.6,47.19,50.16,319.3,40.84,45.5,330.5,46.39,311.7,44.34,313.7,273.8,72.81,
 22.12,14.13,78.08,70.17,532.2,21.23,13.64,71.54,67.38,458,11.82,66.52,58.43,320.1,12.88,64.82,59.77,313.1,70.21,66.53,346,
 60.07,62.37,379.9,57.86,60.86,375.9,59.11,65.93,344.2,54.86,57.45,367.9,39.5,60.98,335.6,53.5,359.9,48.47,315,275.2,84.47,25.8,
 16.58,88.12,86.32,647.2,25.11,16.66,75.17,72.99,444.5,24.01,18.32,79.43,77.08,502.5,11.51,69.46,65.64,396.9,1.526,61.86,62.7,
 342.3,68.93,69.51,382.8,68.19,70.44,432.7,55.87,72.17,381.1,64.37,67.12,424.3,65.02,76.64,382.4,73.05,77.91,386,57.12,379,
 64.01,371.1,338.1,97.01,29.75,19.16,91.52,104.7,811.4,28.9,20.47,90.58,96.94,730.6,29.34,21.75,92.46,92.33,643.4,33.25,21.43,
 87.44,86.65,552.3,12.71,74.11,73.55,481.3,14.13,57.47,67.8,468.1,78.13,84.39,495.2,76.95,79.05,489.1,69.11,80.31,493.7,73.26,
 75.55,480.2,72.13,73.6,475.8,55.42,67.91,482.8,69.3,466.9,44.14,466.7,439,110.4,33.98,22.31,80.04,109.2,970,33.39,21.91,96.95,
 91.68,561.9,34.6,23.07,95.8,91.51,765,35.05,20.19,95.62,90.76,552,36.54,18.25,86.44,82.04,613,8.603,76.82,72.33,511.1,1.534,
 64.71,61.06,441.4,84.34,83.97,460.3,86.35,88.18,549.1,84.7,91.34,459.3,82.8,84.5,539.7,76.21,97.01,461,39.05,104.9,463.3,73.1,
 447.3,74.44,443.2,417.4,124.7,38.54,25.25,164.7,130.2,1135,37.8,24.73,108.5,101.8,626,38.62,26.12,102.5,103.1,840.8,39.58,
 22.69,107.3,101,617.1,37.32,22.49,98.35,93.93,681.9,35.64,18.39,92.33,89.74,642.7,8.091,85.63,80,619.9,3.884,73.68,65.29,448.6,
 92.3,92.73,467.9,96.39,97.83,612.6,75.62,98.84,478.9,92.97,93.98,602.8,57.22,108.1,473.1,82.7,108.6,474.9,81.62,475.6,54.4,468,
 446.8,139.9,3.824,41.38,29.1,266.6,160.2,1171,41.95,29.85,228.6,148.8,1103,40.55,26.7,119.9,112.1,689.7,47.91,32.96,187.6,
 124.9,844.2,43.82,24.66,118.8,111.3,681.7,54.53,36.02,205.4,118.4,814.7,45.76,34.7,136,115.5,719.3,19.3,141.4,106,694.4,13.52,
 173.7,80.84,503.4,155.8,110,515.8,107.1,108,679.6,137.6,111.2,520.4,103.8,104,669.4,105.4,117.9,529.7,95.96,119.9,527.9,89.79,
 544.2,115.5,537.6,517.7,155.9,7.366,44.87,30.12,152.9,120.1,694.7,4.155,44.98,31.76,144.8,133.5,799.7,46.17,35.04,126,148.3,
 1003,51.69,36.45,141.3,137.3,960.6,56.73,37.22,138.4,131.5,863.5,57.49,39.94,130.3,129.7,850.5,60.97,41,120.8,126.8,826.1,
 52.02,39.89,107,124.3,781.6,25.52,129.3,120.3,740.1,16.05,65.11,84.48,680.5,70.52,93.84,701,118.4,118.7,750.3,89.42,115.4,
 698.9,115.2,114.5,739.6,114.4,112.2,734.2,40.41,116.9,686.2,107.2,723.5,93.95,664,629.7,172.9,8.324,9.733,49.36,32.84,159.3,
 118,736.7,3.837,16.11,49.9,32.46,147.2,127.4,766.9,21.83,51.69,35.13,134.8,144.3,1030,54.68,34.51,142.2,137.1,995.7,54.39,
 30.36,144.4,134.4,830.7,50.34,28.99,126.6,124.9,955.5,53.22,28.91,143.4,133.5,824.5,49.41,20.31,110.6,118.2,942.8,47.63,14.99,
 105,113.1,828.3,11.04,100.7,101.9,769.6,3.414,80.91,78.07,668.2,83.13,115.6,678.1,130.4,130,824.4,78.99,127.4,673.9,127.3,
 125.6,813.3,126.9,136.7,675.2,143.5,138.6,668.4,110.9,660.7,119.7,643.9,626.7,190.7,9.184,11.02,54.24,35.62,181.4,117.7,682.6,
 5.011,17.71,54.94,35.18,163,131.2,695,23.99,56.43,37.86,150.8,154.2,1086,26.24,58.72,37,149.7,145.2,1048,59.02,35.45,146.6,
 144.6,1031,59.4,32.58,157.4,146.4,908,54.84,27.86,129.6,131.3,961.3,58.39,31.19,156.5,145.4,902.7,53.37,17.36,116,124.5,945.4,
 52.65,17.12,113.3,125.4,946.9,9.778,112.2,125.9,931.8,3.359,94.04,92.4,739.4,98.52,125,771.2,143,141.7,902.1,108.4,134.1,769,
 140.1,137.2,890.5,129.1,141.9,761.7,168.6,143.3,754.5,138.8,741.5,102.5,885.4,865.6,209.4,10.02,11.46,59.37,36.68,229,96.88,
 655,14.51,60.49,41.23,130.2,137.8,1467,19.13,61.01,43.39,168.5,179.5,1056,27.94,67.51,37.93,172.3,159.9,996.5,31.9,60.53,45.91,
 172.1,163.3,987.5,67.36,47.22,169.2,158,978.2,64.72,34.9,171,158.9,989.8,69.59,50.13,140,151.3,964,63.86,33.57,170.1,157.9,
 985.4,76.87,54.39,121.8,149.2,944.8,65.49,13.96,121.9,129.7,938.1,14.75,121,119.4,879.1,4.802,117.8,90.1,825.4,119.8,127.2,
 849.1,156.2,154,983.4,141.2,143.4,845.4,153.5,149.3,971.4,176.1,143.7,837.9,132.9,159.7,830.1,134.1,814.6,48.2,797.3,766.9,229,
 9.636,7.244,65.67,42.34,186.5,158.8,1035,15.21,65.99,44.62,147.7,201.3,1497,9.947,64.5,42.93,187.1,173.5,1085,34.83,67.76,
 42.56,172.1,164,1224,31.05,72.9,40.27,186.2,173.1,1081,30.98,66.93,38.3,164,155.1,1128,67.54,35.31,157.1,155.3,1122,70.35,
 37.33,185,172,1076,66.44,24.65,142.7,147.6,1110,69.65,36.06,184.3,171,1073,65.44,25.39,142.6,144.7,1111,66.75,11.64,142.4,
 145.4,1115,11.35,142.7,129.5,1103,5.856,131.2,97.48,992.7,139.7,138.8,1017,170.1,166.8,1068,129.7,151.4,1010,167.5,161.9,1056,
 200.6,174.9,1004,221.6,174.8,1003,148.8,995.2,148,982.4,969.5,249.5,11.83,13.11,70.4,41.14,273.7,83.16,875.8,9.306,16.24,71.41,
 43.59,220.7,134.6,868.4,16.04,71.9,49.56,190.8,201.6,1154,36.15,81.15,45.31,201.5,187.2,1174,37.98,68.92,50.19,190.9,178.2,
 1045,34.52,78.61,42.71,200.7,186.7,1171,38.65,76.18,51.84,173.9,166.8,1021,78.18,53.34,174.8,167.9,1021,76.29,39.86,199.7,
 185.6,1167,72.77,23.64,155.5,155.4,1013,75.74,38.64,199,184.6,1164,74.42,22.73,154.1,152.2,1018,72.88,22.86,153.4,143.7,1025,
 13.45,157.2,141.2,1016,5.053,157.6,108.4,995.2,158.8,146.8,1007,184.7,180.2,1157,188.9,145.2,1005,182.2,175.1,1144,240.3,187.9,
 997.3,257.6,195.2,994.6,163.1,991.6,200.4,999.7,966.1,270.9,12.77,14.07,76.3,43.34,294.8,57.27,804.4,10.14,19.33,77.5,46.63,
 228.1,143.1,893.1,26.7,77.77,52.77,202.3,215.8,1245,41.54,76.25,52.56,203.2,205.9,1155,43.69,72.73,52.18,199.6,181.8,1124,
 44.87,72.6,52.77,195.2,180.3,1113,43.32,74.33,53.65,184,169.4,1110,36.52,75.56,54.5,178.8,169.4,1103,77.04,57.01,171.5,170.6,
 1102,79.14,58.54,164.7,171,1098,79.72,60.6,163.2,171.4,1095,81.5,63.29,138.3,172,1096,82.03,67.12,139.2,172.2,1103,79.93,70.77,
 138.7,172.4,1112,40.64,213.6,196.7,1258,28.73,139.6,136.3,1061,141.2,153.4,1067,199.9,194,1249,133.8,162.3,1068,197.5,188.8,
 1235,197,186,1228,80.7,170.8,1066,179.9,1215,187.3,1087,1057,293.2,13.7,15.81,82.56,50.97,283.2,88.88,626.9,2.419,84.08,56.02,
 198.9,170.9,2079,21.81,84.19,54.35,208.9,217.2,1487,45.3,96.21,53.42,232.6,216.6,1365,43.69,83.77,50.17,207.7,184,1313,43.52,
 93.18,50.38,231.9,216.3,1364,44.33,86.95,44.3,201.5,179.7,1295,42.57,90.95,47.9,231.3,215.8,1363,34.61,90.11,37.65,194.3,180.2,
 1287,90.52,30.99,192.2,180.8,1266,89.09,45.22,230.5,214.6,1362,92.35,21.85,192.5,175.5,1264,88.85,44.12,230.1,213.5,1361,99.18,
 25.29,193.1,171.8,1272,135.8,26.52,193.3,172.6,1277,26.9,205.8,167.3,1277,12.62,205.4,143.3,1249,206.2,164.3,1252,215.7,208.4,
 1344,244.8,170.6,1260,213.4,203,1330,215.9,176,1277,192.3,179.9,1275,159.3,1267,235.2,1249,1251,316.3,14.68,6.063,88.96,54.48,
 304.3,113.2,736.6,1.501,90.66,60.24,220.5,180.8,2249,29.71,90.61,58.23,223,232.4,1647,50.23,104.3,57.74,248.8,232,1466,46.94,
 90.36,53.89,221.5,195.6,1463,48.37,101,54.49,248.2,231.8,1466,50.59,93.85,47.56,219.3,194.7,1430,47.35,98.52,51.79,247.6,231.4,
 1466,42.12,102.3,39.85,218.9,193.5,1439,38.39,103.3,34.48,210.9,193.1,1439,100.5,29.01,208.7,192.5,1440,95.96,48.06,246.7,
 229.9,1467,115.8,20.03,208.8,188.6,1439,95.87,47.02,246.4,228.7,1467,154.3,23.75,210.8,184.1,1448,171.1,25.05,218,185.3,1452,
 25.77,232.7,182.4,1442,10.32,233.4,144.4,1383,235.8,144.1,1400,232.1,223.3,1443,214,160.8,1401,230,217.7,1429,213.9,172.6,1415,
 247.3,164.9,1404,181.3,1382,131.4,1373,1349,340.6,14.36,7.279,96.17,61.98,282.6,250.2,1788,15.9,97.5,64.72,232.8,305.4,2378,
 35.41,97.45,62.47,236.9,250.8,2025,55.39,112.7,62.25,265.4,248,1571,58.75,96.5,57.76,236.8,207.2,1572,53.46,109.2,58.78,264.9,
 247.9,1572,56.52,101,50.85,237.6,196.8,1541,52.36,106.4,55.87,264.4,247.6,1573,50,98.21,43.46,238.3,195.8,1546,52.1,104.5,53.5,
 264,247,1574,46.05,101.7,33.08,232.2,198.3,1550,102.3,33.56,229.1,196,1550,103.1,50.99,263.5,245.7,1576,106.4,34.24,230.2,199,
 1553,103.2,50.01,263.2,244.6,1577,120.7,35.31,238.4,197.4,1558,201.1,37.06,240.1,199.3,1563,26.04,261.9,195.3,1558,15.29,265.1,
 162.6,1523,220.6,190.1,1540,249.3,238.8,1546,199.9,200.3,1556,247.3,233,1531,150.3,211.8,1537,109.9,217.4,1516,182.2,1476,
 173.2,1435,1427,365.6,16.73,18.18,102.5,61.95,348.6,97.55,832,14.11,25.93,104.7,64.43,266.4,212.1,1125,36.4,104.5,66.84,251.9,
 268.6,1981,53.76,103.3,64.27,252.8,236.6,1890,63.78,103,61.83,252.3,223.1,1732,58.78,117.7,63.26,282.1,264.5,1682,65.3,109.6,
 54.92,252.2,209.4,1697,57.6,114.7,60.12,281.6,264.2,1683,59.06,101.6,47.21,251.2,211.4,1605,57.27,112.6,57.54,281.3,263.8,1685,
 57.35,104.6,37.05,254.3,202.3,1611,57.49,105.7,37.41,252.9,211.8,1610,107.3,37.8,253.8,213.5,1617,110.6,54.03,280.7,262.1,1689,
 115.3,39.12,259.6,214.2,1617,110.8,53.11,280.6,260.9,1691,140.3,29.85,272,210.1,1633,253.9,31.16,229.6,208.5,1636,31.15,226.5,
 207,1631,19.6,227.3,195.3,1605,225.4,210.4,1636,267,254.8,1652,203.4,217.9,1629,265.1,248.8,1637,138.6,233.5,1607,123,216.8,
 1582,198.9,1565,321.9,1600,1564,391.5],
sigma_0 = [54750,4470,13690,47.74,156.4,320.1,6083,1423,130.6,49.04,97.96,545.8,3422,364.3,15250,
 96.98,153,103.7,42.73,86.05,284.6,2190,1152,4564,74.21,179.2,1184,66.49,121.9,80.67,71.09,81.11,234.4,1521,10290,772.2,47.48,
 172.4,748.3,50.02,94,190.8,52.35,111,70.46,34.47,73.04,151.9,1117,122000,241.5,32.37,518.6,196.2,35.84,231.8,121.7,39.39,161.9,
 106.3,41.23,96.41,57.35,68.96,60.29,132.9,855.4,277500,109.7,22.95,264.9,24.39,47.98,245.8,71.4,31.44,300.5,35.63,53.02,141.9,
 71.49,36.8,84.53,46.68,84.97,48.9,103.9,675.9,16910,56.15,16.64,276.6,40.25,27.83,245.1,37.97,29.43,221.4,41.9,30.27,196.2,
 54.21,30.97,171.7,55.75,32.32,49.78,50.11,28.97,37.19,66.95,547.5,1.46,248.6,11.42,11.19,355.1,26.29,18.89,274.8,19.33,34.66,
 223.2,31.44,25.51,302.8,26.07,37.53,165,42.18,26.54,137.4,48.02,27.45,61.27,36.13,45.31,38.5,39.95,452.5,3.034,202.3,16.71,
 35.61,3.278,204.9,12.9,17.67,210.4,19.14,18.77,222.3,21.18,22.12,213.4,24.48,23.5,290.1,24.34,32.78,244.2,28.32,34.07,354.6,
 37.55,27.21,39.36,36.99,22.28,32.9,61.4,380.2,182.8,5.384,173.5,14.25,22.06,6.324,184.2,16.42,32.95,4.915,199,13.98,29.94,
 218.6,16.19,24.94,213.2,18.65,25.72,288.8,19.82,27.86,194.1,22.74,26.53,222.9,26.04,29.89,170.6,33.93,28.57,35.16,32.23,20.98,
 32.72,40.36,323.9,184.5,11.66,153.2,11.66,11.84,69.75,8.454,183.2,12.63,16.28,8.666,219.7,15.69,35.91,6.083,232.6,14.86,37.84,
 210.9,14.63,31,278.9,16.48,23.98,277.7,18.57,24.74,232.6,20.49,31.35,203.1,23.94,26.41,222.3,26.24,33.43,31.55,28.32,19.66,
 38.8,47.54,279.3,99.25,11.25,151.2,9.931,9.167,25.64,7.93,234.7,13.28,19.9,28.51,8.26,194.9,12.06,15.77,11.53,266.1,14.91,26.4,
 36.88,336.8,15.35,36.71,298,15.6,29.32,271.2,15.53,21.46,369.6,17.71,29.64,238.4,19.52,22.77,179.2,19.66,29.55,78.67,20.79,
 29.14,29.8,30.53,15.19,32.2,36.46,243.3,56.44,10.03,188.3,8.52,6.649,105.4,7.824,189.6,9.659,8.303,37.02,8.208,178,10.48,10.82,
 10.99,9.08,203.7,11.55,15.04,13.63,291.9,14.07,20.29,9.139,471.2,14.95,21.48,275.1,12.91,19.16,261.5,14.62,19.31,297,16.59,
 19.46,221.3,18.12,20.41,168.9,20.3,21.01,164.1,24.51,20.87,25.71,22.33,21.78,22.93,24.53,213.9,45.39,6.628,305.3,10.59,5.255,
 49.93,7.493,205.2,11.67,16.34,30.44,8.302,222.2,12.65,8.653,21.48,9.833,214.4,11.98,17.13,11.95,11.48,270.1,13.91,13.91,20.8,
 334.5,15.2,20.68,67.82,451.2,16.76,28.36,288,13.92,26.45,250.5,13.75,17.46,237.8,15.02,26.68,205.3,16.84,18.4,183.1,16.46,
 26.66,384.5,16.22,26.51,24.49,28.78,13.56,29.56,31.43,189.4,48.72,6.394,83.72,9.185,4.28,60.06,7.104,193.7,10.96,14.89,39.73,
 7.899,228.1,11.64,8.071,30.58,9.306,200.7,11.2,15.48,24.76,10.48,247.4,12.65,12.69,14.18,13.04,278.6,13.13,14.44,25.96,320.4,
 14.24,15.59,32.95,419.8,16.02,31.43,285.5,13.06,29.31,238.9,12.94,15.85,346.6,14.31,27.96,190.4,15.67,16.66,378.4,15.15,28.88,
 100.6,16.06,28.74,22.35,28.83,18.38,30.03,31.08,169,0.7363,26.14,6.377,31.07,6.389,4.54,73.92,6.625,41.99,7.085,5.173,67.84,
 7.455,186.2,10.39,13.83,33.44,7.619,67.7,9.136,9.086,32.14,9.855,192.2,10.61,14.28,16.38,8.591,55.62,9.849,9.835,11.53,9.467,
 144.9,10.17,12.88,16.19,133.2,11.13,13.89,15.38,85.45,13.95,27.85,110.4,11.19,26.85,227.2,12.18,14.46,116.4,13.06,26.46,176.8,
 14.6,15.15,124.4,14.2,25.62,79.63,14.83,25.96,20.67,24.51,8.875,25.34,25.89,151.7,2.373,90.17,7.227,128.2,10.1,15.86,2.235,
 73.14,6.924,144.6,8.939,11.68,79.3,6.596,194.5,7.77,7.079,42.24,7.134,154.2,8.622,7.782,30.44,7.757,162.2,9.126,9.791,24.83,
 8.068,185.5,9.293,10.12,16.48,8.521,218.1,9.589,10.79,11.27,9.257,278.8,9.859,12.17,13.35,193.4,10.52,13.69,14.37,661.6,13.59,
 16.42,590.9,14.13,15.47,215.7,11.47,13.24,333.6,12.87,15.7,164.3,13.63,13.84,119.9,14.9,14.17,521.5,15.52,16.65,17.88,14.86,
 11.15,18.2,19.36,136.9,2.252,248.8,61.76,6.849,147.3,10.65,15.54,1.909,126.9,57.95,6.97,171.1,10.03,14.37,84.59,54.65,6.785,
 201.4,8.939,7.512,53.77,7.316,182,9.401,8.077,45.85,8.55,172.1,9.326,11.85,45.05,9.855,230.3,10.26,8.84,33.48,10.52,176.5,
 9.523,12.1,26.17,15.47,298,10.81,9.117,14.41,21.55,329,11.24,12.06,29.8,356.8,12.15,14.15,59.76,524,14.2,19.15,507.5,11.3,
 18.69,204.5,10.81,12.16,478,11.95,19.09,152.9,12.74,12.69,115.5,12.7,19.18,48.07,13.25,19.71,17.25,20.38,8.82,21.69,21.82,
 124.1,2.167,547.8,59.24,6.517,129.1,10.68,20.29,1.639,259.7,55.61,6.62,158.4,9.886,19.83,173.1,52.84,6.489,183.4,8.585,7.494,
 92.79,52.37,6.988,187.8,9.082,8.085,55.2,7.729,195.2,9.107,8.364,46.65,8.817,165.6,8.835,10.96,45.27,10.99,247.7,9.948,9.746,
 33.48,10.69,169.4,9.019,11.14,26.11,19.41,306.5,10.45,10.11,14.05,21.17,322.2,10.46,10.08,37.26,330.6,10.49,10.45,68.75,455,
 12.91,17.17,425.6,10.7,15.76,193.7,10.19,11.21,297.2,11.54,15.99,142.5,11.92,11.67,126.8,12.4,16.48,38.56,12.96,16.94,14.08,
 17.73,10.38,12.4,12.34,113.1,2.059,3134,75.99,6.69,85.13,12.12,24.2,715.4,52.64,5.887,270.6,9.535,4.369,3556,53.83,5.872,163.4,
 7.208,8.926,207.1,42.55,6.694,155.3,8.204,9.948,83.29,57.62,6.536,157.1,8.009,10.24,49.31,6.956,162.6,8.337,10.44,46.96,9.007,
 159.4,8.371,10.13,35.41,7.844,239.3,8.749,10.77,33.15,10.77,162.8,8.542,10.25,17.69,8.345,313.6,8.956,11.26,11.89,27.95,315.5,
 10.22,11.44,27.55,321.9,10.89,13.15,51.49,341.8,12.92,15.04,336.5,10.58,14.2,183.5,9.623,10.36,199.2,10.98,14.44,133,11.18,
 10.78,74.17,12.33,14.87,73.74,11.78,15.28,14.63,16.06,19.99,16.94,17.57,103.5,0.6532,1485,53.13,5.602,154,8.555,10.25,1031,
 50.21,5.566,238.2,6.869,4.593,1924,53.88,5.797,148,7.72,9.165,222.4,48.56,6.252,179.5,8.333,7.106,226.9,43.27,6.892,150,7.776,
 9.244,120.1,53.68,7.706,196.3,8.723,8.464,56.75,8.881,213,8.726,8.562,46.87,9.131,153.7,7.934,9.366,43.11,14.19,257.3,9.146,
 8.774,32.57,10.78,156.6,8.093,9.448,24.11,16.32,261.8,9.388,8.763,12.46,36.51,264.4,9.421,8.708,38.94,265.2,10.28,8.914,45.33,
 312.4,12.27,11.18,282.6,9.935,10.63,173.8,9.095,9.61,264.6,10.53,10.86,124.4,10.5,9.975,62.9,10.35,11.14,26.92,10.89,11.25,
 13.33,11.56,7.312,11.99,11.85,95.05,1.549,16680,66.97,6.172,74.98,11.56,15.92,1.189,259200,49.84,5.735,118.5,9.823,16.27,
 199100,47.07,5.285,160.9,6.777,8.979,318.8,36.65,5.863,143.3,7.318,8.541,228.9,52.93,5.951,161,7.66,11.01,235.5,43.67,7.047,
 145,7.373,8.591,97.13,49.35,6.791,193.9,8.196,11.54,49.92,7.236,192.1,8.168,11.52,46.45,9.199,148.2,7.522,8.667,41.69,15.85,
 243,8.799,11.72,31.81,10.74,150.8,7.67,8.718,22.25,19.32,250.9,9.043,11.61,12.05,20.34,255.1,9.475,11.45,35.35,245.5,9.67,
 11.68,58.56,246.4,11.48,12.2,246.6,9.533,11.91,164.7,8.605,8.934,138.1,10.87,12.05,116.5,9.871,9.26,47.44,9.744,12.38,21.59,
 9.881,12.55,12.29,12.77,5.352,12.66,13.09,87.59,1.468,18500,62.98,5.921,71.91,10.76,20.55,1.084,36790,46.24,5.434,124.1,9.289,
 16.66,6301,44.22,5.015,159.1,6.454,8.309,368.3,48.1,5.295,157.8,6.695,9.711,266.1,51.43,5.672,163.6,7.623,10.27,208.2,52.46,
 6.047,170.9,7.682,10.49,170.1,52.1,6.463,192,8.138,10.55,132.3,53.36,6.938,203.2,8.158,10.67,55.3,7.237,220.5,8.13,10.68,46.86,
 7.65,239.2,8.135,10.78,39.22,7.963,244.9,8.138,10.84,30.22,8.143,332.5,8.161,10.83,20.92,8.093,331.6,8.199,10.69,11.35,7.969,
 336.7,8.24,10.52,12.28,149.6,7.555,8.097,12.07,338.3,9.94,11.61,336.8,9.191,11.5,156.1,8.15,8.327,300,10,11.55,109.3,9.298,
 8.619,77.2,9.942,8.773,267.5,11.25,11.81,11.38,9.098,5.833,11.57,11.95,80.99,1.555,4931,45.87,4.896,90.75,10.3,35.82,12.75,43,
 4.666,181,8.001,3.02,1205,41.92,4.874,165.1,6.667,6.149,375.5,31.75,5.159,133.2,6.563,7.414,379.2,42.75,5.619,166.6,7.606,
 8.026,348.3,37.33,6.114,134.4,6.587,7.423,274.8,44.6,6.99,176.7,7.74,8.26,229.9,43.68,7.244,135.7,6.639,7.433,173.5,47.89,
 9.284,190.1,7.76,8.375,50.45,11.93,194.9,7.771,8.671,44.87,9.201,138.3,6.772,7.447,35.95,19.54,197.2,8.017,8.703,29.93,10.52,
 140.3,6.9,7.456,17.86,19.79,199.3,8.247,8.593,6.171,19.77,200.2,8.292,8.527,21.22,178.3,8.543,8.536,27.34,180.4,9.597,8.967,
 181.6,8.754,8.903,148.1,7.728,7.779,98.93,9.597,8.868,102.7,8.771,8.042,73.48,10.43,8.703,49.87,10.74,8.806,12.53,9.026,4.599,
 9.384,9.046,75.1,1.437,1186,43.51,4.611,86.11,9.424,28.36,0.9258,40.75,4.394,162.5,7.632,2.776,1965,39.8,4.586,158.8,6.314,
 5.367,391.8,29.63,4.849,128.8,6.226,6.933,436,40.16,5.242,160.8,7.233,6.925,379.7,34.62,5.709,129.7,6.24,6.929,312.4,41.44,
 6.461,163.9,7.253,7.274,298.8,40.31,6.728,130.8,6.277,6.924,259.6,41.97,8.589,165.3,7.316,7.178,171.4,43.72,10.54,178,7.348,
 7.17,47.42,13.4,182.5,7.399,7.163,43.82,9.15,133.7,6.43,6.916,29.06,22.74,185.2,7.589,7.182,28.9,10.37,135.6,6.549,6.912,10.53,
 23.83,184.6,7.825,7.081,4.708,23.61,174.2,7.851,7.048,23.39,153.9,8.013,7.159,36.74,154,9.47,7.837,153,9.448,7.626,140.5,7.335,
 7.283,143.4,9.953,7.673,96.66,8.285,7.52,82.19,10.52,7.6,31.98,11.45,7.788,11.17,8.149,8.142,8.336,8.382,69.83,0.4681,1027,
 42.75,4.164,109.3,5.938,4.87,3992,38.65,4.139,158.8,4.889,2.659,1181,37.68,4.311,153.7,5.929,3.761,402.8,27.69,4.565,124.8,
 5.914,6.497,386.8,38.04,4.896,153.7,6.884,6.44,401,32.17,5.342,125.5,5.918,6.482,345.2,38.6,5.992,152.9,7.138,6.731,347,37.28,
 6.26,126.3,5.944,6.468,304.1,43.14,7.773,152.5,7.186,6.671,209.3,42.86,7.318,127.4,5.993,6.454,145.8,46.46,11.52,161.5,7.172,
 6.63,48.88,12.54,166.6,7.274,6.633,42.66,9.072,129.5,6.11,6.433,33.99,14.41,167.5,7.271,6.609,27.83,10.19,131.1,6.221,6.419,
 15.69,16.68,158.8,7.417,6.561,3.929,16.53,157.7,7.415,6.515,24.26,133.1,7.587,6.567,24.53,130.7,8.723,6.915,190.6,7.838,6.748,
 133.5,6.97,6.833,179.8,8.422,6.65,91.11,7.838,7.047,186.6,8.876,6.899,189.8,9.074,7.156,11.08,7.658,6.336,8.196,8.057,65.1,
 1.236,10170,39.3,4.094,77.84,9.077,25.86,0.8168,3408,36.18,4.075,132.9,6.713,13.82,1498,35.68,4.057,147.8,5.6,4.228,600.7,
 35.82,4.285,146.7,6.21,4.708,417.2,35.98,4.578,147.3,6.48,5.666,415.1,29.96,5.006,121.5,5.621,6.079,345.6,35.76,5.518,147.6,
 6.772,5.935,380.4,34.56,5.837,122.2,5.637,6.057,318.7,41.18,7.066,149,6.748,6.653,277.7,39.61,6.797,123.1,5.673,6.036,203.3,
 44.54,10.08,146.5,6.987,6.591,112.2,46.64,10.98,148.9,6.804,6.593,48.63,11.86,148.7,6.802,6.532,41.42,8.971,125.4,5.81,5.993,
 32.54,13.95,144.3,6.875,6.536,26.76,10,126.9,5.912,5.972,13.61,20.83,133.3,7.058,6.399,2.886,20.73,187.1,7.155,6.381,21.08,193,
 7.256,6.432,19.39,192.8,7.654,6.675,199.2,7.257,6.4,126.9,6.629,6.423,189.3,7.856,6.509,86.01,7.424,6.617,239.6,8.158,6.765,
 166.3,9.068,7.052,10.25,7.294,3.342,7.032,7.165,60.83],
y_a = [32.88,2.199,32.88,20.35,33.82,7.391,32.88,2.708,121.2,34.19,1166,17.19,32.88,9.5,1.352,73.54,
 9.203,79.15,44.33,190.6,21.63,32.88,5.687,1.568,54.98,33.08,3.085,96.09,14.99,74.71,28.28,74.59,21.83,32.88,2.361,2.306,138,
 88.93,2.793,91,115.2,6.264,94.28,17.33,53.42,42.31,55.47,26.27,32.88,1.364,3.241,381.2,17.28,4.02,189.4,71.36,6.156,110.4,
 73.12,9.708,114.1,22.87,54.86,38.96,56.82,25.35,32.88,1.242,4.297,1257,50,26,50,73.29,7.282,123,50,26,50,70.89,11.94,79.33,
 24.57,58.76,60.75,61.37,26.57,32.88,2.442,5.808,204200,41.79,8.495,60.75,69.37,10.54,65.25,72.03,11.39,67.34,73.97,11.17,69.07,
 64.23,14.27,71.67,26,50,38.36,81.81,33.52,32.88,2.557e+07,322.2,239.5,5.642e+07,22.57,11.03,252.7,50,26,50,83.52,12.5,69.77,50,
 26,50,96.49,13.77,72.37,70.62,16.21,75.12,26.44,49.68,45.49,51.98,53.15,32.88,2.625e+07,10790,23.89,23.74,4.341e+07,4112,57.75,
 134.6,78.13,16.13,140.1,93.01,16.75,71.47,95.63,16.36,61.73,50,26,50,50,26,50,61.6,18.73,51.08,26,50,39.04,53.84,27.78,32.88,
 2.797,434.1,11310,30.94,44.05,219.5,2404,22.75,29.05,1.99e+06,508.2,35.5,27.57,83.53,19.89,32.6,86.33,19.04,33.76,50,26,50,
 87.47,19.83,34.78,50,26,50,64.8,19.45,36.6,26,50,39.13,42.95,35.67,32.88,3.849,22.88,5.765e+06,47.42,258,4.907,74.89,353.7,
 36.23,77.06,353.2,116.9,22.96,21.74,1.356e+06,118.8,27.86,20.17,81.99,22,19.79,50,26,50,50,26,50,68.06,21.21,22.6,50,26,50,
 65.21,22.13,24.87,26,50,38.51,30.34,28.48,32.88,6.712,26.13,2230,65.94,256.2,18,180,70,26,50,9.795,705.8,139.2,33.88,61.63,
 230.3,57.6,23.33,29.45,1111,47,24.86,20.1,34.97,18.45,19.91,50,26,50,47.38,21.29,21.22,50,26,50,67.44,24.03,24.02,73.3,25.73,
 26.56,29.88,30.68,37.41,32.57,32.54,32.88,13.21,32.96,71.93,94.69,3734,7.474,239.6,75.38,51.88,748.3,14.45,794.1,149.8,40.07,
 141,36.39,722.4,93.1,32.02,55.27,257,48.64,23.97,38.54,1656,36.1,25.19,40.92,44.95,23.35,35.55,50,26,50,55.19,21.7,35.68,50,26,
 50,50,26,50,68.22,23.06,37.42,26,50,43.65,44.59,44.05,32.88,22.32,184.3,34.98,24.91,1.856e+06,18,180,70,26,50,33.83,245,42.14,
 20.88,271.4,45,180,70,26,50,37.59,197.6,38.9,20.25,55.24,273.8,38.03,20.42,31.67,1497,39.5,21.95,24.02,40.19,20.02,19.91,50,26,
 50,56.55,23.21,20.99,50,26,50,65.13,25.31,22.98,89.84,26.83,25.23,30.44,27.72,37.97,30.58,31.62,32.88,26.4,170,54.52,26.93,
 3.285e+07,18,180,70,26,50,32.08,231.8,43.8,21.34,184.7,45,180,70,26,50,38.76,202.2,42.84,20.82,49.58,39.45,200.2,42.2,20.94,
 42.08,321.2,42.3,21.36,45.03,708.2,44.19,23.36,20.08,45.08,20.86,17.44,50,26,50,53.29,23.51,19.17,50,26,50,76.63,25.72,20.42,
 72.01,26.88,22.35,30.57,26.15,46.56,28.54,30.39,32.88,2.41e+07,214.3,2229,7.187e+06,104.4,6165,16.97,300.1,8.731e+06,65.61,
 2564,18,180,70,26,50,54.26,1323,1.658e+07,32.59,89.29,45,180,70,26,50,91.14,2262,13470,29.72,94.08,63.08,1032,56.18,29.32,
 45.54,322.3,55.93,27.84,48.26,775.2,48.65,26.04,21.12,56.35,24.56,18.37,50,26,50,58.16,24.58,19.58,50,26,50,65.41,26.09,22.13,
 63.68,27.24,23.51,31.41,28.58,37.32,31.29,32.82,32.88,208.2,14.65,173.6,221.7,24.68,25.63,15950,18.98,524.6,134.9,29.14,32.33,
 17.11,790.9,68.19,38.66,110.8,44.15,1200,99.06,31.69,119.1,76.47,1194,88.11,29.28,62.15,78.19,2341,69.93,29.27,64.16,92.36,
 1977,58.16,28.52,57.51,68.21,988.4,47.68,28.23,44.92,180,70,26,50,698.9,43.71,27.02,35.22,41.29,17.8,31.97,50,26,50,57.66,
 24.61,32.18,50,26,50,50,26,50,104.4,26.18,34.43,26,50,39.59,39.79,39.21,32.88,411.8,20.66,21.86,170.9,60.07,21.72,29.4,582.5,
 52.29,24.76,175.7,55.17,22.15,26.29,66.03,27.88,189.8,52.15,24.06,65.45,33.21,188.1,54.28,23.2,71.24,45,180,70,26,50,38.82,
 192.1,45.17,22.49,76.37,45,180,70,26,50,43.44,226.4,43.55,22.78,78.4,44.54,276.1,43.79,23.03,41.05,345.1,44,23.61,34.94,1109,
 47.59,26.94,27.01,47.39,22.96,23.93,50,26,50,66.5,25.04,24.32,50,26,50,81.34,26.7,26.16,85.57,27.85,27.92,31.67,31.09,37.18,
 32.9,34.52,32.88,429.7,16.1,21.51,157.2,66.29,21.44,24.15,509.5,39.09,24.16,161.7,59.14,21.85,19.66,54.74,26.86,176.8,56.6,
 23.7,57.95,68.51,31.31,177.8,49.55,22.89,62.52,34.34,179.2,48.74,23.19,64.68,45,180,70,26,50,39.99,197.4,45.76,22.69,53.27,45,
 180,70,26,50,45.14,269.1,45.92,23.19,51.51,46.48,273.8,46.54,23.68,51.27,422.5,47.01,24.45,52.88,1244,50.03,26.87,28.84,49.55,
 23.64,27.65,50,26,50,66.09,25.33,27.89,50,26,50,86.86,26.93,29.34,91.98,28.42,30.6,31.33,33.85,41.31,82.25,58.93,32.88,391.4,
 7.037,14.88,77.59,503.7,19.05,23.43,17.49,22.16,163.4,43.85,21.95,385.8,5.763,23.53,358,69.66,30.31,32.9,40,45,180,70,26,50,
 136.9,28.83,530.1,69.51,27.02,37.26,45.79,598.8,69.88,25.83,38.41,45,180,70,26,50,64.4,805.7,50.44,25.7,40.05,45,180,70,26,50,
 75.56,913.6,47,26.27,37.75,45.23,347.5,49.59,24.18,36.78,327.7,49.69,24.92,32.87,970.8,51.98,28.83,30.96,52.75,24.34,29.5,50,
 26,50,69.05,26.03,30.2,50,26,50,94.59,27.43,31.51,101.7,28.49,32.84,32.26,35.87,72.58,38.05,38.84,32.88,523.2,9.671,19.81,
 135.6,55.6,22.58,33.43,14.15,21.75,154.2,46.33,25.72,146.5,9,21,180,70,26,50,54.48,28.17,160.5,50.67,22.86,71.79,40,45,180,70,
 26,50,68.11,34.2,171.9,50.19,22.87,55.69,37.29,181.5,50.17,23.2,56.03,45,180,70,26,50,41.98,234.1,50.42,23.58,56.47,45,180,70,
 26,50,48.49,241.9,52.7,24.28,55.76,49.63,441.3,53.17,24.72,55.4,447.5,53.21,25.55,56.57,878.7,54.94,29.68,43.35,55.99,24.96,
 40.93,50,26,50,74.49,26.59,40.93,50,26,50,101.6,27.9,42.38,109.1,28.92,44.57,32.25,48.41,38.8,51.85,53.51,32.88,2.92e+06,4.497,
 14.85,59.28,395.2,21.87,39.65,1.439e+09,2.411,21.3,101.6,92.12,19.7,24.05,2.538,23.56,254.4,61.98,27.56,28.99,40,45,180,70,26,
 50,76.78,25.94,311.7,60.64,25.15,32.01,40,45,180,70,26,50,113.3,38.79,381.1,53.45,23.94,32.19,47.63,445.4,58.09,24.38,32.56,45,
 180,70,26,50,43.24,257.3,53.79,24.03,33.45,45,180,70,26,50,49.74,277.1,55.96,24.74,33.33,52.54,277.2,56.38,25.35,33.13,418,
 56.6,26.17,34.28,1103,58.14,30.03,36.35,59.79,25.64,33.74,50,26,50,79.15,27.37,34.64,50,26,50,114.7,28.25,36.63,123.4,28.37,
 38.24,32.27,42.56,39.34,47.27,48.29,32.88,111600,4.458,14.79,52.93,321.9,27.85,36.33,25620,3.564,21.55,92.71,80.58,20.26,23.81,
 5.385,23.36,213.1,59.9,27.22,31.7,35.29,22.86,233.2,59.33,27.72,33.29,62.73,24.28,248.2,57.19,24.01,35.34,95.79,27.51,279.7,
 56.9,24.15,35.37,97.91,33.06,316.9,52.38,23.16,35.63,92.77,38.55,360.5,52.93,23.45,35.84,43.8,409.2,52.98,23.87,35.76,51.25,
 467.9,52.76,24.4,35.96,55.74,521.4,54.52,25.16,35.98,60.39,563.7,50.9,25.7,36.12,61.28,588.6,52.37,26.26,35.85,58.07,565,52.79,
 26.93,35.62,180,70,26,50,515,54.59,30.26,37.13,55.69,26.91,34.12,50,26,50,70.1,26.75,35.78,50,26,50,50,26,50,119.1,28.39,41.16,
 26,50,38.49,50.86,57.69,32.88,755.9,6.607,19.73,119.8,76.86,27.97,31.61,28.23,20.74,135.8,52.74,22.74,1067,14.34,22.97,137.9,
 54.57,23.53,52.82,40,45,180,70,26,50,43.96,27.73,148.1,55.03,22.98,50.52,40,45,180,70,26,50,58.29,33.24,165.9,55.82,23.44,
 51.04,40,45,180,70,26,50,76.81,38.07,193.6,57.95,23.95,49.57,40.52,225.8,59.95,24.35,46.46,45,180,70,26,50,45.13,312.5,62.28,
 25.08,46.38,45,180,70,26,50,52.16,294.3,64.07,25.92,46.04,47.15,283,64.53,26.27,45.99,316.6,65.3,27.05,47.6,646.3,66.6,29.96,
 49.09,68.6,26.95,46.35,50,26,50,92.63,27.31,47.9,50,26,50,94.57,27.75,52.67,93.92,28.78,55.13,33.66,60.59,38.73,63.78,67.86,
 32.88,741.1,6.823,19.42,115.7,78.68,27.12,36.22,44.42,20.48,131.4,56.53,22.89,1325,9.751,22.47,133,56.5,23.61,67.58,40,45,180,
 70,26,50,38.74,26.99,143.4,56.99,23.15,63.43,40,45,180,70,26,50,54.17,32.16,160.9,58.6,23.82,63.33,40,45,180,70,26,50,69.23,
 36.59,189.2,61.6,24.16,62.53,79.72,38.78,215,62.01,24.5,62.47,41.68,251.5,63.97,24.82,62.46,45,180,70,26,50,43.8,361.1,66.4,
 25.6,62.48,45,180,70,26,50,47.43,378.6,68.74,26.44,61.37,47.9,362.2,69.51,26.79,60.36,345.5,70.45,27.38,61.22,825.2,71.84,
 31.04,58.39,74.51,27.79,55.28,50,26,50,76.28,28.35,56.81,50,26,50,95.11,27.38,62.29,101.4,29.97,63.96,33.32,67.82,45.22,73.47,
 74.31,32.88,2383,7.988,17.47,115.8,66.88,24.02,76.45,7.098,20.22,127.8,57.76,26.75,748.9,13.9,22.11,129.1,58.47,23.84,225.9,40,
 45,180,70,26,50,37.74,26.18,139.1,59.51,23.37,69.13,40,45,180,70,26,50,50.07,31.21,156.3,61.63,23.79,71.96,40,45,180,70,26,50,
 63.55,36.67,182.3,64.17,24.23,69.64,40,45,180,70,26,50,79.67,42.07,234.7,66.69,24.86,69.17,44.83,244.1,68.6,25.02,69.22,45,180,
 70,26,50,50.25,257.9,71.19,26.04,69.23,45,180,70,26,50,58.49,304.2,74.05,26.91,66.91,50.61,291.8,74.68,27.57,66.16,361.8,76.54,
 28.34,66.67,617.2,77.93,31.55,68.7,63.9,27.95,65.12,50,26,50,77.15,28.03,69.87,50,26,50,105.5,28.57,70.45,127.2,28.92,71.57,
 34.17,72.81,40.99,73.79,79.12,32.88,1029,5.288,18.76,108.6,82.98,32.19,44.97,877.5,7.832,21.22,117.5,62.6,23.42,27.65,11.88,
 21.81,125.5,60.9,23.94,105.7,24.29,23.37,129.5,61.44,23.47,120.2,34.62,25.54,135.2,62.05,23.64,87.51,40,45,180,70,26,50,46.08,
 30.34,151.2,63.76,24.01,90.66,40,45,180,70,26,50,58.5,36.06,176.1,65.62,24.43,68.74,40,45,180,70,26,50,71.43,41.8,221.6,69.61,
 25,67.99,78.04,44.77,231.3,71.56,25.36,67.74,47.76,238.9,74,25.75,67.68,45,180,70,26,50,53.08,288.3,77.22,26.31,68.39,45,180,
 70,26,50,60.7,357.7,80.71,27.48,67.53,53.84,344.8,63.76,27.96,66.93,335.5,63.08,28.65,67.91,529.3,63.47,31.73,70.53,66.45,
 28.13,68.64,50,26,50,80.99,28.52,70.85,50,26,50,116.9,27.69,71.73,133,29.77,72.46,35.02,78.35,39.04,92.72,90.95,32.88],
P = [2.963,6.098,2.963,4.423,1.49,2.916,2.963,10.64,1.348,3.738,1.022,3.157,2.963,5.469,12.18,1.438,
 9.374,1.436,3.501,1.21,2.624,2.963,6.336,10.85,1.503,4.15,14.8,1.338,7.489,1.442,4.754,1.428,2.581,2.963,8.821,9.139,1.252,
 3.348,9.956,1.335,3.194,9.711,1.335,6.719,1.552,3.606,1.538,2.315,2.963,11.4,8.037,1.083,4.995,7.999,1.185,3.618,7.271,1.289,
 3.648,8.183,1.287,6.061,1.54,4.943,1.533,2.336,2.963,12.49,7.303,0.9638,4,4.5,1.65,3.596,6.543,1.263,4,4.5,1.65,3.628,6.03,
 1.377,5.771,1.511,5.214,1.501,2.255,2.963,10.43,6.678,0.845,4.029,6.038,1.42,3.645,5.652,1.424,3.563,5.554,1.422,3.563,5.907,
 1.416,3.819,5.562,1.411,4.5,1.65,3.992,1.396,2.002,2.963,3.789,3.57,3.38,0.7736,4.643,5.618,1.12,4,4.5,1.65,3.511,5.405,1.414,
 4,4.5,1.65,3.367,5.458,1.405,3.675,5.395,1.397,5.246,1.579,4.689,1.575,1.678,2.963,3.923,2.96,4.742,1.952,3.61,2.995,3.903,
 1.225,3.63,5.05,1.238,3.513,4.941,1.413,3.438,4.951,1.464,4,4.5,1.65,4,4.5,1.65,4.366,4.95,1.539,4.5,1.65,3.986,1.56,2.161,
 2.963,10.84,4.088,2.762,4.399,1.588,4.481,2.92,4.81,1.898,3.477,3.11,4.267,1.876,3.609,4.749,1.732,3.522,4.763,1.73,4,4.5,1.65,
 3.46,4.594,1.721,4,4.5,1.65,3.927,4.926,1.712,4.5,1.65,4.062,1.676,1.915,2.963,9.721,5.334,2.639,3.933,1.102,9.525,4.676,3.133,
 4.172,1.385,4.208,3.529,4.803,2.123,3.353,3.547,4.565,2.186,3.588,4.611,2.094,4,4.5,1.65,4,4.5,1.65,3.718,4.515,2.02,4,4.5,
 1.65,4.118,4.52,1.982,4.5,1.65,4.148,1.925,2.135,2.963,8.516,5.205,2.795,3.617,1.063,7.2,4.2,3.7,4.5,1.65,8.21,3.682,3.391,
 4.192,1.467,4.447,3.967,4.787,1.901,4.89,4.256,4.787,2.232,4.307,4.938,2.123,4,4.5,1.65,4.244,4.617,2.087,4,4.5,1.65,3.727,
 4.317,2.01,3.527,4.191,1.945,4.45,1.885,3.995,1.869,2.002,2.963,7.513,5.038,3.633,3.346,0.8646,8.421,3.95,3.635,3.738,0.9844,
 7.321,3.591,3.319,3.968,1.221,5.977,3.618,3.565,4.227,1.517,4.361,4.142,4.754,1.736,3.626,4.742,4.826,1.738,4.11,4.545,1.742,4,
 4.5,1.65,4.011,4.615,1.737,4,4.5,1.65,4,4.5,1.65,4,4.707,1.72,4.5,1.65,4.48,1.668,1.765,2.963,6.896,4.196,4.457,4.205,0.7888,
 7.2,4.2,3.7,4.5,1.65,6.261,3.95,4.019,4.703,1.105,5.9,4.2,3.7,4.5,1.65,6.009,4.245,4.239,4.934,1.527,4.689,4.431,5.148,1.86,
 4.756,4.637,5.347,2.121,4.242,4.904,2.182,4,4.5,1.65,3.889,4.515,2.148,4,4.5,1.65,3.852,4.266,2.085,4.272,4.067,2.025,4.421,
 1.999,4.103,1.953,2.037,2.963,6.662,4.223,3.328,4.021,0.7631,7.2,4.2,3.7,4.5,1.65,6.347,3.961,4.046,4.657,1.16,5.9,4.2,3.7,4.5,
 1.65,6.142,4.127,4.125,4.848,1.562,6.194,4.303,4.227,4.919,1.659,4.685,4.329,5.094,1.664,4.645,4.492,5.323,2.315,4.165,4.862,
 2.362,4,4.5,1.65,4.195,4.545,2.271,4,4.5,1.65,4.151,4.251,2.234,3.789,4.173,2.171,4.418,2.074,4.439,2.037,2.092,2.963,4.427,
 5.631,3.587,2.067,3.159,0.8392,7.208,3.891,2.239,3.449,0.8917,7.2,4.2,3.7,4.5,1.65,5.626,3.423,2.363,4.13,1.343,5.9,4.2,3.7,
 4.5,1.65,5.035,3.242,2.349,4.28,1.355,5.52,3.365,3.706,4.325,1.628,4.079,3.674,4.499,1.635,3.884,3.547,4.988,2.275,3.554,4.565,
 2.338,4,4.5,1.65,3.609,4.472,2.282,4,4.5,1.65,3.764,4.262,2.182,3.884,4.169,2.144,4.401,2.005,3.844,1.966,2.018,2.963,4.841,
 7.498,4.165,3.087,4.592,1.966,4.313,7.152,3.771,3.3,4.269,1.744,7.186,3.588,3.77,3.891,1.256,5.864,3.447,3.446,4.152,1.271,
 5.265,3.398,3.521,4.281,1.479,5.203,3.235,3.707,4.3,1.482,5.007,3.218,3.907,4.353,1.531,5.408,3.326,4.2,4.388,1.646,4.2,3.7,
 4.5,1.65,3.857,4.937,5.021,1.842,4.878,5.361,1.858,4,4.5,1.65,4.171,4.561,1.851,4,4.5,1.65,4,4.5,1.65,4.452,4.389,1.823,4.5,
 1.65,4.176,1.777,1.862,2.963,4.588,10.22,7.081,4.207,3.635,4.911,1.937,5.018,7.3,6.87,4.198,3.796,4.755,1.953,6.452,6.613,
 4.068,3.952,4.456,1.41,6.239,4.071,3.854,4.579,1.415,5.9,4.2,3.7,4.5,1.65,6.159,4.175,4.139,4.764,1.426,5.9,4.2,3.7,4.5,1.65,
 6.056,4.346,4.348,4.836,1.428,6.085,4.455,4.411,4.893,1.707,4.545,4.463,5.033,1.825,4.577,4.674,5.297,2.06,4.637,4.853,2.11,4,
 4.5,1.65,4.304,4.532,2.101,4,4.5,1.65,3.618,4.303,2.052,3.452,4.218,2.022,4.371,1.971,4.054,1.96,1.998,2.963,4.552,10.96,7.123,
 4.268,3.531,5.085,2.187,4.958,7.825,6.917,4.26,3.739,4.872,2.29,6.742,6.696,4.124,3.868,4.508,1.469,6.3,6.379,4.126,3.967,
 4.645,1.476,6.226,4.142,4.006,4.635,1.478,5.9,4.2,3.7,4.5,1.65,6.137,4.259,4.216,4.817,1.581,5.9,4.2,3.7,4.5,1.65,6.042,4.42,
 4.355,4.873,1.603,6.025,4.407,4.373,4.83,1.603,4.552,4.376,4.779,1.604,4.559,4.543,5.16,2.02,4.485,4.83,2.005,4,4.5,1.65,4.056,
 4.576,1.997,4,4.5,1.65,3.655,4.373,1.972,3.355,4.278,1.955,4.198,1.915,4.265,1.477,1.622,2.963,4.565,13.77,7.586,4.706,2.767,
 5.757,2.326,10.38,7.058,4.188,4.365,4.927,1.042,11.87,6.904,3.847,3.723,4.183,1.738,7,5.9,4.2,3.7,4.5,1.65,5.672,6.578,3.633,
 3.697,4.427,1.748,5.859,3.553,3.712,4.52,1.745,5.9,4.2,3.7,4.5,1.65,5.46,3.393,4.15,4.587,1.74,5.9,4.2,3.7,4.5,1.65,5.183,3.28,
 4.379,4.576,1.777,5.865,4.463,4.332,4.886,1.793,4.457,4.34,4.98,1.891,4.521,4.329,5.244,1.969,4.289,4.902,1.968,4,4.5,1.65,
 3.834,4.573,1.955,4,4.5,1.65,3.402,4.461,1.935,3.596,4.246,1.918,4.336,1.887,4.724,1.874,1.928,2.963,4.641,15.75,7.258,4.374,
 3.823,4.789,1.822,11.23,7.102,4.23,4.261,4.266,1.189,14.5,7.2,4.2,3.7,4.5,1.65,6.484,6.628,4.224,3.993,4.721,1.455,7,5.9,4.2,
 3.7,4.5,1.65,6.392,6.344,4.245,4.06,4.81,1.58,6.191,4.264,4.114,4.789,1.582,5.9,4.2,3.7,4.5,1.65,6.057,4.369,4.228,4.847,1.586,
 5.9,4.2,3.7,4.5,1.65,5.923,4.286,4.188,4.837,1.589,5.86,4.49,4.181,4.803,1.588,4.496,4.178,4.945,1.588,4.495,4.251,5.204,1.752,
 4.153,4.858,1.756,4,4.5,1.65,3.96,4.591,1.755,4,4.5,1.65,3.33,4.278,1.74,3.193,4.221,1.717,4.31,1.687,4.083,1.668,1.696,2.963,
 4.113,16.46,7.643,4.989,2.793,6.149,1.947,4.095,17.97,7.133,4.563,3.46,5.386,2.207,18.08,6.977,4.005,3.835,4.332,1.859,7,5.9,
 4.2,3.7,4.5,1.65,6.206,6.855,3.866,3.847,4.612,1.902,7,5.9,4.2,3.7,4.5,1.65,5.798,6.116,3.719,4.041,4.767,1.919,5.832,3.624,
 3.973,4.732,1.914,5.9,4.2,3.7,4.5,1.65,6.012,4.393,4.166,4.866,1.909,5.9,4.2,3.7,4.5,1.65,5.829,4.34,4.141,4.856,1.905,5.787,
 4.326,4.14,4.916,1.902,4.462,4.109,4.9,1.894,4.49,4.081,5.139,1.885,4.044,4.859,1.907,4,4.5,1.65,3.619,4.741,1.893,4,4.5,1.65,
 3.19,4.277,1.868,3.086,4.199,1.847,4.294,1.799,3.8,1.749,1.794,2.963,4.112,16.91,7.672,5.129,2.837,6.635,2.118,4.167,15.66,
 7.138,4.64,3.572,5.376,2.262,12.32,7.017,4.093,3.878,4.355,1.826,7.056,7.043,4.036,3.879,4.41,1.869,6.344,7.028,3.991,3.924,
 4.752,1.866,5.914,6.823,3.91,3.953,4.757,1.874,5.905,6.509,3.824,4.082,4.905,1.873,6.15,6.265,3.741,4.107,4.887,1.876,6.059,
 3.643,4.154,4.85,1.878,5.835,3.561,4.204,4.817,1.878,5.729,3.487,4.187,4.774,1.88,5.612,3.421,4.446,4.741,1.878,5.596,3.36,
 4.41,4.71,1.875,5.676,3.329,4.407,4.675,1.871,4.2,3.7,4.5,1.65,3.846,4.366,4.885,1.889,4.328,4.849,1.922,4,4.5,1.65,4.143,
 4.711,1.895,4,4.5,1.65,4,4.5,1.65,4.194,4.5,1.827,4.5,1.65,3.998,1.722,1.718,2.963,4.337,15.32,7.331,4.513,3.416,5.913,2.476,
 18.43,7.214,4.323,4.078,5.044,0.9653,11.14,7.043,4.336,3.992,4.602,1.571,7,5.9,4.2,3.7,4.5,1.65,6.895,6.744,4.342,3.993,4.903,
 1.675,7,5.9,4.2,3.7,4.5,1.65,6.482,6.378,4.35,4.018,4.914,1.682,7,5.9,4.2,3.7,4.5,1.65,6.518,6.102,4.355,4.029,4.878,1.697,
 6.001,4.398,4.013,4.849,1.731,5.9,4.2,3.7,4.5,1.65,5.839,4.423,3.98,4.854,1.732,5.9,4.2,3.7,4.5,1.65,5.592,4.315,3.953,4.84,
 1.73,5.247,4.297,3.945,4.812,1.727,4.21,3.868,4.82,1.714,4.28,3.855,4.928,1.719,3.824,4.851,1.738,4,4.5,1.65,3.432,4.722,1.717,
 4,4.5,1.65,3.535,4.612,1.668,3.631,4.518,1.651,4.493,1.619,3.808,1.609,1.613,2.963,4.342,22.23,7.372,4.548,3.408,5.643,2.316,
 19.28,7.246,4.347,3.99,5.06,0.954,11.4,7.093,4.367,3.977,4.604,1.485,7,5.9,4.2,3.7,4.5,1.65,7.119,6.796,4.37,3.98,4.91,1.583,7,
 5.9,4.2,3.7,4.5,1.65,6.505,6.437,4.38,3.969,4.88,1.599,7,5.9,4.2,3.7,4.5,1.65,6.485,6.079,4.394,3.932,4.868,1.598,6.44,5.977,
 4.412,3.969,4.851,1.599,5.919,4.429,3.955,4.836,1.599,5.9,4.2,3.7,4.5,1.65,5.646,4.438,3.924,4.825,1.598,5.9,4.2,3.7,4.5,1.65,
 5.176,4.245,3.886,4.816,1.599,5.019,4.225,3.842,4.785,1.602,4.232,3.763,4.78,1.602,4.304,3.746,4.981,1.644,3.705,5.132,1.653,4,
 4.5,1.65,3.752,4.878,1.64,4,4.5,1.65,3.615,4.768,1.599,3.435,4.706,1.592,4.426,1.58,4.477,1.558,1.588,2.963,4.224,20.33,7.555,
 4.488,3.668,4.576,1.451,16.16,7.277,4.366,3.997,4.155,0.9937,10.17,7.13,4.39,3.967,4.581,1.184,7,5.9,4.2,3.7,4.5,1.65,6.878,
 6.87,4.397,3.955,4.917,1.56,7,5.9,4.2,3.7,4.5,1.65,6.558,6.491,4.41,3.925,4.979,1.562,7,5.9,4.2,3.7,4.5,1.65,6.458,6.291,4.413,
 3.89,4.96,1.57,7,5.9,4.2,3.7,4.5,1.65,6.285,6.02,4.436,3.889,4.901,1.57,5.919,4.373,3.881,4.908,1.57,5.9,4.2,3.7,4.5,1.65,
 5.723,4.283,3.847,4.826,1.568,5.9,4.2,3.7,4.5,1.65,5.405,4.13,3.78,4.791,1.575,4.854,4.111,3.767,4.743,1.577,4.24,3.661,4.738,
 1.576,4.26,3.637,4.881,1.581,3.955,4.791,1.59,4,4.5,1.65,3.888,4.659,1.559,4,4.5,1.65,3.944,4.513,1.56,4.081,4.445,1.56,4.489,
 1.567,4.316,1.577,1.583,2.963,4.259,17.36,7.456,4.614,3.391,5.888,2.215,4.393,13.63,7.201,4.506,3.864,4.988,2.242,10.82,7.163,
 4.414,3.949,4.574,1.361,7.985,7.076,4.424,3.938,4.803,1.363,7.005,6.932,4.422,3.933,4.891,1.484,7,5.9,4.2,3.7,4.5,1.65,6.555,
 6.524,4.431,3.914,4.975,1.488,7,5.9,4.2,3.7,4.5,1.65,6.43,6.396,4.431,3.896,4.934,1.597,7,5.9,4.2,3.7,4.5,1.65,6.214,6.115,
 4.447,3.838,4.972,1.598,6.097,5.995,4.386,3.823,4.875,1.6,5.875,4.334,3.794,4.839,1.597,5.9,4.2,3.7,4.5,1.65,5.645,4.155,3.739,
 4.8,1.594,5.9,4.2,3.7,4.5,1.65,5.261,4.196,3.661,4.771,1.591,4.592,4.181,3.995,4.758,1.592,4.197,4.019,4.736,1.59,4.225,4.011,
 4.697,1.588,3.978,4.728,1.584,4,4.5,1.65,3.908,4.612,1.574,4,4.5,1.65,4.016,4.506,1.574,4.013,4.514,1.577,4.434,1.557,3.684,
 1.497,1.535,2.963],
y_w = [0,0,0,0,0,0,0,0,0,0,0,0,0,0.5608,0,0,0,0,0,0,0,0,0.4474,0,0,0.5276,0,0,0,0,0,0,0,0,0.4239,0,0,
 0.4209,0,0,0.5496,0,0,0,0,0,0,0,0,0.4103,0,0,0.02182,0,0,0.05538,0,0,0.0376,0,0,0,0,0,0,0,0,0.3857,0,0,0.3,0,0,0.139,0,0,0.3,0,
 0,0.3418,0,0,0,0,0,0,0,0,0.3345,0,0,0.3052,0,0,0.2042,0,0,0.2573,0,0,0.2895,0,0,0.168,0,0,0,0,0,0,0,0,0,0.1465,0,0,0.29,0,0,
 0.3,0,0,0.1644,0,0,0.3,0,0,0.2182,0,0,0.1613,0,0,0,0,0,0,0,0,0,0.01463,0,0,0,2.223e-05,0,0,0.2851,0,0,0.005276,0,0,0.005419,0,
 0,0.3,0,0,0.3,0,0,0.006452,0,0,0,0,0,0,0,0,0.3076,0,0.02337,0,0,0,0.007839,0,0,0,0.02016,0,0,0.09152,0,0,0.1041,0,0,0.3,0,0,
 0.1206,0,0,0.3,0,0,0.1386,0,0,0,0,0,0,0,0,0.2921,0,0.0002774,0,0,0.3169,0,0.000287,0,0,0,0.0008658,0,0,0,0.0008417,0,0,
 6.329e-05,0,0,0.3,0,0,0.3,0,0,2.813e-05,0,0,0.3,0,0,4.76e-05,0,0,0,0,0,0,0,0,0.2765,0,0.0003422,0,0,0.26,0,0.1,0,0,0.329,0,
 0.005039,0,0,0,0.004482,0,0,0,0.001118,0,0,0.001008,0,0,0.3,0,0,0.001094,0,0,0.3,0,0,0.001364,0,0,0.001286,0,0,0,0,0,0,0,0,
 0.2621,0,0.2485,0,0,0.284,0,0.2934,0,0,0.2989,0,0.02142,0,0,0.3761,0,0.009497,0,0,0,0.02785,0,0,0,0.0248,0,0,0.02108,0,0,0.3,0,
 0,0.02178,0,0,0.3,0,0,0.3,0,0,0.04183,0,0,0,0,0,0,0,0,0.2479,0,0.2017,0,0,0.26,0,0.1,0,0,0.2707,0,0.2237,0,0,0.26,0,0.1,0,0,
 0.4036,0,0.2462,0,0,0,0.2396,0,0,0,0.2356,0,0,0.2672,0,0,0.3,0,0,0.3329,0,0,0.3,0,0,0.3987,0,0,0.3017,0,0,0,0,0,0,0,0,0.2355,0,
 0.627,0,0,0.26,0,0.1,0,0,0.2643,0,0.007167,0,0,0.26,0,0.1,0,0,0.1882,0,0.007283,0,0,0.07398,0,0.007408,0,0,0,0.007258,0,0,0,
 0.007712,0,0,0.008883,0,0,0.3,0,0,0.01392,0,0,0.3,0,0,0.01194,0,0,0.005509,0,0,0,0,0,0,0,0,0.0002049,0.2437,0,0.5274,0,0,
 0.2531,0,0.556,0,0,0.26,0,0.1,0,0,0.314,0,0.4546,0,0,0.26,0,0.1,0,0,0.4138,0,0.5284,0,0,0.247,0,0.6504,0,0,0,0.7309,0,0,0,
 0.8892,0,0,0.7774,0,0,0.3,0,0,0.8872,0,0,0.3,0,0,0.928,0,0,1.487,0,0,0,0,0,0,0,0,0.0005841,0.2754,0,0.003343,0,0,0.3539,0.2735,
 0,0.0003358,0,0,0.2658,0,0.0004791,0,0,0.3052,0,0.001107,0,0,0.3557,0,0.0004384,0,0,0.38,0,1.4e-05,0,0,0.4261,0,4.346e-06,0,0,
 0.2204,0,4.591e-06,0,0,0,0.1,0,0,0,7.881e-06,0,0,7.117e-06,0,0,0.3,0,0,1.104e-05,0,0,0.3,0,0,0.3,0,0,1.164e-05,0,0,0,0,0,0,0,0,
 0.0001441,0.3104,0.2671,0,0.003208,0,0,0.01415,0.3103,0.2658,0,0.003289,0,0,0.389,0.2567,0,0.003368,0,0,0.2941,0,0.003269,0,0,
 0.26,0,0.1,0,0,0.006694,0,0.003408,0,0,0.26,0,0.1,0,0,0.00605,0,0.003645,0,0,0.005576,0,0.003675,0,0,0,0.00364,0,0,0,0.00366,0,
 0,0.003716,0,0,0.3,0,0,0.009238,0,0,0.3,0,0,0.0148,0,0,0.04096,0,0,0,0,0,0,0,0,0.003612,0.2972,0.2645,0,5.519e-05,0,0,0.001633,
 0.2982,0.2642,0,1.643e-05,0,0,0.365,0.2571,0,1.691e-05,0,0,0.3297,0.2839,0,1.717e-05,0,0,0.2921,0,1.732e-05,0,0,0.26,0,0.1,0,0,
 0.009213,0,1.795e-05,0,0,0.26,0,0.1,0,0,0.007441,0,1.68e-05,0,0,0.008018,0,1.686e-05,0,0,0,1.686e-05,0,0,0,1.753e-05,0,0,
 1.75e-05,0,0,0.3,0,0,1.465e-05,0,0,0.3,0,0,1.321e-05,0,0,3.372e-05,0,0,0,0,0,0,0,0,0.0005057,0.3417,0.269,0,0.0009964,0,0,
 0.4665,0.2583,0,6.49e-05,0,0,0.5113,0.2586,0,0.01158,0,0,0.3,0.26,0,0.1,0,0,0.1237,0.2283,0,0.01225,0,0,0.3123,0,0.0118,0,0,
 0.26,0,0.1,0,0,0.283,0,0.01006,0,0,0.26,0,0.1,0,0,0.407,0,0.0147,0,0,0.5604,0,0.01503,0,0,0,0.01495,0,0,0,0.01528,0,0,0.01591,
 0,0,0.3,0,0,0.0164,0,0,0.3,0,0,0.01814,0,0,0.02681,0,0,0,0,0,0,0,0,9.332e-05,0.776,0.2601,0,0.005785,0,0,0.4827,0.2576,0,
 0.006156,0,0,0.35,0.26,0,0.1,0,0,0.3436,0.2673,0,0.005538,0,0,0.3,0.26,0,0.1,0,0,0.003113,0.2242,0,0.005648,0,0,0.2358,0,
 0.005706,0,0,0.26,0,0.1,0,0,0.3135,0,0.00596,0,0,0.26,0,0.1,0,0,0.2753,0,0.005989,0,0,0.3057,0,0.005962,0,0,0,0.005905,0,0,0,
 0.006239,0,0,0.005775,0,0,0.3,0,0,0.006187,0,0,0.3,0,0,0.001634,0,0,0.00354,0,0,0,0,0,0,0,0,0.03256,0.3881,0.2659,0,0.04661,0,
 0,0.000874,0.4774,0.2616,0,0.1974,0,0,0.6536,0.2565,0,0.02549,0,0,0.3,0.26,0,0.1,0,0,0.06108,0.1836,0,0.02553,0,0,0.3,0.26,0,
 0.1,0,0,0.1437,0.3087,0,0.01659,0,0,0.3093,0,0.01717,0,0,0.26,0,0.1,0,0,0.3835,0,0.01402,0,0,0.26,0,0.1,0,0,0.4047,0,0.01435,0,
 0,0.0001633,0,0.01436,0,0,0,0.01411,0,0,0,0.01417,0,0,0.01415,0,0,0.3,0,0,0.01713,0,0,0.3,0,0,0.01237,0,0,0.01723,0,0,0,0,0,0,
 0,0,0.03238,0.4039,0.2646,0,0.06314,0,0,0.01598,0.4144,0.2599,0,0.1223,0,0,0.4407,0.2557,0,1.309e-05,0,0,0.3316,0.2449,0,
 1.308e-05,0,0,0.2402,0.1365,0,1.32e-05,0,0,0.02368,0.02105,0,1.33e-05,0,0,0.02604,0.02404,0,1.377e-05,0,0,0.02051,0.02667,0,
 1.398e-05,0,0,0.02673,0,1.508e-05,0,0,0.02574,0,1.574e-05,0,0,0.02828,0,1.594e-05,0,0,0.02772,0,1.114e-05,0,0,0.03441,0,
 1.107e-05,0,0,0.06512,0,1.111e-05,0,0,0,0.1,0,0,0,1.179e-05,0,0,1.177e-05,0,0,0.3,0,0,1.631e-05,0,0,0.3,0,0,0.3,0,0,1.901e-05,
 0,0,0,0,0,0,0,0,0.03355,0.3676,0.2573,0,4.833e-05,0,0,4.806e-05,0.2553,0,0.0001431,0,0,0.01266,0.2547,0,8.474e-06,0,0,0.3,0.26,
 0,0.1,0,0,0.0001079,0.2457,0,8.505e-06,0,0,0.3,0.26,0,0.1,0,0,0.0001086,0.3125,0,8.773e-06,0,0,0.3,0.26,0,0.1,0,0,0.0001706,
 0.4132,0,8.959e-06,0,0,0.462,0,9.089e-06,0,0,0.26,0,0.1,0,0,0.5704,0,9.096e-06,0,0,0.26,0,0.1,0,0,0.5707,0,9.111e-06,0,0,0.824,
 0,9.099e-06,0,0,0,8.972e-06,0,0,0,8.999e-06,0,0,9.07e-06,0,0,0.3,0,0,9.87e-06,0,0,0.3,0,0,1.182e-05,0,0,2.239e-05,0,0,0,0,0,0,
 0,0,0.03908,0.006227,0.2566,0,1.68e-05,0,0,0.03708,0.2545,0,0.0001246,0,0,0.4019,0.2543,0,4.038e-07,0,0,0.3,0.26,0,0.1,0,0,
 0.003263,0.2509,0,4.065e-07,0,0,0.3,0.26,0,0.1,0,0,0.0001508,0.3229,0,4.138e-07,0,0,0.3,0.26,0,0.1,0,0,0.0001416,0.4805,0,
 4.163e-07,0,0,0.000122,0.5276,0,4.273e-07,0,0,0.5228,0,4.274e-07,0,0,0.26,0,0.1,0,0,0.743,0,4.306e-07,0,0,0.26,0,0.1,0,0,
 0.8373,0,4.288e-07,0,0,0.8478,0,4.401e-07,0,0,0,4.221e-07,0,0,0,4.4e-07,0,0,4.687e-07,0,0,0.3,0,0,5.528e-07,0,0,0.3,0,0,
 6.838e-07,0,0,2.062e-06,0,0,0,0,0,0,0,0,0.03736,1.582,0.2599,0,9.174e-07,0,0,0.7439,0.2537,0,2.543e-08,0,0,0.3416,0.2537,0,
 2.525e-08,0,0,0.3,0.26,0,0.1,0,0,0.3091,0.246,0,2.525e-08,0,0,0.3,0.26,0,0.1,0,0,0.002171,0.3331,0,2.511e-08,0,0,0.3,0.26,0,
 0.1,0,0,0.00204,0.2759,0,2.517e-08,0,0,0.3,0.26,0,0.1,0,0,0.004503,0.3632,0,2.403e-08,0,0,0.3996,0,2.421e-08,0,0,0.26,0,0.1,0,
 0,0.4731,0,2.441e-08,0,0,0.26,0,0.1,0,0,0.4555,0,2.405e-08,0,0,0.8398,0,2.341e-08,0,0,0,1.879e-08,0,0,0,1.894e-08,0,0,
 1.558e-08,0,0,0.3,0,0,1.496e-08,0,0,0.3,0,0,1.972e-08,0,0,2.821e-08,0,0,0,0,0,0,0,0,0.03962,0.4667,0.2559,0,1.125e-08,0,0,
 0.0004021,0.3562,0.2534,0,5.913e-09,0,0,0.3519,0.2533,0,6.087e-09,0,0,0.3026,0.247,0,6.071e-09,0,0,0.3108,0.2426,0,6.078e-09,0,
 0,0.3,0.26,0,0.1,0,0,0.2476,0.3502,0,6.079e-09,0,0,0.3,0.26,0,0.1,0,0,0.2472,0.145,0,5.943e-09,0,0,0.3,0.26,0,0.1,0,0,0.3888,
 0.1997,0,5.027e-09,0,0,0.3731,0.2234,0,5.33e-09,0,0,0.2432,0,5.551e-09,0,0,0.26,0,0.1,0,0,0.4056,0,5.496e-09,0,0,0.26,0,0.1,0,
 0,0.5034,0,4.757e-09,0,0,0.8255,0,5.868e-09,0,0,0,6.018e-09,0,0,0,6.017e-09,0,0,6.413e-09,0,0,0.3,0,0,6.713e-09,0,0,0.3,0,0,
 1.579e-08,0,0,2.25e-08,0,0,0,0,0,0,0,0]
};

%%%%%%%%%%%%%%%%%%%%%%%%%%
define ionized_phabs_sigma()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ionized_phabs_sigma}
%\synopsis{calculates the photoabsorption cross section for ionized material}
%\usage{Double_Type sigma = ionized_phabs_sigma(Double_Type E);}
%\description
%    \code{E} is the energy in keV (may be a scalar or an array),
%    \code{sigma} is the corresponding cross section in cm^2.
%    The ion of interest is specified by the following
%\qualifiers{
%\qualifier{Z}{atomic number (1 <= \code{Z} <= 30)}
%\qualifier{Nel}{number of electrons (\code{Z} >= \code{Nel} > 0)}
%\qualifier{n}{principal quantum number for the selection of a subshell (together with l)}
%\qualifier{l}{orbital angular momentum quantum number for the selection of a subshell}
%}
%\seealso{Verner & Yakovlev (1995)}
%!%-
{
  variable E;
  switch(_NARGS)
  { case 1: E = (); E *= 1000; }  % keV / eV
  { help(_function_name()); return; }

  variable Z = qualifier("Z", 1);       % atomic number
  variable Nel = qualifier("Nel", 1);   % number of electrons

  variable sum_sigma = Double_Type[length(E)];  % sum_sigma[*] = 0.;
  variable i = where(VernerYakovlev1995_Table1.Z==Z and VernerYakovlev1995_Table1.Nel==Nel);
  ifnot(length(i))
    vmessage("warning (%s): no cross section available for Z=%S, Nel=%S", _function_name(), Z, Nel);
  else if(qualifier_exists("n") || qualifier_exists("l"))
  {
    variable j = i[ where(    VernerYakovlev1995_Table1.n[i]==qualifier("n")
			  and VernerYakovlev1995_Table1.l[i]==qualifier("l")) ];
    if(length(j))
      i = j;
    else
    {
      vmessage("error (%s): n=%S, l=%S is no valid subshell", _function_name(), qualifier("n"), qualifier("l"));
      message("  possible values for the n and l qualifiers:");
      variable n_l = 100*VernerYakovlev1995_Table1.n[i] + VernerYakovlev1995_Table1.l[i];
      n_l = n_l[unique(n_l)];
      foreach n_l ( n_l[array_sort(n_l)] )
        vmessage("    Z=%d, Nel=%d, n=%d, l=%d", Z, Nel, n_l/100, n_l mod 100);
      return;
    }
  }
  foreach i (i)
  {
    variable l       = VernerYakovlev1995_Table1.l      [i];
    variable E_th    = VernerYakovlev1995_Table1.E_th   [i];
    variable E_0     = VernerYakovlev1995_Table1.E_0    [i];
    variable sigma_0 = VernerYakovlev1995_Table1.sigma_0[i];
    variable y_a     = VernerYakovlev1995_Table1.y_a    [i];
    variable P       = VernerYakovlev1995_Table1.P      [i];
    variable y_w     = VernerYakovlev1995_Table1.y_w    [i];
    if(qualifier_exists("verbose"))
      vmessage("(#%d): Z=%d, Nel=%d, n=%d, l=%d, E_th=%f, E_0=%f, sigma_0=%f, y_a=%f, P=%f, y_w=%f",
	       i, VernerYakovlev1995_Table1.Z[i], VernerYakovlev1995_Table1.Nel[i], VernerYakovlev1995_Table1.n[i], l, E_th, E_0, sigma_0, y_a, P, y_w);
    variable y = E/E_0;
    variable Q = 5.5 + l - 0.5*P;
    variable sigma = [sigma_0 * ( (y-1)^2 + y_w^2 ) * y^(-Q) * ( 1 + sqrt(y/y_a) )^(-P) * 1e-18];  % cm^2 / Mb
    sigma[where(E < E_th)] = 0;
    sum_sigma += sigma;
  }

  if(length(sum_sigma)>1)
    return sum_sigma;
  else
    return sum_sigma[0];
}

%
% Code from src/Mike/isis_emcee_hammer.sl
%
% -*- mode:slang; mode:folding -*-

% v. 1.0.0 -- June 27, 2017 Major change!  chain is now returned as
% Array_Type[nfreepar][nwalkers*nsteps], rather than
% Double_Type[nfreepar,nwalkers*nsteps].  This avoids some potential
% issues of creating chain files to big to store in a S-lang array.
% Such overly large arrays are still possible in the calculation
% steps, but can be somewhat mitigated by using a small sim_count
% value and writing out the chains at intermediate steps.

%%%  NOTE: Breaks backward compatability with v0.7.x !!!
%%%  v0.8.+ uses a new file format for output.  Conversion
%%%  routine included to change old fits files into new format

% A simple MCMC code for ISIS, based upon the parallel "simple
% stretch" method from "emcee: the MCMC Hammer" by Foreman-Mackey et
% al., (2012); arXiv:1202.3665

% Written by M. Nowak, mnowak@space.mit.edu, with modifications by
% Tobias Beuchert, tobias.beuchert@sternwarte.uni-erlangen.de
% Lia Corrales, lia@space.mit.edu
% Matthias Kuehnel, matthias.kuehnel@sternwarte.uni-erlangen.de
% Sebastian Falkner, sebastian.falkner@fau.de
% Extended for MPI-based Parallelization by Ashkbiz Danehkar,
% 08/03/2017 (v. 0.9.4) 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Some definitions for people with older versions of S-lang

#ifnexists wherefirst %{{{
define wherefirst (a)
     {
        variable i = where (a);
        if (length(i))
          return i[0];
        else
          return NULL;
     }
#endif
%}}}
#ifnexists wherefirstmax %{{{
public define wherefirstmax(a)
{
   return wherefirst (a==max(a));
}
#endif
%}}}
#ifnexists wherelast %{{{
define wherelast (a)
     {
        variable i = where (a);
        if (length(i))
          return i[-1];
        else
          return NULL;
     }
#endif
%}}}
#ifnexists wherelast_le %{{{
public define wherelast_le(a,b)
{
   return wherelast (a <= b); 
}
#endif
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Set the free model parameters from an array of parameter values.
% Return a 1 if any of the steps violates a parameter boundary

private define set_pars(fp,x) %{{{
{
   variable i;
   _for i (0,length(x)-1,1)
   {
      if( x[i] < get_par_info(fp[i]).min ||
          x[i] > get_par_info(fp[i]).max    )
      {
         return 1;
      }
      else
      {  
         set_par(fp[i],x[i]);
      }
   }
   return 0;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Evaluate the log probability difference between the current walker
% and the proposal.  This presumes that P(D|M) is given by
% P(M)*exp(-chi^2/2), where P(M) is uniform within the min/max bounds
% of the parameter file, and 0 outside of that range. Note that this
% presumes that for the already known walkers, one has the fit
% statistic already calculated (sent via xstat).

private define gauss_prior(priors) %{{{
{
    % Expected INPUTS:
    % priors    = struct{parlist, priorlist}
    % where
    %    parlist   = list of parameter ids that have priors
    %    priorlist = list of prior structures with field names "center", "sigma"
    % The priors are assumed to be gaussian and are not normalized
    % with respect to the hard limits
    %
    % OUTPUT:
    % Sum of log-probabilities to be added to the log-likelihood of
    % the raw fit statistic (logpn)
    % Note that chi^2 stat used in logpn is positive (negative is
    %   propogated through dlogpn), so this needs to return a positive
    %   value.  Also do not divide by two because of treatment in
    %   logpn
    
    variable i, par, pprob, result = 0.0;
    _for i (0,length(priors.parlist)-1, 1)
    {
	par   = get_par(priors.parlist[i]);
	pprob = priors.priorlist[i];
	result += (par - pprob.mean)^2 / (pprob.sigma^2);
    }

    %message(sprintf("Adding %.2f to the chi-squared stat", result));
    return result;
}
%}}}

private define logpn(fp,xstat,y,z) %{{{
{
    % First, deal with priors
    variable log_prior = 0.0;
    if (qualifier_exists("priors")) log_prior = gauss_prior(qualifier("priors"));

    % Now do the rest
    variable infoy, ystat=1.e32, dlogpn=-1.e32, fv_hold = Fit_Verbose, 
            i, nfp = length(fp);
    
    ifnot(set_pars(fp,y))
    {
	Fit_Verbose = -1;
	
	() = eval_counts(&infoy);  
	ystat  = infoy.statistic + log_prior;  % Log-likelihood = log-fitstat + log-priors
	%message(sprintf("Total ystat = %.2f", ystat));
	dlogpn = (xstat-ystat)/2. +  (nfp-1)*log(z);
	
	Fit_Verbose = fv_hold;
    }

    return dlogpn,ystat;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Revamped output routine so that it appends the file, rather than
% write it from scratch each time

public define write_chain( fname, nw, freepar, cdf_scl_lo, cdf_scl_hi, %{{{
                           chain, chainstat, chng, pfile )
{
   variable i, fp;
   if(qualifier_exists("create"))
   {
      fp = fits_open_file(fname,"c");

      %%% First extension: list of free parameters.
      %%%
      variable extname = "PARAMETERS";
      variable colname = ["FREE_PAR"];
      variable nrows   = length(freepar);
      variable ttype   = ["J"];
      variable tunit   = [" parameter indices"];

      fits_create_binary_table(fp, extname, nrows, colname, ttype, tunit);

      fits_update_key(fp, "PFILE", pfile, "");
      fits_update_key(fp, "MODEL", get_fit_fun, "");

      variable dlist;
      list_data(&dlist);
      ifnot(__is_initialized(&dlist)) dlist="   Data not defined at time of file write";
      variable hist_comment = 
         ["This file contains the results of a Markov chain analysis",
          "generated by the emcee hammer subroutine",
          "This extension contains information about the model",
          "PFILE are the parameter files that started the chain",
          "MODEL is the fit function that was applied to the data",
          "FREE_PAR are the parameter indices of the free parameters",
          "The following is the list of data fit:","",
          strchop(dlist,'\n',0)
         ];
      array_map(Void_Type, &fits_write_comment, fp, hist_comment);

      if(_fits_write_col(fp,fits_get_colnum(fp,"FREE_PAR"),1,1,freepar))
         throw IOError;

      %%% Second extension: chain results for each walker/parameter
      %%%

      extname = "MCMCCHAIN";
      colname = ["FITSTAT","UPDATE"], ttype = ["D","J"], 
      tunit =  [" fit statistics"," update indicator"];
      nrows = length(chainstat);
      variable nfrpar = length(freepar);
      variable nsteps = length(chainstat[*,0]);

      _for i (0,length(freepar)-1,1)
      {
         colname = [colname,"CHAINS"+string(freepar[i])];
         ttype = [ttype,"D"];
         tunit = [tunit, " parameter values"];
      }

      fits_create_binary_table(fp, extname, nrows, colname, ttype, tunit);

      fits_update_key(fp, "NWALKERS", nw, "");
      fits_update_key(fp, "NFREEPAR", nfrpar, "");
      fits_update_key(fp, "NSTEPS", nsteps, "");
      fits_update_key(fp, "CDF_SCL_LO", cdf_scl_lo, "");
      fits_update_key(fp, "CDF_SCL_HI", cdf_scl_hi, "");

      hist_comment = 
         ["A Markov chain generated by the emcee hammer subroutine",
          "NWALKERS is the number of walkers *per* free parameter",
          "NFREEPAR is the number of free parameters",
          "NSTEPS is the number of iterations for each walker",
          "CDF_SCL_LO: The step amplitude distribution goes as 1/sqrt(z),",
          "CDF_SCL_HI: bounded by z = 1/CDF_SCL_LO -> CDF_SCL_HI","",
          "The following file columns are the results unpacked",
          "as 1D vectors of length NWALKERS*NFREEPAR*NSTEPS:","",
          "FITSTAT contains the vector of walker fit statistics",
          "UPDATE is a yes/no answer as to whether a walker updated",
          "CHAINS# are the values for the free parameter given by",
          "parameter index #" 
         ];
      array_map(Void_Type, &fits_write_comment, fp, hist_comment);

      _for i (0,nfrpar-1,1)
      {
         if(_fits_write_col(fp,
             fits_get_colnum(fp,"CHAINS"+string(freepar[i])),1,1,chain[i,*]))
            throw IOError;
      }

      _for i (0,nsteps-1,1)
      {
         if(_fits_write_col(fp,fits_get_colnum(fp,"FITSTAT"),1+i*nw*nfrpar,1,
            chainstat[i,*])) throw IOError;

         if(_fits_write_col(fp,fits_get_colnum(fp,"UPDATE"),1+i*nw*nfrpar,1,
            chng[i,*])) throw IOError;
      }

      %%% Third extension: chain statistics at each iteration
      %%%
      extname = "CHAINSTATS";
      colname = ["FRAC_UPDATE","MIN_STAT","MED_STAT","MAX_STAT"];
      nrows   = length(chainstat[*,0]);
      ttype   = ["D","D","D","D"];
      tunit   = [" fraction"," chi2"," chi2"," chi2"];

      fits_create_binary_table(fp, extname, nrows, colname, ttype, tunit);

      hist_comment = 
         ["This extension contains some useful summary information",
          "for the individual chain steps", 
          "FRAC_UPDATE is the fraction of walkers that updated",
          "MIN_STAT: minimum chi2 for a given step",
          "MED_STAT: median chi2 for a given step",
          "MAX_STAT: maximum chi2 for a given step"
         ];
      array_map(Void_Type, &fits_write_comment, fp, hist_comment);

      variable update=Double_Type[nrows], mnstat=@update, mdstat=@update, mxstat=@update;
      _for i (0,nrows-1,1)
      {
         update[i] = sum(chng[i,*])/length(chng[i,*]);
         mnstat[i] = min(chainstat[i,*]);
         mdstat[i] = median(chainstat[i,*]);
         mxstat[i] = max(chainstat[i,*]);
      }

      if(_fits_write_col(fp,fits_get_colnum(fp,"FRAC_UPDATE"),1,1,update))
         throw IOError;

      if(_fits_write_col(fp,fits_get_colnum(fp,"MIN_STAT"),1,1,mnstat))
         throw IOError;

      if(_fits_write_col(fp,fits_get_colnum(fp,"MED_STAT"),1,1,mdstat))
         throw IOError;

      if(_fits_write_col(fp,fits_get_colnum(fp,"MAX_STAT"),1,1,mxstat))
         throw IOError;
   }
   else
   {
      %%% Second extension (first unchanged): chain results
      %%%
      fp = fits_open_file(fname+"[MCMCCHAIN]","w");

      variable nsteps_prev = fits_read_key(fname+"[MCMCCHAIN]","NSTEPS");

      nfrpar = length(freepar);
      nsteps = length(chainstat[*,0]);

      nrows = nsteps_prev*nfrpar*nw;

      _for i (0,nfrpar-1,1)
      {
         if(_fits_write_col(fp,fits_get_colnum(fp,"CHAINS"+string(freepar[i])),
                                         nrows+1,1,chain[i,*]))
            throw IOError;
      }

      _for i (0,nsteps-1,1)
      {
         if(_fits_write_col(fp,fits_get_colnum(fp,"FITSTAT"),nrows+1+i*nw*nfrpar,1,chainstat[i,*]))
            throw IOError;

         if(_fits_write_col(fp,fits_get_colnum(fp,"UPDATE"),nrows+1+i*nw*nfrpar,1,chng[i,*]))
            throw IOError;
      }

      fits_update_key(fp, "NSTEPS", nsteps+nsteps_prev, "");

      %%% Third extension: chain statistics
      %%%
      fp = fits_open_file(fname+"[CHAINSTATS]","w");

      nrows   = length(chainstat[*,0]);
      update=Double_Type[nrows], mnstat=@update, mdstat=@update, mxstat=@update;
      _for i (0,nrows-1,1)
      {
         update[i] = sum(chng[i,*])/length(chng[i,*]);
         mnstat[i] = min(chainstat[i,*]);
         mdstat[i] = median(chainstat[i,*]);
         mxstat[i] = max(chainstat[i,*]);
      }

      if(_fits_write_col(fp,fits_get_colnum(fp,"FRAC_UPDATE"),nsteps_prev+1,1,update))
         throw IOError;

      if(_fits_write_col(fp,fits_get_colnum(fp,"MIN_STAT"),nsteps_prev+1,1,mnstat))
         throw IOError;

      if(_fits_write_col(fp,fits_get_colnum(fp,"MED_STAT"),nsteps_prev+1,1,mdstat))
         throw IOError;

      if(_fits_write_col(fp,fits_get_colnum(fp,"MAX_STAT"),nsteps_prev+1,1,mxstat))
         throw IOError;
  }
}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private variable fs = stderr;

% A subroutine to read out a chain from a FITS file, of the format
% from the subroutine above.  If the qualifier init_walker is set, it
% will spit out an array suitable for restarting the chain.
   
public define read_chain() %{{{
{
   if(_NARGS==1)
   {
      variable fname = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  (nw, nfreepar, freepar, 
   chain_statistic, chain)       = read_chain(fname);

  (nw, nfreepar, freepar, 
   cdf_scl_lo, cdf_scl_hi,
   pfile, walkers_init, 
   statistic_init, change)       = read_chain(fname; init_walker);

   frac_update                   = read_chain(fname; frac_update);

  (frac_update, min_statistic,
   med_statistic, max_statistic) = read_chain(fname; chain_stats);

   Read the MCMC results stored in a fits file.  Normally, it will
   return just the minimum results required for plotting.  Qualifiers,
   however, can be used to have it return enough information to
   initialize a chain, return an array that has the fraction of
   walkers updated at each step, or return the fractional update array
   along with the min, median, and maximum value of the statistic at
   each step.
`); %}}}
      return;
   }

   variable i, nw, nfreepar, nsteps, cdf_scl_lo, cdf_scl_hi, pfile, 
            freepar, frac_update, chain, chain_stat, chain_chng, 
            tmp_chain, tmp_stat, tmp_chng, indces; 

   variable statinit, walkinit;

   (pfile) = fits_read_key(fname+"[PARAMETERS]","PFILE");

   (nw, nfreepar, nsteps, cdf_scl_lo, cdf_scl_hi ) = 
      fits_read_key(fname+"[MCMCCHAIN]","NWALKERS","NFREEPAR","NSTEPS",
                                        "CDF_SCL_LO","CDF_SCL_HI");

   freepar = fits_read_col(fname+"[PARAMETERS]","FREE_PAR");

   if( qualifier_exists("init_walker") )
   {
      chain = Double_Type[nfreepar,nw*nfreepar]; 
      indces = [nsteps*nw*nfreepar-nw*nfreepar:nsteps*nw*nfreepar-1];

      _for i (0,nfreepar-1,1)
      {
         tmp_chain = fits_read_col(fname+"[MCMCCHAIN]","CHAINS"+string(freepar[i]));
         chain[i,*] = tmp_chain[indces];
         () = __tmp(tmp_chain);
      }

      tmp_stat = fits_read_col(fname+"[MCMCCHAIN]","FITSTAT");
      chain_stat = tmp_stat[indces];
      () = __tmp(tmp_stat);

      tmp_chng = fits_read_col(fname+"[MCMCCHAIN]","UPDATE");
      chain_chng = tmp_chng[indces];
      () = __tmp(tmp_chng);

      return nw, nfreepar, freepar, cdf_scl_lo, cdf_scl_hi, pfile, 
             chain, chain_stat, chain_chng;
   }
   else if(qualifier_exists("frac_update"))
   {
      frac_update = fits_read_col(fname+"[CHAINSTATS]","FRAC_UPDATE");
      return frac_update;
   }
   else if(qualifier_exists("chain_stats"))
   {
      variable min_statistic, med_statistic, max_statistic;
      (frac_update, min_statistic,
       med_statistic, max_statistic)=fits_read_col(fname+"[CHAINSTATS]",
                                           "FRAC_UPDATE", "MIN_STAT",
                                           "MED_STAT", "MAX_STAT"     );
      return frac_update, min_statistic, med_statistic, max_statistic;
   }
   else
   {
      chain = Array_Type[nfreepar];

      _for i (0,nfreepar-1,1)
      {
         chain[i] = fits_read_col(fname+"[MCMCCHAIN]","CHAINS"+string(freepar[i]));
      }
      chain_stat = fits_read_col(fname+"[MCMCCHAIN]","FITSTAT");
      reshape(chain_stat,[nsteps,nw*nfreepar]);

      return nw, nfreepar, freepar, chain_stat, chain;
   }
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The inverse of the Cumulative Distribution Function for 1/sqrt(z),
% if 1/a <= z <= b, to use as a random number generator for a
% 1/sqrt(z) probability distribution.  Parameter steps will have an
% amplitude governed by random draws from this distribution.

private define icdf(x,a,b) %{{{
{
   variable c1 = 1/(sqrt(a*b)-1);
   variable c2 = 1/c1^2/a;
   return c2*(x+c1)^2;
}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Return an array with the indices of the free parameters

private define free_par() %{{{
{
   variable i, free_par = Integer_Type[0];
   _for i (1,get_num_pars,1)
   {
      if(get_par_info(i).freeze == 0 &&
         get_par_info(i).tie == NULL)
      {
         free_par = [free_par,i];
      }
   }
   return free_par;
}
%}}}

   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Initialize nw walkers *per free parameter*, whose indices are in fp.
% For now, we will initialize uniformly, or as a gaussian, over the
% range given for the min/max values of the free parameters, or over
% the min/max values for an input parameter file, both modulo a scale
% factor (==1 for full min/max range, uniform draw)

private define init_walker(fp,nw) %{{{
{
   variable npfile=1, fname = "/tmp/mcmc_pfile_"+string(getpid)+".par";

   try{
      save_par(fname);
   }
   catch AnyError: {
      () = printf("\n\n Failed saving initial parameters \n\n");
       plot_pause;
   }

   if(qualifier_exists("pfile"))
   {
      variable pfile_init, pfiles = strchop(qualifier("pfile"),';',0);
      npfile = length(pfiles);

      try{
         foreach pfile_init (pfiles)
         {
            load_par(pfile_init);
         }
      }
      catch AnyError: {
         () = printf("\n\n Cannot read parameter file: %s \n",pfile_init);
         () = printf(" Retaining original parameter file for initialization. \n\n");
         npfile = 1;
         pfiles = [fname];
         load_par(fname);
      }
   }
   else
   {
      pfiles = [fname];
   }

   variable sc, sig;
   if(qualifier_exists("gaussian"))
   {
      sc = qualifier("scale",1./3.);
   }
   else
   {
      sc = qualifier("scale",1.);
   }

   variable fv_hold = Fit_Verbose;
   Fit_Verbose = -1;

   % The walkers are 2D arrays, with the first index covering the free
   % parameters, and the second index being the specific instances
   % (i.e., the length of the 2nd index is the number of walkers).

   variable i, j, k, info, nfp = length(fp),
            xp    = Double_Type[nfp, npfile], xl=@xp, xh=@xp, 
            xstrt = Double_Type[nfp,nw*nfp],
            xstat = Double_Type[nw*nfp];

   _for k (0,npfile-1,1)
   {   
      load_par(pfiles[k]);
      () = eval_counts;

      _for i (0,nfp-1,1)
      {
         xp[i,k]      = get_par_info(fp[i]).value;
         xl[i,k]      = get_par_info(fp[i]).min;
         xh[i,k]      = get_par_info(fp[i]).max;
      }
   }

   % Reload the original parameter file, which should have the widest
   % allowable min/max values (enough to encompass any ranges present
   % in the input parameter files)

   if(qualifier_exists("pfile"))
   {
      try{ 
         load_par(fname);
      }
      catch AnyError: {
         () = printf("\n\n Failed reloading original  parameters \n\n");
         plot_pause;
      }
      if(eval_counts)
      {
         () = printf("\n\n Failed evaluating original  parameters \n\n");
          plot_pause;
      }
   }

   _for j (0,nw*nfp-1,1)
   {
      k = j*npfile/(nw*nfp);

      _for i (0,nfp-1,1)
      {
         if(qualifier_exists("gaussian"))
         {
            sig = (@(qualifier("grnd", &grand)))();
            if(sig < 0)
            {
               xstrt[i,j] = xp[i,k] + sig*sc*(xp[i,k]-xl[i,k]);
            }
            else
            {
               xstrt[i,j] = xp[i,k] + sig*sc*(xh[i,k]-xp[i,k]);
            }
         }
         else
         {
            xstrt[i,j] = (1-sc)*xp[i,k] + 
              sc*(xl[i,k]+(xh[i,k]-xl[i,k])*(@(qualifier("urnd", &urand)))());
         }

         if(xstrt[i,j]<xl[i,k]) xstrt[i,j] = xl[i,k];
         if(xstrt[i,j]>xh[i,k]) xstrt[i,j] = xh[i,k];
         set_par(fp[i],xstrt[i,j]);
      }

      () = eval_counts(&info);
      variable stat = info.statistic;
      if (qualifier_exists("priors"))
      {
	  stat += gauss_prior(qualifier("priors"));
      }
      xstat[j] = stat;
   }

   Fit_Verbose = fv_hold;

   return xstrt, xstat;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The function to create a proposition (i.e., the new trial
% parameters), and test whether or not that proposition is accepted.

private define update_walker(cdf_scl_lo,cdf_scl_hi,rnd,fp,xa,xstat,xb) %{{{
{
   % xa is a vector of parameters for a specific walker to be updated,
   % while xstat is the fit statistic of that walker. xb is an array
   % of walker parameters (a second, independent group) from which an
   % update is drawn.  The first index of xb spans the parameter,
   % while the second spans specific walker instances.

   variable lb = length(xb[0,*]);                   % No. of walkers
   variable xuse = xb[*,int(rnd[0]*lb)];            % Choose one
   variable z = icdf(rnd[1],cdf_scl_lo,cdf_scl_hi); % Random step size

   % New trial parameters
   xuse += z*(xa-xuse); 

   %variable priors = qualifier("priors",0);
   variable dlogpn, ystat;
   (dlogpn,ystat) = logpn(fp,xstat,xuse,z;; __qualifiers);

   variable ptest = log(rnd[2]);
   if(ptest <= dlogpn)
   {
      return xuse, ystat, 1;             % 1 = changed value
   }
   else
   {
      return xa, xstat, 0;               % 0 = same value
   }
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This variable will be used to communicate with the slaves running
% the code in parallel, as well as receive their results.

private variable Task_List = 
   struct{ index, length, rands, free_par, walkers, wstats, change, 
           proposals, output };

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This is the function that the slaves will be running.  It will
% continously loop, receiving instructions, running the walker update
% for a specific index, and spitting back the results, until it gets
% sent a negative index.

private define slave_task(s, k, cdf_scl_lo, cdf_scl_hi) %{{{
{
   send_msg(s, SLAVE_READY);

%   message("Slave "+string(k)+" Starting");

   variable robjs, index, rands, freepar, walkers, wstats, props, 
            output, nstat, chng;

   forever
   {
      robjs   = recv_objs (s);

      index   = robjs[0];
      if(index < 0) break;

      rands   = robjs[1];   % Vector of three random numbers
      freepar = robjs[2];   % Vector of free parameter indices
      walkers = robjs[3];   % Parameter vector for a single walker
      wstats  = robjs[4];   % Fit statistic for above walker
      props   = robjs[5];   % Parameter vectors for walker group

      %variable priors = qualifier("priors", 0);
      (output, nstat, chng) = update_walker(cdf_scl_lo, cdf_scl_hi, rands, 
                                      freepar, walkers, wstats, props;; __qualifiers);

      send_msg (s, SLAVE_RESULT);
      send_objs (s, index, output, nstat, chng);
   }

%   message("Slave "+string(k)+" Ending");

   return 0;
}
%}}}

     
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Stolen from the conf_loop.sl script in the ISIS distribution (in the
% .../isis-[version-minorversion]/share directory).  This function is
% for the parallelized version of the code.  Once all slaves are
% finished (i.e., show a SLAVE_READY status), shut them down by
% sending a negative index

private define maybe_finished (slaves) %{{{
{
   variable s;
   foreach s (slaves)
   {
      if (s.status != SLAVE_READY) return;
   }

   foreach s (slaves)
   {
      send_objs (s, -1);
   }
}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% A function to send tasks (indices, walker arrays, random number
% arrays, etc.) to the slaves running in parallel

private define send_next_task (s) %{{{
{
   if(Task_List.index < Task_List.length)
   {
      variable index, rands, freepar, walker, wstat, props;

      index   = Task_List.index;
      rands   = Task_List.rands[index];
      freepar = Task_List.free_par;
      walker  = Task_List.walkers[index];
      wstat   = Task_List.wstats[index];
      props   = Task_List.proposals;

      send_objs (s, index, rands, freepar, walker, wstat, props);
      Task_List.index++;
      s.status = SLAVE_RUNNING;
   }
   else maybe_finished (s);
}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% A function to receive the results from the slaves.  Here the return
% status isn't really necessary, but see the ISIS script conf_loop.sl
% for an example of using a return status to decide whether or not to
% start a new set of slaves.

private define recv_slave_result(s) %{{{
{
   variable objs = recv_objs(s);
   Task_List.output[objs[0]] = objs[1];
   Task_List.wstats[objs[0]] = objs[2];
   Task_List.change[objs[0]] = objs[3];
   s.status = SLAVE_READY;
   return 0;
}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This is the simple function that manage_slaves will continously loop
% over.  It will listen to the slaves and either send them a task, or
% receive a result, and then send them a task.  The details of the
% sending and receiving are handled in separate functions.

private define slave_handler(s,msg) %{{{
{
   switch (msg.type)
   {
    case SLAVE_READY: send_next_task (s);
   }
   {
    case SLAVE_RESULT: 
      () = recv_slave_result(s);
      send_next_task(s);
   }
}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The function to do the emcee hammer.  It takes as input the number
% of walkers per free parameter, and the number of simulations per set
% of walkers.  Optionally, it can take a qualifier to read the
% parameter min/max values from other file(s).  Or it can be
% initialized from an existing chain.

public define emcee() %{{{
{
   if(_NARGS==2)
   {
      variable nw, nsim;
      (nw, nsim) = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
   emcee( nw, nsim; pfile="file1;file2;...;filen", 
                    scale=#, gaussian, 
                    cdf_scl_lo=#, cdf_scl_hi=#, 
                    init_chain=file, 
                    output=file, sim_count=#, 
                    num_slaves=#, nice=#, serial,
                    nocopy, priors, urnd=&urand, grnd=&grand );

  Create a set of MCMC chains for an ensemble of walkers (i.e.,
  propositions) consisting of a number of walkers equal to nw X the
  number of free parameters.  This ensemble will be evolved for nsim
  iterations.  All results will be stored in an output file.

  The initial distribution of walkers is taken from the current
  parameter values, or from the parameter values in the files listed
  in pfile (file names, including full paths if not in the local
  directory, separated by semi-colons).  If the gaussian qualifier is
  not set, each walker has its parameters uniformly distributed from
  value-scale*(value-min) to value+scale*(max-value). (Default
  scale=1.)

  Otherwise, the walkers have initial parameters gaussianly
  distributed about value with sigma = scale*(max-value) or sigma =
  scale*(value-min) (i.e., normalized to the possibly asymmetric
  min/max values).  (Default scale=1/3.)

  Alternatively, the walkers can be initialized from the last
  iteration of an existing chain stored in a file, init_chain.

  The walkers are split into groups (0) and (1), with the update
  proposition (i.e., parameter vector) for the group (0) j^th walker
  being:
            X(0)'_j = X(1)_k + z*(X(0)_j-X(1)_k)
  X(1)_k is randomly drawn from group (1) and z is randomly
  distributed as 1/sqrt(z) between values of 1/cdf_scl_lo and
  cdf_scl_hi (default for both=2). Group (1) is then subsequently
  updated in a similar manner.

  The output chain is placed in the file specified by the output
  qualifier (default name is mcmc_results_processid.fits).
  Intermediate results are written to this file every sim_count
  (default=50) iterations.  If the chains were started from a previous
  chain file, the final output file will contain the initial chains as
  well unless the nocopy qualifier has been set.

  If num_slaves is specified, the process will be parallelized with a
  niceness value of nice for each process.  The serial qualifier will
  always override this, and force the default behavior of serial
  evaluation.

  The optional qualifiers grnd and urnd can be set to user defined
  functions to, e.g., allow random numbers drawn from a server or file
  in case of a manual multi-job run (among several machines).

  PRIORS
  One can apply gaussian priors to particular parameters by
  constructing a structure with the following fields
  priors.parlist -- an array of parameter ids
  priors.priorlist -- an array of structures describing each prior. A
     prior structure needs two fields: 'mean' and 'sigma'

  For example:
  myprior = struct{mean=0.1, sigma=0.2}
  emcee(nw, nsim; priors=struct{parlist=[1], priorlist=[myprior]})
`); %}}}
      return;
   }
   % Store initial parameters so that we can restore them at the end
   variable pars = get_params;

   variable num_slaves = int(qualifier("num_slaves",1));
   if(qualifier_exists("serial")) num_slaves=1;

   variable isc=0, sim_count_array, sim_count = int(qualifier("sim_count",50));
   if(sim_count <= 0) sim_count=50;

   sim_count_array=[sim_count-1:nsim-1+sim_count:sim_count];
   variable ochng;

   variable screwup    = qualifier("infile","NONE");
            screwup    = qualifier("in_file",screwup);
            screwup    = qualifier("start",screwup);
            screwup    = qualifier("start_file",screwup);
            screwup    = qualifier("initfile",screwup);
            screwup    = qualifier("init_file",screwup);
            screwup    = qualifier("input",screwup);
            screwup    = qualifier("initchain",screwup);
   variable init_chain = qualifier("init_chain",screwup);

   if ( screwup != "NONE" ) 
      ()=printf("\n *** Using %s for chain initialization  *** \n\n",init_chain);

            screwup    = qualifier("outfile","mcmc_results_");
            screwup    = qualifier("out_file",screwup);
            screwup    = qualifier("stop",screwup);
            screwup    = qualifier("stop_file",screwup);
            screwup    = qualifier("outchain",screwup);
            screwup    = qualifier("out_chain",screwup);
   variable outfile    = qualifier("output",screwup);

   if ( outfile == "mcmc_results_" ) outfile += "_"+string(getpid)+".fits";
   ()=printf("\n *** Using %s for output file *** \n\n",outfile);

   variable pfile      = qualifier("pfile","NONE");
   variable cdf_scl_lo = qualifier("cdf_scl_lo",2.);
   variable cdf_scl_hi = qualifier("cdf_scl_hi",2.);

   nw   = int(nw);
   nsim = int(nsim);

   variable xwalk, xstat, xwalk_a, xwalk_b, freepar = free_par();

   if(init_chain=="NONE")
   {
      (xwalk, xstat) = init_walker(freepar,nw;;__qualifiers);
   }
   else
   {
      variable nwc=0, nfpc=0, fpc;
      try
      {
       (nwc, nfpc, fpc, cdf_scl_lo, cdf_scl_hi, pfile,
        xwalk, xstat, ochng)=read_chain(init_chain;init_walker);

       if(qualifier_exists("nocopy"))
       {
          variable wstat=@xstat, wchng=@ochng;
          reshape(wstat,[1,length(wstat)]);
          reshape(wchng,[1,length(wchng)]);
          write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
                          xwalk, wstat, wchng, pfile;create);
       }
       else
       {
          () = system("cp "+init_chain+" "+outfile);
       }
      }
      catch AnyError:
      {
        ()=printf("\n Failed to read or create chain file \n");
      }
      % check, if present fit corresponds to the one, one wants to append to (init_chain)
      if(nwc!=nw || nfpc!=length(freepar) || max(fpc-freepar)!=0)
      {
         ()=printf("\n Chain file inconsistent, initializing from data \n");
         (xwalk, xstat) = init_walker(freepar,nw;;__qualifiers);
         init_chain="NONE";
         pfile="NONE";
      }
   }

   if(1/cdf_scl_lo >= cdf_scl_hi || cdf_scl_lo <= 0 || cdf_scl_hi <= 0)
   {
      ()=printf("\n1/CDF_SCL_LO must be < CDF_SCL_HI and > 0.\n");
      return;
   }

   variable iwlk_a, iwlk_b, la, lb, im, i, j, k, slaves, s, 
            nwf       = nw*num_free_params,
            xchng     = @xstat,
            xout      = Double_Type[num_free_params,sim_count*nwf],
            xout_stat = Double_Type[sim_count,nwf],
            xout_chng = Double_Type[sim_count,nwf]; 

   iwlk_a = [0:nwf-1:2];
   iwlk_b = [1:nwf-1:2];

   la = length(iwlk_a);
   lb = length(iwlk_b);

   Task_List.free_par  = freepar;
   Task_List.output    = Array_Type[max([la,lb])];
   Task_List.wstats    = Double_Type[max([la,lb])];
   Task_List.change    = Long_Type[max([la,lb])];

   variable array_xwalk      = Array_Type[max([la,lb])];
   variable array_xwalk_stat = Double_Type[max([la,lb])];

   % Seed all the required random numbers at the start, so we don't
   % have issues with stuff when it comes time to parallelize
   variable rands_a = Array_Type[nsim*la];
   variable rands_b = Array_Type[nsim*lb];

   _for i (0,nsim*la-1,1) { rands_a[i] = (@(qualifier("urnd", &urand)))(3); }
   _for i (0,nsim*lb-1,1) { rands_b[i] = (@(qualifier("urnd", &urand)))(3); }

   variable irnd;

   _for i (0,nsim-1,1)
   {
      % Update first half of walkers ...
      irnd = i*la+[0:la-1];

      _for j (0,la-1,1)
      {
         array_xwalk[j]      = xwalk[*,iwlk_a[j]];
         array_xwalk_stat[j] = xstat[iwlk_a[j]];

         if(num_slaves==1)
         {
            if(i==0 && j==0) () = printf("\n Serial Run \n");
            (Task_List.output[j],Task_List.wstats[j],
             Task_List.change[j]) = update_walker(cdf_scl_lo,cdf_scl_hi,
                                                  rands_a[irnd[j]], 
                                                  freepar,
                                                  array_xwalk[j],
                                                  array_xwalk_stat[j],
                                                  xwalk[*,iwlk_b] );
         }
      }

      if(num_slaves>1)
      {
         if(i==0) () = printf("\n Parallel Run \n");

         Task_List.index     = 0;
         Task_List.length    = la;
         Task_List.rands     = rands_a[irnd];
         Task_List.walkers   = array_xwalk;
         Task_List.wstats    = array_xwalk_stat;
         Task_List.proposals = xwalk[*,iwlk_b];

         slaves = new_slave_list();

         _for k (0, num_slaves-1, 1)
         {
            s = fork_slave(&slave_task, k, cdf_scl_lo, cdf_scl_hi 
                                                  ;; __qualifiers );
            s.status = SLAVE_READY;
            append_slave(slaves, s);
         }
         
         manage_slaves(slaves, &slave_handler);
      }

      _for j (0,la-1,1)
      {
         xwalk[*,iwlk_a[j]] = Task_List.output[j];
         xstat[iwlk_a[j]]   = Task_List.wstats[j];
         xchng[iwlk_a[j]]   = Task_List.change[j];
      }

      % Update second half of walkers ...

      irnd = i*lb+[0:lb-1];

      _for j (0,lb-1,1)
      {
         array_xwalk[j]      = xwalk[*,iwlk_b[j]];
         array_xwalk_stat[j] = xstat[iwlk_b[j]];

         if(num_slaves==1)
         {
            (Task_List.output[j],Task_List.wstats[j],
             Task_List.change[j]) = update_walker(cdf_scl_lo,cdf_scl_hi,
                                                  rands_b[irnd[j]], 
                                                  freepar,
                                                  array_xwalk[j],
                                                  array_xwalk_stat[j],
                                                  xwalk[*,iwlk_a]);
         }
      }

      if(num_slaves > 1)
      {
         Task_List.index     = 0;
         Task_List.length    = lb;
         Task_List.rands     = rands_b[irnd];
         Task_List.walkers   = array_xwalk;
         Task_List.wstats    = array_xwalk_stat;
         Task_List.proposals = xwalk[*,iwlk_a];

         slaves = new_slave_list();

         _for k (0, num_slaves-1, 1)
         {
            s = fork_slave(&slave_task, k, cdf_scl_lo, cdf_scl_hi
                                                  ;; __qualifiers );
            s.status = SLAVE_READY;
            append_slave(slaves, s);
         }
         
         manage_slaves(slaves, &slave_handler);
      }

      _for j (0,lb-1,1)
      {
         xwalk[*,iwlk_b[j]] = Task_List.output[j];
         xstat[iwlk_b[j]]   = Task_List.wstats[j];
         xchng[iwlk_b[j]]   = Task_List.change[j];
      }

      im = i mod sim_count;

      xout[*,im*nwf+[0:nwf-1:2]] = xwalk[*,iwlk_a];
      xout[*,im*nwf+[1:nwf-1:2]] = xwalk[*,iwlk_b];

      xout_stat[im,[0:nwf-1:2]] = xstat[iwlk_a];
      xout_stat[im,[1:nwf-1:2]] = xstat[iwlk_b];

      xout_chng[im,[0:nwf-1:2]] = xchng[iwlk_a];
      xout_chng[im,[1:nwf-1:2]] = xchng[iwlk_b];

%      if(i==0){ () = printf("\n Sim loop: \n"); }
%      if( i==sim_count_array[isc] || i==nsim-1 )
%      {
%         if(isc==0 && init_chain=="NONE")
%         {
%            write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
%                             xout, xout_stat, xout_chng, pfile; create);
%         }
%         else 
%         {
%            if (i==nsim-1)
%            {
%               variable icut, 
%                        xout_tmp = Double_Type[num_free_params,(im+1)*nwf],
%                        xout_stat_tmp = Double_Type[im+1,nwf],
%                        xout_chng_tmp = Double_Type[im+1,nwf];
%               _for icut (0,num_free_params-1,1)
%               {
%                  xout_tmp[icut,*]=xout[icut,[0:(im+1)*nwf-1]];
%               }
%               _for icut (0,im,1)
%               {
%                  xout_stat_tmp[icut,*]=xout_stat[icut,*];
%                  xout_chng_tmp[icut,*]=xout_chng[icut,*];
%               }
%               xout=@xout_tmp;
%               xout_stat=@xout_stat_tmp;
%               xout_chng=@xout_chng_tmp;
%            }
%            write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
%                             xout, xout_stat, xout_chng, pfile);
%         }
%      
%         () = printf("%d ... \n",i+1);
%         isc++;
%      }

         if(i==0){ () = printf("\n Sim loop: \n"); }
         if( i==sim_count_array[isc] || i==nsim-1 )
         {
            if(isc==0)
            {
               if ((init_chain=="NONE") or (qualifier_exists("nocopy")))
               {
                  write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
                              xout, xout_stat, xout_chng, pfile; create);
               }
               else
               {
                  write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
                              xout, xout_stat, xout_chng, pfile);
               }
            }
            else
            {
               write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
                                xout, xout_stat, xout_chng, pfile);
            }

            () = printf("%d ... \n",i+1);
            isc++;
         }

  }   

   if(sim_count > 0) () = printf("\n");

   set_params(pars);
}
%}}}

#ifexists merge_struct_arrays
public define emcee_merge() %{{{
%!%+
%\function{emcee_merge}
%\synopsis{Merge emcee fits files}
%\usage{emcee_merge( String_Type[] InFiles, String_Type OutFile)}
%\altusage{emcee_merge( String_Type[] InFiles )}
%\qualifiers{
%\qualifier{chatty}{If given, enable informative output.}
%\qualifier{force}{If given, overwrite existing OutFile.}
%\qualifier{adjuststeps}{If given, do not disregard InFiles with different #simsteps.
%                        Instead cut all InFiles to min(#simsteps).}
%\qualifier{ncut}{[=0] #sim. steps to be cut from the start of the chains.}
%\qualifier{autocut}{If given, 'ncut' is automatically set, s.t. the faulty sim. steps
%                      containing ZERO entries are deleted. These entries corresponds
%                      to the initialisation of the walkers.}
%\qualifier{id[=0]}{Index of the InFiles to check consistency to.}
%}
%\description
%   Merge several FITS Infiles created with emcee into one OutFile. It is required that
%   all InFiles are based on the same model and data. Only InFiles are merged that
%   have the same 'model' key and same 'free_par' indices compared to the first file
%   in InFiles.
%
%   IMPORTANT is that a common random number server was used for all InFiles to
%   ensure their statistical independency! This is not checked automatically!
%
%   InFiles (passed the check) are merged in a sense that the different walkers
%   from all InFiles are incorporated into a common chain, e.g., if InFiles are
%   two files with 5 free Parameters, 11 walkers and 1000 nsteps, respectively,
%   then the OutFile has 5 free Parameters, 22 walkers and 1000 nsteps.
%
%   To be able to merge this way NSTEPS in all InFiles has to be the same. InFiles
%   with different NSTEPS will be disregarded. With the 'adjuststeps' qualifier
%   it is possible to cut all InFiles to the mininmal occuring NSTEPS. NOTE that
%   this means to throw away simulation iterations in all InFiles.
%
%   With 'ncut' the number of simulation iterations at the beginning of the
%   the chains which should be cut away are set.
%
%   An existing OutFile can be overwritten with the 'force' qualifier.
%
%   If no OutFile is given, 'string_intersection' is used to determine the
%   OutFile path/name!
%
%\seealso{emcee}
%!%-
{
  variable chatty = qualifier_exists("chatty");
  variable force  = qualifier_exists("force");
  variable adjuststeps = qualifier_exists("adjuststeps");
  variable autocut = qualifier_exists("autocut");
  variable ncut = qualifier("ncut",0);
  variable id = qualifier("id",0);
  
  variable helptext = %{{{
    `emcee_merge( String_Type[] InFiles, String_Type OutFile
                 ; chatty, force, adjuststeps, ncut=0, autocut, id=0
                );

     Merge several FITS Infiles created with emcee into one OutFile. It is required that
     all InFiles are based on the same model and data. Only InFiles are merged that
     have the same 'model' key and same 'free_par' indices compared to the first file
     in InFiles.

     IMPORTANT is that a common random number server was used for all InFiles to
     ensure there statistical independency! This is not checked automatically!

     InFiles (passed the check) are merged in a sense that the different walkers
     from all InFiles are incorporated into a common chain, e.g., if InFiles are
     two files with 5 free Parameters, 11 walkers and 1000 nsteps, respectively,
     then the OutFile has 5 free Parameters, 22 walkers and 1000 nsteps.

     To be able to merge this way NSTEPS in all InFiles has to be the same. InFiles
     with different NSTEPS will be disregarded. With the 'adjuststeps' qualifier
     it is possible to cut all InFiles to the mininmal occuring NSTEPS. NOTE that
     this means to throw away simulation iterations in all InFiles.

     With 'ncut' the number of simulation iterations at the beginning of the
     the chains which should be cut away are set.
     If 'autocut' is given, 'ncut' is automatically set, s.t. the faulty sim. steps
     containing ZERO entries are deleted. These entries corresponds to the
     initialisation of the walkers.

     'id' determines the index of the InFiles to check consistency to.

     An existing OutFile can be overwritten with the 'force' qualifier.`;
%}}}
 
  variable files, outfile = NULL;
  switch(_NARGS)
  { case 1 : files = ();
    outfile = __get_reference("string_intersection");
    if( _isnull(outfile) ){
      outfile = path_dirname(files[id]) + "/emcee_merge.fits";
    }
    else{
      outfile = [(@outfile)(files;longest)][0] + ".fits";
    }
  }
  { case 2 : (files, outfile) = (); }
  { vmessage(helptext); return; }

  variable N = length(files);
  
  %%%
  %%% Initialize Outfile
  %%%
  if( access( outfile, F_OK )==0 and force==0){
    vmessage("ERROR(%s): 'outfile'=%s already exists, use 'force' qualifier to overwrite!",
	     _function_name, outfile );
    return;
  }
  else if(chatty){ vmessage("INFO(%s): Using 'outfile': %s",_function_name,outfile); }

    
  %%%
  %%% CHECK FILES FOR CONSISTENCY & GET TABLE INFO
  %%%
  variable free_par = fits_read_col( files[id]+"[PARAMETERS]", "FREE_PAR" ); 
  variable model    = fits_read_key( files[id]+"[PARAMETERS]", "MODEL" );

  variable nf = length(free_par);
  variable NW = Integer_Type[N];
  variable NSTEPS = Integer_Type[N];
  NW[id] =  fits_read_key( files[id]+"[MCMCCHAIN]", "NWALKERS" );
  NSTEPS[id] =  fits_read_key( files[id]+"[MCMCCHAIN]", "NSTEPS" );
  
  variable IND = Integer_Type[0];
  variable ii;
  _for ii ( 0, N-1 ){
    %if(chatty){ vmessage("INFO(%s): CHECKING '%s' for consistency!",_function_name,files[ii]); }
    %%% CHECK [PARAMETERS] extension for sameness
    ifnot( model == fits_read_key( files[ii]+"[PARAMETERS]", "MODEL") ){
      vmessage("ERROR(%s): 'MODEL' in '%s' does not match! Skipping ...",
	       _function_name, files[ii] );
      continue;
    }
    else ifnot( _eqs(free_par, fits_read_col( files[ii]+"[PARAMETERS]", "FREE_PAR")) ){
      vmessage("ERROR(%s): 'FREE_PAR' in '%s' do not match! Skipping ...",
	       _function_name, files[ii] );
      continue;
    }

    NW[ii]     = fits_read_key( files[ii]+"[MCMCCHAIN]", "NWALKERS" );
    NSTEPS[ii] = fits_read_key( files[ii]+"[MCMCCHAIN]", "NSTEPS" );
    
    if( NSTEPS[ii] == NSTEPS[id] or adjuststeps==1 ){
      IND = [ IND, ii ];
    }
    else{
      vmessage("ERROR(%s): 'NSTEPS' in '%s' does not match! Use 'adjuststeps' qualifier! Skipping ...",
	       _function_name, files[ii] );
    }
  }

  %%% AUTOCUT
  if( autocut ){
    variable fu = fits_read_col( files[id]+"[CHAINSTATS]", "FRAC_UPDATE" );
    ncut = wherelast_le( fu, 0 ) + 2;
    if( chatty ){vmessage("INFO(%s): Setting 'ncut'=%d",_function_name,ncut);};
  }
  
  %%%
  %%% INITIALIZE OUTPUTFILE
  %%%
  if(chatty){
    vmessage("INFO(%s): Initializing 'outfile': %s",_function_name, outfile );
  }
  
  variable MNW = nint( sum(NW[IND]) );
  variable MNSTEPS = nint(min(NSTEPS[IND])-ncut);
  variable clen = MNW * MNSTEPS * nf ;

  %%% create output file: create extensions and define their structure
  variable extname, colname, ttype, tunit;

  variable ofp = fits_open_file( outfile, "c" );

  extname = "PARAMETERS";
  colname = ["FREE_PAR"];
  ttype   = ["J"];
  tunit   = [" parameter indices"];  
  fits_create_binary_table( ofp, extname, nf, colname, ttype, tunit);
  fits_check_error( _fits_write_col( ofp, fits_get_colnum(ofp,"FREE_PAR"),1,1,free_par ) );
  
  extname = "MCMCCHAIN";
  colname = ["FITSTAT","UPDATE", "CHAINS"+array_map(String_Type,&sprintf,"%d",free_par) ];
  ttype   = ["D","J", (@["D"])[Integer_Type[nf]]];
  tunit   =  [" fit statistics"," update indicator", (@[" parameter values"])[Integer_Type[nf]]];
  fits_create_binary_table( ofp, extname, clen, colname, ttype, tunit);

  extname = "CHAINSTATS";
  colname = ["FRAC_UPDATE","MIN_STAT","MED_STAT","MAX_STAT"];
  ttype   = ["D","D","D","D"];
  tunit   = [" fraction"," chi2"," chi2"," chi2"];
  fits_create_binary_table( ofp, extname, MNSTEPS, colname, ttype, tunit);

  fits_close_file(ofp);

  %%% copy the header key words from InFiles[0] to OutFile without the extension
  %%% dimension information!
  %%% NOTE that there are no functions in isis to do that!
  variable numhdu = fits_get_num_hdus(files[id]);
  _for ii ( 1 , numhdu-1 ){
    () = system ( sprintf("cphead %s+%d %s+%d", files[id], ii, outfile, ii) );
  }
  
  %%%
  %%% MERGE [MCMCCHAIN] EXTENSION
  %%%
  if(chatty){
    vmessage("INFO(%s): Merging %d/%d files:",_function_name, length(IND), N );
  }
  
  %%% initialize indexing
  %%% expand s.t., the outfile has sum(NW) walkers
  variable cindmtrx = _reshape( [0:MNW*nf*MNSTEPS-1], [MNSTEPS,MNW,nf] );
  variable CIND = Array_Type[N];
  variable _CIND = Array_Type[N];
  variable jj;
  _for jj ( 0, length(IND)-1 ){
    _CIND[IND[jj]] = [NW[IND[jj]]*nf*ncut:NW[IND[jj]]*nf*min(NSTEPS[IND[jj]])-1];
    CIND[IND[jj]] = cindmtrx[*, [nint(sum(NW[IND[[0:jj-1]]]))+[0:NW[IND[jj]]-1]], * ];
  }


  %%% Read and write
  ofp = fits_open_file( outfile, "w" );
  fits_check_error( _fits_movnam_hdu ( ofp, _FITS_BINARY_TBL, "MCMCCHAIN", 0 ) );
  fits_update_key( ofp, "NWALKERS", MNW );
  fits_update_key( ofp, "NSTEPS", MNSTEPS );

  %%% [MCMCCHAIN]FITSTAT
  if(chatty){
    vmessage("");
    ()=system(sprintf("echo -n \"INFO(%s): Merging '%s[MCMCCHAIN]/FITSTAT' ... FileID: \"",
		      _function_name,outfile)
	     );
  }
  variable col = Double_Type[clen];
  foreach jj ( IND ){
    if(chatty){ ()=system(sprintf("echo -n %d, ",jj+1)); }
    col[CIND[jj]] = fits_read_col( files[jj]+"[MCMCCHAIN]", "FITSTAT" )[_CIND[jj]];
  }
  fits_check_error( _fits_write_col( ofp, fits_get_colnum(ofp,"FITSTAT"),1,1,col ) );

  %%% [CHAINSTATS] MIN/MED/MAX_STAT
  reshape(col,[MNSTEPS,MNW,nf]);
  variable cstatmin = Double_Type[MNSTEPS];
  variable cstatmed = Double_Type[MNSTEPS];
  variable cstatmax = Double_Type[MNSTEPS];
  _for ii ( 0, MNSTEPS-1 ){
    cstatmin[ii] = min(col[ii,*,*]);
    cstatmax[ii] = max(col[ii,*,*]);
    cstatmed[ii] = median(col[ii,*,*]);
  }
  fits_check_error( _fits_movnam_hdu ( ofp, _FITS_BINARY_TBL, "CHAINSTATS", 0 ) );
  fits_check_error( _fits_write_col( ofp, fits_get_colnum(ofp,"MIN_STAT"),1,1,cstatmin ) );
  fits_check_error( _fits_write_col( ofp, fits_get_colnum(ofp,"MED_STAT"),1,1,cstatmed ) );
  fits_check_error( _fits_write_col( ofp, fits_get_colnum(ofp,"MAX_STAT"),1,1,cstatmax ) );  
  
  %%% [MCMCCHAIN]UPDATE
  fits_check_error( _fits_movnam_hdu ( ofp, _FITS_BINARY_TBL, "MCMCCHAIN", 0 ) );
  if(chatty){
    vmessage("");
    ()=system(sprintf("echo -n \"INFO(%s): Merging '%s[MCMCCHAIN]/UPDATE' ... FileID: \"",
		      _function_name,outfile)
	     );
  }
  col = Double_Type[clen];
  foreach jj ( IND ){
    if(chatty){ ()=system(sprintf("echo -n %d, ",jj+1)); }
    col[CIND[jj]] = fits_read_col( files[jj]+"[MCMCCHAIN]", "UPDATE" )[_CIND[jj]];
  }
  fits_check_error( _fits_write_col( ofp, fits_get_colnum(ofp,"UPDATE"),1,1,col ) );
  
  %%% [CHAINSTATS]FRAC_UPDATE
  fits_check_error( _fits_movnam_hdu ( ofp, _FITS_BINARY_TBL, "CHAINSTATS", 0 ) );
  fu = sum(sum(_reshape(col,[MNSTEPS,MNW,nf]),2),1)/MNW/nf;
  fits_check_error( _fits_write_col( ofp, fits_get_colnum(ofp,"FRAC_UPDATE"),1,1,fu ) );

  
  
  %%% [MCMCCHAIN]CHAINS
  fits_check_error( _fits_movnam_hdu ( ofp, _FITS_BINARY_TBL, "MCMCCHAIN", 0 ) );
  foreach ii ( free_par ){
    if(chatty){
      vmessage("");
      ()=system(sprintf("echo -n \"INFO(%s): Merging '%s[MCMCCHAIN]/CHAINS%d' ... FileID: \"",
			_function_name,outfile,ii)
	       );
    }
    col = Double_Type[clen];
    foreach jj ( IND ){
      if(chatty){ ()=system(sprintf("echo -n %d, ",jj+1)); }
      col[CIND[jj]] = fits_read_col( files[jj]+"[MCMCCHAIN]", "CHAINS"+string(ii) )[_CIND[jj]];
    }
    fits_check_error(_fits_write_col(ofp, fits_get_colnum(ofp,"CHAINS"+string(ii)),1,1,col) );
  }
  fits_close_file(ofp);
  if(chatty){ vmessage(""); }
  
  return;
}
%}}}
public define emcee_chain_hist_collect() %{{{
%!%+
%\function{emcee_chain_hist_collect}
%\synopsis{Collect chain histograms from multiple emcee files}
%\usage{Struct_Type[] chist = emcee_chain_hist_collect( String_Type[] InFiles, Interger_Type[] PID )}
%\altusage{Struct_Type[] chist = emcee_chain_hist_collect( String_Type[] InFiles )}
%\qualifiers{
%\qualifier{chatty}{If given, enable informative output.}
%\qualifier{ncut}{[=0] #sim. steps to be cut from the start of the chains.}
%\qualifier{autocut}{If given, 'ncut' is automatically set, s.t. the faulty sim. steps
%                      containing ZERO entries are deleted. These entries corresponds
%                      to the initialisation of the walkers.}
%\qualifier{id[=0]}{Index of the InFiles to check consistency to.}
%\qualifier{nbin[=100]}{Number of histogram bins.}
%\qualifier{pmin}{Double_Type[length(PID)]: Containing parmater lower limits.}
%\qualifier{pmax}{Double_Type[length(PID)]: Containing parmater upper limits.}
%}
%\description
%   Generate parameter histograms by collecting chains from several FITS Infiles
%   created with emcee. It is required that all InFiles are based on the same model
%   and data. Only InFiles are having the same 'model' key and same 'free_par'
%   indices are taken into account.
%
%   IMPORTANT is that a common random number server was used for all InFiles to
%   ensure there statistical independency! This is not checked automatically!
%
%   Chains from InFiles (passed the check) for each parameter in 'pid' are collected,
%   that is added to a common histogram with 'nbin' bins. With 'pmin' and 'pmax'
%   the limits of the histograms can be set manually, where 'pmin'/'pmax' must
%   be an array of the same length as 'pid'.
%
%   'id' sets the index in InFiles, which is used to perform the consistency check.
%
%   With 'ncut' the number of simulation iterations at the beginning of the
%   the chains which should be cut away are set.
%   'autocut' can be used to set 'ncut' automatically, s.t. all iteration
%   steps related to the walker initialization are cut away.
%
%\seealso{emcee}
%!%-
{
  variable chatty = qualifier_exists("chatty");
  variable autocut = qualifier_exists("autocut");
  variable ncut = qualifier("ncut",0);
  variable id = qualifier("id",0);

  variable nbin = qualifier("nbin",100);
  variable pmin = qualifier("pmin",NULL);
  variable pmax = qualifier("pmax",NULL);
  
  variable helptext = %{{{
    `Struct_Type[] chist = emcee_chain_hist_collect( String_Type[] InFiles, Interger_Type[] PID )
     Struct_Type[] chist = emcee_chain_hist_collect( String_Type[] InFiles )

   QUALIFIERS:
     chatty:   If given, enable informative output.
     ncut[=0]: #sim. steps to be cut from the start of the chains.
     autocut:  If given, 'ncut' is automatically set, s.t. the faulty sim. steps
               containing ZERO entries are deleted. These entries corresponds
               to the initialisation of the walkers.
     id[=0]:   Index of the InFiles to check consistency to.}
     nbin[=100]: Number of histogram bins.
     pmin:     Double_Type[length(PID)]: Containing parmater lower limits.
     pmax:     Double_Type[length(PID)]: Containing parmater upper limits.


   DESCRIPTION:
     Generate parameter histograms by collecting chains from several FITS Infiles
     created with emcee. It is required that all InFiles are based on the same model
     and data. Only InFiles are having the same 'model' key and same 'free_par'
     indices are taken into account.

     IMPORTANT is that a common random number server was used for all InFiles to
     ensure there statistical independency! This is not checked automatically!

     Chains from InFiles (passed the check) for each parameter in 'pid' are collected,
     that is added to a common histogram with 'nbin' bins. With 'pmin' and 'pmax'
     the limits of the histograms can be set manually, where 'pmin'/'pmax' must
     be an array of the same length as 'pid'.

     'id' sets the index in InFiles, which is used to perform the consistency check.

     With 'ncut' the number of simulation iterations at the beginning of the
     the chains which should be cut away are set.
     'autocut' can be used to set 'ncut' automatically, s.t. all iteration
     steps related to the walker initialization are cut away.
`;
%}}}
 
  variable files, pid = NULL;
  switch(_NARGS)
  { case 1 : files = (); }
  { case 2 : (files, pid) = (); }
  { vmessage(helptext); return; }

  variable N = length(files);
  
  %%%
  %%% CHECK FILES FOR CONSISTENCY & GET TABLE INFO
  %%%
  variable free_par = fits_read_col( files[id]+"[PARAMETERS]", "FREE_PAR" );
  if( _isnull(pid) ){
    pid = free_par;
    if( chatty ){vmessage("INFO(%s): Setting 'pid[%d]'=[%s]", _function_name,length(pid),
			  strjoin(array_map(String_Type, &string, pid),",")
			 );
    }
  }
  variable NP = length(pid);

  
  variable model    = fits_read_key( files[id]+"[PARAMETERS]", "MODEL" );

  variable nf = length(free_par);
  variable NW = Integer_Type[N];
  variable NSTEPS = Integer_Type[N];
  variable IND = Integer_Type[0];

  variable ii;
  _for ii ( 0, N-1 ){
    %if(chatty){ vmessage("INFO(%s): CHECKING '%s' for consistency!",_function_name,files[ii]); }
    %%% CHECK [PARAMETERS] extension for sameness
    ifnot( model == fits_read_key( files[ii]+"[PARAMETERS]", "MODEL") ){
      vmessage("ERROR(%s): 'MODEL' in '%s' does not match! Skipping ...",
	       _function_name, files[ii] );
      continue;
    }
    else ifnot( _eqs(free_par, fits_read_col( files[ii]+"[PARAMETERS]", "FREE_PAR")) ){
      vmessage("ERROR(%s): 'FREE_PAR' in '%s' do not match! Skipping ...",
	       _function_name, files[ii] );
      continue;
    }
    else{
      IND = [ IND, ii ];
    }
    
    ( NW[ii], NSTEPS[ii] ) = fits_read_key( files[ii]+"[MCMCCHAIN]", "NWALKERS", "NSTEPS" );
  }

  %%% AUTOCUT
  if( autocut ){
    variable fu = fits_read_col( files[id]+"[CHAINSTATS]", "FRAC_UPDATE" );
    ncut = wherelast_le( fu, 0 ) + 2;
    if( chatty ){vmessage("INFO(%s): Setting 'ncut'=%d",_function_name,ncut);};
  }

  %%% PMIN, PMAX
  variable jj, chain, cind;
  if( _isnull(pmin) or _isnull(pmax) ){
    try{
      pmin = merge_struct_arrays( get_par_info( pid ) );
      pmax = pmin.max;
      pmin = pmin.min;
      if( chatty ){vmessage("INFO(%s): Setting Pmin/Pmax according to active Parameters!",_function_name);};
    }
    catch AnyError:{
      if( chatty ){vmessage("INFO(%s): Determine Pmin/Pmax from InFile[id=%d]='%s' !",_function_name,id,files[id]);};
      pmin = Double_Type[NP];
      pmax = Double_Type[NP];
      _for ii ( 0, NP-1 ){
	cind = [ NW[id] * nf * ncut : ];
	chain = fits_read_col( files[id]+"[MCMCCHAIN]", sprintf("CHAINS%d",pid[ii]) )[cind];
	pmin[ii] = min(chain);
	pmax[ii] = max(chain);
      }
    }
  }
  
  
  %%%
  %%% Collect Chain histograms for parameters 'pid'
  %%%
  variable chist = Struct_Type[NP];
  _for ii ( 0, NP-1 ){
    if( chatty ){() = system(sprintf("echo -n \"INFO(%s): Collecting chain hist for 'pid'=%d (%3d/%3d): InFile ID:\"",
				     _function_name,pid[ii], ii+1,NP)
			    );
    };
    chist[ii] = struct{
      bin_lo, bin_hi,
      value = Integer_Type[nbin],
      pid   = pid[ii]
    };
    ( chist[ii].bin_lo, chist[ii].bin_hi ) = linear_grid( pmin[ii], pmax[ii], nbin );
    _for jj ( 0, length(IND)-1 ){
      if(chatty){ ()=system(sprintf("echo -n %d, ",jj+1)); }
      cind = [ NW[IND[jj]] * nf * ncut : ];
      chain = fits_read_col( files[IND[jj]]+"[MCMCCHAIN]", sprintf("CHAINS%d",pid[ii]) )[cind];

      chist[ii].value += histogram( chain, chist[ii].bin_lo, chist[ii].bin_hi );
    }
    if(chatty){vmessage("");}
  }
  
  
  return chist;
}
%}}}
#endif

%%%%%%%%%%%%%%%%%%%%%%%

private define unity(x) %{{{
{
   return x;
}
%}}}

public define chain_hist() %{{{
{
   if(_NARGS==2)
   {
      variable i, y;
      (i, y) = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  chain_hist(i, chain; fcut=#, lcut=#, nbin=#, color=#, linestyle=#,
                       pmin=#, pmax=#, 
                       xrange={# or NULL,# or NULL}, 
                       yrange={# or NULL,# or NULL},
                       transform=&func, 
                       function=&func, func_vars=[Array],
                       clevel=#, print, return_map, return_median,
                       xlabel=String, ylabel=String, oplot); 

  Plot a normalized histogram of the MCMC chain for the i^th free
  parameter, between values pmin and pmax (defaults to the min/max
  values in the chain; normalization, however, is always for the full
  range of the chain subsample). The plot axes have the ranges given
  in the lists, xrange and yrange (NULL==autoscale). Throw out the
  first fcut fraction of the data (default=0.2) and the last lcut
  fraction of the data (default=0), and use nbin bins (default=50).

  The x-axis variable can be transformed using any monotonic function
  passed as a reference via the transform qualifier (e.g.,
  transform=&log10).  pmin/pmax will then be in reference to this
  transform variable.

  Alternatively, if a multivariate function is passed via the function
  qualifier, with func_vars being an array indicating which of the i
  free parameters serve as variables.  These choices supercede both
  the parameter index input, i, and any transform input.  The function
  should expect input of the form x[i,k], where the i index is for the
  free parameters of the fit, and the k index is a specific instance
  from the chain.

  Plot using pgplot color given by qualifier (default=1, black) and
  linestyle given by qualifier (default=1, solid line).  The plot will
  be an overplot if the oplot qualifier is set. The plot function
  presumes that the appropriate fit model has been loaded.  Custom X/Y
  labels will be applied if the x/ylabel qualifiers have been set.
  The print qualifier has the lower and upper clevel (default=0.9)
  confidence levels printed out. The qualifier return_map, returns the
  MAP (maximum a posteriori) value, i.e., the peak of the
  distribution, return_median the median of the posterior
  distribution.

`); %}}}
      return;
   }
   variable xr = qualifier("xrange",{NULL,NULL});
   variable yr = qualifier("yrange",{NULL,NULL});
   xrange(xr[0],xr[1]); yrange(yr[0],yr[1]);
   connect_points(-1);
   variable fp = free_par;
   variable xlab = qualifier("xlabel",NULL);
   variable ylab = qualifier("ylabel",NULL);
   variable clevel = qualifier("clevel",0.9);
   variable transf = qualifier("transform",&unity);
   if(clevel >= 1) clevel=0.9;
   if(xlab==NULL)
   {
      xlabel("\\fr "+get_par_info(free_par[i]).name);
   }
   else
   {
      xlabel(xlab);
   }
   if(ylab==NULL)
   {
      ylabel("\\fr Probability");
   }
   else
   {
      ylabel(ylab);
   };
   variable lo,hi,lo_fine_wide,hi_fine_wide, real_xmin, real_xmax;
   variable nbin = qualifier("nbin",50);
   variable fcut = qualifier("fcut",0.2);
   variable lcut = qualifier("lcut",0);
   variable col = [qualifier("color",1)][0];
   variable lstyle = [qualifier("linestyle",1)][0];
   variable l = length(y[i]);
   variable ikeep=[int(fcut*l):l-1-int(lcut*l)];
   variable ntot = 1.*length(ikeep);
   variable ykeep;
   if( qualifier_exists("function") && qualifier_exists("func_vars") ) 
   {
      variable func = qualifier("function",&unity);
      variable func_vars = qualifier("func_vars",i);
      variable iv, nv=length(func_vars), ytemp = y[func_vars];
      _for iv (0,nv-1,1) { ytemp[iv] = ytemp[iv][ikeep]; }
      ykeep = @func(ytemp);
   }
   else
   {
      ykeep = @transf(y[i][ikeep]);
   }

   real_xmin = min(ykeep);
   real_xmax = max(ykeep);
   variable xmin = qualifier("pmin",real_xmin);
   variable xmax = qualifier("pmax",real_xmax);
   (lo,hi) = linear_grid(xmin,xmax,nbin);
   (lo_fine_wide,hi_fine_wide) = linear_grid(real_xmin,real_xmax,4*nbin);
   variable n = histogram(ykeep,lo,hi);
   variable n_fine_wide = histogram(ykeep,lo_fine_wide,hi_fine_wide);

   variable csum = cumsum(n_fine_wide)/ntot;

   % determine MAP or median values of posterior distribution
   if(qualifier_exists("return_map") || qualifier_exists("return_median")) {    
     variable grid_mean = (lo_fine_wide+hi_fine_wide)/2.;
     % P. Gregory, "Bayesian Logical Data Analysis for the Physical Sciences": median robust measure!
     variable p_map = grid_mean[where(n_fine_wide==max(n_fine_wide))][0];
     if (length(wherelast(csum <= 0.5))>=1 && wherelast(csum <= 0.5) != NULL) {
       variable p_median = grid_mean[wherelast(csum <= 0.5)][0];
     }
     else {
       p_median = p_map;
       message("Distribution unresolved with current binning. Adapt parameter ranges or change binning. Using MAP value.");
     }
     vmessage("MAP  %.2f   MEDIAN  %.2f",p_map,p_median);
   }
   ifnot(qualifier_exists("oplot") || qualifier_exists("return_map") || qualifier_exists("return_median"))
   { 
      linestyle(lstyle);
      hplot(lo,hi,n/(hi-lo)/ntot,col);
   }
   else ifnot(qualifier_exists("return_map") || qualifier_exists("return_median"))
   {
      linestyle(lstyle);
      ohplot(lo,hi,n/(hi-lo)/ntot,col);
   }
   % If posterior distribution is not clearly centrally peaked but shows secondary maxima,
   % the cummulative sum over the whole distribution may not be the best way.
   % 
   if(qualifier_exists("print"))
   {
      variable ifirst = wherelast(csum <= (1-clevel)/2.);
      variable ilast = wherefirst(csum >= (1+clevel)/2.);
      if(length(ifirst) > 0 && length(ilast) > 0)
      {
         () = printf("\n %.3e confidence levels are: %.4e %.4e \n\n",
                         clevel,hi_fine_wide[ifirst],lo_fine_wide[ilast]);
      }
      else
      {
         () = printf("\n Failed finding confidence levels.");
      }
   }
   if (qualifier_exists("return_map")) return p_map;
   else if (qualifier_exists("return_median")) return p_median;
}
%}}}


public define chain_confidence() %{{{
%Author: Marco Fink -- marco.fink@fau.de
{
   if(_NARGS==2)
   {
      variable i, chain;
      (i, chain) = ();
   }
   else
   {
        () = fprintf(fs,  "%s\n", %{{{
`
  chain_confidence(i, chain; fcut=#, lcut=#, nbin=#,
                       pmin=#, pmax=#, 
                       xrange={# or NULL,# or NULL}, 
                       yrange={# or NULL,# or NULL},
                       transform=&func,
                       inverse=&func,
                       plot=#,
                       clevel=#
                       ); 

  This is a fork of the chain_hist() function by Mike. The Functionality is very 
  similar, but all results are returned to the stack so they can be used for 
  further plotting and processing. This function Calculates posterior 
  probability distribution (pdf) and cumulative distribution function (cdf) 
  for the i-th parameter in the given MCMC-chain.
  
  It returns the histogram grids, pdf and cdf and a struct of parameters 
  specifying the pdf. The struct contains confidence levels (default: 90%), 
  median, mean and maximum values of the pdf.
  
  Usage example:
  (lo, hi, n, csum, stats) = chain_confidence(3,read_chain("mcmc_chain.fits"));
  
  print_struct(stats);
     conflo     median      mean       max    confhi
    57.5784   72.02990   69.8261   73.5922   74.7639

  If the transform and inverse qualifiers are given, the distributions and plots
  are calculated in the transformed parameter space.

  If the 'plot' qualifier is set to >= 1, a pg-plot visualization of the 
  pdf and cdf is shown.
  
  see also:
  chain_hist(), read_chain()

 `);   %}}} 
  return;
   }
  
   variable fp = free_par();
   variable clevel = qualifier("clevel",0.9);
   if(clevel >= 1) clevel=0.9;
  
   variable lo,hi,lo_fine_wide,hi_fine_wide, real_xmin, real_xmax;
   variable nbin = qualifier("nbin",100);
   variable fcut = qualifier("fcut",0.2);
   variable lcut = qualifier("lcut",0);
   variable trafo = qualifier("transform",&unity);
   variable invtrafo = qualifier("inverse",&unity);
   variable plotflag = qualifier("plot",0);

   variable l = length(chain[i]);
   variable ikeep=[int(fcut*l):l-1-int(lcut*l)];
   variable ntotal = 1.*length(ikeep);

   variable chkeep = (chain[i][ikeep]);

   variable xmin = qualifier("pmin",min(chkeep));
   variable xmax = qualifier("pmax",max(chkeep));

   chkeep = @trafo(chkeep); xmin = @trafo(xmin); xmax = @trafo(xmax);

   (lo,hi) = linear_grid(xmin,xmax,nbin);
   variable n = histogram(chkeep,lo,hi);
   variable csum = cumsum(n)/sum(n);
 
   variable ifirst = wherelast(csum <= (1-clevel)/2.);
   variable imed = wherelast(csum <= 0.5);
   variable imax = int(mean(where_max(n)));
   variable ilast = wherefirst(csum >= (1+clevel)/2.);

   if ( ifirst == NULL || imed == NULL ) {
      if ( not qualifier_exists("nonverbose") )
         vmessage("*** Warning: Maximum confidence at lowest bin");
      ifirst = 0;
      imed = 0;
      }
 
   variable stats = struct { 
       conflo = (lo[ifirst]+hi[ifirst])/2., 
       median = (lo[imed]+hi[imed])/2.,
       mean = weighted_mean((lo+hi)/2.,n),
       max = (lo[imax]+hi[imax])/2.,
       confhi = (lo[ilast]+hi[ilast])/2.};
   if (qualifier_exists("xlabel"))
   {
      xlabel(qualifier("xlabel"));
   }
   else
   {
      xlabel("Parameter Value");
   }
   if (qualifier_exists("ylabel"))
   {
      ylabel(qualifier("ylabel"));
   }
   else
   {
      ylabel("CDF");
   }

   if(plotflag > 0){
       linestyle(1); 
       hplot(lo,hi,1.0*n/max(n),4), ohplot(lo,hi,csum);
       linestyle(2);
       oplot([stats.mean,stats.mean],[0,1],4);
       oplot([stats.max,stats.max],[0,1],4);

       oplot([stats.median,stats.median],[0,csum[imed]],2);
       oplot([stats.conflo,stats.conflo],[0,csum[ifirst]],2);
       oplot([stats.confhi,stats.confhi],[0,csum[ilast]],2);
       linestyle(1); 
   }
 
   stats.conflo = @invtrafo(stats.conflo);
   stats.median = @invtrafo(stats.median);
   stats.mean = @invtrafo(stats.mean);
   stats.max = @invtrafo(stats.max);
   stats.confhi = @invtrafo(stats.confhi);
 
   return (@invtrafo(lo),@invtrafo(hi),n,csum,stats);
}
%}}}



public define chain_set_params() %{{{
{
  if(_NARGS==1)
  {
    variable y;
    (y) = ();
  }
  else
  {
    () = fprintf(fs,  "%s\n", %{{{
`
		 chain_set_params(chain; );

  Set parameters based on the posterior probability distributions gained
  in the emcee chain. This function loops over all free parameters and
  uses the function chain_hist to extract either the MAP or the median
  values if one sets the appropriate qualifiers that are passed to the
  chain_hist function. 		 
`		 
		); %}}}
    return;
  }
  ifnot(qualifier_exists("return_map") or qualifier_exists("return_median"))
  {
     ()=fprintf(fs, " Set either the return_map or return_median qualifier\n");
     return;
  }
  variable fp = free_par;
  variable i;
  _for i (0,length(fp)-1,1) {
    set_par(fp[i],chain_hist(i,y;;__qualifiers));
  }  

}
%}}}

public define vec_hist() %{{{
{
   if(_NARGS==1)
   {
      variable i, y;
      y = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  vec_hist(chain_vector; fcut=#, lcut=#, nbin=#, color=#, linestyle=#,
                       pmin=#, pmax=#, 
                       xrange={# or NULL,# or NULL}, 
                       yrange={# or NULL,# or NULL},
                       clevel=#, print, 
                       xlabel=String, ylabel=String, oplot); 

  Plot a normalized histogram for a vector from the MCMC chain,
  between values pmin and pmax (defaults to the min/max values in the
  chain; normalization, however, is always for the full range of the
  chain subsample). The plot axes have the ranges given in the lists,
  xrange and yrange (NULL==autoscale). Throw out the first fcut
  fraction of the data (default=0.2) and the last lcut fraction of the
  data (default=0), and use nbin bins (default=50).

  Plot using pgplot color given by qualifier (default=1, black) and
  linestyle given by qualifier (default=1, solid line).  The plot will
  be an overplot if the oplot qualifier is set.  The print qualifier
  has the lower and upper clevel (default=0.9) confidence levels
  printed out.
`); %}}}
      return;
   }
   variable xr = qualifier("xrange",{NULL,NULL});
   variable yr = qualifier("yrange",{NULL,NULL});
   xrange(xr[0],xr[1]); yrange(yr[0],yr[1]);
   connect_points(-1);
   variable fp = free_par;
   variable xlab = qualifier("xlabel","\\frX");
   variable ylab = qualifier("ylabel","\\frY");
   variable clevel = qualifier("clevel",0.9);
   if(clevel >= 1) clevel=0.9;

   xlabel(xlab);
   ylabel(ylab);

   variable lo,hi,lo_fine_wide,hi_fine_wide, real_xmin, real_xmax;
   variable nbin = qualifier("nbin",50);
   variable fcut = qualifier("fcut",0.2);
   variable lcut = qualifier("lcut",0);
   variable col = [qualifier("color",1)][0];
   variable lstyle = [qualifier("linestyle",1)][0];
   variable l = length(y);
   variable ikeep=[int(fcut*l):l-1-int(lcut*l)];
   variable ntot = 1.*length(ikeep);
   variable ykeep = y[ikeep];
   real_xmin = min(ykeep);
   real_xmax = max(ykeep);
   variable xmin = qualifier("pmin",real_xmin);
   variable xmax = qualifier("pmax",real_xmax);
   (lo,hi) = linear_grid(xmin,xmax,nbin);
   (lo_fine_wide,hi_fine_wide) = linear_grid(real_xmin,real_xmax,4*nbin);
   variable n = histogram(ykeep,lo,hi);
   variable n_fine_wide = histogram(ykeep,lo_fine_wide,hi_fine_wide);
   ifnot(qualifier_exists("oplot"))
   { 
      linestyle(lstyle);
      hplot(lo,hi,n/(hi-lo)/ntot,col);
   }
   else
   {
      linestyle(lstyle);
      ohplot(lo,hi,n/(hi-lo)/ntot,col);
   }
   if(qualifier_exists("print"))
   {
      variable csum = cumsum(n_fine_wide)/ntot;
      variable ifirst = wherefirst(csum >= (1-clevel)/2.);
      variable ilast = wherelast(csum >= (1+clevel)/2.);
      if(length(ifirst) > 0 && length(ilast) > 0)
      {
         () = printf("\n %.3e confidence levels are: %.4e %.4e \n\n",
                         clevel,hi_fine_wide[ifirst],lo_fine_wide[ilast]);
      }
      else
      {
         () = printf("\n Failed finding confidence levels.");
      }
   }
}
%}}}

public define chain_plot() %{{{
{
   if(_NARGS==4)
   {
      variable i, nc, nw, ans;
      (i, nc, nw, ans) = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  chain_plot(i, nc, nw, chain; xrange={# or NULL,# or NULL}, 
                               yrange={# or NULL,# or NULL},
                               xlabel=String, ylabel=String, 
                               color=[#], transform=&func, oplot,
                               ); 

  Plot the nc^th parameter chain for the i^th free parameter from an
  MCMC chain ensemble that was made up of nw walkers per free
  parameter.  Alternatively, plot the chain with a function applied,
  passed as a reference via the transform qualifier (e.g.,
  transform=&log10). Plot using pgplot color given by qualifier
  (default=[1], black).
  The plot ranges can be set via the
  xrange/yrange qualifiers, and the plot labels can be overridden via
  the xlabel/ylabel qualifiers.  The plot will be an overplot if the
  oplot qualifier is set. The plot function presumes that the
  appropriate fit model has been loaded.
`); %}}}
      return;
   }
   connect_points(0);
   variable xr = qualifier("xrange",{NULL,NULL});
   variable yr = qualifier("yrange",{NULL,NULL});
   xrange(xr[0],xr[1]); yrange(yr[0],yr[1]);

   variable xlab = qualifier("xlabel",NULL);
   variable ylab = qualifier("ylabel",NULL);
   if(ylab==NULL)
   {
      ylabel("\\fr "+get_par_info(free_par[i]).name);
   }
   else
   {
      ylabel(ylab);
   }
   if(xlab==NULL)
   {
      xlabel("\\fr Iteration");
   }
   else
   {
      xlabel(xlab);
   };

   variable lans=length(ans[0]);
   variable lfp = length(free_par);
   variable col = qualifier("color",1);
   variable ichain=[nc-1:lans-1:lfp*nw];
   variable y=ans[i][ichain];
   variable x=[0:length(ichain)-1];
   variable transf = qualifier("transform",&unity);
   ifnot(qualifier_exists("oplot"))
   { 
     plot(x,@transf(y),col);
   }
   oplot(x,@transf(y),col);
}
%}}}



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#ifeval (__get_reference("xfig_plot_new")!=NULL)
public define xfig_chain_plot() %{{{
{
   if(_NARGS==1)
   {
      variable nu;
      (nu) = ();
   }  
   else if(_NARGS==4)
   {
      variable i, nc, nw, ans;
      (i, nc, nw, ans) = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  xfig_chain_plot(i, nc, nw, chain; xrange={# or NULL,# or NULL}, 
                               yrange={# or NULL,# or NULL},
                               xlabel=String, ylabel=String, 
                               color=[#], transform=&func); 
  xfig_chain_plot(nu; xrange={# or NULL,# or NULL},
                               yrange={# or NULL,# or NULL},
                               xlabel=String, ylabel=String);

  Xfig version of chain_plot. See the appropriate help for further details.
  This function allows to plot multiple chains with different colors,
  if nc and col are given as arrays.
  The function will return a xfig plot struct.
  All qualifiers are passed to the xfig_plot.plot function.

  For only one argument (e.g. the acceptance fraction),
  this function plots the according array.
`); %}}}
      return;
   }

   variable xr,yr;
   variable col = qualifier("color",[1]);
   variable xlab = qualifier("xlabel",NULL);
   variable ylab = qualifier("ylabel",NULL);
   variable P = xfig_plot_new();
   if(_NARGS==4) {
     connect_points(0);
     if(ylab==NULL)
     {
       ylab = get_par_info(free_par[i]).name;
     }
     if(xlab==NULL)
     {
       xlab = "Iteration";
     }
     variable lans=length(ans[0]);
     variable lfp = length(free_par);
     xr = qualifier("xrange",{0,lans/lfp/nw});
     yr = qualifier("yrange",{0,1});
     P.world(xr[0],xr[1],yr[0],yr[1]);
     variable k;
     ifnot (typeof(nc)==Array_Type) nc = [nc];
     _for k (0,length(nc)-1,1) {
       variable ichain=[nc[k]-1:lans-1:lfp*nw];
       variable y=ans[i][ichain];
       variable x=[0:length(ichain)-1];
       y=y[where(y!=0)];
       x=x[where(y!=0)];          
       variable transf = qualifier("transform",&unity);
       P.plot(x,@transf(y) ;; struct_combine(__qualifiers, struct { color = (typeof(col)==Array_Type ? col[k] : col) }));
     }
%     return P;
   }
  
   if(_NARGS==1) {
     if(ylab==NULL)
     {
       ylab = "Acceptance fraction";
     }
     if(xlab==NULL)
     {
       xlab = "Iteration";
     }     
     xr = qualifier("xrange",{0,length(nu)});
     yr = qualifier("yrange",{0,1});
     P.world(xr[0],xr[1],yr[0],yr[1]);
     if (typeof(col)==Array_Type) col = col[0];
     variable idx = where(nu!=0);
     P.plot([1:length(nu[idx])],nu[idx] ;; struct_combine(__qualifiers, struct { color = col }));     
   }

   P.xlabel(xlab);
   P.ylabel(ylab);
   return P;

}
%}}}
#endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% If we are going to overplot a contour, we must use the same grid
% binning as last time!  (And the same transform function!)
private variable onbinx=50, onbiny=50, oxlo, oxhi, oylo, oyhi, otransfx, otransfy;

public define chain_hist2d() %{{{
{
   if(_NARGS==3)
   {
      variable i, j, y;
      (i, j, y) = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  chain_hist2d(i, j, chain; fcut=#, lcut=#, nbinx=#, nbiny=#, 
                            transform_x=&func_x, transform_y=&func_y, 
                            function_x=&func_x, function_y=&func_y,  
                            func_vars_x = [Array], func_vars_y = [Array],
                            xlo=#, xhi=#, ylo=#, yhi=#,
                            color=Array, linestyle=Array,
                            levels=Array, xlabel=String, ylabel=String,
                            oplot, best_fit, return_cstruct); 

  Plot contours for a normalized, 2D histogram of the MCMC chain for
  the i^th (x) and j^th (y) free parameters. Throw out the first fcut
  fraction of the data (default=0.2) and the last lcut fraction of the
  data (default=0), and use nbinx/nbiny bins for each axis
  (default=50), which are histogrammed over the min/max values of the
  chain, or the input xlo/xhi/ylo/yho values.  

  The x- and y-axes can be transformed by passing functions as
  references via the transform_x and transform_y qualifiers,
  respectively (e.g., transform_x=&log10).

  Alternatively, multivariate functions can be passed via the
  function_x/y qualifiers, with func_vars_x/y being arrays indicating
  which of the i/j free parameters serve as variables.  These choices
  supercede the parameter indices input, i/j, and any transforms
  input.  Each function should expect input of the form x[i] and
  y[j], where the i/j indices are for the free parameters of the fit.

  Contours will use the pgplot color index, color (default=1=black),
  linestyle, line_style (default=1=solid), and will be placed at the
  probability values in the array, levels (default=[0.68,0.9,0.95]).
  Use custom X/Y labels with the x/ylabel qualifiers.  Overplot on an
  existing contour if the oplot qualifier is set. (This also forces
  nbinx/nbiny and xlo/xhi/ylo/yhi and transform_x/transform_y values
  to be the same as the previously plotted values.)  If the best_fit
  qualifier is input, then an x-mark is placed at the location
  indicated by the parameter file (presumed to be the best-fit), not
  at the peak of the marginalized probability distribution.

  If the return_cstruct qualifier is set, the structure containing the
  plotted contours will be returned with the additional field levels
  containing the minimum number of walkers (in one histogram bin)
  corresponding to the given confidence levels.
`); %}}}
      return;
   }
   connect_points(-1);
   variable nbinx = qualifier("nbinx",50);
   variable nbiny = qualifier("nbiny",50);
   variable fcut = qualifier("fcut",0.2);
   variable lcut = qualifier("lcut",0);
   variable levels = qualifier("levels",[0.68,0.9,0.95]);
   variable pg_col = [qualifier("color", [1:length(levels)]+1)];
   variable pg_lstyle = [qualifier("linestyle",ones(length(levels)))];
   pg_lstyle = [qualifier("line_style",pg_lstyle)];
   variable xlab=qualifier("xlabel",NULL);
   variable ylab=qualifier("ylabel",NULL);
   variable transfx = qualifier("transform_x",&unity);
   variable transfy = qualifier("transform_y",&unity);
   variable xbest, ybest;

   if(qualifier_exists("oplot"))
   {
      nbinx = onbinx; nbiny = onbiny; transfx = otransfx; transfy = otransfy;
   }
   else
   {
      onbinx = nbinx; onbiny = nbiny; otransfx = transfx; otransfy = transfy;
   }

   if(xlab==NULL)
   {
      xlabel("\\fr "+get_par_info(free_par[i]).name);
   }
   else
   {
      xlabel(xlab);
   }

   if(ylab==NULL)
   {
      ylabel("\\fr "+get_par_info(free_par[j]).name);
   }
   else
   {
      ylabel(ylab);
   }

   variable xgrid, ygrid, xlo, xhi, ylo, yhi, ycut;
   ycut = length(y[i]);
   ycut = [[int(fcut*ycut):ycut-1-int(lcut*ycut)]];

   variable xkeep, ykeep, func, func_vars;
   if( qualifier_exists("function_x") && qualifier_exists("func_vars_x") ) 
   {
      func = qualifier("function_x",&unity);
      func_vars = qualifier("func_vars_x",i);
      variable inx, nx=length(func_vars), xtemp=y[func_vars];
      _for inx (0,nx-1,1) { xtemp[inx]=xtemp[inx][ycut]; }
      xkeep = @func(xtemp);
   }
   else
   {
      xkeep = @transfx(y[i][ycut]);
   }
   if( qualifier_exists("function_y") && qualifier_exists("func_vars_y") ) 
   {
      func = qualifier("function_y",&unity);
      func_vars = qualifier("func_vars_y",j);
      variable iny, ny=length(func_vars), ytemp=y[func_vars];
      _for iny (0,ny-1,1) { ytemp[iny]=ytemp[iny][ycut]; }
      ykeep = @func(ytemp);
   }
   else
   {
      ykeep = @transfy(y[j][ycut]);
   }

   if(qualifier_exists("oplot"))
  {
      xlo = oxlo;
      xhi = oxhi;
  }
  else
  {
      xlo = min(xkeep);
      xhi = max(xkeep);
      xlo = 1.*qualifier("xlo",xlo);
      xhi = 1.*qualifier("xhi",xhi);
      oxlo = xlo;
      oxhi = xhi;
  }
   xgrid = [0:nbinx]*(xhi-xlo)/(nbinx-1) + xlo; 

   if(qualifier_exists("oplot"))
   {
      ylo = oylo;
      yhi = oyhi;
   }
   else
   {
      ylo = min(ykeep);
      yhi = max(ykeep);
      ylo = 1.*qualifier("ylo",ylo);
      yhi = 1.*qualifier("yhi",yhi);
      oylo = ylo;
      oyhi = yhi;
   }
   ygrid = [0:nbiny]*(yhi-ylo)/(nbiny-1) + ylo; 

   variable n = histogram2d( ykeep, xkeep, ygrid, xgrid );
   variable icull=[0:nbinx-1];
   variable ii;
   _for ii (1,nbiny-1,1) { icull = [icull, ii*(nbinx+1)+[0:nbinx-1]]; }
   variable nn = n[icull];
   reshape (nn,[nbiny,nbinx]);

   variable n1d = @nn;
   reshape(n1d,[length(n1d)]);
   
   variable isort = array_sort(n1d);
   n1d = reverse(n1d[isort]);
   variable nprob = cumsum(n1d);
   nprob /= nprob[-1];

   isort = array_sort(levels);
   levels = levels[isort];

   variable iw, lev, clevels = Double_Type[0], plevels= Double_Type[0];

   foreach lev (levels)
   {
      iw = where(nprob >= lev);
      if(length(iw)>0) 
      {
         clevels = [clevels,n1d[iw[0]]];
         plevels = [plevels,lev];
      }
   }

%   print("Plotting contour levels:");
%   print(plevels);

   variable wfmn = wherefirstmax(nn);
   variable iy = wfmn/nbinx;
   variable ix = wfmn mod nbinx;

   point_style(5); 
   variable ogpo = get_plot_options;
   variable pconf = struct{width, type, color};
   variable cstruct = struct{chisqr, px, py, best, px_best, py_best};
   cstruct.best=0;
   cstruct.px = conf_grid(i,xgrid[1],xgrid[nbinx],nbinx);
   cstruct.py = conf_grid(j,ygrid[1],ygrid[nbiny],nbiny);
   cstruct.chisqr = nn;
   cstruct.px_best = xgrid[0];
   cstruct.py_best = ygrid[0];
   pconf.width = ogpo.line_width;
   pconf.type = pg_lstyle;
   pconf.color = pg_col;
   xlin; ylin; 
   if(qualifier_exists("oplot"))
   {
      % Weird ISIS thing, where the color doesn't seem to get set in
      % an oplot_contour, so do the point first to set the color.
      if(qualifier_exists("best_fit"))
      {
         xbest = @transfx(get_par(free_par[i]));
         ybest = @transfy(get_par(free_par[j]));
         connect_points(0); oplot(xbest,ybest,pg_col);
      }
      else
      {
         connect_points(0); oplot((xgrid[ix]+xgrid[ix+1])/2,(ygrid[iy]+ygrid[iy+1])/2,pg_col);
      }
      connect_points(1); oplot_conf(cstruct,pconf,clevels);
   }
   else
   {
      connect_points(1); plot_conf(cstruct,pconf,clevels);

      if(qualifier_exists("best_fit"))
      {
         xbest = @transfx(get_par(free_par[i]));
         ybest = @transfy(get_par(free_par[j]));
         connect_points(0); oplot(xbest,ybest,pg_col);
      }
      else
      {
         connect_points(0); oplot((xgrid[ix]+xgrid[ix+1])/2,(ygrid[iy]+ygrid[iy+1])/2,pg_col);
      }
   }
   set_plot_options(ogpo);

   if (qualifier_exists("return_cstruct")) {
     return struct_combine(cstruct, struct { levels = clevels });
   }
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% If we are going to overplot a contour, we must use the same grid
% binning as last time!  (And the same transform function!)
private variable onbinx_vec=50, onbiny_vec=50, oxlo_vec, oxhi_vec, oylo_vec, oyhi_vec;

public define vec_hist2d() %{{{
{
   if(_NARGS==2)
   {
      variable yi, yj;
      (yi, yj) = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  vec_hist2d(chain_vector1, chain_vector1; fcut=#, lcut=#, nbinx=#, nbiny=#, 
                            xlo=#, xhi=#, ylo=#, yhi=#, color=#,  line_style=#,
                            levels=Array, xlabel=String, ylabel=String, oplot,
                            return_cstruct); 

  Plot contours for a normalized, 2D histogram for two vectors from
  the MCMC chain.  Throw out the first fcut fraction of the data
  (default=0.2) and the last lcut fraction of the data (default=0),
  and use nbinx/nbiny bins for each axis (default=50), which are
  histogrammed over the min/max values of the chain, or the input
  xlo/xhi/ylo/yho values.

  Contours will use the pgplot color index, color (default=1=black),
  linestyle, line_style (default=1=solid), and will be placed at the
  probability values in the array, levels (default=[0.68,0.9,0.95]).
  Use custom X/Y labels with the x/ylabel qualifiers.  Overplot on an
  existing contour if the oplot qualifier is set. (This also forces
  nbinx/nbiny and xlo/xhi/ylo/yhi and transform_x/transform_y values
  to be the same as the previously plotted values.)

  If the return_cstruct qualifier is set, the function will return the 
  structure used to plot the confidence contours.
`); %}}}
      return;
   }
   connect_points(-1);
   variable nbinx = qualifier("nbinx",50);
   variable nbiny = qualifier("nbiny",50);
   variable fcut = qualifier("fcut",0.2);
   variable lcut = qualifier("lcut",0);
   variable levels = qualifier("levels",[0.68,0.9,0.95]);
   variable pg_col = qualifier("color",1);
   variable pg_lstyle = qualifier("linestyle",1);
   pg_lstyle = qualifier("line_style",pg_lstyle);
   variable xlab=qualifier("xlabel","\\frX");
   variable ylab=qualifier("ylabel","\\frY");
   variable xbest, ybest;

   if(qualifier_exists("oplot"))
   {
      nbinx = onbinx_vec; nbiny = onbiny_vec; 
   }
   else
   {
      onbinx_vec = nbinx; onbiny_vec = nbiny; 
   }

   xlabel(xlab);
   ylabel(ylab);

   variable xgrid, ygrid, xlo, xhi, ylo, yhi, ycut;
   ycut = length(yi);
   ycut = [[int(fcut*ycut):ycut-1-int(lcut*ycut)]];

   variable xkeep = yi[ycut];
   variable ykeep = yj[ycut];

   if(qualifier_exists("oplot"))
   {
      xlo = oxlo_vec;
      xhi = oxhi_vec;
   }
   else
   {
      xlo = min(xkeep);
      xhi = max(xkeep);
      xlo = 1.*qualifier("xlo",xlo);
      xhi = 1.*qualifier("xhi",xhi);
      oxlo_vec = xlo;
      oxhi_vec = xhi;
   }
   xgrid = [0:nbinx]*(xhi-xlo)/(nbinx-1) + xlo; 

   if(qualifier_exists("oplot"))
   {
      ylo = oylo_vec;
      yhi = oyhi_vec;
   }
   else
   {
      ylo = min(ykeep);
      yhi = max(ykeep);
      ylo = 1.*qualifier("ylo",ylo);
      yhi = 1.*qualifier("yhi",yhi);
      oylo_vec = ylo;
      oyhi_vec = yhi;
   }
   ygrid = [0:nbiny]*(yhi-ylo)/(nbiny-1) + ylo; 

   variable n = histogram2d( ykeep, xkeep, ygrid, xgrid );
   variable icull=[0:nbinx-1];
   variable ii;
   _for ii (1,nbiny-1,1) { icull = [icull, ii*(nbinx+1)+[0:nbinx-1]]; }
   variable nn = n[icull];
   reshape (nn,[nbiny,nbinx]);

   variable n1d = @nn;
   reshape(n1d,[length(n1d)]);
   
   variable isort = array_sort(n1d);
   n1d = reverse(n1d[isort]);
   variable nprob = cumsum(n1d);
   nprob /= nprob[-1];

   isort = array_sort(levels);
   levels = levels[isort];

   variable iw, lev, clevels = Double_Type[0], plevels= Double_Type[0];

   foreach lev (levels)
   {
      iw = where(nprob >= lev);
      if(length(iw)>0) 
      {
         clevels = [clevels,n1d[iw[0]]];
         plevels = [plevels,lev];
      }
   }

%   print("Plotting contour levels:");
%   print(plevels);

   variable wfmn = wherefirstmax(nn);
   variable iy = wfmn/nbinx;
   variable ix = wfmn mod nbinx;

   point_style(5); 
   variable ogpo = get_plot_options;
   variable pconf = struct{width, type, color};
   variable cstruct = struct{chisqr, px, py, best, px_best, py_best};
   cstruct.best=0;
   cstruct.px = conf_grid(1,xgrid[1],xgrid[nbinx],nbinx);
   cstruct.py = conf_grid(2,ygrid[1],ygrid[nbiny],nbiny);
   cstruct.chisqr = nn;
   cstruct.px_best = xgrid[0];
   cstruct.py_best = ygrid[0];
   pconf.width = ogpo.line_width;
   pconf.type = pg_lstyle;
   pconf.color = pg_col;

   xlin; ylin; 
   if(qualifier_exists("oplot"))
   {
      % Weird ISIS thing, where the color doesn't seem to get set in
      % an oplot_contour, so do the point first to set the color.

      connect_points(0); oplot((xgrid[ix]+xgrid[ix+1])/2,(ygrid[iy]+ygrid[iy+1])/2,pg_col);
      connect_points(1); oplot_conf(cstruct,pconf,clevels);  %oplot_contour(nn,clevels);
   }
   else
   {
      connect_points(1); plot_conf(cstruct,pconf,clevels); 
      % plot_contour(nn,1,xgrid[[1:nbinx]],ygrid[[1:nbiny]],clevels);
      connect_points(0); oplot((xgrid[ix]+xgrid[ix+1])/2,(ygrid[iy]+ygrid[iy+1])/2,pg_col);
   }
   set_plot_options(ogpo);

   if(qualifier_exists("return_cstruct")) return cstruct;
}
%}}}

public define plot_covar() %{{{
{ 
   if(_NARGS==2)
   {
      variable args = __pop_list(_NARGS);
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  id=plot_covar([pars], chain; fcut=#, lcut=#, nbin=Array, 
                        prange={{lo_0,hi_0},{lo_1,hi_1},...}, 
                        hrange={{lo_0,hi_0},{lo_1,hi_1},...}, 
                        plabel=String_Array, hlabel=String_Array, 
                        pmin=Array, pmax=Array, 
                        levels=Array, color=Array, linestyle=Array,
                        device=String, window=#, 
                        viewport=#, size=#, aspect=#,
                        charsize=#, point_size=#, best_fit);

  Create an NxN grid where along the diagonal 1D probability
  histograms are plotted, and the lower left corner contains the
  N(N-1)/2 2D probability contours for the results of an MCMC chain.
  The return value id is the identifier for the opened pgplot device.

  Parameters to be plotted are passed in the [pars] array, and are
  numbered by their order in the chain (i.e., [0,1,2, ...]).  

  Throw out the first fcut fraction of the data (default=0.2) and the last
  lcut fraction of the data (default=0).  Use nbin[i] bins for
  each parameter axis (default=50). 

  pmin[i], pmax[i] retain the same meaning as in chain_hist(), except
  are now arrays, with each value corresponding to the input
  parameters.

  prange[i] is a List of Lists, with each entry being the range over
  which each parameter will have its histogram/contour plotted.
  (Defaults are pmin[i]/pmax[i], or the min/max value of the parameter
  in the portion of the chain that is used.)

  hrange[i] is a List of Lists, with each entry being the range over
  which each parameter *historgram* will have its y-axis plotted
  (Defaults are to autoscale.)

  plabel[i] is a String_Array of the axis labels associated with each
  parameter.  (Default is to take the parameter name from the model
  parameter file, which is presumed to have been loaded.)

  hlabel[i] is a String_Array of the y-axis labels associated with each
  parameter for the histogram plots (default="\\fr Probability").

  Contours will be plotted for the probability levels given by the
  levels array (default=[0.68,0.9,0.95]).

  Histograms/contours will use the pgplot color indices in the color
  array, with the first entry being for the probability histograms,
  and the remaining entries are for the contour levels
  (default=1=black for all). The linestyle array serves the same
  purpose for the lines (default=1=solid).

  The device qualifier indicates the pgplot device (default="/xw").
  Use, for example, device="covar_plot.ps/vcps" to create a postscript
  plot.  (A postscript plot will later have to be closed via
  close_plot(id).)  The window=id qualifier allows an existing pgplot
  device to be reused, but *not* have its basic grid structure (i.e.,
  number of parameters plotted) changed.

  viewport is the standard pgplot definition *for each column*
  (default=[0.1,0.95,0.1,0.98]).  size and aspect are the inputs to
  the resize(size,aspec) command *for the whole* window (defaults=20,
  1, respectively).

  charsize and point_size are the usual pgplot commands (default=1).
`); %}}}
      return;
   }
   variable pars = args[0];
   variable chain = args[1];
   variable n = length(pars);
   variable pr={}, pr_tmp = qualifier("prange",{{NULL,NULL}});
   variable hr={}, hr_tmp = qualifier("hrange",{{NULL,NULL}});
   variable plab=String_Type[n], 
            plab_tmp = qualifier("plabel",String_Type[0]);
   variable hlab=String_Type[n], 
            hlab_tmp = [qualifier("hlabel","\\fr Probability")];
   variable pbin=Integer_Type[n], 
            pbin_tmp=[qualifier("nbin",Integer_Type[0])];
   variable fcut = qualifier("fcut",0.2);
   variable lcut = qualifier("lcut",0);
   variable l = length(chain[0]);
   variable ikeep = [int(fcut*l):l-1-int(lcut*l)];
   variable pmin=Double_Type[n], pmin_tmp = [qualifier("pmin",Double_Type[0])];
   variable pmax=Double_Type[n], pmax_tmp = [qualifier("pmax",Double_Type[0])];
   variable levels = qualifier("levels",[0.68,0.9,0.95]);
   variable col = [qualifier("color",1)];
            if(length(col)<length(levels)+1)
            { 
               col=[col[[0:length(col)-1]],
                    Integer_Type[length(levels)-length(col)+1]+col[-1]];
            }
   variable pg_lstyle = [qualifier("linestyle",1)];
            pg_lstyle = [qualifier("line_style",pg_lstyle)];
            if(length(pg_lstyle)<length(levels)+1)
            {
               pg_lstyle=[pg_lstyle[[0:length(pg_lstyle)-1]],
                          Integer_Type[length(levels)-length(pg_lstyle)+1]
                          +pg_lstyle[-1]];
            }
   variable csize = qualifier("charsize",1);
   variable psize = qualifier("point_size",1);
            psize = qualifier("pointsize",psize);
   variable device=qualifier("device","/xw");
   variable size=qualifier("size",20);
   variable aspect=qualifier("aspect",1);
   variable id;

   % Simple error checks
   if( typeof(pr_tmp)!=List_Type && 
      (typeof(pr_tmp[0])!=List_Type && pr_tmp[0]!=NULL) )
   { message("\n prange must be a List of Lists \n"); return; }
   if( typeof(hr_tmp)!=List_Type && 
      (typeof(hr_tmp[0])!=List_Type && hr_tmp[0]!=NULL) )
   { message("\n hrange must be a List of Lists \n"); return; }
   if( typeof(plab_tmp)!=Array_Type && 
      (typeof(plab_tmp[0])!=String_Type && plab_tmp[0]!=NULL) )
   { message("\n plabel must be Array of Strings \n"); return; }
   if( typeof(hlab_tmp)!=Array_Type && 
      (typeof(hlab_tmp[0])!=String_Type && hlab_tmp[0]!=NULL) )
   { message("\n hlabel must be Array of Strings \n"); return; }
   if( typeof(pbin_tmp)!=Array_Type && 
      (typeof(pbin_tmp[0])!=Integer_Type && pbin_tmp[0]!=NULL) )
   { message("\n nbin must be an Array of Integers\n"); return; }
   if( typeof(pmin_tmp)!=Array_Type && 
      (typeof(pmin_tmp[0])!=Double_Type && pmin_tmp[0]!=NULL) )
   { message("\n pmin must be an Array of Doubles\n"); return; }
   if( typeof(pmax_tmp)!=Array_Type && 
      (typeof(pmax_tmp[0])!=Double_Type && pmax_tmp[0]!=NULL) )
   { message("\n pmax must be an Array of Doubles\n"); return; }

   variable i, j, iarray;
   
   % Set the defaults
   _for i (0,n-1,1)
   {
      try{
         if(pmin_tmp[i]!=NULL){ pmin[i]=pmin_tmp[i]; }
         else{ pmin[i] = min(chain[i][ikeep]);}
      }
      catch AnyError: {
         pmin[i] = min(chain[i][ikeep]);
      }
      try{
         if(pmax_tmp[i]!=NULL){ pmax[i]=pmax_tmp[i]; }
         else{ pmax[i] = max(chain[i][ikeep]);};
      }
      catch AnyError: {
         pmax[i] = max(chain[i][ikeep]);
      }
      try{ list_append(pr,pr_tmp[i]); }
      catch AnyError: { list_append(pr,{NULL,NULL}); }
      if(length(pr[i])!=2) { pr[i] = {pmin[i],pmax[i]}; }
      if(pr[i][0]==NULL) { pr[i][0]=pmin[i]; }
      if(pr[i][1]==NULL) { pr[i][1]=pmax[i]; }
      try{ list_append(hr,hr_tmp[i]); }
      catch AnyError: { list_append(hr,{NULL,NULL}); }
      try{ 
         if(plab_tmp[i]==NULL)
         {  plab[i]="\\fr "+get_par_info(free_par[pars[i]]).name; }
         else { plab[i]=plab_tmp[i]; }
      }
      catch AnyError: { 
         plab[i]="\\fr "+get_par_info(free_par[pars[i]]).name; 
      }
      try{
         hlab[i]=hlab_tmp[i];
      }
      catch AnyError: {
         hlab[i]="\\fr Probability";
      }
      try{ 
         if(pbin_tmp[i]==NULL)
         {  pbin[i]=50; }
         else { pbin[i]=50; }
      }
      catch AnyError: { 
         pbin[i]=50;
      }
   }

   if(qualifier_exists("window"))
   {
      id = qualifier("window");
      window(id);
   }
   else
   {
      id=open_plot(device,n,1);
   }
   variable ov=qualifier("viewport",[0.1,0.95,0.1,0.98]);
   resize(size,aspect);
   charsize(csize);
   point_size(psize);
%
   variable ov_strct = struct{xmin,xmax,ymin,ymax};
   ov_strct.xmin=ov[0];
   ov_strct.xmax=ov[1];
   ov_strct.ymin=ov[2];

   _for i (0,n-1,1)
   {
      iarray=Integer_Type[n-i]+1;
      multiplot(iarray);
      ov_strct.ymax=ov[2]+(n-i)/(1.*n)*(ov[3]-ov[2]);

      set_outer_viewport(ov_strct);

      _for j (0,n-i-1,1)
      {
         mpane(j+1);
         if(j==0)
         {
            chain_hist(pars[i],chain;fcut=fcut,lcut=lcut,nbin=pbin[i],
               pmin=pmin[i],pmax=pmax[i],xrange=pr[i],yrange=hr[i],
               color=col[0],linestyle=pg_lstyle[0],ylabel=hlab[i]);
         }
         else
         {
            if(qualifier_exists("best_fit"))
            {
               chain_hist2d(pars[i],pars[j+i],chain;fcut=fcut,lcut=lcut,
                  nbinx=pbin[i],nbiny=pbin[j+i],xlo=pr[i][0],
                  xhi=pr[i][1],ylo=pr[j+i][0],yhi=pr[j+i][1],
                  color=col[[1:length(col)-1]],
                  linestyle=pg_lstyle[[1:length(pg_lstyle)-1]],
                  levels=levels,xlabel=plab[i],ylabel=plab[j+i]);
            } 
            else
            {
               chain_hist2d(pars[i],pars[j+i],chain;fcut=fcut,lcut=lcut,
                  nbinx=pbin[i],nbiny=pbin[j+i],xlo=pr[i][0],
                  xhi=pr[i][1],ylo=pr[j+i][0],yhi=pr[j+i][1],
                  color=col[[1:length(col)-1]],
                  linestyle=pg_lstyle[[1:length(pg_lstyle)-1]],
                  levels=levels,xlabel=plab[i],ylabel=plab[j+i]);
            } 
         }
      }
   }
   return id;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Old -> New Chain Format Routines.  To Be Removed Eventually??? %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define old_read_chain() %{{{
{
   if(_NARGS==1)
   {
      variable fname = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  (nw, nfreepar, freepar, chain) = old_read_chain(fname);
  (nw, nfreepar, freepar, 
   cdf_scl_lo, cdf_scl_hi,
   chain, chain_statistic, 
   update_indicator, pfile, 
   walkers_init, statistic_init) = old_read_chain(fname; init_walker);
   frac_update                   = old_read_chain(fname; frac_update);

   Read the MCMC results stored in a fits file of the format produced
   by versions <0.7 of this code.  Normally, it will return just the
   minimum results required for plotting.  Qualifiers, however, can be
   used to have it return enough information to initialize a chain, or
   just return an array that has the fraction of walkers updated at
   each step.
`); %}}}
      return;
   }

   variable nw, nfreepar, freepar, cdf_scl_lo, cdf_scl_hi, pfile,
            chain, chainstat, chng, frac_update, statinit;

   (nw, nfreepar, pfile, cdf_scl_lo, cdf_scl_hi ) = 
      fits_read_key(fname+"[MCMCCHAIN]","NWALKERS","NFREEPAR","PFILE",
                                        "CDF_SCL_LO","CDF_SCL_HI");
   (freepar, chain) = fits_read_col(fname+"[MCMCCHAIN]",
                                    "FREE_PAR","CHAIN"  );
   chainstat        = fits_read_col(fname+"[CHAINSTEPS]","FITSTAT");
   chng             = fits_read_col(fname+"[CHAINSTEPS]","UPDATE");
   frac_update      = fits_read_col(fname+"[CHAINSTEPS]","FRAC_UPDATE");

   if(qualifier_exists("init_walker"))
   {
      variable iwalk   = Double_Type[nfreepar,nw*nfreepar];
      variable clength = length(chain[0,*]);
      iwalk            = chain[*,[clength-nw*nfreepar:clength-1]];
      statinit         = chainstat[-1,*];
      return nw, nfreepar, freepar, cdf_scl_lo, cdf_scl_hi, 
             chain, chainstat, chng, pfile, iwalk, statinit;
   }
   else if(qualifier_exists("frac_update"))
   {
      return frac_update;
   }
   else
   {
      return nw, nfreepar, freepar, chain;
   }
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define old_to_new_chain() %{{{
{
   if(_NARGS==2)
   {
      variable fname_old, fname_new;
      (fname_old, fname_new) = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
   old_to_new_chain(fname_old,fname_new);

   Takes the chain results stored in a version <= 0.7.x file,
   fname_old, and writes them to a version >= 0.8.0 file, fname_new.
`); %}}}
      return;
   }

   variable nw, nfreepar, freepar, cdf_scl_lo, cdf_scl_hi, chain,
    chain_statistic, update_indicator, pfile;

   (nw, nfreepar, freepar, cdf_scl_lo, cdf_scl_hi, 
    chain, chain_statistic, update_indicator, pfile, , ) = 
                      old_read_chain(fname_old; init_walker);

   write_chain(fname_new, nw, freepar, cdf_scl_lo, cdf_scl_hi, chain,
    chain_statistic, update_indicator, pfile; create);
}

%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  MPI Routines                                                   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


#ifexists rcl_mpi_init  %{{{


public define mpi_read_chain() %{{{
{
   if(_NARGS==1)
   {
      variable fname = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
  (nw, nfreepar, freepar, 
   chain_statistic, chain)       = mpi_read_chain(fname);

  (nw, nfreepar, freepar, 
   cdf_scl_lo, cdf_scl_hi,
   pfile, walkers_init, 
   statistic_init, change)       = mpi_read_chain(fname; mpi_init_walker);

   frac_update                   = mpi_read_chain(fname; frac_update);

  (frac_update, min_statistic,
   med_statistic, max_statistic) = mpi_read_chain(fname; chain_stats);

   Read the MCMC results stored in a fits file.  Normally, it will
   return just the minimum results required for plotting.  Qualifiers,
   however, can be used to have it return enough information to
   initialize a chain, return an array that has the fraction of
   walkers updated at each step, or return the fractional update array
   along with the min, median, and maximum value of the statistic at
   each step.
`); %}}}
      return;
   }
   variable i, nw, nfreepar, nsteps, cdf_scl_lo, cdf_scl_hi, pfile, 
            freepar, frac_update, chain, chain_stat, chain_chng, 
            tmp_chain, tmp_stat, tmp_chng, indces; 

   variable statinit, walkinit;

   (pfile) = fits_read_key(fname+"[PARAMETERS]","PFILE");

   (nw, nfreepar, nsteps, cdf_scl_lo, cdf_scl_hi ) = 
      fits_read_key(fname+"[MCMCCHAIN]","NWALKERS","NFREEPAR","NSTEPS",
                                        "CDF_SCL_LO","CDF_SCL_HI");

   freepar = fits_read_col(fname+"[PARAMETERS]","FREE_PAR");

   if( qualifier_exists("mpi_init_walker") )
   {
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     This is what Mike wrote. For some reason it was different than in read_chain
%      chain = Array_Type[nfreepar];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      chain = Double_Type[nfreepar,nw*nfreepar];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      indces = [nsteps*nw*nfreepar-nw*nfreepar:nsteps*nw*nfreepar-1];
      _for i (0,nfreepar-1,1)
      {
         tmp_chain = fits_read_col(fname+"[MCMCCHAIN]","CHAINS"+string(freepar[i]));
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     Again, this is what Mike wrote, different as in read_chain without MPI
%         chain[i] = tmp_chain[indces];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
         chain[i,*] = tmp_chain[indces];
         () = __tmp(tmp_chain);
      }
      
      tmp_stat = fits_read_col(fname+"[MCMCCHAIN]","FITSTAT");
      chain_stat = tmp_stat[indces];
      () = __tmp(tmp_stat);

      tmp_chng = fits_read_col(fname+"[MCMCCHAIN]","UPDATE");
      chain_chng = tmp_chng[indces];
      () = __tmp(tmp_chng);

      return nw, nfreepar, freepar, cdf_scl_lo, cdf_scl_hi, pfile, 
             chain, chain_stat, chain_chng;
   }
   else if(qualifier_exists("frac_update"))
   {
      frac_update = fits_read_col(fname+"[CHAINSTATS]","FRAC_UPDATE");
      return frac_update;
   }
   else if(qualifier_exists("chain_stats"))
   {
      variable min_statistic, med_statistic, max_statistic;
      (frac_update, min_statistic,
       med_statistic, max_statistic)=fits_read_col(fname+"[CHAINSTATS]",
                                           "FRAC_UPDATE", "MIN_STAT",
                                           "MED_STAT", "MAX_STAT"     );
      return frac_update, min_statistic, med_statistic, max_statistic;
   }
   else
   {
      chain = Array_Type[nfreepar];

      _for i (0,nfreepar-1,1)
      {
         chain[i] = fits_read_col(fname+"[MCMCCHAIN]","CHAINS"+string(freepar[i]));
      }
      chain_stat = fits_read_col(fname+"[MCMCCHAIN]","FITSTAT");
      reshape(chain_stat,[nsteps,nw*nfreepar]);

      return nw, nfreepar, freepar, chain_stat, chain;
   }
}
%}}}

private define mpi_init_walker(fp,nw,proc_rank,proc_total) %{{{
{
   variable iret=0; 
   variable npfile=1, fname = "./tmp_mcmc_pfile_"+string(getpid)+".par";
   %variable npfile=1, fname = "./tmp_mcmc_pfile_"+string(proc_rank)+".par";
   variable fname_len=int(1);
   if(proc_rank==0)
   {
      try{
         save_par(fname);
      }
      catch AnyError: {
         () = printf("\n\n Failed saving initial parameters \n\n");
          plot_pause;
      }

      if(qualifier_exists("pfile"))
      {
         variable pfile_init, pfiles = strchop(qualifier("pfile"),';',0);
         npfile = length(pfiles);

         try{
            foreach pfile_init (pfiles)
            {
               load_par(pfile_init);
            }
         }
         catch AnyError: {
            () = printf("\n\n Cannot read parameter file: %s \n",pfile_init);
            () = printf(" Retaining original parameter file for initialization. \n\n");
            npfile = 1;
            pfiles = [fname];
            load_par(fname);
         }
      }
      else
      {
         pfiles = [fname];
      }
      fname_len=strlen(fname);
      fname_len=int(fname_len);
   }
   iret=rcl_mpi_barrier();
   iret=rcl_mpi_bcast_int([npfile],1);
   %iret=rcl_mpi_bcast_int([fname_len],1);
   %iret=rcl_mpi_bcast_char(fname,26);
      
   variable sc, sig;
   if(qualifier_exists("gaussian"))
   {
      sc = qualifier("scale",1./3.);
   }
   else
   {
      sc = qualifier("scale",1.);
   }

   variable fv_hold = Fit_Verbose;
   Fit_Verbose = -1;

   % The walkers are 2D arrays, with the first index covering the free
   % parameters, and the second index being the specific instances
   % (i.e., the length of the 2nd index is the number of walkers).

   variable i, j, k, info, nfp = length(fp),
            xp    = Double_Type[nfp, npfile], xl=@xp, xh=@xp, 
            xstrt = Double_Type[nfp,nw*nfp],
            xstat = Double_Type[nw*nfp];
   
   iret=rcl_mpi_barrier();
   if(proc_rank==0)
   {
      _for k (0,npfile-1,1)
      {   
         load_par(pfiles[k]);
         () = eval_counts;

         _for i (0,nfp-1,1)
         {
            xp[i,k]      = get_par_info(fp[i]).value;
            xl[i,k]      = get_par_info(fp[i]).min;
            xh[i,k]      = get_par_info(fp[i]).max;
         }
      }
   }
   iret=rcl_mpi_barrier();
   iret=rcl_mpi_bcast_double(xp,nfp*npfile);
   iret=rcl_mpi_bcast_double(xl,nfp*npfile);
   iret=rcl_mpi_bcast_double(xh,nfp*npfile);
   iret=rcl_mpi_barrier();
   % Reload the original parameter file, which should have the widest
   % allowable min/max values (enough to encompass any ranges present
   % in the input parameter files)

   if(qualifier_exists("pfile"))
   {
      try{ 
         load_par(fname);
      }
      catch AnyError: {
         () = printf("\n\n Failed reloading original  parameters \n\n");
         plot_pause;
      }
      if(eval_counts)
      {
         () = printf("\n\n Failed evaluating original  parameters \n\n");
          plot_pause;
      }
   }
   variable nw_nfp_section= round(nw*nfp/proc_total);
   variable nw_nfp_min=0;
   variable nw_nfp_max=0;
   if (proc_rank == proc_total-1) 
   {
      nw_nfp_min=int(proc_rank*nw_nfp_section);
      nw_nfp_max=int(nw*nfp-1);
   } 
   else
   {
      nw_nfp_min=int(proc_rank*nw_nfp_section);
      nw_nfp_max=int(((proc_rank+1)*nw_nfp_section)-1);
   }
   %_for j (0,nw*nfp-1,1)
   _for j (nw_nfp_min,nw_nfp_max,1)
   {
      k = j*npfile/(nw*nfp);

      _for i (0,nfp-1,1)
      {
         if(qualifier_exists("gaussian"))
         {
            sig = (@(qualifier("grnd", &grand)))();
            if(sig < 0)
            {
               xstrt[i,j] = xp[i,k] + sig*sc*(xp[i,k]-xl[i,k]);
            }
            else
            {
               xstrt[i,j] = xp[i,k] + sig*sc*(xh[i,k]-xp[i,k]);
            }
         }
         else
         {
            xstrt[i,j] = (1-sc)*xp[i,k] + 
              sc*(xl[i,k]+(xh[i,k]-xl[i,k])*(@(qualifier("urnd", &urand)))());
         }

         if(xstrt[i,j]<xl[i,k]) xstrt[i,j] = xl[i,k];
         if(xstrt[i,j]>xh[i,k]) xstrt[i,j] = xh[i,k];
         set_par(fp[i],xstrt[i,j]);
      }

      () = eval_counts(&info);
      variable stat = info.statistic;
      if (qualifier_exists("priors"))
      {
         stat += gauss_prior(qualifier("priors"));
      }
      xstat[j] = stat;
   }
   iret=rcl_mpi_barrier();
   variable lentag=0;
   if(proc_rank==0)
   {
      _for i (1,proc_total-1,1)
      {
         if (i == proc_total-1) 
         {
            nw_nfp_min=int(i*nw_nfp_section);
            nw_nfp_max=int(nw*nfp-1);
         } 
         else
         {
            nw_nfp_min=int(i*nw_nfp_section);
            nw_nfp_max=int(((i+1)*nw_nfp_section)-1);
         }
         variable xstat_recv=xstat[[nw_nfp_min:nw_nfp_max]];
         iret=rcl_mpi_org_recv_double(xstat_recv, nw_nfp_max-nw_nfp_min+1, i, lentag);
         xstat[[nw_nfp_min:nw_nfp_max]]=xstat_recv[[0:nw_nfp_max-nw_nfp_min]];
      }
   }
   else
   {
      variable xstat_send=xstat[[nw_nfp_min:nw_nfp_max]];
      iret=rcl_mpi_org_send_double(xstat_send, nw_nfp_max-nw_nfp_min+1, 0, lentag);
   }
   iret=rcl_mpi_barrier();
   iret=rcl_mpi_bcast_double(xstat[*],nw*nfp);
   iret=rcl_mpi_barrier();
   _for k (0,nfp-1,1)
   {
      if(proc_rank==0)
      {
         _for i (1,proc_total-1,1)
         {
            if (i == proc_total-1) 
            {
               nw_nfp_min=int(i*nw_nfp_section);
               nw_nfp_max=int(nw*nfp-1);
            } 
            else
            {
               nw_nfp_min=int(i*nw_nfp_section);
               nw_nfp_max=int(((i+1)*nw_nfp_section)-1);
            }
            variable xstrt_recv=xstrt[k,[nw_nfp_min:nw_nfp_max]];
            iret=rcl_mpi_org_recv_double(xstrt_recv, nw_nfp_max-nw_nfp_min+1, i, lentag);
            xstrt[k,[nw_nfp_min:nw_nfp_max]]=xstrt_recv[[0:nw_nfp_max-nw_nfp_min]];
         }
      }
      else
      {
         variable xstrt_send=xstrt[k,[nw_nfp_min:nw_nfp_max]];
         iret=rcl_mpi_org_send_double(xstrt_send, nw_nfp_max-nw_nfp_min+1, 0, lentag);
      }
      iret=rcl_mpi_barrier();
   }
   iret=rcl_mpi_barrier();
   iret=rcl_mpi_bcast_double(xstrt[*,*],nw*nfp*nfp);
   iret=rcl_mpi_barrier();
   Fit_Verbose = fv_hold;
   return xstrt, xstat;
}
%}}}

private define mpi_update_walker(cdf_scl_lo,cdf_scl_hi,rnd,fp,xa,xstat,xb) %{{{
{
   % xa is a vector of parameters for a specific walker to be updated,
   % while xstat is the fit statistic of that walker. xb is an array
   % of walker parameters (a second, independent group) from which an
   % update is drawn.  The first index of xb spans the parameter,
   % while the second spans specific walker instances.

   variable lb = length(xb[0,*]);                   % No. of walkers
   variable xuse = xb[*,int(rnd[0]*lb)];            % Choose one
   variable z = icdf(rnd[1],cdf_scl_lo,cdf_scl_hi); % Random step size

   % New trial parameters
   xuse += z*(xa-xuse); 
   
   %variable priors = qualifier("priors",0);
   variable dlogpn, ystat;
   (dlogpn,ystat) = logpn(fp,xstat,xuse,z;; __qualifiers);

   variable ptest = log(rnd[2]);
   if(ptest <= dlogpn)
   {
      return xuse, ystat, 1;             % 1 = changed value
   }
   else
   {
      return xa, xstat, 0;               % 0 = same value
   }
}
%}}}

private variable MPI_Task_List = 
   struct{ index, length, rands, free_par, walkers, wstats, change, 
           proposals, output };

public define mpi_emcee() %{{{
{
   variable proc_rank=rcl_mpi_init();
   variable proc_total=rcl_mpi_numtasks(); 
   if(_NARGS==2)
   {
      variable nw, nsim;
      (nw, nsim) = ();
   }
   else
   {
      () = fprintf(fs,  "%s\n", %{{{
`
   mpi_emcee( nw, nsim; pfile="file1;file2;...;filen", 
                    scale=#, gaussian, 
                    cdf_scl_lo=#, cdf_scl_hi=#, 
                    init_chain=file, 
                    output=file, sim_count=#, 
                    nocopy, priors, urnd=&urand, grnd=&grand );

  Create a set of MCMC chains for an ensemble of walkers (i.e.,
  propositions) consisting of a number of walkers equal to nw X the
  number of free parameters.  This ensemble will be evolved for nsim
  iterations.  All results will be stored in an output file.

  The initial distribution of walkers is taken from the current
  parameter values, or from the parameter values in the files listed
  in pfile (file names, including full paths if not in the local
  directory, separated by semi-colons).  If the gaussian qualifier is
  not set, each walker has its parameters uniformly distributed from
  value-scale*(value-min) to value+scale*(max-value). (Default
  scale=1.)

  Otherwise, the walkers have initial parameters gaussianly
  distributed about value with sigma = scale*(max-value) or sigma =
  scale*(value-min) (i.e., normalized to the possibly asymmetric
  min/max values).  (Default scale=1/3.)

  Alternatively, the walkers can be initialized from the last
  iteration of an existing chain stored in a file, init_chain.

  The walkers are split into groups (0) and (1), with the update
  proposition (i.e., parameter vector) for the group (0) j^th walker
  being:
            X(0)'_j = X(1)_k + z*(X(0)_j-X(1)_k)
  X(1)_k is randomly drawn from group (1) and z is randomly
  distributed as 1/sqrt(z) between values of 1/cdf_scl_lo and
  cdf_scl_hi (default for both=2). Group (1) is then subsequently
  updated in a similar manner.

  The output chain is placed in the file specified by the output
  qualifier (default name is mcmc_results_processid.fits).
  Intermediate results are written to this file every sim_count
  (default=50) iterations.  If the chains were started from a previous
  chain file, the final output file will contain the initial chains as
  well unless the nocopy qualifier has been set.

  The optional qualifiers grnd and urnd can be set to user defined
  functions to, e.g., allow random numbers drawn from a server or file
  in case of a manual multi-job run (among several machines).

  PRIORS
  One can apply gaussian priors to particular parameters by
  constructing a structure with the following fields
  priors.parlist -- an array of parameter ids
  priors.priorlist -- an array of structures describing each prior. A
     prior structure needs two fields: 'mean' and 'sigma'

  For example:
  myprior = struct{mean=0.1, sigma=0.2}
  mpi_emcee(nw, nsim; priors=struct{parlist=[1], priorlist=[myprior]})
`); %}}}
      return;
   }
   % Store initial parameters so that we can restore them at the end
   variable pars = get_params;

   variable isc=0, sim_count_array, sim_count = int(qualifier("sim_count",50));
   if(sim_count <= 0) sim_count=50;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   sim_count_array=[0:nsim+sim_count:sim_count];
   sim_count_array=[sim_count-1:nsim-1+sim_count:sim_count];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   variable ochain, ostat, ochng;

   variable screwup    = qualifier("infile","NONE");
            screwup    = qualifier("in_file",screwup);
            screwup    = qualifier("start",screwup);
            screwup    = qualifier("start_file",screwup);
            screwup    = qualifier("initfile",screwup);
            screwup    = qualifier("init_file",screwup);
            screwup    = qualifier("input",screwup);
            screwup    = qualifier("initchain",screwup);
   variable init_chain = qualifier("init_chain",screwup);
   
   if ( screwup != "NONE" ) 
      if (proc_rank==0) ()=printf("\n *** Using %s for chain initialization  *** \n\n",init_chain);
      
            screwup    = qualifier("outfile","mcmc_results_");
            screwup    = qualifier("out_file",screwup);
            screwup    = qualifier("stop",screwup);
            screwup    = qualifier("stop_file",screwup);
            screwup    = qualifier("outchain",screwup);
            screwup    = qualifier("out_chain",screwup);
   variable outfile    = qualifier("output",screwup);

   if ( outfile == "mcmc_results_" ) outfile += "_"+string(getpid)+".fits";
   if (proc_rank==0) ()=printf("\n *** Using %s for output file *** \n\n",outfile);

   variable pfile      = qualifier("pfile","NONE");
   variable cdf_scl_lo = qualifier("cdf_scl_lo",2.);
   variable cdf_scl_hi = qualifier("cdf_scl_hi",2.);

   nw   = int(nw);
   nsim = int(nsim);

   variable xwalk, xstat, xwalk_a, xwalk_b, freepar = free_par();
   
   if(init_chain=="NONE")
   {
      (xwalk, xstat) = mpi_init_walker(freepar,nw,proc_rank,proc_total;;__qualifiers);
   }
   else
   {
      variable nwc=0, nfpc=0, fpc;
      try
      {
         (nwc, nfpc, fpc, cdf_scl_lo, cdf_scl_hi, pfile,
          xwalk, xstat, ochng)=mpi_read_chain(init_chain;mpi_init_walker);
         if (proc_rank==0)
         {
            if(qualifier_exists("nocopy"))
            {
               variable wstat=@xstat, wchng=@ochng;
               reshape(wstat,[1,length(wstat)]);
               reshape(wchng,[1,length(wchng)]);
               write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
                           xwalk, wstat, wchng, pfile;create);
            }
            else
            {
               () = system("cp "+init_chain+" "+outfile);
            }
         }
      }
      catch AnyError:
      {
         if (proc_rank==0) ()=printf("\n Failed to read or create chain file \n");
      }
      % check, if present fit corresponds to the one, one wants to append to (init_chain)
      if(nwc!=nw || nfpc!=length(freepar) || max(fpc-freepar)!=0)
      {
         if (proc_rank==0) ()=printf("\n Chain file inconsistent, initializing from data \n");
         (xwalk, xstat) = mpi_init_walker(freepar,nw,proc_rank,proc_total;;__qualifiers);
         init_chain="NONE";
         pfile="NONE";
      }
   }

   if(1/cdf_scl_lo >= cdf_scl_hi || cdf_scl_lo <= 0 || cdf_scl_hi <= 0)
   {
      if (proc_rank==0) ()=printf("\n1/CDF_SCL_LO must be < CDF_SCL_HI and > 0.\n");
      return;
   }

   variable iwlk_a, iwlk_b, la, lb, im, i, j, k, s, 
            nwf       = nw*num_free_params,
            xchng     = @xstat,
            xout      = Double_Type[num_free_params,sim_count*nwf],
            xout_stat = Double_Type[sim_count,nwf],
            xout_chng = Double_Type[sim_count,nwf]; 
   
   iwlk_a = [0:nwf-1:2];
   iwlk_b = [1:nwf-1:2];

   la = length(iwlk_a);
   lb = length(iwlk_b);

   MPI_Task_List.free_par  = freepar;
   MPI_Task_List.output    = Array_Type[max([la,lb])];
   MPI_Task_List.wstats    = Double_Type[max([la,lb])];
   MPI_Task_List.change    = Long_Type[max([la,lb])];

   variable array_xwalk      = Array_Type[max([la,lb])];
   variable array_xwalk_stat = Double_Type[max([la,lb])];

   % Seed all the required random numbers at the start, so we don't
   % have issues with stuff when it comes time to parallelize
   variable rands_a = Array_Type[1];%[nsim*la];
   variable rands_b = Array_Type[1];%[nsim*lb];
   variable rnd_var = Double_Type[3];

   variable rands_a_new = Double_Type[nsim*la,3];
   variable rands_b_new = Double_Type[nsim*lb,3];
   
   %
   if (proc_rank==0)
   {
      _for i (0,nsim*la-1,1) 
      { 
         rands_a[0] = (@(qualifier("urnd", &urand)))(3); 
         rands_a_new[i,0]=rands_a[0][0];
         rands_a_new[i,1]=rands_a[0][1];
         rands_a_new[i,2]=rands_a[0][2];
      }
      _for i (0,nsim*lb-1,1) 
      { 
         rands_b[0] = (@(qualifier("urnd", &urand)))(3); 
         rands_b_new[i,0]=rands_b[0][0];
         rands_b_new[i,1]=rands_b[0][1];
         rands_b_new[i,2]=rands_b[0][2];
      }
   }
   variable iret=0;  
   iret=rcl_mpi_barrier();
   iret=rcl_mpi_bcast_double(rands_a_new,3*nsim*la);
   iret=rcl_mpi_bcast_double(rands_b_new,3*nsim*lb);
   iret=rcl_mpi_barrier();
   
   variable irnd;
   
   
   variable lentag=0;
    
   variable la_section= int(ceil(la/proc_total));
   variable la_min=0;
   variable la_max=0;
   variable iwlk_a_list1= Integer_Type[la_section];
   variable iwlk_a_list2= Integer_Type[la_section];
   
   variable xwalk_recv_a=Double_Type[num_free_params,la_section];
   variable xwalk_recv_a_len=0;
   variable xstat_recv_a= Double_Type[la_section];
   variable xstat_recv_a_len=0;
   variable xchng_recv_a= Double_Type[la_section];
   variable xchng_recv_a_len=0;

   variable xwalk_send_a=Double_Type[num_free_params,la_section];
   variable xwalk_send_a_len=0;
   variable xstat_send_a= Double_Type[la_section];
   variable xstat_send_a_len=0;
   variable xchng_send_a= Double_Type[la_section];
   variable xchng_send_a_len=0;  
     
   variable lb_section= int(ceil(lb/proc_total));
   variable lb_min=0;
   variable lb_max=0;
   variable iwlk_b_list1= Integer_Type[lb_section];
   variable iwlk_b_list2= Integer_Type[lb_section];
   
   variable xwalk_recv_b=Double_Type[num_free_params,lb_section];
   variable xwalk_recv_b_len=0;
   variable xstat_recv_b= Double_Type[lb_section];
   variable xstat_recv_b_len=0;
   variable xchng_recv_b= Double_Type[lb_section];
   variable xchng_recv_b_len=0;
   
   variable xwalk_send_b=Double_Type[num_free_params,lb_section];
   variable xwalk_send_b_len=0;
   variable xstat_send_b= Double_Type[lb_section];
   variable xstat_send_b_len=0;
   variable xchng_send_b= Double_Type[lb_section];
   variable xchng_send_b_len=0;  
   variable xdata_send_b=Double_Type[length(xwalk)+length(xstat)+length(xchng)];
   variable xdata_recv_b=Double_Type[length(xwalk)+length(xstat)+length(xchng)];
   
   variable xwalk_len=0;
   variable xstat_len=0;
   variable xchng_len=0;
       
   iret=rcl_mpi_barrier();
   iret=rcl_mpi_bcast_double(xwalk,length(xwalk));
   iret=rcl_mpi_bcast_double(xstat,length(xstat));
   iret=rcl_mpi_bcast_double(xchng,length(xchng));
   la_section= ceil(la/proc_total);
   lb_section= ceil(lb/proc_total);
   iret=rcl_mpi_barrier();
   _for i (0,nsim-1,1)
   {
      % Update first half of walkers ...
      irnd = i*la+[0:la-1];
      
      if (proc_rank == proc_total-1) 
      {
         la_min=int(proc_rank*la_section);
         la_max=int(la-1);
      } 
      else
      {
         la_min=int(proc_rank*la_section);
         la_max=int(((proc_rank+1)*la_section)-1);
      }
      %_for j (0,la-1,1)
      _for j (la_min,la_max,1)
      {
         array_xwalk[j]      = xwalk[*,iwlk_a[j]];
         array_xwalk_stat[j] = xstat[iwlk_a[j]];
         if(i==0 && j==0) () = printf("\n Start Run \n");
         rnd_var=[rands_a_new[irnd[j],0],rands_a_new[irnd[j],1],rands_a_new[irnd[j],2]];
         (MPI_Task_List.output[j],MPI_Task_List.wstats[j],
          MPI_Task_List.change[j]) = mpi_update_walker(cdf_scl_lo,cdf_scl_hi,
                                               rnd_var, 
                                               freepar,
                                               array_xwalk[j],
                                               array_xwalk_stat[j],
                                               xwalk[*,iwlk_b]);
      }

      %_for j (0,la-1,1)
      _for j (la_min,la_max,1)
      {
         xwalk[*,iwlk_a[j]] = MPI_Task_List.output[j];
         xstat[iwlk_a[j]]   = MPI_Task_List.wstats[j];
         xchng[iwlk_a[j]]   = MPI_Task_List.change[j];
      }
      
      % Update second half of walkers ...
      irnd = i*lb+[0:lb-1];
      
      if (proc_rank == proc_total-1) 
      {
         lb_min=int(proc_rank*lb_section);
         lb_max=int(lb-1);
      } 
      else
      {
         lb_min=int(proc_rank*lb_section);
         lb_max=int(((proc_rank+1)*lb_section)-1);
      }
      
      %_for j (0,lb-1,1)
      _for j (lb_min,lb_max,1)
      {
         array_xwalk[j]      = xwalk[*,iwlk_b[j]];
         array_xwalk_stat[j] = xstat[iwlk_b[j]];
         rnd_var=[rands_b_new[irnd[j],0],rands_b_new[irnd[j],1],rands_b_new[irnd[j],2]];
         (MPI_Task_List.output[j],MPI_Task_List.wstats[j],
          MPI_Task_List.change[j]) = mpi_update_walker(cdf_scl_lo,cdf_scl_hi,
                                               rnd_var, 
                                               freepar,
                                               array_xwalk[j],
                                               array_xwalk_stat[j],
                                               xwalk[*,iwlk_a]);
      }

      %_for j (0,lb-1,1)
      _for j (lb_min,lb_max,1)
      {
         xwalk[*,iwlk_b[j]] = MPI_Task_List.output[j];
         xstat[iwlk_b[j]]   = MPI_Task_List.wstats[j];
         xchng[iwlk_b[j]]   = MPI_Task_List.change[j];
      }
      
      lentag=0;
      iret=rcl_mpi_barrier();
      
      if(proc_rank==0)
      {
         _for k (1,proc_total-1,1)
         {
            if (k == proc_total-1) 
            {
               la_min=int(k*la_section);
               la_max=int(la-1);
            } 
            else
            {
               la_min=int(k*la_section);
               la_max=int(((k+1)*la_section)-1);
            }
            iwlk_a_list1=int(iwlk_a[[la_min:la_max]]);
            xwalk_recv_a_len=int(length(xwalk[*,iwlk_a_list1]));
            xstat_recv_a_len=int(length(xstat[iwlk_a_list1]));
            xchng_recv_a_len=int(length(xchng[iwlk_a_list1]));
            
            if (k == proc_total-1) 
            {
               lb_min=int(k*lb_section);
               lb_max=int(lb-1);
            } 
            else
            {
               lb_min=int(k*lb_section);
               lb_max=int(((k+1)*lb_section)-1);
            }
            iwlk_b_list1=int(iwlk_b[[lb_min:lb_max]]);
            xwalk_recv_b_len=int(length(xwalk[*,iwlk_b_list1]));
            xstat_recv_b_len=int(length(xstat[iwlk_b_list1]));
            xchng_recv_b_len=int(length(xchng[iwlk_b_list1]));
            
            variable xdata_recv_a= Double_Type[xwalk_recv_a_len+3*xstat_recv_a_len+xwalk_recv_b_len+3*xstat_recv_b_len];
            
            iret=rcl_mpi_org_recv_double(xdata_recv_a, xwalk_recv_a_len+3*xstat_recv_a_len + xwalk_recv_b_len+3*xstat_recv_b_len, k, 0);
            
            iwlk_a_list1=int(xdata_recv_a[[0:xstat_recv_a_len-1]]);
            xwalk[*,iwlk_a_list1]=xdata_recv_a[[xstat_recv_a_len+0:xwalk_recv_a_len+xstat_recv_a_len-1]];
            xstat[iwlk_a_list1]=xdata_recv_a[[xwalk_recv_a_len+xstat_recv_a_len:xwalk_recv_a_len+2*xstat_recv_a_len-1]];
            xchng[iwlk_a_list1]=xdata_recv_a[[xwalk_recv_a_len+2*xstat_recv_a_len:xwalk_recv_a_len+3*xstat_recv_a_len-1]];

            iwlk_b_list1=int(xdata_recv_a[[xwalk_recv_a_len+3*xstat_recv_a_len + 0:xwalk_recv_a_len+3*xstat_recv_a_len + xstat_recv_b_len-1]]);
            xwalk[*,iwlk_b_list1]=xdata_recv_a[[xwalk_recv_a_len+3*xstat_recv_a_len + xstat_recv_b_len+0:xwalk_recv_a_len+3*xstat_recv_a_len + xwalk_recv_b_len+xstat_recv_b_len-1]];
            xstat[iwlk_b_list1]=xdata_recv_a[[xwalk_recv_a_len+3*xstat_recv_a_len + xwalk_recv_b_len+xstat_recv_b_len:xwalk_recv_a_len+3*xstat_recv_a_len + xwalk_recv_b_len+2*xstat_recv_b_len-1]];
            xchng[iwlk_b_list1]=xdata_recv_a[[xwalk_recv_a_len+3*xstat_recv_a_len + xwalk_recv_b_len+2*xstat_recv_b_len:xwalk_recv_a_len+3*xstat_recv_a_len + xwalk_recv_b_len+3*xstat_recv_b_len-1]];
         }
         _for k (1,proc_total-1,1)
         {
            xdata_send_b=Double_Type[length(xwalk)+length(xstat)+length(xchng)];
            xdata_send_b[[0:length(xwalk)-1]]=xwalk[*,*];
            xdata_send_b[[length(xwalk):length(xwalk)+length(xstat)-1]]=xstat[*];
            xdata_send_b[[length(xwalk)+length(xstat):length(xwalk)+length(xstat)+length(xchng)-1]]=xchng[*];
            iret=rcl_mpi_org_send_double(xdata_send_b,length(xwalk)+length(xstat)+length(xchng), k, 0); 
         }
      }
      else
      {
         iwlk_a_list2=int(iwlk_a[[la_min:la_max]]);
         xwalk_send_a_len=int(length(xwalk[*,iwlk_a_list2]));
         xstat_send_a_len=int(length(xstat[iwlk_a_list2]));
         xchng_send_a_len=int(length(xchng[iwlk_a_list2]));
         
         iwlk_b_list2=int(iwlk_b[[la_min:la_max]]);
         xwalk_send_b_len=int(length(xwalk[*,iwlk_b_list2]));
         xstat_send_b_len=int(length(xstat[iwlk_b_list2]));
         xchng_send_b_len=int(length(xchng[iwlk_b_list2]));
         
         variable xdata_send_a= Double_Type[xwalk_send_a_len+3*xstat_send_a_len + xwalk_send_b_len+3*xstat_send_b_len];

         xdata_send_a[[0:xstat_send_a_len-1]]=double(iwlk_a_list2);
         xdata_send_a[[xstat_send_a_len+0:xwalk_send_a_len+xstat_send_a_len-1]]=xwalk[*,iwlk_a_list2];
         xdata_send_a[[xwalk_send_a_len+xstat_send_a_len:xwalk_send_a_len+2*xstat_send_a_len-1]]=xstat[iwlk_a_list2];
         xdata_send_a[[xwalk_send_a_len+2*xstat_send_a_len:xwalk_send_a_len+3*xstat_send_a_len-1]]=xchng[iwlk_a_list2];
         
         xdata_send_a[[xwalk_send_a_len+3*xstat_send_a_len + 0:xwalk_send_a_len+3*xstat_send_a_len + xstat_send_b_len-1]]=double(iwlk_b_list2);
         xdata_send_a[[xwalk_send_a_len+3*xstat_send_a_len + xstat_send_b_len+0:xwalk_send_a_len+3*xstat_send_a_len + xwalk_send_b_len+xstat_send_b_len-1]]=xwalk[*,iwlk_b_list2];
         xdata_send_a[[xwalk_send_a_len+3*xstat_send_a_len + xwalk_send_b_len+xstat_send_b_len:xwalk_send_a_len+3*xstat_send_a_len + xwalk_send_b_len+2*xstat_send_b_len-1]]=xstat[iwlk_b_list2];
         xdata_send_a[[xwalk_send_a_len+3*xstat_send_a_len + xwalk_send_b_len+2*xstat_send_b_len:xwalk_send_a_len+3*xstat_send_a_len + xwalk_send_b_len+3*xstat_send_b_len-1]]=xchng[iwlk_b_list2];
         
         iret=rcl_mpi_org_send_double(xdata_send_a, xwalk_send_a_len+3*xstat_send_a_len + xwalk_send_b_len+3*xstat_send_b_len, 0, 0); 
         
         xdata_recv_b=Double_Type[length(xwalk)+length(xstat)+length(xchng)];
         iret=rcl_mpi_org_recv_double(xdata_recv_b, length(xwalk)+length(xstat)+length(xchng), 0, 0);
         xwalk[*,*]=xdata_recv_b[[0:length(xwalk)-1]];
         xstat[*]=xdata_recv_b[[length(xwalk):length(xwalk)+length(xstat)-1]];
         xchng[*]=xdata_recv_b[[length(xwalk)+length(xstat):length(xwalk)+length(xstat)+length(xchng)-1]];
      }
      
      iret=rcl_mpi_barrier();
      
      %xwalk_len=length(xwalk[*,*]);
      %xstat_len=length(xstat[*]);
      %xchng_len=length(xchng[*]);
      %iret=rcl_mpi_bcast_double(xwalk[*,*],xwalk_len);
      %iret=rcl_mpi_bcast_double(xstat[*],xstat_len);
      %iret=rcl_mpi_bcast_double(xchng[*],xchng_len);
      %iret=rcl_mpi_barrier();
      
      if (proc_rank==0)
      {
         im = i mod sim_count;
         
         xout[*,im*nwf+[0:nwf-1:2]] = xwalk[*,iwlk_a];
         xout[*,im*nwf+[1:nwf-1:2]] = xwalk[*,iwlk_b];

         xout_stat[im,[0:nwf-1:2]] = xstat[iwlk_a];
         xout_stat[im,[1:nwf-1:2]] = xstat[iwlk_b];

         xout_chng[im,[0:nwf-1:2]] = xchng[iwlk_a];
         xout_chng[im,[1:nwf-1:2]] = xchng[iwlk_b];
         
         if(i==0){ () = printf("\n Sim loop: \n"); }
         if( i==sim_count_array[isc] || i==nsim-1 )
         {
            if(isc==0)
            {
               if ((init_chain=="NONE") or (qualifier_exists("nocopy")))
               {
                  write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
                              xout, xout_stat, xout_chng, pfile; create);
               }
               else
               {
                  write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
                              xout, xout_stat, xout_chng, pfile);
               }
            }
            else
            {
               write_chain(outfile, nw, freepar, cdf_scl_lo, cdf_scl_hi,
                                xout, xout_stat, xout_chng, pfile);
            }
         
            () = printf("%d ... \n",i+1);
            isc++;
         }
      }
   }
   iret=rcl_mpi_barrier();
   if (proc_rank==0)
   {
      if(sim_count > 0) () = printf("\n");
   }
   set_params(pars);
   
   rcl_mpi_finalize();
   
}
%}}}

%}}}
#else
public define mpi_emcee()
{
   message(" *** error:  MPI EMCEE can not be used as SLMPI is not installed *** ");
}

#endif


%
% Code from src/Mike/isis_fancy_plots.sl
%
% -*- slang -*-

try { require("COPY.sl"); }
catch AnyError:; ;

%  Version 1.4.3

private variable _fancy_plots_version = [1,4,3];
public variable fancy_plots_version=int(sum(_fancy_plots_version*[10000,100,1]));
public variable fancy_plots_version_string=sprintf("%d.%d.%d",
   _fancy_plots_version[0],_fancy_plots_version[1],_fancy_plots_version[2]);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Most recent update: April 2016

% Big block of plotting functions, to allow you to do all sorts of
% XSPEC type plotting, and then some.  Plot parameters can be input as
% either a structure, or as qualifiers, or both.  (For the latter
% case, qualifiers supersede structure choices.)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Functions in This File.  Usage message when function is called
% without arguments.
%
% pg_info          : Writes out useful info about plotting choices.
% pg_color         : Make a few nice colors for pgplot
% sov              : An abbreviation of set_outer_viewport
% apj_size         : An autoset of resize/set_outer_viewport that works
%                    well for a "typical" ApJ one column plot.
% keynote_size     : An autoset of resize/set_outer_viewport that works
%                    well for a "typical" keynote presentation.
% open_print       : A version of open_plot that invokes pg_color first
% close_print      : A version of close_plot that will then display the 
%                    hardcopy via a chosen system utility (e.g., gv)
% set_plot_widths  : Routine to set all the plot line thicknesses
% nice_width       : An autoset of plot widths that works well for 
%                    papers and presentations (especially on Macs)
% set_plot_labels  : Resets all plot labels to their defaults, and can
%                    be used to change the default font.
% new_plot_labels  : Change the plot labels for all plot styles.
% add_plot_unit    : Create new X-/Y-unit combinations for the plots
% fancy_plot_unit  : My variation on ISIS's plot_unit.  Same choices as
%                    in ISIS, but this saves variables (x_unit, y_unit) 
%                    that will be used below.
% data_list        : List version of all_data
% write_plot       : Write the data from the plot functions to ASCII
%                    files
% plot_counts      : Plot background subtracted data as counts/bin,
%                    with choices of three kinds of residuals, or none 
%                    at all. (cstat overides defaults)
% plot_data        : Main plotting routine, to plot background sub-
%                    tracted data in detector counts space.  Counts
%                    per second per x_unit, with choices of three
%                    kinds of residuals, or none at all.
% plot_residuals   : A plot of just the data residuals
% plot_fit_model   : An hplot of just the background subtracted model.
% plot_unfold      : A plot of the unfolded spectra, or powers of 
%                    energy/freq./wavelength times unfolded spectra.
% plotxy           : Simple x-y (o)plots with error bars.
% plot_comps       : Plot individual components of a model
% plot_double      : Use two fancy plot functions at once, while also
%                    plotting individual components of a model

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Private Functions in This File - Used for the Plotting Internals 
%
% pd_set                : Set up variable to store data for ascii dump
% p_reset               : Reset the plot defaults
% multiplotit           : Set up two panes for residual plotting
% vaxis                 : Make a redshift or velocity x-axis
% make_exp_and_back     : Create data/backgrounds exposures & scales
% make_data             : Create plot data (plot_counts/data/residuals)
% hp_loop               : Sets index ranges for hplot so disjoint bits
%                       : of data, model, get plotted separately
% hp_loopit             : Does the actual hplotting for disjoint bits
% xerplot               : Put a bar at 0 or 1 across the residuals plot 
% start_plot            : Kluge to get auto-scaling right for multiple
%                       : data sets
% datplot_err           : Plot data error bars
% trans_strct_and_array : Transfer structure inputs to plot variables
% res_set               : Begin set-up of the residuals plot
% resplot               : Driver for the residual plots
% p_range_label         : Set up the plot ranges and labels
% make_default          : Create default plot parameters
% unpack_list           : Turn a list with internal arrays into a 
%                       : simple list of elements
% trans_strct           : Transfer structure inputs to plot variables
% make_array_a_list     : Turn an array into a list
% set_plot_parms_struct : Ultimately all plot choices are passed via
%                       : a structure, using this subroutine
% make_indices          : Parse the input data indices
% set_plot_parms        : Parse the input to the above routine

% make_flux             : Create the flux corrected data to be plotted
% write_plot_head       : Makes the header for ASCII files containing
%                         data from the plot functions

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

variable fp = stderr;

public define pg_info()
%!%+
%\function{pg_info}
%\synopsis{Print a core dump of some useful pgplot and isis_fancy_plots information.}
%\usage{pg_info; -or- pginfo;}
%\description
%\seealso{Nearly all isis_fancy_plot functions return a use message if invoked without arguments}
%!%-
%!%+
%\function{pginfo}
%\synopsis{Print a core dump of some useful pgplot and isis_fancy_plots information.}
%\usage{pg_info; -or- pginfo;}
%\description
%\seealso{Nearly all isis_fancy_plot functions return a use message if invoked without arguments}
%!%-
{
   () = fprintf(fp,  "%s\n", %{{{
`
 # COLOR   SYMBOL    # COLOR       SYMBOL             # COLOR SYMBOL

 1 default dot       9 yllw green  mdot              17 green filled circle
 2 red     plus     10 green+cyan  fancy square      18 brown filled star
 3 green   cross    11 indigo      diamond           19 pink  big square
 4 blue    circle   12 purple      star           20-27 gold  sized circles
 5 cyan    cross    13 red+magenta solid triangle 28-31       left/right/
 6 magenta square   14 dark grey   hollow cross               up/down arrows
 7 yellow  triangle 15 light grey  star of david 32-127       ASCII symbols
 8 orange  earth    16 black       filled square

 Negative symbol #'s give filled symbols with that many sides
 Colors 17-27 must first be defined with pg_color();

 pointstyle(#)           : Choose one the above symbols.
 connect_points(#)       : -1 = line only / 0 = points only / 1 = both
 linestyle(#)            : (1) full line, (2) dashed, (3) dot-dash,
                           (4) dotted, (5) dash-dot-dot-dot

 line_or_color(#)        : plot(x,y,##), ## = linestyle(##), if # = 0
                                         ## = color, if # != 0 (default)

 set_line_width(#)       : Set the widths of lines in ISIS intrinsic plots
 set_frame_line_width(#) : Set the frame width
 charsize(#)             : Set character sizes in plots
 point_size(#)           : Set datapoint sizes in plots

 _pgsci(#)               : Set color index to #
 _pgtext(x,y,string)     : Write text in plot (use log10 of x/y for log plots)

 xlabel(), ylabel(), title(); with \u, \d, \g = up, down, greek
   \fn, \fr, \fi, \fs = normal, roman, italic, and script fonts
   \A = Angstrom, \. = center dot, \x = multiplication,
   (NOTE: Use \\ if placing within "  ", and \ if placing within `` `` !!!)

 Custom plot functions defined by ISIS fancy plots script (obtain help for
 individual functions, usually by invoking function without any arguments):

   Plotting routines  : plot_counts, plot_data, plot_unfold, plot_residuals,
                        plot_fit_model, plotxy, plot_comps, plot_double
   Output routines    : open_print, close_print (variant of intrinsics,
                        open_plot, close_plot), write_plot (write to ASCII)
   Input routines     : data_list (list version of all_data)
   Unit set routines  : fancy_plot_unit, add_plot_unit
   Color set routine  : pg_color
   Labelling routines : set_plot_labels, new_plot_labels
   Width/size routines: set_plot_widths, nice_width, sov,
                        apj_size, keynote_size
   Flux scale routines: set_power_scale
`); %}}}
}

alias("pg_info","pginfo");

%%%%%%%%%%%%%%%%%%%%%%%%

public define pg_color()
%!%+
%\function{pg_color}
%\synopsis{Set pgplot colors 17, 18, 19, 20 to a green, brown, pink, dark yellow (isis_fancy_plots package)}
%\usage{pg_color; -or- pgcolor;}
%\description
%\seealso{pg_info, pginfo}
%!%-
%!%+
%\function{pgcolor}
%\synopsis{Set pgplot colors 17, 18, 19, 20 to a green, brown, pink, dark yellow (isis_fancy_plots package)}
%\usage{pg_color; -or- pgcolor;}
%\description
%\seealso{pg_info, pginfo}
%!%-
{
   _pgshls(17,255,0.3,0.6);   % Define a decent green color for pgplot
   _pgshls(18,150,0.35,0.4);  % Define a decent brown color for pgplot
   _pgshls(19,120,0.75,0.75); % Define a pink color for pgplot

   % Define a dark yellow color for pgplot
   variable i; _for i (20,27,1){ _pgshls(i,180,0.45,0.65); }
}

alias("pg_color","pgcolor");

%%%%%%%%%%%%%%%%%%%

public define sov()
%!%+
%\function{sov}
%\synopsis{Set pgplot outer viewport (isis_fancy_plots package)}
%\usage{sov(Double_Type, Double_Type, Double_Type, Double_Type);}
%\description
%
% sov(xmin,xmax,ymin,ymax);
%
%   Equivalent to:
%      isis> v=struct{xmin,xmax,ymin,ymax};
%      isis> v.xmin=xmin;
%      isis> v.xmax=xmax;
%      isis> v.ymin=ymin;
%      isis> v.ymax=ymax;
%      isis> set_outer_viewport(v)
%   and:
%      isis> set_outer_viewport(struct{xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax});
%\seealso{set_outer_viewport, apj_size, keynote_size, nice_width, open_print, close_print, pg_color, pg_info}
%!%-
{
   variable v=struct{xmin,xmax,ymin,ymax};
   switch(_NARGS)
   {
    case 4:
      (v.xmin,v.xmax,v.ymin,v.ymax)=();
      set_outer_viewport(v);
      return;
   }
   {
      () = fprintf(fp, "\n%s\n", %{{{
` sov(xmin,xmax,ymin,ymax);

   Equivalent to:
      isis> v=struct{xmin,xmax,ymin,ymax};
      isis> v.xmin=xmin;
      isis> v.xmax=xmax;
      isis> v.ymin=ymin;
      isis> v.ymax=ymax;
      isis> set_outer_viewport(v)
   and:
      isis> set_outer_viewport(struct{xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax});
`                 );            %}}}
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%

public define apj_size()
%!%+
%\function{apj_size}
%\synopsis{Set the pgplot output size to something suitable for ApJ single column (isis_fancy_plots package)}
%\usage{apj_size;}
%\description
%
%   Use as:
%   isis> id = open_print("fig1.ps/vcps"); apj_size; nice_width;
%   isis> plot(x,y);
%   isis> close_print(id,"gv");
%\seealso{sov, keynote_size, nice_width, open_print, close_print, pg_color, pg_info}
%!%-
{
   resize(30,0.9);
   sov(0.13,0.65,0.1,0.65);
   charsize(1.05);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define keynote_size()
%!%+
%\function{keynote_size}
%\synopsis{Set the pgplot output size to something suitable for Keynoe presentations (isis_fancy_plots package)}
%\usage{apj_size;}
%\description
%
%   Use as:
%   isis> id = open_print("fig1.ps/vcps"); keynote_size; nice_width;
%   isis> plot(x,y);
%   isis> close_print(id,"gv");
%\seealso{sov, apj_size, nice_width, open_print, close_print, pg_color, pg_info}
%!%-
{
   resize(30,0.65);
   sov(0.13,0.65,0.1,0.65);
   charsize(1.1);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%

private variable fname = String_Type[0];

public define open_print(a)
%!%+
%\function{open_print}
%\synopsis{Wrapper around open_plot (and pg_color) to allow a system function to call the output file (isis_fancy_plots package)}
%\usage{id = open_print(String_Type);}
%\description
%
%   Use as:
%   isis> id = open_print("fig1.ps/vcps"); keynote_size; nice_width;
%   isis> plot(x,y);
%   isis> close_print(id,"gv");
%\seealso{close_print, sov, open_plot, close_plot, apj_size, keynote_size, nice_width, pg_color, pg_info}
%!%-
{
   variable id;
   variable fname_piece = strchop(a,'/',0);
   variable npiece = length(fname_piece);

   if( npiece < 2 )
   {
      message(" Need to specify a plot device.");
      return;
   } 

   id = open_plot(a);
   if(id>length(fname)-1){loop(id-length(fname)+1){fname=[fname,""];}}

   fname[id] = strjoin(fname_piece[[0:npiece-2]],"/");

   pg_color;
   return id;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define close_print()
%!%+
%\function{close_print}
%\synopsis{Wrapper around close_plot to allow system function to call the output file (isis_fancy_plots package)}
%\usage{close_print(window_id, String_Type);}
%\description
%
%   Use as:
%   isis> id = open_print("fig1.ps/vcps"); keynote_size; nice_width;
%   isis> plot(x,y);
%   isis> close_print(id,"gv");
%\seealso{open_print, sov, open_plot, close_plot, apj_size, keynote_size, nice_width, pg_color, pg_info}
%!%-
{
   variable a,b;
   switch(_NARGS)
   {
    case 1:
      a = ();
      close_plot(a);
      return;
   }
   {
    case 2:
      (a,b) = ();
      close_plot(a);
      () = system(b+" "+fname[a]+" &");
      return;
   }
   {
      message("Incorrect arguments.");
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Structure variable for saving plot data to files

static variable pd=struct{type, index, file, xaxis, yaxis, raxis, dlo,
 dhi, dval, derr, model_on, mlo, mhi, mval, con_mod, res, res_m, res_p, 
weight, exp, instrument, grating, order, part};

% Standard plot options accesible through the plot functions themselves

private variable parm_s = 
                    struct{ dcol, decol, mcol, ccol, cstyle, rcol, recol, dsym, rsym,
                            yrange, xrange, res, power, oplt, bkg, 
                            zshift, vzero, zaxis,
                            xlabel, ylabel, sum_exp, con_mod, gap, scale, no_reset};

% Parameters that should be lists on input

variable list_params=["dcol","decol", "dsym","mcol","ccol","cstyle","rcol","recol", 
                      "rsym","scale","yrange","xrange","bkg","zshift","gap"];

% All the parameters

private variable parms=[list_params,"res","oplt","vzero","zaxis","power",
                        "xlabel","ylabel", "sum_exp", "con_mod", "no_reset"];

private variable dcol_df, mcol_df, power_df, style_df, gap_df, sum_exp_df;

% Note that the following variables, including the list of data set
% indices, essentially get passed around like common block variables

private variable indx, dcol, decol, mcol, rcol, recol, dsym, rsym, 
                yrng, xrng, res, power, oplt, bkg,
                zshift, vzero, zaxis,
                xlabl, ylabl,
                pd_mean_not_sum=0, use_con_flux, rescale, no_reset, gap;

dcol_df = 4;         % Default data color is blue
style_df = 4;        % Default symbol is a circle
mcol_df = 2;         % Default model color red
power_df = 1;        % Default unfolded spectrum is Photons/cm^2/sec/Unit
xlabl=NULL;          % Default is to use the standard X plot labels
ylabl=NULL;          % Default is to use the standard Y plot labels
sum_exp_df = 1;      % Default is to sum exposure times when combining data for plot_data
use_con_flux = 1;    % Default is to smear the model through the response for plot_unfold
no_reset = 0;        % Default is to "reset" plots after plotting, so
		     % the subsequent plot takes place in the next
		     % panel (multi-plots), or redraws the window as a
		     % whole (single plots)
gap_df = 1;          % plots have gaps where the data does

public variable popt = @parm_s;  % Create a public structure variable for choices

% Plot options only accessible through set_plot_widths

private variable d_width, de_width, m_width, r_width, re_width, 
                 ebar_x, ebar_y, data_err;
d_width =1;    % Default widths on line thicknesses.
m_width =1; 
r_width =1;
de_width=1;
re_width=1; 
ebar_x = 0;    % Default is no caps on error bars
ebar_y = 0;
data_err=0;    % Default is no X error bars

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Replaces old way of setting plot widths via global variables

public define set_plot_widths() 
%!%+
%\function{set_plot_widths}
%\synopsis{Sets plot widths for isis_fancy_plots package}
%\usage{set_plot_widths([;qualifiers]);}
%\qualifiers{
%\qualifier{m_width}{Model line width}
%\qualifier{d_width}{Data line width}
%\qualifier{de_width}{Data error bar line width}
%\qualifier{r_width}{Residual line width}
%\qualifier{re_width}{Residual error bar line width}
%\qualifier{ebar_x}{X error bar term cap length}
%\qualifier{ebar_y}{Y error bar term cap length}
%\qualifier{data_err}{!=0 X error bars plotted}
%}
%\description
%
% set_plot_widths(; d_width=#, de_width=#, r_width=#, re_width=#, m_width=#,
%                   ebar_x=#, ebar_y=#, data_err=#);
%
% Sets line widths on data, residuals, error bars, and models, sets the
% length of x/y error bar term caps (ebar_x, ebar_y), and toggles the
% X error bar plotting. Values are retained until explicitly overwritten.
%\seealso{nice_width, pg_info}
%!%-
{
   variable quals=["d_width","de_width","r_width","re_width","m_width",
                   "ebar_x","ebar_y","data_err"];

   if(sum(array_map(Integer_Type,&qualifier_exists,quals;;__qualifiers))==0)
   {
      () = fprintf (fp, "%s\n", %{{{
`
 set_plot_widths(; d_width=#, de_width=#, r_width=#, re_width=#, m_width=#,
                   ebar_x=#, ebar_y=#, data_err=#);

 Sets line widths on data, residuals, error bars, and models, sets the
 length of x/y error bar term caps (ebar_x, ebar_y), and toggles the
 X error bar plotting. Values are retained until explicitly overwritten.

 Current values:
`                   );           %}}}
      ()=fprintf(fp, "  Data                 : %3i\n", d_width);
      ()=fprintf(fp, "  Data Errors          : %3i\n", de_width);
      ()=fprintf(fp, "  Residuals            : %3i\n", r_width);
      ()=fprintf(fp, "  Residual Errors      : %3i\n", re_width);
      ()=fprintf(fp, "  Model                : %3i\n", m_width);
      ()=fprintf(fp, "  X error bar term caps: %3i\n", ebar_x);
      ()=fprintf(fp, "  Y error bar term caps: %3i\n", ebar_y);
      ()=fprintf(fp, "  X error bars on?     : %3i\n\n", data_err);
      return;
   }
   m_width  = int(abs(qualifier("m_width", m_width )))[0];
   d_width  = int(abs(qualifier("d_width", d_width )))[0];
   de_width = int(abs(qualifier("de_width",de_width)))[0];
   r_width  = int(abs(qualifier("r_width", r_width )))[0];
   re_width = int(abs(qualifier("re_width",re_width)))[0];
   ebar_x   = int(abs(qualifier("ebar_x",  ebar_x  )))[0];
   ebar_y   = int(abs(qualifier("ebar_y",  ebar_y  )))[0];
   data_err = int(abs(qualifier("data_err",data_err)))[0];
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%dcol_df = 4;         % Default data color is blue
%style_df = 4;        % Default symbol is a circle
%mcol_df = 2;         % Default model color red
%power_df = 1;        % Default unfolded spectrum is Photons/cm^2/sec/Unit
%pd_mean_not_sum = 0; % Default is to sum exposure times when combining data for plot_data
%use_con_flux = 1;    % Default is to smear the model through the response for plot_unfold
%gap_df = 1;          % plots have gaps where the data does

public define reset_plot_defaults() 
%!%+
%\function{reset_plot_defaults}
%\synopsis{Changes some of the defaults on the isis_fancy_plots package}
%\usage{reset_plot_defaults(;dcol=Integer_Type, ...);}
%\qualifiers{
%\qualifier{dcol}{Data color value}
%\qualifier{dsym}{Data symbol value}
%\qualifier{mcol}{Model color value}
%\qualifier{sum_exp}{!=0, Sum the exposures when combining data}
%\qualifier{use_con_flux}{!=0, the unfolded model includes response smearing}
%\qualifier{gap}{==0, plot models across data gaps.}
%}
%\description
%
% reset_plot_defaults(; dcol=#, dsym=#, mcol=#, sum_exp=#,
%                   use_con_flux=#, gap=#);
%
% Resets some of the plot defaults to a user's specifications. (See
% help messages for individual plotting functions, and pg_info, to
% understand these settings.)  Use with no arguments to see current
% values.
%\seealso{pg_info}
%!%-
{
   variable quals=["dcol","dsym","mcol","power","sum_exp",
                   "use_con_flux","gap"];

   if(sum(array_map(Integer_Type,&qualifier_exists,quals;;__qualifiers))==0)
   {
      () = fprintf (fp, "%s\n", %{{{
`
 reset_plot_defaults(; dcol=#, dsym=#, mcol=#, sum_exp=#,
                   use_con_flux=#, gap=#);

 Resets some of the plot defaults to a user's specifications. (See
 help messages for individual plotting functions, and pg_info, to
 understand these settings.)

 Current values:
`                   );           %}}}
      ()=fprintf(fp, "  dcol          : %3i\n", dcol_df);
      ()=fprintf(fp, "  dsym          : %3i\n", style_df);
      ()=fprintf(fp, "  mcol          : %3i\n", mcol_df);
      ()=fprintf(fp, "  sum_exp       : %3i\n", sum_exp_df);
      ()=fprintf(fp, "  use_con_flux  : %3i\n", use_con_flux);
      ()=fprintf(fp, "  gap           : %3i\n\n", gap_df);
      return;
   }
   dcol_df = int(abs(qualifier("dcol", dcol_df )))[0];
   style_df = int(abs(qualifier("dsym", style_df )))[0];
   mcol_df = int(abs(qualifier("mcol",mcol_df)))[0];
   sum_exp_df = int(abs(qualifier("sum_exp",sum_exp_df)))[0];
   use_con_flux = int(abs(qualifier("use_con_flux",  use_con_flux  )))[0];
   gap_df = int(abs(qualifier("gap",  gap_df  )))[0];
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define nice_width()
%!%+
%\function{nice_width}
%\synopsis{Sets reasonable defaults for plot line widths.}
%\usage{nice_width;}
%\description
%
%  Equivalent to:
%    isis> set_plot_widths(;d_width=2, de_width=2, r_width=2, re_width=2, m_width=2);
%\seealso{set_plot_widths, apj_size, keynote_size, open_print, close_print}
%!%-
{
   variable dw=2, ew=2, fw=2, mw=2;
   set_frame_line_width(fw);
   set_line_width(dw);
   set_plot_widths(;d_width=dw, de_width=ew, r_width=dw, re_width=ew, m_width=mw);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private variable plopt=NULL;

private variable refplt = [&plot, &oplot];
private variable hrefplt = [&hplot, &ohplot];
private variable slangref;

% Beginning of the definition of the standard units we'll use.

private variable y_unit = "photons", x_unit="kev";
private variable zlo, zhi;

private variable ssk, ssw, ssh, vlbl, redlbl, clbl, mllbl;
private variable lk = ["ev","kev","mev","gev","tev",
                      "hz","khz","mhz","ghz",
                      "angstrom","a","nm","um","micron","mm","cm","m"];

private variable ak = Assoc_Type[String_Type], 
                 sk = Assoc_Type[String_Type];
private variable y_scl, y_scl_norm = Assoc_Type[Float_Type]; % For storing Y-axis scalings

unit_add("micron",0,1.e4);  % Same thing as "um"
unit_add("psd",1,1.);       % SITAR defines power spectra as keV based units
                            % since people have in the past used XSPEC to fit them

ak["ev"] = "eV"; ak["kev"] = "keV"; ak["mev"] = "MeV"; 
ak["gev"] = "GeV"; ak["tev"] = "TeV";
ak["hz"] = "Hz"; ak["khz"] = "kHz"; ak["mhz"] = "MHz"; ak["ghz"] = "GHz";
ak["angstrom"] = `\A`; ak["a"] = `\A`; ak["nm"] = "nm";
ak["um"] = `\gmm`; ak["micron"] = `\gmm`; 
ak["mm"] = "mm"; ak["cm"] = "cm"; ak["m"] = "m";
ak["psd"]= "Hz";

private variable xlbl = Assoc_Type[String_Type];
private variable ylbl = Assoc_Type[Array_Type];
private variable rlbl;
private variable pg_font;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Define labels for each combination of possible x_unit & y_unit.

public define set_plot_labels()
%!%+
%\function{set_plot_labels}
%\synopsis{Restore default plot labels of isis_fancy_plots package}
%\usage{set_plot_labels(;pg_font="\\\\fr") -or- set_plot_labels(;pg_font=``\\\\fr``)}
%\qualifiers{
%\qualifier{pg_font}{="\\\\fn", "\\\\fr", "\\\\fs", or "\\\\fi"}
%}
%\description
%\seealso{new_plot_labels, fancy_plot_unit, add_plot_unit}
%!%-

    
{
   pg_font = qualifier("pg_font",`\fr`);
   pg_font = qualifier("pgfont",pg_font);
   
   ssk = pg_font+"Energy";
   ssw = pg_font+"Wavelength";
   ssh = pg_font+"Frequency";

   sk["ev"] = ssk; sk["kev"] = ssk; sk["mev"] = ssk; 
     sk["gev"] = ssk; sk["tev"] = ssk;
   sk["hz"] = ssh; sk["khz"] = ssh; sk["mhz"] = ssh; sk["ghz"] = ssh;
   sk["angstrom"] = ssw; sk["a"] = ssw; sk["nm"] = ssw;
     sk["um"] = ssw; sk["micron"] = ssw; sk["mm"] = ssw; sk["cm"] = ssw; sk["m"] = ssw;
   sk["psd"]= ssh; 

   variable nu_F_nu_lbls
    = [   `F\d\gn\u/\gn\u-2\d`,
          `F\d\gn\u/\gn`,
          `F\d\gn\u`,
       `\gnF\d\gn\u`
      ]+"  (";

   variable l_F_l_lbls
    = [        `F\d\gl\u`,
       `\gl`+  `F\d\gl\u`,
       `\gl\u2\dF\d\gl\u`,
       `\gl\u3\dF\d\gl\u`
      ]+"  (";

   variable ergs=`ergs cm\u-2\d s\u-1\d`;
   variable watts=`Watts cm\u-2\d`;

   variable lbl;
   foreach lbl (lk)
   {
      xlbl[lbl] = sk[lbl]+" ("+ak[lbl]+")";

      variable plot_data_lbl = `Counts s\u-1\d `+ak[lbl]+`\u-1\d`;
      variable plot_counts_lbls = 
         ["",          "", ak[lbl]+" ", ak[lbl]+`\u2\d `]
                          +"Counts/bin"+
         ["/"+ak[lbl], "",          "",               ""];

      variable nu_F_nu_suffix = 
         [" "+ak[lbl]+`\u-3\d)`, " "+ak[lbl]+`\u-2\d)`, " "+ak[lbl]+`\u-1\d)`, ")"];

      variable l_F_l_prefix =
         ["", "", ak[lbl]+" ", ak[lbl]+`\u2\d `];

      variable l_F_l_suffix = 
         [" "+ak[lbl]+`\u-1\d`, "", "", ""]+")";

      ylbl[lbl+"photons"] = pg_font+
         [plot_data_lbl,
          [ak[lbl]+`\u-1\d `,"",ak[lbl]+" ",ak[lbl]+`\u2\d `]
             +"Photons cm\\u-2\\d s\\u-1\\d "+ak[lbl]+"\\u-1\\d",
          plot_counts_lbls                                       ];

      y_scl_norm[lbl+"photons"] = 1.;  % Photons is photons

      ylbl[lbl+"mjy"] = pg_font+
         [plot_data_lbl,
          "I\\d\\gn\\u"+
              ["/"+sk[lbl]+`\u2\d  (mJy/`+ak[lbl]+`\u2\d)`,
                         "/"+sk[lbl]+"  (mJy/"+ak[lbl]+")",
                                                 "  (mJy)",
                       ` \x `+sk[lbl]+`  (mJy\.`+ak[lbl]+")"],
          plot_counts_lbls                                       ];

      y_scl_norm[lbl+"mjy"] = Const_h*1.e26; % The definition of mJy

      if(unit_info(lbl).is_energy==1)
      {
         ylbl[lbl+"ergs"] = pg_font+
            [plot_data_lbl,
             nu_F_nu_lbls+ergs+nu_F_nu_suffix,
             plot_counts_lbls                                    ];

         % Converts plot_unfold(...;power=3) to ergs/cm^2/s

         y_scl_norm[lbl+"ergs"] = Const_eV*1.e3*unit_info(lbl).scale; 

         ylbl[lbl+"watts"] =  pg_font+
            [plot_data_lbl,
             nu_F_nu_lbls+watts+nu_F_nu_suffix,
             plot_counts_lbls                           ];

         % Converts plot_unfold(...;power=3) to Watts/cm^2

         y_scl_norm[lbl+"watts"] = Const_eV*unit_info(lbl).scale; 
      }
      else
      {
         ylbl[lbl+"ergs"] = pg_font+
            [plot_data_lbl,
             l_F_l_lbls+l_F_l_prefix+ergs+l_F_l_suffix,
             plot_counts_lbls                                             ];

         % Converts plot_unfold(...;power=1) to ergs/cm^2/s

         y_scl_norm[lbl+"ergs"] = Const_h*Const_c*1.e8/unit_info(lbl).scale; 

         ylbl[lbl+"watts"] = pg_font+
            [plot_data_lbl,
             l_F_l_lbls+l_F_l_prefix+watts+l_F_l_suffix,
             plot_counts_lbls                                             ];

         % Converts plot_unfold(...;power=1) to Watts/cm^2

         y_scl_norm[lbl+"watts"] = Const_h*Const_c*1.e5/unit_info(lbl).scale; 
      }
   }

   xlbl["psd"] = sk["psd"]+"("+ak["psd"]+")";

   ylbl["psdpsd_leahy"] = pg_font+["PSD Plot","PSD Plot","PSD Plot","PSD Plot","PSD Plot",
                                   ` Power (Leahy)/f`,
                                   ` Power (Leahy)`,
                                   ` f \x Power (Leahy)`,
                                   ` f\u2\d \x Power (Leahy)`];

   y_scl_norm["psdpsd_leahy"] = 1;  % Not really used in PSD plots

   ylbl["psdpsd_rms"] = pg_font+["PSD Plot","PSD Plot","PSD Plot","PSD Plot","PSD Plot",
                                 ` Power/f (RMS\u2\d/Hz\u2\d)`,
                                 ` Power (RMS\u2\d/Hz)`,
                                 ` f \x Power (RMS\u2\d)`,
                                 ` f\u2\d \x Power (RMS\u2\d \. Hz)`];

   y_scl_norm["psdpsd_rms"] = 1;  % Not really used in PSD plots

   rlbl = [pg_font+`\gx`,pg_font+`\gx\u2`,pg_font+"Ratio"];
   rlbl = [rlbl,rlbl];

   vlbl = pg_font+` Velocity  (km s\u-1\d)`;
   redlbl = pg_font+" Redshift  (z)";
   clbl = pg_font+[`\gDC/|\gDC|\u1/2`,`\gDC`];
   mllbl = pg_font+[`\gDML/|\gDML|\u1/2`,`\gDML`];
}

%%%%%%%%%%%%%%%%%%

set_plot_labels();

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Allow the user to ovewrite default labels, and define labels for new
% plot unit types.

public define new_plot_labels()
%!%+
%\function{new_plot_labels}
%\synopsis{Create new plot labels for routines in isis_fancy_plots package.}
%\usage{new_plot_labels(String_Type [, String_Type]);}
%\qualifiers{
%\qualifier{xlabel}{New X-axis label}
%\qualifier{ylabel}{String array with new Y-axis labels for data}
%\qualifier{rlabel}{String array with new Y-axis labels for residuals}
%\qualifier{clabel}{String array with new Y-axis residual labels for Cash statistics}
%\qualifier{mllabel}{String array with new Y-axis residual labels for Maximum Likelihood statistics}
%\qualifier{vlabel}{String with new Doppler velocity X-axis label}
%\qualifier{zlabel}{String with new redshift X-axis label}
%\qualifier{pg_font}{PGPLOT font type (default is \\\\fr)}
%}
%\description
%
% new_plot_labels(x_unit [,y_unit];xlabel=string,ylabel=[string],...,
%                                   pg_font=string);
%
% Changes the default labels for plot_counts, plot_data, plot_unfold
% for the different units set by fancy_plot_unit. pg_font string will
% be *prepended* to all inputs. Use any valid pair of units to set residual,
% Cash statistic, or velocity/redshift axis labels.
%
% Use: 
%
%    set_plot_labels(;pg_font="\\\\fr") -or- set_plot_labels(;pg_font=``\\\\fr``)
%
% to restore defaults.
%
% Inputs:
%
%   x_unit : angstrom,a,nm,um,micron,cm,mm,m, ev,kev,mev,gev,tev,
%            hz,khz,mhz,ghz, psd
%   y_unit : photons (=default), ergs, watts, mjy, psd_leahy, or psd_rms
%   xlabel : String with new X-axis label
%   ylabel : String *array* (up to 9 elements) with new Y-axis labels.
%            Order of the array *must* be: 
%             [plot_data, plot_unfold(power=0->3),plot_counts(power=0->3)]
%   rlabel : String *array* (up to 3 elements) with new Y-axis labels for
%            residuals.  Order of the the array *must* be: 
%             [chi, chi^2, ratio]
%   clabel : String Array with new Cash statistic labels [res=1 or 4, 2 or 5,
%            which yield +/-sqrt(|Delta C|), Delta C, respectively]
%   mllabel: String Array with new Maximum Likelihood statistic labels [res=1 or 4, 
%            2 or 5, which yield +/-sqrt(|Delta ML|), Delta ML, respectively]
%   vlabel : String with new Doppler velocity X-axis label
%   zlabel : String with new Redshift X-axis label
%   pg_font: "\\\\fn", "\\\\fr", "\\\\fi", "\\\\fs" = normal, roman, italic,
%            or script fonts will be used on the labels (Default is \\\\fr.)
%\seealso{set_plot_labels, fancy_plot_unit, add_plot_unit}
%!%-
{
   variable str,xunit,yunit,lbl,lbly,xlbls,ylbls,ly,rlbls,lr,nvlbl,nredlbl,cashlbl;

   switch(_NARGS)
   {
      case 1:
      xunit = ();
      yunit = "photons";
   }
   {
      case 2:
      (xunit,yunit) = ();
   }
   {
      () = fprintf(fp, "%s\n", %%%{{{
`
 new_plot_labels(x_unit [,y_unit];xlabel=string,ylabel=[string],...,
                                   pg_font=string);

 Changes the default labels for plot_counts, plot_data, plot_unfold
 for the different units set by fancy_plot_unit. pg_font string will
 be *prepended* to all inputs. Use any valid pair of units to set residual,
 Cash statistic, or velocity/redshift axis labels.
 Use: 

    set_plot_labels(;pg_font="\\fr") -or- set_plot_labels(;pg_font=``\fr``)

 to restore defaults.

 Inputs:

   x_unit : angstrom,a,nm,um,micron,cm,mm,m, ev,kev,mev,gev,tev,
            hz,khz,mhz,ghz, psd
   y_unit : photons (=default), ergs, watts, mjy, psd_leahy, or psd_rms
   xlabel : String with new X-axis label
   ylabel : String *array* (up to 9 elements) with new Y-axis labels.
            Order of the array *must* be: 
             [plot_data, plot_unfold(power=0->3),plot_counts(power=0->3)]
   rlabel : String *array* (up to 3 elements) with new Y-axis labels for
            residuals.  Order of the the array *must* be: 
             [chi, chi^2, ratio]
   clabel : String Array with new Cash statistic labels [res=1 or 4, 2 or 5,
            which yield +/-sqrt(|Delta C|), Delta C, respectively]
   mllabel: String Array with new Maximum Likelihood statistic labels [res=1 or 4, 
            2 or 5, which yield +/-sqrt(|Delta ML|), Delta ML, respectively]
   vlabel : String with new Doppler velocity X-axis label
   zlabel : String with new Redshift X-axis label
   pg_font: "\\fn", "\\fr", "\\fi", "\\fs" = normal, roman, italic,
            or script fonts will be used on the labels (Default is \\fr.)
`                       );        %%%}}}
      return;
   }

   lbl  = strlow(xunit);
   lbly = strlow(xunit+yunit);
   xlbls   = qualifier("xlabel", NULL);
   ylbls   = qualifier("ylabel", NULL);
   rlbls   = qualifier("rlabel", NULL);
   nvlbl   = qualifier("vlabel", NULL);
   nredlbl = qualifier("zlabel", NULL);
   cashlbl = qualifier("clabel", NULL);
   mllbl   = qualifier("mllabel", NULL);
   pg_font = qualifier("pg_font",`\fr`);
   pg_font = qualifier("pgfont", pg_font);

   ifnot(assoc_key_exists(ylbl,lbly))
   {
      () = fprintf(fp, "\n%s\n", %{{{
`  Unrecognized combination of X- and Y-units. Define them with: 

   add_plot_unit(xunit,yunit;xscale=val,yscale=val);
`                  );            %}}}
      return;
   }

   if(xlbls != NULL)
   {
      xlbls=string([xlbls][0]); 
      xlbl[lbl] = pg_font+xlbls;
   }

   if(ylbls != NULL)
   {
      ylbls = [ylbls];
      ly = length(ylbls);
      if(ly>9){ ly = 9; }   

      ylbl[lbly][[0:ly-1]] = pg_font+ylbls[[0:ly-1]];
   }

   if(rlbls != NULL)
   {
      rlbls = [rlbls];
      lr = length(rlbls);
      if(lr>3){ lr = 3; }   

      rlbl[[0:lr-1]] = pg_font+rlbls[[0:lr-1]];
      rlbl[[4:4+lr-1]] = pg_font+rlbls[[0:lr-1]];
   }

   if(cashlbl != NULL)
   {
      cashlbl=string([cashlbl]); 
      clbl = pg_font+[clbl,clbl];
   }

   if(mllbl != NULL)
   {
      mllbl=string([mllbl]); 
      mllbl = pg_font+[mllbl,mllbl];
   }

   if(nvlbl != NULL)
   {
      nvlbl=string([nvlbl][0]); 
      vlbl = pg_font+nvlbl;
   }

   if(nredlbl != NULL)
   {
      nredlbl=string([nredlbl][0]); 
      redlbl = pg_font+nredlbl;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define add_plot_unit()
%!%+
%\function{add_plot_unit}
%\synopsis{Create a new X-unit,Y-unit pair for isis_fancy_plots package}
%\usage{add_plot_unit(String_Type, String_Type [; ]);}
%\qualifiers{
%\qualifier{SEE BELOW}{}
%}
%\description
%
% add_plot_unit(xunit, yunit; xscale=val, yscale=val, is_energy=val,
%                             xlabel=str, ylabel=str, pgfont=str);
%
% Create a new X-unit and Y-unit pair for plot_counts, plot_data, plot_unfold,
% and plot_fit_model.  (yunit and yscale only affect plot_unfold; plot_counts
% and plot_data are only sensitive to the choice of xunit.) Set is_energy=1 if
% the xunit scales as keV (otherwise it defaults to scaling as Angstroms). xscale
% is the scaling of keV/xunit or A/xunit.  yscale scales *all* of the y-axes in
% the plot_unfold plots, so it must be set appropriately to achieve a desired
% effect.  xunit and yunit will be treated as lower case regardless of input.
% Note that existing X-units will not be overwritten; their already defined
% values will be used regardless of input.  A set of default axis labels will
% be produced, with "ylabel" substituting for:
%
%       "Photons cm^{-2} s^{-1}"   (the default)
%
% However, these labels can be rewritten using the new_plot_labels command.
%
% To get the scalings correct, remember that plot_unfold(...;power=2) is always
% photons/s/cm^2 in the absence of scalings, power=0 is proportional to flux
% for wavelength based X-units, while power=3 is proportional to flux for energy
% based units.  For these latter two cases, the yscale needs to account for
% divisions/multiplications of the scaled X-unit.
%
% Examples:
%
% To produce mJy vs. THz plots, use:
%
%    add_plot_unit("thz", "mjy"; xscale=Const_eV/Const_h*1.e12,
%                  yscale=Const_h*1.e26, xlabel="THz", ylabel="mJy/THz");
%
% Const_eV/Const_h*1.e12 ~ 4.14e-3 is keV per Terra-Hz
% plot_unfold(...;power=2) produces photons/s/cm^2 =>
%    (photons*xunit)/s/cm^2/xunit  (xunit scaling cancels out)
%
% We wish to convert this to mJy:
%
%    (photons*xunit/s/cm^2/xunit) =
%    [photons*(1.e-26 ergs)*Hz/(1.e-26 ergs)]/s/cm^2/Hz =
%    Hz/(1.e-26 ergs)*mJy
%
% hence multiplying by yscale=(1.e-26 ergs)/Hz = 1.e26*Const_h
% produces the desired scaling for plot_unfold.
%
% To produce BTU/Acre/s vs. keV plots, use:
%
% plot_unfold(...;power=3) for X-unit=keV produces keV/s/cm^2,
% which we wish to convert to BTU/Acre/s, hence xscale=1 and:
%
%    erg_p_btu = 1.055056e10                    % ISO ergs/BTU
%    cmsq_p_a = 4.0468564224e7                  % cm^2 per international acre
%    yscale = cmsq_p_a*Const_eV*1.e3/erg_p_btu  % cm^2/Acre*BTU/keV
%
% To achieve this same Y-unit vs. other energy based X-units,
% multiply yscale by keV/xunit, i.e., the value of xscale.
%\seealso{fancy_plot_unit, set_plot_labels, new_plot_labels}
%!%-
{
   variable xunit, yunit, xlbls, ylbls, xscl, is_en, pgfont;
   switch(_NARGS)
   {
      case 2:
      (xunit,yunit) = ();
      xunit = strlow(xunit);
      yunit = strlow(yunit);
      if(assoc_key_exists(ylbl,xunit+yunit))
      {
         () = fprintf(fp, "\n X-unit/Y-unit pair already exists\n");
         return;
      }
      xlbls = qualifier("xlabel","X");
      ylbls = qualifier("ylabel",`Photons cm\u-2\d s\u-1\d `);
      pgfont = qualifier("pgfont",`\fr`);
      is_en = int(qualifier("is_energy",0));
      if(is_en != 1 && is_en != 0){ is_en = 0; }
      xscl = qualifier("xscale",1.);
      y_scl_norm[xunit+yunit] = qualifier("yscale",1.);
      if(unit_exists(xunit))
      {
         xlbls = unit_info.name;
      }
      else
      {
         unit_add(xunit,is_en,xscl);
         xlbl[xunit] = pg_font + xlbls;
      }

      ylbl[xunit+yunit]
       = [pg_font
         +[`Counts s\u-1\d ` + xlbls + `\u-1\d`,
            xlbls + `\u-1\d `+ ylbls + xlbls + `\u-1\d`,
                               ylbls,
            xlbls +            ylbls,
            xlbls + `\u2\d ` + ylbls + xlbls + `\u-1\d`,  %%% <- Mh: The last `\u-1\d` is wrong, I guess.
                         "Counts/bin/"+xlbls,
                         "Counts/bin",
            xlbls +      " Counts/bin",
            xlbls + `\u2\d Counts/bin`
          ],
         ylbls];
   }
   {
      ()=fprintf(fp, "%s\n", %{{{
`
 add_plot_unit(xunit, yunit; xscale=val, yscale=val, is_energy=val,
                             xlabel=str, ylabel=str, pgfont=str);

 Create a new X-unit and Y-unit pair for plot_counts, plot_data, plot_unfold,
 and plot_fit_model.  (yunit and yscale only affect plot_unfold; plot_counts
 and plot_data are only sensitive to the choice of xunit.) Set is_energy=1 if
 the xunit scales as keV (otherwise it defaults to scaling as Angstroms). xscale
 is the scaling of keV/xunit or A/xunit.  yscale scales *all* of the y-axes in
 the plot_unfold plots, so it must be set appropriately to achieve a desired
 effect.  xunit and yunit will be treated as lower case regardless of input.
 Note that existing X-units will not be overwritten; their already defined
 values will be used regardless of input.  A set of default axis labels will
 be produced, with "ylabel" substituting for:

       "Photons cm\u-2\d s\u-1\d"   (the default)

 However, these labels can be rewritten using the new_plot_labels command.

 To get the scalings correct, remember that plot_unfold(...;power=2) is always
 photons/s/cm^2 in the absence of scalings, power=0 is proportional to flux
 for wavelength based X-units, while power=3 is proportional to flux for energy
 based units.  For these latter two cases, the yscale needs to account for
 divisions/multiplications of the scaled X-unit.

 Examples:

 To produce mJy vs. THz plots, use:

    add_plot_unit("thz", "mjy"; xscale=Const_eV/Const_h*1.e12,
                  yscale=Const_h*1.e26, xlabel="THz", ylabel="mJy/THz");

 Const_eV/Const_h*1.e12 ~ 4.14e-3 is keV per Terra-Hz
 plot_unfold(...;power=2) produces photons/s/cm^2 =>
    (photons*xunit)/s/cm^2/xunit  (xunit scaling cancels out)

 We wish to convert this to mJy:

    (photons*xunit/s/cm^2/xunit) =
    [photons*(1.e-26 ergs)*Hz/(1.e-26 ergs)]/s/cm^2/Hz =
    Hz/(1.e-26 ergs)*mJy

 hence multiplying by yscale=(1.e-26 ergs)/Hz = 1.e26*Const_h
 produces the desired scaling for plot_unfold.

 To produce BTU/Acre/s vs. keV plots, use:

 plot_unfold(...;power=3) for X-unit=keV produces keV/s/cm^2,
 which we wish to convert to BTU/Acre/s, hence xscale=1 and:

    erg_p_btu = 1.055056e10                    % ISO ergs/BTU
    cmsq_p_a = 4.0468564224e7                  % cm^2 per international acre
    yscale = cmsq_p_a*Const_eV*1.e3/erg_p_btu  % cm^2/Acre*BTU/keV

 To achieve this same Y-unit vs. other energy based X-units,
 multiply yscale by keV/xunit, i.e., the value of xscale.
`); %}}}
   }
   return;
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define fancy_plot_unit( )
%!%+
%\function{fancy_plot_unit}
%\synopsis{Change the x-axis, and possibly the y-axis units, in the isis_fancy_plots package.}
%\usage{fancy_plot_unit( String_Type [, String_Type]);}
%\description
%
% fancy_plot_unit(xunit [,yunit]);
%
% Change the X-axis plot units to "xunit" (as for the ISIS command
% plot_unit; and change the Y-axis unit to "yunit" (default yunit=
% "photons").  These will be used for the functions: plot_counts,
% plot_data, plot_unfold, plot_fit_model. Unit names are case insensitive.
%
% Available X-units:
%
%      eV, keV, MeV, GeV, TeV,
%      Angstrom, A, nm, um, mm, cm, m,
%      Hz, kHz, MHz, GHz,
%      psd   (used for plotting power spectra from SITAR)
%
% Available Y-units:
%
%      photons (default), mJy, ergs, watts, psd_leahy, psd_rms
%
% Units added via add_plot_unit are also supported.
%
% **NOTE**: Y-units photons/mJy/ergs/watts affect only plot_unfold, while
% psd_* are for plot_counts, but will also affect plot_data/plot_unfold.
%
% Fundamentally, power=1 is proportional to photons/cm^2/s/xunit
% (plot_unfold) or Counts/bin (plot_counts), with higher (lower) powers
% multiplying (dividing) by xunit. plot_data is always Counts/sec/xunit
% ("power" has no effect).
%
% mJy: y-unit for plot_unfold/power=2 is mJy.
%
% ergs: y-unit for plot_unfold/power=3 (xunit=keV, etc.) or power=1
%       (xunit=A, etc.) is ergs/cm^2/sec.
%
% watts: Similar behavior to the ergs unit, but yielding Watts/cm^2.
%
% psd_leahy/psd_rms are for use with SITAR timing routines, and plot
% Power Spectra in Leahy or (RMS/Hz)^2 units vs. Hz, using plot_counts.
%\seealso{plot_unit, add_plot_unit, set_plot_labels, new_plot_labels}
%!%-
{
 variable x_unit_arg, y_unit_arg;
 switch(_NARGS)
 {
  case 1:
    x_unit_arg = ();
    y_unit_arg = "photons";
 }
 {
  case 2:
    (x_unit_arg,y_unit_arg) = ();
 }
 {    ()=fprintf(fp, "%s\n", %{{{
`
 fancy_plot_unit(xunit [,yunit]);

 Change the X-axis plot units to "xunit" (as for the ISIS command
 plot_unit; and change the Y-axis unit to "yunit" (default yunit=
 "photons").  These will be used for the functions: plot_counts,
 plot_data, plot_unfold, plot_fit_model. Unit names are case insensitive.

 Available X-units:

      eV, keV, MeV, GeV, TeV,
      Angstrom, A, nm, um, mm, cm, m,
      Hz, kHz, MHz, GHz,
      psd   (used for plotting power spectra from SITAR)

 Available Y-units:

      photons (default), mJy, ergs, watts, psd_leahy, psd_rms

 Units added via add_plot_unit are also supported.

 **NOTE**: Y-units photons/mJy/ergs/watts affect only plot_unfold, while
 psd_* are for plot_counts, but will also affect plot_data/plot_unfold.

 Fundamentally, power=1 is proportional to photons/cm^2/s/xunit
 (plot_unfold) or Counts/bin (plot_counts), with higher (lower) powers
 multiplying (dividing) by xunit. plot_data is always Counts/sec/xunit
 ("power" has no effect).

 mJy: y-unit for plot_unfold/power=2 is mJy.

 ergs: y-unit for plot_unfold/power=3 (xunit=keV, etc.) or power=1
       (xunit=A, etc.) is ergs/cm^2/sec.

 watts: Similar behavior to the ergs unit, but yielding Watts/cm^2.

 psd_leahy/psd_rms are for use with SITAR timing routines, and plot
 Power Spectra in Leahy or (RMS/Hz)^2 units vs. Hz, using plot_counts.
`); %}}}
    return;
 }

 x_unit_arg=strlow(x_unit_arg);
 y_unit_arg=strlow(y_unit_arg);

 ifnot(assoc_key_exists(ylbl,x_unit_arg+y_unit_arg))
 {
    ()=fprintf(fp, "%s\n", %{{{
`
 X-unit, Y-unit pair does not exist.
 Use add_plot_unit(); to create them.
`             );            %}}}
    return;
 }
 % else: arguments are okay.
 x_unit = x_unit_arg;
 y_unit = y_unit_arg;

 if(x_unit == "psd")
 {
    plot_unit("kev");
 }
 else if(x_unit == "micron")
 {
    plot_unit("um");
 }
 else
 {
    () = fprintf(fp, "\n Also setting ISIS intrinsic plot units to: %s \n\n",x_unit);
    try{ plot_unit(x_unit); } % Try to set all ISIS plots to the X_unit
    catch AnyError: 
    { () = fprintf(fp, "\n Unit not recognized by ISIS intrinsic plot routines. \n\n"); }
 }

 y_scl=y_scl_norm[x_unit+y_unit];

 return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% pd will store the plotted data.  There will be one Array_Type or
% List_Type index for each input data set or group.  Data/models only
% need a single array for each (they are either single data sets, or
% combined).  Residuals need lists to hold an array for each
% combination member.

private define pd_set(indx,indx_ln,ary_sze)
{
   variable pfield, i, j, a;

   variable list=[ "index", "file", "dlo", "dhi", "dval", "derr", 
                   "model_on", "mlo", "mhi", "mval", "weight",
                   "instrument", "grating", "order", "part"];
   
   foreach pfield (list)
   {
      set_struct_field( pd, pfield, Array_Type[indx_ln] );
   }

   pd.xaxis = "NONE";
   pd.yaxis = "NONE";
   pd.raxis = "NONE";
   pd.exp = Double_Type[indx_ln];

   pd.res = array_map(List_Type,&__pop_list,Char_Type[indx_ln]);
   pd.res_m = array_map(List_Type,&__pop_list,Char_Type[indx_ln]);
   pd.res_p = array_map(List_Type,&__pop_list,Char_Type[indx_ln]);

   pd.con_mod = 1;  % Unfolded model defaults to smeared by response

   _for i (0,indx_ln-1,1)
   {
      pd.file[i] = String_Type[ary_sze[i]];
      pd.instrument[i] = String_Type[ary_sze[i]];
      pd.grating[i] = String_Type[ary_sze[i]];
      pd.index[i] = Integer_Type[ary_sze[i]];
      pd.order[i] = Integer_Type[ary_sze[i]];
      pd.part[i] = Integer_Type[ary_sze[i]];
      pd.model_on[i] = ["      no"];

      _for j (0,ary_sze[i]-1,1)
      {
         pd.index[i][j]=[indx[i]][j];
         a = get_data_info([indx[i]][j]);
         pd.file[i][j] = a.file+"/part="+string(a.part)
                         +"/order="+string(a.order);
         pd.instrument[i][j] = a.instrument;
         pd.grating[i][j] = a.grating;
         pd.order[i][j] = a.order;
         pd.part[i][j] = a.part;
      }
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%

private define p_reset()
{
   set_plot_options(plopt);
   ifnot(no_reset) multiplot(1);
}  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define multiplotit(res)
{
   if(res) multiplot([3,1]);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define vaxis(i,bin_lo,bin_hi)
{
   if(unit_info(x_unit).is_energy)
   {
      if(vzero!=NULL)  % Velocity/redshift X-axis relative to vzero
      {
         zhi = (vzero-bin_lo)/bin_lo;
         zlo = (vzero-bin_hi)/bin_hi;

         if(zaxis)
         {
            bin_lo=zlo;
            bin_hi=zhi;
            xlabel(redlbl);
         }
         else 
         {
            bin_lo = ((zlo+1)^2-1)/((zlo+1)^2+1)*Const_c/1.e5;
            bin_hi = ((zhi+1)^2-1)/((zhi+1)^2+1)*Const_c/1.e5;
            xlabel(vlbl);
         }

      }
      else  % Or just redshift the existing X-axis
      {
         bin_lo = bin_lo*(1+zshift[i]);
         bin_hi = bin_hi*(1+zshift[i]);
      }      
   }
   else
   {
      if(vzero!=NULL) % Velocity/redshift X-axis relative to vzero
      {
         zlo = (bin_lo-vzero)/vzero;
         zhi = (bin_hi-vzero)/vzero;

         if(zaxis)
         {
            bin_lo=zlo;
            bin_hi=zhi;
            xlabel(redlbl);
         }
         else
         {
            bin_lo = ((zlo+1)^2-1)/((zlo+1)^2+1)*Const_c/1.e5;
            bin_hi = ((zhi+1)^2-1)/((zhi+1)^2+1)*Const_c/1.e5;
            xlabel(vlbl);
         }
      }
      else  % Or just redshift the existing X-axis
      {
         bin_lo = bin_lo/(1+zshift[i]);
         bin_hi = bin_hi/(1+zshift[i]);
      }
   }  
   return bin_lo, bin_hi;
}     

%%%%%%%%%%%%%%%%%%%

private define I(n)
{
   return n;
}

private variable _revr = [&I, &reverse], _rev;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define make_exp_and_back(hist_index)
{
   variable arfinfo,expos,scl,rb,nt,unbin_back,bin_back,rebin_back,iw;
   variable dscl = get_data_backscale(hist_index);
   variable dexp = get_data_exposure(hist_index); 
   variable bscl = get_back_backscale(hist_index);
   variable bexp = get_back_exposure(hist_index);

   if(bexp == NULL || bscl == NULL || bexp == 0 || min([bscl]) == 0)
   {
      scl = 1.;  % Throw our hands up and concede defeat
   }
   else 
   {
      scl = (dexp*dscl)/(bexp*bscl);
      if(1 < length(dscl)==length(scl))  % if dscl is an array
        scl[where(dscl==0)] = 0;         %   fix possible 0/0 = NaN elements
   }

   % Is scl a vector?  If yes, must bin to the data, such that
   % \sum_group scl*B_scaled = scl'*\sum_group*Bscl

   if(length(scl)>1)
   {
      % First, store the current binning, then unbin the data

      bin_back = get_back(hist_index);

      if(bin_back != NULL && length(bin_back)>1)
      {
         rb = get_data_info(hist_index).rebin;
         nt = get_data_info(hist_index).notice_list;
         rebin_data(hist_index,0);
         unbin_back = get_back(hist_index);

         % Multiply the unbinned, scaled background by scale, then bin
         rebin_back = rebin_array(scl*unbin_back,rb);
         scl = Double_Type[length(rebin_back)];   

         % Only compute scale where bin_back !=0
         iw = where(bin_back != 0);
         scl[iw] = rebin_back[iw]/bin_back[iw];

         % Set the data right again
         rebin_data(hist_index,rb);
         ignore(hist_index);
         notice_list(hist_index,nt);
      }
   }

   expos=dexp;
   arfinfo = get_data_info(hist_index).arfs[0];
   if(arfinfo != 0)
   {
      expos = get_arf_exposure(arfinfo);
      if(expos==NULL || expos<=0.) 
      {
         expos = dexp;
      }
   }

   return expos, scl;
}   

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private variable pscale=1;

public define set_power_scale()
%!%+
%\function{set_power_scale}
%\synopsis{Determine the y-axis is scaled by energy when using the isis_fancy_plot package.}
%\usage{set_power_scale(Integer_Type);}
%\description
%set_power_scale(a);  
%
%  Determine how the y-axis is scaled by energy or wavelength when using
%  the fancy_plot routines. a=1, E|lambda is set to the midpoint of the 
%  bin. a=2, E|lambda is set to the geometric midpoint (i.e., sqrt(Elo*Ehi)).
%  a=3, E|lambda is set to the average value assuming an x^-2 powerlaw.
%  Any other values, and this message is printed.
%\seealso{fancy_plot_unit, add_plot_unit, plot_unfold, plot_counts}
%!%-
{
   variable a = 4;
   if(_NARGS==1)
   {
      a = ();
      a = int(a);
      if(a != 1 && a != 2 && a !=3) a = 4;
   }
   switch(a)
   {case 1: pscale=1; return; }
   {case 2: pscale=2; return; }
   {case 3: pscale=3; return; }
   {case 4:
      () = fprintf(fp, "\n%s\n", %{{{
`
set_power_scale(a);  

  Determine how the y-axis is scaled by energy or wavelength when using
  the fancy_plot routines. a=1, E|lambda is set to the midpoint of the 
  bin. a=2, E|lambda is set to the geometric midpoint (i.e., sqrt(Elo*Ehi)).
  a=3, E|lambda is set to the average value assuming an x^-2 powerlaw.
  Any other values, and this message is printed.
`
 %}}}
      );
      () = fprintf(fp, "   Current setting = %2i\n\n",pscale);
   }
}

private define power_scale(lo,hi,power)
{
   variable pow;
   switch(power)
   { case 0:
      switch(pscale)
      { case 1:      pow =  2./(lo+hi);                    }
      { case 2:      pow =  1./sqrt(lo*hi);                }
      { case 3:      pow =  (1-(lo/hi)^3)/(1-lo/hi)/lo/3.; }
   }
   { case 1:         pow = ones( length(lo) );             }
   { case 2: 
      switch(pscale)
      { case 1:      pow = (lo+hi)/2.;                     }
      { case 2:      pow = sqrt(lo*hi);                    }
      { case 3:      pow = lo*hi/(hi-lo)*log(hi/lo);       }
   }
   { case 3:
      switch(pscale)
      { case 1:      pow = sqr((lo+hi)/2.);                }
      { case 2 or 3: pow = (lo*hi);                        }
   }
   return pow;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define make_data(iindx,res,power,modc,bkgon,rescl)
{
   % Create the X-axis data (can be overridden with Velocity axis)

   % Manfred Hanke wants this done once, outside the loop *and* wants
   % error checking that all data sets have the same grid.  I actually
   % want to allow munging of the grids for things that are close, but
   % not exact, matches (e.g., HEXTE). I know, it's dangerous...

   variable unit_info_x = unit_info(x_unit);
   variable data = get_data_counts(iindx[0]);
   variable bin_lo = data.bin_lo;
   variable bin_hi = data.bin_hi;

   % Flip to energy, but don't reverse quite yet
   if(unit_info_x.is_energy)
     (bin_lo,bin_hi) = (_A(1)/bin_hi, _A(1)/bin_lo);

   bin_lo /= unit_info_x.scale;
   bin_hi /= unit_info_x.scale;

   % Powers of x to multiply against y-axis
      
   variable pow = power_scale(bin_lo,bin_hi,power); 

   % Noticed data indices
   variable dnote = get_data_info(iindx[0]).notice;    

   bkgon = int(bkgon);                                 % Background subtraction?
   variable meth = strtok(get_fit_statistic,";=")[-1]; % Fit Method

   % Running totals of plot variables
   variable tdata=0., tmodl=0., ters=0., tders=0., tbers=0., tbgd=0., texpos=0.;

   variable i; 
   foreach i (iindx)
   {
      data = get_data_counts(i);                   % Data (on wavelength grid)
      variable sys_err_frac = get_sys_err_frac(i); % Systematic errors
      variable sys_err2 = length(sys_err_frac)   
                         ? (sys_err_frac*data.value)^2
                         : 0;

      % Arf exposure and data backscale - the latter can exist
      % independently of an actual background

      variable expos, bdscl;
      (expos, bdscl) = make_exp_and_back(i);

      variable bgd = get_back(i);                  % The background, if it exists

      if(bgd == NULL)
      {
         bgd = Float_Type[length(data.value)];
         bdscl=0.;
      }

      variable modl;
      if(modc > 0) 
      {
         % Running total of model kept separate in case none defined
         modl = get_model_counts(i).value;

         % In ISIS, "background subtraction" does not affect defintion
         % of chi^2 derived from model counts, so tmodl suffices
   
         tmodl += modl; 
      }
      else
      {
         tmodl  += 0.*data.value;  % Make sure modl comes back as array, even
                                    % if empty (avoids various modc=0 checks)
      }

      % ... However, "background subtraction" does affect definition of
      % data error bars.  It increases the data error bars, and is 
      % *always* part of chi^2, if using data as definition of sigma

      variable ders = sqr( meth == "gehrels"
                           ? 1.+sqrt(data.err^2+0.75)
                           : data.err
                         ) + sys_err2;

      % Running total of exposure, data, data errors, and background
      % Note that we "undo" background error for *data*, if including
      % the background in the data plot.
      texpos += expos;
      tdata  += data.value; 
      tders  += ders;
      tbgd   += bgd;
      tbers  += bdscl*bgd;
   }

   if( meth == "model" && modc > 0 )
   {
      ters = tmodl;  % Residuals calculated from model errors
   }
   else
   {
      % Residuals always have background errors, if using data...
      ters = tders+tbers; % ...or data errors, if model not defined/chosen
   }

   % ... but data only have background errors if subtracting background
   ifnot(bkgon) tders += tbers; 

   ters  = sqrt(ters);
   tders = sqrt(tders);
   tbers = sqrt(tbers);

   % Reverse all arrays, if energy unit.
   variable r;
   if(unit_info_x.is_energy) 
     foreach r ([&bin_lo, &bin_hi, &pow, &dnote,
                 &tdata, &tmodl, &ters, &tders, &tbers, &tbgd])
            @r = reverse(@r);

   % Rescale everything
   tdata *= rescl[0]; tders *= rescl[0]; ters *= rescl[0]; tbgd *= rescl[0];
   tbers *= rescl[0]; tmodl *= rescl[0]; 

   variable perunit = ( bin_hi - bin_lo )*texpos;

#ifexists get_bin_corr_factor
   % --- Added by Remeis group, Thomas Duaser: new; for FERMI data ---
   if (get_bin_corr_factor(i) != NULL)
     perunit *= get_bin_corr_factor(i);
   % -------------------------------
#endif

   %  Make sure we only plot where data & model are well-defined

   variable iw;
   if(res <= 0 || modc == 0)
   {
      iw = where(dnote !=0);   % No residuals or no model
   }
   else if(res == 1 || res == 2 || res == 4 || res == 5)
   {
      iw = where(dnote != 0 and ters > 0.); % Chi/chi^2 residuals
   }
   else   % Ratio residuals, which need model defined & !=0
   {
      if( meth == "model" && modc > 0 )
      {
         iw = where(dnote != 0 and tmodl-tbgd-ters > 0.);
      }
      else
      {
         iw = where(dnote != 0 and tmodl-tbgd > 0.); 
      }
   }

   variable resd=0, mres=0, pres=0;
   if(modc > 0)
   {
      switch(res)
      {
       case 3 or case 6:                 % Ratio residuals
         resd = ones(length(tdata))*1.;
         mres = @resd;
         pres = @resd;
       
         % If background not subtracted, ratio the total data & model counts
         if( meth == "model" )
         {
            resd[iw] = (tdata[iw]-(1-_max(0,bkgon))*tbgd[iw])/
                       (tmodl[iw]-(1-_max(0,bkgon))*tbgd[iw]);
            pres[iw] = (tdata[iw]-(1-_max(0,bkgon))*tbgd[iw])/
                       (tmodl[iw]-(1-_max(0,bkgon))*tbgd[iw]-ters[iw]);
            mres[iw] = (tdata[iw]-(1-_max(0,bkgon))*tbgd[iw])/
                       (tmodl[iw]-(1-_max(0,bkgon))*tbgd[iw]+ters[iw]);
         }
         else
         {
            resd[iw] = (tdata[iw]-(1-_max(0,bkgon))*tbgd[iw])/
                       (tmodl[iw]-(1-_max(0,bkgon))*tbgd[iw]);
            mres[iw] = (tdata[iw]-(1-_max(0,bkgon))*tbgd[iw]-ters[iw])/
                       (tmodl[iw]-(1-_max(0,bkgon))*tbgd[iw]);
            pres[iw] = (tdata[iw]-(1-_max(0,bkgon))*tbgd[iw]+ters[iw])/
                       (tmodl[iw]-(1-_max(0,bkgon))*tbgd[iw]);
         }
      }
      {
       case 1 or case 2 or case 4 or case 5:  % Chi/Chi^2 residuals
         resd = ones(length(tdata))*1.;
         mres = @resd;
         pres = @resd;
         resd[iw] = (tdata[iw] - tmodl[iw]) / ters[iw];
         mres[iw] = resd[iw] - 1.;
         pres[iw] = resd[iw] + 1.;
     
         if(res == 2 || res == 5)
         {
            resd = sign(resd) * resd^2;
            mres = sign(mres) * mres^2;
            pres = sign(pres) * pres^2;
         }   

         variable iww, sgn_resd;

         if(meth=="cash") % Override Chi/Chi^2 if Cash statistics chosen
         {
            iww = where(tdata[iw] > 0);
            sgn_resd = sign(resd[iw])*2;
            resd[iw]= (tmodl[iw]-tdata[iw]);

            if(length(iww)>0)
            {
               resd[iw[iww]] += tdata[iw[iww]]*
                                  log(tdata[iw[iww]]/tmodl[iw[iww]]);
            }
            resd[iw] *= sgn_resd;

            if(res == 1 || res == 4) resd[iw] /= sqrt(abs(resd[iw]));
         }

         if(meth=="ml") % Override Chi/Chi^2 if Maximum Likelihood statistics chosen
         {
            iww = where(tmodl[iw] > 0);
            sgn_resd = sign(resd[iw])*2;
            resd[iw]= tmodl[iw] + log(tdata[iw]+1);

            if(length(iww)>0)
            {
               resd[iw[iww]] -= tdata[iw[iww]]*log(tmodl[iw[iww]]);
            }
            resd[iw] *= sgn_resd;

            if(res == 1 || res == 4) resd[iw] /= sqrt(abs(resd[iw]));
         }
      }
   }

   ifnot( bkgon )  % Leave the background in, or subtract it  
   {
      tdata = (tdata-tbgd)*pow;
      if(modc > 0) tmodl = (tmodl-tbgd)*pow;
   }
   else if(bkgon < 0)
   {
      tdata = tbgd*pow;
      tders = tbers;
   }
   else
   {
      tdata = tdata*pow;
      if(modc > 0) tmodl = tmodl*pow;
   }

   tders = tders*pow;

   if(pd_mean_not_sum !=0)  % Sum or average the exposures.
   {
	perunit=perunit/length(iindx);
   }
   return bin_lo,bin_hi,tdata,[tmodl],tders,iw, % Used by plot_counts/unfold
          perunit,[resd],[mres],[pres],         % Used by plot_data/residuals
          texpos;                               % Info Moritz wanted
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

private define hp_loop(iw)
{
   variable wiw, ja, jb;
   wiw = where( (iw - shift(iw,-1)) > 1 );
   if(length(wiw) != 0)
   {
      ja = [0,wiw];
      jb = [wiw-1,length(iw)-1];
   }
   else
   {
      ja = [0];
      jb = [length(iw)-1];
   }
   return ja, jb;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define hp_loopit(xl,xh, y, color, iw, ja, jb, gaptest)
{
   pointstyle(-1);
   variable jj, diw;
   variable plt_opts=get_plot_options;
   variable minx=plt_opts.xmin;
   variable maxx=plt_opts.xmax;
   variable miny=plt_opts.ymin;
   variable maxy=plt_opts.ymax;

   if(gaptest)
   {
      _for jj (0,length(ja)-1,1)
      {
         diw = iw[[ja[jj]:jb[jj]]];
         if(    any(minx <= xl[diw] <= maxx)
             && any(minx <= xh[diw] <= maxx)
             && any(miny <=  y[diw] <= maxy) )
         {
            ohplot( xl[diw], xh[diw], y[diw], color );
         }
      }
   } 
   else
   {
      if(    any(minx <= xl[iw] <= maxx)
          && any(minx <= xh[iw] <= maxx)
          && any(miny <=  y[iw] <= maxy) )
      {
         oplot( (xl[iw]+xh[iw])/2., y[iw], color );
      }
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define xerplot(res,blo,bhi)
{
   variable xer=0.,vlo,lo,mid,hi,vhi;

   set_line_width(2); linestyle(2); connect_points(-1);

   if ( res == 3 || res == 6 ) xer = 1.;
  
   lo=_max(blo,get_plot_options.xmin); vlo=lo;
   hi=_min(bhi,get_plot_options.xmax); vhi=hi;
   mid=(lo+hi)/2.;

   if(-3.e38 < get_plot_options.xmin < lo)
      vlo=get_plot_options.xmin;

   if(hi < get_plot_options.xmin < 3.e38)
      vhi=get_plot_options.xmax;

   oplot( [vlo,lo,mid,hi,vhi],[xer,xer,xer,xer,xer], 1);

   set_line_width(r_width); linestyle(1);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% A kluge to get the autoscaling right for multiple datasets.  Plot
% two small points in white to start the data off.

private define start_plot(axlo, axhi, aylo, ayhi)
{  
   if(oplt!=0) return 1;  % Bug out if this is an overplot
   variable p = get_plot_options, xlo, xhi, ylo, yhi;
   variable xloglo=-3.e38, yloglo=-3.e38;
   if(get_plot_options.logx) xloglo=1.e-38;
   if(get_plot_options.logy) yloglo=1.e-38;
   variable iw = where( axhi >= p.xmin and axlo <= p.xmax and
                        ayhi >= p.ymin and aylo <= p.ymax and
                        axlo > xloglo and aylo > yloglo );

   if(length(iw)==0 || max(abs([p.xmin,p.xmax,p.ymin,p.ymax]))<3.e38)
   {
      return 0;
   }
   else
   {
      xlo = min(axlo[iw]);
      xhi = max(axhi[iw]);
      ylo = min(aylo[iw]);
      yhi = max(ayhi[iw]);
   }

   if(p.ymin > -3.e38) ylo=p.ymin;
   if(p.xmin > -3.e38) xlo=p.xmin;
   if(p.ymax <  3.e38) yhi=p.ymax;
   if(p.xmax <  3.e38) xhi=p.xmax;

   connect_points(0); point_size(0.1); pointstyle(0);
   plot([xlo,xhi],[ylo,yhi],0);
   return 1;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define datplot_err(xm,xp,yy,yp,ym,width,col,isres)
{
   if(width==0 || col==0) return;

   _pgslw(int(width)); _pgsls(1);

   variable xx;

   xx = (xm+xp)/2.;

   if(get_plot_options().logx) 
   { 
      xm[where(xm <= 0)] = 1.e-38;
      xx[where(xx <= 0)] = 2.e-38;
      xp[where(xp <= 0)] = 3.e-38;
      xm = log10(xm); 
      xx = log10(xx); 
      xp = log10(xp); 
   }
   if(get_plot_options().logy) 
   { 
      ym[where(ym <= 0)] = 1.e-38;
      yy[where(yy <= 0)] = 2.e-38;
      yp[where(yp <= 0)] = 3.e-38;
      ym = log10(ym); 
      yy = log10(yy);
      yp = log10(yp); 
   }

   _pgsci(int(col));
   () = _pgerry(length(xx),xx,ym,yp,ebar_y);

   % Only residuals get X-error bars

   if(isres || data_err)
   {
      _pgsci(int(col));
      () = _pgerrx(length(xx),xm,xp,yy,ebar_x);
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define res_set(yrng,res)
{
   variable meth = strtok(get_fit_statistic,";=");
   meth = meth[-1];

   ylin;
   switch (length(yrng))
   {
    case 4:
      yrange(yrng[2],yrng[3]);
   }
   {
    case 3:
      yrange(yrng[2],NULL);
   }
   {
      yrange();
   }
 
   if(ylabl==NULL)
   {
      ylabel(rlbl[abs(res)-1]);
      if(meth=="cash" && (res==1 || res==4))
      {
         ylabel(clbl[0]);
      }
      if(meth=="cash" && (res==2 || res==5))
      {
         ylabel(clbl[1]);
      }
      if(meth=="ml" && (res==1 || res==4))
      {
         ylabel(mllbl[0]);
      }
      if(meth=="ml" && (res==2 || res==5))
      {
         ylabel(mllbl[1]);
      }
   }
   else
   {
      ylabel([ylabl][-1]);
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define resplot(yrng,res,indx,indx_ln,ary_sze,rsym,rcol,bkg)
{
   variable bin_lo,bin_hi,data,modl,ers,resd,mres,pres,iw;
   variable indx_use,i,j,jj,jstop,ja,jb;
   variable meth = strtok(get_fit_statistic,";=");
   meth = meth[-1];

   res_set(yrng,res);

   % If res <= 3, the residuals are to be left uncombined

   if(res > 3)  
   {
      indx_use = indx_ln;
      jstop=Integer_Type[indx_ln];  % Zeros, i.e., only one loop
   }
   else
   {
      indx_use = int(sum(ary_sze));
      jstop=ary_sze-1;              % As many loops as there are data sets
   }

   bin_lo = Array_Type[indx_use];  bin_hi = @bin_lo; 
   resd = @bin_lo; mres = @bin_lo; pres = @bin_lo; iw = @bin_lo;
   variable all_lo=Double_Type[0], all_hi=@all_lo, all_data_lo=@all_lo, 
            all_data_hi=@all_lo, istrt=0;

   jj=0;
   _for i (0,indx_ln-1,1)
   {
      _for j (0,jstop[i],1)
      {
         ifnot(jstop[i])    % Combine the residuals
         {
            (bin_lo[jj], bin_hi[jj],,,,iw[jj],,resd[jj],mres[jj],pres[jj],)= 
               make_data(indx[i],res,1,1,bkg[i],1.*ones(length(indx[i])));
         }
         else               % Don't combine residuals
         {
            (bin_lo[jj], bin_hi[jj],,,,iw[jj],,resd[jj],mres[jj],pres[jj],)= 
               make_data(indx[i][j],res,1,1,bkg[i],[1.]);
         }

         if(vzero!=NULL || zshift[i]!=0.)
         {
            (bin_lo[jj],bin_hi[jj]) = vaxis(i,bin_lo[jj],bin_hi[jj]);
         }

         % Save data to write to ASCII file later

         list_append(pd.res[i],resd[jj][iw[jj]]);
         list_append(pd.res_m[i],mres[jj][iw[jj]]);
         list_append(pd.res_p[i],pres[jj][iw[jj]]);

         pd.dlo[i] = bin_lo[jj][iw[jj]];
         pd.dhi[i] = bin_hi[jj][iw[jj]];

         all_lo = [all_lo,bin_lo[jj][iw[jj]]]; 
         all_hi = [all_hi,bin_hi[jj][iw[jj]]]; 
         all_data_lo = [all_data_lo,mres[jj][iw[jj]]]; 
         all_data_hi = [all_data_hi,pres[jj][iw[jj]]];

         jj++;
      }
   }

   pd.raxis = get_plot_options().ylabel;
   pd.xaxis = get_plot_options().xlabel;

   set_line_width(1);
   istrt = start_plot(all_lo, all_hi, all_data_lo, all_data_hi);

   jj=0;
   _for i (0,indx_ln-1,1)
   {
      _for j (0,jstop[i],1)
      {
         if(res==2 || res==5 || rsym[i][0]==0)   % Histogram plots
         {
            pointstyle(1);
            point_size(0.1);
         }
         else                                    % Data symbol plots
         {
            pointstyle(rsym[i][j]);
            point_size(plopt.point_size);
         }
         connect_points(0);  
         ifnot(min([jj+istrt+oplt,1])) % In case a plot hasn't been started yet...
         {
            slangref = refplt[0];
            @slangref((bin_lo[jj][iw[jj]]+bin_hi[jj][iw[jj]])/2.,resd[jj][iw[jj]],rcol[i][j]);
         }

         ifnot(jj) xerplot(res,min(all_lo),max(all_hi));

         % These should all be histograms without error bars
         if(res==2 || res==5 || ((meth=="cash" || meth=="ml") && (res!=3 && res!=6)))
         {
            connect_points(-1);
            (ja,jb) = hp_loop(iw[jj]);
            hp_loopit(bin_lo[jj], bin_hi[jj], resd[jj], rcol[i][0], iw[jj], ja, jb, 1);
         }
         else if(res==1 || res==3 || res==4 || res==6) % Have error bars, could be hists
         {
            datplot_err(bin_lo[jj][iw[jj]],bin_hi[jj][iw[jj]],resd[jj][iw[jj]],
                        mres[jj][iw[jj]],pres[jj][iw[jj]],re_width,recol[i][j],1);

            point_size(plopt.point_size);
            pointstyle(rsym[i][j]);
            connect_points(0);
            set_line_width(r_width);

            if(rsym[i][j] != 0)   % Plot with symbols
            {
               slangref = refplt[1];
               @slangref((bin_lo[jj][iw[jj]]+bin_hi[jj][iw[jj]])/2.,
                          resd[jj][iw[jj]],rcol[i][j]);	
            }
            else                  % Plot with histograms
            { 
               pointstyle(1);
               connect_points(1);
               (ja,jb) = hp_loop(iw[jj]);
               hp_loopit(bin_lo[jj],bin_hi[jj],resd[jj],rcol[i][j],iw[jj],ja,jb,1);
            }
         }
         jj++;
      }
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define p_range_label(res,power)
{
   % Very important!  I presume plot_bin_integral throughout these 
   % plotting functions!!!

   plot_bin_integral;

   multiplotit(res);

   if(xlabl==NULL)
   {
      xlabel(xlbl[x_unit]);
   }
   else
   {
      xlabel(xlabl);
   }

   if(vzero!=NULL)
   {
      xlin;
   }

   if(length(xrng)>1)
   {
      if((xrng[0]!=NULL) && (xrng[0]<=0))
      {
         xlin;
      }
      xrange(xrng[0],xrng[1]);
   }
   % No else ... so as to respect prior isis> xrange(...); choices,
   % but I will overwrite with new xrange choices.

   plopt = get_plot_options;

   if(ylabl==NULL)
   {
      ylabel(ylbl[x_unit+y_unit][power]);
   }
   else
   {
      ylabel([ylabl][0]);
   }

   if (length(yrng) > 1)
   {
      if((yrng[0]!=NULL) && (yrng[0]<=0))
      {
         ylin;
      }
      yrange(yrng[0],yrng[1]);
   }
   else
   {
      yrange();
   }

   return plopt;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Set defaults on plotting parameters

private define make_default(strct,var,var_df,is_list)
{
   % strct holds user inputs, var_df defaults, and var the values to use

   ifnot(is_list) % if shouldn't be a list, make an array
   {
      if(typeof(strct)==List_Type)
      {
         strct = list_to_array(strct);
      }
   }
   if(strct!=NULL) % if value input, assign ...
   {
      @var=strct;
   }
   else
   {
      @var=var_df; % ... otherwise use default
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define trans_strct_and_array(strct,iln,ary,var,var_df)
{
   % strct holds user inputs, var_df defaults, and var the values to use
   % iln is the number of input data groups, ary the # of indices/group

   variable i, lparm, strc_use, lstrc;

   lparm=0;
   if(strct != NULL) lparm = min( [ length(strct), iln ] );

   % Create lists & arrays with default values, but overwrite
   % those portions that have actual user input

   _for i (0,lparm-1,1)    %% Some user input
   {
      strc_use=Integer_Type[ary[i]]+[var_df[i]][0];
      lstrc=length(strct[i]);
      strc_use[[0:min([lstrc-1,ary[i]-1])]]=
         [strct[i]][[0:min([lstrc-1,ary[i]-1])]];
      list_append(@var,strc_use);
   }
   _for i (lparm,iln-1,1)  %% No user input
   {
      strc_use=Integer_Type[ary[i]]+[var_df[i]][0];
      list_append(@var,strc_use);
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define trans_strct(strct,iln,var,var_df)
{  
   % strct holds user inputs, var_df defaults, and var the values to use
   % iln is the number of input data groups

   variable i, lparm;
   if(length(var_df)<iln) % Make sure defaults cover # of data sets
   {
      var_df = {var_df[0]};
      loop(iln-1){ list_append(var_df,var_df[0]); }
   }

   lparm = 0;
   if(strct != NULL) lparm = min( [ length(strct), iln ] );

   _for i (0,lparm-1,1)
   {
      list_append(@var,[strct[i]][0]);
   }
   _for i (lparm,iln-1,1)
   {
      list_append(@var,var_df[i]);
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Make arrays into lists (for input)

private define make_array_a_list(a)
{
   variable b, c={};
   foreach b (a)
   {
      list_append(c,b);
   }
   return c;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Make all_data into a list

public define data_list()
%!%+
%\function{data_list (isis_fancy_plots package)}
%\synopsis{Turn an array into a list.}
%\usage{data_list(Array);}
%\description
%\seealso{list_to_array}
%!%-
{
   return make_array_a_list(all_data);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The below, in fact, is the way all parameters will ultimately be
% set, and hence has all the error checking.  pstruct is a structure
% variable with the plot options (either input or the defaults), while
% indx is a list with the indices (single values or arrays to be
% combined).

private define set_plot_parms_struct(a, indx_ln, ary_sze)
{
   variable pstruct,i,list,lparm,mcol_df_use,strc_use,lstrc;
    
   pstruct = @a;

   foreach list (parms)
   {
      % An input structure might not have all fields

      ifnot(struct_field_exists( pstruct, list ))
      {
         pstruct = struct_combine( pstruct, list );
      }

      % Qualifiers will overwrite a structure field

      if(qualifier_exists(list))
      {
         set_struct_field(pstruct,list,qualifier(list,NULL));
      }

      % But NULL means "use defaults"

      if(get_struct_field( pstruct, list ) != NULL)
      {
         % Array inputs that should be lists are converted

         if(length(where(list_params==list))!=0)
         {
            if(typeof(get_struct_field(pstruct,list))!=List_Type)
            {
               set_struct_field(pstruct,list,
                   make_array_a_list(get_struct_field(pstruct,list)));
            }
         }
         else if(typeof(get_struct_field(pstruct,list))==List_Type)
         {
            set_struct_field(pstruct,list,
                             list_to_array(get_struct_field(pstruct,list)));
         }
      }
   }

   % No model plotting if models are undefined!

   if( get_fit_fun() != NULL && get_fit_fun() != "null" 
                             && get_fit_fun() != "bin_width(1)" )
   {
      mcol_df_use = mcol_df;
   }
   else
   {
      mcol_df_use = 0;
   }

   dcol={}; decol={}; mcol={}; rcol={}; recol={}; dsym={}; rescale={}; 
   rsym={}; bkg={}; zshift={}; gap={};

   % Transfer structure values over to the variables

   trans_strct(pstruct.dcol,indx_ln,&dcol,dcol_df);
   trans_strct(pstruct.decol,indx_ln,&decol,dcol);
   trans_strct(pstruct.dsym,indx_ln,&dsym,style_df);
   trans_strct(pstruct.mcol,indx_ln,&mcol,mcol_df_use);
   trans_strct(pstruct.zshift,indx_ln,&zshift,0.);
   trans_strct(pstruct.bkg,indx_ln,&bkg,0);
   trans_strct(pstruct.gap,indx_ln,&gap,gap_df);

   % Variables that could have arrays in the input list  

   trans_strct_and_array(pstruct.rcol,indx_ln,ary_sze,&rcol,dcol);
   trans_strct_and_array(pstruct.recol,indx_ln,ary_sze,&recol,rcol);
   trans_strct_and_array(pstruct.rsym,indx_ln,ary_sze,&rsym,dsym);
   trans_strct_and_array(pstruct.scale,indx_ln,ary_sze,&rescale,1.*ones(indx_ln));

   make_default(pstruct.yrange,&yrng,{NULL,NULL,NULL,NULL},1);
   make_default(pstruct.xrange,&xrng,{NULL},1);
   make_default(pstruct.xlabel,&xlabl,NULL,0);
      if(xlabl != NULL) 
      { 
         if(typeof(xlabl)==List_Type) xlabl=list_to_array(xlabl);
         xlabl = [xlabl][0]; 
      }
   make_default(pstruct.ylabel,&ylabl,NULL,0);
      if(ylabl != NULL) 
      { 
         if(typeof(xlabl)==List_Type) ylabl=list_to_array(ylabl);
         ylabl = [ylabl]; 
      }
   make_default(pstruct.power,&power,power_df,0); 
      power=abs(int([power][0]));
      if(power!=0 && power!=1 && power!=2 && power !=3) power=power_df;
   make_default(pstruct.res,&res,0,0); res=abs(int([res][0]));
      if(res!=0 && res!=1 && res!=2 && res!=3 && res!=4 && res!=5 && res!=6) res=0;
   make_default(pstruct.oplt,&oplt,0,0); oplt=abs(int([oplt][0]));
   make_default(pstruct.vzero,&vzero,NULL,0); 
      if(vzero!=NULL){ vzero=abs([vzero][0]); };
   make_default(pstruct.zaxis,&zaxis,0,0); zaxis=abs(int([zaxis][0]));
   make_default(pstruct.con_mod,&use_con_flux,1,0); 
      use_con_flux=abs(int([use_con_flux][0]));
   make_default(pstruct.no_reset,&no_reset,0,0); 
      no_reset=abs(int([no_reset][0]));
   make_default(pstruct.sum_exp,&pd_mean_not_sum,sum_exp_df,0);
      pd_mean_not_sum=abs(int([pd_mean_not_sum][0]-1));
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Make sure the input indices form a list, and that any list within
% this list is replaced with an array of combination members, and that
% all indices are integers

private define make_indices(args)
{
   variable i, indx_ln, ary_sze, idx; 
   indx = args[0];
   if(typeof(indx) != List_Type) indx={ indx };

   indx_ln=length(indx);
   ary_sze=Integer_Type[indx_ln];

   _for i (0,indx_ln-1,1)
   {
      if(typeof(indx[i])==List_Type)
      {
         try
         {
            indx[i] = [combination_members(int(abs(indx[i][0])))];
            ary_sze[i] = length(indx[i]);
         }
         catch AnyError:
         {
            indx_ln = 0;
            return indx, indx_ln, ary_sze;
         }
      }
      else
      {
         foreach idx (indx[i])
         {
            try
            {
               indx[i] = int(abs(indx[i]));
               idx = int(abs(idx));
               () = get_data_info(idx).spec_num;
            }
            catch AnyError:
            {
               indx_ln = 0;
               return indx, indx_ln, ary_sze;
            }
         }
         ary_sze[i] = length(indx[i]);
      }
   }

   if(length(args)>2 || (length(args)>1 && typeof(args[-1])!=Struct_Type))
   {
      message(` 
 Multiple input arguments -- only the first will be plotted. Use list:

   plot_data({1,2,3},popt; qualifiers);

 to plot multiple datasets.
`            );
   }   
   return indx, indx_ln, ary_sze;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define set_plot_parms(nargs,args)
{
   variable indx_ln, ary_sze;

   (indx, indx_ln, ary_sze)  = make_indices(args);

   if(nargs > 1 && typeof(args[-1])==Struct_Type)
   {
      % Presume arguments are passed in a structure, but allow qualifiers
      % to overwrite the structure variables

      set_plot_parms_struct(args[-1],indx_ln,ary_sze;;__qualifiers);
   }
   else  % Or modify default structure based upon only qualifier inputs
   {
      set_plot_parms_struct(parm_s,indx_ln,ary_sze;;__qualifiers()); % Set defaults
   }

   return indx, indx_ln, ary_sze;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Strings for option descriptions that we will use many times

private variable ores_descript=
`  Residuals are units of chi, chi2, or ratio, and will be based upon whether
  one chooses sigma=model, data, or gehrels in set_fit_statistic();
  (data error bars are only affected by the latter two).
  set_fit_statistic("cash"); will alter the residuals to the Cash statistic.
  set_fit_statistic("ml"); will alter the residuals to the Maximum Likelihood statistic.
`;
private variable oopts=`
  Options below refer to structure variables/qualifiers

`;
private variable oindx=
`   indx    = list of data set indices to be plotted. Any indices grouped in
             an array within that list will be *combined* in the data plot.
             Single number in list is combo id, {#} = [combination_members(#)].
`;
private variable odcol=
`   dcol    = (pgplot) color value for data (or list of color values)
`;
private variable odecol=
`   decol   = (pgplot) color value for data error bars (or list of color values)
`;
private variable omcol=
`   mcol    = (pgplot) color value for model (or list of color values)
             0 => No model plotted
`;
private variable orcol=
`   rcol    = (pgplot) color value for residuals (or list of color values; arrays
             within the list allow for individual color values if portions of
             the data are combined, but their associated residuals are not)
`;
private variable orecol=
`   recol   = color for residual error bars (or list of color values; arrays
             within the list act as for residual color inputs)
`;
private variable odsym=
`   dsym    = (pgplot) symbol value for data (or list of symbol values)
             0 => histogram plot
`;
private variable orsym=
`   rsym    = (pgplot) symbol value for residuals (or list of symbol values;
             arrays within the list act as for residual color inputs)
             0 => histogram plot
`;
private variable oyrange=
`   yrange  = List of Y-limits for the data & model and (optionally) residuals
`;
private variable oxrange=
`   xrange  = List of X-limits for the data & model & residuals
             Note: Any X- or Y-range set to NULL is autoscaled
`;
private variable ooplt=
`   oplt    = 0 (default) for new plot, !=0 for overplotting
`;
private variable onrst=
`   no_reset= 0 (default)- plots *will* be reset, i.e., next plot moves to new pane
             (multiplot), next plot redraws window (single plot). no_reset=1 necessary
             for overplotting multiplots (oplt=1 sufficient for single plots).
`;
private variable ores=
`   res     = 0 (default), no residuals; 1, 2, or 3 = chi, chi2, or ratio residuals
             4, 5, or 6 = chi, chi2, or ratio, but combine residuals for combined data
             set_fit_method("cash"); or set_fit_method("ml") will cause res=(2 or 5) 
             or res=(1 or 4) to plot the residual for the Cash or Maximum Likelihood 
             statistic or its square root, respectively
`;
private variable obkg=
`   bkg     = List of 0's (subtract background-default), 1's (include backgrounds),
             or -1's (plot *only* the background [no model plotted in this case]).
             Ratio residuals will include background in data/model, other residuals
             are unaffected. Indices within a combination are treated the same.)
`;
private variable oxlabel=
`   xlabel  = String that will overwrite default X-axis label (default=NULL)
`;
private variable oylabel=
`   ylabel  = String or string array that will overwrite the default Y-axis labels
             (second element of array applies to residuals; default=NULL)
`;
private variable ozshift=
`   zshift  = List of redshifts to be applied to the data (default zshift={0,0,...})
`;
private variable ovzero=
`   vzero   = If set, the reference X-unit value to be defined as zero velocity.
             The X-axis then becomes a velocity axis (km/s) referenced to this
             point (default vzero=NULL; setting vzero/zaxis supersedes zshift)
`;
private variable ozaxis=
`   zaxis   = If not 0, use a redshift axis instead of a velocity axis *if* vzero
             is defined (default zaxis=0)
`;
private variable oscale=
`   scale   = Multiplicatively scale the Y-axis by the values in a list.
             Any arrays in the list should hold the individual scalings for
             data set arrays in the input index list.  **Note:**  these values
             only scale the plots, not the fits.  (Default values are 1.)
`;
private variable osum_exp=
`   sum_exp = If==1, then when combining data sets, sum the exposure times (as
             opposed to using the mean exposure time; default sum_exp=1).
`;
private variable ogap=
`   gap     =  1 (default), models are histograms with gaps where data has gaps,
              0          , model are bin-centered lines, without gaps.
`;
%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define plot_counts()
%!%+
%\function{plot_counts}
%\synopsis{Plot counts per bin (isis_fancy_plots package)}
%\description
%
%plot_counts({indx,[arry],{cid}},pstruct);  % pstruct = struct{ dcol, mcol, rcol, ...}
%plot_counts({indx,[arry],{cid}};dcol={val},mcol={val},rcol={val,[arry],val},...);
%
%  Plot background subtracted data, model, and residuals as counts/bin
%  Residuals are units of chi, chi2, or ratio, and will be based upon whether
%  one chooses sigma=model, data, or gehrels in set_fit_statistic();
%  (data error bars are only affected by the latter two).
%  set_fit_statistic("cash"); will alter the residuals to the Cash statistic.
%  set_fit_statistic("ml"); will alter the residuals to the Maximum Likelihood statistic.
%
%  Options below refer to structure variables/qualifiers
%
%   indx    = list of data set indices to be plotted. Any indices grouped in
%             an array within that list will be *combined* in the data plot.
%             Single number in list is combo id, {#} = [combination_members(#)].
%   dcol    = (pgplot) color value for data (or list of color values)
%   decol   = (pgplot) color value for data error bars (or list of color values)
%   mcol    = (pgplot) color value for model (or list of color values)
%             0 => No model plotted
%   rcol    = (pgplot) color value for residuals (or list of color values; arrays
%             within the list allow for individual color values if portions of
%             the data are combined, but their associated residuals are not)
%
%   recol   = color for residual error bars (or list of color values; arrays
%             within the list act as for residual color inputs)
%   dsym    = (pgplot) symbol value for data (or list of symbol values)
%             0 => histogram plot
%   rsym    = (pgplot) symbol value for residuals (or list of symbol values;
%             arrays within the list act as for residual color inputs)
%             0 => histogram plot
%   xrange  = List of X-limits for the data & model & residuals
%             Note: Any X- or Y-range set to NULL is autoscaled
%   yrange  = List of Y-limits for the data & model and (optionally) residuals
%   oplt    = 0 (default) for new plot, !=0 for overplotting
%   no_reset= 0 (default)- plots *will* be reset, i.e., next plot moves to new pane
%             (multiplot), next plot redraws window (single plot). no_reset=1 necessary
%             for overplotting multiplots (oplt=1 sufficient for single plots).
%   res     = 0 (default), no residuals; 1, 2, or 3 = chi, chi2, or ratio residuals
%             4, 5, or 6 = chi, chi2, or ratio, but combine residuals for combined data
%             set_fit_method("cash"); or set_fit_method("ml") will cause res=(2 or 5) 
%             or res=(1 or 4) to plot the residual for the Cash or Maximum Likelihood 
%             statistic or its square root, respectively
%   power  = 0, 1 (default), 2, or 3 for Counts/bin X
%            (1/Unit, 1, Unit, Unit^2), respectively
%   bkg     = List of 0's (subtract background-default), 1's (include backgrounds),
%             or -1's (plot *only* the background [no model plotted in this case]).
%             Ratio residuals will include background in data/model, other residuals
%             are unaffected. Indices within a combination are treated the same.)
%   xlabel  = String that will overwrite default X-axis label (default=NULL)
%   ylabel  = String or string array that will overwrite the default Y-axis labels
%             (second element of array applies to residuals; default=NULL)
%   zshift  = List of redshifts to be applied to the data (default zshift={0,0,...})
%   vzero   = If set, the reference X-unit value to be defined as zero velocity.
%             The X-axis then becomes a velocity axis (km/s) referenced to this
%             point (default vzero=NULL; setting vzero/zaxis supersedes zshift)
%   zaxis   = If not 0, use a redshift axis instead of a velocity axis *if* vzero
%             is defined (default zaxis=0)
%   scale   = Multiplicatively scale the Y-axis by the values in a list.
%             Any arrays in the list should hold the individual scalings for
%             data set arrays in the input index list.  **Note:**  these values
%             only scale the plots, not the fits.  (Default values are 1.)
%   gap     =  1 (default), models are histograms with gaps where data has gaps,
%              0          , model are bin-centered lines, without gaps.
%\seealso{plot_data, plot_unfold, plot_residuals, plot_fit_model, plotxy, plot_comps, plot_double}
%!%-
{
   variable i,args,indx_ln,iindx,ary_sze;
   variable data,bin_lo,bin_hi,modl,pres,mres,resd,ers,texp;
   variable iw;
   variable ja,jb;

   power_df=1; 

   if(_NARGS > 0)
   {
      args = __pop_list(_NARGS);
      (indx, indx_ln, ary_sze) = set_plot_parms(_NARGS,args;;__qualifiers());
      ifnot(indx_ln)
      {
         message("\n  One or more data sets undefined.\n");
         return;
      }
   }
   else
   {
      () = fprintf(fp, "\n%s\n", %{{{
`
plot_counts({indx,[arry],{cid}},pstruct);  % pstruct = struct{ dcol, mcol, rcol, ...}
plot_counts({indx,[arry],{cid}};dcol={val},mcol={val},rcol={val,[arry],val},...);

  Plot background subtracted data, model, and residuals as counts/bin
`
+ores_descript
+oopts+oindx+odcol+odecol+omcol+orcol+orecol+odsym+orsym+oxrange+oyrange+ooplt+onrst+ores
+`   power  = 0, 1 (default), 2, or 3 for Counts/bin X
            (1/Unit, 1, Unit, Unit^2), respectively
`
+obkg+oxlabel+oylabel+ozshift+ovzero+ozaxis+oscale+ogap); %}}}
      return;
   }

   %  And let the plotting begin ...

   plopt = p_range_label(res,power+5);

   % THE DATA & MODEL:

   pd_set(indx,indx_ln,ary_sze);
   pd.type="plot_counts";

   % Create the data first, in case the X or Y ranges are set to
   % autoscale.  This way we can get the autoscaling correct for the
   % full range of data

   texp = Double_Type[indx_ln]; bin_lo = Array_Type[indx_ln];  bin_hi = @bin_lo; 
   data = @bin_lo; modl = @bin_lo; ers = @bin_lo; iw = @bin_lo;
   variable all_lo=Double_Type[0], all_hi=@all_lo, all_data_lo=@all_lo, 
            all_data_hi=@all_lo, istrt=0;

   _for i (0,indx_ln-1,1)
   {
      (bin_lo[i], bin_hi[i], data[i], modl[i], ers[i], iw[i],,,,,texp[i]) = 
               make_data(indx[i],res,power,mcol[i],bkg[i],rescale[i]);
      
       if(vzero!=NULL || zshift[i]!=0.)
      {
         (bin_lo[i],bin_hi[i]) = vaxis(i,bin_lo[i],bin_hi[i]);
      }

      all_lo = [all_lo,bin_lo[i][iw[i]]]; 
      all_hi = [all_hi,bin_hi[i][iw[i]]]; 
      all_data_lo = [all_data_lo,data[i][iw[i]]-ers[i][iw[i]]]; 
      all_data_hi = [all_data_hi,data[i][iw[i]]+ers[i][iw[i]]];
   }

   istrt = start_plot(all_lo, all_hi, all_data_lo, all_data_hi);

   _for i (0,indx_ln-1,1)
   { 
      % Start the plots by plotting data points.  Choose mid-points of
      % bins, to ensure autoscaling matches between data & residuals

      if( dsym[i][0] == 0 )
      {
         pointstyle(1);
         point_size(0.1);
      }
      else
      {
         pointstyle(dsym[i][0]);
         point_size(plopt.point_size);
      }
      connect_points(0);
      ifnot(min([i+istrt+oplt,1]))  % In case we haven't started a plot yet...
      {
         slangref = refplt[0];
         @slangref((bin_lo[i][iw[i]]+bin_hi[i][iw[i]])/2.,data[i][iw[i]],dcol[i][0]);
      }
      point_size(plopt.point_size);

      % Start saving data for writing to an ASCII file

      pd.xaxis = get_plot_options().xlabel;
      pd.yaxis = get_plot_options().ylabel;
      pd.dlo[i] = bin_lo[i][iw[i]];
      pd.dhi[i] = bin_hi[i][iw[i]];
      pd.dval[i] = data[i][iw[i]];
      pd.derr[i] = ers[i][iw[i]];
      pd.exp[i] = texp[i];

      datplot_err(bin_lo[i][iw[i]],bin_hi[i][iw[i]],data[i][iw[i]],
                  data[i][iw[i]]-ers[i][iw[i]],data[i][iw[i]]+ers[i][iw[i]],
                  de_width,decol[i][0],0);

      set_line_width(d_width);

      if(dsym[i][0] != 0)   % Symbol plots
      {
         connect_points(0);
         pointstyle(dsym[i][0]);
         slangref = refplt[1];
         @slangref( (bin_lo[i][iw[i]]+bin_hi[i][iw[i]])/2., data[i][iw[i]], dcol[i][0] );
      }
      else                  % Histogram plots
      {  
         pointstyle(1);
         connect_points(-1);
         (ja,jb) = hp_loop(iw[i]);
         hp_loopit( bin_lo[i], bin_hi[i], data[i], dcol[i][0], iw[i], ja, jb, 1);
      }

      if(mcol[i][0] != 0 && bkg[i][0] >= 0)   % Only plot model if chosen & exists
      {
         set_line_width(m_width);  connect_points(-1);
         (ja,jb) = hp_loop(iw[i]);
         hp_loopit(bin_lo[i], bin_hi[i], modl[i], mcol[i][0], iw[i], ja, jb, gap[i]); 
         pd.model_on[i] = ["     yes"];
         pd.mlo[i] = bin_lo[i][iw[i]];
         pd.mhi[i] = bin_hi[i][iw[i]];
         pd.mval[i] = modl[i][iw[i]];
      }
   }

   % RESIDUAL PLOTS -

   if(res) resplot(yrng,res,indx,indx_ln,ary_sze,rsym,rcol,bkg);

   p_reset;
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define write_plot_head(fph,i,plt_data,plt_res,plt_mod)
{
   variable lf = length(pd.file[i]),ilf=0;

   () = fprintf(fph,"# Plot Type : %s \n", pd.type);
   () = fprintf(fph,"# \n");
   () = fprintf(fph,"# X-axis    : %s \n", pd.xaxis);
   () = fprintf(fph,"# Y-Axis    : %s \n", pd.yaxis);
   () = fprintf(fph,"# Residual  : %s \n", pd.raxis);
   () = fprintf(fph,"# \n");

   if(vzero != NULL)
   {
      () = fprintf(fph,"# Velocity Axis Zero Point: %10.4e %s \n",vzero,x_unit);
      () = fprintf(fph,"# \n");
   }
   else if(zshift[i][0] != 0)
   {
      () = fprintf(fph,"# Source Redshifted by z = %10.4e \n",zshift[i][0]);
      () = fprintf(fph,"# \n");
   }

   if(lf>1 && plt_data>0)
   {
      () = fprintf(fph,"# Combined data comprised of data indices: \n# \n");
   }
   _for ilf (0,lf-1,1)
   {  
      () = fprintf(fph,"# Index- %3i,  Model Plotted- %s,  Data File: %s,   Scale: %10.4e \n", 
                       pd.index[i][ilf], pd.model_on[i][0], pd.file[i][ilf], rescale[i][ilf]);
   }
   () = fprintf(fph,"# \n");

   variable ans=" no";
   if(bkg[i][0]>0) ans="yes";
   if(bkg[i][0]<0) ans="background only";
   () = fprintf(fph,"# Background included in data and model-   %s \n",ans);
   () = fprintf(fph,"# \n");

   if(plt_data==1 || plt_res==1) % Data and/or residuals plot
   {
      ()=fprintf(fph,"# XAXIS_COLS   :   1 set  (X 2 columns - bin_lo, bin_hi)\n");
   }
   else if (plt_data==0 && plt_mod==1)  % Model only plot
   {
      ()=fprintf(fph,"# XAXIS_COLS   :   1 set  (X 2 columns - mbin_lo, mbin_hi)\n");
   }

   variable nr=length(pd.res[i]);
   variable rs="s";
   if(nr==1){rs=" ";}

   variable xhead="#         X_LO          X_HI";
   variable dhead="    DATA_VALUE    DATA_ERROR";
   variable mhead="   MODEL_VALUE";
   variable rhead="     RES_VALUE   RES_VAL-ERR   RES_VAL+ERR";
 
   % The four possible plot cases are 1) We plot only data and
   % residuals, 2) we plot only residuals, 3) we plot only model, 4)
   % we plot data, model, and residuals

   if(plt_data)
   {
      ()=fprintf(fph,"# DATA_COLS    :   1 set  (X 2 columns - Data, Error)\n");
   }

   if(plt_mod==1 && mcol[i][0] > 0 && bkg[i][0] >= 0 && pd.con_mod >0)
   {
      ()=fprintf(fph,"# MODEL_COLS   :   1 set  (X 1 column  - Value) \n");
   }

   if(plt_res==1 && nr>0)
   {
      ()=fprintf(fph,"# RESIDUAL_COLS: %3i set"+rs+" (X 3 columns - Mean, Mean-1Sigma, Mean+1Sigma)\n",nr);
   }

   () = fprintf(fph,"# \n"+xhead);

   if(plt_data)
   {
      () = fprintf(fph,dhead);
   }

   if(plt_mod==1 && mcol[i][0]>0 && bkg[i][0] >= 0 
                 && (pd.con_mod>0 || (plt_res==0 && plt_data==0)))
   {
      ()=fprintf(fph,mhead);
   }        

   if(plt_res==1 && nr>0)
   {
      loop(nr)
      {
         () = fprintf(fph,rhead);
      }
   }

   () = fprintf(fph,"\n");
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define write_plot()
%!%+
%\function{write_plot}
%\synopsis{Write the data from the last plot made with the isis_fancy_plots package }
%\usage{[pd =] write_plot(root; head=0, data);}
%\qualifiers{
%\qualifier{head}{==0 to suppress header in output file}
%\qualifier{data}{if set, return data as a structure instead}
%}
%\description
%
% [pd =] write_plot(root; head=0, data);
%
% Creates ASCII files with the data from the last plot made (with any
% plot device) using plot_counts, plot_data, plot_residuals, or
% plot_unfold.  Data will be stored in columns in the following order-
% bin_lo, bin_hi, data_value, data_error (single column), model,
% {mean residual, residual-1sigma, residual+1sigma}.  There can be
% multiple sets of columns for the groups of residuals if the data are
% combined, but the residuals are not.
%
% Important notes:
%
%   The X-axis and Y-axis units will reflect those of the plot. The
%   binning and noticed data ranges will reflect data filters applied
%   within ISIS, but not the (usually more limited) plot range filters.
%
%   Each uncombined data set, or group of combined data sets, will be
%   be written to a separate ASCII file.  A file for combined data sets
%   can have multiple columns for residuals if they are left uncombined.
%
%   Combined data or residuals will only write out the chosen
%   combinations, not the individual pieces used to create them.
%
%   Files created from plot_residuals will only output the bin_lo/hi
%   grid and the residual/-/+ columns, not the data or model.
%
%   If the model or residuals were not plotted, they will not be
%   written to the ASCII file.
%
%   Plots made from plot_unfold will create separate files for the
%   data and the unfolded model, as they can be on different grids.
%   The model file will begin with the appropriate bin_lo, bin_hi.
%
% Inputs:
%
%   root- A string with the base file name.  Outputs will be--
%         root_#.dat, root_#.res (for plot residuals), and root_#.mod (for
%         plot_unfold, if the data and model are on separate grids). # will
%         will cycle from 0 to the number of input data groups-1, assigning
%         data sets in the order that they were input to the plot command.
%         ***Old files will be overwritten if 'root' is not a unique name.***
%
%   head- Set as a qualifier only.  If !=0 (default=1), the ASCII data
%         files will have useful header information appended, otherwise,
%         the files will just contain columns of data.
%
%   data- If the data qualifier exists, instead of writing an ASCII file, the 
%         write_plot function will return a structure with the data values from 
%         the last plot generated.
%\seealso{pg_info}
%!%-
{
   variable root;

#ifexists COPY
   if(qualifier_exists("data")) return COPY(pd);
#else
   if(qualifier_exists("data")) return pd;
#endif
   switch(_NARGS)
   {
      case 1:
      root = ();
      root = string(root);
   }
   {
      ()=fprintf(fp, "%s\n", %{{{
`
 [pd =] write_plot(root; head=0, data);

 Creates ASCII files with the data from the last plot made (with any
 plot device) using plot_counts, plot_data, plot_residuals, or
 plot_unfold.  Data will be stored in columns in the following order-
 bin_lo, bin_hi, data_value, data_error (single column), model,
 {mean residual, residual-1sigma, residual+1sigma}.  There can be
 multiple sets of columns for the groups of residuals if the data are
 combined, but the residuals are not.

 Important notes:

   The X-axis and Y-axis units will reflect those of the plot. The
   binning and noticed data ranges will reflect data filters applied
   within ISIS, but not the (usually more limited) plot range filters.

   Each uncombined data set, or group of combined data sets, will be
   be written to a separate ASCII file.  A file for combined data sets
   can have multiple columns for residuals if they are left uncombined.

   Combined data or residuals will only write out the chosen
   combinations, not the individual pieces used to create them.

   Files created from plot_residuals will only output the bin_lo/hi
   grid and the residual/-/+ columns, not the data or model.

   If the model or residuals were not plotted, they will not be
   written to the ASCII file.

   Plots made from plot_unfold will create separate files for the
   data and the unfolded model, as they can be on different grids.
   The model file will begin with the appropriate bin_lo, bin_hi.

 Inputs:

   root- A string with the base file name.  Outputs will be--
         root_#.dat, root_#.res (for plot residuals), and root_#.mod (for
         plot_unfold, if the data and model are on separate grids). # will
         will cycle from 0 to the number of input data groups-1, assigning
         data sets in the order that they were input to the plot command.
         ***Old files will be overwritten if 'root' is not a unique name.***

   head- Set as a qualifier only.  If !=0 (default=1), the ASCII data
         files will have useful header information appended, otherwise,
         the files will just contain columns of data.

   data- If the data qualifier exists, instead of writing an ASCII file, the 
         write_plot function will return a structure with the data values from 
         the last plot generated.
`                  );        %}}}

      return;
   }

   variable head = qualifier("head", 1);

   variable ndata=0, nres=0, i, j, fpd;
   if( pd.dval[0] != NULL ) ndata = length(pd.dval);
   if( pd.res [0] != NULL ) nres  = length(pd.res);

   _for i (0, max([ndata,nres])-1, 1)
   {
      variable columns = { pd.dlo[i], pd.dhi [i] };
      ifnot(pd.type=="plot_residuals")
      {
	 list_append(columns, pd.dval[i]);
	 list_append(columns, pd.derr[i]);
      }
      if( ( pd.type=="plot_counts" || pd.type=="plot_data"
	   || (pd.type=="plot_unfold" && pd.con_mod) )
	  && pd.mlo[i][0]!=NULL )
         list_append(columns, pd.mval[i]);

      _for j (0, length(pd.res[i])-1, 1)
         if(pd.res[i][j][0] != NULL)
         {
	    list_append(columns, pd.res  [i][j]);
	    list_append(columns, pd.res_m[i][j]);
	    list_append(columns, pd.res_p[i][j]);
         }

      switch(pd.type)
      { case "plot_counts" or case "plot_data" or case "plot_unfold":
         fpd = fopen("${root}_${i}.dat"$, "w");
         if(head)  write_plot_head(fpd,i,1,1,1);  % data, residuals, model
      }
      { case "plot_residuals":
         fpd = fopen("${root}_${i}.res"$, "w");
         if(head)  write_plot_head(fpd,i,0,1,0);  % only residuals
      }
      writecol(fpd, __push_list(columns));
      () = fclose(fpd);

      if(pd.type=="plot_unfold" && pd.mlo[i][0]!=NULL && not pd.con_mod)
      {
	 fpd = fopen("${root}_${i}.mod"$, "w");
	 if(head)  write_plot_head(fpd,i,0,0,1);  % model only
	 writecol(fpd, pd.mlo[i], pd.mhi[i], pd.mval[i]);
	 () = fclose(fpd);
      }
   }
}

alias("write_plot","writeplot");

%%%%%%%%%%%%%%%%%%%%%%%%%

public define plot_data()
%!%+
%\function{plot_data}
%\synopsis{Plot counts per unit per second (isis_fancy_plots package)}
%\description
%
%plot_data({indx,[arry],{cid}},pstruct);  % pstruct = struct{ dcol, mcol, rcol, ...}
%plot_data({indx,[arry],{cid}};dcol={val},mcol={val},rcol={val,[arry],val},...);
%
%  Plot background subtracted data, model, and residuals as counts/xunit/sec.
%  Residuals are units of chi, chi2, or ratio, and will be based upon whether
%  one chooses sigma=model, data, or gehrels in set_fit_statistic();
%  (data error bars are only affected by the latter two).
%  set_fit_statistic("cash"); will alter the residuals to the Cash statistic.
%  set_fit_statistic("ml"); will alter the residuals to the Maximum Likelihood statistic.
%
%  Options below refer to structure variables/qualifiers
%
%   indx    = list of data set indices to be plotted. Any indices grouped in
%             an array within that list will be *combined* in the data plot.
%             Single number in list is combo id, {#} = [combination_members(#)].
%   dcol    = (pgplot) color value for data (or list of color values)
%   decol   = (pgplot) color value for data error bars (or list of color values)
%   mcol    = (pgplot) color value for model (or list of color values)
%             0 => No model plotted
%   rcol    = (pgplot) color value for residuals (or list of color values; arrays
%             within the list allow for individual color values if portions of
%             the data are combined, but their associated residuals are not)
%   recol   = color for residual error bars (or list of color values; arrays
%             within the list act as for residual color inputs)
%   dsym    = (pgplot) symbol value for data (or list of symbol values)
%             0 => histogram plot
%   rsym    = (pgplot) symbol value for residuals (or list of symbol values;
%             arrays within the list act as for residual color inputs)
%             0 => histogram plot
%   xrange  = List of X-limits for the data & model & residuals
%             Note: Any X- or Y-range set to NULL is autoscaled
%   yrange  = List of Y-limits for the data & model and (optionally) residuals
%   oplt    = 0 (default) for new plot, !=0 for overplotting
%   no_reset= 0 (default)- plots *will* be reset, i.e., next plot moves to new pane
%             (multiplot), next plot redraws window (single plot). no_reset=1 necessary
%             for overplotting multiplots (oplt=1 sufficient for single plots).
%   res     = 0 (default), no residuals; 1, 2, or 3 = chi, chi2, or ratio residuals
%             4, 5, or 6 = chi, chi2, or ratio, but combine residuals for combined data
%             set_fit_method("cash"); or set_fit_method("ml") will cause res=(2 or 5) 
%             or res=(1 or 4) to plot the residual for the Cash or Maximum Likelihood 
%             statistic or its square root, respectively
%   bkg     = List of 0's (subtract background-default), 1's (include backgrounds),
%             or -1's (plot *only* the background [no model plotted in this case]).
%             Ratio residuals will include background in data/model, other residuals
%             are unaffected. Indices within a combination are treated the same.)
%   xlabel  = String that will overwrite default X-axis label (default=NULL)
%   ylabel  = String or string array that will overwrite the default Y-axis labels
%             (second element of array applies to residuals; default=NULL)
%   zshift  = List of redshifts to be applied to the data (default zshift={0,0,...})
%   vzero   = If set, the reference X-unit value to be defined as zero velocity.
%             The X-axis then becomes a velocity axis (km/s) referenced to this
%             point (default vzero=NULL; setting vzero/zaxis supersedes zshift)
%   zaxis   = If not 0, use a redshift axis instead of a velocity axis *if* vzero
%             is defined (default zaxis=0)
%   scale   = Multiplicatively scale the Y-axis by the values in a list.
%             Any arrays in the list should hold the individual scalings for
%             data set arrays in the input index list.  **Note:**  these values
%             only scale the plots, not the fits.  (Default values are 1.)
%   sum_exp = If==1, then when combining data sets, sum the exposure times (as
%             opposed to using the mean exposure time; default sum_exp=1).
%   gap     =  1 (default), models are histograms with gaps where data has gaps,
%              0          , model are bin-centered lines, without gaps.
%\seealso{plot_data, plot_unfold, plot_residuals, plot_fit_model, plotxy, plot_comps, plot_double}
%!%-
{
   variable i,args,indx_ln,iindx,ary_sze;
   variable data,bin_lo,bin_hi,modl,pres,mres,resd,ers,kcor=1.,perunit,texp;
   variable iw;
   variable ja,jb;

   if(_NARGS > 0)
   {
      args = __pop_list(_NARGS);
      (indx, indx_ln, ary_sze) = set_plot_parms(_NARGS,args;;__qualifiers());
      power=1; % No multiplying by X for plot_data (only plot_counts & plot_unfold)
      ifnot(indx_ln)
      {
         message("\n  One or more data sets undefined.\n");
         return;
      }
   }
   else
   {
      () = fprintf(fp, "%s\n", %{{{
`
plot_data({indx,[arry],{cid}},pstruct);  % pstruct = struct{ dcol, mcol, rcol, ...}
plot_data({indx,[arry],{cid}};dcol={val},mcol={val},rcol={val,[arry],val},...);

  Plot background subtracted data, model, and residuals as counts/xunit/sec.
`
+ores_descript
+oopts+oindx+odcol+odecol+omcol+orcol+orecol+odsym+orsym+oxrange+oyrange+ooplt+onrst+ores
+obkg+oxlabel+oylabel+ozshift+ovzero+ozaxis+oscale+osum_exp+ogap); %}}}
      return;
   }

   %  And let the plotting begin ...

   plopt = p_range_label(res,0);

   % THE DATA & MODEL:

   pd_set(indx,indx_ln,ary_sze);
   pd.type="plot_data";

   % Create the data first, in case the X or Y ranges are set to
   % autoscale.  This way we can get the autoscaling correct for the
   % full range of data

   texp = Double_Type[indx_ln]; bin_lo = Array_Type[indx_ln];  bin_hi = @bin_lo; 
   data = @bin_lo; modl = @bin_lo; ers = @bin_lo; iw = @bin_lo;
   variable all_lo=Double_Type[0], all_hi=@all_lo, all_data_lo=@all_lo, 
            all_data_hi=@all_lo, istrt=0;

   _for i (0,indx_ln-1,1)
   {
      (bin_lo[i],bin_hi[i],data[i],modl[i],ers[i],iw[i],perunit,,,,texp[i]) = 
               make_data(indx[i],res,power,mcol[i],bkg[i],rescale[i]);

      if(vzero==NULL && zshift[i]!=0 && unit_is_energy(x_unit)!=1)
         kcor = 1.+zshift[i];

      data[i] *= kcor/perunit;
       ers[i] *= kcor/perunit;
      modl[i] *= kcor/perunit;

       if(vzero!=NULL || zshift[i]!=0.)
      {
         (bin_lo[i],bin_hi[i]) = vaxis(i,bin_lo[i],bin_hi[i]);
      }

      all_lo = [all_lo,bin_lo[i][iw[i]]]; 
      all_hi = [all_hi,bin_hi[i][iw[i]]]; 
      all_data_lo = [all_data_lo,data[i][iw[i]]-ers[i][iw[i]]]; 
      all_data_hi = [all_data_hi,data[i][iw[i]]+ers[i][iw[i]]];
   }

   istrt = start_plot(all_lo, all_hi, all_data_lo, all_data_hi);

   _for i (0,indx_ln-1,1)
   { 
      % Start the plots by plotting data points.  Choose mid-points of
      % bins, to ensure autoscaling matches between data & residuals

      if( dsym[i][0] == 0 )
      {
         pointstyle(1);
         point_size(0.1);
      }
      else
      {
         pointstyle(dsym[i][0]);
         point_size(plopt.point_size);
      }
      connect_points(0);
      ifnot(min([i+istrt+oplt,1]))   % In case we haven't started a plot yet
      {
         slangref = refplt[0];
         @slangref((bin_lo[i][iw[i]]+bin_hi[i][iw[i]])/2.,data[i][iw[i]],dcol[i][0]);
      }
      point_size(plopt.point_size);

      pd.xaxis = get_plot_options().xlabel;
      pd.yaxis = get_plot_options().ylabel;
      pd.dlo[i] = bin_lo[i][iw[i]];
      pd.dhi[i] = bin_hi[i][iw[i]];
      pd.dval[i] = data[i][iw[i]];
      pd.derr[i] = ers[i][iw[i]];
      pd.exp[i] = texp[i];

      datplot_err(bin_lo[i][iw[i]],bin_hi[i][iw[i]],data[i][iw[i]],
                  data[i][iw[i]]-ers[i][iw[i]],data[i][iw[i]]+ers[i][iw[i]],
                  de_width,decol[i][0],0);

      set_line_width(d_width);

      if(dsym[i][0] != 0)       % Symbol plots
      {
         connect_points(0);
         pointstyle(dsym[i][0]);
         slangref = refplt[1];
         @slangref( (bin_lo[i][iw[i]]+bin_hi[i][iw[i]])/2., data[i][iw[i]], dcol[i][0] );
      }
      else                      % Histogram plots   
      {  
         pointstyle(1);
         connect_points(-1);
         (ja,jb) = hp_loop(iw[i]);
         hp_loopit( bin_lo[i], bin_hi[i], data[i], dcol[i][0], iw[i], ja, jb, 1);
      }

      if(mcol[i][0] != 0 and bkg[i][0] >=0)
      {
         set_line_width(m_width);  connect_points(-1);
         (ja,jb) = hp_loop(iw[i]);
         hp_loopit(bin_lo[i], bin_hi[i], modl[i], mcol[i][0], iw[i], ja, jb, gap[i]); 
         pd.model_on[i] = ["     yes"];
         pd.mlo[i] = bin_lo[i][iw[i]];
         pd.mhi[i] = bin_hi[i][iw[i]];
         pd.mval[i] = modl[i][iw[i]];
      }
   }

   % RESIDUAL PLOTS -

   if(res) resplot(yrng,res,indx,indx_ln,ary_sze,rsym,rcol,bkg);

   p_reset;
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define plot_residuals()
%!%+
%\function{plot_residuals}
%\synopsis{Plot the data residuals (isis_fancy_plots package)}
%\description
%
%plot_residuals({indx,[arry],{cid}},pstruct);  % pstruct = struct{ rcol, rsym, ...}
%plot_residuals({indx,[arry],{cid}};dcol={val},rcol={val,[arry],val},rsym=...);
%
%  Plot data residuals, without plotting the data.  Residuals related to data
%  indices in [arry] will appear in a single ascii file if write_plot(); is
%  used, and will be combined in the plot if res>3 is chosen.
%
%  Options below refer to structure variables/qualifiers
%
%   indx    = list of data set indices to be plotted. Any indices grouped in
%             an array within that list will be *combined* in the data plot.
%             Single number in list is combo id, {#} = [combination_members(#)].
%   rcol    = (pgplot) color value for residuals (or list of color values; arrays
%             within the list allow for individual color values if portions of
%             the data are combined, but their associated residuals are not)
%   recol   = color for residual error bars (or list of color values; arrays
%             within the list act as for residual color inputs)
%   rsym    = (pgplot) symbol value for residuals (or list of symbol values;
%             arrays within the list act as for residual color inputs)
%             0 => histogram plot
%   xrange  = List of X-limits for the data & model & residuals
%             Note: Any X- or Y-range set to NULL is autoscaled
%   yrange  = List of Y-limits for the data & model and (optionally) residuals
%   oplt    = 0 (default) for new plot, !=0 for overplotting
%   no_reset= 0 (default)- plots *will* be reset, i.e., next plot moves to new pane
%             (multiplot), next plot redraws window (single plot). no_reset=1 necessary
%             for overplotting multiplots (oplt=1 sufficient for single plots).
%   res     = 0 (default), no residuals; 1, 2, or 3 = chi, chi2, or ratio residuals
%             4, 5, or 6 = chi, chi2, or ratio, but combine residuals for combined data
%             set_fit_method("cash"); or set_fit_method("ml") will cause res=(2 or 5) 
%             or res=(1 or 4) to plot the residual for the Cash or Maximum Likelihood 
%             statistic or its square root, respectively
%   xlabel  = String that will overwrite default X-axis label (default=NULL)
%   ylabel  = String or string array that will overwrite the default Y-axis labels
%             (second element of array applies to residuals; default=NULL)
%   zshift  = List of redshifts to be applied to the data (default zshift={0,0,...})
%   vzero   = If set, the reference X-unit value to be defined as zero velocity.
%             The X-axis then becomes a velocity axis (km/s) referenced to this
%             point (default vzero=NULL; setting vzero/zaxis supersedes zshift)
%   zaxis   = If not 0, use a redshift axis instead of a velocity axis *if* vzero
%             is defined (default zaxis=0)
%\seealso{plot_counts, plot_data, plot_unfold, plot_fit_model, plotxy, plot_comps, plot_double}
%!%-
{
   variable args,indx_ln,iindx,ary_sze;

   power_df=1; 

   if(_NARGS > 0)
   {
      args = __pop_list(_NARGS);
      (indx, indx_ln, ary_sze) = set_plot_parms(_NARGS,args;;__qualifiers());
      if(length(yrng)==2) yrng={NULL,NULL,yrng[0],yrng[1]}; 
      if(length(yrng)==1) yrng={NULL,NULL,yrng[0],NULL}; 
      ifnot(indx_ln)
      {
         message("\n  One or more data sets undefined.\n");
         return;
      }
   }
   else
   {
      () = fprintf(fp, "%s\n", %{{{
`
plot_residuals({indx,[arry],{cid}},pstruct);  % pstruct = struct{ rcol, rsym, ...}
plot_residuals({indx,[arry],{cid}};dcol={val},rcol={val,[arry],val},rsym=...);

  Plot data residuals, without plotting the data.  Residuals related to data
  indices in [arry] will appear in a single ascii file if write_plot(); is
  used, and will be combined in the plot if res>3 is chosen.
`
+oopts+oindx+orcol+orecol+orsym+oxrange+oyrange+ooplt+onrst+ores
+oxlabel+oylabel+ozshift+ovzero+ozaxis);%}}}
      return;
   }

   %  And let the plotting begin ...

   plopt = p_range_label(0,0);

   % THE DATA & MODEL:

   pd_set(indx,indx_ln,ary_sze);
   pd.type="plot_residuals";

   if(res) resplot(yrng,res,indx,indx_ln,ary_sze,rsym,rcol,bkg);

   p_reset;
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define plot_fit_model()
%!%+
%\function{plot_fit_model}
%\synopsis{Plot background subtracted model as counts/xunit/sec (isis_fancy_plots package).}
%\description
%
%plot_fit_model({indx,[arry],{cid}},pstruct);  % pstruct = struct{ mcol, ...}
%plot_fit_model({indx,[arry],{cid}};mcol={val},...);
%
%  Plot background subtracted model as counts/xunit/sec.
%
%  Options below refer to structure variables/qualifiers
%
%   indx    = list of data set indices to be plotted. Any indices grouped in
%             an array within that list will be *combined* in the data plot.
%             Single number in list is combo id, {#} = [combination_members(#)].
%   mcol    = (pgplot) color value for model (or list of color values)
%             0 => No model plotted
%   xrange  = List of X-limits for the data & model & residuals
%             Note: Any X- or Y-range set to NULL is autoscaled
%   yrange  = List of Y-limits for the data & model and (optionally) residuals
%   oplt    = 0 (default) for new plot, !=0 for overplotting
%   no_reset= 0 (default)- plots *will* be reset, i.e., next plot moves to new pane
%             (multiplot), next plot redraws window (single plot). no_reset=1 necessary
%             for overplotting multiplots (oplt=1 sufficient for single plots).
%   bkg     = List of 0's (subtract background-default), 1's (include backgrounds),
%             or -1's (plot *only* the background [no model plotted in this case]).
%             Ratio residuals will include background in data/model, other residuals
%             are unaffected. Indices within a combination are treated the same.)
%   xlabel  = String that will overwrite default X-axis label (default=NULL)
%   ylabel  = String or string array that will overwrite the default Y-axis labels
%             (second element of array applies to residuals; default=NULL)
%   zshift  = List of redshifts to be applied to the data (default zshift={0,0,...})
%   vzero   = If set, the reference X-unit value to be defined as zero velocity.
%             The X-axis then becomes a velocity axis (km/s) referenced to this
%             point (default vzero=NULL; setting vzero/zaxis supersedes zshift)
%   zaxis   = If not 0, use a redshift axis instead of a velocity axis *if* vzero
%             is defined (default zaxis=0)
%   scale   = Multiplicatively scale the Y-axis by the values in a list.
%             Any arrays in the list should hold the individual scalings for
%             data set arrays in the input index list.  **Note:**  these values
%             only scale the plots, not the fits.  (Default values are 1.)
%   sum_exp = If==1, then when combining data sets, sum the exposure times (as
%             opposed to using the mean exposure time; default sum_exp=1).
%   gap     =  1 (default), models are histograms with gaps where data has gaps,
%              0          , model are bin-centered lines, without gaps.
%\seealso{plot_counts; plot_data, plot_unfold, plot_residuals, plotxy, plot_comps, plot_double}
%!%-
{
   variable i,args,indx_ln,iindx,ary_sze;
   variable bin_lo,bin_hi,modl,kcor=1,perunit,texp;
   variable iw;
   variable ja,jb;

   power_df=1; 

   if(_NARGS > 0)
   {
      args = __pop_list(_NARGS);
      (indx, indx_ln, ary_sze) = set_plot_parms(_NARGS,args;;__qualifiers());
      ifnot(indx_ln)
      {
         message("\n  One or more data sets undefined.\n");
         return;
      }
   }
   else
   {
      () = fprintf(fp, "%s\n", %{{{
`
plot_fit_model({indx,[arry],{cid}},pstruct);  % pstruct = struct{ mcol, ...}
plot_fit_model({indx,[arry],{cid}};mcol={val},...);

  Plot background subtracted model as counts/xunit/sec.
`
+oopts+oindx+omcol+oxrange+oyrange+ooplt+onrst+obkg+oxlabel+oylabel+ozshift+ovzero
+ozaxis+oscale+osum_exp+ogap); %}}}
      return;
   }

   %  And let the plotting begin ...

   plopt = p_range_label(res,0);

   % Create the data first, in case the X or Y ranges are set to
   % autoscale.  This way we can get the autoscaling correct for the
   % full range of data

   texp = Double_Type[indx_ln]; bin_lo=Array_Type[indx_ln];  bin_hi=@bin_lo; modl=@bin_lo; iw=@bin_lo; 
   variable all_lo=Double_Type[0], all_hi=@all_lo, all_modl=@all_lo, istrt=0;

   _for i (0,indx_ln-1,1)
   {
      (bin_lo[i],bin_hi[i],,modl[i],,iw[i],perunit,,,,texp[i]) = 
               make_data(indx[i],res,power,mcol[i],bkg[i],rescale[i]);

      if(vzero==NULL && zshift[i]!=0 && unit_is_energy(x_unit)!=1)
         kcor = 1.+zshift[i];

      modl[i] *= kcor/perunit;

      if(vzero!=NULL || zshift[i]!=0.)
      {
         (bin_lo[i],bin_hi[i]) = vaxis(i,bin_lo[i],bin_hi[i]);
      }

      all_lo = [all_lo,bin_lo[i][iw[i]]]; 
      all_hi = [all_hi,bin_hi[i][iw[i]]]; 
      all_modl = [all_modl,modl[i][iw[i]]];
   }

   istrt = start_plot(all_lo, all_hi, all_modl, all_modl);
   variable istrtII = 0;  % The first model could be 0'd out

   _for i (0,indx_ln-1,1)
   { 
      if(mcol[i][0] != 0)
      {
         pointstyle(1);
         point_size(0.1);
         connect_points(0);
         ifnot(min([istrt+istrtII+oplt,1]));
         {
            slangref = refplt[0];
            @slangref((bin_lo[i][iw[i]]+bin_hi[i][iw[i]])/2.,modl[i][iw[i]],mcol[i][0]);
            istrtII=1;
         }
         point_size(plopt.point_size);
         set_line_width(m_width);  connect_points(-1);
         (ja,jb) = hp_loop(iw[i]);
         hp_loopit(bin_lo[i], bin_hi[i], modl[i], mcol[i][0], iw[i], ja, jb, gap[i]); 
      }
   }
   p_reset;
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define make_flux(iindx,res,power,modc,bkgon,rscl,zshft)
{
   variable bin_hi, bin_lo, mbin_lo=NULL, mbin_hi=NULL, frst_modl, fmodl, 
            iwm=Integer_Type[0], iw, iwv, powm, jj, ja, jb,
            tdata=0., tmc = 0, tmodl=0., ters=0., twght=0., texp,
            wght, gdi, gdcu, gdcb, kcor=1., kp,i;

   (bin_lo,bin_hi,tdata,tmc,ters,iw, , , , , texp) = 
      make_data(iindx,res,power,modc,bkgon,rscl);

   _rev = _revr[0];
   kp=3-power;

   if(unit_is_energy(x_unit))
   {
      _rev = _revr[1];
      kp=(power-1);
   }

   if(vzero==NULL) kcor=(1.+zshft);
   kcor = kcor^kp;

   _for i (0,length(iindx)-1,1)
   {
      flux_corr(iindx[i]);                       % Flux correct the data
      wght = get_flux_corr_weights(iindx[i]);    % Flux correction weights

#ifexists get_bin_corr_factor
   % --- Added by Remeis group, Thomas Duaser: new; for FERMI data ---
     if (get_bin_corr_factor(iindx[i]) != NULL)
	wght *= @_rev(get_bin_corr_factor(iindx[i]));
#endif
 
      gdi = get_data_info(iindx[i]);             % Info about the original
      gdcb = get_data_counts(iindx[i]);          % binned & noticed data

      if(length(gdi.notice) != length(gdi.rebin))
      {
         rebin_data(iindx[i],0);
      }
      gdcu = get_data_counts(iindx[i]);          % The unbinned data

      if(length(gdi.notice) != length(gdi.rebin))
      {
         rebin_data(iindx[i],gdi.rebin);
         ignore(iindx[i]);
         notice_list(iindx[i],gdi.notice_list);  % Restore bins & bounds
      }

      % The spectral weights have to be rebinned to the data binning

      twght += rebin(gdcb.bin_lo,gdcb.bin_hi,gdcu.bin_lo,gdcu.bin_hi,
                     wght)/(gdcb.bin_hi-gdcb.bin_lo);

      if( modc != 0 )
      {
         fmodl = get_model_flux(iindx[i]);

         ifnot(i)       % Only do the bins for the first data index
         {
            frst_modl = @fmodl;

            if(unit_is_energy(x_unit))
            {
               mbin_lo = _A(fmodl.bin_hi)/unit_info(x_unit).scale;
               mbin_hi = _A(fmodl.bin_lo)/unit_info(x_unit).scale;
            }
            else
            {
               mbin_lo = fmodl.bin_lo/unit_info(x_unit).scale;
               mbin_hi = fmodl.bin_hi/unit_info(x_unit).scale;
            }

            %% Need to be more clever here.  Temporary (i.e., semi-
            %% permanent) kludge to have model gaps follow data gaps
  
            (ja,jb) = hp_loop(iw);
            _for jj (0,length(ja)-1,1)
            {
               iwv = iw[[ja[jj]:jb[jj]]]; 
               iwm = [ iwm, where( (mbin_lo >= min(bin_lo[iwv])) and
                                   (mbin_hi <= max(bin_hi[iwv]))  ) ];
            }

            powm = power_scale(mbin_lo,mbin_hi,power);
         }
         else             % All other models follow the first binning
         {
            fmodl.value=rebin(frst_modl.bin_lo,frst_modl.bin_hi,
                              fmodl.bin_lo,fmodl.bin_hi,fmodl.value);
         }
 
         fmodl.value = @_rev( fmodl.value );
         fmodl.value = fmodl.value / (mbin_hi-mbin_lo);
         fmodl.value[iwm] = fmodl.value[iwm]*powm[iwm];

         tmodl += rscl[i]*fmodl.value;
      }
   }
   twght = @_rev(twght);

   iw = iw[where(twght[iw]>0)];  % avoid infinities from zero weights

   variable factor1 = kcor * y_scl_norm[x_unit+y_unit];
   variable factor2 = factor1 / twght / (bin_hi-bin_lo);

   return bin_lo, bin_hi, factor2*tdata, factor2*ters, iw,             % data
          texp, factor2,                                               % the weights
          use_con_flux                                                 % &  either
          ? ( bin_lo, bin_hi, factor2*tmc, iw )                        %    convolved model
          : ( mbin_lo, mbin_hi, [factor1*tmodl/length(iindx)], iwm );  % or bare model
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define plot_unfold()
%!%+
%\function{plot_unfold}
%\synopsis{Plot flux-corrected spectra (isis_fancy_plots package).}
%\description
%
%plot_unfold({indx,[arry],{cid}},pstruct);  % pstruct = struct{ dcol, mcol, rcol, ...}
%plot_unfold({indx,[arry],{cid}};dcol={val},mcol={val},rcol={val,[arry],val},...);
%
%  Plot background subtracted unfolded data, model, and residuals using a
%  variety of X- and Y-units set by fancy_plot_unit(xunit, [yunit]);
%  Residuals are units of chi, chi2, or ratio, and will be based upon whether
%  one chooses sigma=model, data, or gehrels in set_fit_statistic();
%  (data error bars are only affected by the latter two).
%  set_fit_statistic("cash"); will alter the residuals to the Cash statistic.
%  set_fit_statistic("ml"); will alter the residuals to the Maximum Likelihood statistic.
%
%  Options below refer to structure variables/qualifiers
%
%   indx    = list of data set indices to be plotted. Any indices grouped in
%             an array within that list will be *combined* in the data plot.
%             Single number in list is combo id, {#} = [combination_members(#)].
%   dcol    = (pgplot) color value for data (or list of color values)
%   decol   = (pgplot) color value for data error bars (or list of color values)
%   mcol    = (pgplot) color value for model (or list of color values)
%             0 => No model plotted
%   rcol    = (pgplot) color value for residuals (or list of color values; arrays
%             within the list allow for individual color values if portions of
%             the data are combined, but their associated residuals are not)
%   recol   = color for residual error bars (or list of color values; arrays
%             within the list act as for residual color inputs)
%   dsym    = (pgplot) symbol value for data (or list of symbol values)
%             0 => histogram plot
%   rsym    = (pgplot) symbol value for residuals (or list of symbol values;
%             arrays within the list act as for residual color inputs)
%             0 => histogram plot
%   xrange  = List of X-limits for the data & model & residuals
%             Note: Any X- or Y-range set to NULL is autoscaled
%   yrange  = List of Y-limits for the data & model and (optionally) residuals
%   oplt    = 0 (default) for new plot, !=0 for overplotting
%   no_reset= 0 (default)- plots *will* be reset, i.e., next plot moves to new pane
%             (multiplot), next plot redraws window (single plot). no_reset=1 necessary
%             for overplotting multiplots (oplt=1 sufficient for single plots).
%   res     = 0 (default), no residuals; 1, 2, or 3 = chi, chi2, or ratio residuals
%             4, 5, or 6 = chi, chi2, or ratio, but combine residuals for combined data
%             set_fit_method("cash"); or set_fit_method("ml") will cause res=(2 or 5) 
%             or res=(1 or 4) to plot the residual for the Cash or Maximum Likelihood 
%             statistic or its square root, respectively
%   power  = 0, 1 (usual default), 2 (default for mJy) or 3 (default for ergs/Watts
%            vs. energy units)=> photons/cm^2/s/xunit *(1/xunit,1,xunit,xunit^2)
%   bkg     = List of 0's (subtract background-default), 1's (include backgrounds),
%             or -1's (plot *only* the background [no model plotted in this case]).
%             Ratio residuals will include background in data/model, other residuals
%             are unaffected. Indices within a combination are treated the same.)
%   xlabel  = String that will overwrite default X-axis label (default=NULL)
%   ylabel  = String or string array that will overwrite the default Y-axis labels
%             (second element of array applies to residuals; default=NULL)
%   zshift  = List of redshifts to be applied to the data (default zshift={0,0,...})
%   vzero   = If set, the reference X-unit value to be defined as zero velocity.
%             The X-axis then becomes a velocity axis (km/s) referenced to this
%             point (default vzero=NULL; setting vzero/zaxis supersedes zshift)
%   zaxis   = If not 0, use a redshift axis instead of a velocity axis *if* vzero
%             is defined (default zaxis=0)
%   scale   = Multiplicatively scale the Y-axis by the values in a list.
%             Any arrays in the list should hold the individual scalings for
%             data set arrays in the input index list.  **Note:**  these values
%             only scale the plots, not the fits.  (Default values are 1.)
%   gap     =  1 (default), models are histograms with gaps where data has gaps,
%              0          , model are bin-centered lines, without gaps.
%   con_mod= 1 (default), the smear the model by the detector response, otherwise 
%            plot the unsmeared model at the internal resolution of the arf
%
%  Note: Model flux is: ( \\\\int dE S(E) )/dE, while data is 
%  (C(h) - B(h))/(\\\\int R(h,E) A(E) dE)/dh/t, where A(E) is effective area, 
%  R(h,E) is RMF, C(h)/B(h) are total/background counts. Thus, the data 
%  and model will match best only in the limit of a delta function RMF, 
%  and in fact might look different than the residuals (which is the 
%  only proper comparison between data and model, anyhow).
%\seealso{set_power_scale, plot_counts, plot_data, plot_residuals, plot_fit_model, plotxy, plot_comps, plot_double}
%!%-
{
   variable args,i,indx, indx_ln, ary_sze,iw,iwm,ja,jb;
   variable bin_lo,bin_hi,mbin_lo,mbin_hi,data,ers,modl,twgt,texp;

   power_df=1;
   if(y_unit=="mjy")
   {
      power_df=2;
   }
   if((y_unit=="ergs" || y_unit=="watts") && unit_is_energy(x_unit))
      power_df=3;

   if(_NARGS > 0)
   {
      args = __pop_list(_NARGS);
      (indx, indx_ln, ary_sze) = set_plot_parms(_NARGS,args;;__qualifiers());
      ifnot(indx_ln)
      {
         message("\n  One or more data sets undefined.\n");
         return;
      }
   }
   else
   {
      () = fprintf(fp, "%s\n", %%%{{{
`
plot_unfold({indx,[arry],{cid}},pstruct);  % pstruct = struct{ dcol, mcol, rcol, ...}
plot_unfold({indx,[arry],{cid}};dcol={val},mcol={val},rcol={val,[arry],val},...);

  Plot background subtracted unfolded data, model, and residuals using a
  variety of X- and Y-units set by fancy_plot_unit(xunit, [yunit]);
`
+ores_descript
+oopts+oindx+odcol+odecol+omcol+orcol+orecol+odsym+orsym+oxrange+oyrange+ooplt+onrst+ores
+`   power  = 0, 1 (usual default), 2 (default for mJy) or 3 (default for ergs/Watts
            vs. energy units)=> photons/cm^2/s/xunit *(1/xunit,1,xunit,xunit^2)
`
+obkg+oxlabel+oylabel+ozshift+ovzero+ozaxis+oscale+ogap
+`   con_mod= 1 (default), the smear the model by the detector response, otherwise 
            plot the unsmeared model at the internal resolution of the arf

  Note: Model flux is: ( \int dE S(E) )/dE, while data is 
  (C(h) - B(h))/(\int R(h,E) A(E) dE)/dh/t, where A(E) is effective area, 
  R(h,E) is RMF, C(h)/B(h) are total/background counts. Thus, the data 
  and model will match best only in the limit of a delta function RMF, 
  and in fact might look different than the residuals (which is the 
  only proper comparison between data and model, anyhow).
`                  );          %%%}}}
      return;
   }

   %  And let the plotting begin ...
   
   plopt = p_range_label(res,power+1);

   % DATA/MODEL PLOTS - 

   pd_set(indx,indx_ln,ary_sze);
   pd.type="plot_unfold";

   % Create the data first, in case the X or Y ranges are set to
   % autoscale.  This way we can get the autoscaling correct for the
   % full range of data

   texp = Double_Type[indx_ln];
   bin_lo = Array_Type[indx_ln];  bin_hi = @bin_lo; twgt = @bin_lo;
   data = @bin_lo; modl = @bin_lo; ers = @bin_lo; iw = @bin_lo;
   mbin_lo = @bin_lo; mbin_hi = @bin_lo; iwm = @bin_lo;
   variable all_lo=Double_Type[0], all_hi=@all_lo, all_data_lo=@all_lo, 
            all_data_hi=@all_lo, istrt=0;

   _for i (0,indx_ln-1,1)
   {
      (bin_lo[i], bin_hi[i], data[i], ers[i], iw[i], texp[i], twgt[i],
       mbin_lo[i], mbin_hi[i], modl[i], iwm[i]       )=
         make_flux(indx[i],res,power,mcol[i],bkg[i],rescale[i],zshift[i]);

      if(vzero!=NULL || zshift[i]!=0.)
      {
         (bin_lo[i],bin_hi[i]) = vaxis(i,bin_lo[i],bin_hi[i]);
      }

      all_lo = [all_lo,bin_lo[i][iw[i]]]; 
      all_hi = [all_hi,bin_hi[i][iw[i]]];
      all_data_lo = [all_data_lo,data[i][iw[i]]-ers[i][iw[i]]];
      all_data_hi = [all_data_hi,data[i][iw[i]]+ers[i][iw[i]]];

      if(mcol[i][0] > 0 && bkg[i][0] >= 0)
      {
         if(vzero!=NULL || zshift[i]!=0.)
         {
            (mbin_lo[i],mbin_hi[i]) = vaxis(i,mbin_lo[i],mbin_hi[i]);
         }
         all_lo = [all_lo,mbin_lo[i][iwm[i]]]; 
         all_hi = [all_hi,mbin_hi[i][iwm[i]]]; 
         all_data_lo = [all_data_lo,modl[i][iwm[i]]]; 
         all_data_hi = [all_data_hi,modl[i][iwm[i]]];
      }
   }

   istrt = start_plot(all_lo, all_hi, all_data_lo, all_data_hi);

   _for i (0,indx_ln-1,1)
   { 
      % Start the plots by plotting data points.  Choose mid-points of
      % bins, to ensure autoscaling matches between data & residuals

      if( dsym[i][0] == 0 )
      {
         pointstyle(1);
         point_size(0.1);
      }
      else
      {
         pointstyle(dsym[i][0]);
         point_size(plopt.point_size);
      }
      connect_points(0);
      ifnot(min([i+istrt+oplt,1]))  % In case we haven't started a plot yet...
      {
         slangref = refplt[0];
         @slangref((bin_lo[i][iw[i]]+bin_hi[i][iw[i]])/2.,data[i][iw[i]],dcol[i][0]);
      }
      point_size(plopt.point_size);

      % Start saving data for writing to an ASCII file

      pd.xaxis = get_plot_options().xlabel;
      pd.yaxis = get_plot_options().ylabel;
      pd.dlo[i] = bin_lo[i][iw[i]];
      pd.dhi[i] = bin_hi[i][iw[i]];
      pd.dval[i] = data[i][iw[i]];
      pd.derr[i] = ers[i][iw[i]];
      pd.weight[i] = twgt[i][iw[i]];
      pd.exp[i] = texp[i];

      datplot_err(bin_lo[i][iw[i]],bin_hi[i][iw[i]],data[i][iw[i]],
                  data[i][iw[i]]-ers[i][iw[i]],data[i][iw[i]]+ers[i][iw[i]],
                  de_width,decol[i][0],0);

      set_line_width(d_width);

      if(dsym[i][0] != 0)   % Symbol plots
      {
         connect_points(0);
         pointstyle(dsym[i][0]);
         slangref = refplt[1];
         @slangref( (bin_lo[i][iw[i]]+bin_hi[i][iw[i]])/2., data[i][iw[i]], dcol[i][0] );
      }
      else                  % Histogram plots
      {  
         pointstyle(1);
         connect_points(-1);
         (ja,jb) = hp_loop(iw[i]);
         hp_loopit( bin_lo[i], bin_hi[i], data[i], dcol[i][0], iw[i], ja, jb, 1);
      }

      if(mcol[i][0] > 0 && bkg[i][0] >= 0)       % Only plot model if chosen & exists
      {
         set_line_width(m_width);  connect_points(-1);
         (ja,jb) = hp_loop(iwm[i]);
         hp_loopit(mbin_lo[i], mbin_hi[i], modl[i], mcol[i][0], iwm[i], ja, jb, gap[i]); 
         pd.model_on[i] = ["     yes"];
         pd.mlo[i] = mbin_lo[i][iwm[i]];
         pd.mhi[i] = mbin_hi[i][iwm[i]];
         pd.mval[i] = modl[i][iwm[i]];
         pd.con_mod = use_con_flux;
      }
   }

   % RESIDUAL PLOTS -

   if(res) resplot(yrng,res,indx,indx_ln,ary_sze,rsym,rcol,bkg);

   p_reset;
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                % 
%     Definitions for Simple plotxy Function     %
%                                                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%

public define plotxy()
%!%+
%\function{plotxy}
%\synopsis{Generate a simple x/y plot with error bars (isis_fancy_plots package)}
%\description
%
% plotxy(x,dxm,dxp,y,dym,dyp, pstruct); % pstruct = struct{dcol, decol, xrng, ...}
% plotxy(x,dxm,dxp,y,dym,dyp; dcol=#, decol=#, ...);
%
% Also accepts:
%
% plotxy(x,y [,pstruct;qualifiers]);
% plotxy(x,dxm,dxp,y [,pstruct;qualifiers]);
% plotxy(x,,,y,dym,dyp [,pstruct;qualifiers]);
%
%   Plot simple x,y plots with error bars: x-dxm, x+dxp, etc.
%
%   Options below refer to structure variables/associative keys/qualifiers:
%
%   dcol  = (pgplot) color value for data
%   decol = (pgplot) color value for data
%   dsym  = (pgplot) symbol value for data
%           Note: dsym=0 will *not* produce a histogram plot
%   xrange= List of X-limits for the data. Ranges previously input
%           via xrange(); will be respected if this option is not set.
%   yrange= List of Y-limits for the data. Ranges previously input
%           via yrange(); will be respected if this is not set.
%   xlabel= String for the X-axis label
%   ylabel= String for the Y-axis label
%           Note: xlabel(); ylabel(); commands will also work.
%   oplt    = 0 (default) for new plot, !=0 for overplotting
%
%   Further note: plotxy(...); will apply the choices from connect_points(#);
%\seealso{plot_counts, plot_data, plot_unfold, plot_residuals, plot_fit_model, plotxy, plot_comps, plot_double}
%!%-
{
   variable args, ivars, x,xx,xm=NULL,xp=NULL, y,yy,ym=NULL,yp=NULL;
   plopt = get_plot_options;

   if(_NARGS > 0)
   {
      args = __pop_list(_NARGS);
      ivars=_NARGS;

      if(typeof(args[-1])==Struct_Type)
      {
         set_plot_parms_struct(args[-1],1,[1];;__qualifiers);
         ivars--;
      }
      else
      {
         set_plot_parms_struct(parm_s,1,[1];;__qualifiers);
      }

      if(ivars==2)
      {
         x = args[0];
         y = args[1];
      }
      else if(ivars==4 || ivars==5 || ivars==6)
      {
         x = args[0];
         xm = args[1];
         xp = args[2];
         y = args[3];
         if(ivars>=5) ym = args[4];
         if(ivars==6) yp = args[5];
      }
      else
      {
         () = fprintf(fp, "%s\n", %%%{{{
`
 Inconsistent number of input arguments (expecting 2, 4, or 6 inputs
 plus plot options structure and/or qualifiers).
`                     );          %%%}}}
         return;
      }
   }
   else
   {
     ()=fprintf(fp, "%s\n", %{{{
`
 plotxy(x,dxm,dxp,y,dym,dyp, pstruct); % pstruct = struct{dcol, decol, xrng, ...}
 plotxy(x,dxm,dxp,y,dym,dyp; dcol=#, decol=#, ...);

 Also accepts:

 plotxy(x,y [,pstruct;qualifiers]);
 plotxy(x,dxm,dxp,y [,pstruct;qualifiers]);
 plotxy(x,,,y,dym,dyp [,pstruct;qualifiers]);

   Plot simple x,y plots with error bars: x-dxm, x+dxp, etc.

   Options below refer to structure variables/associative keys/qualifiers:

   dcol  = (pgplot) color value for data
   decol = (pgplot) color value for data
   dsym  = (pgplot) symbol value for data
           Note: dsym=0 will *not* produce a histogram plot
   xrange= List of X-limits for the data. Ranges previously input
           via xrange(); will be respected if this option is not set.
   yrange= List of Y-limits for the data. Ranges previously input
           via yrange(); will be respected if this is not set.
   xlabel= String for the X-axis label
   ylabel= String for the Y-axis label
           Note: xlabel(); ylabel(); commands will also work.
`
+ooplt
+`
   Further note: plotxy(...); will apply the choices from connect_points(#);
`); %}}}
      return;
   }

   if(xlabl!=NULL) xlabel(xlabl);
   if(ylabl!=NULL) ylabel(ylabl[0]);

   if(length(xrng)>1)
   {
      if((xrng[0]!=NULL) && (xrng[0]<=0))
      {
         xlin;
      }
      xrange(xrng[0],xrng[1]);
   }
   if(length(yrng)>1)
   {
      if((yrng[0]!=NULL) && (yrng[0]<=0))
      {
         ylin;
      }
      yrange(yrng[0],yrng[1]);
   }

   set_line_width(d_width);
   point_style(dsym[0]);

   if(oplt==0)
   {
      plot(x,y,dcol[0]);
   }
   else
   {
      oplot(x,y,dcol[0]);
   }

   xx = @x*1.;
   yy = @y*1.;

   if(get_plot_options().logx) 
   { 
      xx[where(xx <= 0)] = 2.e-38;
      xx = log10(xx);
   }
   if(get_plot_options().logy) 
   { 
      yy[where(yy <= 0)] = 2.e-38;
      yy = log10(yy);
   }

   if( xm != NULL || xp != NULL )
   {
      if( xm == NULL ) xm = 0.*x;
      if( xp == NULL ) xp = 0.*x;

      xm=x-xm; xp=x+xp;

      if(get_plot_options().logx) 
      { 
         xm[where(xm <= 0)] = 1.e-38;
         xp[where(xp <= 0)] = 3.e-38;

         xm = log10(xm); 
         xp = log10(xp); 
      }

      _pgslw(de_width);
      _pgsci(decol[0]);
      () = _pgerrx(length(x),xm,xp,yy,ebar_x);
   }

   if( ym != NULL || yp != NULL )
   {
      if( ym == NULL ) ym = 0.*y;
      if( yp == NULL ) yp = 0.*y;

      ym=y-ym; yp=y+yp;

      if(get_plot_options().logy) 
      { 
         ym[where(ym <= 0)] = 1.e-38;
         yp[where(yp <= 0)] = 3.e-38;

         ym = log10(ym); 
         yp = log10(yp); 
      }

     _pgslw(de_width);
     _pgsci(decol[0]);
     () = _pgerry(length(x),xx,ym,yp,ebar_y);
   }

   set_line_width(d_width);
   point_style(dsym[0]);

   oplot(x,y,dcol[0]);
   color(dcol[0]);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define plot_comps()
%!%+
%\function{plot_comps}
%\synopsis{Create a data plot with model components explicitly shown (isis_fancy_plots package)}
%\usage{plot_comps({data},&plot_func;dcol={val},mcol={val},ccol={val},cstyle=val,...);}
%\altusage{plot_comps({data},pstrut,&plot_func); where pstruct=struct{dcol, mcol, ...}}
%\description
%  Use a fancy plotting function, e.g., plot_counts or plot_data or
%  plot unfold, passed as a reference, and cycle through all the
%  components with a norm parameter.  Plot each of these as a separate
%  model component.  The plot functions now take two additional
%  optional qualifiers (which alternatively can be passed via the
%  pstruct structure variable): ccol and cstyle.  The ccol parameter
%  gives the color of the model components for each dataset, which can
%  be different from the color of the model for the complete model.
%  The cstyle allows a global change of the line_style for *all* of the
%  model components.  (I.e., only one alternate line_style can be
%  chosen.)  data is the usual combination of integers (=individual
%  data sets), arrays (=data sets to be combined), and lists (=id of
%  combined datasets).
%
%\examples
%    plot_comps({1,[2,3]},popt,&plot_counts;xrange={1,10});
%    plot_comps({5,8},&plot_unfold);
%    plot_comps({{1}},popt,&plot_data);
%\seealso{plot_counts, plot_data, plot_unfold, plot_residuals, plot_fit_model, plotxy, plot_comps, plot_double}
%!%-
{
   variable args;
   if(_NARGS == 2 || _NARGS==3)
   {
      args = __pop_list(_NARGS);
   }
   else
   {
      () = fprintf(fp, "%s\n", %%%{{{
`
plot_comps({data},pstruct,&plot_func);  % pstruct=struct{dcol, mcol, ...}
plot_comps({data},&plot_func;dcol={val},mcol={val},ccol={val},cstyle=val,...);

  Use a fancy plotting function, e.g., plot_counts or plot_data or
  plot unfold, passed as a reference, and cycle through all the
  components with a norm parameter.  Plot each of these as a separate
  model component.  The plot functions now take two additional
  optional qualifiers (which alternatively can be passed via the
  pstruct structure variable): ccol and cstyle.  The ccol parameter
  gives the color of the model components for each dataset, which can
  be different from the color of the model for the complete model.
  The cstyle allows a global change of the line_style for *all* of the
  model components.  (I.e., only one alternate line_style can be
  chosen.)  data is the usual combination of integers (=individual
  data sets), arrays (=data sets to be combined), and lists (=id of
  combined datasets). Examples:

    plot_comps({1,[2,3]},popt,&plot_counts;xrange={1,10});
    plot_comps({5,8},&plot_unfold);
    plot_comps({{1}},popt,&plot_data);
`                  );          %%%}}}
      return;
   }

   % Since plot_residuals might be called, have to make sure that
   % yrange is padded with two NULLS if only two values (defaulted to
   % data plot) are input.  res_yrange ultimately will always be
   % passed as a qualifier to the plot routines.

   variable i, j, res_yrange={NULL,NULL,NULL,NULL};
   variable dores = qualifier("res",0);

   % Allow model components to have different colors for each
   % *dataset* (i.e., all components for the same dataset will have
   % the same color).  Force this to default to the model color if
   % nothing is input.
   variable ccol = qualifier("ccol",NULL);
   variable cstyle = qualifier("cstyle",1);
   variable mcol = qualifier("mcol",NULL);

   % We have a plot structure variable input
   if(_NARGS == 3)
   {
      variable popt_save = @args[1];
      dores = popt_save.res;
      res_yrange = {};

      % Fill res_yrange with popt structure values, then pad with
      % NULLs

      _for i (0,length(popt_save.yrange)-1,1)
      {
         list_append(res_yrange,popt_save.yrange[i]);
      }
      _for i (length(popt_save.yrange),3,1)
      {
         list_append(res_yrange,NULL);
      }

      ccol = qualifier("ccol",popt_save.ccol);
      cstyle = qualifier("cstyle",popt_save.cstyle);
      mcol = qualifier("mcol",popt_save.mcol);

      dores = qualifier("res",dores);
   }

   if( ccol == NULL ) ccol = mcol;
   if( cstyle == NULL || 
       (typeof(cstyle) != Integer_Type && typeof(cstyle) != Long_Type) ) cstyle = 1;

   variable oline_style = get_plot_options().line_style;

   if( length(mcol) > 1 && length(mcol) > length(ccol) )
   {
      _for i (length(ccol),length(mcol)-1,1)
      {
         list_append(ccol,mcol[i]);
      }
   }
      
   variable yrange_tmp = qualifier("yrange",res_yrange);

   % First if block should only get executed if a yrange qualifier was
   % input

   if(length(yrange_tmp)<4)
   {
      res_yrange = {};
      _for i (0,length(yrange_tmp)-1,1)
      {
         list_append(res_yrange,yrange_tmp[i]);
      }
      _for i (length(yrange_tmp),3,1)
      {
         list_append(res_yrange,NULL);
      }
   }
   else
   {
      res_yrange = yrange_tmp;
   }

   variable fv_hold = Fit_Verbose;
   Fit_Verbose = -1;

   variable tmpfile="/tmp/plot_comps_"+string(getpid)+".par";
   save_par(tmpfile);

   % find all components that are a "norm" that aren't already 0,
   % aren't tied to another parameter, aren't a function of another
   % parameter, and won't break if set to 0

   variable normcomps = get_params("*");
   variable iw = Integer_Type[0];
   _for i (0,length(normcomps)-1,1)
   {
      if( normcomps[i].is_a_norm==1 &&
          normcomps[i].tie==NULL    &&
          normcomps[i].fun==NULL    &&
          ((normcomps[i].value>0 && normcomps[i].min<=0) ||  
           (normcomps[i].value<0 && normcomps[i].max>=0)   )
         )
         iw = [iw,i];
   } 
   normcomps = normcomps[iw];
   variable n_comps = length(normcomps);
   variable orig;
   
   variable first_plot=__qualifiers, middle_plot=@first_plot, 
            last_plot=@first_plot, res_plot=@first_plot, qfield;

   if(first_plot==NULL)
   { 
      first_plot = struct{dsym,dcol,decol,mcol,no_reset};
      middle_plot = struct{dsym,dcol,decol,oplt,no_reset};
      last_plot = struct{oplt,no_reset};
      res_plot = struct{yrange};
   }

   variable olist = {}, zlist = {};

   loop(length(args[0]))
   {
      list_append(olist,1);
      list_append(zlist,0);
   }

   % When doing model components, try to keep the data semi-hidden
   foreach qfield (["dsym","dcol","decol","mcol","res","no_reset"])
   {
      ifnot(struct_field_exists(first_plot,qfield))
      { 
         first_plot = struct_combine(first_plot,qfield);
      }
   }
   set_struct_field(first_plot, "dcol", zlist);
   set_struct_field(first_plot, "dsym", olist);
   set_struct_field(first_plot, "mcol", zlist);
   set_struct_field(first_plot, "res", 0);
   set_struct_field(first_plot, "no_reset", 1);

   foreach qfield (["dsym","dcol","decol","mcol","res","oplt","no_reset"])
   {
      ifnot(struct_field_exists(middle_plot,qfield))
      { 
         middle_plot = struct_combine(middle_plot,qfield);
      }
   }
   set_struct_field(middle_plot, "dcol", zlist);
   set_struct_field(middle_plot, "dsym", olist);
   set_struct_field(middle_plot, "mcol", ccol);
   set_struct_field(middle_plot, "res", 0);
   set_struct_field(middle_plot, "oplt", 1);
   set_struct_field(middle_plot, "no_reset", 1);

   foreach qfield (["res","oplt","no_reset"])
   {
      ifnot(struct_field_exists(last_plot,qfield))
      { 
         last_plot = struct_combine(last_plot,qfield);
      }
   }
   set_struct_field(last_plot, "res", 0);
   set_struct_field(last_plot, "oplt", 1);
   set_struct_field(last_plot, "no_reset", 0);

   if (dores)
   {
      set_struct_field(last_plot, "no_reset", 1);
      ifnot(struct_field_exists(res_plot,"yrange")) 
         res_plot = struct_combine(res_plot,"yrange");
      res_plot.yrange = res_yrange;
      multiplot([3,1]);
      mpane(1);
   }

   if(_NARGS<3)
   {
      @args[1](args[0];; first_plot);
   }
   else 
   {
      @args[2](args[0],popt_save;; first_plot);
   }

   % here is the main loop:   
   _for i (0,n_comps-1,1)
   { 
      load_par(tmpfile);
      orig = normcomps[i].value;
      _for j (0,n_comps-1,1) { set_par(normcomps[j].name,0.); };
      set_par(normcomps[i].name, orig);
      ()=eval_counts;

      line_style(cstyle);
      if(_NARGS<3)
      {
         @args[1](args[0];; middle_plot);
      }
      else 
      {
         @args[2](args[0],popt_save;; middle_plot);
      }
   };

   % And the final data plot ...
   load_par(tmpfile);
   ()=eval_counts;

   line_style(oline_style);
   if(_NARGS<3)
   {
      @args[1](args[0];; last_plot);
   }
   else 
   {
      @args[2](args[0],popt_save;; last_plot);
   }

   if (dores)
   {
      mpane(2);
      if(_NARGS<3)
      {
         plot_residuals(args[0];; res_plot);
      }
      else 
      {
         plot_residuals(args[0],popt_save;; res_plot);
      }
   }   
   () = remove(tmpfile);

   Fit_Verbose = fv_hold;
};

%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define plot_double()
%!%+
%\function{plot_double}
%\synopsis{Use two different plot functions in the same figure (isis_fancy_plots package)}
%\description
%
%plot_double({data},pstruct,&plot_funcI,&plot_funcII);  % pstruct=struct{dcol, mcol, ...}
%plot_double({data},&plot_funcI,&plot_funcII;dcol={val},mcol={val},ccol={val},...);
%
%  Using fancy plotting functions, e.g., plot_counts or plot_data or
%  plot unfold, passed as references, apply plot_funcI in the upper
%  panel and plot_funcII in the lower panel.  (If residuals are chosen
%  to be displayed, include a third panel for them beneath the first
%  two plots.)  In each plot, cycle through all the components with a
%  norm parameter.  Plot each of these as a separate model component.
%  The plot functions now take two additional optional qualifiers
%  (which alternatively can be passed via the pstruct structure
%  variable): ccol and cstyle.  The ccol parameter gives the color of
%  the model components for each dataset, which can be different from
%  the color of the model for the complete model.  The cstyle allows a
%  global change of the line_style for *all* of the model components.
%  (I.e., only one alternate line_style can be chosen.)  data is the
%  usual combination of integers (=individual data sets), arrays (=data
%  sets to be combined), and lists (=id of combined datasets). Plot
%  parameters retain their usual meaning, with the exception of yrange
%  and power.  yrange is now a list of up to 6 elements, with the first
%  two applying to plot_funcI, the next two applying to plot_funcII,
%  and the final two applying to the residuals. If power is a list of
%  two elements, then the first element applies to plot_funcI and the
%  second element applies to plot_funcII.
%
%\examples
%    plot_double({1,[2,3]},popt,&plot_unfold,&plot_counts;xrange={1,10});
%    plot_double({5,8},&plot_unfold,&plot_data);
%    plot_double({{1}},popt,&plot_data,&plot_counts);
%\seealso{plot_counts, plot_data, plot_unfold, plot_residuals, plot_fit_model, plotxy, plot_comps, plot_double}
%!%-
{
   variable args;
   if(_NARGS == 3 || _NARGS==4)
   {
      args = __pop_list(_NARGS);
   }
   else
   {
      () = fprintf(fp, "%s\n", %%%{{{
`
plot_double({data},pstruct,&plot_funcI,&plot_funcII);  % pstruct=struct{dcol, mcol, ...}
plot_double({data},&plot_funcI,&plot_funcII;dcol={val},mcol={val},ccol={val},...);

  Using fancy plotting functions, e.g., plot_counts or plot_data or
  plot unfold, passed as references, apply plot_funcI in the upper
  panel and plot_funcII in the lower panel.  (If residuals are chosen
  to be displayed, include a third panel for them beneath the first
  two plots.)  In each plot, cycle through all the components with a
  norm parameter.  Plot each of these as a separate model component.
  The plot functions now take two additional optional qualifiers
  (which alternatively can be passed via the pstruct structure
  variable): ccol and cstyle.  The ccol parameter gives the color of
  the model components for each dataset, which can be different from
  the color of the model for the complete model.  The cstyle allows a
  global change of the line_style for *all* of the model components.
  (I.e., only one alternate line_style can be chosen.)  data is the
  usual combination of integers (=individual data sets), arrays (=data
  sets to be combined), and lists (=id of combined datasets). Plot
  parameters retain their usual meaning, with the exception of yrange
  and power.  yrange is now a list of up to 6 elements, with the first
  two applying to plot_funcI, the next two applying to plot_funcII,
  and the final two applying to the residuals. If power is a list of
  two elements, then the first element applies to plot_funcI and the
  second element applies to plot_funcII.  Examples:

    plot_double({1,[2,3]},popt,&plot_unfold,&plot_counts;xrange={1,10});
    plot_double({5,8},&plot_unfold,&plot_data);
    plot_double({{1}},popt,&plot_data,&plot_counts);
`                  );          %%%}}}
      return;
   }

   variable i, j, res_yrange={NULL,NULL,NULL,NULL,NULL,NULL}, d1_yrange, d2_yrange;
   variable dores = qualifier("res",0);
   variable puse, pdo = qualifier("power",NULL);
   variable ccol = qualifier("ccol",NULL);
   variable cstyle = qualifier("cstyle",1);
   variable mcol = qualifier("mcol",NULL);

   if(_NARGS == 4)
   {
      variable popt_save = @args[1];
      dores = popt_save.res;
      dores = qualifier("res",dores);
      pdo = popt_save.power;
      pdo = qualifier("power",NULL);

      res_yrange = {};
      _for i (0,length(popt_save.yrange)-1,1)
      {
         list_append(res_yrange,popt_save.yrange[i]);
      }
      _for i (length(popt_save.yrange),5,1)
      {
         list_append(res_yrange,NULL);
      }

      ccol = qualifier("ccol",popt_save.ccol);
      cstyle = qualifier("cstyle",popt_save.cstyle);
      mcol = qualifier("mcol",popt_save.mcol);
   }

   if( ccol == NULL ) ccol = mcol;
   if( cstyle == NULL || 
       (typeof(cstyle) != Integer_Type && typeof(cstyle) != Long_Type) ) cstyle = 1;

   variable oline_style = get_plot_options().line_style;

   if( length(mcol) > 1 && length(mcol) > length(ccol) )
   {
      _for i (length(ccol),length(mcol)-1,1)
      {
         list_append(ccol,mcol[i]);
      }
   }

   variable yrange_tmp = qualifier("yrange",res_yrange);
   if(length(yrange_tmp)==1) yrange_tmp={};

   _for i (length(yrange_tmp),5,1)
   {
      list_append(yrange,NULL);
   }

   d1_yrange = {yrange_tmp[0],yrange_tmp[1]};
   d2_yrange = {yrange_tmp[2],yrange_tmp[3]};
   res_yrange = {yrange_tmp[4],yrange_tmp[5]};

   variable fv_hold = Fit_Verbose;
   Fit_Verbose = -1;

   variable tmpfile="/tmp/plot_double_"+string(getpid)+".par";
   save_par(tmpfile);

   % find all components that are a "norm" that aren't already 0,
   % aren't tied to another parameter, aren't a function of another
   % parameter, and won't break if set to 0

   variable normcomps = get_params("*");
   variable iw = Integer_Type[0];
   _for i (0,length(normcomps)-1,1)
   {
      if( normcomps[i].is_a_norm==1 &&
          normcomps[i].tie==NULL    &&
          normcomps[i].fun==NULL    &&
          ((normcomps[i].value>0 && normcomps[i].min<=0) ||  
           (normcomps[i].value<0 && normcomps[i].max>=0)   )
         )
         iw = [iw,i];
   } 
   normcomps = normcomps[iw];
   variable n_comps = length(normcomps);
   variable orig;
   
   variable first_plot=__qualifiers, middle_plot=@first_plot, 
            last_plot=@first_plot, res_plot=@first_plot, qfield;

   if(first_plot==NULL)
   { 
      first_plot = struct{dsym,dcol,decol,mcol,power,yrange,no_reset};
      middle_plot = struct{dsym,dcol,decol,power,yrange,oplt,no_reset};
      last_plot = struct{power,yrange,oplt,no_reset};
      res_plot = struct{yrange};
   }

   variable olist = {}, zlist = {};

   loop(length(args[0]))
   {
      list_append(olist,1);
      list_append(zlist,0);
   }

   % When doing model components, try to keep the data semi-hidden
   foreach qfield (["dsym","dcol","decol","mcol","yrange","res","no_reset","power"])
   {
      ifnot(struct_field_exists(first_plot,qfield))
      { 
         first_plot = struct_combine(first_plot,qfield);
      }
   }
   puse=pdo[0];
   set_struct_field(first_plot, "dcol", zlist);
   set_struct_field(first_plot, "dsym", olist);
   set_struct_field(first_plot, "mcol", zlist);
   set_struct_field(first_plot, "power", puse);
   set_struct_field(first_plot, "yrange", d1_yrange);
   set_struct_field(first_plot, "res", 0);
   set_struct_field(first_plot, "no_reset", 1);

   foreach qfield (["dsym","dcol","decol","yrange","mcol","res","oplt","no_reset","power"])
   {
      ifnot(struct_field_exists(middle_plot,qfield))
      { 
         middle_plot = struct_combine(middle_plot,qfield);
      }
   }
   set_struct_field(middle_plot, "dcol", zlist);
   set_struct_field(middle_plot, "dsym", olist);
   set_struct_field(middle_plot, "yrange", d1_yrange);
   set_struct_field(middle_plot, "power", puse);
   set_struct_field(middle_plot, "mcol", ccol);
   set_struct_field(middle_plot, "res", 0);
   set_struct_field(middle_plot, "oplt", 1);
   set_struct_field(middle_plot, "no_reset", 1);

   foreach qfield (["res","yrange","oplt","no_reset","power"])
   {
      ifnot(struct_field_exists(last_plot,qfield))
      { 
         last_plot = struct_combine(last_plot,qfield);
      }
   }
   set_struct_field(last_plot, "yrange", d1_yrange);
   set_struct_field(last_plot, "power", puse);
   set_struct_field(last_plot, "res", 0);
   set_struct_field(last_plot, "oplt", 1);
   set_struct_field(last_plot, "no_reset", 1);

   if (dores)
   {
      ifnot(struct_field_exists(res_plot,"yrange")) 
         res_plot = struct_combine(res_plot,"yrange");
      res_plot.yrange = res_yrange;
      multiplot([5,5,2]);
   }
   else
   {
      multiplot([1,1]);
   }

%%%  First Plot %%%

   mpane(1);

   if(_NARGS<4)
   {
      @args[1](args[0];; first_plot);
   }
   else 
   {
      @args[2](args[0],popt_save;; first_plot);
   }

   % here is the main loop:   
   _for i (0,n_comps-1,1)
   { 
      load_par(tmpfile);
      orig = normcomps[i].value;
      _for j (0,n_comps-1,1) { set_par(normcomps[j].name,0.); };
      set_par(normcomps[i].name, orig);
      ()=eval_counts;

      line_style(cstyle);
      if(_NARGS<4)
      {
         @args[1](args[0];; middle_plot);
      }
      else 
      {
         @args[2](args[0],popt_save;; middle_plot);
      }
   };

   % And the final data plot ...
   load_par(tmpfile);
   ()=eval_counts;

   line_style(oline_style);
   if(_NARGS<4)
   {
      @args[1](args[0];; last_plot);
   }
   else 
   {
      @args[2](args[0],popt_save;; last_plot);
   }

%%%  Second Plot %%%

   mpane(2);

   if(length(pdo)>1)
   {
      puse=pdo[1];
      set_struct_field(first_plot, "power", puse);
      set_struct_field(middle_plot, "power", puse);
      set_struct_field(last_plot, "power", puse);
   }
   set_struct_field(first_plot, "yrange", d2_yrange);
   set_struct_field(middle_plot, "yrange", d2_yrange);
   set_struct_field(last_plot, "yrange", d2_yrange);

   if(_NARGS<4)
   {
      @args[2](args[0];; first_plot);
   }
   else 
   {
      @args[3](args[0],popt_save;; first_plot);
   }

   % here is the main loop:   
   _for i (0,n_comps-1,1)
   { 
      load_par(tmpfile);
      orig = normcomps[i].value;
      _for j (0,n_comps-1,1) { set_par(normcomps[j].name,0.); };
      set_par(normcomps[i].name, orig);
      ()=eval_counts;

      line_style(cstyle);
      if(_NARGS<4)
      {
         @args[2](args[0];; middle_plot);
      }
      else 
      {
         @args[3](args[0],popt_save;; middle_plot);
      }
   };

   % And the final data plot ...
   load_par(tmpfile);
   ()=eval_counts;

   line_style(oline_style);
   if(_NARGS<4)
   {
      @args[2](args[0];; last_plot);
   }
   else 
   {
      @args[3](args[0],popt_save;; last_plot);
   }

%%% And the residuals %%%

   if (dores)
   {
      mpane(3);
      if(_NARGS<4)
      {
         plot_residuals(args[0];; res_plot);
      }
      else 
      {
         plot_residuals(args[0],popt_save;; res_plot);
      }
   }   
   () = remove(tmpfile);

   Fit_Verbose = fv_hold;
};

%
% Code from src/Mike/isisrc_conf.sl
%
% -*- slang -*-

% Last Updated: July 27, 2015

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Functions in This File.  Usage message (almost always) when 
% function is called without arguments.

% sep_grid           : Set the eval method to SEPARATE_GRID
% mrg_grid           : Set the eval method to MERGED_GRID
% usr_grid           : Set up logarithmically spaced USER_GRIDs
% fstat_mod          : Set the fit statistic to sigma=model
% fstat_gehr         : Set the fit statistic to sigma=gehrels
% fstat_norm         : Set the fit statistic to sigma=data
% lmdif              : Set the fit method to lmdif 
% mpfit              : Set the fit method to mpfit 
% minim              : Set the fit method to minim
% subplex            : Set the fit method to subplex
% marq               : Set the fit method to marquardt
% conf_map_fail_hook : Switch between subplex/lmdif when doing
%                      error contours, and the fit fails
% conf_map           : Call conf_map_counts with conf_map_fail_hook
% corback            : Set the data background to have a fittable 
%                      normalization, using the back_fun utility
% fcorback_fit       : The fit function used by corback
% corfile            : In addition to the data background, set another
%                      background subtracted file as a background with
%                      fittable normalization
% fcorfile_fit       : The fit function used by corfile
% bayes_nconf        : Bayesian estimate of lower/upper counts bound
%                      given observed counts & significance
% bayes_sconf        : Bayesian estimate of lower/upper significance bound
%                      given observed counts & desired lower/upper bounds

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define param_default_structure(value,freeze,pmin,pmax,
                                             hmin,hmax,pstep,prstep)
{
   variable param_def = struct{value, freeze, min, max,
                               hard_min, hard_max, step, relstep};
   param_def.value=value;
   param_def.freeze=freeze;
   param_def.min=pmin;
   param_def.max=pmax;
   param_def.hard_min=hmin;
   param_def.hard_max=hmax;
   param_def.step=pstep;
   param_def.relstep=prstep;
   return param_def;
}

public define sep_grid()
{
   switch(_NARGS)
   {
    case 1:
      variable a = ();
      set_eval_grid_method(SEPARATE_GRID,a);
      return;
   }
   {
      variable fps=stderr;
      () = fprintf(fps, "\n%s\n", %%%{{{           
+" sep_grid(a);\n"
+"\n"
+"   Equivalent to set_eval_grid_method(SEPARATE_GRID,a);\n"); %%%}}}
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%

public define fstat_mod()
{
      set_fit_statistic("chisqr;sigma=model");
      return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define fstat_gehr()
{
      set_fit_statistic("chisqr;sigma=gehrels");
      return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define cash()
{
      set_fit_statistic("cash");
      return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define fstat_norm()
{
      set_fit_statistic("chisqr;sigma=data");
      return;
}

alias("fstat_norm","fstat_data");
alias("fstat_norm","fstat_dat");

%%%%%%%%%%%%%%%%%%%%%%%%

public define mrg_grid()
{
   switch(_NARGS)
   {
    case 1:
      variable a = ();
      set_eval_grid_method(MERGED_GRID,a);
      return;
   }
   {
      variable fps=stderr;
      () = fprintf(fps, "\n%s\n", %%%{{{           
+" sep_grid(a);\n"
+" \n"
+"   Equivalent to set_eval_grid_method(MERGED_GRID,a);\n"); %%%}}}
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static variable lgh_lb = Double_Type[1000], lgh_ub = @lgh_lb, lgh_scl = @lgh_lb;

public define log_grid_hook(id,s)
{
    s.bin_lo = _A(10^[lgh_lb[id]:lgh_ub[id]:lgh_scl[id]]);
    s.bin_hi = make_hi_grid(s.bin_lo);
    return s;
}


public define usr_grid()
{ 
   variable a,lb,ub,scl,cache,i,id;
   switch(_NARGS)
   {
    case 4:
      (a,lb,ub,scl) = ();
      cache = 0;
   }
   {
    case 5:
      (a,lb,ub,scl,cache) = ();
      if(cache!=0){cache = 1;}
   }
   {
      variable fps=stderr;
      () = fprintf(fps, "\n%s\n", %%%{{{           
` usr_grid(a,lb,ub,scl [,cache]; pileup, pile_opts=string);
 
   Create a USER_GRID for each dataset, a[i], such that there is a
   logarithmic energy grid running over keV energies:

      10^[lb[i]:ub[i]:scl[i]]

   Called via set_eval_grid_method(USER_GRID, a, &log_grid_hook [,cache]);
   *Also* sets the kernel to: 
 
      set_kernel(a, \"std;eval=all\");

   *unless* the pileup kernel is also set, in which case it is set to

      set_kernel(a, "pileup"+pile_opts);

   (e.g., use pile_opts=";nterms=10;fracexpo=0.5"). 

   If using a pileup kernel, *do not* set caching and *do not* set
   more than one dataset at a time.

 Inputs:
   a    : Vector indices for which the USER_GRID will be applied
   lb   : Vector of log10 of lower keV bounds
   ub   : Vector of log10 of upper keV bounds
   scl  : Uniform log step size of the grid
   cache: (Optional) If non-zero, cache the model evaluation
`); %%%}}}
      return;
   }

   a = [a]; lb = [lb]; ub = [ub]; scl = [scl];

   if(qualifier_exists("pileup") && 
      (length(a) >1 || cache==1)   )
   {
      variable fpsp = stderr;
      () = fprintf(fps, "\n%s\n", %%%{{{           
` Attempt to mix pileup kernel with caching or multiple datasets.
  See help message for usr_grid.
`); %%%}}}
   }

   variable la = length(a), lbz = lb[0], ubz = ub[0], sclz = scl[0];
   variable llb = length(lb);
   variable lub = length(ub);
   variable lscl = length(scl);
   variable pile_opts = qualifier("pile_opts","");

   if(cache==0)
   {
      if(llb < la)
      {
         lb = [lb,ones(la-llb)*lbz];
      }
      if(lub < la)
      {
         ub = [ub,ones(la-lub)*ubz];
      }
      if(lscl < la)
      {
         scl = [scl,ones(la-lscl)*sclz];
      }
   }
   else
   {
      lb = Double_Type[la]; ub = @lb; scl = @lb;
      lb[*] = lbz;
      ub[*] = ubz;
      scl[*] = sclz;
   }

   i = 0;
   foreach(a)
   {
      id = ();
      lgh_lb[id] = lb[i];
      lgh_ub[id] = ub[i];
      lgh_scl[id] = scl[i];
      i++;
   }

   set_eval_grid_method (USER_GRID, a, &log_grid_hook, cache);
   if(qualifier_exists("pileup"))
   {
      set_kernel(a, "pileup"+pile_opts);
   }
   else
   {
      set_kernel(a, "std;eval=all");
   }
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%

public define subplex()
{
   set_fit_method("subplex");
   return;
}

%%%%%%%%%%%%%%%%%%%%

public define marq()
{
   set_fit_method("marquardt");
   return;
}

%%%%%%%%%%%%%%%%%%%%

public define lmdif()
{
   set_fit_method("lmdif");
   return;
}

%%%%%%%%%%%%%%%%%%%%

public define mpfit()
{
   set_fit_method("mpfit");
   return;
}

%%%%%%%%%%%%%%%%%%%%

public define minim()
{
   set_fit_method("minim");
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define conf_map_fail_hook(p1, p2, best_pars, try_pars,fit_info)
{
   variable save_method = get_fit_method ();
                  
   variable meth = strchop(save_method,';',0)[0];

   if(meth=="marquardt" or meth=="lmdif")
   {
      subplex;
   }
   else
   {
      lmdif;
   }

   () = fit_counts (&fit_info);
                  
   set_fit_method (save_method);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static variable conf_map_info=struct{fail};
conf_map_info.fail = &conf_map_fail_hook;

%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define conf_map(x,y)
{
   return conf_map_counts(x,y,conf_map_info);
}

private variable bkg_data = {};   
private variable bkg_exp = {};
private variable bkg_area = {};
private variable bkg_scl = {};
private variable corfile_data = {};   
private variable corfile_app = {};   

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define create_backscale(hist_index)
{
   variable scl;
   variable dscl = get_data_backscale(hist_index);
   variable dexp = get_data_exposure(hist_index);
   variable bscl = get_back_backscale(hist_index);
   variable bexp = get_back_exposure(hist_index);

   if(dscl==NULL || dexp == NULL || bexp == NULL || bscl == NULL || 
      bexp == 0 || min([bscl]) == 0)
   {
      scl = 1.;  % Throw our hands up and concede defeat
   }
   else 
   {
      scl = (dexp*dscl)/(bexp*bscl);
      if(1 < length(dscl)==length(scl)) % if dscl is an array
        scl[where(dscl==0)] = 0;        %   fix possible 0/0 = NaN elements
   }
   return scl, dexp, dscl, bexp, bscl;
}   

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define create_back(hist_index)
{
   rebin_data(hist_index,0);

   variable b = get_back(hist_index);
   if(hist_index > length(bkg_data)-1)
   {
      loop(hist_index-length(bkg_data)+1)
      {
         list_append(bkg_data,NULL);
         list_append(bkg_exp,NULL);
         list_append(bkg_area,NULL);
         list_append(bkg_scl,NULL);
      }
   }

   if( b==NULL )
   {
      bkg_data[hist_index] = 
      Double_Type[length(get_data_counts(hist_index).bin_lo)];
      bkg_exp[hist_index] = 1.;
      bkg_area[hist_index] = 1.;
      bkg_scl[hist_index] = 1.;
   }
   else
   {
      bkg_data[hist_index] = b;
      bkg_exp[hist_index] = get_back_exposure(hist_index);
      bkg_area[hist_index] = get_back_backscale(hist_index);
      (bkg_scl[hist_index],,,,) = create_backscale(hist_index);
   }

}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Defining fit function for the background %%%

public define fcorback_fit(lo,hi,par)
{
   return par[0]*bkg_data[int(par[1])];
}
	
add_slang_function("fcorback",["norm","id"]);

% The id parameter of fcorback should always be frozen to an 
% integer value.  It's default value is 2 but it should be 
% set to be the id of the dataset it describes and remain 
% frozen there. 

%%%%%%%%%%%%%%%%%%%%%%%%%%

define fcorback_default(i)
{
   switch(i)
   {
    case 0:
    return param_default_structure(1,0,0,2,0,1.e10,1.e-3,1.e-3);
   }
   {
    case 1:
    return param_default_structure(1,1,1,100,1,1000,1,1);
   }
}

set_param_default_hook("fcorback","fcorback_default");

%%%%%%%%%%%%%%%%%%%%%%%

public define corback()
{
   variable fps=stderr;
   variable hist_index=0,wipe=1,b,rb,n;
   switch(_NARGS)
   {
    case 1:
      hist_index=();
   }
   {
    case 2:
      (hist_index,wipe)=();
   }
   {
      ()=__pop_list(_NARGS);
   }

   variable ahist = abs(int(hist_index));

   if(get_data_info(ahist)!=NULL && (hist_index<0 || wipe==NULL))
   {
      back_fun(ahist,NULL);
      
      if(get_data_info(ahist).bgd_file=="#_define_bgd()")
      {
         % Save the binning of the data, wipe it out, and then read
         % back in the existing background for these data, and then
         % rebin.
	      
         rb=get_data_info(ahist).rebin;
         n=get_data_info(ahist).notice_list;

         () = fprintf(fps, "\n%s\n", %%%{{{           

` Background errors are being restored to the data variance via any
 previously stored values.
`); %%%}}}
         rebin_data(ahist,0);
         () = _define_back(ahist, bkg_data[ahist], 
                                  bkg_area[ahist], bkg_exp[ahist]);
         rebin_data(ahist,rb);
         ignore(ahist);
         notice_list(ahist,n);
      }

      return;
   }
   else if(get_data_info(ahist)!=NULL)
   {
      % Save the binning of the data, wipe out any background
      % fit function, create a background for the data, rebin
	      
      rb=get_data_info(ahist).rebin;
      n=get_data_info(ahist).notice_list;

      back_fun(ahist,NULL);
      create_back(ahist);
 
      rebin_data(ahist,rb);
      ignore(ahist);
      notice_list(ahist,n);

      if(qualifier_exists("no_bkg_err"))
      {
         () = define_back(ahist,NULL);
         () = fprintf(fps, "\n%s\n", %%%{{{           
` Background errors will *not* be included in the data variance.
`); %%%}}}
      }

      back_fun(ahist,"fcorback("+string(ahist)+")");    
      set_par("fcorback("+string(ahist)+").id",ahist);
      return;
   }

   () = fprintf(fps, "\n%s\n", %%%{{{           
` corback(hist_index [; no_bkg_err]);

   Sets the background associated with data set hist_index as a
   correction file which can be scaled via a constant using the
   automatically defined function:

      fcorback(hist_index)  
 
   The background errors continue to be applied, presuming a unit
   scaling of the correction file (i.e., the errors do not update
   with fits of the correction normalization constant).

   Background errors can be turned off (i.e., excluded from the data
   variance) by setting the qualifier: no_bkg_err.

   To clear the background correction do:

      isis> corback(hist_index,NULL); ---or--- corback(-hist_index);

   If the background errors previously had been turned off via the 
   no_bkg_err qualifier, undoing the corback call will reset the 
   background via the _define_back(#) function. That is, the name of 
   the original data background file (if any) will no longer be listed; 
   however, the previously stored values for the background will be 
   restored.  (Background errors will then also be included in the data
   variance.)
`); %%%}}}
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define fcorfile_fit(lo,hi,par)
{
   return par[0]*bkg_data[int(par[2])] +
          par[1]*corfile_data[int(par[2])];
}
	
add_slang_function("fcorfile",["norm","cornorm","id"]);

%%%%%%%%%%%%%%%%%%%%%%%%%%

define fcorfile_default(i)
{
   switch(i)
   {
    case 0:
    return param_default_structure(1,0,0,2,0,1000,1.e-3,1.e-3);
   }
   {
    case 1:
    return param_default_structure(1,0,0,10,0,1000,1.e-3,1.e-3);
   }
   {
    case 2:
    return param_default_structure(1,1,1,100,1,1000,1,1);
   }
}

set_param_default_hook("fcorfile","fcorfile_default");

%%%%%%%%%%%%%%%%%%%%%%%

public define corfile()
{
   variable indx, wipe, cindx, rscl, rb, n;
   switch(_NARGS)
   {
    case 2:
      (indx, wipe) = ();

      if(wipe==NULL)
      {
         indx = int(abs(indx));
         back_fun(indx,NULL);

         % Make sure info really exists for background, and 
         % that you are not overwriting a file background
         % (latter should only happen if you've accidentally
         % invoked the correction, or you've overwritten the
         % data file; screwups still possible)

         if(length(bkg_data)>indx && bkg_data[indx]!=NULL &&
            length(corfile_app)>indx && corfile_app[indx]==1 &&
            get_data_info(indx).bgd_file=="#_define_bgd()")
         {
            rb=get_data_info(indx).rebin;
            n=get_data_info(indx).notice_list;
            rebin_data(indx,0);
            () = _define_back(indx,bkg_data[indx]/bkg_scl[indx],
                              bkg_area[indx],bkg_exp[indx]);
            corfile_app[indx]=0;
            rebin_data(indx,rb);
            ignore(indx);
            notice_list(indx,n);
         }
         return;
      }
   }
   {
    case 3:
      (indx, cindx, rscl) = ();

      indx = int(abs(indx));
      cindx = int(abs(cindx));
      rscl = abs(rscl);

      % Minimal check that background & corfile data are sanely
      % defined, and a check to make sure that a corfile isn't already
      % defined

      if(get_data_info(indx)!=NULL && get_data_info(cindx)!=NULL && rscl>0.
         && (length(corfile_app)<indx+1 || corfile_app[indx]!=1)            )
      {
         % Wipe out any existing background function
         back_fun(indx,NULL);

         % The correction file data & background, unbinned, and scale
         rb=get_data_info(cindx).rebin;
         n=get_data_info(cindx).notice_list;
         rebin_data(cindx,0);
         variable cdata = get_data_counts(cindx);
         variable cbgd = get_back(cindx);
         variable cscl;
         (cscl,,,,) = create_backscale(cindx);

         % Even though this will be excluded, set the corfile data
         % binning and noticing back to what it was originally
         rebin_data(cindx,rb);
         ignore(cindx);
         notice_list(cindx,n);

         % If there is no defined background, set it to zero
         if(cbgd==NULL) cbgd = Double_Type[length(cdata.value)];

         % Square of the errors on the correction data, allowing for
         % the fact that we might not be including the background
         % errors in the variance of the correction data.
         variable cbkg_err = qualifier("cbkg_err_scl",1);
         variable cbkg_sys = qualifier("cbkg_sys_frac",0);
         variable cerr2 = cdata.value + cbkg_err*cscl*cbgd 
                                      + (cbkg_sys*cbgd)^2;

         % These are the pieces that will be used in the
         % definition of the back_fun function

         % Make sure there are list elements to store the data

         if(indx > length(corfile_data)-1)
         {
            loop(indx-length(corfile_data)+1)
            {
               list_append(corfile_data,NULL);
               list_append(corfile_app,0);
            }
         }

         corfile_data[indx] = (cdata.value-cbgd);
         corfile_app[indx] = 1;
         exclude(cindx);

         % The background part...

         variable bscl, dexp, darea, bexp, barea;
         (bscl, dexp, darea, bexp, barea)  = create_backscale(indx);
         rb=get_data_info(indx).rebin;
         n=get_data_info(indx).notice_list;
         create_back(indx); 

         % The above defines the back_fun function.  However, we 
         % must redefine the data background so that errors are 
         % properly propagated in the fit.  (The data for dataset
         % indx, however, does not get redefined.  It is still the
         % total counts for the that dataset.)

         % Define the new scaled background as: (old scaled background
         % + corfile scale*[corfile data-scaled corfile background]
         % Define new (back error)^2 as: (old scale*old scaled
         % background + (corfile scale)^2 *[corfile data+corfile
         % backscale*corfile background])

         variable bkg_err = qualifier("bkg_err_scl",1);
         variable bkg_sys = qualifier("bkg_sys_frac",0);
         variable nback = bkg_data[indx] + rscl*(cdata.value-cbgd);
         variable nberr2 = bkg_err*bscl*bkg_data[indx] + 
                           (bkg_sys*bkg_data[indx])^2 + rscl^2*cerr2;

         % But in propagating errors, we need to have an unscaled
         % background (A) and scale (x) so that xA = new background
         % and (x^2*A) = (new backround error)^2. So, define the
         % background as A = (new back)^2/(new back error)^2, x = (new
         % back error)^2/(new back)

         variable dback=nback^2/nberr2;

         % but remember that ISIS will then scale the background by
         % the data area and exposure.  For simplicity, set the
         % background exposure equal to the data exposure, and then
         % define the background area scale so the backscale = x

         variable dscal=darea*nback/nberr2;

         % Help avoid 0/0 : background -> 0 here should take care of
         % these points, since that goes to 0 as nback^2.
         dscal[where(dscal==0.)]=1.;
         dscal[where(nberr2==0.)]=1.;
         dback[where(nberr2==0.)]=0.;
         
         () = _define_back(indx,dback,dscal,dexp);

         % Finish the job...
         
         back_fun(indx,"fcorfile("+string(indx)+")");    
         set_par("fcorfile("+string(indx)+").id",indx);
         set_par("fcorfile("+string(indx)+").cornorm",rscl);

         % Rebin the data back to its original binning

         rebin_data(indx,rb);
         ignore(indx);
         notice_list(indx,n);

         return;
      }
   }
   variable fps=stderr;
   () = fprintf(fps, "\n%s\n", %%%{{{     
` corfile(data_indx, corr_indx, scale [; bkg_err_scl, bkg_sys_frac, 
                                        cbkg_err_scl, cbkg_sys_frac]);
 
   Create new, scalable backgrounds for the dataset represented by
   data_indx that is comprised of its current background file plus the
   scaled, background subtracted data represented by dataset
   corr_indx.  Thus, this function can be used to create difference
   spectra between datasets data_indx and corr_indx.  (For example,
   data_indx could be spectra from a faint Galactic LMXB, while
   corr_indx is an observation of diffuse Galactic ridge emission.)

   Dataset data_indx is to be fitted, while dataset corr_indx is
   automatically excluded.

   scale is used to scale the background subtracted data, and is
   usually the ratio of exposure times for dataset data_indx divided
   by that for dataset corr_indx.

   Thus, the new dataset becomes:

        T_d - F_d*B_d - F_c*(T_c-B_c)

   with F_d and F_c being fittable constants via the newly defined
   function: fcorfile(data_indx).  (T=total counts, B=scaled
   background counts, d=data_indx, c=corr_indx).

   The errors for both background components will be added in
   quadrature and incorporated into the data variance presuming F_d=1
   and F_c=scale.  (I.e., the errors are not updated as the fit
   function normalization constants are updated.)

   This latter behavior can be modified by setting the bkg_err_scl,
   bkg_sys_frac, cbkg_err_scl, and/or the cbkg_sys_frac qualifiers.
   These qualifiers alter the data variance due to the data and
   correction file backgrounds and allow for systematic (i.e.,
   fractional) errors in these backgrounds.  Specifically, the data
   variance is taken as:

        T_d + b_d*s_d*B_d + (f_d*B_d)^2 + 
              scale*( T_c + b_c*s_c*B_c + (f_c*B_c)^2 )

   where s are the usual scaling factors for the scaled backgrounds,
   b_d/b_c (defaults of 1) are set by bkg_err_scl/cbkg_err_scl, and
   f_d/f_c (defaults of 0) are set by bkg_sys_frac/cbkg_sys_frac.  For
   example, bkg_err_scl=cbkg_err_scl=bkg_sys_frac=ckg_sys_frac=0
   yields a data variance that is only dependent upon the measured
   counts. bkg_err_scl=cbkg_err_scl=0, bkg_sys_frac=ckg_sys_frac=0.01
   yields a data variance that incorporates a 1% systematic error in
   both the data background and correction file background counts.

   Systematic data errors (if any) will be continued to be applied to
   dataset data_indx, but will not be incorporated for dataset
   corr_indx.

   No error checking is done to make sure that the corr_indx dataset is 
   compatible with the dataset from data_indx.

   To clear the background correction do:

      isis> corfile(data_indx, NULL);

   The original data background will be restored as a defined
   background via the _define_back(#) function.  The original
   background file name will no longer be seen in a list_data call.
   The background file can be re-read via define_back to restore the
   original file name. (This is not strictly necessary, as
   _define_back(#) will restore all of the original information from
   this file.)
`); %%%}}}
   return;
}

%%%%%%%%%%%%%%%%%%%%

define bayes_nconf()
{
   switch(_NARGS)
   {
    case 2:
       variable n,sig;
      (n,sig) = ();
   }
   {
      variable fps=stderr;
      () = fprintf(fps, "\n%s\n", %%%{{{     
` (nlow,nhigh) = bayes_nconf(nobs,sig);
 
   Given an observed number of counts and a desired significance level
   return the Bayesian prediction for the true model counts, using
   symmetric bounds [i.e., P(n<nlow) = P(n>nhhigh)].
 
 Inputs:
   nobs : The observed counts.
   sig  : The desired significance
 
 Outputs:
   nlow : Lower predicted bound
   nhigh: Upper predicted bound
`); %%%}}}
      return;
   }
   % We'll search 10^-4 to 63
   variable x = 10.^([-4000:1800]/1000.);
   variable ser=0.,fac=1.,lser,ub,wl,wu;
   variable i=0;
   loop(n+1)
   {
      ser = ser + x^i/fac;
      i++;
      fac = fac*i;
   }

   ub = -x+log(ser);

   wl = x[ max( where(ub > log((sig+1)/2.)) ) ];
   wu = x[ min( where(ub < log((1-sig)/2.)) ) ];

   return wl, wu;
}

%%%%%%%%%%%%%%%%%%%%

define bayes_sconf()
{
   switch(_NARGS)
   {
    case 3:
       variable n,xl,xu;
      (n,xl,xu) = ();
   }
   {
      variable fps=stderr;
      () = fprintf(fps, "\n%s\n", %%%{{{     
` (sig_low,sig_high) = bayes_sconf(nobs,nlow,nhigh);

   Given an observed number of counts and a lower counts bound and an
   upper counts bound, return the Bayesian estimate of the associated
   significance level, assuming symmetric bounds for the significance
   [i.e., P(n<nlow) = P(n>nhhigh)].

 Inputs:
   nobs    : The observed counts.
   nlow    : The desired lower counts bound
   nhigh   : The desired upper counts bound

 Outputs:
   sig_low : Estimated significance of lower counts
   sig_high: Estimated significance of upper counts
`); %%%}}}
      return;
   }
   % We'll search 10^-4 to 63
   variable sig = [-13000:0]/1000.; % log(1-sig)
   variable esig = log(2.-exp(sig));
   variable lser=0.,user=0.,fac=1.,lb,ub,wl,wu;
   variable i=0;
   loop(n+1)
   {
      lser = lser + xl^i/fac;
      user = user + xu^i/fac;
      i++;
      fac = fac*i;
   }

   lb = -xl+log(lser);
   ub = -xu+log(user);

   wl = 1.-exp(sig[ min( where(lb > esig - log(2.)) ) ]);
   wu = 1.-exp(sig[ max( where(ub > sig - log(2.)) ) ]);

   return wl,wu;
}

%%%%%%%%%%%%%%%%%%%%

%
% Code from src/Mike/isisrc_data.sl
%
% -*- slang -*-

% Safe guarding against dependencies from other Remeis scripts

try{
}
catch AnyError: {}

% Last Updated: April 25, 2017

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Functions in This File.  Usage message (almost always) when 
% function is called without arguments.

% list_all          : List all the data, arfs, and rmfs
% load_scaled_data  : Load data with non-unity AREASCAL
% pha2_bkg          : Assign background from pha2 file
% read_radio        : Read radio data from an ascii file, load to struct
% load_radio        : Take the structure above, and load it into data
% read_optical      : Read (optical) data from file where the units are 
%                     A_lo, A_hi, mJy, Delta mJy
% clear_all         : Wipe out all data, arfs, and rmfs
% load_data_integral: load INTEGRAL data
% fits_write_pha    : Write a fits pha file; stolen from Manfred Hanke

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define list_all() 
{ 
   list_data;
   list_rmf;
   list_arf; 
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static variable ascal_data = {};

static define areascale_post_model_hook(lo,hi,c,b)
{
   return ascal_data[Isis_Active_Dataset-1]*c+b;
}

private define read_ascl(file)
{
   variable ascl, ascl_key=NULL, ascl_col=NULL;
   
   % First try to read it as a keyword
   try{
      ascl_key = fits_read_key(file,"AREASCAL");
   }
   catch AnyError:
   {
      ascl_key = NULL;
   }
   
   % Next try to read it as a column
   try{
      ascl_col = fits_read_col(file+"[1]","AREASCAL");
   }
   catch AnyError:
   {
      ascl_col = NULL;
   }
   
   % Column supersedes keyword
   if(ascl_col==NULL){ ascl=ascl_key; } else { ascl=ascl_col; }
   
   if(ascl_key==NULL && ascl_col==NULL) return -1;

   return ascl;
}

public define load_scaled_data()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{load_scaled_data}
%\synopsis{Loads a dataset with a non-unity AREASCAL keyword or column}
%\usage{id = load_scaled_data( String_Type file);}
%\qualifiers{
%\qualifier{rmf}{[="rmf.fits", response file if not in data header]}
%\qualifier{arf}{[="arf.fits", effective area file if not in data header]}
%\qualifier{bkg}{[="bkg.fits", background file if not in data header]}
%}
%\description
%    Load a data file with a non-unity AREASCAL, rewrite the
%    data/background to undo the ISIS default application of this value,
%    create a post_model_hook to properly apply it, and rewrite the data
%    and background backscales to properly apply it.  The data itself
%    must either reference a set of responses/backgrounds, or these must
%    be entered via qualifiers upon initial loading of the data.
%\seealso{load_data, _define_back, set_data_backscale}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
   switch(_NARGS)
   {
    case 1:
      variable file = ();
      variable id = load_data(file);
   
      variable dascl = read_ascl(file);
      if(dascl[0] < 0) { message(" No data AREASCAL found."); return -1; }

      if(qualifier_exists("arf"))
      {
         variable aid = load_arf(qualifier("arf","unknown.arf"));
         assign_arf(aid, id);
      }
   
      if(qualifier_exists("rmf"))
      {
         variable rid = load_rmf(qualifier("rmf","unknown.rmf"));
         assign_rmf(rid, id);
      }

      if(qualifier_exists("bkg"))
      {
         variable bid = load_rmf(qualifier("bkg","unknown.bkg"));
         if( define_back(id,bid) ) { message(" Failed to load background."); return -1; }
      }
   
      % Everything should now be loaded, so get its information
      variable farf, garf, dinf = get_data_info(id), 
        data = get_data_counts(id), rb = dinf.rebin, did_rb=0;

      % If data.value and rb are the same length, no automatic
      % rebinning has happened, so no need to undo it, and no need to
      % "correct" the data errors

      if(length(data.value)!= length(rb))
      {
         did_rb=1;
         rebin_data(id,0);
         data = get_data_counts(id);
      }
         

      % Who the hell knows what order the energy bounds are stored in.
      % Read that information from the RMF.
   
      variable emin, emax, alo, ahi, rvrs=0;
      try{
         (emin, emax) = fits_read_col(get_rmf_info(dinf.rmfs[0]).arg_string+"[EBOUNDS]",
                                      "e_min","e_max");

         % Energy order is ascending, wavelength descending: reverse
         % the AREASCAL
         if(emin[1]>emin[0])
         {
            (alo,ahi) = _A(emin,emax);
            if(length(dascl)>1)             
            { 
               reverse(dascl); 
               rvrs=1;
            }
            else
            { dascl = dascl*ones(length(alo)); }
	 }
         % Energy order is descending, wavelength ascending: AREASCAL
         % is in the correct order
         else
         {
            alo = array_map(Double_Type, &_A, emax);
            ahi = array_map(Double_Type, &_A, emin);
	    if(length(dascl)==1) { dascl = dascl*ones(length(alo)); }
         }

         if(length(dascl) != length(alo))
         {
            message(" Incommensurate AREASCAL and EBOUNDS lengths");
            return -1;
         } 

         % Undo the effects of AREASCAL on the data

         if(length(dascl) != length(data.value))
         {
            message(" Incommensurate AREASCAL and data lengths");
            return -1;
         } 

         data.value *= dascl;
         if(did_rb) data.err *= dascl;
         put_data_counts(id,data);
      }
      catch AnyError:
      {
         message(" Failed to read energy bounds from RMF."); 
         return -1;
      }

      % Use the AREASCAL to alter the calculated model counts, via
      % set_post_model_hook

      % Set the list to contain the AREASCAL information
      variable i, lascal_data=length(ascal_data);
      if(lascal_data<id)
      {
         _for i (lascal_data+1,id-1,1) { list_append(ascal_data,1); }
         list_append(ascal_data,dascl);
      }
      else
      {
         ascal_data[id-1]=dascl;
      }

      set_post_model_hook(id, &areascale_post_model_hook);
   
      % Now check to see if there is a background
      if(dinf.bgd_file!="")
      {
         variable bascl = read_ascl(file);
         if(bascl[0] < 0) { message(" No background AREASCAL found."); return -1; }

         % Assume same ordering as data file
         if(rvrs) bascl = reverse(bascl);
   
         try{
            % ISIS doesn't give background directly, only scaled
            % background, so get the pieces to undo that.
            variable bbs = get_back_backscale(id);
            variable be  = get_back_exposure(id);
            variable bs  = get_back(id);

            variable dbs = get_data_backscale(id);
            variable de  = get_back_exposure(id);

            variable inz = where(de*dbs > 0);
       
            % Reconstruct the measured background
            variable bm = @bs;
            bm[inz] = (be*bbs[inz])/(de*dbs[inz])*bascl[inz]*bs[inz];

            % Set the new background, with the proper areascales
            % applied
            () = _define_back(id, bm, bbs*bascl, be);
            set_data_backscale(id,dbs*dascl);
         }
         catch AnyError:
         {
            message(" Failed to update background."); 
            return -1;
         }
      }
 
      if(did_rb) rebin_data(id,rb);

      return id;
   }

   variable fp = stderr;

   () = fprintf(fp,  "%s\n", %{{{
`
  id = load_scaled_data(file;rmf="rmf.fits", arf="arf.fits", bkg="bkg.fits");
  
  Load a data file with a non-unity AREASCAL, rewrite the
  data/background to undo the ISIS default application of this value,
  create a post_model_hook to properly apply it, and rewrite the data
  and background backscales to properly apply it.  The data itself
  must either reference a set of responses/backgrounds, or these must
  be entered via qualifiers upon initial loading of the data.
`
                       );        %%%}}}
}
   

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define pha2_bkg( )
{
   variable id, fname, dinfo, dexp, bsc_up, bsc_down, bup, bdown, blo, cnum;
   switch(_NARGS)
   { 
    case 2:
      (id, fname) = ();
   }
   {
      variable fp=stderr;
      () = fprintf(fp, "\n%s\n", %%%{{{
`  pha2_bkg(data_id, "pha2_file");

   Reads the appropriate background_up/_down columns from the
   file "pha2_file", along with the BACKSCUP and BACKSCDN 
   keywords, and assigns the appropriate background to the
   data set signified by data_id.  Presumes the usual 12-dataset
   (HEG/MEG, -3 -> +3) pha2 file.
`                              ); %%%}}}
      return;
   }
   dinfo = get_data_info(id);
   dexp = get_data_exposure(id);
   (bsc_up,bsc_down) = fits_read_key(fname,"backscup","backscdn");
   (bup,bdown,blo) = fits_read_col(fname,"background_up","background_down","bin_lo");

   cnum = 6*(dinfo.part-1)+dinfo.order+3; 
   if(dinfo.order>0) cnum--;

   % Check that things really are wavelength descending from the file
   if(blo[cnum,1] < blo[cnum,0])
   {
      () = _define_back( id, reverse(bup[cnum,*]+bdown[cnum,*]),
                         bsc_up+bsc_down, get_data_exposure(id) );
   }
   else
   {
      () = _define_back( id, (bup[cnum,*]+bdown[cnum,*]),
                         bsc_up+bsc_down, get_data_exposure(id) );
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Stolen from John Houck & Glenn Allen
%

static variable c_cgs = Const_c;
static variable h_cgs = Const_h;
static variable mjy2cgs = 1.0e-26;

public define read_radio( )
{
   variable in_fil, nbin;
   switch(_NARGS)
   { 
    case 2:
      (in_fil, nbin) = ();
   }
   {
      variable fp=stderr;
      () = fprintf(fp, "\n%s\n", %%%{{{
+" radio_data = read_radio(radio_file,nbins);\n"
+"\n"
+"   WARNING: THIS FUNCTION IS DEPRECATED! USE load_radio2!\n"
+"   Takes data in the ASCII file `radio_file', comprised of three\n"
+"   columns: Hz, mJy, \Delta mJy, and places them in a data structure\n" 
+"   suitable for loading via `load_radio'.  nbins gives the\n"
+"   number of logarithmically spaced wavelength bins to cover the\n"
+"   radio data."); %%%}}}
      return;
   }

   variable in_filp = fopen ( in_fil, "r" ); 

   if ( in_filp == NULL )
   {
      verror ( "\n  ***** Error: Unable to open file %s.\n", in_fil );
   }

   variable frequency, flux_density, flux_density_err;

   ( frequency, flux_density, flux_density_err ) = readcol ( in_fil, 1, 2, 3 );

   () = fclose ( in_filp ); 

   variable n_data_pnts = length ( frequency );

   % Turn Hz into Angstrom, & make sure data is sorted

   variable wavelength = c_cgs / frequency * 1.e8; 

   variable sort_indices = array_sort( wavelength );
   wavelength = wavelength[ sort_indices ]; 

   % Convert to ph cm^-2 s^-1 Hz^-1

   variable flux_differential = flux_density*mjy2cgs/h_cgs/frequency;
   variable flux_differential_err = flux_density_err*mjy2cgs/h_cgs/frequency;

   flux_differential = flux_differential[ sort_indices ];
   flux_differential_err = flux_differential_err[ sort_indices ];

   % Create wavelength grid

   variable wavelength_lo, wavelength_hi;
   variable n_wavelength_bins = nbin;
   variable scal_factor = 1. + (10./double(n_wavelength_bins));

   ( wavelength_lo, wavelength_hi ) = linear_grid( 
                                         log10(min(wavelength)/scal_factor),
                                         log10(max(wavelength)*scal_factor),
                                         n_wavelength_bins 
                                                 );
                                               
   wavelength_lo  = 10.0^wavelength_lo;
   wavelength_hi  = 10.0^wavelength_hi;

   variable flux_integrated     = Double_Type [ n_wavelength_bins ];
   variable flux_integrated_err = Double_Type [ n_wavelength_bins ];

   variable i1min = 0; 
   variable i1max = n_data_pnts;
   variable i1, i2;

   for ( i1 = i1min; i1 < i1max; i1++ )
   {
      i2 = where ( (wavelength_lo <= wavelength[i1]) and
                     (wavelength[i1] < wavelength_hi) );
        
      if( length(i2) > 0 )
      {
        
         %  ph cm^-2 s^-1 (in bin)	     

	flux_integrated[i2] =  flux_differential[i1] * c_cgs * 1.e8 *
	                       (wavelength_hi[i2] - wavelength_lo[i2]) 
                               / wavelength_lo[i2] / wavelength_hi[i2];

	flux_integrated_err[i2] =  flux_differential_err[i1] * c_cgs * 1.e8 *
	                       (wavelength_hi[i2] - wavelength_lo[i2]) 
                               / wavelength_lo[i2] / wavelength_hi[i2];
      }
   }

   % Remove unnecessary bins

   variable flag  = 1;
   i1    = -1;
   variable index = 0;

   while (flag)
   {
      i2 = where ( flux_integrated[[i1+1:]] > 0.0 ) + i1 + 1;
      if ( length(i2) > 0 )
      {
         if ( i2[0] > i1+1 )
	 {
            flux_integrated[index]     = 0.0;
            flux_integrated_err[index] = 0.0;
            wavelength_lo[index]       = wavelength_lo[i1+1];
            wavelength_hi[index]       = wavelength_hi[i2[0]-1];

            i1 = i2[0] - 1;
            index++;
         }
         else
         {
            flux_integrated[index]     = flux_integrated[i2[0]];
            flux_integrated_err[index] = flux_integrated_err[i2[0]];
            wavelength_lo[index]       = wavelength_lo[i2[0]];
            wavelength_hi[index]       = wavelength_hi[i2[0]];

            i1 = i2[0];
            index++;
         }
      }
      else flag = 0;
   }

   i1max = index - 1;
   i1min = where( flux_integrated > 0.0 )[0];
   variable n_out_pnts = i1max - i1min + 1;

   variable radio_data = struct{ bin_lo, bin_hi, value, err };

   radio_data.bin_lo = Double_Type [ n_out_pnts ];
   radio_data.bin_lo = wavelength_lo[[i1min:i1max]];

   radio_data.bin_hi = Double_Type [ n_out_pnts ];
   radio_data.bin_hi = wavelength_hi[[i1min:i1max]];

   radio_data.value  = Double_Type [ n_out_pnts ];
   radio_data.value  = flux_integrated[[i1min:i1max]];

   radio_data.err    = Double_Type [ length(wavelength_lo) ];
   radio_data.err    = flux_integrated_err[[i1min:i1max]];

   radio_data.value = radio_data.value;
   radio_data.err = radio_data.err;

   return radio_data;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define load_radio()
{
    vmessage("WARNING! THIS FUNCTION IS DEPRECATED! USE load_radio2!");
    switch(_NARGS)
    {
	case 1:
	variable radio_struct;
	radio_struct = ();
	variable radio_id = define_counts(radio_struct);
	set_data_exposure(radio_id,1.);
	ignore_list(radio_id, where ( radio_struct.value == 0.));
	return radio_id;
    }
    {
	variable fp=stderr;
	() = fprintf(fp, "\n%s\n", %%%{{{
	    +"radio_id = load_radio(radio_struct);\n"
	    +" \n"
	    +"   WARNING THIS FUNCTION IS DEPRECATED! USE load_radio2! \n"
	    +"   Takes the structure output from `read_radio();' and loads\n"
	    +"   it up for fitting.  Also ignores all the zero count\n"
	    +"   channels.  (Note: Those could come back if any further filters\n"
	    +"   are applied.)\n"); %%%}}}
	    return;
	}
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define read_optical( )
{
   variable in_fil, nbin;
   switch(_NARGS)
   { 
    case 1:
      (in_fil) = ();
   }
   {
      variable fp=stderr;
      () = fprintf(fp, "\n%s\n", %%%{{{
+" id = read_optical(optical_file);\n"
+" \n"
+"   Takes data in the ASCII file `optical_file', comprised of four\n"
+"   columns: A_lo, A_hi, mJy, \Delta mJy, and defines a data set with\n"
+"   identification id.  This will fail horribly for overlapping data bins.\n"); %%%}}}
      return;
   }

   variable in_filp = fopen ( in_fil, "r" ); 

   if ( in_filp == NULL )
   {
      verror ( "\n  ***** Error: Unable to open file %s.\n", in_fil );
   }

   variable alo, ahi, flux_density, flux_density_err;

   ( alo, ahi, flux_density, flux_density_err ) = readcol ( in_fil, 1,2,3,4 );

   () = fclose ( in_filp ); 

   variable isort = array_sort(alo);

   variable f_conv = 1.50918896;

   alo = alo[isort];
   ahi = ahi[isort];
   flux_density = flux_density[isort];
   flux_density_err = flux_density_err[isort];

   variable i=0, lalo = length(alo), lotemp, hitemp;
   loop(lalo-1)
   {
      if(alo[i+1] != ahi[i]);
      {
        lotemp = [alo[[0:i]],ahi[i],alo[[i+1:]]];
        hitemp = [ahi[[0:i]],alo[i+1],ahi[[i+1:]]];
	alo = lotemp;
        ahi = hitemp;
        flux_density = [flux_density[[0:i]],0.,flux_density[[i+1:]]];
        flux_density_err = [flux_density_err[[0:i]],0.,flux_density_err[[i+1:]]];
        i++;
      }
      i++;
   }

   variable flux = flux_density * f_conv * log( ahi/alo );
   variable eflux = flux_density_err * f_conv * log( ahi/alo);
   variable id = define_counts(alo,ahi,flux,eflux);
   set_data_exposure(id,1);
   ignore_list(id, where ( flux == 0.));
   return id;
}

%%%%%%%%%%%%%%%%%%%%%%%%%

public define clear_all()
%!%+
%\function{clear_all}
%\synopsis{Deletes data & corresponding ARFs and RMFs}
%\usage{clear_all();}
%\qualifiers{
%\qualifier{noprompt}{avoid the prompt in scripts}
%}
%\description
%    Deletes all data, the corresponding ARFs and RMFS, as well
%    as the intrinsic correction factors used for Fermi spectra.
%    This function removes the indicated spectra from the internal
%    list; it does not affect the disk files containing the spectra.
%
%
%\seealso{delete_data}
%!%-
{
   ifnot (qualifier_exists ("noprompt"))
   {
      variable fp =stderr;
      () = fprintf(fp, "\n%s\n", %%%{{{
		   +" Clearing all data & responses after the pause.\n"
		   +" ^C^C now if this isn't what you want!!!\n"); %%%}}}
      plot_pause;
   }
   delete_data(all_data);
   delete_arf(all_arfs);
   delete_rmf(all_rmfs);

% The following exists in other Remeis scripts, but won't exist if
% these functions are used outside of the Remeis scripts

#ifexists unfold_corr_factor
   unfold_corr_factor = Assoc_Type[Array_Type]; 
#endif

}


%%%%%%%%%%%%%%%%%%%%%%%%%%%


%
% Code from src/Mike/isisrc_fitfun.sl
%
% -*- slang -*-

% Last Updated: April 12, 2012

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Functions in This File

% simple_gpile_fit : Gratings Pileup correction function
%                    (New, better version than in "ABC Guide to Pileup")

% Manfred Hanke's even better simple_gpile + associated functions

% simple_gpile2_fit 
% use_simple_gpile
% get_unpiled_fit_fun
% get_unpiled_data_flux

% The following work presuming data x-axis follows keV (i.e., Hz -> keV
% for PSD):

% mgauss_fit     : Gaussian absorption that's never < 0 (multiplicative)
% qpo_fit        : Lorentzian, normalized to QPO rms (counts/bin)
% zfc_fit        : Zero-frequency centered Lorentzian (counts/bin)
% plaw_fit       : Counts/bin power-law
% bkn_plaw_fit   : Counts/bin broken power-law
% dbkn_plaw_fit  : Counts/bin doubly broken power-law
% rms_gauss_fit  : Counts/bin Guassian, normalized to RMS.
% sinwave_fit    : Counts/bin sine wave
% four_chip_fit  : Renorm for specific ACIS chips
% three_chip_fit : Renorm for specific ACIS chips
% chip()         : Combining the above two

% The following work presuming x-axis follows Angstrom:

% aqpo_fit       : Lorentzian, normalized to QPO rms (counts/bin)
% azfc_fit       : Zero-frequency centered Lorentzian (counts/bin)
% aplaw_fit      : Counts/bin power-law
% abkn_plaw_fit  : Counts/bin broken power-law
% adbkn_plaw_fit : Counts/bin doubly broken power-law
% asinwave_fit   : Counts/bin sine wave
% asqrwave_fit   : Counts/bin square wave

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define param_default_structure(value,freeze,pmin,pmax,
                                             hmin,hmax,pstep,prstep)
{
   variable param_def = struct{value, freeze, min, max,
                               hard_min, hard_max, step, relstep};
   param_def.value=value;
   param_def.freeze=freeze;
   param_def.min=pmin;
   param_def.max=pmax;
   param_def.hard_min=hmin;
   param_def.hard_max=hmax;
   param_def.step=pstep;
   param_def.relstep=prstep;
   return param_def;
}

%%%%%%%%%%%%%%%%%%%%% 
%%%%%%%%%%%%%%%%%%%%% 
%% Minus Gaussian  %% 
%%%%%%%%%%%%%%%%%%%%% 
%%%%%%%%%%%%%%%%%%%%% 

define mgauss_fit(lo,hi,par)
{  
   % par[0] is unit normalization
   variable norm = par[0];

   % par[1] is line energy in keV
   variable kev = par[1];

   % par[2] is line width (sigma) in keV 
   variable sig = par[2];

   variable eavg = reverse(_A(lo)+_A(hi))/2.;
   variable scl = (1. - norm * exp(-((eavg-kev)/sig)^2/2.) );

   variable iw = where( scl < 0. );
   scl[iw] = 0.;
  
   return scl;
}

add_slang_function("mgauss",["norm","LineE","Sigma"]);

define mgauss_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(1.,0,0.,100.,0.,100.,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(6.7,0,0.001,1000.,1.e-6,1.e6,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(0.1,0,0.0001,10.,1.e-8,1.e6,1.e-3,1.e-3);
   }
}

set_param_default_hook("mgauss","mgauss_defaults");

%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%
%% QPO (Lorentzian) %% 
%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%

% Define a s-lang qpo that works better. Properly averages over bin,
% and thus yields a smoother fit.

define qpo_fit(lo,hi,par)
{
   variable l,r,q,f,al,ah;

   % Go from Angstrom to keV (which we pretend is Fourier Hz),
   % Hence it's F[(al-f)/f] - F[(ah-f)/f] below...

   al = _A(lo);
   ah = _A(hi);

   q = par[1];
   f = par[2];

   % In this formulation, par[0] is rms amplitude

   r = par[0]/(0.5 - atan(-2.*q)/PI);

   l = r^2/(al-ah)/PI * 
       ( atan(2.*q*(al-f)/f) - atan(2.*q*(ah-f)/f) );

   l = reverse(l);

   return l;
}

add_slang_function("qpo",["norm [rms]","Q [f/FWHM]","f [Hz]"]);

define qpo_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,0.,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1,0,0.01,1.e3,1.e-6,1.e9,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(1,0,0.,1.e4,0.,1.e16,1.e-3,1.e-3);
   }
}

set_param_default_hook("qpo","qpo_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Zero Freq. Centered Lorentzian  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define zfc_fit(lo,hi,par)
{
   variable a,f,al,ah,l;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),
   % Hence it's atan(al/f) - atan(ah/f) below...

   al = _A(lo);
   ah = _A(hi);
   f = par[1];
   
   % In this formulation, par[0] is RMS

   a = par[0]^2 * PI/2./f;

   l = a*f/(al-ah) *  ( atan(al/f) - atan(ah/f) );

   l = reverse(l);

   return l;
}

add_slang_function("zfc",["norm [rms]","f [Hz]"]);

define zfc_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1,0,1.e-6,1.e3,0,1.e16,1.e-3,1.e-3);
   }
}

set_param_default_hook("zfc","zfc_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Counts/Bin Powerlaw  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%

define plaw_fit(lo,hi,par)
{
   variable a,s,al,ah,l;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);

   a = par[0];
   s = par[1];
   
   if(s != -1)
   {
      l = a/(s+1) * (al^(s+1) - ah^(s+1));
   }
   else
   {
      l = a * ( log(al) - log(ah) );
   }

   l = reverse(l/(al - ah));

   return l;
}

add_slang_function("plaw",["norm","slope"]);

define plaw_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,0.,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(-2,0,-5,5,-10,10,1.e-3,1.e-3);
   }
}

set_param_default_hook("plaw","plaw_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Broken Counts/Bin Powerlaw  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define bkn_plaw_fit(lo,hi,par)
{
   variable a,sa,b,sb,al,ah,aa,l;
   variable iwl,iwh,liwl,liwh;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);
   aa = (al+ah)/2.;
   l = @al;

   a = par[0];
   sa = par[1];
   b = par[2];
   sb = par[3];

   iwl = where(aa<= b);
   iwh = where(aa > b);

   liwl = length(iwl);
   liwh = length(iwh);
   
   if(liwl != 0)
   {
      l[iwl] = a * aa[iwl]^sa;
   }

   if(liwh != 0)
   {
      l[iwh] = a * b^sa * ( aa[iwh]/b )^sb;
   }

   l = reverse(l);

   return l;
}

add_slang_function("bkn_plaw",["norm","slope_1","f_break [Hz]","slope_2"]);

define bkn_plaw_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(0,0,-5,5,-10,10,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(1,0,0.,1.e4,0.,1.e16,1.e-3,1.e-3);
   }
   {case 3:
      return param_default_structure(-2,0,-5,5,-10,10);
   }
}

set_param_default_hook("bkn_plaw","bkn_plaw_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Double Broken Counts/Bin Powerlaw  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define dbkn_plaw_fit(lo,hi,par)
{
   variable a,sa,ba,sb,bb,sc,al,ah,aa,l;
   variable iwl,iwm,iwh,liwl,liwm,liwh;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);
   aa = (al+ah)/2.;
   l = @al;

   a = par[0];
   sa = par[1];
   ba = par[2];
   sb = par[3];
   bb = par[4];
   sc = par[5];

   iwl = where(aa <= ba);
   iwm = where(aa > ba and aa <= bb);
   iwh = where(aa > bb);

   liwl = length(iwl);
   liwm = length(iwm);
   liwh = length(iwh);
   
   if(liwl != 0)
   {
      l[iwl] = a * aa[iwl]^sa;
   }

   if(liwm != 0)
   {
      l[iwm] = a * ba^sa * ( aa[iwm]/ba )^sb;
   }

   if(liwh != 0)
   {
      l[iwh] = a * ba^sa * (bb/ba)^sb * (aa[iwh]/bb)^sc;
   }

   l = reverse(l);

   return l;
}

add_slang_function("dbkn_plaw",["norm","slope_1","f_brk_1 [Hz]",
                   "slope_2","f_brk_2 [Hz]","slope_3"]);

define dbkn_plaw_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(0,0,-5,5,-10,10,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(1,0,0.,1.e4,0.,1.e16,1.e-3,1.e-3);
   }
   {case 3:
      return param_default_structure(-1,0,-5,5,-10,10,1.e-3,1.e-3);
   }
   {case 4:
      return param_default_structure(10,0,0.,1.e4,0.,1.e16,1.e-3,1.e-3);
   }
   {case 5:
      return param_default_structure(-2,0,-5,5,-10,10,1.e-3,1.e3);
   }
}

set_param_default_hook("dbkn_plaw","dbkn_plaw_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Counts/Bin Gaussian Normalized to RMS  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#ifexists erf

define rms_gauss_fit(lo,hi,par)
{
   variable r,f,sig,al,ah,l;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);

   r = par[0];
   f = par[1];
   sig = sqrt(2.)*par[2];

   l = r^2/2 * ( erf( (al-f)/sig ) - erf( (ah-f)/sig ) );
   l = reverse(l/(al-ah));

   return l;
}

add_slang_function("rms_gauss",["norm [rms]","f_0 [Hz]","sigma [Hz]"]);

define rms_gauss_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1,0,0,1.e4,0.,1.e16,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(0.2,0,1.e-4,1.e3,1.e-16,1.e16,1.e-3,1.e-3);
   }
}

set_param_default_hook("rms_gauss","rms_gauss_defaults");

#endif

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Counts/bin Sine Wave %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%

define sinwave_fit(lo,hi,par)
{
   variable a,f,phs,al,ah,l;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),

   al = _A(lo);
   ah = _A(hi);

   a = par[0];
   f = par[1];
   phs = par[2];

   l = a * ( -cos(2*PI*(f*al-phs)) + cos(2*PI*(f*ah-phs)) )/2/PI/f/(al-ah);
   l = reverse(l);

   return l;
}

add_slang_function("sinwave",["norm","freq","phase"]);

define sinwave_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1,0,0,1.e4,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(0,0,0.,1.,0.,1.,1.e-3,1.e-3);
   }
}

set_param_default_hook("sinwave","sinwave_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define four_chip_fit(lo,hi,par)
{  
   % Input, but frozen, lower bin boundaries
   variable lb = [par[4],par[6],par[8],par[10]];

   % Input, but frozen, upper bin boundaries
   variable ub = [par[5],par[7],par[9],par[11]];

   % Input effective area constants
   variable ec = [par[0],par[1],par[2],par[3]];

   variable scl = 1.*ones(length(lo));
   variable avg = (lo+hi)/2.;

   variable iw = Array_Type[4];
   variable ib = Array_Type[3];

   variable i=0;
   loop(4)
   {
      iw = where(avg > lb[i] and avg <= ub[i]);
      scl[iw] = ec[i];

      if(i !=3)
      {
         ib = where(avg > ub[i] and avg <= lb[i+1]);
         scl[ib] = (avg[ib]-ub[i])/(lb[i+1]-ub[i]) *
                   (ec[i+1]-ec[i]) + ec[i];
      }
      
      i++;
   }

   return scl;
}

add_slang_function("four_chip",["ccd7_nrm","ccd6_nrm","ccd5_nrm","ccd4_nrm",
                         "ccd7_lb","ccd7_ub","ccd6_lb","ccd6_ub",
                         "ccd5_lb","ccd5_ub","ccd4_lb","ccd4_ub"]);

define four_chip_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(1.,0,0.8,1.2,0,2,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1.,0,0.8,1.2,0,2,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(1.,0,0.8,1.2,0,2,1.e-3,1.e-3);
   }
   {case 3:
      return param_default_structure(1.,0,0.8,1.2,0,2,1.e-3,1.e-3);
   }
   {case 4:
      return param_default_structure(1.068,1,1.0,40.0,0,100,1.e-3,1.e-3);
   }
   {case 5:
      return param_default_structure(1.557,1,1.0,40.0,0,100,1.e-3,1.e-3);
   }
   {case 6:
      return param_default_structure(1.597,1,1.0,40.0,0,100,1.e-3,1.e-3);
   }
   {case 7:
      return param_default_structure(7.372,1,1.0,40.0,0,100,1.e-3,1.e-3);
   }
   {case 8:
      return param_default_structure(7.436,1,1.0,40.0,0,100,1.e-3,1.e-3);
   }
   {case 9:
      return param_default_structure(13.202,1,1.0,40.0,0,100,1.e-3,1.e-3);
   }
   {case 10:
      return param_default_structure(13.255,1,1.0,40.0,0,100,1.e-3,1.e-3);
   }
   {case 11:
      return param_default_structure(18.980,1,1.0,40.0,0,100,1.e-3,1.e-3);
   }
}


set_param_default_hook("four_chip","four_chip_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define three_chip_fit(lo,hi,par)
{  
   % Input, but frozen, lower bin boundaries
   variable lb = [par[3],par[5],par[7]];

   % Input, but frozen, upper bin boundaries
   variable ub = [par[4],par[6],par[8]];

   % Input effective area constants
   variable ec = [par[0],par[1],par[2]];

   variable scl = 1.*ones(length(lo));
   variable avg = (lo+hi)/2.;

   variable iw = Array_Type[4];
   variable ib = Array_Type[3];

   variable i=0;
   loop(3)
   {
      iw = where(avg > lb[i] and avg <= ub[i]);
      scl[iw] = ec[i];

      if(i !=2)
      {
         ib = where(avg > ub[i] and avg <= lb[i+1]);
         scl[ib] = (avg[ib]-ub[i])/(lb[i+1]-ub[i]) *
                   (ec[i+1]-ec[i]) + ec[i];
      }
      
      i++;
   }

   return scl;
}

add_slang_function("three_chip",["ccd7_nrm","ccd8_nrm","ccd9_nrm",
                         "ccd7_lb","ccd7_ub","ccd8_lb","ccd8_ub",
                         "ccd9_lb","ccd9_ub"]);

define three_chip_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(1.,0,0.8,1.2,0,10,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1.,0,0.8,1.2,0,10,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(1.,0,0.8,1.2,0,10,1.e-3,1.e-3);
   }
   {case 3:
      return param_default_structure(1.065,1,1.0,40.0,0,100.,1.e-3,1.e-3);
   }
   {case 4:
      return param_default_structure(8.381,1,1.0,40.0,0,100.,1.e-3,1.e-3);
   }
   {case 5:
      return param_default_structure(8.568,1,1.0,40.0,0,100.,1.e-3,1.e-3);
   }
   {case 6:
      return param_default_structure(20.060,1,1.0,40.0,0,100.,1.e-3,1.e-3);
   }
   {case 7:
      return param_default_structure(20.079,1,1.0,40.0,0,100.,1.e-3,1.e-3);
   }
   {case 8:
      return param_default_structure(31.258,1,1.0,40.0,0,100.,1.e-3,1.e-3);
   }
}


set_param_default_hook("three_chip","three_chip_defaults");

%%%%%%%%%%%%%

define chip()
{
   switch(Isis_Active_Dataset)
   {
    case 1:
    return four_chip(1);
   }
   {
    case 2:
    return three_chip(2);
   }
   {
    case 3:
    return four_chip(3);
   }
   {
    case 4:
    return three_chip(4);
   }
   {
     return 1;
   }
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Gratings Pileup, Take II  (Better Version %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define simple_gpile_fit(lo,hi,par,fun)
{  
   % Peak pileup correction goes as:
   %    exp(log(1-pfrac)*[counts/max(counts)])

   variable pfrac = par[0];

   % Pileup scales with model counts from *data set* indx ...
   variable indx = typecast(par[1],Integer_Type);

   if( indx == 0 or pfrac == 0. )
   {
      return fun;   % Quick escape for no changes ...
   }

   % ... but the arf index could be a different number, so get that

   variable arf_indx = get_data_info(indx).arfs;

   % Get arf information

   variable arf = get_arf(arf_indx[0]);

   % In dither regions (or bad pixel areas), counts are down not
   % from lack of area, but lack of exposure.  Pileup fraction
   % therefore should scale with count rate assuming full exposure.
   % Use the arf "fracexpo" column to correct for this effect

   variable fracexpo = get_arf_info(arf_indx[0]).fracexpo;
   if( length(fracexpo) > 1 )
   {
      fracexpo[where(fracexpo ==0)] = 1.;
   }
   else if( fracexpo==0 )
   {
      fracexpo = 1;
   }

   % Rebin arf to input grid, correct for fractional exposure, and 
   % multiply by "fun" to get ("corrected") model counts per bin

   variable mod_cts;
   mod_cts = fun * rebin( lo, hi,
                          arf.bin_lo, arf.bin_hi,
                          arf.value/fracexpo*(arf.bin_hi-arf.bin_lo) )
                / (hi-lo);

   % Go from cts/bin/s -> cts/angstrom/s 

   mod_cts = mod_cts/(hi-lo);

   % Use 2nd and 3rd order arfs to include their contribution.
   % Will probably work best if one chooses a user grid that extends
   % from 1/3 of the minimum wavelength to the maximum, and has
   % at least 3 times the resolution of the first order grid.

   variable mod_ord;
   if(par[2] > 0)
   {
      indx = typecast(par[2],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      if( length(fracexpo) > 1 )
      {
         fracexpo[where(fracexpo ==0)] = 1.;
      }
      else if( fracexpo==0 )
      {
         fracexpo = 1;
      }
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,2*arf.bin_lo,2*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }
   if(par[3] > 0)
   {
      indx = typecast(par[3],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      if( length(fracexpo) > 1 )
      {
         fracexpo[where(fracexpo ==0)] = 1.;
      }
      else if( fracexpo==0 )
      {
         fracexpo = 1;
      }
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,3*arf.bin_lo,3*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }

   variable max_mod = max(mod_cts);
   if(max_mod <= 0){ max_mod = 1.;}

   % Scale maximum model counts to pfrac

   mod_cts = log(1-pfrac)*mod_cts/max_mod;

   % Return function multiplied by exponential decrease

   fun = exp(mod_cts) * fun;
  
   return fun;
}

add_slang_function("simple_gpile",
                   ["pile_frac","data_indx","arf2_indx","arf3_indx"]);

define simple_gpile_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,1,0.,0.5,0.,1.,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1,1,0,48,0,100,1,1);
   }
   {case 2:
      return param_default_structure(2,1,0,48,0,100,1,1);
   }
   {case 3:
      return param_default_structure(3,1,0,48,0,100,1,1);
   }
}

set_param_default_hook("simple_gpile", "simple_gpile_defaults");

set_function_category("simple_gpile", ISIS_FUN_OPERATOR);

%%%%%%%%%%%%%%%%%%%%%%% ANGSTROM COUNTS/BIN FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%
%% QPO (Lorentzian) %% 
%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%

% Define a s-lang qpo that works better. Properly averages over bin,
% and thus yields a smoother fit.

define aqpo_fit(lo,hi,par)
{
   variable l,r,q,f,al,ah;

   al = hi;
   ah = lo;

   q = par[1];
   f = par[2];

   % In this formulation, par[0] is rms amplitude

   r = par[0]/(0.5 - atan(-2.*q)/PI);

   l = r^2/(al-ah)/PI * 
       ( atan(2.*q*(al-f)/f) - atan(2.*q*(ah-f)/f) );

   return l;
}

add_slang_function("aqpo",["norm [rms]","Q [f/FWHM]","f [Angstrom]"]);

define aqpo_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1,0,0.01,1.e3,1.e-16,1.e16,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(1,0,0.,1.e4,-1.e16,1.e16,1.e-3,1.e-3);
   }
}

set_param_default_hook("aqpo","aqpo_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Zero Freq. Centered Lorentzian  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define azfc_fit(lo,hi,par)
{
   variable a,f,al,ah,l;
 
   % Go from Angstrom to keV (which we pretend is Fourier Hz),
   % Hence it's atan(al/f) - atan(ah/f) below...

   al = hi;
   ah = lo;
   f = par[1];
   
   % In this formulation, par[0] is RMS

   a = par[0]^2 * PI/2./f;

   l = a*f/(al-ah) *  ( atan(al/f) - atan(ah/f) );

   return l;
}

add_slang_function("azfc",["norm [rms]","f [Angstrom]"]);

define azfc_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1,0,1.e-6,1.e3,0,1.e16,1.e-3,1.e-3);
   }
}

set_param_default_hook("azfc","azfc_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Counts/Bin Powerlaw  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%

define aplaw_fit(lo,hi,par)
{
   variable a,s,al,ah,l;
 
   al = hi;
   ah = lo;

   a = par[0];
   s = par[1];
   
   if(s != -1)
   {
      l = a/(s+1) * (al^(s+1) - ah^(s+1));
   }
   else
   {
      l = a * ( log(al) - log(ah) );
   }

   l = (l/(al - ah));

   return l;
}

add_slang_function("aplaw",["norm","slope"]);

define aplaw_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(-2,0,-5,5,-10,10,1.e-3,1.e-3);
   }
}

set_param_default_hook("aplaw","aplaw_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Broken Counts/Bin Powerlaw  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define abkn_plaw_fit(lo,hi,par)
{
   variable a,sa,b,sb,al,ah,aa,l;
   variable iwl,iwh,liwl,liwh;
 
   al = hi;
   ah = lo;
   aa = (al+ah)/2.;
   l = @al;

   a = par[0];
   sa = par[1];
   b = par[2];
   sb = par[3];

   iwl = where(aa<= b);
   iwh = where(aa > b);

   liwl = length(iwl);
   liwh = length(iwh);
   
   if(liwl != 0)
   {
      l[iwl] = a * aa[iwl]^sa;
   }

   if(liwh != 0)
   {
      l[iwh] = a * b^sa * ( aa[iwh]/b )^sb;
   }

   return l;
}

add_slang_function("abkn_plaw",["norm","slope_1","f_break [Angstrom]","slope_2"]);

define abkn_plaw_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(0,0,-5,5,-10,10,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(1,0,0.,1.e4,0,1.e16,1.e-3,1.e-3);
   }
   {case 3:
      return param_default_structure(-2,0,-5,5,-10,10,1.e-3,1.e-3);
   }
}

set_param_default_hook("abkn_plaw","abkn_plaw_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Double Broken Counts/Bin Powerlaw  %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define adbkn_plaw_fit(lo,hi,par)
{
   variable a,sa,ba,sb,bb,sc,al,ah,aa,l;
   variable iwl,iwm,iwh,liwl,liwm,liwh;
 
   al = hi;
   ah = lo;
   aa = (al+ah)/2.;
   l = @al;

   a = par[0];
   sa = par[1];
   ba = par[2];
   sb = par[3];
   bb = par[4];
   sc = par[5];

   iwl = where(aa <= ba);
   iwm = where(aa > ba and aa <= bb);
   iwh = where(aa > bb);

   liwl = length(iwl);
   liwm = length(iwm);
   liwh = length(iwh);
   
   if(liwl != 0)
   {
      l[iwl] = a * aa[iwl]^sa;
   }

   if(liwm != 0)
   {
      l[iwm] = a * ba^sa * ( aa[iwm]/ba )^sb;
   }

   if(liwh != 0)
   {
      l[iwh] = a * ba^sa * (bb/ba)^sb * (aa[iwh]/bb)^sc;
   }

   return l;
}

add_slang_function("adbkn_plaw",["norm","slope_1","f_brk_1 [Angstrom]",
                   "slope_2","f_brk_2 [Angstrom]","slope_3"]);

define adbkn_plaw_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(0,0,-5,5,-10,10,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(1,0,0.,1.e4,0,1.e16,1.e-3,1.e-3);
   }
   {case 3:
      return param_default_structure(-1,0,-5,5,-10,10,1.e-3,1.e-3);
   }
   {case 4:
      return param_default_structure(10,0,0.,1.e4,0,1.e16,1.e-3,1.e-3);
   }
   {case 5:
      return param_default_structure(-2,0,-5,5,-10,10,1.e-3,1.e-3);
   }
}

set_param_default_hook("adbkn_plaw","adbkn_plaw_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Counts/bin Sine Wave %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%

define asinwave_fit(lo,hi,par)
{
   variable a,f,phs,al,ah,l;
 
   al = hi;
   ah = lo;

   a = par[0];
   f = par[1];
   phs = par[2];

   l = a * ( -cos(2*PI*(f*al-phs)) + cos(2*PI*(f*ah-phs)) )/2/PI/f/(al-ah);

   return l;
}

add_slang_function("asinwave",["norm","freq","phase"]);

define asinwave_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1,0,0,1.e4,0,1.e16,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(0,0,0.,1.,0,1,1.e-3,1.e-3);
   }
}

set_param_default_hook("asinwave","asinwave_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%% Counts/bin Square Wave %% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

define asqrwave_fit(lo,hi,par)
{
   variable a,c,w,r=@lo,iw;

   r[*]=0;
 
   a = par[0];
   c = par[1];
   w = par[2];

   iw = where( lo>= c-w/2. and hi <= c+w/2. );
   if(length(iw) >0)
   {
      r[iw] = a;
   }

   iw = where( lo< c-w/2. and hi <= c+w/2. );
   if(length(iw) >0)
   {
      r[iw] = a*(hi[iw] - c+w/2.)/(hi[iw]-lo[iw]);
   }

   iw = where( lo >= c-w/2. and hi > c+w/2. );
   if(length(iw) >0)
   {
      r[iw] = a*(c+w/2. - lo[iw])/(hi[iw]-lo[iw]);
   }

   return r;
}

add_slang_function("asqrwave",["norm","center","width"]);

define asqrwave_defaults(i)
{
   switch(i)
   {case 0:
      return param_default_structure(0.1,0,0.,1.e8,-1.e16,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(1,0,0,1.e4,0,1.e16,1.e-3,1.e-3);
   }
   {case 2:
      return param_default_structure(0,0,0.,1.,0,1,1.e-3);
   }
}

set_param_default_hook("asqrwave","asqrwave_defaults");

public define tilt_fit(lo,hi,par)
{
   variable e = (_A(lo)+_A(hi))/2/par[0];
   return reverse( e^par[1] );
}

add_slang_function("tilt",["E0 [keV]","DGamma"]);

define tilt_defaults(i)
{
   switch(i)
   {case 0:
      return(3,1,0.1,10);
   }
   {case 1:
      return(0,0,-0.25,0.25);
   }
}

set_param_default_hook("tilt","tilt_defaults");

%%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile2_fit(lo, hi, par, fun)
%%%%%%%%%%%%%%%%%%%%%%%%
{  
   % 2007, August 14 - fracexpo does not have to be an array

   % 2007, May 03 - correct rebinning of the arf

   % 2007, January 22 - no explicit refering to max(mod_cts)
   
   % 2005, October 25 - New and improved functionality, especially
   %                    in the dithered regions of the chips!

   % Peak pileup correction goes as:
   %    exp(log(1-pfrac)*[counts/max(counts)])
   %  = exp(- beta * counts/S/A )
   variable beta = par[0];

   % Pileup scales with model counts from *data set* indx
   variable indx = typecast(par[1],Integer_Type);

   if( indx == 0 or beta == 0. )
   { return fun; }  % Quick escape for no changes ... 

   % The arf index could be a different number, so get that
   variable arf_indx = get_data_info(indx).arfs;

   % Get arf information
   variable arf = get_arf(arf_indx[0]);

   % In dither regions (or bad pixel areas), counts are down not
   % from lack of area, but lack of exposure.  Pileup fraction
   % therefore should scale with count rate assuming full exposure.
   % Use the arf "fracexpo" column to correct for this effect
   variable fracexpo = get_arf_info(arf_indx[0]).fracexpo;
   if(length(fracexpo)>1)
   { fracexpo[where(fracexpo==0)] = 1.; }
   else
   { if(fracexpo==0) { fracexpo = 1; } }

   % Rebin arf to input grid, correct for fractional exposure, and 
   % multiply by "fun" to get ("corrected") model counts per bin
   variable mod_cts_int;
   mod_cts_int = fun * rebin(lo, hi,
                             arf.bin_lo, arf.bin_hi,
                             arf.value*(arf.bin_hi-arf.bin_lo)/fracexpo)
                                      / (hi-lo);
   % Go from bin-integrated(ph/cm^2/s) * bin-integrated(cm^2)
   % to cts/s/angstrom
   variable mod_cts = mod_cts_int/(hi-lo);

   % Use 2nd and 3rd order arfs to include their contribution.
   % Will probably work best if one chooses a user grid that extends
   % from 1/3 of the minimum wavelength to the maximum, and has
   % at least 3 times the resolution of the first order grid.
   variable mod_ord;
   if(par[2] > 0)
   {
      indx = typecast(par[2],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      if(length(fracexpo)>1)
      { fracexpo[where(fracexpo==0)] = 1.; }
      else
      { if(fracexpo==0) { fracexpo = 1; } }
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,2*arf.bin_lo,2*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }
   if(par[3] > 0)
   {
      indx = typecast(par[3],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      if(length(fracexpo)>1)
      { fracexpo[where(fracexpo==0)] = 1.; }
      else
      { if(fracexpo==0) { fracexpo = 1.; } }
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,3*arf.bin_lo,3*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }

   % Return function multiplied by exponential decrease
   return exp(-beta*mod_cts) * fun;
}


add_slang_function("simple_gpile2",
                   ["beta [s*A/cts]","data_indx","arf2_indx","arf3_indx"]);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile2_defaults(i)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
   switch(i)
   {case 0:
      return param_default_structure(0.05,1,0,10,0,1.e16,1.e-3,1.e-3);
   }
   {case 1:
      return param_default_structure(0,1,0,48,0,1000,1,1);
   }
   {case 2:
      return param_default_structure(0,1,0,48,0,1000,1,1);
   }
   {case 3:
      return param_default_structure(0,1,0,48,0,1000,1,1);
   }
}

set_param_default_hook("simple_gpile2", "simple_gpile2_defaults");

set_function_category("simple_gpile2", ISIS_FUN_OPERATOR);



%%%%%%%%%%%%%%%%%%%%%%%
define use_simple_gpile()
%%%%%%%%%%%%%%%%%%%%%%%
{
  % process arguments
  variable data1, arf2=NULL, arf3=NULL;
  switch(_NARGS)
  { case 1:  data1 = (); }
  { case 2: (data1, arf2) = (); }
  { case 3: (data1, arf2, arf3) = (); }
  { message("usage: use_simple_gpile(data1[, arf2[, arf3]]);"); return; } 

  if(arf2!=NULL) 
  { arf2 = [arf2];
    if(length(arf2)!=length(data1)) { message("error (use_simple_gpile): arrays data1 and arf2 have different length"); return; }
  }
  if(arf3!=NULL)
  { arf3 = [arf3];
    if(length(arf3)!=length(data1)) { message("error (use_simple_gpile): arrays data1 and arf3 have different length"); return; }
  }
 
  variable simple_gpile_fun = "simple_gpile2";
  variable fitFun = get_fit_fun();
  if(substr(fitFun, 1, strlen(simple_gpile_fun)) != simple_gpile_fun)
  { fit_fun(simple_gpile_fun + "(Isis_Active_Dataset, " + fitFun + ")"); }

  % assign model parameters
  variable i;
  foreach i (all_data)
  { 
    variable simple_gpile_instance = simple_gpile_fun + "(" + string(i) + ")";
    set_par(simple_gpile_instance + ".beta", 0, 1);
    set_par(simple_gpile_instance + ".data_indx", 0, 1);
    set_par(simple_gpile_instance + ".arf2_indx", 0, 1);
    set_par(simple_gpile_instance + ".arf3_indx", 0, 1);
  }
  for(i=0; i<length([data1]); i++)
  {
    variable id = [data1][i];
    simple_gpile_instance = simple_gpile_fun + "(" + string(id) + ")";
    set_par(simple_gpile_instance + ".beta", 0.05, 0);
    set_par(simple_gpile_instance + ".data_indx", id, 1);
    if(arf2!=NULL) { set_par(simple_gpile_instance + ".arf2_indx", arf2[i], 1); }
    if(arf3!=NULL) { set_par(simple_gpile_instance + ".arf3_indx", arf3[i], 1); }
  }
}


%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_unpiled_fit_fun()
%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable fitFun = get_fit_fun();
  variable unpiledFitFun;
  (unpiledFitFun,) = strreplace(fitFun, "simple_gpile2(Isis_Active_Dataset, ", "(", 1);
  return unpiledFitFun;
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_unpiled_data_flux()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  % process arguments
  variable id;
  switch(_NARGS)
  { case 1:  id = (); }
  { message("usage: get_unpiled_data_flux(id);"); return; } 

  variable f = get_data_flux(id);
  Isis_Active_Dataset = id;
  variable pm = eval_fun(f.bin_lo, f.bin_hi);
  variable fitFun = get_fit_fun();
  fit_fun(get_unpiled_fit_fun());
  Isis_Active_Dataset = id;
  variable  m = eval_fun(f.bin_lo, f.bin_hi);
  fit_fun(fitFun);
  f.value *= m/pm;
  f.err *= m/pm;
  return f;
}

%
% Code from src/Mike/isisrc_flux.sl
%
% -*- slang -*-

% Last Updated: July 22, 2016


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Public Functions in This File.  Usage message (almost always) when 
% function is called without arguments.

% eqw         : Calculate a line equivalent width in mA and eV
% data_flux   : Calculate the flux between two bounds, based upon the 
%               observed data (does not extend beyond data grid)
% model_flux  : Calculate the flux between two bounds, based upon the 
%               fitted model (does not extend beyond model grid)
% calc_flux   : Calculate the flux based upon the fitted model, using 
%               an arbitrary grid (full grid must be input)
% flux_err    : Monte Carlo estimate of model flux errors (probably
%               wrong in practice in most cases)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%

public define eqw()
{
   variable indx, par, pmin=0, lo=NULL, hi=NULL, eva, eve, bounds, unit;
   variable fv_hold = Fit_Verbose;

   switch(_NARGS)
   { case 2:
      (indx, par) = ();
   }
   {
      variable fp=stderr;
      () = fprintf(fp, "\n%s\n", %%%{{{
+" (ew_ma, ew_ev) = eqw(indx, par [; pmin=#, bounds={val,val}, unit=string, noeval, print]);\n"
+"\n"
+" Calculate the equivalent width for a model component, in milli-angstrom and eV.\n"
+"\n"
+"  Inputs:\n"
+"   indx  = Data set index to which the model is applied\n"
+"   par   = Number for, *or* string with the name of, the *normalization*\n"
+"           parameter for the model component for which the EW will be calculated\n"
+"\n"
+"  Optional Qualifier Inputs:\n"
+"   pmin  = For parameter par, the value to which it should be set when calculating\n"
+"           the continuum flux without the line (default = 0, but this allows \n"
+"           other parameter toggles to be used).\n"
+"   bounds= *Energy* ranges (in keV) over which to restrict the\n"
+"           evaluation of the equivalent width. (Otherwise, uses\n"
+"           the full energy range of the arf; see get_model_flux();)\n"
+"   unit  = Units (case insensitive string) for the values of bounds (default=\"kev\")\n"
+"   noeval- If exists, don't do the final eval_counts (parameters will be reset\n"
+"           to initial values, but not evaluated - i.e., plots won't show\n"
+"           the fit with the line, and may otherwise be screwy)\n"
+"   print - If exists, don't output the results, rather, print to screen\n"
+"\n"
+"  Outputs:\n"
+"   ew_ma, ew_ev : Equivalent widths in milli-angstrom and eV\n"); %%%}}}
      return;
   }

   pmin = qualifier("pmin",0.);
   bounds = qualifier("bounds",NULL);
   unit = qualifier("unit","kev");

   if(length(bounds)==2 && bounds[0] != NULL && bounds [1] != NULL)
   {
      if(unit_info(unit).is_energy)
      {
         lo = _A(max([bounds[0],bounds[1]])/unit_info(unit).scale);
         hi = _A(min([bounds[0],bounds[1]])/unit_info(unit).scale);
      }
      else
      {
         lo = min([bounds[0],bounds[1]])/unit_info(unit).scale;
         hi = max([bounds[0],bounds[1]])/unit_info(unit).scale;
      }
   }

   Fit_Verbose = -1;
   () = eval_counts;

   variable a = get_model_flux(indx);
   variable phold = get_par_info(par);

   set_par(par,pmin,0,min([0.,pmin,phold.min]),max([0,pmin,phold.max]));
   () = eval_counts;
   variable b = get_model_flux(indx);

   set_par(par,phold.value,phold.freeze,phold.min,phold.max);

   ifnot(qualifier_exists("noeval")) () = eval_counts;

   variable iw, iwa, iwe;
   if( lo != NULL && hi != NULL )
   {
      iw = where(a.bin_lo >= lo and a.bin_hi <= hi);
   }
   else
   {
      iw = [0:length(a.bin_lo)-1];
   }
   
   variable diff = a.value - b.value;
   variable bdena = b.value/(a.bin_hi-a.bin_lo);
   variable bdene = reverse(reverse(b.value)/(_A(a.bin_lo)-_A(a.bin_hi)));
   variable fdena = diff/(a.bin_hi-a.bin_lo);
   variable fdene = reverse(reverse(diff)/(_A(a.bin_lo) - _A(a.bin_hi)));
   variable flux = sum( diff[iw] );

   iwa = min(where( abs(fdena[iw]) == max(abs(fdena[iw])) ));
   iwe = min(where( abs(fdene[iw]) == max(abs(fdene[iw])) ));
   
   eva = flux/bdena[iw[iwa]]*1000.;
   eve = flux/bdene[iw[iwa]]*1000.;

   Fit_Verbose = fv_hold;

   ifnot(qualifier_exists("print"))
   {
      return eva, eve;
   }
   else
   {
      () = printf("\n Equivalent Width  (mA):  %8.4e ",eva);
      () = printf("\n Equivalent Width  (eV):  %8.4e \n\n",eve);
     }
  }

%%%%%%%%%%%%%%%%%%%%%%%%%%

public define calc_eqw()
{
   variable i, indx, par, fpar, pars, pmin=0, lo=NULL, hi=NULL, eva, eve, bounds, unit;
   variable fv_hold = Fit_Verbose;

   switch(_NARGS)
   { case 4:
      (indx, par, fpar, pars) = ();
   }
   {
      variable fp=stderr;
      () = fprintf(fp, "\n%s\n", %%%{{{
+" ew = calc_eqw(indx, par, fpar, pars [; pmin=#, bounds={val,val}, unit=string]);\n"
+"\n"
+" Calculate a model component equivalent width, in milli-angstrom or eV (default),\n"
+" given a full set of model free-parameters.  Note: current model parameters will\n"
+" be overwritten with values from this function call.\n"
+"\n"
+"  Inputs:\n"
+"   indx  = Data set index for which the EW is calculated\n"
+"   par   = Number for, *or* string with the name of, the *normalization*\n"
+"           parameter for the model component for which the EW will be calculated\n"
+"   fpar  = Array of indices of free parameters\n"
+"   pars  = Array of values of free parameters\n"
+"\n"
+"  Optional Qualifier Inputs:\n"
+"   pmin  = For parameter par, the value to which it should be set when calculating\n"
+"           the continuum flux without the line (default = 0, but this allows \n"
+"           other parameter toggles to be used).\n"
+"   bounds= *Energy* ranges (in keV) over which to restrict the\n"
+"           evaluation of the equivalent width. (Otherwise, uses\n"
+"           the full energy range of the arf; see get_model_flux();)\n"
+"   unit  = Units (case insensitive string) for the values of bounds (default=\"kev\")\n"
+"   mA    - If exists, output the euivalent width in milliAngstrom\n"
+"\n"
+"  Outputs:\n"
+"   ew: Equivalent width in milli-angstrom or eV (default)\n"); %%%}}}
      return;
   }

   pmin = qualifier("pmin",0.);
   bounds = qualifier("bounds",NULL);
   unit = qualifier("unit","kev");

   if(length(bounds)==2 && bounds[0] != NULL && bounds [1] != NULL)
   {
      if(unit_info(unit).is_energy)
      {
         lo = _A(max([bounds[0],bounds[1]])/unit_info(unit).scale);
         hi = _A(min([bounds[0],bounds[1]])/unit_info(unit).scale);
      }
      else
      {
         lo = min([bounds[0],bounds[1]])/unit_info(unit).scale;
         hi = max([bounds[0],bounds[1]])/unit_info(unit).scale;
      }
   }

   Fit_Verbose = -1;

   _for i (0,length(fpar)-1,1)
   {
      set_par(fpar[i],pars[i]);
   }

   () = eval_counts;

   variable a = get_model_flux(indx);
   variable phold = get_par_info(par);

   set_par(par,pmin,0,min([0.,pmin,phold.min]),max([0,pmin,phold.max]));
   () = eval_counts;
   variable b = get_model_flux(indx);

   set_par(par,phold.value,phold.freeze,phold.min,phold.max);

   variable iw, iwa, iwe;
   if( lo != NULL && hi != NULL )
   {
      iw = where(a.bin_lo >= lo and a.bin_hi <= hi);
   }
   else
   {
      iw = [0:length(a.bin_lo)-1];
   }
   
   variable diff = a.value - b.value;
   variable bdena = b.value/(a.bin_hi-a.bin_lo);
   variable bdene = reverse(reverse(b.value)/(_A(a.bin_lo)-_A(a.bin_hi)));
   variable fdena = diff/(a.bin_hi-a.bin_lo);
   variable fdene = reverse(reverse(diff)/(_A(a.bin_lo) - _A(a.bin_hi)));
   variable flux = sum( diff[iw] );

   iwa = min(where( abs(fdena[iw]) == max(abs(fdena[iw])) ));
   iwe = min(where( abs(fdene[iw]) == max(abs(fdene[iw])) ));
   
   eva = flux/bdena[iw[iwa]]*1000.;
   eve = flux/bdene[iw[iwa]]*1000.;

   Fit_Verbose = fv_hold;

   ifnot(qualifier_exists("mA"))
   {
      return eve;
   }

   return eva;

  }

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define model_flux ( )
{
   variable id, lo, hi, bounds, unit, pc, m, e_avg, pflux, eflux, fp=stderr;

   switch(_NARGS)
   {
    case 3:
      (id,lo,hi) = ();
   }
   {
       () = fprintf(fp, "\n%s\n", %%%{{{
+" (photon_flux, energy_flux) = model_flux(indx,lo,hi [;unit=string, pc=#, print]);\n"
+"\n"
+"   Returns model fluxes for data set indx evaluated over the energy interval (keV)\n"
+"   bounded by lo, hi.  Outputs are units of: \n"
+"\n"
+"      photons/cm^2/sec, ergs/cm^2/sec  (if pc=NULL [DEFAULT])\n"
+"      photons/sec, ergs/sec            (if pc>0)\n"
+"\n"
+"   Optional qualifiers:\n"
+"\n"
+"      unit  = Units (case insensitive string) for the values of lo, hi (default=\"kev\")\n"
+"      pc    = distance, in parsec, to the source (default=NULL)\n"
+"      print - if present, print results to screen without returning values (default off)\n"
+"\n"
+"   See also:  data_flux, calc_flux\n"); %%%}}}
      return;
   }   

   m = get_model_flux(id);

   unit=qualifier("unit","kev");

   if(unit_info(unit).is_energy)
   {
      bounds = _A([min([lo,hi]),max([lo,hi])]/unit_info(unit).scale);
      if(bounds[1] > m.bin_hi[-1])
         () = fprintf(fp, "\n%s\n", %%%{{{
+"***WARNING***: lower energy bound is below that of model data.\n"
+"               *** Use calc_flux(); instead ***\n"); %%%}}}
      if(bounds[0] < m.bin_lo[0])
         () = fprintf(fp, "\n%s\n", %%%{{{
+"***WARNING***: upper energy bound is above that of model data.\n"
+"               *** Use calc_flux(); instead ***\n"); %%%}}}
   }
   else
   {
      bounds = [min([lo,hi]),max([lo,hi])]/unit_info(unit).scale;
      if(bounds[0] < m.bin_lo[0])
         () = fprintf(fp, "\n%s\n", %%%{{{
+"***WARNING***: lower wavelength bound is below that of model data.\n"
+"               *** Use calc_flux(); instead ***\n"); %%%}}}
      if(bounds[1] > m.bin_hi[-1])
         () = fprintf(fp, "\n%s\n", %%%{{{
+"***WARNING***: upper wavelength bound is above that of model data.\n"
+"               *** Use calc_flux(); instead ***\n"); %%%}}}
   }

   pflux = rebin (bounds[0], bounds[1], m.bin_lo, m.bin_hi, m.value)[0];   

   e_avg = _A(1.)*(1./m.bin_hi+1./m.bin_lo)/2.*Const_eV*1.e3;
   eflux = rebin (bounds[0], bounds[1], m.bin_lo, m.bin_hi, m.value*e_avg)[0];  

   pc = qualifier("pc",NULL);

   variable phot_string=" photons/cm^2/sec";
   variable e_string=" ergs/cm^2/sec";

   if( pc != NULL && pc > 0)
   { 
      variable scl = 4*PI*(3.08568025e18*pc)^2;
      pflux *= scl;
      eflux *= scl;
      phot_string=" photons/sec";
      e_string=" ergs/sec";
   }

   if(qualifier_exists("print"))
   {
      () = printf("\n Photon Flux:  %8.4e "+phot_string,pflux);
      () = printf("\n Energy Flux:  %8.4e "+e_string+"\n\n",eflux);
      return;
   }
   else
   {
      return pflux, eflux;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define data_flux ( )
{
   variable id, lo, hi, bounds, unit, pc, m, e_avg, pflux, pflux_err, eflux, eflux_err, fp=stderr;

   switch(_NARGS)
   {
    case 3:
      (id,lo,hi) = ();
   }
   {
      () = fprintf(fp, "\n%s\n", %%%{{{ 
+" (photon_flux, pflux_err, energy_flux, eflux_err) = data_flux(indx,lo,hi [; unit=string, pc=value, print]);\n"
+" \n"
+"   Returns data fluxes (via flux_corr/get_data_flux) for data set indx evaluated over\n"
+"   energy interval (keV) bounded by lo, hi. Error is error of the mean value of the flux.\n"
+"   Outputs are units of: \n"
+"\n"
+"      photons/cm^2/s, photons/cm^2/s, ergs/cm^2/s, ergs/cm^2/s  (if pc=NULL [DEFAULT])\n"
+"      photons/s, photons/s, ergs/s, ergs/s                      (if pc>0)\n"
+"\n"
+"   Optional qualifiers:\n"
+"\n"
+"      unit  = Units (case insensitive string) for the values of lo, hi (default=\"kev\")\n"
+"      pc    = distance, in parsec, to the source (default=NULL)\n"
+"      print - if exists, print results to screen without returning values (default off)\n"
+"\n"
+"   See also:  model_flux, calc_flux\n"); %%%}}}
      return;
   }   

   flux_corr(id);
   m = get_data_flux(id);

   unit=qualifier("unit","kev");

   if(unit_info(unit).is_energy)
   {
      bounds = _A([min([lo,hi]),max([lo,hi])]/unit_info(unit).scale);
      if(bounds[1] > m.bin_hi[-1])
         () = fprintf(fp, "\n%s\n", %%%{{{
+"***WARNING***: lower energy bound is below that of data.\n"
+"               *** Flux is lower limit estimate ***\n"); %%%}}}
      if(bounds[0] < m.bin_lo[0])
         () = fprintf(fp, "\n%s\n", %%%{{{
+"***WARNING***: upper energy bound is above that of data.\n"
+"               *** Flux is lower limit estimate ***\n"); %%%}}}
   }
   else
   {
      bounds = [min([lo,hi]),max([lo,hi])]/unit_info(unit).scale;
      if(bounds[0] < m.bin_lo[0])
         () = fprintf(fp, "\n%s\n", %%%{{{
+"***WARNING***: lower wavelength bound is below that of data.\n"
+"               *** Flux is lower limit estimate ***\n"); %%%}}}
      if(bounds[1] > m.bin_hi[-1])
         () = fprintf(fp, "\n%s\n", %%%{{{
+"***WARNING***: upper wavelength bound is above that of data.\n"
+"               *** flux is lower limit estimate ***\n"); %%%}}}
   }

   pflux = rebin(bounds[0], bounds[1], m.bin_lo, m.bin_hi, m.value)[0];   
   pflux_err = sqrt( rebin(bounds[0], bounds[1], m.bin_lo, m.bin_hi, m.err^2)[0]);

   variable n = length( where(m.bin_lo >= bounds[0] and m.bin_hi <= m.bin_hi) );
   if(n<=1) {  n=2; message("\n ***WARNING***: only one data flux bin in interval. \n"); }

   e_avg = _A(1.)*(1./m.bin_hi+1./m.bin_lo)/2.*Const_eV*1.e3;
   eflux = rebin (bounds[0], bounds[1], m.bin_lo, m.bin_hi, m.value*e_avg)[0];  
   eflux_err = sqrt( rebin(bounds[0], bounds[1], m.bin_lo, m.bin_hi, (m.err*e_avg)^2)[0]);

   pc = qualifier("pc",NULL);

   variable phot_string=" photons/cm^2/sec";
   variable e_string=" ergs/cm^2/sec";

   if( pc != NULL && pc > 0)
   { 
      variable scl = 4*PI*(3.08568025e18*pc)^2;
      pflux *= scl;
      pflux_err *= scl;
      eflux *= scl;
      eflux_err *= scl;
      phot_string=" photons/sec";
      e_string=" ergs/sec";   }

   if(qualifier_exists("print"))
   {
      () = printf("\n Photon Flux:  %8.4e +/- %8.4e"+phot_string,pflux,pflux_err);
      () = printf("\n Energy Flux:  %8.4e +/- %8.4e"+e_string+"\n\n",eflux,eflux_err);
      return;
   }
   else
   {
      return pflux, pflux_err, eflux, eflux_err;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define calc_flux ( )
{
   variable id, lo, hi, bin_lo, bin_hi, unit, pc, m, e_avg, pflux, eflux,  fp=stderr;

   switch(_NARGS)
   {
    case 3:
      (id,lo,hi) = ();
   }
   {
      () = fprintf(fp, "\n%s\n", %%%{{{
+"(photon_flux, energy_flux) = calc_flux(indx,bin_lo,bin_hi [; unit=string, pc=value, print]);\n"
+"\n"
+"   Returns model flux for data set indx evaluated over the grid defined by bin_lo, bin_hi.\n"
+"   Outputs are units of: \n"
+"\n"
+"      photons/cm^2/sec, ergs/cm^2/sec  (if pc=NULL [DEFAULT])\n"
+"      photons/sec, ergs/sec            (if pc>0)\n"
+"\n"
+"   Optional qualifiers:\n"
+"\n"
+"      unit  = Units (case insensitive string) for the values of lo, hi (default=\"kev\")\n"
+"      pc    = distance, in parsec, to the source (default=NULL)\n"
+"      print - if exists, print results to screen without returning values (default off)\n"
+"\n"
+"   See also:  model_flux, data_flux\n"); %%%}}}
      return;
   }   

   unit = qualifier("unit","kev");

   if(unit_info(unit).is_energy)
   {
      bin_lo = _A(hi/unit_info(unit).scale);
      bin_hi = _A(lo/unit_info(unit).scale);
   }
   else
   {
      bin_lo = lo/unit_info(unit).scale;
      bin_hi = hi/unit_info(unit).scale;
   }

   Isis_Active_Dataset=id;
   m = eval_fun(bin_lo,bin_hi);

   pflux = sum(m);

   e_avg = _A(1.)*(1./bin_hi+1./bin_lo)/2.*Const_eV*1.e3;
   eflux = sum(m*e_avg);  

   pc = qualifier("pc",NULL);

   variable phot_string=" photons/cm^2/sec";
   variable e_string=" ergs/cm^2/sec";

   if( pc != NULL && pc > 0)
   { 
      variable scl = 4*PI*(3.08568025e18*pc)^2;
      pflux *= scl;
      eflux *= scl;
      phot_string=" photons/sec";
      e_string=" ergs/sec";
   }

   if(qualifier_exists("print"))
   {
      () = printf("\n Photon Flux:  %8.4e "+phot_string,pflux);
      () = printf("\n Energy Flux:  %8.4e "+e_string+"\n\n",eflux);
      return;
   }
   else
   {
      return pflux, eflux;
   }
}

#iftrue
%%%%%%%%%%%%%%%%%

% Right now, this only use the diagonal elements of the covariance
% matrix (i.e., the single parameter errors).  With changes to the
% ISIS internals, could probably get access to estimates of the other
% values


define flux_err()
{
   variable indx, lo, hi, file, nloop, nbin, delt, pflux, eflux, fp=stderr;

   switch(_NARGS)
   {
    case 3:
      (indx,lo,hi) = ();
   }
   {
      () = fprintf(fp, "\n%s\n", %%%{{{
+" (pflux, eflux) = flux_err(indx,lo,hi [;unit=val, nloop=val, nbin=val,\n"
+"                                        sigma=val, file=string]);\n"
+" \n"
+"   Generates Monte Carlo evaluations of the model flux by varying the\n"
+"   input parameters (assuming gaussian statistics) between lower and\n"
+"   upper bounds for *all* unfrozen and untied parameters.  By default,\n"
+"   the parameter bounds are presumed to be 90% confidence  limits, i.e.,\n"
+"   sigma=+/-sqrt(2.71).  All parameters are also presumed to be statistically\n"
+"   independent (probably a bad assumption in many cases). The parameter\n"
+"   ranges are taken from the min/max limits of the current model parameters,\n"
+"   unless another input file is specified (e.g., the save file from a conf_loop\n"
+"   run).  Results are returned as a structure variable which contains the\n"
+"   fluxes from the individual evaluations, as well as histograms of the values,\n"
+"   statistical properties (mean and standard deviation)\n"
%+"   If the GSL module has been loaded, then results from fitting the histogram\n"
%+"   with a gaussian are also provided.\n"
+"\n"
+"   Inputs: \n"
+"\n"
+"      indx            - Data set index for which the flux will be evaluated\n"
+"      lo, hi          - Low/high unit boundaries for the flux evaluation\n"
+" \n" 
+"   Outputs: \n"
+"\n"
+"    pflux.value     - Photon fluxes in units of photons/cm^2/s\n"
+"    pflux.mean      - Mean value of the photon fluxes in units of photons/cm^2/s\n"
+"    pflux.sdev      - Photon flux standard deviation in units of photons/cm^2/s\n"
+"    pflux.bin_lo    - Array for lower bounds of photon flux histogram\n"
+"    pflux.bin_hi    - Array for upper bounds of photon flux histogram\n"
+"    pflux.hist_val  - Array for histogram values\n"
%+"    pflux.hist_mean - Fitted histogram mean\n"
%+"    pflux.hist_sigma- Fitted histogram sigma\n"
+"\n"
+"    eflux           - Structure variable with the same fields as pflux, but in\n"
+"                      units of ergs/cm^2/sec\n"
+"   Optional qualifiers:\n"
+"\n"
+"    unit  = units of the lo, hi bounds  (default=\"kev\")\n"
+"    nloop = Number of Monte Carlo evaluations to perform (default=1000)\n"
+"    nbin  = Number of histogram bins to use (default=nloop/100)\n"
+"    sigma = The sigma to which the min/max parameter values correspond (default=sqrt(2.71))\n"
+"    file  = The name of the file containing the parameter min/max bounds (default=NULL)\n"
+"\n"
+"   See also:  model_flux, data_flux, calc_flux, conf_loop\n"); %%%}}}
      return;
   }   

   pflux = struct{ value, mean, sdev, bin_lo, bin_hi, hist_vals, hist_mean, hist_sigma };
   eflux = struct{ value, mean, sdev, bin_lo, bin_hi, hist_vals, hist_mean, hist_sigma };

   save_par("/tmp/temp.par");

   if(qualifier_exists("file"))
      load_par(qualifier("file"));

   nloop=qualifier("nloop",1000);
   nbin=qualifier("nbin",nloop/100);
   delt=qualifier("sigma",sqrt(2.71));

   variable fv_hold = Fit_Verbose;
   Fit_Verbose=-1;

   variable i, ii, x, puse, aindx=Integer_Type[0];
   variable par=Double_Type[0], min_par=Double_Type[0], max_par=Double_Type[0];

   _for ii (1,get_num_pars,1)
   {
      puse = get_par_info(ii);
      if(puse.freeze==0 && puse.tie==NULL)
      {
         aindx = [aindx,ii];
         par = [par,puse.value];
         min_par = [min_par,puse.min];
         max_par = [max_par,puse.max];
      }
   }

   pflux.value = Double_Type[nloop];
   eflux.value = Double_Type[nloop];

   _for i (0,nloop-1,1)
   {
      _for ii (0,length(par)-1,1)
      {
         forever
         {
            x = grand(1);
            if(abs(x[0]) < delt) break;
         }
         if( x[0] <= 0 )
         {
            puse = par[ii]+x*(par[ii]-min_par[ii])/delt;
         }
         else
         {
            puse = par[ii]+x*(max_par[ii]-par[ii])/delt;
         }
         set_par(aindx[ii],puse);
      }
      () = eval_counts;
      (pflux.value[i], eflux.value[i]) = model_flux(indx,lo,hi;;__qualifiers);
   }
   load_par("/tmp/temp.par");
   () = eval_counts;
   Fit_Verbose = fv_hold;

   variable stats = moment(pflux.value);
   pflux.mean = stats.ave;
   pflux.sdev = stats.sdev;

   stats = moment(eflux.value);
   eflux.mean = stats.ave;
   eflux.sdev = stats.sdev;

   (pflux.bin_lo,pflux.bin_hi) = linear_grid(min(pflux.value),max(pflux.value),nbin);
   pflux.hist_vals = histogram(pflux.value,pflux.bin_lo,pflux.bin_hi);

   (eflux.bin_lo,eflux.bin_hi) = linear_grid(min(eflux.value),max(eflux.value),nbin);
   eflux.hist_vals = histogram(eflux.value,eflux.bin_lo,eflux.bin_hi);

   return pflux, eflux;
}
#endif

%
% Code from src/Mike/isisrc_group.sl
%

% Last Updated: June 4, 2010

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% These have been superseded by:

%  use_file_group   (replaces apply_x2i_grp)

%  regroup_file     (replaces apply i2x_grp)

% Public Functions in This File.  Usage message (almost always) when 
% function is called without arguments.

% i2x_grp       : Turn an ISIS grouping into a GRPPHA/XSPEC one 
% write_x_grp   : ... take that grouping and apply it to a file
% apply_i2x_grp : Do the above two in one fell swoop
% x2i_grp       : Turn an XSPEC/GRPPHA grouping into an ISIS one
% apply_x2i_grp : ... and apply it in one fell swoop
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define i2x_grp()
{
   switch(_NARGS)
   {
    case 1:
      variable i, lb, ub, id, gdi, lgdi, iw, liw;

      id = ();
      gdi = [reverse(get_data_info(id).rebin),5];
      lgdi = length(gdi)-1;

      variable grp = Integer_Type[lgdi];

      iw = [where( shift(gdi,-1) != gdi )];
      liw = length(iw);
      iw[liw-1] = lgdi;

      grp[iw[[0:liw-2]]] = 1;

      foreach i ([0:liw-2])
      {
         lb = iw[i];
         ub = iw[i+1]-1;
         if(ub>lb)
         {
            grp[[lb+1:ub]] = -1;
         }
      }
      
      return grp;
   }
   {
      variable fp=stderr;
      () = fprintf(fp, "\n%s\n", %%%{{{
+" grouping = i2x_grp(id);\n"
+" \n"
+"   For any dataset, id, grouped in ISIS, create an *energy ordered*\n"
+"   grouping vector that follows the grppha conventions.  (ISIS groups\n"
+"   flagged with 0, however, become *noticed* groups of 1.)\n"); %%%}}}
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define write_x_grp()
{
   switch(_NARGS)
   {
    case 2:
      variable id, grouping, file, fp, colnum;

      (id, grouping) = ();
      file = get_data_info(id).file;

      fp = fits_open_file(file+"[SPECTRUM]","w");
      () = _fits_get_colnum(fp, "grouping", &colnum);
      () = _fits_write_col(fp,colnum,1,1,grouping);
      fits_close_file(fp);

      return;
   }
   {
      variable fpe=stderr;
      () = fprintf(fpe, "\n%s\n", %%%{{{
+" write_x_grp(id,grouping);\n"
+" \n"
+"   Write an XSPEC style grouping to the *file* associated with id.\n"
+"   *Grouping vector must follow grppha conventions and be energy ordered.*\n"
+" \n"
+"   See: grouping = i2x_grp(id);\n"); %%%}}}
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define apply_i2x_grp()
{
   switch(_NARGS)
   {
    case 1:
      variable id, grp;

      id = ();
      grp = i2x_grp(id);
      write_x_grp(id,grp);

      return;
   }
   {
      variable fpe=stderr;
      () = fprintf(fpe, "\n%s\n", %%%{{{
+" apply_i2x_grp(id);\n"
+" \n"
+"   Take the ISIS defined grouping from dataset id, create an XSPEC style\n"
+"   grouping based upon that, and then write that grouping to the\n"
+"   associated file.\n"
+"   *** Superseded by the ISIS intrinsic regroup_file(id); ***\n"); %%%}}}
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%

public define x2i_grp()
{
   switch(_NARGS)
   {
    case 1:
      variable id, file, grp, lgrp, iw, liw, sn, i;

      id = ();
      file = get_data_info(id).file;
      grp = fits_read_col(file,"grouping");

      lgrp = length(grp);
      iw = where(grp == 1);
      liw = length(iw);
      iw = [iw,lgrp];

      sn = 1;
      foreach i ([0:liw-1])
      {
         grp[[iw[i]:max([iw[i],iw[i+1]-1])]] = sn;
         sn = -1*sn;
      }

      return reverse(grp);
   }
   {
      variable fpe=stderr;
      () = fprintf(fpe, "\n%s\n", %%%{{{
+" grouping = x2i_grp(id);\n"
+" \n"
+"   Read an XSPEC style *energy ordered* grouping, associated with the\n"
+"   dataset indicated by id, and convert it to a *wavelength ordered* ISIS\n"
+"   style grouping suitable for use in rebinning, i.e., \n"
+"      isis> rebin_data(indx,grouping);\n"); %%%}}}
      return;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define apply_x2i_grp()
{
   switch(_NARGS)
   {
    case 1:
      variable id, grp; 

      id = ();
      grp = x2i_grp(id);
      rebin_data(id,grp);

      return;
   }
   {
      variable fpe=stderr;
      () = fprintf(fpe, "\n%s\n", %%%{{{
+" apply_x2i_grp(id);\n"
+" \n"
+"   Take the XSPEC style grouping from the *file* associated with\n"
+"   dataset id, convert it to an ISIS style, wavelength ordered\n"
+"   grouping, and apply it to dataset id.\n"
+"   *** Superseded by the ISIS intrinsic use_group_file(id); ***\n"); %%%}}}
      return;
   }
}


%
% Code from src/Mike/isisrc_line_models.sl
%

% only activate if module xspec could be loaded
#ifeval __get_reference("load_xspec_local_models")!=NULL
% -*- slang -*-

% Nov. 26, 2017 -- Version 0.4.2

public define minus(){ return -1.; }

try{
   alias_fun("constant","minus"; names=["value"], values=[1], freeze=[1], min=[0], max=[1]);
   __set_hard_limits("gaussian","norm",-1.e38,1.e38);
   __set_hard_limits("egauss","norm",-1.e38,1.e38);
   __set_hard_limits("agauss","norm",-1.e38,1.e38);
   __set_hard_limits("gauss","norm",-1.e38,1.e38);
   __set_hard_limits("zagauss","norm",-1.e38,1.e38);
   __set_hard_limits("zgauss","norm",-1.e38,1.e38);
   __set_hard_limits("voigt","norm",-1.e38,1.e38);
   __set_hard_limits("Lorentz","norm",-1.e38,1.e38);
}
catch AnyError: {}

variable fp_stde = stderr;

% Line information:

variable line_name=String_Type[0], 
   line_energy=Double_Type[0], line_energy_min=Double_Type[0], line_energy_max=Double_Type[0],
   line_norm=Double_Type[0], line_norm_min=Double_Type[0], line_norm_max=Double_Type[0],
   line_sigma=Double_Type[0], line_sigma_min=Double_Type[0], line_sigma_max=Double_Type[0],
   line_vtherm=Double_Type[0], line_vtherm_min=Double_Type[0], line_vtherm_max=Double_Type[0],
   line_redshift=Double_Type[0];

% ISIS Angstrom gauss model: g_, ISIS keV egauss model: eg_, ISIS
% voigt model: v_, ISIS Lorentz model: L_, XSPEC keV gaussian model:
% xg_, XSPEC Angstrom agauss model: ag_, XSPEC zgauss model: zg_,
% XSPEC zagauss model: zag_

variable etypes=["voigt","egauss","gaussian","zgauss"];
variable atypes=["agauss","lorentz","gauss","zagauss"];
variable ftype_name = Assoc_Type[String_Type];
ftype_name["voigt"]="v_";
ftype_name["egauss"]="eg_";
ftype_name["gaussian"]="xg_";
ftype_name["zgauss"]="zg_";
ftype_name["lorentz"]="L_";
ftype_name["gauss"]="g_";
ftype_name["agauss"]="ag_";
ftype_name["zagauss"]="zag_";

variable mod_comps = Assoc_Type[Array_Type];
variable mod_values = Assoc_Type[Array_Type];
variable mod_types = Assoc_Type[Array_Type];
variable mod_subtract = Assoc_Type[Array_Type];
variable mod_i_a_d = Assoc_Type[Array_Type];
variable mod_eorder = Assoc_Type[Char_Type];

public define line_model_prefix()
{
   variable args;
   if(_NARGS==2)
   {
      args = __pop_list(_NARGS);
      variable mod_type=args[0], prefix=args[1];

      if(typeof(mod_type) != String_Type || typeof(prefix) != String_Type)
      { message("Names of fit function and associated prefix must be strings"); return; }

      ifnot( assoc_key_exists(ftype_name,strlow(mod_type)) )
      { message("Fit function not implemented in database"); return; }

      ftype_name[strlow(mod_type)]=prefix;
      return;
   }
   else
   {
      () = fprintf(fp_stde, "%s\n", %{{{
`
  line_model_prefix(mod_name,prefix)
  
   Change the prefix associated with line model mod_name when using
   the init_line and add_line function to create a summed line model.
   Note that init_line must be rerun, and the line must be deleted
   (delete_line) and then added again (add_line) to the existing line
   model.
`  ); %}}}
      return;
   }
}


public define minus_fit(lo,hi,par)
{
   return par[0];
}

add_slang_function("minus",["norm"]);

define minus_defaults(i)
{
   variable pdef = struct{value, freeze, min, max, hard_min,
    hard_max, step, relstep};
   pdef.value=-1.;
   pdef.freeze=0;
   pdef.min=-1.e8;
   pdef.max=1.e8;
   pdef.hard_min=-1.e38;
   pdef.hard_max=1.e38;
   pdef.step=1.e-3;
   pdef.relstep=1.e-3;
   return pdef;
}

set_param_default_hook("minus","minus_defaults");

public define init_line()
{
   variable lstruct=NULL;
   if(_NARGS==0 && __qualifiers==NULL)
   {
      () = fprintf(fp_stde, "%s\n", %{{{
`
  init_line(struct; line_id="name", center=#, min=#, max=#,
                    norm=#, norm_min=#, norm_max=#,
                    sigma=#, sigma_min=#, sigma_max=#,
                    vtherm=#, vtherm_min=#, vtherm_max=#,
                    redshift=#, unit="type")'

  Add a line identified by line_id to a database accessed by the
  add_line, delete_line, and find_line functions.  Default values for
  the line can be passed by either a structure or by
  qualifiers. Qualifiers will override the values of the structure
  fields.

  Line sigma have the expected meaning for gaussian profiles, while
  Lorentz and Voigt profile width parameters are chosen such that they
  have full width half maximum of 2*sqrt(2*log*(2))*sigma.

  Some roundoff errors may occur when applying line values input in
  one unit to a model whose defaults are in different units.

  Reusing a line_id will overwrite previously initialized values;
  however, any defined fit function must be reentered or reevaluated
  to see the changes applied.

  Structure Fields/Qualifiers:

   line_id:    String by which the line will identified (default "a")
   center:     line centroid, units of unit qualifier/field (default 6.4)
   center_min: lower limit of centroid (default center-sigma_max)
   center_max: upper limit of centroid (default center+sigma_max)
   norm:       line amplitude (photons/s/cm^2) (default 1.e-4)
   norm_min:   lower limit of line amplitude (default 0)
   norm_max:   upper limit of line amplitude (default 0.1)
   sigma:      line width (presuming a gaussian) (default 0.01)
   sigma_min:  lower limit of line width (default 0)
   sigma_max:  upper limit of line width (default 0.1)
   vtherm:     Maxwellian velocity for voigt model (default 1000 km/s)
   vtherm_min: lower limit of Maxwellian velocity (default 0.1)
   vtherm_max: upper limit of Maxwellian velocity (default 100000 km/s)
   redshift:   line redshift, in redshifted model (default 0)
   unit:       String for units of input center/sigma (default "keV")
`  ); %}}}
      return;
   }
   else if(_NARGS==1)
   {
      lstruct=();
   }
   else if(_NARGS>1)
   {
      message("Too many input variables: only single structure expected");
      variable args = __pop_list(_NARGS);
      return;
   }

   if(typeof(lstruct) != Struct_Type && lstruct != NULL)
   {
      print(lstruct);
      message("Input must be a structure and/or qualifiers");
      return;
   }

   variable line_id="a", center=6.4, center_min,
   center_max, norm=1.e-5, norm_min=0., norm_max=0.1,
   sigma=0.01, sigma_min=0., sigma_max=0.1, vtherm=50.,
   vtherm_min=0.1, vtherm_max=500., redshift=0., unit="kev",
   u_info, center_e, center_e_min, center_e_max,
   center_a, center_a_min, center_a_max,
   sigma_e, sigma_e_min, sigma_e_max,
   sigma_a, sigma_a_min, sigma_a_max;

   try{ unit = get_struct_field(lstruct,"unit"); }
   catch AnyError: {}
   unit = qualifier("unit","keV");
   u_info = unit_info(unit);
   
   try{ line_id = get_struct_field(lstruct,"line_id"); }
   catch AnyError: {}
   line_id = qualifier("line_id","a");
   
   try{ center = get_struct_field(lstruct,"center"); }
   catch AnyError: {}
   center = qualifier("center",center);

   try{ sigma = get_struct_field(lstruct,"sigma"); }
   catch AnyError: {}
   sigma = qualifier("sigma",sigma);

   try{ sigma_min = get_struct_field(lstruct,"sigma_min"); }
   catch AnyError: {}
   sigma_min = qualifier("sigma_min",sigma_min);

   try{ sigma_max = get_struct_field(lstruct,"sigma_max"); }
   catch AnyError: {}
   sigma_max = qualifier("sigma_max",sigma_max);

   if(sigma_max<sigma) sigma_max=sigma;
   if(sigma_min>sigma) sigma_min=0;
   center_min = center-sigma_max;
   center_max = center+sigma_max;
 
   try{ center_min = get_struct_field(lstruct,"center_min"); }
   catch AnyError: {}
   center_min = qualifier("center_min",center_min);

   try{ center_max = get_struct_field(lstruct,"center_max"); }
   catch AnyError: {}
   center_max = qualifier("center_max",center_max);

   try{ norm = get_struct_field(lstruct,"norm"); }
   catch AnyError: {}
   norm = qualifier("norm",norm);

   try{ norm_min = get_struct_field(lstruct,"norm_min"); }
   catch AnyError: {}
   norm_min = qualifier("norm_min",norm_min);

   try{ norm_max = get_struct_field(lstruct,"norm_max"); }
   catch AnyError: {}
   norm_max = qualifier("norm_max",norm_max);   

   if(norm_max<norm) norm_max=norm;
   if(norm_min>norm) norm_min=0;

   try{ vtherm = get_struct_field(lstruct,"vtherm"); }
   catch AnyError: {}
   vtherm= qualifier("vtherm",vtherm);

   try{ vtherm_min = get_struct_field(lstruct,"vtherm_min"); }
   catch AnyError: {}
   vtherm_min = qualifier("vtherm_min",vtherm_min);

   try{ vtherm_max = get_struct_field(lstruct,"vtherm_max"); }
   catch AnyError: {}
   vtherm_max = qualifier("vtherm_max",vtherm_max);

   if(vtherm_max<vtherm) vtherm_max=vtherm;
   if(vtherm_min>vtherm) vtherm_min=vtherm;

   try{ redshift = get_struct_field(lstruct,"redshift"); }
   catch AnyError: {}
   redshift = qualifier("redshift",redshift);

   variable iw = where(line_name != line_id);
   line_name = [line_name[iw],line_id];

   % Convert variables to keV and Angstrom units

   if(u_info.is_energy)
   {
      center_e = center*u_info.scale;
      center_e_min = center_min*u_info.scale;
      center_e_max = center_max*u_info.scale;

      sigma_e = sigma*u_info.scale;
      sigma_e_min = sigma_min*u_info.scale;
      sigma_e_max = sigma_max*u_info.scale;

      center_a = _A(center_e);
      center_a_min = _A(center_e_max);
      center_a_max = _A(center_e_min);

      sigma_a = (sigma_e/center_e)*center_a;
      sigma_a_min = (sigma_e_min/center_e)*center_a;
      sigma_a_max = (sigma_e_max/center_e)*center_a;
   }
   else
   {
      center_a = center*u_info.scale;
      center_a_min = center_min*u_info.scale;
      center_a_max = center_max*u_info.scale;

      sigma_a = sigma*u_info.scale;
      sigma_a_min = sigma_min*u_info.scale;
      sigma_a_max = sigma_max*u_info.scale;

      center_e = _A(center_a);
      center_e_min = _A(center_a_max);
      center_e_max = _A(center_a_min);

      sigma_e = (sigma_a/center_a)*center_e;
      sigma_e_min = (sigma_a_min/center_a)*center_e;
      sigma_e_max = (sigma_a_max/center_a)*center_e;
   }

   line_energy = [line_energy[iw], center_e];
   line_energy_min = [line_energy_min[iw], center_e_min];
   line_energy_max = [line_energy_max[iw], center_e_max];

   line_norm = [line_norm[iw], norm];
   line_norm_min = [line_norm_min[iw], norm_min];
   line_norm_max = [line_norm_max[iw], norm_max];

   line_sigma = [line_sigma[iw], sigma_e];
   line_sigma_min = [line_sigma_min[iw], sigma_e_min];
   line_sigma_max = [line_sigma_max[iw], sigma_e_max];

   line_vtherm = [line_vtherm[iw], vtherm];
   line_vtherm_min = [line_vtherm_min[iw], vtherm_min];
   line_vtherm_max = [line_vtherm_max[iw], vtherm_max];

   line_redshift = [line_redshift[iw], redshift];

   % ISIS Line Models
   alias_fun("gauss",ftype_name["gauss"]+line_name[-1];
             names = ["area [photons/s/cm^2]", "center [A]", "sigma [A]"],
             values = [norm, center_a, sigma_a],
             freeze = [0, 0, 0],
             min = [norm_min, center_a_min, sigma_a_min],
             max = [norm_max, center_a_max, sigma_a_max]);

   alias_fun("egauss",ftype_name["egauss"]+line_name[-1];
             names = ["area [photons/s/cm^2]", "center [keV]", "sigma [keV]"],
             values = [norm, center_e, sigma_e],
             freeze = [0, 0, 0],
             min = [norm_min, center_e_min, sigma_e_min],
             max = [norm_max, center_e_max, sigma_e_max]);

   variable fwhm_scl = 2*sqrt(2*log(2));
   alias_fun("Lorentz",ftype_name["lorentz"]+line_name[-1];
             names = ["area [photons/s/cm^2]", "center [A]", "fwhm [A]"],
             values = [norm, center_a, fwhm_scl*sigma_a],
             freeze = [0, 0, 0],
             min = [norm_min, center_a_min, fwhm_scl*sigma_a_min],
             max = [norm_max, center_a_max, fwhm_scl*sigma_a_max]);

   fwhm_scl *= 2.*PI;
   alias_fun("voigt",ftype_name["voigt"]+line_name[-1];
             names = ["norm [photons/s/cm^2]", "energy [keV]", "fwhm [keV]", "vtherm [km/s]"],
             values = [norm, center_e, fwhm_scl*sigma_e, vtherm],
             freeze = [0, 0, 0, 0],
             min = [norm_min, center_e_min, fwhm_scl*sigma_e_min, vtherm_min],
             max = [norm_max, center_e_max, fwhm_scl*sigma_e_max, vtherm_max]);

   % XSPEC Line Models
   alias_fun("zgauss",ftype_name["zgauss"]+line_name[-1];
             names = ["norm [photons/s/cm^2]", "LineE [keV]", "Sigma [keV]","Redshift"],
             values = [norm, center_e, sigma_e, redshift],
             freeze = [0, 0, 0, 1],
             min = [norm_min, center_e_min, sigma_e_min, -0.999],
             max = [norm_max, center_e_max, sigma_e_max, 10]);

   alias_fun("gaussian",ftype_name["gaussian"]+line_name[-1];
             names = ["norm [photons/s/cm^2]", "LineE [keV]", "Sigma [keV]"],
             values = [norm, center_e, sigma_e],
             freeze = [0, 0, 0],
             min = [norm_min, center_e_min, sigma_e_min],
             max = [norm_max, center_e_max, sigma_e_max]);

   alias_fun("agauss",ftype_name["agauss"]+line_name[-1];
             names = ["norm [photons/s/cm^2]", "LineE [A]", "Sigma [A]"],
             values = [norm, center_a, sigma_a],
             freeze = [0, 0, 0],
             min = [norm_min, center_a_min, sigma_a_min],
             max = [norm_max, center_a_max, sigma_a_max]);

   alias_fun("zagauss",ftype_name["zagauss"]+line_name[-1];
             names = ["norm [photons/s/cm^2]", "LineE [A]", "Sigma [A]","Redshift"],
             values = [norm, center_a, sigma_a, redshift],
             freeze = [0, 0, 0, 1],
             min = [norm_min, center_a_min, sigma_a_min, -0.999],
             max = [norm_max, center_a_max, sigma_a_max, 10]);

}

public define add_line()
{
   variable args;
   if(_NARGS==3)
   {
      args=__pop_list(_NARGS);
      variable lname=args[0], fname=args[1], ftype=args[2];

      if(typeof(lname) != String_Type || 
         typeof(fname) != String_Type || 
         typeof(ftype) != String_Type    )
      { message("Input component, model, and line model type values as strings"); return; }

      ftype=strlow(ftype);

      if(length(where([etypes,atypes]==ftype))==0)
      { message("Unrecognized line model type"); return; }

      if(length(where(line_name==lname))==0)
      { message("Line component not initialized"); return; }

      if(assoc_key_exists(mod_values,fname) && 
         length(where(mod_comps[fname]==lname))>0 )
      { message("Line already in model"); return; }

      ifnot( assoc_key_exists(mod_values,fname) )
      {
         mod_comps[fname]=String_Type[0]; 
         mod_values[fname]=Double_Type[0]; 
         mod_types[fname]=String_Type[0]; 
         mod_subtract[fname]=Char_Type[0]; 
         mod_i_a_d[fname]=String_Type[0]; 
         if(length(etypes==ftype))
         {
            mod_eorder[fname]=1; 
         }
         else
         {
            mod_eorder[fname]=0  ; 
         }
      }

      variable i, iline = where(line_name==lname); 
 
      mod_comps[fname] = [mod_comps[fname],line_name[iline]];
      mod_values[fname] = [mod_values[fname],line_energy[iline]];
      mod_types[fname] = [mod_types[fname],ftype];
      if(qualifier_exists("subtract"))
      {
         mod_subtract[fname] = [mod_subtract[fname],1];
      }
      else
      {
         mod_subtract[fname] = [mod_subtract[fname],0];
      }

      variable i_a_d = qualifier("id","1");
      mod_i_a_d[fname] = [mod_i_a_d[fname],i_a_d];

      mod_eorder[fname] = qualifier("eorder",mod_eorder[fname]);

      variable isrt=array_sort(mod_values[fname]);
      ifnot(mod_eorder[fname]) isrt=reverse(isrt);
      mod_comps[fname] = mod_comps[fname][isrt];
      mod_values[fname] = mod_values[fname][isrt];
      mod_types[fname] = mod_types[fname][isrt];
      mod_subtract[fname] = mod_subtract[fname][isrt];
      mod_i_a_d[fname] = mod_i_a_d[fname][isrt];

      variable meval="public define "+fname+"(){ variable y;";

      if(mod_subtract[fname][0])
      {
         meval += "y=-minus("+mod_i_a_d[fname][0]+")*"+ftype_name[mod_types[fname][0]]
                  +mod_comps[fname][0]+"("+mod_i_a_d[fname][0]+");";
      }
      else
      {
         meval += "y="+ftype_name[mod_types[fname][0]]
                  +mod_comps[fname][0]+"("+mod_i_a_d[fname][0]+");";
      }

      _for i (1,length(isrt)-1,1)
      {
         if(mod_subtract[fname][i])
         {
            meval += "y-=minus("+mod_i_a_d[fname][i]+")*"+ftype_name[mod_types[fname][i]]
                     +mod_comps[fname][i]+"("+mod_i_a_d[fname][i]+");";
         }
         else
         {
           meval += "y+="+ftype_name[mod_types[fname][i]]
                    +mod_comps[fname][i]+"("+mod_i_a_d[fname][i]+");";
         }
      }
      meval += "return y; }";
      eval(meval);
      return;
   }
   else
   {
      args = __pop_list(_NARGS);
      () = fprintf(fp_stde, "%s\n", %{{{
`
  add_line(line_id, function_name, line_type;
              id=string, eorder=#, subtract);

  Create or add to the fit function function_name() (a model that
  returns a sum of line components) a line component identified by the
  line_id string. The line_id must first have been created by the
  init_line function.

  The added component must be one of the following ISIS or XSPEC
  models: gauss, egauss, Lorentz, voigt, gaussian, agauss, zgauss,
  zagauss, as identified by the line_type string.

  The added component will have the prefix: g_, eg_, L_, v_, xg_, ag_,
  zg_, or zag_, respectively, coupled with line_id(id). (Default
  prefixes can be changed with line_model_prefix function.)

  Qualifiers:

   id:       String to identify the model instance, e.g., "1", "2",
             "Isis_Active_Dataset" (default "1")
   eorder:   if=0, the line components will be arranged in wavelength order
             (default is last used value, or 0 or 1, based upon whether 
             the first added model is naturally wavelength or energy 
             based, respectively)
   subtract: if present, subtract, not add, the line component
             (indicated by minus(id) function, which is just an
             aliased constant function, before the line component)
`  ); %}}}
      return;
   }
}

public define delete_line()
{
   variable args;
   if(_NARGS==2)
   {
      args=__pop_list(_NARGS);
      variable lname=args[0], fname=args[1];

      if(typeof(lname) != String_Type || 
         typeof(fname) != String_Type    )
      { message("Input component and model type values as strings"); return; }

      if(length(where(line_name==lname))==0)
      { message("Line component not initialized"); return; }

      ifnot(assoc_key_exists(mod_comps,fname))
      { message("Summed line function not initialized"); return; }

      variable i, iw=where(mod_comps[fname]!=lname);

      mod_comps[fname] = mod_comps[fname][iw];
      mod_values[fname] = mod_values[fname][iw];
      mod_types[fname] = mod_types[fname][iw];
      mod_subtract[fname] = mod_subtract[fname][iw];
      mod_i_a_d[fname] = mod_i_a_d[fname][iw];

      variable meval="public define "+fname+"(){ variable y;";

      if(mod_subtract[fname][0])
      {
         meval += "y=-minus("+mod_i_a_d[fname][0]+")*"+ftype_name[mod_types[fname][0]]
                  +mod_comps[fname][0]+"("+mod_i_a_d[fname][0]+");";
      }
      else
      {
         meval += "y="+ftype_name[mod_types[fname][0]]
                  +mod_comps[fname][0]+"("+mod_i_a_d[fname][0]+");";
      }

      _for i (1,length(mod_comps[fname])-1,1)
      {
         if(mod_subtract[fname][i])
         {
            meval += "y-=minus("+mod_i_a_d[fname][i]+")*"+ftype_name[mod_types[fname][i]]
                     +mod_comps[fname][i]+"("+mod_i_a_d[fname][i]+");";
         }
         else
         {
           meval += "y+="+ftype_name[mod_types[fname][i]]
                    +mod_comps[fname][i]+"("+mod_i_a_d[fname][i]+");";
         }
      }
      meval += "return y; }";
      eval(meval);
      return;
   }
   else
   {
      args=__pop_list(_NARGS);
      () = fprintf(fp_stde, "%s\n", %{{{
`
  delete_line(line_id, function_name);

  Delete the line component, identified by the string line_id
  (previously initialized by the init_line function), from the summed
  fit function (previously created by the funtion add_line) given by
  the string, function_name.
`  ); %}}}
      return;
   }
}

public define find_line()
{
   variable args, lo_in, hi_in, lo_tru, hi_tru, e_lo, e_hi;
   if(_NARGS==2)
   {
      args = __pop_list(_NARGS);
      lo_in=args[0], hi_in=args[1];

      if( typeof(lo_in)!=Char_Type && typeof(lo_in)!=Integer_Type && 
          typeof(lo_in)!=Float_Type && typeof(lo_in)!=Double_Type && 
          typeof(hi_in)!=Char_Type && typeof(hi_in)!=Integer_Type && 
          typeof(hi_in)!=Float_Type && typeof(hi_in)!=Double_Type         )
       { message("Inputs need to be numbers"); return; }

      if(lo_in<hi_in)
      {
         lo_tru = lo_in;
         hi_tru = hi_in;
      }
      else if(lo_in>hi_in)
      {
         lo_tru = hi_in;
         hi_tru = lo_in;
      }
      else
      { message("Inputs identical"); return; } 

      variable u_info = unit_info(qualifier("unit","kev")); 
  
      if(u_info.is_energy)
      {
         e_lo = lo_in*u_info.scale;
         e_hi = hi_in*u_info.scale;
      }
      else
      {
         e_lo = _A(hi_in*u_info.scale);
         e_hi = _A(lo_in*u_info.scale);
      }

      variable iw = where(e_lo <= line_energy <= e_hi);
      if(length(iw) == 0)
      {
         ifnot(qualifier_exists("noprint"))
	 { message("No lines found"); }
         return NULL;
      }
      else
      {
         variable isrt=array_sort(line_energy[iw]);
	 ifnot(qualifier_exists("noprint"))
         {
            variable i;
            _for i (0,length(isrt)-1,1)
            {
               if(u_info.name == "keV" || u_info.name == "A" || u_info.name == "Angstrom")
               {
                  () = fprintf(fp_stde," %-12S   %14.4g keV   %14.4g A\n",
                  line_name[iw[i]], line_energy[iw[i]], _A(line_energy[iw[i]]));
               }
               else
               {
                  variable le;
                  if(u_info.is_energy)
                  {
                     le = line_energy[iw[i]]/u_info.scale;
                  }
                  else
                  {
                     le = _A(line_energy[iw[i]])/u_info.scale;
                  }
                  () = fprintf(fp_stde," %-12S   %14.4g keV   %14.4g A   %14.4g   "
                               +u_info.name+"\n",
                  line_name[iw[i]], line_energy[iw[i]], _A(line_energy[iw[i]]), le);
               }
	    }      
	 }
         return line_name[iw[isrt]];
      }   
   }
   else
   {
      args = __pop_list(_NARGS);
      () = fprintf(fp_stde, "%s\n", %{{{
`
  line_ids = find_line(a,b; unit=string, noprint)
  
   Return an array of line_ids from the database, originally
   initialized with init_line, with default centers between bounds
   (a,b), and further print this list to the screen.

  Qualifiers:

   unit:    Units of the bounds (default "keV")
   noprint: If present, don't print the list to the screen
`  ); %}}}
      return;
   }
}

public define save_line_models()
{
   if(_NARGS==1)
   {
      variable i, file=();
      if(typeof(file)!=String_Type)
       { message("Input filename needs to be a string"); return; }

      variable fps = fopen(file,"w");

      () = fprintf(fps,"%% Line Database: \n");
      _for i (0,length(line_name)-1,1)
      {
         () = fprintf(fps,"init_line(;line_id=\"%s\", center=%g, min=%g, max=%g,\n",
                      line_name[i],line_energy[i],line_energy_min[i],line_energy_max[i]);
         () = fprintf(fps,"           norm=%g, norm_min=%g, norm_max=%g,\n",
                      line_norm[i],line_norm_min[i],line_norm_max[i]);
         () = fprintf(fps,"           sigma=%g, sigma_min=%g, sigma_max=%g,\n",
                      line_sigma[i],line_sigma_min[i],line_sigma_max[i]);
         () = fprintf(fps,"           vtherm=%g, vtherm_min=%g, vtherm_max=%g,\n",
                      line_vtherm[i],line_vtherm_min[i],line_vtherm_max[i]);
         () = fprintf(fps,"           unit=\"keV\");\n");
      }
      () = fprintf(fps,"\n");

      variable mod_def, lmkeys = assoc_get_keys(mod_comps);

      foreach mod_def (lmkeys)
      {
         () = fprintf(fps,"%% Line Model: \"%s\" \n",mod_def);
         _for i (0,length(mod_comps[mod_def])-1,1)
         {
            if(mod_subtract[mod_def][i])
            {
               () = fprintf(fps,"add_line(\"%s\",\"%s\",\"%s\"; id=\"%s\", eorder=%i, subtract);\n",
                            mod_comps[mod_def][i],mod_def,mod_types[mod_def][i],
                            mod_i_a_d[mod_def][i],mod_eorder[mod_def]);
            }
            else
            {
               () = fprintf(fps,"add_line(\"%s\",\"%s\",\"%s\"; id=\"%s\", eorder=%i);\n",
                            mod_comps[mod_def][i],mod_def,mod_types[mod_def][i],
                            mod_i_a_d[mod_def][i],mod_eorder[mod_def]);
            }
         }
         () = fprintf(fps,"\n");
      }
      () = fclose(fps);
   }
   else
   {
      variable args = __pop_list(_NARGS);
      () = fprintf(fp_stde, "%s\n", %{{{
`
  save_line_models(file_name);
  
   Create a S-lang script in file_name that when executed initializes
   all lines and models, as they currently exist, from the init_line
   and add_line/delete_line functions.
`  ); %}}}
      return;
   }
}


#iftrue
public define rolling_fit()
{
   if(_NARGS==1)
   {
      variable i=0, comp, mod_name=();
      if(typeof(mod_name)!=String_Type)
       { message("Input model name needs to be a string"); return; }

      ifnot(assoc_key_exists(mod_comps,mod_name))
       { message("Model does not exist"); return; }

      variable fv_hold = Fit_Verbose;
      Fit_Verbose = -1;
      variable nloop = qualifier("nloop",1);
      variable tol = qualifier("tol",0.);

      variable line_mods=ftype_name[mod_types[mod_name]]+
                         mod_comps[mod_name]+"("+
                         mod_i_a_d[mod_name]+")";

      variable mod_pars, mod_par_num, 
               mod_pars_thaw = Array_Type[length(mod_comps[mod_name])],
               unfrz_line_pars = String_Type[0];

      foreach comp (line_mods)
      {
         mod_pars_thaw[i] = String_Type[0];
         foreach mod_par_num (get_fun_params(comp))
         {
            variable gp = get_params[mod_par_num];
            variable ptype = strlow(strtok(gp.name,".")[-1]);

            % As long as the parameter is not tied or frozen, add it
            % to the global list of line parameters that will be
            % frozen on each loop
            if( gp.freeze==0 && gp.tie==0 )
            {
               unfrz_line_pars = [unfrz_line_pars,gp.name];

               % But only if it is in a category of fit parameters
               % that we want to fit, add it to the list of parameters
               % for that component to be unfrozen for the local fit.
               if( (((ptype=="norm") && qualifier_exists("nonorm")) != 1) && 
                   (((ptype=="sigma") && qualifier_exists("nosigma")) != 1) && 
                   (((ptype=="fwhm") && qualifier_exists("nosigma")) != 1) && 
                   (((ptype=="vtherm") && qualifier_exists("novtherm")) != 1)  )
               {
                  mod_pars_thaw[i] = [mod_pars_thaw[i],get_params[mod_par_num].name];
               }
            }
	 }
         i++;
      }

      variable itot=i-2, info;

      () = eval_counts(&info);
      variable delta_stat=1.e16, start_stat = info.statistic;
      variable looped = 0;

      while(looped < nloop && delta_stat > tol)
      {
         _for i (0,itot,1)
         {
            freeze(unfrz_line_pars);
            if(length(mod_pars_thaw[i]))
            {
               thaw(mod_pars_thaw[i]);
               () = fit_counts;
            }
         }
         thaw(unfrz_line_pars);
         () = eval_counts(&info);
         delta_stat = start_stat - info.statistic;
         start_stat = info.statistic;
         looped++;
      }
      Fit_Verbose = fv_hold;
   }
   else
   {
      variable args = __pop_list(_NARGS);
      () = fprintf(fp_stde, "%s\n", %{{{
`
  rolling_fit(line_model_name);
  
   For a line model created with the init_line/add_line/delete_line
   functions and identified by the string line_model_name, perform a
   fit where the continuum+one line component at a time is left
   unfrozen.  Serially loop through a fit of all of the line
   components a specified number of times, or until the fit statistic
   stops changing (comparing full loop to full loop) within a
   specified tolerance range.

  Qualifiers:

   nloop =    Number of fit loops through the line components. 
              (Default nloop=1)
   tol =      Tolerance on changes in fit statistic changes.  If the
              fit statistic has changed by less than this value since
              the beginning of the loop, stop looping regardless of the 
              nloop parameter value. (Default tol=0.)
   nonorm =   Do not fit line normalizations, regardless of whether or
              not they are unfrozen parameters.
   nosigma =  Do not fit line widths (Sigma or fwhm), regardless of 
              whether or not they are unfrozen parameters.
   novtherm = Do not fit line therm widths, regardless of whether or
              not they are unfrozen parameters.
`  ); %}}}
      return;
   }
}  
#endif

define query_user(string_ref)
{
   variable buf, status;
 
   forever {
      if ( fprintf(stdout, "  Enter line name: ") < 0 ) return -1;
      status = fgets(&buf, stdin);

      if(status>1)
      {
         @string_ref = strtrim(buf,"\n");
         return 0;
      }

      if (fprintf(stdout, "  Come again??\n")<0) return -1;
  }
}

public define add_cursor_line()
{
   variable args;
   if(_NARGS==2)
   {
      args = __pop_list(_NARGS);
      variable fun_name = args[0], line_type = args[1];
   }
   else
   {
      () = fprintf(fp_stde, "%s\n", %{{{
`
  add_cursor_line(fun_name, line_type [; qualifiers])

  Initialize (via the init_line function) and add (via the add_line
  function) lines of type line_type to the line model named fun_name
  using the cursor clicking on a plot.  Lines will be given default
  names based upon their centroid value.

  The user can either add only the line centroid values (with line
  lower/upper limits automatically added), or click on three points to
  correspond to line limits and centroid.  Allowed line widths will be
  determined from the lower/upper limits.

  Default keyboard values for lower/centroid/upper values are mapped
  to keyboard triplets: x/c/v ; s/d/f ; w/e/r; 1/2/3; n/m/, ; j/k/l ;
  i/o/p ; 9/0/- with the spacebar or mouse click always being the
  centroid. q quits the session (and cannot be overwritten).

  Qualifiers:

   Any from init_line or add_line will be passed through and retain
   their usual default values, except that center, center_min,
   center_max, sigma, sigma_min, sigma_max will be generated from the
   plot clicks.  Thus, reuse of those qualifiers will generate an
   error.

   res             : Fraction resolution of detector (default=1.e-3),
                     used to guess lower/upper limits [=line center X
                     (1+/-res*res_factor)] and line width [=line
                     center*res*sigma_scale] and line width limits
                     [=line width*res*sigma_factor_lo/_hi]
   res_factor      : Used to determine line limits (default=2)
   sigma_scale     : Used to determine line widths (default=0.25)
   sigma_factor_hi : Used to determine line width upper limit 
                     (default=8)
   sigma_factor_lo : Used to determine line width lower limit
                     (default=0)
   limits          : If exists, determine limits/widths from cursor
                     clicks
   lo_limit        : Change default keys for lower limit clicks (string 
                     array)
   hi_limit        : Change default keys for upper limit clicks
                     (string array)
   center          : Change default keys for centroid clicks (string
                     array)
   prompt_name     : If exists, prompt for name of each line
   fit             : If exists, freeze all parameters except for newly
                     defined line, and perform a fit
 ` ); %}}} 
      return;
   }
 
   variable x, y, ch,line_id;
   variable res = qualifier("res",1.e-3);
   variable res_factor = qualifier("res_factor",2);
   variable sigma_scale = qualifier("sigma_scale",0.25);
   variable sigma_factor_hi = qualifier("sigma_factor_hi",8.);
   variable sigma_factor_lo = qualifier("sigma_factor_lo",0.);
   variable limits = 0;
   if( qualifier_exists("limits") ) limits=1;
   variable lo_limit = [qualifier("lo_limit",["x","s","w","1","n","j","i","9"])];
   variable hi_limit = [qualifier("hi_limit",["v","f","r","3",",","l","p","-"])];
   variable center = [qualifier("center",["c","d","e","2","m","k","o","0","A"," "])];
   variable n=1, j=0;
   if (limits) n=3;

   if ( length(where("q"==lo_limit)) || length(where("q"==hi_limit)) ||
      length(where("q"==center)) ) return;

   variable line_center, line_lo, line_hi;

   forever{
      line_center=-1.; line_lo=-1.; line_hi=-1;
      loop(n)
      {
         ifnot(j) () = printf("Enter line center, min, and max (any order): \n\n");
         if(j==2) { j=0; } else j++;
         
         cursor(&x,&y,&ch);
         if(get_plot_info.xlog) x = 10^x;
         if(ch=="q") break;
         if(length(where(ch==lo_limit)))
         {
            line_lo=x;
         }
         else if(length(where(ch==hi_limit)))
         {
            line_hi=x;
         }
         else if(length(where(ch==center)))
         {
            line_center=x;
         }
         if (n==1) line_center=x;
      }
      if (ch=="q" || line_center <=0) break;

      if (line_lo <= 0 || line_lo >= line_center) 
        line_lo = line_center*(1-res*res_factor);

      if (line_hi <= 0 || line_hi <= line_center) 
        line_hi = line_center*(1+res*res_factor);
      
      () = printf("Selected line value: %.4f + %.4f - %.4f \n\n",
           line_center,line_hi-line_center,line_center-line_lo);

      if(qualifier_exists("prompt_name"))
      {
         () = query_user(&line_id);
      }
      else
      {
         % Multiply wavelength or energy by 1000, truncate, and call
         % that the line name
         line_id = string(int(1.e3*x));
      }

      init_line( ;__qualifiers, line_id=line_id, center=line_center, 
        center_min=line_lo, center_max=line_hi, 
        sigma=line_center*res*sigma_scale, 
        sigma_min=line_center*res*sigma_scale*sigma_factor_lo, 
        sigma_max=line_center*res*sigma_scale*sigma_factor_hi );

      add_line(line_id, fun_name, line_type; __qualifiers);

      if(qualifier_exists("fit"))
      {
         () = eval_counts;
         freeze("*");
         thaw("*"+line_id+"*");
         () = fit_counts;
      }
   }
}






#endif

%
% Code from src/slang/physics/j_ff.sl
%
define j_ff() {
%!%+
%\function{j_ff}
%\synopsis{Calculates the emission coefficient for free-free emission (bremsstrahlung)}
%\usage{Double_Type j = j_ff(nu,T);}
%
%\qualifiers{
%\qualifier{Z}{average nuclear charge (default=1)}
%\qualifier{ne}{electron particle density (cm^-3; default: 1e10)}
%\qualifier{np}{proton particle density (cm^-3; default: 1e10)}
%}
%\description
%    This function returns the emission coefficient for free-free radiation
%    (bremsstrahlung). The frequency, nu, is measured in Hz and can be an array,
%    the temperature T is measured in K.
%
%    Note: per Kirchhoff's law the total bremsstrahlung spectrum from
%    a slab of size R is
%    B_nu(nu,T)*(1.-exp(-R*alpha_ff(nu,T)))
%
%\seealso{alpha_ff}
%!%-

    variable nu,T;
    
    (nu,T)=();

    variable Z=qualifier("Z",1.);
    variable ne=qualifier("ne",1e10);
    variable ni=qualifier("ni",1e10);
    
    variable alpha=alpha_ff(nu,T;ne=ne,ni=ni,Z=Z);
    variable jnu=alpha*B_nu(nu,T);

}

%
% Code from src/slang/date/jd2mjd.sl
%
define jd2mjd ()
%!%+
%\function{jd2mjd}
%\synopsis{Convert Julian Dates to Modified Julian Dates.}
%\usage{jd2mjd(JD);}
%\description
%    Helper function to convert Julian Dates to MJD by
%    subtracting off 2400000.5.
%\seealso{mjd2jd}
%!%-
{
   variable JD;
   switch(_NARGS)
     { case 1: JD = (); }
     { help(_function_name()); return; }
   return JD-2400000.5;
};

define JD2MJD ()
%!%+
%\function{JD2MJD}
%\synopsis{Convert Julian Dates to Modified Julian Dates.}
%\usage{JD2MJD(JD);}
%\description
%    Helper function to convert Julian Dates to MJD by
%    subtracting off 2400000.5.
%\seealso{j2dmjd, mjd2jd}
%!%-
{
   variable JD;
   switch(_NARGS)
     { case 1: JD = (); }
     { help(_function_name()); return; }
   return JD-2400000.5;
};

%
% Code from src/slang/date/jd2year.sl
%
define jd2year()
%!%+
%\function{jd2year}
%\synopsis{transform a JD into a fractional year}
%\usage{Double_Type year = jd2year(Double_Type JD);}
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%}
%\description
%  NOTE: This routine takes into account leap years. This is good for
%  plotting, but this means that there are slight phase shifts that could be
%  problems when doing time series analysis. For such purposes best use
%  EpochofJD!
%\seealso{EpochofJD}
%!%-
{
  variable JD;
  switch(_NARGS)
  { case 1: JD = (); }
  { help(_function_name()); return; }

  if (qualifier_exists("mjd")) {
      JD=mjd2jd(JD);
  }
  
  if(typeof(JD)==Array_Type) {
      return array_map(Double_Type, &jd2year, JD);
  }
  
  variable date=DateOfJD(JD);

  variable jd0=JDofDate(date.year  ,1,1,0,0,0.);
  variable jd1=JDofDate(date.year+1,1,1,0,0,0.);

  return date.year+(JD-jd0)/(jd1-jd0);

}

%
% Code from src/slang/astronomy/jpleph.sl
%

% only activate if module vector could be loaded
#ifeval __get_reference("vector")!=NULL
%
% Functions to handle JPL old-style (binary, non-SPK)
% ephemerides
%
% Joern Wilms, Sept. 2019 - Dec. 2019
%


define jpl_bineph2fits(fitsfile,ephfile) 
%!%+
%\function{jpl_bineph2fits}
%\synopsis{convert a binary JPL ephemeris to FITS}
%\usage{jpl_bineph2fits(ephfile,fitsfile);}
%\description
% This function reads the JPL binary ephemeris in ephfile
% and writes it to the fitsfile.
%
% The JPL ephemeris must be in the "old-style" format.
% SPK files are NOT supported.
%
% The function has to be run on a machine with the same byte
% sex as the ephemeris (JPL typically distributes ephemerides
% for all common types of byte sex). It does not do a lot of
% checking, however, it has been used to successfully
% convert ALL available JPL old-style ephemerides into FITS.
%
% Which also means you should probably ask Joern Wilms for
% the location of the relevant FITS file rather than using this
% function...
%
%\seealso{jpl_eph,jpl_eph_vec}
%!%-
{
    %
    % read data into a large structure
    %
    variable ifile=fopen(ephfile,"rb");

    variable sizeof_double=8;
    variable sizeof_int32=4;

    variable tit;
    ()=fread_bytes(&tit,84,ifile); % read the title

    ()=fseek(ifile,2652, SEEK_SET);  % buffer starts 2652 bytes past the start

    % header constants 
    variable buf;
    % validity interval
    ()=fread_bytes(&buf,sizeof_double,ifile);
    variable ephem_start=unpack("d",buf);
    ()=fread_bytes(&buf,sizeof_double,ifile);
    variable ephem_end=unpack("d",buf);
    ()=fread_bytes(&buf,sizeof_double,ifile);
    variable ephem_step=unpack("d",buf);

    % number of constants
    ()=fread_bytes(&buf,sizeof_int32,ifile);
    variable ncon=unpack("k",buf);
    
    if (ncon>65535) {
	print("byte sex of binary file is wrong!");
	exit;
    }

    % AU
    ()=fread_bytes(&buf,sizeof_double,ifile);
    variable au=unpack("d",buf);
    % earth moon mass ratio
    ()=fread_bytes(&buf,sizeof_double,ifile);
    variable emrat=unpack("d",buf);

    % value from jpleph, I don't know where it is in the header
    variable pvsun_t=-1e80;
    variable ipt=Integer_Type[40,3];
    variable i;
    _for i(0,39,1) {
	()=fread_bytes(&buf,sizeof_int32,ifile);
	ipt[i/3,i mod 3]=unpack("k",buf);
    }

    variable ephemeris_version=atoi(substr(tit,27,strlen(tit)));

    % see jpleph.cpp
    % I'm not sure why the reading isn't done in a different way
    ipt[12,0]=ipt[12,1];
    ipt[12,1]=ipt[12,2];
    ipt[12,2]=ipt[13,0];

    % decision on whether the ephemeris includes TT-TDB data
    if (ephemeris_version >=430 && ncon!=400 ) {
	if (ncon>400) {
	    ()=fseek(ifile, (ncon-400)*6,SEEK_CUR);
	}
	_for i(0,6) {
	    ()=fread_bytes(&buf,sizeof_int32,ifile);
	    ipt[13+i/3,i mod 3]=unpack("k",buf);
	}
    } else {
	ipt[13,0]=-1;
    }

    % check if there's garbage in these values
    if (ipt[13,0] != ipt[12,0]+ipt[12,1]*ipt[12,2]*3 ||
    ipt[14,0] != ipt[13,0]+ipt[13,1]*ipt[13,2]*3 ) {
	()=printf("Info: No valid TT-TDB pointer\n");
	ipt[[13:14],[0:2]]=0;
    }

    variable kernel_size=4;
    _for i(0,14,1) {
	variable dim=3; % most parameters have xyz values
	if (i==11) {
	    dim=2; % nutations: 2 values
	}
	if (i==14) {
	    dim=1; % tdt-tt
	}
	kernel_size+=2*ipt[i,1]*ipt[i,2]*dim;
    }

    variable recsize=kernel_size*4;
    variable ncoeff=kernel_size/2;

    % read constant numbers and values
    ()=fseek(ifile,recsize,SEEK_SET);
    variable val=Double_Type[ncon];
    variable tmp=fread(&val,Double_Type,ncon,ifile);

    variable bufoffset=84*3+400*6+5*sizeof_double+41*sizeof_int32;
    variable nam=String_Type[ncon];
    ()=fseek(ifile,84*3,SEEK_SET);
    _for i(0,ncon-1,1) {
	if (i==400) {
	    ()=fseek(ifile,bufoffset,SEEK_SET);
	}
	()=fread_bytes(&buf,6,ifile);
	nam[i]=strtrim(buf);
    }

    % 
    variable idnam=[
    "Mercury", 
    "Venus",   
    "Earth",   
    "Mars",    
    "Jupiter", 
    "Saturn",  
    "Uranus",  
    "Neptune", 
    "Pluto",   
    "Moon",    
    "Sun",     
    "Nutations",               
    "Lunar libration (Euler) angles",
    "Lunar angular velocity",
    "TT-TDB"                   
    ];

    variable origin="J. Wilms, joern.wilms@sternwarte.uni-erlangen.de";
    variable creator="bineph2fits.sl, version 1.0";

    variable de1_keys=struct {
	EXTNAME="DE1",
	ORIGIN=origin,
	CREATOR=creator
    };

    variable de1_data=struct {
	Cname=nam[*],
	Cvalue=val[*]
    };

    variable de2_keys=struct {
	EXTNAME="DE2",
	ORIGIN=origin,
	CREATOR=creator
    };

    variable onum=[0:length(idnam)-1];
    variable de2_data=struct {
        Object=idnam[*],
        Pointer=ipt[onum,0],
        NumCoeff=ipt[onum,1],
        NumSubIntv=ipt[onum,2]
    };

    % write the first two extensions
    variable fptr=fits_open_file(fitsfile,"c");
    fits_write_binary_table(fptr,"DE1",de1_data,de1_keys);
    fits_write_date(fptr);
    fits_write_chksum(fptr);
    fits_write_binary_table(fptr,"DE2",de2_data,de2_keys);
    fits_write_date(fptr);
    fits_write_chksum(fptr);

    % now read ephemeris records
    % we need to do this piece by piece since some ephemerides
    % can be too long for isis to handle in one sitting

    variable nrec=int((ephem_end-ephem_start)/ephem_step);
    
    fits_create_binary_table(fptr,"DE3",nrec,["ChebCoeffs"],[sprintf("%iD",ncoeff)],[""]);
    fits_update_key(fptr,"EXTNAME","DE3");
    fits_update_key(fptr,"ORIGIN",origin);
    fits_update_key(fptr,"CREATOR",creator);
    fits_update_key(fptr,"TSTART",ephem_start);
    fits_update_key(fptr,"TSTOP",ephem_end);
    fits_update_key(fptr,"TIMEDEL",ephem_step);
    fits_update_key(fptr,"TIMESYS","TT");
    fits_update_key(fptr,"TIMEUNIT","d");
    fits_update_key(fptr,"JDREF",0.);
    fits_update_key(fptr,"TDIM1",sprintf("(%i)",ncoeff));

    variable nr;
    _for nr(0,nrec-1,1) {
        % we could probably omit this
        ()=fseek(ifile,(nr+2)*recsize,SEEK_SET);
        buf=Double_Type[ncoeff];
        ()=fread(&buf,Double_Type,ncoeff,ifile);
        variable coeffs=@buf;
        ()=_fits_write_col(fptr,1,nr+1,1,coeffs[*]);
    }
    fits_write_date(fptr);
    fits_write_chksum(fptr);

    fits_close_file(fptr);
}

private variable jplname=Assoc_Type[Int_Type];

private define jpl_name2jplid(name) {
    %
    % convert the name of an object to the
    % object's number in the old-style JPL ephemeris
    %
    
    %  1 - Mercury
    %  2 - Venus 
    %  3 - Earth (geocenter)
    %  4 - Mars (system barycenter)
    %  5 - Jupiter (system barycenter)
    %  6 - Saturn (system barycenter)
    %  7 - Uranus (system barycenter)
    %  8 - Neptune (system barycenter)
    %  9 - Pluto (system barycenter)
    % 10 - Moon
    % 11 - Sun
    % 12 - Solar System Barycenter
    % 13 - Earth-Moon Barycenter
    % 14 - 1980 IAU nutation angles
    % 15 - Lunar libration (Euler) angles
    % 16 - Lunar angular velocity
    % 17 - TT-TDB (at geocenter)

    if (assoc_key_exists(jplname,"initialized")==0) {
	jplname["MERCURY"]=  1;
	jplname["VENUS"]=  2;
	jplname["EARTH"]=  3;
	jplname["MARS"]=  4;
	jplname["JUPITER"]=  5;
	jplname["SATURN"]=  6;
	jplname["URANUS"]=  7;
	jplname["NEPTUNE"]=  8;
	jplname["PLUTO"]=  9;
	jplname["MOON"]= 10;
	jplname["SUN"]= 11;
	jplname["SOLAR SYSTEM BARYCENTER"]= 12;
	jplname["SSB"]= 12;
	jplname["EARTH-MOON BARYCENTER"]= 13;
	jplname["EARTH-MOON"]= 13;
	jplname["EMB"]= 13;
	jplname["1980 IAU NUTATION ANGLES"]= 14;
	jplname["IAU1980"]= 14;
	jplname["LUNAR LIBRATION (EULER) ANGLES"]= 15;
	jplname["LUNAR LIBRATION"]= 15;
	jplname["LUNAR ANGULAR VELOCITY"]= 16;
	jplname["TT-TDB"]=17;
	jplname["initialized"]=-1;
    }

    if (assoc_key_exists(jplname,strup(name))) {
	return jplname[strup(name)];
    }

    throw UsageError(sprintf("Object name does not exist: %s\n",name));
}

define jpl_initeph() 
%!%+
%\function{jpl_initeph}
%\synopsis{initialize a JPL ephemeris}
%\usage{Struct_Type results = jpl_initeph(file);}
%\description
% This function initializes a JPL ephemeris for later use with
% function jpl_eph.
%
% For an empty argument, the function defaults to DE430 as distributed
% with HEASOFT. This is probably what you want (and since jpl_eph
% initializes to that ephemeris, this means that you probably do NOT
% want to call this function in the first place).
%
% The ephemeris file must be in FITS format, using the HEASOFT
% convention (see function jpl_bineph2fits if you need to generate
% such a file from one of the older JPL ephemeris distributions).
%
% The function first searches for the ephemeris file as given,
% and then searches for the file by prepending ${LHEA_DATA}/
% or ${LHEASOFT}/refdata/. 
%
%\qualifiers{
%  \qualifier{nocache}{Do not cache the ephemeris data in memory,
%                        i.e., keep the FITS file open and seek on it.
%                        This is faster if calculating just a few positions
%                        (e.g., for a few weeks), otherwise, reading the
%                        whole ephemeris first [the default] is faster,
%                        but a memory hog, i.e., it won't work for DE431.}
%}
%\seealso{jpl_eph,jpl_bineph2fits}
%
%!%-
{
    variable ephfile;
    switch(_NARGS)
    {case 0: ephfile="JPLEPH.430";}
    {case 1: ephfile=();}
    { throw UsageError, sprintf("%s: Invalid number of arguments\n",_function_name()); }

    % search for the file
    variable fitsfile=ephfile;
    if (stat_file(fitsfile)==NULL) {
        if (getenv("LHEA_DATA")!=NULL) {
	    fitsfile=getenv("LHEA_DATA")+"/"+ephfile;
	} else {
	    if (getenv("LHEASOFT")!=NULL) {
		fitsfile=getenv("LHEASOFT")+"/refdata/"+ephfile;
	    } else {
		throw UsageError,sprintf("%s: eph not given and LHEA_DATA or LHEASOFT not set!",_function_name());
	    }
	}
	if (stat_file(fitsfile)==NULL) {
	    throw UsageError,sprintf("%s: cannot find ephfile %s",_function_name,ephfile);
	}
     }
    variable fptr=fits_open_file(fitsfile,"r");

    variable ret=_fits_movnam_hdu(fptr,_FITS_BINARY_TBL,"DE1",0);
    variable hdr=fits_read_header(fptr);
    variable const=fits_read_table(fptr);

    variable nn=where(const.cname=="AU");
    if (length(nn)==0) {
	throw RunTimeError,sprintf("%s: Ephemeris file %s does not contain a value for AU.\n",_function_name,fitsfile);
    }
    variable au=const.cvalue[nn[0]];

    nn=where(const.cname=="EMRAT");
    if (length(nn)==0) {
	throw RunTimeError,sprintf("%s: Ephemeris file %s does not contain a value for the Earth-Moon mass ratio.\n",_function_name,fitsfile);
    }
    variable emrat=const.cvalue[nn[0]];

    nn=where(const.cname=="CLIGHT");
    if (length(nn)==0) {
	throw RunTimeError,sprintf("%s: Ephemeris file %s does not contain a value for the speed of light.\n",_function_name,fitsfile);
    }
    variable clight=const.cvalue[nn[0]];

    nn=where(const.cname=="GMS");
    if (length(nn)==0) {
	throw RunTimeError,sprintf("%s: Ephemeris file %s does not contain a value for GM for the Sun.\n",_function_name,fitsfile);
    }
    variable gms=const.cvalue[nn[0]];
    
    ()=_fits_movnam_hdu(fptr,_FITS_BINARY_TBL,"DE2",0);
    variable pars=fits_read_table(fptr);

    ()=_fits_movnam_hdu(fptr,_FITS_BINARY_TBL,"DE3",0);
    variable jdstart=fits_read_key(fptr,"TSTART");
    variable jdstop=fits_read_key(fptr,"TSTOP");
    variable dt=fits_read_key(fptr,"TIMEDEL");

    if (qualifier_exists("nocache")) {
	return struct {
	    pars=pars,
	    fptr=fptr,
	    jdstart=jdstart,
	    jdstop=jdstop,
	    dt=dt,
	    au=au,
	    emrat=emrat,
	    clight=clight,
	    gms=gms,
	    cache=0
	};
    }

    variable coeffs=fits_read_table(fptr);
    fits_close_file(fptr);

    return struct {
	pars=pars,
	coeffs=coeffs,
	jdstart=jdstart,
	jdstop=jdstop,
	dt=dt,
	au=au,
	emrat=emrat,
	clight=clight,
	gms=gms,
	cache=1
    };
}


% cache for chebyshev polynomial
private variable jpl_ephcache=struct {
    tc=-2., 
    cheby=Double_Type[30],    % this is larger than in all published JPL eph 
    dercheby=Double_Type[30], % that used the old JPL format
    ncf=-1,
    chebcoeffs=[0], % cache for ephemeris chebyshev coefficients
    recno=-1        % (only if reading from a file)
};

private define jplinter(jd,jplindex,eph) {
    %
    % perform the interpolation in the JPL ephemeris
    % This is a private routine that should never be called
    % from anywhere else than jpl_eph
    %
    % Because of the way we're reading from the ephemeris and
    % because of the caching of the Chebyshev polynomials, this
    % routine is NOT safe to be used in parallel_map and other
    % parallel routines!!
    %
    
    % ephemeris record to use
    variable fltrecno=(jd[0]-eph.jdstart + jd[1])/eph.dt;
    
    variable recno=typecast(fltrecno,Long_Type);
    variable trel=fltrecno-recno;
    % boundary case (plus possible roundoff correction)
    if (trel<=0.) {
	trel=1.;
	recno=recno-1;
    }
    
    %
    % parameters for this component
    % ncf: number of coefficients per component
    % na : number of sub-intervals in the full interval

    if (length(eph.pars.numcoeff)<= jplindex-1) {
        throw UsageError(sprintf("Ephemeris has no information for jplindex=%i\n"),jplindex);
    }
    
    variable ncf=eph.pars.numcoeff[jplindex-1];
    variable na =eph.pars.numsubintv[jplindex-1];

    if (ncf==0) {
        throw UsageError(sprintf("Ephemeris has no information for jplindex=%i\n"),jplindex);
    }
    
    % number of coefficients for the jplindex
    variable ncm=3;
    % nutations
    if (jplindex==12) {
	ncm=2;
    }
    % TT-TDB
    if (jplindex==15) {
	ncm=1;
    }

    variable tmp=na*trel;
    variable ell=int(tmp);
    variable tc=2.0*(tmp-typecast(tmp,Long_Type))-1.0;
    
    if (int(tmp) >= na) {
	ell-=1;
	tc+=2.;
    }
    
    % here we should have -1<=tc <=+1

    %
    % evaluate the chebyshev polynomials
    % (we reuse earlier values from our cache if applicable)
    %
    variable i;
    variable t2=2*tc;
    variable ncfstart=2;

    if (tc==jpl_ephcache.tc) {
	ncfstart=jpl_ephcache.ncf+1;
    } else {
	jpl_ephcache.tc=tc;
	jpl_ephcache.cheby[0]=1.;
	jpl_ephcache.cheby[1]=tc;
	jpl_ephcache.dercheby[0]=0.0;
	jpl_ephcache.dercheby[1]=1.0;
    }
    
    % calculate the missing Chebyshev coefficients if necessary
    if (ncfstart<ncf) {
	% note: dercheby[ncf] is not needed, but this makes
	%  the overall logic of the code simpler
	_for i(ncfstart,ncf,1) {
	    jpl_ephcache.cheby[i]=t2*jpl_ephcache.cheby[i-1] - jpl_ephcache.cheby[i-2];
	    jpl_ephcache.dercheby[i]=t2*jpl_ephcache.dercheby[i-1]+2.0*jpl_ephcache.cheby[i-1]-jpl_ephcache.dercheby[i-2];
	}
	jpl_ephcache.ncf=ncf;
    }
    
    % apply the chebyshev polynomial
    variable j;
    variable pos=Double_Type[ncm];
    variable vel=Double_Type[ncm];
    variable vfac=2.*na/(eph.dt*86400.);

    %
    % readable version of the code below
    %
    % % running the loop backwards ensures that cndx goes through
    % % chebcoeffs in one direction only (better for the cache in
    % % compiled languages)
    % _for i(ncm-1,0,-1) {
    % 	% NB: -1 is there because slang and C arrays are 0 based, not 1 based
    % 	variable cndx=eph.pars.pointer[jplindex-1]+ncf*(i+ell*ncm)-1;
	
    % 	% this is faster than 2 loops, even though the vel loop would
    % 	% only have to go to 1 and not to 0 (since dercheby[0]=0)
    % 	_for j(ncf-1,0,-1) {
    % 	    variable jj=cndx+j;
    % 	    pos[i]+=jpl_ephcache.cheby[j]*eph.coeffs.chebcoeffs[recno,jj];
    % 	    vel[i]+=jpl_ephcache.dercheby[j]*eph.coeffs.chebcoeffs[recno,jj];
    % 	}
    % 	vel[i]*=vfac;
    % }

    % The complicated handling of the cache here is because
    % copying ephchebcoeffs over to the local variable chebcoeffs
    % speeds up the code by 10%
    variable chebcoeffs;
    if (eph.cache) {
	chebcoeffs=eph.coeffs.chebcoeffs[recno,*];
    } else {
	if (jpl_ephcache.recno!=recno) {
	    jpl_ephcache.chebcoeffs=fits_read_cell(eph.fptr,1,recno+1);
	    jpl_ephcache.recno=recno;
	}
	chebcoeffs=jpl_ephcache.chebcoeffs[*];
    }

    % NB: -1 for 0 based arrays, 0 for 1-based arrays
    variable cndx=eph.pars.pointer[jplindex-1]+ncf*((ell+1)*ncm)-1;
    _for i(ncm-1,0,-1) {
	_for j(ncf-1,0,-1) {
	    cndx--;
	    pos[i]+=jpl_ephcache.cheby[j]*chebcoeffs[cndx];
	    vel[i]+=jpl_ephcache.dercheby[j]*chebcoeffs[cndx];
	}
	vel[i]*=vfac;
    }
    
    return [pos,vel];
    
    %
    % acceleration (if necessary)
    % variable der2cheby=Double_Type[ncf];
    % der2cheby[0]=0.;
    % der2cheby[1]=0.;
    % _for i(2,ncf-1,1) {
    % 	der2cheby[i]=4*eph.dercheby[i-1]+t2*der2cheby[i-1]-der2cheby[i-2];
    % }

    % variable acc=Double_Type[ncm];
    % _for i(0,ncm-1,1) {
    % 	% NB: -1 is there because slang and C arrays are 0 based, not 1 based
    % 	cndx=eph.pars.pointer[jplindex-1]+ncf*(i+ell*ncm)-1;
    
    % 	_for j(ncf-1,0,-1) {
    % 	    jj=cndx+j;
    % 	    acc[i]+=der2cheby[j]*eph.coeffs.chebcoeffs[recno,jj];
    % 	}
    % 	acc[i]*=vfac*vfac;
    % }
}

% save jpl ephemeris (used if eph argument of jpl_eph is not given)
private variable jpl_ephsave;

define jpl_eph ()
%!%+
%\function{jpl_eph}
%\synopsis{calculate planet positions using the JPL ephemeris}
%\usage{Struct_Type results = jpl_eph(jd,ta,ce);}
%\description
%    This function returns the ephemeris  of target body ta
%    as seen from center body ce for the time jd (the JD in the
%    TDB system). 
%
%    NOTE: To calculate apparent planet positions for observing,
%    use the higher level function planetpos.
%
%    NOTE2: If you want a routine that is compatible with
%    function vsop87, uses TT and is vector safe, use  jpl_eph_vec
%
%    jd is either a Double or a 2-element array (to avoid roundoff
%    errors the array can include, e.g., the integer and fractional
%    parts of the JD). It is given in in TDB. If you need to calculate
%    positions for a time TT (=TAI+32.184s), convert with function
%    tt2tdb.
%    Note, however, that for all practical applications except for ms
%    pulsar timing, you can assume that TT=TDB. 
%
%    The target and center bodies ta and ce are identified as outlined
%    below.
%
%    By default, the routine initializes to DE430 as distributed with HEASOFT
%    (and found either in  $LHEA_DATA/JPLEPH.430 or in
%    $LHEASOFT/refdata/JPLEPH.430).
%    The ephemeris in HEASOFT is only a subset of DE430, valid for about 1950--2050.
%    For these ephemerides, the precision for the decades around AD2000 is on
%    the order of meters to tens of meters, and milliarcseconds (see below). 
%    Joern provides full files for DE430, valid for the time between
%    AD1550 and AD2650, and for DE431, which covers years -13200 to +17191
%    (with lower precision for the moon).
%
%    Use the ephemeris qualifier if you want to use another ephemeris.
%
%    For planets, the function returns an array containing the xyz-position,
%    and the velocity components in the ICRF. For angle variables
%    (only contained in a few of the JPL ephemerides), the array elements are
%    the angles and their  derivatives, and for the difference TT-TDB, the time
%    difference and its derivative. Units are km and km/s, rad and deg/day [!],
%    or s and s/day [!]. 
%
%    Target bodies or angles are either identified by a number as listed
%    below or by the indicated case-insensitive string. Note that not all
%    ephemeris files contain all of the possibilities!
%
%     1 - Mercury
%     2 - Venus 
%     3 - Earth
%     4 - Mars
%     5 - Jupiter
%     6 - Saturn
%     7 - Uranus
%     8 - Neptune
%     9 - Pluto
%    10 - Moon
%    11 - Sun
%    12 - Solar System Barycenter
%    13 - Earth-Moon Barycenter
%    14 - 1980 IAU nutation angles
%    15 - Lunar libration (Euler) angles
%    16 - Lunar angular velocity
%    17 - TT-TDB
%
%    where 3 is the Earth's geocenter. For planets with moons, the 
%    position returned is with respect to the barycenter of the respective
%    planetary system.
% 
%    TT-TDB is measured at the geocenter.
%
%    The following abbreviations are also allowed:
%    12: SSB
%    13: Earth-Moon or EMV
%    14: IAU1980
%    15: Lunar Libration
%
% For targets 14-17 the center is ignored
% 
% See Folkner et al., 2014, IPN Progress Report 42-196 for a description
% of the precision that is reached with DE430/DE431. The figures in
% this publication indicate that for the timeframe from about 1970 onwards,
% where precise astrometry and space probe data exist, a precision
% on the order of several 10s of meters is reached (and apparent
% angular positions are precise to milli-arcseconds).
%
% This function is NOT array safe, and it CANNOT be used in
% a parallel_map or other parallelized functions. This is
% because global caches are used that we cannot avoid if we
% want decent performance for single calls. Sorry
%
% \qualifiers{
%    \qualifier{ephemeris}{structure defining the ephemeris, initialized with
%                            jpl_initeph (see there).}
% }
% 
%\seealso{jpl_eph_vec,jpl_initeph, planetpos}
%!%-
{
    variable jdin,ta,ce;

    switch(_NARGS)
    { case 2: (jdin,ta)=(); }
    { case 3: (jdin,ta,ce)=();  }
    { throw UsageError, sprintf("%s: Invalid number of arguments\n",_function_name()); }

    variable eph;
    if (qualifier_exists("ephemeris")) {
	eph=qualifier("ephemeris");
    } else {
	% ephemeris not given: use the cached emphemeris or open an
	% new one
	% (NB: jpl_initeph with no argument defaults to DE430. It
	% also does all of the error handling
	% and also search in LHEA_DATA and LHEASOFT)
	ifnot (__is_initialized(&jpl_ephsave)) {
	    jpl_ephsave=jpl_initeph();
	}
	eph=jpl_ephsave;
    }

    variable target=ta;
    
    if (typeof(target)==String_Type) {
	target=jpl_name2jplid(target);
    }

    if (_NARGS==2) {
	if (target<14 || target>17) {
	    { throw UsageError, sprintf("%s: Must give center\n",_function_name()); }
	}
	ce=0; % dummy value
    }

    variable center=ce;
    if (typeof(center)==String_Type) {
	center=jpl_name2jplid(center);
    }

    variable jd=@jdin;
    if (typeof(jd)==Double_Type) {
	jd=[jd,0.];
    } else {
	if (typeof(jd)==Array_Type && length(jd)!=2) {
	    throw UsageError,sprintf("%s: jd must be double scalar or double array with 2 entries",_function_name());
	}
    }

    % check validity
    variable jdlow=jd[0]+jd[1];
    if (jdlow<eph.jdstart or jdlow>eph.jdstop) {
	throw UsageError,sprintf("%s: Time %15f is outside validity interval (JD %15f to %15f)\n",_function_name(),jdlow,eph.jdstart,eph.jdstop);
    }
    
    
    if (target<=0 || target > 17) {
	throw UsageError,"Target must be between 1 and 17";
    }

    if (center<=0) {
	if (target<14) {
	    throw UsageError,"Must give center info!";
	}
    }
    if (center>13) {
	if (target<14) {
	    throw UsageError,"Center value does not make sense";
	}
    }

    if (target<=13 && center>=14) {
	throw UsageError,"Center not possible for angles";
    }

    if (center<=13) {
	if (center==target) {
	    return [0.,0.,0.,0.,0.,0.];
	}
    }
    
    %
    % Map object index to index in JPL file
    %
    % (so lots of strange bureaucracy)
    %
    variable jplindex;
    variable pv;
    variable pv2;
    variable pvb;
    variable pvm;

    % ignore center for the angles
    if (14<=target<=17) {
	center=0;
    }
    
    if (target==14) {
	% nutation
	jplindex=12;

	if (eph.pars.pointer[jplindex-1]==0 || eph.pars.numcoeff[jplindex-1]*eph.pars.numsubintv[jplindex-1]==0) {
	    % no nutation values present
	    return -1;
	}
	pv=jplinter(jd,jplindex,eph);
	pv[2]*=86400.; % rad/s -> rad/d
	pv[3]*=86400.; % rad/s -> rad/d
	return pv;
    }

    if (target==15 || target==16) {
	% Lunar libration (Euler) angles
	if (target==15) {
	    jplindex=13;
	} else {
	    jplindex=14;
	}

	if (eph.pars.pointer[jplindex-1]==0 || eph.pars.numcoeff[jplindex-1]*eph.pars.numsubintv[jplindex-1]==0) {
	    % no lunar euler angle values present
	    return -1;
	}
	pv=jplinter(jd,jplindex,eph);
	pv[3]*=86400.; % rad/s -> rad/d
	pv[4]*=86400.; % rad/s -> rad/d
	pv[5]*=86400.; % rad/s -> rad/d
	return pv;
    }

    if (target==17) {
	% TT-TDB
	jplindex=15;

	if (eph.pars.pointer[jplindex-1]==0 || eph.pars.numcoeff[jplindex-1]*eph.pars.numsubintv[jplindex-1]==0) {
	    % no TT-TDB value present
	    return -1;
	}
	pv=jplinter(jd,jplindex,eph);
	pv[1]*=86400.; % s/s -> s/d
	return pv;
    }
    
    %
    % Earth/moon cases
    %
    % target=moon, center=geocenter
    if (target==10 && center==3) {
	jplindex=10; % moon wrt. earth
	return jplinter(jd,jplindex,eph);
    }

    % target=earth, center=moon
    if (target==3 && center==10) {
	jplindex=10; % moon wrt. earth
	return -jplinter(jd,jplindex,eph);
    }

    % earth or moon is involved
    if (target==3 || center==3 || target==10 || center==10) {
	variable facte=1./(1.+eph.emrat);
	variable factm=facte-1.0;

	pvm=jplinter(jd,10,eph); % moon
	pvb=jplinter(jd,3,eph); % E+M Barycenter

	if (target==3 || target==10) {
	    
	    if (center<12) {
		% any of the other planets
		% (note e/m bary has already been taken care of)
		pv2=jplinter(jd,center,eph);
	    } else {
		% EM barycenter
		if (center==13) {
		    pv2=jplinter(jd,3,eph);
		} else {
		    pv2=[0.,0.,0.,0.,0.,0.];
		}
	    }
	    if (target==3) {
		% earth
		pv=pvb-facte*pvm-pv2;
	    } else {
		% moon
		pv=pvb-factm*pvm-pv2;
	    }
	    return pv;
	}

	% target is not earth or moon
	if (target<12) {
	    pv=jplinter(jd,target,eph);
	} else {
	    if (target==13) {
		pv=jplinter(jd, 3,eph);
	    } else {
		pv=[0.,0.,0.,0.,0.,0.];
	    }
	}

	if (center==3) {
	    pv=pv-(pvb-facte*pvm);
	} else {
	    pv=pv-(pvb-factm*pvm);
	}
	return pv;
    }

    % cases not involving Earth or Moon
    % (but might include the EM barycenter)
    variable pv1;
    if (target<12) {
	pv1=jplinter(jd,target,eph);
    } else {
	if (target==13) {
	    pv1=jplinter(jd,3,eph);
	} else {
	    pv1=[0.,0.,0.,0.,0.,0.];
	}
    }

    if (center<12) {
	pv2=jplinter(jd,center,eph);
    } else {
	if (center==13) {
	    pv2=jplinter(jd,3,eph);
	} else {
	    pv2=[0.,0.,0.,0.,0.,0.];
	}
    }
    return pv1-pv2;
}

% forward declaration (necessary for recursive call later!)
define jpl_eph_vec();

define jpl_eph_vec()
%!%+
%\function{jpl_eph_vec}
%\synopsis{high precision planetary positions from the JPL ephemerides, vsop87 compatible}
%\usage{Struct_Type jpl_eph_vec(JD,planet;qualifiers)}
%\description
% This routine provides an interface to jpl_eph which allows to
% calculate heliocentric ephemerides from the JPL ephemerides.
% It is compatible with a subset of the functions that are also
% available from function vsop87.
%
% Use planetpos to calculate apparent planetary positions.
%
% The routine returns high precision planetary positions for
% all planets for a given (array) of times (in TT).
%
% Depending on the qualifiers the routine can return rectangular
% coordinates (X,Y,Z) or ecliptical coordinates
% (longitude, latitude, distance) in a heliocentric coordinate system
% for epoch J2000 (the ICRF).
%
% Contrary to vsop87, orbital elements cannot be calculated. 
%
% By default the function returns rectangular heliocentric
% coordinates in a structure with elements x, y, z, vx, vy, vz.
% Here x, y, z are in an ecliptical(!) coordinate system for
% the epoch J2000 (but see the equatorial qualifier), where the x-axis
% is in the direction of the vernal equinox, the z-axis points
% to the ecliptical North pole, and the y-axis forms
% a right handed coordinate system with the x- and z-axes.
% The default units are AU and AU/day (but see the mks and cgs qualifiers).
%
% For heliocentric coordinates the planet can have the values
% "Mercury", "Venus", "Earth", "Mars", "Jupiter", "Saturn",
% "Uranus", "Neptune", and "Pluto" (the latter is not included
% in vsop87!). The value "Earth-Moon" will return the position
% of the Earth-Moon barycenter.
%
% The default unit of time is the Julian Date in TT, to be
% compatible with vsop87. Use the qualifier tdb to avoid the
% conversion TT->TDB, which for pretty much all practical
% applications except for ms pulsar timing results in no
% appreciable difference in the results but is slightly faster.
%
% Use the mjd qualifier to use MJD rather than JD as input.
%
% The default function uses DE430 from HEASOFT, which is valid
% for the time between 1950 and 2050. Joern can provide ephemeris
% files for the full DE430, valid AD1550 and AD2650, and
% also for DE431, which covers years -13200 to +17191
% (with lower precision for the moon). See the ephemeris qualifier.
%
% When using JPL ephemerides, do NOT use this function in
% a parallel_map or other parallelized computations. This is
% because global caches are used that we cannot avoid if we
% want decent performance for single calls. Sorry.
%  
%\qualifiers{
% \qualifier{tdb}{The time argument is in TDB, not TT}
% \qualifier{mjd}{The time argument is in MJD, not in JD}
% \qualifier{barycentric}{Return coordinates with respect to the solar system barycenter rather
%                    than heliocentric coordinates.}
% \qualifier{spherical}{Return spherical coordinates in the ecliptical coordinate system.
%                       In this case the structure returned contains the elements
%                       lat, lon, r, latdot, londot, rdot, where lat and lon are the ecliptical
%                       latitude and longitude (in rad), r is the distance in AU, and
%                       latdot, londot, rdot are their time derivatives (in rad/day or
%                       AU/day, respectively.)}
% \qualifier{equatorial}{return coordinates in the ICRS2 coordinate system, not the
%                       equatorial one (vsop87 does not have this option)}
% \qualifier{mean_equinox}{Coordinates are given in the frame defined by the
%                   mean equinox and ecliptic of the date.
%                   The default is to return J2000 coordinates.}
% \qualifier{deg}{All angle quantities to be returned are in deg, or deg/day.}
% \qualifier{mks}{All distances returned by the function are given in meters or m/s, depending
%                 on the deg qualifier, angles are returned in rad and rad/s or deg and deg/s.}
% \qualifier{cgs}{Dito, but use centimeters instead of meters.}
% \qualifier{ephemeris}{Ephemeris structure initialized with jpl_initeph.}
% \qualifier{forcearray}{Force the tags in the returned struct to be arrays.
%                          The default is that they are arrays only if JD
%                          is an array.}
%}
%\seealso{vsop87,jpl_initeph,jpl_eph}
%!%-
{
    variable jdin,planet;
    switch(_NARGS)
    {
	case 2: (jdin,planet) = ();
    }
    { help(_function_name()); return; }

    % warn people if they expect the elements qualifier
    if (qualifier_exists("elements")) {
	throw UsageError,sprintf("%s: Elements qualifier is not supported\n",_function_name());
    }

    % if ephemeris is set, then pass it through
    variable eph;
    if (qualifier_exists("ephemeris")) {
	eph=qualifier("ephemeris");
    } else {
	eph=jpl_initeph();
    }
    
    % take care of JD being an array
    if (typeof(jdin)==Array_Type) {
	% initialize the ephemeris only once!
	variable qual=__qualifiers();
	if (not qualifier_exists("ephemeris")) {
	    qual=struct_combine(qual,struct {ephemeris=eph});
	}

	% we cannot use parallel map on jpl_eph_vec, so let's do this the
	% old style way
	variable res={};
	variable i;
	_for i(0,length(jdin)-1,1) {
	    list_append(res,jpl_eph_vec(jdin[i],planet;;qual));
	}
	return struct_array_2_struct_of_arrays(list_to_array(res));
    }
    
    % center body: default is heliocentric
    variable cb="Sun";
    if (qualifier_exists("barycentric")) {
	cb="SSB";
    }

    variable JD=@jdin;
    if (qualifier_exists("mjd")) {
	JD=mjd2jd(JD);
    }

    % input is TT, so find the corresponding TDB
    if (not qualifier_exists("tdb")) {
        % do conversion with ephemeris IF the ephemeris provides
        % the time conversion.  otherwise default to Bretagnon.
        if (length(eph.pars.numcoeff)>16 && eph.pars.numcoeff[16]>0) {
            JD=tt2tdb(JD;ephemeris=eph);
        } else {
            JD=tt2tdb(JD);
        }
    }

    % from here on time is in TDB

    % get xyz coordinates
    variable pos=jpl_eph(JD,planet,cb;ephemeris=eph);

    variable p=vector(pos[0],pos[1],pos[2]);
    variable v=vector(pos[3],pos[4],pos[5]);
    
    if (qualifier_exists("cgs")) {
    	p=p*1e5;
    	v=v*1e5;
    } else {
	if (qualifier_exists("mks")) {
     	    p=p*1000.;
     	    v=v*1000.;
	} else {
     	    p=p/eph.au;
     	    v=v*(86400./eph.au);
	}
    }

    % go to position for the equinox of date?
    if (qualifier_exists("mean_equinox")) {
	variable Bmat=gcrs2j2000_matrix(); % frame bias matrix
	variable Pmat=precession_matrix(JD);
	variable mat=Pmat*Bmat;
	p=mat*p;
	v=mat*v;
    }
    
    % go to ecliptical coordinate system?
    if (not qualifier_exists("equatorial")) {
	% conversion matrix from equatorial to ecliptical coordinates
	variable epsilon;
	if (qualifier_exists("mean_equinox")) {
	    epsilon=obliquity(JD);
	} else {
	    epsilon=obliquity(2451545.0);
	}
	variable tmat=matrix33_rot(1,epsilon);
	p=tmat*p;
	v=tmat*v;
    }

    % convert to angles?
    if (not qualifier_exists("spherical")) {
	if (qualifier_exists("forcearray")) {
	    return struct { x=[p.x], y=[p.y], z=[p.z], vx=[v.x],vy=[v.y],vz=[v.z] };
	} else {
	    return struct { x=p.x, y=p.y, z=p.z, vx=v.x,vy=v.y,vz=v.z };
	}
    }
    variable r,lon,lat;
    (r,lon,lat)=vector_to_spherical(p);

%    r=sqrt(p.x*p.x+p.y*p.y+p.z*p.z);
%    lat=asin(p.z/r);
%    lon=atan2(p.y,p.x);
    variable rdot=(v.x*p.x+v.y*p.y+v.z*p.z)/r;
    variable latdot=(r*v.z-p.z*rdot)/(r^2*sqrt((1.-(p.z/r))*(1.+(p.z/r))));
    variable londot=(p.x*v.y-p.y*v.x)/(p.x^2+p.y^2);
    
    if (qualifier_exists("deg")) {
	lat*=180./PI;
	lon*=180./PI;
	latdot*=180./PI;
	londot*=180./PI;
    }
    
    if (qualifier_exists("forcearray")) {
	return struct { lat=[lat], lon=[lon], r=[r], latdot=[latdot], londot=[londot], rdot=[rdot] } ;
    }
    return struct { lat=lat, lon=lon, r=r, latdot=latdot, londot=londot, rdot=rdot } ;
	
}

define planetpos();

define planetpos()
%!%+
%\function{planetpos}
%\synopsis{Calculate the apparent position of the Sun, Moon, and the planets
% (Mercury,..., Neptune, and Pluto).}
%\usage{(r,alp,del) = planetpos(JD;qualifiers);}
%\altusage{(r,alp,del) = planetpos(JD,object;qualifiers);}
%\description
%
% Returns the true geocentric distance (in AU), and apparent right ascension
% and declination of a planet in the geocentric celestial reference system for the
% date JD (in TT or TDB [the default]).
% 
% Use the center-qualifier to choose another center object.
%
% The default coordinate system is the J2000.0 (ICRS) system, to include the
% effects of precession and nutation to get "apparent" positions, use the
% of_date qualifier, to get the mean equinox position (w/o nutation), use
% of_date together with the mean_equinox qualifier.
%
% The true geocentric distance is the distance the planet has at the time of the
% calculation, while the apparent position is corrected for light travel time
% effects. This is the default (and identical to the Astronomical Almanac).
% If the lighttravel qualifier is set, the function returns the
% distance the light traveled. In most cases, this is what you want.
% Use the "no_apparent" to switch off the correction for light travel time
% (e.g., useful for the calculation of heliocentric coordinates, where
% traditionally light travel time effects are are not taken into account)
%
% The default is the JPL ephemeris DE430 (as used in the astronomical almanac
% and as distributed with HEASOFT). Use the ephemeris qualifier to switch to
% another ephemeris.
%
% Alternatively, the less precise VSOP87 can be used (but not for Pluto and
% the Moon). There is no computing time advantage to doing so, however, but
% vsop87 is valid for several 1000 years around AD2000, while most JPL
% ephemerides are limited to less than that.
%
% This function has been checked against the values tabulated in the
% Astronomical Almanac for 2014 and 2015, which also uses DE430.
% The results match exactly.
%
% JD and object can be arrays. In this case the return value is
% an array of Vectors or three arrays of distance, RA and DEC,
% sorted by object first and by date next.
%
%\qualifiers{
% \qualifier{object}{a string (valid values: Sun, Mercury, Venus,
%                      Mars, Jupiter, Saturn, Uranus, Neptune, Pluto, Moon,
%                      default: Sun).}
% \qualifier{center}{center object for the position. Default: Earth.
%                      Either a string (same as "object"), or a
%                      3D or 6D long double array. If 3D: position
%                      in AU in a barycentric coordinate system in
%                      ICRS coordinates (xyz), if 6D: position and
%                      velocity (AU/D) in a barycentric coordinate system.}
% \qualifier{topocentric}{return topocentric positions; if this qualifier
%                      is of type Vector_Type, then the xyz coordinates are
%                      interpreted as the position of the observer in the
%                      IERS 2010 coordinate system, in units of meters (see
%                      function geographic2vector). Otherwise, a structure with
%                      geographic coordinates is used, use tags lat, lon for
%                      the geographic position (in degrees, east is positive),
%                      and tag height as height (in meters) over the WGS 84
%                      geoid (as appropriate for GPS coordinates).}
% \qualifier{of_date}{if set, return the apparent position in the geocentric
%                       celestial reference system for the ecliptic and
%                       equinox of date (this is what is listed in the
%                       astronomical almanac).}
% \qualifier{mean_equinox}{if set together with of_date, return coordinates
%                       for the mean equinox of date (i.e., ignore nutation).}
% \qualifier{no_apparent}{ignore light travel time effects}
% \qualifier{ecliptical}{return coordinates in ecliptical coordinates,
%                        i.e., return r,lon,lat (or the xyz vector in ecliptical coords)
%                       (also takes of_date qualifier into account)}
% \qualifier{vector}{return a geocentric/object centric direction (!) vector
%                      for the position.}
% \qualifier{deg}{return the angles in degrees (default: radian).}
% \qualifier{mks}{return distance in m (default: AU)}
% \qualifier{cgs}{return distance in cm (default: AU)}
% \qualifier{lighttravel}{return distance the light traveled (default: true geocentric dist)}
% \qualifier{mjd}{the given date is the MJD}
% \qualifier{tt}{the given date is TT (default: TDB, difference only matters for pulsar work)}
% \qualifier{vsop87}{use the VSOP87 ephemeris rather than the JPL ephemeris, does not work
%                      for Pluto and the Moon.}
%}
%\seealso{jpl_initeph,vsop87,geographic2vector}
%!%-
%
{

    % fixme: avoid roundoff errors in time conversion
    % TODO: we should move the initializers for the ephemeris and tt
    % before the array calls, since this will save some time
    % 

    variable jdin,objects;

    switch (_NARGS)
    { case 1: jdin=(); }
    { case 2: (jdin,objects)=(); }
    { help(_function_name()); return;}
    
    ifnot (__is_initialized(&objects)) {
	objects=qualifier("object","Sun");
    } else {
	if (qualifier_exists("object")) {
	    throw UsageError,sprintf("%f: cannot give object as argument and qualifier at the same time\n",_function_name());
	}
    }

    %
    % Array arguments
    %
    variable i,j;
    if (typeof(jdin)==Array_Type || typeof(objects)==Array_Type) {
        variable qual=__qualifiers();
	% take care of case that objects were given via the qualifier
        if (typeof(qual)!=Null_Type) {
            if (struct_field_exists(qual,"object")) {
                qual=reduce_struct(qual,"object");
            }
        }

	% force objects to be an array (otherwise, if only one string
	% is given, the array syntax will give the ASCII value of the
	% first character in the string...)
	if (typeof(objects)!=Array_Type) {
	    objects=[objects];
	}
	
	if (qualifier_exists("vector")) {
	    variable res={};
	    _for i(0,length(jdin)-1,1) {
		_for j(0,length(objects)-1,1) {
		    list_append(res,planetpos(jdin[i],objects[j];;qual));
		}
	    }
	    return list_to_array(res);
	}

	variable res_r={};
	variable res_ra={};
	variable res_de={};
	variable r_r,r_ra,r_de;
	
	_for i(0,length(jdin)-1,1) {
	    _for j(0,length(objects)-1,1) {
		(r_r,r_ra,r_de)=planetpos(jdin[i],objects[j];;qual);
		list_append(res_r,r_r);
		list_append(res_ra,r_ra);
		list_append(res_de,r_de);
	    }
	}
	return (list_to_array(res_r),list_to_array(res_ra),list_to_array(res_de));
    }
    
    % from here, jdin and object are definitively scalars
    variable JD=jdin;
    if (qualifier_exists("mjd")) {
	JD=mjd2jd(JD);
    }

    variable planet=strup(objects);

    variable eph;
    if (qualifier_exists("ephemeris")) {
	eph=qualifier("ephemeris");
    } else {
	if (not qualifier_exists("vsop87")) {
	    % ephemeris not given: use the cached ephemeris or open an
	    % new one
	    % (NB: jpl_initeph with no argument defaults to DE430. It
	    % also does all of the error handling
	    % and also search in LHEA_DATA and LHEASOFT)
	    ifnot (__is_initialized(&jpl_ephsave)) {
		jpl_ephsave=jpl_initeph();
	    }
	    eph=jpl_ephsave;
	}
    }

    % we do all calculation in tdb rather than in tt
    % this means that when using vsop87 we convert back to tt, but
    % it makes the structure of the code here much simpler
    if (qualifier_exists("tt")) {
	JD=tt2tdb(JD;ephemeris=eph);
    } 

    variable centerobj=qualifier("center","Earth");
    variable cen;
    if (typeof(centerobj)==Array_Type) {
	if (length(centerobj)!=3 and length(centerobj)!=6) {
	    throw UsageError,sprintf("%s: center is vector and has invalid length\n",_function_name());
	}
	if (length(centerobj)==3) {
	    cen=[centerobj[0],centerobj[1],centerobj[2],0.,0.,0.];
	} else {
	    cen=[centerobj[0],centerobj[1],centerobj[2],centerobj[3],centerobj[4],centerobj[5]];
	}
	centerobj="POSITION";
    } else {
	centerobj=strup(centerobj);
    }

    if (centerobj==planet) {
	throw UsageError,sprintf("%s: Center object and target object cannot be identical.\n",_function_name());
    }

    % speed of light in au/day and GM/c^2 (in AU)
    variable c,muc2,auval;
    if (qualifier_exists("vsop87")) {
	c=173.1446;
	muc2=9.87e-9;
	auval=149597870.66; % value for the DE200 (i.e. DE118 per Standish, 1990)
    } else {
	c=86400.*eph.clight/eph.au;
	muc2=eph.gms/c/c;
	auval=eph.au;
    }

    
    %
    % decide on the ephemeris handlers
    %
    variable ephfunc;
    variable ephqual;
    if (qualifier_exists("vsop87")) {
	ephfunc=&vsop87;
	ephqual=struct {
	    tdb=1,
	    barycentric=1
	};
    } else {
	ephfunc=&jpl_eph_vec;
	ephqual=struct {
	    tdb=1,
	    equatorial=1,
	    ephemeris=eph
	};
    }


    %
    % topocentric position rather than planetocentric?
    %

    % relative position and velocity of observer wrt planet's
    % center
    variable g=vector(0.,0.,0.);
    variable gdot=vector(0.,0.,0.); 

    if (qualifier_exists("topocentric")) {
        if (centerobj!="EARTH") {
            throw UsageError,sprintf("%s: Computation of topocentric positions is currently only possible for Earth.\n",_function_name()); 
        }

        variable topopos;
        if (typeof(qualifier("topocentric"))==Vector_Type) {
            topopos=qualifier("topocentric"); % xyz in m
        } else {
            if (typeof(qualifier("topocentric"))==Struct_Type) {
                variable tp=qualifier("topocentric");
                topopos=geographic2vector(tp.lon,tp.lat,tp.height;deg,datum="WGS 84");
            } else {
                throw UsageError,sprintf("%s: Topocentric coordinates must be given as Vector or struct\n");
            }
        }

        %
        % compute observer's offset from Earth's center
        variable ut1=tt2ut1(tdb2tt(JD));
        variable beta=GMST(ut1;gast)/86400. * 2.*PI; % rad
        variable betadot=7.292115e-5; % rad/s (angular rotation of Earth)

        % geocentric position of the observer, ignoring
        % polar motion (per expl. suppl., p 296, this is
        % at the mas level for the moon, and smaller for
        % all other objects...)
        variable sb,cb;
        (sb,cb)=sincos(beta);
        g=matrix33_rot(3,-beta)*topopos;
        gdot=betadot * matrix33_new(-sb,-cb,0.,cb,-sb,0.,0.,0.,0.) * topopos;
        % m -> au  / m/s -> au/day
        g=g/(auval*1000.);
        gdot=gdot*86400./(auval*1000.);
    }
    
    variable p2; % will contain the geocentric position at the end
    variable ebary,sbary,qbary,ehelio;
    variable truedist=0.;
    
    if (planet!="SUN" and centerobj!="SUN") {
	% barycentric position of the center object and the Sun
	if (centerobj=="POSITION") {
	    ebary=struct {x=cen[0],y=cen[1],z=cen[2],vx=cen[3],vy=cen[4],vz=cen[5]};
	} else {
	    ebary=(@ephfunc)(JD,centerobj;;ephqual);
	}

	sbary=(@ephfunc)(JD,"Sun";;ephqual);

        % barycentric position and velocity of the Earth
	variable Eb=vector(ebary.x,ebary.y,ebary.z);
	variable Ev=vector(ebary.vx,ebary.vy,ebary.vz);
        
	% barycentric position of the Sun now
	variable Sb0=vector(sbary.x,sbary.y,sbary.z);

        % correction for topocentric positions
        % (g and gdot are 0 if we remain geocentric)
        Eb+=g;
        Ev+=gdot;
        
	% heliocentric position of the Earth now
	variable E=Eb-Sb0;
	variable ee=abs(E);
	
	%
	% adjust position for the light travel time
	% in an iterative way
	%
	variable tau=0.;
	do {

	    % Vector from the center to the planet
	    qbary=(@ephfunc)(JD-tau,planet;;ephqual);
	    variable Qb=vector(qbary.x,qbary.y,qbary.z);
	    variable P=Qb-Eb;
	    variable pp=abs(P);
	    if (truedist==0.) { % remember true geocentric distance
		truedist=pp;
	    }

	    % barycentric position of the Sun
	    sbary=(@ephfunc)(JD-tau,"Sun";;ephqual);
	    variable Sb=vector(sbary.x,sbary.y,sbary.z);

	    % heliocentric position of the planet
	    variable Q=Qb-Sb;
	    variable qq=abs(Q);
	    
	    % light travel time, including gravitational field of the sun
	    % (Shapiro delay [except for case of the Sun])
	    variable tauold=tau;
	    tau=(pp+2.*muc2*log((ee+pp+qq)/(ee-pp+qq)))/c;
	    if (qualifier_exists("no_apparent")) {
		tauold=tau;
	    }

	} while ( abs(tau-tauold) > 1e-9);
	    
	%
	% correction for relativistic deflection of light
	%

	% unit vectors
	variable pnorm=P/abs(P);
	variable qnorm=Q/abs(Q);
	variable enorm=E/ee;
	    
	variable g1=2*muc2/ee;
	variable g2=1.+qnorm*enorm;

	% geocentric direction of the planet, corrected for light
	% deflection due to the Sun
	variable p1=pnorm+g1/g2 * ((pnorm*qnorm)*enorm - (enorm*pnorm)*qnorm); 
	    
	%
	% aberration of light
	% (relativistic calculation, which is overkill for most applications)
	%
	variable vel=Ev*(1./c);
	variable vv=abs(vel);
    	variable beta1=sqrt((1.-vv)*(1.+vv)); % 1/beta

	variable f1=p1*vel;
	variable f2=1.+f1/(1.+beta1);
	
	p2=(beta1*p1+ f2*vel)/(1.+f1);
    } else {
	%
	% special treatment we are either calculating
	% heliocentric coordinates or the position of the
	% Sun
	%
	variable ccobj=centerobj;
	if (ccobj=="SUN") {
	    ccobj=planet;
	}

        tau=0.;
	% work around a s-lang problem with function pointers apparently not
	% allowing named qualifiers
	variable ephqual2=struct_combine(ephqual,struct{heliocentric=1});
	do {
	    ehelio=(@ephfunc)(JD-tau,ccobj;;ephqual2);
            variable Eh=vector(ehelio.x,ehelio.y,ehelio.z)+g;
	    tauold=tau;
	    if (truedist==0.) {
		truedist=abs(Eh);
	    }
	    tau=abs(Eh)/c;
	    if (qualifier_exists("no_apparent")) {
		tauold=tau;
	    }
	} while (abs(tau-tauold) > 1e-9);
	if (ccobj=="SUN") {
	    p2=Eh;
	} else {
	    p2=-Eh;
	}
    }

    if (qualifier_exists("vsop87")) {
	% conversion matrix from ecliptical to equatorial coordinates
	% (Bretagnon & Francou, 1988, eq 1)
	%
	% NB because of the phi rotation this is slighly different than the
	% rotation back to ecliptical below, 
	% so we will always do this 
	variable eps=dms2deg(23,26,21.4091;radian);
	variable phi=dms2deg(0.,0.,-0.0930;radian);

	variable ecl2icrf=matrix33_rot(3,phi)*matrix33_rot(1,-eps);
    
	p2=ecl2icrf*p2; % vector is now in the GCRF
    }

    % we now have the GCRS position
    
    % 
    % convert GCRS position into
    % position for the equinox of date?
    %
    variable p3;
    if (qualifier_exists("of_date")) {
	variable Bmat=gcrs2j2000_matrix(); % frame bias matrix
	variable Pmat=precession_matrix(JD);

	if (qualifier_exists("mean_equinox")) {
	    p3=Pmat*Bmat*p2;
	} else {
	    variable Nmat=nutation_matrix(JD);
	    p3=Nmat*Pmat*Bmat*p2;
	}
    } else {
	p3=p2;
    }

    % this is faster than using equatorial2ecliptical
    if (qualifier_exists("ecliptical")) {
	variable epsilon;
	if (qualifier_exists("of_date")) {
	    if (qualifier_exists("mean_equinox")) {
		epsilon=obliquity(JD);
	    } else {
		epsilon=obliquity(JD;true);
	    }
	} else {
	    epsilon=obliquity(2451545.0;true);
	}
	variable tmat=matrix33_rot(1,epsilon);
	p3=tmat*p3;
    }
    
    if (qualifier_exists("vector")) {
	% return vector quantity
	return p3; 
    }

    % default: RA, and DEC
    variable r,alp,del;

    if (qualifier_exists("deg")) {
	(r,alp,del)=vector_to_spherical(p3;deg);
    } else {
	(r,alp,del)=vector_to_spherical(p3);
    }

    if (qualifier_exists("lighttravel")) {
	truedist=tau*c;
    }

    if (qualifier_exists("mks")) {
	truedist*=auval*1000.;
    }

    if (qualifier_exists("cgs")) {
	truedist*=auval*1e5;
    }
    
    return (truedist,alp,del);
}

#endif

%
% Code from src/plot/keV2erg.sl
%
define keV2erg(value)
%!%+
%\function{keV2erg}
%\synopsis{convert flux to erg}
%\usage{Double_Type new_value = kev2erg(Double_Type old_value)}
%\qualifiers{
% \qualifier{y_fac}{: divide the value by 10^{y_fac} }
%}
%\seealso{plot_unfold}
%!%- 
{
   variable y_fac = qualifier("y_fac",0);

   variable fac0 = 1./624150647.996;

   return value*fac0/(10^y_fac);
}
%
% Code from src/slang/math/kendall_tau_censored.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%%%%%%%%%%%%%%%%%%%%
define kendall_tau_censored()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{kendall_tau_censored}
%\synopsis{calculation of Kendall rank correlation coefficient}
%\usage{Struct_Type cc = kendall_tau(Double_Type x[], Double_Type y[]);}
%\altusage{Struct_Type cc = kendall_tau(Double_Type x[], Double_Type y[], Char_Type y_is_upper_limit[]);}
%\altusage{Struct_Type cc = kendall_tau(Double_Type x[], Char_Type x_is_upper_limit[], Double_Type y[], Char_Type y_is_upper_limit[]);}
%\description
%    This function calculates the Kendall rank correlation coefficient \code{tau},
%    as defined by M. G. Kendall (1938, Biometrika, 30, 81), between two
%    arrays of equal length.
%    The result \code{tau} is within the interval [-1,1], where 1 means a perfect
%    agreement between both rankings, -1 means perfect disagreement, and values
%    around 0 indicate that both arrays are independent.
%    The value of \code{tau} is the difference of concordant and discordant pairs
%    divided by the total number of pairs.
%    In the definition of Helser (Dennis R. Helsel, "Nondetects and Data Analysis",
%    Wiley, 2005) only the number of determined (valid) pairs is used instead of
%    the total number. The structure returned by this function includes the values
%    of \code{tau}, \code{sigma}, and \code{pval} in both ways.
%
%    The value \code{sigma} characterizes the expected width of the distribution of \code{tau}
%    for random data. (If the "gsl" module is not available, the p-value is not calculated.
%    It can be obtained from \code{tau} and \code{sigma}.)
%
%    See statistic module for further functions:%\examples
%    variable x = [1:2:#100];
%    variable y = grand(100)+x;
%    variable cc_no_limits   = kendall_tau_censored(x,y);
%    variable cc_incl_limits = kendall_tau_censored(x,y,nint(urand(100)));
%\seealso{kendall_tau, pearson_r, spearman_r, spearmanrho, correlation_coefficient}
%!%-
{
  variable x, x_is_lim, y, y_is_lim;
  variable i,j;
  variable n;
  variable tau = 0;

  variable nondet_x = 0;
  variable nondet_y = 0;

  switch(_NARGS)
  { case 2: (x,y) = ();
    n = length(x);
    _for i (1,n-1,1)
      _for j (0,i-1,1)
      {
	tau += sign(x[i] - x[j]) * sign(y[i] - y[j]);
	if ((x[i] == x[j])) nondet_x +=1;
	if ((y[i] == y[j])) nondet_y +=1;
      }
  }
  { case 3: (x, y, y_is_lim) = ();
    n = length(x);
    _for i (1,n-1,1)
      _for j (0,i-1,1)
      {
	tau +=   sign(x[i] - x[j]) * ( (y_is_lim[i] ? 0 : 1)*(y[j] < y[i] ? 1 : 0) - (y_is_lim[j] ? 0 : 1)*(y[i] < y[j] ? 1 : 0) );
	if ((x[i] == x[j])) nondet_x +=1;
	if ((y[i] == y[j]) or (y_is_lim[i] and y[i]>y[j]) or (y_is_lim[j] and y[j]>y[i])) nondet_y +=1;
      }
  }
  { case 4: (x, x_is_lim, y, y_is_lim) = ();
    n = length(x);
    _for i (1,n-1,1)
      _for j (0,i-1,1)
      {
	tau += ( (x_is_lim[i] ? 0 : 1)*(x[j] < x[i] ? 1 : 0) - (x_is_lim[j] ? 0 : 1)*(x[i] < x[j] ? 1 : 0)) *
	  ( (y_is_lim[i] ? 0 : 1)*(y[j] < y[i] ? 1 : 0) - (y_is_lim[j] ? 0 : 1)*(y[i] < y[j] ? 1 : 0) );
	if ((x[i] == x[j]) or (x_is_lim[i] and x[i]>x[j]) or (x_is_lim[j] and x[j]>x[i])) nondet_x +=1;
	if ((y[i] == y[j]) or (y_is_lim[i] and y[i]>y[j]) or (y_is_lim[j] and y[j]>y[i])) nondet_y +=1;
      }
  }
  { help(_function_name()); return; }

  variable ktau = tau/ (0.5*n*(n-1));
  variable sigma = sqrt( (4*n+10) / (9.0*n*(n-1)) );

  variable nr_valid_pairs = sqrt( (0.5*n*(n-1) - nondet_x) * (0.5*n*(n-1) - nondet_y) );
  variable htau = tau/ nr_valid_pairs;
  variable n_red = 0.5*(1+sqrt(1+8*nr_valid_pairs));
  variable hsigma = sqrt( (4*n_red+10) / (9.0*n_red*(n_red-1)) );
  
  variable pval  = gsl-> erfc (sqrt(2)* abs(ktau/sigma) *0.5);
  variable hpval = gsl-> erfc (sqrt(2)* abs(htau/hsigma) *0.5);

  return struct {
    kendall_tau   = ktau,
    kendall_sigma = sigma,
    kendall_pval  = pval,
    helser_tau    = htau,
    helser_sigma  = hsigma,
    helser_pval   = hpval
  };
}
#endif

%
% Code from src/slang/physics/kerr_lp_redshift.sl
%
%!%+
%\function{kerr_lp_redshift}
%\synopsis{calculates the redshift from a Lamp Post source (point-like)
% on the rotational axis of a spining black hole (spin a) to the  observer}
%\usage{kerr_lp_redshift(height, a)}
%\description
%   z_obs = 1 / ( sqrt( 1 - 2*h/(h^2 + a^2) )  - 1 
%!%-
define kerr_lp_redshift(){
   variable height, a;
   
   switch(_NARGS)
   { case 2: (height, a) = ();}
   { help(_function_name()); return; }


   return 1.0 / sqrt( 1.0 - 2*height / (height^2 + a^2))-1.0;
}

%
% Code from src/slang/physics/kerr_rms.sl
%
%!%+
%\function{kerr_rms}
%\synopsis{calculates the radius of marginal stability of a black hole
%    in units of GM/c^2}
%\usage{kerr_rms(a)}
%!%-
define kerr_rms() {
   variable a;
   
   switch(_NARGS)
   { case 1: a = ();}
   { help(_function_name()); return; }
   
%   if (abs(a) > 1) {
%      message("Only values of -1 < a < 1 are reasonable!");  
%      return;
%   }
   
   variable z1 = 1+ (1 - a^2)^(1./3)*((1+a)^(1./3) + (1-a)^(1./3)  );
   variable z2 = sqrt(3.*a^2 + z1^2);
   
   return (3 + z2 - sign(a)*sqrt( (3-z1)*(3+z1+2*z2)));
}

%
% Code from src/slang/physics/kerr_rplus.sl
%
%!%+
%\function{kerr_rplus}
%\synopsis{calculates the event horizon of a black hole in units of GM/c^2}
%\usage{kerr_rplus(a)}
%!%-
define kerr_rplus() {
   variable a;
   
   switch(_NARGS)
   { case 1: a = ();}
   { help(_function_name()); return; }
   
   return 1. + sqrt((1.-a)*(1.+a));
}

%
% Code from src/slang/misc/keyinput.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define keyinput ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{keyinput}
%\synopsis{reads input from the keyboard}
%\usage{String_Type = keyinput([String_Type message]);}
%\qualifiers{
%    \qualifier{silent}{characters are not echoed, e.g. useful for password input}
%    \qualifier{time}{read timout after given seconds. Returns an empty string}
%    \qualifier{nchr}{returns automatically after given number
%              of chars have been read}
%    \qualifier{prompt}{output the given string before reading}
%    \qualifier{default}{initial text to be read}
%}
%\description
%    Prompts the user to input a line, which is ended
%    with return. If a limited number of chars is given
%    using the `nchr' qualifier, the line is automatically
%    ended after reaching this number without pressing return. 
%    It is also possbile to limit the time for an input with the 
%    `time' qualifier and to disable echoing the input using 
%    the `silent' qualifier. The latter is useful, e.g., for password 
%    inputs or for single key events.
%    If `nchr==1', the returned keycode is
%    checked against special keys such as the arrow keys.
%    If one of these is detected, the returned string contains
%    the name if the pressed key.
%!%-
{
 if (_NARGS == 1) { variable ms = __pop_args(1); message(typecast(ms[0].value, String_Type)); }

 variable pp, inpt, cmd;
 % create command to be executed
 cmd="/bin/bash -c 'read -r"; % -r option disables \ as escape parameter
 if (qualifier_exists("silent")) cmd=cmd+" -s";
 if (qualifier_exists("time")) cmd=sprintf("%s -t %.1f",cmd,qualifier("time"));
 if (qualifier_exists("nchr")) cmd=sprintf("%s -n %d",cmd,qualifier("nchr"));
 if (qualifier_exists("prompt")) cmd=sprintf("%s -p \"%s\"",cmd,qualifier("prompt"));
 if (qualifier_exists("default")) cmd=sprintf("%s -e -i \"%s\"",cmd,qualifier("default"));
 cmd=cmd+" line; echo $line'";
 % execute command and store output in variable
 pp = popen(cmd,"r");
 inpt = fgetslines(pp);
 () = pclose(pp);
 % if nchr = 1 test on special keys
 if (qualifier("nchr", 0) == 1)
 {
   variable a;
   if (inpt[0] == "\x1B\n") % if a special key was hit, it consists of at least 3 bytes -> read missing 2 bytes
   {
     pp = popen("/bin/bash -c 'read -r -s -n 1 line; echo $line'", "r");
     a = fgetslines(pp);
     () = pclose(pp);
     if (a[0] == "[\n") % ARROWS, HOME, etc.
     {
       pp = popen("/bin/bash -c 'read -r -s -n 1 line; echo $line'", "r");
       a = fgetslines(pp);
       () = pclose(pp);
       if (a[0] == "1\n") % F5-F8 (5 bytes)
       {
         pp = popen("/bin/bash -c 'read -r -s -n 2 line; echo $line'", "r");
         a = fgetslines(pp);
         () = pclose(pp);
       }
     }
     else if (a[0] == "O\n") % F1-F4
     {
       pp = popen("/bin/bash -c 'read -r -s -n 1 line; echo $line'", "r");
       a = fgetslines(pp);
       () = pclose(pp);
     }
     if (a[0] == "2\n") % INS, F8-F12
     {
       pp = popen("/bin/bash -c 'read -r -s -n 1 line; echo $line'", "r");
       a = substr(a[0],1,strlen(a[0])-1) + fgetslines(pp);
       () = pclose(pp);
     }
     a = substr(a[0],1,strlen(a[0])-1);
     % transform to special key
     switch (a)
       { case "A": a = "UP_ARROW"; }
       { case "B": a = "DOWN_ARROW"; }
       { case "D": a = "LEFT_ARROW"; }
       { case "C": a = "RIGHT_ARROW"; }
       { case "5": a = "PAGE_UP"; pp = popen("/bin/bash -c 'read -r -s -n 1 line;'", "r"); ()=pclose(pp); } % (4 bytes)
       { case "6": a = "PAGE_DOWN"; pp = popen("/bin/bash -c 'read -r -s -n 1 line;'", "r"); ()=pclose(pp); }
       { case "H": a = "HOME"; }
       { case "F": a = "END"; }
       { case "3": a = "DEL"; pp = popen("/bin/bash -c 'read -r -s -n 1 line;'", "r"); ()=pclose(pp); }
       { case "2~": a = "INS"; }
       { case "P": a = "F1"; }
       { case "Q": a = "F2"; }
       { case "R": a = "F3"; }
       { case "S": a = "F4"; }
       { case "5~": a = "F5"; }
       { case "7~": a = "F6"; }
       { case "8~": a = "F7"; }
       { case "9~": a = "F8"; }
       { case "20": a = "F9"; pp = popen("/bin/bash -c 'read -r -s -n 1 line;'", "r"); ()=pclose(pp); }
       { case "21": a = "F10"; pp = popen("/bin/bash -c 'read -r -s -n 1 line;'", "r"); ()=pclose(pp); }
       { case "23": a = "F11"; pp = popen("/bin/bash -c 'read -r -s -n 1 line;'", "r"); ()=pclose(pp); }
       { case "24": a = "F12"; pp = popen("/bin/bash -c 'read -r -s -n 1 line;'", "r"); ()=pclose(pp); }
       { a = sprintf("UNKNOWN_%s", a); }
     inpt = [a + " "];
   }
 }
 % return this variable without line break
 return substr(inpt[0],1,strlen(inpt[0])-1);
}

%
% Code from src/data/timing/lag_energy_spectrum.sl
%
% -*- mode: slang; mode: fold -*-

private define filterFouQuantForFrequencyRange(fouquant, fmin, fmax) {
  %{{{
  variable idx = where(fmin <= fouquant.freq <= fmax);
  variable idxmin = min(idx);
  variable idxmax = max(idx);
  variable df = fouquant.freq[[idxmin+1:idxmax]] - fouquant.freq[[idxmin:idxmax-1]];
  
  if (length(idx) == 0) vmessage("ERROR (lag_energy_spectrum): No frequency bins between %g-%g in Fourier products!", fmin, fmax);
  if (qualifier("verbose", 0) > 1){
    vmessage("  ***(lag_energy_spectrum): Minimal frequency bin %d at %g Hz", idxmin, fouquant.freq[idxmin]);
    vmessage("  ***(lag_energy_spectrum): Maximal frequency bin %d at %g Hz", idxmax, fouquant.freq[idxmax]);
    vmessage("  ***(lag_energy_spectrum): Number of frequency bins: %d", length(idx));
  }
  variable ret = struct_filter(fouquant, idx; copy);
  return struct_combine(ret, struct{df = df, idxmin = idxmin, idxmax = idxmax, n_freqs = length(idx)});
} %}}}

private define verifyLightcurveTime(lc1, lc2){ 
  %{{{
  %% Test whether the time interval of reference and energy-resolved lightcurve is the same
  if (length(lc1.time)!=length(lc2.time) or (lc1.time[0]!=lc2.time[0]) or (lc1.time[-1]!=lc2.time[-1])){
    throw DataError, "ERROR: The time arrays of the reference and small band lightcurve seem to be different!";
  }
} %}}}

private define getLightcurve(lc, dt, dimseg){ 
  %{{{

  variable data = typeof(lc) == String_Type ? fits_read_table(lc) : lc;

  if (struct_field_exists(data, "time") * struct_field_exists(data, "rate") != 1) {
    throw UsageError, sprintf("ERROR (%s): %s does not have a time and rate column! Exiting...", _function_name(), lc);
  }
  
  return segment_lc_for_psd(data, dt, dimseg;; __qualifiers);
} %}}}

private define calculateSummedReferenceLightcurve(lc_list, index, dt, dimseg){ 
  %{{{
  % Calculate the reference lightcurve as sum of all
  % energy-resolved lightcurves but the current channel-of-interest.
  % This way there is no correlation of the Poisson noise between the
  % subject and reference bands (see also Uttley+14 Sect 3.2).

  %% todo: clean up this horrible if-else mess!
  
  variable channel_of_interest = typeof(lc_list[index]) == String_Type ? fits_read_table(lc_list[index]) : lc_list[index];

  variable jj, lc;
  variable summed_rate = Double_Type[length(channel_of_interest.rate)];
  
  if (qualifier_exists("elo") * qualifier_exists("ehi") * qualifier_exists("elo_ref") * qualifier_exists("ehi_ref") == 1) {
    %% If we have energy information about the ref LC we want to
    %% only subtract the COI in case it's within the ref LC borders.
    variable elo_ref = qualifier("elo_ref");
    variable ehi_ref = qualifier("ehi_ref");
    variable elo = qualifier("elo");
    variable ehi = qualifier("ehi");
    
    variable idx_lo = wherefirst(elo >= elo_ref);
    variable idx_hi = wherelast(ehi <= ehi_ref);

    _for jj (idx_lo, idx_hi, 1) {
      if ((elo[index] >= elo[jj]) and (ehi[index] <= ehi[jj])) {
	if (qualifier("verbose", 0) > 1) vmessage("  ***(%s): COI LC (index %d, %g-%g keV) is within Ref. LC (%g-%g keV). Not adding COI lightcurve to Ref. LC.",
						  _function_name(), index, elo[index], ehi[index], elo_ref, ehi_ref);
	continue;
      } else {
	lc = typeof(lc_list[jj]) == String_Type ? fits_read_table(lc_list[jj]) : lc_list[jj];
	verifyLightcurveTime(channel_of_interest, lc);
	if (qualifier("verbose", 0) > 1) vmessage("  ***(%s): Summing LC (index %d, %g-%g keV) to Ref. LC (%g-%g keV).",
						  _function_name(), jj, elo[jj], ehi[jj], elo_ref, ehi_ref);
	summed_rate += lc.rate;
      }
    }	     
  } else {
    %% If no energy information about the reference lightcurve is
    %% provided, we assume that it spans the full energy range.
    _for jj (0, length(lc_list)-1, 1){
      if (jj == index){
	if (qualifier("verbose", 0) > 1) vmessage("  ***(%s): Not adding COI LC (index %d) to Ref. LC, assuming that the Ref. LC goes over all energies.",
						  _function_name(), index);
	continue;
      } else {
	lc = typeof(lc_list[jj]) == String_Type ? fits_read_table(lc_list[jj]) : lc_list[jj];
	verifyLightcurveTime(channel_of_interest, lc);
	summed_rate += lc.rate;
      }
    }
  }
  
  return segment_lc_for_psd(struct{time = channel_of_interest.time, rate = summed_rate},
			    dt, dimseg;; __qualifiers);
} %}}}

private define write_cpd_to_ascii(realcpd, realcpd_err, imagcpd, imagcpd_err, fmin, fmax){
  %{{{
  % ASCII files of the cross-spectrum are needed for fitting with RELTRANS
  variable fptr_imag = fopen(sprintf("freq_cross_spec_imag_%f_%fHz.ascii", fmin, fmax), "w");
  variable fptr_real = fopen(sprintf("freq_cross_spec_real_%f_%fHz.ascii", fmin, fmax), "w");
  variable ii;
  _for ii (0, length(realcpd)-1, 1){
    () = fputs(sprintf("%i\t%f\t%f\n", ii+1, imagcpd[ii], imagcpd_err[ii]), fptr_imag);
    () = fputs(sprintf("%i\t%f\t%f\n", ii+1, realcpd[ii], realcpd_err[ii]), fptr_real);
  }
  () = fclose(fptr_imag);
  () = fclose(fptr_real);
} %}}}

private define verify_frequency_binning(dat){
  %{{{
  %% Check that the frequency binning of every energy-resolved
  %% lightcurve was the same for the individual frequency bins. If
  %% this is not the case this means that the energy-resolved
  %% lightcurves were extracted with different binning/time
  %% resolution.
  variable ii;
  variable N_energies = array_shape(dat.lag)[0];
  _for ii (1, N_energies-1, 1){
    if (_eqs(dat.n_segments[ii], dat.n_segments[0]) == 0 or
	_eqs(dat.n_frequencies[ii,*], dat.n_frequencies[0,*]) == 0 or
	_eqs(dat.f_min_dft[ii,*], dat.f_min_dft[0,*]) == 0 or
	_eqs(dat.f_max_dft[ii,*], dat.f_max_dft[0,*]) == 0){
      vmessage("WARNING (%s): The DFT binning of the energy-resolved lightcurves differs! Check the time resolutios of your individual lightcurves!",
	       _function_name());
    }
  }
} %}}}

private define write_to_outfile(dat, f_lo, f_hi, keywords, lc_list){
  %{{{
  variable fp = fits_open_file(qualifier("outfile"), "c");

  variable N_energies, N_les;
  (N_energies, N_les) = array_shape(dat.lag)[0], array_shape(dat.lag)[1];
  
  variable jj;
  _for jj (0, N_les-1, 1){
    % write_cpd_to_ascii(dat.mean_realcpds[*,jj], dat.mean_realcpd_errs[*,jj], 
    % 		     dat.mean_imagcpds[*,jj], dat.mean_imagcpd_errs[*,jj], 
    % 		     f_lo[jj], f_hi[jj]);  
    
    variable extname = sprintf("BAND_%f_%fHz", f_lo[jj], f_hi[jj]);
    variable sdata = struct{
      "ENERGY_LO"         = qualifier("elo", Double_Type[N_energies]),
      "ENERGY_HI"         = qualifier("ehi", Double_Type[N_energies]),
      "LAG"               = dat.lag[*,jj], 
      "LAG_ERR"           = dat.errlag[*,jj], 
      "MEAN_IMAG_CPD"     = dat.mean_imagcpds[*,jj],
      "MEAN_IMAG_CPD_ERR" = dat.mean_imagcpd_errs[*,jj],
      "MEAN_REAL_CPD"     = dat.mean_realcpds[*,jj],
      "MEAN_REAL_CPD_ERR" = dat.mean_realcpd_errs[*,jj],
      "MEAN_PHASE"        = dat.mean_phases[*,jj],
      "RMS"               = dat.rms[*,jj],
      "RMS_ERR"           = dat.rms_err[*,jj],
    };
    
    variable keys = struct_combine(struct{
      F_MIN = f_lo[jj], F_MAX=f_hi[jj], N_SEGS = dat.n_segments[0],
      N_FREQS = dat.n_frequencies[0,jj], FMIN_DFT = dat.f_min_dft[0,jj], 
      FMAX_DFT = dat.f_max_dft[0,jj]}, keywords);
    variable hist = struct{
      history = [sprintf("Wrote lag-energy spectrum in frequency range %f-%fHz", f_lo[jj], f_hi[jj])],
      comment = ["Time unit is seconds, energy unit is keV, frequency unit is Hertz", 
		 "Analyzed lightcurves:", lc_list]
    };
    
    fits_write_binary_table(fp, extname, sdata, keys, hist);
  }
  
  %% Write power spectrum and coherence into "COHERENCE" & "PSD" extension
  variable psd_colnames = String_Type[N_energies+1];
  variable cof_colnames = String_Type[2*N_energies+1];
  variable lag_colnames = String_Type[2*N_energies+1];
  (psd_colnames[0], cof_colnames[0], lag_colnames[0]) = "freq", "freq", "freq";

  variable ii;
  _for ii (0, N_energies-1, 1){
    psd_colnames[ii+1] = sprintf("psd%i", ii);
    cof_colnames[ii+1] = sprintf("cof%i", ii);
    cof_colnames[N_energies+ii+1] = sprintf("errcof%i", ii);
    lag_colnames[ii+1] = sprintf("lag%i", ii);
    lag_colnames[N_energies+ii+1] = sprintf("errlag%i", ii);
  }
    
  variable psd_data = @Struct_Type(psd_colnames);
  variable cof_data = @Struct_Type(cof_colnames);
  variable lag_data = @Struct_Type(lag_colnames);
  
  %% Frequency grids are all the same (because LCs have same binning)
  set_struct_field(psd_data, "freq", dat.freq[0]);
  set_struct_field(cof_data, "freq", dat.freq[0]);
  set_struct_field(lag_data, "freq", dat.freq[0]);
  
  _for ii (0, N_energies-1, 1){
    set_struct_field(psd_data, psd_colnames[ii+1], dat.psd[ii]);
    set_struct_field(cof_data, cof_colnames[ii+1], dat.cof[ii]);
    set_struct_field(cof_data, cof_colnames[N_energies+ii+1], dat.errcof[ii]);
    set_struct_field(lag_data, lag_colnames[ii+1], dat.freqres_lags[ii]);
    set_struct_field(lag_data, lag_colnames[N_energies+ii+1], dat.freqres_errlags[ii]);
  }
  
  fits_write_binary_table(fp, "PSD", psd_data, struct{"N_SEGS"=dat.n_segments[0]});
  fits_write_binary_table(fp, "COHERENCE", cof_data, struct{"N_SEGS"=dat.n_segments[0]});
  fits_write_binary_table(fp, "LAGS", lag_data, struct{"N_SEGS"=dat.n_segments[0]});
  
  fits_close_file(fp);
  () = system(sprintf("fthedit %s\[0] keyword=N_LES operation=add value=%i comment=\"Number of frequency-resolved lag-energy spectra\"",
		      qualifier("outfile"), N_les));
  vmessage("Wrote lag-energy spectrum data to %s", qualifier("outfile"));
} %}}}


define lag_energy_spectrum(lc_list, dimseg) {
%!%+
%\function{lag_energy_spectrum}
%\synopsis{Calculates the time lag of energy-resolved lightcurves}
%\usage{Struct_Type les = lag_energy_spectrum(String_Type[lcs], dimseg)}
%\description
%    * lc_list: Array of file names of the energy-resolved
%               lightcurves (must have same length and time binning)
%    * dimseg:  Segmentation length in bins, needed to segment the
%               lightcurve and do the Fourier calculation
%    
%    returns: lag-energy-spectrum as a struct with fields lag, errlag
%\qualifiers{
%    \qualifier{dt}{Time resolution of the lightcurve in seconds
%                [default: TIMEDEL keyword]}
%    \qualifier{deadtime}{Detector deadtime in seconds [default: 0.0s]}
%    \qualifier{f_lo}{Array_Type. For frequency-resolved lag-energy
%                spectra. Default: lowest sampled frequency [an array
%                containing only fmin=1/(dt*dimseg)]. Unit: Hz}
%    \qualifier{f_hi}{See f_lo. Default: Nyquist frequency [1/(2*dt)]. Unit: Hz}
%    \qualifier{verbose}{Increase verbosity [default=0]}
%    \qualifier{outfile}{String_Type. If set, write various timing products
%                into a FITS file.}
%    \qualifier{elo, ehi}{Double_Type: Energy grid}
%    \qualifier{elo_ref, ehi_ref}{Boundaries of reference lightcurve}
%}
%\notes
%    * The default error computation is Nowak et al., 1999, ApJ, 510,
%      874 (Sect. 4).
%    * All input lightcurves are accounted for their gaps and segmented
%      by segment_lc_for_psd. 
%    * The cross-spectrum (CPD) and lags are calculated with foucalc.
%    * The PSD normalization is fixed to Miyamoto.
%    * The average time lag in the frequency interval is calculated by
%      the normal mean on the imaginary and real part of the
%      cross-spectrum: 
%      atan2(mean(Imag(CPD)), mean(Real(CPD)))/(PI*(fmin+fmax))
%    * Reference lightcurve: For each energy bin, the function takes the
%      summed lightcurves of all but the current energy bands (always
%      excluding the lightcurve for the cross-spectrum).
%    * The function does *not* use any information about the energy
%      grid. This has to be created by yourself (and must be the same as
%      the lightcurves in lc_list)! The elo, ehi qualifiers are only
%      for the records in the output file
%    
%    If you want to add energy information to the output FITS file,
%    you can pass two Double_Type arrays (same length as lc_list) via
%    the elo, ehi qualifiers. The columns of the created extension,
%    called BAND_<f_lo>_<f_hi>Hz, are:
%  
%    ENERGY_LO           - Energy channel (keV), taken from elo qualifier
%    ENERGY_HI           - Energy channel (keV), taken from ehi qualifier
%    MEAN_PHASE          - Phase (radians), computed as atan2(mean(Im(CPD)), mean(Re(CPD)))
%    LAG                 - Time lag (s), computed as phase/(PI*(fmin+fmax))
%    LAG_ERR             - Error computed as in Nowak et al., 1999, ApJ, 510, 874 Eq. 16
%    LAG_ERR_PROPERR     - Error computed from Gaussian error propagation
%    MEAN_REAL/IMAG_CPD  - real/imaginary part of the mean of the
%                          segment-averaged cross-spectrum
%    MEAN_REAL/IMAG_CPD_ERR - standard error on the mean of the
%                          real/imaginary segment-averaged cross-spectrum 
%    RMS/RMS_ERR         - Root Mean Squared variability of the channel-of-interest 
%                          lightcurve. The error is calculated using
%                          Vaughan et al., MNRAS 345, 1271, 2003 Eq. 11.
%                          
%    Following keywords are written into each extension:
%    
%    N_SEGS              - Number of segments averaged over (foucalc.numavgall)
%    N_FREQS             - Number of frequencies averaged over (as in
%                          "mean of the CPD")
%    F_MIN/F_MAX         - Minimum and maximum frequency as given by
%                          f_lo, f_hi qualifiers
%    FMIN_DFT/FMAX_DFT   - Minimal/Maximal frequency of the frequency-filtered
%                          Fourier products - this can be different from
%                          F_MIN/F_MAX due to the discretization of the
%                          *Discrete* Fourier Transform
%                         
%    If you find any bugs or have questions, please contact ole.koenig@fau.de
%
%    The function also accepts all qualifiers of segment_lc_for_psd.
%
%\example
%    %% Example for frequency-resolved lag-energy spectrum
%    (f_lo, f_hi) = log_grid(1, 50, n_freqs);  % (Hz)
%    % Energy grid must be the same as extracted lightcurves
%    (elo, ehi) = log_grid(0.5, 10, n_energies);  % (keV)
%    variable les = lag_energy_spectrum(lc_names, dimseg
%                                       ; f_lo = f_lo, f_hi=f_hi,
%                                       verbose=2, outfile="test.fits", elo=elo, ehi=ehi);
%    _for ii (0, n_freqs-1, 1) {
%      ohplot_with_err(elo, ehi, les.lag[*,ii], les.errlag[*,ii]);   
%    }
%\seealso{segment_lc_for_psd, foucalc, colacal}
%!%-
  try {
    variable dt = qualifier_exists("dt") ? qualifier("dt") : fits_read_key(lc_list[0], "TIMEDEL");
  } catch TypeMismatchError:
  {
    return vmessage("ERROR (%s): Couldn't receive the time resolution from TIMEDEL, need dt qualifier!", _function_name());
  }
  
  variable f_lo = qualifier("f_lo", [1/(dt*dimseg)]);
  variable f_hi = qualifier("f_hi", [1/(2*dt)]);  % Nyquist frequency
  variable verbose = qualifier("verbose", 0);
  variable deadtime = qualifier("deadtime", 0.0);
  variable normtype = "Miyamoto";

  variable N_energies = length(lc_list);
  variable N_les = length(f_lo);

  variable dat = struct{
    %% Fields for frequency-averaged data
    lag               = Double_Type[N_energies, N_les],
    errlag            = Double_Type[N_energies, N_les],
    rms               = Double_Type[N_energies, N_les],
    rms_err           = Double_Type[N_energies, N_les],
    min_freqs         = Double_Type[N_energies, N_les],
    max_freqs         = Double_Type[N_energies, N_les],
    mean_realcpds     = Double_Type[N_energies, N_les],
    mean_realcpd_errs = Double_Type[N_energies, N_les],
    mean_imagcpds     = Double_Type[N_energies, N_les],
    mean_imagcpd_errs = Double_Type[N_energies, N_les],
    mean_phases       = Double_Type[N_energies, N_les],
    errlag_properr    = Double_Type[N_energies, N_les],
    f_min_dft         = Double_Type[N_energies, N_les],
    f_max_dft         = Double_Type[N_energies, N_les],
    n_frequencies     = Integer_Type[N_energies, N_les],

    %% Fields for frequency-resolved data
    n_segments        = Integer_Type[N_energies],
    freq              = Array_Type[N_energies],
    psd               = Array_Type[N_energies],
    cof               = Array_Type[N_energies],
    errcof            = Array_Type[N_energies],
    freqres_lags      = Array_Type[N_energies],
    freqres_errlags   = Array_Type[N_energies]
  };

  %% Loop through energy-resolved lightcurves (channel-of-interests)
  variable ii, jj;
  _for ii (0, N_energies-1, 1){
    
    if (verbose > 0) vmessage("  ***(%s): Calculate time lags for %s", _function_name(), lc_list[ii]);
    
    variable channel_of_interest = getLightcurve(lc_list[ii], dt, dimseg;; __qualifiers);

    variable lc_ref = calculateSummedReferenceLightcurve(lc_list, ii, dt, dimseg;; __qualifiers);     

    variable fouquant = foucalc(struct{time  = channel_of_interest.time,
                                       rate1 = lc_ref.rate,
                                       rate2 = channel_of_interest.rate},
				dimseg; normtype=normtype, deadtime=deadtime);

    variable numseg = fouquant.numavgall[0];
    
    dat.freq[ii] = fouquant.freq;
    dat.psd[ii] = fouquant.signormpsd2;
    dat.cof[ii] = fouquant.cof12;
    dat.errcof[ii] = fouquant.errcof12;
    dat.freqres_lags[ii] = fouquant.lag12;
    dat.freqres_errlags[ii] = fouquant.errlag12;
    dat.n_segments[ii] = numseg;

    %% Loop through frequency ranges
    _for jj (0, N_les-1, 1){
      if (verbose > 1) vmessage("  ***(%s): Calculating lag-energy spectrum in frequency range %g-%g Hz", 
				_function_name(), f_lo[jj], f_hi[jj]);

      variable fouquant_f = filterFouQuantForFrequencyRange(fouquant, f_lo[jj], f_hi[jj]; verbose=verbose);
      
      % RMS calculation (same as in foucalc but we want it for every
      % energy lightcurve and frequency range - running foucalc
      % everytime drastically increases the runtime due to the DFT, so
      % copy-paste the calculation here!)
      variable sigrms = sqrt( sum(fouquant.signormpsd2[[fouquant_f.idxmin:fouquant_f.idxmax-1]] * fouquant_f.df) );
      variable noirms = sqrt( sum(fouquant.noinormpsd2[[fouquant_f.idxmin:fouquant_f.idxmax-1]] * fouquant_f.df) );
      variable errrms = sqrt((4.*sigrms^2*noirms^2+2.*noirms^4)/(numseg*fouquant_f.n_freqs*sigrms^2));  % Vaughan+03 Eq. 11

      dat.rms[ii,jj] = sigrms;
      dat.rms_err[ii,jj] = errrms;
      
      %% Coherence and lag calculation on averaged cross-spectrum
      variable cola = colacal(0.5*(f_lo[jj]+f_hi[jj]),  % mid frequency
			      mean(fouquant_f.realcpd12)+mean(fouquant_f.imagcpd12)*1i,
			      mean(fouquant_f.noicpd12),
			      mean(fouquant_f.rawpsd1), mean(fouquant_f.rawpsd2), 
			      mean(fouquant_f.noipsd1), mean(fouquant_f.noipsd2),
			      mean(fouquant_f.sigpsd1), mean(fouquant_f.sigpsd2),
			      length(fouquant_f.freq)*numseg);  % K*M with M=number of frequencies averaged over and K=number of segments
      (dat.lag[ii,jj], dat.errlag[ii,jj]) = cola.lag, cola.errlag;
	  
      variable m_ReCPD = mean(fouquant_f.realcpd12);
      variable m_ReCPD_err = mean(fouquant_f.errrealcpd12);
      variable m_ImCPD = mean(fouquant_f.imagcpd12);
      variable m_ImCPD_err = mean(fouquant_f.errimagcpd12);
      
      %% Calculation from Gaussian error propagation
      variable m_phase = atan2(m_ImCPD, m_ReCPD);  % [radians]
      dat.errlag_properr[ii,jj]=1/(PI*(f_lo[jj]+f_hi[jj]))*
	sqrt((m_ReCPD/(m_ImCPD^2+m_ReCPD^2))^2*m_ImCPD_err^2 +
	     (m_ImCPD/(m_ImCPD^2+m_ReCPD^2))^2*m_ReCPD_err^2);
      
      dat.mean_realcpds[ii,jj]     = m_ReCPD;
      dat.mean_realcpd_errs[ii,jj] = m_ReCPD_err;
      dat.mean_imagcpds[ii,jj]     = m_ImCPD;
      dat.mean_imagcpd_errs[ii,jj] = m_ImCPD_err;
      dat.mean_phases[ii,jj]       = m_phase;
      dat.f_min_dft[ii,jj]         = min(fouquant_f.freq);
      dat.f_max_dft[ii,jj]         = max(fouquant_f.freq);
      dat.n_frequencies[ii,jj]     = length(fouquant_f.freq);
    }
  }

  verify_frequency_binning(dat);
  
  if (qualifier_exists("outfile")){
    variable keywords = struct{DIMSEG=dimseg, TIMEDEL=dt, NORMTYPE=normtype};
    write_to_outfile(dat, f_lo, f_hi, keywords, lc_list;; __qualifiers());
  }
  
  return dat;
}

%
% Code from src/slang/math/lagrange.sl
%
define chebyshev_nodes (n)
%!%+
%\function{chebyshev_nodes}
%\synopsis{Get nodes of Chebyshev polynomial}
%\usage{Double_Type[] nodes = chebyshev_nodes(Int_Type n);}
%\qualifiers{
%  \qualifier{second}{if given, return nodes of second order polynomal.}
%  \qualifier{min}{[=-1] rescale to min}
%  \qualifier{max}{[=1] rescale to max}
%}
%\description
%  Given an integer \code{n} this function returns the nodes
%  of the Chebyshev polynomial of the first kind of order \code{n}.
%
%  Or, if the \code{second} qualifier is given, the corresponding nodes
%  of the second kind polynomial.
%
%  The nodes are distributed between -1 and 1 per default. If the \code{min}
%  and \code{max} qualifiers are given the nodes are rescaled accordingly.
%\seealso{chebyshev_lagrange_weights}
%!%-
{
  variable x, r;

  x = [0:n-1];
  if (qualifier_exists("second"))
    r = -cos(x*1./(n-1)*PI);
  else
    r = -cos((2.0*x+1.0)/n*PI/2);

  variable l = qualifier("min", -1.0);
  variable h = qualifier("max", 1.0);

  return 0.5*(l+h)+0.5*(h-l)*r;
}

define chebyshev_lagrange_weights (n)
%!%+
%\function{chebyshev_lagrange_weights}
%\synopsis{Get Lagrange weights for the chebyshev_nodes}
%\usage{Double_Type[] chebyshev_lagrange_weights(Int_Type n);}
%\qualifiers{
%  \qualifier{second}{if given, return weights of second order polynomal.}
%}
%\description
%  Given an integer \code{n} this function returns the weights for the
%  barycentric Lagrange polynomial for nodes placed the the Chebyshev nodes.
%
%  In case the \code{second} qualifier is given, the second kind Chebyshev
%  polynomial defines the weights, else the first kind.
%\seealso{chebyshev_nodes,lagrange_weights,lagrange_poly}
%!%-
{
  variable x, w;
  x = [0:n-1];
  if (qualifier_exists("second"))
    w = -(-1)^x*[0.5, 1.+Double_Type[n-2], 0.5];
  else
    w = -(-1)^x*sin((2.0*x+1)/n*PI/2);

  return w;
}

define lagrange_weights (x0)
%!%+
%\function{lagrange_weights}
%\synopsis{Get Lagrange polynomial weights}
%\usage{Double_Type[] lagrange_weights(Double_Type[] x0);}
%\description
%  Given an array of evaluation points \code{x0} this function
%  calculates the weights for the barycentric Lagrange polynomial.
%\seealso{lagrange_poly}
%!%-
{
  variable w = Double_Type[length(x0)]+1;
  variable i,j;
  _for i (0, length(x0)-1)
  {
    _for j (0,i-1)
      w[j] *= 1.0/(x0[i]-x0[j]);
    _for j (i+1,length(x0)-1)
      w[j] *= 1.0/(x0[i]-x0[j]);
  }

  return w;
}

define lagrange_poly (x, x0, y0)
%!%+
%\function{lagrange_poly}
%\synopsis{Interpolate points with Lagrange polynomial}
%\usage{Double_Type[] lagrange_poly(Double_Type[] x, x0, y0);}
%\qualifiers{
%  \qualifier{w}{calculated weights can be passed directly}
%}
%\description
%  Calculate the Lagrange interpolation at points \code{x} given
%  points \code{x0},\code{y0}.
%
%  If multiple invocations of the interpolation for the same pair
%  \code{x0},\code{y0} are required it is better to calculate the
%  weights beforehand (see \code{lagrange_weights}) and pass them
%  via the \code{w} qualifier.
%
%  The Lagrange interpolation is quite powerfull, however, as any
%  polynomial interpolation scheme it suffers from Runges phenomenon.
%  It can be shown that the interpolation of a function on a uniform
%  grid is close to the worst case, while the evaluation on a grid
%  given by the Chebyshev nodes is close to optimal. Therefore, when
%  interpolating a function not restircted to a specific grid you'd
%  be best advised to use \code{chebyshev_nodes} and
%  \code{chebyshev_lagrange_weights}.
%
%\example
%  variable x = chebyshev_nodes(20; min=-2, max=2);
%  variable w = chebyshev_lagrange_weights(20);
%  variable f = sin(x);
%  % we can extrapolate, but this will diverge quickly
%  variable p = lagrange_poly([-5:5:#100], x, f; w=w);
%
%\seealso{lagrange_weights,chebyshev_nodes,chebyshev_lagrange_weights}
%!%-
{
  variable r = Double_Type[length(x)];
  variable s, t;
  variable w;

  if (length(x0) != length(y0))
    throw UsageError, "Lengths of x0 and y0 do not match";

  w = qualifier("w", lagrange_weights(x0));

  variable i;
  _for i (0, length(x)-1)
  {
    s = w/(x[i]-x0);
    t = where(isinf(s));

    if (length(t))
      r[i] = 1.0*y0[t[0]];
    else
      r[i] = sum(s*y0)/sum(s);
  }

  return length(r)==1 ? r[0]: r;
}

define lagrange_poly_deriv (x, y)
%!%+
%\function{lagrange_poly_deriv}
%\synopsis{Calculate y values of the derivative of a Lagrange polynomial}
%\usage{dy = lagrange_poly_deriv(x, y);}
%\qualifiers{
%  \qualifier{w}{calculated weights can be passed directly}
%}
%\description
%  Returns the y values of the derivative of the lagrange polynomial at the
%  given points x according to Berrut & Trefethen 2004. Interpolating those
%  points with the same weights gives the full derivative function.
%
%\example
%  variable x = chebyshev_nodes(20; min=-2, max=2);
%  variable w = chebyshev_lagrange_weights(20);
%  variable f = sin(x);
%  variable p = lagrange_poly([-5:5:#100], x, f; w=w);
%  variable df = lagrange_poly_deriv(x, f);
%  variable dp = lagrange_poly([-5:5:#100], x, df; w=w);
%
%\seealso{lagrange_poly,lagrange_weights}
%!%-
{
  variable r = Double_Type[length(y)];
  variable t, w, i;

  w = qualifier("w", lagrange_weights(x));

  _for i (0, length(y)-1)
  {
    t = [[0:i-1],[i+1:length(y)-1]];
    r[i] = -sum(w[t]/w[i]*(y[t]-y[i])/(x[t]-x[i]));
  }

  return length(r)==1 ? r[0] : r;
}

%
% Code from src/data/lightcurve/lc_from_events.sl
%
private define _lc_from_event_single(ff,b_lo,b_hi,band_max)
{
   variable s_filt;
   % if the upper boundary is equal to the maximal value: include it! 
   % (otherwise bins are lost)
   if (b_hi == band_max)
   {	 
      s_filt = struct_filter (ff, where(ff.pi >= b_lo and ff.pi <= b_hi ); copy);
   }
   else
   {
      s_filt = struct_filter (ff, where(ff.pi >= b_lo and ff.pi < b_hi ); copy);
   }
   return s_filt;
}

%%%%%%%%%%%%%%%%%%%%%%%
define lc_from_events()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{lc_from_events}
%\synopsis{extracts light curves from an event list}
%\usage{Struct_Type dat = lc_from_events(Struct_Type evts [, Array_Type bands]);}
%\qualifiers{
%\qualifier{dt}{time resolution in sec [default: 100]}
%\qualifier{back}{Struct_Type b_evts : subtract background events}
%\qualifier{gti}{struct { Double_Type start, stop } : GTIs for events}
%\qualifier{minfracexp}{minimum allowed fractional exposure of a time bin, requires GTIs to be set}
%}
%\description
%    This function extracts light curves from an event list given as the
%    following structure:
%      Double_Type[] time - the event arrival times (in seconds)
%      Double_Type[] pi   - the associated event energies (in eV) 
%    If \code{bands}
%    is given, an array of structures containing light curves of each band
%    is returned. The energy bands have to be given in eV. To receive
%    two lightcurves with energies between 1-2keV and 2-3keV, e.g., 
%
%    If GTIs are provided, the fractional exposure 'fracexp' will be added
%    to the output structure.
%    
%    \code{dat = lc_from_events(evts, [1000,2000,3000]);}
%\seealso{color_color_data, histogram}
%!%-
{

   variable ff, bands;
   switch (_NARGS)
   {case 1: ff = ();}
   {case 2: (ff, bands) = ();}
   { help(_function_name()); return;}

%   print(typeof(ff));
   if (typeof(ff) == String_Type) { ff = fits_read_table(ff);}
   
   variable band_min = min(ff.pi); variable band_max = max(ff.pi);
   
   ifnot (__is_initialized(&bands))
   {
      bands = [band_min, band_max];
   }

   variable i, band_lo, band_hi;
   variable n_bands = length(bands)-1;
   variable lc_bands = Struct_Type[n_bands];
   variable s_filt, b_filt;


   % establish the time-grid 
   variable dt = qualifier("dt",100);
   variable tmin = min(ff.time); variable tmax = max(ff.time);
   
   variable t_lo = [tmin:tmax:dt];
   variable t_hi = [t_lo[[1:]],tmax];
   
   % GTI filtering
   variable fracexp = NULL;
   if (qualifier_exists("gti")) {
     variable qual = struct { fracexp = &fracexp };
     if (qualifier_exists("minfracexp")) {
       qual = struct_combine(qual, struct { minfracexp = qualifier("minfracexp") });
     }
     i = filter_gti(t_lo, t_hi, qualifier("gti");; qual);
     t_lo = t_lo[i];
     t_hi = t_hi[i];
   }

   _for i(0,n_bands-1,1)
   {
      band_lo = bands[i];
      band_hi = bands[i+1];

      % make the light curve
      s_filt = _lc_from_event_single(ff,band_lo,band_hi,band_max);
      variable cts = histogram(s_filt.time, t_lo, t_hi);

      % exculde the back ground?
      variable bg = 0; variable q = qualifier("back");
      if (q != NULL)
      {
	 b_filt = _lc_from_event_single(q,band_lo,band_hi,band_max);
	 bg = histogram(b_filt.time, t_lo, t_hi);
      }
      
      variable dt_bins = t_hi-t_lo;
      
      lc_bands[i] = struct {
	 time = 0.5*(t_lo+t_hi),
	 rate = (cts-bg)/dt_bins,
	 err = sqrt(cts+bg)/dt_bins
      };
      if (fracexp != NULL) {
	lc_bands[i] = struct_combine(lc_bands[i], struct { fracexp = fracexp });
      }
   }

   
   return (length(lc_bands) > 1) ? lc_bands : lc_bands[0] ;
}


%
% Code from src/data/timing/lcseg.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% ISIS timing-tools
%
% lcseg.sl - Tools for reading FITS-lightcurves from extraction directories
% 	             - cutting lightcurves into segments of given size
% 	             - sync'ing lightcurves to latest start and earliest stop
% 	             - search (time)arrays for gaps
% 	             - write gzipped, segmented FITS-lightcurve set to HD
%
% Functions: timegap()
% 	     timeseg()
% 	     lcsync()
% 	     lcseg()
% 	     lcsegcombine()
% 	     lcsegwrap()
%
% Short Descriptions:
%
% (gaps, gapdura, dblock) = timegap (timearray)
%
% Qualifiers: tolerance (Changes the maximum allowed relative deviation from the bintime)
% 	      bintime (Changes the bintime - standard: zeit[1]-zeit[0])
% Outputs:    gaps (Starting bins of gaps in timearray)
% 	      gapdura (Duration of the gaps)
% 	      dblock (Duration of the uninterupted parts of the timearray)
%
% (ndx, gaps, gapdura) = timeseg (timearray, dimseg)
%
% Qualifiers: tolerance (Changes the maximum allowed relative deviation from the bintime)
%  	      searchgap (Switches the use of timegap() on/off)
% Outputs:    gaps (Starting bins of gaps in timearray)
% 	      gapdura (Duration of the gaps)
% 	      ndx (Array of indices that can be used to segment the lightcurve)
%
% () = lcsegcombine (lcname, dimseg)
%
% Qualifiers: checktimes (When activated, the timearrays are checked for inconsistencies)
% Outputs:    None!
% Effects:    Writes a gzipped FITS file to disk, containing the complete
% XTE/PCA lightcurve structure for foucalc()
%
% lcsegcombine() has to be executed in the extraction-root-path to
% work correctly:
% (crux:/data/user/pirner/x-ray/cygx-1/rxte/)
% (pulsar:/proj.stand/mpai45/cygx1/extractions/)
%
% (lc) = lcsync (lclo,lchi)
%
% Inputs:  Structures of 4 low PCA(B) channels and 2 high PCA(E) channels
% Outputs: Structure with all 6 channels (sync'ed)
%
% () = lcsegwrap (lcname,dimseg[n])
%
% Effects: Same as lcsegcombine, dimseg may be an array of dimsegs
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define XTEtoRealtime (t)
{
    variable st;
    variable rt;
    st = dateOfMJD(MJDref_satellite("XTE")+(t/86400.));
    rt = string(st.year)+"-"+sprintf("%02d",st.month)+"-"+sprintf("%02d",st.day)+"_"+sprintf("%02d",st.hour)+"-"+sprintf("%02d",st.minute)+"-"+sprintf("%02.f",floor(st.second));
    return rt;
}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define gapsync (tlo,thi)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable bintime = qualifier ("bintime", tlo[1]-tlo[0]);
  variable tolerance = qualifier ("tolerance", 1e-8);

  variable lo_bad = where(tlo<thi[0] or tlo>thi[-1]);
  variable hi_bad = where(thi<tlo[0] or thi>tlo[-1]);

  message ("gapsync: Syncing common times in both EAs");
  variable igap;

  variable deltalo = (shift(tlo,1)-tlo)[[:-2]];
  foreach igap (where(abs(deltalo-bintime)/bintime > tolerance))
      hi_bad = [hi_bad, where( tlo[igap] < thi < tlo[igap+1] )];

  variable deltahi = (shift(thi,1)-thi)[[:-2]];
  foreach igap (where(abs(deltahi-bintime)/bintime > tolerance))
      lo_bad = [lo_bad, where( thi[igap] < tlo < thi[igap+1] )];

  % Create index arrays for lower and upper times

  return complementary_array([0:length(tlo)-1], lo_bad),
  complementary_array([0:length(thi)-1], hi_bad);
}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define gapsync2 (tlo,thi)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{

  variable bintime = qualifier ("bintime", tlo[1]-tlo[0]);
  variable tolerance = qualifier ("tolerance", 1e-8);
  variable deltalo = (shift(tlo,1)-tlo)[[0:length(tlo)-2]];
  variable deltahi = (shift(thi,1)-thi)[[0:length(thi)-2]];
  variable gaplo = where(abs(deltalo-bintime)/bintime > tolerance);
  variable gaphi = where(abs(deltahi-bintime)/bintime > tolerance);
  variable lengthlo = length (tlo);
  variable lengthhi = length (thi);

  variable lowstart=0;

    if (tlo[0] < thi[0])
    {
            lowstart = where(tlo == thi[0])[0];
    }

    variable highstart=0;

    if (tlo[0] > thi[0])
    {
            highstart = where (thi == tlo[0])[0];
    }

  message ("gapsync: Syncing common times in both EAs");

  variable tgaplo = tlo[gaplo];
  variable tgaphi = thi[gaphi];
  variable tgaploend = tlo[gaplo+1];
  variable tgaphiend = thi[gaphi+1];

  variable startlo = lowstart;
  variable starthi = highstart;
  variable stoplo = Integer_Type[0];
  variable stophi = Integer_Type[0];

  variable gap;
  foreach gap (tgaphiend)
    {
       startlo = [startlo, where(tlo == gap)];
    }

  foreach gap (tgaploend)
    {
      starthi = [starthi, where(thi == gap)];
    }

  foreach gap (tgaphi)
    {
      stoplo = [stoplo, where(tlo == gap)];
    }

  foreach gap (tgaplo)
    {
      stophi = [stophi, where(thi == gap)];
    }

  variable lowstop = lengthlo-1;;

  if (tlo[-1] > thi[-1])
    {
      lowstop = where(tlo == thi[-1])[0];
    }

  variable highstop = lengthhi-1;

  if (tlo[-1] < thi[-1])
    {
      highstop = where(thi == tlo[-1])[0];
    }

  stoplo = [stoplo,lowstop];
  stophi = [stophi,highstop];

  % Create index arrays for lower and upper times

  variable blockslo = length (startlo);
  variable blockshi = length (starthi);
  variable ndxlo = Integer_Type[0];
  variable ndxhi = Integer_Type[0];

  % Create lower indices

  variable i;
  for (i=0;i<blockslo;i++)
    {
      ndxlo = [ndxlo, [startlo[i]:stoplo[i]]];
    }
  for (i=0;i<blockshi;i++)
    {
      ndxhi = [ndxhi, [starthi[i]:stophi[i]]];
    }

  return ndxlo,ndxhi;

}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define timesync (timelo,timehi)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable idxlo = Integer_Type[0];
  variable idxhi = Integer_Type[0];
  variable pointerlo;
  variable pointerhi;
  variable lengthlo = length(timelo);
  variable lengthhi = length(timehi);

  for (pointerlo=0,pointerhi=0;pointerlo<lengthlo && pointerhi<lengthhi;)
    {
      if (timelo[pointerlo] == timehi[pointerhi])
	{
	  idxlo = [idxlo,pointerlo];
	  idxhi = [idxhi,pointerhi];
	  pointerlo++;
	  pointerhi++;
	}
      else if (timelo[pointerlo] < timehi[pointerhi])
	{
	  pointerlo++;
	}
      else
	{
	  pointerhi++;
	}
    }

  return idxlo, idxhi;
}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define timegap (zeit)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable tolerance = qualifier ("tolerance", 1e-8);
  variable bintime = qualifier ("bintime", zeit[1]-zeit[0]);
  variable deltaT = (shift(zeit,1)-zeit)[[0:length(zeit)-2]];
  variable gaps = where(abs(deltaT-bintime)/bintime > tolerance);
  variable numgaps = length (gaps);
  variable dblock;
  variable gapdura;

% Check for bins shorter than the bintime!
  variable warn = length(where(deltaT < bintime));
  if (warn > 0)
    {
      message ("timegap: There are time differences shorter than the bintime!");
    }

  if (length(gaps)==0)
    {
      message ("timegap: There are no gaps in the time array");
      dblock = length(zeit);
      gapdura = Integer_Type[0];
    }
  else
    {
      message ("timegap: There are "+string(numgaps)+" gaps in the time array");
      dblock = gaps[0]+1;
      dblock = [dblock, shift(gaps,1)-gaps];
      dblock[-1] = length(zeit)-(gaps[-1]+1);
      gapdura = zeit[gaps+1]-zeit[gaps];
    }

  return gaps, gapdura, dblock;
}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define timeseg (zeit, dimseg)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable tolerance = qualifier ("tolerance", 1e-8);
  variable searchgap = qualifier ("searchgap", 1);
  variable bintime = qualifier ("bintime", zeit[1]-zeit[0]);
  variable ndx;
  variable gaps;
  variable gapdura;
  variable dblock;
  variable numseg;
  variable numtime;
  variable dstart;
  variable zseg;
  variable zndx;
  variable start;
  variable i;
  variable j;
  variable nn;
  variable tdseg;
  variable dtdseg;
  variable ndxtest;

  if (searchgap != 1)
    {
      numseg=int(length(zeit)/dimseg);
      numtime=numseg*dimseg;
      ndx=[0:numtime-1];
      gaps = NULL;
      gapdura = NULL;
    }
  else
    {
      (gaps, gapdura, dblock) = timegap(zeit;tolerance=tolerance,bintime=bintime);

      if (length(gaps) > 0)
	{
	  dstart=[0,gaps+1];
	}
      else
	{
	  dstart = 0;
	}

      zseg = int(dblock/dimseg);
      zndx = where (dblock >= dimseg);
      nn = length (zndx);

      if (nn > 0)
	{
	  dstart = dstart[zndx];
          dblock = dblock[zndx];
	  zseg = zseg[zndx];
	  start = -1;

	  for (i=0;i<nn;i++)
	    {
	      start = [ start , dstart[i] + dimseg*[0:zseg[i]-1] ];
	    }

	  start = start[[1:]];

      	}
      else
	{
	  message ("timeseg: Warning! Data do not contain a valid segment...");
	  return NULL,NULL,NULL, "noseg";
	}

      ndx = Int_Type[length(start)*dimseg];

      for (i=0; i<length(start); i++)
	{
	  ndx[[(i*dimseg):(((i+1)*dimseg)-1)]] = start[i]+[0:dimseg-1];
  	}

      tdseg = zeit[ndx[0]] - zeit [ndx[dimseg-1]];
      dtdseg = Double_Type[length(ndx)/dimseg];

      for (i=0; i<(length(ndx)/dimseg);i++)
	{
          dtdseg[i] = tdseg -(zeit[ndx[i*dimseg]] - zeit[ndx[((i+1)*dimseg)-1]]);
	}

      ndxtest = where ( dtdseg != 0.);

      if (length(ndxtest) != 0)
	{
	  message ("timeseg: dt != bintime in following indices:");
	  print (ndxtest);
	}

      start=0.;
    }

  return ndx, gaps, gapdura, NULL;

}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define lcseg (structarray,dimseg)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable checktimes = qualifier ("checktimes", 0);
  variable segsize = string(dimseg);
  variable searchgap = qualifier ("searchgap", 1);
  variable i;
  variable ndx;
  variable gaps;
  variable gapdura;

  (ndx,gaps,gapdura) = timeseg(structarray.time,dimseg;searchgap=searchgap);

  structarray.time = structarray.time[ndx];

  variable rarray;
  foreach rarray (get_struct_field_names(structarray)[[1:]])
    {
      set_struct_field (structarray, rarray, get_struct_field(structarray, rarray)[ndx]);
    }

  return structarray,gaps,gapdura;
}

%
% Code from src/slang/date/leapyear.sl
%
define leapyear()
%!%+
%\function{leapyear}
%\synopsis{tells whether a year is a leapyear in the Gregorian calendar}
%\usage{Char_Type leapyear(Integer_Type y)}
%\description
%    For a leapyear in the Gregorian calendar, \code{y} is either
%    divisible by 4 but not by 100,
%    or \code{y} is divisible by 400.
%\seealso{daysPerMonth}
%!%-
{
  variable y;
  switch(_NARGS)
  { case 1: y = (); }
  { help(_function_name()); return; }

  return (y mod 4==0 and y mod 100>0) or y mod 400==0;
}

%
% Code from src/fitting/linear_fit_xerr_yerr_data.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define linear_fit_xerr_yerr_data()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{linear_fit_xerr_yerr_data}
%\synopsis{fits a linear function to data points with both x- and y-errors}
%\usage{(Double_Type a, b) = linear_fit_xerr_yerr_data(Double_Type X[], Xerr[], Y[], Yerr[][, Double_Type b0]);
%\altusage{(Double_Type a, aerr, b, berr) = linear_fit_xerr_yerr_data(Double_Type X[], Xerr[], Y[], Yerr[][, Double_Type b0]; with_errors);}
%}
%\qualifiers{
%\qualifier{verbose}{}
%\qualifier{n}{number of iterations}
%\qualifier{with_errors}{The 90% confidence interval (Delta chi^2 = 2.7) is computed as well.}
%}
%\description
%    The data points \code{(X[i]+-Xerr[i], Y[i]+-Yerr[i])}
%    are iteratively fitted with the linear function \code{y = a + b*x}.
%    If no initial value for the slope b is specified,
%       \code{b0 = s * [max(Y)-min(Y)]/[max(X)-min(X)]}
%    is used.
%    The best fit is found iteratively after the algorithm
%    by Fasano & Vio (1988), BICDS 35, p.191.
%\seealso{linear_regression}
%!%-
{
  variable X, Xerr, Y, Yerr, a, b;

  switch(_NARGS)
  { case 4: (X, Xerr, Y, Yerr   ) = (); b = (max(Y)-min(Y))/(max(X)-min(X)); }
  { case 5: (X, Xerr, Y, Yerr, b) = (); }
  { help(_function_name()); return; }

  variable i, W, avX, avY, alpha, beta, gamma;
  if(qualifier_exists("verbose"))  vmessage("b=%g", b);
  for(i=0; i<qualifier("n", 20); i++)
  {
    W = 1/(b^2*Xerr^2 + Yerr^2);
    avX = weighted_mean(X, W);
    avY = weighted_mean(Y, W);
    a = avY - b * avX;
    alpha = sum( W^2 *  Xerr^2 * (Y-avY) * (X-avX) );
    beta  = sum( W^2 * (Yerr^2 * (X-avX)^2 - Xerr^2 * (Y-avY)^2) );
    gamma = sum( W^2 *  Yerr^2 * (Y-avY) * (X-avX) );
    b = (-beta + sqrt(beta^2+4*alpha*gamma))/(2*alpha);
    if(qualifier_exists("verbose"))  vmessage("a=%g, b=%g", a, b);
  }

  if(qualifier_exists("with_errors"))
  {
    variable deltaChi2 = 2.7055;
    variable delta_a = sqrt(deltaChi2 / sum( W ) );
    variable delta_b = sqrt(deltaChi2 / sum( ( X^2*Yerr^4 + 3*b^2*Xerr^4*Y^2 + 3*b^2*Xerr^4*a^2 - Xerr^2*Y^2*Yerr^2 - Xerr^2*a^2*Yerr^2 - 3*X^2*b^2*Xerr^2*Yerr^2
					      -2*X*b^3*Xerr^4*Y + 2*X*b^3*Xerr^4*a - 6*b^2*Xerr^4*Y*a + 2*Xerr^2*Y*a*Yerr^2 + 6*X*b*Xerr^2*Y*Yerr^2 - 6*X*b*Xerr^2*a*Yerr^2
					     )/(b^2*Xerr^2+Yerr^2)^3 ) );
    return (a, delta_a, b, delta_b);
  }
  else
    return (a, b);
}

%
% Code from src/slang/math/linear_regression.sl
%
define linear_regression()
%!%+
%\function{linear_regression}
%\synopsis{computes a linear regression fit}
%\usage{Struct_Type = linear_regression(Double_Type[] x, y[, err]);}
%\description
%    \code{err} is the uncertainty of the \code{y} values.
%    If \code{err} is not specified, all data points are weighted equally.
%
%    The parameters \code{a} and \code{b} of the best fit \code{y = a + b*x}
%    are   \code{a = (Sy*Sxx - Sx*Sxy)/D}\n
%      and \code{b = (S *Sxy - Sx*Sy )/D}\n
%    where \code{S  = sum(1/err^2)}\n
%          \code{Sx = sum(x/err^2)},  \code{Sxx = sum(x*x/err^2)}\n
%          \code{Sy = sum(y/err^2)},  \code{Sxy = sum(x*y/err^2)}\n
%      and \code{D = S*Sxx - Sx^2}.
%
%    The returned structure contains the fields \code{a} and \code{b},
%    and the errors \code{da} and \code{db}, respectively,
%    as well as the reduced chi square \code{chisq} of the fit.
%\seealso{linear_fit_xerr_yerr_data}
%!%-
{
  variable x, y, err;
  switch(_NARGS)
  { case 2: (x, y) = (); err = Integer_Type[length(y)] + 1; }
  { case 3: (x, y, err) = (); }
  { return help(_function_name()); }

  variable err2 = sqr(err);
  variable S   = sum( 1 /err2);
  variable Sx  = sum( x /err2);
  variable Sy  = sum( y /err2);
  variable Sxx = sum(x*x/err2);
  variable Sxy = sum(x*y/err2);

  variable D = S*Sxx - Sx^2;
  variable a = (Sy*Sxx-Sx*Sxy)/D;
  variable b = (S*Sxy-Sx*Sy)/D;
  return struct {
    a  = a,
    b  = b,
    da = sqrt(Sxx/D),
    db = sqrt(S/D),
    chisq = sumsq((a+b*x-y)/err) / (length(x)-2)  % <- should rather be called chisq_red or red_chi2
  };
}

%
% Code from src/fitting/xyfit/fit-functions/linear_xyfit.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define linear_xyfit()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{linear_xyfit}
%\synopsis{linear xy fit function to be used with xyfit_fun}
%\usage{xyfit_fun ("linear");}
%\description
%    This function is not meant to be called directly!
%    
%    Calling \code{xyfit_fun ("linear");} sets up a linear fit
%    function for xy-data. It has the form \code{y = a*x + b},
%    where \code{a} and \code{b} are the two fit parameters.
%\seealso{xyfit_fun, define_xydata, plot_xyfit, linear_regression}
%!%-
{
  variable xref, yref, par;
  switch(_NARGS)
  { case 0: return ["a [coefficient of x]", "b [additive constant]"]; } % fit parameters
  { case 3: (xref, yref, par)=(); }
  { return help(_function_name); }

  @yref = par[0] * @xref + par[1];
}

define linear_xyfit_default(i)
{
  switch(i)
  { case 0: return (1, 0, -1e5, 1e5); }
  { case 1: return (0, 0, -1e5, 1e5); }
}

%
% Code from src/fitting/fit-functions/lines/lines.sl
%
%%%%%%%%%%%%%%%%
define lines_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%
%!%+
%\function{lines}
%\synopsis{model for gaussian line profiles of highly ionized ions' transitions}
%\usage{fit_fun("lines(id)");}
%\description
%     The return values are the bin-averages of  1 + sum_i gauss_i(lambda) ,
%     i.e., lines is a multiplicative model.
%\seealso{gauss, set_lines_par_fun}
%!%-
{
  variable i, g = 0*bin_lo;
  for(i=0; i<199; i++)
  { variable EW = par[4*i+1]/1e3;
    if(EW!=0) %                                  gauss.sigma = FWHM / (2*sqrt{2*ln(2)})
    { g += eval_fun2("gauss", bin_lo, bin_hi, [EW, par[4*i], par[4*i+2]*0.0004246609]); }
  }
  return 1. + g/(bin_hi-bin_lo);
}

%%%%%%%%%%%%%%%%%%%%%
define lines_defaults(i)
%%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
% O8a (O VIII alpha)
  { case 0: return (18.9689, 1, 18.9057, 19.0321); }
  { case 1: return (0, 1, -100, 0); }
  { case 2: return (10, 1, 5, 76); }
  { case 3: return (0, 1, -1, 10); }
% O8b (O VIII beta)
  { case 4: return (16.0059, 1, 15.9525, 16.0593); }
  { case 5: return (0, 1, -100, 0); }
  { case 6: return (10, 1, 5, 64); }
  { case 7: return (0, 1, -1, 10); }
% O8g (O VIII gamma)
  { case 8: return (15.1762, 1, 15.1256, 15.2268); }
  { case 9: return (0, 1, -100, 0); }
  { case 10: return (10, 1, 5, 61); }
  { case 11: return (0, 1, -1, 10); }
% O8d (O VIII delta)
  { case 12: return (14.8206, 1, 14.7712, 14.8700); }
  { case 13: return (0, 1, -100, 0); }
  { case 14: return (10, 1, 5, 59); }
  { case 15: return (0, 1, -1, 10); }
% O7f (O VII f)
  { case 16: return (22.0977, 1, 22.0241, 22.1714); }
  { case 17: return (0, 1, -100, 0); }
  { case 18: return (10, 1, 5, 88); }
  { case 19: return (0, 1, -1, 10); }
% O7i (O VII i)
  { case 20: return (21.8036, 1, 21.7310, 21.8763); }
  { case 21: return (0, 1, -100, 0); }
  { case 22: return (10, 1, 5, 87); }
  { case 23: return (0, 1, -1, 10); }
% O7a (O VII alpha)
  { case 24: return (21.6015, 1, 21.5295, 21.6735); }
  { case 25: return (0, 1, -100, 0); }
  { case 26: return (10, 1, 5, 86); }
  { case 27: return (0, 1, -1, 10); }
% O7b (O VII beta)
  { case 28: return (18.6270, 1, 18.5649, 18.6891); }
  { case 29: return (0, 1, -100, 0); }
  { case 30: return (10, 1, 5, 75); }
  { case 31: return (0, 1, -1, 10); }
% O7g (O VII gamma)
  { case 32: return (17.7680, 1, 17.7088, 17.8272); }
  { case 33: return (0, 1, -100, 0); }
  { case 34: return (10, 1, 5, 71); }
  { case 35: return (0, 1, -1, 10); }
% O7d (O VII delta)
  { case 36: return (17.3960, 1, 17.3380, 17.4540); }
  { case 37: return (0, 1, -100, 0); }
  { case 38: return (10, 1, 5, 70); }
  { case 39: return (0, 1, -1, 10); }
% Ne10a (Ne X alpha)
  { case 40: return (12.1339, 1, 12.0934, 12.1743); }
  { case 41: return (0, 1, -100, 0); }
  { case 42: return (10, 1, 5, 49); }
  { case 43: return (0, 1, -1, 10); }
% Ne10b (Ne X beta)
  { case 44: return (10.2389, 1, 10.2047, 10.2730); }
  { case 45: return (0, 1, -100, 0); }
  { case 46: return (10, 1, 5, 41); }
  { case 47: return (0, 1, -1, 10); }
% Ne10g (Ne X gamma)
  { case 48: return (9.7082, 1, 9.6758, 9.7405); }
  { case 49: return (0, 1, -100, 0); }
  { case 50: return (10, 1, 5, 39); }
  { case 51: return (0, 1, -1, 10); }
% Ne10d (Ne X delta)
  { case 52: return (9.4807, 1, 9.4491, 9.5123); }
  { case 53: return (0, 1, -100, 0); }
  { case 54: return (10, 1, 5, 38); }
  { case 55: return (0, 1, -1, 10); }
% Ne10e (Ne X epsilon)
  { case 56: return (9.3616, 1, 9.3304, 9.3928); }
  { case 57: return (0, 1, -100, 0); }
  { case 58: return (10, 1, 5, 37); }
  { case 59: return (0, 1, -1, 10); }
% Ne10z (Ne X zeta)
  { case 60: return (9.2912, 1, 9.2603, 9.3222); }
  { case 61: return (0, 1, -100, 0); }
  { case 62: return (10, 1, 5, 37); }
  { case 63: return (0, 1, -1, 10); }
% Ne9f (Ne IX f)
  { case 64: return (13.6990, 1, 13.6533, 13.7446); }
  { case 65: return (0, 1, -100, 0); }
  { case 66: return (10, 1, 5, 55); }
  { case 67: return (0, 1, -1, 10); }
% Ne9i (Ne IX i)
  { case 68: return (13.5531, 1, 13.5079, 13.5983); }
  { case 69: return (0, 1, -100, 0); }
  { case 70: return (10, 1, 5, 54); }
  { case 71: return (0, 1, -1, 10); }
% Ne9a (Ne IX alpha)
  { case 72: return (13.4473, 1, 13.4025, 13.4921); }
  { case 73: return (0, 1, -100, 0); }
  { case 74: return (10, 1, 5, 54); }
  { case 75: return (0, 1, -1, 10); }
% Ne9b (Ne IX beta)
  { case 76: return (11.5440, 1, 11.5055, 11.5825); }
  { case 77: return (0, 1, -100, 0); }
  { case 78: return (10, 1, 5, 46); }
  { case 79: return (0, 1, -1, 10); }
% Ne9g (Ne IX gamma)
  { case 80: return (11.0010, 1, 10.9643, 11.0377); }
  { case 81: return (0, 1, -100, 0); }
  { case 82: return (10, 1, 5, 44); }
  { case 83: return (0, 1, -1, 10); }
% Ne9d (Ne IX delta)
  { case 84: return (10.7650, 1, 10.7291, 10.8009); }
  { case 85: return (0, 1, -100, 0); }
  { case 86: return (10, 1, 5, 43); }
  { case 87: return (0, 1, -1, 10); }
% Ne9e (Ne IX epsilon)
  { case 88: return (10.6426, 1, 10.6071, 10.6781); }
  { case 89: return (0, 1, -100, 0); }
  { case 90: return (10, 1, 5, 43); }
  { case 91: return (0, 1, -1, 10); }
% Ne9z (Ne IX zeta)
  { case 92: return (10.5650, 1, 10.5298, 10.6002); }
  { case 93: return (0, 1, -100, 0); }
  { case 94: return (10, 1, 5, 42); }
  { case 95: return (0, 1, -1, 10); }
% Na11a (Na XI alpha)
  { case 96: return (10.0250, 1, 9.9916, 10.0584); }
  { case 97: return (0, 1, -100, 0); }
  { case 98: return (10, 1, 5, 40); }
  { case 99: return (0, 1, -1, 10); }
% Na11b (Na XI beta)
  { case 100: return (8.4595, 1, 8.4313, 8.4877); }
  { case 101: return (0, 1, -100, 0); }
  { case 102: return (10, 1, 5, 34); }
  { case 103: return (0, 1, -1, 10); }
% Na11g (Na XI gamma)
  { case 104: return (8.0211, 1, 7.9943, 8.0478); }
  { case 105: return (0, 1, -100, 0); }
  { case 106: return (10, 1, 5, 32); }
  { case 107: return (0, 1, -1, 10); }
% Na11d (Na XI delta)
  { case 108: return (7.8332, 1, 7.8071, 7.8593); }
  { case 109: return (0, 1, -100, 0); }
  { case 110: return (10, 1, 5, 31); }
  { case 111: return (0, 1, -1, 10); }
% Na10f (Na X f)
  { case 112: return (11.1900, 1, 11.1527, 11.2273); }
  { case 113: return (0, 1, -100, 0); }
  { case 114: return (10, 1, 5, 45); }
  { case 115: return (0, 1, -1, 10); }
% Na10i (Na X i)
  { case 116: return (11.0800, 1, 11.0431, 11.1169); }
  { case 117: return (0, 1, -100, 0); }
  { case 118: return (10, 1, 5, 44); }
  { case 119: return (0, 1, -1, 10); }
% Na10a (Na X alpha)
  { case 120: return (11.0027, 1, 10.9660, 11.0394); }
  { case 121: return (0, 1, -100, 0); }
  { case 122: return (10, 1, 5, 44); }
  { case 123: return (0, 1, -1, 10); }
% Na10b (Na X beta)
  { case 124: return (9.4330, 1, 9.4016, 9.4644); }
  { case 125: return (0, 1, -100, 0); }
  { case 126: return (10, 1, 5, 38); }
  { case 127: return (0, 1, -1, 10); }
% Na10g (Na X gamma)
  { case 128: return (8.9828, 1, 8.9529, 9.0127); }
  { case 129: return (0, 1, -100, 0); }
  { case 130: return (10, 1, 5, 36); }
  { case 131: return (0, 1, -1, 10); }
% Na10d (Na X delta)
  { case 132: return (8.7884, 1, 8.7591, 8.8177); }
  { case 133: return (0, 1, -100, 0); }
  { case 134: return (10, 1, 5, 35); }
  { case 135: return (0, 1, -1, 10); }
% Mg12a (Mg XII alpha)
  { case 136: return (8.4210, 1, 8.3929, 8.4491); }
  { case 137: return (0, 1, -100, 0); }
  { case 138: return (10, 1, 5, 34); }
  { case 139: return (0, 1, -1, 10); }
% Mg12b (Mg XII beta)
  { case 140: return (7.1062, 1, 7.0825, 7.1298); }
  { case 141: return (0, 1, -100, 0); }
  { case 142: return (10, 1, 5, 28); }
  { case 143: return (0, 1, -1, 10); }
% Mg12g (Mg XII gamma)
  { case 144: return (6.7379, 1, 6.7154, 6.7604); }
  { case 145: return (0, 1, -100, 0); }
  { case 146: return (10, 1, 5, 27); }
  { case 147: return (0, 1, -1, 10); }
% Mg12d (Mg XII delta)
  { case 148: return (6.5801, 1, 6.5582, 6.6020); }
  { case 149: return (0, 1, -100, 0); }
  { case 150: return (10, 1, 5, 26); }
  { case 151: return (0, 1, -1, 10); }
% Mg11f (Mg XI f)
  { case 152: return (9.3143, 1, 9.2833, 9.3454); }
  { case 153: return (0, 1, -100, 0); }
  { case 154: return (10, 1, 5, 37); }
  { case 155: return (0, 1, -1, 10); }
% Mg11i (Mg XI i)
  { case 156: return (9.2312, 1, 9.2004, 9.2620); }
  { case 157: return (0, 1, -100, 0); }
  { case 158: return (10, 1, 5, 37); }
  { case 159: return (0, 1, -1, 10); }
% Mg11a (Mg XI alpha)
  { case 160: return (9.1687, 1, 9.1382, 9.1993); }
  { case 161: return (0, 1, -100, 0); }
  { case 162: return (10, 1, 5, 37); }
  { case 163: return (0, 1, -1, 10); }
% Mg11b (Mg XI beta)
  { case 164: return (7.8503, 1, 7.8241, 7.8765); }
  { case 165: return (0, 1, -100, 0); }
  { case 166: return (10, 1, 5, 31); }
  { case 167: return (0, 1, -1, 10); }
% Mg11g (Mg XI gamma)
  { case 168: return (7.4730, 1, 7.4481, 7.4979); }
  { case 169: return (0, 1, -100, 0); }
  { case 170: return (10, 1, 5, 30); }
  { case 171: return (0, 1, -1, 10); }
% Mg11d (Mg XI delta)
  { case 172: return (7.3101, 1, 7.2857, 7.3345); }
  { case 173: return (0, 1, -100, 0); }
  { case 174: return (10, 1, 5, 29); }
  { case 175: return (0, 1, -1, 10); }
% Al13a (Al XIII alpha)
  { case 176: return (7.1728, 1, 7.1489, 7.1967); }
  { case 177: return (0, 1, -100, 0); }
  { case 178: return (10, 1, 5, 29); }
  { case 179: return (0, 1, -1, 10); }
% Al13b (Al XIII beta)
  { case 180: return (6.0530, 1, 6.0328, 6.0731); }
  { case 181: return (0, 1, -100, 0); }
  { case 182: return (10, 1, 5, 24); }
  { case 183: return (0, 1, -1, 10); }
% Al13g (Al XIII gamma)
  { case 184: return (5.7393, 1, 5.7202, 5.7585); }
  { case 185: return (0, 1, -100, 0); }
  { case 186: return (10, 1, 5, 23); }
  { case 187: return (0, 1, -1, 10); }
% Al13d (Al XIII delta)
  { case 188: return (5.6049, 1, 5.5862, 5.6235); }
  { case 189: return (0, 1, -100, 0); }
  { case 190: return (10, 1, 5, 22); }
  { case 191: return (0, 1, -1, 10); }
% Al12f (Al XII f)
  { case 192: return (7.8721, 1, 7.8459, 7.8984); }
  { case 193: return (0, 1, -100, 0); }
  { case 194: return (10, 1, 5, 31); }
  { case 195: return (0, 1, -1, 10); }
% Al12i (Al XII i)
  { case 196: return (7.8070, 1, 7.7809, 7.8330); }
  { case 197: return (0, 1, -100, 0); }
  { case 198: return (10, 1, 5, 31); }
  { case 199: return (0, 1, -1, 10); }
% Al12a (Al XII alpha)
  { case 200: return (7.7573, 1, 7.7314, 7.7832); }
  { case 201: return (0, 1, -100, 0); }
  { case 202: return (10, 1, 5, 31); }
  { case 203: return (0, 1, -1, 10); }
% Al12b (Al XII beta)
  { case 204: return (6.6350, 1, 6.6129, 6.6571); }
  { case 205: return (0, 1, -100, 0); }
  { case 206: return (10, 1, 5, 27); }
  { case 207: return (0, 1, -1, 10); }
% Al12g (Al XII gamma)
  { case 208: return (6.3140, 1, 6.2930, 6.3350); }
  { case 209: return (0, 1, -100, 0); }
  { case 210: return (10, 1, 5, 25); }
  { case 211: return (0, 1, -1, 10); }
% Al12d (Al XII delta)
  { case 212: return (6.1750, 1, 6.1544, 6.1956); }
  { case 213: return (0, 1, -100, 0); }
  { case 214: return (10, 1, 5, 25); }
  { case 215: return (0, 1, -1, 10); }
% Si14a (Si XIV alpha)
  { case 216: return (6.1822, 1, 6.1616, 6.2028); }
  { case 217: return (0, 1, -100, 0); }
  { case 218: return (10, 1, 5, 25); }
  { case 219: return (0, 1, -1, 10); }
% Si14b (Si XIV beta)
  { case 220: return (5.2172, 1, 5.1998, 5.2346); }
  { case 221: return (0, 1, -100, 0); }
  { case 222: return (10, 1, 5, 21); }
  { case 223: return (0, 1, -1, 10); }
% Si14g (Si XIV gamma)
  { case 224: return (4.9469, 1, 4.9304, 4.9634); }
  { case 225: return (0, 1, -100, 0); }
  { case 226: return (10, 1, 5, 20); }
  { case 227: return (0, 1, -1, 10); }
% Si14d (Si XIV delta)
  { case 228: return (4.8311, 1, 4.8150, 4.8472); }
  { case 229: return (0, 1, -100, 0); }
  { case 230: return (10, 1, 5, 19); }
  { case 231: return (0, 1, -1, 10); }
% Si13f (Si XIII f)
  { case 232: return (6.7403, 1, 6.7178, 6.7628); }
  { case 233: return (0, 1, -100, 0); }
  { case 234: return (10, 1, 5, 27); }
  { case 235: return (0, 1, -1, 10); }
% Si13i (Si XIII i)
  { case 236: return (6.6882, 1, 6.6659, 6.7105); }
  { case 237: return (0, 1, -100, 0); }
  { case 238: return (10, 1, 5, 27); }
  { case 239: return (0, 1, -1, 10); }
% Si13a (Si XIII alpha)
  { case 240: return (6.6479, 1, 6.6258, 6.6701); }
  { case 241: return (0, 1, -100, 0); }
  { case 242: return (10, 1, 5, 27); }
  { case 243: return (0, 1, -1, 10); }
% Si13b (Si XIII beta)
  { case 244: return (5.6805, 1, 5.6616, 5.6994); }
  { case 245: return (0, 1, -100, 0); }
  { case 246: return (10, 1, 5, 23); }
  { case 247: return (0, 1, -1, 10); }
% Si13g (Si XIII gamma)
  { case 248: return (5.4045, 1, 5.3865, 5.4225); }
  { case 249: return (0, 1, -100, 0); }
  { case 250: return (10, 1, 5, 22); }
  { case 251: return (0, 1, -1, 10); }
% Si13d (Si XIII delta)
  { case 252: return (5.2850, 1, 5.2674, 5.3026); }
  { case 253: return (0, 1, -100, 0); }
  { case 254: return (10, 1, 5, 21); }
  { case 255: return (0, 1, -1, 10); }
% S16a (S XVI alpha)
  { case 256: return (4.7292, 1, 4.7134, 4.7449); }
  { case 257: return (0, 1, -100, 0); }
  { case 258: return (10, 1, 5, 19); }
  { case 259: return (0, 1, -1, 10); }
% S16b (S XVI beta)
  { case 260: return (3.9912, 1, 3.9779, 4.0045); }
  { case 261: return (0, 1, -100, 0); }
  { case 262: return (10, 1, 5, 16); }
  { case 263: return (0, 1, -1, 10); }
% S16g (S XVI gamma)
  { case 264: return (3.7845, 1, 3.7718, 3.7971); }
  { case 265: return (0, 1, -100, 0); }
  { case 266: return (10, 1, 5, 15); }
  { case 267: return (0, 1, -1, 10); }
% S16d (S XVI delta)
  { case 268: return (3.6959, 1, 3.6836, 3.7082); }
  { case 269: return (0, 1, -100, 0); }
  { case 270: return (10, 1, 5, 15); }
  { case 271: return (0, 1, -1, 10); }
% S15f (S XV f)
  { case 272: return (5.1015, 1, 5.0845, 5.1185); }
  { case 273: return (0, 1, -100, 0); }
  { case 274: return (10, 1, 5, 20); }
  { case 275: return (0, 1, -1, 10); }
% S15i (S XV i)
  { case 276: return (5.0665, 1, 5.0496, 5.0834); }
  { case 277: return (0, 1, -100, 0); }
  { case 278: return (10, 1, 5, 20); }
  { case 279: return (0, 1, -1, 10); }
% S15a (S XV alpha)
  { case 280: return (5.0387, 1, 5.0219, 5.0555); }
  { case 281: return (0, 1, -100, 0); }
  { case 282: return (10, 1, 5, 20); }
  { case 283: return (0, 1, -1, 10); }
% S15b (S XV beta)
  { case 284: return (4.2990, 1, 4.2847, 4.3133); }
  { case 285: return (0, 1, -100, 0); }
  { case 286: return (10, 1, 5, 17); }
  { case 287: return (0, 1, -1, 10); }
% S15g (S XV gamma)
  { case 288: return (4.0883, 1, 4.0747, 4.1019); }
  { case 289: return (0, 1, -100, 0); }
  { case 290: return (10, 1, 5, 16); }
  { case 291: return (0, 1, -1, 10); }
% S15d (S XV delta)
  { case 292: return (3.9980, 1, 3.9847, 4.0113); }
  { case 293: return (0, 1, -100, 0); }
  { case 294: return (10, 1, 5, 16); }
  { case 295: return (0, 1, -1, 10); }
% Ar18a (Ar XVIII alpha)
  { case 296: return (3.7329, 1, 3.7205, 3.7454); }
  { case 297: return (0, 1, -100, 0); }
  { case 298: return (10, 1, 5, 15); }
  { case 299: return (0, 1, -1, 10); }
% Ar18b (Ar XVIII beta)
  { case 300: return (3.1506, 1, 3.1401, 3.1611); }
  { case 301: return (0, 1, -100, 0); }
  { case 302: return (10, 1, 5, 13); }
  { case 303: return (0, 1, -1, 10); }
% Ar18g (Ar XVIII gamma)
  { case 304: return (2.9875, 1, 2.9775, 2.9974); }
  { case 305: return (0, 1, -100, 0); }
  { case 306: return (10, 1, 5, 12); }
  { case 307: return (0, 1, -1, 10); }
% Ar18d (Ar XVIII delta)
  { case 308: return (2.9176, 1, 2.9078, 2.9273); }
  { case 309: return (0, 1, -100, 0); }
  { case 310: return (10, 1, 5, 12); }
  { case 311: return (0, 1, -1, 10); }
% Ar17f (Ar XVII f)
  { case 312: return (3.9942, 1, 3.9808, 4.0075); }
  { case 313: return (0, 1, -100, 0); }
  { case 314: return (10, 1, 5, 16); }
  { case 315: return (0, 1, -1, 10); }
% Ar17i (Ar XVII i)
  { case 316: return (3.9694, 1, 3.9561, 3.9826); }
  { case 317: return (0, 1, -100, 0); }
  { case 318: return (10, 1, 5, 16); }
  { case 319: return (0, 1, -1, 10); }
% Ar17a (Ar XVII alpha)
  { case 320: return (3.9491, 1, 3.9359, 3.9622); }
  { case 321: return (0, 1, -100, 0); }
  { case 322: return (10, 1, 5, 16); }
  { case 323: return (0, 1, -1, 10); }
% Ar17b (Ar XVII beta)
  { case 324: return (3.3650, 1, 3.3538, 3.3762); }
  { case 325: return (0, 1, -100, 0); }
  { case 326: return (10, 1, 5, 13); }
  { case 327: return (0, 1, -1, 10); }
% Ar17g (Ar XVII gamma)
  { case 328: return (3.2000, 1, 3.1893, 3.2107); }
  { case 329: return (0, 1, -100, 0); }
  { case 330: return (10, 1, 5, 13); }
  { case 331: return (0, 1, -1, 10); }
% Ar17d (Ar XVII delta)
  { case 332: return (3.1280, 1, 3.1176, 3.1384); }
  { case 333: return (0, 1, -100, 0); }
  { case 334: return (10, 1, 5, 13); }
  { case 335: return (0, 1, -1, 10); }
% Ca20a (Ca XX alpha)
  { case 336: return (3.0203, 1, 3.0102, 3.0304); }
  { case 337: return (0, 1, -100, 0); }
  { case 338: return (10, 1, 5, 12); }
  { case 339: return (0, 1, -1, 10); }
% Ca20b (Ca XX beta)
  { case 340: return (2.5494, 1, 2.5409, 2.5579); }
  { case 341: return (0, 1, -100, 0); }
  { case 342: return (10, 1, 5, 10); }
  { case 343: return (0, 1, -1, 10); }
% Ca20g (Ca XX gamma)
  { case 344: return (2.4174, 1, 2.4093, 2.4255); }
  { case 345: return (0, 1, -100, 0); }
  { case 346: return (10, 1, 5, 10); }
  { case 347: return (0, 1, -1, 10); }
% Ca20d (Ca XX delta)
  { case 348: return (2.3609, 1, 2.3530, 2.3687); }
  { case 349: return (0, 1, -100, 0); }
  { case 350: return (9, 1, 5, 9); }
  { case 351: return (0, 1, -1, 10); }
% Ca19f (Ca XIX f)
  { case 352: return (3.2110, 1, 3.2003, 3.2217); }
  { case 353: return (0, 1, -100, 0); }
  { case 354: return (10, 1, 5, 13); }
  { case 355: return (0, 1, -1, 10); }
% Ca19i (Ca XIX i)
  { case 356: return (3.1927, 1, 3.1821, 3.2034); }
  { case 357: return (0, 1, -100, 0); }
  { case 358: return (10, 1, 5, 13); }
  { case 359: return (0, 1, -1, 10); }
% Ca19a (Ca XIX alpha)
  { case 360: return (3.1772, 1, 3.1666, 3.1877); }
  { case 361: return (0, 1, -100, 0); }
  { case 362: return (10, 1, 5, 13); }
  { case 363: return (0, 1, -1, 10); }
% Ca19b (Ca XIX beta)
  { case 364: return (2.7050, 1, 2.6960, 2.7140); }
  { case 365: return (0, 1, -100, 0); }
  { case 366: return (10, 1, 5, 11); }
  { case 367: return (0, 1, -1, 10); }
% Ca19g (Ca XIX gamma)
  { case 368: return (2.5710, 1, 2.5624, 2.5796); }
  { case 369: return (0, 1, -100, 0); }
  { case 370: return (10, 1, 5, 10); }
  { case 371: return (0, 1, -1, 10); }
% Ca19d (Ca XIX delta)
  { case 372: return (2.5140, 1, 2.5056, 2.5224); }
  { case 373: return (0, 1, -100, 0); }
  { case 374: return (10, 1, 5, 10); }
  { case 375: return (0, 1, -1, 10); }
% Fe26a (Fe XXVI alpha)
  { case 376: return (1.7799, 1, 1.7739, 1.7858); }
  { case 377: return (0, 1, -100, 0); }
  { case 378: return (7, 1, 5, 7); }
  { case 379: return (0, 1, -1, 10); }
% Fe26b (Fe XXVI beta)
  { case 380: return (1.5028, 1, 1.4977, 1.5078); }
  { case 381: return (0, 1, -100, 0); }
  { case 382: return (6, 1, 5, 6); }
  { case 383: return (0, 1, -1, 10); }
% Fe26g (Fe XXVI gamma)
  { case 384: return (1.4251, 1, 1.4203, 1.4298); }
  { case 385: return (0, 1, -100, 0); }
  { case 386: return (6, 1, 5, 6); }
  { case 387: return (0, 1, -1, 10); }
% Fe26d (Fe XXVI delta)
  { case 388: return (1.3918, 1, 1.3871, 1.3964); }
  { case 389: return (0, 1, -100, 0); }
  { case 390: return (6, 1, 5, 6); }
  { case 391: return (0, 1, -1, 10); }
% Fe25f (Fe XXV f)
  { case 392: return (1.8682, 1, 1.8620, 1.8744); }
  { case 393: return (0, 1, -100, 0); }
  { case 394: return (7, 1, 5, 7); }
  { case 395: return (0, 1, -1, 10); }
% Fe25i (Fe XXV i)
  { case 396: return (1.8595, 1, 1.8533, 1.8657); }
  { case 397: return (0, 1, -100, 0); }
  { case 398: return (7, 1, 5, 7); }
  { case 399: return (0, 1, -1, 10); }
% Fe25a (Fe XXV alpha)
  { case 400: return (1.8504, 1, 1.8442, 1.8566); }
  { case 401: return (0, 1, -100, 0); }
  { case 402: return (7, 1, 5, 7); }
  { case 403: return (0, 1, -1, 10); }
% Fe25b (Fe XXV beta)
  { case 404: return (1.5731, 1, 1.5679, 1.5783); }
  { case 405: return (0, 1, -100, 0); }
  { case 406: return (6, 1, 5, 6); }
  { case 407: return (0, 1, -1, 10); }
% Fe25g (Fe XXV gamma)
  { case 408: return (1.4950, 1, 1.4900, 1.5000); }
  { case 409: return (0, 1, -100, 0); }
  { case 410: return (6, 1, 5, 6); }
  { case 411: return (0, 1, -1, 10); }
% Fe25d (Fe XXV delta)
  { case 412: return (1.4610, 1, 1.4561, 1.4659); }
  { case 413: return (0, 1, -100, 0); }
  { case 414: return (6, 1, 5, 6); }
  { case 415: return (0, 1, -1, 10); }
% Fe1Ka (Fe I Kalpha)
  { case 416: return (1.9370, 1, 1.9305, 1.9435); }
  { case 417: return (0, 1, -100, 0); }
  { case 418: return (8, 1, 5, 8); }
  { case 419: return (0, 1, -1, 10); }
% Fe24_1062A (161758)
  { case 420: return (10.6190, 1, 10.5836, 10.6544); }
  { case 421: return (0, 1, -100, 0); }
  { case 422: return (10, 1, 5, 42); }
  { case 423: return (0, 1, -1, 10); }
% Fe24_679A (161756)
  { case 424: return (6.7887, 1, 6.7661, 6.8113); }
  { case 425: return (0, 1, -100, 0); }
  { case 426: return (10, 1, 5, 27); }
  { case 427: return (0, 1, -1, 10); }
% bl_1066A (161753, 18204) = (Fe XXIV, Fe XVII)
  { case 428: return (10.6600, 1, 10.6245, 10.6955); }
  { case 429: return (0, 1, -100, 0); }
  { case 430: return (10, 1, 5, 43); }
  { case 431: return (0, 1, -1, 10); }
% Fe24_799A (161759, 161754)
  { case 432: return (7.9908, 1, 7.9642, 8.0175); }
  { case 433: return (0, 1, -100, 0); }
  { case 434: return (10, 1, 5, 32); }
  { case 435: return (0, 1, -1, 10); }
% bl_1102A (149890, 18163) = (Fe XXIII, Fe XVII)
  { case 436: return (11.0225, 1, 10.9858, 11.0592); }
  { case 437: return (0, 1, -100, 0); }
  { case 438: return (10, 1, 5, 44); }
  { case 439: return (0, 1, -1, 10); }
% bl_1099A (149893, 131886) = (Fe XXIII, Fe XXII)
  { case 440: return (10.9872, 1, 10.9506, 11.0239); }
  { case 441: return (0, 1, -100, 0); }
  { case 442: return (10, 1, 5, 44); }
  { case 443: return (0, 1, -1, 10); }
% Fe23_830A (149914)
  { case 444: return (8.3038, 1, 8.2761, 8.3315); }
  { case 445: return (0, 1, -100, 0); }
  { case 446: return (10, 1, 5, 33); }
  { case 447: return (0, 1, -1, 10); }
% Fe22_1225A (131880)
  { case 448: return (12.2519, 1, 12.2111, 12.2927); }
  { case 449: return (0, 1, -100, 0); }
  { case 450: return (10, 1, 5, 49); }
  { case 451: return (0, 1, -1, 10); }
% Fe22_1143A (132636)
  { case 452: return (11.4270, 1, 11.3889, 11.4651); }
  { case 453: return (0, 1, -100, 0); }
  { case 454: return (10, 1, 5, 46); }
  { case 455: return (0, 1, -1, 10); }
% Fe22_897A (132642)
  { case 456: return (8.9748, 1, 8.9449, 9.0047); }
  { case 457: return (0, 1, -100, 0); }
  { case 458: return (10, 1, 5, 36); }
  { case 459: return (0, 1, -1, 10); }
% Fe22_1193A (132890)
  { case 460: return (11.9320, 1, 11.8922, 11.9718); }
  { case 461: return (0, 1, -100, 0); }
  { case 462: return (10, 1, 5, 48); }
  { case 463: return (0, 1, -1, 10); }
% bl_1177A (132626, 75499) = (Fe XXII, Fe XX)
  { case 464: return (11.7660, 1, 11.7268, 11.8052); }
  { case 465: return (0, 1, -100, 0); }
  { case 466: return (10, 1, 5, 47); }
  { case 467: return (0, 1, -1, 10); }
% Fe22_1149A (131888, 132634)
  { case 468: return (11.4900, 1, 11.4517, 11.5283); }
  { case 469: return (0, 1, -100, 0); }
  { case 470: return (10, 1, 5, 46); }
  { case 471: return (0, 1, -1, 10); }
% Fe22_873A (131902, 132644)
  { case 472: return (8.7307, 1, 8.7016, 8.7598); }
  { case 473: return (0, 1, -100, 0); }
  { case 474: return (10, 1, 5, 35); }
  { case 475: return (0, 1, -1, 10); }
% Fe22_786A (131918, 132676)
  { case 476: return (7.8650, 1, 7.8388, 7.8912); }
  { case 477: return (0, 1, -100, 0); }
  { case 478: return (10, 1, 5, 31); }
  { case 479: return (0, 1, -1, 10); }
% Fe21_1228A (128657)
  { case 480: return (12.2840, 1, 12.2431, 12.3249); }
  { case 481: return (0, 1, -100, 0); }
  { case 482: return (10, 1, 5, 49); }
  { case 483: return (0, 1, -1, 10); }
% Fe21_1198A (128681)
  { case 484: return (11.9750, 1, 11.9351, 12.0149); }
  { case 485: return (0, 1, -100, 0); }
  { case 486: return (10, 1, 5, 48); }
  { case 487: return (0, 1, -1, 10); }
% Fe21_1304A (128653)
  { case 488: return (13.0444, 1, 13.0009, 13.0879); }
  { case 489: return (0, 1, -100, 0); }
  { case 490: return (10, 1, 5, 52); }
  { case 491: return (0, 1, -1, 10); }
% bl_920A (128775, 75070, 74587) = (Fe XXI, Fe XX, Fe XX)
  { case 492: return (9.1967, 1, 9.1661, 9.2274); }
  { case 493: return (0, 1, -100, 0); }
  { case 494: return (10, 1, 5, 37); }
  { case 495: return (0, 1, -1, 10); }
% Fe21_857A (128847, 131702, 129682)
  { case 496: return (8.5740, 1, 8.5454, 8.6026); }
  { case 497: return (0, 1, -100, 0); }
  { case 498: return (10, 1, 5, 34); }
  { case 499: return (0, 1, -1, 10); }
% Fe21_1233A (131540)
  { case 500: return (12.3270, 1, 12.2859, 12.3681); }
  { case 501: return (0, 1, -100, 0); }
  { case 502: return (10, 1, 5, 49); }
  { case 503: return (0, 1, -1, 10); }
% Fe20_1292A (75469, 74908)
  { case 504: return (12.9165, 1, 12.8735, 12.9596); }
  { case 505: return (0, 1, -100, 0); }
  { case 506: return (10, 1, 5, 52); }
  { case 507: return (0, 1, -1, 10); }
% Fe20_1286A (74900, 75472)
  { case 508: return (12.8550, 1, 12.8121, 12.8978); }
  { case 509: return (0, 1, -100, 0); }
  { case 510: return (10, 1, 5, 51); }
  { case 511: return (0, 1, -1, 10); }
% Fe20_1283A (74354, 74360)
  { case 512: return (12.8255, 1, 12.7827, 12.8683); }
  { case 513: return (0, 1, -100, 0); }
  { case 514: return (10, 1, 5, 51); }
  { case 515: return (0, 1, -1, 10); }
% Fe20_1258A (74924, 75496)
  { case 516: return (12.5760, 1, 12.5341, 12.6179); }
  { case 517: return (0, 1, -100, 0); }
  { case 518: return (10, 1, 5, 50); }
  { case 519: return (0, 1, -1, 10); }
% Fe20_1000A (74457, 75036, 74463, 75579)
  { case 520: return (9.9992, 1, 9.9659, 10.0326); }
  { case 521: return (0, 1, -100, 0); }
  { case 522: return (10, 1, 5, 40); }
  { case 523: return (0, 1, -1, 10); }
% Fe20_1297A (75461, 74351)
  { case 524: return (12.9652, 1, 12.9220, 13.0084); }
  { case 525: return (0, 1, -100, 0); }
  { case 526: return (10, 1, 5, 52); }
  { case 527: return (0, 1, -1, 10); }
% bl_1012A (75587, 39003, 18189) = (Fe XX, Fe XIX, Fe XVII)
  { case 528: return (10.1203, 1, 10.0865, 10.1540); }
  { case 529: return (0, 1, -100, 0); }
  { case 530: return (10, 1, 5, 40); }
  { case 531: return (0, 1, -1, 10); }
% bl_1301A (74897, 39144) = (Fe XX, Fe XIX)
  { case 532: return (13.0070, 1, 12.9636, 13.0504); }
  { case 533: return (0, 1, -100, 0); }
  { case 534: return (10, 1, 5, 52); }
  { case 535: return (0, 1, -1, 10); }
% Fe19_1466A (39118)
  { case 536: return (14.6640, 1, 14.6151, 14.7129); }
  { case 537: return (0, 1, -100, 0); }
  { case 538: return (10, 1, 5, 59); }
  { case 539: return (0, 1, -1, 10); }
% Fe19_1380A (39132)
  { case 540: return (13.7950, 1, 13.7490, 13.8410); }
  { case 541: return (0, 1, -100, 0); }
  { case 542: return (10, 1, 5, 55); }
  { case 543: return (0, 1, -1, 10); }
% Fe19_1364A (39124)
  { case 544: return (13.6450, 1, 13.5995, 13.6905); }
  { case 545: return (0, 1, -100, 0); }
  { case 546: return (10, 1, 5, 55); }
  { case 547: return (0, 1, -1, 10); }
% Fe19_1350A (38915)
  { case 548: return (13.4970, 1, 13.4520, 13.5420); }
  { case 549: return (0, 1, -100, 0); }
  { case 550: return (10, 1, 5, 54); }
  { case 551: return (0, 1, -1, 10); }
% Fe19_1346A (38603)
  { case 552: return (13.4620, 1, 13.4171, 13.5069); }
  { case 553: return (0, 1, -100, 0); }
  { case 554: return (10, 1, 5, 54); }
  { case 555: return (0, 1, -1, 10); }
% Fe19_1342A (39134)
  { case 556: return (13.4230, 1, 13.3783, 13.4677); }
  { case 557: return (0, 1, -100, 0); }
  { case 558: return (10, 1, 5, 54); }
  { case 559: return (0, 1, -1, 10); }
% Fe19_1294A (38634)
  { case 560: return (12.9450, 1, 12.9018, 12.9881); }
  { case 561: return (0, 1, -100, 0); }
  { case 562: return (10, 1, 5, 52); }
  { case 563: return (0, 1, -1, 10); }
% Fe19_1082A (39172)
  { case 564: return (10.8160, 1, 10.7799, 10.8521); }
  { case 565: return (0, 1, -100, 0); }
  { case 566: return (10, 1, 5, 43); }
  { case 567: return (0, 1, -1, 10); }
% Fe19_1013A (38692)
  { case 568: return (10.1309, 1, 10.0971, 10.1647); }
  { case 569: return (0, 1, -100, 0); }
  { case 570: return (10, 1, 5, 41); }
  { case 571: return (0, 1, -1, 10); }
% Fe19_986A (39223)
  { case 572: return (9.8552, 1, 9.8224, 9.8881); }
  { case 573: return (0, 1, -100, 0); }
  { case 574: return (10, 1, 5, 39); }
  { case 575: return (0, 1, -1, 10); }
% Fe19_1352A (39128, 38611)
  { case 576: return (13.5163, 1, 13.4712, 13.5614); }
  { case 577: return (0, 1, -100, 0); }
  { case 578: return (10, 1, 5, 54); }
  { case 579: return (0, 1, -1, 10); }
% Fe19_1293A (38941, 39146)
  { case 580: return (12.9320, 1, 12.8889, 12.9752); }
  { case 581: return (0, 1, -100, 0); }
  { case 582: return (10, 1, 5, 52); }
  { case 583: return (0, 1, -1, 10); }
% Fe18_1453A (38000)
  { case 584: return (14.5340, 1, 14.4856, 14.5824); }
  { case 585: return (0, 1, -100, 0); }
  { case 586: return (10, 1, 5, 58); }
  { case 587: return (0, 1, -1, 10); }
% Fe18_1437A (37997)
  { case 588: return (14.3730, 1, 14.3251, 14.4209); }
  { case 589: return (0, 1, -100, 0); }
  { case 590: return (10, 1, 5, 57); }
  { case 591: return (0, 1, -1, 10); }
% Fe18_1457A (37942)
  { case 592: return (14.5710, 1, 14.5224, 14.6196); }
  { case 593: return (0, 1, -100, 0); }
  { case 594: return (10, 1, 5, 58); }
  { case 595: return (0, 1, -1, 10); }
% Fe18_1562A (37995)
  { case 596: return (15.6250, 1, 15.5729, 15.6771); }
  { case 597: return (0, 1, -100, 0); }
  { case 598: return (10, 1, 5, 62); }
  { case 599: return (0, 1, -1, 10); }
% Fe18_1332A (37835)
  { case 600: return (13.3230, 1, 13.2786, 13.3674); }
  { case 601: return (0, 1, -100, 0); }
  { case 602: return (10, 1, 5, 53); }
  { case 603: return (0, 1, -1, 10); }
% Fe18_1426A (37830, 38002)
  { case 604: return (14.2560, 1, 14.2085, 14.3035); }
  { case 605: return (0, 1, -100, 0); }
  { case 606: return (10, 1, 5, 57); }
  { case 607: return (0, 1, -1, 10); }
% Fe18_1421A (37944, 37998, 38112)
  { case 608: return (14.2056, 1, 14.1582, 14.2529); }
  { case 609: return (0, 1, -100, 0); }
  { case 610: return (10, 1, 5, 57); }
  { case 611: return (0, 1, -1, 10); }
% Fe18_1153A (37961, 38014)
  { case 612: return (11.5270, 1, 11.4886, 11.5654); }
  { case 613: return (0, 1, -100, 0); }
  { case 614: return (10, 1, 5, 46); }
  { case 615: return (0, 1, -1, 10); }
% Fe18_1133A (37845, 37963, 38012)
  { case 616: return (11.3260, 1, 11.2882, 11.3638); }
  { case 617: return (0, 1, -100, 0); }
  { case 618: return (10, 1, 5, 45); }
  { case 619: return (0, 1, -1, 10); }
% Fe17_1678A (20125)
  { case 620: return (16.7800, 1, 16.7241, 16.8359); }
  { case 621: return (0, 1, -100, 0); }
  { case 622: return (10, 1, 5, 67); }
  { case 623: return (0, 1, -1, 10); }
% Fe17_1526A (20128)
  { case 624: return (15.2610, 1, 15.2101, 15.3119); }
  { case 625: return (0, 1, -100, 0); }
  { case 626: return (10, 1, 5, 61); }
  { case 627: return (0, 1, -1, 10); }
% Fe17_1501A (20127)
  { case 628: return (15.0140, 1, 14.9640, 15.0640); }
  { case 629: return (0, 1, -100, 0); }
  { case 630: return (10, 1, 5, 60); }
  { case 631: return (0, 1, -1, 10); }
% Fe17_1227A (18155)
  { case 632: return (12.2660, 1, 12.2251, 12.3069); }
  { case 633: return (0, 1, -100, 0); }
  { case 634: return (10, 1, 5, 49); }
  { case 635: return (0, 1, -1, 10); }
% Fe17_1212A (18151)
  { case 636: return (12.1240, 1, 12.0836, 12.1644); }
  { case 637: return (0, 1, -100, 0); }
  { case 638: return (10, 1, 5, 48); }
  { case 639: return (0, 1, -1, 10); }
% Fe17_1536A (25066)
  { case 640: return (15.3597, 1, 15.3085, 15.4109); }
  { case 641: return (0, 1, -100, 0); }
  { case 642: return (10, 1, 5, 61); }
  { case 643: return (0, 1, -1, 10); }
% Fe17_1382A (20130)
  { case 644: return (13.8250, 1, 13.7789, 13.8711); }
  { case 645: return (0, 1, -100, 0); }
  { case 646: return (10, 1, 5, 55); }
  { case 647: return (0, 1, -1, 10); }
% bl_1384A (38918, 74885) = (Fe XIX, Fe XX)
  { case 648: return (13.8410, 1, 13.7949, 13.8871); }
  { case 649: return (0, 1, -100, 0); }
  { case 650: return (10, 1, 5, 55); }
  { case 651: return (0, 1, -1, 10); }
% Al11Ka (Al XI Kalpha)
  { case 652: return (7.8850, 1, 7.8587, 7.9113); }
  { case 653: return (0, 1, -100, 0); }
  { case 654: return (10, 1, 5, 32); }
  { case 655: return (0, 1, -1, 10); }
% Al10Ka (Al X Kalpha)
  { case 656: return (7.9640, 1, 7.9375, 7.9905); }
  { case 657: return (0, 1, -100, 0); }
  { case 658: return (10, 1, 5, 32); }
  { case 659: return (0, 1, -1, 10); }
% Al9Ka (Al IX Kalpha)
  { case 660: return (8.0500, 1, 8.0232, 8.0768); }
  { case 661: return (0, 1, -100, 0); }
  { case 662: return (10, 1, 5, 32); }
  { case 663: return (0, 1, -1, 10); }
% Al8Ka (Al VIII Kalpha)
  { case 664: return (8.1290, 1, 8.1019, 8.1561); }
  { case 665: return (0, 1, -100, 0); }
  { case 666: return (10, 1, 5, 33); }
  { case 667: return (0, 1, -1, 10); }
% Al7Ka (Al VII Kalpha)
  { case 668: return (8.2030, 1, 8.1757, 8.2303); }
  { case 669: return (0, 1, -100, 0); }
  { case 670: return (10, 1, 5, 33); }
  { case 671: return (0, 1, -1, 10); }
% Al6Ka (Al VI Kalpha)
  { case 672: return (8.2690, 1, 8.2414, 8.2966); }
  { case 673: return (0, 1, -100, 0); }
  { case 674: return (10, 1, 5, 33); }
  { case 675: return (0, 1, -1, 10); }
% Al5Ka (Al V Kalpha)
  { case 676: return (8.3280, 1, 8.3002, 8.3558); }
  { case 677: return (0, 1, -100, 0); }
  { case 678: return (10, 1, 5, 33); }
  { case 679: return (0, 1, -1, 10); }
% Al4Ka (Al IV Kalpha)
  { case 680: return (8.3320, 1, 8.3042, 8.3598); }
  { case 681: return (0, 1, -100, 0); }
  { case 682: return (10, 1, 5, 33); }
  { case 683: return (0, 1, -1, 10); }
% Al3Ka (Al III Kalpha)
  { case 684: return (8.3360, 1, 8.3082, 8.3638); }
  { case 685: return (0, 1, -100, 0); }
  { case 686: return (10, 1, 5, 33); }
  { case 687: return (0, 1, -1, 10); }
% Al2Ka (Al II Kalpha)
  { case 688: return (8.3390, 1, 8.3112, 8.3668); }
  { case 689: return (0, 1, -100, 0); }
  { case 690: return (10, 1, 5, 33); }
  { case 691: return (0, 1, -1, 10); }
% Si12Ka (Si XII Kalpha)
  { case 692: return (6.7500, 1, 6.7275, 6.7725); }
  { case 693: return (0, 1, -100, 0); }
  { case 694: return (10, 1, 5, 27); }
  { case 695: return (0, 1, -1, 10); }
% Si11Ka (Si XI Kalpha)
  { case 696: return (6.8130, 1, 6.7903, 6.8357); }
  { case 697: return (0, 1, -100, 0); }
  { case 698: return (10, 1, 5, 27); }
  { case 699: return (0, 1, -1, 10); }
% Si10Ka (Si X Kalpha)
  { case 700: return (6.8820, 1, 6.8591, 6.9049); }
  { case 701: return (0, 1, -100, 0); }
  { case 702: return (10, 1, 5, 28); }
  { case 703: return (0, 1, -1, 10); }
% Si9Ka (Si IX Kalpha)
  { case 704: return (6.9470, 1, 6.9238, 6.9702); }
  { case 705: return (0, 1, -100, 0); }
  { case 706: return (10, 1, 5, 28); }
  { case 707: return (0, 1, -1, 10); }
% Si8Ka (Si VIII Kalpha)
  { case 708: return (7.0070, 1, 6.9836, 7.0304); }
  { case 709: return (0, 1, -100, 0); }
  { case 710: return (10, 1, 5, 28); }
  { case 711: return (0, 1, -1, 10); }
% Si7Ka (Si VII Kalpha)
  { case 712: return (7.0630, 1, 7.0395, 7.0865); }
  { case 713: return (0, 1, -100, 0); }
  { case 714: return (10, 1, 5, 28); }
  { case 715: return (0, 1, -1, 10); }
% Si6Ka (Si VI Kalpha)
  { case 716: return (7.1120, 1, 7.0883, 7.1357); }
  { case 717: return (0, 1, -100, 0); }
  { case 718: return (10, 1, 5, 28); }
  { case 719: return (0, 1, -1, 10); }
% Si5Ka (Si V Kalpha)
  { case 720: return (7.1170, 1, 7.0933, 7.1407); }
  { case 721: return (0, 1, -100, 0); }
  { case 722: return (10, 1, 5, 28); }
  { case 723: return (0, 1, -1, 10); }
% Si4Ka (Si IV Kalpha)
  { case 724: return (7.1210, 1, 7.0973, 7.1447); }
  { case 725: return (0, 1, -100, 0); }
  { case 726: return (10, 1, 5, 28); }
  { case 727: return (0, 1, -1, 10); }
% Si3Ka (Si III Kalpha)
  { case 728: return (7.1240, 1, 7.1003, 7.1477); }
  { case 729: return (0, 1, -100, 0); }
  { case 730: return (10, 1, 5, 28); }
  { case 731: return (0, 1, -1, 10); }
% Si2Ka (Si II Kalpha)
  { case 732: return (7.1260, 1, 7.1022, 7.1498); }
  { case 733: return (0, 1, -100, 0); }
  { case 734: return (10, 1, 5, 29); }
  { case 735: return (0, 1, -1, 10); }
% Ar16Ka (Ar XVI Kalpha)
  { case 736: return (3.9950, 1, 3.9817, 4.0083); }
  { case 737: return (0, 1, -100, 0); }
  { case 738: return (10, 1, 5, 16); }
  { case 739: return (0, 1, -1, 10); }
% Ar15Ka (Ar XV Kalpha)
  { case 740: return (4.0250, 1, 4.0116, 4.0384); }
  { case 741: return (0, 1, -100, 0); }
  { case 742: return (10, 1, 5, 16); }
  { case 743: return (0, 1, -1, 10); }
% Ar14Ka (Ar XIV Kalpha)
  { case 744: return (4.0570, 1, 4.0435, 4.0705); }
  { case 745: return (0, 1, -100, 0); }
  { case 746: return (10, 1, 5, 16); }
  { case 747: return (0, 1, -1, 10); }
% Ar13Ka (Ar XIII Kalpha)
  { case 748: return (4.0890, 1, 4.0754, 4.1026); }
  { case 749: return (0, 1, -100, 0); }
  { case 750: return (10, 1, 5, 16); }
  { case 751: return (0, 1, -1, 10); }
% Ar12Ka (Ar XII Kalpha)
  { case 752: return (4.1190, 1, 4.1053, 4.1327); }
  { case 753: return (0, 1, -100, 0); }
  { case 754: return (10, 1, 5, 16); }
  { case 755: return (0, 1, -1, 10); }
% Ar11Ka (Ar XI Kalpha)
  { case 756: return (4.1470, 1, 4.1332, 4.1608); }
  { case 757: return (0, 1, -100, 0); }
  { case 758: return (10, 1, 5, 17); }
  { case 759: return (0, 1, -1, 10); }
% Ar10Ka (Ar X Kalpha)
  { case 760: return (4.1740, 1, 4.1601, 4.1879); }
  { case 761: return (0, 1, -100, 0); }
  { case 762: return (10, 1, 5, 17); }
  { case 763: return (0, 1, -1, 10); }
% Ar9Ka (Ar IX Kalpha)
  { case 764: return (4.1780, 1, 4.1641, 4.1919); }
  { case 765: return (0, 1, -100, 0); }
  { case 766: return (10, 1, 5, 17); }
  { case 767: return (0, 1, -1, 10); }
% Ar8Ka (Ar VIII Kalpha)
  { case 768: return (4.1800, 1, 4.1661, 4.1939); }
  { case 769: return (0, 1, -100, 0); }
  { case 770: return (10, 1, 5, 17); }
  { case 771: return (0, 1, -1, 10); }
% Ar7Ka (Ar VII Kalpha)
  { case 772: return (4.1840, 1, 4.1701, 4.1979); }
  { case 773: return (0, 1, -100, 0); }
  { case 774: return (10, 1, 5, 17); }
  { case 775: return (0, 1, -1, 10); }
% Ar6Ka (Ar VI Kalpha)
  { case 776: return (4.1860, 1, 4.1720, 4.2000); }
  { case 777: return (0, 1, -100, 0); }
  { case 778: return (10, 1, 5, 17); }
  { case 779: return (0, 1, -1, 10); }
% Ar5Ka (Ar V Kalpha)
  { case 780: return (4.1890, 1, 4.1750, 4.2030); }
  { case 781: return (0, 1, -100, 0); }
  { case 782: return (10, 1, 5, 17); }
  { case 783: return (0, 1, -1, 10); }
% Ar4Ka (Ar IV Kalpha)
  { case 784: return (4.1900, 1, 4.1760, 4.2040); }
  { case 785: return (0, 1, -100, 0); }
  { case 786: return (10, 1, 5, 17); }
  { case 787: return (0, 1, -1, 10); }
% Ar3Ka (Ar III Kalpha)
  { case 788: return (4.1920, 1, 4.1780, 4.2060); }
  { case 789: return (0, 1, -100, 0); }
  { case 790: return (10, 1, 5, 17); }
  { case 791: return (0, 1, -1, 10); }
% Ar2Ka (Ar II Kalpha)
  { case 792: return (4.1930, 1, 4.1790, 4.2070); }
  { case 793: return (0, 1, -100, 0); }
  { case 794: return (10, 1, 5, 17); }
  { case 795: return (0, 1, -1, 10); }
}

add_slang_function("lines",
[
         "O8a_lam [A (18.9689 A)]",         "O8a_EW [mA]",         "O8a_FWHM [mA]",         "O8a_A",  %  @ 18.9689 A (O VIII alpha)
         "O8b_lam [A (16.0059 A)]",         "O8b_EW [mA]",         "O8b_FWHM [mA]",         "O8b_A",  %  @ 16.0059 A (O VIII beta)
         "O8g_lam [A (15.1762 A)]",         "O8g_EW [mA]",         "O8g_FWHM [mA]",         "O8g_A",  %  @ 15.1762 A (O VIII gamma)
         "O8d_lam [A (14.8206 A)]",         "O8d_EW [mA]",         "O8d_FWHM [mA]",         "O8d_A",  %  @ 14.8206 A (O VIII delta)
         "O7f_lam [A (22.0977 A)]",         "O7f_EW [mA]",         "O7f_FWHM [mA]",         "O7f_A",  %  @ 22.0977 A (O VII f)
         "O7i_lam [A (21.8036 A)]",         "O7i_EW [mA]",         "O7i_FWHM [mA]",         "O7i_A",  %  @ 21.8036 A (O VII i)
         "O7a_lam [A (21.6015 A)]",         "O7a_EW [mA]",         "O7a_FWHM [mA]",         "O7a_A",  %  @ 21.6015 A (O VII alpha)
         "O7b_lam [A (18.6270 A)]",         "O7b_EW [mA]",         "O7b_FWHM [mA]",         "O7b_A",  %  @ 18.6270 A (O VII beta)
         "O7g_lam [A (17.7680 A)]",         "O7g_EW [mA]",         "O7g_FWHM [mA]",         "O7g_A",  %  @ 17.7680 A (O VII gamma)
         "O7d_lam [A (17.3960 A)]",         "O7d_EW [mA]",         "O7d_FWHM [mA]",         "O7d_A",  %  @ 17.3960 A (O VII delta)
       "Ne10a_lam [A (12.1339 A)]",       "Ne10a_EW [mA]",       "Ne10a_FWHM [mA]",       "Ne10a_A",  %  @ 12.1339 A (Ne X alpha)
       "Ne10b_lam [A (10.2389 A)]",       "Ne10b_EW [mA]",       "Ne10b_FWHM [mA]",       "Ne10b_A",  %  @ 10.2389 A (Ne X beta)
       "Ne10g_lam [A ( 9.7082 A)]",       "Ne10g_EW [mA]",       "Ne10g_FWHM [mA]",       "Ne10g_A",  %  @  9.7082 A (Ne X gamma)
       "Ne10d_lam [A ( 9.4807 A)]",       "Ne10d_EW [mA]",       "Ne10d_FWHM [mA]",       "Ne10d_A",  %  @  9.4807 A (Ne X delta)
       "Ne10e_lam [A ( 9.3616 A)]",       "Ne10e_EW [mA]",       "Ne10e_FWHM [mA]",       "Ne10e_A",  %  @  9.3616 A (Ne X epsilon)
       "Ne10z_lam [A ( 9.2912 A)]",       "Ne10z_EW [mA]",       "Ne10z_FWHM [mA]",       "Ne10z_A",  %  @  9.2912 A (Ne X zeta)
        "Ne9f_lam [A (13.6990 A)]",        "Ne9f_EW [mA]",        "Ne9f_FWHM [mA]",        "Ne9f_A",  %  @ 13.6990 A (Ne IX f)
        "Ne9i_lam [A (13.5531 A)]",        "Ne9i_EW [mA]",        "Ne9i_FWHM [mA]",        "Ne9i_A",  %  @ 13.5531 A (Ne IX i)
        "Ne9a_lam [A (13.4473 A)]",        "Ne9a_EW [mA]",        "Ne9a_FWHM [mA]",        "Ne9a_A",  %  @ 13.4473 A (Ne IX alpha)
        "Ne9b_lam [A (11.5440 A)]",        "Ne9b_EW [mA]",        "Ne9b_FWHM [mA]",        "Ne9b_A",  %  @ 11.5440 A (Ne IX beta)
        "Ne9g_lam [A (11.0010 A)]",        "Ne9g_EW [mA]",        "Ne9g_FWHM [mA]",        "Ne9g_A",  %  @ 11.0010 A (Ne IX gamma)
        "Ne9d_lam [A (10.7650 A)]",        "Ne9d_EW [mA]",        "Ne9d_FWHM [mA]",        "Ne9d_A",  %  @ 10.7650 A (Ne IX delta)
        "Ne9e_lam [A (10.6426 A)]",        "Ne9e_EW [mA]",        "Ne9e_FWHM [mA]",        "Ne9e_A",  %  @ 10.6426 A (Ne IX epsilon)
        "Ne9z_lam [A (10.5650 A)]",        "Ne9z_EW [mA]",        "Ne9z_FWHM [mA]",        "Ne9z_A",  %  @ 10.5650 A (Ne IX zeta)
       "Na11a_lam [A (10.0250 A)]",       "Na11a_EW [mA]",       "Na11a_FWHM [mA]",       "Na11a_A",  %  @ 10.0250 A (Na XI alpha)
       "Na11b_lam [A ( 8.4595 A)]",       "Na11b_EW [mA]",       "Na11b_FWHM [mA]",       "Na11b_A",  %  @  8.4595 A (Na XI beta)
       "Na11g_lam [A ( 8.0211 A)]",       "Na11g_EW [mA]",       "Na11g_FWHM [mA]",       "Na11g_A",  %  @  8.0211 A (Na XI gamma)
       "Na11d_lam [A ( 7.8332 A)]",       "Na11d_EW [mA]",       "Na11d_FWHM [mA]",       "Na11d_A",  %  @  7.8332 A (Na XI delta)
       "Na10f_lam [A (11.1900 A)]",       "Na10f_EW [mA]",       "Na10f_FWHM [mA]",       "Na10f_A",  %  @ 11.1900 A (Na X f)
       "Na10i_lam [A (11.0800 A)]",       "Na10i_EW [mA]",       "Na10i_FWHM [mA]",       "Na10i_A",  %  @ 11.0800 A (Na X i)
       "Na10a_lam [A (11.0027 A)]",       "Na10a_EW [mA]",       "Na10a_FWHM [mA]",       "Na10a_A",  %  @ 11.0027 A (Na X alpha)
       "Na10b_lam [A ( 9.4330 A)]",       "Na10b_EW [mA]",       "Na10b_FWHM [mA]",       "Na10b_A",  %  @  9.4330 A (Na X beta)
       "Na10g_lam [A ( 8.9828 A)]",       "Na10g_EW [mA]",       "Na10g_FWHM [mA]",       "Na10g_A",  %  @  8.9828 A (Na X gamma)
       "Na10d_lam [A ( 8.7884 A)]",       "Na10d_EW [mA]",       "Na10d_FWHM [mA]",       "Na10d_A",  %  @  8.7884 A (Na X delta)
       "Mg12a_lam [A ( 8.4210 A)]",       "Mg12a_EW [mA]",       "Mg12a_FWHM [mA]",       "Mg12a_A",  %  @  8.4210 A (Mg XII alpha)
       "Mg12b_lam [A ( 7.1062 A)]",       "Mg12b_EW [mA]",       "Mg12b_FWHM [mA]",       "Mg12b_A",  %  @  7.1062 A (Mg XII beta)
       "Mg12g_lam [A ( 6.7379 A)]",       "Mg12g_EW [mA]",       "Mg12g_FWHM [mA]",       "Mg12g_A",  %  @  6.7379 A (Mg XII gamma)
       "Mg12d_lam [A ( 6.5801 A)]",       "Mg12d_EW [mA]",       "Mg12d_FWHM [mA]",       "Mg12d_A",  %  @  6.5801 A (Mg XII delta)
       "Mg11f_lam [A ( 9.3143 A)]",       "Mg11f_EW [mA]",       "Mg11f_FWHM [mA]",       "Mg11f_A",  %  @  9.3143 A (Mg XI f)
       "Mg11i_lam [A ( 9.2312 A)]",       "Mg11i_EW [mA]",       "Mg11i_FWHM [mA]",       "Mg11i_A",  %  @  9.2312 A (Mg XI i)
       "Mg11a_lam [A ( 9.1687 A)]",       "Mg11a_EW [mA]",       "Mg11a_FWHM [mA]",       "Mg11a_A",  %  @  9.1687 A (Mg XI alpha)
       "Mg11b_lam [A ( 7.8503 A)]",       "Mg11b_EW [mA]",       "Mg11b_FWHM [mA]",       "Mg11b_A",  %  @  7.8503 A (Mg XI beta)
       "Mg11g_lam [A ( 7.4730 A)]",       "Mg11g_EW [mA]",       "Mg11g_FWHM [mA]",       "Mg11g_A",  %  @  7.4730 A (Mg XI gamma)
       "Mg11d_lam [A ( 7.3101 A)]",       "Mg11d_EW [mA]",       "Mg11d_FWHM [mA]",       "Mg11d_A",  %  @  7.3101 A (Mg XI delta)
       "Al13a_lam [A ( 7.1728 A)]",       "Al13a_EW [mA]",       "Al13a_FWHM [mA]",       "Al13a_A",  %  @  7.1728 A (Al XIII alpha)
       "Al13b_lam [A ( 6.0530 A)]",       "Al13b_EW [mA]",       "Al13b_FWHM [mA]",       "Al13b_A",  %  @  6.0530 A (Al XIII beta)
       "Al13g_lam [A ( 5.7393 A)]",       "Al13g_EW [mA]",       "Al13g_FWHM [mA]",       "Al13g_A",  %  @  5.7393 A (Al XIII gamma)
       "Al13d_lam [A ( 5.6049 A)]",       "Al13d_EW [mA]",       "Al13d_FWHM [mA]",       "Al13d_A",  %  @  5.6049 A (Al XIII delta)
       "Al12f_lam [A ( 7.8721 A)]",       "Al12f_EW [mA]",       "Al12f_FWHM [mA]",       "Al12f_A",  %  @  7.8721 A (Al XII f)
       "Al12i_lam [A ( 7.8070 A)]",       "Al12i_EW [mA]",       "Al12i_FWHM [mA]",       "Al12i_A",  %  @  7.8070 A (Al XII i)
       "Al12a_lam [A ( 7.7573 A)]",       "Al12a_EW [mA]",       "Al12a_FWHM [mA]",       "Al12a_A",  %  @  7.7573 A (Al XII alpha)
       "Al12b_lam [A ( 6.6350 A)]",       "Al12b_EW [mA]",       "Al12b_FWHM [mA]",       "Al12b_A",  %  @  6.6350 A (Al XII beta)
       "Al12g_lam [A ( 6.3140 A)]",       "Al12g_EW [mA]",       "Al12g_FWHM [mA]",       "Al12g_A",  %  @  6.3140 A (Al XII gamma)
       "Al12d_lam [A ( 6.1750 A)]",       "Al12d_EW [mA]",       "Al12d_FWHM [mA]",       "Al12d_A",  %  @  6.1750 A (Al XII delta)
       "Si14a_lam [A ( 6.1822 A)]",       "Si14a_EW [mA]",       "Si14a_FWHM [mA]",       "Si14a_A",  %  @  6.1822 A (Si XIV alpha)
       "Si14b_lam [A ( 5.2172 A)]",       "Si14b_EW [mA]",       "Si14b_FWHM [mA]",       "Si14b_A",  %  @  5.2172 A (Si XIV beta)
       "Si14g_lam [A ( 4.9469 A)]",       "Si14g_EW [mA]",       "Si14g_FWHM [mA]",       "Si14g_A",  %  @  4.9469 A (Si XIV gamma)
       "Si14d_lam [A ( 4.8311 A)]",       "Si14d_EW [mA]",       "Si14d_FWHM [mA]",       "Si14d_A",  %  @  4.8311 A (Si XIV delta)
       "Si13f_lam [A ( 6.7403 A)]",       "Si13f_EW [mA]",       "Si13f_FWHM [mA]",       "Si13f_A",  %  @  6.7403 A (Si XIII f)
       "Si13i_lam [A ( 6.6882 A)]",       "Si13i_EW [mA]",       "Si13i_FWHM [mA]",       "Si13i_A",  %  @  6.6882 A (Si XIII i)
       "Si13a_lam [A ( 6.6479 A)]",       "Si13a_EW [mA]",       "Si13a_FWHM [mA]",       "Si13a_A",  %  @  6.6479 A (Si XIII alpha)
       "Si13b_lam [A ( 5.6805 A)]",       "Si13b_EW [mA]",       "Si13b_FWHM [mA]",       "Si13b_A",  %  @  5.6805 A (Si XIII beta)
       "Si13g_lam [A ( 5.4045 A)]",       "Si13g_EW [mA]",       "Si13g_FWHM [mA]",       "Si13g_A",  %  @  5.4045 A (Si XIII gamma)
       "Si13d_lam [A ( 5.2850 A)]",       "Si13d_EW [mA]",       "Si13d_FWHM [mA]",       "Si13d_A",  %  @  5.2850 A (Si XIII delta)
        "S16a_lam [A ( 4.7292 A)]",        "S16a_EW [mA]",        "S16a_FWHM [mA]",        "S16a_A",  %  @  4.7292 A (S XVI alpha)
        "S16b_lam [A ( 3.9912 A)]",        "S16b_EW [mA]",        "S16b_FWHM [mA]",        "S16b_A",  %  @  3.9912 A (S XVI beta)
        "S16g_lam [A ( 3.7845 A)]",        "S16g_EW [mA]",        "S16g_FWHM [mA]",        "S16g_A",  %  @  3.7845 A (S XVI gamma)
        "S16d_lam [A ( 3.6959 A)]",        "S16d_EW [mA]",        "S16d_FWHM [mA]",        "S16d_A",  %  @  3.6959 A (S XVI delta)
        "S15f_lam [A ( 5.1015 A)]",        "S15f_EW [mA]",        "S15f_FWHM [mA]",        "S15f_A",  %  @  5.1015 A (S XV f)
        "S15i_lam [A ( 5.0665 A)]",        "S15i_EW [mA]",        "S15i_FWHM [mA]",        "S15i_A",  %  @  5.0665 A (S XV i)
        "S15a_lam [A ( 5.0387 A)]",        "S15a_EW [mA]",        "S15a_FWHM [mA]",        "S15a_A",  %  @  5.0387 A (S XV alpha)
        "S15b_lam [A ( 4.2990 A)]",        "S15b_EW [mA]",        "S15b_FWHM [mA]",        "S15b_A",  %  @  4.2990 A (S XV beta)
        "S15g_lam [A ( 4.0883 A)]",        "S15g_EW [mA]",        "S15g_FWHM [mA]",        "S15g_A",  %  @  4.0883 A (S XV gamma)
        "S15d_lam [A ( 3.9980 A)]",        "S15d_EW [mA]",        "S15d_FWHM [mA]",        "S15d_A",  %  @  3.9980 A (S XV delta)
       "Ar18a_lam [A ( 3.7329 A)]",       "Ar18a_EW [mA]",       "Ar18a_FWHM [mA]",       "Ar18a_A",  %  @  3.7329 A (Ar XVIII alpha)
       "Ar18b_lam [A ( 3.1506 A)]",       "Ar18b_EW [mA]",       "Ar18b_FWHM [mA]",       "Ar18b_A",  %  @  3.1506 A (Ar XVIII beta)
       "Ar18g_lam [A ( 2.9875 A)]",       "Ar18g_EW [mA]",       "Ar18g_FWHM [mA]",       "Ar18g_A",  %  @  2.9875 A (Ar XVIII gamma)
       "Ar18d_lam [A ( 2.9176 A)]",       "Ar18d_EW [mA]",       "Ar18d_FWHM [mA]",       "Ar18d_A",  %  @  2.9176 A (Ar XVIII delta)
       "Ar17f_lam [A ( 3.9942 A)]",       "Ar17f_EW [mA]",       "Ar17f_FWHM [mA]",       "Ar17f_A",  %  @  3.9942 A (Ar XVII f)
       "Ar17i_lam [A ( 3.9694 A)]",       "Ar17i_EW [mA]",       "Ar17i_FWHM [mA]",       "Ar17i_A",  %  @  3.9694 A (Ar XVII i)
       "Ar17a_lam [A ( 3.9491 A)]",       "Ar17a_EW [mA]",       "Ar17a_FWHM [mA]",       "Ar17a_A",  %  @  3.9491 A (Ar XVII alpha)
       "Ar17b_lam [A ( 3.3650 A)]",       "Ar17b_EW [mA]",       "Ar17b_FWHM [mA]",       "Ar17b_A",  %  @  3.3650 A (Ar XVII beta)
       "Ar17g_lam [A ( 3.2000 A)]",       "Ar17g_EW [mA]",       "Ar17g_FWHM [mA]",       "Ar17g_A",  %  @  3.2000 A (Ar XVII gamma)
       "Ar17d_lam [A ( 3.1280 A)]",       "Ar17d_EW [mA]",       "Ar17d_FWHM [mA]",       "Ar17d_A",  %  @  3.1280 A (Ar XVII delta)
       "Ca20a_lam [A ( 3.0203 A)]",       "Ca20a_EW [mA]",       "Ca20a_FWHM [mA]",       "Ca20a_A",  %  @  3.0203 A (Ca XX alpha)
       "Ca20b_lam [A ( 2.5494 A)]",       "Ca20b_EW [mA]",       "Ca20b_FWHM [mA]",       "Ca20b_A",  %  @  2.5494 A (Ca XX beta)
       "Ca20g_lam [A ( 2.4174 A)]",       "Ca20g_EW [mA]",       "Ca20g_FWHM [mA]",       "Ca20g_A",  %  @  2.4174 A (Ca XX gamma)
       "Ca20d_lam [A ( 2.3609 A)]",       "Ca20d_EW [mA]",       "Ca20d_FWHM [mA]",       "Ca20d_A",  %  @  2.3609 A (Ca XX delta)
       "Ca19f_lam [A ( 3.2110 A)]",       "Ca19f_EW [mA]",       "Ca19f_FWHM [mA]",       "Ca19f_A",  %  @  3.2110 A (Ca XIX f)
       "Ca19i_lam [A ( 3.1927 A)]",       "Ca19i_EW [mA]",       "Ca19i_FWHM [mA]",       "Ca19i_A",  %  @  3.1927 A (Ca XIX i)
       "Ca19a_lam [A ( 3.1772 A)]",       "Ca19a_EW [mA]",       "Ca19a_FWHM [mA]",       "Ca19a_A",  %  @  3.1772 A (Ca XIX alpha)
       "Ca19b_lam [A ( 2.7050 A)]",       "Ca19b_EW [mA]",       "Ca19b_FWHM [mA]",       "Ca19b_A",  %  @  2.7050 A (Ca XIX beta)
       "Ca19g_lam [A ( 2.5710 A)]",       "Ca19g_EW [mA]",       "Ca19g_FWHM [mA]",       "Ca19g_A",  %  @  2.5710 A (Ca XIX gamma)
       "Ca19d_lam [A ( 2.5140 A)]",       "Ca19d_EW [mA]",       "Ca19d_FWHM [mA]",       "Ca19d_A",  %  @  2.5140 A (Ca XIX delta)
       "Fe26a_lam [A ( 1.7799 A)]",       "Fe26a_EW [mA]",       "Fe26a_FWHM [mA]",       "Fe26a_A",  %  @  1.7799 A (Fe XXVI alpha)
       "Fe26b_lam [A ( 1.5028 A)]",       "Fe26b_EW [mA]",       "Fe26b_FWHM [mA]",       "Fe26b_A",  %  @  1.5028 A (Fe XXVI beta)
       "Fe26g_lam [A ( 1.4251 A)]",       "Fe26g_EW [mA]",       "Fe26g_FWHM [mA]",       "Fe26g_A",  %  @  1.4251 A (Fe XXVI gamma)
       "Fe26d_lam [A ( 1.3918 A)]",       "Fe26d_EW [mA]",       "Fe26d_FWHM [mA]",       "Fe26d_A",  %  @  1.3918 A (Fe XXVI delta)
       "Fe25f_lam [A ( 1.8682 A)]",       "Fe25f_EW [mA]",       "Fe25f_FWHM [mA]",       "Fe25f_A",  %  @  1.8682 A (Fe XXV f)
       "Fe25i_lam [A ( 1.8595 A)]",       "Fe25i_EW [mA]",       "Fe25i_FWHM [mA]",       "Fe25i_A",  %  @  1.8595 A (Fe XXV i)
       "Fe25a_lam [A ( 1.8504 A)]",       "Fe25a_EW [mA]",       "Fe25a_FWHM [mA]",       "Fe25a_A",  %  @  1.8504 A (Fe XXV alpha)
       "Fe25b_lam [A ( 1.5731 A)]",       "Fe25b_EW [mA]",       "Fe25b_FWHM [mA]",       "Fe25b_A",  %  @  1.5731 A (Fe XXV beta)
       "Fe25g_lam [A ( 1.4950 A)]",       "Fe25g_EW [mA]",       "Fe25g_FWHM [mA]",       "Fe25g_A",  %  @  1.4950 A (Fe XXV gamma)
       "Fe25d_lam [A ( 1.4610 A)]",       "Fe25d_EW [mA]",       "Fe25d_FWHM [mA]",       "Fe25d_A",  %  @  1.4610 A (Fe XXV delta)
       "Fe1Ka_lam [A ( 1.9370 A)]",       "Fe1Ka_EW [mA]",       "Fe1Ka_FWHM [mA]",       "Fe1Ka_A",  %  @  1.9370 A (Fe I Kalpha)
  "Fe24_1062A_lam [A (10.6190 A)]",  "Fe24_1062A_EW [mA]",  "Fe24_1062A_FWHM [mA]",  "Fe24_1062A_A",  %  @ 10.6190 A (161758)
   "Fe24_679A_lam [A ( 6.7887 A)]",   "Fe24_679A_EW [mA]",   "Fe24_679A_FWHM [mA]",   "Fe24_679A_A",  %  @  6.7887 A (161756)
    "bl_1066A_lam [A (10.6600 A)]",    "bl_1066A_EW [mA]",    "bl_1066A_FWHM [mA]",    "bl_1066A_A",  %  @ 10.6600 A (161753, 18204) = (Fe XXIV, Fe XVII)
   "Fe24_799A_lam [A ( 7.9908 A)]",   "Fe24_799A_EW [mA]",   "Fe24_799A_FWHM [mA]",   "Fe24_799A_A",  %  @  7.9908 A (161759, 161754)
    "bl_1102A_lam [A (11.0225 A)]",    "bl_1102A_EW [mA]",    "bl_1102A_FWHM [mA]",    "bl_1102A_A",  %  @ 11.0225 A (149890, 18163) = (Fe XXIII, Fe XVII)
    "bl_1099A_lam [A (10.9872 A)]",    "bl_1099A_EW [mA]",    "bl_1099A_FWHM [mA]",    "bl_1099A_A",  %  @ 10.9872 A (149893, 131886) = (Fe XXIII, Fe XXII)
   "Fe23_830A_lam [A ( 8.3038 A)]",   "Fe23_830A_EW [mA]",   "Fe23_830A_FWHM [mA]",   "Fe23_830A_A",  %  @  8.3038 A (149914)
  "Fe22_1225A_lam [A (12.2519 A)]",  "Fe22_1225A_EW [mA]",  "Fe22_1225A_FWHM [mA]",  "Fe22_1225A_A",  %  @ 12.2519 A (131880)
  "Fe22_1143A_lam [A (11.4270 A)]",  "Fe22_1143A_EW [mA]",  "Fe22_1143A_FWHM [mA]",  "Fe22_1143A_A",  %  @ 11.4270 A (132636)
   "Fe22_897A_lam [A ( 8.9748 A)]",   "Fe22_897A_EW [mA]",   "Fe22_897A_FWHM [mA]",   "Fe22_897A_A",  %  @  8.9748 A (132642)
  "Fe22_1193A_lam [A (11.9320 A)]",  "Fe22_1193A_EW [mA]",  "Fe22_1193A_FWHM [mA]",  "Fe22_1193A_A",  %  @ 11.9320 A (132890)
    "bl_1177A_lam [A (11.7660 A)]",    "bl_1177A_EW [mA]",    "bl_1177A_FWHM [mA]",    "bl_1177A_A",  %  @ 11.7660 A (132626, 75499) = (Fe XXII, Fe XX)
  "Fe22_1149A_lam [A (11.4900 A)]",  "Fe22_1149A_EW [mA]",  "Fe22_1149A_FWHM [mA]",  "Fe22_1149A_A",  %  @ 11.4900 A (131888, 132634)
   "Fe22_873A_lam [A ( 8.7307 A)]",   "Fe22_873A_EW [mA]",   "Fe22_873A_FWHM [mA]",   "Fe22_873A_A",  %  @  8.7307 A (131902, 132644)
   "Fe22_786A_lam [A ( 7.8650 A)]",   "Fe22_786A_EW [mA]",   "Fe22_786A_FWHM [mA]",   "Fe22_786A_A",  %  @  7.8650 A (131918, 132676)
  "Fe21_1228A_lam [A (12.2840 A)]",  "Fe21_1228A_EW [mA]",  "Fe21_1228A_FWHM [mA]",  "Fe21_1228A_A",  %  @ 12.2840 A (128657)
  "Fe21_1198A_lam [A (11.9750 A)]",  "Fe21_1198A_EW [mA]",  "Fe21_1198A_FWHM [mA]",  "Fe21_1198A_A",  %  @ 11.9750 A (128681)
  "Fe21_1304A_lam [A (13.0444 A)]",  "Fe21_1304A_EW [mA]",  "Fe21_1304A_FWHM [mA]",  "Fe21_1304A_A",  %  @ 13.0444 A (128653)
     "bl_920A_lam [A ( 9.1967 A)]",     "bl_920A_EW [mA]",     "bl_920A_FWHM [mA]",     "bl_920A_A",  %  @  9.1967 A (128775, 75070, 74587) = (Fe XXI, Fe XX, Fe XX)
   "Fe21_857A_lam [A ( 8.5740 A)]",   "Fe21_857A_EW [mA]",   "Fe21_857A_FWHM [mA]",   "Fe21_857A_A",  %  @  8.5740 A (128847, 131702, 129682)
  "Fe21_1233A_lam [A (12.3270 A)]",  "Fe21_1233A_EW [mA]",  "Fe21_1233A_FWHM [mA]",  "Fe21_1233A_A",  %  @ 12.3270 A (131540)
  "Fe20_1292A_lam [A (12.9165 A)]",  "Fe20_1292A_EW [mA]",  "Fe20_1292A_FWHM [mA]",  "Fe20_1292A_A",  %  @ 12.9165 A (75469, 74908)
  "Fe20_1286A_lam [A (12.8550 A)]",  "Fe20_1286A_EW [mA]",  "Fe20_1286A_FWHM [mA]",  "Fe20_1286A_A",  %  @ 12.8550 A (74900, 75472)
  "Fe20_1283A_lam [A (12.8255 A)]",  "Fe20_1283A_EW [mA]",  "Fe20_1283A_FWHM [mA]",  "Fe20_1283A_A",  %  @ 12.8255 A (74354, 74360)
  "Fe20_1258A_lam [A (12.5760 A)]",  "Fe20_1258A_EW [mA]",  "Fe20_1258A_FWHM [mA]",  "Fe20_1258A_A",  %  @ 12.5760 A (74924, 75496)
  "Fe20_1000A_lam [A ( 9.9992 A)]",  "Fe20_1000A_EW [mA]",  "Fe20_1000A_FWHM [mA]",  "Fe20_1000A_A",  %  @  9.9992 A (74457, 75036, 74463, 75579)
  "Fe20_1297A_lam [A (12.9652 A)]",  "Fe20_1297A_EW [mA]",  "Fe20_1297A_FWHM [mA]",  "Fe20_1297A_A",  %  @ 12.9652 A (75461, 74351)
    "bl_1012A_lam [A (10.1203 A)]",    "bl_1012A_EW [mA]",    "bl_1012A_FWHM [mA]",    "bl_1012A_A",  %  @ 10.1203 A (75587, 39003, 18189) = (Fe XX, Fe XIX, Fe XVII)
    "bl_1301A_lam [A (13.0070 A)]",    "bl_1301A_EW [mA]",    "bl_1301A_FWHM [mA]",    "bl_1301A_A",  %  @ 13.0070 A (74897, 39144) = (Fe XX, Fe XIX)
  "Fe19_1466A_lam [A (14.6640 A)]",  "Fe19_1466A_EW [mA]",  "Fe19_1466A_FWHM [mA]",  "Fe19_1466A_A",  %  @ 14.6640 A (39118)
  "Fe19_1380A_lam [A (13.7950 A)]",  "Fe19_1380A_EW [mA]",  "Fe19_1380A_FWHM [mA]",  "Fe19_1380A_A",  %  @ 13.7950 A (39132)
  "Fe19_1364A_lam [A (13.6450 A)]",  "Fe19_1364A_EW [mA]",  "Fe19_1364A_FWHM [mA]",  "Fe19_1364A_A",  %  @ 13.6450 A (39124)
  "Fe19_1350A_lam [A (13.4970 A)]",  "Fe19_1350A_EW [mA]",  "Fe19_1350A_FWHM [mA]",  "Fe19_1350A_A",  %  @ 13.4970 A (38915)
  "Fe19_1346A_lam [A (13.4620 A)]",  "Fe19_1346A_EW [mA]",  "Fe19_1346A_FWHM [mA]",  "Fe19_1346A_A",  %  @ 13.4620 A (38603)
  "Fe19_1342A_lam [A (13.4230 A)]",  "Fe19_1342A_EW [mA]",  "Fe19_1342A_FWHM [mA]",  "Fe19_1342A_A",  %  @ 13.4230 A (39134)
  "Fe19_1294A_lam [A (12.9450 A)]",  "Fe19_1294A_EW [mA]",  "Fe19_1294A_FWHM [mA]",  "Fe19_1294A_A",  %  @ 12.9450 A (38634)
  "Fe19_1082A_lam [A (10.8160 A)]",  "Fe19_1082A_EW [mA]",  "Fe19_1082A_FWHM [mA]",  "Fe19_1082A_A",  %  @ 10.8160 A (39172)
  "Fe19_1013A_lam [A (10.1309 A)]",  "Fe19_1013A_EW [mA]",  "Fe19_1013A_FWHM [mA]",  "Fe19_1013A_A",  %  @ 10.1309 A (38692)
   "Fe19_986A_lam [A ( 9.8552 A)]",   "Fe19_986A_EW [mA]",   "Fe19_986A_FWHM [mA]",   "Fe19_986A_A",  %  @  9.8552 A (39223)
  "Fe19_1352A_lam [A (13.5163 A)]",  "Fe19_1352A_EW [mA]",  "Fe19_1352A_FWHM [mA]",  "Fe19_1352A_A",  %  @ 13.5163 A (39128, 38611)
  "Fe19_1293A_lam [A (12.9320 A)]",  "Fe19_1293A_EW [mA]",  "Fe19_1293A_FWHM [mA]",  "Fe19_1293A_A",  %  @ 12.9320 A (38941, 39146)
  "Fe18_1453A_lam [A (14.5340 A)]",  "Fe18_1453A_EW [mA]",  "Fe18_1453A_FWHM [mA]",  "Fe18_1453A_A",  %  @ 14.5340 A (38000)
  "Fe18_1437A_lam [A (14.3730 A)]",  "Fe18_1437A_EW [mA]",  "Fe18_1437A_FWHM [mA]",  "Fe18_1437A_A",  %  @ 14.3730 A (37997)
  "Fe18_1457A_lam [A (14.5710 A)]",  "Fe18_1457A_EW [mA]",  "Fe18_1457A_FWHM [mA]",  "Fe18_1457A_A",  %  @ 14.5710 A (37942)
  "Fe18_1562A_lam [A (15.6250 A)]",  "Fe18_1562A_EW [mA]",  "Fe18_1562A_FWHM [mA]",  "Fe18_1562A_A",  %  @ 15.6250 A (37995)
  "Fe18_1332A_lam [A (13.3230 A)]",  "Fe18_1332A_EW [mA]",  "Fe18_1332A_FWHM [mA]",  "Fe18_1332A_A",  %  @ 13.3230 A (37835)
  "Fe18_1426A_lam [A (14.2560 A)]",  "Fe18_1426A_EW [mA]",  "Fe18_1426A_FWHM [mA]",  "Fe18_1426A_A",  %  @ 14.2560 A (37830, 38002)
  "Fe18_1421A_lam [A (14.2056 A)]",  "Fe18_1421A_EW [mA]",  "Fe18_1421A_FWHM [mA]",  "Fe18_1421A_A",  %  @ 14.2056 A (37944, 37998, 38112)
  "Fe18_1153A_lam [A (11.5270 A)]",  "Fe18_1153A_EW [mA]",  "Fe18_1153A_FWHM [mA]",  "Fe18_1153A_A",  %  @ 11.5270 A (37961, 38014)
  "Fe18_1133A_lam [A (11.3260 A)]",  "Fe18_1133A_EW [mA]",  "Fe18_1133A_FWHM [mA]",  "Fe18_1133A_A",  %  @ 11.3260 A (37845, 37963, 38012)
  "Fe17_1678A_lam [A (16.7800 A)]",  "Fe17_1678A_EW [mA]",  "Fe17_1678A_FWHM [mA]",  "Fe17_1678A_A",  %  @ 16.7800 A (20125)
  "Fe17_1526A_lam [A (15.2610 A)]",  "Fe17_1526A_EW [mA]",  "Fe17_1526A_FWHM [mA]",  "Fe17_1526A_A",  %  @ 15.2610 A (20128)
  "Fe17_1501A_lam [A (15.0140 A)]",  "Fe17_1501A_EW [mA]",  "Fe17_1501A_FWHM [mA]",  "Fe17_1501A_A",  %  @ 15.0140 A (20127)
  "Fe17_1227A_lam [A (12.2660 A)]",  "Fe17_1227A_EW [mA]",  "Fe17_1227A_FWHM [mA]",  "Fe17_1227A_A",  %  @ 12.2660 A (18155)
  "Fe17_1212A_lam [A (12.1240 A)]",  "Fe17_1212A_EW [mA]",  "Fe17_1212A_FWHM [mA]",  "Fe17_1212A_A",  %  @ 12.1240 A (18151)
  "Fe17_1536A_lam [A (15.3597 A)]",  "Fe17_1536A_EW [mA]",  "Fe17_1536A_FWHM [mA]",  "Fe17_1536A_A",  %  @ 15.3597 A (25066)
  "Fe17_1382A_lam [A (13.8250 A)]",  "Fe17_1382A_EW [mA]",  "Fe17_1382A_FWHM [mA]",  "Fe17_1382A_A",  %  @ 13.8250 A (20130)
    "bl_1384A_lam [A (13.8410 A)]",    "bl_1384A_EW [mA]",    "bl_1384A_FWHM [mA]",    "bl_1384A_A",  %  @ 13.8410 A (38918, 74885) = (Fe XIX, Fe XX)
      "Al11Ka_lam [A ( 7.8850 A)]",      "Al11Ka_EW [mA]",      "Al11Ka_FWHM [mA]",      "Al11Ka_A",  %  @  7.8850 A (Al XI Kalpha)
      "Al10Ka_lam [A ( 7.9640 A)]",      "Al10Ka_EW [mA]",      "Al10Ka_FWHM [mA]",      "Al10Ka_A",  %  @  7.9640 A (Al X Kalpha)
       "Al9Ka_lam [A ( 8.0500 A)]",       "Al9Ka_EW [mA]",       "Al9Ka_FWHM [mA]",       "Al9Ka_A",  %  @  8.0500 A (Al IX Kalpha)
       "Al8Ka_lam [A ( 8.1290 A)]",       "Al8Ka_EW [mA]",       "Al8Ka_FWHM [mA]",       "Al8Ka_A",  %  @  8.1290 A (Al VIII Kalpha)
       "Al7Ka_lam [A ( 8.2030 A)]",       "Al7Ka_EW [mA]",       "Al7Ka_FWHM [mA]",       "Al7Ka_A",  %  @  8.2030 A (Al VII Kalpha)
       "Al6Ka_lam [A ( 8.2690 A)]",       "Al6Ka_EW [mA]",       "Al6Ka_FWHM [mA]",       "Al6Ka_A",  %  @  8.2690 A (Al VI Kalpha)
       "Al5Ka_lam [A ( 8.3280 A)]",       "Al5Ka_EW [mA]",       "Al5Ka_FWHM [mA]",       "Al5Ka_A",  %  @  8.3280 A (Al V Kalpha)
       "Al4Ka_lam [A ( 8.3320 A)]",       "Al4Ka_EW [mA]",       "Al4Ka_FWHM [mA]",       "Al4Ka_A",  %  @  8.3320 A (Al IV Kalpha)
       "Al3Ka_lam [A ( 8.3360 A)]",       "Al3Ka_EW [mA]",       "Al3Ka_FWHM [mA]",       "Al3Ka_A",  %  @  8.3360 A (Al III Kalpha)
       "Al2Ka_lam [A ( 8.3390 A)]",       "Al2Ka_EW [mA]",       "Al2Ka_FWHM [mA]",       "Al2Ka_A",  %  @  8.3390 A (Al II Kalpha)
      "Si12Ka_lam [A ( 6.7500 A)]",      "Si12Ka_EW [mA]",      "Si12Ka_FWHM [mA]",      "Si12Ka_A",  %  @  6.7500 A (Si XII Kalpha)
      "Si11Ka_lam [A ( 6.8130 A)]",      "Si11Ka_EW [mA]",      "Si11Ka_FWHM [mA]",      "Si11Ka_A",  %  @  6.8130 A (Si XI Kalpha)
      "Si10Ka_lam [A ( 6.8820 A)]",      "Si10Ka_EW [mA]",      "Si10Ka_FWHM [mA]",      "Si10Ka_A",  %  @  6.8820 A (Si X Kalpha)
       "Si9Ka_lam [A ( 6.9470 A)]",       "Si9Ka_EW [mA]",       "Si9Ka_FWHM [mA]",       "Si9Ka_A",  %  @  6.9470 A (Si IX Kalpha)
       "Si8Ka_lam [A ( 7.0070 A)]",       "Si8Ka_EW [mA]",       "Si8Ka_FWHM [mA]",       "Si8Ka_A",  %  @  7.0070 A (Si VIII Kalpha)
       "Si7Ka_lam [A ( 7.0630 A)]",       "Si7Ka_EW [mA]",       "Si7Ka_FWHM [mA]",       "Si7Ka_A",  %  @  7.0630 A (Si VII Kalpha)
       "Si6Ka_lam [A ( 7.1120 A)]",       "Si6Ka_EW [mA]",       "Si6Ka_FWHM [mA]",       "Si6Ka_A",  %  @  7.1120 A (Si VI Kalpha)
       "Si5Ka_lam [A ( 7.1170 A)]",       "Si5Ka_EW [mA]",       "Si5Ka_FWHM [mA]",       "Si5Ka_A",  %  @  7.1170 A (Si V Kalpha)
       "Si4Ka_lam [A ( 7.1210 A)]",       "Si4Ka_EW [mA]",       "Si4Ka_FWHM [mA]",       "Si4Ka_A",  %  @  7.1210 A (Si IV Kalpha)
       "Si3Ka_lam [A ( 7.1240 A)]",       "Si3Ka_EW [mA]",       "Si3Ka_FWHM [mA]",       "Si3Ka_A",  %  @  7.1240 A (Si III Kalpha)
       "Si2Ka_lam [A ( 7.1260 A)]",       "Si2Ka_EW [mA]",       "Si2Ka_FWHM [mA]",       "Si2Ka_A",  %  @  7.1260 A (Si II Kalpha)
      "Ar16Ka_lam [A ( 3.9950 A)]",      "Ar16Ka_EW [mA]",      "Ar16Ka_FWHM [mA]",      "Ar16Ka_A",  %  @  3.9950 A (Ar XVI Kalpha)
      "Ar15Ka_lam [A ( 4.0250 A)]",      "Ar15Ka_EW [mA]",      "Ar15Ka_FWHM [mA]",      "Ar15Ka_A",  %  @  4.0250 A (Ar XV Kalpha)
      "Ar14Ka_lam [A ( 4.0570 A)]",      "Ar14Ka_EW [mA]",      "Ar14Ka_FWHM [mA]",      "Ar14Ka_A",  %  @  4.0570 A (Ar XIV Kalpha)
      "Ar13Ka_lam [A ( 4.0890 A)]",      "Ar13Ka_EW [mA]",      "Ar13Ka_FWHM [mA]",      "Ar13Ka_A",  %  @  4.0890 A (Ar XIII Kalpha)
      "Ar12Ka_lam [A ( 4.1190 A)]",      "Ar12Ka_EW [mA]",      "Ar12Ka_FWHM [mA]",      "Ar12Ka_A",  %  @  4.1190 A (Ar XII Kalpha)
      "Ar11Ka_lam [A ( 4.1470 A)]",      "Ar11Ka_EW [mA]",      "Ar11Ka_FWHM [mA]",      "Ar11Ka_A",  %  @  4.1470 A (Ar XI Kalpha)
      "Ar10Ka_lam [A ( 4.1740 A)]",      "Ar10Ka_EW [mA]",      "Ar10Ka_FWHM [mA]",      "Ar10Ka_A",  %  @  4.1740 A (Ar X Kalpha)
       "Ar9Ka_lam [A ( 4.1780 A)]",       "Ar9Ka_EW [mA]",       "Ar9Ka_FWHM [mA]",       "Ar9Ka_A",  %  @  4.1780 A (Ar IX Kalpha)
       "Ar8Ka_lam [A ( 4.1800 A)]",       "Ar8Ka_EW [mA]",       "Ar8Ka_FWHM [mA]",       "Ar8Ka_A",  %  @  4.1800 A (Ar VIII Kalpha)
       "Ar7Ka_lam [A ( 4.1840 A)]",       "Ar7Ka_EW [mA]",       "Ar7Ka_FWHM [mA]",       "Ar7Ka_A",  %  @  4.1840 A (Ar VII Kalpha)
       "Ar6Ka_lam [A ( 4.1860 A)]",       "Ar6Ka_EW [mA]",       "Ar6Ka_FWHM [mA]",       "Ar6Ka_A",  %  @  4.1860 A (Ar VI Kalpha)
       "Ar5Ka_lam [A ( 4.1890 A)]",       "Ar5Ka_EW [mA]",       "Ar5Ka_FWHM [mA]",       "Ar5Ka_A",  %  @  4.1890 A (Ar V Kalpha)
       "Ar4Ka_lam [A ( 4.1900 A)]",       "Ar4Ka_EW [mA]",       "Ar4Ka_FWHM [mA]",       "Ar4Ka_A",  %  @  4.1900 A (Ar IV Kalpha)
       "Ar3Ka_lam [A ( 4.1920 A)]",       "Ar3Ka_EW [mA]",       "Ar3Ka_FWHM [mA]",       "Ar3Ka_A",  %  @  4.1920 A (Ar III Kalpha)
       "Ar2Ka_lam [A ( 4.1930 A)]",       "Ar2Ka_EW [mA]",       "Ar2Ka_FWHM [mA]",       "Ar2Ka_A",  %  @  4.1930 A (Ar II Kalpha)
]);
set_param_default_hook("lines", "lines_defaults");


%%%%%%%%%%%%%%%%%%%%%%%%%%
define set_lines_par_fun()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{set_lines_par_fun}
%\synopsis{sets the derived amplitude parameter in a lines-model}
%\usage{set_lines_par_fun([Integer_Type id]);}
%\description
%    The amplitude parameters of the lines-model are set for every line:\n
%    \code{set_par_fun("lines(id).line_A", "lines(id).line_EW/lines(id).line_FWHM");}\n
%    If code{id} is not specified, id=1 is used.
%\seealso{gauss, lines, unset_lines_par_fun}
%!%-
{
  variable id=1;
  switch(_NARGS)
  { case 0: ; }
  { case 1: id = (); }
  { help(_function_name()); return; }

  variable lines = "lines("+string(id)+").";
  % Amplitude = EW/[sigma*sqrt{2*pi}] = EW/FWHM * 2*sqrt{ln(2)/pi}
  set_par_fun(lines+"O8a_A", lines+"O8a_EW/"+lines+"O8a_FWHM*0.9394372787");
  set_par_fun(lines+"O8b_A", lines+"O8b_EW/"+lines+"O8b_FWHM*0.9394372787");
  set_par_fun(lines+"O8g_A", lines+"O8g_EW/"+lines+"O8g_FWHM*0.9394372787");
  set_par_fun(lines+"O8d_A", lines+"O8d_EW/"+lines+"O8d_FWHM*0.9394372787");
  set_par_fun(lines+"O7f_A", lines+"O7f_EW/"+lines+"O7f_FWHM*0.9394372787");
  set_par_fun(lines+"O7i_A", lines+"O7i_EW/"+lines+"O7i_FWHM*0.9394372787");
  set_par_fun(lines+"O7a_A", lines+"O7a_EW/"+lines+"O7a_FWHM*0.9394372787");
  set_par_fun(lines+"O7b_A", lines+"O7b_EW/"+lines+"O7b_FWHM*0.9394372787");
  set_par_fun(lines+"O7g_A", lines+"O7g_EW/"+lines+"O7g_FWHM*0.9394372787");
  set_par_fun(lines+"O7d_A", lines+"O7d_EW/"+lines+"O7d_FWHM*0.9394372787");
  set_par_fun(lines+"Ne10a_A", lines+"Ne10a_EW/"+lines+"Ne10a_FWHM*0.9394372787");
  set_par_fun(lines+"Ne10b_A", lines+"Ne10b_EW/"+lines+"Ne10b_FWHM*0.9394372787");
  set_par_fun(lines+"Ne10g_A", lines+"Ne10g_EW/"+lines+"Ne10g_FWHM*0.9394372787");
  set_par_fun(lines+"Ne10d_A", lines+"Ne10d_EW/"+lines+"Ne10d_FWHM*0.9394372787");
  set_par_fun(lines+"Ne10e_A", lines+"Ne10e_EW/"+lines+"Ne10e_FWHM*0.9394372787");
  set_par_fun(lines+"Ne10z_A", lines+"Ne10z_EW/"+lines+"Ne10z_FWHM*0.9394372787");
  set_par_fun(lines+"Ne9f_A", lines+"Ne9f_EW/"+lines+"Ne9f_FWHM*0.9394372787");
  set_par_fun(lines+"Ne9i_A", lines+"Ne9i_EW/"+lines+"Ne9i_FWHM*0.9394372787");
  set_par_fun(lines+"Ne9a_A", lines+"Ne9a_EW/"+lines+"Ne9a_FWHM*0.9394372787");
  set_par_fun(lines+"Ne9b_A", lines+"Ne9b_EW/"+lines+"Ne9b_FWHM*0.9394372787");
  set_par_fun(lines+"Ne9g_A", lines+"Ne9g_EW/"+lines+"Ne9g_FWHM*0.9394372787");
  set_par_fun(lines+"Ne9d_A", lines+"Ne9d_EW/"+lines+"Ne9d_FWHM*0.9394372787");
  set_par_fun(lines+"Ne9e_A", lines+"Ne9e_EW/"+lines+"Ne9e_FWHM*0.9394372787");
  set_par_fun(lines+"Ne9z_A", lines+"Ne9z_EW/"+lines+"Ne9z_FWHM*0.9394372787");
  set_par_fun(lines+"Na11a_A", lines+"Na11a_EW/"+lines+"Na11a_FWHM*0.9394372787");
  set_par_fun(lines+"Na11b_A", lines+"Na11b_EW/"+lines+"Na11b_FWHM*0.9394372787");
  set_par_fun(lines+"Na11g_A", lines+"Na11g_EW/"+lines+"Na11g_FWHM*0.9394372787");
  set_par_fun(lines+"Na11d_A", lines+"Na11d_EW/"+lines+"Na11d_FWHM*0.9394372787");
  set_par_fun(lines+"Na10f_A", lines+"Na10f_EW/"+lines+"Na10f_FWHM*0.9394372787");
  set_par_fun(lines+"Na10i_A", lines+"Na10i_EW/"+lines+"Na10i_FWHM*0.9394372787");
  set_par_fun(lines+"Na10a_A", lines+"Na10a_EW/"+lines+"Na10a_FWHM*0.9394372787");
  set_par_fun(lines+"Na10b_A", lines+"Na10b_EW/"+lines+"Na10b_FWHM*0.9394372787");
  set_par_fun(lines+"Na10g_A", lines+"Na10g_EW/"+lines+"Na10g_FWHM*0.9394372787");
  set_par_fun(lines+"Na10d_A", lines+"Na10d_EW/"+lines+"Na10d_FWHM*0.9394372787");
  set_par_fun(lines+"Mg12a_A", lines+"Mg12a_EW/"+lines+"Mg12a_FWHM*0.9394372787");
  set_par_fun(lines+"Mg12b_A", lines+"Mg12b_EW/"+lines+"Mg12b_FWHM*0.9394372787");
  set_par_fun(lines+"Mg12g_A", lines+"Mg12g_EW/"+lines+"Mg12g_FWHM*0.9394372787");
  set_par_fun(lines+"Mg12d_A", lines+"Mg12d_EW/"+lines+"Mg12d_FWHM*0.9394372787");
  set_par_fun(lines+"Mg11f_A", lines+"Mg11f_EW/"+lines+"Mg11f_FWHM*0.9394372787");
  set_par_fun(lines+"Mg11i_A", lines+"Mg11i_EW/"+lines+"Mg11i_FWHM*0.9394372787");
  set_par_fun(lines+"Mg11a_A", lines+"Mg11a_EW/"+lines+"Mg11a_FWHM*0.9394372787");
  set_par_fun(lines+"Mg11b_A", lines+"Mg11b_EW/"+lines+"Mg11b_FWHM*0.9394372787");
  set_par_fun(lines+"Mg11g_A", lines+"Mg11g_EW/"+lines+"Mg11g_FWHM*0.9394372787");
  set_par_fun(lines+"Mg11d_A", lines+"Mg11d_EW/"+lines+"Mg11d_FWHM*0.9394372787");
  set_par_fun(lines+"Al13a_A", lines+"Al13a_EW/"+lines+"Al13a_FWHM*0.9394372787");
  set_par_fun(lines+"Al13b_A", lines+"Al13b_EW/"+lines+"Al13b_FWHM*0.9394372787");
  set_par_fun(lines+"Al13g_A", lines+"Al13g_EW/"+lines+"Al13g_FWHM*0.9394372787");
  set_par_fun(lines+"Al13d_A", lines+"Al13d_EW/"+lines+"Al13d_FWHM*0.9394372787");
  set_par_fun(lines+"Al12f_A", lines+"Al12f_EW/"+lines+"Al12f_FWHM*0.9394372787");
  set_par_fun(lines+"Al12i_A", lines+"Al12i_EW/"+lines+"Al12i_FWHM*0.9394372787");
  set_par_fun(lines+"Al12a_A", lines+"Al12a_EW/"+lines+"Al12a_FWHM*0.9394372787");
  set_par_fun(lines+"Al12b_A", lines+"Al12b_EW/"+lines+"Al12b_FWHM*0.9394372787");
  set_par_fun(lines+"Al12g_A", lines+"Al12g_EW/"+lines+"Al12g_FWHM*0.9394372787");
  set_par_fun(lines+"Al12d_A", lines+"Al12d_EW/"+lines+"Al12d_FWHM*0.9394372787");
  set_par_fun(lines+"Si14a_A", lines+"Si14a_EW/"+lines+"Si14a_FWHM*0.9394372787");
  set_par_fun(lines+"Si14b_A", lines+"Si14b_EW/"+lines+"Si14b_FWHM*0.9394372787");
  set_par_fun(lines+"Si14g_A", lines+"Si14g_EW/"+lines+"Si14g_FWHM*0.9394372787");
  set_par_fun(lines+"Si14d_A", lines+"Si14d_EW/"+lines+"Si14d_FWHM*0.9394372787");
  set_par_fun(lines+"Si13f_A", lines+"Si13f_EW/"+lines+"Si13f_FWHM*0.9394372787");
  set_par_fun(lines+"Si13i_A", lines+"Si13i_EW/"+lines+"Si13i_FWHM*0.9394372787");
  set_par_fun(lines+"Si13a_A", lines+"Si13a_EW/"+lines+"Si13a_FWHM*0.9394372787");
  set_par_fun(lines+"Si13b_A", lines+"Si13b_EW/"+lines+"Si13b_FWHM*0.9394372787");
  set_par_fun(lines+"Si13g_A", lines+"Si13g_EW/"+lines+"Si13g_FWHM*0.9394372787");
  set_par_fun(lines+"Si13d_A", lines+"Si13d_EW/"+lines+"Si13d_FWHM*0.9394372787");
  set_par_fun(lines+"S16a_A", lines+"S16a_EW/"+lines+"S16a_FWHM*0.9394372787");
  set_par_fun(lines+"S16b_A", lines+"S16b_EW/"+lines+"S16b_FWHM*0.9394372787");
  set_par_fun(lines+"S16g_A", lines+"S16g_EW/"+lines+"S16g_FWHM*0.9394372787");
  set_par_fun(lines+"S16d_A", lines+"S16d_EW/"+lines+"S16d_FWHM*0.9394372787");
  set_par_fun(lines+"S15f_A", lines+"S15f_EW/"+lines+"S15f_FWHM*0.9394372787");
  set_par_fun(lines+"S15i_A", lines+"S15i_EW/"+lines+"S15i_FWHM*0.9394372787");
  set_par_fun(lines+"S15a_A", lines+"S15a_EW/"+lines+"S15a_FWHM*0.9394372787");
  set_par_fun(lines+"S15b_A", lines+"S15b_EW/"+lines+"S15b_FWHM*0.9394372787");
  set_par_fun(lines+"S15g_A", lines+"S15g_EW/"+lines+"S15g_FWHM*0.9394372787");
  set_par_fun(lines+"S15d_A", lines+"S15d_EW/"+lines+"S15d_FWHM*0.9394372787");
  set_par_fun(lines+"Ar18a_A", lines+"Ar18a_EW/"+lines+"Ar18a_FWHM*0.9394372787");
  set_par_fun(lines+"Ar18b_A", lines+"Ar18b_EW/"+lines+"Ar18b_FWHM*0.9394372787");
  set_par_fun(lines+"Ar18g_A", lines+"Ar18g_EW/"+lines+"Ar18g_FWHM*0.9394372787");
  set_par_fun(lines+"Ar18d_A", lines+"Ar18d_EW/"+lines+"Ar18d_FWHM*0.9394372787");
  set_par_fun(lines+"Ar17f_A", lines+"Ar17f_EW/"+lines+"Ar17f_FWHM*0.9394372787");
  set_par_fun(lines+"Ar17i_A", lines+"Ar17i_EW/"+lines+"Ar17i_FWHM*0.9394372787");
  set_par_fun(lines+"Ar17a_A", lines+"Ar17a_EW/"+lines+"Ar17a_FWHM*0.9394372787");
  set_par_fun(lines+"Ar17b_A", lines+"Ar17b_EW/"+lines+"Ar17b_FWHM*0.9394372787");
  set_par_fun(lines+"Ar17g_A", lines+"Ar17g_EW/"+lines+"Ar17g_FWHM*0.9394372787");
  set_par_fun(lines+"Ar17d_A", lines+"Ar17d_EW/"+lines+"Ar17d_FWHM*0.9394372787");
  set_par_fun(lines+"Ca20a_A", lines+"Ca20a_EW/"+lines+"Ca20a_FWHM*0.9394372787");
  set_par_fun(lines+"Ca20b_A", lines+"Ca20b_EW/"+lines+"Ca20b_FWHM*0.9394372787");
  set_par_fun(lines+"Ca20g_A", lines+"Ca20g_EW/"+lines+"Ca20g_FWHM*0.9394372787");
  set_par_fun(lines+"Ca20d_A", lines+"Ca20d_EW/"+lines+"Ca20d_FWHM*0.9394372787");
  set_par_fun(lines+"Ca19f_A", lines+"Ca19f_EW/"+lines+"Ca19f_FWHM*0.9394372787");
  set_par_fun(lines+"Ca19i_A", lines+"Ca19i_EW/"+lines+"Ca19i_FWHM*0.9394372787");
  set_par_fun(lines+"Ca19a_A", lines+"Ca19a_EW/"+lines+"Ca19a_FWHM*0.9394372787");
  set_par_fun(lines+"Ca19b_A", lines+"Ca19b_EW/"+lines+"Ca19b_FWHM*0.9394372787");
  set_par_fun(lines+"Ca19g_A", lines+"Ca19g_EW/"+lines+"Ca19g_FWHM*0.9394372787");
  set_par_fun(lines+"Ca19d_A", lines+"Ca19d_EW/"+lines+"Ca19d_FWHM*0.9394372787");
  set_par_fun(lines+"Fe26a_A", lines+"Fe26a_EW/"+lines+"Fe26a_FWHM*0.9394372787");
  set_par_fun(lines+"Fe26b_A", lines+"Fe26b_EW/"+lines+"Fe26b_FWHM*0.9394372787");
  set_par_fun(lines+"Fe26g_A", lines+"Fe26g_EW/"+lines+"Fe26g_FWHM*0.9394372787");
  set_par_fun(lines+"Fe26d_A", lines+"Fe26d_EW/"+lines+"Fe26d_FWHM*0.9394372787");
  set_par_fun(lines+"Fe25f_A", lines+"Fe25f_EW/"+lines+"Fe25f_FWHM*0.9394372787");
  set_par_fun(lines+"Fe25i_A", lines+"Fe25i_EW/"+lines+"Fe25i_FWHM*0.9394372787");
  set_par_fun(lines+"Fe25a_A", lines+"Fe25a_EW/"+lines+"Fe25a_FWHM*0.9394372787");
  set_par_fun(lines+"Fe25b_A", lines+"Fe25b_EW/"+lines+"Fe25b_FWHM*0.9394372787");
  set_par_fun(lines+"Fe25g_A", lines+"Fe25g_EW/"+lines+"Fe25g_FWHM*0.9394372787");
  set_par_fun(lines+"Fe25d_A", lines+"Fe25d_EW/"+lines+"Fe25d_FWHM*0.9394372787");
  set_par_fun(lines+"Fe1Ka_A", lines+"Fe1Ka_EW/"+lines+"Fe1Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Fe24_1062A_A", lines+"Fe24_1062A_EW/"+lines+"Fe24_1062A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe24_679A_A", lines+"Fe24_679A_EW/"+lines+"Fe24_679A_FWHM*0.9394372787");
  set_par_fun(lines+"bl_1066A_A", lines+"bl_1066A_EW/"+lines+"bl_1066A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe24_799A_A", lines+"Fe24_799A_EW/"+lines+"Fe24_799A_FWHM*0.9394372787");
  set_par_fun(lines+"bl_1102A_A", lines+"bl_1102A_EW/"+lines+"bl_1102A_FWHM*0.9394372787");
  set_par_fun(lines+"bl_1099A_A", lines+"bl_1099A_EW/"+lines+"bl_1099A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe23_830A_A", lines+"Fe23_830A_EW/"+lines+"Fe23_830A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe22_1225A_A", lines+"Fe22_1225A_EW/"+lines+"Fe22_1225A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe22_1143A_A", lines+"Fe22_1143A_EW/"+lines+"Fe22_1143A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe22_897A_A", lines+"Fe22_897A_EW/"+lines+"Fe22_897A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe22_1193A_A", lines+"Fe22_1193A_EW/"+lines+"Fe22_1193A_FWHM*0.9394372787");
  set_par_fun(lines+"bl_1177A_A", lines+"bl_1177A_EW/"+lines+"bl_1177A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe22_1149A_A", lines+"Fe22_1149A_EW/"+lines+"Fe22_1149A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe22_873A_A", lines+"Fe22_873A_EW/"+lines+"Fe22_873A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe22_786A_A", lines+"Fe22_786A_EW/"+lines+"Fe22_786A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe21_1228A_A", lines+"Fe21_1228A_EW/"+lines+"Fe21_1228A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe21_1198A_A", lines+"Fe21_1198A_EW/"+lines+"Fe21_1198A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe21_1304A_A", lines+"Fe21_1304A_EW/"+lines+"Fe21_1304A_FWHM*0.9394372787");
  set_par_fun(lines+"bl_920A_A", lines+"bl_920A_EW/"+lines+"bl_920A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe21_857A_A", lines+"Fe21_857A_EW/"+lines+"Fe21_857A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe21_1233A_A", lines+"Fe21_1233A_EW/"+lines+"Fe21_1233A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe20_1292A_A", lines+"Fe20_1292A_EW/"+lines+"Fe20_1292A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe20_1286A_A", lines+"Fe20_1286A_EW/"+lines+"Fe20_1286A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe20_1283A_A", lines+"Fe20_1283A_EW/"+lines+"Fe20_1283A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe20_1258A_A", lines+"Fe20_1258A_EW/"+lines+"Fe20_1258A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe20_1000A_A", lines+"Fe20_1000A_EW/"+lines+"Fe20_1000A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe20_1297A_A", lines+"Fe20_1297A_EW/"+lines+"Fe20_1297A_FWHM*0.9394372787");
  set_par_fun(lines+"bl_1012A_A", lines+"bl_1012A_EW/"+lines+"bl_1012A_FWHM*0.9394372787");
  set_par_fun(lines+"bl_1301A_A", lines+"bl_1301A_EW/"+lines+"bl_1301A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1466A_A", lines+"Fe19_1466A_EW/"+lines+"Fe19_1466A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1380A_A", lines+"Fe19_1380A_EW/"+lines+"Fe19_1380A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1364A_A", lines+"Fe19_1364A_EW/"+lines+"Fe19_1364A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1350A_A", lines+"Fe19_1350A_EW/"+lines+"Fe19_1350A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1346A_A", lines+"Fe19_1346A_EW/"+lines+"Fe19_1346A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1342A_A", lines+"Fe19_1342A_EW/"+lines+"Fe19_1342A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1294A_A", lines+"Fe19_1294A_EW/"+lines+"Fe19_1294A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1082A_A", lines+"Fe19_1082A_EW/"+lines+"Fe19_1082A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1013A_A", lines+"Fe19_1013A_EW/"+lines+"Fe19_1013A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_986A_A", lines+"Fe19_986A_EW/"+lines+"Fe19_986A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1352A_A", lines+"Fe19_1352A_EW/"+lines+"Fe19_1352A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe19_1293A_A", lines+"Fe19_1293A_EW/"+lines+"Fe19_1293A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe18_1453A_A", lines+"Fe18_1453A_EW/"+lines+"Fe18_1453A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe18_1437A_A", lines+"Fe18_1437A_EW/"+lines+"Fe18_1437A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe18_1457A_A", lines+"Fe18_1457A_EW/"+lines+"Fe18_1457A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe18_1562A_A", lines+"Fe18_1562A_EW/"+lines+"Fe18_1562A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe18_1332A_A", lines+"Fe18_1332A_EW/"+lines+"Fe18_1332A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe18_1426A_A", lines+"Fe18_1426A_EW/"+lines+"Fe18_1426A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe18_1421A_A", lines+"Fe18_1421A_EW/"+lines+"Fe18_1421A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe18_1153A_A", lines+"Fe18_1153A_EW/"+lines+"Fe18_1153A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe18_1133A_A", lines+"Fe18_1133A_EW/"+lines+"Fe18_1133A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe17_1678A_A", lines+"Fe17_1678A_EW/"+lines+"Fe17_1678A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe17_1526A_A", lines+"Fe17_1526A_EW/"+lines+"Fe17_1526A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe17_1501A_A", lines+"Fe17_1501A_EW/"+lines+"Fe17_1501A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe17_1227A_A", lines+"Fe17_1227A_EW/"+lines+"Fe17_1227A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe17_1212A_A", lines+"Fe17_1212A_EW/"+lines+"Fe17_1212A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe17_1536A_A", lines+"Fe17_1536A_EW/"+lines+"Fe17_1536A_FWHM*0.9394372787");
  set_par_fun(lines+"Fe17_1382A_A", lines+"Fe17_1382A_EW/"+lines+"Fe17_1382A_FWHM*0.9394372787");
  set_par_fun(lines+"bl_1384A_A", lines+"bl_1384A_EW/"+lines+"bl_1384A_FWHM*0.9394372787");
  set_par_fun(lines+"Al11Ka_A", lines+"Al11Ka_EW/"+lines+"Al11Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Al10Ka_A", lines+"Al10Ka_EW/"+lines+"Al10Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Al9Ka_A", lines+"Al9Ka_EW/"+lines+"Al9Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Al8Ka_A", lines+"Al8Ka_EW/"+lines+"Al8Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Al7Ka_A", lines+"Al7Ka_EW/"+lines+"Al7Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Al6Ka_A", lines+"Al6Ka_EW/"+lines+"Al6Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Al5Ka_A", lines+"Al5Ka_EW/"+lines+"Al5Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Al4Ka_A", lines+"Al4Ka_EW/"+lines+"Al4Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Al3Ka_A", lines+"Al3Ka_EW/"+lines+"Al3Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Al2Ka_A", lines+"Al2Ka_EW/"+lines+"Al2Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si12Ka_A", lines+"Si12Ka_EW/"+lines+"Si12Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si11Ka_A", lines+"Si11Ka_EW/"+lines+"Si11Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si10Ka_A", lines+"Si10Ka_EW/"+lines+"Si10Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si9Ka_A", lines+"Si9Ka_EW/"+lines+"Si9Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si8Ka_A", lines+"Si8Ka_EW/"+lines+"Si8Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si7Ka_A", lines+"Si7Ka_EW/"+lines+"Si7Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si6Ka_A", lines+"Si6Ka_EW/"+lines+"Si6Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si5Ka_A", lines+"Si5Ka_EW/"+lines+"Si5Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si4Ka_A", lines+"Si4Ka_EW/"+lines+"Si4Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si3Ka_A", lines+"Si3Ka_EW/"+lines+"Si3Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Si2Ka_A", lines+"Si2Ka_EW/"+lines+"Si2Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar16Ka_A", lines+"Ar16Ka_EW/"+lines+"Ar16Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar15Ka_A", lines+"Ar15Ka_EW/"+lines+"Ar15Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar14Ka_A", lines+"Ar14Ka_EW/"+lines+"Ar14Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar13Ka_A", lines+"Ar13Ka_EW/"+lines+"Ar13Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar12Ka_A", lines+"Ar12Ka_EW/"+lines+"Ar12Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar11Ka_A", lines+"Ar11Ka_EW/"+lines+"Ar11Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar10Ka_A", lines+"Ar10Ka_EW/"+lines+"Ar10Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar9Ka_A", lines+"Ar9Ka_EW/"+lines+"Ar9Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar8Ka_A", lines+"Ar8Ka_EW/"+lines+"Ar8Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar7Ka_A", lines+"Ar7Ka_EW/"+lines+"Ar7Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar6Ka_A", lines+"Ar6Ka_EW/"+lines+"Ar6Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar5Ka_A", lines+"Ar5Ka_EW/"+lines+"Ar5Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar4Ka_A", lines+"Ar4Ka_EW/"+lines+"Ar4Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar3Ka_A", lines+"Ar3Ka_EW/"+lines+"Ar3Ka_FWHM*0.9394372787");
  set_par_fun(lines+"Ar2Ka_A", lines+"Ar2Ka_EW/"+lines+"Ar2Ka_FWHM*0.9394372787");
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define unset_lines_par_fun()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{unset_lines_par_fun}
%\synopsis{removes the par_fun's in a lines-model to speed up the model}
%\usage{unset_lines_par_fun([Integer_Type id]);}
%\description
%    The amplitude parameters of the lines-model are reset for every line:\n
%    \code{set_par_fun("lines(id).line_A", NULL);}\n
%    If \code{id} is not specified, code{id=1} is used.
%\seealso{gauss, lines, set_lines_par_fun}
%!%-
{
  variable id=1;
  switch(_NARGS)
  { case 0: ; }
  { case 1: id = (); }
  { help(_function_name()); return; }

  variable lines = "lines("+string(id)+").";
  set_par_fun(lines+"O8a_A", NULL); freeze(lines+"O8a_A");
  set_par_fun(lines+"O8b_A", NULL); freeze(lines+"O8b_A");
  set_par_fun(lines+"O8g_A", NULL); freeze(lines+"O8g_A");
  set_par_fun(lines+"O8d_A", NULL); freeze(lines+"O8d_A");
  set_par_fun(lines+"O7f_A", NULL); freeze(lines+"O7f_A");
  set_par_fun(lines+"O7i_A", NULL); freeze(lines+"O7i_A");
  set_par_fun(lines+"O7a_A", NULL); freeze(lines+"O7a_A");
  set_par_fun(lines+"O7b_A", NULL); freeze(lines+"O7b_A");
  set_par_fun(lines+"O7g_A", NULL); freeze(lines+"O7g_A");
  set_par_fun(lines+"O7d_A", NULL); freeze(lines+"O7d_A");
  set_par_fun(lines+"Ne10a_A", NULL); freeze(lines+"Ne10a_A");
  set_par_fun(lines+"Ne10b_A", NULL); freeze(lines+"Ne10b_A");
  set_par_fun(lines+"Ne10g_A", NULL); freeze(lines+"Ne10g_A");
  set_par_fun(lines+"Ne10d_A", NULL); freeze(lines+"Ne10d_A");
  set_par_fun(lines+"Ne10e_A", NULL); freeze(lines+"Ne10e_A");
  set_par_fun(lines+"Ne10z_A", NULL); freeze(lines+"Ne10z_A");
  set_par_fun(lines+"Ne9f_A", NULL); freeze(lines+"Ne9f_A");
  set_par_fun(lines+"Ne9i_A", NULL); freeze(lines+"Ne9i_A");
  set_par_fun(lines+"Ne9a_A", NULL); freeze(lines+"Ne9a_A");
  set_par_fun(lines+"Ne9b_A", NULL); freeze(lines+"Ne9b_A");
  set_par_fun(lines+"Ne9g_A", NULL); freeze(lines+"Ne9g_A");
  set_par_fun(lines+"Ne9d_A", NULL); freeze(lines+"Ne9d_A");
  set_par_fun(lines+"Ne9e_A", NULL); freeze(lines+"Ne9e_A");
  set_par_fun(lines+"Ne9z_A", NULL); freeze(lines+"Ne9z_A");
  set_par_fun(lines+"Na11a_A", NULL); freeze(lines+"Na11a_A");
  set_par_fun(lines+"Na11b_A", NULL); freeze(lines+"Na11b_A");
  set_par_fun(lines+"Na11g_A", NULL); freeze(lines+"Na11g_A");
  set_par_fun(lines+"Na11d_A", NULL); freeze(lines+"Na11d_A");
  set_par_fun(lines+"Na10f_A", NULL); freeze(lines+"Na10f_A");
  set_par_fun(lines+"Na10i_A", NULL); freeze(lines+"Na10i_A");
  set_par_fun(lines+"Na10a_A", NULL); freeze(lines+"Na10a_A");
  set_par_fun(lines+"Na10b_A", NULL); freeze(lines+"Na10b_A");
  set_par_fun(lines+"Na10g_A", NULL); freeze(lines+"Na10g_A");
  set_par_fun(lines+"Na10d_A", NULL); freeze(lines+"Na10d_A");
  set_par_fun(lines+"Mg12a_A", NULL); freeze(lines+"Mg12a_A");
  set_par_fun(lines+"Mg12b_A", NULL); freeze(lines+"Mg12b_A");
  set_par_fun(lines+"Mg12g_A", NULL); freeze(lines+"Mg12g_A");
  set_par_fun(lines+"Mg12d_A", NULL); freeze(lines+"Mg12d_A");
  set_par_fun(lines+"Mg11f_A", NULL); freeze(lines+"Mg11f_A");
  set_par_fun(lines+"Mg11i_A", NULL); freeze(lines+"Mg11i_A");
  set_par_fun(lines+"Mg11a_A", NULL); freeze(lines+"Mg11a_A");
  set_par_fun(lines+"Mg11b_A", NULL); freeze(lines+"Mg11b_A");
  set_par_fun(lines+"Mg11g_A", NULL); freeze(lines+"Mg11g_A");
  set_par_fun(lines+"Mg11d_A", NULL); freeze(lines+"Mg11d_A");
  set_par_fun(lines+"Al13a_A", NULL); freeze(lines+"Al13a_A");
  set_par_fun(lines+"Al13b_A", NULL); freeze(lines+"Al13b_A");
  set_par_fun(lines+"Al13g_A", NULL); freeze(lines+"Al13g_A");
  set_par_fun(lines+"Al13d_A", NULL); freeze(lines+"Al13d_A");
  set_par_fun(lines+"Al12f_A", NULL); freeze(lines+"Al12f_A");
  set_par_fun(lines+"Al12i_A", NULL); freeze(lines+"Al12i_A");
  set_par_fun(lines+"Al12a_A", NULL); freeze(lines+"Al12a_A");
  set_par_fun(lines+"Al12b_A", NULL); freeze(lines+"Al12b_A");
  set_par_fun(lines+"Al12g_A", NULL); freeze(lines+"Al12g_A");
  set_par_fun(lines+"Al12d_A", NULL); freeze(lines+"Al12d_A");
  set_par_fun(lines+"Si14a_A", NULL); freeze(lines+"Si14a_A");
  set_par_fun(lines+"Si14b_A", NULL); freeze(lines+"Si14b_A");
  set_par_fun(lines+"Si14g_A", NULL); freeze(lines+"Si14g_A");
  set_par_fun(lines+"Si14d_A", NULL); freeze(lines+"Si14d_A");
  set_par_fun(lines+"Si13f_A", NULL); freeze(lines+"Si13f_A");
  set_par_fun(lines+"Si13i_A", NULL); freeze(lines+"Si13i_A");
  set_par_fun(lines+"Si13a_A", NULL); freeze(lines+"Si13a_A");
  set_par_fun(lines+"Si13b_A", NULL); freeze(lines+"Si13b_A");
  set_par_fun(lines+"Si13g_A", NULL); freeze(lines+"Si13g_A");
  set_par_fun(lines+"Si13d_A", NULL); freeze(lines+"Si13d_A");
  set_par_fun(lines+"S16a_A", NULL); freeze(lines+"S16a_A");
  set_par_fun(lines+"S16b_A", NULL); freeze(lines+"S16b_A");
  set_par_fun(lines+"S16g_A", NULL); freeze(lines+"S16g_A");
  set_par_fun(lines+"S16d_A", NULL); freeze(lines+"S16d_A");
  set_par_fun(lines+"S15f_A", NULL); freeze(lines+"S15f_A");
  set_par_fun(lines+"S15i_A", NULL); freeze(lines+"S15i_A");
  set_par_fun(lines+"S15a_A", NULL); freeze(lines+"S15a_A");
  set_par_fun(lines+"S15b_A", NULL); freeze(lines+"S15b_A");
  set_par_fun(lines+"S15g_A", NULL); freeze(lines+"S15g_A");
  set_par_fun(lines+"S15d_A", NULL); freeze(lines+"S15d_A");
  set_par_fun(lines+"Ar18a_A", NULL); freeze(lines+"Ar18a_A");
  set_par_fun(lines+"Ar18b_A", NULL); freeze(lines+"Ar18b_A");
  set_par_fun(lines+"Ar18g_A", NULL); freeze(lines+"Ar18g_A");
  set_par_fun(lines+"Ar18d_A", NULL); freeze(lines+"Ar18d_A");
  set_par_fun(lines+"Ar17f_A", NULL); freeze(lines+"Ar17f_A");
  set_par_fun(lines+"Ar17i_A", NULL); freeze(lines+"Ar17i_A");
  set_par_fun(lines+"Ar17a_A", NULL); freeze(lines+"Ar17a_A");
  set_par_fun(lines+"Ar17b_A", NULL); freeze(lines+"Ar17b_A");
  set_par_fun(lines+"Ar17g_A", NULL); freeze(lines+"Ar17g_A");
  set_par_fun(lines+"Ar17d_A", NULL); freeze(lines+"Ar17d_A");
  set_par_fun(lines+"Ca20a_A", NULL); freeze(lines+"Ca20a_A");
  set_par_fun(lines+"Ca20b_A", NULL); freeze(lines+"Ca20b_A");
  set_par_fun(lines+"Ca20g_A", NULL); freeze(lines+"Ca20g_A");
  set_par_fun(lines+"Ca20d_A", NULL); freeze(lines+"Ca20d_A");
  set_par_fun(lines+"Ca19f_A", NULL); freeze(lines+"Ca19f_A");
  set_par_fun(lines+"Ca19i_A", NULL); freeze(lines+"Ca19i_A");
  set_par_fun(lines+"Ca19a_A", NULL); freeze(lines+"Ca19a_A");
  set_par_fun(lines+"Ca19b_A", NULL); freeze(lines+"Ca19b_A");
  set_par_fun(lines+"Ca19g_A", NULL); freeze(lines+"Ca19g_A");
  set_par_fun(lines+"Ca19d_A", NULL); freeze(lines+"Ca19d_A");
  set_par_fun(lines+"Fe26a_A", NULL); freeze(lines+"Fe26a_A");
  set_par_fun(lines+"Fe26b_A", NULL); freeze(lines+"Fe26b_A");
  set_par_fun(lines+"Fe26g_A", NULL); freeze(lines+"Fe26g_A");
  set_par_fun(lines+"Fe26d_A", NULL); freeze(lines+"Fe26d_A");
  set_par_fun(lines+"Fe25f_A", NULL); freeze(lines+"Fe25f_A");
  set_par_fun(lines+"Fe25i_A", NULL); freeze(lines+"Fe25i_A");
  set_par_fun(lines+"Fe25a_A", NULL); freeze(lines+"Fe25a_A");
  set_par_fun(lines+"Fe25b_A", NULL); freeze(lines+"Fe25b_A");
  set_par_fun(lines+"Fe25g_A", NULL); freeze(lines+"Fe25g_A");
  set_par_fun(lines+"Fe25d_A", NULL); freeze(lines+"Fe25d_A");
  set_par_fun(lines+"Fe1Ka_A", NULL); freeze(lines+"Fe1Ka_A");
  set_par_fun(lines+"Fe24_1062A_A", NULL); freeze(lines+"Fe24_1062A_A");
  set_par_fun(lines+"Fe24_679A_A", NULL); freeze(lines+"Fe24_679A_A");
  set_par_fun(lines+"bl_1066A_A", NULL); freeze(lines+"bl_1066A_A");
  set_par_fun(lines+"Fe24_799A_A", NULL); freeze(lines+"Fe24_799A_A");
  set_par_fun(lines+"bl_1102A_A", NULL); freeze(lines+"bl_1102A_A");
  set_par_fun(lines+"bl_1099A_A", NULL); freeze(lines+"bl_1099A_A");
  set_par_fun(lines+"Fe23_830A_A", NULL); freeze(lines+"Fe23_830A_A");
  set_par_fun(lines+"Fe22_1225A_A", NULL); freeze(lines+"Fe22_1225A_A");
  set_par_fun(lines+"Fe22_1143A_A", NULL); freeze(lines+"Fe22_1143A_A");
  set_par_fun(lines+"Fe22_897A_A", NULL); freeze(lines+"Fe22_897A_A");
  set_par_fun(lines+"Fe22_1193A_A", NULL); freeze(lines+"Fe22_1193A_A");
  set_par_fun(lines+"bl_1177A_A", NULL); freeze(lines+"bl_1177A_A");
  set_par_fun(lines+"Fe22_1149A_A", NULL); freeze(lines+"Fe22_1149A_A");
  set_par_fun(lines+"Fe22_873A_A", NULL); freeze(lines+"Fe22_873A_A");
  set_par_fun(lines+"Fe22_786A_A", NULL); freeze(lines+"Fe22_786A_A");
  set_par_fun(lines+"Fe21_1228A_A", NULL); freeze(lines+"Fe21_1228A_A");
  set_par_fun(lines+"Fe21_1198A_A", NULL); freeze(lines+"Fe21_1198A_A");
  set_par_fun(lines+"Fe21_1304A_A", NULL); freeze(lines+"Fe21_1304A_A");
  set_par_fun(lines+"bl_920A_A", NULL); freeze(lines+"bl_920A_A");
  set_par_fun(lines+"Fe21_857A_A", NULL); freeze(lines+"Fe21_857A_A");
  set_par_fun(lines+"Fe21_1233A_A", NULL); freeze(lines+"Fe21_1233A_A");
  set_par_fun(lines+"Fe20_1292A_A", NULL); freeze(lines+"Fe20_1292A_A");
  set_par_fun(lines+"Fe20_1286A_A", NULL); freeze(lines+"Fe20_1286A_A");
  set_par_fun(lines+"Fe20_1283A_A", NULL); freeze(lines+"Fe20_1283A_A");
  set_par_fun(lines+"Fe20_1258A_A", NULL); freeze(lines+"Fe20_1258A_A");
  set_par_fun(lines+"Fe20_1000A_A", NULL); freeze(lines+"Fe20_1000A_A");
  set_par_fun(lines+"Fe20_1297A_A", NULL); freeze(lines+"Fe20_1297A_A");
  set_par_fun(lines+"bl_1012A_A", NULL); freeze(lines+"bl_1012A_A");
  set_par_fun(lines+"bl_1301A_A", NULL); freeze(lines+"bl_1301A_A");
  set_par_fun(lines+"Fe19_1466A_A", NULL); freeze(lines+"Fe19_1466A_A");
  set_par_fun(lines+"Fe19_1380A_A", NULL); freeze(lines+"Fe19_1380A_A");
  set_par_fun(lines+"Fe19_1364A_A", NULL); freeze(lines+"Fe19_1364A_A");
  set_par_fun(lines+"Fe19_1350A_A", NULL); freeze(lines+"Fe19_1350A_A");
  set_par_fun(lines+"Fe19_1346A_A", NULL); freeze(lines+"Fe19_1346A_A");
  set_par_fun(lines+"Fe19_1342A_A", NULL); freeze(lines+"Fe19_1342A_A");
  set_par_fun(lines+"Fe19_1294A_A", NULL); freeze(lines+"Fe19_1294A_A");
  set_par_fun(lines+"Fe19_1082A_A", NULL); freeze(lines+"Fe19_1082A_A");
  set_par_fun(lines+"Fe19_1013A_A", NULL); freeze(lines+"Fe19_1013A_A");
  set_par_fun(lines+"Fe19_986A_A", NULL); freeze(lines+"Fe19_986A_A");
  set_par_fun(lines+"Fe19_1352A_A", NULL); freeze(lines+"Fe19_1352A_A");
  set_par_fun(lines+"Fe19_1293A_A", NULL); freeze(lines+"Fe19_1293A_A");
  set_par_fun(lines+"Fe18_1453A_A", NULL); freeze(lines+"Fe18_1453A_A");
  set_par_fun(lines+"Fe18_1437A_A", NULL); freeze(lines+"Fe18_1437A_A");
  set_par_fun(lines+"Fe18_1457A_A", NULL); freeze(lines+"Fe18_1457A_A");
  set_par_fun(lines+"Fe18_1562A_A", NULL); freeze(lines+"Fe18_1562A_A");
  set_par_fun(lines+"Fe18_1332A_A", NULL); freeze(lines+"Fe18_1332A_A");
  set_par_fun(lines+"Fe18_1426A_A", NULL); freeze(lines+"Fe18_1426A_A");
  set_par_fun(lines+"Fe18_1421A_A", NULL); freeze(lines+"Fe18_1421A_A");
  set_par_fun(lines+"Fe18_1153A_A", NULL); freeze(lines+"Fe18_1153A_A");
  set_par_fun(lines+"Fe18_1133A_A", NULL); freeze(lines+"Fe18_1133A_A");
  set_par_fun(lines+"Fe17_1678A_A", NULL); freeze(lines+"Fe17_1678A_A");
  set_par_fun(lines+"Fe17_1526A_A", NULL); freeze(lines+"Fe17_1526A_A");
  set_par_fun(lines+"Fe17_1501A_A", NULL); freeze(lines+"Fe17_1501A_A");
  set_par_fun(lines+"Fe17_1227A_A", NULL); freeze(lines+"Fe17_1227A_A");
  set_par_fun(lines+"Fe17_1212A_A", NULL); freeze(lines+"Fe17_1212A_A");
  set_par_fun(lines+"Fe17_1536A_A", NULL); freeze(lines+"Fe17_1536A_A");
  set_par_fun(lines+"Fe17_1382A_A", NULL); freeze(lines+"Fe17_1382A_A");
  set_par_fun(lines+"bl_1384A_A", NULL); freeze(lines+"bl_1384A_A");
  set_par_fun(lines+"Al11Ka_A", NULL); freeze(lines+"Al11Ka_A");
  set_par_fun(lines+"Al10Ka_A", NULL); freeze(lines+"Al10Ka_A");
  set_par_fun(lines+"Al9Ka_A", NULL); freeze(lines+"Al9Ka_A");
  set_par_fun(lines+"Al8Ka_A", NULL); freeze(lines+"Al8Ka_A");
  set_par_fun(lines+"Al7Ka_A", NULL); freeze(lines+"Al7Ka_A");
  set_par_fun(lines+"Al6Ka_A", NULL); freeze(lines+"Al6Ka_A");
  set_par_fun(lines+"Al5Ka_A", NULL); freeze(lines+"Al5Ka_A");
  set_par_fun(lines+"Al4Ka_A", NULL); freeze(lines+"Al4Ka_A");
  set_par_fun(lines+"Al3Ka_A", NULL); freeze(lines+"Al3Ka_A");
  set_par_fun(lines+"Al2Ka_A", NULL); freeze(lines+"Al2Ka_A");
  set_par_fun(lines+"Si12Ka_A", NULL); freeze(lines+"Si12Ka_A");
  set_par_fun(lines+"Si11Ka_A", NULL); freeze(lines+"Si11Ka_A");
  set_par_fun(lines+"Si10Ka_A", NULL); freeze(lines+"Si10Ka_A");
  set_par_fun(lines+"Si9Ka_A", NULL); freeze(lines+"Si9Ka_A");
  set_par_fun(lines+"Si8Ka_A", NULL); freeze(lines+"Si8Ka_A");
  set_par_fun(lines+"Si7Ka_A", NULL); freeze(lines+"Si7Ka_A");
  set_par_fun(lines+"Si6Ka_A", NULL); freeze(lines+"Si6Ka_A");
  set_par_fun(lines+"Si5Ka_A", NULL); freeze(lines+"Si5Ka_A");
  set_par_fun(lines+"Si4Ka_A", NULL); freeze(lines+"Si4Ka_A");
  set_par_fun(lines+"Si3Ka_A", NULL); freeze(lines+"Si3Ka_A");
  set_par_fun(lines+"Si2Ka_A", NULL); freeze(lines+"Si2Ka_A");
  set_par_fun(lines+"Ar16Ka_A", NULL); freeze(lines+"Ar16Ka_A");
  set_par_fun(lines+"Ar15Ka_A", NULL); freeze(lines+"Ar15Ka_A");
  set_par_fun(lines+"Ar14Ka_A", NULL); freeze(lines+"Ar14Ka_A");
  set_par_fun(lines+"Ar13Ka_A", NULL); freeze(lines+"Ar13Ka_A");
  set_par_fun(lines+"Ar12Ka_A", NULL); freeze(lines+"Ar12Ka_A");
  set_par_fun(lines+"Ar11Ka_A", NULL); freeze(lines+"Ar11Ka_A");
  set_par_fun(lines+"Ar10Ka_A", NULL); freeze(lines+"Ar10Ka_A");
  set_par_fun(lines+"Ar9Ka_A", NULL); freeze(lines+"Ar9Ka_A");
  set_par_fun(lines+"Ar8Ka_A", NULL); freeze(lines+"Ar8Ka_A");
  set_par_fun(lines+"Ar7Ka_A", NULL); freeze(lines+"Ar7Ka_A");
  set_par_fun(lines+"Ar6Ka_A", NULL); freeze(lines+"Ar6Ka_A");
  set_par_fun(lines+"Ar5Ka_A", NULL); freeze(lines+"Ar5Ka_A");
  set_par_fun(lines+"Ar4Ka_A", NULL); freeze(lines+"Ar4Ka_A");
  set_par_fun(lines+"Ar3Ka_A", NULL); freeze(lines+"Ar3Ka_A");
  set_par_fun(lines+"Ar2Ka_A", NULL); freeze(lines+"Ar2Ka_A");
}


%%%%%%%%%%%%%%%%%%%%%%%%%
define set_gauss_line_par()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{set_gauss_line_par}
%\synopsis{initializes a line in the lines-model with the parameters of a gauss-line}
%\usage{set_gauss_line_par([id,] line, area, center, sigma);}
%\description
%    If \code{id} is not specified, \code{id=1} is used.
%    \code{line} is the name in the lines-model, appearing as parameters
%    \code{line_lam}, \code{line_EW}, \code{line_FWHM} and \code{line_A}.
%\seealso{gauss, lines}
%!%-
{
  variable id=1, line, area, center, sigma;
  switch(_NARGS)
  { case 4: (line, area, center, sigma) = (); }
  { case 5: (id, line, area, center, sigma) = (); }
  { help(_function_name()); return; }

  if(line=="") { message(sprintf("warning (%s): line @ %.4f A without a name is skipped", _function_name(), center)); return; }

  line = sprintf("lines(%d).%s_", id, line);
  % message(line);
  if( howmany(array_struct_field(get_params(), "name")==line) == 0 )  { message("error (%s): %s does not exist", _function_name(), line); return; }
  variable info = get_par_info(line+"lam");
  % print(info);
  if(center < info.min or center > info.max)
  { set_par(line+"lam", center, 0, _min(info.min, center), _max(info.max, center));
    message(sprintf("warning (%s): line @ %.2f A is out of range [%.4f:%.4f]", _function_name(), center, info.min, info.max));
  }
  else
  { set_par(line+"lam", center, 0); }
  set_par(line+"EW", abs(area)*1e3, 0);
  variable A = area/sqrt(2*PI)/sigma;
  if(A<-1) { message(sprintf("warning (%s): line @ %.2f A had amplitude %.1f< -1. It is now set to -1.", _function_name(), center, A)); A = -1; }
  if(A>10) { message(sprintf("warning (%s): line @ %.2f A had amplitude %g > 10. It is now set to 10.", _function_name(), center, A)); A = 10; }
  set_par(line+"A", A, 0);
}


%%%%%%%%%%%%%%%
define fit_line()
%%%%%%%%%%%%%%%
%!%+
%\function{fit_line}
%\synopsis{activates a line in the lines model and fits its parameters}
%\usage{fit_line([id,] line);}
%\seealso{lines}
%!%-
{
  variable id=1, line;
  switch(_NARGS)
  { case 1:      line  = (); }
  { case 2: (id, line) = (); }
  { help(_function_name()); return; }

  line = sprintf("lines(%d).%s_", id, line);

  variable t = thawedParameters();
  freeze("*");
  thaw(line+"lam");
  if(get_par(line+"EW")==0)  { set_par(line+"EW", -10, 0); }  else  { thaw(line+"EW"); }
  thaw(line+"FWHM");

  message("before fitting");
  list_free;

  fit_counts;
  message("after fitting");
  list_free;

  thaw(t);
}
%%%%%%%%%%%%%%%%%%%%
define dont_use_line()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{dont_use_line}
%\synopsis{sets the EW / amplitude of a line in the lines-model to zero}
%\usage{dont_use_line([id,] line);}
%\description
%    If \code{id} is not specified, \code{id=1} is used.
%    \code{line} is the name in the lines-model, appearing as parameters
%    \code{line_lam}, \code{line_EW}, \code{line_FWHM} and \code{line_A}.
%\seealso{lines}
%!%-
{
  variable id=1, line;
  switch(_NARGS)
  { case 1: (line) = (); }
  { case 2: (id, line) = (); }
  { help(_function_name()); return; }

  variable l;
  foreach l ([line])
  { l = sprintf("lines(%d).%s_", id, l);
    set_par(l+"EW", 0, 1);
    set_par(l+"A", 0, 1);
  }
}


%%%%%%%%%%%%%%%%%%%
define compare_pars(a, b)
%%%%%%%%%%%%%%%%%%%
{ a = get_par(a);
  b = get_par(b);
  if(a<=b) { return -1; } else { if(a==b) { return +1; } else { return +1; } }
}


%%%%%%%%%%%%%%%%%%%
define get_line_ids()
%%%%%%%%%%%%%%%%%%%
{
  variable line;
  switch(_NARGS)
  { case 1: line = (); }
  { help(_function_name()); return; }

  switch(line)
  { case "O8a": return [99080821]; }  %  (O VIII alpha)
  { case "O8b": return [99080822]; }  %  (O VIII beta)
  { case "O8g": return [99080823]; }  %  (O VIII gamma)
  { case "O8d": return [99080824]; }  %  (O VIII delta)
  { case "O7f": return [99080719]; }  %  (O VII f)
  { case "O7i": return [99080720]; }  %  (O VII i)
  { case "O7a": return [99080721]; }  %  (O VII alpha)
  { case "O7b": return [99080722]; }  %  (O VII beta)
  { case "O7g": return [99080723]; }  %  (O VII gamma)
  { case "O7d": return [99080724]; }  %  (O VII delta)
  { case "Ne10a": return [99101021]; }  %  (Ne X alpha)
  { case "Ne10b": return [99101022]; }  %  (Ne X beta)
  { case "Ne10g": return [99101023]; }  %  (Ne X gamma)
  { case "Ne10d": return [99101024]; }  %  (Ne X delta)
  { case "Ne10e": return [99101025]; }  %  (Ne X epsilon)
  { case "Ne10z": return [99101026]; }  %  (Ne X zeta)
  { case "Ne9f": return [99100919]; }  %  (Ne IX f)
  { case "Ne9i": return [99100920]; }  %  (Ne IX i)
  { case "Ne9a": return [99100921]; }  %  (Ne IX alpha)
  { case "Ne9b": return [99100922]; }  %  (Ne IX beta)
  { case "Ne9g": return [99100923]; }  %  (Ne IX gamma)
  { case "Ne9d": return [99100924]; }  %  (Ne IX delta)
  { case "Ne9e": return [99100925]; }  %  (Ne IX epsilon)
  { case "Ne9z": return [99100926]; }  %  (Ne IX zeta)
  { case "Na11a": return [99111121]; }  %  (Na XI alpha)
  { case "Na11b": return [99111122]; }  %  (Na XI beta)
  { case "Na11g": return [99111123]; }  %  (Na XI gamma)
  { case "Na11d": return [99111124]; }  %  (Na XI delta)
  { case "Na10f": return [99111019]; }  %  (Na X f)
  { case "Na10i": return [99111020]; }  %  (Na X i)
  { case "Na10a": return [99111021]; }  %  (Na X alpha)
  { case "Na10b": return [99111022]; }  %  (Na X beta)
  { case "Na10g": return [99111023]; }  %  (Na X gamma)
  { case "Na10d": return [99111024]; }  %  (Na X delta)
  { case "Mg12a": return [99121221]; }  %  (Mg XII alpha)
  { case "Mg12b": return [99121222]; }  %  (Mg XII beta)
  { case "Mg12g": return [99121223]; }  %  (Mg XII gamma)
  { case "Mg12d": return [99121224]; }  %  (Mg XII delta)
  { case "Mg11f": return [99121119]; }  %  (Mg XI f)
  { case "Mg11i": return [99121120]; }  %  (Mg XI i)
  { case "Mg11a": return [99121121]; }  %  (Mg XI alpha)
  { case "Mg11b": return [99121122]; }  %  (Mg XI beta)
  { case "Mg11g": return [99121123]; }  %  (Mg XI gamma)
  { case "Mg11d": return [99121124]; }  %  (Mg XI delta)
  { case "Al13a": return [99131321]; }  %  (Al XIII alpha)
  { case "Al13b": return [99131322]; }  %  (Al XIII beta)
  { case "Al13g": return [99131323]; }  %  (Al XIII gamma)
  { case "Al13d": return [99131324]; }  %  (Al XIII delta)
  { case "Al12f": return [99131219]; }  %  (Al XII f)
  { case "Al12i": return [99131220]; }  %  (Al XII i)
  { case "Al12a": return [99131221]; }  %  (Al XII alpha)
  { case "Al12b": return [99131222]; }  %  (Al XII beta)
  { case "Al12g": return [99131223]; }  %  (Al XII gamma)
  { case "Al12d": return [99131224]; }  %  (Al XII delta)
  { case "Si14a": return [99141421]; }  %  (Si XIV alpha)
  { case "Si14b": return [99141422]; }  %  (Si XIV beta)
  { case "Si14g": return [99141423]; }  %  (Si XIV gamma)
  { case "Si14d": return [99141424]; }  %  (Si XIV delta)
  { case "Si13f": return [99141319]; }  %  (Si XIII f)
  { case "Si13i": return [99141320]; }  %  (Si XIII i)
  { case "Si13a": return [99141321]; }  %  (Si XIII alpha)
  { case "Si13b": return [99141322]; }  %  (Si XIII beta)
  { case "Si13g": return [99141323]; }  %  (Si XIII gamma)
  { case "Si13d": return [99141324]; }  %  (Si XIII delta)
  { case "S16a": return [99161621]; }  %  (S XVI alpha)
  { case "S16b": return [99161622]; }  %  (S XVI beta)
  { case "S16g": return [99161623]; }  %  (S XVI gamma)
  { case "S16d": return [99161624]; }  %  (S XVI delta)
  { case "S15f": return [99161519]; }  %  (S XV f)
  { case "S15i": return [99161520]; }  %  (S XV i)
  { case "S15a": return [99161521]; }  %  (S XV alpha)
  { case "S15b": return [99161522]; }  %  (S XV beta)
  { case "S15g": return [99161523]; }  %  (S XV gamma)
  { case "S15d": return [99161524]; }  %  (S XV delta)
  { case "Ar18a": return [99181821]; }  %  (Ar XVIII alpha)
  { case "Ar18b": return [99181822]; }  %  (Ar XVIII beta)
  { case "Ar18g": return [99181823]; }  %  (Ar XVIII gamma)
  { case "Ar18d": return [99181824]; }  %  (Ar XVIII delta)
  { case "Ar17f": return [99181719]; }  %  (Ar XVII f)
  { case "Ar17i": return [99181720]; }  %  (Ar XVII i)
  { case "Ar17a": return [99181721]; }  %  (Ar XVII alpha)
  { case "Ar17b": return [99181722]; }  %  (Ar XVII beta)
  { case "Ar17g": return [99181723]; }  %  (Ar XVII gamma)
  { case "Ar17d": return [99181724]; }  %  (Ar XVII delta)
  { case "Ca20a": return [99202021]; }  %  (Ca XX alpha)
  { case "Ca20b": return [99202022]; }  %  (Ca XX beta)
  { case "Ca20g": return [99202023]; }  %  (Ca XX gamma)
  { case "Ca20d": return [99202024]; }  %  (Ca XX delta)
  { case "Ca19f": return [99201919]; }  %  (Ca XIX f)
  { case "Ca19i": return [99201920]; }  %  (Ca XIX i)
  { case "Ca19a": return [99201921]; }  %  (Ca XIX alpha)
  { case "Ca19b": return [99201922]; }  %  (Ca XIX beta)
  { case "Ca19g": return [99201923]; }  %  (Ca XIX gamma)
  { case "Ca19d": return [99201924]; }  %  (Ca XIX delta)
  { case "Fe26a": return [99262621]; }  %  (Fe XXVI alpha)
  { case "Fe26b": return [99262622]; }  %  (Fe XXVI beta)
  { case "Fe26g": return [99262623]; }  %  (Fe XXVI gamma)
  { case "Fe26d": return [99262624]; }  %  (Fe XXVI delta)
  { case "Fe25f": return [99262519]; }  %  (Fe XXV f)
  { case "Fe25i": return [99262520]; }  %  (Fe XXV i)
  { case "Fe25a": return [99262521]; }  %  (Fe XXV alpha)
  { case "Fe25b": return [99262522]; }  %  (Fe XXV beta)
  { case "Fe25g": return [99262523]; }  %  (Fe XXV gamma)
  { case "Fe25d": return [99262524]; }  %  (Fe XXV delta)
  { case "Fe1Ka": return [99260101]; }  %  (Fe I Kalpha)
  { case "Fe24_1062A": return [161758]; }  %  (161758)
  { case "Fe24_679A": return [161756]; }  %  (161756)
  { case "bl_1066A": return [161753, 18204]; }  %  (161753, 18204) = (Fe XXIV, Fe XVII)
  { case "Fe24_799A": return [161759, 161754]; }  %  (161759, 161754)
  { case "bl_1102A": return [149890, 18163]; }  %  (149890, 18163) = (Fe XXIII, Fe XVII)
  { case "bl_1099A": return [149893, 131886]; }  %  (149893, 131886) = (Fe XXIII, Fe XXII)
  { case "Fe23_830A": return [149914]; }  %  (149914)
  { case "Fe22_1225A": return [131880]; }  %  (131880)
  { case "Fe22_1143A": return [132636]; }  %  (132636)
  { case "Fe22_897A": return [132642]; }  %  (132642)
  { case "Fe22_1193A": return [132890]; }  %  (132890)
  { case "bl_1177A": return [132626, 75499]; }  %  (132626, 75499) = (Fe XXII, Fe XX)
  { case "Fe22_1149A": return [131888, 132634]; }  %  (131888, 132634)
  { case "Fe22_873A": return [131902, 132644]; }  %  (131902, 132644)
  { case "Fe22_786A": return [131918, 132676]; }  %  (131918, 132676)
  { case "Fe21_1228A": return [128657]; }  %  (128657)
  { case "Fe21_1198A": return [128681]; }  %  (128681)
  { case "Fe21_1304A": return [128653]; }  %  (128653)
  { case "bl_920A": return [128775, 75070, 74587]; }  %  (128775, 75070, 74587) = (Fe XXI, Fe XX, Fe XX)
  { case "Fe21_857A": return [128847, 131702, 129682]; }  %  (128847, 131702, 129682)
  { case "Fe21_1233A": return [131540]; }  %  (131540)
  { case "Fe20_1292A": return [75469, 74908]; }  %  (75469, 74908)
  { case "Fe20_1286A": return [74900, 75472]; }  %  (74900, 75472)
  { case "Fe20_1283A": return [74354, 74360]; }  %  (74354, 74360)
  { case "Fe20_1258A": return [74924, 75496]; }  %  (74924, 75496)
  { case "Fe20_1000A": return [74457, 75036, 74463, 75579]; }  %  (74457, 75036, 74463, 75579)
  { case "Fe20_1297A": return [75461, 74351]; }  %  (75461, 74351)
  { case "bl_1012A": return [75587, 39003, 18189]; }  %  (75587, 39003, 18189) = (Fe XX, Fe XIX, Fe XVII)
  { case "bl_1301A": return [74897, 39144]; }  %  (74897, 39144) = (Fe XX, Fe XIX)
  { case "Fe19_1466A": return [39118]; }  %  (39118)
  { case "Fe19_1380A": return [39132]; }  %  (39132)
  { case "Fe19_1364A": return [39124]; }  %  (39124)
  { case "Fe19_1350A": return [38915]; }  %  (38915)
  { case "Fe19_1346A": return [38603]; }  %  (38603)
  { case "Fe19_1342A": return [39134]; }  %  (39134)
  { case "Fe19_1294A": return [38634]; }  %  (38634)
  { case "Fe19_1082A": return [39172]; }  %  (39172)
  { case "Fe19_1013A": return [38692]; }  %  (38692)
  { case "Fe19_986A": return [39223]; }  %  (39223)
  { case "Fe19_1352A": return [39128, 38611]; }  %  (39128, 38611)
  { case "Fe19_1293A": return [38941, 39146]; }  %  (38941, 39146)
  { case "Fe18_1453A": return [38000]; }  %  (38000)
  { case "Fe18_1437A": return [37997]; }  %  (37997)
  { case "Fe18_1457A": return [37942]; }  %  (37942)
  { case "Fe18_1562A": return [37995]; }  %  (37995)
  { case "Fe18_1332A": return [37835]; }  %  (37835)
  { case "Fe18_1426A": return [37830, 38002]; }  %  (37830, 38002)
  { case "Fe18_1421A": return [37944, 37998, 38112]; }  %  (37944, 37998, 38112)
  { case "Fe18_1153A": return [37961, 38014]; }  %  (37961, 38014)
  { case "Fe18_1133A": return [37845, 37963, 38012]; }  %  (37845, 37963, 38012)
  { case "Fe17_1678A": return [20125]; }  %  (20125)
  { case "Fe17_1526A": return [20128]; }  %  (20128)
  { case "Fe17_1501A": return [20127]; }  %  (20127)
  { case "Fe17_1227A": return [18155]; }  %  (18155)
  { case "Fe17_1212A": return [18151]; }  %  (18151)
  { case "Fe17_1536A": return [25066]; }  %  (25066)
  { case "Fe17_1382A": return [20130]; }  %  (20130)
  { case "bl_1384A": return [38918, 74885]; }  %  (38918, 74885) = (Fe XIX, Fe XX)
  { case "Al11Ka": return [99131101]; }  %  (Al XI Kalpha)
  { case "Al10Ka": return [99131001]; }  %  (Al X Kalpha)
  { case "Al9Ka": return [99130901]; }  %  (Al IX Kalpha)
  { case "Al8Ka": return [99130801]; }  %  (Al VIII Kalpha)
  { case "Al7Ka": return [99130701]; }  %  (Al VII Kalpha)
  { case "Al6Ka": return [99130601]; }  %  (Al VI Kalpha)
  { case "Al5Ka": return [99130501]; }  %  (Al V Kalpha)
  { case "Al4Ka": return [99130401]; }  %  (Al IV Kalpha)
  { case "Al3Ka": return [99130301]; }  %  (Al III Kalpha)
  { case "Al2Ka": return [99130201]; }  %  (Al II Kalpha)
  { case "Si12Ka": return [99141201]; }  %  (Si XII Kalpha)
  { case "Si11Ka": return [99141101]; }  %  (Si XI Kalpha)
  { case "Si10Ka": return [99141001]; }  %  (Si X Kalpha)
  { case "Si9Ka": return [99140901]; }  %  (Si IX Kalpha)
  { case "Si8Ka": return [99140801]; }  %  (Si VIII Kalpha)
  { case "Si7Ka": return [99140701]; }  %  (Si VII Kalpha)
  { case "Si6Ka": return [99140601]; }  %  (Si VI Kalpha)
  { case "Si5Ka": return [99140501]; }  %  (Si V Kalpha)
  { case "Si4Ka": return [99140401]; }  %  (Si IV Kalpha)
  { case "Si3Ka": return [99140301]; }  %  (Si III Kalpha)
  { case "Si2Ka": return [99140201]; }  %  (Si II Kalpha)
  { case "Ar16Ka": return [99181601]; }  %  (Ar XVI Kalpha)
  { case "Ar15Ka": return [99181501]; }  %  (Ar XV Kalpha)
  { case "Ar14Ka": return [99181401]; }  %  (Ar XIV Kalpha)
  { case "Ar13Ka": return [99181301]; }  %  (Ar XIII Kalpha)
  { case "Ar12Ka": return [99181201]; }  %  (Ar XII Kalpha)
  { case "Ar11Ka": return [99181101]; }  %  (Ar XI Kalpha)
  { case "Ar10Ka": return [99181001]; }  %  (Ar X Kalpha)
  { case "Ar9Ka": return [99180901]; }  %  (Ar IX Kalpha)
  { case "Ar8Ka": return [99180801]; }  %  (Ar VIII Kalpha)
  { case "Ar7Ka": return [99180701]; }  %  (Ar VII Kalpha)
  { case "Ar6Ka": return [99180601]; }  %  (Ar VI Kalpha)
  { case "Ar5Ka": return [99180501]; }  %  (Ar V Kalpha)
  { case "Ar4Ka": return [99180401]; }  %  (Ar IV Kalpha)
  { case "Ar3Ka": return [99180301]; }  %  (Ar III Kalpha)
  { case "Ar2Ka": return [99180201]; }  %  (Ar II Kalpha)
  { vmessage("error (%s): line %s is unknown.", _function_name(), line); return; }
}


%%%%%%%%%%%%%%%%%%%%%%%
define get_line_lambdas()
%%%%%%%%%%%%%%%%%%%%%%%
{
  variable line;
  switch(_NARGS)
  { case 1: line = (); }
  { help(_function_name()); return; }

  switch(line)
  { case "O8a": return [18.968912]; }  %  (O VIII alpha)
  { case "O8b": return [16.005899]; }  %  (O VIII beta)
  { case "O8g": return [15.176164]; }  %  (O VIII gamma)
  { case "O8d": return [14.820574]; }  %  (O VIII delta)
  { case "O7f": return [22.097725]; }  %  (O VII f)
  { case "O7i": return [21.803638]; }  %  (O VII i)
  { case "O7a": return [21.601503]; }  %  (O VII alpha)
  { case "O7b": return [18.627001]; }  %  (O VII beta)
  { case "O7g": return [17.768000]; }  %  (O VII gamma)
  { case "O7d": return [17.396000]; }  %  (O VII delta)
  { case "Ne10a": return [12.133888]; }  %  (Ne X alpha)
  { case "Ne10b": return [10.238855]; }  %  (Ne X beta)
  { case "Ne10g": return [9.708176]; }  %  (Ne X gamma)
  { case "Ne10d": return [9.480745]; }  %  (Ne X delta)
  { case "Ne10e": return [9.361633]; }  %  (Ne X epsilon)
  { case "Ne10z": return [9.291233]; }  %  (Ne X zeta)
  { case "Ne9f": return [13.698976]; }  %  (Ne IX f)
  { case "Ne9i": return [13.553110]; }  %  (Ne IX i)
  { case "Ne9a": return [13.447307]; }  %  (Ne IX alpha)
  { case "Ne9b": return [11.544000]; }  %  (Ne IX beta)
  { case "Ne9g": return [11.001000]; }  %  (Ne IX gamma)
  { case "Ne9d": return [10.765000]; }  %  (Ne IX delta)
  { case "Ne9e": return [10.642600]; }  %  (Ne IX epsilon)
  { case "Ne9z": return [10.565000]; }  %  (Ne IX zeta)
  { case "Na11a": return [10.025000]; }  %  (Na XI alpha)
  { case "Na11b": return [8.459500]; }  %  (Na XI beta)
  { case "Na11g": return [8.021067]; }  %  (Na XI gamma)
  { case "Na11d": return [7.833167]; }  %  (Na XI delta)
  { case "Na10f": return [11.190000]; }  %  (Na X f)
  { case "Na10i": return [11.080000]; }  %  (Na X i)
  { case "Na10a": return [11.002700]; }  %  (Na X alpha)
  { case "Na10b": return [9.433000]; }  %  (Na X beta)
  { case "Na10g": return [8.982800]; }  %  (Na X gamma)
  { case "Na10d": return [8.788400]; }  %  (Na X delta)
  { case "Mg12a": return [8.421013]; }  %  (Mg XII alpha)
  { case "Mg12b": return [7.106155]; }  %  (Mg XII beta)
  { case "Mg12g": return [6.737902]; }  %  (Mg XII gamma)
  { case "Mg12d": return [6.580088]; }  %  (Mg XII delta)
  { case "Mg11f": return [9.314339]; }  %  (Mg XI f)
  { case "Mg11i": return [9.231208]; }  %  (Mg XI i)
  { case "Mg11a": return [9.168750]; }  %  (Mg XI alpha)
  { case "Mg11b": return [7.850300]; }  %  (Mg XI beta)
  { case "Mg11g": return [7.473000]; }  %  (Mg XI gamma)
  { case "Mg11d": return [7.310100]; }  %  (Mg XI delta)
  { case "Al13a": return [7.172800]; }  %  (Al XIII alpha)
  { case "Al13b": return [6.052967]; }  %  (Al XIII beta)
  { case "Al13g": return [5.739333]; }  %  (Al XIII gamma)
  { case "Al13d": return [5.604867]; }  %  (Al XIII delta)
  { case "Al12f": return [7.872118]; }  %  (Al XII f)
  { case "Al12i": return [7.806957]; }  %  (Al XII i)
  { case "Al12a": return [7.757301]; }  %  (Al XII alpha)
  { case "Al12b": return [6.635000]; }  %  (Al XII beta)
  { case "Al12g": return [6.314000]; }  %  (Al XII gamma)
  { case "Al12d": return [6.175000]; }  %  (Al XII delta)
  { case "Si14a": return [6.182241]; }  %  (Si XIV alpha)
  { case "Si14b": return [5.217208]; }  %  (Si XIV beta)
  { case "Si14g": return [4.946905]; }  %  (Si XIV gamma)
  { case "Si14d": return [4.831070]; }  %  (Si XIV delta)
  { case "Si13f": return [6.740294]; }  %  (Si XIII f)
  { case "Si13i": return [6.688187]; }  %  (Si XIII i)
  { case "Si13a": return [6.647947]; }  %  (Si XIII alpha)
  { case "Si13b": return [5.680500]; }  %  (Si XIII beta)
  { case "Si13g": return [5.404500]; }  %  (Si XIII gamma)
  { case "Si13d": return [5.285000]; }  %  (Si XIII delta)
  { case "S16a": return [4.729170]; }  %  (S XVI alpha)
  { case "S16b": return [3.991193]; }  %  (S XVI beta)
  { case "S16g": return [3.784465]; }  %  (S XVI gamma)
  { case "S16d": return [3.695877]; }  %  (S XVI delta)
  { case "S15f": return [5.101501]; }  %  (S XV f)
  { case "S15i": return [5.066492]; }  %  (S XV i)
  { case "S15a": return [5.038726]; }  %  (S XV alpha)
  { case "S15b": return [4.299000]; }  %  (S XV beta)
  { case "S15g": return [4.088300]; }  %  (S XV gamma)
  { case "S15d": return [3.998000]; }  %  (S XV delta)
  { case "Ar18a": return [3.732922]; }  %  (Ar XVIII alpha)
  { case "Ar18b": return [3.150621]; }  %  (Ar XVIII beta)
  { case "Ar18g": return [2.987481]; }  %  (Ar XVIII gamma)
  { case "Ar18d": return [2.917574]; }  %  (Ar XVIII delta)
  { case "Ar17f": return [3.994153]; }  %  (Ar XVII f)
  { case "Ar17i": return [3.969363]; }  %  (Ar XVII i)
  { case "Ar17a": return [3.949075]; }  %  (Ar XVII alpha)
  { case "Ar17b": return [3.365000]; }  %  (Ar XVII beta)
  { case "Ar17g": return [3.200000]; }  %  (Ar XVII gamma)
  { case "Ar17d": return [3.128000]; }  %  (Ar XVII delta)
  { case "Ca20a": return [3.020304]; }  %  (Ca XX alpha)
  { case "Ca20b": return [2.549359]; }  %  (Ca XX beta)
  { case "Ca20g": return [2.417399]; }  %  (Ca XX gamma)
  { case "Ca20d": return [2.360854]; }  %  (Ca XX delta)
  { case "Ca19f": return [3.211031]; }  %  (Ca XIX f)
  { case "Ca19i": return [3.192747]; }  %  (Ca XIX i)
  { case "Ca19a": return [3.177153]; }  %  (Ca XIX alpha)
  { case "Ca19b": return [2.705000]; }  %  (Ca XIX beta)
  { case "Ca19g": return [2.571000]; }  %  (Ca XIX gamma)
  { case "Ca19d": return [2.514000]; }  %  (Ca XIX delta)
  { case "Fe26a": return [1.779852]; }  %  (Fe XXVI alpha)
  { case "Fe26b": return [1.502752]; }  %  (Fe XXVI beta)
  { case "Fe26g": return [1.425067]; }  %  (Fe XXVI gamma)
  { case "Fe26d": return [1.391783]; }  %  (Fe XXVI delta)
  { case "Fe25f": return [1.868194]; }  %  (Fe XXV f)
  { case "Fe25i": return [1.859517]; }  %  (Fe XXV i)
  { case "Fe25a": return [1.850399]; }  %  (Fe XXV alpha)
  { case "Fe25b": return [1.573100]; }  %  (Fe XXV beta)
  { case "Fe25g": return [1.495000]; }  %  (Fe XXV gamma)
  { case "Fe25d": return [1.461000]; }  %  (Fe XXV delta)
  { case "Fe1Ka": return [1.937000]; }  %  (Fe I Kalpha)
  { case "Fe24_1062A": return [10.619000]; }  %  (161758)
  { case "Fe24_679A": return [6.788690]; }  %  (161756)
  { case "bl_1066A": return [10.663000, 10.657000]; }  %  (161753, 18204) = (Fe XXIV, Fe XVII)
  { case "Fe24_799A": return [7.985700, 7.996000]; }  %  (161759, 161754)
  { case "bl_1102A": return [11.019000, 11.026000]; }  %  (149890, 18163) = (Fe XXIII, Fe XVII)
  { case "bl_1099A": return [10.981000, 10.993500]; }  %  (149893, 131886) = (Fe XXIII, Fe XXII)
  { case "Fe23_830A": return [8.303800]; }  %  (149914)
  { case "Fe22_1225A": return [12.251900]; }  %  (131880)
  { case "Fe22_1143A": return [11.427000]; }  %  (132636)
  { case "Fe22_897A": return [8.974800]; }  %  (132642)
  { case "Fe22_1193A": return [11.932000]; }  %  (132890)
  { case "bl_1177A": return [11.770000, 11.762000]; }  %  (132626, 75499) = (Fe XXII, Fe XX)
  { case "Fe22_1149A": return [11.490000, 11.490000]; }  %  (131888, 132634)
  { case "Fe22_873A": return [8.725410, 8.736000]; }  %  (131902, 132644)
  { case "Fe22_786A": return [7.865000, 7.865000]; }  %  (131918, 132676)
  { case "Fe21_1228A": return [12.284000]; }  %  (128657)
  { case "Fe21_1198A": return [11.975000]; }  %  (128681)
  { case "Fe21_1304A": return [13.044400]; }  %  (128653)
  { case "bl_920A": return [9.194400, 9.197920, 9.197850]; }  %  (128775, 75070, 74587) = (Fe XXI, Fe XX, Fe XX)
  { case "Fe21_857A": return [8.574000, 8.574000, 8.574000]; }  %  (128847, 131702, 129682)
  { case "Fe21_1233A": return [12.327000]; }  %  (131540)
  { case "Fe20_1292A": return [12.912000, 12.921100]; }  %  (75469, 74908)
  { case "Fe20_1286A": return [12.846000, 12.864000]; }  %  (74900, 75472)
  { case "Fe20_1283A": return [12.824000, 12.827000]; }  %  (74354, 74360)
  { case "Fe20_1258A": return [12.576000, 12.576000]; }  %  (74924, 75496)
  { case "Fe20_1000A": return [9.997740, 10.000400, 9.993450, 10.005400]; }  %  (74457, 75036, 74463, 75579)
  { case "Fe20_1297A": return [12.965000, 12.965400]; }  %  (75461, 74351)
  { case "bl_1012A": return [10.120300, 10.119500, 10.121000]; }  %  (75587, 39003, 18189) = (Fe XX, Fe XIX, Fe XVII)
  { case "bl_1301A": return [12.992000, 13.022000]; }  %  (74897, 39144) = (Fe XX, Fe XIX)
  { case "Fe19_1466A": return [14.664000]; }  %  (39118)
  { case "Fe19_1380A": return [13.795000]; }  %  (39132)
  { case "Fe19_1364A": return [13.645000]; }  %  (39124)
  { case "Fe19_1350A": return [13.497000]; }  %  (38915)
  { case "Fe19_1346A": return [13.462000]; }  %  (38603)
  { case "Fe19_1342A": return [13.423000]; }  %  (39134)
  { case "Fe19_1294A": return [12.945000]; }  %  (38634)
  { case "Fe19_1082A": return [10.816000]; }  %  (39172)
  { case "Fe19_1013A": return [10.130900]; }  %  (38692)
  { case "Fe19_986A": return [9.855220]; }  %  (39223)
  { case "Fe19_1352A": return [13.518000, 13.514600]; }  %  (39128, 38611)
  { case "Fe19_1293A": return [12.931100, 12.933000]; }  %  (38941, 39146)
  { case "Fe18_1453A": return [14.534000]; }  %  (38000)
  { case "Fe18_1437A": return [14.373000]; }  %  (37997)
  { case "Fe18_1457A": return [14.571000]; }  %  (37942)
  { case "Fe18_1562A": return [15.625000]; }  %  (37995)
  { case "Fe18_1332A": return [13.323000]; }  %  (37835)
  { case "Fe18_1426A": return [14.256000, 14.256000]; }  %  (37830, 38002)
  { case "Fe18_1421A": return [14.208000, 14.208000, 14.200700]; }  %  (37944, 37998, 38112)
  { case "Fe18_1153A": return [11.527000, 11.527000]; }  %  (37961, 38014)
  { case "Fe18_1133A": return [11.326000, 11.326000, 11.326000]; }  %  (37845, 37963, 38012)
  { case "Fe17_1678A": return [16.780001]; }  %  (20125)
  { case "Fe17_1526A": return [15.261000]; }  %  (20128)
  { case "Fe17_1501A": return [15.014000]; }  %  (20127)
  { case "Fe17_1227A": return [12.266000]; }  %  (18155)
  { case "Fe17_1212A": return [12.124000]; }  %  (18151)
  { case "Fe17_1536A": return [15.359700]; }  %  (25066)
  { case "Fe17_1382A": return [13.825000]; }  %  (20130)
  { case "bl_1384A": return [13.839000, 13.843000]; }  %  (38918, 74885) = (Fe XIX, Fe XX)
  { case "Al11Ka": return [7.885000]; }  %  (Al XI Kalpha)
  { case "Al10Ka": return [7.964000]; }  %  (Al X Kalpha)
  { case "Al9Ka": return [8.050000]; }  %  (Al IX Kalpha)
  { case "Al8Ka": return [8.129000]; }  %  (Al VIII Kalpha)
  { case "Al7Ka": return [8.203000]; }  %  (Al VII Kalpha)
  { case "Al6Ka": return [8.269000]; }  %  (Al VI Kalpha)
  { case "Al5Ka": return [8.328000]; }  %  (Al V Kalpha)
  { case "Al4Ka": return [8.332000]; }  %  (Al IV Kalpha)
  { case "Al3Ka": return [8.336000]; }  %  (Al III Kalpha)
  { case "Al2Ka": return [8.339000]; }  %  (Al II Kalpha)
  { case "Si12Ka": return [6.750000]; }  %  (Si XII Kalpha)
  { case "Si11Ka": return [6.813000]; }  %  (Si XI Kalpha)
  { case "Si10Ka": return [6.882000]; }  %  (Si X Kalpha)
  { case "Si9Ka": return [6.947000]; }  %  (Si IX Kalpha)
  { case "Si8Ka": return [7.007000]; }  %  (Si VIII Kalpha)
  { case "Si7Ka": return [7.063000]; }  %  (Si VII Kalpha)
  { case "Si6Ka": return [7.112000]; }  %  (Si VI Kalpha)
  { case "Si5Ka": return [7.117000]; }  %  (Si V Kalpha)
  { case "Si4Ka": return [7.121000]; }  %  (Si IV Kalpha)
  { case "Si3Ka": return [7.124000]; }  %  (Si III Kalpha)
  { case "Si2Ka": return [7.126000]; }  %  (Si II Kalpha)
  { case "Ar16Ka": return [3.995000]; }  %  (Ar XVI Kalpha)
  { case "Ar15Ka": return [4.025000]; }  %  (Ar XV Kalpha)
  { case "Ar14Ka": return [4.057000]; }  %  (Ar XIV Kalpha)
  { case "Ar13Ka": return [4.089000]; }  %  (Ar XIII Kalpha)
  { case "Ar12Ka": return [4.119000]; }  %  (Ar XII Kalpha)
  { case "Ar11Ka": return [4.147000]; }  %  (Ar XI Kalpha)
  { case "Ar10Ka": return [4.174000]; }  %  (Ar X Kalpha)
  { case "Ar9Ka": return [4.178000]; }  %  (Ar IX Kalpha)
  { case "Ar8Ka": return [4.180000]; }  %  (Ar VIII Kalpha)
  { case "Ar7Ka": return [4.184000]; }  %  (Ar VII Kalpha)
  { case "Ar6Ka": return [4.186000]; }  %  (Ar VI Kalpha)
  { case "Ar5Ka": return [4.189000]; }  %  (Ar V Kalpha)
  { case "Ar4Ka": return [4.190000]; }  %  (Ar IV Kalpha)
  { case "Ar3Ka": return [4.192000]; }  %  (Ar III Kalpha)
  { case "Ar2Ka": return [4.193000]; }  %  (Ar II Kalpha)
  { message("error ("+_function_name()+"): line "+line+" is unknown."); return; }
}


%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_line_transition()
%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable line;
  switch(_NARGS)
  { case 1: line = (); }
  { help(_function_name()); return; }

  switch(line)
  { case "O8a": return [8, 8, 21]; }  %  (O VIII alpha)
  { case "O8b": return [8, 8, 22]; }  %  (O VIII beta)
  { case "O8g": return [8, 8, 23]; }  %  (O VIII gamma)
  { case "O8d": return [8, 8, 24]; }  %  (O VIII delta)
  { case "O7f": return [8, 7, 19]; }  %  (O VII f)
  { case "O7i": return [8, 7, 20]; }  %  (O VII i)
  { case "O7a": return [8, 7, 21]; }  %  (O VII alpha)
  { case "O7b": return [8, 7, 22]; }  %  (O VII beta)
  { case "O7g": return [8, 7, 23]; }  %  (O VII gamma)
  { case "O7d": return [8, 7, 24]; }  %  (O VII delta)
  { case "Ne10a": return [10, 10, 21]; }  %  (Ne X alpha)
  { case "Ne10b": return [10, 10, 22]; }  %  (Ne X beta)
  { case "Ne10g": return [10, 10, 23]; }  %  (Ne X gamma)
  { case "Ne10d": return [10, 10, 24]; }  %  (Ne X delta)
  { case "Ne10e": return [10, 10, 25]; }  %  (Ne X epsilon)
  { case "Ne10z": return [10, 10, 26]; }  %  (Ne X zeta)
  { case "Ne9f": return [10, 9, 19]; }  %  (Ne IX f)
  { case "Ne9i": return [10, 9, 20]; }  %  (Ne IX i)
  { case "Ne9a": return [10, 9, 21]; }  %  (Ne IX alpha)
  { case "Ne9b": return [10, 9, 22]; }  %  (Ne IX beta)
  { case "Ne9g": return [10, 9, 23]; }  %  (Ne IX gamma)
  { case "Ne9d": return [10, 9, 24]; }  %  (Ne IX delta)
  { case "Ne9e": return [10, 9, 25]; }  %  (Ne IX epsilon)
  { case "Ne9z": return [10, 9, 26]; }  %  (Ne IX zeta)
  { case "Na11a": return [11, 11, 21]; }  %  (Na XI alpha)
  { case "Na11b": return [11, 11, 22]; }  %  (Na XI beta)
  { case "Na11g": return [11, 11, 23]; }  %  (Na XI gamma)
  { case "Na11d": return [11, 11, 24]; }  %  (Na XI delta)
  { case "Na10f": return [11, 10, 19]; }  %  (Na X f)
  { case "Na10i": return [11, 10, 20]; }  %  (Na X i)
  { case "Na10a": return [11, 10, 21]; }  %  (Na X alpha)
  { case "Na10b": return [11, 10, 22]; }  %  (Na X beta)
  { case "Na10g": return [11, 10, 23]; }  %  (Na X gamma)
  { case "Na10d": return [11, 10, 24]; }  %  (Na X delta)
  { case "Mg12a": return [12, 12, 21]; }  %  (Mg XII alpha)
  { case "Mg12b": return [12, 12, 22]; }  %  (Mg XII beta)
  { case "Mg12g": return [12, 12, 23]; }  %  (Mg XII gamma)
  { case "Mg12d": return [12, 12, 24]; }  %  (Mg XII delta)
  { case "Mg11f": return [12, 11, 19]; }  %  (Mg XI f)
  { case "Mg11i": return [12, 11, 20]; }  %  (Mg XI i)
  { case "Mg11a": return [12, 11, 21]; }  %  (Mg XI alpha)
  { case "Mg11b": return [12, 11, 22]; }  %  (Mg XI beta)
  { case "Mg11g": return [12, 11, 23]; }  %  (Mg XI gamma)
  { case "Mg11d": return [12, 11, 24]; }  %  (Mg XI delta)
  { case "Al13a": return [13, 13, 21]; }  %  (Al XIII alpha)
  { case "Al13b": return [13, 13, 22]; }  %  (Al XIII beta)
  { case "Al13g": return [13, 13, 23]; }  %  (Al XIII gamma)
  { case "Al13d": return [13, 13, 24]; }  %  (Al XIII delta)
  { case "Al12f": return [13, 12, 19]; }  %  (Al XII f)
  { case "Al12i": return [13, 12, 20]; }  %  (Al XII i)
  { case "Al12a": return [13, 12, 21]; }  %  (Al XII alpha)
  { case "Al12b": return [13, 12, 22]; }  %  (Al XII beta)
  { case "Al12g": return [13, 12, 23]; }  %  (Al XII gamma)
  { case "Al12d": return [13, 12, 24]; }  %  (Al XII delta)
  { case "Si14a": return [14, 14, 21]; }  %  (Si XIV alpha)
  { case "Si14b": return [14, 14, 22]; }  %  (Si XIV beta)
  { case "Si14g": return [14, 14, 23]; }  %  (Si XIV gamma)
  { case "Si14d": return [14, 14, 24]; }  %  (Si XIV delta)
  { case "Si13f": return [14, 13, 19]; }  %  (Si XIII f)
  { case "Si13i": return [14, 13, 20]; }  %  (Si XIII i)
  { case "Si13a": return [14, 13, 21]; }  %  (Si XIII alpha)
  { case "Si13b": return [14, 13, 22]; }  %  (Si XIII beta)
  { case "Si13g": return [14, 13, 23]; }  %  (Si XIII gamma)
  { case "Si13d": return [14, 13, 24]; }  %  (Si XIII delta)
  { case "S16a": return [16, 16, 21]; }  %  (S XVI alpha)
  { case "S16b": return [16, 16, 22]; }  %  (S XVI beta)
  { case "S16g": return [16, 16, 23]; }  %  (S XVI gamma)
  { case "S16d": return [16, 16, 24]; }  %  (S XVI delta)
  { case "S15f": return [16, 15, 19]; }  %  (S XV f)
  { case "S15i": return [16, 15, 20]; }  %  (S XV i)
  { case "S15a": return [16, 15, 21]; }  %  (S XV alpha)
  { case "S15b": return [16, 15, 22]; }  %  (S XV beta)
  { case "S15g": return [16, 15, 23]; }  %  (S XV gamma)
  { case "S15d": return [16, 15, 24]; }  %  (S XV delta)
  { case "Ar18a": return [18, 18, 21]; }  %  (Ar XVIII alpha)
  { case "Ar18b": return [18, 18, 22]; }  %  (Ar XVIII beta)
  { case "Ar18g": return [18, 18, 23]; }  %  (Ar XVIII gamma)
  { case "Ar18d": return [18, 18, 24]; }  %  (Ar XVIII delta)
  { case "Ar17f": return [18, 17, 19]; }  %  (Ar XVII f)
  { case "Ar17i": return [18, 17, 20]; }  %  (Ar XVII i)
  { case "Ar17a": return [18, 17, 21]; }  %  (Ar XVII alpha)
  { case "Ar17b": return [18, 17, 22]; }  %  (Ar XVII beta)
  { case "Ar17g": return [18, 17, 23]; }  %  (Ar XVII gamma)
  { case "Ar17d": return [18, 17, 24]; }  %  (Ar XVII delta)
  { case "Ca20a": return [20, 20, 21]; }  %  (Ca XX alpha)
  { case "Ca20b": return [20, 20, 22]; }  %  (Ca XX beta)
  { case "Ca20g": return [20, 20, 23]; }  %  (Ca XX gamma)
  { case "Ca20d": return [20, 20, 24]; }  %  (Ca XX delta)
  { case "Ca19f": return [20, 19, 19]; }  %  (Ca XIX f)
  { case "Ca19i": return [20, 19, 20]; }  %  (Ca XIX i)
  { case "Ca19a": return [20, 19, 21]; }  %  (Ca XIX alpha)
  { case "Ca19b": return [20, 19, 22]; }  %  (Ca XIX beta)
  { case "Ca19g": return [20, 19, 23]; }  %  (Ca XIX gamma)
  { case "Ca19d": return [20, 19, 24]; }  %  (Ca XIX delta)
  { case "Fe26a": return [26, 26, 21]; }  %  (Fe XXVI alpha)
  { case "Fe26b": return [26, 26, 22]; }  %  (Fe XXVI beta)
  { case "Fe26g": return [26, 26, 23]; }  %  (Fe XXVI gamma)
  { case "Fe26d": return [26, 26, 24]; }  %  (Fe XXVI delta)
  { case "Fe25f": return [26, 25, 19]; }  %  (Fe XXV f)
  { case "Fe25i": return [26, 25, 20]; }  %  (Fe XXV i)
  { case "Fe25a": return [26, 25, 21]; }  %  (Fe XXV alpha)
  { case "Fe25b": return [26, 25, 22]; }  %  (Fe XXV beta)
  { case "Fe25g": return [26, 25, 23]; }  %  (Fe XXV gamma)
  { case "Fe25d": return [26, 25, 24]; }  %  (Fe XXV delta)
  { case "Fe1Ka": return [26, 1, 1]; }  %  (Fe I Kalpha)
  { case "Fe24_1062A": return [26, 24, -161758]; }  %  (161758)
  { case "Fe24_679A": return [26, 24, -161756]; }  %  (161756)
  { case "bl_1066A": return NULL; }  %  (161753, 18204) = (Fe XXIV, Fe XVII)
  { case "Fe24_799A": return NULL; }  %  (161759, 161754)
  { case "bl_1102A": return NULL; }  %  (149890, 18163) = (Fe XXIII, Fe XVII)
  { case "bl_1099A": return NULL; }  %  (149893, 131886) = (Fe XXIII, Fe XXII)
  { case "Fe23_830A": return [26, 23, -149914]; }  %  (149914)
  { case "Fe22_1225A": return [26, 22, -131880]; }  %  (131880)
  { case "Fe22_1143A": return [26, 22, -132636]; }  %  (132636)
  { case "Fe22_897A": return [26, 22, -132642]; }  %  (132642)
  { case "Fe22_1193A": return [26, 22, -132890]; }  %  (132890)
  { case "bl_1177A": return NULL; }  %  (132626, 75499) = (Fe XXII, Fe XX)
  { case "Fe22_1149A": return NULL; }  %  (131888, 132634)
  { case "Fe22_873A": return NULL; }  %  (131902, 132644)
  { case "Fe22_786A": return NULL; }  %  (131918, 132676)
  { case "Fe21_1228A": return [26, 21, -128657]; }  %  (128657)
  { case "Fe21_1198A": return [26, 21, -128681]; }  %  (128681)
  { case "Fe21_1304A": return [26, 21, -128653]; }  %  (128653)
  { case "bl_920A": return NULL; }  %  (128775, 75070, 74587) = (Fe XXI, Fe XX, Fe XX)
  { case "Fe21_857A": return NULL; }  %  (128847, 131702, 129682)
  { case "Fe21_1233A": return [26, 21, -131540]; }  %  (131540)
  { case "Fe20_1292A": return NULL; }  %  (75469, 74908)
  { case "Fe20_1286A": return NULL; }  %  (74900, 75472)
  { case "Fe20_1283A": return NULL; }  %  (74354, 74360)
  { case "Fe20_1258A": return NULL; }  %  (74924, 75496)
  { case "Fe20_1000A": return NULL; }  %  (74457, 75036, 74463, 75579)
  { case "Fe20_1297A": return NULL; }  %  (75461, 74351)
  { case "bl_1012A": return NULL; }  %  (75587, 39003, 18189) = (Fe XX, Fe XIX, Fe XVII)
  { case "bl_1301A": return NULL; }  %  (74897, 39144) = (Fe XX, Fe XIX)
  { case "Fe19_1466A": return [26, 19, -39118]; }  %  (39118)
  { case "Fe19_1380A": return [26, 19, -39132]; }  %  (39132)
  { case "Fe19_1364A": return [26, 19, -39124]; }  %  (39124)
  { case "Fe19_1350A": return [26, 19, -38915]; }  %  (38915)
  { case "Fe19_1346A": return [26, 19, -38603]; }  %  (38603)
  { case "Fe19_1342A": return [26, 19, -39134]; }  %  (39134)
  { case "Fe19_1294A": return [26, 19, -38634]; }  %  (38634)
  { case "Fe19_1082A": return [26, 19, -39172]; }  %  (39172)
  { case "Fe19_1013A": return [26, 19, -38692]; }  %  (38692)
  { case "Fe19_986A": return [26, 19, -39223]; }  %  (39223)
  { case "Fe19_1352A": return NULL; }  %  (39128, 38611)
  { case "Fe19_1293A": return NULL; }  %  (38941, 39146)
  { case "Fe18_1453A": return [26, 18, -38000]; }  %  (38000)
  { case "Fe18_1437A": return [26, 18, -37997]; }  %  (37997)
  { case "Fe18_1457A": return [26, 18, -37942]; }  %  (37942)
  { case "Fe18_1562A": return [26, 18, -37995]; }  %  (37995)
  { case "Fe18_1332A": return [26, 18, -37835]; }  %  (37835)
  { case "Fe18_1426A": return NULL; }  %  (37830, 38002)
  { case "Fe18_1421A": return NULL; }  %  (37944, 37998, 38112)
  { case "Fe18_1153A": return NULL; }  %  (37961, 38014)
  { case "Fe18_1133A": return NULL; }  %  (37845, 37963, 38012)
  { case "Fe17_1678A": return [26, 17, -20125]; }  %  (20125)
  { case "Fe17_1526A": return [26, 17, -20128]; }  %  (20128)
  { case "Fe17_1501A": return [26, 17, -20127]; }  %  (20127)
  { case "Fe17_1227A": return [26, 17, -18155]; }  %  (18155)
  { case "Fe17_1212A": return [26, 17, -18151]; }  %  (18151)
  { case "Fe17_1536A": return [26, 17, -25066]; }  %  (25066)
  { case "Fe17_1382A": return [26, 17, -20130]; }  %  (20130)
  { case "bl_1384A": return NULL; }  %  (38918, 74885) = (Fe XIX, Fe XX)
  { case "Al11Ka": return [13, 11, 1]; }  %  (Al XI Kalpha)
  { case "Al10Ka": return [13, 10, 1]; }  %  (Al X Kalpha)
  { case "Al9Ka": return [13, 9, 1]; }  %  (Al IX Kalpha)
  { case "Al8Ka": return [13, 8, 1]; }  %  (Al VIII Kalpha)
  { case "Al7Ka": return [13, 7, 1]; }  %  (Al VII Kalpha)
  { case "Al6Ka": return [13, 6, 1]; }  %  (Al VI Kalpha)
  { case "Al5Ka": return [13, 5, 1]; }  %  (Al V Kalpha)
  { case "Al4Ka": return [13, 4, 1]; }  %  (Al IV Kalpha)
  { case "Al3Ka": return [13, 3, 1]; }  %  (Al III Kalpha)
  { case "Al2Ka": return [13, 2, 1]; }  %  (Al II Kalpha)
  { case "Si12Ka": return [14, 12, 1]; }  %  (Si XII Kalpha)
  { case "Si11Ka": return [14, 11, 1]; }  %  (Si XI Kalpha)
  { case "Si10Ka": return [14, 10, 1]; }  %  (Si X Kalpha)
  { case "Si9Ka": return [14, 9, 1]; }  %  (Si IX Kalpha)
  { case "Si8Ka": return [14, 8, 1]; }  %  (Si VIII Kalpha)
  { case "Si7Ka": return [14, 7, 1]; }  %  (Si VII Kalpha)
  { case "Si6Ka": return [14, 6, 1]; }  %  (Si VI Kalpha)
  { case "Si5Ka": return [14, 5, 1]; }  %  (Si V Kalpha)
  { case "Si4Ka": return [14, 4, 1]; }  %  (Si IV Kalpha)
  { case "Si3Ka": return [14, 3, 1]; }  %  (Si III Kalpha)
  { case "Si2Ka": return [14, 2, 1]; }  %  (Si II Kalpha)
  { case "Ar16Ka": return [18, 16, 1]; }  %  (Ar XVI Kalpha)
  { case "Ar15Ka": return [18, 15, 1]; }  %  (Ar XV Kalpha)
  { case "Ar14Ka": return [18, 14, 1]; }  %  (Ar XIV Kalpha)
  { case "Ar13Ka": return [18, 13, 1]; }  %  (Ar XIII Kalpha)
  { case "Ar12Ka": return [18, 12, 1]; }  %  (Ar XII Kalpha)
  { case "Ar11Ka": return [18, 11, 1]; }  %  (Ar XI Kalpha)
  { case "Ar10Ka": return [18, 10, 1]; }  %  (Ar X Kalpha)
  { case "Ar9Ka": return [18, 9, 1]; }  %  (Ar IX Kalpha)
  { case "Ar8Ka": return [18, 8, 1]; }  %  (Ar VIII Kalpha)
  { case "Ar7Ka": return [18, 7, 1]; }  %  (Ar VII Kalpha)
  { case "Ar6Ka": return [18, 6, 1]; }  %  (Ar VI Kalpha)
  { case "Ar5Ka": return [18, 5, 1]; }  %  (Ar V Kalpha)
  { case "Ar4Ka": return [18, 4, 1]; }  %  (Ar IV Kalpha)
  { case "Ar3Ka": return [18, 3, 1]; }  %  (Ar III Kalpha)
  { case "Ar2Ka": return [18, 2, 1]; }  %  (Ar II Kalpha)
  { message("error ("+_function_name()+"): line "+line+" is unknown."); return; }
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define _get_lambda_parameters_of_lines()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable ids=1, lMin=get_plot_options().xmin, lMax=get_plot_options().xmax, all=0;
  switch(_NARGS)
  { case 0: ; }
  { case 1: ids = (); }
  { case 2: (lMin, lMax) = (); }
  { case 3: (ids, lMin, lMax) = (); }
  { case 4: (ids, lMin, lMax, all) = (); }
  { return; }

  if(lMin>lMax) { (lMin, lMax) = (lMax, lMin); }

  variable par_ids = Integer_Type[0];
  variable par_info, par_infos = get_params();
  foreach par_info (par_infos)
  {
    variable id;
    foreach id ([ids])
    {
      if(    string_match(par_info.name, "^lines("+string(id)+")\..*_lam", 1)
         && ((lMin <= par_info.min <= lMax) or (lMin<=par_info.max<=lMax))
         && (all || get_par(par_info.index+1)!=0)
        )
      { par_ids = [par_ids, par_info.index]; }
    }
  }
  if(length(par_ids)==0) { return NULL; }

  return par_ids[ array_sort(par_ids, &compare_pars) ];
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_lambda_parameters_of_lines()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_lambda_parameters_of_lines}
%!%-
{
  variable ids=1, lMin=get_plot_options().xmin, lMax=get_plot_options().xmax;
  switch(_NARGS)
  { case 0: ; }
  { case 1: ids = (); }
  { case 2: (lMin, lMax) = (); }
  { case 3: (ids, lMin, lMax) = (); }
  { help(_function_name()); return; }

  return _get_lambda_parameters_of_lines(ids, lMin, lMax, 0);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_lambda_parameters_of_all_lines()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_lambda_parameters_of_all_lines}
%!%-
{
  variable ids=1, lMin=get_plot_options().xmin, lMax=get_plot_options().xmax;
  switch(_NARGS)
  { case 0: ; }
  { case 1: ids = (); }
  { case 2: (lMin, lMax) = (); }
  { case 3: (ids, lMin, lMax) = (); }
  { help(_function_name()); return; }

  return _get_lambda_parameters_of_lines(ids, lMin, lMax, 1);
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_lambda_parameters_of_lines_from_one_ion(ion)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_lambda_parameters_of_lines_from_one_ion}
%!%-
{
 return get_lambda_parameters_of_lines[where(array_map(Char_Type, &string_match, array_struct_field( get_params(get_lambda_parameters_of_lines), "name"), ion, 1) )];
}


%%%%%%%%%%%%%%%%%%%%%
define list_all_lines()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{lists_all_lines}
%\synopsis{shows the parameters of a lines-model's lines in a given wavelength range}
%\usage{list_all_lines([ids][, lMin, lMax]);}
%\seealso{lines, get_lambda_parameters_of_all_lines, list_lines}
%!%-
{
  variable ids=1, lMin=get_plot_options().xmin, lMax=get_plot_options().xmax;
  switch(_NARGS)
  { case 0: ; }
  { case 1: ids = (); }
  { case 2: (lMin, lMax) = (); }
  { case 3: (ids, lMin, lMax) = (); }
  { help(_function_name()); return; }

  variable par_ids = get_lambda_parameters_of_all_lines(ids, lMin, lMax);
  variable par_id, first=1;
  foreach par_id (par_ids)
  { if(first) { first = 0; } else { message(""); }
    list_Par( par_id+[0,1,2] );
  }
}


%%%%%%%%%%%%%%%%%
define list_lines()
%%%%%%%%%%%%%%%%%
%!%+
%\function{list_lines}
%\synopsis{shows the parameters of a lines-model's used lines in a given wavelength range}
%\usage{list_lines([ids][, lMin, lMax]);}
%\seealso{lines, get_lambda_parameters_of_lines, list_all_lis}
%!%-
{
  variable ids=Integer_Type[0], lMin=get_plot_options().xmin, lMax=get_plot_options().xmax;
  switch(_NARGS)
  { case 0: ; }
  { case 1: ids = (); }
  { case 2: (lMin, lMax) = (); }
  { case 3: (ids, lMin, lMax) = (); }
  { help(_function_name()); return; }

  if(length(ids)==0)
  {
    variable ff = get_fit_fun();
    forever
    {
      variable m = string_matches(ff, `lines(\([0-9]*\))\(.*\)`);
      if(m==NULL)  break;
      ids = [ids, integer(m[1])];
      ff = m[2];
    }
  }

  variable par_ids = get_lambda_parameters_of_lines(ids, lMin, lMax);
  variable par_id, first=1;
  foreach par_id (par_ids)
  { if(first) { first = 0; } else { message(""); }
    list_Par( par_id+[0,1,2] );
  }
}


%%%%%%%%%%%%%%%%%%%%%%
define get_line_labels()
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_line_labels}
%\qualifiers{
%\qualifier{label_Ly_He}{ [=1]}
%\qualifier{print_list}{}
%}
%\seealso{lines}
%!%-
{
  variable lambdas = Double_Type[0], labels = String_Type[0], Z = Char_Type[0], ion = Char_Type[0], nr = Char_Type[0];

  foreach ( get_lambda_parameters_of_lines() )
  { variable line = ();
    line = string_matches(get_par_info(line).name, `lines([0-9]+)\.\(.*\)_lam`)[1];
    variable tr = get_line_transition( line );
    variable lab;
    if(tr!=NULL)
    {
      Z   = [Z,   tr[0]];
      ion = [ion, tr[1]];
      nr  = [nr,  tr[2]];
      if(tr[2]>0)
      {
        lab = sprintf("%s %s %s", atom_name(tr[0]), Roman(tr[1]), ext_line_type(tr[2], 1));
        switch(tr[1])
        { case tr[0]:
            if(qualifier("label_Ly_He", 1))
            { lab = sprintf("%s Ly %s", atom_name(tr[0]), ext_line_type(tr[2], 1)); }
        }
        { case (tr[0]-1):
            if(qualifier("label_Ly_He", 1))
            { lab = sprintf("%s He %s", atom_name(tr[0]), ext_line_type(tr[2], 1)); }
        }
      }
      else
      { lab = sprintf("%s %s", atom_name(tr[0]), Roman(tr[1])); }
    }
    else
    {
      lab = string_matches(line, `\([a-zA-Z0-9]+\)`)[1];
      Z   = [Z,   -1];
      ion = [ion, -1];
      nr  = [nr,  -1];
      if(lab == "bl")
        lab = "blend";
      else
        lab = string_matches(lab, `\([a-zA-Z]+\)`)[1] + " " + Roman( integer(string_matches(lab, `\([0-9]+\)`)[1]) );
    }

    lambdas = [lambdas, mean(get_line_lambdas(line))];
    labels  = [labels,  lab];
  }

  if(qualifier_exists("print_list"))
  { _for $1 (0, length(lambdas)-1, 1)
    { ()=printf(" { %.4f, \"%s\", %d, %d, %d },\n", lambdas[$1], labels[$1], Z[$1], ion[$1], nr[$1]); }
  }
  else
    return lambdas, labels;
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%
define Doppler_velocity()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{Doppler_velocity}
%\synopsis{calculates a Doppler velocity shift from a wavelength and the rest wavelength}
%\usage{Double_Type v = Doppler_velocity(Double_Type lambda, Double_Type lambda0);}
%\description
%    \code{v = (lambda-lambda0)/lambda_0 * c;  %} speed of light \code{c = 299792} km/s
%\seealso{get_line_velocity}
%!%-
{
  variable lambda, lambda0;
  switch(_NARGS)
  { case 2: (lambda, lambda0) = (); }
  { help(_function_name()); return; }

   return (lambda/lambda0-1)*299792.;
}


%%%%%%%%%%%%%%%%%%%%%%%
define get_line_velocity()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_line_velocity}
%\synopsis{calculates the velocity shift in a line of the lines-model}
%\usage{Double_Type v = get_line_velocity([Integer_Type i,] String_Type line);}
%\description
%    \code{lambda = get_par("lines(i).line_lam");}\n
%    \code{lambda0 = mean( get_line_lambdas(line) );}\n
%    \code{v = (lambda-lambda0)/lambda_0 * c;  %} speed of light \code{c = 299792} km/s
%\seealso{Doppler_velocity}
%!%-
{
  variable i=1, line;
  switch(_NARGS)
  { case 1: line = (); }
  { case 2: (i, line) = (); }
  { help(_function_name()); return; }

  variable parameter_name = "lines(" + string(i) + ")." + line  + "_lam";
  if( howmany(array_struct_field(get_params(), "name")==parameter_name) == 0 )  { vmessage("error (%s): line %s does not exist within the current model", _function_name(), line); return NULL; }
  variable lambda = get_par(parameter_name);
  variable lambda0 = mean( get_line_lambdas(line) );
  return (lambda/lambda0-1)*299792.;
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define get_column_density_from_line()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_column_density_from_line}
%\synopsis{}
%\usage{Double_Type N = get_column_density_from_line([Integer_Type i,] String_Type line);}
%\description
%     N  =  mc^2/[pi e^2] * W_lambda / [f_lu * lambda^2];\n
%        =  1.13e17/cm^2 * (W_lambda/mA) / [f_lu * (lambda/A)^2]\n
%                                        %  f_lu = mc / [8 pi^2 e^2] * lambda^2 * g_u / g_l * A_ul\n
%                                        %       = 1.4992e-16 * (lambda/A)^2 * g_u / g_l * (A_ul/s^{-1})\n
%!%-
{
  variable i=1, line;
  switch(_NARGS)
  { case 1: line = (); }
  { case 2: (i, line) = (); }
  { help(_function_name()); return; }

  variable parameter_name = "lines("+string(i)+")."+line+"_EW";
  if( howmany(array_struct_field(get_params(), "name")==parameter_name) == 0 )  { vmessage("error (%s): %s does not exist within the current model", _function_name(), line); return NULL; }
  variable EW = get_par("lines("+string(i)+")."+line+"_EW");
  if(EW>=0) { vmessage("warning (%s): %s (#%d) is not an absorption line", _function_name(), line, i); return 0; }
  variable transition = get_line_transition(line);
  if(transition == NULL)  { vmessage("warning (%s): transition(s) causing %s unknown", _function_name(), line); return NULL; }
  variable info = ext_line_info(transition[0], transition[1], transition[2]);
  if(info.gf == NULL)  { vmessage("warning (%s): atomic-data of %s unknown", _function_name(), line); return NULL; }
  variable lambda0 = info.lambda;
  variable f_ij = info.gf/info.g_low;
  return -1.13e17/f_ij/lambda0^2 * EW;
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%
define HeTriplet_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%%%%%
{
  return 0*bin_lo;  % additive dummy-model
}

%%%%%%%%%%%%%%%%%%%%
define HeTriplet_defaults(i)
%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
  { case 0: return (1.5, 0, 0, 5); }
  { case 1: return (1.5, 0, 0, 5); }
}

add_slang_function("HeTriplet", ["Mg11_R", "S15_R"]);
set_param_default_hook("HeTriplet", "HeTriplet_defaults");


%%%%%%%%%%%%%%%%%%%%
define use_HeTriplet(id, ion)
%%%%%%%%%%%%%%%%%%%%
{
  id = "(" + string(id) + ")." + ion;
  variable i = get_par("lines" + id + "i_EW");
  if(i!=0) { set_par("HeTriplet" + id + "_R",  get_par("lines" + id + "f_EW") / i); }  % R = f/i
  set_par("lines" + id + "f_EW",  "lines" + id + "i_EW * HeTriplet" + id + "_R");      % f = i * R
}

%
% Code from src/sixte/lissajous.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define lissajous_pattern(){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{lissajous_pattern}
%\synopsis{calculates a Lissajous pattern for 3:4 ratio with PI/4 offset}
%\usage{variable li = lissajous_pattern();}
%\qualifiers{
%\qualifier{amplitude}{[1.0]: amplitude by default from -1 to 1}
%\qualifier{tstart}{[0.0]: tstart}
%\qualifier{tstop}{[1.0]: tstop}
%\qualifier{x0}{[0.0]: x0}
%\qualifier{y0}{[0.0]: y0}
%}
%\description
%    An ARF is built from a mirror area, filter-,
%    support grid-, detector layer transmission
%    information. The transmission files-parameters
%    have to be file names of transmission tables 
%    in the format:
%
%!%-

   variable ampl = qualifier("amplitude",1.0);
   variable tstart = qualifier("tstart",0.0);
   variable tstop  = qualifier("tstop",1.0);
   
   variable a = 3/4.;
   variable b = 4/4.;

   variable offset=PI/4.;
   
   variable t = [0:1.0:#1000];

   variable x = sin(a*t*2*PI*4+offset)*ampl;
   variable y = sin(b*t*2*PI*4)*ampl;

   t *= tstop;
   t += tstart;

   x+= qualifier("x0",0.0);
   y+= qualifier("y0",0.0);
   
   return struct{t=t,x=x,y=y};
}

%
% Code from src/fitting/parameters/list_Par.sl
%
define list_Par()
%!%+
%\function{list_Par}
%\synopsis{lists parameters of the current fit-function as ISIS-commands}
%\usage{list_Par([ pars[] ]);}
%\qualifiers{
%\qualifier{fit_fun}{lists also the fit-function}
%\qualifier{fmt}{format statement for min/max/value}
%}
%!%-
{
  variable pars;
  switch(_NARGS)
  { case 0: pars = freeParameters(); }
  { % else:  % collect all arguments
      pars = __pop_list(_NARGS);
  }

  % convert String_Type arguments to indices of matching parameters
  variable p, par, param, params = get_params();
  variable n = length(params);
  variable list_of_pars = {};
  foreach par (pars)
    switch(_typeof(par))
    { case String_Type:
	foreach p ([par])
          foreach param (params)
            if(string_match(param.name, glob_to_regexp(p), 1))
              list_append(list_of_pars, param.index);
    }
    { case Integer_Type:
	foreach p (par)
	  if(0 < p <= n)
            list_append(list_of_pars, p);
          else
            vmessage("warning (%s): ignoring parameter %d", _function_name(), p);
    }
    { % else:
        vmessage("warning (%s): ignoring %S argument %S", _function_name(), typeof(par), par);
    }

  variable fit_fun = get_fit_fun();
  if(qualifier_exists("fit_fun"))
    vmessage(`fit_fun( "%s" );`, (fit_fun==NULL ? "" : fit_fun));

  n = length(list_of_pars);
  if(n==0 || fit_fun==NULL)  return;

  variable fmt = qualifier("fmt", "%.8g");

  variable names = String_Type[n],  values  = String_Type[n],
           freezes = Char_Type[n],
           mins = String_Type[n],  maxs = String_Type[n],
           steps = String_Type[n], relsteps = String_Type[n],
           ties = String_Type[n],  funs = String_Type[n],
           indxs = String_Type[n],  units = String_Type[n];
  _for p (0, n-1, 1)
  { variable info = get_par_info(list_of_pars[p]);
      names[p] = info.name;
     values[p] = sprintf(fmt, info.value);
    freezes[p] = info.freeze;
       mins[p] = (info.min==-DOUBLE_MAX ? "-DOUBLE_MAX" : sprintf(fmt, info.min));
       maxs[p] = (info.max== DOUBLE_MAX ?  "DOUBLE_MAX" : sprintf(fmt, info.max));
      steps[p] = sprintf(fmt, info.step);
   relsteps[p] = sprintf(fmt, info.relstep);
       ties[p] = (info.tie==NULL ? "" : info.tie);
       funs[p] = (info.fun==NULL ? "" : info.fun);
      indxs[p] = sprintf("%d", info.index);
      units[p] = info.units;
  }

  variable max_chars_name    = max(array_map(Integer_Type, &strlen, names));
  variable max_chars_value   = max(array_map(Integer_Type, &strlen, values));
  variable max_chars_min     = max(array_map(Integer_Type, &strlen, mins));
  variable max_chars_max     = max(array_map(Integer_Type, &strlen, maxs));
  variable max_chars_step    = max(array_map(Integer_Type, &strlen, steps));
  variable max_chars_relstep = max(array_map(Integer_Type, &strlen, relsteps));
  variable max_chars_unit    = max(array_map(Integer_Type, &strlen, units));
  variable max_chars_indx    = max(array_map(Integer_Type, &strlen, indxs));
  _for p (0, n-1, 1)
  {
    vmessage(`set_par( "%s",%s %s%s,  %d,  %s%s, %s%s; step=%s%s, relstep=%s%s); %% %s%s %s(par # %s)`,
              names[p], multiple_string(max_chars_name-strlen(names[p]), " "),
              multiple_string(max_chars_value-strlen(values[p]), " "), values[p],
              freezes[p],
              multiple_string(max_chars_min-strlen(mins[p]), " "), mins[p],
              multiple_string(max_chars_max-strlen(maxs[p]), " "), maxs[p],
              multiple_string(max_chars_step-strlen(steps[p]), " "), steps[p],
              multiple_string(max_chars_relstep-strlen(relsteps[p]), " "), relsteps[p],
              units[p], multiple_string(max_chars_unit-strlen(units[p]), " "),
              multiple_string(max_chars_indx-strlen(indxs[p]), " "), indxs[p]
	     );
    if(ties[p]!="")  vmessage(`tie( "%s", "%s");`, ties[p], names[p]);
    if(funs[p]!="")  vmessage(`set_par( "%s", "%s");`, names[p], funs[p]);
  }
}

%
% Code from src/data/spectrum.loaddata/loadDataset.sl
%
variable rmf_absolute_paths = String_Type[0];

%%%%%%%%%%%%%%%%%%
define loadDataset()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{loadDataset}
%\synopsis{loads and assigns one spectrum and its detector response}
%\usage{Integer_Type id = loadDataset(phaFile[, rmfFile[, arfFile[, backFile[, row]]]]);}
%\description
%
%!%-
{
  % process arguments
  %%%%%%%%%%%%%%%%%%%
  variable phaFile=NULL, rmfFile=NULL, arfFile=NULL, backFile=NULL, row=NULL, systematicUncert=0;
  switch(_NARGS)
  { case 1:  phaFile = (); }
  { case 2: (phaFile, rmfFile) = (); }
  { case 3: (phaFile, rmfFile, arfFile) = (); }
  { case 4: (phaFile, rmfFile, arfFile, backFile) = (); }
  { case 5: (phaFile, rmfFile, arfFile, backFile, row) = (); }
  { case 6: (phaFile, rmfFile, arfFile, backFile, row, systematicUncert) = (); }
  { help(_function_name()); return; }

  % load data
  %%%%%%%%%%%
  variable spec;
  if(row!=NULL) { spec = load_data(phaFile, row); } else  { spec = load_data(phaFile); }
  % save infos
  variable info = get_data_info(spec);
  variable exposure = get_data_exposure(spec);

  if(fits_key_exists(phaFile, "BACKFILE"))
  {
    variable FITS_backFile = fits_read_key(phaFile, "BACKFILE");
    if(FITS_backFile != ""  and  FITS_backFile != "none")
    {
      if(backFile!=NULL)
      { message("warning (loadDataset): background loaded via FITS-header keyword; argument of loadDataset ignored"); }
      backFile = FITS_backFile;
    }
  }

  if(backFile != NULL)
  { % load background
    if(fits_key_exists(phaFile, "BACKSCAL") and fits_key_exists(phaFile, "BACKSCUP") and fits_key_exists(phaFile, "BACKSCDN"))
    { % define_back cannot be used for type 2 PHA files
      variable back;
      if(row!=NULL) { back = load_data(backFile, row); } else { back = load_data(backFile); }
      variable s = get_data_counts(spec);
      variable b = get_data_counts(back);
      variable bp = fits_read_key(phaFile, "BACKSCAL"); % background scale
      variable bbu= fits_read_key(phaFile, "BACKSCUP");
      variable bbd= fits_read_key(phaFile, "BACKSCDN");
      variable scale = bp/(bbu+bbd);
      s.value -= b.value*scale;
      variable err = sqrt(s.value + scale*scale*b.value);
      delete_data(spec);
      delete_data(back);
      spec = define_counts(s.bin_lo, s.bin_hi, s.value, err);
      set_data_info(spec, info);
      set_data_exposure(spec, exposure);
    }
    else
    { ()=define_back(spec, backFile);
    }
  }
  set_rebin_error_hook(spec, "rebin_error_hook");

  % load response: ARF & RMF = RSP
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  variable arf=NULL, rmf=NULL;

%  print(info);

  if(info.rmfs[0]>0)
  { if(rmfFile!=NULL)
    { message("warning (loadDataset): RMF already loaded via FITS-header keywords; argument of loadDataset ignored");
      rmfFile = NULL;
    }
    rmf = info.rmfs[0];
    unassign_rmf(rmf);  % will be assigned at the end
  }
  if(info.arfs[0]>0)
  { if(arfFile != NULL)
    { message("warning (loadDataset): ARF already loaded via FITS-header keywords; argument of loadDataset ignored");
      arfFile = NULL;
    }
    arf = info.arfs[0];
    unassign_arf(arf);  % will be assigned at the end
  }

  % loadRMF
  %%%%%%%%%
  if(rmfFile != NULL)
  { rmf = load_rmf(rmfFile); }

  % use absolute paths
  variable rmf_info = get_rmf_info(rmf);
  if(rmf_info.arg_string[0] != '/')
  { rmf_info.arg_string = getcwd() + rmf_info.arg_string;
    set_rmf_info(rmf, rmf_info); % does not (yet at ISIS 1.4.9-13) update arg_string due to src/rmf.c:Rmf_set_info
  }
  rmf_absolute_paths = [rmf_absolute_paths, rmf_info.arg_string];


  % load ARF
  %%%%%%%%%%

  % 1. ARF file
  if(arfFile!=NULL) { arf = load_arf(arfFile); }

  % use absolute paths
  if(arf!=NULL)
  {
    variable arf_info = get_arf_info(arf);
    if(arf_info.file[0] != '/')
    { arf_info.file = getcwd() + arf_info.file;
      set_arf_info(arf, arf_info);
    }
  }

%
%  % 2. XMM/RGS areascal
%  if(fits_key_exists(phaFile, "TELESCOP") and fits_key_exists(phaFile, "INSTRUME"))
%  { if(    substr(fits_read_key(phaFile, "TELESCOP"), 1, 3)=="XMM"
%       and substr(fits_read_key(phaFile, "INSTRUME"), 1, 3)=="RGS")
%    {
%      variable  arfGrid = get_rmf_arf_grid(rmf);
%      variable dataGrid = get_rmf_data_grid(rmf);
%      arfData = struct_combine(arfGrid, "value", "err");
%      variable fracexpo = rebinDensity(arfGrid.bin_lo, arfGrid.bin_hi,
%                                       dataGrid.bin_lo, dataGrid.bin_hi, fits_read_table(phaFile).areascal);
%      arfData.value = fracexpo;
%      arfData.err = 0*fracexpo;
%
%     if(arf==NULL)
%      { arf = define_arf(arfData); }
%      else
%      { arfData.value *= get_arf(arf).value;
%        put_arf(arf, arfData);
%      }
%
%      variable arf_info = get_arf_info(arf);
%      arf_info.fracexpo *= fracexpo;
%      if(arf_info.file==NULL) { arf_info.file = ""; }
%      set_arf_info(arf, arf_info);
%    }
%  }

  % 3. RSP-factors
  if(rmf!=NULL)
  {
    (arf, rmf) = factorized_arf_rmf(arf, rmf);
  }

  % assign response
  %%%%%%%%%%%%%%%%%
  if(rmf!=NULL) { assign_rmf(rmf, spec); }

  if(arf!=NULL)
  { assign_arf(arf, spec);
    if(get_arf_exposure(arf)<0) { set_arf_exposure(arf, get_data_exposure(spec)); }
  }

  % useful commands
  ignore_bad(spec);
  flux_corr(spec);
%  if(fits_key_exists(phaFile, "RFLORDER"))
%  { info = get_data_info(spec);
%    info.order = fits_read_key(phaFile, "RFLORDER");
%    set_data_info(spec, info);
%  }

  return spec;
}

%
% Code from src/data/spectrum.loaddata/loadHETG.sl
%

define loadHETGspec()
{
%!%+
%\function{loadHETGspec}
%\synopsis{Load locally extracted Chandra HETG spectra and responses}
%\usage{Struct_Type ids = loadHETGspec();}
%\qualifiers{
%\qualifier{dir [="."]}{     : path to the extracted spectra}
%\qualifier{heg [=1]}{       : load HEG spectra (=0: don't load)}
%\qualifier{meg [=1]}{       : load MEG spectra (=0: don't load)}
%\qualifier{order [=[-1,1]]}{: diffraction orders to load. Sign matters.}
%}
%\description
%   Load Chandra HETG spectra extracted with the (new) Remeis
%   extraction scripts and return a structure with the dataset indices.
%   
%   The new (summer 2014) Remeis Chandra gratings extraction scripts
%   do not longer specify observation / extraction specific information 
%   in the file name, but only the arm and diffraction order of the
%   spectrum, e.g., heg_m1.pha or meg_p1.pha. The old extraction
%   scripts used to generate a loaddata0.sl file for convenience. With
%   the more generic file names, this function replaces the load
%   script. To ensure backwards compatibility with dataset-index
%   sensitive par files, the function loads the arms and orders in the
%   exact same order as the old load scripts used to: 
%      heg_m1 heg_p1 meg_m1 meg_p1 heg_m2 heg_p2 meg_m2 meg_p2 ...
%   
%   The returned structure is then in the form (in case of the defaults):
%      struct{
%        dir = ".",
%        orders = [-1,1],
%        arms = ["heg","meg"],
%        heg_m1 = 1,
%        heg_p1 = 2,
%        meg_m1 = 3,
%        ...
%      }
%   The function returns -1 if both arms are ignored, i.e., no spectra
%   chosen.
%      
%   Negative and positive diffraction orders can be chosen independently
%   of each other. However, all selected orders are applied to all
%   selected grating arms: You can load HEG without MEG and vice versa. 
%   But if you load them both in a single call of loadHETGspec, the
%   same orders will be loaded for both of them. 
%      
%   If you would like to load configurations like [heg_m1, meg_p1],
%   where different orders are loaded for each grating arms, you have
%   to call the function multiple times or load the spectra by hand. 
%   
%\seealso{loadHETGlcs, loadHETGlc_sum}
%!%-
   variable dir = qualifier("dir",".")+"/";
   variable heg = qualifier("heg",1);
   variable meg = qualifier("meg",1);
   variable order = int(qualifier("order",[-1,1]));

   variable arm = String_Type[0];
   if(heg) arm = [arm,"heg"];
   if(meg) arm = [arm,"meg"];
   if(length(arm)==0) return -1;
   
   variable aos = abs(order[unique(abs(order))]);
   aos = aos[array_sort(aos)];
   
   variable ids = struct{dir = dir, orders=order, arms = arm};
   variable o,a,v, w,str, did;
   foreach o (aos)
	 {
		foreach a (arm)
		  {
			 foreach v([0,1])
			   {
				  w = where(order == o * [-1,1][v]);
				  if(length(w)>0)	
					{				  
					   str = sprintf("%s_%s%d",a,["m","p"][v],o);
					   ids = struct_combine(ids,str);	
					   did = load_data(dir+str+".pha");
					   set_struct_field(ids,str,did);
					}
				  
			   }
		  }
		
	 }
   return ids;
}

define loadHETGlcs()
{
%!%+
%\function{loadHETGlcs}
%\synopsis{Load Chandra HETG lightcurves}
%\usage{lcs = loadHETGlcs();}
%\qualifiers{
%\qualifier{dir [="."]}{     : path to the extracted spectra}
%\qualifier{heg [=1]}{       : load HEG spectra (=0: don't load)}
%\qualifier{meg [=1]}{       : load MEG spectra (=0: don't load)}
%\qualifier{order [=[-1,1]]}{: diffraction orders to load. Sign matters.}
%\qualifier{energyband [=[500,10000]]}{: extracted energy bands. See
%                                 description below for details.}
%\qualifier{unit [="eV"]}{   : Unit for the energyband: eV, keV, A}
%}
%\description   
%   Load Chandra HETG lightcurves extracted with the (new) Remeis
%   extraction scripts and return a structure with the results.
%   
%   The new (summer 2014) Remeis Chandra gratings extraction scripts
%   do not longer specify observation / extraction specific
%   information in the file name, but only the arm, diffraction order,
%   and energyband of the lightcurve, e.g., heg_m1_500-1000.lc or
%   meg_p1_3500-10000.lc. 
%   
%   loadHETGlcs reads the corresponding lightcurve for each specified arm,
%   diffraction order, and enegeryband and stores it as a field in a
%   structure. 
%   The returned structure is then in the form (in case of the defaults):
%      struct{
%        dir = ".",
%        orders = [-1,1],
%        arms = ["heg","meg"],
%        energyband = [500, 10000],
%        unit = "eV",
%        heg_m1_500_10000 = Struct_Type, % content of the fits file
%        heg_p1_500_10000 = Struct_Type,
%        meg_m1_500_10000 = Struct_Type,
%        ...
%      }
%   The function returns -1 if both arms are ignored, i.e., no
%   lightcurves chosen.
%   
%   The energyband can be specified in multiple ways:
%     * a list or array of limits: [0.5,1.5,3,10] keV. In this case, the
%       energybands for the lightcurves are assumed to be 0.5-1.5,
%       1.5-3, and 3-10 keV, i.e., there are no gaps.
%     * a list or an array of energy pairs: {[0.5,1.5],[3,10]} or
%       [{0.5,1.5},{3,10}] keV. Then the bands are taken as they are,
%       i.e., it is possible to have gaps between energy bands. 
%  
%\seealso{loadHETGlc_sum, loadHETGspec}
%!%-
   variable dir = qualifier("dir",".")+"/";
   variable heg = qualifier("heg",1);
   variable meg = qualifier("meg",1);
   variable order = qualifier("order",[-1,1]);
   variable enb = qualifier("energyband",[500,10000]);
   variable unit = qualifier("unit","eV");
   variable en,i;
   if(typeof(enb[0]) == Double_Type or typeof(enb[0]) == Integer_Type)
	 {
		en = Array_Type[length(enb)-1];
		_for i (1,length(enb)-1,1)
		  {
			 en[i-1] = enb[[i-1,i]]; 
		  }
	 }else{
		en = enb;
	 }
   _for i (0,length(en),1)
	 {
		if(unit == "keV") en[i] *=1e3;
		if(unit == "A") en[i] = 12398.42/en[i];
	 }
   
   variable arm = String_Type[0];
   if(heg) arm = [arm,"heg"];
   if(meg) arm = [arm,"meg"];
   if(length(arm)==0) return -1;
   
   variable aos = abs(order[unique(abs(order))]);
   aos = aos[array_sort(aos)];
   
   variable lcs = struct{dir = dir, orders=order, arms = arm,
	  energyband = enb, unit=unit};
   variable o,a,v,e, w,str, did;
   foreach o (aos)
	 {
		foreach a (arm)
		  {
			 foreach v([0,1])
			   {
				  w = where(order == o * [-1,1][v]);
				  if(length(w)>0)	
					{
					   _for e (0,length(en)-1,1)
						 {
							str = sprintf("%s_%s%d_%04d-%04d",
										  a,["m","p"][v],o,
										  int(en[e][0]),int(en[e][1]));
							did = fits_read_table(dir+str+".lc");
							str = sprintf("%s_%s%d_%04d_%04d",
										  a,["m","p"][v],o,
										  int(en[e][0]),int(en[e][1]));
							lcs = struct_combine(lcs,str);	
							set_struct_field(lcs,str,did);
						 }
					}
			   }
		  }
	 }
   return lcs;
}

define loadHETGlc_sum()
{
%!%+
%\function{loadHETGlc_sum}
%\synopsis{ADD Chandra HETG lightcurves}
%\usage{Struct_Type lcsum = loadHETGlcs( [lcs] );}
%\qualifiers{
%\qualifier{dir [="."]}{     : path to the extracted spectra}
%\qualifier{heg [=1]}{       : load HEG spectra (=0: don't load)}
%\qualifier{meg [=1]}{       : load MEG spectra (=0: don't load)}
%\qualifier{order [=[-1,1]]}{: diffraction orders to load. Sign matters.}
%\qualifier{energyband [=[500,10000]]}{: extracted energy bands. See
%                                 description below for details.}
%\qualifier{unit [="eV"]}{   : Unit for the energyband: eV, keV, A}
%}
%\description   
%   Sum multiple Chandra HETG lightcurves extracted with the (new) Remeis
%   extraction scripts and return a structure with the result.
%   
%   The new (summer 2014) Remeis Chandra gratings extraction scripts
%   do not longer specify observation / extraction specific
%   information in the file name, but only the arm, diffraction order,
%   and energyband of the lightcurve, e.g., heg_m1_500-1000.lc or
%   meg_p1_3500-10000.lc, and only extract lightcurves on a per
%   arm/order basis. 
%   
%   Often, we are more interested in the combined lightcurve of
%   multiple arms / orders. 
%   loadHETGlc_sum can be used in 2 ways:
%     - call the function with no arguments and use the qualifiers to
%       specify which lightcurves you are interested in. The
%       qualifiers are passed to loadHETGlcs to load the lightcurves
%       first.
%     - call the function with 1 argument: the lightcurves produced by
%       a previous call of loadHETGlcs. In this case, all lightcurves
%       contained in the structure will be summed, and the qualifiers
%       will have no effect [in the future, they could be used to
%       choose a subset of the given lightcurves, but this is not
%       implemented yet].
%   
%   The returned structure has the same fields as simply reading a
%   single lightcurve file would produce. If reading of the
%   lightcurves fails, the return value is -1. 
%   
%   The function assumes that the lightcurves have not been tempered
%   with, i.e., that they are correctly (Chandra) formatted
%   lightcurves and have the same length, time grid, etc. No checks
%   are performed. Fields like time and exposure are taken from the
%   first lightcurve in the structure, since they are expected to be
%   the same for various extractions of the same ObsID. 
%   
%   The energyband can be specified in multiple ways:
%     * a list or array of limits: [0.5,1.5,3,10] keV. In this case, the
%       energybands for the lightcurves are assumed to be 0.5-1.5,
%       1.5-3, and 3-10 keV, i.e., there are no gaps.
%     * a list or an array of energy pairs: {[0.5,1.5],[3,10]} or
%       [{0.5,1.5},{3,10}] keV. Then the bands are taken as they are,
%       i.e., it is possible to have gaps between energy bands. 
%  
%\seealso{loadHETGlcs, loadHETGspec}
%!%-   
   variable qq, lcs;
   if(_NARGS==1) { lcs = (); }
   else
	 {
		qq = __qualifiers();
		lcs = loadHETGlcs(;;qq);
	 }
   
   if( not( typeof(lcs)==Struct_Type)) return -1;
   
   variable fields = get_struct_field_names(lcs);
   variable fld;
   foreach fld (["dir","orders","arms","energyband","unit"])
	 {fields[where(fields==fld)] = "";}
   fields = fields[where(fields!="")];

%   variable lcfields = get_struct_field_names(get_struct_field(lcs,fields[0]));
%   variable ssum = @Struct_Type(lcfields);
   variable ssum = get_struct_field(lcs,fields[0]);
   ssum.stat_err = ssum.stat_err^2;
   ssum.count_rate_err = ssum.count_rate_err^2;
   variable dum,i;
   _for i (1,length(fields)-1,1)
	 {
		dum = get_struct_field(lcs,fields[i]);
		ssum.counts += dum.counts;
		ssum.stat_err += dum.stat_err^2;
		ssum.count_rate += dum.count_rate;
		ssum.count_rate_err += dum.count_rate_err^2;
	 }
   ssum.stat_err = sqrt(ssum.stat_err);
   ssum.count_rate_err = sqrt(ssum.count_rate_err);
   return ssum;
}

%
% Code from src/data/spectrum.loaddata/load_1storderHETGS_datasets.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define load_1storderHETGS_datasets()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{load_1storderHETGS_datasets}
%\synopsis{loads MEG+-1 & HEG+-1 Chandra HETGS spectra}
%\usage{Integer_Type ids[] = load_1storderHETGS_datasets(String_Type specpath, RMFpath);}
%!%-
{
  variable specpath, RMFpath;
  switch(_NARGS)
  { case 2: (specpath, RMFpath) = (); }
  { help(_function_name());  return;}
  return load_HETGS_datasets(specpath, RMFpath, 1);
}

%
% Code from src/data/spectrum.loaddata/load_HETGS_datasets.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define load_HETGS_datasets()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{load_HETGS_datasets}
%\synopsis{loads Chandra HETGS spectra from a type II pha file}
%\usage{Integer_Type ids[] = load_HETGS_datasets(String_Type specpath, RMFpath[, Integer_Type ms[]]);}
%\description
%    \code{ids} is an array containing negative/positive MEG/HEG spectra
%!%-
{
  variable specpath, RMFpath, ms = [1,2,3];
  switch(_NARGS)
  { case 2: (specpath, RMFpath) = (); }
  { case 3: (specpath, RMFpath, ms) = (); }
  { help(_function_name()); return;}
  specpath += "/";
  RMFpath += "/";

  % 1st index: [0, 1, 2, 3] = grating and +/- direction, see variable grat
  variable grat = ["MEG_-", "MEG_", "HEG_-", "HEG_"];
  % 2nd index: [1,2,3] = order
  variable row = Integer_Type[4,4];  % in pha2-file
  row[[0:3],1] = [9, 10, 3, 4];  % 1st order
  row[[0:3],2] = [8, 11, 2, 5];  % 2nd-order
  row[[0:3],3] = [7, 12, 1, 6];  % 3rd-order

  variable m, s, specs = Integer_Type[0];
  foreach m ([ms])
    _for s (0, 3, 1)
      specs = [specs, loadDataset(specpath + "pha2.fits",
                                  RMFpath + grat[s] + string(m) + ".rmf",
                                  specpath + grat[s] + string(m) + "_garf.fits",
                                  specpath+"bkg2.fits",
                                  row[s, m]
                                 )
               ];
  return specs;
}

%
% Code from src/data/load_data_combined.sl
%
%%%%%%%%%%%%%%%%%%%%%
define load_data_combined()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{load_data_combined}
%\synopsis{reads and combines the given rows from a Fits Type II pha file}
%\usage{Integer_Type load_data_combined(String_Type, pha_filename, Integer_Type[] rows);}
%\description
%    A combination of load_data and combine_datasets, with
%    the exception that the rows appear as a single dataset.
%
%    All qualifiers are passed to load_data.
%
%\seealso{load_data, combine_datasets}
%!%-
{
  if (_NARGS != 2) { help(_function_name); return; }
  variable phafile, rows;
  (phafile, rows) = ();

  % first, load the data the classical way
  variable did = load_data(phafile, rows;; __qualifiers);
  % get the combined exposure time
  variable exposure = sum(array_map(Double_Type, &get_data_exposure, did)); 
  % combine them and retrieve the combined data structure
  variable gid = combine_datasets(did);
  variable data = _A(get_combined(gid, &get_data_counts));
  uncombine_datasets(gid);
  delete_data(did);
  % define the data
  variable nid = define_counts(data);
  set_data_exposure(nid, exposure);
  variable backfile = fits_read_key(phafile, "BACKFILE");
  if (backfile != "none") {
    if (path_dirname(backfile) == ".") backfile = path_dirname(phafile) + "/" + backfile;
    ()=define_back(nid, backfile);
  }
  variable ancrfile = fits_read_key(phafile, "ANCRFILE");
  if (ancrfile != "none") {
    if (path_dirname(ancrfile) == ".") ancrfile = path_dirname(phafile) + "/" + ancrfile;
    assign_arf(load_arf(ancrfile), nid);
  }
  variable respfile = fits_read_key(phafile, "RESPFILE");
  if (respfile != "none") {
    if (path_dirname(respfile) == ".") respfile = path_dirname(phafile) + "/" + respfile;
    assign_rmf(load_rmf(respfile), nid);
  }
  return nid;
}

%
% Code from src/data/spectrum.loaddata/load_data_integral.sl
%
define load_data_integral()
%!%+
%\function{load_data_integral}
%\synopsis{reads an INTEGRAL spectrum from a PHA file and performs the necessary tweaks}
%\usage{Integer_Type id = load_data_integral(String_Type pha_filename);}
%\description
%    ISIS's load_data often does not work with INTEGRAL spectra for the following reasons:\n
%    - The extension of the PHA file is not called 'SPECTRUM'.\n
%    - The extensions of the RMF file are not called 'MATRIX' and 'EBOUNDS'.\n
%    - The extension of the ARF file is not called 'SPECRESP'.\n
%    - The columns ENERG_LO and ENERG_HI of the ARF file contain only zeros.\n
%    The function load_data_integral tries to make the appropriate changes
%    (modify the extension names, write a new ARF file) and finally uses load_data.
%\qualifiers{
%\qualifier{verbose}{[=1] show changes}
%}
%\seealso{load_data}
%!%-
{
  variable phafile;
  switch(_NARGS)
  { case 1: phafile = (); }
  { help(_function_name()); }

  variable verbose = qualifier("verbose", 1);

  % correct name of the PHA-file's SPECTRUM extension
  variable EXTNAME, newEXTNAME, F = fits_open_file(phafile, "w");
  ()=_fits_movabs_hdu(F, 1);
  do
  { EXTNAME = fits_read_key(F, "EXTNAME");
    if(any(["ISGR-PHA1-SPE", "JMX1-PHA1-SPE", "JMX2-PHA1-SPE", "SPI.-PHA1-SPE"] == EXTNAME))
    { newEXTNAME = "SPECTRUM";
      ()=_fits_update_key(F, "EXTNAME", newEXTNAME, "name of the binary extension [updated]");
      ()=_fits_write_history(F, "EXTNAME "+EXTNAME+" was updated by ISIS:"+_function_name());
      if(verbose)
	vmessage("%s: changing extension name of %s from '%s' to '%s'", _function_name(), phafile, EXTNAME, newEXTNAME);
    }
  } while(_fits_movrel_hdu(F, 1)==0);
  fits_close_file(F);

  %%vg: if data loaded not from directory with data, but by giving an
  %%(absolute) path to the data, take care of the path to be used for
  %%RESPONDE and ARF
  %%BUT: don't have to do it, if absolute pathes are already given
  variable path;
  variable pathparts = strchop(phafile,'/',0);
  if (length(pathparts) == 1) {
    path = "";
  }
  else {    
    path = strjoin(pathparts[[0:length(pathparts)-2]],"/")+"/";
  }

  variable ok;
  variable RESPFILE = fits_read_key(phafile+"[SPECTRUM]", "RESPFILE");
  ifnot (strncmp(RESPFILE,"/",1)==0) {RESPFILE = path+RESPFILE;};
  if(RESPFILE!=NULL && RESPFILE!="NONE")
  { % correct name of the RMF-file's MATRIX/EBOUNDS extensions
    ok = 1;
    try { F = fits_open_file(RESPFILE, "w"); }
    catch FitsError: { ok = 0; vmessage("warning (%s): FITS error while opening %s in 'w' mode (Do you have write permission?)", _function_name(), RESPFILE); }
    if(ok)
    {
      ()=_fits_movabs_hdu(F, 1);
      do
      { EXTNAME = fits_read_key(F, "EXTNAME");
        if(any(["ISGR-RMF.-RSP",
		"JMX1-RMF.-RSP", "JMX2-RMF.-RSP",
		"SPI.-RMF.-RSP",
		"COMP-SRMF-RSP"] == EXTNAME))
        { newEXTNAME = "MATRIX";  % "SPECRESP MATRIX" ?
  	  ()=_fits_update_key(F, "EXTNAME", newEXTNAME, "name of the binary extension [updated]");
          ()=_fits_write_history(F, "EXTNAME "+EXTNAME+" was updated by ISIS:"+_function_name());
          if(verbose)
  	    vmessage("%s: changing extension name of %s from '%s' to '%s'", _function_name(), RESPFILE, EXTNAME, newEXTNAME);
        }
        if(any(["ISGR-EBDS-MOD",
		"JMX1-FBDS-MOD", "JMX2-FBDS-MOD",
		"SPI.-EBDS-SET",
		"COMP-SEBD-MOD"] == EXTNAME))
        { newEXTNAME = "EBOUNDS";
  	  ()=_fits_update_key(F, "EXTNAME", newEXTNAME, "name of the binary extension [updated]");
          ()=_fits_write_history(F, "EXTNAME "+EXTNAME+" was updated by ISIS:"+_function_name());
          if(verbose)
	    vmessage("%s: changing extension name of %s from '%s' to '%s'", _function_name(), RESPFILE, EXTNAME, newEXTNAME);
        }
      } while(_fits_movrel_hdu(F, 1)==0);
      fits_close_file(F);
    }
  }

  variable ANCRFILE = path+fits_read_key(phafile, "ANCRFILE");
  ifnot (strncmp(ANCRFILE,"/",1)==0) {ANCRFILE = path+ANCRFILE;};
  if(ANCRFILE!=NULL && ANCRFILE!="NONE")
  { % correct name of the ARF-file's EBOUNDS extensions
    ok = 1;
    try { F = fits_open_file(ANCRFILE, "w"); }
    catch FitsError: { ok = 0; vmessage("warning (%s): FITS error while opening %s in 'w' mode (Do you have write permission?)", _function_name(), ANCRFILE); }
    if(ok)
    {
      ()=_fits_movabs_hdu(F, 1);
      do
      { EXTNAME = fits_read_key(F, "EXTNAME");
        if(any(["ISGR-ARF.-RSP",
		"JMX1-AXIS-ARF", "JMX2-AXIS-ARF",
		"COMP-SARF-RSP"] == EXTNAME))
        { newEXTNAME = "SPECRESP";
  	  ()=_fits_update_key(F, "EXTNAME", newEXTNAME, "name of the binary extension [updated]");
          ()=_fits_write_history(F, "EXTNAME "+EXTNAME+" was updated by ISIS:"+_function_name());
          if(verbose)
  	    vmessage("%s: changing extension name of %s from '%s' to '%s'", _function_name(), ANCRFILE, EXTNAME, newEXTNAME);
        }
      } while(_fits_movrel_hdu(F, 1)==0);
      fits_close_file(F);

      variable arf = fits_read_table(ANCRFILE);
      if(max(arf.energ_lo)==0)
      { if(verbose)
          vmessage("%s: changing %s due to zero-columns ENERG_LO and ENERG_HI", _function_name(), ANCRFILE);
        (arf.energ_lo, arf.energ_hi) = fits_read_col(RESPFILE+"[MATRIX]", "energ_lo", "energ_hi");
        variable oldANCRFILE = ANCRFILE+".old";
        ()=rename(ANCRFILE, oldANCRFILE);
        F = fits_open_file(ANCRFILE, "c");
        variable Fold = fits_open_file(oldANCRFILE, "r");
        ()=_fits_movabs_hdu(Fold, 1);
        ()=_fits_copy_header(Fold, F);
        ()=_fits_movabs_hdu(Fold, 2);
        variable records = fits_read_records(Fold);
        fits_write_binary_table(F, newEXTNAME, arf);
        ()=_fits_movabs_hdu(F, 2);
        fits_write_records(F, records);
        fits_close_file(F);
       fits_close_file(Fold);
      }
    }
  }

  variable OGIPcomp = Rmf_OGIP_Compliance;
  Rmf_OGIP_Compliance = 0;
  variable id = load_data(phafile);
  Rmf_OGIP_Compliance = OGIPcomp;

  set_rebin_error_hook(id, "rebin_error_hook");

  return id;
}

%
% Code from src/data/load_fermi.sl
%
% -*- mode: slang; mode: fold -*-

private define fermi_new_centered_energy_grid(lo,hi,cener) %{{{
{
   variable i,n = length(lo);
   variable nlo = Double_Type[n];
   variable nhi = @nlo;
   
   nlo[0] = lo[0];
   nhi[0] = 2.0*cener[0]-nlo[0];
   
   _for i(1,n-1,1)
   {
      nlo[i] = nhi[i-1];
      nhi[i] = 2.0*cener[i]-nlo[i];      
   }      
   return nlo,nhi;
}
%}}}
private define log_interpol(new_x,x,y) %{{{
{
   return exp(interpol(log(new_x),log(x),log(y));;__qualifiers);
}
%}}}
private define integrate_diff_flux(dat,mo) %{{{
% ****** Differtial -> Integrated Flux ****** %
{
   variable n = length(dat.bin_lo);
   variable corr_fac = Double_Type[n];
   variable i,new_x,new_y,f_e_center,integ_flux;
   _for i(0,n-1,1)
   {
      new_x = exp([log(dat.bin_lo[i]):log(dat.bin_hi[i]):#10]);
      new_y = log_interpol(new_x,mo.ener,mo.flux);
      f_e_center = log_interpol(0.5*(dat.bin_lo[i]+dat.bin_hi[i]),mo.ener,mo.flux);   

      integ_flux = integrate_trapez(new_x,new_y);

      corr_fac[i] = integ_flux / (f_e_center * (dat.bin_hi[i]-dat.bin_lo[i]));      
   }

   variable integ_bin = (dat.bin_hi-dat.bin_lo)*corr_fac;
   
   variable str = struct{bin_lo=dat.bin_lo,bin_hi=dat.bin_hi,value=dat.flux*integ_bin,err=dat.err*integ_bin};
   
   return str, corr_fac;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%
define load_fermi()%{{{
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{load_fermi}
%\synopsis{loads fermi data produced by the "Fermi-SED Script"@Remeis}
%\usage{Integer_Type data_id = load_fermi(String_Type filename);}
%\qualifiers{
%\qualifier{rmf_factor}{[=10]: define a fine RMF to properly evaluate the model}
%}
%\description
%  Loads fermi data produced by the "Fermi-SED Script"@Remeis. Note
%  that special care is taken here, as the energy bins are relatively
%  large. Therefore the function set_bin_corr_factor is used to
%  calculate a correction factor. This factor is used automatically
%  for loading the data. In order to plot the data properly,
%  plot_data/unfold has to be used.
%  WARNING: Use clear_all to delete the correction factors (as well 
%  as the data). This is necessary as otherwise the correction factors 
%  are applied to a newly loaded dataset with the same ID.
%
%  This function is not working if you have computed a Fermi spectrum
%  with Fermipy!
%
%\seealso{load_fermi_spectrum,load_fermi_catalog}
%!%-
{
   variable in_fil;
   switch(_NARGS)
   { case 1: in_fil = ();}
   { help(_function_name()); return; }
   
   variable raw_dat = fits_read_table(in_fil);
   % make sure that no "strange" bins enter the analysis (SED analysis; Nela)
   struct_filter(raw_dat, where(raw_dat.center_energy >0.0)); 
   
   if ( raw_dat.center_energy == NULL )
   {verror ( "\n  ***** Error: Unable to open file %s.\n", in_fil );return NULL; }
   
   variable blo,bhi;
   (blo,bhi) =  fits_read_col(in_fil+"[2]","Bin Low Edge", "Bin High Edge");

   variable model = fits_read_table(in_fil+"[3]");

   % convert the bins such that for the new grid: 
   (blo,bhi) = fermi_new_centered_energy_grid(blo,bhi,raw_dat.center_energy);

   % make it keV here!!
   variable GeV2keV = 1e6;
   
   variable str_dat = struct{bin_lo=blo*GeV2keV,bin_hi=bhi*GeV2keV,flux=@raw_dat.dn_de/GeV2keV, err=@raw_dat.dn_de_error/GeV2keV};
   %struct_filter(str_dat, where(str_dat.err >0.0));   
   variable str_mo = struct{ener=model.center_energy*GeV2keV, flux=model.model/GeV2keV};
   
   Minimum_Stat_Err=1.e-20;   

   % properly Bin-Integrate the values
   variable dat, corr_factor;
   (dat,corr_factor) = integrate_diff_flux(str_dat,str_mo);

   % load data and set the correction factor
   variable fermi_id = define_counts(_A(dat));
   set_bin_corr_factor(fermi_id,corr_factor);
      
   set_data_exposure(fermi_id,1.);
   ignore_list(fermi_id, where ( dat.value == 0.));

   assign_rmf (make_fine_rmf(fermi_id,qualifier("rmf_factor",10)),fermi_id);

   return fermi_id;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define load_fermi_spectrum()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{load_fermi_spectrum}
%\synopsis{loads fermi data produced by any fermipy script}
%\usage{Integer_Type data_id = load_fermi_spectrum(String_Type filename);}
%\qualifiers{
%\qualifier{rmf_factor}{[=10]: define a fine RMF to properly evaluate the model}
%\qualifier{ts_min}{[=9]: set TS threshold for spectral bins to be loaded in as data points.}
%}
%\description
%  Loads fermi data produced by any Fermi spectrum produced with fermipy. Note
%  that special care is taken here, as the energy bins are relatively
%  large. Therefore the function set_bin_corr_factor is used to
%  calculate a correction factor. This factor is used automatically
%  for loading the data. In order to plot the data properly,
%  plot_data/unfold has to be used.
%
%  A note regarding upper limits: you can set yourself at which significance
%  you consider a spectral to be valid and not use the 2 sigma upper limit (UL).
%  The default value is TS = 9, which is ~3 sigma. The SED plot produced by 
%  the Fermi-LAT analysis script has a lower threshold, which is why you might
%  get less data points in your spectrum, when you load it in isis with the
%  default value of 'ts_min'. 
%  Spectral bins with a lower TS value will be treated as upper limits and 
%  not loaded in isis. For plotting purposes, the upper limits need to be 
%  added from the [*]_sed.fits file (column e2dnde_ul) and multiplied by a factor
%  of 1.60218e-6 to convert from MeV to erg.
%
%  This function supersedes the previous function 'load_fermi', which can only
%  be used to load the Fermi Spectra produced by the original "Fermi-SED"
%  script.
%
%  WARNING: Use clear_all to delete the correction factors (as well 
%  as the data). This is necessary as otherwise the correction factors 
%  are applied to a newly loaded dataset with the same ID.
%
%
%
%\seealso{load_fermi}
%!%-
{
        variable in_fil;
        switch(_NARGS)
        { case 1: in_fil = ();}
        { help(_function_name()); return; }

        variable raw_dat = fits_read_table(in_fil); % Energy of input bins are given in MeV!
        % Filter out 'strange' bins and upper limits
        variable tsfilter = qualifier("ts_min",9);
        struct_filter(raw_dat, where(raw_dat.e_ref>0.0));
        struct_filter(raw_dat, where(raw_dat.ts>=tsfilter));

        if (raw_dat.e_ref == NULL){
                vmessage("* Error: Unable to open file, no spectral bins detected in %s \n", in_fil);
        }

        % Load Model flux
        variable model = fits_read_table(in_fil+"[2]");

        % convert bins to fit to new grid
        variable blo, bhi;
        (blo, bhi) = fermi_new_centered_energy_grid(raw_dat.e_min, raw_dat.e_max, raw_dat.e_ref);

        % Convert from MeV to keV
        variable MeV2keV = 1e3;

        variable str_dat = struct{bin_lo=blo*MeV2keV, bin_hi=bhi*MeV2keV, flux=@raw_dat.dnde/MeV2keV, err=@raw_dat.dnde_err/MeV2keV};
        variable str_mo = struct{ener=model.energy*MeV2keV, flux=model.dnde/MeV2keV};

        Minimum_Stat_Err=1.e-20;

        % properly bin-integrate the values
        variable dat, corr_factor;
        (dat, corr_factor) = integrate_diff_flux(str_dat, str_mo);

        % Load data and set the correction factor
        variable fermi_id = define_counts(_A(dat));
        set_bin_corr_factor(fermi_id, corr_factor);

        set_data_exposure(fermi_id, 1.);
        ignore_list(fermi_id, where(dat.value==0.));
        assign_rmf(make_fine_rmf(fermi_id, qualifier("rmf_factor",10)), fermi_id);

        return fermi_id;
}



private define get_corr_factor(dat,gam) %{{{
{
   variable ec = 0.5*(dat.bin_lo+dat.bin_hi);   
   return
     ( dat.bin_hi^(1.-gam) - dat.bin_lo^(1.-gam))
     / ( (ec^(-gam)) * (dat.bin_hi - dat.bin_lo) * (1 - gam));   
}

%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%
define load_fermi_catalog() %{{{
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{load_fermi_catalog}
%\synopsis{loads Fermi data from a catalog for a given source}
%\usage{Integer_Type data_id = load_fermi_catalog(String_Type sourcename);
%}
%\qualifiers{
%\qualifier{list}{return list of available sources}
%\qualifier{ul}{Load includes upper limits}
%\qualifier{catalog}{specify the catalog (default: 2FGL_point_source@Remeis) }
%\qualifier{fgl3}{Use this when reading in the 3FGL catalog, either with 
%                   the catalog qualifier or without }
%}
%\description
%  WARNING: Use clear_all to delete the correction factors (as well 
%  as the data). This is necessary as otherwise the correction factors 
%  are applied to a newly loaded dataset with the same ID.
%
%\seealso{load_fermi}
%!%-
{

   variable src=NULL;
   switch(_NARGS)
   { case 0: ifnot(qualifier_exists("list")) { help(_function_name()); return; } }
   { case 1: src = (); }
   { help(_function_name()); return; }
   
   variable path_to_data = NULL;

   if (qualifier_exists("catalog"))
   {
      path_to_data = qualifier("catalog");
      if(stat_file(path_to_data)==NULL)
      {
	 vmessage("error (%s): Fermi catalog not at %s", _function_name(),path_to_data);
	 return -1;
      }
   }
   else
   {      

       if (qualifier_exists("fgl3"))
       {
     try {
	   foreach path_to_data ([local_paths.fermi_catalog_3fgl])
      if(stat_file(path_to_data)!=NULL)  break;   
     } catch NotImplementedError: { vmessage("error (%s): local_paths.fermi_catalog_3fgl not defined",_function_name()); return NULL; };
	   if(stat_file(path_to_data)==NULL)
	   { 
	       vmessage("error (%s): Fermi catalog not found on the system", _function_name());
	       return -1;
	   }
       }
       else
       {
	   try {
	   foreach path_to_data ([local_paths.fermi_catalog])
      if(stat_file(path_to_data)!=NULL)  break;   
     } catch NotImplementedError: { vmessage("error (%s): local_paths.fermi_catalog not defined",_function_name()); return NULL; };
	   if(stat_file(path_to_data)==NULL)
	   { 
	       vmessage("error (%s): Fermi catalog not found on the system", _function_name());
	       return -1;
	   }
       }
   }


   
   variable r = fits_read_table (path_to_data);   
   
   if (qualifier_exists("list"))
   return r.source_name;
   


   Minimum_Stat_Err=1e-20;
   struct_filter(r, where(r.source_name==src);dim=0);
  


   ifnot (length(r)>0)
   { 
      vmessage("error (%s): Source %s not found", _function_name(),src);
      return -1;
   }
   else if (length(r) > 1)
   {
      % although name should be unique, make sure nothing strange happens
      vmessage("error (%s): Source %s not unique!", _function_name(),src);
      return -1;
   }

   variable dat,i;
   if (qualifier_exists("fgl3"))
   {
       variable fluxes_unc = [
       r.unc_flux100_300[0,0],
       r.unc_flux100_300[0,1],
       r.unc_flux300_1000[0,0],
       r.unc_flux300_1000[0,1],
       r.unc_flux1000_3000[0,0],
       r.unc_flux1000_3000[0,1],
       r.unc_flux3000_10000[0,0],
       r.unc_flux3000_10000[0,1],
       r.unc_flux10000_100000[0,0],
       r.unc_flux10000_100000[0,1],
       ];
       variable temperr = Double_Type[0];
       _for i(0,4,1)
       {


	   if (isnan(fluxes_unc[2*i]) == 1 )
	   {
	       temperr = [temperr,0.0];
	   }
	   else
	   {
	       temperr = [temperr,max([abs(fluxes_unc[2*i]),fluxes_unc[2*i+1]])];
	   }
	   
       }
       dat = struct
       {
	   bin_lo = [100,300,1000,3000,10000]         * 1e3, % convert from MeV to keV
	   bin_hi = [    300,1000,3000,10000, 100000] * 1e3,
	   value  = [
	   r.flux100_300[0],
	   r.flux300_1000[0],
	   r.flux1000_3000[0],
	   r.flux3000_10000[0],
	   r.flux10000_100000[0],
	   ],
	   err = temperr,
       };
       }
       else
       {
	   dat = struct
	   {
	       bin_lo = [30,100,300,1000,3000,10000]         * 1e3, % convert from MeV to keV
	       bin_hi = [   100,300,1000,3000,10000, 100000] * 1e3,
	       value  = [r.flux30_100[0],
	       r.flux100_300[0],
	       r.flux300_1000[0],
	       r.flux1000_3000[0],
	       r.flux3000_10000[0],
	       r.flux10000_100000[0]
	       ],
	       err    = [r.unc_flux30_100[0],
	       r.unc_flux100_300[0],
	       r.unc_flux300_1000[0],
	       r.unc_flux1000_3000[0],
	       r.unc_flux3000_10000[0],
	       r.unc_flux10000_100000[0]
	       ]
	   };
	   }       
	   
	   if (not qualifier_exists("ul"))
	   {

	       if (length(dat.err) == 0)
	       {
		   vmessage("Only Upper Limits found, cannot load data");
		   
	       }
	   }


	   % warning: this is only a crude estimation, i.e., plotted values
	   % might not fit 100% the residuals!!
	   
	   variable gam = r.spectral_index[0];
	   variable corr_factor = get_corr_factor(dat,gam);
	   
	   dat = _A(dat); % change grid to angstrom, and do corresponding reversement of values
	   variable fermi_id = define_counts(dat);
	   set_bin_corr_factor(fermi_id,corr_factor);

	   % ignore upper limits (err=0) and unvalid bins
	   set_data_exposure (fermi_id,1); % should be already by default (but to be sure)

	   %IS THIS CAUSING TROUBLE? PROBABLY NOT
	   assign_rmf (make_fine_rmf(fermi_id,qualifier("rmf_factor",10)),fermi_id); % define a fine RMF

	   if (not qualifier_exists("ul"))
	   {
	       ignore_list (fermi_id, where(dat.err==0 or isnan(dat.value))); 

	   }


	   return fermi_id;
       
}

%}}}

%
% Code from src/fitting/load_grouped_xmm_data.sl
%
define load_grouped_xmm_data()
%!%+
%\function{load_grouped_xmm_data}
%\synopsis{wrapper for load_data, which sets already a few basic options}
%\usage{Integer_Type Index = load_grouped_xmm_data([String_Type Path], [E_lo, E_hi]);}
%\qualifiers{
%\qualifier{min_sn[=5]}{[\code{=5}]: Set the minimal S/N ratio (see help("group");) }
%\qualifier{min_chan[=2]}{[\code{=2}]: Set the minimal number of Channels do be rebinned (see help("group");) }
%\qualifier{delete_data}{If set, any data sets loaded previously will be deleted.}
%}
%\description
%   This functions loads data (by default "src_sd.pha"), groups it,
%   and notices the energy bins between E_lo[=1keV] and E_hi[=10keV]. 
%   When changing all default values, it can be easily used for any
%   other satellite.
%!%-
{
   variable spec, emin=1.,emax=10.;
   switch (_NARGS)
   {case 0: spec ="src_sd.pha" ; }
   {case 1: spec = (); }
   {case 2: spec ="src_sd.pha" ; (emin, emax) = (); }
   {case 3: (spec, emin, emax) = (); }
   { help(_function_name()); return; }

   if(qualifier_exists("delete_data")) delete_data(all_data);
   
   variable id = load_data(spec);

   
   group(id;
	 min_sn=qualifier("min_sn",5),
	 min_chan=qualifier("min_chan",2));
   variable rng = qualifier("range",[1,10]);
   ignore_en(id,,emin);
   ignore_en(id,emax,);

   return id;
}

%
% Code from src/fitting/parameters/load_par_from_FITS_header.sl
%
define load_par_from_FITS_header()
%!%+
%\function{load_par_from_FITS_header}
%\synopsis{loads the fit-function and parameters from a FITS file}
%\usage{load_par_from_FITS_header(String_Type filename);}
%\description
%    This function reads keywords from a FITS header
%    created with \code{save_par_to_FITS_header_struct}.
%\seealso{save_par_to_FITS_header_struct, save_par, load_par}
%!%-
{
  variable file;
  if(_NARGS==1)  file = ();
  else { help(_function_name()); return; }

  fit_fun( fits_read_key(file, "fit_fun") );
  variable i;
  _for i (1, fits_read_key(file, "num_pars"), 1)
  {
    variable istr = string(i);

    variable par_fun = fits_read_key(file, "par"+istr+"_fun");
    set_par_fun(i, (__is_numeric(par_fun) ? NULL : par_fun));

    variable par_min = fits_read_key(file, "par"+istr+"_min");  ifnot(__is_numeric(par_min))  par_min = -DOUBLE_MAX;
    variable par_max = fits_read_key(file, "par"+istr+"_max");  ifnot(__is_numeric(par_max))  par_max =  DOUBLE_MAX;
    set_par(i, fits_read_key(file, "par"+istr, "par"+istr+"_frz"), par_min, par_max);

    variable par_tie = fits_read_key(file, "par"+istr+"_tie");
    if(__is_numeric(par_tie))  untie(i);  else  tie(par_tie, i);
  }
}

%
% Code from src/fitting/load_par_tbnew.sl
%
define load_par_tbnew(filename)
%!%+
%\function{load_par_tbnew}
%\synopsis{load a parameter file after decreasing the upper limit for tbnew.PL}
%\usage{load_par_tbnew(String_Type filename);}
%\description
%    Early versions of the \code{TBnew} absorption model allowed
%    for a too large upper limit of the \code{PL} parameter of 5.
%    This function reduces the \code{PL} parameter to 3.999,
%    rewrites the parameter file, and loads it with \code{load_par}.
%\seealso{load_par}
%!%-
{
  variable fp = fopen(filename, "r");
  variable line, lines = fgetslines(fp);
  ()=fclose(fp);

  variable s = "[ \t]+";  % white spaces
  variable num = `\d*\.?\d*`;  % a number
  fp = fopen(filename, "w");
  foreach line (lines)
  {
    variable m = string_matches(line, `^\(.*[Tt][Bb]new(\d+)\.PL$s\d+$s[01]$s$num$s$num$s\)\($num\)`$);
    if(m!=NULL)
      ()=fprintf(fp, "%s%S\n", m[1], _min(3.999, atof(m[2])));
    else
      ()=fprintf(fp, "%s", line);
  }
  ()=fclose(fp);

  load_par(filename);
}

%
% Code from src/data/spectrum.loaddata/load_piled1storderHETGS_datasets.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define load_piled1storderHETGS_datasets()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{load_piled1storderHETGS_datasets}
%\synopsis{loads 1st order Chandra-HETGS data and sets up the simple_gpile model}
%\usage{Integer_Type ids[] = load_1storderHETGS_dataset(specpath, RMFpath);}
%!%-
{
  variable specpath, RMFpath;
  switch(_NARGS)
  { case 2: (specpath, RMFpath) = (); }
  { help(_function_name()); return; }

  variable ids = load_piledHETGS_datasets(specpath, RMFpath);
  delete_data(ids[[4:11]]);
  return ids[[0:3]];
}

%
% Code from src/data/spectrum.loaddata/load_piledHETGS_datasets.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define load_piledHETGS_datasets()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{load_piledHETGS_datasets}
%\synopsis{loads Chandra HETGS spectral and sets up the simple_gpile model}
%\usage{Integer_Type ids[] = load_1storderHETGS_dataset(specpath, RMFpath);}
%\seealso{load_HETGS_datasets, use_simple_gpile}
%!%-
{
  variable specpath, RMFpath;
  switch(_NARGS)
  { case 2: (specpath, RMFpath) = (); }
  { help(_function_name()); return;}

  variable s = load_HETGS_datasets(specpath, RMFpath);
  reshape(s, [3,4]);
  variable s1 = s[0,*];
  variable a2 = Integer_Type[4];
  variable a3 = Integer_Type[4];
  variable i;
  _for i (0, 3, 1)
  { a2[i] = get_data_info(s[1,i]).arfs[0];
    a3[i] = get_data_info(s[2,i]).arfs[0];
  }
  if(get_fit_fun() == "")  fit_fun("1");
  use_simple_gpile(s1, a2, a3);
  reshape(s, [12]);
  return s;
}

%
% Code from src/data/load_radio2.sl
%
define load_radio2()
%%%%%%%%%%%%%%%%
%!%+
%\function{load_radio2}
%
%\synopsis{reads and loads radio data}
%\usage{Integer_Type data_id = load_radio (String_Type filename);}
%
%\description
%    Takes data an ASCII file , comprised of three  columns: 
%    Frequency[Hz], Flux density [mJy], Error [mJy], and loads
%    the data as a dataset.
%    WARNING: The standard bin width is set to 10%
%
%    It can be changed with the qualifier binwidth (given in
%    fraction of the frequency). Since a log frequency grid is
%    used, it will not be exactly the size given.
%    WARNING: Bins cannot overlap, as grid needs to be continuous!
%    
%\qualifiers{
%\qualifier{binwidth}{   specify bin width}
%\qualifier{use_struct}{ use an ISIS structure with the
%                          following fields: freq[Hz],
%                          bandwidth[Hz], flux[mJy], err[mJy];
%                          with full bandwidth given}
%}
%
%\example
%    isis> variable data_file = "/some/path/radio_file.txt";
%    isis> variable radio_data = load_radio2(data_file);
%
%    example using a textfile that includes bin widths:
%
%    isis> variable abc = ascii_read_table ("radio_dat.txt",[{"%F","freq"},{"%F", "bandwidth"},{"%F","flux"},{"%F","err"}]);
%    isis> variable radio_data = load_radio2(abc; use_struct);
%
%\seealso{read_radio, load_radio}
%!%-
{
    variable in_fil;
    switch(_NARGS)
    { case 1: in_fil  = (); }
    { help(_function_name()); return; }

    variable binwidth = qualifier("binwidth", NULL);
    variable usestruct = qualifier("use_struct", NULL);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 1. Try loading the user specified file

    if (not qualifier_exists("use_struct"))
    {
	try
	{
	    variable load_ascii = ascii_read_table(in_fil,[{"%F", "freq"},{"%F", "flux"},{"%F","err"}]);
	}
	catch AnyError:
	{
	    vmessage("*** Error: File %s could not be opened", in_fil);
	}
    }
    else
    {
	load_ascii = @in_fil;
    }

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 2. Units: Flux density mJy -> ph/cm^2/s/Hz

    variable mjy2cgs = 1.0e-26;
    load_ascii.flux = load_ascii.flux/Const_h/load_ascii.freq*mjy2cgs;
    load_ascii.err = load_ascii.err/Const_h/load_ascii.freq*mjy2cgs;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 3. Units: Frequency -> Angstrom & sort

    if (qualifier_exists("use_struct"))
    {
	%Get fractional bin width
	variable bin_w_struct = load_ascii.bandwidth/load_ascii.freq;
    }

    variable wavelength = Const_c/load_ascii.freq*1e8;
    load_ascii = struct_combine(load_ascii,struct{wave = wavelength});
    variable sorted = sort_struct_arrays(load_ascii, "wave");

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 4. Make lo & hi grid

    variable lo_grid, hi_grid;

    if (qualifier_exists("use_struct"))
    {
	lo_grid = sorted.wave-bin_w_struct*sorted.wave/2.;	
	hi_grid = 10^(log10(sorted.wave-lo_grid))+sorted.wave;
    }
    else if (not qualifier_exists("binwidth") && 
    not qualifier_exists("use_struct"))
    {
	binwidth=0.1;
	lo_grid = sorted.wave-binwidth*sorted.wave/2.;
	hi_grid = 10^(log10(sorted.wave-lo_grid))+sorted.wave;
    }
    else if (qualifier_exists("binwidth") )
    {
	lo_grid = sorted.wave-binwidth*sorted.wave/2.;
	hi_grid = 10^(log10(sorted.wave-lo_grid))+sorted.wave;
    }

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 5. Make continuous grid

    variable full_grid_lo = Double_Type[0];
    variable full_grid_hi = Double_Type[0];
    variable full_flux = Double_Type[0];
    variable full_err = Double_Type[0];

    variable i;
    _for i(0, length(lo_grid)-1,1)
    {
	if (i<length(lo_grid)-1)
	{
	    full_grid_lo=[full_grid_lo,lo_grid[i]];
	    full_grid_hi=[full_grid_hi,hi_grid[i]];
	    full_flux = [full_flux,sorted.flux[i]];
	    full_err = [full_err,sorted.err[i]];

	    % Check if bins are continuous
	    %if not add empty bin

	    if (lo_grid[i+1] != hi_grid[i])
	    {
		full_grid_lo=[full_grid_lo,hi_grid[i]];
		full_grid_hi=[full_grid_hi,lo_grid[i+1]];
		full_flux = [full_flux,0.];
		full_err = [full_err,0.];
	    }
   	}
	else
	{
	    full_grid_lo=[full_grid_lo,lo_grid[i]];
	    full_grid_hi=[full_grid_hi,hi_grid[i]];
	    full_flux = [full_flux,sorted.flux[i]];
	    full_err = [full_err,sorted.err[i]];
	}
    }
    


    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 6. Integrate flux and make final structure

    variable flux_int, err_int, final_struct;
    
    flux_int = full_flux*Const_c*1.e8*(full_grid_hi-full_grid_lo)/full_grid_lo/full_grid_hi;
    err_int = full_err*Const_c*1.e8*(full_grid_hi-full_grid_lo)/full_grid_lo/full_grid_hi;

    final_struct = struct{bin_lo = full_grid_lo, bin_hi = full_grid_hi, value=flux_int, err = err_int};

    variable radio_id = define_counts(final_struct);
    set_data_exposure(radio_id,1.);
    ignore_list(radio_id, where (final_struct.value == 0.));

    return radio_id; 
}

%
% Code from src/slang/compatibility/log1p.sl
%
#ifnexists log1p
define log1p {
    return log(1.+p);
}
#endif

%
% Code from src/histogram/log_grid.sl
%
%%%%%%%%%%%%%%%
define log_grid()
%%%%%%%%%%%%%%%
%!%+
%\function{log_grid}
%\synopsis{generate a logarithmic histogram grid}
%\usage{(bin_lo[], bin_hi[]) = log_grid(min, max, nbins);}
%\description
%    This function is an shorthand form of:\n
%       \code{(log_bin_lo, log_bin_lo) = linear_grid( log(min), log(max), nbins );}\n
%       \code{bin_lo = exp( log_bin_lo );}\n
%       \code{bin_hi = exp( log_bin_hi );}
%\seealso{linear_grid}
%!%-
{
  variable mn, mx, nbins;
  switch(_NARGS)
  { case 3: (mn, mx, nbins) = (); }
  { help(_function_name()); return; }

  variable log_bin_lo, log_bin_hi;
  (log_bin_lo, log_bin_hi) = linear_grid( log(mn), log(mx), nbins );
  return (exp(log_bin_lo), exp(log_bin_hi));
}

%
% Code from src/fitting/fit-functions/log_parabola.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%!%+
%\function{log_par}
%\synopsis{log-log parabolic fit function}
%\description
%    Fit function of the form F(x)= exp ( a*(log(x) - center)^2 + peak );
%    For fitting the function is implemented in integrated from.
%    \code{center} has to be specified in the unit in which the bins are given.
%    The function is a parabola in the log-log space of the unit of the bins.
%    Currently the function only works for negative curvature \code{a<0}.
%
%    For a log-parabola in the energy/frequency regime the function \code{log_par_en}
%    has to be used.
%    
%\seealso{log_par_en}
%!%-

define log_par_fit (lo, hi, par)
{
  variable t = par[0];
  variable s = par[1];
  variable a = par[2];

  if (a == 0) return t*(hi-lo);
  s = log(s);
  t = log(t);
  
  variable L = log(lo);
  variable H = log(hi);

  variable inv_sqrt_a = 1./sqrt(abs(a));
  variable b = 1.-2*a*s;
  if (a < 0)
  {
    return exp(t+a*s^2 - 0.25* (b^2)/a )*0.5*(sqrt(PI)* (gsl->erf(0.5* (2*a*L+b)*inv_sqrt_a)-
							       gsl->erf(0.5* (2*a*H+b)*inv_sqrt_a)))*inv_sqrt_a;
  }
  if (a > 0)
  {
    message("a>0 does not work yet (complex error function required, Faddeeva function)");
    return -exp(t+a*s^2)*0.5*(sqrt(PI)* exp((-0.25* (b^2)/a))* (gsl->erf(0.5* (2*a*L+b)*inv_sqrt_a)-
							       gsl->erf(0.5* (2*a*H+b)*inv_sqrt_a)))*inv_sqrt_a;
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%
define log_par_defaults(i)
%%%%%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
  { case 0: return    1, 0, 1e-30, 1e10; }
  { case 1: return    2, 0, 1e-10, 1e30; }
  { case 2: return   -1, 0,  -1e4,    0; }
}

add_slang_function("log_par", ["peak","center","curvature"], [0]);
set_param_default_hook("log_par", "log_par_defaults");



%!%+
%\function{log_par_en}
%\synopsis{log-log parabolic fit function in the energy/frequency regime}
%\description
%    Fit function of the form F(nu)= exp ( a*(log(nu) - center)^2 + peak );
%    For fitting the function is implemented in integrated from.
%    \code{center} has to be specified in Hz. The bins are expected to be
%    given in Angstrom (as it is typically the case in ISIS).
%    Currently the function only works for negative curvature \code{a<0}.
%    
%\seealso{log_par}
%!%-

define log_par_en_fit (bin_lo, bin_hi, par)
{
  variable s = par[1] * 4.135667516e-18;% convert s from Hz to keV (multiply with h = 4.135667516e-18 [keV * s] )
  return reverse( log_par_fit (_A(bin_hi) , _A(bin_lo) , [par[0], s, par[2]] ));
}

%%%%%%%%%%%%%%%%%%%%%%%%
define log_par_en_defaults(i)
%%%%%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
  { case 0: return    1, 0, 1e-30, 1e10; }
  { case 1: return    1e18, 0, 1e-10, 1e30; }
  { case 2: return   -1, 0,  -1e4,    0; }
}

add_slang_function("log_par_en", ["peak [ph/cm^2/s^1/keV at center]","center [Hz]","curvature"], [0]);
set_param_default_hook("log_par_en", "log_par_en_defaults");


#endif

%
% Code from src/data/logging.sl
%
define logging()
%!%+
%\function{logging}
%
%\synopsis{write a log file}
%\usage{logging ("filename.log", "Log message");}
%
%\description
%    This function writes output to a log file (which will be created if it does
%    not exist. It can also handle input in the sprintf format (see example).
%    By default all input to logging is also printed to stdout.
%    Repeated calls to the function add the new message at the end of the file. It
%    is therefore recommended to specify the date and time in the name of the log
%    file.
%    
%\qualifiers{
%\qualifier{v}{verbosity: [default: 1] print output also to stdout, set this
%                              qualifier to 0 for no output to stdout}
%}
%
%\example
%    isis> variable cutime = strftime("%Y-%m-%d_%H:%M:%S");
%    isis> variable l = cutime+".log";
%    isis> logging(l, "* Found %.u observations for %s", length(src_info), src);
%    isis> logging(l, "***Error in function %s", _function_name() );
%
%\seealso{strftime, _function_name()}
%!%-
{

    if (_NARGS < 2)
    { 
	help(_function_name()); return; 
    }


    variable filename, msg;
    variable v = qualifier("v", 1);
    
    variable args;
    args=__pop_args(_NARGS);
    
    filename = __push_args(args[0]);
    msg = __push_args(args[1]);
 
    variable args2 = array_remove(args,0);
    args2 = array_remove(args2,0);

    variable a = sprintf(msg, __push_args(args2) );

    if (v == 1)
    {
	message(a);
    }

    variable F = fopen(filename, "a+");
    () = fwrite(a,F);
    () = fwrite ("\n",F);
    () = fclose(F);
}

%
% Code from src/misc/math/lorentz_trafo.sl
%

% only activate if module vector could be loaded
#ifeval __get_reference("vector")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%
define lorentz_trafo()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{lorentz_trafo}
%\synopsis{lorentz transformation for a boost in any direction}
%\usage{lorentz_trafo( Double_Type ct,
%                      Vector_Type x,
%                      Vector_Type B
%                    );
%}
%\description
%    This function performs a Lorentz transformation:
%      ( ct, x ) -> ( ct', x')
%      ct' = gamma ( ct - B*x )
%      x'  = x + ( (gamma-1)/B^2 * B*x - gamma * ct ) * B
%    where x is the spartial vector, gamma the Lorentz-factor
%    and B=v/c the velocity vector.
%    
%\seealso{Vector_Type}
%!%-
{
  variable ct, x, B;
  switch(_NARGS)
  { case 3: ( ct, x, B) = (); }
  { help(_function_name()); return; }

  variable BB = dotprod( B, B );
  variable xB = dotprod( x, B );

  variable g = 1. / sqrt( 1. - BB );
  
  
  variable ctp = g * ( ct - xB );
  variable xp = x + vector_mul( (g-1)/BB*xB - g*ct, B );

  return ( ctp, xp );
}

#endif

%
% Code from src/fitting/xyfit/fit-functions/lorentz_xyfit.sl
%
define lorentz_xyfit()
%!%+
%\function{lorentz_xyfit}
%\synopsis{Lorentzian xy fit function to be used with xyfit_fun}
%\usage{xyfit_fun ("lorentz");}
%\description
%    Calling \code{xyfit_fun ("lorentz");} sets up a Lorentzian fit
%    function for xy-data. It has the form 
%    \code{y = norm * sigma/(2pi) * 1/((x-center)^2 + (sigma/2)^2)}
%\example
%    variable id = define_xydata(x, y, yerr);
%    xyfit_fun("lorentz");
%    () = fit_counts;
%    variable xfit, yfit;
%    (xfit, yfit) = eval_xyfun([min(x) : max(x) : #1000]);
%    plot(xfit, yfit);
%\seealso{xyfit_fun, define_xydata, plot_xyfit, linear_regression}
%!%-
{
  variable xref, yref, par;
  switch(_NARGS)
  { case 0: return ["norm", "center [center]", "sigma [sigma]"];} % fit parameters
  { case 3: (xref, yref, par) = (); }
  { return help(_function_name); }

  @yref = par[0] * par[2]/(2*PI) * 1./((@xref - par[1])^2 + (par[2]/2.)^2);
}

define lorentz_xyfit_default(i)
{
  switch(i)
  { case 0: return (0.01, 0, -1e10 , 1e10); }
  { case 1: return (0   , 0,     0 , 1e5); }
  { case 2: return (1   , 0,  1e-6 , 1e6); }
}

%
% Code from src/fitting/fit-functions/lorentzmb.sl
%
%%%%%%%%%%%%%%%%%%%%
define lorentzmb_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{lorentzmb (fit-function)}
%\synopsis{implements a Lorentzian profile}
%\description
%    \code{L(f)  =  [2 rms^2 Q f0]/[pi/2 + atan(2*Q)] * 1/[f0^2 + 4 Q^2 (f-f0)^2]}\n
%    [see also Eq. (1) in Pottschmidt et al. (2003), A&A 407, 1039-1058]\n
%    where:\n
%    \code{rms} = contribution to the root mean square variability\n
%    \code{ Q } = quality factor\n
%    \code{ f0} = resonance frequency = \code{nu0/sqrt(1 + 1/[4 Q^2])}\n
%    \code{nu0} = peak frequency (maximum of \code{f * L(f)})
%
%    The parameters of the lorentzmb fit-function are
%    "rms" (= \code{rms}), "peakfr" (= \code{nu0}) and "quality" (= \code{Q}).
%
%    The integrated fit-function is:\n
%    \code{int L(f) df  =  rms^2/[pi/2 + atan(2*Q)] * arctan{2Q/pi*(f/f0-1)}}
%!%-
{
   variable rms = par[0];
   variable f = par[1]/sqrt(1.0+1/(4*par[2]^2));
   variable q = par[2];
   return rms^2/(0.5*PI + atan(2*q)) * ( atan( 2.0*q/f*(-bin_lo+f)) - atan( 2.0*q/f*(-bin_hi+f) ) );
}

%%%%%%%%%%%%%%%%%%%%%%%%
define lorentzmb_default(i)
%%%%%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
  { case 0: return (0   , 0   , 0   , 1   ); }
  { case 1: return (1   , 0   , 1e-5, 1e+3); }
  { case 2: return (0.5 , 0   , 0   , 10  ); }
}

add_slang_function("lorentzmb", ["rms", "peakfr [Hz]", "quality"]);
set_param_default_hook("lorentzmb", &lorentzmb_default);

%
% Code from src/fitting/fit-functions/lorentzmb_old.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define lorentzmb_old_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{lorentzmb_old (fit-function)}
%\synopsis{implements a Lorentzian profile}
%\description
%    \code{L(f)  =  1/pi * [2 R^2 Q f0] / [f0^2 + 4 Q^2 (f-f0)^2]}\n
%    [Eq. (1) in Pottschmidt et al. (2003), A&A 407, 1039-1058]\n
%    where:\n
%    \code{ R } = normalization constant\n
%    \code{ Q } = quality factor\n
%    \code{ f0} = resonance frequency = \code{nu0/sqrt(1 + 1/[4 Q^2])}\n
%    \code{nu0} = peak frequency (maximum of \code{f * L(f)})
%
%    The parameters of the lorentzmb fit-function are
%    "norm" (= \code{R}), "peakfr" (= \code{nu0}) and "quality" (= \code{Q}).
%
%    The integrated fit-function is:\n
%    \code{int L(f) df  =  R^2/pi * arctan{2Q/pi*(f/f0-1)}}
%!%-
{
   variable R  = par[0];
   variable Q  = par[2];
   variable f0 = par[1]/sqrt(1+0.25/Q^2);
   return R^2/PI * ( atan(2.0*Q*(bin_hi/f0-1)) - atan(2.0*Q*(bin_lo/f0-1)) );
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define lorentzmb_old_default(i)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
  { case 0: return (0   , 0   , 0   , 1   ); }
  { case 1: return (1   , 0   , 1e-5, 1e+3); }
  { case 2: return (0.5 , 0   , 0   , 10  ); }
}

add_slang_function("lorentzmb_old", ["norm", "peakfr [Hz]", "quality"]);
set_param_default_hook("lorentzmb_old", &lorentzmb_old_default);

%
% Code from src/slang/astronomy/lumdist.sl
%
%%%%%%%%%%%%%%%%%
private define ldist(z)
%%%%%%%%%%%%%%%%%
{
  variable omega_lambda = qualifier("omega_lambda",   0.7);
  variable           q0 = qualifier("q0"          , -0.55);

  variable term1 = (1.+z)^2;
  variable term2 =  1.+2.*(q0+omega_lambda)*z;
  variable term3 = z*(2.+z)*omega_lambda;
  variable denom = (term1*term2 - term3);

  return 1./sqrt(denom);
}

%%%%%%%%%%%%%%%%%
define lumdist()
%%%%%%%%%%%%%%%%%
%!%+
%\function{lumdist}
%\synopsis{Calculate luminosity distance (in Mpc) of an object given its redshift}
%\usage{result = lumdist(z);}
%\qualifiers{
%\qualifier{silent}{If set, the program will not display adopted
%                       cosmological parameters at the terminal.}
%\qualifier{h0}{[=70] Hubble parameter in km/s/Mpc}
%\qualifier{omega_k}{curvature constant, normalized to the closure density.
%                       Default is 0, indicating a flat universe.}
%\qualifier{omega_m}{Matter density, normalized to the closure density,
%                       default is 0.3. Must be non-negative.}
%\qualifier{omega_lambda}{: Cosmological constant, normalized to the
%                       closure density (default: 0.7).}
%\qualifier{q0}{[=-0.55] Deceleration parameter, numeric scalar = -R*(R'')/(R')^2}
%}
%\description
%    INPUTS:    z = redshift (positive scalar or vector)
%    OUTPUTS:   The result of the function is the luminosity distance (in Mpc)
%               for each input value of z.
%    The luminosity distance in the Friedmann-Robertson-Walker model is 
%    taken from  Caroll, Press, and Turner (1992, ARAA, 30, 499), p. 511.
%    Uses a closed form (Mattig equation) to compute the distance when the 
%    cosmological constant is zero.   Otherwise integrates the function using
%    QSIMP.	
%    See help for cosmo_param for a description of the cosmological parameters,
%    note that only two out of the four parameters should be given!
%
%    The routine can fail to converge at high redshift for closed universes with
%    non-zero lambda. This can presumably be fixed by replacing QSIMP with
%    an integrator that can handle a singularity.
%
%\example
%    %  Plot the distance of a galaxy in Mpc as a function of redshift out 
%    %  to z = 5.0, assuming the default cosmology (Omega_m=0.3, Lambda = 0.7,
%    %  H0 = 70 km/s/Mpc)
%    variable z = [0:5:0.1];
%    plot (z, lumdist(z));
%
%    %  Now overplot the relation for zero cosmological constant and 
%    %  Omega_m=0.3
%    oplot (z, lumdist(z ; omega_lambda=0 , omega_m=0.3) );
%\seealso{cosmo_param, qsimp}
%!%-
{
  variable           h0 = qualifier("h0"     ,         70);
  variable      omega_k = qualifier("omega_k",       NULL); 
  variable      omega_m = qualifier("omega_m",       NULL);
  variable omega_lambda = qualifier("omega_lambda",  NULL);
  variable           q0 = qualifier("q0"     ,       NULL);
  variable            c = qualifier("c"      ,   2.9979e5); % speed of light in km/s 

  variable z;
  switch(_NARGS)
  {
   case 1: z = ();
  }
  { help(_function_name()); return; }

  variable n = length(z);

  % obtain cosmological params from cosmo_param
  variable cosmopar = cosmo_param(; omega_k = omega_k, omega_lambda = omega_lambda,
				  omega_m = omega_m, q0 = q0);
  omega_k      = cosmopar.omega_k;
  omega_lambda = cosmopar.omega_lambda;
  omega_m      = cosmopar.omega_m;
  q0           = cosmopar.q0;
  
  ifnot (qualifier_exists ("silent"))
    vmessage("lumdist:\nh0: %5.2f, omega_m: %5.2f, omega_lambda: %5.2f, q0: %5.2f, omega_k: %5.2f"
	     , h0, omega_m, omega_lambda, q0, omega_k );

  % For the case of lambda = 0, we use the closed form from equation 5.238 of
  % Astrophysical Formulae (Lang 1998). This avoids terms that almost cancel
  % at small q0*z better than the more familiar Mattig formula.
  variable dlum;
    
  if (omega_lambda == 0)
  {
    variable denom = sqrt(1.+2.*q0*z) + 1. + q0*z;
    dlum = (c*z/h0)*(1. + z*(1.-q0)/denom);
    return dlum;
  } else {
    % For non-zero lambda 
    variable i;
    variable tz = typeof(z);
    if (tz != Array_Type) z = [z];
    dlum = z*0.;
    _for i(0,n-1,1)
    {
      if (z[i] <= 0.0) dlum[i] = 0.0;
      else
      {
	dlum[i] = qsimp ( &ldist, 0, z[i] ;; struct_combine ( __qualifiers, struct {q0 = q0, omega_lambda = omega_lambda}) );
      }
    }
    
    if (omega_k > 0.) dlum = sinh(sqrt(omega_k)*dlum)/sqrt(omega_k);
    if (omega_k < 0.) 
    {
      dlum = sin(sqrt(-omega_k)*dlum)/sqrt(-omega_k);
      if (dlum < 0) dlum = 0;
    }

    if (tz != Array_Type) { z = z[0]; dlum = dlum[0]; }
    return, c*(1.+z)*dlum/h0;
  }
}

%
% Code from src/data/spectrum.datastructures/luminosity.sl
%
%%%%%%%%%%%%%%%%%
define luminosity()
%%%%%%%%%%%%%%%%%
%!%+
%\function{luminosity}
%\synopsis{computes a source luminosity assuming the current fit-function and a distance}
%\usage{Double_Type luminosity(Double_Type Emin, Emax, d);}
%\qualifiers{
%\qualifier{factor}{[=\code{1.001}] step of logarithmic energy grid}
%\qualifier{Emin}{minimum energy of (extended) grid}
%\qualifier{Emax}{maximum energy of (extended) grid}
%}
%\description
%    returns \code{4pi (d kpc)^2 * int_Emin^Emax E*S_E(E) dE}  (in erg/s)
%
%    Use flux2lum to calculate luminosities for extragalactic sources.
%
%\seealso{energyflux, flux2lum}
%!%-
{
  variable Emin, Emax, d;
  switch(_NARGS)
  { case 3: (Emin, Emax, d) = (); }
  { help(_function_name()); return; }

  return 4 * PI * (d*3.08568025e21)^2   % kpc -> cm
         * energyflux(Emin, Emax;; __qualifiers) * 1.60217646e-9;   % keV -> erg
}

%
% Code from src/data/make_fine_rmf.sl
%
%%%%%%%%%%%%%%%%
%!%+
%\function{make_fine_rmf}         
%\synopsis{creates a fine RMF }
%\usage{make_fine_rmf (Data_Id [Integer], Factor [Integer]);}
%\description
%   Creates a fine RMF. Therefore each data bin will be split into 
%   a certain number of bins, specified by the variable "Factor".
%\seealso{load_slang_rmf}   
%!%-
private define rmf_profile (bin_lo, bin_hi, x, parms)
{
   variable rmf = Double_Type[length(bin_lo)];
   variable i = where (bin_lo < x <= bin_hi);
   rmf[i] = 1.0;
   return rmf/sum(rmf);
}

define make_fine_rmf()
{
   variable iDat,fac;
   switch(_NARGS)
   { case 2: (iDat,fac) = ();}
   { help(_function_name()); return; }
   variable dat = _A(get_data_counts(iDat));
   fac = int(fac);
   
   if (fac <= 0) 
     return -1;
   
   variable i, ind, n = length(dat.bin_lo);
   variable lo = Double_Type[n*fac], hi = @lo;
   
   _for i(0,n-1,1)
   {
      ind = [i*fac:(i+1)*fac-1];
      (lo[ind],hi[ind]) = log_grid(dat.bin_lo[i],dat.bin_hi[i],fac);
   }   
   return load_slang_rmf (&rmf_profile,dat.bin_lo,dat.bin_hi,lo,hi;parms=fac,grid="en");
}

%
% Code from src/sixte/make_movie_from_evtList.sl
%

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

private define write_movie_img_png(fname,img,cmap){ %{{{

   variable rgb_img,pl,fov,t;
   if (not qualifier_exists("labels")){
      rgb_img  = png_gray_to_rgb(img, cmap) ;
      png_write_flipped(fname, rgb_img) ;

   } else {
   
      pl = xfig_plot_new(10,10);
      if (qualifier_exists("gmax"))
	pl.plot_png (img; cmap="hot",gmax=qualifier("gmax"));
      else
	pl.plot_png (img; cmap="hot");
      fov = qualifier("fov",1.);
      pl.world(-fov/2,fov/2,-fov/2,fov/2);
      pl.xaxis(;off,color="white");
      pl.yaxis(;off,color="white");
      pl.plot([0,0],[0,1];world0,color="white",width=2);
      pl.plot([1,1],[0,1];world0,color="white",width=2);
      pl.plot([0,1],[0,0];world0,color="white",width=2);
      pl.plot([0,1],[1,1];world0,color="white",width=2);

      t = qualifier("t",NULL);
      if (t != NULL){
	  pl.add_object(xfig_new_text(
				      sprintf(`$t=%.3f$\,ksec`,t*1e-3)
				     ;color="white",fontsize=`small`),
		       0.96,0.96,0.5,0.5;world0);
      }
      
      pl.render(fname);
   }
}
%}}}



%%%%%%%%%%%%%%%%%%%%%%%
define make_movie_from_evtList(fevt,dt,output){ %{{{
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{make_movie_from_evtList}
%\synopsis{makes a simple movie from a given event list}
%\usage{Double_Type total_img = make_movie_from_evtList(string eventFile, double dt, string movieFile);}
%\description
%    Make a movie from the eventlist "eventFile". The time (in
%    seconds) of each frame is adjusted by "dt". The output file is
%    written to "movieFile" (best use a mp4 file). 
%\qualifiers{
%\qualifier{cmap}{["hot"] color map}
%\qualifier{thres}{[0.0] threshold of counts/frame, when a frame should be discarded}
%\qualifier{fps}{frames per secondswitch on the scale}
%\qualifier{size}{[10,10] size of the image in cm}
%\qualifier{tstart}{start time of the movie (sec)}
%\qualifier{tstop}{end time of the movie (sec)}
%\qualifier{labels}{draw some labels (white frame and time in ksec)}
%}
%\seealso{xfig_plot_new,fitswcs_get_img_wcs}
%!%-

   variable evt = fits_read_table(fevt);
   
   variable fpng = qualifier("fpng","evt_img_%06d.png");
   variable fps = qualifier("fps",30);
   variable tstart = qualifier("tstart",min(evt.time));
   variable tstop = qualifier("tstop",max(evt.time));
   
   variable nx = max(evt.rawx)+1, ny=max(evt.rawy)+1;
   variable img = Double_Type[nx,ny];
   
   variable i, n = int((tstop-tstart)/dt)+1;
   
   variable ind,t0,t1,id;
   variable img_sum = @img;

   img_sum[*,*]=0;

   variable gmax=0.0;
   
   variable t = Double_Type[0];
   variable img_ar = Array_Type[0];
   variable img_tmp = Array_Type[1];
   _for i(0,n-1,1){
   
      t0 = tstart+i*dt;
      t1 = tstart+(i+1)*dt;
      ind = where( t0 <= evt.time < t1);
      
      img[*,*] = 0;
      foreach id(ind){
	 img[evt.rawx[id],evt.rawy[id]]++;
      }      

      if (sum(img) > qualifier("thres",0.0)){
	 gmax = max([gmax,max(img)]);
	 img_tmp[0] = [@img];
	 img_ar = [img_ar,img_tmp];
	 t = [t,0.5*(t0+t1)];
	 img_sum += img;
      }
      
   }

   variable cmap = qualifier("cmap","hot");
   
   variable npng = length(img_ar);
   _for i(0,npng-1,1){
      if (qualifier_exists("labels"))
	write_movie_img_png(sprintf(fpng,i),log(img_ar[i]),cmap
		      ;labels,t=t[i],gmax=log(gmax) );
      else
	write_movie_img_png(sprintf(fpng,i),log(img_ar[i]),cmap;
			    gmax=log(gmax) );
   }

      
      
   variable cmd = sprintf("ffmpeg -y -f image2 -r %.2f -i %s %s"$,fps,fpng,output);
   message(cmd);
   () = system(cmd);

   _for i(0,npng-1,1){
      () = remove(sprintf(fpng,i));
   }

   return img_sum;
}
%}}}


#endif

#endif

%
% Code from src/slang/system/make_path.sl
%
define make_path ()
%!%+
%\function{make_path}
%\synopsis{Recursively create multiple directories}
%\usage{make_path (dirs)}
%\description
%   This function creates multiple directory paths. It is
%   similar to mkdir_rec, but does not internally change
%   directories, allows multiple path specifications, and
%   individual modes.
%
%   The function throws an IO error if a path component exists
%   and is not a directory.
%
%\qualifiers{
%\qualifier{mode}{mode for the paths to be generated. If mode is an 
%                 array, it contains the mode for each individual path
%                 specified. Default: 0777}
%\qualifier{separator}{path separator. Default is /, so you will
%                  probably not have to use it...}
%}
%\example
% make_path(["./test1/test2","./test1/../test3"];mode=[0777,0700]);
%\seealso{mkdir,mkdir_rec}
%!%-
{
    variable pp=();
    variable mode=qualifier("mode",0777);

    % special case of one string argument
    if (typeof(pp)==String_Type) {
	pp=[pp];
    }

    % sanity check on mode
    if (length(mode)>1) {
	if (length(mode) != length(pp) ) {
	   throw UsageError,sprintf("%s: Must either give one mode or one per path.\n",_function_name());
	}
    } else {
        mode=mode+Integer_Type[length(pp)];
    }

    variable sep=qualifier("separator","/");

    variable i,j;
    _for i(0,length(pp)-1,1) {
	variable comp=strsplit(pp[i],sep);
	variable pa="";
	_for j(0,length(comp)-1,1) {
	    pa+=comp[j];

            % generate the directory. Catch
            % failures except for those where
            % a directory already exists

	    if (pa != "") {
		variable ret=mkdir(pa,mode[i]);
		if (ret==-1) {
		    if (errno!=EEXIST) {
			throw IOError,sprintf("mkdir %s failed: %s\n",pa,errno_string(errno));
		    } else {
			% give an error if the path component exists and
			% is NOT a directory
			variable st=stat_file(pa);
			if (stat_is("dir",st.st_mode)==0) {
			    throw IOError,sprintf("mkdir %s failed: path component exists and is not a directory",pa);
			}
		    }
		}
	    }
            % don't add the separator at the start of the j-loop:
            % this will cause the stat_file call above to fail
            pa+=sep; 
	}
    }
}

%
% Code from src/misc/radio/make_spix.sl
%
%%%%%%%%%%%%%%%%
define make_spix()
%%%%%%%%%%%%%%%%
%!%+
%\function{make_spix}
%\synopsis{creates a spectral index map using two .fits files (provided by DIFMAP)}
%\usage{make_spix(String_Type \code{name_lo}, String_Type \code{name_hi});}
%\qualifiers{
%\qualifier{iterations}{[=1] set number of translations to average over}
%\qualifier{cc}{[=cc.fits] name of the cross correlation map.  Will generate
%                              one if the file does not exist.}
%\qualifier{lothreshold}{[=NULL] if image has a pixel below this value it's set to
%                              this value. also criterion for showing spectral index if
%                              req_both_bands is set to 1}
%\qualifier{hithreshold}{[=NULL] if image has a pixel below this value it's set to
%                              this value. also criterion for showing spectral index if
%                              req_both_bands is set to 1}
%\qualifier{shift}{[={NULL,NULL}] define the shift of the second image wrt the first
%                              in format {x,y} in mas}
%\qualifier{crpix_shift}{[=0.] Extra shift for converting mas in pixel. In the past, 0.5 was used, 
%					but default is now  0., as only a relative is needed}
%\qualifier{dec_step}{[=NULL] pixel size in declination (value>0). 
%				 	Default value is the largest pixel size of the two images}
%\qualifier{dec_size}{[=NULL] mapsize size in declination. 
%					Default value is the largest map size of the two images}
%\qualifier{ra_step}{[=NULL] pixel size in right ascension (value>0). 
%					Default value is the largest pixel size of the two images}
%\qualifier{ra_size}{[=NULL] mapsize size in right ascension
%					Default value is the largest map size of the two images}
%\qualifier{beam}{[=[NULL,NULL,NULL]] defines beam for restoring the two images 
%						[semi-major axis, semi-minor axis, position angle] 
%						in mas (major and minor axis) and degree (position angle)}
%\qualifier{req_both_bands}{[=0] only make calculations if both bands are above the
%                              noise level}
%\qualifier{fit_noise}{[=1] set to 0 to use fits header for noise information}
%\qualifier{lo_nsigma}{[=3] number of standard deviations from mean to define
%                              noise limit.  Only matters if fit_noise==1,
%                              overrides lothreshold}
%\qualifier{hi_nsigma}{[=3] number of standard deviations from mean to define
%                              noise limit.  Only matters if fit_noise==1,
%                              overrides hithreshold}
%\qualifier{n_beams}{[=3] size of beam to exclude from core on correlation
%                              calculation}
%\qualifier{excl_core}{[=1] set to 0 if you do not want the program to
%                              automatically exclude the core from correlation calculations}
%\qualifier{overwrite}{overwrite the restore fits files (if already existing)}
%}
%\description
%    This function creates a spectral index (F=v^a) map from two fits files provided by
%    DIFMAP.  A cross correlation fits image is generated and the images are shifted
%    by the best (iterations) translations and the spectral index, etc. are
%    calculated for each translations. The values are averaged with the weights
%    given by the corresponding value in the cc image.
%    If the provided images do not have the same size, resolution and beam, the
%    functions difmap_restore and enclosing_ellipse are used to obtain images with
%    these properties. It is also possible to specify a desired map size and 
%    pixel size for both axes individually, otherwise  the default value from the comparison 
%    of both images will be used. Note, that changing the pixel size may require changing the maps size as well.
%
%    It returns a structure that holds the information for the following values\n
%    Use the structure as an input to\n
%    struct_name.spec_map      ->   matrix of calculated spectral index values\n
%    struct_name.stdev         ->   matrix of weighted stdevs of calculated values\n
%    struct_name.avg_lum       ->   matrix of weighted average of the summed BRIGHTNESS of the images\n
%    struct_name.weights       ->   matrix of sum of weights used for calculations in each pixel\n
%    struct_name.avg_shift     ->   average shifts in a 2-element array [x_shift,y_shift]\n
%    struct_name.shift_weight  ->   weights for calculating avg shift [x_weight,y_weight]\n
%    struct_name.avg_shift_pixel -> average shift in pixel [x_shift_pixel, y_shift_pixel]\n
%    struct_name.ra_px_center  ->   x index of center pixel\n
%    struct_name.ra_steps      ->   x mas per pixel\n
%    struct_name.dec_px_center ->   y index of center pixel\n
%    struct_name.dec_steps     ->   y mas per pixel\n
%    struct_name.major         ->   semi major axis of beam in mas
%    struct_name.minor         ->   semi minor axis of beam in mas
%    struct_name.source        ->   source name
%    struct_name.date          ->   dates of images
%    struct_name.posang        ->   position angle of beam in degrees
%\seealso{plot_spix, write_spix, read_spix, difmap_restore, enclosing_ellipse}
%!%-
{
  variable name_lo, name_hi;
  switch(_NARGS)
  { case 2: (name_lo, name_hi) = (); }                                          
  { help(_function_name()); return; }
  
  variable naxis1_lo = fits_read_key (name_lo, "NAXIS1");
  variable cdelt1_lo = fits_read_key (name_lo, "CDELT1");
  variable naxis2_lo = fits_read_key (name_lo, "NAXIS2");
  variable cdelt2_lo = fits_read_key (name_lo, "CDELT2");
  variable bmaj_lo   = fits_read_key (name_lo, "BMAJ")  ;
  variable bmin_lo   = fits_read_key (name_lo, "BMIN")  ;
  variable bpa_lo    = fits_read_key (name_lo, "BPA")   ;

  variable naxis1_hi = fits_read_key (name_hi, "NAXIS1");
  variable cdelt1_hi = fits_read_key (name_hi, "CDELT1");
  variable naxis2_hi = fits_read_key (name_hi, "NAXIS2");
  variable cdelt2_hi = fits_read_key (name_hi, "CDELT2");
  variable bmaj_hi   = fits_read_key (name_hi, "BMAJ")  ;
  variable bmin_hi   = fits_read_key (name_hi, "BMIN")  ;
  variable bpa_hi    = fits_read_key (name_hi, "BPA")   ;
  
  %% check if beams, map sizes, and resolution are identical in both images
  %% and create images with such properties if this is not the case
  if (
      naxis1_lo  !=  naxis1_hi or cdelt1_lo  !=  cdelt1_hi or
      naxis2_lo  !=  naxis2_hi or cdelt2_lo  !=  cdelt2_hi or
      bmaj_lo    !=  bmaj_hi   or bmin_lo    !=  bmin_hi   or
      bpa_lo     !=  bpa_hi    
     )
  {
    vmessage("Beam, map size, or resolution are not identical in both images!");
    vmessage("Using function difmap_restore to obtain two comparable images.");
    variable new_smajor, new_sminor, new_posang; % semimajor axis, difmap uses major axis -> *0.5
    if(qualifier_exists("beam")) {
	variable new_beam = qualifier("beam");
	new_smajor = new_beam[0];
	new_sminor = new_beam[1];
	new_posang = new_beam[2];
    }
    else {
	vmessage("Calculating enclosing beam.");
	(new_smajor, new_sminor, new_posang) = enclosing_ellipse( 3.6e6*0.5*bmaj_lo, 3.6e6*0.5*bmin_lo, bpa_lo/180.*PI,
							      3.6e6*0.5*bmaj_hi, 3.6e6*0.5*bmin_hi, bpa_hi/180.*PI);
	new_posang*=180./PI;
   }
    variable ra_pixel   = _max(naxis1_lo,naxis1_hi);
    variable dec_pixel  = _max(naxis2_lo,naxis2_hi);
    variable ra_length  = 3.6e6*_max( abs(naxis1_lo*cdelt1_lo) , abs(naxis1_hi*cdelt1_hi) );
    variable dec_length = 3.6e6*_max( abs(naxis2_lo*cdelt2_lo) , abs(naxis2_hi*cdelt2_hi) );
    variable restore_qualifiers = struct {xsize = ra_pixel , xstep = ra_length / ra_pixel ,
			      ysize = dec_pixel, ystep = dec_length/dec_pixel};
    if(qualifier_exists("ra_step"))
	restore_qualifiers.xstep = qualifier("ra_step");
    if(qualifier_exists("ra_size"))
	restore_qualifiers.xsize = qualifier("ra_size");
    if(qualifier_exists("dec_step"))
	restore_qualifiers.ystep = qualifier("dec_step");
    if(qualifier_exists("dec_size"))
	restore_qualifiers.ysize = qualifier("dec_size");

    if (qualifier_exists("overwrite")) restore_qualifiers = struct_combine (restore_qualifiers, "overwrite");
    name_lo = difmap_restore (name_lo, new_smajor, new_sminor, new_posang;; restore_qualifiers);
    name_hi = difmap_restore (name_hi, new_smajor, new_sminor, new_posang;; restore_qualifiers);
    if (name_lo == NULL or name_hi == NULL)
    {
      vmessage("ERROR %s: could not restore both images!\nYou can try to provide restored images.", _function_name());
      return NULL;
    }
  }

  %% insert filenames and other values
  variable name_corr      = qualifier("cc","cc.fits");     %name of the cross correlation coefficient
  variable N              = qualifier("iterations", 1);    %pick the best N transformations to average over
  variable req_both_bands = qualifier("req_both_bands",0); %spix just for pixels in which both bands above noise
  variable fit_noise      = qualifier("fit_noise",1);      %only does something if req_both_bands==1 (otherwise fitting in plot_spix messes up)
  variable lo_nsigma      = qualifier("lo_nsigma",3);      %only does something if fit_noise==1
  variable hi_nsigma      = qualifier("hi_nsigma",3);      %only does something if fit_noise==1  
  variable excl_core      = qualifier("excl_core",1);
  variable n_beams        = qualifier("n_beams",3);        %only does something if excl_core==1
  variable name_lo_corr   = name_lo;                       %names for files used in cc calculation. If region is excluded this changes to new file with excluded region
  variable name_hi_corr   = name_hi;                       %names for files used in cc calculation. If region is excluded this changes to new file with excluded region

  %% 2D arrays for images, correlation coefficients, etc
  variable mlo = fits_read_img(name_lo);  mlo = mlo[0,0,*,*]; % remove the 2 DIFMAP dimensions for FREQ and STOKES
  variable mhi = fits_read_img(name_hi);  mhi = mhi[0,0,*,*]; 
  variable length_x     = length(mlo[*,0]);
  variable length_y     = length(mlo[0,*]);
  variable weights      = Double_Type[length_x,length_y];  %sum of weights used per pixel for normalizing
  variable spec_map     = Double_Type[length_x,length_y];
  variable spec_map_2   = Double_Type[length_x, length_y]; %expectation value of values squared - for stdev
  variable avg_lum      = Double_Type[length_x, length_y]; %weighted average BRIGHTNESS over translations
  variable stdev        = Double_Type[length_x, length_y];
  variable avg_shift    = [0,0];
  variable shift_weight = 0;
  variable corrcoff;
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %PERFORMING CROSS CORRELATION CALCULATIONS
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  %% use beam from x image to make region file to exclude region from correlation calculation
  variable major     = fits_read_key (name_lo, "BMAJ")*(3.6e+6)/2.0; % clean beam major axis in mas (semi major!), used at end of script
  variable minor     = fits_read_key (name_lo, "BMIN")*(3.6e+6)/2.0; % clean beam minor axis ins mas (semi minor!), used at end of script
  variable posang    = fits_read_key (name_lo, "BPA");               % clean beam postion angle in degrees, used at end of script
  variable ra_steps  = fits_read_key (name_lo, "CDELT1")*(3.6e+6);   % RA in mas corresponding to one pixel, used at end of script
  variable dec_steps = fits_read_key (name_lo, "CDELT2")*(3.6e+6);   % DEC in mas corresponding to one pixel, arbitrarily chosen, used at end of script
  variable major_pix = n_beams*major / dec_steps;
  variable minor_pix = n_beams*minor / dec_steps;

  variable in_shift = qualifier("shift", {NULL,NULL});

  %% prepare files for excluding core
  if(excl_core==1 and in_shift[0]==NULL and in_shift[1]==NULL){
    %generate filenames which don't exist yet
    variable exclude_reg  = "exclude.reg";  while (stat_file(exclude_reg) != NULL)  exclude_reg="X"+exclude_reg;
    variable lo_tmp_fits  = "lo_tmp.fits";  while (stat_file(lo_tmp_fits) != NULL)  lo_tmp_fits="X"+lo_tmp_fits;
    variable hi_tmp_fits  = "hi_tmp.fits";  while (stat_file(hi_tmp_fits) != NULL)  hi_tmp_fits="X"+hi_tmp_fits;
    variable lo_lst_fits  = "lo_lst.fits";  while (stat_file(lo_lst_fits) != NULL)  lo_lst_fits="X"+lo_lst_fits;
    variable hi_lst_fits  = "hi_lst.fits";  while (stat_file(hi_lst_fits) != NULL)  hi_lst_fits="X"+hi_lst_fits;
    variable lo_excl_fits = "lo_excl.fits"; while (stat_file(lo_excl_fits) != NULL) lo_excl_fits="X"+lo_excl_fits;
    variable hi_excl_fits = "hi_excl.fits"; while (stat_file(hi_excl_fits) != NULL) hi_excl_fits="X"+hi_excl_fits;

    %generating region file -> an ellipse for each band
    variable temp = where_max(mlo);
    variable dec_beam_center  = temp[0]  /  length(mlo[0,*]) ;
    variable ra_beam_center = temp[0] mod length(mlo[0,*]) ;
    variable reg_file = fopen(exclude_reg,"w");
    fprintf(reg_file, "-ellipse(%i,%i,%f,%f,%f)\n",nint(ra_beam_center+1.0), nint(dec_beam_center+1.0), major_pix, minor_pix, posang+90); %adding 1 because ds9 starts at (1,1)
    
    temp = where_max(mhi);
    dec_beam_center  = temp[0]  /  length(mhi[0,*]) ;
    ra_beam_center = temp[0] mod length(mhi[0,*]) ;
    fprintf(reg_file, "-ellipse(%i,%i,%f,%f,%f)\n",int(ra_beam_center+1.0), int(dec_beam_center+1.0), major_pix, minor_pix, posang+90); %adding 1 because ds9 starts at (1,1)
    ()=fclose(reg_file);
    
    %generating fits files with region excluded
    ()=fits_write_image_hdu(lo_tmp_fits,"TEMP",mlo);
    ()=fits_write_image_hdu(hi_tmp_fits,"TEMP",mhi);
    ()=system("fim2lst "+lo_tmp_fits+" "+lo_lst_fits+" > /dev/null");
    ()=system("fim2lst "+hi_tmp_fits+" "+hi_lst_fits+" > /dev/null");
    ()=system("ftcopy \'"+lo_lst_fits+"[regfilter(\""+exclude_reg+"\")]\' "+lo_excl_fits);
    ()=system("ftcopy \'"+hi_lst_fits+"[regfilter(\""+exclude_reg+"\")]\' "+hi_excl_fits);
    name_lo_corr = lo_excl_fits;
    name_hi_corr = hi_excl_fits;
  }

  %% if the shift is defined, don't go about finding or generating CC map
  if(in_shift[0]!=NULL or in_shift[1]!=NULL)
  {
    if(in_shift[0]==NULL or in_shift[1]==NULL)
    {
      message("Must give both shift coordinates {x,y}");
      return NULL;
    }
    %in_shift[0] = nint(in_shift[0]/(fits_read_key(name_lo, "CDELT1")*(3.6e+6))+0.5); % converting RA  shift from mas to pixels
    %in_shift[1] = nint(in_shift[1]/(fits_read_key(name_lo, "CDELT2")*(3.6e+6))+0.5); % converting DEC shift from mas to pixels
    %RS edit: make +0.5 term optional as it should not be necessary for VLBI maps from Difmap
    in_shift[0] = nint(in_shift[0]/(fits_read_key(name_lo, "CDELT1")*(3.6e+6))+qualifier("crpix_shift",0.)); % converting RA  shift from mas to pixels
    in_shift[1] = nint(in_shift[1]/(fits_read_key(name_lo, "CDELT2")*(3.6e+6))+qualifier("crpix_shift",0.)); % converting DEC shift from mas to pixels
    if(abs(in_shift[0])>length_x or abs(in_shift[1])>length_y)
    {
      message("Shifts must be within image boundaries");
      return NULL;
    }
    corrcoff = Integer_Type[2*length_x-1,2*length_y-1];
    corrcoff[*,*] = 1;
  }
  else{
    if (stat_file(name_corr) == NULL){
      ()=system(sprintf("echo \'%s\\n%s\\n%s\\n%d\\n%d\\n%d\\n%d\\n\' | /home/grossberger/fpipe/target/bin/FPImages2CC > /dev/null 2>&1",
			name_lo_corr,name_hi_corr,name_corr,length_x,length_y,length_x,length_y));
    }
    if (stat_file(name_corr) == NULL){
      vmessage("ERROR %s: Cross correlation failed! No file %s!\n\tCheck program FPImages2CC!",_function_name(), name_corr);
      return NULL;
    }
    corrcoff = fits_read_img(name_corr);
  }

  %% clean up from excluding core
  if(excl_core==1 and in_shift[0]==NULL and in_shift[1]==NULL){
    ()=remove(exclude_reg );
    ()=remove(lo_tmp_fits );
    ()=remove(hi_tmp_fits );
    ()=remove(lo_lst_fits );
    ()=remove(hi_lst_fits );
    ()=remove(lo_excl_fits);
    ()=remove(hi_excl_fits);
  }
    
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %FITTING NOSE ON BOTH IMAGES
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  variable lothreshold = qualifier("lothreshold",NULL);  %if given image has a pixel below this value, it's set to this value
  variable hithreshold = qualifier("hithreshold",NULL);

  %% better threshold levels are only done if req_both_bands==1 so fitting in
  %% plot_spix doesn't mess up (only use brightness scaling if req_both_bands==0)
  %% 
  %% only do calculation if threshold is not predefined
  if(req_both_bands==1){
%     if(lothreshold==NULL) lothreshold = fits_read_key(name_lo, "NOISE");
%     if(hithreshold==NULL) hithreshold = fits_read_key(name_hi, "NOISE");
    if(fit_noise==1){
      variable mu , sigma;
      if(lothreshold==NULL){
	message("Fit for first image noise:");
	(mu,sigma) = fit_gauss_to_img_noise (mlo);
	lothreshold = mu + lo_nsigma*sigma;
      }
      if(hithreshold==NULL){	
	message("Fit for second image noise:");
	(mu,sigma) = fit_gauss_to_img_noise (mhi);
	hithreshold = mu + hi_nsigma*sigma;
      }
    }
  }
  
  %% if neither threshold is set by another method, set them to 1e-5
  if(lothreshold==NULL) lothreshold=1e-5;
  if(hithreshold==NULL) hithreshold=1e-5;
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %CALCULATING SPECTRAL INDEX, ETC
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  %% make log maps, remove noise
  variable log_mlo = log(_max (mlo, lothreshold));
  variable log_mhi = log(_max (mhi, hithreshold));
  variable fac = log(fits_read_key(name_lo,"CRVAL3")/1.e9)-log(fits_read_key(name_hi,"CRVAL3")/1.e9); %width of frequency range on log scale 
  
  %% initializing limits of translated arrays
  variable limhi_left,  limhi_right,  limhi_top,  limhi_bottom;
  variable limlo_left,  limlo_right,  limlo_top,  limlo_bottom;
  variable left, right, top, bottom;

  %% finding and sorting the best translations
  variable a;
  if( in_shift[0]!=NULL or in_shift[1]!=NULL ){
    a = (in_shift[1]+length_y-1)*length(corrcoff[0,*]) + in_shift[0]+length_x-1;
  }
  else if (N==1){                                               %avoid forcing shift in both directions if only one shift 
    a = where_max(corrcoff);
  }
  else{
    a = @corrcoff;
    variable b;
    if(length((corrcoff[0,*]-1)/2 mod 2 == 1)) b=0;            %which cc values are set to 0 depends on whether there are an even or odd number of pixels in a row       
    else b=1;
    a[[b:length(a[*,0])-1-b:2],            *            ]=0;   %set every element of the ccc map that we can't translate to to zero so it's
    a[          *             , [b:length(a[0,*])-1-b:2]]=0;   %not chosen as a shift when the array is sorted
    a = _reshape(a,[length(a[*,*]),1])[*,0];                   %put corr. coeff into 1d array
    a = array_sort(a)[[-N:]];                                  %gives 1d index of N top corr. coeff
  }
  
  %% translate hi image wrt lo image, calculate spectral index, record weights used per pixel
  %% ONE ITERATION PER TRANSLATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  variable mhi_temp, mlo_temp, log_mlo_temp, log_mhi_temp, spec_map_temp, spec_map_2_temp, avg_lum_temp, weights_temp, good_points;
  variable x, y, index;
  foreach index (a){
    x= index  /  length(corrcoff[0,*])-length_y+1;
    y= index mod length(corrcoff[0,*])-length_x+1;

    %% defining limits for translated images
    limhi_left   = -min([0,x]);
    limhi_right  = length_x-1-max([0,x]);
    limhi_top    = length_y-1-max([0,y]);
    limhi_bottom = -min([0,y]);
    
    limlo_left   = max([0,x]);
    limlo_right  = length_x-1+min([0,x]);
    limlo_top    = length_y-1+min([0,y]);
    limlo_bottom = max([0,y]);

    if(in_shift[0]!=NULL or N==1){
      left   = limlo_left;
      right  = limlo_right;
      top    = limlo_top;
      bottom = limlo_bottom;
    }
    else{
      left   = abs(x)/2;
      right  = length_x-1-abs(x)/2;
      top    = length_y-1-abs(y)/2;
      bottom = abs(y)/2;
    }

    if(req_both_bands==1){
      %% making temporary arrays
      mlo_temp     = mlo    [ [limlo_left:limlo_right] , [limlo_bottom:limlo_top] ];
      mhi_temp     = mhi    [ [limhi_left:limhi_right] , [limhi_bottom:limhi_top] ];
      log_mlo_temp = log_mlo[ [limlo_left:limlo_right] , [limlo_bottom:limlo_top] ];
      log_mhi_temp = log_mhi[ [limhi_left:limhi_right] , [limhi_bottom:limhi_top] ];
      
      spec_map_temp   = Double_Type [ right-left+1, top-bottom+1 ];
      spec_map_2_temp = Double_Type [ right-left+1, top-bottom+1 ];
      avg_lum_temp    = Double_Type [ right-left+1, top-bottom+1 ];
      weights_temp    = Double_Type [ right-left+1, top-bottom+1 ];
      
      %% calculating temporary values
      good_points = where( mlo_temp > lothreshold+1e-10  and  mhi_temp > hithreshold+1e-10);
      
      spec_map_temp[good_points]  = corrcoff[length_x+x-1,length_y+y-1] * (log_mlo_temp[good_points] - log_mhi_temp[good_points])/fac;
      spec_map_2_temp[good_points]= corrcoff[length_x+x-1,length_y+y-1] *((log_mlo_temp[good_points] - log_mhi_temp[good_points])/fac)^2;
      avg_lum_temp[good_points]   = corrcoff[length_x+x-1,length_y+y-1] *     (mlo_temp[good_points] + mhi_temp[good_points]);
      weights_temp[good_points]   = corrcoff[length_x+x-1,length_y+y-1];
      
      %% adding temporary values to actual arrays
      spec_map   [[left:right],[bottom:top]] += spec_map_temp;
      spec_map_2 [[left:right],[bottom:top]] += spec_map_2_temp;
      avg_lum    [[left:right],[bottom:top]] += avg_lum_temp;
      weights    [[left:right],[bottom:top]] += weights_temp;
      
      avg_shift    += corrcoff[length_x+x-1, length_y+y-1]*[x,y];
      shift_weight += corrcoff[length_x+x-1, length_y+y-1];
    }
    else{
      spec_map [[left:right],[bottom:top]] += corrcoff[length_x+x-1,length_y+y-1] / fac *
	(  log_mlo[[limlo_left:limlo_right],[limlo_bottom:limlo_top]] -
	   log_mhi[[limhi_left:limhi_right],[limhi_bottom:limhi_top]]   );
      
      spec_map_2[[left:right],[bottom:top]] += (spec_map[[left:right],[bottom:top]])^2;
      
      avg_lum[[left:right],[bottom:top]] += corrcoff[length_x+x-1,length_y+y-1] *
	( mlo[ [limlo_left:limlo_right] , [limlo_bottom:limlo_top] ] +
	  mhi[ [limhi_left:limhi_right] , [limhi_bottom:limhi_top] ]   );
      
      weights[[left:right],[bottom:top]] += corrcoff[length_x+x-1,length_y+y-1];
      
      avg_shift    += corrcoff[length_x+x-1, length_y+y-1]*[x,y];
      shift_weight += corrcoff[length_x+x-1, length_y+y-1];
    }
  }
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  %% finish calculating various values
  avg_lum    /= weights;
  avg_shift  /= shift_weight;
  spec_map   /= weights;
  spec_map_2 /= weights;
  good_points = where(abs(spec_map_2 - spec_map^2)>1e-10);   %takes care of "nan" due to errors from machine precision
  stdev[good_points] = sqrt(spec_map_2[good_points] - spec_map[good_points]^2);
  
  %% anything that was divided by zero is set to 0
  variable indx = where( weights < 1e-5 );
  spec_map  [indx] = 0;
  spec_map_2[indx] = 0;
  stdev     [indx] = 0;
  avg_lum   [indx] = 0;
  
  %% shift map_high according to the avg_shift for return
  variable j;
  _for j(0,length_y-1,1) mhi[*,j] = shift(mhi[*,j], -nint(x));
  _for j(0,length_x-1,1) mhi[j,*] = shift(mhi[j,*], -nint(y));
  
  %% ouputting for use by plot_spix
  return struct{
    spec_map       = spec_map,
    lo_map         = mlo,
    hi_map_shifted = mhi,
    stdev          = stdev,
    avg_lum        = avg_lum,
    avg_shift      = avg_shift,
    shift_weight   = shift_weight,
    avg_shift_pixel = [-nint(x), -nint(y)], %shift in pixel
    weights        = weights,
    ra_px_center   = fits_read_key (name_lo, "CRPIX1"),     % declination center pixel
    ra_steps       = ra_steps,                              % RA in mas corresponding to one pixel
    dec_px_center  = fits_read_key (name_lo, "CRPIX2"),     % declination center pixel
    dec_steps      = dec_steps,                             % DEC in mas corresponding to one pixel
    major          = major,                                 % clean beam major axis in mas (semi major!)
    minor          = minor,                                 % clean beam minor axis ins mas (semi minor!)
    posang         = posang,                                % clean beam postion angle in degrees
    source         = fits_read_key (name_lo, "OBJECT"),     % source name
    date           = strftime_MJD( "%Y-%m-%d", MJDofDateString(fits_read_key (name_lo, "DATE-OBS"))) +" / "+
                     strftime_MJD( "%Y-%m-%d", MJDofDateString(fits_read_key (name_hi, "DATE-OBS"))),
    hithreshold    = hithreshold,
    lothreshold    = lothreshold,
  };
}


%
% Code from src/data/timing/makepsd.sl
%
%%%%%%%%%%%%%%
define makepsd()
%%%%%%%%%%%%%%
%!%+
%\function{makepsd}
%\synopsis{Calculate the power density spectrum of a single energy band.}
%\usage{(psd, nipsd) = makepsd(rate, timeseg, dimseg);}
%\description
%    Input:
%      rate - input rate array
%      timeseg - real time of one segment of rate
%      dimseg  - number of bins in each segment of rate
%
%    Output:
%      psd   - unnormalized, averaged PSD
%      nipsd - averaged PSD, individual segments normalized
%\qualifiers{
%\qualifier{avgbkg}{Average background count rate for Miyamoto normalization}
%\qualifier{normtype}{"Miyamoto", "Leahy", "Schlittgen" normalization type}
%}
%!%-
{
  variable rate, timeseg, dimseg;
  switch(_NARGS)
  { case 3: (rate, timeseg, dimseg) = (); }
  { help(_function_name()); return; }

  variable numseg = int(length(rate)/dimseg);
  variable normtype = qualifier("normtype", "Miyamoto");
  variable avgbkg = qualifier("avgbkg", 0.);
  variable startbin = 0;
  variable endbin = dimseg - 1;
  variable outarray = 0.;
  variable normoutarray = 0.;
  loop(numseg)
  {
    variable temppsd = sqr(abs( dorDFT(rate[[startbin:endbin]]) ));
    variable indivavgrate = mean(rate[[startbin:endbin]]);
    normoutarray += normpsd(temppsd, normtype, indivavgrate, avgbkg, timeseg, dimseg);
    outarray += temppsd;
    startbin += dimseg;
    endbin += dimseg;
  }
  return outarray/(numseg*dimseg), normoutarray/(numseg*dimseg);
}

%
% Code from src/slang/color/maps.sl
%
define get_color_map (name)
%!%+
%\function{get_color_map}
%\synopsis{Get an intrinsic color map or construct one}
%\usage{UInt_Type[256] map = get_color_map(name);}
%\qualifiers{
%  \qualifier{reverse}{Reverse returned map}
%}
%\description
%  Retreive a colormap, either from the internal list, or
%  constructed from a color palette.
%
%  Qualifiers are passed to the constructor.
%
%\seealso{get_color_palette, png_get_colormap_names}
%!%-
{
  variable map = NULL;

#ifexists png_add_colormap
  ifnot (qualifier_exists("palette"))
  {
    try
      map = png_get_colormap(name);
    catch AnyError;
  }

  if (NULL != map)
    return map;
#endif

  try
    map = get_color_palette(name, 256;; __qualifiers);
  catch AnyError:
    throw UsageError, "Unknown palette name '$name'";

  if (length(map)<256)
    map = get_color_palette("blend", 256; colors=map);

  if (qualifier_exists("reverse"))
    map = reverse(map);

  if (_typeof(map) == String_Type)
    return rgb2hex(hex2rgb(map));
  return map;
}

%
% Code from src/slang/astronomy/mass_function.sl
%
define mass_function()
%!%+
%\function{mass_function}
%\synopsis{Calculates the mass function of a binary.}
%\usage{Double_Type   mass = mass_function(Double_Type porb, asini[, error_porb, error_asini]);
% or Double_Type[]    i = mass_function(Double_Type mass, Double_Type[] Mopt[, error_mass, error_Mopt]; i);}
%\qualifiers{
%    \qualifier{i}{calculate and return the inclination (case 2a)}
%    \qualifier{Mx}{neutron star mass in units of the solar mass
%             (default: 1.4)}
%    \qualifier{rad}{return the inclination angle in radian
%             (default: degrees)}
%    \qualifier{chatty}{set to zero to suppress output messages}
%}
%\description
%    The mass function of a binary is given by
%    (see, e.g., Hilditch, 2001)
%
%    f(M) = (M_opt sin i)^3 / (M_x + M_opt)^2
%         = (4 Pi^2 (a sin i)^3) / (G P_orb^2)
%
%    with the masses of the neutron star, M_x, and its optical
%    ompanion star, M_opt, the inclination, i, the semi-major axis,
%    a, the orbital period, P_orb, and the gravitational constant, G.
%
%    There are two ways this function can be used:
%    1)  The orbital period (in days) and the semi-major axis (in
%        lt-s) are provided, which then is used to calculate the
%        right hand side of the above equation. The returned mass is
%        given in units of the solar mass.
%    2)  The value of the mass function, mass, and the companion
%        mass(es), Mopt, are provided and the i-qualifier is set.
%        The returned value(s) is (are) the orbital inclination(s), i.
%    In case of 2) the neutron star mass is assumed to be Mx = 1.4
%    solar masses by default, but can be changed by the Mx-qualifier.
%    The thirs possible case, that the companion mass is returned,
%    is formally solving a third order polynomial, which is yet not
%    implemented here.
%
%    If uncertainties for the given parameters are provided the error
%    propagated values are returned as well for the specific case via
%      (value,error) = mass_funtion(..., error(s));
%!%-
{
  variable arg1, arg2, err1 = NULL, err2 = NULL;
  switch(_NARGS)
  { case 2: (arg1, arg2) = (); }
  { case 4: (arg1, arg2, err1, err2) = (); }
  { help(_function_name()); return; }

  % qualifiers
  variable Mx = qualifier("Mx", 1.4);
  variable chatty = qualifier("chatty", 1);
  variable errprop = err1 != NULL && err2 != NULL;

  % case
  variable icase = qualifier_exists("mopt") ? 3
                 : qualifier_exists("i") ? 2 : 1;
  variable value, err;
  switch (icase) 
  % porb and asini given, case 1
  { case 1:
    if (chatty) { message("orbital period and semi-major axis given -> returning mass function"); }
    value = 1.0735e-3 * arg2^3 / arg1^2;
    if (errprop) {
      err = 1.0735e-3 * sqrt(sqr(3*arg2^2/arg1^2 * err2) + sqr(2*arg2^3/arg1^3 * err1));
    }
  }
  % companion mass given, case 2
  { case 2:
    if (chatty) { message("companion mass given, returning inclination"); }
    value = asin(((Mx + arg2)^2 * arg1)^(1./3) / arg2);
    ifnot (qualifier_exists("rad")) { value = value*180./PI; }
    if (errprop) {
      variable c = 3*arg2*sqrt(1 - (arg1^(2./3)*(Mx+arg2)^(4./3))/arg2^2);
      err = sqrt(
        sqr((2*arg2*(arg1/(Mx+arg2))^(1./3) - 3*arg1^(1./3)*(Mx+arg2)^(2./3))
	   /(arg2*c) * err2)
      + sqr(((Mx+arg2)/arg1)^(2./3)/c * err1)
      );
    }
  }
  { case 3:
    if (chatty) { message("inclination given, returning companion mass"); }
    message("not implemented yet");
  }

  if (errprop) { return (value, err); }
  else { return value; }
}

%
% Code from src/slang/arrays/matrix2array.sl
%
define matrix2array()
%!%+
%\function{matrix2array}
%\synopsis{Transforms a matrix to an Array}
%\usage{Array_Type[] matrix2array( Any_Type[...,d,...] M, d );}
%\description
%     From the given dimension d of a matrix M an array is created,
%     i.e., the given dimension of the matrix is shifted into the
%     array which then contains matrices without this dimension.
%\example
%     variable M = _reshape( [1:2*3*4], [2,3,4] );
%     variable A0 = matrix2array(M,0);
%     variable A2 = matrix2array(M,2);
%     vmessage("A0");print(A0);
%     vmessage("A2");print(A2);
%!%-
{
  variable M, d;
  switch(_NARGS)
  { case 2: ( M, d ) = (); }
  { help(_function_name); return; }

  variable dim = array_shape(M);
  variable dlen= length(dim);
  
  variable I1 = Array_Type[d==0?0:d];
  I1[*] = [*];
  variable I2 = Array_Type[d==dlen-1?0:dlen-d-1];
  I2[*] = [*];
  
  variable A = Array_Type[dim[d]];
  _for $1 ( 0, dim[d]-1 ){
    A[$1] = M[__push_array(I1), $1, __push_array(I2) ];
  }

  return A;
}

%
% Code from src/slang/math/matrix33.sl
%

% only activate if module vector could be loaded
#ifeval __get_reference("vector")!=NULL
% 
% Implementation of 3x3 matrices and some vector routines


%%%%%%%%%%%%%%%%%%%%%%%%
define vector_astro ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{vector_astro}
%\synopsis{vector initializer that is more appropriate for astronomy}
%\usage{Vector_Type vector_astro(Double_Type x, y, z)}
%\altusage{Vector_Type vector_astro(Double_Type r, phi, theta; sph)}
%\altusage{Vector_Type vector_astro(Double_Type r, lon, lat; astro)}
%\altusage{Vector_Type vector_astro(Double_Type phi, theta; sph)}
%\altusage{Vector_Type vector_astro(Double_Type lon, lat; astro)}
%\qualifiers{
%\qualifier{astro}{consider the given coordinates to be astronomical (r, lon, lat)
%                    following the astronomical definition (i.e., lon ranges from
%                    0 to 2pi, lat from -pi/2 to pi/2; default for two arguments)}
%\qualifier{sph}{consider the given coordinates to be spherical (r, phi, theta)
%                  following the mathematical definition (i.e., theta ranges from
%                  0 to pi)}
%\qualifier{deg}{interpret angular arguments in degrees}
%}
%\description
%  The components of the vector are returned within the Vector_Type
%  and are accessible like a structure with the fields x, y, and z. 
%
%  If the astro qualifier is given, cartesian coordinates are
%  calculated as
%#v+
%    [x, y, z] =
%        r * [cos(lon)*cos(lat), sin(lon)*cos(lat), sin(lat)]
%#v-
%  If the sph qualifier is given, the cartesian coordinates are
%  calculated as
%#v+
%    [x, y, z] =
%        r * [cos(phi)*sin(theta), sin(phi)*sin(theta), cos(theta)]
%#v-
%
% If the function is called with only two arguments, then these represent the
% spherical or astronomical coordinates, and we assume that r=1.
%
% If the initializer is called with a single argument of type Vector_Type,
% the corresponding Vector_Type is returned
%
%\seealso{Vector_Type,Matrix33_Type,vector_to_spherical,dms2deg,hms2deg}
%!%-
{
    variable x,y,z;
    switch(_NARGS)
    {case 1: x=();}
    {case 2: (x,y)=();}
    {case 3: (x,y,z) = ();}
    {return help(_function_name());}

    if (Vector_Type == typeof(x)) {
	return vector(x.x,x.y,x.z);
    }
    
    if (__is_numeric (x) != 2) {
	x = typecast (x, Double_Type);
    }
    if (__is_numeric (y) != 2) {
	y = typecast (y, Double_Type);
    }

    if (_NARGS==3) {
	if (__is_numeric (z) != 2) {
	    z = typecast (z, Double_Type);
	}
    }
    
    % spherical coordinates given?
    variable sinphi, cosphi, sintheta, costheta;
    if (_NARGS==2 or qualifier_exists("sph") or qualifier_exists("astro")) {
	if (_NARGS==2) {
	    % no radius given
	    z=y; y=x; x=1.;
	}
	if (qualifier_exists("deg")) {
	    (sinphi, cosphi) = sincos(y*PI/180.);
	    (sintheta, costheta) = sincos(z*PI/180.);
	} else {
	    (sinphi, cosphi) = sincos(y);
	    (sintheta, costheta) = sincos(z);
	}
	if (qualifier_exists("sph")) {
	    return vector(x * cosphi * sintheta, x * sinphi * sintheta, x * costheta);
	};
        % phi: alpha, theta: delta
	return vector(x * cosphi * costheta, x * sinphi * costheta, x * sintheta);
    }
    
    return vector(x,y,z);
}

%%%%%%%%%%%%%%%%%%%%%%%%
define vector_to_spherical ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{vector_to_spherical}
%\synopsis{Returns the spherical coordinates corresponding to a 3D Vector}
%\usage{(r,lon,lat)=vector_to_spherical(Vector_Type v)}
%\qualifiers{
%\qualifier{astro}{return coordinates using the astronomical convention (the default)}
%\qualifier{sph}{return coordinates (r,theta,phi) using the mathematical definition
%                  (i.e., theta ranges from 0 to pi)}
%\qualifier{deg}{return angular arguments in degrees (default: radian)}
%}
%
%\description
%   This function returns the spherical coordinates corresponding to a vector.
%   See the description of function vector_astro for the relevant equations.
%   
%
%\seealso{Vector_Type,vector_astro,dms2deg,hms2deg}
%!%-
{
    variable v;
    switch(_NARGS) {
	case 1: v=();
    }
    { return help(_function_name());}

    variable r,lon,lat;
    r=hypot(v.x,v.y,v.z);
    
    if (qualifier_exists("sph")) {
	lat=acos(v.z/r);
    } else {
	lat=asin(v.z/r);
    }
    
    lon=atan2(v.y,v.x);
    if (lon<0.) {
	lon+=2.*PI;
    }
    if (qualifier_exists("deg")) {
	lon*=180./PI;
	lat*=180./PI;
    }

    return (r,lon,lat);
}

private define vector_div_scalar () {
    %
    % divide a vector by a scalar
    %
    variable v,a;
    (v,a)=();
    return vector(v.x/a,v.y/a,v.z/a);
}

%!%+
%\datatype{Matrix33_Type}
%\synopsis{3x3 matrix type}
%\description
%  3x3 Matrix type that is compatible with Vector_Type
%
%  The following operations are defined for the Matrix33_Type:
%    * M1+M2: Addition and subtraction of matrices
%    * -M1: Changing sign of a matrix
%    * M1*M2: Matrix - Matrix multiplication
%    * M1*V: Matrix - Vector multiplication (V is a Vector_Type)
%    * M1*f and f*M1: Matrix - real multiplication
%    * M1/M2: Multiply M1 with the inverse of M2. Do NOT use this...
%
%  Objects of type Matrix33_Type can be instantiated with the 
%  following functions (see there for detailed descriptions):
%    * matrix33_new: general initialization
%    * matrix33_diag: return a diagonal matrix
%    * matrix33_scalar: return a scalar matrix
%    * matrix33_null: return a null matrix
%    * matrix33_identity: return an identity matrix
%    * matrix33_rot: return a rotation matrix for the x-, y-, or z-axis
%    * matrix33_reflect: return a reflection matrix for the x-, y-, or z-axis
%
%  Other functions operating on matrices (functions marked with + are also 
%      available through accessor functions):
%    * matrix33_determinant: calculate the determinant of the matrix (+)
%    * matrix33_get_diag: return the diagonal elements of the matrix (+)
%    * matrix33_get_trace: return the sum of the diagonal elements (+)
%    * matrix33_transpose: return the transpose of the matrix (+)
%    * matrix33_adjoint: return the adjoint of the matrix
%    * matrix33_cofactors: return the matrix of cofactors
%
%  The Matrix33_Type object has the following accessors:
%    * m.determinant(): return the determinant of the matrix
%    * m.as_array(): return the elements of the matrix as a 3x3 array
%    * m.transpose(): return the transpose of the matrix (not in place!)
%    * m.diag(): return the diagonal elements as a vector
%    * m.trace(): return the trace of the matrix
%    * m.inverse(): return the inverse of the matrix (not in place!)
%
%!%-
if (0 == is_defined("Matrix33_Type")) {
    typedef struct { m,
            determinant, diag, as_array, transpose, trace, inverse } Matrix33_Type;
}

define matrix33_new();

define matrix33_diag()
%!%+
%\function{matrix33_diag}
%\synopsis{return a diagonal-matrix}
%\usage{ Matrix33_Type=matrix33_diag(m11,m22,m33);}
%\description
%  Returns a 3x3 diagonal matrix. If m11,m22,m33 are given,
%  then the three diagonal values are initialized to these
%  three values. If only m11 is given, then a scalar
%  matrix where all three elements are equal to m11 is
%  returned.
%
%\seealso{Vector_Type, Matrix33_Type}
%!%-
{
    variable m11,m22,m33;
    switch(_NARGS) {
	case 1: m11=();
	m11=double(m11);
	m22=m11;
	m33=m11;
    }
    {
	case 3: (m11,m22,m33)=();
	m11=double(m11);
	m22=double(m22);
	m33=double(m33);
    }
    {
	help(_function_name());
	return;
    }
    return matrix33_new(m11,0.,0.,0.,m22,0.,0.,0.,m33);
}

define matrix33_scalar(val)
%!%+
%\function{matrix33_scalar}
%\synopsis{return a scalar-matrix}
%\usage{ Matrix33_Type=matrix33_scala(m11);}
%\description
%  Returns a 3x3 scalar matrix (a matrix where all diagonal elements have the
%  same value and where all other elements are zero)
%
%\seealso{Vector_Type, Matrix33_Type}
%!%-
{
    return matrix33_diag(double(val));
}

define matrix33_null() 
%!%+
%\function{matrix33_null}
%\synopsis{return a null-matrix}
%\usage{ Matrix33_Type=matrix33_null();}
%\description
%  Returns a 3x3 null-matrix
%
%\seealso{Matrix33_Type}
%!%-
{ 
   return matrix33_diag(0.);
}

define matrix33_identity()
%!%+
%\function{matrix33_identity}
%\synopsis{return a identity-matrix}
%\usage{ Matrix33_Type=matrix33_identity();}
%\description
%  Returns a 3x3 identity matrix.
%
%\seealso{Matrix33_Type}
%!%-
{
    return matrix33_diag(1.);
}


private define matrix33_mat_add_mat(m1,m2) {
    %
    % add two 3x3 matrices
    %
    variable m=matrix33_new();

    m.m=m1.m+m2.m;

    return m;
}

private define matrix33_mat_sub_mat(m1,m2) {
    %
    % subtract a 3x3 matrix from another
    %
    variable m=matrix33_new();

    m.m=m1.m-m2.m;

    return m;
}

private define matrix33_chs(m1) {
    %
    % change the sign of all elements of a 3x3 matrix
    %
    variable m=matrix33_new();

    m.m=-m1.m;

    return m;
}

private define matrix33_mat_mul_mat(m1,m2) {
    %
    % multiply two 3x3 matrices with each other
    %
    variable m=matrix33_new();
    m.m=m1.m # m2.m;
    return m ;
}

private define matrix33_mat_mul_real(m1,fac) {
    %
    % multiply a 3x3 matrix with a number
    %
    variable m=matrix33_new();
    m.m=m1.m*double(fac);
    return m;
}

private define matrix33_real_mul_mat(fac,m1) {
    %
    % multiply a 3x3 matrix with a number
    %
    variable m=matrix33_new();
    m.m=m1.m*double(fac);
    return m;
}

private define matrix33_mat_mul_vector(m1,v) {
    %
    % multiply a 3x3 matrix with a Vector
    %
    return vector(
      m1.m[0,0]*v.x+m1.m[0,1]*v.y+m1.m[0,2]*v.z,
      m1.m[1,0]*v.x+m1.m[1,1]*v.y+m1.m[1,2]*v.z,
      m1.m[2,0]*v.x+m1.m[2,1]*v.y+m1.m[2,2]*v.z
    );
}

private define matrix33_real_mul_mat(fac,m1) {
    %
    % multiply a number with a 3x3 matrix
    %
    return matrix33_mat_mul_real(m1,fac);
}


private define matrix33_mat_div_real(m1,fac) {
    %
    % divide a 3x3 matrix by a number
    %
    variable m=matrix33_new();

    m.m=m1.m/double(fac);

    return m;
}

private define matrix33_real_div_mat(fac,m1) {
    %
    % divide a 3x3 matrix by a number
    %
    return fac*m1.inverse();
}

private define matrix33_mat_div_mat(m1,m2) {
    %
    % multiply a matrix with the inverse of the other matrix
    %
    return matrix33_mat_mul_mat(m1,m2.inverse());
}



private define matrix33_as_string(m1) {
    return sprintf("[[%12g %12g %12g],\n [%12g %12g %12g],\n [%12g %12g %12g]]\n",
                                  m1.m[0,0],m1.m[0,1],m1.m[0,2],
                                  m1.m[1,0],m1.m[1,1],m1.m[1,2],
                                  m1.m[2,0],m1.m[2,1],m1.m[2,2]
    );
}


define matrix33_determinant() 
%!%+
%\function{matrix33_determinant}
%\synopsis{return the determinant of a 3x3 matrix}
%\usage{ det=matrix33_determinant(m);}
%\description
%  Calculates the determinant of a 3x3 matrix. 
%
%\seealso{vector, Vector_Type, Matrix33_Type}
%!%-
{
    variable m;
    switch(_NARGS) {
	case 1: m=();
    }
    {
	help(_function_name());
	return;
    }
    
    return m.m[0,0]*(m.m[1,1]*m.m[2,2]-m.m[1,2]*m.m[2,1])
          -m.m[0,1]*(m.m[1,0]*m.m[2,2]-m.m[1,2]*m.m[2,0])
          +m.m[0,2]*(m.m[1,0]*m.m[2,1]-m.m[1,1]*m.m[2,0]);
}

define matrix33_transpose() 
%!%+
%\function{matrix33_transpose}
%\synopsis{return the transpose of a 3x3 matrix}
%\usage{ Matrix33_Type=matrix33_transpose(m);}
%\description
%  Returns the transpose of a 3x3 matrix
%
%\seealso{vector, Vector_Type, matrix33_new}
%!%-
{
    variable m;
    switch(_NARGS) {
	case 1: m=();
    }
    {
	help(_function_name());
	return;
    }

    return matrix33_new(m.m[0,0],m.m[1,0],m.m[2,0],
                        m.m[0,1],m.m[1,1],m.m[2,1],
                        m.m[0,2],m.m[1,2],m.m[2,2]);
}

define matrix33_cofactors()
{
    variable m=();
    return matrix33_new(
      +(m.m[1,1]*m.m[2,2]-m.m[2,1]*m.m[1,2]),
      -(m.m[1,0]*m.m[2,2]-m.m[2,0]*m.m[1,2]),
      +(m.m[1,0]*m.m[2,1]-m.m[2,0]*m.m[1,1]),
      -(m.m[0,1]*m.m[2,2]-m.m[2,1]*m.m[0,2]),
      +(m.m[0,0]*m.m[2,2]-m.m[2,0]*m.m[0,2]),
      -(m.m[0,0]*m.m[2,1]-m.m[2,0]*m.m[0,1]),
      +(m.m[0,1]*m.m[1,2]-m.m[1,1]*m.m[0,2]),
      -(m.m[0,0]*m.m[1,2]-m.m[1,0]*m.m[0,2]),
      +(m.m[0,0]*m.m[1,1]-m.m[1,0]*m.m[0,1]));
}

define matrix33_adjoint()
{
    variable m=();
    return matrix33_cofactors(m).transpose();
}

define matrix33_inverse()
{
    variable m=();
    return matrix33_adjoint(m)* 1./m.determinant();
}


define matrix33_as_array() 
%!%+
%\function{matrix33_as_array}
%\synopsis{return the contents of the matrix as a 3x3 matrix}
%\usage{ arr=matrix33_as_array(m);}
%\description
%  Return the contents of the matrix as a 3x3 array
%
%\seealso{vector, Vector_Type, matrix33_new}
%!%-
{
    variable m;
    switch(_NARGS) {
	case 1: m=();
    }
    {
	help(_function_name());
	return;
    }
    return m.m[*,*];
}

define matrix33_get_diag() 
%!%+
%\function{Vector_Type=matrix33_get_diag(Matrix33_Type)}
%\synopsis{return the diagonal elements of a 3x3 matrix}
%\usage{ dia=matrix33_get_diag(m);}
%\description
%  Return the diagonal elements of a Matrix33_Type as a Vector_Type.
%  Also available through the accessor function m.diag
%
%\seealso{vector,Vector_Type, matrix33_new}
%!%-
{
    variable m;
    switch(_NARGS) {
	case 1: m=();
    }
    {
	help(_function_name());
	return;
    }
    
    return vector(m.m[0,0],m.m[1,1],m.m[2,2]);
}

define matrix33_get_trace() 
%!%+
%\function{Vector_Type=matrix33_get_trace(Matrix33_Type)}
%\synopsis{return the trace of a 3x3 matrix}
%\usage{trac=matrix33_get_trace(m);}
%\description
%  Return the sum of the diagonal elements of a Matrix33_Type
%  Also available through the accessor function m.trace
%
%\seealso{vector,Vector_Type, matrix33_new}
%!%-
{
    variable m;
    switch(_NARGS) {
	case 1: m=();
    }
    {
	help(_function_name());
	return;
    }
    
    return m.m[0,0]+m.m[1,1]+m.m[2,2];
}


define matrix33_new()
%!%+
%\function{matrix33_new}
%\synopsis{instantiate a 3x3 matrix}
%\usage{ Matrix33_Type=matrix33_new();}
%\description
% Instantiates a new 3x3 matrix object.
%  The following initializers are available:
%    * No arguments: a zero-Matrix is returned
%    * One argument:
%        If the argument is of type Matrix33_Type: a copy of the argument
%           is returned
%        If the argument is a scalar: all matrix elements are initialized to
%           this scalar (use matrix33_diag to initialize a diagonal matrix!)
%        If the argument is an array with 9 elements: the matrix is initialized
%           to these elements
%    * Nine arguments: matrix elements m11,m12,m13,m21,m22,m23,m31,m32,m33,
%      i.e., the elements are in row order and the matrix is:
%
%               m11 m12 m13
%          M =  m21 m22 m23
%               m31 m32 m33
%
%\seealso{Matrix33_Type, Vector_Type}
%!%-
{
    variable m=@Matrix33_Type;
    m.m=Double_Type[3,3];
    
    variable mat;
    switch(_NARGS)
    {
	case 0:
	m.m[*,*]=0.;
    }
    {
	case 1:
	mat=();
	if (Matrix33_Type==typeof(mat)) {
	    m.m[*,*]=mat.m[*,*];
	} else {
	    if (Array_Type==typeof(mat)) {
		if ((array_shape(mat))[0]==9) {
		    m.m[*,*]=double(mat);
		} else {
		    throw UsageError,sprintf("%s: array argument must have 9 elements\n",_function_name());
		}
	    } else {
		if (Double_Type==typeof(mat)) {
		    m.m[*,*]=mat;
		} else {
		    throw UsageError,sprintf("%s: single argument must be of Matrix33_Type or an array\n",_function_name());
		}
	    }
	}
    }
    {
	case 9:
	variable m11,m12,m13,m21,m22,m23,m31,m32,m33;
	(m11,m12,m13,m21,m22,m23,m31,m32,m33)=();
	m.m[*,*]=[[double(m11),double(m12),double(m13)],
                  [double(m21),double(m22),double(m23)],
                  [double(m31),double(m32),double(m33)]];

    }

    m.determinant=&matrix33_determinant;
    m.diag=&matrix33_get_diag;
    m.as_array=&matrix33_as_array;
    m.transpose=&matrix33_transpose;
    m.trace=&matrix33_get_trace;
    m.inverse=&matrix33_inverse;
    return m;    
}


%%%%%%%%%%%%%%%%%
define matrix33_reflect()
%%%%%%%%%%%%%%%%%
%!%+
%\function{matrix33_reflect}
%\synopsis{return a 3x3 reflection matrix to change the handedness of the ith axis}
%\usage{ Matrix33_Type=matrix33_reflect(i);}
%\description
% Returns a reflection matrix, i.e., a matrix that changes the handedness
% of the ith axis (where i=1: x-axis, i=2: y-axis, and i=3: z-axis)
% when multipliying it with a vector.
%
%\seealso{vector, Vector_Type, matrix33_new}
%!%-
{
    variable ax;
    
    switch(_NARGS) {
	case 1: ax=();
    }
    {
	help(_function_name());
	return;
    }

    ax=int(ax);
 
    if (ax<1 || ax>3) {
	throw UsageError,sprintf("%s: axis must be 1,2 or 3\n",_function_name());
    }

    if (ax==1) {
	return matrix33_diag(-1.,+1.,+1.);
    }
    if (ax==2) {
	return matrix33_diag(+1.,-1.,+1.);
    }
    return matrix33_diag(+1.,+1.,-1.);
    
}
    
%%%%%%%%%%%%%%%%%
define matrix33_rot()
%%%%%%%%%%%%%%%%%
%!%+
%\function{matrix33_rot}
%\synopsis{return a standard rotation matrix about the x-, y-, or z-axis}
%\usage{ Matrix33_Type=matrix33_rot(i,angle;qualifiers);}
%\qualifiers{
% \qualifier{deg}{angle is given in deg [default: radians]}
%}
%\description
% Returns a rotation matrix to transform a column-3 vector from
% one cartesian coordinate system to another. The new coordinate
% system is given by rotating the original system in a counter
% clockwise way around the ith axis (where i=1: x-axis,
% i=2: y-axis, and i=3: z-axis)
%
% The nomenclature follows Kaplan et al, The IAU Resolutions
% on Astronomical Reference Systems, Time Scales, and Earth
% Rotation Models, USNO circular 179, 2005.
%
%\seealso{vector, Vector_Type, matrix33_new}
%!%-
{
    variable ax,ang;
    
    switch(_NARGS) {
	case 2: (ax,ang)=();
    }
    {
	help(_function_name());
	return;
    }

    if (ax<1 || ax>3) {
	throw UsageError,sprintf("%s: axis must be 1,2 or 3\n",_function_name());
    }
    
    variable si,co;
    if (qualifier_exists("deg")) {
	(si,co)=sincos(ang*PI/180.);
    } else {
	(si,co)=sincos(ang);
    }

    ax=int(ax);

    if (ax==1) {
	return matrix33_new(1.,0.,0.,0.,co,si,0.,-si,co);
    }
    if (ax==2) {
	return matrix33_new(co,0.,-si,0.,1.,0.,si,0.,co);
    }
	
    return matrix33_new(co,si,0.,-si,co,0.,0.,0.,1.);
}

% Operator overloading
#ifexists __add_unary
__add_binary("*",Vector_Type, &matrix33_mat_mul_vector, Matrix33_Type, Vector_Type);
__add_binary("*",Matrix33_Type, &matrix33_mat_mul_mat, Matrix33_Type, Matrix33_Type);
__add_binary("*",Matrix33_Type, &matrix33_mat_mul_real, Matrix33_Type, Any_Type);
__add_binary("*",Matrix33_Type, &matrix33_real_mul_mat, Any_Type,Matrix33_Type);

__add_binary("/",Matrix33_Type, &matrix33_mat_div_mat, Matrix33_Type, Matrix33_Type);
__add_binary("/",Matrix33_Type, &matrix33_mat_div_real, Matrix33_Type, Any_Type);
__add_binary("/",Matrix33_Type, &matrix33_real_div_mat, Any_Type,Matrix33_Type);

__add_binary("+",Matrix33_Type, &matrix33_mat_add_mat, Matrix33_Type, Matrix33_Type);
__add_binary("-",Matrix33_Type, &matrix33_mat_sub_mat, Matrix33_Type, Matrix33_Type);
__add_unary("-",Matrix33_Type, &matrix33_chs, Matrix33_Type);

__add_string(Matrix33_Type,&matrix33_as_string);

% addition to the vector type
__add_binary("/",Vector_Type, &vector_div_scalar, Vector_Type, Any_Type);

#endif

#endif

%
% Code from src/slang/arrays/2d/matrixmul.sl
%
define matrixmul()
%!%+
%\function{matrixmul}
%\synopsis{Matrix multiplication}
%\usage{ M[n,m] = matrixmul( A[n,j], B[j,m] );}
%\description
%    (Efficient) Calculation of the product of two matrices A and B.
%    If A is a (n x j)-Matrix and B a (j x m)-Matrix the resulting
%    Matrix has the dimensions (n x m).
%
%    If A or B is one dimensional, the neccessary second dimension
%    is assumed to be of length 1,
%    i.e., A[n]*B[m] =  A[n,1]*B[1,m] = M[n,m]
%
%    This function is identical to the ISIS intrinsic operator #
%    or the inner product and only available for backwards
%    compatibility!
%
%\seealso{transpose, inner_product}
%!%-
{
  
  variable _a, _b;
  switch(_NARGS)
  { case 2: ( _a, _b ) = (); }
  { help(_function_name); return;}

  if( length(array_shape(_a))==length(array_shape(_b))==1 ){
    return _a # transpose(_b);
  }else{
    return _a # _b;
  }  
}

%
% Code from src/slang/structs/tables/max_struct_field.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define max_struct_field()
%%%%%%%%%%%%%%%%%%%%%%%
%!+
%\function{max_struct_field}
%\synopsis{returns the maximal field value of an array of structures}
%\usage{Double_Type mx = max_struct_field(Struct_Type structs[], String_Type fieldname);}
%\description
%    \code{structs} is an array of structures containing the field called \code{fieldname}.
%    \code{mx} is computed in the following way:\n
%    \code{mx = max( array_map(Double_Type, &max, array_map(Array_Type, &get_struct_field, [structs], fieldname)) );}
%\seealso{min_struct_field, min_time, max_time}
%!%-
{
  variable structs, fieldname;
  switch(_NARGS)
  { case 2: (structs, fieldname) = (); }
  { help(_function_name()); return; }

  return max( array_map(Double_Type, &max, array_map(Array_Type, &get_struct_field, [structs], fieldname)) );
}

%
% Code from src/slang/structs/tables/lightcurves/max_time.sl
%
define max_time()
%!%+
%\function{max_time}
%\synopsis{finds the latest time in an array of structures with a time field}
%\usage{Double_Type tmax = max_time(Struct_Type structs[]);}
%\description
%    \code{structs} is an array of structures containing the field \code{time}.
%    \code{tmax} is computed in the following way:\n
%    \code{tmax = max( array_map(Double_Type, &max, array_map(Array_Type, &get_struct_field, [structs], "time")) );}
%\seealso{min_time, min_struct_field, max_struct_field}
%!%-
{
  variable structs;
  switch(_NARGS)
  { case 1: structs = (); }
  { help(_function_name()); return; }

  return max( array_map(Double_Type, &max, array_map(Array_Type, &get_struct_field, [structs], "time")) );
}

%
% Code from src/misc/simulation/mc_sig.sl
%

% only activate if module rand could be loaded
#ifeval __get_reference("rand_flat")!=NULL

%%%%%%%%%%%%%%%%%%%%%
define mc_sig()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{mc_sig}
%\synopsis{calculates the significance of a spectral component doing a Monte Carlo (MC) simulation}
%\usage{Struct_Type mc_sig(String_Type withComponent.fits, String_Type withoutComponent.fits);
% or Struct_Type mc_sig(String_Type torqueDir);}
%\qualifiers{
%    \qualifier{mcruns}{number of MC loops (default: 10)}
%    \qualifier{beforeData}{script to be called before any data is loaded (default: NULL)}
%    \qualifier{afterData}{script to be called after the data have been loaded (default: NULL)}
%    \qualifier{beforeFit}{array of scripts to be called right before a fit of faked data to
%                 the model without [0] and with [1] the component (default: NULL)}
%    \qualifier{id}{override the dataset ID(s) after fits_load_fit with the given one(s)}
%    \qualifier{ignbinning}{do not apply the same binning after faking the data}
%    \qualifier{ignnotice}{do not use the same energy ranges after faking the data}
%    \qualifier{torque}{calculate significance using given number of torque jobs, i.e., total
%                 number of runs = mcruns*torque (default: 0, i.e., don't use torque)}
%    \qualifier{walltime}{walltime of each torque job in minutes (default: 3 minutes times
%                 number of MC runs per job)}
%    \qualifier{dontSubmit}{do not submit the torque job-file (implies qualifier 'dontwaint')
%                 (default: submit)}
%    \qualifier{dontWait}{do not wait on the the torque jobs to complete (implies qualifier
%                 'dontClean') (default: wait)}
%    \qualifier{torqueDir}{temporary directory to save necessary torque files
%                 NOTE: this directory will be deleted afterwards completely!
%                 (default: ~/.isis_mc_sig/)}
%    \qualifier{dontClean}{do not delete temporary torque directory afterwards}
%    \qualifier{chatty}{a number >0 means more chatty (default: 0)}
%    \qualifier{ }{additional qualifiers are passed to fits_load_fit}
%}
%\description
%    This function calculates the significance of a spectral component
%    found in real data. During each Monte Carlo loop, spectra data
%    without the component are simulated (for each detector) and this
%    data are then fitted with a model containing the component (that
%    needs to be tested for) and separately fitted with a model without
%    it.
%    The resulting simulated differences in chi square between these
%    fits are returned and compared to the measured difference: the
%    number of simulated chi squares below the measured one corresponds
%    to the significance, that the spectral component is real (i.e. in
%    80 cases out of 100 runs the simulated chi square difference is
%    below the measured chi square difference, the significance is 80%).
%
%    If two FITS-files created with fits_save_fit are provided, the
%    first includes the model "with" the component to be tested and
%    the second one "without" it.
%
%    If only one argument is given, this has to be a temporary directory
%    with results of a previous torque run (e.g. if "dontWait" was set).
%    The function tries to collect and return the results as usual.
%
%    The returned structure is defined as follows:
%      readdchisqr  - the measured difference in chi square
%      fakedchisqr  - an array of simulated differences in chi square
%      nfalse       - the number of detected false positives
%      significance - the resulting significance as defined above
%\example
%    % FITS-files created by fitting a cutoffpl and iron line to
%    % RXTE-PCA, -HEXTE and Swift-XRT data (Rmf_OGIP_Compliance = 0 to
%    % load XRT data, see help of fits_load_fit)
%    sig = mc_sig("rxte_swift_cutoffpl_ironline.fits",
%                 "rxte_swift_cutoffpl.fits";
%                 mcruns = 1, ROC = [2,2,0],
%                 beforeData = "defineMyModels.sl",
%                 afterData = "setDataHooks.sl",
%                 chatty = 2);
%\seealso{fits_save_fit, fits_load_fit, fakeit}
%!%-
{
  variable withFits, woFits;
  switch (_NARGS)
    % try to collect results from a previous torque run
    { case 1: withFits = ();
      % check if all files are there
      ifnot (access(withFits, X_OK) == 0 && access(sprintf("%s/collect.sl", withFits), F_OK) == 0
          && access(sprintf("%s/mc_sig.sl", withFits), F_OK) == 0 && access(sprintf("%s/torque.job", withFits), F_OK) == 0) {
	vmessage("error (%s): %s does not contain the necessary files", _function_name, withFits); return;
      }
      % collect and return the result
      () = evalfile(sprintf("%s/collect.sl", withFits));
      return eval("collect;");
    }
    { case 2: (withFits, woFits) = (); }
    { help(_function_name); return; }

  variable mcruns = qualifier("mcruns", 100);

  % additional loading scripts and qualifiers
  variable beforeData = qualifier("beforeData", NULL);
  variable afterData = qualifier("afterData", NULL);
  variable beforeFit = qualifier("beforeFit", NULL);
  variable chatty = qualifier("chatty", 0); if (chatty == NULL) chatty = 1;
  variable torque = qualifier("torque", NULL);

  % some error handling
  if (access(withFits, F_OK) != 0) { vmessage("error (%s): %s not found", _function_name, withFits); return; }
  if (access(woFits, F_OK) != 0) { vmessage("error (%s): %s not found", _function_name, woFits); return; }
  if (mcruns < 1) { vmessage("error (%s): number of MC runs has to be greater zero", _function_name); return; }
  if (torque != NULL && torque < 1) { vmessage("error (%s): number of torque jobs has to be greater zero", _function_name); return; }
  if (beforeData != NULL) {
    if (typeof(beforeData) != String_Type) { vmessage("error (%s): beforeData has to be of String_Type", _function_name); return; }
    if (access(beforeData, F_OK) != 0) { vmessage("error (%s): beforeData (%s) not found", _function_name, beforeData); return; }
  }
  if (afterData != NULL) {
    if (typeof(afterData) != String_Type) { vmessage("error (%s): afterData has to be of String_Type", _function_name); return; }
    if (access(afterData, F_OK) != 0) { vmessage("error (%s): afterData (%s) not found", _function_name, afterData); return; }
  }
  if (beforeFit != NULL) {
    if (length(beforeFit) != 2) { vmessage("error (%s): beforeFit has to be an array containing exactly two file names", _function_name); return; }
    if (_typeof(beforeFit) != String_Type) { vmessage("error (%s): beforeFit has to be an array of String_Type", _function_name); return; }
    variable i; _for i (0, 1, 1) {
      if (access(beforeFit[i], F_OK) != 0) { vmessage("error (%s): beforeFit[%d] (%s) not found", _function_name, i, beforeFit[i]); return; }
    }
  }

  % get the real delta chi square
  variable statWith = fits_read_col(withFits, "chi")[0];
  variable statWO = fits_read_col(woFits, "chi")[0];
  variable realdchisqr = statWO - statWith;
  if (chatty) vmessage("measured delta chi square = %.2f", realdchisqr);
  variable dchisqr; % will contain the simulated chi squares

  %% WRITE TORQUE JOB-SCRIPT %%
  if (torque) {
    variable walltime = qualifier("walltime", 3*mcruns);
    if (walltime < 0) { vmessage("error (%s): walltime has to be greater zero", _function_name); return; }

    % create the temporary torque directory
    variable torquedir = qualifier("torqueDir", NULL);
    if (torquedir == NULL) {   
      % get current user
      variable user, pp = popen("whoami", "r");
      ()=fgets(&user, pp);
      ()=pclose(pp);
      user = strreplace(user, "\n", "");
      torquedir = sprintf("/home/%s/.isis_mc_sig", user);
    }
    ()=system(sprintf("mkdir -p %s/log", torquedir));
    
    if (chatty) message("writing torque job-script and -file");

    % get current directory
    variable pwd;
    pp = popen("pwd", "r");
    ()=fgets(&pwd, pp);
    ()=pclose(pp);
    pwd = strreplace(pwd, "\n", "");

    % write calculation script
    variable scriptstring = "";
    scriptstring += "require(\"isisscripts\");\n";
    scriptstring += "seed_random(atoi(__argv[-2]));\n";
    scriptstring += "variable torquedir = path_dirname(__argv[-3]);\n";
    scriptstring += sprintf("variable sig = mc_sig(\"%s\", \"%s\"; mcruns = %d", withFits, woFits, mcruns);
    if (beforeData != NULL) scriptstring += sprintf(", beforeData = \"%s\"", beforeData);
    if (afterData != NULL) scriptstring += sprintf(", afterData = \"%s\"", afterData);
    if (beforeFit != NULL) scriptstring += sprintf(", beforeFit = [\"%s\", \"%s\"]", beforeFit[0], beforeFit[1]);
    % handle the additional qualifiers to fits_load_fit
    variable q;
    foreach q (get_struct_field_names(__qualifiers)) {
      if (wherefirst(q == ["mcruns","beforeData","afterData","beforeFit","torque","walltime","dontSubmit","dontWait","torqueDir","dontClean","chatty"]) == NULL) {
	scriptstring += sprintf(", %s", q);
	if (q == "ROC") scriptstring += sprintf(" = [%s]", strjoin(array_map(String_Type, &sprintf, "%d", get_struct_field(__qualifiers, q)), ","));
      }
    }
    scriptstring += ");\n";
    scriptstring += "fits_write_binary_table(sprintf(\"%s/run%s.fits\", torquedir, __argv[-1]), , struct { fakedeltachisqr = sig.fakedeltachisqr });\n";
    variable fp = fopen(sprintf("%s/mc_sig.sl", torquedir), "w+");
    ()=fputs(scriptstring, fp);
    ()=fclose(fp);

    % write collection script
    if (qualifier_exists("dontWait") || qualifier_exists("dontSubmit")) {
      scriptstring = "";
      scriptstring += "variable collect = struct {\n";
      scriptstring += sprintf("  realdeltachisqr = %f,\n", realdchisqr);
      scriptstring += sprintf("  fakedeltachisqr = merge_struct_arrays(array_map(Struct_Type, &fits_read_table, glob(\"%s/*.fits\"))).fakedeltachisqr,\n", torquedir);
      scriptstring += sprintf("  nfalse = 0,\n");
      scriptstring += sprintf("  significance = 0.\n");
      scriptstring += "};\n";
      scriptstring += "variable runs = length(collect.fakedeltachisqr);\n";
      scriptstring += sprintf("if (runs != %d) {\n", torque*mcruns);
      scriptstring += sprintf("  vmessage(\"warning (%%s): there are results missing (found %%d, should be %d)\", _function_name, runs);\n", torque*mcruns);
      scriptstring += "}\n";
      scriptstring += "collect.nfalse = length(where(collect.fakedeltachisqr >= collect.realdeltachisqr));\n";
      scriptstring += "collect.significance = 1. - (collect.nfalse == 0 ? 1./runs : 1.*collect.nfalse/runs);\n";
      scriptstring += "vmessage(\"false positives = %d\", collect.nfalse);\n";
      scriptstring += "vmessage(\"-> significance %s %.3f\", collect.nfalse == 0 ? \">=\" : \"=\", collect.signif);\n";
      
      fp = fopen(sprintf("%s/collect.sl", torquedir), "w+");
      ()=fputs(scriptstring, fp);
      ()=fclose(fp);
    }
    
    % write jobfile
    variable jobstring = "";
    jobstring += "#!/bin/bash\n#\n";
    jobstring += "#PBS -S /bin/bash -V\n";
    jobstring += sprintf("#PBS -t 0-%d%%10000\n", torque-1);
    jobstring += "#PBS -l nodes=1\n";
    jobstring += "#PBS -l arch=x86_64\n";
    jobstring += sprintf("#PBS -l walltime=00:%02d:00\n", walltime);
    jobstring += "#PBS -N mc_sig\n";
    jobstring += sprintf("#PBS -o %s/log/out\n", torquedir);
    jobstring += sprintf("#PBS -e %s/log/err\n", torquedir);
    jobstring += "export HOST=`hostname`\n";
    jobstring += "export HOSTNAME=$HOST\n\n";
    jobstring += sprintf("cd %s\n", pwd);
#ifexists rand
    variable random_nrs = rand(torque);
#else
    variable random_nrs = typecast (urand(torque)*UINT_MAX, UInteger_Type);
#endif
    _for i (0, torque-1, 1)
    jobstring += sprintf("COMMAND[%d]=\"isis-script %s/mc_sig.sl %d %d\"\n", i, torquedir, random_nrs[i], i+1);
    jobstring += "/usr/bin/nice -n +15 ${COMMAND[$PBS_ARRAYID]}\n";
    fp = fopen(sprintf("%s/torque.job", torquedir), "w+");
    ()=fputs(jobstring, fp);
    ()=fclose(fp);

    ifnot (qualifier_exists("dontSubmit")) {
      % submit job-script
      if (chatty) message("submitting job-file and waiting on exit");
      ()=system(sprintf("qsub %s/torque.job", torquedir));

      ifnot (qualifier_exists("dontWait")) {
        % waint until all fits-files appeared
        variable n = 0, t = 0;
        do {
          sleep(1);
          variable resultfits = glob(sprintf("%s/*.fits", torquedir));
          _for i (1, length(resultfits)-n, 1) ()=system("echo -n .");
          n = length(resultfits); t++;
  	  if (1.*t/60 > walltime) {
  	    vmessage("error (%s): the walltime of %d min. has passed!\n", _function_name, walltime);
	    vmessage("error (%s): please check %s\n", _function_name, torquedir);
	    vmessage("error (%s): exiting without results", _function_name);
	    return 0;
	  }
        } while (n < torque);

        % collect the results
        if (chatty) message("\ncollecting results");
        dchisqr = merge_struct_arrays(array_map(Struct_Type, &fits_read_table, resultfits)).fakedchisqr;
        mcruns *= torque;

        % clean
        ifnot (qualifier_exists("dontClean")) ()=system(sprintf("rm -r %s/", torquedir));
      } else { if (chatty) vmessage("you can find all files in %s", torquedir); return 1; }
    } else { if (chatty) vmessage("you can find all files in %s", torquedir); return 1; }
  }
  
  %% SIGNIFICANCE CALCULATION %%
  else {
    % evaluate a scriptfile before any data is loaded
    if (beforeData != NULL) { if (chatty) vmessage("calling %s", beforeData); ()=evalfile(beforeData); }
  
    % load the data and best-fit model
    % (containing the component needed to be checked)
    if (chatty) vmessage("loading data and best-fit model");
    variable id = fits_load_fit(withFits;; struct_combine(struct { noeval }, __qualifiers));
    if (qualifier_exists("id")) { id = [qualifier("id")]; }
    
    % evaluate a scriptfile before any model is evaluated
    if (afterData != NULL) { if (chatty) vmessage("calling %s", afterData); ()=evalfile(afterData); }

    % allow ISIS to overwrite the loaded spectra with fake data
    if (chatty) { vmessage("fake datatset ID(s) = [%s]",
			   strjoin(array_map(String_Type, &sprintf, "%d", id), ",")); }
    array_map(Void_Type, &set_fake, id, 1);

    % remember noticing and binning
    variable info = array_map(Struct_Type, &get_data_info, id);

    % plot
    if (qualifier_exists("plot")) {
      xlog;ylog;
      title("real data and best-fit model");
      plot_data({__push_array(all_data)}; res = 1);
      ()=keyinput(; nchr = 1, silent, prompt = "press any key to continue");
      message("");
    }
    
    % preapre MC loop
    variable woParams, withParams, woFun = NULL, withFun = NULL;
    % Monte Carlo loop
    variable mc; dchisqr = Double_Type[mcruns];
    if (chatty) message("starting Monte Carlo run");
    _for mc (0, mcruns-1, 1) {
      if (chatty) vmessage("  run %d of %d", mc+1, mcruns);
      
      % load or set the model WITHOUT the component
      if (mc == 0) {
        ()=fits_load_fit(woFits; nodata);
	woFun = get_fit_fun();
	woParams = get_params();
      } else {
	fit_fun(woFun);
        set_params(woParams);
      }
      
      % fake the spectra WITHOUT the component
      fakeit;

      % apply binning and noticing
      ifnot (qualifier_exists("ignbinning")) {
	array_map(Void_Type, &group, id);
        array_map(Void_Type, &rebin_data, id, array_struct_field(info, "rebin"));
      }
      ifnot (qualifier_exists("ignnotice")) {
	array_map(Void_Type, &ignore, id);
        array_map(Void_Type, &notice_list, id, array_struct_field(info, "notice_list"));
      }
      
      % evaluate a scriptfile before fitting
      if (beforeFit != NULL) { if (chatty) vmessage("    calling %s", beforeFit[0]); ()=evalfile(beforeFit[0]); }
      % fit the fake spectra WITHOUT the compoment
      ()=fit_counts(&statWO; fit_verbose = -1);

      % plot
      if (qualifier_exists("plot")) {
        title("fake data and fit without component");
        plot_data({__push_array(all_data)}; res = 1);
        ()=keyinput(; nchr = 1, silent, prompt = "    -> press any key to continue");
        message("");
      }

      % load or set the model *with* the component
      if (mc == 0) {
        ()=fits_load_fit(withFits; nodata);
	withFun = get_fit_fun();
	withParams = get_params();
      } else {
	fit_fun(withFun);
        set_params(withParams);
      }
    
      % evaluate a scriptfile before fitting
      if (beforeFit != NULL) { if (chatty) vmessage("    calling %s", beforeFit[1]); ()=evalfile(beforeFit[1]); }
      % fit the fake spectra *with* the component
      ()=fit_counts(&statWith; fit_verbose = -1);

      % plot
      if (qualifier_exists("plot")) {
        title("fake data and fit with component");
        plot_data({__push_array(all_data)}; res = 1);
        ()=keyinput(; nchr = 1, silent, prompt = "    -> press any key to continue");
        message("");
      }
      
      dchisqr[mc] = statWO.statistic - statWith.statistic;
      if (chatty) vmessage("    simulated delta chi square = %.3f", mc+1, dchisqr[mc]);
    }
  }

  % exit and return result
  variable signif = 1.*length(where(dchisqr <= realdchisqr)) / mcruns;
  if (chatty) vmessage("significance = %.3f", signif);
  return struct { realdeltachiqsr = realdchisqr, fakedeltachisqr = dchisqr, significance = signif };
}

#endif

%
% Code from src/slang/compatibility/mean.sl
%
#ifnexists mean
define mean(x)
{
  return sum(x)/length(x);
}
#endif

%
% Code from src/slang/arrays/2d/median_2d.sl
%
define median_2d()
%!%+
%\function{median_2d}
%\synopsis{computes the median image of a list of images (2d arrays)}
%\usage{med_img = median_2d(img1, ...);}
%\description
%!%-
{
  if(_NARGS<1) { help(_function_name()); return; }

  variable args = __pop_list(_NARGS);
  variable m = @(args[0]);
}

%
% Code from src/slang/structs/tables/struct-arrays/merge_struct_arrays.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define merge_struct_arrays()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{merge_struct_arrays}
%\synopsis{creates a structure whose fields are merged from all structures' fields}
%\usage{Struct_Type merged_s = merge_struct_arrays(Struct_Type s[]);}
%\description
%    All elements of s have to be structures with the same fields.
%    The return value merged_s is another structure of this kind, and
%    merged_s.field = [s[0].field, s[1].field, ..., s[-1].field];
%    holds for every field. (If s[i].field is NULL, it is skipped
%    unless "keep_null" qualifier is set.)
%\qualifiers{
%\qualifier{remove_excess_fields}{: remove fields not present in all
%       structures.}
%\qualifier{reshape = Integer_Type dim}{: reshape the merged fields to the
%       dimensions of the original fields, using the sum of dimensiom 'dim'
%       to account for the increased array. Care has to be taken that the
%       other dimension need to have the same length in all structures. }
%\qualifier{keep_null}{: keep fields which contain NULL.}
%}
%\seealso{append_struct_arrays, get_intersection, reshape}
%!%-
{
  variable s;
  switch(_NARGS)
  { case 1: s = (); }
  { help(_function_name()); return; }

  variable resh = qualifier("reshape", -1) ;
  variable field, i;
  if(qualifier_exists("remove_excess_fields"))
  {
    variable fieldnames = get_struct_field_names(s[0]);
    _for i (1, length(s)-1, 1)
    { variable i1;
      (i1, ) = get_intersection(fieldnames, get_struct_field_names(s[i]));
      fieldnames = fieldnames[i1];
    }
    _for i (0, length(s)-1, 1)
    { variable new_s = @Struct_Type(fieldnames);
      foreach field (fieldnames)
	set_struct_field(new_s, field, get_struct_field(s[i], field));
      s[i] = new_s;
    }
  }

  variable merged_s = @(s[0]);
  variable dim;
  foreach field (get_struct_field_names(s[0]))
  {
    variable  findim = 0  ;
    variable a = get_struct_field(s[0], field);
    if(a!=NULL or qualifier_exists("keep_null")){  a = [a];
     (dim,,)= array_info(a) ;
    findim += dim[resh] ; 
    }
    _for i (1, length(s)-1, 1)
    { variable a_ = get_struct_field(s[i], field);
      if(a_!=NULL or qualifier_exists("keep_null")) { a = [a, a_];
	(dim,,)= array_info(a_) ;
	findim += dim[resh] ;
      }
    }
    if(resh>=0 && length(dim)>resh){
      dim[resh] = findim  ;
      reshape(a,dim) ;
    }
    set_struct_field(merged_s, field, a);
  }
  return merged_s;
}

%
% Code from src/slang/astronomy/midas_bary_helio_corr.sl
%
define midas_bary_helio_corr()
%!%+
%\function{midas_bary_helio_corr}
%\usage{Struct_Type midas_bary_helio_corr(Y, m, d, H, M, S,  RAh, RAm, RAs,  decdeg, decmin, decsec)}
%\description
%    Y, m, d, H, M, S  specify the date in UT.
%
%    The fields of the returned structure have the following meaning:\n
%     bary_corr_time: Barycentric correction time, in days\n
%    helio_corr_time: Heliocentric correction time, in days\n
%       bary_RV_corr: Total barycentric RV correction, in km/s\n
%      helio_RV_corr: Total heliocentric RV correction, in km/s\n
%    diurnal_RV_corr: (incl.) diurnal RV correction, in km/s
%!%-
{
  variable Y, m, d, H, M, S,  RAh, RAm, RAs,  decdeg, decmin, decsec;
  switch(_NARGS)
  { case 12: (Y, m, d, H, M, S,  RAh, RAm, RAs,  decdeg, decmin, decsec) = ();  }
  { return help(_function_name()); }
  
  variable fp = popen(sprintf("echo 'com/bary %d,%d,%d %d,%d,%f %d,%d,%f %d,%d,%f' | inmidas -p 47", 
			      Y, m, d, H, M, S,  RAh, RAm, RAs,  decdeg, decmin, decsec),
		      "r");
  variable info = struct { 
     bary_corr_time = atof(string_matches(fgetslines(fp, 1)[0], `Barycentric correction time: *\(.*\) day`, 1)[1]),
    helio_corr_time = atof(string_matches(fgetslines(fp, 2)[0], `Heliocentric correction time: *\(.*\) day`, 1)[1]),
       bary_RV_corr = atof(string_matches(fgetslines(fp, 1)[0], `Total barycentric RV correction: *\(.*\) km/s`, 1)[1]),
      helio_RV_corr = atof(string_matches(fgetslines(fp, 1)[0], `Total heliocentric RV correction: *\(.*\) km/s`, 1)[1]),
    diurnal_RV_corr = atof(string_matches(fgetslines(fp, 1)[0], `(incl. diurnal RV correction of *\(.*\) km/s)`, 1)[1]),
  };
  ()=pclose(fp);
  return info;
}

%
% Code from src/slang/arrays/min_max.sl
%
%%%%%%%%%%%%%%
define min_max()
%%%%%%%%%%%%%%
%!%+
%\function{min_max}
%\synopsis{returns the minimum and maximum value of an array}
%\usage{(mn, mx) = min_max(Double_Type a[]);
%\altusage{(mn, mx) = min_max(Double_Type a1[], a2[], ...);}
%}
%\qualifiers{
%\qualifier{pad}{[=0] additive fraction to be padded on both sides:\n
%                   \code{-dmin = dmax = (max-min) * pad}\n
%                This qualifier may be overwritten by the following ones:}
%\qualifier{dmin}{[=0] difference added to the minimum}
%\qualifier{dmax}{[=0] difference added to the maximum}
%
%\qualifier{logpad}{[=0] multiplicative fraction to be padded on both sides:\n
%                      \code{fmin^-1 = fmax = (max/min)^logpad}\n
%                   All array elements must be either positive or negative.
%                   (For negative arrays \code{-A},  \code{[min_max(-A; logpad=lp)]}
%                    is the same as  \code{array_reverse([min_max(A; logpad=lp))} .)
%                   This qualifier may be overwritten by the following ones:}
%\qualifier{fmin}{[=1] factor to multiply the minimum}
%\qualifier{fmax}{[=1] factor to multiply the maximum}
%}
%\description
%     \code{mn = min(a) * fmin + dmin;}\n
%     \code{mx = max(a) * fmax + dmax;}
%\seealso{min, max, moment}
%!%-
{
  variable a;
  switch(_NARGS)
  { case 1: a = (); }
  { case 0: help(_function_name); return; }
  { % else:
      a = ();
      loop (_NARGS-1)
      {
	variable arg = ();
	a = [arg, a];
      }
  }
  variable mn = min(a), mx = max(a);
  variable pad = (mx-mn)*qualifier("pad", 0);
  variable logpad = 1;
  if(qualifier_exists("logpad"))
  {
    if(-2 < sign(mn) + sign(mx) < 2)
      throw UsageError, "all array elements must be either positive or negative when using logpad";
    logpad = (1.*mx/mn)^qualifier("logpad", 0);
  }
  return mn * qualifier("fmin", 1/logpad) + qualifier("dmin", -pad),
         mx * qualifier("fmax",   logpad) + qualifier("dmax",  pad);
}

%
% Code from src/slang/structs/tables/min_struct_field.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define min_struct_field()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{min_struct_field}
%\synopsis{returns the minimal field value of an array of structures}
%\usage{Double_Type mn = min_struct_field(Struct_Type structs[], String_Type fieldname);}
%\description
%    \code{structs} is an array of structures containing the field called \code{fieldname}.
%    \code{mn} can be computed in the following way:\n
%    \code{mn = min( array_map(Double_Type, &min, array_map(Array_Type, &get_struct_field, [structs], fieldname)) );}
%\seealso{max_struct_field, min_time, max_time}
%!%-
{
  variable structs, fieldname;
  switch(_NARGS)
  { case 2: (structs, fieldname) = (); }
  { help(_function_name()); return; }

  return min( array_map(Double_Type, &min, array_map(Array_Type, &get_struct_field, [structs], fieldname)) );
}

%
% Code from src/slang/structs/tables/lightcurves/min_time.sl
%
define min_time()
%!%+
%\function{min_time}
%\synopsis{finds the earliest time in an array of structures with a time field}
%\usage{Double_Type tmin = min_time(Struct_Type structs[]);}
%\description
%    \code{structs} is an array of structures containing the field \code{time}.
%    \code{tmin} is computed in the following way:\n
%    \code{tmin = min( array_map(Double_Type, &min, array_map(Array_Type, &get_struct_field, [structs], "time")) );}
%\seealso{max_time, min_struct_field, max_struct_field}
%!%-
{
  variable structs;
  switch(_NARGS)
  { case 1: structs = (); }
  { help(_function_name()); return; }

  return min( array_map(Double_Type, &min, array_map(Array_Type, &get_struct_field, [structs], "time")) );
}

%
% Code from src/fitting/missing_contour_trq.sl
%
define missing_contour_trq(){
%%%%%%%%%%%%%%%%%%%%%%%%%%    
%!%+
%\function{missing_contour_trq}
%\synopsis{calculates remaining contour points from function
%    contour_trq. It is used in such cases where countour_trq does
%    finnished before all point are calculated due to torque
%    issues.} 
%\usage{missing_contour_trq(String_Type inputDir);}
%\description
%    -\code{inputDir}   directory where all torque job files are,
%                       created by function contour_trq
%
%    It may be required to run the function several times.
%    The output (number of files) needs to be checked manualy.
%\example           
%    missing_contour_trq("torqueFiles");
%          
%\seealso{contour_trq}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%%%    
   
   variable inDir=String_Type;
   variable infile2=String_Type;
   variable i,j,k,in_1,in_2,lines_1,lines_2;
   variable part_1,part_2;
   variable wallTime=qualifier("walltime","00:30:00");

   switch(_NARGS)
     { case 1 : (inDir)=();}
     { help (_function_name() );return;}
   
   variable out="torque.missing.txt";
   
   system(sprintf("ls %s/Files/*trq | xargs -n1 basename \> temp1.txt ",inDir));
   system(sprintf("ls %s/*trq | xargs -n1 basename \> temp2.txt",inDir));
   system("diff -w -y --suppress-common-lines temp1.txt temp2.txt \> temp.out");


   variable tempIn=fopen("temp.out","r");
   variable tempOut=fopen(out,"w");
   variable newLine, outNew;
   lines_1=fgetslines(tempIn);
   
   _for i(0,length(lines_1)-1,1){%
	    newLine=strchop(lines_1[i],'_',0);
	    outNew=sprintf("%s_%s_trq \n",newLine[0],newLine[1]);
	    outNew=strtrim(outNew,"\n");
	    ()=fprintf(tempOut,"isis %s/Files/%s\n",inDir,outNew);
   }

   ()=fclose(tempIn);
   ()=fclose(tempOut);
   system(sprintf("qsub_array %s  --walltime=$wallTime --arch=x86_64 --pretend",out));
   ()=fprintf(stdout,"Wrote %d jobs to %s.\n",length(lines_1),out);
   system("rm temp1.txt temp2.txt temp.out");
}

%
% Code from src/slang/date/mjd2jd.sl
%
define mjd2jd ()
%!%+
%\function{mjd2jd}
%\synopsis{Convert Modified Julian Dates to Julian Dates.}
%\usage{mjd2jd(MJD);}
%\description
%    Helper function to convert MJD to Julian Dates 
%    by adding 2400000.5.
%\seealso{jd2mjd}
%!%-
{
   variable MJD;
   switch(_NARGS)
     { case 1: MJD = (); }
     { help(_function_name()); return; }
   return MJD+2400000.5;
 };

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define MJD2JD ()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{MJD2JD}
%\synopsis{Convert Modified Julian Dates to Julian Dates.}
%\usage{MJD2JD(MJD);}
%\description
%    Helper function to convert MJD to Julian Dates 
%    by adding 2400000.5.
%\seealso{jd2mjd, JD2MJD, MJD2JD}
%!%-
{
   variable MJD;
   switch(_NARGS)
     { case 1: MJD = (); }
     { help(_function_name()); return; }
   return MJD+2400000.5;
 };

%
% Code from src/slang/system/mkdir_rec.sl
%
define mkdir_rec( )
%!%+
%\function{mkdir_rec}
%\synopsis{Create a new directory (recursively)}
%\usage{Int_Type mkdir_rec (String_Type dir [,Int_Type mode])}
%\description
%   Does the same as 'mkdir' with the exception that this
%   function creates also subdirectories.
%   See the help of 'mkdir' for a detailed description.
%\seealso{mkdir}
%!%-
{
  variable args = __pop_list(_NARGS);

  % Remember the current working directory
  variable cwd = getcwd;

  % Check if path is absolut or relative
  variable abrel = args[0][0] == '/' ? "/" : "";

  % Split the path into individual directories
  variable dirs = strtok(args[0],"/");
  dirs[0] = abrel + dirs[0];
  
  % Try to change into each (sub-)dir, in case of failure create that dir
  variable i, stat = 0;
  _for i ( 0, length(dirs)-1, 1 ){
    if( chdir( dirs[i] ) == -1 and stat == 0 ){
      args[0] = dirs[i];
      stat = mkdir( __push_list(args) );
    }
  }
  % change back to the original working directory
  () = chdir(cwd);
  
  return stat;
}

%
% Code from src/fitting/parameters/modelComponents.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define modelComponents()
%%%%%%%%%%%%%%%%%%%%%%
{
  variable i, n = get_num_pars();
  variable components = String_Type[0];
  _for i (1, n, 1)
  { variable component = string_matches(get_par_info(i).name, `^\([^\.]*\)\.`, 1)[1];
    if(all(components!=component))  components = [components, component];
  }
  return components;
}

%
% Code from src/fitting/mpi_fit_pars.sl
%

%%% import("%%slmpi");
%%% require("%%fork_socket");

#ifeval __get_reference("rcl_mpi_init")!=NULL && __get_reference("fork_slave")!=NULL 


%  ==== already define SUBS   ===== %
private define kill_all_slaves ();
private define wait_for_slave_to_exit();
private define print_log();
private define init_setup();
private define clean_messages();
private define forked_conf();
private define mpi_gather_results();
private define mpi_write_conf_file();
private define get_fit_options_from_qualifiers();

% ================================================================ %
% #1: Initialize important variables know to ALL routines  ======= %
private variable n_param, mpi_host,mpi_rank;
private variable mpi_stdout_file;
private variable MPI_FINISHED = 0;
private variable MPI_ABORT = 0;
private variable MPI_TAG_STATE = 10;
private variable MPI_TAG_PARAMS = 20;
private variable MPI_READ_MESSAGES_TIME_INTERVAL = 0.01;
private variable MPI_USE_FORK=1;
% ================================================================ %

public define mpi_new_best_fit_hook();

%  =================================== %
private define client_main(pars,n_param)
%  ===================================  %
{
   variable info = get_params(pars);
   variable param = merge_struct_arrays(info).value;
   
   variable all_info = get_params();
   variable all_param = merge_struct_arrays(all_info).value;
   variable n_all_param = length(all_info);
   
   variable state = Integer_Type;
   variable i_param = pars[mpi_rank];
   variable master_has_best_fit=0;
   variable bufid;
   variable states = Integer_Type[n_param];

   variable level,tolerance,fitmethod;
   (level,tolerance,fitmethod) = get_fit_options_from_qualifiers(;;__qualifiers);
   set_fit_method(fitmethod);

   print_log(0,sprintf("Calculating Parameter < %s >",get_par_info(i_param).name));

   % ============  DO IT: the LOOP  ==================== %
   do
   {
      master_has_best_fit=0;
      % #1 === ALL === Calculate the Confidence Interval ================
      variable lo, hi,name;
      if (MPI_USE_FORK)  (lo, hi) = forked_conf (i_param,level,tolerance); %===  ALL  ====
      else               (lo, hi) = conf (i_param,level,tolerance);


      % #2 === new best fit found?  =====================================
      if( lo == hi && (rcl_mpi_iprobe_tag(MPI_TAG_PARAMS) == 0) )
      {
	 % store the new parameters in the variable
	 all_param = merge_struct_arrays(get_params()).value;

	 % exit slaves and program if anything went wrong on the
	 % conf-calculation
	 % XXX still need to find a better solution !!! XXX
	 if (lo==NULL) param[0] = DOUBLE_MAX;

	 variable stat;()=eval_counts(&stat;fit_verbose=-1);stat = stat.statistic;
	 if (hi!=NULL) 
	   print_log(0,sprintf(" ****** New Best Fit Found [chi2=%.2f]for < %s >: <%.4f>",stat,get_par_info(i_param).name,hi));
	 
	 % only send the values if NOT MASTER
	 if (rcl_mpi_master()) master_has_best_fit=1;  %  ====  MASTER  =====
	 else
	   {                                           %  ====  CLIENTS  ====
	      rcl_mpi_return_param(all_param,n_all_param,MPI_TAG_PARAMS);
	   }
      }
      else ifnot (rcl_mpi_master())                    %  ====  CLIENTS  ====
      {
	 rcl_mpi_return_state([0],1,MPI_TAG_STATE);
      }

     if (lo!=hi) print_log(0,sprintf(" < %s > finished: <%.4f,%.4f>",get_par_info(i_param).name,lo,hi));


      % #3a: MASTER test for new parameters             % ==== MASTER ==== %
      %     (only if parameters and not status sent !!! )
      if (rcl_mpi_master() && master_has_best_fit==0)
      {
	% master assumes everybody is still working
	% (as no messages were received yet)
	states[*]=1;
	% master is finshed (and does not have the a new best fit)
	states[rcl_mpi_rank_master()]=0;

	 % wait for parameters/messages from Clients
	 forever
	 {
	   if (rcl_mpi_iprobe())
	   {
	     variable RANK = rcl_mpi_status_src();
	     variable FLAG = rcl_mpi_status_tag();
	     if (FLAG == MPI_TAG_PARAMS)
	     {
	       rcl_mpi_recv_param(all_param,n_all_param);

		% XXXX check if any error occured: has to be fixed XXXXX
		% XXX still need to find a better solution !!! XXX
		if (param[0] == DOUBLE_MAX)
		  {
		     MPI_FINISHED=1;
		     print_log(0," ******* An Error Occured: Exiting mpi_fit_pars()!");
		  }
		
		clean_messages(n_all_param);
		break;
	     }
	     else if (FLAG == MPI_TAG_STATE)
	     {
	       variable mpi_rank_client = Integer_Type;
	       rcl_mpi_recv_state(&mpi_rank_client,1);

	       states[RANK]=0;
	       print_log(0,sprintf("Parameters left < %d >",int(sum(states)));single);

	       % jump out of the loop if all process are finished
	       % (i.e, == 0)
	       if (int(sum(states))==0)
	       {
		 MPI_FINISHED=1;
		 break;
	       }
	     }
	   }
	   else  sleep(MPI_READ_MESSAGES_TIME_INTERVAL);
	 }
      }  % ===================================================================== %


     % #4: SEND the PARAMS and STATE to the clients
     % === BCAST PARAMS  =============
      if (rcl_mpi_master()) rcl_mpi_dist_param(all_param,n_all_param,MPI_TAG_PARAMS);
      else  rcl_mpi_recv_param(all_param,n_all_param);
      rcl_mpi_barrier();

     % === BCAST STATE  =============
      variable bcast_int = [MPI_FINISHED];
      rcl_mpi_bcast_int(bcast_int,1);
      MPI_FINISHED = bcast_int[0];

      
      % XXXX check if any error occured: has to be fixed XXXXX
      % XXX still need to find a better solution !!! XXX
      if (param[0] == DOUBLE_MAX) MPI_ABORT=1;

      
      % #5a: === JUMP out of the loop if FINISHED
      if (MPI_FINISHED) break;

      % #5b: === reset everything and load new parameters for the NEXT ROUND
      % set the NEW PARAMETERS AND START ALL OVER
      variable i;
      _for i(0,n_all_param-1,1)
      {
	 all_info[i].value = all_param[i];
      }
      % ... and load them
      set_params(all_info);

      % =========  get rid of useless messages     % ==== MASTER ==== %
      if (rcl_mpi_master())
      {
	 % ====== Hook for saving Data %
	 mpi_new_best_fit_hook();
	 clean_messages(n_all_param);
      }


   } while (MPI_FINISHED != 1);
  % =================================================== %

   if (MPI_ABORT) return NULL,NULL;

   variable b_lo = Double_Type[n_param];
   variable b_hi = Double_Type[n_param];
   b_lo[*] = 0;
   b_hi[*] = 0;

   lo = [lo]; hi = [hi];
   rcl_mpi_gather_conf(lo,hi,b_lo,b_hi,n_param);
   ifnot (rcl_mpi_master()) {b_lo=NULL;b_hi=NULL;}

   rcl_mpi_barrier();
   return b_lo,b_hi;
}






%  ===================================  %
define mpi_fit_pars()
%  ===================================  %
%!%+
%\function{mpi_fit_pars}
%\synopsis{computes confidence intervals using MPI}
%\usage{Struct_Type results = mpi_fit_pars([Integer_Type pars[]]);}
%\description
%    The function \code{mpi_fit_pars} is designed to provide a similar
%    interface then \code{pvm_fit_pars}, written by M. Hanke for the
%    isis-scripts. Please pay special attention to the notes marked IMPORTANT
%    below.
%
%    \code{pars} is an array of parameters, for which the confidence levels are to be fitted.
%    If \code{pars} is not specified, all free parameters of the current model are used.
%    The best fit which is eventually found is always saved in
%    <\code{dir}>/<\code{basefilename}>\code{_best.par}; the confidence limits are
%    written in plain-text to <\code{dir}>/<\code{basefilename}>\code{_conf.txt} and as a
%    FITS table to <\code{dir}>/<\code{basefilename}>\code{_conf.fits}.
%
%    The verbosity of \code{mpi_fit_pars} is controlled by the intrinsic variable \code{Fit_Verbose}.
%
%    The return value \code{results = struct { index, name, value, min, max, conf_min, conf_max, buf_below, buf_above, tex }}
%    is a table with the following information for each parameter:\n
%    \code{min} and \code{max} are the minimum/maximum values allowed.
%    \code{conf_min} and \code{conf_max} are the confidence limits.
%    \code{buf_below} (\code{buf_above}) is the fraction of the allowed range \code{[min:max]}
%    which seperates the lower (upper) confidence limit from \code{min} (\code{max}).
%    If one of these buffers is 0, your confidence interval has bounced.
%
%    Perhaps more usefully, this information is written to the files
%    <\code{dir}>/<\code{basefilename}>\code{_conf.txt} and
%    <\code{dir}>/<\code{basefilename}>\code{_conf.fits}.  In case of any
%    error, the return value is \code{NULL}. If run via slurm/torque, these
%    files will only be written by the host node once all threads have
%    completed - i.e., there's no danger of the results being written by
%    several machines at the same time.
%
%    NOTE: The function "mpi_new_best_fit_hook()" can be defined by the
%    user. This function will be called whenever a new best fit is found. This
%    can be used to, e.g, store the newly found best fit (although the usual
%    caveats about writing files to disk while running via MPI still apply...).
%
%    IMPORTANT: Rember that the whole script containing the call
%    "mpi_fit_pars()" is typically run on N > 1 computers. Therefore only the
%    *minimum* amount of code necessary to run mpi_fit_pars() should be
%    included in the script.
%
%    IMPORTANT: This function relies on MPI. This means that it only
%    works when started externally via, e.g., \code{mpiexec isis-script
%    my_script.sl} or via a slurm jobfile. Slurm or mpiexec must be able to
%    start as many processes as you have parameters for which you want to
%    calculate confidence limits (e.g., if you want error bars on four
%    parameters, you need four tasks - no more, no less).
%
%    IMPORTANT: This code is still in <beta> test. It might happen
%    that the function ends unexpectedly.
%\qualifiers{
%\qualifier{level}{confidence level to be computed.
%             As for conf, 0 means 68%, 1 means 90% [default], and 2 means 99% confidence level.}
%\qualifier{tolerance}{the tolerance for chi^2 improvements without interrupting the search
%                 for the confidence intervals, see \code{help("conf");}. The default is \code{1e-3}.}
%\qualifier{fitmethod}{fit-method to be used, see \code{help("set_fit_method");}
%                 Default is the currently used fit-method returned by \code{get_fit_method()}.}
%\qualifier{dir}{[="."] specifies the directory in which the logfiles shall be stored.
%                  It may be a relative path to the current working directory.}
%\qualifier{basefilename}{[=startdate]}
%\qualifier{verbose}{Files with the stdout and the inital parameters are kept after program
%                     is finished. }
%\qualifier{forked}{[=0] Use a separate (forked) process to
%                  calculate the confidence levels. It speeds up the calculation, but
%                  might cause troubles for models which write files to disk, as this
%                  process is killed  immediatelly when asked to restart
%                  calculation and not stopped smoothly. This option can also
%                  cause your job to be killed if run via slurm, as it requests
%                  additional threads beyond those allocated by slurm.  }
%\qualifier{do_not_finalize}{if set, "mpi_finalize()" is not called
%                  within the routine. In this case the code following the
%                  mpi_fit_pars() call will be executed on *all* nodes. Only the
%                  master process will return the confidence structure (see above);
%                  all other processes will return NULL. This can be used to start
%                  more than one mpi_fit_pars() in one script. Only use it for very
%                  quick evaluations and if you *know that you're doing*. At the end
%                  of such an script, it is important to call "mpi_finalize()" manually.}
%\example
%     % EXAMPLE 1
%     % Simple script for confidence level calcualtion with mpi_fit_pars()
%     variable id = load_data("my_data.pha");   % load the data
%     xnotice_en(id,0.5,16);                    % initialize data
%     load_par("my_best_fit.par");
%
%     variable result = mpi_fit_pars(pars);     % do calculation for pars[]
%     fits_save_fit("result.fits",result);      % save all information
%                                               % to disk
%                                               % Note that mpi_fit_pars() alone will save information to disk,
%                                               % by default in files named <date>_best.par, <date>_conf.txt,
%                                               % and <date>_conf.fits
%
%     % EXAMPLE 2 - Uses slurm to schedule jobs on the Remeis cluster.
%     % This requires two files - mpi_fit.sl, containing the ISIS code to be
%     % executed, and mpi_fit.slurm, containing the job specification for slurm.
%     % This would be run by calling "sbatch mpi_fit.slurm" at the command line.
%     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     % ISIS script (mpi_fit.sl)
%     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     variable id = load_data("my_data.pha");   % load the data
%     xnotice_en(id,0.5,16);                    % initialize data
%     load_par("my_best_fit.par");
%     variable pars = [1,2];                    % obviously you will pick whatever parameters you need here
%     () = mpi_fit_pars(pars;                   % Do the confidence limit calculations. Results will be saved to
%       basefilename="my_fit");                 % "my_fit_conf.fits","my_fit_conf.txt", and "my_fit_best.par".
%                                               % Note that unlike the above example, we do not manually save
%     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%            % the results to a file - mpi_fit_pars does this on its own.
%     % Jobfile (mpi_fit.slurm)
%     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     #!/bin/bash
%     #SBATCH --job-name my_fit                 # name of the job to be run by slurm
%     #SBATCH --time 00:10:00                   # You might need more (or less) time.
%     #SBATCH --ntasks 2                        # we want erors on two parameters (see above), so we need 2 cores
%     #SBATCH --mem-per-cpu 1G                  # 1 gig of memory per core (tweak to your liking)
%     srun /usr/bin/nice -n +19 isis mpi_fit.sl # run the script with srun (and nice +19 because you don't 
%                                               # want to bog down someone else's machine)
%                                               
%}
%\seealso{fit_pars; conf, set_fit_method, pvm_fit_pars}
%!%-
{
   variable status = 1;
   variable SetupFile=NULL, pars = NULL;
   MPI_USE_FORK = qualifier("forked",0);

   switch(_NARGS)
   {case 0: pars = NULL;
%      if(length(glob("init.sl"))>0) { SetupFile = "init.sl"; }
%      if(length(glob("loaddata.sl"))>0) { SetupFile = "loaddata.sl"; }
%      if(SetupFile==NULL) { help(_function_name()); return; }
%      message(_function_name()+": using SetupFile " + SetupFile);
   }
   { case 1:  pars = (); }
%   { case 2: (SetupFile, pars) = (); }
   { help(_function_name()); return; }

   % log file:
   variable dir = qualifier("dir", ".");
   variable startdate = strftime("%Y-%m-%d_%H:%M:%S", localtime(_time()));
   variable basefilename = qualifier("basefilename", startdate);
   variable logbasefilename = dir + "/" + basefilename;
   variable Best_filename = logbasefilename+"_best.par";
   variable logbasefilename_stdout = logbasefilename+"_stdout.log";
   mpi_stdout_file = fopen(logbasefilename_stdout, "w");

   % are there enough MPI processes available?
   %  ===  2 === initialize MPI  =========================================================
   if ( rcl_mpi_init() < 0) { print_log(0,"MPI could not be initialzied. Exiting ..."); status = 0; return NULL;}

   % initialize variables depending on the single hosts
   pars = init_setup(pars);

   % check the number of parameters
   if (n_param < 2)
   { status=0;
      if (rcl_mpi_master()) print_log(0,sprintf("Only %d free parameter(s); no caculation with N < 2 is possible.",n_param);single);
   }

   if (rcl_mpi_numtasks() != n_param )
   { status=0;
      if (rcl_mpi_master()) print_log(0,sprintf("MPI Tasks (%d) not equal to # parameters (%d).",rcl_mpi_numtasks(), n_param);single);
   }
   if ( status != 1 )
   {
      if (rcl_mpi_master()) print_log(0,"MPI can not be set up. Exiting mpi_fit_pars ...";single);
      rcl_mpi_finalize();
      return NULL;
   }


  %  ===  3  ===
  % start the single processes, if everything is set up correctly
  if (rcl_mpi_rank() == 0) print_log(0,"Starting "+string(rcl_mpi_numtasks())+ " MPI tasks ... ";single);

  %  ===== MAIN ===========  %
  variable conf_lo,conf_hi;
   (conf_lo,conf_hi) = client_main(pars,n_param;;__qualifiers);

   % if something went wrong; quit immediatelly
   if (MPI_ABORT)
   {
      rcl_mpi_finalize();
      return NULL;
   }
   
   % QUIT all other processes except the MASTER
   ifnot (qualifier_exists("do_not_finalize"))
   {
      rcl_mpi_finalize();
      ifnot (rcl_mpi_master()) { exit(1);}
   }


   %  ===  5  ===  Build Resulting Structure
   variable results = NULL;
   if (rcl_mpi_master())
   {
      save_par(Best_filename);
      results = mpi_gather_results(pars,conf_lo,conf_hi);

      variable stat;()=eval_counts(&stat;fit_verbose=-1);
      fits_write_binary_table(logbasefilename+"_conf.fits", "pvm_fit_pars-results", results, struct {
	 chi2 = stat.statistic,
	 num_bins = stat.num_bins,
	 n_var_pars = stat.num_variable_params,
	 dof = stat.num_bins-stat.num_variable_params,
	 chi2red = stat.statistic/(stat.num_bins-stat.num_variable_params)
      } );


      mpi_stdout_file = fopen(logbasefilename+"_conf.txt", "w");
      mpi_write_conf_file(results,stat,logbasefilename);
      ()=fclose(mpi_stdout_file);

      ifnot(qualifier_exists("verbose"))
      {
	 ()=remove(logbasefilename+"_stdout.log");
	 ()=remove(logbasefilename+"_initial.par");
      }

   }
   return results;
}



% ================================================== %
% ==============   SUBS   ========================== %
% ================================================== %

% =============   FORKED SLAVES  =================== %
private define mpi_slave_fun(s,lo,hi,i,level,tolerance)
{
   variable e;
   try(e)
   {
      (lo,hi) = conf(i,level,tolerance);
   }
   catch AnyError:
   {
      print_log(0,e.message);
      lo = NULL; hi = NULL;
   } 

   send_objs(s,lo,hi,get_params);

   return 0;
}

private define fork_probe_status(s)
{
   variable status = select(s.sock,NULL,NULL,0);
   if (status != NULL && status.nready) return 1;
   return 0;
}

%  ===================================  %
private define forked_conf(i_param,level,tolerance)
%  ===================================  %
{
   variable lo=NULL,hi=NULL;
   variable param = get_params();

   () = new_slave_list;
   variable s = fork_slave(&mpi_slave_fun,lo,hi,i_param,level,tolerance);

   forever
   {
      if ( rcl_mpi_iprobe_tag(MPI_TAG_PARAMS))
      {
	 break;
      }
      else if (fork_probe_status(s))
      {
	 variable obs = recv_objs(s);
	 lo = obs[0]; hi = obs[1];
	 param=obs[2];	 set_params(param);
	 break;
      }
      else
      {
	 sleep(MPI_READ_MESSAGES_TIME_INTERVAL);
      }
   }
   % end slave here !
   kill_all_slaves(s);

%   if (lo!= NULL) () = print_log(0,sprintf("Returned value from forked_conf: <%.3f , %.3f>",lo,hi));

   return lo,hi;
}

private define wait_for_slave_to_exit(s)
{
   variable w;
   do
   {
      w = waitpid (s.pid, WNOHANG|WUNTRACED);
   } while (w.pid == 0);

   return;
}

private define kill_all_slaves (slaves)
{
   variable s;
   foreach s (slaves)
   {
      if (kill (s.pid, 0) == 0)
      {
	 if (kill (s.pid, SIGTERM) == 0)
	   wait_for_slave_to_exit(s);
      }
      s.status = SLAVE_EXITED;
      s.sock = NULL;
      s.fp = NULL;
   }
   return;
}


%  =================================== %


private define clean_messages(n_param)
{
   variable dump = Double_Type[n_param];
   variable i_dump = Double_Type;
   while(rcl_mpi_iprobe()>0)
   {
      if (rcl_mpi_iprobe_tag(MPI_TAG_PARAMS)) { rcl_mpi_recv_param(dump,n_param); }
      else if (rcl_mpi_iprobe_tag(MPI_TAG_STATE)) {rcl_mpi_recv_state(&i_dump,1); }
   }
}


public define isis_fit_improved_hook()
{
   if (MPI_USE_FORK) return 0;  % always return normally when using the FORK method

   if ( rcl_mpi_iprobe_tag(MPI_TAG_PARAMS))
   {
      return 1;
   }
   return 0;
}


private define init_setup(pars)
{
   mpi_rank = rcl_mpi_rank();
   mpi_host = uname.nodename;

   % if not specified differently: use all free parameters
   if (pars == NULL) pars = freeParameters();


   variable p = get_params(pars);
   n_param = (p!= NULL)?length(p):0;

   return pars;
}


private define print_log(thres, s)
{
   variable msg;
   variable t = strftime("%Y-%m-%d_%H:%M:%S", localtime(_time()));
   if (qualifier_exists("single")) msg= "%s\n"$;
   else msg= "@$mpi_host-$mpi_rank($t): %s\n"$;

   if(Fit_Verbose >= thres)  () = printf(msg,s);
  ()=fprintf(mpi_stdout_file, msg, s);
  ()=fflush(mpi_stdout_file);
}




%  GATHER the INFORMATION

private define mpi_gather_results(pars,conf_min,conf_max)
{
  variable info = get_params(pars);

  % collect results
  variable results = struct {
             index     = Integer_Type[0],
             name      = String_Type[0],
             value     = Double_Type[0],
             min       = Double_Type[0],
             max       = Double_Type[0],
             conf_min  = Double_Type[0],
             conf_max  = Double_Type[0],
             buf_below = Double_Type[0],
             buf_above = Double_Type[0],
             tex       = String_Type[0],
           };
  variable i,d,cmin,cmax;
  _for i(0,n_param-1,1)
  {
    d = info[i]; cmin = conf_min[i]; cmax = conf_max[i];

    variable buf_below = (cmin-d.min)/(d.max-d.min);
    variable buf_above = (d.max-cmax)/(d.max-d.min);
    results.index     = [results.index,     d.index];
    results.name      = [results.name,      d.name];
    results.value     = [results.value,     d.value];
    results.conf_min  = [results.conf_min,  cmin];
    results.conf_max  = [results.conf_max,  cmax];
    results.min       = [results.min,       d.min];
    results.max       = [results.max,       d.max];
    results.buf_below = [results.buf_below, buf_below];
    results.buf_above = [results.buf_above, buf_above];
    results.tex       = [results.tex,       TeX_value_pm_error(d.value, cmin, cmax)];
  }

  return results;
}

private define mpi_write_conf_file(results,stat,logbasefilename)
{
   % write confidence file
   print_log(-1, get_fit_fun();single);
   print_log(-1, sprintf("(%10s <=) %10s < %38s < %10s (<= %10s)", "min.all.", "conf.min.", "parameter value", "conf.max.", "max.all.");single);
   for($1=0; $1<length(results.name); $1++)
     print_log(-1, sprintf("(%10g <=) %10g < %25s = %10g < %10g (<= %10g)   [%4.1f%% (conf)%5.1f%%]   %s",
			  results.min[$1], results.conf_min[$1], results.name[$1], results.value[$1], results.conf_max[$1], results.max[$1],
			  100*results.buf_below[$1], 100*results.buf_above[$1], results.tex[$1]);single);
   print_log(-1, sprintf("\nstatistic/dof = %f/%d = %f",
			stat.statistic, stat.num_bins-stat.num_variable_params, stat.statistic/(stat.num_bins-stat.num_variable_params));single);
   return;
}




private define  get_fit_options_from_qualifiers()
{
   variable level = qualifier("level", 1);
   if(all(level!=[0:2]))
   {
      message("warning ("+_function_name()+"): level has to be 0 (68%), 1 (90%), or 2 (99%).\nUsing default level = 1.");
      level = 1;
   }
   if(level==NULL) { level = 1; }
   variable tolerance = qualifier("tolerance", 1e-3);
   variable fitmethod = qualifier("fitmethod", get_fit_method());
   return level,tolerance,fitmethod;
}

#endif

%
% Code from src/fitting/mpi_fit_pars_trqscript_gen.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define mpi_fit_pars_trqscript_gen(){
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{mpi_fit_pars_trqscript_gen}
%\synopsis{Creates torque scripts for derivation of confidence intervals}
%\usage{mpi_fit_pars_trqscript_gen(String_Type fitfile, String_Type resultfile, String_Type scriptfile)}
%\description
%    This function generates 1) an executable S-Lang file
%    that is loading a previous fit saved with \code{fits_save_fit}, 
%    calling \code{mpi_fit_pars} with the current number of free parameters
%    and saving the result via \code{fits_save_fit} and \code{save_pars},
%    2) writing the .cmd and 3) the .cmd.job file
%    that can be submitted to Torque via 'qsub *.cmd.job'.
%
%    If only two arguments are given
%    (the resultfile and the scriptfile), the function expects
%    that you want to calculate the uncertainties out of isis
%    In that case the function will itself save the fit to a
%    file in the results-directory. When mpi_fit_pars is finished,
%    you have to delete it by your own.
%
%    ATTENTION: make sure that the min/max parameter borders
%    are set wide enough in your previous fit but still with
%    a reasonable width to allow mpi to find the best fit and
%    ncertainties in a reasonable amount of time.
%
%    NOTE: All given qualifiers are passed to the function
%    fits_load_fit. 
%\qualifiers{
%\qualifier{walltime}{[="00:30:00"]: String_Type, wallime for the torque}
%\qualifier{ROC}{Rmf OGIP Compliance (see fits_load_fit}
%\qualifier{submit}{1-submit to torque, 0-only create job files}%
%}
%\example
%    isis>
%
%    isis> mpi_fit_pars_trqscript_gen("input.fits","result.fits","trq_scripts/mpi_script.sl";
%                                      walltime="01:00:00")
%    OR 
%    isis> mpi_fit_pars_trqscript_gen("result.fits","trq_scripts/mpi_script.sl";
%                                      walltime="01:00:00")
%
%    THEN
%    
%    qsub trq_scripts/mpi_script.cmd.job
%
%\seealso{fits_save_fit, mpi_fit_pars, fits_list_fit_pars}
%!%-

  variable fitfile,resultfile,scriptfile,parfile,year,month,day,hr,mn,sc;
  switch(_NARGS)
  { case 2:  (resultfile,scriptfile) = (); }    
  { case 3:  (fitfile,resultfile,scriptfile) = (); }
  { help(_function_name()); return; }
  
  variable walltime = qualifier("walltime", "00:30:00");
  variable submit = qualifier("submit",0);
  variable logdir = path_dirname(resultfile);

  % create directories if non existing
  system(sprintf("mkdir -p %s %s",path_dirname(resultfile), path_dirname(scriptfile)));
  
  sc = string(localtime(_time).tm_sec);
  mn = string(localtime(_time).tm_min);
  hr = string(localtime(_time).tm_hour);
  day = string(localtime(_time).tm_mday);
  month = string(localtime(_time).tm_mon+1);
  year = string(localtime(_time).tm_year+1900);
  
  %   /////////////////////////
  %  //// SCRIPT GENERATOR ///
  % /////////////////////////
  
  % I    err scripts for all obsid's
  % II   cmd script for trq combining all err calc scripts
  % III  job script out of cmd script for torque ('qsub xxx.cmd.job')
  
  variable i,j,f,cmd,fitfun,freePars,job,roc_arr,strct,roc_strng;

  % initialize files
  f = fopen(scriptfile, "w");
  cmd = fopen(strreplace(scriptfile,".sl",".cmd"), "w");

  switch(_NARGS)
  { case 2: fits_save_fit(path_dirname(resultfile)+"/fit_tmp.fits"); fitfile=path_dirname(resultfile)+"/fit_tmp.fits";}
  
  % load attempt file
  ()=fits_load_fit(fitfile;; __qualifiers() );
  variable roc = __qualifiers().ROC;
  fitfun = get_fit_fun;
  
  %    if (is_substr(fitfun,"pexrav") > 0){
  %      % to file
  %      fputs(sprintf("fit_fun(\"%s\");\n", fitfun), f);
  %      fputs("__set_hard_limits (\"pexrav(1).rel_refl\",-1,1e6);\n\n", f);
  %      % execute here for getting freePars
  %      fit_fun(sprintf("%s", fitfun));
  %      __set_hard_limits("pexrav(1).rel_refl",-1,1e6);
  %      }
  
  % write to mpi_fit_pars exec file ----------------------------------------------------------------------
  roc_strng="[";
  for (i=0;i<length(roc);i++){
    if (i!=length(roc)-1) roc_strng+=string(roc[i])+",";
    if (i==length(roc)-1) roc_strng+=string(roc[i]);
  }
  roc_strng+="]";
  ()=fputs(sprintf("fits_load_fit(\"%s\";ROC=%s);\n\n", fitfile,roc_strng), f);
  freePars = length(freeParameters);
  
  % ERROR CALCULATION VIA MPI
  ()=fputs("variable pars = freeParameters;\n", f);
  ()=fputs(sprintf("system(\"mkdir -p %s/logdir/\");\n", logdir), f);
  ()=fputs(sprintf("variable result = mpi_fit_pars(pars; verbose, dir=\"%s/logdir/\");\n", logdir), f);
  ()=fputs(sprintf("fits_save_fit(\"%s\",result);\n",resultfile), f);
  parfile = strreplace(resultfile,".fits",".par");
  ()=fputs(sprintf("save_par(\"%s\");", parfile), f);
  ()=fclose(f);
  % ------------------------------------------------------------------------------------------------------
  
  % write to command file --------------------------------------------------------------------------------
  ()=fputs(sprintf("mpiexec isis-script %s\n",scriptfile), cmd);
  % ------------------------------------------------------------------------------------------------------

  
  % write job script -------------------------------------------------------------------------------------
  job = fopen(strreplace(scriptfile,".sl",".cmd.job"), "w");
  ()=fputs("#!/bin/bash\n", job);
  ()=fputs("#\n", job);
  ()=fputs("#PBS -S /bin/bash -V\n", job);
  ()=fputs("#PBS -t 0-0%10000\n", job);
  ()=fputs(sprintf("#PBS -l nodes=%d\n", freePars), job);
  ()=fputs(sprintf("#PBS -l walltime=%s\n", walltime), job);
  ()=fputs(sprintf("#PBS -N %s\n", strreplace(scriptfile,".sl",".cmd")), job);
  ()=fputs(sprintf("#PBS -o /tmp/pbs/beuchert/err_calc_%s-%s-%s_%s:%s:%s.cmd.out\n",year,month,day,hr,mn,sc), job);
  ()=fputs(sprintf("#PBS -e /tmp/pbs/beuchert/err_calc_%s-%s-%s_%s:%s:%s.cmd.err\n",year,month,day,hr,mn,sc), job);
  ()=fputs("export HOST=hostname\n\n", job);
  ()=fputs("export HOSTNAME=$HOST\n\n\n", job);
  ()=fputs(sprintf("COMMAND[0]=\"mpiexec isis-script %s\"\n", scriptfile), job);
  ()=fputs("/usr/bin/nice -n +15 ${COMMAND[$PBS_ARRAYID]}\n", job);
  ()=fclose(job);
  % ------------------------------------------------------------------------------------------------------
  
  delete_data(1);
  delete_arf(1);
  delete_rmf(1);
  
  ()=fclose(cmd);

  if (submit == 1) system(sprintf("qsub %s",strreplace(scriptfile,".sl",".cmd.job")));

}

%
% Code from src/misc/mpi_master_only.sl
%
#ifexists rcl_mpi_init
%%%%%%%%%%%%%%%%%%%%%%%%%
define mpi_master_only ()
%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{mpi_master_only}
%\synopsis{Use function with on one host only in an MPI job}
%\usage{mpi_master_only (Reference_Type function, arg1, arg2, ...)}
%\qualifiers{
%\qualifier{verbose}{show name of host which acts as MPI master}
%\qualifier{veryverbose}{show name of host which acts as MPI master, PID and parent PID (overwrites verbose)}
%}
%\description
%	Sometimes, for example by saving files, it is required that a function is only called once
%	in an MPI job. mpi_master_only does this for a function with an arbitrary number of arguments.
%	The function has to be passed by reference, the arguments simply as arguments.
%\example
%	mpi_master_only(&sprintf, "Pi is approximately %f and Eulers number is %f", PI, E);
%	output:
%		Pi is approximately 3.141593 and Eulers number is 2.718281828459045
%	But this only once
%!%-
{
	variable args = __pop_list(_NARGS);
	variable todo = args[0];
	list_delete (args, 0);
	rcl_mpi_init;
	if (rcl_mpi_master)
	{
		variable hostname = strchop(fgetslines(popen("hostname", "r"))[0], '\n', 0)[0];
		if (qualifier_exists("veryverbose"))
			vmessage("*** mpi_master_only on: "+hostname+" PID "+sprintf("%d", getpid)+" parent PID "+sprintf("%d", getppid));
		else if (qualifier_exists("verbose"))
			vmessage("*** mpi_master_only on: "+hostname);
		(@todo)(__push_list(args));
	}
	rcl_mpi_finalize;
}
#endif

%
% Code from src/misc/simulation/mpi_mc_sig.sl
%
% -*- mode:slang; mode:fold -*-
#ifexists rcl_mpi_init
private define verbexec (script){
  %{{{
  variable chatty = qualifier("chatty", 0); if (chatty == NULL) chatty = 1;
  if (script != NULL) {
    if (chatty){ vmessage("calling %s", script);}
    ()=evalfile(script); 
  }
}
%}}}
private define compare_stat(beforeData, afterData, beforeFit, info, id, withFun, withParams, woFun, woParams){
  %{{{
  % apply binning and noticing
  variable statWO, statWith;
  variable chatty = qualifier("chatty", 0); if (chatty == NULL) chatty = 1;
  ifnot (qualifier_exists("ignbinning")) {
    array_map(Void_Type, &group, id);
    array_map(Void_Type, &rebin_data, id, array_struct_field(info, "rebin"));
  }
  ifnot (qualifier_exists("ignnotice")) {
    array_map(Void_Type, &ignore, id);
    array_map(Void_Type, &notice_list, id, array_struct_field(info, "notice_list"));
  }
  % load or set the model *without* the component
  fit_fun(woFun);
  set_params(woParams);
  % evaluate a scriptfile before fitting
  if (beforeFit != NULL) { if (chatty) vmessage("    calling %s", beforeFit[0]); ()=evalfile(beforeFit[0]); }
  % fit the fake spectra WITHOUT the compoment
  ()=fit_counts(&statWO; fit_verbose = -1);
  % load or set the model *with* the component
  fit_fun(withFun);
  set_params(withParams);
  % evaluate a scriptfile before fitting
  if (beforeFit != NULL) { if (chatty) vmessage("    calling %s", beforeFit[1]); ()=evalfile(beforeFit[1]); }
  % fit the fake spectra *with* the component
  ()=fit_counts(&statWith; fit_verbose = -1);

  return  (statWO.statistic - statWith.statistic);
}
%}}}
private define receive_spectrum(id){
  %{{{
  % Loop over IDs and receive all spectra as array of counts
  variable ii;
  variable origspec, nbins;
  foreach ii (id){
    group(ii); %Unset grouping before overwriting data. Data will be rebinned before fitting anyway
    origspec = get_data_counts(ii);
    nbins = length(get_data_info(ii).rebin);
    variable value = Double_Type[nbins];
    variable error = Double_Type[nbins];
    () = rcl_mpi_recv_double(value, nbins);
    () = rcl_mpi_recv_double(error, nbins);
    put_data_counts(ii, origspec.bin_lo, origspec.bin_hi, value, error); 
  }
}
%}}}
private define send_spectrum(rank, id){
  %{{{
  % Loop over IDs and send all spectra as array of counts
  variable ii;
  variable spec;
  foreach ii (id){
    spec = get_data_counts(ii);
    () = rcl_mpi_send_double(spec.value, length(spec.value), rank,1);
    () = rcl_mpi_send_double(spec.err, length(spec.err), rank,1);
  }
}
%}}}
private define distribute_runs(mcruns, nclients){
  %{{{
  % Distribute the mc runs on the different clients.
  % This is important if the number of runs is not the
  % same for every client which happens if the
  % the number of runs is not an integer multiple
  % of the number of nodes.
  variable dist = Integer_Type[nclients];
  dist[*] = int (mcruns /nclients);
  variable rest = mcruns mod nclients;
  dist[[0:rest-1]] +=1;
  return dist;
}
%}}}
private define save_results(realdchisqr, fakedchisqr, nfalse, significance, file){
  %{{{
  variable fp = fits_open_file(file, "c");
  fits_write_binary_table(fp, "REALDELTACHISQUARE", struct{realdchisqr = realdchisqr});
  fits_write_binary_table(fp, "FAKEDELTACHISQUARE", struct{fakedchisqr = fakedchisqr});
  fits_write_binary_table(fp, "FALSEPOSITIVES", struct{nfalse = nfalse});
  fits_write_binary_table(fp, "SIGNIFICANCE", struct{significance =significance});
  fits_close_file(fp);
}
%}}}
%
%%%%%%%%%%%%%%%%%%%%%
define mpi_mc_sig()
%%%%%%%%%%%%%%%%%%%%%
% HELP
%{{{
%!%+
%\function{mpi_mc_sig}
%\synopsis{calculates the significance of a spectral component doing a Monte Carlo (MC) simulation}
%\usage{Struct_Type mpi_mc_sig(String_Type withComponent.fits, String_Type withoutComponent.fits [, String_Type results.fits]);}
%\qualifiers{
%    \qualifier{mcruns}{number of MC loops (default: 10)}
%    \qualifier{beforeData}{script to be called before any data is loaded (default: NULL)}
%    \qualifier{afterData}{script to be called after the data have been loaded (default: NULL)}
%    \qualifier{beforeFit}{array of scripts to be called right before a fit of faked data to
%                 the model without [0] and with [1] the component (default: NULL)}
%    \qualifier{id}{override the dataset ID(s) after fits_load_fit with the given one(s)}
%    \qualifier{ignbinning}{do not apply the same binning after faking the data}
%    \qualifier{ignnotice}{do not use the same energy ranges after faking the data}
%    \qualifier{chatty}{a number >0 means more chatty (default: 0)}
%    \qualifier{seed}{seed for random number generator}
%    \qualifier{ }{additional qualifiers are passed to fits_load_fit}
%}
%\description
%    This function calculates the significance of a spectral component
%    found in real data. During each Monte Carlo loop, spectra data
%    without the component are simulated (for each detector) and these
%    data are then fitted with a model containing the component (that
%    needs to be tested for) and separately fitted with a model without
%    it.
%    The resulting simulated differences in chi-square between these
%    fits are returned and compared to the measured difference: the
%    number of simulated chi squares below the measured one corresponds
%    to the significance, that the spectral component is real (i.e. in
%    80 cases out of 100 runs the simulated chi square difference is
%    below the measured chi square difference, the significance is 80%).
%
%    Two FITS-files created with fits_save_fit must be provided, the
%    first includes the model "with" the component to be tested and
%    the second one "without" it. An optional third filename may be given
%    to store the results to. Otherwise, a default file will be created.
%
%    This function is supposed to be used for parallel computation with
%    MPI. The mcruns qualifier specifies the total number of MC runs which
%    is split accordingly among the available nodes.
%
%    The function will create a FITS file containing the following extensions:
%      realdeltachisquare - the measured difference in chi square
%      fakedeltachisquare - an array of simulated differences in chi square
%      falsepositives     - the number of detected false positives
%      significance       - the resulting significance as defined above
%\example
%    % FITS-files created by fitting a cutoffpl and iron line to
%    % RXTE-PCA, -HEXTE and Swift-XRT data (Rmf_OGIP_Compliance = 0 to
%    % load XRT data, see help of fits_load_fit)
%
%     mpi_mc_sig("rxte_swift_cutoffpl_ironline.fits",
%                "rxte_swift_cutoffpl.fits",
%                "significance.fits";
%                 mcruns = 100, ROC = [2,2,0],
%                 beforeData = "defineMyModels.sl",
%                 afterData = "setDataHooks.sl",
%                 chatty = 2);
%    % A typical call from the command line could look like this
%
%    mpiexec -n 4 isis <script>
%
%    to use 4 cores for the MC run. However, it is highly
%    recommended to use a job scheduling system (e.g., SLURM)
%    for these kind of simulations.
%                              
%\seealso{fits_save_fit, fits_load_fit, fakeit, mc_sig}
%!%-
%}}}
{
  % Arguments and qualifiers
  %{{{
  variable withFits, woFits, resfile = "significance.fits";
  switch (_NARGS)
  { case 2: (withFits, woFits) = ();}
  { case 3: (withFits, woFits, resfile) = (); }
  { help(_function_name); return; }
  
  variable mcruns = qualifier("mcruns", 100);
  
  % additional loading scripts and qualifiers
  variable beforeData = qualifier("beforeData", NULL);
  variable afterData = qualifier("afterData", NULL);
  variable beforeFit = qualifier("beforeFit", NULL);
  variable chatty = qualifier("chatty", 0); if (chatty == NULL) chatty = 1;
  variable seed = qualifier("seed", NULL); % seed for random number generator
  %}}}
  % some error handling
  %{{{
  if (access(withFits, F_OK) != 0) { vmessage("error (%s): %s not found", _function_name, withFits); return; }
  if (access(woFits, F_OK) != 0) { vmessage("error (%s): %s not found", _function_name, woFits); return; }
  if (mcruns < 1) { vmessage("error (%s): number of MC runs has to be greater zero", _function_name); return; }
  if (beforeData != NULL) {
    if (typeof(beforeData) != String_Type) { vmessage("error (%s): beforeData has to be of String_Type", _function_name); return; }
    if (access(beforeData, F_OK) != 0) { vmessage("error (%s): beforeData (%s) not found", _function_name, beforeData); return; }
  }
  if (afterData != NULL) {
    if (typeof(afterData) != String_Type) { vmessage("error (%s): afterData has to be of String_Type", _function_name); return; }
    if (access(afterData, F_OK) != 0) { vmessage("error (%s): afterData (%s) not found", _function_name, afterData); return; }
  }
  if (beforeFit != NULL) {
    if (length(beforeFit) != 2) { vmessage("error (%s): beforeFit has to be an array containing exactly two file names", _function_name); return; }
    if (_typeof(beforeFit) != String_Type) { vmessage("error (%s): beforeFit has to be an array of String_Type", _function_name); return; }
    variable i; _for i (0, 1, 1) {
      if (access(beforeFit[i], F_OK) != 0) { vmessage("error (%s): beforeFit[%d] (%s) not found", _function_name, i, beforeFit[i]); return; }
    }
  }
  %}}}  

  variable PROC_RANK=rcl_mpi_init();
  variable PROC_TOTAL=rcl_mpi_numtasks();

  variable REQUEST=-255; % Just some random integer signal
  variable COLLECT=128;
  variable MASTER=0;
  variable SIG=Integer_Type[2];
  variable FAKE_FINISHED=0;
  variable statWith, statWO;
  variable realdchisqr = fits_read_col(woFits, "chi")[0] - fits_read_col(withFits, "chi")[0]; 
  
  if (chatty){vmessage("measured delta chi square = %.2f", realdchisqr);}
  if (seed != NULL){seed_random(seed);}

  %%% SIGNIFICANCE CALCULATION %%%
  
  % evaluate a scriptfile before any data is loaded
  verbexec(beforeData;; __qualifiers);
  
  % load the data and best-fit model (containing the component needed to be checked)
  variable id = fits_load_fit(withFits;; struct_combine(struct { noeval }, __qualifiers));
  if (qualifier_exists("id")) { id = [qualifier("id")]; }
  
  % evaluate a scriptfile before any model is evaluated
  verbexec(afterData;; __qualifiers);
  
  % remember noticing and binning
  variable info = array_map(Struct_Type, &get_data_info, id);

  % preapre MC loop
  variable woParams, withParams, woFun, withFun, mc, simdchisqr = Double_Type[0];
  variable fake_counter = 0; %keep track of how many spectra have been faked

  % Save fit-functions and parameters
  ()=fits_load_fit(withFits; nodata);
  withFun = get_fit_fun();
  withParams = get_params();
  
  ()=fits_load_fit(woFits; nodata);
  woFun = get_fit_fun();
  woParams = get_params();

  %%%%% The master process (rank = 0) is only responsible for the faking and for coordinating  %%%%%

  if (PROC_RANK == MASTER){
    
    % Tell the clients how many runs they'll have to do
    variable dist = distribute_runs(mcruns, PROC_TOTAL - 1);
    variable client;
    _for client (1, PROC_TOTAL-1){
      () = rcl_mpi_send_int([dist[client-1]],1,client,1);
    }
    
    % allow ISIS to overwrite the loaded spectra with fake data
    if (chatty) { vmessage("fake datatset ID(s) = [%s]",
			   strjoin(array_map(String_Type, &sprintf, "%d", id), ",")); 
    }
    array_map(Void_Type, &set_fake, id, 1);
    array_map(Void_Type, &group, id); % Unset grouping for faking
    
    % Wait for fake request from clients
     while(FAKE_FINISHED !=1 ) { 
      () = rcl_mpi_recv_int(SIG,2);
      if (SIG[0] == REQUEST){
	% fake the spectra WITHOUT the component
	fit_fun(woFun);
	set_params(woParams);
	fakeit();
	% Send faked spectra to client
	send_spectrum(SIG[1], id);
	fake_counter++;
	if (fake_counter == mcruns){FAKE_FINISHED = 1;}
      }
    }
  }
  %%%%% The clients are only responsible for the fitting %%%%%
  
  if (PROC_RANK != MASTER){
    % First receive the number of runs from master
    variable mccl;
    () = rcl_mpi_recv_int(&mccl, 1);
    variable results = Double_Type[mccl];
    _for mc (0, mccl-1, 1){
      % Send request to master
      () = rcl_mpi_send_int([REQUEST, PROC_RANK],2,MASTER,1);
      receive_spectrum(id); % Listen to reply from master
      results[mc] = compare_stat(beforeData, afterData, beforeFit, info, id, withFun, withParams, woFun, woParams;; __qualifiers);
    }
    if (chatty){vmessage("Process %d finished fitting", PROC_RANK);}
  }
  rcl_mpi_barrier(); % Wait for all clients to finish

  % Finally, the master process collects all results
  % Since the number of runs may differ from one client to the other,
  % we do this only upon request from the master who knows how
  % many results to expect
  
  if (PROC_RANK != MASTER){
    () = rcl_mpi_recv_int(SIG,2); % wait for request from master
    if (SIG[0] == COLLECT){
      () = rcl_mpi_send_double(results, mccl, MASTER, 1); % Send results from MC runs to master
    }
  }

  % Collect results
  if (PROC_RANK == MASTER){
    _for client (1, PROC_TOTAL-1){
      variable tmpres = Double_Type[dist[client-1]];
      () = rcl_mpi_send_int([COLLECT, PROC_RANK],2,client,1);
      () = rcl_mpi_recv_double(tmpres, dist[client-1]);
      simdchisqr = [simdchisqr,tmpres];
    }
    % return result
    variable nfalse = length(where(simdchisqr >= realdchisqr));
    variable signif = 1. - (nfalse == 0 ? 1./mcruns : 1.*nfalse/mcruns);

    if (chatty){
      vmessage("false positives = %d", nfalse);
      vmessage("-> significance %s %.3f", nfalse == 0 ? ">=" : "=", signif);
    }
    save_results(realdchisqr, simdchisqr, nfalse, signif, resfile);
  }
  rcl_mpi_finalize();
}
#endif
%
% Code from src/Mike/msg.sl
%
public define msg(str_array)
{
   () = printf("\n");
   foreach(str_array)
   {
      variable str = ();
      () = printf(" %s\n", str);
   }
   () = printf("\n");
   return;
}

%
% Code from src/fitting/fit-functions/multibknpower.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define multibknpower_fit(bin_lo, bin_hi, pars)
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{multibknpower (fit-function)}
%\synopsis{implements a multiply-broken powerlaw}
%\description
%    Only break energies >0 and the corresponding photon indices are considered.
%    \code{PhoIndx0} applies for energies below the first break.
%    Each \code{PhoIndex}i applies for energies above \code{BreakE}i, but below the next break.
%    As for \code{bknpower}, the normalization is the flux of the first power law at 1 keV
%    (not necessarily the broken power law model)  in ph/s/cm^2/keV.
%!%-
{
  variable norm    = pars[ 0    ];
  variable Gamma   = pars[ 1    ];
  variable Ebreaks = pars[[2::2]];
  variable Gammas  = pars[[3::2]];

  variable ind = where(Ebreaks>0);
  ind = ind[array_sort(Ebreaks[ind])];
  Ebreaks = [Ebreaks[ind], _Inf];
  Gammas = Gammas[ind];
  variable n = length(ind);

  variable Elo = _A(bin_hi);
  variable Ehi = _A(bin_lo);
  variable value = 0*bin_lo;
  variable Ebinwidth = Ehi - Elo;

  variable E0 = 0;
  variable i;
  _for i (0, n, 1)
  {
    ind = where(Ehi>E0 and Elo<Ebreaks[i]);
    if(length(ind))
    {
      variable ELO = _max(E0,         Elo[ind]);
      variable EHI = _min(Ebreaks[i], Ehi[ind]);
      variable one_minus_Gamma = 1 - Gamma;
      value[ind] += norm * (one_minus_Gamma==0 ? log(EHI/ELO) : (EHI^one_minus_Gamma - ELO^one_minus_Gamma)/one_minus_Gamma);
    }
    if(i<n)
    {
      E0 = Ebreaks[i];
      norm *= E0^(Gammas[i]-Gamma);
      Gamma = Gammas[i];
    }
  }
  return reverse(value);
}

add_slang_function("multibknpower", ["norm", "PhoIndx0",
  "BreakE1", "PhoIndx1",
  "BreakE2", "PhoIndx2",
  "BreakE3", "PhoIndx3",
  "BreakE4", "PhoIndx4",
  "BreakE5", "PhoIndx5",
  "BreakE6", "PhoIndx6",
  "BreakE7", "PhoIndx7",
  "BreakE8", "PhoIndx8",
  "BreakE9", "PhoIndx9",
  "BreakE10", "PhoIndx10",
], 0);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define multibknpower_default(i)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  if(i==0)    return (1, 0,  0, 1e10); % norm
  if(i==1)    return (2, 0, -2, 9);    % PhoIndx0
  if(i mod 2) return (2, 1, -2, 9);    % PhoIndx[i]
              return (0, 1,  0, 1e6);  % BreakE[i]
}

set_param_default_hook("multibknpower", &multibknpower_default);

%
% Code from src/slang/strings/multiple_string.sl
%
define multiple_string(n, s)
{
  variable res = "";
  for(; n>0; n--) { res += s; }
  return res;
}

%
% Code from src/plot/multiplot_data_m_lam_counts.sl
%
define multiplot_data_m_lam_counts()
{
  variable ids, xmin=NULL, xmax=NULL;
  switch(_NARGS)
  { case 1: ids = (); }
  { case 3: (ids, xmin, xmax) = (); }
  {%else:
      message("usage: plot_m_lambda(ids[, m_lam_min, m_lam_max]); % where ids = [id1, id2, ...];");
      return;
  }

  multiplot(0*[ids]+1);
  xrange(xmin, xmax);
  xlabel(`order m  \x  wavelength \gl  [\A]`);
  plot_bin_integral;
  variable m;
  for(m=1; m<=length(ids); m++)
  {
    ylabel("counts/bin (m="+string(m)+")");
    variable data = get_data_counts([ids][m-1]);
    hplot(m*data.bin_lo, m*data.bin_hi, data.value);
  }
}

%
% Code from src/plot/multiplot_flux_counts_res.sl
%
define multiplot_flux_counts_res()
%!%+
%\function{multiplot_flux_counts_res}
%\synopsis{creates a plot of model flux, data counts and residuals in 3 panels}
%\usage{multiplot_flux_counts_res(data_ind[, rel_size]);}
%\description
%    If data_ind is an array of data-indices, the data will be rebinned
%    to the grid of the first index.
%!%-
{
  variable data_ind, rel_size = [2, 3, 1];
  switch(_NARGS)
  { case 1:  data_ind = (); }
  { case 2: (data_ind, rel_size) = (); }
  { help(_function_name()); return; }

  variable f = get_model_flux(1);
  variable d = get_data_counts([data_ind][0]);
  variable m = get_model_counts([data_ind][0]);
  variable i;
  _for i (1, length(data_ind)-1, 1)
  { variable tmp_d = get_data_counts(data_ind[i]);
    variable tmp_m = get_model_counts(data_ind[i]);
    d.value += rebin(d.bin_lo, d.bin_hi,  tmp_d.bin_lo, tmp_d.bin_hi, tmp_d.value);
    d.err = sqrt(d.err^2 + rebin(d.bin_lo, d.bin_hi,  tmp_d.bin_lo, tmp_d.bin_hi, tmp_d.err)^2);
    m.value += rebin(m.bin_lo, m.bin_hi,  tmp_m.bin_lo, tmp_m.bin_hi, tmp_m.value);
  }
  variable res = (d.value-m.value)/d.err;

  variable plot_opt = get_plot_options();
  multiplot(rel_size);
  ylabel(`Flux [ph/s/cm\u2\d/\A]`);
  plot_bin_density;
  hplot(f);

  plot_bin_integral;
  ylabel("Counts [/bin]");
  hplot(d);
  ohplot(m);

  xlabel(`Wavelength [\A]`);
  ylabel(`\gx`);
  ylin;
  hplot(d.bin_lo, d.bin_hi, res);
  oplot([plot_opt.xmin, plot_opt.xmax], [0, 0]);

  set_plot_options(plot_opt);
}

%
% Code from src/fitting/fit-functions/narrowline.sl
%
define narrowline_fit (bin_lo,bin_hi,par)
%!%+
%\function{narrowline (fit-function)}
%\synopsis{multiplicative model; implements unresolved absorption/emission line}
%\description
%   simple multiplicative line-function to efficiently model
%   unresolved absorption and emission lines. Allows to direclty fit
%   the equivalent width of the line.
%
%   Caution: the function does not check whether the line is
%   unresolved; user needs to take care of that before using the
%   function. In particular, for absorption lines the absolute value
%   of the equivalent width should be smaller than the resolution of
%   the data.
%
%   The fit parameters are
%
%      center [A]   line position
%      eqw    [A]   equivalent width of the line
%!%-
{

  %array that we will return in the end; 1 everywhere
  variable dd = Double_Type[length(bin_lo)];
  dd[*] = 1.0;
  
  variable idx;

  %only do stuff if your line is within the energy range of your data
  %at all
  if (par[0] > bin_lo[0] and par[0] < bin_hi[-1]) {
    
    %if emission line, do ...
    if (par[1] >= 0) {
      
      %%last entry where the lower limit of the bin is below the
      %%centroid energy of the lone
      idx = where(bin_lo < par[0])[-1];
      
      %determine the multiplicative factors representing the line
      dd[idx] = 1+par[1]/(bin_hi[idx]-bin_lo[idx]);
      
    }
    %if absorption line, do ...
    else{
      
      %maximal and minimal wavelength
      variable lmin = par[0]-abs(par[1])/2.;
      variable lmax = par[0]+abs(par[1])/2.;
      
      %use where and not fancy max/min logic which would be shorter code
      %but much less efficient
      
      %note that the inner where always have to be one that stars with
      %zero and counts up without interruption; otherwise we are getting
      %ugly stuff
      variable idx1 = where (bin_hi[where( bin_lo < lmin)] > lmin);
      variable idx2 = where (bin_hi[where( bin_lo < lmax)] > lmax);
      variable idx3 = where (bin_lo[where( bin_hi < lmax)] > lmin);
      
      %% next step: union function of the three idx
      idx = union(idx1,idx2,idx3);
      
      %determine the multiplicative factors representing the line
      variable i;
      _for i (0,length(idx)-1,1) {
	dd[idx[i]]=1-_max((_min(bin_hi[idx[i]],lmax) -
			   _max(bin_lo[idx[i]],lmin)),0)/
	  (bin_hi[idx[i]]-bin_lo[idx[i]]);
      }
      
    }
    
  }
  
  return dd;
}

%%%%%%%%%%%%%%%%%%%%%%%%
add_slang_function("narrowline", ["center [A]", "eqw [A]"]);
%%%%%%%%%%%%%%%%%%%%%%%%

define narrowline_default(i)
{
  switch(i)
    { case 0: return ( 9.15, 0, 0.1, 100 ); }
    { case 1: return ( -0.004, 0, -1, 1); }
}
set_param_default_hook("narrowline", &narrowline_default);

%
% Code from src/slang/arrays/normalized_in_0_1.sl
%
define normalized_in_0_1()
%!%+
%\function{normalized_in_0_1}
%\synopsis{computes a normalized array with min=0, max=1}
%\usage{Double_Type a_norm[] = normalized_in_0_1(Double_Type a[]);}
%\description
%    \code{a_norm = (a - min(a)) / (max(a)-min(a));}
%!%-
{
  variable a;
  switch(_NARGS)
  { case 1: a = (); }
  { help(_function_name()); return; }

  variable mn = min(a), mx = max(a);
  return (a-mn)/(mx-mn);
}

%
% Code from src/data/timing/normpsd.sl
%
define normpsd()
%!%+
%\function{normpsd}
%\usage{psd_norm = normpsd(psd, normtype, avrate, avbackrate, timeseg, dimseg);}
%\description
%    \code{psd} is the unnormalized power spectrum.
%    \code{avrate} is the average countrate of corresponding lightcurve.
%    \code{avbackrate} is the average background countrate of corresponding lightcurve.
%    \code{timeseg} is the real time of one corresponding lightcurve segment.
%    \code{dimseg} is the number of bins in one lightcurve segment.
%
%    String_Type \code{normtype} can be:\n
%    - "Miaymoto" (default) [Miyamoto et al. (1991), ApJ 383, 784]\n
%      \code{psd *= 2 * timeseg / (dimseg * [avrate-avbackrate]^2)}\n
%    - "Leahy" [Leahy et al. (1983), ApJ 266, 160]\n
%      \code{psd *= 2 * timeseg / (dimseg^2 * avrate)}\n
%    - "Schlittgen" [Schlittgen, H.J., Streitberg, B. (1995), Zeitreihenanalyse, R. Oldenbourg]
%      \code{psd /= dimseg}\n
%!%-
{
  variable psd, normtype, avrate, avbackrate=0, timeseg, dimseg;
  switch(_NARGS)
  { case 5: (psd, normtype, avrate, timeseg, dimseg) = ();
            vmessage("warning (%s): avbackrate=%g is assumed", _function_name(), avbackrate);
  }
  { case 6: (psd, normtype, avrate, avbackrate, timeseg, dimseg) = (); }
  { help(_function_name()); return; }

  switch(normtype)
  { case "Schlittgen": return psd                /  dimseg; }
  { case "Leahy":      return psd * 2. * timeseg / (dimseg * dimseg * avrate); }
  { % case "Miyamoto":  % and default
                       return psd * 2. * timeseg / (dimseg * (avrate-avbackrate))^2; }

}

%
% Code from src/data/notice_human2isis.sl
%
% ================= %
define notice_human2isis(sl)
% ================= %
%!%+
%\function{notice_human2isis}
%\synopsis{create isis notice_list from str = notice_isis2human}
%\usage{Array_Type notice_list = notice_human2isis(String_Type str);}
%\description
%   This routine converts a notice_list string created with
%   notice_isis2human back to the ISIS conform binning array
%   (see e.g. get_data_info(idx).notice_list). It can be used to
%   restore the otcing of data directly, by using it as argument
%   in the function notice_list(...).
%   
%\seealso{fits_save_fit, get_data_info, notice_isis2human}
%!%-
{
  
  variable nl = Integer_Type[0];
  
  variable tk = strtok(sl,",");

  variable i,j;
  _for i(0, length(tk)-1, 1)
  {
    variable ttk = strtok(tk[i],"-");
    if (length(ttk) == 1)
    {
      nl = [nl, integer(ttk)];
    }
    else
    {
      variable tlo = integer(ttk[0]);
      variable thi = integer(ttk[1]);
      
      _for j(tlo,thi,1)
      {
	nl = [nl, j];	
      }

    }
  }
   
  return nl;
}

%
% Code from src/data/notice_isis2human.sl
%
define notice_isis2human(nl)
%!%+
%\function{notice_isis2human}
%\synopsis{converts isis notice_list in a human readable string}
%\usage{String_Type str = notice_isis2human(Array_Type notice_list);}
%\description
%   This routine converts the notice_list as used by ISIS (see e.g. 
%   get_data_info(idx).notice_list) to a string which is easy to read.
%   
%   The routine fits_save_fit() uses this string to save the noticed
%   bins of the data. It can be converted back by using 
%   notice_human2isis().
%\seealso{fits_save_fit, get_data_info, notice_human2isis}
%!%-
{
  variable sl = "";
  variable comma = "";
  
  variable n = length(nl);
  variable i=0;
  
  % initialize
  variable rem = nl[0];
  
  _for i(1,n-1,1)
  {
    if (nl[i] != nl[i-1]+1)
    {
      if ( (nl[i] - rem) <= 1)
      {
	sl += comma+string(nl[i-1]);
      }
      else 
      {
	sl += comma+string(rem)+"-"+string(nl[i-1]);
      }
      comma = ",";
      rem = nl[i];
    }
    
  }

  % print what is on the "stack"
  if ( (nl[-1] - rem) < 1)
  {
    sl += comma+string(nl[i]);
  }
  else 
  {
    sl += comma+string(rem)+"-"+string(nl[-1]);
  }
  comma = ",";
  rem = nl[-1];    
  
  return sl;
}

%
% Code from src/data/notice_listFromString.sl
%
define notice_listFromString(str)
%!%+
%\function{notice_listFromString}
%\synopsis{creates a isis notice_list from a string containing the
%noticed bins separated with ",".}
%\usage{Array_Type notice = notice_listFromString(String_Type str);}                
%\seealso{notice_list, notice_listToString, get_data_info, fits_save_fit}       
%!%-
{
  
  variable sub = strtok(str,",");
  variable n = length(sub);
  variable nl = Integer_Type[n];
  
  variable i;
  _for i(0,n-1,1)
  {
    nl[i] = integer(sub[i]);
  }

  return nl;
}

%
% Code from src/data/notice_listToString.sl
%
define notice_listToString(nl)
%!%+
%\function{notice_listToString}
%\synopsis{converts an isis notice_list to a string}
%\usage{String_Type str = notice_listToString(Array_Type notice_list);}                
%\seealso{notice_list, notice_listFromString, get_data_info, fits_save_fit}       
%!%-
{
  variable result = "";
  variable comma = "";
  
  variable i;
  _for i(0,length(nl)-1,1)
  {
    result += comma+string(nl[i]);
    comma = ",";    
  }
  return result;
}

%
% Code from src/slang/trees/ntree.sl
%
define ntree();
%!%+
%\function{ntree}
%\synopsis{creates a new node of a tree with n children}
%\usage{Struct_Type ntree(DataType_Type objectType[, Struct_Type ntreeType]);
% or Struct_Type ntree(DataType_Type objectType[, Integer_Type numChilds]);}
%\description
%  An n-tree is a special kind of a tree data structure.
%  Each node has exactly n children, which are n-tree
%  nodes themselves. The number of children is specified,
%  e.g., by the 'numChilds' parameter, which is 2 by
%  default. Initially, all children are null pointers.
%  Most importantly, a node contains a list of objects,
%  which is an array of the type specified by the
%  'objectType' parameter. Using the functions provided
%  in the n-tree structure, objects can be inserted or
%  removed.
%  
%  The n-tree structure contains the following functions:
%    insert    - appends an object to the list.
%                Usage: ntree.insert(objectType object);
%    remove    - removes an object from the list and
%                returns 1 on success, 0 otherwise.
%                Usage: Integer_Typer ntree.remove(objectType object);
%    get       - returns all objects stored in the node.
%                Usage: objectType[] ntree.get();
%    getChilds - returns all objects stored in the children.
%                Usage: objectType[] ntree.getChilds();
%    getAll    - iterates over all nodes recursively to
%                return all objects stored in the tree.
%                Usage: objectType[] ntree.getAll();
%    childN    - returns the N's child of the actual node.
%                If the child does not exist, NULL will be
%                returned. The function names can be
%                overwritten by the type definition (see
%                below).
%                Usage: Struct_Type ntree.childN();
%                Qualifiers:
%                  new    - creates the child if it does
%                           not exist
%                  renew  - creates a new child no matter
%                           if it exists already
%                  delete - deletes the child
%
%  In addition, the following fields exist:
%    objects   - array of objects stored in the node
%    childs    - array of length N containing the
%                children of the node
%    type      - structure describing the type of the
%                n-tree (see below)
%
%  The type of the n-tree can be specified by the
%  optional 'ntreeType' parameter. This structure
%  defines additional behaviours of the n-tree, and
%  the number of children and their the return function
%  names eventually. The default type sets these names
%  to the ones described above (childN). Additional
%  informations about the usage of n-tree types can
%  be found in the help of the 'ntree_type' function.
%\example
%  % creates an 4-tree (called quad tree) with one
%  % initial node and an object array of Integer_Type
%  tree = ntree(Integer_Type, 4);
%
%  % insert '5' into the initial node
%  tree.insert(5);
%
%  % create and insert '10' into the third child
%  tree.child3(; new).insert(10);
%
%  % checks if child2 exists
%  if (tree.child2() == NULL)
%    message("child2 does not exist");
%
%  % insert '8' into child3, but it is not
%  % created again since it exists already
%  tree.child3(; new).insert(8);
%
%  % get all objects of the children
%  % (will return [10,8])
%  print(tree.getChilds());
%\seealso{ntree_type}
%!%-


%%%
% private functions
%%%

% inserts an object into the tree
private define ntree_insert_method(nargs) {
  variable node, new;
  variable addargs = (nargs > 2 ? __pop_args(nargs-2) : Struct_Type[0]); % additional parameters for type specific insert function
  switch (nargs - length(addargs))
    { case 2: (node, new) = (); }
    { vmessage("error (%s): object to be inserted not given", _function_name); return; }

  % call insert callback function
  if (struct_field_exists(node.type, "insert")) {
    variable insertfun = node.type.insert; % prevent type structure to be pushed onto the stack
    node = @insertfun(node, new, __push_args(addargs) ;; __qualifiers);
  }
  
  if (node != NULL) node.objects = [node.objects, typecast(new, _typeof(node.objects))];
}

% removes an object from the tree
private define ntree_remove_method(nargs) {
  variable node, del;
  variable addargs = (nargs > 2 ? __pop_args(nargs-2) : Struct_Type[0]); % additional parameters for type specific delete function
  switch (nargs - length(addargs))
    { case 2: (node, del) = (); }
    { vmessage("error (%s): object to be deleted not given", _function_name); return; }

  variable len = length(node.objects);

  % call delete callback function
  if (struct_field_exists(node.type, "remove")) {
    variable removefun = node.type.insert; % prevent type structure to be pushed onto the stack
    node = @removefun(node, del, __push_args(addargs) ;; __qualifiers);
  }
  
  if (node != NULL) node.objects = node.objects[wherenot(node.objects == del)];

  return (len != length(node.objects) && node != NULL);
}

% returns the objects within a node
private define ntree_get_method(nargs) {
  variable node;
  switch(nargs)
    { case 1: (node) = (); }

  return node.objects;
}

% returns the objects within all children
private define ntree_getChilds_method(nargs) {
  variable node;
  switch(nargs)
    { case 1: (node) = (); }

  variable obj = _typeof(node.objects)[0];
  variable i;
  _for i (0, length(node.childs)-1, 1)
    if (node.childs[i] != NULL) obj = [obj, node.childs[i].objects];
  return obj;
}

% returns all objects within the n-tree
private define ntree_getAll_method();
private define ntree_getAll_method(nargs) {
  variable node, visited;
  switch(nargs)
    { case 1: (node) = (); visited = Struct_Type[0]; }
    { case 2: (node, visited) = (); }

  variable obj = _typeof(node.objects)[0];
  % proceed if node has not been visited yet
  if (wherefirst(visited == node) == NULL || length(visited) == 0) {
    visited = [visited, node]; % set this node as visited
    obj = node.objects;
    % recurse children
    variable child;
    foreach child (node.childs) if (child != NULL) obj = [obj, ntree_getAll_method(child, visited, 2)];
  };

  return obj;
}

% returns the n^th child node structure
private variable ntree_child_method_definition = `private define ntree_child_method(n, nargs) {
  variable node;
  switch(nargs)
    { case 1: (node) = (); }

  % eventually create new child
  % qualifiers important since they are passed to the type dependent creation function
  if ((qualifier_exists("new") && node.childs[n-1] == NULL) || qualifier_exists("renew")) node.childs[n-1] = ntree(_typeof(node.objects); parent = node, index = n-1);
  % delete a child (note: children will NOT be deleted recursively)
  if (qualifier_exists("delete")) node.childs[n-1] = NULL;

  return node.childs[n-1];
}`;

%%%
% handler functions
%%%

private define ntree_insert() {
  variable args = __pop_args(_NARGS);
  return ntree_insert_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define ntree_remove() {
  variable args = __pop_args(_NARGS);
  return ntree_remove_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define ntree_get() {
  variable args = __pop_args(_NARGS);
  return ntree_get_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define ntree_getChilds() {
  variable args = __pop_args(_NARGS);
  return ntree_getChilds_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define ntree_getAll() {
  variable args = __pop_args(_NARGS);
  return ntree_getAll_method(__push_args(args), _NARGS ;; __qualifiers);
}
private variable ntree_child_definition = `private define ntree_child%d() {
  variable args = __pop_args(_NARGS);
  return ntree_child_method(__push_args(args), %d, _NARGS ;; __qualifiers);
}`;


%%%
% main functions
%%%

% returns a new n-tree node
define ntree() {
  variable objtype, type;
  variable addargs = (_NARGS > 2 ? __pop_args(_NARGS-2) : Struct_Type[0]); % additional parameters for type specific creation function
  switch (_NARGS - length(addargs))
    { case 1: (objtype) = (); }
    { case 2: (objtype, type) = (); }
    { help(_function_name); return; }

  variable parent = qualifier("parent", NULL);
  variable index  = qualifier("index", NULL);
  % default structure defining a binary tree
  ifnot (__is_initialized(&type))
    type = (parent == NULL ? struct { numChilds = 2 } : parent.type);
  % number of children given
  else if (typeof(type) == Integer_Type)
    type = struct { numChilds = type };
  % type checks
  else if (typeof(type) != Struct_Type)
    { vmessage("error (%s): either a type structure or the number of children has to be given", _function_name); return; }
  ifnot (struct_field_exists(type, "numChilds")) { vmessage("error (%s): the number of children has to be given", _function_name); return; }
  ifnot (struct_field_exists(type, "childNames"))
    type = struct_combine(type, struct { childNames = array_map(String_Type, &sprintf, "child%d", [1:type.numChilds]) });
  if (length(type.childNames) != type.numChilds) { vmessage("error (%s): the number of child names has to be %d", _function_name, type.numChilds); return; }

  % build child function structure
  variable childfun = struct_combine(type.childNames);
  % loop to define the private child functions
  % is this a cheat for function nesting?
  variable n;
  eval(ntree_child_method_definition, "isis");
  _for n (1, type.numChilds, 1) {
    eval(sprintf(ntree_child_definition, n, n), "isis");
    set_struct_field(childfun, type.childNames[n-1], eval(sprintf("&ntree_child%d", n), "isis"));
  }
  
  % build node structure
  variable node = struct_combine(
    struct {
      insert    = &ntree_insert,
      remove    = &ntree_remove,
      get       = &ntree_get,
      getChilds = &ntree_getChilds,
      getAll    = &ntree_getAll
    },
    childfun,
    struct {
      objects = objtype[0],
      childs  = Struct_Type[type.numChilds],
      type    = type
    }
  );

  % call creation callback function
  if (struct_field_exists(type, "new")) {
     variable newfun = type.new; % prevent type structure to be pushed onto the stack
     node = @newfun(node, index, parent, __push_args(addargs) ;; __qualifiers);
  }
  
  return node;
}

%
% Code from src/slang/trees/ntree_types.sl
%
% -*- mode: slang; mode: fold -*-

% map quadtree %{{{
%!%+
%\function{ntree_type_mapQuad}
%\synopsis{returns the n-tree definition of a map quadtree}
%\usage{Struct_Type ntree(DataType_Type objectType, ntree_type("mapQuad"), Double_Type size[, Double_Type x0, y0]); }
%\description
%  A map quadtree is a 4-tree, which divides a 2d-space
%  into squares with an edge length of 'size'. Therefore,
%  each node of the tree is identified by a certain
%  coordinate x,y. The coordinate of the tree's root can
%  be specified by the x0 and y0 parameters and are (0|0)
%  by default. The children are named "top", "right",
%  "bottom" and "left". Due to that structure the tree
%  will be an undirected graph, meaning that, e.g., the
%  left node is connected to the right node and vise versa:
%
%   -------            ----            -------
%  |x0-size| -right-> | x0 | -right-> |x0+size|
%  |  y0   | <--left- | y0 | <--left- |  y0   |
%   -------            ----            -------
%                       ^|
%                   top || bottom
%                       |v
%                    -------
%                   |   x0  |
%                   |y0-size|
%                    -------
%
%  The map quadtree extends the n-tree structure by the
%  following functions:
%    lookup(x,y) - returns the node fitting the given
%      x and y coordinates
%    insert(object[,x,y]) - overwrites the usual insert
%      function such, that the object is inserted into
%      the node fitting the optional x and y coordinates
%
%  Note: deleting a child by the 'delete' qualifier,
%    e.g. tree.left(; delete) does NOT cascade! That is,
%    the left node still points to the right one (here
%    called tree). It might still be accessable via
%    tree.top().left().bottom()
%\example
%\seealso{ntree, ntree_type}
%!%-
private define ntree_type_mapQuad_lookup_method(nargs) {
  variable node, x, y;
  switch(nargs)
    { case 3: (node, x, y) = (); }
    { vmessage("error (%s): x and y coordinates have to be given", _function_name); return; }

  % return yourself if no coordinates are given
  if (x == NULL && y == NULL) return node;
  % return yourself if coordinates are within
  if (abs(node.x - x) <= .5*node.size && abs(node.y - y) <= .5*node.size) return node;
  % pass request to child
  else { 
    if (x < node.x - .5*node.size) return node.left(; new).lookup(x, y);
    else if (x > node.x + .5*node.size) return node.right(; new).lookup(x, y);
    else if (y < node.y + .5*node.size) return node.bottom(; new).lookup(x, y);
    else return node.top(; new).lookup(x, y);
  }
}

private define ntree_type_mapQuad_lookup() {
  variable args = __pop_args(_NARGS);
  return ntree_type_mapQuad_lookup_method(__push_args(args), _NARGS ;; __qualifiers);
}

private define ntree_type_mapQuad_new() {
  variable node, c, parent, size, x, y;
  switch (_NARGS)
    { case 3: (node, c, parent) = (); x = 0; y = 0; }
    { case 4: (node, c, parent, size) = (); x = 0; y = 0; }
    { case 6: (node, c, parent, size, x, y) = (); }
    { vmessage("error (%s): uncorrect number of arguments", _function_name); return; }

  if (parent == NULL && (not __is_initialized(&size))) { vmessage("error (%s): size is not specified", _function_name); return; }

  % extend node structure
  node = struct_combine(struct {
    x      = (parent == NULL ? x : parent.x + (c == 1 ? 1 : c == 3 ? -1 : 0)*parent.size),
    y      = (parent == NULL ? y : parent.y + (c == 0 ? 1 : c == 2 ? -1 : 0)*parent.size),
    size   = (parent == NULL ? size : parent.size),
    lookup = &ntree_type_mapQuad_lookup
  }, node);
  % set coordinates
  % set one child as parent accordingly
  if (parent != NULL) node.childs[[2,3,0,1][c]] = parent;

  return node;
}

private define ntree_type_mapQuad_insert() {
  variable node, new, x, y;
  switch (_NARGS)
    { case 2: (node, new) = (); return node; }
    { case 4: (node, new, x, y) = (); return node.lookup(x, y); }
    { vmessage("error (%s): uncorrect number of arguments", _function_name); }
}%}}}


% available n-tree types
private variable ntree_types = struct {
  mapQuad = struct {
    numChilds  = 4,
    childNames = ["top", "right", "bottom", "left"],
    new        = &ntree_type_mapQuad_new,
    insert     = &ntree_type_mapQuad_insert
  }
};

%%%%%%%%%%%%%%%
define ntree_type()
%%%%%%%%%%%%%%%
%!%+
%\function{ntree_type}
%\synopsis{returns a specific stucture defining an n-tree type}
%\usage{Struct_Type ntree_type(String_Type typeName); }
%\description
%  This functions returns the structure defining
%  the specific n-tree type named 'typeName'. A
%  list of all available types is shown if no
%  name is passed. The 'help' qualifier shows
%  additional information about the given type.
%
%  The type of an n-tree is specified and can be
%  extended by a structure with the following
%  fields:
%  
%  numChilds
%    Defines the number of children of each node.
%    node. It has the same effect as the
%    'numChilds' parameter of the 'ntree' function.
%
%  childNames (optional)
%    Array of strings defining the return function
%    names for each children. The default names are
%    "child1" to "childN".
%
%  new (optional)
%    Reference to a callback function, which gets
%    called right before a new node created by
%    'ntree()' is returned. Parameters passed are
%    (1) the new node as a structur, (2) as which
%    child number it is created (starting at 1,
%    NULL if it is the root) and (3) the parent
%    node it belongs to (NULL if there is none).
%    The function has to return the new node.
%    
%  insert (optional)
%    Reference to a callback function, which gets
%    called before an object is inserted into a
%    node by 'node.insert()'. Parameters passed are
%    (1) the node as a structure and (2) the object,
%    which should be inserted. The function has to
%    return the node. If NULL is returned, the
%    object will not be inserted into the node.
%
%  remove (optional)
%    Reference to a callback function, which gets
%    called before an object is removed from a node
%    by 'node.remove()'. Parameters passed are (1)
%    the node as a structure and (2) the object,
%    which should be removed. The function has to
%    return the node. If NULL is returned, the
%    object will not be removed from the node.
%\example
%  % first define a callback function, which
%  % prevents inserting negative numbers
%  define ntree_myType_insert(node, object) {
%    if (object < 0) return NULL;
%    else return node;
%  }
%
%  % structure defining a 4-tree (called quadtree)
%  % and the children return functions are renamed
%  % to match directions. The insert callback
%  % function defined above is assigned also.
%  variable myType = struct {
%    numChilds = 4,
%    childNames = ["left","right","top","bottom"],
%    insert = &ntree_myType_insert
%  };
%
%  % create a new n-tree of myType
%  variable tree = ntree(Double_Type, myType);
%\seealso{ntree}
%!%-
{
  variable type, t;
  switch(_NARGS)
    { case 1: (type) = (); }
    { help(_function_name); message(" AVAILABLE TYPES"); foreach t (get_struct_field_names(ntree_types)) vmessage("    %s", t); return; }

  ifnot (struct_field_exists(ntree_types, type)) {
    vmessage("type '%s' not known\navailable types:", type);
    foreach t (get_struct_field_names(ntree_types)) vmessage("  %s", t);
    return;
  }

  if (qualifier_exists("help")) help(sprintf("ntree_type_%s", type));
  else return get_struct_field(ntree_types, type);
}

%
% Code from src/data/nuFnu.sl
%

define nuFnu()
%%%%%%%%%%%%%%%%
%!%+
%\function{nuFnu}
%
%\synopsis{changes an Energy/Flux Spectrum to a Frequency/Flux*Frequency Spectrum}
%\usage{Struct_Type = nuFnu (hist_index, E_min(keV), E_max(keV));}
%
%\description
%    Use this function on a dataset that has been modeled.
%    This function uses get_data_flux/eval_fun_keV to load
%    spectral data into isis and convert energy bins in a 
%    given range to frequency (s^-1) bins as well as Flux 
%    to flux*frequency.
%    If no range specified: E_min=0.5 keV, E_max=10 keV.
%    Values and Errors are given in erg s^-1 cm^-2.
%
%    WARNING: Currently only frequency has been implemented 
%             as x-unit.
%             
%    For the evaluation of a deabsorbed component use the 
%    qualifier "deabs". deabs will try to calculate the 
%    factor between absorbed and deabsorbed model 
%    components and correct the data.
%
%    WARNING: This does not work when a convolution 
%             normalization (e.g., cflux, enflux, phflux)
%             is part of the model!
%
%    The "numbin" qualifier allows to specify the number
%    of output bins. Currently only a logarithmic grid 
%    has been implemented. If numbin is not specified 
%    the flux for the current grid will be returned.
%
%    The "group" qualifier allows to specify the S/N that
%    will be used for the flux calculation. Flux calculations
%    are only correct for large number of bins. The final 
%    grid can be smaller (qualifier: numbin).
%
%
%    WARNING: Currently tested only on low counts spectra.
%             S/N might have to be increased for large 
%             number of counts.
%
%\qualifiers{
%\qualifier{deabs}{   evaluate deabsorbed flux}
%\qualifier{ff}{      define deabsorbed fit function}
%\qualifier{numbin}{  define number of bins}
%\qualifier{group}{  define S/N for grouping}
%}
%
%\example
%    isis>xray = load_data("data.pha");
%    isis>load_par("x.par");
%    isis>() = fit_counts;
%    isis>new=nuFnu(xray,2,10);	 %change spectrum (2-10 keV) into frequency spectrum
%
%    %Plot data:
%    isis>hplot(new.lo, new.hi, new.val);
%
%    % Evaluate Function for deabsorbed model
%    % (WARNING: DOES NOT WORK IF ENFLUX IS USED IN FIT FUNCTION)
%
%    isis>xray = load_data("data.pha");
%    isis>fit_fun("tbnew_simple_z(1)*powerlaw(1)");
%    isis>() = fit_counts;
%    isis>new = nuFnu(xray,2,10; deabs, ff="powerlaw(1)");
%    isis>%Calculate a deabsorbed powerlaw
%
%\seealso{get_data_flux, rebin_mean, eval_fun_keV, group}
%!%-
{
    variable dset;
    variable emin = NULL, emax = NULL;
    switch(_NARGS)
    { case 1: dset  = (); }
    { case 3: (dset,emin,emax) = ();}
    { help(_function_name()); return; }

    variable deabs = qualifier("deabs");  
    variable ffq = qualifier("ff", NULL);
    variable numbin = qualifier("numbin", NULL);  
    variable sngroup = qualifier("group", NULL); 

    % Check if emin & emax & numbin have been specified

    if (emin == NULL) 
    {
	emin=0.5;
    }
    if (emax == NULL) 
    {
	emax=10.;
    }


    % Define Constants 

    variable pla = 4.13566751e-18; %planck in keV*s
    variable newflu, newerr,newfac,blo,bhi;

    % Rebin here to low S/N
    % Find optimal binning!!

    variable gdi = get_data_info(dset);
    if(length(gdi.notice) != length(gdi.rebin))
    {	
	if (sngroup != NULL)
	{
	    group(dset;min_sn=sngroup,bounds=emin,unit="keV");
	}
    }


    % Get data flux

    flux_corr(dset);

    variable datflux = _A(get_data_flux(dset));
    struct_filter(datflux, where(datflux.bin_lo > 0.99*emin));
    struct_filter(datflux, where(datflux.bin_hi < emax));

    % Central bin energy in erg
    variable bincen = keV2erg(0.5*(datflux.bin_lo+datflux.bin_hi));

    %# FIX THIS LATER!!!!
    % Get bins in Hz
    datflux.bin_lo = datflux.bin_lo/pla;
    datflux.bin_hi = datflux.bin_hi/pla;


    % Central bin energy in Hz
    variable bincenhz = 0.5*(datflux.bin_lo+datflux.bin_hi);

    % Calculate flux values
    datflux.value = datflux.value/(datflux.bin_hi-datflux.bin_lo)*bincen*bincenhz;
    datflux.err = datflux.err/(datflux.bin_hi-datflux.bin_lo)*bincen*bincenhz;


    % Make grid - optional
    if (numbin != NULL)
    {	
	variable logrid,higrid;
        (logrid,higrid) = log_grid(emin/pla,emax/pla,numbin);
	newflu = rebin_mean(logrid,higrid,datflux.bin_lo,datflux.bin_hi,
	datflux.value);
	newerr = rebin_mean(logrid,higrid,datflux.bin_lo,datflux.bin_hi,datflux.err);
	blo = logrid;
	bhi = higrid;
    }
    else
    {
	% Restore bins & bounds
	rebin_data(dset,gdi.rebin);
	notice_list(dset,gdi.notice_list); 

	% Get original binning
	variable datcounts = _A(get_data_counts(dset));
	struct_filter(datcounts, where(datcounts.bin_lo > 0.99*emin));
	struct_filter(datcounts, where(datcounts.bin_hi < emax));

	% Get bins in Hz
	datcounts.bin_lo = datcounts.bin_lo/pla;
	datcounts.bin_hi = datcounts.bin_hi/pla;


	blo = datcounts.bin_lo;
	bhi = datcounts.bin_hi;
	
	if (length(datcounts.bin_lo) != length(datflux.bin_lo))
	{
	    newflu = rebin_mean(datcounts.bin_lo,datcounts.bin_hi,datflux.bin_lo,datflux.bin_hi,datflux.value);
	    newerr = rebin_mean(datcounts.bin_lo,datcounts.bin_hi,datflux.bin_lo,datflux.bin_hi,datflux.err);
	}
	else
	{
	    newflu = datflux.value;
	    newerr = datflux.err;
	}


    }

    % Get Absorption Factor

    if (qualifier_exists("deabs"))
    {
	variable gff = get_fit_fun;
	save_par("yxzylabyxpar.par");
	if (gff == NULL) 
	{
	    vmessage("No Fit Function specified");
	    variable all_data_n;
	}
	else
	{
	    % Get Fit Functions
	    variable agd = _A(get_data_counts(1));
	    struct_filter(agd, where(agd.bin_lo > 0.99*emin));
	    struct_filter(agd, where(agd.bin_hi < emax));
	    variable abs2 = eval_fun_keV(agd.bin_lo, agd.bin_hi);
	    fit_fun(ffq);
	    variable newgff = get_fit_fun;
	    vmessage(sprintf("Deabsorbed Fit Function found: %s", newgff));
	    vmessage(sprintf("Fit Function found: %s", gff));

	    variable fverb = Fit_Verbose;
	    Fit_Verbose=-1;
	    () = eval_counts;
	    Fit_Verbose=fverb;
	   
	    variable deabs2 = eval_fun_keV(agd.bin_lo, agd.bin_hi);

	    %Calculate deabsorbing factor
	    variable factor=deabs2/abs2;

	    if (numbin != NULL)
	    {
		newfac = rebin_mean(logrid,higrid,datflux.bin_lo,datflux.bin_hi,factor);
	    }
	    else 
	    {

		if (length(datcounts.bin_lo) != length(datflux.bin_lo))
		{
		    newfac = rebin_mean(datcounts.bin_lo,datcounts.bin_hi,datflux.bin_lo,datflux.bin_hi,factor);
		}
		else
		{
		    newfac=factor;
		}
		
	    }

	    newflu = newflu*newfac;
	    newerr = newerr*newfac;

	}
    }
    
    variable new_struct = struct {bin_lo = blo, bin_hi = bhi, value=newflu, err=newerr };

    return new_struct;
}


%
% Code from src/data/num_bin.sl
%
%%%%%%%%%%%%%%%%
define num_bin()
%%%%%%%%%%%%%%%%
%!%+
%\function{num_bin}
%
%\synopsis{Number of noticed bins}
%\usage{Double_Type = num_bin (hist_index);}
%\description
%       Use this function to retrieve number of
%       noticed bins.
%\qualifiers{
%    \qualifier{dof}{gives degrees of freedom instead of number of bins}
%}
% 
%\example
%       isis>xray = load_data("data.pha");
%       isis>variable num = num_bin(1);
%
%\seealso{dof}
%!%-
{
    variable dset;

    switch(_NARGS)
    { case 1: dset  = (); }
    { help(_function_name()); return; }

    variable ff = get_fit_fun();
    variable fv = Fit_Verbose;
    variable tmp_stat;

    Fit_Verbose=-1;
    () = eval_counts(&tmp_stat);
    Fit_Verbose = fv;

    variable nbin = typecast (tmp_stat.num_bins, Double_Type);
    variable nvar = typecast (tmp_stat.num_variable_params, Double_Type);
    variable number = nbin;
    variable dof1 = nbin-nvar;

    if (qualifier_exists("dof")){
	return dof1;
    }
    else {
	return number;
    }
}  

%
% Code from src/slang/astronomy/nutation_angles.sl
%
%
% Implement the IAU 2000A Nutation theory
%

%
% private variables to cache the nutation theories
%
private variable nutation2000a_ls;
private variable nutation2000a_pl;
private variable nutation2000k_ls;
private variable nutation2000k_pl;
private variable nutation2000b_ls;

define nutation_angles();

define nutation_angles()
%!%+
%\function{nutation_angles}
%\synopsis{calculate the nutation in longitude and obliquity for a given date}
%\usage{ (dpsi,deps)=nutation_angles(JD;qualifiers)}
%\qualifiers{
%   \qualifier{JD_frac}{fractional part of the Julian Date (for highest precision)}
%   \qualifier{theory}{nutation theory to use. Possible string values are
%           IAU2000A - full IAU2000A theory, i.e., an implementation of the
%                      theory of Matthews et al. (2002, J. Geophys. Res, 107, 2068)
%           IAU2000B - truncated IAU2000A theory per McCarthy & Luzum (2003,
%                      Cel. Mech. Dyn. Astron, 85, 37); accuracy is
%                      1 mas in 1995-2050
%           NU2000K  - truncated IAU2000A theory from Kaplan (2009, USNO
%                      circular 181); accuracy is 0.1mas in 1700-2300
%                      [the default]}
%   \qualifier{mjd}{the time argument is in MJD, not in JD}
%   \qualifier{JD_frac}{fractional part of the time argument. Use this if
%                     highest precision is needed. The nutation angles are then
%                     returned for the date JD+JD_frac.}
%   \qualifier{deg}{return nutation angles in degrees}
%   \qualifier{arcsec}{return nutation angles in arcseconds}
%   \qualifier{mas}{return nutation angles in milliarcseconds}
%   \qualifier{iaufile}{FITS file containing the theory coefficients
%                   (usually found in ISISSCRIPTS_REFPATH)}
% 
%}
%\description
% This routine calculates the nutation angles for the forced nutation
% of the non-rigid earth according to one of the nutation theories described
% above for date given by the JD-argument (which is interpreted as TT).
% For most applications, usage of the NU2000K or IAU2000B routines
% is fully sufficient.
%
% This routine is array safe (but JD_frac MUST be a scalar in this case!),
% the angles returned are in radians unless one of the deg, arcsec, or mas
% keywords is given.
%
% The routine is based on the NOVAS 3.1 C-code (Bangert et al., 2011, see
% http://aa.usno.navy.mil/software/novas/novas_info.php )
%
%\seealso{precess}
%!%-
{
    variable JD;
    switch(_NARGS)
    { case 1 : JD=(); }
    { help(_function_name()); return;}

    %
    % if called with an array, call this routine for each time argument
    %
    if (Array_Type==typeof(JD)) {
	variable i;
	variable dpsi=Double_Type[length(JD)];
	variable deps=Double_Type[length(JD)];
	_for i(0,length(JD)-1,1) {
	    (dpsi[i],deps[i])=nutation_angles(JD[i];;__qualifiers());
	}
	return dpsi,deps;
    }
    
    %
    % read FITS file?
    %
    ifnot (__is_initialized(&nutation2000a_ls)) {
	variable searchpath=getenv("ISISSCRIPTS_REFPATH");
	if (searchpath==NULL and qualifier_exists("iaufile")==0) {
          throw UsageError,sprintf("%s: iaufile qualifier not given and ISISSCRIPTS_REFPATH not found\n",_function_name());
      }
      if (searchpath==NULL) {
          searchpath=".";
      }
      variable nutfi=qualifier("iaufile",searchpath+"/nutation_iau2000.fits.gz");
      % the following opens the file multiple times
      nutation2000a_ls=fits_read_table(nutfi+"[IAU2000A_LS]");
      nutation2000a_pl=fits_read_table(nutfi+"[IAU2000A_PL]");
      nutation2000b_ls=fits_read_table(nutfi+"[IAU2000B_LS]");
      nutation2000k_ls=fits_read_table(nutfi+"[IAU2000K_LS]");
      nutation2000k_pl=fits_read_table(nutfi+"[IAU2000K_PL]");
    }


    % get the theory coefficients
    variable theory_ls, theory_pl;
    variable theory=qualifier("theory","NU2000K");
    switch (theory)
    { case "IAU2000A":
	theory_ls=nutation2000a_ls;
	theory_pl=nutation2000a_pl;
    }	
    { case "IAU2000B":
	theory_ls=nutation2000b_ls;
	theory_pl=NULL;
    }
    { case "NU2000K":
	theory_ls=nutation2000k_ls;
	theory_pl=nutation2000k_pl;
    }
    {throw UsageError,sprintf("%s: Unknown theory for nutation.\n",_function_name()); }

    variable T;
    if (qualifier_exists("mjd")) {
	T=((JD-51544.5)+qualifier("JD_frac",0.))/36525.0;
    } else {
	T=((JD-2451545.)+qualifier("JD_frac",0.))/36525.0;
    }

    % conversion parameters
    variable asec360=3600.*360.;
    variable asec2rad=PI/648000.;

    variable L, Lprime, F, D, Om;
    if (theory=="IAU2000B") {
	% Delaunay angles from Simon et al. (1994)
	% (used in IAU 2000B)


	% Mean anomaly of the Moon
	L = ((485868.249036 + T * 1717915923.2178) mod asec360) * asec2rad;

	% Mean anomaly of the Sun
	Lprime = ((1287104.79305 + T * 129596581.0481) mod asec360) * asec2rad;

	% Mean argument of the latitude of the Moon
	F = ((335779.526232 + T * 1739527262.8478) mod asec360) * asec2rad;

	% Mean elongation of the Moon from the Sun.
	D = ((1072260.70369 + T * 1602961601.2090) mod asec360) * asec2rad;

	% Mean longitude of the ascending node of the Moon.
	Om = ((450160.398036 - T * 6962890.5431) mod asec360) * asec2rad;
    } else {	
	% Fundamental arguments from the Simon et al. (1994) lunar theory
	% full polynomials are used for IAU 2000A and NU 2000K

	% Mean anomaly of the Moon
        L = ((485868.249036 +
                   T * (1717915923.2178 +
                   T * (        31.8792 +
                   T * (         0.051635 +
                   T * (       - 0.00024470))))) mod asec360) * asec2rad;

	% Mean anomaly of the Sun
        Lprime= ((1287104.79305 +
                       T * ( 129596581.0481 +
                       T * (       - 0.5532 +
                       T * (         0.000136 +
                       T * (       - 0.00001149))))) mod asec360) * asec2rad;

	% Mean argument of the latitude of the Moon
        F = ((335779.526232 +
                   T * (1739527262.8478 +
                   T * (      - 12.7512 +
                   T * (      -  0.001037 +
                   T * (         0.00000417))))) mod asec360) * asec2rad;

	% Mean elongation of the moon from the Sun
	D = ((1072260.70369 +
                   T * (1602961601.2090 +
                   T * (       - 6.3706 +
                   T * (         0.006593 +
                   T * (       - 0.00003169))))) mod asec360) * asec2rad;

	% Mean longitude of the Moon's ascending node
	Om= ((450160.398036 +
                   T * ( - 6962890.5431 +
                   T * (         7.4722 +
                   T * (         0.007702 +
                   T * (       - 0.00005939))))) mod asec360) * asec2rad;
    } 
	
    %
    % calculate the luni-solar nutation
    %
    variable twopi=2.*PI;
    
    variable arg=(theory_ls.l*L+theory_ls.lprime*Lprime+
                 theory_ls.f*F+theory_ls.d*D+theory_ls.om*Om) mod twopi;
    variable sc,cc;
    (sc,cc)=sincos(arg);

    %
    % should probably do a reverse sort first
    %
    variable dp_ls=sum((theory_ls.l_sin+T*theory_ls.l_tsin)*sc+theory_ls.l_cos*cc);
    variable de_ls=sum((theory_ls.e_cos+T*theory_ls.e_tcos)*cc+theory_ls.e_sin*sc);
    
    %
    % planetary nutation
    %
    variable dp_pl,de_pl;
    if (NULL!=theory_pl) {
	variable al,alsu,af,ad,aom,apa,alme,alve,alea,alma,alju,alsa,alur,alne;

	if (theory=="IAU2000A") {
	    %
	    % anomalies and longitudes for the IAU2000A theory
	    %
	    
	    % mean anomaly of the Moon 
	    al = (2.35555598 + 8328.6914269554 * T) mod twopi;

	    % Mean anomaly of the Sun
	    alsu=(6.24006013 + 628.301955 * T) mod twopi;

	    % Mean argument of the latitude of the Moon
	    af = (1.627905234 + 8433.466158131 * T) mod twopi;

	    % Mean elongation of the Moon from the Sun
	    ad = (5.198466741 + 7771.3771468121 * T) mod twopi;

	    % Mean longitude of the ascending node of the Moon.
	    aom = (2.18243920 - 33.757045 * T) mod twopi;

	    % General accumulated precession in longitude.
	    apa = (0.02438175 + 0.00000538691 * T) * T;

	    % Planetary longitudes, Mercury through Neptune (Souchay et al. 1999).
	    alme = (4.402608842 + 2608.7903141574 * T) mod twopi;
	    alve = (3.176146697 + 1021.3285546211 * T) mod twopi;
	    alea = (1.753470314 +  628.3075849991 * T) mod twopi;
	    alma = (6.203480913 +  334.0612426700 * T) mod twopi;
	    alju = (0.599546497 +   52.9690962641 * T) mod twopi;
	    alsa = (0.874016757 +   21.3299104960 * T) mod twopi;
	    alur = (5.481293871 +    7.4781598567 * T) mod twopi;
	    alne = (5.321159000 +    3.8127774000 * T) mod twopi;
	} else {
	    %
	    % Terms for the IAU 2000K theory
	    %

	    %
	    % lunar terms (same as for IAU 2000A terms above)
	    %
	    al=L;
	    alsu=Lprime;
	    af=F;
	    ad=D;
	    aom=Om;
	    
	    % planetary longitudes from mercury through neptune wrt
	    % dynamical ecliptic and equinox of J2000, higher order
	    % terms omitted (Simon et al., 1994, Sect. 5.8.1-5.8.8)
	    alme = (4.402608842461 + 2608.790314157421 * T) mod twopi;
	    alve = (3.176146696956 + 1021.328554621099 * T) mod twopi;
	    alea = (1.753470459496 +  628.307584999142 * T) mod twopi;
	    alma = (6.203476112911 +  334.061242669982 * T) mod twopi;
	    alju = (0.599547105074 +   52.969096264064 * T) mod twopi;
	    alsa = (0.874016284019 +   21.329910496032 * T) mod twopi;
	    alur = (5.481293871537 +    7.478159856729 * T) mod twopi;
	    alne = (5.311886286677 +    3.813303563778 * T) mod twopi;

	    % General precession in longitude (Simon et al. 1994)
	    apa = (0.024380407358 + 0.000005391235 * T) * T;
	}

	% theory arguments 
	arg = (al*theory_pl.l   + alsu*theory_pl.lprime+af*theory_pl.f+
               ad*theory_pl.d   + aom*theory_pl.om+
               alme*theory_pl.me+alve*theory_pl.ve+
               alea*theory_pl.e +alma*theory_pl.ma+
               alju*theory_pl.ju+alsa*theory_pl.sa+
               alur*theory_pl.ur+alne*theory_pl.ne+
               apa*theory_pl.pre) mod twopi;
	(sc,cc)=sincos(arg);

	% should sort the array before summing
	dp_pl=sum(theory_pl.l_sin*sc+theory_pl.l_cos*cc);
	de_pl=sum(theory_pl.e_sin*sc+theory_pl.e_cos*cc);
    } else {
	%
	% fixed terms accounting for omission of long-period planetary
	% terms in the IAU 2000B theory
	%
	dp_pl=-1350.0; % 1e-7 arcsec
	de_pl= 3880.0; % 1e-7 arcsec
    }

    variable dp,de;
    if (qualifier_exists("deg")) {
	dp=(dp_ls+dp_pl)*1e-7*3600.;
	de=(de_ls+de_pl)*1e-7*3600.;
    } else {
	if (qualifier_exists("arcsec")) {
	    dp=(dp_ls+dp_pl)*1e-7;
	    de=(de_ls+de_pl)*1e-7;
	} else {
	    if (qualifier_exists("mas")) {
		dp=(dp_ls+dp_pl)*1e-1;
		de=(de_ls+de_pl)*1e-1;
	    } else {
		% radians
		dp=(dp_ls+dp_pl)*1e-7*asec2rad;
		de=(de_ls+de_pl)*1e-7*asec2rad;
	    }
	}
    }
    return (dp,de);
}

%
% Code from src/misc/play/obfuscate_string.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define obfuscate_string()
%%%%%%%%%%%%%%%%%%%%%%%
{
  variable s = ();

  variable c = Char_Type[0];
  foreach $1 (s)  if(not any(c==$1))  c = [c, $1];
  seed_random(_time());
  c = c[array_permute(c)];

  $1 = 2*length(c)/3;
  c = [c[[0:$1-1]], [32,101,114,114,111,144,32], c[[$1:]]];

  variable result = `("`;
  foreach $1 (c)  result += sprintf("%c", $1);
  result += `")[[`;
  variable i = String_Type[strlen(s)];
  _for $1 (0, strlen(s)-1, 1)
  { $2 = where(c==s[$1])[0];
    if($2<10)
      i[$1] = sprintf("%d", $2);
    else
      i[$1] = sprintf("0x%X", $2);
  }
  result += strjoin(i, ",") + "]]";

  message(result);
  return result;
}

%
% Code from src/slang/astronomy/object_visibility_satellite.sl
%
define object_visibility_satellite(ra,dec,mjdstart,mjdstop) {
%!%+
%\function{object_visibility_satellite}
%\synopsis{Calculate the time intervals when an object is observable with an astronomical satellite}
%\usage{visibility=object_visibility_satellite(ra,dec,mjdstart,mjdstop);}
%\qualifiers{
%\qualifier{dt}{Time resolution of the visibility calculation (days, default: 1)}
%\qualifier{numerical}{see below} 
%
%    Qualifiers to select the satellite. Constraints are ANDed if multiple
%    satellite qualifiers are given, i.e., the joint visibility is returned
%\qualifier{swift}{46-180 deg from Sun; so far ignores lunar constraint (>23deg)}
%\qualifier{suzaku}{70-110 deg from Sun}
%\qualifier{xmm}{70-110 deg from Sun}
%\qualifier{chandra}{>46deg from Sun; ignores lunar constraint (>6deg)}
%\qualifier{hst}{>50deg from Sun}
%\qualifier{nustar}{no solar avoidance constraints(!)}
%}
%\description
% This routine calculates the time intervals during which an astronomical
% source with position ra, dec (both given in degrees) is visible to a given
% astronomical satellite. Most popular facilities are supported (see above).
% For each time between mjdstart and mjdstop the routine calculates the
% angular separation between the source and the Sun. By default it returns
% an array of structures giving the time intervals during which a source is
% visible. The tags of the structure describing each interval are 
%    MJDstart and MJDstop: start and stop MJD of the visibility interval
%    MJDstring           : human readable string of the above
% An empty array is returned if the object is not visible during the time
% interval specified by mjdstart and mjdstop.
%
% If the qualifier "numerical" is set, the function returns a structure with
% the following tags
%    mjd: MJD
%    theta: separation betwee source and Sun (deg)
%    visible: boolean visibility (0: no, 1: visible)
% instead of the array of time intervals
%
% Note: if the visibility interval starts before mjdstart or extends beyond
%   mjdend, the start and stop times of the search interval are shown instead.
% Note 2: several satellites also have moon avoidance angles, or avoidance
%   angles for the Earth. Both are NOT (yet) taken into account.
%
%\example
% % convert source position into degrees
% variable ra=hms2deg(19,49,35.49);
% variable dec=dms2deg(+30,12,31.8);
%
% % visibility during the year 2013
% variable tstart=MJDofDate(2013,1,1);
% variable tstop=tstart+365;
%
% variable vis=object_visibility_satellite(ra,dec,tstart,tstop;suzaku,dt=0.1);
%
% variable i;
% for(i=0;i<length(vis);i++) {
%     printf("%s\n",vis[i].MJDstring);
% }
%!%-
    variable dt=qualifier("dt",1.);
    variable tstart=mjdstart;

    variable thmin=0.;
    variable thmax=360.;

    if (qualifier_exists("swift")) {
	thmin=max([thmin,46.]); 
	thmax=min([thmax,180.]);
	% also: moon: 23deg (???)
    }

    if (qualifier_exists("suzaku")) {
	thmin=max([thmin,70.]); 
	thmax=min([thmax,110.]);
    }

    if (qualifier_exists("xmm") || qualifier_exists("xmm-newton")) {
	thmin=max([thmin,70.]); 
	thmax=min([thmax,110.]);
    }

    if (qualifier_exists("chandra")) {
	thmin=max([thmin,46.]); 
	thmax=min([thmax,180.]);
	% also: moon: outside 6deg
    }

    if (qualifier_exists("hst")) {
	thmin=max([thmin,50.]); 
	thmax=min([thmax,180.]);
    }

    if (qualifier_exists("nustar")) {
	thmin=max([thmin,0.]); 
	thmax=min([thmax,180.]);
    }

    variable visibility=struct {mjd=[mjdstart:mjdstop:dt]};
    variable ilen=length(visibility.mjd);

    visibility=struct_combine(visibility,
                           struct { theta=Double_Type[ilen],
                                    visible=Int_Type[ilen]
                                  }
    );

    variable i;
    _for i (0, ilen-1, 1) {
	variable eqsun,eclsun;
	(eqsun,eclsun)=sunpos(visibility.mjd[i];MJD);

	variable dist=greatcircle_distance(ra,dec,eqsun.ra,eqsun.dec;unit="deg");
	dist*=180./PI;

	visibility.theta[i]=dist;
	visibility.visible[i]=0; % false
	if (dist>=thmin && dist <=thmax) {
	    visibility.visible[i]=1; % true
	}
    }

    if (qualifier_exists("numerical")) {
	return visibility;
    }

    %
    % convert visibility list into something easy to understand
    %
    variable vislist={};

    variable vis=0;
    variable visint=struct{MJDstart,MJDstop,MJDstring};
    _for i(0,ilen-1,1) {
	% new visibility interval?
	if (vis==0 && visibility.visible[i]==1) {
	    visint.MJDstart=visibility.mjd[i];
	    vis=1;
	}
	% end of current visibility interval
	% NB: i is guaranteed to be >1
	if (vis==1 && visibility.visible[i]==0) {
	    visint.MJDstop=visibility.mjd[i-1];
	    visint.MJDstring=strftime_MJD(visint.MJDstart)+" -- "+
	    strftime_MJD(visint.MJDstop);
	    list_append(vislist,@visint);
	    vis=0;
	}
    }
    % end of loop, is the interval still open?
    if (vis==1) {
	visint.MJDstop=visibility.mjd[ilen-1];
	visint.MJDstring=strftime_MJD(visint.MJDstart)+" -- "+
 	                 strftime_MJD(visint.MJDstop);
	list_append(vislist,visint);
    }
    if (length(vislist)==0) {
	visibility=[0:-1];
    } else {
	visibility=list_to_array(vislist);
    }

    return(visibility);
}

%
% Code from src/slang/astronomy/obliquity.sl
%
define obliquity ()
%!%+
%\function{obliquity}
%\synopsis{calculate the obliquity of the ecliptic}
%\usage{eps=obliquity(JD;mjd);}
%\qualifiers{
%  \qualifier{mjd}{if set, the argument is in MJD, not in JD}
%  \qualifier{laskar}{use the polynomial expression by Laskar (1986,
%            Astron. Astrophys. 157, 59), which is valid for a time
%            span of around 10000 years around J2000, rather than
%            the IAU 2006 expression.}
%  \qualifier{true}{return the true obliquity for the date by applying
%            nutation}
%  \qualifier{theory}{nutation theory to use, see help for
%            function nutation_angles; default: NU2000K}
%  \qualifier{deg}{return the obliquity in degrees }
%  \qualifier{arcsec}{return the obliquity in arcseconds}
%}
%\description
% This routine calculates the obliquity of the ecliptic for
% the Julian Date(s) JD using the IAU 2006 resolutions, which
% are based on Hilton et al., (2006, Cel. Mech. Dyn. Astron. 94, 351).
% JD can be an array, it is formally measured in TT.
%
% The IAU 2006 expression is good to better than 1 arcsecond
% for +/-500 years around J2000. Outside of this interval
% use the polynomial expression by Laskar (laskar keyword).
%
% The angles returned are in rad, use the deg or arcsec qualifiers
% if you want them in degrees or arc seconds.
%
%\seealso{nutation_angles}
%!%-
{
    variable JD,T;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {help(_function_name()); return; }

    if (qualifier_exists("mjd")) {
	JD=mjd2jd(JD);
    }

    if (qualifier_exists("deg") && qualifier_exists("arcsec")) {
	throw UsageError, sprintf("Usage error in %s: Only one of 'deg' and 'arcsec' qualifiers is allowed\n",_function_name());
    }

    % time in centuries since J2000.0
    T=(JD-2451545.0)/36525.;
    
    variable eps;
    
    if(qualifier_exists("laskar")) {
	% Laskar (1986, Table 8; T is per 10000 yr!)
	T=T/100.;
	eps=(((((((((2.45*T+5.79)*T+27.87)*T+7.12)*T-39.05)*T-249.67)*T-51.38)*T+1999.25)*T-1.55)*T-4680.93)*T;
    } else {
	% IAU 2006 equations
	% (T is per century!)
	eps=((((-4.34E-8*T-5.76E-7)*T+0.00200340)*T-0.0001831*T)-46.836769)*T;
    }
    eps+=84381.406;

    if (qualifier_exists("true")) {
	variable dpsi,deps;
	(dpsi,deps)=nutation_angles(JD;theory=qualifier("theory","NU2000K"),arcsec);
	eps+=deps;
    }
    
    if (qualifier_exists("arcsec")) {
	return eps;
    }

    if(qualifier_exists("deg")) {
	return eps/3600.;
    }

    return eps*PI/648000.; % arcsec -> rad
}

%
% Code from src/plot/ohplot_filled.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define ohplot_filled(lo,hi,val)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ohplot_filled}
%\synopsis{over-plot a filled histogram defined by slang arrays}
%\usage{ohplot_filled(Array_Type bin_lo, Array_Type bin_hi, Array_Type values)
%}
%\qualifiers{
%\qualifier{fill_style}{[\code{=1}] set the fill style:\n
%                       \code{FS = 1} \code{=>} solid (default)\n
%                       \code{FS = 2} \code{=>} outline\n
%                       \code{FS = 3} \code{=>} hatched (cannot be used with ylog;)\n
%                       \code{FS = 4} \code{=>} cross-hatched (cannot be used with ylog;)\n}
%\qualifier{ymin}{[\code{=min(values)}] set the lower y-value to which the areas are filled}
%\qualifier{angle}{[\code{=degrees}] sets the angle of the hatched lines for FS=3 [default = 45] }
%}
%\description
%    This function overplots a histogram described by three 1-D S-Lang arrays of size N.
%    The area below the histogram is filled. The fill style can be selected with a qualifier.
%\examples
%    \code{ohplot_filled([1:5],[2:6],[1:5]);}\n
%\seealso{hplot_filled, hplot}
%!%-
{
  ohplot(lo,hi,val);
  variable fs = qualifier("fill_style",1);

  %get hatching style and save it in order to reset it in the end
  variable hangle, hsep, hphase ;
  (hangle, hsep, hphase) = _pgqhs ;
  
  _pgsfs(fs);  
  if (fs == 3) { _pgshs(qualifier("angle",45), 1, 0);}
  
  variable i;
  if (get_plot_options.logx)
  {
    lo = log10(lo);
    hi = log10(hi);
  }
  if (get_plot_options.logy) val = log10(val);
  variable y_min = qualifier("ymin",max([min(val),get_plot_options.ymin ]));
  if (get_plot_options.logy) y_min = log10(y_min);
  if (get_plot_options.logy && (fs > 2)) 
  {
    printf("warning: hatched fill_style not allowed with ylog;\nfill_style was set to 1\n");
    _pgsfs(1);
  }
  _for i (0, length(lo)-1, 1) 
  {
    ()=_pgpoly(4,[lo[i],lo[i],hi[i],hi[i]],[y_min,val[i],val[i],y_min]);
  }

  %reset hatching style to default values:
  _pgshs(hangle, hsep, hphase);
}

%
% Code from src/plot/ohplot_with_err.sl
%

define ohplot_with_err()
%!%+
%\function{ohplot_with_err}
%\synopsis{overplots histogram data points with errorbars}
%\description
%    This function passes all its arguments and qualifiers to the
%    \code{hplot_with_err} function, but adds the \code{overplot} qualifier.
%\seealso{[o][h]plot_with_err, [o][h]plot}
%!%-
{
  if(_NARGS)
  { variable args = __pop_list(_NARGS);
    hplot_with_err(__push_list(args);; struct_combine(__qualifiers(), "overplot"));
  }
  else
    help(_function_name());
}

%
% Code from src/plot/opdc.sl
%
alias("oplot_data_counts", "opdc");

%
% Code from src/plot/opdf.sl
%
define opdf()
{
  variable i, ids;
  switch(_NARGS)
  { case 0: ids = all_data; }
  { case 1: ids = (); ids = [ids]; }

  foreach i (ids) { oplot_data_flux(i); }
}

%
% Code from src/plot/open_plot_aa1col.sl
%
%%%%%%%%%%%%%%%%
define open_plot_aa1col(name)
%%%%%%%%%%%%%%%%
{
   setPGPLOTenv(20,20);
   ()=open_plot(name);
   charsize(1.5);
   _pgscf(2);
}

%
% Code from src/plot/open_plot_aa2col.sl
%
define open_plot_aa2col(name)
{
   setPGPLOTenv(18,20);
   ()=open_plot(name);
   charsize(0.85);
   _pgscf(2);
}

%
% Code from src/plot/open_plot_ps2eps.sl
%
private variable open_plot_ps2eps_cmd = "";

define open_plot_ps2eps()
%!%+
%\function{open_plot_ps2eps}
%\synopsis{opens a plot and saves .ps filename for latter use of ps2eps}
%\usage{Integer_Type id = open_plot_ps2eps(device[, nxpanes[, nypanes]]);}
%\description
%    \code{open_plot_ps2eps} passes its arguments to \code{open_plot}.
%    If used together with \code{close_plot_ps2eps}, an .eps file
%    is finally produced from an usual PGPLOT .ps output
%    through the external tool ps2eps. Therefore, \code{device} has to be
%    a \code{.ps} file with \code{/}[\code{v}][\code{c}]\code{ps} specification.
%\qualifiers{
%\qualifier{ps2epsopt}{ [="-R=+ -B -f"]: option for ps2eps}
%\qualifier{noremoveps}{the .ps file will not be removed after conversion}
%\qualifier{pre_enlargeBB}{enlarge bounding box by specified value [=1, if none specified] before ps2eps}
%\qualifier{pre_run_cmd=cmd}{command to run before ps2eps.
%                       The ps-file is passed to \code{cmd} as an argument.
%                       \code{cmd} is expected to write the modified ps file to \code{stdout}.}
%\qualifier{enlargeBB}{enlarge bounding box by specified value [=1, if none specified] after ps2eps}
%}
%\seealso{open_plot, close_plot_ps2eps}
%!%-
{
  variable device, nxpanes=1, nypanes=1;
  switch(_NARGS)
  { case 1: device = (); }
  { case 2: (device, nxpanes) = (); }
  { case 3: (device, nxpanes, nypanes) = (); }
  { help(_function_name()); return; }

  variable filename = string_matches(device, `\(.*\)\.ps/`, 1);
  if(filename==NULL)
    vmessage(`warning (%s): device "%s" does not contain a .ps file`, _function_name(), device);
  else
  {
    filename = filename[1];
    open_plot_ps2eps_cmd = "";
    variable delta;
    if(qualifier_exists("pre_enlargeBB"))
    {
      delta = string(qualifier("pre_enlargeBB"));
      if(delta=="NULL")  delta = "1";
      open_plot_ps2eps_cmd = "/usr/bin/awk '"+
	                     + `$1=="%%BoundingBox:" && $2!="(atend)" { $2-=`+delta+"; $3-="+delta+"; $4+="+delta+"; $5+="+delta+"; }"
                             + "{ print; }"
	                     +"' $filename.ps > $filename.ps.tmp; mv $filename.ps.tmp $filename.ps; "$;
    }

    variable pre_run_cmd = qualifier("pre_run_cmd");
    if(pre_run_cmd!=NULL)
      open_plot_ps2eps_cmd += "$pre_run_cmd $filename.ps > $filename.ps.tmp; mv $filename.ps.tmp $filename.ps; "$;

    open_plot_ps2eps_cmd += "ps2eps "+qualifier("ps2epsopt", "-R=+ -B -f")+" $filename.ps"$;

    ifnot(qualifier_exists("noremoveps"))
      open_plot_ps2eps_cmd += "; /bin/rm $filename.ps"$;

    if(qualifier_exists("enlargeBB"))
    {
      delta = string(qualifier("enlargeBB"));
      if(delta=="NULL")  delta = "1";
      open_plot_ps2eps_cmd += "; /usr/bin/awk '"+
	                     +  `$1=="%%BoundingBox:" && $2!="(atend)" { $2-=`+delta+"; $3-="+delta+"; $4+="+delta+"; $5+="+delta+"; }"
                             +  "{ print; }"
	                     +"' $filename.eps > $filename.eps.tmp; mv $filename.eps.tmp $filename.eps"$;
    }
  }
  return open_plot(device, nxpanes, nypanes);
}


define close_plot_ps2eps()
%!%+
%\function{close_plot_ps2eps}
%\synopsis{closes a plot calls ps2eps}
%\usage{close_plot_ps2eps();}
%\seealso{close_plot, close_plot_ps2eps}
%!%-
{
  close_plot();
  if(open_plot_ps2eps_cmd != "")
  {
    if(qualifier_exists("verbose"))  message(open_plot_ps2eps_cmd);
    ()=system(open_plot_ps2eps_cmd);
    open_plot_ps2eps_cmd = "";
  }
}

%
% Code from src/plot/oplot_contour_lines.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
define oplot_contour_lines()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{oplot_contour_lines}
%\synopsis{overplots contour lines of a 2d array (image)}
%\usage{oplot_contour_lines(f, f0[, X, Y]);}
%\qualifiers{
%\qualifier{save}{=filename: saves/restores the contour lines in a FITS file}
%\qualifier{pgplot}{uses \code{_pgline} instead of \code{oplot}}
%}
%\description
%    \code{f} has to be a two-dimensional array (an image).
%    \code{f0} is the value of the contour lines \code{f[y,x] = f0}.
%    The arrays \code{X} and \code{Y}, if present, transform the array-indices \code{x} and \code{y}
%    to the coordinate system used for plotting.
%\seealso{get_contour_lines}
%!%-
{
  variable f, f0, X=NULL, Y=NULL;
  switch(_NARGS)
  { case 2: (f, f0) = (); }
  { case 4: (f, f0, X, Y) = (); }
  { help(_function_name()); return; }

  if(X==NULL)
  { X = [0:length(f[0,*])-1];
    Y = [0:length(f[*,0])-1];
  }

  variable filename = qualifier("save");
  variable pgplot = qualifier_exists("pgplot");

  variable col = get_plot_options.color;
  variable cline, data, F;
  if(filename==NULL || stat_file(filename)==NULL)
  {
    if(filename!=NULL)  F = fits_open_file(filename, "c");
    foreach cline (get_contour_lines(f, f0))
    {
      if(pgplot)
        ()=_pgline(length(cline.x), X[cline.x], Y[cline.y]);
      else
      { color(col); oplot(X[cline.x], Y[cline.y]); }

      if(filename!=NULL)
        fits_write_binary_table(F, "contour-line", struct { x=X[cline.x], y=Y[cline.y] });
    }
  }
  else
  {
    F = fits_open_file(filename, "r");
    while(_fits_movrel_hdu(F, 1)==0)
    { data = fits_read_table(F);
      if(pgplot)
        ()=_pgline(length(data.x), data.x, data.y);
      else
      { color(col); oplot(data.x, data.y); }
    }
  }
  if(filename!=NULL)  fits_close_file(F);
}

%
% Code from src/plot/oplot_rect.sl
%
define oplot_rect()
%!%+
%\function{oplot_rect}
%\synopsis{overplots a rectangle}
%\usage{oplot_rect(Double_Type x1, y1, x2, y2);
%\altusage{oplot_rect(Double_Type [x1, x2], [y1, y2]);}
%}
%!%-
{
  variable x1, y1, x2, y2;
  switch(_NARGS)
  { case 2:
      variable x, y; (x, y) = ();
      x1=x[0]; y1=y[0]; x2=x[1]; y2=y[1];
  }
  { case 4: (x1, y1, x2, y2) = (); }
  { help(_function_name()); return; }

  variable plot_opt = get_plot_options();
  connect_points(-1);
  oplot([x1,x2,x2,x1,x1], [y1,y1,y2,y2,y1]);
  connect_points(plot_opt.connect_points);
}

%
% Code from src/plot/oplot_struct_arrays.sl
%
define oplot_struct_arrays()
%!%+
%\function{oplot_struct_arrays}
%\synopsis{overplots the fields of a structure agains each other}
%\usage{oplot_struct_arrays(Struct_Type s, String_Type fieldnameX, String_Type fieldnameY);}
%\qualifiers{
%\qualifier{xoffset}{}
%\qualifier{yoffset}{}
%}
%\description
%    \code{oplot(s.fieldnameX - xoffset,  s.fieldnameY - yoffset);}
%!%-
{
  variable s, fieldnameX, fieldnameY;
  switch(_NARGS)
  { case 3: (s, fieldnameX, fieldnameY)= (); }
  { help(_function_name()); return; }

  oplot(get_struct_field(s, fieldnameX)-qualifier("xoffset", 0),
        get_struct_field(s, fieldnameY)-qualifier("yoffset", 0));
}

%
% Code from src/plot/oplot_with_err.sl
%
%%%%%%%%%%%%%%%%%%%%%
define oplot_with_err()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{oplot_with_err}
%\synopsis{overplots data points with errorbars}
%\description
%    This function passes all its arguments and qualifiers to the
%    \code{plot_with_err} function, but adds the \code{overplot} qualifier.
%\seealso{[o][h]plot_with_err, [o][h]plot}
%!%-
{
  if(_NARGS)
  { variable args = __pop_list(_NARGS);
    plot_with_err(__push_list(args);; struct_combine(__qualifiers(), "overplot"));
  }
  else
    help(_function_name());
}

%
% Code from src/plot/oplot_xline.sl
%
define oplot_xline()
%!%+
%\function{oplot_xline}
%\synopsis{overplots one or more vertical line(s)}
%\usage{oplot_xline(Double_Type x1[, x2, ...]);}
%\qualifiers{
%\qualifier{ymin}{[=<minimum of current \code{yrange}>]}
%\qualifier{ymax}{[=<maximum of current \code{yrange}>]}
%\qualifier{color}{number of the color to use}
%}
%\description
%    For each \code{x} of \code{x1}[, \code{x2}, ...] (which may be arrays, too),
%    a line is overplotted from \code{ymin} to \code{ymax} with the same \code{color}.
%!%-
{
  if(_NARGS<1) { help(_function_name()); return; }

  variable args = __pop_list(_NARGS);
  variable plot_opt = get_plot_options();
  variable col = qualifier("color", plot_opt.color);
  variable ymin = qualifier("ymin", plot_opt.ymin);
  variable ymax = qualifier("ymax", plot_opt.ymax);
  connect_points(-1);
  variable x;
  loop(_NARGS)
    foreach x ([list_pop(args)])
    { color(col);
      oplot([x, x], [ymin, ymax]);
    }
  connect_points(plot_opt.connect_points);
}

%
% Code from src/plot/oplot_yline.sl
%
%%%%%%%%%%%%%%%%%%
define oplot_yline()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{oplot_yline}
%\synopsis{overplots one or more horizontal line(s)}
%\usage{oplot_yline(Double_Type y1[, y2, ...]);}
%\qualifiers{
%\qualifier{xmin}{[=<minimum of current \code{xrange}>]}
%\qualifier{xmax}{[=<maximum of current \code{xrange}>]}
%\qualifier{color}{number of the color to use}
%}
%\description
%    For each \code{y} of \code{y1}[, \code{y2}, ...] (which may be arrays, too),
%    a line is overplotted from \code{xmin} to \code{xmax} with the same \code{color}.
%!%-
{
  if(_NARGS<1) { help(_function_name()); return; }

  variable args = __pop_list(_NARGS);
  variable plot_opt = get_plot_options();
  variable col = qualifier("color", plot_opt.color);
  variable xmin = qualifier("xmin", plot_opt.xmin);
  variable xmax = qualifier("xmax", plot_opt.xmax);
  connect_points(-1);
  variable y;
  loop(_NARGS)
    foreach y ([list_pop(args)])
    { color(col);
      oplot([xmin, xmax], [y, y]);
    }
  set_plot_options(plot_opt);
}

%
% Code from src/plot/opmc.sl
%
define opmc()
{
  variable id, ids;
  switch(_NARGS)
  { case 0: ids = all_data; }
  { case 1: ids = (); ids = [ids]; }
  { message("usage: opmc([ids]);"); return; }

  foreach id (ids) { oplot_model_counts(id); }
}

%
% Code from src/plot/opmf.sl
%
define opmf()
{
  variable id, ids;
  switch(_NARGS)
  { case 0: ids = all_data; }
  { case 1: ids = (); ids = [ids]; }
  { message("usage: opmf([ids]);"); return; }

  foreach id (ids) { oplot_model_flux(id); }
}

%
% Code from src/data/binning/optimal_binning.sl
%
%%% functions for model and data binning following certain criteria
%%% 
%%%

%%% optimal binning after Kaastra & Bleeker 2016

%%%
private define binWidthKaastraBleeker (numCounts, numResolver)
%%%
%%+
%\function{binWidthKaastraBleeker}
%\synopsis{Caluculate optimized bin width per data resolution}
%\usage{Double_Type[] binWidthKaastraBleeker (Double_Type[] numCounts,
%                                             Double Type numResolver)}
%\description
%    Kaaster & Bleeker (2016) describe an aproximation to data binning to
%    minimize the number of bins by keeping the resolving power for near
%    resolution limit features.
%    
%    The approximation is based on a Monte Carlo approach to band limited
%    data. The result gives the optimized bin width per FWHM for each data
%    channel. It depends on the number of counts for each channel in the
%    FWHM of the instrument <numCounts> and the number of resolution elemets
%    <numResolver> (roughly sum 1/c_i where c_i is the FWHM in channels for
%    channel i)
%\seealso{}
%%-
{
  variable r = 1 + .2*log(numResolver);
  variable t = log(numCounts*r);
  variable d = 1.;

  if (t>2.110)
    d = (0.08 + 7./t + 1.8/(t^2))/(1+5.9/t);

  return d;
}
%%%

%%%
private define countsInResolver (c1, c2, response, counts)
%%%
%%+
%\function{countsInResolver}
%\synopsis{Calculate approximated number of counts per FWHM}
%\usage{Double_Type[] countsInResolver (Double_Type[] c1,
%                                       Double_Type[] c2,
%                                       Double_Type[][] response,
%                                       Double_Type[] counts)}
%\description
%
%\seealso{}
%%-
{
  variable i1, i2;
  variable size;
  variable c, h;

  i1 = int(c1);
  i2 = int(c2);
  size = length(response);
  c = sum(counts[[i1:i2]]);
  h = sum(response)/sum(response[[(i1-((i1 == 0)?0:1)):(i2+((i2==size-1)?0:1))]]);

  return c*h;
}

%%%
private define buildOptimalBins (id, minCounts)
%%%
{
  variable cResponse;
  variable channel;
  variable c1,c2;
  variable resWidth;
  variable numResolver;
  variable counts, resolverCnts;

  variable channelWidth;
  variable fakeId, grid, data, fit;
  variable to, size;
  variable rebin, flip;

  if (all(all_data() != id))
    throw UsageError, "Dataset $id not defined"$;

  (fakeId, grid, data) = responseEvalSetup(id);
  counts = get_data(fakeId).value;
  include(fakeId);
  fit = open_fit();

  size = length(grid.bin_lo);
  channel = 0;
  rebin = Int_Type[size];
  flip = -1;

  while (channel < size) {
    cResponse = responseEval(fakeId, grid, channel, fit);
    (c1, c2) = hist_fwhm_index(cResponse, channel; interp);
    resWidth = c2-c1;
    numResolver = sum(1./resWidth);
    resolverCnts = countsInResolver(c1, c2, cResponse, counts);
    channelWidth = int(binWidthKaastraBleeker(resolverCnts, numResolver)*resWidth);
    if (channelWidth < 1) channelWidth = 1;

    to = min([channel+channelWidth-1, size-1]);
    while (sum(counts[[channel:to]])<minCounts && to<(size-1))
      to++;

    rebin[[channel:to]] = flip;
    flip *= -1;

    channel = to+1;
  }

  % remove temporary values
  fit.close();
  responseEvalCleanup(fakeId, data);

  return rebin;
}

%%%
define rebin_dataset_optimal (histIdx)
%%%
%!%+
%\function{rebin_dataset_optimal}
%\usage{rebin_dataset_optimal (Int_Type index);}
%\altusage{rebin_dataset_optimal (Int_Type[] index);}
%\synopsis{Rebin datasets addressed by <index> based on Kaastra & Bleeker 2016}
%\qualifiers{
%\qualifier{min_counts}{[=0]: Ensure a minimum number of counts per bin (alters statistical properties).}
%}
%\description
%  Rebin a dataset to a numerical optimum defined by the associated
%  response matrix. The algorithm tries to balance the maximum
%  possible information with the smallest possible grid to represent
%  it.
%\seealso{rebin_data}
%!%-
{
  variable idx;
  variable mask;

  variable min_counts = qualifier("min_counts", 0);

  foreach idx (histIdx) {
    mask = buildOptimalBins(idx, min_counts);
    rebin_data(idx, mask); % The binning mask, applied on a wavelength grid
  }
}

%%%
private define buildOptimalBinsGroup (id, minCounts)
%%%
{
  variable cResponse;
  variable channel;
  variable c1,c2;
  variable resWidth;
  variable numResolver;
  variable counts, resolverCnts;

  variable channelWidth;
  variable fakeId, grid, data, fit;
  variable to, size;
  variable rebin, flip;

  % for grouping we have to set all fake datasets
  variable dataIds, numData, fid, did, weights, i, countsId, errorValue = NULL;

  try (errorValue) { dataIds = combination_members(id); }
  catch TypeMismatchError;
  if (NULL != errorValue)
    throw UsageError, "Dataset group $id not defined"$;

  numData = length(dataIds);
  fakeId = Int_Type[numData];
  data = Struct_Type[numData];
  weights = Double_Type[numData];
  countsId = Array_Type[numData];

  % setup all data
  _for i (0, numData-1) {
    (fid, grid, did) = responseEvalSetup(dataIds[i]);
    weights[i] = get_data_info(fid).combo_weight;
    countsId[i] = get_data(fid).value;
    fakeId[i] = fid;
    data[i] = did;
    ifnot (__is_initialized(&counts))
      counts = Double_Type[length(grid.bin_lo)];
    counts += countsId[i]*weights[i];
  }

  include(fakeId);
  fit = open_fit();

  % all grids have to be the same
  size = length(grid.bin_lo);
  channel = 0;
  rebin = Int_Type[size];
  flip = -1;

  while (channel < size) {
    cResponse = Double_Type[size];
    _for i (0, numData-1) {
      cResponse += responseEval(fakeId[i], grid, channel, fit)*weights[i];
    }

    (c1, c2) = hist_fwhm_index(cResponse, channel; interp);
    resWidth = c2-c1;
    numResolver = sum(1./resWidth);
    resolverCnts = countsInResolver(c1, c2, cResponse, counts);
    channelWidth = int(binWidthKaastraBleeker(resolverCnts, numResolver)*resWidth);
    if (channelWidth < 1) channelWidth = 1;

    to = min([channel+channelWidth-1, size-1]);
    if (minCounts>0) {
      _for i (0, numData-1) {
	while (sum(countsId[i][[channel:to]])<minCounts && to<(size-1))
	  to++;
      }
    }

    rebin[[channel:to]] = flip;
    flip *= -1;

    channel = to+1;
  }

  % remove temporary values
  fit.close();

  _for i (0, numData-1)
    responseEvalCleanup(fakeId[i], data[i]);
  exclude(all_data());
  _for i (0, numData-1)
    include(data[i].includes);

  return rebin;
}

%%%
define rebin_combined_optimal (groupsIdx)
%%%
%!%+
%\function{rebin_combined_optimal}
%\usage{rebin_combined_optimal (Int_Type group_index);
%\altusage{rebin_combined_optimal Int_Type[] group_index);}}
%\synopsis{Rebin a dataset combination based on Kaastra & Bleeker 2016}
%\qualifiers{
%\qualifier{min_counts}{[=0]: Ensure a minimum number of counts per bin per dataset.}
%}
%\description
%  Rebin a dataset combination using the numerical optimum defined by the
%  response. From the combination a combined response is calculated respecting
%  the weights.
%\seealso{combine_datasets, rebin_combined}
%!%-
{
  variable gid;
  variable mask;

  variable min_counts = qualifier("min_counts", 0);

  foreach gid (groupsIdx) {
    mask = buildOptimalBinsGroup(gid, min_counts);
    rebin_combined(gid, mask);
  }
}

%
% Code from src/misc/astronomy/orbit_calculator.sl
%
define AS() %{{{
%!%+
%\function{AS}
%\synopsis{Evaluate equations of motion, total energy, or circular velocity derived from a revised
%    Allen & Santillan potential}
%\usage{AS(Double_Types t, m[6,n]; qualifiers)}
%\qualifiers{
%\qualifier{coords}{[\code{="cyl"}] Use cylindrical ("cyl") or cartesian ("cart") coordinates.}
%\qualifier{eomecd}{[\code{="eom"}] Return equations of motion ("eom"), total energy ("energy"),
%      circular velocity ("circ"), or Sun-Galactic center distance ("sgcd").}
%\qualifier{Mb}{[\code{=409}] Mass of bulge in Galactic mass units, see Irrgang et al. 2013.}
%\qualifier{Md}{[\code{=2856}] Mass of disc in Galactic mass units, see Irrgang et al. 2013.}
%\qualifier{Mh}{[\code{=1018}] Mass scale factor of halo in Galactic mass units,
%               see Irrgang et al. 2013.}
%\qualifier{bb}{[\code{=0.23}] Bulge scale length, see Irrgang et al. 2013.}
%\qualifier{ad}{[\code{=4.22}] Disc scale length, see Irrgang et al. 2013.}
%\qualifier{bd}{[\code{=0.292}] Disc scale length, see Irrgang et al. 2013.}
%\qualifier{ah}{[\code{=2.562}] Halo scale length, see Irrgang et al. 2013.}
%\qualifier{exponent}{[\code{=2}] Exponent in the halo mass distribution, see Irrgang et al. 2013.}
%\qualifier{cutoff}{[\code{=200}] Halo cutoff, see Irrgang et al. 2013.}
%}
%\description
%    Evaluate the equations of motion, the total energy, or the circular velocity at time 't'
%    derived from the revised Galactic gravitational potential by Allen & Santillan (see Model I
%    in Irrgang et al., 2013, A&A, 549, A137) using either cylindrical coordinates (r [kpc],
%    phi [rad], z [kpc]) and their canonical momenta vr [kpc/Myr], Lz [kpc^2/Myr], vz [kpc/Myr])
%    or cartesian coordinates (x [kpc], y [kpc], z [kpc], vx [kpc/Myr], vy [kpc/Myr], vz [kpc/Myr]),
%    see qualifier 'coords'.  Conservation of angular momentum Lz is implemented in the equations
%    of motion for cylindrical coordinates only. The total energy E_total [kpc^2/Myr^2] is not
%    used to integrate the equations of motion although being a conserved quantity, too. Therefore,
%    conservation of energy, i.e., of E_total, is a measure for the precision of the numerical
%    methods applied.
%
%    For computing orbits with n different initial conditions, the input parameter m is
%    a [6,n]-matrix with (qualifier("coords")=="cyl")   or (qualifier("coords")=="cart")
%       m[0,*] = r;                                        m[0,*] = x;
%       m[1,*] = phi;                                      m[1,*] = y;
%       m[2,*] = z;                                        m[2,*] = z;
%       m[3,*] = vr;                                       m[3,*] = vx;
%       m[4,*] = Lz;                                       m[4,*] = vy;
%       m[5,*] = vz;                                       m[5,*] = vz;
%    If the qualifier 'eomecd' is set to "eom", the function returns a [6,n]-matrix delta with
%       delta[0,*] = vr;                                   delta[0,*] = vx;
%       delta[1,*] = Lz/r^2; % = vphi                      delta[1,*] = vy;
%       delta[2,*] = vz;                                   delta[2,*] = vz;
%       delta[3,*] = -d/dr (Potential(r,z) + Lz^2/r^2);    delta[3,*] = -d/dx Potential(x,y,z);
%       delta[4,*] = 0; % -d/dphi Potential(r,z)           delta[4,*] = -d/dy Potential(x,y,z);
%       delta[5,*] = -d/dz Potential(r,z);                 delta[5,*] = -d/dz Potential(x,y,z);
%    If the qualifier 'eomecd' is set to "energy", the function returns a [1,n]-array storing
%    the total energy for each orbit:
%       E_total(r,z,vr,vz,Lz) = Double_Type[n] = 0.5*(vr^2+vz^2+Lz^2/r^2) + Potential(r,z)
%    or
%       E_total(x,y,z,vx,vy,vz) = Double_Type[n] = 0.5*(vx^2+vy^2+vz^2) + Potential(x,y,z)
%    If the qualifier 'eomecd' is set to "circ", the function returns a [1,n]-array storing
%    the circular velocity for each orbit:
%       v_circ(r,z) = Double_Type[n] = sqrt( r * d/dr Potential(r,z) )
%    If the qualifier 'eomecd' is set to "sgcd", the function returns the Sun-Galactic center
%    distance found to fit best to this potential.
%\example
%    delta = AS(0, m);
%    energy = AS(0, m; eomecd="energy");
%    v_circ = AS(0, m; eomecd="circ");
%    sgcd = AS(; eomecd="sgcd");
%\seealso{orbit_calculator, MN_NFW, MN_TF, plummer_MW}
%!%-
{
  % equations of motion, energy, circular velocity or Sun-Galactic center distance:
  variable eomecd = qualifier("eomecd", "eom");
  if(eomecd == "sgcd")
  {
    return 8.4;
  }
  else if(_NARGS!=2)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable coords = qualifier("coords", "cyl");
    variable t, input;
    (t, input) = ();
    input *= 1.;
    variable r, z, vz;
    if(coords=="cart")
    {
      variable x, y, vx, vy;
      x = input[0,*];
      y = input[1,*];
      z = input[2,*];
      vx = input[3,*];
      vy = input[4,*];
      vz = input[5,*];
      r = sqrt(x^2+y^2);
    }
    else % coords=="cyl"
    {
      variable phi, vr, Lz;
      r = input[0,*];
      phi = input[1,*];
      z = input[2,*];
      vr = input[3,*];
      Lz = input[4,*];
      vz = input[5,*];
    }
    % definition of gravitational potential parameters:
    variable Mb = qualifier("Mb", 409.);  % mass of the bulge
    variable Md = qualifier("Md", 2856.); % mass of the disk
    variable Mh = qualifier("Mh", 1018.); % mass scale length of the halo
    variable bb = qualifier("bb", 0.23);  % scale length bulge
    variable ad = qualifier("ad", 4.22);  % scale length disk
    variable bd = qualifier("bd", 0.292); % scale length disk
    variable ah = qualifier("ah", 2.562); % scale length halo
    variable exponent = qualifier("exponent", 2.); % exponent in halo mass distribution
    variable cutoff = qualifier("cutoff", 200.); % halo cutoff

    % variable kpcmyr_to_kms = 977.7736364875057; % = conversion factor from kpc/myr to km/s = 3.0856775975*10^16 / (10^6 * 3.15582 * 10^7);
    variable constant = 1.0459799346702096e-04; % 100./kpcmyr_to_kms^2; factor 100 because potential is given in 100km^2/s^2
    % store frequently appearing expressions to save runtime:
    variable r2 = r^2;
    variable z2 = z^2;
    variable r2z2 = r2+z2;
    variable sr2z2 = sqrt(r2z2);
    variable sz2bd2 = sqrt(z2+bd^2);
    variable cutoff_ah = cutoff/ah;
    variable exp_minus_one = exponent-1.;
    variable ind1, ind11, ind12, ind2, ind3, ind4;
    ind1 = where( sr2z2 < cutoff, &ind2 );
    variable eps = 1e-5;
    ind11 = where( (sr2z2[ind1]) > eps, &ind12 ); % account for expressions with 1/sr2z2; e.g., right-hand side of equations of motion at origin are set to zero to solve singularity
    if(eomecd == "eom")
    {
      % store frequently appearing expressions to save runtime:
      variable dPot_bulge = -Mb/(r2z2+bb^2)^1.5;
      variable dPot_disc = -Md/(r2+(ad+sz2bd2)^2)^1.5;
      variable dPot_halo_inner = -Mh*(sr2z2/ah)^exp_minus_one/(r2z2)/(1.+(sr2z2/ah)^exp_minus_one)/ah;
      variable dPot_halo_outer = -Mh*(cutoff_ah)^exponent/(r2z2)^1.5/(1.+(cutoff_ah)^exp_minus_one);

      % calculate derivatives of the potential with respect to z (dPotdz):
      variable dPotdz = constant*( dPot_bulge*z + dPot_disc*(ad+sz2bd2)/sz2bd2*z );
      dPotdz[ind1[ind11]] += constant*dPot_halo_inner[ind1[ind11]]*z[ind1[ind11]];
      dPotdz[ind2] += constant*dPot_halo_outer[ind2]*z[ind2];

      variable ret = Double_Type[6,length(r)];

      if(coords=="cart")
      {
	% calculate derivatives of the potential with respect to x and y (dPotdx, dPotdy):
	variable dPotdx = constant*(dPot_bulge + dPot_disc)*x;
	variable dPotdy = constant*(dPot_bulge + dPot_disc)*y;
	dPotdx[ind1[ind11]] += constant*dPot_halo_inner[ind1[ind11]]*x[ind1[ind11]];
	dPotdy[ind1[ind11]] += constant*dPot_halo_inner[ind1[ind11]]*y[ind1[ind11]];
	dPotdx[ind2] += constant*dPot_halo_outer[ind2]*x[ind2];
	dPotdy[ind2] += constant*dPot_halo_outer[ind2]*y[ind2];

	ret[0,*] = vx;
	ret[1,*] = vy;
	ret[2,*] = vz;
	ret[3,*] = dPotdx;
	ret[4,*] = dPotdy;
	ret[5,*] = dPotdz;
	return ret;
      }
      else % coords=="cyl"
      {
	ind3 = where( Lz!=0, &ind4 );
	% calculate derivatives of the potential with respect to r (dPotdr):
	variable dPotdr = constant*(dPot_bulge + dPot_disc)*r;
	dPotdr[ind1[ind11]] += constant*dPot_halo_inner[ind1[ind11]]*r[ind1[ind11]];
	dPotdr[ind2] += constant*dPot_halo_outer[ind2]*r[ind2];
	dPotdr[ind3] += (Lz[ind3])^2/(r[ind3])^3;

	ret[0,*] = vr;
	ret[1,ind3] = (Lz[ind3])/(r2[ind3]);
	ret[1,ind4] = 0;
	ret[2,*] = vz;
	ret[3,*] = dPotdr;
	ret[4,*] = 0*Lz;
	ret[5,*] = dPotdz;
	return ret;
      }
    }
    else if(eomecd == "energy")
    {
      % calculate the kinetic and potential energies (Ekin, Epot):
      variable Ekin;
      if(coords=="cart")
      {
	Ekin = 0.5*(vx^2+vy^2+vz^2);
      }
      else % coords=="cyl"
      {
	Ekin = 0.5*(vr^2+vz^2);
	ind3 = where( Lz!=0 );
	Ekin[ind3] += 0.5*(Lz[ind3])^2/(r2[ind3]);
      }
      variable Epot = -Mb/sqrt(r2z2+bb^2)-Md/sqrt(r2+(ad+sz2bd2)^2);                                                 % phi_bulge + phi_disc
      Epot[ind1[ind11]] += Mh*log((1.+(sr2z2[ind1[ind11]]/ah)^(exp_minus_one))/(1.+cutoff_ah^(exp_minus_one)))/(exp_minus_one)/ah
                         - Mh*cutoff_ah^(exp_minus_one)/(1.+cutoff_ah^(exp_minus_one))/ah;                           % phi_bulge + phi_disc + phi_halo
      Epot[ind1[ind12]] += Mh*log((1.+(eps/ah)^(exp_minus_one))/(1.+cutoff_ah^(exp_minus_one)))/(exp_minus_one)/ah
                         - Mh*cutoff_ah^(exp_minus_one)/(1.+cutoff_ah^(exp_minus_one))/ah;                           % phi_bulge + phi_disc + phi_halo
      Epot[ind2] += -Mh*cutoff_ah^exponent/sr2z2[ind2]/(1.+cutoff_ah^exp_minus_one);                                 % phi_bulge + phi_disc + phi_halo
      return Ekin + constant*Epot;
    }
    else if(eomecd == "circ")
    {
      dPotdr = Mb/(r2z2+bb^2)^1.5*r + Md/(r2+(ad+sz2bd2)^2)^1.5*r;
      dPotdr[ind1[ind11]] += Mh*(sr2z2[ind1[ind11]]/ah)^exp_minus_one/(r2z2[ind1[ind11]])/(1.+(sr2z2[ind1[ind11]]/ah)^exp_minus_one)/ah*r[ind1[ind11]];
      dPotdr[ind2] += Mh*(cutoff_ah)^exponent/(r2z2[ind2])^1.5/(1.+(cutoff_ah)^exp_minus_one)*r[ind2];
      return 10.*sqrt(r*dPotdr);
    }
  }
}%}}}

define MN_NFW() %{{{
%!%+
%\function{MN_NFW}
%\synopsis{Evaluate equations of motion, total energy, or circular velocity derived from a
%    potential with a Miyamoto & Nagai bulge and disk component and a Navarro, Frenk,
%    & White dark matter halo}
%\usage{MN_NFW(Double_Types t, m[6,n]; qualifiers)}
%\qualifiers{
%\qualifier{coords}{[\code{="cyl"}] Use cylindrical ("cyl") or cartesian ("cart") coordinates.}
%\qualifier{eomecd}{[\code{="eom"}] Return equations of motion ("eom"), total energy ("energy"),
%      circular velocity ("circ"), or Sun-Galactic center distance ("sgcd").}
%\qualifier{Mb}{[\code{=439}] Mass of bulge in Galactic mass units, see Irrgang et al. 2013.}
%\qualifier{Md}{[\code{=3096}] Mass of disc in Galactic mass units, see Irrgang et al. 2013.}
%\qualifier{Mh}{[\code{=142200}] Mass scale factor of halo in Galactic mass units, see Irrgang et al. 2013.}
%\qualifier{bb}{[\code{=0.236}] Bulge scale length, see Irrgang et al. 2013.}
%\qualifier{ad}{[\code{=3.262}] Disc scale length, see Irrgang et al. 2013.}
%\qualifier{bd}{[\code{=0.289}] Disc scale length, see Irrgang et al. 2013.}
%\qualifier{ah}{[\code{=45.02}] Halo scale length, see Irrgang et al. 2013.}
%}
%\description
%    Evaluate the equations of motion, the total energy, or the circular velocity at time 't'
%    derived from a potential with a Miyamoto & Nagai bulge and disk component and a Navarro,
%    Frenk, & White dark matter halo (see Model III in Irrgang et al., 2013, A&A, 549, A137)
%    using either cylindrical coordinates (r [kpc], phi [rad], z [kpc]) and their canonical
%    momenta vr [kpc/Myr], Lz [kpc^2/Myr], vz [kpc/Myr]) or cartesian coordinates (x [kpc],
%    y [kpc], z [kpc], vx [kpc/Myr], vy [kpc/Myr], vz [kpc/Myr]), see qualifier 'coords'.
%    Conservation of angular momentum Lz is implemented in the equations of motion for
%    cylindrical coordinates only. The total energy E_total [kpc^2/Myr^2] is not used to
%    integrate the equations of motion although being a conserved quantity, too. Therefore,
%    conservation of energy, i.e., of E_total, is a measure for the precision of the numerical
%    methods applied.
%
%    For computing orbits with n different initial conditions, the input parameter m is
%    a [6,n]-matrix with (qualifier("coords")=="cyl")   or (qualifier("coords")=="cart")
%       m[0,*] = r;                                        m[0,*] = x;
%       m[1,*] = phi;                                      m[1,*] = y;
%       m[2,*] = z;                                        m[2,*] = z;
%       m[3,*] = vr;                                       m[3,*] = vx;
%       m[4,*] = Lz;                                       m[4,*] = vy;
%       m[5,*] = vz;                                       m[5,*] = vz;
%    If the qualifier 'eomecd' is set to "eom", the function returns a [6,n]-matrix delta with
%       delta[0,*] = vr;                                   delta[0,*] = vx;
%       delta[1,*] = Lz/r^2; % = vphi                      delta[1,*] = vy;
%       delta[2,*] = vz;                                   delta[2,*] = vz;
%       delta[3,*] = -d/dr (Potential(r,z) + Lz^2/r^2);    delta[3,*] = -d/dx Potential(x,y,z);
%       delta[4,*] = 0; % -d/dphi Potential(r,z)           delta[4,*] = -d/dy Potential(x,y,z);
%       delta[5,*] = -d/dz Potential(r,z);                 delta[5,*] = -d/dz Potential(x,y,z);
%    If the qualifier 'eomecd' is set to "energy", the function returns a [1,n]-array storing
%    the total energy for each orbit:
%       E_total(r,z,vr,vz,Lz) = Double_Type[n] = 0.5*(vr^2+vz^2+Lz^2/r^2) + Potential(r,z)
%    or
%       E_total(x,y,z,vx,vy,vz) = Double_Type[n] = 0.5*(vx^2+vy^2+vz^2) + Potential(x,y,z)
%    If the qualifier 'eomecd' is set to "circ", the function returns a [1,n]-array storing
%    the circular velocity for each orbit:
%       v_circ(r,z) = Double_Type[n] = sqrt( r * d/dr Potential(r,z) )
%    If the qualifier 'eomecd' is set to "sgcd", the function returns the Sun-Galactic center
%    distance found to fit best to this potential.
%\example
%    delta = MN_NFW(0, m);
%    energy = MN_NFW(0, m; eomecd="energy");
%    v_circ = MN_NFW(0, m; eomecd="circ");
%    sgcd = MN_NFW(; eomecd="sgcd");
%\seealso{orbit_calculator, AS, MN_TF, plummer_MW}
%!%-
{
  % equations of motion, energy, circular velocity or Sun-Galactic center distance:
  variable eomecd = qualifier("eomecd", "eom");
  if(eomecd == "sgcd")
  {
    return 8.33;
  }
  else if(_NARGS!=2)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable coords = qualifier("coords", "cyl");
    variable t, input;
    (t, input) = ();
    input *= 1.;
    variable r, z, vz;
    if(coords=="cart")
    {
      variable x, y, vx, vy;
      x = input[0,*];
      y = input[1,*];
      z = input[2,*];
      vx = input[3,*];
      vy = input[4,*];
      vz = input[5,*];
      r = sqrt(x^2+y^2);
    }
    else % coords=="cyl"
    {
      variable phi, vr, Lz;
      r = input[0,*];
      phi = input[1,*];
      z = input[2,*];
      vr = input[3,*];
      Lz = input[4,*];
      vz = input[5,*];
    }
    % definition of gravitational potential parameters:
    variable Mb = qualifier("Mb", 439.);    % mass of the bulge
    variable Md = qualifier("Md", 3096.);   % mass of the disk
    variable Mh = qualifier("Mh", 142200.); % mass scale factor of the halo
    variable bb = qualifier("bb", 0.236);   % scale length bulge
    variable ad = qualifier("ad", 3.262);   % scale length disk
    variable bd = qualifier("bd", 0.289);   % scale length disk
    variable ah = qualifier("ah", 45.02);   % scale length halo

    % variable kpcmyr_to_kms = 977.7736364875057; % = conversion factor from kpc/myr to km/s = 3.0856775975*10^16 / (10^6 * 3.15582 * 10^7);
    variable constant = 1.0459799346702096e-04; % 100./kpcmyr_to_kms^2; factor 100 because potential is given in 100km^2/s^2
    % store frequently appearing expressions to save runtime:
    variable r2 = r^2;
    variable z2 = z^2;
    variable r2z2 = r2+z2;
    variable sr2z2 = sqrt(r2z2);
    variable sz2bd2 = sqrt(z2+bd^2);
    variable ind1, ind2, ind3, ind4;
    ind1 = where(sr2z2 < 1e-5, &ind2); % ind1 to avoid singularities when falling into the Galactic Center
    if(eomecd == "eom")
    {
      % store frequently appearing expressions to save runtime:
      variable dPot_bulge = -Mb/(r2z2+bb^2)^1.5;
      variable dPot_disc = -Md/(r2+(ad+sz2bd2)^2)^1.5;
      variable dPot_halo = Mh/r2z2/ah/(1.+sr2z2/ah)-Mh*log(1.+sr2z2/ah)/(r2z2)^1.5;

      % calculate derivatives of the potential with respect to z (dPotdz):
      variable dPotdz = constant*(dPot_bulge*z+dPot_disc*(ad+sz2bd2)/sz2bd2*z);
      dPotdz[ind2] += constant*dPot_halo[ind2]*z[ind2];

      variable ret = Double_Type[6,length(r)];

      if(coords=="cart")
      {
	% calculate derivatives of the potential with respect to x and y (dPotdx, dPotdy):
	variable dPotdx = constant*(dPot_bulge + dPot_disc)*x;
	variable dPotdy = constant*(dPot_bulge + dPot_disc)*y;
	dPotdx[ind2] += constant*dPot_halo[ind2]*x[ind2];
	dPotdy[ind2] += constant*dPot_halo[ind2]*y[ind2];

	ret[0,*] = vx;
	ret[1,*] = vy;
	ret[2,*] = vz;
	ret[3,*] = dPotdx;
	ret[4,*] = dPotdy;
	ret[5,*] = dPotdz;
	return ret;
      }
      else % coords=="cyl"
      {
	ind3 = where( Lz!=0, &ind4 );
	% calculate derivatives of the potential with respect to r (dPotdr):
	variable dPotdr = constant*(dPot_bulge+dPot_disc)*r;
	dPotdr[ind2] += constant*dPot_halo[ind2]*r[ind2];
	dPotdr[ind3] += (Lz[ind3])^2/(r[ind3])^3;

	ret[0,*] = vr;
	ret[1,ind3] = (Lz[ind3])/(r2[ind3]);
	ret[1,ind4] = 0;
	ret[2,*] = vz;
	ret[3,*] = dPotdr;
	ret[4,*] = 0*Lz;
	ret[5,*] = dPotdz;
	return ret;
      }
    }
    else if(eomecd == "energy")
    {
      % calculate the kinetic and potential energies (Ekin, Epot):
      variable Ekin;
      if(coords=="cart")
      {
	Ekin = 0.5*(vx^2+vy^2+vz^2);
      }
      else % coords=="cyl"
      {
	Ekin = 0.5*(vr^2+vz^2);
	ind3 = where( Lz!=0 );
	Ekin[ind3] += 0.5*(Lz[ind3])^2/(r2[ind3]);
      }
      variable Epot = -Mb/sqrt(r2z2+bb^2)-Md/sqrt(r2+(ad+sz2bd2)^2); % phi_bulge + phi_disc
      Epot[ind1] += -Mh/ah;                                          % phi_bulge + phi_disc + phi_halo
      Epot[ind2] += -Mh*log(1.+sr2z2[ind2]/ah)/sr2z2[ind2];          % phi_bulge + phi_disc + phi_halo
      return Ekin + constant*Epot;
    }
    else if(eomecd == "circ")
    {
      dPotdr = Mb/(r2z2+bb^2)^(1.5)*r+Md/(r2+(ad+sz2bd2)^2)^(1.5)*r;
      dPotdr[ind2] += -Mh/(r2z2[ind2])/ah*r[ind2]/(1.+sr2z2[ind2]/ah)+Mh*log(1.+sr2z2[ind2]/ah)/(r2z2[ind2])^(1.5)*r[ind2];
      return 10.*sqrt(r*dPotdr);
    }
  }
}%}}}

define MN_TF() %{{{
%!%+
%\function{MN_TF}
%\synopsis{Evaluate equations of motion, total energy, or circular velocity derived from a
%    potential with a Miyamoto & Nagai bulge and disk component and a truncated, flat
%    rotation curve halo model}
%\usage{MN_TF(Double_Types t, m[6,n]; qualifiers)}
%\qualifiers{
%\qualifier{coords}{[\code{="cyl"}] Use cylindrical ("cyl") or cartesian ("cart") coordinates.}
%\qualifier{eomecd}{[\code{="eom"}] Return equations of motion ("eom"), total energy ("energy"),
%      circular velocity ("circ"), or Sun-Galactic center distance ("sgcd").}
%\qualifier{Mb}{[\code{=175}] Mass of bulge in Galactic mass units, see Irrgang et al. 2013.}
%\qualifier{Md}{[\code{=2829}] Mass of disc in Galactic mass units, see Irrgang et al. 2013.}
%\qualifier{Mh}{[\code{=69725}] Mass of halo in Galactic mass units, see Irrgang et al. 2013.}
%\qualifier{bb}{[\code{=0.184}] Bulge scale length, see Irrgang et al. 2013.}
%\qualifier{ad}{[\code{=4.85}] Disc scale length, see Irrgang et al. 2013.}
%\qualifier{bd}{[\code{=0.305}] Disc scale length, see Irrgang et al. 2013.}
%\qualifier{ah}{[\code{=200}] Halo scale length, see Irrgang et al. 2013.}
%}
%\description
%    Evaluate the equations of motion, the total energy, or the circular velocity at time 't'
%    derived from a potential with a Miyamoto & Nagai bulge and disk component and a truncated,
%    flat rotation curve dark matter halo model (see Model II in Irrgang et al., 2013, A&A,
%    549, A137) using either cylindrical coordinates (r [kpc], phi [rad], z [kpc]) and their
%    canonical momenta vr [kpc/Myr], Lz [kpc^2/Myr], vz [kpc/Myr]) or cartesian coordinates
%    (x [kpc], y [kpc], z [kpc], vx [kpc/Myr], vy [kpc/Myr], vz [kpc/Myr]), see qualifier
%    'coords'. Conservation of angular momentum Lz is implemented in the equations of motion
%    for cylindrical coordinates only. The total energy E_total [kpc^2/Myr^2] is not used to
%    integrate the equations of motion although being a conserved quantity, too. Therefore,
%    conservation of energy, i.e., of E_total, is a measure for the precision of the numerical
%    methods applied.
%
%    For computing orbits with n different initial conditions, the input parameter m is
%    a [6,n]-matrix with (qualifier("coords")=="cyl")   or (qualifier("coords")=="cart")
%       m[0,*] = r;                                        m[0,*] = x;
%       m[1,*] = phi;                                      m[1,*] = y;
%       m[2,*] = z;                                        m[2,*] = z;
%       m[3,*] = vr;                                       m[3,*] = vx;
%       m[4,*] = Lz;                                       m[4,*] = vy;
%       m[5,*] = vz;                                       m[5,*] = vz;
%    If the qualifier 'eomecd' is set to "eom", the function returns a [6,n]-matrix delta with
%       delta[0,*] = vr;                                   delta[0,*] = vx;
%       delta[1,*] = Lz/r^2; % = vphi                      delta[1,*] = vy;
%       delta[2,*] = vz;                                   delta[2,*] = vz;
%       delta[3,*] = -d/dr (Potential(r,z) + Lz^2/r^2);    delta[3,*] = -d/dx Potential(x,y,z);
%       delta[4,*] = 0; % -d/dphi Potential(r,z)           delta[4,*] = -d/dy Potential(x,y,z);
%       delta[5,*] = -d/dz Potential(r,z);                 delta[5,*] = -d/dz Potential(x,y,z);
%    If the qualifier 'eomecd' is set to "energy", the function returns a [1,n]-array storing
%    the total energy for each orbit:
%       E_total(r,z,vr,vz,Lz) = Double_Type[n] = 0.5*(vr^2+vz^2+Lz^2/r^2) + Potential(r,z)
%    or
%       E_total(x,y,z,vx,vy,vz) = Double_Type[n] = 0.5*(vx^2+vy^2+vz^2) + Potential(x,y,z)
%    If the qualifier 'eomecd' is set to "circ", the function returns a [1,n]-array storing
%    the circular velocity for each orbit:
%       v_circ(r,z) = Double_Type[n] = sqrt( r * d/dr Potential(r,z) )
%    If the qualifier 'eomecd' is set to "sgcd", the function returns the Sun-Galactic center
%    distance found to fit best to this potential.
%\example
%    delta = MN_TF(0, m);
%    energy = MN_TF(0, m; eomecd="energy");
%    v_circ = MN_TF(0, m; eomecd="circ");
%    sgcd = MN_TF(; eomecd="sgcd");
%\seealso{orbit_calculator, AS, MN_NFW, plummer_MW}
%!%-
{
  % equations of motion, energy, circular velocity or Sun-Galactic center distance:
  variable eomecd = qualifier("eomecd", "eom");
  if(eomecd == "sgcd")
  {
    return 8.35;
  }
  else if(_NARGS!=2)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable coords = qualifier("coords", "cyl");
    variable t, input;
    (t, input) = ();
    input *= 1.;
    variable r, z, vz;
    if(coords=="cart")
    {
      variable x, y, vx, vy;
      x = input[0,*];
      y = input[1,*];
      z = input[2,*];
      vx = input[3,*];
      vy = input[4,*];
      vz = input[5,*];
      r = sqrt(x^2+y^2);
    }
    else % coords=="cyl"
    {
      variable phi, vr, Lz;
      r = input[0,*];
      phi = input[1,*];
      z = input[2,*];
      vr = input[3,*];
      Lz = input[4,*];
      vz = input[5,*];
    }
    % definition of gravitational potential parameters:
    variable Mb = qualifier("Mb", 175.);   % mass of the bulge
    variable Md = qualifier("Md", 2829.);  % mass of the disk
    variable Mh = qualifier("Mh", 69725.); % mass of the halo
    variable bb = qualifier("bb", 0.184);  % scale length bulge
    variable ad = qualifier("ad", 4.85);   % scale length disk
    variable bd = qualifier("bd", 0.305);  % scale length disk
    variable ah = qualifier("ah", 200.);   % scale length halo

    % variable kpcmyr_to_kms = 977.7736364875057; % = conversion factor from kpc/myr to km/s = 3.0856775975*10^16 / (10^6 * 3.15582 * 10^7);
    variable constant = 1.0459799346702096e-04; % 100./kpcmyr_to_kms^2; factor 100 because potential is given in 100km^2/s^2
    % store frequently appearing expressions to save runtime:
    variable r2 = r^2;
    variable z2 = z^2;
    variable r2z2 = r2+z2;
    variable sr2z2 = sqrt(r2z2);
    variable sz2bd2 = sqrt(z2+bd^2);
    variable sr2z2ah2 = sqrt(r2z2+ah^2);
    variable ind1, ind2, ind3, ind4;
    variable eps = 1e-5;
    ind1 = where(sr2z2 < eps, &ind2); % ind1 to avoid singularities when falling into the Galactic center
    if(eomecd == "eom")
    {
      % store frequently appearing expressions to save runtime:
      variable dPot_bulge = -Mb/(r2z2+bb^2)^1.5;
      variable dPot_disc = -Md/(r2+(ad+sz2bd2)^2)^1.5;
      variable dPot_halo = -Mh/sr2z2ah2/r2z2;

      % calculate derivatives of the potential with respect to z (dPotdz):
      variable dPotdz = constant*(dPot_bulge*z+dPot_disc*(ad+sz2bd2)/sz2bd2*z);
      dPotdz[ind2] += constant*dPot_halo[ind2]*z[ind2];

      variable ret = Double_Type[6,length(r)];

      if(coords=="cart")
      {
	% calculate derivatives of the potential with respect to x and y (dPotdx, dPotdy):
	variable dPotdx = constant*(dPot_bulge + dPot_disc)*x;
	variable dPotdy = constant*(dPot_bulge + dPot_disc)*y;
	dPotdx[ind2] += constant*dPot_halo[ind2]*x[ind2];
	dPotdy[ind2] += constant*dPot_halo[ind2]*y[ind2];

	ret[0,*] = vx;
	ret[1,*] = vy;
	ret[2,*] = vz;
	ret[3,*] = dPotdx;
	ret[4,*] = dPotdy;
	ret[5,*] = dPotdz;
	return ret;
      }
      else % coords=="cyl"
      {
	ind3 = where( Lz!=0, &ind4 );
	% calculate derivatives of the potential with respect to r (dPotdr):
	variable dPotdr = constant*(dPot_bulge+dPot_disc)*r;
	dPotdr[ind2] += constant*dPot_halo[ind2]*r[ind2];
	dPotdr[ind3] += (Lz[ind3])^2/(r[ind3])^3;

	ret[0,*] = vr;
	ret[1,ind3] = (Lz[ind3])/(r2[ind3]);
	ret[1,ind4] = 0;
	ret[2,*] = vz;
	ret[3,*] = dPotdr;
	ret[4,*] = 0*Lz;
	ret[5,*] = dPotdz;
	return ret;
      }
    }
    else if(eomecd == "energy")
    {
      % calculate the kinetic and potential energies (Ekin, Epot):
      variable Ekin;
      if(coords=="cart")
      {
	Ekin = 0.5*(vx^2+vy^2+vz^2);
      }
      else % coords=="cyl"
      {
	Ekin = 0.5*(vr^2+vz^2);
	ind3 = where( Lz!=0 );
	Ekin[ind3] += 0.5*(Lz[ind3])^2/(r2[ind3]);
      }
      variable Epot = -Mb/sqrt(r2z2+bb^2)-Md/sqrt(r2+(ad+sz2bd2)^2); % phi_bulge + phi_disc
      Epot[ind1] += -Mh/ah*log((sqrt(eps^2+ah^2)+ah)/eps);           % phi_bulge + phi_disc + phi_halo
      Epot[ind2] += -Mh/ah*log((sr2z2ah2[ind2]+ah)/sr2z2[ind2]);     % phi_bulge + phi_disc + phi_halo
      return Ekin + constant*Epot;
    }
    else if(eomecd == "circ")
    {
      dPotdr = Mb/(r2z2+bb^2)^(1.5)*r+Md/(r2+(ad+sz2bd2)^2)^(1.5)*r;
      dPotdr[ind2] += Mh/sr2z2ah2[ind2]/r2z2[ind2]*r[ind2];
      return 10.*sqrt(r*dPotdr);
    }
  }
}%}}}

define plummer_MW() %{{{
%!%+
%\function{plummer_MW}
%\synopsis{Alternative model potential for the function 'orbit_calculator'}
%\usage{plummer_MW(Double_Types t, m[6,n]; qualifiers)}
%\qualifiers{
%\qualifier{plummer_spheres}{Structure whose fields are again structures with fields 't' [Myr],
%      'x' [kpc], 'y' [kpc], 'z' [kpc], 'psa' [Msun/Mgal*constant] (see notes), and 'psb'
%      [kpc^2] describing the orbits and shapes of the Plummer spheres.
%      Always make sure that the tabulated times 't[*]' are in monotonic increasing order
%      if 't[-1]' is positive or in monotonic decreasing order if 't[-1]' is negative.}
%\qualifier{MW_potential}{[\code{="AS"}]: Function ("AS", "MN_NFW", or "MN_TF"), which evaluates the
%      equations of motion that result from a model for the gravitational potential of
%      the Milky Way.}
%\qualifier{All qualifiers from the Milky Way model potential (see qualifier 'MW_potential').}{}
%\qualifier{All qualifiers from the function 'plummer_interaction_kernel'.}{}
%}
%\description
%    This function provides an alternative model for the gravitational potential of the
%    Milky Way which can be used by the function 'orbit_calculator'. The gravitational
%    forces of a standard Milky Way potential (see qualifier 'model') are combined with
%    those arising from the interaction with a number of moving Plummer spheres (see
%    function 'plummer_interaction_kernel' and qualifier 'plummer_spheres') to determine
%    the acceleration of n independent test particles at time 't'.
%\notes
%    Because of the unit convention used for the potentials of the Milky Way, the field
%    'psa' of the qualifier 'plummer_spheres', which is assumed to be the mass of the
%    respective Plummer sphere in solar masses, has to be converted to Galactic mass units
%    and then multiplied with a constant accounting for the remaining unit conversions
%    (see example below). The units of 'psb' have to be kpc^2.
%\example
%    % Test particle affected by the Milky Way and satellite galaxies:
%    t_end = -100; % integration time in Myr
%    model = "AS"; % Milky Way mass model
%    s = properties_satellite_galaxies();
%    i = struct{ x, y, z, vx, vy, vz, psa, psb };
%    SunGCDist = (@(__get_reference(model)))(; eomecd="sgcd"); % Sun-GC distance of chosen mass model
%    temp = [SunGCDist,0,0,0,0,0]; reshape(temp, [6,1]);
%    vlsr = (@(__get_reference(model)))(0, temp; eomecd="circ")[0]; % Local standard of rest velocity of chosen mass model
%    (i.x, i.y, i.z, i.vx, i.vy, i.vz) = cel2gal(s.ah, s.am, s.as, s.dd, s.dm, s.ds, s.dist, s.vrad, s.pma_cos_d, s.pmd; SunGCDist=SunGCDist, vlsr=vlsr);
%    kpcmyr_to_kms = 977.7736364875057; % = conversion factor from kpc/myr to km/s = 3.0856775975*10^16 / (10^6 * 3.15582 * 10^7);
%    i.vx /= kpcmyr_to_kms;
%    i.vy /= kpcmyr_to_kms;
%    i.vz /= kpcmyr_to_kms;
%    i.psa = s.Pl_mass/2.325131802556774e+07; % conversion from solar masses to Galactic mass units Mgal to have G=1
%    % Mgal = 2.325131802556774e+07 = 1e8*3.0856775975*1e19/6.6742/1e-11/1.9884/1e30, see Irrgang et al., 2013, A&A, 549, A137
%    const = 100./kpcmyr_to_kms^2; % factor 100 because potential is given in 100 km^2/s^2, see Irrgang et al. 2013
%    i.psa *= const;
%    i.psb = (s.Pl_radius)^2;
%    ps = N_body_simulation(i, t_end; kernel="N_body_simulation_MW_kernel", psa=i.psa, psb=i.psb, model=model);
%    % ps contains time-dependent coordinates of Plummer spheres
%    % add information about shape of the Plummer spheres:
%    j = 0;
%    foreach field (get_struct_field_names(ps))
%    {
%      temp = struct{ psa=i.psa[j], psb=i.psb[j] };
%      set_struct_field(ps, field, struct_combine(get_struct_field(ps, field), temp));
%      j++;
%    };
%    % compute trajectories of test particles:
%    s = orbit_calculator(4,38,12.8,-54,33,12,61,723,0.86,0.57,t_end; set, model="plummer_MW", MW_potential=model, plummer_spheres=ps);
%    plot(s.tr.o0.x,s.tr.o0.y);
%    % without satellite galaxies:
%    s = orbit_calculator(4,38,12.8,-54,33,12,61,723,0.86,0.57,t_end; set, model=model);
%    oplot(s.tr.o0.x,s.tr.o0.y);
%\seealso{orbit_calculator, plummer_interaction_kernel, AS, MN_NFW, MN_TF}
%!%-
{
  variable MW_potential = qualifier("MW_potential", "AS"); % function evaluating the equations of motions of the Milky Way potential
  if(__get_reference(MW_potential)==NULL or (MW_potential!="AS" and MW_potential!="MN_NFW" and MW_potential!="MN_TF"))
  {
    _pop_n(_NARGS); % remove function arguments from stack
    throw UsageError, sprintf("Usage error in '%s': Invalid input for qualifier 'MW_potential'.", _function_name());
  }
  MW_potential = __get_reference(MW_potential);
  if(qualifier("eomecd")=="sgcd")
    return (@MW_potential)(;; __qualifiers);
  else if(_NARGS!=2)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': Wrong number of arguments.", _function_name());
  }
  variable ps;
  if(qualifier_exists("plummer_spheres"))
    ps = qualifier("plummer_spheres");
  else % create structure with no fields
  {
    ps = @Struct_Type(String_Type[0]);
  }
  variable t, m;
  (t, m) = ();
  variable r = (@MW_potential)(t,m;; __qualifiers); % [6,n]-matrix with velocities and forces
  if(qualifier("eomecd")=="circ")
    return r;
  variable r2 = plummer_interaction_kernel(t,m,ps;; __qualifiers); % [3,n]-matrix with forces only
  if(qualifier("eomecd")=="eom")
  {
    r[3,*] += r2[0,*]; % sum forces for first coordinate
    r[4,*] += r2[1,*]; % sum forces for second coordinate
    r[5,*] += r2[2,*]; % sum forces for third coordinate
  }
  else if(qualifier("eomecd")=="energy")
    r += r2; % sum energies
  return r;
}%}}}

define orbit_calculator() %{{{
%!%+
%\function{orbit_calculator}
%\synopsis{Calculate orbits of test particles in a Galactic gravitational potential}
%\usage{orbit_calculator(Double_Types ah[], am[], as[], dd[], dm[], ds[], dist[], vrad[], pma_cos_d[], pmd[], t_end; qualifiers)
%    % or (for error propagation)
%    orbit_calculator(Double_Types ah, am, as, dd, dm, ds, dist, dist_sigma[], vrad, vrad_sigma[], pma_cos_d, pma_cos_d_sigma[], pmd, pmd_sigma[], t_end; qualifiers)}
%\altusage{orbit_calculator(Double_Types x[], y[], z[], vx[], vy[], vz[], t_end; qualifiers)}
%\qualifiers{
%\qualifier{coords}{[\code{="cyl"}]: Use cylindrical ("cyl") or cartesian ("cart") coordinates for the internal
%      computations. Note that circular orbits like that of the Sun are computed faster in cylindrical
%      coordinates whereas cartesian coordinates are much more efficient for straight-line trajectories
%      or those that come very close to the z-axis where angular momentum terms slow down cylindrical
%      calculations.}
%\qualifier{dt}{Deactivate adaptive mode and use this fixed stepsize instead.}
%\qualifier{model}{[\code{="AS"}]: Function ("AS", "MN_NFW", "MN_TF", or "plummer_MW") evaluating the equations
%      of motion, circular velocity, and energy of the model potential.}
%\qualifier{MC_runs}{[\code{=nint(10^5)}]: Number of Monte Carlo realizations in the case that 1-sigma errors are given.}
%\qualifier{ODE_solver}{[\code{="RKCK"}]: Choose among three different Runge-Kutta (RK) integration methods:
%      "RKF": RK-Fehlberg, "RKCK": RK-Cash-Karp, "RKDP": RK-Dormand-Prince.}
%\qualifier{parallax_pma_corr}{[\code{=0}]: Correlation between parallax and proper motion in right ascension.}
%\qualifier{parallax_pmd_corr}{[\code{=0}]: Correlation between parallax and proper motion in declination.}
%\qualifier{pma_pmd_corr}{[\code{=0}]: Correlation between proper motion in right ascension and in declination.}
%\qualifier{disk}{[\code{=struct{radius = 0, height = 0.1, x = 0, y = 0, z = 0, crossings = 1}}]: If disk.radius > 0,
%      orbit integration will be stopped at the moment a trajectory has crossed a horizontal plane
%      located at z = disk.z + disk.height * Gaussian-random-number inside a circle of radius
%      disk.radius centered at (disk.x, disk.y) for the disk.crossings-th time.}
%\qualifier{seed}{[\code{=_time()}]: Seed the random number generator via the function 'seed_random'.}
%\qualifier{set}{[\code{=0}]: If present, trajectories will be saved ("Save entire trajectories"). As
%      this can be very memory-consuming for a large number of orbits, one can additionally
%      use the value of this qualifier to specify a lower limit on the time difference of
%      two consecutive moments of time that will be saved.}
%\qualifier{stff}{"Save to fits files": Prefix of fits files to which initial and final structures are written.}
%\qualifier{SunGCDist}{By default, the Sun-Galactic center distance is taken from the current model.
%      Use this qualifier to explicitly set a distance in kpc.}
%\qualifier{tolerance}{[\code{=1e-8}]: Absolute error control tolerance; lower limit: 1e-15.}
%\qualifier{verbose}{Show intermediate times t.}
%\qualifier{Any qualifiers from 'cel2gal' and the model potential function except 'vlsr' and 'eomecd'.}{
%      Important note: for consistency, the local standard of rest velocity vlsr is calculated from the
%      circular velocity of the current model potential evaluated at (r=SunGCDist, z=0).}
%}
%\description
%    Calculate orbits of test particles in a Galactic gravitational potential from given
%    initial conditions. The latter can be given either in celestial or Galactic cartesian
%    coordinates (see the help on the function 'cel2gal' for format and unit conventions).
%    Integration starts at time t=0 and ends at t=t_end [Myr], which is the last input
%    parameter. A negative t_end implies backward integration, which is e.g. useful to
%    trace back orbits in time. The potential and its equations of motion are outsourced
%    to a function specified by the qualifier 'model'. In this way, switching from one model
%    to another one is simply done by changing the before mentioned qualifier. Make always
%    use of cylindrical coordinates (r [kpc], phi [rad], z [kpc]) and their canonical momenta
%    (vr [kpc/Myr], Lz [kpc^2/Myr], vz [kpc/Myr]) as well as of cartesian coordinates (x [kpc],
%    y [kpc], z [kpc], vx [kpc/Myr], vy [kpc/Myr], vz [kpc/Myr]) to set up the equations of
%    motion (see the qualifier 'coords'). To numerically integrate the coupled differential
%    equations, an adaptive Runge-Kutta method of fourth/fifth order is used (see qualifier
%    'ODE_solver'). The stepsize is hereby controlled such that for each step an absolute
%    accuracy in coordinates (in units of kpc) and velocity components (in units of km/s) is
%    achieved that is smaller than given by the qualifier 'tolerance'. The function is optimized
%    for multi-orbit calculations, i.e., when the input parameters are either arrays or 1-sigma
%    uncertainties are given in addition. The latter are used to create Gaussian distributed
%    initial conditions to perform error propagation based on a number of Monte Carlo runs as
%    specified by the qualifier 'MC_runs'. To assign asymmetric uncertainties, use an array
%    of length two, i.e., [sigma_plus, sigma_minus], instead of a single number. The Gaussian
%    distribution is then split up into two Gaussian distributions, one with standard deviation
%    sigma_plus (for values larger than the respective input parameter) and one with sigma_minus
%    (else). To account for correlations between the distance (or alternatively the parallax,
%    see the qualifier 'parallax' in the function 'cel2gal'), proper motions in right ascension,
%    and proper motion in declination, make use of the qualifiers 'parallax_pma_corr',
%    'parallax_pmd_corr', and 'pma_pmd_corr'. Note that asymmetric uncertainties and
%    parameter correlations are mutually exclusive for individual parameters. All orbits are
%    computed simultaneously and on the same time grid. Stepsize control is hereby based on
%    the worst-offender principle. The function returns one structure with the two fields
%    "i" (initial) and "f" (final) (both again structures) containing the initial and final
%    Galactic cartesian coordinates (in kpc) and velocities (in km/s) as well as the z-component
%    of angular momentum Lz (in kpc^2/Myr) and total energy E_total = E_kin + E_pot (in kpc^2/Myr^2).
%    To see whether conservation of energy or conservation of angular momentum is implemented
%    in the equations of motion or not, have a look at the help of the outsourced potential
%    function. The final structure contains also the minimum and maximum Galactocentric distances
%    Rmin = min(R(t)) and Rmax = max(R(t)) with R(t) = sqrt(x(t)^2+y(t)^2+z(t)^2) and - if the
%    qualifier 'coords' is set to "cyl" - the number of revolutions about the Galactic z-axis
%    (negative values imply a motion in direction of Galactic rotation, i.e. prograde orbits,
%    while positive ones imply retrograde orbits). To save the initial and final structures to
%    fits files, set the qualifier 'stff' to the desired prefix of the filename. If not only
%    the initial and final situation is of interest, but rather the entire trajectories, use
%    the qualifier 'set'. In this case, the additional field "tr" (trajectory) (again a structure,
%    one field per orbit ("o1", "o2", ...)) is added to the returned structure containing all
%    orbits. For tracing back orbits to the Galactic disk, use the 'disk' qualifier to stop
%    individual calculations at the moment a trajectory has crossed the disk for a certain
%    number of times.
%\example
%    s = orbit_calculator(12,22,29.6,40,49,36,3.078,262,-13.52,16.34,-1000; disk = struct{radius=50, height=0.2, x=0, y=0, z=0, crossings=1});
%    s = orbit_calculator(12,22,29.6,40,49,36,3.078,[0.6,0.3],262,5,-13.52,1.31,16.34,1.37,-1000; disk = struct{radius=50, height=0.2, x=0, y=0, z=0, crossings=1}, stff="HIP60350", MC_runs=100);
%    s = orbit_calculator(-8.4,0,0,0,242,0,250; set, model="MN_TF");
%    plot(s.tr.o0.x, s.tr.o0.y);
%    s = orbit_calculator([8:10:#500], [0:1:#500], [4:5:#500], [0:0:#500], [210:230:#500], [-10:-50:#500], 1000);
%
%    % Example using parallax, proper motions, and correlation parameters from Gaia:
%    s = orbit_calculator(12,22,29.6,40,49,36,0.325,0.3,262,5,-13.52,1.31,16.34,1.37,-1; parallax, parallax_pma_corr=0.1, parallax_pmd_corr=0.2, pma_pmd_corr=0.75);
%
%    % Example using the spectroscopic distance and only proper motions from Gaia:
%    s = orbit_calculator(12,22,29.6,40,49,36,3.077,[0.6,0.3],262,5,-13.52,1.31,16.34,1.37,-1; pma_pmd_corr=0.75);
%\seealso{cel2gal, AS, MN_NFW, MN_TF, plummer_MW, xfig_3d_orbit_on_cube}
%!%-
{
  % =====================
  % processing the input:
  variable model = qualifier("model", "AS"); % function evaluating the equations of motions, circular velocity, energy of the model potential
  if(__get_reference(model)==NULL or (model!="AS" and model!="MN_NFW" and model!="MN_TF" and model!="plummer_MW"))
  {
    _pop_n(_NARGS); % remove function arguments from stack
    throw UsageError, sprintf("Usage error in '%s': wrong input for qualifier 'model'.", _function_name());
  }
  variable coords = qualifier("coords", "cyl");
  if(coords != "cyl" and coords != "cart")
  {
    _pop_n(_NARGS); % remove function arguments from stack
    throw UsageError, sprintf("Usage error in '%s': wrong input for qualifier 'coords'.", _function_name());
  }
  model = __get_reference(model); % function evaluating the equations of motions, circular velocity, energy of the model potential
  variable dt_fixed = qualifier_exists("dt"); % fixed stepsize for integrator
  variable eps = _max( qualifier("tolerance", 1e-8), 1e-15 ); % error control tolerance
  variable disk = qualifier("disk", struct{radius = 0, height = 0.1, x = 0, y = 0, z = 0, crossings = 1});
  % If disk.radius > 0, orbit integration will be stopped at the moment a trajectory has crossed a horizontal plane located at
  % z = disk.z + disk.height * Gaussian-random-number inside a circle of radius disk.radius centered at (disk.x, disk.y) for the disk.crossings-th time.
  if(typeof(disk)!=Struct_Type)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    throw UsageError, sprintf("Usage error in '%s': The qualifier 'disk' has to be a Struct_Type.", _function_name());
  }
  if(struct_field_exists(disk, "radius")==0 or struct_field_exists(disk, "height")==0 or struct_field_exists(disk, "x")==0 or struct_field_exists(disk, "y")==0 or struct_field_exists(disk, "z")==0 or struct_field_exists(disk, "crossings")==0)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    throw UsageError, sprintf("Usage error in '%s': The qualifier 'disk' has to be a Struct_Type with fields 'radius', 'height', 'x', 'y', 'z', and 'crossings'.", _function_name());
  }
  if(typeof(disk.crossings)!=Integer_Type || disk.crossings<1)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    throw UsageError, sprintf("Usage error in '%s': The qualifier 'disk.crossings' has to be a positive Integer_Type.", _function_name());
  }
  variable set = qualifier_exists("set"); % "save entire trajectory"
  variable set_threshold = qualifier("set", 0);
  if(typeof(set_threshold)==Null_Type) % to cover cases where the qualifier 'set' is present but without a specified value
    set_threshold = 0;
  variable verbose = qualifier_exists("verbose");
  % algorithm for solving the ordinary differential equations (ODEs): Runge-Kutta-Fehlberg ("RKF"), Runge-Kutta-Cash-Karp ("RKCK"), Runge-Kutta-Dormand-Prince ("RKDP"):
  variable method = qualifier("ODE_solver", "RKCK");
  % ============================================================================
  variable qualies = __qualifiers;
  qualies = struct_combine(qualies, "eomecd");
  qualies = struct_combine(qualies, "vlsr");
  qualies = struct_combine(qualies, "SunGCDist");
  variable SunGCDist = (@model)(; eomecd="sgcd"); % Sun-Galactic center distance from the model
  if(qualifier_exists("SunGCDist"))
    SunGCDist = qualifier("SunGCDist");
  set_struct_field(qualies, "SunGCDist", SunGCDist); % qualifier 'SunGCDist' in function cel2gal is set according to the current potential
  variable temp0 = 0*Double_Type[6,length(SunGCDist)]; temp0[0,*] = SunGCDist;
  set_struct_field(qualies, "eomecd", "circ"); % important for the following line
  set_struct_field(qualies, "vlsr", ((@model)(0, temp0;; qualies))[0] ); % qualifier 'vlsr' in function cel2gal is set according to the current potential
  set_struct_field(qualies, "eomecd", "eom"); % set back to default
  qualies = struct_combine(qualies, "coords");
  set_struct_field(qualies, "coords", coords);
  % ============================================================================
  if(qualifier_exists("stff")) % some keywords for fits-header to know what input parameters or qualifiers were used for this computation
    variable keys = struct{SunGCDist = string(median(SunGCDist))};
  variable x, y, z, vx, vy, vz, t_end; % t_end: maximum integration time in Myrs

  if(_NARGS==11)
  {
    variable ah, am, as, dd, dm, ds, dist, rv, pma, pmd;
    (ah, am, as, dd, dm, ds, dist, rv, pma, pmd, t_end) = ();
    ah = [1.*ah]; am = [1.*am]; as = [1.*as]; dd = [1.*dd]; dm = [1.*dm]; ds = [1.*ds]; dist = [1.*dist]; rv = [1.*rv]; pma = [1.*pma]; pmd = [1.*pmd];
    (x, y, z, vx, vy, vz) = cel2gal(ah, am, as, dd, dm, ds, dist, rv, pma, pmd;; qualies);
  }

  else if(_NARGS==15)
  {
    variable dist_sigma, rv_sigma, pma_sigma, pmd_sigma;
    (ah, am, as, dd, dm, ds, dist, dist_sigma, rv, rv_sigma, pma, pma_sigma, pmd, pmd_sigma, t_end) = ();
    variable MC_runs = nint(qualifier("MC_runs", nint(10^5)));
    variable parallax_pma_corr = qualifier("parallax_pma_corr", 0);
    if(abs(parallax_pma_corr)>1)
      throw UsageError, sprintf("Usage error in '%s': Qualifier 'parallax_pma_corr' has to be between -1 and 1.", _function_name());
    variable parallax_pmd_corr = qualifier("parallax_pmd_corr", 0);
    if(abs(parallax_pmd_corr)>1)
      throw UsageError, sprintf("Usage error in '%s': Qualifier 'parallax_pmd_corr' has to be between -1 and 1.", _function_name());
    variable pma_pmd_corr = qualifier("pma_pmd_corr", 0);
    if(abs(pma_pmd_corr)>1)
      throw UsageError, sprintf("Usage error in '%s': Qualifier 'pma_pmd_corr' has to be between -1 and 1.", _function_name());
    variable seed = nint(qualifier("seed", _time()));
    seed_random(seed); % initialize the seed according to qualifier or time
    variable rv_distr;
    if(length(rv_sigma)==2)
    {
      rv_distr = grand(MC_runs);
      rv_distr[where(rv_distr>0)] *= rv_sigma[0];
      rv_distr[where(rv_distr<0)] *= rv_sigma[1];
      rv_distr += rv;
    }
    else
      rv_distr = rv+grand(MC_runs)*rv_sigma[0];
    variable cov_mat = 0.*Double_Type[3,3]; % initialize covariance matrix between parallax (or distance), pma, and pmd with zeros
    variable distr = Double_Type[3,MC_runs];
    variable len = length(dist_sigma);
    if(len==2 and (parallax_pma_corr!=0 or parallax_pmd_corr!=0))
      throw UsageError, sprintf("Usage error in '%s': Asymmetric uncertainties and correlation parameters cannot be used together.", _function_name());
    if(parallax_pma_corr==0 and parallax_pmd_corr==0)
    {
      cov_mat[0,0] = 1; cov_mat[1,0] = 0; cov_mat[0,1] = 0; cov_mat[2,0] = 0; cov_mat[0,2] = 0;
      if(len==2)
      {
	distr[0,*] = grand(MC_runs);
	distr[0,where(distr[0,*]>0)] *= dist_sigma[0];
	distr[0,where(distr[0,*]<0)] *= dist_sigma[1];
      }
      else
	distr[0,*] = grand(MC_runs)*dist_sigma[0];
    }
    else
    {
      cov_mat[0,0] = (dist_sigma[0])^2; cov_mat[1,0] = parallax_pma_corr*dist_sigma[0]*pma_sigma[0]; cov_mat[0,1] = cov_mat[1,0];
      cov_mat[2,0] = parallax_pmd_corr*dist_sigma[0]*pmd_sigma[0]; cov_mat[0,2] = cov_mat[2,0];
      distr[0,*] = grand(MC_runs);
    }
    len = length(pma_sigma);
    if(len==2 and (parallax_pma_corr!=0 or pma_pmd_corr!=0))
      throw UsageError, sprintf("Usage error in '%s': Asymmetric uncertainties and correlation parameters cannot be used together.", _function_name());
    if(parallax_pma_corr==0 and pma_pmd_corr==0)
    {
      cov_mat[1,1] = 1; cov_mat[1,0] = 0; cov_mat[0,1] = 0; cov_mat[1,2] = 0; cov_mat[2,1] = 0;
      if(len==2)
      {
	distr[1,*] = grand(MC_runs);
	distr[1,where(distr[1,*]>0)] *= pma_sigma[0];
	distr[1,where(distr[1,*]<0)] *= pma_sigma[1];
      }
      else
	distr[1,*] = grand(MC_runs)*pma_sigma[0];
    }
    else
    {
      cov_mat[1,1] = (pma_sigma[0])^2; cov_mat[1,0] = parallax_pma_corr*dist_sigma[0]*pma_sigma[0]; cov_mat[0,1] = cov_mat[1,0];
      cov_mat[1,2] = pma_pmd_corr*pma_sigma[0]*pmd_sigma[0]; cov_mat[2,1] = cov_mat[1,2];
      distr[1,*] = grand(MC_runs);
    }
    len = length(pmd_sigma);
    if(len==2 and (parallax_pmd_corr!=0 or pma_pmd_corr!=0))
      throw UsageError, sprintf("Usage error in '%s': Asymmetric uncertainties and correlation parameters cannot be used together.", _function_name());
    if(parallax_pmd_corr==0 and pma_pmd_corr==0)
    {
      cov_mat[2,2] = 1; cov_mat[2,0] = 0; cov_mat[0,2] = 0; cov_mat[1,2] = 0; cov_mat[2,1] = 0;
      if(len==2)
      {
	distr[2,*] = grand(MC_runs);
	distr[2,where(distr[2,*]>0)] *= pmd_sigma[0];
	distr[2,where(distr[2,*]<0)] *= pmd_sigma[1];
      }
      else
	distr[2,*] = grand(MC_runs)*pmd_sigma[0];
    }
    else
    {
      cov_mat[2,2] = (pmd_sigma[0])^2; cov_mat[2,0] = parallax_pmd_corr*dist_sigma[0]*pmd_sigma[0]; cov_mat[0,2] = cov_mat[2,0];
      cov_mat[1,2] = pma_pmd_corr*pma_sigma[0]*pmd_sigma[0]; cov_mat[2,1] = cov_mat[1,2];
      distr[2,*] = grand(MC_runs);
    }
    variable e;
    try(e)
    {
      distr = cholesky_decomposition(cov_mat)#distr; % matrix product of the Cholesky decomposition of the covariance matrix and a vector containing random numbers, see the function 'cholesky_decomposition' for details
    }
    catch AnyError:
    {
      vmessage(e.message);
      throw UsageError, sprintf("Usage error in '%s': The given parameter correlations are not physical.", _function_name());
    }
    variable dist_distr = distr[0,*] + dist;
    variable pma_distr  = distr[1,*] + pma;
    variable pmd_distr  = distr[2,*] + pmd;
    (x, y, z, vx, vy, vz) = cel2gal(ah*ones(MC_runs), am*ones(MC_runs), as*ones(MC_runs), dd*ones(MC_runs), dm*ones(MC_runs), ds*ones(MC_runs), dist_distr, rv_distr, pma_distr, pmd_distr;; qualies);
    variable ind = where(dist_distr>0); len = length(ind); % omit negative distances or parallaxes
    if(len!=MC_runs)
    {
      vmessage("Warning in '%s': only %d out of %d Monte Carlo runs are performed because initial conditions with negative distances have been omitted.", _function_name(), len, MC_runs);
      x = x[ind]; y = y[ind]; z = z[ind]; vx = vx[ind]; vy = vy[ind]; vz = vz[ind];
    }
    if(qualifier_exists("stff")) % some keywords for fits-header to know what input parameters or qualifiers were used for this computation
    {
      keys = struct_combine( keys, struct{ maximum_integration_time = string(t_end), rah=string(ah), ras=string(am), ram=string(as), decd=string(dd), decm=string(dm), decs=string(ds),
	dist=string(dist), dist_sigma_plus=string(dist_sigma[0]), dist_sigma_minus=string(dist_sigma[-1]), rv=string(rv), rv_sigma_plus=string(rv_sigma[0]), rv_sigma_minus=string(rv_sigma[-1]),
	pma_cos_delta=string(pma), pma_cos_delta_sigma_plus=string(pma_sigma[0]), pma_cos_delta_sigma_minus=string(pma_sigma[-1]), pmd=string(pmd), pmd_sigma_plus=string(pmd_sigma[0]),
	pmd_sigma_minus=string(pmd_sigma[-1]) });
    }
  }

  else if(_NARGS==7)
  {
    (x, y, z, vx, vy, vz, t_end) = ();
  }

  else
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }

  variable s = sign(t_end);
  if(s==0) s=1; % to account for t_end==0
  t_end *= (1.*s); vx = [1.*s*vx]; vy = [1.*s*vy]; vz = [1.*s*vz]; % for backward integration, i.e. t_end < 0, reverse t_end and the sign of the velocity components
  x = [1.*x]; y = [1.*y]; z = [1.*z]; % cast to Double_Type[]
  % =====================

  variable ret;
  if(set) ret = struct{i, f, tr};
  else ret = struct{i, f};
  % ==============================================================================================================
  % preparations:
  variable kms_to_kpcMyr = 0.0010227316044154543; % km/s = 1 / (3.0856775975*10^16 / 10^6 / 3.15582 / 10^7) kpc/Myr;
  vx *= kms_to_kpcMyr; vy *= kms_to_kpcMyr; vz *= kms_to_kpcMyr;
  ret.i = struct{ t, x, y, z, vx, vy, vz, Lz, Energy };
  len = length(x);
  ret.i.t = 0.*Double_Type[len];
  ret.i.x = x;
  ret.i.y = y;
  ret.i.z = z;
  ret.i.vx = vx;
  ret.i.vy = vy;
  ret.i.vz = vz;
  variable disk_crossings = 0*Integer_Type[len];
  variable disk_positions = disk.z + disk.height*grand(len);
  variable Rmin = [sqrt(x^2+y^2+z^2)]; % initialize Rmin as Array_Type
  variable Rmax = [sqrt(x^2+y^2+z^2)]; % initialize Rmax as Array_Type
  variable m = Double_Type[6, len];
  if(coords=="cart")
  {
    m[0,*]=x, m[1,*]=y, m[2,*]=z, m[3,*]=vx, m[4,*]=vy, m[5,*]=vz;
    ret.i.Lz = @m[0,*]*@m[4,*] - @m[1,*]*@m[3,*]; % Lz = x*vy - y*vx
  }
  else % transform cartesian coordinates to cylindrical ones
  {
    (m[0,*], m[1,*], m[2,*], m[3,*], m[4,*], m[5,*]) = cart2cyl(x, y, z, vx, vy, vz); % r, phi, z, vr, vphi, vz = ...
    m[4,*] *= (m[0,*])^2; % conversion from vphi to Lz: Lz = vphi*r^2
    ret.i.Lz = @m[4,*];
    variable phi_start = m[1,*];
  }
  variable qualies_energy = @qualies;
  set_struct_field(qualies_energy, "eomecd", "energy"); % important for the following line
  ret.i.Energy = (@model)(0, m;; qualies_energy); % calculate initial total energy using the model-function -> function 'orbit_calculator' is model independent
  % ==============================================================================================================

  % ============================
  % initializing the ODE-solver:
  variable k1, k2, k3, k4, k5, k6, k7, dx_c, dx_e; % dx_c = dx continue, dx_e = dx error estimation
  variable coef = Double_Type[7,7]; % modified Butcher tableau
  if(method == "RKF")
  {
    coef[0,0]= 1./4;   coef[0,1]=  1./4;
    coef[1,0]= 3./8;   coef[1,1]=  3./32;      coef[1,2]=  9./32;
    coef[2,0]= 12./13; coef[2,1]=  1932./2197; coef[2,2]= -7200./2197;  coef[2,3]=  7296./2197;
    coef[3,0]= 1.;     coef[3,1]=  439./216;   coef[3,2]= -8.;          coef[3,3]=  3680./513;    coef[3,4]= -845./4104;
    coef[4,0]= 1./2;   coef[4,1]= -8./27;      coef[4,2]=  2.;          coef[4,3]= -3544./2565;   coef[4,4]=  1859./4104;   coef[4,5]= -11./40;
    coef[5,0]= 0;      coef[5,1]=  25./216;    coef[5,2]=  1408./2565;  coef[5,3]=  2197./4104;   coef[5,4]= -1./5;         coef[5,5]= 0; % fourth order solution
    coef[6,0]= 0;      coef[6,1]=  16./135;    coef[6,2]=  6656./12825; coef[6,3]=  28561./56430; coef[6,4]= -9./50;        coef[6,5]= 2./55; % fifth order solution
  }

  else if(method == "RKDP")
  {
    coef[0,0]= 1./5;  coef[0,1]= 1./5;
    coef[1,0]= 3./10; coef[1,1]= 3./40;       coef[1,2]=  9./40;
    coef[2,0]= 4./5;  coef[2,1]= 44./45;      coef[2,2]= -56./15;      coef[2,3]= 32./9;
    coef[3,0]= 8./9;  coef[3,1]= 19372./6561; coef[3,2]= -25360./2187; coef[3,3]= 64448./6561; coef[3,4]= -212./729;
    coef[4,0]= 1.;    coef[4,1]= 9017./3168;  coef[4,2]= -355./33;     coef[4,3]= 46732./5247; coef[4,4]=  49./176;       coef[4,5]= -5103./18656;
    coef[5,0]= 1.;    coef[5,1]= 35./384;     coef[5,2]= 500./1113;    coef[5,3]=  125./192;   coef[5,4]= -2187./6784;    coef[5,5]= 11./84; % fourth order solution
    coef[6,0]= 0;     coef[6,1]= 5179./57600; coef[6,2]= 7571./16695;  coef[6,3]=  393./640;   coef[6,4]= -92097./339200; coef[6,5]= 187./2100;    coef[6,6]= 1./40; % fitfh order solution
  }

  else % use RKCK if method is none of the two above methods
  {
    coef[0,0]= 1./5;  coef[0,1]= 1./5;
    coef[1,0]= 3./10; coef[1,1]= 3./40;       coef[1,2]= 9./40;
    coef[2,0]= 3./5;  coef[2,1]= 3./10;       coef[2,2]= -9./10;       coef[2,3]= 6./5;
    coef[3,0]= 1.;    coef[3,1]= -11./54;     coef[3,2]= 5./2;         coef[3,3]= -70./27;       coef[3,4]= 35./27;
    coef[4,0]= 7./8;  coef[4,1]= 1631./55296; coef[4,2]= 175./512;     coef[4,3]= 575./13824;    coef[4,4]= 44275./110592; coef[4,5]= 253./4096;
    coef[5,0]= 0;     coef[5,1]= 2825./27648; coef[5,2]= 18575./48384; coef[5,3]= 13525./55296;  coef[5,4]= 277./14336;    coef[5,5]= 1./4; % fourth order solution
    coef[6,0]= 0;     coef[6,1]= 37./378;     coef[6,2]= 250./621;     coef[6,3]= 125./594;      coef[6,4]= 0;             coef[6,5]= 512./1771; % fitfh order solution
  }
  % ============================

  % =======================================
  % solving the equations of motions:
  variable t = 0;
  variable dt = 0.001;
  if(dt_fixed) dt = qualifier("dt"); % use fixed stepsize when qualifier "dt" is present
  variable t_eoc = t_end + 0*Double_Type[len]; % t_eoc: time at the end or moment of crossing the disk
  ind = [0:len-1:1]; % array of indices, important for bookkeeping of which orbits already crossed the disk and thus for qualifier 'disk.radius'
  if(set)
  {
    variable t_last_save = 0;
    variable cols = String_Type[len];
    variable i, j;
    _for i(0, len-1, 1)
    {
      cols[i] = sprintf("o%d", i);
    }
    ret.tr = @Struct_Type(cols);
    variable temp1 = struct{ t, x, y, z, vx, vy, vz, Lz, Energy };
    _for i(0, len-1, 1)
    {
      temp1.t = {t};
      temp1.x = {m[0,i]};
      temp1.y = {m[1,i]};
      temp1.z = {m[2,i]};
      temp1.vx = {m[3,i]};
      temp1.vz = {m[5,i]};
      temp1.Lz = {m[4,i]};
      temp1.Energy = {ret.i.Energy[i]};
      set_struct_field(ret.tr, cols[i], @temp1);
    }
  }
  while(t < t+dt <= t_end and length(ind)>0)
  {
    k1 = dt * (@model)(s*t, m[*,ind];; qualies);
    k2 = dt * (@model)(s*(t + coef[0,0]*dt), m[*,ind] + coef[0,1]*k1;; qualies);
    k3 = dt * (@model)(s*(t + coef[1,0]*dt), m[*,ind] + coef[1,1]*k1 + coef[1,2]*k2;; qualies);
    k4 = dt * (@model)(s*(t + coef[2,0]*dt), m[*,ind] + coef[2,1]*k1 + coef[2,2]*k2 + coef[2,3]*k3;; qualies);
    k5 = dt * (@model)(s*(t + coef[3,0]*dt), m[*,ind] + coef[3,1]*k1 + coef[3,2]*k2 + coef[3,3]*k3 + coef[3,4]*k4;; qualies);
    k6 = dt * (@model)(s*(t + coef[4,0]*dt), m[*,ind] + coef[4,1]*k1 + coef[4,2]*k2 + coef[4,3]*k3 + coef[4,4]*k4 + coef[4,5]*k5;; qualies);
    if(method == "RKDP") % for RKPD use fifth order solution to continue integration
    {
      dx_e = coef[5,1]*k1 + coef[5,2]*k3 + coef[5,3]*k4 + coef[5,4]*k5 + coef[5,5]*k6; % fourth order solution
      k7 = dt * (@model)(s*(t + coef[5,0]*dt), m[*,ind] + dx_e;; qualies);
      dx_c = coef[6,1]*k1 + coef[6,2]*k3 + coef[6,3]*k4 + coef[6,4]*k5 + coef[6,5]*k6 + coef[6,6]*k7; % fifth order solution
    }
    else % for RKF, RKCK use fourth order solution to continue integration
    {
      dx_c = coef[5,1]*k1 + coef[5,2]*k3 + coef[5,3]*k4 + coef[5,4]*k5 + coef[5,5]*k6; % fourth order solution
      dx_e = coef[6,1]*k1 + coef[6,2]*k3 + coef[6,3]*k4 + coef[6,4]*k5 + coef[6,5]*k6; % fifth order solution
    }
    variable scale = 0.9*(eps/max(abs(dx_c-dx_e)))^(0.2); % 0.9 is safety factor; derivation see Numerical Recipes, Section "Adaptive Stepsize Control for Runge-Kutta"
    if(dt_fixed) scale = 1; % use fixed stepsize when qualifier "dt" is present
    % variable scale = ( 0.5*eps*dt/max(abs(dx_c-dx_e)) )^0.25;  % "derivation of this formula can be found in advanced books on numerical analysis"
    if(scale>=1 or dt<=1e-14*t) % scale>1: accept stepsize only if correction factor is equal/larger than 1
    {                           % dt<=abs(1e-14*t) in order to avoid infinite loops due to Double_Type precision limit at dt=1e-16*t
      if(disk.radius>0)
      {
	variable m_last = @(m[*,ind]);
	variable t_last = @t;
      }
      m[*,ind] += dx_c;
      t += dt;
      % ---------------------
      % update Rmin and Rmax:
      variable temp_R;
      if(coords=="cart")
	temp_R = sqrt((m[0,ind])^2+(m[1,ind])^2+(m[2,ind])^2); % temp_R = R = sqrt(x^2+y^2+z^2)
      else % coords = "cyl"
	temp_R = sqrt((m[0,ind])^2+(m[2,ind])^2); % temp_R = R = sqrt(r^2+z^2)
      temp0 = where(temp_R>Rmax[ind]);
      Rmax[ind[temp0]] = temp_R[temp0];
      temp0 = where(temp_R<Rmin[ind]);
      Rmin[ind[temp0]] = temp_R[temp0];
      % ---------------------
      if(disk.radius>0)
      {
	% identify orbits that have crossed the disk:
	variable ind_cro, ind_con; % indices of orbits which have crossed plane and of those for which calculation is continued
	if(coords=="cart") % (z-z_disk) * (z_last-z_disk) <= 0 but z!=z_last!=0 and r<=disk.radius -> orbit crossed disk inside radius of disk.radius during last step:
	  ind_cro = where((m[2,ind]-disk_positions[ind]) * (m_last[2,*]-disk_positions[ind]) <= 0 and m[2,ind]!=m_last[2,*] and
			  ( (m_last[0,*]-disk.x)^2 + (m_last[1,*]-disk.y)^2 <= disk.radius^2 or (m[0,ind]-disk.x)^2 + (m[1,ind]-disk.y)^2 <= disk.radius^2));
	else % coords = "cyl"
	  ind_cro = where((m[2,ind]-disk_positions[ind]) * (m_last[2,*]-disk_positions[ind]) <= 0 and m[2,ind]!=m_last[2,*] and ( (m_last[0,*]*cos(m_last[1,*])-disk.x)^2 + (m_last[0,*]*sin(m_last[1,*])-disk.y)^2 <= disk.radius^2 or
																  (m[0,ind]*cos(m[1,ind])-disk.x)^2 + (m[0,ind]*sin(m[1,ind])-disk.y)^2 <= disk.radius^2));
	disk_crossings[ind[ind_cro]] += 1;
	ind_cro = where(disk_crossings[ind]==disk.crossings, &ind_con);
	variable len_ind_cro = length(ind_cro);
	if(len_ind_cro>0)
	{
	  % =====================================================================
	  % interpolate to moment of crossing and save the result in m and t_end:
	  variable t_travel_last = (disk_positions[ind[ind_cro]]-m_last[2,ind_cro])/m_last[5,ind_cro]; % t_travel_last := t_cross - t_last = (disk_positions - z_last)/vz_last <=> t_travel_last*vz_last  = disk_positions - z_last <= linear interpolation
	  variable t_travel = (m[2,ind[ind_cro]]-disk_positions[ind[ind_cro]])/m[5,ind[ind_cro]]; % t_travel = t - t_cross = (z-disk_positions)/vz <=> vz*t_travel = z - disk_positions <= linear interpolation
	  % average with weights: the larger the estimated travel time, the worse the linear approximation
	  variable weight = t_travel / ( t_travel + t_travel_last );
	  t_eoc[ind[ind_cro]] = weight*t_last + (1.-weight)*t;
	  variable t_remaining = Double_Type[6,len_ind_cro];
	  t_remaining[0,*] = t_eoc[ind[ind_cro]] - t_last;
	  t_remaining[1,*] = t_remaining[0,*];
	  t_remaining[2,*] = t_remaining[0,*];
	  t_remaining[3,*] = t_remaining[0,*];
	  t_remaining[4,*] = t_remaining[0,*];
	  t_remaining[5,*] = t_remaining[0,*];
	  % re-calculate the last step but with adopted timestep t_remaining instead of dt
	  k1 = t_remaining * (@model)(s*t_last, m_last[*,ind_cro];; qualies);
	  k2 = t_remaining * (@model)(s*(t_last + coef[0,0]*t_remaining[0,*]), m_last[*,ind_cro] + coef[0,1]*k1;; qualies);
	  k3 = t_remaining * (@model)(s*(t_last + coef[1,0]*t_remaining[0,*]), m_last[*,ind_cro] + coef[1,1]*k1 + coef[1,2]*k2;; qualies);
	  k4 = t_remaining * (@model)(s*(t_last + coef[2,0]*t_remaining[0,*]), m_last[*,ind_cro] + coef[2,1]*k1 + coef[2,2]*k2 + coef[2,3]*k3;; qualies);
	  k5 = t_remaining * (@model)(s*(t_last + coef[3,0]*t_remaining[0,*]), m_last[*,ind_cro] + coef[3,1]*k1 + coef[3,2]*k2 + coef[3,3]*k3 + coef[3,4]*k4;; qualies);
	  k6 = t_remaining * (@model)(s*(t_last + coef[4,0]*t_remaining[0,*]), m_last[*,ind_cro] + coef[4,1]*k1 + coef[4,2]*k2 + coef[4,3]*k3 + coef[4,4]*k4 + coef[4,5]*k5;; qualies);
	  if(method == "RKDP") % for RKPD use fifth order solution to continue integration
	  {
	    dx_e = coef[5,1]*k1 + coef[5,2]*k3 + coef[5,3]*k4 + coef[5,4]*k5 + coef[5,5]*k6; % fourth order solution
	    k7 = t_remaining * (@model)(s*(t_last + coef[5,0]*t_remaining[0,*]), m_last[*,ind_cro] + dx_e;; qualies);
	    dx_c = coef[6,1]*k1 + coef[6,2]*k3 + coef[6,3]*k4 + coef[6,4]*k5 + coef[6,5]*k6 + coef[6,6]*k7; % fifth order solution
	  }
	  else % for RKF, RKCK use fourth order solution to continue integration
	  {
	    dx_c = coef[5,1]*k1 + coef[5,2]*k3 + coef[5,3]*k4 + coef[5,4]*k5 + coef[5,5]*k6; % fourth order solution
	    dx_e = coef[6,1]*k1 + coef[6,2]*k3 + coef[6,3]*k4 + coef[6,4]*k5 + coef[6,5]*k6; % fifth order solution
	  }
	  m[*,ind[ind_cro]] = m_last[*,ind_cro] + dx_c;
	  % =====================================================================
	}
      }
      if(set)
      {
	if(t-t_last_save > set_threshold or t==t_end)
	{
	  t_last_save = t;
	  variable temp_energy = (@model)(s*t, m[*,ind];; qualies_energy);
	  _for i(0, length(ind)-1, 1)
	  {
	    list_append((get_struct_field(ret.tr, cols[ind[i]])).t, t);
	    list_append((get_struct_field(ret.tr, cols[ind[i]])).x, m[0,ind[i]]);
	    list_append((get_struct_field(ret.tr, cols[ind[i]])).y, m[1,ind[i]]);
	    list_append((get_struct_field(ret.tr, cols[ind[i]])).z, m[2,ind[i]]);
	    list_append((get_struct_field(ret.tr, cols[ind[i]])).vx, m[3,ind[i]]);
	    list_append((get_struct_field(ret.tr, cols[ind[i]])).vz, m[5,ind[i]]);
	    list_append((get_struct_field(ret.tr, cols[ind[i]])).Lz, m[4,ind[i]]);
	    list_append((get_struct_field(ret.tr, cols[ind[i]])).Energy, temp_energy[i]);
	  }
	}
      }
      if(disk.radius>0)
      {
	if(len_ind_cro>0)
	{
	  % continue orbit calculation only for those orbits in m that did not cross the disk yet:
	  ind = ind[ind_con];
	}
      }
      if(verbose) vmessage("t=%g", s*t);
    }
    dt = _max(dt*scale, t*1e-15); % abs(1e-15*t) in order to avoid infinite loops due to Double_Type precision limit at dt=1e-16*t
    if(t+dt > t_end) dt = t_end-t;
  }
  % =======================================
  ret.i.vx /= (s*kms_to_kpcMyr); % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
  ret.i.vy /= (s*kms_to_kpcMyr); % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
  ret.i.vz /= (s*kms_to_kpcMyr); % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
  ret.i.Lz *= s; % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components

  if(set)
  {
    _for i(0, len-1, 1)
    {
      temp1.t = list_to_array((get_struct_field(ret.tr, cols[i])).t, Double_Type);
      temp1.t[-1] = t_eoc[i];
      temp1.t *= s; % for backward integration, i.e. t_end < 0, re-reverse t_end
      temp1.Energy = list_to_array((get_struct_field(ret.tr, cols[i])).Energy, Double_Type);
      if(coords=="cart")
      {
	temp1.x = list_to_array((get_struct_field(ret.tr, cols[i])).x, Double_Type);
	temp1.y = list_to_array((get_struct_field(ret.tr, cols[i])).y, Double_Type);
	temp1.z = list_to_array((get_struct_field(ret.tr, cols[i])).z, Double_Type);
	temp1.vx = list_to_array((get_struct_field(ret.tr, cols[i])).vx, Double_Type);
	temp1.vy = list_to_array((get_struct_field(ret.tr, cols[i])).Lz, Double_Type);
	temp1.vz = list_to_array((get_struct_field(ret.tr, cols[i])).vz, Double_Type);
	temp1.Lz = temp1.x*temp1.vy - temp1.y*temp1.vx; % Lz = x*vy - y*vx
      }
      else % coords = "cyl"
      {
	variable temp2 = struct{ r, phi, z, vr, vphi, vz };
	temp2.r = list_to_array((get_struct_field(ret.tr, cols[i])).x, Double_Type);
	temp2.phi = list_to_array((get_struct_field(ret.tr, cols[i])).y, Double_Type);
	temp2.z = list_to_array((get_struct_field(ret.tr, cols[i])).z, Double_Type);
	temp2.vr = list_to_array((get_struct_field(ret.tr, cols[i])).vx, Double_Type);
	temp1.Lz = list_to_array((get_struct_field(ret.tr, cols[i])).Lz, Double_Type);
	temp2.vphi = @temp1.Lz; ind = where(temp2.r != 0); temp2.vphi[ind] /= ((temp2.r)^2)[ind];
	temp2.vz = list_to_array((get_struct_field(ret.tr, cols[i])).vz, Double_Type);
	(temp1.x, temp1.y, temp1.z, temp1.vx, temp1.vy, temp1.vz) =
	  cyl2cart(temp2.r, temp2.phi, temp2.z, temp2.vr, temp2.vphi, temp2.vz);
      }
      temp1.Lz *= s; % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components by multiplying s
      temp1.vx /= (s*kms_to_kpcMyr); % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
      temp1.vy /= (s*kms_to_kpcMyr); % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
      temp1.vz /= (s*kms_to_kpcMyr); % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
      set_struct_field(ret.tr, cols[i], @temp1);
    }
  }

  ret.f = struct{ t, x, y, z, vx, vy, vz, Lz, Energy, Rmin, Rmax };
  ret.f.t = s*t_eoc; % for backward integration, i.e. t_end < 0, re-reverse t_end and the sign of the velocity components
  if(coords=="cart")
  {
    ret.f.x = m[0,*]; ret.f.y = m[1,*]; ret.f.z = m[2,*]; ret.f.vx = m[3,*]; ret.f.vy = m[4,*]; ret.f.vz = m[5,*];
    ret.f.Lz = s*(ret.f.x*ret.f.vy - ret.f.y*ret.f.vx); % Lz = x*vy - y*vx
  }
  else % coords = "cyl"
  {
    variable temp_vphi = @m[4,*]; ind = where( m[0,*] != 0 ); temp_vphi[ind] /= (m[0,ind])^2;
    (ret.f.x, ret.f.y, ret.f.z, ret.f.vx, ret.f.vy, ret.f.vz) = cyl2cart(m[0,*], m[1,*], m[2,*], m[3,*], temp_vphi, m[5,*]);
    ret.f.Lz = s*m[4,*]; % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
    ret.f = struct_combine(ret.f, "Nrev");
    ret.f.Nrev = s*(m[1,*] - phi_start)/(2*PI);
  }
  ret.f.Energy = (@model)(s*t, m;; qualies_energy);
  ret.f.vx /= (s*kms_to_kpcMyr); % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
  ret.f.vy /= (s*kms_to_kpcMyr); % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
  ret.f.vz /= (s*kms_to_kpcMyr); % for backward integration, i.e. t_end < 0, re-reverse the sign of the velocity components
  ret.f.Rmin = Rmin;
  ret.f.Rmax = Rmax;

  if(qualifier_exists("stff")) % some keywords for fits-header to know what input parameters or qualifiers were used for this computation
  {
    keys = struct_combine( keys, struct{ vxs, vys, vzs, Mb, Md, Mh, bb, ad, bd, ah, exponent, cutoff, model, method, eps}, disk );
    variable field;
    foreach field (["vxs", "vys", "vzs", "Mb", "Md", "Mh", "bb", "ad", "bd", "ah", "exponent", "cutoff"])
    {
      if(qualifier_exists(field))
	set_struct_field(keys, field, string(median(qualifier(field))));
      else
	set_struct_field(keys, field, "Default value");
    }
    keys.model = string(model);
    keys.method = method;
    keys.eps = string(eps);
    if(__is_initialized(&seed))
      keys = struct_combine( keys, struct{ random_seed = string(seed) } );
    if(qualifier_exists("dt"))
      keys = struct_combine( keys, struct{ dt = string(qualifier("dt")) } );
    fits_write_binary_table(qualifier("stff")+"_initial.fits", "Initial_conditions", ret.i, keys);
    fits_write_binary_table(qualifier("stff")+"_final.fits", "Final_conditions", ret.f, keys);
  }

  return ret;
}%}}}

%
% Code from src/slang/date/orbitalphase.sl
%

define orbitalphase()
%!%+
%\function{orbitalphase}
%\synopsis{calculates an orbital phase at a given MJD time from an ephemeris}
%\usage{Double_Type phi = orbitalphase(Double_Type MJD, T0, P0, [Pdot, [Pddot ]]);
%\altusage{Double_Type phi = orbitalphase(Double_Type MJD, String_Type ephemeris);}
%}
%\description
%    \code{phi = (MJD-T0)/P - floor( (MJD-T0)/P );}
%    where P = P0 + ( Pdot/2. + Pddot*MJD/6. )*MJD ;
%    The string \code{ephemeris} can be used to refer to an internally
%    stored (T0, P) pair or, if available, (T0, P, Pdot, Pddot) quadruple.
%
%    If \code{orbitalphase} is called with an unknown \code{ephemeris},
%    the allowed values are shown by \code{get_ephemeris}.
%\qualifiers{
%\qualifier{numpulse}{return pulse number instead of phase}
%}
%\seealso{get_ephemeris}
%!%-
{
  variable MJD, T0, P, ephemeris=NULL, Pdot=0, Pddot=0;
  switch(_NARGS)
  { case 2: (MJD, ephemeris) = ();
      try { variable ephem = get_ephemeris(ephemeris); }
      catch AnyError: { return; }
      T0 = ephem.T0 ;
      P = ephem.P ;
      if(struct_field_exists(ephem, "Pdot"))   Pdot=ephem.Pdot ;
      if(struct_field_exists(ephem, "Pddot"))  Pddot=ephem.Pddot ;
  }
  { case 3: (MJD, T0, P) = ();              }
  { case 4: (MJD, T0, P, Pdot) = ();        }
  { case 5: (MJD, T0, P, Pdot, Pddot) = (); }
  { help(_function_name()); return; }

  MJD -= 1.*T0;
  variable phi = MJD/( P + (Pdot/2. + Pddot*MJD/6.)*MJD );
  if(qualifier_exists("numpulse"))  return phi;
  return phi-floor(phi);
}

%
% Code from src/slang/arrays/multi-dimensional/outer_product.sl
%
%%%%%%%%%%%%%%%%%%%%
define outer_product()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{outer_product}
%\synopsis{computes the outer product of a number of vectors}
%\usage{Double_Type P[] = outer_product(Double_Type f1[], f2[], ..., fn[]);}
%\description
%    The \code{n} arguments \code{f1}, \code{f2}, ..., \code{fn} have to be one-dimensional arrays.
%    The return value \code{P} is an \code{n}-dimensional array with
%    \code{P[i1, i2, ..., in] = f1[i1] * f2[i2] * ... fn[in]}.
%!%-
{
  if(_NARGS==0)
  { help(_function_name()); return; }

  variable n = _NARGS;
  variable arg = __pop_args(n);
  arg = array_map( Array_Type, &get_struct_field, arg, "value" );
  variable dim = array_map( Integer_Type, &length, arg );

  variable P = Double_Type[nint(prod(dim))];
  P[[0:dim[0]-1]] = arg[0];
  
  variable I, len, _len = dim[0];
  variable ii;
  _for ii ( 1, length(dim)-1 ){
    len = nint(prod(dim[[0:ii]]));
    I = [0:len-1];
    P[I] = P[I/dim[ii] mod _len] * arg[ii][I mod dim[ii]];
    _len = @len;
  }
  
  reshape(P, dim);
  return P;
}

% %%%%%%%%%%%%%%%%%%%%
% define outer_product_old()
% %%%%%%%%%%%%%%%%%%%%
% %!%+
% %\function{outer_product}
% %\synopsis{computes the outer product of a number of vectors}
% %\usage{Double_Type P[] = outer_product(Double_Type f1[], f2[], ..., fn[]);}
% %\description
% %    The \code{n} arguments \code{f1}, \code{f2}, ..., \code{fn} have to be one-dimensional arrays.
% %    The return value \code{P} is an \code{n}-dimensional array with
% %    \code{P[i1, i2, ..., in] = f1[i1] * f2[i2] * ... fn[in]}.
% %!%-
% {
%   if(_NARGS==0)
%   { help(_function_name()); return; }

%   variable n = _NARGS;
%   variable arg = __pop_args(n);
%   variable dim = Integer_Type[n];
%   variable d, total_length = 1;
%   _for d (0, n-1, 1)
%   { dim[d] = array_shape(arg[d].value)[0];
%     total_length *= dim[d];
%   }
%   variable P = Double_Type[total_length];
%   variable i = Integer_Type[n];
%   variable j;
%   _for j (0, total_length-1, 1)
%   { $1 = 1;
%     _for d (0, n-1, 1)
%       $1 *= arg[d].value[i[d]];
%     P[j] = $1;
%     _for d (n-1, 0, -1)
%     { if(i[d]<dim[d]-1)
%       {	i[d]++;
% 	break;
%       }
%       else
%         i[d] = 0;
%     }
%   }
%   reshape(P, dim);
%   return P;
% }

%
% Code from src/slang/arrays/multi-dimensional/outer_products.sl
%
define outer_products()
%!%+
%\function{outer_products}
%\synopsis{computes multi-dimensional arrays from one-dimensional ones}
%\usage{(Double_Type P1[], P2[], ...) = outer_products(Double_Type f1[], f2[], ..., fn[]);}
%\description
%    \code{P1[i1, i2, ..., in] = f1[i1];}\n
%    \code{P2[i1, i2, ..., in] = f2[i2];}\n
%    ...\n
%    \code{Pn[i1, i2, ..., in] = fn[i1];}\n
%
%    The return value \code{Pj} is calculated as outer product of \code{x1, x2, ..., xn},
%    where \code{xj} = \code{fj} and
%          \code{xi = [1, 1, ..., 1]} (same length as \code{fi}) for \code{i!=j}.
%\seealso{outer_product}
%!%-
{
  if(_NARGS==0)
  { help(_function_name()); return; }

  variable n = _NARGS;
  variable arg = __pop_args(n);
  variable ones = Array_Type[n];
  variable i, j;
  _for i (0, n-1, 1)
    ones[i] = Integer_Type[length(arg[i].value)] + 1;
  _for i (0, n-1, 1)
  { variable list = {};
    _for j (0, n-1, 1)
      if(i==j)
        list_append(list, arg[j].value);
      else
        list_append(list, ones[j]);
    outer_product( __push_list(list) );  % return value is left on stack
  }
}

%
% Code from src/misc/radio/overplot_clean_model.sl
%

% only activate if module gcontour could be loaded
#ifeval __get_reference("gcontour_compute")!=NULL

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% convert pixel to mas
private define convMAS(pix,ref_pix,del) {
  return (pix - ref_pix+0.5)*del*3.6e6;}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define overplot_clean_model() 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{overplot_clean_model}
%\synopsis{creates an overlay of an clean VLBI image and the modelfit components}
%\usage{overplot_clean_model(String_Type \code{cleanfile}, String_Type \code{modfile}, String_Type \code{outputfile});}
%\qualifiers{
%\qualifier{ra_mas}{[={20,-20}] {left,right} limits of image in mas}
%\qualifier{dec_mas}{[={-20,20}] {top,bottom} limits of image in mas}
%\qualifier{plot_size}{[=15] size of plot}
%\qualifier{n_sigma}{[=3.0] lowest contour of clean image}
%\qualifier{sourcename}{[=default] name of the source, by default the name is read from the .fits file,\n
% 				set to NULL for not plotting a source name}
%\qualifier{obs_date}{[=default] the observation date, by default the observation date is read from the .fits file, \n
%				set to NULL for not plotting a observation date, set to "mjd" when MJD format required}
%\qualifier{cont_color}{[="gray"] color of contours of clean image}
%\qualifier{cont_scl}{[="2"] set factor to change the separation between contour levels}
%\qualifier{cont_lvl}{[=[c1,c2,..] set contour levels [Jy] manually, overwrites other contour parameters}
%\qualifier{model_color}{[="black"] color of Gaussian ellipses}
%\qualifier{center_symbol}{[="+"] symbol stating the ellipse center, set to NULL for no symbol}
%\qualifier{comp_label}{[=0] set to 1 when labelling of components depending on distance from [0,0]
% 				is requested}
%\qualifier{ind_inverse}{[=0] if the labels should be in inverse order set to 1}
%\qualifier{label_alt}{[=NULL] give list of alternative labels}
%\qualifier{label_color}{[="red"] color of component labels}
%\qualifier{ex_counterjet}{[=0]	set to 1 if counterjet components (and core) should be excluded and define
% 				the corresponding quadrant of the plot via the CJ-coordinates}
%\qualifier{counterjet_ra}{[=NULL] define counterjet coordinates in mas}
%\qualifier{counterjet_dec}{[=NULL] define counterjet coordinates in mas}
%}
%\description
%    This function creates an overlay of a VLBI-clean image and the corresponding\n
%    model of Gaussian components which are over-plotted as ellipses.\n
%    The required input format are fits-file for both the clean and the modelfit images.
%    The format of the output file depends on the suffix of the given\n
%    \code{filename}. Possible formats of the output file are PDF, EPS,\n
%    PNG, GIF, etc.
%    If labelling = 1 is set, the components are labeled with J0,J1,J2,... depending on 
%    their distance to [0,0].
%    The counterjet components (and core) can be excluded by defining the corresponding 
%    quadrant of the the plot via counterjet_ra and _dec.
%!%-
{
variable cleanfile,modfile,outputfile;
  switch(_NARGS)
  { case 3: (cleanfile,modfile,outputfile) = (); }
  { help(_function_name()); return; }
%%%%%%%%%%%%
variable ra_mas		= qualifier("ra_mas", [20,-20]);
variable dec_mas	= qualifier("dec_mas", [-20,20]); 
variable nsigma    	= qualifier("n_sigma",3.0);
variable sourcename	= qualifier("sourcename", fits_read_key (cleanfile, "OBJECT"));
variable obs_date       = qualifier("obs_date", fits_read_key (cleanfile, "DATE-OBS") );
variable comp_label	= qualifier("comp_label", 0);
variable label_color	= qualifier("label_color", "red");
variable ex_counterjet	= qualifier("ex_counterjet",0);
variable counterjet_ra 	= qualifier("counterjet_ra", NULL);
variable counterjet_dec = qualifier("counterjet_dec", NULL);
variable ind_inverse	= qualifier("ind_inverse",0);
variable label_alt	= qualifier("label_alt", NULL);
%%%%
variable mjd = fits_read_key(cleanfile,"DATE-OBS");
variable Y, m, d;
()=sscanf(mjd, "%4d-%2d-%2d", &Y, &m, &d);
mjd = MJDofDate(Y, m, d);
if (obs_date == "mjd"){ obs_date = string(mjd);}
%%%
variable image		= fits_read_img(cleanfile)[0,0,*,*];
variable noise		= fits_read_key(cleanfile,  "NOISE");
variable min_treshhold 	= 1e-9;
variable rms_image 	= nsigma*noise;
% %find reference pixel in map and create array:
variable  pix1 = fits_read_key(cleanfile,"CRPIX1");
variable  pix2 = fits_read_key(cleanfile,"CRPIX2");
variable  delta1 = fits_read_key(cleanfile,"CDELT1");
variable  delta2 = fits_read_key(cleanfile,"CDELT2");
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
variable model = fits_read_table(modfile);
variable factor = 3.6e6;
variable RA = model.deltax*factor;
variable Decl = model.deltay*factor;
%% Counterjet exclusion:
%% The user defines the quadrant where the CJ and core components are located.
%% This quadrant is excluded in the following calculations.
variable excl;
if (ex_counterjet==1){
	%1.Quadrant:
	if (max(counterjet_ra)<=0 and (min(counterjet_dec)>=0)){
% 	theta = +0-90
	excl = wherenot(RA<=0.0 and Decl>=0);
	}
	%2.Quadrant:
	if (min(counterjet_ra)>=0 and (min(counterjet_dec)>=0)){
% 	theta = -0-90% 	
	excl = wherenot(RA>=0.0 and Decl>=0);	
	}
	%3.Quadrant:
	if (min(counterjet_ra)>=0 and (max(counterjet_dec)<=0)){
% 	theta -90-180
	excl = wherenot(RA>=0.0 and Decl<=0);
	}
	%4.Quadrant:
	if (max(counterjet_ra)<=0 and (max(counterjet_dec)<=0)){
% % 	theta = +90-180
	excl = wherenot(RA<=0.0 and Decl<=0);
	}
model = struct_filter (model,excl;copy);
}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% add field for distance and sort components by distance
model = struct_combine (model, struct{distance=Double_Type[length(model.flux)]});
%%%%
variable distance = sqrt((model.deltax*factor)^2+(model.deltay*factor)^2);
variable a;
for (a=0; a<length(model.flux); a++) {model.distance[a]=distance[a];};

variable model_sort = struct_filter(model, array_sort(model.distance);copy);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
image[where(image < min_treshhold)] = min_treshhold;
variable x, y;
(x,y)= get_coordinatearrays_of_image(image);
(x,y)=(convMAS(x,pix1,delta1),convMAS(y,pix2,delta2));
variable xplotmin = ra_mas[0]; variable xplotmax = ra_mas[1];
variable yplotmin = dec_mas[0]; variable yplotmax = dec_mas[1];
variable xmin =  wherelast (x[0,*]>xplotmin);
variable xmax =  wherefirst (x[0,*]<xplotmax);
variable ymin =  wherelast(y[*,0] < yplotmin);
variable ymax =  wherefirst(y[*,0]> yplotmax);
image   = image [[ymin:ymax],[xmin:xmax]];
x   = x [[ymin:ymax],[xmin:xmax]];
y   = y [[ymin:ymax],[xmin:xmax]];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % ellipses (in new order!)
variable smaj = factor*model_sort.major_ax;
variable smin = factor*model_sort.minor_ax;
variable posangle = (model_sort.posangle/180.) * PI;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% fit background noise (for details see plot_vlbi_map)
    variable lo,hi;
    variable map = image;
    variable map_min = min(map);
    variable map_max = max(map);
    (lo,hi) = log_grid(1e-5,map_max+(1e-5-map_min),500);
    variable nr = histogram (map-map_min+1e-5,lo,hi);
    
    variable px_dat = define_counts(lo,hi,nr,sqrt(nr)+1);
    fit_fun("gauss");
    % guessing good start parameters for the fit from the statistics of the map
    % works well if only a small region of the image is jet emission
    set_par("gauss(1).center",median(map-map_min+1e-5)        , 0 , 1e-6, 10           );
    set_par("gauss(1).area"  ,0.99*length(map)                , 0 , 1   , 2*length(map));
    set_par("gauss(1).sigma" ,quantile (0.84, _reshape(map-map_min+1e-5,length(dup))) - get_par("gauss(1).center")
	    , 0 , 1e-6, 1 );
    exclude(all_data);
    include(px_dat); % other methods/checks to find best fit?
    variable stat, fitmeth = get_fit_method;
    ()=fit_counts(&stat);
    if ( stat.statistic/(stat.num_bins-stat.num_variable_params) > 8)
    {
      if (Fit_Verbose >= 0) vmessage("Reduced chi-square > 8\t\t--> fitting continued");
      set_fit_method ("subplex"); ()=fit_counts;
      set_fit_method ("mpfit"); ()=fit_counts;
      set_fit_method (fitmeth);
    }
    
    variable  mu    = get_par("gauss(1).center") +map_min-1e-5;
    variable  sigma = get_par("gauss(1).sigma" );
    delete_data(px_dat);
    variable clrcut       = qualifier("clrcut",1);
    variable clrmin  = qualifier("clrmin",clrcut ? mu+nsigma*sigma : map_min);
    variable clrmax  = qualifier("clrmax",map_max);

    variable cl    = qualifier("cont_scl", 2.);
    variable n_pos = int(log( (map_max - mu) / (nsigma*sigma)) / log(cl));
    variable n_neg = int(log( (mu - map_min) / (nsigma*sigma)) / log(cl));
    variable lvls  = qualifier("cont_lvl", mu + [-(cl^[0:n_neg]),cl^[0:n_pos]]*nsigma*sigma );


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % define plot 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
variable i,j;
variable width = qualifier("plot_size",15.);
variable height = width/abs(xplotmin-xplotmax)*abs(yplotmax-yplotmin);
variable plot = xfig_plot_new(width, height);
plot.world(xplotmin, xplotmax, yplotmin, yplotmax);	% world coordinates in mas
plot.xlabel("relative RA [mas]");
plot.ylabel("relative DEC [mas]");
variable major_ticks = [-100:100:5];
variable minor_ticks= [-100:100:1];
plot.x1axis(; major=major_ticks, minor= minor_ticks);
plot.y1axis(; major=major_ticks, minor= minor_ticks);

%% adding source and date
  if (sourcename != NULL){
    variable name_label = xfig_new_text(sourcename;color="black");
    name_label.scale(1.3);
    name_label.translate(vector(0.15*width,0.96*height,0));
    plot.add_object(name_label);
  }
  if (obs_date != NULL){
    variable date_label = xfig_new_text(obs_date;color="black");
    date_label.scale(1.3);
    date_label.translate(vector(0.85*width,0.96*height,0));
    plot.add_object(date_label);
  }
%% % labelling the components:
variable ii;
if (comp_label ==1 and label_alt!=NULL and length(label_alt)!=length(model_sort.distance)){print("number of labels does not match number of components");}
if (comp_label ==1 and label_alt!=NULL and length(label_alt)==length(model_sort.distance)){
for (ii=0; ii<length(label_alt); ii++){
		if (xplotmin >model_sort.deltax[ii]*factor> xplotmax and yplotmin <model_sort.deltay[ii]*factor< yplotmax){
		plot.xylabel(model_sort.deltax[ii]*factor, model_sort.deltay[ii]*factor, string(label_alt[ii]), 0.5, 0.5;color=qualifier("label_color","red"));};
	};}
variable index = [0:length(model_sort.distance)];
if (ind_inverse == 1){index = [(length(model_sort.distance)-1):0:-1];}
if (comp_label ==1 and label_alt==NULL){
	for (ii=0; ii<length(model_sort.distance); ii++){
		if (xplotmin >model_sort.deltax[ii]*factor> xplotmax and yplotmin <model_sort.deltay[ii]*factor< yplotmax){
		plot.xylabel(model_sort.deltax[ii]*factor, model_sort.deltay[ii]*factor, "J"+string(index[ii]), 0.5, 0.5;color=qualifier("label_color","red"));};
	};}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % plot contours of clean image
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



variable xref = pix1-xmin; variable yref = pix2-ymin;
% % determine contour levels

% % 
variable linewidth = Integer_Type[length(lvls)]; linewidth[*] = 1;
variable line = Integer_Type[length(lvls)]; line[*] = 0;
variable col = String_Type[length(lvls)]; col[*] = qualifier("cont_color","gray") ;
% % plot contours
 variable gcl = gcontour_compute(map, lvls);
_for i (0, length(gcl)-1, 1) {
  _for j (0, length(gcl[i].x_list)-1, 1) {
    plot.plot(convMAS(gcl[i].x_list[j],xref,delta1),
	     convMAS(gcl[i].y_list[j],yref,delta2);
	     width=linewidth[i], line=line[i], color=col[i]);
  }
}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % overplot Gaussian modelcomponents
variable phi=[0:2*PI:#200]; 
variable ex, ey;
_for i (0, length(model_sort.deltax)-1,1){
(ex, ey) = (ellipse(smaj[i]/2.0,smin[i]/2.0,PI/2.0-posangle[i],phi) );
plot.plot(model_sort.deltax[i]*factor,model_sort.deltay[i]*factor; sym=qualifier("center_symbol", "+"), symcolor=qualifier("model_color","black")); 
plot.plot(model_sort.deltax[i]*factor+ex,model_sort.deltay[i]*factor+ey) ;
}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % plot beam
variable beam_maj      = fits_read_key (cleanfile, "BMAJ")*(3.6e+6)/2.0; % clean beam major axis in mas (semi major!)
variable beam_min      = fits_read_key (cleanfile, "BMIN")*(3.6e+6)/2.0; % clean beam minor axis in mas (semi minor!)
variable beam_ang      = fits_read_key (cleanfile, "BPA")/180*PI; % clean beam postion angle

variable beam = xfig_new_ellipse(beam_min,beam_maj);  % major axis to north, pos_angle = 0
variable beam_clr = "gray";
beam.set_pen_color(beam_clr);
beam.rotate(vector(0,0,1),beam_ang); % rotation for position angle in rad ("rotate" rotates clock wise)
beam.scale((1.*width) /abs(-convMAS(xmin,pix1,delta1)+convMAS(xmax,pix1,delta1)) ,
	   (1.*height)/abs(-convMAS(ymin,pix2,delta2)+convMAS(ymax,pix2,delta2)) , 0); 
beam.translate(vector(1.5-beam_maj*width/(xmax-xmin),1.5+beam_maj*height/(ymax-ymin),0));
beam.area_fill=20;
beam.fill_color=xfig_lookup_color(beam_clr);
plot.add_object(beam);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
xfig_new_hbox_compound (plot, 0.4).render(outputfile);
%%
}


#endif

#endif

#endif

%
% Code from src/slang/pack_obj.sl
%
% pack format from input object type
private define _pack_obj_fmt(obj) {
  switch (typeof(obj))
    { case Int16_Type:  return "j"; } % 16bit == [Short_Type]
    { case Int32_Type:  return "k"; } % 32bit == [Integer_Type, Int_Type]
    { case Int64_Type:  return "q"; } % 64bit == [Long_Type]
    { case UInt16_Type: return "J"; } % 16bit
    { case UInt32_Type: return "K"; } % 32bit == [UInteger_Type, UInt_Type]
    { case UInt64_Type: return "Q"; } % 64bit == [ULong_Type, ULLong_Type]
    { case Char_Type:   return "c"; }
    { case UChar_Type:  return "C"; }
    { case Float_Type:  return "f"; } % 32bit == [Float32_Type]
    { case Double_Type: return "d"; } % 64bit == [Float64_Type]
    { case String_Type: return sprintf("s%d", strlen(obj)); }
  return NULL;
}
% object type from input pack format
private define _unpack_obj_fmt(fmt) {
  switch (fmt)
    { case "j": return Int16_Type; }
    { case "k": return Int32_Type; }
    { case "q": return Int64_Type; }
    { case "J": return UInt16_Type; }
    { case "K": return UInt32_Type; }
    { case "Q": return UInt64_Type; }
    { case "c": return Char_Type; }
    { case "C": return UChar_Type; }
    { case "f": return Float_Type; }
    { case "d": return Double_Type; }
    { case "s": return String_Type; }
  return NULL;
}

define pack_obj();
%%%%%%%%%%%%%%%%%%%%
define pack_obj()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pack_obj}
%\synopsis{converts an SLang object into a binary string}
%\usage{BString_Type[] pack_obj(Any_Type obj);}
%\description
%    Uses the `pack' function to convert a single SLang object into
%    an array of binary strings (BString_Type). The format specifier
%    is included as first character of the strings.
%    
%    If the object is a single number or string then the returned
%    array consists of a single item only. In case the given object
%    is an array or a structure, the returned array starts with the
%    type of the input array and its length or the field structure
%    definition, respectively. The remaining items are the input
%    array items or the structure items. For structures, the
%    contained objects are converted recursively.
%
%    The following data-types are supported:
%    Int16_Type, Int32_Type, Int64_Type, Float_Type, Double_Type,
%    Char_Type, String_Type, Array_Type, Struct_Type
%    and its aliases.
%\example
%#v+
%#p+
%    s = pack_obj(PI);        % s[0] = "d\030-DT\373!\011@"
%
%    s = pack_obj([1:3] + 4); % s[0] = "ak\003\000\000\000"
%                             % s[1] = "\005\000\000\000"
%                             % s[2] = "\006\000\000\000"
%                             % s[3] = "\007\000\000\000"
%
%    variable obj = struct {
%      number = 598105,
%      float  = 341.12e8,
%      more   = struct {
%        msg    = "hello"
%      }
%    };
%    s = pack_obj(obj);
%    print(s);
%#p-
%#v-
%\seealso{unpack_obj, pack, unpack}
%!%-
{
  variable obj;
  switch (_NARGS)
    { case 1: obj = (); }
    { help(_function_name); return; }

  % check input type
  variable type = typeof(obj), fmt, s, len, i;
  switch (type)
  % array type
  { case Array_Type:
    % item type and its format
    type = _typeof(obj);
    fmt = _pack_obj_fmt(obj[0]);
    if (fmt == NULL) {
      vmessage("error(%s): array of %s not supported", _function_name, string(type));
      return;
    }
    % array length
    len = length(obj);
    % init string array
    s = BString_Type[len+1];
    % array specification in first item: array indicator 'a', item format, length
    s[0] = pack("ccK", 'a', fmt[0], len);
    % pack all items
    _for i (1, len, 1) {
      s[i] = type == String_Type ? obj[i-1] : pack(fmt, obj[i-1]);
    }
  }
  % struct type
  { case Struct_Type:
    % fieldnames and number of fields
    variable fields = get_struct_field_names(obj);
    len = length(fields);
    % init string array with field names
    s = BString_Type[len+1];
    % struct field specification in first item: struct indicator 't', (empty), number of fields
    s[0] = pack("cK", 't', len);
    % insert field names
    s[[1:]] = fields;
    % recursively add field values
    _for i (0, len-1, 1) {
      s = [s, pack_obj(get_struct_field(obj, fields[i]))];
    }
  }
  % all other types, i.e., basic items
  {
    fmt = _pack_obj_fmt(obj);
    s = [pack(sprintf("c%s", fmt), fmt[0], obj)];
  }
  
  return s;
}


define unpack_obj();
%%%%%%%%%%%%%%%%%%%%
define unpack_obj()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{unpack_obj}
%\synopsis{converts a binary string back into an SLang object}
%\usage{Any_Type pack_obj(BString_Type obj);}
%\description
%    An input array of binary strings (BString_Type), which has
%    been created by the `pack_obj' function, is converted back
%    into an SLang object using the `unpack' function.
%
%    See the documentation of `pack_obj` for more details.
%\example
%#v+
%#p+
%    s = pack_obj(PI);        % s[0] = "d\\030-DT\\373!\\011@"
%
%    a = unpack_obj(s);       % a = 3.141592653589793;
%#p-
%#v-
%\seealso{pack_obj, unpack, pack}
%!%-
{
  variable s; % input array of strings
  variable q = 0;  % current item index of s
  variable p = &q; % and its pointer
  % process arguments
  switch (_NARGS)
    { case 1: s = (); }
    { case 2: (s,p) = (); }
    { help(_function_name); return; }

  % sanity checks
  if (typeof(s) != Array_Type) { s  = [s]; }
  if (_typeof(s) != BString_Type) {
    vmessage("error(%s): input has to be of BString_Type", _function_name);
    return;
  }
  variable obj, fmt, len;

  % extract format specification (first character)
  fmt = substr(s[@p], 1, 1);

  switch (fmt)
  % array
  { case "a":
    % format and length of the array
    (,fmt,len) = unpack("ccK", s[@p]);
    (@p)++;
    fmt = sprintf("%c", fmt);
    variable form = _unpack_obj_fmt(fmt);
    variable i;
    % init array
    obj = form[len];
    % fill array
    _for i (0, len-1, 1) {
      obj[i] = unpack_obj(fmt + s[@p+i]);
    }
    (@p) += len;
  }
  % struct
  { case "t":
    % get struct field names
    (,len) = unpack("cK", s[@p]);
    (@p)++;
    variable fields = s[[@p:@p+len-1]];
    (@p) += len;
    % init struct
    obj = @Struct_Type(fields);
    % fill struct
    _for i (0, length(fields)-1, 1) {
      set_struct_field(obj, fields[i], unpack_obj(s, p));
    }
    return obj;
  }
  % string
  { case "s":
    obj = substr(s[@p], 2, -1);
    (@p)++;
  }
  % all the other formats, i.e., basic formats
  {
    (,obj) = unpack(sprintf("c%s", fmt), s[@p]);
    (@p)++;
  }

  return obj;
}

%
% Code from src/slang/color/palettes.sl
%
% -*- mode: fold; mode: slang; -*-

private variable ColorPalettes = Assoc_Type[Any_Type, [0x000000, 0xff0000, 0x00ff00, 0x0000ff, 0xffff00, 0xff00ff, 0x00ffff, 0xffffff]];
define add_color_palette (name, values)
%!%+
%\function{add_color_palette}
%\synopsis{Add a color palette}
%\usage{add_color_palette (String_Type name, UInt_Type[] colors);
%\altusage{add_color_palette (String_Type name, Ref_Type generator);}}
%\description
%  Add an array of hex values representing a color palette to the
%  system.
%
%  Alternatively, a function can be passed with this signature
%  \code{UInt_Type[] palette = generator(UInt_Type n);}
%  where \code{n} is the number of colors returned. It is okay to
%  return fewer colors than \code{n}, but never more.
%\seealso{get_color_palette}
%!%-
{
  ColorPalettes[name] = values;
}

% seaborn colors %{{{
add_color_palette("sb:deep", [0x4c72b0, 0xdd8452, 0x55a868, 0xc44e52, 0x8172b3, 0x937860,
			      0xda8bc3, 0x8c8c8c, 0xccb974, 0x64b5cd]);
add_color_palette("sb:muted", [0x4878d0, 0xee854a, 0x6acc64, 0xd65f5f, 0x956cb4, 0x8c613c,
			       0xdc7ec0, 0x797979, 0xd5bb67, 0x82c6e2]);
add_color_palette("sb:pastel", [0xa1c9f4, 0xffb482, 0x8de5a1, 0xff9f9b, 0xd0bbff, 0xdebb9b,
			       0xfab0e4, 0xcfcfcf, 0xfffea3, 0xb9f2f0]);
add_color_palette("sb:bright", [0x023eff, 0xff7c00, 0x1ac938, 0xe8000b, 0x8b2be2, 0x9f4800,
				0xf14cc1, 0xa3a3a3, 0xffc400, 0x00d7ff]);
add_color_palette("sb:dark", [0x001c7f, 0xb1400d, 0x12711c, 0x8c0800, 0x591e71, 0x592f0d,
			      0xa23582, 0x3c3c3c, 0xb8850a, 0x006374]);
add_color_palette("sb:colorblind", [0x0173b2, 0xde8f05, 0x029e73, 0xd55e00, 0xcc78bc, 0xca9161,
				    0xfbafe4, 0x949494, 0xece133, 0x56b4e9]);
%!%+
%\function{palette--sb}
%\synopsis{Get Seaborn color paletts}
%\usage{UInt_Type[] get_color_paletts("sb:"+name, UInt_Type n);}
%\description
%  The seaborn paletts are a very good set of colors for qualitative
%  representation of data (like lines, points, etc.).
%!%-

%}}}

private define hsl_palette (n)
%!%+
%\function{palette--hsl}
%\synopsis{Generate palette from HSL space}
%\usage{UInt_Type[] palette = get_color_palette ("hsl", UInt_Type n);}
%\qualifiers{
%\qualifier{h[=0.01]}{Starting hue}
%\qualifier{s[=0.65]}{Saturation}
%\qualifier{l[=0.60]}{Lightness}
%}
%\description
%  Create a color palette with \code{n} entries. Control the
%  appearence with the HSL qualifiers.
%\seealso{hsl2rgb}
%!%-
{
  variable hues = ([0:1:#n]+qualifier("h", 0.01)) mod 1.0;
  hues -= int(hues);

  return rgb2hex(hsl2rgb(hues, qualifier("s", 0.65), qualifier("l", 0.6)));
}

private define hsluv_palette (n)
%!%+
%\function{palette--hsluv}
%\synopsis{Generate palette from HSLuv space}
%\usage{UInt_Type[] palette = get_color_palette ("hsluv", UInt_Type n);}
%\qualifiers{
%\qualifier{h[=0.01]}{Starting hue}
%\qualifier{s[=0.90]}{Saturation}
%\qualifier{l[=0.60]}{Lightness}
%}
%\description
%  Create a color palette with \code{n} entries. Control the
%  appearence with the HSL qualifiers.
%\seealso{hsluv2rgb}
%!%-
{
  variable hues = ([0:1:#n]+qualifier("h", 0.01)) mod 1.0;
  hues -= int(hues);

  return rgb2hex(hsluv2rgb(hues, qualifier("s", 0.9), qualifier("l", 0.6)));
}

private define hpluv_palette (n)
%!%+
%\function{palette--hpluv}
%\synopsis{Generate palette from HSLuv space}
%\usage{UInt_Type[] palette = get_color_palette ("hpluv", UInt_Type n);}
%\qualifiers{
%\qualifier{h[=0.01]}{Starting hue}
%\qualifier{s[=0.90]}{Saturation}
%\qualifier{l[=0.60]}{Lightness}
%}
%\description
%  Create a color palette with \code{n} entries. Control the
%  appearence with the HSL qualifiers.
%\seealso{hpluv2rgb}
%!%-
{
  variable hues = ([0:1:#n]+qualifier("h", 0.01)) mod 1.0;
  hues -= int(hues);

  return rgb2hex(hsluv2rgb(hues, qualifier("s", 0.9), qualifier("l", 0.6)));
}

% HSL, HSLuv palettes %{{{
add_color_palette("hsl", &hsl_palette);
add_color_palette("hsluv", &hsluv_palette);
add_color_palette("hpluv", &hpluv_palette);
%}}}

% SRON (Paul Tol) palettes %{{{
private define sron_palette (n)
%!%+
%\function{palette--sron-rainbow}
%\synopsis{Get SRON rainbow palette from Paul Tol}
%\usage{UInt_Type[] palette = get_color_palette("sron-rainbow", UInt_Type n);}
%\description
%  Returns a set of colors more or less like rainbow colors with good contrast.
%  \code{n} makes a smart subset up to 23.
%!%-
{
  variable colors = [0xe8ecfb, 0xd9cce3, 0xd1bbd7, 0xcaaccb, 0xba8db4, 0xae76a3,
		     0xaa6f9e, 0x994f88, 0x882e72, 0x1965b0, 0x437dbf, 0x5289c7,
		     0x6195cf, 0x7bafde, 0x4eb265, 0x90c987, 0xcae0ab, 0xf7f056,
		     0xf7cb45, 0xf6c141, 0xf4a736, 0xf1932d, 0xee8026, 0xe8601c,
		     0xe65518, 0xdc050c, 0xa5170e, 0x72190e, 0x42150a, 0x777777];
  variable selector = {
    [10],
    [10, 26],
    [10, 18, 26],
    [10, 15, 18, 26],
    [10, 14, 15, 18, 26],
    [10, 14, 15, 17, 18, 26],
    [9, 10, 14, 15, 17, 18, 26],
    [9, 10, 14, 15, 17, 18, 23, 26],
    [9, 10, 14, 15, 17, 18, 23, 26, 28],
    [9, 10, 14, 15, 17, 18, 21, 24, 26, 28],
    [9, 10, 12, 14, 15, 17, 18, 21, 24, 26, 28],
    [3, 6, 9, 10, 12, 14, 15, 17, 18, 21, 24, 26],
    [3, 6, 9, 10, 12, 14, 15, 16, 17, 18, 21, 24, 26],
    [3, 6, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26],
    [3, 6, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 28],
    [3, 5, 7, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 28],
    [3, 5, 7, 8, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 28],
    [3, 5, 7, 8, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 27, 28],
    [2, 4, 5, 7, 8, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 27, 28],
    [2, 4, 5, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 20, 22, 24, 26, 27, 28],
    [2, 4, 5, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 21, 23, 25, 26, 27, 28],
    [2, 4, 5, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 21, 23, 25, 26, 27, 28, 29],
    [1, 2, 4, 5, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 21, 23, 25, 26, 27, 28, 29]};

  if (n > length(selector))
    n = length(selector);

  return colors[selector[n-1]];
}
add_color_palette("sron-rainbow", &sron_palette);

add_color_palette("sron:bright", [0x4477aa, 0x66ccee, 0x228833, 0xccbb44, 0xee6677, 0xaa3377, 0xbbbbbb]);
add_color_palette("sron:high-contrast", [0xffffff, 0xddaa33, 0xbb5566, 0x004488, 0x000000]);
add_color_palette("sron:vibrant", [0x0077bb, 0x33bbee, 0x009988, 0xee7733, 0xcc3311, 0xee3377,
				   0xbbbbbb]);
add_color_palette("sron:muted", [0x332288, 0x88ccee, 0x44aa99, 0x117733, 0x999933, 0xddcc77,
				 0xcc6677, 0x882255, 0xaa4499, 0xdddddd]);
add_color_palette("sron:medium-contrast", [0xffffff, 0xeecc66, 0xee99aa, 0x6699cc, 0x997700, 0x994455,
					   0x004488, 0x000000]);
add_color_palette("sron:pale", [0xbbccee, 0xcceeff, 0xccddaa, 0xeeeebb, 0xffcccc, 0xdddddd]);
add_color_palette("sron:dark", [0x222255, 0x225555, 0x225522, 0x666633, 0x663333, 0x555555]);
add_color_palette("sron:light", [0x77aadd, 0x99ddff, 0x44bb99, 0xbbcc33, 0xaaaa00, 0xeedd88,
				 0xee8866, 0xffaabb, 0xdddddd]);
add_color_palette("sron:sunset", [0x364b9a, 0x4a7bb7, 0x6ea6cd, 0x98cae1, 0xc2e4ef, 0xeaeccc,
				  0xfeda8b, 0xfdb366, 0xf67e4b, 0xdd3d2d, 0xa50026, 0xffffff]);
add_color_palette("sron:burd", [0x2166ac, 0x4393c3, 0x92c5de, 0xd1e5f0, 0xf7f7f7, 0xfddbc7,
				0xf4a582, 0xd6604d, 0xb2182b, 0xffee99]);
add_color_palette("sron:prgn", [0x762a83, 0x9970ab, 0xc2a5cf, 0xe7d4e8, 0xf7f7f7, 0xd9f0d3,
				0xacd39e, 0x5aae61, 0x187837, 0xffee99]);
add_color_palette("sron:ylorbr", [0xffffe5, 0xfff7bc, 0xfee391, 0xfec44f, 0xfb9a29, 0xec7014,
				  0xcc4c02, 0x993404, 0x662506, 0x888888]);
add_color_palette("sron:iridescent", [0xfefbe9, 0xfcf7d5, 0xf5f3c1, 0xeaf0b5, 0xddecbf, 0xd0e7ca,
				      0xc2e3d2, 0xb5ddd8, 0xa8d8dc, 0x9bd2e1, 0x8dcbe4, 0x81c4e7,
				      0x7bbce7, 0x7eb2e4, 0x88a5dd, 0x9398d2, 0x9b8ac4, 0x9d7db2,
				      0x9a709e, 0x906388, 0x805770, 0x684957, 0x46353a, 0x999999]);
add_color_palette("sron:smooth-rainbow", [0xe8ecfb, 0xddd8ef, 0xd1c1e1, 0xc3a8d1, 0xb58fc2, 0xa778b4,
					  0x9b62a7, 0x8c4e99, 0x6f4c9b, 0x6059a9, 0x5568b8, 0x4e79c5,
					  0x4d8ac6, 0x4e96bc, 0x549eb3, 0x59a5a9, 0x60ab9e, 0x69b190,
					  0x77b77d, 0x8cbc68, 0xa6be54, 0xbebc48, 0xd1b541, 0xddaa3c,
					  0xe49c39, 0xe78c35, 0xe67932, 0xe4632d, 0xdf4828, 0xda2222,
					  0xb8221e, 0x95211b, 0x721e17, 0x521a13, 0x666666]);
add_color_palette("sron:landcover", [0x5566aa, 0x117733, 0x44aa66, 0x55aa22, 0x668822, 0x99bb55,
				     0x558877, 0x88bbaa, 0xaaddcc, 0x44aa88, 0xddcc66, 0xffdd44,
				     0xffee88, 0xbb0011]);
%!%+
%\function{palette--sron}
%\synopsis{Get SRON palettes}
%\usage{UInt_Type[] get_color_palette("sron:"+name, UInt_Type n);}
%\description
%  The SRON colors are defined in Paul Tol's technical note (\code{https://personal.sron.nl/~pault/data/colourschemes.pdf})
%  and provides a set of colors for qualitative, diverging, and sequential
%  representations.
%
%  Note: To get the original set of colors (was provided by \code{get_sron_colors})
%  use \code{get_color_palette("sron-rainbow", n);} for diverging and sequential
%  there are no direct matches. See the other color palette options.
%
%\seealso{get_color_palette_names, palette--sron-rainbow}
%!%-
%}}}

% Tableau10 palettes %{{{
add_color_palette("tab:10", [0x4e79a7, 0xf28e2b, 0xe15759, 0x76b7b2, 0x59a14f, 0xedc948,
			    0xb07aa1, 0xff9da7, 0x9c755f, 0xbab0ac]);
add_color_palette("tab:20", [0x4e79a7, 0xa0cbe8, 0xf28e2b, 0xffbe7d, 0x59a14f, 0x8cd17d,
			    0xb6992d, 0xf1ce63, 0x499894, 0x86bcb6, 0xe15759, 0xff9d9a,
			    0x79706e, 0xbab0ac, 0xd37295, 0xfabfd2, 0xb07aa1, 0xd4a6c8,
			    0x9d7660, 0xd7b5a6]);
add_color_palette("tab:colorblind", [0x1170aa, 0xfc7d0b, 0xa3acb9, 0x57606c, 0x5fa2ce, 0xc85200,
				     0x7b848f, 0xa3cce9, 0xffbc79, 0xc8d0d9]);
add_color_palette("tab:seattle-grays", [0x767f8b, 0xb3b7b8, 0x5c6060, 0xd3d3d3, 0x989ca3]);
add_color_palette("tab:traffic", [0xb60a1c, 0xe39802, 0x309143, 0xe03531, 0xf0bd27, 0x51b364,
				  0xff684c, 0xffda66, 0x8ace7e]);
add_color_palette("tab:miller-stone", [0x4f6980, 0x849db1, 0xa2ceaa, 0x638b66, 0xbfbb60, 0xf47942,
				       0xfbb04e, 0xb66353, 0xd7ce9f, 0xb9aa97, 0x7e756d]);
add_color_palette("tab:superfishel-stone", [0x6388b4, 0xffae34, 0xef6f6a, 0x8cc2ca, 0x55ad89, 0xc3bc3f,
					    0xbb7693, 0xbaa094, 0xa9b5ae, 0x767676]);
add_color_palette("tab:nuriel-stone", [0x8175aa, 0x6fb899, 0x31a1b3, 0xccb22b, 0xa39fc9, 0x94d0c0,
				       0x959c9e, 0x027b8e, 0x9f8f12]);
add_color_palette("tab:jewel-bright", [0xeb1e2c, 0xfd6f30, 0xf9a729, 0xf9d23c, 0x5fbb68, 0x64cdcc,
				       0x91dcea, 0xa4a4d5, 0xbbc9e5]);
add_color_palette("tab:summer", [0xbfb202, 0xb9ca5d, 0xcf3e53, 0xf1788d, 0x00a2b3, 0x97cfd0,
				 0xf3a546, 0xf7c480]);
add_color_palette("tab:winter", [0x90728f, 0xb9a0b4, 0x9d983d, 0xcecb76, 0xe15759, 0xff9888,
				 0x6b6b6b, 0xbab2ae, 0xaa8780, 0xdab6af]);
add_color_palette("tab:gree-orange-teal", [0x4e9f50, 0x87d180, 0xef8a0c, 0xfcc66d, 0x3ca8bc, 0x98d9e4,
					   0x94a323, 0xc3ce3d, 0xa08400, 0xf7d42a, 0x26897e, 0x8dbfa8]);
add_color_palette("tab:red-blue-brown", [0x466f9d, 0x91b3d7, 0xed444a, 0xfeb5a2, 0x9d7660, 0xd7b5a6,
					 0x3896c4, 0xa0d4ee, 0xba7e45, 0x39b87f, 0xc8133b, 0xea8783]);
add_color_palette("tab:purple-pink-gray", [0x8074a8, 0xc6c1f0, 0xc46487, 0xffbed1, 0x9c9290, 0xc5bfbe,
					   0x9b93c9, 0xddb5d5, 0x7c7270, 0xf498b6, 0xb173a0, 0xc799bc]);
add_color_palette("tab:hue-circle", [0x1ba3c6, 0x2cb5c0, 0x30bcad, 0x21b087, 0x33a65c, 0x57a337,
				     0xa2b627, 0xd5bb21, 0xf8b620, 0xf89217, 0xf06719, 0xe03426,
				     0xf64971, 0xfc719e, 0xeb73b3, 0xce69be, 0xa26dc2, 0x7873c0,
				     0x4f7cba]);
%!%+
%\function{palette--tab}
%\synopsis{Get Tableau10 color palettes}
%\usage{UInt_Type[] palette = get_color_palette("tab:"+name, UInt_Type n);}
%\description
%  Color palettes used by the Tableau10 software for visualization. Usuall only
%  good for qualitative data.
%!%-
%}}}

% Seaborn palette generators %{{{
private define space_converter (r, g, b)
{
  variable space = qualifier("space", "rgb");
  variable back = NULL;

  if (typeof(space) == String_Type)
  {
    if (space == "rgb")
    {
      (r,g,b) = (r/255., g/255., b/255.);
    }
    else if (space == "hsl")
    {
      (r, g, b) = rgb2hsl(r, g, b);
      back = &hsl2rgb;
    }
    else if (space == "hsluv")
    {
      (r, g, b) = rgb2hsluv(r, g, b);
      back = &hsluv2rgb;
    }
    else if (space == "hpluv")
    {
      (r, g, b) = rgb2hpluv(r, g, b);
      back = &hpluv2rgb;
    }
    else
      throw UsageError, "Color space '$space' is not recognized"$;
  }
  else if (typeof(space) == List_Type and length(space)==2
	   and typeof(space[0]) == Ref_Type and typeof(space[1]) == Ref_Type)
  {
    (r, g, b) = (@space[0])(r, g, b);
    back = space[1];
  }
  else
    throw UsageError, "Unable to perform color space conversion with given functions.";

  return (r,g,b,back);
}

private define blend_palette (n)
%!%+
%\function{palette--blend}
%\synopsis{Blend one or more colors linearly (in RGB space).}
%\usage{UInt_Type get_color_palette("blend", UInt_Type n);}
%\qualifiers{
%  \qualifier{colors[=[red,gray,blue]]}{The colors to interpolate, must be given as HEX}
%  \qualifier{set[=[0:1:#length(colors)]]}{The setpoints of the colors. Can be unsorted.}
%  \qualifier{space[="hsluv"]}{The color space in which interpolation is done.}
%}
%\description
%  Blend colors to a palette from given colors. The setpoints can be
%  used to shift the color at the same position to higher or lower
%  interpolation values.
%
%  The qualifier \code{space} allows to control in what space the interpolation
%  is done. This either must be one of the known color spaces "hsl", "hsluv", 
%  "hpluv", or "rgb", or a tuple (list with 2 elements) with references
%  to two functions. The first must take three arguments (R,G,B) as 8 bit
%  encoded colors and convert it to the color space variables (x,y,z). The
%  second converts back (x,y,z) -> (R,G,B).
%!%-
{
  variable colors = qualifier("colors", [0x2525a5, 0xa5a5a5, 0xa52525]);
  variable weights = _max(0., _min(1.,
			qualifier("set", [0:1:#length(colors)])
		     ));
  if (length(colors) != length(weights))
    throw UsageError, sprintf("Length of colors (%d) and length of setpoints (%d) do not match",
			      length(colors), length(weights));
  variable sort = array_sort(weights);
  weights = weights[sort];
  colors = colors[sort];

  variable r,g,b;
  (r,g,b) = hex2rgb(colors);

  variable back;
  (r, g, b, back) = space_converter(r, g, b; space=qualifier("space", "hsluv"));

  % linear interpolation of the color space
  variable dw = weights[[1:]]-weights[[:-2]];
  variable mr = (r[[1:]]-r[[:-2]])*1./dw;
  variable mg = (g[[1:]]-g[[:-2]])*1./dw;
  variable mb = (b[[1:]]-b[[:-2]])*1./dw;
  variable winf = where(dw==0);
  mr[winf] = 0.;
  mg[winf] = 0.;
  mb[winf] = 0.;

  weights = weights[[:-2]];
  variable idx = [0, int(weights*(n-1)), n-1];
  variable tr = r[[:-2]]-mr*weights;
  variable tg = g[[:-2]]-mg*weights;
  variable tb = b[[:-2]]-mb*weights;
  variable grid = [0:1:#n];
  r = Double_Type[n];
  g = Double_Type[n];
  b = Double_Type[n];
  variable w, i, m, t;

  w = [0:idx[0]-1];
  r[w] = _max(0., _min(grid[w]*mr[0]+tr[0], 1.));
  g[w] = _max(0., _min(grid[w]*mg[0]+tg[0], 1.));
  b[w] = _max(0., _min(grid[w]*mb[0]+tb[0], 1.));
  _for i (1, length(idx)-3)
  {
    w = [idx[i]:idx[i+1]-1];
    r[w] = _max(0., _min(grid[w]*mr[i-1]+tr[i-1], 1.));
    g[w] = _max(0., _min(grid[w]*mg[i-1]+tg[i-1], 1.));
    b[w] = _max(0., _min(grid[w]*mb[i-1]+tb[i-1], 1.));
  }
  w = [idx[-2]:idx[-1]];
  r[w] = _max(0., _min(grid[w]*mr[-1]+tr[-1], 1.));
  g[w] = _max(0., _min(grid[w]*mg[-1]+tg[-1], 1.));
  b[w] = _max(0., _min(grid[w]*mb[-1]+tb[-1], 1.));

  if (back == NULL)
    return rgb2hex(r, g, b);
  else
    return rgb2hex(@back(r, g, b));
}
add_color_palette("blend", &blend_palette);

private define dark_palette (n)
%!%+
%\function{palette--dark}
%\synopsis{Get dark blend palette with given color}
%\usage{UInt_Type[] palette = get_color_palette("dark", UInt_Type n);}
%\qualifiers{
%  \qualifier{color[=red]}{Color to blend with dark pattern (as HEX value).}
%  \qualifier{reverse}{If given, blend from color to dark}
%}
%\description
%  Get a color palette from the chosen color fading in from
%  a dark tone.
%!%-
{
  variable color = qualifier("color", 0xa52525);
  variable r, g, b;
  (r, g, b) = hex2rgb(color);
  variable h,s;
  (h,s,) = rgb2hsluv(r, g, b);
  variable gray = rgb2hex(hsluv2rgb(h, 0.15*s, 0.15));

  variable colors;
  if (qualifier_exists("reverse"))
    colors = [rgb2hex(r, g, b), gray];
  else
    colors = [gray, rgb2hex(r, g, b)];

  return blend_palette(n;; struct { @(__qualifiers()), colors=colors});
}
add_color_palette("dark", &dark_palette);

private define light_palette (n)
%!%+
%\function{palette--light}
%\synopsis{Get light blend palette with given color}
%\usage{UInt_Type[] palette = get_color_palette("light", UInt_Type n);}
%\qualifiers{
%  \qualifier{color[=red]}{Color to blend with dark pattern (as HEX value).}
%  \qualifier{reverse}{If given, blend from color to light}
%}
%\description
%  Get a color palette from the chosen color fading in from
%  a light tone.
%!%-
{
  variable color = qualifier("color", 0xa52525);
  variable r, g, b;
  (r, g, b) = hex2rgb(color);
  variable h,s;
  (h,s,) = rgb2hsluv(r, g, b);
  variable gray = rgb2hex(hsluv2rgb(h, 0.15*s, 0.95));

  variable colors;
  if (qualifier_exists("reverse"))
    colors = [rgb2hex(r, g, b), gray];
  else
    colors = [gray, rgb2hex(r, g, b)];

  return blend_palette(n;; struct { @(__qualifiers()), colors=colors});
}
add_color_palette("light", &light_palette);

private define diverging_palette (n)
%!%+
%\function{palette--diverging}
%\synopsis{Create a diverging palette between two colors}
%\usage{UInt_Type palette = get_color_palette("diverging", UInt_Type n);}
%\qualifiers{
%  \qualifier{dark}{If given, midpoint is a dark tone}
%  \qualifier{h_pos[=0.7]}{Hue of positive color}
%  \qualifier{h_neg[=0.2]}{Hue of negative color}
%  \qualifier{sep[=0.01]}{Size of midpoint region}
%  \qualifier{s[=0.75]}{Saturation of both colors}
%  \qualifier{l[=0.5]}{Luminosity of both colors}
%}
%\description
%  Create a color palette useful to represent diverging data.
%
%\seealso{hsluv2rgb}
%!%-
{
  variable blender = (qualifier_exists("dark")) ?
    &dark_palette: &light_palette;
  variable mid = (qualifier_exists("dark")) ?
    rgb2hex(hsluv2rgb(0.133, 0.133, 0.133)) : rgb2hex(hsluv2rgb(0.95, 0.95, 0.95));

  variable sep = _max(0.0, _min(1.0, qualifier("sep", 0.01)));
  variable s = qualifier("s", 0.75);
  variable l = qualifier("l", 0.5);
  variable neg = rgb2hex(hsluv2rgb(qualifier("h_neg", 0.611), s, l));
  variable pos = rgb2hex(hsluv2rgb(qualifier("h_pos", 0.055), s, l));
  variable sepw = int(sep*255);
  variable half = int(128-sepw/2);

  neg = (@blender)(half; color=neg, reverse);
  pos = (@blender)(half; color=pos);

  return blend_palette(n; colors=[neg, Int_Type[sepw]+mid, pos]);
}
add_color_palette("diverging", &diverging_palette);

private define cubehelix_palette (n)
%!%+
%\function{palette--cubehelix}
%\synopsis{Create palette using 'cubehelix' description}
%\usage{UInt_Type[] pattern = get_color_palette("cubehelix", UInt_Type n);}
%\qualifiers{
%  \qualifier{start[=0.0]}{Start color of the palette}
%  \qualifier{s[=0.8]}{Saturation of the colors}
%  \qualifier{rot[=0.4]}{How far the scale rotates in color (hue)}
%  \qualifier{dark[=0.15]}{Dark end of the palette}
%  \qualifier{light[=0.85]}{Light end of the palette}
%}
%\description
%  Calculates a color palette according to the description
%  in D. A. Green's, 'A colour scheme for the display of
%  astronomical intensity images', (Bull. Astr. Soc. India (2011)).
%!%-
{
  variable _r = [-0.14861, 1.78277];
  variable _g = [-0.29227, -0.90649];
  variable _b = [1.97294, 0.0];

  variable r, g, b;
  variable start = _max(0., _min(1., qualifier("start", 0.0)));
  variable s = _max(0.0, _min(1., qualifier("s", 0.8)));
  variable rot = qualifier("rot", 0.4);
  variable gamma = qualifier("gamma", 1.0);
  variable grid = [_max(0.0, _min(1.0, qualifier("dark", 0.15))):
		   _max(0.0, _min(qualifier("light", 0.85))):#n];
  variable lambda = grid^gamma;

  variable phi = 2*PI*(start + rot*grid);
  variable a = s*lambda*(1.0-lambda)/2.0;
  variable sphi = sin(phi);
  variable cphi = cos(phi);

  r = _max(0., _min(1., lambda + a*(_r[0]*sphi+_r[1]*cphi)));
  g = _max(0., _min(1., lambda + a*(_g[0]*sphi+_g[1]*cphi)));
  b = _max(0., _min(1., lambda + a*(_b[0]*sphi+_b[1]*cphi)));

  return rgb2hex(r, g, b);
}
add_color_palette("cubehelix", &cubehelix_palette);
%}}}

private define colormap_palette (n, name)
%!%+
%\function{palette--map}
%\synopsis{Retrieve a set of colors from a colormap}
%\usage{UInt_Type[] palette = get_color_palette (String_Type name, UInt_Type len);}
%\qualifiers{
%  \qualifier{start[=0.0]}{Starting "color" (lowest value)}
%  \qualifier{range[=1.0]}{Range of "color", larger than one causes repetitions}
%}
%\description
%  This function is useful if a range of colors is required.
%  It returns \code{len} values from a linear mapping of the
%  given colormap. The colors are constructed from palette--blend
%  and forwards all qualifiers to this constructor.
%
%  Note: This function is used only if the given \code{name} is
%  a colormap (and does not match a palette name, see \code{png_get_colormap_names}),
%  to construct a %  set of colors from a color array see
%  palette--blend.
%
%\seealso{png_get_colormap_names, palette--name}
%!%-
{
#ifexists png_add_colormap
  variable map;
  try
    map = png_get_colormap(name);
  catch AnyError:
    throw UsageError, "Unknown palette: '$name'"$;

  variable start = qualifier("start", 0.0);
  variable range = qualifier("range", 1.0);
  while (start<0) start += 1.0;

  start = int(start*(length(map)-1));
  range = int(range*(length(map)-1));

  variable ind = [start:start+range] mod length(map);

  return blend_palette(n;; struct { @__qualifiers(),  colors=map[ind] });
#else
  return ColorPalettes[""];
#endif
}

define get_color_palette (name, n)
%!%+
%\function{get_color_palette}
%\synopsis{Get a number of colors from a color palette}
%\usage{UInt_Type[] = palette get_color_palette(String_Type name, UInt_Type n);}
%\qualifiers{
%  \qualifier{repeat}{If given, repeat colors to match \code{n}}
%  \qualifier{cmap}{If given, force colormap search}
%  \qualifier{string}{Return colormap as HEX string instead of value}
%}
%\description
%  Get \code{n} colors (as hex encoded values) from a palette. Known
%  palette names can be obtained by \code{get_color_palette_names}.
%
%  This function forwards all qualifiers to the generator functions (if any).
%  If no palette matching \code{name} is found, a colormap is searched.
%  To generate a palette from a colormap only, use the \code{cmap}
%  qualifier. For information how palettes are generated from colormaps
%  see \code{palette--map}, else see \code{palette--name}, where \code{name}
%  is the first argument.
%  
%  If it is not a colormap nor a generator function, a static color map is
%  used. They are usually named by like \code{family:palette}.
%  For most of them, information can be obtained via \code{palette--family}.
%
%  Eventually the used map only has a finite number of colors. If \code{n}
%  exceeds this number, only the maximum number of colors are returned. To
%  repeat colors up to \code{n} times, use the \code{repeat} qualifier.  
%
%\seealso{palette--hsl,palette--map,palette--sron}
%!%-
{
  variable pal;

  if (assoc_key_exists(ColorPalettes, name) and not qualifier_exists("cmap"))
    pal = ColorPalettes[name];
  else
    pal = colormap_palette(n, name;; __qualifiers());

  if (typeof(pal) == Ref_Type)
    pal = @pal(n;; __qualifiers());

  if (qualifier_exists("repeat") and length(pal)<n)
    pal = pal[[0:n-1] mod length(pal)];

  if (qualifier_exists("string"))
    return array_map(String_Type, &sprintf, "%06x", pal[[0:n-1]]);

  return pal[[0:n-1]];
}

define get_color_palette_names ()
%!%+
%\function{get_color_palette_names}
%\synopsis{Get list of available color palette names}
%\usage{String_Type[] names = get_color_palette_names();}
%\seealso{get_color_palette}
%!%-
{
  return assoc_get_keys(ColorPalettes);
}

%
% Code from src/misc/imagedownload/panstarrs.sl
%

% only activate if module curl could be loaded
#ifeval __get_reference("curl_new")!=NULL

% only activate if module expat could be loaded
#ifeval __get_reference("xml_new")!=NULL

% only activate if module pcre could be loaded
#ifeval __get_reference("pcre_compile")!=NULL


% Accessing http://plpsipp1v.stsci.edu/cgi-bin/ps1cutouts via curl, parsing
% the output and dowloading the images. The usage of the exposed function
% is supposed to mimic the usage of the gui available on the page itself

% After I wrote this entire mechanism I discovered that there is additional
% information about an easier API in the lower sections on
% https://outerspace.stsci.edu/display/PANSTARRS/PS1+Image+Cutout+Service
% but well ...

% The base URL of the page itself
private variable __panstarrsBase = "http://ps1images.stsci.edu/cgi-bin/ps1cutouts";

% The following three variable represent the options available on the page
% They can be used as options in the URL
private variable __panstarrsFilters = ["g", "r", "i", "z", "y"];
private variable __panstarrsFiletypes = ["stack", "warp"];
private variable __panstarrsAuxiliary = ["data", "mask", "wt", "exp", "expwt", "num"];

% Pixel scale in "/px
private variable __panstarrsScale = 0.25;

% Irrelevant for the curl based query, but has to be set
private variable __panstarrsOutputSize = 256;
private variable __panstarrsVerbose = 0;
private variable __panstarrsAutoscale = 99.5;

% Callback for curl data retrieval
% Just appends the data which is supposed to be a BString to the string
% ref is pointing to
private define __panstarrsDownloadCallback(ref, data) {
  @ref += data;
  return strlen(data) + 1;
}

% Callback for durl cata retrieven
% Write the content to a filepointer
private define __panstarrsDownloadWriteCallback(ref, data) {
  if ( -1 == fwrite(data, @ref) ) {
    variable errsv = errno;
    throw RunTimeError, sprintf("Error writing to file: %s", errno_string(errsv));
  }
  return strlen(data) + 1;
}

% Pick the qualifiers apart to get the valid options for the query
private define __panstarrsGetOptions(pos, size) {
  variable filters = {}, fileTypes = {}, auxiliaries = {};
  variable filter, fileType, auxiliary;

  if ( qualifier_exists("color") ) {
    throw UsageError, "Combined color images can not be downloaded";
  }

  foreach filter(__panstarrsFilters) {
    if ( qualifier_exists(filter) ) {
      list_append(filters, filter);
    }
  }
  !if (length(filters)) {
    throw UsageError, "At least one filter is required for PanSTARRS image queries";
  }

  foreach fileType(__panstarrsFiletypes) {
    if ( qualifier_exists(fileType) ) {
      list_append(fileTypes, fileType);
    }
  }
  !if ( length(fileTypes) ) {
    throw UsageError, "At least one file type is required for PanSTARRS image queries" +
      "\n(did you mean \"stack\"?)";
  }

  foreach auxiliary(__panstarrsAuxiliary) {
    if ( qualifier_exists(auxiliary) ) {
      list_append(auxiliaries, auxiliary);
    }
  }
  !if ( length(auxiliaries) ) {
    throw UsageError, "At least one auxiliary type is required for PanSTARRS image queries" +
      "\n(did you mean \"data\"?)";
  }
  return (filters, fileTypes, auxiliaries);
}

%!%+
%\function{__assemblePanstarrsUrl}
%\synopsis{Assemble a URL for the PanSTARRS image cutout server}
%\usage{String_Type __assemblePanstarrsUrl(String_Type position, Integer_Type size)}
%\description
%    position is either a set of coordinates of the form "ra+/-dec"
%    (e.g. "124.45+34") or a string which is used by the cutout server itself
%    to perform a simbad query.
%    size is the pixel number along the edge of the square image.
%    The qualifiers for the function are the set of options for filter, filetype,
%    and auxiliary data given in __panstarrsFilters, __panstarrsFiletypes, and
%    __panstarrsAuxiliary
%!%-
private define __assemblePanstarrsUrl(pos, size, filters, fileTypes, auxiliaries) {
  variable options;
  variable filter, fileType, auxiliary;

  options = {
    sprintf("pos=%s", pos),
    sprintf("size=%d", size),
    sprintf("output_size=%d", __panstarrsOutputSize),
    sprintf("verbose=%d", __panstarrsVerbose),
    sprintf("autoscale=%f", __panstarrsAutoscale),
    "catlist="
  };
  foreach filter(filters) {
    list_append(options, sprintf("filter=%s", filter));
  }
  foreach fileType(fileTypes) {
    list_append(options, sprintf("filetypes=%s", fileType));
  }
  foreach auxiliary(auxiliaries) {
    list_append(options, sprintf("auxiliary=%s", auxiliary));
  }
  options = list_to_array(options);
  return __panstarrsBase + "?" + strjoin(options, "&");
}

%!%+
%\function{__panstarrsRetrieveSite}
%\synopsis{Download a html page of the PanSTARRS cutout server}
%\usage{String_Type __panstarrsRetrieveSite(String_Type URL)}
%\description
%    Simply takes a url and downloads the html content of the site using curl,
%    returning it as a string. Throws an exception upon error.
%!%-
private define __panstarrsRetrieveSite(url) {
  variable c = curl_new(url);
  variable str = "";
  curl_setopt(c, CURLOPT_WRITEFUNCTION, &__panstarrsDownloadCallback, &str);
  curl_perform(c);
  return str;
}

% XML parser character handler
private define __panstarrsXMLCharHandler(p, data) {
  p.userdata.currentText = data;

  % we just entered a new image section in the table
  if ( p.userdata.currentElement == "th" ) {
    % Store the name of the current image (e.g. stack 2469.010 g, ...)
    p.userdata.currentThing = data;
  }
}

% XML parser start element handler
private define __panstarrsXMLStartElementHandler(p, element, attrs) {
  variable ind;
  p.userdata.currentElement = element;
  p.userdata.currentAttrs = attrs;

  % Array of struct -> struct of array
  attrs = struct {
    names = array_map(String_Type, &get_struct_field, attrs, "name"),
    values = array_map(String_Type, &get_struct_field, attrs, "value")
  };
  ind = where(attrs.names == "title");
  variable titleAttrs = struct_filter(attrs, ind; copy);
  ind = where(titleAttrs.values == "Download FITS cutout");
  !if ( length(ind) ) {
    % Here is not a fits cutout
    % Important: There are no colored cutouts available, only single band images
    return;
  }
  if ( length(ind) > 1 ) {
    % Two "Download FITS cutout" in one table entry, that's very wrong
    throw DataError, "Found multiple cutouts for one image";
  }
  ind = where(attrs.names == "href");
  variable key = p.userdata.currentThing;
  if ( assoc_key_exists(p.userdata.images, key) ) {
    throw DataError, "Found multiple images for " + key;
  }
  p.userdata.images[key] = attrs.values[ind[0]];
}

% Insert a stack only match
private define __panstarrsInsertStackMatch(assoc, col, url) {
  !if ( assoc_key_exists(assoc, "stack") ) {
    assoc["stack"] = Assoc_Type[Assoc_Type];
  }
  !if ( assoc_key_exists(assoc["stack"], "colors") ) {
    assoc["stack"]["colors"] = Assoc_Type[String_Type];
  }
  assoc["stack"]["colors"][col] = url;
}

% Insert a stack and auxiliary data match
private define __panstarrsInsertStackAuxMatch(assoc, aux, col, url) {
  !if ( assoc_key_exists(assoc, "stack") ) {
    assoc["stack"] = Assoc_Type[Assoc_Type];
  }
  !if ( assoc_key_exists(assoc["stack"], "aux") ) {
    assoc["stack"]["aux"] = Assoc_Type[Assoc_Type];
  }
  !if ( assoc_key_exists(assoc["stack"]["aux"], aux) ) {
    assoc["stack"]["aux"][aux] = Assoc_Type[String_Type];
  }
  assoc["stack"]["aux"][aux][col] = url;
}

% Insert a warp only match
private define __panstarrsInsertWarpMatch(assoc, col, field, url) {
  !if ( assoc_key_exists(assoc, "warp") ) {
    assoc["warp"] = Assoc_Type[Assoc_Type];
  }
  !if ( assoc_key_exists(assoc["warp"], "colors") ) {
    assoc["warp"]["colors"] = Assoc_Type[Assoc_Type];
  }
  !if ( assoc_key_exists(assoc["warp"]["colors"], col) ) {
    assoc["warp"]["colors"][col] = Assoc_Type[String_Type];
  }
  assoc["warp"]["colors"][col][field] = url;
}

% Insert a warp and auxiliary data match
private define __panstarrsInsertWarpAuxMatch(assoc, aux, col, field, url) {
  !if ( assoc_key_exists(assoc, "warp") ) {
    assoc["warp"] = Assoc_Type[Assoc_Type];
  }
  !if ( assoc_key_exists(assoc["warp"], "aux") ) {
    assoc["warp"]["aux"] = Assoc_Type[Assoc_Type];
  }
  !if ( assoc_key_exists(assoc["warp"]["aux"], aux) ) {
    assoc["warp"]["aux"][aux] = Assoc_Type[Assoc_Type];
  }
  !if ( assoc_key_exists(assoc["warp"]["aux"][aux], col) ) {
    assoc["warp"]["aux"][aux][col] = Assoc_Type[String_Type];
  }
  assoc["warp"]["aux"][aux][col][field] = url;
}

% Parse the names of the images and sort the accordingly
% The names are named according to this scheme:
%
% stack \d+\.\d+ (g|i|r|y|z)
% stack\.(exp|expwt|mask|num|wt) \d+\.\d+ (g|i|r|y|z)
% warp \d+\.\d+ (g|i|r|y|z) \d+\.\d+
% warp.(mask|wt) \d+\.\d+ (g|i|r|y|z) \d+\.\d+
%
% where the first \d+\.\d+ should always have the same values but the second
% \d+\.\d+ for the warps can differ and give a lot of results. Due to this reason
% the first sets of digits will be removed but the second one will stay.
% The regex for the colors and the auxiliary types is assembled dynamically from
% __panstarrsFilters and __panstarrsAuxiliary
% Returns an assoc of the form:
%assoc
%├── stack
%│   ├── aux
%│   │   ├── exp
%│   │   │   ├── g=url
%│   │   │   ├── ...
%│   │   ├── expwt
%│   │   │   ├── g=url
%│   │   │   ├── ...
%│   │   ├── mask
%│   │   │   ├── g=url
%│   │   │   ├── ...
%│   │   ├── num
%│   │   │   ├── g=url
%│   │   │   ├── ...
%│   │   └── wt
%│   │       ├── g=url
%│   │       ├── ...
%│   └── colors
%│       ├── g=url
%│       ├── ...
%└── warp
%    ├── aux
%    │   ├── mask
%    │   │   ├── g
%    │   │   │   └── 55275.32206=url
%    │   │   ├── i
%    │   │   │   └── 57035.57214=url
%    │   │   ├── r
%    │   │   │   └── 56303.54493=url
%    │   │   ├── y
%    │   │   │   └── 55903.56165=url
%    │   │   └── z
%    │   │       └── 55932.46071=url
%    │   └── wt
%    │       ├── g
%    │       │   └── 55275.32281=url
%    │       ├── i
%    │       │   └── 57035.57214=url
%    │       ├── r
%    │       │   └── 55268.34759=url
%    │       ├── y
%    │       │   └── 55903.57351=url
%    │       └── z
%    │           └── 56618.57846=url
%    └── colors
%        ├── g
%        │   ├── 55238.43916=url
%        │   ├── ...
%        ├── i
%        │   ├── 55259.45618=url
%        │   ├── ...
%        ├── r
%        │   ├── 55268.33816=url
%        │   ├── ...
%        ├── y
%        │   ├── 55521.60728=url
%        │   ├── ...
%        └── z
%            ├── 55546.50549=url
%            ├── ...
%#p-
%#v-
% Where all nodes are Assoc_Type[Assoc_Type] except each deepst leave which is
% an Assoc_Type[String_Type]. Note that not all of the branches might be there,
% depending on what has been asked for to query and what is actually available
private define __panstarrsSortUrls(urls) {
  variable names = assoc_get_keys(urls), name;
  variable colorCaptureRegex = "(" + strjoin(__panstarrsFilters, "|") + ")";
  variable auxCaptureRegex = "(" + strjoin(__panstarrsAuxiliary, "|") + ")";
  variable numDotNumCaptureRegex = `(\d+\.\d+)`;
  variable stackRegexStr = "^" + strjoin(["stack", numDotNumCaptureRegex,
      colorCaptureRegex], `\s+`) + "$";
  variable stackAuxRegexStr = "^" + strjoin(["stack\." + auxCaptureRegex,
      numDotNumCaptureRegex, colorCaptureRegex], `\s+`) + "$";
  variable warpRegexStr = "^" + strjoin(["warp", numDotNumCaptureRegex,
      colorCaptureRegex, numDotNumCaptureRegex], `\s+`) + "$";
  variable warpAuxRegexStr = "^" + strjoin(["warp\." + auxCaptureRegex,
      numDotNumCaptureRegex, colorCaptureRegex, numDotNumCaptureRegex], `\s+`) + "$";
  variable stackRegex = pcre_compile(stackRegexStr);
  variable stackAuxRegex = pcre_compile(stackAuxRegexStr);
  variable warpRegex = pcre_compile(warpRegexStr);
  variable warpAuxRegex = pcre_compile(warpAuxRegexStr);
  variable colorName, auxName, fieldName;
  variable ret = Assoc_Type[Assoc_Type];
  foreach name(names) {
    if ( pcre_exec(stackRegex, name ) ) {
      colorName = pcre_nth_substr(stackRegex, name, 2);
      __panstarrsInsertStackMatch(ret, colorName, urls[name]);
    } else if ( pcre_exec(stackAuxRegex, name) ) {
      auxName = pcre_nth_substr(stackAuxRegex, name, 1);
      colorName = pcre_nth_substr(stackAuxRegex, name, 3);
      __panstarrsInsertStackAuxMatch(ret, auxName, colorName, urls[name]);
    } else if ( pcre_exec(warpRegex, name) ) {
      colorName = pcre_nth_substr(warpRegex, name, 2);
      fieldName = pcre_nth_substr(warpRegex, name, 3);
      __panstarrsInsertWarpMatch(ret, colorName, fieldName, urls[name]);
    } else if ( pcre_exec(warpAuxRegex, name) ) {
      auxName = pcre_nth_substr(warpAuxRegex, name, 1);
      colorName = pcre_nth_substr(warpAuxRegex, name, 3);
      fieldName = pcre_nth_substr(warpAuxRegex, name, 4);
      __panstarrsInsertWarpAuxMatch(ret, auxName, colorName, fieldName, urls[name]);
    } else {
      vmessage("Warning: Skipping unrecognized image type: %s", name);
    }
  }
  return ret;
}

%!%+
%\function{__parsePanstarrsPage}
%\synopsis{Parse the html of a PanSTARRS cutout server page}
%\usage{Assoc_Type __parsePanstarrsPage(String_Type html)}
%\description
%    Takes the result of __panstarrsRetrieveSite() and parses it in order to
%    determine which images are available and what their specific download
%    URLs are
%!%-
private define __parsePanstarrsPage(str) {
  variable _lines = strsplit(str, "\n");
  variable n = length(_lines);
  variable newstr = "";
  variable ii = 0;

  % Remove the javascript section as it contains invalid XML, e.g. i<inputs.length;
  while ( ii<n ) {
    if ( _lines[ii] == "<script type=\"text/javascript\">" ) {
      while ( _lines[ii] != "</script>" ) {
	ii++;
      }
      ii++;
    }
    newstr += _lines[ii] + "\n";
    ii++;
  }

  % Escape the &
  newstr = strreplace(newstr, "&", "&#038;");

  variable xmlp = xml_new();
  xmlp.userdata = struct {
    images = Assoc_Type[String_Type],
    currentElement,
    currentAttrs,
    currentText,
    currentThing
  };
  xmlp.startelementhandler = &__panstarrsXMLStartElementHandler;
  xmlp.characterdatahandler = &__panstarrsXMLCharHandler;
  xml_parse(xmlp, newstr, 0);
  variable urls = xmlp.userdata.images;
  variable key;
  foreach key(assoc_get_keys(urls)) {
    % Replace the escaped &'s with actual &'s
    variable url = strreplace(urls[key], "&amp;", "&");
    % The resulting URL's have leading // but no http:, so add it
    urls[key] = "http:" + url;
  }
  variable keys = assoc_get_keys(urls);
  return __panstarrsSortUrls(urls);
}

% Download a file from a url and save it as fname
private define __panstarrsDownloadFile(url, fname) {
  if ( stat_file(fname) != NULL ) {
    throw RunTimeError, sprintf("File %s already exists", fname);
  }
  !if ( qualifier_exists("quiet") ) {
    vmessage("Downloading as %s: %s", fname, url);
  }
  variable fptr = fopen(fname, "w");
  if ( fptr == NULL ) {
    variable errsv = errno;
    throw RunTimeError, sprintf("Error opening %s for writing: %s", fname, errno_string(errsv));
  }
  variable c = curl_new(url);
  curl_setopt(c, CURLOPT_WRITEFUNCTION, &__panstarrsDownloadWriteCallback, &fptr);
  curl_perform(c);
  if ( -1 == fclose(fptr) ) {
    throw RunTimeError, sprintf("Error closing file %s: %s", fname, errno_string(errsv));
  }
}

% Download all found stacks
private define __panstarrsDownloadStacks(urls, outdir, colors, assoc) {
  variable filter;
  !if ( assoc_key_exists(urls, "stack") ) {
    return;
  }
  !if ( assoc_key_exists(urls["stack"], "colors") ) {
    return;
  }
  foreach filter(colors) {
    !if ( assoc_key_exists(urls["stack"]["colors"], filter) ) {
      continue;
    }
    variable outfile = sprintf("%sstack_%s.fits", outdir, filter);
    __panstarrsDownloadFile(urls["stack"]["colors"][filter], outfile;; __qualifiers);
    __panstarrsInsertStackMatch(assoc, filter, outfile);
  }
}

% Download all found auxiliary stacks
private define __panstarrsDownloadStacksAux(urls, outdir, auxiliaries, colors, assoc) {
  variable aux, filter;
  !if ( assoc_key_exists(urls, "stack") ) {
    return;
  }
  !if ( assoc_key_exists(urls["stack"], "aux") ) {
    return;
  }
  foreach aux(auxiliaries) {
    !if ( assoc_key_exists(urls["stack"]["aux"], aux) ) {
      continue;
    }
    foreach filter(colors) {
      !if ( assoc_key_exists(urls["stack"]["aux"][aux], filter) ) {
	continue;
      }
      variable outfile = sprintf("%sstack_%s_%s.fits", outdir, aux, filter);
      __panstarrsDownloadFile(urls["stack"]["aux"][aux][filter], outfile;; __qualifiers);
      __panstarrsInsertStackAuxMatch(assoc, aux, filter, outfile);
    }
  }
}

% Download all found warps
private define __panstarrsDownloadWarps(urls, outdir, colors, assoc) {
  variable filter, field;
  !if ( assoc_key_exists(urls, "warp") ) {
    return;
  }
  !if ( assoc_key_exists(urls["warp"], "colors") ) {
    return;
  }
  foreach filter(colors) {
    !if ( assoc_key_exists(urls["warp"]["colors"], filter) ) {
      continue;
    }
    foreach field(assoc_get_keys(urls["warp"]["colors"][filter])) {
      variable outfile = sprintf("%swarp_%s_%s.fits", outdir, filter, field);
      __panstarrsDownloadFile(urls["warp"]["colors"][filter][field], outfile;; __qualifiers);
      __panstarrsInsertWarpMatch(assoc, filter, field, outfile);
    }
  }
}

% Download all found warp auxiliaries
private define __panstarrsDownloadWarpsAux(urls, outdir, auxiliaries, colors, assoc) {
  variable aux, filter, field;
  !if ( assoc_key_exists(urls, "warp") ) {
    return;
  }
  !if ( assoc_key_exists(urls["warp"], "aux") ) {
    return;
  }
  foreach aux(auxiliaries) {
    !if ( assoc_key_exists(urls["warp"]["aux"], aux) ) {
      continue;
    }
    foreach filter(colors) {
      !if ( assoc_key_exists(urls["warp"]["aux"][aux], filter) ) {
	continue;
      }
      foreach field(assoc_get_keys(urls["warp"]["aux"][aux][filter])) {
	variable outfile = sprintf("%swarp_%s_%s_%s.fits", outdir, aux, filter, field);
	__panstarrsDownloadFile(urls["warp"]["aux"][aux][filter][field], outfile;; __qualifiers);
	__panstarrsInsertWarpAuxMatch(assoc, aux, filter, field, outfile);
      }
    }
  }
}

% Download the images into a directory and name them
% The directory already must exist
private define __panstarrsDownloadImages(urls, outdir, filters, fileTypes, auxiliaries) {
  variable ret = Assoc_Type[Assoc_Type];
  __panstarrsDownloadStacks(urls, outdir, filters, ret;; __qualifiers);
  __panstarrsDownloadStacksAux(urls, outdir, auxiliaries, filters, ret;; __qualifiers);
  __panstarrsDownloadWarps(urls, outdir, filters, ret;; __qualifiers);
  __panstarrsDownloadWarpsAux(urls, outdir, auxiliaries, filters, ret;; __qualifiers);
  return ret;
}

%!%+
%\function{panstarrsImageDownload}
%\synopsis{Download PanSTARRS image cutouts}
%\usage{Assoc_Type fileNames = panstarrsImageDownload(String_Type position, Double_Type size, String_Type outdir)
%    Assoc_Type fileNames = panstarrsImageDownload(Double_Type ra, Double_Type dec, Double_Type size, String_Type outdir)}
%\qualifiers{
%\qualifier{g}{Download images taken with the g filter}
%\qualifier{r}{Download images taken with the r filter}
%\qualifier{i}{Download images taken with the i filter}
%\qualifier{z}{Download images taken with the z filter}
%\qualifier{y}{Download images taken with the y filter}
%\qualifier{stack}{Download final stacks}
%\qualifier{warp}{Download individual warps (single epoch images)}
%\qualifier{data}{Download main data products}
%\qualifier{mask}{Download masks}
%\qualifier{wt}{Download weight images}
%\qualifier{exp}{Download exposure maps}
%\qualifier{expwt}{Download weighted exposure maps (undocumented)}
%\qualifier{num}{Download num images (undocumented)}
%}
%\notes
%    Please note that PanSTARRS only covers the sky north of DEC=-30 deg.
%    At least one filter is required.
%    At least one of stack or warp is required.
%    At least one of data, mask, wt, exp, expwt or num is required.
%    The function only throws an exception on real runtime errors. If some data
%    just aren't available no exceptions is thrown, the images just don't exist
%    and are therefore not downloaded and not entered into the returned assoc.
%\description
%    Access the PanSTARRS DR1 image cutout server located at
%    http://plpsipp1v.stsci.edu/cgi-bin/ps1cutouts to
%    download different combinations of data products and filters.
%    The usage of this function is analogous to the usage of this website.
%    position can be a generic string which is resolved by the server itself
%    using Simbad or other means. Alternatively RA and DEC can be supplied
%    directly to the function in degrees.
%    outdir must supply a directory where the output files shall be written.
%    The function returns an Assoc_Type which contains the information about
%    the downloaded files. If all options are set the assoc looks like this:
%#v+
%#p+
%    assoc
%    ├── stack
%    │   ├── aux
%    │   │   ├── exp
%    │   │   │   ├── g=path
%    │   │   │   ├── ...
%    │   │   ├── expwt
%    │   │   │   ├── g=path
%    │   │   │   ├── ...
%    │   │   ├── mask
%    │   │   │   ├── g=path
%    │   │   │   ├── ...
%    │   │   ├── num
%    │   │   │   ├── g=path
%    │   │   │   ├── ...
%    │   │   └── wt
%    │   │       ├── g=path
%    │   │       ├── ...
%    │   └── colors
%    │       ├── g=path
%    │       ├── ...
%    └── warp
%        ├── aux
%        │   ├── mask
%        │   │   ├── g
%        │   │   │   ├── 55860.59225=path
%        │   │   │   ├── ...
%        │   │   ├── i
%        │   │   │   ├── 55200.51443=path
%        │   │   │   ├── ...
%        │   │   ├── r
%        │   │   │   ├── 55911.49788=path
%        │   │   │   ├── ...
%        │   │   ├── y
%        │   │   │   ├── 55518.62770=path
%        │   │   │   ├── ...
%        │   │   └── z
%        │   │       ├── 55302.23917=path
%        │   │       ├── ...
%        │   └── wt
%        │       ├── g
%        │       │   ├── 55860.59225=path
%        │       │   ├── ...
%        │       ├── i
%        │       │   ├── 55200.51443=path
%        │       │   ├── ...
%        │       ├── r
%        │       │   ├── 55911.49788=path
%        │       │   ├── ...
%        │       ├── y
%        │       │   ├── 55518.62770=path
%        │       │   ├── ...
%        │       └── z
%        │           ├── 55302.23917=path
%        │           ├── ...
%        └── colors
%            ├── g
%            │   ├── 55860.59225=path
%            │   ├── ...
%            ├── i
%            │   ├── 55200.51443=path
%            │   ├── ...
%            ├── r
%            │   ├── 55911.49788=path
%            │   ├── ...
%            ├── y
%            │   ├── 55518.62770=path
%            │   ├── ...
%            └── z
%                ├── 55302.23917=path
%                ├── ...
%#p-
%#v-
%    The structure should be self explanatory, except the set of numbers at
%    the lowest level of the warps. These are the MJDs when the exposure
%    was completed.
%!%-
define panstarrsImageDownload() {
  variable pos, size, outdir, ra, dec;
  switch(_NARGS)
  { case 3:
    (pos, size, outdir) = ();
  }
  { case 4:
    (ra, dec, size, outdir) = ();
    !if ( typeof(ra) == Integer_Type || typeof(ra) == Double_Type || typeof(ra) == Float_Type ) {
      throw UsageError, "ra needs to be float, double or int";
    }
    !if ( typeof(dec) == Integer_Type || typeof(dec) == Double_Type || typeof(dec) == Float_Type ) {
      throw UsageError, "dec needs to be float, double or int";
    }
    if ( ra < 0. || ra >= 360. ) {
      throw UsageError, "ra must be in [0.0, 360)";
    }
    if ( dec < -90. || dec > 90. ) {
      throw UsageError, "dec must be [90.0, 90.0]";
    }
    if ( dec <= -30. ) {
      throw UsageError, "PanSTARRS is only available at DEC > -30 deg";
    }
    pos = sprintf("%f%+f", ra, dec);
  }
  { throw NumArgsError, "Invalid number of arguments"; }

  !if ( typeof(pos) == String_Type ) {
    throw UsageError, "position needs to be a string";
  }
  !if ( typeof(size) == Integer_Type || typeof(size) == Double_Type || typeof(size) == Float_Type ) {
    throw UsageError, "size needs to be float, double or int";
  }
  !if ( size > 0.0 ) {
    throw UsageError, "size must be larger than 0.0";
  }
  !if ( typeof(outdir) == String_Type ) {
    throw UsageError, "outdir needs to be a string";
  }

  variable dstat = stat_file(outdir);
  if ( dstat == NULL ) {
    if ( -1 == mkdir(outdir) ) {
      variable errsv = errno;
      throw RunTimeError, sprintf("Error creating directory %s: %s",
	  outdir, errno_string(outdir));
    }
  } else {
    !if ( stat_is("dir", dstat.st_mode ) ) {
      throw UsageError, sprintf("File %s exists but is not a directory", outdir);
    }
  }

  !if ( outdir[-1] == '/' ) {
    outdir += "/";
  }

  size = int(ceil(size / __panstarrsScale));
  variable filters, fileTypes, auxiliaries;
  (filters, fileTypes, auxiliaries) = __panstarrsGetOptions(pos, size;; __qualifiers);
  variable url = __assemblePanstarrsUrl(pos, size, filters, fileTypes, auxiliaries);
  variable siteData = __panstarrsRetrieveSite(url);
  variable imageUrls = __parsePanstarrsPage(siteData);
  !if ( length(assoc_get_keys(imageUrls)) ) {
    vmessage("Warning: No images found to downloaded");
  }
  return __panstarrsDownloadImages(imageUrls, outdir, filters, fileTypes, auxiliaries;; __qualifiers);
}

#endif

#endif

#endif

%
% Code from src/fitting/par.sl
%
%%! Author: Fritz-Walter Schwarm <fritz.schwarm@sternwarte.uni-erlangen.de>

%%! Filename of currently loaded parameter file
private variable par_file = "";
%%! Parameter file last modification
private variable par_mtime = 0;
%%! Parameter file stat struct
private variable par_stat;
%%! Lines of the last used parameter file
private variable par_lines = Array_Type[0];
%%! Names of all components in parameter file
private variable par_names = Array_Type[0];
%%! Number of components in parameter file
private variable par_N = 0;
%%! Index of current component
private variable par_idx = 0;
%%! Global list of renaming operations
private variable par_rename = 0;
%%! Regular expression for parameter from parameter listing
private variable par_regexp = "^ *\\([0-9]\+\\) \+\\([\^. ()]\+\\)" +
	"(\\([0-9]\+\\))\.\\([\^ ]\+\\) \+\\([0-9]\+\\) \+\\([0-9]\+\\) \+" +
	"\\([-+.eE0-9]\+\\) \+\\([-+.eE0-9]\+\\) \+\\([-+.eE0-9]\+\\) \+\\([\^ ]*\\)";

%%! Read a parameter file to an array of strings
private define par_read(fname)
{
	variable fp, lines;

	fp = fopen(fname, "r");

	if (fp == NULL) {
		throw OpenError, "fopen failed to open $fname for reading"$;
	}

	lines = fgetslines(fp);

	if (lines == NULL) {
		throw ReadError, "fgetslines failed to read lines from $fname"$;
	}

	() = fclose(fp);

	return lines;
}

%%! Get components from an array of strings
private define par_cmp(lines)
{
	variable i, pos, strs, len, first = 1, list = {};

	for (i = 0; i < length(lines); i++) {

		strs = string_matches(lines[i], "^#\[\(.*\)\]"R);

		if (NULL != strs) {

			list_append(list, strs[1]);

		} else if ((0 == string_match(lines[i], "^ *[\^#]"R)) and (1 == first)) {

			list_insert(list, "DEFAULT");
			first = 0;
		}
	}

	return list;
}

%%%%%%%%%%%%%%%%%
define par_load()
%%%%%%%%%%%%%%%%%
%!%+
%\function{par_load}
%\synopsis{Loads fit function and parameters for given component from file}
%\usage{par_load(String_Type filename [, String_Type component]);}
%\qualifiers{
%    \qualifier{clean}{[=1]:   Remove temporary parameter files for individual
%                       components after calling load_par.}
%    \qualifier{dryrun}{[=0]:  Generate parameter file for component but
%                       do not load it, i.e., do not call load_par.
%                       Implies clean=0.}
%    \qualifier{iterate}{[=0]: Iterate through components. Returns the name of
%                       the current component or NULL if no more components
%                       are available.}
%}
%\description
%    Creates a copy of the given parameter file and modifies it such that only
%    the fit function associated with the given component is active. This file
%    is then loaded using the standart load_par function.
%    This way individual model components, or alternative fit functions, can
%    be defined in the same parameter file using the following format:
%    	#[name1] fitFunction1
%    	#[name2] fitFunction2
%    	defaultFitFunction
%    The rest of the file contains the parameters in the usual format used by
%    list_par. The name "DEFAULT" is reserved for the (uncommented) default
%    fit function. The function returns the name of the loaded component. It
%    returns NULL if the desired component does not exist or if no more
%    components are available during iteration.
%\seealso{par_save, load_par, list_par}
%!%-
{
	variable fname, cmp, active_cmps = NULL, cmp_pars, i, j, k, par, active, fp, buf, strs, fcmp, stat, first = 1;

	EXIT_BLOCK {
		if (par_idx < par_N) {
			return par_names[par_idx];
		} else {
			return NULL;
		}
	}

	%%! Check if component parameter was given
	if (_NARGS == 0 or _NARGS > 2) {

		usage("[cmp] = par_load(filename [, component])");

	} else if (_NARGS == 1) {

		fname = ();
		cmp = NULL;

	} else {
		cmp = ();
		fname = ();
	}

	par_stat = stat_file(fname);

	if (par_stat == NULL) {
		throw IOError, "Unable to stat $fname"$;
	}

	if ((0 == strcmp("", par_file)) or 
		(0 != strcmp(fname, par_file)) or 
			((par_stat != NULL) and (par_stat.st_mtime != par_mtime))) {

		vmessage("Loading file $fname ..."$);

		par_file = fname;

		par_lines = par_read(fname);

		par_names = par_cmp(par_lines);

		par_N = length(par_names);

		par_idx = 0;

		par_mtime = par_stat.st_mtime;

	} else if ((1 == qualifier_exists("iterate")) and (0 != qualifier("iterate", 0))) {

		par_idx++;
	}

	if (NULL != cmp) {
		par_idx = 0;

		while (strcmp(cmp, par_names[par_idx]) != 0 and par_idx < par_N) { par_idx++; }
	}

	if (par_idx >= par_N) {
		return;
	}

	cmp = par_names[par_idx];

	fcmp = sprintf("%s_%s", fname, cmp);

	fp = fopen(fcmp, "w");

	if (fp == NULL) {
		throw OpenError, "fopen failed to open $fcmp for writing"$;
	}

	for (i = 0; i < length(par_lines); i++) {

		strs = string_matches(par_lines[i], "^#\\[$cmp\\]\\(.*\\)"$);

		if (NULL != strs) {

			%%! This line contains the desired component -> uncomment
			if (-1 == fprintf(fp, "%s\n", strs[1])) {
				throw WriteError, "fprintf failed to write to $fcmp: "$ + errno_string(errno);
			}

			%%! Setting the fit function makes it possible to use get_fun_components
			fit_fun(strs[1]);

			%%! Obtain individual components
			active_cmps = get_fun_components();

		} else if ((0 == string_match(par_lines[i], "^ *[\^#]"R)) and (1 == first)) {

			%%! First non-comment line -> comment out unless DEFAULT component is requested
			if (0 == strcmp("DEFAULT", cmp)) {
				if (-1 == fprintf(fp, "%s", par_lines[i])) {
					throw WriteError, "fprintf failed to write to $fcmp: "$ + errno_string(errno);
				}

				%%! Setting the fit function makes it possible to use get_fun_components
				fit_fun(par_lines[i]);

				%%! Obtain individual components
				active_cmps = get_fun_components();
			} else {
				if (-1 == fprintf(fp, "#[DEFAULT] %s", par_lines[i])) {
					throw WriteError, "fprintf failed to write to $fcmp: "$ + errno_string(errno);
				}
			}

			first = 0;
		} else {
			%%! Either a comment, header, or parameter line
			par = string_matches(par_lines[i], "^ *\\([0-9]\+\\) \+\\([\^. ]\+\\)");

			if (NULL != par) {
				%%! Parameter line -> transfer only active parameters (otherwise ISIS throws an error on load_par)
				k = 0;
				active = 0;

				while ((NULL != active_cmps) and (k < length(active_cmps)) and (0 == active)) {
					if (0 == strcmp(par[2], active_cmps[k])) {
						active = 1;
					} else {
						k++;
					}
				}

				if (active) {
					%%! Parameter is needed for current component -> transfer unaltered
					if (-1 == fprintf(fp, "%s", par_lines[i])) {
						throw WriteError, "fprintf failed to write to $fcmp: "$ + errno_string(errno);
					}
				} else {
					%%! Parameter is not needed for current component -> comment out
					if (-1 == fprintf(fp, "# %s", par_lines[i])) {
						throw WriteError, "fprintf failed to write to $fcmp: "$ + errno_string(errno);
					}
				}

			} else {
				%%! Must be the a comment or header line -> transfer unaltered
				if (-1 == fprintf(fp, "%s", par_lines[i])) {
					throw WriteError, "fprintf failed to write to $fcmp: "$ + errno_string(errno);
				}
			}
		}
	}

	() = fclose(fp);

	if ((0 == qualifier_exists("dryrun")) or (0 == qualifier("dryrun", 0))) {

		load_par(fcmp);

		if ((0 == qualifier_exists("clean")) or (0 != qualifier("clean", 1))) {
			() = system("rm $fcmp"$);
		}
	}

	return;
}

private define par_array_idx(array, element) {
	variable i;

	for (i = 0; i < length(array); i++) {
		if (array[i] == element) {
			return i;
		}
	}

	return -1;
}

private define par_format_text(str) {

	variable i, buffer, match, tmp, left, right, last = {};
	variable strs;
	variable pos;
	variable greek = [ "alpha", "beta", "gamma", "delta", "epsilon", "zeta",
	                   "eta", "theta", "iota", "kappa", "lambda", "mu", "nu",
	                   "xi", "pi", "rho", "sigma", "tau", "upsilon", "phi",
	                   "chi", "psi", "omega" ];
	variable words = [ "para",       "perp",   "PhoIndex", "nH",   "area", "center", "cutoffE", "foldE"    ];
	variable syms  = [ "\\parallel", "\\perp", "\\Gamma",  "$N_H$", "A",   "E",      "$E_cut$", "$E_fold$" ];

	%%! Copy input string to buffer
	buffer = str;

	%%! User defined renaming
	for (i = 0; i < length(par_rename); i++) {
		buffer = strreplace(buffer, par_rename[i][0], par_rename[i][1]);
	}

	%%! Replace greek letter words by greek symbols
	for (i = 0; i < length(greek); i++) {
		buffer = strreplace(buffer, greek[i], sprintf("$\\%s$", greek[i]));
	}

	%%! Replace common words by symbols
	for (i = 0; i < length(words); i++) {
		buffer = strreplace(buffer, words[i], sprintf("$%s$", syms[i]));
	}

	%%! Replace underscore with subscription
	while (match = string_matches(buffer, "\\(\$?\\)\\([\\\\a-zA-Z0-9]\+\\)\\(\\1\\)_\\(\$?\\)\\([\\\\a-zA-Z0-9]\+\\)\\(\\4\\)"), NULL != match) {

		if ("" == match[1] and "" == match[3]) {
			left = sprintf("\\mathrm{%s}", match[2]);
		} else {
			left = match[2];
		}
	
		if ("" == match[4] and "" == match[6]) {
			right = sprintf("\\mathrm{%s}", match[5]);
		} else {
			right = match[5];
		}
	
%		if (("$" == match[1] and "" == match[3]) or ("" == match[4] and "6" == match[6])) {
		buffer = sprintf("${%s}_{%s}$", left, right);
%%!FS: Currently only one subscripted variable is supported per input string
%		buffer = strreplace(buffer, match[0], sprintf("${%s}_{%s}$", left, right));
	}

	%%! Replace carret with superscription
	while (match = string_matches(buffer, "\\(\$?\\)\\([\\\\a-zA-Z0-9]\+\\)\\(\$?\\)\^\\(\$?\\)\\([\\\\a-zA-Z0-9]\+\\)\\(\$?\\)"), NULL != match) {

		if ("" == match[1] and "" == match[3]) {
			left = sprintf("\\mathrm{%s}", match[2]);
		} else {
			left = match[2];
		}
	
		if ("" == match[4] and "" == match[6]) {
			right = sprintf("\\mathrm{%s}", match[5]);
		} else {
			right = match[5];
		}
	
		buffer = strreplace(buffer, match[0], sprintf("${%s}^{%s}$", left, right));
	}

	return buffer;
}

private define par_format_number(number) {

	variable order = int(floor(log10(number)));
	variable places, str, value;

	if (number == 0.0) {
		return sprintf("0");
	}

	if ((-1 <= order and order < 2) or (0.0 == number - round(number))) {
		value = number;
	} else {
		value = number / 10^order;
	}

	if (0.0 == number - round(number)) {
		places = 0;

		str = sprintf("%.*lf", places, value);
	} else {
		places = 2;

		str = sprintf("%.*lf", places, value);

		if (0.0 == atof(str)) {
			return "0";
		}

		while (0 == strcmp("0", substr(str, strlen(str), 1))) {
			str = substr(str, 1, strlen(str)-1);
		}
	}

	if ((-1 <= order and order < 2) or (0.0 == number - round(number))) {
		return sprintf("%s", str);
%		return sprintf("%.*lf", places, number);
	}

	if (order < -1) {
		return sprintf("%s \\times 10^{%d}", str, order);
%		return sprintf("%.*lf \\times 10^{%d}", places, number / 10^order, order);
	}

	return sprintf("%s \\times 10^{%d}", str, order);
%	return sprintf("%.*lf \\times 10^{%d}", places, number / 10^order, order);
}

%%%%%%%%%%%%%%%%%
define par_save()
%%%%%%%%%%%%%%%%%
%!%+
%\function{par_save}
%\synopsis{Save fit function and parameters to labeled component in file}
%\usage{Int_Type par_save(String_Type filename [, String_Type component]);}
%\qualifiers{
%    \qualifier{dryrun}{[=0]: generate parameter file for component but
%                      do not load it, i.e., do not call load_par.
%                      Implies clean=0.}
%}
%\description
%    Saves the current fit function and parameter list
%    to the given parameter file. If a component name is given, the current fit
%    function will be saved as a named and commented out component, which can
%    loaded using par_load. Otherwise the main difference to save_par is that
%    comment lines, which are only loaded if par_load is used, are saved to the
%    parameter file as well. Calling this function is equivalent to calling
%    save_par if no comment lines are present or if the parameter file has been
%    loaded using load_par.
%    The function returns the number of components written to the file, or -1
%    upon error.
%\seealso{par_load, save_par, load_par, list_par}
%!%-
{
	variable fname, cmp, fp, strs, par, buf, buf_lines, i, k, active_cmps, active, written = 0, found = 0, first = 1;

	vmessage("WARNING: Do not use this function at the moment!\n");

	%%! Check if component parameter was given
	if (_NARGS == 0 or _NARGS > 2) {

		usage("status = par_save(filename [, component])");

	} else if (_NARGS == 1) {

		fname = ();
		cmp = NULL;

	} else {
		cmp = ();
		fname = ();
	}

	fp = fopen(fname, "w");

	if (fp == NULL) {
		throw OpenError, "fopen failed to open $fname for writing"$;
	}

	buf = get_fit_fun();

	if ((NULL != cmp) and (NULL != buf)) {
		%%! Write current fit function to commented out named component
		if (-1 == fprintf(fp, "#[%s] %s\n", cmp, buf)) {
			throw WriteError, "fprintf failed to write to $fname: "$ + errno_string(errno);
		}

		written++;
	}

	%%! Obtain individual components of current fit function
	active_cmps = get_fun_components();

	%%! Write comment lines loaded via par_load
	for (i = 0; i < length(par_lines); i++) {

		strs = string_matches(par_lines[i], "^#\\[$cmp\\]\\(.*\\)"$);

		if (NULL != strs) {

			if (written < 1) {
				%%! This line contains the old component with the given name -> ignore if overwritten already
				if (-1 == fprintf(fp, "%s\n", strs[1])) {
					throw WriteError, "fprintf failed to write to $fname: "$ + errno_string(errno);
				}
	
				%%! Setting the fit function makes it possible to use get_fun_components
				fit_fun(strs[1]);

				%%! Obtain individual components
				active_cmps = get_fun_components();
	
				written++;
			}

		} else if (0 < string_match(par_lines[i], "^ *#\\[.*\\]")) {

			%%! Commented out component line with other name -> transfer unaltered
			if (-1 == fprintf(fp, "%s", par_lines[i])) {
				throw WriteError, "fprintf failed to write to $fname: "$ + errno_string(errno);
			}

			written++;

		} else if (0 < string_match(par_lines[i], "^ *#")) {

			%%! Comment line -> transfer unaltered but do not count
			if (-1 == fprintf(fp, "%s", par_lines[i])) {
				throw WriteError, "fprintf failed to write to $fname: "$ + errno_string(errno);
			}

		}
	}

	%%! Read current parameter listing to buffer
	list_par(&buf);
	buf_lines = strtok(buf, "\n");

	for (i = 0; i < length(buf_lines); i++) {

		if ((0 == string_match(buf_lines[i], "^ *[\^#]"R)) and (1 == first)) {

			%%! First non-comment line -> save to file as default component
			if (-1 == fprintf(fp, "%s\n", buf_lines[i])) {
				throw WriteError, "fprintf failed to write to $fname: "$ + errno_string(errno);
			}

			first = 0;
			written++;
		} else {
			%%! Either the header or a parameter line
			par = string_matches(buf_lines[i], "^ *\\([0-9]\+\\) \+\\([\^. ]\+\\)");

			if (NULL != par) {
				%%! Parameter line -> transfer only active parameters (otherwise ISIS throws an error on load_par)
				k = 0;
				active = 0;

				while ((k < length(active_cmps)) and (0 == active)) {
					if (0 == strcmp(par[2], active_cmps[k])) {
						active = 1;
					} else {
						k++;
					}
				}

				if (active) {
					%%! Parameter is needed for current component -> transfer unaltered
					if (-1 == fprintf(fp, "%s\n", buf_lines[i])) {
						throw WriteError, "fprintf failed to write to $fname: "$ + errno_string(errno);
					}
				} else {
					%%! Parameter is not needed for current component -> comment out
					if (-1 == fprintf(fp, "# %s\n", buf_lines[i])) {
						throw WriteError, "fprintf failed to write to $fname: "$ + errno_string(errno);
					}
				}

			} else {
				%%! Must be the header line -> transfer unaltered
				if (-1 == fprintf(fp, "%s\n", buf_lines[i])) {
					throw WriteError, "fprintf failed to write to $fname: "$ + errno_string(errno);
				}
			}
		}
	}

	() = fclose(fp);

	return written;
}

%%%%%%%%%%%%%%%%%
define par_list()
%%%%%%%%%%%%%%%%%
%!%+
%\function{par_list}
%\synopsis{Print a parameter file listing}
%\usage{Int_Type par_list([String_Type filename | Ref_Type buffer | File_Type fd]);}
%\qualifiers{
%    \qualifier{tex}{[=0]: Format parameter listing in form of a LATEX tabular
%                       environment. This does currently not work for a Ref_Type
%                       argument.}
%    \qualifier{pdf}{[=0]: Compile a pdf file containing the parameter listing
%                       as a table.}
%    \qualifier{all}{[=0]: Add all parameters to tex and pdf file. By default
%                       parameters with the following name - value combinations
%                       are excluded:
%                           norm    1}
%    \qualifier{exclude}{[=0]: Array or list of parameters to be excluded in
%                       tex and pdf listings. These are even excluded if the
%                       "all" qualifier above is set.}
%    \qualifier{include}{[=0]: Array or list of additional parameters to be
%                       included in tex and pdf listings.}
%}
%\description
%    Prints all internally stores parameter file lines to screen. This includes
%    out-commented fit functions and parameter lines, which are only loaded if
%    par_load is used instead of load_par. Writes parameter listing to file if
%    a String_Type parameter is given, which will be interpreted as filename.
%    If a File_Type parameter is given the listing will be written to this
%    file. If the argument is of Ref_Type, the listing will be returned in this
%    buffer. This argument functionality works analog to the default list_par
%    function.
%\seealso{par_load, par_save, save_par, load_par, list_par}
%!%-
{
	variable i, j, fpar = NULL, buffer, line, idx, modname, modinst, parname, tie, freeze, parval, parmin, parmax, parunit, strs, fp = stdout, written = -1, total = 0, tex = 0, pdf = NULL, pdftex_fp, all = 0;

	variable exclude_name  = [ "norm" ];
	variable exclude_value = [  1.0   ];
	variable exclude_idx, exclude_user, include_user = {};

	%%! Check if component parameter was given
	if (_NARGS > 1) {
		usage("par_list([filename|buffer|fp])");
	}

	if ((1 == qualifier_exists("tex")) and (0 != qualifier("tex", 0))) {
		tex = 1;
	}

	if ((1 == qualifier_exists("all")) and (0 != qualifier("all", 0))) {
		all = 1;
	}

	if (1 == qualifier_exists("exclude")) {
		exclude_user = qualifier("exclude", {});
	}

	if (1 == qualifier_exists("rename")) {
		par_rename = qualifier("rename", {});
	}

	if (1 == qualifier_exists("include")) {
		include_user = qualifier("include", {});
	}

	if (_NARGS == 1) {

		fpar = ();

		if (String_Type == typeof(fpar)) {

			fp = fopen(fpar, "w");
		
			if (fp == NULL) {
				throw OpenError, "fopen failed to open $fpar for writing"$;
			}

		} else if (File_Type == typeof(fpar)) {

			fp = fpar;

		} else if (Ref_Type == typeof(fpar)) {
			@fpar = "";
		}
	}

	if (pdf = qualifier("pdf", NULL), NULL != pdf) {

		pdftex_fp = fopen("$pdf.tex"$, "w");

		if (NULL == pdftex_fp) {
			throw OpenError, "fopen failed to open $pdf.tex for writing"$;
		}
	
		buffer = par_format_text("\\documentclass{article}\n\\usepackage{booktabs}\n\\begin{document}\n\\begin{table}\n\\centering\n");

		if (-1 == fprintf(pdftex_fp, "%s", buffer)) {
			throw WriteError, "printf failed to write to $pdf.tex: "$ + errno_string(errno);
		}
	}

	buffer = par_format_text("\\begin{tabular}[b]{ccc}\n\\toprule\nParameter & Value & Unit \\\\\n\\midrule\n");

	if (NULL != pdf) {
		if (-1 == fprintf(pdftex_fp, "%s", buffer)) {
			throw WriteError, "printf failed to write to $pdf.tex: "$ + errno_string(errno);
		}
	}

	if (tex) {

		if (Ref_Type == typeof(fpar)) {
			@fpar += buffer;
		} else {
			if (written = fprintf(fp, "%s", buffer), written == -1) {
				throw WriteError, "printf failed to write to $fp: "$ + errno_string(errno);
			}
	
			total += written;
		}
	}

	for (i = 0; i < length(par_lines); i++) {

		strs = string_matches(par_lines[i], par_regexp);

		if (NULL != strs) {

			line    = strs[0];
			idx     = integer(strs[1]);
			modname = strs[2];
			modinst = integer(strs[3]);
			parname = strs[4];
			tie     = integer(strs[5]);
			freeze  = integer(strs[6]);
			parval  = atof(strs[7]);
			parmin  = atof(strs[8]);
			parmax  = atof(strs[9]);

			strs = string_matches(strs[0], ".* \+\\([\^ ]\+\\)$");

			if (length(strs) > 1) {
				parunit = strs[1];
			} else {
				parunit = "";
			}

			%%! Exclude parameters from exclude list
			if (0 <= par_array_idx(exclude_user, parname)) {
				continue;
			}

			%%! Excluded parameter/value combinations
			if (exclude_idx = par_array_idx(exclude_name, parname),
				(not all) and (exclude_idx >= 0) and
				(parval == exclude_value[exclude_idx])) {
				continue;
			}

			%%! Exclude enflux E_min/E_max (included in enflux parameter below)
			if ((0 == strcmp("enflux", modname)) and
				((0 == strcmp("E_min", parname)) or (0 == strcmp("E_max", parname)))) {
				continue;
			}

			%%! Show enflux model parameter in one row
			if (0 == strcmp("enflux", parname)) {

				variable tmp, tmp2, tmp_unit;
				variable E_min = NULL, E_max = NULL;
				variable E_min_unit = NULL, E_max_unit = NULL;

				for (j = 0; j < length(par_lines); j++) {
					tmp = string_matches(par_lines[j], par_regexp);

					if (tmp != NULL) {

						if ((0 == strcmp(modname, tmp[2])) and
							(modinst == integer(tmp[3]))) {

							tmp2 = string_matches(tmp[0], ".* \+\\([\^ ]\+\\)$");

							if (length(tmp2) > 1) {
								tmp_unit = tmp2[1];
							} else {
								tmp_unit = "";
							}

							if (0 == strcmp("E_min", tmp[4])) {

								E_min = atof(tmp[7]);
								E_min_unit = tmp_unit;

							} else if (0 == strcmp("E_max", tmp[4])) {

								E_max = atof(tmp[7]);
								E_max_unit = tmp_unit;
							}

							if ((NULL != E_min) and (NULL != E_max)) {
								break;
							}
						}
					}
				}

				parname = sprintf("$F_{%s\\,\\mathrm{%s}}^{%s\\,\\mathrm{%s}}$",
					par_format_number(E_min), par_format_text(E_min_unit),
					par_format_number(E_max), par_format_text(E_max_unit));
			}

			buffer = sprintf("%s & $%s$ & %s \\\\\n", par_format_text(parname),
				par_format_number(parval), par_format_text(parunit));

			if (NULL != pdf) {
				if (-1 == fprintf(pdftex_fp, "%s", buffer)) {
					throw WriteError, "printf failed to write to $pdf.tex: "$ + errno_string(errno);
				}
			}
		} else {
			buffer = NULL;
		}

		if (tex and (NULL != buffer)) {
			if (Ref_Type == typeof(fpar)) {
				@fpar += buffer;
			} else {
				if (written = fprintf(fp, "%s", buffer), written == -1) {
					throw WriteError, "printf failed to write to $fp: "$ + errno_string(errno);
				}
			}
		} else if (not tex) {
			if (Ref_Type == typeof(fpar)) {
				@fpar += par_lines[i];
			} else {
				if (written = fprintf(fp, "%s", par_lines[i]), written == -1) {
					throw WriteError, "printf failed to write to $fp: "$ + errno_string(errno);
				}
			}
		}

		total += written;
	}

	for (i = 0; i < length(include_user); i++) {

		buffer = sprintf("%s & $%s$ & %s \\\\\n", par_format_text(include_user[i][0]),
			par_format_number(include_user[i][1]), par_format_text(include_user[i][2]));

		if (NULL != pdf) {
			if (-1 == fprintf(pdftex_fp, "%s", buffer)) {
				throw WriteError, "printf failed to write to $pdf.tex: "$ + errno_string(errno);
			}
		}

		if (tex and (NULL != buffer)) {
			if (Ref_Type == typeof(fpar)) {
				@fpar += buffer;
			} else {
				if (written = fprintf(fp, "%s", buffer), written == -1) {
					throw WriteError, "printf failed to write to $fp: "$ + errno_string(errno);
				}
			}
		}

		total += written;
	}

	buffer = par_format_text("\\bottomrule\\end{tabular}\n");

	if (NULL != pdf) {
		if (-1 == fprintf(pdftex_fp, "%s", buffer)) {
			throw WriteError, "printf failed to write to $pdf.tex: "$ + errno_string(errno);
		}
	}

	if (tex) {
		if (Ref_Type == typeof(fpar)) {
			@fpar += buffer;
		} else {
			if (written = fprintf(fp, "%s", buffer), written == -1) {
				throw WriteError, "printf failed to write to $fp: "$ + errno_string(errno);
			}
		}

		total += written;
	}

	if (NULL != pdf) {
		buffer = par_format_text("\\end{table}\n\\end{document}");

		if (-1 == fprintf(pdftex_fp, buffer)) {
			throw WriteError, "printf failed to write to $pdf.tex: "$ + errno_string(errno);
		}
	}

	if (String_Type == typeof(fpar)) {
		() = fclose(fp);
	}

	if (NULL != pdf) {
		() = fclose(pdftex_fp);
		vmessage("pdflatex --shell-escape $pdf"$);
		() = system("pdflatex --shell-escape $pdf.tex"$);
		() = system("rm $pdf.log $pdf.aux");
	}

	if (Ref_Type == typeof(fpar)) {
		total = strlen(@fpar);
	}

	if (fpar != NULL) {
		return total;
	} else {
		return;
	}
}

%
% Code from src/slang/structs/tables/par2tex.sl
%

% only activate if module pcre could be loaded
#ifeval __get_reference("pcre_compile")!=NULL

private define get_mod_num(name) {
  return pcre_matches ([0-9],name)[0];
}

define par_names_to_tex() {
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %!%+
  %\function{par_names_to_tex}
  %\synopsis{Extract the function name, LaTeX parameter name
  %and unit from a parameter string such as
  %'powerlaw(2).PhoIndex' as provided by fits_load_fit_struct.}
  %\usage{ String_Type (name,fun,unit) = par_names_to_tex(paramater-string,unit);}
  %\description
  %   This function is required by par2tex.
  %\example
  %   par_names_to_tex("powerlaw(2).PhoIndex");
  %   output: "PL 2", "$\Gamma$", ""
  %\seealso{fits_save_fit, fits_load_fit_struct, par2tex, table_print_TeX}
  %!%-
  variable name, unit, fun;
  switch(_NARGS)
  {case 2 : (name,unit)=();}
  { help(_function_name()); return; }

  if(is_substr(name, "detconst") > 0)
  {
    if (is_substr(name , "XIS0") >0) fun = "XIS\,0"R;
    if (is_substr(name , "XIS1") >0) fun = "XIS\,1"R;
    if (is_substr(name , "XIS3") >0) fun = "XIS\,3"R;
    if (is_substr(name , "HXD") >0) fun = "HXD";
    if (is_substr(name , "PIN") >0) fun = "HXD";
    name = "Detconst";
  }
  if(is_substr(name, "absori") > 0)
  {
    if (is_substr(name , "PhoIndex") >0) fun = "$\\Gamma$";
    if (is_substr(name , "nH") >0) fun = "$N_\text{H}$\,($10^{22}\mathrm{cm}^{-2}$)"R;
    if (is_substr(name , "Redshift") >0) fun = "$z$";
    if (is_substr(name , "Temp_abs") >0) fun = "$T_{\\text{abs}}$ (K)";
    if (is_substr(name , "xi") >0) fun = "$\\xi$";
    if (is_substr(name , "Fe_abund") >0) fun = "${\tt Fe}-abund$";
    name = sprintf("WA %s",get_mod_num(name));
  }
  if(is_substr(name, "zpcfabs") > 0)
  {
    if (is_substr(name , "nH") >0) fun = "$N_\text{H}$\,($10^{22}\mathrm{cm}^{-2}$)"R;
    if (is_substr(name , "CvrFract") >0) fun = "f$_\\text{cvr}$";
    if (is_substr(name , "Redshift") >0) fun = "$z$";
    name = "Partial covering with cold absorption";
  }
  if(is_substr(name , "zxipcf") >0)
  {
    if (is_substr(name , "Nh") >0) fun = "$N_\text{H}$\,($10^{22}\mathrm{cm}^{-2}$)"R;
    if (is_substr(name , "log") >0) fun = "$\log\xi$\,($\mathrm{erg}\,\mathrm{cm}\,\mathrm{s}^{-1}$)"R;
    if (is_substr(name , "CvrFract") >0) fun = "f$_\\text{cvr}$";
    if (is_substr(name , "Redshift") >0) fun = "$z$";
    name = sprintf("WA %s",pcre_matches ([0-9],name)[0]);
  }

  if(is_substr(name , "ztbabs")>0)
  {
    if (is_substr(name , "nH") >0) fun = "$N_\text{H}$\,($10^{22}\mathrm{cm}^{-2}$)"R;
    if (is_substr(name , "Redshift") >0) fun = "$z$";
    name = "Internal cold absorption";
  }
  if(is_substr(name , "powerlaw")>0)
  {
    if (is_substr(name , "norm") >0) fun = "norm";
    if (is_substr(name , "PhoIndex") >0) fun = "$\\Gamma$";
    name = sprintf("PL %s",get_mod_num(name));
  }
  if(is_substr(name , "cutoffpl")>0)
  {
    if (is_substr(name , "norm") >0) fun = "norm";
    if (is_substr(name , "PhoIndex") >0) fun = "$\\Gamma$";
    if (is_substr(name , "HighECut") >0) fun = "$\\mathrm{E}_\mathrm{c}$";
    name = sprintf("PL %s",get_mod_num(name));
  }
  if(is_substr(name , "zgauss")>0)
  {
    if (is_substr(name , "norm") >0) fun = "norm";
    if (is_substr(name , "LineE") >0) fun = "Energy\\,(keV)";
    if (is_substr(name , "Sigma") >0) fun = "$\\sigma$\\,(keV)";
    if (is_substr(name , "Redshift") >0) fun = "$z$";
    name = sprintf("Gauss %s",get_mod_num(name));
  }
  if(is_substr(name , "redge")>0)
  {
    if (is_substr(name , "norm") >0) fun = "norm";
    if (is_substr(name , "edge") >0) fun = "Energy\\,(keV)";
    if (is_substr(name , "kT") >0) fun = "$kT$\\,(keV)";
    name = sprintf("Redge %s",get_mod_num(name));
  }
  if(is_substr(name , "egauss")>0)
  {
    if (is_substr(name , "area") >0) fun = "area";
    if (is_substr(name , "center") >0) fun = "Energy\\,(keV)";
    if (is_substr(name , "sigma") >0) fun = "$\\sigma$\\,(keV)";
    name = sprintf("Gauss %s",get_mod_num(name));
  }
  if(is_substr(name , "tbabs")>0)
  {
    if (is_substr(name , "nH") >0) fun = "$N_\text{H,Gal}$\,($10^{22}\mathrm{cm}^{-2}$)"R;
    name = "Galactic absorption";
  }
  if((is_substr(name , "tbnew_simple")>0) && (is_substr(name , "tbnew_simple_z")==0))
  {
    if (is_substr(name , "nH") >0) fun = "$N_\text{H,Gal}$\,($10^{22}\mathrm{cm}^{-2}$)"R;
    name = "CA";
  }
  if(is_substr(name , "tbnew_simple_z")>0)
  {
    if (is_substr(name , "nH") >0) fun = "$N_\text{H,int}$\,($10^{22}\mathrm{cm}^{-2}$)"R;
    if (is_substr(name , ".z") >0) fun = "$z$";
    name = sprintf("CA",get_mod_num(name));
  }
  if(is_substr(name , "comptt")>0)
  {
    if (is_substr(name , "norm") >0) fun = "norm";
    if (is_substr(name , "T0") >0) fun = "T$_{0}$\\,(keV)";
    if (is_substr(name , "kT") >0) fun = "$k_\\text{B}T$\\,(keV)";
    if (is_substr(name , "taup") >0) fun = "$\\tau$";
    if (is_substr(name , "Redshift") >0) fun = "$z$";
    if (is_substr(name , "approx") >0) fun = "";
    name = "Comptonization of soft photons";
  }
  if(is_substr(name , "constant")>0)
  {
    if (is_substr(name , "factor") >0) fun = "f$_\\text{cvr}$";
    name = "Covering Factor";
  }
  if(is_substr(name , "xillver")>0)
  {
    if (is_substr(name , "norm") >0) fun = "norm";
    if (is_substr(name , "Gamma") >0) fun = "$\\Gamma$";
    if (is_substr(name , "logXi") >0) fun = "$\log\xi$\,($\mathrm{erg}\,\mathrm{cm}\,\mathrm{s}^{-1}$)"R;
    if (is_substr(name , "A_Fe") >0) fun = "$A_\\mathrm{Fe}$";
    if (is_substr(name , "redshift") >0) fun = "$z$";
    name = "Ion. Refl.";
  }
  if(is_substr(name , "relxill")>0)
  {
    if (is_substr(name , "norm") >0) fun = "norm";
    if (is_substr(name , "Index1") >0) fun = "Index 1";
    if (is_substr(name , "Index2") >0) fun = "Index 2";
    if (is_substr(name , "Rbr") >0) fun = "break rad.";
    if ((is_substr(name , ".a") >0) && (is_substr(name ,".angleon") ==0)) fun = "$a$";
    if (is_substr(name , "Incl") >0) fun = "$i$";
    if (is_substr(name , "Rin") >0) fun = "$R_\mathrm{in}$"R;
    if (is_substr(name , "Rout") >0) fun = "$R_\mathrm{out}$"R;
    if (is_substr(name , ".z") >0) fun = "$z$";
    if (is_substr(name , "gamma") >0) fun = "$\Gamma$"R;
    if (is_substr(name , "logxi") >0) fun = "$\log\xi$\,($\mathrm{erg}\,\mathrm{cm}\,\mathrm{s}^{-1}$)"R;
    if (is_substr(name , "Afe") >0) fun = "$Z_\\mathrm{Fe}$";
    if (is_substr(name , "Ecut") >0) fun = "$E_\mathrm{cut}$"R;
    if (is_substr(name , "refl_frac") >0) fun = "$R_\mathrm{f}$";
    if (is_substr(name , "angleon") >0) fun = "angle averaging";
    name = "Relativ. Refl.";
  }
  if(is_substr(name , "relxilllp")>0)
  {
    if (is_substr(name , "norm") >0) fun = "norm";
    if (is_substr(name , ".h") >0) fun = "height";
    if ((is_substr(name , ".a") >0) && (is_substr(name ,".angleon") ==0)) fun = "$a$";
    if (is_substr(name , "Incl") >0) fun = "$i$";
    if (is_substr(name , "Rin") >0) fun = "$R_\mathrm{in}$"R;
    if (is_substr(name , "Rout") >0) fun = "$R_\mathrm{out}$"R;
    if (is_substr(name , ".z") >0) fun = "$z$";
    if (is_substr(name , "gamma") >0) fun = "$\Gamma$"R;
    if (is_substr(name , "logxi") >0) fun = "$\log\xi$\,($\mathrm{erg}\,\mathrm{cm}\,\mathrm{s}^{-1}$)"R;
    if (is_substr(name , "Afe") >0) fun = "$Z_\\mathrm{Fe}$";
    if (is_substr(name , "Ecut") >0) fun = "$E_\mathrm{cut}$"R;
    if (is_substr(name , "refl_frac") >0) fun = "$R_\mathrm{f}$"R;
    if (is_substr(name , "fixReflFrac") >0) fun = "freeze refl. fract.";
    if (is_substr(name , "angleon") >0) fun = "angle averaging";
    name = "Relativ. Refl.";
  }

  if(unit!=NULL && is_substr(unit , "10^22")>0)
  {
    unit = "$\\times\\,10^{22}$";
  }

  return name,fun,unit;

}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define par2tex()
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %!%+
  %\function{par2tex}
  %\synopsis{Write a LaTeX parameter table from a fit saved with fits_save_fit}
  %\usage{ Integer_Type status = par2tex(fits-fit);}
  %\qualifiers{
  %\qualifier{Obsid}{observation identifier (e.g. obsid)}
  %\qualifier{Src}{source string}
  %}
  %\description
  %   Uses fit saved with fits_save_fit (basename.fits) to write
  %   a LaTeX parameter table (basename.table). The functions
  %   must be defined in the called function 'par_names_to_tex'!
  %   ---- This is the updated version of table_print_TeX ------
  %\example
  %   par2tex("basename.fits";"0102030405",Src="NGC 1234");
  %\seealso{fits_save_fit, par_names_to_tex, table_print_TeX}
  %!%-
{
  variable fits;
  switch(_NARGS)
  {case 1 : fits=();}
  { help(_function_name()); return; }

  variable obsid = qualifier("Obsid","");

  variable i,j,k,l,p,f,fitfun,fall;

  fall=fopen(sprintf("%s/%s.table",path_dirname(fits),path_basename_sans_extname(fits)),"w");
  (qualifier_exists("loadscript") ? evalfile(qualifier("loadscript")) : fits_load_fit(sprintf("%s",fits);ROC=0));
  variable fs = fits_load_fit_struct(sprintf("%s",fits));
  fitfun = get_fit_fun;

  ()=fputs("\\begin{tabular}[ht]{llll}\n",fall);
  ()=fputs("\\midrule\\midrule\n",fall);
  ()=fputs(sprintf("\\multicolumn{4}{c}{%s}\\\\\n",fitfun),fall);
  ()=fputs("\\midrule\n",fall);
  ()=fputs("model component & Parameter & freeze & value\\\\\n",fall);
  ()=fputs("\\midrule\n",fall);

  p = get_params;
  variable fun = String_Type[length(p)];
  variable count = Integer_Type[length(p)];
  variable stat_info;
  eval_counts(&stat_info);
  variable stat = stat_info.statistic;
  variable free = stat_info.num_bins - stat_info.num_variable_params;

  delete_data(1);
  delete_arf(1);
  delete_rmf(1);

  ()=fputs(sprintf("\\multicolumn{4}{c}{%s: $\\chi^2$\\,(dof) = %.2f\\,(%u)}\\\\\n",obsid,stat,free),fall);
  ()=fputs("\\midrule\n",fall);

  variable conf_name_split;
  variable conf_name = String_Type[length(p)];
  % read conf levens from fits_save
  for (i=0;i<length(p);i++) {
    conf_name_split = pcre_matches (`(.*?)(\(.*?\).)(.*)`, p[i].name);
    conf_name[i] = sprintf("%s_%s_%s_conf",conf_name_split[1],pcre_matches (`[0-9]`,p[i].name)[0],conf_name_split[3]);
    conf_name[i] = strreplace(conf_name[i],"(","_");
    conf_name[i] = strreplace(conf_name[i],")","_");
    %    vmessage("%s: %.2e < %.2e < %.2e",conf_name[i],get_struct_field(fs,conf_name[i])[0,0],p[i].value,get_struct_field(fs,conf_name[i])[0,1]);
    p[i].min = p[i].value-get_struct_field(fs,conf_name[i])[0,0];
    p[i].max = get_struct_field(fs,conf_name[i])[0,1]-p[i].value;
    %    vmessage("%s: min=%.2e, max=%.2e",conf_name[i],p[i].min,p[i].max);
  }

  for (i=0;i<length(p);i++)
  {
    variable name_tmp,fun_tmp,unit_tmp;
    (name_tmp,fun_tmp,unit_tmp)=par_names_to_tex(p[i].name,p[i].units);
    %    vmessage("fun = %s",fun_tmp);
    p[i].name=name_tmp;
    p[i].units=unit_tmp;
    fun[i]=fun_tmp;
  }

  % eliminate double par names
  count = Integer_Type[length(p)];
  for (j=0;j<length(p)-1;j++)
  {
    if (p[j+1].name == p[j].name)
    {
      count[j+1] = count[j]+1;
    }
  }
  for (j=0;j<length(p);j++)
  {
    if (count[j] > 0) p[j].name = "";
  }

  for (i=0;i<length(p);i++)
  {
    %    vmessage("parname=%s",p[i].name);
    vmessage("%s",conf_name[i]);
    if (get_struct_field(fs,conf_name[i])[0,0] != 0.) {
      variable tex = TeX_value_pm_error (p[i].value,p[i].value-p[i].min,p[i].value+p[i].max);
      ()=fputs(sprintf("%s & %s & %d & %s\\\\\n",p[i].name,fun[i],p[i].freeze,tex),fall);
    }
    else {
      %      vmessage("%.2f | min=%.2f  max=%.2f",p[i].value,p[i].min,p[i].max);
      ()=fputs(sprintf("%s & %s & %d & %.3g\\\\\n",p[i].name,fun[i],p[i].freeze,p[i].value),fall);
    }
  }

  ()=fputs("\\bottomrule\n",fall);
  ()=fputs("\\end{tabular}\n\n",fall);
  fclose(fall);

  variable status;
  if (length(glob(sprintf("%s/%s.table",path_dirname(fits),path_basename_sans_extname(fits)))) == 1) status = 1;
  else status = 0;

  return status;
}

#endif

%
% Code from src/fitting/parameters/par_bounds.sl
%
% there is 'check_parameter_hit' but it is a: not documented and b: not very useful

define par_bounds ()
%!%+
%\function{par_bounds}
%\synopsis{Return parameter index of parameters hitting bounds}
%\usage{Int_Type[] par_bounds();}
%\qualifiers{
%  \qualifier{tolerance}{[=0.01] Boundary hit tolerance}
%}
%\description
%  This function returns the parameter index for every parameter with satisfies
%  (par.max-par.value)/(par.max-par.min) <= tolerance or
%  (par.value-par.min)/(par.max-par.min) <= tolerance.
%\seealso{list_par_bounds, list_free_bounds}
%!%-
{
  variable tol = _min(_max(qualifier("tolerance", 0.01), 1.), 0);

  variable par, p, r = {};

  par = get_params();
  if (NULL == par)
    return NULL;

  foreach p (par) {
    if (tol >= (p.max-p.value)/(p.max-p.min)
	|| tol >= (p.value-p.min)/(p.max-p.min))
      list_append(r, p.index);
  }

  return (length(r)) ? list_to_array(r) : Int_Type[0];
}

private variable __bounds_header = "%s\n idx%-*s          min             <bound>            max";
private variable __bounds_format = "%3u  %-*s  %10.7g %30s %-10.7g";
private variable __bounds_default = ('-')[[0:29]/30];

private define __bounds_lines (p, tol, max_name_len) {
  variable idicator;
  variable left, right;
  variable sector;
  variable factor;
  variable str;

  if (p.min == p.max) {
    left = '|'; right = '|';
    sector = 12;
    factor = 0.;
  } else if (tol >= (p.max-p.value)/(p.max-p.min)) {
    left = '<'; right = '|';
    factor = +.5;
    sector = 24;
  } else if (tol >= (p.value-p.min)/(p.max-p.min)) {
    left = '|'; right = '>';
    factor = -.5;
    sector = 0;
  } else {
    left = '<'; right = '>';
    factor = (p.value-.5*(p.max+p.min))/(p.max-p.min);
    sector = int(round((p.value-p.min)/(p.max-p.min)*24));
  }

  str = @__bounds_default;
  str[[sector:sector+5]] = string_to_wchars(sprintf("<%c.%02d>",
							(sign(factor)<0) ? '-' : '+',
							int(abs(factor*100))));
  str[0] = left; str[-1] = right;
  vmessage(__bounds_format, p.index, max_name_len, p.name, p.min, wchars_to_string(str), p.max);
}

define list_par_bounds ()
%!%+
%\function{list_par_bounds}
%\synopsis{List parameter range bounds}
%\usage{list_par_bounds();}
%\qualifiers{
%  \qualifier{tolerance}{[=0.01] Boundary hit tolerance}
%}
%\description
%  This function reports the values of the fit parameters relative to the
%  allowed parameter range. The value is indicated in a graphical indicator
%  where the parameter lies (from -0.5 to +0.5 around the center). If the
%  parameter value is within <tolerance> close to the range minimum or maximum
%  the indicator at the correspongind limit changes from '<' or '>' to '|'.
%  In the case of min == max both ends are displayed as '|'.
%\seealso{list_free_bounds, par_bounds}
%!%-
{
  variable tol = _min(_max(qualifier("tolerance", 0.01), 1.), 0);

  variable par,p;
  variable len = 0;

  par = get_params();
  if (NULL == par)
    return;

  foreach p (par)
    len = max([len, strlen(p.name)]);

  vmessage(__bounds_header, get_fit_fun(), len, "  param");

  variable idicator;
  variable left, right;
  variable sector;
  variable factor;
  foreach p (get_params)
    __bounds_lines (p, tol, len);
}

define list_free_bounds ()
%!%+
%\function{list_free_bounds}
%\synopsis{List parameter range bounds for free parameters}
%\usage{list_free_bounds();}
%\qualifiers{
%  \qualifier{tolerance}{[=0.01] Boundary hit tolerance}
%}
%\description
%  This function reports the parameter boundaries for all free parameters.
%  For more details see help for \code{list_par_bounds}.
%\seealso{list_par_bounds, par_bounds}
%!%-
{
  variable tol = _min(_max(qualifier("tolerance", 0.01), 1.), 0);

  variable par,p;
  variable len = 0;

  par = get_params();
  if (NULL == par)
    return;

  foreach p (par)
    len = max([len, strlen(p.name)]);

  vmessage(__bounds_header, get_fit_fun(), len, "  param");

  variable idicator;
  variable left, right;
  variable sector;
  variable factor;
  foreach p (get_params) {
    ifnot (p.freeze || p.tie != NULL || p.fun != NULL)
      __bounds_lines (p, tol, len);
  }
}

%
% Code from src/slang/date/parseDateString.sl
%
define parseDateString()
%!%+
%\function{parseDateString}                                                                                         
%\synopsis{parses a string including year, month, day [, hour [, minute [, second]]]}
%\usage{Struct_Type parseDateString(String_Type date_string);}
%\qualifiers{
%\qualifier{no_2digit_correction}{prevents the correction of two digit year specification,
%                              e.g., 83/12/13 instead of 1983/12/13. If this qualifier
%                              is not set, 2000 is added to years from 00-19 and
%                              1900 to years 20-99.}
%}
%\description
%    This function tries to interprete the given string as a data and returns
%    a structure consistent with the return value of `localtime'. It is
%    assumed that the first number occuring in the string specifies the year,
%    the second the month, and the third the day. If further numbers are found,
%    they are interpreted as hour, minute, second and fraction of seconds. Note
%    that the seconds contained in the structure as defined in `localtime' has
%    to be an integer, while here a double is used in case of fraction of
%    seconds. There can be any separators between the numbers (except numbers).
%\example
%    parseDateString("69/07/21");
%    parseDateString("1969-07-21 02:56:30.44");
%    parseDateString("sdf1969--X07..21QQ_02/56((30,,44xyz");
%\seealso{localtime}
%!%-
{
  variable str;
  switch(_NARGS)
  { case 1: str = (); }
  { help(_function_name()); return;}

  if (length(str)>1)
  { throw RunTimeError, _function_name()+": vector argument cannot be handled"; }
  
  variable matched = 1;
  variable res = {};
  variable len, start, l;
  
  while (matched)
  {
    len = strcharlen (str);
    matched = string_match (str,"\([0-9]+\)"R);
    if (matched)
    {
      (start, l) = string_match_nth(1);
      list_append (res, str[[start:start+l-1]]);
      if ( start+l >= len)
	matched = 0;
      else
	str = str[[start+l:]];
    }
  }

  res = array_map (Integer_Type, &atoi, list_to_array(res));
  variable len_res = length (res);

  % correct two digit years, e.g., 83/12/13 for 1983-12-13 (often used in old radio data)
  ifnot (qualifier_exists("no_2digit_correction"))
    if (len_res >=1 && res[0] < 100) res[0] += (res[0]<20 ? 2000: 1900);
  
  if (len_res < 3)
  { throw RunTimeError, _function_name()+": no proper date found (less than three numbers)"; }
  
  variable i;
  variable d = Integer_Type[7];
  _for i (0,nint(_min(len_res-1,6)),1)
  {
    d[i] = res[i];
  }
  variable sec = d[5];
  if ( d[6] > 0 ) sec = d[5] + double(d[6])/10^(ceil(log10(d[6])));

  % calculate day of year
  variable yday = 0;
  _for i (1, d[1]-1, 1) { yday += daysPerMonth(i, d[0]); }
  yday += d[2];

  % calculate day of week
  variable wday = dayOfWeek(d[0], d[1], d[2]);
  
  return struct {
    tm_sec = sec, tm_min = d[4], tm_hour = d[3],
    tm_mday = d[2], tm_mon = d[1] - 1, tm_year = d[0] - 1900,
    tm_wday = wday mod 7, tm_yday = yday, tm_isdst = -1
  };
}

%
% Code from src/data/partial_correlation.sl
%
private define h_akritas(i,j,k,l,idat,dat)
{
    variable cj1, cj2;    
    cj1 = - idat[j,k];

    if (dat[i,k] < dat[j,k])	
    {
	cj1=idat[i,k];
    }

    cj2 = - idat[j,l];
    if(dat[i,l] < dat[j,l]) 
    {
	cj2=idat[i,l];
    }
    
    return cj1*cj2;
}

private define tau_kendall(k,l,ntot,idat,dat)
{
    variable sum = 0.0;
    variable i,j;
    variable ac = 2.0/(ntot*(ntot-1));

    _for j (0, ntot-1,1)
    {
	_for i (0,ntot-1,1)
	{
	    if (i >= j)
	    {
		break;
	    }
	    sum = sum + h_akritas(i,j,k,l,idat,dat);
	}
    }
    return sum * ac;
}

private define tau_kendall_partial(ntot,k1,k2,k3,idat,dat)
{
    return (tau_kendall(k1,k2,ntot,idat,dat)-tau_kendall(k1,k3,ntot,idat,dat)*
    tau_kendall(k2,k3,ntot,idat,dat)) / 
    sqrt((1.0-tau_kendall(k1,k3,ntot,idat,dat)^2)*
    (1.0-tau_kendall(k2,k3,ntot,idat,dat)^2));
}

private define an_akritas(ntot,idat,dat,k1,k2,k3)
{
    variable c1 = 16.0/(ntot-1);
    variable c2 = 6.0/((ntot-1)*(ntot-2)*(ntot-3));
    variable asum = 0.0;
    variable ave = 0.0;
    variable aasum = Double_Type[ntot];
    variable i1, i2, j1, j2, i;

    _for i1 (0, ntot-1,1)
    {
	_for j1 (0, ntot-3,1)
	{
	    %%% inner summation with j1<i2<j2 and all != i1
	    if(j1 == i1) 
	    {
		continue;
	    }
	    
	    _for j2(j1+2, ntot-1,1)
	    {
		if(j2 == i1)
		{
		    continue;
		}
		_for i2 (j1+1, j2-1,1)
		{
		    if(i2 == i1)
		    {
			continue;
		    }
		    
		    variable cj1, cj2, cj3, cj4, cj5, cj6, cj7;
		    variable gtsum = 0.0;

		    cj1=- idat[j1,k1];
		    if(dat[i1,k1]<dat[j1,k1]) cj1=idat[i1,k1];
		    cj2=- idat[j1,k2];
		    if(dat[i1,k2]<dat[j1,k2]) cj2=idat[i1,k2];
		    cj3=- idat[j1,k3];
		    if(dat[i1,k3]<dat[j1,k3]) cj3=idat[i1,k3];
		    cj4=- idat[j2,k2];
		    if(dat[i2,k2]<dat[j2,k2]) cj4=idat[i2,k2];
		    cj5=- idat[j2,k3];
		    if(dat[i2,k3]<dat[j2,k3]) cj5=idat[i2,k3];
		    cj6=- idat[i2,k2];
		    if(dat[j2,k2]<dat[i2,k2]) cj6=idat[j2,k2];
		    cj7=- idat[i2,k3];
		    if(dat[j2,k3]<dat[i2,k3]) cj7=idat[j2,k3];
		    gtsum=cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    cj1=- idat[j2,k1];
		    if(dat[i1,k1]<dat[j2,k1]) cj1=idat[i1,k1];
		    cj2=- idat[j2,k2];
		    if(dat[i1,k2]<dat[j2,k2]) cj2=idat[i1,k2];
		    cj3=- idat[j2,k3];
		    if(dat[i1,k3]<dat[j2,k3]) cj3=idat[i1,k3];
		    cj4=- idat[j1,k2];
		    if(dat[i2,k2]<dat[j1,k2]) cj4=idat[i2,k2];
		    cj5=- idat[j1,k3];
		    if(dat[i2,k3]<dat[j1,k3]) cj5=idat[i2,k3];
		    cj6=- idat[i2,k2];
		    if(dat[j1,k2]<dat[i2,k2]) cj6=idat[j1,k2];
		    cj7=- idat[i2,k3];
		    if(dat[j1,k3]<dat[i2,k3]) cj7=idat[j1,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[i2,k1];
		    if(dat[i1,k1]<dat[i2,k1]) cj1=idat[i1,k1];
		    cj2=- idat[i2,k2];
		    if(dat[i1,k2]<dat[i2,k2]) cj2=idat[i1,k2];
		    cj3=- idat[i2,k3];
		    if(dat[i1,k3]<dat[i2,k3]) cj3=idat[i1,k3];
		    cj4=- idat[j1,k2];
		    if(dat[j2,k2]<dat[j1,k2]) cj4=idat[j2,k2];
		    cj5=- idat[j1,k3];
		    if(dat[j2,k3]<dat[j1,k3]) cj5=idat[j2,k3];
		    cj6=- idat[j2,k2];
		    if(dat[j1,k2]<dat[j2,k2]) cj6=idat[j1,k2];
		    cj7=- idat[j2,k3];
		    if(dat[j1,k3]<dat[j2,k3]) cj7=idat[j1,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[i1,k1];
		    if(dat[j1,k1]<dat[i1,k1]) cj1=idat[j1,k1];
		    cj2=- idat[i1,k2];
		    if(dat[j1,k2]<dat[i1,k2]) cj2=idat[j1,k2];
		    cj3=- idat[i1,k3];
		    if(dat[j1,k3]<dat[i1,k3]) cj3=idat[j1,k3];
		    cj4=- idat[j2,k2];
		    if(dat[i2,k2]<dat[j2,k2]) cj4=idat[i2,k2];
		    cj5=- idat[j2,k3];
		    if(dat[i2,k3]<dat[j2,k3]) cj5=idat[i2,k3];
		    cj6=- idat[i2,k2];
		    if(dat[j2,k2]<dat[i2,k2]) cj6=idat[j2,k2];
		    cj7=- idat[i2,k3];
		    if(dat[j2,k3]<dat[i2,k3]) cj7=idat[j2,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[i2,k1];
		    if(dat[j1,k1]<dat[i2,k1]) cj1=idat[j1,k1];
		    cj2=- idat[i2,k2];
		    if(dat[j1,k2]<dat[i2,k2]) cj2=idat[j1,k2];
		    cj3=- idat[i2,k3];
		    if(dat[j1,k3]<dat[i2,k3]) cj3=idat[j1,k3];
		    cj4=- idat[j2,k2];
		    if(dat[i1,k2]<dat[j2,k2]) cj4=idat[i1,k2];
		    cj5=- idat[j2,k3];
		    if(dat[i1,k3]<dat[j2,k3]) cj5=idat[i1,k3];
		    cj6=- idat[i1,k2];
		    if(dat[j2,k2]<dat[i1,k2]) cj6=idat[j2,k2];
		    cj7=- idat[i1,k3];
		    if(dat[j2,k3]<dat[i1,k3]) cj7=idat[j2,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[j2,k1];
		    if(dat[j1,k1]<dat[j2,k1]) cj1=idat[j1,k1];
		    cj2=- idat[j2,k2];
		    if(dat[j1,k2]<dat[j2,k2]) cj2=idat[j1,k2];
		    cj3=- idat[j2,k3];
		    if(dat[j1,k3]<dat[j2,k3]) cj3=idat[j1,k3];
		    cj4=- idat[i2,k2];
		    if(dat[i1,k2]<dat[i2,k2]) cj4=idat[i1,k2];
		    cj5=- idat[i2,k3];
		    if(dat[i1,k3]<dat[i2,k3]) cj5=idat[i1,k3];
		    cj6=- idat[i1,k2];
		    if(dat[i2,k2]<dat[i1,k2]) cj6=idat[i2,k2];
		    cj7=- idat[i1,k3];
		    if(dat[i2,k3]<dat[i1,k3]) cj7=idat[i2,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[i1,k1];
		    if(dat[i2,k1]<dat[i1,k1]) cj1=idat[i2,k1];
		    cj2=- idat[i1,k2];
		    if(dat[i2,k2]<dat[i1,k2]) cj2=idat[i2,k2];
		    cj3=- idat[i1,k3];
		    if(dat[i2,k3]<dat[i1,k3]) cj3=idat[i2,k3];
		    cj4=- idat[j2,k2];
		    if(dat[j1,k2]<dat[j2,k2]) cj4=idat[j1,k2];
		    cj5=- idat[j2,k3];
		    if(dat[j1,k3]<dat[j2,k3]) cj5=idat[j1,k3];
		    cj6=- idat[j1,k2];
		    if(dat[j2,k2]<dat[j1,k2]) cj6=idat[j2,k2];
		    cj7=- idat[j1,k3];
		    if(dat[j2,k3]<dat[j1,k3]) cj7=idat[j2,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[j1,k1];
		    if(dat[i2,k1]<dat[j1,k1]) cj1=idat[i2,k1];
		    cj2=- idat[j1,k2];
		    if(dat[i2,k2]<dat[j1,k2]) cj2=idat[i2,k2];
		    cj3=- idat[j1,k3];
		    if(dat[i2,k3]<dat[j1,k3]) cj3=idat[i2,k3];
		    cj4=- idat[j2,k2];
		    if(dat[i1,k2]<dat[j2,k2]) cj4=idat[i1,k2];
		    cj5=- idat[j2,k3];
		    if(dat[i1,k3]<dat[j2,k3]) cj5=idat[i1,k3];
		    cj6=- idat[i1,k2];
		    if(dat[j2,k2]<dat[i1,k2]) cj6=idat[j2,k2];
		    cj7=- idat[i1,k3];
		    if(dat[j2,k3]<dat[i1,k3]) cj7=idat[j2,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[j2,k1];
		    if(dat[i2,k1]<dat[j2,k1]) cj1=idat[i2,k1];
		    cj2=- idat[j2,k2];
		    if(dat[i2,k2]<dat[j2,k2]) cj2=idat[i2,k2];
		    cj3=- idat[j2,k3];
		    if(dat[i2,k3]<dat[j2,k3]) cj3=idat[i2,k3];
		    cj4=- idat[j1,k2];
		    if(dat[i1,k2]<dat[j1,k2]) cj4=idat[i1,k2];
		    cj5=- idat[j1,k3];
		    if(dat[i1,k3]<dat[j1,k3]) cj5=idat[i1,k3];
		    cj6=- idat[i1,k2];
		    if(dat[j1,k2]<dat[i1,k2]) cj6=idat[j1,k2];
		    cj7=- idat[i1,k3];
		    if(dat[j1,k3]<dat[i1,k3]) cj7=idat[j1,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[i1,k1];
		    if(dat[j2,k1]<dat[i1,k1]) cj1=idat[j2,k1];
		    cj2=- idat[i1,k2];
		    if(dat[j2,k2]<dat[i1,k2]) cj2=idat[j2,k2];
		    cj3=- idat[i1,k3];
		    if(dat[j2,k3]<dat[i1,k3]) cj3=idat[j2,k3];
		    cj4=- idat[i2,k2];
		    if(dat[j1,k2]<dat[i2,k2]) cj4=idat[j1,k2];
		    cj5=- idat[i2,k3];
		    if(dat[j1,k3]<dat[i2,k3]) cj5=idat[j1,k3];
		    cj6=- idat[j1,k2];
		    if(dat[i2,k2]<dat[j1,k2]) cj6=idat[i2,k2];
		    cj7=- idat[j1,k3];
		    if(dat[i2,k3]<dat[j1,k3]) cj7=idat[i2,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[j1,k1];
		    if(dat[j2,k1]<dat[j1,k1]) cj1=idat[j2,k1];
		    cj2=- idat[j1,k2];
		    if(dat[j2,k2]<dat[j1,k2]) cj2=idat[j2,k2];
		    cj3=- idat[j1,k3];
		    if(dat[j2,k3]<dat[j1,k3]) cj3=idat[j2,k3];
		    cj4=- idat[i1,k2];
		    if(dat[i2,k2]<dat[i1,k2]) cj4=idat[i2,k2];
		    cj5=- idat[i1,k3];
		    if(dat[i2,k3]<dat[i1,k3]) cj5=idat[i2,k3];
		    cj6=- idat[i2,k2];
		    if(dat[i1,k2]<dat[i2,k2]) cj6=idat[i1,k2];
		    cj7=- idat[i2,k3];
		    if(dat[i1,k3]<dat[i2,k3]) cj7=idat[i1,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    cj1=- idat[i2,k1];
		    if(dat[j2,k1]<dat[i2,k1]) cj1=idat[j2,k1];
		    cj2=- idat[i2,k2];
		    if(dat[j2,k2]<dat[i2,k2]) cj2=idat[j2,k2];
		    cj3=- idat[i2,k3];
		    if(dat[j2,k3]<dat[i2,k3]) cj3=idat[j2,k3];
		    cj4=- idat[j1,k2];
		    if(dat[i1,k2]<dat[j1,k2]) cj4=idat[i1,k2];
		    cj5=- idat[j1,k3];
		    if(dat[i1,k3]<dat[j1,k3]) cj5=idat[i1,k3];
		    cj6=- idat[i1,k2];
		    if(dat[j1,k2]<dat[i1,k2]) cj6=idat[j1,k2];
		    cj7=- idat[i1,k3];
		    if(dat[j1,k3]<dat[i1,k3]) cj7=idat[j1,k3];
		    gtsum=gtsum+cj1*(2.0*cj2 - cj3*(cj4*cj5+cj6*cj7) );
		    
		    aasum[i1]=aasum[i1]+1.0/24.0*gtsum; %
		}
		
		
	    }
	}
	ave = ave + c2*aasum[i1];
    }
    ave=ave/ntot;

    _for i (0,ntot-1,1)
    {
	asum=asum+(c2*aasum[i]-ave)^2;
	
    }
    return asum*c1;
}


private define sigma_akritas(ntot,idat,dat,k1,k2,k3)
{
    variable sig2 = an_akritas(ntot,idat,dat,k1,k2,k3)/
    (ntot*(1.0-tau_kendall(k1,k3,ntot,idat,dat)^2)*
    (1.0-tau_kendall(k2,k3,ntot,idat,dat)^2));
    return sqrt(sig2);
}




define partial_correlation (data2)
%!%+
%\function{partial_correlation}
%\synopsis{Tests two luminosities for partial correlation due to redshift}
%\usage{partial_correlations(String_Type filename);}
%\description
%       This function tests if a correlation of two parameters is due to
%       the redshift. Description of method and code adapted from
%       Akritas & Siebert, 1996, MNRAS, 278, 919.
%       Works with FITS and ASCII files.
%
%       The input file needs 6 columns with luminosity1, UL, 
%       luminosity2, UL, redshift, UL. The upper limit columns should 
%       have a 1 for a detection, and a 0 for an upper limit.
%
%\examples
%    partial_correlation("data.txt")
%  
%    partial_correlation("data.fits")
%
%\seealso{}
%!%-
{
    %Get local time
    variable mn_a = localtime(_time).tm_min;
    variable hr_a = localtime(_time).tm_hour;
    
    
    %######################################
    % READ FILE

    variable abc = glob(data2);

    if (length(abc) != 1)
    {
	vmessage("File not found!");
	return 0;
    }

    variable testfits = strchop (abc[0],'.',0);
    if (testfits[1] == "FITS" || testfits[1] == "fits")
    {
	variable data = fits_read_table (data2,
	[{"%F","lum1"},{"%F","lum1_UL"},{"%F","lum2"},{"%F","lum2_UL"},
	{"%F","z"},{"%F","z_UL"}]);
    }
    else
    {
	data = ascii_read_table (data2,
	[{"%F","lum1"},{"%F","lum1_UL"},{"%F","lum2"},{"%F","lum2_UL"},
	{"%F","z"},{"%F","z_UL"}]);
    }


    %######################################
    % SETUP VARIABLES

    variable ntot = length(data.lum1);
    variable dat = Double_Type[ntot,3];
    variable idat = Double_Type[ntot,3];

    variable lumtemp = Struct_Type[3];
    lumtemp[0] = struct {lum = data.lum1};
    lumtemp[1] = struct {lum = data.lum2};
    lumtemp[2] = struct {lum = data.z};
    variable lumultemp = Struct_Type[3];
    lumultemp[0] = struct {lumul = data.lum1_UL};
    lumultemp[1] = struct {lumul = data.lum2_UL};
    lumultemp[2] = struct {lumul = data.z_UL};

    variable p,i;

    _for p (0, ntot-1,1)
    {
	dat[p,0]=-log(data.lum1[p])/log(10);	% CHANGE TO RIGHT CENSORING  
	dat[p,1]=-log(data.lum2[p])/log(10);	% CHANGE TO RIGHT CENSORING
	dat[p,2]=-data.z[p];			% CHANGE TO RIGHT CENSORING
	
	idat[p,0]=data.lum1_UL[p];
	idat[p,1]=data.lum2_UL[p];
	idat[p,2]=data.z_UL[p];
    }

    %######################################
    % COMPUTE PARTIAL CORRELATION COEFFICIENT

    variable k1 = 0;
    variable k2 = 1;
    variable k3 = 2;

    vmessage("\n---COMPUTE PARTIAL CORRELATION COEFFICIENT---");
    vmessage("Tau(1,2): %.2f", tau_kendall(k1,k2,ntot,idat,dat));
    vmessage("Tau(1,3): %.2f", tau_kendall(k1,k3,ntot,idat,dat));
    vmessage("Tau(2,3): %.2f", tau_kendall(k2,k3,ntot,idat,dat));

    variable res = tau_kendall_partial(ntot,k1,k2,k3,idat,dat);
    vmessage("Partial Kendalls tau: %.2f", res);
    vmessage("\n***************************");
    vmessage("Calculating variance...this takes some time...");
    variable s = sigma_akritas(ntot,idat,dat,k1,k2,k3);
    vmessage("Square root of variance (sigma): %.2f", s);

    if(abs(res/s) > 1.96) 
    {
	vmessage("Zero partial correlation rejected at level 0.05\n");
    }
    else 
    {
	vmessage("Null hypothesis cannot be rejected!");
	vmessage("--> No correlation present, if influence of third variable is excluded\n");
    }



    %######################################
    %End time
    variable mn_e = localtime(_time).tm_min;
    variable hr_e = localtime(_time).tm_hour;
    

    if (hr_e-hr_a == 0)
    {
	if (mn_e-mn_a == 0)
	{
	    vmessage("Running: < 1 min");
	}
	else
	{
	    vmessage("Running: "+string(mn_e-mn_a));
	}
    }
    else
    {
	vmessage("Running: "+string(hr_e-hr_a)+":"+string(mn_e-mn_a));
    }
}

%
% Code from src/slang/path_realpath.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define path_realpath (path)
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{path_realpath}
%\synopsis{Get full path of specified path}
%\usage{String_Type realpath = path_realpath(String_Type path);}
%\description
%  Turn the given path into an asbolute path (replacing '..' and '.').
%\seealso{path_concat,path_is_absolut}
%!%-
{
  variable cwd = getcwd();
  variable realpath;

  if (path_is_absolute(path))
    realpath = path;
  else
    realpath = path_concat(cwd, path);

  % we utilize that path_dirname handles relative instructions at top level
  variable file = path_basename(realpath); % if it is a file we need this
  variable d = path_dirname(realpath);
  variable b = path_basename(d);
  variable trace = {file};

  while ("" != b) {
    list_append(trace, b);
    d = path_dirname(d);
    b = path_basename(d);
  }

  list_reverse(trace);
  variable t;
  realpath = d; % this is root
  foreach t (trace)
    realpath = path_concat(realpath, t);

  return realpath;
}

%
% Code from src/plot/pbd.sl
%
alias("plot_bin_density",  "pbd");

%
% Code from src/plot/pbi.sl
%
alias("plot_bin_integral", "pbi");

%
% Code from src/plot/pdc.sl
%
define pdc()
{
  variable i, ids;
  switch(_NARGS)
  { case 0: ids = all_data; }
  { case 1: ids = (); ids = [ids]; }

  plot_bin_integral;
  plot_data_counts(ids[0]);
  _for i (1, length(ids)-1, 1) { oplot_data_counts(ids[i]); }
}

%
% Code from src/plot/pdf.sl
%
define pdf()
{
  variable ids;
  switch(_NARGS)
  { case 0: ids = all_data; }
  { case 1: ids = (); ids = [ids]; }
  { message("usage: pdf([ids]);"); return; }

  plot_bin_density;
  plot_data_flux(ids[0]);
  variable i;
  _for i (1, length(ids)-1, 1) { oplot_data_flux(ids[i]); }
}

%
% Code from src/slang/astronomy/period_search_via_string_length.sl
%
define period_search_via_string_length() %{{{
%!%+
%\function{period_search_via_string_length}
%\synopsis{A period-finding method for sparse randomly spaced observations}
%\usage{(Double_Type p[], SL[]) = period_search_via_string_length(Double_Type t[], s[], pmin, pmax; qualifiers)}
%\description
%    This function may be useful to determine the period of a variable
%    signal 's' observed at a relatively small number 'N' of randomly
%    spaced observations at times 't' possibly taken many periods apart.
%
%    To identify candidate periods in the given range between 'pmin' and
%    'pmax', string-lengths 'SL' for trial periods 'p' (sampled in equal
%    frequency steps, see qualifier "delta_phi_max") are computed and
%    tabulated. The trial period giving the lowest string-length results
%    in the most smooth-looking curve and is thus the most likely period.
%\notes
%    The underlying idea is to minimize the so-called "string-length" in
%    a phase diagram. The phase diagram for a trial period 'p' is created
%    by computing the phase 'phi' according to
%      phi = (t-min(t))/p mod 1.
%    The string-length 'SL' is then simply the sum of the lengths of line
%    segments joining successive points (phi_i, s_i) in a phase diagram:
%      SL = sum ( ((s_i-s_i-1)^2+(phi_i-phi_i-1)^2)^(1/2) )
%         + ((s_1-s_N)^2+(phi_1-phi_N+1)^2)^(1/2).
%    To get rid of different scales caused by different units in ordinate
%    (e.g., km/s or magnitude) and abscissa (unit = trial period), the
%    observations are scaled so that
%      s' = 0.5*(s-min(s))/(max(s)-min(s))-0.25
%    in order to give equal weight to measures and phases (for details,
%    see Dworetsky 1983, MNRAS, 203, 917).
%
%    For reference: the string-length of a sinusoidal string is 1.4637.
%\qualifiers{
%\qualifier{delta_phi_max [=0.01]}{: Trial periods 'p' are sampled in equal
%      frequency steps of
%        delta_freq = delta_phi_max/(max(t)-min(t)).
%      This ensures that phase errors are always less than 'delta_phi_max'
%      since
%        phi = (t-min(t))/p mod 1 = (t-min(t))*freq mod 1
%        -> delta_phi = (t-min(t))*delta_freq
%        -> delta_phi_max = (max(t)-min(t))*delta_freq.}
%}
%\example
%    t = 1000*urand(50); % time of observations
%    s = sin(t)+0.05*grand(length(t)); % signal of observations; true period is 2*PI
%    connect_points(0); point_style(2); plot(t, s);
%    (p, SL) = period_search_via_string_length(t, s, 1, 10; delta_phi_max=0.01);
%    connect_points(-1); point_style(-1); plot(p, SL);
%    ind = where(SL==min(SL)); SL[ind[0]]; p[ind[0]]; % index, value, and period of minimum string-length
%    connect_points(0); point_style(2); plot((t-min(t))/p[ind[0]] mod 1, s); % phase diagram
%\seealso{}
%!%-
{
  if(_NARGS!=4)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable t, s, pmin, pmax;
  (t, s, pmin, pmax) = ();
  if(length(t)!=length(s))
    throw UsageError, sprintf("Usage error in '%s': The first two arguments have to be arrays of the same length.", _function_name());
  if(pmin>=pmax)
    throw UsageError, sprintf("Usage error in '%s': The minimum period has to be lower than the maximum period.", _function_name());
  % ensure that times start at 0 and are positive:
  t -= min(t);
  % scale signal according to Eq. (3) in Dworetsky 1983, MNRAS, 203, 917:
  variable mis = min(s), mas = max(s);
  s = 0.5*(s-mis)/(mas-mis)-0.25;
  % create array with trial periods:
  variable delta_freq = qualifier("delta_phi_max", 0.01)/max(t); % min(t)=0, see above
  variable len = (1./pmin-1./pmax)/delta_freq;
  if(len>100000000) % to avoid runtime problems because of huge random access memory usage; Double_Type[100000000] consumes roughly 1GB
    throw LimitExceededError, sprintf("Usage error in '%s': Random access memory usage too large. Choose a smaller interval for the trial periods.", _function_name());
  len = nint(len);
  variable periods = 1./[1./pmax:1./pmin:#len];
  variable string_length = Double_Type[len];
  % phase data according to trial period and compute string length:
  variable i;
  _for i(0, len-1, 1)
  {
    % phase data according to trial period:
    variable phase = t/periods[i] mod 1;
    % sort in ascending order with phase:
    variable ind = array_sort(phase); phase = phase[ind];
    variable phased_s = s[ind];
    % compute string-length:
    string_length[i] = sum(sqrt((phased_s[[1:]]-phased_s[[:-2]])^2+(phase[[1:]]-phase[[:-2]])^2))+sqrt((phased_s[0]-phased_s[-1])^2+(phase[0]-phase[-1]+1)^2);
  }
  return (periods, string_length);
}%}}}

%
% Code from src/data/timing/pfold.sl
%
define phaseOfTime (time, p, dp, ddp) {
%!%+
%\function{phaseOfTime}
%\synopsis{Compute the phase corresponding to given time}
%\usage{Double_Type[] phaseOfTime(Double_Type[] time, Double_Type, p, dp, ddp);}
%\description
%  Given a time and a period evolution by the period P \code{p},
%  first, and second derivative of P, \code{dp} and \code{ddp},
%  respectively, calculate the phase.
%
%  Note: This function does a simple taylor expansion of the phase
%  dphi = 1./P(t)dt. Time zero correpsonds to phase zero. For
%  long time arrays this function suffers from precision errors.
%
%  Since it is expected that the period changes monotonically, this
%  function throws an error if the given time is outside of the extrema
%  (if any). This only applies to non-zero \code{ddp}.
%\seealso{timeOfPhase}
%!%-
  variable a1 = 1./double(p);

  if (abs(dp)<1e-16 && abs(ddp)<1e-16)
    return time*a1;

  variable a2 = -0.5*a1*a1*dp;

  if (abs(ddp)<1e-16)
    return (time*a2+a1)*time;

  variable a3 = dp*dp*a1*a1*a1/3.-ddp*a1*a1/6.;

  if ((a2*a2-3*a3*a1>0) &&
      not all((-a2-sqrt(a2*a2-3*a3*a1))/3./a3 < time < (-a2+sqrt(a2*a2-3*a3*a1))/3./a3))
    throw DomainError, "Period evolution is invalid for the given time frame";

  return ((time*a3+a2)*time+a1)*time;
}

define timeOfPhase (phase, p, dp, ddp) {
%!%+
%\function{timeOfPhase}
%\synopsis{Compute the time corresponding to given phase}
%\usage{Double_Type[] timeOfPhase(Double_Type[] phase, Double_Type, p, dp, ddp);}
%\description
%  Given a phase and a period evolution by the period P \code{p},
%  first, and second derivative of P, \code{dp} and \code{ddp},
%  respectively, calculate the time.
%
%  Note: This function is essentially the inverse of \code{timeOfPhase}.
%  For non-zero \code{ddp} the inverse is only approximated through
%  series inversion!
%\seealso{phaseOfTime}
%!%-
  variable b1 = p;

  if (abs(dp)<1e-16 && abs(ddp)<1e-16)
    return phase*b1;

  if (abs(ddp)<1e-16)
    return p*(1.-sqrt(1.-2*dp*phase))/dp;

  % through series inversion
  variable b2 = 0.5*b1*dp;
  variable b3 = (p*dp*dp+p*p*ddp)/6.;

  if ((b2*b2-3*b3*b1>0) &&
      not all((-b2-sqrt(b2*b2-3*b3*b1))/3./b3 < time < (-b2+sqrt(b2*b2-3*b3*b1))/3./b3))
    throw DomainError, "Period evolution is invalid for the given period frame";

  return ((b3*phase+b2)*phase+b1)*phase;
}

private define periodFoldEvents (time, p) {
% relatively easy task:
%   1) sort in the events to the phase histogram `value'
%   2) calculate exposure time for each phase bin `exposure'
%     => value/exposure is the pulse profile
  variable verbose = qualifier("verbose", 1);

  % no gti given, assume all events are in a gti (edge are taken sharply)
  if (not qualifier_exists("gti") && verbose>0)
    vmessage("*** Warning: No GTIs given, assuming full time between first and last event\n  This is likely not what you want");
  variable gti = qualifier("gti", struct {start=min(time),stop=max(time)});
  variable dp = qualifier("pdot", 0.); % P dot (for Taylor approximation)
  variable ddp = qualifier("pddot", 0.); % P dot dot (also for Taylor)
  variable nbins = qualifier("nbins", 30); % number of phase bins
  variable t0 = qualifier("t0", min(gti.start)); % zero point of period evolution

  variable grid = [0:1:#nbins+1]; % overflow bin does not matter here
  variable value = UInt_Type[nbins+1];
  variable exposure = Double_Type[nbins];

  % working variables
  variable k, j, startj, n, pStart, pStop, nStart,
    nStop, inStop, inStart, tLeft, tRight,
    phase, gtiStart, gtiStop, eTime, bTime,
    gtiIdx;

  () = histogram(time, gti.start, gti.stop, &gtiIdx);

  _for k (0, length(gtiIdx)-1) {
    eTime = time[gtiIdx[k]]-t0;
    gtiStart = gti.start[k]-t0;
    gtiStop  = gti.stop[k]-t0;
    pStart = phaseOfTime(gtiStart, p, dp, ddp);
    pStop = phaseOfTime(gtiStop, p, dp, ddp);

    % get index in pulse profile of start and stop
    nStart = int(floor(pStart));
    nStop = int(floor(pStop));

    inStart = wherefirst(((pStart-nStart)-grid)<0.)-1;
    inStop = wherefirst(((pStop-nStop)-grid)<0.)-1;

    % need to go through each covered profile bin
    % and calculate the relative exposure. This is necessary
    % because if dp and ddp are non-zero, the coverage is
    % not linear.
    j = inStart;
    n = nStart;

    if ((inStart == inStop) && (nStart==nStop)) {
      % gti window is completely inside bin
      exposure[inStart] += gtiStop-gtiStart;
    } else if ((nStop - nStart) > 5) {
      % if many periods are covered in one interval it is better to loop over the phase bins only once
      if (qualifier_exists("exact"))
      {
	n = [nStart:nStop]; %nStop-nStart+1;
	_for j (0, nbins-1)
	{
	  bTime = timeOfPhase([grid[j]+n, grid[j+1]+n], p, dp, ddp);
	  exposure[j] += sum(bTime[[length(n)+1:2*length(n)-2]]-bTime[[1:length(n)-2]]);

	  if (j==inStart) % endpoints need special treatment
	    exposure[j] += (bTime[length(n)]-gtiStart);
	  if (j==inStop)
	    exposure += (gtiStop-bTime[length(n)-1]);
	}
      } else { % assume constant p over GTI such that there is a linear relation between dPhi and dT
	bTime = timeOfPhase([nStart+grid,nStop+grid], p, dp, ddp);
	exposure += (bTime[nbins+1]-bTime[nbins])/nbins;
	exposure[[inStart:nbins-1]] += (bTime[nbins]-gtiStart)/(nbins-inStart);
	exposure[[0:inStop]] += (gtiStop-bTime[nbins+1])/(inStop+1);
      }
    } else {
      bTime = timeOfPhase(n+grid, p, dp, ddp);
      startj = j;
      do {
	if (j==nbins) { % next period?
	  % phase grid change, so write it first
	  exposure[[startj:nbins-1]] += _min(gtiStop,bTime[[startj+1:nbins]])-_max(gtiStart,bTime[[startj:nbins-1]]);
	  n++;
	  bTime = timeOfPhase(n+grid, p, dp, ddp);
	  j = 0; startj = j;
	}
	j++;
      } while ((j <= inStop) || (n < nStop));
      % last partial period not covered yet, j==0 if last bin had been covered
      if (j!=0)
	exposure[[startj:j-1]] += _min(gtiStop,bTime[[startj+1:j]])-_max(gtiStart,bTime[[startj:j-1]]);
    }

    phase = phaseOfTime(eTime, p, dp, ddp);

    if (length(phase))
      value += histogram(phase-floor(phase), grid);
  }

  return struct {
    bin_lo = grid[[:-2]],
    bin_hi = grid[[1:]],
    value = 1.*value[[:-2]]/exposure, % count rate
    error = sqrt(value[[:-2]])/exposure, % count rate error
    counts = value[[:-2]],
    npts=value[[:-2]], % not sure what this is used for
    ttot = exposure % per bin exposure
  };
}

private define periodFoldLC (time, rate, rateerr, p) {
  variable verbose = qualifier("verbose", 1);

  variable dp  = qualifier("dp", 0.);
  variable ddp = qualifier("ddp", 0.);
  variable nbins = qualifier("nbins", 30);

  if (not qualifier_exists("dt") && verbose>0)
    vmessage("*** Warning: Grid spacing not given, assuming dt=diff(time)");
  variable dt = qualifier("dt"); % if not given, dt is t_k+1-t_k, can be single value
  variable fracexp = qualifier("fracexp", 1.); % if not given, full time coverage of dt is assumed
  variable t0 = qualifier("t0", min(time)); % zero point of period evolution
  variable fast = not qualifier_exists("exact");

  if (NULL == dt)
    dt = mean(time[[1:]]-time[[:-2]]);

  variable timeend, phaseLo, phaseHi;
  variable grid = [0:1:#nbins+1];
  variable value = Double_Type[nbins];
  variable exposure = Double_Type[nbins];
  variable err = Double_Type[nbins];
  variable npts;

  % working variables
  variable tStart, tStop, pStart, pStop, nStart,
    nStop, inStart, inStop, n, j, startj, k, cFracExp, cDt,
    bTime, tLeft, tRight, tVal, tTime, fastP, fastIdx;

  phaseLo = phaseOfTime(time-t0, p, dp, ddp);

  % we can speed up the calculation by alot when we ignore
  % that lc bins are finite
  if (fast) {
    fastP = phaseLo-floor(phaseLo);
    npts = histogram(fastP, grid, ,&fastIdx)[[:-2]];
    _for j (0, length(fastIdx)-2) {
      cDt = (length(dt) == 1) ? dt : dt[fastIdx[j]];
      cFracExp = (length(fracexp) == 1) ? fracexp*cDt : fracexp[fastIdx[j]]*cDt;
      value[j] = sum((rate[fastIdx[j]])*cFracExp);
      exposure[j] = sum(cFracExp);
      err[j] = (NULL != rateerr) ? sum(sqr((rateerr[fastIdx[j]])*cFracExp)) : value[j];
    }
  } else {
    timeend = time-t0+dt;
    phaseHi = phaseOfTime(timeend, p, dp, ddp);
    npts = Int_Type[nbins];

    _for k (0, length(phaseLo)-1) {
      tStart = time[k]-t0;
      % nummerical issue here, large timestamps + dt creates deviations
      tStop = tStart + ((length(dt)==1) ? dt : dt[k]);
      pStart = phaseLo[k];
      pStop = phaseHi[k];
      nStart = int(floor(pStart));
      nStop = int(floor(pStop));

      % phase bins lc bins fall into
      inStart = wherefirst(((pStart-nStart)-grid)<0)-1;
      inStop = wherefirst(((pStop-nStop)-grid)<0)-1;

      if (NULL == inStart || NULL == inStop)
	continue;

      npts[inStart]++;

      n = nStart;
      j = inStart;

      cFracExp = (length(fracexp)==1) ? fracexp : fracexp[k];

      if ((inStart == inStop) && (nStart == nStop)) {
	tTime = (tStop-tStart)*cFracExp;
	tVal = rate[k]*tTime;
	value[j] += tVal;
	exposure[j] += tTime;
	err[j] += (NULL != rateerr) ? sqr(rateerr[k]*tTime) : tVal;
      } else if ((nStop - nStart) > 5) {
	% if many periods are covered in one interval it is better to loop over the phase bins only once
	% and for the LC it is safe to assume the period does not change significantly over one bin
	bTime = timeOfPhase([nStart+grid,nStop+grid], p, dp, ddp);
	exposure += (bTime[nbins+1]-bTime[nbins])/nbins;
	exposure[[inStart:nbins-1]] += (bTime[nbins]-tStart)/(nbins-inStart);
	exposure[[0:inStop]] += (tStop-bTime[nbins+1])/(inStop+1);
      } else {
	% otherwise we loop through the phases until we hit the interval limit
	bTime = timeOfPhase(n+grid, p, dp, ddp);
	startj = j;
	do {
	  if (j == nbins) {
	    tTime = (_min(tStop,bTime[[startj+1:nbins]])-_max(tStart,bTime[[startj:nbins-1]]))*cFracExp;
	    tVal = rate[k]*tTime;
	    exposure[[startj:nbins-1]] += tTime;
	    value[[startj:nbins-1]] += tVal;
	    err[[startj:nbins-1]] += (NULL != rateerr) ? rateerr[k]*tTime : tVal;
	    n++;
	    bTime = timeOfPhase(n+grid, p, dp, ddp);
	    j=0; startj = j;
	  }
	  j++;
	} while ((j <= inStop) || (n < nStop));
	% last coverage might not be handled
	if (j!=0) {
	  tTime = (_min(tStop,bTime[[startj+1:j]])-_max(tStart,bTime[[startj:j-1]]))*cFracExp;
	  tVal = rate[k]*tTime;
	  exposure[[startj:j-1]] += tTime;
	  value[[startj:j-1]] += tVal;
	  err[[startj:j-1]] += (NULL != rateerr) ? sqr(rateerr[k]*tTime) : tVal;
	}
      }
    }
  }

  return struct {
    bin_lo = grid[[:-2]],
    bin_hi = grid[[1:]],
    value = 1.*value/exposure, % count rate
    error = sqrt(err)/exposure, % count rate error
    counts = value, % counts
    npts=npts, % how often a bin was hit by start time (not sure what this is used for)
    ttot = exposure, % per bin exposure
  };
}

%%%%%%%%%%%%%%%%%%%%%%%%
define pfold ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pfold}
%\synopsis{folds a lightcurve or event list on a given period}
%\usage{Struct_Type pp = pfold(Double_Type t, r, p);
%\altusage{Struct_Type pp = pfoldpfold(Double_Type t, r, p, e);}
%\altusage{Struct_Type pp = pfold(Double_Type t, p);}}
%\qualifiers{
%\qualifier{nbins}{[=30] number of bins for the pulse profile}
%\qualifier{exact}{take finite lightcurve bins into account (LC case).
%             take differantial phase grid into account (Events case).}
%\qualifier{dt}{lightcurve bin size.}
%\qualifier{fracexp}{fractional exposure per lightcurve bin.}
%\qualifier{t0}{set reference time.}
%\qualifier{pdot}{[=0] first derivative of pulse period p.}
%\qualifier{pddot}{[=0] second derivative of pulse period p.}
%\qualifier{gti}{GTIs for event data, given as struct{start=Double_Type, stop=Double_Type}}
%}
%\description
%   Calculates the pulse profile of a given lightcurve or event list
%   for times \code{t} and rate \code{r}. In case of an event list
%   normally the use of GTIs is necessary to ensure correct results.
%
%   If the qualifiers \code{pdot} or \code{pddot} are given and not
%   zero a taylor expansion of the period evolution is used to
%   calculate the mapping to phase (from \code{t0}). If \code{pddot}
%   is non-zero the result is only approximately correct. See
%   \code{phaseOfTime}.
%
%   The input arrays \code{t}, \code{r}, and \code{e} can also be
%   given as qualifiers.
%
%   If the \code{exact} qualifier is given the finite size of lightcurve
%   bins is taken into account by linear interpolation. However, the
%   computation time is increased. For event lists, the code will take the
%   differentail phase grid change into account. Otherwise the exposure
%   time is approximated assuming a constant period over one GTI frame.
%
%   The returned structure contains the fields bin_lo, bin_hi, value
%   and err, ready for plotting the pulse profile (as rate). For further
%   diagnostics the \code{counts} (counts) and \code{ttot} (per-bin exposure)
%   are given. The field \code{npts} contains the number of lightcurve bins
%   starting in each bin.
%\seealso{phaseOfTime, timeOfPhase}
%!%-
{
  variable chatty = qualifier("chatty",qualifier("verbose", _Isisscripts_Verbose)) ;
  variable t,r = NULL,e = NULL,p;

  switch(_NARGS)
  {case 2:
    (t,p) = ();
    (r,e) = (qualifier("r", r), qualifier("e", e));
  }
  {case 3:
    (t,r,p) = ();
    e = qualifier("e", e);
  }
  {case 4:
    (t,r,p,e) = ();
  }
  {() = (); help(_function_name()); return; }

  if (NULL == r) {
    % event data
    return periodFoldEvents(t, p;; __qualifiers);
  } else {
    % lightcurve
    if (length(t) != length(r) || (e != NULL && length(t)!=length(e)))
      throw UsageError, "Time and rate (or error) arrays are not of equal length";
    return periodFoldLC(t, r, e, p;; __qualifiers);
  }
}

%
% Code from src/data/GTI/phasebin_GTI.sl
%
%%%%%%%%%%%%%%%%%%%%
define phasebin_GTIs()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{phasebin_GTIs}
%\synopsis{computes the GTI time in phase bins}
%\usage{Double_Type dt[] = phasebin_GTIs(tstart, tstop, T0, P, n);}
%!%-
{
  variable tstart, tstop, T0, P, n;
  switch(_NARGS)
%  { case 2: (gti, n) = (); }
%  { case 3: (tstart, tstop, n) = (); }
%  { case 4: (gti, n, T0, P) = (); }
  { case 5: (tstart, tstop, T0, P, n) = (); }
  { help(_function_name()); return; }

  variable dt = Double_Type[n];
  variable dt_bin = P/n;
  variable i_start = int(floor( (tstart-T0)/P * n ));
  variable i_stop  = int(floor( (tstop -T0)/P * n ));
  variable i;
  _for i (0, length(tstart)-1, 1)
    if(i_start[i] == i_stop[i])  % start and stop fall in the same bin
      dt[i_start[i] mod n] += tstop[i] - tstart[i];
    else
    {
      dt[i_start[i] mod n] += T0 + (i_start[i]+1)*dt_bin - tstart[i];  % first bin, partially exposed
      dt[i_stop[i]  mod n] += tstop[i] - (T0 + i_stop[i]*dt_bin);  % last bin, partially exposed
      foreach i ([i_start[i]+1 : i_stop[i]-1] mod n)  % intermediate bins,
        dt[i] += dt_bin;                              % fully exposed
    }
  return dt;
}

%
% Code from src/fitting/fit-functions/phflux.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define phflux_fit(lo, hi, par, y)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{phflux (fit-function)}
%\synopsis{fits the photon flux in a given energy range}
%\description
%    This function can be used as a convolution model to determine
%    the photon flux [ph/s/cm^2] of the model in the energy range
%    given by \code{E_min} and \code{E_max}. Only bins within(!) this energy range
%    are considered for the calculation of the flux, thus the model
%    should be evaluated on a grid including the values defining the
%    energy range, i.e. \code{E_min} and \code{E_max} are elements
%    of \code{[bin_lo, bin_hi]} of the grid. A user grid can be used
%    for this purpose.
%    As this function fits the normalization of the total convolved
%    model, the normalizations of its components are not defined
%    absolutely, but only relativ to each others. For that reason it
%    is meaningful to freeze the normalization of one component, at
%    best the one of the continuum to avoid ambiguities during
%    fitting.
%
%\examples
%    % data definition:
%    \code{variable lo = _A([1:10]);}
%    \code{variable hi = make_hi_grid(lo);}
%    \code{variable my_data = define_counts(lo,hi,lo,sqrt(lo));}\n
%    \code{variable my_emin = 2.5;}
%    \code{variable my_emax = 6.5;}
%    % defining a grid containing the energy limits for the flux:
%    % (in this way only valid if energy range is covered by the grid
%    % and if these values are not already element of bin_lo, bin_hi)
%    \code{define my_grid(id, s)}
%    \code{\{}
%    \code{   variable mygdc = get_data_counts(id);}
%    \code{   mygdc.bin_lo = [mygdc.bin_lo,_A(my_emax),_A(my_emin)];}
%    \code{   mygdc.bin_hi = [mygdc.bin_hi,_A(my_emax),_A(my_emin)];}
%    \code{   s.bin_lo = mygdc.bin_lo[array_sort(mygdc.bin_lo)];}
%    \code{   s.bin_hi = mygdc.bin_hi[array_sort(mygdc.bin_hi)];}
%    \code{   return s;}
%    \code{\}}
%    
%    \code{set_eval_grid_method (USER_GRID, my_data, &my_grid);}\n
%    \code{fit_fun("phflux(1,powerlaw(1))");}
%
%    \code{set_par("phflux(1).E_min",  my_emin,  1); % keV}
%    \code{set_par("phflux(1).E_max",  my_emax,  1); % keV}
%    \code{freeze("powerlaw(1).norm");}
%    \code{()=fit_counts();}
%    \code{list_par;}
%
%
%    % It is also possible to determine only the flux of certain
%    % model components, e.g., the unabsorbed flux:
%    \code{fit_fun(phabs(1)*phflux(1, powerlaw(1)));}
%
%\seealso{enflux, set_eval_grid_method}
%!%-
{
  variable flux = par[0];  % <-- int_E1^E2 S_E(E) dE  =  int_l1^l2 S_l(l) dl  [in ph/s/cm^2]
  variable l1 = _A(par[2]);
  variable l2 = _A(par[1]);
  if (Fit_Verbose >= 0 && (l1 < lo[0] or l2 > hi[-1]))
    vmessage("warning: enflux energy range (%.2e-%.2e keV) not covered by data grid (%.2e-%.2e keV)",
	     par[1],par[2],_A(hi[-1]),_A(lo[0]));
  variable i = where(l1 <= lo and hi <= l2);
  return y * flux / sum(y[i]);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define phflux_default(i)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{ switch(i)
  { case 0: return struct{value=0.1, freeze=0, min=0, max=1e20, hard_min=0, hard_max=1e100, step=1e-5, relstep=1e-3};}
  { case 1: return (3.0 , 1, 0, 1e20 ); }
  { case 2: return (10. , 1, 0, 1e20 ); }
}

add_slang_function("phflux", ["phflux [ph/s/cm^2]", "E_min [keV]", "E_max [keV]"], [0]);
set_param_default_hook("phflux", "phflux_default");
set_function_category("phflux", ISIS_FUN_OPERATOR);

%
% Code from src/data/photon_flux.sl
%
define photon_flux()
%!%+
%\function{photon_flux}
%
%\synopsis{}
%\usage{String_Type = photon_flux (hist_index, E_min, E_max);}
%\description
%	This function returns the integrated photon flux of
%	a defined Energy range. 
%	If no energy range specified E_min=0.5, E_max=10.
%
%\seealso{get_data_flux;}
%!%-
 { 
variable dset;
variable emin = NULL, emax = NULL;

switch(_NARGS)
  { case 1: dset  = (); }
  { case 3: (dset,emin,emax) = ();}
  { help(_function_name()); return; }

    
    %%%%%%%%%%%%%
    %Check if Emin & Emax have been specified
    %%%%%%%%%%
    
    if (emin == NULL) emin=0.5;
    if (emax == NULL) emax=10;
        
    %Get Data Counts in Angstrom,
    %Change Bins to keV
    
    flux_corr(dset);
    variable gdfa = _A(get_data_flux(dset)); 
  %  variable gdfk =_A(gdfa);
    
    %%%%%%%%%%
    %% Only notice defined energy range
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    struct_filter(gdfa, where(gdfa.bin_lo>emin));
    struct_filter(gdfa, where(gdfa.bin_hi<emax));


      
    variable gdf =sum(gdfa.value);

    return gdf;    
}

%
% Code from src/slang/physics/physical_quantity.sl
%
#ifnexists PhysicalQuantity_Type

%%%%%%%
typedef struct {
%%%%%%%
  numerical_value,
  leng_unit, leng_dim,
  time_unit, time_dim,
  mass_unit, mass_dim,
  curr_unit, curr_dim,
  temp_unit, temp_dim,
} PhysicalQuantity_Type;

#endif

private variable physical_quantity_baseunits = struct {
       length_in_m  = Assoc_Type[Double_Type],
         time_in_s  = Assoc_Type[Double_Type],
         mass_in_kg = Assoc_Type[Double_Type],
      current_in_A  = Assoc_Type[Double_Type],
  temperature_in_K  = Assoc_Type[Double_Type],
         unit       = Assoc_Type[PhysicalQuantity_Type],
};

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        %%%%%%%%%%%%%%%%%%%%%%%
private define parse_unit_power(str)
        %%%%%%%%%%%%%%%%%%%%%%%
{
  variable m = string_matches(str, `\(.*\)^\(.*\)`, 1);
  if(m!=NULL)
    (m[1], atoi(str_delete_chars(m[2], "{[()]}")));  % (unit, power);  % left on stack
  else
    if(str=="")
      ("", 0);   % (unit, power);  % left on stack
    else
      (str, 1);  % (unit, power);  % left on stack
}


private define handle_physical_quantity_units(q, units);

%%%%%%%%%%%%%%%%%%%%%%%%
define physical_quantity()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{physical_quantity}
%\synopsis{initializes a physical quantity with number and units}
%\usage{PhysicalQuantity_Type physical_quantity(Double_Type number)}
%\qualifiers{
%\qualifier{leng}{length unit and dimension}
%\qualifier{time}{time unit and dimension}
%\qualifier{mass}{mass unit and dimension}
%\qualifier{curr}{electrical current unit and dimension}
%\qualifier{temp}{temperature unit and dimension}
%\qualifier{unit}{array of compound units (see \code{physical_quantity_unit})}
%}
%\description
%    The units are specified as strings with their dimensionality
%    indicated by "^" as powers, i.e., "[unit]^[dim]".
%    [dim] can be a negative number as well.
%    For dim==1, "^1" can be omitted.
%\example
%    variable c = physical_quantity(299792.458; length="km", time="s^-1");
%\seealso{physical_quantity_unit}
%!%-
{
  variable number;
  switch(_NARGS)
  { case 1: number = (); }
  { help(_function_name()); return; }

  variable quant = @PhysicalQuantity_Type;
  quant.numerical_value = number;
  (quant.leng_unit, quant.leng_dim) = parse_unit_power(qualifier("leng", ""));
  (quant.time_unit, quant.time_dim) = parse_unit_power(qualifier("time", ""));
  (quant.mass_unit, quant.mass_dim) = parse_unit_power(qualifier("mass", ""));
  (quant.curr_unit, quant.curr_dim) = parse_unit_power(qualifier("curr", ""));
  (quant.temp_unit, quant.temp_dim) = parse_unit_power(qualifier("temp", ""));

  if(qualifier_exists("unit"))
    quant = handle_physical_quantity_units(quant, [qualifier("unit")]);

  return quant;
}


%%%%%%%%%%%%%%%
define SIprefix(prefix)
%%%%%%%%%%%%%%%
%!%+
%\function{SIprefix}
%\usage{Double_Type SIprefix(String_Type prefix)}
%!%-
{
  switch(prefix)
  { case "Y": return 1e24; }
  { case "Z": return 1e21; }
  { case "E": return 1e18; }
  { case "P": return 1e15; }
  { case "T": return 1e12; }
  { case "G": return 1e9; }
  { case "M": return 1e6; }
  { case "k": return 1e3; }
  { case "h": return 1e2; }
% { case "da": return 1e1; }
  { case "" : return 1; }
  { case "d": return 1e-1; }
  { case "c": return 1e-2; }
  { case "m": return 1e-3; }
  { case "µ": return 1e-6; }
  { case "n": return 1e-9; }
  { case "p": return 1e-12; }
  { case "f": return 1e-15; }
  { case "a": return 1e-18; }
  { case "z": return 1e-21; }
  { case "y": return 1e-24; }
  if(qualifier_exists("verbose"))  vmessage("error (%s): unknown prefix '%s'", _function_name(), prefix);
  return 0;
}

        %%%%%%%%%%%%%%%%%%%%%%%
private define unit_to_baseunit(unit, baseunits)
        %%%%%%%%%%%%%%%%%%%%%%%
{
  variable strlen_unit = strlen(unit);
  variable baseunit, basevalue, result=NULL, interpretation={};
  foreach baseunit, basevalue (baseunits) using ("keys", "values")
  { variable strlen_baseunit = strlen(baseunit);
    if(strlen_baseunit<=strlen_unit && unit[[-strlen_baseunit:]] == baseunit)
    {
      variable SIprefix_name = unit[[:-strlen_baseunit-1]];
      variable SIprefix_value = SIprefix(SIprefix_name);
      if(SIprefix_value!=0)
      {
	list_append(interpretation, (SIprefix_name=="" ? "" : SIprefix_name + "-") + baseunit);
        if(__is_numeric(basevalue))
          result = SIprefix_value * basevalue;
        else
        { result = @basevalue;
	  result.numerical_value *= SIprefix_value;
	}
      }
    }
  }
  if(qualifier_exists("get_interpretation"))
    return interpretation;

  if(result==NULL)
    vmessage("error (%s): unknown unit '%s'", _function_name(), unit);
  if(length(interpretation)>1)
    vmessage("warning: ambiguous unit '%s'\n         possible interpretations: %s",
	     unit, strjoin(list_to_array(interpretation), " or "));
  return result;
}


%%%%%%%%%%%%%%%%%%%%
define convert_units()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{convert_units}
%\synopsis{convers units of a physical quantity}
%\usage{convert_units(PhysicalQuantity_Type q);
%\altusage{PhysicalQuantity_Type convert_units(PhysicalQuantity_Type q; copy)}
%}
%\qualifiers{
%\qualifier{leng}{unit of length}
%\qualifier{time}{unit of time}
%\qualifier{mass}{unit of mass}
%\qualifier{curr}{unit of electrical current}
%\qualifier{temp}{unit of temperature}
%\qualifier{SI}{sets  leng="m", time="s", mass="kg", curr="A", temp="K"}
%\qualifier{copy}{does not change unit of q, but returns a copy of q}
%}
%\description
%   The allowed units are specified by the associative arrays
%         baseunits_length_in_m, baseunits_time_in_s,
%         baseunits_mass_in_kg, baseunits_current_in_A,
%     and baseunits_temperature_in_K
%   which specify the factor to the corresponding SI units.
%!%-
{
  variable q;
  switch(_NARGS)
  { case 1: q = (); }
  { help(_function_name()); return; }

  variable leng_unit = qualifier("leng", "");
  variable time_unit = qualifier("time", "");
  variable mass_unit = qualifier("mass", "");
  variable curr_unit = qualifier("curr", "");
  variable temp_unit = qualifier("temp", "");
  if(qualifier_exists("SI"))
    (leng_unit, time_unit, mass_unit, curr_unit, temp_unit)
    = ("m",     "s",       "kg",      "A",       "K");

  variable copy = qualifier_exists("copy");
  if(copy)  q = @q;

  variable factor1, factor2;
  if(leng_unit!="" && q.leng_dim!=0)
  {
    factor1 = unit_to_baseunit(q.leng_unit, physical_quantity_baseunits.length_in_m);
    factor2 = unit_to_baseunit(  leng_unit, physical_quantity_baseunits.length_in_m);
    if(factor1!=NULL && factor1!=0 && factor2!=NULL && factor2!=0)
      q.numerical_value *= (factor1/factor2)^q.leng_dim,
      q.leng_unit = leng_unit;
  }
  if(time_unit!="" && q.time_dim!=0)
  {
    factor1 = unit_to_baseunit(q.time_unit, physical_quantity_baseunits.time_in_s);
    factor2 = unit_to_baseunit(  time_unit, physical_quantity_baseunits.time_in_s);
    if(factor1!=NULL && factor1!=0 && factor2!=NULL && factor2!=0)
      q.numerical_value *= (factor1/factor2)^q.time_dim,
      q.time_unit = time_unit;
  }
  if(mass_unit!="" && q.mass_dim!=0)
  {
    factor1 = unit_to_baseunit(q.mass_unit, physical_quantity_baseunits.mass_in_kg);
    factor2 = unit_to_baseunit(  mass_unit, physical_quantity_baseunits.mass_in_kg);
    if(factor1!=NULL && factor1!=0 && factor2!=NULL && factor2!=0)
      q.numerical_value *= (factor1/factor2)^q.mass_dim,
      q.mass_unit = mass_unit;
  }
  if(curr_unit!="" && q.curr_dim!=0)
  {
    factor1 = unit_to_baseunit(q.curr_unit, physical_quantity_baseunits.current_in_A);
    factor2 = unit_to_baseunit(  curr_unit, physical_quantity_baseunits.current_in_A);
    if(factor1!=NULL && factor1!=0 && factor2!=NULL && factor2!=0)
      q.numerical_value *= (factor1/factor2)^q.curr_dim,
      q.curr_unit = curr_unit;
  }
  if(temp_unit!="" && q.temp_dim!=0)
  {
    factor1 = unit_to_baseunit(q.temp_unit, physical_quantity_baseunits.temperature_in_K);
    factor2 = unit_to_baseunit(  temp_unit, physical_quantity_baseunits.temperature_in_K);
    if(factor1!=NULL && factor1!=0 && factor2!=NULL && factor2!=0)
      q.numerical_value *= (factor1/factor2)^q.temp_dim,
      q.temp_unit = temp_unit;
  }

  if(copy)  return q;
}


        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_string(quant)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable str = string(quant.numerical_value);
  variable unit;
  foreach unit ([{quant.leng_dim, quant.leng_unit},
	         {quant.time_dim, quant.time_unit},
		 {quant.mass_dim, quant.mass_unit},
		 {quant.curr_dim, quant.curr_unit},
		 {quant.temp_dim, quant.temp_unit}])
    if(unit[0]!=0)
    {
      str = sprintf("%s %s", str, unit[1]);
      if(unit[0]!=1)
        str = sprintf("%s^%d", str, unit[0]);
    }
  return str;
}
__add_string(PhysicalQuantity_Type, &physical_quantity_string);

        %%%%%%%%%%%%%%%%%%%%%%
private define unit_qualifiers(q)
        %%%%%%%%%%%%%%%%%%%%%%
{
  variable pow = qualifier("power", 1);

  if(qualifier_exists("nodim"))
    return struct {
      leng = q.leng_unit,
      time = q.time_unit,
      mass = q.mass_unit,
      curr = q.curr_unit,
      temp = q.temp_unit
    };
  else
    return struct {
      leng = q.leng_unit+"^"+string(int(round(q.leng_dim*pow))),
      time = q.time_unit+"^"+string(int(round(q.time_dim*pow))),
      mass = q.mass_unit+"^"+string(int(round(q.mass_dim*pow))),
      curr = q.curr_unit+"^"+string(int(round(q.curr_dim*pow))),
      temp = q.temp_unit+"^"+string(int(round(q.temp_dim*pow)))
    };
}

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_add(q1, q2)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  if(   q1.leng_dim != q2.leng_dim
     || q1.time_dim != q2.time_dim
     || q1.mass_dim != q2.mass_dim
     || q1.curr_dim != q2.curr_dim
     || q1.temp_dim != q2.temp_dim
    )
    throw MathError, sprintf("error (%s): cannot add %S and %S", _function_name(), q1, q2);
  % else
  if(   (q1.leng_dim!=0  &&  q1.leng_unit != q2.leng_unit)
     || (q1.time_dim!=0  &&  q1.time_unit != q2.time_unit)
     || (q1.mass_dim!=0  &&  q1.mass_unit != q2.mass_unit)
     || (q1.curr_dim!=0  &&  q1.curr_unit != q2.curr_unit)
     || (q1.temp_dim!=0  &&  q1.temp_unit != q2.temp_unit)
    )
  {
%    q1 = convert_units(q1; SI, copy);
%    q2 = convert_units(q2; SI, copy);
    q2 = convert_units(q2;; struct_combine(unit_qualifiers(q1; nodim), struct { copy }));
  }
  return physical_quantity(q1.numerical_value + q2.numerical_value;; unit_qualifiers(q1));
}
__add_binary("+", PhysicalQuantity_Type, &physical_quantity_add, PhysicalQuantity_Type, PhysicalQuantity_Type);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_chs(q)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  return physical_quantity(-q.numerical_value;; unit_qualifiers(q));
}
__add_unary("-", PhysicalQuantity_Type, &physical_quantity_chs, PhysicalQuantity_Type);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_sub(q1, q2)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  return q1 + (-q2);
}
__add_binary("-", PhysicalQuantity_Type, &physical_quantity_sub, PhysicalQuantity_Type, PhysicalQuantity_Type);


        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_mul(q1, q2)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  if(q1.leng_dim!=0  &&  q2.leng_dim!=0  &&  q1.leng_unit != q2.leng_unit)
    q2 = convert_units(q2; leng=q1.leng_unit, copy);
  if(q1.leng_dim==0  &&  q2.leng_dim!=0)  q1.leng_unit = q2.leng_unit;

  if(q1.time_dim!=0  &&  q2.time_dim!=0  &&  q1.time_unit != q2.time_unit)
    q2 = convert_units(q2; time=q1.time_unit, copy);
  if(q1.time_dim==0  &&  q2.time_dim!=0)  q1.time_unit = q2.time_unit;

  if(q1.mass_dim!=0  &&  q2.mass_dim!=0  &&  q1.mass_unit != q2.mass_unit)
    q2 = convert_units(q2; mass=q1.mass_unit, copy);
  if(q1.mass_dim==0  &&  q2.mass_dim!=0)  q1.mass_unit = q2.mass_unit;

  if(q1.curr_dim!=0  &&  q2.curr_dim!=0  &&  q1.curr_unit != q2.curr_unit)
    q2 = convert_units(q2; curr=q1.curr_unit, copy);
  if(q1.curr_dim==0  &&  q2.curr_dim!=0)  q1.curr_unit = q2.curr_unit;

  if(q1.temp_dim!=0  &&  q2.temp_dim!=0  &&  q1.temp_unit != q2.temp_unit)
    q2 = convert_units(q2; temp=q1.temp_unit, copy);
  if(q1.temp_dim==0  &&  q2.temp_dim!=0)  q1.temp_unit = q2.temp_unit;

  return physical_quantity(q1.numerical_value * q2.numerical_value;
				leng=q1.leng_unit+"^"+string(q1.leng_dim+q2.leng_dim),
				time=q1.time_unit+"^"+string(q1.time_dim+q2.time_dim),
				mass=q1.mass_unit+"^"+string(q1.mass_dim+q2.mass_dim),
				curr=q1.curr_unit+"^"+string(q1.curr_dim+q2.curr_dim),
				temp=q1.temp_unit+"^"+string(q1.temp_dim+q2.temp_dim),
			       );
}
__add_binary("*", PhysicalQuantity_Type, &physical_quantity_mul, PhysicalQuantity_Type, PhysicalQuantity_Type);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_mul_sq(s, q)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  return physical_quantity(s*q.numerical_value;; unit_qualifiers(q));
}
__add_binary("*", PhysicalQuantity_Type, &physical_quantity_mul_sq, Any_Type, PhysicalQuantity_Type);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_mul_qs(q, s)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  return physical_quantity(s*q.numerical_value;; unit_qualifiers(q));
}
__add_binary("*", PhysicalQuantity_Type, &physical_quantity_mul_qs, PhysicalQuantity_Type, Any_Type);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_power(q, p)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable dim;
  if(__is_numeric(p)!=1)  % Integer_Type
    foreach dim ([q.leng_dim, q.time_dim, q.mass_dim, q.curr_dim, q.temp_dim]*p)
      if( abs(dim - round(dim)) > 1e-6 )
        throw MathError, sprintf("error (%s): cannot compute (%S)^%S", _function_name(), q, p);
  return physical_quantity(q.numerical_value^p;; unit_qualifiers(q; power=p));
}
__add_binary("^", PhysicalQuantity_Type, &physical_quantity_power, PhysicalQuantity_Type, Any_Type);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_div(q1, q2)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  return q1 * (@q2)^-1;
}
__add_binary("/", PhysicalQuantity_Type, &physical_quantity_div, PhysicalQuantity_Type, PhysicalQuantity_Type);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_div_qs(q, s)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  return q * s^-1;
}
__add_binary("/", PhysicalQuantity_Type, &physical_quantity_div_qs, PhysicalQuantity_Type, Any_Type);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define physical_quantity_div_sq(s, q)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  return s * q^-1;
}
__add_binary("/", PhysicalQuantity_Type, &physical_quantity_div_sq, Any_Type, PhysicalQuantity_Type);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define handle_physical_quantity_units(q, units)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable unit, dim;
  foreach unit (units)
  { (unit, dim) = parse_unit_power(unit);
    q *= unit_to_baseunit(unit, physical_quantity_baseunits.unit) ^ dim;
  }
  return q;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define physical_quantity_length_in_m()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{physical_quantity_length_in_m}
%\synopsis{defines a new length unit or gets the conversion for an existing one}
%\usage{physical_quantity_length_in_m(String_Type name, Double_Type value);
%\altusage{Double_Type value = physical_quantity_length_in_m(String_Type name);}
%}
%\description
%    The unit  \code{name} = \code{value} m  can be used in
%       \code{physical_quantity(x; leng=name);  % = x * value} m
%\seealso{physical_quantity}
%!%-
{
  variable name, value;
  switch(_NARGS)
  { case 1: name = ();
      return physical_quantity_baseunits.length_in_m[name];
  }
  { case 2: (name, value) = ();
      physical_quantity_baseunits.length_in_m[name] = value;
  }
  { help(_function_name);
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define physical_quantity_time_in_s()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{physical_quantity_time_in_s}
%\synopsis{defines a new time unit or gets the conversion for an existing one}
%\usage{physical_quantity_time_in_s(String_Type name, Double_Type value);
%\altusage{Double_Type value = physical_quantity_time_in_s(String_Type name);}
%}
%\description
%    The unit  \code{name} = \code{value} s  can be used in
%       \code{physical_quantity(x; time=name);  % = x * value} s
%\seealso{physical_quantity}
%!%-
{
  variable name, value;
  switch(_NARGS)
  { case 1: name = ();
      return assoc_key_exists(physical_quantity_baseunits.time_in_s, name)
             ? physical_quantity_baseunits.time_in_s[name]
             : NULL;
  }
  { case 2: (name, value) = ();
      physical_quantity_baseunits.time_in_s[name] = value;
  }
  { help(_function_name);
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define physical_quantity_mass_in_kg()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{physical_quantity_mass_in_kg}
%\synopsis{defines a new mass unit or gets the conversion for an existing one}
%\usage{physical_quantity_mass_in_kg(String_Type name, Double_Type value);
%\altusage{Double_Type value = physical_quantity_mass_in_kg(String_Type name);}
%}
%\description
%    The unit  \code{name} = \code{value} kg  can be used in
%       \code{physical_quantity(x; mass=name);  % = x * value} kg
%\seealso{physical_quantity}
%!%-
{
  variable name, value;
  switch(_NARGS)
  { case 1: name = ();
      return assoc_key_exists(physical_quantity_baseunits.mass_in_kg, name)
             ? physical_quantity_baseunits.mass_in_kg[name]
             : NULL;
  }
  { case 2: (name, value) = ();
      physical_quantity_baseunits.mass_in_kg[name] = value;
  }
  { help(_function_name);
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define physical_quantity_current_in_A()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{physical_quantity_current_in_A}
%\synopsis{defines a new current unit or gets the conversion for an existing one}
%\usage{physical_quantity_current_in_A(String_Type name, Double_Type value);
%\altusage{Double_Type value = physical_quantity_current_in_A(String_Type name);}
%}
%\description
%    The unit  \code{name} = \code{value} A  can be used in
%       \code{physical_quantity(x; curr=name);  % = x * value} A
%\seealso{physical_quantity}
%!%-
{
  variable name, value;
  switch(_NARGS)
  { case 1: name = ();
      return assoc_key_exists(physical_quantity_baseunits.current_in_A, name)
             ? physical_quantity_baseunits.current_in_A[name]
             : NULL;
  }
  { case 2: (name, value) = ();
      physical_quantity_baseunits.current_in_A[name] = value;
  }
  { help(_function_name);
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define physical_quantity_temperature_in_K()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{physical_quantity_temperature_in_K}
%\synopsis{defines a new temperature unit or gets the conversion for an existing one}
%\usage{physical_quantity_temperature_in_K(String_Type name, Double_Type value);
%\altusage{Double_Type value = physical_quantity_temperature_in_K(String_Type name);}
%}
%\description
%    The unit  \code{name} = \code{value} K  can be used in
%       \code{physical_quantity(x; curr=name);  % = x * value} K
%\seealso{physical_quantity}
%!%-
{
  variable name, value;
  switch(_NARGS)
  { case 1: name = ();
      return assoc_key_exists(physical_quantity_baseunits.temperature_in_K, name)
             ? physical_quantity_baseunits.temperature_in_K[name]
             : NULL;
  }
  { case 2: (name, value) = ();
      physical_quantity_baseunits.temperature_in_K[name] = value;
  }
  { help(_function_name);
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define physical_quantity_unit()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{physical_quantity_unit}
%\synopsis{defines a new compound unit or gets the value of an existing one}
%\usage{physical_quantity_unit(String_Type name, Double_Type value;; qualifiers);
%\altusage{physical_quantity_unit(String_Type name, PhysicalQuantity_Type value);}
%\altusage{PhysicalQuantity_Type value = physical_quantity_unit(String_Type name);}
%}
%\description
%    If the second argument of the \code{physical_quantity_unit} function
%    is a double value, the value of the unit is
%    constructed with the \code{physical_quantity} function,
%    i.e., all its qualifiers can be applied.
%
%    The unit  \code{name} = \code{value}  can be used in
%       \code{physical_quantity(x; unit=name);  % = x * value}
%\seealso{physical_quantity}
%!%-
{
  variable name, value;
  variable baseunits = physical_quantity_baseunits.unit;
  switch(_NARGS)
  { case 1: name = ();
      return assoc_key_exists(baseunits, name)
             ? baseunits[name]
             : NULL;
  }
  { case 2: (name, value) = ();
      variable interpretation = unit_to_baseunit(name, baseunits; get_interpretation);
      ifnot(assoc_key_exists(baseunits, name) || length(interpretation)==0)
        vmessage("warning (%s): '%s' could be confused with %s",
		 _function_name. name, strjoin(list_to_array(interpretation), " or "));

      baseunits[name]
       = __is_numeric(value)
         ? physical_quantity(value;; __qualifiers)
         : value;
  }
  { help(_function_name);
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

variable physical_quantity_const = struct {
  BOLTZMANN = physical_quantity(1.3806503e-23; leng="m^2", time="s^-2", mass="kg", temp="K^-1"),
         kB = physical_quantity(1.3806503e-23; leng="m^2", time="s^-2", mass="kg", temp="K^-1"),
  ELECTRON_CHARGE = physical_quantity(1.602176462e-19; time="s", curr="A"),
                e = physical_quantity(1.602176462e-19; time="s", curr="A"),
  GRAVITATIONAL_CONSTANT = physical_quantity(6.673e-11; leng="m^3", time="s^-2", mass="kg^-1"),
                       G = physical_quantity(6.673e-11; leng="m^3", time="s^-2", mass="kg^-1"),
  MASS_ELECTRON = physical_quantity(9.10938188e-31; mass="kg"),
            mEl = physical_quantity(9.10938188e-31; mass="kg"),
  MASS_NEUTRON = physical_quantity(1.67492716e-27; mass="kg"),
            mN = physical_quantity(1.67492716e-27; mass="kg"),
  MASS_PROTON = physical_quantity(1.67262158e-27; mass="kg"),
           mP = physical_quantity(1.67262158e-27; mass="kg"),
  PLANCKS_CONSTANT_H = physical_quantity(6.62606876e-34; leng="m^2", time="s^-1", mass="kg"),
                   h = physical_quantity(6.62606876e-34; leng="m^2", time="s^-1", mass="kg"),
  PLANCKS_CONSTANT_HBAR = physical_quantity(1.05457159642e-34; leng="m^2", time="s^-1", mass="kg"),
                   hbar = physical_quantity(1.05457159642e-34; leng="m^2", time="s^-1", mass="kg"),
  SPEED_OF_LIGHT = physical_quantity(299792458; leng="m", time="s^-1"),
               c = physical_quantity(299792458; leng="m", time="s^-1"),
  VACUUM_PERMITTIVITY = physical_quantity(8.854187817e-12; leng="m^-3", time="s^4", mass="kg^-1", curr="A^2"),
             epsilon0 = physical_quantity(8.854187817e-12; leng="m^-3", time="s^4", mass="kg^-1", curr="A^2"),
};

physical_quantity_baseunits.length_in_m["m"] = 1.;
physical_quantity_baseunits.length_in_m["A"] = 1e-10;
physical_quantity_baseunits.length_in_m["in"] = 0.0254;
physical_quantity_baseunits.length_in_m["ft"] = 0.30480; % = 12 in
physical_quantity_baseunits.length_in_m["yd"] = 0.91440; % =  3 ft
physical_quantity_baseunits.length_in_m["ly"] = 9.4607304725808e15;
physical_quantity_baseunits.length_in_m["pc"] = 3.0857e16;
physical_quantity_baseunits.length_in_m["R[sun]"] = 6.955e8;
physical_quantity_baseunits.length_in_m["AU"] = 1.49597870691e11;

physical_quantity_baseunits.time_in_s["s"] = 1.;
physical_quantity_baseunits.time_in_s["min"] = 60.;
physical_quantity_baseunits.time_in_s["h"] = 3600.;
physical_quantity_baseunits.time_in_s["d"] = 86400.;
physical_quantity_baseunits.time_in_s["yr"] = 3.15576e7;  % 3.1556952e7 = 365.2425 d

physical_quantity_baseunits.mass_in_kg["g"] = 1e-3;
physical_quantity_baseunits.mass_in_kg["u"] = 1.66053873e-27;
physical_quantity_baseunits.mass_in_kg["M[sun]"] = 1.98892e30;
%physical_quantity_baseunits.mass_in_kg["eV"] = 1.78266173e-36;

physical_quantity_baseunits.current_in_A["A"] = 1;

physical_quantity_baseunits.temperature_in_K["K"] = 1;

physical_quantity_unit("N", 1; mass="kg", leng="m", time="s^-2");
physical_quantity_unit("J", 1; unit="N", leng="m");
physical_quantity_unit("W", 1; unit="J", time="s^-1");
physical_quantity_unit("V", 1; unit="W", curr="A^-1");

physical_quantity_unit("erg", 1e-7; unit="J");
physical_quantity_unit("eV", physical_quantity_const.e * physical_quantity(1; unit="V"));
physical_quantity_unit("Jy", 1e-26; unit="J", leng="m^-2");
physical_quantity_unit("Hz", 1; time="s^-1");

%
% Code from src/plot/plotGTI.sl
%
define plotGTI()
%!%+
%\function{plotGTI}
%\synopsis{visualizes Good Time Intervals}
%\usage{plotGTI(Struct_Type gti[, Double_Type offset]);}
%\qualifiers{
%\qualifier{ylevel}{plot GTIs at constant y-level}
%\qualifier{overplot}{}
%}
%!%-
{
  variable gti, offset=0;
  switch(_NARGS)
  { case 1: gti = (); }
  { case 2: (gti, offset) = (); }
  { help(_function_name()); return; }

  variable overplot = qualifier_exists("overplot");
  variable col = get_plot_options().color;
  variable ngtis = length(gti.start);
  variable ymin = offset-1;
  variable ymax = ngtis+offset;

  if (overplot){
    (,,ymin, ymax) = _pgqwin(); % Get ranges of current plot
  }
  variable dy = (ymax-ymin)/double(ngtis+1);

  ifnot(overplot){
    xrange(min(gti.start), max(gti.stop));
    yrange(ymin, ymax);
  }
  
  variable ii;
  _for ii (0, ngtis-1, 1){
    color(col); oplot([gti.start[ii], gti.stop[ii]], ymin+(ii+1)*dy*[1, 1]+offset);
  }
}

define oplotGTIs()
%!%+
%\function{oplotGTIs}
%\synopsis{visualizes Good Time Intervals}
%\usage{oplotGTIs(Struct_Type gti[, Double_Type offset]);}
%!%-
{
  variable gti, offset=0;
  switch(_NARGS)
  { case 1: gti = (); }
  { case 2: (gti, offset) = (); }
  { help(_function_name()); return;}

  plotGTI(gti, offset;; struct_combine(__qualifiers, struct{overplot}));
}
%
% Code from src/data/spectrum.datastructures/plot_combined_data_model_residuals.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define plot_combined_data_model_residuals()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{plot_combined_data_model_residuals}
%\synopsis{plots data and model counts, and residuals for a combination of data sets}
%\usage{plot_combined_data_model_residuals([Integer_Type id[]]);}
%\qualifiers{
%\qualifier{dcol}{color of data [default=1]}
%\qualifier{mcol}{color of model [default=2]}
%}
%\description
%    If no indices \code{id} of data sets are specified, \code{all_data} are used.\n
%    The combined data, model and residuals are plotted in a two panel multiplot.
%\seealso{get_combined_data_model_residuals}
%!%-
{
  variable datasets;
  switch(_NARGS)
  { case 0: datasets = all_data; }
  { case 1: datasets = (); }

  variable d, m, r;
  (d, m, r) = get_combined_data_model_residuals(datasets);

  multiplot([4,1]);
  plot_bin_density;
  color(qualifier("dcol", 1)); plot_with_err(d);
  color(qualifier("mcol", 2)); ohplot(m);
  plot_bin_integral;
  color(qualifier("dcol", 1)); plot_with_err(r);
  r.value *= 0;
  color(qualifier("mcol", 1)); ohplot(r);
}

%
% Code from src/misc/radio/plot_component.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

define plot_component() {
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{plot_component}
%\synopsis{to plot modelfit parameters of different components}
%\usage{plot_component([array of \code{struct comp}]);}
%\qualifiers{
%\qualifier{outputfile}{outputfile name and directory by hand}
%\qualifier{mjd_min}{[=53736.0] lower limit of time axis in mjd (default: 1/1/2006)}
%\qualifier{mjd_max}{[=55927.0] upper limit of time axis in mjd (default: 1/1/2012)}
%\qualifier{date_min}{[=2006.0] lower limit of time axis (default: 1/1/2006)}
%\qualifier{date_max}{[=2012.0] upper limit of time axis (default: 1/1/2012)}
%\qualifier{distance_min}{[=-2] lower limit of distance axis in mas}
%\qualifier{distance_max}{[=20] upper limit of distance axis in mas}
%\qualifier{flux_min}{[=1e-5] lower limit of flux axis in Jy}
%\qualifier{flux_max}{[=5] upper limit of flux axis in Jy}
%\qualifier{TB_min}{[=10^5] lower limit of brightness temperature axis in K}
%\qualifier{TB_max}{[=10^15] upper limit of brightness temperature axis in K}
%\qualifier{size_min}{[=1e-3] lower limit of size axis in mas^2}
%\qualifier{size_max}{[=100] upper limit of size axis in mas^2}
%\qualifier{pos_min}{[=-180] lower limit of pos angle in degrees}
%\qualifier{pos_max}{[=180] upper limit of pos angle in degrees}
%\qualifier{labels}{[=default] provide array of custom labels for the components}
%\qualifier{sym}{[=default] provide array of custom symbols for the components}
%\qualifier{symcolor}{[=default] provide array of custom colors for the components}
%\qualifier{symsize}{[=1] provide array of custom symbol sizes or one value to apply size to all}
%\qualifier{legend}{create a legend}
%\qualifier{lpos_x}{set the x position of the legend in world0 system (0 = left, 1=right)}
%\qualifier{lpos_y}{set the y position of the legend in world0 system (0 = bottom, 1=top)}
%\qualifier{linestyle}{[=default] set to 0 if the flux values should not be conntected (or to another value for 
% 					another line style}
%\qualifier{nocomp}{provide a comp structure (or array of structures) to plot not identified components in black}
%\qualifier{plots}{[=[9,10,3,4,11,12]]: provide an array of numbers to generated specified plots:
%					1:	distance vs mjd
%					2:	flux vs mjd
%					3:	T_B vs distance
%					4:	flux vs distance
%					5:	size vs distance 
%					6:	pos angle vs distance	
%					7:	size vs mjd
%					8:	pos angle vs mjd
%					9:	distance vs time
%					10:	flux vs time
%					11:	size vs time
%					12:	pos_angle vs time }
%}
%\description
%This functions creates overview plots for modelfit components. The required input format is the comp structure which can be obtained with get_component or an array of the comp structure.
%!%-
variable comp;
  switch(_NARGS)
  { case 1: (comp) = (); }
  { help(_function_name()); return; }

variable ii, jj;
variable major_ticks, minor_ticks;
variable mjd_min	= qualifier("mjd_min", 53736.0); % 1.1.2006
variable mjd_max	= qualifier("mjd_max", 55927.0); % 1.1.2012
variable date_min	= qualifier("date_min", 2006.0);
variable date_max	= qualifier("date_max", 2012.0);
variable distance_min	= qualifier("distance_min", -2);
variable distance_max	= qualifier("distance_max", 20);
variable flux_min	= qualifier("flux_min", 1e-5); % Jy
variable flux_max	= qualifier("flux_max", 5); % Jy
variable tb_min		= qualifier("tb_min", 10^5);
variable tb_max		= qualifier("tb_max", 10^15);
variable size_min	= qualifier("size_min", 1e-3);
variable size_max	= qualifier("size_max", 100);
variable pos_min	= qualifier("pos_min", -180);
variable pos_max	= qualifier("pos_max", 180);
variable dlabels	= qualifier("labels");
variable dsym		= qualifier("sym");
variable dsymcolor	= qualifier("symcolor");
variable dsymsize	= qualifier("symsize");
variable plots		= qualifier("plots",[9,10,3,4,11,12]);
variable x_legend	= qualifier("lpos_x",0.875);
variable y_legend	= qualifier("lpos_y",0.95);

if(qualifier_exists("symsize") == 0) {
	print("default symbol sizes selected");
	dsymsize = ones(length(comp));
}

if(qualifier_exists("symsize") == 1) {
  if(length(dsymsize) == 1) {
    variable copy_dsymsize = dsymsize;
    dsymsize = Double_Type [length(comp)];
    _for ii (0, length(comp)-1, 1){
      dsymsize[ii] = copy_dsymsize;
    }
  }
}

if(qualifier_exists("sym") == 0) {
	print("default symbols selected");
	dsym = Integer_Type [length(comp)];
	_for jj (0, length(comp)-1, 1){
	dsym[jj] = jj mod 9 + 1;
	}	
}

if(qualifier_exists("symcolor") == 0) {
	print("default symbol colors selected");
	dsymcolor = Integer_Type [length(comp)];
	dsymcolor = [16,5,23,17,11,8,2,20,14,13,19,31,33,32,18,12,6,15,21];
}

if(qualifier_exists("labels") == 0) {
	print("default labels selected");
	dlabels = ["A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z"];
	if(length(comp) > length(dlabels)) {
	print("NO default labels possible!");
	exit(-1);
	}
}

variable pl, pl_comp, pl_all, npl, npl_comp, obj, box;

npl = length(plots);
pl = Struct_Type [npl];
obj = Struct_Type [npl];
box = Struct_Type [npl];

_for ii (0, npl-1, 1){
pl[ii] = xfig_plot_new(10., 10.);

switch(plots[ii]) 
	{ case 1:
	pl[ii].world( mjd_min, mjd_max,distance_min,distance_max);
	pl[ii].xlabel("Epoch [MJD]");
	pl[ii].ylabel("Distance [mas]");
	major_ticks = [-100:100:5];
	minor_ticks = [-100:100:1];
	pl[ii].y1axis(; major=major_ticks, minor= minor_ticks, world1);
	}
	{ case 2:
	pl[ii].world(mjd_min, mjd_max, flux_min, flux_max; ylog);
	pl[ii].xlabel("Epoch [MJD]");
	pl[ii].ylabel("Flux [Jy]");
	}
	{ case 3:
	pl[ii].world(distance_min, distance_max, tb_min, tb_max; ylog);
	pl[ii].xlabel("Distance [mas]");
	pl[ii].ylabel("$\mathrm{T_B}$ [K]"R);
	}
	{ case 4:
	pl[ii].world(distance_min, distance_max, flux_min, flux_max; ylog);
	pl[ii].xlabel("Distance [mas]");
	pl[ii].ylabel("Flux [Jy]");
	}
	{ case 5:
	pl[ii].world(distance_min, distance_max, size_min, size_max; ylog);
	pl[ii].xlabel("Distance [mas]");
	pl[ii].ylabel("A $\mathrm{[mas^2]}$"R);
	}
	{ case 6:
	pl[ii].world(distance_min, distance_max, pos_min, pos_max);
	pl[ii].xlabel("Distance [mas]");
	pl[ii].ylabel("Position Angle [deg]");
	}
	{ case 7:
	pl[ii].world(mjd_min, mjd_max, size_min, size_max;ylog);
	pl[ii].xlabel("Epoch [MJD]");
	pl[ii].ylabel("A $\mathrm{[mas^2]}$"R);
	}
	{ case 8:
	pl[ii].world(mjd_min, mjd_max, pos_min, pos_max);
	pl[ii].xlabel("Epoch [MJD]");
	pl[ii].ylabel("Position Angle [deg]");
	}
	{ case 9:
	pl[ii].world( date_min, date_max, distance_min,distance_max);
	pl[ii].xlabel("Epoch [y]");
	pl[ii].ylabel("Distance [mas]");
	major_ticks = [-100:100:5];
	minor_ticks = [-100:100:1];
	pl[ii].y1axis(; major=major_ticks, minor= minor_ticks, world1);
	}
	{ case 10:
	pl[ii].world(date_min, date_max, flux_min, flux_max; ylog);
	pl[ii].xlabel("Epoch [y]");
	pl[ii].ylabel("Flux [Jy]");
	}
	{ case 11:
	pl[ii].world(date_min, date_max, size_min, size_max;ylog);
	pl[ii].xlabel("Epoch [y]");
	pl[ii].ylabel("A $\mathrm{[mas^2]}$"R);
	}
	{ case 12:
	pl[ii].world(date_min, date_max, pos_min, pos_max);
	pl[ii].xlabel("Epoch [y]");
	pl[ii].ylabel("Position Angle [deg]");
	}

%create a rectangle for the legend in every plot
if(qualifier_exists("legend")==1){	
  box[ii] = xfig_new_rectangle (2.25, 0.4*(1+length(comp)));
  box[ii].set_area_fill (-1);
  box[ii].set_fill_color ("white");
  pl[ii].add_object (box[ii], x_legend, y_legend-length(comp)/50.;world00);
}

}

% loop over all components
_for jj (0, length(comp)-1, 1) {
  %Legend
  _for ii (0, npl-1, 1){
    if(qualifier_exists("legend")==1){
      obj[ii] = xfig_new_text (dlabels[jj]);
      pl[ii].add_object(obj[ii], x_legend + 0.025 , y_legend-0.025-jj/25.;world00);
      pl[ii].plot(x_legend-0.075,  y_legend-0.025-jj/25.; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj],world00);
    }
  switch(plots[ii]) 
  { case 1:
  pl[ii].plot(comp[jj].mjd, comp[jj].distance; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj],line = qualifier("linestyle",2), world1);
  }
  { case 2:
  pl[ii].plot(comp[jj].mjd, comp[jj].flux, 0.15*comp[jj].flux; eb_color=dsymcolor[jj], size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  { case 3:
  pl[ii].plot(comp[jj].distance, comp[jj].tb; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  { case 4:
  pl[ii].plot(comp[jj].distance, comp[jj].flux, 0.15*comp[jj].flux; eb_color=dsymcolor[jj], size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  { case 5:
  pl[ii].plot(comp[jj].distance, comp[jj].size; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  { case 6:
  pl[ii].plot(comp[jj].distance, comp[jj].posangle; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  { case 7:
  pl[ii].plot(comp[jj].mjd, comp[jj].size; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  { case 8:
  pl[ii].plot(comp[jj].mjd, comp[jj].posangle; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  { case 9:
  pl[ii].plot(comp[jj].date, comp[jj].distance; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj],line = qualifier("linestyle",2), world1);
  }
  { case 10:
  pl[ii].plot(comp[jj].date, comp[jj].flux, 0.15*comp[jj].flux; eb_color=dsymcolor[jj], size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  { case 11:
  pl[ii].plot(comp[jj].date, comp[jj].size; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  { case 12:
  pl[ii].plot(comp[jj].date, comp[jj].posangle; size=dsymsize[jj], sym=dsym[jj], symcolor=dsymcolor[jj], line = qualifier("linestyle",2), world1);
  }
  }
}

if(qualifier_exists("nocomp")) {
variable nocomp = qualifier("nocomp",NULL);
  _for ii (0, npl-1, 1){
    switch(plots[ii]) 
    { case 1:
    pl[ii].plot(nocomp.mjd, nocomp.distance; size=0.7, sym=5, symcolor="black", world1);
    }
    { case 2:
    pl[ii].plot(nocomp.mjd, nocomp.flux, 0.15*nocomp.flux; eb_color="black", size=0.7, sym=5, symcolor="black", world1);
    }
    { case 3:
    pl[ii].plot(nocomp.distance, nocomp.tb;  size=0.7, sym=5, symcolor="black", world1);
    }
    { case 4:
    pl[ii].plot(nocomp.distance, nocomp.flux, 0.15*nocomp.flux; eb_color="black", size=0.7, sym=5, symcolor="black", world1);
    }
    { case 5:
    pl[ii].plot(nocomp.distance, nocomp.size; size=0.7, sym=5, symcolor="black", world1);
    }
    { case 6:
    pl[ii].plot(nocomp.distance, nocomp.posangle; size=0.7, sym=5, symcolor="black", world1);
    }
    { case 7:
    pl[ii].plot(nocomp.mjd, nocomp.size;  size=0.7, sym=5, symcolor="black", world1);
    }
    { case 8:
    pl[ii].plot(nocomp.mjd, nocomp.posangle; size=0.7, sym=5, symcolor="black", world1);
    }
    { case 9:
    pl[ii].plot(nocomp.date, nocomp.distance; size=0.7, sym=5, symcolor="black", world1);
    }
    { case 10:
    pl[ii].plot(nocomp.date, nocomp.flux, 0.15*nocomp.flux; eb_color="black", size=0.7, sym=5, symcolor="black", world1);
    }
    { case 11:
    pl[ii].plot(nocomp.date, nocomp.size;  size=0.7, sym=5, symcolor="black", world1);
    }
    { case 12:
    pl[ii].plot(nocomp.date, nocomp.posangle; size=0.7, sym=5, symcolor="black", world1);
    }
  }
}



if(npl mod 2 == 1) {
  npl_comp = npl / 2 + 1;
  pl_comp = Struct_Type [npl_comp];
    _for ii (0, npl / 2 - 1, 1) {
	pl_comp[ii] = xfig_new_hbox_compound (pl[ 2 * ii ], pl[2 * ii + 1], 0.5);
    }
   pl_comp[npl / 2] = xfig_new_hbox_compound (pl[npl -1], 0.5);
} else {
  npl_comp = npl / 2;
  pl_comp = Struct_Type [npl_comp];
  _for ii (0, npl_comp-1, 1) {
    pl_comp[ii] = xfig_new_hbox_compound (pl[ 2 * ii ], pl[2 * ii+1], 0.5);
  }
}


switch(npl_comp) 
  { case 1:
  pl_all = xfig_new_vbox_compound (pl_comp[0], 0.5);
  }
  { case 2:
  pl_all = xfig_new_vbox_compound (pl_comp[0], pl_comp[1], 0.5);
  }
  { case 3:
  pl_all = xfig_new_vbox_compound (pl_comp[0], pl_comp[1], pl_comp[2], 0.5);
  }
  { case 4:
  pl_all = xfig_new_vbox_compound (pl_comp[0], pl_comp[1], pl_comp[2], pl_comp[3], 0.5);
  }
  { case 5:
  pl_all = xfig_new_vbox_compound (pl_comp[0], pl_comp[1], pl_comp[2], pl_comp[3], pl_comp[4], 0.5);
  }
  { case 6:
  pl_all = xfig_new_vbox_compound (pl_comp[0], pl_comp[1], pl_comp[2], pl_comp[3], pl_comp[4], pl_comp[5], 0.5);
  }

if(qualifier_exists("outputfile")==1) {
variable outfile = qualifier("outputfile");
pl_all.render(outfile);
}

return pl_all;
}


#endif

%
% Code from src/plot/plot_contour_trq.sl
%
define plot_contour_trq(){
%%%%%%%%%%%%%%%%%%%%%%%%%%    
%!%+
%\function{plot_contour_trq}
%\synopsis{creates a contour plot out of the results from function
%    contour_trq.}
%\usage{plot_contour_trq(String_Type inputDir);}
%\qualifiers{
%\qualifier{out}{   post script outputname}
%\qualifier{image}{ fits image outputname}
%\qualifier{x}{     x label}
%\qualifier{y}{     y label}
%}
%\description
%       - \code{inputDir} directory where all torque job files are,
%                  created by function contour_trq.
%
%       Be aware of the issues with the torque (look at the description
%       of contour_trq function), as it may not have all bins^2
%       values, and hence give wrong contour plot.
%        
%       The default post script file name is cont_default.ps.
%       Output fits image file is cont_default.fits.
%
% EXAMPLE
%
%
%       plot_contour_trq("fileDirectory";out="my_out",x="x_lab",y="y_lab");
%
%       creates a plot "my_out.ps" with x-axis label named "x_lab" and y_axis
%       label named "y_lab".                   
%\seealso{contour_trq, missing_contour_trq}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%%%    
   variable inputDir=String_Type;
   variable tempOut="outTemp.txt";
   variable i=Int_Type;
   variable lines=String_Type;
   variable k=Int_Type;
   variable j;
   switch(_NARGS)
     { case 1 : (inputDir)=();}
     { help (_function_name() ); return;}
   
   
   %GET LINES AND SORT THEM
   system("ls $inputDir -1 > $tempOut"$); %get all output files to one file
   variable input=fopen("$tempOut"$,"r");
   lines=fgetslines(input);
   variable chose=[0:length(lines)-2:1];
   lines=lines[chose];
   variable sout=fopen("test.txt","w");
   variable binsSq=sqrt(length(lines));

   %IF ALL NEEDED VALUES ARE CALCULATED OR SOME ARE MISSING
   variable nonZeroDigit=strchop(string(binsSq),'.',0);   %check if number of lines is bins^2
   if(strcmp(nonZeroDigit[1],"0")){%
      ()=fprintf(stdout,"======================\nYou do not have all contour_plot values! Your plot is wrong!\nCheck description of contour_trq function.\nYou may need to run missing_contour_trq function.\n======================\n");
       help (_function_name() );
   }
   binsSq=int(binsSq);
   
   %IF ARRAY HAS NEGATIVE FIRST VALUES
   variable neg=where(strncmp(lines,"-",1)==0); %negative starting numbers
   if(length(neg)!=0){%
      %POS:
      variable pos=where(strncmp(lines,"-",1)!=0); %positive starting numbers
      variable posArr=lines[pos];%all positive lines
      %NEG:
      neg=reverse(neg);
      variable negArr=lines[neg];%all negative lines
      variable start=0;
      variable bins=binsSq;
      variable tempo;
      
      _for j(0,(length(neg)/binsSq)-1,1){
	 variable binPart=[start:bins-1:1];
	 variable neg_temp_Arr=negArr[binPart];
	 neg_temp_Arr=reverse(neg_temp_Arr);
	 ifnot(start){
	    tempo=[neg_temp_Arr];
	 }else{
	    tempo=[tempo,neg_temp_Arr];
	 }
	 start=start+binsSq;
	 bins=bins+binsSq;
      }
      variable negArr_1=@tempo;
      lines=[negArr_1,posArr];
   }else{
      variable sorted=array_sort(lines);
      lines=lines[sorted];
   }

   %SORTED TABLE TO FITS FILE
   variable arr_temp=Double_Type[length(lines)];
   variable arr_x=Double_Type[length(lines)];
   variable arr_y=Double_Type[length(lines)];
   variable arr=Double_Type[binsSq,binsSq];
   variable best_chi_value=1e5;
   _for i(0,length(lines)-1,1){	    
      variable partLines=strchop(lines[i],'_',0);
      lines[i]=strtrans(lines[i],"\n","");
      variable chiFile=fopen(inputDir+"/"+lines[i],"r");
      
      variable chiLine=fgetslines(chiFile);
      k=0;
      while(k<length(chiLine)){%
	 if(string_matches(chiLine[k],"Chi",1)!=NULL){
	    variable chi=strchop(chiLine[k],'=',0);
	    break;
	 }
	 k++;
      }
      arr_temp[i]=atof(chi[1]);
      if(arr_temp[i] < best_chi_value){%
	 best_chi_value=arr_temp[i];
	 variable best_x=atof(partLines[0]);
	 variable best_y=atof(partLines[1]);
      }
      
      arr_x[i]=atof(partLines[0]);
      arr_y[i]=atof(partLines[1]);
      ()=fclose(chiFile);
   }

   %PART ARRAY TO bins * bins
   variable x,y;
   variable b=0;
   _for x(0,binsSq-1,1){%
      _for y(0,binsSq-1,1){%
	 arr[y,x]=(arr_temp[b]-min(arr_temp));
	 b++;
      }
   }
   system("rm $tempOut"$);
   ()=fclose(input);

   %CREATE FITS IMAGE
   fits_write_image("test.fits",arr);
   variable fd=fits_open_file("test.fits","wr");
   ()=fprintf(stdout,"Best minimum value is %f.\n",best_chi_value);
   fits_update_key(fd,"BESTSTAT",best_chi_value);%min(arr_temp));
   fits_update_key(fd,"BEST_X",best_x);
   fits_update_key(fd,"BEST_Y",best_y);
   fits_update_key(fd,"PXMIN",min(arr_x));
   fits_update_key(fd,"PXMAX",max(arr_x));
   fits_update_key(fd,"PXNUM",binsSq);
   fits_update_key(fd,"PYMIN",min(arr_y));
   fits_update_key(fd,"PYMAX",max(arr_y));
   fits_update_key(fd,"PYNUM",binsSq);
   fits_close_file(fd);

   variable image=load_conf("test.fits");

   variable outputImage=qualifier("image",NULL);
   if(outputImage!=NULL){
      save_conf(image,outputImage+".fits");
   }else{
      save_conf(image,"cont_default.fits");
   }
   
   %save_conf(image,"cont_default.fits");
   variable outputname=qualifier("out",NULL);
   if(outputname!=NULL){
      open_plot(outputname+".ps/cps");
   }else{
      open_plot("cont_default.ps/cps");
   }
   variable xlab=qualifier("x",NULL);
   variable ylab=qualifier("y",NULL);
   if(xlab!=NULL){
      xlabel(xlab);
   }
   if(ylab!=NULL){
      ylabel(ylab);
   }
   plot_conf(image);
   ()=close_plot;
}

%
% Code from src/misc/radio/plot_jet_speed.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define plot_jet_speed ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{plot_jet_speed}
%\synopsis{visualizes the jet_speed_struct used by init_jet_fit}
%\usage{plot_jet_speed (Struct_Type jet_speed_struct);}
%\qualifiers{
%\qualifier{xrange}{set the xrange (provide a list or array)}
%\qualifier{yrange}{set the yrange (provide a list or array)}
%}
%\seealso{init_jet_fit}
%!%-
%
{
  variable jss;
  switch(_NARGS)
  { case 1: jss = ();}
  { help(_function_name()); return; }
  variable one_day_in_years = 0.002737803091986241; % = 1./365.256363004; % sidereal year
  variable xmin,xmax;
  if (qualifier_exists("xrange"))
  { xmin = qualifier("xrange")[0];
    xmax = qualifier("xrange")[1];  }
  else
  { ( xmin, xmax ) = min_max( jss.time );
    xrange (xmin - 0.05*(xmax-xmin) , xmax + 0.05*(xmax-xmin) );  }
  variable ymin, ymax;
  if (qualifier_exists("yrange"))
  { ymin = qualifier("yrange")[0];
    ymax = qualifier("yrange")[1];  }
  else
  { ymin = min( jss.distance - jss.derr);
    ymax = max( jss.distance + jss.derr); }
  yrange (ymin - 0.05*(ymax-ymin) , ymax + 0.05*(ymax-ymin) );
  xlin; ylin;
  variable ci = unique(jss.component);
  variable i,k, counter = 0 , clr, pst;
  init_plot;
  variable pnt_styles = [1,2,4,8,10,11,12,14,15];
  variable colors     = [1:19];
  foreach i (ci)
  {
    k = where( jss.component == jss.component[i]);
    clr = pnt_styles[(counter mod 9)];
    pst = colors[(counter mod  19)];
    color (clr); point_style(pst);
    counter += 1;
    plot_with_err ( jss.time[k] , jss.distance[k]-jss.offset[k]
		    , jss.derr[k] ; overplot);
    connect_points(-1); color(clr);
    oplot([_max(xmin, jss.t0[i]), max(jss.time[k]) ],
    [(_max(xmin, jss.t0[i]) - jss.t0[i])*one_day_in_years*jss.speed[i],
     (max(jss.time[k]) - jss.t0[i])*one_day_in_years*jss.speed[i]
    ]);
  }
}

%
% Code from src/misc/radio/plot_spix.sl
%

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

%%%%%%%%%%%%%%%%
define plot_spix()
%%%%%%%%%%%%%%%%
%!%+
%\function{plot_spix}
%\synopsis{plots a spectral index map using the struct provided by make_spix}
%\usage{plot_spix(Struct_Type values, String_Type outputname);}
%\qualifiers{
%\qualifier{inv_color}{[=0] inverts color scale (i.e., roles of red and blue switch)}
%\qualifier{min_spix}{[=min(values.spec_map)] minimum spectral index to be plotted}
%\qualifier{max_spix}{[=max(values.spec_map)] maximum spectral index to be plotted}
%\qualifier{lum_scale}{[=1] use the sum of the intensity of both images to scale
%                      the brightness of the spectral index map, set to 0 in order to
%                      show the significant regions at uniform brightness}
%\qualifier{nsigma [=0]}{spectral index values are only shown for pixels where
%                      the emission (in both bands) is above the specified
%                      significance (e.g., nsigma=3 means that the emission
%                      has to be 3*sigma above the noise level)}
%\qualifier{nsigma_lo}{[=0] emission limit for low frequency  band separately}
%\qualifier{nsigma_hi}{[=0] emission limit for high frequency band separately}
%\qualifier{low_lum_limit}{[=0] number of stdevs above the median that luminosity data
%                                        begins to be displayed}
%\qualifier{hi_lum_limit}{[=0] number of (e^n) stdevs above the median at which color
%                                        is fully saturated (0 -> no limit)}
%\qualifier{ra_frac}{[={0.0,1.0}] fraction of RA to display with {left,right} limits,
%                                        will keep scaling}
%\qualifier{dec_frac}{[={0.0,1.0}] fraction of DEC to display with {left,right} limits,
%                                        will keep scaling}
%\qualifier{ra_mas}{[={NULL,NULL}] {left,right} limits of image in mas}
%\qualifier{dec_mas}{[={NULL,NULL}] {top,bottom} limits of image in mas}
%\qualifier{size}{[=14] size of plot (maximum dimension)}
%\qualifier{print_mode}{[=0] set =1 for white background}
%\qualifier{blue_bkgr}{[=0] for a dark blue background set to 1}
%\qualifier{quadratic}{create a quadratic image, by default the scaling is such that
%                                        scaling in RA and DEC is equal}
%\qualifier{no_labels}{plot no labels}
%\qualifier{pmodecolor}{[=1] color saturation decreases wth this root of luminosity
%                                        (i.e., 2=sqrt) in print mode only}
%\qualifier{beam_color}{[="gray"] beam color}
%\qualifier{src_name}{[=values.source] manually set source name in plot}
%\qualifier{obs_date}{[=values.date] manually set date in plot}
%\qualifier{xyunit}{[="mas"] unit of the x/y labels, can be changed between "mas", "arcsec", "arcmin" and "deg";
%                                               assumption: FITS header sets unit "mas"}                                       
%\qualifier{xfig_tmp_dir}{[=xfig_get_tmp_dir] set the path for the tmp directory used by xfig}
%\qualifier{xfig_autoeps_dir}{[=xfig_get_autoeps_dir] set the path for the autoeps directory}
%\qualifier{color_scheme}{[="hot"] setting another colorscheme (lum_scale=0 is automatically set)}
%\qualifier{h_scalebar}{if given, a horizontal scalebar is plotted above the spixmap}
%\qualifier{no_scalebar}{if given, no scalebar is plotted}
%}
%\description
%    This function creates a spectral image plot with xfig from the
%    structure generated by make_spix. The log of the average brightness
%    is displayed as color saturation and the spectral index as hue so
%    that both spectral index and brightness are apparent in the image.
%    It will ouput a .pdf file of the plot with a color scale bar on the
%    side.  For the luminosity scaling, the data below the median value
%    is cut out and the data above the median is fit to a gaussian. The
%    hi_lum_limit and low_lum_limit are based on the stdev from this fit.
%\seealso{make_spix, read_spix, write_spix}
%!%-
{  
  variable values,outname;
  switch(_NARGS)
  { case 2: (values,outname) = (); }
  { help(_function_name()); return; }
  
  %% moving input data into local variables so it's not modified
  variable avg_lum  = @values.avg_lum;
  variable spec_map = @values.spec_map;
  variable mlo      = @values.lo_map;
  variable mhi      = @values.hi_map_shifted;
  variable length_x = length(mlo[*,0]);
  variable length_y = length(mlo[0,*]);

  variable low_lum_limit = qualifier("low_lum_limit", 0); % sets the number of standard deviations above the median that luminosity data begins to be displayed
  variable hi_lum_limit  = qualifier("hi_lum_limit", 0);  % sets the number of (e^n) standard deviations at which color is fully saturated (default 0 -> no limit)
  variable lum_scale     = qualifier("lum_scale", 1);     % set to 0 to turn off brightness scaling
  variable inv_color     = qualifier("inv_color",0);      % if set to 1 blue -> optically thin (low spectral index)
  variable image_size    = qualifier("size", 14.);        % size of resulting image (maximum dimension)
  variable print_mode    = qualifier("print_mode", 0);    % makes white background if = 1;
  variable pmodecolor    = qualifier("pmodecolor", 1);    % causes color saturation to decrease with this root of luminosity (i.e. 2=sqrt) in print mode only
  variable src_name      = qualifier("src_name", values.source);
  variable obs_date      = qualifier("obs_date", values.date);
  variable dec_frac      = qualifier("dec_frac",[0.,1.]);
  variable ra_frac       = qualifier("ra_frac",[0.,1.]);
  variable color_scheme  = qualifier("color_scheme","hot");
  variable nsigma        = qualifier("nsigma", 0 );        % if set, only spix values, where the emission (in both bands) is above the specified number of sigma, are shown
  variable nsigma_lo     = qualifier("nsigma_lo", nsigma); % emission in low  freq band has to above the specified number of sigma       
  variable nsigma_hi     = qualifier("nsigma_hi", nsigma); % emission in high freq band has to above the specified number of sigma       
  variable no_labels	= qualifier("no_labels");
  variable xyunit       = qualifier("xyunit","mas");
  % if you want to have a dark blue background:
  variable bluebkgr = qualifier("blue_bkgr", 0);
    
  %% fitting noise in avg_lum if parameters are used later
  %% assumes median value is in noise. Fits a gaussian to the (hopefully) noise
  if(lum_scale){ %low_lum_limit!=0 or hi_lum_limit!=0){
    variable sigma_lum, med_lum;
    (med_lum , sigma_lum) = fit_gauss_to_img_noise (avg_lum);
  }
  %% fitting noise of the low  frequency map if parameters are used later
  if(nsigma_lo != 0){
    variable sigma_lo, med_lo;
    ( med_lo , sigma_lo) = fit_gauss_to_img_noise (mlo);
  }
  %% fitting noise of the high frequency map if parameters are used later
  if(nsigma_hi != 0){
    variable sigma_hi, med_hi;
    ( med_hi , sigma_hi) = fit_gauss_to_img_noise (mhi);
  }

  %% displaying average luminosity as color saturation
  %% in the next steps the values for avg_lum are set
  if(lum_scale){
    avg_lum = _max( avg_lum , med_lum + low_lum_limit * sigma_lum );
    if(hi_lum_limit > 0){
      hi_lum_limit = exp(hi_lum_limit);
      avg_lum = _min ( avg_lum , med_lum + hi_lum_limit * sigma_lum );
    }
    avg_lum = log( _max(avg_lum, min(avg_lum[where(avg_lum>0)]) ) );
    avg_lum -= min(avg_lum);
    if(print_mode) avg_lum = avg_lum^pmodecolor;
    avg_lum /= max(avg_lum)*2;
  }
  else {
    avg_lum[*,*] = 0.5; % all values have the full color
  }
  variable bkgr = (print_mode ? 1 : 0); % saturation for unvalied spix pixels, 1 = white background, 0 = black background
  if(nsigma_lo != 0) avg_lum[where(mlo < med_lo + sigma_lo * nsigma_lo)] = bkgr;
  if(nsigma_hi != 0) avg_lum[where(mhi < med_hi + sigma_hi * nsigma_hi)] = bkgr;
  if(print_mode){    avg_lum = 1-avg_lum;  }


  %%%%%%%%%%%%%%%%%%%%% GENERATING PLOTS %%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  %setting plot boundaries
  variable dec_len       = length(spec_map[*,0])-1;
  variable ra_len        = length(spec_map[0,*])-1;
  variable ra_px_center  = @values.ra_px_center;
  variable ra_steps      = @values.ra_steps;
  variable dec_px_center = @values.dec_px_center;
  variable dec_steps     = @values.dec_steps;    
  variable ra_min        = -(ra_px_center-0.5)*ra_steps;
  variable ra_max        = (ra_px_center-0.5)*ra_steps;
  variable dec_min       = -(dec_px_center-0.5)*dec_steps;
  variable dec_max       = (dec_px_center-0.5)*dec_steps;
  
  variable unit_fac,unit_lab_x,unit_lab_y;
  switch (xyunit)
  { case "mas" : unit_fac=1.; }
  { case "arcsec" : unit_fac=1./1e3; }
  { case "arcmin" : unit_fac=1./1e3/60; }
  { case "deg" : unit_fac=1./1e3/3600; }
  unit_lab_x=sprintf("Relative RA [%s]",xyunit);
  unit_lab_y=sprintf("Relative DEC [%s]",xyunit);
  
  % correct ra/dec steps and beam smaj/smin for given unit
  ra_steps *= unit_fac;
  dec_steps *= unit_fac;
  
  
  %% cutting image
  variable indxdec, indxra;
  if (qualifier_exists("ra_mas") or qualifier_exists("dec_mas"))
  {
    variable ra_mas = qualifier("ra_mas", {ra_min,ra_max});
    if (ra_mas[0] < ra_mas[1])  ra_mas = { ra_mas[1], ra_mas[0] };
    ra_mas  = { _min( ra_mas[0], ra_min) , _max( ra_mas[1], ra_max) };
    variable dec_mas = qualifier("dec_mas", {dec_min,dec_max});
    if (dec_mas[0] > dec_mas[1])  dec_mas = { dec_mas[1], dec_mas[0] };
    dec_mas = { _max(dec_mas[0],dec_min) , _min(dec_mas[1],dec_max) };
    indxdec = [int(_max(dec_mas[0]/dec_steps + (dec_px_center-0.5) , 0)) :
	       int(_min(dec_mas[1]/dec_steps + (dec_px_center-0.5) , dec_len))];
    indxra  = [int(_max( ra_mas[0]/ ra_steps + ( ra_px_center-0.5) , 0)) :
	       int(_min( ra_mas[1]/ ra_steps + ( ra_px_center-0.5) , ra_len))];
    ra_px_center  -= int( ra_mas[0]/ra_steps  + (ra_px_center-0.5) );
    dec_px_center -= int(dec_mas[0]/dec_steps + (dec_px_center-0.5) );
    ra_min  =  ra_mas[0];  ra_max = ra_mas[1];
    dec_min = dec_mas[0]; dec_max = dec_mas[1];
  }
  else
  {
    indxdec = [int(dec_len*dec_frac[0]) : int(dec_len*dec_frac[1])];
    indxra  = [int( ra_len* ra_frac[0]) : int( ra_len* ra_frac[1])];
  }
  spec_map = spec_map[ indxdec , indxra ];
  avg_lum  = avg_lum [ indxdec , indxra ];
  mlo      = mlo     [ indxdec , indxra ];
  mhi      = mhi     [ indxdec , indxra ];
  
  %% setting scale
  variable width  = abs(ra_min  - ra_max);
  variable height = abs(dec_min - dec_max);
  if (qualifier_exists("quadratic")){width = 1; height = 1;} % create quadratic image
  variable image_scale = _max(width, height);
  width  *= image_size/image_scale;
  height *= image_size/image_scale;
  variable pl = xfig_plot_new(width,height);
  variable beam_dec_scale = image_size/(image_scale*abs(dec_max-dec_min));  %for scaling the beam later
  variable beam_ra_scale  = image_size/(image_scale*abs(ra_max-ra_min));    %for scaling the beam later
  
  xfig_new_color("grey",0x616161);
  variable color = (print_mode ? "black" : "grey");
  pl.world(ra_min,ra_max,dec_min,dec_max; color=color);
  pl.axis(; color =color, ticlabels2=0);
  ifnot (qualifier_exists("no_labels")){
  pl.xlabel(unit_lab_x);
  pl.ylabel(unit_lab_y);}
  variable spec_map_name = "spec_map.png"; while (stat_file(spec_map_name) != NULL) spec_map_name="X"+spec_map_name;
  
  %%%%%%%%%%%%%%%%%% changing spec_map accoring to the required color scale %%%%%%%%%%%%%%%%%%%%
  % so far there should not have been changes to the spec_map
  %
  variable min_spix      = qualifier("min_spix", min(spec_map[where(avg_lum>0)]));
  variable max_spix      = qualifier("max_spix", max(spec_map[where(avg_lum>0)]));
  %% setting user specified color scheme, if required
  if (qualifier_exists("color_scheme"))
  {
    variable h, s, l; 
    (h,s,l) = rgb2hsl(png_gray_to_rgb(spec_map, color_scheme; gmin= min_spix, gmax=max_spix)); % color scheme just in valid spix range
    if(nsigma_lo != 0) l[where(mlo < med_lo + sigma_lo * nsigma_lo)] = bkgr;
    if(nsigma_hi != 0) l[where(mhi < med_hi + sigma_hi * nsigma_hi)] = bkgr;
    
    if (bluebkgr ==1){
        if(nsigma_lo != 0) h[where(mlo < med_lo + sigma_lo * nsigma_lo)] = 0.66; 
	if(nsigma_lo != 0) s[where(mlo < med_lo + sigma_lo * nsigma_lo)] = 1; 
	if(nsigma_lo != 0) l[where(mlo < med_lo + sigma_lo * nsigma_lo)] = 0.2; 
    
	if(nsigma_hi != 0) h[where(mhi < med_hi + sigma_hi * nsigma_hi)] = 0.66;
	if(nsigma_hi != 0) s[where(mhi < med_hi + sigma_hi * nsigma_hi)] = 1; 
	if(nsigma_hi != 0) l[where(mhi < med_hi + sigma_hi * nsigma_hi)] = 0.2; 
    }
    
    png_write_flipped(spec_map_name, hsl2rgb(h, s, l));
  } 
  %% otherwise use the default color schemes, setting the spectral index to colors between 0 (red) and 2/3 (blue)
  else
  {
    if(max(values.avg_lum)==0) {
      message("NO EMISSION???");
      spec_map[*,*]=0;
      avg_lum[*,*]=0;
    }
    else{
      if(max_spix<min_spix) message("MIN_SPIX SHOULD NOT BE LARGER THAN MAX_SPIX");
      spec_map = _min(_max(spec_map,min_spix), max_spix);
      if(min_spix*max_spix>=0){          %if min and max spix are the same sign, limit color scale to their range
	spec_map -= min_spix;
	spec_map /= (max_spix-min_spix)*3./2;
      }
      else{ % setting green to be at spix=0
	spec_map[where(spec_map>0)] /= max_spix*3;
	spec_map[where(spec_map<0)] /= -min_spix*3;
	spec_map += 1./3;
      }
      
      if(inv_color==1) spec_map = 2./3. - spec_map;  %inverting colors so red is a larger spectral index
    }
    %% where the spectral index is outside of the range, make luminosity 0
    %avg_lum[where(spec_map<min_spix-1e-5)]=0;
    %avg_lum[where(spec_map>max_spix+1e-5)]=0;
    png_write_flipped(spec_map_name, array_map(Integer_Type, &hsl2rgb, spec_map, 1, avg_lum));
  }
  pl.plot_png(spec_map_name);

  %% making color scale bar
  ifnot (qualifier_exists("no_scalebar")){
    variable scale_name = "scale.png"; while (stat_file(scale_name) != NULL) scale_name="X"+scale_name;
    
    variable cs = xfig_plot_new (width/25., height);
    cs.world (0, 1, min_spix, max_spix);
    cs.xaxis (;off);
    cs.y1axis (;off);
    cs.y2axis (;on,color= color);
    ifnot (qualifier_exists("no_labels")) {
      cs.y2label ("Spectral Index");}
    
    if (qualifier_exists("h_scalebar")){
      cs = xfig_plot_new (width, height/25.);
      cs.world (min_spix, max_spix,0, 1);
      cs.yaxis (;off);
      cs.x1axis (;off);
      cs.x2axis (;on,color= color);
      ifnot (qualifier_exists("no_labels")) {
	cs.x2label ("Spectral Index");
      }
    }

    variable N = 256;
    variable scale;
    variable cs_shape = (qualifier_exists("h_scalebar") ? [1,N] : [N,1]);
    
    if (qualifier_exists("color_scheme")) {
      scale = [min_spix: max_spix: #N];
      png_write_flipped(scale_name, png_gray_to_rgb(_reshape(scale,cs_shape),color_scheme));
    } 
    else {
      if(min_spix*max_spix>=0){          %if min and max spix are the same sign, limit color scale to their range
	scale = [0:2./3.:#N];
      }
      else {
	variable N_neg = nint( abs(min_spix/(max_spix-min_spix)) *N);
	scale = [[0:1:#N_neg],[1:2.:#(N+1-N_neg)][[1:]]] / 3.0;
      }
      if(inv_color==1) scale = 2./3 - scale;       %inverting scale
      png_write_flipped(scale_name, hsl2rgb( _reshape(scale,cs_shape), 1, 0.5));
    }
    cs.plot_png(scale_name);
  }
  
  %% adding source and date
  if (src_name != NULL){
    pl.xylabel(0.05,0.95,src_name,-0.5,0.5; world0, depth = 30, color=color);
  }
  if (obs_date != NULL){
    pl.xylabel(0.95,0.95,obs_date, 0.5,0.5; world0, depth = 30, color=color);
  }

  %% making beam
  variable beam_maj = @values.major;%fits_read_key (name8, "BMAJ")*(3.6e+6)/2.0;   % clean beam major axis in mas (semi major!)
  variable beam_min = @values.minor;%fits_read_key (name8, "BMIN")*(3.6e+6)/2.0;   % clean beam minor axis ins mas (semi minor!)
  variable beam_ang = @values.posang*PI/180.;%fits_read_key (name8, "BPA")/180*PI; % clean beam postion angle
  variable beam     = xfig_new_ellipse(beam_min,beam_maj);  % major axis to north, pos_angle = 0
  variable beam_clr = qualifier("beam_color",  (print_mode ? "gray" : "grey") );
  % apply custom units set with xy_unit
  beam_maj *= unit_fac;
  beam_min *= unit_fac;
  
  beam.set_pen_color(beam_clr);
  beam.rotate(vector(0,0,1),beam_ang);
  beam.scale( beam_ra_scale, beam_dec_scale , 0);
  beam.translate(vector(0.5-beam_maj*width/(ra_max-ra_min),0.5+beam_maj*height/(dec_max-dec_min),0));
  beam.area_fill=20;
  beam.fill_color=xfig_lookup_color(beam_clr);
  pl.add_object(beam);
  
  %% outputting file and cleaning up
  if (qualifier_exists("no_scalebar"))
    pl.render (outname);
  if (qualifier_exists("h_scalebar")) {xfig_new_vbox_compound (cs, pl, 0.4).render (outname);
      ()=remove(scale_name);}
  else {
    xfig_new_hbox_compound (pl, cs, 0.4).render (outname);
    ()=remove(scale_name);
  }
  ()=remove(spec_map_name);
}


#endif

#endif

%
% Code from src/plot/plot_struct_arrays.sl
%
define plot_struct_arrays()
%!%+
%\function{plot_struct_arrays}
%\synopsis{plots the fields of a structure agains each other}
%\usage{plot_struct_arrays(Struct_Type s, String_Type fieldnameX, String_Type fieldnameY);}
%\qualifiers{
%\qualifier{xoffset}{}
%\qualifier{yoffset}{}
%}
%\description
%    \code{plot(s.fieldnameX - xoffset,  s.fieldnameY - yoffset);}
%!%-
{
  variable s, fieldnameX, fieldnameY;
  switch(_NARGS)
  { case 3: (s, fieldnameX, fieldnameY)= (); }
  { help(_function_name()); return; }

  plot(get_struct_field(s, fieldnameX)-qualifier("xoffset", 0),
       get_struct_field(s, fieldnameY)-qualifier("yoffset", 0));
}

%
% Code from src/plot/plot_table_columns.sl
%
define plot_table_columns()
%!%+
%\function{plot_table_columns}
%\synopsis{plots columns of a table against each other}
%\usage{plot_table_columns(Struct_Type table);}
%\qualifiers{
%\qualifier{x}{array of columns to be used for the x-axis [default: all columns]}
%\qualifier{y}{array of columns to be used for the y-axis [default: all columns]}
%\qualifier{path}{path to save the postscript plots}
%\qualifier{multiplot}{produce one single multiplot (does not call open_plot)}
%}
%!%-
{
  variable table;
  switch(_NARGS)
  { case 1: table = (); }
  { help(_function_name()); return; }

  variable columns = get_struct_field_names(table);
  variable Xcolumn, Xcolumns = [qualifier("x", columns)];
  variable Ycolumn, Ycolumns = [qualifier("y", columns)];
  variable path = qualifier("path", ".");
  if(stat_file(path)==NULL)
    mkdir(path);
  if(qualifier_exists("multiplot"))
  {
    variable n = 0;
    foreach Xcolumn (Xcolumns)
      foreach Ycolumn (Ycolumns)
        if(Xcolumn != Ycolumn)
          n++;
    multiplot(ones(n));
  }
  foreach Xcolumn (Xcolumns)
    foreach Ycolumn (Ycolumns)
      if(Xcolumn != Ycolumn)
      {
	ifnot(qualifier_exists("multiplot"))
	  ()=open_plot(path + "/" + Xcolumn + "-" + Ycolumn + ".ps/ps");
	xlabel(Xcolumn);
	ylabel(Ycolumn);
        plot(get_struct_field(table, Xcolumn), get_struct_field(table, Ycolumn));
	ifnot(qualifier_exists("multiplot"))
	  close_plot;
      }
}

%
% Code from src/misc/radio/plot_vlbi_map.sl
%

% only activate if module gcontour could be loaded
#ifeval __get_reference("gcontour_compute")!=NULL

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DEFINITION OF WCS SYSTEM FOR AXIS SCALING %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
  private define mJy_wcs_func (map, cd)
  {
    variable mu = cd.mu, sigma = cd.sigma, nsigma = cd.nsigma, mx = cd.max, mn = cd.min;
    variable clrcut = cd.clrcut, neglog = cd.neglog, frac   = cd.frac;
    variable lim    = mu+nsigma*sigma;
    variable scl    = clrcut ? 1 : frac*log (mx/lim)/(lim-mn)/(1.-frac);
    variable shft   = log ((lim-mu)/sigma) - lim*scl ; 

    variable is_array = typeof (map)==Array_Type;
    variable x = 1.0*[@map]; % make sure that it is Double_Type, otherwise TypeMismatchError possible
    variable i1 = where(x >  lim);
    variable i2;
    if (neglog)
    {
      i2 = where(-lim <= x <= lim);
      variable i3 = where(x <  -lim);
      x[i2] = clrcut ? lim*scl : x[i2] * scl;
      x[i3] = clrcut ? lim*scl : -log( (-x[i3]+mu)/sigma ) + shft;
    }
    else
    {
      i2 = where(x <=  lim);
      x[i2] = clrcut ? lim*scl : x[i2] * scl;
    }
    x[i1] = log( (x[i1]-mu)/sigma ) - shft;
    return is_array ? x : x[0] ;
  }

  private define mJy_wcs_invfunc (map,cd)
  {
    variable mu = cd.mu, sigma = cd.sigma, nsigma = cd.nsigma, mx = cd.max, mn = cd.min;
    variable clrcut = cd.clrcut, neglog = cd.neglog, frac   = cd.frac;
    variable lim    = mu+nsigma*sigma;
    variable scl    = clrcut ? 1 : frac*log (mx/lim)/(lim-mn)/(1.-frac);
    variable shft   = log ((lim-mu)/sigma) - lim*scl ; 
    lim    = lim * scl;

    variable is_array = typeof (map)==Array_Type;
    variable x = 1.0*[@map];
    variable i1 = where(x >  lim);
    variable i2;
    if (neglog)
    {
      i2 = where(-lim <= x <= lim);
      variable i3 = where(x <  -lim);
      x[i2] = x[i2] / scl;
      x[i3] = -exp(-(x[i3]-shft))*sigma +mu;
    }
    else
    {
      i2 = where(x <=  lim);
      x[i2] = x[i2] / scl;
    }
    x[i1] = exp(x[i1]+shft)*sigma + mu;
    return is_array ? x : x[0] ;
  }

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PREVIOUSLY USED: atan for axis scaling:
%%%%%%
%  private define mJy_wcs_func (x,cd) {atan(x/20.);}
%  private define mJy_wcs_invfunc (x,cd) {tan(x)*20.;}
%  xfig_plot_add_transform ("mJy", &mJy_wcs_func, &mJy_wcs_invfunc, struct{mu=mu,sigma=sigma,nsigma=nsigma};);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define plot_vlbi_map()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{plot_vlbi_map}
%\synopsis{creates an image of the VLBI map with transparent background using a .fits file (provided by DIFMAP)}
%\usage{plot_vlbi_map(String_Type \code{fitsfile}, [String_Type \code{fitsfile (polflux)}], String_Type \code{filename})}
%\altusage{plot_vlbi_map(Struct_Type \code{img_struct}, String_Type \code{filename})}
%\qualifiers{
%\qualifier{color_scheme [="ds9b"]}{  select a color scheme for the image}
%\qualifier{dec_frac [={0.0,1.0}]}{   declination range of the image by
%                                selecting a fraction of the file's range}
%\qualifier{ra_frac [={0.0,1.0}]}{    right ascension range of the image by
%                                selecting a fraction of the file's range}
%\qualifier{dec_mas [={min,max}]}{    set declination range of the image directly in mas.
%                                overwrites \code{dec_frac}}
%\qualifier{ra_mas [={min,max}]}{     set right ascension range of the image directly in mas
%                                (analog to \code{dec_mas})}
%\qualifier{fit_noise [=1]}{          fit the noise of a map}                                
%\qualifier{n_sigma [=3.0]}{          start color scaling at \code{n_sigma*sigma}}
%\qualifier{cont_scl [=2.0]}{         set factor to change the separation between\n
%                                contour levels, the levels are placed at:\n
%                                \code{cont_scl^[0,1,...]*n_sigma*sigma}}
%\qualifier{cont_lvl [=[c1,c2,..]}{   set contour levels [Jy] manually, overwrites\n
%                                other contour parameters}
%\qualifier{cont_width [=2]}{         set width of contour lines}
%\qualifier{plot_cont [=1]}{          set to 0 in order to plot without contour lines}
%\qualifier{cont_depth [=2]}{         set contour-line depth}
%\qualifier{plot_vec [=0]}{           set to 1 in order to plot with vectors
%                                (requires second fits file with polarized flux density)}
%\qualifier{vec_density [=5]}{        density of EVPA vectors; possible values: [1..10]}
%\qualifier{vec_width [=1]}{          width of EVPA vectors}
%\qualifier{vec_color [="black"]}{    color of EVPA vectors}
%\qualifier{plot_clr_img [=1]}{       set to 0 in order to plot without colored image}
%\qualifier{plot_clr_key [=1]}{       set to 0 in order to plot without key for color scale}
%\qualifier{plot_scale_arrows [=1]}{  set to 1 in order to plot with arrows denoting the size scalesin the map}
%\qualifier{clrcut [=0]}{             set to 1 in order to start the color scale at the
%                                significant emission (i.e., no color scaling below)}
%\qualifier{clrmin [=min(image)]}{    set minimum value [Jy] for color scale (see png_gray_to_rgb)}
%\qualifier{clrmax [=max(image)]}{    set maximum value [Jy] for color scale (see png_gray_to_rgb)}
%\qualifier{clrmu [from fit_gauss_to_img_noise]}{set noise level of the image [Jy] for color scale
%                                (mean value of regions without significant emission)}
%\qualifier{clrsig [from fit_gauss_to_img_noise]}{set width of noise (1-sigma) [Jy] for color scale,
%                                for identical color scale (e.g., in order to compare
%                                different images) the following qualifiers have to be set:
%                                clrmin, clrmax, clrmu, clrsig}
%\qualifier{colmap_depth}{            set depth of png map}
%\qualifier{bkg_white}{               set white map below clrmin, if clrcut is set, clrmin equals
%                                n_sigma above the mean background.}
%\qualifier{quadratic}{               create a quadratic image, by default the scaling is such that
%                                scaling in RA and DEC is equal}
%\qualifier{plotsize [=14]}{          size of the plot (scaling of image relative to font size)}
%\qualifier{src_name [=default]}{     set the name of the source, by default the name\n
%                                is read from the .fits file
%                                set to NULL for not plotting a source name}
%\qualifier{obs_date [=default]}{     set the observation date, by default the observation\n
%                                date is read from the .fits file
%                                set to NULL for not plotting a observation date}
%\qualifier{axis_color [="gray"]}{    set axis color}
%\qualifier{source_color [="gray"]}{  set color of source name}
%\qualifier{date_color [="gray"]}{    set color of observation date}
%\qualifier{date_depth [=1]}{         set depth of observation date}
%\qualifier{source_depth [=1]}{       set depth of source name}
%\qualifier{cont_color [="gray"]}{    set color of contour lines}
%\qualifier{neg_cont_color [="gray"]}{set color of contour lines below noise level}
%\qualifier{date_xy [=[0.95,0.95]]}{  set world0 coordinates of date string}
%\qualifier{plot_beam [=1]}{          set to 1 in order to plot the beam}
%\qualifier{plot_components}{         set to plot model components on top of image}
%\qualifier{model_circs}{             set to plot model component circles on top of image}
%\qualifier{pos_comp_color [="seagreen"]}{set for color of model components with positive flux}
%\qualifier{neg_comp_color [="seagreen"]}{set for color of model components with negative flux}
%\qualifier{beam_color [="gray"]}{    set color of the beam}
%\qualifier{no_labels}{:	        plot no labels}
%\qualifier{colmap_label}{:	        set label of colormap}
%\qualifier{neglog [=0]}{             use log scale for negative flux values}
%\qualifier{frac [=0.1]}{:	        fraction to which the linear region below n_sigma*sigma
%                                is scaled in the color scale (if clrcut != 0)}
%\qualifier{funit [="mJy"]}{          unit of the color scale bar, can be changed from "mJy" to "Jy"}
%\qualifier{xyunit [="mas"]}{         unit of the x/y labels, can be changed between "mas", "arcsec", "arcmin" and "deg";
%                                       assumption: FITS header sets unit "mas"}
%\qualifier{return_xfig}{             set qualifier to return the xfig object(s) instead than
%                                rendering directly}
%}
%\description
%    This function creates an image of a jet using the .fits file provided
%    by DIFMAP.
%    Alternatively a structure with the required fields (as obtained
%    with the function \code{read_difmap_img}) can be given directly to
%    \code{plot_vlbi_map} instead of the name of a fits file.
%    The color scheme can be selected and contour lines are calculated.
%    The format of the output file depends on the suffix of the given
%    \code{filename}. Possible formats of the output file are PDF, EPS,
%    PNG, GIF, etc.
%
%    If plot_vlbi_map is called with three arguments, the first two of them beiing maps,
%    the code will determine the noise level of the second map and cut the first
%    map at the given sigma level. A typical application would be to plot the
%    distribution of electric vectors or the degree of polarization only where
%    the polarized flux is significant.
%\seealso{read_difmap_fits,fit_gauss_to_img_noise}
%!%-
{
  variable s,pol,filename;
  switch(_NARGS)
  { case 2: (s,filename) = (); }
  { case 3: (s,pol,filename) = (); }
  { help(_function_name()); return; }

  variable color_scheme = qualifier("color_scheme", "ds9b");
  variable dec_frac     = qualifier("dec_frac",{0.,1.});
  variable ra_frac      = qualifier("ra_frac",{0.,1.});
  variable fit_noise    = qualifier("fit_noise",1); 
  variable nsigma       = qualifier("n_sigma",3.0);  
  variable plot_cont    = qualifier("plot_cont",1);
  variable cont_depth   = qualifier("cont_depth",2);
  variable plot_vec     = qualifier("plot_vec",0);
  variable plot_pdeg     = qualifier("plot_pdeg",0);  
  variable plot_clr_img = qualifier("plot_clr_img",1);
  variable plot_clr_key = qualifier("plot_clr_key",1);
  variable plot_scale_arrows = qualifier("plot_scale_arrows",0);  
  variable clrcut       = qualifier("clrcut",0);
  variable bkg_white    = qualifier("bkg_white",0);  
  variable no_labels	= qualifier("no_labels");
  variable colmap_label	= qualifier("colmap_label",NULL);
  variable funit	= qualifier("funit","mJy");
  variable xyunit	= qualifier("xyunit","mas");  
  variable frac         = _max ( 1e-6, _min(1, qualifier("frac"  ,0.1)) );
  variable neglog       = qualifier("neglog",0);  
  variable plot_components = qualifier("plot_components");
  variable comps;

  % if filename instead of struct provided read the fitsfile:
  if (typeof(s) == String_Type)
  {
    if ((plot_cont or plot_clr_img) and fit_noise!=0) {
      comps = fits_read_table(s);
      s = read_difmap_fits (s ; fit_noise);
    }
    else {
      comps = fits_read_table(s);
      s = read_difmap_fits (s);
    }
  }

  variable unit_fac,unit_lab_x,unit_lab_y;
  switch (xyunit)
  { case "mas" : unit_fac=1.; }
  { case "arcsec" : unit_fac=1./1e3; }
  { case "arcmin" : unit_fac=1./1e3/60; }
  { case "deg" : unit_fac=1./1e3/3600; }  
  unit_lab_x=sprintf("Relative RA [%s]",xyunit);
  unit_lab_y=sprintf("Relative DEC [%s]",xyunit);  
  
  % correct ra/dec steps and beam smaj/smin for given unit
  s.ra_steps *= unit_fac;
  s.dec_steps *= unit_fac;
  s.beam_smaj *= unit_fac;
  s.beam_smin *= unit_fac;  
  
  if (plot_cont==1 and fit_noise==0)
    throw RunTimeError, sprintf("%s: provided structure includes no proper noise information for contour determination. Set plot_cont=0 instead!", _function_name());
  
  s.src_name   = qualifier("src_name", s.src_name );
  s.date_str   = qualifier("obs_date", s.date_str );

  variable map          = @(s.img); % in order not to modify the original image. 
  variable shape = array_shape(map);
  if ( length(shape) == 2 )
  {
    variable dec_len = shape[0] - 1;
    variable ra_len  = shape[1] - 1;
  }
  else
  {
    throw RunTimeError, sprintf("%s: multiple images for different FREQ or STOKES?\n\t-> modify the function %s",
				_function_name(), _function_name());
  }
  
  if (plot_cont or plot_clr_img)
  {
    variable map_min = min(map);
    variable map_max = max(map);

    if (s.mu == NULL or s.sigma == NULL)
      vmessage(sprintf("%s: provided structure includes no proper noise information", _function_name()));
%      throw RunTimeError, sprintf("%s: provided structure includes no proper noise information", _function_name());

    variable clrmin = qualifier("clrmin", clrcut ? s.mu+nsigma*s.sigma : map_min);
    variable clrmax  = qualifier("clrmax", map_max );
    variable clrmu   = qualifier("clrmu",  (s.mu!=NULL ? s.mu : 0 ));
    variable clrsig  = qualifier("clrsig", (s.sigma!=NULL ? s.sigma : 1e-3));

    if (qualifier_exists("bkg_white")) map[where(map<clrmin)]=-1000; % set white background level
    
    variable cd = struct{mu=clrmu,sigma=clrsig,nsigma=nsigma,max=clrmax,min=clrmin,clrcut=clrcut,
      frac = frac, neglog=neglog};
  }

  % cut maps on user demand --------
  variable idx_cut_x,idx_cut_y;
  ifnot (qualifier_exists("ra_mas") or qualifier_exists("dec_mas"))
  {
    idx_cut_x = [int(dec_len*dec_frac[0]):int(dec_len*dec_frac[1])];
    idx_cut_y = [int(ra_len*ra_frac[0]):int(ra_len*ra_frac[1])];
    map = map[idx_cut_x,idx_cut_y];
  }
  % if map has been cut, the index of the center pixel has to be corrected:
  variable ra_px_center  = s.ra_px_center  - int(ra_len*ra_frac[0]);   % right ascensin center pixel
  variable dec_px_center = s.dec_px_center - int(dec_len*dec_frac[0]); % declination center pixel
  
  variable axis_clr = qualifier("axis_color","gray"); 
  variable axis_dep = qualifier("axis_depth",100); 
  variable map_shape = array_shape(map);
  variable ra_min  = -( ra_px_center  - 0.5                 )*s.ra_steps; % adding half pixel (0.5) such that pixel centers correspond to integers for WCS conversion
  variable ra_max  =  ( map_shape[1]  - (ra_px_center-0.5)  )*s.ra_steps;
  variable dec_min = -( dec_px_center - 0.5                 )*s.dec_steps;
  variable dec_max =  ( map_shape[0]  - (dec_px_center-0.5) )*s.dec_steps;

  if (qualifier_exists("ra_mas") or qualifier_exists("dec_mas"))
  {
    variable ra_mas = qualifier("ra_mas",  {ra_min,ra_max});
    if (ra_mas[0] < ra_mas[1]) { (ra_mas[0],ra_mas[1]) = (ra_mas[1],ra_mas[0]);}
    ra_mas  = { _min( ra_mas[0],ra_min) ,  _max(  ra_mas[1],ra_max) };
    variable dec_mas = qualifier("dec_mas", {dec_min,dec_max});
    if (dec_mas[0] > dec_mas[1]) { (dec_mas[0],dec_mas[1]) = (dec_mas[1],dec_mas[0]);}
    dec_mas = { _max(dec_mas[0],dec_min) , _min(dec_mas[1],dec_max) };
    idx_cut_x = [int(dec_mas[0]/s.dec_steps + (dec_px_center-0.5) ) : int(dec_mas[1]/s.dec_steps + (dec_px_center-0.5) )-1];
    idx_cut_y = [int( ra_mas[0]/ s.ra_steps + ( ra_px_center-0.5) ) : int( ra_mas[1]/ s.ra_steps + ( ra_px_center-0.5) )-1];      
    map = map[idx_cut_x,idx_cut_y];
    ra_px_center  -= int( ra_mas[0]/s.ra_steps  + (ra_px_center-0.5) );    
    dec_px_center -= int(dec_mas[0]/s.dec_steps + (dec_px_center-0.5) );
    ra_min  =  ra_mas[0];  ra_max = ra_mas[1];
    dec_min = dec_mas[0]; dec_max = dec_mas[1];
  }

  if (plot_cont)
  {
    variable cl    = qualifier("cont_scl", 2.);
    variable n_pos = int(log( (map_max - s.mu) / (nsigma*s.sigma)) / log(cl));
    variable n_neg = int(log( (s.mu - map_min) / (nsigma*s.sigma)) / log(cl));
    variable lvls  = qualifier("cont_lvl", s.mu + [-(cl^[0:n_neg]),cl^[0:n_pos]]*nsigma*s.sigma );
    variable  gct  = gcontour_compute(map,lvls);  
  }

  variable image_size  = qualifier("plotsize",14);

  variable pwidth  = abs(ra_min  - ra_max);
  variable pheight = abs(dec_min - dec_max);
  if (qualifier_exists("quadratic")){pwidth = 1; pheight = 1;} % create quadratic image
  variable image_scale = _max(pwidth, pheight);
  pwidth  *= image_size/image_scale;
  pheight *= image_size/image_scale;

  variable w1 = xfig_plot_new (pwidth, pheight);
  w1.world( ra_min , ra_max, dec_min, dec_max ;color=axis_clr);
  w1.axis(; color =axis_clr, depth=axis_dep, ticlabels2=0);

  if (plot_scale_arrows) {
    variable forw_arr = xfig_create_arrow(; arrow_type=2, arrow_style=1);
    variable back_arr = xfig_create_arrow(; arrow_type=2, arrow_style=1);
    w1.add_object(xfig_new_polyline([-2,10], [0,0]; forward_arrow=forw_arr,backward_arrow=back_arr));
  }

  if (plot_clr_img)
  {
    map = mJy_wcs_func(map,cd);
    variable colmap = png_get_colormap(color_scheme);
    %colmap[0] = 16777215; % set base always to white
    png_add_colormap ("ds9b",colmap);

    % do w1.plot_png separately to account for alpha channel: 
    variable rgbmap = png_gray_to_rgb (map, colmap;
				       gmin = mJy_wcs_func(clrmin,cd),
				       gmax = mJy_wcs_func(clrmax,cd));
    % make white background transparent for overlay images
    variable alpha = @rgbmap;    
    alpha[where(rgbmap==0xFFFFFF)]=0;
    alpha[where(rgbmap!=0xFFFFFF)]=255;
    variable r=(rgbmap & 0x00FF0000) shr 16;
    variable g=(rgbmap & 0x0000FF00) shr 8;
    variable b=(rgbmap & 0x000000FF) shr 0;

    variable random = rand_int(1,1e3,1)[0];
    png_write_flipped (sprintf("%spng_alpha_%u.png",xfig_get_tmp_dir,random), (alpha<<24)+(r<<16)+(g<<8)+b, 1);
    w1.plot_png(sprintf("%spng_alpha_%u.png",xfig_get_tmp_dir,random);
		cmap=color_scheme,
		gmin = mJy_wcs_func(clrmin,cd),
		gmax = mJy_wcs_func(clrmax,cd),
	        depth = qualifier("colmap_depth",100)); % image in the background, other components should have smaller depth
  }
  variable i,ii,k;
  if (plot_cont)
  {
    map_shape = array_shape(map);
    _for i(0,length(lvls)-1,1)
      _for k(0,length(gct[i].x_list)-1,1) 
      {
	if (lvls[i] > s.mu)
	  w1.plot( (gct[i].x_list[k]+0.5) / map_shape[1] * ( ra_max-  ra_min) +  ra_min,
		   (gct[i].y_list[k]+0.5) / map_shape[0] * (dec_max- dec_min) + dec_min;
		   color= qualifier("cont_color","gray"), depth=cont_depth,width=qualifier("cont_width",2));
        else
	  w1.plot( (gct[i].x_list[k]+0.5) / map_shape[1] * ( ra_max-  ra_min) +  ra_min,
		   (gct[i].y_list[k]+0.5) / map_shape[0] * (dec_max- dec_min) + dec_min;
		   color=qualifier("neg_cont_color","gray"),line = 1, depth=cont_depth,width=qualifier("cont_width",2));
	% added 0.5 in order to have pixel centers as integers in WCS conversion
      }
  }

  %%%% if two maps are loaded, use the second one and its noise level to cut the first map  %%%%
  % the first map (EVPA map) will not be plotted in color scale but translated to EVPA vectors

  
  if (plot_vec and _NARGS==2) throw RunTimeError, sprintf("%s: second difmap fits file for polarized flux required for plotting EVPA vectors.", _function_name()); 
  else if (plot_vec and _NARGS==3)
  {
    variable P = read_difmap_fits (pol ; fit_noise); % load fits map with polarized flux
    variable pmap = @(P.img);
    pmap = pmap[idx_cut_x,idx_cut_y]; % cut P-map to same shape as loaded map
    %vmessage("pmap: ra_size=%u   dec_size=%u",length(pmap[0,*]),length(pmap[*,0]));
    %vmessage("pmap: min=%.3f   max=%.3f   P.mu+nsigma*P.sigma=%.3f",min(pmap),max(pmap),P.mu+nsigma*P.sigma);
    map[where(pmap<P.mu+nsigma*P.sigma)]=0.0; % set map to zero where P-map is below noise-defined threshold, map size remains the same.

    %vmessage("chimap: %u  entries == 0",length(where(map == 0.0)));
    %vmessage("chimap: %u  entries != 0",length(where(map != 0.0)));    
    map = map * PI/180.0; % deg -> rad
    variable arrows = xfig_new_polyline_list ();
    variable ra_dim = length(map[0,*]);
    variable dec_dim = length(map[*,0]);
    %vmessage("ra_min=%.3f  ra_max=%.3f  dec_min=%.3f  dec_max=%.3f",ra_min,ra_max,dec_min,dec_max);
    variable ra_pos = Double_Type[0];
    variable dec_pos = Double_Type[0];
    variable ra_dist = Double_Type[0];
    variable dec_dist = Double_Type[0];        
    _for i (0,dec_dim-1,1) {
      _for ii (0,ra_dim-1,1) {
	if (map[i,ii] != 0.0) {	  
	  dec_pos = [dec_pos,dec_min+(i+0.5)*s.dec_steps];
	  ra_pos = [ra_pos,ra_min+(ii+0.5)*s.ra_steps];
	  variable len_vec = min([abs(ra_min-ra_max),abs(dec_min-dec_max)])*0.2*(pmap[i,ii]/max(pmap));
	  ra_dist = [ra_dist,sin(map[i,ii])*len_vec/2.];
	  dec_dist = [dec_dist,cos(map[i,ii])*len_vec/2.];
	  %vmessage("sin(%.3f)=%.3f   cos(%.3f)=%.3f",map[i,ii]*180./PI,sin(map[i,ii]),map[i,ii]*180./PI,cos(map[i,ii]));
	}
      }
    }
    if ( 1<= qualifier("vec_density",5) <= 10 ) variable idx = int([0:length(ra_pos)-1:#int(length(ra_pos)/(11.-qualifier("vec_density",5)))]);
    else throw RunTimeError, sprintf("%s: qualifier \"vec_density\" requires values within [1,10]", _function_name());
    ra_pos=ra_pos[idx];
    dec_pos=dec_pos[idx];
    ra_dist=ra_dist[idx];
    dec_dist=dec_dist[idx];    
    _for i (0,length(ra_pos)-1,1) w1.plot([ra_pos[i]-ra_dist[i],ra_pos[i]+ra_dist[i]],[dec_pos[i]-dec_dist[i],dec_pos[i]+dec_dist[i]];line=0,width=qualifier("vec_width",1),color=qualifier("vec_color","black"),depth=0); 
  }

    
  ifnot (qualifier_exists("no_labels")){
  w1.xlabel(unit_lab_x);
  w1.ylabel(unit_lab_y);}
  
  if (s.src_name != NULL)
  {
    w1.xylabel(0.05,0.95,s.src_name,-0.5,0.5; world0, depth=qualifier("source_depth",1), color=qualifier("source_color","gray"));
  }
  if (s.date_str != NULL)
  {
    w1.xylabel(qualifier("date_xy",[0.95,0.95])[0],qualifier("date_xy",[0.95,0.95])[1],s.date_str, 0.5,0.5; world0, depth=qualifier("date_depth",1), color=qualifier("date_color","gray"));
  }

  if (qualifier("plot_beam",1) == 1) {
    variable beam = xfig_new_ellipse(s.beam_smin,s.beam_smaj);  % major axis to north, pos_angle = 0
    %  vmessage("minor: %f \tmajor: %f\tangle %f",s.beam_smin,s.beam_smaj,s.beam_pang);  % major axis to north, pos_angle = 0
    variable beam_clr = qualifier("beam_color","gray");
    beam.set_pen_color(beam_clr);
    beam.rotate(vector(0,0,1),s.beam_pang); % rotation for position angle in rad ("rotate" rotates counter clockwise),
    beam.scale( pwidth/abs(ra_max-ra_min) , pheight/abs(dec_max-dec_min) , 0);
    %  vmessage("RA:  %f-%f",ra_min,ra_max);
    %  vmessage("DEC: %f-%f",dec_min,dec_max);
    beam.translate(vector(0.5-s.beam_smaj*pwidth/(ra_max-ra_min),0.5+s.beam_smaj*pheight/(dec_max-dec_min),0));
    beam.area_fill=20;
    beam.fill_color=xfig_lookup_color(beam_clr);
    w1.add_object(beam);
  }
  
  if (qualifier_exists("plot_components")) {
    comps.deltax *= 3.6e6;
    comps.deltay *= 3.6e6;
    variable pos_comps = struct_filter(comps, where(comps.flux > 0); copy);
    variable neg_comps = struct_filter(comps, where(comps.flux < 0); copy);
    w1.plot(pos_comps.deltax, pos_comps.deltay; sym="+",width=2,color=qualifier("pos_comp_color","seagreen"),depth=1);
    w1.plot(neg_comps.deltax, neg_comps.deltay; sym="*",width=2,color=qualifier("neg_comp_color","seagreen"),depth=1);
    if (qualifier_exists("model_circs")) {
      _for k (0,length(pos_comps.deltax)-1,1) w1.plot(pos_comps.deltax[k], pos_comps.deltay[k];
						      sym="circle",size=3.*pos_comps.major_ax[k]*3.6e6,width=2,color=qualifier("pos_comp_color","seagreen"),depth=1);
      _for k (0,length(neg_comps.deltax)-1,1) w1.plot(neg_comps.deltax[k], neg_comps.deltay[k];
						      sym="circle",size=3.*pos_comps.major_ax[k]*3.6e6,width=2,color=qualifier("neg_comp_color","seagreen"),depth=1);
    }
  }
  

  if (plot_clr_key && plot_clr_img)
  {
    xfig_plot_add_transform ("mJy", &mJy_wcs_func, &mJy_wcs_invfunc, cd;);
    variable w2 = xfig_plot_new (pwidth/25., pheight);
    if (clrcut)
    {
      w2.world (0, 1, clrmin*(funit=="mJy" ? 1000 : 1), clrmax*(funit=="mJy" ? 1000 : 1) );
      w2.y1axis (;ticlabels=0, wcs="log", color = "white");
      w2.y2axis (;on,wcs="log",color="black");
    }
    else
    {
      variable minor=Double_Type[0];
      variable major=Double_Type[0];
      variable majbase, minbase;
      variable logmin = nint(floor(log10(clrmu+nsigma*clrsig)));
      variable logmax = nint(floor(log10(clrmax)));
      if (logmax-logmin < 4) { majbase = [1,2,5]; minbase = [3,4,6,7,8,9];}
      else  { majbase = [1]; minbase = [2:9];}
    
      variable j;
      if (logmax-logmin < 10) % arbitrary value to avoid giant loops, if logmin --> -inf
      { _for j(logmin, logmax,1)
	{
	  major = [major, (10^j)*majbase];
	  minor = [minor, (10^j)*minbase];
	}
      }
      if (neglog) {
	minor = [minor[where(minor != 0)],-minor[where(minor != 0)]];
	major = [0,major[where(clrmu+nsigma*clrsig <= major <= clrmax)],-major[where(clrmin <= -major <= clrmu-nsigma*clrsig)]];
      }
      else {
	minor = [minor, major[where(clrmu+nsigma*clrsig > major)], -minor, -major];
	major = [0,major[where(clrmu+nsigma*clrsig <= major <= clrmax)]];
      }
      variable tl = major;
      if(funit=="mJy") tl *= 1000;
      w2.world (0, 1, clrmin, clrmax);
      w2.y1axis (;      wcs="mJy", color = "white", major=major, minor=minor, ticlabels=0 );
      w2.y2axis (;  on, wcs="mJy", color = "black", major=major, minor=minor, ticlabels=tl);
    }
    w2.xaxis (;off);
    if (qualifier_exists("no_labels")==0 and colmap_label==NULL){ w2.y2label ("Total Intensity ["+(funit=="mJy" ? funit : "Jy")+"/beam]");}
    else if (qualifier_exists("no_labels")==0 and colmap_label != NULL) w2.y2label(colmap_label);
    
    variable scale_array = _reshape([mJy_wcs_func(clrmin, cd):
		     mJy_wcs_func(clrmax, cd):#256],[256,1]);
%    scale_array = mJy_wcs_invfunc (scale_array,cd);
    w2.plot_png (scale_array; cmap = color_scheme);
    if (qualifier_exists("return_xfig"))
      return w1,w2;
    else
      xfig_new_hbox_compound (w1, w2, 0.4).render(filename);
  }
  else
  {
    if (qualifier_exists("return_xfig"))
      return w1;
    else
      w1.render(filename);
  }
}


#endif

#endif

#endif

%
% Code from src/misc/radio/plot_vlbi_pol_map.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

% ISSUES: qualifiers given to plot_vlbi_map in struct{ } forces them, if this script is called multiple times,
% the last enforcement will be also valid for the upcoming call of plot_vlbi_map.
% The question, which qualifier should be free is currently under debate.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define plot_vlbi_pol_map ()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{plot_vlbi_pol_map}
%\synopsis{creates a polarimetric image of a VLBI map using according .fits files (provided by DIFMAP)}
%\usage{(Struct_Type xfig-map, Struct_Type xfig-colormap) = plot_vlbi_pol_map(String_Type \code{fitsfile flux-map-1},
%                  String_Type \code{filename flux-map-2}, String_Type \code{filename plot-product})}
%\altusage{(Struct_Type xfig-map, Struct_Type xfig-colormap, Struct_Type xfig-vectormap) =
%                  plot_vlbi_pol_map(String_Type \code{fitsfile flux-map-1}, String_Type \code{fitsfile EVPA-map},
%                  String_Type \code{filename flux-map-2}, String_Type \code{filename plot-product})}
%\qualifiers{
%\qualifier{n_sigma_pol}{[=5] start color scaling of polarized flux at \code{n_sigma*sigma}}
%\qualifier{n_sigma_vec}{[=5] start plotting of EVPA vectors at \code{n_sigma*sigma} in reference to the polarized flux}
%\qualifier{colmap_pol}{if set, plot most significant polarized flux (depending on n_sigma_pol),
%                                      recommended: color map "iceandfire"}
%\qualifier{colmap_back}{if set, plot polarized flux or total intensity flux, recommended: color map "ds9b"}
%\qualifier{pol_cont}{set to 1 in order to plot contours of flux-map-1}
%\qualifier{pol_cont_col}{[="white"] color of contours of flux-map-1}
%\qualifier{pol_cont_depth}{[=0] depth of contours of flux-map-1}
%\qualifier{pol_translate}{[=0.4] if plot_pol_colmap=0, the contours of polarized flux will be translated
%                                      by \code{pol_translate*range_in_declination}}
%\qualifier{render_object}{[=1] set to 1 in order to render the plot object}
%}
%\description
%   This function uses the existing maps flux-map-1 (polarized flux),
%   (the EVPA_map for the electric vector position angle) and
%   flux_map_2 (polarized flux or total intensity flux) to plot the polarized flux
%   as color-coded map (and the EVPA as vectors) - cutted at the most significant contours
%   depending on n_sigma_pol - on top of the total intensity contours or as contours
%   vertically translated to them.
%   One has the flexibility to plot the polarized flux distribution instead of the total
%   intensity flux color-coded depending on the map given for flux-map-2.
%   All qualifiers of \code{plot_vlbi_map} are forwarded to this function.
%\seealso{plot_vlbi_map}
%!%-
{
  variable pfits,chifits,ifits,plotpath;

  switch(_NARGS)
  {case 3: (pfits,ifits,plotpath) = (); }
  {case 4: (pfits,chifits,ifits,plotpath) = (); }

  variable def_pol_cmap = "iceandfire";
  variable def_I_cmap = "ds9b";  
  
  variable n_sigma_vec = qualifier("n_sigma_vec",5.);
  variable n_sigma_pol = qualifier("n_sigma_pol",5.);
  variable colmap_pol = qualifier("colmap_pol");
  variable colmap_pol_depth = qualifier("colmap_pol_depth",100);  
  variable colmap_back = qualifier("colmap_back");
  variable pol_cont = qualifier("pol_cont",0);  
  variable pol_cont_col = qualifier("pol_cont_col","white");
  variable pol_cont_depth = qualifier("pol_cont_depth",1);      
  variable pol_translate = qualifier("pol_translate",0.4); % percentage of dec-range    
  variable plot_components = qualifier("plot_components");
  variable render_object = qualifier("render_object",1);
  
  variable P,chi,xf_P,xf_PP,xf_I,xf_II,xf_vec;
  variable k;
  
  % clrcut qualifier sets minimum of map to 3sigma above the background;  color map always ranges from min(map) to max(map), here: min=black!
  if (_NARGS==4)
  {
    (xf_vec) = plot_vlbi_map(chifits,pfits,"I_map.pdf";; struct_combine (__qualifiers, struct {axis_color="white",axis_depth=1000,source_color="white",no_labels,
      return_xfig,plot_clr_img=0,plot_cont=0,plot_beam=0,plot_vec=1,n_sigma=n_sigma_vec,src_name=NULL,obs_date=NULL}));
  }
  else xf_vec = xfig_plot_new; % dummy

  if (qualifier_exists("colmap_back"))  
  {
    (xf_I,xf_II) = plot_vlbi_map(ifits,"I_map.pdf";; struct_combine (__qualifiers, struct {axis_color="black",
      plot_cont=1,return_xfig}));    
    (xf_P) = plot_vlbi_map(pfits,"P_map.pdf";; struct_combine (__qualifiers, struct {axis_color="white",axis_depth=1000,date_color="white",source_color="white",
      no_labels,plot_cont=pol_cont,cont_color=pol_cont_col,plot_beam=0,plot_clr_img=0,n_sigma=n_sigma_pol,src_name=NULL,obs_date=NULL,return_xfig}));
    xf_II.y2label("Total Intensity [mJy/beam]");
  }
  else if (qualifier_exists("colmap_pol"))
  {
    (xf_I) = plot_vlbi_map(ifits,"I_map.pdf";; struct_combine (__qualifiers, struct {axis_color="black",
      plot_cont=1,cont_depth=2,plot_clr_img=0,return_xfig}));
    (xf_P,xf_PP) = plot_vlbi_map(pfits,"P_map.pdf";; struct_combine (__qualifiers, struct {axis_color="white",axis_depth=1000,source_color="white",
      no_labels,plot_cont=pol_cont,cont_color=pol_cont_col,plot_beam=0,clrcut=1,n_sigma=n_sigma_pol,bkg_white,src_name=NULL,obs_date=NULL,return_xfig,
      colmap_depth=colmap_pol_depth,cont_depth=pol_cont_depth}));
    xf_PP.y2label("\begin{center}$\mathrm{Polarized~Intensity}$ \\ $\mathrm{[mJy/beam]}$\end{center}"R);
  }
  else
  {
    (xf_I) = plot_vlbi_map(ifits,"I_map.pdf";; struct_combine (__qualifiers, struct {axis_color="black",axis_depth=1000,
      return_xfig,plot_clr_img=0,plot_cont=1}));
    (xf_P) = plot_vlbi_map(pfits,"P_map.pdf";; struct_combine (__qualifiers, struct {axis_color="white",axis_depth=1000,source_color="white",
      no_labels,plot_cont=pol_cont,cont_color=pol_cont_col,plot_beam=0,n_sigma=n_sigma_pol,plot_clr_img=0,clrcut=1,src_name=NULL,obs_date=NULL,return_xfig}));
  }
  if (qualifier_exists("pol_translate"))
  {
    xf_P.translate(vector(0, -(qualifier("dec_mas")[1]-qualifier("dec_mas")[0])*pol_translate,0));
    xf_vec.translate(vector(0, -(qualifier("dec_mas")[1]-qualifier("dec_mas")[0])*pol_translate,0));
  }
  xf_P.axes(;off);
  xf_vec.axes(;off);
%  xf_P.set_depth(1);
%  xf_vec.set_depth(5);
  xf_I.add_object(xf_P); % total I contours and P-color map with color map in PP
  xf_I.add_object(xf_vec);

  
  
%  if (qualifier_exists("plot_components")) {
%    variable comps = fits_read_table(ifits);
%    comps.deltax *= 3.6e6;
%    comps.deltay *= 3.6e6;
%    variable pos_comps = struct_filter(comps, where(comps.flux > 0); copy);
%    variable neg_comps = struct_filter(comps, where(comps.flux < 0); copy);
%    xf_I.plot(pos_comps.deltax, pos_comps.deltay; sym="+",width=2,color=qualifier("pos_comp_color"),depth=1);
%    xf_I.plot(neg_comps.deltax, neg_comps.deltay; sym="*",width=2,color=qualifier("neg_comp_color"),depth=1);
%    if (qualifier_exists("model_circs")) {
%      _for k (0,length(pos_comps.deltax)-1,1) xf_I.plot(pos_comps.deltax[k], pos_comps.deltay[k];
%							sym="circle",size=3.*pos_comps.major_ax[k]*3.6e6,width=2,color=qualifier("pos_comp_color"),depth=1);
%      _for k (0,length(neg_comps.deltax)-1,1) xf_I.plot(neg_comps.deltax[k], neg_comps.deltay[k];
%							sym="circle",size=3.*pos_comps.major_ax[k]*3.6e6,width=2,color=qualifier("neg_comp_color"),depth=1);
%    }
%  }
    
  if (render_object==1) {
    if (qualifier_exists("colmap_pol") == 0 && qualifier_exists("colmap_back") == 0) xfig_new_hbox_compound(xf_I).render(plotpath);
    if (qualifier_exists("colmap_back")) xfig_new_hbox_compound(xf_I,xf_II).render(plotpath);
    if (qualifier_exists("colmap_pol")) xfig_new_hbox_compound(xf_I,xf_PP).render(plotpath);
    if (qualifier_exists("colmap_back") && qualifier_exists("colmap_pol"))
      throw RunTimeError, sprintf("%s: can't plot both full- and polarized intensity color maps", _function_name(), _function_name());
  }

  if (_NARGS==4 && qualifier_exists("colmap_back")) return (xf_I,xf_II,xf_vec);
  if (_NARGS==4 && qualifier_exists("colmap_pol")) return (xf_I,xf_PP,xf_vec);  
  if (_NARGS==3 && qualifier_exists("colmap_back")) return (xf_I,xf_II);
  if (_NARGS==3 && qualifier_exists("colmap_pol")) return (xf_I,xf_PP);  
}


#endif

%
% Code from src/plot/plot_with_err.sl
%
define plot_with_err()
%!%+
%\function{plot_with_err}
%\synopsis{plots data points with their errorbars}
%\usage{plot_with_err(x, [xErr,] y, yErr);
%\altusage{plot_with_err(Struct_Type s);}
%}
%\qualifiers{
%\qualifier{xerr}{change 3-argument-syntax to \code{plot_with_err(x, xErr, y);}}
%\qualifier{xminmax}{changes the meaning of \code{xErr} -- and \code{x}, if \code{xErr} is not a list}
%\qualifier{yminmax}{changes the meaning of \code{yErr} -- and \code{y}, if \code{yErr} is not a list}
%\qualifier{minmax}{equivalent to both \code{x}- and \code{yminmax}}
%\qualifier{i}{index-array of subset of data points to be plotted}
%\qualifier{set_xrange=frac}{ (default: 0.05): set the \code{xrange} from the lowest to highest x-value
%                  with additional padding (given as a fraction fo the x-range) on both sides}
%\qualifier{set_yrange=frac}{ (default: 0.05): set the \code{xrange} from the lowest to highest y-value
%                  with additional padding (given as a fraction of the y-range) on both sides}
%\qualifier{set_ranges}{equivalent to both \code{set_xrange} and \code{set_yrange}}
%\qualifier{overplot}{The data will be overplotted.}
%\qualifier{connect_points}{data points are also connected}
%\qualifier{histogram}{draw histogram lines, too}
%\qualifier{error_color}{draw error bars in a different color}
%}
%\description
%    In order to use asymetric errors for x and/or y,
%    the correspondig \code{Err} argument has to be a list \code{{ Err1, Err2 }}.
%    If one of the \code{minmax} qualifiers is used,
%    the corresponding \code{Err} list contains directly minimum and maximum values.
%
%    If one of the \code{minmax} qualifiers is used, but \code{Err} is not a list,
%    the value and \code{Err} arguments actually mean minimum and maximum values.
%    The actual value is infered to be the mean of minimum and maximum.
%\examples
%    % examples with symmetrical errorbars:\n
%     plot_with_err(x,       y, yErr);\n
%     plot_with_err(x, xErr, y, yErr);\n
%     plot_with_err(x, xErr, y,     ; xerr);\n
%    \n
%     % examples with unspecified x and/or y value:\n
%      plot_with_err(xMin, xMax, y   , yErr; xminmax);  % => x = (xMin+xMax)/2\n
%      plot_with_err(x[, xErr],  yMin, yMax; yminmax);  % => y = (yMin+yMax)/2\n
%      plot_with_err(xMin, xMax, yMin, yMax;  minmax);  % => [both inferences]\n
%      s = struct { bin_lo=x-xErr, bin_hi=x+xErr, value=y, err=yErr };\n
%      plot_with_err(s);
%    \n
%    % examples with asymmetrical errorbars:
%     plot_with_err(x, {xErr1, xErr2}, y,  {yErr1, yErr2});\n
%     plot_with_err(x, {xMin,  xMax }, y,  {yMin,  yMax}; minmax);\n
%    \n
%    % example combining the above features:\n
%     plot_with_err(xMin, xMax, y, {yMin, yMax}; minmax);  % => x = (xMin+xMax)/2\n
%\seealso{[o][h]plot_with_err, [o][h]plot}
%!%-
{
  variable plot_options = get_plot_options();  % will be overwritten after possibly setting {x/y}range

  variable s, x, xErr=NULL, y, yErr=NULL;
  switch(_NARGS)
  { case 1:
      s = ();
      x    = (s.bin_lo + s.bin_hi)/2.;
      xErr = (s.bin_hi - s.bin_lo)/2.;
      y    =  s.value;
      if(struct_field_exists(s, "err"))    yErr = s.err;  % <- standard
      else if(struct_field_exists(s, "error"))  yErr = s.error;  % <- convenient (?)
      if(plot_options.use_bin_density)
      { y /= (s.bin_hi-s.bin_lo);
        if(yErr!=NULL)  yErr /= (s.bin_hi-s.bin_lo);
      }
  }
  { case 2:
      vmessage("warning (%s): only two arguments specified, trying normal plot", _function_name());
      (x, y) = ();
      plot(x, y);
      return;
  }
  { case 3:
      if(qualifier_exists("xerr"))
        (x, xErr, y      ) = ();
      else
        (x,       y, yErr) = ();
  }
  { case 4: (x, xErr, y, yErr) = (); }
  { help(_function_name()); return; }

  variable xminmax = _NARGS>1 && (qualifier_exists("minmax") || qualifier_exists("xminmax"));
  variable yminmax = _NARGS>1 && (qualifier_exists("minmax") || qualifier_exists("yminmax"));
  variable xMin, xMax, yMin, yMax;
  if(xErr==NULL)
    (xMin, xMax) = (x, x);
  else
    if(typeof(xErr)!=List_Type)
      if(xminmax)
       (xMin, xMax, x) = (x, xErr, (x+xErr)/2.);
      else
       (xMin, xMax) = (x-xErr, x+xErr);
    else
      if(xminmax)
        (xMin, xMax) = (xErr[0], xErr[1]);
      else
        (xMin, xMax) = (x-xErr[0], x+xErr[1]);

  if(yErr==NULL)
    (yMin, yMax) = (y, y);
  else
    if(typeof(yErr)!=List_Type)
      if(yminmax)
        (yMin, yMax, y) = (y, yErr, (y+yErr)/2.);
      else
        (yMin, yMax) = (y-yErr, y+yErr);
    else
      if(yminmax)
        (yMin, yMax) = (yErr[0], yErr[1]);
      else
        (yMin, yMax) = (y-yErr[0], y+yErr[1]);

  % select data points
  variable i = qualifier("i");
  if(i!=NULL)
    (x,    xMin,    xMax,    y,    yMin,    yMax   ) =
    (x[i], xMin[i], xMax[i], y[i], yMin[i], yMax[i]);

  % set ranges
  variable pad = qualifier("set_ranges");
  variable xpad = qualifier("set_xrange", pad);
  if(qualifier_exists("set_xrange") || qualifier_exists("set_ranges"))
  {
    variable qx = (plot_options.logx ? struct { logpad } : struct { pad });
    set_struct_fields(qx, (xpad!=NULL ? xpad : 0.05));
    xrange(min_max([xMin, xMax];; qx));
  }
  variable ypad = qualifier("set_yrange", pad);
  if(qualifier_exists("set_yrange") || qualifier_exists("set_ranges"))
  {
    variable qy = (plot_options.logy ? struct { logpad } : struct { pad });
    set_struct_fields(qy, (ypad!=NULL ? ypad : 0.05));
    yrange(min_max([yMin, yMax];; qy));
  }

  plot_options = get_plot_options();  % only after possibly setting {x/y}range

  % draw points (has indeed to be before selection of visible data points!)
  variable hist = qualifier_exists("histogram");
  connect_points( qualifier_exists("connect_points") || hist);
  plot_bin_integral;
  variable col = plot_options.start_color;
  if(qualifier_exists("overplot"))
  { col = plot_options.color;
    if(hist)
      ohplot(xMin, xMax, y);
    else
      oplot(x, y);
  }
  else
    if(hist)
      hplot(xMin, xMax, y);
    else
      plot(x, y);
  connect_points(-1);

%  % draw histogram lines
%  if(qualifier_exists("histogram"))
%    color(col), ohplot(xMin, xMax, y);

  % select visible data points
  i = where(    xMin<=plot_options.xmax and xMax>=plot_options.xmin
            and yMin<=plot_options.ymax and yMax>=plot_options.ymin);
  (x,    xMin,    xMax,    y,    yMin,    yMax   ) =
  (x[i], xMin[i], xMax[i], y[i], yMin[i], yMax[i]);

  % draw error bars
  variable n_minus_1 = length(x)-1;
  variable err_col = qualifier("error_color", col);
  ifnot(xErr==NULL || hist)
    _for i (0, n_minus_1, 1)
      color(err_col), oplot([xMin[i], xMax[i]], y[[i,i]]);
  if(yErr!=NULL)
    _for i (0, n_minus_1, 1)
      color(err_col), oplot(x[[i,i]], [yMin[i], yMax[i]]);

  % restore previous options
  ()=change_plot_options(; connect_points=plot_options.connect_points,
			   color=col+1,
			   start_color=plot_options.start_color,
			   use_bin_density=plot_options.use_bin_density
			);
}

%
% Code from src/plot/plot_with_y2axis.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define plot_with_y2axis()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{plot_with_y2axis}
%\synopsis{plot two functions with different y-axes}
%\usage{plot_with_y2axis(x1, y1, [ x2,] y2);}
%\qualifiers{
%\qualifier{color}{set the color of the 2nd y-axis [default: red]}
%\qualifier{yspace}{space above and below the last data point in units of the total y-range.}
%}
%\description
%    This function plots two different parameters y1 and y2, which depend
%    on the same paramter x. By default, the same x is taken for
%    both parameters.\n
%    \n
%    The y2-axis and the ranges are set automatically.\n
%    \n
%    Additionally the function returns the rescaled y2 values, which
%    now lie in the range of the y1 values.
%!%-
{
  variable X,Y,Xold,Yold;
  switch(_NARGS)
  { case 3:
      (X,Yold,Y) = ();
      Xold = X;
  }
  { case 4:
      (Xold,Yold,X,Y) = ();
  }
  { % else:
      help(_function_name()); return;
  }

  variable col = qualifier("color", 2);
  variable yspace = qualifier("yspace", 0);

  variable ymnOld = min(Yold);
  variable ymxOld = max(Yold);

  variable ymn = min(Y);
  variable ymx = max(Y);

  variable scale = 1.*(ymxOld - ymnOld)/ (ymx - ymn);
  variable Ynew = (Y - ymn)*scale + ymnOld;

  variable offset = (ymxOld-ymnOld)*yspace;

  xrange(min_max([X,Xold]));
  yrange (ymnOld-offset, ymxOld+offset);

  variable plo = change_plot_options(; yopt="BNST");
  plot(Xold, Yold);
  color(col);
  oplot(X, Ynew);
  y2axis(ymn,ymx;color=col);

  set_plot_options(plo);

  return Ynew;
}

%
% Code from src/misc/astronomy/plummer_interaction_kernel.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

define plummer_interaction_kernel()
%!%+
%\function{plummer_interaction_kernel}
%\synopsis{Evaluate equations of motion or potential energy from a number of Plummer spheres}
%\usage{Double_Type eom[3,n] = plummer_interaction_kernel(Double_Types t, m[6,n], Struct_Type ps; qualifiers)}
%\altusage{Double_Type energy[n] = plummer_interaction_kernel(Double_Types t, m[6,n], Struct_Type ps; qualifiers)}
%\qualifiers{
%\qualifier{coords}{[\code{="cyl"}] Use cylindrical ("cyl") or cartesian ("cart") coordinates.}
%\qualifier{eomecd}{[\code{="eom"}] Return equations of motion ("eom") or potential energy ("energy").}
%}
%\description
%    This function computes the equations of motion or potential energy of n test particles
%    at time 't' caused by the interaction with a number of moving Plummer spheres. Depending
%    on whether cylindrical coordinates (r,phi,z) and their canonical momenta (vr,Lz,vz) or
%    cartesian coordinates (x,y,z,vx,vy,vz; see qualifier 'coords') are used, the second
%    input parameter 'm' is a [6,n]-matrix with
%    (qualifier("coords")=="cyl")                       or (qualifier("coords")=="cart")
%       m[0,*] = r;                                        m[0,*] = x;
%       m[1,*] = phi;                                      m[1,*] = y;
%       m[2,*] = z;                                        m[2,*] = z;
%       m[3,*] = vr;                                       m[3,*] = vx;
%       m[4,*] = Lz;                                       m[4,*] = vy;
%       m[5,*] = vz;                                       m[5,*] = vz;
%    If the qualifier 'eomecd' is set to "eom", the function returns a [3,n]-matrix 'delta' with
%       delta[0,*] = -d/dr Phi(r,phi,z);                   delta[0,*] = -d/dx Phi(x,y,z);
%       delta[1,*] = -d/dphi Phi(r,phi,z);                 delta[1,*] = -d/dy Phi(x,y,z);
%       delta[2,*] = -d/dz Phi(r,phi,z);                   delta[2,*] = -d/dz Phi(x,y,z);
%    whereby Phi is the sum over all Plummer potentials.
%    If the qualifier 'eomecd' is set to "energy", the function returns an array of length n
%    storing the potential energy for each orbit:
%       E(r,phi,z) = Double_Type[n] = Phi(r,phi,z)
%    or
%       E(x,y,z) = Double_Type[n] = Phi(x,y,z)
%
%    For each Plummer sphere, the third input parameter 'ps', which is a structure, contains
%    a field which is again a structure with fields 't', 'x', 'y', and 'z' (all are arrays of
%    the same length) that list the time-dependent positions of the sphere. Additionally, the
%    shape of the respective sphere (see notes below) is given by the two fields 'psa' and
%    'psb' (both scalars).
%\notes
%    The potential of a Plummer sphere at distance r is
%       Phi(r) = -psa*(psb+r^2)^(-1/2)
%    The resulting acceleration in radial direction is
%       d^2/dt^2 r = -d/dr Phi = -psa*(psb+r^2)^(-3/2)*r
%
%    Cubic spline interpolation is used to determine the positions of the Plummer spheres
%    at time 't' if the GSL-module is available. Otherwise, linear interpolation is applied.
%    Extrapolation is not allowed. Always make sure that the tabulated times '.t[*]' are in
%    monotonic increasing order if '.t[-1]' is positive or in monotonic decreasing order if
%    '.t[-1]' is negative.
%\example
%    ps = struct{ o0, o1 };
%    ps.o0 = struct{ t=[0,1,2], x=[-1,0,1], y=[0,1,2], z=[0,1,2], psa=1, psb=4 };
%    ps.o1 = struct{ t=[0,1,2], x=[1,0,-1], y=[0,1,2], z=[0,1,2], psa=1, psb=4 };
%    m = Double_Type[6,1];
%    m[0,0] = 0; m[1,0] = 2; m[2,0] = 0; m[3,0] = 0; m[4,0] = 0; m[5,0] = 0;
%    r = plummer_interaction_kernel(0,m,ps; coords="cart");
%    r = plummer_interaction_kernel(0,m,ps; eomecd="energy");
%\seealso{N_body_simulation_std_kernel, plummer_MW}
%!%-
{
  if(_NARGS!=3)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': Wrong number of arguments.", _function_name());
  }
  variable t, m, ps;
  (t, m, ps) = ();
  if(_is_struct_type(ps)!=1)
    throw UsageError, sprintf("Usage error in '%s': Third input parameter has to be Struct_Type.", _function_name());
  variable eomecd = qualifier("eomecd", "eom");
  variable coords = qualifier("coords", "cyl");
  %
  variable cols = get_struct_field_names(ps);
  variable N = length(m[0,*]);
  variable r;
  if(eomecd == "eom")
    r = Double_Type[3,N]; % equations of motion
  else if(eomecd == "energy")
    r = Double_Type[N]; % energy
  else
    throw UsageError, sprintf("Usage error in '%s': Invalid value for qualifier 'eomecd'.", _function_name());
  %
  variable intpolfunc = &interpol;
#ifeval __get_reference("gsl->interp_cspline")!=NULL; % use cubic spline interpolation if GSL-module is available, otherwise linear interpolation
    intpolfunc = &gsl->interp_cspline;
#endif
  %
  variable col;
  foreach col (cols) % loop over all Plummer spheres
  {
    % --------------------------------------------------------
    % determine the position of the Plummer spheres at time t:
    variable temp1 = t;
    variable temp2 = @((get_struct_field(ps, col)).t);
    variable s = sign(temp2[-1]);
    if(s<0) % function 'intpolfunc' expects old_x to be in monotonic increasing order
    {
      temp1 *= s;
      temp2 *= s;
    }
    ifnot( temp2[0] <= temp1 <= temp2[-1] )
      throw UsageError, sprintf("Usage error in '%s': Orbit of Plummer sphere '%s' does not cover requested time t=%g.", _function_name(), col, t);
    variable x_ps = @intpolfunc(temp1, temp2, (get_struct_field(ps, col)).x);
    variable y_ps = @intpolfunc(temp1, temp2, (get_struct_field(ps, col)).y);
    variable z_ps = @intpolfunc(temp1, temp2, (get_struct_field(ps, col)).z);
    % --------------------------------------------------------
    % --------------------------------------
    % compute equations of motion or energy:
    variable psa = (get_struct_field(ps, col)).psa;
    variable psb = (get_struct_field(ps, col)).psb;
    variable dx, dy;
    if(coords=="cart")
    {
      dx = x_ps - m[0,*];
      dy = y_ps - m[1,*];
    }
    else % coords=="cyl"
    {
      variable cos_phi = cos(m[1,*]);
      variable sin_phi = sin(m[1,*]);
      dx = x_ps - m[0,*]*cos_phi;
      dy = y_ps - m[0,*]*sin_phi;
    }
    variable dz = z_ps - m[2,*];
    variable d2 = dx^2 + dy^2 + dz^2;
    if(eomecd == "eom")
    {
      variable interaction = psa*(psb+d2)^(-1.5);
      if(coords=="cart")
      {
	r[0,*] += interaction*dx;
	r[1,*] += interaction*dy;
	r[2,*] += interaction*dz;
      }
      else % coords=="cyl"
      {
	r[0,*] += interaction*(x_ps*cos_phi + y_ps*sin_phi - m[0,*]);
	r[1,*] += interaction*(y_ps*cos_phi - sin_phi*x_ps)*m[0,*];
	r[2,*] += interaction*dz;
      }
    }
    else if(eomecd == "energy")
    {
      r += -psa/sqrt(psb+d2); % potential energy of Plummer sphere is -a/sqrt(b+dx^2+dy^2+dz^2)
    }
    % --------------------------------------
  }
  return r;
}

#endif

%
% Code from src/plot/pmc.sl
%
define pmc(i)
{ plot_bin_integral;
  plot_model_counts(i);
}

%
% Code from src/plot/pmf.sl
%
define pmf(i)
{ plot_bin_density;
  plot_model_flux(i);
}

%
% Code from src/slang/images/png_read_curve.sl
%

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

%%%%%%%%%%%%%%%%%%%%%
define png_read_curve()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{png_read_curve}
%\synopsis{reads a curve from a x-y plot in an image}
%\usage{(X, Y) = png_read_curve(filename, xpix1, xval1, xpix2, xval2, ypix1, yval1, ypix2, yval2);}
%\description
%   \code{filename} is a png file containing a color-defined curve.
%   Its \code{x} and \code{y} coordinates are calibrated by \code{pix}/\code{val} pairs
%   of pixel coordinate and corresponding value assuming a linear scale.
%   The pixel coordinates start with (0, 0) in the upper left corner.
%   The return values are calibrated x-values and y-values
%   averaged over all pixels of the curve in the corresponding column.
%\qualifiers{
%\qualifier{color}{[\code{=0x000000} (black)]: considered color of the curve}
%\qualifier{wherenot}{consider any color except the one specified above}
%}
%!%-
{
  variable filename, xpix1, xval1, xpix2, xval2, ypix1, yval1, ypix2, yval2;
  switch(_NARGS)
  { case 9: (filename, xpix1, xval1, xpix2, xval2, ypix1, yval1, ypix2, yval2) = (); }
  { help(_function_name()); return; }

  variable verbose = qualifier_exists("verbose");
  variable col = qualifier("color", 0x000000);
  variable use_wherenot = qualifier_exists("wherenot");

  variable img = png_read(filename);
  variable x, X = {}, Y = {};
  _for x (0, array_shape(img)[1]-1, 1)
  {
    variable iy = use_wherenot ? wherenot(img[*,x]==col) : where(img[*,x]==col);
    variable len = length(iy);
    if(len)
    {
      list_append(X, xval1 + 1.*(      x -xpix1)/(xpix2-xpix1)*(xval2-xval1));
      list_append(Y, yval1 + 1.*(mean(iy)-ypix1)/(ypix2-ypix1)*(yval2-yval1));
    }
    if(verbose)
      vmessage("xpix=%d, #{ypix}=%d" + (len ? " => x=%f, y=%f" : ""),
	       x, (len ? (len, X[-1], Y[-1]) : 0) );
  }

  return (list_to_array(X), list_to_array(Y));
}


#endif

%
% Code from src/plot/point_distance2_from_line.sl
%
define point_distance2_from_line(xp,yp,x1,y1,x2,y2) 
%!%+
%\function{point_distance2_from_line}
%\synopsis{Calculate the squared distance of a point from a line}
%\usage{d2=point_distance2_from_line(xp,yp,x1,y1,x2,y2)}
%\description
% Calculate the  distance of point (xp,yp) from the line
% defined by the points P1=(x1,y1) and P2=(x2,y2), where
% P1!=P2. This condition is not tested for speed
% reasons and will result in a division by zero.
%!%-
{
    variable dy=y2-y1;
    variable dx=x2-x1;

    variable nom=dy*xp-dx*yp+x2*y1-y2*x1;
    
    return nom*nom/(dy*dy+dx*dx);
}


%
% Code from src/slang/geometry/point_in_polygon.sl
%
%
% routines to determine whether a point is located inside or outside
% a polygon
%

define crossing_number_polygon() 
%!%+
%\function{crossing_number_polygon}
%\usage{cn=crossing_number_polygon(P,V)}
%\synopsis{return the crossing number for a point P in a polygon}
%\description
% Return the crossing number for a point P in a (potentially
% complex polygon defined by the vertex points V).
% The polygon has to be closed, i.e. V.x[n]==V.x[0] and
% V.y[n]==V.y[0] where n is the number of polygon points.
%
% The crossing number is the number of times that a ray starting 
% from point P crosses the polygon.  The point is outside of the
% polygon if the crossing number is even, and inside if the
% crossing number is odd (even-odd rule).
%
% The point P is a struct{x,y}, while the polygon V is 
% defined by its vertex points, which are stored in a
% struct{x[],y[]}
% where the arrays are the x- and y-coordinates.
%
% See the URL below for more explanations.
%
% Based on code by Dan Sunday,
% http://geomalgorithms.com/a03-_inclusion.html
% and patterned after Randolph Franklin (2000),
% http://www.ecse.rpi.edu/Homepages/wrf/research/geom/pnpoly.html
%
%\seealso{winding_number_polygon,point_in_polygon,simplify_polygon}
%!%-
{
    if (_NARGS != 2) {
	help(_function_name());
    }

    variable P,V;
    (P,V)=();

    % number of vertex points
    variable n=length(V.x);
    
    if (n<1) {
	throw UsageError,sprintf("%s: Polygon must have at least 2 vertex points.\n",_function_name());
    }

    if (V.x[0]!=V.x[n-1] or V.y[0]!=V.y[n-1])  {
	throw UsageError,sprintf("%s: Polygon is not closed.\n",_function_name());
    }
    
    if (length(V.x) != length(V.y) ) {
	throw UsageError,sprintf("%s: Vertex arrays x and y must be of same length.\n",_function_name());
    }

    variable cn=0; % crossing number counter
    variable i;
    % loop over all edges of the polygon
    _for i(0, n-2, 1) {
	% upward or downward crossing?
	if (((V.y[i]<=P.y) and (V.y[i+1]>P.y)) or 
	((V.y[i]>P.y) and  (V.y[i+1]<=P.y))) {
	    % edge-ray intersect coordinate
	    variable vt=(P.y-V.y[i])/(V.y[i+1]-V.y[i]);
	    % P.x < intersect x coordinate?
	    if (P.x < V.x[i]+vt*(V.x[i+1]-V.x[i]) ) {
		cn++;
	    }
	}
    }
    return cn;
}

define winding_number_polygon()
%!%+
%\function{winding_number_polygon}
%\usage{wn=winding_number_polygon(P,V)}
%\synopsis{return the winding number for a point P in a polygon}
%\description
% return the winding number for a point P in a (potentially
% complex polygon V)
%
% The winding number counts the number of times the polygon V winds
% around point P. The point is outside if the winding number is 0.
%
% The algorithm used here is as efficient as the determination
% of the crossing number.
%
% The point P is a struct{x,y}, while the polygon is a
% struct{x[],y[]} where the arrays are the x- and y-coordinates.
% The polygon has to be closed, i.e. V.x[n]==V.x[0] and
% V.y[n]==V.y[0] where n is the number of polygon points.
%
% See the URL below for more explanations.
%
% Based on code by Dan Sunday,
% http://geomalgorithms.com/a03-_inclusion.html
%
%\seealso{crossing_number_polygon,point_in_polygon,simplify_polygon}
%!%-
{
    if (_NARGS != 2) {
	help(_function_name());
    }

    variable P,V;
    (P,V)=();

    variable n=length(V.x);

    if (n<2) {
	throw UsageError,sprintf("%s: Polygon must have at least 2 vertex points.\n",_function_name());
    }

    if (V.x[0]!=V.x[n-1] or V.y[0]!=V.y[n-1])  {
	throw UsageError,sprintf("%s: Polygon is not closed.\n",_function_name());
    }
    
    if (length(V.x) != length(V.y) ) {
	throw UsageError,sprintf("%s: Vertex arrays x and y must be of same length.\n",_function_name());
    }
    
    variable wn=0;
    % loop over all edges
    variable i;
    _for i(0, n-2, 1) {
	if (V.y[i]<=P.y) {
	    % start y<=P.y
	    % ...upward crossing?
	    if (V.y[i+1] > P.y) {
		% Is point left of edge?
		if (point_is_left_of_line(V.x[i],V.y[i],V.x[i+1],V.y[i+1],P.x,P.y) > 0) {
		    wn++;
		}
	    }
	} else {
	    % start y>P.y
	    % ...downward crossing?
	    if (V.y[i+1]<=P.y) {
		if (point_is_left_of_line(V.x[i],V.y[i],V.x[i+1],V.y[i+1],P.x,P.y) < 0) {
		    wn--;
		}
	    }
	}
    }
    return wn;
}

define point_in_polygon()
%!%+
%\function{point_in_polygon}
%\usage{ret=point_in_polygon(p0,V);}
%\altusage{ret=point_in_polygon(x,y,Vx,Vy);}
%\synopsis{determine whether a point is in a polygon}
%\qualifiers{
%\qualifier{evenodd}{use the crossing number method}
%\qualifier{crossing}{use the crossing number method}
%\qualifier{winding}{use the winding number method (the default)}
%}
%\description
% The function returns 1 if the point p0 is located inside the
% polygon defined by the vertices V, and 0 if it is located
% outside of the polygon.
%
% The polygon has to be closed, i.e. V.x[n]==V.x[0] and
% V.y[n]==V.y[0] where n is the number of polygon points.
%
% Either the winding number (default) or the even-odd-rule
% define what is meant by inside.
%
% The point is either defined by a struct{x,y} and the vertices
% by a struct{x[],y[]}, or the coordinates can be directly
% given in the respective arrays.
%
% See the URL below for more explanations.
%
% Based on code by Dan Sunday,
% http://geomalgorithms.com/a03-_inclusion.html
%
%\seealso{crossing_number_polygon,point_in_polygon,simplify_polygon}
%!%-
%
{
    variable P,V;
    if (_NARGS==2) {
	(P,V)=();
    } else {
	if (_NARGS==4) {
	    variable x,y,Vx,Vy;
	    (x,y,Vx,Vy)=();
	    P=struct{x=x,y=y};
	    V=struct{x=@Vx,y=@Vy};
	} else {
	    throw UsageError,sprintf("%s: Wrong number of arguments.\n",_function_name());
	}
    }
    if (qualifier_exists("evenodd") or qualifier_exists("crossing")) {
	variable cn=crossing_number_polygon(P,V);
	printf("CN: %i\n",cn);
	return ( (cn mod 2)==1 );
    }

    % winding number method
    variable wn=winding_number_polygon(P,V);
    return ( wn!=0);
}

%
% Code from src/slang/geometry/point_is_left_of_line.sl
%
define point_is_left_of_line()
%!%+
%\function{point_is_left_of_line}
%\usage{ret=point_is_left_of_line(p0,p1,p2);}
%\altusage{ret=point_is_left_of_line(p0x,p0y,p1x,p1y,p2x,p2y);}
%\synopsis{determines whether point p2 is left of a line through p0 and p1}
%\description
% This function tests if point p2 is to the left of an infinite line defined
% by points p0 and p1. The points are defined by structs p=struct{x,y},
% where x is the x-coordinate and y is the y-coordinate.
%
% The function returns
%   +1 if P2 is left of the line
%    0 if P2 is on the line
%   -1 -1 if P2 is right of the line
%
% Based on code by Dan Sunday, http://geomalgorithms.com/a03-_inclusion.html
%
%\seealso{crossing_number_polygon,winding_number_polygon,point_in_polygon,simplify_polygon}
%!%-
{
    variable p0,p1,p2;
    variable cross;
    if (_NARGS==6) {
	variable p0x,p0y,p1x,p1y,p2x,p2y;
	(p0x,p0y,p1x,p1y,p2x,p2y)=();
	cross= (p1x-p0x)*(p2y-p0y) - (p2x-p0x)*(p1y-p0y);
    } else {
	if (_NARGS==3) {
	    (p0,p1,p2)=();
	    cross= (p1.x-p0.x)*(p2.y-p0.y) - (p2.x-p0.x)*(p1.y-p0.y);
	} else {
	    throw UsageError,sprintf("%s: Wrong number of arguments.\n",_function_name());
	}
    }
    return sign(cross);
}

%
% Code from src/slang/astronomy/polytropic_standard_model.sl
%
define polytropic_standard_model() %{{{
%!%+
%\function{polytropic_standard_model}
%\synopsis{Compute the structure of a polytropic standard model}
%\usage{Struct_Type s = polytropic_standard_model(Double_Type M, R, X, Y)}
%\description
%    Based on the stellar mass 'M' (in solar masses), stellar
%    radius 'R' (in solar radii), hydrogen mass fraction 'X',
%    and helium mass fraction 'Y', this function computes the
%    structure of a polytropic standard model. The output
%    structure 's' contains the fields "r" (radial coordinate
%    in solar radii), "rho" (density in g per cm^3), "P"
%    (pressure in dynes per cm^2), and "T" (temperature in K),
%    i.e., the structure of the polytrope.
%\notes
%    Polytropes are gaseous spheres in hydrostatic equilibrium
%    which obey a polytropic equation of state
%      P = K rho^((n+1)/n)
%    where 'P' is the pressure, 'rho' the density, 'K' the
%    polytropic constant, and 'n' the polytropic index. The
%    so-called polytropic standard model is a polytropic model
%    of index n=3. Its structure has a surprisingly good
%    resemblence to those of main sequence stars and is, thus,
%    used as a starting point to study the internal structure
%    of stars. For details, see, e.g., Chapter 2.4 in the book
%    "Principles of Stellar Evolution and Nucleosynthesis" by
%    D.D. Clayton.
%\example
%    s = polytropic_standard_model(1, 1, 0.73, 0.26);
%\seealso{solve_Lane_Emden_equation, solve_Eddington_quartic_equation}
%!%-
{
  if(_NARGS!=4)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable M, R, X, Y;
  (M, R, X, Y) = ();
  M *= 1.; R *= 1.; % cast to Double_Type
  variable mu = 2./(1.+3.*X+0.5*Y); % mean molecular weight, see Equation (2-15) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  variable beta = solve_Eddington_quartic_equation(M, mu);
  variable K = ((8.314510e7/mu)^4*3/7.5646e-15*(1.-beta)/beta^4)^(1./3.); % polytropic constant of a standard model, see Equation (2-302) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  variable Pc = 1.24e17*M^2/R^4; % central pressure of a polytropic standard model in dynes/cm^2, see Equation (2-315) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  variable rhoc = (Pc/K)^(3./4.); % central density of a polytropic standard model in g/cm^3, polytropic equation of state for n=3
  variable Lane_Emden_function = solve_Lane_Emden_equation(3);
  variable s = struct{ r, rho, P, T };
  s.r = Lane_Emden_function.xi/Lane_Emden_function.xi[-1]*R; % radius in solar radii
  s.rho = rhoc*Lane_Emden_function.theta^3; % density in g/cm^3, see Equation (2-291) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  s.P = K*s.rho^(4./3.); % pressure in dynes/cm^2, polytropic equation of state for n=3
  s.T = (8.314510e7/mu*3/7.5646e-15*(1.-beta)/beta*s.rho)^(1./3.); % temperature in K, see Equation (2-287) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  return s;
}%}}}

%
% Code from src/slang/math/pos_modulo.sl
%
define pos_modulo()
%!%+
%\function{pos_modulo}
%\synopsis{computes the modulo and ensures that it is positive}
%\usage{m = pos_modulo(a, b);}
%\description
%    \code{a} and \code{b} can either be arrays or scalars of a numerical type.
%
%    \code{m =  a mod b;  %} if m is positive, otherwise:\n
%    \code{m = (a mod b) + b;}
%!%-
{
  variable a, b;
  switch(_NARGS)
  { case 2: (a, b) = (); }
  { help(_function_name()); return; }

  variable m = a mod b;
  if(typeof(m)==Array_Type)
    m[where(m<0)] += b;
  else
    if(m<0)  m += b;
  return m;
}

%
% Code from src/slang/astronomy/position_angle.sl
%
define position_angle()
%!%+
%\function{position_angle}
%\usage{ sep=position_angle(ra1,dec1,ra2,dec2);}
%\synopsis{calculates the position angle of object1 with respect to object2}
%\qualifiers{
%\qualifier{deg}{if set, the input coordinates and output are in degrees (default: radian)}
%\qualifier{radian}{if set, the input coordinates and output are in radian (the default))}
%}
%\description
% This routine calculates the position angle on the sky of the point 
% with coordinates ra1/dec1 with respect to the object with coordinates
% ra2/dec2.
%
% Following the IAU conventions, the position angle is 0 if object1 is
% directly north of 2 and it is counted in the N-E-S-W-N direction, i.e.,
% it is pi/2 / 90deg if object 1 is East of object 2.
%
% Note that per IAU convention, "position angle" always refers to the
% relative position with respect to the equatorial system. This
% routine is agnostic to the type of coordinate system, i.e., the
% position angle can be computed in all coordinate systems, but the
% term "position angle" should be avoided for other coordinate systems.
%
% The position angle is not defined if object 2 is at one of the poles,
% it is 0 if object1 and object2 are at the same position.
%
% The position angle is returned in the units of the input coordinate system.
%
% This function is array safe for either ra1/dec1 or for ra2/dec2.
%
%\seealso{hms2deg,dms2deg,angle2string,greatcircle_distance}
%!%-
{
    variable ra1in,dec1in,ra2in,dec2in;
    variable ra1,dec1,ra2,dec2;
    switch(_NARGS)
    { case 4: (ra1in,dec1in,ra2in,dec2in)=(); }
    { help(_function_name()); return; }

    if (qualifier_exists("deg")) {
	ra1=ra1in*PI/180.;
	dec1=dec1in*PI/180.;
	ra2=ra2in*PI/180.;
	dec2=dec2in*PI/180.;
    } else {
	ra1=@ra1in;
	dec1=@dec1in;
	ra2=@ra2in;
	dec2=@dec2in;
    }
    
    variable s,c,td1,s2,c2;
    (s,c)=sincos(ra1-ra2);
    td1=tan(dec1);
    (s2,c2)=sincos(dec2);
    variable pa= atan2(s,c2*td1-s2*c);

    if (typeof(pa)==Array_Type) {
        variable ndx=where(pa<0.);
        if (length(ndx)>0) {
            pa[ndx]+=2*PI;
        }
    } else {
        if (pa<0.) {
            pa+=2*PI;
        }
    }
    
    if (qualifier_exists("deg")) {
	return pa*180./PI;
    }
    return pa;
}

%
% Code from src/misc/simulation/powerlaw_noise.sl
%
define powerlaw_noise()
%!%+
%\function{powerlaw_noise}
%\synopsis{simulates a light curve with a power law distributed power spectrum}
%\usage{Double_Type rate[] = powerlaw_noise(Integer_Type n);}
%\qualifiers{
%\qualifier{beta}{[= 1.5]: power law index of the power spectrum}
%\qualifier{mean}{[= 0]: mean rate of the simulated light curve}
%\qualifier{sigma}{[= 1]: standard deviation of the simulated light curve}
%}
%\description
%    \code{n} is the number of bins of the simulated lightcurve.
%    It should be a power of two best performance of the fast Fourier transform.
%\seealso{Timmer & Koenig (1995): "On generating power law noise", A&A 300, 707-710}
%!%-
{
  variable n;
  switch(_NARGS)
  { case 1: n = (); }
  { help(_function_name()); return; }

  variable n_2 = int(n/2);
  if(2*n_2 != n)
  { vmessage("error (%s): nt must be even", _function_name()); return; }

  variable betaph = qualifier("beta", 1.5);  % power law index of periodogram
  variable mean   = qualifier("mean",   0);  % mean count rate of the simulated lightcurve
  variable sigma  = qualifier("sigma",  1);  % standard deviation of the lightcurve to be simulated

  % calculating powerlaw distributed periodogram
  variable fac = (1.*[1:n_2]/n)^(-betaph/2.);  % 1/T, ..., 1/dt
  variable pos_real = grand(n_2)*fac;
  variable pos_imag = grand(n_2)*fac;
  pos_imag[n_2-1] = 0;
  variable real = [ 0., pos_real,  pos_real[[n_2-2:0:-1]] ];
  variable imag = [ 0., pos_imag, -pos_imag[[n_2-2:0:-1]] ];

  % simulate lc from its Fourier transform
  variable rate = Real(fft(real + imag*1i, 1));

  % normalize to desired mean count rate and variance
  variable m = moment(rate);
  return mean + (rate-m.ave)/m.sdev * sigma;
}

%
% Code from src/fitting/xyfit/fit-functions/powerlaw_xyfit.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define powerlaw_xyfit()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{powerlaw_xyfit}
%\synopsis{linear xy fit function to be used with xyfit_fun}
%\usage{xyfit_fun ("powerlaw");}
%\description
%    This function is not meant to be called directly!
%    
%    Calling \code{xyfit_fun ("powerlaw");} sets up a powerlaw fit
%    function for xy-data. It has the form \code{y = norm*x^{-index}}
%\seealso{xyfit_fun, define_xydata, plot_xyfit, linear_regression}
%!%-
{
  variable xref, yref, par;
  switch(_NARGS)
  { case 0: return ["norm", "index"]; } % fit parameters
  { case 3: (xref, yref, par)=(); }
  { return help(_function_name); }

  @yref = par[0] * (@xref)^(-par[1]);
}

define powerlaw_xyfit_default(i)
{
  switch(i)
  { case 0: return (1, 0, -1e5, 1e5); }
  { case 1: return (2, 0, -1, 4); }
}

%
% Code from src/slang/math/primefactors.sl
%
define primefactors(x)
%!%+
%\function{primefactors}
%\synopsis{factorizes an integer number into primes}
%\usage{Integer_Type factors[] = primefactors(Integer_Type x);}
%\description
%    The array of prime factors will be ordered:
%    \code{factor[i] <= factor[j]  %} for \code{i<j}.
%!%-
{
  variable factors = Integer_Type[0];
  variable p;
  foreach p ([2,3,5,7])
    while(x mod p==0)
    { factors = [factors, p];
      x /= p;
    }
  p = 11;
  while(x>1)
  { while(x mod p==0)
    { factors = [factors, p];
      x /= p;
    }
    do  p+=2;  while((p mod 3)*(p mod 5)*(p mod 7)==0);
    if(p*p>x)  p = x;
  }
  return factors;
}

%
% Code from src/misc/math/principal_components.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%%%%%
define principal_components()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{principal_components}
%\synopsis{performs a principal components analysis}
%\usage{Struct_Type PCA = principal_components(Struct_Type s);}
%\description
%    The normalized components (which are stored in \code{PCA.components.c}#i)
%    are calculated from the fields of the structure \code{s}
%    such that they have a mean of 0 and a variance of 1.
%    From them, the covariance matrix \code{PCA.cov_matrix} is calculated,
%    which is diagonalized (see \code{PCA.eigenvalues} and \code{PCA.eigenvectors}).
%    The principal components are stored in \code{PCA.components.pc}#i
%    in ascending order of their contribution to the total variance.
%\qualifiers{
%\qualifier{table}{[="tab"]: name of the structure \code{s}}
%}
%\seealso{cov_matrix}
%!%-
{
  variable s;
  switch(_NARGS)
  { case 1: s = (); }
  { help(_function_name()); return; }

  variable tablename = qualifier("table", "tab");

  variable fieldnames = get_struct_field_names(s);
  variable n_comp = length(fieldnames);
  variable n_data = length(get_struct_field(s, fieldnames[0]));
  variable X = Double_Type[n_data, n_comp];
  variable mom = Struct_Type[n_comp];

  variable PCA = struct { components, cov_matrix, eigenvectors, eigenvalues, frac_variance };
  PCA.components = @Struct_Type(["c" +array_map(String_Type, &string, [1:n_comp]),
                                 "pc"+array_map(String_Type, &string, [1:n_comp])]);
  variable comp;
  _for comp (0, n_comp-1, 1)
  { variable a = get_struct_field(s, fieldnames[comp]);
    mom[comp] = moment(a);
    mom[comp].sdev *= sqrt(mom[comp].num/(mom[comp].num+1.));  % think about that...
    variable c = (a-mom[comp].ave)/mom[comp].sdev;
    set_struct_field(PCA.components, "c"+string(comp+1), c);
    X[*, comp] = c;
  }
  PCA.cov_matrix = cov_matrix(X);
  (PCA.eigenvectors, PCA.eigenvalues) = gsl->eigen_symmv(PCA.cov_matrix);
  PCA.frac_variance = PCA.eigenvalues/sum(PCA.eigenvalues);

  variable i;
  _for i (0, n_comp-1, 1)
  { vmessage("principal component #%d:  [fract. variance: %4.1f%%]", i, 100*PCA.frac_variance[i]);
    variable fmt = "%g*%s.%s";
    variable pc = Double_Type[n_data];
    variable const = 0;
    _for comp (0, n_comp-1, 1)
    { ()=printf(fmt, PCA.eigenvectors[comp,i]/mom[comp].sdev, tablename, fieldnames[comp]);
      pc += PCA.eigenvectors[comp,i] * X[*, comp];
      const -= PCA.eigenvectors[comp,i] * mom[comp].ave/mom[comp].sdev;
      fmt = " %+g*%s.%s";
    }
    vmessage(" %+g\n", const);
    set_struct_field(PCA.components, "pc"+string(i+1), pc);
  }

  return PCA;
}


#endif

%
% Code from src/slang/arrays/print_array.sl
%
define print_array(arr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{print_array}
%\synopsis{Prints the entries of an array in one line.}
%\usage{print_array(Array_Type);}
%\description
%     Prints the entries of an array in one line instead of different lines.
%     Helpful for displaying dependent values of different arrays like a table.
%\example
%     UVOT-filter names and their wavelengths.
%
%     isis> filters=["UVW2","UVM2","UVW1","U","B","V"];
%     isis> wave=[2078.546871,2257.186625,2659.009185,3475.482080,4359.034280,5429.548507];
%     isis> print_array(filters); print_array(wave);
%     UVW2    UVM2    UVW1    U       B       V
%     2078.546871     2257.186625     2659.009185     3475.482080     4359.034280     5429.548507
%!%-   
{
    variable ty; switch(_typeof(arr))
     {                                                                                                                 
         case
          Integer_Type: ty = "%d";
     }                            
     {                            
      case Double_Type: if (min(arr) < 1e-4) ty ="%e";
          else ty = "%f";                             
     }                                                
     {                                                
         case                                         
          String_Type: ty = "%s";                     
     }                                                
   ; ()=array_map(Integer_Type, &printf,              
                  sprintf("%s\t", ty), arr); message("");
} %}}}                                                       

%
% Code from src/misc/math/print_statistics.sl
%
define print_statistics()
%!%+
%\function{print_statistics}
%\synopsis{shows statistical information on an array of numbers}
%\usage{print_statistics(Double_Type a[]);
%\altusage{print_statistics(Struct_Type s);}
%}
%\description
%    If the argument of \code{print_statistics} is an array \code{a},
%    its minimum, average, standard deviation and maximum 
%    are shown. If \code{a} contains irregular numbers (\code{nan} or \code{+/-inf}),
%    the same quantities are also shown for the regular numbers only.
%
%    If \code{print_statistics} is called with a structure argument,
%    the above mentioned task is performed on all array fields.
%\seealso{moment}
%!%-
{
  variable a;
  switch(_NARGS)
  { case 1: a = (); }
  { help(_function_name()); return; }

  variable m, i;
  if(typeof(a)==Struct_Type)
  {
    variable field = get_struct_field_names(a);
    foreach field (field)
    {
      ()=printf("%s=", field);  % field name
      field = get_struct_field(a, field);  % field value
      ()=printf("%S", _typeof(field));
      if(typeof(field)==Array_Type)
      {
	()=printf("[%s]", strjoin(array_map(String_Type, &string, array_shape(field)), ","));
	if(0<__is_numeric(field)<3)
	{
	  m = moment(field);
	  ()=printf(" {min=%g, av=%g, sdev=%g, max=%g}", m.min, m.ave, m.sdev, m.max);
	  i = wherenot(isnan(field) or isinf(field));
	  if(length(i)<length(field))
	  {
	    m = moment(field[i]);
	    ()=printf(" (%d regular numbers {min=%g, av=%g, sdev=%g, max=%g})", length(i), m.min, m.ave, m.sdev, m.max);
	  }
	}
      }
      ()=printf("\n");
    }
    return;
  }

  variable dims = array_shape(a);
  ()=printf("\n       %d", dims[0]);
  _for i (1, length(dims)-1, 1)  ()=printf(" x %d", dims[i]);
  if(length(dims)>1)  ()=printf(" = %d", length(a));
  m = moment(a);
  i = wherenot(isnan(a) or isinf(a));  % regular numbers
  if(length(i)<length(a))  % there are irregular numbers
  { variable mreg = moment(a[i]);
    ()=printf("  (%d regular numbers)\n", length(i));
    ()=printf("min  = %-20g (%g)\n", m.min,  mreg.min);
    ()=printf(" av  = %-20g (%g)\n", m.ave,  mreg.ave);
    ()=printf("sdev = %-20g (%g)\n", m.sdev, mreg.sdev);
    ()=printf("max  = %-20g (%g)\n", m.max,  mreg.max);
  }
  else  % there are no irregular numbers
  { ()=printf("\n");
    ()=printf("min  = %g\n", m.min);
    ()=printf(" av  = %g\n", m.ave);
    ()=printf("sdev = %g\n", m.sdev);
    ()=printf("max  = %g\n", m.max);
  }
}

%
% Code from src/slang/structs/tables/struct-arrays/print_struct.sl
%
%%%%%%%%%%%%%%%%%%%
define print_struct()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{print_struct}
%\synopsis{prints the fields of a structure as columns of a table}
%\usage{print_struct([File_Type F,] Struct_Type s);}
%\description
%    If \code{s} is a structure of arrays or lists,
%    these are displayed as columns of a table.
%    The output is written to \code{stdout} unless another \code{F}
%    is specified, which may either be a file pointer
%    or a string containing the filename.
%\qualifiers{
%\qualifier{i}{array of rows which are to be shown at all}
%\qualifier{mark}{array of rows which are to be marked}
%\qualifier{fields}{array of fieldnames (columns) which are to be shown}
%\qualifier{fmt}{format string(s) for the columns, see \code{help("sprintf");}}
%\qualifier{sep}{string separator between the columns (default = \code{"   "})}
%\qualifier{initial}{initial separator on a line (default = \code{""})}
%\qualifier{final}{final separator on a line (default = \code{""})}
%\qualifier{html}{set \code{sep}, \code{initial} and \code{final} such that an HTML table is produced}
%\qualifier{tex}{set \code{sep}, \code{initial} and \code{final} such that a TeX table is produced}
%\qualifier{nohead}{don't show head line with field names}
%}
%\examples
%    \code{print_struct(   s);                       %} The fields of \code{s} are printed.\n
%    \code{print_struct(F, s);                       %} The fields of \code{s} are printed to the file \code{F}.\n
%    \code{print_struct( , s; fmt="%.2f");           %} The fields are floats which are printed with 2 decimals.\n
%    \code{print_struct( , s; fmt="%.2f", sep=" ");  %} As before, but with smaller separation between the columns.
%\seealso{writecol, sprintf}
%!%-
{
  variable F=stdout, s;
  switch(_NARGS)
  { case 1:     s  = (); }
  { case 2: (F, s) = (); }
  { help(_function_name()); return; }

  variable fcloseF = 0;
  if(typeof(F)==String_Type)  (F, fcloseF) = (fopen(F, "w"), 1);

  % find fields to be shown
  variable fields = qualifier("fields");
  if(fields==NULL)
  { fields = String_Type[0];
    variable field;
    foreach field ( get_struct_field_names(s) )
      if(any([Array_Type, List_Type]==typeof(get_struct_field(s, field))))
	fields = [fields, field];
  }
  if(length(fields)==0)  % no array or list fields found
    fields = get_struct_field_names(s);  % print scalar fields instead
  variable nFields = length(fields);

  % get arrays
  variable arrays = Array_Type[nFields];
  variable col;
  _for col (0, nFields-1, 1)
  {
    variable val = get_struct_field(s, fields[col]);
    if(typeof(val)==List_Type)  val = list_to_array(val);
    arrays[col] = [ val ];
  }

  % formats
  variable fmt, fmt0 = qualifier("fmt", "%S");
  if(typeof(fmt0)==Array_Type && length(fmt0)==nFields)
    fmt = fmt0;
  else
  { fmt = String_Type[nFields];
    if(typeof(fmt0)!=Array_Type)
      fmt[*] = fmt0;
    else
    { vmessage("warning (%s): %d format(s) given for %d field(s)", _function_name(), length(fmt0), nFields);
      fmt = String_Type[nFields];
      if(length(fmt0)<nFields)
      { fmt[[0:length(fmt0)-1]] = fmt0;
	fmt[[length(fmt0):]] = "%S";
      }
      else
        fmt = fmt0[[0:nFields-1]];
    }
  }
  variable mark = qualifier("mark", Integer_Type[0]);
  variable indices = qualifier("i", [0:length(arrays[0])-1]);

  % format table values
  variable row, nRows = length(arrays[0][indices]);
  ifnot(qualifier_exists("nohead"))  nRows++;
  variable str = String_Type[nRows, 1+nFields];
  variable y = 0;
  ifnot(qualifier_exists("nohead"))
  { str[y,0] = "";
    _for col (1, nFields, 1)  str[y,col] = fields[col-1];
    y++;
  }
  foreach row (indices)
  { if( any(mark==row) )  str[y,0] = "*";  else  str[y,0] = "";
    _for col (1, nFields, 1)  str[y,col] = sprintf(fmt[col-1], arrays[col-1][row]);
    y++;
  }

  % fill up with spaces
  variable max_len = Integer_Type[1+nFields];
  _for col (0, nFields, 1)
  { max_len[col] = max( array_map(Integer_Type, &strlen, str[*,col]) );
    _for y (0, nRows-1, 1)
      str[y,col] = multiple_string(max_len[col]-strlen(str[y,col]), " ") + str[y,col];
  }

  % print table
  variable sep = qualifier("sep", "   ");
  variable cols = where(max_len>0);
  variable initial = qualifier("initial", "");
  variable final = qualifier("final", "");
  variable text_after_table = "";
  if(qualifier_exists("html"))
  { (initial, sep, final) = ("<TR><TD> ", " </TD><TD> ", " </TD></TR>");
    ()=fprintf(F, "<TABLE>\n");
    text_after_table = "</TABLE>\n";
  }
  else
    if(qualifier_exists("tex"))
    { (initial, sep, final) = ("", " & ", ` \\`);
      variable align = qualifier("tex");
      if(align==NULL)  align = "l";
      ()=fprintf(F, `\begin{tabular}{` + multiple_string(length(cols), align) + "}\n");
      text_after_table =  "\\end{tabular}\n";
    }
  _for y (0, nRows-1, 1)
    ()=fprintf(F, "%s%s%s\n", initial, strjoin(str[y, cols], sep), final);
  ()=fprintf(F, text_after_table);
  if(fcloseF)  ()=fclose(F);
}

%
% Code from src/plot/printplot.sl
%
define new_printplot()
%!%+
%\function{new_printplot}
%\synopsis{returns a new printplot-structure}
%\usage{Struct_Type new_printplot(Double_Type x0, x1, y0, y1);}
%\qualifiers{
%    \qualifier{W, H}{the width, i.e. the number of columns,
%                 and the height, i.e. the number of rows,
%                 of the plotting area without any axes or
%                 ticmarks (default: 60 and 10)}
%    \qualifier{[x,y]axis}{if set to zero do not draw the axis}
%    \qualifier{[x,y]tics}{if set to zero do not draw the ticmark}
%    \qualifier{[x,y]format}{sprintf format of the tics
%                 (default: "%f")}
%    \qualifier{[x,y]strlen}{maximum string length of the ticmarks
%                 (default: 4)}
%}
%\description
%    For printing a simple plot into the terminal a char
%    matrix is returned by this function, which can be
%    later printed by 'printplot_out'.
%    Per default, the matrix is 20x4 in size. Changing its
%    size can be done by the W- and H-qualifiers. The
%    arguments specify the x- and y-ranges of the plotting
%    area. The minimum and maximum values are the only
%    ticmarks which are added. Note that the char-matrix
%    is enlarged to fit the axes and ticmarks.
%\seealso{printplot_out, printplot, printhplot}
%!%-
{
  variable x0, y0, x1, y1;
  switch (_NARGS)
    { case 4: (x0,x1,y0,y1) = (); }
    { help(_function_name); return; }

  % qualifiers
  variable W = qualifier("W", 60);
  variable H = qualifier("H", 10);
  variable xaxis = qualifier("xaxis", 1);
  variable yaxis = qualifier("yaxis", 1);
  variable xtics = qualifier("xtics", 1);
  variable ytics = qualifier("ytics", 1);
  variable xform = qualifier("xformat", "%f");
  variable yform = qualifier("yformat", "%f");
  variable xlen = qualifier("xstrlen", 4);
  variable ylen = qualifier("ystrlen", 4);

  % offsets due to axes and ticmarks
  variable xoff = 0, yoff = 0;
  if (xtics) { yoff += 1; } % the x-tics fit in one row
  if (ytics) { xoff += ylen; } % the y-tics require more
  if (yaxis) { xoff++; }
  % define the char-matrix
  variable out = Char_Type[H+yoff,W+xoff];
  out[*,*] = ' '; % fill with spaces
  % draw axis
  if (xaxis) { out[H-1,[xoff:]] = '_'; }
  if (yaxis) { out[[:H-1],xoff-1] = '|'; }
  % draw tics
  variable tstr;
  if (xtics) {
    tstr = substr(sprintf(xform, x0), 1, xlen);
    out[H,[:strlen(tstr)-1]] = unpack(sprintf("c%d", strlen(tstr)), tstr);
    tstr = substr(sprintf(xform, x1), 1, xlen);
    out[H,[-strlen(tstr):]] = unpack(sprintf("c%d", strlen(tstr)), tstr);
  }
  if (ytics) {
    tstr = substr(sprintf(yform, y0), 1, ylen);
    out[H-1,[:strlen(tstr)-1]] = unpack(sprintf("c%d", strlen(tstr)), tstr);
    tstr = substr(sprintf(yform, y1), 1, ylen);
    out[0,[:strlen(tstr)-1]] = unpack(sprintf("c%d", strlen(tstr)), tstr);
  }

  return struct {
    matrix = out, W = W, H = H,
    world = struct {
      x0 = x0, x1 = x1, y0 = y0, y1 = y1,
      xoff = xoff, yoff = yoff
    }
  };
}

%%%%%%%%%%%%%%%
define printplot_out()
%%%%%%%%%%%%%%%
%!%+
%\function{printplot_out}
%\synopsis{print a char-matrix into the terminal}
%\usage{printplot_out(Char_Type[] matrix);}
%\seealso{new_printplot}
%!%-
{
  variable out, n;
  switch (_NARGS)
    { case 1: (out) = (); }
    { help(_function_name); return; }

  _for n (0, length(out[*,0])-1, 1) {
    vmessage(strjoin(array_map(String_Type, &sprintf, "%c", out[n,*])));
  }
}

%%%%%%%%%%%%%%%
define printplot()
%%%%%%%%%%%%%%%
%!%+
%\function{printplot}
%\synopsis{print a plot of xy-values into the terminal}
%\usage{printplot(Double_Type[] x, y[, Struct_Type printplot]);}
%\qualifiers{
%    \qualifier{sym}{the char used for a data point (default: 'x')}
%    \qualifier{get}{return the structure instead of printing}
%}
%\description
%    Prints a very simpel ASCII-version of plotting the
%    given xy-values into the terminal. The x- and y-ranges
%    are defined automatically by the input, but can be
%    specified via the optional printplot-structure (see
%    'new_printplot'). The only drawn ticmarks represent
%    these ranges.
%
%    All qualifiers are also passed to new_printplot.
%\example
%    % plot a sin
%    x = [0:2*PI:#100];
%    printplot(x, sin(x));
%
%    % overplot a cosin
%    printplot(x, .5*cos(x), printplot(x, sin(x); get); sym = '*');
%\seealso{new_printplot, printhplot}
%!%-
{
  variable x, y, plot = NULL;
  switch (_NARGS)
    { case 2: (x,y) = (); }
    { case 3: (x,y,plot) = (); }
    { help(_function_name); return; }

  % sanity checks
  if (length(x) != length(y)) {
    vmessage("error (%s): x and y must be of equal length!", _function_name);
    return;
  }

  % qualifiers
  variable sym = qualifier("sym", 'x');

  % create a new printplot structure if not already given
  if (plot == NULL) {
    plot = new_printplot(min(x), max(x), min(y), max(y);; __qualifiers);
  }

  % plot the data
  variable i, xd, yd;
  _for i (0, length(x)-1, 1) {
    if (plot.world.x0 < x[i] < plot.world.x1 && plot.world.y0 < y[i] < plot.world.y1) {
      xd = int(plot.world.xoff + plot.W*(x[i]-plot.world.x0)/(plot.world.x1 - plot.world.x0));
      yd = plot.H - int(plot.world.yoff + plot.H*(y[i]-plot.world.y0)/(plot.world.y1 - plot.world.y0));
      plot.matrix[yd, xd] = sym;
    }
  }

  % print or return
  if (qualifier_exists("get")) { return plot; }
  else { printplot_out(plot.matrix); }
}

%%%%%%%%%%%%%%%
define printhplot()
%%%%%%%%%%%%%%%
%!%+
%\function{printhplot}
%\synopsis{print a plot of a histogram into the terminal}
%\usage{printplot(Double_Type[] lo, hi, values[, Struct_Type printplot]);}
%\qualifiers{
%    \qualifier{sym}{the char used for the bar (default: 'o')}
%    \qualifier{get}{return the structure instead of printing}
%}
%\description
%    Prints a very simple ASCII-version of plotting the
%    given histogram into the terminal. The x- and y-ranges
%    are defined automatically by the input, but can be
%    specified via the optional printplot-structure (see
%    'new_printplot'). The only drawn ticmarks represent
%    these ranges.
%
%    All qualifiers are also passed to new_printplot
%\example
%    % plot gaussian distributed random numbers
%    (lo,hi) = linear_grid(-3, 3, 40);
%    printhplot(lo, hi, histogram(grand(10000), lo, hi); W=40);
%\seealso{new_printplot, printplot}
%!%-
{
  variable lo, hi, value, plot = NULL;
  switch (_NARGS)
    { case 3: (lo,hi,value) = (); }
    { case 4: (lo,hi,value,plot) = (); }
    { help(_function_name); return; }

  % sanity checks
  if (length(lo) != length(hi) || length(hi) != length(value)) {
    vmessage("error (%s): lo, hi, and values must be of equal length!", _function_name);
    return;
  }

  % qualifiers
  variable sym = qualifier("sym", 'o');

  % create a new printplot structure if not already given
  if (plot == NULL) {
    plot = new_printplot(lo[0], hi[-1], min(value), max(value);; __qualifiers);
  }

  % plot the histogram
  variable i, bd, hd;
  _for i (0, length(lo)-1, 1) {
    if (lo[i] >= plot.world.x0 && lo[i] <= plot.world.x1) {
      if (value[i] > plot.world.y0) {
        bd = int(round(plot.world.xoff + plot.W*(lo[i]-plot.world.x0)/(plot.world.x1 - plot.world.x0)));
        if (value[i] > plot.world.y1) { hd = 0; }
        else {
          hd = plot.H - int(plot.world.yoff + round(plot.H*(value[i]-plot.world.y0)/(plot.world.y1 - plot.world.y0)));
        }
	if (bd > plot.W+plot.world.xoff-1) { bd = plot.W+plot.world.xoff-1; }
        if (hd > plot.H-1) { hd = plot.H-1; }
        if (hd < 0) { hd = 0; }
        plot.matrix[[hd:plot.H-1], bd] = sym;
      }
    }
  }

  % print or return
  if (qualifier_exists("get")) { return plot; }
  else { printplot_out(plot.matrix); }
}

%
% Code from src/histogram/probmap.sl
%
private define w2m(obj, val, dim) {
  % dimension dependent boundaries
  variable mn, mx, bins, ddim;
  (mn,mx,bins,ddim) = (dim == 'x' ? (&obj.world.xMin, &obj.world.xMax, &obj.world.xbins, &obj.world.dx)
		                  : (&obj.world.yMin, &obj.world.yMax, &obj.world.ybins, &obj.world.dy));
  % loop over all input values and convert
  variable con = Integer_Type[length(val)], i;
  _for i (0, length(val)-1 ,1)
    con[i] = int(ceil((val[i] - @mn) / @ddim) - 1);

  return length(con) == 1 ? con[0] : con;
}

private define m2w(obj, val, dim) {
  % dimension dependent boundaries
  variable mn, mx, bins, ddim;
  (mn,mx,bins,ddim) = (dim == 'x' ? (&obj.world.xMin, &obj.world.xMax, &obj.world.xbins, &obj.world.dx)
		                  : (&obj.world.yMin, &obj.world.yMax, &obj.world.ybins, &obj.world.dy));
  % loop over all input values and convert
  variable con = Double_Type[length(val)], i;
  _for i (0, length(val)-1 ,1)
    con[i] = (val[i] + .5)*@ddim + @mn;

  return length(con) == 1 ? con[0] : con;
}


% 2d gauss
private define probmap_2dgauss(x,y,mux,muy,sigx,sigy) {
  return exp(-.5*( (sigx == 0 ? 0 : ((x-mux)/sigx)^2) + (sigy == 0 ? 0 : ((y-muy)/sigy)^2))) / (2*PI*(sigx == 0 ? 1 : sigx)*(sigy == 0 ? 1 : sigy));
}


% normalization functions
private define probmap_calcmap_maxNorm(raw, sum) { variable mx = 1.*max(sum);  return (mx == 0. ? 1. : 1./mx); }

% calculate the probability map from the given data
private define probmap_calcmap(obj) {
  % set world if not done yet
  ifnot (obj.world_defined()) {
    obj.xrange(min(array_flatten(array_struct_field(obj.data.raw, "x") - array_struct_field(obj.data.raw, "dx"))),
	       max(array_flatten(array_struct_field(obj.data.raw, "x") + array_struct_field(obj.data.raw, "dx"))));
    obj.yrange(min(array_flatten(array_struct_field(obj.data.raw, "y") - array_struct_field(obj.data.raw, "dy"))),
	       max(array_flatten(array_struct_field(obj.data.raw, "y") + array_struct_field(obj.data.raw, "dy"))));
  }

  % recalculate map if needed
  if (obj.data.needsUpdate) {
    variable i, d, sum = Double_Type[obj.world.ybins, obj.world.xbins];
    % loop over data
    _for i (0, length(obj.data.raw)-1, 1)
    {
      % calculate lower left corner of the error ellipse
      variable x0 = w2m(obj, obj.data.raw[i].x - 3*obj.data.raw[i].dx, 'x');
      variable y0 = w2m(obj, obj.data.raw[i].y - 3*obj.data.raw[i].dy, 'y');
      % calculate upper right corner where error is given
      variable x1 = @x0, y1 = @y0, n;
      n = where(obj.data.raw[i].dx != 0.);
      x1[n] = w2m(obj, obj.data.raw[i].x[n] + 3*obj.data.raw[i].dx[n], 'x');
      n = where(obj.data.raw[i].dy != 0.);
      y1[n] = w2m(obj, obj.data.raw[i].y[n] + 3*obj.data.raw[i].dy[n], 'y');
      % loop over coordinates
      _for d (0, length(x0)-1, 1) {
	% loop over error ellipse
	variable x,y;
	_for x (x0[d], x1[d], 1) _for y (y0[d], y1[d], 1) {
          if (0 <= x < obj.world.xbins && 0 <= y < obj.world.ybins) {
	    sum[y,x] += probmap_2dgauss(x, y, .5*(x0[d]+x1[d]), .5*(y0[d]+y1[d]), 1.*(x1[d]-x0[d])/6, 1.*(y1[d]-y0[d])/6);
	    % count point?
	    if (struct_field_exists(obj.data, "npoints"))
	      obj.data.npoints[y,x] += 1;
	  }
	}
      }
    }
    variable normFun = obj.norm_fun;
    obj.data.map = sum * @normFun(obj.data.raw, sum ;; struct_field_exists(obj.data, "npoints") ? struct { npoints = @(obj.data.npoints) } : NULL);
    
    obj.data.needsUpdate = 0;
  }
}


% retrieve the probability density map
private define probmap_getmap_method(nargs) {
  variable obj;
  switch(nargs)
    { case 1: (obj) = (); }

  probmap_calcmap(obj);
  return obj.data.map;
}


% retrieve one point within the probability map
private define probmap_getpoint_method(nargs) {
  variable obj,x,y;
  switch(nargs)
    { case 3: (obj,x,y) = (); }
    { message("Usage: probmap.getpoint(x, y);"); return; }

  probmap_calcmap(obj);

  ifnot (obj.world.xMin <= x <= obj.world.xMax && obj.world.yMin <= y <= obj.world.yMax) {
    message("error (probmap.getpoint): coordinates are out of world"); return; 
  }
  return obj.data.map[w2m(y,obj,'y'),w2m(x,obj,'x')];
}


% add the given data to the map (including error ellipses)
private define probmap_add_method(nargs) {
  variable obj, x, y, dx, dy;
  switch(nargs)
    { case 3: (obj,x,y) = (); dx = Integer_Type[length(x)]; dy = dx; }
    { case 4: (obj,x,y,dy) = (); dx = Integer_Type[length(x)]; }
    { case 5: (obj,x,y,dx,dy) = (); }
    { message("Usage: probmap.add(x, y[, dx, dy]);"); return; }

  obj.data.raw = [obj.data.raw, struct { x=x, y=y, dx=dx, dy=dy }];
  
  obj.data.needsUpdate = 1;

  return length(obj.data.raw);
}


% delete the data identified by its index from the map
private define probmap_delete_method(nargs) {
  variable obj, i;
  switch(nargs)
    { case 2: (obj,i) = (); }
    { message("Usage: probmap.delete(index);"); return; }

  ifnot (0 <= i < length(obj.data.raw)) { message("error (probmap.delete): given index is out of range"); return; }
  obj.data.raw = array_remove(obj.data.raw, i);
  
  obj.data.needsUpdate = 1;
}


% set the xrange
private define probmap_xrange_method(nargs) {
  variable obj, x0, x1;
  if (nargs == 3) {
    (obj,x0,x1) = ();
    if (obj.world.xMin == NULL && obj.world.xMax == NULL) {
      if (x0 < x1) {
        obj.world.xMin = 1.*x0;
        obj.world.xMax = 1.*x1;
	obj.world.dx = 1.*(x1-x0)/obj.world.xbins;
      } else { message("error (probmap.xrange): min has to smaller than max"); }
    } else { message("error (probmap.xrange): range has been defined already"); }
  } else { message("error (probmap.xrange): ranges not given"); }
}


% set the yrange
private define probmap_yrange_method(nargs) {
  variable obj, y0, y1;
  if (nargs == 3) {
    (obj,y0,y1) = ();
    if (obj.world.yMin == NULL && obj.world.yMax == NULL) {
      if (y0 < y1) {
        obj.world.yMin = 1.*y0;
        obj.world.yMax = 1.*y1;
        obj.world.dy = 1.*(y1-y0)/obj.world.ybins;
      } else { message("error (probmap.yrange): min has to smaller than max"); }
    } else { message("error (probmap.yrange): range has been defined already"); }
  } else { message("error (probmap.yrange): ranges not given"); }
}


% find the most probable path between two given points
private define probmap_calcpath_estimate(x0, y0, x1, y1) { return 0.; }
private define probmap_calcpath_method(nargs) {
  variable obj, x0, y0, x1, y1;
  if (nargs != 5) { message("error (probmap.calcpath): coordinates not given properly (x0,y0,x1,y1)"); return; }
  (obj, x0, y0, x1, y1) = ();
  % convert into map coordinates
  variable goX = w2m(obj,x0,'x'), goY = w2m(obj,y0,'y');
  if (goX < 0 || goX > obj.world.xbins-1 || goY < 0 || goY > obj.world.ybins-1)
    { message("error (probmap.calcpath): starting point out of range"); return; }
  variable desX = w2m(obj,x1,'x'), desY = w2m(obj,y1,'y');
  if (desX < 0 || desX > obj.world.xbins-1 || desY < 0 || desY > obj.world.ybins-1)
    { message("error (probmap.calcpath): ending point out of range"); return; }
  probmap_calcmap(obj);
  variable infind = where(obj.data.map == 0.);
  variable nmap = @obj.data.map;
  nmap[infind] = -_Inf;
  variable best = aStar(nmap, goY, goX, desY, desX;; struct_combine(__qualifiers, struct { estimate = &probmap_calcpath_estimate, max, meanCost }));
  return struct { x = m2w(obj,best.y,'x'), y = m2w(obj,best.x,'y'), prob = best.cost };
}


% check if the world has been defined yet
private define probmap_worlddefined_method(nargs) {
  variable obj;
  if (nargs == 1) {
    (obj) = ();
    return (obj.world.xMin != NULL
         && obj.world.xMax != NULL
         && obj.world.yMin != NULL
         && obj.world.yMax != NULL
    );
  }
}


% visible object methodes, which have to call
% internal methods to handle the stack correctly
% -> the calling object itself gets passed to
%    that functions
private define probmap_add() {
  variable args = __pop_args(_NARGS);
  return probmap_add_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define probmap_delete() {
  variable args = __pop_args(_NARGS);
  return probmap_delete_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define probmap_xrange() {
  variable args = __pop_args(_NARGS);
  return probmap_xrange_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define probmap_yrange() {
  variable args = __pop_args(_NARGS);
  return probmap_yrange_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define probmap_getmap() {
  variable args = __pop_args(_NARGS);
  return probmap_getmap_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define probmap_getpoint() {
  variable args = __pop_args(_NARGS);
  return probmap_getpoint_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define probmap_calcpath() {
  variable args = __pop_args(_NARGS);
  return probmap_calcpath_method(__push_args(args), _NARGS ;; __qualifiers);
}
private define probmap_worlddefined() {
  variable args = __pop_args(_NARGS);
  return probmap_worlddefined_method(__push_args(args), _NARGS ;; __qualifiers);
}


%%%%%%%%%%%%%%%%%%%%%
define probmap() {
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{probmap}
%\synopsis{histogram-like function for 2d-data including uncertainties}
%\usage{Struct_Type probmap([Double_Type x0, x1, y0, y1]);}
%\qualifiers{
%    \qualifier{xbins}{number of bins in x-direction (preferred over 'bins')}
%    \qualifier{ybins}{number of bins in y-direction (preferred over 'bins')}
%    \qualifier{bins}{number of bins in both directions (default: 201)}
%}
%\description
%    Using this object, a histogram of multiple xy-datasets
%    including uncertainties for both, x- and y-direction,
%    can be created. The uncertainties are handled as 2d-
%    gaussians with sigma_x and sigma_y equal to the given
%    errors.
%
%    Calling this function creates a structure with the
%    following functions among other fields:
%      add      - adds xy-data to the map and returns the dataset
%                 number of the added data
%                 Usage: probmap.add(Double_Type[] x, y[, dy]);
%                     or probmap.add(Double_Type[] x, y[, dx, dy]);
%      delete   - deletes data identified by its dataset number
%                 Usage: probmap.delete(Integer_Type number);
%      xrange   - sets the x-range
%                 Usage: probmap.xrange(Double_Type xMin, xMax);
%      yrange   - sets the y-range
%                 Usage: probmap.yrange(Double_Type yMin, yMax);
%      npoints  - count the number of xy-data added in each point
%                 of the map. If set the density map is available
%                 at .data.npoints, and it is passed as qualifier
%                 'npoints' to the norm_fun (see next qualifier).
%      norm_fun - reference to a function used for normalization.
%                 The parameters are an array of all datasets,
%                 given as a structure (x,y,dx,dy), and the
%                 histogram map resulting by adding all data.
%                 The function has to return either a single
%                 normalization factor or a map of factors.
%                 Default factor: 1./max(histogram)
%      getmap   - calculates the histogram map from the added data
%                 and returns it as 2d-array
%                 Usage: probmap.getmap();
%      getpoint - returns the value of the histogram map at a
%                 specific point
%                 Usage: probmap.getpoint(x,y);
%      calcpath - calculates the path with the highest averaged
%                 probability between to points using the
%                 A*-algorithm returning a structure with the
%                 best path (x,y)
%\example
%    % creata a new map and set x- and y-range
%    variable pm = probmap(0.1, 100, 0, 20);
%    % add two xy-datasets including uncertainties
%    pm.add(x1, y1, dx1, dy1);
%    pm.add(x2, y2, dx2, dy2);
%    % plot the resulting histogram
%    plot_image(pm.getmap());
%\seealso{histogram2d, histogram_gaussian_probability, aStar}
%!%-
  variable obj = struct {
    add           = &probmap_add,
    delete        = &probmap_delete,
    xrange        = &probmap_xrange,
    yrange        = &probmap_yrange,
    getmap        = &probmap_getmap,
    getpoint      = &probmap_getpoint,
    calcpath      = &probmap_calcpath,
    world_defined = &probmap_worlddefined,
    world         = struct { xMin, xMax, yMin, yMax, dx, dy, xbins, ybins },
    norm_fun      = &probmap_calcmap_maxNorm,
    data          = struct { raw = Struct_Type[0], map, needsUpdate = 0 }
  };
   
  % initialize map
  obj.world.xbins = qualifier("xbins", qualifier("bins", 201));
  obj.world.ybins = qualifier("ybins", qualifier("bins", 201));
  obj.data.map = Double_Type[obj.world.ybins, obj.world.xbins];

  % count point density?
  if (qualifier_exists("npoints"))
    obj.data = struct_combine(obj.data, struct { npoints = Integer_Type[obj.world.ybins, obj.world.xbins] });

  % set world if given
  if (_NARGS == 4) {
    variable x0, x1, y0, y1;
    (x0,x1,y0,y1) = ();
    obj.xrange(x0, x1);
    obj.yrange(y0, y1);
  };

  return obj;
}

%
% Code from src/slang/misc/progress_bar.sl
%
define draw_progress_bar(cur, goal)
%!%+
%\function{draw_progress_bar}
%\synopsis{Draw a progress bar}
%\usage{draw_progress_bar( position , maximum )}
%\qualifiers{
%\qualifier{tip}{String to draw tip of the arrow (default: ">")}
%\qualifier{bar}{String to draw bar of the arrow (default: "=")}
%\qualifier{front}{String to draw space in front of arrow (default: ".")}
%\qualifier{append}{String to append to each drawn progress bar (default: "")}
%\qualifier{columns}{Columns to use to write progress bar (default: Terminal width)}
%\qualifier{fmt}{Format to use for writing the percentage info in front of the bar.
%		   The printing function gets passed the percentage of the current
%		   state to the function as the second argument after the format
%		   (default: "%.1f%%", for example 01.3%)}
%}
%\description
%	Draw a progress bar across the prompt.
%	The width of the bar (arrow) is calculated from the fraction of
%	position/maximum. In case position is greater than maximum nothing happens.
%	Note that the function automatically determines the width of the current terminal.
%	This procedure takes some tens of milliseconds. The function can be sped by
%	setting the "columns" qualifier manually.
%!%-
{
  if ( cur > goal )
  {
    return ;
  }
  variable tip = qualifier("tip", ">");
  variable tiplen = strlen(tip);
  variable bar = qualifier("bar", "=");
  variable barlen = strlen(bar);
  variable front = qualifier("front", ".");
  variable frontlen = strlen(front);
  variable append = qualifier("append", "");
  variable fmt = qualifier("fmt", "%.1f%%");

  variable cols;

  if ( qualifier_exists("columns") )
  {
    cols = qualifier("columns");
  } else {
    % Determine columns of current terminal
    variable str;
    variable fptr = popen("stty size", "r");
    () = fgets(&str, fptr);
    () = fclose(fptr);
    () = sscanf(str, "%*d %d\n", &cols);
  }

  variable frac = double(cur)/double(goal);
  () = printf("\r");
  variable printed = printf(fmt+"[", frac * 100);
  variable printbar = int(frac * (cols - printed - 1)) - tiplen;
  variable ii;
  _for ii(1, printbar/barlen, 1)
  {
    printed += printf("%s", bar);
  }
  printed += printf("%s", tip);
  while ( (printed + tiplen + frontlen ) < cols + 1)
  {
    printed += printf("%s", front);
  }
  while (printed < cols - 1)
  {
    printed += printf(" ");
  }
  () = printf("]");
  () = printf("%s", append);
  () = fflush(stdout);
}

%
% Code from src/slang/astronomy/properties_satellite_galaxies.sl
%
define properties_satellite_galaxies()
%!%+
%\function{properties_satellite_galaxies}
%\synopsis{Retrieve the properties of several satellite galaxies of the Milky Way}
%\usage{Struct_Type s = properties_satellite_galaxies()}
%\description
%    Retrieve the properties (current kinematics, spatial extent, mass) of several satellite
%    galaxies of the Milky Way.
%
%    The output structure contains for each object the celestial coordinates in right ascension
%    [h, m, s] and declination [deg, arcmin, arcsec], distance [kpc], radial velocity [km/s],
%    proper motion in right ascension times cosine of declination [mas/yr], proper motion in
%    declination [mas/yr]), half-light radius [kpc], dynamcial mass inside half-light radius
%    [solar masses], Plummer softening radius [kpc], and the Plummer mass [solar masses].
%\notes
%    Positions, distances, radial velocities, half-light radii, and dynamcial masses are from
%    McConnachie 2012 when available. Proper motions (PMs) are available only for a subset of
%    satellites and are set to zero otherwise.
%
%    The parameters of the Plummer potentials (softening radius, mass) are determined under
%    the assumption that the dynamcial mass inside the half-light radius is 0.85 times the
%    total mass, which gives a softening radius of about one third of the half-light radius.
%
%    Main reference: McConnachie 2012, AJ, 144, 4.
%    Additional references or comments:
%    - LMC, SMC:         PMs and Plummer parameters from Kallivayalil et al. 2013, ApJ, 764, 161.
%                        Half-light radius and mass derived from Plummer parameters.
%    - Canis Major:      PMs from Dinescu et al. 2005, AJ, 631, L49.
%                        Dynamical mass assumed to be ten times stellar mass, half-light radius set to 1 kpc.
%    - Sagittarius dSph: PMs from Pryor et al. 2010, AJ, 139, 839.
%    - Fornax:           PMs from Piatek et al. 2007, AJ, 133, 818.
%    - Sextans:          PMs from Walker et al. 2008, ApJ, 688, L75.
%    - Sculptor:         PMs from Piatek et al. 2006, AJ, 131, 1445.
%    - Draco:            PMs from Pryor et al. 2014, AJ, submitted [arXiv:1407.3509].
%    - Ursa Minor:       PMs from Piatek et al. 2005, AJ, 130, 95.
%    - Carina:           PMs from Piatek et al. 2003, AJ, 126, 2346.
%    - Leo I:            PMs from Sohn et al. 2013, ApJ, 768, 139.
%    - Leo II:           PMs from Lepine et al. 2011, ApJ, 741, 100.
%    - Bootes III:       Dynamical mass assumed to be ten times stellar mass, half-light radius set to 10 pc.
%\example
%    s = properties_satellite_galaxies();
%    print_struct(s);
%\seealso{N_body_simulation_MW_kernel}
%!%-
{
  if(_NARGS!=0)
  {
    % _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': Wrong number of arguments.", _function_name());
  }
  %
  variable s = struct{ object, ah, am, as, dd, dm, ds, dist, vrad, pma_cos_d, pmd, hl_radius, hl_mass, Pl_radius, Pl_mass };
  s.object    = [    "LMC",    "SMC", "CanisMajor", "SagittariusdSph", "Fornax", "Sextans", "Sculptor",  "Draco", "UrsaMinor", "Carina",   "LeoI", "LeoII", "Segue", "UrsaMajorII", "BootesII", "SegueII", "Willman1",
		 "ComaBerenices", "Bootes", "UrsaMajor", "Hercules", "LeoIV", "CanesVenatici", "CanesVenaticiII", "BootesIII"];
  s.ah        = [    5.   ,    0.   ,         7.  ,             18.  ,    2.   ,     10.  ,       1.  ,   17.   ,       15.  ,     6.  ,  10.    ,  11.   ,   10.  ,          8.  ,      13.  ,      2.  ,      10.  ,
		 12.  ,    14.  ,       10.  ,      16.  ,   11.  ,           13.  ,             12.  ,       13.  ];
  s.am        = [   23.   ,   52.   ,        12.  ,             55.  ,   39.   ,     13.  ,       0.  ,   20.   ,        9.  ,    41.  ,   8.    ,  13.   ,    7.  ,         51.  ,      58.  ,     19.  ,      49.  ,
		 26.  ,     0.  ,       34.  ,      31.  ,   32.  ,           28.  ,             57.  ,       57.  ];
  s.as        = [   34.5  ,   44.8  ,        35.  ,             19.5 ,   59.3  ,      3.  ,       9.4 ,   12.4  ,        8.5 ,    36.7 ,  28.1   ,  28.8  ,    4.0 ,         30.  ,       0.  ,     16.  ,      21.  ,
  		 59.  ,     6.  ,       52.8 ,       2.  ,   57.  ,            3.5 ,             10.  ,       12.  ];
  s.dd        = [  -69.   ,  -72.   ,       -27.  ,            -30.  ,  -34.   ,     -1.  ,     -33.  ,   57.   ,       67.  ,   -50.  ,  12.    ,  22.   ,   16.  ,         63.  ,      12.  ,     20.  ,      51.  ,
		 23.  ,    14.  ,       51.  ,      12.  ,    0.  ,           33.  ,             34.  ,       26.  ];
  s.dm        = [   45.   ,   49.   ,        40.  ,             32.  ,   26.   ,     36.  ,      42.  ,   54.   ,       13.  ,    57.  ,  18.    ,   9.   ,    4.  ,          7.  ,      51.  ,     10.  ,       3.  ,
  		 54.  ,    30.  ,       55.  ,      47.  ,  -32.  ,           33.  ,             19.  ,       48.  ];
  s.ds        = [   22.   ,   43.   ,         0.  ,             43.  ,   57.   ,     53.  ,      33.  ,   55.   ,       21.  ,    58.  ,  23.    ,   6.   ,   55.  ,         48.  ,       0.  ,     31.  ,       0.  ,
  		 15.  ,     0.  ,       12.  ,      30.  ,    0.  ,           21.  ,             15.  ,        0.  ];
  s.dist      = [   51.   ,   64.   ,         7.  ,             26.  ,  147.   ,     86.  ,      86.  ,   76.   ,       76.  ,   105.  , 254.    , 233.   ,   23.  ,         32.  ,      42.  ,     35.  ,      38.  ,
		 44.  ,    66.  ,       97.  ,     132.  ,  154.  ,          218.  ,            160.  ,       47.  ];
  s.vrad      = [  262.2  ,  145.6  ,        87.  ,            140.  ,   55.3  ,    224.2 ,     111.4 , -291.   ,     -246.9 ,   222.9 , 282.5   ,  78.   ,  208.5 ,       -116.5 ,    -117.  ,    -39.2 ,     -12.3 ,
		 98.1 ,    99.  ,      -55.3 ,      45.2 ,  132.3 ,           30.9 ,           -128.9 ,      197.5 ];
  s.pma_cos_d = [    1.910,    0.772,        -1.61,             -2.75,    0.476,     -0.26,       0.09,    0.177,       -0.50,     0.22,  -0.1140,   0.104,    0.  ,          0.  ,       0.  ,      0.  ,       0.  ,
		 0.  ,     0.  ,        0.  ,       0.  ,    0.  ,            0.  ,              0.  ,        0.  ];
  s.pmd       = [    0.229,   -1.117,         0.84,             -1.65,   -0.360,      0.10,       0.02,   -0.221,        0.22,     0.15,  -0.1256,  -0.033,    0.  ,          0.  ,       0.  ,      0.  ,       0.  ,
		 0.  ,     0.  ,        0.  ,       0.  ,    0.  ,            0.  ,              0.  ,        0.  ];
  s.hl_radius = [   24.   ,    3.   ,         1.  ,              2.6 ,    0.7  ,      0.7 ,       0.28,    0.22 ,        0.18,     0.25,   0.25  ,   0.18 ,    0.03,          0.15,       0.05,      0.04,       0.03,
		 0.08,     0.24,        0.32,       0.33,    0.21,            0.56,              0.08,        0.01];
  s.hl_mass   = [25500.   , 2550.   ,       490.  ,            190.  ,   56.   ,     25.  ,      14.  ,   11.   ,        9.5 ,     6.3 ,  12.    ,   4.6  ,    0.26,          3.9 ,       3.3 ,      0.23,       0.27,
		 0.94,     0.81,       11.  ,       2.6 ,    1.3 ,           19.  ,              0.91,        0.17]*10^6;
  s.Pl_radius = s.hl_radius/3.;
  s.Pl_mass = s.hl_mass/0.85;
  return s;
}

%
% Code from src/misc/simulation/psd_lc.sl
%
private define psd_lc_normalized_lightcurve(rate, mean, sigma, poisson, c)
{
  variable m = moment(rate);
  % renormalize to desired mean count rate and variance
  if (poisson)
  {
    % the real count rate is Poisson distributed, therefore the time
    % resolution and number of PCUs must be taken into account
    if( (sigma*c)^2 < mean*c)
       return vmessage("error (%s): sigma%s is too small", _function_name(), qualifier("id", ""));
    sigma = sqrt(((sigma*c)^2-mean*c))/c;

    return prand((mean + (rate-m.ave)/m.sdev * sigma)*c)/c;
  }
  else
    return        mean + (rate-m.ave)/m.sdev * sigma;
}


%%%%%%%%%%%%%
define psd_lc()
%%%%%%%%%%%%%
%!%+
%\function{psd_lc}
%\synopsis{simulate a random light curve that follows a given PSD using the algorithm of Timmer and Koenig}
%\usage{Double_Type rate[] = psd_lc(Integer_Type n, Double_Type dt, Ref_Type PSD);
%\altusage{(rate1, rate2) = psd_lc(Integer_Type n, Double_Type dt, Ref_Type PSD; time_lag_spectrum=...);}
%}
%\qualifiers{
%\qualifier{mean}{[= 100]: mean count rate of the simulated lightcurve}
%\qualifier{sigma}{[= 20]: standard deviation of the simulated lightcurve}
%\qualifier{poisson}{if Poisson noise should be applied on the lightcrurve}
%\qualifier{nr_PCUs}{[= 1]: number of PCUs}
%\qualifier{time_lag_spectrum}{spectrum of time lags (two lightcurves will be returned)}
%\qualifier{mean_2}{[= 100]: mean count rate of the second lightcurve}
%\qualifier{sigma_2}{[= 20]: standard deviation of the second lightcurve}
%}
%\description
%    \code{n} is the number of bins of the simulated lightcurve.
%    It should be a power of two for best performance of the FFT.
%    \code{dt} is the time resolution.
%    \code{PSD} is a reference to a function which takes one argument
%    -- the frequency -- and calculates the corresponding PSD value.
%    The number of PCUs is needed for the calculation of Poisson noise
%    if a mean RXTE count rate is given in counts/PCU.
%
%    If \code{time_lag_spectrum} is reference to a function of one argument
%    -- the frequency -- that calculates the time lag spectrum,
%    an additional second lightcurve is returned that has
%    the corresponding time lag with respect to the first one.
%
%    see Timmer & Koenig (1995): "On generating power law noise",
%    A&A 300, 707-710
%!%-
{
  variable n, dt, PSD;
  switch(_NARGS)
  { case 3: (n, dt, PSD) = (); }
  { return help(_function_name()); }

  variable n_2 = int(n/2);
  if(2*n_2 != n)
    return vmessage("error (%s): n must be even", _function_name());

  variable time_lag_spectrum = qualifier("time_lag_spectrum");  % if this qualifier is set, 2 lightcurves will be returned
  variable poisson = qualifier_exists("poisson");  % if poisson noise should be applied on the simulated lightcurve
  variable c = dt * qualifier("nr_PCUs", 1);

  % generate frequencies {1/T,...,1/2t}
  variable f = 1.*[1:n_2]/(n*dt);
  variable fa = @PSD(f ;;__qualifiers() );
  variable fac = sqrt(1/2.*fa);
  % multiply by gaussian distributed random number to get real and imaginary part
  variable pos_real = grand(n_2)*fac;
  variable pos_imag = grand(n_2)*fac;
  pos_imag[n_2-1] = 0;
  % get the values for negative frequencies as g(-f)=g*(f)
  variable real = [ 0., pos_real,  pos_real[[n_2-2:0:-1]] ];
  variable imag = [ 0., pos_imag, -pos_imag[[n_2-2:0:-1]] ];
  variable z = real + imag*1i;

  psd_lc_normalized_lightcurve
    (
	Real( fft(z, 1) ),       % simulated light curve from its Fourier transform
	qualifier("mean", 100),  % mean count rate of the simulated lightcurve
	qualifier("sigma", 20),  % standard deviation of the lightcurve to be simulated
	poisson,
	c
    );  % first return value left on stack

  if(time_lag_spectrum == NULL)
    return;  % no timelag given => no second lightcurve

  %
  % simulate the second lc with a time lag
  %
  variable f1 = [0, f];
  variable f2 = f[[n_2-2:0:-1]];
  psd_lc_normalized_lightcurve
    (
	Real( fft(z*[exp( 2i*PI*f1 * @time_lag_spectrum(f1 ;;__qualifiers() )),
		     exp(-2i*PI*f2 * @time_lag_spectrum(f2 ;;__qualifiers() ))], 1) ),
	qualifier("mean_2", 100),  % mean count rate of the simulated lightcurve
	qualifier("sigma_2", 20),  % standard deviation of the lightcurve to be simulated
	poisson,
	c
      ; id="_2"
    );  % second return value left on stack
}

%
% Code from src/data/timing/psdcorr_zhang.sl
%
private define zhangfunction(acofidx, sumidx, deadtime, mincrate, bintime)
{
  variable tau = 1. / mincrate ;
  variable t = acofidx * bintime - sumidx * deadtime ;
  variable x = t / tau ;
  if(t >= 0.)
  {
    variable nsumidx = int(sumidx);
    variable g = 0.;
    variable i;
    _for i (0, nsumidx, 1)
      g += (double(nsumidx - i)/factorial(i)) * x^i;
    return (acofidx) - (sumidx * (deadtime + tau) / bintime) + ((tau * exp(-x) * g) / bintime);
  }
  return 0.;
}

%%%%%%%%%%%%%%%%%%%%
define psdcorr_zhang(totrate, tseg, dimseg)
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{psdcorr_zhang}
%\synopsis{Timing Tools: Poisson Noise and Deadtime Correction (Zhang)}
%\usage{(freq, psd) = psdcorr_zhang(totrate, tseg, dimseg);}
%\qualifiers{
%\qualifier{nonparalyzable}{Set this qualifier if the deadtime is non-paralyzable}
%}
%\description
%    The deadtime is assumed to be paralyzable by default.
%    
%    Inputs:
%      totrate - total countrate of all instruments
%      tseg    - realtime length of the lc segments used for psd calculation
%      dimseg  - bincount of the lc segments used for psd calculation
%
%    Outputs:
%      freq    - fourier frequency array
%      noipsd  - array of observational noise of the psd
%\seealso{Zhang et al. (1995), ApJ, 449, 930}
%!%-
{
  variable detrate = NULL;
  variable bintime = tseg / dimseg;
  variable timearray = [1:dimseg] * bintime;

  variable numinst = qualifier("numinst", 1.);
  variable deadtime = qualifier("deadtime", 1e-5);

  %%better not get avgrate values from the qualifier - the qualifier
  %%from fourcalc will be used, which is a reference and the whole thing
  %%will break down
  %variable avgrate = qualifier("avgrate", totrate);
  variable avgrate = @totrate;
  
  variable incrate = qualifier("incrate", NULL);
  variable normtype = qualifier("normtype", "Miyamoto");
  variable avgbkg = qualifier("avgbkg", 0.);

  if(incrate==NULL)
    detrate = totrate/numinst;
  else
    incrate = totrate/numinst;

  variable freq = foufreq(timearray);
  variable om = 2. * PI * bintime * freq;
  variable zhangpsd = Double_Type[dimseg/2];
  variable i, j;

  if(qualifier_exists("nonparalyzable"))
  {
    if(incrate!=NULL)
      detrate = incrate / (1. + incrate * deadtime);
    else
      incrate = detrate / (1. - detrate * deadtime);

    variable dim = int((dimseg * bintime / deadtime ) + 2.);
    variable h = Double_Type[dimseg+1, dim+1];
    _for i (0, dimseg, 1)
      _for j (1, dim, 1)
	h[i,j] = zhangfunction(double(i), double(j), deadtime, incrate, bintime);

    variable a = Double_Type[dimseg];
    variable sumdim = int(bintime/deadtime + 1.);
    a[0] = detrate * bintime * (1. + 2.*sum(h[1, [1:sumdim+1]]));
    _for i (1, dimseg-1, 1)
    {
      sumdim = int( (double(i) + 1.) * bintime / deadtime + 1.);
      a[i] = detrate * bintime * sum( h[i+1,[1:sumdim+1]] - 2.*h[i,[1:sumdim+1]] + h[i-1,[1:sumdim+1]] );
    }

    variable b = 4. * (a - detrate * detrate * bintime * bintime) / (detrate * bintime);
    b[0] = b[0] / 2.;

    _for i (1, dimseg-1, 1)
      zhangpsd += (double(dimseg-i) / double(dimseg)) * b[i] * cos(om*double(i));
    zhangpsd += b[0];
  }
  else
  {
    if(incrate!=NULL)
      detrate = incrate * exp(-incrate*deadtime);
    else
      incrate = detrate / (1. - detrate * deadtime);

    if(bintime >= deadtime)
      zhangpsd = 2. * (  1.
		       - 2. * detrate * deadtime * (1. - deadtime/(2.*bintime))
		       - ( (dimseg - 1.) / dimseg) * detrate * deadtime * (deadtime / bintime) * cos(om));
    else
    { variable m = 1.*int(deadtime/bintime);
      zhangpsd = 2. - 2 * detrate * bintime * (  1.
                                               - (dimseg-m)/dimseg * (m+1.-deadtime/bintime)^2 * cos(m*om)
                                               + (dimseg-m-1.)/dimseg *  (m-deadtime/bintime)^2 * cos((m+1.)*om)
                                               + 2. * cos((m+1.)/2.*om) * sin(m/2.*om)/sin(om/2.)
                                               - (m+1.)/dimseg * sin((2.*m+1.)/2.*om)/sin(om/2.)
                                               + 1./dimseg * (sin((m+1.)/2.*om)/sin(om/2.))^2.
		                              );
    }
  }

  % zhangpsd is Leahy-normalzed => unnormalized (raw) noipsd
  variable noipsd = (zhangpsd*dimseg*dimseg*avgrate)/(2.*tseg);

%  if(not qualifier_exists("unnormalized"))
%    noipsd = normpsd(noipsd, normtype, avgrate, avgbkg, tseg, dimseg);
%   return freq, noipsd;
  return freq, noipsd, normpsd(noipsd, normtype, avgrate, avgbkg, tseg, dimseg);
}

%
% Code from src/data/GTI/pulsar_GTI.sl
%
% -*- mode:slang; mode:fold -*-

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define phasebin_to_gti(tstart, tstop, tref, p, pdot, pddot, phase_lo, phase_hi, satellite){ %{{{
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  variable d2s = 86400.;
  variable eph = struct{t0, p0 = p, pdot = pdot, p2dot = pddot, p3dot =0.}; % orbit structure required by pulse_time

  if (qualifier_exists("MJD")){
    eph.t0 = double(tref); 
    tref = (tref-MJDref_satellite(satellite))*d2s; % Converting the reference time in the  staellite's time format
    tstart = (tstart-MJDref_satellite(satellite))*d2s; % Start and stop time are converted to seconds, because
    tstop = (tstop-MJDref_satellite(satellite))*d2s;   % the pulse period will always be in seconds!
    
  }else{
    eph.t0 = tref/d2s + MJDref_satellite(satellite); % The reference time in the ephemeris structure must always be in MJD
  }
  variable nxp = [ int((tstart-tref)/p)-2 : int((tstop-tref)/p) +2] ; %index array of pulses during observation
  
  variable gti = struct {
    start = Double_Type[length(nxp)] ,
    stop = Double_Type[length(nxp)]
  };
  
  %calculate pulse arrival times
  gti.start = pulse_time (nxp + phase_lo; eph = eph); % pulse_time returns arrival times in MJD
  gti.stop =  pulse_time (nxp + phase_hi; eph = eph); % Make sure the MJD qualifier is NOT passed to pulse time!!!
  
  %convert to satellite's time system
  gti.start = (gti.start - MJDref_satellite(satellite))*d2s; % convert back to seconds
  gti.stop = (gti.stop - MJDref_satellite(satellite))*d2s;
  
  return gti;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define convert_GTI(nobarevt, barevt, gti){ %{{{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  % Do linear interpolation to calculate the GTIs in the satellite's time format
  variable nobargti = struct{
    start = interpol (gti.start, barevt.time, nobarevt.time),
    stop = interpol (gti.stop, barevt.time, nobarevt.time)};    

  return nobargti;
};
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define pulsar_GTI(){ %{{{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulsar_GTI}
%\synopsis{Calculates GTIs for pulse phase resolved spectroscopy}
%\usage{pulsar_GTI(Double_Type tstart, tstop, String_Type satellite, basefilename, Double_Type t0, p, phase_lo, phase_hi);}
%\qualifiers{
%\qualifier{pdot}{first derivative of the pulse period in s/s (default: 0) }
%\qualifier{pddot}{second derivative of the pulse period in s/s^2 (default: 0) }
%\qualifier{MJD}{If set, tstart, tstop and t0 are assumed to in MJD}
%\qualifier{local}{create GTIs in satellite's local time system. Requires qualifiers 'nobarevt' and 'barevt' to be set.}
%\qualifier{barevt}{event file in barycentered time frame}
%\qualifier{nobarevt}{event file in local time frame}
%\qualifier{}{all other qualifiers are passed to BinaryPos}
%}
%\description
%   This function calculates the GTIs for pulse phase resolved spectroscopy. Input arguments are\n
%   
%      tstart - start time of the time interval to be covered (typically the start of the observation)
%      tstart - end time of the time interval to be covered (typically the end of the observation) 
%      satellite - the name of the satellite (needed for setting the reference MJD correctly)
%      basefilename - the file created will be named 'basefilename.gti'
%      t0 - the reference time (time of phase zero)
%      p - the pulse period
%      phase_lo - the lower phase boundary
%      phase_hi - the upper phase boundary
%   \n
%   Upon qualifier request, tstart, tstop, and t0 can be given in MJD instead of seconds. The pulse
%   period is always in seconds.\n
%   \n
%   If the pulse period and the reference time are corrected for the binary
%   motion, the orbital parameters should be provided by qualifiers such that
%   the GTIs can be transformed into the observers time system. The
%   qualifiers are passed and equal to the \code{BinaryPos} function.
%   \n
%   If GTIs in the satellite's local reference time system are needed (e.g., for Suzaku-XIS),
%   an event file in both local and barycenterd time can be passed via qualifiers and the GTIs
%   are interpolated to the local time system. Header keywords are set accordingly.
%\example
%   pulsar_GTI (2.7082e8, 2.7097e8, "nustar", "phase_0.25-0.35", 2.708242e8, 443.07, 0.25, 0.35);
%   \n
%   creates a GTI file named 'phase_0.25-0.35.gti', ranging from 2.7082e8--2.7097e8 seconds
%   in NuSTAR's mission specific time system for the pulsar 4U 1907+09 with pulse period 443.07 seconds
%   for the pulse phase interval 0.25--0.35 where phase 0.0 is set to be at 2.708242e8 seconds.
%
%\seealso{MJDref_satellite, BinaryPos, pulse_time}
%!%-
  
  variable nobarevt, barevt, name, t0, p, phase_lo, phase_hi, tstart, tstop, satellite;
  switch(_NARGS)
  {case 8: (tstart, tstop, satellite, name, t0, p, phase_lo, phase_hi) = (); }
  {help(_function_name()); return; }
  
  variable fn_local = qualifier("nobarevt", NULL);
  variable local = qualifier_exists("local");
  variable fn_bary = qualifier("barevt", NULL);
  variable pdot = qualifier("pdot", 0.);
  variable pddot = qualifier("pddot", 0.);
  variable d2s = 86400.; % days to seconds
  
  if (local){
    if ( fn_local == NULL or fn_bary == NULL){
      fprintf(stderr, "%s: ERROR: Conversion to LOCAL time system requires qualifiers 'nobarevt' and 'barevt' to be set!\n", _function_name);
      return;
    }
    %Check the arguments   
    if(typeof(fn_local) ==  String_Type){
      if(fits_read_key(fn_local, "TIMEREF") != "LOCAL"){
	()= printf("ERROR: Event file has wrong reference time! \n");
	return -1;
      };
      nobarevt = fits_read_table(fn_local);
    };
    if(typeof(fn_bary) == String_Type){
      if(fits_read_key(fn_bary, "TIMEREF") != "SOLARSYSTEM"){
	()=printf("ERROR: Event file has wrong reference time! \n");
	return -1;
      };
      barevt = fits_read_table(fn_bary);
    };
    %Check the arguments
    if (length(nobarevt.time) != length(barevt.time)){
      () = printf("ERROR: Event files do not contain an equal number of events! \n");
      return -1;
    };
    
    
    struct_filter(nobarevt, array_sort(barevt.time));
    struct_filter(barevt, array_sort(barevt.time));   
  }
  
  variable bargti =  phasebin_to_gti(tstart, tstop, t0, p, pdot, pddot, phase_lo, phase_hi, satellite;;__qualifiers);  
  
  % binary correction
  if (qualifier_exists("asini") and qualifier_exists("porb")) {
    variable qual = reduce_struct(__qualifiers, "satellite");
    % Do the binary correction in MJD 
    bargti.start = bargti.start/d2s + MJDref_satellite(satellite);
    bargti.stop = bargti.stop/d2s + MJDref_satellite(satellite);   
    
    bargti.start += BinaryPos(bargti.start;; qual)/d2s;
    bargti.stop  += BinaryPos(bargti.stop ;; qual)/d2s;
    
    bargti.start = (bargti.start - MJDref_satellite(satellite))*d2s;
    bargti.stop =  (bargti.stop - MJDref_satellite(satellite))*d2s;
  }
  
  if (local){
    variable nobargti =  convert_GTI(nobarevt, barevt, bargti);
  }
  
  % Write the FITS file and update some keywords  
  fits_write_gti(name +".gti", local ? nobargti : bargti, MJDref_satellite(satellite); date = time() );
  fits_update_key(name +".gti", "TELESCOP", satellite, "Mission name");
  fits_update_key(name +".gti", "TIMEREF", local ? "LOCAL" : "SOLARSYSTEM");
  fits_update_key(name +".gti[1]", "TIMEREF", local ? "LOCAL" : "SOLARSYSTEM");
  fits_update_key(name +".gti", "MJDREFI", int(MJDref_satellite(satellite)), "MJD reference day");
  fits_update_key(name +".gti", "MJDREFF", MJDref_satellite(satellite) - int(MJDref_satellite(satellite)), "fractional part of the MJD reference");  
  fits_update_key(name +".gti[1]", "MJDREFI", int(MJDref_satellite(satellite)), "MJD reference day");
  fits_update_key(name +".gti[1]", "MJDREFF", MJDref_satellite(satellite) - int(MJDref_satellite(satellite)), "fractional part of the MJD reference");
  fits_write_comment (name +".gti", "Phase_lo = " + string(phase_lo));
  fits_write_comment (name +".gti", "Phase_hi = " + string(phase_hi));
}
%}}}
%
% Code from src/fitting/fit-functions/pulsar/pulsarorbit.sl
%
%%%%%%%%%%%%%%%%%%%%%
private define pulsarorbit_fit(lo, hi, par)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulsarorbit}
%\synopsis{fit-function modelling a neutron star pulse period evolution}
%\usage{fit_fun("pulsarorbit");}
%\description
%    DEPRECATED, use 
%      fit_fun("dopplerorbit*pulsartorque");
%    from now on.
%!%-
{
  message("*** DEPRECATED ***");
  return eval_fun2("dopplerorbit", lo, hi, [par[[6:]], 0])
       * eval_fun2("pulsartorque", lo, hi, [par[[:5]], 0.]);
}


% add fit-function
add_slang_function("pulsarorbit", [
  "p [s]", "t0 [MJD]", "a [s/s]", "b [s/s @ (L0,p)]", "alpha", "L0",
  "asini [lt-s]", "porb [days]", "tau [MJD]", "ecc", "omega [degrees]"
]);
set_function_category("pulsarorbit", ISIS_FUN_ADDMUL);


% define parameter defaults
private define pulsarorbit_defaults(i) {
  return [
    struct { value = 100.0, freeze = 0, hard_min = 0.0, hard_max = 10000, min = 0, max = 300, step = 1e-6, relstep = 1e-8 },
    struct { value = 56000.0, freeze = 1, hard_min = 15000, hard_max = int(ceil(UNIXtime2MJD(_time))), min = 15000, max = int(ceil(UNIXtime2MJD(_time))), step = 1e-1, relstep = 1e-3 },
    struct { value = 0.0, freeze = 1, hard_min = -1e-7, hard_max = 1e-7, min = 0, max = 1e-7, step = 1e-10, relstep = 1e-12 },
    struct { value = 0.0, freeze = 1, hard_min = 0, hard_max = 1e-7, min = 0, max = 1e-7, step = 1e-10, relstep = 1e-12 },
    struct { value = 6./7., freeze = 1, hard_min = 0, hard_max = 10, min = 6./7., max = 1., step = 1e-2, relstep = 1e-3 },
    struct { value = 1., freeze = 1, hard_min = 0, hard_max = DOUBLE_MAX, min = 0, max = 0, step = 0, relstep = 0 },
    struct { value = 100.0, freeze = 0, hard_min = 0.0, hard_max = 10000, min = 0, max = 1000, step = 1e-2, relstep = 1e-4 },
    struct { value = 10.0, freeze = 0, hard_min = 0.0, hard_max = 10000, min = 0, max = 1000, step = 1e-2, relstep = 1e-4 },
    struct { value = 56000.0, freeze = 0, hard_min = 15000, hard_max = int(ceil(UNIXtime2MJD(_time))), min = 15000, max = int(ceil(UNIXtime2MJD(_time))), step = 1e-1, relstep = 1e-3 },
    struct { value = 0.0, freeze = 0, hard_min = 0.0, hard_max = 1.0, min = 0, max = 1.0, step = 1e-2, relstep = 1e-4 },
    struct { value = 0.0, freeze = 0, hard_min = -360, hard_max = 360, min = -180, max = 180, step = 1e-1, relstep = 1e-3 }
  ][i];
}
set_param_default_hook("pulsarorbit", &pulsarorbit_defaults);


%%%%%%%%%%%%%%%%%%%%%
private define pulsar_fit(lo, hi, par)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulsar}
%\synopsis{fit-function modelling a neutron star pulse period evolution}
%\usage{fit_fun("pulsar");}
%\description
%    Deprecated, use 
%      fit_fun("pulsartorque");
%    from now on.
%!%-
{
  message("*** DEPRECATED ***");
  return eval_fun2("pulsartorque", lo, hi, [par, 0.]);
}


% add fit-function
add_slang_function("pulsar", [
  "p [s]", "t0 [MJD]", "a [s/s]", "b [s/s @ (L0,p)]", "alpha", "L0"]
);
set_function_category("pulsar", ISIS_FUN_ADDMUL);


% define parameter defaults
set_param_default_hook("pulsar", &pulsarorbit_defaults);

%
% Code from src/fitting/fit-functions/pulsar/pulsarorbit_fluxerror_mc.sl
%
%%%%%%%%%%%%%%%%%%%%%
define pulsarorbit_fluxerror_mc()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulsarorbit_fluxerror_mc}
%\synopsis{estimates additional uncertainties caused by the flux measurements}
%\usage{Double_Type[] pulsarorbit_fluxerror_mc(Integer_Type dataset);}
%\qualifiers{
%    \qualifier{runs}{number of MC loops (default: 100)}
%    \qualifier{save}{save the uncertainty estimation to the
%              assigned FITS-filename}
%    \qualifier{collect}{file-pattern used by 'glob' to read and
%              merge all FITS-files from previous runs}
%    \qualifier{modify}{add the estimated uncertainties to the
%              dataset(s) assigned to this qualifier
%              (see below for a detailed description)}
%    \qualifier{chatty}{be chatty if > 0 (default: 0)}
%}
%\description
%    The flux measurements used to calculate the evolution
%    of the spin-period in the fit-function 'pulsarorbit'
%    usually have uncertainties. These uncertainties are
%    not taken into account during an ordinary fit. This
%    function performs Monte Carlo simulations to estimate
%    the uncertainties of the modelled period induced by
%    the flux uncertainties.
%
%    During each run the flux evolution associated to the
%    dataset 'id' is varied using
%      flux = flux + grand * flux_err
%    such that synthetic flux evolutions are created. Then
%    a fit is performed using only the given dataset 'id'.
%    This results in many modelled period evolutions. Their
%    standard deviation at each time is finally considered
%    as an additional uncertainty in the period space.
%    These uncertainties are returned by the function.
%
%    If the 'modify'-qualifier is set, the dataset(s)
%    assigned to that qualifier or, in case of NULL, the
%    dataset 'id' is modified as follows:
%      new_error = sqrt(sqr(data_error) + sqr(mc_error))
%    where data_error is the current 'error'-field of the
%    dataset and mc_error is the estimated additional
%    uncertainty as calculated by this function. In case
%    of multiple given datasets, the time range of 'id'
%    should include all of these datasets to get a proper
%    period evolution.
%\seealso{pulsarorbit}
%!%-
{
  variable id;
  if (_NARGS != 1) { help(_function_name); return; }
  id = ();

  variable chatty = qualifier("chatty", 0);
  
  % either perform the simulation or load from previous runs
  variable sim, n;
  if (qualifier_exists("collect")) {
    if (chatty) { message("collecting previous runs"); }
    sim = array_map(Array_Type, &get_struct_field,
      array_map(Struct_Type, &fits_read_table, glob(qualifier("collect"))), "periods"
    );
    variable dim = [0,array_shape(sim[0])[1]];
    _for n (0, length(sim)-1, 1) {
      dim[0] += array_shape(sim[n])[0];
    };
    sim = _reshape(array_flatten(sim), dim);
  } else {  
    % get flux evolution
    variable L = get_dataset_metadata(id);
    if (L == NULL) {
      vmessage("error (%s): metadata-structure is not set", _function_name); return;
    }
    if (not struct_field_exists(L, "time") or not struct_field_exists(L, "flux") or not struct_field_exists(L, "flux_err")) {
      vmessage("error (%s): metadata-structure does not have the required fields", _function_name); return;
    }

    variable pars = get_params;
    variable incl = all_data[where(merge_struct_arrays(get_data_info(all_data)).exclude == 0)];
    exclude(all_data);
    include(id);
    
    % Monte Carlo uncertainty estimation
    seed_random(_time);
    if (chatty) { message("starting MC runs"); }
    variable runs = qualifier("runs", 100);
    sim = Array_Type[runs];
    _for n (1, runs, 1) {
      variable mcflux = L.flux + grand(length(L.flux))*L.flux_err;
      mcflux[where(mcflux < 0)] = 0.;

      set_dataset_metadata(id, struct_combine(L, struct { flux = mcflux } ));

      ()=fit_counts(; fit_verbose = -1);

      sim[n-1] = get_model_counts(id).value;
    }

    % restore original flux data and parameter
    set_dataset_metadata(id, L);
    set_params(pars);
    exclude(id);
    include(incl);

    % reshape
    sim = _reshape(array_flatten(sim), [length(sim), length(sim[0])]);
  }

  if (chatty) {
    vmessage("time length = %d, runs = %d", length(sim[0,*]), length(sim[*,0]));
  }
  % eventually save
  if (qualifier_exists("save")) {
    if (chatty) { message("saving runs"); }
    fits_write_binary_table(qualifier("save"), "mcflux", struct { periods = sim });
  }
  
  % calculate standard deviation
  variable err = Double_Type[length(sim[0,*])];
  _for n (0, length(err)-1, 1) {
    err[n] = moment(sim[*,n]).sdev;
  }

  % modify dataset
  if (qualifier_exists("modify")) {
    variable mid = qualifier("modify");
    if (mid == NULL) { mid = [id]; }
    if (typeof(mid) != Array_Type or (_typeof(mid) != Integer_Type and _typeof(mid) != UInteger_Type)) {
      vmessage("error (%s): 'modify' must be an array of integers", _function_name); return;
    }

    % loop and modify
    variable t = get_data_counts(id).bin_lo;
    foreach n (mid) {
      if (chatty) { vmessage("modifying dataset %d", n); }
      variable cts = get_data_counts(n);
      cts.err = sqrt(sqr(cts.err) + sqr(interpol(cts.bin_lo, t, err)));
      put_data(n, cts);
    }
  }
  
  return err;
}

%
% Code from src/fitting/fit-functions/pulsar/pulsartaylor.sl
%
%%%%%%%%%%%%%%%%%%%%%
private define pulsartaylor_fit(lo, hi, par)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulsartaylor}
%\synopsis{fit-function modelling the pulse period evolution with a Taylor series}
%\usage{fit_fun("pulsartaylor");}
% altusage{fit_fun("dopplerorbit*pulsartaylor");}
%\description
%    This fit-function computes the pulse period evolution of, e.g.,
%    a neutron star based on a Taylor series. That is
%      p(t) = p0 + Pdot*(t-t0) + .5*Pddot*(t-t0)^2 + ...
%    Here, p0 is the pulse period at the reference time t0 and
%    Pdot, Pddot,... are the higher order derivatives.
%
%    The parameters of the fit-function are:
%    p0    - spin-period at t0 (s)
%    t0    - reference time t0 (MJD), has to be fixed
%    pdot  - period derivative (s/s)
%    p2dot - 2nd period derivative (s/s^2)
%    etc.
%
%    By default the Taylor series is computed up to the second
%    order. If your data requires higher orders, you can change this
%    limit using 'pulsartaylor_set_order'.
%
%    NOTE: in case you add the Doppler shift of orbital motion to the
%          model using the 'dopplerorbit' fit-function, make sure that
%          this fit-function is calculated *first* in order to
%          transform the times ('lo' parameter of the fit-function)
%          into the barycenter of the binary! These corrected times
%          are shared among these fit-functions via the ISISscripts
%          caching extension. Then the reference time, t0, is
%          interpreted as binary corrected!
%
%    NOTE: a Taylor series describes the pulse period evolution
%          phenomenologically. In case of mass-accretion from a donor
%          star the evolution is driven by the mass accretion rate and
%          a Taylor series does not model this correctly. It is likely
%          that further fit-parameters, such as orbital parameters,
%          get biased due to this imperfect modelling (see, e.g., PhD
%          thesis of M. Bissinger)!
%\example
%    % define the measured period evolution as ISIS dataset
%    id = define_counts(time, make_hi_grid(time), period, period_err);
%
%    % set the fit-function
%    fit_fun("pulsartaylor(1)");
%\seealso{pulseperiod, taylor, pulsartaylor_set_order, fitfun_cache}
%!%-
{
  % input parameters
  variable t0 = par[0], p0 = par[1];

  % get binary corrected times
  variable orbitcache = fitfun_get_cache("dopplerorbit");
  variable binarytime = lo;
  if (orbitcache != NULL && orbitcache.binarytime != NULL) {
    binarytime = orbitcache.binarytime;
  }

  % define period structure
  variable p = struct { t0 = t0, p0 = p0 };
  if (length(par) > 1) {
    p = struct_combine(p, struct { pdot = par[2] });
  }
  if (length(par) > 2) {
    variable f = array_map(String_Type, &sprintf, "p%ddot", [[2:length(par)-2]]);
    p = struct_combine(p, @Struct_Type(f));
    array_map(Void_Type, &set_struct_field, p, f, par[[3:]]);
  }
  
  % return model on the data time grid (as an array always!)
  return [ pulseperiod(binarytime, p) ];
}


% define parameter defaults
private define pulsartaylor_defaults(i) {
  switch (i)
  { case 0:
    return struct { value = 56000.0, freeze = 1, hard_min = 15000, hard_max = 100000, min = 15000, max = int(ceil(UNIXtime2MJD(_time))), step = 1e-1, relstep = 1e-3 };
  }
  { case 1:
    return struct { value = 100.0, freeze = 0, hard_min = 1e-6, hard_max = 10000, min = 1e-3, max = 300, step = 1e-6, relstep = 1e-8 };
  }
  {
    return struct { value = 0.0, freeze = i > 2, hard_min = -1e-3, hard_max = 1e-3, min = -10^(-2-3*i), max = 10^(-2-3*i), step = 10^(-5-3*i), relstep = 10^(-7-3*i) };
  }
}

private variable _pulsartaylor_handle = &pulsartaylor_fit;
private variable _pulsartaylor_defaults_handle = &pulsartaylor_defaults;


%%%%%%%%%%%%%%%%%%%%%
define pulsartaylor_set_order()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulsartaylor_set_order}
%\synopsis{changes the order of the Taylor series used in 'pulsartaylor'}
%\usage{pulsartaylor_set_order(Integer_Type new_order);}
%\description
%    By default the Taylor series used in the 'pulsartaylor' fit-
%    function is computed up to the second order. This function allows
%    to change this order starting at zero up to the 7th order. The
%    number of fit-parameters are adapted and named accordingly.
%
%    NOTE: changing a fit-function and its parameters on the fly is
%          not supported by ISIS. In order to still achieve this
%          feature here, the fit-function is first deleted using
%          'del_function' and defined again using 'add_slang_function'.
%          Testing revealed, however, that this trick only works once!
%\seealso{pulsartaylor}
%!%-
{
  if (_NARGS != 1) { help(_function_name); }
  if (_pulsartaylor_handle == NULL || _pulsartaylor_defaults_handle == NULL) {
    vmessage("error (%s): setting the order only works once!", _function_name);
    return;
  }
  
  % get and check order
  variable order = ();
  if (order < 0) {
    vmessage("error (%s): order has to be positive");
    return;
  }
  if (order > 7) {
    vmessage("error (%s): only up to the 7th order allowed");
    return;
  }

  % delete old fit-function
  if (any(_isis->_function_list == "pulsartaylor")) {
    del_function("pulsartaylor");
  }
  
  % add fit-function with new order
  variable args = String_Type[order+2];
  args[[0,1]] = ["t0", "p0"];
  if (order > 0) { args[2] = "pdot [s/s]"; }
  if (order > 1) {
    args[[3:]] = array_map(String_Type, &sprintf, "p%ddot [s/s^%d]", [2:order], [2:order]);
  }
  add_slang_function("pulsartaylor", _pulsartaylor_handle, args);
  set_function_category("pulsartaylor", ISIS_FUN_ADDMUL);
  set_param_default_hook("pulsartaylor", _pulsartaylor_defaults_handle);

  % reset fit-function in order to update the parameter list
  fit_fun(get_fit_fun);

  % since re-defining is working only once (acutally the second time,
  % the parameter list is wrong after having called this function the
  % third time), we set the pointers to NULL. Maybe there is a way to
  % let ISIS purge the parameter/fit-function memory?
  if (not qualifier_exists("anyway")) {
    _pulsartaylor_handle = NULL;
    _pulsartaylor_defaults_handle = NULL;
  }
}

% add the fit-function with a default order of 2
pulsartaylor_set_order(2; anyway);

%
% Code from src/fitting/fit-functions/pulsar/pulsartorque.sl
%
% define the simplified ordinary differential equation
%   \dot{P} = a - b * P^2 L^alpha
% after Ghosh & Lamb (1979), which will be solved by the
% fit-function below.
private define _ode_torque(p, t) {
  variable p0 = qualifier("p0");
  variable a = qualifier("a");
  variable b = qualifier("b");
  variable alpha = qualifier("alpha");
  variable lc = qualifier("lc");

  % select the rate at 't'
  variable rate = interpol(t, lc.time, lc.rate);
  
  return a - b * (p/p0)^2 * sign(rate)*abs(rate)^alpha;
}

%%%%%%%%%%%%%%%%%%%%%
private define pulsartorque_fit(lo, hi, par)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulsartorque}
%\synopsis{fit-function modelling the accretion torque of a neutron star (simple way)}
%\usage{fit_fun("pulsartorque");}
% altusage{fit_fun("dopplerorbit*pulsartorque");}
%\description
%    The spin-up of an accreting neutron star is connected with its
%    luminosity via
%      Pdot = - b * P^2 L^alpha
%    as found by Ghosh & Lamb (1979). This fit-function solves this
%    differential equation in order to calculate the spin-evolution of
%    the neutron star.
%
%    The parameters of the fit-function are:
%    p     - spin-period at t0 (s)
%    t0    - reference time t0 (MJD), has to be fixed
%    a     - constant spin-up or -down (s/s)
%    b     - torque strength (s/s @ Lnorm and Pnorm)
%    alpha - exponent of b (try to freeze it)
%    Lnorm - luminosity used for normalization, has to be fixed
%    Pnorm - spin-period user for normaliziation, has to be fixed
%
%    In order to express the torque strength, b, in the usual unit of
%    a spin-change (s/s), the equation is modfied by two normalization
%    constants:
%      Pdot = - b * (P/Pnorm)^2 (L/Lnorm)^alpha
%    where Pnorm is a reference spin-period and Lnorm a reference
%    luminosity. For instance, Pnorm could be tied to parameter p,
%    i.e., the period p at t0. This should be avoided, however, when
%    multiple dataset are fitted with the same torque strength b, but
%    different values for the parameter p. Here Pnorm has to be the
%    same among all datasets! The same applies to Lnorm, which should
%    be fixed to an (arbitrary) flux or rate, for instance
%    corresponding to the flux of the Crab.
%    NOTE: setting Pnorm = 0 (the default) uses Pnorm = p internally
%          in order to ensure backward compatibility.
%
%    The exponent, alpha, of the torque strength, b, depends on the
%    accretion mechanism. After Ghosh & Lamb, alpha=6/7 for disk-, and
%    alpha=1 for wind- accretion.
%
%    In order to calculate the spin-evolution, the count rate over
%    time, i.e., a light curve has to be assigned to the dataset using
%    'set_dataset_metadata' (see example below). The period
%    measurements have to be within the time range of the light
%    curve. The structure assigned to the dataset must contain the
%    following fields:
%      time - time (MJD, ordered)
%      rate - count rate at 'time'
%
%    NOTE: in case you add the Doppler shift of orbital motion to the
%          model using the 'dopplerorbit' fit-function, make sure that
%          this fit-function is calculated *first* in order to
%          transform the times ('lo' parameter of the fit-function
%          /and/ light curve 'time' grid) into the barycenter of the
%          binary! These corrected times are shared among these
%          fit-functions via the ISISscripts caching extension. Then
%          the reference time, t0, is interpreted as binary corrected!
%
%    NOTE: the full Ghosh & Lamb accretion torque theory provides
%          equations for the torque strength, b, as a function of the
%          neutron star parameters. This is not implemented here for
%          simplicity. The full theory is implemented in the
%          'pulsarGL79' fit-function.
%
%    NOTE: the model is calculated on the light curve time grid in
%          order to ensure that the integration includes the reference
%          time, t0. This model is interpolated onto the period time
%          grid in the end.
%\example
%    % define the measured period evolution as ISIS dataset
%    id = define_counts(time, make_hi_grid(time), period, period_err);
%
%    % assign the light curve to this dataset
%    set_dataset_metadata(id, struct {
%      time = lc.time,
%      rate = lc.rate
%    });
%
%    % examples for setting the fit-function
%    fit_fun("pulsartorque(1)");
%    fit_fun("dopplerorbit(1)*pulsartorque(1)"); % add orbital Doppler shift
%\seealso{pulsarGL79, dopplerorbit, set_dataset_metadata, fitfun_cache}
%!%-
{
  % input parameters
  variable p0 = par[0], t0 = par[1], a = par[2]*86400, b = par[3]*86400,
           alpha = par[4], Lnorm = par[5], Pnorm = par[6];
  if (Pnorm < 4*DOUBLE_EPSILON) { Pnorm = p0; }

  % get and check meta-structure
  variable meta = get_dataset_metadata(Isis_Active_Dataset);
  if (typeof(meta) != Struct_Type) {
    throw InvalidParmError, "metadata-structure is not defined!"$;
  }
  if (sum(array_map(Integer_Type, &struct_field_exists, meta,["time", "rate", "flux"])) < 2) {
    throw InvalidParmError, "metadata-structure does not contain required fields!"$;
  }

  % define variables
  variable lcrate;
  variable lctime = @(meta.time);
  if (not struct_field_exists(meta, "rate") && struct_field_exists(meta, "flux")) {
    lcrate = @(meta.flux) / Lnorm; % backward compatibility
  } else  { lcrate = @(meta.rate) / Lnorm; }
  if (lo[0] < meta.time[0] || lo[-1] > meta.time[-1]) {
    throw InvalidParmError, "light curve does not include pulse period data"$;
  }

  % define cache
  variable cache = fitfun_get_cache();

  % get binary corrected times
  variable orbitcache = fitfun_get_cache("dopplerorbit");
  variable binarytime = lo;
  if (orbitcache != NULL) {
    if (orbitcache.binarytime != NULL) {
      binarytime = orbitcache.binarytime;
    }
    if (struct_field_exists(orbitcache.metacor, "time") && orbitcache.metacor.time != NULL) {
      lctime = orbitcache.metacor.time;
    }
  }

  % calculate intrinsic spin period evolution by solving the
  % differential equation on the light curve time grid. this is
  % necessary because the reference time might be outside of the
  % period time range -> using the light curve we still can predict
  % the period at these times. a user-defined time grid using
  % set_eval_grid_method does not work because ISIS rebins (=sums up!)
  % the resulting model onto the data time grid.
  if (b != 0. && (cache.spin == NULL || any(cache.lastpar != par))) {
    cache.spin = solveODEbyIntegrate(
      &_ode_torque, lctime; t0 = t0, x0 = p0, qualifiers = struct {
        p0 = Pnorm, a = a, b = b, alpha = alpha,
        lc = struct { time = lctime, rate = lcrate }
      }
    );
  }
  % b = 0
  else if (cache.spin == NULL || any(cache.lastpar[[:2]] != par[[:2]])) {
    cache.spin = a*(lctime - t0) + p0;
%    if (errorprop) { meta.errorprop = 0; }
  }

  
      % % error propagation
      % if (errorprop) {
      %   variable dt = diff(L.time); dt = [dt, mean(dt)];
      %   _for l (0, length(L.time)-1, 1) {
      %     if (l == n) { dLint[l] = 0.; }
      %     else {
      % 	    i = (n < l ? [n:l] : [l:n]);
      %       dLint[l] = sqrt(sum(sqr(
      %         .5 * (pit[i]/p0)^2 * sign(L.flux[i])*abs(L.flux[i])^(alpha-1) * L.dflux[i] * dt[i]
      %       )));
      % 	  }
      % 	}
      %   dLint = abs(dLint) / Lnorm^(alpha-1);
      % 	% period uncertainties
      % 	meta.errorprop = [interpol(lot, L.time,  b*alpha/Lnorm * dLint*86400)];
      % }
      % cache
      % if (cache) { meta.cache.pit = pit; }

  % update cached parameters
  cache.lastpar = par;

  % return model on the data time grid (as an array always!)
  return [ interpol(binarytime, lctime, cache.spin) ];
}


% add fit-function
add_slang_function("pulsartorque", [
  "p [s]", "t0 [MJD]", "a [s/s]", "b [s/s @ (Lnorm,Pnorm)]", "alpha", "Lnorm", "Pnorm [s]"
]);
set_function_category("pulsarorbit", ISIS_FUN_ADDMUL);


% define parameter defaults
private define pulsartorque_defaults(i) {
  return [
    struct { value = 100.0, freeze = 0, hard_min = 0.0, hard_max = 10000, min = 0, max = 300, step = 1e-6, relstep = 1e-8 },
    struct { value = 56000.0, freeze = 1, hard_min = 15000, hard_max = 100000, min = 15000, max = int(ceil(UNIXtime2MJD(_time))), step = 1e-1, relstep = 1e-3 },
    struct { value = 0.0, freeze = 1, hard_min = -1e-7, hard_max = 1e-7, min = 0, max = 1e-7, step = 1e-10, relstep = 1e-12 },
    struct { value = 0.0, freeze = 1, hard_min = 0, hard_max = 1e-5, min = 0, max = 1e-5, step = 1e-10, relstep = 1e-12 },
    struct { value = 6./7., freeze = 1, hard_min = 0, hard_max = 10, min = 6./7., max = 1., step = 1e-2, relstep = 1e-3 },
  struct { value = 1., freeze = 1, hard_min = 0, hard_max = DOUBLE_MAX, min = 0, max = 0, step = 0, relstep = 0 },
  struct { value = 0.0, freeze = 1, hard_min = 0, hard_max = 10000, min = 0, max = 300, step = 0, relstep = 0 }
  ][i];
}
set_param_default_hook("pulsartorque", &pulsartorque_defaults);


% setup the fit-function's cache
fitfun_init_cache("pulsartorque", &pulsartorque_fit, struct { spin = NULL, lctime, lastpar });


% chi-square statistics taking the model uncertainties into account
define pulsarorbit_fit_constraint(stat, pars) {
  vmessage("error (%s): is being re-implemented", _function_name);
  return;
  % variable dmdl = array_flatten(array_struct_field(array_map(
  %   Struct_Type, &get_dataset_metadata, all_data
  % ), "errorprop"));
  % variable data = array_map(
  %   Struct_Type, &get_data_counts, all_data
  % );
  % variable err = array_flatten(array_struct_field(data, "err"));
  % data = array_flatten(array_struct_field(data, "value"));
  % variable mdl = array_flatten(array_struct_field(array_map(
  %   Struct_Type, &get_model_counts, all_data
  % ), "value"));
  % return sum(sqr(data - mdl) / (sqr(err) + sqr(dmdl)));
}

%
% Code from src/data/pulsar/pulse2pulse_flux_lc.sl
%

%%%%%%%%%%%%%%%%%%%%%%%%
define pulse2pulse_flux_lc()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulse2pulse_flux_lc}
%\synopsis{averages the given lightcurve over the pulsar's period}
%\usage{Struct_Type = pulse2pulse_flux_lc(
%      Struct_Type lightcurve,
%      Double_Type or Struct_Type period,
%      [, Struct_Type orbit]
%    );}
%\qualifiers{
%    \qualifier{remap}{interpolate the resulting flux lightcurve
%               on the input time grid (default: no)}
%    \qualifier{interpol}{function reference to perform the time
%               interpolation (default: &interpol_points)}
%    \qualifier{dphitol}{minimum phase coverage (dphi) a pulse in
%               the lc has to have at least (default: .95)}
%    \qualifier{gaptol}{minimum time difference between bins which
%               defines a gap (default: 2*min(diff(lc.time)))}
%    \qualifier{t0}{time of pulse phase zero (can be provided
%               using the ephemeris structure as well;
%               default: first time bin of the lightcurve)}
%    \qualifier{phase}{array of the pulse phases corresponding to
%               the lightcurve time array (default: its
%               calculated using 'pulseperiod2phase'). If
%               the qualifier is set to a reference the
%               calculated phases are assigned to the
%               given variable.}
%}
%\description
%    The underlying slope in a lightcurve of a pulsar
%    might affect any timing analysis of its pulsations
%    due to pulse to pulse variations of the luminosity.
%    This functions averages the count rates over each
%    pulse to get this luminosity dependance, which is
%    afterwards interpolated to the binning of the input
%    lightcurve.
%
%    The input lightcurve must be of struct {
%      Double_Type[] time, rate, error
%    }
% 
%    The input pulse period may be either a single number
%    implying a constant pulse period or a structure
%    containing the pulse ephemeris as defined in
%    'check_pulseperiod_orbit_struct'. In addition, the
%    orbital parameters might be also given as a
%    structure to take also the Doppler shift of the
%    pulse period into account. This is necessary only
%    in case of a non binary corrected lightcurve.
%
%    The output lightcurve is a struct {
%      Double_Type[] time (center of pulse), rate, error,
%      dphi (phase coverage of each pulse, <1 for a gap)
%    }
%
%    By default, bad sampled pulses are ignored in the
%    final lightcurve, which would lead to wrong mean
%    count rates otherwise.
%
%    Note, that all parameters must have the same time
%    unit, if applicable.
%
%    Note further, that the lightcurve's time grid has
%    to be evenly spaced!
%\seealso{pulseperiod2phase, interpol_points}
%!%-
{
  variable lc, period, orb = NULL;
  switch(_NARGS)
    { case 2: (lc,period) = (); }
    { case 3: (lc,period,orb) = (); }
    { help(_function_name()); return; }

  % convert a constant given period into a ephemeris structure
  variable t0 = qualifier("t0", lc.time[0]); % time of phase zero
  if (typeof(period) == Double_Type) {
    period = struct { p0 = period, t0 = t0 };
  }

  % check input structures
  if (not check_pulseperiod_orbit_struct(period)) {
    vmessage("error (%s): pulse period is not given properly", _function_name);
    return;
  }
  if (orb != NULL && not check_pulseperiod_orbit_struct(orb)) {
    vmessage("error (%s): orbit is not given properly", _function_name);
    return;
  }

  % calculate the pulse phase at each time of the lightcurve
  variable phiqual = qualifier("phase", NULL);
  variable phi = typeof(phiqual) == Array_Type ? phiqual : pulseperiod2phase(lc.time, period; sameunit);
  if (typeof(phiqual) == Ref_Type) { @(phiqual) = phi; }
  if (length(phi) != length(lc.time)) {
    vmessage("error (%s): length of phase grid does not match the time grid", _function_name);
    return;
  }

  % initialize output structure
  variable sl = struct {
    time = Double_Type[0], rate = Double_Type[0],
    error = Double_Type[0], dphi = Integer_Type[0]
  };

  % loop over lightcurve
  variable ci = 0; % first index of current pulse
  variable phi0 = phi[0]; % phase of first index of current pulse
  variable gaptol = qualifier("gaptol", 2.*min(diff(lc.time)));
  variable i, ix, frac, bt, ct, gap;
  variable last = struct { ct = lc.time[0], frac = 0, bt, rate = 0, error = 0, gap = 1 };
  _for i (1, length(lc.time)-1, 1) {
    % check on a gap (end of the lightcurve is equal to a gap as well)
    % note that gap=0 if a gap is detected
    gap = not (lc.time[i] - lc.time[i-1] > gaptol || i == length(lc.time)-1);
    % -> multiply last index fraction by zero, see below
    % check if current phase is greater than phi0+1 or a gap was detected
    % -> one full pulse -> average flux and add to output
    if (phi[i] >= phi0+1 || gap==0) {
      % indices completely within the pulse
      ix = [ci:i-1];
      % current index is already within the next pulse
      % -> fraction to the current pulse
      frac = (phi[ci]+1. - phi[i-1]) / (phi[i] - phi[i-1]) * gap; % note gap=0 here
      % time of phase+1
      ct = lc.time[i-1] + frac*(lc.time[i]-lc.time[i-1]);
      % time binning of the indices (including last one)
      bt = diff([lc.time[ix], ct]);
      % time binning of the last index of the former pulse
      last.bt = (lc.time[ci]-last.ct)*last.gap; % note last.gap=0 here
      % add to output and take former pulse into account
      sl.time  = [sl.time, last.ct + .5*(last.bt + sum(bt))];
      sl.rate  = [sl.rate, weighted_mean([last.rate, lc.rate[ix]], [last.bt, bt])];
      sl.error = [sl.error, sqrt(sqr(last.error*last.bt)+sum(sqr(lc.error[ix]*bt)))
		  / (last.bt+sum(bt))]; % error propagation of weighted_mean
      sl.dphi  = [sl.dphi, gap ? 1. : phi[i-1] - phi0]; % phase coverage
      % resets first index and remember last index properties, note gap=0 here
      ci = i;
      phi0 = (gap ? phi0+1 : phi[i]); % note gap==1 means _no_ gap
      last.ct = (gap ? ct : lc.time[i]);
      last.frac = (1.-frac)*gap;
      last.rate = lc.rate[i-1]*gap;
      last.error = lc.error[i-1]*gap;
      last.gap = gap;
    }
  }

  % ignore bad covered pulses
  variable dphitol = qualifier("dphitol", .95);
  if (dphitol > 0) {
    struct_filter(sl, where(sl.dphi >= dphitol));
  }
  
  % interpolate time grid
  if (qualifier_exists("remap"))
  {
    variable intfun = qualifier("interpol", &interpol_points);
    sl.rate = @intfun(lc.time, sl.time, sl.rate);
    % estimate uncertainties
    variable oldslerr = @sl.error, n;
    sl.error = Double_Type[length(lc.time)];
    _for n (0, length(lc.time)-1, 1) {
      sl.error[n] = oldslerr[where_min(abs(sl.time - lc.time[n]))[0]];
    }
    sl.time = lc.time;
  }
  
  return sl;
}

%
% Code from src/data/pulsar/pulseperiod.sl
%
% -*- mode: slang; mode: fold -*-

%%%%%%%%%%%%%%%%%%%%%%%%
define check_pulseperiod_orbit_struct()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{check_pulseperiod_orbit_struct}
%\synopsis{checks if the pulse period- and/or orbit-structure are valid}
%\usage{Integer_Type check_pulseperiod_orbit_struct(Struct_Type structure);
% or Struct_Type check_pulseperiod_orbit_struct(Double_Type pulseperiod);}
%\description
%    This function performs a checks whether the given
%    structure matches one of the following definitions
%    of a pulse period- or orbital parameter structure:
%
%    1) pulseperiod struct { % given as period over time
%        Double_Type[] time, % in MJD
%        Double_Type[] period % in seconds
%       }
%
%    2) pulseperiod struct { % given as taylor coefficients
%        Double_Type t0, % reference time in MJD
%        Double_Type p0[, % pulse period at t0 in seconds
%        Double_Type pdot[, % first derivative in s/s
%        Double_Type p2dot[, % second derivative in s/s^2
%        ...
%        Double_Type pNdot]]] % higher orders
%       }
%
%    3) orbit struct {
%        Double_Type tau or t90, % time of periastron passage
%                    % or mean longitude of 90 degrees in MJD
%        Double_Type porb, % orbital period in days
%        Double_Type asini, % projected semi-major axis in lt-s
%        Double_Type ecc[, % eccentricity
%        Double_Type omega] % longitude of periastron in degrees
%                                          % required if ecc > 0     
%       }
%
%    On a successful match the number of the definition
%    (1-3) is returned, 0 otherwise.
%    Note that additional field names are allowed.
%    
%    If an input pulse period (as a number) is given
%    instead a structure, a new pulse period structure
%    (type 2) is returned with t0 set to 0 and the
%    pulse period as given.
%
%    Note that units cannot checked by this function, but
%    are a suggestion. Read the help of any function using
%    the same definition on their required units.
%\seealso{pulseperiod}
%!%-
{
  variable s;
  switch(_NARGS)
    { case 1: (s) = (); }
    { help(_function_name()); return; }

  % return value depends on input type
  switch (typeof(s))
    % new pulse period structure
    { case Double_Type: return struct { p0 = s, t0 = 0. }; }
    % check input structure
    { case Struct_Type:
      variable fields = get_struct_field_names(s);

      %%% check required field names on existence
      % type 1
      if (length(complement(["time","period"], fields)) == 0) {
        if (all([typeof(s.time), typeof(s.period)] == Array_Type)) {
          if (all(s.period > 0)) { return 1; }
	  else {
	    vmessage("warning (%s): 'time' and 'period' have to be greater zero", _function_name);
	  }
        }
        vmessage("warning (%s): 'time' and 'period' have to be arrays", _function_name);
      }
      % type 2
      else if (length(complement(["t0","p0"], fields)) == 0) {
        if (all([typeof(s.t0), typeof(s.p0)] == Double_Type)) {
 	  if (s.p0 > 0) { return 2; }
	  else {
	    vmessage("warning (%s): 't0' and 'p0' have to be greater zero", _function_name);
	  }
        }
        vmessage("warning (%s): 't0' and 'p0' have to be numbers", _function_name);
      }
      % type 3
      else if (length(complement(["porb","asini","ecc"], fields)) == 0
 	       && length(complement(["tau", "t90"], fields)) == 1) {
        if (all([typeof(s.porb), typeof(s.asini), typeof(s.ecc)] == Double_Type)) {
	  if (s.porb < 0 or s.asini < 0) {
            vmessage("warning (%s): 'porb' and 'asini' have to be greater zero", _function_name);
	    return 0;
	  }
          if (s.ecc != 0) {
	    ifnot (0 < s.ecc < 1) {
	      vmessage("warning (%s): 'ecc' has to be between 0 and 1", _function_name);
	      return 0;
	    }
	    if (any(fields == "omega")) {
              if (typeof(s.omega) == Double_Type) { return 3; }
	      vmessage("warning (%s): 'omega' has to be a number", _function_name);
	      return 0;
	    }
            vmessage("warning (%s): 'omega' has to exist because 'ecc' > 0", _function_name);
	    return 0;
          }
	  return 3;
        }
        vmessage("warning (%s): 'porb', 'asini', 'ecc', 'tau'/'t90' have to be numbers", _function_name);
      }
     
      return 0;
    }
    % neither structure nor pulse period given
    { vmessage("warning (%s): input has to be a structure or a single pulse period", _function_name); return 0; }
  % end of switch
  
  vmessage("error (%s): this should not happen", _function_name);
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define pulseperiod_transform()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{pulseperiod_transform}
%\synopsis{transforms a period evolution given as taylor coefficients to a new t0}
%\usage{Struct_Type pulse_transform(Double_Type new_t0, Struct_Type pulseperiod);}
%\qualifiers{
%    \qualifier{sameunit}{set if the unit of 'time' is in the same
%               unit as the pulse period or set to the
%               conversion factor from days to 'time'}
%}
%\description
%    The structure containing the pulse period and its
%    derivatives at a certain time t0 must fullfil the
%    conditions given in 'check_pulseperiod_orbit_struct'.
%    This description of the pulse period evolution is
%    transformed to a new given t0. In case this time is
%    given as an array an array of transformed structures
%    is returned.
%
%    Note, that the uncertainty of the transformed pulse
%    ephemeris scales with (t - new_t0)^N * pNdot with
%    the highest order N of the taylor series.
%\seealso{check_pulseperiod_orbit_struct}
%!%-
{
  variable t, s;
  switch(_NARGS)
    { case 2: (t,s) = (); }
    { help(_function_name()); return; }

  % check input structure
  if (check_pulseperiod_orbit_struct(s) != 2) {
    vmessage("error (%s): input structure is not a valid type 2 period evolution", _function_name);
    return;
  }

  variable tconv = qualifier("sameunit", 86400.);
  if (tconv == NULL) { tconv = 1.; }

  % get taylor coefficients from structure
  variable coeff = taylorcoeff_from_struct(s, "p0", "p\([0-9]*\)dot"R);

  % loop over input times and build output structures
  variable i, fnames = [
    "p0", length(coeff) > 1 ? "pdot" : String_Type[0],
    array_map(String_Type, &sprintf, "p%ddot", [2:length(coeff)-1])
  ];
  variable f, snew = struct_array(length(t), struct_combine(struct { t0 }, fnames));
  _for i (0, length(t)-1, 1) {
    % transform coefficients
    snew[i].t0 = t[i];
    _for f (0, length(coeff)-1, 1) {
      set_struct_field(snew[i], fnames[f],
        taylor((t[i]-s.t0)*tconv, coeff[[f:]])[0]
      );
    }
  }

  return typeof(t) == Double_Type ? snew[0] : snew;
} %}}}


%%%%%%%%%%%%%%%%%%%%%%%%
define pulseperiod()
%%%%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{pulseperiod}
%\synopsis{calculates the pulse period at the given time depending
%    on the given pulse period evolution}
%\usage{Double_Type[] pulse_period(Double_Type[] time, Struct_Type pulseperiod[, Struct_Type orbit]);}
%\qualifiers{
%    \qualifier{interpol}{reference to a function to interpolate the
%               pulse period evolution on the requested
%               time grid (default: &interpol_points)}
%    \qualifier{sameunit}{set if the unit of 'time' is in the same
%               unit as the pulse period or set to the
%               conversion factor from days to 'time'}
%}
%\description
%    Calculates the expected pulse period at the given time
%    (in days), which may be a single value or an array.
%    The structures containing the pulse period and the
%    optional orbital parameters must follow the conditions
%    described in 'check_pulseperiod_orbit_struct'.
%
%    If the pulse period is given as evolution (time vs.
%    period) the period at the requested time is calculated
%    by interpolation. Otherwise the pulse period is
%    calculated by a 'taylor' series using a given period
%    and its derivatives at a reference time in. In any
%    case the requested time has to be in MJD and the period
%    (and its Nth derivative) in seconds (/seconds^N)
%
%    Finally, if any orbital parameters are provided, the
%    returned period gets modified by the binary motion.
%\seealso{check_pulseperiod_orbit_struct, taylor, radial_velocity_binary}
%!%-
{
  variable t, period, orb = NULL;
  switch(_NARGS)
    { case 2: (t,period) = (); }
    { case 3: (t,period,orb) = ();}
    { help(_function_name()); return; }

  % check input structures
  variable type = check_pulseperiod_orbit_struct(period);
  if (typeof(type) == Struct_Type) {
    period = type; type = 2;
  } else {
    if (type == 0) {
      vmessage("error (%s): period is not given properly", _function_name);
    }
  }
  if (orb != NULL && check_pulseperiod_orbit_struct(orb) == 0) {
    vmessage("error (%s): orbit is not given properly", _function_name);
  }

  variable retper;

  variable tconv = qualifier("sameunit", 86400.);
  if (tconv == NULL) { tconv = 1.; }

  % period over time given
  if (type == 1) {
    variable inter = qualifier("inter", &interpol_points);
    retper = @inter(t, period.time, period.period);
  }
  % taylor coefficients given - get those from the structure
  else {
    variable coeff = taylorcoeff_from_struct(period, "p0", "p\([0-9]*\)dot"R);
    retper = taylor((t-period.t0)*tconv, coeff);
  }

  % take orbit into account
  if (orb != NULL) {
    % build qualifier-structure
    variable qual = struct { P = orb.porb, e = orb.ecc, asini = orb.asini };
    if (struct_field_exists(orb, "tau")) {
      qual = struct_combine(qual, struct { T0 = orb.tau });
    } else { qual = struct_combine(qual, struct { T90 = orb.t90 }); }
    if (struct_field_exists(orb, "omega")) {
      qual = struct_combine(qual, struct { omega = orb.omega, degrees });
    }
    retper *= (1. + radial_velocity_binary(t;; qual));
  }

  return retper;
} %}}}

%
% Code from src/data/pulsar/pulseperiod2phase.sl
%

%%%%%%%%%%%%%%%%%%%%%%%%
define pulseperiod2phase()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulseperiod2phase}
%\synopsis{calculates the pulse phase from a given pulse period}
%\usage{Double_Type[] pulseperiod2phase(
%      Double_Type[] time,
%      Struct_Type pulseperiod[, Struct_Type orbit]
%    );}
%\qualifiers{
%    \qualifier{interpol}{interpolation method to re-map
%               the period- onto the input time-
%               grid (default: &interpol_points)}
%    \qualifier{getphi}{variable reference to return the
%               phase on the given period over time
%               grid (method (a), see below).}
%    \qualifier{sameunit}{set if the unit of 'time' is in the same
%               unit as the pulse period or set to the
%               conversion factor from days to 'time'}
%}
%\description
%    The pulse period p(t) of a pulsar and its pulse
%    phase phi(t) are connected by
%      dphi(t) / dt = 1. / p(t)
%    Thus, from a given pulse period the phase can be
%    calculated by integration, which this function
%    provides.
%
%    The pulse period may be given as
%      a) the pulse period over time
%         struct { time, period }
%      b) the taylor coefficients of the pulse
%         ephemeris via
%         struct { p0[, t0, pdot[, p2dot]] }
%    as defined in 'check_pulseperiod_orbit_struct'.
%
%    It is assumed that the unit of the pulse period
%    is a factor of 86400 larger than the time unit of
%    the light curve, i.e., light curve in days and
%    pulse period in seconds. If both have the same
%    unit (regardless whether its seconds or days)
%    use the sameunit-qualifier.
%    
%    In case a), the phase is numerically integrated
%    using the trapez method on the full provided period
%    time grid. Finally, the calculated phase is re-mapped
%    onto the input time grid by interpolation.
%
%    In case b), the phase is calculated analytically by
%    a taylor-series:
%      phi(t) = (t-t0)/p0 - pdot/p0^2*(t-t0)^2 + ...
%
%    Note, that the analytical calculation only supports
%    a pulse ephemeris up the the second order (p2dot).
%    If you need higher orders, you can first calculate
%    the pulse period over time via a taylor-series and
%    finally use method a).
%
%    In case orbital elements are provided, the additional
%    phase shift is calculated after Hilditch Eq 3.43:
%      delta phi = z(t)/c (f0 zdot(t)/c - f(t-t0))
%    with the projected position z(t) of the neutron star
%    and its spin frequency evolution f(t) = 1 / p(t),
%    where p(t) is the given pulse period evolution.
%\seealso{pulse_period, taylor, BinaryPos, radial_velocity_binary}
%!%-
{
  variable t, p, orb = NULL;
  switch (_NARGS)
    { case 2: (t,p) = (); }
    { case 3: (t,p,orb) = (); }
    { help(_function_name); return; }

  % determine input type and sanity checks
  if (typeof(p) != Struct_Type) {
    vmessage("error(%s): pulse period must be of Struct_Type", _function_name);
    return;
  }
  variable type = check_pulseperiod_orbit_struct(p);
  if (type == 1) {
    if (t[0] < p.time[0] or t[-1] > p.time[-1]) {
     vmessage("error(%s): pulse period over time does not cover the required time range", _function_name);
      return;
    }
  } else if (type == 2) {
    if (struct_field_exists(p, "p3dot")) {
      vmessage("error(%s): analytical solution only accounts orders up to p2dot", _function_name);
      return;
    }
  } else {
    vmessage("error(%s): pulse period over time is not given properly", _function_name);
    return;
  }
  if (orb != NULL && check_pulseperiod_orbit_struct(orb) != 3) {
    vmessage("error(%s): orbital parameters are not given properly", _function_name);
  }

  variable tconv = qualifier("sameunit", 86400.);
  if (tconv == NULL) { tconv = 1.; }

  % calculate neutron star position and its radial velocity if required
  variable zpos = 0., vrad = 0.;
  if (orb != NULL) {
    zpos = BinaryPos(t; asini = orb.asini, porb = orb.porb, t0 = orb.tau,
		     eccentricity = orb.ecc, omega = orb.omega);
    vrad = radial_velocity_binary(t; asini = orb.asini, P = orb.porb, T0 = orb.tau,
				  e = orb.ecc, omega = orb.omega, degrees);
  }

  % period over time given
  if (type == 1) {
    % calculate pulse phase by integrating 1/p(t)
    variable i, phi = Double_Type[length(p.time)];
%    variable integrate = qualifier("integrate", &integrate_trapez);
    _for i (1, length(phi)-1, 1) {
      % trapez (or triangle?) method
      % using integrate_trapez is much slower because of the loop
      phi[i] = phi[i-1] +
               .5*tconv*(p.time[i]-p.time[i-1])*(1./p.period[i]+1./p.period[i-1]);
%      phi[i] = @integrate(
%        p.time[[:i]], 1./p.period[[:i]]
%      );
    }

    if (qualifier_exists("getphi")) {
      @(qualifier("getphi")) = phi;
    }
    
    % interpolate on input time grid
    variable interpol = qualifier("interpol", &interpol_points);
    return @interpol(
      t, p.time, phi
    ) + vrad;
  }

  % period ephemeris given
  if (type == 2) {
    variable p0 = p.p0;
    variable t0 = struct_field_exists(p, "t0") ? p.t0 : t[0];
    variable pdot = struct_field_exists(p, "pdot") ? p.pdot : 0;
    variable p2dot = struct_field_exists(p, "p2dot") ? p.p2dot : 0;

    return taylor((t-t0)*tconv, [0, 1./p0, -pdot/p0^2, (2*p0*pdot^2 - p2dot*p0^2)/(p0^4)]);
  }

  vmessage("error(%s): case unknown, this should not happen", _function_name);
  return;
}

%
% Code from src/data/pulsar/pulseperiod_epfold.sl
%

%%%%%%%%%%%%%%%%%%%%%%%%
define pulseperiod_epfold() {
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulseperiod_epfold}
%\synopsis{UNDER DEVELOPMENT; performs an automatic pulse period search on the given light curve(s)}
%\usage{Struct_Type pulseperiod_epfold(Struct_Type[] lc, Double_Type p0);}
%\qualifiers{
%    \qualifier{nbins}{profile bins (default: 32)}
%    \qualifier{fracexp}{minimum fracexp all bins should have (default: 1.0)}
%    \qualifier{dpscale}{period search range relative to formal resolution
%               of epfold (default: 3)}
%    \qualifier{dpmin}{minimum period search range (default: 0.01 s)}
%    \qualifier{gapscale}{factor for maximum allowed gap length relative to
%               formal resolution o f epfold (default: .5)}
%    \qualifier{goodness}{threshold for the goodness of any signal
%               (default: 3)}
%    \qualifier{pbins}{mininum number of consecutive bins in epfold with
%               'goodness' to define a signal (default: 3)}
%    \qualifier{chatty}{chattiness (default: 1)}
%    \qualifier{plotlc}{reference to function(Struct_Type[] lc)}
%    \qualifier{plotepf}{reference to
%               function(Struct_Type epf, Double_Type median, norm)}
%}
%\description
%    UNDER DEVELOPMENT, USE WITH CAUTION! Send questions or bugs to
%    matthias.kuehnel@sternwarte.uni-erlangen.de
%
%    lc[] = struct { time, rate, error, fracexp }
%    with time in MJD
%
%    p0 = pulse period in seconds
%    
%    returns struct {
%      time (mean in MJD)
%      period (in s)
%      error (in s)
%      epfold (structure returned by epfold)
%      lc (input light curves, but maybe splitted)
%    }    
%\seealso{epfold}
%!%-
  variable lc, p0;
  switch (_NARGS)
    { case 2: (lc, p0) = (); }
    { help(_function_name); return; }

  % sanity checks
  if (typeof(lc) != Array_Type) { lc = [lc]; }
  if (_typeof(lc) != Struct_Type) { vmessage("error(%s): light curves have to be an array of structures!", _function_name); return; }
  if (typeof(p0) != Double_Type) { vmessage("erros(%s): pulse period has to be a floating point number!", _function_name); return; }
  if (p0 <= 0) { vmessage("error(%s): pulse period has to be > 0", _function_name); return; }

  % pre-process input
  variable nlc = length(lc);
  p0 /= 86400; % s -> d
  lc = array_map(Struct_Type, &struct_combine, COPY(lc), struct { split = 0, splitnum = 0 });
  % qualifiers
  variable nbins = qualifier("nbins", 32);
  variable fracexp = qualifier("fracexp", 1.);
  variable sdp = qualifier("dpscale", 4.);
  if (sdp < 3.) { vmessage("warning(%s): setting sdp=3 (minimum recommended value)", _function_name); sdp = 3.; }
  variable dpmin = qualifier("dpmin", .01/86400);
  if (dpmin < 0) { vmessage("error(%s): dpmin has to be > 0",  _function_name); return; }
  variable gs = qualifier("gapscale", .5);
  variable qsig = qualifier("goodness", 3.);
  variable nsig = int(qualifier("pbins", 3));
  variable chatty = qualifier("chatty", 1);
  
  % light curve properties and eventually split them
  if (chatty) { vmessage("condition checks"); }
  variable lcp = struct {
    time = Double_Type[nlc], % mean time
    dt = Double_Type[nlc], % time resolution
    len = Double_Type[nlc], % length
    bins = Integer_Type[nlc], % number of time bins
    dp = Double_Type[nlc], % formal epfold period resolution
  };
  % loop over light curve(s) (if splitted below until all segments were looped)
  variable i = 0;
  while (i < nlc) {
    % field checks
    if (any(array_map(Integer_Type, &struct_field_exists, lc[i], ["time", "rate", "error", "fracexp"]) == 0)) { vmessage("erros(%s): light curve does not have all required fields!"); return; }
    % filter on fractional exposure
    struct_filter(lc[i], where(lc[i].fracexp >= fracexp));
    % determine properties
    lcp.dt[i] = lc[i].time[1]-lc[i].time[0];
    lcp.len[i] = lc[i].time[-1]-lc[i].time[0] + lcp.dt[i];
    lcp.bins[i] = length(lc[i].time);
    lcp.time[i] = lc[i].time[0] + .5*lcp.len[i];
    lcp.dp[i] = p0^2 / lcp.len[i];
    % check on gaps and split light curve
    variable splitted = 0;
    if (lcp.len[i] - lcp.dt[i]*lcp.bins[i] > lcp.dt[i]) {
      variable gt = gs*(p0^2/lcp.dp[i] - p0); % gap threshold
      if (chatty) { vmessage("  light curve [%d] contains gaps (threshold %.0f s)", i, gt*86400); }
      variable slc = split_lc_at_gaps(lc[i], gt);
      if (qualifier_exists("plotlc")) {
	(@(qualifier("plotlc")))(slc);
      }
      % splitted
      if (length(slc) > 1) {
        splitted = 1;
        slc[0].splitnum = length(slc)-1;
        if (chatty) { vmessage("    splitted into %d segments", slc[0].splitnum+1); }
        variable j;
        _for j (1, length(slc)-1, 1) {
          slc[j].split = j;
          slc[j].splitnum = slc[0].splitnum;
        }
        % insert into main array
        lc = [
          i>0 ? lc[[:i-1]] : Struct_Type[0], % up to i-1
          slc, % splitted ones
          lc[[i+1:]] % from i+1
        ];
        % increase number of light curves and property structure
        struct_filter(lcp, [[0:nlc-1], ones(slc[0].splitnum)*0]);
        nlc += slc[0].splitnum;
      }
    }
    % go on
    if (splitted == 0) { i++; }
  }
  
  % prepare output structure
  variable out = struct {
    time = Double_Type[nlc], period = Double_Type[nlc],
    error = Double_Type[nlc], epfold = Struct_Type[nlc],
    lc = lc
  };
  
  % epoch folding
  if (chatty) { vmessage("epoch folding"); }
  _for i (0, nlc-1, 1) {
    if (qualifier_exists("skiphack")) {
      if (any(qualifier("skiphack") == i)) { continue; }
    }
    if (p0-lcp.dp[i]*sdp < 0 || p0+lcp.dp[i]*sdp > lcp.len[i]) {
      vmessage("warning(%s): period search range too large, skipping light curve [%d]", _function_name, i);
      continue;
    }
    % epoch folding
    variable psr = _max(lcp.dp[i]*sdp, dpmin); % period search range
    variable epf = epfold(
      lc[i].time, lc[i].rate, p0-psr, p0+psr;;
      struct_combine(struct {
	nbins = nbins, dt = lcp.dt[i]*lc[i].fracexp, exact, chatty = -1
%	nsrch = int(lcp.dp[i]*sdp*2 / lcp.dt[i] * 10)
      }, __qualifiers)
    );
    % normalize
    variable n = where(abs(epf.p - epf.p[where_max(epf.stat)]) > lcp.dp[i]);
    variable med = median(epf.stat[n]);
    variable norm = sqrt(sum(sqr(epf.stat[n] - med))*1./(length(n)-1));
    if (qualifier_exists("plotepf")) {
      (@(qualifier("plotepf")))(epf, med, norm);
    }
    % save into output
    out.epfold[i] = struct_combine(epf, struct { median = med, norm = norm });
    % where normalized chi^2 > threshold
    variable w = (epf.stat - med) / norm >= qsig;
    % number of consecutive bins where chi^2 > threshold
    variable k, ci = @w;
    _for k (1, length(w)-1, 1) {
      ci[k] = ci[k-1] == 0 ? w[k] : w[k]*(ci[k-1]+1);
    }
    % where consecutive bins > threshold
    w = where(ci > nsig);
    if (length(w) == 0) {
      vmessage("warning(%s): no period signal found in light curve [%d]", _function_name, i);
      continue;
    }
    if (length(w) != max(ci[w])-nsig) { % max(ci) should be equal the expected number of consecutive bins
      vmessage("warning(%s): multiple signals found in light curve [%d]", _function_name, i);
      continue;
    }
    % result
    out.time[i]   = mean(lc[i].time);
    out.period[i] = mean(epf.p[[w[0]-nsig:w[-1]]]) * 86400;
    out.error[i]  = lcp.dp[i] * 86400;
  }

  % sort by time
  n = array_sort(lcp.time);
  struct_filter(lcp, n);
  struct_filter(out, n);
  
  if (qualifier_exists("getlcp")) {
    (@(qualifier("getlcp"))) = lcp;
  }
  
  return out;
}

%
% Code from src/data/pulsar/pulseperiod_magic.sl
%
define pulseperiod_magic() {
%!%+
%\function{pulseperiod_magic}
%\synopsis{UNDER HEAVY DEVELOPMENT; USAGE ON YOUR OWN RISK}
%\usage{Struct_Type pulseperiod_magic(Struct_Type[] lc, Double_Type p0, dpmax);}
%\qualifiers{
%    \qualifier{ccflim}{cross-correlation threshold (default: .7)}
%}
%\description
%    UNDER DEVELOPMENT, USE WITH CAUTION! Send questions or bugs to
%    matthias.kuehnel@sternwarte.uni-erlangen.de
%    see pulseperiod_epfold
%
%    lc[] = struct { time, rate, error, fracexp }
%    with time in MJD
%
%    p0 = pulse period in seconds
%    dpmax = maximum allowed period derivative
%    
%    returns struct {
%      time (mean in MJD)
%      period (in s)
%      error (in s)
%    }    
%\seealso{pulseperiod_epfold, pulseperiod_phase_connect, bayesian_blocks}
%!%-
  variable lc, p0, dpmax;
  switch (_NARGS)
    { case 3: (lc, p0, dpmax) = (); }
    { help(_function_name); return; }

  % epoch folding of all light curves
  % - sanity checks done
  % - also splits at gaps if needed
  % - determines light curve/segment properties (lcp)
  variable lcp;
  variable epfres = pulseperiod_epfold(
    lc, p0;; struct_combine(__qualifiers, struct { getlcp = &lcp })
  );
  lc = epfres.lc;
  variable nlc = length(lc);
  p0 /= 86400; % s -> d

  % qualifiers
  variable nbins = qualifier("nbins", 32);
  variable clim = qualifier("ccflim", .7);

  % prepare output structure
  variable out = struct {
    time = lcp.time[[:-2]], period, error,
    epfold = epfres, % epoch folding result
    pblock, % periods from bayesian blocks
    phasecon % all possible periods from phase connection
  };
  
  % times of constant period using bayesian_blocks
  % periods are in seconds here (as returned by pulseperiod_epfold)
  variable n = where(epfres.period > 0);
  variable pblock = get_blocks_data(bayesian_blocks(struct {
    tt = epfres.time[n],
    cell_data = { epfres.period[n], epfres.error[n] }
  }));
  out.pblock = struct {
    time_lo = pblock.time, time_hi = pblock.time + pblock.dt,
    period = pblock.data, error = pblock.error
  };

  % first iteration period evolution structure
  variable final = struct {
    time = @(lcp.time),
    period = Array_Type[nlc], % arrays as the determined phase shift below
    error = Array_Type[nlc],  % is degenerated by integers
    block = Integer_Type[nlc] % block index
  };
  variable i;
  _for i (0, nlc-1, 1) {
    % use period of the corresponding (bayesian) block
    n = wherefirst(pblock.time <= final.time[i] <= pblock.time+pblock.dt);
    if (n == NULL) {
      vmessage("warning(%s): light curve [%d] outside of the period blocks; assuming nearest period", _function_name, i);
      n = where_min(abs(pblock.time + .5*pblock.dt - final.time[i]))[0];
    }
    final.block[i]  = n;
    final.period[i] = [pblock.data[n]/86400];  % back into days
    final.error[i]  = [pblock.error[n]/86400];
  }
  % cross-correlation
  vmessage("cross correlating");
  % error propagate the period uncertainties into a phase shift
  variable dphi = Double_Type[nlc-1];
  _for i (0, nlc-2, 1) {
    dphi[i] = (final.time[i+1] - final.time[i])
            * (1./(final.period[i][0] - final.error[i][0]) - 1./final.period[i][0]);
  }
%  % loop over all blocks
%  _for i (0, length(pblock.data)-1, 1) {
    % sanity check for min(dphi) < 1 (as we cannot measure the integer part)
    variable start = where_min(dphi)[0];
    if (dphi[start] > 1) {
      % use a qualifier here to continue? i.e. providing a specific period?
      vmessage("error(%s): times between light curves are too large,", _function_name);
      vmessage("resulting in pulse phase uncertainties of > 1, can't cross-correlate!");
      return;
    }
    % sort by distance to start and loop
    variable j;
    foreach j (array_sort(abs(final.time[[:-2]] - final.time[start]))) {
      % left (-1), right side (+1), or at (0) the starting point?
      variable s = sign(final.time[j] - final.time[start]);
      % create pulse profile of current light curve and its right neighbor
      % using a commond reference time and period
      variable k = [j+1, j, j-1][s+1]; % index of period to use
      % we have to use every period estimate
      variable l;
      variable pnew = Double_Type[0], perr = @pnew;  %, pdot = @pnew;
      _for l (0, length(final.period[k])-1, 1) {
	% vmessage("%f %f %f", final.period[k][l], mean(lc[j].time), final.time[j+1]);
        variable ref = pfold(
          lc[j].time, lc[j].rate, final.period[k][l], lc[j].error;
          nbins = nbins, dt = lcp.dt[j]*lc[j].fracexp, exact, t0 = final.time[j]
        );
        variable nei = pfold(
          lc[j+1].time, lc[j+1].rate, final.period[k][l], lc[j+1].error;
          nbins = nbins, dt = lcp.dt[j+1]*lc[j+1].fracexp, exact, t0 = final.time[j]
	);
	% hplot(ref);()=keyinput(;nchrs=1);
	% hplot(nei);()=keyinput(;nchrs=1);	
        % error propagate phase shift again
        dphi[j] = (final.time[j+1] - final.time[j])
                * (1./(final.period[k][l] - final.error[k][l]) - 1./final.period[k][l]);
        % determine phase shift
        variable phi, ccf;
        (phi,ccf) = pulseprofile_phase_connect(nei,ref);
	if (ccf < clim) {
	  vmessage("warning(%s): skipping [%d] (ccf = %.2f)", _function_name, j, ccf); continue;
	}
        % shift if measured shift is larger than expected (only applies if dphi < 1)
        if (abs(phi) > dphi[j]) {
	  phi = phi - [0, sign(phi)];
	  % select the least worse difference
	  phi = phi[where_min(abs(phi))];
	}
        % add integers within expected phase shift range
        phi += [-int(floor(dphi[j])):int(floor(dphi[j]))];
        % calculate new period assuming dot{P} = 0
	variable padd = 1./(phi/(final.time[j+1] - final.time[j]) + 1./final.period[k][l]);
        pnew = [pnew, padd];
        perr = [perr, padd*lcp.dt[j]/(final.time[j+1] - final.time[j])];
      }
      if (length(pnew) == 0) { continue; }
      final.period[j] = pnew;
      final.error[j] = perr;
    }

    % remove last entry from final evolution (we lost one degree of freedom)
    struct_filter(final, [:-2]);
    % copy into output
    out.phasecon = COPY(final);
    % now we have multiple possible period evolutions,
    % first build a list of all possible indice combinations
    variable ind  = list_new;
    list_append(ind, ones(length(final.period))-1); % start with first period indices
    while (not all(ind[-1] == array_map(Integer_Type, &length, final.period)-1)) {
%    while (ind[-1][-1] != length(final.period[-1])-1) {
      % increment
      variable nind = @(ind[-1]);
      j = 0;
      forever {
	nind[j]++;
        if (nind[j] == length(final.period[j])) { nind[j] = 0; j++; }
	else { break; }
      }
      list_append(ind, nind);
    }
    vmessage("checking %d possible solutions", length(ind));
    % check each solution against pdot constraints
    n = Integer_Type[length(ind)];
    variable stat = Double_Type[length(ind)] + _Inf;
    _for j (0, length(ind)-1, 1) {
      pnew = Double_Type[length(final.period)]; perr = @pnew;
      _for k (0, length(pnew)-1, 1) {
	pnew[k] = final.period[k][ind[j][k]];
	perr[k] = final.error[k][ind[j][k]];
      }
      % use a linear regression to estimate pdot
      variable lin = linear_regression(final.time, pnew, perr);
      if (abs(lin.b > dpmax)) { continue; }
      stat[j] = lin.chisq;
      n[j] = 1;
    }
    n = where(n == 1);
    vmessage("  %d left after pdot-constraint", length(n));
    % select set of indices corresponding to best statistic
    n = n[where_min(stat[n])[0]];
    _for k (0, length(pnew)-1, 1) {
      pnew[k] = final.period[k][ind[n][k]];
      perr[k] = final.error[k][ind[n][k]];
    }
    vmessage("  selecting most likelihood solution (chi^2_red = %.2f)", stat[n]);
    % return best solution
    final.period = pnew*86400;
    final.error = perr*86400;
%  }
  out.period = final.period;
  out.error  = final.error;
  
  return out;
}

%
% Code from src/data/timing/pulseperiod_search.sl
%
define pulseperiod_search ( ) %{{{
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulseperiod_search}
%\synopsis{Looks for periodic signal in a lightcurve.}
%\usage{Struct_Type pulseperiod_search( lc [, p0 [, sigma ] ] )};
%\description
%
% The input lightcurve should be a structure of the form
%     lc = struct { time, rate, [ error ], [ fracexp ] }
% If error resp. fracexp are not given, they will be filled with
% sqrt(rate) resp. ones. All fields are arrays of doubles of the
% same length.
% p0 and sigma determine an initial guess for the period and an
% approximate error.
% The output is a structure of the form
%     { ep, pp, lc } 
% Each field is a list, in which each element corresponds to one
% segment of the lightcurve (multiple elements if the lightcurve
% contains considerable gaps).
% ep[i] contains epoch folding information, pp[i] the folded pulse profile
% and lc[i] the lightcurve under consideration.
% The function works the following way: (the functions used are given)
%
%     (1) If no p0 is given, a Fourier method is used to find an
%         initial guess for the period. A splitting is performed first
%         if necessary
%            * split_and_epfold_lc (; split_only), foucalc
%     (2) The input lightcurve is split if it contains gaps.
%            * split_and_epfold_lc (; split_only)
%     (3) A correction for any underlying variation in flux of the
%         lightcurve is performed.
%            * pulse2pulse_flux_lc
%     (4) For each segment an epochfolding is performed.
%            * split_and_epfold_lc
%     (5) The peak in the statistics of epfold is found.
%            * find_peak
%
% Qualifiers for the individual functions can be passed as structures
% with the names given below. See their helps for further information.
%
%\qualifiers{
%\qualifier{compact}{output contains only the essential information}
%\qualifier{epfold_fourier_qu}{contains qualifiers for step (1)}
%\qualifier{fourier_qu}{contains qualifiers for step (1)}
%\qualifier{epfold_split_qu}{contains qualifiers for step (2)}
%\qualifier{pulse2pulse_slope_qu}{contains qualifiers for step (3)}
%\qualifier{epfold_qu}{contains qualifiers for step (4)}
%\qualifier{find_peak_qu}{contains qualifiers for step (5)}
%\qualifier{no_slope_correction}{step (3) is not done}
%\qualifier{no_splitting}{step (2) is not done}
%\qualifier{exact}{forces epfold to use exact}
%\qualifier{not_exact}{forces epfold not to use exact}
%\qualifier{fourier}{fourier is used, even if p0 is given -- can be used to pass sigma}
%\qualifier{chatty}{boolean value (default: 1)}
%
%}
%\seealso{foucalc, split_and_epfold_lc, pulse2pulse_flux_lc, find_peak}
%!%-
{
   variable fourier = 0;
   variable lc;
   variable p0 = -1.;
   variable sigma = DOUBLE_MAX;
   switch (_NARGS)
   { case 1: lc = (); fourier = 1; }
   { case 2: (lc, p0) = (); }
   { case 3: (lc, p0, sigma) = (); }
   { help(_function_name); return; }
   
   % Sanity checks %{{{
   ifnot (typeof(lc) == Struct_Type) {
      vmessage("Error(%s): lightcurve is no Struct_Type", _function_name); return;
   }
   ifnot (struct_field_exists(lc, "time") and struct_field_exists(lc, "rate")) {
      vmessage("Error(%s): lightcurve has to contain fields named time and rate", _function_name); return;
   }
   ifnot (typeof(lc.time)==Array_Type and typeof(lc.rate)==Array_Type and _typeof(lc.time)==Double_Type and _typeof(lc.rate)==Double_Type) {
      vmessage("Error(%s): time and rate must be arrays of doubles", _function_name); return;
   }
   ifnot (length(lc.time) == length(lc.rate)) {
      vmessage("Error(%s): time and rate must have the same length"); return;
   }
   ifnot (struct_field_exists(lc, "error")) {
      vmessage("Warning(%s): lightcurve does not contain a field named error, setting it to sqrt(rate)", _function_name);
      variable err_struct = struct { error = sqrt(lc.rate) };
      lc = struct_combine ( lc, err_struct );
   }
   ifnot (struct_field_exists(lc, "fracexp")) {
      vmessage("Warning(%s): lightcurve does not contain a field named fracexp, setting it to ones", _function_name);
      variable fracexp_struct = struct { fracexp = double(ones(length(lc.time))) };
      lc = struct_combine ( lc, fracexp_struct );
   }
   ifnot (typeof(lc.error)==Array_Type and typeof(lc.fracexp)==Array_Type and _typeof(lc.error)==Double_Type and (_typeof(lc.fracexp)==Double_Type or _typeof(lc.fracexp)==Integer_Type)) {
      vmessage("Error(%s): error and fracexp must be arrays of numbers", _function_name); return;
   }
   ifnot ((length(lc.error) == length(lc.time)) and (length(lc.fracexp) == length(lc.time))) {
      vmessage("Error(%s): error and fracexp must have the length of time and rate", _function_name); return;
   }
   %}}}

   variable fourier_qualifiers = qualifier( "fourier_qu", empty_struct() );
   variable epfold_split_fourier_qualifiers = qualifier( "epfold_fourier_qu", empty_struct() );
   variable epfold_split_qualifiers = qualifier( "epfold_split_qu", empty_struct() );
   variable pulse2pulse_slope_qualifiers = qualifier( "pulse2pulse_slope_qu", empty_struct() );
   variable epfold_qualifiers = qualifier( "epfold_qu", empty_struct() );
   variable find_peak_qualifiers = qualifier( "find_peak_qu", empty_struct() );


   variable no_smoothing = (qualifier_exists("no_slope_correction")) ? 1 : 0;
   variable no_splitting = (qualifier_exists("no_splitting")) ? 1 : 0;
   fourier = (qualifier_exists("fourier")) ? 1 : fourier;
   variable exact = (qualifier_exists("exact")) ? 1 : 0;
   if (exact == 1) { epfold_qualifiers = struct_combine ( epfold_qualifiers, "exact" ); }
   variable not_exact = (qualifier_exists("not_exact")) ? 1 : 0;
   if (not_exact == 1) { epfold_qualifiers = struct_combine ( epfold_qualifiers, "not_exact" ); }
   ifnot (exact * not_exact == 0) { vmessage("Error(%s): you cannot give exact and not_exact", _function_name); return; }
   variable compact = (qualifier_exists("compact")) ? 1 : 0;
   variable chatty = qualifier("chatty", 1);
   
   if (fourier) { %{{{
      if(chatty) {vmessage("(%s): First look for period via Fourier", _function_name);}
      
      variable dt = median ( (lc.time - shift(lc.time,-1))[[0:length(lc.time)-1]] );
      variable split_fourier = split_and_epfold_lc ( lc, 4.*dt; epfold_qualifiers = struct { split_only = 1 });
      variable ii;
      variable f;
      variable fourier_periods = {};
      _for ii (0, length(split_fourier.lc)-1, 1) {
         variable lc_fourier = struct { time = split_fourier.lc[ii].time, rate1 = split_fourier.lc[ii].rate };
         variable ix = 0;
         variable power = 20;
         while (ix == 0) {
            try {
               fourier_qualifiers = struct_combine ( fourier_qualifiers, "compact" );
               f = foucalc ( lc_fourier, nint(2.^power);; fourier_qualifiers);
               ix = 1;
            } catch IndexError: {
               power += -1;
            }
            if (power<5) {
               f  = 0;
               ix = 1;
            }
         }
         if (typeof(f) == Integer_Type) { continue; }
         variable fourier_period = 1. / f.freq[where_max(f.signormpsd1)];
         list_append ( fourier_periods, fourier_period );
      }
      if (length(fourier_periods) == 0) { vmessage("Error(%s): Fourier search failed: No periods found.", _function_name); return; }
      fourier_periods = list_to_array ( fourier_periods );
      p0 = mean ( fourier_periods );
   } %}}}
   
   % prepare output
   variable out_p = {};
   variable out_lc = {};
   variable out_pp = {};

   % Epfold : split only. If no splitting, astronomical gap scale
   variable gs;
   if (no_splitting ==1) {
      gs = DOUBLE_MAX;
   } else {
      gs = 0.5;
   }

   variable split_add = struct { gap_scale = gs, split_only = 1, chatty = chatty };
   epfold_split_qualifiers = struct_combine( split_add, epfold_split_qualifiers );
   variable split = split_and_epfold_lc ( lc, p0; epfold_qualifiers = epfold_split_qualifiers ); 

   % For each segment do smoothing ( unless no_smoothing == 1 )
   variable jj;
   _for jj (0, length(split.lc) - 1, 1) {
      if ( typeof(split.lc[jj]) == Null_Type ) {
         vmessage("Warning(%s): Segment %d corrupted", _function_name, jj);
         continue;
      }
      if (no_smoothing == 0) {
         variable lc_red = struct {
            time = split.lc[jj].time,
            rate = split.lc[jj].rate,
            error = split.lc[jj].error };
         variable lc_avg = pulse2pulse_flux_lc ( lc_red, p0;; pulse2pulse_slope_qualifiers );
         if ( length(lc_avg.time) > 1 ) {
            if(chatty){vmessage("(%s): slope correction successful for segment %d", _function_name, jj);}
            lc_avg.rate = interpol_points(lc_red.time, lc_avg.time, lc_avg.rate);
            split.lc[jj].rate += mean(lc_avg.rate) - lc_avg.rate;
         } else {
            if(chatty){vmessage("Warning(%s): slope correction failed for segment %d", _function_name, jj);}
         }
         split.lc[jj].fracexp = ones(length(split.lc[jj].time));
      } else {
         if(chatty){vmessage("(%s): No slope correction is performed as you have chosen", _function_name);}
      }
      % For each segment do epochfolding, astronomical gap scale
      variable add_struct = struct { gap_scale = DOUBLE_MAX, chatty = chatty };
      epfold_qualifiers = struct_combine ( add_struct, epfold_qualifiers );
      if(chatty){vmessage("(%s): Performing epochfolding for segment %d", _function_name, jj);}
      variable ep = split_and_epfold_lc ( split.lc[jj], p0; epfold_qualifiers = epfold_qualifiers );
      % For each segment do find_peak
      if ( typeof(ep.epfold[0] ) == Null_Type ) {
         vmessage("Warning(%s): Epochfolding failed for segment %d", _function_name, jj);
         continue;
      }
      variable find_peak_in = struct { lc = ep.lc[0], p = ep.epfold[0].p, stat = ep.epfold[0].stat, nbins_epfold = ep.epfold[0].nbins, expectation = p0, sigma = sigma };
      variable find_peak_add = struct { pfold_not_exact = 1, exact = ep.exact[0], chatty=chatty };
      find_peak_qualifiers = struct_combine ( find_peak_add, find_peak_qualifiers );
      variable p = find_peak ( find_peak_in; find_peak_qualifiers = find_peak_qualifiers ); 
      
      if (struct_field_exists(p, "profile")) {
         list_append ( out_pp, p.profile );
         p = reduce_struct ( p, "profile" );
      }
      list_append ( out_p, p );
      list_append ( out_lc, ep.lc[0] );
   }

   variable output;
   if (compact) {
      variable mean_period = 0;
      variable mean_period_bad = 0;
      variable err_period=0;
      variable err_period_bad = 0;
      variable sum = 0;
      variable kk;
      _for kk (0, length(out_p) - 1, 1) {
         if (out_p[kk].badness < 2) {
            mean_period += out_p[kk].period;
            err_period += sqr(out_p[kk].error);
            sum += 1;
         }
         mean_period_bad += out_p[kk].period;
         err_period_bad += sqr(out_p[kk].error);
      }
      if (mean_period ==0) {
         mean_period = mean_period_bad/length(out_p);
         err_period = sqrt(err_period_bad) / length(out_p);
      } else {
         mean_period /= double(sum);
         err_period = sqrt(err_period) / double(sum);
      }
      output = struct {
         period = mean_period,
         error = err_period, 
         pp = out_pp };
   } else {
      output = struct {
         ep = out_p,
         pp = out_pp,
         lc = out_lc };
      if (length(output.pp) == 0) {
         output = reduce_struct(output, "pp");
      }
   }
   return output;
} %}}}

%
% Code from src/data/pulsar/pulseprofile_de-compose.sl
%
%%%%%%%%%%%%%%%
define pulseprofile_compose()
%%%%%%%%%%%%%%%
%!%+
%\function{pulseprofile_compose}
%\synopsis{composes a pulse profile from sine and cosine functions}
%\usage{Struct_Type pulseprofile_compose(Struct_Type decomposition);}
%\description
%    Inverse of pulseprofile_decompose, see its help for details.
%\seealso{pulseprofile_decompose, pfold}
%!%-
{
  variable nbins, phi0, a, b, ampl;
  switch (_NARGS)
    { case 1: (phi0) = ();
      nbins = phi0.nbins;
      ampl = phi0.ampl;
      a = phi0.a;
      b = phi0.b;
      phi0 = phi0.phi0;
    }
    { help(_function_name()); return; }

  % calculate profile
  variable output = ampl * ones(nbins);
  variable i, s, c, phi = 2*PI*[0:1:#nbins];
  _for i (1, length(a), 1) {
    (s,c) = sincos(i*phi - phi0*2*PI);
    output += a[i-1]*s + b[i-1]*c;
  }
  
  return struct {
    bin_lo = [0:1.*(nbins-1)/nbins:#nbins],
    bin_hi = [1./nbins:1:#nbins],
    value = output
  };
}

private variable _pulseprofile_decompose_fitwrapper_freep = struct {
  ai = NULL, bi = NULL, an = 0, bn = 0
};
private define _pulseprofile_decompose_fitwrapper(x, pars) {
  % only specific orders given?
  variable fwf = _pulseprofile_decompose_fitwrapper_freep;
  variable spec = (fwf.ai != NULL && fwf.bi != NULL);
  % highest order
  variable n = spec ? max([fwf.an,fwf.bn]) : (length(pars)-2)/2;
  % array of coefficients
  variable a = Double_Type[n], b = Double_Type[n];
  if (spec) {
    if (fwf.an > 0) { a[fwf.ai[[:fwf.an-1]]-1] = pars[[2:fwf.an+1]]; }
    if (fwf.bn > 0) { b[fwf.bi[[:fwf.bn-1]]-1] = pars[[fwf.an+2:]]; }
  } else {
    a[*] = pars[[2:n+1]];
    b[*] = pars[[-n:]];
  }

  % return composition
  return pulseprofile_compose(struct {
    nbins = length(x), phi0 = pars[0],
    ampl = pars[1], a = a, b = b
  }).value;
}

%%%%%%%%%%%%%%%
define pulseprofile_decompose()
%%%%%%%%%%%%%%%
%!%+
%\function{pulseprofile_decompose}
%\synopsis{decomposes a pulse profile into sine and cosine functions}
%\usage{Struct_Type pulseprofile_decompose(
%      Struct_Type profile[, Integer_Type[] a_orders, b_orders]
%    );}
%\qualifiers{
%    \qualifier{amin/amax}{array of min/max values for the allowed range
%                of coefficient a_n during the fit (same order
%                as a_orders). Needs a_orders and b_orders to
%                be specified (default: NULL)}
%    \qualifier{bmin}{same as amin/amax for b_n}
%    \qualifier{phi0rng}{range of allowed phase offets [min,max]
%                (default: [-1,1])}
%    \qualifier{amplrng}{range of allowed amplitudes [min,max]
%                (default: [-_Inf,+_Inf])}
%    \qualifier{initpars}{array of initial parameters in the form
%                [phi0[, ampl[, a_n..., b_n...]]]
%                The values for a_n and b_n only apply if
%                a_orders and b_orders are specified
%                (default: [0, mean(profile.value), 0..., 0...])}
%    \qualifier{maxord}{highest order to use (if no specific orders for
%            a or b are given; default: nbins/2-1)}
%}
%\description
%    The given pulse profile, F, of the form
%       struct { Double_Type[] bin_lo, bin_hi, value }
%    is fitted by a series of sine and cosine functions,
%    
%    F(phi) = ampl + sum_n^N a_n*sin(2PI*(phi+phi_0)*n)
%                            + b_n*cos(2PI*(phi+phi_0)*n)
%
%    where phi is the phase bin, ampl is the mean flux, N is
%    the highest order to use, phi_0 is a phase offset, and
%    a_n and a_b are the coefficients of the sine and cosine,
%    respectively.
%    By default the series is calculated up to the highest
%    possible order, which is related to the number of phase
%    bins (Nyquist frequency). It is also possible to specify
%    the orders, which has to be taken into account during the
%    fitting, for the sine and cosine function (a_orders and
%    b_orders, respectively).
%    The returned structure contains the resulting fit
%    parameters,
%    
%    struct {
%      Integer_Type nbins, % number of phase bins
%      Double_Type phi0,   % phase offset
%      Double_Type ampl,   % mean amplitude (zero order)
%      Double_Type[] a,    % sine coefficients
%      Double_Type[] b     % cosine coefficients
%    }
%
%    where the indices of the arrays a and b specify the order,
%    n, of the coefficient (starting with n=1). This structure
%    can be passed to 'pulseprofile_compose' to calculate the
%    modelled pulse profile from the coefficients.
%\example
%    % synthetic example profile
%    variable nbins = 32;
%    variable prof = struct { bin_lo, bin_hi, value };
%    (prof.bin_lo, prof.bin_hi) = linear_grid(0, 1, nbins);
%    prof.value  = 2 + 8*sin((prof.bin_lo)*2*PI)^2;
%    prof.value += 4*sin((prof.bin_lo)*PI);
%    hplot(prof);
%    
%    % decomposition using the first order sine and second
%    % order cosine coefficients only
%    variable decomp = pulseprofile_decompose(prof, [1], [2]);
%    ohplot(pulseprofile_compose(decomp));
%\seealso{pfold, pulseprofile_compose}
%!%-
{
  variable fwf = _pulseprofile_decompose_fitwrapper_freep;
  variable prof, spec = 0, ai = NULL, bi;
  switch(_NARGS)
    { case 1: (prof) = (); }
    { case 3: (prof, ai, bi) = ();
      fwf.ai = @ai;
      fwf.bi = @bi;
      spec = 1;
    }
    { help(_function_name()); return; }

  variable nbins = length(prof.value);

  % highest order (< Nyquist frequency)
  variable n = spec ? max([ai,bi]) : qualifier("maxord", nbins/2-1);
  if (n > nbins/2-1) { vmessage("error (%s): maxord needs to be <=d%!", _function_name, nbins/2-1); return; }
  
  % parameter ranges
  variable amin = spec ? qualifier("a_min", NULL) : NULL;
  variable amax = spec ? qualifier("a_max", NULL) : NULL;
  variable bmin = spec ? qualifier("b_min", NULL) : NULL;
  variable bmax = spec ? qualifier("b_max", NULL) : NULL;
  if (spec) {
    if (amin != NULL && length(amin) != length(ai)) { vmessage("error (%s): a_min has to have the same length as a_orders!", _function_name); return ; }
    if (amax != NULL && length(amax) != length(ai)) { vmessage("error (%s): a_max has to have the same length as a_orders!", _function_name); return ; }
    if (bmin != NULL && length(bmin) != length(bi)) { vmessage("error (%s): b_min has to have the same length as a_orders!", _function_name); return ; }
    if (bmax != NULL && length(bmax) != length(bi)) { vmessage("error (%s): b_max has to have the same length as a_orders!", _function_name); return ; }
  }
  variable phi0rng = qualifier("phi0rng", [-1,1]);
  if (length(phi0rng) != 2) { vmessage("error (%s): phi0rng has to have 2 elements", _function_name); return; }
  variable amplrng = qualifier("amplrng", [-_Inf,_Inf]);
  if (length(amplrng) != 2) { vmessage("error (%s): amplrng has to have 2 elements", _function_name); return; }
  
  % starting values (phi0, ampl)
  variable initpars = qualifier("initpars", Double_Type[0]);
  if (length(initpars) < 1) { initpars = [initpars, 0.]; }
  if (length(initpars) < 2) { initpars = [initpars, 2*mean(prof.value)]; }
  variable pars = [initpars[0], initpars[1]];
  variable a = Double_Type[0], b = Double_Type[0];
  % iterative fit
  variable i, k;
  _for i (1, n, 1) {
    % increase order
    if (spec) {
      k = wherefirst(ai == i);
      if (k != NULL) { a = [a, length(initpars) > 2 ? initpars[k] : 0]; }
      k = wherefirst(bi == i);
      if (k != NULL) { b = [b, length(initpars) > 2 ? initpars[length(ai)+k] : 0]; }
      fwf.an = length(a);
      fwf.bn = length(b);
    } else {
      a = [a, 0]; b = [b, 0];
    }
    pars = [pars[[0,1]], a, b];

    % fit
    (pars,) = array_fit(
      prof.bin_lo, prof.value, NULL, pars,
      [phi0rng[0], amplrng[0],
       amin == NULL || length(a) == 0 ? -_Inf*ones(length(a)) : amin[[:length(a)-1]],
       bmin == NULL || length(b) == 0 ? -_Inf*ones(length(b)) : bmin[[:length(b)-1]]],
      [phi0rng[1], amplrng[1],
       amax == NULL || length(a) == 0 ? +_Inf*ones(length(a)) : amax[[:length(a)-1]],
       bmax == NULL || length(b) == 0 ? +_Inf*ones(length(b)) : bmax[[:length(b)-1]]],
      &_pulseprofile_decompose_fitwrapper
    );
    a = pars[[2:length(a)+1]];
    b = pars[[length(a)+2:]];
  }

  % reset and return
  if (spec) {
    a = Double_Type[n]; b = Double_Type[n];
    if (fwf.an > 0) { a[fwf.ai[[:fwf.an-1]]-1] = pars[[2:fwf.an+1]]; }
    if (fwf.bn > 0) { b[fwf.bi[[:fwf.bn-1]]-1] = pars[[fwf.an+2:]]; }
    fwf.ai = NULL;
    fwf.bi = NULL;
    fwf.an = 0;
    fwf.bn = 0;
  }
  return struct {
    nbins = nbins,
    phi0 = pars[0],
    ampl = pars[1],
    a = a,
    b = b
  };
}

%
% Code from src/data/pulsar/pulseprofile_energy_map.sl
%

% only activate if module gcontour could be loaded
#ifeval __get_reference("gcontour_compute")!=NULL

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL
% -*- mode: slang; mode: fold -*-


% TODO
% - detailed help is missing for all functions!
%   add examples, for after-editing of plots as well
% - allow user phase-grids (code already done, but needs to be
%   activated once pfold accepts user phase-grids)
% - if two maps are provided (Struct_Type[2]?) to the plotting
%   routine, it will plot these side by side and skip showing
%   one map twice (one can plot a raw- vs. interpolated-map)
% - plotting routine allows mutiple maps for different
%   detectors (= energy ranges) to be given (List_Type?)
% - include energy shift (lag) of the spectra (y-direction) in
%   the xfig plot (qualifier e_lag)
% - put the color scale to the top of the plot if needed
% - adaptive energy grid
% - check spectrum interpolation (y-direction!): how to
%   interpolate a histogram properly? what about the edges?
% - instead of lag calculate relative difference to reference
%   like sum(abs(current/reference - 1))

%%%%%%%%%%%%%%%%%%%%%
define pulseprofile_energy_normalize()
%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{pulseprofile_energy_normalize}
%\synopsis{normalize a given pulse profile, spectrum, or pulseprofile-energy-map}
%\usage{Double_Type[] pulseprofile_energy_normalize(
%       Double_Type[] profile_or_spectrum,
%       String_Type method
%     );}
%\altusage{Double_Type[egrid,pgrid] pulseprofile_energy_normalize(
%          Double_Type[egrid,pgrid] map,
%          String_Type method,
%          Char_Type dimension
%          );}
%\description
%    A function to normalize either a 1D array representing a pulse profile or
%    spectrum, or a 2D pulse profile energy map. In the latter case the dimension
%    the normalization should apply to has to be specified, as the normalization
%    is performed in 1D, i.e., either for each phase normalize the energies
%    (dim = 'e') or for each energy normalize the pulse profile (dmi = 'p').
%    
%    Following normalization methods are available:
%
%       "sdev": Substracts the mean value and devides by the standard deviation:
%               isis> mom = moment( value );
%               isis> normvalue = ( value - mom.ave ) / mom.sdev;
%
%     "minmax": Substracts the min. value and deviedes by the min-max range.
%               isis> normvalue = ( value - min(value) ) / (max(value)-min(value));
%    
%\example
%\seealso{pulseprofile_energy_map}
%!%-
{
  variable value, method, dim = NULL;
  switch (_NARGS)
    { case 2: (value, method) = (); }
    { case 3: (value, method, dim) = (); }
    { help(_function_name); return; }

  value = COPY(value);
  variable shape = array_shape(value), tmp;
  switch (length(shape))
  % single array (profile or spectrum) given
  { case 1:
    switch (method)
    { case "sdev": 
      tmp = moment(value);
      return (value - tmp.ave) / tmp.sdev;
    }
    { case "minmax":
      return (value - min(value)) / (max(value)-min(value));
    }
    % { case "pfrac":
    % 	return (value - min(value)) / (max(value)-min(value));
    % }

    { vmessage("error(%s): method '%s' not known!", _function_name, method); return; }
  }
  % map given
  { case 2:
    variable i, idx, idy;
    dim = wherefirst(dim == ['p','e']);
    if (dim == NULL) { vmessage("error(%s): dimension has to be either 'p' oder 'e'", _function_name); return; }
    switch (method)
	    { case "sdev":
      _for i (0, shape[dim]-1, 1) {
	(idy, idx) = dim == 0 ? ([i],[*]) : ([*],[i]);
	tmp = moment(value[idy,idx]);
	value[idy,idx] = (value[idy,idx] - tmp.ave) / tmp.sdev;
      }
    }
    { case "minmax":
      _for i (0, shape[dim]-1, 1) {
        (idy, idx) = dim == 0 ? ([i],[*]) : ([*],[i]);
        value[idy,idx] = (value[idy,idx] - min(value[idy,idx])) / (max(value[idy,idx])-min(value[idy,idx]));
      }
    }
    { vmessage("error(%s): method '%s' not known!", _function_name, method); return; }    
  }
  % error
  { vmessage("error(%s): either single profile/spectrum nor map given!", _function_name); return; }

  return value;
} %}}}


%%%%%%%%%%%%%%%%%%%%%
define pulseprofile_energy_map(); % recursion needed for background calculation
define pulseprofile_energy_map()
%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{pulseprofile_energy_map}
%\synopsis{sorts the given events into a pulsephase-energy-histogram}
%\usage{Double_Type[egrid,phigrid] pfold_event_energy_map(
%      Double_Type[] events, energies, Double_Type pulse_period;
%      gti = Struct_Type, egrid = Struct_Type, pgrid = Struct_Type
%    );}
%\description
%\example
%\seealso{pfold}
%!%-
{
  variable events, energies, p0, backevts = NULL, backenerg = NULL;
  switch (_NARGS)
    { case 3: (events, energies, p0) = (); }
    { case 5: (events, energies, p0, backevts, backenerg) = (); }
    { help(_function_name); return; }

  % sanity checks
  if (typeof(events) != Array_Type || typeof(energies) != Array_Type) {
    vmessage("error(%s): 'events' and 'energies' have to be an array of doubles!", _function_name);
    return;
  }
  if (length(events) != length(energies)) {
    vmessage("error(%s): 'events' and 'energies' have to be of equal length!", _function_name);
  }
  
  % define energy grid
  variable e;
  variable egrid = qualifier("egrid", NULL);
  if (typeof(egrid) != Struct_Type) {
    e = typeof(egrid) == Integer_Type ? egrid : 16; % number of energy bins
    egrid = struct { bin_lo, bin_hi };
    (egrid.bin_lo, egrid.bin_hi) = log_grid(min(energies), max(energies), e);
  }

  % define phase grid
  variable pgrid = qualifier("pgrid", NULL);
  if (typeof(pgrid) != Struct_Type) {
    e = typeof(pgrid) == Integer_Type ? pgrid : 16; % number of phase bins
    pgrid = struct { bin_lo, bin_hi };
    (pgrid.bin_lo, pgrid.bin_hi) = linear_grid(0., 1., e);
  }

  % create pulse profile map
  variable map = Double_Type[length(egrid.bin_lo),length(pgrid.bin_lo)];
  variable countmap = @map; % total count map
  variable t0 = qualifier("t0", NULL);
  if (t0 == NULL) {
    t0 = events[0];
    vmessage("warning(%s): no reference time given, setting t0 = %f", _function_name, t0);
  }
  _for e (0, length(egrid.bin_lo)-1, 1) {
    variable nents = where(egrid.bin_lo[e] <= energies < egrid.bin_hi[e]);
    if (length(nents) > 1) {
      variable pp = pfold(
        events[where(egrid.bin_lo[e] <= energies < egrid.bin_hi[e])], p0;;
% once pfold can handle user phase grids      struct_combine(struct { t0 = t0, phigrid = pgrid }, __qualifiers)
        struct_combine(struct { t0 = t0, nbins = length(pgrid.bin_lo) }, __qualifiers)
      );
      map[e,*] = pp.value;
      countmap[e,*] = pp.value*pp.ttot; % total number of events in each pixel
    }
  }

  % initialize return structure
  variable ret = struct {
    map, counts, signalquality, egrid = egrid, pgrid = pgrid
  };

  ifnot (qualifier_exists("isbackground")) {
    % eventually calculate background map and subtract it
    if (backevts == NULL) {
      vmessage("!!!\n!!! warning(%s): please provide background events!\n!!!", _function_name);
    } else {
      variable backmap = pulseprofile_energy_map(
          backevts, backenerg, p0;; struct_combine(__qualifiers, struct {
          egrid = egrid, pgrid = pgrid, t0 = t0, isbackground
        })
      );
      ifnot (qualifier_exists("dontsubtract")) {
        map -= backmap.map;
        %countmap -= backmap.counts;
      }
      ret = struct_combine(ret, struct {
        backmap = backmap
      });
    }

    % calculate signal-to-noise (including background)
    ret.signalquality =
      (countmap - (backevts == NULL ? 0 : backmap.counts))
      / (backevts == NULL ? sqrt(countmap) : backmap.counts);
    
    % normalize
    variable norm = qualifier("norm", "sdev");
    variable dim = qualifier("normdim", 'p');
    if (norm != NULL) {
      map = pulseprofile_energy_normalize(map, norm, dim);
    }
  }
  ret.map = map;
  ret.counts = countmap;
  
  return ret;
} %}}}


%%%%%%%%%%%%%%%%%%%%%
define pulseprofile_energy_interpolate()
%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{pulseprofile_energy_interpolate}
%\synopsis{interpolates a pulsephase-energy-histogram to a finer grid}
%\usage{Double_Type[egrid*e_interp,phigrid*p_interp] pulseprofile_energy_interpolate(
%      Struct_Type map, Integer_Type p_interp, Integer_Type e_interp
%    );}
%\description
%\example
%\seealso{pulseprofile_energy_map}
%!%-
{
  variable map, p_interp, e_interp;
  switch (_NARGS)
    { case 3: (map,p_interp,e_interp) = (); }
    { help(_function_name); return; }

  variable p_ref = qualifier("p_ref", NULL);
  variable e_ref = qualifier("e_ref", NULL);
  
  variable shape, nmap = map.map, i, ngrid, tcmap, tsmap;
  % interpolate map - in phase direction
  if (p_interp > 1) {
    shape = array_shape(map.map);
    nmap = Double_Type[shape[0], shape[1]*p_interp];
    ngrid = struct {
      bin_lo = 1.*[0:shape[1]*p_interp-1] / (shape[1]*p_interp),
      bin_hi = 1.*[1:shape[1]*p_interp] / (shape[1]*p_interp),
    };		    
    _for i (0, shape[0]-1, 1) {
      nmap[i,*] = gsl->interp_cspline_periodic( % profile is periodic!
        .5*(ngrid.bin_lo + ngrid.bin_hi),
	.5*([map.pgrid.bin_lo[-1]-1,map.pgrid.bin_lo,map.pgrid.bin_lo[0]+1]+[map.pgrid.bin_hi[-1]-1,map.pgrid.bin_hi,map.pgrid.bin_hi[0]+1]),
	[map.map[i,-1], map.map[i,*], map.map[i,0]]
      );
    }
     % interpolat reference profile
    if (p_ref != NULL) {
      @p_ref = gsl->interp_cspline_periodic(
        .5*(ngrid.bin_lo + ngrid.bin_hi),
	.5*([map.pgrid.bin_lo[-1]-1,map.pgrid.bin_lo,map.pgrid.bin_lo[0]+1]+[map.pgrid.bin_hi[-1]-1,map.pgrid.bin_hi,map.pgrid.bin_hi[0]+1]),
	[(@p_ref)[-1], @p_ref, (@p_ref)[0]]
      );
    }
    map.map = nmap;
    map.pgrid = ngrid;

    % enlarge counts- and signalquality-maps
    tcmap = Double_Type[shape[0], shape[1]*p_interp]; tsmap = @tcmap;
    _for i (0, p_interp-1) {
      tcmap[*,[i::p_interp]] = map.counts;
      tsmap[*,[i::p_interp]] = map.signalquality;
    }
    map.counts = tcmap;
    map.signalquality = tsmap;
  }

  % interpolate map - in energy direction
  if (e_interp > 1) {
    shape = array_shape(map.map);
    variable l = shape[0]*e_interp-2;
    ngrid = gsl->interp_polynomial(
      1.*[0:l]/l,
      1.*[0:shape[0]-1]/(shape[0]-1),
      .5*(map.egrid.bin_lo+map.egrid.bin_hi)
    );
    ngrid = struct { bin_lo = ngrid[[:-2]], bin_hi = ngrid[[1:]] };
    nmap = Double_Type[l, shape[1]]; % l -> l+2 for edge extrapolation
    _for i (0, shape[1]-1, 1) {
      nmap[*,i] = gsl->interp_linear( % spectrum is non-periodic (it better is!)
        .5*(ngrid.bin_lo + ngrid.bin_hi), % [*,i] -> [1:l],i for edge extrapolation
        .5*(map.egrid.bin_lo + map.egrid.bin_hi),
     	map.map[*,i]
      );
      % % extrapolate edges (either: edge raw flux half, or: missing edge flux)
      % variable totraw = sum(map.map[*,i]*(map.egrid.bin_hi-map.egrid.bin_lo));
      % variable totint = sum(nmap[[1:l],i]*(ngrid.bin_hi-ngrid.bin_lo));
      % variable left  = [
      % 	map.map[0,i]*(map.egrid.bin_hi[0]-map.egrid.bin_lo[0])/2,
      %   map.map[0,i]*(map.egrid.bin_hi[0]-map.egrid.bin_lo[0])-nmap[1,i]*(ngrid.bin_hi[0]-ngrid.bin_lo[0])
      % ];
      % variable right = [
      % 	map.map[-1,i]*(map.egrid.bin_hi[-1]-map.egrid.bin_lo[-1])/2,
      % 	map.map[-1,i]*(map.egrid.bin_hi[-1]-map.egrid.bin_lo[-1])-nmap[-2,i]*(ngrid.bin_hi[-1]-ngrid.bin_lo[-1])
      % ];
      % % which method results in better flux conversation?
      % variable w = where_min(abs(totraw - (totint+left+right)));
      % totint += (left[w]+right[w]);
      % nmap[0,i]  = left[w];
      % nmap[-1,i] = right[w];
      % % vmessage("warning(%s): interpolated flux differs by %.3f%%", _function_name, (totint/totraw-1)*100);
    }
    % add bins at the edges
    % ngrid.bin_lo = [map.egrid.bin_lo[0], ngrid.bin_lo, ngrid.bin_hi[-1]];
    % ngrid.bin_hi = [ngrid.bin_lo[1], ngrid.bin_hi, map.egrid.bin_hi[-1]];
    % % correct bin integrated flux in edges
    % nmap[0,*]  /= (ngrid.bin_hi[0] -ngrid.bin_lo[0] );
    % nmap[-1,*] /= (ngrid.bin_hi[-1]-ngrid.bin_lo[-1]);
    
    % interpolat reference spectrum
    if (e_ref != NULL) {
      @e_ref = gsl->interp_cspline(
        .5*(ngrid.bin_lo + ngrid.bin_hi),
        .5*(map.egrid.bin_lo + map.egrid.bin_hi),
    	@e_ref
      );
    }
    map.map = nmap;
    map.egrid = ngrid;

    % enlarge counts- and signalquality-maps
    tcmap = Double_Type[shape[0]*e_interp, shape[1]]; tsmap = @tcmap;
    _for i (0, e_interp-1, 1) {
      tcmap[[i::e_interp],*] = map.counts;
      tsmap[[i::e_interp],*] = map.signalquality;
    }
    map.counts = tcmap[[1:l],*]; % remove for edge extrapolation
    map.signalquality = tsmap[[1:l],*]; % remove for edge extrapolation
  }
  
  return map;
} %}}}


%%%%%%%%%%%%%%%%%%%%%
define pulseprofile_energy_lag()
%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{pulseprofile_energy_lag}
%\synopsis{calculates a lag (=shift) in a pulsephase-energy-histogram}
%\usage{Double_Type[] pulseprofile_energy_lag(
%      Struct_Type map, 'p' or 'e'[, Double_Type[] reference]
%    );}
%\description
%\example
%\seealso{pulseprofile_energy_map, CCF_1d}
%!%-
{
  variable map, dim, ref = NULL;
  switch (_NARGS)
    { case 2: (map,dim) = (); }
    { case 3: (map,dim,ref) = (); }
    { help(_function_name); return; }

  dim = wherefirst(dim == ['p','e']);
  if (dim == NULL) { vmessage("error(%s): dimension has to be either 'p' oder 'e'", _function_name); return; }

  % collapse map to get the reference (average pulse profile or spectrum)
  variable i;
  if (ref == NULL) {
    vmessage("warning (%s): no reference given, collapsing map", _function_name);
    ref = Double_Type[length(dim == 0 ? map.map[0,*] : map.map[*,0])]; 
    _for i (0, length(ref)-1, 1) {
      ref[i] = sum(dim == 0 ? map.map[*,i] : map.map[i,*]);
    }
  } else { ref = COPY(ref); }

  % shift rerence to avoid jumping around phase 0
  variable sh = qualifier("shift", .5);
  sh = int(length(ref)*sh);
  ref = shift(ref, sh);
  
  % calculate lag
  variable lag = Double_Type[length(dim == 0 ? map.map[*,0] : map.map[0,*])], ccf;
  _for i (0, length(lag)-1, 1) {
    ccf = CCF_1d(dim == 0 ? map.map[i,*] : map.map[*,i], ref);
    lag[i] = mean(where_max(ccf));
  }
  lag = 1.*lag/length(ccf) - 1.*sh/length(ref);

  return struct {
    bin_lo = dim == 0 ? map.egrid.bin_lo : map.pgrid.bin_lo,
    bin_hi = dim == 0 ? map.egrid.bin_hi : map.pgrid.bin_hi,
    lag    = lag
  };
} %}}}


%%%%%%%%%%%%%%%%%%%%%
define xfigplot_pulseprofile_energy_map()
%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{xfigplot_pulseprofile_energy_map}
%\synopsis{returnes an Xfig plot object containing the pulse-profile-energy-map}
%\usage{Xfig_Object xfigplot_pulseprofile_energy_map(Struct_Type map);}
%\description
%  Struct_Type map = struct{
%     pgrid = struct{ bin_lo = Double_Type[np],
%                     bin_hi = Double_Type[np]
%                   },
%     egrid = struct{ bin_lo = Double_Type[ne],
%                     bin_hi = Double_Type[ne]
%                   },
%     map = Double_Type[ne,np]
%  };
%     
%\example
%\seealso{pulseprofile_energy_map}
%!%-
{
  variable map;
  switch (_NARGS)
    { case 1: (map) = (); }
    { help(_function_name); return; }
  
  variable plag = qualifier("p_lag", NULL),
           elag = qualifier("e_lag", NULL);

  % renorm map
  variable nmap = COPY(map.map);
  variable scalefun = qualifier("scalefun", NULL);
  variable l;
  if (scalefun != NULL) {
    nmap = (@scalefun)(nmap);
    if (struct_field_exists(__qualifiers, "gmin")) {
      __qualifiers.gmin = (@scalefun)([__qualifiers.gmin,map.map])[0];
    }
    if (struct_field_exists(__qualifiers, "gmax")) {
      __qualifiers.gmax = (@scalefun)([__qualifiers.gmax,map.map])[0];
    }
  }
  
  % convert the map (with a known grid) into an image (which has pixels
  % and thus has a *linear* grid by definition) -> we need a fine pixel
  % grid and we have to lookup the color in the original grid.

  % TODO: do that for the signalquality-map as well!
  variable fine   = qualifier("fineness", 10);
  variable shape  = array_shape(nmap)*fine;
  variable pixmap = Double_Type[__push_array(shape)];
  variable pixqual = @pixmap; % fine grid for signalquality
  variable pixgrid = struct { % grid of the pixels
    p = struct { bin_lo, bin_hi }, % pulse phase (always linear here!)
    e = struct { bin_lo, bin_hi }  % energy      (either linear or logarithmic)
  };
  (pixgrid.p.bin_lo, pixgrid.p.bin_hi) = linear_grid(
    map.pgrid.bin_lo[0], map.pgrid.bin_hi[-1], shape[1]
  );
  (pixgrid.e.bin_lo, pixgrid.e.bin_hi) = (@(qualifier_exists("lin") ? &linear_grid : &log_grid))(
    map.egrid.bin_lo[0], map.egrid.bin_hi[-1], shape[0]
  );
  variable x,y,ix,iy;
  _for y (0, shape[0]-1, 1) _for x (0, shape[1]-1, 1) {
    ix = wherefirst(map.egrid.bin_lo <= .5*(pixgrid.e.bin_lo[y]+pixgrid.e.bin_hi[y]) <= map.egrid.bin_hi);
    iy = wherefirst(map.pgrid.bin_lo <= .5*(pixgrid.p.bin_lo[x]+pixgrid.p.bin_hi[x]) <= map.pgrid.bin_hi);
    pixmap[y,x] = nmap[ix,iy];
  }

  % show profile twice
  nmap = Double_Type[shape[0],shape[1]*2];
  nmap[*,[:shape[1]-1]] = pixmap;
  nmap[*,[shape[1]:]]   = pixmap;

  % initialize xfig-plot
  variable W = qualifier("W", 14);
  variable H = qualifier("H", 10);
  variable xlabel = qualifier("xlabel", "Pulse Phase $\varphi$"R );
  variable ylabel = qualifier("ylabel", "Energy (keV)" );
  
  variable pl = xfig_plot_new(W,H);
  pl.xlabel(xlabel);
  pl.ylabel(ylabel);
  pl.world(0, 2, map.egrid.bin_lo[0], map.egrid.bin_hi[-1]);
  ifnot (qualifier_exists("lin")) { pl.yaxis(;log); pl.y2axis(; ticlabels = 0); }
  variable tics = [0,.5,1.0,1.5], ticl = COPY(tics); ticl[where(ticl > 1)] -= 1;
  pl.x2axis(; major = tics, minor = tics+.25, ticlabels = 0,
    major_color = "white", minor_color = "white");
  pl.x1axis(; major = tics, minor = tics+.25, ticlabels = array_map(
    String_Type, &sprintf, "%.1f", ticl
  ));
  % plot the map
  pl.plot_png(nmap;; struct_combine(struct { depth = 100 }, __qualifiers));
  if (qualifier_exists("plotegrid")) {
    _for l (1, length(map.egrid.bin_lo)-1, 1) {
      pl.plot([0,2], [1,1]*map.egrid.bin_lo[l]; color = "white", line = 1);
    }
  }

  %%% SIGNALQUALITY
  % plot a boarder where the signalquality drops below a certain value
  if( qualifier_exists("snthres") ){
    variable snthres = qualifier("snthres",5);
    variable sncols = qualifier("sncols","darkgreen");

    variable snshape  = array_shape(map.signalquality);
    variable snmap = Double_Type[snshape[0],snshape[1]*2];
    snmap[*,[:snshape[1]-1]] = map.signalquality;
    snmap[*,[snshape[1]:]]   = map.signalquality;

    variable sncont   = gcontour_compute( snmap, snthres );

    variable n,i;
    _for n ( 0, length(sncont)-1 ){
      _for i ( 0, length(sncont[n].x_list)-1 ){
	x = sncont[n].x_list[i]/(snshape[1]*2.);
	y = sncont[n].y_list[i]/snshape[0];

	pl.plot( x, y;
		 world00, width=3, depth = 98,
		 color=[sncols][length([sncols])==length(sncont)? n : 0],
		 line=length(sncont)-n-1
	       );
	pl.plot( x, y; world00, width=1, depth = 99 );

      }
    }
  }

  % color scale
  variable cmap_fsize = qualifier("cmapfontsize","scriptsize");
  variable pc = xfig_plot_new(0.333*W, 0.05*H);
  l = length(png_get_colormap(qualifier("cmap", "gray")));
  pc.world(min(map.map),max(map.map),0,1);
  pc.yaxis(; major = 0, minor = 0);
  pc.x1axis(; ticlabel_size = cmap_fsize,maxtics=5);
  pc.x2axis(; major_color = "white", minor_color = "white", ticlabels = 0);
  if (qualifier_exists("cmap_label")) {
    pc.x2label(qualifier("cmap_label"); size = cmap_fsize);
  }
  variable s = [min(map.map):max(map.map):#l];
  if (scalefun != NULL) { s = (@scalefun)(s); }
  pc.plot_png(_reshape(s, [1,l]);;
    struct_combine(struct { depth = 96 }, __qualifiers )
  );
  % draw a white box behind the color scale
  variable x1,x2,y1,y2,x0,y0;
  (x1,x2,y1,y2,,) = pc.get_bbox();
  variable bord = (x2-x1)*.03;
  variable box = xfig_new_rectangle(x2-x1+2*bord, y2-y1+2*bord);
  box.set_area_fill(20); box.set_fill_color("white");
  box.set_depth(97);
  box.translate(vector(x1-bord, y1-bord, 0.));
  pc.add_object(box);
  pl.add_object(pc, .93, .05, .5, -.5; world0);

  % phase lag
  if (plag != NULL) {
    % histogram values in y-direction
    plag = struct_combine(plag, struct { nlag = Double_Type[length(plag.bin_lo)*2], e = Double_Type[length(plag.bin_lo)*2] });
    _for l (0, length(plag.lag)-1, 1) {
      plag.nlag[[2*l,2*l+1]] = [1,1]*plag.lag[l];
      plag.e[[2*l,2*l+1]] = [plag.bin_lo[l], plag.bin_hi[l]];
    }
    % plot
    pc = xfig_plot_new(0.15*W, H);
    pc.world(min(plag.lag), max(plag.lag), plag.bin_lo[0], plag.bin_hi[-1];;
      struct_combine(struct { padx = .15 }, qualifier_exists("lin") ? NULL : struct { ylog })
    );
    pc.xlabel("$\Delta \varphi$"R);
    pc.plot(plag.nlag, plag.e); 
    pc.y1axis(; ticlabels = 0); 
    pc.translate(vector(W,0,0));
    pl.add_object(pc);
  }
  
  return pl;
} %}}}


#endif

#endif

#endif

#endif

%
% Code from src/data/pulsar/pulseprofile_phase_connect.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

% helper function: interpolate the cross-correlation
% given by the ccf-qualifier at the point x
private define _pulseprofile_phase_connect_ccfint(x) {
  variable ccf = qualifier("ccf");
  variable phi = 1.*[0:length(ccf)]/length(ccf);
  if (x < 0) { x++; }
  if (x >= 1) { x--; }
  variable f = qualifier("f");
  return gsl->interp_cspline_periodic(x, phi, [ccf, ccf[0]]);
}

%%%%%%%%%%%%%%%%%%%%%%%%
define pulseprofile_phase_connect() {
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pulseprofile_phase_connect}
%\synopsis{UNDER DEVELOPMENT; returns the phase shift between two pulse profiles}
%\usage{(Double Type phi, ccf) = pulseprofile_phase_connect(Struct_Type prof, ref);}
%\qualifiers{
%    \qualifier{shift}{automatically shifts the given profile in order to match
%            the phase of the reference profile (caution: overwrites
%            the input!)}
%}
%\description
%    UNDER DEVELOPMENT, USE WITH CAUTION! Send questions or bugs to
%    matthias.kuehnel@sternwarte.uni-erlangen.de
%
%    Determines the phase shift of the pulse profile 'prof' with
%    respect to the reference profile 'ref' by a cross-correlation.
%    The cross-correlation is interpolated to enhance the precision
%    beyond the pulse profile binning. Both pulse profiles need to
%    have the same number of bins.
%
%    The input structures have to have the same fields as returned
%    by 'epfold'.
%    
%    The resulting phase shift 'phi' and the maximum value of the
%    cross-correlation 'ccf' are returned.
%\seealso{pfold, CCF_1d}
%!%-
  variable prof, ref;
  switch (_NARGS)
    { case 2: (prof, ref) = (); }
    { help(_function_name); return; }

  % sanity check
  variable nbins = length(ref.value);
  if (length(prof.value) != nbins) {
    vmessage("error(%s): different number of phase bins", _function_name);
  }
  % calculate cross correlation function
  variable ccf = CCF_1d(ref.value, prof.value); 
  % find and interpolate maximum
  variable k = where_max(ccf)[0];
  variable ki = find_function_maximum(
    &_pulseprofile_phase_connect_ccfint, (k-1.)/nbins, (k+1.)/nbins;
    qualifiers = struct { ccf = ccf }
  );
  variable ccfi = _pulseprofile_phase_connect_ccfint(ki; ccf = ccf);
  % shift profile
  if (qualifier_exists("shift")) {
    if (ccfi > qualifier("ccflim")) {
      prof.value = shift_intpol(prof.value, -ki*nbins);
      prof.error = shift_intpol(prof.error, -ki*nbins);
    } else {
      vmessage("warning(%s): match below limit in order to shift the profile, skipping...");
    }
  }
  return (ki, ccfi);
}

#endif

%
% Code from src/slang/strings/pushKommaString.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define pushKommaString(a, b)
%%%%%%%%%%%%%%%%%%%%%%
{
  pushString(a, b, ",  ");
}

%
% Code from src/slang/strings/pushString.sl
%
define pushString(str, new, sep)
{
  if(@str==NULL or @str=="") { @str = new; } else { @str = @str + sep + new; }
}

%
% Code from src/slang/arrays/pushStructFieldArray.sl
%
define pushStructFieldArray()
%!%+
%\function{pushStructFieldArray}
%\synopsis{appends a value to an array which is the field of structure}
%\usage{pushStructFieldArray(s, field, value);}
%\description
%    \code{s.field = [s.field, value];  %} if \code{s.field} was not \code{NULL}
%!%-
{
  variable s, field, value;
  switch(_NARGS)
  { case 3: (s, field, value) = (); }
  { help(_function_name()); return; }

  variable a = get_struct_field(s, field);
  if(a==NULL)
    set_struct_field(s, field, [value]);
  else
    set_struct_field(s, field, [a, value]);
}

%
% Code from src/fitting/pvm_fit_pars.sl
%

% only activate if module pvm_ms could be loaded
#ifeval __get_reference("pvm_mytid")!=NULL

% only activate if module pvm_ms could be loaded
#ifeval __get_reference("pvm_mytid")!=NULL

private variable USER_SLAVE_NEWTASK = 1;
private variable USER_SLAVE_RESULT  = 2;
private variable USER_SLAVE_EXIT    = 3;

private variable Statistic;
private variable Num_Params, Pars_Left;
private variable Best_Statistic = NULL, Best_Params = NULL;
private variable Best_filename, stdout_file;
private variable Limits;
private variable Host=Assoc_Type[String_Type], StillWorking=Assoc_Type[Char_Type];

public variable pvm_fit_pars_last_results = NULL;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define print_stdout_and_file(thres, s)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  if(Fit_Verbose >= thres)  message(s);
  ()=fprintf(stdout_file, "%s\n", s);
  ()=fflush(stdout_file);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define restart_slaves_with_new_params(params)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  Pars_Left = Num_Params;
  variable par;
  foreach par (assoc_get_keys(Host))
    StillWorking[par] = 1;
  set_params(params);
  save_par(Best_filename);
  variable master_tid = pvm_mytid();
  foreach  (pvm_tasks(0).ti_tid)
  { variable tid = ();
    if(tid != master_tid)
    { pvm_psend(tid, USER_SLAVE_NEWTASK);
      pvm_send_obj(tid, USER_SLAVE_RESULT, params);
    }
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define exit_all_slaves()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable master_tid = pvm_mytid();
  variable s = pvm_tasks(0);
  variable i = where(s.ti_tid != master_tid);
  array_map(Void_Type, &pvm_psend, s.ti_tid[i], USER_SLAVE_EXIT);
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define handle_user_message(msgid, tid)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable cl;
  switch(msgid)
  { case USER_SLAVE_NEWTASK:
      cl = pvm_recv_obj();
      if(Best_Statistic == NULL || cl.statistic < Best_Statistic)
      { print_stdout_and_file(1, sprintf("found improved fit (%s=%4.6f) @ %s [%s] for %s = %g",
					 Statistic, cl.statistic,
					 cl.hostname, strftime("%Y-%m-%d %H:%M:%S", localtime(_time())),
					 cl.name, cl.value)
			     );
        Best_Statistic = cl.statistic;
        Best_Params = cl.params;
        restart_slaves_with_new_params(Best_Params);
      }
      else
      { pvm_psend(tid, USER_SLAVE_NEWTASK);
        pvm_send_obj(tid, USER_SLAVE_RESULT, Best_Params);
      }
   }
   { case USER_SLAVE_RESULT:
       cl = pvm_recv_obj();
       variable info = get_par_info(cl.name);
       Pars_Left--;
       StillWorking[cl.name] = 0;
       variable msg = Pars_Left ? "\nworking on $Pars_Left more parameter"$ + (Pars_Left>1 ? "s" : "") : "";
       if(0 < Pars_Left <= 10)
       {
	 msg += " @ ";
	 variable par, first=1;;
	 foreach par (assoc_get_keys(Host))
	   if(StillWorking[par])
	     msg += (first ? "" : ", ") + Host[par],  first = 0;
       }
       print_stdout_and_file(0, sprintf("%10s finished [%s]:\n(%10g <=) %10g < %25s = %10g < %10g (<= %10g)   %s%s",
					cl.hostname, strftime("%Y-%m-%d %H:%M:%S", localtime(_time())),
					info.min,
					cl.lo, cl.name, cl.value, cl.hi,
					info.max,
					TeX_value_pm_error(cl.value, cl.lo, cl.hi),
					msg
				       )
			    );
       Limits[cl.name] = [cl.value, cl.lo, cl.hi];
       ifnot(Pars_Left)  exit_all_slaves();
   }
   return 1;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define slave_spawned_callback(tid, host, argv)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable par = argv[1];
  Host[par] = host;
  StillWorking[par] = 1;
  print_stdout_and_file(1, sprintf("%10s spawned for %s", host, par) );
}


%%%%%%%%%%%%%%%%%%%
define pvm_fit_pars()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pvm_fit_pars}
%\synopsis{computes confidence intervals with PVM}
%\usage{Struct_Type results = pvm_fit_pars(String_Type SetupFile[, Integer_Type pars[]]);}
%\description
%    With SetupFile, one has to provide an ISIS-script which loads/rebins the spectra
%    and loads/assigns the response. If the model requires additional modules,
%    they have to be activated as well in \code{SetupFile}. It is, however, not necessary
%    to define/load the model itself, as the currently defined model will be saved
%    into a file <\code{dir}>/<\code{basefilename}>\code{_initial.par}.
%
%    \code{pars} is an array of parameters, for which the confidence levels are to be fitted.
%    If \code{pars} is not specified, all free parameters of the current model are used.
%    The best fit which is eventually found is always saved in <\code{dir}>/<\code{basefilename}>\code{.par}.
%
%    The verbosity of \code{pvm_fit_pars} is controlled by the intrinsic variable \code{Fit_Verbose}.
%
%    The return value \code{results = struct { index, name, value, min, max, conf_min, conf_max, buf_below, buf_above, tex }}
%    is a table with the following information for each parameter:\n
%    \code{min} and \code{max} are the minimum/maximum values allowed.
%    \code{conf_min} and \code{conf_max} are the confidence limits.
%    \code{buf_below} (\code{buf_above}) is the fraction of the allowed range \code{[min:max]}
%    which seperates the lower (upper) confidence limit from \code{min} (\code{max}).
%    If one of these buffers is 0, your confidence interval has bounced.
%
%    The same infomation is stored in the files <\code{dir}>/<\code{basefilename}>\code{_conf.txt}
%    and <\code{dir}>/<\code{basefilename}>\code{_conf.fits}.
%    In case of any error, the return value is \code{NULL}.
%\qualifiers{
%\qualifier{level}{confidence level to be computed.
%             As for conf, 0 means 68%, 1 means 90% [default], and 2 means 99% confidence level.}
%\qualifier{tolerance}{the tolerance for chi^2 improvements without interrupting the search
%                 for the confidence intervals, see \code{help("conf");}. The default is \code{1e-3}.}
%\qualifier{fitmethod}{fit-method to be used, see \code{help("set_fit_method");}
%                 Default is the currently used fit-method returned by \code{get_fit_method()}.}
%\qualifier{nph}{the number of processes per host, see \code{pvm_ms}.}
%\qualifier{debug}{[=1] prints additional debug information from \code{pvm_ms}. (Default=0)}
%\qualifier{dir}{[="."] specifies the directory in which the logfiles shall be stored.
%                  It may be a relative path to the current working directory.}
%\qualifier{basefilename}{[=startdate]}
%\qualifier{verbose}{log every output -- from the master script or any slave --
%               in the file <\code{dir}>/<\code{basefilename}>\code{_stdout.log},
%               and keep the initial parameters.}
%\qualifier{isisscript}{[="isis-script"] command to start ISIS for slaves.}
%}
%\seealso{fit_pars; conf, set_fit_method, cl_master/cl_slave [Houck/Noble], pvm_ms [S-Lang module]}
%!%-
{
  variable SetupFile=NULL, pars = freeParameters();
  switch(_NARGS)
  { case 0:
      if(length(glob("init.sl"))>0) { SetupFile = "init.sl"; }
      if(length(glob("loaddata.sl"))>0) { SetupFile = "loaddata.sl"; }
      if(SetupFile==NULL) { help(_function_name()); return; }
      message(_function_name()+": using SetupFile " + SetupFile);
  }
  { case 1:  SetupFile = (); }
  { case 2: (SetupFile, pars) = (); }
  { help(_function_name()); return; }

  variable level = qualifier("level", 1);
  if(all(level!=[0:2]))
  {
    message("warning ("+_function_name()+"): level has to be 0 (68%), 1 (90%), or 2 (99%).\nUsing default level = 1.");
    level = 1;
  }
  if(level==NULL) { level = 1; }
  variable tolerance = qualifier("tolerance", 1e-3);
  variable fitmethod = qualifier("fitmethod", get_fit_method());
  variable debug_PVM = qualifier("debug", 0);
  variable num_processes_per_host = qualifier("nph", 2);
  variable dir = qualifier("dir", ".");
  variable isisscript = qualifier("isisscript", "isis-script");
  variable nicelevel = qualifier("nice", 19);
  variable remove_slave = 1;

  variable ok=1;
  try { variable pvmconfig = pvm_config(); } catch AnyError: { ok=0; };
  ifnot(ok)
  { message("error ("+_function_name()+"): PVM has not been started");
    return NULL;
  }
  variable num_hosts = length(pvmconfig.hi_tid);
  Num_Params = length(pars);

  if(num_hosts*num_processes_per_host < Num_Params)
  { vmessage("error (%s): only %d PVM-hosts available; %d processes/host is not enough for %d parameters", _function_name(), num_hosts, num_processes_per_host, Num_Params);
    return NULL;
  }

  variable CWD = getcwd();
  variable startdate = strftime("%Y-%m-%d_%H:%M:%S", localtime(_time()));
  variable basefilename = qualifier("basefilename", startdate);
  variable logbasefilename = dir + "/" + basefilename;
  stdout_file = fopen(logbasefilename+"_stdout.log", "w");
  print_stdout_and_file(1, "started on " + startdate);

  thaw(pars);
  save_par(logbasefilename+"_initial.par");
  Best_filename = logbasefilename+".par";
  Statistic = strtok( get_fit_statistic(), ";")[0];
  Limits = Assoc_Type[Any_Type, NULL];

  % setting up the PVM
  pvm_ms_set_debug(debug_PVM);
  pvm_ms_set_num_processes_per_host(num_processes_per_host);
  pvm_ms_set_message_callback(&handle_user_message);
  pvm_ms_set_slave_spawned_callback(&slave_spawned_callback);

  variable master_tid = pvm_mytid();
  USER_SLAVE_NEWTASK += master_tid * 10;
  USER_SLAVE_RESULT  += master_tid * 10;
  USER_SLAVE_EXIT    += master_tid * 10;

  % setting up the slave script
  variable filename_slave = startdate + "_slave";
  variable F = fopen(filename_slave, "w");
  ()=fputs(
`#! /usr/bin/env $isisscript
% -*- mode: SLang -*-
% $filename_slave generated by pvm_fit_pars() [Manfred Hanke <Manfred.Hanke@sternwarte.uni-erlangen.de>],
% which is based on:
%
% cl_slave: Slave program for parallel computation of single-parameter
%           confidence limits in ISIS, using PVM, as described in
%           "Using the Parallel Virtual Machine for Everyday Analysis,"
%           by Noble et al 2006 (http://arxiv.org/abs/astro-ph/0510688)
%
% Authors:  John C. Houck <houck@space.mit.edu>
%           Michael S. Noble <mnoble@space.mit.edu>
%
% Version:  Thu Mar  6 11:24:42 EST 2008


pvm_sigterm_enable(1);

private variable Have_New_Params = 0;

public define isis_fit_improved_hook()
{
  variable bufid = pvm_nrecv($master_tid, $USER_SLAVE_NEWTASK);  % non-blocking receive
  if(bufid > 0)
  { Have_New_Params = 1;
    return 1;
  }
  return 0;
}

private variable info = struct { name, value, lo, hi, hostname, params, statistic };

private define slave_next_task();
private define slave_next_task()
{
   variable bufid = pvm_recv($master_tid, -1);  % blocking receive
   variable msgid; (, msgid, ) = pvm_bufinfo(bufid);
   switch(msgid)
   { case $USER_SLAVE_RESULT:
       variable param, params = pvm_recv_obj();
       message("receiving new parameters:");
       foreach param (params)
         if(param.value != get_par(param.name))
           ()=printf("%s = %S; ", param.name, param.value);
       ()=printf("\n");
       set_params(params);
       return 1;
   }
   { case $USER_SLAVE_NEWTASK:
       return slave_next_task();
   }
   return 0;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

()=system("renice +$nicelevel " + string(getpid()) + " >> /dev/null");

()=chdir("$CWD");
()=evalfile("$SetupFile");
load_par("${logbasefilename}_initial.par");
set_fit_method("$fitmethod");

info.name = __argv[1];
info.hostname = uname().nodename;

variable status = 1;
do
{
   Have_New_Params = 0;

   % do it
   (info.lo, info.hi) = conf (info.name, $level, $tolerance);
   info.value = get_par (info.name);
   ifnot(Have_New_Params)
   {
     if(info.lo == info.hi)  % Found better fit
     {
       variable x; ()=fit_counts(&x);
       info.statistic = x.statistic;
       info.params = get_params();
       pvm_send_obj($master_tid, $USER_SLAVE_NEWTASK, info);
     }
     else
     {
       info.params = NULL;
       info.statistic = NULL;
       pvm_send_obj($master_tid, $USER_SLAVE_RESULT, info);
     }
   }
} while(0 != slave_next_task() );

pvm_ms_slave_exit(status);
exit(status);
`$, F); % fputs
  ()=fclose(F);
  ()=system("chmod u+x "+filename_slave);

  % starting slaves
  Pars_Left = Num_Params;
  variable slave_argvs = Array_Type[length(pars)];
  _for $1 (0, Num_Params-1, 1)  slave_argvs[$1] = [CWD + filename_slave, get_par_info(pars[$1]).name];
  tic;
  variable exit_status = pvm_ms_run_master(slave_argvs);

  % finished
  variable runtime = int(toc()); % runtime, in seconds
  print_stdout_and_file(1, strftime("finished on %Y-%m-%d %H:%M:%S", localtime(_time())));
  if(remove_slave)  ()=remove(filename_slave);

  variable hours = runtime mod (24*60*60);
  variable mins = hours mod 3600;
  print_stdout_and_file(1, sprintf("runtime: %dd %dh %dmin %dsec", runtime/86400, hours/3600, mins/60, mins mod 60));
  message("");

  foreach $1 (exit_status)
  { ()=fprintf(stdout_file, "------------------------------------------------------------------------\n%s: (exit_status %d)\n%S\n\n", $1.host, $1.exit_status, $1.stdout); }

  ()=fclose(stdout_file);

  % collect results
  variable results = struct {
             index     = Integer_Type[0],
             name      = String_Type[0],
             value     = Double_Type[0],
             min       = Double_Type[0],
             max       = Double_Type[0],
             conf_min  = Double_Type[0],
             conf_max  = Double_Type[0],
             buf_below = Double_Type[0],
             buf_above = Double_Type[0],
             tex       = String_Type[0],
           };
  variable name, v;
  foreach name, v (Limits) using ("keys", "values")
  {
    variable info = get_par_info(name);
    variable buf_below = (v[1]-info.min)/(info.max-info.min);
    variable buf_above = (info.max-v[2])/(info.max-info.min);
    results.index     = [results.index,     info.index];
    results.name      = [results.name,      name];
    results.value     = [results.value,     v[0]];
    results.conf_min  = [results.conf_min,  v[1]];
    results.conf_max  = [results.conf_max,  v[2]];
    results.min       = [results.min,       info.min];
    results.max       = [results.max,       info.max];
    results.buf_below = [results.buf_below, buf_below];
    results.buf_above = [results.buf_above, buf_above];
    results.tex       = [results.tex,       TeX_value_pm_error(v[0], v[1], v[2])];
    set_par(name, v[0]);
  }
  save_par(Best_filename);
  variable stat, FV = Fit_Verbose;
  Fit_Verbose = -1;
  ()=eval_counts(&stat);
  Fit_Verbose = FV;

  struct_filter(results, array_sort(results.index));
  pvm_fit_pars_last_results = results;
  fits_write_binary_table(logbasefilename+"_conf.fits", "pvm_fit_pars-results", results, struct {
    chi2 = stat.statistic,
    num_bins = stat.num_bins,
    n_var_pars = stat.num_variable_params,
    dof = stat.num_bins-stat.num_variable_params,
    chi2red = stat.statistic/(stat.num_bins-stat.num_variable_params)
  } );

  % write confidence file
  stdout_file = fopen(logbasefilename+"_conf.txt", "w");
  print_stdout_and_file(0, get_fit_fun());
  print_stdout_and_file(0, sprintf("(%10s <=) %10s < %38s < %10s (<= %10s)", "min.all.", "conf.min.", "parameter value", "conf.max.", "max.all."));
  for($1=0; $1<length(results.name); $1++)
    print_stdout_and_file(0, sprintf("(%10g <=) %10g < %25s = %10g < %10g (<= %10g)   [%4.1f%% (conf)%5.1f%%]   %s",
				      results.min[$1], results.conf_min[$1], results.name[$1], results.value[$1], results.conf_max[$1], results.max[$1],
				      100*results.buf_below[$1], 100*results.buf_above[$1], results.tex[$1]));
  print_stdout_and_file(0, sprintf("\nstatistic/dof = %f/%d = %f", stat.statistic, stat.num_bins-stat.num_variable_params, stat.statistic/(stat.num_bins-stat.num_variable_params)));
  ()=fclose(stdout_file);

  ifnot(qualifier_exists("verbose"))
  {
    ()=remove(logbasefilename+"_stdout.log");
    ()=remove(logbasefilename+"_initial.par");
  }

  return results;
}


#endif

#endif

%
% Code from src/fitting/pvm_fit_pars_txt2fits.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define pvm_fit_pars_txt2fits()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{pvm_fit_pars_txt2fits}
%\usage{pvm_fit_pars_txt2fits(String_Type txt_filename);}
%!%-
{
  variable txt_filename;
  switch(_NARGS)
  { case 1: txt_filename = (); }
  { help(_function_name()); return; }

  variable fits_filename;
  if(string_match(txt_filename, "\.txt", 1))
    (fits_filename, ) = strreplace(txt_filename, ".txt", ".fits", 1);
  else
    fits_filename = txt_filename + ".fits";

  variable t = ascii_read_table(txt_filename, [{"%F","min"},{"%s",""},{"%F","conf_min"},{"%s",""},{"%s","name"},{"%s",""},{"%F","value"},{"%s",""},{"%F","conf_max"},{"%s",""},{"%F","max"}];startline=2);
  fits_write_binary_table(fits_filename, "confidence intervals", t);
}

%
% Code from src/slang/math/qromb.sl
%
define polint( xa, ya, x)
%!%+
% NAME:
%\function{polint}
%\synopsis{Polynomial Interpolation and Extrapolation with error estimation}
%\usage{ (Double_Type y, dy ) = polint (Double_Type[] xa, ya, Double_Type x);}
%\description
%    Returns an interpolated value \code{y} at the given point \code{x} and an
%    error estimate \code{dy} for the given arrays \code{xa} and \code{ya}.
%    If P(x) is the polynomial of degree n-1 such that P(xa_i) = ya_i, i=0,n-1,
%    then the retruned value y=P(x).
%    Adapted from algorithm in Numerical Recipes,
%    by Press et al. (1992, 2nd edition), Section 3.1. 
%
%\seealso{qromb}
%!%-
{
  variable n = length([xa]);
  
  variable c = @ya;
  variable d = @ya;

  variable dift = abs( x - xa );
  variable ns = where_min( dift );
  variable dif = dift[ns];
  
  variable y = ya[ns];
  ns -= 1;

  
  variable i,m;
  variable dy;
  variable den, ho, hp, w;
  
  _for m ( 1, n-1 ){
    i   = [0:n-m-1];
    ho  = xa[i] - x;
    hp  = xa[i+m] - x;
    w   = c[i+1] - d[i];
    den = ho-hp;
    if( length(where(den == 0.)) ){
      vmessage("<%s> : ERROR (Occurs only if xa[i]=xa[j], i!=j)",_function_name);
      return 0;
    }
    den = w/den;
    d[i] = hp*den;
    c[i] = ho*den;

    if( 2*(ns+1) < n-m ){
      dy = c[ns+1];
    }
    else{
      dy = d[ns];
      ns -= 1;
    }
    y += dy;
  }
  return ( y, dy );
}

define qromb()
%!%+
% NAME:
%\function{qromb}
%\synopsis{Integrate using Rombergs's rule to specified accuracy.}
%\usage{Double_Type int = qromb (Ref_Type function, Double_Type min, Double_Type max);}
%\description
%    Integrate a function between the limits \code{min} and \code{max} to specified
%    accuracy using the extended trapezoidal rule. Adapted from algorithm
%    in Numerical Recipes, by Press et al. (1992, 2nd edition), Section 4.3.
%    The precision and number of maximal iterations can be set via qualifiers.
%    Any other keywords are passed directly to the user-supplied function.
%    NOTE: Romberg is more efficient then Simpson. 
%\qualifiers{
%\qualifier{qromb_eps [=1e-6]}{:    Scalar specifying the fractional accuracy before
%                            ending the iteration.}
%\qualifier{qromb_max_iter [=16]}{: Integer specifying the total number iterations
%                            at which \code{qsimp} will terminate even if the
%                            specified accuracy has not yet been met.
%                            The maximum number of function evaluations is
%                            2^(qromb_max_iter-2)-1.}
%}
%\qualifier{k [=5]}{:    Integration is performed by Romberg’s method
%                       of order 2k, where, e.g., k=2 is Simpson’s rule.
%}
%\example
%    %  Compute the integral of sin(x) from 0 to PI/3.
%    %  The value obtained should be cos(PI/3) = 0.5
%    variable val = qromb( &sin, 0, PI/3);
%
%\seealso{integrate_trapez,polint,qsimp}
%!%-
{
  variable a,b,func;
  switch(_NARGS)
  { case 3: (func, a, b) = (); }
  { help(_function_name()); return; }
  
   
  variable eps   = qualifier("qromb_eps",1e-6);
  variable jmax  = qualifier("qromb_max_iter",16);
  variable k     = qualifier("k",5);

  variable ss, dss;
  variable h = Double_Type[jmax+1];
  variable s = Double_Type[jmax+1];

  h[0] = 1.0;

  variable j;
  _for j ( 0, jmax-1, 1 ){
    ifnot(j)
      s[j] = 0.5*(b-a)*(@func(a;; __qualifiers)+@func(b;; __qualifiers));
    else
      s[j] = integrate_trapez (func, a, b, s[j];;
			       struct_combine (__qualifiers, struct { steps=2^(j-1)} ) );
    if( j >= k ){
      (ss,dss) = polint( h[[j+1-k:j]], s[[j+1-k:j]], 0. );
      if( abs(dss) <= abs(eps*ss) ) return ss;
    }
    s[j+1] = s[j];
    h[j+1] = 0.25*h[j];
  }
  vmessage ("<%s>: WARNING - Sum did not converge to precision of %.2e after %d steps",
	    _function_name, eps,jmax);
  vmessage ("                Change in last step was: delta_int/int = %e", abs(dss)/ss);
  return ss;
}

%
% Code from src/slang/math/qsimp.sl
%
define qsimp()
%!%+
% NAME:
%\function{qsimp}
%\synopsis{Integrate using Simpson's rule to specified accuracy.}
%\usage{Double_Type int = qsimp (Ref_Type function, Double_Type min, Double_Type max);}
%\description
%    Integrate a function between the limits \code{min} and \code{max}to specified
%    accuracy using the extended trapezoidal rule. Adapted from algorithm
%    in Numerical Recipes, by Press et al. (1992, 2nd edition), Section 4.2.
%    The precision and number of maximal iterations can be set via qualifiers.
%    Any other keywords are passed directly to the user-supplied function.
%\qualifiers{
%\qualifier{qsimp_eps [=1e-6]}{:    Scalar specifying the fractional accuracy before
%                            ending the iteration.}
%\qualifier{qsimp_max_iter [=16]}{: Integer specifying the total number iterations
%                            at which \code{qsimp} will terminate even if the
%                            specified accuracy has not yet been met.
%                            The maximum number of function evaluations is
%                            2^(qsimp_max_iter+5)-1.}
%}
%\example
%    %  Compute the integral of sin(x) from 0 to PI/3.
%    %  The value obtained should be cos(PI/3) = 0.5
%    variable val = qsimp( &sin, 0, PI/3);
%       
%\seealso{integrate_trapez,qromb}
%!%-  
{
  variable A,B,func;
  switch(_NARGS)
  { case 3: (func, A, B) = (); }
  { help(_function_name()); return; }

  variable qsimp_max_iter = qualifier("qsimp_max_iter",  16);
  variable qsimp_eps      = qualifier("qsimp_eps",     1e-6);

  variable oldS  = -1e30;
  variable delta;
  variable i,S = NULL;
  _for i (0, qsimp_max_iter -1 ,1)
  {
    S = integrate_trapez (func, A, B, S;; struct_combine (__qualifiers, struct { steps = 2^(i+4)} ) );
    delta = abs (S - oldS);
    if ( delta <= abs(qsimp_eps*oldS) ) return S;
    oldS = S;
  }
  vmessage ("<%s>: WARNING - Sum did not converge to precision of %.2e after %d steps",
	    _function_name,qsimp_eps,qsimp_max_iter);
  vmessage ("                Change in last step was: delta_int/int = %e", delta/S);
  return S;
}

%
% Code from src/fitting/xyfit/fit-functions/quadratic_xyfit.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define quadratic_xyfit()
%%%%%%%%%%%%%%%%%%%%%%%%
{
  variable xref, yref, par;
  switch(_NARGS)
  { case 0: return ["a [coefficient of x^2]", "b [coefficient of x]", "c [additive constant]"]; } % fit parameters
  { case 3: (xref, yref, par)=(); }
  { return help(_function_name); }

  @yref = par[0]*(@xref)^2 + par[1]*(@xref) + par[2];
}

define quadratic_xyfit_default(i)
{
  switch(i)
  { case 0: return (1, 0, -1e5, 1e5); } % coefficient of x^2
  { case 1: return (1, 0, -1e5, 1e5); } % coefficient of x
  { case 2: return (0, 0, -1e5, 1e5); } % additive constant
}

%
% Code from src/slang/arrays/quantile.sl
%
%%%%%%%%%%%%%%%
define quantile()
%%%%%%%%%%%%%%%
%!%+
%\function{quantile}
%\synopsis{gets an arbitrary quantile of an array}
%\usage{Any_Type quantile(Double_Type p, Any_Type a[]);}
%\description
%    \code{quantile(0, a) = min(a)} and \code{quantile(1, a) = max(a)}.
%    For values \code{0 < p < 1}, intermediate values of the \code{a} will be returned.
%\seealso{median}
%!%-
{
  variable p, a;
  switch(_NARGS)
  { case 2: (p, a) = (); }
  { help(_function_name()); return; }

  variable n = length(a);
  a = _reshape(a, n);
  return a[array_sort(a)[ int(_max(0, _min(p*(n-1), n-1))) ]];
}

%
% Code from src/fitting/fit-functions/quasar.sl
%

public variable is_table_accr_loaded = 0;
public variable quasar_table;



define quasar_accr_fit(lo,hi,par)
%!%+
%\function{quasar_accr (fit-function)}
%\synopsis{fits a composite quasar spectra in the optical/UV}
%\description        
%	This function fits a composite quasar spectrm to the data, taking into
%       account the redshift z. see Vanden Berk et al. 2001
%  
%
%\examples
%    % data definition:
%    load_data("optical.pha");
%    fit_fun("quasar_accr(1)+powerlaw(1)");
%    
%
%!%-
{

 if ( is_table_accr_loaded == 0)
 {
     quasar_table =  ascii_read_table("/userdata/data/krauss/tanami/host_galaxy/datafile1.txt", [{"%F", "wavelength"}, {"%F", "Flux"} ]); %, {"%F", "Err"}]);

     is_table_accr_loaded = 1;
 }


variable a = quasar_table;

    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5
% GET LO-HI GRID in ANGSTROM

variable b_lo=Double_Type[length(a.wavelength)];
variable b_hi=Double_Type[length(a.wavelength)];

b_lo = a.wavelength;
b_hi = make_hi_grid(b_lo);

% Angstrom rest frame -> Angstrom observed wavelength
%%%%%%%%%%%%%%%%%%%%%
variable z0 = par[1];
variable f0 = par[0];


variable blo = b_lo*(z0+1);
variable bhi = b_hi*(z0+1);

variable mblo = min(blo);
variable mablo = max(blo);



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% REBIN DATA TO FIT DATA GRID

variable flu=a.Flux;
variable new = rebin_mean (lo, hi, blo, bhi, flu);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% USE ONLY OPTICAL DATA IN DATA GRID
variable i;


_for i (0, length(new)-1,1){
    if (lo[i]<mblo){
	new[i] = 0;
    }
}

_for i (0, length(new)-1,1){
    if (hi[i]>mablo){
	new[i] = 0;
    }
}


variable f1 = new*f0;


return f1;

}
add_slang_function("quasar_accr", ["f [ ]","z [ ]"]);

private define quasar_accr_defaults(i)
{
   switch(i)
     {case 0 : return (1, 0, 0.001, 3000); } % Factor
     {case 1 : return (0.01, 1, 0, 10);  } %REDSHIFT
 
}

set_param_default_hook("quasar_accr", &quasar_accr_defaults);

%
% Code from src/slang/astronomy/radial_velocity_binary.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define radial_velocity_binary()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{radial_velocity_binary}
%\synopsis{computes the radial velocity of a binary system as a function of orbital phase or time}
%\usage{Double_Type v = radial_velocity_binary(Double_Type x);}
%\qualifiers{
%\qualifier{v0}{systemic velocity}
%\qualifier{K}{velocity semi amplitude, in km/s}
%\qualifier{asini}{value of a * sin i, in kilometers}
%\qualifier{P}{orbital period, in days}
%\qualifier{T0}{epoch of periastron, x means time in days}
%\qualifier{T90}{epoch of mean longitude 90 degr, x means time in days}
%\qualifier{e}{eccentricity (0 <= e < 1)}
%\qualifier{omega}{longitude of periastron in radian (unless degrees is set), default = 0}
%\qualifier{degrees}{omega is measured in degrees instead of radian}
%}
%\description
%    x has the meaning of orbital phase (0 <= x < 1) unless T0 or T90 is specified.
%\seealso{Ch. 3 of R.W. Hilditch, An Introduction to Close Binary Stars, Cambridge Univ. Press, 2001}
%!%-
{
  variable x;
  switch(_NARGS)
  { case 1: x = (); }
  { help(_function_name()); return; }

  variable e = qualifier("e", 0);
  variable K = qualifier("K", 1);
  variable asini = qualifier("asini", 0);
  variable P = qualifier("P", 0);
  variable omega = qualifier("omega", 0);
  if(qualifier_exists("degrees"))  omega *= PI/180.;
  variable v0 = qualifier("v0", 0);

  if(e<0. || e>=1.)
  { vmessage("error (%s): Eccentricity must be between 0 and 1.", _function_name());
    return;
  }
  if(qualifier_exists("T0") && qualifier_exists("T90"))
  { vmessage("error (%s): Only one qualifier of 'T0' and 'T90' is allowed.", _function_name());
    return;
  }
  if(qualifier_exists("K"))
  { if(asini!=0)
    { vmessage("error (%s): Either  K  *or*  asini and P  have to be specified.", _function_name());
      return;
    }
  }
  else
    if(asini!=0 && P==0)
    { vmessage("error (%s): To compute K, asini *and* P have to be specified.", _function_name());
      return;
    }

  variable M;
  if(qualifier_exists("T0") || qualifier_exists("T90"))
  {
    if(P==0)
    { vmessage("error (%s): In order to use the qualifiers 'T0' or 'T90', 'P' has to be specified as well.", _function_name());
      return;
    }
    variable T0 = qualifier("T0");
    if(qualifier_exists("T90"))  T0 = qualifier("T90") + (omega - PI/2.)/(2*PI) * P;
    M = 2*PI * (x - T0)/P;
  }
  else
    M = 2*PI * x;

  if(asini!=0)
    K = 2*PI * asini / (P*86400*sqrt((1.-e)*(1.+e)));

  variable eccanom = KeplerEquation(M, e);  % eccentric anomaly from Kepler's Equation
  variable theta = 2.*atan( sqrt( (1.+e)/(1.-e) )*tan(eccanom/2.) );  % true anomaly from eccentric anomaly
  return v0 + K * (cos(theta+omega) + e*cos(omega));  % velocity amplitude
}

%
% Code from src/misc/radio/radio_mod2img.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%%%%%%%%% subroutine for calculating a 2D-Gaussian profile (beam) with position angle
%%%%%%%%% (add this function as a separate usable funtion to the isisscripts?)
private define mb_gauss_2d (x, y, major, minor, theta, x0, y0)
{
  variable a =  sin(theta)^2/(2.0*major^2) + cos(theta)^2/(2.0*minor^2);
  variable b =  sin(2*theta)/(4.0*major^2) - sin(2*theta)/(4.0*minor^2);
  variable c =  cos(theta)^2/(2.0*major^2) + sin(theta)^2/(2.0*minor^2);

  variable A = 1/(2*PI*major*minor);
  return A * exp (- ( a*(x-x0)^2 + 2*b*(x-x0)*(y-y0) + c*(y-y0)^2 ) );
}

%%%%%%%%%%%%%%%%%%%%%%%%
define radio_mod2img ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{radio_mod2img}
%\synopsis{creates an image using a model and a beam}
%\usage{Struct_Type img = radio_mod2img( Struct_Type \code{mdl}, Double_Type \code{beam});}
%\qualifiers{
%\qualifier{src_name}{[=NULL] name of the source}
%\qualifier{date_mjd}{[=_NaN] observation date (MJD)}
%\qualifier{obs_date}{[=NULL] observation date (string). if only date_mjd or only obs_date
%                         are set, the other one is calculated.}
%\qualifier{nrpix_ra}{[=512] number of pixels for right ascension}
%\qualifier{nrpix_dec}{number of pixels in declination (by default calculated from RA-DEC-range)}
%\qualifier{ra}{[=[ra_min,ra_max]]  RA range in mas}
%\qualifier{dec}{[=[dec_min,dec_max]] DEC range in mas, by default the RA-DEC-range is
%                         calculated from the distribution of model components and the beam size}
%\qualifier{delt}{[=[ra_delt,dec_delt]] resolution in mas/pixel, overwrites nrpix qualifiers}
%\qualifier{sigma}{[=1e-3*max(img)] value for 1 sigma (used by plot_vlbi_map)}
%}
%\description
%    This function uses a model to generate an image and convolves it with the beam.
%    Here the "beam" is simply a Gaussian profile, which is given as an array:
%    \code{beam = [smajor_axis,sminor_axis,position_angle];}
%    The model has to have the fields:
%        \code{flux}   flux of each model component [Jy]
%        \code{ra}     relative RA of component (change to delta_x?)
%        \code{dec}    relative RA of component (change to delta_y?)
%        \code{smajor} component size (smajor axis) [mas] (0 for point source)
%        \code{sminor} component size (sminor axis) [mas]
%        \code{pang}   position angle of component's smajor axis
%    Currently the last two fields (\code{sminor}, and \code{pang}) are
%    ignored and only circular components (\code{smajor}) are used.
%\example
%    variable mdl = struct {
%        flux   = [0.6 , 0.9,  0.2,  1.2,  3],
%        ra     = [2   , 1.4,  0.7, 0.25,  0],
%        dec    = [1   , 0.6, 0.35,  0.1,  0],
%        smajor = [0.3 , 0.2,    0,    0,  0],
%        sminor = [0.3 , 0.2,    0,    0,  0],
%       pang   = [0    ,   0,    0,    0,  0] };
%    variable beam = [0.2, 0.07, 0.4];
%    variable img = radio_mod2img (mdl, beam);
%    plot_vlbi_map (img, "test.pdf");
%\seealso{plot_vlbi_map, read_difmap_fits}
%!%-
{
  variable mdl, beam_val;
  switch (_NARGS)
  { case 2: (mdl,beam_val) = (); }
  { help(_function_name()); return; }

  %%%%%%%% use beam:
  variable smajor = beam_val[0];
  variable sminor = beam_val[1];
  variable theta  = beam_val[2];

  variable ra_r  = qualifier ("ra",  [min( mdl.ra  - mdl.smajor - 5*smajor), max( mdl.ra  + mdl.smajor + 5*smajor)]);
  variable ra_min = max(ra_r);    variable ra_max = min(ra_r); % RA decreasing to west
  variable dec_r = qualifier ("dec", [min( mdl.dec - mdl.smajor - 5*smajor), max( mdl.dec + mdl.smajor + 5*smajor)]);
  variable dec_min = min(dec_r);  variable dec_max = max(dec_r);

  variable nrpix_ra  = qualifier("nrpix_ra",  512);
  variable nrpix_dec = qualifier("nrpix_dec", nint (512.0*abs(double(dec_max-dec_min)/(ra_max-ra_min)) ));
  
  variable RA, DEC;
  variable ra_delt,dec_delt;
  variable delt = qualifier("delt",NULL);
  variable ra_grid, dec_grid;

  if (delt == NULL)
  {
    ra_grid  = [ra_min :ra_max :#nrpix_ra ];
    dec_grid = [dec_min:dec_max:#nrpix_dec];
    ra_delt  = (ra_max  - ra_min)  / nrpix_ra ;
    dec_delt = (dec_max - dec_min) / nrpix_dec;
  }
  else
  {
    switch (length(delt))
    { case 1: ra_delt = delt[0]; dec_delt = delt[0]; }
    { case 2: ra_delt = delt[0]; dec_delt = delt[1]; }
    { throw RunTimeError, sprintf("%s: no proper stepsize (delt) provided", _function_name()); }
    
    ra_grid  = [ra_min :ra_max :ra_delt ];
    dec_grid = [dec_min:dec_max:dec_delt];
  }
  (RA, DEC) = get_grid( ra_grid, dec_grid);

  %%%%%%% create image and fill it with the model components:
  variable img_size = array_shape(RA);
  variable img = Double_Type[img_size[0],img_size[1]];
  variable nr_cmp = length (mdl.flux);
  variable i;
  _for i (0, nr_cmp-1, 1)
  {
    if (mdl.smajor[i] < 0.3*_min(abs(ra_delt),abs(dec_delt))) % point source
    {
      img[where( (RA + 0.5*ra_delt < mdl.ra[i] <= RA - 0.5*ra_delt) and (DEC - 0.5*dec_delt < mdl.dec[i] <= DEC + 0.5*dec_delt))]
	+= mdl.flux[i];
    }
    if (mdl.smajor[i] >= 0.3*_min(abs(ra_delt),abs(dec_delt))) % spherical (!) Gaussian component (add case for elliptical)
    {
      
      variable valx = 0.5 * ( gsl->erf(( ra_grid - 0.5*ra_delt - mdl.ra[i] )/(sqrt(2)*mdl.smajor[i])) -
			    gsl->erf(( ra_grid + 0.5*ra_delt - mdl.ra[i] )/(sqrt(2)*mdl.smajor[i])));
      variable valy = 0.5 * ( gsl->erf(( dec_grid + 0.5*dec_delt - mdl.dec[i] )/(sqrt(2)*mdl.smajor[i])) -
			      gsl->erf(( dec_grid - 0.5*dec_delt - mdl.dec[i] )/(sqrt(2)*mdl.smajor[i])));
      
      variable valXX, valYY;  (valXX, valYY) = get_grid(valx, valy);
      img += mdl.flux[i] * valXX * valYY;
    }
  }

  variable b_nr = ceil(5*smajor/_min(abs(ra_delt),abs(dec_delt)));
  variable beam = mb_gauss_2d ( get_grid([-b_nr:b_nr]*ra_delt, [-b_nr:b_nr]*dec_delt),
				smajor, sminor, theta, 0, 0 );
  beam /= max(beam); % keep peak flux (or keep total flux by /=sum(beam))

  variable date_mjd =  qualifier( "date_mjd", _NaN );
  variable date_str =  qualifier( "obs_date", (isnan(date_mjd) ? NULL : strftime_MJD( "%Y-%m-%d", date_mjd ) ) );
  if (isnan(date_mjd) && date_str != NULL) date_mjd = MJDofDateString (date_str);
    
  variable res = struct {
%    trueimg= img,
%    beam = beam,
    img= gsl->convolve (img, beam),

    src_name      = qualifier( "src_name", NULL ),
    date_mjd      = date_mjd,
    date_str      = date_str,
    
    ra_center     = 0,
    ra_px_center  = -sign( ra_min  *ra_delt ) * ( abs( ra_min /ra_delt )),
    ra_steps      = ra_delt,
    
    dec_center    = 0,
    dec_px_center = -sign( dec_min *dec_delt) * ( abs( dec_min/dec_delt)),
    dec_steps     = dec_delt,
    
    beam_smaj     = smajor,
    beam_smin     = sminor,
    beam_pang     = theta,
    
    mu            = 0.0,
    sigma
  };
  
  res.sigma = qualifier("sigma",1e-3*max(res.img)); 
  
  return res;
}

#endif

%
% Code from src/slang/astronomy/radius_to_unit.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define radius_to_unit()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{radius_to_unit}
%\synopsis{converts a radius from a given unit into another}
%\usage{Double_Type = radius_to_unit(Double_Type radius, String_Type from_unit, to_unit); }
%\qualifiers{
%  \qualifier{asini}{projected semi major axis (lt-s)}
%  \qualifier{i}{inclination (degrees)}
%  \qualifier{mq}{mass ratio of star to companion}
%}
%\description
%  Converts the 'radius' from a given unit 'from_unit'
%  into another unit 'to_unit'. There might be additional
%  information needed for the calculation, e.g. the
%  projected semi major axis, which is passed via
%  qualifiers.
%
%  The following units are supported (with additional
%  needed qualifiers):
%    rsun - solar radii
%    cm   - centimeters
%    lts  - light seconds
%    disp - binary displacement (asini, i, mq)
%    fill - fill factor = radius/critical radius (asini, i, mq)
%
%  Note:
%  It might be possible, that less qualifiers may be passed,
%  if the variables are canceled during the calculation. For
%  example, converting from binary displacement to fill factor
%  requires the mass ratio 'mq' only.
%
%  Note:
%  Let the projected semi major axis be the distance from CM
%  to M1. Then, mq is defined as M1/M2.
%  That means, to convert the radius of the companion, mq has
%  to be inverted and asini has to be the distance from CM
%  to M2, which is simply asini*mq[not inverted]
%\seealso{Roche_potential, Roche_critical}
%!%-
{
 variable radius, fromunit, tounit;
 switch(_NARGS)
   { case 3: (radius, fromunit, tounit) = (); }
   { help(_function_name()); return; }

 variable known_units = ["rsun","lts","disp","fill","cm"];
 % check given units
 if (wherefirst(known_units == fromunit) == NULL) { vmessage("%s: unit '%s' unkown!", _function_name, fromunit); return; }
 if (wherefirst(known_units == tounit) == NULL) { vmessage("%s: unit '%s' unkown!", _function_name, tounit); return; }
  
 variable rsol = 6.9599e10; % solar radius in cm (https://www.cfa.harvard.edu/)

 % determine needed qualifiers by specifying the conversion formula:
 % index corresponds to a parameter (rsol, asini, sin(i), (1+mq), rcrit, Const_c) and value to multiplication or division (1,-1)
 variable convert_from = [0,0,0,0,0,0], convert_to = [0,0,0,0,0,0];
 % convert input to lt-s
 switch(fromunit)
   { case "rsun": convert_from[[0,5]] = [1,-1]; } % *rsol/Const_c
   { case "cm": convert_from[5] = -1; } % /Const_c
   { case "disp": convert_from[[1,2,3]] = [1,-1,1]; } % *asini/sin(i)*(1+mq)
   { case "fill": convert_from[[1,2,3,4]] = [1,-1,1,1]; } % *asini/sin(i)*(1+mq)*rcrit
 % convert radius in lt-s to output unit
 switch(tounit)
   { case "rsun": convert_to[[0,5]] = [-1,1]; } % /rsol*Const_c
   { case "cm": convert_to[5] = 1; } % *Const_c
   { case "disp": convert_to[[1,2,3]] = [-1,1,-1]; } % /asini*sin(i)/(1+mq)
   { case "fill": convert_to[[1,2,3,4]] = [-1,1,-1,-1]; } % /asini*sin(i)/(1+mq)/rcrit
 % combine to get final conversion formula
 variable convert = convert_from + convert_to;

 % perform conversation and check on needed qualifiers
 variable asini, i, mq, rcrit;
 if (convert[0] != 0) % rsol
 {
   if (convert[0] == 1) radius *= rsol; else radius /= rsol;
 }
 if (convert[1] != 0) % asini
 {
   ifnot (qualifier_exists("asini")) { vmessage("%s: qualifier 'asini' needed to convert radius!", _function_name); return; }
   asini = qualifier("asini");
   if (convert[1] == 1) radius *= asini; else radius /= asini;
 }
 if (convert[2] != 0) % i
 {
   ifnot (qualifier_exists("i")) { vmessage("%s: qualifier 'i' needed to convert radius!", _function_name); return; }
   i = qualifier("i");
   if (convert[2] == 1) radius *= sin(1.*i*PI/180); else radius /= sin(1.*i*PI/180);
 }
 if (convert[3] != 0) % mq
 {
   ifnot (qualifier_exists("mq")) { vmessage("%s: qualifier 'mq' needed to convert radius!", _function_name); return; }
   mq = qualifier("mq");
   if (convert[3] == 1) radius *= (1.+mq); else radius /= (1.+mq);
 }
 if (convert[4] != 0) % rcrit
 {
   ifnot (qualifier_exists("mq")) { vmessage("%s: qualifier 'mq' needed to convert radius!", _function_name); return; }
   mq = qualifier("mq");
   rcrit = Roche_critical(mq).rcrit;
   if (convert[4] == 1) radius *= rcrit; else radius /= rcrit;
 }
 if (convert[5] != 0) % Const_c
 {
   if (convert[5] == 1) radius *= 2.99792458e10; else radius /= 2.99792458e10;
 }

 return radius;
}

%
% Code from src/misc/satellites/rate2mdot.sl
%
define rate2mdot (rate, sat, obj, erg)
{
  variable ratio = 0;
  if(sat=="INTEGRAL" && obj=="Vela X-1" && erg=="20-60")
    ratio = 1.04019e+35;  % This is only valid for the 20--60keV band using the spectrum from kreykenbohm08a and a rebinnend RMF version 1992a.
  else if(sat=="INTEGRAL" && obj=="Vela X-1" && erg=="20-30")
    ratio = 1.85141e+35;
  else if(sat=="INTEGRAL" && obj=="Vela X-1" && erg=="40-60")
    ratio = 7.11778e+35;
  else
    message("No known ratio value could be found!");

  variable eta = qualifier("eta", 0.1);  % accretion effiency
  variable c = 2.9979e10;  % speed of light in cm/s
  variable msun = 1.989e33;  % mass of sun in g
  variable mdot = (ratio * rate) / (eta * c^2) * (86400 * 365.2425) / msun;

  return mdot;  % in solar masses/year
}

%
% Code from src/slang/math/ratio_error_prop.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define ratio_error_prop()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{ratio_error_prop}
%\synopsis{calculates a ratio and error propagation}
%\usage{(rat, err) = ratio_error_prop(a, a_err, b, b_err);
%\altusage{(rat, err) = ratio_error_prop(a, b; Poisson);}
%}
%\description
%    \code{rat = a/b}\n
%    \code{err = sqrt[ (1/b * a_err)^2 + (a/b^2 * b_err)^2 ]}
%\qualifiers{
%\qualifier{Poisson}{\code{a_err = sqrt(a);  b_err = sqrt(b);}}
%}
%\seealso{hardnessratio_error_prop}
%!%-
{
  variable a, a_err, b, b_err;
  switch(_NARGS)
  { case 2: if(qualifier_exists("Poisson"))
            { (a, b) = (); a_err = sqrt(a); b_err = sqrt(b); }
            else
            { help(_function_name()); return; }
  }
  { case 4: (a, a_err, b, b_err) = (); }
  { help(_function_name()); return; }

  return 1.*a/b,  sqrt( (a_err/b)^2 + (a/b^2*b_err)^2 );
}

%
% Code from src/plot/read_data_from_write_plot.sl
%
define read_data_from_write_plot(fname)
%!%+
%\function{read_data_from_write_plot}
%\synopsis{reads data that was saved with "write_plot" in a structure}
%\usage{Structure str = read_data_from_write_plot(String_Type file)}
%\qualifiers{
%\qualifier{no_res}{: no residuals should be loaded}
%\qualifier{no_mod}{: no model included in dat file, implies no_res}
%\qualifier{y_fac[=0]}{: scale the y-axis by 10^{y_fac}}
%}
%\description
%    This function reads data from a plot saved with "write_plot" and returns
%    it in a structure which can be used for plotting.
%   
%    The structure tags are:
%      lo:  low energy
%      hi:  high energy
%      val: data value
%      err: uncertainty of the data value
%      model: model value of fit
%      res: residual
%      res_min: lower error bar
%      res_max: upper error bar
%
%    IMPORTANT: The filename has to be given without the ".dat" ending!
%
%    Note that if the data were saved with plot_unfold(...;...,power=3);
%    the values are automatically converted to ergs/s/cm^2
%\seealso{xfig_plot_unfold,xfig_plot_data,read_col,write_plot}
%!%- 
{
   if ( string_matches (fname,"[0-9.]\.dat"R) == NULL) fname = sprintf("%s.dat",fname);
  % convert keV to erg ???
   variable y_fac = qualifier("y_fac",0);
   variable keV2erg_fac = 1./(10^y_fac);
   
   variable b;
   variable F = fopen(fname,"r");
   while (-1 != fgets (&b, F))
   {
      if (b == "# Y-Axis    : \frkeV\u2\d Photons cm\u-2\d s\u-1\d keV\u-1\d \n"R)
      {
	 message("Using ergs/s/cm^2 by default now ...");
	 keV2erg_fac = keV2erg(1.;;__qualifiers);
	 break;
      }
   }

   
  variable lo,hi,val,err;  
  (lo, hi, val, err) =  readcol(fname, 1, 2, 3, 4);
  
  % convert to ergs
  val *= keV2erg_fac;
  err *= keV2erg_fac;

  % write it in a structure
  variable s = struct{lo=lo, hi=hi, val=val, err=err};
  
  ifnot(qualifier_exists("no_mod")){
    variable model;
    model =  readcol(fname, 5);
    model *= keV2erg_fac;

    s = struct_combine(s,struct{model=model});
    
    % are residuals given in the data-table???
    ifnot (qualifier_exists("no_res"))
    {
      variable res,res_min,res_max;
      (res, res_min, res_max) =   readcol(fname, 6,7,8);
      variable s_res = struct{  res=res, res_min=res_min, res_max=res_max  };
      s = struct_combine(s,s_res);
    }
  }
    
  return s;
}

%
% Code from src/misc/radio/read_difmap_fits.sl
%

private define try_fits_read_key_or_fill (file, key, fill)
{
  return fits_key_exists (file, key) ? fits_read_key (file, key) : fill;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define read_difmap_fits ()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{read_difmap_fits}
%\synopsis{read a fits image provided by DIFMAP}
%\usage{Struct_Type img_struct = read_difmap_fits(String_Type \code{fitsfile_name})}
%\qualifiers{
%\qualifier{fit_noise}{fit properties of the image noise}
%}
%\description
%    This function reads a .fits file provided by DIFMAP and returns
%    a structure containing the image and keywords.
%    The returned structure can be used as input for the function
%    \code{plot_vlbi_map}.
%\seealso{plot_vlbi_map}
%!%-
{
  switch(_NARGS)
  { case 1:   variable fitsfile = (); }
  { help(_function_name()); return; }
  
  variable s = struct {
    img           = fits_read_img(fitsfile),

    src_name      = try_fits_read_key_or_fill (fitsfile, "OBJECT", NULL),
    date_mjd      = MJDofDateString( try_fits_read_key_or_fill (fitsfile, "DATE-OBS", "0000-01-01")),
    date_year,
    date_str,

    ra_center     = try_fits_read_key_or_fill (fitsfile, "CRVAL1",0),            % right ascension of the center
    ra_px_center  = try_fits_read_key_or_fill (fitsfile, "CRPIX1",0),            % right ascensin center pixel
    ra_steps      = try_fits_read_key_or_fill (fitsfile, "CDELT1",0)*(3.6e+6),   % RA in mas corresponding to one pixel

    dec_center    = try_fits_read_key_or_fill (fitsfile, "CRVAL2",0),            % declination of the center
    dec_px_center = try_fits_read_key_or_fill (fitsfile, "CRPIX2",0),            % declination center pixel
    dec_steps     = try_fits_read_key_or_fill (fitsfile, "CDELT2",0)*(3.6e+6),   % DEC in mas corresponding to one pixel
    
    beam_smaj     = try_fits_read_key_or_fill (fitsfile, "BMAJ",0)*(3.6e+6)/2.0, % clean beam semi (!) major axis in mas
    beam_smin     = try_fits_read_key_or_fill (fitsfile, "BMIN",0)*(3.6e+6)/2.0, % clean beam semi (!) minor axis ins mas
    beam_pang     = try_fits_read_key_or_fill (fitsfile, "BPA" ,0)/180*PI,        % clean beam postion angle

    mu            = NULL,
    sigma         = NULL,
    sigma_difmap  = try_fits_read_key_or_fill (fitsfile, "NOISE", _NaN),
  };

  s.date_str     = strftime_MJD( "%Y-%m-%d", s.date_mjd );
  s.date_year    = yearOfMJD ( s.date_mjd );
  
  % remove 2 DIFMAP extra dimensions (FREQ and STOKES):
  variable shape = array_shape(s.img);
  shape = shape[ where( shape > 1 ) ];
  if ( length(shape) == 2 )
    reshape (s.img, shape);
  else
    throw RunTimeError, sprintf("%s: multiple images for different FREQ or STOKES?\n\t-> modify the function %s",
				_function_name(), _function_name());

  if (qualifier_exists("fit_noise"))
    (s.mu,s.sigma) = fit_gauss_to_img_noise (s.img);

  return s;
}

%
% Code from src/misc/radio/read_ep.sl
%
define read_ep(){
%!%+
%\function{read_ep}
%\synopsis{read in the epoch files}
%\usage{read_ep([array of epoch fits files]);}
%\qualifiers{
%\qualifier{quadrant}{[=1] Quadrant which is defined as a positive distance ( RA, DEC > 0 == 1 || RA < 0, DEC > 0 == 2 || RA < 0, DEC < 0 == 3 || RA > 0, DEC < 0 == 4)
%}
%}
%\description
% This functions returns a structure of an Epoch . The require input is an array of epoch fits files.
%!%-
variable ep;
switch(_NARGS)
{ case 1: (ep) = (); }
{ help(_function_name()); return; }

variable Epoch, filename, nepochs, factor, ii, jj, dist;
variable Y, m, d;
variable unit = qualifier("unit","mas");
variable quadrant = qualifier("quadrant",1);
nepochs = length(ep);
Epoch = Struct_Type[nepochs];
switch (unit)
{
case "mas":
factor = 3.6e6;
}
{
case "degree":
factor = 1.0;
}

_for ii (0, nepochs-1, 1){
	filename = ep[ii];
	Epoch[ii] = fits_read_table (filename);
	if(quadrant == 1 || quadrant == 4){
	struct_filter(Epoch[ii],array_sort(Epoch[ii].deltax;dir=-1));
	} else {
	struct_filter(Epoch[ii],array_sort(Epoch[ii].deltax));
	}
	Epoch[ii] = struct_combine(Epoch[ii],"distance","mjd","tb","freq");
	Epoch[ii].mjd = fits_read_key(filename,"DATE-OBS");
	()=sscanf(Epoch[ii].mjd, "%4d-%2d-%2d, %2d:%2d", &Y, &m, &d);
	Epoch[ii].mjd = MJDofDate(Y, m, d);
	Epoch[ii].freq = fits_read_key(filename,"CRVAL3")/1.e9; %in GHz
	ifnot (qualifier_exists("zero")) {
	Epoch[ii].deltax = factor * Epoch[ii].deltax;
	Epoch[ii].deltay = factor * Epoch[ii].deltay;
	} else {
	%still does not work for counterjets
	Epoch[ii].deltax = factor * (Epoch[ii].deltax - Epoch[ii].deltax[length(Epoch[ii].deltax) - 1]);
	Epoch[ii].deltay = factor * (Epoch[ii].deltay - Epoch[ii].deltay[length(Epoch[ii].deltay) - 1]);
	}
	Epoch[ii].distance = sqrt(sqr(Epoch[ii].deltax) + sqr(Epoch[ii].deltay));
	Epoch[ii].major_ax = factor * Epoch[ii].major_ax;
	Epoch[ii].minor_ax = factor * Epoch[ii].minor_ax;
	Epoch[ii].tb = 1.22*1e+12*Epoch[ii].flux/(sqr(Epoch[ii].freq)*(Epoch[ii].major_ax)*(Epoch[ii].minor_ax)); % see Condon et al. 1982
	_for jj (0, length(Epoch[ii].flux) - 1, 1){
	  if(quadrant == 1){
	    if(Epoch[ii].deltax[jj] < 0 || Epoch[ii].deltay[jj] < 0){
	      Epoch[ii].distance[jj] = - sqrt(sqr(Epoch[ii].deltax[jj]) + sqr(Epoch[ii].deltay[jj]));
	  }}
	  if(quadrant == 2){
	    if(Epoch[ii].deltax[jj] > 0 && Epoch[ii].deltay[jj] < 0){
	      Epoch[ii].distance[jj] = - sqrt(sqr(Epoch[ii].deltax[jj]) + sqr(Epoch[ii].deltay[jj]));
	  }}
	  if(quadrant == 3){
	    if(Epoch[ii].deltax[jj] > 0 || Epoch[ii].deltay[jj] > 0){
	      Epoch[ii].distance[jj] = - sqrt(sqr(Epoch[ii].deltax[jj]) + sqr(Epoch[ii].deltay[jj]));
	  }}
	  if(quadrant == 4){
	    if(Epoch[ii].deltax[jj] < 0 && Epoch[ii].deltay[jj] > 0){
	      Epoch[ii].distance[jj] = - sqrt(sqr(Epoch[ii].deltax[jj]) + sqr(Epoch[ii].deltay[jj]));
	  }}
	}

};

return Epoch;
}

%
% Code from src/slang/structs/histograms/read_histo.sl
%
define read_histo(inf){
%!%+
%\function{read_histo}
%\synopsis{read text data into histogram structure}
%\usage{Struct_Type hist = read_histo(String_Type filename);}
%\qualifiers{
%\qualifier{cols}{[=4] number of columns in the data file}
%\qualifier{collist}{array of column index for bin_lo, bin_hi,
%                    value, err in that order. Needs col=-1.}
%\qualifier{bin_lo}{column index for bin_lo. Needs col=-1.}
%\qualifier{bin_hi}{column index for bin_hi. Needs col=-1.}
%\qualifier{value}{column index for value. Needs col=-1.}
%\qualifier{err}{column index for err. Needs col=-1.}
%}
%\description
%    Read a text file with column data directly into a histogram
%    structure struct{bin_lo,bin_hi,value,err};. 
%    The default assumption is that the text file contains the four
%    columns bin_lo, bin_hi, value, and err in that order as the first
%    four columns in the file (further columns being ignored). 
%    Non-standard files can be processed via qualifiers. Missing
%    columns are populate with assumptions, e.g., the grid reflecting
%    the row numbers and the uncertainty assuming Poisson statistics.
%    If all columns are present, but out of order / other column
%    numbers, a list with column numbers can be supplied.
%    
%    Qualifier cols:
%        cols=4:  bin_lo, bin_hi, value, err
%        cols=1:  value. 
%                 Then bin_lo=[0:length(value)-1], bin_hi=bin_lo+1,
%                 err=sqrt(value). 
%        cols=2:  bin_lo, value. 
%                 Then bin_hi=make_hi_grid(lo), err=sqrt(value).
%        cols=3:  bin_lo, bin_hi, value.
%                 Then err = sqrt(value).
%        cols=-1: Either use collist qualifier to supply an array of
%                 column indices for all of bin_lo, bin_hi, value, err 
%                 (in this order). Or, for a sub-selection, use the
%                 bin_lo, bin_hi, value, and err to supply column
%                 indices individually. If any of those four equals 0,
%                 this field is ignored. Missing columns are populated
%                 as described above. 
%                 Presence of collist takes presedence over the others. 
%\seealso{init_histo, add_hist, shift_hist, 
%scale_hist, stretch_hist}
%!%-   
   variable cols = qualifier("cols",4);
   variable s = struct{bin_lo,bin_hi,value, err};
   switch(cols)
	 { case 1: 
		  s.value = readcol(inf,1);
		s.err = sqrt(s.value);
		s.bin_lo = [0: length(s.value)-1];
		s.bin_hi = [1: length(s.value)];
	 }
   	 { case 2: 
		  (s.bin_lo, s.value) = readcol(inf,1,2);
		s.err = sqrt(s.value);
		s.bin_hi = make_hi_grid(s.bin_lo);
	 }
   	 { case 3: 
		  (s.bin_lo, s.bin_hi, s.value) = readcol(inf,1,2,3);
		s.err = sqrt(s.value);
	 }
	 { case 4: 
		  (s.bin_lo, s.bin_hi, s.value, s.err) = readcol(inf, 1,2,3,4);
	 }
	 { case -1:
		  variable bl, bh, val, err, all;
		if(qualifier_exists("collist")){
		   all = qualifier("collist");
		   bl = all[0]; bh = all[1]; val = all[2]; err=all[3];
		}else{
		   bl = qualifier("bin_lo",0);
		   bh = qualifier("bin_hi",0);
		   val = qualifier("value",0);
		   err = qualifier("err",0);
		}
		if(val) s.value = readcol(inf, val);
		if(err) s.err = readcol(inf, err); else s.err = sqrt(s.value);
		if(bl) s.bin_lo = readcol(inf, bl); else s.bin_lo = [0:length(s.value)-1];
		if(bh) s.bin_hi = readcol(inf,bh); else s.bin_hi = make_hi_grid(s.bin_lo);
	 }
   return @s;
}

%
% Code from src/sixte/read_sixte_lc.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define read_sixte_lc(fname){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{read_sixte_lc}
%\synopsis{Reads a LC created by the SIXTE makelc tool}
%\description
%       makelc writes the lightcurve file according to the first OGIP
%       standard, i.e. only a COUNTS column is written and the timing
%       information is stored in the TSTART, TSTOP, TIMEDEL, and
%       TIMEPIXR keywords. SIXTE uses the convention TIMEPIXR=0, which
%       means that the time field of the lightcurve returned by
%       read_sixte_lc equals the beginning of the bin (time_lo).
%\usage{read_sixte_lc(filename);}
%!%-
  variable lc = struct{time, rate, error};
  variable raw = fits_read_table(fname);
  
  variable dt = fits_read_key(fname, "TIMEDEL");
  variable timepixr = fits_read_key(fname, "TIMEPIXR");
  
  variable n = length(raw.counts);
  variable time_lo = dt*[0:n-1];
  % variable time_hi = dt*[1:n];
  
  if (timepixr != 0.){
    vmessage("WARNING: TIMEPIXR equals %g! Assuming TIMEPIXR=0 (start of time bin)", timepixr);
  }
  
  lc.time = time_lo;
  lc.rate = raw.counts/dt;
  lc.error = sqrt(raw.counts)/dt;
  
  return lc;
}

%
% Code from src/misc/radio/read_spix.sl
%
define read_spix()
%!%+
%\function{read_spix}
%\synopsis{reads the FITS file created with write_spix}
%\usage{Stuct_Type \code{spix} = read_spix( String_Type \code{filename});}
%\seealso{make_spix, write_spix, plot_spix}
%!%-
{
  variable filename;
  switch(_NARGS)
  { case 1: (filename) = (); }
  { help(_function_name()); return; }


  variable spixm = fits_read_table ( filename );

  variable spix    = struct {
    spec_map       = spixm.spec_map,
    lo_map         = spixm.lo_map,
    hi_map_shifted = spixm.hi_map_shifted,
    stdev          = spixm.stdev,
    avg_lum        = spixm.avg_lum,
    weights        = spixm.weights,
    avg_shift      = [fits_read_key(filename , "avg_xshift" ),fits_read_key(filename , "avg_yshift" )],
    shift_weight   = fits_read_key(filename , "shift_weight" ),
    ra_px_center   = fits_read_key(filename , "ra_px_center" ),
    ra_steps       = fits_read_key(filename , "ra_steps" ),
    dec_px_center  = fits_read_key(filename , "dec_px_center" ),
    dec_steps      = fits_read_key(filename , "dec_steps" ),
    major          = fits_read_key(filename , "major" ),
    minor          = fits_read_key(filename , "minor" ),
    posang         = fits_read_key(filename , "posang" ),
    source         = fits_read_key(filename , "source" ),
    date           = fits_read_key(filename , "date" )
  };

  return spix;
}

%
% Code from src/histogram/rebinDensity.sl
%
%%%%%%%%%%%%%%%%%%%
define rebinDensity()
%%%%%%%%%%%%%%%%%%%
{ variable newLo, newHi, lo, hi, value;
  if(_NARGS!=5)
  { message("usage: newValue = rebinDensity(newLo, newHi,  lo, hi, value);"); return; }
  else
  { (newLo, newHi, lo, hi, value) = ();
    return rebin(newLo, newHi,  lo, hi, value*(hi-lo))/(newHi-newLo);
  }
}

%
% Code from src/data/spectrum.datastructures/rebinGroup.sl
%
define rebinGroup()
%!%+
%\function{rebinGroup}
%\synopsis{rebins a \code{{bin_lo, bin_hi, value, err} struct}ure by a grouping factor}
%\usage{Struct_Type s_ = rebinGroup(Struct_Type s, Integer_Type n);}
%\seealso{rebin}
%!%-
{ variable s, lo, hi, value, err=NULL, nGrouping;
  switch(_NARGS)
  { case 2:
    (s, nGrouping) = ();
    if(_typeof(s) != Struct_Type) { message("usage: groupedStruct = rebinDensity(lo, hi, value[, err], nGrouping);"); return; }
    if(length(get_struct_field_names(s))>3)
    { return rebinGroup(s.bin_lo, s.bin_hi, s.value, s.err, nGrouping); }
    else
    { return rebinGroup(s.bin_lo, s.bin_hi, s.value, nGrouping); };
  }
  { case 4: (lo, hi, value, nGrouping) = (); }
  { case 5: (lo, hi, value, err, nGrouping) = (); }
  { help(_function_name()); return; }

  if(err==NULL) { s = struct { bin_lo, bin_hi, value }; } else { s = struct { bin_lo, bin_hi, value, err }; }
  s.bin_lo = lo[ [0:length(lo)-1:nGrouping] ];
  s.bin_hi = make_hi_grid(s.bin_lo);
  s.value = rebin(s.bin_lo, s.bin_hi,  lo, hi, value);
  if(err!=NULL) { s.err = sqrt( rebin(s.bin_lo, s.bin_hi,  lo, hi, err^2) ); }
  return s;
}

%
% Code from src/data/spectrum.loaddata/rebin_error_hook.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define rebin_error_hook(orig_cts, orig_stat_err, grouping)
%%%%%%%%%%%%%%%%%%%%%%%
{
  return sqrt(rebin_array(orig_stat_err^2, grouping));
}

%
% Code from src/data/timing/rebin_fouquan.sl
%
define rebin_fouquan()
%!%+
%\function{rebin_fouquan}
%\synopsis{rebins a Fourier quantity from timing analysis with foucalc}
%\usage{Struct_Type rebin_fouquan(freq, fouquan, numseg)}
%\description
%    \code{freq} is an array of the original Fourier frequencies.
%    \code{fouquan} is the array of original Fourier quantities.
%    \code{numseg} is the number of seqments used to calculate \code{fouquant}.
%
%    The output structure has the following fields, which are arrays:
%    - \code{freq_lo} and \code{freq_hi} define the new frequency bin.
%    - \code{freq} is the average original frequency in each bin.
%    - \code{n} is the number of original frequencies, and
%    - \code{n_tot} is the total number of values (including segmentation) contributing to each bin.
%    - \code{value} is the average Fourier quantity.
%    - \code{error = value/sqrt(n_tot)} is the 1 sigma error for the \code{n_tot} values.
%    - \code{sigma} is the standard deviation of the original quanities in each bin
%\qualifiers{
%\qualifier{logfreq}{}
%\qualifier{linfreq}{}
%\qualifier{newfreq}{array of new frequencies}
%\qualifier{nofreq}{no rebinning}
%\qualifier{verbose}{}
%}
%!%-
{
  variable fou, fields, field, freq, fouquan, numseg;
  variable new = struct { freq_lo, freq_hi, freq, n=NULL, n_tot };
  switch(_NARGS)
  { case 1: fou = ();
      fields = String_Type[0];
      foreach field (get_struct_field_names(fou))
	if(string_match(field, "^signormpsd", 1))
	  fields = [fields, field];
      return rebin_fouquan(fou, fields;; __qualifiers());
  }
  { case 2: (fou, field) = ();
      switch(typeof(field))
      { case String_Type:
          freq = fou.freq;
          fouquan = get_struct_field(fou, field);
          numseg = get_struct_field(fou, "numavgall");
      }
      { case Array_Type:
	  fields = field;
	  foreach field (fields)
	  {
	    variable tmp = rebin_fouquan(fou, field;; __qualifiers());
	    if(new.n==NULL)
	    {
	        (new.freq_lo, new.freq_hi, new.freq, new.n, new.n_tot)
	      = (tmp.freq_lo, tmp.freq_hi, tmp.freq, tmp.n, tmp.n_tot);
	    }
	    variable tmp_new = @Struct_Type(field+["", "_error", "_sigma"]);
	    set_struct_fields(tmp_new, tmp.value, tmp.error, tmp.sigma);
	    new = struct_combine(new, tmp_new);
	  }
	  return new;
      }
      { % else:
	  vmessage("error (%s): %S is not valid for field", _function_name(), typeof(field));
          return;
      }
  }
  { case 3: (freq, fouquan, numseg) = (); }
  { help(_function_name()); return; }

  variable fmin = min(freq);
  variable fmax = max(freq);

  % return structure
  new = struct_combine(new, struct { value, error, sigma });

  % check for qualifiers
  variable verbose = qualifier_exists("verbose");
  variable linfreq = qualifier("linfreq", NULL);
  variable logfreq = qualifier("logfreq", NULL);
  variable newfreq = qualifier("newfreq", NULL); % <<< qualifier newfreq has to be an array!

  % computing new frequency array
  if(qualifier_exists("nofreq"))
    new.freq_lo = freq;
  else
  {
    if((logfreq==NULL) && (linfreq==NULL) && (newfreq==NULL))
    { vmessage("warning (%s): no frequency rebinning specified. Assuming logfreq=0.15.", _function_name());
      logfreq = 0.15;
    }
    if(newfreq != NULL)
      new.freq_lo = newfreq;
    if(linfreq != NULL)
      new.freq_lo = fmin + [0:linfreq]*(fmax-fmin)/linfreq;
    if(logfreq != NULL)
      new.freq_lo = fmin * (1.+logfreq)^[0:log(fmax/fmin)/log(1.+logfreq)];
  }

  new.freq_hi = make_hi_grid(new.freq_lo);
  variable i, n = length(new.freq_lo);  % length of new arrays
  new.freq    = Double_Type[n];
  new.n       = Integer_Type[n];
  new.n_tot   = Integer_Type[n];
  new.value   = Double_Type[n];
  new.error   = Double_Type[n];
  new.sigma   = Double_Type[n];
  _for i (0, n-1, 1)
  {
    variable index = where(new.freq_lo[i] <= freq < new.freq_hi[i]);
    variable n_freqs = length(index);
    new.n[i] = n_freqs;
    if(typeof(numseg)==Array_Type)
      new.n_tot[i] = int(sum(numseg[index]));
    else
      new.n_tot[i] = n_freqs * numseg;
    if(n_freqs == 0)
    { if(verbose)
	vmessage("  ***(%s): Empty frequency-bin (#%d) found in %g-%g Hz.", _function_name(), i, new.freq_lo[i], new.freq_hi[i]);
    }
    else
    {
      variable av_value = sum(fouquan[index])/n_freqs;
      new.value[i] = av_value;
      new.error[i] = av_value/sqrt(new.n_tot[i]);  % one sigma error for n_tot measurements
      if(new.n_tot[i]>1)
        new.sigma[i] = sqrt( sum(sqr(fouquan[index])-av_value^2) / (new.n_tot[i]-1.)); % standard deviation of all the old psd values in the new frequency bin
      else
        new.sigma[i] = new.psd[i];

%     variable av_f_value = sum(fouquan[index]*freq[index])/n_freqs;
%     new.freq[i] = av_f_value / av_value;  % weighted average
      new.freq[i] = sum(freq[index])/n_freqs;
    }
  }

  index = where(new.n==0);
  while(length(index)>0)  % filtering for empty bins
  {
    if(index[0]==n-1)
      struct_filter(new, [:-2]);  % just discard the last bin
    else
    { new.freq_lo[index[0]+1] = new.freq_lo[index[0]];  % enlarge next bin
      struct_filter(new, [[0:index[0]-1], [index[0]+1:n-1]]);  % remove the empty bin
    }
    n--;
    index = index[[1:]]-1;
  }
  % struct_filter(newPSD, where(newPSD.n>0));

  return new;
}

%
% Code from src/data/rebin_human2isis.sl
%
define rebin_human2isis(reb)
%!%+
%\function{rebin_human2isis}
%\synopsis{create isis binning from str = rebin_human2isis}
%\usage{Array_Type binning = rebin_human2isis(String_Type str);}
%\description
%   This routine converts a binning string created with
%   rebin_isis2human back to the ISIS conform binning array
%   (see e.g. get_data_info(idx).rebin). It can be used to
%   restore the binning of data directly.
%   
%\seealso{fits_save_fit, get_data_info, rebin_isis2human}
%!%-
{

  variable tp = strtok(reb,",");
  
  variable s0 = Integer_Type[0];
  variable sg = 1;
  
  variable t,sub,is,i_lo,i_hi,n,i,j,del;
  foreach t(tp)
  {
    sub = strtok(t, ":");
    if (length(sub) == 1)
    {
      s0 = [s0, sg];
      sg *=-1;
    }
    else
    {
      is = strtok(sub[0],"-");
      i_lo = integer(is[0]);
      i_hi = integer(is[1]);
      del = i_hi-i_lo+1;
      n = integer(sub[1]);
            
      _for i(1,del,1)
      {
	s0 = [s0, sg];	  
	if (i mod n == 0) sg *=-1;
      }
      
    }
  }

  return s0;
}

%
% Code from src/data/rebin_isis2human.sl
%
define rebin_isis2human(reb)
%!%+
%\function{rebin_isis2human}
%\synopsis{converts isis binning in a human readable string}
%\usage{String_Type str = rebin_isis2human(Array_Type binning);}
%\description
%   This routine converts the binning as used by ISIS (see e.g. 
%   get_data_info(idx).rebin) to a string which is easy to read.
%   
%   The routine fits_save_fit() uses this string to save the binning
%   of the data. It can be converted back by using rebin_human2isis().
%\seealso{fits_save_fit, get_data_info, rebin_human2isis}
%!%-
{
  variable ind = 0;
  variable ind_0 = ind;  
  variable comma = "";
  variable result = "";
  variable str;
  variable ct = 0;
  variable diff=-1;
  variable diff_0 = -1;
  variable sg = sign(reb[0]);
  
  variable i;
  _for i(1,length(reb),1)
  {
    if (i == length(reb) || sign(reb[i]) != sg )
    {
      
      ct++;
      diff = i-ind;
      if ((diff != diff_0))
      {
	if ( diff_0 != -1) 
	{
	  if (ct == 1 and diff_0 == 1)
	  {
	    str = comma+string(ind-1);
	  }
	  else
	  {
	    str= comma+string(ind_0)+"-"+string(ind-1)+":"+string(diff_0);
	  }
	  result +=str;
	  comma = ",";
	}
	if (i < (length(reb)-1)) ct = 0;
	ind_0 = ind;
	diff_0 = diff;	  
      }
      
      ind = i;
      if (i < length(reb)) sg = sign(reb[i]);
      
    }
  }
  
  if (ct == 1 and diff_0 == 1)
  {
    str = comma+string(ind-1);
  }
  else
  {
    str= comma+string(ind_0)+"-"+string(ind-1)+":"+string(diff_0);
  }
  result +=str;
  
  return result;
}

%
% Code from src/data/lightcurve/rebin_lc.sl
%
define rebin_lc()
%!%+
%\function{rebin_lc}
%\synopsis{rebins a lightcurve structure to a new time resolution}
%\usage{Struct_Type new_lc = rebin_lc(Struct_Type lc, Double_Type new_dt);}
%\qualifiers{
%\qualifier{time}{[= \code{"time"}] field name in the structure containing the time}
%\qualifier{rate}{[= \code{"rate"}] field name(s) in the structure containing the rate(s)}
%\qualifier{error}{[= \code{"error"}] field name(s) containig the corresponding error(s)}
%\qualifier{float}{type cast rate(s) and error(s) to Float_Type}
%\qualifier{verbose}{shows the assumed time resolution of the initial light curve}
%}
%\description
%    The original light curve \code{lc} may contain discontinuities.
%    \code{new_lc.rate[i]} contains the average \code{lc.rate}
%    where \code{new_lc.time[i] <= lc.time < new_lc.time[i]+new_dt}.
%    \code{error^2} is rebinned accordingly.
%
%    The structure fields "\code{time, rate, error}" may have arbitrary names,
%    but these must then be specified by the according qualifiers.
%    \code{new_lc} will also have  these field names. In addition, a "\code{time_hi}"
%    field is added, which contains the upper boundary of the time bins.
%\example
%    \code{lc1 = struct { time=[1:10], rate=[1:10]^2, , error=[1:10] };}\n
%    \code{LC1 = rebin_lc(lc1, 2);}\n
%
%    \code{lc2 = struct { t=[1:10], r1=[1:10]^2, r2=[1:10]^3, e1=[1:10], e2=[1:10]^1.5 };}\n
%    \code{LC2 = rebin_lc(lc2, 2; time="t", rate=["r1", "r2"], error=["e1", "e2"]);}\n
%\seealso{rebin}
%!%-
{
  variable lc, new_dt;
  switch(_NARGS)
  { case 2: (lc, new_dt) = (); }
  { help(_function_name()); return; }

  variable timefield = qualifier("time", "time");
  ifnot(struct_field_exists(lc, timefield))
    return vmessage("error (%s): structure has no field '%s'", _function_name(), timefield);
  variable time_lo = get_struct_field(lc, timefield);
  variable dt = min( time_lo[[1:]] - time_lo[[:-2]] );
  variable time_hi = time_lo + dt;  % Do not use make_hi_grid here; light curve might contain gaps!
  if(qualifier_exists("verbose"))
    vmessage("%s assumes initial time resolution %g", _function_name(), dt);
  if(dt==0)
    vmessage("error (%s): minimum time difference of input light curve is zero", _function_name());
  if(new_dt < dt)
    vmessage("warning (%s): new_dt=%g < dt=%g (=> interpolation)", _function_name, new_dt, dt);

  variable t = [min(time_lo) : max(time_lo) : new_dt];  %| new
  variable t_hi = make_hi_grid(t);                      %| bins
  variable lc_ = @Struct_Type([timefield+["", "_hi"], "n"]);
  set_struct_fields(lc_, t, t_hi, rebin(t, t_hi, time_lo, time_hi, 0*time_lo+1));
  lc_.n[where(lc_.n==0)] = -1;

  variable ratefields  = [ qualifier("rate",  "rate")  ];  %| always make ratefields
  variable errorfields = [ qualifier("error", "error") ];  %| and errorfields be arrays
  variable n_rate = length(ratefields);  %| here /number/ of rate and error fields;
  variable n_err = length(errorfields);  %| but will change below (twice)
  if(n_rate != n_err)
  { variable m = min([n_rate, n_err]);
    vmessage("warning (%s): length(rate fields)=%d != length(error fields)=%d, only using first %d fields", _function_name(), n_rate, n_err, m);
    (n_rate, n_err) = (m, m);
  }

  variable float = qualifier_exists("float");

  variable i;
  _for i (0, n_rate-1, 1)  % /numbers/, won't be needed below this line
  { (n_rate, n_err) = (ratefields[i], errorfields[i]);  % now /name/ of rate and error fields; but will change below
    ifnot(struct_field_exists(lc, n_rate))
      return vmessage("error (%s): structure has no field '%s'", _function_name(), n_rate);
    ifnot(struct_field_exists(lc, n_err))
      return vmessage("error (%s): structure has no field '%s'", _function_name(), n_err);
    variable s = @Struct_Type([n_rate, n_err]);         % /names/, won't be needed below this line
    (n_rate, n_err) = (      rebin(t, t_hi,  time_lo, time_hi, get_struct_field(lc, n_rate) ) /lc_.n,   % now /new/ rate field
                       sqrt( rebin(t, t_hi,  time_lo, time_hi, get_struct_field(lc, n_err)^2))/lc_.n ); % now /new/ error field
    if(float)
     (n_rate, n_err) = (typecast(n_rate, Float_Type), typecast(n_err, Float_Type));
    set_struct_fields(s, n_rate, n_err);
    lc_ = struct_combine(lc_, s);
  }
  struct_filter(lc_, where(lc_.n>0));
  return lc_;
}

%
% Code from src/histogram/rebin_mean.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define rebin_mean()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{rebin_mean}
%\synopsis{rebins units like intensity, where you expect the binning
%to take the mean of the value, properly}
%\usage{intens_new = rebin_mean(r_nlo,r_nhi,r_lo,r_hi,intens);}
%!%-
% 
{
   variable r_nlo,r_nhi,r_lo,r_hi,intens;
   switch(_NARGS)
   { case 5: (r_nlo,r_nhi,r_lo,r_hi,intens) = (); }
   {    help(_function_name()); return; }
   
   
   variable rm = 0.5*(r_lo+r_hi);
   variable rm_new = 0.5*(r_nlo+r_nhi);
   variable l_new = length(rm_new);
   variable l = length(rm);
   variable intens_new = Double_Type[l_new];
  
  
  
  variable i,ind;  
  _for i(0,l_new-1,1)
  {    
    if (rm_new[i]<rm[0])
    {
      %special case: LOWER END extrapolation if new grid is a little larger than
      %the old one
      intens_new[0] = (intens[1]-intens[0])/(rm[1]-rm[0])*(rm[0]-rm_new[i])+intens[0];
    }
    else if (rm_new[i]>rm[-1]) 
    {
     %special case: UPPER END
     %extrapolation if new grid is a little larger than the old one
      intens_new[i] = (intens[l-2]-intens[l-1])/(rm[l-2]-rm[l-1])*(rm_new[i]-rm[l-1])+intens[l-1];
    }
    else
    {
      ind = where(rm<=rm_new[i])[-1];
      
      intens_new[i] = ((rm[ind+1]-rm_new[i])*intens[ind]+(rm_new[i]-rm[ind])*intens[ind+1])/
	(rm[ind+1]-rm[ind]);
    }
  }
    
  return intens_new;
}

%
% Code from src/data/rebin_to_energy_grid.sl
%
define rebin_to_energy_grid(id,elo,ehi)
%!%+
%\function{rebin_to_energy_grid}
%\synopsis{rebin data to a given energy grid}
%\usage{rebin_to_energy_grid(Dataset ID, bin_lo, bin_hi);}
%\description
%   This routine bins the data to a given energy grid. Bins outside
%   the given energy grid are left unbinned.
%\seealso{rebin_data,rebin_satellite, rebin_human2isis}
%!%-
{
   variable info = get_data_info(id).rebin;
   % go to the original binning
   rebin_data(id,0);

   % get the (un-binned) grid 
   variable e = _A(get_data_counts(id)); % take care to be in energy space
   variable nelo = e.bin_lo;   variable nehi = e.bin_hi;

   variable n = length(nelo); variable m = length(elo);
   variable n_reb = Integer_Type[n];
   
   variable i,j;
   % initalize 
   variable val = +1;j=0;

   %  ====  LOOP  ====  %
   _for i(0,n-1,1)
   {
      % see if the bin has to be changed
      if (elo[j]<nelo[i])
      {
	 val*=-1;
	 if (j < m-1) j++;
      }
      if (elo[0]>nelo[i]) val*=-1; % leave it unbinned outside the energy range
      
      n_reb[i] = val;
   }
   %  ================  %
   
   % rebin spectrum
   rebin_data(id,n_reb);
      
   return reverse(n_reb);
}

%
% Code from src/data/rebin_to_instrument.sl
%
define rebin_to_instrument(ni,no)
%!%+
%\function{rebin_to_instrument}
%\synopsis{rebin data from one instrument to another instrument}
%\usage{rebin_to_instrument(Dataset id_reference, Dataset id_to_be_rebinned);}
%\description
%   This routine tries its best to rebin the given dataset with the ID
%   "id_to_be_rebinned" to the dataset with id "id_reference".
%\seealso{rebin_to_energy_grid,rebin_data}
%!%-
{
   variable d = _A(get_data_counts(ni));
   rebin_to_energy_grid(no,d.bin_lo,d.bin_hi);
   
   return;
}

%
% Code from src/slang/arrays/2d/rectangle_where.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define rectangle_where()
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{rectangle_where}
%\synopsis{finds the rectangle in a 2d array, where an expression is true}
%\usage{(Integer_Type Y[], X[]) = rectangle_where(Integer_Type expr);}
%\description
%    The rectangle in the 2d array defined by the index-arrays
%    \code{Y} and \code{X} contains all elements for which \code{expr!=0},
%    i.e., one can crop the elements for which \code{expr==0}.
%    (Note that \code{all(expr[Y,X])} is not necessarily true.)
%\example
%    img = img[rectangle_where(img>0)];
%!%-
{
  variable expr;
  switch(_NARGS)
  { case 1: expr = (); }
  { help(_function_name()); return; }

  variable dims = array_shape(expr);
  variable w=dims[1], h=dims[0];
  variable xmin = 0;    while(xmin<w-1 && all(expr[*,xmin]==0))  xmin++;
  variable xmax = w-1;  while(xmax>0   && all(expr[*,xmax]==0))  xmax--;
  variable ymin = 0;    while(ymin<h-1 && all(expr[ymin,*]==0))  ymin++;
  variable ymax = h-1;  while(ymax>0   && all(expr[ymax,*]==0))  ymax--;

  return [ymin:ymax], [xmin:xmax];
}

%
% Code from src/slang/geometry/rectangles_overlap.sl
%
define rectangles_overlap (src,clp)
%!%+
%\function{rectangles_overlap}
%\synopsis{Checks whether two rectangles overlap}
%\usage{res=rectangles_overlap(src,clp)}
%\description
%
% This function checks whether the two rectangles src and clp
% overlap (return value=1) or not (return value =0).
%
% The rectangles are defined by (xmin,ymin,xmax,ymax)
% either as a 4 element array in this order or as a
% struct with these tags. Both boxes must be defined
% using the same format.
%
%!%-
{
    if (typeof(src)!=typeof(clp)) {
        throw UsageError,"%s: src and clp must have the same type!",_function_name();
    }

    variable xmin,xmax,ymin,ymax;
    if (typeof(src)==Array_Type) {
        %  (sxmax<=cxmin || sxmin>=cxmax)
        if (src[2]<=clp[0] || src[0]>=clp[2]) {
            return 0;
        }
        
        % if (symax<=cymin || symin>=cymax)
        if (src[3]<=clp[1] || src[1]>=clp[3]) {
            return 0;
        }
        return 1;
    }

    if (src.xmax<=clp.xmin || src.xmin>=clp.xmax) {
        return 0;
    }
        
    if (src.ymax<=clp.ymin || src.ymin>=clp.ymax) {
        return 0;
    }

    return 1;

}

%
% Code from src/fitting/fit-functions/redden.sl
%
define redden_fit(lo, hi, par)
%!%+
%\function{redden (fit-function)}
%\synopsis{fits the reddening of optical/UV data}
%\description
%	This function uses fm_unred to fit the reddening of optical/UV data.
%	See fm_unred for details.
%
%\examples
%    % data definition:
%    load_data("optical.pha");
%    fit_fun("redden(1)*powerlaw(1)");
%    
%
%\seealso{fm_unred}
%!%-
{
  %%%%%%%%%%%%%%%%%%%
  %  variable wave= struct{  V=5468, U=3465, B=4392, UVW1=2600, UVW2=1928, UVM2=2246 };
  %  We don't need the wavelength, since middle of bin energy \sim  wavelength%%%%
  %%%%%%%%%%%%%%%%%%%%%%%%5

    variable flux=lo*0+1.;       %We only need the factor
    variable wave = 0.5* (lo+hi);

    %%%%%%%%%%%%%%%%%%%%%%
    variable nh=par[1]*1e22;
    variable rv=par[0];
    variable fac = fm_unred(wave, flux; N_H=nh, R_V=rv,wilm);
    fac[where (fac>1e308)]=1;  %should not be this high -> Need to fix energy range sometime soon
    return 1.0/(fac);
}

add_slang_function("redden", ["R [ ]","nH [10^22/cm^2]"]);

private define redden_defaults(i)
{
    switch(i)
    {case 0 : return (3.1, 1, 2, 8); } % Extinktion: Default R_V= 3.1
    {case 1 : return (0.1, 1, 0, 50);  }
    
}

set_param_default_hook("redden", &redden_defaults);

%
% Code from src/slang/arrays/2d/reduce_image_resolution.sl
%
define reduce_image_resolution()
{
  variable img, nx, ny;
  switch(_NARGS)
  { case 2: (img, nx) = (); ny = nx; }
  { case 3: (img, nx, ny) = (); }
  { help(_function_name()); return; }

  variable size = array_shape(img);
  variable h = size[0]/ny;
  variable w = size[1]/nx;
  variable factor = 1.0/nx/ny;
  variable img_new = Double_Type[h, w];
  variable x, y;
  _for y (0, h-1, 1)
    _for x (0, w-1, 1)
      img_new[y,x] = sum( img[y*ny+[0:ny-1], x*ny+[0:nx-1]] ) * factor;
  return img_new;
}

%
% Code from src/slang/structs/reduce_struct.sl
%
%%%%%%%%%%%%%%%%%%%%
define reduce_struct()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{reduce_struct}
%\synopsis{remove one or more fields from a strucure}
%\usage{Struct_Type reduced_struct = reduce_struct(Struct_Type s, String_Type fieldsnames[]);}
%\qualifiers{
%\qualifier{extract:}{If given, the returned struct does only contain the fields 'fieldnames'!}
%}
%\description
%  Either removing the fields 'fieldnames' from the given structure 's' (if they even exist) or
%  if the qualifier 'extract' is given removing all other fields but those given with 'fieldnames'!
%!%-
{
  variable s, removeFields;
  switch(_NARGS)
  { case 2: (s, removeFields) = (); }
  { help(_function_name()); return; }
  
  variable fields = get_struct_field_names(s);
  variable ind;
  if( qualifier_exists("extract") ){
    ind = intersection(fields,removeFields);
  }
  else{
    ind = complement(fields,removeFields);
    ind = ind[array_sort(ind)];
  }
  fields = fields[ind];
  
  variable new_s = @Struct_Type(fields);
  variable field;
  foreach field (fields){
    set_struct_field(new_s, field, get_struct_field(s, field) );
  }
  return new_s;
}

%
% Code from src/fitting/relxill/reflection_fraction_relxill.sl
%
% -*- mode: slang; mode: fold -*-
%!%+
%\function{reflection_fraction_relxill}
%\synopsis{calculates the reflection fraction as defined in relxill, using the rel_lp_table of the relxill code}
%\usage{Double_Type fR = 
%        reflection_fraction_relxill(double a, double/array height, double rin, double rout);}
%\qualifiers{
%\qualifier{path}{[getenv("RELXILL_TABLE_PATH")]: path to the table}
%\qualifier{table}{[rel_lp_table_v0.5b.fits: name of the table}
%\qualifier{struct}{if set returns full structure returning fR, f_inf, and f_bh }
%}
%\seealso{kerr_rms,kerr_rplus}
%\description
%   Assumptions:
%   - height, rin, and rout have to be given in R_g, a is the
%     dimensionless spin parameter
%   - height, rin, and rout can also be given in negative values,
%     which is interpreted the same way as the relxill definition:
%     negative heights are in units of the event horizon (kerr_rplus) and
%     and radii in units of the ISCO (kerr_rms)
%   - photons are not allowed to cross the disk plane
%   - produces identical results to relxill
%   
%   This function 
%   
%   If relxill and the RELXILL_TABLE_PATH environment variable is set
%   up correctly, this function will work out of the box. Otherwise
%   the path needs to be set manually.
%   
%   Note: this function has been tested only for
%   rel_lp_table_v0.5b.fits. It is not recommended to change this. Any 
%   deviations are < 0.01.
%   
%   Questions: contact Thomas Dauser
%   
%   Reference: Dauser et al., 2016, A&A, 590, A76
%!%-

private define relat_aberration(th, bet){ %{{{
        return acos((cos(th)-bet) / (1-bet*cos(th)));
}
%}}}


private define interpol_frac(ilo,agrid, a){ %{{{
   return (a - agrid[ilo]) / ( agrid[ilo+1] - agrid[ilo] );
}
%}}}

private define get_reflfrac_hgrid(ind_a,ind,dat,r,rin,rout){ %{{{
   

   variable del = get_struct_field(dat,sprintf("del%i",ind+1))[ind_a,*];

   variable rind = where(rin <= r < rout);
   
   if (length(rind) <= 1){
      message(" *** error: calculating reflection fraction not possible for a disk from rin=%.3f to rout=%.3f",
	      rin,rout);
      return 0.0;
   }
      
   
   variable irlo = rind[0];
   variable irhi = rind[-1];

   variable del_bh,del_ad,fac;
   if (irlo>0){
      fac = interpol_frac(irlo-1,r,rin);      
      del_bh = (1-fac)*del[irlo-1] + fac*del[irlo];
   } else {
      del_bh =  del[irlo];
   }
   
   if (irlo<(length(r)-1)){
      fac = interpol_frac(irhi,r,rout);      
      del_ad = (1-fac)*del[irhi] + fac*del[irhi];
   } else {
      del_ad = del[irhi];
   }

   variable del_ad_max = del[-1];

   variable beta = qualifier("beta",0.0);

   if (beta > 1e-3){
      del_bh = relat_aberration(del_bh, -1. * beta);
      del_ad = relat_aberration(del_ad, -1. * beta);
      del_ad_max = relat_aberration(del_ad_max, -1. * beta);
   }

   
   variable f_bh  = 0.5*(1.0 - cos(del_bh));
   variable f_ad  = 0.5*(cos(del_bh) - cos(del_ad));

   %% photons are not allowed to cross the disk plane
   variable f_inf = 0.5*(1.0 + cos(del_ad_max));
		   
   %%% photons are not allowed to cross the disk plane                                                                                                            
   if (f_inf > 0.5){
     f_inf = 0.5;
   }

   if (qualifier_exists("verbose"))
     vmessage(" R=%.3f [f_ad=%.3f, f_inf=%.3f] for a=%.3f, h=%.3f, rin=%.3f, rout=%.3f",
	      f_ad/f_inf,f_ad,f_inf,dat.a[ind_a],dat.hgrid[ind_a,ind],rin,rout);
   
   return struct {fR = f_ad/f_inf, f_inf=f_inf, f_bh=f_bh};
}
%}}}

private define ipol_reflFracStruct(slo, shi, frac){ %{{{
        
   variable fR    = slo.fR*(1-frac) + shi.fR*frac;
   variable f_inf = slo.f_inf*(1-frac) + shi.f_inf*frac;
   variable f_bh  = slo.f_bh*(1-frac) + shi.f_bh*frac;
   
   return struct{fR=fR, f_inf=f_inf, f_bh=f_bh};
}
%}}}

private define get_reflfrac_agrid(ind_a,dat,h,rin,rout){ %{{{
   
   variable hgrid = dat.hgrid[ind_a,*];
   variable r = dat.r[ind_a,*];

   
   variable ilo = wherefirst(h<hgrid);

   if (ilo != 0){
      ilo--;
   }   
   
   variable frac_h = interpol_frac(ilo,hgrid,h);
   
   variable r_rlo = get_reflfrac_hgrid(ind_a,ilo,dat,r,rin,rout;;__qualifiers);
   variable r_rhi = get_reflfrac_hgrid(ind_a,ilo+1,dat,r,rin,rout;;__qualifiers);

   return ipol_reflFracStruct(r_rlo, r_rhi, frac_h);
}
%}}}

private define get_reflfrac_struct(a,h,rin,rout){ %{{{
   variable path = qualifier("path",getenv("RELXILL_TABLE_PATH"));
   variable fname = qualifier("table","rel_lp_table_v0.5b.fits");
   variable dat = fits_read_table(path+"/"+fname);

      
   variable ilo = wherefirst(a < dat.a);
   if (ilo == length(dat.a)-1){
      ilo--;
   }   
   variable frac_a = interpol_frac(ilo,dat.a,a);

   
   variable ii, n = length(h);
   variable refl_frac = Double_Type[n];
   variable refl_frac_struct = Struct_Type[n];
   
   _for ii(0,n-1){
      variable r_alo = get_reflfrac_agrid(ilo,dat,h[ii],rin,rout;;__qualifiers);
      variable r_ahi = get_reflfrac_agrid(ilo+1,dat,h[ii],rin,rout;;__qualifiers);      

      refl_frac_struct[ii] = ipol_reflFracStruct(r_alo, r_ahi, frac_a);

      refl_frac[ii] = refl_frac_struct[ii].fR;
      if (qualifier_exists("verbose")){
	 vmessage(" R=%.3f for a=%.3f, h=%.3f, rin=%.3f, rout=%.3f",
		 refl_frac[ii],a,h[ii],rin,rout);
      }

   }


   if (n==1){
      return refl_frac_struct[0];
   } else {   
      return merge_struct_arrays(refl_frac_struct);
   }
}
%}}}

private define convert_negative_radius_to_rg(r,a){ %{{{
   if (r<0){
      return kerr_rms(a)*abs(r);
   } else {
      return r;
   }
}
%}}}

private define convert_negative_height_to_rg(h,a){ %{{{

   variable ind_h_negative = where(h<0);
   
   if ( length(ind_h_negative) > 0 ){
      
      h[ind_h_negative] *= -1.0*kerr_rplus(a);
   }

}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define reflection_fraction_relxill(){

   variable a, height,rin,rout;
   variable rout_max = 1000.;
   
   switch(_NARGS) {
    case 2: 
      (a, height)  = ();
      rin = kerr_rms(a);
      rout = rout_max;
   }
   {
    case 3: 
      (a, height, rin)  = (); 
      rout = rout_max;   
   }
   { 
    case 4: 
      (a, height, rin, rout)  = (); 
   }
   { 
      help(_function_name()); return; 
   }
   
   %% convert height to a double-array, as it is required by the
   %% get_reflfrac_struct
   variable height_array = 1.0*[height];

   
   rin = convert_negative_radius_to_rg(rin,a);
   rout = convert_negative_radius_to_rg(rout,a);

   convert_negative_height_to_rg(height_array,a);
   
   variable refl_frac_struct = get_reflfrac_struct(a,height_array,rin,rout;;__qualifiers());

   
   
   if(qualifier_exists("struct")){
      
      if (length(refl_frac_struct)==1){
	 return refl_frac_struct[0];
      } else {   
	 return refl_frac_struct;
      }      
      
   } else {   
      if (length(refl_frac_struct)==1){
	 return refl_frac_struct[0].fR;
      } else {   
	 return refl_frac_struct.fR;
      }
   }
   
}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Code from src/slang/physics/refraction.sl
%

private define atmosphere_hohenkerk(h) {

    variable r=qualifier("rrefract");
   
    variable rr=r.Rearth+h;
    
    variable mu;
    variable T,Pw,P;
    
    if (rr<=r.Rtropo) {
	T=r.T0-r.alpha*(rr-r.R0);
	variable trat=(T/r.T0);
	Pw=r.Pw0*trat^r.delta;
	P=(r.P0+r.pfac*r.Pw0)*trat^r.gamma - r.pfac*Pw;
    } else {
	% Troposphere
	T=r.Ttropo;
	Pw=0.;
	P=r.Ptropo*exp(-r.tropscal/T * (rr-r.Rtropo));
    }

    return (T,P,Pw);
}

private define refract_mu(h) {
    variable r=qualifier("rrefract");

    variable T,P,Pw;
    (T,P,Pw)=atmosphere_hohenkerk(h;rrefract=r);

    return refractive_index_air(r.lambda;mum,pressure=P,pressure_water=Pw,temperature=T,hPa);
}

private define refract_integrand(psi)
{
    variable r=qualifier("rrefract");
    
    %
    % for a given angle psi [rad],
    % return the value of the integrand
    % (equation 3 of Auer and Standish)
    
    variable rtry=1.001*r.Rearth;
    variable sinpsi=sin(psi);
    
    %
    % newton raphson to find the correct height
    % We limit this to 8 steps (necessary since sometimes newton
    % raphson goes into oscillations and does not converge;
    % typically convergence happens after 4 steps)
    %
    variable maxnum=10;
    do {
	variable h=rtry-r.Rearth;
	variable mu=refract_mu(h;rrefract=r);

	variable dh=2.; % meter
	variable mu1,dmudr;

	% make sure that we are not crossing the
	% tropospheric boundary when calculating the
	% derivative
	if (rtry < r.Rtropo && rtry+dh > r.Rtropo) {
	    mu1=refract_mu(h-dh;rrefract=r);
	    dmudr=(mu-mu1)/dh;
	} else {
	    mu1=refract_mu(h+dh;rrefract=r);
	    dmudr=(mu1-mu)/dh;
	}
	variable fr=mu*rtry-r.mu0*r.R0*r.sinz0/sinpsi;
	variable dfr=dmudr*rtry+mu;
	variable ddr=fr/dfr;
	
	rtry-=ddr;
	maxnum--;
    } while (abs(ddr)>1e-10 && (maxnum>0));

    
    % d ln(mu(r))/dr = d mu(r)/dr * 1/mu(r)
    % dr / d ln r = r
    % therefore
    % dln(mu(r))/d(ln(r)) = d ln(mu(r))/dr * dr/d(ln r)
    %                     = d mu(r)/dr * r/mu(r)
    % and this then allows us to
    % rewrite the integrand of Auer as
    % r dmu/dr/(mu+r*dmu/dr)
    %
    
    return rtry*dmudr/(mu+rtry*dmudr);
}

define refraction() {
%!%+
%\function{refraction}
%\synopsis{calculate the correction for astronomical refraction}
%\usage{Double_Type refraction(z0;qualifiers)}
%\qualifiers{
%    \qualifier{lambda}{wavelength (in A, between 3000 and 17000 A)}
%    \qualifier{mum}{wavelength is in microns}
%    \qualifier{nm}{wavelength is in nanometers}
%    \qualifier{temperature}{temperature at observer [K; default: 288.15K=15C]}
%    \qualifier{lapse_rate}{temperature lapse rate [K/m; default: 0.0065K/m = 6.5K/km]}
%    \qualifier{centigrade}{temperature is given in C}
%    \qualifier{pressure}{ambient total pressure at observer (Pa, default: 1013.25kPa)}
%    \qualifier{kPa}{ambient pressure is in kPa}
%    \qualifier{hPa}{ambient pressure is in hPa (or mbar)}
%    \qualifier{rel_humidity}{relative humidity at observer (between 0 and 1)}
%    \qualifier{altitude}{altitude of observer above geoid (m; below 11000m)}
%    \qualifier{latitude}{geographical latitude of the observer (rad; default: 0)}
%    \qualifier{deg}{all angles are given in degrees, not radians}
%    \qualifier{exact}{use numerical integration also for z0 below 80deg}
%}
%\description
% For a given zenith distance z and local observing conditions, this function calculates
% the refraction angle, that is the difference R=z-z0 where z is the unrefracted zenith
% distance that would be measured if the Earth did not have an atmosphere, and z0 is the
% observed zenith distance.
%
% The most common use of this function will be to calculate z0 for a given topocentric
% zenith distance. For all practical purposes, R is so small and changes slowly enough,
% such that users can call the function with z and determine z0 from the return value.
% Note that the default arguments are in radians, use the deg qualifier to switch to
% degrees.
%
% The refraction depends very slightly on the atmospheric properties, i.e., its temperature
% profile. These can be set with the respective qualifiers. 
%
% The default settings of the routine are that for zenith distances smaller than 80 degrees
% the approximations given by Saastamoinen (1972, Bull. Geodesique 105, 279 and 1972, Bull.
% Geodesique 106, 383) are used. 
%
% For larger zenith distances, or if the "exact"-qualifier is set, a numerical approach
% is chosen, following the approach discussed by Auer & Standish (2000, AJ 119, 2472
% [first submitted in 1979!]), Hohenkerk and Sinclair (1985, HM Nautical Almanac Office,
% NAO Technical Note 63), and Mangum and Wallace (2015, PASP 127, 74). The function uses
% the atmospheric model discussed by Hohenkerk and Sinclair, but uses the exact refraction
% formula for air. This exact approach yields good results for zenith distances up to a
% few degrees larger than 90 degrees (i.e., observation of the horizon from a mountain
% top) and is the one on which the refraction formulae of the Astronomical Almanac are based.
%
% For large zenith angles, there are slight differences at the arcsecond or less level
% between the results discussed here and the numbers listed by Hohenkerk or Auer. These
% are due to the different treatment of numerical instabilities and the use of numerical
% differentiation formulae. Given that for large zenith angles the simplified atmospheric
% model of Hohenkerk (constant temperature lapse rate in the troposphere, constant temperature
% in the stratosphere) results in larger systematic errors anyway (see Nauenberg, 2017,
% PASP 129, 44503), this should not be seen as an error of the function.
%
% As a caveat, the calculations using the Saastamoinen formulae assume that all pressure
% terms there are in Pa, rather than in mbar. This reproduces the results with respect
% to the numerical simulations and values tabulated elsewhere to <1". But it is not
% what Saastamoinen claims. I (J. Wilms) am puzzled...
%
% If the exact qualifier is set or for large z, all input into the function MUST be
% scalars - in this case this routine is NOT array safe.
%
% For the Saastamoinen formulae, the routine is array safe in all relevant parameters.
%
%!%-;
    variable z0;
    switch (_NARGS)
    { case 1: z0=(); }
    { help(_function_name()); return; }

    if (qualifier_exists("deg")) {
	z0*=PI/180.;
    }

    % temperature at observer [K]
    variable T0;
    if (qualifier_exists("centigrade")) {
	T0=qualifier("temperature",0.)+273.15;
    } else {
	T0=qualifier("temperature",273.15);
    }
    % pressure at observer [Pa]
    variable P0;
    if (qualifier_exists("kPa")) {
	P0=1000.*qualifier("pressure",101.325); % kPa -> Pa
    } else {
	if (qualifier_exists("hPa")) { % =mbar
	    P0=100.*qualifier("pressure",1013.25); % hPa -> Pa
	} else {
	    P0=qualifier("pressure",101325.); % default: Pa
	}
    }
    variable h0=qualifier("altitude",0.);    % height of observer above geoid [m]

    variable lat=qualifier("latitude",0.);
    if (qualifier_exists("deg")) {
	lat*=PI/180.;
    }
    
    variable Rh=qualifier("rel_humidity",0.);
    if (Rh<0. or  Rh>1.) {
	throw RunTimeError, sprintf("%s: relative humidity is out of bounds\n",_function_name());
    }
    % partial pressure of water at observer (Pa)
    % see Mangum and Wallace (2015; appendix C)
    variable Pw0=Rh*(T0/247.1)^18.36;

    % wavelength for the observation 
    variable lambda=0.574; % mum
    if (qualifier_exists("lambda")) {
	if (qualifier_exists("mum")) {
	    lambda=qualifier("lambda");
	} else {
	    lambda=qualifier("lambda")/10000.; % A->Mum
	}
    }

    % atmospheric structure
    variable htropo=qualifier("Htropo",11000.);   % altitude of the tropopause above geoid [h]
    variable hstrato=qualifier("Hstrato",80000.);  % altitude of the stratosphere above geoid [h]

    if (h0>=htropo) {
	()=printf("%s: WARNING: altitude is above topopause. Returning 0!\n",_function_name());
	return 0.;
    }

    % WGS84 radius (for speed reasons we do not use geographic2vector)
    variable Rearth=6378137.*(1.-sin(lat)^2./298.257223563); 
    variable R0=Rearth+h0;
    
    variable Rstrato=hstrato+Rearth; % radius of the stratosphere
    variable Rtropo=Rearth+htropo;
    variable alpha=qualifier("lapse_rate",0.0065); % Temperature lapse rate [K/m]
    variable Ttropo=T0-alpha*(Rtropo-R0); % temperature at the tropopause
    
    % Saastamoinen 1972 standard formula
    % use if
    % a) qualifier "exact" is not set and
    % b) z0<80 degrees (1.4 rad)
    if (not qualifier_exists("exact") && z0<1.4) {
	variable tanz=tan(z0);
	variable tanz2=tanz*tanz;

	% note: Saastamoinen claims that the pressures should be
	% in mbar/hPa, but results agreeing with numerical calculations
	% agree only when using Pa here
	% I am confused...
	variable pterm=(P0-0.156*Pw0)/T0;
	
	% latitude term (Saastamoinen 72, equation above eq 37)
	% corrected by 1e3 to avoid a division below
	variable lterm=0.07485e-3*(1.+0.0060*cos(2.*lat)+0.00012e-3*h0);
	
	% basic equation (Saastamoinen, eq. 30a)
	variable dz=16.271*tanz*(1.+0.0000394*tanz2*pterm)*pterm-lterm*(tanz2+1.)*tanz*P0;

	% correction for z in65-80 deg
	% (this should be Saastamoinen, Table III, but I can't fully
	% reproduce his values)
	if (z0>1.13) {
	    variable T1=T0-alpha*(htropo-h0);
	    variable p1=P0*(T1/T0)^5.26;

	    variable tanz3=tanz2*tanz;
	    variable tanz5=tanz3*tanz2;
	    variable tanz7=tanz3*tanz2;
	    variable tanz9=tanz7*tanz2;

	    variable del=0.000288e-6*(3.*tanz5+5.*tanz3)*(p1*T1+0.190*P0*T0)
	                -0.013e-6*tanz5*(p1*p1/T1)
	                -0.014e-12*tanz7*(p1*T1*T1+0.64*P0*T0*T0)
	                +0.0003e-15*tanz9*(p1*T1*T1*T1+2.*P0+T0*T0*T0);
	    
	    dz+=del;
	}
	
	% wavelength correction (Saastamoinen, eq 38 [p388])
	dz*=170.2649/(173.3-1./lambda/lambda);

	
	if (qualifier_exists("deg")) {
	    return dz/3600.;
	}
	
	return dz/3600 * PI/180.;
    }


    % local acceleration
    variable g=g_earth(lat;mks,altitude=h0);


    % exponent of temperature dependency on water pressure,
    % between 18 and 21 (only by accident the same as the exponent
    % in the equation for Pw)
    variable delta=18.36;
    
    % Universal gas constant [J/Mole K]
    variable Runiv=8314.36; 
    
    % Molecular weight of dry air and of water vapor [g/mole]
    variable Md=28.9644; 
    variable Mw=18.0152;

    % atmospheric structure
    
    variable tropscal=g*Md/Runiv;
    variable gamma=tropscal/alpha;
    % pressure at the tropopause
    variable pfac=(1.-Mw/Md)*gamma/(delta-gamma);
    variable Ptropo=(P0+pfac*Pw0)*(Ttropo/T0)^gamma; 

    
    % structure containing all of the parameters
    % needed for the evaluation of the atmospheric
    % structure and other helpers for the
    % refraction
    variable r;
    r=struct
    {
	Rearth=Rearth, % Radius of the Earth, m
	Rtropo=Rtropo, % radius of the troposphere
	lambda=lambda,
	P0=P0,         % pressure at observer, mb
	T0=T0,         % Temperature at observer, K
	Pw0=Pw0,       % partial pressure of water vapor at observer, mb
	alpha=alpha,   % temperature lapse rate
	delta=delta,   % humidity exponent
	gamma=gamma,   %
	pfac=pfac,     % helper
	tropscal=tropscal, % helper for the pressure scaling in the troposphere
	Ttropo=Ttropo, % Temperature at tropopause, K
	Ptropo=Ptropo, % Pressure at tropopause, mb
	R0=R0,         % radius of observer from center of Earth, m
	z0=z0,         % zenith angle, rad
	sinz0=sin(z0), % 
	mu0=0.
    };
    r.mu0=refract_mu(h0;rrefract=r); % refract_mu does not need mu0!
    %
    % angle at the tropopause in the troposphere
    %
    variable mutropo=refract_mu(htropo;rrefract=r);
    variable sintropo=r.mu0*R0*r.sinz0/(mutropo*Rtropo);
    variable psitropo=asin(sintropo);
	
    %
    % angle at the end of the stratosphere
    %
    variable mustrato=refract_mu(hstrato;rrefract=r);
    variable sinstrato=r.mu0*r.R0*r.sinz0/(mustrato*Rstrato);
    variable psistrato=asin(sinstrato);

    variable refstrato=-qromb(&refract_integrand,psitropo,z0;rrefract=r);
    variable reftropo=-qromb(&refract_integrand,psistrato,psitropo;rrefract=r);

    variable reftot=refstrato+reftropo;

    if (qualifier_exists("deg")) {
	return reftot*180./PI;
    }
	
    return reftot;
}

%
% Code from src/slang/physics/refractive_index_air.sl
%
define refractive_index_air() {
%!%+
%\function{refractive_index_air}
%\synopsis{calculate the refractive index of dry and moist air}
%\usage{Double_Type refractive_index_air(lambda;qualifiers)}
%\qualifiers{
%    \qualifier{mum}{wavelength is in microns}
%    \qualifier{nm}{wavelength is in nanometers}
%    \qualifier{temperature}{temperature [K; default: 288.15K]}
%    \qualifier{centigrade}{temperature is given in C}
%    \qualifier{pressure}{ambient total pressure (Pa, default: 1013.25kPa)}
%    \qualifier{kPa}{all pressure arguments are in kPa}
%    \qualifier{hPa}{all pressure arguments are in hPa (or mbar)}
%    \qualifier{CO2_ppm}{CO2 fraction in ppm (default: 450)}
%    \qualifier{water_pressure}{partial water vapor pressure}
%    \qualifier{rel_humidity}{relative humidity}
%    \qualifier{silent}{do not emit warning messages}
%}
%\description
% This function calculates the phase refractive index of dry and moist air
% for light in the optical and IR following the standard paper by
% Ciddor (1996, Appl. Optics 35(9), 1566) and the discussion by Stone
% and Zimmerman (NIST Engineering Metrology Toolbox;
% http://emtoolbox.nist.gov/Wavelength/Documentation.asp) for a given
% wavelength (default A, but see the mum and nm qualifiers) and ambient
% conditions (temperature, pressure,  humidity or partial vapor pressure,
% and CO2 concentration). The function has been verified against the values 
% given by Ciddor and in the NIST Engineering Metrology Toolbox.
%
% The relative uncertainty of the approximations used here is claimed to be 
% around 2e-8. The range of validity is 3000 A<lambda<17000 A, temperatures
% between -40C and 100C (233-373K), and pressures between 10 and 140kPa.
% The CO2 fraction is allowed to vary between 0 and 2000ppm.
%
% This routine is array safe. If lambda is an array, the other qualifiers
% can be either arrays of the same length as lambda or single valued.
%
%!%-;
    variable lambda=(); % default: A

    if (qualifier_exists("mum") and qualifier_exists("nm")) {
	throw RunTimeError,sprintf("%s: only one of nm and mum qualifiers is allowed!",_function_name());
    }

    if (qualifier_exists("kPa") and qualifier_exists("hPa")) {
	throw RunTimeError,sprintf("%s: only one of kPa and hPa qualifiers is allowed!",_function_name());
    }

    if (qualifier_exists("mum")) {
	lambda*=10000.; % mum -> A
    }

    if (qualifier_exists("nm")) {
	lambda*=10.; % nm -> A
    }

    variable chatty=(qualifier_exists("silent")==1);
    
    if (chatty && length(lambda)!=length(where(3000<=lambda<=17000))) {
	()=printf("%s: WARNING: Wavelength outside of validity interval 3000-17000A\n",_function_name());
    }


    % convert lambda to mum
    lambda=lambda/10000.;
    
    variable T; % temperature in K (default: 15c [standard atmosphere]
    if (qualifier_exists("centigrade")) {
	T=qualifier("temperature",15.)+273.15;
    } else {
	T=qualifier("temperature",288.15);
    }

    if (chatty && length(T) !=length(where(233.15<=T<=373.15))) {
	()=printf("%s: WARNING: Temperature outside of validity interval 233-373K (-40-100C)\n",_function_name());
    }
 
    % pressure in P
    variable P;
    if (qualifier_exists("kPa")) {
	P=1000.*qualifier("pressure",101.325); % kPa -> Pa
    } else {
	if (qualifier_exists("hPa")) { % =mbar
	    P=100.*qualifier("pressure",1013.25); % hPa -> Pa
	} else {
	    P=qualifier("pressure",101325.); % default: Pa
	}
    }


    if (chatty && length(P)!=length(where(10e3<=P<=140e3))) {
	()=printf("%s: WARNING: Pressure outside of validity interval 10-140kPa\n",_function_name());
    }

    % CO2 content (ppm)
    variable xc=qualifier("CO2_ppm",450.);
    if (chatty && length(xc) != length(where(0.<=xc<=2000.))) {
	()=printf("%s: WARNING: CO2 fraction outside of validity interval 0-2000ppm\n",_function_name());
    }
    
    if (qualifier_exists("water_pressure") and qualifier_exists("rel_humidity")) {
	throw RunTimeError,sprintf("%s: only one of water_pressure and rel_humidity qualifiers is allowed!",_function_name());
    }

    
    % mole fraction of water vapor
    variable xw=0.;
    if (qualifier_exists("water_pressure") or qualifier_exists("rel_humidity")) {
	% enhancement factor of water vapor in air
	variable alpha=1.00062;
	variable beta=3.14e-8; % Pa^-1
	variable gamma=5.60e-7; % C^-2

	variable f=alpha+beta*P+gamma*(T-273.15)^2.;
	
	if (qualifier_exists("water_pressure")) {
	    %
	    % calculation for a given water partial pressure
	    %
	    variable Pv=qualifier("water_pressure");
	    if (qualifier_exists("kPa")) {
		Pv*=1000.; % kPa -> Pa
	    }
	    if (qualifier_exists("hPa")) {
		Pv*=100.; % hPa -> Pa
	    }
	    xw=f*Pv/P; % mole fraction
	    
	} else {
	    %
	    % relative humidity known
	    %
	    variable Rh=qualifier("rel_humidity",0.);
	    if (length(Rh)!=length(where(0<=Rh<=1.))) {
		throw RunTimeError, sprintf("%s: relative humidity is out of bounds\n",_function_name());
	    }

	    % saturation vapor pressure of water vapor in air
	    variable svp=SaturationVaporPressure(T;iapws);

	    % mole fraction 
	    xw=Rh*f*svp/P;
	}
    }

    %
    % from here on: T in K, lambda in mum, P in Pa
    %
    
    % wavenumber squared
    variable s2=1./(lambda*lambda);
    
    %
    % constants for standard phase and group refractivities
    % (Ciddor, Appendix A)
    %
    
    % dry air (mum^-2)
    variable k=[238.0185, 5792105., 57.362, 167917.];

    % water vapor (mum^-2)
    variable w=[295.235, 2.6422, -0.032380, 0.004028];

    % compressibility
    variable a=[1.58123e-6, -2.9331e-8, 1.1043e-10];
    variable b=[5.707e-6, -2.051e-8];
    variable c=[1.9898e-4, -2.376e-6];
    variable d=1.83e-11;
    variable e=-0.765e-8;


    % Gas constant
    variable R=8.314472; % J/mol/K

    variable Mv=0.018015; % 

    %
    % refractive index for standard air
    % (15C, 1013.25 hPa, 0% humidity)
    %
    variable Pr1=101325.; % Pa
    variable Tr1=288.15; % K
    
    % n_as-1 : dry standard air, 450ppm CO2
    variable nas=1e-8*(k[1]/(k[0]-s2) + k[3]/(k[2]-s2));

    variable nvs1=1.022e-8*(((w[3]*s2+w[2])*s2+w[1])*s2+w[0]);
    
    % molar mass of dry air
    variable Ma=0.0289635+1.2011e-8*(xc-400.);

    % n_axs-1: dry standard air, != 450ppm CO2
    variable naxs1=nas*(1.+5.34e-7*(xc-450.));
    
    % Celsius temperature
    variable TT=T-273.15;

    % Compressibility of dry air
    % (this is the equation for Zm below evaluated with
    % T=288.15, P=101325, xw=0)
    variable Za=0.9995922115;
    
    % Ciddor, eq 12 - compressibility of moist air
    variable Zm=1.-(P/T)*(a[0]+a[1]*TT+a[2]*TT^2.+(b[0]+b[1]*TT)*xw+
               (c[0]+c[1]*TT)*xw^2)+(P/T)^2.*(d+e*xw^2);

    variable rho_vs=0.00985938;
    % density of dry air at 15C, 131325 Pa w/xw=0
    variable rho_axs=Pr1*Ma/(Za*R*Tr1);
    % density of the vapor component
    variable rho_v=xw*P*Mv/(Zm*R*T);
    % density of the dry componnt
    variable rho_a=(1.-xw)*P*Ma/(Zm*R*T);

    % index of refraction
    variable n=1.+(rho_a/rho_axs)*naxs1+(rho_v/rho_vs)*nvs1;

    return n;
}

%
% Code from src/slang/strings/removeLeadingZeros.sl
%
define removeLeadingZeros(s)
{
  variable res = string_matches(s, `0*\([^0].*\)`, 1);
  return res==NULL ? s : res[1];
}

%
% Code from src/slang/structs/rename_struct_fields.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
define rename_struct_fields()
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{rename_struct_fields}
%\usage{Struct_Type new_s = rename_struct_fields(Struct_Type s, String_Type fieldnames[]);}
%!%-
{
  variable s, fieldnames;
  switch(_NARGS)
  { case 2: (s, fieldnames) = (); }
  { help(_function_name()); return; }

  if(length(get_struct_field_names(s)) != length(fieldnames))
  { message("error ("+_function_name()+"): The number of field names is not the same.");
    return;
  }

  variable new_s = @Struct_Type(fieldnames);
  set_struct_fields(new_s, get_struct_fields(s, __push_array(get_struct_field_names(s))));
  return new_s;
}

%
% Code from src/fitting/parameters/renormalizeTo.sl
%
define renormalizeTo()
{
  variable id, norms;
  switch(_NARGS)
  { case 2: (id, norms) = (); }
  { % else:
      message("usage: renormalizeTo(id, norm);");
      return;
  }
  variable factor = get_par("constant("+string(id)+").factor");
  variable norm;
  foreach norm ([norms])
  { set_par(norm, get_par(norm)*factor); }
  variable data;
  foreach data (all_data)
  { variable par = "constant("+string(data)+").factor";
    if(get_par(par)==1) { thaw(par); }
    set_par(par, get_par(par)/factor);
    if(data==id) { freeze(par); }
  }
}

%
% Code from src/slang/structs/tables/struct-arrays/replicate_table.sl
%
define replicate_table()
%!%+
%\function{replicate_table}
%\synopsis{repeats columns of a table, possibly with a periodic shift}
%\usage{Struct_Type replicate_table(Struct_Type t);}
%\qualifiers{
%\qualifier{P}{[= 1] period}
%\qualifier{back}{[= 0] repeat periods backwards}
%\qualifier{ahead}{[= 1] repeat periods forwards}
%\qualifier{periodic}{[\code{= ["bin_lo", "bin_hi"]}] periodic structure fields}
%}
%\description
%    The return value is a structure with the same fields as \code{t}.
%    All array fields are repeated (back+1+ahead) times; periodic ones as\n
%      \code{[ val-back*P, ..., val-P, val, val+P, ..., val+P*ahead ]} ,\n
%    while other ones are just replicated:\n
%      \code{[ val       , ..., val  , val, val  , ..., val         ]} .\n
%!%-
{
  variable t;
  switch(_NARGS)
  { case 1: t = (); }
  { help(_function_name()); return; }
  
  variable P = qualifier("P", 1);
  variable back = qualifier("back", 0);
  variable ahead = qualifier("ahead", 1);
  variable periodic = qualifier("periodic", ["bin_lo", "bin_hi"]);
  
  variable T = @t;
  variable i, n=back+1+ahead;
  variable field;
  foreach field ( get_struct_field_names(t) )
  {
    variable val = get_struct_field(t, field);
    if(typeof(val)!=Array_Type)
      continue;
    variable isperiodic = any(field == periodic);
    variable len = length(val);
    variable a = @ _typeof(val)[n*len];
    _for i (-back, ahead, 1)
      a[ [(i+back)*len : (i+back+1)*len-1 ] ] = (isperiodic ? val + i*P : val);
    set_struct_field(T, field, a);
  }
  return T;
}

%
% Code from src/misc/atomic/require_atoms.sl
%
define require_atoms()
%!%+
%\function{require_atoms}
%\usage{require_atoms();}
%\description
%    This function loads the atomic data from the
%    Astrophysical Plasma Emission Database via\n
%       \code{atoms(aped);}\n
%    unless \code{_isis->Dbase} is already initialized.
%\seealso{atoms}
%!%-
{
  if(_isis->Dbase.dir == NULL)  atoms(aped);
}

%
% Code from src/slang/math/rescale_range.sl
%
define rescale_range()
%!%+
%\function{rescale_range}
%\synopsis{rescales a value}
%\usage{Double_Type y = rescale_range(Double_Type x);}
%\qualifiers{
%\qualifier{in}{inputrange}
%\qualifier{out}{outputrange}
%}
%\description
%    possibilities for in/out and the corresponding scaling function:\n
%    \code{"0:1"}, no scaling\n
%    \code{"-inf:inf"}, arctan
%!%-
{
  variable x;
  switch(_NARGS)
  { case 1: x = (); }
  { help(_function_name()); return; }

  variable in  = qualifier("in", "-inf:inf");
  variable out = qualifier("out", "0:1");

  variable x01;
  switch(in)
  { case "0:1": x01 = x; }
  { case "-inf:inf": x01 = 0.5 + atan(x)/PI; }
  { vmessage(`error (%s): input range "%s" unknown`, _function_name(), in); return; }

  switch(out)
  { case "0:1": return x01; }
  { case "-inf:inf": return tan(PI*(x01-0.5)); }
  { vmessage(`error (%s): output range "%s" unknown`, _function_name(), out); return; }
}

%
% Code from src/fitting/statistics/residual_runs.sl
%
#ifexists normal_cdf

private define __runs_info (seq)
{
  variable confidence = [.10,.05,.01];

  variable n2;
  variable n1 = where(seq>0, &n2);

  variable s = Int_Type[length(seq)];
  s[n1] = 1;
  s[n2] = 0;
  variable w = where(s[[1:]]-s[[:-2]]);
  variable r = length(w)+1;
  variable l = (1 == r) ? length(seq) : w-[-1, w[[:-2]]]; 
  variable p;
  n1 = length(n1);
  n2 = length(n2);

  variable info = struct {
    runs = -1,
    over = n1,
    under = n2,
    min_run = -1,
    max_run = -1,
    test = [0,0,0],
    over_test = [0,0,0],
    under_test = [0,0,0],
    confidence = confidence,
  };

  if (n1 < 2 or n2 < 2 or r < 2 or r > n1 + n2)
    return info;

  p = __runs_p(n1, n2, r);

  if (NULL == p)
    return info;

  info.runs = r;
  info.min_run = min(l);
  info.max_run = max(l);
  info.test = confidence/2.<p and confidence/2.<(1-p);
  info.under_test = confidence < p;
  info.over_test = confidence < (1-p);

  return info;
}

%%%%%%%%%%%%%%%%%%%%%%%
define residual_runs ()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{residual_runs}
%\synopsis{Perform runs test on model and data}
%\usage{status = residual_runs([&stat]);}
%\description
%    While the Chi2 test tests for the absolute difference
%    of model and data, the runs test (Wald-Wolfowitz test)
%    accounts for the sign difference.
%
%    The test is in particular usefull for checking if the
%    model misses essential features of the data. As a simple
%    example concider data coming from a linear relation
%    described by a constant function. Although the chi2 test
%    might give a reasonable result the improper model is
%    easily spotted in the residuals.
%
%    The results are presented for 3 different confidence
%    regions. Each model passing the test is added to the
%    entries of the result matrix.
%
%    Note: The Confidence region is the total confidence. It
%    can happen that the test fails for the two tailed test but
%    succeeds for both one tailed tests!
%
%    If a reference is given as argument the same information
%    is stored as struct in that reference.
%\seealso{runs_test, normal_cdf}
%!%-
{
  variable res;
  switch (_NARGS)
  { case 0: res = NULL; }
  { case 1: res = (); }
  { help("residual_runs"); return; }

  variable ret, save_verbose = Fit_Verbose;
  Fit_Verbose = -1;
  ret = eval_counts();
  Fit_Verbose = save_verbose;

  variable active = all_data(1);
  variable id,model,data,noti;
  variable w, s, dinfo;
  variable coll = {};

  variable info = struct {
    num_bins = 0,
    runs = 0,
    over = 0,
    under = 0,
    min_run = INT_MAX,
    max_run = 0,
    test = [0,0,0],
    over_test = [0,0,0],
    under_test = [0,0,0],
    confidence = [.10,.05,.01],
    skipped = Int_Type[0],
  };

  foreach id (active) {
    noti = get_data_info(id).notice_list;
    model = get_model(id).value;
    data = get_data(id).value;

    s = data/model-1;
    w = where(s != 0); % ignore exact matches

    dinfo = __runs_info(s);
    dinfo = struct { id = id, @dinfo, num_bins = length(noti) };
    list_append(coll, dinfo);
  }

  foreach s (coll) {
    info.num_bins += s.num_bins;
    info.runs += s.runs;
    info.over += s.over;
    info.under += s.under;
    if (-1 != s.min_run) {
      info.min_run = min([info.min_run, s.min_run]);
      info.max_run = max([info.max_run, s.max_run]);
    } else
      info.skipped = [info.skipped, s.id];
    info.test += s.test;
    info.over_test += s.over_test;
    info.under_test += s.under_test;
  }


  if (Fit_Verbose >= 0 && ret == 0) {
    variable msg =
`       Number of runs = %s
 Num. of above points = %d
 Num. of below points = %d
      Min. run length = %s
      Max. run length = %s
            Data bins = %d
  Successful tests: Two tail, Lower tail, Upper tail 
       Confidence
     (%05.2f)%%:    % 10d, % 10d, %10d
     (%05.2f)%%:    % 10d, % 10d, %10d
     (%05.2f)%%:    % 10d, % 10d, %10d
     Skipped data sets = %d`;
    vmessage(msg, info.runs==-1 ? sprintf("NaN") : sprintf("%d", info.runs),
	     info.over,
	     info.under,
	     info.min_run == INT_MAX ? sprintf("NaN") : sprintf("%d", info.min_run),
	     info.max_run == 0 ? sprintf("NaN") : sprintf("%d", info.max_run),
	     info.num_bins,
	     info.confidence[0]*100, info.test[0], info.under_test[0], info.over_test[0],
	     info.confidence[1]*100, info.test[1], info.under_test[1], info.over_test[1],
	     info.confidence[2]*100, info.test[2], info.under_test[2], info.over_test[2],
	     length(info.skipped));
  }

  if (NULL != res && Ref_Type != typeof(res))
    throw UsageError, "Requires a reference";

  if (Ref_Type == typeof(res))
    @res = info;

  return ret;
}

#endif
%
% Code from src/data/resolution_rebin.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
% Author: Marco Fink
% Mail: marco.fink@fau.de


%%%%%%%%%%%%%%%%%%%%%%%%
define energy_resolution(E_keV, instrument){
%%%%%%%%%%%%%%%%%%%%%%%%
% Helper function that returns Spectral resolution in keV at a given
% point in keV. Second argument is a string specifying the instrument.

    variable dE_eV = 0;

    switch(instrument)
    % approximate Energy resolution for NuStar FPM,
    % linear interpolation of values given on:
    % https://heasarc.gsfc.nasa.gov/docs/nustar/nustar_tech_desc.html
      {case "fpm": dE_eV = 8.6*E_keV + 310; }

    % approximate Energy resolution for XMM-Newton, assuming:
    % linear interpolation of values given on:
    % https://xmm-tools.cosmos.esa.int/external/xmm_user_support/documentation/uhb/epic_specres.html
      {case "epn": dE_eV = 8.9*E_keV + 91; }

    % approximate Energy resolution for Suzaku XIS, 
    % linear interpolation of values given on:
    % http://space.mit.edu/XIS/about/  
      {case "xis": dE_eV = 13.3*E_keV + 47; }
      { return -1; }

    % catch negative cases and convert to keV
    if(dE_eV <= 0){
      dE_eV = 0.1;
    }

    return 0.001*dE_eV;
}


%%%%%%%%%%%%%%%%%%%%%%%%
define resolution_rebin(id, instrument) {
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{resolution_rebin}
%\synopsis{Rebin spectra based on approximated detector resolution.}
%\usage{Array_Type resolution_rebin(dataset_id, instrument; oversampling, min_counts);}
%\qualifiers{
%\qualifier{sampling}{Float_Type, oversampling factor of detector resolution/bin width, default: 3}
%\qualifier{min_sn}{Float_Type, minimum signal-to-noise ratio, default: 5}
%}
%\description
%    This function rebins a dataset to fulfill the following two criteria:\n
%    1: Each bin contains enough counts so that - assuming Poissonian noise -
%       Signal-to-Noise is larger than the given minimum.\n
%    2: If criterion 1 is fulfilled, the bin size is chosen as a fraction of the 
%       detector resolution, an oversampling of 3 is recommended (Kaastra 2016).\n
%
%    Required arguments are the dataset ID and a string that specifies the used 
%    instrument. Supported are:\n
%      XMM EPICpn: "epn"\n
%      Suzaku XIS: "xis"\n
%      NuSTAR FPM: "fpm"\n
%
%\examples
%     Load and plot EPICpn data with a histogram oversampling detector resolution by 5:\n
%     variable EPN_data = load_data("src_s.pha");\n
%     resolution_rebin(EPN_data,"epn"; sampling = 5);\n
%     plot_data(EPN_data; dsym=0);\n
%
%\seealso{rebin_data, group}
%!%-

  % check for detector support
  if(energy_resolution(0, instrument) == -1){
    vmessage("Detector not supported");
    return [0];
  }

  variable sampling = qualifier("sampling", 3);
  variable signal_noise = qualifier("min_sn", 5);
  
  rebin_data(id,0);

  % assuming Poissonian distribution for signal-to-noise
  variable min_counts = signal_noise^2;
  variable data = get_data(id);

  variable E_min = _A(data).bin_lo[0];
  variable E_max = _A(data).bin_hi[-1];
  
  % calculate the rebin grid from the energy resolution
  variable res_bin_lo = [E_min]; 
  variable next_bin = 0;
    
  while(res_bin_lo[-1] < E_max){
    next_bin = res_bin_lo[-1] + (1.0/sampling)*energy_resolution(res_bin_lo[-1], instrument);
    res_bin_lo = array_insert(res_bin_lo, next_bin, length(res_bin_lo));
  };
  
  % go to Angstrom space
  res_bin_lo = _A(res_bin_lo);
  
  variable index_array = Integer_Type[length(data.bin_lo)];
  variable i;
  variable flag = 1;
  variable j = 0;
  variable counts = 0;

  % fill the index array for later use by rebin_data
  for(i=0; i < length(index_array)-1; i++){
    if(data.bin_lo[i] > res_bin_lo[j]){
      if(counts >= min_counts){
	counts = 0;
	flag *= -1;
      }
      j += 1;
    }else{
      counts += data.value[i];
    }
    
    index_array[i] = flag;
  };
  
  rebin_data(id, index_array);
  return index_array;
}

%
% Code from src/slang/color/rgb2hex.sl
%
define rgb2hex (r, g, b)
%!%+
%\function{rgb2hex}
%\synopsis{converts r, g, b values to 24 bit color values.}
%\usage{Integer_Type rgb2hex(Double/Int_Type r, g, b)}
%\qualifiers{
%  \qualifier{string}{returns value as hex encoded string.}
%}
%\description
%  Convert given color values to one 24 bit value.
%  If values are doubles they are interpreted as ranging
%  from 0 to 1. If integers they are interpreted as
%  8 bit (0-255) color values.
%\seealso{rgb2hex}
%!%-
{
  if (_typeof(r) == Double_Type) r *= 255;
  if (_typeof(g) == Double_Type) g *= 255;
  if (_typeof(b) == Double_Type) b *= 255;

  variable col = (int(_max(0, _min(r, 255)))<<16)
    + (int(_max(0, _min(g, 255)))<<8)
      + (int(_max(0, _min(b, 255))));

  if (qualifier_exists("str") or qualifier_exists("string"))
    col = array_map (String_Type, &sprintf, "%06x", col) ;
  else if (qualifier_exists("Str") or qualifier_exists("String")
	   or qualifier_exists("STR") or qualifier_exists("STRING"))
    col = array_map (String_Type, &sprintf, "%06X", col) ;
  

  return length(col)==1 ? col[0] : col;
}

private define map_hexstr (str)
{
  variable r, g, b;
  () = sscanf(str, "%02x%02x%02x", &r, &g, &b);
  return r, g, b;
}

define hex2rgb (col)
%!%+
%\function{hex2rgb}
%\synopsis{converts a 24 bit color value to r, g, b, values.}
%\usage{Integer_Type r, g, b hex2rgb(Int/String_Type col)}
%\qualifiers{
%  \qualifier{float}{If given, the channels are instead returned in the range (0-1).}
%}
%\description
%  Converts a hex color to RGB values. Can be either 24 bit
%  value or string.
%\seealso{rgb2hex}
%!%-
{
  variable r, g, b;
  if (_typeof(col) == String_Type)
    (r, g, b) = array_map(UInt_Type, UInt_Type, UInt_Type, &map_hexstr, strlow(col));
  else
    (r, g, b) = (col & (0xff0000))>>16, (col & 0xff00)>>8, col & 0xff;

  if (qualifier_exists("float"))
    (r, g, b) = (r/255., g/255., b/255.);

  return length(r) == 1 ? (r[0], g[0], b[0]) : (r, g, b);
}

%
% Code from src/misc/radio/ridge_line.sl
%
define ridge_line ()
%!%+
%\function{ridge_line}
%\synopsis{calculates the ridge line}
%\usage{Struct_Type ridge_line = ridge_line (Array_Type \code{img})}
%\qualifiers{
%\qualifier{ref_pix}{[=[y,x]] starting pixel of the ridge line (index convention: img[y,x])
%                 brightest pixel used by default}
%\qualifier{dr}{[=0.8] step size in pixels}
%\qualifier{steps}{number of steps (default is maximal length of dimensions)}
%}
%\description
%    This function calculates the ridge line in a (jet) image.
%    Starting from the brightest point (assumed to be the core),
%    or a point specified by the \code{ref_pix} qualifier, the ridge
%    line is calculated. The ridge line is given in different ways:
%    The brightest points at each distance (in steps of \code{dr} pixels)
%    from the reference pixel, are given by the fields \code{peak_x} and
%    \code{peak_y} of the returned structure. The field \code{peak_flux}
%    contains the corresponding pixel value.
%    As an alternative to the peak ridge line, a flux-centered ridge
%    line is provided. The fields \code{flux_cent_x} and \code{flux_cent_y}
%    specify the points, at which the integrated flux (along the same
%    distance from the refence pixel) to the left is equal to that on
%    the right side.
%\notes
%    - elliptical beams will create artifacts (wiggles in the ridge line),
%      in order to avoid this effect, obtain the ridge line from an
%      image convolved with a circular beam
%    - valid points can be obtained, e.g., by filtering the radii at which
%      the peak flux exceeds the map's noise level (fit_gauss_to_img_noise)
%    - currently no smoothing of the points is done (e.g., fit spline to
%      ridge line)
%    - there can be problems if the radius (used for ridge line calculation)
%      lies completely within the jet/beam, \code{ref_pix} qualifier can be
%      used to select another starting point (on the "jet axis")
%\example
%    variable mdl = struct {
%                   flux   = [0.3, 0.5, 0.1,  0.2, 0.3,  0.9,    3],
%                   ra     = [3.7, 2.8, 2.2,  1.4, 0.9,  0.5,    0],
%                   dec    = [1.2, 0.8, 0.8,  0.7, 0.6,  0.4,    0],
%                   smajor = [0.4, 0.2, 0.1, 0.03, 0.1, 0.01, 0.01],
%                   sminor = [0.4, 0.2, 0.1, 0.03, 0.1, 0.01, 0.01],
%                   pang   = [0,      0,   0,    0,   0,    0,    0] };
%    variable beam = [0.3, 0.3, 0.7];
%    variable img = radio_mod2img (mdl, beam);
%    variable r = ridge_line (img.img;);
%    struct_filter(r, where(r.peak_value>1e-4));
%    plot_image (log(img.img+1e-4));
%    color(4);  oplot(r.peak_x,r.peak_y);
%    color(13); oplot(r.flux_cent_x,r.flux_cent_y);
%\seealso{fit_gauss_to_img_noise, radio_mod2img, plot_vlbi_map}
%!%-
{
  switch(_NARGS)
  { case 1: variable img = (); }
  { help(_function_name()); return; }
  
  % reduce to 2D image (remove difmap FREQ and STOKES)
  variable shape = array_shape(img);
  shape = shape[ where( shape > 1 ) ];
  if ( length(shape) == 2 )
    reshape (img, shape);
  
  variable len_x = shape[1];
  variable len_y = shape[0];
  
  variable i = where_max(img)[0];
  variable ref_pix = qualifier( "ref_pix", [( i / len_x),(i mod len_x)]);
  variable ref_pix_x = ref_pix[1];
  variable ref_pix_y = ref_pix[0];
  
  variable pang = 0.0;
  
  variable dr = qualifier("dr", 0.8);
  
  variable len = qualifier("steps",nint(_max(len_x,len_y)/dr));
  variable d = [1:len];
  variable steps, phi, x, y, r, sel, peak_index, len_sel;
  
  variable ridge = struct {
    i = Integer_Type[len],
    peak_x      = Integer_Type[len],
    peak_y      = Integer_Type[len],
    peak_value  = Double_Type[len],
    flux_cent_x = Double_Type[len],
    flux_cent_y = Double_Type[len],
  };

  variable lx,ly,rx,ry,lsum,rsum,sum_diff;
  _for i (0,len-1,1)
  {
    r = d[i]*dr ;
    ridge.i[i]=i;
    steps = nint(ceil(2*PI*r* 2)); % *2 to cover for sure every pixel
    phi = [0:2*PI:#(steps+1)];
    x = nint ( ref_pix_x + sin(pang+phi)*r );
    y = nint ( ref_pix_y + cos(pang+phi)*r );
    sel = unique(len_x*y + x);
    x = x[sel];
    y = y[sel];
    sel = where(0 <= x < len_x and 0 <= y < len_y);
    if (length(sel)<1) break;
    x = x[sel];
    y = y[sel];
    peak_index = where_max (img[y*len_x + x])[0];
    ridge.peak_value[i] = max (img[y*len_x + x]);
    ridge.peak_x[i] = x [peak_index];
    ridge.peak_y[i] = y [peak_index];

    %%%% determine ridge line with integral method
    pang = atan2 (ref_pix_x - ridge.peak_x[i] , ridge.peak_y[i] - ref_pix_y ); % to get proper position angle
    steps = nint(ceil(1*0.5*PI*r* 2)); % to cover for sure every pixel
    phi = [0:1*0.5*PI:#(steps+1)];

    (lsum,rsum) = (0,0);
    do {
      sum_diff = lsum -rsum;
      (lx,ly) = ( nint ( ref_pix_x - sin(pang+phi)*r ) , nint ( ref_pix_y + cos(pang+phi)*r ) );
      (rx,ry) = ( nint ( ref_pix_x - sin(pang-phi)*r ) , nint ( ref_pix_y + cos(pang-phi)*r ) );
      sel = where(0<=lx<len_x and 0<=ly<len_y and 0<=rx<len_x and 0<=ry<len_y);
      (rx, ry, lx, ly) = (rx[sel], ry[sel], lx[sel], ly[sel]);
      (lsum , rsum)= ( sum(img[ly*len_x + lx]) , sum(img[ry*len_x + rx]) );
      if (lsum == rsum) break;
      if (lsum > rsum) pang += _min(0.1,dr/r);
      else pang -= _min(0.1,dr/r);
    } while (sum_diff*(lsum -rsum) >= 0);
    
    ridge.flux_cent_x[i] = (length(rx)>0 ? rx [0] : 0);
    ridge.flux_cent_y[i] = (length(ry)>0 ? ry [0] : 0);
    }
  
  return ridge;
}

%
% Code from src/misc/simulation/rndbknpwrlc.sl
%
define rndbknpwrlc()
%!%+
%\function{rndbknpwrlc}
%\synopsis{simulates a light curve with a broken power law distributed power spectrum}
%\usage{(t, rate) = rndbknpwrlc(Integer_Type nt);}
%\qualifiers{
%\qualifier{beta1}{[=1.5]: first power law index of the power spectrum}
%\qualifier{beta1}{[=1.0]: second power law index of the power spectrum}
%\qualifier{perc}{[]}
%\qualifier{mean}{[=0]: mean rate of the simulated light curve}
%\qualifier{sigma}{[=1]: standard deviation of the simulated light curve}
%\qualifier{dt}{[=1]: time resolution of the simulated light curve}
%}
%\description
%    \code{nt} is the number of bins of the simulated lightcurve.
%    It should be a power of two best performance of the fast Fourier transform.
%!%-
{
  variable nt;
  switch(_NARGS)
  { case 1: nt = (); }
  { help(_function_name()); return; }

  % nt must be even and should also be a power of 2 for calculation speed reasons
  if(nt mod 2 != 0)
  { vmessage("error (%s): nt must be even", _function_name()); return; }

  variable betaph1 = qualifier("beta1", 1.5); % power law index of periodogram
  variable betaph2 = qualifier("beta2", 1.0); % power law index of periodogram
  variable perc = qualifier("perc", 0.8); % Percent of PSD after which break occurs
  variable mean = qualifier("mean", 0); %  Mean count rate of the simulated lightcurve
  variable sigma = qualifier("sigma", 1); % Standard deviation of the lightcurve to be simulated
  variable dt = qualifier("dt", 1); %  Time resolution of the lc

  variable minin = min((([0:(nt/2.)*perc]+1.)/(dt*nt))^(-(betaph1/2.)) );
  variable maxin = max((([0:(nt/2.)*(1.-perc)]+floor(nt/2.*perc)+2.)/(dt*nt))^(-(betaph2/2.)));

  %calculating powerlaw distributed periodogram
  variable   fac = [(([0:(nt/2.)*perc]+1.)/(dt*nt))^(-(betaph1/2.)),
		    ((([0:(nt/2.)*(1.-perc)]+floor(nt/2.*perc)+2.)/(dt*nt))^(-(betaph2/2.)))/maxin*minin ]; % 1/T,...,1/dt
  variable pos_real = grand(nt/2)*fac;
  variable pos_imag = grand(nt/2)*fac;
  pos_imag[nt/2-1] = 0;
  variable neg_real = pos_real[[0:nt/2-2]];
  variable neg_imag = -(pos_imag[[0:nt/2-2]]);
  variable real = [0., pos_real, reverse(neg_real)];
  variable imag = [0., pos_imag, reverse(neg_imag)];

  % simulate lc from its Fourier transform
  variable rate = Real(fft((real+(imag)*1i),1));

  % normalize to desired mean count rate and variance
  variable m = moment(rate);
  variable t = dt*[0:nt-1];
  rate = mean + (rate-m.ave)/m.sdev*sigma;

  return (t, rate);
}

%
% Code from src/misc/simulation/rndpwrlc.sl
%
define rndpwrlc()
%!%+
%\function{rndpwrlc}
%\synopsis{simulates a light curve with a power law distributed power spectrum}
%\usage{(t, rate) = rndpwrlc(Integer_Type nt);}
%\qualifiers{
%\qualifier{beta}{[=1.5] power law index of the power spectrum}
%\qualifier{mean}{[=0] mean rate of the simulated light curve}
%\qualifier{sigma}{[=1] standard deviation of the simulated light curve}
%\qualifier{dt}{[=1] time resolution of the simulated light curve}
%}
%\description
%    \code{nt} is the number of bins of the simulated lightcurve.
%    It should be a power of two best performance of the fast Fourier transform.
%    See also Timmer & Koenig (1995): "On generating power law noise",
%    A&A 300, 707-710
%!%-
{
  variable nt;
  switch(_NARGS)
  { case 1: nt = (); }
  { help(_function_name()); return; }

  variable nt_2 = int(nt/2);
  if(2*nt_2 != nt)
  { vmessage("error (%s): nt must be even", _function_name()); return; }

  variable dt = qualifier("dt", 1);  % time resolution of the lc

  return dt*[0:int(nt)-1], powerlaw_noise(nt;; __qualifiers());
}

%
% Code from src/slang/math/round2.sl
%
define round2()
%!%+
%\function{round2}
%\synopsis{Round to the nearest integral value or to given digit}
%\usage{Double_Type[] = round2( Double_Type[] value);}
%\altusage{Double_Type[] = round2( Double_Type[] value, Integer_Type digit );}
%\description
%   This function rounds its argument to the nearest integral value and
%   returns it as a floating point result. If the argument is an array,
%   an array of the corresponding values will be returned.
%   If a 2nd argument is given it is used as digit the value is supposed
%   to be rounded to.
%\seealso{round, floor2, ceil2, nint}
%!%-
{
  variable val, dig = 0;
  switch (_NARGS)
  { case 1: val = (); }
  { case 2: (val,dig)=();}
  { help(_function_name); return; }

  dig = nint(dig);
  return round( val * 10^(-dig) )*10^dig;  
}

define ceil2()
  %!%+
  %\function{ceil2}
  %\synopsis{Round value up to the nearest integral value or to given digit}
  %\usage{Double_Type[] = ceil2( Double_Type[] value);}
  %\altusage{Double_Type[] = ceil2( Double_Type[] value, Integer_Type digit );}
  %\description
  %   This function rounds its argument up to the nearest integral value and
  %   returns it as a floating point result. If the argument is an array,
  %   an array of the corresponding values will be returned.
  %   If a 2nd argument is given it is used as digit the value is supposed
  %   to be rounded up to.
  %\seealso{ceil, floor2, round2, nint}
  %!%-
{
    variable val, dig = 0;
    switch (_NARGS)
  { case 1: val = (); }
  { case 2: (val,dig)=();}
  { help(_function_name); return; }

    dig = nint(dig);
    return ceil( val * 10^(-dig) )*10^dig;
}

define floor2()
  %!%+
  %\function{floor2}
  %\synopsis{Round value down to the nearest integral value or to given digit}
  %\usage{Double_Type[] = floor2( Double_Type[] value);}
  %\altusage{Double_Type[] = floor2( Double_Type[] value, Integer_Type digit );}
  %\description
  %   This function rounds its argument down to the nearest integral value and
  %   returns it as a floating point result. If the argument is an array,
  %   an array of the corresponding values will be returned.
  %   If a 2nd argument is given it is used as digit the value is supposed
  %   to be rounded down to.
  %\seealso{floor, ceil2, round2, nint}
  %!%-
{
    variable val, dig = 0;
    switch (_NARGS)
  { case 1: val = (); }
  { case 2: (val,dig)=();}
  { help(_function_name); return; }

    dig = nint(dig);
    return floor( val * 10^(-dig) )*10^dig;
}

%
% Code from src/slang/physics/round_conf.sl
%
%%%%%%%%%%%%%%%%%%%
define round_conf()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{round_conf}
%\synopsis{converts confidence intervals after DIN 1333 and gives the rounded decimal place.}
%\usage{Struct_Type round_conf(Double_Type conf_lo, conf_val, conf_hi);}
%\altusage{Struct_Type round_conf(Double_Type val, sym_err);}
%\description
%    \code{conf_lo} is the lower confidence limit
%    \code{conf_val} is the best fit
%    \code{conf_hi} is the upper confidence limit
%    The alternative usage with two arguemts allows to specify a
%    value \code{val} with a symmetric uncertainty \code{sym_err}.
%
%    The returned structure contains the fields
%    - \code{err_lo} (the rounded lower error)
%    - \code{value_val} (the rounded best fit)
%    - \code{err_hi} (the rounded upper error)
%    - \code{digit} (the rounded decimal place of the error)
%
%\seealso{round_err, TeX_value_pm_error}
%!%-
{
  variable val_lo, val, val_hi, err, min_digit;
  variable err_lo, err_hi;
  switch(_NARGS)
  { case 3: (val_lo,val,val_hi) = ();
    if(val-val_lo<0 or val_hi-val<0)
    { vmessage("error in %s: value has to be within confidence limits (check order of arguments)",_function_name());
      return;
    }
     err_lo = round_err(val-val_lo;;__qualifiers);
     err_hi = round_err(val_hi-val;;__qualifiers);
    
    variable digit1 = err_lo.digit;
    variable digit2 = err_hi.digit;
    min_digit = _min(digit1,digit2);
    
    if (digit1!=digit2) {
      err_lo = round_err(val-val_lo; digits=min_digit);
      err_hi = round_err(val_hi-val; digits=min_digit);
    }

  }
  { case 2: (val,err) = ();
    err_lo = round_err(abs(err));
    err_hi = err_lo;
    min_digit = err_lo.digit;
  }
  { help(_function_name()); return; }
  
  val = round(val*10^(-min_digit))*10^min_digit;
  % account for numerical problems such as val=4.145, digits=-1 => round(val*10^(-min_digit))*10^min_digit = 4.1000000000000005 via fmt:
  variable fmt;
  if(min_digit<0)
    fmt = sprintf("%%.%df", nint(-min_digit));
  else
    fmt = "%.0f";
  val = atof(sprintf(fmt, val));
  return struct{err_lo=err_lo.value,value=val,err_hi=err_hi.value,digit=min_digit};
}

%
% Code from src/slang/physics/round_err.sl
%
%%%%%%%%%%%%%%%%%%%
define round_err()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{round_err}
%\synopsis{rounds an error after DIN 1333 and gives the rounded decimal place}
%\usage{Struct_Type round_err(Double_Type x);}
%\qualifiers{
%    \qualifier{digits}{decimal place where rounding will be applied (suspends DIN 1333)}
%}
%\description
%    \code{x} is the error which will be rounded
%
%    The returned structure contains the fields
%    - \code{value} (the rounded error)
%    - \code{digit} (the rounded decimal place of the error)
%
% EXAMPLES
%     round_err(0.1278) will return value=0.13 and digit=-2
%     round_err(1278)   will return value=1300 and digit= 2
%     since after DIN 1333, the decimal point may only be right after the rounded digit or left of it,
%     the error should be in LaTeX assigned as $13.0\\times 10^{2}$
%     (? comment: in order to have the correct number of significant digits shouldn't it be:
%     $0.13\\times 10^{4}$, $1.3\\times 10^{3}$, or $13\\times 10^{2}$)
%\seealso{round_conf}
%!%-
{
  variable x;
  switch(_NARGS)
  { case 1: x = (); }
  { help(_function_name()); return; }
  variable digits = 0;
  ifnot (x==0)
  {
    if(qualifier_exists("digits"))
      digits = qualifier("digits");
    else
    {
      digits = floor(log10(x)); % first significant decimal place
      if( round2( x/(10^digits),-14 ) < 3 ) % DIN 1333: if first significant decimal place is smaller than 3, then next decimal place is the one to be rounded
	digits--;
    }
    % account for numerical problems such as x=4.7-4.1 = 0.6000000000000005 causing ceil(x) = 7 instead of 6
    x = x*10^(-digits);
    if( x-int(x) > 10^(-15) ) % all decimal places beyond 15 are not trustworthy due to numerics
      x = ceil(x);
    x *= 10^digits;
    % account for numerical problems such as x=0.6, digits=-1 => ceil(x*10^(-digits))*10^digits = 0.6000000000000001 via fmt:
    variable fmt;
    if(digits<0)
      fmt = sprintf("%%.%df", nint(-digits));
    else
      fmt = "%.0f";
    x = atof(sprintf(fmt, x));
  }
  return struct{value=x, digit=digits};
}

%
% Code from src/slang/math/runs_test.sl
%
#ifexists normal_cdf

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define __runs_p (n1, n2, r)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Implement runs probability after
% "Statistics for management and economics"
% Mendenhall, Reinmush & Beaver 1989
% Use table in Appendix for small values
% of n1,n2 and use Gaussian approximation
% for larger.
{
  % for 2 <= n1, n2 <= 10
  variable small_p = Array_Type[45];
  % n1 = 2, n2 = 2 .. 10
  small_p[0] = [.333, .666, 1.];
  small_p[1] = [.200, .500, .900, 1.];
  small_p[2] = [.133, .400, .800, 1.];
  small_p[3] = [.095, .333, .714, 1.];
  small_p[4] = [.071, .286, .643, 1.];
  small_p[5] = [.056, .250, .583, 1.];
  small_p[6] = [.044, .222, .533, 1.];
  small_p[7] = [.036, .200, .491, 1.];
  small_p[8] = [.030, .182, .455, 1.];

  % n1 = 3, n2 = 3 .. 10
  small_p[9] = [.100, .300, .700, .900, 1.];
  small_p[10] = [.057, .200, .543, .800, .971, 1.];
  small_p[11] = [.036, .143, .429, .714, .929, 1.];
  small_p[12] = [.024, .107, .345, .643, .881, 1.];
  small_p[13] = [.017, .083, .283, .583, .833, 1.];
  small_p[14] = [.012, .067, .236, .533, .788, 1.];
  small_p[15] = [.009, .055, .200, .491, .745, 1.];
  small_p[16] = [.007, .045, .171, .455, .706, 1.];

  % n1 = 4, n2 = 4 .. 10
  small_p[17] = [.029, .114, .371, .629, .886, .971, 1.];
  small_p[18] = [.016, .071, .262, .500, .786, .929, .992, 1.];
  small_p[19] = [.010, .048, .190, .405, .690, .881, .976, 1.];
  small_p[20] = [.006, .033, .142, .333, .606, .833, .954, 1.];
  small_p[21] = [.004, .024, .109, .279, .533, .788, .929, 1.];
  small_p[22] = [.003, .018, .085, .236, .471, .745, .902, 1.];
  small_p[23] = [.002, .014, .068, .203, .419, .706, .874, 1.];

  % n1 = 5, n2 = 5 .. 10
  small_p[24] = [.008, .040, .167, .357, .643, .833, .960, .992, 1.];
  small_p[25] = [.004, .024, .110, .262, .522, .738, .911, .976, .998, 1.];
  small_p[26] = [.003, .015, .076, .197, .424, .652, .854, .955, .992, 1.];
  small_p[27] = [.002, .010, .054, .152, .347, .576, .793, .929, .984, 1.];
  small_p[28] = [.001, .007, .039, .119, .287, .510, .734, .902, .972, 1.];
  small_p[29] = [.001, .005, .029, .095, .239, .455, .678, .874, .958, 1.];

  % n1 = 6, n2 = 6 .. 10
  small_p[30] = [.002, .013, .067, .175, .392, .608, .825, .933, .987, .998, 1.];
  small_p[31] = [.001, .008, .043, .121, .296, .500, .733, .879, .966, .992, .999, 1.];
  small_p[32] = [.001, .005, .028, .086, .226, .413, .646, .821, .937, .984, .998, 1.];
  small_p[33] = [.000, .003, .019, .063, .175, .343, .566, .762, .902, .972, .994, 1.];
  small_p[34] = [.000, .002, .013, .047, .137, .288, .497, .706, .864, .958, .990, 1.];

  % n1 = 7, n2 = 7 .. 10
  small_p[35] = [.001, .004, .025, .078, .209, .383, .617, .791, .922, .975, .996, .999, 1.];
  small_p[36] = [.000, .002, .015, .051, .149, .296, .514, .704, .867, .949, .988, .998, 1.00, 1.];
  small_p[37] = [.000, .001, .010, .035, .108, .231, .427, .622, .806, .916, .975, .994, .999, 1.];
  small_p[38] = [.000, .001, .006, .024, .080, .182, .355, .549, .743, .879, .957, .990, .998, 1.];

  % n1 = 8, n2 = 8 .. 10
  small_p[39] = [.000, .001, .009, .032, .100, .214, .405, .595, .786, .900, .968, .991, .999, 1.00, 1.];
  small_p[40] = [.000, .001, .005, .020, .069, .157, .319, .500, .702, .843, .939, .980, .996, .999, 1.00, 1.];
  small_p[41] = [.000, .000, .003, .013, .048, .117, .251, .419, .621, .782, .903, .964, .990, .998, 1.00, 1.];

  % n1 = 9, n2 = 9 .. 10
  small_p[42] = [.000, .000, .003, .012, .044, .109, .238, .399, .601, .762, .891, .956, .988, .997, 1.00, 1.00, 1.];
  small_p[43] = [.000, .000, .002, .008, .029, .077, .179, .319, .510, .681, .834, .923, .974, .992, .999, 1.00, 1.00, 1.];

  % n1 = 10, n2 = 10
  small_p[44] = [.000, .000, .001, .004, .019, .051, .128, .242, .414, .586, .758, .872, .949, .981, .996, .999, 1.00, 1.00, 1.];

  variable p;
  if (n1 < 2 or n2 < 2)
    throw DomainError, "Only defined for n1,n2 >= 2";
  if (r < 2 or r > n1+n2)
    throw DomainError, "r must be in range 2 .. n1 + n2";

  % swap if needed
  if (n1 > n2)
    (n1,n2) = (n2,n1);

  if (n1 < 11 and n2 < 11)
    return small_p[(n1-2)*9+(n2-2)-((n1-2)*(n1-1))/2][r-2]; % access upper triangle matrix

  variable ex = 2.*n1*n2/(n1+n2) + 1;
  variable var2 = (ex-1)*(ex-2)/(n1+n2-1);

  variable z = (r-ex)/sqrt(var2);

  return round(1e3*normal_cdf(z))*1e-3; % table has only 3 significants
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%
define runs_test (sequence)
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{runs_test}
%\synopsis{Perform runs test on sequence}
%\usage{Int_Type = runs_test(Num_Type[]);}
%\qualifiers{
%  \qualifier{confidence}{[=0.05]: Critical test probability}
%  \qualifier{overmixing}{Test for overmixing}
%  \qualifier{undermixing}{Test for undermixing}
%}
%\description
%    This function performs the runs test (Wald-Wolfowitz test)
%    on the given array. The sequence is assumend to represent
%    a dichotom set with n1 where sequence > 0 and n2 the
%    complementary.
%
%    The test returns true if the sequence is sufficiently
%    randomly divided into the two sets (determined by the
%    confidence qualifier).
%
%    Per default the sequence is tested against over- and
%    undermixing but can be adjusted to only test for one with
%    the appropriate qualifier.
%
%\seealso{normal_cdf}
%!%-
{
  variable confidence = qualifier("confidence", 0.05);
  variable mode = qualifier_exists("overmixing") ? 1
    : (qualifier_exists("undermixing") ? -1 : 0);

  variable n2;
  variable n1 = where(sequence>0, &n2);

  variable s = Int_Type[length(sequence)];
  s[n1] = 1;
  s[n2] = 0;
  variable r = length(where(s[[1:]]-s[[:-2]]))+1;
  variable p;
  n1 = length(n1);
  n2 = length(n2);

  if (n1 < 2 or n2 < 2 or r < 2 or r > n1 + n2)
    p = NULL;
  else
    p = __runs_p(n1, n2, r);

  if (NULL == p)
    return 0;

  switch (mode)
  { case 0: return confidence/2.<p && confidence/2.<(1-p); }
  { case -1: return confidence < p; }
  { case 1: return confidence < (1-p); }
}
#endif

%
% Code from src/fitting/parameters/save_par_to_FITS_header_struct.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define save_par_to_FITS_header_struct()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{save_par_to_FITS_header_struct}
%\synopsis{saves fit-function and paramters to a FITS header structure}
%\usage{Struct_Type save_par_to_FITS_header_struct()}
%\description
%    The returned structure can be used as header keys
%    that are written to a FITS file.
%    This header can be read with \code{load_par_from_FITS_header}.
%\seealso{load_par_from_FITS_header, load_par, save_par, fits_write_binary_table}
%!%-
{
  variable p = get_params();
  variable s = struct { fit_fun = get_fit_fun(), num_pars=length(p) };
  variable i = 1;
  foreach p (p)
  {
    variable si = @Struct_Type( "par"+string(i)+["", "_min", "_max", "_frz", "_tie", "_fun"] );
    set_struct_fields(si,
		      p.value,
		      p.min == -DOUBLE_MAX ? "-inf" : p.min,
		      p.max ==  DOUBLE_MAX ? "+inf" : p.max,
		      p.freeze,
		      p.tie==NULL ? 0 : p.tie,
		      p.fun==NULL ? 0 : p.fun
		     );
    s = struct_combine(s, si);
    i++;
  }
  return s;
}

%
% Code from src/plot/save_plot.sl
%
define save_plot()
%!%+
%\function{save_plot}
%\synopsis{saves ISIS spectral data into a FITS file}
%\usage{save_plot([filename[, ids]]);}
%\qualifiers{
%\qualifier{A}{Save the data in Angstrom and not in keV.}
%}
%\description
%   This functions saves all data and model points as currently
%   noticed and rebinned to the file 'filename'.fits. Each dataset is
%   stored in an own extension. The current model is save in
%   'filename'.par. By default the name 'save_plot' is chosen.
%   
%   Hereby the values are given in counts/bin and in 
%   photons/s/cm^2/keV. To calculate the flux from the observed 
%   data, the function get_convolved_model_flux() was used.
%   
%   Additionally, information on the instrument, the target, the grating,
%   the filename of the model, the functions used for the model 
%   and the frame time are stored in the header.
%!%-
{
  variable ids=all_data, filename=NULL;
  switch(_NARGS)
  { case 1: filename = (); }
  { case 2: (filename, ids) = (); }
  { help(_function_name()); return; }

   if(filename==NULL)  filename = "save_plot";
  
   %% open the fits file
   variable F = fits_open_file(filename+".fits", "c");
  
   %% get info of the data
   variable data_info;
   variable FitVerbose = Fit_Verbose;
   Fit_Verbose = -1;
   variable fit_stat;
   () = eval_counts(&fit_stat);
   variable chi2 = fit_stat.statistic/(fit_stat.num_bins - fit_stat.num_variable_params);
		    
  variable sd = Struct_Type[length(ids)];
  variable si = Struct_Type[length(ids)];
  variable i,ind;
  variable plotunit = "keV";
  if (qualifier_exists("A")) {
    plotunit = "A";
  }
  
   for (i = 0; i < length(ids); i++) {
      
      
      %%% - get data info - %%%
      data_info = get_data_info(ids[i]);
      variable filt = data_info.notice_list;

      
      
      %%% - get data counts - %%%
      variable d = get_data_counts(ids[i]);
      struct_filter(d,filt);
     if ( not qualifier_exists("A")) {
       d = _A(d);
     }
     
     %%% - get model counts - %%%
     variable m = get_model_counts(ids[i]);
      struct_filter(m,filt);
     if ( not qualifier_exists("A")) {
       m = _A(m);
     }
      
      %%% - get data flux - %%%
      variable mf,mf_re;
      variable f = struct {bin_lo=d.bin_lo,bin_hi=d.bin_hi,
	 value = m.value*0 +1, err = m.value*0+1};
      mf_re = m.value*0 +1;
      try {	 
	 flux_corr(ids[i]);
	 f = get_data_flux(ids[i]);
	 struct_filter(f,filt);
	 if ( not qualifier_exists("A")) {
	    f = _A(f);
	 }
	 
	 % original rebinning
	 rebin_data(ids[i],0);
	 () = eval_counts;
	 % apply flux correction
	 flux_corr_model_counts(ids[i]);
	 mf = get_convolved_model_flux(ids[i]);
	 if ( not qualifier_exists("A")) {
	    mf = _A(mf);
	 }
	 mf_re = rebin(d.bin_lo,d.bin_hi,mf.bin_lo,mf.bin_hi,mf.value);
	 
      } catch AnyError:{
	 message("No flux correction possible for this configuration. \n");
      }
      
      
      % restore binning of the data
      rebin_data(ids[i], data_info.rebin);
      ignore(ids[i]);
      notice_list(ids[i], data_info.notice_list);
      
      %%% - write structures - %%%
      sd[i] = struct {
	 bin_lo     = d.bin_lo,
	 bin_hi     = d.bin_hi,
	 value      = d.value,
	 err        = d.err,
	 model      = m.value,
	 model_flux = mf_re/(f.bin_hi-f.bin_lo),
	 flux       = f.value/(f.bin_hi-f.bin_lo),
	 flux_err   = f.err / (f.bin_hi-f.bin_lo)
      };
      
      
      si[i] = struct {
	 TUNIT1 = plotunit,
	 TUNIT2 = plotunit,
	 TUNIT3 = "counts/bin",
	 TUNIT4 = "counts/bin",
	 TUNIT5 = "counts/bin",
	 TUNIT6 = "photons/s/cm^2/keV",
	 TUNIT7 = "photons/s/cm^2/keV",
	 TUNIT8 = "photons/s/cm^2/keV"
      };
      fits_write_binary_table(F, "spectrum "+string(ids[i]), sd[i], si[i]);
      
      %% - write additionaly data info to the extension - %%%
 
      fits_update_key(F, "NUMDAT",length(ids));
      fits_update_key(F, "MODEL",+filename+".par");
      fits_update_key(F, "FITFUN",get_fit_fun);
      fits_update_key(F, "INSTRUME", data_info.instrument);
      fits_update_key(F, "TARGET", data_info.target);
      fits_update_key(F, "GRATING", data_info.grating);
      fits_update_key(F, "FRAMET", data_info.frame_time);
      fits_update_key(F, "EXPOSURE",get_data_exposure(i+1),"exposure in sec");
      fits_update_key(F, "CHI2RED", chi2, "reduced chi2 of the model");
      fits_update_key(F, "TSTART", data_info.tstart, "start of the observation");
      fits_update_key(F, "SPECPATH", getcwd, "path to the spectral files");
      fits_update_key(F, "SPECNAME", data_info.file, "spectral file");
      fits_update_key(F, "BGDNAME", data_info.bgd_file, "background file");
   }
   %% - write model - %%%
   save_par(filename+".par");
   
   fits_close_file(F);   
   Fit_Verbose = FitVerbose;
}

%
% Code from src/slang/save_slang_variable.sl
%
% -*- mode: slang; mode: fold -*-

% known data-types, which code can be written using a simple format code %{{{
private define _slang_primitive_datatype_format_string(t)
{
  return t == Int16_Type    ? "%dH"
      :  t == UInt16_Type   ? "%dUH"
      :  t == Int32_Type    ? "%dL"
      :  t == UInt32_Type   ? "%ldUL"
      :  t == Int64_Type    ? "%lldLL"
      :  t == UInt64_Type   ? "%lldULL"
      :  t == Double_Type   ? "%S"
      :  t == Float_Type    ? "%Sf"
      :  t == Complex_Type  ? "%S"
      :  t == Char_Type     ? "typecast(%d,Char_Type)"
      :  t == UChar_Type    ? "'%c'"
      :  t == String_Type   ? `"%s"`
      :  t == BString_Type  ? `"%s"B`
      :  t == Null_Type     ? "%S"
      :  t == Void_Type     ? "%S"
      :  t == DataType_Type ? "%S"
      : NULL;
}

private define _is_primitive_datatype(t)
{
  return _slang_primitive_datatype_format_string(t) != NULL;
}
%}}}

% recursive function for converting the given S-lang
% variable into a string and writing into a file
private define _slang_variable_to_file();
private define _slang_variable_to_file(varname, var, fp)
{
  variable fmt = _slang_primitive_datatype_format_string(typeof(var));
  if (fmt != NULL) { % primitive datatype %{{{
    % print out variable declaration
    if (varname != NULL) { ()=fprintf(fp, "variable %s = ", varname); }
    % check on infinity or NaN
    if (__is_numeric(var) && isinf(var)) {
      ()=fprintf(fp, "%s_Inf", sign(var) == -1 ? "-" : "");
    } else if (__is_numeric(var) && isnan(var)) {
      ()=fprintf(fp, "%s_NaN", substr(string(var),1,1) == "-" ? "" : "-");
    } else {
      ()=fprintf(fp, fmt, var);
    }
    % ending
    if (varname != NULL) { ()=fprintf(fp, ";\n"); }
    return;
  }
%}}}
  variable type = typeof(var);
  % objects
  variable i, n, elementvarname, keys, values, ind;
  if(type == Ref_Type)  % references (as values of struct fields) %{{{
  {
    % convert reference to executable string
    ind = string(var);
    % check on private function/variable
    if (__get_reference(substr(ind, 2, -1)) == NULL) {
      % FIXME: ps has not been assigned!
      % if (ps == NULL) { vmessage("error: '%s' is a private function/variable", ind); return; }
    }
    ()=fprintf(fp, "variable %s = %s;\n", varname, ind);
  }
%}}}
  else if(type == Array_Type)  % arrays %{{{
  {
    variable len = length(var);
    variable dims, ndims, datatype; (dims, ndims, datatype) = array_info(var);
    % initialize empty array
    ()=fprintf(fp, "variable %s = %S[%d];\n", varname, datatype, len);
    if(ndims > 1)  { reshape(var, len); }  % flatten array
    if (_is_primitive_datatype(datatype)) {
      % print elements in packages
      variable ps = qualifier("packagesize", datatype == String_Type ? 5 : 10);
      _for n (0, len/ps, 1) {
	ind = [n*ps:(n+1)*ps < len ? ((n+1)*ps)-1 : len-1];
	if (length(ind) > 0) {
	  ()=fprintf(fp, "%s[[%d:%d]] = [\n  ", varname, ind[0], ind[-1]);
	  _for i (0, length(ind)-1, 1) {
	    _slang_variable_to_file(NULL, var[ind[i]], fp);
	    if (i < length(ind)-1) { ()=fprintf(fp, ","); }
	  }
	  ()=fprintf(fp, "\n];\n");
	}
      }
    }
    else { % array contains objects -> every element will be a variable
      _for n (0, len-1, 1) {
	% variable name for current element
	elementvarname = sprintf("%s_i%d", varname, n);
	% define variable for element and build it recursively
	()=fprintf(fp, "%% array element %d %%{{{\n", n);
	_slang_variable_to_file(elementvarname, var[n], fp;; __qualifiers);
	()=fprintf(fp, "%%}}}\n");
	()=fprintf(fp, "%s[%d] = __tmp(%s);\n", varname, n, elementvarname);
      }
    }
    if(ndims > 1)
    { % reshape finally
      ()=fprintf(fp, "%% reshape to array's dimensions\nreshape(%s, [%s]);\n",
		 varname, strjoin(array_map(String_Type, &string, dims), ","));
      reshape(var, dims);
    }
  }
%}}}
  else if(type == Assoc_Type)  % associative arrays %{{{
  {
    % get keys and values
    keys = assoc_get_keys(var);
    values = assoc_get_values(var);
    len = length(keys);
    % init associative array
    ()=fprintf(fp, "variable %s = Assoc_Type[%s];\n", varname, string(_typeof(values)));
    % fill array and take value's data type into account
    if (_is_primitive_datatype(_typeof(values))) {
      _for n (0, len-1, 1) {
        ()=fprintf(fp, `%s["%s"] = `, varname, keys[n]);
	_slang_variable_to_file(NULL, values[n], fp);
	()=fprintf(fp, ";\n");
      }
    }
    else { % objects
      _for n (0, len-1, 1) {
	% variable name for current element
	elementvarname = sprintf("%s_i%d", varname, n);
	% define element variable and call yourself recursively
	()=fprintf(fp, "%% associative array's element %d %%{{{\n", n);
	_slang_variable_to_file(elementvarname, values[n], fp;; __qualifiers);
	()=fprintf(fp, "%%}}}\n");
	()=fprintf(fp, "%s[\"%s\"] = __tmp(%s);\n", varname, keys[n], elementvarname);
      }
    }
  }
%}}}
#ifeval __get_reference("vector") != NULL
  else if(type == Vector_Type) %{{{
  {
    variable vector_constructor = "vector("B;
    % each component is a variable
    _for n ('x', 'z', 1) {
      elementvarname = sprintf("%s_v%c", varname, n);
      ()=fprintf(fp, "%% %c component%%{{{\n", n);
      _slang_variable_to_file(elementvarname, get_struct_field(var, char(n)), fp;; __qualifiers);
      ()=fprintf(fp, "%%}}}\n");
      vector_constructor += sprintf("\n  __tmp(%s)%s", elementvarname, (n!='z' ? "," : ""));
    }
    % create the vector
    ()=fprintf(fp, "%% create vector\n");
    ()=fprintf(fp, "variable %s = %s\n);\n", varname, vector_constructor);
  }
%}}}
#endif
  else if(is_struct_type(var))  % also works with typedef'ed struct's %{{{
  {
    % get struct field names and define them
    keys = get_struct_field_names(var);
    variable fieldsvarname = sprintf("%s_flds", varname);
    ()=fprintf(fp, "%% field names %%{{{\n");
    _slang_variable_to_file(fieldsvarname, keys, fp;; __qualifiers);
    ()=fprintf(fp, "variable %s = @Struct_Type(__tmp(%s));\n", varname, fieldsvarname);
    ()=fprintf(fp, "%%}}}\n");
    % loop over fields
    _for n (0, length(keys)-1, 1) {
      variable value = get_struct_field(var, keys[n]);
      if (_is_primitive_datatype(typeof(value))) {
	% assign primitive value directly
	()=fprintf(fp, "set_struct_field(%s, \"%s\", ", varname, keys[n]);
        _slang_variable_to_file(NULL, value, fp);
	()=fprintf(fp, ");\n");
      }
      else { % objects get their own variables
	elementvarname = sprintf("%s_f%d", varname, n);
	()=fprintf(fp, "%% field value %d %%{{{\n", n);
	_slang_variable_to_file(elementvarname, value, fp;; __qualifiers);
	()=fprintf(fp, "%%}}}\n");
	()=fprintf(fp, "set_struct_field(%s, \"%s\", __tmp(%s));\n", varname, keys[n], elementvarname);
      }
    }
  }
%}}}
  else if(type == List_Type) %{{{
  {
    % init empty list
    ()=fprintf(fp, "variable %s = list_new();\n", varname);
    _for n (0, length(var)-1, 1) {
      if (_is_primitive_datatype(typeof(var[n]))) {
	% append primitive value directly
	()=fprintf(fp, "list_append(%s, ", varname);
        _slang_variable_to_file(NULL, var[n], fp);
	()=fprintf(fp, ");\n");
      }
      else { % objects get their own variables
	elementvarname = sprintf("%s_l%d", varname, n);
	()=fprintf(fp, "%% list element %d %%{{{\n", n);
	_slang_variable_to_file(elementvarname, var[n], fp;; __qualifiers);
	()=fprintf(fp, "%%}}}\n");
	()=fprintf(fp, "list_append(%s, __tmp(%s));\n", varname, elementvarname);
      }
    }
  }
%}}}
  else
  {
    vmessage("error: data-type %S not supported", typeof(var));
    return;
  }
}


%%%%%%%%%%%%%%%%%%%%%%%%
define save_slang_variable()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{save_slang_variable}
%\synopsis{allows to save S-Lang variables into a file}
%\usage{save_slang_variable(file, &var1, &var2, ...);}
%\qualifiers{
%    \qualifier{edit}{open an editor to modify the variables;
%             in that case all variables have to be passed
%             as references and will be set to their new
%             values after the editor is closed}
%    \qualifier{delete}{delete the file after editing (requires the
%             'edit' qualifier to be set); note that the
%             given variables are still modified!}
%}
%\description
%    The S-Lang code defining the given variables is saved
%    into a file, specified by either the filename or an
%    already opened file-pointer. In order to handle arrays
%    with a large number (>1000) of items as well as
%    complex structures, the S-Lang code uses temporary
%    variable names. Their values are assigned step by step
%    to avoid a stack overflow. The file can be evaluated
%    later to push the saved variables onto the stack (see
%    the example).
%
%    This function allows to modify the given variables as
%    well. In that case the 'edit'-qualifier has to be set
%    and all passed variables have to be given as references.
%    The file the variables are saved into is shown in the
%    editor specified by the EDITOR environment variable
%    or jed, if EDITOR is undefined. After saving the file
%    and closing the editor, the file is evaluated, which
%    should push the (modified) S-Lang objects onto the
%    stack. These objects are finally assigned to the given
%    variables.
%
%    NOTE: the latter feature is based on the function
%          'edit_var', which does the same except that the
%          main purpose is to edit the variables using a
%          temporary file. Here, the S-lang code is human
%          readable as well, since no temporary variables
%          are used to assign the values stepwise. In that
%          case, however, stack overflow errors may occure.
%          But who wants to edit such large variables...?
%    
%    The function supports the following data-types:
%      Integer_Type, Double_Type, Complex_Type,
%      Char_Type, String_Type, BString_Type,
%      Null_Tpye, Void_Type (=Undefined_Type),
%    as well as
%      Array_Type, Assoc_Type, Struct_Type, List_Type
%      Vector_Type, Ref_Type (as structure fields)
%\example
%    % define a structure 
%    variable a = struct { example = "foo" };
%    
%    % save the structure into a file
%    save_slang_variable("mystruct.sl", a);
%
%    % restore the variable into a new one
%    variable b;
%    (b,) = evalfile("mystruct.sl");
%    
%    % edit the original variable
%    % using a temporary filename
%    save_slang_variable("/tmp/myedit", &a; edit, delete);
%
%    % this operation can be performed using
%    % 'edit_var' as well (but better readable)
%    edit_var(&a);
%\seealso{evalfile, edit_var}
%!%-
{
  if (_NARGS < 2) { help(_function_name()); return; }

  % get filename and given variables
  variable args = __pop_list(_NARGS-1);
  variable fname = ();

  % given variables should be edited
  variable n, edit = qualifier_exists("edit");
  % check which given variables are references
  variable varref = Ref_Type[length(args)];
  _for n (0, length(args)-1, 1) {
    if (typeof(args[n]) == Ref_Type) { varref[n] = args[n]; }
  }
  % if in edit mode every variable has to be given as reference
  % however, if none references are given at all, ignore edit mode
  if (edit and any(_isnull(varref))) {
    message("error: editing requires references exclusively");
    return;
  }
  % dereference
  _for n (0, length(args)-1, 1) {
    if (varref[n] != NULL) { args[n] = @(args[n]); }
  }

  % pick temporary variable names for each given variable
  variable varname = String_Type[length(args)];
  _for n (0, length(varname)-1, 1) {
    (,varname[n]) = varref[n] != NULL ? get_variable_name(varref[n]) : (0,sprintf("v%02d", n+1));
    varname[n] = sprintf("_%s_%02d", varname[n], n+1);
  }

  % recursively write variables and their values into the file
  variable fp = (typeof(fname) == File_Type ? fname : fopen(fname, "w+"));
  ()=fprintf(fp, "%% -*- mode: slang; mode: fold -*-\n\n");
  _for n (0, length(varname)-1, 1) {
    ()=fprintf(fp, "%% %d%s variable %%{{{\n", n+1, n == 10 ? "th" : n == 11 ? "th" : n mod 10 == 0 ? "st" : n mod 10 == 1 ? "nd" : n mod 10 == 2 ? "rd" : "th", varname[n]);
    _slang_variable_to_file(varname[n], args[n], fp;; __qualifiers);
    ()=fprintf(fp, "%%}}}\n\n");
  }

  % finally, push variables onto the stack and uninitialize them
  ()=array_map(Integer_Type, &fprintf, fp, "__tmp(%s);  %% left on stack\n", varname);

  if (typeof(fname) == String_Type) { ()=fclose(fp); }

  % eventually open editor
  if (edit) {
    variable e = getenv("EDITOR");
    if (e == NULL) { e = "jed"; }
    if (system(sprintf("%s %s", e, fname))) {
      message("warning: running $EDITOR raised an error");
      return;
    }

    try(e) { ()=evalfile(fname); } % will push all variables onto the stack
    catch AnyError:
    {
      message("error: unable to evaluate the file");
      vmessage(`"%s" exception:
%s
%s:%d`, e.descr, e.message, e.file, e.line);
    }

    % assign the new values to the given variables
    args = __pop_list(length(varref));
    _for n (0, length(args)-1, 1) {
      @(varref[n]) = args[n];
    }

    % eventually remove the file
    if (qualifier_exists("delete")) {
      if (typeof(fname) == File_Type) { vmessage("warning: 'file' has to be given as filename to be deleted"); }
      else { ()=remove(fname); }
    }
  }
}


%%%%%%%%%%%%%%%%%%%%%%%%
define storevar()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{storevar}
%\synopsis{use `save_slang_variable' instead}
%!%-
{
  message("\nuse save_slang_variable instead\n");
  return "use save_slang_variable instead";
}

%
% Code from src/fitting/parameters/save_statistics.sl
%
define save_statistics()
%!%+
%\function{save_statistics}
%\synopsis{saves the fit-statistic in a textfile}
%\usage{save_statistics(String_Type filename);}
%\seealso{eval_counts}
%!%-
{
  variable filename;
  switch(_NARGS)
  { case 1: filename = (); }
  { help(_function_name()); return; }

  variable F = fopen(filename, "w");
  variable s;
  variable FV = Fit_Verbose;
  Fit_Verbose = -1;
  ()=eval_counts(&s);
  Fit_Verbose = FV;
  ()=fprintf(F, "Parameters [Variable] = %d [%d]\n", length(get_params()), s.num_variable_params);
  ()=fprintf(F, "            Data bins = %d\n", s.num_bins);
  variable stat = strtok(get_fit_statistic(), ";")[0];
  if(stat == "chisqr")
  {
    ()=fprintf(F, "           Chi-square = %0.7g\n", s.statistic);
    if(s.num_bins > s.num_variable_params)
      ()=fprintf(F, "   Reduced chi-square = %0.7g\n", s.statistic/(s.num_bins-s.num_variable_params));
  }
  else
    ()=fprintf(F, "%20s = %0.7g\n", "Statistic ("+stat+")", s.statistic);

  ()=fclose(F);
}

%
% Code from src/data/savitzky_golay.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

% if the references exist the next block will be evaluated,
% otherwise it won't be loaded (this is the reason why we need "#"-statement here.

% see Numerical Recipes, Third Edition, Section 14.9
% see Savitzky & Golay 1964, Analytical Chemistry, 36, 1627
% see Ziegler 1981, Applied Spectroscopy, 35, 88
% see Bromba & Ziegler 1981, Analytical Chemistry, 53, 1583

define savitzky_golay_coefficients() %{{{
%!%+
%\function{savitzky_golay_coefficients}
%\synopsis{Calculate the Savitzky-Golay coefficients used for data smoothing}
%\usage{Double_Type[] c = savitzky_golay_coefficients(positive Integer_Type nl, nr, p; qualifiers)}
%\description
%    Calculate the Savitzky-Golay coefficients used for data smoothing. The arguments
%    `nl'/`nr' are hereby the data points to the left/right while `p' is the polynomial
%    degree (all have to be positive integer numbers; otherwise their absolute value is
%    rounded to the next nearest integer). The quanity `p' must not exceed `nl'+`nr'!
%    The returned array of coefficients is ordered as [c_-nl, ..., c_0, ..., c_nr].
%\notes
%    Requires GSL module.
%\qualifiers{
%\qualifier{derivative [=0]}{Savitzky-Golay coefficients used for calculating the numerical
%      derivative of the corresponding order (must not exceed `p', i.e., the order of
%      the polynomial).}
%}
%\example
%    % coefficients to smooth data:
%    c = savitzky_golay_coefficients(15,15,4);
%    % -> coefficients to compute first derivative:
%    c = savitzky_golay_coefficients(15,15,4; derivative=1);
%\seealso{savitzky_golay_smoothing}
%!%-
{
  if(_NARGS!=3)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable nl, nr, p; % nl/nr: number of points to the left/right used for smoothing, p: polynomial degree
    (nl, nr, p) = ();
    nl = nint( abs( nl ) ); % make sure that number of data points to the left is a positive integer
    nr = nint( abs( nr ) ); % make sure that number of data points to the right is a positive integer
    p  = nint( abs( p  ) ); % make sure that degree of polynomial is a positive integer
    variable ld = qualifier("derivative", 0); % which derivative is requested, zero for data fitting
    ld = nint( abs( ld ) ); % make sure that derivative is a positive integer
    if(p>nl+nr)             % p>nl+nr: degree of polynomial is larger than number of data points fitted to
      throw UsageError, sprintf("Usage error in '%s': p must not exceed nl+nr.", _function_name());
    if(ld>p)                % ld>p: desired derivative is larger than degree of polynomial
      throw UsageError, sprintf("Usage error in '%s': The order of the derivative must not exceed p.", _function_name());
    % see Numerical Recipes, Third Edition, Section 14.9 for the following:
    variable matrix = Double_Type[p+1,p+1];
    variable k = [-nl:nr:1];
    variable i,j;
    _for i(0,p,1)
    {
      _for j(0,i,1) % matrix is symmetric
      {
	matrix[i,j] = sum(k^(i+j)); % eq. 14.9.4 in Numerical Recipes, Third Edition
	matrix[j,i] = matrix[i,j];
      }
    }
    variable inverse_matrix = gsl->linalg_LU_invert(gsl->linalg_LU_decomp(matrix));
    variable temp = [0:p:1];
    variable c = Double_Type[nl+nr+1];
    _for i(-nl,nr,1)
    {
      c[i+nl] = sum(inverse_matrix[ld,*]*i^temp); % eq. 14.9.6 in Numerical Recipes, Third Edition
    }
    return prod([2:ld])*c; % prod([2:ld]) = factorial(ld) due to computing the ld-th derivative of a polynomial
  }
}%}}}

define savitzky_golay_smoothing() %{{{
%!%+
%\function{savitzky_golay_smoothing}
%\synopsis{Smooth noisy data by using a Savitzky-Golay filter}
%\usage{Double_Type[] smoothed_data = savitzky_golay_smoothing(Double_Type[] data,
%                                  positive Integer_Type nl, nr, p; qualifiers)}
%\description
%    The idea of Savitzky-Golay filtering is to approximate the given data within
%    a moving window (ranging from `nl' data points to the left to `nr' data points
%    to the right) by a polynomial of order `p'. The respective polynomials are -
%    in principle - determined by least-squares fits to the window data points.
%    Luckily, Savitzky & Golay found a way to replace the fitting and evaluating
%    of the polynomial by taking just linear combinations of neighboring data points
%    tremendously speeding up the smoothing process. However, their method, which
%    is implemented here, is valid only for regularly spaced data points. For more
%    details on the method, see ``Numerical Recipes'', Third Edition, Section 14.9.
%
%    Note that the input parameters `nl', `nr', and `p' have to be positive integer
%    numbers (otherwise their absolute value is automatically rounded to the next
%    nearest integer). The quantity `p' must not exceed the minimum window semi-
%    length min(`nl',`nr') and `nl'+`nr' has to be smaller than or equal to the
%    number of total data points.
%
%    Practical hint: Best results are obtained when the full width of the degree 4
%    Savitzky-Golay filter is between 1 and 2 times the full width at half maximum
%    of the desired features in the data.
%\notes
%    Requires GSL module.
%\qualifiers{
%\qualifier{derivative [=0]}{Apart from smoothing noisy data, the Savitzky-Golay method
%      is also capable to compute numerical derivatives from the fitted polynomials.
%      In order to calculate the `k'-th derivative of the data, set this qualifier
%      equal to `k' and divide the returned array by the data stepsize to the power
%      of `k'. Note that `k' has to be a positive integer (otherwise it is replaced
%      by the next nearest integer of its absolute value). Note that `k' must not
%      exceed `p'.}
%\qualifier{periodic}{Set this qualifer to assume periodic boundary conditions for your
%      data avoiding special treatment of data points close to the edges and thus
%      speeding up the computation. In this case, the condition `p'<min(`nl',`nr')
%      is replaced by `p'<`nl'+`nr'.}
%}
%\example
%    x = [0:20:#1000];
%    data = sin(x);
%    data = data + 0.1*grand(length(x));
%    smoothed_data = savitzky_golay_smoothing(data,100,100,4);
%    first_derivative = savitzky_golay_smoothing(data,100,100,4; derivative=1)/(20./1000.);
%    second_derivative = savitzky_golay_smoothing(data,100,100,4; derivative=2)/(20./1000.)^2;
%\seealso{savitzky_golay_coefficients}
%!%-
{
  if(_NARGS!=4)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable data, nl, nr, p; % data: data to smooth, nl/nr: number of ponts to the left/right used for smoothing, p: polynomial degree
    (data, nl, nr, p) = ();
    nl = nint( abs( nl ) ); % make sure that number of data points to the left is a positive integer
    nr = nint( abs( nr ) ); % make sure that number of data points to the right is a positive integer
    p  = nint( abs( p  ) ); % make sure that degree of polynomial is a positive integer
    variable ld = qualifier("derivative", 0); % which derivative is requested, zero for data fitting
    ld = nint( abs( ld ) ); % make sure that derivative is a positive integer
    if(p>nl+nr)             % p>nl+nr: degree of polynomial is larger than number of data points fitted to
      throw UsageError, sprintf("Usage error in '%s': p must not exceed nl+nr.", _function_name());
    if(ld>p)                % ld>p: desired derivative is larger than degree of polynomial
      throw UsageError, sprintf("Usage error in '%s': The order of the derivative must not exceed p.", _function_name());
    if(nl+nr>length(data))  % the number of data points used for fitting is larger than the number of available data points
      throw UsageError, sprintf("Usage error in '%s': The number of data points required for fitting is larger than the number of available data points. Reduce nl and/or nr.", _function_name());
    variable coeffs = savitzky_golay_coefficients(nl,nr,p; derivative=ld); % get the Savitzky-Golay-Coefficients
    variable len = length(data);
    % instead of using a for-loop to smooth over neighboring data, create new arrays containing the shifted data
    % then, e.g., shift(data,-1)+shift(data,0)+shift(data,1) is the average over next neighbors
    variable temp = Double_Type[ len, nl+nr+1 ];
    variable i;
    _for i( -nl, nr, 1)
    {
      temp[*,i+nl] = shift(data,i)*coeffs[i+nl];
    }
    variable data_smooth = sum(temp,1);
    % so far: periodic boundary conditions due to use of function 'shift'
    % to have non-periodic boundary condition, the data points at the beginning and end need special treatment:
    if(qualifier_exists("periodic")!=1)
    {
      if(p>nl or p>nr) % p>nl, p>nr: degree of polynomial is larger than the number of data points fitted to at the edges of the data array
	throw UsageError, sprintf("Usage error in '%s': p must not exceed the minimum window semi-length min(nl,nr).", _function_name());
      _for i( 0, nl-1, 1) % data points at the beginning
      {
	variable nr_min = nint(_min(nr,len-i-1));
	coeffs = savitzky_golay_coefficients(i, nr_min, p; derivative=ld);
	data_smooth[i] = sum(data[[0:nr_min+i:1]]*coeffs);
      }
      variable i_temp = i+1;
      _for i( nint(_max(i_temp,len-nr-1)), len-1, 1) % data points at the end
      {
	coeffs = savitzky_golay_coefficients(nl, len-1-i, p; derivative=ld);
	data_smooth[i] = sum(data[[i-nl:len-1:1]]*coeffs);
      }
    }
    return data_smooth;
  }
}%}}}


#endif

%
% Code from src/slang/structs/histograms/scale_hist.sl
%
define scale_hist(hist, scal){
%!%+
%\function{scale_hist}
%\synopsis{scale a histogram}
%\usage{Struct_Type hist = scale_hist(hist, scal);}
%\description
%    Scale a histogram structure with fields bin_lo, bin_hi, value, and
%    err. The value field is scaled by the factor scal. The err field,
%    if present is scaled accordingly. If other fields are present,
%    those are preserved and passed on. Only presence of value and 
%    err are checked for, but err is optional.
%\seealso{add_hist, shift_hist, stretch_hist}
%!%-   
   if(_NARGS==0) {  help(_function_name()); return; }
   
   variable ns = struct_copy(hist);
   ifnot(struct_field_exists(ns, "value")){
	  help(_function_name()); return;
   }
   ns.value *= scal;
   if(struct_field_exists(ns,"err")){
	  ns.err *= scal;
   }
   return ns;
}

%
% Code from src/data/timing/scargle.sl
%
define scargle()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{scargle}
%\synopsis{Computes the lomb-scargle periodogram of an unevenly sampled lightcurve}
%\usage{Struct_Type res = scargle (t, c); % where t and c contain time and counts of the lc}
%\qualifiers{
%\qualifier{fmin}{minimum frequency to be used (NOT ANGULAR FREQ!), has precede over pmin, pmax}
%\qualifier{fmax}{maximum frequency to be used (NOT ANGULAR FREQ!), has precede over pmin, pmax}
%\qualifier{pmin}{minimum period to be used}
%\qualifier{pmax}{maximum period to be used}
%\qualifier{omega}{array of angular frequencies for which the PSD values are desired;
%                    if set, value for numf will be reset to length of omega during code}
%\qualifier{noise}{for the normalization of the periodogram.
%                    if not set, equal to the variance of the original lc.}
%\qualifier{numf}{number of independent frequencies}
%\qualifier{old}{if set computing the periodogram according to J.D. Scargle, 1982, ApJ 263, 835
%                  if not set, computing the periodogram with the fast algorithm
%                  of W.H. Press and G.B. Rybicki, 1989, ApJ 338, 277.}
%\qualifier{nu}{if set, output structure also contains frequency}
%\qualifier{om}{if set, output structure also contains angluar frequency}
%}
%\description
%     (transcribed from IDL-program scargle.pro)
%
%     The Lomb Scargle PSD is computed according to the definitions
%     given by Scargle, 1982, ApJ 263, 835, and Horne and Baliunas,
%     1986, ApJ 302, 757. Beware of patterns and clustered data
%     points as the Horne results break down in this case! Read and
%     understand the papers and this code before using it! For the
%     fast algorithm read W.H. Press and G.B. Rybicki, 1989, ApJ 338,
%     277.
%
%     The code is still stupid in the sense that it wants normal
%     frequencies, but returns angular frequency...
%
%     The transcribed version is version 1.7, 2000.07.28
%
%     Unlike the IDL function, the isis code returns a structure
%     containing the power spectral density (psd) and period and, if
%     specified by qualifiers, also the frequency (nu) and angular
%     frequency (om).
%
%\example
%     variable t = [0:100:0.1];
%     variable c = sin(t)+((rand(1000)/(2^32-1)*0.2)+0.9);
%     variable res = scargle(t, c; nu);
%     plot(res.period,res.psd);
%     plot(res.nu,res.psd);
%!%-
{
   
   % read in light curve given to the function
   variable whatyouget;
   variable i;
   variable m;
   variable t,c;
   switch(_NARGS)
     {case 2: (t,c)=();}
     {help(_function_name()); return;}
   
   % read qualifiers given, ordered like they are used in the code
   variable noise = qualifier("noise", sqrt(moment(c).var));
   
   % makes time managable (Scargle periodogram is time-shift invariant)
   
   variable time = t-t[0];
   
   % number of independent frequencies (Horne and Baliunas, eq. 13)
   variable n0 = length(time);
   variable horne = typecast((-6.362+1.193*n0+0.00098*n0^2.), Long_Type);
   
   % set numf or use default value if qualifier does not exist
   variable numf = int( qualifier("numf", ( horne<0 ? 5 : horne ) ) );

   % min. freq.  is 1/T
   variable fmin = double( qualifier("fmin", 1.0/qualifier("pmax", max(time) ) ) );
   
   % max. freq.: approx. to Nyquist frequency
   variable fmax = double( qualifier("fmax", 1.0/qualifier("pmin", (2.0*max(time))/n0 ) ) );

   % if fmin > fmax, calculations go crap
   if (fmin > fmax)
     {
	sprintf("Scargle Error: fmin > fmax! Check your inputs!");
     }
   
   
   % if omega is not given, compute it
   variable om;
   ifnot (qualifier_exists("omega"))
     {
	om = 2.*PI*(fmin+((fmax-fmin)*[0:numf-1]/(numf-1.)));
     }
   else
     {
	om = double(qualifier("omega"));
	numf = length(om);
     }

   % Periodogram
   variable px,nu,period;
   if (qualifier_exists("old"))
     {
	% subtract the mean from data
	variable cn = c-mean(c);
	
	% compute the periodogram
	px = Double_Type[numf];
	for (i=0; i<numf; i++)
	  {
	     variable tau = atan(sum(sin(2.*(om)[i]*time))/sum(cos(2.*(om)[i]*time)));
	     tau = tau/(2.*(om)[i]);
	     
	     variable co = cos(om[i]*(time-tau));
	     variable si = sin(om[i]*(time-tau));
	     
	     px[i] = 0.5*(sum(cn*co)^2/sum(co^2) + sum(cn*si)^2/sum(si^2));
	  }
	
	% correct normalization
	variable var = moment(cn).var;
	if (var!=0)
	  {
	     px=px/var;
	  }
	else
	  {
	     sprintf("Scargle Warning: Variance is zero");
	  }

	% some other nice helpers
	% computed here due to memory usage reasons

	nu = om/(2.*PI);
	period = 1./nu;

        whatyouget = struct { psd=px, period=period };
        if (qualifier_exists("nu")) { whatyouget = struct_combine( whatyouget, struct {nu=nu});}
        if (qualifier_exists("om")) { whatyouget = struct_combine( whatyouget, struct {om=om});}

        return whatyouget;
     }
   
   % Ref.: W.H. Press and G.B. Rybicki, 1989, ApJ 338, 277
   
   % Eq. (6); s2, c2
   variable s2 = Double_Type[numf];
   variable c2 = Double_Type[numf];
   
   for (i=0; i<numf; i++)
     {
	s2[i] = sum(sin(2.*om[i]*time));
	c2[i] = sum(cos(2.*om[i]*time));
     }

   % Eq. (2): Definition => tan(2omtau)
   % --- tan(2omtau) = s2/c2
   variable omtau = atan(s2/c2) / 2.;
  
   % cos(tau), sin(tau)
   variable cosomtau = cos(omtau);
   variable sinomtau = sin(omtau);
   
   % Eq. (7); sum(cos(t-tau)^2) and sum(sin(t-tau)^2)
   variable tmp = c2*cos(2.*omtau) + s2*sin(2.*omtau);
   variable tc2 = 0.5*(n0+tmp);      % sum(cos(t-tau)^2)
   variable ts2 = 0.5*(n0-tmp);      % sum(sin(t-tau)^2)
   
   % clean up
   tmp = 0.;
   omtau = 0.;
   s2 = 0.;
   c2 = 0.;
  
   % computing the periodogram for the original lc
   
   % subtract mean from data
   cn = c - mean(c);
   
   % Eq. (5); sh and ch
   variable sh = Double_Type[numf];
   variable ch = Double_Type[numf];
   
   for (i=0; i<numf; i++)
     {
	sh[i] = sum(cn*sin(om[i]*time));
	ch[i] = sum(cn*cos(om[i]*time));
     }
   
   % Eq. (3)
   px = (ch*cosomtau + sh*sinomtau)^2 / tc2
     + (sh*cosomtau - ch*sinomtau)^2 / ts2;
   
   % correct normalization
   px = 0.5*px/(noise^2);
   
   % some other nice helpers
   % computed here due to memory usage reasons

   nu = om/(2.*PI);
   period = 1./nu;

   whatyouget = struct { psd=px, period=period };
   if (qualifier_exists("nu")) { whatyouget = struct_combine( whatyouget, struct {nu=nu});}
   if (qualifier_exists("om")) { whatyouget = struct_combine( whatyouget, struct {om=om});}

   return whatyouget;
}

%
% Code from src/data/timing/segment_lc_for_psd.sl
%
define segment_lc_for_psd(lc, dt, dimseg) { %{{{
%!%+
%\function{segment_lc_for_psd}
%\synopsis{Function to segment a lightcurve in preparation for a PSD,
%    filtering out the segments which do not fit into multiples of
%    the segmentation length.}
%\usage{segment_lc_for_psd(lc,dt,dimseg)}
%\qualifiers{
%    \qualifier{gapfactor}{relative factor, telling how large the gap is relative
%                to dt [default: 1.1]}
%    \qualifier{ratefield}{Name of rate field in lightcurve [default: rate]}
%    \qualifier{verbose}{Increase verbosity to display which parts of the
%                lightcurve and what proportion was rejected [default=0].}
%}
%\description
%    Given a lightcurve with gaps, this function segments it such that
%    only intervals of length "dimseg" (in bins) are present. Data
%    that does not fit into integer multiples of the segmentation
%    length are cut off.
%    
%    ATTENTION: Never use a discrete Fourier transform (e.g., 
%    foucalc) with a segmentation length ("dimseg") larger than chosen
%    in this function!
%\notes
%    The time array is not altered, so the lightcurve will still
%    contain gaps, however none that have different length than
%    "dimseg". In other words: The output data arrays can begin at
%    arbitrary times, so are *not* sorted to always start at integer
%    multiples of the segmentation length.
%\example
%    dt=1.; % (s)
%    len=64.; % (s)
%    offset=10; % background level
%    T=5.; % (s), sinusoid periodidity
%    omega=2*PI/T; % Angular frequency (rad/s)
%    time_arr=[0:len:dt];
%    lc = struct{ time = time_arr,
%                 rate = offset+sin(omega*time_arr)+0.1*grand(int(len/dt)) };
%    lc_gaps=struct{
%             time=[lc.time[[0:20]],lc.time[[30:55]]],
%             rate=[lc.rate[[0:20]],lc.rate[[30:55]]] };
%    dimseg=16; % (bins)
%    lc_split = segment_lc_for_psd(lc_gaps, dt, dimseg);
%    res=foucalc(struct{time=lc_split.time,rate1=lc_split.rate},dimseg);
%\seealso{split_lc_at_gaps, foucalc}
%!%-

  variable gapfactor = qualifier("gapfactor", 1.1); 
  variable ratefield = qualifier("ratefield", "rate");
  variable verbose = qualifier("verbose", 0);

  variable split_lc = split_lc_at_gaps(lc, dt*gapfactor);

  variable segmented_lc = struct{
    time=Double_Type[0],
    rate=Double_Type[0],
    error=Double_Type[0],
    rejected_time, n_segments, total_time
  };
  
  %% keep a record of the total time and the time thrown away
  variable total_time=0.; % (s)
  variable rejected_time=0.;
  variable n_segments=0;

  variable ii;
  _for ii (0,length(split_lc)-1,1) {
    
    variable exposure=split_lc[ii].time[-1]-split_lc[ii].time[0];
    total_time += exposure;
    
    % get how often the segment fits into a given part of the split lightcurve
    variable nn = int(length(split_lc[ii].time)/dimseg);

    % only do this if there is at least one segment
    if (nn > 0) {
      % note that each segment goes from 0 to nn*dimseg-1
      % (the -1 is the important part!)
      n_segments++;
      segmented_lc.time=[segmented_lc.time,split_lc[ii].time[[0:nn*dimseg-1]]];      
      segmented_lc.rate=[segmented_lc.rate,  
			 get_struct_field(split_lc[ii],ratefield)[[0:nn*dimseg-1]]];
      if (struct_field_exists(split_lc[ii],"error")){
	segmented_lc.error=[segmented_lc.error, split_lc[ii].error[[0:nn*dimseg-1]]];
      }
    } else {
      rejected_time += exposure;
      if (verbose > 1){
	vmessage("  ***(%s): Skipping lightcurve segment with index %i, length %gs", _function_name(), ii, exposure);
      }
    }
  }

  if (verbose > 0) {
    vmessage("  ***(%s): Rejected time: %g s (%g percent of total data time=%gs)",
	     _function_name(), rejected_time, rejected_time/total_time*100, total_time);
    vmessage("  ***(%s): Number of segments with length>%gs (dimseg=%d, dt=%gs): %d",
	     _function_name(), dt*dimseg, dimseg, dt, n_segments);
  }
  
  segmented_lc.rejected_time = rejected_time;
  segmented_lc.total_time = total_time;
  segmented_lc.n_segments = n_segments;
  
  return segmented_lc; 
} %}}}
%
% Code from src/plot/setPGPLOTenv.sl
%
%%%%%%%%%%%%%%%%%%%
define setPGPLOTenv()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{setPGPLOTenv}
%\synopsis{sets environment variables used by PGPLOT}
%\usage{setPGPLOTenv(Double_Type w, h[, hoff[, voff]]);}
%\qualifiers{
%\qualifier{gif}{sets the PGPLOT_GIF enviroment variables}
%}
%\description
%    The environment variables for the PGPLOT postscript driver,
%    \code{PGPLOT_PS_{WIDTH,HEIGHT,HOFFSET,VOFFSET}}, are set
%    to the specified width \code{w} and height \code{h}. The default
%    horizontal and vertical offset is \code{hoff=0} and \code{voff=0}.
%    The parameters \code{w}, \code{h}, \code{hoff}, \code{voff} specify the size in cm,
%    while the enviroment variables are measured in milli-inches.
%    
%    If the \code{gif} qualifier is given, \code{PGPLOT_GIF_{WIDTH,HEIGHT}}
%    are set to \code{w} and \code{h} (in pixels).
%\seealso{putenv, http://www.astro.caltech.edu/~tjp/pgplot/devices.html}
%!%-
{
  variable w=NULL, h=NULL, hoff=NULL, voff=NULL;
  switch(_NARGS)
  { case 0: ; }
  { case 2: (w, h) = (); }
  { case 3: (w, h, hoff) = (); }
  { case 4: (w, h, hoff, voff) = (); }
  { message("usage: setPGPLOTenv([w, h[, hoff[, voff]]]);"); }

  if(qualifier_exists("gif"))
  {
    if(w==NULL)  w = 800;
    if(h==NULL)  h = 600;
    putenv(sprintf("PGPLOT_GIF_WIDTH=%.0f", w));
    putenv(sprintf("PGPLOT_GIF_HEIGHT=%.0f", h));
  }
  else
  {
    if(w==NULL)  w = 21;
    if(h==NULL)  h = 29.7;
    if(hoff==NULL)  hoff = 0;
    if(voff==NULL)  voff = 0;
    putenv(sprintf("PGPLOT_PS_WIDTH=%.0f", w/0.00254));
    putenv(sprintf("PGPLOT_PS_HEIGHT=%.0f", h/0.00254));
    putenv(sprintf("PGPLOT_PS_HOFFSET=%.0f", hoff/0.00254));
    putenv(sprintf("PGPLOT_PS_VOFFSET=%.0f", voff/0.00254));
  }
}

%
% Code from src/fitting/set_par_from_confmap.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define set_par_from_confmap_table()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{set_par_from_confmap_table}
%\synopsis{set parameters from a table obtained by get_confmap}
%\usage{set_par_from_confmap_results(Struct_Type table, Integer_Type i);}
%\description
%    The function sets the parameters of \code{table}'s row \code{i}.
%    As parameter names are infered from column names, the FITS
%    table may have to be read with the \code{casesen} qualifier.
%\seealso{get_confmap}
%!%-
{
  variable table, i;
  switch(_NARGS)
  { case 2: (table, i) = (); }
  { return help(_function_name()); }

  variable field;
  foreach field (get_struct_field_names(table)[[:-2]])
  {
    variable part = strtok(field, "_");
    variable par = "";
    while(not('0' <= part[0][-1] <= '9'))
    {
      par += part[0]+"_";
      part = part[[1:]];
    }
    variable m = string_matches(part[0], `\(.*\)\(\d+\)`);
    par = sprintf("%s%s(%s).%s", par, m[1], m[2], part[1]);
    variable val = get_struct_field(table, field)[i];
    vmessage(`set_par("%s", %S);`, par, val);
    try { set_par(par, val); }
    catch AnyError: ;
  }
}

%
% Code from src/slang/arrays/2d/shift_2d.sl
%
define shift_2d()
{
  variable a, dy, dx;
  switch(_NARGS)
  { case 3: (a, dy, dx) = (); }
  { help(_function_name()); return; }

  variable n = array_shape(a);  % n[0] = ny,  n[1] = nx;
  dx = -dx mod n[1];  if(dx<0)  dx += n[1];
  dy = -dy mod n[0];  if(dy<0)  dy += n[0];
  variable N = n[0] * n[1];
  variable i = [ 0 : N-1 ];
  a = a[ (i + (i+dx) mod n[1] - i mod n[1] + dy*n[1]) mod N ];
  reshape(a, n);
  return a;
}

define shift_2d_old()
{
  variable a, dy, dx;
  switch(_NARGS)
  { case 3: (a, dy, dx) = (); }
  { help(_function_name()); return; }

  variable ny = length(a[*,0]);
  variable nx = length(a[0,*]);
  variable a_ = typeof(a[0,0])[ny, nx];
  variable x, y;
  _for y (0, ny-1, 1)  a_[y,*] = shift(a [y,*], dx);
  _for x (0, nx-1, 1)  a_[*,x] = shift(a_[*,x], dy);
  return a_;
}

%
% Code from src/slang/structs/histograms/shift_hist.sl
%
define shift_hist(hist, shft){
%!%+
%\function{shift_hist}
%\synopsis{shift a histogram}
%\usage{Struct_Type hist = shift_hist(hist, shift);}
%\description
%    Shift the grid of a histogram structure with fields bin_lo, 
%    bin_hi, value, and err. The bin_lo and bin_hi fields are shifted
%    by the amount shift. If other fields are present,
%    those are preserved and passed on. Only presence of bin_lo and 
%    bin_hi are checked for. 
%\seealso{add_hist, scale_hist, stretch_hist}
%!%-   
   if(_NARGS==0) {  help(_function_name()); return; }
   
   variable ns = struct_copy(hist);
   ifnot(struct_field_exists(ns, "bin_lo") and 
		 struct_field_exists(ns, "bin_hi")){
	  help(_function_name()); return;
   }
   
   ns.bin_lo += shft;
   ns.bin_hi += shft;
   
   return ns;
}

%
% Code from src/slang/misc/show_slang_code.sl
%
%%%%%%%%%%%%%%%%%%%%%%
define show_slang_code()
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{show_slang_code}
%\usage{show_slang_code(String_Type function);}
%\description
%    The function \code{show_slang_code} reads all \code{.sl} files
%    in the directories contained in the S-Lang load path,
%    and *tries* to find the definition of \code{function}
%    by parsing the code for {}-brackets and comments.
%    In its current version, \code{show_slang_code} gets confused, e.g.,
%    from {}-brackets and % characters in strings.
%!%-
{
  variable function;
  switch(_NARGS)
  { case 1: function = (); }
  { help(_function_name()); return; }

  variable show_slang_code_code = "";  % could be made a private global variable
  variable show_slang_code_len = 0;    % could be made a private global variable

  if(show_slang_code_code=="")
  {
    variable dir, slfile;
    foreach dir (strtok(get_slang_load_path(), ":"))
      foreach slfile (glob(dir+"/*.sl"))
      { % message(slfile);
	variable F = fopen(slfile, "r");
	if (typeof(F) == File_Type) {
          show_slang_code_code += strjoin(fgetslines(F), "");
          ()=fclose(F);
	}
      }
    show_slang_code_len = strlen(show_slang_code_code);
  }
  variable code = show_slang_code_code;
  variable len =  show_slang_code_len;

  variable startpos = 0;
  forever
  {
    startpos = string_match(code, "define +"+function+" *\(", startpos+1);
    if(startpos==0)  break;
    variable pos=startpos;
    variable comment = 0;
    while(pos<len)
    { if(code[pos]=='%')  comment = 1;
      if(code[pos]=='\n') comment = 0;
      if(code[pos]=='{' and not comment)  break;
      pos++;
    }
    variable brackets = 1;
    pos++;
    while(brackets>0 && pos<len)
    { if(code[pos]=='%')  comment = 1;
      if(code[pos]=='\n') comment = 0;
      if(code[pos]=='{' && not comment)  brackets++;
      if(code[pos]=='}' && not comment)  brackets--;
      pos++;
    }
    if(pos<len)
      message("\n"+code[[startpos-1:pos-1]]+"\n");
  }
}

%
% Code from src/misc/image/simbad2ds9.sl
%
%% This function intends on converting a SIMBAD cone search list into
%% a DS9 region file. The input is an ASCII file.
%%
%% Author: Ole Koenig (ole.koenig@fau.de)
%% Date:   2019-10-17


public define simbad2ds9(){
%!%+
%\function{simbad2ds9}
%\synopsis{Converts a SIMBAD cone search ASCII file into a DS9 region
%    file (default outfile: "ds9.reg")}
%\qualifiers{
%\qualifier{radius}{circle radius (arcsec), default: 20arcsec}
%\qualifier{type}{regex to filter for source type, default ""}
%}
%\notes
%    See http://simbad.u-strasbg.fr/simbad/sim-display?data=otypes
%    for a list of type abbreviations
%\example
%    simbad2ds9("simbadascii.txt","output.reg" ; radius=10, type="HXB");
%\usage{simbad2ds9(String_Type filename, [String_Type outfile]);}
%!%-

  variable filename;
  variable outfile="ds9.reg";
  switch(_NARGS)
  { case 1: (filename) = (); }
  { case 2: (filename, outfile) = (); }
  { help(_function_name()); return; }

  %% Test whether file exists
  if ( system(sprintf("[ -f %s ]", filename)) != 0 ){
    vmessage("Input file does not exist");
  }
  
  %% Radius with which the circles are drawn (ideally this should be
  %% the 90% error radius of each measurement)
  variable radius = qualifier("radius", 20.0); % (arcsec)
  %% A regular expression to filter the list for specific type
  %% E.g. filter for HMX via "HMX"
  %% See http://simbad.u-strasbg.fr/simbad/sim-display?data=otypes
  variable type = qualifier("type", ""); % Reg. Expression
  
  variable fpIn = fopen(filename,"r");
  variable data = fgetslines(fpIn);

  variable fpOut = fopen(outfile,"w");
  %% Write header of DS9 region file
  ()=fputs("# Region file format: DS9 version 4.1\n",fpOut);
  ()=fputs("global color=green dashlist=8 3 width=1 font=\"helvetica 10 normal roman\" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1\n",fpOut);
  ()=fputs("fk5\n",fpOut);
 
  %% Define regular expression to format string
  % First four entries: # |   dist|identifier   |type
  variable p1="\(\d\) *| *\(\d+.\d+\)|\(.+ .+\)|\(.+\)"R;
  % Coordinates: |ra dec   |
  variable p2="|\(\d\{2\} \d\{2\} \d\{2\}\.\d+\) \(-?\d\{2\} \d\{2\} \d\{2\}\.?\d*\)"R;
  
  variable line;
  % Skip first 9 and last 3 lines
  _for line (0,length(data)-1,1) {
    variable query = string_matches(data[line], p1+p2) ;
    if (query==NULL) { continue; }
    
    variable out = struct {
      raw        = query[0],
      idx        = query[1],
      dist       = atof(query[2]),
      identifier = strtrim(query[3]),
      type       = query[4],
      %% Format R.A. and Dec. from "18 58 25.216" to "18:48:25.216"
      ra         = sprintf("%s:%s:%s", __push_array(strsplit(query[5]," "))),
      dec        = sprintf("%s:%s:%s", __push_array(strsplit(query[6]," ")))
    };

    if ( string_matches(out.type, sprintf(".*%s.*"R, type) )!=NULL ) {
      ()=fputs(sprintf("circle(%s,%s,%.2f\") # text={%s, %s}\n",
		       out.ra, out.dec, radius, out.identifier, out.type) ,fpOut);
    }
  }

  ()=fclose(fpIn);
  ()=fclose(fpOut);

}
%
% Code from src/fitting/fit-functions/simple_gpile/simple_gpile.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile_fit(lo,hi,par,fun)
%%%%%%%%%%%%%%%%%%%%%%%
{
   % October 25, 2005 - New and improved functionality, especially in
   % the dithered regions of the chips!

   % Peak pileup correction goes as:
   %    exp(log(1-pfrac)*[counts/max(counts)])
   variable pfrac = par[0];

   % Pileup scales with model counts from *data set* indx ...
   variable indx = typecast(par[1],Integer_Type);

   if( indx == 0 or pfrac == 0. )
      return fun;   % Quick escape for no changes ...

   % ... but the arf index could be a different number, so get that
   variable arf_indx = get_data_info(indx).arfs;

   % Get arf information
   variable arf = get_arf(arf_indx[0]);

   % In dither regions (or bad pixel areas), counts are down not
   % from lack of area, but lack of exposure.  Pileup fraction
   % therefore should scale with count rate assuming full exposure.
   % Use the arf "fracexpo" column to correct for this effect
   variable fracexpo = get_arf_info(arf_indx[0]).fracexpo;
   fracexpo[where(fracexpo ==0)] = 1.;

   % Rebin arf to input grid, correct for fractional exposure, and
   % multiply by "fun" to get ("corrected") model counts per bin
   variable mod_cts;
   mod_cts = fun * rebin(lo,hi,arf.bin_lo, arf.bin_hi,arf.value/fracexpo);

   % Go from cts/bin/s -> cts/angstrom/s
   mod_cts = mod_cts/(hi-lo);

   % Use 2nd and 3rd order arfs to include their contribution.
   % Will probably work best if one chooses a user grid that extends
   % from 1/3 of the minimum wavelength to the maximum, and has
   % at least 3 times the resolution of the first order grid.
   variable mod_ord;
   if(par[2] > 0)
   {
      indx = typecast(par[2],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      fracexpo[where(fracexpo ==0)] = 1.;
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,2*arf.bin_lo,2*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }
   if(par[3] > 0)
   {
      indx = typecast(par[3],Integer_Type);
      arf = get_arf(indx);
      fracexpo=get_arf_info(indx).fracexpo;
      fracexpo[where(fracexpo ==0)] = 1.;
      mod_ord = arf.value/fracexpo*
                rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
      mod_ord = rebin(lo,hi,3*arf.bin_lo,3*arf.bin_hi,mod_ord)/(hi-lo);
      mod_cts = mod_cts+mod_ord;
   }

   % In case something funky happens while wandering parameter space ...
   variable max_mod = max(mod_cts);

   if(max_mod <= 0){ max_mod = 1.;}

   % Scale maximum model counts to pfrac
   mod_cts = log(1-pfrac)*mod_cts/max_mod;

   % Return function multiplied by exponential decrease
   fun = exp(mod_cts) * fun;
   return fun;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile_defaults(i)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
   switch(i)
   {case 0:
      return (0.1,1,0.,0.5);
   }
   {case 1:
      return (1,1,0,100);
   }
   {case 2:
      return (2,1,0,100);
   }
   {case 3:
      return (3,1,0,100);
   }
}

add_slang_function("simple_gpile", ["pile_frac","data_indx","arf2_indx","arf3_indx"]);

set_function_category("simple_gpile", ISIS_FUN_OPERATOR);

set_param_default_hook("simple_gpile", "simple_gpile_defaults");

%
% Code from src/fitting/fit-functions/simple_gpile/simple_gpile2.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile2_fit(lo, hi, par, fun)
%%%%%%%%%%%%%%%%%%%%%%%%
{
  % 2007, August 14 - fracexpo does not have to be an array

  % 2007, May 03 - correct rebinning of the arf

  % 2007, January 22 - no explicit refering to max(mod_cts)

  % 2005, October 25 - New and improved functionality, especially
  %                    in the dithered regions of the chips!

  % Peak pileup correction goes as:
  %    exp(log(1-pfrac)*[counts/max(counts)])
  %  = exp(- beta * counts )
  variable beta = par[0];

  % Pileup scales with model counts from *data set* indx
  variable indx = typecast(par[1], Integer_Type);

  if( indx == 0 or beta == 0. )
    return fun;  % Quick escape for no changes ...

  % The arf index could be a different number, so get that
  variable arf_indx = get_data_info(indx).arfs;

  % Get arf information
  variable arf = get_arf(arf_indx[0]);

  % In dither regions (or bad pixel areas), counts are down not
  % from lack of area, but lack of exposure.  Pileup fraction
  % therefore should scale with count rate assuming full exposure.
  % Use the arf "fracexpo" column to correct for this effect
  variable fracexpo = get_arf_info(arf_indx[0]).fracexpo;
  if(length(fracexpo)>1)
    fracexpo[where(fracexpo==0)] = 1.;
  else
  { if(fracexpo==0)  fracexpo = 1; }

  % Rebin arf to input grid, correct for fractional exposure, and
  % multiply by "fun" to get ("corrected") model counts per bin
  variable mod_cts_int;
  mod_cts_int = fun * rebin(lo, hi,
                            arf.bin_lo, arf.bin_hi,
                            arf.value*(arf.bin_hi-arf.bin_lo)/fracexpo)
                                     / (hi-lo);
  % Go from bin-integrated(ph/cm^2/s) * bin-integrated(cm^2)
  % to cts/s/angstrom
  variable mod_cts = mod_cts_int/(hi-lo);

  % Use 2nd and 3rd order arfs to include their contribution.
  % Will probably work best if one chooses a user grid that extends
  % from 1/3 of the minimum wavelength to the maximum, and has
  % at least 3 times the resolution of the first order grid.
  variable mod_ord;
  if(par[2] > 0)
  {
    indx = typecast(par[2], Integer_Type);
    arf = get_arf(indx);
    fracexpo=get_arf_info(indx).fracexpo;
    if(length(fracexpo)>1)
    { fracexpo[where(fracexpo==0)] = 1.; }
    else
    { if(fracexpo==0) { fracexpo = 1; } }
    mod_ord = arf.value/fracexpo*
              rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
    mod_ord = rebin(lo,hi,2*arf.bin_lo,2*arf.bin_hi,mod_ord)/(hi-lo);
    mod_cts = mod_cts+mod_ord;
  }
  if(par[3] > 0)
  {
    indx = typecast(par[3], Integer_Type);
    arf = get_arf(indx);
    fracexpo=get_arf_info(indx).fracexpo;
    if(length(fracexpo)>1)
    { fracexpo[where(fracexpo==0)] = 1.; }
    else
    { if(fracexpo==0) { fracexpo = 1.; } }
    mod_ord = arf.value/fracexpo*
              rebin(arf.bin_lo,arf.bin_hi,lo,hi,fun);
    mod_ord = rebin(lo,hi,3*arf.bin_lo,3*arf.bin_hi,mod_ord)/(hi-lo);
    mod_cts = mod_cts+mod_ord;
  }

  % Return function multiplied by exponential decrease
  return exp(-beta*mod_cts) * fun;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile2_defaults(i)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
   switch(i)
   {case 0:
      return (0.05,1,0,10);
   }
   {case 1:
      return (0, 1, 0, 100);
   }
   {case 2:
      return (0, 1, 0, 100);
   }
   {case 3:
      return (0, 1, 0, 100);
   }
}

add_slang_function("simple_gpile2", ["beta [s*A/cts]", "data_indx", "arf2_indx", "arf3_indx"]);
set_function_category("simple_gpile2", ISIS_FUN_OPERATOR);
set_param_default_hook("simple_gpile2", "simple_gpile2_defaults");

%
% Code from src/fitting/fit-functions/simple_gpile/simple_gpile3.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile3_fit(lo, hi, par, fun)
%%%%%%%%%%%%%%%%%%%%%%%%
{
  % 2011-02    - parameterize beta relative to  beta0 = 3 * Dlambda * Tframe
  % 2007-08-14 - fracexpo does not have to be an array
  % 2007-05-03 - correct rebinning of the arf
  % 2007-01-22 - no explicit refering to max(mod_cts)
  % 2005-10-25 - New and improved functionality, especially
  %              in the dithered regions of the chips!

  variable beta = par[0];
  if(beta == 0.)
    return fun;  % quick escape; nothing to be changed

  variable info = get_data_info(Isis_Active_Dataset);

  % get ARF index (not necessarily the same as Isis_Active_Dataset)
  variable arf_indx = info.arfs;

  % get ARF data
  variable arf = get_arf(arf_indx[0]);

  % In dither regions (or bad pixel areas), counts are down not
  % from lack of area, but lack of exposure.  Pileup fraction
  % therefore should scale with count rate assuming full exposure.
  % Use the arf "fracexpo" column to correct for this effect
  variable fracexpo = get_arf_info(arf_indx[0]).fracexpo;
  if(length(fracexpo)>1)
    fracexpo[where(fracexpo==0)] = 1.;
  else
    if(fracexpo==0)  fracexpo = 1.;

  % get ("corrected") model counts spectrum
  variable mod_cts
    = fun * rebin(lo, hi,
	          arf.bin_lo, arf.bin_hi,
	          arf.value/fracexpo*(arf.bin_hi-arf.bin_lo)   % = "bin-integrated" ARF (cm^2*A), needed for rebinning
	         )                  /sqr(     hi-        lo);  % one factor (hi-lo) to get again a bin-averaged ARF
                                                               % second to go from  bin-integrated(ph/cm^2/s) * bin-average(cm^2)
                                                               %                to  bin-density(cts/s/A)

  % Use 2nd and 3rd order arfs to include their contribution.
  % Will probably work best if one chooses a user grid that extends
  % from 1/3 of the minimum wavelength to the maximum, and has
  % at least 3 times the resolution of the first order grid.
  variable indx, mod_ord;
  if(par[1] > 0)
  {
    indx = int(par[1]);
    arf = get_arf(indx);
    fracexpo = get_arf_info(indx).fracexpo;
    if(length(fracexpo)>1)
      fracexpo[where(fracexpo==0)] = 1.;
    else
      if(fracexpo==0)  fracexpo = 1.;
    mod_ord = arf.value/fracexpo * rebin(arf.bin_lo, arf.bin_hi,  lo, hi,  fun);
    mod_cts += rebin(lo, hi,
		     2*arf.bin_lo, 2*arf.bin_hi,
		     mod_ord
		    )/(hi-lo);
  }
  if(par[2] > 0)
  {
    indx = int(par[2]);
    arf = get_arf(indx);
    fracexpo = get_arf_info(indx).fracexpo;
    if(length(fracexpo)>1)
      fracexpo[where(fracexpo==0)] = 1.;
    else
      if(fracexpo==0)  fracexpo = 1.;
    mod_ord = arf.value/fracexpo * rebin(arf.bin_lo, arf.bin_hi,  lo, hi,  fun);
    mod_cts += rebin(lo, hi,
		     3*arf.bin_lo, 3*arf.bin_hi,
		     mod_ord
		    )/(hi-lo);
  }

  % Peak pileup correction goes as:
  %    exp( - beta * [spectral count rate] )
  % where [spectral count rate] in given in units of cts/s/A
  % and    beta  ~  beta0 = 3 * Dlambda * Tframe
  %     Dlambda  =   5.5 mA   for HEG  (info.part == 1)
  %              =  11   mA   for MEG  (info.part == 2)
  if(1 <= info.part <= 2)  % for Chandra HETGS spectrum
    beta *= 3 * 0.0055*info.part * info.frame_time;
  else
    vmessage("warning: get_data_info(%d).part==%d\n =>  beta is in units of [s*A/cts] and not [beta0 = 3*Dlambda * Tframe",
	     Isis_Active_Dataset, info.part);
  return exp(-beta * mod_cts) * fun;

  % References: Nowak et al. (2008), ApJ, 689 : 1199-1214
  %             Hanke et al. (2009), ApJ, 690 : 330-346
}

add_slang_function("simple_gpile3", ["beta [beta0]", "arf2_indx", "arf3_indx"]);
set_function_category("simple_gpile3", ISIS_FUN_OPERATOR);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define simple_gpile3_defaults(i)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
{
   switch(i)
   { case 0: return (1, 1, 0,  10); }
   { case 1: return (0, 1, 0, 100); }
   { case 2: return (0, 1, 0, 100); }
}

set_param_default_hook("simple_gpile3", "simple_gpile3_defaults");

%
% Code from src/slang/geometry/simplify_polygon.sl
%
define douglas_peucker(x,y,maxdist2);
define douglas_peucker(x,y,maxdist2) {
%!%+
%\function{douglas_peucker}
%\synopsis{Simplify polygons using the Douglas Peucker Algorithm}
%\usage{(xx,yy)=douglas_peucker(x,y,d2);}
%\description
%   This function implements an algorithm to simplify complex
%   polygons (Douglas & Peucker, 1973, Cartographica 10(2), 112).
%   A recursive version presented by Hershberger & Snoeyink (1992,
%   Proc. 5th Intl. Symp. on Spatial Data Handling, 134-143) is
%   used.
%   For a polygon P defined by positions in given by the arrays (x,y),
%   the algorithm returns the polygon P2 defined by coordinates (xx,yy) 
%   with the property that the maximum distance between the lines segments
%   of P and P2 is smaller than sqrt(d2). 
%   Note 1:
%   The algorithm only yields a good result if P is not self-intersecting.
%   Note 2:
%   As discussed by Hershberger & Snoeyink, the worst performance of
%   the Douglas-Peucker-algorithm is O(N^2). Hershberger & Snoeyink (1998,
%   Computational Geometry 11(3-4), 175-185) present a O(N log N) algorithm.
%   Unfortunately for me (J. Wilms), this algorithm is too complex to be
%   implementable on the ICE between Hamburg and Bamberg...
%   Note 3:
%   This function is called recursively and does not perform sanity checks on
%   x,y,d2. If you do not know about the properties of your data, use
%   simplify_polygon.
%   
%\seealso{simplify_polygon}
%!%-
    variable npt=length(x);
    %
    % segment cannot be simplified
    %
    if (npt<=2) {
	return x,y;
    }
    
    %
    % find point with maximum distance from the polygon
    %
    variable i;
    variable isplit=0;
    variable maxdi2=0.;

    _for i(1,npt-2,1) {
	variable dist2=point_distance2_from_line(x[i],y[i],x[0],y[0],x[npt-1],y[npt-1]);
	if (dist2>maxdi2) {
	    maxdi2=dist2;
	    isplit=i;
	}
    }

    if (maxdi2>=maxdist2) {
	%
	% recursively simplify polygon further
	%
	variable xx1,yy1,xx2,yy2;
	(xx1,yy1)=douglas_peucker(x[[0:isplit]],y[[0:isplit]],maxdist2);
	(xx2,yy2)=douglas_peucker(x[[isplit:npt-1]],y[[isplit:npt-1]],maxdist2);

	return ([xx1,xx2[[1:]]] , [yy1,yy2[[1:]]] );
    }

    return([x[0],x[npt-1]],[y[0],y[npt-1]]);
}

define simplify_polygon() 
%!%+
%\function{simplify_polygon}
%\synopsis{Simplify polygons using the Douglas Peucker Algorithm}
%\usage{(xx,yy)=simplify_polygon(x,y,d);}
%\description
% Use the Douglas Peucker Algorithm to simplify the polygon P
% defined by the positions in the arrays x,y such that the
% maximum distance between all segments  of the resulting
% polygon P2 defined by (xx,yy) is smaller than d.
% See the help for the function douglas_peucker for caveats
% Note that the argument d in simplify_polyon defines the
% distance, while the corresponding argument in douglas_peucker
% defines the distance squared!
%
% \seealso{douglas_peucker}
%!%-
{

    if (_NARGS != 3 ) {
	help(_function_name());
	return;
    }

    variable x,y,maxdist;
    (x,y,maxdist)=();
    
    if (length(maxdist)!=1) {
	throw UsageError,sprintf("%s: maxdist must be a scalar\n",_function_name());
    }

    if (maxdist<=0) {
	throw UsageError, sprintf("%s: maxdist must be positive\n",_function_name());
    }

    if (length(x)!=length(y)) {
	throw UsageError, sprintf("%s: x, y must have the same length\n",_function_name());
    }
    
    if (length(x)<2) {
	throw UsageError, sprintf("%s: x, y must include at least two points\n",_function_name());
    }

    douglas_peucker(x,y,maxdist*maxdist);
}

%
% Code from src/sixte/simput_athenacrab.sl
%
define simput_athenacrab(flux){   
%!%+
%\function{simput_athenacrab}
%\synopsis{returns command to create a standard Athena Crab Simput File}
%\usage{String_Type cmd = simput_athenacrab(Double_Type flux);}
%\description
%    Flux has to be given in Crab. The naming convention is that the
%    flux (in micro Crab = 1e-6 Crab or in erg/cm^2/s if unit=cgs) is
%    encoded in the filename. The spectrum has an absorption of 4E21
%    cm^-2, norm of 9.5 ph/kev/cm^2/s at 1keV, and powerlaw index of
%    2.1. The source position is RA=Dec=0.
%\qualifiers{
%\qualifier{unit}{crab or cgs, default: crab}
%\qualifier{dir}{To append a full directory path, default: cwd}
%\qualifier{logEgrid}{use a logarithmic energy grid (from Elow to
%            Eup with Nbins), default: no}
%\qualifier{Nbins}{number of energy bins created from Elow to Eup,
%            default: 1000}
%}
%!%-
  variable unit = qualifier("unit", "crab");

  variable crabnorm = 2.0521231034620645e-08; % To get a flux of 9.5ph/kev/cm^2/s at 1keV
  
  variable srcname, plFlux;
  if (unit == "cgs"){
    srcname = sprintf("athenacrab_flux%ecgs.simput", flux);
    plFlux = sprintf("plFlux=%e", flux);
  } else {
    srcname = sprintf("athenacrab_flux%07imuCrab.simput", nint(1e6*(flux)));
    plFlux = sprintf("plFlux=%e", flux*crabnorm);
  }
  srcname = qualifier("dir", "") + srcname;
  
  variable energy_string = sprintf("logEgrid=%s Nbins=%i", 
				   qualifier("logEgrid", "no"), qualifier("Nbins", 1000));

  variable simputcom = "simputfile " + 
    "simput=" + srcname + " RA=0.0 Dec=0.0 " + 
    "Emin=2. Emax=10. plPhoIndex=2.1 NH=0.4 " +
    energy_string + " " + plFlux + 
    " clobber=yes";
  
  return simputcom;
}

%
% Code from src/sixte/simputfile.sl
%
% -*- mode: slang; mode: fold -*-

private variable fluxInCrab = 2.4e-8; % for 2--10 keV
private define get_simputfile_empty_struct(){ %{{{

   variable arr = ["Simput","Src_ID","Src_Name","RA","Dec",
		   "srcFlux","Elow","Eup","Estep",
		   "Nbins","logEgrid",
		   "plPhoIndex","plFlux","bbkT","bbFlux",
		   "flSigma","flFlux","rflSpin","rflFlux",
		   "NH","Emin","Emax",
		   "ISISFile","ISISPrep","XSPECFile","PHAFile",
		   "ASCIIFile","LCFile","MJDREF",
		   "PSDnpt","PSDfmin","PSDfmax","LFQ","LFrms",
		   "HBOf","HBOQ","HBOrms","Q1f","Q1Q","Q1rms",
		   "Q2f","Q2Q","Q2rms","Q3f","Q3Q","Q3rms","PSDFile",
		   "ImageFile","chatter","clobber","history"];

   return @Struct_Type(arr);
}
%}}}
private define set_simputfile_flux(){ %{{{

   variable emin = 2.0, emax = 10.;
   variable str,flux;
   switch(_NARGS)
   {case 2: (str,flux) = (); }
   {case 4: (str,flux,emin,emax) = (); }
   { help(_function_name()); return; }

   str.Emin=emin;
   str.Emax=emax;

   
   if (qualifier_exists("crab"))
     flux *= fluxInCrab;

   str.srcFlux=flux;

   return;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%
define get_simputfile_struct(Simput,RA,Dec,flux){ %{{{
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_simputfile_struct}
%\synopsis{get a very basic SIMPUT structure for further editing}
%\usage{get_simputfile_struct(filename, RA, Dec flux);}
%\description
%    Note: All parameters and there usage are according to the simputfile
%    routine of the SIMPUT package. E.g., flux is given in erg/cm^2/s.
%\qualifiers{
%\qualifier{crab}{flux is given in units of Crab}
%\qualifier{emin}{[2.0]: lower limit of the energy band of the given flux in keV}
%\qualifier{emax}{[10.0]: lower limit of the energy band of the given flux in keV}
%}
%\seealso{create_basic_simputfile,eval_simputfile,set_simputfile_model_grid,set_simputfile_flux}
%!%-
   variable str = get_simputfile_empty_struct();
   str.Simput=Simput;
   str.RA=RA;
   str.Dec=Dec;
   
   if (qualifier_exists("emin") && qualifier_exists("emax"))
     set_simputfile_flux(str,flux,qualifier("emin"),qualifier("emax") ;; __qualifiers);
   else
     set_simputfile_flux(str,flux ;; __qualifiers);

   if (not qualifier_exists("noclobber"))
     str.clobber="yes"; 
   
   return str;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%
define set_simputfile_model_grid(){ %{{{
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{set_simputfile_model_grid}
%\synopsis{set the model grid in the SIMPUT structure}
%\usage{set_simputfile_model_grid(Struct_Type str);}
%%\altusage{set_simputfile_model_grid(Struct_Type str, Elow, Eup, Estep);}
%\description
%    All energies are given in keV. By default, the spectrum is
%    evaluated from 0.05-24.0 keV in steps of 0.01 keV.
%\seealso{create_basic_simputfile,get_simputfile_struct,eval_simputfile,set_simputfile_flux}
%!%-

   variable str;
   variable emin=0.05,emax=25,estep=0.01;
   
   switch(_NARGS)
   {case 1: str = (); }
   {case 4: (str,emin,emax,estep) = (); }
   { help(_function_name()); return; }

   str.Elow=emin;
   str.Eup=emax;
   str.Estep=estep;
   
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%
define eval_simputfile(str){ 
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{eval_simputfile}
%\synopsis{evaluates the SIMPUT structure}
%\usage{Integer_Type sucess = eval_simputfile(Struct_Type str);}
%\description
%    This function evalutes the SIMPUT structure, created for example
%    with "get_simputfile_struct". Generally, all fields of the
%    structure == NULL are skipped.
%\qualifiers{
%\qualifier{quiet}{don't show any output}
%}
%\seealso{create_basic_simputfile,get_simputfile_struct,set_simputfile_model_grid,set_simputfile_flux}
%!%-

   variable cmd = "simputfile";

   variable sf,sfv;
   foreach sf(get_struct_field_names(str)){

      sfv = get_struct_field(str,sf);
      if (sfv !=NULL){
	 if (not (typeof(sfv) == String_Type))
	   sfv = string(sfv);
	 
	 cmd += sprintf(" %s=%s",sf,sfv);
      }
      
   }

   % for backwards compatibility...
   if (not (qualifier_exists("quite"))) {
       if (not (qualifier_exists("quiet"))) {
	   message("\n"+cmd);
       }
   }
   
   return system(cmd);
}

%%%%%%%%%%%%%%%%%%%%%%%
define create_basic_simputfile(Simput,RA,Dec,flux,nh,pl){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{create_basic_simputfile}
%\synopsis{creates a basic SIMPUT file}
%\usage{Struct_Type str = create_basic_simputfile(filename, RA, Dec, srcFlux, nH, powerLawIndex);}
%\description
%    This function creates a structure, which contains all necessary
%    info for a simple "simputfile" command to create a basic SIMPUT
%    file for a given point source. The flux is assumed to be given in
%    the 2-10 keV band by default. The file can be directly created by
%    adding the qualifier "eval". Otherwise a structure is returned.
%    This structure can be further modifed, and then evaluated with
%    the function "eval_simputfile".
%\qualifiers{
%\qualifier{crab}{flux is given in units of Crab}
%\qualifier{eval}{directly evalute the returned structure (similar to call eval_simputfile(str); }
%\qualifier{emin}{lower limit of the energy band of the given flux in keV}
%\qualifier{emax}{lower limit of the energy band of the given flux in keV}
%\qualifier{quite}{don't show any output}
%}
%\examples
%    \code{variable str = create_basic_simputfile("velaX-1.fits",135.528583,40.554722,100e-3,1.8,5.0;crab);}
%\seealso{eval_simputfile,get_simputfile_struct,set_simputfile_model_grid,set_simputfile_flux}
%!%-

   if (qualifier_exists("crab"))
     flux *= fluxInCrab;

   variable str = get_simputfile_struct(Simput,RA,Dec,flux);

   str.NH=nh;
   str.plPhoIndex=pl;
   str.plFlux = flux;
   
   set_simputfile_model_grid(str ;; __qualifiers);

   if(qualifier_exists("eval"))
     () = eval_simputfile(str);
   
   return str;
}

%
% Code from src/data/spectrum.loaddata/simulate_data.sl
%
%%%%%%%%%%%%%%%%%%%%
define simulate_data()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{simulate_data}
%\synopsis{fakes a spectrum}
%\usage{Integer_Type id = simulate_data(String_Type RSPfile, Double_Type exposure);}
%\description
%    The currently defined fit-function is used to fake the spectrum.
%\seealso{fakeit}
%!%-
{
  variable RMFfile, exposure;
  switch(_NARGS)
  { case 2: (RMFfile, exposure) = (); }
  { help(_function_name()); return; }

  variable data_index;
  if(all_data==NULL) { data_index = 1; } else { data_index = max(all_data)+1; }
  variable r = load_rmf(RMFfile);
  variable a = factor_rsp(r);
  assign_arf(a, data_index);
  assign_rmf(r, data_index);
  set_arf_exposure(a, exposure);
  fakeit;
  return data_index;
}

%
% Code from src/data/simultaneous_fit.sl
%
% -*- mode: slang; mode: fold; -*-

define data_map_function()
%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{data_map_function}
%\synopsis{executes a function on specific datasets}
%\usage{data_map_function(String_Type filter, Ref_Type function, arguments...; qualifiers);}
%\description
%    Loops over all defined datasets and matches each data
%    information against the given filter. In the simplest
%    way, this filter is a string respresenting the wanted
%    instrument as specified in the field of 'get_data_info'.
%    If the 'fi' qualifier is provided, the given filter is
%    interpreted as if-statement, which has to be fullfiled.
%    Thereby, the associated data information can be accessed
%    via the 'info' variable. The given function is applied
%    to each matching dataset in the form
%      function(dataset, arguments...; qualifiers);
%\example
%    % set the energy ranges for all RXTE-PCA sepctra
%    data_map_function("PCA", &xnotice_en, 3.5, 50);
%
%    % apply grouping to all Suzaku-PIN spectra
%    data_map_function(`is_substr(info.file, "hxd_pin")`,
%      &group; min_sn = 40, fi);
%\seealso{get_data_info, eval, array_map}
%!%-
{
  if (_NARGS < 2) { help(_function_name); return; }
  variable dtctr, funref, args = __pop_args(_NARGS - 2);
  variable ifstat = qualifier_exists("fi");
  (dtctr, funref) = ();

  % loop over all data
  variable d;
  _for d (0, length(all_data)-1, 1) {
    variable info = get_data_info(all_data[d]);
    if (ifstat) eval(sprintf("variable info = get_data_info(all_data[%d]);", d));
    if (info.instrument == dtctr || (ifstat && eval(dtctr))) (@funref)(all_data[d], __push_args(args) ;; __qualifiers);
  }
} %}}}

%%%%%%%%%%%%%%%%%%%%%
define simfit_namespace()
%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+
%\function{simfit_namespace}
%\synopsis{implements all SimFit-functions into a namespace}
%\usage{simfit_namespace(Struct_Type SimFit);}
%\qualifiers{
%    \qualifier{name}{name of the new namespace (default: simfit)}
%    \qualifier{chatty}{chattiness of this function (default: 1)}
%}
%\description
%    Takes a simultaneous fit structure as input and
%    implements all available functions defined in there
%    into a new namespace. In this way the user no longer
%    has to type the structure's name in front of the
%    functions, but use the functions directly.
%
%    Note that in this namespace the ISIS intrinsic
%    functions, such as 'fit_fun', 'set_par', etc. are
%    being overwritten by the SimFit-versions.
%
%    Furthermore, 'eval_counts' is defined as a combination
%    of 'eval_groups' and 'eval_global', and 'fit_counts'
%    now points to 'fit_smart'.
%
%    To access the original ISIS functions just switch the
%    namespace back to 'isis' using 'use_namespace' or
%    access the functions via isis->function_name
%
%    WARNING: there is no check implemented yet that the
%    given structure actually is a SimFit-structure, so
%    every structure containing reference to functions may
%    be passed, which might HARM your ISIS-session or
%    machine!
%
%    FINALLY, if there will be any error the namespace is
%    most likely set to 'isis' afterwards (we tried to
%    catch errors in the SimFit-functions but this does
%    not work for, e.g., syntax errors within the shell).
%\seealso{simultaneous_fit, use_namespace}
%!%-
{
  if (_NARGS != 1) { help(_function_name); return; }
  variable s = ();
  variable space   = qualifier("name", "simfit");
  variable current = "Global";
  variable chatty = qualifier("chatty", 1);
  
  % check on existence of the namespace
  if (any(_get_namespaces == space)) {
    vmessage("error (%s): namespace '%s' already exists", _function_name, space);
    return;
  }

  % define a new variable holding the simfit structure
  % into the current namespace
  variable simfit = sprintf("SimFit_Structure_%s", space);
  if (chatty) { vmessage("copying structure to variable %s->%s", current, simfit); }
  eval(sprintf("variable %s", simfit), current);
  use_namespace(current);
  @(__get_reference(simfit)) = s;

  % define new namespace
  if (chatty) { vmessage("implementing namespace '%s'", space); }
  implements(space);
  % implement functions
  if (chatty) { vmessage("defining functions in namespace '%s'", space); }
  variable fun, code;
  foreach fun (get_struct_field_names(s)) {
    variable ref = get_struct_field(s, fun);
    if (typeof(ref) == Ref_Type) {
      code  = sprintf("define %s() {\n", fun);
      code += sprintf("  variable args = __pop_args(_NARGS);\n");
      code += sprintf("  use_namespace(\"%s\");\n", current);
      code += sprintf("  variable e = NULL;\n");
      code += sprintf("  try(e) { %s.%s(__push_args(args);; __qualifiers); }\n", simfit, fun);
      code += sprintf("  catch AnyError: { message(e.message); };\n");
      code += sprintf("  use_namespace(\"%s\");\n", space);
      code += sprintf("}\n");
      eval(code, space);
    }
  }
  % implement eval_counts
  code  = sprintf("define eval_counts() {\n");
  code += sprintf("  variable args = __pop_args(_NARGS);\n");
  code += sprintf("  use_namespace(\"%s\");\n", current);
  code += sprintf("  variable e = NULL;\n");
  code += sprintf("  try(e) { variable a = %s.eval_groups(__push_args(args);; struct_combine(struct { skipfinaloutput }, __qualifiers));\n", simfit);
  code += sprintf("  variable b = %s.eval_global(;; __qualifiers); }\n", simfit);
  code += sprintf("  catch AnyError: { message(e.message); };\n");
  code += sprintf("  use_namespace(\"%s\");\n", space);
  code += sprintf("  return (a && b);\n");
  code += sprintf("}\n");
  eval(code, space);
  % implement fit_counts
  code  = sprintf("define fit_counts() {\n");
  code += sprintf("  variable args = __pop_args(_NARGS);\n");
  code += sprintf("  use_namespace(\"%s\");\n", current);
  code += sprintf("  variable e = NULL;\n");
  code += sprintf("  try(e) { %s.fit_smart(__push_args(args);; __qualifiers); }\n", simfit);
  code += sprintf("  catch AnyError: { message(e.message); };\n");
  code += sprintf("  use_namespace(\"%s\");\n", space);
  code += sprintf("}\n");
  eval(code, space);
} %}}}

%%% TODO %%%
% list_par (what was this about?)
% [a-b] place holder
% hack substructure (see below in main function simultaneous_fit)
% group_stats -> list or return d.o.f. for all groups (mu-factor!)

variable SimFit_Verbose = -1; % verbosity during [eval/fit]_counts;
variable SimFit_Verbose_Histogram = 0; % show (0) or hide (-1) histogram after [eval/fit]_*

private define _simultaneous_fit_help(nargs, fname) %{{{
{ % display help to a specific function (adopted from xfgi module)
  if (nargs == 0 && qualifier_exists("help") == 0) return 0;
  _pop_n(nargs);
  help(strreplace(fname, "eous_fit_", "eous_fit."));
  if (qualifier_exists("help")) return 1;
} %}}}

private define _simultaneous_fit_parname(parname, dataid);
private define _simultaneous_fit_parname(parname, dataid) %{{{
{ % evaluates a parameter name with placeholders
  % check if placeholder exists
  if (is_substr(parname, "%")) {
    % matches on parameters WITHOUT %
    variable nparname = string_matches(parname, "\(.*\)(\([^%]*\))\.\(.*\).*"R);
    % if no parameter was found, try to evaluate the placeholder
    if (nparname == NULL) {
      nparname = string_matches(parname, "\(.*\)(\(.*\))\.\(.*\).*"R);
      if (nparname == NULL) return parname;
      else return sprintf("%s(%d).%s", _simultaneous_fit_parname(nparname[1], dataid),
			               eval(strreplace(nparname[2], "%", sprintf("%d", dataid))),
			               _simultaneous_fit_parname(nparname[3], dataid));
    }
    % otherwise, repeat with strings before and after the found non-% parameter
    else return sprintf("%s(%s).%s", _simultaneous_fit_parname(nparname[1], dataid),
			             nparname[2],
			             _simultaneous_fit_parname(nparname[3], dataid));
  }
  % return input without changing anything
  else return parname;
} %}}}

private define _simultaneous_fit_get_instances(ff, comp) %{{{
{ % extracts all instances of the given component from the fit-function
  variable rules = String_Type[0];
  variable z = 1;
  while (z < strlen(ff)) {
    variable s = is_substr(substr(ff, z, -1), sprintf("%s(", comp)) - 1;
    if (s >= 0) {
      variable p = 1, l = strlen(comp), e = 1;
      while(p > 0) {
        switch(substr(ff, z+s+l+e, 1))
          { case "(": p++; }
          { case ")": p--; }
          { case ",": if (p == 1) p = 0; }
          { e++; }
      }
      rules = [rules, substr(ff, z+s+l+1, e-1)];
      z += s+l+e+1;
    } else z = strlen(ff);
  }
  return rules;
} %}}}

private define simultaneous_fit_setrestore() %{{{
%!%+    
%\function{simultaneous_fit.setrestore}
%\synopsis{To be written}
%\usage{simultaneous_fit.setrestore();}
%\description
%    To be written...
%\seealso{simultaneous_fit.restore}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: (s) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  s.model.exclude = all_data[where(array_struct_field(array_map(Struct_Type, &get_data_info, all_data), "exclude") == 1)];
  s.model.params = get_params();
} %}}}

private define simultaneous_fit_restore() %{{{
%!%+    
%\function{simultaneous_fit.restore}
%\synopsis{To be written}
%\usage{simultaneous_fit.restore();}
%\description
%    To be written...
%\seealso{simultaneous_fit.setrestore}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: (s) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  include(array_flatten(list_to_array(s.data, Array_Type)));
  exclude(s.model.exclude);
  if (qualifier_exists("params")) {
    set_params(s.model.params);
  } else {
    variable fstate = array_struct_field(s.model.params, "freeze");
    freeze(freeParameters);
    thaw(array_struct_field(s.model.params, "name")[where(fstate == 0)]);
  }
} %}}}

private define simultaneous_fit_apply_logic() %{{{
%!%+    
%\function{simultaneous_fit.apply_logic}
%\synopsis{ties parameters of simultaneous data to each other}
%\usage{simultaneous_fit.apply_logic();}
%\qualifiers{
%    \qualifier{keepparfun}{do not reset all parameter functions}
%    \qualifier{chatty}{chattiness of this function (default: 1)}
%}
%
%\description
%    For each defined data an individual set of parameters
%    exist. If some of the data were taken simultaneously
%    the parameters should be the same. This function applies
%    this logic by tieing the parameters of these data to
%    each other by setting the corresponding parameter
%    functions via set_par_fun. To do so the functions for
%    all parameters are set to NULL first (this can be
%    inhibited using the keepparfun-qualifier).
%\seealso{simultaneous_fit.add_data, simultaneous_fit.fit_fun,
%    set_par_fun}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: (s) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  variable chatty = qualifier("chatty", 1);

  if (s.model.fit_fun == NULL) return;
  % the updategroup-qualifier is used internally and, thus,
  % not listed in the help
  variable update = qualifier("updategroup", NULL);
  % delete all parameter functions (can be reset eventually
  % with .set_par_fun_history if the user has used
  % .set_par_fun only)
  if (update == NULL) {
    ifnot (qualifier_exists("keepparfun")) {
      if (chatty) { message("setting all parameter functions to NULL"); }
      array_map(Void_Type, &set_par_fun,
	array_struct_field(get_params, "name"), NULL);
      s.sort_params(; updategroup = 0, chatty = 0);
    }
  }

  % loop over all parameters of a data group
  % and tie parameters to each other
  variable g, check = 0;
  if (chatty) { message("tieing group parameters together"); }
  _for g (update == NULL ? 0 : update, update == NULL ? length(s.model.group)-1 : update, 1) {
    variable p;
    _for p (0, length(s.model.group[g])-1, 1) {
      if (s.model.tieto[g][p] != NULL)
        set_par_fun(s.model.group[g][p], s.model.tieto[g][p]);
      else if (update != NULL && get_par_info(s.model.group[g][p]).fun != NULL) {
	set_par_fun(s.model.group[g][p], NULL);
	check = 1;
      }
    }
  }

  if (check) {
    vmessage("warning (%s): parameter logic of group %d changed, please check parameter functions:\n", _function_name, update+1);
    s.list_groups(update+1; tied);
  }
} %}}}

private define simultaneous_fit_sort_params() %{{{
%!%+    
%\function{simultaneous_fit.sort_params}
%\synopsis{sort all fit parameters into global and group ones}
%\usage{simultaneous_fit.sort_params();}
%\qualifiers{
%    \qualifier{chatty}{chattyness of this function (default: 1)}
%}
%\description
%    The defined data may contain both, simultaneous and non-
%    simultaneous data. Thus, there are parameters which apply
%    to individual data on one hand (called group parameters),
%    and on the other hand parameters which act on all defined
%    data (called global parameters). This function determines
%    the relationship between the defined data in order to sort
%    the existing parameters into global and group parameters.
%\seealso{simultaneous_fit.add_data, simultaneous_fit.apply_logic}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: (s) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  if (s.model.fit_fun == NULL) { return; }
  if (length(s.data) == 0) { return; }
  variable chatty = qualifier("chatty", 1);
  variable update = qualifier("updategroup", NULL);

  % remember excluded datasets
  s.setrestore();
  
  % idea: include the data groups one after another and
  %       extract the names of any new parameters
  if (chatty) { message("analyzing parameters"); }
  variable g, ids = array_flatten(list_to_array(s.data, Array_Type));
  exclude(ids);
  exclude(ids); % needed to update list of parameters properly (Isis_Active_Dataset -> 0)
  % get global parameters
  variable globalpars = get_params;
  if (length(globalpars) > 0) { s.model.global = array_struct_field(globalpars, "name"); }
  % remove parameters, which don't exist after including all data (e.g. component(0).parameter)
  include(ids);
  variable n = Integer_Type[0];
  _for g (0, length(s.model.global)-1, 1) if (_get_index(s.model.global[g]) > -1) n = [n, g];
  s.model.global = s.model.global[n];
  if (chatty) { vmessage("  -> %d global parameters", length(s.model.global)); }

  if (update == 0) { return; }
  % get parameteres assigned to specific data groups
  if (update == NULL) { % forget the known logic
    s.model.instances = Assoc_Type[Array_Type];
    s.model.group = Array_Type[length(s.data)];
    s.model.tieto = Array_Type[length(s.data)];
    s.model.stat = struct_array(length(s.data)+1, struct {
      statistic = 0,
      num_variable_params = 0,
      num_bins = 0
    }); % +1 for globals (index 0)
  }
  _for g (update == NULL ? 0 : update, update == NULL ? length(s.data)-1 : update, 1) {
    variable d, thispars = String_Type[0], tiesto = String_Type[0];
    _for d (0, length(s.data[g])-1, 1) {
      exclude(ids);
      include(s.data[g][d]);
      variable actpars = get_params;
      if (length(actpars) > 0) {
        actpars = array_struct_field(actpars, "name");
        % remove global parameters
        n = Integer_Type[0];
	variable t;
        _for t (0, length(actpars)-1, 1) if (wherefirst(s.model.global == actpars[t]) == NULL) n = [n, t];
        thispars = [thispars, actpars[n]];
	% check if parameter can/may be tied to first dataset of this group
	_for t (0, length(n)-1, 1) {
	  variable comp;
          if (d == 0) { % if first dataset of this group, extract all instances of the fit components
	    tiesto = [tiesto, NULL];
	    comp = substr(actpars[n[t]], 1, is_substr(actpars[n[t]], "(")-1);
	    ifnot (assoc_key_exists(s.model.instances, comp)) s.model.instances[comp] = _simultaneous_fit_get_instances(s.model.fit_fun, comp);
	  }
	  else {
	    % find parameter name of first data within this group
            comp = string_matches(actpars[n[t]], "\(.*\)(\(\d*\))\.\(.*\).*"R);
	    variable tieparname = NULL;
	    if (assoc_key_exists(s.model.instances, comp[1])) {
	      % find instance string, which results into the current parameter instance
	      variable inst = s.model.instances[comp[1]];
              variable i = wherefirst(array_map(Integer_Type, &eval, array_map(String_Type, &strreplace, inst, "%", sprintf("%d", s.data[g][d]))) == atoi(comp[2]));
	      if (i != NULL) tieparname = sprintf("%s(%d).%s", comp[1], eval(strreplace(inst[i], "%", sprintf("%d", s.data[g][0]))), comp[3]);
	    }
	    tiesto = [tiesto, tieparname];
	  }
	}
      }
    }
    s.model.group[g] = @thispars;
    s.model.tieto[g] = @tiesto;
  }
  if (chatty) {
    n = int(sum(array_map(Integer_Type, &length, s.model.group)));
    t = 1. * n / length(s.model.group);
    vmessage("  -> %d group parameters, %.0f per group", n, t);
    if (t - int(t) > 0) {
      message("  warning: number of parameters per group is no integer!");
    }
  }
  
  s.restore();
} %}}}

private define simultaneous_fit_add_data_old() %{{{
%!%+    
%\function{simultaneous_fit.add_data_old}
%\synopsis{adds the given data to the simultaneous/combined fit (old version)}
%\usage{simultaneous_fit.add_data(String_Type[] phafiles);
% or simultaneous_fit.add_data(Struct_Type[] { bin_lo, bin_hi, value, error });
% or simultaneous_fit.add_data(List_Type datagroups);}
%\qualifiers{
%    \qualifier{loadfun}{function reference to load the given filename
%              (default &load_data), further arguments are passed}
%    \qualifier{loadfunlist}{indicates that 'loadfun' already returns an
%              array of datagroups, i.e. a list with integer
%              arrays representing the datasets in each group}
%    \qualifier{nosort}{do not sort the parameters (implies 'nologic')}
%    \qualifier{nologic}{do not apply parameter logic}
%    \qualifier{ROC}{value for Rmf_OGIP_Compliance for each data,
%              has to be given in the same structure as the
%              input data, i.e., as integer array or list of
%              integer arrays}
%}
%\description
%    This function is deprecated and is superseded by the default
%    one in the simultaneous-fit-structure.
%
%    The given data will be loaded or defined and added to the
%    simultaneous/combined fit. This data can be given as either
%    a file name to the spectrum as accepted by 'load_data' or a
%    structure as accepted by 'define_counts'.
%
%    If an array of data is given, this data will be treated
%    as simultaneously recorded data (called a data group).
%    That means, if any parameter logic is applied afterwards,
%    the parameters of all datasets are tied to each other.
%    This logic will be applied automatically unless the
%    'nologic' qualifier is given. To apply it later,
%    'simultaneous_fit.apply_logic' may be called. This logic,
%    however, has to be known first, which is done by
%    'simultaneous_fit.sort_params' automatically. Again,
%    this may be skipped by the 'nosort' qualifier.
%
%    Further on, a list of several data groups can be given.
%    This list may contain both, filenames to spectra and
%    structures and arrays of these as well.
%\example
%    % define a simultaneous fit first
%    variable simfit = simultaneous_fit();
%    
%    % add a single RXTE observation (consisting of a PCA-
%    % and a HEXTE-spectrum, both taken simultaenously, of
%    % course)
%    simfit.add_data(["pca.pha", "hexte.pha"]);
%
%    % add an RXTE (PCA and HEXTE) and a Swift-XRT
%    % observation (both observations were NOT taken
%    % simultaneously -> List_Type)
%    simfit.add_data({["pha.pha", "hexte.pha"], "xrt.pha"});
%\seealso{load_data, define_counts, simultaneous_fit.sort_params,
%    simultaneous_fit.apply_logic}
%!%-
{
  variable s, data, addargs = Struct_Type[0];
  if (_NARGS == 2) (s,data) = ();
  else if (_NARGS > 2) {
    addargs = __pop_args(_NARGS-2);
    (s,data) = ();
  }
  else { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  variable loadfun = qualifier("loadfun", &load_data);
  % loadfun given and that returns a list of data groups
  variable loadfunlist = qualifier_exists("loadfun") && qualifier_exists("loadfunlist");
  
  % convert input string(s), struct(s), or integer(s) into a list
  if (any(_typeof(data) == [String_Type, Struct_Type, Integer_Type]))
    { data = { data }; }
  else if (_typeof(data) != List_Type)
    { vmessage("error (%s): wrong type of 'data'", _function_name); return; }

  % same for Rmf_OGIP_Compliance
  variable roc = qualifier("ROC", NULL);
  variable defroc = Rmf_OGIP_Compliance;
  if (roc != NULL) {
    if (_typeof(roc) == Integer_Type) roc = { roc };
    else if (_typeof(roc) != List_Type) { vmessage("error (%s): wrong type of 'roc' qualifier", _function_name); return; }
    if (length(roc) != length(data) || length(array_flatten(list_to_array(roc))) != length(array_flatten(list_to_array(data))))
      { vmessage("error (%s): 'roc' qualifier has to have the same structure and number of elements as the input data", _function_name); return; }
  }
    
  % loop over data groups
  variable g, nid = list_new;
  _for g (0, length(data)-1, 1) {
    % convert input into array (will not affect an already given array)
    if (typeof(data[g]) != Array_Type) data[g] = [data[g]];
    % load the data or define the counts
    variable d, ids = loadfunlist ? List_Type[0] : Integer_Type[0];
    _for d (0, length(data[g])-1, 1) {
      if (roc != NULL) Rmf_OGIP_Compliance = roc[g][d];
      ids = [ids, typeof(data[g][d]) == String_Type ? (@loadfun)(data[g][d], __push_args(addargs);; __qualifiers) : define_counts(data[g][d] ;; __qualifiers)];
      if ((loadfunlist ? ids[-1][0] : ids[-1]) == -1) { vmessage("error (%s): something went wrong during loading or defining data", _function_name); return; }
    }
    % eventually set data info
    if (qualifier_exists("datainfo")) {
      foreach d (ids)
	set_data_info(d, struct_combine(get_data_info(d), qualifier("datainfo")));
    }
    % remember data indices
    if (loadfunlist) {
      _for d (0, length(ids)-1, 1) {
	while (length(ids[d]) > 0) {
          list_append(s.data, [list_pop(ids[d])]);
          list_append(nid, s.data[-1]);
	}
      }
    }
    else {
      list_append(s.data, ids);
      list_append(nid, s.data[-1]);
    }
  }

  % parameter stuff
  ifnot (qualifier_exists("nosort")) {
    s.sort_params();
    ifnot (qualifier_exists("nologic")) s.apply_logic();
  }

  Rmf_OGIP_Compliance = defroc;
  return nid;
} %}}}

private define simultaneous_fit_add_data() %{{{
%!%+    
%\function{simultaneous_fit.add_data}
%\synopsis{adds the given data to the simultaneous/combined fit}
%\usage{simultaneous_fit.add_data(List_Type data_for_one_group);
% or simultaneous_fit.add_data(List_Type[] multiple_groups);}
%\qualifiers{
%    \qualifier{nosort}{do not sort the parameters (implies 'nologic')}
%    \qualifier{nologic}{do not apply parameter logic}
%}
%\description
%    NOTE: compared to the previous version of this function the data
%          has to be provided as a list or array of lists!
%
%    The given data will be loaded or defined and added to the
%    simultaneous/combined fit. The data can be provided in different
%    formats inside a surrounding list. All data within this list is
%    treated as a single data group. Further groups can be defined by
%    either calling add_data again or by providing an array of lists,
%    where each item corresponds to a group. The data inside the
%    list(s) can be given as
%    - String_Type filename: the given filename is loaded by a call
%      to 'load_data'.
%    - Integer_Type dataset: the ID of an already existing ISIS
%      dataset.
%    - Struct_Type, the field layouts allowed are
%      - { Double_Type[] bin_lo, bin_hi, value, error }: is passed
%        to 'define_counts' by default.
%      - { String_Type filename }: is passed to 'load_data' by
%        default.
%      Further fields allow additional features:
%        - { ..., Integer_Type roc }: Rmf_OGIP_Compliance is set to
%           'roc' before the data is defined or loaded.
%        - { ..., Ref_Type loadfun }: reference to a function which
%          should be used instead of 'load_data' or 'define_counts'.
%          This function has to return the new ISIS dataset ID.
%        - { ..., Struct_Type qualifiers } - the structure is passed
%          as qualifiers to the function for loading or defining
%          the data.
%     
%    After the data and corresponding data groups have been defined
%    the resulting parameter logic is analyzed by
%    'simultaneous_fit.sort_params'. As this step might take a long
%    time and is only needed once (during the last call to 'add_data'
%    or manually), you can skip this step using the 'nosort'
%    qualifier.
%    Finally, the analyzed parameter logic is applied to the defined
%    model and parameters using 'simultaneous_fit.apply_logic' unless
%    the 'nologic' qualifier is set.
%\example
%    % define a simultaneous fit structure first
%    variable simfit = simultaneous_fit();
%    
%    % add a single RXTE observation (consisting of a PCA-
%    % and a HEXTE-spectrum, both taken simultaenously, of
%    % course)
%    simfit.add_data({"pca.pha", "hexte.pha"});
%
%    % add two data groups consisting of an RXTE (PCA and HEXTE) and
%    % a Swift-XRT observation
%    simfit.add_data([ % note the array
%      {"pha.pha", "hexte.pha"}, % first data group
%      {struct { filename = "xrt.pha", roc = 0 }} % second data group
%    ]);
%
%    % add one already existing ISIS datasets and load an PHA-file
%    % both defining a single data group
%    % fit grouped into two data groups
%    simfit.add_data({1, "file.pha"});
%\seealso{load_data, define_counts, simultaneous_fit.sort_params,
%    simultaneous_fit.apply_logic}
%!%-
{
  variable s, data;
  if (_NARGS == 2) { (s,data) = (); }
  else { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) { return; }

  % sanity checks
  if (typeof(data) == List_Type) { data = [data]; }
  else if (typeof(data) != Array_Type || _typeof(data) != List_Type) {
    vmessage("error (%s): 'data' have to be given as list or array of lists", _function_name);
    return;
  }

  variable troc = Rmf_OGIP_Compliance;

  % loop over array times (= data groups)
  variable g, d, ids = List_Type[length(data)];
  _for g (0, length(data)-1, 1) {
    % loop over list itmes (= datasets wihtin the group)
    ids[g] = list_new;
    _for d (0, length(data[g])-1, 1) {
      switch (typeof(data[g][d]))
      % load file
      { case String_Type:
	list_append(ids[g], load_data(data[g][d]));
      }
      % add existing dataset
      { case Integer_Type:
	list_append(ids[g], data[g][d]);
      }
      { case Struct_Type:
	variable fields = get_struct_field_names(data[g][d]);
	
	% determine struct type
	variable type;
	if (any(fields == "filename")) { type = 1; }
	else if (any(fields == "bin_lo") && any(fields == "bin_hi") &&
		 any(fields == "value") && any(fields == "err")) { type = 2; }
	else {
	  vmessage(
            "error (%s): given structure layout is not allowed", _function_name
	  ); return;
	}
	
	% define load function
	variable funref = any(fields == "loadfun") ? data[g][d].loadfun
                        : type == 1 ? &load_data : type == 2 ? &define_counts : NULL;
	if (funref == NULL) {
	  vmessage(
            "error (%s): function to load/define the data is NULL", _function_name
	  ); return;
	}
	% additional qualifiers
	variable qual = any(fields == "qualifiers") ? data[g][d].qualifiers : NULL;
	% custom Rmf_OGIP_Compliance
	if (any(fields == "roc")) { Rmf_OGIP_Compliance = data[g][d].roc; }

	% now load or define the data
	list_append(ids[g], (@funref)(
	  type == 1 ? data[g][d].filename : data[g][d]
	;; qual));

	% restore Rmf_OGIP_Compliance
	if (any(fields == "roc")) { Rmf_OGIP_Compliance = troc; }
      }
    }
    
    % add IDs to internal simfit structure
    list_append(s.data, list_to_array(ids[g], Integer_Type));
  }
  
  ifnot (qualifier_exists("nosort")) {
    % analyze parameter logic
    s.sort_params();
    % apply logic
    ifnot (qualifier_exists("nologic")) { s.apply_logic(); }
  }

  return ids;
} %}}}

private define simultaneous_fit_delete_data() %{{{
%!%+    
%\function{simultaneous_fit.delete_data}
%\synopsis{deletes the given data from the simultaneous/combined fit}
%\usage{simultaneous_fit.delete_data(Integer_Type[] group[, Integer_Type[] data]);}
%\qualifiers{
%    \qualifier{keep}{the associated data are not deleted by
%           'delete_data', thus only the logic of the
%           simultaneous fit is modified}
%}
%\description
%    The given data 'group' is removed from the simultaneous
%    fit and the associated data is deleted (using the ISIS
%    internal function 'delete_data'. The second, optional
%    argument can be used to delete specific data from a
%    group specified by its number within the group (starting
%    at one).
%
%    If the first group is deleted, the global parameters
%    might change. In that case a warning message will be
%    raised and you should check the parameter dependencies!
%
%    Further on, if the first dataset of a group is deleted,
%    the parameter logic will change and paramaters will be
%    tied again according to that logic (but of that group
%    only). Please check the resulting parameter dependencies
%    (in particular the value-functions of global parameters)
%    as suggested by the warning message!
%\example
%    % delete data group 3 and 5
%    simultaneous_fit.delete_data([3,5]);
%
%    % delete data group 1
%    % this might cause that a global parameter gets deleted
%    simultaneous_fit.delete_data(1);
%    
%    % delete the first spectrum from group 4 only
%    % this will probably change the group parameter dependencies
%    simultaneous_fit.delete_data(4, 1);
%\seealso{delete_data, simultaneous_fit.add_data}
%!%-
{
  variable s, grp, data = NULL;
  switch (_NARGS)
    { case 2: (s,grp) = (); }
    { case 3: (s,grp,data) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  if (typeof(grp) == Array_Type && typeof(data) == Array_Type)
    { vmessage("error (%s): either 'group' or 'data' may be an array, not both", _function_name, length(s.data)); return; }
  if (typeof(grp) != Array_Type) grp = [grp];
  if (min(grp) < 1 || max(grp) > length(s.data))
    { vmessage("error (%s): allowed data groups range from 1 to %d", _function_name, length(s.data)); return; }
  if (data != NULL) {
    if (typeof(data) != Array_Type) data = [data];
    if (min(data) < 1 || max(data) > length(s.data[grp[0]-1]))
      { vmessage("error (%s): allowed data in group %d range from 1 to %d", _function_name, grp[0], length(s.data[grp[0]-1])); return; }
  }
      
  % loop over given groups
  variable g, ng = 0;
  variable keep = qualifier_exists("keep");
  foreach g (grp-1) {
    % remember global parameters
    variable global = @(s.model.global);
    % remove data
    if (data == NULL) {
      ifnot (keep) { delete_data(s.data[g-ng]); }
      list_delete(s.data, g-ng);
      if (s.model.fit_fun != NULL) {
        s.model.group = array_remove(s.model.group, g-ng);
        s.model.tieto = array_remove(s.model.tieto, g-ng);
        s.sort_params(; updategroup = 0); % update global parameters (if necessary)
      }
      ng++;
    } else {
      ifnot (keep) { delete_data(s.data[g-ng][data-1]); }
      data = data[array_sort(data)];
      variable d;
      _for d (0, length(data)-1, 1)
	s.data[g-ng] = array_remove(s.data[g-ng], data[d]-d-1);
      % whole group was deleted
      if (length(s.data[g-ng]) == 0) {
        list_delete(s.data, g-ng);
        if (s.model.fit_fun != NULL) {
          s.model.group = array_remove(s.model.group, g-ng);
          s.model.tieto = array_remove(s.model.tieto, g-ng);
          s.sort_params(; updategroup = 0); % update global parameters (if necessary)
	}
	ng++;
      } else {
        if (s.model.fit_fun != NULL) {
          s.sort_params(; updategroup = g-ng); % update global and group parameters
          s.apply_logic(; updategroup = g-ng);
	}
      }
    }
    % check on changed global parameters
    variable p;
    foreach p (global)
      if (wherefirst(s.model.global == p) == NULL)
        vmessage("warning (%s): global parameter %s has been deleted and removed from the list!", _function_name, p);
  }
} %}}}

private define simultaneous_fit_fit_fun() %{{{
%!%+    
%\function{simultaneous_fit.fit_fun}
%\synopsis{defines the model applied within the simultaneous fit}
%\usage{simultaneous_fit.fit_fun(String_Type fit_fun);}
%\qualifiers{
%    \qualifier{nosort}{do not sort the parameters
%                (implies 'nologic' and 'nohistory')}
%    \qualifier{nologic}{do not apply parameter logic}
%    \qualifier{nohistory}{do not apply the set_par_fun-history}
%    \qualifier{ask}{see set_par_fun_history}
%    \qualifier{chatty}{chattiness of this function (default: 1)}
%}
%\description
%    The given string is used to define the fit function.
%    Usually, a lot of data is added to the simultaneous
%    fit, which might lead to more complicated fit functions.
%    In order to simplify its definition, some place holders
%    can be used within the fit function:
%      % - will be replaced by Isis_Active_Dataset
%    More placeholders will be implemented in the future.
%
%    Afterwards, the parameters are sorted into global and
%    group ones and the parameter logic is applied to the
%    model (see simultaneous_fit.sort_params and
%    simultaneous_fit.apply_logic).
%
%    Finally, any parameter functions previously define
%    with simultaneous_fit.set_par_fun are restored by
%    calling simultaneous_fit.set_par_fun_history(; apply).
%\example
%    % the model shall be an absorbed powerlaw, where there
%    % are independent parameters of the powerlaw for each
%    % data group, but a single absorber, which is applied
%    % to all data
%    simultaneous_fit.fit_fun("tbnew(1)*powerlaw(%)");
%\seealso{fit_fun, simultaneous_fit.sort_params,
%    simultaneous_fit.apply_logic, simultaneous_fit.set_par_fun_history}
%!%-
{
  variable s, ff;
  switch (_NARGS)
    { case 2: (s,ff) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  variable chatty = qualifier("chatty", 1);
  
  % modify and set fit function
  if (chatty) { message("setting fit-function"); }
  s.model.fit_fun = ff;
  ff = strreplace(ff, "%", "Isis_Active_Dataset");
  fit_fun(ff);

  % parameter stuff
  ifnot (qualifier_exists("nosort")) {
    s.sort_params(; chatty = chatty);
    if (qualifier_exists("nohistory")) {
      ifnot (qualifier_exists("nologic")) { s.apply_logic(; chatty = chatty); }
    } else {
      s.set_par_fun_history(;; struct_combine(
	struct { apply, chatty = chatty }, [
          qualifier_exists("ask") ? "ask" : String_Type[0],
          qualifier_exists("nologic") ? "nologic" : String_Type[0],
        ])
      );
    }
  }
} %}}}

private define simultaneous_fit_get_par() %{{{
%!%+    
%\function{simultaneous_fit.get_par}
%\synopsis{get the value of a/many fit parameter/parameters}
%\usage{Double_Type[] simultaneous_fit.get_par(String_Type parameter);}
%\description
%    Does the same as 'get_par' with the modification that
%    the following place holders are allowed:
%      % - set the value for all defined data
%    Note, that here the parameter name has to be provided,
%    not the parameter id!
%
%    All qualifiers are passed to get_par.
%\example
%    % get the normalization of all powerlaws
%    norms = simultaneous_fit.get_par("powerlaw(%).norm");
%\seealso{get_par}
%!%-
{
  variable s, parname;
  switch (_NARGS)
    { case 2: (s,parname) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % loop over all data groups
  variable g, values = Double_Type[0];
  _for g (0, length(s.data)-1, 1) {
    % loop over datasets within the group
    variable d;
    _for d (0, length(s.data[g])-1, 1) {
      % current parameter name
      variable nparname = _simultaneous_fit_parname(parname, s.data[g][d]);
      % check if this parameter exists
      if (_get_index(nparname) > -1) {
        variable pinfo = get_par_info(nparname);
        % get this parameter
        if (pinfo.fun == NULL) values = [values, get_par(nparname)];
      }
    }
  }

  return values;
} %}}}

private define simultaneous_fit_set_par() %{{{
%!%+    
%\function{simultaneous_fit.set_par}
%\synopsis{set the value of a/many fit parameter/parameters}
%\usage{simultaneous_fit.set_par(String_Type parameter[, Double_Type value[, Integer_Type freeze[, Double_Type min, max]]]);}
%\description
%    Does the same as 'set_par' with the modification that
%    the following place holders are allowed:
%      % - set the value for all defined data
%    Note, that here the parameter name has to be provided,
%    not the parameter id!
%
%    All qualifiers are passed to set_par.
%\example
%    % set the normalization of all powerlaws
%    simultaneous_fit.set_par("powerlaw(%).norm", 1, 0, 0, 10);
%\seealso{set_par}
%!%-
{
  variable s, parname, value, freeze, min, max;
  switch (_NARGS)
    { case 2: (s,parname) = (); value = NULL; }
    { case 3: (s,parname,value) = (); }
    { case 4: (s,parname,value,freeze) = (); }
    { case 6: (s,parname,value,freeze,min,max) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % loop over all data groups
  variable nparname = String_Type[0];
  variable g,d;
  _for g (0, length(s.data)-1, 1) {
    % loop over datasets within the group
    _for d (0, length(s.data[g])-1, 1) {
      % current parameter name
      nparname = [ nparname, _simultaneous_fit_parname(parname, s.data[g][d]) ];
      % check if this parameter exists
    }
  }
  
  value = (value == NULL ? get_par(nparname) : value);
  % set this parameter
  switch (_NARGS)
  { case 2: set_par(nparname ;; __qualifiers); }
  { case 3: set_par(nparname, value ;; __qualifiers); }
  { case 4: set_par(nparname, value, freeze ;; __qualifiers); }
  { case 6: set_par(nparname, value, freeze, min, max ;; __qualifiers); }
  
} %}}}

private define simultaneous_fit_copy_par() %{{{
%!%+    
%\function{simultaneous_fit.copy_par}
%\synopsis{copy the parameters of one group to another}
%\usage{simultaneous_fit.copy_par(Integer_Type from_group, to_group);}
%\qualifiers{
%    \qualifier{limits}{copy parameter limits as well}
%}
%\description
%    The values of the group parameters of the first given
%    datagroup are copied to the second datagroup. Parameter
%    limits are not copied unless the limits-qualifier is
%    set.
%
%    Instead of a single target data group, 'to_group', an
%    array of groups, Integer_Type[], may be given.
% \example
%    % copy the parameter values of group 5 to group 6
%    simultaneous_fit.copy_par(5, 6);
%\seealso{simultaneous_fit.list_groups}
%!%-
{
  variable s, g1, g2;
  switch (_NARGS)
    { case 3: (s,g1,g2) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  variable pars = merge_struct_arrays(get_params(s.model.group[g1-1]));
  if (typeof(g2) != Array_Type) { g2 = [g2]; }
  variable g;
  foreach g (g2) {
    if (qualifier_exists("limits")) {
      variable frz = array_struct_field(get_params(s.model.group[g-1]), "freeze");
      array_map(Void_Type, &set_par, s.model.group[g-1], pars.value, frz, pars.min, pars.max);
    } else {
      array_map(Void_Type, &set_par, s.model.group[g-1], pars.value);
    }
  }
} %}}}

private define simultaneous_fit_set_par_fun() %{{{
%!%+    
%\function{simultaneous_fit.set_par_fun}
%\synopsis{Define the value of a/many fit parameter.parameters using a function}
%\usage{simultaneous_fit.set_par_fun(String_Type parameter, String_Type valuefunction);}
%\description
%    In princible, this function does the same as 'set_par_fun'
%    with the modifcation that place holders are allowed:
%      % - set the function for all defined data
%    Note, that here the parameter name has to be provided,
%    not the parameter id!
%
%    This function tries to recognize, if the value function
%    is a parameter itself, which should be applied to a set
%    of the same parameters (e.g. the photon indices are set
%    to a single one). In that case the parameter given in
%    the value function will be treated as a global one from
%    now on.
%
%    On the other side, if the value functions of a set of
%    parameters are deleted (valuefunction=NULL), which were
%    tied to a global parameter before, this parameter is
%    treated as a group one.
%
%    In both cases, a warning message will be shown in order
%    to inform the user which parameter has been treated as
%    global or group parameter.
%
%    In an upcoming version, the place holders are allowed
%    in the value function as well.
%
%    Any call to this function is remembered in an internal
%    history for applying the calls later again (see
%    simultaneous_fit.set_par_fun_history).
%\example
%    % use a single photon index for all data, while
%    % keeping individual normalizations
%    simultaneous_fit.fit_fun("powerlaw(%)");
%    simultaneous_fit.set_par_fun("powerlaw(%).PhoIndex",
%      "powerlaw(1).PhoIndex");
%
%    % allow individual photon indices again
%    simultaneous_fit.set_par_fun("powerlaw(%).PhoIndex",
%      NULL);
%
%    % more fascinating, tie the photon index to a function
%    % of the powerlaw normalization
%    simultaneous_fit.set_par_fun("powerlaw(%).PhoIndex",
%      "constant(1).factor + powerlaw(%).norm * constant(2).factor");
%\seealso{set_par_fun, simultaneous_fit.set_par_fun_history}
%!%-
{
  variable s, parname, parfun;
  switch (_NARGS)
    { case 3: (s,parname,parfun) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % add call to history
  list_append(s.model.tieto_history, {parname, parfun});
  
  % check if parfun is a parameter and it should be applied to a set
  % of parameters -> it is likely that parfun might become a global parameter
  variable global = 0;
  if (parfun != NULL && strlen(parfun) < 32 && _get_index(parfun) > -1 && is_substr(parname, "%") && qualifier_exists("ignoreglobal") == 0) {
    vmessage("warning (%s): interprete %s as global parameter", _function_name, parfun);
    global = 1;
    % move parameter to global list (if not already there)
    if (wherefirst(s.model.global == parfun) == NULL) s.model.global = [s.model.global, parfun];
    else vmessage("warning (%s): %s is already a global parameter", _function_name, parfun);
    % corresponding parameter within each group is tied below
  }
  % if parfun is NULL, than a global parameter might become
  % a group parameter (all corresponding ones)
  if (parfun == NULL) global = Integer_Type[0];

  % loop over all data groups
  variable g;
  _for g (0, length(s.data)-1, 1) {
    % loop over datasets within the group
    variable d;
    _for d (0, length(s.data[g])-1, 1) {
      % current parameter name
      variable nparname = _simultaneous_fit_parname(parname, s.data[g][d]);
      % check if this parameter exists
      if (_get_index(nparname) > -1) {
        variable pinfo = get_par_info(nparname), n;
	% set parameter function
	if (nparname != parfun || qualifier_exists("all"))
	  set_par_fun(nparname, parfun == NULL ? NULL : _simultaneous_fit_parname(parfun, s.data[g][d]));
        % if parfun is a global parameter tie corresponding group parameter
	if (typeof(global) == Integer_Type && global == 1) {
	  n = wherefirst(s.model.group[g] == nparname);
	  if (n != NULL) s.model.tieto[g][n] = parfun;
	}
	% if parfun is NULL check if *all* parameters are tied to the same
	% global parameter (which will become a group parameter afterwards)
	if (typeof(global) == Array_Type) {
	  n = wherefirst(s.model.group[g] == nparname);
	  if (n != NULL && ((length(global) == 0 && s.model.tieto[g][n] != NULL) || (length(global > 0) && s.model.tieto[g][n] == s.model.tieto[0][global[0]])))
	    global = [global, n];
	  else global = 0;
	}
      }
    }
  }
  % change a global parameter to a group one
  if (typeof(global) == Array_Type) {
    vmessage("warning (%s): interprete that %s is no longer a global parameter", _function_name, s.model.tieto[0][global[0]]);
    s.model.global = array_remove(s.model.global, wherefirst(s.model.global == s.model.tieto[0][global[0]]));
    n = 0;
    _for g (0, length(s.data)-1, 1)
      _for d (0, length(s.data[g])-1, 1) {
	if (d == 0) s.model.tieto[g][global[n]] = NULL;
	else {
	  % parameters of remaining data within this group should be
	  % tied to first dataset (according to the logic)
	  s.model.tieto[g][global[n]] = s.model.group[g][global[n-d]];
	  set_par_fun(s.model.group[g][global[n]], s.model.tieto[g][global[n]]);
	}
	n++;
      }
  }
} %}}}

private define simultaneous_fit_set_par_fun_history() %{{{
%!%+    
%\function{simultaneous_fit.set_par_fun_history}
%\synopsis{applies or clears the history of set_par_fun}
%\usage{simultaneous_fit.set_par_fun_history(; qualifiers);}
%\qualifiers{
%    \qualifier{get}{returns the history of calls to set_par_fun}
%    \qualifier{print}{print the history}
%    \qualifier{apply}{applies all calls of the history again}
%    \qualifier{ask}{prompts the user before applying each call
%              of the history (apply-qualifier required)}
%    \qualifier{clear}{clears the history}
%    \qualifier{chatty}{chattiness of this function (default: 1)}
%    \qualifier{nologic}{do not call apply_logic during the history
%              is applied (apply-qualifier)}
%}
%\description
%    Whenever simultaneous_fit.set_par_fun is called the
%    arguments passed are remembered in an internal
%    history. This functions allows to apply each call of
%    the history again, retrieve the history, or clear it.
%
%    This is particularly useful once the parameter logic
%    is set by simultaneous_fit.apply_logic. Note that
%    simultaneous_fit.fit_fun applies the history auto-
%    matically.
%    Note that applying the history first resets the
%    parameter logic with simultaneous_fit.apply_logic.
%
%    In case the history should be returned a list is
%    returned, where each item is a call to set_par_fun
%    with the two strings which have been passed.
%\seealso{simultaneous_fit.set_par_fun,simultaneous_fit.apply_logic}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: s = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  variable chatty = qualifier("chatty", 1);
  
  % return history
  if (qualifier_exists("get")) {
    return s.model.tieto_history;
  }

  % clear history
  if (qualifier_exists("clear")) {
    s.model.tieto_history = list_new;
    if (chatty) { message("cleaned the set_par_fun-history"); }
    return;
  }

  variable i;
  % print history
  if (qualifier_exists("print")) {
    if (chatty) { message("history of set_par_fun:"); }			 
    _for i (0, length(s.model.tieto_history)-1, 1) {
      vmessage("  set_par_fun(\"%s\", \"%s\");",
	       s.model.tieto_history[i][0],
	       s.model.tieto_history[i][1]);
    }
    return;
  }

  % apply history
  if (qualifier_exists("apply")) {
    ifnot (qualifier_exists("nologic")) { s.apply_logic(; chatty = chatty); }
    variable hist = s.model.tieto_history;
    s.model.tieto_history = list_new;
    variable ask = qualifier_exists("ask"), str;
    if (chatty and length(hist) > 0) { message("applying the set_par_fun-history"); }
    _for i (0, length(hist)-1, 1) {
      str = sprintf("  set_par_fun(\"%s\", \"%s\");"R, hist[i][0], hist[i][1]);
      if (ask) {
	switch (keyinput(; nchr = 1, prompt = str + " % y/n? "))
          { case "y": message(""); }
          { message(""); continue; }
      } else { message(str); }
      s.set_par_fun(hist[i][0], hist[i][1]);
    }
    return;
  }

  ()=_simultaneous_fit_help(0, _function_name; help);
} %}}}

private define simultaneous_fit_set_global() %{{{
%!%+    
%\function{simultaneous_fit.set_global}
%\synopsis{Sets a specific group parameter to be fitted globally}
%\usage{simultaneous_fit.set_global(String_Type grouppar);}
%\description
%    A wrapper around simultaneous_fit.set_par_fun, which
%    simply ties the given group parameter name (using the
%    %-placeholder instead of the dataset index) to a single,
%    global parameter.
%    If successful, a warning message will be shown.
%\example
%    % use a single photon index for all data
%    simultaneous_fit.fit_fun("powerlaw(%)");
%    simultaneous_fit.set_global("powerlaw(%).PhoIndex");
%\seealso{simultaneous_fit.unset_global,simultaneous_fit.set_par_fun}
%!%-
{
  variable s, parname;
  switch (_NARGS)
    { case 2: (s,parname) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  s.set_par_fun(parname, _simultaneous_fit_parname(parname, 1));
} %}}}

private define simultaneous_fit_unset_global() %{{{
%!%+    
%\function{simultaneous_fit.unset_global}
%\synopsis{A specific group parameter is fitted individually again}
%\usage{simultaneous_fit.unset_global(String_Type grouppar);}
%\description
%    A wrapper around simultaneous_fit.set_par_fun, which
%    simply unties the given group parameter name (using the
%    %-placeholder instead of the dataset index), which was
%    tied to a single global parameter before.
%    If successful, a warning message will be shown.
%\example
%    % use a photon index for each data group
%    simultaneous_fit.unset_global("powerlaw(%).PhoIndex");
%\seealso{simultaneous_fit.set_global,simultaneous_fit.set_par_fun}
%!%-
{
  variable s, parname;
  switch (_NARGS)
    { case 2: (s,parname) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  s.set_par_fun(parname, NULL);
} %}}}

private define simultaneous_fit_freeze() %{{{
%!%+    
%\function{simultaneous_fit.freeze}
%\synopsis{freezes a/many parameter/parameters}
%\usage{simultaneous_fit.freeze(String_Type parameter);}
%\description
%    In princible, this function does the same as 'freeze'
%    with the modifcation that place holders are allowed:
%      % - set the function for all defined data
%    Note, that here the parameter name has to be provided,
%    not the parameter id!
%\seealso{freeze}
%!%-
{
  variable s, parname;
  switch (_NARGS)
    { case 2: (s, parname) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  s.set_par(parname, NULL, 1);
} %}}}

private define simultaneous_fit_thaw() %{{{
%!%+    
%\function{simultaneous_fit.thaw}
%\synopsis{thawes a/many parameter/parameters}
%\usage{simultaneous_fit.thaw(String_Type parameter);}
%\description
%    In princible, this function does the same as 'thaw'
%    with the modifcation that place holders are allowed:
%      % - set the function for all defined data
%    Note, that here the parameter name has to be provided,
%    not the parameter id!
%\seealso{thaw}
%!%-
{
  variable s, parname;
  switch (_NARGS)
    { case 2: (s, parname) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  s.set_par(parname, NULL, 0);
} %}}}

private define simultaneous_fit_list_data() %{{{
%!%+    
%\function{simultaneous_fit.list_data}
%\synopsis{lists the data groups and their associated dataset IDs}
%\usage{simultaneous_fit.list_data([&variable]);}
%\description
%    Lists the data groups and their associated dataset IDs and
%    number of group parameters. If the optional reference to a
%    variable is provided, then the IDs as an array of lists
%    is put into this variable, where each item corresponds to
%    a data group.
%!%-
{
  variable s, ids = NULL;
  switch (_NARGS)
    { case 1: (s) = (); }
    { case 2: (s,ids) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  
  % list all parameters
  variable list;
  list_par(&list);
  if (list == "null") { message(list); return; }
  list = strchop(list, '\n', 0);

  % loop over data groups
  variable g;
  if (typeof(ids) == Ref_Type) { (@ids) = List_Type[length(s.data)]; }
  _for g (0, length(s.data)-1, 1) {
    if (typeof(ids) == Ref_Type) { (@ids)[g] = { __push_array(s.data[g]) }; }
    vmessage("grp %s: IDs=[%s], #pars=%d",
      sprintf(sprintf("%%0%ds", int(ceil(log10(length(s.data))))), sprintf("%d", g+1)),
      strjoin(array_map(String_Type, &sprintf, "%d", s.data[g]), ","),
      length(wherenot(_isnull(s.model.group[g])))
    );
  }
} %}}}

private define simultaneous_fit_list_global() %{{{
%!%+    
%\function{simultaneous_fit.list_global}
%\synopsis{lists all global parameters}
%\usage{simultaneous_fit.list_global();}
%\description
%    Like 'list_par' this function lists all global parameters.
%\seealso{list_par}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: (s) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % this might be done in a more elegant way...
  
  % list all parameters
  variable list;
  list_par(&list);
  if (list == "null") { message(list); return; }
  list = strchop(list, '\n', 0);

  % selective output
  variable l;
  message(list[0]); message(list[1]);
  _for l (2, length(list)-1, 1)
    if (substr(list[l], 1, 3) != "#=>")
      if (wherefirst(array_map(Integer_Type, &is_substr, list[l], " "+s.model.global+" ")) != NULL)
        message(list[l]);
} %}}}

private define simultaneous_fit_list_groups() %{{{
%!%+    
%\function{simultaneous_fit.list_groups}
%\synopsis{lists all parameters of specific data groups}
%\usage{simultaneous_fit.list_groups([Integer_Type[] groups]);}
%\qualifiers{
%    \qualifier{tied}{list parameters, which are determined by a
%           value function, also}
%}
%\description
%    Like 'list_par' this function lists all parameters of
%    one or more specific data groups. Note, that unlike
%    arrays the index of the first data group is 1. If no
%    data groups are given the parameters of all groups
%    are listed. By default, parameters without a value
%    function are listed only.
%\seealso{list_par}
%!%-
{
  variable s, groups;
  switch (_NARGS)
    { case 1: (s) = ();
      groups = s.model.current_groups[0] == -1
             ? [[1:length(s.data)]]
             : s.model.current_groups; }
    { case 2: (s,groups) = (); if (typeof(groups) != Array_Type) groups = [groups]; }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % this might be done in a more elegant way...
  
  % list all parameters
  variable list;
  list_par(&list);
  if (list == "null") { message(list); return; }
  list = strchop(list, '\n', 0);

  % get all group parameters
  variable g, grouppar = String_Type[0];
  _for g (0, length(groups)-1, 1) grouppar = [grouppar, s.model.group[groups[g]-1][where(_isnull(s.model.tieto[groups[g]-1]))]];

  % selective output
  variable l;
  message(list[0]); message(list[1]);
  _for l (2, length(list)-1, 1)
    if (substr(list[l], 1, 3) != "#=>")
      if (wherefirst(array_map(Integer_Type, &is_substr, list[l], grouppar)) != NULL) {
	variable tie = (l < length(list)-2 && substr(list[l+1], 1, 3) == "#=>" ? qualifier_exists("tied") ? 1 : 2 : 0);
	if (tie < 2) { message(list[l]);
	  if (tie) message(list[l+1]);
	}
      }
} %}}}

private define simultaneous_fit_plot_group() %{{{
%!%+    
%\function{simultaneous_fit.plot_group}
%\synopsis{plots data and model of a specific data group}
%\usage{simultaneous_fit.plot_group([Integer_Type group]);}
%\description
%    This function plots the data and model of the given
%    data group using the 'plot_data' function.
%
%    If no group number is given the currently selected
%    group is plotted (if it is the only one selected).
%    A specific group can be selected using .select_groups
%
%    All qualifiers are passed to plot_data
%
%\seealso{plot_data, simultaneous_fit.select_groups}
%!%-
{
  variable s, grp;
  switch (_NARGS)
    { case 1: (s) = ();
      if (length(s.model.current_groups) == 1 && s.model.current_groups != -1) {
	grp = s.model.current_groups;
      } else {
        vmessage("error (%s): only a single selected group can be plotted", _function_name); return;
      }
    }
    { case 2: (s,grp) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  if (grp < 1 || grp > length(s.data))
    { vmessage("error (%s): allowed data group range from 1 to %d", _function_name, length(s.data)); return; }

  variable l = list_new();
  array_map(&list_append, l, s.data[grp-1]);
  plot_data(l ;; __qualifiers);
} %}}}

private define _simultaneous_fit_comb_chi_sqr() %{{{
{ % calculates and prints the combined chi-square statistic
  % after Kuehnel et al, 2015b, Acta. Pol., subm.
  variable s = ();

  % get and merge all statistics
  variable chi = s.model.stat;
  chi = merge_struct_arrays(array_map(Struct_Type, &reduce_struct, chi, "covariance_matrix"));
  if (any(chi.num_bins == 0)) {
    vmessage("error(%s): the global parameters or some groups have not been evaluated or fitted", _function_name);
    return;
  }

  s.model.cstat.statistic = chi.statistic[[1:]];
  s.model.cstat.num_group_params = chi.num_variable_params[[1:]];
  s.model.cstat.num_global_params = chi.num_variable_params[0];
  s.model.cstat.num_bins = int(chi.num_bins[[1:]]);
  variable N = length(chi.statistic)-1; % number of groups
  variable mu = s.model.cstat.weight_fun == NULL % weightening of global parameters
    ? (s.model.cstat.num_bins - s.model.cstat.num_group_params)
      / sum(s.model.cstat.num_bins - s.model.cstat.num_group_params)
    : @(s.model.cstat.weight_fun)(
	s.model.cstat.statistic, s.model.cstat.num_bins,
	s.model.cstat.num_group_params, s.model.cstat.num_global_params
      );
  s.model.cstat.combredchi = 1./N * sum(
    s.model.cstat.statistic / (s.model.cstat.num_bins
      - s.model.cstat.num_group_params - mu * s.model.cstat.num_global_params
  ));

  if (SimFit_Verbose > -2 && qualifier_exists("recalc") == 0) {
    vmessage(" Var. global/group pars. = %d/%d", s.model.cstat.num_global_params, int(sum(s.model.cstat.num_group_params)));
    vmessage("               Data bins = %d", int(sum(s.model.cstat.num_bins)));
    vmessage("              Chi-square = %s", substr(sprintf("%.7f", sum(s.model.cstat.statistic)), 1, 9));
    vmessage("    Comb.red. chi-square = %s", substr(sprintf("%.7f", s.model.cstat.combredchi), 1, 9));
  }
} %}}}

private define _simultaneous_fit_map_groups() %{{{
{ % loops over all groups and calls the given function reference
  variable s, groups, dtype, fun, params = Struct_Type[0];
  if (_NARGS > 4) { params = __pop_args(_NARGS-4); }
  (s, groups, dtype, fun) = ();

  % remember free parameters and exclude state
  s.setrestore(); 
  variable freep = freeParameters;
  variable ids = array_flatten(list_to_array(s.data)); 

  try {
    % loop over all given groups
    variable g, ret = (dtype)[length(groups)];
    _for g (0, length(groups)-1, 1) {
      variable p;
      % print group number
      if (SimFit_Verbose > -2) {
	()=system(sprintf("echo -n %d..", groups[g]));
      }
      % freeze all parameters except those associated to this data group (and are not tied)
      variable grouppar = array_map(Integer_Type, &_get_index, s.model.group[groups[g]-1][where(_isnull(s.model.tieto[groups[g]-1]))]);
      foreach p (freep) if (wherefirst(grouppar == p) == NULL) freeze(p);
      % exclude all datasets except those associated with the given group
      exclude(ids);
      _for p (0, length(s.data[groups[g]-1])-1) { ifnot (any(s.model.exclude == s.data[groups[g]-1][p])) { include(s.data[groups[g]-1][p]); } }
      % call the function
      variable tparams = COPY(params);
      _for p (0, length(tparams)-1) { if (length(tparams[p].value) == length(groups)) { tparams[p].value = tparams[p].value[g]; } }
      (@fun)(__push_args(tparams) ;; __qualifiers);
       % get the returned values from that function
      if (dtype != Void_Type) { ret[g] = (); }
      % restore free parameters and exclude state
      s.restore();
    }

    ifnot (qualifier_exists("skipfinaloutput")) {
      if (SimFit_Verbose > -2) { message(""); }
      % print combined statistic
      _simultaneous_fit_comb_chi_sqr(s);
       % print histogram
      if (SimFit_Verbose_Histogram == 0) { ()=s.group_stats(); }
    }
    if (dtype != Void_Type) { return ret; }
    return;
  }
  catch UserBreakError: {
    if (SimFit_Verbose > -2) { message(""); }
    vmessage("\nuser break during group %d: restoring simfit status", g+1);
    s.restore(; params);
    if (dtype != Void_Type) { return dtype[1]; }
  }
} %}}}

private define _simultaneous_fit_map_global() %{{{
{ % calls the given function reference for the global parameters only
  variable s, dtype, fun, params = Struct_Type[0];
  if (_NARGS > 3) { params = __pop_args(_NARGS-3); }
  (s, dtype, fun) = ();

  % remember free parameters
  variable freep = freeParameters;
  s.setrestore();

  try {
    % print global indicator
    if (SimFit_Verbose > -2) {
      ()=system("echo -n global..");
    }
    % freeze all parameters except the global ones
    variable p, global = array_map(Integer_Type, &_get_index, s.model.global);
    foreach p (freep) if (wherefirst(p == global) == NULL) freeze(p);
    % call the function
    variable ret;
    (@fun)(__push_args(params) ;; __qualifiers);
    if (dtype != Void_Type) { ret = (); }
    % restore free parameters
    thaw(freep);

    % update fit-statistic of all groups
    ifnot (qualifier_exists("skipupdatestat")) {
      variable i, fv = Fit_Verbose;;
      exclude(all_data); Fit_Verbose = -1;
      _for i (0, length(s.data)-1, 1) {
	include(s.data[i]);
	s.model.stat[i+1].statistic = eval_stat_counts().statistic;
	exclude(s.data[i]);
      }
      include(all_data); Fit_Verbose = fv;
    }

    ifnot (qualifier_exists("skipfinaloutput")) {
      if (SimFit_Verbose > -2) { message(""); }
      % print combined statistic
      _simultaneous_fit_comb_chi_sqr(s);
       % print histogram
      if (SimFit_Verbose_Histogram == 0) { ()=s.group_stats(); }
    }
    if (dtype != Void_Type) { return ret; }
    return;
  }
  catch UserBreakError: {
    if (SimFit_Verbose > -2) { message(""); }
    message("\nuser break: restoring simfit status");
    s.restore(; params);
  }
} %}}}

private define simultaneous_fit_group_stats() %{{{
%!%+    
%\function{simultaneous_fit.group_stats}
%\synopsis{prints a summary about the groups current fit-statistics}
%\usage{Integer_Type[] simultaneous_fit.group_stats();}
%\description
%    A histogram of the current red. chi-squares of all
%    groups is printed into the terminal. The histogram
%    ranges from the smallest to the largest found
%    red. chi-square. This range is shown as two ticmarks
%    below the x-Axis. The detailed statistics of each
%    group can be found in the .model.stat-field.
%    The returned array holds the numbers of groups
%    sorted by their red. chi-square, starting at the
%    worst.
%\seealso{printhplot}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: (s) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % get and merge the red. chi-squares of all groups
  variable chis = merge_struct_arrays(s.model.stat[[1:]]);
  if (any(chis.num_bins == 0)) {
    vmessage("error(%s): some groups have not been evaluated or fitted", _function_name);
    return;
  }
  chis = chis.statistic / (chis.num_bins - chis.num_variable_params);
  % create a histrogram
  variable lo, hi;
  (lo,hi) = linear_grid(min(chis), max(chis), 30);
  variable hist = histogram(chis, lo, hi);
  % printplot it
  printhplot(lo, hi, hist; W = length(lo), H = 5, yformat = "%d", ystrlen = 2);

  return array_sort(chis; dir = -1) + 1;
} %}}}

private define simultaneous_fit_filter_groups() %{{{
%!%+    
%\function{simultaneous_fit.filter_groups}
%\synopsis{return the group numbers matching a filter}
%\usage{simultaneous_fit.filter_groups([Integer_Type or String_Type filter]);}
%\description
%    Returns the group numbers which reduced chi-square
%    matches the given filter. The filter can be one of
%    the following three options:
%    1) no filter given - the group with the worst fit
%       statistic is returned
%    2) integer (n) given - sorts the groups after their
%       fit statistic (beginning at the worst) and returns
%       the first n groups
%    3) string given - has to be given in the format
%       "[operator][number]". Those group number are
%       returned which reduced chi-squares (chi) matches
%         chi [operator] [number]
%       The operators >, >=, <, <=, and == are possible.
%       Multiple of those rules can be specified if
%       separated with ','.
%
%    This function can be very useful if combined with
%    .select_groups
%\example
%    % initialize the simultaneous fit
%    simpi = simulteneous_fit();
%    ...
%    
%    % return the three worst fitted groups
%    grps = simpi.filter_groups(3);
%
%    % return those which have 1.7 <= chi^2_red <= 3
%    grps = simpi.filter_groups(">=1.7,<=3");
%
%    % directly select the worst group for fitting,
%    % plotting, etc.
%    simpi.select_groups(simpi.filter_groups());
%\seealso{simultaneous_fit.select_groups}
%!%-
{
  variable s, filter = NULL;
  switch (_NARGS)
    { case 1: (s) = (); }
    { case 2: (s, filter) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % get and merge the red. chi-squares of all groups
  variable chis = merge_struct_arrays(s.model.stat[[1:]]);
  if (any(chis.num_bins == 0)) {
    vmessage("error(%s): some groups have not been evaluated or fitted", _function_name);
    return;
  }
  chis = chis.statistic / (chis.num_bins - chis.num_variable_params);

  % return the worst group
  if (filter == NULL) {
    return where_max(chis) + 1;
  }
  % return the n worst groups
  if (typeof(filter) == Integer_Type) {
    return array_sort(chis; dir = -1)[[0:filter-1]] + 1;
  }
  % return matching user-defined filter
  filter = strchop(filter, ',', 0);
  variable i = [0:length(s.data)-1], n;
  _for n (0, length(filter)-1, 1) {
    variable rule = string_matches(filter[n], "\([^0-9]*\)\([0-9\.]*\)"R);
    if (length(rule) != 3 || rule[2] == "") {
      vmessage("error(%s): could not parse the filter %s", _function_name, filter[n]);
      return;
    }
    switch (rule[1])
      { case ">": i = i[where(chis[i] > atof(rule[2]))]; }
      { case ">=": i = i[where(chis[i] >= atof(rule[2]))]; }
      { case "<": i = i[where(chis[i] < atof(rule[2]))]; }
      { case "<=": i = i[where(chis[i] <= atof(rule[2]))]; }
      { case "==": i = i[where(chis[i] == atof(rule[2]))]; }
      { vmessage("error(%s): unkown operater '%s'", _function_name, rule[1]); return; }
  }
  return i+1;
} %}}}

private define simultaneous_fit_select_groups() %{{{
%!%+    
%\function{simultaneous_fit.select_groups}
%\synopsis{select the default groups to be worked on}
%\usage{simultaneous_fit.select_groups([Integer_Type[] groups]);}
%\description
%    Some functions within in the simultaneous structure
%    accept an optional array of group numbers, on which
%    they perform their taks. For instance, .fit_groups
%    either performs a fit for the given groups only. If
%    none are given, the default is to fit all groups.
%    This function, however, changes this default such
%    that the here given group(s) are used by default.
%    If no groups are given the default is set back to
%    all groups. To get the current set default you may
%    check the .model.current_groups variable (-1 means
%    all groups are used).
%!%-
{
  variable s, g;
  switch (_NARGS)
    { case 1: (s) = (); s.model.current_groups = -1; }
    { case 2: (s, g) = (); s.model.current_groups = g; }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;  
} %}}}

private define simultaneous_fit_eval_groups() %{{{
%!%+    
%\function{simultaneous_fit.eval_groups}
%\synopsis{evaluates the model for each (given) data group}
%\usage{Integer_Type simultaneous_fit.eval_groups([Integer_Type[] groups]);}
%\description
%    Loops over each (given) data group and evaulates the
%    model using only those parameters assigned to that
%    group. All other groups and parameters are excluded.
%    If no group is given, every group is evaluated.
%    Note, that the number of the first data group is 1
%    (not 0 as for arrays).
%    The displayed fit statistics are those of the
%    individual group only. The statistics are available
%    afterwards in the model.stat struct-array.
%
% All qualifiers are passed to eval_counts
%\seealso{simultaneous_fit.eval_global, eval_counts}
%!%-
{
  variable s, groups;
  switch (_NARGS)
    { case 1: (s) = ();
      groups = s.model.current_groups[0] == -1
             ? [1:length(s.data)]
             : s.model.current_groups; }
    { case 2: (s,groups) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % check given groups
  if (min(groups) < 1 || max(groups) > length(s.data)) { vmessage("error (%s): allowed data groups range from 1 to %d", _function_name, length(s.data)); return NULL; }

  %    if (fit_counts(&(s.model.stat[g]);; struct_combine(struct { fit_verbose = SimFit_Verbose }, __qualifiers)) == -1);
  variable ref = Ref_Type[length(groups)], fc;
  _for fc (0, length(groups)-1) { ref[fc] = &(s.model.stat[groups[fc]]); }
  fc = _simultaneous_fit_map_groups(s, groups, Integer_Type, &eval_counts, ref;;
                                    struct_combine(struct { fit_verbose = SimFit_Verbose }, __qualifiers));
 
  return min(fc); % <0 if any fit failed
} %}}}

private define simultaneous_fit_fit_groups() %{{{
%!%+    
%\function{simultaneous_fit.fit_groups}
%\synopsis{performs a fit of each (given) data group}
%\usage{Integer_Type simultaneous_fit.fit_groups([Integer_Type[] groups]);}
%\description
%    Loops over each (given) data group and performs a fit
%    of only those parameters assigned to that group. All
%    other groups and parameters are excluded from the fit.
%    In particular, all global parameters are fixed! If no
%    group is given, a fit for every group is performed.
%    Note, that the number of the first data group is 1
%    (not 0 as for arrays).
%    The displayed fit statistics are those of the
%    individual group only. The statistics are available
%    afterwards in the model.stat struct-array.
%
%    All qualifiers are passed to fit_counts.
%
%\seealso{simultaneous_fit.fit_global, fit_counts}
%!%-
{
  variable s, groups;
  switch (_NARGS)
    { case 1: (s) = ();
      groups = s.model.current_groups[0] == -1
             ? [1:length(s.data)]
             : s.model.current_groups; }
    { case 2: (s,groups) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % check given groups
  if (min(groups) < 1 || max(groups) > length(s.data)) { vmessage("error (%s): allowed data groups range from 1 to %d", _function_name, length(s.data)); return NULL; }

  %    if (fit_counts(&(s.model.stat[g]);; struct_combine(struct { fit_verbose = SimFit_Verbose }, __qualifiers)) == -1);
  variable ref = Ref_Type[length(groups)], fc;
  _for fc (0, length(groups)-1) { ref[fc] = &(s.model.stat[groups[fc]]); }
  fc = _simultaneous_fit_map_groups(s, groups, Integer_Type, &fit_counts, ref;;
                                    struct_combine(struct { fit_verbose = SimFit_Verbose }, __qualifiers));

  return min(fc); % <0 if any fit failed
} %}}}

private define simultaneous_fit_eval_global() %{{{
%!%+    
%\function{simultaneous_fit.eval_global}
%\synopsis{evaluates the model for all global parameters}
%\usage{Integer_Type simultaneous_fit.eval_global();}
%\description
%    All data groups are included within the evaulation
%    of the model, but the only free parameters are the
%    global parameters.
%    The displayed fit statistic includes all data, but
%    the global parameters only. The statistics are
%    available afterwards in the model.stat struct-array.
%
%    All qualifiers are passed to eval_counts.
%
%\seealso{simultaneous_fit.eval_groups, eval_counts}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: (s) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  return _simultaneous_fit_map_global(s, Integer_Type, &eval_counts, &(s.model.stat[0]);;
				      struct_combine(struct { fit_verbose = SimFit_Verbose }, __qualifiers));
} %}}}

private define simultaneous_fit_fit_global() %{{{
%!%+    
%\function{simultaneous_fit.fit_global}
%\synopsis{performs a fit of all global parameters}
%\usage{Integer_Type simultaneous_fit.fit_global();}
%\description
%    All data groups are included within the fit, but the
%    only free parameters are the global parameters. In
%    particular, all group parameters are fixed!
%    The displayed fit statistic includes all data, but
%    the global parameters only. The statistics are
%    available afterwards in the model.stat struct-array.
%
%    All qualifiers are passed to fit_counts.
%
%\seealso{simultaneous_fit.fit_group, fit_counts}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: (s) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  return _simultaneous_fit_map_global(s, Integer_Type, &fit_counts, &(s.model.stat[0]);;
				      struct_combine(struct { fit_verbose = SimFit_Verbose }, __qualifiers));
} %}}}

private define simultaneous_fit_fit_smart() %{{{
%!%+    
%\function{simultaneous_fit.fit_smart}
%\synopsis{combines group- and global-fits to achieve the best-fit}
%\usage{Integer_Type simultaneous_fit.fit_smart();}
%\qualifiers{
%    \qualifier{maxiter}{maximum number of iterations (default: 10)}
%    \qualifier{tol}{favored difference in chi square (default: 0.1)}
%    \qualifier{chatty}{output statistics for each iteration (default: 1)}
%}
%\description
%    By alternating between fitting the group parameters
%    and then the global parameters only, a best-fit is
%    tried to achieve. The fit is successful, if the
%    difference in delta chi square compared to the
%    previous iteration is less than the given tolerance.
%    The loop interrupts, if a maximum number of allowed
%    iterations is reached (the default number is low on
%    purpose).
%
%    Some tests showed that the final best-fit here is worse
%    than that achieved by 'fit_counts' (in the order of 10%
%    in reduced chi square). If a lot of data is fitted,
%    this method is, however, much faster than a simple
%    'fit_counts' (which might take days...).
%
%    All qualifiers not listed above are passed to fit_counts
%\seealso{simultaneous_fit.fit_global, simultaneous_fit.fit_groups}
%!%-
{
  variable s;
  switch (_NARGS)
    { case 1: (s) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % default values
  variable maxiter = qualifier("maxiter", 10);
  variable tol = qualifier("tol", .1);
  variable chatty = qualifier("chatty", 1);

  % initialize fit
  if (s.model.stat[0].num_bins == 0) {
    vmessage("error(%s): the global parameters have not been evaluated or fitted", _function_name);
    return;
  }
  variable last = s.model.stat[0].statistic;
  variable i = 0;

  % iterative loop
  do {
    % perform the fits
    ()=s.fit_groups(;; struct_combine(struct { fit_verbose = SimFit_Verbose }, __qualifiers));
    ()=s.fit_global(;; struct_combine(struct { fit_verbose = SimFit_Verbose }, __qualifiers));
    % compute statistics
    variable this = s.model.stat[0].statistic;
    variable change = last - this;
    last = this; i++;
    if (chatty) {
      vmessage(sprintf("iteration %%d: chi^2 = %%.%df, change = %%.%df", abs(int(log10(tol)))+1, abs(int(log10(tol)))+1), i, this, change);
    }
  } while (change >= tol && i < maxiter);
  return -(i == maxiter);
} %}}}

private define simultaneous_fit_fit_pars() %{{{
%!%+    
%\function{simultaneous_fit.fit_pars}
%\synopsis{runs fit_pars for the given data group}
%\usage{Struct_Type simultaneous_fit.fit_pars( Integer_Type group );}
%\altusage{Struct_Type simultaneous_fit.fit_pars( String_Type parname );}
%\qualifiers{
%\qualifier{frozen}{[=0] Perform fit_pars also for frozen Parameters.
%                   NOTE: just works if 'parname' is given!}
%\qualifier{force}{If group=0 (globals) force to perform fit_pars for all
%                  globals. Without this qualifier the parname of ONE global
%                  parameter is required!}
%}
%\description
%    Calculates the confidence level(s) of the parameter(s)
%    associated to the given data 'group' or 'parname'.
%    The function uses 'fit_pars' and returns its result.
%
%    Instead of the number of a data group, the name of
%    a parameter may be given instead. In that
%    case the confidence level of this parameter is
%    calculated only.
%
%    Note that for a global parameter the 'parname' has
%    to be given!
%
%    Also note that normally frozen parameters are ignored.
%    If this function is called with 'parname' and the
%    'frozen' qualifier is given, the confidence level
%    is calculated nevertheless.
%
%    It is important to keep in mind, that a 'fit_pars'
%    can lead to a new best fit. If the confidence levels
%    of a group as a whole are calculated a new best fit
%    of one of the group parameter is automatically taken
%    into account, but not if the confidence level of a
%    single parameter (e.g., using 'parname') is looked
%    at. This is a problem especially for the global
%    parameters. Use 'mpi_fit_pars' for your advantage!
%    
%    Note, that a usual 'fit_pars' will take a tremendous
%    amount of time (like 'fit_counts'). Thus, this
%    function accepts one data group only and it is
%    recommended to use only one parameter in addition.
%    It might be helpful to reduce the accuracy with which
%    fitting methods try to calculate chi square values
%    and/or parameters (see set_fit_method).
%
%    Further on, this approach is ideal to be used with
%    Torque (see 'simultaneous_fit.fit_pars_jobfile').
%
%    All qualifiers are also passed to fit_pars.
%
%\seealso{fit_pars, simultaneous_fit.list_groups,
%    simultaneous_fit.list_global, simultaneous_fit.fit_pars_jobfile}
%!%-
{
  variable frozen = qualifier("frozen",0);  if( qualifier_exists("frozen") ) frozen = 1;

  variable freep = freeParameters;
  variable paridx = Integer_Type[0];
  variable OUT = NULL, p;
  
  variable s, grp = NULL, parname = NULL;
  switch (_NARGS)
  { case 2: (s,grp) = ();
    if( typeof(grp) == String_Type ){
      % find according group:
      parname = grp;
      grp == NULL;
      ifnot( wherefirst( parname == s.model.global ) == NULL )
	grp = 0;
      else{
	_for p ( 1, length(s.model.group), 1 )
	  ifnot( wherefirst( parname == s.model.group[p-1] ) == NULL )
	    grp = p;
      }
      if( grp == NULL )
      { vmessage("error (%s): %s is no parameter!", _function_name, parname); return OUT; }
    }
  }
  %  { case 3: (s,grp,parname) = (); }
  { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  s.setrestore();

  try {
  % global parameter
  if (grp == 0) {
    if (parname == NULL ){
      ifnot( qualifier_exists("force") ){
	vmessage("error (%s): global parameter name required (or use qualifier 'force')", _function_name);
        return OUT;
      }
    }
    else if( wherefirst(_get_index(parname)==freeParameters) == NULL and frozen==0 ){
      vmessage("error (%s): %s is frozen! Set qualifier 'frozen' to perform confidence calculations!",
	       _function_name, parname);
      return OUT;
    }
    
    variable globalpar = array_map(Integer_Type, &_get_index, s.model.global);

    % freeze groups
    foreach p (freep) if (wherefirst(p == globalpar) == NULL) freeze(p);
    if (parname == NULL)
      paridx = freeParameters;
    else if( wherefirst(_get_index(parname)==freeParameters) != NULL or frozen==1 )
      paridx = _get_index(parname);
    % perform confidence calculation
    OUT = fit_pars( paridx ;; __qualifiers);
    s.model.stat[grp] = reduce_struct( eval_stat_counts(), "covariance_matrix");
    % restore free parameters
    thaw(freep);
  }
  % group parameter(s)
  else if( grp <= length(s.model.group) ) {
    variable grouppar  = array_map(Integer_Type, &_get_index, s.model.group[grp-1][where(_isnull(s.model.tieto[grp-1]))]);
    variable ids = array_flatten(list_to_array(s.data));
    variable excl = array_struct_field(get_data_info(ids), "exclude");

    % freeze all parameters except those associated to this data group (and are not tied)
    foreach p (freep) if (wherefirst(p == grouppar) == NULL) freeze(p);
    % only use given data group during fit
    exclude(ids);
    _for p (0, length(s.data[grp-1])-1, 1) ifnot (excl[wherefirst(all_data == s.data[grp-1][p])]) include(s.data[grp-1][p]);
    % perform confidence calculation
    if (parname == NULL)
      paridx = freeParameters;
    else if( wherefirst(_get_index(parname)==freeParameters) != NULL or frozen==1 )
      paridx = _get_index(parname);
    OUT = fit_pars( paridx ;; __qualifiers);
    s.model.stat[grp] = reduce_struct( eval_stat_counts(), "covariance_matrix");
    % restore free parameters and exclude state
    _for p (0, length(ids)-1, 1) if (excl[p] == 0) include(ids[p]);
    thaw(freep);
  }
  else
  { vmessage("error (%s): There is no group with idx %d!", _function_name, grp); return OUT; }  

  return OUT;
  }
  catch UserBreakError: {
    message("\nuser break: restoring simfit status");
    s.restore(; params);
  }
  catch AnyError: {
    message("\nsome error occured: consider calling .restore");
  }
} %}}}

private define _simultaneous_fit_steppar_fit() %{{{
{
  % Fit function for steppar replacing fit_counts
  variable s, grp;
  switch (_NARGS)
  { case 2: (s, grp) = (); }
  { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;
  
  if(grp==0){
    ()=s.fit_global();
  }
  else{
    ()=s.fit_groups(grp);
  }
  return s.fit_smart(;;__qualifiers);
}
%}}}

private define simultaneous_fit_steppar() %{{{
%!%+    
%\function{simultaneous_fit.steppar}
%\synopsis{runs steppar based on the simultanous fit logic}
%\usage{Struct_Type simultaneous_fit.steppar( String_Type parname );}
%\altusage{Struct_Type simultaneous_fit.steppar( Integer_Type idx );}
%\qualifiers{
%\qualifier{keys}{An additional structure is returned that can be used as FITS header keys}
%\qualifier{frozen}{[=0] Perform steppar also for frozen Parameters.}
%\qualifier{range}{[parmin,parmax] Stepping range. Default is the
%                   minimal/maximal allowed parameter value.}
%\qualifier{nsteps}{[=10] Number of steps the parameter 'par' is
%                   stepped from range[0] to range[1].}
%\qualifier{reset}{If given after each step the initial parameter set, which was
%                    valid before this function was called, is restored.}
%\qualifier{fit}{[=&fit_counts] Reference to the function running the fit algorithm.}
%\qualifier{fitargs}{Arguments required by the 'fit' function. See __push_list for
%                    format information!}
%\qualifier{rerun}{Reruns a stepping procedure based on results of a previous run,
%                    i.e., before each step the according parameter set of the previous
%                    run is loaded, e.g., to improve the results using another fit_method.}
%\qualifier{resume}{Missing steps in the given steppar-file will be calculated based
%                     on the previous steps, i.e., resuming the stepping where it was
%                     stopped.}
%\qualifier{check}{[=0] Before each step saved steppar informations of other stepped
%                        parameters are gathered and checked for a parameter set with a
%                        better chi2 (using steppar_get_bestparams). In case a better
%                        parameter set was found the stepping will be restarted.
%                        * ATTENTION: 'save' qualifier is required !!!
%                        * Gathered are steppar-files, which match the pattern
%                          given with 'save',e.g., if save is "steppar_PID001.fits"
%                          all files "steppar_PID???.fits" are globed!
%                          Note that the affix "_PID???" is automaticcaly appended to
%                          the save string if it does not exist (see 'save' qualifier!)
%                        * The Integer 'check' is set to is the maximal number of
%                          restarts.
%                        * A parameter set is considered better if
%                          chi2_new < chi2_init * ( 1 - dchi2 )
%                        * NOTE: Parameter grouping is possible (see steppar_get_bestparams)
%                        }
%\qualifier{dchi2}{[=0.1] Fractional limit for the chi2 of a new parameter set to be
%                  considered as a better parameter set (see 'check'):
%                  chi2_new < chi2_init * ( 1 - dchi2 ).}
%\qualifier{save}{After each stepping the result is saved in the file given
%                   (as String_Type) with this qualifier. The chi2 of undone steps
%                   are set to 0.! Also note that the filename is appended with
%                   the parameterindex: 'steppar.fits' -> 'steppar_PID001.fits'}
%\qualifier{force}{Forces to overwrite an existing file given with 'save'.}
%\qualifier{chatty}{Prints fitting information.}
%}
%\description
%      This function is calling 'steppar' with an adjusted fitting function:
%      fit = &_simultaneous_fit_steppar_fit which first calls either %.fit_groups
%      or %.fit_global depending on the affiliation of the given parameter.
%      Afterwards %.fit_smart is called (with the passed qualifiers).
%      For more detailed information see the help of 'steppar'!
%
%      Qualifiers are also passed to 'steppar' & '%.fit_smart'!
%
%\seealso{%.fit_smart, steppar}
%!%-
{
  variable s, parname;
  switch (_NARGS)
  { case 2: (s, parname) = (); }
  { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;
  
  try {
    
    %%% Parameter check:
    % check for parameter existence (if multiple, chosing first)
    variable parinfo = get_par_info(parname)[0];
    if( parinfo == NULL ){
      vmessage("ERROR <%s>: No parameter to given parname=%s found!",
	       _function_name,parname );
      return NULL;
    }
    parname = parinfo.name;

    % find group of parameter
    variable i = 0, j;
    variable grp = NULL;
    while( grp == NULL ){
      j = where( s.model.group[i] == parname );
      i++;
      ifnot( length(j) == 0 ){
	grp = i;
	j = j[0];
      }
    }
    % check if global parameter
    if( any( s.model.tieto[grp-1][j] == s.model.global ) ){
      vmessage("WARNING <%s>: '%s' is tied to global parameter '%s' (now used)!",
	       _function_name, parname, s.model.tieto[grp-1][j] );
      grp = 0;
      parname == s.model.tieto[grp-1][j];
      parinfo = get_par_info(parname);
    }

    variable params = get_params;
    variable fitargs = { s, grp };
    variable qualis = struct{
      fit=&_simultaneous_fit_steppar_fit,
      fitargs=fitargs
    };
    if( grp != 0 ){
      qualis = struct_combine( qualis, struct{ groups = s.model.group[grp-1] } );
    }
    
    return steppar( parname ;; struct_combine( qualis, __qualifiers ));
  }
  catch UserBreakError: {
    message("\nuser break: restoring simfit status");
    set_params(params);
    return;
  }
  catch AnyError: {
    message("\nsome error occured: consider calling .restore");
    return;
  }
} %}}}

private define simultaneous_fit_fit_pars_jobfile() %{{{
%!%+    
%\function{simultaneous_fit.fit_pars_jobfile}
%\synopsis{writes a Torque-jobfile to calculate confidence levels}
%\usage{simultaneous_fit.fit_pars_jobfile(String_Type jobfile, scriptfile);}
%\qualifiers{
%    \qualifier{force}{overwrites an existing jobfile or script
%                  (default: don't overwrite)}
%    \qualifier{writescript}{writes a template for the script into
%                  the given filename (file will be overwritten)}
%}
%\description
%    Writes a Torque 'jobfile' to calculate the confidence
%    levels of all (free) parameters. Therefore, the filename
%    of the script for loading the data, defining the model
%    and running simultaneous_fit.fit_pars is mandatory.
%    This script will be called by each job with a
%    command line argument, which is either
%      - the data group index, for which the confidence levels
%        should be calculated for or
%      - the parameter name to compute the level
%        for a single parameter in the global group.
%        This means that the parameters of a data group are
%        treated in parallel by multiple jobs.
%    In case of a global parameter, the data group argument
%    is omitted and the name of the global parameter is
%    given instead. The command lines arguments can be
%    accessed using '__argv' and '__argc'.
%    
%    If the qualifier 'writescript' exists, a template for
%    the script described above will be written into the
%    given 'scriptfile'name. This has to be modified
%    afterwards to ensure compatibility and should be
%    treated as a support!
%    Warning: in case of an already existing filename this
%             file will be overwritten!
%\seealso{simultaneous_fit.fit_pars}
%!%-
{
  variable s, jobfile, scriptfile;
  switch (_NARGS)
    { case 3: (s,jobfile,scriptfile) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;


  % QUALIFIER JOBFILES
  variable walltime = qualifier("walltime","00:01:00");
  variable jobname  = qualifier("jobname","sim_fitpars");
  variable logpath  = qualifier("logpath",jobname);
  
  % Check TYPE
  if (typeof(jobfile) != String_Type){ vmessage("error (%s): given jobfile has to be a string", _function_name); return; }
  if (typeof(scriptfile) != String_Type){ vmessage("error (%s): given scriptfile has to be a string", _function_name); return; }
  
  % Scriptfiles with arguments
  variable script = strtok(scriptfile)[0];
  variable nargs  = length(strtok(scriptfile)[[1:]]);
  
  % Check files
  if (access(jobfile, F_OK) == 0 && qualifier_exists("force") == 0)
    { vmessage("error (%s): jobfile '%s' already exists", _function_name, jobfile); return; }
  if ( not(qualifier_exists("writescript")) && access(script, F_OK) == -1 )
  { vmessage("error (%s): scriptfile '%s' does not exists", _function_name, script); return; }
  if (qualifier_exists("writescript") && access(script, F_OK) == 0 && qualifier_exists("force") == 0)
  { vmessage("error (%s): scriptfile '%s' already exists", _function_name, script); return; }

  
  % get current user
  variable user, pp = popen("whoami", "r");
  ()=fgets(&user, pp);
  ()=pclose(pp);
  user = strreplace(user, "\n", "");

  % get current directory
  variable pwd;
  pp = popen("pwd", "r");
  ()=fgets(&pwd, pp);
  ()=pclose(pp);
  pwd = strreplace(pwd, "\n", "");
  
  % CHECK LOGPATH
  if( is_substr( logpath,pwd ) == 0 )
    logpath = pwd+"/"+logpath;
  if( logpath[-1] == '/' )
    logpath = logpath[[:-2]];
  mkdir(logpath);

 
  % build jobfile
  variable job = "#!/bin/bash\n#\n";
  job += "#PBS -S /bin/bash -V\n";
  job += "#PBS -t 0-%d%%10000\n";
  job += "#PBS -l nodes=1\n";
  job += "#PBS -l arch=x86_64\n";
  job += sprintf("#PBS -l walltime=%s\n", walltime);
  job += sprintf("#PBS -N %s\n",jobname);
  job += sprintf("#PBS -o %s/%s_groups.out\n", logpath, jobname);
  job += sprintf("#PBS -e %s/%s_groups.err\n", logpath, jobname);
  job += sprintf("\ncd %s\n", pwd);

  variable n = 0, g;
  % loop over each data group
  _for g (1, length(s.data), 1) {
    % a fit_pars for all parameters of the group
    job += sprintf("COMMAND[%d]=\"isis-script %s %d\"\n", n, scriptfile, g);
    n++;
  }
  % add global parameters to jobfile
  _for g (0, length(s.model.global)-1, 1) {
    job += sprintf("COMMAND[%d]=\"isis-script %s %s\"\n", n, scriptfile, s.model.global[g]);
    n++;
  }
  job += "/usr/bin/nice -n +15 ${COMMAND[$PBS_ARRAYID]}";
  job = sprintf(job, n-1);

  % write jobfile
  variable fp = fopen(jobfile, "w+");
  ()=fputs(job, fp);
  ()=fclose(fp);


  % QUALIFIER SCRIPTFILE
  variable parpath = qualifier("parpath",logpath);
  % CHECK PARPATH
  if( is_substr( parpath,pwd ) == 0 )
    parpath = pwd+"/"+parpath;
  if( parpath[-1] == '/' )
    parpath = parpath[[:-2]];
  mkdir(parpath);
  
  % eventually, write script template
  variable wscrpt;
  if (qualifier_exists("writescript")) {
    wscrpt  = "require(\"isisscripts\");\n";
    wscrpt += "% TIME LOG\n";
    wscrpt += "tic;vmessage(\"%s : Calling: '%s'\",time,strjoin(__argv,\" \"));\n\n";
    
    wscrpt += "%%% initialize the simultaneous fit, adding the %%%\n";
    wscrpt += "%%% data and defining/loading the model here    %%%\n\n";
    
    wscrpt += "% get command line arguments (last argument has to be the group index or parname!)\n";
    wscrpt += "variable grp = __argv[-1];\n";
    wscrpt += "if( atoi(grp) != 0 ) grp = atoi(grp);\n";

    wscrpt += "\n%%% set the directory for the output fits-files here %%%\n";
    wscrpt += "%%% make sure that this directory exists!            %%%\n";
    wscrpt += sprintf("variable parpath = %s;\n",parpath);
    wscrpt += "% check on existence of parpath\n";
    wscrpt += "if (access(parpath, F_OK) != 0) {\n";
    wscrpt += "  vmessage(\"error: directory %s does not exist, aborting...\", parpath);\nexit;\n}\n";

    wscrpt += "\n% calculate uncertainties\n";
    wscrpt += "%%% replace 'simfit' by the name of your simultaneous-fit-structure %%%\n";
    wscrpt += "variable fpars = simfit.fit_pars( grp ; basefilename = sprintf(\"%s/"+jobname+"_%04d.fits\", parpath, grp));\n";

    wscrpt += "\n% TIME LOG\n";
    wscrpt += "variable elaptime = toc;\n";
    wscrpt += "vmessage(\"%s : Finished '%s' after %d s < %d %s\",\n";
    wscrpt += "time,strjoin(__argv,\" \"),nint(elaptime),time_array(elaptime)[-1]+1,\n";
    wscrpt += "[\"sec\",\"min\",\"h\",\"d\",\"y\"][length(time_array(elaptime))-1]  );";
    
    
    % write scriptfile
    fp = fopen(script, "w+");
    ()=fputs(wscrpt, fp);
    ()=fclose(fp);
  }
} %}}}

#ifeval __get_reference("mpi_fit_pars")!=NULL
private define simultaneous_fit_mpi_fit_pars() %{{{
%!%+    
%\function{simultaneous_fit.mpi_fit_pars}
%\synopsis{runs mpi_fit_pars for the given data group}
%\usage{Struct_Type simultaneous_fit.mpi_fit_pars( Integer_Type group );}
%\description
%    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
%    ATTENTION: This function does not work in an open
%               isis session, as it links to 'mpi_fit_pars'!
%               It is supposed to be used in a script file
%               submitted as a torque job only!
%    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
%
%    Calculates the confidence levels of the FREE parameters
%    associated to the given data group.
%    This function uses the function 'mpi_fit_pars' and its
%    result is returned.
%
%    The argument <group> represents the index  of the
%    data GROUP, of which the confidence levels are supposed
%    to be calculated.
%    group = 0 corresponds to the 'group' of the GLOBAL
%    parameters!
%    
%    Note, that a usual 'fit_pars' will take a tremendous
%    amount of time (like 'fit_counts'). Thus, this
%    function accepts one data group only and it is
%    recommended to use only one parameter in addition.
%    It might be helpful to reduce the accuracy with which
%    fitting methods try to calculate chi square values
%    and/or parameters (see set_fit_method).
%
%    Further on, this approach is ideal to be used with
%    Torque (see 'simultaneous_fit.mpi_fit_pars_jobfiles').
%
%    All qualifiers are also passed to mpi_fit_pars
%
%\seealso{fit_pars, mpi_fit_pars, simultaneous_fit.list_groups,
%    simultaneous_fit.list_global, simultaneous_fit.mpi_fit_pars_jobfiles}
%!%-
{
  variable s, grp;
  switch (_NARGS)
  { case 2: (s,grp) = ();}
  { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;
  
  variable err1 = sprintf("<%s> ERROR: #freeParameters does not match %%.model.stat.num_variable_params, i.e.,\n"
			  +"since the last fit parameters have been frozen/thawed!\n"
			  +"Please perform a fit_counts or %%.fit_smart/groups/global!",_function_name);
  
  variable freep = freeParameters;
  variable OUT = NULL, p;
  % global parameter (grp==0)
  if (grp == 0) {
    variable global = array_map(Integer_Type, &_get_index, s.model.global);
    % freeze groups
    foreach p (freep) if (wherefirst(p == global) == NULL) freeze(p);
    % perform confidence calculation ON FREE PARAMETERS
    if( s.model.stat[grp].num_variable_params != length(freeParameters) )
    { vmessage(err1); }
    else{
      OUT = mpi_fit_pars( freeParameters ;; __qualifiers);
      s.model.stat[grp] = reduce_struct( eval_stat_counts(), "covariance_matrix");
    }
    % restore free parameters
    thaw(freep);
  }
  % group parameter(s)
  else {
    variable ids = array_flatten(list_to_array(s.data)); 
    variable excl = array_struct_field(get_data_info(ids), "exclude");
    % freeze all parameters except those associated to this data group (and are not tied)
    variable grouppar = array_map(Integer_Type, &_get_index, s.model.group[grp-1][where(_isnull(s.model.tieto[grp-1]))]);
    foreach p (freep) { if (wherefirst(grouppar == p) == NULL) { freeze(p); } }
    % only use given data group during fit
    exclude(ids);
    _for p (0, length(s.data[grp-1])-1, 1) { ifnot (excl[wherefirst(all_data == s.data[grp-1][p])]) { include(s.data[grp-1][p]); } }
    % perform confidence calculation
    if( s.model.stat[grp].num_variable_params != length(freeParameters) )
    { message(err1); }
    else{
      OUT = mpi_fit_pars( freeParameters ;; __qualifiers);
      s.model.stat[grp] = reduce_struct( eval_stat_counts(), "covariance_matrix");
    }
    % restore free parameters and exclude state
    _for p (0, length(ids)-1, 1) if (excl[p] == 0) include(ids[p]);
    thaw(freep);
  }
  return OUT;
} %}}}

private define simultaneous_fit_mpi_fit_pars_jobfiles() %{{{
%!%+    
%\function{simultaneous_fit.mpi_fit_pars_jobfiles}
%\synopsis{writes a Torque-jobfile to calculate confidence levels}
%\usage{simultaneous_fit.mpi_fit_pars_jobfiles(String_Type jobfile, scriptfile);}
%\qualifiers{
%\qualifier{walltime: }{[="00:01:00"] Torque walltime for global & group
%                     jobfile}
%\qualifier{wt_global: }{[=walltime] Torque walltime for global jobfile}
%\qualifier{wt_groups: }{[=walltime] Torque walltime for groups jobfile}
%\qualifier{jobname: }{[="simpi_fitpars"] Name of Torque job (& files!)}
%\qualifier{logpath: }{[=pwd+jobname] Directory for log files}
%\qualifier{force: }{overwrites an existing jobfile or script
%                  (default: don't overwrite)}
%\qualifier{writescript: }{ writes a template for the script into
%                  the given filename (file will be overwritten)}
%}
%\description
%    Writes a Torque 'jobfile' to calculate the confidence
%    levels of all parameters. Therefore, the filename of
%    the script for loading the data, defining the model
%    and running simultaneous_fit.fit_pars is mandatory.
%    This script will be called by each job with a
%    command line arguments, which represents the data
%    group, for which the confidence levels should be
%    calculated for. The command line arguments can be
%    accessed using '__argv' and '__argc'.
%
%    In case the 'scriptfile' requires additional command
%    line arguments add them to the 'scriptfile'-string,e.g.,
%    like "scriptfile.sl arg1 arg2", and make sure these
%    are asigned correctly!
%    
%    If the qualifier 'writescript' exists, a template for
%    the script described above will be written into the
%    given 'scriptfile'name. This has to be modified
%    afterwards to ensure compatibility and should be
%    treated as a support!
%    Warning: in case of an already existing filename this
%             file will be overwritten!
%\seealso{simultaneous_fit.fit_pars}
%!%-
{
  variable s, jobfile, scriptfile;
  switch (_NARGS)
  { case 3: (s,jobfile,scriptfile) = (); }
  { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;
  
  % QUALIFIER JOBFILES
  variable walltime = qualifier("walltime","00:01:00");
  variable wt_global= qualifier("wt_global",walltime);
  variable wt_groups= qualifier("wt_groups",walltime);
  variable jobname  = qualifier("jobname","simpi_fitpars");
  variable logpath  = qualifier("logpath",jobname);
  
  % Check TYPE
  if (typeof(jobfile) != String_Type){ vmessage("error (%s): given jobfile has to be a string", _function_name); return; }
  if (typeof(scriptfile) != String_Type){ vmessage("error (%s): given scriptfile has to be a string", _function_name); return; }

  variable N = length(s.model.group), g;
  variable iformat = "_%04d";
  
  % Jobfiles
  variable jobfile_ext   = path_extname(jobfile);
  if( jobfile_ext == "" ) { jobfile_ext = ".job"; }
  variable jobfiles = array_map(String_Type,&sprintf,path_sans_extname(jobfile)+iformat+jobfile_ext,[0:N]);

  % Scriptfiles with arguments
  variable script = strtok(scriptfile)[0];
  variable nargs  = length(strtok(scriptfile)[[1:]]);
  
  % Check files
  _for g ( 0, N, 1 ){
    if (access(jobfiles[g], F_OK) == 0 && qualifier_exists("force") == 0)
    { vmessage("error (%s): jobfile '%s' already exists", _function_name, jobfiles[g]); return; }

  }
  if ( not(qualifier_exists("writescript")) && access(script, F_OK) == -1 )
  { vmessage("error (%s): scriptfile '%s' does not exists", _function_name, script); return; }
  if (qualifier_exists("writescript") && access(script, F_OK) == 0 && qualifier_exists("force") == 0)
  { vmessage("error (%s): scriptfile '%s' already exists", _function_name, script); return; }
  
  % get current user
  variable user, pp = popen("whoami", "r");
  ()=fgets(&user, pp);
  ()=pclose(pp);
  user = strreplace(user, "\n", "");

  % get current directory
  variable pwd = getcwd[[:-2]];
  
  % CHECK LOGPATH
  if( is_substr( logpath,pwd ) == 0 )
    logpath = pwd+"/"+logpath;
  if( logpath[-1] == '/' )
    logpath = logpath[[:-2]];
  ()=mkdir(logpath);
  
  % build jobfiles
  variable job;
  variable nodes,index;

  % loop over each data group
  _for g (0, N, 1) {
    
    % finde & test number of nodes[=#freeParameters in each group] (MUST be the same)
    nodes = s.model.stat[g].num_variable_params;

    index = sprintf(iformat,g);
    
    job = "#!/bin/bash\n#\n";
    job += "#PBS -S /bin/bash -V\n";
    job += sprintf("#PBS -l nodes=%d\n",nodes);
    %job += "#PBS -l arch=x86_64\n";
    job += sprintf("#PBS -l walltime=%s\n", g==0? wt_global : wt_groups );
    job += sprintf("#PBS -N %s%s\n",jobname,index);
    job += sprintf("#PBS -o %s/%s%s.out\n", logpath, jobname, index);
    job += sprintf("#PBS -e %s/%s%s.err\n", logpath, jobname, index);
    job += sprintf("\ncd %s\n", pwd);

    % add group parameter to job file
    job += sprintf("COMMAND=\"mpiexec isis-script %s --grp=%d\"\n", scriptfile, g);
    job += "/usr/bin/nice -n +15 ${COMMAND}";
    
    % write GROUP jobfile
    variable fp = fopen( sprintf("%s",jobfiles[g]), "w+");
    ()=fputs(job, fp);
    ()=fclose(fp);
  }

  % QUALIFIER SCRIPTFILE
  variable parpath = qualifier("parpath",logpath);
  % CHECK PARPATH
  if( is_substr( parpath,pwd ) == 0 )
    parpath = pwd+"/"+parpath;
  if( parpath[-1] == '/' )
    parpath = parpath[[:-2]];
  ()=mkdir(parpath);
  
  % eventually, write script template
  variable wscrpt;
  if (qualifier_exists("writescript")) {
    wscrpt  = "require(\"isisscripts\");\n";
    wscrpt += "% TIME LOG\n";
    wscrpt += "tic;vmessage(\"%s : Calling: '%s'\",time,strjoin(__argv,\" \"));\n\n";
    
    wscrpt += "%%% initialize the simultaneous fit, adding the %%%\n";
    wscrpt += "%%% data and defining/loading the model here    %%%\n\n";
    
    wscrpt += "% get command line arguments (last argument has to be the group index!)\n";
    wscrpt += "variable grp = atoi(get_arg_struct().grp);\n";

    wscrpt += "\n%%% set the directory for the output fits-files here %%%\n";
    wscrpt += "%%% make sure that this directory exists!            %%%\n";
    wscrpt += sprintf("variable parpath = \"%s\";\n",parpath);
    wscrpt += "% check on existence of parpath\n";
    wscrpt += "if (access(parpath, F_OK) != 0) {\n";
    wscrpt += "  vmessage(\"error: directory %s does not exist, aborting...\", parpath);\nexit;\n}\n";

    wscrpt += "\n% calculate uncertainties\n";
    wscrpt += "%%% replace 'simfit' by the name of your simultaneous-fit-structure %%%\n";
    wscrpt += "variable fpars = simfit.mpi_fit_pars( grp ; dir=parpath,";
    wscrpt += "basefilename = sprintf(\""+jobname+iformat+".fits\", grp));\n";

    wscrpt += "\n% TIME LOG\n";
    wscrpt += "variable elaptime = toc;\n";
    wscrpt += "vmessage(\"%s : Finished '%s' after %d s < %d %s\",\n";
    wscrpt += "time,strjoin(__argv,\" \"),nint(elaptime),time_array(elaptime)[-1]+1,\n";
    wscrpt += "[\"sec\",\"min\",\"h\",\"d\",\"y\"][length(time_array(elaptime))-1]  );";
    
    
    % write scriptfile
    fp = fopen(script, "w+");
    ()=fputs(wscrpt, fp);
    ()=fclose(fp);
  }
} %}}}
#endif

private define simultaneous_fit_fit_pars_collect() %{{{
%!%+    
%\function{simultaneous_fit.fit_pars_collect}
%\synopsis{reads the results of a former fit_pars or mpi_fit_pars
%    and checks for a new best-fit}
%\usage{Struct_Type simultaneous_fit.fit_pars_collect(String_Type FITSpattern);}
%\qualifiers{
%    \qualifier{chitol}{how much a delta chi-square  has to be
%                     to actually consider this fit as a new
%                     best-fit or a worse-fit (default: 1e-10)}
%    \qualifier{setnewbestpars}{sets the parameters to those of a
%                     possible new best-fit (see below)}
%    \qualifier{bestpars}{reference to a variable where all found
%                     parameter values and other helpful
%                     information is returned}
%    \qualifier{parspriority}{in case parameters of a new best-fit
%                     are set, this qualifiers decides
%                     whether the changed "global"- or
%                     "group"-parameters are set (both are
%                     not possible!) (default: "global")}
%    \qualifier{PARpattern}{the string which replaces "_conf.fits"
%                     to match the parameter file
%                     (default: "_best.par")}
%    \qualifier{silent}{suppress warning messages (BIT wise)
%                     1: "fit ... is worse"
%                     2: "new best-fit ... found"
%                     4: "no uncertainties ... found"
%                     8: "setting new ... re-fit ..."}
%}
%\description
%    If the uncertainties of a simultaneous fit have been
%    calculated using Torque-jobfiles (e.g., by using the
%    mpi_fit_pars_jobfile function), the results will be
%    saved in several FITS-files.
%    This function reads all FITS-files matching the given
%    'FITSpattern' and returns a single structure containing
%    all results (similar to a single fit_pars structure).
%    Furthermore, if a new best-fit was found during one
%    of the calculations, a warning message will be shown.
%    In case of a crashed calculation (resulting in missing
%    uncertainties) a warning message will be shown as well.
%
%    If the qualifier "setnewbestpars" is set, the parameters
%    of a new best-fit will be set automatically. However,
%    new global parameters causes the group parameters to be
%    fitted again and vise versa. For this reason, new group
%    parameters are discarded in favor for new global ones
%    (to change this priority use the 'parspriority'
%    qualifier).
%
%    Note, that the par-files saved by fit_pars are needed to
%    set the remaining parameters in case of a new best-fit!
%\seealso{simultaneous_fit.mpi_fit_pars_jobfile, simultaneous_fit.fit_pars_jobfile}
%!%-
{
  variable s, pattern;
  switch (_NARGS)
    { case 2: (s,pattern) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  variable silent = qualifier("silent", 0);
  variable chitol = qualifier("chitol", 1e-10);
  % find all FITS-files
  variable files = glob(pattern);
  files = files[ array_sort(files) ];
  if (length(files) == 0) { vmessage("error (%s): did not find any FITS-file", _function_name); return NULL; }

  % loop over all files
  variable file, result = NULL;
  variable newbest = struct { group = Integer_Type[0], file = String_Type[0], params = Struct_Type[0], deltachi = Double_Type[0], missing = Struct_Type[0], statistic = Double_Type[0], isbetter = Integer_Type[0] };
  foreach file (files) {
    % load FITS-file
    variable fits = fits_read_table(file);
    % merge with results
    result = result == NULL ? fits : merge_struct_arrays([result, fits]);
    % check if FITS-file contains global parameters
    variable par, grp = 1;
    foreach par (fits.name) { if (grp && sum(array_map(Integer_Type, &is_substr, s.model.global, par)) > 0) { grp = 0; } }
    % if no global parameters are found, find the corresponding data group
    if (grp) {
      grp = Integer_Type[length(fits.name)];
      % loop over each parameter in FITS-file and find the corresponding group (black one-line magic)
      _for par (0, length(fits.name)-1) {
	grp[par] = wherefirst(array_map(Double_Type, &sum, array_map(Array_Type, &is_substr, s.model.group, fits.name[par])) == 1) + 1;
      }
      grp = int(array_unique(grp));
      % something weird happened if parameter is found in multiple groups...
      if (length(grp) != 1) {
        vmessage("error (%s): FITS-file %s contains parameters of several: groups %s", _function_name, file, strjoin(array_map(String_Type, &sprintf, "%d", grp), ","));
        return NULL;
      }
      grp = grp[0];
    }
    % compare chi-square values of that group (0 = global paramater)
    variable deltachi = fits_read_key(file, "chi2") - s.model.stat[grp].statistic;
    if (deltachi > chitol) {
      ifnot (silent & 1) {
	vmessage("warning (%s): fit during fit_pars of group %d is worse (delta chisqr = %.2f)", _function_name, grp, deltachi);
      }
    }
    else if (deltachi < -chitol) { % new best-fit found
      ifnot (silent & 2) {
	vmessage("warning (%s): new best-fit of %s found (delta chisqr = %.2f)", _function_name, grp == 0 ? "a global parameter" : sprintf("datagroup %d", grp), deltachi);
      }
    }
    % load parameters of the fit
    variable parfile = strreplace(file, "_conf.fits", qualifier("PARpattern", "_best.par"));
    if (access(parfile, F_OK) == 0) {
      variable newpars = NULL;
      try { 
        newpars = ascii_read_table(parfile, [{"%d","idx"},{"%s","name"},{"%d","tieto"},{"%d", "freeze"},{"%f","value"}]; startline = 3);
      }
      catch AnyError: {
        vmessage("warning (%s): could not parse parameter file %s, cannot set new best-fit parameters!", _function_name, parfile);
      }
      if (newpars != NULL) {
        % remember best-fit
        newbest = merge_struct_arrays([newbest, struct {
          group = grp, file = file, params = newpars, deltachi = deltachi, missing = NULL, statistic = 0, isbetter = int(deltachi < -chitol)
        }]);
        % load statistics of best-fit
        variable txtfile = strreplace(file, "_conf.fits", qualifier("TXTpattern", "_conf.txt"));
        % remember new statistics
        newbest.statistic[-1] = fits_read_key(file, "chi2");
      }
    } else { vmessage("warning (%s): parameter file %s not found, cannot set new best-fit parameters!", _function_name, parfile); }
  }

  % check on missing parameters
  foreach par ([array_flatten(s.model.group), s.model.global]) {
    newbest.missing = String_Type[0];
    if (get_par_info(par).freeze == 0 and wherefirst(result.name == par) == NULL) {
      ifnot (silent & 4) { vmessage("warning (%s): no uncertainties of parameter %s found!", _function_name, par); }
      newbest.missing = [newbest.missing, par];
    }
  }
  % eventually return info structure
  if (qualifier_exists("bestpars")) {
    @(qualifier("bestpars")) = newbest;
  }

  % eventually set new best parameters
  variable newbesti = where(newbest.isbetter); % better groups
  if (qualifier_exists("setnewbestpars") && length(newbesti) > 0) {
    variable isgl = wherefirst(newbest.group[newbesti] == 0); % better globals found?
    variable prio = qualifier("parspriority", "global") == "global";
    % set global parameters or...
    if (isgl != NULL && prio == 1) {
      isgl = newbesti[isgl];
      ifnot (silent & 8) { message("setting new GLOBAL parameters, please re-fit the groups"); }
      _for par (0, length(newbest.params[isgl].idx)-1) {
	if (newbest.params[isgl].freeze[par] == 0) {
	  set_par(newbest.params[isgl].name[par], newbest.params[isgl].value[par]);
	}
      }
      if (newbest.statistic[isgl] > 0) {
        s.model.stat[0].statistic = newbest.statistic[isgl];
      }
    }
    % ..set group parameters instead
    else {
      ifnot (silent & 8) { message("setting new GROUP parameters, you might re-fit the globals"); }
      foreach grp (newbesti) {
	if (newbest.group[grp] != 0) {
          _for par (0, length(newbest.params[grp].idx)-1) {
            if (newbest.params[grp].freeze[par] == 0) {
              set_par(newbest.params[grp].name[par], newbest.params[grp].value[par]);
	    }
          }
          if (newbest.statistic[grp] > 0) {
            s.model.stat[newbest.group[grp]].statistic = newbest.statistic[grp];
	  }
	}
      }
    }
  }
  
  return result;
} %}}}

private define simultaneous_fit_fit_pars_run_job() %{{{
%!%+    
%\function{simultaneous_fit.fit_pars_run_job}
%\synopsis{submits the (MPI-)fit_pars-jobfiles until no better fit is found}
%\usage{Struct_Type simultaneous_fit.fit_pars_run_job(
%                  String_Type or String_Type[] globaljobs, groupjobs,
%                  String_Type FITSpattern, String_Type or Ref_Type save
%                );}
%\qualifiers{
%    \qualifier{first}{submits either the jobs for the global- (=0)
%              or the group-parameters (=1, default) first}
%    \qualifier{maxiter}{maximum number of iterations (default: 3)}
%    \qualifier{wait}{seconds to wait before it is checked if the
%              Torque-jobs have completed (default: 10)}
%}
%\description
%    Usually, the uncertainties of the group- and global-
%    parameters should be calculated using Torque-jobs.
%    However, if one job finds a new best-fit of a, e.g.,
%    group parameter, the uncertainties of the global
%    parameters have to be re-calculated (and vise versa).
%    Note, that the calculations of remaining groups in
%    that case, which might still be running, don't have
%    to be canceled, because the global parameters might
%    stay the same even a better group parameter is found!
%
%    This function submits the Torque-jobs for different
%    kinds of parmeters (given as filename-patterns or the
%    filename(s) for the 'globaljobs' and 'groupjobs')
%    until no further best-fit has been found. By default,
%    the group jobs are submitted first. If a new best-fit
%    was found, the internal .save function is called with
%    the given 'save'-filenam to ensure that the current
%    (better) parameters can be loaded by the  re-submitted
%    jobs. If the 'save'-parameter is a reference to a
%    function, this one will be called instead of saving
%    the fit automatically. This function does not get any
%    parameters and should save parameter AND the model
%    field of the simultaneous fit.
%    Once no better fit is found a structure similar to
%    that of fit_pars is returned.
%
%    As a protection for an (infinite) loop, the above
%    procedure is repeated only a few times (adjustable
%    using the 'maxiter' qualifier).
%
%    IMPORTANT:
%    This function is only a control function! It has to
%    know the current parameters and dataset logic, but
%    does not require any CPU power! Thus, to free memory,
%    the complete loaded DATA, ARFs, and RMFs are DELETED!
%    So you CANNOT work with your data afterwards! If you
%    agree with that procedure please set the 'agree'-
%    qualifier. Otherwise, the function will exit.
%\seealso{simultaneous_fit.fit_pars_jobfile}
%!%-
{
  variable s, glbjob, grpjob, pattern, savefun;
  switch (_NARGS)
  { case 5: (s,glbjob,grpjob,pattern,savefun) = (); }
  { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % user has to agree on deleting all data
  ifnot (qualifier_exists("agree")) { vmessage("error (%s): read the help first", _function_name); return; }
  ifnot (typeof(savefun) == String_Type || typeof(savefun) == Ref_Type) {
    vmessage("error (%s): 'save' has to be a filename or a reference to a function", _function_name); return;
  }
  % replace all data by dummies to save memory, since this ISIS
  % instance should just be a control job, so no resources needed
  variable adata = @all_data, ad;
  delete_data(all_data);
  foreach ad ([1:max(adata)]) { ()=define_counts(1, 2, 3, 4); }
  delete_data(complement(all_data, adata)+1);
  delete_arf(all_arfs);
  delete_rmf(all_rmfs);

  variable maxiter = qualifier("maxiter", 3);
  
  % find all job-files
  variable jobs = {
    typeof(glbjob) != Array_Type ? glob(glbjob) : glbjob,
    typeof(grpjob) != Array_Type ? glob(grpjob) : grpjob
  };
  if (length(jobs[0]) == 0) { vmessage("error (%s): no global jobfiles found!", _function_name); return; }
  if (length(jobs[1]) == 0) { vmessage("error (%s): no group jobfiles found!", _function_name); return; }
  vmessage("found %d global- and %d group-jobfiles", length(jobs[0]), length(jobs[1]));

  % check on empty working directory
  variable workdir = path_dirname(pattern);
  if (length(glob(workdir + "/*")) > 0) { vmessage("error (%s): working directory not empty!", _function_name); return; }

  % run-order (0 = globals, 1 = groups)
  variable runord = qualifier("first", 1);
  
  % run-loop
  variable best = 1, jobids, pp, i, j, tmp, nloop = 0;
  do {
    jobids = Array_Type[2];
    % submit the jobs
    _for i (0,1) { % handle globals and groups the same way
      jobids[i] = Integer_Type[0];
      % submit the jobs?
      if (runord == i) {
	% loop over jobfiles of that kind
	_for j (0, length(jobs[i])-1) {
	  % submit the job and get its ID
          pp = popen(sprintf("qsub %s", jobs[i][j]), "r");
	  tmp = strjoin(fgetslines(pp));
	  ()=pclose(pp);
	  tmp = string_matches(tmp, "^\([0-9]*\)\[?\]?\."R);
	  % submit successful?
	  if (length(tmp) != 2) { vmessage("error (%s): got no job-ID after submitting %s\nis Torque running?", _function_name, jobs[i][j]); return; }
	  jobids[i] = [jobids[i], atoi(tmp[1])];
        }
	vmessage("submitted %d %s-jobs (IDs: %s)", length(jobids[i]), i == 0 ? "global" : "group", strjoin(array_map(String_Type, &sprintf, "%d", jobids[i]), ","));
      }
    }
    jobids = array_flatten(jobids);
    j = ones(length(jobids));

    % wait for the jobs to finish
    do {
      sleep(qualifier("wait", 10));
      pp = popen("qstat", "r");
      tmp = fgetslines(pp);
      ()=pclose(pp);
      % (too?) simple checks if command was successful
      if (length(tmp) < 3) { vmessage("error (%s): list of Torque jobs is empty (even completed ones)!", _function_name); return; }
      if (is_substr(tmp[0], "Job id") == 0) { vmessage("error (%s): couldn't get Torque status! Is Torque running?", _function_name); return; }
      % check output of qstat on substrings containing the job IDs     
      _for i (0,length(jobids)-1) {
	if (j[i]) { % only check if job is believed to be still running
	  % find output line where the current job is listed and check if it status isn't "R"
	  pp = [array_map(Void_Type, &string_matches, tmp, sprintf("^%d\[?\]?\."R, jobids[i]))];
	  pp = wherefirst(strlen(pp) > 0); % index where the job is listed
	  if (pp == NULL) { vmessage("error (%s): could not get status of job %d", _function_name, jobids[i]); return; }
          if (is_substr(tmp[pp], " C ")) {
	    j[i] = 0;
	    vmessage("job %d completed, %d of %d jobs left", jobids[i], int(sum(j)), length(jobids));
	  }
	}
      }
    } while(sum(j) > 0);
    
    % all jobs completed, collect results now
    vmessage("collecting results");
    variable results, info;
    results = s.fit_pars_collect(pattern;; struct_combine(struct { bestpars = &info, setnewbestpars, silent = 1+4+8 }, __qualifiers));
	    if (results == NULL) { vmessage("error (%s): collecting results failed", _function_name); return; }
    % check on new best-fit
    if (wherefirst(info.isbetter) == NULL) {
      best = 0;
      % save parameters of new best-fit
      vmessage("saving fit-parameters");
      switch (typeof(savefun))
        { case String_Type: s.save(savefun); }
        { case Ref_Type: (@savefun)(); }
    } else { best = 1; }
    % move all files created by fit_pars 
    vmessage(sprintf("moving files into loop%03d subfolder", nloop/2));
    if (access(sprintf("%s/loop%03d", workdir, nloop/2), F_OK) != 0) {
      ()=system(sprintf("mkdir %s/loop%03d", workdir, nloop/2));
    }
    ()=system(sprintf("mv %s/*.* %s/loop%03d/", workdir, workdir, nloop/2));
    % fit the other parameter type next time
    runord = not runord;
    nloop++;
  } while ((best == 0 || nloop mod 2 == 1) && nloop < maxiter*2); % nloop == 1: make sure that groups- and global are fit_parsed
  if (best == 1) { vmessage("no new best-fit found"); }
  else { vmessage("maximum number of iterations reached"); }
  % collect results of last loop
  % but for that we need to copy the files of a previous loop in order to
  % get missing parameter types (imagine loop0: best-fit in groups, best-fit
  % in globals, loop1: no best-fit in groups -> results of the globals are
  % missing in loop1 subfolder)
  nloop--;
  if (nloop > 1) { % copy result from last loop here (don't overwrite existing files!)
    ()=system(sprintf("cp -n %s/loop%03d/*.* %s/loop%03d/", workdir, nloop/2-1, workdir, nloop/2));
  }
  % collect the (merged) results
  results = s.fit_pars_collect(sprintf("%s/loop%03d/%s", workdir, nloop/2, path_basename(pattern)); silent = 1);
  if (length(info.missing) > 0) {
    vmessage("warning: uncertainties of some parameters are missing!");
     message("         please collect manually and check the results!");
  }
  if (length(info.group) > 0 && best == 1) {
    vmessage("warning: last loop still results in a better fit!");
     message("         please collect manually and check the results!");
     message("         something strange happened...");
  }
  return results;
}
%}}}

private define simultaneous_fit_load() %{{{
%!%+    
%\function{simultaneous_fit.load}
%\synopsis{loads a simultaneous fit from a FITS-table}
%\usage{simultaneous_fit.load(String_Type filename);}
%\description
%    Uses 'fits_load_fit' to first restore the fit and then
%    to load the model-field of the simultaneous fit from
%    the FITS-table. Because the data associations don't
%    have to be analyzed, any additional use of
%    'simultaenous_fit.add_data' before loading the fit (if
%    the data is defined by, e.g, 'define_counts') should
%    be called with the 'nosort' qualifier to save the time
%    needed to create the model-field.
%
%    The function recognizes if the simfit was saved with
%    the 'alt'-qualifier of  simultaneous_fit.save. In this
%    case 'fits_load_fit' is not used.
%
%    All qualifiers are passed to fits_load_fit.
%
%\example
%    % initialize a new simultaneous_fit
%    simfit = simultaneous_fit();
%    % define the data
%    simfit.add_data({
%      [define_counts(...), ...],
%      ...
%    }; nosort);
%    % restore the fit (parameters and model-field)
%    % from a file previously created by simfit.save
%    simfit.load("best-fit.fits"; nodata);
%\seealso{simultaneous_fit.save, simultaneous_fit, fits_load_fit}
%!%-
{
  variable s, filename;
  switch (_NARGS)
    { case 2: (s,filename) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  % load the fit first
  variable fit, tmp;
  if (fits_read_key(filename + "[0]", "alt") == NULL) {
    ()=fits_load_fit(filename;; struct_combine(__qualifiers, struct { strct = &fit }));
  } else { % simfit was saved in an alternative way
    fit = fits_load_fit_struct(filename);
    fit_fun(fits_read_key(filename + "[save_par]", "fit_fun"));
    variable j, i = where(fit.tie != " ", &j);
    array_map(Void_Type, &tie, fit.tie[i], fit.name[i]);
    array_map(Void_Type, &untie, fit.name[j]);
    i = where(fit.fun != " ", &j);
    array_map(Void_Type, &set_par_fun, fit.name[i], fit.fun[i]);
    array_map(Void_Type, &set_par_fun, fit.name[j], NULL);
    array_map(Void_Type, &set_par, get_struct_fields(fit,
      "name", "value", "freeze", "min", "max"
    ));
  }
  if (qualifier_exists("strct")) {
    tmp = qualifier("strct");
    @tmp = fit;
  }

  %%% restore simultaneous fit informations
  % data indices and fit-function
  s.data = list_new;
  array_map(Void_Type, &list_append, s.data, array_map(
    Array_Type, &atoi, array_map(Array_Type, &strchop, strchop(fit.simfit_data[0], ';', 0), ',', 0)
  ));
  s.model.fit_fun = fit.simfit_fit_fun[0];
  % global and group parameters and to what they are tied
  s.model.global = String_Type[0];
  s.model.group = Array_Type[length(s.data)];
  s.model.tieto = Array_Type[length(s.data)];
  _for tmp (0, length(s.data)-1, 1) {
    s.model.group[tmp] = String_Type[0];
    s.model.tieto[tmp] = String_Type[0];
  }
  variable par, grp;
  foreach par (get_params) {
    tmp = "simfit_" + fits_conv_to_legal_char(par.name) + "_group";
    if (struct_field_exists(fit, tmp)) {
      grp = get_struct_field(fit, tmp)[0];
      % global parameter (grp == 0 AND grp < 0, i.e. associated, but global)
      if (grp <= 0) {
	s.model.global = [s.model.global, par.name];
      }
      % group parameter (also applicable to a global one)
      grp = abs(grp);
      if (grp > 0) {
	grp--;
        s.model.group[grp] = [s.model.group[grp], par.name];
        tmp = "simfit_" + fits_conv_to_legal_char(par.name) + "_tieto";
  	if (struct_field_exists(fit, tmp)) {
	  tmp = get_struct_field(fit, tmp)[0];
	  s.model.tieto[grp] = [s.model.tieto[grp], tmp == " " ? NULL : tmp];
	} else {
          vmessage("warning (%s): did not find tie-to field of parameter %s", _function_name, par.name);
	}
      }
    } else {
      vmessage("warning (%s): did not find a group association of parameter %s", _function_name, par.name);
    }
  }
  % statistic
  s.model.stat = Struct_Type[length(s.data)+1];
  tmp = fit.simfit_statistic;
  _for grp (0, length(s.model.stat)-1, 1) {
    s.model.stat[grp] = struct {
      statistic = fit.simfit_statistic[0,grp],
      num_variable_params = fit.simfit_num_variable_params[0,grp],
      num_bins = fit.simfit_num_bins[0,grp]
    };
  }
  _simultaneous_fit_comb_chi_sqr(s; recalc);
  % instances
  s.model.instances = Assoc_Type[Array_Type];
  foreach tmp (strchop(fit.simfit_instances[0], ';', 0)) {
    grp = strchop(tmp, '=', 0);
    s.model.instances[grp[0]] = strchop(grp[1], ',', 0);
  }

  % tieto-history
  s.model.tieto_history = list_new;
  if (struct_field_exists(fit, "simfit_tieto_history")) {
    if (fit.simfit_tieto_history[0] != " ") {
      foreach grp (array_map(Array_Type, &strchop, strchop(fit.simfit_tieto_history[0], ';', 0), ',', 0)) {
	if (length(grp) == 2) {
          list_append(s.model.tieto_history, { grp[0], grp[1] });
	} else {
	  % there seems to be a bug in saveing the history
	  message("warning: error in saved set_par_fun-history, could not restore history completely");
	}
      }
    }
  } else { message("warning: could not find a set_par_fun-history"); }
} %}}}

private define simultaneous_fit_load_model() %{{{
%!%+
%\function{simultaneous_fit.load_model}
%\description
%    DEPRECATED, use simultaneous_fit.load
%\seealso{simultaneous_fit.load}
%!%-
{
  variable s, filename;
  switch (_NARGS)
    { case 2: (s,filename) = (); }
    { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  message("*** DEPRECATED, use simultaneous_fit.load ***");
  ifnot (qualifier_exists("go")) { throw RunTimeError, "***  to continue set the 'go' qualifier   ***"; }

  (s.model,) = evalfile(filename,current_namespace);
  % check on Null_Type of 'stat' field
  ifnot (struct_field_exists(s.model, "stat")) { s.model = struct_combine(s.model, struct { stat = NULL }); }
  if (_typeof(s.model.stat) == Null_Type) { s.model.stat = Struct_Type[length(s.model.group)+1]; }
} %}}}

private define simultaneous_fit_save() %{{{
%!%+    
%\function{simultaneous_fit.save}
%\synopsis{saves the simultaneous fit to a FITS-table}
%\usage{simultaneous_fit.save_model(String_Type filename[, Struct/String_Type conf]);}
%\qualifiers{
%    \qualifier{alt}{use an alternative to 'fits_save_fit', see text;
%  all other qualifiers are passed to 'fits_save_fit'}
% }
%\description
%    Uses 'fits_save_fit' to save the simultaneous fit and,
%    most importantly, its model-field. The creation of this
%    field may take some time due to the fact that the fit-
%    function as well as the data associations have to be
%    analyzed. Thus, saving it to a FITS-table speeds up
%    the loading process if 'simultaneous_fit.load' is used.
%
%    However, the large number of parameters of a simfit
%    can lead to "out of memory" exception when using
%    'fits_save_fit'. In this case, the 'alt'-qualifier
%    uses an alternative FITS-structure for saving, i.e.,
%    'fits_save_fit' is not used. Note that in this case
%    only the parameters (like in 'save_par') are saved,
%    but _not_ the loaded data, noticed energy bins, etc.
%\seealso{simultaneous_fit.load, fits_save_fit}
%!%-
{
  variable s, args;
  if (2<= _NARGS <=3) {
    args = __pop_args(_NARGS-1);
    s = ();
  } else { _simultaneous_fit_help(_NARGS, _function_name); return; }
  if (_simultaneous_fit_help(0, _function_name ;; __qualifiers)) return;

  %%% build model structure to be appended to the FITS-table
  variable info = struct {
    simfit_fit_fun = [s.model.fit_fun],
    % convert the .data list of integer-arrays as a string like "1,2;3,4,5;..."
    simfit_data = [strjoin(array_map( % loop over groups
      String_Type, &strjoin, array_map(  % loop over dataset within group
        Array_Type, &array_map,
	  String_Type, &sprintf, "%d", list_to_array(s.data)
      ), ","),
    ";")]
  };
  % global parameters
  variable par;
  foreach par (s.model.global) {
    info = create_struct_field(info, "simfit_" + fits_conv_to_legal_char(par) + "_group", [0]);
  }
  % group parameters and to what they are tied
  variable grp;
  _for grp (0, length(s.model.group)-1, 1) {
    _for par (0, length(s.model.group[grp])-1, 1) {
      variable tmp = "simfit_" + fits_conv_to_legal_char(s.model.group[grp][par]) + "_group";
      % if field does not exist yet -> group parameter
      % else -> global parameter, save group association, but with a negative sign
      ifnot (struct_field_exists(info, tmp)) {
        info = create_struct_field(info, tmp, [grp+1]);
      } else {
	set_struct_field(info, tmp, -[grp+1]);
      }
      info = create_struct_field(info, "simfit_" + fits_conv_to_legal_char(s.model.group[grp][par]) + "_tieto", [s.model.tieto[grp][par] == NULL ? "" : s.model.tieto[grp][par]]);
    }
  }
  % statistic
  par = merge_struct_arrays(s.model.stat);
  info = struct_combine(info, ["simfit_statistic", "simfit_num_variable_params", "simfit_num_bins"]);
  grp = length(par.statistic);
  info.simfit_statistic = _reshape(par.statistic, [1,grp]);
  info.simfit_num_variable_params = _reshape(par.num_variable_params, [1,grp]);
  info.simfit_num_bins = _reshape(par.num_bins, [1,grp]);
  % instances
  grp = String_Type[0];
  foreach par (assoc_get_keys(s.model.instances)) {
    grp = [grp, sprintf("%s=%s", par, strjoin(s.model.instances[par], ","))];
  }
  info = create_struct_field(info, "simfit_instances", [strjoin(grp, ";")]);
  % tieto-history
  % There seems to be a bug here where the first array_map returns
  % a string array with one element only (no "tie-to"-parameter seems
  % to be present), i.e. the string contains "...;global;..." instead
  % of "...;global,tieto;..." -> workaround in the load-function
  info = create_struct_field(info, "simfit_tieto_history",
    length(s.model.tieto_history) == 0 ? [""] : [strjoin(
      array_map(String_Type, &strjoin, array_map(
        Array_Type, &list_to_array, list_to_array(s.model.tieto_history)
      ), ","), ";"
    )]
  );
  
  % eventually add a given info structure
  if (qualifier_exists("info")) {
    info = struct_combine(qualifier("info"), info);
  }

  % save the fit
  ifnot (qualifier_exists("alt")) {
    fits_save_fit(__push_args(args);; struct_combine(__qualifiers, struct { info = info }));
  } else {
    fits_save_fit_write(args[0].value, info);
    % save all parameters into the same file
    % merge_struct_arrays will not work here because NULL field-values
    % are skipped, which is the case for .tie and .fun in many cases
    variable pars = get_params();
    pars = struct {
      name = array_map(String_Type, &get_struct_field, pars, "name"),
      index = array_map(Integer_Type, &get_struct_field, pars, "index"),
      value = array_map(Double_Type, &get_struct_field, pars, "value"),
      min = array_map(Double_Type, &get_struct_field, pars, "min"),
      max = array_map(Double_Type, &get_struct_field, pars, "max"),
      freeze = array_map(Integer_Type, &get_struct_field, pars, "freeze"),
      tie = array_map(String_Type, &get_struct_field, pars, "tie"),
      fun = array_map(String_Type, &get_struct_field, pars, "fun")
    };
    pars.tie[where(_isnull(pars.tie))] = "";
    pars.fun[where(_isnull(pars.fun))] = "";
    variable fp = fits_open_file(args[0].value, "w");
    fits_update_key(fp, "alt", 1, "alternative format");
    fits_write_binary_table(fp, "save_par", pars, struct { fit_fun = get_fit_fun });
    if (length(args) > 1) { fits_write_binary_table(fp, "fit_pars", args[1].value); }
    fits_close_file(fp);
  }
} %}}}

%%%%%%%%%%%%%%%%%%%%%
define simultaneous_fit()
%%%%%%%%%%%%%%%%%%%%% %{{{
%!%+    
%\function{simultaneous_fit}
%\synopsis{initializes a structure to handle simultaneous/combined
%    fits of a lot of data}
%\usage{Struct_Type simultaneous_fit();}
%\description
%    The advantage of fitting a lot of data at the same is
%    that fit parameters, which seem to be the same for all
%    data, can be fitted properly. That results in a reduced
%    number of free parameters for individual observations
%    or data groups, and thus constrains parameters even
%    better at low data quality.
%
%    The disadvantages are, that a) a lot of painful work
%    has to be done on tieing and freezing parameters and
%    b) a fit using 'fit_counts' will take hours to days.
%    There are several functions implemented within the
%    returned structure, which can help to solve that issues.
%
%    To access the help of a function, simply call it with
%    the 'help' qualifier (like for the xfig functions).
%
%    References for simultaneous fits in ISIS are
%      Kuehnel et al. 2015, Acta Polytechnica 55(2), 123­127
%      Kuehnel et al. 2016, Acta Polytechnica 56(1), 41­46
%    
%    NOTE: The simultaneous fits are still in development.
%!%-
{
  if (_NARGS > 0) { help(_function_name); return; }

  message("NOTE: a new version of the add_data-function has been implemented.");
  message("      You may set the \"old\"-qualifier to fall back to the old");
  message("      function for compatibility reasons.");
  
  return struct {
    add_data = qualifier_exists("old") ? &simultaneous_fit_add_data_old : &simultaneous_fit_add_data,
    delete_data = &simultaneous_fit_delete_data,
    sort_params = &simultaneous_fit_sort_params,
    apply_logic = &simultaneous_fit_apply_logic,
    fit_fun = &simultaneous_fit_fit_fun,
    get_par = &simultaneous_fit_get_par,
    set_par = &simultaneous_fit_set_par,
    freeze = &simultaneous_fit_freeze,
    thaw = &simultaneous_fit_thaw,
    copy_par = &simultaneous_fit_copy_par,
    set_par_fun = &simultaneous_fit_set_par_fun,
    set_par_fun_history = &simultaneous_fit_set_par_fun_history,
    set_global = &simultaneous_fit_set_global,
    unset_global = &simultaneous_fit_unset_global,
    list_data = &simultaneous_fit_list_data,
    list_global = &simultaneous_fit_list_global,
    list_groups = &simultaneous_fit_list_groups,
    plot_group = &simultaneous_fit_plot_group,
    group_stats = &simultaneous_fit_group_stats,
    filter_groups = &simultaneous_fit_filter_groups,
    select_groups = &simultaneous_fit_select_groups,
    eval_groups = &simultaneous_fit_eval_groups,
    eval_global = &simultaneous_fit_eval_global,
    fit_groups = &simultaneous_fit_fit_groups,
    fit_global = &simultaneous_fit_fit_global,
    fit_smart = &simultaneous_fit_fit_smart,
    fit_pars = &simultaneous_fit_fit_pars,
    steppar = &simultaneous_fit_steppar, 
#ifeval __get_reference("mpi_fit_pars")!=NULL
    mpi_fit_pars = &simultaneous_fit_mpi_fit_pars,
    mpi_fit_pars_jobfiles = &simultaneous_fit_mpi_fit_pars_jobfiles,
#endif
    fit_pars_collect = &simultaneous_fit_fit_pars_collect,
    fit_pars_run_job = &simultaneous_fit_fit_pars_run_job,
    load_model = &simultaneous_fit_load_model,
    load = &simultaneous_fit_load,
    save = &simultaneous_fit_save,
    setrestore = &simultaneous_fit_setrestore,
    restore = &simultaneous_fit_restore,
    model = struct {
      fit_fun = NULL,
      global = String_Type[0],
      group = Array_Type[0],
      stat = Struct_Type[0],
      cstat = struct {
	statistic = 0, num_group_params = 0, num_global_params = 0,
	num_bins = 0, combredchi = 0, weight_fun = NULL
      },
      tieto = Array_Type[0],
      tieto_history = list_new,
      instances = Assoc_Type[Array_Type], % key = component, value = instance string
      exclude = Integer_Type[0],
      params = Struct_Type[0],
      current_groups = -1
    },
    data = list_new
    % hack = struct {
    %   % move "internal" functions into this structure, like
    %   sort_params, apply_logic,
    %   set_weights % function to modify the weightening "mu" of the globals
    %   % functions for editing the group parameters?
    % }
  };
} %}}}

%
% Code from src/slang/compatibility/sincos.sl
%
#ifnexists sincos
public define sincos(x)
{
   return sin(x), cos(x);
}
#endif

%
% Code from src/fitting/fit-functions/sinwave.sl
%
%%%%%%%%%%%%%%%%%%
define sinwave2_fit(bin_lo, bin_hi, pars)
%%%%%%%%%%%%%%%%%%
%!%+
%\function{sinwave2 (fit-function)}
%\description
%    a*sin(b*alpha+c)+d
%    (ISIS fit-function sinwave seems to be defined for keV space:
%    sinwave({a,b,c,d};alpha) = sinwave2({a,b,c,d};_A(alpha)))
%!%-
{
   variable a  = pars[0];
   variable b  = pars[1];
   variable c  = pars[2]; 
   variable d  = pars[3];
   
%   bin_hi -= bin_lo[0];
%   bin_lo -= bin_lo[0];
%   bin_lo *= 2*PI/bin_hi[-1];
%   bin_hi *= 2*PI/bin_hi[-1];
   
   variable alpha = 0.5*(bin_lo+bin_hi)*2*PI;
   
   return a*sin(b*alpha+c)+d;
}

%%%%%%%%%%%%%%%%%%%%%%
define sinwave2_default(i)
%%%%%%%%%%%%%%%%%%%%%%
{ switch(i)
  { case 0: return (1., 0, 1e-6, 1e6 ); }
  { case 1: return (1., 0, 1e-6, 1e6 ); }
  { case 2: return (0., 0, -2*PI, 2*PI ); }
  { case 3: return (0., 0, -1, 1 ); }
}

add_slang_function("sinwave2", ["a","b","c","d"]);
set_param_default_hook("sinwave2", "sinwave2_default");

%
% Code from src/Mike/sitar.sl
%

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

%  SITAR - S-lang/ISIS Timing Analysis Routines 
% ***  Version 1.2.8 *** Oct. 18, 2017  ***

private variable __sitar_version = [1,2,8];
public variable _sitar_version=int(sum(__sitar_version*[10000,100,1]));
public variable _sitar_version_string=sprintf("%d.%d.%d",
   __sitar_version[0],__sitar_version[1],__sitar_version[2]);

% (Mostly) Alphabetical List of Routines in this File

% epfold_rate_use 
% fstat           
% glob_opt_use    
% lngamma       
% log_post_XXX
% mdc_use         
% msg
% null_it         
% pfold_rate_use  
% readasm_use     
% reb_rat_use     
% sitar_avg_cpd               
% sitar_avg_psd   
% sitar_bin_events            
% sitar_define_psd            
% sitar_epfold_rate           
% sitar_global_optimum        
% sitar_lags                  
% sitar_lbin_cpd              
% sitar_lbin_psd              
% sitar_make_data_cells       
% sitar_pfold_event            
%  _sitar_event_profile       
%  _sitar_pfold_event          
% sitar_pfold_rate            
%  _sitar_pfold_profile       
%  _sitar_pfold_rate          
% sitar_readasm               
% sitar_rebin_rate            

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% If you've got the s-lang GNU Scientific Module loaded,
% uncomment the following *before* running.  Necessary for statistics
% of epoch fold (dummy function used if left uncommented), and for
% additional functionality of the Bayesian Blocks routine (i.e.,
% `event mode' with alpha & beta > 0, **which won't crash, but won't
% give real results, either, if used without GSL**).

try{ require("gsl","gsl"); } catch AnyError: {};

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

provide("sitar");

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define null_it(a,i)
{
   variable j = Char_Type[ length(a) ];
   j[i] = 1;
   return a[ where(j==0) ];
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

variable fp = stderr;

private define mdc_use()
{
   () = fprintf(fp,  "%s\n", %{{{
`
  cell = sitar_make_data_cells(tt,type,max_delt,frame,tstart,tstop);
    
  Inputs: 
    tt          : The times of the detected events
    type        : Describe cells as 'events' (type=1 or 2), with the
                  size (i.e., normalized duration) being the distance
                  from halfway from the previous event to halfway to
                  the next event (type=1), or from the current event
                  to right before the subsequent event (type=2).
                  Alternatively, events can be assigned to bins of
                  uniform size (type=3).
    max_delt    : Events closer together than max_delt are grouped
                  together in a single cell.
    frame       : The output cell sizes are in units of frame, or this
                  is the bin size for type=3.
    tstart/tstop: start/stop times for making the output cells.
    
  Outputs: 
    cell.pops   : Array with the number of events per cell  
    cell.size   : Event mode: size (i.e. duration) of a cell in units of
                  frame.  Deadtime/efficiency should be folded into this
                  definition.  E.g., size= total "good time" / total "good
                  time" for an individual frame.  (Here, we use the equivalent
                  for uniform deadtime, size = total time / total frame time)
                  Binned mode: size is mean (over whole observation) counts
                  per bin.  (Again, here is where you would put in efficiency 
                  factors, *if* efficiency/deadtime is not uniform bin to bin)
    cell.lo/hi_t: cell start and stop times
    cell.dtcor  : Array, currently set to unity, for storing 'dead 
                  time' or 'efficiency' corrections for the cells.
                  Must be set, after the fact, by the user.
                  Currently only used in sitar_global_optimum to
                  give output block rates corrected for dead time    
`);  %}}}    
   return;
}
  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_make_data_cells( )
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_make_data_cells}
%\synopsis{Loads a dataset with a non-unity AREASCAL keyword or column}
%\usage{cell = sitar_make_data_cells(tt,type,max_delt,frame,tstart,tstop);}
%\description
%    Create data cells from event data to be used with the Bayesian
%    blocks algorithm. tt are the times of the events; type=1,2, or 3
%    determines cell breaks (midway between events; right before
%    subsequent event, or binned); events closer than max_delt are
%    grouped together in a single cell; output cell sizes are in units
%    of frame (or = bin size for type=3), tstart/tstop are the
%    stop/start times for the output cells.
%
%    Output is a structure with cell.pops (number of events per cell),
%    cell.size (duration of cell in units for frame), cell.lo/hi_t
%    (start/stop times of cell), cell.dtcor (array, currently set to
%    unity, for storing cell deadtime corrections).
%\seealso{sitar_global_optimum}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{ 
%     Based upon Jeff Scargle's MATLAB routines to perform a 
%     Bayesian blocks decomposition of an ordered array.
%
%     Paper: "Studies in Astronomical Time Series Analysis. VI.
%             Optimum Partition of the Interval: Bayesian Blocks,
%	      Histograms, and Triggers", J.D. Scargle et al., 2003
%             (to be submitted).
%		
%     Web Site:  http://astrophysics.arc.nasa.gov/~jeffrey

   switch ( _NARGS )
   {
      case 0:
      mdc_use( );
      return;
   }
   {
      variable tt,type,max_delt,frame,tstart,tstop;
      (tt,type,max_delt,frame,tstart,tstop) = ( );
   }

   %  Initially, one datum per cell.

   variable cell = struct{ pops, size, lo_t, hi_t, dtcor };
   tt = tt[ where( tt >= tstart and tt <= tstop ) ];
   cell.pops = ones( length(tt) );

   variable mintt = min(tt);
   variable maxtt = max(tt);
   variable difftt = shift(tt,1) - tt;
   difftt = difftt[ [0:length(difftt)-2] ];

   variable ii_close = where( difftt < max_delt );
   variable ii_start, ii_beyond, ii_end, ii_clump;
   variable  clump_pop, clump_tt, ind_null;

   if ( type !=3 ) while( length(ii_close) )
   {
      ii_start = ii_close[0];
      ii_beyond = where( shift(ii_close,1) - ii_close > 1 );

      if ( length(ii_beyond) == 0 )
      {
         ii_end = ii_start + length(ii_close);
      }
      else
      {
         ii_end = ii_start + ii_beyond[0] + 1;
      }

      ii_clump = [ii_start:ii_end];

      clump_pop = sum( cell.pops[ii_clump] );
      clump_pop = typecast(clump_pop,Integer_Type);
      clump_tt = mean( tt[ii_clump] );

      % Put the members of a clump into one cell

      cell.pops[ii_start] = clump_pop;
      tt[ii_start] = clump_tt;

      % Null the cells evacuated by this operation

      if ( ii_end > ii_start+1 )
      {
         ind_null = [ii_start+1:ii_end];
      }
      else
      {
         ind_null = [ii_end:ii_end];
      }
      cell.pops = null_it(cell.pops,ind_null);
      tt = null_it(tt,ind_null);

      difftt = shift(tt,1) - tt;
      difftt = difftt[ [0:length(difftt)-2] ];
      ii_close = where( difftt < max_delt );
   }

   % Now define the cell sizes and locations.

   variable dt = shift(tt,1) - tt;
   variable ndt = length(dt)-1;
   dt = dt[[0:ndt-1]];
   variable cstart = 1, cstop = ndt-1;

   switch (type)
   { 
    case 1:

    % type = 1: Set to midpoints

      cell.size = 0.5 * (dt[[0:ndt-2]] + dt[[1:ndt-1]]);
      cell.lo_t = tt[[1:ndt-1]]-dt[[0:ndt-2]]/2.;
      cell.hi_t = tt[[1:ndt-1]]+dt[[1:ndt-1]]/2.;

      % If tstart is < first event, add in the first event.  Design
      % feature for sparse lightcurves, where a long wait time until
      % the first event might mean something.

      if ( tstart < mintt )
      {
         cell.size = [ tt[0]-tstart+dt[0]/2., cell.size ];
         cell.lo_t = [ tstart, cell.lo_t ];
         cell.hi_t = [ tt[0]+dt[0]/2., cell.hi_t ];
         cstart = 0;
      }

      % If tstop is > last event, add in the last event.  Design
      % feature for sparse lightcurves, where a long drought at the
      % end might mean something.

      if ( tstop > maxtt )
      {
         cell.size = [ cell.size, tstop-tt[ndt]+dt[ndt-1] ];
         cell.lo_t = [ cell.lo_t, tt[ndt-1]+dt[ndt-1]/2. ];
         cell.hi_t = [ cell.hi_t, tstop ];
         cstop = ndt;
      }

      % Now make cell.pops match the above.

      cell.pops = cell.pops[ [cstart:cstop] ];
   }

   {
    case 2:

    % type =2: Intervals instead of midpoints. 

      cell.size = dt;
      cell.lo_t = tt[ [0:ndt-1] ];
      cell.hi_t = tt[ [1:ndt] ];
      cstart = 0;
      cstop = ndt-1;

      % If tstop is > last event, add in the last event.  Design
      % feature for sparse lightcurves, where a long drought at the
      % end might mean something.

      if ( tstop > maxtt )
      {
         cell.size = [ cell.size, tstop-tt[ndt] ];
         cell.lo_t = [ cell.lo_t, tt[ndt] ];
         cell.hi_t = [ cell.hi_t, tstop ];
         cstop = ndt;
      }

      % I'm not entirely happy with any choices for the endpoints for
      % sparse lightcurves.  But one has to choose *something*.  For
      % tstart < mintt, append that time to the first bin.  Thought
      % for alternative and type=2: add a bin with 0 photons going
      % from tstart to right before the first photon?  (Should be OK
      % with the statistic?)

      if ( tstart < mintt )
      {
         cell.size[0] = cell.size[0] + ( tt[0]-tstart );
         cell.lo_t[0] = tstart;
      }
      cell.pops = cell.pops[ [cstart:cstop] ];
   }

   {
    case 3:

    % type = 3: Bin using ISIS functions

      variable nbins = typecast( (tstop-tstart)/frame, Integer_Type );
      ( cell.lo_t, cell.hi_t ) = linear_grid( tstart, tstop, nbins);

      cell.pops = histogram( tt, cell.lo_t, cell.hi_t );
      cell.dtcor = ones( length(cell.pops) );
      cell.size = @cell.dtcor;

      % Based upon my thoughts of applying this to gratings data, I've
      % decided upon the following modification.  This forces the
      % "rate per cell size" to be typically less than one.  Or, in
      % terms of the marginalized likelihood, this puts a sum of
      % "model counts per block" in the denominator (presuming that
      % your zeroth order model is `constant rate over the whole
      % observation').

      cell.size = cell.size * ( sum(cell.pops)/length(cell.size) );
      return cell;
   }

   {

    % If it's not type 1, 2, or 3, something is wrong! 

      error("Type not 1, 2, or 3 in 'sitar_make_data_cells'");
   }

   % The return for Type 1 or 2

   cell.size = cell.size/frame;
   cell.dtcor = ones( length(cell.pops) );	
   return cell;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#ifnexists gsl->GSL_EUNDRFLW
% If GSL is not defined, define our own
private define lngamma(z)
{
   variable c = [76.18009173,-86.50532033,24.01409822,-1.231739516,
                 1.20858003e-3,-5.36382e-6];
   variable csum=1.;
   foreach(c)
   {  
      csum+=()/z;
      z+=1.;
   }
   csum = log( sqrt(2.*PI)*csum ) - (z-1.5) + (z-6.5)*log(z-1.5);
   return csum;
}
# else
private define lngamma(z)
{
   return gsl->lngamma(z);
}
#endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#ifexists gsl->GSL_EUNDRFLW
   gsl->gsl_set_error_disposition (gsl->GSL_EUNDRFLW, 0);
#endif

#ifnexists beta_inc
private define beta_inc(x,y,z)
{
   return 10.;
}
#else
private define beta_inc(x,y,z)
{
   return gsl->beta_inc(x,y,z);
}
#endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% log_post_XXX functions
%
% PURPOSE:
%     Used in Bayesian decomposition of *time tagged event* data,
%     or *binned* data into blocks with statistically uniform "rate".
%     Note: this does not have to be used on a lightcurve, but instead
%     could be used, e.g., on a spectrum to identify potential lines.  
%     The only requirement is that the data be sequential, with an
%     independent variable (e.g., time, wavelength, etc.), and a dependent
%     variable (e.g., counts, rate).
%
%     This function calculates the log of the probability function (or 
%     fitness function) for a given decomposition.  Called as a subroutine
%     from "global_optimum".
%
% CALLING SEQUENCE:
%     log_prob = log_post_XXX(data_pops, data_size, ncp_prior, alpha, beta);
%
% INPUTS:
%     data_pops  : Array with cell populations (i.e., number of "events" 
%                  per cell)
%     data_size  : Array with cell sizes (i.e., number of "ticks" 
%                  per cell; can contain window function/efficiency)
%     ncp_prior  : log parameter for number of change points.  It is
%                  assumed that the prior probability of # of change
%                  points goes as gamma^N_cp, then:
%	              ncp_prior = -log(gamma)
%                  MAN suggests: ncp_prior = log(# of bins in lightcurve/
%                                             maximum # of desired cells)
%                  ncp_prior=7 is ~ 10^-3 significance for each new block
%     alpha, beta: Parameters that affect choices on priors.  
%                  See glob_opt_use.
%
% OUTPUTS:
%     log_prob   : The logarithm of the probability function for a given
%                  decomposition of the lightcurve.
% 
% MODIFICATION HISTORY:
%     Based upon Jeff Scargle's MATLAB routines to perform a 
%     Bayesian blocks decomposition of an ordered array.
%
%     Paper: "Studies in Astronomical Time Series Analysis. VI.
%             Optimum Partition of the Interval: Bayesian Blocks,
%	      Histograms, and Triggers", J.D. Scargle et al., 2003
%             (to be submitted).
%		
%     Web Site:  http://astrophysics.arc.nasa.gov/~jeffrey
%
% Note: no error checking, since this subroutine shouldn't be called 
% directly, only through "sitar_global_optimum();"

private define log_post_bin( data_pops, data_size, ncp_prior, alpha, beta)
{
   variable dpop_plus_alpha = data_pops + alpha;
   variable addenda = alpha*log( beta ) - lngamma( alpha ) - ncp_prior;
   variable log_prob = lngamma( dpop_plus_alpha ) 
                       - dpop_plus_alpha*log( data_size + beta ) + addenda;
   return log_prob;   
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define log_post_max( data_pops, data_size, ncp_prior, alpha, beta )
{
   variable log_prob = data_pops*( log( (data_pops+1.e-12)/data_size ) - 1) - ncp_prior;
   return log_prob;   
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define log_post_evt_rate( data_pops, data_size, ncp_prior, alpha, beta )
{
   variable log_prob;

   variable arg = data_size - data_pops;
   variable ii = where( arg > 0 );
   variable addenda = -log( beta - alpha ) - ncp_prior;

   % Slightly different tack than Scargle here.  His MATLAB code
   % sets this to ~0.  I set this M=N in Eq. (23) of Scargle (1998,
   % ApJ, 504, pp. 405-418).  Neither is strictly correct, but I
   % don't know if it makes a substantial difference.  (If it does,
   % you should be using the binned version instead.)  

   % The following represent my own thoughts for this case.

   variable ealpha = @data_size, ebeta = @ealpha;
   ealpha[*] = exp(-alpha);
   ebeta[*] = exp(-beta);
      
   % This is only an approximate expression, to make up for my
   % lack of integration skills
         
   log_prob = data_size * log(1.-ebeta) + log(beta)
              - log( beta - alpha );

   % This, however, I believe, is correct.  It *will not*
   % evaluate correctly if the GSL module is not initialized

   log_prob[ii] = lngamma( arg[ii] ) +
                  lngamma( data_pops[ii] + 1 ) -
                  lngamma( data_size[ii] + 1 ) +
                  log( 
                      beta_inc(arg[ii],data_pops[ii]+1,ealpha[ii]) -
                      beta_inc(arg[ii],data_pops[ii]+1,ebeta[ii])  
                      );
                  
   log_prob += addenda;

   return log_prob;   
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define log_post_evt_mean( data_pops, data_size, ncp_prior, alpha, beta )
{
   variable log_prob;

   variable arg = data_size - data_pops;
   variable ii = where( arg > 0 );
   variable dsize_plus_one_over_alpha = data_size + 1./alpha;
   variable dsize_plus_one_over_alpha_plus_one = __tmp(dsize_plus_one_over_alpha) + 1;
   variable addenda = -log(alpha) - ncp_prior;

   log_prob = lngamma( dsize_plus_one_over_alpha ) 
            - lngamma( dsize_plus_one_over_alpha_plus_one );

   log_prob[ii] = lngamma( data_pops[ii] + 1 ) 
                + lngamma( arg[ii] + 1./alpha )
                - lngamma( dsize_plus_one_over_alpha_plus_one[ii] );

   log_prob += addenda;

   return log_prob;   
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define log_post_evt_alpha( data_pops, data_size, ncp_prior, alpha, beta)
{
   variable log_prob;

   variable arg = data_size - data_pops;
   variable ii = where( arg > 0 );
   variable dsize_plus_alpha_plus_two = data_size + alpha + 2;
   variable addenda = log( alpha+1 ) - ncp_prior;

   log_prob = lngamma( data_size + 1 ) + lngamma( alpha + 1 )
            - lngamma( dsize_plus_alpha_plus_two );

   log_prob[ii] = lngamma( data_pops[ii] + 1 ) 
                + lngamma( arg[ii] + alpha + 1 )
                - lngamma( dsize_plus_alpha_plus_two[ii] );

   log_prob += addenda;

   return log_prob;   
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define log_post_evt( data_pops, data_size, ncp_prior, alpha, beta )
{
   variable log_prob;

   variable arg = data_size - data_pops;
   variable ii = where( arg > 0 );
   variable dsize_plus_two = data_size + 2.;

   log_prob = lngamma( data_size + 1. ) 
            - lngamma( dsize_plus_two );

   log_prob[ii] = lngamma( data_pops[ii] + 1 ) 
                + lngamma( arg[ii] + 1.  )
                - lngamma( dsize_plus_two[ii] );
 
   log_prob -= ncp_prior;

   return log_prob;   
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


private define glob_opt_use()
{
   () = fprintf(fp,  "%s\n", %{{{
`
   ans = sitar_global_optimum( cell, ncp_prior, type 
                               [; first, mean_prior, rate_prior, alpha=#, beta=#] );
    
   Inputs:
     cell      : A structure, with cell.pops, cell.size, cell.lo_t
                 cell.hi_t,cell.dtcor (see sitar_make_data_cells)
     ncp_prior : Parameter for prior on number of 'blocks' 
     type      : Identical to type from sitar_make_data_cells

   Qualifiers:
     first     : If set, the code runs in 'trigger' mode, and returns upon
                 the first sign of a change, so long as it is greater
                 than 'first' cells from the beginning of the
                 lightcurve

                EVENT MODE PRIORS (type=1 or 2)-

                 Default prior is that p_1, the probability of one or
                 more photons in a frame, is uniformly dsitributed
                 from 0->1.

     alpha     : >=0 implies p_1 prior is (1+alpha) (1-p_1)^alpha 

     mean_prior: if set, prior on *rate* is exp(-Lambda/Lambda_0),
                 where Lambda_0 is the mean rate from the entire
                 observation.  Supercedes any choice on alpha.

     rate_prior: if set, prior is chosen to be uniform between
     rate_low,   rates ranging from rate_low -- rate_high, with
     rate_high   defaults of rate_low = 1/3 mean rate and 
                 rate_high = 3 times mean rate. Supercedes any
                 choice of alpha or mean_prior.

                BINNED MODE PRIORS AND POSTERIORS (type=3)-

     alpha,    : The prior on the bin rate (sort of) goes as
     beta        (Lambda)^(alpha-1) Exp(-beta*Lambda), where
                 Lambda=rate, and the default is alpha=beta=1

     max_like  : Use maximum likelihood, log(Pmax) = N log(N/T) -N,
                 as the posterior test function.  Overrides any
                 choice of alpha & beta for binned mode.

   Outputs:
     results.cpt  : Array of change point locations for the maximum 
                    likelihood solution (indices are specific to the
                    cell input);
     results.last,: (Diagnostic purposes only) Arrays of the
            .best   location of the last change point and the
                    associated maximum log probability
     results.cts  : Counts in each block
     results.rate : Rate in each block
     results.err  : Poisson error for the block rate
     results.lo_t,: Times of lower (>=) and upper (<) block
            .hi_t   boundaries.
`); %}}}
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_global_optimum()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_global_optimum}
%\synopsis{Loads a dataset with a non-unity AREASCAL keyword or column}
%\usage{ans = sitar_global_optimum( cell, ncp_prior, type [; first, mean_prior, rate_prior, alpha=Double_Type, beta=Double_Type] );}
%\description
%   Inputs:
%     cell      : A structure, with cell.pops, cell.size, cell.lo_t
%                 cell.hi_t,cell.dtcor (see sitar_make_data_cells)
%     ncp_prior : Parameter for prior on number of 'blocks' 
%     type      : Identical to type from sitar_make_data_cells
%
%   Qualifiers:
%     first     : If set, the code runs in 'trigger' mode, and returns upon
%                 the first sign of a change, so long as it is greater
%                 than 'first' cells from the beginning of the
%                 lightcurve
%
%                EVENT MODE PRIORS (type=1 or 2)-
%
%                 Default prior is that p_1, the probability of one or
%                 more photons in a frame, is uniformly dsitributed
%                 from 0->1.
%
%     alpha     : >=0 implies p_1 prior is (1+alpha) (1-p_1)^alpha 
%
%     mean_prior: if set, prior on *rate* is exp(-Lambda/Lambda_0),
%                 where Lambda_0 is the mean rate from the entire
%                 observation.  Supercedes any choice on alpha.
%
%     rate_prior: if set, prior is chosen to be uniform between
%     rate_low,   rates ranging from rate_low -- rate_high, with
%     rate_high   defaults of rate_low = 1/3 mean rate and 
%                 rate_high = 3 times mean rate. Supercedes any
%                 choice of alpha or mean_prior.
%
%                BINNED MODE PRIORS AND POSTERIORS (type=3)-
%
%     alpha,    : The prior on the bin rate (sort of) goes as
%     beta        (Lambda)^(alpha-1) Exp(-beta*Lambda), where
%                 Lambda=rate, and the default is alpha=beta=1
%
%     max_like  : Use maximum likelihood, log(Pmax) = N log(N/T) -N,
%                 as the posterior test function.  Overrides any
%                 choice of alpha & beta for binned mode.
%
%   Outputs:
%     results.cpt  : Array of change point locations for the maximum 
%                    likelihood solution (indices are specific to the
%                    cell input);
%     results.last,: (Diagnostic purposes only) Arrays of the
%            .best   location of the last change point and the
%                    associated maximum log probability
%     results.cts  : Counts in each block
%     results.rate : Rate in each block
%     results.err  : Poisson error for the block rate
%     results.lo_t,: Times of lower (>=) and upper (<) block
%            .hi_t   boundaries.
%\seealso{sitar_make_data_cells}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
%     Used in Bayesian decomposition of *time tagged event* data,
%     or *binned* data into blocks with statistically uniform "rate".
%     Note: this does not have to be used on a lightcurve, but instead
%     could be used, e.g., on a spectrum to identify potential lines.  
%     The only requirement is that the data be sequential, with an
%     independent variable (e.g., time, wavelength, etc.), and the 
%     dependent variable (i.e., counts)
%
%     This is the main routine to determine the optimal partitioning of 
%     the sequential data.
% 
% MODIFICATION HISTORY:
%     Based upon Jeff Scargle's MATLAB routines to perform a 
%     Bayesian blocks decomposition of an ordered array.
%
%     Paper: "Studies in Astronomical Time Series Analysis. VI.
%             Optimum Partition of the Interval: Bayesian Blocks,
%	      Histograms, and Triggers", J.D. Scargle et al., 2003
%             (to be submitted).
%		
   % Some really basic error checking for the input

   variable cell, ev_type, ncp_prior, first, alpha, beta, log_prob, lo, ab_default;
   switch ( _NARGS )
   {
    case 3:
      ( cell, ncp_prior, ev_type ) = ();
   }
   {
      glob_opt_use();
      if(_NARGS != 0)
      {
         () = fprintf(fp, "\n Incorrect number of arguments. \n");
         _pop_n(_NARGS);
      }
      return;
   }

   if ( is_struct_type(cell) )
   {
      variable fields = get_struct_field_names(cell);
      if ( length( where( fields == "pops" or fields == "size" or
                          fields == "lo_t" or fields == "hi_t" or
                          fields == "dtcor" ) ) != 5 )
      {
         () = fprintf(fp, "\n Fields on cell structure not properly set.\n");
         return;
      }
   }
   else
   {
      () = fprintf(fp, "\n 'cell' was not a structure\n");
      return;
   }

   if (
        typeof(cell.pops) != Array_Type or 
        typeof(cell.size) != Array_Type or 
        typeof(cell.lo_t) != Array_Type or 
        typeof(cell.hi_t) != Array_Type or 
        typeof(cell.dtcor)  != Array_Type 
      ) 
   {
       
      () = fprintf(fp,"\n Error: All fields of structure 'cell' must be Array_Type.\n");
      return;
   }

   variable num_cells = length(cell.pops);

   if ( num_cells < 3 ) 
   {
      () = fprintf(fp, "\n Error: length of lightcurve is %d. Needs >= 3.\n", num_cells);
      return;
   }

   if (
        num_cells != length(cell.size) ||
        num_cells != length(cell.lo_t) ||
        num_cells != length(cell.hi_t) ||
        num_cells != length(cell.dtcor) 
      ) 
   {
      () = fprintf(fp, "\n Error: All arrays in cell structure must have same length\n");  
      return;
   }

   if ( ev_type != 1 && ev_type != 2 && ev_type != 3 ) 
   {
      () = fprintf(fp, "\n Error: 'ev_type' set to %d, but must be 1, 2, or 3", ev_type);
      return;
   }

   first=0;
   if(qualifier_exists("first")) first=1;

   variable lpost_fun;

   if(ev_type == 3 && qualifier_exists("max_like"))
   {
      alpha = 1.;
      beta = 1.;
      lpost_fun = &log_post_max;
   }
   else if(ev_type == 3)
   {
      alpha = qualifier("alpha",1.);
      beta = qualifier("beta",1.);
      lpost_fun = &log_post_bin;
   }
   else if(qualifier_exists("rate_prior"))
   {
      alpha = qualifier("rate_low",sum(cell.pops)/sum(cell.size)/3.);
      beta = qualifier("rate_high",3.*sum(cell.pops)/sum(cell.size));
      lpost_fun = &log_post_evt_rate;
   }
   else if(qualifier_exists("mean_prior"))
   {
      alpha = qualifier("rate_low",sum(cell.pops)/sum(cell.size));
      beta = 1.;
      lpost_fun = &log_post_evt_mean;
   }
   else if(qualifier_exists("alpha"))
   {
      alpha = qualifier("alpha",1);
      beta = 1.;
      lpost_fun = &log_post_evt_alpha;
   }
   else
   {
      alpha = 1.;
      beta = 1.;
      lpost_fun = &log_post_evt;
   }

   if ( alpha < 0. || beta < 0. )
   {
      () = fprintf(fp, "\n Error: qualifier values must be >= 0.\n");
      return;
   }

   variable ans = struct{cpt,last,best,cts,rate,err,lo_t,hi_t};
   variable iw, intvl;
   variable best=[0.], last=[0];

   % Long ago there was a reason I did it this way ...

   best[0] =  max( @lpost_fun( cell.pops[[0:0]],cell.size[[0:0]],
                               ncp_prior, alpha, beta ) );

   % Now onto the meat of the job. Slightly longer than the one line
   % in MATLAB.

   variable besttest;
   variable i, irange;
   for( i=1; i<= num_cells-1 ; i++ )
   {
      irange = [i:0:-1];
      besttest =  [0.,best] +
                  reverse(
                            @lpost_fun( 
                                        cumsum( cell.pops[irange] ),
                                        cumsum( cell.size[irange] ),
                                        ncp_prior, alpha, beta
                                      )
                         );

      best = [best, max(besttest)];
      last = [last, min( where( besttest == best[i] ) )];

      % Is this trigger mode?

      if ( first and last[i] ) 
      {
         ans.best = best;
         ans.last = last;

         ans.cpt = [ 0, last[i] ];
	 ans.cts = Double_Type[2];
	 ans.rate = @ans.cts;
	 ans.err = @ans.cts;
	 ans.lo_t = @ans.cts;
	 ans.hi_t = @ans.cts;

         variable ia = Array_Type[2];
         ia[0] = [ 0: last[i]-1 ];
         ia[1] = [ last[i]: i ];
         i = 0;
	 loop(2)
         {
            ans.cts[i] = sum( cell.pops[ ia[0] ] );
            ans.err[i] = ( 1. + sqrt(0.75 + ans.cts[i] ) ); 

            ans.lo_t[i] = cell.lo_t[ min( ia[0] ) ];
            ans.hi_t[i] = cell.hi_t[ max( ia[0] ) ];

            intvl = sum( ( cell.hi_t[ ia[0] ] - cell.lo_t [ ia[0] ] ) 
                         * cell.dtcor[ ia[0] ] );
            ans.rate[i] = ans.cts[i] / intvl;
            ans.err[i] = ans.err[i] / intvl;

	    i++;
         }
         ans.cts = typecast(ans.cts,Integer_Type);
         return ans;
      }
   }

   % Find the optimum partition

   variable index=last[num_cells-1];
   ans.cpt = [index];
   index = last[index-1];	

   % Long ago there was a reason why I did it this way...

   while ( index )
   {
      ans.cpt = [ index, ans.cpt ];
      index = last[index-1];
   }

   % Always return the first changepoint index as 0, if not already 0
   % (as opposed to MATLAB, the latter happens because of how we
   % initialized 'last').

   if ( ans.cpt[0] )
   {
      ans.cpt = [ 0, ans.cpt ];
   }
   ans.best=best;
   ans.last=last;

   % Make those useful things for plotting, etc.

   variable cpt_length = length(ans.cpt);
   ans.rate = Double_Type[cpt_length];
   ans.err = @ans.rate;
   ans.cts = @ans.rate;
   ans.lo_t = @ans.rate;
   ans.hi_t = @ans.rate;

   ans.lo_t[0] = cell.lo_t[0];
   ans.hi_t[cpt_length-1]=cell.hi_t[length(cell.hi_t)-1];
   i = 0;
   loop ( cpt_length-1 )
   {
      iw = [ ans.cpt[i]:ans.cpt[i+1] - 1 ];

      ans.cts[i] = sum( cell.pops[iw] );
      ans.err[i] = ( 1. + sqrt(0.75 + ans.cts[i] ) ); 

      ans.hi_t[i] = cell.hi_t[ ans.cpt[i+1] - 1 ];
      ans.lo_t[i+1] = cell.lo_t[ ans.cpt[i+1] ];
%
      intvl = sum( ( cell.hi_t[iw] - cell.lo_t [iw] ) * cell.dtcor[iw] );
      ans.rate[i] = ans.cts[i] / intvl;
      ans.err[i] = ans.err[i] / intvl;

      i++;
   }

   % Mop up that endpoint

   iw = [ ans.cpt[i]:length(cell.hi_t) - 1 ];

   ans.cts[i] = sum( cell.pops[iw] );
   ans.err[i] = ( 1. + sqrt(0.75 + ans.cts[i] ) ); 

   intvl = sum( ( cell.hi_t[iw] - cell.lo_t [iw] ) * cell.dtcor[iw] );
   ans.rate[i] = ans.cts[i] / intvl;
   ans.err[i] = ans.err[i] / intvl;

   ans.cts = typecast(ans.cts,Integer_Type);
   return ans;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define readasm_use()
{
   () = fprintf(fp,  "%s\n", %{{{
`
   event = sitar_readasm( file [; tstart=#, tstop=#, maxchi2=#, chnl=#, mjd, jd] );
    
     Variables omitted or set to 0 take on default values.
    
   Inputs:
     file      : ASM filename, e.g. 'xa_cygx1_d1.lc' or 'xa_cygx1_d1.col'

   Qualifiers:
     tstart    : Start time to be read (units of MET or MJD)
     tstop     : Stop time to be read (units of MET or MJD)
     maxchi2   : The maximum reduced chi2 for an 'ASM solution' which
                 will be accepted to consider a data point
     chnl      :  =0 for total band (source.lc files = default)
                 !=0 for three ASM colors + total (source.col files) 
     mjd       : If set changes from default of Mission Elapsed Time (days)
                 to MJD = Julian Date - 2,400,000.5
     jd       : If set changes from default of Mission Elapsed Time (days)
                 to Julian Date.  Supersedes mjd flag.
    
   Outputs: 
     event.time: Time of each event
     event.rate: Total ASM count-rate
     event.err : Uncertainty in count rate
     event.chi2: Chi2 of ASM solution
    
   Optional Outputs:
     event.[ch1,ch2,ch3]_rate: ASM count rate in each channel
     event.[ch1,ch2,ch3]_err : ASM count rate error in each channel
     event.[ch1,ch2,ch3]_chi2: Chi2 for ASM solution in each channel
                               (Overrides event.chi2, which won't be output)
`); %}}} 
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_readasm( )
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_readasm}
%\synopsis{Read and RXTE ASM data file}
%\usage{event = sitar_readasm( file );}
%\qualifiers{
%\qualifier{tstart}{[Start time to be read; units of MET or MJD]}
%\qualifier{tstop}{[Stop time to be read; units of MET or MJD]}
%\qualifier{maxchi2}{[Maximum acceptable chi2 for ASM solution]}
%\qualifier{chnl}{[0 for total band; !=0 for three ASM colors+total]}
%\qualifier{mjd}{[Changes from default of Mission Elapsed Time (days) to MJD]}
%\qualifier{jd}{[Changes from default of Mission Elapsed Time (days) to JD]}
%}
%\description
%   Inputs:
%     file      : ASM filename, e.g. 'xa_cygx1_d1.lc' or 'xa_cygx1_d1.col'
%
%   Outputs: 
%     event.time: Time of each event
%     event.rate: Total ASM count-rate
%     event.err : Uncertainty in count rate
%     event.chi2: Chi2 of ASM solution
%    
%   Optional Outputs:
%     event.[ch1,ch2,ch3]_rate: ASM count rate in each channel
%     event.[ch1,ch2,ch3]_err : ASM count rate error in each channel
%     event.[ch1,ch2,ch3]_chi2: Chi2 for ASM solution in each channel
%                               (Overrides event.chi2, which won't be output)
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
   variable file, event;
   variable tstart  = qualifier("tstart",-1.e32);
   variable tstop   = qualifier("tstop",1.e32);
   variable maxchi2 = qualifier("maxchi2",1.e6);
   variable chnl    = qualifier("chnl",0);
   variable mjd     = qualifier_exists("mjd");
   variable jd      = qualifier_exists("jd");

   if (jd ) mjd=1;

   switch (_NARGS)
   {
    case 1:
      ( file ) = ();
   }
   {
      if(_NARGS != 0)
      {
         () = fprintf(fp, "\n Incorrect number of arguments. \n");
         _pop_n(_NARGS);
      }
      readasm_use();
      return;
   }

   if ( orelse{tstart == NULL}{tstart == 0.})
   {
      tstart = -1.e32;
   }
   if ( orelse{tstop == NULL}{tstop == 0.}{tstop<tstart})
   {
      tstop = 1.e32;
   }
   if ( orelse{maxchi2 == NULL}{maxchi2 <= 0.})
   {
      maxchi2 = 1.e6;
   }

   file = strtrans(file," ","");

   variable t, rate, err, chi2, minchan, maxchan, band;

   (t,rate,err,chi2,minchan,maxchan,band) = 
      fits_read_col(file,"TIME","RATE","ERROR","RDCHI_SQ","MINCHAN","MAXCHAN", "BAND");

   variable mjdrefi = fits_read_key(file,"MJDREFI");
   variable mjdrefr = fits_read_key(file,"MJDREFF");
   mjdrefr = mjdrefr + mjdrefi;

   if ( mjd )     % Convert Mission Elapsed time to MJD if Flag Set
   {
      t += mjdrefr;
   }
   if ( jd )
   {
      t += 2400000.5;
   }

   variable mint = min(t), maxt = max(t);
   if( maxt < tstart or mint > tstop )
   {
      () = fprintf(fp, "\n No times specified within [tstart,tstop]\n");
      () = fprintf(fp, "   [tstart,tstop] = [%e,%e]\n",tstart,tstop);   
      () = fprintf(fp, "   [min(t),max(t)] = [%e,%e]\n",mint,maxt); 
      return NULL;
   }

   variable mincolor = [ 410, 410,1189,1861];
   variable maxcolor = [4750,1189,1861,4750];
   variable ndx;

   if ( chnl != 0 )
   {
      event = struct{time, rate, ch1_rate, ch2_rate, ch3_rate, 
                     err, ch1_err, ch2_err, ch3_err, 
                     ch1_chi2, ch2_chi2, ch3_chi2};

      variable ndxa=where(band=="A");
      variable ndxb=where(band=="B");
      variable ndxc=where(band=="C");

      variable ta = t[ndxa], tb = t[ndxb], tc = t[ndxc];
      variable iw = where( ta == tb and tb == tc );

      if( length(iw) != length(ta) )
      {
         () = fprintf(fp, "\n Color channels not of equal length\n");
	return NULL;
      }

      event.time   = t[ndxa];

      event.ch1_rate =  rate[ndxa];
      event.ch1_err  = err[ndxa];
      event.ch1_chi2 =  chi2[ndxa];

      event.ch2_rate =  rate[ndxb];
      event.ch2_err  = err[ndxb];
      event.ch2_chi2 =  chi2[ndxb];

      event.ch3_rate =  rate[ndxc];
      event.ch3_err  = err[ndxc];
      event.ch3_chi2 =  chi2[ndxc];

      ndx = where( event.ch1_chi2 <= maxchi2 and
                   event.ch2_chi2 <= maxchi2 and
                   event.ch3_chi2 <= maxchi2 and
		   event.time >= tstart and event.time <=tstop );
      if( length(ndx) == 0 )
      {
        () = fprintf(fp, "\n No solutions with chi2 < %f found\n", maxchi2);
        return NULL;
      }

      event.time   =   event.time[ndx];
      event.ch1_rate = event.ch1_rate[ndx];
      event.ch1_err  =  event.ch1_err[ndx];
      event.ch1_chi2 = event.ch1_chi2[ndx];

      event.ch2_rate = event.ch2_rate[ndx];
      event.ch2_err  =  event.ch2_err[ndx];
      event.ch2_chi2 = event.ch2_chi2[ndx];

      event.ch3_rate = event.ch3_rate[ndx];
      event.ch3_err  =  event.ch3_err[ndx];
      event.ch3_chi2 = event.ch3_chi2[ndx];

      event.rate = event.ch1_rate + event.ch2_rate + event.ch3_rate;
      event.err = sqrt( event.ch1_err^2 + 
                        event.ch2_err^2 + 
                        event.ch3_err^2 );
   }
   else
   {
      event = struct{time, rate, err, chi2};

      ndx = where(band=="SUM");

      if( length(ndx) == 0 )
      {
         () = fprintf(fp, "\n Empty lightcurve!\n");
        return NULL;
      }
      event.time =     t[ndx];
      event.rate =  rate[ndx];
      event.err  = err[ndx];
      event.chi2 =  chi2[ndx];

      ndx = where( event.chi2 <= maxchi2 and
		   event.time >= tstart and event.time <= tstop );

      if( length(ndx) == 0 )
      {
        () = fprintf(fp, "\n No solutions with chi2 < %f found\n", maxchi2);
        return NULL;
      }

      event.time = event.time[ndx];
      event.rate = event.rate[ndx];
      event.err  =  event.err[ndx];
      event.chi2 = event.chi2[ndx];
   }

   return event;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


private define reb_rat_use()
{
   () = fprintf(fp,  "%s\n", %{{{
`
   lc = sitar_rebin_rate(t,dt,rate [,err] [;tstart=#,tstop=#,minbin=#,
                                            user_bin=array,weight,delgap]);
    
     Variables in [] are optional, but are order specific unless a qualifier.
    
   Inputs:
     t          : Discrete times at which rates are measured
     dt         : Width of evenly spaced time bins in rebinned lightcurve
                  (Ignored if a user_bin is input.)
     rate       : (Presumed GTI & Exposure Corrected) rates

   Optional Input:
     err        : Error on the rates

   Qualifiers:
     tstart     : Beginning of first output time bin
     tstop      : End of last output time bin
     minbin     : Minimum required events per bin, else the bin is
                  set to 0, or ignored. (Default=1.)
     user_bin.lo: Lower time bounds for user defined grid
     user_bin.hi: Upper time bounds for a user defined grid
                  (Overrides dt, tstart, and tstop inputs)
     delgap     : If set, delete empty bins from the lightcurve,
                  otherwise, set them to 0 (default)
     weight     : If set and err input, the mean is weighted by 
                  the error.  I.e., mean = (\sum rate/err^2)/(\sum 1/err^2),
                  and the output variance becomes 
                  (\sum (mean-rate/err^2)^2)/(\sum 1/err^2)^2
    
   Outputs:
     lc.rate    : Mean rate of resulting rebinning
     lc.bin_lo  : Lower time bounds of rebinned lightcurve
     lc.bin_hi  : Upper time bounds of rebinned lightcurve
     lc.num     : Number of events going into a bin
     lc.var     : Variance of the rate in a time bin. (Only calculated
                  where lc.num >= 2, otherwise set to zero.) 
    
   Optional Outputs:
     lc.err     : Rate errors combined in quadrature, i.e., it's 
                  sqrt{\sum{ err^2 }}/N, where N is the number of points
                  in that particular bin (i.e., lc.num).  Only computed if 
                  err is input [otherwise, just use sqrt(lc.var)].
`);  %}}}    
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_rebin_rate( )
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_rebin_rate}
%\synopsis{Rebin a rate lightcurve}
%\usage{lc = sitar_rebin_rate(t,dt,rate [,err]);}
%\qualifiers{
%\qualifier{tstart}{[Beginning of first output time bin]}
%\qualifier{tstop}{[End of last output time bin]}
%\qualifier{minbin}{[Minimum required events per bin, else the bin is set to 0,  or ignored. Default=1.]}
%\qualifier{user_bin.lo}{[Lower time bounds for user defined grid]}
%\qualifier{user_bin.hi}{[Upper time bounds for a user defined grid. Overrides  dt, tstart, and tstop inputs]}
%\qualifier{delgap}{[If set, delete empty bins from the lightcurve, otherwise,  set them to 0]}
%\qualifier{weight}{[If set and err input, the mean is weighted by the error.  I.e., mean = (\\\\sum rate/err^2)/(\\\\sum 1/err^2), and the output variance becomes (\\\\sum (mean-rate/err^2)^2)/(\\\\sum 1/err^2)^2]}
%}
%\description
%     Variables in [] are optional, but are order specific unless a qualifier.
%    
%   Inputs:
%     t          : Discrete times at which rates are measured
%     dt         : Width of evenly spaced time bins in rebinned lightcurve
%                  (Ignored if a user_bin is input.)
%     rate       : (Presumed GTI & Exposure Corrected) rates
%
%   Optional Input:
%     err        : Error on the rates
%
%   Outputs:
%     lc.rate    : Mean rate of resulting rebinning
%     lc.bin_lo  : Lower time bounds of rebinned lightcurve
%     lc.bin_hi  : Upper time bounds of rebinned lightcurve
%     lc.num     : Number of events going into a bin
%     lc.var     : Variance of the rate in a time bin. (Only calculated
%                  where lc.num >= 2, otherwise set to zero.) 
%    
%   Optional Outputs:
%     lc.err     : Rate errors combined in quadrature, i.e., it's 
%                  sqrt{\\\\sum{ err^2 }}/N, where N is the number of points
%                  in that particular bin (i.e., lc.num).  Only computed if 
%                  err is input [otherwise, just use sqrt(lc.var)].
%\seealso{sitar_bin_events}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
   % Bin a rate lightcurve to a new scale

   variable t, rate; 
   variable err=NULL, dt, custf=0, istck;
   variable lc;

   variable tstart = qualifier("tstart",NULL);
   variable tstop  = qualifier("tstop",NULL);
   variable minbin = qualifier("minbin",1);
   variable cust   = qualifier("user_bin",NULL);
   variable wght   = qualifier_exists("weight");
   variable gap    = qualifier_exists("delgap");

   if ( gap !=0 and minbin == 0 ) { minbin = 1; };

   switch (_NARGS)
   {
    case 3:
      ( t, dt, rate ) = ( );
   }
   {
    case 4:
      ( t, dt, rate, err ) = ( );
   }
   {
      reb_rat_use();
      _pop_n(_NARGS);
      return;
   }

   if ( cust != NULL )
   {
      if ( is_struct_type(cust) == 1 && struct_field_exists(cust,"lo")
                                     && struct_field_exists(cust,"hi") )
      { 
         custf = 1;
         dt = 0;
      }
      else 
      {
         reb_rat_use();
         () = fprintf(fp,  "%s\n", %{{{
`
  Custom user_bin is incorrect.  Input structure requires: 
     variable user_bin = struct{ lo, hi };
`); %}}}
         return;
      }
   }

   if (length(t) != length(rate) )
   {
      reb_rat_use();
      () = fprintf(fp, "\n Incommensurate array lengths for time and rate\n\n");
      return;
   }

   if ( dt == NULL ) dt = 0;

   variable mint = min(t);
   variable maxt = max(t);

   if ( tstart == NULL ) tstart = mint;
   if ( tstop == NULL )  tstop = maxt;

   if( maxt < tstart or mint > tstop or mint > maxt)
   {
      reb_rat_use( );
      () = fprintf(fp, "\n Data outside of bounds of start & stop times:\n");
      () = fprintf(fp, "      [tstart,tstop] = [%e,%e]\n",tstart,tstop);
      () = fprintf(fp, "      [min_t, max_t] = [%e,%e]\n\n",mint,maxt);
      return;
   }

   lc = struct{ rate, bin_lo, bin_hi, num, var };
   variable errf = 0;

   if( err != NULL && length(err) == length(rate) )
   {
      lc = struct_combine(lc,"err");
      errf = 1;
   }
   else if ( err != NULL )
   {
      reb_rat_use();
      () = fprintf(fp, "\n Incommensurate array length for error array\n\n");
      return;
   }

   if( custf == 0 && dt > 0)
   {
      variable nfrac = ( tstop - tstart )/dt;
      variable nbins = int(nfrac);
      lc.bin_lo = [0:nbins-1]*dt+tstart;
      lc.bin_hi = [1:nbins]*dt+tstart;
      if( nfrac-nbins > 0 )
      {
         lc.bin_lo = [lc.bin_lo,lc.bin_hi[nbins-1]];
         lc.bin_hi = [lc.bin_hi,tstop]; 
      }
   }
   else if ( custf )
   {
      lc.bin_lo = @cust.lo;
      lc.bin_hi = @cust.hi;
   }
   else
   {
      reb_rat_use();
      () = fprintf(fp, "\n New time bin or custom bin array must be input.\n\n");
      return;
   }

   variable rev;

   lc.num = histogram(t, lc.bin_lo, lc.bin_hi, &rev);
   lc.rate = Double_Type[length(lc.num)]; 
   lc.var = @lc.rate;
   if ( errf ) lc.err = @lc.rate;

   variable i=0, mom, mome;
   loop( length(lc.num) )
   {
      if ( lc.num[i] )
      {
         if( errf !=0 && wght !=0 )
         {
            mom = moment( rate[rev[i]]/sqr(err[rev[i]]));
            mome = moment( 1/sqr(err[rev[i]]) );
         }
         else
         {
            mom = moment( rate[rev[i]] );
            mome = @mom;
            mome.ave = 1.;
            mome.var = 1.;
         }

         lc.rate[i] = mom.ave/mome.ave;
	 lc.var[i] = mom.var/(mome.ave)^2;

	 if ( errf )
	 {
	    lc.err[i] = sqrt(sum( (err[rev[i]])^2 ))/lc.num[i];
	 }
      }
      i++;
   }
   
   variable idx;
   if ( minbin )
   {
      if ( gap )
      {
         idx = where( lc.num >= minbin );

	 if ( length(idx) )
         {
            lc.rate = lc.rate[idx];
	    lc.bin_lo = lc.bin_lo[idx];
	    lc.bin_hi = lc.bin_hi[idx];
	    lc.var = lc.var[idx];
	    lc.num = lc.num[idx];

            if ( errf ) lc.err = lc.err[idx];
         }
	 else
	 {
            reb_rat_use( );
            () = fprintf(fp, "\n Minimum events requirement yields null lightcurve\n\n");
	    return;
	 }
      }
      else
      {
         idx = where( lc.num < minbin );

         if ( length(idx) )
         {
            lc.rate[idx] = 0.;
            lc.var[idx] = 0.;

            if ( errf )
            {
               lc.err[idx] = 0.;
            }
         }
      }
   }

   return lc;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define pfold_rate_use()
{
   () = fprintf(fp,  "%s\n", %{{{
`
   profile  = sitar_pfold_rate(t,rate,p [;pdot=#,pddot=#,nphs=#,tzero=#,
                                          phase_lo=array, phase_hi=array]);
    
   Inputs:
     t             : Times at which rates are measured
     rate          : Lightcurve rate
     p             : Period on which to fold the lightcurve

   Qualifiers:
     pdot          : Period derivative
     pddot         : Period second derivative
     nphs          : Number of phase bins in output folded lightcurve.
                     Default=20.
     tzero         : Time of zero phase. Default = t[0] 
     phase_lo,     : Arrays for phase bin boundaries (e.g., for uneven phase bins);
     phase_hi      :   must be same length with matched boundaries. Supercedes nphs.
    
   Outputs:
     profile.bin_lo: Start value of phase bin
     profile.bin_hi: Stop value of phase bin
     profile.mean  : Mean rate in phase bin
     profile.var   : Variance of rate in phase bin
     profile.sdm   : Standard deviation of mean rate in a phase bin
     profile.num   : Number of data points in phase bin
`); %}}}    
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define _sitar_pfold_profile(nphs)
{
   variable profile = struct{bin_lo, bin_hi, mean, var, sdm, num};
   if(qualifier_exists("phase_lo") && qualifier_exists("phase_hi"))
   {
      profile.bin_lo = qualifier("phase_lo",[0:19]/20.);
      profile.bin_hi = qualifier("phase_hi",[1:20]/20.);
      nphs = length(profile.bin_lo);
   }
   else
   {
      (profile.bin_lo, profile.bin_hi) = linear_grid(0,1,nphs);
   }
   profile.num  = Integer_Type[nphs];
   profile.mean = Double_Type[nphs];
   profile.var = Double_Type[nphs];
   profile.sdm = Double_Type[nphs];
   return profile;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define _sitar_pfold_rate( t, rate, p, pdot, pddot, nphs, profile)
{
   % Convert time to phase

   variable phs;

   switch ( 2*(pddot!=0)+(pdot!=0) )
   {
    case 0:
      phs = p;
   }
   {
    case 1:
      phs = p + pdot*t/2. ; 
   }
   {
      phs = p + ( pdot/2. + pddot*t/6. )*t ; 
   }

   phs = ( t mod phs ) / phs;

   variable ndx = where( phs < 0. );
   if( length(ndx) > 0 ) 
   {
      phs[ndx] = phs[ndx]+1.;
   }

   variable rev;

   profile.num = histogram(phs, profile.bin_lo, profile.bin_hi, &rev);

   variable i, j, mom;
   _for i (0, nphs-1, 1 )
   {
      j = rev[i];
      if ( length(j) )
      {
         mom = moment( rate[j] );
         profile.mean[i] = mom.ave;
         profile.var[i] = mom.var;
         profile.sdm[i] = mom.sdom;
      } 
   }

   return profile;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_pfold_rate( )
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_pfold_rate}
%\synopsis{Fold a rate-based lightcurve on a given period (with derivatives).}
%\usage{profile  = sitar_pfold_rate(t,rate,p);}
%\qualifiers{
%\qualifier{pdot}{[Period derivative.]}
%\qualifier{pddot}{[Period second derivative.]}
%\qualifier{nphs}{[Number of phase bins in output folded lightcurve.]}
%\qualifier{tzero}{[Time of zero phase. Default = t[0].]}
%\qualifier{phase_lo}{[Arrays for phase bin boundaries (e.g., for uneven phase bins).]}
%\qualifier{phase_hi}{[Arrays must be same length with matched boundaries. Supercedes nphs.]}
%}
%\description
%   Inputs:
%     t             : Times at which rates are measured
%     rate          : Lightcurve rate
%     p             : Period on which to fold the lightcurve
%
%   Outputs:
%     profile.bin_lo: Start value of phase bin
%     profile.bin_hi: Stop value of phase bin
%     profile.mean  : Mean rate in phase bin
%     profile.var   : Variance of rate in phase bin
%     profile.sdm   : Standard deviation of mean rate in a phase bin
%     profile.num   : Number of data points in phase bin
%\seealso{sitar_pfold_event, sitar_epfold_rate}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
   variable t, rate, p, profile;

   variable pdot  = qualifier("pdot",0);
   variable pddot = qualifier("pddot",0);
   variable nphs  = qualifier("nphs",20);
   variable tzero = qualifier("tzero",-1.e32);

   switch (_NARGS)
   {
    case 3:
      ( t, rate, p ) = ( );
   }
   {
      pfold_rate_use();
      _pop_n(_NARGS);
      return;
   }

   variable ltime = length(t), lrate = length(rate);
   if( ltime < 2 || lrate < 2 || ltime != lrate )
   {
      pfold_rate_use();
      () = fprintf(fp, "\n Time and rate must be of equal length > 2\n");
      return;
   }

   if ( orelse{nphs == NULL }{ nphs < 1 }{nphs > ltime} )
   {
      nphs = min([20,ltime]);
   }
   else
   {
      nphs = typecast(nphs,Integer_Type);
   }

   if(qualifier_exists("phase_lo") && qualifier_exists("phase_hi"))
   {
      variable bin_lo = qualifier("phase_lo",[0]);
      variable bin_hi = qualifier("phase_hi",[1]);
      variable lblo = length(bin_lo), lbhi = length(bin_hi);
      if(lblo != lbhi || lblo <= 1)
      {
         pfold_rate_use();
         return;
      }
      if(length(where(bin_lo[[1:lblo-1]]==bin_hi[[0:lbhi-2]])) != lblo-1)
      {
         pfold_rate_use();
         return;
      }
      nphs = lblo;
   }

   if ( orelse{ tzero == NULL }{ tzero < -0.9e18 } )
   {
      tzero = t[0];
   }

   t = t - tzero;

   profile = _sitar_pfold_profile(nphs;;__qualifiers);
   profile = _sitar_pfold_rate( t, rate, p, pdot, pddot, nphs, profile);

   return profile;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define pfold_event_use()
{
    () = fprintf(fp,  "%s\n", %{{{
`
   profile  = sitar_pfold_event(t,p, gti_lo, gti_hi [;pdot=#,pddot=#,nphs=#,tzero=#,
                                                      phase_lo=array, phase_hi=array]);
    
   Inputs:
     t             : Times at which rates are measured
     p             : Period on which to fold the lightcurve
     gti_lo        : Array of lower boundaries of good time intervals
     gti_hi        : Array of upper boundaries of good time intervals

   Qualifiers:
     pdot          : Period derivative
     pddot         : Period second derivative
     nphs          : Number of phase bins in output folded lightcurve. (Default=20)
     tzero         : Time of zero phase. (Default = t[0])
     phase_lo,     : Arrays for phase bin boundaries (e.g., for uneven phase bins);
     phase_hi      :   must be same length with matched boundaries. Supercedes nphs.
    
   Outputs:
     profile.bin_lo: Start value of phase bin
     profile.bin_hi: Stop value of phase bin
     profile.cts   : Counts in phase bin
     profile.var   : Square root of counts in phase bin
     profile.expos : Integrated exposure in phase bin
`); %}}}    
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define _sitar_pfold_event_profile(nphs)
{
   variable profile = struct{bin_lo, bin_hi, cts, var, expos};
   if(qualifier_exists("phase_lo") && qualifier_exists("phase_hi"))
   {
      profile.bin_lo = qualifier("phase_lo",[0:19]/20.);
      profile.bin_hi = qualifier("phase_hi",[1:20]/20.);
      nphs = length(profile.bin_lo);
   }
   else
   {
      (profile.bin_lo, profile.bin_hi) = linear_grid(0,1,nphs);
   }
   profile.cts  = Integer_Type[nphs];
   profile.var = Double_Type[nphs];
   profile.expos = Double_Type[nphs];

   return profile;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define _sitar_pfold_event( t, p, pdot, pddot, nphs, profile, 
                                                  gti_strt, gti_stop )
{
   % Convert time to phase

   variable per, per_gti_strt, per_gti_stop, phs, 
            phs_gti_strt, phs_gti_stop, phs_lo, phs_hi;

   switch ( 2*(pddot!=0)+(pdot!=0) )
   {
    case 0:
      per = p;
      per_gti_strt = p;
      per_gti_stop = p;
   }
   {
    case 1:
      per = p + ( pdot/2. )*t ; 
      per_gti_strt = p + ( pdot/2. )*gti_strt ; 
      per_gti_stop = p + ( pdot/2. )*gti_stop ; 
   }
   {
      per = p + ( pdot/2. + pddot*t/6. )*t ; 
      per_gti_strt = p + ( pdot/2. + pddot*gti_strt/6. )*gti_strt ; 
      per_gti_stop = p + ( pdot/2. + pddot*gti_stop/6. )*gti_stop ; 
   }

   phs = ( t mod per ) / per;

   variable ndx = where( phs < 0 );
   if( length(ndx) ) 
   {
      phs[ndx] = phs[ndx]+1.;
   }

   profile.cts = histogram(phs, profile.bin_lo, profile.bin_hi);
   profile.var = sqrt(profile.cts);

   % The (absolute) start & stop phases of each Good Time Interval
   phs_gti_strt = gti_strt / per_gti_strt;
   phs_gti_stop = gti_stop / per_gti_stop;

   % Loop through each GTI

   variable i, j, phs_gti_min, phs_gti_max, tot_phs, tlo, thi, expos, 
            jfrst, jlast, jprf_strt, jprf_last, isum, phs_jfrst, phs_jlast;

   _for i (0,length(phs_gti_strt)-1,1)
   {
      % The first integer phase before the start of the Good Time
      % Intervals
      if(phs_gti_strt[i]<0)
      {
         phs_gti_min = int(phs_gti_strt[i]-1);
      }
      else
      {
         phs_gti_min = int(phs_gti_strt[i]);
      }
      % The first integer phase after the end of the Good Time
      % Intervals
      if(phs_gti_stop[i]<0)
      {
         phs_gti_max = int(phs_gti_stop[i]);
      }
      else
      {
         phs_gti_max = int(phs_gti_stop[i]+1);
      }

      % The total number of periods bracketing the Good Time Interval
      tot_phs = phs_gti_max - phs_gti_min;
      phs_lo = Double_Type[tot_phs*nphs];
      phs_hi = @phs_lo;
      _for j (0,tot_phs-1,1)
      {
         phs_lo[j*nphs+[[0:nphs-1]]] = j+profile.bin_lo;
         phs_hi[j*nphs+[[0:nphs-1]]] = j+profile.bin_hi;
      }

      phs_lo += phs_gti_min;
      phs_hi += phs_gti_min;

      % Convert the phases to times (including period derivatives to
      % leading order - perhaps slight overkill), and get bin exposure

      switch ( 2*(pddot!=0)+(pdot!=0) )
      {
       case 1:
         tlo = p*phs_lo;
         thi = p*phs_hi;
      }
      {
       case 2:
         tlo = p*phs_lo * (1 + pdot*phs_lo/2);
         thi = p*phs_hi * (1 + pdot*phs_hi/2);
      }
      {
         tlo = p*phs_lo * (1 + (pdot/2+pddot*p*phs_lo/3)*phs_lo);
         thi = p*phs_hi * (1 + (pdot/2+pddot*p*phs_hi/3)*phs_hi);
      }
      expos = thi-tlo;

      % The first full phase bin in the GTI
      jfrst = min(where(phs_lo>=phs_gti_strt[i]));
      % The last full phase bin in the GTI
      jlast = max(where(phs_hi<=phs_gti_stop[i]));

      % Where does the first phase bin start in the profile?
      phs_jfrst = phs_lo[jfrst];

      if(phs_jfrst<0)
      {
         phs_jfrst = (phs_jfrst mod 1) + 1;
      }
      else
      {
         phs_jfrst = (phs_jfrst mod 1);
      }

      % Where does this phase start in the output profile?  (We use
      % the profile hi, rather than the profile lo bin, to minimize
      % roundoff errors.)
      jprf_strt = min(where(profile.bin_hi>phs_jfrst));

      % Loop through all of the phase bins
      _for j (0,nphs-1,1)
      {
         isum = [jfrst+j:jlast:nphs];
         if(length(isum))
         {
            profile.expos[((jprf_strt+j) mod nphs)] = sum(expos[isum]);
         }
      }

      % The end points might have some missed exposure
      % Did we miss some exposure at the start?
      if(phs_gti_strt[i] < phs_lo[jfrst])
      {
         profile.expos[jprf_strt-1] += thi[jfrst-1]-gti_strt[i];
      }

      % Did we miss some exposure at the end?
      if(phs_gti_stop[i] > phs_hi[jlast])
      {
         phs_jlast = phs_hi[jlast];
   
         if(phs_jlast<0)
         {
            phs_jlast = (phs_jlast mod 1) + 1;
         }
         else
         {
            phs_jlast = (phs_jlast mod 1);
         }

         % Next bin center has to be greater than this (end bin, upper
         % value) phase.  I'm not sure if the try/catch is needed -
         % this is in case the last bin wraps around to 0 phase.

         try{ jprf_last = min(where((profile.bin_hi+profile.bin_lo)/2 >phs_jlast)); }
         catch AnyError: { jprf_last = 0; };

         profile.expos[jprf_last] += gti_stop[i]-thi[jlast];
      }
   }
   return profile;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_pfold_event( )
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_pfold_event}
%\synopsis{Fold an event-based lightcurve on a given period (with derivatives).}
%\usage{profile  = sitar_pfold_event(t, p, gti_lo, gti_hi);}
%\qualifiers{
%\qualifier{pdot}{[Period derivative.]}
%\qualifier{pddot}{[Period second derivative.]}
%\qualifier{nphs}{[Number of phase bins in output folded lightcurve.]}
%\qualifier{tzero}{[Time of zero phase. Default = t[0].]}
%\qualifier{phase_lo}{[Arrays for phase bin boundaries (e.g., for uneven phase bins).]}
%\qualifier{phase_hi}{[Arrays must be same length with matched boundaries. Supercedes nphs.]}
%}
%\description
%   Inputs:
%     t             : Times at which rates are measured
%     rate          : Lightcurve rate
%     p             : Period on which to fold the lightcurve
%     gti_lo        : Array of lower boundaries of good time intervals
%     gti_hi        : Array of upper boundaries of good time intervals
%
%   Outputs:
%     profile.bin_lo: Start value of phase bin
%     profile.bin_hi: Stop value of phase bin
%     profile.cts   : Counts in phase bin
%     profile.var   : Square root of counts in phase bin
%     profile.expos : Integrated exposure in phase bin
%\seealso{sitar_pfold_rate, sitar_epfold_rate}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
%
%     Fold events on a given period (plus first and second derivatives)
%
   variable profile;
   variable t, p, pdot, pddot, nphs, tzero, gti_lo, gti_hi;

   switch (_NARGS)
   {
    case 4:
      ( t, p, gti_lo, gti_hi ) = ( );
   }
   {
      pfold_event_use();
      _pop_n(_NARGS);
      return;
   }

   pdot = qualifier("pdot",0);
   pddot = qualifier("pddot",0);
   nphs = qualifier("nphs",0);
   tzero = qualifier("tzero",-1.e18);

   if ( orelse{ nphs == NULL }{ nphs < 1 } )
   {
      nphs = 20;
   }
   else
   {
      nphs = typecast(nphs,Integer_Type);
   }

   if(qualifier_exists("phase_lo") && qualifier_exists("phase_hi"))
   {
      variable bin_lo = qualifier("phase_lo",[0]);
      variable bin_hi = qualifier("phase_hi",[1]);
      variable lblo = length(bin_lo), lbhi = length(bin_hi);
      if(lblo != lbhi || lblo <= 1)
      {
         pfold_event_use();
         return;
      }
      if(length(where(bin_lo[[1:lblo-1]]==bin_hi[[0:lbhi-2]])) != lblo-1)
      {
         pfold_event_use();
         return;
      }
      nphs = lblo;
   }

   if ( orelse{ tzero == NULL }{ tzero < -0.9e18 } )
   {
      tzero = t[0];
   }

   t -= tzero;
   gti_lo -= tzero;
   gti_hi -= tzero;

   profile = _sitar_pfold_event_profile(nphs;;__qualifiers);
   profile = _sitar_pfold_event( t, p, pdot, pddot, nphs, profile, gti_lo, gti_hi);

   return profile;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define fstat(v,dfn,dfd)
{
   % Modelled after IDL f_pdf, which computes the probabilty (p) such
   % that: Probability(X <= v) = p, where X is a random variable from
   % the F distribution with (dfn) and (dfd) degrees of freedom.
   % beta_inc, is taken from the Gnu Scientific Library

   variable x, logq=10.;
   if ( v > 0 )
   {  
      logq = beta_inc( dfd/2.0, dfn/2.0, dfd/( dfd + dfn*v ) );
      logq = log10( logq );
      return logq;
   }
   else
   {
      return 1.;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define epfold_rate_use()
{
   () = fprintf(fp,  "%s\n", %{{{
`
   fold  = sitar_epfold_rate(t,rate [,pstart,pstop] [;nphs=#,nsrch=#,loggrid,prds=array]);
    
     Variables omitted or set to 0 take on default values.
     Variables in [] are optional, but are order specific.
    
   Inputs:
     t         : Times at which rates are measured
     rate      : Lightcurve rate

   (Semi-)Optional Inputs:
     pstart    : Start value of periods to search
     pstop     : Stop value of periods to search

     pstart & pstop  ** or ** prds=[array] must be input

   Qualifiers:
     nphs      : Number of phase bins in folded lightcurve. (Default=20)
     nsrch     : nsrch = number of periods to search. (Default=20)
     loggrid   : If input, the period search grid is logarithmic
     prds      : Alternative array of periods to search
                 prds supersedes pstart & pstop inputs
    
   Outputs:
     fold.prds : Array of trial periods
     fold.lstat: Array of L-statistic for trial period
     fold.lsig : Array of log10 of single trial significance levels.
                 (Note: requires use of the GNU Scientific Library 
                 package, otherwise, defaults to a value of 1.)
     fold.nphs : Array of number of phases (<= nphs) at a given
                 period that had data.
`); %}}}   
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_epfold_rate( )
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\functio{sitar_epfold_rate}
%\synopsis{Epoch fold a rate based lightcurve.}
%\usage{fold  = sitar_epfold_rate(t,rate [,pstart,pstop]);}
%\qualifiers{
%\qualifier{nphs}{[Number of phase bins in folded lightcurve. Default=20.]}
%\qualifier{nsrch}{[Number of periods to search. Default=20.]}
%\qualifier{loggrid}{[If input, the period search grid is logarithmic]}
%\qualifier{prds}{[Alternative array of periods to search prds supersedes pstart/pstop inputs]}
%}
%\description
%     Variables omitted or set to 0 take on default values.
%     Variables in [] are optional, but are order specific.
%    
%   Inputs:
%     t         : Times at which rates are measured
%     rate      : Lightcurve rate
%
%   (Semi-)Optional Inputs:
%     pstart    : Start value of periods to search
%     pstop     : Stop value of periods to search
%
%     pstart & pstop  ** or ** prds=[array] must be input
%
%   Outputs:
%     fold.prds : Array of trial periods
%     fold.lstat: Array of L-statistic for trial period
%     fold.lsig : Array of log10 of single trial significance levels.
%                 (Note: requires use of the GNU Scientific Library 
%                 package, otherwise, defaults to a value of 1.)
%     fold.nphs : Array of number of phases (<= nphs) at a given
%\seealso{sitar_pfold_rate, sitar_pfold_event}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
   variable fold = struct{prds,lstat,lsig,nphs};
   variable t, rate, pstart=NULL, pstop=NULL;

   variable nphs  = qualifier("nphs",20);
   variable nsrch = qualifier("nsrch",20);
   variable prds  = qualifier("prds",NULL);
   variable lgrid = qualifier_exists("loggrid");

   switch (_NARGS)
   {
    case 2:
      ( t, rate ) = ( );
   }
   {
    case 4:
      ( t, rate, pstart, pstop ) = ( );
   }
   {
      epfold_rate_use();
      _pop_n(_NARGS);
      return;
   }

   variable ltime = length(t), lrate = length(rate);
   if ( ltime < 2 || lrate < 2 || ltime != lrate )
   {
      epfold_rate_use();
      () = fprintf(fp, "\n Time and rate must be of equal length > 2 \n");
      return;
   }

   % Going with input periods
   if ( prds == NULL )
   {  
      if ( orelse{ pstart == NULL }{ pstop == NULL }{ pstart<=0 }{ pstop<=0 } )
      {
         epfold_rate_use();
         () = fprintf(fp, "\n Start and stop periods must be set to positive values.\n");
         return;
      }
      else
      {
         variable pers = [pstart,pstop];
         pstart = min(pers);
         pstop = max(pers);
      }

      if ( orelse{ nphs == NULL }{ nphs < 1 }{ nphs > ltime } )
      {
         nphs = min([20,ltime]);
         () = fprintf(fp, "\n Setting nphs=20.\n");
      }
      else
      {
         nphs = typecast(nphs,Integer_Type);
      }

      if ( orelse{ nsrch == NULL }{ nsrch <= 0 } )
      {
         nsrch = 20;
         () = fprintf(fp, "\n Setting nsrch=20.\n");
      }
      else
      {
         nsrch = typecast(nsrch,Integer_Type);
      }

      variable b;
      if ( lgrid )
      {
         (,b) = linear_grid(0.,log10(pstop/pstart),nsrch-1);
         prds = pstart*[1.,10.^b];
      }
      else
      {
         (,b) = linear_grid(pstart,pstop,nsrch-1);
         prds = [pstart,b];
      }
   }
   else
   {
      if ( typeof(prds) != Array_Type )
      {
         epfold_rate_use();
         () = fprintf(fp,"\n Optional periods array 'prds' must be 'Array_Type'");
	 return;
      }
   }

   variable i, prof, lprds=length(prds);
   variable dltime=1.*ltime-nphs;
   variable factor = dltime / ( nphs -1. );

   % No loss of generality with subtracting the mean

   rate = rate - mean(rate);
   fold.lstat = @Double_Type[lprds];
   fold.lsig = @fold.lstat;
   fold.prds = prds;
   factor =  ( ltime - nphs ) / ( nphs -1. );

   prof = _sitar_pfold_profile(nphs);

   _for i (0, lprds-1, 1 )
   {
      prof = _sitar_pfold_rate(t,rate,prds[i],0,0,nphs,prof);

      fold.lstat[i] = factor * ( sum( prof.num * prof.mean^2 ) ) /
                      ( sum( prof.var * ( prof.num - 1. ) ) );

      fold.lsig[i] = fstat( fold.lstat[i], nphs-1, dltime );
   }

   fold.nphs = nphs;

   return fold;
} 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% If the GSL module is loaded, it's fft will overwrite ISIS's, and we
% have to use a different normalization.

private variable gsl_fft = 0;

#ifexists _gsl_fft_complex

gsl_fft = 1;

#endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_avg_cpd()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_avg_cpd}
%\synopsis{Create an averaged cross power spectral density.}
%\usage{(f,psd_a,psd_b,cpd,navg,avg_a,avg_b,) = sitar_avg_cpd(cnts_a,cnts_b,l);}
%\qualifiers{
%\qualifier{dt}{[Length of evenly spaced bins. Default == 1.]}
%\qualifier{times}{[Times of measurements (otherwise presumed to have no gaps).]}
%\qualifier{norm}{[Determine PSD normalization convention. =1, 'Leahy normalization'. Poisson noise PSD == 2, in absence of deadtime effects. !=1, rms or Belloni-Hasinger or Miyamoto normalization, i.e., PSD == (rms)^2/Hz & Poisson noise== 2/Rate.]}
%}
%\description
%     Take an evenly spaced lightcurve (presumed counts vs. time), and
%     calculate the PSD and CPD in segments of length l, averaged over
%     the whole lightcurve.  Segments with data gaps are skipped.
%     Deadtime corrections to Poisson noise are *not* made. Poisson
%     noise is *not* subtracted from average PSDs.
%
%   Inputs:
%     cnts_a/b: Arrays of total counts in each time bin
%     l       : Length of individual PSD segments (use a power of 2!!!)
%
%   Outputs:
%     f       : PSD frequencies ( == 1/Input Time Unit)
%     psd_a/b : Average PSDs 
%     cpd     : Normalized Cross Power Spectral Density (complex array)
%     navg    : Number of data segments going into the averages
%     avg_a/b : Average number of counts per segment of length l
%\seealso{sitar_avg_psd, sitar_lbin_cpd, sitar_lbin_psd, sitar_define_psd, sitar_lags}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
%     Take two evenly spaced lightcurves (presumed counts vs. time),
%     and calculate the PSD and CPD in segments of length l, averaged
%     over the whole lightcurve.  Segments with data gaps are skipped.
%     Deadtime corrections to Poisson noise are *not* made. Poisson
%     noise is *not* subtracted from average PSD.

   variable j,s,f,l,k,alen,ntrans;
   variable transa,transb,transc,stransa,stransb;
   variable freq,atrans=0,btrans=0,ctrans=0.+0.i,ntot=0;
   variable mcounta=0,mcountb=0;
   variable dt, wdt, nwdt, wwdt;

   variable counts_a, counts_b, len;

   variable tbin = qualifier("tbin",1.);
            tbin = qualifier("dt",tbin);
   variable bin_time = qualifier("times",NULL);
   variable donorm = qualifier("norm",1);

   switch(_NARGS)
   {
    case 3:
      (counts_a, counts_b, len) = ();
   }
   {
     () = fprintf(fp, "%s\n", %{{{
`
   (f,psd_a,psd_b,cpd,navg,avg_a,avg_b [,transa,transb,transc] ) 
     = sitar_avg_cpd(cnts_a,cnts_b,l [;dt=#,times=array,norm=#,cospec]);

     Take an evenly spaced lightcurve (presumed counts vs. time), and
     calculate the PSD and CPD in segments of length l, averaged over
     the whole lightcurve.  Segments with data gaps are skipped.
     Deadtime corrections to Poisson noise are *not* made. Poisson
     noise is *not* subtracted from average PSDs.

   Inputs:
     cnts_a/b: Arrays of total counts in each time bin
     l       : Length of individual PSD segments (use a power of 2!!!)

   Qualifiers:
     dt      : Length of evenly spaced bins (default == 1)
     times   : Times of measurements (otherwise presumed to have no gaps)
     norm    : Determine PSD normalization convention.
                =1, 'Leahy normalization'. Poisson noise PSD == 2, in 
                    absence of deadtime effects).
               !=1, 'rms' or 'Belloni-Hasinger' or 'Miyamoto' normalization, 
                    i.e., PSD == (rms)^2/Hz & Poisson noise== 2/Rate.
     cospec   : If present, also output transa/b and transc arrays.

   Outputs:
     f        : PSD frequencies ( == 1/Input Time Unit)
     psd_a/b  : Average PSDs 
     cpd      : Normalized Cross Power Spectral Density (complex array)
     navg     : Number of data segments going into the averages
     avg_a/b  : Average number of counts per segment of length l
     transa/b : Array of Complex Arrays of the individual transforms
     transc   : Array of Complex Arrays of the cross spectrum of AxB*
`);  %}}}
      return;
   }

   len = typecast(len,Long_Type);

   alen = len/2;

   % Lengths of the counts and time arrays, and the number of
   % transforms we will expect to be able to do

   variable lcounts_a, lcounts_b, ltime;

   lcounts_a = length(counts_a);
   lcounts_b = length(counts_b);

   if(bin_time != NULL)
   {
      ltime = length(bin_time);
   }
   else
   {
      bin_time = [0:lcounts_a-1]*tbin;
      ltime = lcounts_a;
   }

   if(lcounts_a != ltime or lcounts_b != lcounts_a) 
   { 
      () = fprintf(fp, "\n Arrays have unequal lengths\n");
      return; 
   }

   if(lcounts_a < len) 
   { 
      () = fprintf(fp, "\n Lightcurve too short for desired segment size\n");
      return; 
   }

   % Guess at total number of transforms
   variable lexp = lcounts_a/len+1;

   variable transa_array = Array_Type[lexp], transb_array = Array_Type[lexp], 
     transc_array = Array_Type[lexp];

   dt = bin_time - shift(bin_time,-1);
   wdt = where( dt > 1.5*tbin );

   if(length(wdt) == 0)
   {
      ntrans = lcounts_a/len;

      _for j (1,ntrans,1)
      {
         s = (j-1)*len;
         f = j*len -1;

         transa = counts_a[[s:f]];
         stransa = sum(transa);
         mcounta = mcounta + stransa;

         transb = counts_b[[s:f]];
         stransb = sum(transb);
         mcountb = mcountb + stransb;
         
%         transa = shift( fft(transa,1) , 1 );
%         atrans = atrans + 2*( abs(transa[[0:alen-1]]) )^2 / stransa;
         transa = (shift( fft(transa,1) , 1 )*sqrt(2/stransa))[[0:alen-1]];
         atrans = atrans + ( abs(transa) )^2;

%         transb = shift( fft(transb,1) , 1 );
%         btrans = btrans + 2*( abs(transb[[0:alen-1]]) )^2 / stransb;
         transb = (shift( fft(transb,1) , 1 )*sqrt(2/stransb))[[0:alen-1]];
         btrans = btrans + ( abs(transb) )^2;


         if ((stransa*stransb) > 0)
         {
            transc = 2*transa*Conj(transb)/sqrt(stransa*stransb);
            ctrans = ctrans + transc;
         }
         else
         {
            transc = 2*transa*Conj(transb)/(sqrt(stransa*stransb*(-1))*(0.+1.0i));
            ctrans = ctrans + transc;
         }

         transa_array[ntot] = transa*sqrt(len);
         transb_array[ntot] = transb*sqrt(len);
         transc_array[ntot] = transc*len;
 
         ntot = ntot + 1;
      }
   }
   else
   {
      nwdt = length(wdt) + 1;
      wwdt = [0,wdt,lcounts_a-1];

      _for k (1,nwdt,1)
      {
         ntrans = ( wwdt[k] - wwdt[k-1] )/len;
            
         _for l (1,ntrans,1)
         {
            s = (l-1) * len + wwdt[k-1];
            f = l * len + wwdt[k-1] - 1;

            transa = counts_a[[s:f]];
            stransa = sum(transa);
            mcounta = mcounta + stransa;

            transb = counts_b[[s:f]];
            stransb = sum(transb);
            mcountb = mcountb + stransb;
         
%            transa = shift( fft(transa,1) , 1 );
%            atrans = atrans + 2*( abs(transa[[0:alen-1]]) )^2 / stransa;
            transa = (shift( fft(transa,1) , 1 )*sqrt(2/stransa))[[0:alen-1]];
            atrans = atrans + ( abs(transa) )^2;

%            transb = shift( fft(transb,1) , 1 );
%            btrans = btrans + 2*( abs(transb[[0:alen-1]]) )^2 / stransb;
            transb = (shift( fft(transb,1) , 1 )*sqrt(2/stransb))[[0:alen-1]];
            btrans = btrans + ( abs(transb) )^2;

            if ((stransa*stransb) > 0)
            {
               transc = 2*transa*Conj(transb)/sqrt(stransa*stransb);
	       ctrans = ctrans + transc;
            }
            else
            {
               transc = 2*transa*Conj(transb)/(sqrt(stransa*stransb*(-1))*(0.+1.0i));
               ctrans = ctrans + transc;
            }

            transa_array[ntot] = transa*sqrt(len);
            transb_array[ntot] = transb*sqrt(len);
            transc_array[ntot] = transc*len;

            ntot = ntot + 1;
         }
      }
   }

   if( ntot == 0 )
   {
      () = fprintf(fp, "\n No transforms performed!\n");
      return;
   }

   transa_array = transa_array[[0:ntot-1]];
   transb_array = transb_array[[0:ntot-1]];
   transc_array = transc_array[[0:ntot-1]];

   freq = [1:alen] / (len * tbin);
   mcounta = mcounta / ntot;
   mcountb = mcountb / ntot;

   % I believe that this will give "Leahy Normalization", i.e.,
   % Poisson noise = 2, in absence of deadtime effects.

   atrans = (atrans * len) / ntot;
   btrans = (btrans * len) / ntot;
   ctrans = (ctrans * len) / ntot;

   ctrans = ctrans[[0:alen-1]];

   if( donorm != 1 )
   {
      atrans = atrans * tbin*len / mcounta;
      btrans = btrans * tbin*len / mcountb;
      ctrans = ctrans * tbin*len / sqrt(mcounta*mcountb);

      _for l (0,ntot-1,1)
      {
         transa_array[l] *= sqrt(tbin*len/mcounta);
         transb_array[l] *= sqrt(tbin*len/mcountb);
         transc_array[l] *= tbin*len/sqrt(mcounta*mcountb);
      }
   }

   if( gsl_fft == 1)
   {
      atrans = atrans/len;
      btrans = btrans/len;
      ctrans = ctrans/len;

      _for l (0,ntot-1,1)
      {
         transa_array[l] /= sqrt(len);
         transb_array[l] /= sqrt(len);
         transc_array[l] /= len;
      }
   }   

   if(qualifier_exists("cospec"))
   {
      return freq, atrans, btrans, ctrans, ntot, mcounta, mcountb, 
             transa_array, transb_array, transc_array;
   }
   else
   {
      return freq, atrans, btrans, ctrans, ntot, mcounta, mcountb;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_avg_psd()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_avg_psd}
%\synopsis{Create an averaged power spectral density.}
%\usage{(f,psd,navg,avg_cnts [,psd_err]) = sitar_avg_psd(cnts,l);}
%\qualifiers{
%\qualifier{dt}{[Length of evenly spaced bins. Default == 1.]}
%\qualifier{times}{[Times of measurements (otherwise presumed to have no gaps).]}
%\qualifier{norm}{[Determine PSD normalization convention. =1, Leahy normalization. Poisson noise PSD == 2, in absence of deadtime effects. !=1, rms or Belloni-Hasinger or Miyamoto normalization, i.e., PSD == (rms)^2/Hz & Poisson noise== 2/Rate.]}
%\qualifier{err}{[If it exists, also return the PSD error calculated directly by assuming each sample has 100% error.]}
%}
%\description
%     Take an evenly spaced lightcurve (presumed counts vs. time), and
%     calculate the PSD in segments of length l, averaged over the
%     whole lightcurve.  Segments with data gaps are skipped.  Deadtime
%     corrections to Poisson noise are *not* made. Poisson noise is
%     *not* subtracted from average PSD.
%
%   Inputs:
%     cnts    : Array of total counts in each time bin
%     l       : Length of individual PSD segments (use a power of 2!!!)
%
%   Outputs:
%     f       : PSD frequencies ( == 1/Input Time Unit)
%     psd     : Average PSD.
%     navg    : Number of data segments going into the average
%     avg_cnts: Average number of counts per segment of length l
%
%   Optional Output:
%     psd_err : Estimated error on the PSD, calculated from lightcurve sample
%\seealso{sitar_avg_cpd, sitar_lbin_cpd, sitar_lbin_psd, sitar_define_psd, sitar_lags}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
%     Take an evenly spaced lightcurve (presumed counts vs. time), and
%     calculate the PSD in segments of length l, averaged over the
%     whole lightcurve.  Segments with data gaps are skipped.
%     Deadtime corrections to Poisson noise are *not* made. Poisson
%     noise is *not* subtracted from average PSD.

   variable j,s,f,l,k,alen,ntrans,trans,strans,freq,atrans=0,atrans_err=0,
            ntot=0,mcount=0,dt,wdt,tmp_trans,nwdt,wwdt,counts,len;

   variable tbin = qualifier("tbin",1.); % Some earlier versions used tbin, 
            tbin = qualifier("dt",tbin); %  allow but overide with dt
   variable bin_time = qualifier("times",NULL);
   variable donorm = qualifier("norm",1);

   switch(_NARGS)
   {
    case 2:
      (counts, len) = ();
   }
   {
      () = fprintf(fp,  "%s\n", %{{{
`
   (f,psd,navg,avg_cnts [,psd_err]) = sitar_avg_psd(cnts,l [;dt=#,times=array,norm=#,err]);

     Take an evenly spaced lightcurve (presumed counts vs. time), and
     calculate the PSD in segments of length l, averaged over the
     whole lightcurve.  Segments with data gaps are skipped.  Deadtime
     corrections to Poisson noise are *not* made. Poisson noise is
     *not* subtracted from average PSD.

   Inputs:
     cnts    : Array of total counts in each time bin
     l       : Length of individual PSD segments (use a power of 2!!!)

   Qualifiers:
     dt      : Length of evenly spaced bins (default == 1)
     times   : Times of measurements (otherwise presumed to have no gaps)
     norm    : Determine PSD normalization convention.
                =1, 'Leahy normalization'. Poisson noise PSD == 2, in 
                    absence of deadtime effects).
               !=1, 'rms' or 'Belloni-Hasinger' or 'Miyamoto' normalization, 
                    i.e., PSD == (rms)^2/Hz & Poisson noise== 2/Rate.
     err     : If it exists, also return the PSD error calculated directly by
               assuming each sample has 100% error 

   Outputs:
     f       : PSD frequencies ( == 1/Input Time Unit)
     psd     : Average PSD.
     navg    : Number of data segments going into the average
     avg_cnts: Average number of counts per segment of length l

   Optional Output:
     psd_err : Estimated error on the PSD, calculated from lightcurve sample
`);   %}}}
      return;
   }

   len = typecast(len,Long_Type);

   alen = len/2;

   % Lengths of the counts and time arrays, and the number of
   % transforms we will expect to be able to do

   variable lcounts, ltime;

   lcounts = length(counts);

   if(bin_time != NULL)
   {
      ltime = length(bin_time);
   }
   else
   {
      bin_time = [0:lcounts-1]*tbin;
      ltime = lcounts;
   }

   if(lcounts != ltime ) 
   { 
      () = fprintf(fp, "\n Arrays have unequal lengths \n");
      return; 
   }

   if(lcounts < len) 
   { 
      () = fprintf(fp, "\n Lightcurve too short for desired segment size \n");
      return; 
   }

   dt = bin_time - shift(bin_time,-1);
   wdt = where( dt > 1.5*tbin );

   if(length(wdt) == 0)
   {
      ntrans = lcounts/len;

      _for j (1,ntrans,1)
      {
         s = (j-1)*len;
         f = j*len -1;

         trans = counts[[s:f]];
         strans = sum(trans);
         mcount = mcount + strans;
         
         trans = shift( fft(trans,1) , 1 );
         tmp_trans = 2*( abs(__tmp(trans)[[0:alen-1]]) )^2 / strans;
         atrans += tmp_trans;
         atrans_err += __tmp(tmp_trans)^2;
           
         ntot = ntot + 1;
      }
   }
   else
   {
      nwdt = length(wdt) + 1;
      wwdt = [0,wdt,lcounts-1];

      _for k (1,nwdt,1)
      {
         ntrans = ( wwdt[k] - wwdt[k-1] )/len;
            
         _for l (1,ntrans,1)
         {
            s = (l-1) * len + wwdt[k-1];
            f = l * len + wwdt[k-1] - 1;

            trans = counts[[s:f]];
            strans = sum(trans);
            mcount = mcount + strans;

            trans = shift( fft(trans,1) , 1 );
            tmp_trans = 2*( abs(__tmp(trans)[[0:alen-1]]) )^2 / strans;
            atrans += tmp_trans;
            atrans_err += __tmp(tmp_trans)^2;

            ntot = ntot + 1;
         }
      }
   }

   if( ntot == 0 )
   {
      () = fprintf(fp, "\n No transforms performed! \n");
      return;
   }

   freq = [1:alen] / (len * tbin);
   mcount = mcount / ntot;

   % I believe that this will give "Leahy Normalization", i.e.,
   % Poisson noise = 2, *** in the absence of deadtime effects ***

   atrans = (atrans * len) / ntot;
   atrans_err = (sqrt(atrans_err) * len) / ntot;

   if( donorm != 1 )
   {
      atrans = atrans * tbin*len / mcount;
      atrans_err = atrans_err * tbin*len / mcount;
   }

   if( gsl_fft == 1)
   {
      atrans = atrans/len;
      atrans_err = atrans_err/len;
   }

   if(qualifier_exists("err"))
   {
      return freq, atrans, ntot, mcount, atrans_err;
   }
   else
   {
      return freq, atrans, ntot, mcount;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_lags()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_lags}
%\synopsis{Create time lags and coherence function from input power spectra.}
%\usage{(lag,dlag,g,dg) = sitar_lags(freq,psda,psdb,cpd,noisea,noiseb [,navg]);}
%\description
%     Given two input PSD (without noise subtracted), their CPD, and
%     their associated noise levels, all as functions of Fourier
%     frequency, calculate the time lag and coherence function (and
%     associated errors) vs. f
%
%   Inputs:
%     freq    : Fourier frequency array
%     psda/b  : Power spectra array associated with two lightcurves
%               (Noise subtraction *not* applied!)
%     cpd     : Cross power spectra array associated with two lightcurves
%     noisea/b: Power spectra noise level. Array *or* a constant.
%
%   Optional Unput:
%     navg    : Number of averages (over independent data segments and frequency
%               bins) associated with each input frequency bin. Can be a constant
%               vs. f (defaulted to 1).
%
%   Outputs:
%     lag     : Time lag vs. frequency. Negative values mean psdb lags psda
%     dlag    : Associated error on the lag.
%     g       : Coherence function
%     dg      : Error on the coherence function
%\seealso{sitar_avg_cpd, sitar_lbin_cpd, sitar_avg_psd, sitar_lbin_psd, sitar_define_psd}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
%     Given two input PSD (without noise subtracted), their CPD, and
%     their associated noise levels, all as functions of Fourier
%     frequency, calculate the time lag and coherence function (and
%     associated errors) vs. f
%
%     For background discussion, see:
%
%        Vaughan, B. A. & Nowak, M. A.  1997, ApJ, 474, L43
%
%     and references therein (especially Bendat & Piersol 1986)
%
   variable freq,psda,psdb,cpd,noisea,noiseb,navg=1;

   switch(_NARGS)
   {
    case 6:
      (freq,psda,psdb,cpd,noisea,noiseb) = ();
   }
   {
    case 7:
      (freq,psda,psdb,cpd,noisea,noiseb,navg) = ();
   }
   {
      () = fprintf(fp,  "%s\n", %{{{
`
   (lag,dlag,g,dg) = sitar_lags(freq,psda,psdb,cpd,noisea,noiseb [,navg]);

     Given two input PSD (without noise subtracted), their CPD, and
     their associated noise levels, all as functions of Fourier
     frequency, calculate the time lag and coherence function (and
     associated errors) vs. f

   Inputs:
     freq    : Fourier frequency array
     psda/b  : Power spectra array associated with two lightcurves
               (Noise subtraction *not* applied!)
     cpd     : Cross power spectra array associated with two lightcurves
     noisea/b: Power spectra noise level. Array *or* a constant.

   Optional Unput:
     navg    : Number of averages (over independent data segments and frequency
               bins) associated with each input frequency bin. Can be a constant
               vs. f (defaulted to 1).

   Outputs:
     lag     : Time lag vs. frequency. Negative values mean psdb lags psda
     dlag    : Associated error on the lag.
     g       : Coherence function
     dg      : Error on the coherence function
`);   %}}}
      return;
   }  

   variable lfreq = length(freq), lnoisea = length(noisea), lnoiseb = length(noiseb),
            lnavg = length(navg);

   if( lnavg==1 )
   {
      navg = ones(lfreq)*navg;
      lnavg = lfreq;
   }

   if( lnoisea==1 )
   {
      noisea = ones(lfreq)*noisea;
      lnoisea = lfreq;
   }

   if( lnoiseb==1 )
   {
      noiseb = ones(lfreq)*noiseb;
      lnoiseb = lfreq;
   }

   if( lnavg == 1 )
   {
      navg = ones(lfreq)*navg;
   }

   variable lpsda = length(psda), lpsdb = length(psdb), lcpd = length(cpd);

   if( lnoisea != lnoiseb or lnoiseb != lpsda or lpsda != lpsdb or
       lpsdb != lcpd or lcpd != lnavg or lnavg != lfreq )
   {
      () = fprintf(fp, "\n Inconsistent array sizes \n");
      return;
   }

   variable sa, sb, ns, gs, dgs, gerr, lag, dlag, acpd = abs(cpd);

   % Define noise subtracted PSDs

   sa = psda - noisea;
   sb = psdb - noiseb;

   ns = ( sa*noiseb + sb*noisea + noisea*noiseb ) / navg;

   % Estimate of intrinsic coherence, high signal to noise

   gs = ( acpd^2 - ns ) / sa / sb;

   % Error on intrinsic coherence, squared, sans Poisson noise

   dgs = 2. * ( 1 - gs )^2 / gs^2;

   % Overall error estimate

   gerr = sqrt( ( 2. * ns^2 * navg / ( acpd^2 - ns )^2 +
                  ((noisea+noiseb)/2)^2 * (1/sa^2 + 1/sb^2) + dgs )
                / navg );

   % Calculate the magnitude of the phase lag
   % (Note: PI is a s-lang intrinsic)

   lag = Real(cpd) / acpd;
   lag = acos( lag ) / 2 / PI / freq * Imag(cpd)/abs( Imag(cpd) );

   % Phase lag errors

   dlag = acpd^2 / psda / psdb;
   dlag = sqrt(1-dlag) / sqrt( dlag * 2 * navg ) / 2 / PI / freq;

   return lag, dlag, gs, dgs;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_lbin_cpd()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_lbin_cpd}
%\synopsis{Logarithmically rebin a cross power spectral density.}
%\usage{(aflo,afhi,apsda,apsdb,acpd,nf,[anoisea,anoiseb]) =  sitar_lbin_cpd(f,psda,psdb,cpd,dff,[noisea,noiseb,&rev]);}
%\description
%     Logarithmically rebin two PSDs (and, optionally, their associated
%     Poisson noise levels) and the associated Cross Power Spectral
%     Density
%
%   Inputs:
%     f        : Array of Fourier frequencies
%     psda/b   : Arrays of PSD values
%     dff      : Delta f/f value for logarithmic bin spacing
%
%   Optional Inputs:
%     noisea/b : Arrays (*or single values*) of Poisson noise levels
%     rev      : If declared (i.e., 'isis> variable rev;') and input,
%                rev returns the reverse indices for the binning (i.e.,
%                (apsda[i] = mean( psda[ rev[i] ] ), etc.)
%
%   Outputs:
%     aflo     : Lower boundary of Fourier frequency bin
%     afhi     : Upper boundary of Fourier frequency bin
%     apsda/b  : Rebinned Power Spectral Density values
%     acpd     : Rebinned Cross Power Spectral Density values
%     nf       : Number of frequencies going into the bin
%     anoisea/b: Rebinned noise level (array, even for single value input)
%\seealso{sitar_avg_cpd, sitar_avg_psd, sitar_lbin_psd, sitar_define_psd, sitar_lags}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
%     Logarithmically rebin two PSDs (and, optionally, their
%     associated Poisson, noise levels) and the associated CPD

   variable freq, psda, psdb, cpd, dff, noisea = NULL, noiseb = NULL, rrev = NULL;

   switch(_NARGS)
   {
    case 5:
      (freq, psda, psdb, cpd, dff) = ();
   }
   {
    case 6:
      (freq, psda, psdb, cpd, dff, noisea) = ();
   }
   {
    case 7:
      (freq, psda, psdb, cpd, dff, noisea, noiseb) = ();
   }
   {
    case 8:
      (freq, psda, psdb, cpd, dff, noisea, noiseb, rrev) = ();
   }
   {
      () = fprintf(fp,  "%s\n", %{{{
`
   (aflo,afhi,apsda,apsdb,acpd,nf,[anoisea,anoiseb]) = 
     sitar_lbin_cpd(f,psda,psdb,cpd,dff,[noisea,noiseb,&rev]);

     Logarithmically rebin two PSDs (and, optionally, their associated
     Poisson noise levels) and the associated Cross Power Spectral
     Density

   Inputs:
     f        : Array of Fourier frequencies
     psda/b   : Arrays of PSD values
     dff      : Delta f/f value for logarithmic bin spacing

   Optional Inputs:
     noisea/b : Arrays (*or single values*) of Poisson noise levels
     rev      : If declared (i.e., 'isis> variable rev;') and input,
                rev returns the reverse indices for the binning (i.e.,
                (apsda[i] = mean( psda[ rev[i] ] ), etc.)

   Outputs:
     aflo     : Lower boundary of Fourier frequency bin
     afhi     : Upper boundary of Fourier frequency bin
     apsda/b  : Rebinned Power Spectral Density values
     acpd     : Rebinned Cross Power Spectral Density values
     nf       : Number of frequencies going into the bin
     anoisea/b: Rebinned noise level (array, even for single value input)
`);  %}}}
      return;
   }

   variable lf, lpsda, lpsdb, lcpd, lnoisea, lnoiseb;
   variable fmin, fmax, nmax, flo, fhi, iw, rev;
   variable nfreq, afreq, afreqlo, afreqhi, apsda, apsdb, acpd, 
            anoisea, anoiseb;

   lf = length(freq);
   lpsda = length(psda);
   lpsdb = length(psdb);
   lcpd = length(cpd);

   if( noisea != NULL )
   {
      lnoisea = length(noisea);
      if( lnoisea == 1 )
      {
         noisea = ones(lf) * noisea;
      }
      else if( lnoisea != lpsda )
      {
         () = fprintf(fp, "\n Noise A and PSD arrays have unequal lengths \n");
         return; 
      }
   }

   if( noiseb != NULL )
   {
      lnoiseb = length(noisea);
      if( lnoisea == 1 )
      {
         noiseb = ones(lf) * noiseb;
      }
      else if( lnoiseb != lpsdb )
      {
         () = fprintf(fp, "\n Noise B and PSD arrays have unequal lengths \n");
         return; 
      }
   }

   if( lf != lpsda or lpsda != lpsdb or lpsdb != lcpd )
   {
      () = fprintf(fp, "\n Inconsistent Array Sizes \n");
      return;
   }

   fmin = min(freq);
   fmax = max(freq);

   nmax = typecast((log10(fmax)-log10(fmin))/log10(1+dff)+0.5,Integer_Type);

   flo = fmin * 10.^([0:nmax-1]*log10(1+dff));
   fhi = make_hi_grid(flo);

   nfreq = histogram(freq, flo, fhi, &rev);

   iw = where(nfreq !=0);

   variable liw = length(iw);

   nfreq = nfreq[iw];
   afreqlo = Double_Type[liw];
   apsda = @afreqlo;
   apsdb = @afreqlo;
   acpd  = Complex_Type[liw];

   rev = rev[iw];

   variable i;

   if( noisea == NULL and noiseb == NULL)
   {
      _for i (0, liw-1, 1)
      {
         afreqlo[i] = min(freq[rev[i]]);
         apsda[i] = sum( psda[rev[i]] );
         apsdb[i] = sum( psdb[rev[i]] );
         acpd[i] = sum( cpd[rev[i]] );
      }
   } 
   else if( noiseb == NULL )
   {
      anoisea = @afreqlo;
      anoiseb = NULL;

      _for i (0, liw-1, 1)
      {
         afreqlo[i] = min(freq[rev[i]]);
         apsda[i] = sum( psda[rev[i]] );
         apsdb[i] = sum( psdb[rev[i]] );
         acpd[i] = sum( cpd[rev[i]] );
         anoisea[i] = sum( noisea[rev[i]] );
      }

      anoisea = anoisea / nfreq;
   }
   else if( noisea == NULL )
   {
      anoiseb = @afreqlo;
      anoisea = NULL;

      _for i (0, liw-1, 1)
      {
         afreqlo[i] = min(freq[rev[i]]);
         apsda[i] = sum( psda[rev[i]] );
         apsdb[i] = sum( psdb[rev[i]] );
         acpd[i] = sum( cpd[rev[i]] );
         anoiseb[i] = sum( noiseb[rev[i]] );
      }

      anoiseb = anoiseb / nfreq;
   }   
   else
   {
      anoisea = @afreqlo;
      anoiseb = @afreqlo;

      _for i (0, liw-1, 1)
      {
         afreqlo[i] = min(freq[rev[i]]);
         apsda[i] = sum( psda[rev[i]] );
         apsdb[i] = sum( psdb[rev[i]] );
         acpd[i] = sum( cpd[rev[i]] );
         anoisea[i] = sum( noisea[rev[i]] );
         anoiseb[i] = sum( noiseb[rev[i]] );
      }

      anoisea = anoisea / nfreq;
      anoiseb = anoiseb / nfreq;
   }   


   afreqhi = make_hi_grid(afreqlo);
   afreqhi[liw-1] = max(freq[rev[liw-1]]);

   apsda = apsda / nfreq;
   apsdb = apsdb / nfreq;
   acpd = acpd / nfreq;

   if(rrev != NULL)
   {
     @rrev = rev;
   }

   if( noisea == NULL and noiseb == NULL )
   {
      return afreqlo, afreqhi, apsda, apsdb, acpd, nfreq;
   } 
   else
   {
      return afreqlo, afreqhi, apsda, apsdb, acpd, nfreq, 
             anoisea, anoiseb;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_lbin_psd()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_lbin_psd}
%\synopsis{Logarithmically rebin a power spectral density.}
%\usage{(aflo,afhi,apsd,nf [,anoise,apsd_err]) = sitar_lbin_psd(f,psd,dff);}
%\qualifiers{
%\qualifier{noise}{[Array (*or single value*) of Poisson noise levels.]}
%\qualifier{rev}{[If declared (i.e., 'isis> variable rev;') and input, rev returns the reverse indices for the binning (i.e., af[i] = mean( freq[ rev[i] ] ), etc.).]}
%\qualifier{psd_err}{[Array of errors for the PSD, in which case a PSD error array will be returned.]}
%}
%\description
%     Logarithmically rebin a PSD (and, optionally, its associated
%Poisson noise and error values)
%
%   Inputs:
%     f      : Array of Fourier frequencies
%     psd    : Array of PSD values
%     dff    : Delta f/f value for logarithmic bin spacing 
%
%   Outputs:
%     aflo   : Lower bounds of Fourier frequency bins
%     afhi   : Upper bounds of Fourier frequency bins
%     apsd   : Rebinned PSD values
%     nf     : Number of frequencies going into the bin
%
%   Optional Outputs:
%     anoise  : Rebinned noise level (array, even for single value input)
%     apsd_err: Rebinned PSD error array
%\seealso{sitar_avg_cpd, sitar_lbin_cpd, sitar_avg_psd, sitar_define_psd, sitar_lags}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
%     Logarithmically rebin a PSD (and, optionally, its associated
%     Poisson noise)

   variable freq, psd, dff, noise = NULL, rrev = NULL, psd_err = NULL;

   switch(_NARGS)
   {
    case 3:
      (freq, psd, dff) = ();
   }
% For backwards compatibility, maintain these two choice has "hidden"
% options
   {
    case 4:
      (freq, psd, dff, noise) = ();
   }
   {
     case 5:
      (freq, psd, dff, noise, rrev) = ();
   }
   {
      () = fprintf(fp, "%s\n", %{{{
`


   (aflo,afhi,apsd,nf [,anoise,apsd_err]) = sitar_lbin_psd(f,psd,dff;
                            noise=value, rev_indx=&Array, err=Array);

     Logarithmically rebin a PSD (and, optionally, its associated
Poisson noise and error values)

   Inputs:
     f      : Array of Fourier frequencies
     psd    : Array of PSD values
     dff    : Delta f/f value for logarithmic bin spacing 

   Qualifiers:
     noise  : Array (*or single value*) of Poisson noise levels
     rev    : If declared (i.e., 'isis> variable rev;') and input,
              rev returns the reverse indices for the binning (i.e.,
              (af[i] = mean( freq[ rev[i] ] ), etc.)
     err    : Array of errors for the PSD, in which case a PSD error
              array will be returned

   Outputs:
     aflo   : Lower bounds of Fourier frequency bins
     afhi   : Upper bounds of Fourier frequency bins
     apsd   : Rebinned PSD values
     nf     : Number of frequencies going into the bin

   Optional Outputs:
     anoise  : Rebinned noise level (array, even for single value input)
     apsd_err: Rebinned PSD error array
`);   %}}}
      return;
   }

   noise = qualifier("noise",noise);
   rrev = qualifier("rev_indx",rrev);
   psd_err = qualifier("err",NULL);

   variable lf, lpsd, lnoise, fmin, fmax, nmax, flo, fhi, iw, rev,
   nfreq, lnf, afreq, afreqlo, afreqhi, apsd, anoise, apsd_err;

   lf = length(freq);
   lpsd = length(psd);

   if(lf != lpsd) 
   { 
      () = fprintf(fp, "\n Frequency and PSD arrays have unequal lengths \n");
      return; 
   }

   if(psd_err != NULL) 
   { 
      if(lpsd != length(psd_err))
      {
         () = fprintf(fp, "\n PSD and error arrays have unequal lengths \n");
         return; 
      }
   }

   if( noise != NULL )
   {
      lnoise = length(noise);
      if( lnoise == 1 )
      {
         noise = ones(lpsd)*noise;
      }
      else if( lnoise != lpsd )
      {
         () = fprintf(fp, "\n Noise and PSD arrays have unequal lengths \n");
         return; 
      }
   }

   fmin = min(freq);
   fmax = max(freq);

   nmax = typecast((log10(fmax)-log10(fmin))/log10(1+dff)+0.5,Integer_Type);

   flo = fmin * 10.^([0:nmax-1]*log10(1+dff));
   fhi = make_hi_grid(flo);

   nfreq = histogram(freq, flo, fhi, &rev);

   iw = where(nfreq !=0);

   nfreq = nfreq[iw];
   lnf = length(nfreq);

   apsd = Double_Type[lnf];
   apsd_err = @apsd;
   anoise = @apsd;
   afreqlo = @apsd;
   afreqhi = @apsd;

   rev = rev[iw];

   variable i;

   _for i (0, lnf-1, 1)
   {
      afreqlo[i] = min( freq[rev[i]]);
      apsd[i] = sum( psd[rev[i]] );

      if( noise != NULL ) anoise[i] = sum( noise[rev[i]] )/nfreq[i];

      if( psd_err != NULL ) apsd_err[i] = sqrt(sum( (psd_err[rev[i]])^2 ))/nfreq[i];
   } 

   afreqhi = make_hi_grid(afreqlo);
   afreqhi[lnf-1] = max(freq[rev[lnf-1]]);

   apsd = apsd / nfreq;

    if(rrev != NULL)
    {
      @rrev = rev;
    }

   if( noise != NULL && psd_err != NULL)
   {
      return afreqlo, afreqhi, apsd, nfreq, anoise, apsd_err;
   } 
   else if( noise != NULL )
   {
      return afreqlo, afreqhi, apsd, nfreq, anoise;
   }
   else if( psd_err != NULL )
   {
      return afreqlo, afreqhi, apsd, nfreq, apsd_err;
   }
   else
   {
      return afreqlo, afreqhi, apsd, nfreq;
   }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_define_psd()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_define_psd}
%\synopsis{Store a power spectral density as a fittable dataset.}
%\usage{id = sitar_define_psd(flo,fhi,psd,err [,noise]);}
%\qualifiers{
%\qualifier{noise}{[Same as optional input, noise.]}
%}
%\description
%   Inputs:
%     f_lo   : Low value of PSD frequency bin (Hz -> keV)
%     f_hi   : High value of PSD frequency bin (Hz -> keV)
%     psd    : Array of PSD values (Power -> Counts/bin)
%     err    : Array of PSD Errors
%
%   Optional Input:
%     noise  : Array (*or single value*) of Poisson noise levels.
%              If undefined, background is undefined.
%
%   Outputs:
%     id     : Dataset Index
%\seealso{sitar_avg_cpd, sitar_lbin_cpd, sitar_avg_psd, sitar_define_psd, sitar_lags}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
%     Define a PSD to be a fittable dataset in ISIS, with frequency
%     values set to keV.

   variable f_lo, f_hi, psd, err, noise=NULL;

   switch(_NARGS)
   {
    case 4:
      (f_lo, f_hi, psd, err) = ();
   }
   {
    case 5:
      (f_lo, f_hi, psd, err, noise) = ();
   }   
   {
      () = fprintf(fp,  "%s\n", %{{{
`
   id = sitar_define_psd(flo,fhi,psd,err [,noise]; noise=# or [Array]);

     Define psd to be an ISIS counts data set.

   Inputs:
     f_lo   : Low value of PSD frequency bin (Hz -> keV)
     f_hi   : High value of PSD frequency bin (Hz -> keV)
     psd    : Array of PSD values (Power -> Counts/bin)
     err    : Array of PSD Errors

   Optional Input:
     noise  : Array (*or single value*) of Poisson noise levels.
              If undefined, background is undefined.

   Qualifiers:
     noise  : Same as above.  Supercedes optional input.

   Outputs:
     id     : Dataset Index
`);
      return;
   }

   variable id;

   noise = qualifier("noise",noise);

   if(noise !=NULL)
   {
      variable lnoise = length(noise); 
      if( lnoise == 1 )
      {
         noise = ones(length(psd))*noise;
         id = define_counts(_A(f_hi),_A(f_lo),reverse(psd-noise),
                            reverse(err));
      }
      else if(lnoise != length(psd))
      {
         print("Noise array length differs from psd length");
         return;
      }
      else
      {
         id = define_counts(_A(f_hi),_A(f_lo),reverse(psd-noise),
                            reverse(err));
      }
   }
   else
   {
      id = define_counts(_A(f_hi),_A(f_lo),reverse(psd),reverse(err));
   }
   return id;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The GSL library has to be imported in order to run this routine, so,
% do a check, and if it exists, load the Gregory-Loredo function.

#ifexists gsl->interp_akima

% t = event times; tmin & tmax =  max & min time of lightcurve,
% mmax = maximum partition number in trial lightcurves
% ta, adt = effective area/deadtime times and values
% dodither = Toggle for applying dither correction to lightcurve
% thresh = helps determine truncation of partitioned lightcurve

public define sitar_glvary()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sitar_glvary}
%\synopsis{Create an optimal lightcurve using the Gregory-Loredo algorithm.}
%\usage{gl = sitar_glvary(t);}
%\qualifiers{
%\qualifier{tmin}{[The minimum time to consider. Default=min(t)-1]}
%\qualifier{tmax}{[The maximum time to consider. Default=max(t)+1]}
%\qualifier{mmax}{[Consider lightcurve divisions from 2 to mmax evenly spaced bins. Default=300.]}
%\qualifier{thresh }{[Truncate the number of partitions by ignoring those for which \\\\sum_m(odds ratio)<max(\\\\sum_n(odds ratio))/exp(thresh), where n = 2 -> mmax.  thresh sets a minimum probability: (1-p_min) ~ exp(thresh)*(1-p_peak). Default=2.]}
%\qualifier{nbins}{[Create an output lightcurve with nbins. Default=mmax.]}
%\qualifier{texp, frac_exp}{[Array pair that give for fractional exposure as a function of time, which will be interpolated and used to correct lightcurve rates. Arrays must have a minimum of five entries. Default is for no correction.]}
%}
%\description
%     Use the Gregory-Loredo algorithm to find the odds ratios that
%     even divisions of a lightcurve are better descriptions of the
%     data than a constant lightcurve.  A 'best estimate' lightcurve
%     can also be output for a lightcurve with nbins.
%
%   Inputs:
%     t      : Array of event times
%
%   Outputs:
%     gl.p         : Total probability that some evenly partitioned lightcurve, with up
%                    to gl.mmax bins, is a better description than a constant lightcurve
%     gl.ppart     : The probability for an individual evenly partitioned lightcurve
%                    that it is a better description than a constant lightcurve
%     gl.lodds_sum : The natural logarithm of the sum of the odds ratios comparing 
%                    lightcurves with two or more partitions to a constant lightcurve.
%                    gl.p == exp(gl.lodds_sum)/[1+exp(gl.lodds_sum)]
%     gl.mpeak     : The number of partitions for the evenly partitioned lightcurve 
%                    with the maximum probability.
%     gl.mmax      : The maximum number of partitions actually used (influenced by
%                    the setting of the thresh parameter)
%     gl.m         : The number of partitions corresponding to each evenly partitioned
%                    lightcurve considered (=[2:mmax])
%     gl.pm        : Total probability that some evenly partitioned lightcurve
%                    is a better description than a constant lightcurve for each
%                    maximum number of partitions considered ([2:mmax])
%     gl.nj        : The counts histogram corresponding to each partitioning above
%     gl.aj        : The integrated fractional exposure for each partitioning above
%     gl.a_avg     : The averaged fractional exposure.
%     gl.tmin      : The value of tmin actually used (maximum of [tmin,min(texp)]);
%     gl.tmax      : The value of tmax actually used (minimum of [tmax,max(texp)]);
%     gl.tlc       : The output lightcurve times (an array with input nbins bins)
%     gl.rate      : Best estimate of the lightcuve rates at the above times
%     gl.erate     : Best estimate of the lightcuve rate errors at the above times
%\seealso{sitar_global_optimum}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
   variable t, dodither=0; 

   switch(_NARGS)
   {
    case 1:
      t = ();
   }
   {
      () = fprintf(fp, "%s\n", %{{{
`
   gl = sitar_glvary(t [;tmin=#, tmax=#, mmax=#, thresh=#,
                         nbins=#, texp=array, frac_exp=array]);

     Use the Gregory-Loredo algorithm to find the odds ratios that
     even divisions of a lightcurve are better descriptions of the
     data than a constant lightcurve.  A 'best estimate' lightcurve
     can also be output for a lightcurve with nbins.

   Inputs:
     t      : Array of event times

   Qualifiers:
     tmin     : The minimum time to consider (Default=min(t)-1)
     tmax     : The maximum time to consider  (Default=max(t)+1)
     mmax     : Consider lightcurve divisions from 2 to mmax evenly spaced 
                bins. (Default=300)
     thresh   : Truncate the maximum number of partitions considered by ignoring
                partitionings for which--
                   \sum_m(odds ratio) < max(\sum_n(odds ratio))/exp(thresh) , 
                where n = 2 -> mmax.  I.e., the more partitionings you have,
                the less significant the results tend to become.  thresh
                sets a minimum probability: (1-p_min) ~ exp(thresh)*(1-p_peak).
                (Default=2.)
     nbins    : Create an output lightcurve with nbins. (Default=mmax)
     texp,    : A pair of arrays that give values for the fractional exposure
     frac_exp   as a function of time.  *** Only non-zero values of exposure 
                will be retained ***, which will then be interpolated and used
                to correct the lightcurve rates. These arrays must have a
                minimum of five entries. (Default is for no correction.)

   Outputs:
     gl.p         : Total probability that some evenly partitioned lightcurve, with up
                    to gl.mmax bins, is a better description than a constant lightcurve
     gl.ppart     : The probability for an individual evenly partitioned lightcurve
                    that it is a better description than a constant lightcurve
     gl.lodds_sum : The natural logarithm of the sum of the odds ratios comparing 
                    lightcurves with two or more partitions to a constant lightcurve.
                    gl.p == exp(gl.lodds_sum)/[1+exp(gl.lodds_sum)]
     gl.mpeak     : The number of partitions for the evenly partitioned lightcurve 
                    with the maximum probability.
     gl.mmax      : The maximum number of partitions actually used (influenced by
                    the setting of the thresh parameter)
     gl.m         : The number of partitions corresponding to each evenly partitioned
                    lightcurve considered (=[2:mmax])
     gl.pm        : Total probability that some evenly partitioned lightcurve
                    is a better description than a constant lightcurve for each
                    maximum number of partitions considered ([2:mmax])
     gl.nj        : The counts histogram corresponding to each partitioning above
     gl.aj        : The integrated fractional exposure for each partitioning above
     gl.a_avg     : The averaged fractional exposure.
     gl.tmin      : The value of tmin actually used (maximum of [tmin,min(texp)]);
     gl.tmax      : The value of tmax actually used (minimum of [tmax,max(texp)]);
     gl.tlc       : The output lightcurve times (an array with input nbins bins)
     gl.rate      : Best estimate of the lightcuve rates at the above times
     gl.erate     : Best estimate of the lightcuve rate errors at the above times
`);   %}}}
      return;
}

   variable tmin     = qualifier("tmin",min(t)-1.);
   variable tmax     = qualifier("tmax",max(t)+1.);
   variable mmax     = qualifier("mmax", 300);
   variable thresh   = qualifier("thresh",2);
   variable ta       = qualifier("texp",NULL);
   variable adt      = qualifier("frac_exp",NULL);
   variable nbins    = qualifier("nbins",mmax);

   if(length(ta)==length(adt) && length(adt) >=5) dodither=1;

   if(dodither)
   {
      variable inz = where(adt>0);
      variable adt_use = [inz[0]:inz[length(inz)-1]];

      ta = ta[adt_use];
      adt = adt[adt_use];

      variable na = length(adt);
      if(na < 5)
      {
         () = fprintf(fp,"\n Fractional exposure array is too short. \n");
         return;
      }

      if(tmin < ta[0]){ tmin = ta[0]; }
      if(tmax > ta[na-1]){ tmax = ta[na-1]; }
   }

   % Only look at times within tmin & tmax

   t = t[where(t>=tmin and t<tmax)];

   variable j;
   variable n=length(t);
   variable dt_int = tmax - tmin;
   variable a_avg=0.;

   if(dodither)
   {
      % Define the spline of the effective area curve

      variable spline_adt = gsl->interp_akima_init(ta,adt);
      a_avg = log(gsl->interp_eval_integ(spline_adt,tmin,tmax)/dt_int);

      % Do integration of spline in neighboring points, then sum

      variable iadt = Double_Type[na];
      _for j (1,na-1,1)
      {
         iadt[j] = gsl->interp_eval_integ(spline_adt,ta[j-1],ta[j]);
      }
      iadt=cumsum(iadt);

      % Define the spline of the integrated effective area curve

      variable spline_iadt = gsl->interp_akima_init(ta,iadt);
   }

   variable nj = Array_Type[int(mmax)-1];
   variable aj = @nj;
   variable m=[2:int(mmax):1];
   variable lods=Double_Type[int(mmax)-1];

   variable lo,hi,im;

   _for im (2,int(mmax),1)
   {
      % Create grid for partitioning lightcurve into im bins

       lo = [0:im-1];
       hi = [1:im];
       lo = __tmp(lo)*(dt_int/im);
       hi = __tmp(hi)*(dt_int/im);

      if(dodither)
      {
         % For each partitioning, create average deadtime/effective
         % area per bin.  Dead bins are set equal to the average 
         % effective area, so as not to contribute to the sum 

         aj[im-2] = (gsl->interp_eval(spline_iadt,hi)
                    -gsl->interp_eval(spline_iadt,lo))*(im/dt_int);

         aj[im-2][where(aj[im-2]==0.)] = a_avg;
      }
      else
      {
         aj[im-2] = Double_Type[im]+1;
      }

      % For each lightcurve partition, the arrays of counts per bin

      nj[im-2] = histogram(t,lo,hi);

      % The odds ratio for each partitioning.  The nj[im-2]*()
      % term is removed if effective area variation is unimportant

      if(dodither)
      {
         lods[im-2] = sum( nj[im-2]*(a_avg-log(aj[im-2])) + 
                           lngamma(nj[im-2]+1) ) +   
                      n*log(im) + lngamma(im) - lngamma(n+im);
      }
      else 
      {
         lods[im-2] = sum( lngamma(nj[im-2]+1) ) +
                      n*log(im) + lngamma(im) - lngamma(n+im);
      }
   }

   % This bit uses the return pieces from above to find what the
   % maximum odds ratio is, and temporarily takes that out (lomax),
   % and then truncates the number of lightcurve binnings kept

   variable iw = where(lods==max(lods));
   variable lomax = lods[min(iw)];

   lods = exp(lods-lomax);
   variable csum = cumsum(lods)/(m-1);

   % Truncate the number of partitionings of the lightcurve

   if(thresh < 0) thresh==2.;
   variable imax = max(where(csum >= max(csum)/exp(thresh)));

   % Return the probability (p), the odds ratio for each partioning
   % (ppart), the log of the summed odds (lodds_sum), the # of
   % partitions for the peak odds ratio (mpeak), the # of partitions
   % for each (m), the histogrammed counts for each partitioning (nj),
   % the integrated and averaged effective area for each partioning 
   %(aj, a_avg), and the used min/max times (tmin/tmax)

   variable gl_struct=   
     struct{p, ppart, lodds_sum, mpeak, mmax, pm, m, nj,
            aj, a_avg, tmin, tmax, tlc, rate, erate};

   % The # of partitions of the lightcurve with the highest odds ratio.

   gl_struct.mpeak = min(iw)+2;

   % Calculate the total probability of variability (p) and the
   % probability for each partitioning of the lightcurve (ppart).

   variable msum = csum[imax];
   variable lsum = log(msum) + lomax;
   gl_struct.p = 1/(exp(-lsum)+1);
   gl_struct.ppart = __tmp(lods)[[0:imax]]/
                            ((imax+1)*(msum+exp(-lomax)));
   gl_struct.pm = 1/(exp(-lomax)/csum+1);

   % The rest of the return values

   gl_struct.lodds_sum = lsum;
   gl_struct.mmax = imax+2;
   gl_struct.m = __tmp(m);
   gl_struct.nj = __tmp(nj);
   gl_struct.aj = __tmp(aj);
   gl_struct.a_avg = exp(a_avg);
   gl_struct.tmin = tmin;
   gl_struct.tmax = tmax;

   % Return a lightcurve with nbin bins

   variable tfrac = [0:nbins-1]/(nbins*1.);
   variable rate = Double_Type[nbins];
   variable erate=@rate;

   % We might not have used all the data ...

   variable ntot=sum(gl_struct.nj[0]); 

   variable nj_ii_k, oratio_ii, aj_ii_k, drate;

   % Best estimate of the lightcurve is calculated here.
   % Fractional area/deadtime correction is included.

   variable ii;
   _for ii (0,imax,1)
   {
      variable mloop=ii+2;
      variable k = int( tfrac*mloop );

      nj_ii_k=gl_struct.nj[ii][k]; 
      oratio_ii=gl_struct.ppart[ii];
      aj_ii_k = gl_struct.aj[ii][k];
      drate = (mloop/(ntot+mloop))*
              (__tmp(oratio_ii)*(nj_ii_k+1)/aj_ii_k);
      rate = __tmp(rate) + drate; 
      erate = __tmp(erate) + (mloop/(ntot+mloop+1))*
              (__tmp(drate)*(__tmp(nj_ii_k)+2)/__tmp(aj_ii_k));
   }

   % Here I differ from Gregory & Loredo, who only include the
   % variable part of the lightcurve (i.e., partitionings with >=2
   % bins), reasonable for p~1. But p can be lower, therefore the
   % estimated constant lightcurve part should be included.

   rate = __tmp(rate) + (1.-gl_struct.p)/gl_struct.a_avg;
   erate = __tmp(erate) + (1.-gl_struct.p)/gl_struct.a_avg^2;
   erate = sqrt(__tmp(erate)-rate^2);
 
   gl_struct.tlc = __tmp(tfrac)*(tmax-tmin)+tmin;
   gl_struct.rate = __tmp(rate)*(ntot/(tmax-tmin));
   gl_struct.erate = __tmp(erate)*(ntot/(tmax-tmin));

   return gl_struct; 
}
#endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


private define reb_evt_use()
{
   () = fprintf(fp,  "%s\n", %{{{
`
   lc = sitar_bin_events(t,dt,gti_lo,gti_hi [;tstart=#,tstop=#,obin=#,
                                              user_bin=array,delgap]);
    
   Inputs:
     t          : Discrete times at which rates are measured
     dt         : Width of evenly spaced time bins in rebinned lightcurve
                  (Ignored if a user_bin is input.)

   Qualifiers:
     tstart     : Beginning of first output time bin (default: min(t))
     tstop      : End of last output time bin (default: max(t))
     obin       : For a tstop, last bin width might be < dt.  Include this
                  "overflow" bin if width is > obin*dt (default: obin=0.1)
     user_bin.lo: Lower time bounds for user defined grid
     user_bin.hi: Upper time bounds for a user defined grid
                  (Overrides dt, tstart, and tstop inputs)
     minexp     : Do not include bins with exposure < minexp (default: 0)
     delgap     : If set, delete 0 exposure bins from the lightcurve,
                  otherwise set their rate & error to 0 (default)
    
   Outputs:
     lc.rate    : Mean rate of resulting binning
     lc.err     : Error on bin rate (minimum = -log(0.32)/(bin exposure))
     lc.cts     : Counts in a bin
     lc.bin_lo  : Lower time bounds of binned lightcurve
     lc.bin_hi  : Upper time bounds of binned lightcurve
     lc.expos   : Exposure of a time bin (i.e., intersection with
                  good time intervals)
`);  %}}}    
   return;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define sitar_bin_events( )
%!%+
%\function{sitar_bin_events}
%\synopsis{Bin an event-based lightcurve.}
%\usage{lc = sitar_bin_events(t,dt,gti_lo,gti_hi);}
%\qualifiers{
%\qualifier{tstart}{[Beginning of first output time bin. Default: min(t).]}
%\qualifier{tstop}{[End of last output time bin. Default: max(t).]}
%\qualifier{obin}{[For a tstop, last bin width might be < dt.  Include this overflow bin if width is > obin*dt. Default: obin=0.1.]}
%\qualifier{user_bin.lo}{[Lower time bounds for user defined grid.]}
%\qualifier{user_bin.hi}{[Upper time bounds for a user defined grid. Overrides dt, tstart, and tstop inputs.]}
%\qualifier{minexp}{[Do not include bins with exposure < minexp. Default: 0.]}
%\qualifier{delgap}{[If set, delete 0 exposure bins from the lightcurve, otherwise set their rate & error to 0.]}
%}
%\description
%   Inputs:
%     t          : Discrete times at which rates are measured
%     dt         : Width of evenly spaced time bins in rebinned lightcurve
%                  (Ignored if a user_bin is input.)
%
%   Outputs:
%     lc.rate    : Mean rate of resulting binning
%     lc.err     : Error on bin rate (minimum = -log(0.32)/(bin exposure))
%     lc.cts     : Counts in a bin
%     lc.bin_lo  : Lower time bounds of binned lightcurve
%     lc.bin_hi  : Upper time bounds of binned lightcurve
%     lc.expos   : Exposure of a time bin (i.e., intersection with
%                  good time intervals)
%\seealso{sitar_rebin_rate}
%!%-
%%%%%%%%%%%%%%%%%%%%%%%%
{
   % Bin a rate lightcurve to a new scale

   variable t, dt, gti_lo, gti_hi; 
   variable custf=0, istck;
   variable lc = struct{ rate, err, cts, bin_lo, bin_hi, expos };

   variable tstart = qualifier("tstart",NULL);
   variable tstop  = qualifier("tstop",NULL);
   variable bintol = qualifier("obin",0.1);
   variable minexp = qualifier("minexp",0);
   variable cust   = qualifier("user_bin",NULL);
   variable gap    = qualifier_exists("delgap");

   switch (_NARGS)
   {
    case 4:
      ( t, dt, gti_lo, gti_hi ) = ( );
   }
   {
      reb_evt_use();
      _pop_n(_NARGS);
      return;
   }

   if ( cust != NULL )
   {
      if ( is_struct_type(cust) == 1 && struct_field_exists(cust,"lo")
                                     && struct_field_exists(cust,"hi") )
      { 
         custf = 1;
         dt = 0;
      }
      else 
      {
         reb_evt_use();
         () = fprintf(fp,  "%s\n", %{{{
`
  Custom user_bin is incorrect.  Input structure requires: 
     variable user_bin = struct{ lo, hi };
`); %}}}
         return;
      }
   }

   if ( dt == NULL ) dt = 0;

   variable mint = min(t);
   variable maxt = max(t);

   if ( tstart == NULL ) tstart = mint;
   if ( tstop == NULL )  tstop = maxt;

   if( maxt < tstart || mint > tstop)
   {
      reb_evt_use( );
      () = fprintf(fp, "\n Data outside of bounds of start & stop times:\n");
      () = fprintf(fp, "      [tstart,tstop] = [%e,%e]\n",tstart,tstop);
      () = fprintf(fp, "      [min_t, max_t] = [%e,%e]\n\n",mint,maxt);
      return;
   }

   % Remove tstart from all the times to begin with, add back in at the end
   t -= tstart;
   gti_lo -= tstart;
   gti_hi -= tstart;

   if( custf == 0 && dt > 0)
   {
      variable nfrac = ( tstop - tstart )/dt;
      variable nbins = int(nfrac);
      lc.bin_lo = [0:nbins-1]*dt;
      lc.bin_hi = [1:nbins]*dt;
      if( nfrac-nbins > bintol )
      {
         lc.bin_lo = [lc.bin_lo,lc.bin_hi[nbins-1]];
         lc.bin_hi = [lc.bin_hi,tstop-tstart]; 
      }
   }
   else if ( custf )
   {
      lc.bin_lo = @cust.lo-tstart;
      lc.bin_hi = @cust.hi-tstart;
   }
   else
   {
      reb_evt_use();
      () = fprintf(fp, "\n Time bin or custom bin array must be input.\n\n");
      return;
   }

   % Create the histogram of the times to creat counts per bin
   lc.cts = histogram(t, lc.bin_lo, lc.bin_hi);
   lc.rate = 0.*lc.cts;
   lc.err = @lc.rate;
   lc.expos = @lc.rate;

   % Figure out the bins within the Good Time Intervals

   variable i,j,ilo,ihi;
   _for i (0,length(gti_lo)-1,1)
   {
      ilo=where(lc.bin_hi>gti_lo[i]);
      if(length(ilo))
      {
         ihi=where(lc.bin_lo<gti_hi[i]);
         if(length(ihi))
         {
            ilo=min(ilo);
            ihi=max(ihi);

            _for j (ilo+1,ihi-1,1)
            {
               lc.expos[j] += lc.bin_hi[j]-lc.bin_lo[j];
            }
            if(ilo==ihi)
            {
               lc.expos[ilo] += min([gti_hi[i],lc.bin_hi[ihi]]) -
                                max([gti_lo[i],lc.bin_lo[ilo]]);
            }
            else
            {
               lc.expos[ilo] += lc.bin_hi[ilo]-max([gti_lo[i],lc.bin_lo[ilo]]);
               lc.expos[ihi] += min([gti_hi[i],lc.bin_hi[ihi]])-lc.bin_lo[ihi];
            }
	 }
      }
   }

   % Figure out which exposure bins aren't empty, and make rates & errors

   variable idx = where( lc.expos > minexp);
   if ( length(idx) )
   {
      lc.rate[idx] = lc.cts[idx]/lc.expos[idx];
      lc.err[idx] =  sqrt(abs(lc.cts[idx]+log(0.32)/2)-log(0.32)/2)/lc.expos[idx];
   }

   lc.bin_lo += tstart;
   lc.bin_hi +=tstart;

   if (gap)
   {
      lc.cts = lc.cts[idx];
      lc.rate = lc.rate[idx];
      lc.err = lc.err[idx];
      lc.expos = lc.expos[idx];
      lc.bin_lo = lc.bin_lo[idx];
      lc.bin_hi = lc.bin_hi[idx];
   }

   return lc;
}

#endif

%
% Code from src/sixte/sixte_create_img.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define sixte_create_img_xifu(ra,dec,fevt,fimg){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sixte_create_img_xifu}
%\synopsis{returns cmd to create an Image from an EventFile for the XIFU Baseline config }
%\usage{String cmd = sixte_create_img_xifu(double ra, double dec, string event_file, string image_file);}
%!%-
   return
     "imgev " +
     "EventList=$fevt "$ +
     "Image=$fimg "$ +
     "CoordinateSystem=0 " +
     "Projection=TAN " +
     "CUNIT1=deg CUNIT2=deg " +
     "NAXIS1=60 NAXIS2=64 " +
     sprintf("CRVAL1=%lf CRVAL2=%lf ", ra, dec) +
     "CDELT1=-0.001265 CDELT2=0.001265 " +
     "CRPIX1=32.5 CRPIX2=32.5 " +
     "clobber=yes ";
}


%%%%%%%%%%%%%%%%%%%%%%%
define sixte_create_img_wfi(ra,dec,fevt,fimg){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sixte_create_img_wfi}
%\synopsis{returns cmd to create an Image from an EventFile for the XIFU Baseline config }
%\usage{string cmd = sixte_create_img_wfi(double ra, double dec, string event_file, string image_file);}
%\qualifiers{
%\qualifier{mode}{["large"]: change mode of the WFI}
%}
%!%-

   if (qualifier("mode","large") != "large"){
      message("Only *large* mode implemented for the WFI yet");
      return;      
   }
   
   return
     "imgev " +
     "EventList=$fevt "$ +
     "Image=$fimg "$ +
     "CoordinateSystem=0 " +
     "Projection=TAN " +
     "CUNIT1=deg CUNIT2=deg " +
     "NAXIS1=512 NAXIS2=512 " +
     sprintf("CRVAL1=%lf CRVAL2=%lf ", ra, dec) +
     "CDELT1=-0.000621 CDELT2=0.000621 " +
     "CRPIX1=256.5 CRPIX2=256.5 " +
     "clobber=yes ";
}

%
% Code from src/sixte/wrapper/sixte_psfgen.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define sixte_psfgen(hew, npix, resolution_meter, focallength, filename){ %{{{
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sixte_psfgen(hew, npix, resolution_meter, focallength, filename)}
%\synopsis{simple wrapper to Sixte-Tool "psfgen"}
%\description
%   hew: HEW of PSF [arcsec]
%   npix: Number of pixels in X and Y direction
%   resolution_meter: width of one pixel in the focal plane [meter]
%   focallength: focal length of telescope [meter]
%   filename: name of output file
%   
%   Note: this functions requires Sixte to be installed!
%!%-


       variable command="psfgen "+
       			sprintf("Width=%d ", npix)+
			sprintf("PixelWidth=%e ", resolution_meter)+
			sprintf("FocalLength=%f ", focallength,)+
			sprintf("PSFFile=%s ", filename)+
			sprintf("HEW=%f ", hew)+
			"type=1 history=true";

   if (qualifier_exists("verbose"))  {
      message(command);
   }
	()=system(command);
}
%}}}

%
% Code from src/misc/sltable.sl
%
% This sets which type of table we're creating.
private variable sltableType = "tabular";
% end-of-line character
private define sltableEOL() {
  return "\n";
}
% these keep track of indentation
private variable sltableNTabs = 0;
private define sltableTabs() {
  variable i;
  variable t = "";
  if(sltableNTabs > 0) {
    _for i (0,sltableNTabs-1,1) {
      t += "  ";
    }
  }
  return t;
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableSetType
%
% SYNOPSIS
%   Defines output of subsequent function calls.
%
% USAGE
%   table_set_type(String_Type newTableType);
%
% DESCRIPTION
%   This function sets the internal "sltableType" variable, which is checked by
%   the other functions in this package to determine what type of LaTeX code
%   they output. E.g., sltableNoteMark("*") will output "\tablenotemark{*}"
%   command if sltableType == "deluxetable", but will output "$^{\text{*}}$" if
%   sltableType == "tabular".
%
% INPUTS
%   String_Type newTableType: type of tables to be generated by subsequent
%     commands. Must be either "deluxetable" or "tabular".
%
% OUTPUTS
%   This function produces no output.
%
% SEE ALSO
%   table_get_type
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableSetType() {
  variable newTableType;
  (newTableType) = ();
  if( _NARGS != 1 || 
      typeof(newTableType) != String_Type ||
      (sltableType != "deluxetable" and sltableType != "tabular")
    ) {
    usage("sltableSetType(String_Type sltableType)");
    throw InvalidParmError;
  }
  sltableType = newTableType;
}
%%%%%%%%%%%%%%%%%%%%%%%%
% sltableGetType
%
% SYNOPSIS
%   Returns the current table type
%
% USAGE
%   String_Type curTableType = table_get_type()
%
% DESCRIPTION
%   Returns the current value of the internal "sltableType" variable.
%
% INPUTS
%   No inputs
%
% OUTPUTS
%   No outputs
%
% SEE ALSO
%   table_set_type
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableGetType() {
  return sltableType;
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableNoteMark
%
% SYNOPSIS
%   Produces a table note mark
%
% USAGE
%   String_Type markText = sltableNoteMark(String_Type mark)
%
% DESCRIPTION
%   Produces table note mark (e.g., a superscripted character or symbol) code
%   depending on the value of sltableType. If sltableType == "deluxetable", uses
%   deluxetable's \tablenotemark command. If sltableType == "tabular", uses math
%   mode and \text to produce a superscripted mark.
%
% INPUTS
%   String_Type mark: mark to be used
%
% OUTPUTS
%   String_Type markText: formatted table note mark
%
% SEE ALSO
%   sltableNoteText
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableNoteMark(mark) {
  if(sltableType == "deluxetable") 
    return sprintf("\\tablenotemark{%s}",mark);
  else
    return sprintf("$^{\\text{%s}}$",mark);
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableNoteText
%
% SYNOPSIS
%   Produces a table footnote
%
% USAGE
%   String_Type footnoteText = sltableNoteText(String_Type mark, String_Type
%   text)
%
% DESCRIPTION
%   Produces code for a table footnote depending on the value of sltableType. If
%   sltableType == "deluxetable", uses deluxetable's \tablenotetext command. If
%   sltableType == "tabular", uses \flushleft and \footnotesize to produce a
%   footnote.
%
% INPUTS
%   String_Type mark: mark to be used
%   String_Type text: footnote text
%
% OUTPUTS
%   String_Type foonoteText: formatted table note text
%
% SEE ALSO
%   sltableNoteMark
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableNoteText(mark,text) {
  if(sltableType == "deluxetable") 
    return sprintf("\\tablenotetext{%s}{%s}",mark,text);
  else
    return sprintf("\\flushleft\\footnotesize{$^{\\text{%s}}$%s}",mark,text);
}

% Table header things - captions, column headers
%%%%%%%%%%%%%%%%%%%%%%%%
% sltableCaption
%
% SYNOPSIS
%   Creates an appropriate table caption statement with label
%
% USAGE
%   String_Type caption = table_caption(String_Type text, String_Type label);
%
% DESCRIPTION
%   Creates a table caption of the form \caption{<text> \label{<label>}} (for
%   tabular) or \tablecaption{<text> \label{<label>}} (for deluxetable).
%
% INPUTS
%   String_Type text: The table's caption.
%   String_Type label: The LaTeX label for the table
%
% OUTPUTS
%   String_Type caption: The formatted LaTeX table caption
%
% SEE ALSO
%
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableCaption(text,label) {
  variable continueFlag = qualifier("continue",0);
  if(sltableType == "deluxetable")
    return sprintf("\\tablecaption{%s\\label{%s}}",text,label);
  else
    if(continueFlag) {
      return sprintf("\\contcaption{%s\\label{%s}}",text,label+"_continued");
    } else {
      return sprintf("\\caption{%s\\label{%s}}",text,label);
    }
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableColHead
%
% SYNOPSIS
%   Create a formatted table column header.
%
% USAGE
%   String_Type colhead = sltableColHead(String_Type text)
%
% DESCRIPTION
%   Create a properly-formatted column header containing the supplied text for
%   the currently-selected table type. Note that if sltableType == "tabular",
%   this just returns the supplied string, as tabular does not treat the column
%   headers any differently from the other parts of the table body. For
%   deluxetables, it uses the \colhead command.
%
% INPUTS
%   String_Type text: Text for the column header.
%
% OUTPUTS
%   String_Type colhead: The formatted column header.
%
% SEE ALSO
%   sltableColHeads
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableColHead(text) {
  if(sltableType == "deluxetable")
    return sprintf("\\colhead{%s}",text);
  else
    return sprintf("%s",text);
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableColHeads
%
% SYNOPSIS
%   Create a formatted table header statement
%
% USAGE
%   String_Type sltableColHeads = sltableColHeads(String_Type colHeads)
%
% DESCRIPTION
%   Given the full text of the column headers (as a single string!), returns
%   the table header. Note that tabular doesn't treat the column headers and
%   table header any differently than any other table lines, so if sltableType ==
%   "tabular", this function just returns the supplied string. For
%   deluxetables, it uses the \tablehead command.
%
% INPUTS
%   String_Type colHeads: a single string (not an array!) containing all the
%   column headers separated by &'s.
%
% OUTPUTS
%   String_Type sltableColHeads: Formatted table header line
%
% SEE ALSO
%   sltableColHead
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableColHeads(text) {
  if(sltableType == "deluxetable")
    return sprintf("\\tablehead{%s}",text);
  else
    return sprintf("%s",text);
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableColJust
%
% SYNOPSIS
%   Build column-justification string for n columns
%
% USAGE
%   String_Type justString = table_column_just(Integer_Type n)
%
% DESCRIPTION
%   Build a column-justification string with the first column left-justified
%   and the remaining columns right-justified
%
% INPUTS
%   Integer_Type n: number of columns
%
% OUTPUTS
%   String_Type justString: column justification string
%
% SEE ALSO
%
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableColJust(n) {
  variable i;
  variable justString = "";
  if(n < 1) {
    message("Error: trying to make a table with no columns.");
    throw UsageError;
  }
  _for i (0,n-1,1) {
    if(i == 0) {
      justString += "l";
    } else {
      justString += "r";
    }
  }
  return justString;
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableRoundToDigit
%
% SYNOPSIS
%   Round a value to a number of digits.
%
% USAGE
%   String_Type val = sltableRoundToDigit(Double_Type value, Integer_Type
%   numDigits);
%
% DESCRIPTION
%   Produces a LaTeX string for "value" rounded to "numDigits" decimal digits.
%
% INPUTS
%   Double_Type value: the value to be rounded.
%   Integer_Type numDigits: the desired number of decimal places.
%
% OUTPUTS
%   String_Type val: a string containing the rounded value.
%
% SEE ALSO
%   sltableRound
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableRoundToDigit(val,sf) {
  variable totDig,decDig;
  variable fmt;
  variable neg = (val < 0 ? "-" : "");
  variable zeroString = "";
  % Only do the following if val is non-zero
  if(val == 0) { % If val is zero, simpler to format
    fmt = sprintf("%%%i.%if",sf+1,sf);
    return(sprintf(fmt,val));
  } else {
    val = abs(val);
    variable dig = int(floor(log10(val))); % power of 10 of leading digit
    % if number is less than 1: leading zero, but no other integer digits
    if(dig < 0) { 
      dig *= -1;
      dig += sf-1;
      fmt = "%.${dig}f"$;
    } else if (dig < sf-1) { % leading digit is at least 10^0 but less than 10^sf
      totDig = sf + 1;
      decDig = sf - dig - 1;
      fmt = "%${totDig}.${decDig}f"$;
    } else { % Bigger than 10^sf - format as integer now
      % need to divide by power of ten and then add zeros to end of string
      variable pow10 = dig - sf + 1;
      val *= 10^-pow10;
      val = nint(val);
      fmt = "%${sf}i"$;
      zeroString = (pow10 ? sprintf("%0${pow10}i"$,0) : "");
    }
    return(neg + sprintf(fmt,val) + zeroString);
  }
}


%%%%%%%%%%%%%%%%%%%%%%%%
% sltableRound
%
% SYNOPSIS
%   Round a value to its uncertainty or to a number of digits and return a
%   LaTeX string.
%
% USAGE
%   String_Type val = sltableRound(value[,minimum,maximum]; qualifiers)
%
% DESCRIPTION
%   Produces a LaTeX string for a value, possibly with errors. Uses
%   TeX_value_pm_error() for the "standard" case (a value with an asymmetric
%   confidence interval) and sltableRoundToDigit() otherwise. If the value has no
%   errors, it is printed to three significant figures. If it is frozen, it is
%   surrounded by parentheses (or whatever you provide as the "frozendelim"
%   qualifier) and printed to three sig figs.
%   Upper/lower limits are determined automatically by a parameter's min/max
%   values crossing zero and the "limit" field of the provided struct being set
%   to 1. If "limit" is zero, the upper and lower limits having different sign
%   is not interpreted as having any significance (so for something like a
%   photon index, which can be zero without meaning anything special, you'd
%   want to set limit = 0).
%
% INPUTS
%
% OUTPUTS
%
% SEE ALSO
%   sltableRoundToDigit
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableRound() {
  variable value, mini, maxi, limit = 0, nodata = 0, frz = 0;
  if(_NARGS == 1) {
    (value) = ();
  } else if(_NARGS == 3) {
    (value,mini,maxi) = ();
  } else {
    usage("String_Type val = sltableRound(value[,minimum, maximum]; qualifiers)");
  }

  variable frozenDelim = qualifier("frozendelim",["(",")"]);
  variable nodataSymb = qualifier("nodata",sltableType=="deluxetable"?"\\nodata":"\\ldots");

  if(typeof(value) == Struct_Type && typeof(value.value) == String_Type) {
    nodata = struct_field_exists(value,"nodata") ? value.nodata : 0;
    value = value.value;
  } else if(typeof(value) == Struct_Type && typeof(value.value) != String_Type) {
    mini = struct_field_exists(value,"min") ? value.min : value.value;
    maxi = struct_field_exists(value,"max") ? value.max : value.value;
    frz = struct_field_exists(value,"freeze") ? value.freeze : 0;
    limit = struct_field_exists(value,"limit") ? value.limit : 1;
    nodata = struct_field_exists(value,"nodata") ? value.nodata : 0;
    value = value.value;
  } else ifnot(__is_initialized(&mini)) {
    frz = 0;
    limit = 0;
    nodata = 0;
  } else {
    mini = value;
    maxi = value;
    frz = 0;
    limit = 0;
    nodata = 0;
  }

  if(nodata) { % first, if "nodata" was set, output the nodataSymb
    return nodataSymb;
  } else if (typeof(value) == String_Type) { % output strings as-is
      return value;
  } else if (mini == maxi and frz == 0) {  % parameter has no errors
    return sprintf("$%s$",sltableRoundToDigit(value,3));
  } else if(frz) {            % parameter is frozen - output (value)
    return sprintf("$%s%s%s$",frozenDelim[0],sltableRoundToDigit(value,3),frozenDelim[1]);
  } else if (limit and maxi > 0 and mini <= 0) {    % upper limit
    return sprintf("$< %s$",sltableRoundToDigit(maxi,3));
  } else if (limit and mini < 0 and maxi >= 0) {    % lower limit
    return sprintf("$> %s$",sltableRoundToDigit(mini,3));
  } else {
    return TeX_value_pm_error(value,mini,maxi;sci=3,noparens);
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%
define sltableBuildBody()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sltableBuildBody}
%\synopsis{Build a 2D array of table cells}
%\usage{String_Type table[] = sltableBuildBody(Array_Type par1[, Array_Type par2,...])}
%\qualifiers{
%\qualifier{horiz}{Integer_Type. If zero, each argument defines one column of
%           the table. If nonzero, each argument defines one row. Default 0.}
%\qualifier{frozendelim}{String_Type[2]. Defines strings to use before and
%           after a frozen parameter value. Default ["(",")"].}
%\qualifier{nodata}{String_Type. Defines string to use when no data is
%           present in a cell. Default is "\\\\nodata" for deluxetables and
%           "\\\\ldots" for tabular tables.}
%\qualifier{rownames}{String_Type[] or List_Type. Names for rows of table.
%         If given as a List_Type of String_Type arrays, each element of the
%         list will be included in its own column (this is useful for, e.g.,
%         putting the parameter name and the units in separate columns).}
%}
%
%\description
%   Builds the "body" of a table (in the form of a 2D array of strings,
%   containing the fields of the table).
%
%   Arguments should be either 1D arrays or structs, the same as one would give
%   to sltable(). An array of strings will be used as-is. Arrays of numbers
%   will be rounded to 3 significant figures.  Structs allow the handling of
%   error bars and offer considerably more flexibility, as defined below:
%
%   Struct arguments should, at minimum, have a "value" field, which must be
%   an array containing the value of the parameter for each row (or column, if
%   the table is horizontally-aligned). Further fields can be used to modify
%   the output. All should be arrays of the same length as the value field,
%   and they have the following names and uses:
%
%   min: minimum value of the parameter.
%   max: maximum value of the parameter. Min & max determine rounding & errors.
%   freeze: if 1, parameter is frozen, if 0, parameter is thawed.
%   limit: if 0, parameter can be zero without being an upper/lower limit.
%
%   Note that most of these are the same as those in the struct returned by
%   get_par_info(). The exception is the "limit" field, which determines how a
%   value of 0 for the parameter is handled. If "limit" is 1, then a confidence
%   interval containing zero is interpreted as an upper or lower limit. If
%   "limit" is 0, the value and error bars are typeset as a standard confidence
%   interval. E.g., one would set limit to 1 for a power-law normalization and
%   to 0 for a power-law index (the system is not flexible enough at the moment
%   to handle pathological cases like upper/lower limits on logarithmic
%   parameters like photon indices).
%
%   The number of "rows" in the table is determined by the length of the first
%   array or the first struct's value field. All later arrays and struct fields
%   should have the same number of elements.
%
%\seealso{sltable}
%!%-
{
  if(_NARGS < 1) {
    usage("String_Type table[] = sltableBuildBody(Struct_Type par1[, Struct_Type par2,...])");
    throw StackUnderflowError;
  }
  % pop all arguments into parList
  variable parList = __pop_list(_NARGS);
  variable horiz = qualifier("horiz",0);
  variable frozenDelim = qualifier("frozendelim",["(",")"]);
  variable nodataSymb = qualifier("nodata",sltableType=="deluxetable"?"\\nodata":"\\ldots");
  variable rowNames = qualifier("rownames",String_Type[0]);
  if(length(rowNames) && typeof(rowNames) != List_Type) rowNames = {rowNames};
  variable i,j;
  variable value;
  variable numRows, numCols;
  variable theTable;

  if(typeof(parList[0]) != Struct_Type) {
    numRows = length(parList[0]);
  } else {
    numRows = length(parList[0].value);
  }
  numCols = length(parList);
  if(horiz) {
    numRows += length(rowNames);
  } else {
    numCols += length(rowNames);
  }

  theTable = String_Type[numRows,numCols];

  _for i (0,length(rowNames)-1,1) {
    _for j (0,length(rowNames[i])-1,1) {
      if(horiz) {
        theTable[i,j] = rowNames[i][j];
      } else {
        theTable[j,i] = rowNames[i][j];
      }
    }
  }

  _for i (0,length(parList)-1,1) {
    variable iNdx,jNdx,isRowName;
    _for j (0,numRows-1,1) {
      if(horiz) {
        if(j < length(rowNames)) {
          isRowName = 1;
        } else {
          isRowName = 0;
          iNdx = i;
          jNdx = j - length(rowNames);
        }
      } else {
        iNdx = i + length(rowNames);
        jNdx = j;
        isRowName = 0;
      }
      if(not(isRowName)) {
        if(typeof(parList[i]) == Struct_Type) {
          value = struct_filter(parList[i],jNdx;copy);
        } else {
          value = parList[i][jNdx];
        }
        theTable[j,iNdx] = sltableRound(value;frozendelim=frozenDelim,nodata=nodataSymb);
      }
    }
  }
  return theTable;
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableBuildColHead
%
% SYNOPSIS
%   Build the column headers to a LaTeX table.
%
% USAGE
%   String_Type header = sltableBuildColHead();
%
% DESCRIPTION
%   Returns a string containing a LaTeX table column header. Use sltableSetType
%   to change whether this creates a deluxetable column header (using \tabhead
%   and \colhead) or a standard tabular table.
%
% INPUTS
%
% OUTPUTS
%   String_Type header: the LaTeX column headers.
%
% SEE ALSO
%   sltableSetType, sltableBuildBody, sltableBuildFooter
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableBuildColHead() {
  variable colNames = qualifier("colnames",String_Type[0]);
  variable numCols = qualifier("numCols",0);
  if(length(colNames) && typeof(colNames) != List_Type) colNames = {colNames};
  variable i,j;
  variable line = "";

  _for i (0,length(colNames)-1,1) {
    _for j (0,numCols-1,1) {
      if(j < numCols - length(colNames[i])) {
        line += "& ";
      } else {
        line += sltableColHead(colNames[i][j - (numCols - length(colNames[i]))]);
        if(j - (numCols - length(colNames[i])) < length(colNames[i])-1) 
          line += "&";
      }
    }
    if(i < length(colNames)-1 or sltableType == "tabular") 
      line += ` \\ `;
  }
  line = sltableColHeads(line);
  return line;
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableBuildHeader
%
% SYNOPSIS
%   Build the header to a LaTeX table.
%
% USAGE
%   String_Type header = sltableBuildHeader();
%
% DESCRIPTION
%   Returns a string containing a LaTeX table header. Use sltableSetType to
%   change whether this creates a deluxetable or a standard tabular table.
%
% INPUTS
%
% OUTPUTS
%   String_Type header: the LaTeX table header.
%
% SEE ALSO
%   sltableSetType, sltableBuildBody, sltableBuildFooter
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableBuildHeader() {
  variable header = "";
  variable longtable = qualifier("longtable",0);
  variable landscape = qualifier("landscape",0);
  variable numRows = qualifier("numRows");
  variable numCols = qualifier("numCols");
  variable tabletypesize = qualifier("tabletypesize","footnotesize");
  variable caption = qualifier("caption","");
  variable label = qualifier("label","placeholder");
  variable star = qualifier("star",0);
  variable horiz = qualifier("horiz",0);
  variable quals = __qualifiers;
  if(star) {
    star = "*";
  } else {
    star = "";
  }
  if(sltableType == "tabular") {
    if(landscape) {
      header += sltableTabs + `\begin{landscape}` + sltableEOL;
      sltableNTabs++;
    }
    header += sltableTabs + `\begin{table` + star + `}` + sltableEOL;
    sltableNTabs++;
  } else if(sltableType == "deluxetable") {
    if(longtable) header += sltableTabs + `\startlongtable` + sltableEOL;
    header += sltableTabs + `\begin{deluxetable` + star + `}{` + sltableColJust(numCols) + `}` + sltableEOL;
    sltableNTabs++;
    if(landscape) header += sltableTabs + `  \rotate` + sltableEOL;
    header += sltableTabs + `\tabletypesize{\` + tabletypesize + `}` + sltableEOL;
    header += sltableTabs + `\tablewidth{0pt}` + sltableEOL;
  }

  % Table caption + label
  header += sltableTabs + sltableCaption(caption,label) + sltableEOL;

  if(sltableType == "tabular") {
    header += sltableTabs + `\begin{tabular}{` + sltableColJust(numCols) + `}` + sltableEOL;
    header += sltableTabs + `\toprule` + sltableEOL;
  }

  % Add line of column heads
  header += sltableTabs + sltableBuildColHead(;;quals) + sltableEOL;

  % Startdata (or tabular equivalent)
  if(sltableType == "tabular") {
    header += sltableTabs + `\midrule` + sltableEOL;
  } else if(sltableType == "deluxetable") {
    header += sltableTabs + `\startdata` + sltableEOL;
  }

  sltableNTabs++;
  return header;
}

%%%%%%%%%%%%%%%%%%%%%%%%
% sltableBuildFooter
%
% SYNOPSIS
%   Build the footer to a LaTeX table.
%
% USAGE
%   String_Type footer = sltableBuildFooter();
%
% DESCRIPTION
%   Returns a string containing a LaTeX table footer. Use sltableSetType to
%   change whether this creates a deluxetable or a standard tabular table.
%
% INPUTS
%
% OUTPUTS
%   String_Type footer: the LaTeX table footer.
%
% SEE ALSO
%   sltableSetType, sltableBuildBody, sltableBuildHeader
%%%%%%%%%%%%%%%%%%%%%%%%
private define sltableBuildFooter() {
  variable noteText = qualifier("noteText",String_Type[0]);
  variable noteSym = qualifier("noteSym",String_Type[0]);
  variable longtable = qualifier("longtable",0);
  variable landscape = qualifier("landscape",0);
  variable star = qualifier("star",0);
  if(star) {
    star = "*";
  } else {
    star = "";
  }
  variable footer = "";
  if(sltableType == "tabular") {
    footer += sltableTabs + `\bottomrule` + sltableEOL;
    sltableNTabs--;
    footer += sltableTabs + `\end{tabular}` + sltableEOL;
  } else if(sltableType == "deluxetable") {
    sltableNTabs--;
    footer += sltableTabs + `\enddata` + sltableEOL;
  }
  variable i;
  _for i (0,length(noteText)-1,1)
    footer += sltableTabs + sltableNoteText(noteSym[i],noteText[i]) + sltableEOL;
  sltableNTabs--;
  if(sltableType == "tabular") {
    footer += sltableTabs + `\end{table` + star + `}` + sltableEOL;
  } else if (sltableType == "deluxetable") {
    footer += sltableTabs + `\end{deluxetable` + star + `}` + sltableEOL;
  }
  return footer;
}


%%%%%%%%%%%%%%%%%%%%%%%%
define sltable()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sltable}
%\synopsis{Generate a LaTeX table}
%\usage{String_Type table = sltable(par1[,par2,par3,...])}
%\description
%   sltable() is intended to streamline the production of tables of,
%   e.g., spectral parameters with error bars, although it should be flexible
%   enough to produce most simply-structured tables. There is a set of examples
%   on the Remeis wiki:
%   http://www.sternwarte.uni-erlangen.de/wiki/doku.php?id=isis:sltable
%
%   By default this builds tables using the "deluxetable" package; this can be
%   disabled to produce standard "tabular" tables by setting the "deluxe"
%   qualifier to zero. The deluxetable environment has been tested using the
%   latest aastex61.cls stylefile. Tabular tables will require the "booktabs"
%   package, as they provide the toprule, midrule, and
%   bottomrule LaXeX macros used here.
%
%   Each argument given to sltable() provides the values of one
%   column of the table (or one row, if the "horiz" qualifier is nonzero).
%   The length of each argument (or the length of the "value" field of a
%   Struct_Type argument) defines the number of rows (or columns, if horiz=1).
%   The arguments can be String_Type (in which case they are printed
%   literally), Double_Type (or some other numerical type), in which case they
%   are printed to three decimal places (TODO: figure out a better, more
%   flexible way of doing this), or Struct_Type, as defined below.
%
%   Struct_Type arguments should at least have a "value" field, which should be
%   an array of some type. If only the "value" field is present, the value
%   printed to the table follows the rules for String_Type and Double_Type
%   arguments above. Further fields which can be provided are:
%     min: the minimum values of this parameter
%     max: the maximum values of this parameter
%     freeze: if 1, the parameter is frozen, if 0, it is thawed
%     limit: if 1, this can be an upper/lower limit
%     nodata: if 1, there is no data here and "..." should be printed
%   These should all be arrays of the same length as the "value" field, and
%   allow for the printing of error bars and indicating upper/lower limits and
%   frozen parameters.
%
%   If the min and max fields are provided and the parameter is not frozen and
%   not an upper or lower limit, the confidence intervals are printed using the
%   TeX_value_pm_error() function. If it is an upper or lower limit, it is
%   rounded to three decimal places and indicated accordingly with "<" or ">"
%   symbols. If frozen, it is rounded to three decimal places and surrounded by
%   parentheses (these delimiters can be customized by changing the
%   "frozendelim" qualifier).
%
%   If the "nodata" field is nonzero, then a "no data" indicator will be
%   printed. By default, for a deluxetable, this is "\\\\nodata" and for a
%   tabular table it is "\\\\ldots". This is useful if you are putting multiple
%   models which do not share the same parameters into the same table.
%
%   This is a lot of words, so here is an
%   EXAMPLE
%   This will produce a simple 3x3 table with values with error bars, some
%   frozen parameters, some upper limits, and some cells in the table filled
%   with literal strings.
%   \code{
%   variable par1 = struct{value=[1,2,3],min=[0.9,1.9,0.0],max=[1.1,2.3,3.2],
%     freeze=[0,1,0],limit=[1,1,1]};
%   variable par2 = struct{value=[1,2,3],min=[0.9,1.9,2.9],max=[1.1,2.3,3.2],
%     freeze=[1,0,0],limit=[1,1,1]};
%   variable par3 = struct{value=["foo","bar","baz"]};
%   variable parnames = ["par1","par2","par3"];
%   variable parunits = ["unit1","unit2","unit3"];
%   variable obsNames = ["obs1","obs2","obs3"];
%   variable notes = ["note1","note2"];
%   variable noteSym = ["1","2"];
%   variable t = sltable(par1,par2,par3;
%     colnames={parnames,parunits},rownames=obsNames,
%     notes={notes,noteSym},label="tab:example",
%     caption="this is the table's caption");
%   }
%   The variable "t" now contains the full LaTeX table - drop it into a LaTeX
%   document and see what it ends up looking like. Note the use of the
%   "colnames" and "rownames" qualifiers to define the names and units for the
%   columns and rows, and how List_Type and Array_Type values are interpreted
%   differently there. Footnotes are also possible via the "notes" qualifier
%   (although currently we can't place the footnote markers into the table
%   automatically - you'll have to handle that yourself).
%
%   This script is under development. Contact Paul Hemphill (pbh@space.mit.edu)
%   if you find any bugs.
%
%\qualifiers{
%\qualifier{deluxe}{Integer_Type. If nonzero, we'll produce a deluxetable.
%        Otherwise, tabular. Tabular tables aren't as fancy and aren't
%        implemented as well at the moment. In theory we could add other table
%        types. Default 1.}
%\qualifier{horiz}{Integer_Type. If nonzero, arguments to sltable
%        indicate _rows_ of the table. If 0, they indicate _columns_.}
%\qualifier{rownames}{String_Type[] or List_Type[]. Names for the rows
%        of the table. Multiple columns for the row names (e.g., parameter names
%        and units in separate columns) can be handled by providing a List_Type
%        containing multiple String_Type arrays.}
%\qualifier{colnames}{String_Type[] or List_Type[]. Names for the columns of
%        the table (i.e., the column headers). Similar to rownames above,
%        multi-line headers can be produced by passing a List_Type to this
%        qualifier.} 
%\qualifier{caption}{String_Type. LaTeX caption for the table. Default is
%        blank. Don't include a label statement here - that's handled by the 
%        "label" qualifier below.}
%\qualifier{label}{String_Type. LaTeX label for the table. Default is
%        "placeholder," so make sure to set this.}
%\qualifier{nodata}{String_Type, set to what should be printed if there is no
%         data for a particular field (e.g., if you are putting multiple models
%         into the same table and they do not share the same parameters).}
%\qualifier{frozendelim}{String_Type[], two-element string array containing
%        opening and closing delimiters for a frozen value. By default this is
%        ["(",")"], i.e., the frozen value is surrounded by parentheses.}
%\qualifier{star}{Integer_Type. If nonzero, the table will be a table* or
%        deluxetable*. If 0, the table will be a table or deluxetable.}
%\qualifier{tabletypesize}{String_Type. Sets the font size of the table.
%        Should be a LaTeX font size without the backslash. Default is
%        "footnotesize".}
%\qualifier{longtable}{Integer_Type. If nonzero (and we're making a
%        deluxetable), this is a long table, so a startlongtable statement will
%        be included in the table header. I don't think tabular tables do
%        anything with this right now. Default 0.}
%\qualifier{landscape}{Integer_Type. If nonzero, this is a
%        landscape-orientation table. Includes a \\\\rotate command if
%        deluxetable, or wraps everything in a begin/end landscape block if
%        tabular. Default 0.}
%\qualifier{notes}{List_Type[2]. Element 0 should be a String_Type[] array of
%        the text of any footnotes; element 1 should be a String_Type[] array of
%        the symbols associated with those notes.}
%}
%
%\seealso{TeX_value_pm_error}
%!%-
{
  sltableNTabs = 0; % reset tabs to zero when sltable is called
  variable theHeader,theTable,theFooter;
  variable args = __pop_list(_NARGS);
  variable quals = __qualifiers;
  variable horiz = qualifier("horiz",0);
  if(_isnull(quals)) quals = empty_struct;

  if(qualifier("deluxe",1)) {
    sltableSetType("deluxetable");
  } else {
    sltableSetType("tabular");
  }

  theTable = sltableBuildBody(__push_list(args);;quals);
  if(horiz) theTable = transpose(theTable);

  variable numRows = length(theTable[*,0]),
           numCols = length(theTable[0,*]);
  quals = create_struct_field(quals,"numRows",numRows);
  quals = create_struct_field(quals,"numCols",numCols);

  theHeader = sltableBuildHeader(;;quals);
  variable i,j;
  variable line = theHeader;
  _for i (0,numRows-1,1) {
    _for j (0,numCols-1,1) {
      if(j == 0)
        line += sltableTabs;
      line += theTable[i,j];
      if(j < numCols-1) {
        line += " & ";
      } else {
        if(i < numRows-1 or sltableType == "tabular" or sltableType == "deluxetable")
          line += ` \\ `;
        line += sltableEOL;
      }
    }
  }
  if(qualifier_exists("notes")) {
    variable noteText,noteSym,notes;
    notes = qualifier("notes");
    noteText = notes[0];
    noteSym = notes[1];
    quals = create_struct_field(quals,"noteText",noteText);
    quals = create_struct_field(quals,"noteSym",noteSym);
  }
  theFooter = sltableBuildFooter(;;quals);
  line += theFooter;
  return line;
}

%
% Code from src/misc/socket_send_object.sl
%
% -*- mode: slang; mode: fold -*-

% LAYOUT
%
% socket_get_object                  socket_send_object
% -> receive object                   -> send object
%   no  qualifier: via _sso_client     no  qualifier: via _sso_server
%  `re' qualifier: via _sso_server    `re' qualifier: via _sso_client


%%% VARIABLES %{{{

% global variable for the object: we need this globally to
% use the object during the hook functions (e.g., _sso_send)
private variable _sso_obj = NULL;
% name of the receiver (_sso_server) or the sender (_sso_client)
% and the expected username on the host 
private variable _sso_host = "";
private variable _sso_user = NULL; % if NULL -> disable greet messages
%}}}

%%% HOOKS %{{{

% greet_hook of _sso_server:
% 1) send own username
% 2) check re-greet message with expected username
private define _server_greet() {
  variable s, c, greet;
  switch (_NARGS)
    { case 2: (s,c) = (); return getenv("USER"); } % case 1)
    { case 3: (s,c,greet) = (); % case 2)
      if (greet == _sso_user) { return 1; } % accepted
      s.shutdown(; trigger); % trigger shutdown after server's main loop
      return 0; % greet not accepted
    }
}

% greet_hook of _sso_client: check greet message against
% expected username and send own username
private define _client_greet(c, greet) {
  if (greet == _sso_user) {  return getenv("USER"); } % accepted
  return 0; % not accepted
}

% established_hook of _sso_server: send/get the object
% and shut down the server afterwards
private define _server_send(s, c) {
  c.config.chatty = 1; % enable progress bar
  ()=c.send(_sso_obj);
  ()=s.shutdown();
  % we do not need to set c.config.chatty back to 0 since this does not affect
  % the chattiness of the server itself (which is set in s.config.chatty (c->s))
}
private define _server_get(s, c) {
  c.config.chatty = 1;
  () = c.send("1"); % see footnote *1
  _sso_obj = c.receive();
  ()=s.shutdown();
}

% connect hook of _sso_client:  get/send the object
private define _client_get(c) {
  c.config.chatty = 1; % enable progress bar
  _sso_obj = c.receive();
  c.config.chatty = 0; % disable further client messages
}
private define _client_send(c) {
  c.config.chatty = 1;
  if (c.receive() == "1") { % see footnote *1
    ()=c.send(_sso_obj);
  } else { throw RunTimeError, "unexpected communication"; }
  c.config.chatty = 0;
}
%}}}

%%% MAIN ROUTINES %{{{

% establish a server and either send (default)
% or get the object (`re' qualifier)
define _sso_server() {
  ()=tcp_server(;; struct_combine(struct {
    maxclients = 1, % we do not expect more connections
    chatty = 0,
    greet_hook = &_server_greet,
    established_hook = qualifier_exists("re")
		? &_server_get  % get the object or
		: &_server_send % send the object by this hook
  }, _sso_user == NULL ? struct { nogreet } : NULL));
  % return the object
  if (qualifier_exists("re")) { return _sso_obj; }
}

% connect to the server and either get (default)
% or send the object (`re' qualifier)
define _sso_client() {
  variable client;
  % try to connect to the sender until a connection is established
  do {
    client = tcp_client(_sso_host;; struct_combine(struct {
      chatty = 0,
      greet_hook = &_client_greet,
      connect_hook = qualifier_exists("re")
		? &_client_send  % get the object or
		: &_client_get % send the object by this hook
    }, _sso_user == NULL ? struct { nogreet } : NULL));
    % client == 0 -> connection failed, e.g., server is not running
    if (typeof(client) == Integer_Type && client == 0) { sleep(1); }
  } while (typeof(client) != Struct_Type);
  % disconnect from the server
  ()=client.disconnect();
  % return the object
  ifnot (qualifier_exists("re")) { return _sso_obj; }
}

% split the given "username@host" in _sso_host into
% the username and host
define _sso_split_input() {
  variable n = is_substr(_sso_host, "@");
  if (n > 0) {
    _sso_user = substr(_sso_host, 1, n-1);
    _sso_host = substr(_sso_host, n+1, -1);
  }
}

%%%%%%%%%%%%%%%%%%%%
define socket_send_object()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{socket_send_object, socket_get_object}
%\synopsis{send or get an SLang object to or from another machine via a TCP socket}
%\usage{         socket_send_object("[username@]to_host", Any_Type object);
%    Any_Type socket_get_object("[username@]from_host");}
%\description
%    Established a connection to another machine in order to send or
%    receive an SLang object. Of course, this requires user input
%    from an active ISIS environment on the host. See the help of
%    `pack_obj' for the supported object formats.
%
%    socket_send_object:
%      A TCP server is started and as soon as a client connects the
%      object is sent. The server is shut down afterwards.
%    socket_get_object:
%      A TCP client tries to connect to the host in order to receive
%      the object. The connection is closed after the object has been
%      received.
%
%    For security reasons, it is recommended to provide the username
%    expected on the host. In this case the server and the client
%    exchange and check the corresponding username before the object
%    is transferred. If the username does not match the expected one
%    the connection is not trusted and closed automatically.
%
%    It might occur that machine A which is about to send an object
%    is behind a router/firewall or does not has a public internet IP
%    address, while the retrieving machine B does. Using the qualifier
%    `re' machine B starts a server for retrieving the object and
%    machine A a client connection for sending purposes (i.e., the
%    connection is established with server/client inverted with
%    respect to the default behavior).
%\example
%    % receive an object from user "falkner" on "indus"
%    obj = socket_get_object("falkner@indus");
%    
%    % send an array of doubles to "ara" without a greet message
%    socket_send_object("ara", Double_Type[10000]);
%
%    % send a structure `s' to an outside machine, while the own
%    % machine "laptop" is behind a router
%    socket_send_object("volans.sternwarte.uni-erlangen.de", s; re);
%    % the call on "volans" then would be
%    obj = socket_get_object("laptop"; re); % starts a server
%\seealso{tcp_server, tcp_client}
%!%-
{
  switch (_NARGS)
    { case 2: (_sso_host, _sso_obj) = (); }
    { help(_function_name); }

  _sso_split_input(); % split _sso_host into username and host
  % send object
  qualifier_exists("re") ? _sso_client(; re) : _sso_server();
}

define socket_get_object() {
  switch (_NARGS)
    { case 1: (_sso_host) = (); }
    { help(_function_name); }
  
  _sso_split_input(); % split _sso_host into username and host
  % get object
  return qualifier_exists("re") ? _sso_server(; re) : _sso_client();
}
%}}}


% *1) %{{{
%     In case the client sends the object, the connect_hook _may not_
%     send any message/object in the first place! That is the server
%     waits for the re-greet message of the client every 100 ms.
%     During this time the greet message of the server get accepted
%     by the client probably, which then would send the re-greet _and_
%     a message/object. The server would first receive the latter,
%     which of course is not the expected greet message and closes
%     the connection! Thus, whenever the client should send a message
%     first, it _has to_ wait for a "go"-message by the server!
%}}}


%
% Code from src/misc/atomic/solarAbund.sl
%
define solarAbund()
%!%+
%\function{solarAbund}
%\synopsis{returns the solar abundance of an element}
%\usage{Double_Type solarAbund(Integer_Type Z)}
%\description
%    \code{Z} is the element's proton number.
%\seealso{Grevesse et al., 1996: Standard Abundances}
%!%-
{
  variable Z;
  switch(_NARGS)
  { case 1: Z = (); }
  { help(_function_name()); return; }

  switch(Z)
  { case  1: return 10^(12.00-12); }
  { case  2: return 10^(10.99-12); }
  { case  6: return 10^( 8.55-12); }
  { case  7: return 10^( 7.97-12); }
  { case  8: return 10^( 8.87-12); }
  { case 10: return 10^( 8.08-12); }
  { case 12: return 10^( 7.58-12); }
  { case 13: return 10^( 6.47-12); }
  { case 14: return 10^( 7.55-12); }
  { case 16: return 10^( 7.33-12); }
  { case 18: return 10^( 6.52-12); }
  { case 20: return 10^( 6.36-12); }
  { case 26: return 10^( 7.50-12); }
  "Don't know, please tell me.";
}

%
% Code from src/slang/math/solveODEbyIntegrate.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define solveODEbyIntegrate()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{solveODEbyIntegrate}
%\synopsis{solves an ordinary differential equation on a grid by integration}
%\usage{x = solveODEbyIntegrate(&f, tgrid);}
%\qualifiers{
%    \qualifier{rdt}{integration step size relative to 'tgrid' binsize (default: 0.1)}
%    \qualifier{x0}{integration constant (default: 0)}
%    \qualifier{t0}{time reference for x0 (default: first 'tgrid' bin)}
%    \qualifier{method}{integratipn method (default: &integrateRK4)}
%}
%\description
%    The integral of the ordinary differential equation (ODE)\n
%      dx/dt = f(x(t), t)  with  x(t0) = x0\n
%    reads\n
%      x(t) = x0 + int_{t0}^{t} f(x(t'), t') dt'
%
%    \code{&f} is a reference to a function with two arguments:
%       \code{define f(x, t)}\n
%       \code{\{}\n
%       \code{  return ...;}\n
%       \code{\}}
%\seealso{integrateRK4}
%!%-
{
  variable f, tgrid;
  switch(_NARGS)
    { case 2: (f, tgrid) = (); }
    { help(_function_name()); return; }

  % step size and integration constant
  variable rdt = qualifier("rdt", .1);
  variable t0 = qualifier("t0", tgrid[0]);
  variable x0 = qualifier("x0", 0);
  variable method = qualifier("method", &integrateRK4);
  variable qual = qualifier("qualifiers", NULL);
    
  % index of reference time, t0, in time grid, tgrid
  if (tgrid[0] > t0 || tgrid[-1] < t0) {
    vmessage("error (%s): t0 (%.2f) not within tgrid (%.2f - %.2f)", _function_name, t0, tgrid[0], tgrid[-1]);
    return;
  }
  variable i0 = wherefirstmin(abs(tgrid - t0));

  % set-up solution: calculate starting point
  variable fx = Double_Type[length(tgrid)]; % will be the solution f(x)
  variable dt = abs(t0 - tgrid[i0]) * rdt;
  if (dt < 1e-3*rdt) {
    fx[i0] = x0;
  } else if (t0 < tgrid[i0]) {
    fx[i0] = (@method)(f, t0, tgrid[i0], dt, x0;; qual);
  } else {
    fx[i0] = (@method)(f, tgrid[i0], t0, -dt, x0;; qual);
  }

  % solve remaining bins
  variable i;
  % forward integration
  _for i (i0+1, length(tgrid)-1, 1) {
    dt = (tgrid[i] - tgrid[i-1]) * rdt;
    fx[i] = (@method)(f, tgrid[i-1], tgrid[i], dt, fx[i-1];; qual);
  }
  % backward integration
  _for i (i0-1, 0, -1) {
    dt = (tgrid[i+1] - tgrid[i]) * rdt;
    fx[i] = (@method)(f, tgrid[i+1], tgrid[i], -dt, fx[i+1];; qual);
  }

  return fx;
}

%
% Code from src/slang/math/solve_2d_system_of_equations.sl
%
define solve_2d_system_of_equations(ax1, ay1, c1, ax2, ay2, c2)
%!%+
%\function{solve_2d_system_of_equations}
%\usage{(x, y) = solve_2d_system_of_equations(ax1, ay1, c1, ax2, ay2, c2)}
%\description
%    \code{ax1 * x  +  ay1 * y  =  c1}\n
%    \code{ax2 * x  +  ay2 * y  =  c2}\n
%!%-
{
  variable D  = ax1 * ay2 - ax2 * ay1;
  variable Dx =  c1 * ay2 -  c2 * ay1;
  variable Dy = ax1 *  c2 - ax2 *  c1;
  if(D!=0)
    return Dx/D,  Dy/D;
  else
    if(Dx!=NULL)
      return NULL, NULL;  % 0 solutions
    else
      return NULL, NULL;  % infinitely many solutions
}

%
% Code from src/slang/astronomy/solve_Eddington_quartic_equation.sl
%
define solve_Eddington_quartic_equation() %{{{
%!%+
%\function{solve_Eddington_quartic_equation}
%\synopsis{Solve the Eddington quartic equation}
%\usage{Double_Type beta = solve_Eddington_quartic_equation(Double_Type M, mu)}
%\description
%    For a given mass 'M' (in solar masses) and mean molecular
%    weight 'mu', solve the Eddington quartic equation
%      1-beta = 0.003*M^2*mu^4*beta^4
%    numerically for 'beta' by using Newton's method to find
%    the root of a function.
%\notes
%    According to Equation (2-15) in the book "Principles of
%    Stellar Evolution and Nucleosynthesis" by D.D. Clayton,
%    the mean molecular weight can be approximated by
%      mu ~ 2/(1+3*X+0.5*Y)
%    where 'X' and 'Y' are the mass fractions of hydrogen
%    and helium.
%
%    The Eddington quartic equation needs to be solved to
%    compute the polytropic constant of the polytropic standard
%    model, see the function 'polytropic_standard_model'.
%\example
%    X = 0.73; Y = 0.26; mu = 2/(1+3*X+0.5*Y);
%    beta = solve_Eddington_quartic_equation(1, mu);
%\seealso{polytropic_standard_model}
%!%-
{
  if(_NARGS!=2)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': Wrong number of arguments.", _function_name());
  }
  variable M, mu;
  (M, mu) = ();
  variable beta_old = 0, beta = 1; % initial guess for beta
  variable M2mu4 = 0.003*M^2*mu^4;
  % Newton method for finding the root of the function f = 0.003*M^2*mu^4*beta^4 + \beta - 1
  while(abs(beta_old-beta)>1e-10)
  {
    beta_old = beta;
    beta = beta_old - (M2mu4*beta_old^4+beta_old-1.)/(4.*M2mu4*beta_old^3+1.);
  }
  return beta;
}%}}}

%
% Code from src/slang/astronomy/solve_Lane_Emden_equation.sl
%
private define Lane_Emden_eom(t,y,n) %{{{
{
  return [y[1], -(y[0])^n-2./t*y[1]]; % the Lane-Emden equation written as a system of first-order differential equations
}%}}}

define solve_Lane_Emden_equation() %{{{
%!%+
%\function{solve_Lane_Emden_equation}
%\synopsis{Solve the Lane-Emden equation numerically}
%\usage{Struct_Type s = solve_Lane_Emden_equation(Double_Type n)}
%\description
%    The Lane-Emden equation for index 0<'n'<5 is a second order
%    differential equation of the form
%
%         1   d (     dtheta)
%      ---- --- (xi^2 ------) = -theta^n
%      xi^2 dxi (        dxi)
%
%    with the boundary conditions
%
%                 dtheta
%      theta = 1, ------ = 0 at xi = 0 .
%                    dxi
%
%    The solution 'theta' satisfying these boundary conditions
%    is called Lane-Emden function of index 'n' and is returned
%    by this function together with its derivative dtheta.
%
%    The Lane-Emden functions are used to study the structures
%    of so-called polytropes, that is, gaseous spheres in hydro-
%    static equilibrium which obey a polytropic equation of state:
%      pressure = constant times density^((n+1)/n).
%    For details, see, e.g., Chapter 2.4 in the book "Principles
%    of Stellar Evolution and Nucleosynthesis" by D.D. Clayton.
%    The output structure contains also the polytropic constants
%      Rn = first zero of the Lane-Emden function of index n,
%
%                 dtheta
%      Mn = -xi^2 ------ evaluated at Rn.
%                    dxi
%
%            ( 3 dtheta)^(-1)
%      Dn = -(-- ------)      evaluated at Rn,
%            (xi    dxi)
%
%           Rn^((n-3)/n)(3Dn)^((3-n)/3n)
%      Bn = ---------------------------- .
%                (n+1)Mn^((n-1)/n)
%\notes
%    The numerical solution of the Lane-Emden equation is achieved
%    here by rewritting it as a system of first order differential
%    equations
%           d
%         --- [y_0,y_1] = [y_1, -(y_0)^n - 2/xi y_1 ]
%         dxi
%    which then allows an adaptive Runge-Kutta method of 4./5. order
%    to be applied.
%\qualifiers{
%\qualifier{tol}{[=1e-10] Absolute error control tolerance. Lower limit is 1e-15.}
%}
%\example
%    s = solve_Lane_Emden_equation(3);
%    print(s);
%\seealso{polytropic_standard_model}
%!%-
{
  if(_NARGS!=1)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable n = ();
  ifnot(0<n<5)
    throw UsageError, sprintf("Usage error in '%s': The polytropic index 'n' has to fulfill 0<n<5.", _function_name());
  variable tol = qualifier("tol", 1e-10);                                           % numerical tolerance
  variable dt = 1e-5;                                                               % initial stepsize
  variable t = 1e-5;                                                                % independent variable, initialized slightly off-center
  variable y = [1-1./6.*t^2+n/120.*t^4, -1./3.*t+n*4./120.*t^3];                    % initial condition, see Problem 2-56 in D.D. Claytons's book "Principles of Stellar Evolution and Nucleosynthesis"
  variable s = struct{ xi={0,t}, theta={1,y[0]}, dtheta={0,y[1]}, Rn, Mn, Dn, Bn }; % bookkeeping of results
  %
  variable method = qualifier("method", "RKCK");
  variable verbose = qualifier_exists("verbose");
  % ============================
  % initializing the ODE-solver:
  variable k1, k2, k3, k4, k5, k6, k7, dy_c, dy_e; % dy_c = dy continue, dy_e = dx error estimation
  variable coef = Double_Type[7,7]; % modified Butcher tableau for Runge-Kutta Cash-Karp method
  coef[0,0]= 1./5;  coef[0,1]= 1./5;
  coef[1,0]= 3./10; coef[1,1]= 3./40;       coef[1,2]= 9./40;
  coef[2,0]= 3./5;  coef[2,1]= 3./10;       coef[2,2]= -9./10;       coef[2,3]= 6./5;
  coef[3,0]= 1.;    coef[3,1]= -11./54;     coef[3,2]= 5./2;         coef[3,3]= -70./27;       coef[3,4]= 35./27;
  coef[4,0]= 7./8;  coef[4,1]= 1631./55296; coef[4,2]= 175./512;     coef[4,3]= 575./13824;    coef[4,4]= 44275./110592; coef[4,5]= 253./4096;
  coef[5,0]= 0;     coef[5,1]= 2825./27648; coef[5,2]= 18575./48384; coef[5,3]= 13525./55296;  coef[5,4]= 277./14336;    coef[5,5]= 1./4; % fourth order solution
  coef[6,0]= 0;     coef[6,1]= 37./378;     coef[6,2]= 250./621;     coef[6,3]= 125./594;      coef[6,4]= 0;             coef[6,5]= 512./1771; % fitfh order solution
  % ============================
  % =================
  % solving the ODEs:
  while(y[0]>tol)
  {
    k1 = dt * Lane_Emden_eom(t, y, n);
    k2 = dt * Lane_Emden_eom(t + coef[0,0]*dt, y + coef[0,1]*k1, n);
    k3 = dt * Lane_Emden_eom(t + coef[1,0]*dt, y + coef[1,1]*k1 + coef[1,2]*k2, n);
    k4 = dt * Lane_Emden_eom(t + coef[2,0]*dt, y + coef[2,1]*k1 + coef[2,2]*k2 + coef[2,3]*k3, n);
    k5 = dt * Lane_Emden_eom(t + coef[3,0]*dt, y + coef[3,1]*k1 + coef[3,2]*k2 + coef[3,3]*k3 + coef[3,4]*k4, n);
    k6 = dt * Lane_Emden_eom(t + coef[4,0]*dt, y + coef[4,1]*k1 + coef[4,2]*k2 + coef[4,3]*k3 + coef[4,4]*k4 + coef[4,5]*k5, n);
    dy_c = coef[5,1]*k1 + coef[5,2]*k3 + coef[5,3]*k4 + coef[5,4]*k5 + coef[5,5]*k6; % fourth order solution
    dy_e = coef[6,1]*k1 + coef[6,2]*k3 + coef[6,3]*k4 + coef[6,4]*k5 + coef[6,5]*k6; % fifth order solution
    variable scale = 0.9*(tol/max(abs(dy_c-dy_e)))^(0.2); % 0.9 is safety factor; derivation see Numerical Recipes, Section "Adaptive Stepsize Control for Runge-Kutta"
    % variable scale = ( 0.5*tol*dt/max(abs(dy_c-dy_e)) )^0.25;  % "derivation of this formula can be found in advanced books on numerical analysis", statement in Numerical Methods Using Matlab, 4th Edition, 2004
    if(scale>1 or dt<=1e-15*t) % scale>1: accept stepsize only if correction factor is larger than 1
    {                          % dt<=abs(1e-15*t) in order to avoid infinite loops due to Double_Type precision limit at dt=1e-16*t
      y += dy_c;
      t += dt;
      list_append(s.xi, t);
      list_append(s.theta, y[0]);
      list_append(s.dtheta, y[1]);
    }
    dt = _min(_max(dt*scale, t*1e-15), abs(y[0]/y[1])); % abs(1e-15*t) in order to avoid infinite loops due to Double_Type precision limit at dt=1e-16*t
  }                                                     % abs(y[0]/y[1]) is based on Newton's method to find the root of a function and is thus important when theta approaches zero
  s.xi = list_to_array(s.xi, Double_Type);
  s.theta = list_to_array(s.theta, Double_Type);
  s.dtheta = list_to_array(s.dtheta, Double_Type);
  s.Rn = s.xi[-1];
  s.Mn = -s.Rn^2*s.dtheta[-1];
  s.Dn = 1./(-3/s.Rn*s.dtheta[-1]);
  s.Bn = s.Rn^((n-3.)/n)*(3.*s.Dn)^((3.-n)/3./n)/(n+1.)/s.Mn^((n-1.)/n);
  return s;
}%}}}

%
% Code from src/slang/arrays/sort_arrays.sl
%
define sort_arrays()
%!%+
%\function{sort_arrays}
%\synopsis{}
%\usage{Integer_Type[] sort_arrays( Array_Type[] A1 [, A2 [, A3 ...]] );}
%\qualifiers{
%\qualifier{dir [=1]:}{1: Ascending sorting; -1: Descending sorting}
%\qualifier{method [='msort']:}{ Choose sorting method, 'msort' for merge-
%                       and 'qsort' for quick-sort.}
%}
%\description
%   Similar to 'array_sort' this funtion sorts arrays, but also sorts
%   multiple appearing array entries based on an addition given array
%   and so on. In other words this function allows to sort several arrays
%   (e.g., columns in a FITS-file) depending on the sorting of the previous
%   array.
%
%   This is especially useful if the arrays represent parameter combinations
%   and you want to perform a resorting. See example below.
%
%   NOTE THAT all arrays must have the same length!
%   
%\example
%   % Example 1:
%    a1 = [ 5,6,1,7,5,6,7 ];
%    a2 = [ 8,9,4,4,1,2,3 ];
%    i12= sort_arrays(a1,a2);
%    array_map( Void_Type, &vmessage, "%.2d %.2d", a1[i12], a2[i12] );
%    i21 = sort_arrays(a2,a1);
%    array_map( Void_Type, &vmessage, "%.2d %.2d", a1[i21], a2[i21] );
%
%   % Example 2: resorting parameter combinations:
%    par = struct{ a = [0:1:#3],
%                  b = ["hello","world"],
%                  c = [ 2001, 2012]
%                };
%    parcomb = get_par_combinations( par );
%    print(parcomb);
%
%    parcomb =  struct_array_2_struct_of_arrays(parcomb);
%    struct_filter( parcomb, sort_arrays(parcomb.b,parcomb.a,parcomb.c) );
%    print(struct_of_arrays_2_struct_array(parcomb));
%    
%\seealso{array_sort, get_par_combinations}
%!%-
{
  
  if( _NARGS == 0 ){
    help(_function_name);
    return NULL;
  }
  %% Get Arguments
  variable arg_list = __pop_list(_NARGS);

  %% Initialize sorting index array
  variable isort = array_sort( arg_list[0] ;; __qualifiers );
  variable len = length(isort);
  
  variable ii, istart, istopp;
  _for ii ( 0, _NARGS-2 ){ % loop over all given arrays
    istart = 0;
    while( istart < len ){
      % the ii-th array is sorted, [istart:istopp] gives
      % the global interval where the ii-th array has the same entry
      % not that it is crucial to use wherefirst here, which only works if the array is sorted!
      istopp = wherefirst( arg_list[ii][isort][[istart:]] != arg_list[ii][isort][istart] );
      if( _isnull(istopp) ){
	istopp = len-1;
      }else{
	istopp += (istart-1);
      }

      % now sort the ii+1th array in the [istart:istopp] range until the end of the array is hit
      isort[ [istart:istopp] ] = isort[ [istart:istopp] ][ array_sort( arg_list[ii+1][isort][[istart:istopp]] ;; __qualifiers ) ];
      istart = istopp+1;

      % when the ii+1th array is sorted go to the next array, i.e., the ii+1th array will be the ii-th array!
    }
  }

  return isort;      
}

%
% Code from src/slang/structs/tables/struct-arrays/sort_struct_arrays.sl
%
define sort_struct_arrays()
%!%+
%\function{sort_struct_arrays}
%\usage{Struct_Type sort_struct_arrays(Struct_Type s, String_Type fieldname);}
%!%-
{
  variable s, fieldname;
  switch(_NARGS)
  { case 2: (s, fieldname)= (); }
  { help(_function_name()); return; }

  return struct_filter(s, array_sort(get_struct_field(s, fieldname)); copy);

%  variable i = array_sort(get_struct_field(s, fieldname));
%  variable sorted_s = @s;
%  foreach fieldname (get_struct_field_names(s))
%  { set_struct_field(sorted_s, fieldname, get_struct_field(s, fieldname)[i]); }
%  return sorted_s;
}

%
% Code from src/slang/structs/sort_struct_fields.sl
%
define sort_struct_fields() {
%!%+
%\function{sort_struct_fields}
%\synopsis{sorts structure tags in a predefined order}
%\usage{ Struct_Type newstruc = empty_struct(Struct_Type s, String_Type taglist[]);}
%\description
% This function creates a new structure newstruc with the fiels given by the
% array taglist, and fills them with the values from structure s.
%
% The order in which the fields are added is given by taglist, fields not
% listed in taglist are ignored.
%
% The function is useful, e.g., when using fits_binary_table to create a FITS-file 
% from a structure. In this case the order in which the FITS columns are created
% corresponds to the order in which they are contained in the structure. One can then
% use this function to create a FITS-file with the desired order of columns.
%
%!%-
    variable s,fieldarr;
    (s,fieldarr)=();

    variable snew=@Struct_Type(fieldarr);
    variable i;
    _for i(0,length(fieldarr)-1,1) {
	set_struct_field(snew,fieldarr[i],get_struct_field(s,fieldarr[i]));
    }
    
    return snew;
}

%
% Code from src/data/timing/spearmanrho.sl
%
%%%%%%%%%%%%%%%%%%
define spearmanrho()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{spearmanrho}
%\synopsis{Timing Tools: Spearman's Rank Correlation Coefficient}
%\usage{(rho) = spearmanrho (a,b);}
%\description
% rho = 1 - frac {6 sum d_i^2} {n(n^2 -1)}
%
% d_i = a_i - b_i :  difference between the ranks of corresponding values
%               n :  number of values
%
% Alternative: The statistics module (\code{}) includes
% several functions for determining (rank) correlations coefficients,
% such as \code{spearman_r}, which provides the p-value (as well as the
% correlation coefficient \code{rho}).
%!%-
{
  variable a,b;
  switch(_NARGS)
  { case 2: (a,b) = (); }
  { help(_function_name());
    throw UsageError, sprintf("%s Usage Error: wrong number of arguments", _function_name() );
  }
  
   if (length(a) != length(b))
	 {
	   throw UsageError, sprintf("%s Usage Error: array lengths differ!", _function_name() );
	 }
   

   if(length(a)!=length(unique(a)) or length(b)!=length(unique(b)))
	 {
		%message ("Warning: arrays not unique!");
		variable rgx, rgy,dx,dy, sa,sb;
		sa = array_sort(a);
		sb = array_sort(b);
		rgx = a[sa];
		rgy = b[sb];
		variable tmp = Double_Type[length(rgx)];
		variable i,ndx,u = unique(rgx);
		_for i (0,length(u)-1,1)
		  {
			 ndx = where( rgx==rgx[u[i]]);
			 tmp[ndx] = mean(ndx+1);
		  }		
		dx = tmp[array_sort(sa)] - mean(tmp);
		u = unique(rgy);
		tmp[*] = 0.;
		_for i (0,length(u)-1,1)
		  {
			 ndx = where( rgy==rgy[u[i]]);
			 tmp[ndx] = mean(ndx+1);
		  }
		dy = tmp[array_sort(sb)] - mean(tmp);
		return sum(dx*dy)/sqrt(sum(sqr(dx))*sum(sqr(dy)));
	 }

   return 1-(6*sum(sqr(array_sort(array_sort(a))-array_sort(array_sort(b))))/(length(a)*(sqr(length(a))-1)));
}

%
% Code from src/slang/math/sphere_urand.sl
%

% only activate if module rand could be loaded
#ifeval __get_reference("rand_flat")!=NULL

%%%%%%%%%%%%%%%%%%%
define sphere_urand()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{sphere_urand}
%\synopsis{generate spherical coordinates uniformly distributed on a sphere}
%\usage{(Double_Type theta, phi) = sphere_urand();
%\altusage{(Double_Type theta[], phi[]) = sphere_urand(Integer_Type n);}
%}
%\description
%    The function generates \code{n} (default: \code{1}) pairs \code{(theta[i], phi[i])}
%    of spherical coordinates -- such that the points
%     \code{[ sin(theta)*cos(phi), sin(theta)*sin(phi), cos(theta) ]}
%    are uniformly distributed on the surface of the unit sphere.
%!%-
{
  variable n;
  switch(_NARGS)
  { case 0: n = 1; }
  { case 1: n = (); }
  { help(_function_name()); return; }

  variable theta = acos(rand_flat(-1, 1, n)),
           phi = rand_flat(0, 2*PI, n);

  if(n==1)
    return (theta[0], phi[0]);
  else
    return (theta, phi);
}

#endif

%
% Code from src/slang/math/spherical_volume.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define spherical_volume()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{spherical_volume}
%\synopsis{computes the volume of an object parameterized in spherical coordinates}
%\usage{Double_Type vol = spherical_volume(Ref_Type &r);}
%\qualifiers{
%\qualifier{Ntheta}{[=180]}
%\qualifier{Nphi}{[=180]}
%\qualifier{logfile}{[=NULL]: If logfile is a filehandle, the function evaluations are logged.}
%}
%\description
%    \code{r} has to be a real function with two arguments (theta, phi).\n
%    \code{vol = int_0^2pi dphi  int_0^pi dtheta sin(theta)  int_0^r(theta, phi) dr r^2}\n
%    \code{    = int_0^2pi dphi  int_0^pi dtheta sin(theta)  r(theta, phi)^3/3}\n
%    The numerical integration of phi and theta is performed on the following grid:\n
%    \code{theta = [0 : Pi : #Ntheta];   phi = [0 : 2*Pi : #Nphi];}
%!%-
{
  variable r;
  switch(_NARGS)
  { case 1: r = (); }
  { help(_function_name()); return; }

  variable Ntheta = qualifier("Ntheta", 180);
  variable Nphi = qualifier("Nphi", 180);
  variable logfile = qualifier("logfile", NULL);

  variable dtheta = PI/Ntheta;
  variable dphi = 2*PI/Nphi;
  variable itheta, iphi, theta, phi, rr;
  variable volume = 0;
  _for itheta (0, Ntheta-1, 1)
  { theta = itheta * dtheta;
    _for iphi (0, Nphi-1, 1)
    { phi = iphi * dphi;
      rr = @r(theta, phi);
      if(logfile==NULL)
        ()=fprintf(logfile, "%g %g %g\n", theta, phi, rr);
      volume += rr^3/3. * sin(theta)*dtheta*dphi;
    }
    if(logfile==NULL)
      ()=fprintf(logfile, "\n");
  }
  return volume;
}

%
% Code from src/data/timing/split_and_epfold_lc.sl
%

private define get_qualifier ( qualifier_structure, qualifier_name, default_value ) { %{{{
   try {
      return get_struct_field ( qualifier_structure, qualifier_name );
   } catch InvalidParmError: {
      return default_value;
   }
}
%}}}

%%%%%%%%%%%%%%%%%%%%%
define split_and_epfold_lc()%{{{
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{split_and_epfold_lc}
%\synopsis{Takes lightcurve and pulse period, splits if necessary and executes epfold for each segment.}
%\usage{Struct_Type split_and_epfold_lc(Struct_Type OR Array_Type lc, Double_Type p0), all times in the same units}
%\qualifiers{
%\qualifier{exact}{Forces execution of pfold with exact qualifier}
%\qualifier{not_exact}{Forces execution of pfold without exact qualifier}
%\qualifier{exact_threshold}{see description (default: 6.)}
%\qualifier{split_only}{Only the splitting is performed}
%\qualifier{nbins}{Sampling of pfold (default: min(32, p0 / dt), dt time resolution}
%\qualifier{nsrch}{Number of trial periods for epfold (default: 1e4)}
%\qualifier{grid_steps}{Period resolution for epfold (default: not set)}
%\qualifier{search_range}{Period interval considered left/right of p0 in terms of dp=p0^2/T (default: 2.)}
%\qualifier{dpmin}{For very long lightcurves, this can replace dp (default: p0e-3)}
%\qualifier{gap_scale}{the smaller, the higher is the sensitivity for gaps (default: .5)}
%\qualifier{fracexp}{Only timebins with fracexp >= this value will be considered (default: 1.)}
%\qualifier{chatty}{chattiness (default: 1)}
%}
%\description
%
% The input lightcurve(s) should be a structure or an array of
% structures of the form:
%     lc = struct { time, rate, fracexp, error }
% where each field is an array of doubles. 
% The output is a structure of the form:
%     out = struct { time, [ epfold ], lc, dp, exact }
% Each field is an array, where each element corresponds to one of the
% segments the split has produced. The individual fields are:
%     * time: average time of segment
%     * epfold: struct { p, stat, nbins, badp }
%       (not output if 'split_only' qualifier has been used)
%     * lc: of the same form as input structure lc
%     * dp: either p^2 / T or dpmin
%     * exact: 1 if exact qualifier was used for epfold, 0 otherwise
% Remarks regarding qualifiers:
%     * by default, exact is only chosen if there are less than exact_threshold
%       pulses in the timespan considered. However, as of Sep 16,
%       there was a bug in pfold, so that exact is ONLY used if exact_threshold
%       the 'exact_threshold' qualifier exists. When this bug is fixed, please
%       change the constant at the beginning of the script
%     * nsrch and grid_steps are mutually exclusive. If you set nsrch
%       to 0, the default epfold routine will be
%       used
%     * All qualifiers can also be passed using a structure named
%       'epfold_qualifiers'. If this qualifier is present, its
%       content will overwrite all other qualifiers given. If passed
%       using this structure, all qualifiers must be assigned values
%       (e.g. exact = 0, split_only = 1). Makes scripts easier to
%       read.
%
%\seealso{pfold, epfold, pulseperiod_epfold, pulseperiod_search, find_peak}
%!%-
{
   variable exact_bug = 1; % set this to 0 when the bug in pfold (; exact) has been fixed
                                                                                                                                                                                                  
   variable lc, p0;                              % definition of lightcurve array of structures and initial trial period
   switch (_NARGS)                                     % the input is checked
      { case 2: (lc, p0) = (); }
      { help(_function_name); return; }
                                                                                                                                                                                                  
   % sanity checks
   if (typeof(lc) != Array_Type) { lc = [lc]; }        % further checks of the input data
   variable iii;
   _for iii (0, length(lc) - 1, 1) {
      try {
         variable timex = lc[iii].time;
         variable ratex = lc[iii].rate;
         variable fracexpx = lc[iii].fracexp;
         variable errorx = lc[iii].error;
         variable sumx = timex + ratex + fracexpx + errorx;
      } catch AnyError: {
         vmessage("Error(%s): lc must have fields time, rate, error, fracexp. These have to be arrays of numbers of the same length", _function_name);
         return;
      }
   }
   if (_typeof(lc) != Struct_Type) { vmessage("error(%s): light curves have to be an array of structures!", _function_name); return; }
   if (typeof(p0) != Double_Type) { vmessage("error(%s): pulse period has to be a floating point number!", _function_name); return; }
   if (p0 <= 0) { vmessage("error(%s): pulse period has to be > 0", _function_name); return; }
                                                                                                                                                                                                  
   % pre-process input
   variable nlc = length(lc);                          % nlc is the number of structures given, each structure contains one lightcurve etc.
   lc = array_map(Struct_Type, &struct_combine, COPY(lc), struct { split = 0, splitnum = 0 }); % the fields split and splitnum are added to each of nlc lc structures
   % qualifiers %{{{
   variable epfold_qualifiers_default = empty_struct (); 
   variable aa;
   if (qualifier_exists("exact_threshold")) { aa=struct{exact_threshold=qualifier("exact_threshold")};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("fracexp")) { aa=struct{fracexp=qualifier("fracexp")};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("search_range")) { aa=struct{search_range=qualifier("search_range")};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("dpmin")) { aa=struct{dpmin=qualifier("dpmin")};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("gap_scale")) { aa=struct{gap_scale=qualifier("gap_scale")};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("chatty")) { aa=struct{chatty=qualifier("chatty")};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("grid_steps")) { aa=struct{grid_steps=qualifier("grid_steps")};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("nsrch")) { aa=struct{nsrch=qualifier("nsrch")};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("nbins")) { aa=struct{nbins=qualifier("nbins")};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("split_only")) { aa=struct{split_only=1};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("exact")) { aa=struct{exact=1};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }
   if (qualifier_exists("not_exact")) { aa=struct{not_exact=1};epfold_qualifiers_default=struct_combine(epfold_qualifiers_default,aa); }

   if (length(get_struct_field_names(epfold_qualifiers_default)) != 0 and qualifier_exists("epfold_qualifiers") ) {
      vmessage("Warning(%s): You have provided qualifiers both as a structure and as individual arguments, I will only consider the structure", _function_name);
   }

   variable epfold_qualifiers = qualifier ( "epfold_qualifiers", epfold_qualifiers_default );
   if (typeof(epfold_qualifiers) != Struct_Type) {
      vmessage("Error(%s): epfold_qualifiers must be a structure", _function_name);
      return;
   }
   variable exact_threshold = get_qualifier ( epfold_qualifiers, "exact_threshold", 6.);
   if (struct_field_exists(epfold_qualifiers, "exact_threshold")) { exact_bug = 0; }
   variable fracexp = get_qualifier ( epfold_qualifiers, "fracexp", 1. ); % fractional exposure, if less than the defined value the data bin will not be considered
   variable sdp = get_qualifier ( epfold_qualifiers, "search_range", 2. ); % scales the period search range
   variable dpmin = get_qualifier ( epfold_qualifiers, "dpmin", 0.001*p0 ); % the minimum expected width of the chi2 peak (for high T p^2/T can be misleading)
   variable gs = get_qualifier ( epfold_qualifiers, "gap_scale", 0.5 );  % measure for the threshold at which a gap is detected
   variable chatty = get_qualifier ( epfold_qualifiers, "chatty", 1 );           % determines the program's behaviour
   variable grid_steps = get_qualifier ( epfold_qualifiers, "grid_steps", -1. );
   variable nsrch = get_qualifier ( epfold_qualifiers, "nsrch", struct_field_exists(epfold_qualifiers,"grid_steps") ? -1. : 1e4 );
   variable nbins = get_qualifier ( epfold_qualifiers, "nbins", -1 );
   variable split_only = get_qualifier ( epfold_qualifiers, "split_only", 0);
   variable exact = get_qualifier ( epfold_qualifiers, "exact", 0);
   variable not_exact = get_qualifier ( epfold_qualifiers, "not_exact", 0);
   % Sanity checks %{{{
   ifnot (typeof(exact_threshold) == Double_Type or typeof(exact_threshold) == Integer_Type) {
      vmessage("Warning(%s): exact_threshold must be a number, setting it to default", _function_name);
      exact_threshold = 6.;
   }
   exact_threshold = double(exact_threshold);
   ifnot (typeof(fracexp) == Double_Type or typeof(fracexp) == Integer_Type) {
      vmessage("Warning(%s): fracexp must be a number, setting it to default", _function_name);
      fracexp = 1.;
   }
   fracexp = double(fracexp);
   if (fracexp<0 or fracexp>1) {
      vmessage("Warning(%s): fracexp must be [0,1], setting it to default", _function_name);
      fracexp = 1.;
   }
   ifnot (typeof(sdp) == Double_Type or typeof(sdp) == Integer_Type) {
      vmessage("Warning(%s): sdp must be a number, setting it to default", _function_name);
      sdp = 2.;
   }
   sdp = double(sdp);
   if (sdp < 2.) { vmessage("warning(%s):  search_range should usually be chosen >= 2", _function_name); }
   ifnot (typeof(dpmin) == Double_Type or typeof(sdp) == Integer_Type) {
      vmessage("Warning(%s): dpmin must be a number, setting it to default", _function_name);
      dpmin = 0.001*p0;
   }
   dpmin = double(dpmin);
   if (dpmin < 0) {
      vmessage("Warning(%s): dpmin has to be > 0, setting it to default",  _function_name);
      dpmin = 0.001*p0;
   }
   ifnot (typeof(chatty) == Integer_Type) {
      vmessage("Warning(%s): chatty has to be either 1 or 0, setting it to default", _function_name);
      chatty = 1;
   }
   ifnot (chatty == 1 or chatty == 0) {
      vmessage("Warning(%s): chatty has to be either 1 or 0, setting it to default", _function_name);
      chatty = 1;
   }
   ifnot (typeof(nsrch) == Integer_Type or typeof(nsrch) == Double_Type) {
      vmessage("Warning(%s): nsrch must be a number, setting it to default", _function_name);
      nsrch = 1e4;
   }
   nsrch = nint(nsrch);
   ifnot (typeof(grid_steps) == Integer_Type or typeof(grid_steps) == Double_Type) {
      vmessage("Warning(%s): grid_steps must be number, setting it to default", _function_name);
      grid_steps = 0.001*p0;
   }
   if (nsrch*grid_steps > 0) { vmessage("error(%s): grid_steps and nsrch qualifiers are mutually exclusive", _function_name ); return; }
   ifnot (typeof(nbins) == Integer_Type or typeof(nbins) == Double_Type) {
      vmessage("Warning(%s): grid_steps must be number, setting it to default", _function_name);
      nbins = -1;
   }
   nbins = nint(nbins);
   ifnot (typeof(split_only) == Integer_Type) {
      vmessage("Warning(%s): split_only can only be 1 or 0, setting it to default", _function_name);
      split_only = 0;
   }
   ifnot (split_only == 0 or split_only == 1) {
      vmessage("Warning(%s): split_only can only be 1 or 0, setting it to default", _function_name);
      split_only = 0;
   }
   ifnot (typeof(exact) == Integer_Type) {
      vmessage("Warning(%s): exact can only be 1 or 0, setting it to default", _function_name);
      exact = 0;
   }
   ifnot (exact == 0 or exact == 1) {
      vmessage("Warning(%s): exact can only be 1 or 0, setting it to default", _function_name);
      exact = 0;
   }
   ifnot (typeof(not_exact) == Integer_Type) {
      vmessage("Warning(%s): not_exact can only be 1 or 0, setting it to default", _function_name);
      not_exact = 0;
   }
   ifnot (not_exact == 0 or not_exact == 1) {
      vmessage("Warning(%s): not_exact can only be 1 or 0, setting it to default", _function_name);
      not_exact = 0;
   }
   ifnot (exact * not_exact == 0) { vmessage("Error(%s): exact and not exact are mutually exclusive", _function_name); return; }
   %}}}
   %}}}
                                                                                                                                                                                                  
   % light curve properties and eventually split them
   variable lcp = struct {
      time = Double_Type[nlc],                          % mean time - just the centre of the time grid
      dt = Double_Type[nlc],                            % time resolution - width of the time bins
      len = Double_Type[nlc],                           % length - total length of the time grid
      bins = Integer_Type[nlc],                         % number of time bins - integer number of time bins
      dp = Double_Type[nlc],                            % formal epfold period resolution - the theoretical p^2/T limit, can be replaced by dpmin
   };
   % loop over light curve(s) (if splitted below until all segments were looped)
   variable i = 0;                                     % loop variable, loops over the individual structures in array lc
   while (i < nlc)
   { %{{{                                  % for all structures
      % field checks
      if (any(array_map(Integer_Type, &struct_field_exists, lc[i], ["time", "rate", "error", "fracexp"]) == 0)) { vmessage("erros(%s): light curve does not have all required fields!"); return; }
      % filter on fractional exposure
      struct_filter(lc[i], where(lc[i].fracexp >= fracexp));  % only those parts of the ith structure are allowed where the fracexp criterion is met
      % determine properties
      lcp.dt[i] = median ( (lc[i].time - shift(lc[i].time,-1))[[0:length(lc[i].time)-1]] );                % the aforementioned variables are filled with values
      lcp.len[i] = lc[i].time[-1]-lc[i].time[0] + lcp.dt[i];
      lcp.bins[i] = length(lc[i].time);
      lcp.time[i] = lc[i].time[0] + .5*lcp.len[i];
      lcp.dp[i] = p0^2 / lcp.len[i];
      % check on gaps and split light curve
      variable splitted = 0;                                  % counter variable
      if (lcp.len[i] - lcp.dt[i]*lcp.bins[i] > lcp.dt[i]) {   % if the total time span is larger than the number of bins times time step , do ... (i.e. there exist gaps)
         variable gt = max ( [ lcp.dt[i], gs*(p0^2/lcp.dp[i] - p0) ] );               % gap threshold - equals 'gapscale' times ( total time span - trial period )
         %if (chatty) { vmessage("  light curve [%d] contains gaps (threshold %.0f s)", i, gt); }
         variable slc = split_lc_at_gaps(lc[i], gt);           % where the gaps in the time grid are larger than the threshold, the structure is split -->
                                                               % returns an array of structures called slc (each of the contained structures doesn't contain gaps anymore)
         % splitted
         if (length(slc) > 1) {                                % if the lightcurve has been split , do ...
            splitted = 1;                                       % counter variable is changed
            slc[0].splitnum = length(slc)-1;                    % the first structure is assigned splitnum = number of structures - 1
            if (chatty) { vmessage("(%s): light curve splitted into %d segments", _function_name, slc[0].splitnum+1); }
            variable j;
            _for j (1, length(slc)-1, 1) {                      % iteration over all structures except for the very first one
               slc[j].split = j;                                 % variable split is set to the index of the structure in slc
               slc[j].splitnum = slc[0].splitnum;                % all splitnum in the array are the same
            }
            % insert into main array
            lc = [                                             % at the position of the previously unsplit structure the split structures are inserted 
               i>0 ? lc[[:i-1]] : Struct_Type[0], % up to i-1    % if i==0, no data is written, otherwise the ith lc-structures until i-1 are written
               slc, % splitted ones                              % the splitted structures are written
               lc[[i+1:]] % from i+1                             % the following lc-structures are written
            ];
            % increase number of light curves and property structure
            struct_filter(lcp, [[0:nlc-1], ones(slc[0].splitnum)*0]); % the number of fields in lcp (the property structure) is correspondingly increased
            nlc += slc[0].splitnum;                                   % the number of structures in the array lc is correspondingly increased
         }
      }
      % go on
      if (splitted == 0) { i++; }                             % if no splitting was necessary, the next structure can be approached, otherwise the current one is processed
   }                                                         % again (check???)
   %}}}
                                                                                                                                                                                                  
   % prepare output structure
   variable out = struct { %{{{                                  % out contains:
      time = Double_Type[nlc], 
      epfold = Struct_Type[nlc],    % error of period, the folding output for each structure
      lc = lc,                                                 % the input lightcurves (splitted now!)
      dp = Double_Type[nlc],
      exact = Integer_Type[nlc], }; %}}}
   
   % epoch folding
   _for i (0, nlc-1, 1) { %{{{                                   % it is iterated over all structures (splitted now!)
      if (lcp.len[i] < 2. * p0) { continue; }
      if (p0-lcp.dp[i]*sdp < 0 || p0+lcp.dp[i]*sdp > lcp.len[i]) {
        vmessage("warning(%s): period search range too large, skipping light curve [%d]", _function_name, i);
        % sdp = min ( [ ( lcp.len[i] - p0 ) / lcp.dp[i], p0 / lcp.dp[i] ] );
        continue;
      }
      % epoch folding
      variable psr = _max(lcp.dp[i]*sdp, dpmin); % period search range - it's the maximum of the theoretical one (p^2/T*scaling) and the minimum allowed one
      
      variable pmin = _max(p0-0.4*p0, p0-psr, 0);                        % no negative periods allowed! -- added by Leander
      variable pmax = _min(p0+psr, p0+0.8*p0);
      variable epf;
                                                                                                                                                                                                   
      if ( split_only ) {
         if (i==0) { vmessage("(%s): Skipping epoch folding, only splitting of the lc has been performed", _function_name); }
         epf = empty_struct();
      } else { %{{{         
         nbins = (nbins < 0) ? (min ( [ 32, nint(p0 / lcp.dt[i]) ] )) : nbins; % nbins is set. The current p0/dt version is very small sometimes, but avoids problems with oversampling
         variable exct;
         if ( exact ) {
            exct = 1.;
         } else {
            if  ( exact ) { 
               exct = -1.;
            } else {
               if  ( ( ( max(lc[i].time) - min(lc[i].time) ) / p0  < exact_threshold ) and not exact_bug) { % normally, only the first criterion should be necessary. When 'exact' is working again remove the 2nd
                  exct = 1.;
               } else {
                  exct = -1.;
               }
            }
         }                                                                                                                                                                                                 
         if (exct == 1) {
            out.exact[i] = 1;
         } else {
            out.exact[i] = 0;
         }

         % For different configurations epfolding is carried out.
         if (nsrch > 0 and exct > 0.) { %{{{
            vmessage("(%s): epochfolding with exact, nsrch=%d, nbins=%d", _function_name, nsrch, nbins);
            epf = epfold(                                 % the epochfolding is calculated with the usual ISIS function
               lc[i].time, lc[i].rate, pmin, pmax;        % the required paramaters are given
               nsrch=nsrch, dt = lcp.dt[i]*lc[i].fracexp, nbins=nbins, chatty = -1, exact,
                        ); %}}}
         } else if (nsrch < 0 and exct > 0.) { %{{{ 
            vmessage("(%s): epochfolding with exact, grid_steps=%lf, nbins=%d", _function_name, grid_steps, nbins);
            epf = epfold(                                 % the epochfolding is calculated with the usual ISIS function
               lc[i].time, lc[i].rate, pmin, pmax;        % the required paramaters are given
               dp=grid_steps, dt = lcp.dt[i]*lc[i].fracexp, nbins=nbins, chatty = -1, exact,
                        ); %}}}
         } else if (nsrch > 0 and exct < 0.) {%{{{ 
            vmessage("(%s): epochfolding without exact, nsrch=%d, nbins=%d", _function_name, nsrch, nbins);
            epf = epfold(                                 % the epochfolding is calculated with the usual ISIS function
               lc[i].time, lc[i].rate, pmin, pmax;        % the required paramaters are given
               nsrch=nsrch, dt = lcp.dt[i]*lc[i].fracexp, nbins=nbins, chatty = -1,
                        ); %}}}
         } else if (nsrch < 0 and exct < 0.) {%{{{ 
            vmessage("(%s): epochfolding without exact, grid_steps=%lf, nbins=%d", _function_name, grid_steps, nbins);
            epf = epfold(                                 % the epochfolding is calculated with the usual ISIS function
               lc[i].time, lc[i].rate, pmin, pmax;        % the required paramaters are given
               dp=grid_steps, dt = lcp.dt[i]*lc[i].fracexp, nbins=nbins, chatty = -1,
                        ); %}}}
         } else if (nsrch == 0 and exct > 0.) {%{{{
            vmessage("(%s): epochfolding with exact, default, nbins=%d", _function_name, nbins);
            epf = epfold(                                 % the epochfolding is calculated with the usual ISIS function
               lc[i].time, lc[i].rate, pmin, pmax;        % the required paramaters are given
               dt = lcp.dt[i]*lc[i].fracexp, nbins=nbins, chatty = -1, exact,
                        ); %}}}
         } else if (nsrch == 0 and exct < 0.) {%{{{
            vmessage("(%s): epochfolding without exact, default, nbins=%d", _function_name, nbins);
            epf = epfold(                                 % the epochfolding is calculated with the usual ISIS function
               lc[i].time, lc[i].rate, pmin, pmax;        % the required paramaters are given
               dt = lcp.dt[i]*lc[i].fracexp, nbins=nbins, chatty = -1,
                        ); %}}}
         } else {
            vmessage ("error(%s): corrupted qualifier set", _function_name);
            return;
         }                                                                                                                                                                                                
      }%}}}

                                                                                                                                                                                                      
      out.dp[i]=lcp.dp[i];
      % save into output
      out.epfold[i] = epf;   % these values are written into the output structure
      out.time[i]   = mean(lc[i].time);                                            % the out.time is the average of the time scale
   } %}}}
   if (split_only) {out=reduce_struct(out, "epfold");}
                                                                                                                                                                                                  
   % sort by time
   variable n = array_sort(lcp.time);
   struct_filter(lcp, n);
   struct_filter(out, n);
   
   return out;                                                                    % that's it
}
%}}}

%
% Code from src/slang/structs/tables/lightcurves/split_lc_at_gaps.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define split_lc_at_gaps()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{split_lc_at_gaps}
%\synopsis{splits a light curve at gaps of a certain length}
%\usage{Struct_Type split_lc[] = split_lc_at_gaps(Struct_Type lc, Double_Type gap_threshold);}
%\qualifiers{
%\qualifier{time}{[\code{="time"}] time field}
%}
%\description
%    \code{lc} has to be a structure containing a \code{time} field, which is an array of ascending values,
%    and other fields, which are arrays of the same length. As soon as the difference of
%    sequential times is larger than \code{gap_threshold}, the structure is split, such that finally
%    an array of structures like \code{lc} is returned.
%
%   The following constructions are equivalent:\n
%      \code{split_lc_at_gaps(lc, gap_threshold)}\n
%      \code{split_struct(lc, blocks_between_gaps(lc.time, gap_threshold))}
%\seealso{split_struct, blocks_between_gaps}
%!%-
{
  variable lc, deltaT;
  switch(_NARGS)
  { case 2: (lc, deltaT)= (); }
  { help(_function_name()); return; }

  variable lcs = Struct_Type[0];
  variable t = get_struct_field(lc, qualifier("time", "time"));
  variable T0 = t[0];
  variable i, len = length(t);
  _for i (1, len, 1)
    if(i==len || t[i]>t[i-1]+deltaT)
    { % split
      variable T1 = (i<len ? t[i] : t[i-1]+deltaT);
      lcs = [lcs, struct_filter(lc, where(T0 <= t < T1); copy)];
      T0 = T1;
    }
  return lcs;
}

%
% Code from src/slang/structs/tables/split_struct.sl
%
define split_struct()
%!%+
%\function{split_struct}
%\synopsis{splits a structure of related arrays into several structures of the same kind}
%\usage{Struct_Type structs[] = split_struct(Struct_Type s, Integer_Type group[]);}
%\description
%    All fields of the structure \code{s}, as well as \code{group}, have to be related arrays
%    of equal length. The array elements with the same corresponding \code{group} value
%    are selected to form a new structure, unless the \code{group} value is negative.
%    In the latter case (\code{group<0}), the array elements are discarded.
%    The split structures are returned in an array, indexed by the \code{group} value.
%    In other words:\n
%       \code{structs[g].field = s.field[where(group==g)]}\n
%    for \code{0 <= g <= max(group)} and any \code{field} of the structure \code{s}.
%\examples
%    \code{variable prime_struct = struct { i, p };}\n
%    \code{prime_struct.i = [ 1,  2,  3,  4,  5,  6,  7,  8,  9, 10];}\n
%    \code{prime_struct.p = [ 2,  3,  5,  7, 11, 13, 17, 19, 23, 29];}\n
%
%    \code{variable group = [-1,  1, -1,  2,  0,  1,  2,  3,  1,  3];}\n
%    \code{% which might be obtained from the following selection:}\n
%    \code{group = Integer_Type[length(prime_struct.p)];}\n
%    \code{group[*] = -1;}\n
%    \code{group[where(prime_struct.p mod 10 == 1)] = 0;}\n
%    \code{group[where(prime_struct.p mod 10 == 3)] = 1;}\n
%    \code{group[where(prime_struct.p mod 10 == 7)] = 2;}\n
%    \code{group[where(prime_struct.p mod 10 == 9)] = 3;}\n
%
%    \code{variable primes = split_struct(prime_struct, group);}\n
%    \code{writecol(stdout, primes[1].i, primes[1].p);  % primes with last digit 3}\n
%\seealso{split_lc_at_gaps}
%!%-
{
  variable s, group;
  switch(_NARGS)
  { case 2: (s, group)= (); }
  { help(_function_name()); return; }

  % lots of checks necessary...

  variable g, structs = Struct_Type[0];
  _for g (0, max(group), 1)
    structs = [ structs, struct_filter(s, where(group==g); copy) ];
  return structs;
}

%
% Code from src/slang/structs/split_struct_cols.sl
%
define split_struct_cols(str,num)
%!%+
%\function{split_struct_cols}
%\synopsis{split structure in an array of structures with maximal num
%columns}
%\usage{Struct_Type str_array = split_struct_cols(Struct_Type str,
%Integer_Type num);}
%\description
%
%     This function splits a structure in an array of structures,
%     with maximal 'num' columns in each structure.
%
%     This can be useful, as fits-tables only allow 999 colums to
%     be written in the same extension. With split_struct_cols these
%     can be splitted an written into different extensions. Afterwards
%     they can be easily combined with struct_combine().
%     
%\seealso{fits_save_fit, struct_combine}
%!%-
{  
  
  variable names = get_struct_field_names(str);
  
  variable n = int((length(names)-1)/num)+1;
  
  variable strs = Struct_Type[n];
  variable i;
  _for i(0,n-1,1)
  {
    variable ind,ss;
    ind = (i < n-1) ? (names[[0:num-1]+num*i]) : (names[[num*i:]]);
    
    strs[i] = @Struct_Type(ind);
    foreach ss(ind)
    {
      set_struct_field(strs[i],ss,get_struct_field(str,ss));
    }
  }

  return strs;
}

%
% Code from src/fitting/xyfit/fit-functions/sqrsinc_xyfit.sl
%
define sqrsinc_xyfit() { %{{{
   variable xref, yref, par;
   switch (_NARGS)
   { case 0: return ["center [center]", "norm [norm]", "width [width]", "offset [offset]", "asymmetry [asymmetry]" ]; }
   { case 3: (xref, yref, par) = (); }
   { return; }
   variable x = @xref - par[0];
   variable xx = x*(1+(1+sign(x))*par[4])/par[2];
   @yref = par[3] + par[1] * sqr( sin(xx)/(xx) );
}
%}}}

%
% Code from src/misc/radio/stacking_vlbi_images.sl
%

% only activate if module gcontour could be loaded
#ifeval __get_reference("gcontour_compute")!=NULL

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% convert pixel to mas
private define convMAS(pix,ref_pix,del) {
  return (pix - ref_pix+0.5)*del*3.6e6;}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define stacking_vlbi_images() 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{stacking_vlbi_images}
%\synopsis{stack VLBI images of same size}
%\usage{stacking_vlbi_images(Array_Type \code{string(cleanfiles)}, String_Type \code{outputfile});}
%\qualifiers{
%\qualifier{ra_mas}{[={20,-20}] {left,right} limits of image in mas}
%\qualifier{dec_mas}{[={-20,20}] {top,bottom} limits of image in mas}
%\qualifier{plot_size}{[=15] size of plot}
%\qualifier{n_sigma}{[=3.0] lowest contour of clean image}
%\qualifier{sourcename}{[=default] name of the source, by default the name is read from the
%                             .fits file, set to NULL for not plotting a source name}
%\qualifier{obs_date1}{[=default] observation date, by default the date of the first
%                              observation is used}
%\qualifier{obs_date2}{[=default] 2nd observation date, by default the date of the last
%                              observation is used}
%\qualifier{cont_scl}{[="2"] set factor to change the separation between contour levels}
%\qualifier{cont_lvl}{[=[c1,c2,..] contour levels [Jy] manually, overwrites other contour parameters}
%\qualifier{major}{specify a vector for the major tic marks of the plot}
%\qualifier{minor}{specify a vector for the minor tic marks of the plot}
%}
%\description
%    This function creates a stacked image of individual VLBI-clean images.\n 
%    Important note: all images need to be of the same size! Different beam sizes are 
%    NOT recognized or corrected.\n
%    The required input format are fits-file for both the clean and the modelfit images.
%    The format of the output file depends on the suffix of the given\n
%    \code{filename}. Possible formats of the output file are PDF, EPS,\n
%    PNG, GIF, etc.
%!%-
{
variable cleanfiles, outputfile;

  switch(_NARGS)
  { case 2: (cleanfiles,outputfile) = (); }
  { help(_function_name()); return; }
  %%%%%%%%%%%%
  variable nsigma    	  = qualifier("n_sigma",3.0);
  variable sourcename	  = qualifier("sourcename", fits_read_key (cleanfiles[0], "OBJECT"));
  variable comp_label	  = qualifier("comp_label", 0);
  variable label_color	  = qualifier("label_color", "red");
  variable ra_mas         = qualifier("ra_mas",  [ 20,-20]);
  variable dec_mas	  = qualifier("dec_mas", [-20, 20]); 

  % MB: the following variables are not used?
  %  variable ex_counterjet  = qualifier("ex_counterjet",0);
  %  variable counterjet_ra  = qualifier("counterjet_ra", NULL);
  %  variable counterjet_dec = qualifier("counterjet_dec", NULL);
  %  variable ind_inverse    = qualifier("ind_inverse",0);
  %  variable label_alt      = qualifier("label_alt", NULL);

  variable n = length(cleanfiles);

  variable obs_date = array_map (String_Type, &fits_read_key,   cleanfiles, "DATE-OBS");
  variable mjd      = array_map (Double_Type, &MJDofDateString, obs_date);
  variable obs_date1       = qualifier("obs_date1", strftime_MJD("%Y-%m-%d", min(mjd))  );
  variable obs_date2       = qualifier("obs_date2", strftime_MJD("%Y-%m-%d", max(mjd))  );

  % MB: estimate noise from all images:
  %  variable noise_i  = array_map (Double_Type, &fits_read_key, cleanfiles, "NOISE");
  %  variable noise    = hypot(noise_i);
  
  variable ii;
  _for ii (0,n-1,1){
    variable file = fits_read_img(cleanfiles[ii])[0,0,*,*];
    %%% add images:
    variable image;
    if (ii==0) {image =@file;}
    else image +=@file;
  }

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % fit background noise in order to determine contour levels
  % (for details see plot_vlbi_map)
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %
  % MB:    this should be done before cutting (and modifying the image), because
  %        then more background emission is included and the fitted noise value
  %        is less affected from significant jet emission
  variable map_min = min(image);
  variable map_max = max(image);
  variable mu, sigma;
  (mu,sigma) = fit_gauss_to_img_noise (image);
  variable cl    = qualifier("cont_scl", 2.);
  variable n_pos = int(log( (map_max - mu) / (nsigma*sigma)) / log(cl));
  variable n_neg = int(log( (mu - map_min) / (nsigma*sigma)) / log(cl));
  variable lvls  = qualifier("cont_lvl", mu + [-(cl^[0:n_neg]),cl^[0:n_pos]]*nsigma*sigma );

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % get coordinates and cut image
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  variable  pix1   = fits_read_key(cleanfiles[0],"CRPIX1");
  variable  pix2   = fits_read_key(cleanfiles[0],"CRPIX2");
  variable  delta1 = fits_read_key(cleanfiles[0],"CDELT1");
  variable  delta2 = fits_read_key(cleanfiles[0],"CDELT2");
  variable x, y;
  (x,y) = get_coordinatearrays_of_image(image);
  (x,y) = (convMAS(x,pix1,delta1),convMAS(y,pix2,delta2));
  variable xplotmin =  ra_mas[0]; variable xplotmax =  ra_mas[1];
  variable yplotmin = dec_mas[0]; variable yplotmax = dec_mas[1];
  variable xmin =  wherefirst (x[0,*] <= xplotmin);
  variable xmax =  wherelast  (x[0,*] >= xplotmax);
  variable ymin =  wherefirst (y[*,0] >= yplotmin);
  variable ymax =  wherelast  (y[*,0] <= yplotmax);
  image   = image [[ymin:ymax],[xmin:xmax]];
  x       = x     [[ymin:ymax],[xmin:xmax]];
  y       = y     [[ymin:ymax],[xmin:xmax]];

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % define plot 
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  variable i,j;
  variable width = qualifier("plot_size",15.);
  variable height = width/abs(xplotmin-xplotmax)*abs(yplotmax-yplotmin);
  variable plot = xfig_plot_new(width, height);
  plot.world(xplotmin, xplotmax, yplotmin, yplotmax);	% world coordinates in mas
  plot.xlabel("relative RA [mas]");
  plot.ylabel("relative DEC [mas]");

  if (qualifier_exists("major") or qualifier_exists("minor"))
  {
    variable major_ticks = qualifier("major", [-100:100:5]);
    variable minor_ticks = qualifier("minor", [-100:100:1]);
    plot.xaxis(; major=major_ticks, minor= minor_ticks);
    plot.yaxis(; major=major_ticks, minor= minor_ticks);
    plot.y2axis(; ticlabels=0);
    plot.x2axis(; ticlabels=0);
  }
  
  % add source and date:
  variable xpos_text = 0.06;
  variable ypos_text = 0.93;
  if (sourcename != NULL){
    plot.xylabel(xpos_text,ypos_text, sourcename, -0.5,0.5; world0, color="black");
  }
  plot.xylabel(1-xpos_text,ypos_text, obs_date1+" to \n"+obs_date2, 0.5,0.5; world0, color="black");

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % plot contours of clean image
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  variable xref = pix1-xmin; variable yref = pix2-ymin;
  variable linewidth = Integer_Type[length(lvls)]; linewidth[*] = 1;
  variable line = Integer_Type[length(lvls)]; line[*] = 0;
  variable col = String_Type[length(lvls)]; col[*] = qualifier("cont_color","black") ;
  % % plot contours
  variable gcl = gcontour_compute(image, lvls);
  _for i (0, length(gcl)-1, 1) {
    _for j (0, length(gcl[i].x_list)-1, 1) {
      plot.plot(convMAS(gcl[i].x_list[j],xref,delta1),
		convMAS(gcl[i].y_list[j],yref,delta2);
		width=linewidth[i], line=line[i], color=col[i]);
    }
  }

  plot.render(outputfile);
}

#endif

#endif

#endif

%
% Code from src/fitting/fit-functions/step1.sl
%
%%%%%%%%%%%%%%%%
define step1_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%
%!%+
%\function{step1 (fit-function)}
%\synopsis{a step function which is 1 above x0}
%\description
%    If one bin encloses \code{x0}, the value of step1 function
%    is the fraction of the bin which is above \code{x0}.
%\seealso{step2}
%!%-
{
  variable value = 0*bin_lo;
  value[where(bin_lo>=par[0])] = 1;

  variable i = where(bin_lo < par[0] < bin_hi);
  value[i] = (bin_hi[i]-par[0])/(bin_hi[i]-bin_lo[i]);
  return value;
}

add_slang_function("step1", ["x0"]);

%
% Code from src/fitting/fit-functions/step2.sl
%
%%%%%%%%%%%%%%%%
define step2_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%
%!%+
%\function{step2 (fit-function)}
%\synopsis{a step function which is 1 between x1 and x2}
%    If one bin encloses \code{x0} or \code{x2}, the value of the step2 function
%    is the fraction of the bin which is between \code{x1} and \code{x2}.
%\seealso{step1}
%!%-
{
  variable value = 0*bin_lo;

  % par[0]  bin_lo                  bin_hi  par[1]
  % {       [----------------------------]       }
  value[ where(par[0] < bin_lo and bin_hi < par[1]) ] = 1;

  %         bin_lo  par[0]          bin_hi  par[1]
  %         [       {--------------------]       }
  variable i = where(bin_lo < par[0] < bin_hi < par[1]);
  value[i] = (bin_hi[i]-par[0])/(bin_hi[i]-bin_lo[i]);

  % par[0]  bin_lo          par[1]  bin_hi
  % {       [--------------------}       ]
  i = where(par[0] < bin_lo < par[1] < bin_hi);
  value[i] = (par[1]-bin_lo[i])/(bin_hi[i]-bin_lo[i]);

  %         bin_lo  par[0]  par[1]  bin_hi
  %         [       {------------}       ]
  i = where(bin_lo < par[0] < par[1] < bin_hi);
  value[i] = (par[1]-par[0])/(bin_hi[i]-bin_lo[i]);

  return value;
}

add_slang_function("step2", ["x1", "x2"]);

%
% Code from src/fitting/steppar.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL
% -*- mode: slang; mode: fold -*-

private variable steppar_mainkeys = %{{{
  % list of the main keys for steppar
  strlow(["extname",
	  "fit_method",
	  "fit",
	  "statistic",
	  "num_variable_params",
	  "num_bins",
	  "pname",
	  "pidx",
	  "pval",
	  "pmin",
	  "pmax"
	 ]);
%}}}
private define struct_fieldnames_to_lowercase( s ){ %{{{
  % converts the fieldnames of a Struct_Type to lower cases!
  variable fn = get_struct_field_names( s );
  variable ls = @Struct_Type(strlow(fn));

  variable f;
  foreach f(fn){
    set_struct_field( ls, strlow(f), get_struct_field( s, f ) );
  }
  return ls;
}
%}}}


%%%%%%%%%%%%%%%%%%%%
define steppar_save() %{{{
%%%%%%%%%%%%%%
%!%+
%\function{steppar_save}
%\synopsis{Saves steppar information (& keys) to a *.fits file}
%\usage{steppar_save( String_Type file, Struct_Type[] steppar [, keys]);}
%\description
%      Saves the output of 'steppar' to a *.fits file. 
%\seealso{steppar}
%!%-
{
  variable file, sp, keys=NULL;
  switch(_NARGS)
  { case 2: (file, sp) = (); }
  { case 3: (file, sp, keys) = (); }
  { help(_function_name);return;}

  variable nf = length(sp);
  variable extnames;
  if( keys==NULL ){
    keys = struct_array(nf,@Struct_Type(String_Type[0]));
    extnames = array_map( String_Type, &sprintf, "PID%05d",[1:nf] );
  }
  else{
    extnames = array_struct_field ( keys, "extname");
  }
  variable i = wherenot(is_substr(extnames,"PID"));
  extnames[i] = "PID"+extnames[i];
  
  variable fp = fits_open_file(file,"c");
  _for i ( 0, nf-1, 1 ){
    fits_write_binary_table( fp, extnames[i], sp[i], keys[i] );
  }
  fits_close_file(fp);
}
%}}}

%%%%%%%%%%%%%%%%%%%%%
define steppar_load() %{{{
%%%%%%%%%%%%%%
%!%+
%\function{steppar_load}
%\synopsis{loads saved steppar information}
%\usage{ Struct_Type[] steppar = steppar_load( String_Typep[] pat ); }
%\altusage{ Struct_Type[] steppar, keys = steppar_load( String_Type[] pat ; keys ); }
%\qualifiers{
%\qualifier{keys}{If given also the key structure is returned (using fits_read_header).
%                   If keys is a List_Type with keynames, only these keys are read using
%                   fits_read_key_struct, which is faster!}
%\qualifier{ext}{Only loads extension fitting 'ext' (String_Type)!}
%}
%\description
%     Loads a *.fits file storing the steppar information saved with
%     steppar_save. The given string 'pat' can be a single file
%     including several extension, an array with files or used as
%     a pattern for multiple files in globbing format (see glob).
%     
%\seealso{steppar, steppar_save, glob}
%!%-
{
  variable ext = qualifier("ext",NULL);
  variable chatty = qualifier_exists("chatty") ? 1:0;
  
  variable file;
  switch(_NARGS)
  { case 1: file = (); }
  { help(_function_name);return;}

  % Find all files matching the given pattern 'file' and sort them
  file = glob(file);
  file = file[array_sort(file)];
  
  variable sp = Struct_Type[0];
  variable sk = Struct_Type[0];
  
  variable keys = qualifier("keys");

  variable N=0,next;
  variable extind = [*];
  variable extread;
  
  % LOOP over each file
  variable f,i;
  foreach f(file){

    ifnot( access( f, R_OK )==0 ){
      vmessage( "ERROR <%s>: No permission to read file:\n %s",
		_function_name,f);
      continue;
    }
    
    next = fits_nr_extensions (f);
    if( next < 0 ){
      vmessage( "WARNING <%s>: No extension in file:\n %s",
		_function_name,f);
      continue;
    }
    sp   = [ sp, Struct_Type[next] ];
    if( qualifier_exists("keys") != NULL ){
      sk = [ sk, Struct_Type[next] ];
    }

    if( chatty ){
      vmessage(" INFO <%s>: Loading '%s' with %d extensions!",
	       _function_name, f, next );
    }
    
    % LOOP over each extension of current file:
    _for i ( 1, next, 1 ){
      % If a specific extension name is given, only load this extension!
      extread = fits_read_key( sprintf("%s[%d]",f,i), "extname" );
      if( ext != NULL and ext != extread ){
	continue;
      }
      
      sp[i-1+N]   = fits_read_table( sprintf("%s[%d]",f,i); casesen );
      
      % Only read header(keys), if qualifier exists and convert names to lower case
      if( qualifier_exists("keys") ){
	if( typeof(keys) == List_Type ){
	  sk[i-1+N] = fits_read_key_struct( sprintf("%s[%d]",f,i), __push_list(keys) );
	}
	else{
	  sk[i-1+N] = fits_read_header( sprintf("%s[%d]",f,i) );
	  sk[i-1+N] = struct_fieldnames_to_lowercase( sk[i-1+N] );
	}
      }
      
      if( ext == extread ){
	extind = i-1+N;
	break;
      }
    }
    % If a specific extension name is given, stop loading if it was found!
    if( ext == extread ){
      break;
    }
    N += next;
  }
  
  if( ext != NULL and length(extind) == 0 ){
    vmessage("WARNING: <%s>: No extension '%s' found!",_function_name,ext);
    extind = Integer_Type[0];
  }
  
  if( qualifier_exists("keys") )
    return sp[extind],sk[extind];
  else
    return sp[extind];
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define _steppar_get_params() %{{{
{
  variable chatty = qualifier_exists("chatty") ? 1 : 0;
  variable paridx = qualifier("paridx");
  
  variable sp, ind;
  switch(_NARGS)
  { case 2: (sp,ind) = (); }
  { return NULL; }
  
  variable params = get_params;
  variable newparams = Struct_Type[0];
  
  variable i, pname, val;
  _for i ( 0, length(params)-1, 1 ){
    if(qualifier_exists("paridx") && not(any(paridx==params[i].index)) ){
      continue;
    }    
    pname = escapedParameterName(params[i].name);
    if( struct_field_exists( sp, pname ) ){
      val = get_struct_field( sp, pname )[ind];
      if( chatty ){
	vmessage(" Setting: (%.3d) %s = %g  was  %g",
		 params[i].index, params[i].name, val, params[i].value );
      }
      newparams = [ newparams, Struct_Type[1] ];
      newparams[-1] = COPY( params[i] );
      set_struct_field( newparams[-1], "value", val );
    }
  }
  return newparams;
}
%}}}
private define _steppar_get_bestparams() %{{{
{
  variable fv = Fit_Verbose;
  Fit_Verbose = -1;
  variable minchi2 = qualifier("minchi2",eval_stat_counts.statistic);
  Fit_Verbose = fv;

  variable chatty = qualifier_exists("chatty") ? 1 : 0;
  
  variable sp;
  switch(_NARGS)
  { case 1: sp = (); }
  { help(_function_name); return; }

  if( _typeof( sp ) == String_Type )
    sp = steppar_load( sp );
  else ifnot( _typeof(sp)==Struct_Type ){
    vmessage("ERROR: <%s>: Argument has to be String or Struct_Type!",
	     _function_name); return NULL;
  }
  
  % Find indices where chi2 is minimal
  variable imin = NULL;
  variable jmin, chi2, noz;
  variable _vmin, vmin = minchi2;
 
  variable i;
  _for i ( 0, length(sp)-1, 1 ){
    chi2 = get_struct_field( sp[i], "chi2" );
    noz = where(chi2>0);
    _vmin = min(chi2[noz]);
    if( _vmin < vmin){
      vmin = _vmin;
      imin = i;
      jmin = noz[where_min(chi2[noz])][0];
    }
  }
  if( imin == NULL ){
    if( chatty )
      vmessage(" WARNING <%s>: No better parameter set found, which beats chi2=%g",
	       _function_name, minchi2 );
    
    return Struct_Type[0];
  }
  else if ( chatty ){
    vmessage(" INFO <%s>: Better parameter set for '%s' with chi2=%g found (%d,%d)!",
	     _function_name,
	     get_struct_field_names(sp[imin])[0],
	     vmin,imin,jmin
	    );
  }
  return _steppar_get_params( sp[imin], jmin ;; __qualifiers );
}
%}}}%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define steppar_get_bestparams() %{{{
%%%%%%%%%%%%%%
%!%+
%\function{steppar_get_bestparams}
%\synopsis{ Extracts new best fit parameter values out of steppar information }
%\usage{ Struct_Type[] params = steppar_get_bestparams( String_Type[] File );}
%\altusage{ Struct_Type[] params = steppar_get_bestparams( Struct_Type[] steppar );}
%\qualifiers{
%\qualifier{chatty}{Information output}
%\qualifier{minchi2}{Chi2 limit the new parameter set has to beat
%                      to be taken into accounts!
%                      Default: eval_stat_counts.statistic}
%\qualifier{groups}{Allows to extract best parameter sets for defined parameter
%                     groups. 'groups' has to be an Array_Type, where each entry
%                     represents a group and contains an Integer/String array
%                     with the according parameter indices/names.
%                     Useful in combination with 'simultaneous_fit' settings, e.i.
%                     groups = %.model.groups!}
%}
%\description
%      This function searches for the minimal chisqr value within the given
%      files/steppar information and returns a parameter structure with the
%      according parameters (see get_params). With the groups qualifier a
%      groupping can be specified, i.e., the search for a minimal chi2 is
%      done for each individual groups. The returned structure array
%      only contains those parameter, which lead to the better chi2 value.
%      If no better chi2 was found or there was no better chi2 than the
%      specified minchi2 value, the function returns Struct_Type[0].
%
%      ATTENTION:
%      It is neccessary that the model with which the steppar was executed
%      is loaded as this functions uses get_params and only changes those
%      values of the freeParameters as the values of the frozen ones are
%      not saved!
%      If another parameter set is loaded there will be a difference in
%      the new chi2 value this function gives (use 'chatty') and that
%      chi2 an eval_counts will return!
%\seealso{get_params}
%!%-
{
  variable chatty = qualifier_exists("chatty") ? 1 : 0;
  
  variable groups = Array_Type[1];
  groups[0] = [1:get_num_pars];
  groups = qualifier("groups",groups);

  variable sp;
  switch(_NARGS)
  { case 1: sp = (); }
  { help(_function_name); return; }

  if( _typeof( sp ) == String_Type )
    sp = steppar_load( sp );
  else ifnot( _typeof(sp)==Struct_Type ){
    vmessage("ERROR: <%s>: Argument has to be String or Struct_Type!",
	     _function_name); return NULL;
  }
  
  variable escnames, grpind, ind;

  variable newparams = Struct_Type[0];

  % get names of stepped parameter
  variable spnames = String_Type[0];
  variable i, j;
  _for i ( 0, length(sp)-1, 1 ){
    spnames = [ spnames, get_struct_field_names( sp[i] )[0] ];
  }
  % get the parameters for each group
  _for i ( 0, length(groups)-1, 1 ){
    if( chatty and qualifier_exists("groups") ){
      vmessage(" INFO <%s>: Group %d/%d:",
	       _function_name, i+1, length(groups) );
    }

    % ensure that groups are parameter indices!
    grpind    = array_struct_field( get_par_info( groups[i] ), "index" );
    % get escaped parameter names
    escnames  = array_struct_field( get_par_info( groups[i] ), "name" );
    escnames  = array_map( String_Type, &escapedParameterName, escnames );
    % determine steppar array indices
    ind = Integer_Type[0];
    _for j ( 0, length(spnames)-1, 1 ){
      if( any( spnames[j] == escnames ) ){
	ind = [ ind, j ];
      }
    }
    newparams = [ newparams,
		 _steppar_get_bestparams( sp[ind];;struct_combine(__qualifiers,struct{paridx=grpind}))
		];
  }
  return newparams;
}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define array_split_at_extrema( a ) %{{{
%!%+
%\function{array_split_at_extrema}
%\synopsis{Splits an array into monotone chucks}
%\usage{ Array_Type[] I = array_split_at_extrema( array );}
%!%-
{
  variable I = Array_Type[1];
  I[0] = [0,1];
  
  variable signum = sign((shift(a,1)-a)[[:-2]]);
  signum[where(signum==0)]=1;
  
  variable i, ind=0;
  _for i ( 1, length(signum)-1, 1 ){
    if( signum[i] == signum[i-1] ){
      I[ind] = [ I[ind], i+1 ];
    }
    else{
      ind++;
      I = [ I, Array_Type[1] ];
      I[ind] = [i,i+1];
    }
  }
  return I;
}
%}}}
define steppar_get_conf() %{{{
%%%%%%%%%%%%%%
%!%+
%\function{steppar_get_conf}
%\synopsis{Obtains confidence limits out of steppar information}
%\usage{ Struct_Type[] conf = steppar_get_conf( String_Type[] File );}
%\altusage{ Struct_Type[] conf = steppar_get_conf( Struct_Type[] steppar );}
%\qualifiers{
%\qualifier{chatty}{Information output}
%\qualifier{dchi2}{[=2.71] Delta Chisqr}
%}
%\description
%      Tries to obtain confidence level related to the given 'dchi2'
%      value (default: 2.71) based on the steppar information.
%      To do so this function calculates interpolated intersections
%      of the chi2 landscape and min(chi2)+'dchi2'.
%      Only if 2 or more such intersections are found the confidence
%      limits are set. In case of more then 2 solutions always the
%      minimal/maximal solutions are taken.
%\seealso{interpol}
%!%-
{
  variable chatty = qualifier_exists("chatty") ? 1 : 0;
  variable dchi2   = qualifier("dchi2",2.71);
  
  variable sp;
  switch(_NARGS)
  { case 1: sp = (); }
  { help(_function_name); return; }

  if( _typeof( sp ) == String_Type )
    sp = steppar_load( sp );
  else ifnot( _typeof(sp)==Struct_Type ){
    vmessage("ERROR: <%s>: Argument has to be String or Struct_Type!",
	     _function_name); return;
  }

  variable N = length(sp);
  variable s = struct{
    index, name, value, min, max, conf_min, conf_max    
  };
  s = struct_array( N, s );

  % OBTAIN PAR INDICES
  variable pnames = array_map( String_Type, &escapedParameterName, array_struct_field( get_params, "name" ) );
  
  variable fn, p, pmin, pmax, chimin;
  variable sa, isort;
  variable _isec, isecs;
  variable i,j;
  _for i ( 0, N-1, 1 ){
    fn = get_struct_field_names(sp[i])[0];
    p = get_struct_field( sp[i], fn );
    s[i].min   = min(p);
    s[i].max   = max(p);
    s[i].name  = fn;
    s[i].index = wherefirst( fn == pnames )+1;
    
    chimin = min( sp[i].chi2 );
    sa = array_split_at_extrema(sp[i].chi2);
    isecs = Double_Type[0];
    _for j ( 0, length(sa)-1, 1 ){
      isort = array_sort(sp[i].chi2[sa[j]]);
      _isec = interpol( chimin+dchi2, sp[i].chi2[sa[j]][isort], p[sa[j]][isort] );
      pmin = min(p[sa[j]]);
      pmax = max(p[sa[j]]);
      if( pmin <= _isec <= pmax ){
	isecs = [ isecs, _isec ];
      }
    }
    if( length(isecs) > 1 ){
      s[i].conf_min = isecs[0];
      s[i].conf_max = isecs[-1];
    }
  }
  
  return s;
}
%}}}

%%%%%%%%%%%%%%%
define steppar() %{{{
%%%%%%%%%%%%%%
%!%+
%\function{steppar}
%\synopsis{performs a fit while stepping the value of a fit parameter through a given range}
%\usage{Struct_Type info = steppar(String/Integer_Type par [, Double_Type val1, val2, step] );}
%\altusage{(Struct_Type info, keys) = steppar( String/Integer_Type par [, Double_Type val1, val2, step]; keys);}
%\qualifiers{
%\qualifier{keys}{An additional structure is returned that can be used as FITS header keys}
%\qualifier{frozen}{[=0] Perform steppar also for frozen Parameters.}
%\qualifier{range}{[parmin,parmax] Stepping range. Default is the
%                   minimal/maximal allowed parameter value.}
%\qualifier{nsteps}{[=10] Number of steps the parameter 'par' is
%                   stepped from range[0] to range[1].}
%\qualifier{reset}{If given after each step the initial parameter set, which was
%                    valid before this function was called, is restored.}
%\qualifier{fit}{[=&fit_counts] Reference to the function running the fit algorithm.}
%\qualifier{fitargs}{Arguments required by the 'fit' function. See __push_list for
%                    format information!}
%\qualifier{rerun}{Reruns a stepping procedure based on results of a previous run,
%                    i.e., before each step the according parameter set of the previous
%                    run is loaded, e.g., to improve the results using another fit_method.}
%\qualifier{resume}{Missing steps in the given steppar-file will be calculated based
%                     on the previous steps, i.e., resuming the stepping where it was
%                     stopped.}
%\qualifier{check}{[=0] Before each step saved steppar informations of other stepped
%                        parameters are gathered and checked for a parameter set with a
%                        better chi2 (using steppar_get_bestparams). In case a better
%                        parameter set was found the stepping will be restarted.
%                        * ATTENTION: 'save' qualifier is required !!!
%                        * Gathered are steppar-files, which match the pattern
%                          given with 'save',e.g., if save is "steppar_PID00001.fits"
%                          all files "steppar_PID?????.fits" are globed!
%                          Note that the affix "_PID???" is automaticcaly appended to
%                          the save string if it does not exist (see 'save' qualifier!)
%                        * The Integer 'check' is set to is the maximal number of
%                          restarts.
%                        * A parameter set is considered better if
%                          chi2_new < chi2_init * ( 1 - dchi2 )
%                        * NOTE: Parameter grouping is possible (see steppar_get_bestparams)
%                        }
%\qualifier{dchi2}{[=0.1] Percental limit for the chi2 of a new parameter set to be
%                  considered as a better parameter set (see 'check'):
%                  chi2_new < chi2_init * ( 1 - dchi2 ).}
%\qualifier{save}{After each stepping the result is saved in the file given
%                   (as String_Type) with this qualifier. The chi2 of undone steps
%                   are set to 0.! Also note that the filename is appended with
%                   the parameterindex: 'steppar.fits' -> 'steppar_PID00001.fits'}
%\qualifier{force}{Forces to overwrite an existing file given with 'save'.}
%\qualifier{chatty}{Prints fitting information.}
%}
%\description
%    The given parameter is stepped through in the given parameter 'range',
%    which is devided into 'nsteps' equidistant value points. If 'val1', 'val2'
%    and 'step' are given 'range' = ['val1','val2'] and
%    'nsteps' = ('val2'-'val1')/'step'.
%    At each of this points a fit is performed. The fitting starts with the
%    value point closed to the best fit value and alternatingly progresses outwards
%    to each side. The 'reset' qualifier restores the initial parameter set
%    after each step, otherwise the parameter set of the nearest stepping point
%    is used.
%    The 'rerun' qualifier allows to rerun a stepping using the steppar information
%    of a previous run, i.e., at each step the according parameter set of that
%    previous run is loaded before the fit algorithm is started.
%    In case the steppar function was killed, the 'resume' qualifier
%    can be used to resume the stepping (given that 'save' was used!).
%    If the 'save' qualifier is given the results of the stepping are stored
%    as a .fits file after each step! NOTE that the affix "_PID?????" is
%    automatically appended to the filename (if not already included), where
%    the ??? are the ID of the stepped parameter. If the specified file
%    already exists an error is thrown if the 'force' qualifier for overwriting
%    is not set.
%    The 'save' qualifier is also requiered if the 'check' for a better
%    parameter set is enabled, i.e., before each step it is checked if there
%    is a parameter set amongst other stepped parameters with
%    chi2 < chi2_init * ( 1 - 'dchi2') leading to a restart of the stepping
%    using the better parameter set, where chi2_init is the chi2 of the parameter
%    set the stepping was initialized with (which will be updated after a restart).
%    The maximal number of restarts is given by 'check'.
%    This qualifier is useful if several steppar processes are running at the same
%    time saving their results after each step to a fits-file, as in this case
%    a better fit will be automatically applied! NOTE THAT using 'check'
%    can lead to inconsistent steppar resulsts in terms of inital parameter sets,
%    e.g., one steppar process is already finished and afterwards in another one
%    a better parameter set is found leading to a reastart in the remaining
%    steppar processes!
%
%    => After using 'check' either make sure all steppar run with the same initial
%       parameter set or manually extract the best parameter set and run all steppars
%       once more!    
%    
%    The fitting itself is performed using the 'fit' function, to which
%    all qualifiers are passed, which gets the 'fitargs' as arguments (see
%    __push_list).
%    
%    IMPORTANT:
%      steppar is based on the current version of the parameter set!
%       
%    KEYS:
%     fit_method: fit method used for fitting (see get_fit_method)
%     statistic:  chisqr according to the parameter set the function was callled with
%     num_variable_params:
%     num_bins:
%     pval:       initial value of the stepped parameter    
%     pname:      exact name of the stepped parameter
%     %_freeze, %_min, %_max: additional information about the parameters, where
%                             % is the escapedParameterName
%\seealso{}
%!%-
{
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %%% QUALIFIERS

  % fkt is a function pointer executed in each step, called
  % with fitargs as arguments (see __push_list)
  variable fit     = qualifier("fit",&fit_counts);
  variable fitargs = qualifier("fitargs",{});
  
  variable frozen = qualifier_exists("frozen") ? 1 : 0;
  variable reset  = qualifier_exists("reset") ? 1 : 0;
  variable nsteps = nint(qualifier("nsteps",10));
  variable range  = qualifier("range");

  variable chatty = qualifier_exists("chatty") ? 1 : 0;
  variable fitverbosity = Fit_Verbose;
  
  variable save   = qualifier("save");
  variable rerun = qualifier("rerun",NULL);
  variable resume = qualifier("resume",NULL);
  variable force  = qualifier_exists("force") ? 1 : 0;

  variable check  = qualifier("check",0);
  variable dchi2  = qualifier("dchi2",0.1);
  if( check and not(qualifier_exists("save")) ){
    vmessage("ERROR <%s>: 'check' requires 'save' qualifier!",
	     _function_name);
    return;
  }
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  %%%  ARGUMENTS
  variable par, val1, val2, step;
  switch(_NARGS)
  { case 1: par = (); }
  { case 4: (par, val1, val2, step) = ();
    range  = [_min(val1,val2),_max(val1,val2)];
    nsteps = nint(abs(val2-val1)/step);
  }
  { help(_function_name); return; }

  ifnot( chatty ){ Fit_Verbose = -1; }
  
  %%% Parameter checks:
  % check for parameter existence (if multiple, chosing first)
  variable parinfo = get_par_info(par)[0];
  if( parinfo == NULL ){
    vmessage("ERROR <%s>: No parameter to given par=%s found!",
	     _function_name,par );
    return NULL;
  }
  par = parinfo.name;
  
  variable pidstr = sprintf("PID%05d",parinfo.index);

  % SAVE: Preparation
  variable path, fname, base;  
  if( qualifier_exists("save") ){
    % rearrange save path
    path = path_dirname(save) + "/";
    fname = path_basename_sans_extname(save);
    ifnot( is_substr(fname,pidstr) ){
      fname += "_" + pidstr;
    }
    fname += ".fits";
    save = path+fname;
    
    % Check Directory Tree (create it if not existent)
    if( access(path,F_OK) == -1 ){
      ()=mkdir_rec(path);
    }
    
    % Does file already exist?
    if( access( save, F_OK) == 0 and force == 0 ){
      vmessage("ERROR: <%s> File('%s') does already exist (Use 'force' qualifier to overwrite!",
	       _function_name, save );
      return NULL;
    }
  }

  % check if frozen
  if( parinfo.freeze == 1 ){
    if( frozen ){
      vmessage("WARNING <%s>: '%s' frozen: continnuing, due to 'frozen' qualifier!",
	       _function_name, par);
    }
    else{
      vmessage("WARNING <%s>: '%s' frozen: QUITTING!",
	       _function_name, par);
      return NULL;
    }
  }
  % Check & set par range
  if( qualifier_exists("range") or _NARGS==4 ){
    range = [ min(range), max(range) ];
    if( range[0] < parinfo.hard_min ){
      vmessage("WARNING <%s>: min(range) = %.3f < parmin = %.3f! Setting min(range) = parmin!",
	       _function_name, range[0], parinfo.hard_min );
      range[0] = parinfo.hard_min;
    }
    if( range[1] > parinfo.hard_max ){
      vmessage("WARNING <%s>: max(range) = %.3f > parmax = %.3f! Setting max(range) = parmax!",
	       _function_name, range[1], parinfo.hard_max );
      range[1] = parinfo.hard_max;
    }
  }
  else{
    range = [ parinfo.min, parinfo.max ];
  }

  variable originalparams = get_params();
  
  % Initializing output struct
  variable ifields, kfields, param, paramname;
  ifields = [ escapedParameterName(par), "chi2", "chi2red" ];
  foreach param (originalparams){
    if(param.freeze==0){
      paramname = escapedParameterName(param.name);
      if(paramname != ifields[0])
	ifields = [ifields, paramname];
    }
  }
  variable info = @Struct_Type(ifields);
  foreach param (ifields){
    set_struct_field(info, param, Double_Type[nsteps]);
  }

  variable init_stat;
  () = eval_counts(&init_stat);

  % Keys
  variable keys = @Struct_Type(steppar_mainkeys);
  keys.extname  = pidstr;
  keys.fit_method = get_fit_method();
  (,keys.fit) = get_variable_name(&fit);
  %keys.statistic = init_stat.statistic;
  keys.num_variable_params = init_stat.num_variable_params;
  keys.num_bins = init_stat.num_bins;
  keys.pname = parinfo.name;
  keys.pidx  = parinfo.index;
  %keys.pval = parinfo.value;
  keys.pmin = parinfo.min;
  keys.pmax = parinfo.max;

  kfields = String_Type[0];
  foreach param (originalparams)
    kfields = [kfields, escapedParameterName(param.name) + ["_freeze", "_min", "_max"]];
  keys = struct_combine( keys, @Struct_Type(kfields) );
  foreach param (originalparams){
    paramname = escapedParameterName(param.name);
    set_struct_field(keys, paramname+"_freeze", param.freeze);
    set_struct_field(keys, paramname+"_min",    param.min);
    set_struct_field(keys, paramname+"_max",    param.max);
  }

  %%% Resume/Rerun file check
  % - In case both qualifiers are given, rerun overwrites resume!
  % - stephistory: Ref-array containing pointer to steppar info
  %                in which paramter set for initialization are
  %                searched ('info' is the current one). If file
  %                check is successfull loaded information is added
  %                to this array.
  % - rsm_sp,rsm_k: Struct_Types which store loaded steppar files
  %                 rsm_sp is set to info to ensure rsm_sp has
  %                 the same struct fields!
  
  variable stephistory = [&info];
  variable rsm_sp = COPY(info);
  variable rsm_k, lsp, lk;
  
  variable refile = qualifier("rerun",qualifier("resume",NULL));
  if( refile != NULL && access( refile, F_OK) == 0 ){
    (lsp,lk) = steppar_load( refile ; keys, ext=pidstr );
    if( length(lsp) != 0 ){
      rsm_sp = lsp[0];
      rsm_k = lk[0];
      
      stephistory = [ stephistory, &rsm_sp ];
      % rerun test, keys should be equal!
      foreach param(["extname","num_variable_params","num_bins","pname","pidx"]){
	ifnot( _eqs( get_struct_field(keys,param), get_struct_field(rsm_k,param) ) ){
	  vmessage("ERRO: <%s>: Key '%s' of file-to-rerun/resume ('%s') does not match to Parameterset!",
		   _function_name, param, save );
	  return NULL;
	}
      }
    }
  }
  else if( refile != NULL ){
    vmessage("WARNING: <%s> File('%s') to rerun does not exist!",
	     _function_name, refile );
  }

  % steppar stepping variables
  variable stepping = [range[0]:range[1]:#nsteps];
  % set the stepped parameter values
  set_struct_field(info, escapedParameterName(parinfo.name), stepping );

  
  variable iorder, ileft, iright, imax;
  variable stat, val;
  variable pdiff, pind, pdiffmin;
  variable useparams,params,newparams,qualis;
  variable checkfiles;
  
  variable nrestarts = 0, i, p;
  % LOOP over restarts. If check>0 after each step a check for new best parameters
  % is performed and the stepping is restarted.
  % NOTE THAT the check for new best params does not wait for the other steppar
  % processes, i.e., if the current stepping is finished earlier than others it will
  % finish and possible best params found afterwards are not taken into account!
  while( nrestarts <= check ){

    % ATTENTION! Some already set values change due to new best parameter set
    params = get_params;
    
    keys.pval = get_par(par);
    ()=eval_counts(&stat);
    keys.statistic = stat.statistic;

    %%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Define stepping order
    % order in which stepping is stepped (altering right/left with respect to pval)
    if( qualifier_exists("resume") ){
      % only take steps into account, which are not calculated yet
      ileft  = reverse(where( parinfo.value >  stepping and rsm_sp.chi2 == 0 ));
      iright =         where( parinfo.value <= stepping and rsm_sp.chi2 == 0 );
    }
    else{
      ileft  = reverse(where( parinfo.value >  stepping ));
      iright =         where( parinfo.value <= stepping );
    }
    % apply the altering order
    imax = min([length(iright),length(ileft)]) - 1;
    iorder = Integer_Type[0];
    _for i ( 0, imax, 1 ){
      iorder = [ iorder, iright[i], ileft[i] ];
    }
    iorder = [ iorder, iright[[imax+1:]] ];
    iorder = [ iorder, ileft[[imax+1:]] ];

    
    qualis = struct{
      minchi2 = stat.statistic*( 1. - dchi2),
    };


    % get files in which new best parameters are searched (only runs if 'save' is given)
    checkfiles = NULL;
    if( check > 0 and qualifier_exists("save") ){
      checkfiles = string_matches (save,"PID[0-9]+"R);
      if( checkfiles !=NULL ){
	checkfiles = strreplace( save, checkfiles[0], "PID?????" );
	checkfiles = glob(checkfiles);
	checkfiles = checkfiles[array_sort(checkfiles)];
	checkfiles = checkfiles[where(checkfiles != save)];
      }
    }

    % start the stepping
    foreach i ( iorder ){
      % check other parameter files for a better parameter set and restart the stepping
      if( nrestarts < check and checkfiles!=NULL ){
	newparams = steppar_get_bestparams(checkfiles;; struct_combine(__qualifiers,qualis) );
	if( length(newparams) > 0 ){
	  set_params(newparams);
	  nrestarts++;
	  break;
	}
      }
      
      val = stepping[i];
      
      % SEARCH best parameter set in means of smallest difference to the stepped par value
      useparams = params;
      if( reset==0 ){
	pdiffmin = abs( val - parinfo.value );
	foreach p ( stephistory ){
	  pdiff = abs( val - get_struct_field(@p,escapedParameterName(parinfo.name)));
	  pind = where( (@p).chi2 != 0 );
	  if( length(pind)>0 && min(pdiff) < pdiffmin ){
	    useparams = _steppar_get_params( @p, pind[where_min( pdiff[pind] )] );
	    pdiffmin = min(pdiff);
	  }
	}
      }
      set_params( useparams );
      
      % SET & STEP
      set_par( par, val, 1, range[0], range[1] );
      () = (@fit)(__push_list(fitargs) ;; __qualifiers );
      
      % WRITE info
      stat = eval_stat_counts;
      foreach param (get_params){
	if(param.freeze==0 || param.name==par){
	  get_struct_field(info, escapedParameterName(param.name))[i] = param.value;
	}
      }
      info.chi2[i] = stat.statistic;
      info.chi2red[i] = stat.statistic/(stat.num_bins-stat.num_variable_params);
      
      if( chatty ){
	%print_struct(info; i=i);
	vmessage("INFO <%s>: Stepping %s (%d); restarts %3d/%3d: step %3d/%3d",
		 _function_name, parinfo.name, parinfo.index,
		 nrestarts, check, i+1, length(iorder)
		);
      }
      if( qualifier_exists("save") ){
	fits_write_binary_table( save, pidstr, info , keys );
      }
    }
    % In case no new best params are found during stepping make sure while loop
    % will be left
    if( length(iorder)==0 || i == iorder[-1] ){
      nrestarts = 2*(abs(check)+1);
    }
  }
  % Restore parameters
  set_params(originalparams);
  ()=eval_counts;
  Fit_Verbose=fitverbosity;
  
  if(qualifier_exists("keys"))
  {
    return info, keys;
  }
  else
    return info;
}
%}}}


%%%%%%%%%%%%%%%%%%%%%
define steppar_plot() %{{{
%%%%%%%%%%%%%%
%!%+
%\function{steppar_plot}
%\synopsis{xfig plot of the chi2 landscape of (a) stepped parameter(s)}
%\usage{steppar_plot( Struct_Type steppar, key );}
%\altusage{steppar_plot( String_Type stepparfile );}
%\qualifiers{
%\qualifier{norender}{If given xfig plots are returned instead of rendered}
%\qualifier{pdfunite}{If given the rendered plots are united into a single one}
%\qualifier{ignorez}{Ignores Chi2 values equal to Zero}
%\qualifier{size}{=[15,11] Size of the plot}
%\qualifier{path}{="steppar.pdf": Path for the rendered file}
%\qualifier{ext}{=".pdf": File Type, e.g., "png"}
%\qualifier{yoff}{Constant offset for the y-axis. Set to the initial statistic
%                   by default}
%\qualifier{yrange}{Range of the y-axis. Either Double_Type[1] or [2].
%                     If only one value is given it is taken as ymax! }
%\qualifier{dchi2}{=2.71: Delta Chisqr}
%}
%!%-
{
  variable sp, keys;
  switch(_NARGS)
  { case 1: sp = (); }
  { case 2: (sp, keys) = (); }
  { help(_function_name); return; }
  
  if( _typeof( sp ) == String_Type )
    ( sp, keys ) = steppar_load( sp ; keys );
  else ifnot( _typeof(sp)==Struct_Type ){
    vmessage("ERROR: <%s>: Argument has to be String or Struct_Type!",
	     _function_name); return;
  }

  variable norender = qualifier_exists("norender") ? 1 : 0;
  variable pdfunite = qualifier_exists("pdfunite") ? 1 : 0;
  variable ignorez  = qualifier_exists("ignorez") ? 1 : 0;
  variable size = qualifier("size",[15,11]);
  variable path = qualifier("path","steppar.pdf");
  variable ext  = qualifier("ext",path_extname(path));
  ext = ext == "" ? ".pdf" : ext;
  path = path_sans_extname(path);

  variable yoff = qualifier("yoff");
  variable yrange = [qualifier("yrange")];
  variable ymin = NULL, ymax = NULL;
  if( qualifier_exists("yrange") ){
    ymax = yrange[-1];
    if( length(yrange) > 1 ){
      ymin = yrange[0];
    }
  }

  variable dchi2 = qualifier("dchi2",2.71);
  
  variable N = length(sp);
  variable xf = Struct_Type[N];

  variable spconf = steppar_get_conf( sp; dchi2=dchi2 );
  
  variable x, y, chimin, xmin, xmax, dx, dy;
  variable wz;
  variable i, fn, I;
  _for i ( 0, N-1, 1 ){
    fn = escapedParameterName(keys[i].pname);
    x  = get_struct_field( sp[i], fn );
    y  = sp[i].chi2;

    wz = where( y < DOUBLE_MAX);
    if(ignorez){
      wz = where( 0 < y < DOUBLE_MAX );
    }

    x = x[wz];
    y = y[wz];

    
    xmin = min(x);
    xmax = max(x);
    dx = 0.0*(xmax-xmin);
    xmin -= dx;
    xmax += dx;

    if( ymin == NULL ){
      ymin = min(y);
    }
    if( ymax == NULL ){
      ymax = max(y);
    }
    
    dy = 0.05*(ymax-ymin);
    ymin -= dy;
    ymax += dy;
    
    chimin = min(y);
    
    ifnot(qualifier_exists("yoff")){
      yoff = keys[i].statistic;
    }
    
    xf[i] = xfig_plot_new( size[0], size[1] );
    xf[i].world( xmin,xmax , ymin-yoff,ymax-yoff ; padx=0, pady=0 );
    xf[i].ylabel(sprintf("$\\chi^2 - %.3f$",yoff));
    xf[i].xlabel( strreplace(keys[i].pname,"_","\\\_")+sprintf(" [%s]",keys[i].extname) );

    % Chi2
    xf[i].plot( x, y-yoff
		; depth=30, color="black", width=4 );
    % initial stat
    xf[i].plot( [xmin, xmax] , keys[i].statistic*[1,1]-yoff
		; depth=25, color="red");
    % initial par value
    xf[i].plot( keys[i].pval*[1,1], [ymin, ymax]-yoff
		; depth=25, color="red", width=2 );
    xf[i].xylabel( keys[i].pval, ymax-dy-yoff, 
		   sprintf("$p_\\mathrm{init}=%.3g$",keys[i].pval),
		   nint((xmax-keys[i].pval)/(xmax-xmin))==0 ? .5 : -.5,
		   .5
		   ; color="red",rotate=90
		 );
    
    % fill dchi2 area
    xf[i].plot( [xmin,xmax], chimin-yoff+[dchi2,dchi2]
		; depth=27, color="black" , width=2 );
    xf[i].xylabel( xmin+dx, chimin+dchi2-yoff
		   , sprintf("$\\Delta\\chi^2 = %.2f$",dchi2)
		   , -.5, -.5
		 ); 
    if( __is_initialized(&spconf) ){
      if( spconf[i].conf_min != NULL and spconf[i].conf_max != NULL ){
	xf[i].plot( spconf[i].conf_min*[1,1], [xmin,chimin+dchi2]-yoff
		    ; depth=27, color="black" );
	xf[i].plot( spconf[i].conf_max*[1,1], [xmin,chimin+dchi2]-yoff
		    ; depth=27, color="black" );
	I = where( y-chimin <= dchi2 );
	xf[i].shade_region( [spconf[i].conf_min, x[I], spconf[i].conf_max],
			    [chimin+dchi2, y[I], chimin+dchi2]-yoff
			    ; fillcolor="black");
      }
    }
    ifnot( norender ){
      %xf[i].render(path+sprintf("_%03d",keys[i].pidx)+ext);
      xf[i].render(path+sprintf("_%s",keys[i].extname)+ext);
    }
  }
  
  if(norender){
    return xf;
  }
  if(pdfunite and N>1 ){
    ()=system("pdfunite "+path+"_PID?????"+ext+" "+path+ext);
    ()=system("rm "+path+"_PID?????"+ext);
  }
}
%}}}



#endif

%
% Code from src/slang/structs/histograms/stretch_hist.sl
%
define stretch_hist(hist, scal){
%!%+
%\function{stretch_hist}
%\synopsis{stretch a histogram grid}
%\usage{Struct_Type hist = stretch_hist(hist, scal);}
%\description
%    Stretch the grid of a histogram structure with fields bin_lo, 
%    bin_hi, value, and err. The bin_lo and bin_hi fields are scaled 
%    by the factor scal. If other fields are present, those are 
%    preserved and passed on. Only presence of bin_lo and bin_hi are
%    checked for. If scal is an array with [a0,a1,a2,a3...], the grid
%    is stretched with a polynomial function of 
%       a0 + a1*bin_lo + a2*bin_lo^2 + ... .
%\seealso{add_hist, shift_hist, scale_hist}
%!%-   
   if(_NARGS==0) {  help(_function_name()); return; }

   variable ns = struct_copy(hist);
   ifnot(struct_field_exists(ns, "bin_lo") and 
		 struct_field_exists(ns, "bin_hi")){
	  help(_function_name()); return;
   }

   if(typeof(scal) == Integer_Type or typeof(scal)==Double_Type){
	 % ns.bin_lo *=scal;
	 % ns.bin_hi *=scal;
	 % return ns;
	  scal = [0, scal];
   }
   
   variable xl = scal[-1];
   variable xh = scal[-1];
   variable ii;
   _for ii (length(scal)-2, 0, -1){
	  xl = scal[ii] + xl * ns.bin_lo;
	  xh = scal[ii] + xh * ns.bin_hi;
   }
   ns.bin_lo = xl;
   ns.bin_hi = xh;
   return ns;
}

%
% Code from src/slang/date/strftime_MJD.sl
%
%%%%%%%%%%%%%%%%%%%
define strftime_MJD()
%%%%%%%%%%%%%%%%%%%
%!%+
%\function{strftime_MJD}
%\synopsis{formats an Modified Julian Date as a string}
%\usage{String_Type strftime_MJD([String_Type fmt,] Double_Type MJD)}
%\description
%    If \code{fmt} is not specified, \code{"%Y-%m-%d, %H:%M:%S"} is assumed.
%\seealso{strftime, gmtime, MJD2UNIXtime}
%!%-
{
  variable fmt="%Y-%m-%d, %H:%M:%S", MJD;
  switch(_NARGS)
  { case 1: MJD = (); }
  { case 2: (fmt, MJD) = (); }
  { help(_function_name()); return; }
  
  if(typeof(MJD)==Array_Type)
    return array_map(String_Type, &strftime, fmt, array_map(Struct_Type, &gmtime, MJD2UNIXtime(MJD)));

  
  return strftime(fmt,gmtime( MJD2UNIXtime(MJD)));
}

%
% Code from src/data/string2sys_err_array.sl
%
% =================== %
define string2sys_err_array(str)
% =================== %
%!%+
%\function{string2sys_err_array}
%\synopsis{create sys_err_array from str = sys_err_array2string}
%\usage{Array_Type sys_err = string2sys_err_array(String_Type str);}
%\description
%
%   This routines converts a string created by the function
%   "sys_err_array2string" back to an array, which can then be applied
%   to the data by the function set_sys_err_frac.
%   
%\seealso{fits_save_fit, set/get_sys_err_frac, sys_err_array2string}
%!%- 
{
   variable delim = qualifier("delim",",");   

   variable arr = Double_Type[0];

   variable tp = strtok(str,delim);

   variable t,sub,is,i_lo,i_hi,fac,del,i;
   foreach t(tp)
   {
      sub = strtok(t, ":");
      if (length(sub) != 2)
      {
	 vmessage("sys_err_array_to_string: String %s not compatible with the function!",str);
	 return NULL;
      }
      is = strtok(sub[0],"-");
      i_lo = integer(is[0]);
      i_hi = integer(is[1]);
      fac = atof(sub[1]);

      del = i_hi-i_lo+1;
      _for i(0,del-1,1)
      {
	 arr = [arr, fac];
      }
   }

   return arr;
}

%
% Code from src/slang/strings/string_intersection.sl
%
define string_intersection()
%!%+
%\function{string_intersection}
%\synopsis{Returns the common substrings within all strings of the given array}
%\usage{String_Type string_intersection(String_Type[] strings);}
%\qualifiers{
%    \qualifier{minlen}{restrict the returned substrings to a minimum length}
%    \qualifier{longest}{return the longest substring only}
%}
%\example
%    a = string_intersection(["mister", "twister"]; minlen = 3);
%    % -> ["ister", "ster", "ter", "ste", "iste", "ist"]
%!%-
{
  variable str;
  switch (_NARGS)
    { case 1: str = (); }
    { help(_function_name); return; }

  variable minlen = qualifier("minlen", 1);

  % find the shortest string in the given array
  variable lens = array_map(Integer_Type, &strlen, str);
  variable s = wherefirstmin(lens);
  % build up substring matrix of the shortest string
  variable n = lens[s]-minlen+1;
  variable match = String_Type[n*(n-1)/2+n];
  variable i, j, k = 0;
  _for i (0, n-1, 1) _for j (0, n-i-1, 1) {
    match[k] = substr(str[s], i+1, minlen+j);
    k++;
  }

  % loop over remaining strings
  str = array_remove(str, s);
  i = 0;
  while (i < length(str) && any(match != "")) {
    % remove non-substrings from the matrix (no an array)
    match = match[where(
      array_map(Integer_Type, &is_substr, str[i], match) > 0
    )];
    i++;
  }
  
  % reduce
  match = match[where(match != "")];
  match = match[unique(match)];
  if (qualifier_exists("longest")) {
    match = match[where_max(array_map(Integer_Type, &strlen, match))];
  }

  return match;
}

%
% Code from src/slang/strings/string_match_perl.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define string_match_perl()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{string_match_perl}
%\synopsis{matches a string against a RE and stores the results in $1, $2, ...}
%\usage{Integer_Type string_match(String_Type str, pat[, Integer_Type pos])}
%\description
%    The string \code{str} is matched against the regular expression \code{pat},
%    starting at the position \code{pos} (numbered from 1, which is the default).
%    The function returns the position of the start of the match in \code{str}.
%    The entire matching string is stored in the global variable $0,
%    the substrings of patterns enclosed by "\\\\( ... \\\\)" pairs
%    are stored in $1, $2, ..., $9  (as in Perl).
%
%    \code{string_match_perl} is deprecated. Use S-Lang's string_matches instead.
%\example
%    \code{if( string_match_perl("hello world", "\\\\([a-z]+\\\\) \\\\([a-z]+\\\\)"R) )}\n
%    \code{  ()=printf("1=%s, 2=%s\n", $1, $2);}
%
%    \code{%} Equivalent way using \code{string_matches}:\n
%    \code{variable m = string_matches("hello world", `\\([a-z]+\\) \\([a-z]+\\)`, 1);}\n
%    \code{if(m!=NULL)  vmessage("1=%s, 2=%s", m[1], m[2]);}
%\seealso{string_matches}
%!%-
{
  message("string_match_perl is deprecated. Use S-Lang's string_matches instead.");

  variable str, pat, pos=1, len;
  switch(_NARGS)
  { case 2: (str, pat) = (); }
  { case 3: (str, pat, pos) = (); }
  { return help(_function_name()); }

  variable str_m = string_match(str, pat, pos);
  if(str_m)
  {
    (pos, len) = string_match_nth(0);
    $0  = substr(str, pos+1, len);

    variable ok=1;
    try { (pos, len) = string_match_nth(1); } catch AnyError: { ok = 0; }
    $1 = substr(str, pos+1, len);

    ok=1;
    try { (pos, len) = string_match_nth(2); } catch AnyError: { ok = 0; }
    if(ok)  $2 = substr(str, pos+1, len);  else  return str_m;

    try { (pos, len) = string_match_nth(3); } catch AnyError: { ok = 0; }
    if(ok)  $3 = substr(str, pos+1, len);  else  return str_m;

    try { (pos, len) = string_match_nth(4); } catch AnyError: { ok = 0; }
    if(ok)  $4 = substr(str, pos+1, len);  else  return str_m;

    try { (pos, len) = string_match_nth(5); } catch AnyError: { ok = 0; }
    if(ok)  $5 = substr(str, pos+1, len);  else  return str_m;

    try { (pos, len) = string_match_nth(6); } catch AnyError: { ok = 0; }
    if(ok)  $6 = substr(str, pos+1, len);  else  return str_m;

    try { (pos, len) = string_match_nth(7); } catch AnyError: { ok = 0; }
    if(ok)  $7 = substr(str, pos+1, len);  else  return str_m;

    try { (pos, len) = string_match_nth(8); } catch AnyError: { ok = 0; }
    if(ok)  $8 = substr(str, pos+1, len);  else  return str_m;

    try { (pos, len) = string_match_nth(9); } catch AnyError: { ok = 0; }
    if(ok)  $9 = substr(str, pos+1, len);  else  return str_m;
  }

  return str_m;
}

%
% Code from src/slang/strings/strjoin_wrap.sl
%
define strjoin_wrap()
%!%+
%\function{strjoin_wrap}
%\synopsis{concatenates elements of an array and inserts linebreaks}
%\usage{Sting_Type strjoin_wrap(Array_Type x, String_Type delim);}
%\qualifiers{
%\qualifier{maxlen}{[=72]: maximum number of characters in a line}
%\qualifier{newline}{[="  "]: new line string}
%\qualifier{initial}{[=""]: initial delimiter}
%\qualifier{final}{[=""]: final delimiter}
%}
%\seealso{strjoin}
%!%-
{
  variable x, delim=", ";
  switch(_NARGS)
  { case 1: x = (); }
  { case 2: (x, delim) = (); }
  { help(_function_name()); return; }

  variable newline = qualifier("newline", "  ");
  variable maxlen = qualifier("maxlen", 72) - strlen(delim);
  variable result = "";
  variable i, buf = qualifier("initial", "");
  _for i (0, length(x)-2, 1)
  { variable xstr = string(x[i]);
    if(strlen(buf+xstr)>maxlen)
    { result += buf+"\n";
      buf = newline;
    }
    buf += xstr+delim;
  }
  if(strlen(buf+string(x[-1]))>maxlen)
  { result += buf+"\n";
    buf = newline;
  }
  return result + buf + string(x[-1]) + qualifier("final", "");
}

%
% Code from src/slang/strings/strreplace1.sl
%
define strreplace1()
%!%+
%\function{strreplace1}
%\synopsis{replaces one occurence of a substring in a string by another string}
%\usage{String_Type res = strreplace1(String_Type str, search_str, repl_str);}
%\description
%    replaces the first occurence of \code{search_str} in \code{str} by \code{repl_str}:\n
%    \code{(res, ) = strreplace(str, search_str, repl_str, 1);}
%!%-
{
  variable str, search_str, repl_str;
  switch(_NARGS)
  { case 3: (str, search_str, repl_str) = (); }
  { help(_function_name()); return; }

  variable res;
  (res, ) = strreplace(str, search_str, repl_str, 1);
  return res;
}

%
% Code from src/slang/strings/strreplaces.sl
%
define strreplaces() {
%!%+
%\function{strreplaces}
%\synopsis{Replace one or more substrings}
%\usage{String_Type strreplaces(String_Type a, String_Type[] b, String_Type[] c[, Integer_Type max_n]);}
%\description
%    This is a modificaiton to 'strreplace', where
%    multiple, different substrings (given as arrays)
%    are replaced at the same time.
%\seealso{strreplace}
%!%-
  variable a, b, c, n = NULL;
  switch (_NARGS)
    { case 3: (a,b,c) = (); }
    { case 4: (a,b,c,n) = (); }
    { help(_function_name); return; }
  
  if (typeof(b) != Array_Type) b = [b];
  if (typeof(c) != Array_Type) c = [c];
  if (length(b) != length(c)) { vmessage("error (%s): given arrays must be of equal length", _function_name); return; }
  variable i;
  _for i (0, length(b)-1, 1)
    if (n == NULL) a = strreplace(a, b[i], c[i]);
    else (a,) = strreplace(a, b[i], c[i], n);
  return a;
}

%
% Code from src/slang/strings/strsplit.sl
%
define strsplit()
%!%+
%\function{strsplit}
%\synopsis{split a string at a separator and return contents as an array}
%\usage{Array_Type strsplit(String_Type s, String_Type separator);}
%\description
%This routine takes a string s and splits it into parts separated by
%a separator character.
%\example
%variable ra="12:34:56.78";
%variable rastr,hh,mm,ss;
%rastr=strsplit(ra,":");
%hh=atof(rastr[0]); mm=atof(rastr[1]); ss=atof(rastr[2]);
%ra=hms2deg(hh,mm,ss);
%!%-
{
    variable s,separator;
    switch(_NARGS)
    { case 2: (s,separator) = (); }
    { return help(_function_name()); }

    if (length(separator)!=1) {
	throw RunTimeError,sprintf("%s: Separator must be exactly one character\n",_function_name());
    }
    variable retarr={};
    variable str=s;
    variable pos=is_substr(str,separator);
    while ( pos != 0 ) {
	list_append(retarr,substr(str,1,pos-1));
	str=substr(str,pos+1,strlen(str)-pos);
	pos=is_substr(str,separator);
    }
    list_append(retarr,str);

    return list_to_array(retarr);
}

%
% Code from src/slang/structs/struct_array.sl
%
define struct_array(n,s) {
%!%+
%\function{struct_array}
%\synopsis{creates an array of structures}
%\usage{Struct_Type[] struct_array(Integer_Type n, Struct_Type s);}
%\description
%    An array of n structures is created,
%    where each structure is a copy of the
%    given one s. 
%\example
%    variable sarr = struct_array(4, struct { firstname, lastname });
%    sarr[2].firstname = "Karl";
%    sarr[2].lastname = "Remeis";
%
%    print(sarr);
%    % will return
%    % {firstname=NULL, lastname=NULL}
%    % {firstname=NULL, lastname=NULL}
%    % {firstname="Karl", lastname="Remeis"}
%    % {firstname=NULL, lastname=NULL}
%\seealso{table_copy}
%!%-
  variable sn = Struct_Type[n];
  _for n (0, n-1, 1) sn[n] = COPY(s);
  return sn;
}

%
% Code from src/slang/arrays/struct_array_2_struct_of_arrays.sl
%

define struct_array_2_struct_of_arrays()
%!%+
%\function{struct_array_2_struct_of_arrays}
%\synopsis{Converts a Struct_Type[] to Struct_Type with array fields}
%\usage{Struct_Type struct_array_2_struct_of_arrays(Struct_Type[] SA);}
%\qualifiers{
%\qualifier{depth [0]:}{ If a field is a Struct_Type itself,
%                        the this function calls itself iteratively
%                        'depth' times going down the rabbit hole.}
%\qualifier{dim [0]:}{ If a field is multi-dimensional 'dim' is the
%                       dimension to which the array is appended to.}
%}
%\description
%   This function converts a Struct-Array 'Struct_Type[]' to
%   a single Struct_Type by converting the COMMON fields of
%   all structs into arrays.
%   The function works iteratively in case a field of the given
%   struct array SA itself is a Struct_Type. The function will
%   call itself maximaly 'depth' times.
%   In case a field is multi-dimensional the array will be appended
%   to the 'dim'-th dimension.
%
%   NOTE that this function runs orders faster than 'merge_struct_arrays'
%   while offering the same features as, e.g., reshaping multi-dimensional
%   arrays (the reshaping in this function also works if not every field
%   is multidimensional!).
%
%   The only drawback is that the function fails if an entry is NULL as
%   this case cannot be catched. In that case you may use
%   'merge_struct_arrays'!
%\examples
%
%   %%%%% SIMPLE example
%   variable sa = struct_array( 3, struct{ a=1, b=2 } );
%   sa[0] = struct_combine( struct{c}, sa[0] );
%   variable s = struct_array_2_struct_of_arrays( sa );
%   print(s);
%   {a=Integer_Type[3],
%    b=Integer_Type[3]}
%    
%   %%%%% COMPLEX example
%   variable sa = struct_array( 30000, struct{ a=1,
%                                              b=_reshape([1:120],[1,2,60]),
%                                              c=struct{ d="hello" }
%                                            }
%                             );
%   %%% only merge the parent structure
%   variable s = struct_array_2_struct_of_arrays( sa ; depth=0 );
%   isis>    print(s);
%   %  {a=Integer_Type[30000],
%   %   b=Integer_Type[30000,2,60],
%   %   c=Struct_Type[30000]}
%
%   %%% merge also the first generation     
%   variable s = struct_array_2_struct_of_arrays( sa ; depth=1 );
%   print(s);
%   %  {a=Integer_Type[30000],
%   %   b=Integer_Type[30000,2,60],
%   %   c=Struct_Type with 1 fields}
%   print(s.c);
%   %  {d=String_Type[3]}
%
%   %%% use the second dimension of mudlti-dimensional arrays for merging:
%   variable s = struct_array_2_struct_of_arrays( sa ; dim=1 );
%   print(s);
%   %  {a=Integer_Type[30000],
%   %   b=Integer_Type[1,60000,60],
%   %   c=Struct_Type[30000]}
%
%   %%% Runtime comparision to 'merge_struct_array'
%    tic;  s = struct_array_2_struct_of_arrays( sa );
%    vmessage(" sa2soa took %g seconds to call",toc);
%    print(s);
%    tic;  m = merge_struct_arrays( sa );
%    vmessage(" msa took %g seconds to call",toc);
%    print(m);
%      
%\seealso{merge_struct_array, array_struct_field, array_flatten, unique, intersection}
%!%-
{
  variable _sa;
  switch(_NARGS)
  { case 1: _sa = (); }
  { help(_function_name); }
  
  %%% if 'sa' is no array just return it
  ifnot( typeof (_sa) == Array_Type){
    return array_flatten( _sa[wherenot(_isnull(_sa))] );
  }

  %%% if length(sa) == 1 return sa[0]
  if( length(_sa) == 1 ){
    return _sa[0];
  }

  %%% remove NULL entries
  variable sa = _sa[wherenot(_isnull(_sa))];
  
  %%% QUALIFIERS
  variable depth = qualifier("depth",0);
  variable rdim  = qualifier("dim",0);
  
  %%% get all fieldnames
  variable fnames = array_map( Array_Type, &get_struct_field_names, sa );
  fnames = fnames[unique(fnames)];
  
  %%% find the common fields which are present in all Structs in the array
  variable cfnames = length(fnames)==1 ? fnames[0] : fnames[0][ intersection( __push_array(fnames) ) ];
  
  %%% Do the conversion to a Struct of array_s
  variable s;
  if( _typeof(sa) == Struct_Type ){
    s = @(_typeof(sa))(cfnames);
  }
  else{
    s = @_typeof(sa);
  }
  
  variable fn, fa, fadim, fadimarray;
  foreach fn(cfnames){
    fa = array_struct_field( sa, fn );
    if( _typeof(fa) == Array_Type ){
      fa = fa[wherenot(_isnull(fa))];
      fadim = array_shape(fa[0]);
      if( length(fadim) > 1 ){
	fadimarray = array_flatten( array_map( Array_Type, &array_shape, fa) );
	fadim[rdim] = nint(sum(fadimarray[[rdim:length(fadimarray)-1:length(fadim)]]));
	fa = array_flatten( fa ;; __qualifiers );
	reshape( fa, fadim );
      }
      else{
	fa = array_flatten( fa ;; __qualifiers );
      }
    }
    else if( _typeof(fa) == Struct_Type && depth > 0 ){
      fa = struct_array_2_struct_of_arrays( fa ; depth=depth-1 );
    }
    set_struct_field( s, fn, fa );
  }
  return s;
}

%
% Code from src/slang/structs/struct_field_or_default.sl
%
define struct_field_or_default(s,tag,def) {
%!%+
%\function{struct_field_or_default}
%\synopsis{return value of a structure field or a default value}
%\usage{value=struct_field_or_default(Struct_Type s, String_Type tag, default);}
%\description
%    return s.tag if it exists, otherwise default
%!%-
return struct_field_exists(s,tag) ? get_struct_field(s,tag) : def;
}

%
% Code from src/slang/structs/struct_fields_list_to_array.sl
%
define struct_fields_list_to_array()
%!%+
%\function{struct_fields_list_to_array}
%\synopsis{convert all fields of a structure from lists to arrays}
%\usage{struct_fields_list_to_array(Struct_Type s[, DataType_Type t]);}
%\seealso{list_to_array}
%!%-
{
  variable s, t=NULL;
  switch(_NARGS)
  { case 1: s = (); }
  { case 2: (s, t) = (); }
  { return help(_function_name()); }

  if(typeof(s)!=Struct_Type)
    return vmessage("error (%s): expecting Struct_Type, found %S", _function_name(), typeof(s));

  variable n;
  foreach n (get_struct_field_names(s))
  {
    variable v = get_struct_field(s, n);
    if(typeof(v)==List_Type)
      set_struct_field(s, n, list_to_array(t==NULL ? v : (v, t)));
  }
}

%
% Code from src/slang/arrays/struct_of_arrays_2_struct_array.sl
%
define struct_of_arrays_2_struct_array()
%!%+
%\function{struct_of_arrays_2_struct_array}
%\synopsis{Converts a Struct_Type with array fields to a Struct_Type[]}
%\usage{Struct_Type[] struct_of_arrays_2_struct_array(Struct_Type S);}
%\description
%   This function converts a 'Struct_Type' containing fields
%   which are arrays (all arrays must have the same dimension!)
%   to a Struct-array 'Struct_Type[]' with the same fields but
%   one dimensional. That means the returned Struct_Type[] has
%   then the dimension (array_shape) of the fields.
%   The returned Struct-Array will contain all fields the original
%   'S' also had, BUT for those fields with a different
%   array_shape then the first field and the value will be set to the
%   first entry of its field!
%\example
%   variable s = struct{ a=[1:3], b=[4:6], c=[3,20] };
%   variable sa = struct_of_arrays_2_struct_array( s );
%   print(sa);
%   {a=1, b=4, c=3}
%   {a=2, b=5, c=3}
%   {a=3, b=6, c=3}
%\seealso{array_map}
%!%-
{
  variable s;
  switch(_NARGS)
  { case 1: s = (); }
  { help(_function_name); }

  %%% Get the field names
  variable fnames = get_struct_field_names( s );
  variable fields = get_struct_fields( s, fnames );

  %%% Create a Struct-Array with the dimensions of the first field
  variable type = typeof(s);
  variable n = length(fields[0]);

  variable sa;
  if( type == Struct_Type ){
    sa = struct_array( n, @type(fnames) );
  }
  else{
    sa = @type[n];
  }
  
  variable i;
  _for i ( 0, length(fnames)-1 ){
    array_map( Void_Type, &set_struct_field, sa, fnames[i],  fields[i] );
  }
  return sa;
}

%
% Code from src/misc/math/struct_print_field_statistics.sl
%
define struct_print_field_statistics()
%!%+
%\function{struct_print_field_statistics}
%\usage{struct_print_field_statistics(Struct_Type s);}
%\seealso{moment}
%!%-
{
  variable s;
  switch(_NARGS)
  { case 1: s = (); }
  { help(_function_name()); return; }

  variable fieldnames = get_struct_field_names(s);
  variable fmt = sprintf("%%%ds: ",  max(array_map(Integer_Type, &strlen, fieldnames)));
  variable fieldname;
  foreach fieldname (fieldnames)
  {
    ()=printf(fmt, fieldname);
    variable a = get_struct_field(s, fieldname);
    variable dims, type;  (dims, , type) = array_info(a);
    foreach $1 (dims) { ()=printf("%d x ", $1); }
    ()=printf("%S", type);
    if( __is_datatype_numeric(type) )
    { variable m = moment(a);
      ()=printf(" [min=%g, av=%g, sdev=%f, max= %g]", m.min, m.ave, m.sdev, m.max);
    }
    ()=printf("\n");
  }
}

%
% Code from src/slang/arrays/struct_to_assoc_array.sl
%
define struct_to_assoc_array()
%!%+
%\function{struct_to_assoc_array}
%\synopsis{converts a structure to an associative array}
%\usage{Assoc_Type struct_to_assoc_array(Struct_Type structure[, DataType_Type data-type]);}
%\description
%    The given structure is converted into an associative
%    array. Thereby, the keys of the returned array will
%    be equal to the field names of the structure.
%
%    If the second argument is provided, the values of the
%    associative array will be of that type. Otherwise,
%    the type of the first field of the structure will be
%    used. Note, that all fields have to be of the same
%    data-type!
%\example
%    c = struct_to_assoc_array(struct {
%      pi = 3.141592653589793,
%      c = 2.99792458e10,
%      h = 6.62606876e-27
%    });
%    foreach key (c) using ("keys")
%      vmessage("c[%s] = %e", key, c[key]); ;
%\seealso{Assoc_Type, assoc_get_keys, get_struct_field_names}
%!%-
{
  variable s, type;
  switch(_NARGS)
    { case 1: (s) = (); }
    { case 2: (s, type) = (); }
    { help(_function_name()); return; }

  if (typeof(s) != Struct_Type) { vmessage("error (%s): input is not of Struct_Type", _function_name); return; }

  % get names of struct fields and initialize array
  variable fields = get_struct_field_names(s);
  ifnot (__is_initialized(&type)) type = typeof(get_struct_field(s, fields[0]));
  else if (typeof(type) != DataType_Type) { vmessage("error (%s): given type is not of DataType_Type", _function_name); return; }
  variable a = Assoc_Type[type];

  % loop over fields and add values to array
  variable field;
  foreach field (fields) a[field] = get_struct_field(s, field);
  
  return a;
}

%
% Code from src/data/lightcurve/sum_lc.sl
%
%%%%%%%%%%%%%
define sum_lc()
%%%%%%%%%%%%%
%!%+
%\function{sum_lc}
%\synopsis{combines light curves from one structure}
%\usage{Struct_Type sum_lc(Struct_Type lc, Array_Type inds)}
%\qualifiers{
%\qualifier{time}{name of the time field (default: \code{"time"})}
%\qualifier{rate}{format string for the rate field (default: \code{"rate%S"})}
%\qualifier{error}{format string for the error field (default: \code{"error%S"})}
%}
%\description
%    The rates in the fields of \code{lc} (named by \code{inds} 
%    via the \code{rate} format statement) are summed up.
%    The corresponding errors (named by the \code{error} format statement)
%    are added in quadrature.
%    The returned structure has the fields \code{time, rate, error}.
%\examples
%    % 1. Assume you have
%      lc = struct { time, rate1, error1, rate2, error2, rate3, error3 };
%    % Then
%      slc = sum_lc(lc, [1:3]);
%    % is equivalent to
%      slc = struct {
%        time = lc.time,
%        rate = lc.rate1 + lc.rate2 + lc.rate3,
%        error = sqrt( lc.error1^2 + lc.error2^2 + lc.error3^2 )
%      };
%
%    % 2. Assume you have
%      lc = struct { time, rate_a, err_a, rate_b, err_b, rate_c, err_c };
%    % Then
%      slc = sum_lc(lc, ["a", "b", "c"]; rate="rate_%s", error="err_%s");
%    % is equivalent to
%      slc = struct {
%        time = lc.time,
%        rate = lc.rate_a + lc.rate_b + lc.rate_c,
%        error = sqrt( lc.err_a^2 + lc.err_b^2 + lc.err_c^2 )
%      };
%!%-
{
  variable lc, ind, inds;
  switch(_NARGS)
  { case 2: (lc, inds) = (); }
  { help(_function_name()); return; }

  variable fmtrate = qualifier("rate", "rate%S");
  variable fmterr = qualifier("error", "error%S");
  variable slc = struct {
    time = get_struct_field(lc, qualifier("time", "time")),
    rate = Double_Type[length(lc.time)],
    error = Double_Type[length(lc.time)]
  };
  foreach ind (inds)
  {
    slc.rate += get_struct_field(lc, sprintf(fmtrate, ind));
    slc.error += get_struct_field(lc, sprintf(fmterr, ind))^2;
  }
  slc.error = sqrt( slc.error );

  return slc;
}

%
% Code from src/slang/astronomy/sunpos.sl
%
define sunpos () 
%!%+
%\function{sunpos}
%\synopsis{computes the RA and Dec of the Sun at a given date.}
%\usage{(Double_Type eq, Double_Type ecl) = sunpos(Double_Type JD);}
%\description
%   Computes the position of the Sun at the date JD (or dates, if JD
%   is an array), given in Julian Date. Position calculation takes
%   into account pertubations from the planets, aberration and
%   nutation.  Output value 'eq' is a structure in equatorial
%   coordinates: eq = struct{ra, dec}, 'ecl' is in ecliptic plane
%   coordinates: ecl = struct{lon, obl} (Longitude and true
%   obliquity).
%   
%   Note: Function adopted from the IDL function with the same name.
%   See there for accuracy tests, which should also apply here.
%\qualifiers{
%\qualifier{MJD}{give time in MJD instead of JD}
%\qualifier{radian}{return all variables as radians rather than degrees}
%\qualifier{hour}{return RA in hours rather than degrees (not possible when radian is set).}
%}
%\example
%   (eq, ecl) = sunpos(2455055.5) ;
%   ()=printf("Right ascension: %.4f deg, Declination: %.4f deg\n", eq.ra, eq.dec) ;
%!%-
{
   variable jd   ;

   switch(_NARGS) 
     { case 1: (jd) = () ; }
     { help(_function_name()); return; }

    if (qualifier_exists("mjd") or qualifier_exists("MJD")) jd+=2400000.5  ; % translate MJD to JD
    
    variable dtor = PI/180.0  ; %       ;(degrees to radian, double precision)
  
  % ;  form time in Julian centuries from 1900.0
   variable t = (jd - 2415020.0)/36525.0 ;
  
  % ;  form sun's mean longitude
   variable l = (279.696678+((36000.768925*t) mod 360.0))*3600.0 ;

% ;  allow for ellipticity of the orbit (equation of centre)
% ;  using the Earth's mean anomaly ME
   variable me = 358.475844 + ((35999.049750*t) mod 360.0) ;
   variable ellcor  = (6910.1 - 17.2*t)*sin(me*dtor) + 72.3*sin(2.0*me*dtor) ;
     
   l = l + ellcor ;

% ; allow for the Venus perturbations using the mean anomaly of Venus MV
  variable mv = 212.603219 + ((58517.803875*t) mod 360.0) ; 
   
  variable vencorr = 4.8 * cos((299.1017 + mv - me)*dtor) + 
          5.5 * cos((148.3133 +  2.0 * mv  -  2.0 * me )*dtor) + 
          2.5 * cos((315.9433 +  2.0 * mv  -  3.0 * me )*dtor) + 
          1.6 * cos((345.2533 +  3.0 * mv  -  4.0 * me )*dtor) + 
          1.0 * cos((318.15   +  3.0 * mv  -  5.0 * me )*dtor) ;

   l = l + vencorr ;

% ;  Allow for the Mars perturbations using the mean anomaly of Mars MM
   variable mm = 319.529425  +  (( 19139.858500 * t)  mod  360.0 ) ;
 
   variable marscorr = 2.0 * cos((343.8883 -  2.0 * mm  +  2.0 * me)*dtor ) + 
            1.8 * cos((200.4017 -  2.0 * mm  + me) * dtor) ;
 
   l = l + marscorr ;

% ; Allow for the Jupiter perturbations using the mean anomaly of
% ; Jupiter MJ
   variable mj = 225.328328  +  (( 3034.6920239 * t)  mod  360.0 ) ;
 
   variable jupcorr = 7.2 * cos(( 179.5317 - mj + me )*dtor) + 
          2.6 * cos((263.2167  -  mj ) *dtor) + 
          2.7 * cos(( 87.1450  -  2.0 * mj  +  2.0 * me ) *dtor) + 
          1.6 * cos((109.4933  -  2.0 * mj  +  me ) *dtor) ;
 
   l = l + jupcorr ;

% ; Allow for the Moons perturbations using the mean elongation of
% ; the Moon from the Sun D
   variable d = 350.7376814  + (( 445267.11422 * t)  mod  360.0 ) ;
   variable mooncorr  = 6.5 * sin(d*dtor) ;
  
   l = l + mooncorr ;

% ; Allow for long period terms
   variable longterm  = + 6.4 * sin(( 231.19  +  20.20 * t )*dtor) ;
   l  =    l + longterm ;
   l  =  ( l + 2592000.0)  mod  1296000.0 ;
   variable longmed = l/3600.0 ;

% ; Allow for Aberration
   l  =  l - 20.5 ;

% ; Allow for Nutation using the longitude of the Moons mean node OMEGA
   variable omega = 259.183275 - (( 1934.142008 * t ) mod 360.0 ) ;
   l  =  l - 17.2 * sin(omega*dtor) ;

% ; Form the True Obliquity
   variable oblt  = 23.452294 - 0.0130125*t + (9.2*cos(omega*dtor))/3600.0 ;

% ; Form Right Ascension and Declination
   l = l/3600.0 ;

   %forcing RA to be an array, so that indexing works always
   variable ra  = [atan2( (sin(l*dtor) * cos(oblt*dtor)) , (cos(l*dtor)) )] ;
   
   variable neg = where(ra < 0.0) ;
   
   if (length(neg)) ra[neg] += 2.0*PI ;

   variable dec = asin(sin(l*dtor) * sin(oblt*dtor)) ;
 
   if(qualifier_exists("radian"))
     {
	oblt *= dtor ; longmed *= dtor ;
     }
   else 
     {
	if(qualifier_exists("hour")) ra /= 15. ; %transfer RA degrees to hours, only if radian is not set
	ra /= dtor ;  dec /= dtor ;
     }

   %returing RA to a single double value if array has only one entry.
   if (length(ra) == 1) { return struct{ra=ra[0], dec=dec[0]}, struct{lon=longmed[0], obl=oblt[0]} ; }
   else { return struct{ra=ra, dec=dec}, struct{lon=longmed, obl=oblt} ;} 
}

%
% Code from src/fitting/parameters/switch_tbnew_rel_abund.sl
%

% only activate if module xspec could be loaded
#ifeval __get_reference("load_xspec_local_models")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define switch_tbnew_rel_abund()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{switch_tbnew_rel_abund}
%\usage{switch_tbnew_rel_abund([String_Type parnames]);}
%\description
%    Unless one or more \code{parnames} are given, all elements are switched.
%\qualifiers{
%\qualifier{inst}{[=1] instance of tbnew}
%\qualifier{verbose}{}
%}
%!%-
{
  variable parnames;
  switch(_NARGS)
  { case 0: parnames = "tbnew(" + string(qualifier("inst", 1)) + ")."
                       + ["He", "C", "N", "O", "Ne", "Na", "Mg", "Al", "Si", "S", "Cl", "Ar", "Ca", "Cr", "Fe", "Co", "Ni"]; }
  { case 1: parnames = (); }
  { help(_function_name()); return; }

  parnames = [parnames];
  variable parname_parts = strtok(parnames[0], ".");
  variable NH = get_par(parname_parts[0] + ".nH");
  if(NH>0)  NH *= -1e22;

  if(qualifier_exists("verbose"))
  {
    message("% before switching:");
    list_Par(parnames);
  }
  variable parname;
  foreach parname (parnames)
  { variable parinfo = get_par_info(parname);
    $1 = NH*xspec_elabund(strtok(parname, ".")[1]);
    if(parinfo.value<0)  $1 = 1./$1;
    set_par(parname, parinfo.value*$1, parinfo.freeze, parinfo.max*$1, parinfo.min*$1);
  }
  if(qualifier_exists("verbose"))
  {
    message("\n% after switching:");
    list_Par(parnames);
  }
}


#endif

%
% Code from src/data/sys_err_array2string.sl
%
define sys_err_array2string(arr)
%!%+
%\function{sys_err_array2string}
%\synopsis{creates a string from an Array returned by "get_sys_err_frac"}
%\usage{String_Type str = sys_err_array2string(Array_Type sys_err);}
%\description
%
%    This routine creates a String from the Array of systematic errors
%    apllied to the data. Such an array is returned by the function
%    "get_sys_err_frac".
%   
%\seealso{fits_save_fit, set/get_sys_err_frac, string2sys_err_array}
%!%- 
{
   variable delim = "";
   variable str_sys = "";
   
   variable i_beg = 0;
   variable i_end = 0;
   
   variable n = length(arr), i;
   if(n) {
   _for i(0,n-1,1)
   {

      
      if (i > 0)
      {
	 if (abs(arr[i] - arr[i-1]) > 1e-8)
	 {
	    str_sys += delim+string(i_beg)+"-"+string(i_end)+":"+string(arr[i_beg]);
	    delim = qualifier("delim",",");   
	    i_beg = i; i_end=i;
	 }
	 else
	 {
	    i_end++;
	 }
	 
      }
   }
   % write the last line, too
   str_sys += delim+string(i_beg)+"-"+string(i_end)+":"+string(arr[i_beg]);   
   } else {
    % if length(arr) is zero, there are no systematic errors defined. This
    % string should result in zero systematic errors when run through
    % string2sys_err_array() and set_sys_err_frac().
    str_sys = "0-0:0";   
   }
   return str_sys;
}

%
% Code from src/slang/system/system_message.sl
%
define message_system()
%!%+
%\function{message_system}
%\usage{Integer_Type message_system(String_Type cmd)}
%\description
%    \code{message(cmd);}\n
%    \code{system(cmd);}
%!%-
{
  variable cmd;
  switch(_NARGS)
  { case 1: cmd = (); }
  { help(_function_name()); return; }

  message(cmd);
  return system(cmd);
}

%
% Code from src/slang/structs/tables/table_add_row.sl
%
define table_add_row()
%!%+
%\function{table_add_row}
%\usage{table_add_row(table, ...);}
%!%-
{
  if(_NARGS<2) { help(_function_name()); return; }

  variable values = __pop_list(_NARGS-1);
  variable t = ();

  variable fieldnames, i, n = length(values);
  switch(typeof(t))
  { case Struct_Type:
      fieldnames = get_struct_field_names(t);
      if(length(fieldnames) == n)
        _for i (0, n-1, 1)
          set_struct_field(t, fieldnames[i], [get_struct_field(t, fieldnames[i]), values[i] ]);
      else
        vmessage("error (%s): The table is a structure of %d arrays, but %d values have been given.", _function_name(), length(fieldnames), n);
  }
  { case Array_Type:
      vmessage("warning (%s): The array cannot be modified, pass a reference to it (&) instead.", _function_name());
  }
  { case Ref_Type:
      if(typeof(@t)!=Array_Type)
      { vmessage("warning (%s): Table has to be a reference to an array (of structures).", _function_name());
        return;
      }
      variable s = (@t)[0];
      if(typeof(s) == Struct_Type)
      { s = @s;
        fieldnames = get_struct_field_names(s);
        if(length(fieldnames) == n)
        { _for i (0, n-1, 1) { set_struct_field(s, fieldnames[i], values[i]); }
          @t = [@t, s];
        }
        else
        { vmessage("error (%s): The table is an array of structures with %d fields, but %d values have been given.", _function_name(), length(fieldnames), n); }
      }
      else
      { if(n==1)
          @t = [@t, values[0]];
        else
          vmessage("error (%s): The table is just an array, but %d values have been given.", _function_name(), n);
      }
  }
  { % else:
      vmessage("error (%s): The table is neither a structure of arrays nor an array of structures, but a %S.", _function_name(), typeof(t));
  }
}

%
% Code from src/slang/structs/tables/struct-arrays/table_copy.sl
%
% -*- mode: slang; mode: fold -*-

define COPY() %{{{
%!%+
%\function{COPY}
%\synopsis{makes a copy of a nested Struct/Array/List_Type}
%   \usage{Struct_Type copy = COPY(Struct_Type copyme);}
%\altusage{Array_Type  copy = COPY(Array_Type  copyme);}
%\altusage{List_Type   copy = COPY(List_Type   copyme);}
%\altusage{Assoc_Type  copy = COPY(Assoc_Type   copyme);}
%\description
%    \code{COPY} returns a properly dereferenced copy of an
%    arbitrarily nested \code{Struct/Array or List_Type} with
%    all its entries. 
%    Depending on the Data_Type of the initial given
%    copyme and the Data_Type of its entries, \code{COPY}
%    iteratively calls the  according sub-function
%    (\code{struct_copy}, \code{array_copy} or \code{list_copy}).
%    If the given Data_Type or one of its entries is not
%    one of previously mentioned ones, \code{COPY} returns
%    @(Data_Type) or respectively @(entry).
%    
%    NOTE: * Ref_Type entries will not be dereferenced, which
%            allows to copy XFIG-OBJECTS !!!
%          * Double_Type[] is an Array_Type
%
%\example
%    Array_Type:
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9];
%     copy = COPY(s); copy[0].a[[0]] = ["modified"];
%     print(s[0].a);
%
%    List_Type:
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9];
%     l = [{ array_copy(s), 1., [1:10] }, {"abs"}];
%     copy = COPY(l); copy[0][0][0].a[[0]] = ["modified"];
%     print(l[0][0][0].a);
%
%    Struct_Type:
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9]; S = struct{ f=s };     
%     C = struct_copy(S); C.f[0].a[[0]] = ["modified"];
%     print(S.f[0].a);
%
%\seealso{struct_copy, array_copy, list_copy, assoc_copy}
%!%-
{
  variable copyme;
  switch(_NARGS)
  { case 1: copyme = (); }
  { help(_function_name()); return; }

  variable type = typeof(copyme);
  if( type  == Struct_Type )
    return struct_copy(copyme);
  else if( type == Array_Type )
    return array_copy(copyme);
  else if( type == List_Type )
    return array_copy(copyme);
  else if( type == Assoc_Type )
    return assoc_copy(copyme);
  else{
    vmessage("WARNING: <%s>: Unsupported Data_Type=%S, retruning @copyme",
	     _function_name,type);
    return @copyme;
  }
}
%}}}

define struct_copy() %{{{
%!%+
%\function{struct_copy}
%\synopsis{makes a copy of a struct with all its fields}
%\usage{Struct_Type copy = struct_copy(Struct_Type struct);}
%\description
%    \code{struct_copy} copies a \code{Struct_Type} with all its
%    fields, which can be of any Type! If a field is an Array_Type[]
%    \code{struct_copy} calls \code{array_copy} and if a field is
%    \code{Struct_Type}, \code{struct_copy} calls itself.
%    Further it copies Ref_Type, which allows copying XFIG-Objects.
%\example
%    s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%    s[0].a[[0]]=[0:9]; S = struct{ f=s };     
%    C = struct_copy(S); C.f[0].a[[0]] = ["modified"];
%    print(S.f[0].a);
%\seealso{array_copy, list_copy, COPY, assoc_copy}
%!%-
{
  variable table;
  switch(_NARGS)
  { case 1: table = (); }
  { help(_function_name()); return; }
  
  variable f, fieldnames = get_struct_field_names(table);
  variable field , ftype;
  variable copy = @Struct_Type(fieldnames);
  foreach f (fieldnames){
    field = get_struct_field(table, f);
    ftype = typeof(field);
    if( ftype == Struct_Type )
      set_struct_field(copy, f, struct_copy(field) );
    else if( ftype == Array_Type )
      set_struct_field(copy, f, array_copy(field) );
    else if( ftype == List_Type )
      set_struct_field(copy, f, list_copy(field) );
    else if( ftype == Assoc_Type )
      set_struct_field(copy, f, assoc_copy(field) );
    else if ( (ftype == String_Type) ||  (ftype == Ref_Type) )
      set_struct_field(copy, f, field );	
    else
      set_struct_field(copy, f, @field );
  }
  return copy;
}
%}}}
define array_copy() %{{{
%!%+
%\function{array_copy}
%\synopsis{makes a copy of a nested array}
%\usage{Array_Type copy = array_copy(Array_Type[] array);}
%\description
%    \code{array_copy} copies an \code{Array_Type[]} with all its
%    entries, which can be of any Type! If an entry is a Struct_Type
%    \code{array_copy} calls \code{struct_copy} and if an entrie is
%    an \code{Array_Type[]}, \code{array_copy} calls itself.
%\example
%    s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%    s[0].a[[0]]=[0:9];
%    copy = array_copy(s); copy[0].a[[0]] = ["modified"];
%    print(s[0].a);
%\seealso{struct_copy}
%!%-
{
  variable array;
  switch(_NARGS)
  { case 1: array = (); }
  { help(_function_name()); return; }
  
  variable type = typeof(array);
  ifnot( type == Array_Type ){
    if( type == Struct_Type )
      return struct_copy( array );
    if( type == List_Type )
      return list_copy( array );
    else if( type == Ref_Type || type == String_Type )
      return array;
    else
      return @array;
  }
  
  variable len  = length(array);
  variable dim  = array_shape(array);
  type = _typeof(array);
  variable copy = type[len];
  reshape( copy, dim );
  
  variable i;
  _for i ( 0, len-1, 1 ){
    copy[[i]] = array_copy( array[[i]][0] );
  }
  return copy;
}
%}}}

define list_copy() %{{{
%!%+
%\function{list_copy}
%\synopsis{makes a copy of a nested list}
%\usage{List_Type copy = list_copy(List_Type list);}
%\description
%    \code{list_copy} returns a properly dereferenced copy of an
%    arbitrarily nested \code{List_Type} with all its entries. 
%    Depending on the Data_Type of the entries, \code{list_copy}
%    iteratively calls the according sub-function
%    (\code{struct_copy}, \code{array_copy} or \code{list_copy}).
%    If the Data_Type of one of its entries is not
%    one of privouse mentioned ones, \code{list_copy} returns
%    @(Data_Type) or respectively @(entry).
%\example
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9];
%     l = [{ array_copy(s), 1., [1:10] }, {"abs"}];
%     copy = COPY(l); copy[0][0][0].a[[0]] = ["modified"];
%     print(l[0][0][0].a);
%\seealso{COPY, struct_copy, array_copy, assoc_copy}
%!%-
{
  variable list;
  switch(_NARGS)
  { case 1: list = (); }
  { help(_function_name()); return; }
  
  variable type = typeof(list);
  ifnot( type == List_Type ){
    if( type == Struct_Type )
      return struct_copy( list );
    else if( type == Array_Type )
      return array_copy( list );
    else if( type == Assoc_Type )
      return assoc_copy( list );
    else if( type == Ref_Type || type == String_Type )
      return list;
    else
      return @list;
  }
  
  variable len  = length(list);
  variable copy = {};
  
  variable i;
  _for i ( 0, len-1, 1 ){
    list_append( copy, list_copy(list[i]) );
  }
  return copy;
}
%}}}

define table_copy() %{{{
%!%+
%\function{table_copy}
%\synopsis{makes a copy of all columns of a table}
%\usage{Struct_Type copy = table_copy(Struct_Type table);}
%\description
%    \code{copy.field = @(table.field);  %} for every \code{field}
%    NOTE: This also works for XFIG-PLOT-OBJECTS
%!%-
{
  variable table;
  switch(_NARGS)
  { case 1: table = (); }
  { help(_function_name()); return; }
  
  vmessage("<%s> is deprecated! Use <struct_copy> instead!",_function_name);
  return struct_copy(table);
}
%}}}

define assoc_copy()
%!%+
%\function{assoc_copy}
%\synopsis{makes a copy of an associated array}
%\usage{Assoc_Type copy = assoc_copy(Assoc_Type list);}
%\description
%    \code{assoc_copy} returns a properly dereferenced copy of an
%    arbitrarily nested \code{Assoc_Type} with all its entries. 
%    Depending on the Data_Type of the entries, \code{assoc_copy}
%    iteratively calls the according sub-function
%    (\code{struct_copy}, \code{array_copy} or \code{list_copy}).
%    If the Data_Type of one of its entries is not
%    one of previously mentioned ones, \code{assoc_copy} returns
%    @(Data_Type) or respectively @(entry).
%\example
%     s = Struct_Type[1]; s[0]=struct{ a=Array_Type[1,2] };
%     s[0].a[[0]]=[0:9];
%     l = [{ array_copy(s), 1., [1:10] }, {"abs"}];
%     copy = COPY(l); copy[0][0][0].a[[0]] = ["modified"];
%     print(l[0][0][0].a);
%\seealso{COPY, struct_copy, array_copy, list_copy}
%!%-
{
    variable copyme;
    switch(_NARGS)
    { case 1: copyme = (); }
    { help(_function_name()); return; }

    variable new=Assoc_Type[];
    
    variable s=assoc_get_values(copyme);
    variable key;
    foreach key ( s ) {
        new[key]=COPY(s[key]);
    }
    return new;
}




%
% Code from src/xspec/table_models.sl
%
% -*- mode: slang; mode: fold; -*- %

private define tableModelInvalid (self) %{{{
{
  throw UsageError, "Table model object has become invalid!";
}
%}}}

private define tableModelReset (table) %{{{
{
  table.close();
  table.set_function = &tableModelInvalid;
  table.set_additional = &tableModelInvalid;
  table.set_grid = &tableModelInvalid;
  table.add_parameter = &tableModelInvalid;
  table.add_additional = &tableModelInvalid;
  table.numbvals = NULL;
  table.parameter_list = {};
  table.additional_parameter_list = {};

  % remove is safe, we have created it deliberately
  remove(table.filename);
}
%}}}

private define tableModelParameterList (list, name, values) %{{{
{
  variable param = struct {
    name = name,
    values = NULL,
    method = qualifier_exists("log"),
    unit = qualifier("unit", ""),
    minimum = qualifier("minimum", 0.),
    bottom = qualifier("minimum", 0.),
    top = qualifier("top", 1.),
    maximum = qualifier("maximum", 1.),
    initial = qualifier("inital", 0.),
    delta = qualifier("delta", 1e-3),
  };
  
  if (NULL != values) {
    param.values=  values[array_sort(values)];
    param.minimum= qualifier("minimum", min(values));
    param.bottom=  qualifier("bottom", min(values));
    param.top=     qualifier("top", max(values));
    param.maximum= qualifier("maximum", max(values));
    param.initial= qualifier("initial", mean(values));
    param.delta=   qualifier("delta", (max(values)-min(values))*1e-3);
  };

  list_append(@list, param);
}
%}}}

private define tableModelUniqueParameter (table, name) %{{{
{
  variable p;
  foreach p (table.parameter_list)
    if (p.name == name)
      throw UsageError, sprintf("Parameter with name '%s' already exists", name);
  foreach p (table.additional_parameter_list)
    if (p.name == name)
      throw UsageError, sprintf("Parameter with name '%s' already exists", name);
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define tableModelAddParameter (self, name, values)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{table_model.add_parameter}
%\synopsis{Add interpolation parameter to table model}
%\usage{table_model.add_parameter(String_Type name, Double_Type[] values);}
%#c%{{{
%\qualifiers{
%  \qualifier{log}{if given, parameter is interpolated logarithmically rather than linearly}
%  \qualifier{unit}{[=""] Set the unit of the parameter}
%  \qualifier{minimum}{[=min(values)] Set hard lower limit}
%  \qualifier{bottom}{[=min(values)] Set minimum parameter range}
%  \qualifier{top}{[=max(values)] Set maximum parameter range}
%  \qualifier{maximum}{[=max(values)] Set hard upper limit}
%  \qualifier{initial}{[=mean(values)] Set initial parameter value}
%  \qualifier{delta}{[=(max(values)-min(values))*1e-2] Set parameter step hint}
%}
%\description
%  Each call to this function adds a new parameter with the given name
%  and given interpolation points for this parameter. The table model gets
%  interpolated for this given points.
%
%\seealso{table_model.add_additional}
%!%-
{
  tableModelUniqueParameter(self, name);
  tableModelParameterList(&(self.parameter_list), name, values;; __qualifiers);
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define tableModelAddAdditionalParameter (self, name)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{table_model.add_additional}
%\synopsis{Add additional contribution to interpolated spectra}
%\usage{table_model.add_additional(String_Type name);}
%#c%{{{
%\qualifiers{
%  \qualifier{unit}{[=""] Set the unit of the parameter}
%  \qualifier{minimum}{[=min(values)] Set hard lower limit}
%  \qualifier{bottom}{[=min(values)] Set minimum parameter range}
%  \qualifier{top}{[=max(values)] Set maximum parameter range}
%  \qualifier{maximum}{[=max(values)] Set hard upper limit}
%  \qualifier{initial}{[=mean(values)] Set initial parameter value}
%  \qualifier{delta}{[=(max(values)-min(values))*1e-2] Set parameter step hint}
%}
%\description
%  Add additional contributions to the interpolated spectra. These parameters
%  are simply a normalization for the contributions, e.g., 0 means there is
%  no contribution, 1 means there is full contribution. For each of this
%  contributions an additional spectrum is added to the interpolated model.
%
%\seealso{table_model.add_parameter}
%!%-
{
  tableModelUniqueParameter(self, name);
  tableModelParameterList(&(self.additional_parameter_list), name, NULL;; __qualifiers);
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define tableModelSetEnergyGrid (self, bin_lo, bin_hi)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{table_model.set_grid}
%\synopsis{Set the energy grid for the table model}
%\usage{table_model.set_grid(bin_lo, bin_hi);}
%#c%{{{
%\qualifiers{
%  \qualifier{lolimit}{[=0] The model value if lower than lowest bin}
%  \qualifier{hilimit}{[=0] The model value if higher than highest bin}
%}
%\description
%  Set the energy grid on which the model gets interpolated.
%  Boundaries of the bins have to match.
%
%\seealso{table_model.set_function, table_model.set_additional}
%!%-
{
  self.lolimit = 1.*qualifier("lolimit", 0.);
  self.hilimit = 1.*qualifier("hilimit", 0.);

  if (any(array_sort(bin_lo) != [0:length(bin_lo)-1])
      || any(array_sort(bin_hi) != [0:length(bin_hi)-1]))
    throw UsageError, "Bin arrays not in ascending order";
  if (length(bin_lo) != length(bin_hi))
    throw UsageError, "Bin arrays have unequal length";
  if (any(bin_lo[[1:]] != bin_hi[[:-2]]))
    throw UsageError, "Bin boundaries have to match";
  if (any(bin_lo<=0))
    throw UsageError, "Energy grid can not have negative values";

  self.energy_grid = [bin_lo, bin_hi[-1]];
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define tableModelSetFunction (self, function)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{table_model.set_function}
%\synopsis{Set interpolation model function}
%\usage{table_model.set_function(Ref_Type function);}
%#c%{{{
%\description
%  Set the table model function that gets stored in the table
%  on the defined energy grid. The function must be of the form
%  \code{define modelfunction (bin_lo, bin_hi, params)}
%  where bin_lo and bin_hi are the bin boundary arrays and
%  params is the parameter array set by table_model.add_parameter.
%  The function may use qualifiers.
%
%\seealso{table_model.set_additional}
%!%-
{
  self.function = function;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define tableModelSetAdditionalFunction (self, function)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{table_model.set_additional}
%\synopsis{Set interpolation model contributions}
%\usage{table_model.set_additional(Ref_Type function);}
%#c%{{{
%\description
%  Set the table model contributions function that gets stored
%  in the table on the defined energy grid. The function must
%  be of the form
%  \code{define modelfunction (bin_lo, bin_hi, params, k)}
%  where bin_lo and bin_hi are the bin boundary arrays and
%  params is the parameter array set by table_model.add_parameter.
%  The additional argument k is set to the number of the additional
%  parameter. So k=1 for the first additonal parameter, k=2 for the
%  second and so on.
%
%  The function is expected to calculate the k-th contribution with
%  the parameter for this contribution set to 1.
%
%\seealso{table_model.set_additional}
%!%-
{
  self.additional = function;
}
%}}}

private define tableModelWriteHeader (table) %{{{
{
  if (_fits_movabs_hdu(table.fitsfile, 1)) {
    tableModelReset(table);
    throw IOError, "Unable to write to header";
  }

  variable name = table.model_name;
  variable unit = qualifier("unit", "photons/cm^2/s");

  variable header = struct {
    HDUCLASS= "OGIP",
    HDUCLAS1= "XSPEC TABLE MODEL",
    HDUVERS= "1.1.0",
    MODLNAME= name,
    MODLUNIT= unit,
    REDSHIFT= not not qualifier_exists("redshift"),
    ADDMODEL= not qualifier_exists("mult"),
    LOELIMIT= (table.use_double) ? qualifier("loLimit", 0.0) : 1.f*qualifier("loLimit", 0.0),
    HIELIMIT= (table.use_double) ? qualifier("hiLimit", 0.0) : 1.f*qualifier("hiLimit", 0.0)
  };

  variable k;
  foreach k (get_struct_field_names(header))
    fits_update_key(table.fitsfile, k, get_struct_field(header, k));

  fits_update_key(table.fitsfile, "LOELIMIT", table.lolimit, "Model value outside of energy grid (lower)");
  fits_update_key(table.fitsfile, "HIELIMIT", table.hilimit, "Model value outside of energy grid (higher)");
}
%}}}

private define tableModelWriteParameterHeader (table) %{{{
{
  variable parameters_header = struct {
    HDUCLASS= "OGIP",
    HDUCLAS1= "XSPEC TABLE MODEL",
    HDUCLAS2= "PARAMETERS",
    HDUVERS= "1.0.0",
    NINTPARM= length(table.parameter_list),
    NADDPARM= length(table.additional_parameter_list)
  };

  variable k;
  variable max_unitlen = 1;
  table.numbvals = Int_Type[length(table.parameter_list)];
  _for k (0, length(table.parameter_list)-1) {
    max_unitlen = max([max_unitlen, strlen(table.parameter_list[k].unit)]);
    table.numbvals[k] = length(table.parameter_list[k].values);
  }
  _for k (0, length(table.additional_parameter_list)-1)
    max_unitlen = max([max_unitlen, strlen(table.additional_parameter_list[k].unit)]);
  variable max_parlen = max(table.numbvals);

  variable colnames = ["NAME",
		       "METHOD",
		       "INITIAL",
		       "DELTA",
		       "MINIMUM",
		       "BOTTOM",
		       "TOP",
		       "MAXIMUM",
		       "NUMBVALS",
		       "VALUE",
		       "UNIT"];
  variable colform = table.use_double ? ["12A","J","D","D","D","D","D","D","J",
					 sprintf("%dE", max_parlen),
					 sprintf("%dA", max_unitlen)]
                                      : ["12A","J","E","E","E","E","E","E","J",
					 sprintf("%dE", max_parlen),
					 sprintf("%dA", max_unitlen)];
  variable colcoms = String_Type[length(colform)];

  fits_create_binary_table(table.fitsfile, "PARAMETERS", 0, colnames, colform, colcoms);

  foreach k (get_struct_field_names(parameters_header))
    fits_update_key(table.fitsfile, k, get_struct_field(parameters_header, k));
}
%}}}

private define tableModelWriteParameterValues (table) %{{{
{
  variable k;
  variable aType = (table.use_double) ? Double_Type : Float_Type;
  variable allValues = aType[length(table.parameter_list), max(table.numbvals)];
  _for k (0, length(table.parameter_list)-1)
    allValues[k, [:table.numbvals[k]-1]] = table.parameter_list[k].values;
  if (_fits_write_col(table.fitsfile, 10, 1, 1, allValues)) { % values at 10
    tableModelReset(table);
    throw IOError, "Unable to write paramter values";
  }
}
%}}}

private define tableModelWriteParameterMethods (table) %{{{
{
  variable k;
  variable allValues = Int_Type[length(table.parameter_list)];
  _for k (0, length(table.parameter_list)-1)
    allValues[k] = table.parameter_list[k].method;
  if (_fits_write_col(table.fitsfile, 2, 1, 1, allValues)) { % methods at 2
    tableModelReset(table);
    throw IOError, "Unable to write method values";
  }
}
%}}}

private define tableModelWriteParameterNames (table) %{{{
{
  variable k;
  variable allValues = String_Type[length(table.parameter_list)+length(table.additional_parameter_list)];
  _for k (0, length(table.parameter_list)-1)
    allValues[k] = table.parameter_list[k].name;
  _for k (0, length(table.additional_parameter_list)-1)
    allValues[length(table.parameter_list)+k] = table.additional_parameter_list[k].name;
  if (_fits_write_col(table.fitsfile, 1, 1, 1, allValues)) { % names at 1
    tableModelReset(table);
    throw IOError, "Unable to write parameter names";
  }
}
%}}}

private define tableModelWriteParameterUnits (table) %{{{
{
  variable k;
  variable allValues = String_Type[length(table.parameter_list)+length(table.additional_parameter_list)];
  _for k (0, length(table.parameter_list)-1)
    allValues[k] = table.parameter_list[k].unit;
  _for k (0, length(table.additional_parameter_list)-1)
    allValues[length(table.parameter_list)+k] = table.additional_parameter_list[k].unit;
  if (_fits_write_col(table.fitsfile, 11, 1, 1, allValues)) { % units at 11
    tableModelReset(table);
    throw IOError, "Unable to write parameter units";
  }
}
%}}}

private define tableModelWriteParameterInitials (table) %{{{
{
  variable aType = table.use_double ? Double_Type : Float_Type;
  variable k;

  variable npar=       length(table.parameter_list);
  variable nadd=       length(table.additional_parameter_list);
  variable total_pars= npar+nadd;
  variable initV=      aType[total_pars];
  variable deltaV=     aType[total_pars];
  variable minimumV=   aType[total_pars];
  variable bottomV=    aType[total_pars];
  variable topV=       aType[total_pars];
  variable maximumV=   aType[total_pars];

  _for k (0, npar-1) {
    initV[k]=    table.parameter_list[k].initial;
    deltaV[k]=   table.parameter_list[k].delta;
    minimumV[k]= table.parameter_list[k].minimum;
    bottomV[k]=  table.parameter_list[k].bottom;
    topV[k]=     table.parameter_list[k].top;
    maximumV[k]= table.parameter_list[k].maximum;
  }
  _for k (0, nadd-1) {
    initV[k+npar]=    table.additional_parameter_list[k].initial;
    deltaV[k+npar]=   table.additional_parameter_list[k].delta;
    minimumV[k+npar]= table.additional_parameter_list[k].minimum;
    bottomV[k+npar]=  table.additional_parameter_list[k].bottom;
    topV[k+npar]=     table.additional_parameter_list[k].top;
    maximumV[k+npar]= table.additional_parameter_list[k].maximum;
  }
  if (_fits_write_col(table.fitsfile, 3, 1, 1, initV)
      || _fits_write_col(table.fitsfile, 4, 1, 1, deltaV)
      || _fits_write_col(table.fitsfile, 5, 1, 1, minimumV)
      || _fits_write_col(table.fitsfile, 6, 1, 1, bottomV)
      || _fits_write_col(table.fitsfile, 7, 1, 1, topV)
      || _fits_write_col(table.fitsfile, 8, 1, 1, maximumV)) {
    tableModelReset(table);
    throw IOError, "Unable to write paramter vector";
  }
}
%}}}

private define tableModelWriteParameter (table) %{{{
{
  tableModelWriteParameterHeader(table);
  tableModelWriteParameterValues(table);
  tableModelWriteParameterMethods(table);
  tableModelWriteParameterNames(table);
  tableModelWriteParameterUnits(table);
  tableModelWriteParameterInitials(table);
  if (_fits_write_col(table.fitsfile, 9, 1, 1, table.numbvals)) { % numbvals at 9
    tableModelReset(table);
    throw IOError, "Unable to write number of values";
  }
}
%}}}

private define tableModelWriteGrid (table) %{{{
{
  fits_create_binary_table(table.fitsfile, "ENERGIES", 0,
    ["ENERG_LO", "ENERG_HI"], table.use_double ? ["D", "D"] : ["E", "E"], ["keV", "keV"]);

  variable energies_header = struct {
    HDUCLASS="OGIP",
    HDUCLAS1="XSPEC TABLE MODEL",
    HDUCLAS2="ENERGIES",
    HDUVERS="1.0.0",
  };

  variable k;
  foreach k (get_struct_field_names(energies_header))
    fits_update_key(table.fitsfile, k, get_struct_field(energies_header, k));

  if (_fits_write_col(table.fitsfile, 1, 1, 1, table.energy_grid[[:-2]])
      || _fits_write_col(table.fitsfile, 2, 1, 1, table.energy_grid[[1:]])) {
    tableModelReset(table);
    throw IOError, "Unable to write energy grid";
  }
}
%}}}

private define iterateParameterVector (par_set, par_has) %{{{
{
  % reduce the par_set vector one by one, using par_has
  % as reference. So we step as [2,2] -> [2,1] -> [1,2] -> [1,1]
  % return 1 if the iterating vector is to be use, else 0
  % This is used to exit the while loop
  ifnot (length(par_set))
    return 0;

  variable p = length(par_set)-1;
  forever {
    par_set[p]--;
    if (par_set[p])
      break;
    p--;
    if (p<length(par_set)-1)
      par_set[p+1] = par_has[p+1];
    if (p<0)
      return 0;
  }

  return 1;
}
%}}}

private define tableModelWriteSpectraHeader (table) %{{{
{
  variable npar=    length(table.parameter_list);
  variable nadd=    length(table.additional_parameter_list);
  variable npoints= length(table.energy_grid)-1;
  variable k;

  variable colnams = ["PARAMVAL",
		      "INTPSPEC",
		      array_map(String_Type, &sprintf, "ADDSP%03d", [1:nadd])];
  variable colform = table.use_double ? [sprintf("%dD", npar),
					 sprintf("%dD", npoints),
					 array_map(String_Type, &sprintf, "%dD", Int_Type[nadd]+npoints)]
                                      : [sprintf("%dE", npar),
					 sprintf("%dE", npoints),
					 array_map(String_Type, &sprintf, "%dE", Int_Type[nadd]+npoints)];
  variable colcoms = String_Type[length(colform)];

  fits_create_binary_table(table.fitsfile, "SPECTRA", 0, colnams, colform, colcoms);

  variable spectra_header = struct {
    HDUCLASS="OGIP",
    HDUCLAS1="XSPEC TABLE MODEL",
    HDUCLAS2="MODEL SPECTRA",
    HDUVERS="1.0.0"
  };

  foreach k (get_struct_field_names(spectra_header))
    fits_update_key(table.fitsfile, k, get_struct_field(spectra_header, k));
}
%}}}

private define tableModelWriteSpectra (table) %{{{
{
  tableModelWriteSpectraHeader(table);

  variable aType = table.use_double ? Double_Type : Float_Type;

  variable npar=    length(table.parameter_list);
  variable nadd=    length(table.additional_parameter_list);
  variable npoints= length(table.energy_grid)-1;
  variable nval=    table.numbvals[[:-2]]; % we iterate over the last parameter

  variable par_set= @nval; % parameter vector, used as iterator
  variable specs=   aType[table.numbvals[-1], npoints];
  variable parmat=  aType[table.numbvals[-1], npar];
  variable addspec= aType[table.numbvals[-1], npoints];
  variable pV=      aType[npar];

  variable i,k;
  
  variable firstRow= 1;

  % use par_set as iterator
  do {
    _for i (0, npar-2)
      pV[i] = table.parameter_list[i].values[nval[i]-par_set[i]];
    _for i (0, table.numbvals[-1]-1) {
      pV[-1] = table.parameter_list[-1].values[i];
      parmat[i,*] = pV;
      specs[i,*] = @(table.function)(table.energy_grid[[:-2]], table.energy_grid[[1:]],
				     pV;; __qualifiers);

      if (_fits_write_col(table.fitsfile, 1, firstRow, 1, parmat)
	  || _fits_write_col(table.fitsfile, 2, firstRow, 1, specs)) {
	tableModelReset(table);
	throw IOError, "Unable to write table values";
      }
    }
    _for k (0, nadd-1) {
      _for i (0, table.numbvals[-1]-1) {
	addspec[i,*] = @(table.additional)(table.energy_grid[[:-2]],
					   table.energy_grid[[1:]],
					   parmat[i,*],
					   k;; __qualifiers);
      }

      if (_fits_write_col(table.fitsfile, k+3, firstRow, 1, addspec)) {
	tableModelReset(table);
	throw IOError, "Unable to write table values";
      }
    }
    firstRow += table.numbvals[-1];
  } while (iterateParameterVector(par_set, nval));
}
%}}}

private define tableModelSummary (table) %{{{
{
  variable msg = `
Generating table model '%s':
  Parameter:            %d
  Additional Parameter: %d
  Energy grid bins:     %d

  Lowest energy:        %lg keV
  Highest energy:       %lg keV
  Model limit low:      %lg
  Model limit high:     %lg

 > %s`;
  vmessage(msg,
	   table.model_name,
	   length(table.parameter_list),
	   length(table.additional_parameter_list),
	   length(table.energy_grid)-1,
	   min(table.energy_grid),
	   max(table.energy_grid),
	   table.lolimit,
	   table.hilimit,
	   table.filename);
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define tableModelWriteTable (self)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{table_model.write}
%\synopsis{Write table model to file}
%\usage{table_model.write();}
%#c%{{{
%\qualifiers{
%  \qualifier{redshift}{if given, set redshift parameter flag}
%  \qualifier{mult}{if given, model is multiplicative (default additive)}
%  \qualifier{verbose}{show additional information}
%}
%\description
%  Given the model function and additional contributions, this
%  function evaluates the model and contributions on the given
%  parameter values and writes the resulting spectra to the
%  table file. Any qualifier passed to this function is passed
%  on to the model functions.
%
%  IMPORTANT: This function calls the evaluation functions
%  several times and may take a long time to finish. Depending
%  on the evaluation time of the model functions.
%
%\seealso{table_model.close}
%!%-
{
  if (length(self.parameter_list) && NULL == self.function)
    throw UsageError, "No model function given";
  if (length(self.additional_parameter_list) && NULL == self.additional)
    throw UsageError, "No additional function given";
  if (NULL == self.energy_grid)
    throw UsageError, "Energy grid not set";

  if (qualifier_exists("verbose"))
    tableModelSummary(self);

  tableModelWriteParameter(self);
  tableModelWriteGrid(self);
  tableModelWriteSpectra(self;; __qualifiers);
  tableModelWriteHeader(self;; __qualifiers);
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define tableModelClose (self)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{table_model.close}
%\synopsis{Finish table model}
%\usage{table_model.close();}
%#c%{{{
%\description
%  Write the checksums to each HDU and close the
%  table model file.
%!%-
{
  variable nhdus = fits_get_num_hdus(self.fitsfile);
  variable k;
  _for k (1, nhdus) {
    if (_fits_movabs_hdu(self.fitsfile, k)) {
      tableModelReset(self);
      throw IOError, "Unable to write checksums";
    }
    fits_write_chksum(self.fitsfile);
  }

  fits_close_file(self.fitsfile);
}
%}}}

private define tableModelOpenFile (table) %{{{
{
  variable stat = stat_file(table.filename);
  if ( (NULL != stat && stat_is("reg", stat.st_mode))
     && not qualifier_exists("overwrite"))
    throw IOError, "File exists";

  table.fitsfile = fits_open_file(table.filename, "c");

  if (NULL == table.fitsfile)
    throw IOError, "Unable to creat fits file";
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define new_table_model (filename)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{new_table_model}
%\synopsis{Initialize object and file to create table model}
%\usage{Struct_Type model = new_table_model(String_Type filename)}
%#c%{{{
%\qualifiers{
%  \qualifier{name}{[=<filename-no-ext>] set model name}
%  \qualifier{double}{if set, stores all values as doubles instead of floats}
%  \qualifier{overwrite}{if set, ignores existing file with same name}
%}
%\description
%  Giving a file name the function creates a fits file with the given
%  name and returns a handle that eases the process of creating a
%  table model according to the XSPEC format
%     OGIP Memo 92-009 (XSPEC Table Model File Format)
%
%  To create a file one has to give the interpolation points of the
%  model, the additional contributions and the model spectra.
%
%\example
%  variable table = new_table_model("test_model.fits");
%
%  % create "parameter1" with given interpolation points
%  table.add_parameter("parameter1", [0:2:0.1]);
%  table.add_parameter("parameter2", [-1:1:0.001]); % create "parameter2"
%  % create additional parameter "additional"
%  table.add_additional("additional");
%
%  table.set_grid(bin_lo, bin_hi); % define the valid energy grid for the model
%  table.set_function(&my_fancy_model); % set the model function
%  % set additional contributions
%  % (only necessary if additional parameters are given)
%  table.set_additional(&my_fancy_contributions);
%
%  table.write(); % fill table
%  table.close(); % finish table
%\seealso{table_model.add_parameter, table_model.add_additional,
%  table_model.set_grid, table_model.set_function, table_model.set_additional,
%  table_model.write, table_model.close}
%!%-
{
  variable table = struct {
    filename = path_realpath(filename),
    model_name = qualifier("name", path_basename_sans_extname(filename)),
    fitsfile = NULL,

    add_parameter = &tableModelAddParameter,
    add_additional = &tableModelAddAdditionalParameter,
    set_grid = &tableModelSetEnergyGrid,
    set_function = &tableModelSetFunction,
    set_additional = &tableModelSetAdditionalFunction,
    write = &tableModelWriteTable,
    close = &tableModelClose,

    function = NULL,
    additional = NULL,
    energy_grid = NULL,
    numbvals = NULL,
    use_double = qualifier_exists("double"),
    lolimit = 0.,
    hilimit = 0.,
    parameter_list = {},
    additional_parameter_list = {}
  };

  tableModelOpenFile(table;;__qualifiers);

  return table;
}
%}}}

%
% Code from src/slang/structs/tables/table_print_TeX.sl
%
define table_print_TeX()
%!%+
%\function{table_print_TeX}
%\synopsis{prints certain columns of a structure as a TeX table}
%\usage{table_print_TeX(Struct_Type tab[, String_Type cols[]]);}
%\description
%    If no columns are specified by the string-array \code{cols},
%    all columns of the table tab are used.
%\qualifiers{
%\qualifier{align}{alignment of all columns: "l", "c", "r" [default: "l"]}
%\qualifier{heading}{[=\code{cols}]: headings for TeX table}
%\qualifier{output}{[=stdout]: output can be written in a file by stating
%                       a File_Type or the filename (String_Type)}
%}
%!%-
{
  variable tab, cols;
  switch(_NARGS)
  { case 1: tab = (); cols = get_struct_field_names(tab); }
  { case 2: (tab, cols) = (); }
  { help(_function_name()); return; }

  variable x, y, nx = length(cols);
  variable heading = qualifier("heading", cols);
  variable out = qualifier("output");
  out = ( typeof(out) == File_Type ? out  : ( typeof(out)==String_Type ? fopen(out,"w") : stdout));
  
  ()=fprintf(out,"\\begin{tabular}{");
  loop (nx)  ()=fprintf(out,"%s", qualifier("align", "l"));
  ()=fprintf(out,"}\n");

  ()=fprintf(out," \\hline\n \\hline\n");
  ()=fprintf(out," %s", heading[0]);
  _for x (1, nx-1, 1)  ()=fprintf(out," & %s", heading[x]);
  ()=fprintf(out," \\\\\n \\hline\n");

  _for y (0, length(get_struct_field(tab, cols[0]))-1, 1)
  { ()=fprintf(out," %S", get_struct_field(tab, cols[0])[y]);
    _for x (1, nx-1, 1)  ()=fprintf(out," & %S", get_struct_field(tab, cols[x])[y]);
    ()=fprintf(out," \\\\\n");
  }
  ()=fprintf(out," \\hline\n");
  ()=fprintf(out,"\\end{tabular}\n");
  
  if (out != stdout) ()=fclose (out);
}

%
% Code from src/slang/date/tai2tt.sl
%
define tai2tt()
%!%+
%\function{tai2tt}
%\synopsis{Convert a (M)JD in International Atomic Time into Terrestrial Time (TT, also known as TDT or ET)}
%\usage{tt=tai2tt(JD)}
%\description
%  This routine converts a Julian Date that is given in International Atomic
%  Time (TAI) into Terrestrial Time, which is roughly corresponding to
%  Terrestrial Dynamic Time or Ephemeris Time. This is done by
%  adding 32.184s to the TT. This routine is array safe.
%
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the TT-TAI in seconds, rather than the corrected (M)JD.}
%}
%\seealso{utc2tai,tai2utc,tt2tai,tt2tdb}
%!%-
{
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    variable diff=32.184;

    if (qualifier_exists("deltat")) {
	if (typeof(JD)==Array_Type) {
	    return diff+Double_Type[length(JD)];
	}
	return diff;
    }
    return JD+diff/86400.0;
}

%
% Code from src/slang/math/taylor.sl
%
define taylor()
%!%+
%\function{taylor}
%\synopsis{Returns the taylor series for given coefficients}
%\usage{Double_Type taylor(Double_Type[] x, coefficients)}
%\description
%    Returns the taylor series around x with the given
%    'coefficients', which lengths determines the order.
%    The sum is evaluated using the Horner (1819) schema
%    for numerical accuracy and computation speed:
%
%    c0 + c1*x + 1/2!*c2*x^2 + 1/3!*c3*x^3 + ...
%    = (((... + c3)*x/3 + c2)*x/2 + c1)*x + c0
%\example
%    x = [0, 1];
%    c = [1, 4e-3, 0, 3e-6]; % up to third order
%
%    taylor(x, c); % returns [1.0, 1.0040005]
%!%-
{
  variable x,c;
  switch (_NARGS)
    { case 2: (x,c) = (); }
    { help(_function_name); return; }

  c = [c];
  variable n = length(c);

  % we use the Horner schema here to
  % achieve a higher numerical precission
  % and faster computation (Horner, 1819)

  variable y = typeof(x) == Array_Type ? Double_Type[length(x)] : 0;
  variable i;
  _for i (n-1, 1, -1) {
    y = (y + c[i])*x/i;
  }
  y += c[0];

  return y;
}


%%%%%%%%%%%%%%%%%%%%%%%%
define taylorcoeff_from_struct()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{taylorcoeff_from_struct}
%\synopsis{extracts taylor coefficients from the field names of a structure}
%\usage{Double_Type[] taylorcoeff_from_struct(Struct_Type struct[,
%                    String_Type 0th_pattern, String_Type Nth_pattern]);}
%\description
%    If a structure has fields which names reflect the
%    different orders of a taylor series, this function
%    extracts these coefficients by matching the field
%    names against two patterns.
%
%    '0th_pattern' is the field name of the zero order
%    and 'Nth_pattern' this of all higher order terms.
%    Both patterns have to be regular expressions. In
%    particular, the expression of the Nth order must
%    contain a number extraction two determine the order
%    N of that field.
%
%
%    By default, '0th_pattern' is set to "[a-zA-Z]0" and
%    'Nth_pattern' to "[a-zA-Z]\\([0-9]*\\)dot"R. If the
%    latter _pattern matches, but the extracted number
%    is an empty string N=1 is assumed.
%
%    The returned array contains the taylor coefficients
%    in ascending order as used by the 'taylor'-function.
%\seealso{taylor, string_matches}
%!%-
{
  variable s, pat0 = "[a-zA-Z]0", patN = "[a-zA-Z]\([0-9]*\)dot"R;
  switch(_NARGS)
    { case 1: (s) = (); }
    { case 3: (s,pat0,patN) = (); }
    { help(_function_name()); return; }

  % extract and loop over struct's fields
  variable fields = get_struct_field_names(s);
  variable coeff = Double_Type[0];
  variable i;
  _for i (0, length(fields)-1, 1) {
    % check if field matches 0th order
    if (string_matches(fields[i], pat0) != NULL) {
      if (length(coeff) == 0) { coeff = Double_Type[1]; }
      coeff[0] = get_struct_field(s, fields[i]);
    }
    % check if field matches Nth order
    else {
      variable match = string_matches(fields[i], patN, 1);
      if (match != NULL) {
        % get the N and eventually increase coeff-array
        variable n = match[1] == "" ? 1 : atoi(match[1]);
        if (length(coeff) <= n) { coeff = [coeff, Double_Type[n-length(coeff)+1]]; }
        coeff[n] = get_struct_field(s, fields[i]);
      }
    }
  }

  return coeff;
}

%
% Code from src/fitting/fit-functions/tbnew_metal.sl
%
#ifeval is_defined("tbnew") || is_defined("TBnew")

define tbnew_metal_fit(lo, hi, par)
{                     %  H,      He,       C,       N        O,      Ne,      Na,      Mg,      Al,      Si,       S,      Cl,      Ar,      Ca,      Cr,      Fe,      Co,      Ni,
  variable tbnew_par1 = [1, 9.77e-2,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0];
  variable tbnew_par2 = [0,       0, 2.40e-4, 7.59e-5, 4.90e-4, 8.71e-5, 1.45e-6, 2.51e-5, 2.14e-6, 1.86e-5, 1.23e-5, 1.32e-7, 2.57e-6, 1.58e-6, 3.24e-7, 2.69e-5, 8.32e-8, 1.12e-6];  % wilm
%      			          0, 3.63e-4, 1.12e-4, 8.51e-4, 1.23e-4, 2.14e-6, 3.80e-5, 2.95e-6, 3.55e-5, 1.62e-5, 3.16e-7, 3.63e-6, 2.29e-6, 4.68e-7, 4.68e-5, 8.32e-8, 1.78e-6];  % angr
  variable other_tbnew_par = [0.2, 1, 0.025, 0.25, 3.5, 1, 1, 0.5, 1, 0.6, 1, 0.25, 0.2, 0.02, 0.1, 0.6, 0.5, 1, 0.003, 0.03, 0.3, 0.05, 0.04, 0];
  variable par1 = [-par[0]*tbnew_par1, other_tbnew_par];
  variable par2 = [-par[1]*tbnew_par2, other_tbnew_par];
#ifeval is_defined("tbnew")
  return  tbnew_fit(lo, hi, par1) * tbnew_fit(lo, hi, par2);
#else
  return  TBnew_fit(lo, hi, par1) * TBnew_fit(lo, hi, par2);
#endif
}
add_slang_function("tbnew_metal", ["NH_HHe", "NH_met"]);

private define tbnew_metal_defaults(i)
{
  return (1e22, 0, 0, 1e24);
}
set_param_default_hook("tbnew_metal", &tbnew_metal_defaults);

#endif

%
% Code from src/fitting/fit-functions/tbnew_simple.sl
%
#ifeval is_defined("tbnew") || is_defined("TBnew_fit")
define tbnew_simple_fit(lo, hi, par)
{
  variable abs_NH = abs(par[0]);
  variable exp_minus_tau =
#ifeval is_defined("tbnew")
    tbnew_fit
#else
    TBnew_fit
#endif
    (lo, hi, [
    abs_NH, % nH
         1, % He
         1, % C
         1, % N
         1, % O
         1, % Ne
         1, % Na
         1, % Mg
         1, % Al
         1, % Si
         1, % S
         1, % Cl
         1, % Ar
         1, % Ca
         1, % Cr
         1, % Fe
         1, % Co
         1, % Ni
       0.2, % H2
         1, % rho
     0.025, % amin
      0.25, % amax
       3.5, % PL
         1, % H_dep
         1, % He_dep
       0.5, % C_dep
         1, % N_dep
       0.6, % O_dep
         1, % Ne_dep
      0.25, % Na_dep
       0.2, % Mg_dep
      0.02, % Al_dep
       0.1, % Si_dep
       0.6, % S_dep
       0.5, % Cl_dep
         1, % Ar_dep
     0.003, % Ca_dep
      0.03, % Cr_dep
       0.3, % Fe_dep
      0.05, % Co_dep
      0.04, % Ni_dep
         0  % redshift
  ]);
  return (par[0]>=0 ? exp_minus_tau : 1./exp_minus_tau);
}
add_slang_function("tbnew_simple", ["nH [10^22/cm^2]"]);

private define tbnew_simple_defaults(i)
{
  return (1, 0, 0, 100);
}
set_param_default_hook("tbnew_simple", &tbnew_simple_defaults);

#endif

%
% Code from src/fitting/fit-functions/tbnew_simple_z.sl
%

#ifeval is_defined("tbnew") || is_defined("TBnew")

define tbnew_simple_z_fit(lo, hi, par)
{
  variable abs_NH = abs(par[0]);
   variable z = par[1];
  variable exp_minus_tau =
#ifeval is_defined("tbnew")
    tbnew_fit
#else
    TBnew_fit
#endif
    (lo, hi, [
    abs_NH, % nH
         1, % He
         1, % C
         1, % N
         1, % O
         1, % Ne
         1, % Na
         1, % Mg
         1, % Al
         1, % Si
         1, % S
         1, % Cl
         1, % Ar
         1, % Ca
         1, % Cr
         1, % Fe
         1, % Co
         1, % Ni
       0.2, % H2
         1, % rho
     0.025, % amin
      0.25, % amax
       3.5, % PL
         1, % H_dep
         1, % He_dep
       0.5, % C_dep
         1, % N_dep
       0.6, % O_dep
         1, % Ne_dep
      0.25, % Na_dep
       0.2, % Mg_dep
      0.02, % Al_dep
       0.1, % Si_dep
       0.6, % S_dep
       0.5, % Cl_dep
         1, % Ar_dep
     0.003, % Ca_dep
      0.03, % Cr_dep
       0.3, % Fe_dep
      0.05, % Co_dep
      0.04, % Ni_dep
         z  % redshift
  ]);
  return (par[0]>=0 ? exp_minus_tau : 1./exp_minus_tau);
}
add_slang_function("tbnew_simple_z", ["nH [10^22/cm^2]","z [redshift]"]);

private define tbnew_simple_z_defaults(i)
{
   switch(i)
   {case 0:  return (1, 0, 0, 100); }
   {case 1:  return (0, 0, 0, 10); }   
}
set_param_default_hook("tbnew_simple_z", &tbnew_simple_z_defaults);

#endif

%
% Code from src/misc/tcp.sl
%

% only activate if module socket could be loaded
#ifeval __get_reference("socket")!=NULL

% global variable for the server.
% only one server allowed, i.e., NULL = no server running
private variable _TCP_SERVER = NULL;
private variable _TCP_DEFAULT_PORT = 1149;
private variable _TCP_BUF_SIZE = 252; % maximum string length of SLang
private variable _TCP_CODE = struct {
  disconnect = "D", greet = "G", object = "O"
};

private define ccode(code) { return sprintf("%c%s", 2, code); }
private define dcode(code) {
  if (strlen(code) == 2 && code[0] == 2) { return char(code[1]); } else { return NULL; }
}

% print error message
private define _print_err(err) {
  vmessage("%s\n%s:%s:%d %s", err.message, err.file, err.function, err.line, err.descr);
  if (err.traceback != "") { message(err.traceback); }
}

private define _tcp_send_msg();
private define _tcp_receive_msg();

% send a message to the server or a client
private define _tcp_send_msg(sc, buf) {
  if (sc.connected || qualifier_exists("greet")) {
    variable len, err;
    if (bstrlen(buf) > _TCP_BUF_SIZE) {
      vmessage("error(%s): message exceeds buffer size of %d", _TCP_BUF_SIZE);
      return;
    }
    try (err) {
      % send the message
      len = write(sc.config.socket, buf);%fputs(buf, sc.config.fw);
%      ()=fflush(sc.config.fw);
      % wait for a receipt
      if (qualifier_exists("receipt")) {
	if (_tcp_receive_msg(sc) != "1") {
          throw (RunTimeError, "receipt failed");
	}
      }
    }
    catch AnyError: {
      _print_err(err);
      vmessage("error(%s): could not send the message!", _function_name);
      return -1;
    }
    return len;
  }
  throw (RunTimeError, "not accepted");
}

% send an object to the server or a client
private define _tcp_send_obj(sc, obj) {
  if (sc.connected) {
    sc.config.busyobj = 1;
    % convert the object into a string
    variable str = pack_obj(obj), len = 0;
    % send object code
    if (sc.config.chatty) { ()=system("echo -n 'sending object '"); }
    len += _tcp_send_msg(sc, ccode(_TCP_CODE.object); receipt);
    % send array length
    len += _tcp_send_msg(sc, pack_obj(length(str))[0]; receipt);
    % send the strings
    variable i;
    _for i (0, length(str)-1, 1) {
      len += _tcp_send_msg(sc, str[i]; receipt);
      if (sc.config.chatty && i mod 1000 == 0) {
	()=system(sprintf("echo -n '%3.0f%%\b\b\b\b'", 100.*i/length(str)));
      }
    }
    if (sc.config.chatty) { message("100%"); }
    sc.config.busyobj = 0;
    return len;
  }
  throw (RunTimeError, "not accepted");
}

% either send a message or an object
private define _tcp_send(sc, obj) {
  if (typeof(obj) == BString_Type || typeof(obj) == String_Type) {
    return _tcp_send_msg(sc, obj;; __qualifiers);
  }
  return _tcp_send_obj(sc, obj);
}

% receive a message from a client or the server
private define _tcp_receive_msg(sc) {
  if (sc.config.busy) { throw (RunTimeError, "busy"); }
  % skip if a connection is in progress
  if (sc.connected || qualifier_exists("greet")) {
    sc.config.busy = 1;
    % check the ability to read from the socket
    variable stat;
    (,stat) = select(0, sc.config.socket, NULL, NULL, 0);
    variable buf = NULL;
    % read the message
    if (length(stat.iread) > 0 || not qualifier_exists("dontwait")) {
      ()=read(sc.config.socket, &buf, _TCP_BUF_SIZE);
    }
    % send a receipt
    if (qualifier_exists("receipt")) {
      ()=_tcp_send_msg(sc, "1");
    }
    % return
    sc.config.busy = 0;
    return buf;
  }
  throw (RunTimeError, "not accepted");
}

% receive an object from the server or a client
private define _tcp_receive_obj(sc) {
  if (sc.connected) {
    sc.config.busyobj = 1;
    if (sc.config.chatty) { ()=system("echo -n 'receiving object '"); }
    variable str = BString_Type[unpack_obj(_tcp_receive_msg(sc; receipt))];
    % loop until the object code has been sent
    variable i;
    _for i (0, length(str)-1, 1) {
      str[i] = _tcp_receive_msg(sc; receipt);
      if (sc.config.chatty && i mod 1000 == 0) {
	()=system(sprintf("echo -n '%3.0f%%\b\b\b\b'", 100.*i/length(str)));
      }
    }
    if (sc.config.chatty) { message("100%"); }
    % convert the string into an object
    sc.config.busyobj = 0;
    return unpack_obj(str);
  }
  throw (RunTimeError, "not accepted");
}

% either receive a message or an object
private define _tcp_receive(sc) {
  variable msg = _tcp_receive_msg(sc;; __qualifiers);
  if (msg == NULL) { return NULL; }
  if (qualifier_exists("ignorecode")) { return msg; }
  % react on control codes
  switch (dcode(msg))
    { case _TCP_CODE.disconnect:
      if (sc.config.chatty) {
        vmessage("connection closed by server");
      }
      ()=close(sc.config.socket);
      sc.connected = 0;
      if (struct_field_exists(sc.hook, "closed") && sc.hook.closed != NULL) {
	(@sc.hook.closed)(sc);
      }
      return NULL;
    }
    { case _TCP_CODE.object:
      ifnot (qualifier_exists("receipt")) { ()=_tcp_send_msg(sc, "1"); }
      return _tcp_receive_obj(sc);
    }
    { return msg; }
}


% kick a client or disconnect from the server
private define _tcp_disconnect(sc) {
  % send disconnect code
  ()=_tcp_send_msg(sc, ccode(_TCP_CODE.disconnect);; __qualifiers);
  % close socket
  variable err;
  try (err) {
    ()=close(sc.config.socket);
    % mark as disconnected
    sc.connected = 0;
  }
  catch AnyError: {
    _print_err(err);
    vmessage("error(%s): could not close socket!", _function_name);
    return 0;
  }
  return 1;
}

% shutdown the server
private define _tcp_server_shutdown(s) {
  % triggered shutdown
  if (qualifier_exists("trigger")) { _TCP_SERVER.running = -1; return 1; }
  % normal shutdown
  variable err;
  try (err) {
    % close all client connections
    variable client;
    foreach client (s.client) { if (client.connected == 1) { ()=client.kick(); } }
    % close main socket
    ()=close(s.config.socket);
  }
  catch AnyError: {
    _print_err(err);
    vmessage("error(%s): could not shutdown server!", _function_name);
    return 0;
  }
  _TCP_SERVER.running = 0; % also modifies the user's variable
  if (_TCP_SERVER.config.chatty) {
    vmessage("shutdown of server on port %d", _TCP_SERVER.config.port);
  }
  ifnot (struct_field_exists(_TCP_SERVER, "active")) { alarm(0); _TCP_SERVER = NULL; }
  return 1;
}

% remove not connected clients from the list
private define _tcp_server_cleanclients(s) {
  s.client = s.client[where(array_struct_field(s.client, "connected") == 1)];
}

% server handler: accept connections and receive messages
% from the clients. is called as a result from the alarm signal
% in case no handler is provided
private define _tcp_server_handle(dummy) {
  % check on running sever
  if (_TCP_SERVER == NULL) {
    vmessage("error (%s): server is not running, cancelling handler!", _function_name);
    ifnot (struct_field_exists(_TCP_SERVER, "active")) { alarm(0); }
    return;
  }

  variable client;
  
  % check on pending connections
  variable stat; % check the ability to read from the main socket
  (,stat) = select(0, _TCP_SERVER.config.socket, NULL, NULL, 0);
  if (length(stat.iread) > 0) {
    % prepare client structure
    client = struct {
      send = &_tcp_send,
      receive = &_tcp_receive,
      kick = &_tcp_disconnect,
      connected = 0,
      number = length(_TCP_SERVER.client) == 0 ? 0
             : max(array_struct_field(_TCP_SERVER.client, "number")) + 1,
      config = struct {
        socket, ip, port, chatty = _TCP_SERVER.config.chatty,
        busy = 0, busyobj = 0
      },
      unhmsg = BString_Type[0],
      unhobj = list_new()
    };
    % hook 
    if (_TCP_SERVER.hook.connect != NULL
      && (@_TCP_SERVER.hook.connect)(_TCP_SERVER, client) == 0) {
        ()=close(accept(_TCP_SERVER.config.socket));
        if (_TCP_SERVER.config.chatty) {
          vmessage("client #%d from %s on port %d refused by hook", client.number,
		   client.config.ip, client.config.port);
        }
    }
    % restrict number of clients
    else if (client.number >= _TCP_SERVER.config.maxclients) {
      ()=close(accept(_TCP_SERVER.config.socket));
      if (_TCP_SERVER.config.chatty) {
        vmessage("client #%d from %s on port %d exceeded maximum number of connections",
          client.number, client.config.ip, client.config.port);
      }
    }
    % accept
    else {
      client.config.socket = accept(
        _TCP_SERVER.config.socket, &(client.config.ip), &(client.config.port)
      );
      if (_TCP_SERVER.config.chatty) {
        vmessage("client #%d connected from %s on port %d", client.number,
	         client.config.ip, client.config.port);
      }
      % send greet
      if (_TCP_SERVER.config.nogreet == 0) {
	()=_tcp_send_msg(client, _TCP_SERVER.hook.greet != NULL
          ? (@_TCP_SERVER.hook.greet)(_TCP_SERVER, client) : ccode(_TCP_CODE.greet)
        ; greet);
      } else { % skip greet
        client.connected = 1;
	% call "main" function after a client's connection is established
	if (_TCP_SERVER.hook.established != NULL) {
	  (@_TCP_SERVER.hook.established)(_TCP_SERVER, client);
	}
      }
      % add to list of clients
      _TCP_SERVER.client = [_TCP_SERVER.client, client];
    }
  }

  % check for pending messages
  foreach client (_TCP_SERVER.client) {
    variable msg;
    % client is not trustworthy
    if (client.connected == -1) { continue; }
    % greet
    if (client.connected == 0) {
      msg = _tcp_receive_msg(client; dontwait, greet);
      if (typeof(msg) == BString_Type) {
  	% greet accepted
  	if (dcode(msg) == "G" || (_TCP_SERVER.hook.greet != NULL
	      && (@_TCP_SERVER.hook.greet)(_TCP_SERVER, client, msg))) {
          client.connected = 1;
          if (_TCP_SERVER.config.chatty) {
            vmessage("client #%d's greet accepted", client.number);
          }
	  % call "main" function after a client's connection is established
	  if (_TCP_SERVER.hook.established != NULL) {
	    (@_TCP_SERVER.hook.established)(_TCP_SERVER, client);
	  }
        }
        % client's greet not as expected -> not trustworthy!
  	else {
          ()=client.kick(; greet);
          client.connected = -1;
          if (_TCP_SERVER.config.chatty) {
            vmessage("client #%d's greet refused", client.number);
	  }
	}
      }
      continue;
    }
    % proceed normally
    if (client.config.busy || client.config.busyobj) { continue; }
    msg = _tcp_receive_msg(client; dontwait, ignorecode);
    if (typeof(msg) == BString_Type) {
      switch (dcode(msg))
  	% client disconnected
        { case _TCP_CODE.disconnect:
          client.connected = 0;
          ()=close(client.config.socket);
  	  % hook
	  if (_TCP_SERVER.config.chatty) {
            vmessage("client #%d disconnected", client.number);
	  }
	  if (_TCP_SERVER.hook.disconnect != NULL) {
	    (@_TCP_SERVER.hook.disconnect)(_TCP_SERVER, client);
	  }
  	}
        % object received
        { case _TCP_CODE.object:
	  ()=_tcp_send_msg(client, "1");
	  variable obj = _tcp_receive_obj(client);
          % call client handler
          if (_TCP_SERVER.hook.obj_handler != NULL) {
            (@_TCP_SERVER.hook.obj_handler)(_TCP_SERVER, client, obj);
          }
          % append to received objects
	  else {
	    list_append(client.unhobj, obj);
	  }
	}
        % call client handler or
  	% append to unhandled messages
        {
	  if (_TCP_SERVER.hook.msg_handler != NULL) {
            (@_TCP_SERVER.hook.msg_handler)(_TCP_SERVER, client, msg);
	  } else {
            client.unhmsg = [client.unhmsg, msg];
	  }
        }
    }
    if (_TCP_SERVER.running == 0) { return; }
  }
  
  % re-schedule alarm
  ifnot (struct_field_exists(_TCP_SERVER, "active")) { alarm(1); }
}

%%%%%%%%%%%%%%%%%%%%
define tcp_server()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{tcp_server}
%\synopsis{implements a basic TCP server}
%\usage{Integer_Type tcp_server();
% or  Struct_Type tcp_server(; background);}
%\qualifiers{
%    \qualifier{port}{bind port (default: 1149)}
%    \qualifier{maxclients}{maximum allowed number of clients connected
%      simultaneously (default: 10)}
%    \qualifier{nogreet}{disable greet messages (see below)}
%    \qualifier{background}{run server in the background (see below)}
%    \qualifier{ip}{bind address (default: "0.0.0.0", i.e., accept
%      connections from outside; only change if you know what
%      you are doing)}
%    \qualifier{chatty}{chattiness (default: 1)}
%    \qualifier{COMMENT:}{Further qualifiers are listed at HOOKS below.}
%}
%\description
%    Uses the `socket' module to implement a basic TCP server. It
%    listens for clients trying to connect and accepts a connection
%    after a greet message has been exchanged successfully. The
%    greet message can be disabled using the `nogreet' qualifier,
%    which in this case needs to be set for a `tcp_client' as well.
%    Furthermore, any message or object received from a client is
%    read automatically. After any of these actions taken by the
%    server, a user-defined function may be called in order to handle
%    these events further (see the HOOK section below).
%
%    The `background' qualifier allows the server to be started in
%    background mode, such that commands can still be entered into
%    the ISIS prompt. This background mode is realized by calling
%    the main server function, which handles client events, each
%    second scheduling an `alarm` signal (SIGALRM).
%    NOTE: the alarm signals are triggered in the main program loop
%          of ISIS, i.e., while it's "working". That means the
%          main server function is not executed while ISIS awaits
%          input from the prompt (ISIS "sleeps" here). Thus, you
%          might want to do at least something, e.g, press enter.
%    WARNING: further alarms should not be scheduled using `alarm'
%             or `signal'. Otherwise, the server might crash.
%
%    If not in background mode, `tcp_server' returns either 1 after 
%    the server has been shut down successfully or 0 otherwise. Else
%    the structure of the following form is returned. Note that this
%    structure is also passed to each hook (see below):
%    
%      Struct_Type[] client - list of clients
%      Integer_Type &shutdown()
%        Function to shut down the server. After all clients are
%        requested to disconnect, all sockets are closed. Returns
%        whether the shut down was successful (1) or not (0).
%        Qualifiers:
%          trigger - shut down the server at the end of the main
%            loop, which does not interrupt ongoing communication
%      Void_Type &cleanclients()
%        Function to remove clients no longer connected from the list.
%      Integer_Type running - boolean value indicating the server's
%        state (1: running; 0: shut down)
%      Struct_Type config - internal configuration
%      Struct_Type hook - defined hooks
%
%    Each `client' structure has the following fields:
%    
%      Integer_Type &send(String_Type msg or Any_Type obj)
%        Function for sending a string message or an SLang object
%        to a client (see `pack_obj' for list of supported types).
%        Returns the number of bytes sent.
%        Warning: Currently, sending objects is slow (~40 kB/s,
%                 i.e., ~20 s for 100,000 doubles).
%        Qualifiers:
%          receipt - wait for a confirmation by the client for the
%            receipt (the client's `receive' function needs the
%            same qualifier to be set).
%      BString_Type or Any_Type &receive()
%        Wait for and receive a message from the client. Returns
%        either the message itself (BString_Type) or in case of a
%        received object the object itself.
%        Qualifiers:
%          dontwait - do not wait for a message and return NULL in
%            case no message or object is pending.
%          receipt  - send a receipt to the client after having
%            received a message or an object
%      Integer_Type &kick()
%        Kick the client, i.e., close the communication socket.
%        Returns 1 in success or 0 otherwise.
%      Integer_Type connected - boolean value indicating whether
%        the client is still connected (1) or not (0)
%      Integer_Type number - the client's number, which is a serial
%        increasing identifier starting with zero
%      Struct_Type config - internal configuration
%      BString_Type[] unhmsg - in case the `msg_handler'-hook is not
%        set, messages of the client, which have been received
%        without a call to `receive` during the server's main loop
%        are appended here
%      List_Type unhobj - in case the `obj_handler'-hook is not set,
%        objects without a call to `receive' are appended here
%
% HOOKS
%    The following functions are called via qualifiers of the same
%    name, which need to be set to references (Ref_Type) to user-
%    defined functions (see below for an example). The passed
%    `server' and `client' are the server's and corresponding
%    client's structures as defined above.
%    
%    Integer_Type &connect_hook(server, client)
%      Called after a client tries to connect to the server. Should
%      return 1 or 0 for accepting or rejecting the connection,
%      respectively.
%    String_Type  &greet_hook(server, client)
% or Integer_Type &greet_hook(server, client, greetmsg)
%      Called after accepting a client's connection, which should
%      return the greet message to be sent to the client. Is called
%      again after the re-greet message has been received from the
%      client, which should return 1 or 0 for accepting or rejecting
%      the greet, i.e, the connection, respectively. The hook is not
%      called if the server was set up with the `nogreet' qualifier.
%    Void_Type &established_hook(server, client)
%      After a connection has been accepted including the greet, this
%      function allows to start the communication between the client
%      and the server by, e.g., sending messages to the client in
%      combination with the `msg_handler' hook.
%    Void_Type &disconnect_hook(server, client)
%      Called after a client disconnected from the server by itself,
%      i.e., the client has not been kicked.
%    Void_Type &msg_handler(server, client, msg)
%      Called after a message (BString_Type) has been received from
%      the client. If this hook is not set, then the message is
%      appended to the `unhmsg` field of the client's structure.
%    Void_Type &obj_handler(server, client, obj)
%      Called after an object has been received from the client. If
%      this hook is not set, then the object is appended to the
%      `unhobj` field of the client's structure.
%\example
%    % after a client's connection has been accepted,
%    % send a data structure and disconnect client
%    define do_client(s,c) {
%      vmessage("sending data to client #%d", c.number);
%      ()=c.send("welcome client! sending data...");
%      ()=c.send(struct { data = [1:5]*10, err = [1:5] });
%      ()=c.kick();
%      % remove kicked client from the list
%      s.cleanclients();
%    }
%    % start the server
%    variable stat = tcp_server(; established_hook = &do_client);
%    % check its exit status
%    if (stat == 0) { message("server exited unexpectedly"); }
%
%    % a more sophisticated example can be found on the Remeis wiki
%    % www.sternwarte.uni-erlangen.de/wiki/doku.php?id=isis:socket
%\seealso{tcp_client, socket, alarm, pack_obj}
%!%-
{
  % check on already running server
  if (_TCP_SERVER != NULL) {
    vmessage("error: server already running on port %d!", _TCP_SERVER.config.port);
    return 0;
  }

  % prepare server structure
  variable server = struct {
    client = Struct_Type[0],
    shutdown = &_tcp_server_shutdown,
    cleanclients = &_tcp_server_cleanclients,
    running = 0,
    config = struct {
      socket, ip = qualifier("ip", "0.0.0.0"),
      port = qualifier("port", _TCP_DEFAULT_PORT),
      maxclients = qualifier("maxclients", 10),
      nogreet = qualifier_exists("nogreet"),
      chatty = qualifier("chatty", 1)
    },
    hook = struct {
      connect = qualifier("connect_hook", NULL),
      greet = qualifier("greet_hook", NULL),
      established = qualifier("established_hook", NULL),
      disconnect = qualifier("disconnect_hook", NULL),
      msg_handler = qualifier("msg_handler", NULL),
      obj_handler = qualifier("obj_handler", NULL)
    }
  };
  % create and bind a socket
  variable err;
  try (err) {
    server.config.socket = socket(PF_INET, SOCK_STREAM, 0);
    bind(server.config.socket, server.config.ip, server.config.port);
  }
  catch AnyError: {
    _print_err(err);
    vmessage("error: could not launch server!");
    return 0;
  }

  % start listening
  listen(server.config.socket, server.config.maxclients);
  server.running = 1;
  if (server.config.chatty) {
    vmessage("server starts listening on port %d", server.config.port);
  }
    
  % put handler into the background
  _TCP_SERVER = server;
  if (qualifier_exists("background")) {
    signal(SIGALRM, &_tcp_server_handle);
    alarm(1);
    return server;
  }

  % main loop: call handler
  _TCP_SERVER = struct_combine(server, struct { active });
  variable stime = qualifier("sleep", .1);
  try (err) {
    while (_TCP_SERVER.running == 1) {
      _tcp_server_handle(NULL);
      sleep(stime); % avoid 100% CPU
    }
    % shutdown has been triggered
    if (_TCP_SERVER.running == -1) { ()=_TCP_SERVER.shutdown(); }
  }
  catch UserBreakError: {
    return _TCP_SERVER.shutdown();
  }
  catch AnyError: {
    _print_err(err);
    return 0;
  }
  _TCP_SERVER = NULL;
  return 1; % exit successfully
}

%%%%%%%%%%%%%%%%%%%%
define tcp_client()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{tcp_client}
%\synopsis{connects to a TCP server}
%\usage{Struct_Type  tcp_client(String_Type host[, Integer_Type port]);
%\altusage{Integer_Type tcp_client(String_Type host[, Integer_Type port];
%                   msg_handler = Ref_Type, obj_handler = Ref_Type);}}
%\qualifiers{
%    \qualifier{nogreet}{disable greet messages (see below)}
%    \qualifier{chatty}{chattiness (default: 1)}
%    \qualifier{COMMENT:}{Further qualifiers are listed at HOOKS below.}
%}
%\description
%    Uses the `socket' module to connect to a TCP server. After a
%    connection attempt, a greet message is exchanged with the server
%    by default unless the `nogreet' qualifier is set. In the latter
%    case the `tcp_server' has be set up with the same qualifier.
%    After the connection has been established, a structure with
%    functions for communicating with the server is returned.
%    Alternatively, if the `msg_handler` and `obj_handler` hooks are
%    set (see the HOOK section below) the client remains in a main
%    loop until the connection to the server has been closed. In this
%    case, the function returns 0 if any error occured during the
%    lifetime of the connection or 1 otherwise. The client structure,
%    which is passed to all hooks and returned if these hooks are no
%    set, has the following fields:
%    
%      Integer_Type &send(String_Type msg or Any_Type obj)
%        Function for sending a string message or an SLang object
%        to the server (see `pack_obj' for list of supported types).
%        Returns the number of bytes sent.
%        Warning: Currently, sending objects is slow (~40 kB/s,
%                 i.e., ~20 s for 100,000 doubles).
%        Qualifiers:
%          receipt - wait for a confirmation by the client for the
%            receipt (the server's `receive' function needs the
%            same qualifier to be set).
%      BString_Type or Any_Type &receive()
%        Wait for and receive a message from the server. Returns
%        either the message itself (BString_Type) or in case of a
%        received object the object itself.
%        Qualifiers:
%          dontwait - do not wait for a message and return NULL in
%            case no message or object is pending.
%          receipt  - send a receipt to the client after having
%            received a message or an object
%      Integer_Type &disconnect()
%        Disconnect from the server, i.e., close the communication
%        socket. Returns 1 in success or 0 otherwise.
%      Integer_Type connected - boolean value indicating whether
%        the connection is still established (1) or not (0)
%      Struct_Type config - internal configuration
%      Struct_Type hook - defined hooks
%
% HOOKS
%    The following functions are called via qualifiers of the same
%    name, which need to be set to references (Ref_Type) to user-
%    defined functions (see below for an example). The passed
%    `client' is the client's structure as defined above.
%    
%    String_Type or Integer_Type &greet_hook(client, greetmsg)
%      Called after having received the greet message from the server
%      (if not disabled by the `nogreet' qualifier). Should return
%      the greet message to be sent back to the server. Return 0 in
%      case the greet message from the server should be rejected,
%      which will cancel the connection.
%    Void_Type &connect_hook(client)
%      Called after a connection to the server is established. Can
%      be used to start the communication with the server in
%      combination with the `msg_handler' hook.
%    Void_Type &closed_hook(client)
%      Called after the connection has been closed from the server's
%      side.
%    Void_Type &msg_handler(client, msg)
%      Called after a message (BString_Type) has been received from
%      the server.
%    Void_Type &obj_handler(client, obj)
%      Called after an object has been received from the server.
%\example
%    % simple functions for receiving messages and objects
%    define getmsg(c, msg) { vmessage("message from server: %s", msg); };
%    define getobj(c, obj) { message("received an object:"); print(obj); };
%    % start the client
%    variable stat = tcp_client("localhost"; msg_handler = &getmsg, obj_handler = &getobj);
%    % check its exit status
%    if (stat == 0) { message("lost connection to server"); }
%
%    % a more sophisticated example can be found on the Remeis wiki
%    % www.sternwarte.uni-erlangen.de/wiki/doku.php?id=isis:socket
%\seealso{tcp_server, socket, unpack_obj}
%!%-
{
  variable host, port = _TCP_DEFAULT_PORT;
  switch (_NARGS)
    { case 1: host = (); }
    { case 2: (host, port) = (); }
    { help(_function_name); return; }

  % prepare client structure
  variable client = struct {
    send = &_tcp_send,
    receive = &_tcp_receive,
    disconnect = &_tcp_disconnect,
    connected = 0,
    config = struct{
      host = host, port = port, socket, busy = 0, busyobj = 0,
      nogreet = qualifier_exists("nogreet"),
      chatty = qualifier("chatty", 1)
    },
    hook = struct {
      greet = qualifier("greet_hook", NULL),
      connect = qualifier("connect_hook", NULL),
      closed = qualifier("closed_hook", NULL),
      msg_handler = qualifier("msg_handler", NULL),
      obj_handler = qualifier("obj_handler", NULL)
    }
  };

  % connect
  variable err;
  try (err) {
    client.config.socket = socket(PF_INET, SOCK_STREAM, 0);
    connect(client.config.socket, host, port);
  }
  catch SocketError: {
    if (client.config.chatty) {
      vmessage("error: connection to %s on port %d failed!", host, port);
    }
    return 0;
  }

  % exchange greet message
  if (client.config.nogreet == 0) {
    variable msg = _tcp_receive_msg(client; greet);
    variable tosend = client.hook.greet != NULL
      ? (@client.hook.greet)(client, msg) : ccode(_TCP_CODE.greet);
    if (client.hook.greet == NULL ? dcode(msg) != "G"
	: (typeof(tosend) == Integer_Type && tosend == 0)) {
      if (client.config.chatty) {
        vmessage("error: greet from %s refused", host);
      }
      return 0;
    }
    ()=_tcp_send_msg(client, tosend; greet);
  }

  % established
  client.connected = 1;
  if (client.config.chatty) {
    vmessage("connection to %s on port %d established", host, port);
  }
  if (client.hook.connect != NULL) { (@client.hook.connect)(client); }

  if (client.hook.msg_handler == NULL || client.hook.obj_handler == NULL) { return client; }

  % main loop: call handler
  while (client.connected) {
    try (err) {
      variable obj = client.receive();
      if (client.connected == 0) { return 0; }
      if (typeof(obj) == BString_Type) { (@client.hook.msg_handler)(client, obj); }
      else { (@client.hook.obj_handler)(client, obj); }
    }
    catch AnyError: {
      _print_err(err);
      return 0;
    }
  }
  return 1;
}

%%%%%%%%%%%%%%%%%%%%
define tcp_get_server_handle()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{tcp_get_server_handle}
%\synopsis{get a handle to a running TCP server}
%\usage{Struct_Type tcp_get_server_handle();}
%\qualifiers{
%    \qualifier{kill}{shut down / kill the server}
%}
%\description
%    In case the handle, i.e., the structure of a running
%    `tcp_server' got lost by accident, this function
%    returns this structure again. Alternatively, the
%    `kill' qualifier tries to shut down the server. If
%    this fails all communication sockets are closed,
%    which basically means to kill the server.
%\seealso{tcp_server}
%!%-
{
  % check on a running server
  if (_TCP_SERVER == NULL) {
    vmessage("warning(%s): no running TCP server found!", _function_name);
    return NULL;
  }
  % shut down the server
  if (qualifier_exists("kill")) {
    vmessage("%s: sending a shut down request to the server", _function_name);
    if (_TCP_SERVER.shutdown() == 0) {
      vmessage("%s: shut down failed, closing all sockets", _function_name);
      variable i, err;
      _for i (0, length(_TCP_SERVER.clients)-1, 1) {
        try (err) {
          ()=close(_TCP_SERVER.clients[i].config.socket);
	}
        catch AnyError: {
          _print_err(err);
        }
      }
      try (err) {
        ()=close(_TCP_SERVER.config.socket);
      }
      catch AnyError: {
        _print_err(err);
      }
      vmessage("%s: all sockets closed, i.e., server killed", _function_name);
      _TCP_SERVER = NULL;
    } else { _TCP_SERVER = NULL; }
  }
  % return the structure
  return _TCP_SERVER;
}


#endif

%
% Code from src/fitting/parameters/thawedParameters.sl
%
define thawedParameters()
%!%+
%\function{thawedParameters}
%\synopsis{find all parameters of the current fit-function that are not frozen}
%\usage{Integer_Type[] thawedParameters()}
%\description
%    Note that \code{thawedParameters} may include ones that are
%    tied to another one, or derived as functions of other parameters.
%\seealso{freeParameters}
%!%-
{
  variable params = get_params();
  if(params!=NULL)
    return where( array_struct_field(params, "freeze")==0 )
           + 1;  % parameter numbers start with 1
  else
    return Integer_Type[0];
}

%
% Code from src/slang/date/time_array.sl
%
define time_array( time )
%!%+
%\function{time_array}
%\synopsis{Converts a time [s] (e.g., _time) into an array [sec,min, ...]}
%\usage{Integer_Type[] = time_array( Integer_Type );}
%\altusage{String_Type = time_array( Integer_Type ; str);}
%\qualifiers{
%\qualifier{str}{Instead of a time array a String_Type is returned!}
%}
%\description
%   Converts a time given in seconds into a time array, where
%   the first entry corresponds to the seconds, the second to
%   the minutes, etc. The length of the array is variable,
%   i.e., if the given time is of the order of hours, the array
%   will not include days and years.
%
%   If the qualifier 'str' is given the array will be converted
%   into a string (see example).
%\example
%   isis> ta = format_time ( _time ); ta;
%   Integer_Type[5];
%   isis> print(ta);
%   54
%   1
%   14
%   302
%   43
%
%   %or
%   isis> ts = format_time ( _time ;str ); ts;
%   43y 302d 14h  1m 54s
%   
%\seealso{_time}
%!%-
{
  variable ftime = Integer_Type[0];
  variable tmod = [ 60, 60, 24, 365.25 ];

  variable rest = @time;
  variable i;
  for( i=0 ; i<length(tmod) and rest!=0 ; ++i ){
    ftime = [ ftime, rest mod tmod[i] ];
    rest -= ftime[-1];
    rest /= tmod[i];
    }
  if( rest != 0 )
    ftime = [ ftime, rest ];
  ftime = int(ftime);

  if( qualifier_exists("str") ){
    variable sftime="";
    variable unit = ["s","m","h","d","y"];
    _for i ( length(ftime)-1, 0, -1 ){
      sftime += sprintf("%2d%s ",ftime[i],unit[i]);
	  }
    return sftime[[:-2]];
    }
  return ftime;
}

%
% Code from src/misc/transpose_table.sl
%
define transpose_table()
%!%+
%\function{transpose_table}
%\synopsis{transforms a structure of arrays into an array of structures and vice versa}
%\usage{Array_Type table_by_rows    = transpose_table(Struct_Type table_by_columns);
%\altusage{Struct_Type table_by_columns = transpose_table(Array_Type table_by_rows);}
%}
%\description
%    \code{table_by_columns.field[i] = table_by_rows[i].field;}  % for each \code{field} and index \code{i}
%!%-
{
  variable table;
  switch(_NARGS)
  { case 1: table = (); }
  { help(_function_name()); return NULL; }

  variable field, fields;
  if(typeof(table)==Struct_Type)
  { fields = get_struct_field_names(table);
    variable i, l = length(get_struct_field(table, fields[0]));
    variable rows = Struct_Type[l];
    _for i (0, l-1, 1)
    { rows[i] = @Struct_Type(fields);
      foreach field (fields)
        set_struct_field( rows[i], field,  get_struct_field(table, field)[i] );
    }
    return rows;
  }
  else
  { fields = get_struct_field_names(table[0]);
    variable cols = @Struct_Type(fields);
    foreach field (fields)
      set_struct_field(cols, field,  array_struct_field(table, field) );
    return cols;
  }
}

%
% Code from src/slang/misc/try_or_timeout.sl
%
#ifeval __get_reference("TimeoutError") == NULL
new_exception ("TimeoutError", RunTimeError, "Operation has timed out");
#endif

private define __alarm_signal_handler(sig) {
  throw TimeoutError;
}

%!%+
%\function{try_or_timeout}
%\usage{try_or_timeout(UInteger_Type seconds, Ref_Type function, ...);}
%\description
%    Runs the function referenced by \code{function} for at most
%    \code{seconds} seconds. If \code{function} finishes before
%    \code{seconds} have elapsed it returns the return values of
%    \code{function}. If \code{function} does not finish in that time
%    it throws a \code{TimoutError}.
%    Additional arguments passed on to \code{function} may simply be
%    specified as additional arguments to \code{try_or_timeout}.
%    Qualifiers can be passed on as usual.
%!%-
define try_or_timeout() {
  if ( _NARGS <= 1 ) {
    throw NumArgsError, "Invalid number of arguments";
  }
  variable args = __pop_list(_NARGS);
  variable seconds = args[0];
  list_delete(args, 0);
  variable old_handler;
  try {
    signal(SIGALRM, &__alarm_signal_handler, &old_handler);
    alarm(seconds);
    variable todo = args[0];
    list_delete(args, 0);
    (@todo)(__push_list(args);; __qualifiers);
  } finally {
    alarm(0);
    if ( typeof(old_handler) == Ref_Type || typeof(old_handler) == Integer_Type ) {
      signal(SIGALRM, old_handler);
    } else {
      signal(SIGALRM, SIG_DFL);
    }
  }
}

%
% Code from src/slang/date/tt2tai.sl
%
define tt2tai()
%!%+
%\function{tt2tai}
%\synopsis{Convert a (M)JD in Terrestrial Time (TT, aka TDT or ET) into TAI}
%\usage{tai=tt2tai(JD)}
%\description
%This routine converts a Julian Date that is given in Terrestrial Time
%(formerly known as Terrestrial Dynamic Time or Ephemeris Time) into
%TAI. This is done by subtracting 32.184s from the TT.
% This routine is array safe.
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the TAI-TT in seconds, rather than the corrected (M)JD.}
%}
%\seealso{utc2tai,tai2utc,tai2tt,tt2tdb}
%!%-
{
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    variable diff=-32.184;

    if (qualifier_exists("deltat")) {
	if (typeof(JD)==Array_Type) {
            return diff+Double_Type[length(JD)];
	}
	return diff;
    }
    return JD+diff/86400.0;
}

%
% Code from src/slang/date/tt2tcg.sl
%
define tt2tcg()
%!%+
%\function{tt2tcg}
%\synopsis{Convert a (M)JD in Terrestrial Time (TT, aka TDT or ET) into TCG}
%\usage{tai=tt2tcg(JD)}
%\description
%  This routine converts a Julian Date that is given in Terrestrial Time
%  (formerly known as Terrestrial Dynamic Time or Ephemeris Time) into
%  geocentric coordinate time (TCG). The two time scales differ by a constant
%  rate(!) because TT uses the SI second at the surface of the geoid, and TCG uses
%  the SI second at the potential of the center of the Earth.
%  This routine is array safe.
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the TCG-TT in seconds, rather than the corrected (M)JD.}
%}
%\seealso{tcg2tt,utc2tai,tai2utc,tai2tt,tt2tdb}
%!%-
{
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    % rate factor (exact by definition)
    variable lg=6.969290134e-10;
    
%    variable diff=-32.184;

    % zero point of TT, TCG, and TCB
    % (1977 Jan 1 00:00:00 TAI)
    variable t0;
    if (qualifier_exists("mjd")) {
        t0=43144.0+32.184/86400.; % TT
    } else {
        t0=2443144.5+32.184/86400.; % TT
    }
    % Kaplan, eq 2.3
    % TT= TCG-lg*(TCG-t0);
    % TT = TCG - lg*TCG + lg*t0
    % TT = TCG*(1-lg) + lg*t0
    % TT - lg*t0 = TCG*(1-lg)
    % (TT-lg*t0)/(1-lg) = TCG

    variable tcg=(JD-lg*t0)/(1.-lg);

    if (qualifier_exists("deltat")) {
	return (tcg-JD)*86400.;
    }
    return tcg;
}

define tcg2tt()
%!%+
%\function{tcg2tt}
%\synopsis{Convert a (M)JD in TCG to Terrestrial Time (TT, aka TDT or ET)}
%\usage{tai=tcg2tt(JD)}
%\description
%  This routine converts a Julian Date that is given in geocentric coordinate
%  time (TCG) to Terrestrial Time. The two time scales differ by a constant
%  rate(!) because TT uses the SI second at the surface of the geoid, and TCG
%  uses the SI second at the potential of the center of the Earth.
%  This routine is array safe.
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the TT-TCG in seconds, rather than the corrected (M)JD.}
%}
%\seealso{tcg2tt,utc2tai,tai2utc,tai2tt,tt2tdb}
%!%-
{
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    % rate factor (exact by definition)
    variable lg=6.969290134e-10;
    
    % zero point of TT, TCG, and TCB
    % (1977 Jan 1 00:00:00 TAI)
    variable t0;
    if (qualifier_exists("mjd")) {
        t0=43144.0+32.184/86400.; % TT
    } else {
        t0=2443144.5+32.184/86400.; % TT
    }
    % Kaplan, eq 2.3
    % TT= TCG-lg*(TCG-t0);
    % TT = TCG - lg*TCG + lg*t0
    % TT = TCG*(1-lg) + lg*t0
    % TT - lg*t0 = TCG*(1-lg)
    % (TT-lg*t0)/(1-lg) = TCG

    variable tt=JD-lg*(JD-t0);

    if (qualifier_exists("deltat")) {
	return (tt-JD)*86400.;
    }
    return tt;
}


%
% Code from src/slang/date/tt2tdb.sl
%
define tt2tdb()
%!%+
%\function{tt2tdb}
%\synopsis{Convert a (M)JD in Terrestrial Time (TT, aka TDT or ET) into Terrestrial Barycentric Time (TDB)}
%\usage{tdb=tt2tdb(JD)}
%\description
%  This function converts a Julian Date that is given in Terrestrial
%  Time into Terrestrial Dynamic Barycentric Time.
%
%  By default the routine uses equation 2.6 of Kaplan (2005, USNO Circular 179), which
%  is a truncated version of the series provided by Fairhead and Bretagnon,
%  1990, A&A 229, 240 and which is better than +/-10mus between AD1600 and AD2200.
%
%  If the qualifier ephemeris is given, and points to a JPL ephemeris that
%  does contain TT-TDB information, then the value determined from the
%  ephemeris is computed (which has a higher precision than Fairhead & Bretagnon,
%  but is slower to evaluate).
%
%  Note that for virtually all practical applications apart from barycentering
%  ms radio pulsar data it is sufficient to use TT as the argument to DE405
%  or VSOP1987.
%
%  This routine is array safe.
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the difference TDB-TT in seconds, rather than the corrected (M)JD.}
%\qualifier{ephemeris}{if set and no argument, then correct using DE430, otherwise
%                        use the ephemeris provided (see function jpl_initeph).}
%}
%\seealso{tdb2tt,utc2tai,tai2utc,tai2tt,jpl_initeph}
%!%-
{
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    variable dif;
    if (qualifier_exists("ephemeris")) {

	variable ephstruct;
	if (typeof(qualifier("ephemeris"))==Struct_Type) {
	    ephstruct=struct {ephemeris=qualifier("ephemeris")}; % pass to jpl_eph
	} else {
	    ephstruct=struct {dummynameinvalid=0}; % cheat...
	}
	
	variable jdtmp=@JD;
	if (qualifier_exists("mjd")) {
	    jdtmp=jd2mjd(jdtmp);
	}
	
	dif=Double_Type[length(jdtmp)];

	variable i;
	_for i(0,length(jdtmp)-1,1) {
	    variable ddd=(jpl_eph(jdtmp[i],"TT-TDB";;ephstruct))[0]; % ignore time derivative

	    % now iterate until ns precision is reached
	    % (typically only 2 iterations are needed)
	    do {
		variable dold=ddd;
		ddd=(jpl_eph(jdtmp[i]-ddd/86400.,"TT-TDB";;ephstruct))[0]; % ignore time derivative
	    } while (abs(dold-ddd) > 1e-9);
	    dif[i]=-ddd; % TT-TDB -> TDB-TT
	}
	if (typeof(jdtmp)!=Array_Type) {
	    dif=dif[0];
	}
    } else {
	% Fairhead and Bretagnon series
	variable T;
	if (qualifier_exists("mjd")) {
	    T=(JD-51544.5)/36525.;
	} else {
	    T=(JD-2451545.0)/36525.;
	}	
	
        dif=+0.001657*sin( 628.3076*T+6.2401)
            +0.000022*sin( 575.3385*T+4.2970)
            +0.000014*sin(1256.6152*T+6.1969)
            +0.000005*sin( 606.9777*T+4.0212)
            +0.000005*sin(  52.9691*T+0.4444)
            +0.000002*sin(  21.3299*T+5.5431)
            +0.000010*T*sin(628.3076*T+4.2490);
    }
    if (qualifier_exists("deltat")) {
	return dif;
    }
    return JD+dif/86400.0;
}

define tdb2tt()
%!%+
%\function{tdb2tt}
%\synopsis{Convert a (M)JD in Terrestrial Barycentric Time (TDB) into Terrestrial Time (TT, aka TDT)}
%\usage{tt=tdb2tt(JD)}
%\description
%  This function converts a Julian Date that is given in Terrestrial Barycentric
%  time (TDB) into Terrestrial Time (TT).
%
%  By default the routine uses equation 2.6 of Kaplan (2005, USNO Circular 179),
%  which is a truncated version of the series provided by Fairhead and Bretagnon,
%  1990, A&A 229, 240 and which is better than +/-10mus between AD1600 and
%  AD2200.
%
%  Since that series is for TT-TDB, we formally calculate the deltat for the TT
%  that equals the given JD, we make no attempt to correct for the slight
%  change in deltat between JD(TT) and JD(TDB), since the correction is smaller
%  than the precision of the Fairhead and Bretagnon series. Use the JPL ephemeris
%  for ms pulsar work.
%
%  If the qualifier ephemeris is given, and points to a JPL ephemeris that
%  does contain TT-TDB information, then the value determined from the
%  ephemeris is computed (which has a higher precision than Fairhead & Bretagnon,
%  but is slower to evaluate).
%
%  Note that for virtually all practical applications apart from barycentering
%  ms radio pulsar data it is sufficient to use TT as the argument to DE405
%  or VSOP1987.
%
%  This routine is array safe.
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the difference TT-TDB in seconds, rather than the corrected (M)JD.}
%\qualifier{ephemeris}{if set and no argument, then correct using DE430, otherwise
%                        use the ephemeris provided (see function jpl_initeph).}
%}
%\seealso{tdb2tt,utc2tai,tai2utc,tai2tt,jpl_initeph}
%!%-
{
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    variable dif;
    if (qualifier_exists("ephemeris")) {

	variable ephstruct;
	if (typeof(qualifier("ephemeris"))==Struct_Type) {
	    ephstruct=struct {ephemeris=qualifier("ephemeris")}; % pass to jpl_eph
	} else {
	    ephstruct=struct {dummynameinvalid=0}; % cheat...
	}
	
	variable jdtmp=@JD;
	if (qualifier_exists("mjd")) {
	    jdtmp=jd2mjd(jdtmp);
	}
	
	dif=Double_Type[length(jdtmp)];

	variable i;
	_for i(0,length(jdtmp)-1,1) {
	    dif[i]=(jpl_eph(jdtmp[i],"TT-TDB";;ephstruct))[0]; % ignore time derivative
	}
	if (typeof(jdtmp)!=Array_Type) {
	    dif=dif[0];
	}
    } else {
	% Fairhead and Bretagnon series for TDB-TT
	variable T;
	if (qualifier_exists("mjd")) {
	    T=(JD-51544.5)/36525.;
	} else {
	    T=(JD-2451545.0)/36525.;
	}	
	
        dif=+0.001657*sin( 628.3076*T+6.2401)
            +0.000022*sin( 575.3385*T+4.2970)
            +0.000014*sin(1256.6152*T+6.1969)
            +0.000005*sin( 606.9777*T+4.0212)
            +0.000005*sin(  52.9691*T+0.4444)
            +0.000002*sin(  21.3299*T+5.5431)
	    +0.000010*T*sin(628.3076*T+4.2490);
	dif=-dif;
    }
    if (qualifier_exists("deltat")) {
	return dif;
    }
    return JD+dif/86400.0;
}

%
% Code from src/slang/arrays/unique_n.sl
%
%%%%%%%%%%%%%%%%%%
define unique_n()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{unique_n}
%\synopsis{finds the indicies of unique tupels}
%\usage{Integer_Type[] unique_n(Array_Type a1, ..., an)}
%\description
%    The arrays \code{a1}, ..., \code{an} (with \code{n>=1}) must have the same length,
%    but may contain different data types.
%    The return value is an array of all indices \code{i}
%    of all unique tupels \code{(a1[i], ..., an[i])}.
%!%-
{
  variable arg = __pop_list(_NARGS);

  variable a, i, j;
  switch(_NARGS)
  { case 0: return help(_function_name()); }
  { case 1:
      a = arg[0];
      % after John Houck's unique function from isis/share/mathmisc.sl
      variable len = length(a);
      if(len <= 1)  return [0:len-1];
      a = _reshape (__tmp(a), len);
      i = array_sort(a);
      a = a[i];
      j = where(a[[1:]] != a[[:-2]]);
      return i[[j,-1]];
  }
  variable u = Integer_Type[0];
  a = list_pop(arg);
  variable unique_a = Char_Type[length(a)] + 1;
  _for i (0, length(a)-1, 1)
    if(unique_a[i])
    {
      i = where(a==a[i]);
      unique_a[i] = 0;
      variable bi = {};
      _for j (0, _NARGS-2, 1)
        list_append(bi, arg[j][i]);
      u = [u, i[unique_n(__push_list(bi))]];
    }
  return u;
  % return u[array_sort(u)];
}

%
% Code from src/fitting/fit-functions/unshift.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%
define shift_intpol()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{shift_intpol}
%\synopsis{Shifts the elements of an array continuously}
%\usage{Array_Type shift_intpol(Array_Type array, Double_Type n)}
%\description
%    This function does in principle work like the 'shift'
%    function, with the exception that the amount of the
%    shift may be a floating point number. The values of
%    the resulting array are in that case re-distributed
%    by linear interpolation. Thereby, the the sum of the
%    array values is still preserved.
%\seealso{shift}
%!%-
{
  variable x, n;
  switch(_NARGS)
    { case 2: (x,n) = (); }
    { help(_function_name()); return; }
  
  variable len = length(x);
  ifnot (len) return x;

  % allow n to be negative and large (adopted from 'shift')
  n = n mod len + len;
  % fraction of a value to be distributed into following index
  variable c = n - int(n);
  % rotate and redistribute
  n = int(n); n = [n:n+len-1];
  return x[n mod len]*(1.-c) + x[(n+1) mod len]*c;
}


%%%%%%%%%%%%%%%%%%%%%%%%
private define unshift_fit(lo, hi, pars) {
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{unshift}
%\synopsis{fit function to determine the shift of an array to a reference one}
%\usage{unshift(id)}
%\description
%    The function 'shift' rotates an array by a given number
%    of indices. The fit function described here tries to
%    get this number back: the counts of the dataset 'id'
%    are assumed to be a shifted version of a reference
%    array. In order to do the fit, this reference array
%    has to be associated to the dataset via:
%    
%      set_dataset_metadata(id, reference_array)
%      
%    It is also possible to scale the array values to match
%    the reference ones or to add an offest to the array.
%    Summarized, the fit parameters are as follows:
%      shift  - relative shift of the counts to the
%               reference array
%      scale  - factor the reference is multiplied with
%      offset - offset added to the model counts
% 
%    The model counts are calculated by
%      model = scale*shift(reference_array, shift) + offset
%\seealso{backshift, shift_intpol, set_dataset_metadata, shift}
%!%-
  % get in array
  variable in = get_dataset_metadata(Isis_Active_Dataset);
  ifnot (typeof(in) == Array_Type) { vmessage("error (%s): metadata has to be an array! aborting...", _function_name); return in; }
  return shift_intpol(in, pars[0]*length(in))*pars[1] + pars[2];
}
add_slang_function("unshift", ["shift","scale","offset"]);
private define unshift_default_hook(i) {
  variable var = [
    struct { value = 0.2, freeze = 0, hard_min = -2, hard_max = 2, min = 0.0, max = 1.0, step = 1e-1, relstep = 1e-3 },
    struct { value = 1.0, freeze = 1, hard_min = 0, hard_max = 2, min = 0.0, max = 1.0, step = 1e-3, relstep = 1e-5 },
    struct { value = 0.0, freeze = 1, hard_min = -DOUBLE_MAX, hard_max = DOUBLE_MAX, min = 0.0, max = 0.0, step = 1e-3, relstep = 1e-5 }
  ];
  return var[i];
}
set_param_default_hook("unshift", &unshift_default_hook);


%%%%%%%%%%%%%%%%%%%%%%%%
define backshift()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{backshift}
%\synopsis{determines the shift of an array to a reference one}
%\usage{Struct_Type backshift(Double_Type[] in, Double_Type[] ref[, Double_Type[] error]);}
% \altusage{Double_Type[] backshift(Double_Type[] in, Double_Type[] ref[, Double_Type[] error]; retarr);}
%\qualifiers{
%    \qualifier{retarr}{function returns the unshifted array instead
%             of the structure defined below}
%}
%\description
%    The function 'shift' rotates an array by a given number
%    of indices. The function described here tries to get
%    this number back: an input array 'in' is assumed to be
%    a shifted version of the reference array 'ref'. The
%    shift is then determined by using the fit-function
%    'unshift', also taking optional 'error's into acocunt.
%    The returned structure, if the 'retarr'-qualifier is
%    not set, is defined as follows:
%      shift         - shift in number of indices
%      shift_conf    - its [lower, upper] confidence limits
%      relative      - relative shift between 0 and 1
%      relative_conf - its [lower, upper] confidence limits
%      redChiSqr     - reduced chi-square value of the fit
%\seealso{unshift, shift_intpol, shift}
%!%-
{
  variable in, ref, err, noerr = 0;
  switch(_NARGS)
    { case 2: (in,ref) = (); err = _typeof(in)[length(in)] + DOUBLE_EPSILON; noerr = 1; }
    { case 3: (in,ref,err) = (); }
    { help(_function_name()); return; }

  ifnot (length(in) == length(ref) && length(ref) == length(err)) { vmessage("error (%s): input arrays have to be of equal length", _function_name); return; }

  % save actual fit function and exclude all data from fit
  variable ff = get_fit_fun;
  if (ff == NULL) ff = "";
  exclude(all_data);
  % define the data
  if (min(err) < Minimum_Stat_Err || Minimum_Stat_Err == -1) {
    vmessage("warning (%s): changed Minimum_Stat_Err to match given input uncertainties", _function_name);
    Minimum_Stat_Err = .9 * min(err);
  }
  variable i = define_counts([1:length(in)], [2:length(in)+1], in, err);
  set_dataset_metadata(i, ref);
  % set the fit function and fit
  fit_fun("unshift(1)");
  set_par("unshift(1).shift", 0.2, 0, 0.0, 1.0);
  ()=fit_counts(; fit_verbose = -1);
  % error estimation
  set_par("unshift(1).shift", get_par("unshift(1).shift"), 0, get_par("unshift(1).shift")-.5, get_par("unshift(1).shift")+.5);
  variable mn, mx, stat;
  ifnot (noerr) (mn,mx) = conf_loop("unshift(1).shift"; cl_verbose = -1, fit_verbose = -1);
  ()=eval_counts(&stat; fit_verbose = -1);
  % return value
  variable ret = qualifier_exists("retarr") ? shift(in, -int(round(get_par("unshift(1).shift")*length(in)))) : struct {
    shift = get_par("unshift(1).shift")*length(in),
    shift_conf = (noerr ? NULL : [mn[0], mx[0]]*length(in)),
    relative = get_par("unshift(1).shift"),
    relative_conf = (noerr ? NULL : [mn[0], mx[0]]),
    redChiSqr = stat.statistic / (stat.num_bins - stat.num_variable_params)
  };
  % delete the previously defined data and restore fit function
  set_dataset_metadata(i, NULL);
  delete_data(i);
  fit_fun(ff);
  include(all_data);
  return ret;
}

%
% Code from src/fitting/fit-functions/simple_gpile/use_simple_gpile.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define use_simple_gpile()
%%%%%%%%%%%%%%%%%%%%%%%
{
  variable data1, arf2=NULL, arf3=NULL;
  switch(_NARGS)
  { case 1:  data1 = (); }
  { case 2: (data1, arf2) = (); }
  { case 3: (data1, arf2, arf3) = (); }
  { message("usage: use_simple_gpile(data1[, arf2[, arf3]]);"); return; }

  if(arf2!=NULL)
  { arf2 = [arf2];
    if(length(arf2)!=length(data1)) { message("error (use_simple_gpile): arrays data1 and arf2 have different length"); return; }
  }
  if(arf3!=NULL)
  { arf3 = [arf3];
    if(length(arf3)!=length(data1)) { message("error (use_simple_gpile): arrays data1 and arf3 have different length"); return; }
  }

  variable simple_gpile_fun = "simple_gpile2";
  variable fitFun = get_fit_fun();
  if(substr(fitFun, 1, strlen(simple_gpile_fun)) != simple_gpile_fun)
    fit_fun(simple_gpile_fun + "(Isis_Active_Dataset, " + fitFun + ")");

  % assign model parameters
  variable i;
  foreach i (all_data)
  {
    variable simple_gpile_instance = simple_gpile_fun + "(" + string(i) + ")";
    set_par(simple_gpile_instance + ".beta", 0, 1);
    set_par(simple_gpile_instance + ".data_indx", 0, 1);
    set_par(simple_gpile_instance + ".arf2_indx", 0, 1);
    set_par(simple_gpile_instance + ".arf3_indx", 0, 1);
  }
  _for i (0, length([data1])-1, 1)
  {
    variable id = [data1][i];
    simple_gpile_instance = simple_gpile_fun + "(" + string(id) + ")";
    set_par(simple_gpile_instance + ".beta", 0.05, 0);
    set_par(simple_gpile_instance + ".data_indx", id, 1);
    if(arf2!=NULL)  set_par(simple_gpile_instance + ".arf2_indx", arf2[i], 1);
    if(arf3!=NULL)  set_par(simple_gpile_instance + ".arf3_indx", arf3[i], 1);
  }
}

%
% Code from src/slang/date/ut12tt.sl
%
private variable morrison_deltat_y=[
-500.,-400.,-300.,-200.,-100.,0.,100.,200.,300.,400.,500.,600.,700.,800.,
900.,1000.,1100.,1200.,1300.,1400.,1500.,1600.,1700.,1710.,1720.,1730.,1740.,
1750.,1760.,1770.,1780.,1790.,1800.,1810.,1820.,1830.,1840.,1850.,1860.,
1870.,1880.,1890.,1900.,1910.,1920.,1930.,1940.,1950.,1960.,1970.,1980.,1990.,2000.];

private variable morrison_deltat_d=[
17190.,15530.,14080.,12790.,11640.,10580.,9600.,8640.,7680.,6700.,
5710.,4740.,3810.,2960.,2200.,1570.,1090.,740.,490.,320.,200.,120.,9.,
10.,11.,11.,12.,13.,15.,16.,17.,17.,14.,13.,12.,8.,6.,7.,8.,2.,-5.,
-6.,-3.,10.,21.,24.,24.,29.,33.,40.,51.,57.,65.];


define ut12tt() {
%!%+
%\function{ut12tt}
%\synopsis{Convert a (M)JD in UT1 into TT}
%\usage{utc=ut12tt(JD)}
%
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the TT-UT1 in seconds, rather than the corrected (M)JD.}
%\qualifier{morrison}{Perform the correction using a linear interpolation
%                       to the points by Morrison and Stephenson, 2004,
%                       Journal for the History of Astronomy 35, 327.
%                       Not really good after 2000}
%}
%\description
% This function converts a (Modified) Julian Date that is given in UT1,
% i.e., the timescale that defines the earth rotation angle,
% into a (M)JD in TT.
%
% By default, for years before 2005 the function uses the polynomial
% expressions for TT-UT1 published by F. Espenak and J. Meeus, 2006,
% Five Millenium Canon of Solar Eclipses: -1999 to +3000,
% NASA TP-2006-214141, for years after 2005 the polynomials published
% by Espenak at http://www.eclipsewise.com/help/deltatpoly2014.html are
% used.
%
% Alternatively, a linar interpolation in the solar eclipse derived
% data by Morrison and Stephenson (2000) can be used (see morrison
% qualifier).
%
% The uncertainty of the result is 1s or less from 1780 until 1830,
% less than 1s from 1830-2000. It increases quadratically backwards
% from 1820, sigma ~32((year-1820)/100)^2.
%
% This function is array safe.
%\seealso{tt2ut1,utc2tai,tai2utc,tai2tt}
%!%-
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    variable year;
    if (qualifier_exists("mjd")) {
	year=EpochofJD(JD;mjd);
    } else {
	year=EpochofJD(JD);
    }

    variable deltat;
    
    if (qualifier_exists("morrison")) {
	% TT-UT1 in seconds	
	deltat=interpol(year,morrison_deltat_y,morrison_deltat_d);

	if (qualifier_exists("deltat")) {
	    return deltat;
	}
	return JD+deltat/86400.;
    }


    deltat=Double_Type[length(year)];
    variable u,tmp;
    variable ndx=where(year<-500.);
    if (length(ndx)>0) {
	u=(year[ndx]-1820.)/100.;
	deltat[ndx]=-20.+32.*u*u;
    }
    ndx=where(-500<=year<500.);
    if (length(ndx)>0) {
	u=year[ndx]/100.;
	deltat[ndx]=(((((0.0090316521*u+0.022174192)*u-0.1798452)*u-5.952053)*u+33.78311)*u-1014.41)*u+10583.6;
    }
    ndx=where(+500<=year<1600);
    if (length(ndx)>0) {
	u=(year[ndx]-1000.)/100.;
	deltat[ndx]=(((((0.0083572073*u-0.005050998)*u-0.8503463)*u+0.319781)*u+71.23472)*u-556.01)*u+1574.2;
    }
    ndx=where(+1600<=year<1700);
    if (length(ndx)>0) {
	u=year[ndx]-1600.;
	deltat[ndx]=((1./7129.*u-0.01532)*u-0.9808)*u+120.;
    }
    ndx=where(+1700<=year<1800.);
    if (length(ndx)>0) {
	u=year[ndx]-1700.;
	deltat[ndx]=(((-1./1174000.*u+0.00013336)*u-0.0059285)*u+0.1603)*u+8.83;
    }
    ndx=where(1800<=year<1860.);
    if (length(ndx)>0) {
	u=year[ndx]-1800.;
	deltat[ndx]=((((((0.000000000875*u-0.0000001699)*u+0.0000121272)*u-0.00037436)*u+0.0041116)*u+0.0068612)*u-0.332447)*u+13.72;
    }
    ndx=where(1860.<=year<1900.);
    if (length(ndx)>0) {
	u=year[ndx]-1860.;
	deltat[ndx]=((((1./233174.*u-0.0004473624)*u+0.01680668)*u-0.251754)*u+0.5737)*u+7.62;
    }
    ndx=where(1900.<=year<1920.);
    if (length(ndx)>0) {
	u=year[ndx]-1900.;
	deltat[ndx]=(((-0.000197*u+0.0061966)*u-0.0598939)*u+1.494119)*u-2.79;
    }
    ndx=where(1920.<=year<1941.);
    if (length(ndx)>0) {
	u=year[ndx]-1920.;
	deltat[ndx]=((0.0020936*u-0.076100)*u+0.84493)*u+21.20;
    }
    ndx=where(1941<=year<1961);
    if (length(ndx)>0) {
	u=year[ndx]-1950.;
	deltat[ndx]=((1./2547*u-1./233)*u+0.407)*u+29.07;
    }
    ndx=where(1961<=year<1986);
    if (length(ndx)>0) {
	u=year[ndx]-1975.;
	deltat[ndx]=((-1./718*u-1./260.)*u+1.067)*u+45.45;
    }
    ndx=where(1986<=year<2005.);
    if (length(ndx)>0) {
	u=year[ndx]-2000.;
	deltat[ndx]=((((0.00002373599*u+0.000651814)*u+0.0017275)*u-0.060374)*u+0.3345)*u+63.86;
    }
    ndx=where(2005.<=year<2015.);
    if (length(ndx)>0) {
	u=year[ndx]-2005.;
	deltat[ndx]=64.69+0.2930*u ;
	
    }
    ndx=where(2015.<=year);
    if (length(ndx)>0) {
	u=year[ndx]-2015.;
	deltat[ndx]=(0.0039755*u+0.3645)*u+67.62;
    }

    % ensure correct return type
    if (typeof(JD)!=Array_Type) {
	deltat=deltat[0];
    }

    if (qualifier_exists("deltat")) {
        % deltat is TT-UT1
        % minus sign: we want to be consistent with the other
        % conversion routines, so need to return UT1-TT
	return -deltat;
    }
    % deltat=TT-UT1 -> TT = deltat+UT1
    return JD+deltat/86400.;
}

define tt2ut1() {
%!%+
%\function{tt2ut1}
%\synopsis{Convert a (M)JD in TT into UT1}
%\usage{ut1=tt2ut1(JD)}
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the TT-UT1 in seconds, rather than the corrected (M)JD.}
%\qualifier{morrison}{Perform the correction using a linear interpolation
%                       to the points by Morrison and Stephenson, 2004,
%                       Journal for the History of Astronomy 35, 327.
%                       Not good after 2000.}
%}
%\description
%
% This function converts a (Modified) Julian Date that is given in
% TT into UT1, i.e., the timescale that defines the earth rotation angle.
%
% By default, for years before 2005 the function uses the polynomial
% expressions for TT-UT1 published by F. Espenak and J. Meeus, 2006,
% Five Millenium Canon of Solar Eclipses: -1999 to +3000,
% NASA TP-2006-214141, for years after 2005 the polynomials published
% by Espenak at http://www.eclipsewise.com/help/deltatpoly2014.html are
% used.
%
% Alternatively, a linar interpolation in the solar eclipse derived
% data by Morrison and Stephenson (2000) can be used (see morrison
% qualifier).
%
% The uncertainty of the result is 1s or less from 1780 until 1830,
% less than 1s from 1830-2000. It increases quadratically backwards
% from 1820, sigma ~32((year-1820)/100)^2.
%
% This routine is array safe.
%
% Note: Because of the uncertainty in TT-UT1, the same
% polynomial is used for the TT->UT1 and UT1->TT conversion,
% and therefore it generally is NOT true that JD==ut12tt(tt2ut1(JD)).
%
%\seealso{ut12tt,utc2tai,tai2utc,tai2tt}
%!%-
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    % obtain UT1-TT
    variable deltat;
    if (qualifier_exists("mjd")) {
	deltat=ut12tt(JD;mjd,deltat);
    } else {
	deltat=ut12tt(JD;deltat);
    }
    
    if (qualifier_exists("deltat")) {
        % return TT-UT1
	return -deltat;
    }
    
    return JD+deltat/86400.;
}

%
% Code from src/slang/date/utc2tai.sl
%
private variable tai2utc_data;

private define read_leapfile(leapfile) {
    % helper function to read a leapfile
    % the argument is the file to read.
    
    variable statret=stat_file(leapfile);
    if (statret==NULL) {
	throw IOError,sprintf("%s: problem with leapfile %s or leapfile not found: Errno %i: %s\n",
	                      _function_name(),leapfile,errno,errno_string(errno));
    }
    variable refjd={};
    variable diffsec={};
    variable mjdoff={};
    variable jdoff={};
    variable rate={};
    % the leapfile is a fixed format file, it is more efficient to read it
    % directly rather than using the readascii function
    variable leap=fopen(leapfile,"r");
    if (leap==NULL) {
	throw IOError, sprintf("%s: error opening leapfie %s: Errno %i: %s\n",
	                       _function_name(),leapfile,errno,errno_string(errno));
    }
    while (not feof(leap)) {
	variable line;
	variable num=fgets(&line,leap);
	if (num>0) {
	    list_append(refjd,atof(substr(line,18,9)));
	    list_append(diffsec,atof(substr(line,37,12)));
	    list_append(mjdoff,atof(substr(line,60,7)));
	    list_append(rate,atof(substr(line,70,10)));
	}
    }
    ()=fclose(leap);
    mjdoff=list_to_array(mjdoff),
    refjd=list_to_array(refjd),
    tai2utc_data=struct{
	refjd=refjd,
	refmjd=jd2mjd(refjd),
	diffsec=list_to_array(diffsec),
	mjdoff=mjdoff,
	jdoff=mjd2jd(mjdoff),
	rate=list_to_array(rate)
    };
}

define utc2tai()
%!%+
%\function{utc2tai}
%\synopsis{Convert a (M)JD in civil time (UTC) into International Atomic Time}
%\usage{tai=utc2tai(JD)}
%\description
%This routine converts a Julian Date that is in civil time (Coordinated
%Universal Time, UTC) into International Atomic Time (TAI) by applying
%the relevant correction for leap seconds. This routine is array safe.
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the TAI-UTC in seconds, rather than the corrected (M)JD.}
%\qualifier{leapfile}{path to a file containing the leap second
%          information. This file is available, e.g., from
%          ftp://maia.usno.navy.mil/ser7/tai-utc.dat
%          Most astronomical software systems distribute this file. The default
%          is to look for it in getenv("LHEASOFT")/refdata}
%}
%\seealso{utc2tai,tt2tai,tai2tt}
%!%-
{
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    if ((not __is_initialized(&tai2utc_data)) or qualifier_exists("leapfile")) {
	read_leapfile(qualifier("leapfile",getenv("LHEASOFT")+"/refdata/tai-utc.dat"));
    }

    variable ndx;
    variable diff={};
    variable jj;

    if (qualifier_exists("mjd")) {
	foreach jj (JD) {
	    if (jj<tai2utc_data.refmjd[0]) {
		throw UsageError,sprintf("%s: TAI is only defined after MJD %15.1f\n",
		                         _function_name(),tai2utc_data.refmjd[0]);
	    }
	    ndx=(where(tai2utc_data.refmjd<=jj))[-1];
	    list_append(diff,tai2utc_data.diffsec[ndx]+ (jj-tai2utc_data.mjdoff[ndx])*tai2utc_data.rate[ndx]);
	}
    } else {
	foreach jj (JD) {
	    if (jj<tai2utc_data.refjd[0]) {
		throw UsageError,sprintf("%s: TAI is only defined after JD %15.1f\n",
		                         _function_name(),tai2utc_data.refjd[0]);
	    }
	    ndx=(where(tai2utc_data.refjd<=jj))[-1];
	    list_append(diff,tai2utc_data.diffsec[ndx]+ (jj-tai2utc_data.jdoff[ndx])*tai2utc_data.rate[ndx]);
	}
    }

    if (typeof(JD)==Array_Type) {
	diff=list_to_array(diff);
    } else {
	diff=diff[0];
    }

    if (qualifier_exists("deltat")) {
	return diff;
    }
    return JD+diff/86400.0;
}

define tai2utc()
%!%+
%\function{tai2utc}
%\synopsis{Converts a (M)JD in International Atomic Time into civil time (UTC)}
%\usage{utc=tai2utc(JD)}
%\description
%This routine converts a Julian Date that is given in international
%atomic time into Coordinated Universal Time (UTC) by applying the
%relevant correction for leap seconds. This routine is array safe.
%
%NOTE: This function is only approximately correct. Within about a
%minute of a leap second this routine can be off by as much as
%one second.
%\qualifiers{
%\qualifier{mjd}{The argument is given in MJD, not in JD.}
%\qualifier{deltat}{Return the UTC-TAI in seconds, rather than the corrected (M)JD.}
%\qualifier{leapfile}{Path to a file containing the leap second
%          information. This file is available, e.g., from
%          ftp://maia.usno.navy.mil/ser7/tai-utc.dat
%          Most astronomical software systems distribute this file. The default
%          is getenv("LHEASOFT")/refdata}
%}
%\seealso{utc2tai,tt2tai,tai2tt}
%!%-
{
    variable JD;
    switch(_NARGS)
    {
	case 1: JD=();
    }
    {
	help(_function_name()); return;
    }

    if ((not __is_initialized(&tai2utc_data)) or qualifier_exists("leapfile")) {
	read_leapfile(qualifier("leapfile",getenv("LHEASOFT")+"/refdata/tai-utc.dat"));
    }

    variable diff;
    if(qualifier_exists("mjd")) {
	diff=utc2tai(JD;mjd,deltat);
    } else {
	diff=utc2tai(JD;deltat);
    }

    if (qualifier_exists("deltat")) {
      return -1.*diff;
    }

    % FIXME: the result is wrong if we are close to a leap second
    return JD-diff/86400.;
}

%
% Code from src/slang/astronomy/v_relativistic_to_optical.sl
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define v_relativistic_to_optical ()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{v_relativistic_to_optical}
%\synopsis{Convert relativistic velocity to optical velocity.}
%\usage{v_relativistic_to_optical(V_rel);}
%\description
%    Calculate the optical velocity definition from a given relativistic velocity.
%    Velocities must be given in km/s.
%!%-
{
   variable V_rel;
   switch(_NARGS)
     { case 1: V_rel = (); }
     { help(_function_name()); return; }
     %  Go from relativistic velocity definition to optical velocity...
     variable c1e5=Const_c/1e5;
   variable gamma = 1.0/sqrt( (1.0 - (V_rel/c1e5))*(1.0+(V_rel/c1e5)) );
   return c1e5 * ( gamma*(1.0 + (V_rel/c1e5)) - 1.0 );
 };
 
%
% Code from src/slang/astronomy/vacuum_to_air.sl
%
define vacuum_to_air()
%!%+
%\function{vacuum_to_air}
%\synopsis{Convert vacuum wavelengths to air wavelengths}
%\usage{Double_Type l_a[] = vacuum_to_air(Double_Type l_v[])}
%\description
%    Convert vacuum wavelengths l_v (in Angstroem) to air wavelengths l_a (in Angstroem)
%    for dry air at 15 degrees Celsius, 101.325 Pa, and with 450 parts per million CO2
%    content. Valid in the wavelength range between 2300 to 17000 Angstroem.
%\notes
%    According to Equation 1 in Ciddor 1996, Applied Optics, 35, 1566:
%       1e8*(l_v-l_a)/l_a = n-1 = k1/(k0-l_v^(-2)) + k3/(k2-l_v^(-2))
%    => l_a = l_v / ( k1*1e-8/(k0-l_v^(-2)) + k3*1e-8/(k2-l_v^(-2)) + 1 )
%       with k0 = 238.0185 * 1e-8 / Angstroem^2,
%            k1 = 5792105  * 1e-8 / Angstroem^2,
%            k2 = 57.362   * 1e-8 / Angstroem^2,
%            k3 = 167917   * 1e-8 / Angstroem^2.
%\qualifiers{
%\qualifier{verbose [=1]}{: Set to 0 to suppress warnings.}
%}
%\example
%    l_a = vacuum_to_air([2000,4000,8000,16000]);
%\seealso{air_to_vacuum}
%!%-
{
  if(_NARGS!=1)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable verbose = qualifier("verbose", 1);
  variable l_v = ();
  l_v = [l_v]; % cast to array
  if( min(l_v)<2300 or max(l_v)>17000 )
    if(verbose)
      vmessage("Warning in '%s': Wavelength outside of valid range 2300 < l_v < 17000 Angstroem!", _function_name());
  variable k0 = 2.380185e-06;
  variable k1 = 5.792105e-10;
  variable k2 = 5.7362e-07;
  variable k3 = 1.67917e-11;
  variable divisor = ( k1/(k0-l_v^(-2)) + k3/(k2-l_v^(-2)) + 1. );
  variable lo = 2300.;
  variable hi = 17000.;
  divisor[where(l_v < lo)] = ( k1/(k0-lo^(-2)) + k3/(k2-lo^(-2)) + 1. );
  divisor[where(l_v > hi)] = ( k1/(k0-hi^(-2)) + k3/(k2-hi^(-2)) + 1. );
  return l_v / divisor;
}

define air_to_vacuum()
%!%+
%\function{air_to_vacuum}
%\synopsis{Convert air wavelengths to vacuum wavelengths}
%\usage{Double_Type l_v[] = air_to_vacuum(Double_Type l_a[])}
%\description
%    Convert air wavelengths l_a (in Angstroem) to vacuum wavelengths l_v (in Angstroem)
%    for dry air at 15 degrees Celsius, 101.325 Pa, and with 450 parts per million CO2
%    content. Valid in the wavelength range between 2300 to 17000 Angstroem.
%\notes
%    According to Equation 1 in Ciddor 1996, Applied Optics, 35, 1566:
%       1e8*(l_v-l_a)/l_a = n-1 = k1/(k0-l_v^(-2)) + k3/(k2-l_v^(-2))
%    => l_a = l_v / ( k1*1e-8/(k0-l_v^(-2)) + k3*1e-8/(k2-l_v^(-2)) + 1 )
%       with k0 = 238.0185 * 1e-8 / Angstroem^2,
%            k1 = 5792105  * 1e-8 / Angstroem^2,
%            k2 = 57.362   * 1e-8 / Angstroem^2,
%            k3 = 167917   * 1e-8 / Angstroem^2.
%    This function is not analytically solvable for l_v. To zeroth order,
%    however, l_v ~ l_a so that
%       l_v ~ l_a * ( k1*1e-8/(k0-1/l_a^2) + k3*1e-8/(k2-1/l_a^2) + 1 ).
%\qualifiers{
%\qualifier{verbose [=1]}{: Set to 0 to suppress warnings.}
%}
%\example
%    l_v = [2000,4000,8000,16000];
%    print(l_v - air_to_vacuum(vacuum_to_air(l_v)));
%\seealso{vacuum_to_air}
%!%-
{
  if(_NARGS!=1)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable verbose = qualifier("verbose", 1);
  variable l_a = ();
  l_a = [l_a]; % cast to array
  if( l_a[0]<2300 or l_a[-1]>17000 )
    if(verbose)
      vmessage("Warning in '%s': Wavelength outside of valid range 2300 < l_a < 17000 Angstroem!", _function_name());
  variable k0 = 2.380185e-06;
  variable k1 = 5.792105e-10;
  variable k2 = 5.7362e-07;
  variable k3 = 1.67917e-11;
  variable factor = ( k1/(k0-l_a^(-2)) + k3/(k2-l_a^(-2)) + 1. );
  variable lo = 2300.;
  variable hi = 17000.;
  factor[where(l_a < lo)] = ( k1/(k0-lo^(-2)) + k3/(k2-lo^(-2)) + 1. );
  factor[where(l_a > hi)] = ( k1/(k0-hi^(-2)) + k3/(k2-hi^(-2)) + 1. );
  return l_a * factor;
}

%
% Code from src/slang/astronomy/velocity_wrt_LSR.sl
%
define radial_velocity_LSR()
%!%+
%\function{radial_velocity_LSR}
%\synopsis{computes the radial velocity component of the local standard of rest}
%\usage{Double_Type radial_velocity_LSR(Doule_Type RA, dec, MJD)
%\altusage{Double_Type radial_velocity_LSR(Doule_Type az, el, MJD, longw, lat)}
%}
%\description
%    right ascension and declination, Modified Julian Date
%    azimuth, elevation, geographic longitude (west), geographic lattitude
%!%-
{ % after SRT/geom.java
  variable RA, dec, az, el, MJD, lat, longw;
  switch(_NARGS)
  { case 3: (RA, dec, MJD) = (); }
  { case 5: (az, el, MJD, lat, longw) = ();
            (RA, dec) = RAdec_from_AzEl(az, el, MJD,  longw, lat);
  }
  { help(_function_name()); return; }

  message("THIS FUNCTION DOES NOT WORK YET!");
  message("Use at your own risk...");
  
  variable cos_dec = cos(dec);
  variable sin_dec = sin(dec);
  variable cos_RA  = cos(RA);
  variable sin_RA  = sin(RA);
                                     % sun 20km/s towards ra=18h dec=30.0
  variable x = 0.;           % 20. * cos(18.*PI/12) * cos(30.*PI/180);
  variable y = -17.3205081;  % 20. * sin(18.*PI/12) * cos(30.*PI/180);
  variable z = 10.;          % 20. * sin(30.*PI180);
  variable vsun = -x * cos_RA * cos_dec - y * sin_RA * cos_dec - z * sin_dec;

  x = cos_dec * cos_RA;
  y = cos_dec * sin_RA;
  z = sin_dec;
  (y, z) = (y * cos(23.5*PI/180)  +  z * sin(23.5*PI/180),
	    z * cos(23.5*PI/180)  -  y * sin(23.5*PI/180));
  variable soulat = atan2(z, hypot(x, y));
  variable soulong = atan2(y, x);
  variable sunlong = ((MJD-40587.)*86400.*360./(365.25*86400.) + 280.0)*PI/180;  % long=280 day 1
  return vsun - 30. * cos(soulat) * sin(sunlong - soulong);
}

%
% Code from src/misc/math/voigt_profile.sl
%
#ifeval _isis_version < 20000

%%%%%%%%%%%%%%%%%%%%%
define voigt_profile()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{voigt_profile}
%\synopsis{computes a normalized Voigt profile}
%\usage{Double_Type voigt_profile(Double_Type x, [[N,] x0,] sigma, gamma)}
%\description
%    The Voigt profile V is the convolution
%    of a Gaussian profile (with width \code{sigma>0})\n
%       \code{G(x; sigma)  =  exp{-x^2/(2 sigma^2)} / [sqrt(2 pi) sigma]}\n
%    and a Lorentzian profile (with width gamma>0)\n
%       \code{L(x; gamma)  =  gamma/pi / (x^2 + gamma^2)}\n
%    i.e.:\n
%       \code{V(x; sigma, gamma)  =  int G(t; sigma) * L(x-t; gamma) dt}
%
%    The \code{voigt_profile} function computes\n
%       \code{N * V(x-x0; sigma, gamma)}\n
%    i.e., \code{x0} is the center of the Voigt profile (default: 0)
%    and \code{N} is its normalization (default: 1), i.e.\n
%       \code{int voigt_profile(x, N, x0, sigma, gamma) dx  =  N} .
%
%    Do not confuse this "Voigt profile" with the "Voigt function H(a, u)".
%\seealso{voigt, get_cfun2}
%!%-
{
  variable x, N=1, x0=NULL, sigma, gamma;
  switch(_NARGS)
  { case 3: (x,        sigma, gamma) = (); }
  { case 4: (x,    x0, sigma, gamma) = (); }
  { case 5: (x, N, x0, sigma, gamma) = (); }
  { help(_function_name()); return; }
  if(x0==NULL)  x0 = 0;

  variable IVIN = Isis_Voigt_Is_Normalized;
  Isis_Voigt_Is_Normalized = 1;
   variable dx = 1-_min(min(x), x0);
   x0 += dx;
   variable y = get_cfun2("voigt", Const_keV_A/(x+dx), [N, x0, 4*PI*gamma, sqrt(2)*Const_c/1e5*sigma/x0]);
  Isis_Voigt_Is_Normalized = IVIN;
  return y;
}

#endif

%
% Code from src/fitting/fit-functions/vradbinary.sl
%
%%%%%%%%%%%%%%%%%%%%%
define vradbinary_fit(bin_lo, bin_hi, par)
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{vradbinary (fit-function)}
%\synopsis{describes the radial velocity in a binary system}
%\description
%    Note that only bin_lo is taken into account.
%\seealso{radial_velocity_binary}
%!%-
{
  return radial_velocity_binary(bin_lo; v0=par[0], K=par[1], P=par[2], T0=par[3], e=par[4], omega=par[5], degrees);
}

%%%%%%%%%%%%%%%%%%%%%%%%%
define vradbinary_default(i)
%%%%%%%%%%%%%%%%%%%%%%%%%
{
  switch(i)
  { case 0: return (    0, 0, -1000,    1000); }  % v0
  { case 1: return (   10, 0,     0,    1000); }  % K
  { case 2: return (   10, 0,  0.01,    1000); }  % P
  { case 3: return (51544, 0,     0, 2500000); }  % T0
  { case 4: return (    0, 0,     0,       1); }  % e
  { case 5: return (    0, 1,     0,     360); }  % omega
}

add_slang_function("vradbinary", ["v0 [km/s]", "K [km/s]", "P [days]", "T0 [days]", "e", "omega [degrees]"]);
set_param_default_hook("vradbinary", &vradbinary_default);

%
% Code from src/slang/astronomy/vsop87.sl
%
private variable vsop87_data;

define vsop87()
%!%+
%\function{vsop87}
%\synopsis{calculates high precision planetary positions using the VSOP87 theory}
%\usage{Struct_Type vsop87(JD,planet;qualifiers)}
%\description
% This routine implements the planetary theory VSOP87 of
% Bretagnon and Francou (1988, A&A 202, 309).
% The routine can return high precision planetary positions for all planets
% for a given (array) of times (in TT).
%
% Depending on the qualifiers the routine can return rectangular coordinates
% (X,Y,Z) or ecliptical coordinates (longitude, latitude, distance) in a
% heliocentric coordinate system either for epoch J2000 or for the epoch
% of date. In addition, elliptical orbital elements and standard orbital
% elements can also be calculated.
%
% By default the routine returns rectangular heliocentric coordinates in a
% structure with elements x, y, z, vx, vy, vz.
% Here x, y, z are in an ecliptical coordinate system for the epoch J2000,
% where the x-axis is in the direction of the vernal equinox, the z-axis
% points to the ecliptical North pole, and the y-axis forms a right handed
% coordinate system with the x- and z-axes. The default units are AU and AU/day.
%
% For heliocentric coordinates the planet can have the values
% "Mercury", "Venus", "Earth", "Mars", "Jupiter", "Saturn",
% "Uranus", and "Neptune". For heliocentric rectangular coordinates
% the value "Earth-Moon" will return the position of the
% Earth-Moon barycenter.
%
% In the case of solar system barycentric data, it is not possible to
% obtain the elements of the Earth-Moon barycenter, only the coordinates
% of the Earth can be queried. In addition, for this set of
% coordinates it is also possible to request the rectangular
% position of the Sun.
%
% For the years 1900-2100 interval the precision of the positions
% is at the level of a few milli-arcseconds, with the exception of
% Saturn, for which a precision of "only" 0.1'' is reached.
%
% The precision is better than 1" for the following time intervals
% around AD2000:
% Mercury, Venus, Earth-Moon Barycenter, Mars: 4000 years
% Jupiter, Saturn: 2000 years
% Uranus, Neptune: 6000 years
%
% The time argument to the routine is the Julian Date in Terrestrial Time (TT),
% which corresponds to the international atomic time TAI+32.184s.
%
% This is a powerful routine with a large number of options. A set of helper
% functions provide simpler interfaces.
%
%\qualifiers{
% \qualifier{mjd}{The argument is in MJD(TT), not in JD(TT)}
% \qualifier{tdb}{The argument is in TDB, not TT}
% \qualifier{barycentric}{Return coordinates with respect to the solar system barycenter rather
%                    than heliocentric coordinates. This implies that coordinates are for J2000.}
% \qualifier{spherical}{Return spherical coordinates in the ecliptical coordinate system.
%                       Only possible for heliocentric data.
%                       In this case the structure returned contains the elements
%                       lat, lon, r, latdot, londot, rdot, where lat and lon are the ecliptical
%                       latitude and longitude (in rad), r is the distance in AU, and
%                       latdot, londot, rdot are their time derivatives (in rad/day or
%                       AU/day, respectively.)}
% \qualifier{mean_equinox}{Coordinates are given in the frame defined by the
%                   mean equinox and ecliptic of the date. In VSOP87, this
%                   system is derived from the J2000 one by pure precessional
%                   motion. The default is to return J2000 coordinates.}
% \qualifier{elements}{Return the orbital elements. The structure contains
%                   the following tags:
%                      a: semi-major axis (in AU)
%                      l: mean longitude (rad)
%                      k=e cos(p): elliptical element k
%                      h=e sin(p): elliptical element h
%                      q=sin(g) cos(omega): elliptical element q
%                      p=sin(g) sin(omega): elliptical element p
%                      e: orbital eccentricity
%                      plon: perihelion longitude (in rad)
%                      g: semi inclination (rad)
%                      i: inclination (rad)
%                      omega: longitude of the ascending node (also called G)
%                 }
% \qualifier{deg}{All angle quantities to be returned are in deg, or deg/day.}
% \qualifier{mks}{All distances returned by the function are given in meters or m/s, depending
%                 on the deg qualifier, angles are returned in rad and rad/s or deg and deg/s.}
% \qualifier{cgs}{Dito, but use centimeters instead of meters.}
% \qualifier{vsopfile}{Path to file containing the VSOP quantities; only necessary if
%                 this is a non-standard installation of isisscripts. This qualifier is
%                 only used if the vsop87 data have not yet been read.
%                 By default, the file with the coefficients is searched for in the
%                 directory pointed to by the environment variable ISISSCRIPTS_REFPATH.}
% \qualifier{forcearray}{Force the tags in the returned struct to be arrays. The default is
%                 that they are arrays only if JD is an array.}
%}
%\seealso{planetpos,jpl_eph,jpl_eph_vec}
%!%-
{
variable jdin,planet;
  switch(_NARGS)
  {
   case 2: (jdin,planet) = ();
  }
  { help(_function_name()); return; }

  % protect the JD argument from being changed on the fly
  variable JD=@jdin;
  
  if (qualifier_exists("mjd")) {
     JD=mjd2jd(JD);
  }

  if (qualifier_exists("tdb")) {
      JD=tdb2tt(JD); % use Bretagnon & Francou conversion to TT
  }
  
  if (qualifier_exists("equatorial")) {
     if (qualifier_exists("mean_equinox") or qualifier_exists("spherical")) {
	 throw UsageError,sprintf("%s: Only J2000 allowed for equatorial coordinates\n",_function_name());
     }
  }
    
  % type:
  % 0 -- VSOP87  Heliocentric elliptic elements J2000
  % 1 -- VSOP87A Heliocentric rectangular variables J2000
  % 2 -- VSOP87B Heliocentric spherical variables J2000
  % 3 -- VSOP87C Heliocentric rectangular variables of date
  % 4 -- VSOP87D Heliocentric spherical variables of date
  % 5 -- VSOP87E Barycentric rectangular variables J2000

  variable type=1;
  
  % figure out what theory to use
  if (qualifier_exists("elements")) {
      type=0;
  } else {
      if (qualifier_exists("barycentric")) {
	  type=5;
	  % spherical are handled later
      } else {
	  % heliocentric coordinates
	  if (qualifier_exists("spherical")) {
	      type=2;
	      if (qualifier_exists("mean_equinox")) {
		  type=4;
	      } 
	  } else {
	      % cartesian coordinates
	      type=1;
	      if (qualifier_exists("mean_equinox")) {
		  type=3;
	      }
	  }
      }
  }


  % translation of body code to planet
  variable planets=["","MERCURY","VENUS","EARTH","MARS","JUPITER","SATURN","URANUS","NEPTUNE",""];

  switch(type)
  { case 0: planets[3]="EARTH-MOON";}
  { case 1: planets[9]="EARTH-MOON";}
  { case 5: planets[9]="SUN";}

  variable bcode=where(planets==strup(planet));

  if (length(bcode)==0) {
      variable i;
      variable erstr="";
      variable comma="";
      _for i(1,length(planets)-1,1) {
	  if (planets[i]!="") {
	      erstr+=comma+planets[i];
	  }
	  comma=", ";
      }
      throw UsageError,sprintf("%s: planet qualifier must be ONE of %s\n",_function_name(),erstr);
  }
  bcode=bcode[0];

  %
  % variables of the theory (will be used later as the structure elements)
  % further 
  variable varname;
  variable varnum; % number of variables in the theory
  if (type==0) {
      varnum=6;
      varname={"a","l","k","h","q","p"};
  } else {
      varnum=3;
      if (type==1 or type==3 or type==5) {
	  varname={"x","y","z","vx","vy","vz"};
      } else {
	  varname={"lon","lat","r","londot","latdot","rdot"};
      }
  }
  
  ifnot (__is_initialized(&vsop87_data)) {
      variable searchpath=getenv("ISISSCRIPTS_REFPATH");
      if (searchpath==NULL and qualifier_exists("vsopfile")==0) {
	  throw UsageError,sprintf("%s: vsopfile qualifier not given and ISISSCRIPTS_REFPATH not found\n",_function_name());
      }
      if (searchpath==NULL) {
	  searchpath=".";
      }
      vsop87_data=fits_read_table(qualifier("vsopfile",searchpath+"/vsop87.fits.gz[VSOP87]"));
  }

  variable century=365250.;

  variable T=(JD-2451545.0)/century;

  variable tlen=length(T);

  variable tmpres=Double_Type[tlen,varnum];
  variable derivres=Double_Type[tlen,varnum];

  variable ii,jj;

  _for ii(0,varnum-1,1) {
      variable ndx=where(vsop87_data.type==type and vsop87_data.body==bcode and vsop87_data.ic==ii+1);
      if (length(ndx)>0) {
	  _for jj(0,tlen-1,1) {
	      variable co=cos(vsop87_data.b[ndx]+vsop87_data.c[ndx]*T[jj]);
	      variable ndx1;
	      variable ndx0=where(vsop87_data.alpha[ndx]==0,&ndx1);
	      if (length(ndx0)>0) {
		  tmpres[jj,ii]=sum(vsop87_data.a[ndx[ndx0]]*co[ndx0]);
	      }
	      if (length(ndx1)>0) {
 	          tmpres[jj,ii]+=sum(T[jj]^vsop87_data.alpha[ndx[ndx1]]*vsop87_data.a[ndx[ndx1]]*co[ndx1]);
	      }
	      % derivatives
	      if (type>0) {
		  variable si=sin(vsop87_data.b[ndx]+vsop87_data.c[ndx]*T[jj]);
		  if (length(ndx0)>0) {
		      derivres[jj,ii]=sum(-vsop87_data.a[ndx[ndx0]]*vsop87_data.c[ndx[ndx0]]*si[ndx0]);
		  }
		  if (length(ndx1)>0) {
		      derivres[jj,ii]+=sum(vsop87_data.alpha[ndx[ndx1]]*T[jj]^(vsop87_data.alpha[ndx[ndx1]]-1)*vsop87_data.a[ndx[ndx1]]*co[ndx1]-
		                          T[jj]^vsop87_data.alpha[ndx[ndx1]]*vsop87_data.a[ndx[ndx1]]*vsop87_data.c[ndx[ndx1]]*si[ndx1]);
		  }
	      }
	  }
      }
  }

  % normalize derivative to per day
  if (type>0) {
      derivres=derivres/century;
  }

  % normalize longitudes
  if (type==0 or type==2 or type==4) {
      ndx=(type==0) ? 1 : 0;
      tmpres[*,ndx]=tmpres[*,ndx] mod (2*PI);
      variable nn=where(tmpres[*,ndx]<0);
      if (length(nn)>0) {
	  tmpres[nn,ndx]+=2*PI;
      }
  }

  %

  %
  % generate structure for returning the values
  %

  % missing structure tags for elliptical elements
  if (type==0) {
      list_append(varname,"e");
      list_append(varname,"plon");
      list_append(varname,"g");
      list_append(varname,"i");
      list_append(varname,"omega");
  }
  varname=list_to_array(varname);
  
  % generate the structure
  variable retval=@Struct_Type(varname);
  if (type==0) {
      if (typeof(JD)==Array_Type or qualifier_exists("forcearray")) {
	  _for i(0,5,1) {
	      set_struct_field(retval,varname[i],tmpres[*,i]);
	  }
      } else {
	  _for i(0,5,1) {
	      set_struct_field(retval,varname[i],tmpres[0,i]);
	  }
      }
      retval.plon=atan2(retval.h,retval.k);
      retval.omega=atan2(retval.plon,retval.q);
      % hoping that the cosine isn't 0 somewhere...
      retval.e=retval.k/cos(retval.plon); 
      retval.g=retval.q/cos(retval.omega);
      retval.i=2*retval.g;
  } else {
      if (typeof(JD)==Array_Type or qualifier_exists("forcearray")) {
	  _for i(0,2,1) {
	      set_struct_field(retval,varname[i],tmpres[*,i]);
	      set_struct_field(retval,varname[i+3],derivres[*,i]);
	  }
      } else {
	  _for i(0,2,1) {
	      set_struct_field(retval,varname[i],tmpres[0,i]);
	      set_struct_field(retval,varname[i+3],derivres[0,i]);
	  }
      }
  }

  variable tag;
  if (qualifier_exists("deg")) {
      foreach tag(["l","p","plon","g","i","omega","lat","lon","latdot","londot"]) {
	  if (struct_field_exists(retval,tag)) {
	      set_struct_field(retval,tag,get_struct_field(retval,tag)*180./PI);
	  }
      }
  }
  if (qualifier_exists("mks") or qualifier_exists("cgs")) {
      variable au=149597870691.; % meters, DE405 value, per Kaplan, 2005, USNO Circular 179
      if (qualifier_exists("cgs")) {
	  au*=100.;
      }
      % convert lengths
      foreach tag(["x","y","z","vx","vy","vz","r","rdot","a"]) {
	  if (struct_field_exists(retval,tag)) {
	      set_struct_field(retval,tag,get_struct_field(retval,tag)*au);
	  }
      }
      % convert rates
      foreach tag(["vx","vy","vz","rdot"]) {
	  if (struct_field_exists(retval,tag)) {
	      set_struct_field(retval,tag,get_struct_field(retval,tag)/86400.0);
	  }
      }
  }

  return retval;
}


%
% Code from src/misc/image/wcs_axes_plot.sl
%

% only activate if module wcsfuns could be loaded
#ifeval __get_reference("wcsfuns_project")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

private define get_tic_pos_dezimal(x)
{
  variable n = qualifier("n",5); % preferred number of major tics
  
  variable coverage  = abs(max(x)-min(x));
  variable fac       = 10^floor(log10(coverage ));
  variable steps     = [  1, 2, 5, 10, 20, 50];
  variable min_steps = [0.2, 1, 1,  2, 10, 10];
  do { % allow the situation of many tic marks:
    steps *= 0.1; min_steps *= 0.1;
    variable nr = coverage/(steps*fac);
    variable i = where_min (abs(n -  nr) );
  } while (i == 0);

  variable major = min(x)-(min(x) mod (steps[i]*fac)) + [-1 : 1+ceil(nr[i])]*(steps[i]*fac);
  variable minor = min(x)-(min(x) mod (min_steps[i]*fac))
    + [-1 : 1+ceil( coverage/(min_steps[i]*fac) )]*(min_steps[i]*fac);
  return major, minor;
}

#ifnexists WCSaxis_Type
typedef struct { ra , dec, x, y, label, major, maj_ra, maj_dec, minor, min_ra, min_dec} WCSaxis_Type;
#endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define wcs_axes_plot ()
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{wcs_axes_plot}
%\synopsis{gets WCS coordinates of axes and returns xfig plot object}
%\usage{xfig_plot_object = wcs_axes_plot (String_Type image_filename); }
%\altusage{xfig_plot_object = wcs_axes_plot (image, wcs); }
%\qualifiers{
%\qualifier{width}{set plot width}
%\qualifier{height}{[=14] set plot height, unless width and height are both
%                     set, the other one is calculated based on the
%                     format of the image (currently CDELT1 != CDELT2
%                     is not taken into account, but only #pixels)}
%\qualifier{x1label}{[=WCS of major] set ticlabels of x1-axis}
%\qualifier{x2label}{[=0] set ticlabels of x2-axis}
%\qualifier{y1label}{[=WCS of major] set ticlabels of y1-axis}
%\qualifier{y2label}{[=0] set ticlabels of y2-axis}
%\qualifier{axis_color}{[="white"] set the color of the axes}
%\qualifier{x1major}{set WCS values for major tics of x1-axis.
%                     similar qualifiers exist for x2, y1, and y2-axes.
%                     unless x1 and x2 are both set, the qualifier affects
%                     both x-axes. the same holds for the y-axes and the
%                     minor qualifiers.}
%\qualifier{x1minor}{set WCS values for minor tics of x1-axis.
%                     similar qualifiers exist for x2, y1, and y2 axes}
%\qualifier{return_axes}{set qualifier to additionally return
%                     data (Assoc_Type) including axes information.
%                     the returned object \code{wcsaxes} can be accessed
%                     with \code{wcsaxes["x1"]} (keys are ["x1","x2","y1","y2"])}
%\qualifier{info}{print the WCS (RA-DEC) ranges of the axes
%                     (helpful for setting minor and major by hand)}
%}
%\description
%    Creates an xfig plot object based on a sky image with WCS coordinates
%    and determines the major and minor tics as well as the tic labels.
%    The underlying coordinates in the xfig plot are pixels (with integers
%    corresponding to the pixel centers).
%    The tic marks of each axis correspond to the "leading" (most varying)
%    coordinate along this axis (see wcs.xtype). There can/will be problems
%    when poles (and similar things) are within the field of view.
%    It is possible (and recommended) to specify major and minor tics and
%    the ticlabels (e.g., in order to use sexagesimal labels) of each axis
%    using the available qualifiers.
%    The alternative usage allows to specify the image and the wcs
%    information separately. They can be given either as a string of the
%    filename or directly as the image (array) and the WCS struct. In the
%    latter case they can be modified before.
%
%    WARNINGS:
%     - check WCS coordinates (the wcsfuns module uses less keywords
%       than, e.g., ds9)
%     - when plotting sub-images (i.e., cutting the image before
%       calling this function), make sure that the WCS structure is
%       consistent (e.g., modified center pixel)
%\example
%    variable filename = "img.fits";
%    variable p = wcs_axes_plot (filename);
%    p.plot_png (log(img); cmap="ds9b");
%    p.render("img.pdf");
%
%    % if, e.g., WCS has to be modified/corrected:
%    variable wcs = fitswcs_get_img_wcs(filename);
%    wcs.cdelt[1] *= -1; % can be necessary if a keyword was not read
%    variable p = wcs_axes_plot ( filename, wcs );
%
%    % use sexagesimal axis labels (example for declination from 44° 8' - 44° 20')
%    variable dec_major = [8:20:2]; % arcmin
%    variable y1major = 44 + dec_major / 60.0;
%    variable y1minor = 44 + [8:20:1] / 60.0; % minor in 1' steps
%    variable dec_ticlabels = array_map (String_Type, &sprintf, "$44^\\circ\\,%02d'$", dec_major);
%    variable p = wcs_axes_plot (filename; y1major = y1major, y1minor = y1minor, y1label = dec_ticlabels);
%    
%!%-
{
  variable img, wcs;
  
  switch(_NARGS)
  { case 1 : (img)       = (); wcs = img;}
  { case 2 : (img,wcs)   = (); }
  { help(_function_name()); return; }

  variable info = qualifier_exists("info");
  variable clr  = qualifier("axis_color","white");

  variable x,y;
  if (typeof(img) == String_Type) {
    try {
      x = fits_read_key (img, "ZNAXIS1"); % check if x and y swapped?
      y = fits_read_key (img, "ZNAXIS2");
    }
    catch AnyError: { vmessage("%s: error: could not read image size (keywords ZNAXIS1, ZNAXIS2) from file\n\ttry alternative usage and provide image",_function_name); };
  }
  else {
    variable size = array_shape(img);
    x = size[1]; y = size[0];
  }
  
  if (typeof(wcs) == String_Type) {
    try { wcs = fitswcs_get_img_wcs(wcs); }
    catch AnyError: { vmessage("%s: error: could not read WCS from file\n\ttry alternative usage and provide wcs directly",_function_name); };
  }
  %%%%%%%%%% get plot format %%%%%%%%%%%%
  % (unless both width and height are specified, the format is
  % calculted using the image dimensions)
  % -> currently CDELT1 != CDELT2 is not taken into account
  variable pw = qualifier("width",  14);
  variable ph = qualifier("height", 14);
  if (qualifier_exists("width")){
      ifnot (qualifier_exists("height"))
      ph = pw / (x+1.0) * (y+1); % added 0.5 on both sides (pixel centers are integer)
  }
  else {
    pw = ph / (y+1.0) * (x+1);
  }
  
  variable p = xfig_plot_new (pw,ph);
  
  p.world(-0.5 ,x+0.5, -0.5, y+0.5);
  % added 0.5 on both sides (pixel centers are integer)  

  %%%%%%%%%%%%%%%%%%%%%%%%%%% determine wcs of coordinate axis
  variable x0 = Integer_Type[y+1];
  variable y0 = Integer_Type[x+1];
  variable x_arr = [0:x]-0.5;
  variable y_arr = [0:y]-0.5;

  variable axes = Assoc_Type[WCSaxis_Type];
  variable k;
  foreach k (["x1","x2","y1","y2"]) axes[k] = @WCSaxis_Type;
  
  ( axes["x1"].y, axes["x1"].x ) = ( y0-0.5   , x_arr    );
  ( axes["x2"].y, axes["x2"].x ) = ( y0+y+0.5 , x_arr    );
  ( axes["y1"].y, axes["y1"].x ) = ( y_arr    , x0-0.5   );
  ( axes["y2"].y, axes["y2"].x ) = ( y_arr    , x0+x+0.5 );
  
  foreach k (axes) using ("values") {
    (k.ra , k.dec) = wcsfuns_deproject (wcs, k.y , k.x ); k.ra = pos_modulo(k.ra,360);
  }
  if (info){
    message("WCS coordinates of the plot axes:");
    foreach k (["x1","x2","y1","y2"]) {
      vmessage ("%s\tWCS (from %s): %.4e - %.4e\n\tWCS (from %s): %.4e - %.4e",
		k, wcs.ctype[1], min_max(axes[k].ra), wcs.ctype[0], min_max(axes[k].dec) );
    }
  }
  
  %%% get values for minor and major tics (in RA DEC)
  variable xmajor, xminor;
  (xmajor, xminor) = get_tic_pos_dezimal ([axes["x1"].ra,  axes["x2"].ra ]; n = pw/2.8);
  variable x1major = qualifier("x1major", qualifier("x2major", xmajor ) );
  variable x2major = qualifier("x2major", qualifier("x1major", xmajor ) );
  variable xmajor_is_set = qualifier_exists("x1major") or qualifier_exists("x2major") ;
  variable x1minor = qualifier("x1minor", qualifier("x2minor", xmajor_is_set ? 0 : xminor ) );
  variable x2minor = qualifier("x2minor", qualifier("x1minor", xmajor_is_set ? 0 : xminor ) );
  axes["x1"].maj_ra = x1major;
  axes["x1"].min_ra = x1minor;
  axes["x2"].maj_ra = x1major;
  axes["x2"].min_ra = x1minor;
    
  variable ymajor, yminor;
  (ymajor, yminor) = get_tic_pos_dezimal ([axes["y1"].dec, axes["y2"].dec]; n = ph/2.5);
  variable y1major = qualifier("y1major", qualifier("y2major", ymajor ) );
  variable y2major = qualifier("y2major", qualifier("y1major", ymajor ) );
  variable ymajor_is_set = qualifier_exists("y1major") or qualifier_exists("y2major") ;
  variable y1minor = qualifier("y1minor", qualifier("y2minor", ymajor_is_set ? 0 : yminor ) );
  variable y2minor = qualifier("y2minor", qualifier("y1minor", ymajor_is_set ? 0 : yminor ) );
  axes["y1"].maj_dec = y1major;
  axes["y1"].min_dec = y1minor;
  axes["y2"].maj_dec = y1major;
  axes["y2"].min_dec = y1minor;

  
  %%% calculate x,y values of tic marks
  foreach k (["x1","x2"]) {
    axes[k].maj_dec = interpol (axes[k].maj_ra, axes[k].ra, axes[k].dec);
    axes[k].min_dec = interpol (axes[k].min_ra, axes[k].ra, axes[k].dec);
    
    ( ,axes[k].major) = wcsfuns_project (wcs, axes[k].maj_ra, axes[k].maj_dec);
    ( ,axes[k].minor) = wcsfuns_project (wcs, axes[k].min_ra, axes[k].min_dec);
  }

  foreach k (["y1","y2"]) {
    axes[k].maj_ra = interpol (axes[k].maj_dec, axes[k].dec, axes[k].ra);
    axes[k].min_ra = interpol (axes[k].min_dec, axes[k].dec, axes[k].ra);
    
    ( axes[k].major,) = wcsfuns_project (wcs, axes[k].maj_ra, axes[k].maj_dec);
    ( axes[k].minor,) = wcsfuns_project (wcs, axes[k].min_ra, axes[k].min_dec);
  }

  %%%%%%%%% set tic labels
  % here it is currently assumed that the most varying coordinate on the x-axes is
  % RA (or wcs.ctype[1]) and DEC (or wcs.ctype[0]) for the y-axes
  axes["x1"].label = qualifier("x1label", axes["x1"].maj_ra  );
  axes["x2"].label = qualifier("x2label", 0 );
  axes["y1"].label = qualifier("y1label", axes["y1"].maj_dec );
  axes["y2"].label = qualifier("y2label", 0 );
  
  %%%%%%%%% set axes parameters
  p.x1axis(; major = axes["x1"].major, minor=axes["x1"].minor, ticlabels=axes["x1"].label, color=clr);
  p.x2axis(; major = axes["x2"].major, minor=axes["x2"].minor, ticlabels=axes["x2"].label, color=clr);
  p.y1axis(; major = axes["y1"].major, minor=axes["y1"].minor, ticlabels=axes["y1"].label, color=clr);
  p.y2axis(; major = axes["y2"].major, minor=axes["y2"].minor, ticlabels=axes["y2"].label, color=clr);

  if (qualifier_exists("return_axes"))
    return p, axes;
  else
    return p;
}


#endif

#endif

%
% Code from src/slang/arrays/weighted_mean.sl
%
define weighted_mean()
%!%+
%\function{weighted_mean}
%\usage{Double_Type m = weighted_mean(x, w);
%\altusage{Double_Type m = weighted_mean(x; err=err);  % or weighted_mean(x, err; err)}
%}
%\description
%    \code{m = sum(w*x) / sum(w);}\n
%    \code{%} for Gaussian errors:\n
%    \code{w = 1/err^2;}
%!%-
{
  variable x, w=NULL, err=qualifier("err", NULL);
  switch(_NARGS)
  { case 1: x = (); }
  { case 2:
      if(qualifier_exists("err"))
      { if(err!=NULL)  vmessage("warning (%s): too many arguments -- qualifier-value for err ignored", _function_name()); 
        (x, err) = ();
      }
      else
        (x, w) = ();
  }
  { help(_function_name()); return; }

  if(err!=NULL)  w = 1/err^2;
  if(w==NULL)  w = 0*x + 1;

  return sum(w*x)/sum(w);
}

%
% Code from src/slang/arrays/where_max.sl
%
%%%%%%%%%%%%%%%%
define where_max()
%%%%%%%%%%%%%%%%
%!%+
%\function{where_max}
%\synopsis{finds the indices of an array where the values are maximal}
%\usage{Integer_Type i[] = where_max(Array_Type a);}
%\description
%    \code{i = where(a==max(a));  % =>  a[i] == max(a)}
%\seealso{where_min}
%!%-
{
  variable a;
  switch(_NARGS)
  { case 1: a = (); }
  { help(_function_name()); return; }

  if(length(a)==0)  return Integer_Type[0];
  variable m = where(a==max(a));
  if(length(m)==1)  return m[0];  else  return m;
}

%
% Code from src/slang/arrays/where_min.sl
%
define where_min()
%!%+
%\function{where_min}
%\synopsis{finds the indices of an array where the values are minimal}
%\usage{Integer_Type i[] = where_min(Array_Type a);}
%\description
%    \code{i = where(a==min(a));  % =>  a[i] == min(a)}
%\seealso{where_max}
%!%-
{
  variable a;
  switch(_NARGS)
  { case 1: a = (); }
  { help(_function_name()); return; }
 
  if(length(a)==0)  return Integer_Type[0];
  variable m = where(a==min(a));
  if(length(m)==1)  return m[0];  else  return m;
}

%
% Code from src/slang/compatibility/wherefirst.sl
%
#ifnexists wherefirst
define wherefirst (a)
     {
        variable i = where (a);
        if (length(i))
          return i[0];
        else
          return NULL;
     }
#endif
%
% Code from src/slang/compatibility/wherefirstmax.sl
%
#ifnexists wherefirstmax
public define wherefirstmax(a)
{
   return wherefirst (a==max(a));
}
#endif

%
% Code from src/slang/compatibility/wherefirstmin.sl
%
% compatibility for users not using the
% newest (or development) version of S-Lang
#ifnexists wherefirstmin
public define wherefirstmin(a)
{
   return wherefirst (a==min(a));
}
#endif

%
% Code from src/slang/compatibility/wherelast.sl
%
#ifnexists wherelast %{{{
define wherelast (a)
     {
        variable i = where (a);
        if (length(i))
          return i[-1];
        else
          return NULL;
     }
#endif
%}}}

%
% Code from src/slang/compatibility/wherelast_le.sl
%
#ifnexists wherelast_le %{{{
public define wherelast_le(a,b)
{
   return wherelast (a <= b); 
}
#endif
%}}}

%
% Code from src/sixte/write_simput_dummy_bkg.sl
%

% only activate if module fitswcs could be loaded
#ifeval __get_reference("fitwcs_get_img_wcs")!=NULL

private define get_arcmin_area(file){

   variable wcs = fitswcs_get_img_wcs(file);
   variable keys = fits_read_key_struct(file,"naxis1","naxis2");

   variable xarcmin = abs(wcs.cdelt[0]*keys.naxis1)*60.;
   variable yarcmin = abs(wcs.cdelt[1]*keys.naxis2)*60.;

   
   return xarcmin*yarcmin;
}

private define write_dummy_bkg_ctsmap(fsrc,fbkg){

   variable src = fits_read_img(fsrc);

   variable wcs = fitswcs_get_img_wcs(fsrc);

   fits_write_image(fbkg,"IMAGE",src*0+1.0);

   fitswcs_put_img_wcs(fbkg,wcs);
   
   
}


private define local_eval_simputfile(ff_par,Simput,RA,Dec,fbkg,Src_Name,srcflux){ 

   variable emin = qualifier("emin",0.5);
   variable emax = qualifier("emax",10.);
   
   variable simput_str = get_simputfile_struct(Simput,RA,Dec,srcflux;emin=emin,emax=emax);
   simput_str.Src_Name = Src_Name;
   simput_str.ImageFile = fbkg;
   simput_str.ISISFile = ff_par;
   
   simput_str.Elow = 0.1;
   simput_str.Eup = 15;
   simput_str.Estep = 0.001;
   
   if (qualifier_exists("overwrite") || (stat_file(simput_str.Simput)==NULL)) {
      eval_simputfile(simput_str);
   } else {
      vmessage("Skipping creating of SIMPUT File %s as it already exists",Simput);
   }
   
}

private define get_bkg_src_flux(fpar,fbkg){

   
   variable arcmin_area = qualifier("arcmin_area",1.);

   variable emin = qualifier("emin",0.5);
   variable emax = qualifier("emax",10.);
      
   load_par(fpar);
   variable flu = energyflux(emin,emax;cgs);
   variable img_area = get_arcmin_area(fbkg);

   return flu*img_area/arcmin_area;
}

%%%%%%%%%%%%%%%%%%%%%%%
define write_simput_dummy_bkg(fsimput,fsrc,fpar){
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{write_simput_dummy_bkg}
%\synopsis{plots a sky image, including an optional scale and/or grid}
%\usage{write_simput_dummy_bkg(fsimput,fsrc,fpar);}
%\description
%    fsimput:  Name of the ouput simput-File
%    fsrc:     Name of a SRC-FITS-Image (with FK5 WCS Header)
%    fpar:     ISIS-par File (with proper normalization) of the BKG
%\qualifiers{
%\qualifier{src_name}{["BKG"] name of the source in the SIMPUT file}
%\qualifier{overwrite}{if given, already exsiting simput files are overwritten}
%\qualifier{arcmin_area}{=[1.0] area in arcmin specifying the flux for the given spectrum }
%}
%\seealso{xfig_plot_new,fitswcs_get_img_wcs}
%!%-

   variable srcName = qualifier("src_name","BKG");
   
   variable fbkg = "dummy_tmp_simput_bkg.bkg";
   write_dummy_bkg_ctsmap(fsrc,fbkg);
   
   variable ra=0.0,dec=0.0;
%   if (fits_read_key(fbkg,"RADESYS") != "FK5"){
%      verror("wcs coords in %s not in FK5. Exiting ... ",fbkg);
%   }
   ra = fits_read_key(fbkg,"CRVAL1");
   dec = fits_read_key(fbkg,"CRVAL2");

   variable bkg_flux = get_bkg_src_flux(fpar,fbkg);
   local_eval_simputfile(fpar,fsimput,ra,dec,fbkg,srcName,bkg_flux;;__qualifiers());
   
   () = system("rm -f $fbkg"$);
}



#endif

%
% Code from src/slang/write_slurm_script.sl
%
%%%%%%%%%%%%%%%
define write_slurm_script(){
%%%%%%%%%%%%%%%
%!%+
%\function{write_slurm_script}
%\synopsis{writes a slurm job script, which can be submitted with squeue}
%\usage{write_slurm_script(file,jobname,walltime,cmds);}
%\qualifiers{
%\qualifier{queue}{set the queue (remeis by default)}
%\qualifier{silent}{do not show any output}
%\qualifier{serial}{instead of parallel, execute commands one after the other}
%\qualifier{option}{setting commands or environment variables before executing the actual commands}
%\qualifier{ntaks}{[=1] number of tasks per CPU}
%\qualifier{memory}{[=1000] MByte in memory allocated (--mem-per-cpu in slurm)}
%\qualifier{output}{output logfile name}
%\qualifier{error}{error logfile name}
%\qualifier{account}{if the selected queue needs a specific account set this qualifier accordingly}
%\qualifier{dir}{[=getcwd()] absolut path where the script should be run (default is cwd)}
%}
%\description
%    Writing a slurm job script, which can be submitted with squeue.
%    The function might yet not include all options and works best for
%    simple cases. The cmds variable is a String_Type array, with each
%    field being one command line. The 
%\example
%    % define your commands
%    variable cmds = 
%      [ "echo 123" , "echo 456" ]; 
%    % call the function
%    variable file = "slurmscript.slurm";
%    variable jobname = "script";
%    variable walltime = "00:30:00";
%    write_slurm_script(file,jobname,walltime,cmds);
%!%-

   variable file,name,walltime,cmds;
   switch (_NARGS)
   { case 4: (file,name,walltime,cmds) = (); }
   { help(_function_name); return; }

   variable fp = fopen(file,"w+");

   variable queue = qualifier("queue","remeis");
   
   () = fprintf(fp,"#!/bin/bash");
   () = fprintf(fp,"\n#SBATCH --partition %s",queue);
   if (qualifier_exists("account")) {
      fprintf(fp,"\n#SBATCH --account %s",qualifier("account","") );
   }
   () =  fprintf(fp,"\n#SBATCH --job-name %s",name);
   () = fprintf(fp,"\n#SBATCH --ntasks=%i",qualifier("ntasks",1));
   () = fprintf(fp,"\n#SBATCH --time %s",walltime);
   () = fprintf(fp,"\n#SBATCH --mem-per-cpu=%i",nint(qualifier("memory",1000.0)));
   () = fprintf(fp,"\n#SBATCH --output %s",qualifier("output",getcwd+name+".out-%a"));
   () = fprintf(fp,"\n#SBATCH --error %s",qualifier("error",getcwd+name+".err-%a"));
   if (not qualifier_exists("serial")){      
      () = fprintf(fp,"\n#SBATCH --array 0-%i",length(cmds)-1);
   }

   if (qualifier_exists("option")){
      () = fprintf(fp,"\n\n### general settings ###\n%s \n",qualifier("option"));
   }

   
   () = fprintf(fp,"\n\ncd %s \n",qualifier("dir",getcwd()));
   
   variable ii;

   %%% do we want to do it in parallel or in serial mode %%%
   if (qualifier_exists("serial")){      
      _for ii(0,length(cmds)-1){
	 () = fprintf(fp,"srun %s\n",cmds[ii]);
      }	   
   } else {
      _for ii(0,length(cmds)-1){
	 () = fprintf(fp,"\nCOMMAND[%i]=\"%s\"",ii,cmds[ii]);
      }	   
      () = fprintf(fp,"\n\nsrun ${COMMAND[$SLURM_ARRAY_TASK_ID]} \n");
   }
   fclose(fp);      


   ifnot (qualifier_exists("silent")){
      () = system("cat "+file);
   }
   
   return;
}

%
% Code from src/misc/radio/write_spix.sl
%
%%%%%%%%%%%%%%%%
define write_spix()
%%%%%%%%%%%%%%%%
%!%+
%\function{write_spix}
%\synopsis{writes the structure provided by make_spix in a FITS file}
%\usage{write_spix(Stuct_Type \code{spix}, String_Type \code{filename});}
%\seealso{make_spix, read_spix, plot_spix}
%!%-
{
  variable spix, filename;
  switch(_NARGS)
  { case 2: (spix, filename) = (); }
  { help(_function_name()); return; }


  variable spixm   = struct {
    spec_map       = spix.spec_map,
    lo_map         = spix.lo_map,
    hi_map_shifted = spix.hi_map_shifted,
    stdev          = spix.stdev,
    avg_lum        = spix.avg_lum,
    weights        = spix.weights,
  };
  
  variable keys = struct {
    avg_xshift    = spix.avg_shift[0],    % to be corrected
    avg_yshift    = spix.avg_shift[1],    % to be corrected
    shift_weight  = spix.shift_weight,    % to be corrected
    ra_px_center  = spix.ra_px_center,
    ra_steps      = spix.ra_steps,
    dec_px_center = spix.dec_px_center,
    dec_steps     = spix.dec_steps,
    major         = spix.major,
    minor         = spix.minor,
    posang        = spix.posang,
    source        = spix.source,
    date          = spix.date
  };

  fits_write_binary_table(filename, "spix", spixm, keys);
}

%
% Code from src/plot/x1axis.sl
%
%%%%%%%%%%%%%
define x1axis()
%%%%%%%%%%%%%
%!%+
%\function{x1axis}
%\synopsis{(re)draws a first x-axis}
%\usage{x1axis(Double_Type min_value, Double_Type max_value[, Double_Type step]);}
%\qualifiers{
%\qualifier{nsub}{[=0] number of minor tick marks within each major divison (nsub=0 => no minor ticks)}
%\qualifier{maji}{[=0.5] length of major tick marks, drawn inwards (in units of the character height)}
%\qualifier{majo}{[=0] length of major tick marks, drawn outwards (in units of the character height)}
%\qualifier{fmin}{[=0.5] length of minor tick marks (as fraction of the major tick marks)}
%\qualifier{disp}{[=0.5] displacement of baseline of tick labels to the axis (in units of the character height)}
%\qualifier{angle}{[=0] orientation of the text (in degrees)}
%\qualifier{color}{[=1] color of the axis}
%\qualifier{linewidth}{[=1] line width of the axis}
%\qualifier{opt}{[="N"/"LN"] pgaxis-options: "L" (log), "N" (numbers), "1" (force decimal), "2" (force EE)}
%\qualifier{yrel}{[=0] relative position of the x1axis (0=bottom, 1=top)}
%\qualifier{xmin}{[=xmin] absolute x-position of the axis}
%\qualifier{xmax}{[=xmax] absolute x-position of the axis}
%}
%\seealso{x2axis (and references therein)}
%!%-
{
  variable po = get_plot_options();
  variable min_value=po.xmin, max_value=po.xmax, step=0;
  switch(_NARGS)
  { case 0: ; }
  { case 1: step = (); }
  { case 2: (min_value, max_value) = (); }
  { case 3: (min_value, max_value, step) = (); }
  { help(_function_name()); return; }

  variable opt = "N";
  variable xmin = qualifier("xmin", po.xmin);
  variable xmax = qualifier("xmax", po.xmax);
  if(po.logx)
  { xmin = log10(xmin);
    xmax = log10(xmax);
    opt = "LN";
    min_value = log10(min_value);
    max_value = log10(max_value);
  }
  variable y = coordY_in_box( qualifier("yrel", 0) );
  if(po.logy)  y = log10(y);
  _pgsci( qualifier("color", 1) );
  _pgslw( qualifier("linewidth", 1) );
  _pgaxis(qualifier("opt", opt), xmin, y, xmax, y,
          min_value, max_value,  step, qualifier("nsub", 0),
          qualifier("maji", 0.5), qualifier("majo", 0), qualifier("fmin", 0.5), qualifier("disp", 0.5),
          qualifier("angle", 0)
         );
}

%
% Code from src/plot/x2axis.sl
%
%%%%%%%%%%%%%
define x2axis()
%%%%%%%%%%%%%
%!%+
%\function{x2axis}
%\synopsis{draws a second x-axis}
%\usage{x2axis(Double_Type min_value, Double_Type max_value[, Double_Type step]);}
%\qualifiers{
%\qualifier{nsub}{[=0] number of minor tick marks within each major divison (nsub=0 => no minor ticks)}
%\qualifier{maji}{[=0.5] length of major tick marks, drawn inwards (in units of the character height)}
%\qualifier{majo}{[=0] length of major tick marks, drawn outwards (in units of the character height)}
%\qualifier{fmin}{[=0.5] length of minor tick marks (as fraction of the major tick marks)}
%\qualifier{disp}{[=0.5] displacement of baseline of tick labels to the axis (in units of the character height)}
%\qualifier{angle}{[=0] orientation of the text (in degrees)}
%\qualifier{color}{[=1] color of the axis}
%\qualifier{linewidth}{[=1] line width of the axis}
%\qualifier{opt}{[="N"/"LN"] pgaxis-options: "L" (log), "N" (numbers), "1" (force decimal), "2" (force EE)}
%\qualifier{yrel}{[=1] relative position of the x2axis (0=bottom, 1=top)}
%\qualifier{xmin}{[=xmin] absolute x-position of the axis}
%\qualifier{xmax}{[=xmax] absolute x-position of the axis}
%}
%\description
%    \code{x2axis} has to be called after the plot window is drawn.
%    It relies on the world coordinates set by \code{xrange}, \code{yrange}.
%    The automatical x2axis can be switched off with \code{change_plot_options}.
%\example
%    ()=change_plot_options(; xopt="BNST");  % default: "BCNST", see _pgbox
%    xrange(0, 360); yrange(-1.1, 1.1);
%    plot([0:360], sin([0:2*PI:#361]));
%    x2axis(0, 2);
%\seealso{_pgaxis, change_plot_options, coords_in_box, y2axis, x1axis, y1axis}
%!%-
{
  variable po = get_plot_options();
  variable min_value, max_value, step=0;
  switch(_NARGS)
  { case 2: (min_value, max_value) = (); }
  { case 3: (min_value, max_value, step) = (); }
  { help(_function_name()); return; }

  variable opt = "N";
  variable xmin = qualifier("xmin", po.xmin);
  variable xmax = qualifier("xmax", po.xmax);
  if(po.logx)
  { xmin = log10(xmin);
    xmax = log10(xmax);
    opt = "LN";
    min_value = log10(min_value);
    max_value = log10(max_value);
  }
  variable y = coordY_in_box( qualifier("yrel", 1) );
  if(po.logy)  y = log10(y);
  _pgsci( qualifier("color", 1) );
  _pgslw( qualifier("linewidth", 1) );
  _pgaxis(qualifier("opt", opt), xmin, y, xmax, y,
          min_value, max_value,  step, qualifier("nsub", 0),
          qualifier("majo", 0), qualifier("maji", 0.5), qualifier("fmin", 0.5), -qualifier("disp", 0.5),
          qualifier("angle", 0)
         );
}

%
% Code from src/plot/x2axis_with_ticks.sl
%
define x2axis_with_tics()
%!%+
%\function{x2axis_with_tics}
%\usage{x2axis_with_tics(Double_Type X[][, Label]);}
%\qualifiers{
%\qualifier{ftick}{[=0.08]: fraction of box-height that tick marks are below axis}
%\qualifier{flabel}{[=0.05]: fraction of box-height that labels are above axis}
%}
%!%-
{
  variable X, Label;
  switch(_NARGS)
  { case 1:  X = (); Label = X; }
  { case 2: (X, Label) = (); }
  { help(_function_name()); return; }

  if(length(X)>0)
    switch( typeof([Label][0]) )
    { case String_Type:
      ; % Label is already ok.
    }
    { case Null_Type:
        Label = String_Type[length(X)];
        Label[*] = "";
    }
    { % else:
        Label = array_map(String_Type, &string, Label);
    }

  variable plot_opt = get_plot_options();
  variable x1 = plot_opt.xmin;
  variable x2 = plot_opt.xmax;
  variable y1 = plot_opt.ymin;
  variable y2 = plot_opt.ymax;
  variable yt = coordY_in_box(1-qualifier("ftick", 0.08));
  variable yl = coordY_in_box(1+qualifier("flabel", 0.05));

  color(1);
  set_line_width(1);
  oplot([x1, x2], [y2, y2]);
  variable i, I = where(x1 <= X <= x2);
  foreach i (I)
  {
%    _pgsci(1);
%    variable v = (X[i]-x1)/(x2-x1);  % if linear
%    _pgtick(x1, y2, x2, y2, v, 0.5, 0, 0.5, 0, "a");
    color(1); oplot(X[i]*[1,1], [yt, y2]);
    color(1); xylabel(X[i], yl, Label[i], 0, 0.5);
  }

  set_plot_options(plot_opt);
}

%
% Code from src/plot/x2label.sl
%

define x2label()
%!%+
%\function{x2label}
%\synopsis{labels the second x-axis}
%\usage{x2label(String_Type s);}
%\qualifiers{
%\qualifier{color}{[=1] textcolor}
%\qualifier{f}{[=0.03] fraction of the plot-area's height that the label is above}
%}
%\seealso{xylabel}
%!%-
{
  variable s;
  switch(_NARGS)
  { case 1: s = (); }
  { help(_function_name()); return; }

  variable plot_opt = get_plot_options();
  color( qualifier("color", 1) );
  xylabel(coords_in_box(0.5, 1+qualifier("f", 0.03)), s, 0, 0.5);
}

%
% Code from src/plot/xfig_3d_orbit_on_cube.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

% --------------------------------------------------------
% determine roll angle (kindly provided by Manfred Hanke):
private define projected_angle_for_eye(dist, theta, phi, roll, x, y, z)
{
  xfig_set_eye(dist, theta, phi, roll);
  variable v = vector([0,x], [0,y], [0,z]);
  variable px, py; (px, py) = xfig_project_to_xfig_plane(v);
  return atan2(py[0]-py[1], px[1]-px[0]) * 180/PI;  % Xfig's y increases downwards
}

private define computed_roll_angle(dist, theta, phi, v, expected_angle)
{
  variable roll = [0:360:#3601] [[:-2]];
  variable angle = array_map(Double_Type, &projected_angle_for_eye, dist, theta, phi, roll, v.x, v.y, v.z);
  variable delta = (angle - expected_angle) mod 360;
  delta[where(delta < -180)] += 360;
  delta[where(delta > +180)] -= 360;
  roll = roll[wherefirstmin(abs(delta))];
  return roll;
}
private define xfig_set_eye_with_computed_roll_angle(dist, theta, phi)
{
  xfig_set_eye(dist, theta, phi, computed_roll_angle(dist, theta, phi, vector(0,0,1), 90));
}
% --------------------------------------------------------

define xfig_3d_orbit_on_cube()
%!%+
%\function{xfig_3d_orbit_on_cube}
%\synopsis{Visualize 3d orbits by projecting them onto a cube}
%\usage{xfig = xfig_3d_orbit_on_cube(Struct_Type orbits [, Struct_Type text]; qualifiers)}
%\description
%    The orbits have to be given in cartesian coordinates (x,y,z) and are passed
%    to the function via a structure. Each individual orbit is itself a structure
%    with fields 'x', 'y', and 'z'. The orbits are then visualized by plotting the
%    projections to the xy-, xz-, and yz-plane on the surfaces of a cube. Use the
%    qualifiers to change the viewing angle, labels, or properties of the cube and
%    the background grid. The function's optional second argument can be used to
%    add text or symbols to the plot (see example section below).
%\notes
%    If one of the three coordinates is missing, the projection on the plane defined
%    by the remaining two coordinates is still plotted. This feature can be used to
%    add lines or text to one specific plane only.
%
%    The properties of the orbits (line style, color, ...) can be changed by adding the
%    field 'qualies' to the structure that decribes the orbit. The field 'qualies' is
%    again a structure whose fields are qualifiers of the function 'xfig_new_polyline'.
%\qualifiers{
%\qualifier{phi [=-45]}{Azimuthal angle in the x-y-plane in degrees (-90 < phi < 0).}
%\qualifier{theta [=60]}{Polar angle from the z-axis in degrees (0 < theta < 90).}
%\qualifier{dist [=1e6]}{Distance of the eye from the focus.}
%\qualifier{scale [=4]}{Factor to scale dimensions with respect to the character size (scale > 0).}
%\qualifier{x_label [="$x$\\,(kpc)"]}{x-label.}
%\qualifier{y_label [="$y$\\,(kpc)"]}{y-label.}
%\qualifier{z_label [="$z$\\,(kpc)"]}{z-label.}
%\qualifier{digits [=1]}{Digits for ticmarks (a non-negative integer).}
%\qualifier{cube}{Modify the cube by providing a structure whose fields are qualifiers of the function 'xfig_new_polyline'.}
%\qualifier{grid}{Modify the grid by providing a structure whose fields are qualifiers of the function 'xfig_new_polyline'.}
%\qualifier{center}{Center the cube at (0,0,0).}
%\qualifier{z_adjust}{Use separate scale for z-axis.}
%}
%\example
%   % basic example:
%   o = struct{ orbit1, orbit2, orbit3 };
%   o.orbit1 = struct{ x=[0,1], y=[0,1], z=[0,1] };
%   o.orbit2 = struct{ x=[1,1], y=[4,1], z=[-3,0], qualies=struct{color="blue"} };
%   o.orbit3 = struct{ x=[-1,0,1], y=[-1,1,-1], qualies=struct{closed, fillcolor="tomato", depth=5} }; % filled triangle in the xy-plane
%   xfig = xfig_3d_orbit_on_cube(o);
%   xfig.render("test.pdf");
%   xfig_set_eye(1e6,0,0,0); % restore the default position of the eye
%
%   % adding text to xy-plane:
%   t = struct{ triangle=struct{text="triangle", x0=0, y0=0, depth=4} };
%   xfig_3d_orbit_on_cube(o,t).render("test.pdf");
%   xfig_set_eye(1e6,0,0,0); % restore the default position of the eye
%
%   % example involving the function 'orbit_calculator':
%   o = orbit_calculator([12,12],[22,22],[29.6,29.6],[40,40],[49,49],[36,36],[3.078,4],[262,262],[-13.52,-13.52],[16.34,16.34],-1000; r_disk=50, set);
%   set_struct_field(o.tr, "o0", struct_combine(o.tr.o0, "qualies")); set_struct_field(o.tr.o0, "qualies", struct{ color="red", depth=2, backward_arrow });
%   set_struct_field(o.tr, "o1", struct_combine(o.tr.o1, "qualies")); set_struct_field(o.tr.o1, "qualies", struct{ color="blue", depth=1, backward_arrow });
%   xfig_3d_orbit_on_cube(o.tr).render("test.pdf");
%   xfig_set_eye(1e6,0,0,0); % restore the default position of the eye
%
%   % changing the appearance of the cube and grid:
%   xfig_3d_orbit_on_cube(o.tr; cube=struct{color="red",width=2}, grid=struct{line=0}).render("test.pdf");
%   xfig_set_eye(1e6,0,0,0); % restore the default position of the eye
%
%   % adding symbols:
%   t = struct{ sun, gc };
%   t.sun = struct{ text="$\\odot$", x0=-8.4, y0=0, z0=0, depth=3 };
%   t.gc = struct{ text="+", x0=0, y0=0, z0=0 };
%   xfig_3d_orbit_on_cube(o.tr, t).render("test.pdf");
%   xfig_set_eye(1e6,0,0,0); % restore the default position of the eye
%\seealso{orbit_calculator}
%!%-
{
  variable o, s;
  if(_NARGS==1)
    o = ();
  else if(_NARGS==2)
    (o, s)  = ();
  else
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  ifnot(is_struct_type(o))
    throw UsageError, sprintf("Usage error in '%s': first input parameter has to be a structure.", _function_name());
  % -----------
  % qualifiers:
  variable digits = qualifier("digits", 1);
  if(typeof(digits)!=Integer_Type or digits<0)
  {
    vmessage("Warning in '%s': Qualifier 'digits' needs to be a non-negative Integer_Type. Setting it to default value digits=1.", _function_name());
    digits = 1;
  }
  variable fmt = sprintf("%%.%df", digits);
  variable label_x = qualifier("x_label", "$x$\\,(kpc)");
  variable label_y = qualifier("y_label", "$y$\\,(kpc)");
  variable label_z = qualifier("z_label", "$z$\\,(kpc)");
  variable scale = qualifier("scale", 4);
  if(scale<=0)
  {
    vmessage("Warning in '%s': Qualifier 'scale' out of valid range scale>0. Setting it to default value scale=4.", _function_name());
    scale = 4;
  }
  % -----------------------
  % set eye and roll angle:
  variable phi = qualifier("phi", -45);
  ifnot(-90<phi<0)
  {
    vmessage("Warning in '%s': Qualifier 'phi' out of valid range -90<phi<0. Setting it to default value phi=-45.", _function_name());
    phi = -45;
  }
  variable theta = qualifier("theta", 60);
  ifnot(0<theta<90)
  {
    vmessage("Warning in '%s': Qualifier 'theta' out of valid range 0<theta<90. Setting it to default value theta=60.", _function_name());
    theta = 60;
  }
  xfig_set_focus(vector(0,0,0));
  xfig_set_eye_with_computed_roll_angle(qualifier("dist", 1e6), theta, phi);
  % -----------------------
  % -----------
  % -------------
  % preparations:
  variable xmin = +_Inf, xmax = -_Inf;
  variable ymin = +_Inf, ymax = -_Inf;
  variable zmin = +_Inf, zmax = -_Inf;
  % determine plot ranges:
  variable field;
  variable orbits = get_struct_field_names(o);
  foreach field (orbits)
  {
    % x:
    if(struct_field_exists( get_struct_field( o, field ), "x"))
    {
      xmin = _min(xmin,min(get_struct_field( get_struct_field( o, field ), "x")));
      xmax = _max(xmax,max(get_struct_field( get_struct_field( o, field ), "x")));
    }
    % y:
    if(struct_field_exists( get_struct_field( o, field ), "y"))
    {
      ymin = _min(ymin,min(get_struct_field( get_struct_field( o, field ), "y")));
      ymax = _max(ymax,max(get_struct_field( get_struct_field( o, field ), "y")));
    }
    % z:
    if(struct_field_exists( get_struct_field( o, field ), "z"))
    {
      zmin = _min(zmin,min(get_struct_field( get_struct_field( o, field ), "z")));
      zmax = _max(zmax,max(get_struct_field( get_struct_field( o, field ), "z")));
    }
  }
  if(is_struct_type(s))
  {
    variable symbols = get_struct_field_names(s);
    foreach field (symbols)
    {
      % x:
      if(struct_field_exists( get_struct_field( s, field ), "x0"))
      {
	xmin = _min(xmin, get_struct_field(get_struct_field( s, field ), "x0"));
	xmax = _max(xmax, get_struct_field(get_struct_field( s, field ), "x0"));
      }
      % y:
      if(struct_field_exists( get_struct_field( s, field ), "y0"))
      {
	ymin = _min(ymin, get_struct_field(get_struct_field( s, field ), "y0"));
	ymax = _max(ymax, get_struct_field(get_struct_field( s, field ), "y0"));
      }
      % z:
      if(struct_field_exists( get_struct_field( s, field ), "z0"))
      {
	zmin = _min(zmin, get_struct_field(get_struct_field( s, field ), "z0"));
	zmax = _max(zmax, get_struct_field(get_struct_field( s, field ), "z0"));
      }
    }
  }
  % -----------------------------
  % if no data are to be plotted:
  if(xmin==+_Inf)
    xmin = -1;
  if(ymin==+_Inf)
    ymin = -1;
  if(zmin==+_Inf)
    zmin = -1;
  if(xmax==-_Inf)
    xmax = 1;
  if(ymax==-_Inf)
    ymax = 1;
  if(zmax==-_Inf)
    zmax = 1;
  % -----------------------------
  variable buffer = 0.05;
  xmin -= buffer*(xmax-xmin);
  xmax += buffer*(xmax-xmin);
  ymin -= buffer*(ymax-ymin);
  ymax += buffer*(ymax-ymin);
  zmin -= buffer*(zmax-zmin);
  zmax += buffer*(zmax-zmin);
  %
  if(qualifier_exists("center")) % center at origin
  {
    variable xy_dim = _max(abs(xmin), abs(xmax), abs(ymin), abs(ymax));
    variable z_dim = _max(abs(zmin), abs(zmax));
    if(qualifier_exists("z_adjust"))
    {
      xmin = -xy_dim;
      xmax =  xy_dim;
      ymin = -xy_dim;
      ymax =  xy_dim;
      zmin = -z_dim;
      zmax =  z_dim;
    }
    else
    {
      variable dim = _max(xy_dim,z_dim);
      xmin = -dim;
      xmax =  dim;
      ymin = -dim;
      ymax =  dim;
      zmin = -dim;
      zmax =  dim;
    }
  }
  %
  variable xy_unit = _max( ceil((xmax-xmin)*10^digits/4.), ceil((ymax-ymin)*10^digits/4.) )*2./10^digits;
  variable z_unit = ceil((zmax-zmin)*10^digits/4.)*2./10^digits;
  ifnot(qualifier_exists("z_adjust"))
  {
    z_unit = _max(xy_unit,z_unit);
    xy_unit = z_unit;
  }
  variable x_center = round(0.5*(xmin+xmax)*10^digits)/10^digits;
  variable y_center = round(0.5*(ymin+ymax)*10^digits)/10^digits;
  variable z_center = round(0.5*(zmin+zmax)*10^digits)/10^digits;
  % -------------
  % -----
  % xfig:
  variable xfig = xfig_plot_new();
  xfig.axis(; off);
  % ----------------
  % reference system
  variable cube = struct{ depth=99 };
  if(qualifier_exists("cube") && is_struct_type(qualifier("cube")))
    cube = struct_combine(cube, qualifier("cube"));
  xfig.add_object( xfig_new_polyline( scale*vector([-1,1] ,[1,1]  ,[1,1]  );; cube) );
  xfig.add_object( xfig_new_polyline( scale*vector([-1,1] ,[-1,-1],[1,1]  );; cube) );
  xfig.add_object( xfig_new_polyline( scale*vector([-1,1] ,[-1,-1],[-1,-1]);; cube) );
  xfig.add_object( xfig_new_polyline( scale*vector([1,1]  ,[-1,1] ,[1,1]  );; cube) );
  xfig.add_object( xfig_new_polyline( scale*vector([-1,-1],[-1,1] ,[1,1]  );; cube) );
  xfig.add_object( xfig_new_polyline( scale*vector([1,1]  ,[-1,1] ,[-1,-1]);; cube) );
  xfig.add_object( xfig_new_polyline( scale*vector([-1,-1],[-1,-1],[-1,1] );; cube) );
  xfig.add_object( xfig_new_polyline( scale*vector([1,1]  ,[-1,-1],[-1,1] );; cube) );
  xfig.add_object( xfig_new_polyline( scale*vector([1,1]  ,[1,1]  ,[-1,1] );; cube) );
  variable i, tics = 2; % total number of ticmarks is 2*tics-1
  variable grid = struct{ line=1, color="gray", depth=100 };
  if(qualifier_exists("grid") && is_struct_type(qualifier("grid")))
    grid = struct_combine(grid, qualifier("grid"));
  _for i(-tics,tics,1)
  {
    xfig.add_object( xfig_new_polyline( scale*vector([-1,1],i/(1.*tics)*[1,1],[1,1]);;   grid) );
    xfig.add_object( xfig_new_polyline( scale*vector([-1,1],[-1,-1],i/(1.*tics)*[1,1]);; grid) );
    xfig.add_object( xfig_new_polyline( scale*vector([1,1],i/(1.*tics)*[1,1],[-1,1]);;   grid) );
    xfig.add_object( xfig_new_polyline( scale*vector(i/(1.*tics)*[1,1],[-1,-1],[-1,1]);; grid) );
    xfig.add_object( xfig_new_polyline( scale*vector([1,1],[-1,1],i/(1.*tics)*[1,1]);;   grid) );
    xfig.add_object( xfig_new_polyline( scale*vector(i/(1.*tics)*[1,1],[-1,1],[1,1]);;   grid) );
  }
  xfig.add_object( xfig_new_polyline( scale*vector([-.5,.5],[-1.1,-1.1],[-1,-1]);;     struct_combine( cube, struct{ forward_arrow }))); % Arrow parallel to x-axis
  xfig.add_object( xfig_new_polyline( scale*vector([1.1,1.1],[-.5,.5],[-1,-1]);;       struct_combine( cube, struct{ forward_arrow }))); % Arrow parallel to y-axis
  xfig.add_object( xfig_new_polyline( scale*vector([1.05,1.05],[1.05,1.05],[-.5,.5]);; struct_combine( cube, struct{ forward_arrow }))); % Arrow parallel to z-axis
  xfig.add_object( xfig_new_text(label_x;; struct_combine( cube, struct{ x0=0, y0=-1.15*scale, z0=-1*scale, just=[0,1.5] }))); % x-label
  xfig.add_object( xfig_new_text(label_y;; struct_combine( cube, struct{ x0=1.15*scale, y0=0, z0=-1*scale, just=[-0.5,0] }))); % y-label
  xfig.add_object( xfig_new_text(label_z;; struct_combine( cube, struct{ x0=1.12*scale, y0=1.12*scale, z0=0, just=[-0.1,-0.1], rotate=90 }))); % z-label
  xfig.add_object( xfig_new_text(sprintf(fmt, -0.5*xy_unit+x_center);; struct_combine( cube, struct{ x0=-.5*scale,    y0=1.05*scale, z0=1*scale,    just=[0,-1.]  }))); % x-ticmark
  xfig.add_object( xfig_new_text(sprintf(fmt,  0.0*xy_unit+x_center);; struct_combine( cube, struct{ x0=0,            y0=1.05*scale, z0=1*scale,    just=[0,-1.]  }))); % x-ticmark
  xfig.add_object( xfig_new_text(sprintf(fmt,  0.5*xy_unit+x_center);; struct_combine( cube, struct{ x0=.5*scale,     y0=1.05*scale, z0=1*scale,    just=[0,-1.]  }))); % x-ticmark
  xfig.add_object( xfig_new_text(sprintf(fmt, -0.5*xy_unit+y_center);; struct_combine( cube, struct{ x0=-1.075*scale, y0=-.5*scale,  z0=1*scale,    just=[0.5,0]  }))); % y-ticmark
  xfig.add_object( xfig_new_text(sprintf(fmt,  0.0*xy_unit+y_center);; struct_combine( cube, struct{ x0=-1.075*scale, y0=0,          z0=1*scale,    just=[0.5,0]  }))); % y-ticmark
  xfig.add_object( xfig_new_text(sprintf(fmt,  0.5*xy_unit+y_center);; struct_combine( cube, struct{ x0=-1.075*scale, y0=.5*scale,   z0=1*scale,    just=[0.5,0]  }))); % y-ticmark
  xfig.add_object( xfig_new_text(sprintf(fmt, -0.5*z_unit+z_center);;  struct_combine( cube, struct{ x0=-1*scale,     y0=-1*scale,   z0=-0.5*scale, just=[0.7,0.7]}))); % z-ticmark
  xfig.add_object( xfig_new_text(sprintf(fmt,  0.0*z_unit+z_center);;  struct_combine( cube, struct{ x0=-1*scale,     y0=-1*scale,   z0=0,          just=[0.7,0.7]}))); % z-ticmark
  xfig.add_object( xfig_new_text(sprintf(fmt,  0.5*z_unit+z_center);;  struct_combine( cube, struct{ x0=-1*scale,     y0=-1*scale,   z0=0.5*scale,  just=[0.7,0.7]}))); % z-ticmark
  % ----------------
  % -------
  % orbits:
  foreach field (orbits)
  {
    variable temp = get_struct_field(o, field);
    variable qualies = struct{ color="black" };
    if(struct_field_exists( temp, "qualies" ))
      qualies = struct_combine( qualies, get_struct_field(temp, "qualies") );
    if(struct_field_exists(temp, "x") and struct_field_exists(temp, "y"))
      xfig.add_object( xfig_new_polyline(vector( (temp.x-x_center)/xy_unit*scale, (temp.y-y_center)/xy_unit*scale, 0*temp.x+scale );; qualies) ); % xy-plane
    if(struct_field_exists(temp, "x") and struct_field_exists(temp, "z"))
      xfig.add_object( xfig_new_polyline(vector( (temp.x-x_center)/xy_unit*scale, 0*temp.x-scale, (temp.z-z_center)/z_unit*scale ) ;; qualies) ); % xz-plane
    if(struct_field_exists(temp, "y") and struct_field_exists(temp, "z"))
      xfig.add_object( xfig_new_polyline(vector( 0*temp.y+scale, (temp.y-y_center)/xy_unit*scale, (temp.z-z_center)/z_unit*scale)  ;; qualies) ); % yz-plane
  }
  % -------
  % --------
  % symbols:
  if(is_struct_type(s))
  {
    foreach field (symbols)
    {
      temp = get_struct_field(s, field);
      variable x0=0, y0=0, z0=0;
      if(struct_field_exists(temp, "x0"))
	x0 = (get_struct_field(temp, "x0")-x_center)/xy_unit*scale;
      if(struct_field_exists(temp, "y0"))
	y0 = (get_struct_field(temp, "y0")-y_center)/xy_unit*scale;
      if(struct_field_exists(temp, "z0"))
	z0 = (get_struct_field(temp, "z0")-z_center)/z_unit*scale;
      if(struct_field_exists(temp, "x0") and struct_field_exists(temp, "y0")) % xy-plane
	xfig.add_object(xfig_new_text(temp.text;; struct_combine(temp, struct{ x0=x0, y0=y0, z0=scale }) ));
      if(struct_field_exists(temp, "x0") and struct_field_exists(temp, "z0")) % xz-plane
	xfig.add_object(xfig_new_text(temp.text;; struct_combine(temp, struct{ x0=x0, y0=-scale, z0=z0 }) ));
      if(struct_field_exists(temp, "y0") and struct_field_exists(temp, "z0")) % yz-plane
	xfig.add_object(xfig_new_text(temp.text;; struct_combine(temp, struct{ x0=scale, y0=y0, z0=z0 }) ));
    }
  }
  % --------
  % -----
  return xfig;
}


#endif

%
% Code from src/plot/xfig_compress_eps.sl
%
define xfig_compress_eps(file)
%!%+
%\function{xfig_compress_eps}
%\synopsis{Compresses an EPS-File via "pdftops -level3 -eps"}
%\usage{int = xfig_compress_eps(String_Type file)}
%\description
%    PS and EPS-Files created with xfig are usually much too large.
%    Using this function the size can be reduced by a factor of few. Simply
%    call this function after creating a PS or EPS with ".render(file)".
%!%-
{
  if (stat_file(file) == NULL)
  {
    vmessage("ERROR: %s(): %s does not exist", _function_name(), file);
    return -1;
  }
  if ("application/postscript" != strtrim(fgetslines(popen("file $file --mime-type -b"$,"r"))[0]))
  {
    vmessage("ERROR: %s(): %s is no [e]ps file", _function_name(), file);
    return -1;
  }

  variable tmp_pdf = "_temp.pdf";  % make sure that no existing files are overwritten with tmp_pdf:
  while (stat_file(tmp_pdf) != NULL) { tmp_pdf = char(nint(65+urand()*25)) + tmp_pdf; }
  
%  return system("ps2pdf -dEPSCrop $file $tmp_pdf; pdf2ps -dLanguageLevel=3 $tmp_pdf $file; rm -f $tmp_pdf"$ );
  return system("ps2pdf -dEPSCrop $file $tmp_pdf; pdftops -level3 -eps $tmp_pdf $file; rm -f $tmp_pdf"$ );
  % MB: the latter case produces smaller output and the bounding box should be set correctly

  % Mh: with pdftops >= 0.13.3, one could avoid tmp_pdf altogether:
  % return system("ps2pdf -dEPSCrop $file - | pdftops -level3 -eps - $file"$);
}

%
% Code from src/plot/xfig_draw_orbit.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

define xfig_draw_orbit ()
%!%+
%\function{xfig_draw_orbit}
%\synopsis{draws the orbit of a binary system into an exsiting xfig-object and
%    allow also to highlight specific phases}
%\usage{xfig_draw_orbit(Xfig_Object xf, Double_Type asini, i, ecc, omega, rstar)
%}
%\qualifiers{
%\qualifier{line_color}{[\code{="gray"}] color of the lines showing the
%           coordinate main axes and the orientation of the semi-major axis of
%           the orbit}
%\qualifier{phases}{[\code{=\{\}}] list of Double_Type[2] arrays indicating
%           the phases which should be highlighted}
%\qualifier{scale}{[\code{=1.05}]  compression/streching for the highlighted
%           phases over/under the orbit}
%\qualifier{phase_color}{[\code{="red"}] color of the highlighted phases,
%           can be a single string or an array of strings, with the same length
%           as phases so that every highlighted phase can be drawn in
%           a different color.}
%\qualifier{phase_width}{[\code{=2}] width of the highlighted phases, can
%           also be an array, just as phase_color.}
%\qualifier{star_color}{[\code{=lightblue}]  fill-color of the companion star}
%\qualifier{trueanom}{if set, phases are given in true anomaly, i.e., equal
%           delta-phi values will correspond to equal delta-t on the
%           orbit.}
%\qualifier{zerolines}{[\code{=1}] plot x- and y-axis lines through [0,0]}
%\qualifier{phaselines}{draw a lightgray line every phase n*0.1 and label it}
%\qualifier{label_phase_switch}{[\code{=0}] : phase at witch phase labels
%           are switched from before the line to after the line.}
%\qualifier{label_pos}{[\code{=0.85}] distance of the phase labels from the 
%           orbit line, in fraction of distance to center.}
%\qualifier{nolabel} { : removes the coordinate system }
%}
%           
%\description
%    This function plots the orbit of a binary system as seen from
%    above into a given xfig-object. The orbital parameters asini
%    (semi-major axis, in lt-sec), i (inclination, in deg), ecc (eccentricity),
%    omega (argument of periastron, in deg) and the radius rstar (in r_sun)
%    of the companion star need to be supplied.
%    Qualifiers are passed through the xfig-plotting routine of the
%    orbit.
%        
%\examples
%     %orbit parameters of GX 301-2 according to Koh et al (1997).
%     variable asini = 368.3 ; % lt-sec (semi major axis)
%      variable i = 66; % degree to rad
%      variable ecc = 0.462 ;  %
%      variable omega = 310.4 ; % degrees
%      variable rstar = 43 ; % solar radii  in lt-sec
%
%      variable xfgx = xfig_plot_new(12,12) ;
%      xfgx.world(-600,450,-450, 600) ;
%      xfig_draw_orbit(xfgx, asini, i, ecc, omega, rstar ;
%        phases={[0.9184,0.9315]}, phase_color="red", phase_width=4, scale=1) ;
%      xfgx.plot([0,0],[-100,-370] ; line=1, forward_arrow, depth = 10,
%        arrow_thickness =2) ;
%      xfgx.xylabel(0,-400, "To Earth");
%      xfgx.render("plots/gx301_skizze2.eps") ;
%
%  \seealso{ellipse}
%!%-
{
variable xfgx, asini, i, ecc, omega, rstar ;
  switch(_NARGS)
    % [asini]=lt-sec, [i]=deg, [ecc]=N/A, [omega]=deg, [rstar]=R_sun
    { case 6: (xfgx, asini, i, ecc, omega, rstar) = (); }
    { help(_function_name()); return; }
      xfgx.axis(;off) ;

    variable c=  2.998e8 ; %speed of light in cm/s
    variable rsun=1.3914e9/2. ; %solar radius in cm

    i = i*PI/180.; % degree to rad

    variable a = asini/sin(i) ;
  
    variable b = sqrt( a^2 * (1.-ecc)*(1.+ecc) ); % semi-minor axis, lt-sec
    omega = omega*PI/180.;

    rstar = rstar*rsun/c ; % solar radii  in lt-sec

    variable scal = qualifier("scale",1.05) ;
    variable lcol = qualifier("line_color","gray") ;
  
    variable phases = qualifier("phases", {}) ;
    variable ph ;
    variable phcol = [qualifier("phase_color", "red")] ;
    variable phwid = [qualifier("phase_width", 2)] ;

    if (length(phcol)==1)
    {phcol = phcol[Integer_Type[length(phases)]];}
  if (length(phwid)==1)
     {phwid = phwid[Integer_Type[length(phases)]];}

if (length(phcol) < length(phases)) {  message("FATAL ERROR (sysdraw): phases color needs at least as many entries as phase list!") ; return ;}
if (length(phwid) < length(phases)) {  message("FATAL ERROR (sysdraw): phases width needs at least as many entries as phase list!") ; return ;}
variable ox1, oy1, ox, oy, phax, phay ;
variable stx, sty ;

(ox,oy) = ellipse(a, b, omega, [0:2*PI:#256]) ;
  
  (stx,sty) = ellipse(rstar,rstar,0,[0:2*PI:#256]) ;

  variable xymm = xfgx.get_world() ;
  
variable starcol = qualifier("star_color", "#"+rgb2hex(0.4, 0.6,1.0 ; string) );

xfgx.shade_region(stx,sty ; color=starcol, depth = 25) ;

  %all ellipses are shifted so that the star sits at [0,0].

  variable shiftx = a*ecc*cos(omega) ;
  variable shifty = a*ecc*sin(omega) ;

  xfgx.plot((ox-a*ecc*cos(omega)),(oy-a*ecc*sin(omega)) ;; __qualifiers) ;
  variable j = 0 , M, eccanom, theta  ;
  foreach ph (phases) {
    if (qualifier_exists("trueanom")){
      M = [0:2*PI:#2048] ; %mean anomaly, highly oversampled
      eccanom = KeplerEquation (M, ecc) ;
      (phax,phay) = ellipse(a*scal,b*scal,omega, eccanom[where(ph[0] <= M/2/PI < ph[1])]) ;
    }
    else{
      (phax,phay) = ellipse(a*scal, b*scal, omega, [ph[0]:ph[1]:#256]*2*PI) ;
    }
    xfgx.plot(phax-shiftx,phay-shifty ; color=phcol[j], depth = 5, width=phwid[j]) ;
    j++ ;
  }


if(qualifier_exists("phaselines")) {
  variable eccanom_cgrid = KeplerEquation([0.0:1:0.1]*2*PI, ecc) ;
  variable theta_cgrid = 2.*atan( sqrt( (1.+ecc)/(1.-ecc) )*tan(eccanom_cgrid/2.) );  % true anomaly from eccentric anomaly
  variable xlab,ylab ;
  variable labscal = qualifier("label_pos",0.85) ;
  variable darkgray = "#" + rgb2hex(0.4,0.4,0.4 ; string) ; % used for the phase labels

  variable phaswitch = qualifier("label_phase_switch", 0) ;
  ifnot (0<=phaswitch<1) {
    message("WARNING (xfig_draw_orbit): label_phase_switch must be in [0,1[, setting it now to default (=0)") ;
    phaswitch=0 ;
  }
  phaswitch = min(where([0:1:0.1] >= phaswitch)) ;

  
  _for j (0,phaswitch-1, 1){
    %phase indicator lines
    xfgx.plot([0,cos(theta_cgrid[j]+omega)]*1500,[0,sin(theta_cgrid[j]+omega)]*1500 ; color="gray", line=3) ;
    %their labels
    (xlab, ylab) = ellipse(a, b, omega, eccanom_cgrid[j]+0.07)  ;
    xfgx.xylabel((xlab-shiftx)*labscal, (ylab-shifty)*labscal,sprintf("%.1f", j/10. ) ;
  		 rotate=(theta_cgrid[j]+omega)/2/PI*360., color=darkgray, size ="\scriptsize"R ) ;
  }

  _for j (phaswitch,length(theta_cgrid)-1, 1){
    %phase indicator lines
    xfgx.plot([0,cos(theta_cgrid[j]+omega)]*1500,[0,sin(theta_cgrid[j]+omega)]*1500 ; color="gray", line=3, depth=81) ;
    %their labels
    (xlab, ylab) = ellipse(a, b, omega, eccanom_cgrid[j]-0.07)  ;
    xfgx.xylabel((xlab-shiftx)*labscal, (ylab-shifty)*labscal,sprintf("%.1f", j/10. ) ;
		 rotate=(theta_cgrid[j]+omega)/2/PI*360.+180, color=darkgray, size ="\scriptsize"R,depth=20 ) ;
}
}
  variable zerolines = qualifier("zerolines", 1) ;
  
  if (zerolines){
xfgx.plot([xymm[0],xymm[1]],[0,0] ; line=0, color=lcol);
xfgx.plot([0,0],[xymm[2],xymm[3]] ; line=0 , color=lcol);
  }
% xfgx.plot([xymm[0]:xymm[1]:#256],[xymm[0]:xymm[1]:#256]*tan(omega) ; line=3, color=lcol) ;

  ifnot (qualifier_exists("nolabel")) {
    xfgx.xlabel("[lt-sec]") ;
    xfgx.ylabel("[lt-sec]") ;
    xfgx.x1axis(;on) ;
    xfgx.x2axis(;on,ticlabels=0) ;
    xfgx.y1axis(;on) ;
    xfgx.y2axis(;on,ticlabels=0) ;
  }
  return asini, i ;
}

#endif

%
% Code from src/plot/xfig_draw_orbit_3D.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

% function to shift, incline, and rotate orbit
private define _xfig_draw_orbit_3D_ellipse(x, foc, inc, omega) {
  x += vector(-foc, 0, 0);
  ifnot (qualifier_exists("noinc")) {
    x = vector_rotate(x, vector(1, 0, 0), inc * PI/180);
  }
  ifnot (qualifier_exists("noomega")) {
    x = vector_rotate(x, vector_rotate(vector(0, 0, 1), vector(1, 0, 0),
      inc * PI/180), omega * PI/180);
  }
  return x;
}

% projection of a 3D vector onto the camera field of view
private define _xfig_draw_orbit_3D_proj(r, cx, cy) {
  return (dotprod(r, cx), dotprod(r, cy));
}

%%%%%%%%%%%%%%%%%
define xfig_draw_orbit_3D()
%%%%%%%%%%%%%%%%%
%!%+
%\function{xfig_draw_orbit_3D}
%\synopsis{draws the orbit of a binary system in 3D}
%\usage{Xfig_Object xfig_draw_orbit([Xfig_Object xf,] Double_Type asini, i, ecc, omega);}
%\qualifiers{
%    \qualifier{caminc}{the camera inclination to the line of sight
%              (in degrees; default: 90-i)}
%    \qualifier{camroll}{the camera roll angle around the z-axis (along the
%              line of sight; in degress; default: 0)}
%}
%\description
%    Similar to 'xfig_draw_orbit' this function plots the orbit of a
%    binary, but seen from a user-defined direction and, thus, in 3D.
%    In order to draw the orbit the projected semi-major axis (asini),
%    inclination (i), eccentricity (ecc), and longitude of periastron
%    (omega) have to be provided.
%    
%    Additionally to the orbital plane the tangent plane of the sky
%    is drawn as well. This plane is defined to be perpendicular to
%    the line of sight and through the center of mass, i.e. the focal
%    point of the eccentric orbit nearest to the periastron (P).
%    Thereby, the observer is located below the tangent plane of the
%    sky, indicated by the arrow pointing to Earth.
%    All dashed lines are within the orbital plane, which are the
%    intersection of the orbital and tangent plane of the sky, the
%    semi-major and -minor axis (in red), and the line connecting the
%    orbit's center and the periastron. The projected semi-major axis
%    (asini) is drawn as dotted dashed line.
%    
%    The camera is controlled via qualifiers and by default the binary
%    is seen face-on.
%
%    Note that so far only basic functionallity have been implemented.
%\seealso{xfig_draw_orbit, ellipse}
%!%-
{
  variable pl = NULL, aSini, inc, ecc, omega;
  switch (_NARGS)
    { case 4: (aSini, inc, ecc, omega) = (); }
    { case 5: (pl, aSini, inc, ecc, omega) = (); }
    { help(_function_name()); return; }

  % camera
  variable cam  = qualifier("caminc", inc);   % inclination, 0 = onto tangent plane from above
  variable roll = qualifier("camroll", 0.); % rotation around z
  % camera unit vectors
  variable cx = vector(1, 0, 0);
  variable cy = vector(0, cos(cam*PI/180), sin(cam*PI/180));
  cx = vector_rotate(cx, vector(0,0,1), roll*PI/180);
  cy = vector_rotate(cy, vector(0,0,1), roll*PI/180);

  % orbit parameters
  variable a = aSini / sin(inc);
  % calculate focal point (center of mass)
  variable foc = sqrt(a^2 - (a*(1-ecc))^2);
  % calculate orbit
  variable phi = [0:2*PI:#200];
  variable r = vector(
    Double_Type[length(phi)], Double_Type[length(phi)], Double_Type[length(phi)]
  );
  (r.x, r.y) = ellipse(a, a*(1-ecc), 0, [0:2*PI:#200]);
  r = _xfig_draw_orbit_3D_ellipse(r, foc, inc, omega);

  %%% lines
  % intersection with tangent plane
  variable n = array_sort(abs(r.z))[[0,1]];
  variable ints = vector(r.x[n], r.y[n], [0,0]);
  % semi-* axes
  variable smaj = _xfig_draw_orbit_3D_ellipse(vector([0,-a], [0,0], [0,0]), foc, inc, omega);
  variable smin = _xfig_draw_orbit_3D_ellipse(vector([0,0], [0,a*(1-ecc)], [0,0]), foc, inc, omega);
  variable asini = vector(smaj.x[-1] * [1,1], smaj.y[-1] * [1,1], smaj.z);
  variable acosi = vector(smaj.x, smaj.y, smaj.z[0] * [1,1]);
  % position of periastron
  variable prast = _xfig_draw_orbit_3D_ellipse(vector(a, 0, 0), foc, inc, omega);
  % line from center of orbit to periastron
  variable cmpe = vector([smaj.x[0], prast.x], [smaj.y[0], prast.y], [smaj.z[0], prast.z]);
  % line from tg to apastron minus asini
  variable apast = vector(asini.x, asini.y, [0, asini.z[0]]);
  % line within the tg connecting the apastron and center of mass
  variable apcm = vector([apast.x[0], 0], [apast.y[0], 0], [0,0]);
  % center of mass
  variable cm = vector(0, 0, 0);

  % tangent plane
  variable wrld = a*(1+ecc);
  variable tp = vector([-1,-1,1,1,-1]*wrld, [-1,1,1,-1,-1]*wrld, [0,0,0,0,0]);

  %%% xfig plot
  if (pl == NULL) { pl = xfig_plot_new(12,12); }
  pl.world(-wrld, wrld, -wrld, wrld; padx = .1, pady = .1);
  pl.axes(; major = 0, minor = 0, color = "white");
  % orbit
  pl.plot(_xfig_draw_orbit_3D_proj(r, cx, cy));
  % tangent plane
  pl.plot(_xfig_draw_orbit_3D_proj(tp, cx, cy));
  % lines
  pl.plot(_xfig_draw_orbit_3D_proj(ints, cx, cy); line = 1);
  pl.plot(_xfig_draw_orbit_3D_proj(smaj, cx, cy); color = "red", line = 1);
  pl.plot(_xfig_draw_orbit_3D_proj(smin, cx, cy); color = "red", line = 1);
  pl.plot(_xfig_draw_orbit_3D_proj(asini, cx, cy); color = "red", line = 2);
  pl.plot(_xfig_draw_orbit_3D_proj(cmpe, cx, cy); line = 1);
  pl.plot(_xfig_draw_orbit_3D_proj(acosi, cx, cy); line = 2);
  pl.plot(_xfig_draw_orbit_3D_proj(apast, cx, cy); line = 2);
  pl.plot(_xfig_draw_orbit_3D_proj(apcm, cx, cy); line = 2);
  pl.plot(_xfig_draw_orbit_3D_proj(cm, cx, cy); sym = "x");
  % to Earth
  variable Earth = vector([1,1]*.8*wrld, -[1,1]*.8*wrld, -[.1,.25]*wrld);
  pl.plot(_xfig_draw_orbit_3D_proj(vector(Earth.x, Earth.y, [0, Earth.z[0]]), cx, cy); line = 2);
  pl.plot(_xfig_draw_orbit_3D_proj(Earth, cx, cy); forward_arrow);
  pl.xylabel(
    _xfig_draw_orbit_3D_proj(vector(Earth.x[1], Earth.y[1], Earth.z[1]), cx, cy), "to Earth",
    0, .7; size = "scriptsize"
  );
  % labels
  pl.xylabel(_xfig_draw_orbit_3D_proj(wrld*vector(0,1.05,.05), cx, cy), "tangent plane of the sky"R;
             size = "scriptsize");
  pl.xylabel(_xfig_draw_orbit_3D_proj(vector(prast.x*1.4, prast.y*1.4, prast.z), cx, cy), "P"; size = "scriptsize");
  pl.xylabel(_xfig_draw_orbit_3D_proj(vector(smaj.x[1]*1.05, smaj.y[1]*1.05, smaj.z[1]), cx, cy), "A";
	     size = "scriptsize");

  return pl;
}


#endif

%
% Code from src/plot/xfig_mix_colors.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

define xfig_mix_colors(color1,color2,fraction) {
%!%+
%\function{xfig_mix_colors}
%\synopsis{mix two named xfig colors}
%\usage{xfig_mix_colors(color1,color2,fraction)}
%\qualifiers{
%\qualifier{name}{xfig name of the color mix}
%}
%\description
% This function mixes two named xfig colors in rgb space. The function looks up
% the rgb values of color1 and color2, and mixes their rgb values according
% to 
% new color=color1*fraction+color2*(1-fraction)
% The operations are analoguous to the color mixing performed by the xcolor
% package of LaTeX (the operation is similar to LaTeX's color1!fraction!color2
% syntax). 
% The function returns the xfig name of the new color, or nothing if the
% name qualifier is given (recommended).
%\seealso{mix_rgb_colors}
%!%-
  variable rgb=color_mix(xfig_lookup_color_rgb(color1),
			 xfig_lookup_color_rgb(color2),fraction);

  variable colname;
  if (qualifier_exists("name")) {
      xfig_new_color(qualifier("name"),rgb);
      return;
  } else {
      colname=color1+"_"+sprintf("%3.3i",int(fraction*1000))+"_"+color2;
      xfig_new_color(colname,rgb);
      return colname;
  }
}


#endif

%
% Code from src/plot/xfig_multibox.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

define xfig_multibox()
%!%+
%\function{xfig_multibox}
%\synopsis{combines M x N xfig-objects in one box}
%\usage{pl = xfig_multibox(XFig_Plot_Type pl[M,N] [, Double_Type spacing])}
%\qualifiers{
%\qualifier{rotate}{switch the meaning of M and N}
%}
%\seealso{xfig_new_vbox_compound, xfig_new_hbox_compound}
%!%-
{

   variable pl,dx=0.0;
   switch(_NARGS)
   { case 1:  pl      = (); }
   { case 2:  (pl,dx)      = (); }
   { help(_function_name()); return; }

   variable ind = array_shape(pl);
   
   if (length(ind)==1)
   {
      if (qualifier_exists("rotate"))
	reshape(pl, [1,ind[0]]);
      else
	reshape(pl, [ind[0],1]);
   }
   else if (length(ind)==2 && qualifier_exists("rotate"))
     transpose(pl);

   ind = array_shape(pl);

   if (length(ind)!=2)  % check if the array has now reasonable shape
     return pl;

   variable m = ind[0], n=ind[1];
   variable i,j,pl_rows=Struct_Type[m];

   variable li_rows,li_sum;

   li_sum={};
   _for j(0,m-1,1)
   {
      li_rows={};
      _for i(0,n-1,1)
	list_append(li_rows,pl[j,i]);
      pl_rows[j]=xfig_new_hbox_compound(__push_list(li_rows),dx);
      list_append(li_sum,pl_rows[j]);
   }
   variable pl_sum = xfig_new_vbox_compound(__push_list(li_sum),dx);
   
   return pl_sum;
}

#endif

%
% Code from src/plot/xfig_normalized_coordinates.sl
%
define xfig_get_normalized_x_position(p,xval) {
%!%+
%\function{xfig_get_normalized_x_position}
%\synopsis{Compute normalized (world0) x-position from world1 x-position}
%\usage{xnorm=xfig_get_normalized_x_position(p,xval)}
%\description
%For a given xfig plot p and x-coordinate value xval given in the world1
%coordinate system, compute the corresponding normalized world0 x-value.
%
%This function is necessary since the xfig module does not export
%its world normalization functions. It is heavily based on these.
%
%\seealso{xfig_get_normalized_position,xfig_get_normalized_y_position}
%!%-
    variable xwcs=p.plot_data.x1axis.wcs_transform;
    variable cd=xwcs.client_data;
    variable ww=double(p.get_world());
    variable f=xwcs.wcs_func;
    variable t0 = (@f)(ww[0], cd);
    variable t1 = (@f)(ww[1], cd);
    variable xnorm=((@f)(double(xval), cd) - t0)/(t1-t0);
    return xnorm;
}

define xfig_get_normalized_y_position(p,yval) {
%!%+
%\function{xfig_get_normalized_y_position}
%\synopsis{Compute normalized (world0) y-position from world1 y-position}
%\usage{ynorm=xfig_get_normalized_y_position(p,yval)}
%\description
%For a given xfig plot p and y-coordinate value yval given in the world1
%coordinate system, compute the corresponding normalized world0 y-value.
%
%This function is necessary since the xfig module does not export
%its world normalization functions. It is heavily based on these.
%
%\seealso{xfig_get_normalized_position,xfig_get_normalized_x_position}
%!%-
    variable ywcs=p.plot_data.y1axis.wcs_transform;
    variable cd=ywcs.client_data;
    variable ww=double(p.get_world());
    variable f=ywcs.wcs_func;
    variable t0 = (@f)(ww[2], cd);
    variable t1 = (@f)(ww[3], cd);

    variable ynorm=((@f)(double(yval), cd) - t0)/(t1-t0);
    return ynorm;
}

define xfig_get_normalized_position(p,xval,yval) {
%!%+
%\function{xfig_get_normalized position}
%\synopsis{Compute normalized (world0) x,y-position from world1 x,y-position}
%\usage{xnorm=xfig_get_normalized_position(p,xval,yval)}
%\description
%For a given xfig plot p and coordinate values xval,yval given in the world1
%coordinate system, compute the corresponding normalized world0 x,y-position.
%
%This function is necessary since the xfig module does not export
%its world normalization functions. It is heavily based on these.
%
%\seealso{xfig_get_normalized_x_position,xfig_get_normalized_y_position}
%!%-
    variable xwcs=p.plot_data.x1axis.wcs_transform;
    variable ywcs=p.plot_data.y1axis.wcs_transform;
    variable xcd=xwcs.client_data;
    variable ycd=ywcs.client_data;
    variable ww=double(p.get_world());
    variable xf=ywcs.wcs_func;
    variable xt0 = (@xf)(ww[0], xcd);
    variable xt1 = (@xf)(ww[1], xcd);

    variable yf=ywcs.wcs_func;
    variable yt0 = (@yf)(ww[2], ycd);
    variable yt1 = (@yf)(ww[3], ycd);

    variable xnorm=((@xf)(double(xval), xcd) - xt0)/(xt1-xt0);
    variable ynorm=((@yf)(double(yval), ycd) - yt0)/(yt1-yt0);
    return xnorm,ynorm;
}

%
% Code from src/plot/xfig_plot_conf.sl
%

% only activate if module gcontour could be loaded
#ifeval __get_reference("gcontour_compute")!=NULL

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

define xfig_plot_conf()
%!%+
%\function{xfig_plot_conf}
%\synopsis{plots contour levels stored with save_conf}
%\usage{XFig_Plot_Type pl = xfig_plot_conf(String_Type file[, XFig_Plot_Type pl])}
%\qualifiers{
%\qualifier{lvls}{specify the Delta chi-sqr values for the contour levels,
%            default: [2.30, 4.61, 9.21], i.e., 1 sigma, 90% and 99% CL}
%\qualifier{color}{specify the colors of the contour lines
%           The array's last value is used for the best fit's cross (if
%           length of array > number of levels).}
%\qualifier{line}{[=0] specify the line styles of the contour lines}
%\qualifier{width}{[=1] specify the width of the contour lines
%                  The array's last value is used for the best fit's cross.}
%\qualifier{smooth}{[=1] this integer>=1 specifies the interpolation factor
%                   for the image used to calculate the contours}
%\qualifier{W}{width of the plot if created by \code{xfig_plot_contours}}
%\qualifier{H}{height of the plot if created by \code{xfig_plot_contours}}
%\qualifier{worldXY}{any world qualifier to the xfig-plot will be passed,
%                    which requires an existing SLxfig plot object to be given}
%\qualifier{transpose}{transposes the countour map (exchange x- and y-values)}
%\qualifier{nocross}{do not plot best fit cross}
%\qualifier{noclip}{if used with a sltikz plot object, do not clip}
%}
%\description
%    If no SLxfig plot object \code{pl} is given, a new one is created
%    and its world coordinate system is defined according to the
%    ranges used to calculate the confidence map. x- and y-labels
%    are initialized with the parameter names, but can be changed
%    afterwards.
%\seealso{get_confmap, xfig_plot_new, enlarge_image}
%!%-
{
   % TODO: It should also be possible to use ISIS' conf_map structure
   %       as argument for this function, not only a save_conf'ed file.
   %       The values that are currently read from the FITS keywords
   %       can certainly also be derived from a load_conf'ed structure.
   %
   %       [Mh]

   variable file, pl=NULL;
   switch(_NARGS)
   { case 1:  file      = (); }
   { case 2: (file, pl) = (); }
   { help(_function_name()); return; }

   variable lvls = qualifier("lvls",[2.2957,4.6052,9.21034]);
   variable n_levels = length(lvls);
   variable col = qualifier("color", [1:n_levels]);
   variable line = qualifier("line", 0);
   variable width = qualifier("width", 1);
   variable fac = qualifier("smooth",1);
   variable clrmap = qualifier("colormap","ds9b");
   variable trns = qualifier_exists("transpose"); 
   variable cross = not(qualifier_exists("nocross"));
   variable qnames = get_struct_field_names(__qualifiers);
   variable world = qnames[where(
   array_map(String_Type, &substr, qnames, 1, 5) == "world")];

   if (qualifier_exists("noclip")) {
       world=struct_combine(world,"noclip");
   }
   
   if(n_levels>1)
   {
     if(length(col)==1)
     {
       variable cola = typeof(col)[n_levels];
       cola[*] = col;
       col = cola;
     }
     if(length(line)==1)  line += Integer_Type[n_levels];
     if(length(width)==1)  width += Integer_Type[n_levels];
   }

   % #1: read the file and get the information
   variable map = fits_read_img (file);
   if (trns) map = transpose(map);

   variable xmin = fits_read_key(file,trns ? "pymin" : "pxmin");
   variable xmax = fits_read_key(file,trns ? "pymax" : "pxmax");
   variable xnum = fits_read_key(file,trns ? "pynum" : "pxnum");

   variable ymin = fits_read_key(file,trns ? "pxmin" : "pymin");
   variable ymax = fits_read_key(file,trns ? "pxmax" : "pymax");
   variable ynum = fits_read_key(file,trns ? "pxnum" : "pynum");

   variable del_x = fits_read_key(file,trns ? "CDELT2P" : "CDELT1P");
   variable del_y = fits_read_key(file,trns ? "CDELT1P" : "CDELT2P");

   variable best_x = fits_read_key(file,trns ? "BEST_Y" : "BEST_X");
   variable best_y = fits_read_key(file,trns ? "BEST_X" : "BEST_y");

   % #1b: enlarge map and hence smooth contour
   if (qualifier_exists("smooth"))
   {
      map = enlarge_image(map,fac,fac);
      del_x /= fac;
      del_y /= fac;
   }

   % #2: compute the confidence levels
   variable gct  = gcontour_compute(map,lvls);
   variable dims = array_shape(map);

%   variable del_x = 1.*(xmax-xmin)/(dims[0]);
%   variable del_y = 1.*(ymax-ymin)/(dims[1]);


   % create a new plot object if none (or NULL) given
   if(pl==NULL)
   {
     pl = xfig_plot_new(qualifier("W", 14), qualifier("H", 10));
     pl.world(xmin, xmax-del_x, ymin, ymax-del_y);
     pl.xlabel(`\verb|`+fits_read_key(file,trns ? "CTYPE2P" : "CTYPE1P")+"|");
     pl.ylabel(`\verb|`+fits_read_key(file,trns ? "CTYPE1P" : "CTYPE2P")+"|");
   }

   % #3: plot the confidence levels
   variable i,j;
   _for i(0, n_levels-1, 1)
   {
      _for j(0,length(gct[i].x_list)-1,1)
      {
	 pl.plot(del_x*gct[i].x_list[j]+xmin,
	         del_y*gct[i].y_list[j]+ymin;; struct_combine(world,
                 struct { color=col[i], line=line[i], width=width[i] }
	 ));
      }
   }
  if(cross){
    pl.plot(best_x,best_y;; struct_combine(world,
      struct { color=length(col) > n_levels ? col[-1] : "black",
               width=width[-1], sym=4 }));
  }
#ifexists png_gray_to_rgb
   if (qualifier_exists("plot_map"))
    {
      if (qualifier_exists("crop_map")){
	 pl.plot_png(png_gray_to_rgb (map [ [1:2*(length(map[*,0])-1)]/2, [1:2*(length(map[0,*])-1)]/2 ] , clrmap));
      } else {
        pl.plot_png(png_gray_to_rgb (sqrt(map), clrmap);cmap=clrmap);
      }
    }
#endif
   return pl;
}


#endif

#endif

#endif

%
% Code from src/plot/xfig_plot_confmap.sl
%

% only activate if module gcontour could be loaded
#ifeval __get_reference("gcontour_compute")!=NULL

% only activate if module gsl could be loaded
#ifeval __get_reference("gsl->convolve")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL
% try to load  modules if available


define xfig_plot_confmap() %{{{
%!%+
%\function{xfig_plot_confmap}
%\synopsis{Create an xfig plot of a confidence map}
%\usage{xfig_plot_new fig = xfig_plot_confmap(String_Type cm; qualifiers)}
%\description
%    Create an xfig plot from a pre-calculated confidence map `cm' computed and saved with
%    the function `get_confmap'.
%\qualifiers{
%\qualifier{best_fit}{Mark the best fit with a cross.}
%\qualifier{CDF}{If set to a positive integer, the chi^2 values are converted to a cumulative
%      distribution function with the degree of freedom given by this qualifier. For
%      instance, CDF=1 implies that chi^2 values of 1, 2.71, and 6.63 are converted
%      to 0.68, 0.9, and 0.99 (this qualifier requires the gsl-module).}
%\qualifier{colormap [="haxby"]}{Choose, e.g., "ds9sls", "seis", "rainbow", "haxby", "topo",
%      "globe" (see the function `png_get_colormap' for more information).}
%\qualifier{colormap_gmin, colormap_gmax}{Minimum and maximum grayscale value for the colormap,
%      see `png_gray_to_rgb'.}
%\qualifier{chi2 [=Double_Type[0]]}{Chi^2 values for which contour lines should be drawn (this
%      qualifier requires the gcontour-module).}
%\qualifier{chi2_color [="magenta"]}{Color of the contour lines and the best fit cross.}
%\qualifier{factor [=1]}{Factor by which the x- and y-sampling of the confidence map is increased.}
%\qualifier{field [="chisqr"]}{Which field of the confidence map shall be plotted.}
%\qualifier{width [=8], height [=8]}{Width and height of the confidence map.}
%\qualifier{latex_package [="txfonts"]}{Load a package in the preamble of LaTeX documents.}
%\qualifier{reverse_axes}{If present, both axes are reversed.}
%\qualifier{[xyz]label}{Add a x-, y-, or z-label to the confidence map.}
%\qualifier{[xyz]axis [=struct{ticlabels_confine=0, ticlabels2=0}]}{Modify the x-, y-, or z-axis
%      by providing a structure whose fields are qualifiers of the function `xfig_plot.axis'.
%      Note that the `log' qualifier (set logarithmic axis scale) does not work.}
%\qualifier{nomultiplot}{Return the confidence map and its corresponding colormap in two separate
%      xfig objects instead of one multiplot object.}
%}
%\example
%    fig = xfig_plot_confmap("confmap.fits"; chi2=[1, 2.71, 6.63], zlabel="$\\chi^2$"R, zaxis=struct{format="%g", ticlabels1=0});
%    fig.render("test.pdf");
%\seealso{get_confmap}
%!%-
{
  if(_NARGS!=1)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable fp;
  fp = ();
  variable width = qualifier("width", 8);
  variable height = qualifier("height", 8);
  variable factor = qualifier("factor", 1);
  variable colormap = qualifier("colormap", "haxby");
  variable chi2 = qualifier("chi2", Double_Type[0]);
  variable chi2_color = qualifier("chi2_color", "magenta");
  variable field_to_be_plotted = qualifier("field", "chisqr");
  variable reverse = qualifier_exists("reverse_axes");
  xfig_add_latex_package(qualifier("latex_package", "txfonts")); % txfonts as used in A&A
  variable i;
  % -----------
  % read input:
  variable cm = fits_read_table(fp; casesen);
  ifnot(struct_field_exists(cm,field_to_be_plotted))
  {
    print(get_struct_field_names(cm));
    throw UsageError, sprintf("Usage error in '%s': Field '%s' does not exist. Check the list of available fields above.", _function_name(), field_to_be_plotted);
  }
  variable keys = fits_read_key_struct(fp, "pxnum",  "pynum", "pxname", "pyname", "best_x", "best_y", "beststat");
  %keys.pxname = strreplace(keys.pxname, "(1).", "1_");
  %keys.pyname = strreplace(keys.pyname, "(1).", "1_");
  keys.pxname = escapedParameterName(keys.pxname);
  keys.pyname = escapedParameterName(keys.pyname);
  variable min_chi = min(get_struct_field(cm, "chisqr"));
  if(min_chi < keys.beststat)
  {
    variable ind_min = where(get_struct_field(cm, "chisqr")==min(get_struct_field(cm, "chisqr")));
    vmessage("Warning in '%s': Found fit better than best-fit assumed by confidence map:", _function_name());
    vmessage("Warning in '%s': Original best-fit with chi-square of %f at position (%f, %f)", _function_name(), keys.beststat, keys.best_x, keys.best_y);
    _for i(0, length(ind_min)-1, 1)
      vmessage("Warning in '%s': Found chi-square of                  %f at position (%f, %f)", _function_name(), get_struct_field(cm, "chisqr")[ind_min[i]],
	       get_struct_field(cm, keys.pxname)[ind_min[i]], get_struct_field(cm, keys.pyname)[ind_min[i]]);
    % update keys.beststat, keys.best_x, keys.best_y:
    keys.beststat = min_chi;
    keys.best_x = get_struct_field(cm, keys.pxname)[ind_min[i]];
    keys.best_y = get_struct_field(cm, keys.pyname)[ind_min[i]];
  }
  % -----------
  % ------------------------------------
  % Smooth confmap via 2D interpolation:
  set_struct_field(cm, "chisqr", sqrt(get_struct_field(cm, "chisqr")-min_chi)); % convert `chi^2' to `sqrt(delta chi^2)' ~ `delta chi' -> linear interpolation (see below) works better
  variable fieldnames = union(keys.pyname, keys.pxname, field_to_be_plotted, "chisqr");
  struct_filter(cm, array_sort(get_struct_field(cm, keys.pyname))); % sort by y-values
  struct_filter(cm, array_sort(get_struct_field(cm, keys.pxname))); % sort by x-values -> order of equal elements in y-values is preserved, see method `merge-sort' in help on function `array_sort'
  variable old_x = union(get_struct_field(cm, keys.pxname));
  % divide each interval in [a:b:#n] in x parts: [a:b:#(n*x-x+1)] because (n-1) segments need (x-1) additional points plus the origional number of points n: (n-1)*(x-1)+n=n*x-x+1
  variable new_x = [old_x[0]:old_x[-1]:#(keys.pxnum*factor-factor+1)];
  variable old_y = union(get_struct_field(cm, keys.pyname));
  variable new_y = [old_y[0]:old_y[-1]:#(keys.pynum*factor-factor+1)];
  variable field;
  foreach field (fieldnames)
  {
    reshape(get_struct_field(cm, field), [keys.pxnum,keys.pynum]);
    variable temp1 = Double_Type[length(new_x), keys.pynum]; % temporary variable
    _for i(0, keys.pynum-1, 1)
      temp1[*,i] = interpol(new_x, old_x, (get_struct_field(cm, field))[*,i]); % add additional points for x-grid
    variable temp2 = Double_Type[length(new_x), length(new_y)]; % finer 2D-array for field
    _for i(0, length(new_x)-1, 1)
      temp2[i,*] = interpol(new_y, old_y, temp1[i,*]); % add additional points for y-grid
    temp2 = transpose(temp2); % to have correct x- and y-axes
    if(reverse)
      array_reverse(temp2); % reverse axes
    set_struct_field(cm, field, temp2);
  }
  if(reverse)
  {
    array_reverse(new_x); % to account for array_reverse(temp2)
    array_reverse(new_y); % to account for array_reverse(temp2)
  }
  set_struct_field(cm, "chisqr", get_struct_field(cm, "chisqr")^2); % convert `delta chi' to `delta chi^2'
  % ------------------------------------
#ifeval __get_reference("gsl->cdf_chisq_P")!=NULL
  if(qualifier_exists("CDF") && typeof(qualifier("CDF"))==Integer_Type)
  {
    set_struct_field(cm, "chisqr", gsl->cdf_chisq_P(get_struct_field(cm, "chisqr"), qualifier("CDF")));
    chi2 = gsl->cdf_chisq_P(chi2, qualifier("CDF"));
  }
#endif
  variable fig = xfig_new_compound();
  % --------
  % colormap
  variable w1 = xfig_plot_new(0.075*width, height);
  variable mi, ma;
  if(qualifier_exists("colormap_gmin"))
    mi = qualifier("colormap_gmin");
  else
    mi = min(get_struct_field(cm, field_to_be_plotted));
  if(qualifier_exists("colormap_gmax"))
    ma = qualifier("colormap_gmax");
  else
    ma = max(get_struct_field(cm, field_to_be_plotted));
  w1.world(0, 0.075*width, mi, ma);
  w1.xaxis(; ticlabels=["0","0"], major=[0, 0.075*width], ticlabel_color="white", ticlabels2=0);
  if(qualifier_exists("zaxis") && typeof(qualifier("zaxis"))==Struct_Type)
    w1.yaxis(;; reduce_struct(qualifier("zaxis"), "log"));
  else
    w1.yaxis(; ticlabels_confine=0, ticlabels1=0);
  if(qualifier_exists("zlabel"))
    w1.xlabel(qualifier("zlabel"));
  w1.plot_png(_reshape([0:255],[256,1]); depth=100, cmap=colormap);
  variable CL;
  foreach CL (chi2)
  {
    if(field_to_be_plotted=="chisqr")
      w1.plot([0,0.075*width], [CL,CL] ; line=0, width=2, depth=99, color=chi2_color);
  }
  w1.translate(vector(1.01*width, 0, 0));
  fig.append(w1);
  % --------
  % -------
  % confmap
  variable w2 = xfig_plot_new(width, height);
  if(reverse)
    w2.world(new_x[0]+.5*(new_x[0]-new_x[1]), new_x[-1]-.5*(new_x[0]-new_x[1]), new_y[0]+.5*(new_y[0]-new_y[1]), new_y[-1]-.5*(new_y[0]-new_y[1]));
  else
    w2.world(new_x[0]-.5*(new_x[1]-new_x[0]), new_x[-1]+.5*(new_x[1]-new_x[0]), new_y[0]-.5*(new_y[1]-new_y[0]), new_y[-1]+.5*(new_y[1]-new_y[0]));
  if(qualifier_exists("xaxis") && typeof(qualifier("xaxis"))==Struct_Type)
    w2.xaxis(;; reduce_struct(qualifier("xaxis"),"log"));
  else
    w2.xaxis(; ticlabels_confine=0, ticlabels2=0);
  if(qualifier_exists("yaxis") && typeof(qualifier("yaxis"))==Struct_Type)
    w2.yaxis(;; reduce_struct(qualifier("yaxis"),"log"));
  else
    w2.yaxis(; ticlabels_confine=0, ticlabels2=0);
  w2.plot_png(get_struct_field(cm, field_to_be_plotted); cmap=colormap, depth=100, gmin=mi, gmax=ma);
  if(qualifier_exists("xlabel"))
    w2.xlabel(qualifier("xlabel"));
  if(qualifier_exists("ylabel"))
    w2.ylabel(qualifier("ylabel"));
  % --------------------
#ifeval __get_reference("gcontour_compute")!=NULL
  % chi^2 contour lines:
  foreach CL (gcontour_compute(get_struct_field(cm, "chisqr"), chi2))
  {
    variable cl;
    _for cl (0, length(CL.x_list)-1, 1)
      w2.plot( new_x[0] + (new_x[1]-new_x[0]) * CL.x_list[cl], new_y[0] + (new_y[1]-new_y[0]) * CL.y_list[cl]; line=0, width=2, depth=99, color=chi2_color);
  }
#endif
  % --------------------
  % -------
  % -----------------------
  % Mark best fit with "+":
  if(qualifier_exists("best_fit"))
    w2.plot(keys.best_x, keys.best_y; sym="+", width=1, color=chi2_color);
  % -----------------------
  fig.append(w2);
  %
  if(qualifier_exists("nomultiplot"))
    return w2, w1;
  else
    return fig;
}%}}}


#endif

#endif

#endif

%
% Code from src/plot/xfig_plot_epfpd.sl
%

% only activate if module gcontour could be loaded
#ifeval __get_reference("gcontour_compute")!=NULL

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

define xfig_plot_epfpd (ep, lc)
%!%+
%\function{xfig_plot_epfpd}
%\synopsis{plot the results of epfoldpdot with Xfig, including the
%pulse profile of the stronges P/P-dot values}
%\usage{xfig_plot_epfpd(Struct_Type eppd, Struct_Type lc)
%}
%\qualifiers{
%\qualifier{W}{width of resulting plot (default=12)}
%\qualifier{H}{height of resulting plot (default=W)}
%\qualifier{pdstart}{start of P-dot axis (default read from eppd)}
%\qualifier{pdstop}{end of P-dot axis (default read from eppd)}
%\qualifier{pstart}{start of P axis (default read from eppd)}
%\qualifier{pstop}{end of P axis (default read from eppd)}
%\qualifier{p0}{center of P  axis (default mean of periods in eppd)}
%\qualifier{cmap}{[\code{="iceandfire"}] Color-map of Chi^2 landscape }
%\qualifier{cont}{switch to plot FWHM contour on map }
%\qualifier{pp}{switch to plot pulse profile below map }
%\qualifier{t0}{[\code{="0"}] t0 for pulse profile }
%\qualifier{nbins}{[\code{="12"}] number of phase bins for pulse profile }
%}
%           
%\description
%    This function plots the results for "epfoldpdot" into a nice
%    colorful xfig image, with the cuts in P and P-dot over the most
%    significant period as well as (if desired) the pulse profile of
%    that period. This style of plot is inspired by the results plots
%    of PRESTO.
%
%    This code is still under development and not very flexible yet.
%    Feel free to improve it.       
%
%  \seealso{epfoldpdot}
%!%-
{
   variable W = qualifier("W", 12) ;
   variable H = qualifier("H", W) ;
   
   variable xf = xfig_plot_new(W,H) ;

   variable pdstart = qualifier("pdstart", min(ep.pd)) ;
   variable pdstop =  qualifier("pdstop", max(ep.pd)) ;
   variable pstart = qualifier("pstart", min(ep.p)) ;
   variable pstop =  qualifier("pstop", max(ep.p)) ;

   variable p0 = qualifier("p0", mean(ep.p)) ;

   xf.world(pdstart, pdstop,pstart-p0,pstop-p0) ;
   xf.xlabel("$\\dot{P}$ [s/s]") ;
   xf.ylabel(sprintf("$P$ [s] - %.5f", p0)) ;

   xf.plot_png(ep.stat ; cmap=qualifier("cmap","iceandfire")) ;

#ifexists gcontour_compute
   if (qualifier_exists("cont"))
     {
	variable fwhm = gcontour_compute(ep.stat, [max(ep.stat)/2.]) ;

	xf.plot(fwhm[0].x_list[0]*(ep.pd[1]-ep.pd[0]) + ep.pd[0],
		(fwhm[0].y_list[0]*(ep.p[1]-ep.p[0]) + ep.p[0] - p0) ;
		line=0, width=2 , color="green2") ;

	xf.plot(min(fwhm[0].x_list[0]*(ep.pd[1]-ep.pd[0]) + ep.pd[0])[[0,0]], [-1,1] ; color="red") ;
     }
#endif

   variable pdmaxn = where_max(max(ep.stat,0)) ;
   variable pmaxn = where_max(max(ep.stat,1)) ;


   xf.plot(ep.pd[pdmaxn], ep.p[pmaxn] ; sym="+") ;

   variable x = ep.stat[pmaxn,*] ;
   variable y = ep.stat[*,pdmaxn] ;

   variable xx = xfig_plot_new(W, H/4.) ;
   xx.world(pdstart,pdstop,min(ep.stat), max(ep.stat) );
   xx.plot(ep.pd, x) ;
   xx.xaxis(;ticlabels=0) ;
   xx.ylabel("$\\chi^2$") ;

   variable xy = xfig_plot_new(W/4., H) ;
   xy.world(min(ep.stat), max(ep.stat),pstart,pstop) ;
   xy.plot(y, ep.p) ;
   xy.yaxis(;ticlabels=0) ;
   xy.xlabel("$\\chi^2$") ;

   variable xf1 = xfig_new_vbox_compound(xx,xf);
   variable xf2= xfig_new_hbox_compound(xf1,xy ; just=-0.93) ;
    
      
   if (qualifier_exists("pp"))
     {
	variable pp = pfold(lc.time, lc.rate, ep.p[pmaxn], lc.error ; pdot=ep.pd[pdmaxn],  nbins=qualifier("nbins", 12),
			    t0=qualifier("t0", 0), dt=lc.fracexp) ;

	variable xpp=xfig_plot_new(W, H/4.);
	xpp.xlabel("Phase") ;
	xpp.ylabel("cts/s") ;
   
	xpp.hplot([pp.bin_lo, pp.bin_lo+1],[pp.value, pp.value],[pp.error,pp.error]) ;
   
	return xfig_new_vbox_compound(xf2,xpp) ;
     }

   else
     return xf2 ;

}

#endif

#endif

#endif

%
% Code from src/plot/xfig_plot_image.sl
%

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

define xfig_plot_image()
%!%+
%\function{xfig_plot_image}
%\synopsis{Plots an image (with x/y-grid)}
%\usage{Struct_Type xfig_plot_new( Double_Type[n,m] IMG, [Double/Struct_Type x, y] );}
%\altusage{xfig_plot_new( Double_Type[n,m] IMG, [Double/Struct_Type x, y [, Struct_Type xf]] );}
%\qualifiers{
%\qualifier{cmap}{[="ds9b"] Either a name (String_Type) of a colormap
%                     (see png_get_colormap_names), or a self defined
%                      colormap (Integer_Type[]).}
%\qualifier{gmin}{[=min(IMG)] Minimal value of image to be plotted}
%\qualifier{gmax}{[=max(IMG)] Maximal value of image to be plotted}
%\qualifier{xmin}{[=min(x)] Minimal value of x-axis}
%\qualifier{xmax}{[=max(x)] Maximal value of x-axis}
%\qualifier{ymin}{[=min(y)] Minimal value of y-axis}
%\qualifier{ymax}{[=max(y)] Maximal value of y-axis}
%\qualifier{dx}{[=0.] Relative justification of pixels in x-direction. Used if 'x' is no grid.}
%\qualifier{dy}{[=dx] Relative justification of pixels in y-direction. Used if 'y' is no grid.}
%\qualifier{fill}{[=20] global filling method (see %.shade_region) }
%\qualifier{fmap}{[=Integer_Type[n,m]+fill] individual filling method for each pixel}
%%\qualifier{nancol}{[="black"] Extra color for IMG pixel values, which are NAN.}
%\qualifier{ratio}{[=1] Desired ratio of x and y-axis scale}
%\qualifier{W}{[=14] Plot width [cm]. 'ratio' is overwritten if W & H are given!}
%\qualifier{H}{[=W/ratio] Plot height [cm]. 'ratio' is overwritten if W & H are given!}
%\qualifier{colorscale}{If given an inlay colorscale is plotted using xfig_plot_colorscale}
%\qualifier{format}{Format for colormap axis ticlabels, see 'help xfig_plot.axis'}
%\qualifier{just}{[=[.95,.05,-.5,.5]] Position (world00) & relative justification
%         of the colorscale.}
%\qualifier{scale}{[=[.33,.05]] : Relative size of the colorscale in respect to W & H.}
%\qualifier{xlabel}{Label of x-axis}
%\qualifier{ylabel}{Label of y-axis}
%\qualifier{depth}{[=150] xfig depth}
%}
%\description
%   Other than the %plot_png function of the xfig_plot structure this function
%   aims to plot images with a (given) ARBITRARY x/y-grid in a correct way
%   (NOTE that .plot_png does not care for x/y values!). This also allows to
%   plot images with logarithmic scales without any problems! In addition
%   it is easy to plot gabbed images by just giving the x and y grid as bin_lo
%   and bin_hi.
%
%   ARGUMENTS
%    IMG:
%    In any case a IMaGe has to be given, which can arbitrary dimensions n and m
%    (it even can be 1-dimensional).
%
%    x, y:
%    x and y-grids are optional arguments but cannot be given individually.
%    If no grid is given the grids are set to the pixel number assuming a linear
%    grid! The grids can be Array_Types either of length n and m, respectively, in
%    which case the bin_lo and bin_hi is created by assuming a linear grid.
%    Otherwise their length must be n+1, m+1!
%    It is also possible that x and y are Struct_Types already containing the bin_lo
%    and bin_hi (it is only required that the bin_lo filed has 'lo' in its name and
%    the bin_hi field 'hi', respectively). In that last case it is also possible to
%    plot images with gabs!
%    
%    xf:
%    Must be and xfig_plot structure. If given the image is plotted into xf and there
%    will be no return value. This way all automatic plot settings of this function will
%    be disregarded. Otherwise a new xfig_plot is created (qualifiers are
%    passed through).
%    
%\example
%   % Example 1: logscale
%   variable x = [PI:10*PI:#100];
%   variable y = [PI:5*PI:#80];
%   variable IMG = sin(x) # transpose(cos(y));
%   variable xf = xfig_plot_image( IMG, x*180/PI, y*180/PI;
%                                         padx=0.025, pady=0.025,
%      		                          dx=-.5,
%                                         xlog, ylog,
%				          fill=20
%				        );
%   xf.render("/tmp/test.pdf");
%
%   % Example 2: image of Example 1 but with gaps
%   variable tmp = [[1:30],[35:65],[75:115]];
%   variable x = struct{ lo = tmp/11.5*PI, hi = (tmp+1)/11.5*PI };
%   tmp = [[1:40],[50:90]];
%   variable y = struct{ lo = tmp/18.*PI, hi = (tmp+1)/18.*PI };
%   variable IMG = sin(x.lo) # transpose(cos(y.lo));
%   variable xf = xfig_plot_image( IMG, x, y ; cmap="drywet");
%   xf.render("/tmp/test.pdf");
%
%\seealso{xfig_plot_new, %.shade_region, xfig_plot_colormap}
%
%!%-
{
  %%%%% ARGUMENTS
  variable IMG,
    x  = NULL,
    y  = NULL,
    xf = NULL;
  
  switch(_NARGS)
  { case 1: IMG = (); }
  { case 3: ( IMG, x, y ) = (); }
  { case 2: ( IMG, xf ) = (); }
  { case 4: ( IMG, x, y, xf ) = (); }

  %%%%% QUALIFIERS

  %% LOAD ColorMap
  variable cmap = qualifier("cmap","ds9b");
  if( _typeof(cmap) == String_Type ){
    if( int(sum(cmap==png_get_colormap_names)) == 0 ){
      vmessage("EEROR(%s): Colormap '%s' does not exist!",
	       _function_name, cmap );
      return NULL;
    }
    cmap = png_get_colormap(cmap);
  }
  %% Scaling in z-direction
  variable gmin = qualifier("gmin",min(IMG));
  variable gmax = qualifier("gmax",max(IMG));
  % variable scalefun = qualifier("scalefun", NULL);
  % if (scalefun != NULL) {
  %   gmin = (@scalefun)(gmin);
  %   gmax = (@scalefun)(gmax);
  %   IMG = (@scalefun)(IMG);
  % }
  
  %% Fillmethod
  variable fmap = qualifier("fmap",NULL);
  variable fill = qualifier("fill",20);
  variable nancol = qualifier("nancol","black");
  
  %% Pixel positioning
  variable dx = qualifier("dx",0.);
  variable dy = qualifier("dy",dx);

  %% Plot size
  variable ratio = qualifier("ratio",1);
  variable W = qualifier("W",14);
  variable H = qualifier("H",NULL); 
  variable padx = qualifier("padx",0.);
  variable pady = qualifier("pady",0.);

  variable depth = qualifier("depth",150);

  %% Label
  variable xlabel = qualifier("xlabel");
  variable ylabel = qualifier("ylabel");
  
  %% COLORscale
  variable colorscale = qualifier("colorscale");
  variable just = qualifier("just", [.95,.05,.5,-.5] );
  variable scale = qualifier("scale", [.33,.05] );
  
  %%%%% LET THE MAGIC HAPPEN
  
  %% Obain Image dimensions
  variable dim = array_shape( IMG );
  if( length(dim) == 1 ){
    dim = [dim[0],1];
    reshape( IMG, dim );
  }
  variable len = int(prod(dim));

  %% Create FILLmap
  if( fmap == NULL ){
    fmap = Integer_Type[dim[0],dim[1]] + fill;
  }
  
  %% Obtain x grids
  variable xlo, xhi, ylo, yhi;
  variable fname, del;
  if( x == NULL ){
    xlo = [0:dim[0]-1];
    xhi = xlo + 1;
  }
  else if( typeof(x) == Struct_Type ){
    fname = get_struct_field_names(x);
    xlo = get_struct_field( x, fname[where(is_substr(fname,"lo"))][0] );
    xhi = get_struct_field( x, fname[where(is_substr(fname,"hi"))][0] );
  }
  else if( typeof(x) == Array_Type ){ 
    if( length(x) == dim[0]+1 ){
      xlo = x[[:-2]];
      xhi = x[[1:]];
    }
    else if( length(x) == dim[0] ){
      del = diff(x);
      xlo = x - (.5+dx) * [del,del[-1]];
      xhi = x + (.5-dx) * [del,del[-1]];
    }
    else{
      vmessage("ERROR(%s): Dimension of x-grid does not match IMG!",_function_name);
    }
  }
  %% Obtain y grids
  if( y == NULL ){
    ylo = [0:dim[1]-1];
    yhi = ylo + 1;
  }
  else if( typeof(y) == Struct_Type ){
    fname = get_struct_field_names(y);
    ylo = get_struct_field( y, fname[where(is_substr(fname,"lo"))][0] );
    yhi = get_struct_field( y, fname[where(is_substr(fname,"hi"))][0] );
  }
  else if( typeof(y) == Array_Type ){
    if( length(y) == dim[1]+1 ){
      ylo = y[[:-2]];
      yhi = y[[1:]];
    }
    else if( length(y) == dim[1] ){
      del = diff(y);
      ylo = y - (.5+dy) * [del,del[-1]];
      yhi = y + (.5-dy) * [del,del[-1]];
    }
    else{
      vmessage("ERROR(%s): Dimension of y-grid does not match IMG!",_function_name);
    }
  }

  variable xmin = qualifier("xmin",xlo[0]);
  variable xmax = qualifier("xmax",xhi[-1]);
  variable ymin = qualifier("ymin",ylo[0]);
  variable ymax = qualifier("ymax",yhi[-1]);
  
  %% If not given, initialize a xfig_plot
  variable ret = 0;
  if( xf == NULL ){
    if( H == NULL ){
      H = abs(1.*W / ( ratio*(xhi[-1]-xlo[0])/(yhi[-1]-ylo[0])*(1+padx)/(1+pady) ));
    }
    else if( qualifier_exists("H") and not(qualifier_exists("W")) ){
      W = abs(1.*H * ( ratio*(xhi[-1]-xlo[0])/(yhi[-1]-ylo[0])*(1+padx)/(1+pady) ));
    }
    xf = xfig_plot_new( W, H ;; struct_combine(__qualifiers,struct{padx=padx,pady=pady}) );
    xf.world( xmin, xmax, ymin, ymax ;; __qualifiers );
    xf.xlabel(xlabel);
    xf.ylabel(ylabel);
    ret = 1;
  }  

  %% Obtain colors from IMG & CMAP accounting for z-scaling
  variable col = png_gray_to_rgb ( IMG , cmap ; gmin=gmin, gmax=gmax );
  col = array_map(String_Type, &sprintf, "#%06X", col );
  col[ where(isnan(IMG) or isinf(IMG)) ] = nancol;
  
  if( qualifier_exists("group") ){ %% TODO
    %% Group pixel of same color to one shade region if possibile!
    variable U = unique(col);
    variable I, ix, iy, xa, ya;
    _for $1 ( 0, length(U)-1 ){
    }
  }else{
    %% Plot each pixel individually!
    _for $1 ( 0, length(xlo)-1 ){
      _for $2 ( 0, length(ylo)-1 ){
	xf.shade_region( xlo[$1], xhi[$1], ylo[$2], yhi[$2] ;;
			 struct_combine( struct{ width=0, depth=depth,
			   color=col[$1,$2],
			   fillcolor=col[$1,$2],
			   fill=fmap[$1,$2] },
					 __qualifiers
				       )
		       );
      }
    }
  }

  %% Add colorscale
  if( qualifier_exists("colorscale") ){
    variable quals = reduce_struct(__qualifiers,["colorscale","xlog","ylog","fill","fmap"]);
    quals = struct_combine(quals, struct{depth=depth-10,W=scale[0]*W,H=scale[1]*H});
    variable cs = xfig_plot_colormap( IMG ;; quals );
    xf.add_object( cs, __push_array(just) ; world00 );
  }
  
  if( ret ){
    return xf;
  }
}


define xfig_plot_colormap()
%!%+
%\function{xfig_plot_colormap}
%\synopsis{Plots a colormap/colorscale}
%\usage{Struct_Type xfig_plot_new( Double_Type[n,m] IMG );}
%\qualifiers{
%\qualifier{cmap}{[="ds9b"] Either a name (String_Type) of a colormap
%                     (see png_get_colormap_names), or a self defined
%                      colormap (Integer_Type[]).}
%\qualifier{gmin}{[=min(IMG)] Minimal value of image to be plotted}
%\qualifier{gmax}{[=max(IMG)] Maximal value of image to be plotted}
%\qualifier{W}{[=14] Plot width [cm]. 'ratio' is overwritten if W & H are given!}
%\qualifier{H}{[=W/ratio] Plot height [cm]. 'ratio' is overwritten if W & H are given!}
%\qualifier{orientation}{[=1] Horizontal if 1, vertical if 0. Set Horizontal if W/H > 1,
%             otherwise set to vertical.}
%\qualifier{label}{Label for the colormap}
%\qualifier{format}{Format for colormap axis ticlabels, see 'help xfig_plot.axis'}
%\qualifier{depth}{[=80] Depth of the colormap}
%\qualifier{fontsize}{[="scriptsize"] Font size of tic & axis label}
%\qualifier{ticlen}{[=.15] Length of major tic marks (minor tic are .5*ticlen).}
%\qualifier{maxtics}{[=5] Maximal number of labeled tics}
%\qualifier{box}{Draws a box behind the colormap}
%\qualifier{border}{[=.01] Relative size of box excess length}
%\qualifier{boxcolor}{[="#FFFFFF"] Color of the box}
%}
%\description
%   This function uses xfig_plot_image to create a colormap based on the given
%   IMaGe (with dimension n and m).
%   
%\example
%   variable x = [PI:10*PI:#100];
%   variable y = [PI:5*PI:#80];
%   variable IMG = sin(x) # transpose(cos(y));
%   variable xf = xfig_plot_colormap( IMG );
%   xf.render("/tmp/test.pdf");
%
%\seealso{xfig_plot_new, %.shade_region, xfig_plot_image}
%
%!%-
{
  %%%%% ARGUMENTS
  variable IMG,
    grid  = NULL;
  
  switch(_NARGS)
  { case 1: IMG = (); }

  %%%%% QUALIFIERS

  variable W = qualifier("W",4);
  variable H = qualifier("H",0.5);
  variable box = qualifier("box");
  variable border = qualifier("border",0.01);
  variable boxcolor = qualifier("boxcolor","#FFFFFF");
  variable fontsize = qualifier("fontsize","scriptsize");
  variable label = qualifier("label");
  variable depth = qualifier("depth",80);
  variable maxtics = qualifier("maxtics",5);
  variable tlen   = qualifier("ticlen",.15);
  variable format = qualifier("format","%g");
  variable orientation = qualifier("orientation", W/H > 1 ? 1 : 0 );

  if( typeof(orientation) == String_Type ){
    if( is_substr("horizontal",orientation) ){
      orientation = 1;
    }
    else{
      orientation = 0;
    }
  }
  
  %% LOAD ColorMap
  variable cmap = qualifier("cmap","ds9b");
  if( _typeof(cmap) == String_Type ){
    if( int(sum(cmap==png_get_colormap_names)) == 0 ){
      vmessage("EEROR(%s): Colormap '%s' does not exist!",
	       _function_name, cmap );
      return NULL;
    }
    cmap = png_get_colormap(cmap);
  }
  variable gmin = qualifier("gmin",min(IMG));
  variable gmax = qualifier("gmax",max(IMG));

  variable sm = [gmin:gmax:#length(cmap)];
  variable lohi = [gmin:gmax:#length(cmap)+1];
  
  variable axisquals = struct{
    ticlabel_size = fontsize,
    maxtics       = maxtics,
    major_len     = tlen,
    minor_len     =.5*tlen,
    depth         = depth-1,
    tic_depth     = depth-1,
    ticlabel_color="black",
    format        = format,
  };

  % color scale
  variable s = xfig_plot_new( W, H );
  
  %% Horizontal
  if( orientation == 1 ){
    s.world( gmin, gmax, 0, 1 );
    s.yaxis(; major = 0, minor = 0, depth=depth-1, tic_depth=depth-1 );
    s.x1axis(;;axisquals);
    s.x2axis(;;struct_combine(axisquals,struct{major_color = "white", minor_color = "white", ticlabels = 0}) );
    s.x2label( label ; size = fontsize , depth=depth-1 );
    
    xfig_plot_image( sm, lohi , [0,1], s ;; struct_combine(struct{depth=depth},reduce_struct(__qualifiers,"colorscale")) );
  }
  %% Vertical
  else{
    s.world( 0, 1, gmin, gmax );
    s.xaxis(; major = 0, minor = 0, depth=depth-1, tic_depth=depth-1);
    s.y2axis(;;axisquals);
    s.y1axis(;;struct_combine(axisquals,struct{major_color = "white", minor_color = "white", ticlabels = 0}) );
    s.ylabel( label ; size = fontsize , depth=depth-1);
    
    xfig_plot_image( transpose(sm), [0,1], lohi, s ;; struct_combine(struct{depth=depth},reduce_struct(__qualifiers,"colorscale")) );
  }

  % draw a white box behind the color scale
  if( qualifier_exists("box") ){
    variable bbox = [s.get_bbox()];
    variable bord = border*(bbox[1]-bbox[0]);
    variable b = xfig_new_rectangle( 2*bord + bbox[1]-bbox[0],
				     2*bord + bbox[3]-bbox[2]
				   );
    b.set_area_fill(20);
    b.set_fill_color(boxcolor);
    b.set_depth(depth+1);
    b.translate( vector( bbox[0]-bord,
			 bbox[2]-bord,
			 0.)
	       );
    s.add_object(b);
  }

  
  return s;
}

#endif

#endif

%
% Code from src/plot/xfig_plot_ionization_structure.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL
% -*- mode: slang; mode: fold -*-


%% necessary to make it work with slightly older SLang
%% 
#ifeval __get_reference("string_to_wchars")!=NULL

private define get_ions(abund, element){ %{{{
  variable allnames = get_struct_field_names(abund);

  variable pattern = sprintf("^%s_", strlow(element));
  variable ionnames = allnames[where(array_map(Integer_Type, &string_match, allnames, pattern))];

  return ionnames;
}
%}}}

private define capitalize(str){ %{{{
  return strsub(str, 1, toupper(string_to_wchars(str)[0]));
}
%}}}

private define filter_ionization_states(ionlabels, ionsel){ %{{{

  variable ii;
  variable nions = length(ionlabels);
  variable ions = String_Type[nions];
  variable ind = Integer_Type[nions];
  
  _for ii (0, nions - 1, 1){
    ions[ii] = strlow(ionlabels[ii][1]);
  }
  
  _for ii (0, length(ionsel)-1, 1){
    ind[where( ions == ionsel[ii])] = 1;
  }
  
  return where(ind);
}
%}}}

private define plot_single_ion(abund, element){ %{{{

  variable WIDTH = qualifier("width", 10);
  variable HEIGHT = qualifier("height", 3);
  variable XAXIS = qualifier ("xaxis", "radius");
  variable ion_states = qualifier("ion_states");
  variable xpow = qualifier("xpow", 10);

  variable abundthresh = qualifier("athresh", 0.3); % Threshold for ion abundance to be plotted
  
  if (XAXIS == "logxi"){ XAXIS = "ion_parameter";};

  ifnot (XAXIS == "radius"  or XAXIS == "ion_parameter"){
    () = printf("WARNING(%s): unknown xaxis qualifier! Using default: 'radius'\n", _function_name);
    XAXIS = "radius";
  }
    
  variable ionnames = get_ions(abund, element);
  variable ii;

  variable ionlabels = array_map (Array_Type, &strchop, ionnames, '_', 0);

  if (qualifier_exists("ion_states")){
    ionnames = ionnames[filter_ionization_states(ionlabels, ion_states)];
    ionlabels = ionlabels[filter_ionization_states(ionlabels, ion_states)];
    
  }

  variable pl = xfig_plot_new(WIDTH, HEIGHT);

  variable xval = get_struct_field(abund,XAXIS);

  if (XAXIS == "radius"){ 
    xval /= 10^(int(xpow));
  }
  
  pl.world(min(xval), max(xval), -0.1, 1.3);
  pl.yaxis(; format = "%.1f", ticlabels2 = 0);
  pl.xlabel(XAXIS == "radius" ? sprintf("radius~[$10^{%d}$cm]", int(xpow))+""R : "$\log\left(\xi\right)$"R);
  pl.ylabel(capitalize(ionlabels[0][0]));
  variable lpos;
  
  _for ii (0, length(ionnames) - 1, 1){
    pl.plot(xval, get_struct_field(abund, ionnames[ii]);; __qualifiers);

    if (max(get_struct_field(abund, ionnames[ii])) > abundthresh){ % only print labels for those ionization states that are visible
      lpos = mean( [xval[wherefirst(get_struct_field(abund, ionnames[ii]) >= abundthresh)], xval[wherelast(get_struct_field(abund, ionnames[ii]) >= abundthresh) ]] );
      % pl.xylabel(lpos, 1.1, capitalize(ionlabels[ii][0] + strup(ionlabels[ii][1])));
      pl.xylabel(lpos, 1.1, strup(ionlabels[ii][1]));
    }
  }

  % return ionlabels;
  return pl;
}
%}}}

private define plot_info_field(abund,info){ %{{{

  variable WIDTH = qualifier("width", 10);
  variable HEIGHT = qualifier("height", 3);
  variable XAXIS = qualifier ("xaxis", "radius");
  variable xpow = qualifier("xpow", 10);
  
  if (XAXIS == "logxi"){ XAXIS = "ion_parameter";};
  
  ifnot (XAXIS == "radius"  or XAXIS == "ion_parameter"){
    () = printf("WARNING(%s): unknown xaxis qualifier! Using default: 'radius'\n", _function_name);
    XAXIS = "radius";
  }

  variable pl = xfig_plot_new(WIDTH, HEIGHT);

  variable xval = get_struct_field(abund,XAXIS);
  
  if (XAXIS == "radius"){
    xval /= 10^(int(xpow));
  }
  
  variable yval = get_struct_field(abund,info);
  variable ylabel;
  
  switch (info)
  { case "ion_parameter": ylabel = "$\log\left(\xi\right)$"R; }
  { case "x_e": ylabel = "$x_\mathrm{e}$)$"R; }
  { case "n_p": ylabel = "$n_\mathrm{p}$"R; }
  { case "pressure": ylabel = "$p$"R; }
  { case "temperature": ylabel = "$T~[10^4\,\mathrm{K}]$"R; }
  {() = printf("ERROR(%s): Unknown column in abundance FITS file selected! Aborting!", _function_name); return NULL;}
  
  pl.world(min(xval), max(xval), .9*min(yval), 1.1*max(yval));
  pl.ylabel(ylabel);
  pl.xlabel(XAXIS == "radius" ? sprintf("radius~[$10^{%d}$cm]", int(xpow))+""R : "$\log\left(\xi\right)$"R);
  pl.plot(xval, yval;; __qualifiers);
  return pl;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%
define xfig_plot_ionization_structure (){ %{{{
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{xfig_plot_ionization_structure}
%\synopsis{plots the ionization structure from an abundance file produced by xstar}
%\usage{plot_ionization_structure(String_Type abundace_file, String_Type[] elements, String_Type ouput_file)}
%\qualifiers{
% \qualifier{xpow}{plot radius in units of 10^xpow (default: 10)}
% \qualifier{xaxis}{use 'logxi' or 'radius' to set the abscissa (default: radius)}
% \qualifier{temp}{add temperature panel}
% \qualifier{press}{add pressure panel}
% \qualifier{xee}{add electron fraction panel}
% \qualifier{dens}{add density panel}
% \qualifier{athresh}{only ionization states are plotted which exceed a certain threshold (default: 0.3)}
% \qualifier{ion_states}{string array of ionization states in roman numbers}
% \qualifier{width}{plot witdh}
% \qualifier{height}{plot height}
%
% All other qualifiers are passed to xfig_plot.plot()
%}
%\description
% This function produces and xfig plot of the ion abundance structure of
% an xstar calculation. The input parameters are the xstar abundance FITS
% file, a string array of the elements of interest and an output file name.
%
%\examples
%
%  xfig_plot_ionization_structure ("xout_abund1.fits", ["si", "FE", "Mg", "O"], "ion_abundance.pdf"; ion_states = ["iii", "v"], temp );
%
% plots the ion abundance of the ionization states III and V for the elements Silicon, Iron, Magnesium and Oxygen
% and the temperature as a function of radius.
%!%-

  variable abundfile, elements, filename;
  switch(_NARGS)
  {case 3: (abundfile, elements, filename) = ();}
  {help(_function_name()); return; }
  
  variable temp = qualifier_exists("temp") or qualifier_exists("temperature");
  variable press = qualifier_exists("press") or qualifier_exists("pressure");
  variable logxi = qualifier_exists("logxi");
  variable xe = qualifier_exists("xe") or qualifier_exists("x_e");
  variable np = qualifier_exists("dens") or qualifier_exists("density");
    
  variable abund = fits_read_table(abundfile + "[ABUNDANCES]");

  variable info =  where([logxi, xe, np, press, temp ]);
  variable ninfo = length(info);
  variable infoplts = Struct_Type[ninfo];

  variable infofields = ["ion_parameter", "x_e", "n_p", "pressure", "temperature"];
  
  variable ii;
  _for ii (0, ninfo - 1, 1){
    infoplts[ii] = plot_info_field(abund,infofields[info[ii]];; __qualifiers);
  }

  
  elements = [elements];
  variable nelements = length (elements);
  variable ionplts = Struct_Type[nelements];

  _for ii (0, nelements -1, 1){
    % print(elements[ii]);
    ionplts[ii] = plot_single_ion(abund, elements[ii];; __qualifiers);    
  }

  xfig_multiplot([infoplts,ionplts]).render(filename);
}
%}}}

#endif
#endif

%
% Code from src/plot/xfig_plot_params.sl
%

% only activate if module png could be loaded
#ifeval __get_reference("png_add_colormap")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL
% -*- mode:slang; mode:fold -*-

define xfig_plot_params()
%!%+
%\function{xfig_plot_params}
%\synopsis{Returns a xfig-plot of the given parameter (value, limits, [conf, chi2])}
%\usage{ Struct_Type xplot = xfig_plot_params( Struct_Type params1, params2, ... );}
%\qualifiers{
%\qualifier{tmpdir}{[='/tmp/'] Directory for temporar files}
%\qualifier{size}{[=[12,9]]: Dimension in cm of the (main) plot}
%\qualifier{space}{[=.05*size[0]]: Space between parameter columns.
%                    Dimension in cm of the (main) plot}
%\qualifier{xwidth}{[=.5]: Width of the parameter symbols (same units as x values).}
%\qualifier{x}{Alternative values for the x-axis (e.g., time in MJD)}
%\qualifier{xlabel}{Label for the alternative x-axis.}
%\qualifier{connect}{Connects the individual parameter values with a line.}
%\qualifier{conf}{Confidence level for the given parameters are plotted, if given.
%                 conf must be a Struct_Type[] with fields: the 'index' of the
%                 parameter and the corresponding confidence (absolut) limits
%                 'conf_min' and 'conf_max'.}
%\qualifier{steppar}{Steppar information (see 'steppar'). Either steppar filename
%                      which is loaded with steppar_load or a List_Type with two
%                      elemtents 1. stepparinformation and 2. the keys.}
%\qualifier{chi2log}{Logarithmic scales for chi2 plots}
%\qualifier{grid}{Enables gridlines on the chi2 plots}
%\qualifier{chi2max}{Custom max value for chi2 range}
%\qualifier{chi2min}{Custom min value for chi2 range}
%\qualifier{cmap}{[='ds9b'] Colormap for the colorcoded chi2 landscapes}
%\qualifier{ticmap}{[='rainbow'] Colormap for the ticlabels/chi2 landscapes}
%}
%\description
%     This function creates and returns a xfig-plot showing a compact overview
%     of the given parameter information.
%     The function takes a arbitrary number of arguments, which however are
%     expected to be all parameter struct arrays (see get_params)! For each
%     argument/parameter-array a individual xfig-plot will be created, which
%     in the end will be combined in an xfig_new_vbox_combound spaced by 'space'.
%
%     The main plot shows the parameter values
%     within its limits (y-axis). Its dimension are determined by the 'size'
%     qualifier. By default the x-axis is either the parameter index or the
%     number/dataset of the parameter, if the names of all given parameter are
%     the same.
%     The values are represented by horizontal lines, where their width
%     can be specified by the 'xwidth' qualifier (measured in x units). If
%     'xwidth' exceeds the minimal distant of adjacent x-values, 'xwidth' is set
%     to that value.
%     The 'x' qualifier allows to use custom x-values, e.g., the time at which
%     the parameters are valid. The corresponding label can be set with 'xlabel'.
%     If the 'x' qualifier is used, a second x-axis will be added to the plot,
%     which will be used to plot the parameter information, i.e., the spacing
%     is given be those 'x' values!
%     It is possible to give additional information about the parameters. Firstly
%     the 'conf' qualifier can be used to give confidence levels of the parameters.
%     'conf' must be a Struct_Type[] with fields 'index' (index of the parameter),
%     'conf_min' and 'conf_max' corresponding to the absolut values of the confidence
%     levels. The 'index' field is needed to find the corresponding parameter! Only
%     matches will be plotted!
%     Further information about the chi2 landscape can be given with the 'steppar'
%     qualifier, which expect filename(s) or filepattern(s) to load steppar
%     information produced with 'steppar'. The files are loaded using the
%     'steppar_load' function. Using 'steppar' qualifier creates a multiplot
%     with two additional plots besides main plot (caution, the return xfig-plot
%     will then be a xfit-multiplot!) showing the individual chi2 landscales
%     of the parameters, if existent and a mean chi2 of those landscapes.
%     With the 'chi2log' qualifier the chi2 landscapes are plotted logarithmic.
%     'grid' will extent the tics to a grid. To set custom limits for the chi2 range
%     use 'chi2min' and 'chi2max'! To easily visualize which chi2 landscape
%     belongs to which parameter, the parameter x-labels are colorized
%     accordingly using the colormap given by 'ticmap'.
%     Furthermore, color-coded chi2 landscapes are added to the main plot using
%     the colormap given with 'cmap'. The colormaps can be given by either a
%     name (String_Type) of an existing colormap (png_get_colormap_names) or
%     a colormap itself.
%\seealso{steppar, steppar_load, png_get_colormap_names}
%
%!%-
{
  %%% CAPTURE ARGUMENTS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  variable args = __pop_list(_NARGS);
  
  %%% PRELOAD QUALIFIERS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  variable tmpdir = qualifier("tmpdir","/tmp/"+getenv("USER")+"/");
  if( tmpdir[-1] != '/' ){
    tmpdir += "/";
  }
  ()=mkdir_rec(tmpdir);
  
  variable size   = qualifier("size",[12,9]);
  variable space  = qualifier("space", 0.05*size[0] );
  variable xwidth = qualifier("xwidth",0.5);
  variable xlabel = qualifier("xlabel","time [MJD]");
  variable connect= qualifier_exists("connect") ? 1:0;
  variable conf   = qualifier("conf",[struct{index}]);

  variable steppar = qualifier("steppar");
  if( qualifier_exists("steppar") ){
    variable sp,sk;
    if( typeof(steppar) == String_Type ){
      % Read steppar information
      (sp,sk) = steppar_load( steppar ;; struct_combine(__qualifiers,struct{keys="pname"}) );
    }
    else if( typeof(steppar) == List_Type and length(steppar)==2 ){
      sp = steppar[0];
      sk = steppar[1];
    }
    else{
      help(_function_name);return;
    }
    variable spname = array_struct_field(sk, "pname");
  }

  variable chi2log= qualifier_exists("chi2log") ? 1:0;
  variable grid   = qualifier_exists("grid") ? 1:0;

  variable cmap = qualifier("cmap","ds9b");
  if( typeof(cmap) == String_Type ){
    cmap = png_get_colormap( cmap );
  }
  cmap = reverse(cmap);

  %%% ARGUMENTS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  variable xplot = {};
  variable params;
  foreach params( args ){
    % check if args
    
    ifnot( _typeof(params) == Struct_Type ){
      help(_function_name);return;
    }

    
    % Extracting parameter information from the parameter array
    variable nparams = length(params);
    variable pname = array_struct_field( params,"name");
    variable pname_lbl = array_map( String_Type, &strreplace, pname, "_", "-" ); 
    variable p     = array_struct_field( params,"value");
    variable pind  = array_struct_field( params,"index");
    variable pmin  = array_struct_field( params,"min");
    variable pmax  = array_struct_field( params,"max");

    % Obtain a overall parameter label 'plbl':
    % If all parameter names are matching plbl <model>*<parametername>
    % is set. Otherwise
    variable pname_tok = array_map( Array_Type, &strtok, pname_lbl, "()." );
    variable pmdl = array_flatten( pname_tok )[[0::length(pname_tok[0])]];
    variable pnum = array_flatten( pname_tok )[[1::length(pname_tok[0])]];
    variable pnick= array_flatten( pname_tok )[[2::length(pname_tok[0])]];
    variable punit = array_struct_field( params,"units");
    punit = array_map( String_Type, &strreplace, punit, "^", "" );     

    % get unique entries (attention unique returns unsorted arrays!)
    variable utmp;
    utmp = unique(pmdl);
    pmdl  = pmdl[ utmp[array_sort(utmp)] ];
    utmp = unique(pnum);
    pnum  = pnum[ utmp[array_sort(utmp)] ];
    utmp = unique(pnick);
    pnick = pnick[ utmp[array_sort(utmp)] ];
    utmp = unique(punit);
    punit = punit[ utmp[array_sort(utmp)] ];
    
    % determin x/y labels:
    % - if for all given parameter the model and the name are the same, than
    %   ylabel = <model>*<name> [unit] and x-axis/label is the parameter number/dataset
    % - otherwise the x-axis is the parameter index
    variable _x, xtics, x1label, plbl;
    if( length(pmdl) and length(pnick) and length(pnum)==nparams ){
      if( length(punit)==1 and punit[0] != "" ){
	punit = " ["+punit+"]";
      }
      
      plbl = [ strjoin( [pmdl,pnick], "*" ) + punit ][0];
      _x = atoi(pnum);
      xtics = "("+pnum+")";
      x1label = "dataset";
    }
    else{
      plbl = "parameter values";
      _x = [1:nparams];
      xtics = array_map( String_Type, &sprintf, "%d", pind );
      x1label = "parameter index";
    }
    _x = [1:nparams];

    %%% QUALIFIERS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    variable x  = qualifier("x",_x);
    variable chi2max= qualifier("chi2max",-DOUBLE_MAX);
    variable chi2min= qualifier("chi2min",DOUBLE_MAX);

    variable ticmap = qualifier("ticmap","rainbow");
    if( typeof(ticmap) == String_Type ){
      ticmap = png_get_colormap( ticmap );
    }
    ticmap = ticmap[nint([0:length(ticmap)-1:#nparams])];

    %%% PLOT PREPARATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % check if xwidth is larger than distant of adjacent x and if need be set
    % xwidth to that minimal distant
    variable _xwidth = xwidth;
    if( nparams > 1 ){
      _xwidth = min( [shift(x,1)-x][[:-2]] );
    }
    if( xwidth > _xwidth ){
      xwidth = _xwidth;
    }

    % Colorize xticlabels, IF chi2 plots are plotted
    if( qualifier_exists("steppar") ){
      xtics = "\\color[rgb]{"+array_map(String_Type, &sprintf, "%0.3f,%03f,%03f", hex2rgb(ticmap;float))+"}{"+xtics+"}";
    }

    % Set x/y plotting ranges
    variable xmin = min(x) - .5*xwidth;
    variable xmax = max(x) + .5*xwidth;
    variable ymin = min(pmin);
    variable ymax = max(pmax);
    variable padx = 0.05;
    variable pady = 0.05;

    variable cind = array_struct_field( conf, "index");
    variable _cind, cmin, cmax;

    % Initialize main plot 'xf':
    variable xf = xfig_plot_new( size[0], size[1] );
    xf.world( xmin, xmax, ymin, ymax ; padx=padx, pady=pady );
    xf.xlabel( x1label );
    xf.ylabel( plbl );

    % Axis management: If 'x' qualifier is not given only x1-axis is on
    xf.x1axis(;
	      major = x,
	      ticlabels = xtics,
	     );
    if( qualifier_exists("x") ){
      xf.x2axis(; on );
      xf.x2label( xlabel );
    }
    else{
      xf.x2axis(; major = x, ticlabels = 0 );
    }

    %%%%%%%%%%%%%%%%%%%
    %%% CREATE MAIN PLOT:
    variable i;
    foreach i ( [0:nparams-1] ){
      % parameter value
      xf.plot( x[i]+[-.5,.5]*xwidth, p[[i,i]] ; depth=10, width=3 );
      % confidence level (if given/existent)
      _cind = where(cind==pind[i]);
      if( length( _cind ) ){
	cmin = conf[_cind[0]].conf_min;
	cmax = conf[_cind[0]].conf_max;
	xf.plot( x[[i,i]], [cmin,cmax] ; depth=10, width=2, color="#000000", line=0 );
	xf.plot( x[i]+[-.25,.25]*xwidth, [cmax,cmax] ; depth=10, width=3, color="#000000" );
	xf.plot( x[i]+[-.25,.25]*xwidth, [cmin,cmin] ; depth=10, width=3, color="#000000" );
      }
      % parameter limits
      xf.plot( x[[i,i]], [pmin[i],pmax[i]] ; depth=40, width=1, color="#999999", line=1 );
      xf.plot( x[i]+[-.5,.5]*xwidth, pmax[[i,i]] ; depth=40,width=3, color="#999999" );
      xf.plot( x[i]+[-.5,.5]*xwidth, pmin[[i,i]] ; depth=40,width=3, color="#999999" );
    }
    if( connect ){
        xf.plot( x, p ; depth=15, width=1, color="#000000", line=0 );
    }

    
    %%%%%%%%%%%%%%%%%%%%%%%%
    %%% Adding Steppar plots
    if( qualifier_exists("steppar") ){

      % qualifiers for the worlds of both chi2 plots. The padding is neccessary to
      % ensure that the yranges are matching
      variable squal = struct{ padx=padx, pady=pady };
      if( chi2log ){
	squal = struct_combine( squal, struct{xlog} );
      }

      % Obtain steppar informations for the given parameters if available
      variable spinfo = Struct_Type[nparams];
      variable spidx  = Integer_Type[0];

      variable nsteps = Integer_Type[0];

      variable j;
      _for i ( 0, nparams-1 ){
	% for each given parameter check if steppar information is available in 'sp'
	j = where( pname[i] == spname );
	if( length(j) ){
	  j = j[0]; % array to single number
	  spidx = [ spidx, i ]; % remember the index of the corresponding parameter

	  % store the necessary information
	  spinfo[i] = struct{

	    pval = get_struct_field( sp[j], escapedParameterName( spname[j] ) ),
	    pval_min,
	    pval_max,

	    chi2 = sp[j].chi2

	  };
	  spinfo[i].pval_min = min(spinfo[i].pval);
	  spinfo[i].pval_max = max(spinfo[i].pval);

	  nsteps = nint(min([ nsteps, length(spinfo[i].pval)]));

	  ifnot( qualifier_exists("chi2min") ){
	    chi2min = min( [ chi2min, sp[j].chi2[where(0<sp[j].chi2<1e100)] ] );
	  }
	  ifnot( qualifier_exists("chi2max") ){
	    chi2max = max( [ chi2max, sp[j].chi2[where(0<sp[j].chi2<1e100)] ] );
	  }
	}
      }
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Overplot the Chi2 as colormap
      variable chi2map,chi2png;
      variable xcm, dx, dy;

      % individual chi2 colormaps
      foreach i (spidx){
	% create the colormap
	chi2map  = tmpdir+"tmp_";
	chi2map += strftime("%j%H%M_");
	chi2map += escapedParameterName(params[i].name);
	chi2map += ".png";
	chi2png = png_gray_to_rgb ( _reshape(spinfo[i].chi2, [length(spinfo[i].chi2),1]),
				    cmap
				    ; gmin=chi2min, gmax=chi2max );
	png_write_flipped ( chi2map, chi2png );

	% create a png of the colormap
	xcm = xfig_new_png( chi2map );
	% resize the png to fit the pmin,pmax range
	( dx, dy ) = xcm.get_pict_bbox();
	xcm.scale_pict( 1./(1+2*padx)*xwidth/(xmax-xmin) * xf.plot_data.plot_width/dx ,
			1./(1+2*pady)*(spinfo[i].pval_max-spinfo[i].pval_min)/(ymax-ymin) * xf.plot_data.plot_height/dy );
	% add the chi2 colormap to the main plot
	xf.add_object( xcm, x[i], spinfo[i].pval_min, 0, -.5 ; depth=30 );
      }

      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Create plot with individual chi2 landscapes
      variable xs = xfig_plot_new( 2./3.*size[0], size[1] );
      xs.x1axis(;grid=grid);
      xs.y1axis(;grid=grid);
      xs.xlabel( "$\chi^2$"R );
      xs.ylabel( plbl );
      xs.world( chi2min, chi2max, ymin, ymax ;; squal );
      foreach i (spidx){
	% the color of the individual chi2 landscapes matches the color of the
	% ticlabels in the main plot
	xs.plot( spinfo[i].chi2, spinfo[i].pval ;
		 color = sprintf("#%06x", ticmap[i]),
		 depth = 25
	       );
      }

      % CREATE inplot scale
      chi2map  = tmpdir+"tmp_";
      chi2map += strftime("%j%H%M_");
      chi2map += string(length(xplot)+1)+"-"+string(_NARGS) + "_scale.png";
      i = [0:1:#length(cmap)];
      if( qualifier_exists("chi2log") ){
	i = 10^i;
      }
      chi2png = png_gray_to_rgb ( _reshape(i,[1,length(cmap)]) , cmap );
      png_write_flipped ( chi2map, chi2png );

      xcm = xfig_plot_new( 1/(1+2*padx) *xs.plot_data.plot_width,
			   0.85*pady*xs.plot_data.plot_height );
      xcm.world ( chi2min, chi2max , 0, 1 ;; reduce_struct(squal,"padx") );
      xcm.xaxis (;off);
      xcm.x2axis(;on,ticlabels=0);
      xcm.yaxis(;off);
      xcm.plot_png ( chi2map );

      xs.add_object( xcm, chi2min, 0, -.5, -.5 ; world10, depth=20 );

      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Create plot of the mean summed chi2 landscape
      variable chi2sum = [0.:0.:#nsteps];
      variable _chi2sum;
      % To sum the individual chi2 landscapes a temporary grid between the minimal
      % and maximal occuring pvalue is defined with #bin='nsteps', which is the minimal
      % occuring amount of pvalues.
      variable p_lo, p_hi;
      (p_lo,p_hi) = linear_grid( ymin,ymax,nsteps);

      variable xm = xfig_plot_new( 1./3.*size[0], size[1] );
      xm.x1axis(;grid=grid);
      xm.y1axis(;grid=grid);
      xm.xlabel( "$\Sigma_n\chi^2/n$"R );
      xm.ylabel( plbl );
      foreach i (spidx){
	% rebin the individual chi2 landscapes to the new grid
	_chi2sum = rebin_mean( p_lo, p_hi,
			       spinfo[i].pval,
			       spinfo[i].pval + (spinfo[i].pval[-1]-spinfo[i].pval[-2]),
			       spinfo[i].chi2
			     );
	_chi2sum[where(_chi2sum==0)] = max(_chi2sum);
	chi2sum += _chi2sum;

      }
      chi2sum /= length(spidx);
      % ger rid of inf entries
      chi2sum[where(isinf(chi2sum))] = max(chi2sum[wherenot(isinf(chi2sum))]);
      
      xm.world( qualifier_exists("chi2min") ? chi2min : min(chi2sum),
		qualifier_exists("chi2max") ? chi2max : max(chi2sum),
		ymin, ymax ;; squal
	      );
      xm.plot( chi2sum, p_lo ; width = 3 );

      xf = xfig_multiplot( xf, xs, xm ; cols=3 );
    }
    list_append( xplot, xf );
  }

  variable xcomp = xfig_new_vbox_compound( __push_list(xplot), space ;; __qualifiers );
  return xcomp;
}

#endif

#endif

%
% Code from src/sixte/xfig_plot_sky_img.sl
%

% only activate if module wcsfuns could be loaded
#ifeval __get_reference("wcsfuns_project")!=NULL

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

private define deg2dms_rounded(deg){
   variable d,m,s;
   (d,m,s) =  deg2dms(deg;;__qualifiers);
   if (s > 59.99){	 
      m++;
      s = 0.0;
      if (m == 60) {
	 m=0;
	 d++;
      }
   }
   return d,m,s;
}

private define slog(a){
   return log(a+1);   
}

private define _xfig_plot_sky_img(img0){ %{{{
   variable img = img0;
   if (typeof(img0) == String_Type)
     img = fits_read_img(img0);
   
   variable dims;
   (dims, ,) = array_info(img);
   
   variable size = qualifier("size",[10,10./dims[1]*dims[0] ]);
   variable pl = xfig_plot_new(size[0],size[1]);
   variable cmap = qualifier("cmap","hot");

   
   variable func;
   if (qualifier_exists("lin")) {
      pl.plot_png (img; cmap=cmap);
   }
   else if (qualifier_exists("func")){
      func = qualifier("func",&log);
      pl.plot_png (@func(img); cmap=cmap,
		   gmin=@func(qualifier("gmin",min(img);))
		  );
   }   
   else {
      pl.plot_png (slog(img); cmap=cmap,
		   gmin=log(qualifier("gmin",log(2)))
		  );
   }

   variable fimg,wcs,lab,wi,arcmin,xar,yar,img_size;
   variable del=0.04;
   variable col_scale = qualifier("scale_color","white");
   if (qualifier_exists("scale")){
      fimg = qualifier("ref_img",img0);
      wcs = fitswcs_get_img_wcs(fimg);
      arcmin = qualifier("arcmin",5);

      img_size = array_shape(img);
      
      wi = (arcmin/60.) / (abs(wcs.cdelt[1])*img_size[1]);

      xar = [1-del-wi,1-del];
      yar = [0.05,0.05];
      pl.plot(xar,yar;color=col_scale,width=3,world0);
      
      lab = sprintf(`$%.0f'$`,arcmin);
      pl.add_object(xfig_new_text(lab;color=col_scale,size=`\footnotesize`),
		    0.5*(xar[0]+xar[1]),yar[0],0,-1.1;world0);
		    
   }
   
   pl.yaxis(; off);
   pl.xaxis(; off);
   return pl;
}
%}}}


%%%%%%%%%%%%%%%%%%%%%%%
define xfig_plot_sky_img(fimg){ %{{{
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{xfig_plot_sky_img}
%\synopsis{plots a sky image, including an optional scale and/or grid}
%\usage{Xfig_Plot_object pl = xfig_plot_sky_img(String_type fits_img);}
%\description
%    This function uses xfig to plot a sky image, with a given color
%    scale ("hot" by default). With a separate qualifier, a scale can
%    be switched on. The 
%\qualifiers{
%\qualifier{cmap}{["hot"] color map}
%\qualifier{scale}{switch on the scale}
%\qualifier{arcmin}{[=5] length of the scale in arcmin}
%\qualifier{lin}{linear image scale instead of logarithmic}
%\qualifier{func}{giva a reference to an arbitrary function to be applied to the image}
%\qualifier{size}{[10,10] size of the image in cm}
%\qualifier{ref_img}{[fits_img] if not given as direct argument, a reference FITS image has to be provided for the scale (filename!)}
%\qualifier{grid}{draw a grid and corresponding labels}
%\qualifier{step_ra}{[=1] step size of the RA-grid (given in min)}
%\qualifier{step_ra}{[=5] step size of the Dec-grid (given in arcmin)}
%\qualifier{grid_color}{[#BBBBBB] color of the grid}
%\qualifier{scale_color}{[white] color of the scale}
%}
%\seealso{xfig_plot_new,fitswcs_get_img_wcs}
%!%-

   
   variable pl = _xfig_plot_sky_img(fimg;;__qualifiers());

   variable size = array_shape(fits_read_img(fimg));
   variable p_xmin = 0.5;
   variable p_xmax = size[1]+0.5;
   variable p_ymin = 0.5;
   variable p_ymax = size[0]+0.5;

   if (not (qualifier_exists("grid"))) {
      return pl;
   }
   

   variable wcs = fitswcs_get_img_wcs (fimg);
   variable gridclr = qualifier("grid_color","#BBBBBB");

   variable arcmin_ra = qualifier("step_ra",1.);
   variable arcmin_dec = qualifier("step_dec",5.);
   

   variable ramin,ramax,dmin,dmax;
   (ramin,dmin) = wcsfuns_deproject(wcs,p_xmin,p_ymin);
   (ramax,dmax) = wcsfuns_deproject(wcs,p_xmax,p_ymax);
   if (ramin < 0){
      ramin +=360;
   }
   if (ramax < 0){
      ramax +=360;      
   }
    
   pl.world(0,size[1],0,size[0]);
   
   variable xmaj_hms = [[int(ramin*24./360):
			 int(ramax*24./360)+1:
			 arcmin_ra/60.]];
   
   variable xmaj = xmaj_hms*360./24.;
   
   variable ymaj = [[int(dmin):int(dmax)+1:arcmin_dec/60.]];

   variable ii,d,m,s;
   variable n = length(xmaj);
   variable xlab = String_Type[n];
   _for ii(0,n-1,1){
      (d,m,s) =  deg2dms_rounded(xmaj[ii];hours);
      xlab[ii] = sprintf("$%i\mathrm{h}%02i\mathrm{m}%02i\mathrm{s}$"R,
			 d,m,nint(s));
   }
   
   variable ny = length(ymaj);
   variable ylab = String_Type[ny];
   _for ii(0,ny-1,1){
      (d,m,s) =  deg2dms_rounded(ymaj[ii]);
      ylab[ii] = sprintf("$%i^\circ%02i'%02i''$"R,d,m,nint(s));
   }

   variable xmaj1_wcs,xmaj2_wcs,ymaj1_wcs,ymaj2_wcs;
   ( ,xmaj1_wcs)= wcsfuns_project(wcs,xmaj,xmaj*0+dmin);
   ( ,xmaj2_wcs)= wcsfuns_project(wcs,xmaj,xmaj*0+dmax);
   (ymaj1_wcs, )= wcsfuns_project(wcs,ymaj*0+ramin,ymaj);
   (ymaj2_wcs, )= wcsfuns_project(wcs,ymaj*0+ramax,ymaj);
   
   pl.axis(;on,color=gridclr);
 
   pl.x1axis(;major=xmaj1_wcs,color=gridclr,ticlabels=xlab);
   pl.y1axis(;major=ymaj1_wcs,color=gridclr,ticlabels=ylab);
   pl.x2axis(;major=xmaj2_wcs,color=gridclr,ticlabels=0);
   pl.y2axis(;major=ymaj2_wcs,color=gridclr,ticlabels=0);

   variable ra,dec,fac = 0.5;
     ra = [ramin-(ramax-ramin)*fac:
	   ramax+(ramax-ramin)*fac:#200];
   dec = [dmin-(dmax-dmin)*fac:
	  dmax+(dmax-dmin)*fac:#200];   

   variable pos_y,pos_x;
   variable src_pos = qualifier("src",NULL);
   if (src_pos != NULL){
      (pos_y,pos_x) = wcsfuns_project (wcs, src_pos[0],src_pos[1]);
      pl.plot(pos_x,pos_y;width=1,line=1,color=qualifier("src_color","white"),
	      depth=10,sym="+");      
   }
      
   variable v;
   foreach v (ymaj)   {      
      (pos_y,pos_x) = wcsfuns_project (wcs, ra,dec*0 + v);
      pl.plot(pos_x,pos_y;  width=1,line=1,color=gridclr,depth=10);
   }
   foreach v (xmaj)  {
      (pos_y,pos_x) = wcsfuns_project (wcs, ra*0+v,dec);
      pl.plot(pos_x,pos_y;  width=1,line=1,color=gridclr,depth=10);
   }
   
   return pl;
}

%}}}


#endif

#endif

%
% Code from src/plot/xfig_plot_spectrum.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL
try { require("xfig"); }
catch AnyError: {}

#ifexists xfig_plot_new

% prototype some subroutines
private define xfps_read_args();
private define xfps_setup_gaps();
private define xfps_create_xfig_object();
private define xfps_plot_data();
private define xfps_plot_model();
private define xfps_plot_res();
private define xfps_create_string_array();
private define xfps_labels();

private define __print_xfig_plot_spectrum_version() {
  variable __xfig_plot_spectrum_version = [0,9,0];
  vmessage("xfig_plot_spectrum version %s",
    strjoin(array_map(String_Type,&string,__xfig_plot_spectrum_version),"."));
}

% Minimum for (pathological) spectra with no points above zero and logarithmic
% y-axis scale
private variable XFPS_Y_DEFAULT_MINIMUM = 1e-6;
% Easier labels for axes on spectral plots
private variable XFPS_LABEL_X_KEV = "Energy (keV)";
private variable XFPS_LABEL_Y_KEV = "Counts s$^{-1}$ keV$^{-1}$";
private variable XFPS_LABEL_X_A = "Wavelength (\\AA)";
private variable XFPS_LABEL_Y_A = "Counts s$^{-1}$ \\AA$^{-1}$";

define xfig_plot_spectrum()
%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{xfig_plot_spectrum}
%\synopsis{Makes an array of slxfig objects from a set of plots saved with write_plot}
%\usage{Struct_Type[] pl = xfig_plot_spectrum(Struct_Type plt[],[...])}
%\description
%   Generates an array of SLxfig plot objects from a given set of plot data
%   structs (as generated by write_plot).
%
%   Arguments should be structs as returned by write_plot(;data). The first
%   argument should be the "best-fit" model - i.e., the spectrum and model you
%   want to plot in the top pane. Subsequent arguments should be structs
%   returned by write_plot(;data) for "intermediate" residuals (e.g. with other
%   spectral features turned on or off).
%
%   The function returns an array of SLxfig plot objects, which can then be
%   rendered using, e.g., xfig_multiplot().render. The colors, symbols,
%   linestyles, and other parameters can be adjusted using qualifiers (see
%   below). Since the xfig objects themselves are returned, you can also modify
%   them yourself after the fact (in order to, e.g., add labels or do more
%   plotting).
%
%   EXAMPLE: consider a scenario where we fit some model containing a Gaussian
%   emission line (using an "egauss" component) to a dataset of three spectra.
%   We want to create a plot that has a top pane (occupying 50% of the total
%   plot area) showing the spectrum and best-fit model, a middle pane showing
%   the residuals without the emission line, and a bottom panel showing the
%   best-fit residuals. The three spectra will be plotted in black, red, and
%   blue, with the model overplotted in green, purple, and orange, and the
%   output will be a PDF called "spectrum.pdf".
%
%   \code{
%   plot_data({1,2,3};res=1);                 % plot data and best-fit model with residuals
%   variable best = write_plot(;data);        % write plot to a structure
%   set_par("egauss(1).area",0.0,1);          % Turn off the emission line
%   () = fit_counts;                          % re-fit without the line
%   plot_data({1,2,3};res=1);                 % plot data without the line
%   variable noline = write_plot(;data);      % write the last plot to another structure
%   pl = xfig_plot_spectrum(best,noline;      % call xfig_plot_spectrum to create the xfig objects
%     colors=["black","red","blue"],          % data colors
%     mcolors=["green","purple","orange"],    % model colors
%     topPanelFrac=0.5);                      % top panel (data+model) occupies 50% of total plot area
%   xfig_multiplot(pl).render("spectrum.pdf");% render the final plot as a PDF
%   }
%
%   More qualifiers (see below) are available to fine-tune how the plot
%   appears. The colors, style, and size modifiers (dsym, symsize, mstyle,
%   mscale, dataDepth, modelDepth) can be given as arrays, with one element for
%   each spectrum plotted, or as single values, in which case each spectrum is
%   plotted with the same symbol/style/color/size.
%
%   By default, the x- and y-ranges for the resulting plot will depend on the
%   energy (or wavelength) ranges which were _noticed_ when you called
%   write_plot(). The simplest way to change the range of the resulting plot is
%   to only notice the wavelength/energy range you want to plot with xnotice()
%   or xnotice_en() - this will ensure that the y-axis and residuals scale
%   properly with what is plotted. There are also "xrange" and "yrange"
%   qualifiers to set the range of the top panel; however, the residual ranges
%   are still determined based on the full range of data.
%
%   Labels for the individual panels can be enabled via the "labels" qualifier.
%   If labels is present, but not set (or if its first element is NULL), then
%   the panels will be automatically labeled with letters starting from 'a'.
%   The "firstPanelLabel" qualifier can be used to change the starting label;
%   it should be a single letter. If the "labels" qualifier is not present at
%   all, the panels will be unlabeled. The x/y positions of the labels can be
%   set via the label_x and label_y qualifiers, which should be arrays of
%   Double_Type values indicating the positions of the lower-right corners of
%   the labels for each panel in 'world0' coordinates for that panel (see
%   'xfig_plot--wcs'). By default these are [0.94,0.90] for the data+model
%   panel and [0.94,0.78] for the residuals.
%
%   The "gaps" qualifier lets you specify where the model should not be plotted
%   (e.g., between 1.7 and 2.4 keV in the Suzaku-XIS). It should be given as a
%   List_Type list of Array_Type arrays, one array for each spectrum plotted.
%   The arrays should be a list of start and stop energies for the gaps (this
%   means they need to have even length); for instruments with no gaps, pass
%   \code{Double_Type[0]}.  E.g., for a Suzaku observation with XIS0, XIS1,
%   XIS3, and the HXD/PIN where you don't want to plot the model between 1.6
%   and 2.3 keV in the XIS spectra, you would pass
%   \code{gaps={[1.6,2.3],[1.6,2.3],[1.6,2.3],Double_Type[0]}}.
%
%   This script is mostly feature-complete, but please contact Paul Hemphill
%   (pbh@space.mit.edu) regarding bugs or missing features (please do not
%   report missing bugs; any bugs that are not present are missing
%   intentionally).
%
%\qualifiers{
%\qualifier{topPanelFrac}{Double_Type, fraction of plot taken up by top panel, default 0.5}
%\qualifier{height}{Double_Type, height of plot in cm, default 14.2}
%\qualifier{width}{Double_Type, width of plot in cm, default 12.6}
%\qualifier{colors}{String_Type[] or Integer_Type[], xfig colors for each spectrum plotted, by default uses get_sron_colors() to get a set of colors}
%\qualifier{mcolors}{String_Type[] or Integer_Type[], xfig colors for each model plotted. By default each model is plotted with the same color as its associated data.}
%\qualifier{dsym}{String_Type[] or Integer_Type[], xfig symbols for data for each spectrum plotted, defaults to "point"}
%\qualifier{symsize}{Integer_Type[], size of symbols for each spectrum plotted. Can be a single value or an array (one element for each spectrum).}
%\qualifier{mstyle}{Integer_Type[], xfig linestyle for overplotted best-fit model for each spectrum. mstyle = -1 plots the model as a histogram. As with symsize, can be single value or array. Default -1}
%\qualifier{mwidth}{Integer_Type[], xfig line width for overplotted best-fit model. As with symsize, can be single value or array. Default 1.}
%\qualifier{mscale}{Double_Type[], multiplicative scale for model (e.g., if you want to plot your model above your data points). Single value or array.}
%\qualifier{dataDepth}{Integer_Type[], xfig depth for data points, either a single value or an array specifying the depth for each spectrum. Default 1}
%\qualifier{modelDepth}{Integer_Type[], xfig depth for model lines. See dataDepth. Default 0 (this means models are plotted over data by default)}
%\qualifier{ratio}{If present, residuals are ratio residuals, otherwise residuals are assumed to be some variant of data-model.}
%\qualifier{no_residuals}{If set, no residuals are plotted.}
%\qualifier{gaps}{List_Type, indicates where gaps in an instrument's spectrum mean the model should not be plotted.}
%\qualifier{xrange}{Double_Type[2], range for x-axis (default autoscaled)}
%\qualifier{yrange}{Double_Type[2], range for y-axis of top panel (default autoscaled)}
%\qualifier{rrange}{List_Type[]. Should be a list of two-element arrays containing the minimum and maximum for the y-axis of each residual plot, in the order that they appear on the plot.}
%\qualifier{xlog}{Axes are linear by default. If "xlog" is set, or if it is set to a nonzero number, the x-axis will be logarithmic. If "xlog" is not set, or is set to zero, the x-axis will be linear.}
%\qualifier{ylog}{See xlog; sets log/linear scale for y-axis on top panel. Residual plots always have linear y-axes.}
%\qualifier{unit}{String_Type, either "keV" (energy, default) or "A" (wavelength).}
%\qualifier{xlabel}{String_Type, label for x-axis, default depends on units. If unit="keV", "Energy (keV)"; if unit="A", "Wavelength (\\AA)". Set to NULL for no label.}
%\qualifier{ylabel}{String_Type, label for y-axis of spectrum, default depends on units. If unit="keV", "Counts s$^-1$ keV$^-1$"; if unit="A", "Counts s$^{-1}$ \\AA$^{-1}$".}
%\qualifier{rlabel}{String_Type, label for y-axis of residuals. If the "ratio" qualifier is set, default is "Ratio", otherwise default is "$\\chi$".}
%\qualifier{labels}{String_Type[], labels for individual panels.}
%\qualifier{firstPanelLabel}{String_Type or Integer_Type, label for first panel (further panels will increment this value, so if you give 'b' the next panels will be 'c', 'd', etc).}
%\qualifier{label_x}{Double_Type[], x-positions (in world0 coordinates) of panel labels.}
%\qualifier{label_y}{Double_Type[], y-positions (in world0 coordinates) of panel labels.}
%\qualifier{arrowY}{Double_Type[2], lower and upper device coordinates for arrows for bins extending off the screen, default [0.01,0.1].}
%\qualifier{version}{Display version information and exit.}
%}
%\seealso{write_plot, xfig_multiplot, xfig_plot_unfold, xfig_plot--wcs}
%!%-
{
  if(qualifier_exists("version") || _NARGS == 0) {
    if(qualifier_exists("version")) {
      __print_xfig_plot_spectrum_version();
      return;
    } else {
      usage("Struct_Type[] = plot_spectrum(String_Type files[],[...])");
    }
  }

  % counter variables (they're gonna be used sooner or later)
  variable i=0,j=0;

  variable spectra;
  spectra = xfps_read_args(__pop_args(_NARGS));

  variable nores = qualifier_exists("no_residuals");
  variable ratio = qualifier_exists("ratio");
  % If no_residuals is set, then we only have one panel. Otherwise, it's
  % length(spectra)+1 (i.e., the number of arguments + 1), so that we have
  % one panel for the data+model, one for each "intermediate" set of residuals,
  % and one for the best-fit residuals.
  variable nPanels = length(spectra) + 1;
  if(length(spectra[0].res[0]) == 0) nores = 1;
  if(nores) nPanels = 1;
  % spectra[0].index is an array containing the dataset indices of the
  % plotted datasets, so length(spectra[0].index) gives us the number of
  % spectra plotted.
  variable numInst = length(spectra[0].index);

  % Plot dimensions
  variable totWidth = qualifier("width",12.6);
  variable totHeight = qualifier("height",14.2);
  variable topPanelFrac = qualifier("topPanelFrac",0.5);
  % Log plots?
  variable xLog = qualifier("xlog",0);
  if(_isnull(xLog) || xLog) xLog = 1;
  variable yLog = qualifier("ylog",0);
  if(_isnull(yLog) || yLog) yLog = 1;
  % Custom ranges?
  variable xRange = qualifier("xrange",NULL);
  variable yRange = qualifier("yrange",NULL);
  variable rRange;
  if(qualifier_exists("rrange")) {
    rRange = qualifier("rrange");
  } else {
    rRange = {};
    _for i (0,nPanels-1,1)
      list_append(rRange,NULL);
  }
  % Color/style options
  % These are set up so that if the user provides a single string (rather than
  % an array) the same color/style is used for all spectra
  variable colors = qualifier("colors", get_color_palette("sron-rainbow", numInst; repeat));
  if(typeof(colors) == String_Type or typeof(colors) == Integer_Type)
    colors = xfps_create_string_array(numInst,colors);
  variable mcolors = qualifier("mcolors",colors);
  if(typeof(mcolors) == String_Type or typeof(mcolors) == Integer_Type)
    mcolors = xfps_create_string_array(numInst,mcolors);
  variable dsym = qualifier("dsym","point");
  if(typeof(dsym) == String_Type)
    dsym = xfps_create_string_array(numInst,"point");
  variable symsize = qualifier("symsize",Double_Type[numInst]+1.0);
  if(length(symsize) == 1) symsize = Double_Type[numInst] + symsize[0];
  variable mstyle = qualifier("mstyle",Integer_Type[numInst]-1);
  if(length(mstyle) == 1) mstyle = Integer_Type[numInst] + mstyle[0];
  variable mwidth = qualifier("mwidth",Integer_Type[numInst]+1);
  if(length(mwidth) == 1) mwidth = Integer_Type[numInst] + mwidth[0];
  % Scaling for model
  variable mscale = qualifier("mscale",Double_Type[numInst]+1.0);
  if(length(mscale) == 1) mscale = Double_Type[numInst] + mscale[0];
  % Depth of model/data
  variable dataDepth = qualifier("dataDepth",Integer_Type[numInst]+1);
  if(length(dataDepth) == 1) dataDepth = Integer_Type[numInst] + dataDepth[0];
  variable modelDepth = qualifier("modelDepth",Integer_Type[numInst]);
  if(length(modelDepth) == 1) modelDepth = Integer_Type[numInst] + modelDepth[0];
  % Labels and units
  variable units = strlow(qualifier("unit","kev"));
  variable xLab = qualifier("xlabel",xfps_labels(units)[0]);
  variable yLabData = qualifier("ylabel",xfps_labels(units)[1]);
  variable yLabRes = qualifier("rlabel",(ratio ? "Ratio" : "$\\chi$"));
  variable firstPanelLabel = qualifier("firstPanelLabel",'a'); % using a char so I can increment it later
  if(typeof(firstPanelLabel) == String_Type) firstPanelLabel = firstPanelLabel[0];
  variable labelX = qualifier("label_x",Double_Type[nPanels]+0.94);
  variable labelY = qualifier("label_y",[0.90,Double_Type[nPanels-1]+0.78]);
  variable labels = String_Type[nPanels];
  if(qualifier_exists("labels")) {
    labels = qualifier("labels",NULL);
    if(_isnull(labels[0])) {
      % need to convert to strings here - not sure if slxfig will print a
      % Char_Type as a string or an integer.
      labels = array_map(String_Type,&char,[firstPanelLabel:firstPanelLabel+nPanels-1]);
    }
  }
  variable plotTitle = qualifier("title",NULL);
  % Arrows for negative lower limits on data points
  variable arrowY = qualifier("arrowY",[0.01,0.10]);
  % setting up some variables for xfig object setup later
  variable panelHeight;
  % Gaps in data?
  variable gaps = qualifier("gaps",NULL);
  gaps = xfps_setup_gaps(gaps,numInst);

  % Stick all the plot options into a struct to make subroutines easier to write
  variable plotOptions = struct{
    colors=colors,mcolors=mcolors,dsym=dsym,symsize=symsize,mstyle=mstyle,
    mscale=mscale,mwidth=mwidth,dataDepth=dataDepth,modelDepth=modelDepth,
    arrowY=arrowY,xLog=xLog,yLog=yLog,gaps=gaps,ratio=ratio,totWidth=totWidth,
    totHeight=totHeight,topPanelFrac=topPanelFrac,nores=nores,nPanels=nPanels,
    yrange=yRange,rrange=rRange,xrange=xRange
  };

  variable pl = Struct_Type[nPanels];
  variable spec;
  _for i (0,nPanels-1,1) {
    spec = struct_copy(spectra[i<nPanels-1?i:0]);
    % Set up plot object and axes' ranges
    pl[i] = xfps_create_xfig_object(spec,i,plotOptions);
    % Axes labels
    if(i==nPanels-1) pl[i].xlabel(xLab);
    pl[i].ylabel( (i==0 ? yLabData : yLabRes) );
    % Loop through spectra (there are length(spec.dval) of them) and plot data/model or residuals
    if(i==0) {
      pl[i] = xfps_plot_data(pl[i],spec,plotOptions);
      pl[i] = xfps_plot_model(pl[i],spec,plotOptions);
    } else {
      pl[i] = xfps_plot_res(pl[i],spec,plotOptions);
    }
    if(not(_isnull(labels[i])) && not(string_match(labels[i],`^[ \t\n]*$`)))
      pl[i].xylabel(labelX[i],labelY[i],labels[i],0.5,-0.5;world0);
  }
  return pl;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% xfps_read_args()
% Read in arguments to xfig_plot_spectrum
% In the future this should be able to read both the text files created by
% write_plot() and the structs created by write_plot(;data). Right now it only
% does the structs.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define xfps_read_args(args) {
  variable arg;
  variable p = Struct_Type[0];  % will store the structs containing data/model/residuals
  foreach arg (args) {
    if(typeof(arg.value) == String_Type or typeof(arg.value) == Array_Type) {
      % argument is a string or array -> read files made by write_plot()
      % for now we'll just fail because I don't want to write this yet
      % (it'll entail converting a read_data_from_write_plot() struct into the
      % struct returned by write_plot(;data) and that's just gonna be a pain)
      throw NotImplementedError,
        "*** Error: Reading files not implemented yet.",arg.value;
    } else if(typeof(arg.value) == Struct_Type) {
      % argument is a struct -> this is a struct from write_plot()
      p = [p,struct_copy(arg.value)];
    }
  }
  return p;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% xfps_setup_gaps()
% Check and set up the "gaps" variable, which defines where gaps exist in the
% spectrum. Should result in gaps being a list of numInst zero-length arrays if
% it wasn't given.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define xfps_setup_gaps(gaps,numInst) {
  variable i;
  if(typeof(gaps) == List_Type and length(gaps) != numInst) {
    vmessage("Error: %i sets of gaps specified, but you have %i instruments to plot.\n",length(gaps),numInst);
    throw InvalidParmError;
  } else if (typeof(gaps) == List_Type) {
    for(i=0;i<length(gaps);i++) {
      if(length(gaps[i]) mod 2) {
        vmessage("Error: gaps for each instrument should be given as an even-length array of start and stop energies.\n");
        throw InvalidParmError;
      }
    }
  } else {
    % If gaps are not given, set to a list of zero-length arrays
    gaps = {};
    for(i=0;i<numInst;i++) list_append(gaps,Double_Type[0]);
  }
  return gaps;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% xfps_create_xfig_object()
% Calls xfig_plot_new and sets the height and width properly.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define xfps_create_xfig_object(theSpectrum,i,popt) {
  variable panelWidth,panelHeight;
  variable xlo,xhi,y,dy,xmin,xmax,ymin,ymax;
  panelWidth = popt.totWidth;
  if(popt.nores) {
    panelHeight = popt.totHeight;
  } else if(i == 0) {
    panelHeight = popt.totHeight * popt.topPanelFrac;
  } else {
    panelHeight = popt.totHeight * (1.0-popt.topPanelFrac)/(double(popt.nPanels-1));
  }
  if(_isnull(popt.xrange)[0]) {
    xmin = min(array_flatten(theSpectrum.dlo));
    xmax = max(array_flatten(theSpectrum.dhi));
  } else {
    xmin = popt.xrange[0];
    xmax = popt.xrange[1];
  }
  if(i != 0) { % working on residuals
    variable rrange;
    if(_isnull(popt.rrange[i-1])) {
      variable j;
      variable rmin = Double_Type[0];
      variable rmax = Double_Type[0];
      _for j (0,length(theSpectrum.res)-1,1) {
        rmin = [rmin,theSpectrum.res_m[j][0]];
        rmax = [rmax,theSpectrum.res_p[j][0]];
      }
      ymin = min(rmin);
      ymax = max(rmax);
    } else {
      ymin = popt.rrange[i-1][0];
      ymax = popt.rrange[i-1][1];
    }
  } else { % working on data
    if(_isnull(popt.yrange)[0]) {
      % We want to scale to the noticed x-range, not the full range.
      % First, we grab the x-axis values and compare them to the x-range, which was set above.
      xlo = array_flatten(theSpectrum.dlo);
      xhi = array_flatten(theSpectrum.dhi);
      variable ndx = where(xlo >= xmin and xhi <= xmax); 
      % Then, we grab the y-values and their errors and select only the ones that are within [xmin,xmax]
      dy = array_flatten(theSpectrum.derr)[ndx];
      y = array_flatten(theSpectrum.dval)[ndx];
      % The maximum is easy in any scenario - it's just the largest value.
      % (unless you gave me a spectrum with all negative numbers, I'm now
      % realizing. Please don't do that?)
      ymax = max(y+dy);
      % For the minimum, if the y-axis isn't logarithmic, we're fine - we just
      % take the smallest value and call it a day.
      if(not(popt.yLog)) {
        ymin = min(y-dy);
      } else {
        % But if the y-axis is log-scaled, we need to get rid of anything <= 0.
        ymin = y-dy;
        % First try to find places where y-dy is above 0 and use the smallest one of those
        if(length(where(ymin >= 0.))) {
          ymin = min(ymin[where(ymin >= 0.)]);
        } else {
          % If that fails, find places where y is greater than zero
          ymin = y;
          if(length(where(ymin >= 0.))) {
            ymin = min(ymin[where(ymin >= 0.)]);
          } else {
            % Otherwise, use some sort of hard minimum.
            % Note that if we get to this point, we're just going to be
            % plotting a bunch of downward arrows, because everything will be
            % off the scale.
            ymin = XFPS_Y_DEFAULT_MINIMUM;
            message("*** Warning: No data points greater than zero and logarithmic y-axis requested.");
          }
        }
      }
    } else {
      ymin = popt.yrange[0];
      ymax = popt.yrange[1];
    }
  }
  % print(sprintf("xmin = %f, xmax = %f, ymin = %f, ymax = %f",xmin,xmax,ymin,ymax));
  variable pltemp = xfig_plot_new(panelWidth,panelHeight);
  pltemp.world(xmin,xmax,ymin,ymax;xlog=popt.xLog,ylog=(i==0?popt.yLog:0),pady=0.1);
  return pltemp;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% xfps_plot_data()
% Plots data on provided xfig plot object
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define xfps_plot_data(pl,spec,popt) {
  variable j;
  _for j (0,length(spec.dval)-1,1) {
    variable dlo,dhi,dval,derr;
    dlo = spec.dlo[j];
    dhi = spec.dhi[j];
    dval = spec.dval[j];
    derr = spec.derr[j];
    % Also get the mins/maxes from the xfig object:
    variable w = pl.get_world(); % w is [xmin,xmax,ymin,ymax]
    % First, only plot points which are above ymin and below ymax
    variable ndx = where(dval-derr >= w[-2] and dval+derr <= w[-1]);
    if(length(ndx))
      pl.plot((dlo[ndx]+dhi[ndx])/2.0,dval[ndx],{dlo[ndx],dhi[ndx]},derr[ndx];
        xminmax,sym=popt.dsym[j],size=popt.symsize[j],color=popt.colors[j],
        depth=popt.dataDepth[j]);

    variable k; % counter
    variable x,y,ymin,ymax; % for arrow coords
    % Now plot arrows for anything that extends below ymin:
    ndx = where(dval-derr <= w[-2]);
    if(length(ndx)) {
      % plot the points _without error bars_ first
      pl.plot((dlo[ndx]+dhi[ndx])/2.0,dval[ndx];sym=popt.dsym[j],size=popt.symsize[j],color=popt.colors[j],depth=popt.dataDepth[j]);
      % then draw a downward arrow - down to arrowY[0] of the way from the
      % bottom axis, starting either at the top of the error bar or arrowY[1]
      % of the way from the bottom axis, whichever is larger
      variable yFrac;
      _for k (0,length(ndx)-1,1) {
        x = (dlo[ndx[k]]+dhi[ndx[k]])/2.0;
        y = dval[ndx[k]] + derr[ndx[k]];
        if(popt.yLog) {
          yFrac = (log10(y)-log10(w[-2]))/(log10(w[-1])-log10(w[-2]));
        } else {
          yFrac = (y-w[-2])/(w[-1]-w[-2]);
        }
        ymin = popt.arrowY[0];
        ymax = max([popt.arrowY[1],yFrac]);
        pl.plot([x,x],[ymax,ymin];world10,size=popt.symsize[j],color=popt.colors[j],depth=popt.dataDepth[j],forward_arrow);
      }
    }
    % And same for anything that extends above ymax:
    ndx = where(dval+derr >= w[-1]);
    if(length(ndx)) {
      % plot the points _without error bars_ first
      pl.plot((dlo[ndx]+dhi[ndx])/2.0,dval[ndx];sym=popt.dsym[j],size=popt.symsize[j],color=popt.colors[j],depth=popt.dataDepth[j]);
      % then draw an upward arrow - up to arrowY[0] of the way from the
      % top axis, starting either at the bottom of the error bar or arrowY[1]
      % of the way from the top axis, whichever is smaller
      _for k (0,length(ndx)-1,1) {
        x = (dlo[ndx[k]]+dhi[ndx[k]])/2.0;
        y = dval[ndx[k]] - derr[ndx[k]];
        if(popt.yLog) {
          yFrac = (log10(y)-log10(w[-2]))/(log10(w[-1])-log10(w[-2]));
        } else {
          yFrac = (y-w[-2])/(w[-1]-w[-2]);
        }
        ymin = min([1.0-popt.arrowY[1],yFrac]);
        ymax = 1.0-popt.arrowY[0];
        pl.plot([x,x],[ymin,ymax];world10,size=popt.symsize[j],color=popt.colors[j],depth=popt.dataDepth[j],forward_arrow);
      }
    }
  }
  return pl;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% xfps_plot_model()
% Plots model on provided xfig plot object
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define xfps_plot_model(pl,spec,popt) {
  variable j;
  _for j (0,length(spec.dval)-1,1) {
    % If spec.mval[j] is NULL, then there's no model! Only keep going if that's not the case!
    if(not(_isnull(spec.mval[j])[0])) {
      variable dlo,dhi,mval;
      dlo = spec.dlo[j];
      dhi = spec.dhi[j];
      mval = spec.mval[j] * popt.mscale[j];
      % Loop through gaps[j] two elements at a time, getting the starting and
      % ending points where we should plot the model.
      % I.e., if gaps[j] is [1.6,2.3], we want to plot the model in the ranges
      % [emin,1.6) and (2.3,emax].
      % If length(gaps[j]) is zero, we still run through this loop once (since
      % we start with k = -1).
      variable k,mstart,mend,ndx;
      _for k (-1,length(popt.gaps[j])-1,2) {
        if(length(popt.gaps[j]) == 0) {
          mstart = min(dlo)-1.0;% subtract 1 just so we know we get all the points
          mend = max(dhi)+1.0;  % add 1 just so we know we get all the points
        } else if(k == -1) {
          mstart = min(dlo)-1.0;
          mend = popt.gaps[j][k+1];
        } else if (k == length(popt.gaps[j])-1) {
          mstart = popt.gaps[j][k];
          mend = max(dhi)+1.0;
        } else {
          mstart = popt.gaps[j][k];
          mend = popt.gaps[j][k+1];
        }
        ndx = where(dlo >= mstart and dhi <= mend);
        if(popt.mstyle[j] == -1) {
          pl.hplot(dlo[ndx],mval[ndx];
            color=popt.mcolors[j],depth=popt.modelDepth[j],
            width=popt.mwidth[j],y_first=mval[ndx[0]],y_last=mval[ndx[-1]]);
        } else {
          pl.plot((dlo[ndx]+dhi[ndx])/2.0,mval[ndx];
            line=popt.mstyle[j],width=popt.mwidth[j],
            color=popt.mcolors[j],depth=popt.modelDepth[j]);
        }
      }
    }
  }
  return pl;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% xfps_plot_res()
% Plots residuals on provided plot object
% TODO: clean this up and make it resemble xfps_plot_data() more.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define xfps_plot_res(pl,spec,popt) {
  variable j;
  _for j (0,length(spec.res)-1,1) {
    variable dlo,dhi,res,res_m,res_p;
    dlo = spec.dlo[j];
    dhi = spec.dhi[j];
    res = spec.res[j][0];
    res_m = spec.res_m[j][0];
    res_p = spec.res_p[j][0];
    pl.plot((dlo+dhi)/2.0,res,{dlo,dhi},{res_m,res_p};minmax,sym=popt.dsym[j],color=popt.colors[j],size=popt.symsize[j],depth=popt.dataDepth[j]);
  }
  pl.plot((dlo+dhi)/2.0,Double_Type[length(dlo)]+(popt.ratio?1.0:0.0);color="black",line=XFIG_LINESTYLE_DASHED,depth=min(popt.dataDepth)-1);
  return pl;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% xfps_create_string_array()
% Initialize an array with a given value
% (mainly to do a String_Type[] array more easily)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define xfps_create_string_array(n,str) {
  variable arr = typeof(str)[n];
  variable i;
  _for i (0,n-1,1) arr[i] = str;
  return arr;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% xfps_labels()
% Return appropriate labels for units
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
private define xfps_labels(unit) {
  variable labelArray;
  switch (unit)
  { case "kev" : labelArray = [XFPS_LABEL_X_KEV,XFPS_LABEL_Y_KEV]; }
  { case "a" : labelArray = [XFPS_LABEL_X_A,XFPS_LABEL_Y_A]; }
  { labelArray = ["UNRECOGNIZED UNITS: "+unit,"UNRECOGNIZED UNITS: "+unit]; }
  return labelArray;
}

#endif

#endif

%
% Code from src/plot/xfig_plot_unfold.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

%  ===================  %
private define prepare_data_for_pp_plot(d)
%  ===================  %
{
  d = [d];

  variable len = length(d);

  variable dfieldnames,fn,f;
  variable pp = Struct_Type[len];
  variable pmin,pmax;

  variable i;
  _for i ( 0, len-1, 1 ){
    dfieldnames = get_struct_field_names(d[i]);

    pp[i] = struct_combine(dfieldnames);
    pmin = min(d[i].lo);
    pmax = max(d[i].hi);

    foreach fn(dfieldnames){
      f = get_struct_field( d[i], fn );

      if( fn=="lo" or fn=="hi" )
	set_struct_field( pp[i], fn, [f,f + pmax-pmin] - pmin );
      else
	set_struct_field( pp[i], fn, [f,f] );
    }
  }

  return len == 1 ? pp[0] : pp;
}

%  ===================  %
define xfig_plot_data(pl,d)
%  ===================  %
%!%+
%\function{xfig_plot_data}
%\synopsis{plots data which was loaded by "read_data_from_write_plot"}
%\usage{XFig_Plot_Type pl = xfig_plot_data(XFig_Plot_Type pl, Struct_Type dat)}
%\qualifiers{
%\qualifier{color}{specify the color}
%\qualifier{width}{line width}
%\qualifier{depth}{depth in plot}
%\qualifier{pp}{data are plotted twice (pulse profile)}
%}
%\seealso{xfig_plot_unfold,read_data_from_write_plot,read_col,write_plot}
%!%-
{
  if( qualifier_exists("pp") )
    d = prepare_data_for_pp_plot(d);

  pl.plot((d.lo+d.hi)*0.5, d.val,
	{d.lo, d.hi}, d.err;
	  sym=qualifier("sym","point"),
	  size=0,
	  color=qualifier("color",1),
	  xminmax,
	  width=qualifier("width",3),
	  depth=qualifier("depth",50)
	 );

  return pl;
}

%  ===================  %
define xfig_plot_model(pl,d)
%  ===================  %
%!%+
%\function{xfig_plot_model}
%\synopsis{plots the model which was loaded by "read_data_from_write_plot"}
%\usage{XFig_Plot_Type pl = xfig_plot_model(XFig_Plot_Type pl, Struct_Type dat)}
%\qualifiers{
%\qualifier{color}{specify the color}
%\qualifier{width}{line width}
%\qualifier{depth}{depth in plot}
%\qualifier{hplot}{make it a histogram plot}
%\qualifier{pp}{model is plotted twice (pulse profile)}
%}
%\seealso{xfig_plot_data,read_data_from_write_plot,read_col,write_plot}
%!%-
{
  if( qualifier_exists("pp") )
    d = prepare_data_for_pp_plot(d);

   if (qualifier_exists("hplot")){
      pl.hplot([d.lo,d.hi[-1]], d.model;
	       color=qualifier("color","black"),
	       width=qualifier("width",3),
	       depth=qualifier("depth",50)
	      );
   } else {      
      pl.plot((d.lo+d.hi)*0.5, d.model;
	      color=qualifier("color","black"),
	      width=qualifier("width",3),
	      depth=qualifier("depth",50)
	     );
   }

  return pl;
}

%  ===================  %
define xfig_plot_res(pl,s)
%  ===================  %
%!%+
%\function{xfig_plot_res}
%\synopsis{plots the residuals which were loaded by "read_data_from_write_plot"}
%\usage{XFig_Plot_Type pl = xfig_plot_model(XFig_Plot_Type pl, Struct_Type dat)}
%\qualifiers{
%\qualifier{color}{specify the color}
%\qualifier{width}{line width}
%\qualifier{depth}{depth in plot}
%\qualifier{chi}{specifies that the residuals are in chi}
%\qualifier{ratio}{specifies that the residuals are a ratio}
%\qualifier{pp}{residuals are plotted twice (pulse profile)}
%}
%\seealso{xfig_plot_data,read_data_from_write_plot,read_col,write_plot}
%!%-
{
  if( qualifier_exists("pp") )
    s = prepare_data_for_pp_plot(s);

  pl.plot((s.lo+s.hi)*0.5, s.res,{s.lo, s.hi}, {s.res_min, s.res_max};
	  sym="point",
	  size=0,color=qualifier("color",1),
	  eb_factor=0,
	  minmax,
	  width=qualifier("width",3),
	  depth=qualifier("depth",50));

  pl.plot([0,1],[0,0];depth=90, width=3, line=0, world01);

  if (qualifier_exists("chi"))
  {
    pl.ylabel(`$\chi$`);
    pl.plot([0,1],[0,0];world01,width=3);
  }
  if (qualifier_exists("ratio"))
  {
    pl.ylabel(`Ratio`);
    pl.plot([0,1],[1,1];world01,width=3);
  }

  return pl;
}

%  ===================  %
define xfig_plot_unfold(d)
%  ===================  %
%!%+
%\function{xfig_plot_unfold}
%\synopsis{tries to plot the complete plot loaded by "read_data_from_write_plot"}
%\usage{XFig_Plot_Type pl = xfig_plot_unfold(Struct_Type dat)}
%\qualifiers{
%\qualifier{size}{[dx,dy,dr] size of the xfig-plot}
%\qualifier{dcol}{color of the data}
%\qualifier{mcol}{color of the model}
%\qualifier{xrng}{[xmin,xmax]}
%\qualifier{yrng}{[ymin,ymax]}
%\qualifier{rrng}{[rmin,rmax]}
%\qualifier{ranges}{[xmin,xmax,ymin,ymax,rmin,rmax]}
%\qualifier{chi}{specifies that the residuals are chi}
%\qualifier{ratio}{specifies that the residuals are ratio}
%\qualifier{y_label}{set a y-label for the data/model plot}
%\qualifier{keV2erg_fac}{if given, the y-label is given in units of
%                          ergs/s/cm^2 x 10^keV2erg_fac }
%\qualifier{pp}{data is plotted twice (pulse profile)}
%}
%\description
%    With "read_data_from_write_plot", previously stored data can be
%    loaded into a structure. This structure ("dat" in the example
%    above) can now be plotted with "xfig_plot_unfold".
%
%    The routines xfig_plot_data, xfig_plot_model, and
%    xfig_plot_res do the single steps on its own.
%\seealso{xfig_plot_data,xfig_plot_model,xfig_plot_res,read_data_from_write_plot,write_plot}
%!%-
{
  % make it always an array
  d = [d];
  variable n = length(d);
  
  variable size = qualifier("size",[14,8,4]);
  variable pl = xfig_plot_new(size[0],size[1]);
  variable pl_res = xfig_plot_new(size[0],size[2]);

  variable xlabel = qualifier("xlabel",`Energy [keV]`);
  variable pp = d;
  if( qualifier_exists("pp") ){
    pp = prepare_data_for_pp_plot(d);
    xlabel = qualifier("xlabel","pulse phase");
  }

  % Ranges
  variable xrng = [1,-1] * DOUBLE_MAX;
  variable yrng = [1,-1] * DOUBLE_MAX;
  variable rrng = [1,-1] * DOUBLE_MAX;

  variable i;
  _for i ( 0, n-1, 1 ){
    xrng[0] = min( [ xrng[0], pp[i].lo ] );
    xrng[1] = max( [ xrng[1], pp[i].hi ] );
    yrng[0] = min( [ yrng[0], pp[i].val-pp[i].err ] );
    yrng[1] = max( [ yrng[1], pp[i].val+pp[i].err ] );
    rrng[0] = min( [ rrng[0], pp[i].res_min ] );
    rrng[1] = max( [ rrng[1], pp[i].res_max ] );
  }
  xrng = qualifier("xrng",xrng);
  yrng = qualifier("yrng",yrng);
  rrng = qualifier("rrng",rrng);
  variable ranges = qualifier("ranges",[xrng,yrng,rrng]);

  variable quali = struct_combine( struct{ padx=0.01, pady=0.05 }, __qualifiers );
  
  if (qualifier_exists("pp")){
    pl.world(ranges[0],ranges[1],ranges[2],ranges[3];; quali );
    pl_res.world(ranges[0],ranges[1],ranges[4],ranges[5];; quali );
  }
  else{
    pl.world(ranges[0],ranges[1],ranges[2],ranges[3];;
	     struct_combine( struct{loglog}, quali ) );
    pl_res.world(ranges[0],ranges[1],ranges[4],ranges[5];;
		 struct_combine( struct{xlog}, quali) ) ;
  }

  variable dcol = [qualifier("dcol",[1:n])];
  variable mcol = [qualifier("mcol",n==1? ["black"] : dcol )];
  
  _for i(0,n-1,1){
    quali = struct_combine(__qualifiers,struct{col=dcol[i]});
    pl = xfig_plot_data(pl,d[i];;quali);
    pl_res = xfig_plot_res(pl_res,d[i];;quali);
    
    quali.col = mcol[i];
    pl = xfig_plot_model(pl,d[i];;quali);
  }
  
  variable y_fac = qualifier("keV2erg_fac",NULL);
  if (y_fac != NULL) {   pl.ylabel(sprintf(`$ \nu F_\nu \; [10^{%d} \times$ ergs\;\;s$^{-1}$\;cm$^{-2}]$`,y_fac)); }
  
  variable y_label = qualifier("y_label",NULL);
  if (y_label!=NULL) pl.ylabel(y_label);

  return xfig_multiplot(pl,pl_res;;
			struct_combine(struct{xlabel=xlabel},__qualifiers));
}


#endif

%
% Code from src/plot/xfig_polarplot_new.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

define xfig_polarplot_new()
%!%+
%\function{xfig_polarplot_new}
%\synopsis{creates a xfig_plot with polar axis}
%\usage{Struct_Type xfig_plot_new( Double_Type Size )}
%\altusage{Struct_Type xfig_plot_new()}
%\qualifiers{
%\qualifier{S}{[=10] Size of quadratic plot in width and height [cm].
%                      Is overwritten by argument size!}
%\qualifier{min}{[=0] Minimal angle of polar plot [degree]}
%\qualifier{max}{[=180] Maximal angle of polar plot [degree]}
%\qualifier{origin}{[=90] Origin of angular axis measured from x-axis [degree]}
%\qualifier{dir}{[=-1] Clockwise direction of polar plot axis (-1/1)}
%\qualifier{ticlabels}{[=.6] If 0 ticlabels are turned of. If not 0 it is used
%                       as relative justification}
%\qualifier{ticlabelsize}{[="small"] Size of ticlabels}
%\qualifier{ticlabelrotate}{[=1] Rotate polar ticlabels corresponding to their position (0/1)}
%\qualifier{ticthickness}{[=2] Thickness of tics}
%\qualifier{smallticinc}{[=2] Increment for small tics [degree]}
%\qualifier{medticinc}{[=10] Increment for medium tics [degree]}
%\qualifier{bigticinc}{[=20] Increment for big tics [degrees]}
%\qualifier{aticlables}{[=[min:max:bigticinc]] Ticlabels of angular axis}
%\qualifier{aticformat}{["$%d^\\circ$"] Format for the angular ticlabels}
%\qualifier{rlabel}{[=NULL] Label for the radial axis}
%\qualifier{rtics}{[=11] Number of radial tics}
%\qualifier{rticlabel}{[ =[0:1:#rtics/2] ] Radial tic labels. The 0th rtic label will not be drawn!}
%\qualifier{norticlabel}{If given, disabels radial tic labels.}
%\qualifier{rticlabelrotate}{[=0] Rotate polar ticlabels corresponding to their position (0/1)}
%\qualifier{padr}{[=0.05] Padding of radial axis}
%\qualifier{grid}{If given grid lines are plotted}
%\qualifier{gridcolor}{[="#BBBBBB"] Color of the grid lines}
%\qualifier{debug}{If given the x-y-axis of the underlying xfig_plot are shown}
%}
%\description
%   This function returns a xfig_plot_new structure but imprinted with polar axis.
%   This way one can use the functionality of xfig_plot_new. The polar axis
%   can be modified with the qualifiers above.
%
%   NOTE that when using %.(h)plot still requries 'x' and 'y' values, not 'radius'
%   and 'angle'. That means the user has to do the transformation manually accounting
%   for the 'origin', 'direction' and 'minimal'/'maximal' values (see examples).
%
%   TO BE IMPLEMENTED:
%   * functionality to us 'radius' and 'angle' for plotting!
%
%\example
%   % Plot coordinate transformation
%   variable min = 0,
%            max = 180,
%            dir = 1,
%            org = 45;
%   variable ang = [min:max:#100];
%   variable rad = 0.5+0.5*sin(ang*PI/180);
%   variable xf = xfig_polarplot_new(; min=min, max=max, origin=org, dir=dir, grid );
%   xf.plot( rad*cos( dir*(ang+org)*PI/180 ),
%            rad*sin( dir*(ang+org)*PI/180 ) ; color="red" );
%   xf.render("/tmp/test.pdf");
%
%\seealso{xfig_plot_new}
%
%!%-
{

  %%%%% QUALIFIERS & ARGUMENTS

  %% Plot Size
  variable S = qualifier("S",10);
  if(_NARGS){
    S = __pop_args(1);
    S = S[0].value;
    _pop_n(_NARGS-1); % get rid of all other arguments
  }

  %% Plot Limits & Orientation
  variable min = int(qualifier("min",0)) mod 360;
  variable max = int(qualifier("max",180)+.5) mod 360;
  if( min == max ) max += 360;
  variable origin = nint(qualifier("origin",90));
  variable dir  = sign(qualifier("dir",-1));
  
  variable ticlabels = qualifier("ticlabels",.6);
  variable ticlabelsize = qualifier("ticlabelsize","small");
  variable ticlabelrotate = qualifier("ticlabelrotate",1);
  variable ticthickness = qualifier("ticthickness",2);
  variable smallticinc  = int(qualifier("smallticinc",2));
  variable medticinc    = int(qualifier("medticinc",10));
  variable bigticinc    = int(qualifier("bigticinc",20));

  variable tlabels = qualifier("aticlabels", [min:max:bigticinc] );
  variable tlabelformat = qualifier("aticformat","$%d^\\circ$");
  
  variable rlabel = qualifier("rlabel");
  variable rtics = int(qualifier("rtics",11));
  variable rticlabelrotate = qualifier("rticlabelrotate",0);
  variable rticlabel = qualifier("rticlabel", array_map( String_Type, &sprintf, "$%.1f$",[0:1:#int(ceil(rtics/2.))]) );
  
  variable padr = qualifier("padr",0.05);
  padr = qualifier("padx",padr);
  padr = qualifier("pady",padr);

  variable gridlines = qualifier_exists("grid") ? 1 : 0;
  variable gridcolor = qualifier("gridcolor","#BBBBBB");
  
  %% ANGULAR & RADIAL TICSIZES
  variable
    rmax = .5*S,
    bigtic = rmax*.05,
    medtic = 0.6*bigtic,
    smalltic = 0.3*bigtic;

  variable rad = [0:rmax*(1-padr):#rtics];
  
  variable obj = xfig_new_compound_list ();
  variable tics = xfig_new_polyline_list ();
  variable grid = xfig_new_polyline_list ();

  %% ANGULAR AXIS, TICS & GRID
  variable theta, theta_rad, ticsize, draw_label, ii=-1;
  _for theta ( min, max, smallticinc ){
    theta_rad = (origin+dir*theta) * PI/180.0;
    draw_label = 0;
    ticsize = smalltic;
    if ((theta mod medticinc) == 0){
      ticsize = medtic;
    }
    if ((theta mod bigticinc) == 0){
      ++ii;
      ticsize = bigtic;
      draw_label = 1;
    }
    
    variable rmin = rmax - ticsize;
    variable
      xs = [rmin, rmax]*cos(theta_rad),
      ys = [rmin, rmax]*sin(theta_rad),
      zs = [0, 0];
    tics.insert (vector (xs, ys, zs));
    
    if (draw_label and ticlabels!=0 and not(theta==max and (max mod 360) == (min mod 360) ) ){
      variable label = xfig_new_text ( sprintf(tlabelformat,tlabels[ii mod ((max-min)/bigticinc+1)] ); size   = ticlabelsize );
      variable lheight = _diff(__push_array([label.get_bbox()][[2,3]]));
      
      label = xfig_new_text ( sprintf(tlabelformat,tlabels[ii mod ((max-min)/bigticinc+1)] );
			      size   = ticlabelsize,
			      rotate = ticlabelrotate*(theta_rad-PI/2)*180/PI,
			      x0 = xs[1] + ticlabelrotate*ticlabels*lheight*cos(theta_rad),
			      y0 = ys[1] + ticlabelrotate*ticlabels*lheight*sin(theta_rad),
			      just = -not(ticlabelrotate)*ticlabels*[cos(theta_rad),sin(theta_rad)],
			    );
      label.set_depth (0);
      obj.insert (label);
      
      %% ANGULAR GRID
      if( gridlines ){
	grid.insert( vector( [S*0.01,rmin], theta_rad*[1,1], PI/2*[1,1] ; sph ) );
      }
    }
  }
  tics.insert( vector( rmax*[1:1:#100], (origin+dir*[min:max:#100])*PI/180., PI/2*[1:1:#100] ; sph ) );
  
  
  %% RADIAL TICS
  variable r;
  foreach theta ( [min,max] ){
    theta_rad = (origin+dir*theta) * PI/180.0;
    tics.insert (vector ( [0,rmax], theta_rad*[1,1], PI/2*[1,1] ; sph));

    _for r ( 1, rtics-1 ){
      ticsize = smalltic;
      draw_label = 0;
    
      if ((r mod 2) == 0){
	ticsize = medtic;
	draw_label = 1;
      }
      tics.insert( vector_rotate(vector( (theta==min?-1:(max mod 360 == min?-1:1))*dir*ticsize*[0,1],
					 rad[[r,r]],
					 [0,0]),
				 vector(0,0,1),
				 theta_rad-PI/2 )
		 );
      if (draw_label and ticlabels!=0 and theta == min ){
	%label = xfig_new_text ( sprintf("$%.1f$",1.*r/(rtics-1)); size = ticlabelsize );
	ifnot( qualifier_exists("norticlabel") ){
	  lheight = _diff(__push_array([label.get_bbox()][[0,1]]));
	  label = xfig_new_text( rticlabel[r/2];%sprintf("$%.1f$",1.*r/(rtics-1));
				 size   = ticlabelsize,
				 rotate = rticlabelrotate*(theta_rad*180/PI - sign(sin(theta_rad))*90),
				 x0   = rad[r] * cos(theta_rad) + rticlabelrotate*dir*ticlabels*lheight * sin(theta_rad),
				 y0   = rad[r] * sin(theta_rad) - rticlabelrotate*dir*ticlabels*lheight * cos(theta_rad),
				 just = not(rticlabelrotate)*dir*ticlabels*[-sin(theta_rad),cos(theta_rad)],
			       );
	  obj.insert( label );
	}

	%% RADIAL GRID
	if( gridlines ){
	  grid.insert( vector( rad[r]*[1:1:#100], (origin+dir*[min:max:#100])*PI/180., PI/2*[1:1:#100] ; sph ) );
	}
      }
    }
  }
  if( qualifier_exists("rlabel") ){
    theta_rad = (origin+dir*max) * PI/180.0;
    
    label = xfig_new_text ( rlabel; size = ticlabelsize );
    lheight = _diff(__push_array([label.get_bbox()][[2,3]]));
    label = xfig_new_text ( rlabel;
			    size   = ticlabelsize,
			    rotate = theta_rad*180/PI - (sign(cos(theta_rad)) <= 0 ? 180 : 0),
			    x0 = .5*rmax*cos(theta_rad) - dir*ticlabels*lheight * sin(theta_rad),
			    y0 = .5*rmax*sin(theta_rad) + dir*ticlabels*lheight * cos(theta_rad),
			    just = [0,0]
			  );
    obj.insert( label );
  }
    
  tics.set_thickness (ticthickness);
  tics.set_depth (20);
  obj.insert (tics);

  grid.set_thickness( nint(_max(1,ticthickness-2)) );
  grid.set_depth (100);
  grid.set_pen_color (gridcolor);
  obj.insert (grid);
  
  %% MERGE TICS & AXIS into a xfig_plot object for further plotting
  variable xf = xfig_plot_new(S,S);
  xf.world(-1./(1-padr),1./(1-padr),-1./(1-padr),1./(1-padr);padx=0,pady=0);
  ifnot( qualifier_exists("debug") )
  {
    xf.axes(;off);
  }
  else{
    xf.plot(0,0);
  }
    
  
  %% Add circle at origin of both plots as marker of positioning
  variable circle = xfig_new_ellipse (S*0.01, S*0.01);
  obj.insert (@circle);
  xf.add_object(@circle, 0, 0 );

  %%% Positioning of xfig_plot_new and the polarplot object
  variable x0,y0;
  (x0,,y0,,,) = obj.get_bbox();
  xf.add_object( obj, 0.5+x0/S, 0.5+y0/S, -.5, -.5 ; world00 );

  return xf;
}

#endif

%
% Code from src/misc/astronomy/xfigplot_hardnessratio_grid.sl
%

% only activate if module xfig could be loaded
#ifeval __get_reference("xfig_plot_new")!=NULL

%%%%%%%%%%%%%%%%%%%%%%%%
define xfigplot_hardnessratio_grid()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{xfigplot_hardnessratio_grid}
%\synopsis{plots hardness-grid and/or hardness-datapoints}
%\usage{xfigplot_hardnessratio_grid(Struct_Type tracks);
% or xfigplot_hardnessratio_grid(Struct_Type tracks, String_Type fits_hr_table
%    );}
%\qualifiers{
%\qualifier{outputdir}{String_Type, name of output directory,
%          default: 'testplot.pdf'}
%\qualifier{W}{Integer_Type, width of new xfig-plot}
%\qualifier{H}{Integer_Type, height of new xfig-plot}
%\qualifier{xrange}{Double_Type[2], range on x-axis}
%\qualifier{yrange}{Double_Type[2], range on y-axis}
%\qualifier{padx}{Double_Type, percentage of padding of whole plot in x direction,
%          default: 0.05}
%\qualifier{pady}{Double_Type, percentage of padding of whole plot in y direction,
%          default: 0.05}
%\qualifier{xlabel}{String_Type, label on x-axis}
%\qualifier{ylabel}{String_Type, label on y-axis}
%\qualifier{par1label}{String_Type, label within plot of par1}
%\qualifier{par1label_x}{Double_Type, x-position of 'par1label' as percentage 
%          (needs to be set, if 'par1label' is given)}
%\qualifier{par1label_y}{Double_Type, y-position of 'par1label' as percentage
%          (needs to be set, if 'par1label' is given)}
%\qualifier{par1label_c}{Integer_Type/String_Type, color of 'par1label' as RGB,
%          default: 0x000000 (black)}
%\qualifier{par2label}{String_Type, label within graph of par2}
%\qualifier{par2label_x}{Double_Type, x-position of 'par2label' as percentage
%          (needs to be set, if 'par2label' is given)}
%\qualifier{par2label_y}{Double_Type, y-position of 'par2label' as percentage
%          (needs to be set, if 'par2label' is given)}
%\qualifier{par2label_c}{Integer_Type/String_Type, color of 'par2label' as RGB,
%          default: 0x000000 (black)}
%\qualifier{extralabel}{String_Type, extra label within graph}
%\qualifier{extralabel_x}{Double_Type, x-position of 'extralabel' as percentage
%          (needs to be set, if 'extralabel' is given)}
%\qualifier{extralabel_y}{Double_Type, y-position of 'extralabel' as percentage
%          (needs to be set, if 'extralabel' is given)}
%\qualifier{extralabel_c}{Integer_Type/String_Type, color of 'extralabel' as RGB,
%          default: 0x000000 (black)}
%\qualifier{extralabel2}{String_Type, extra label within graph}
%\qualifier{extralabel2_x}{Double_Type, x-position of 'extralabel2' as percentage
%          (needs to be set, if 'extralabel' is given)}
%\qualifier{extralabel2_y}{Double_Type, y-position of 'extralabel2' as percentage
%          (needs to be set, if 'extralabel' is given)}
%\qualifier{extralabel2_c}{Integer_Type/String_Type, color of 'extralabel2' as RGB,
%          default: 0x000000 (black)}
%\qualifier{cstart}{Integer_Type/String_Type, start color of grid as RGB,
%          default: 0xC1CDCD (gray)}
%\qualifier{cstop}{Integer_Type/String_Type, stop color of grid as RGB}
%          default: 0xC1CDCD (gray)}
%\qualifier{lend1}{if exists, switch end (or begin) of tracks1 
%          (where each tracks' value will be assigned)}
%\qualifier{lend2}{if exists, switch end (or begin) of tracks2
%          (where each tracks' value will be assigned)}
%\qualifier{shift1x}{Double_Type, x-position of par1-values at end of track, default: 0.}
%\qualifier{shift1y}{Double_Type, y-position of par1-values at end of track, default: 0.}
%\qualifier{shift2x}{Double_Type, x-position of par2-values at end of track, default: 0.}
%\qualifier{shift2y}{Double_Type, y-position of par2-values at end of track, default: 0.}
%\qualifier{hr_x_col}{String_Type, name of hr_x-column ('fits_hr_table') for plotting datapoints}
%\qualifier{hr_y_col}{String_Type, name of hr_y-column ('fits_hr_table') for plotting datapoints}
%\qualifier{err_x_col}{String_Type, name of error_x-column ('fits_hr_table') for plotting datapoints}
%\qualifier{err_y_col}{String_Type, name of error_y-column  ('fits_hr_table)' for plotting datapoints}
%\qualifier{err_c}{Integer_Type/String_Type, color of errorbars as RGB
%          default: 0x000000 (black)}
%\qualifier{symbol_shape}{String_Type, shape of datapoints}
%\qualifier{symbol_size}{Double_Type, size of datapoints, default: 1.}
%\qualifier{symbol_cstart}{Integer_Type/String_Type, (start) color of datapoints as RGB}
%\qualifier{symbol_cstop}{Integer_Type/String_Type, stop color of datapoints as RGB,
%          if not given 'symbol_cstart' is color of all}
%\qualifier{source_col}{String_Type, name of source-column ('fits_hr_table') for plotting datapoints,
%          needs to be given only if different colors for different sources}
%\qualifier{extralabel_sources_x}{Double_Type, x-position of source_names as percentage
%          (if given, source names are displayed in their respective color)}
%\qualifier{extralabel_sources_y}{Double_Type, y-position of source_names as percentage
%\qualifier{tracks1}{Integer_Type, if given, only tracks1 are plottet}
%\qualifier{tracks2}{Integer_Type, if given, only tracks2 are plottet}
%\qualifier{notracks}{Integer_Type, if given, no tracks are plottet, 'fits_hr_table' must be
%          given for datapoints}
%}
%\description
%	- for only grid: give only tracks
%	- for grid with datapoints: give tracks and FITS-table with hardnessratios (and errors)
%       - for only datapoints: give tracks and FITS-table, but use 'notracks'-qualifier
%\seealso{hardnessratio_from_dataset}
%!%-
{
    variable i;
    variable pl, tracks, fits_hr_table;
    switch (_NARGS)
    { case 1: (tracks) = (); }
    { case 2: (tracks,fits_hr_table) = (); }
    { help(_function_name); return; }

    % create empty plot
    pl = xfig_plot_new(qualifier("W", 14), qualifier("H", 10)); 

    % set output directory  
    variable outputdir = qualifier("outputdir", "testplot.pdf");

    % check x/yrange-input
    if(qualifier_exists("xrange")){
      variable xrange = qualifier("xrange", NULL);
      if (length(xrange) != 2 ){
       vmessage("error (%s): xrange-qualifier must be array with 2 elements", _function_name);
       return;
      }
    }
    if(qualifier_exists("yrange")){
      variable yrange = qualifier("yrange", NULL);
      if (length(yrange) != 2 ){
        vmessage("error (%s): yrange-qualifier must be array with 2 elements", _function_name);
        return;
      }
    }

    % labels?
    variable xlabel = qualifier("xlabel", "x-axis");
    variable ylabel = qualifier("ylabel", "y-axis");
    variable par1label;
    variable par2label;
    variable extralabel;
    variable extralabel2;
    if(qualifier_exists("par1label")){par1label=qualifier("par1label");}
    if(qualifier_exists("par2label")){par2label=qualifier("par2label");}
    if(qualifier_exists("extralabel")){extralabel=qualifier("extralabel");}
    if(qualifier_exists("extralabel2")){extralabel2=qualifier("extralabel2");}
  
    % colors of grid?
    variable cstart = qualifier("cstart", 0xC1CDCD);
    variable cstop = qualifier("cstop", 0xC1CDCD);
    xfig_new_color("start",cstart);
    xfig_new_color("stop",cstop);
    variable col=String_Type[length(tracks)];
    _for i (0,length(tracks)-1,1) {
      col[i]=xfig_mix_colors("start","stop",i/(length(tracks)-1.));
    }

    % set world?
    variable padx = qualifier("padx",0.05);
    variable pady = qualifier("pady",0.05);
    if (pl.plot_data.world1_inited == 0) { 
      variable merge = merge_struct_arrays(tracks);
      if(qualifier_exists("xrange") && qualifier_exists("yrange")){
        pl.world(xrange[0],xrange[1],yrange[0],yrange[1]; padx=padx, pady=pady);
      }else if(qualifier_exists("xrange")){
        pl.world(xrange[0],xrange[1], min_max(merge.hr2); padx=padx, pady=pady);
      }else if(qualifier_exists("yrange")){
        pl.world(min_max(merge.hr1), yrange[0],yrange[1]; padx=padx, pady=pady);
      }else{
        pl.world(min_max(merge.hr1), min_max(merge.hr2); padx=padx, pady=pady);
      }
    }
  
    % lineend (where the numbers are plotted at the end of each track)
    variable lend1 = qualifier_exists("lend1") ? -1 : 0;
    variable lend2 = qualifier_exists("lend2") ? -1 : 0;

    % shift of numbers at each track
    variable shift1x =  qualifier("shift1x",0.);
    variable shift1y =  qualifier("shift1y",0.);
    variable shift2x =  qualifier("shift2x",0.);
    variable shift2y =  qualifier("shift2y",0.);

    variable ind1 = where( struct_array_2_struct_of_arrays( tracks ).linetype < 0 );
    variable ind2 = where( struct_array_2_struct_of_arrays( tracks ).linetype > 0 );
    % loop over all tracks

    variable tr1=0;
    variable tr2=0;
    variable notr=0;
    if(qualifier_exists("notracks")){notr=1;}
    if(qualifier_exists("tracks1")){tr1=1;}
    if(qualifier_exists("tracks2")){tr2=1;}

   if(notr == 0){
      if((tr1 == 1 && tr2 == 0) || (tr1 == 0 && tr2 == 0)){
       foreach i (ind1) {
         col[i]=xfig_mix_colors("start","stop",i/(length(ind1)-1.));
         pl.plot(tracks[i].hr1, tracks[i].hr2; line=0, color = col[i]);%,depth=1,width=3);
         pl.xylabel( tracks[i].hr1[lend1], tracks[i].hr2[lend1],
    	        sprintf("%.2f",tracks[i].par1[lend1]),shift1x,shift1y;color=col[i]);%,depth=1,width=3);
       }
      }

      if((tr2 == 1 && tr1 == 0) || (tr1 == 0 && tr2 == 0)){
       foreach i (ind2) {
         col[i]=xfig_mix_colors("start","stop",(i-min(ind2))/(length(ind2)-1.));
         pl.plot(tracks[i].hr1, tracks[i].hr2; line=0, color = col[i]);%,depth=1,width=3);
         pl.xylabel( tracks[i].hr1[lend2], tracks[i].hr2[lend2],
    	        sprintf("%.2f",tracks[i].par2[lend2]),shift2x,shift2y;color=col[i]);%,depth=1,width=3);
       }  
     }

   }

    % ensure all numbers have the same format
    pl.y1axis(; format="%.2f");
    pl.x1axis(; format="%.2f");

    % labels 
    pl.xlabel(xlabel);
    pl.ylabel(ylabel);
    if(qualifier_exists("par1label")){
      variable par1label_x = qualifier("par1label_x", 0.1);
      variable par1label_y = qualifier("par1label_y", 0.9);
      variable par1label_c = qualifier("par1label_c", 0x000000);
      pl.xylabel(par1label_x,par1label_y,sprintf(par1label);world0, color=par1label_c);%,depth=1,width=3);
    }
  
    if(qualifier_exists("par2label")){
      variable par2label_x = qualifier("par2label_x", 0.1);
      variable par2label_y = qualifier("par2label_y", 0.8);
      variable par2label_c = qualifier("par2label_c", 0x000000);
      pl.xylabel(par2label_x,par2label_y,sprintf(par2label);world0, color=par2label_c);%,depth=1,width=3); 
    }

    if(qualifier_exists("extralabel")){
      variable extralabel_x = qualifier("extralabel_x", 0.1);
      variable extralabel_y = qualifier("extralabel_y", 0.5);
      variable extralabel_c = qualifier("extralabel_c", 0x000000);
      pl.xylabel(extralabel_x,extralabel_y,sprintf(extralabel);world0, color=extralabel_c);%,depth=1); 
    }

    if(qualifier_exists("extralabel2")){
      variable extralabel2_x = qualifier("extralabel2_x", 0.1);
      variable extralabel2_y = qualifier("extralabel2_y", 0.5);
      variable extralabel2_c = qualifier("extralabel2_c", 0x000000);
      pl.xylabel(extralabel2_x,extralabel2_y,sprintf(extralabel2);world0, color=extralabel2_c);%,depth=1); 
    }

    pl.render(outputdir);

    %plot datapoints if given
    switch (_NARGS)
    { case 2: 
      %check whether names of hardnessratio-columns in FITS-file are given
      variable hr_x = qualifier("hr_x_col", NULL); 
      variable hr_y = qualifier("hr_y_col", NULL);
      if (hr_x == NULL || hr_y == NULL ){
       vmessage("error (%s): column names (hr_x_col,hr_y_col) of fits-table for plotting datapoints must be given", _function_name);
       return;
      } 
      %check whether names of hardnessratio_error-columns in FITS-file are given    
      variable err_x = qualifier("err_x_col", NULL); 
      variable err_y = qualifier("err_y_col", NULL);
      if (err_x == NULL || err_y == NULL ){
       vmessage("error (%s): column names (err_x_col,err_y_col) of fits-table for errorbars must be given", _function_name);
       return;
      }  

      %qualifiers
      variable fitsfile = fits_open_file(fits_hr_table,"r");
      variable x_col = fits_get_colnum(fitsfile,hr_x);  
      variable y_col = fits_get_colnum(fitsfile,hr_y);
      variable errx_col = fits_get_colnum(fitsfile,err_x);
      variable erry_col = fits_get_colnum(fitsfile,err_y);
      variable numrows = fits_get_num_rows(fitsfile);
      variable symbol_shape =  qualifier("symbol_shape", "point"); 
      variable symbol_size =  qualifier("symbol_size", 1.); 
      variable symbol_cstart =  qualifier("symbol_cstart", 0x000000); 
      variable err_c =  qualifier("err_c", 0x000000);      
      variable extralabel_sources_x;
      variable extralabel_sources_y;
      extralabel_sources_x=qualifier("extralabel_sources_x",0.5);
      extralabel_sources_y=qualifier("extralabel_sources_y",0.95);

      %for mixing colors: create list with all entries init: symbol_start
      variable col_symbol=(@["symbol_start"])[Integer_Type[numrows+1]];
      xfig_new_color("symbol_start",symbol_cstart);
      
      %needed for labeling of source names -> needs to be set outside loop
      

      %create mixed colors only if 'symbol_cstop is given'
      if(qualifier_exists("symbol_cstop")){  
        %column name of 'sources' must be given, in order to determine which hr's
	 %belong to the same source, e.g. same color       
        variable source_col = qualifier("source_col", NULL);
        if (source_col == NULL){
         vmessage("error (%s): column name (source_col) must be given", _function_name);
         return;
        }

        variable s_col = fits_get_colnum(fitsfile,source_col);
	variable symbol_cstop =  qualifier("symbol_cstop", 0x000000);
	xfig_new_color("symbol_stop",symbol_cstop);
	variable source_buffer=fits_read_cell(fitsfile,s_col,1);	

	variable count_sources=1;
        %get amount of different sources in fitsfile (for color-mixing)
        _for i (1,numrows,1) {
	    if(fits_read_cell(fitsfile,s_col,i) != source_buffer){
	      source_buffer=fits_read_cell(fitsfile,s_col,i);
	      count_sources++;
	    }       
        } 
              
        variable sources_label=struct{name=String_Type[count_sources+1],color=String_Type[count_sources+1]}; 
	  %color differs from source_col below (# of entries)		
	
	%reset source buffer
	source_buffer=fits_read_cell(fitsfile,s_col,1);
	%source name array for labeling
	sources_label.name[1]=source_buffer;
	sources_label.color[1]=col_symbol[1];
	
	%mixed color-tab
	variable count=1; %count of current source
	_for i (1,numrows,1) {
	   if(fits_read_cell(fitsfile,s_col,i) != source_buffer){ %only get new color, if source name differs
	     variable fraction=count*1./count_sources*1.;

	     if(fraction<1/4.)
   	      col_symbol[[i:]]=xfig_mix_colors("red","gold",fraction*4);
 	     else if(1/4.<=fraction<2/4.)
    	      col_symbol[[i:]]=xfig_mix_colors("green","red",(fraction-1/4.)*4);
  	     else
    	      col_symbol[[i:]]=xfig_mix_colors("blue","green",(fraction-2/4.)*4);

	     %col_symbol[[i:]]=xfig_mix_colors("symbol_start","symbol_stop",count/(count_sources*1.));
	     source_buffer=fits_read_cell(fitsfile,s_col,i);
	     count++;
	     sources_label.name[count]=source_buffer; 
	     sources_label.color[count]=col_symbol[i]; 
	   }       
      	}

	%label: source names if x-position is given
        if(qualifier_exists("extralabel_sources_x")){
	  variable y=0.; %height of each line
          _for i (1,count_sources,1) {	  
            pl.xylabel(extralabel_sources_x,extralabel_sources_y-y,sprintf(sources_label.name[i]);world0, color=sources_label.color[i]); 
	    y+=0.03;
    	  }
        }

      }  


      _for i (1,numrows,1) { 
      	  
	 pl.plot(fits_read_cell(fitsfile,x_col,i), fits_read_cell(fitsfile,y_col,i),
	 	fits_read_cell(fitsfile,errx_col,i),fits_read_cell(fitsfile,erry_col,i)
		; sym=symbol_shape,
	        symcolor=col_symbol[i],
		color=err_c,
		size=symbol_size, fill);%, depth=2); 	
      }

    }

    pl.render(outputdir);
}


#endif

%
% Code from src/fitting/xspec_to_isis.sl
%

% only activate if module xspec could be loaded
#ifeval __get_reference("load_xspec_local_models")!=NULL

% only activate if module xspec could be loaded
#ifeval __get_reference("load_xspec_local_models")!=NULL

private define read_xspec (file_name)

{
	if ( not qualifier_exists("nonverbose") )
		vmessage("Running Xspec.................");
	variable file_stream = popen("echo '@"+file_name+"' | xspec", "r");
	variable all_lines = fgetslines(file_stream);
	variable i=0;
	variable start_index, end_index;
try
{
	_for i(0, length(all_lines)-1, 1)
	{
		if ( string_match(all_lines[i], "===") == 1 )
			start_index=i;
		if ( string_match(all_lines[i], "___") == 1 )
			end_index=i;
	}
	variable return_parameters=String_Type[end_index-start_index-4];
	variable counter=0;
	_for i(start_index+4, end_index-1, 1)
	{
		return_parameters[i-start_index-4] = all_lines[i];
		counter++;
	}
	variable return_header = all_lines[start_index+2];
	variable model_line=all_lines[start_index+1];
	variable return_model=substr(model_line, string_match(model_line, " ")+1, string_match(model_line, "Source")-string_match(model_line, " ")-1);
	return (return_parameters, return_header, return_model);
} catch AnyError: {
	vmessage("Evaluation of "+file_name+" in Xspec failed:");
	_for i(0, length(all_lines)-1, 1)
		vmessage(all_lines[i]);
	throw ApplicationError;
}
}

private define search (input, pattern)
{
        variable count=0;
        variable i=0;
        variable len=strlen(input);
        _for i(1, len, 1)
        {
                if (substr(input, i, 1) == pattern) {count++;};
        }
        variable positions=Integer_Type[count];
        count=0;
        _for i(1, len, 1)
        {
                if (substr(input, i, 1) == pattern) {positions[count]=i; count++;};
        }
        if ( length(positions) != 0) {return positions;} else {return NULL;};   %CAREFULL: returns position for the ">" for use with substr
}

private define model_start (input, position_start)                              %position_start refers to the position of the closing ">" behind the model
{
        variable start=position_start-3;
        while ( isalnum(substr(input, start, 1) ) == 1) {start--;}
        return start;
}

private define correct_model (model_string_raw)
{
        variable len = strlen(model_string_raw);
        variable all_characters = String_Type[len];
        variable i=0;
        _for i(1, len, 1)
                all_characters[i-1]=substr(model_string_raw, i, 1);
        variable pos = search(model_string_raw, ">");
        variable models = String_Type[length(pos)];
        variable between = String_Type[length(pos)+1];
        variable start = Integer_Type[length(pos)];
        variable end = Integer_Type[length(pos)];
        variable model_string="";
        _for i(0, length(pos)-1, 1)
        {
                start[i] = model_start(model_string_raw, pos[i])+1;
                end[i] = pos[i]-model_start(model_string_raw, pos[i]);
                if ( i == 0 )
                        between[i] = substr(model_string_raw, 1, start[i]-1);
                else    
                        between[i] = substr(model_string_raw, start[i-1]+end[i-1], start[i]-start[i-1]-end[i-1]);
                models[i] = substr(model_string_raw, start[i], end[i]);
                if ( substr(between[i], 1, 1) == "(" )
                        between[i] = "*"+between[i];
                if ( substr(between[i], strlen(between[i]), 1) == ")" )
                        between[i] = between[i]+"*";
                model_string = model_string + between[i] + models[i];
        }
        i=length(pos)-1;
        between[i+1] = substr(model_string_raw, start[i]+end[i], len-end[i]-start[i]+1);
        model_string = model_string+between[i+1];
	return model_string;
}


private define get_string_starts(input)
{
        variable len=strlen(input);
        variable test;
        variable count=0;
        variable was_before=0;
        variable i=1;
        variable char_at_start=isspace(input);                                          %if there is a normal character at the start, this value is not 0
        _for i(1, len-1, 1)
        {       
                test = isspace(substr(input, i, 1));
                if ( ( (was_before == 0) )and (test == 0) )
                {       
                        was_before = 1;
                        count++;
                }
                else if ( (was_before == 1) and (test == 1) )
                {
                        was_before = 0;
                }
        }
        variable starts=Integer_Type[count];
        variable stops=Integer_Type[count];
        variable count2=0;
        _for i(1, len-1, 1)
        {       
                test = isspace(substr(input, i, 1)); 
                if ( ( (was_before == 0) ) and (test == 0) )
                {       
                        was_before = 1;
                        starts[count2] = i;
                        count2++;
                }
                else if ( (was_before == 1) and (test == 1))
                {       
                        stops[count2-1] = i;
                        was_before = 0;
                }
        }
        variable got_substrings = String_Type[length(starts)];
        if (char_at_start != 1)
        {
                starts=shift(starts, -1);
                stops=shift(stops, -1);
                starts[0] = starts[0]+1;
        }
        _for i(0, length(starts)-1, 1)
                {
                        got_substrings[i] = substr(input, starts[i], (stops[i]-starts[i]));
                }
        return (starts, stops, got_substrings);
}


private define get_index (input_array, input_value)
{
	variable i;
	_for i(0, length(input_array)-1, 1)
	{
		if ( (i<length(input_array)) and (input_array[i] == input_value ) ) {return i;};
	}
	return -1;
}

%%%%%%%%%%%%%%%%%%%%%%%%%
define xspec_to_isis (name)
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{xspec_to_isis}
%\synopsis{Parse xspec parameter file to isis parameter file}
%\usage{Integer_Type t = xspec_to_isis(String_Type xspec_par)}
%\qualifiers{
%\qualifier{set}{set loadad parameters}
%\qualifier{save}{ [=String_Type filename] save parameters as isis parameter file
%	if filename is not specified it is saved as the .xcm file but with .par suffix}
%\qualifier{nonverbose}{suppress any output not produced by errors}
%}
%\description
%	Parse an xspec parameter file \code{.xcm} to an isis parameter file.
%	If desired parameters can be set after parsing.
%	*** Warning:
%	Xspec tying expressions are only supported in the form of
%	\code{= p1}
%	*** Warning:
%	Convolution models are not supported!
%	If tied parameters is out of range of the respective parameter
%	the limits are set to fit and a warning is given.
%\example
%	isis> t=xspec_to_isis ("test.xcm"; save);
%	Running Xspec.................
%	parameters saved to test.par as:
%	diskbb(1) + nthComp(2)
%	 idx  param            tie-to  freeze         value         min         max
%	  1  diskbb(1).norm        0     0          21.7378           0       1e+10  
%	  2  diskbb(1).Tin         0     0         0.230239       0.001          10  keV
%	  3  nthComp(2).norm       0     0       0.00019705           0       1e+10  
%	  4  nthComp(2).Gamma      0     0           1.9752       1.001           5  
%	  5  nthComp(2).kT_e       0     0          1.68497     1.68497        1000  keV
%	  6  nthComp(2).kT_bb      2     0         0.230239       0.001          10  keV
%	  7  nthComp(2).inp_type   0     1                1           0           1  0/1
%	  8  nthComp(2).Redshift   0     1                0      -0.999          10  
%\seealso{set_par, load_par}
%!%-
{
%Assuming a .xcm file, then getting the name without .xcm
	if (string_match(name, ".xcm") == 0) {vmessage(name+" is not an xcm-file"); return -1;};
	variable name_file=strreplace(name, ".xcm", "");
	variable old_fun = get_fit_fun;
	variable old_pars = get_params;
	try
	{
	%Splitting the .xcm file
	        variable cut_lines, header_string, model_string_raw;
		try
		{
			(cut_lines, header_string, model_string_raw) = read_xspec(name ;; __qualifiers);
		} catch AnyError: {
			return -1;
		}
		variable model_string = correct_model(model_string_raw);
		model_string = strreplace(strreplace(model_string, "<", "("), ">", ")" );
	        variable i=0;
	        _for i (0, length(cut_lines)-1, 1)
	        {
	                cut_lines[i] = substr(cut_lines[i], 1, strlen(cut_lines[i])-1);                 %removing the \n
	        };
		variable all_lines=Array_Type[length(cut_lines)];
		variable all_lines_starts=Array_Type[length(cut_lines)];
		variable all_lines_stops=Array_Type[length(cut_lines)];
		_for i(0, length(all_lines)-1, 1)
		{
			(all_lines_starts[i], all_lines_stops[i], all_lines[i]) = get_string_starts(cut_lines[i]);
		}
		variable value_starts = string_match(header_string, "Value");
		variable parameter_indices = String_Type[length(all_lines)];
		variable parameter_model_indices = String_Type[length(all_lines)];
		variable parameter_models = String_Type[length(all_lines)];
		variable parameter_parameters = String_Type[length(all_lines)];
		variable parameter_values = Double_Type[length(all_lines)];
		variable parameter_frozen = Integer_Type[length(all_lines)];
		_for i (0, length(parameter_indices)-1, 1)
		{
			parameter_indices[i] = all_lines[i][0];
			parameter_model_indices[i] = all_lines[i][1];
			parameter_models[i] = all_lines[i][2];
			parameter_parameters[i] = all_lines[i][3];
		};
		_for i(0, length(all_lines)-1, 1) 
		{
			parameter_values[i] = atof(all_lines[i][get_index(all_lines_starts[i], value_starts)]);
			if ( all_lines[i][get_index(all_lines_starts[i], value_starts)+1] == "frozen" ) { parameter_frozen[i] = 1;}
		};
		variable parameter_strings = String_Type[length(all_lines)];
		_for i(0, length(all_lines)-1, 1)
		{
			parameter_strings[i] = parameter_models[i]+"("+parameter_model_indices[i]+")."+parameter_parameters[i];
		}
		fit_fun(model_string);
		variable temp_pars=get_params;
		_for i(0, length(all_lines)-1, 1)
		{
			try
			{
				if ( (get_par_info(parameter_strings[i]).min ) > parameter_values[i] )
				{
					set_par(parameter_strings[i], parameter_values[i], parameter_frozen[i], parameter_values[i], get_par_info(parameter_strings[i]).max);
				}
				else if ( (get_par_info(parameter_strings[i]).max ) < parameter_values[i] )
				{
					set_par(parameter_strings[i], parameter_values[i], parameter_frozen[i],get_par_info(parameter_strings[i]).min, parameter_values[i]);
				}
				else
				{
					set_par(parameter_strings[i], parameter_values[i], parameter_frozen[i]);
				}
			} catch AnyError: {
				vmessage("Error setting Parameter %i: %s = %f", i, parameter_strings[i], parameter_values[i]);
			}
		}
	%Getting which parameters are tied to another
	        variable tied_count=0;
	        variable which_tied = Integer_Type[length(all_lines)];
	        _for i(0, length(all_lines)-1, 1)
	        {
	                if ( all_lines[i][get_index(all_lines_starts[i], value_starts)+1] == "=" )
	                {
	                        if ( string_match(all_lines[i][get_index(all_lines_starts[i], value_starts)+2], "p") != 0 )
	                        {
	                                tied_count++;
	                        }
	                }
	        }
	        variable tied_from=Integer_Type[tied_count];
	        variable tied_to=Integer_Type[tied_count];
	        tied_count=0;
	        _for i(0, length(all_lines)-1, 1)
	        {
	                if ( all_lines[i][get_index(all_lines_starts[i], value_starts)+1] == "=" )
	                {
	                        if ( string_match(all_lines[i][get_index(all_lines_starts[i], value_starts)+2], "p") != 0 )
	                        {
	                                tied_from[tied_count] = i;
	                                tied_to[tied_count] = typecast(atof(strreplace(all_lines[i][get_index(all_lines_starts[i], value_starts)+2], "p", ""))-1, Integer_Type);
	                                tied_count++;
	                        }
	                }
	        }
	        _for i(0, length(tied_to)-1, 1)
	        {
			try
			{
		                tie(parameter_strings[tied_to[i]], parameter_strings[tied_from[i]]);
				
				if (get_par_info(parameter_strings[tied_to[i]]).min < get_par_info(parameter_strings[tied_from[i]]).hard_min)
				{
					if (not qualifier_exists("nonverbose"))
					{
						vmessage("***Warning: Minimum of '%s': %f violates hard limit of '%s': %f, setting new Minimum", parameter_strings[tied_to[i]],									%<<--- new line
						get_par_info(parameter_strings[tied_to[i]]).min, parameter_strings[tied_from[i]], get_par_info(parameter_strings[tied_from[i]]).hard_min);
					}
					set_par(parameter_strings[tied_to[i]], get_par(parameter_strings[tied_to[i]]), parameter_frozen[tied_to[i]], get_par_info(parameter_strings[tied_from[i]]).hard_min, get_par_info(parameter_strings[tied_to[i]]).max);
				}
				if (get_par_info(parameter_strings[tied_to[i]]).max > get_par_info(parameter_strings[tied_from[i]]).hard_max)
	                        {
					if (not qualifier_exists("nonverbose"))
					{
	                                	vmessage("***Warning: Maximum of '%s': %f violates hard limit of '%s': %f, setting new Maximum", parameter_strings[tied_to[i]], 
		                                get_par_info(parameter_strings[tied_to[i]]).max, parameter_strings[tied_from[i]], get_par_info(parameter_strings[tied_from[i]]).hard_max);
					}
	        	                set_par(parameter_strings[tied_to[i]], get_par(parameter_strings[tied_to[i]]), parameter_frozen[tied_to[i]], get_par_info(parameter_strings[tied_to[i]]).min, get_par_info(parameter_strings[tied_from[i]]).hard_max);
	                        }
	
			} catch AnyError: {
				vmessage("Error tieing parameters. Expression in Xspec: %s", all_lines[i][get_index(all_lines_starts[i], value_starts)+2]);
			}
	        }
		if (qualifier_exists("save") )
		{
			if ( qualifier("save")==NULL )
				{
		        	save_par(name_file+".par");
				if ( not qualifier_exists("nonverbose") )
				        vmessage("parameters saved to "+name_file+".par as:");
				}
			else
			{
				save_par(qualifier("save"));
				if ( not qualifier_exists("nonverbose") )
	                               vmessage("parameters saved to "+qualifier("save")+".par as:");
			}
		}
		if ( not qualifier_exists("nonverbose") )
		        list_par;
	} catch AnyError: {
		vmessage("Fatal Error");
		return -1;
	}
	if ( not qualifier_exists("set") )
	{
		set_params(temp_pars);
		fit_fun(old_fun);
		set_params(old_pars);
	}
	return 0;
}

#endif

#endif

%
% Code from src/fitting/xyfit/xyfit.sl
%
% -*- mode: slang; mode: fold -*-

% Defining a xyfit function should now be much closer to the 'fit_fun'
% to define a fit function for a binned dataset. The general idea is
% to define one fit function which operate on the metadata of the dataset
% instead of the actual data. To increase the features the so defined function
% works on a stack that is builed up during the function call. Each stack
% element is therefore the result of a subfunction in the xyfit function.
% after all stack elements are calculated the complete function gets called
% but with the sub functions replaced by the individual stack elements.
% This has the benefit, that here we do not have to implement the slang
% interpreter again, but can just use it. The tradeoff is that we use
% memory to store the sub function results. Overall this should not be to bad
% as long as the number of sub functions times number of gridpoints is small.
%
% Be aware that this implementation may break older uses of the xyfit_fun.
% Simple function calls should not be affected, but the workaround with
%
% > xyfit_fun("foo(1)");
% > xyfit_fun("bar(2)");
% > fit_fun("foo(1) + bar(2)");
%
% does not work anymore. But now
%
% > xyfit_fun ("foo(1)+bar(2)");
%
% works as expected.

% input class definition
private variable NUMBER = 0x4;
private variable NUMBER_EXP = 0x1;
private variable NUMBER_DEC = 0x2;

private variable OPERAND = 0x8;
private variable FUNCTION = 0x2;
private variable FUNCTION_XY = 0x4;
private variable FUNCTION_OPEN = 0x1;

private variable _STK = Assoc_Type[List_Type];
private variable _PAR = Assoc_Type[Int_Type];
private variable _FCT = NULL;

private variable Fmt_Colheads, Fmt_Params;
Fmt_Colheads = " idx%-*stie-to  freeze         value         min         max";
Fmt_Params   = "%3u  %-*s  %2d    %2d   %14.7g  %10.7g  %10.7g  %s%s";

%%% helper functions %{{{
define __xy () {
  variable key, el;
  switch(_NARGS)
  { case 0: _STK = Assoc_Type[List_Type];}
  { case 1: key = ();
    if (typeof(key) != String_Type) {
      variable k = assoc_get_keys(_STK);
      if (length(k) != 0)
	return _STK[k[0]];
      else
	return NULL;
    }
    %if (assoc_key_exists(_STK, key))
    %  return _STK[key];
    return 0;
  }
  { case 2: (key, el) = ();
    if (typeof(el) == String_Type) {
      % string(eval()) because number might be a variable
      variable f = sprintf("%s(%s)", key, string(eval(el)));
      if (assoc_key_exists(_STK, f))
	return _STK[f];
      return 0;
    }
    _STK[key] = el;
    return 0;
  }
  { return 0;}
}

private define __f_access (s) {
  variable global_funs = _apropos("Global", s, 3);
  variable current_funs = _apropos(current_namespace, s, 3);
  return (length(global_funs)+length(current_funs)) ? 1 : 0;
}
%}}}

private define define_xyfun (fname, fnumb) { %{{{
  variable ffname = sprintf("%s_xy_fit", fname);
  if (__get_reference(ffname)==NULL) {
    eval(`
define ${ffname}(bin_lo, bin_hi, par)
{
  variable md = get_dataset_metadata(Isis_Active_Dataset);
  ifnot(_is_struct_type(md))
    throw UsageError, "No metadata associated with dataset #$data_id. (Was it defined via define_xydata?)"$;
  variable model = md.xy_model;
  variable quali = @md.xyfit_qualifier;
  variable key = sprintf("${fname}(%d)", Isis_Active_Function_Id);
  variable x = @(model.x);
  variable y = @(model.y)*0;
  ${fname}_xyfit(&x, &y, par;; md.xyfit_qualifier);
  () = __xy(key, {x, y});
  return 0.;
}
`$, current_namespace);
    variable xyname = sprintf("%s_xyfit", fname);
    variable parnames = @(__get_reference(xyname))();
    variable normpars = Integer_Type[0];
    if (typeof(parnames) == Struct_Type) {
      normpars = parnames.norm;
      parnames = parnames.pars;
      if (_typeof(normpars) == String_Type) {
	variable newnormpars = Integer_Type[length(normpars)];
	variable i;
	_for i (0, length(normpars)-1, 1) {
	  newnormpars[i] = wherefirst(parnames == normpars[i]);
	}
	normpars = newnormpars;
      }
    }
    try {
      add_slang_function(
        fname, __get_reference(ffname), parnames, normpars
      );
    }
    catch AnyError: {
      vmessage("%s error: %s(); does not provide parameter description?", _function_name(), xyname);
      return;
    };
    set_param_default_hook(fname, __get_reference(sprintf("%s_default", xyname)));
  }
  return sprintf("+%s%s", fname, fnumb);
} %}}}

private define message_param_structs (pars) { %{{{
  % this is unfortunately a copy of the isis intrinsic function, which
  % is not accessible.
  variable len = 0;
  foreach (pars) {
    variable p = ();
    len = max([len, strlen(p.name)]);
    if (p.tie != NULL) {
      p.tie = _get_index(p.tie);
    }
    else p.tie = 0;
    
    if (p.fun != NULL) {
      p.fun = sprintf ("\n#=>  %s", strtrim(p.fun));
    }
    else p.fun = "";
    
    if (andelse
      {p.min == -_isis->DBL_MAX}
	{p.max ==  _isis->DBL_MAX})
    {
      p.min = 0;
      p.max = 0;
    }
  }
  return len;
}%}}}

private define define_generic (call) { %{{{
  eval(`
define F_fit (bin_lo, bin_hi, par) {
  variable e;
  variable md = get_dataset_metadata(Isis_Active_Dataset);
  ifnot (_is_struct_type(md))
    throw UsageError, "No metadata associated with dataset #$data_id. (Was it defined via define_xydata?)"$;
  variable model = md.xy_model;
  variable x_part = __xy(1)[0];
  model.y = ${call};
  model.x = (x_part != NULL) ? x_part : model.x;
  if (length(model.y) == 1) {
    model.y = @Double_Type[length(model.x)]+1*model.y;
  }
  __xy;
  try (e) {
    return xyfit_residuals(Isis_Active_Dataset);
  } catch AnyError:
  {
    return 0;
  }
}`$, current_namespace);
  add_slang_function("F", assoc_get_keys(_PAR)[assoc_get_values(_PAR)]);
  return "+F(1)";
} %}}}

private define tokenize (call) { %{{{
  % TODO: needs cleanup
  call = strreplace(call, " ", ""); % remove all whitespace
  variable mode = 0; % mode flag
  variable ops = ['-', '+', '*', '/', '^', '(', ')', ',']; % operator charactes
  variable tokens = {}; % buffer for found tokens
  variable c1 = 0, c2 = 0; % starting and stoping token character position
  variable callen = strlen(call);
  variable bracematch = 0; % counter for matching braces (only within xyfit fun)
  while (c2<callen) {
%{{{ loop over all characters    
    while ((callen > c2) && all(ops!=call[c2]) ) {
%{{{ loop over all non-operator characters in a row
      if (mode & OPERAND) {
%{{{ if after operand
	ifnot ((('0' <= call[c2]) && (call[c2] <= '9')) || (('a' <= call[c2]) &&
	       (call[c2] <= 'z')) || (('A' <= call[c2]) && (call[c2] <= 'Z')) ||
	       (call[c2] == '_')) { % if not symbol expression
	  ifnot ((mode & FUNCTION_XY) && ((call[c2] == '(') || (call[c2] == ')'))) { % this seems not really usefull as c2 never is an operator here change to test for expression in xyfun
	    throw UsageError, sprintf("Unsupported character in expression: '%c'", call[c2]);
	  }
	}
%}}}
      } else if (mode & NUMBER) {
%{{{ if in number expression
	if (not(mode & NUMBER_EXP) && ((call[c2] == 'e') || (call[c2] == 'E'))) { % if exponential number expression
	  mode |= NUMBER_DEC | NUMBER_EXP;
	} else if (not(mode & NUMBER_EXP) && call[c2] == '.') { % if decimal number expression
	  mode |= NUMBER_DEC;
	} else if (('0' <= call[c2]) && (call[c2] <= '9')) { % if a number
	  
	} else { % if not a number
	  throw UsageError, sprintf("Not a number: found '%c'", call[c2]);
	}
%}}}
      } else if (mode == 0) {
%{{{ if in no mode
	if (('0'<=call[c2] && call[c2]<='9')) { % if a number
	  mode = NUMBER;
	} else if (call[c2] == '.') { % if a decimal number
	  mode = NUMBER | NUMBER_DEC;
	} else { % else is a function
	  mode = OPERAND;
	}
      }
%}}}
      c2++;
    }
%}}}
    if (callen <= c2) {
%{{{ if reached end of expression
      if (mode & OPERAND) {
%{{{ if in operand
	if (__f_access(sprintf("%s_xyfit", call[[c1:c2-1]]))) {
	  list_append(tokens, sprintf("%s(1)", call[[c1:c2-1]]));
	} else if (mode & FUNCTION_XY) {
	  if ((call[c2-1] == ')') && (mode & FUNCTION_OPEN)) {
	    list_append(tokens, call[[c1:c2]]);
	  } else if (not (mode & FUNCTION_OPEN)) {
	    list_append(tokens, sprintf("%s(1)", call[[c1:c2-1]]));
	  } else {
	    throw UsageError, sprintf("Unrecognized character in xy-function: '%c'", call[c2-1]);
	  }
	} else {
	  list_append(tokens, call[[c1:c2-1]]);
	}
%}}}
      } else if (mode & NUMBER) {
%{{{ if in number
	if (mode & NUMBER_DEC) {
	  list_append(tokens, atof(call[[c1:c2-1]]));
	} else {
	  list_append(tokens, atoi(call[[c1:c2-1]]));
	}
%}}}
      } else {
%{{{ else append part as string
	list_append(tokens, call[[c1:c2-1]]);
      }
%}}}
      return tokens;
%}}}
    } else if (mode & OPERAND) {
%{{{ if in operand
      if (__f_access(sprintf("%s_xyfit", call[[c1:c2-1]]))) {
%{{{ if xyfun
	mode |= FUNCTION_XY;
	if ((call[c2] == '(')) {
	  mode |= FUNCTION_OPEN;
	  bracematch++;
	} else if ( any(ops == call[c2]) ) {
	  list_append(tokens, sprintf("%s(1)", call[[c1:c2]]));
	  c1 = c2+1;
	  mode = 0;
	} else {
	  throw UsageError, sprintf("Unrecognized character after xy-function: '%c'", call[c2]);
	}
%}}}
      } else if (mode & FUNCTION_XY) {
%{{{ if in xyfun
	if ((call[c2] == '(') && (mode & FUNCTION_OPEN)) {
	  bracematch++;
	} else if ((call[c2] == ')') && (mode & FUNCTION_OPEN)) {
	  bracematch--;
	  ifnot (bracematch) {
	    list_append(tokens, call[[c1:c2]]);
	    c1 = c2+1;
	    mode = 0;
	  }
	} else {
	  %throw UsageError, sprintf("Unrecognized character in xy-function: '%c'", call[c2]);
	}
%}}}
      } else if (not(mode & FUNCTION_XY) && __f_access(call[[c1:c2-1]])) {
%{{{ if in slang fun
	list_append(tokens, __get_reference(call[[c1:c2-1]]));
	list_append(tokens, call[c2]);
	c1 = c2+1;
	mode = 0;
%}}}
      } else {
%{{{ else append magic variables
	list_append(tokens, call[[c1:c2-1]]);
	list_append(tokens, call[c2]);
	c1 = c2+1;
	mode = 0;
%}}}
      }
%}}}
    } else if (mode & NUMBER) {
%{{{ if in number
      if ((mode & NUMBER_EXP) && ((call[c2] == '+') || (call[c2] == '-'))) {
	
      } else if (mode & NUMBER_DEC) {
	list_append(tokens, atof(call[[c1:c2-1]]));
	list_append(tokens, call[c2]);
	c1 = c2+1;
	mode = 0;
      } else {
	list_append(tokens, atoi(call[[c1:c2-1]]));
	list_append(tokens, call[c2]);
	c1 = c2+1;
	mode = 0;
      }
%}}}
    } else if (mode == 0) {
%{{{ if none append operator (?)
      list_append(tokens, call[c2]);
      c1 = c2+1;
    }
%}}}
    c2++;
  }
%}}}
  if (callen <= c2) {
%{{{ if at end of expression
% (seems useless to do it again, but meanwhile eoe might be reached) change if
%  possible
    if (mode & OPERAND) { % if in operand
      if (mode & FUNCTION_XY) {
	if ((call[c2-1] == ')') && (mode & FUNCTION_OPEN)) {
	  list_append(tokens, call[[c1:c2-1]]);
	} else if (not (mode & FUNCTION_OPEN)) {
	  list_append(tokens, sprintf("%s(1)", call[[c1:c2-1]]));
	} else {
	  throw UsageError, sprintf("Unrecognized character in xy-function: '%c'", call[c2-1]);
	}
      }
    } else if (mode & NUMBER) { % if in number
      if (mode & NUMBER_DEC) {
	list_append(tokens, atof(call[[c1:c2-1]]));
      } else {
	list_append(tokens, atoi(call[[c1:c2-1]]));
      }
    } else if (c1 == c2) { % if empty string (what to return here?)
      
    } else { % else append token
      list_append(tokens, call[[c1:c2-1]]);
    }
  }
%}}}
  return tokens;
}
%}}}

private define setup_xyfun (flist) { %{{{
  _PAR = Assoc_Type[Int_Type];
  variable i, newcall = "";
  variable ffcall = "";
  variable npar = -1;
  _for i (0, length(flist)-1, 1) {
    if (typeof(flist[i]) == String_Type) {
      if (string_match(flist[i], "[a-zA-Z_0-9]+\\(([-+*/()a-zA-Z_0-9]+)\\)")) {
	variable pos, len;
	(pos, len) = string_match_nth(1);
	ffcall += define_xyfun(flist[i][[:pos-1]], flist[i][[pos:pos+len-1]]);
	newcall += sprintf("__xy(\"%s\", \"%s\")[1]", flist[i][[:pos-1]], flist[i][[pos+1:pos+len-2]]);
      } else if (flist[i][0] == '#') { % currently only recognized as x-axis
	newcall += "model.x";
      } else {
	if (assoc_key_exists(_PAR, flist[i])) {
	  newcall += sprintf("par[%d]", _PAR[flist[i]]);
	} else {
	  npar++;
	  newcall += sprintf("par[%d]", npar);
	  _PAR[flist[i]] = npar;
	}
      }      
    } else if (typeof(flist[i]) == Ref_Type) {
      newcall += sprintf("(@(%s))", string(flist[i]));
    } else if (typeof(flist[i]) == UChar_Type) {
      newcall += sprintf("%c", flist[i]);
    } else {
      if ((typeof(flist[i]) == Double_Type) || (typeof(flist[i]) == Float_Type)) {
	newcall += sprintf("%f", flist[i]);
      } else if (typeof(flist[i]) == Int_Type) {
	newcall += sprintf("%d", flist[i]);
      }
    }
  }
  ffcall += define_generic(newcall);
  fit_fun(ffcall);
} %}}}

private define param_string (p, len) { %{{{
  return sprintf (Fmt_Params, p.index, len, p.name,
		  p.tie, p.freeze, p.value, p.min, p.max, p.units, p.fun);
}%}}}

%%%%%%%%%%%%%%%%%%%%
public define get_xyfit_fun()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_xyfit_function}
%\synopsis{returns xy-fit-function}
%#c%{{{
%\usage{String_Type get_xyfit_fun();}
%\seealso{xyfit_fun, list_xypar}
%!%-
{
  if ((_FCT == "") || (_FCT == "null") || (_FCT == NULL))
    _FCT = NULL;
  return _FCT;
}
%}}}

private define list_xypar_string () { %{{{
  variable s = get_xyfit_fun();
  s = (s == NULL) ? String_Type[0] : [strtrim(s)];
  variable ps = String_Type[0];  
  variable pars = get_params();
  if (pars != NULL) {
    variable len = message_param_structs (pars);
    variable colheads;
    colheads = sprintf (Fmt_Colheads, len, "  param");
    ps = array_map (String_Type, &param_string, pars, len);
    s = strjoin ([s, colheads, ps], "\n");
  }
  if (length(s) == 0)
    return NULL;
  else if (length(s) == 1 && typeof(s) == Array_Type)
    s = s[0];
  return s;
}%}}}

%%%%%%%%%%%%%%%%%
define list_xypar()
%%%%%%%%%%%%%%%%%
%!%+
%\function{list_xypar}
%\synopsis{list current xy-parameter and xy-function}
%#c%{{{
%\usage{list_xypar();}
%\seealso{get_xyfit_fun, save_xypar, xyfit_fun}
%!%-
{
  variable dest = NULL;
  if (_NARGS) dest = ();
  _isis->put_string (dest, _function_name, list_xypar_string());
}
%}}}

%%%%%%%%%%%%%%%%%
define xyfit_fun()
%%%%%%%%%%%%%%%%%
%!%+
%\function{xyfit_fun}
%\synopsis{define xy-function to fit data defined via \code{define_xydata}}
%#c%{{{
%\usage{xyfit_fun(String_Type function_expression);}
%\description
%    Setting up an xy-fit-function with \code{fitfun} with \code{xyfit_fun("fitfun");}
%    is done by interpreting the string "fitfun" and creates a fit function that can
%    be understood by the isis routines. The string is searched for known functions
%    and special symbols. Every thing else is interpreted as a new fit parameter belonging
%    to a generic function \code{F}.
%
%    The known functions can either be a registered isis function with name \code{fun_xyfit}
%    or an slang intrinsic (see examples). If a symbol starts with '#' it is replaced
%    with the x-axis of the current data set.
%
%    If the xy-fit-function describes the graph a function,
%    it usually computes \code{@yref} in terms of \code{@xref} and the \code{par}-array.
%    This is the only option for xy-data without \code{xerr}, see \code{define_xydata}.
%
%    If the xy-fit-function describes a parameterized curve,
%    it usually computes \code{(@xref, @yref)} in terms of \code{par}
%    and probably further (constant) parameters passed to \code{fitfun_xyfit}
%    as qualifiers, which have been defined by \code{set_xyfit_qualifier}
%
%    If a function of the form \code{fitfun_xyfit} is loaded and a function
%    \code{fitfun_xyfit_default} exists, this will be passed to
%    \code{set_param_default_hook}.
%
%    Importand Notes:
%    The routine will first search for functions with the appendix _xyfit
%    and afterwards for intrinsic functions. This means if a function \code{fun} exists
%    and there is also a function \code{fun_xyfit} registered the routine will use the
%    latter.
%    
%    Calling a parameterized function together with functions of the form f(x) = y or
%    two parameterized functions redults in undefined behavior. Calling a parameterized
%    alone behaves as expected.
%
%    The internal structure of the routine requires it that a call to \code{list_par},
%    \code{save_par} or \code{load_par} do not work as expected. Equivalent _xypar functions
%    exist.
%\examples
%    xyfit_fun("linear"); % or
%    xyfit_fun("linear(1)"); % sets up xyfit-function linear_xyfit
%                             % in this form it is the xy-equivalent to fit_fun
%
%    % example for defining a proper _xyfit function:
%    % (example for y = x*a+b; as available already with: xyfit_fun("linear");)
%    define another_linear_regression_xyfit ()
%    {
%        variable xref, yref, par;
%        switch(_NARGS)
%        { case 0: return ["a [unit_a]", "b [unit_b]"];}
%        { case 3: (xref, yref, par) = (); }
%
%        @yref = par[0] * @xref + par[1];
%    }
%    define another_linear_regression_xyfit_default(i)
%    {
%    switch(i)
%        { case 0: return (1, 0, -10, 10); }
%        { case 1: return (0, 0,  -5, 10); }
%    }
%    xyfit_fun("another_linear_regression");
%    list_xypar;
%
%    % example for specifying normalization parameters
%    % (see 'norm_indexes' parameter of 'add_slang_function')
%    define fun_with_norm_xyfit ()
%    {
%        variable xref, yref, par;
%        switch(_NARGS)
%        { case 0: return struct { pars = ["a [unit_a]", "b [unit_b]"], norm = [0] }; }
%        { case 3: (xref, yref, par) = (); }
%
%        @yref = par[0] * (@xref ^ par[1]);
%    }
%    xyfit_fun("fun_with_norm");
%    
%    % example for a simple function call
%    xyfit_fun("Norm*sin(#x)/exp(#x-xoff)"); % results in fit-function of the form
%                                             % Norm*sin(x)/exp(x-xoff)
%                                             % with two parameters: Norm, xoff
%
%    % example of a combination of the previous cases
%    xyfit_fun("linear(1)+tan(#x^2-xoff)/linear(2)");
%    list_xypar; % output:
%                %   linear(1)+tan(#x^2-xoff)/linear(2)
%                %   idx  param    tie-to  freeze         value         min         max
%                %   1  linear(1).a   0     0                1     -100000      100000  coefficient of x
%                %   2  linear(1).b   0     0                0     -100000      100000  additive constant
%                %   3  linear(2).a   0     0                1     -100000      100000  coefficient of x
%                %   4  linear(2).b   0     0                0     -100000      100000  additive constant
%                %   5  F(1).xoff     0     0                0           0           0
%\seealso{define_xydata, set_xyfit_qualifier, list_xypar, save_xypar, load_xypar, plot_xyfit}
%!%-
{
  variable f;
  switch (_NARGS)
  { case 1: f = () ; }
  { return help(_function_name()); }
  if ((f == NULL) || (f == "")) {
    _FCT = NULL;
    fit_fun("");
    return;
  }
  variable tok = tokenize(f);
  setup_xyfun(tok);
  _FCT = f;
}
%}}}

%%%%%%%%%%%%%%%%%
define save_xypar()
%%%%%%%%%%%%%%%%%
%!%+
%\function{save_xypar}
%\synopsis{save current xy-fit-parameter in file}
%#c%{{{
%\usage{save_xypar(String_Type file);}
%\description
%    Save current xy-parameter and xy-function in file.
%\seealso{load_xypar, get_xyfit_fun, list_xypar}
%!%-
{
  if (_NARGS == 0) {
    usage ("%s (filename)", _function_name);
  }
  variable s, file = ();
  variable fp = fopen (file, "w");
  if (fp == NULL)
    throw IsisError, "Failed opening $file for writing"$;
  _isis->put_string (fp, _function_name, list_xypar_string());
  if (2 == is_defined ("isis_save_par_hook")) {
    eval ("isis_save_par_hook (\"$file\")"$);
    s = ();
    () = fputs (s, fp);
  }
  if (-1 == fclose (fp)) {
    variable err = "Failed closing $file"$;
    if (errno) err += sprintf (" (%s)", errno_string(errno));
    throw IsisError, err;
  }
}
%}}}

%%%%%%%%%%%%%%%%%
define load_xypar()
%%%%%%%%%%%%%%%%%
%!%+
%\function{load_xypar}
%\synopsis{load xy-parameter and xy-function from parameter file}
%#c%{{{
%\usage{load_xypar(String_Type file);}
%\description
%    Load parameter file saved with \code{save_xypar("file.par")}. To display
%    the load parameters one should use \code{list_xypar} instead of
%    \code{list_par}.
%\seealso{save_xypar, load_par}
%!%-
{
  _isis->error_if_fit_in_progress (_function_name);
  variable msg = "load_xypar (\"filename\")";
  if (_isis->chk_num_args (_NARGS, 1, msg))
    return;
  variable fname = ();
  variable fp = fopen(fname, "r");
  if (fp == NULL)
    throw IOError, "Failed opening file '${fname}'"$;
  variable xypar = fgetslines(fp);
  ()=fclose(fp);
  variable line = 0;
  while (xypar[line][0] == '#') line++;
  if (xypar[line][-1] != '\n')
    throw InternalError, "String expression of function to long.";
  variable fcall = xypar[line][[:-2]];
  xyfit_fun(fcall);
  xypar[line] = get_fit_fun+"\n";
  variable tmp = popen("mktemp", "r");
  if (tmp == NULL)
    tmp = popen("mktemp -p /home/"+getenv("USER"), "r");
  if (tmp == NULL)
    throw InternalError, "Could not process parameter file";
  variable tmpname;
  () = fgets(&tmpname, tmp); tmpname = tmpname[[:-2]];
  if (pclose(tmp) != 0)
    throw IOError, "Failed to load file '${fname}'"$;
  variable tfp = fopen(tmpname, "w");
  ()=fputslines(xypar, tfp);
  () = fclose(tfp);
  variable status,e;
  try (e) {
    status = _isis->_load_par (tmpname);
  } catch AnyError: {}
  () = remove(tmpname);
}
%}}}

%%%%%%%%%%%%%%%%%%%%
define define_xydata()
%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{define_xydata}
%\synopsis{defines an xy-dataset to be modeled with xyfit_fun}
%#c%{{{
%   \usage{Integer_Type data_id = define_xydata ( x[], y[] [, yerr[]] ); }
%\altusage{Integer_Type data_id = define_xydata ( x[], xerr[], y[], yerr[] ); }
%\qualifiers{
%\qualifier{N[=1000]}{: number of curve points, when xerr is considered}
%\qualifier{x_mdl [default: array covering \code{x} with \code{N} steps]}{:
%            (initial) x-values of the model, when \code{xerr} is considered}
%\qualifier{y_mdl [default: array covering \code{y} with \code{N} steps]}{:
%            (initial) y-values of the model, when \code{xerr} is considered}
%}
%\description
%    This function creates a dummy spectral dataset for ISIS and stores
%    the xy-dataset in its metadata, which are only considered when
%    fitting with a dedicated xy-fit-function defined by \code{xyfit_fun}.
%    Fitting via \code{fit_counts} will minimize the sum of squared residuals,
%    see \code{xyfit_residuals}.
%    
%    If no \code{xerr} is specified, the xy-fit-function will only be evaluated
%    at the x-values of the data. (yerr defaults to 1, if not specified.)
%    
%    If \code{xerr} is given as well, a curve is constructed in order to compute
%    the residuals of the data points. The xy-fit-function may produce a
%    (reasonably smooth!) graph (x, y(x)) or parameterized curve (x(t), y(t)).
%    It operates on the model points and possibly (constant) parameters
%    defined by \code{set_xy_qualifier}. Model points for an xy-dataset
%    are initially defined by the \code{N} or \code{x_mdl} and \code{y_mdl} qualifiers.
%
%    If xerr, yerr are lists with two array entries, than the first is
%    interpreted as lower and second as upper uncertainty
%
%\example
%    N = 150; (x, y) = ellipse(8, 5, 0.2*PI, grand(N)*1.5);
%    x += grand(N)+4; y += grand(N)+1;
%    id = define_xydata (x, ones(N), y, ones(N));
%    xyfit_fun("ellipse_xy");
%    set_xyfit_qualifier(id; curve_parameter=[0:2*PI:#3000]);
%    ()=fit_counts;
%    set_par("ellipse_xy(1).pos_angle", 30); ()=fit_counts; % help with angles...
%    plot_xyfit(id);
%    % fit_interactive(&plot_xyfit);
%\seealso{xyfit_fun, set_xyfit_qualifier, xyfit_residuals, plot_xyfit}
%!%-
{
  variable d = struct { x, xerr, y, yerr, n, xsyserr = 0, ysyserr = 0 };

  switch (_NARGS)
  { case 2: (d.x,         d.y        ) = (); }
  { case 3: (d.x,         d.y, d.yerr) = (); }
  { case 4: (d.x, d.xerr, d.y, d.yerr) = (); }
  { return help(_function_name()); }

  if(d.yerr == NULL) { d.yerr = 0 * d.y + 1; }
  if(typeof(d.xerr) == List_Type) { % put upper and lower uncertainties in 2d array
    d.xerr = _reshape([abs(d.xerr[0]), d.xerr[1]], [2, length(d.xerr[0])]);
  }
  if(typeof(d.yerr) == List_Type) {
    d.yerr = _reshape([abs(d.yerr[0]), d.yerr[1]], [2, length(d.yerr[0])]);
  }
  
  if(length(d.x) == length(d.y) && (length(d.yerr) == length(d.y) || length(d.yerr) == 2*length(d.y))
     && (d.xerr == NULL || (length(d.xerr) == length(d.x) || length(d.xerr) == 2*length(d.x)))) {
    d.n = length(d.x);
  } else {
    vmessage("error (%s): input data have inconsistent array lengths", _function_name);
    return;
  }

  variable data_id = define_counts([1:d.n], dup+1, dup*0, dup+1);  % dummy dataset for ISIS

  variable N = qualifier("N", 1000); % some default number of steps for modelgrid
  variable model = struct {
    x = d.xerr == NULL ? d.x : qualifier("x_mdl", [min(d.x-5*(length(d.xerr) == d.n ? d.xerr : d.xerr[0,*])) : max(d.x+5*(length(d.xerr) == d.n ? d.xerr : d.xerr[1,*])) : #N]),
    y = d.xerr == NULL ? d.y : qualifier("y_mdl", [min(d.y-5*(length(d.yerr) == d.n ? d.yerr : d.yerr[0,*])) : max(d.y+5*(length(d.yerr) == d.n ? d.yerr : d.yerr[1,*])) : #N]),
  };
  set_dataset_metadata (data_id, struct { xy_data=d, xy_model=model, xyfit_qualifier });

  return data_id;
}
%}}}

%%%%%%%%%%%%%%%%
define ignore_xy ()
%%%%%%%%%%%%%%%%
%!%+
%\function{ignore_xy}
%\synopsis{ignore points from xy-dataset}
%c#%{{{
%\usage{ignore_xy (index [, low, high]);}
%\description
%    Wraper function of the ignore function for datasets defined with
%    define_xydata. Ignores data points of dataset \code{index}
%    (in the range low to high) for fitting.
%\seealso{notice_xy, ignore, ignore_en, define_xydata}
%!%-
{
  variable id, lo=NULL, hi;
  switch (_NARGS)
  { case 1: id = ();               }
  { case 3: (id, lo, hi) = ();     }
  { return help(_function_name()); }

  variable i,n,nlist, md;
  _for i (0, length(id)-1, 1) {
    if (_isnull(lo)) {
      ignore(id[i]);
    } else {
      md = get_dataset_metadata(id[i]);
      if (typeof(md) != Struct_Type) {
	vmessage("No metadata associated with dataset %d, have you defined it via 'define_xydata'?", id[i]);
	return;
      }
      variable xn = get_data_info(id[i]).notice;
      n = (md.xy_data.x>=lo and md.xy_data.x<=hi);
      _isis->_set_notice_using_mask(int((xn-n)>0), id[i]);
    }
  }
}
%}}}

%%%%%%%%%%%%%%%%
define notice_xy ()
%%%%%%%%%%%%%%%%
%!%+
%\function{notice_xy}
%\synopsis{notice points from xy-dataset}
%#c%{{{
%\usage{notice_xy (index [, low, high]);}
%\description
%    Wraper function of the notice function for datasets defined with
%    define_xydata. Include data points of dataset \code{index}
%    (in the range low to high) for fitting.
%\seealso{ignore_xy, notice, notice_en, define_xydata}
%!%-
{
  variable id, lo=NULL, hi;
  switch (_NARGS)
  { case 1: id = ();               }
  { case 3: (id, lo, hi) = ();     }
  { return help(_function_name()); }

  variable i,n,nlist, md;
  _for i (0, length(id)-1, 1) {
    if (_isnull(lo)) {
      notice(id[i]);
    } else {
      md = get_dataset_metadata(id[i]);
      if (typeof(md) != Struct_Type) {
	vmessage("No metadata associated with dataset %d, have you defined it via 'define_xydata'?", id[i]);
	return;
      }
      variable xn = get_data_info(id[i]).notice;
      n = (md.xy_data.x>=lo and md.xy_data.x<=hi);
      _isis->_set_notice_using_mask(int((n+xn)>0), id[i]);
    }
  }
}
%}}}

private define get_xyfit_metadata(data_id) %{{{
{
  variable md = get_dataset_metadata(data_id);
  ifnot(_is_struct_type(md))
    throw UsageError, "No metadata associated with dataset #$data_id. (Was it defined via define_xydata?)"$;
  ifnot(   struct_field_exists(md, "xy_data")
	&& struct_field_exists(md, "xy_model")
	&& struct_field_exists(md, "xyfit_qualifier"))
    throw UsageError, "Metadata assosicated with dataset #$data_id were not defined via define_xydata."$;
  return md;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%
define set_xyfit_qualifier(data_id)
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{set_xyfit_qualifier}
%\synopsis{modify the meta data of an xy-dataset defined by \code{define_xydata} and used for an xy-fit}
%#c%{{{
%\usage{set_xyfit_qualifier(data_id; qualifiers); }
%\description
%    This function can be used to modify the information for the xy-data.
%    The used qualifiers are combined with the data structure. Normally
%    the qualifiers to be set should be \code{x_mdl} or \code{curve_parameter}.
%\example
%    set_xyfit_qualifier(id; curve_parameter=[0:2*PI:#3000]);
%\seealso{define_xydata, xyfit_fun, plot_xyfit}
%!%-
{
  % TODO: The xyfit_qualifiers should be specific to the xy-fit-function
  %       and not to the xy-dataset.
  variable md = get_xyfit_metadata(data_id);
  md.xyfit_qualifier = struct_combine(md.xyfit_qualifier, __qualifiers);
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%%%%%
define set_xyfit_sys_err_frac()
%%%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{set_xyfit_sys_err_frac}
%\synopsis{adds systematic uncertainties to an xy-dataset defined by \code{define_xydata}}
%#c%{{{
%\usage{set_xyfit_sys_err_frac(data_id, [xsyserr,] ysyserr);}
%\description
%    A systematic uncertainty is added in quadrature to either
%    the x- and y-data or to the latter only. The combined
%    uncertainty considered by \code{xyfit_residuals} then is
%
%      err_new = sqrt( err^2 + (data * syserr)^2 )
%
%    where \code{data} is the x- or y-data as defined using
%    \code{define_xydata} and \code{err} is the corresponding
%    defined uncertainty.
%
%    By default, no systematics uncertainties are considered.
%\example
%    % adds 0.5% systematics to the y-data only
%    set_xyfit_sys_err_frac(1, .005);
%    
%    % adds systematics to both, x- (0.5%) and y-data (1%)
%    set_xyfit_sys_err_frac(1, .005, 0.01);
%\seealso{xyfit_residuals, define_xydata, set_sys_err_frac}
%!%-
{
  variable data_id, xsyserr = 0, ysyserr;
  switch (_NARGS)
    { case 2: (data_id, ysyserr) = (); }
    { case 3: (data_id, xsyserr, ysyserr) = (); }
    { help(_function_name); }
  
  variable md = get_xyfit_metadata(data_id);
  md.xy_data.xsyserr = xsyserr;
  md.xy_data.ysyserr = ysyserr;
}
%}}}

%%%%%%%%%%%%%%%%%%%%%%
define xyfit_residuals(data_id)
%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{xyfit_residuals}
%\synopsis{calculates the difference between xy-data and xy-model}
%#c%{{{
%\usage{Double_Type res[] = xyfit_residuals(Integer_Type data_id);}
%\description
%    The residuals \code{res[i]} are determined differently for xy-data with
%    uncertainties in y only or in both dimensions.
%    
%    If the xy-data have no x-uncertainties \code{xerr}:\n
%       \code{res[i] = (data.y[i] - model.y[i]) / data.yerr[i];}
%       
%    If the xy-data have uncertainties \code{xerr} as well:\n
%       \code{res[i] = min( relative_distance_from_curve[i] );}\n
%    where \code{relative_distance_from_curve[i]} is composed of 
%       \code{(model.x - data.x[i]) / data.xerr[i]} and
%       \code{(model.y - data.y[i]) / data.yerr[i]}.
%\seealso{define_xydata, xyfit_fun}
%!%-
{
  variable md = get_xyfit_metadata(data_id);
  variable data = md.xy_data;
  variable model = md.xy_model;

  % filter notice list
  variable noti = get_data_info(data_id).notice;

  % add systematics to y-data (if set)
  variable yerr = @(data.yerr);
  if (typeof(data.ysyserr) == Array_Type or data.ysyserr > 0) {
    yerr = length(yerr) == data.n % symmetric error bars?
         ? sqrt(yerr*yerr + data.ysyserr*data.ysyserr*data.y*data.y)
	 : sqrt(yerr*yerr + outer_product([1,1], data.ysyserr*data.ysyserr*data.y*data.y));
  }

  % no x-errors given
  variable residuals;
  if (data.xerr == NULL) {
    residuals = data.y - model.y;
    residuals /= length(yerr) == data.n % symmetric error bars?
      ? yerr : yerr[[0:data.n-1] + data.n*(sign(-residuals)+1)/2];
    return residuals[where(noti)];
  }
  
  % else: fitting data points with uncertainties in both dimensions

  % add systematics to x-data (if set)
  variable xerr = @(data.xerr);
  if (typeof(data.xsyserr) != Double_Type or data.xsyserr > 0) {
    xerr = length(xerr) == data.n % symmetric error bars?
         ? sqrt(xerr*xerr + data.xsyserr*data.xsyserr*data.x*data.x)
	 : sqrt(xerr*xerr + outer_product([1,1], data.xsyserr*data.xsyserr*data.x*data.x));
  }

  variable i, resx, resy;
  residuals = Double_Type[data.n];
  _for i (0, data.n-1, 1)
  {
    %MB: currently we assume a 2D-Gaussian profile of uncertainties, but use an ellipse with
    %    the semi-axes xerr and yerr
    %    take care of effective_error=0? fix problems with strong excentricities?
    %    better solution? the calculations above might be computationally expensive for
    %    fine models with many points

    % variable distance_to_curve = hypot ( model.x - data.x[i], model.y - data.y[i] );
    % variable angle_to_curve    = atan2 ( model.y - data.y[i], model.x - data.x[i] );
    % variable effective_error   = hypot( data.xerr*cos(angle_to_curve), data.yerr*sin(angle_to_curve) );
    % residuals[i] = min ( distance_to_curve / effective_error );
    % MB: the residuals above are only valid in certain cases, better use the Mh version
    resx = model.x - data.x[i];
    resx /= length(xerr) == data.n ? xerr[i] : xerr[(sign(resx)+1)/2,i];
    resy = model.y - data.y[i];
    resy /= length(yerr) == data.n ? yerr[i] : yerr[(sign(resy)+1)/2,i];
    residuals[i] = min( hypot( resx, resy ) );
  }
  return residuals[where(noti)];
}
%}}}

%%%%%%%%%%%%%%%%%
define get_xydata(data_id)
%%%%%%%%%%%%%%%%%
%!%+
%\function{get_xydata}
%\synopsis{provides the xy-data, which has been defined with \code{define_xydata}}
%#c%{{{
%   \usage{(x[],         y[], yerr[]) = get_xydata(Integer_Type data_id);}
%\altusage{(x[], xerr[], y[], yerr[]) = get_xydata(Integer_Type data_id);}
%\description
%    This function returns the xy-data of dataset # \code{data_id}
%    previously defined with \code{define_xydata}.
%    If no x-uncertainty was defined, no \code{xerr} is returned.
%\seealso{define_xydata, get_xymodel}
%!%-
{
  variable data = get_xyfit_metadata(data_id).xy_data;
  variable n = qualifier_exists("noticed") ? where(get_data_info(data_id).notice) : [0:data.n-1];
  return (data.xerr == NULL)
    ? (@data.x[n], @data.y[n], length(data.yerr) == data.n ? @data.yerr[n] : {data.yerr[0,n], data.yerr[1,n]})
    : (@data.x[n], length(data.xerr) == data.n ? @data.xerr[n] : {data.xerr[0,n], data.xerr[1,n]}, @data.y[n], length(data.yerr) == data.n ? @data.yerr[n] : {data.yerr[0,n], data.yerr[1,n]});
}
%}}}

%%%%%%%%%%%%%%%%%%
define get_xymodel(data_id)
%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_xymodel}
%\synopsis{provides the xy-model \code{(x_mdl[], y_mdl[])} given by an xy-fit}
%#c%{{{
%\usage{(x_mdl[], y_mdl[]) = get_xymodel(Integer_Type data_id);}
%\description
%    This function returns the xy-model (\code{(x_mdl[],y_mdl[])})
%    provided by the last evaluation (by \code{fit_counts} / \code{eval_counts})
%    of the xy-fit-function specified with \code{xyfit_fun}.
%\seealso{xyfit_fun, get_xydata, define_xydata}
%!%-
{
  variable model = get_xyfit_metadata(data_id).xy_model;
  return (@model.x, @model.y);
}
%}}}

%%%%%%%%%%%%%%%%%
define eval_xyfun ()
%%%%%%%%%%%%%%%%%
%!%+
%\function{eval_xyfun}
%\synopsis{evaluate current xy-function for points x [y]}
%#c%{{{
%\usage{(Double_Type[] x, Double_Type[] y) = eval_xyfun(Double_Type[] x [, Double_Type[] y]);}
%\qualifiers{
%\qualifier{id[=Isis_Active_Dataset]:}{ uses parameter and model for dataset #id}
%}
%\description
%Evaluate the current xy-model on the points (x,y)
%\seealso{define_xydata, xyfit_fun}
%!%-
{
  variable x,y=NULL;
  switch (_NARGS)
  { case 1: x = ();              }
  { case 2: (x,y) = ();          }
  { help(_function_name); return;}

  variable id = qualifier("id", Isis_Active_Dataset);

  if (y!=NULL) {
    if (length(x) != length(y)) {
      message("x- and y-arrays must have the same size");
      return;
    }
  } else {
    y = @x;
  }

  variable md = get_dataset_metadata(id);
  if (typeof(md) != Struct_Type) {
    vmessage("No metadata associated with dataset %d, have you defined it via 'define_xydata'?", id);
    return;
  }

  variable x_save, y_save;
  (x_save, y_save) = get_xymodel(id);
  md.xy_model.x = @x;
  md.xy_model.y = @y;
  set_dataset_metadata(id, md);
  variable grid = get_data_counts(id);
  variable reset_active = Isis_Active_Dataset;
  Isis_Active_Dataset = id;
  ()=eval_fun(grid.bin_lo, grid.bin_hi);
  Isis_Active_Dataset = reset_active;

  variable xr, yr;
  (xr,yr) = get_xymodel(id);
  md.xy_model.x = x_save;
  md.xy_model.y = y_save;
  set_dataset_metadata(id, md);
  return xr, yr;
}%}}}

%%%%%%%%%%%%%%%%%%
define eval_xyfun2 ()
%%%%%%%%%%%%%%%%%%
%!%+
%\function{eval_xyfun2}
%\synopsis{evaluate valid xyfit_fun string with parameters}
%#c%{{{
%\usage{(Double_Type[] x, Double_Type[] y) = eval_xyfun2 (handle, x[ , y, par])}
%\qualifiers{
%\qualifier{qualifier:}{ pass qualifier structure to function}
%}
%\description
%Evaluates function \code{handle} (either the name of a xy-function or reference)
%on the points x (and y) with parameters \code{par}.
%\seealso{eval_xyfun, define_xydata, xyfit_fun}
%!%-
{
  variable expr,x,y,par;
  variable q = qualifier("qualifier");
  switch (_NARGS)
  { case 3: (expr,x,par) = (); y=@Int_Type[length(x)]*0.; }
  { case 4: (expr,x,y,par) = ();                          }
  { help(_function_name); return;                         }
  if (length(x) != length(y)) {
    message("x- and y-array must have the same length");
    return;
  }

  variable f = (typeof(expr)==Ref_Type) ? expr : __get_reference(expr+"_xyfit");
  variable xr = @x, yr = @y;
  if (_isnull(f)) {
    vmessage("function %s not found", expr);
    return;
  }
  @f(&xr, &yr, par;; q);
  return (xr, yr);
}%}}}

private define plot_xyresiduals(data_id) %{{{
{
  if (get_xyfit_fun == NULL) {
    message("No model defined");
    return;
  }
  variable dcol = qualifier("dcol", 4);
  variable res = qualifier("res", 1); % residual mode
  variable rcol = qualifier("rcol", dcol);
  variable recol = qualifier("recol", rcol);
  variable dsym = qualifier("dsym", 4);
  variable rsym = qualifier("rsym", dsym);
  variable ovrplt = qualifier("ovrplt");
  variable xlbl = qualifier("xlabel", "x");

  variable popt = get_plot_options;

  variable cdcol, crcol, crecol, crsym;
  variable rsign, xyres,noti,my,md,data;

  xlabel(xlbl);
  if (res == 1) {
    ylabel("\\gx");
  } else if (res == 2) {
    ylabel("\\gx\\u2");
  } else {
    ylabel("Ratio");
  }

  variable i;
  % yplot range
  variable yminr, ymaxr;
  _for i (0, length(data_id)-1, 1) {
    md = get_dataset_metadata(data_id[i]);
    data = md.xy_data;
    (,my) = eval_xyfun(data.x, data.y; id=data_id[i]);
    noti = get_data_info(data_id[i]).notice_list;
    rsign = sign(data.y[noti] - my[noti]);
    xyres = xyfit_residuals(data_id[i]);
    xyres = sign(xyres)*xyres;
    if (i) {
      if (res == 1) {
	yminr = min([yminr, rsign*xyres - ones(length(noti)) ] );
	ymaxr = max([ymaxr, rsign*xyres + ones(length(noti)) ] );
      } else if (res == 2) {
	yminr = min([rsign*sqr(xyres),yminr]);
	ymaxr = max([rsign*sqr(xyres),ymaxr]);
      } else {
	yminr = min( [ yminr, (data.y[noti]-((data.n==length(data.yerr))?data.yerr[noti] : data.yerr[0,*][noti]))/my[noti] ] );
	ymaxr = max( [ ymaxr, (data.y[noti]+((data.n==length(data.yerr))?data.yerr[noti] : data.yerr[1,*][noti]))/my[noti] ] );
      }
    } else {
      if (res == 1) {
	yminr = min(rsign*xyres - ones(length(noti)) );
	ymaxr = max(rsign*xyres + ones(length(noti)) );
      } else if (res == 2) {
	yminr = min(rsign*sqr(xyres));
	ymaxr = max(rsign*sqr(xyres));
      } else {
	yminr = min( (data.y[noti]-((data.n==length(data.yerr))?data.yerr[noti] : data.yerr[0,*][noti]))/my[noti] );
	ymaxr = max( (data.y[noti]+((data.n==length(data.yerr))?data.yerr[noti] : data.yerr[1,*][noti]))/my[noti] );
      }
    }
  }
  yrange(yminr-(ymaxr-yminr)*.1, ymaxr+(ymaxr-yminr)*.1);
  
  _for i (0, length(data_id)-1, 1) {
    variable id = data_id[i];
    cdcol = length(dcol)>i ? dcol[i] : 4;
    crcol = length(rcol)>i ? rcol[i] : cdcol;
    crecol = length(recol)>i ? recol[i] : crcol;
    crsym = length(rsym)>i ? rsym[i] : 4;
    
    md = get_dataset_metadata(id);
    variable model = md.xy_model;
    data = md.xy_data;
    noti = get_data_info(id).notice_list;
      
    (,my) = eval_xyfun(data.x, data.y; id=id);
    rsign = sign(data.y[noti] - my[noti]);
    
    variable qstruct = struct{error_color=crecol};
    if (ovrplt || i) {
      qstruct = struct {@qstruct, overplot=NULL};
    }
    
    % plot residuals with errors
    xyres = xyfit_residuals(id);
    xyres = sign(xyres)*xyres;
    
    color(crcol);
    pointstyle(rsym);
    if (res == 1) {
      if (data.xerr == NULL) {
	plot_with_err (data.x[noti], rsign*xyres, ones(length(noti));; qstruct);
      } else {
	plot_with_err (data.x[noti], data.xerr[noti], rsign*xyres, ones(length(noti));; qstruct);
      }
    } else if (res == 2) {
      connect_points(0);
      if (ovrplt || i) {
	oplot (data.x[noti], rsign*sqr(xyres));
      } else {
	plot (data.x[noti], rsign*sqr(xyres));
      }
    } else {
      if (data.xerr == NULL) {
	plot_with_err (data.x[noti], data.y[noti]/my[noti],
		       data.n == length(data.yerr) ? data.yerr[noti]/my[noti] :
		     {data.yerr[0,*][noti]/my[noti], data.yerr[1,*][noti]/my[noti]};; qstruct);
      } else {
	plot_with_err (data.x[noti], data.n == length(data.xerr) ? data.xerr[noti] :
		     {data.xerr[0,*][noti], data.xerr[1,*][noti]},
		       data.y[noti]/my[noti], data.n == length(data.yerr) ? data.yerr[noti]/my[noti] :
		     {data.yerr[0,*][noti]/my[noti], data.yerr[1,*][noti]/my[noti]};; qstruct);
      }
    }
  }
  
  % plot ideal line
  variable cpo = get_plot_options;
  color(-1);
  connect_points(-1);
  linestyle(2);
  if (res == 3) {
    oplot([cpo.xmin, cpo.xmax], [1, 1]);
  } else {
    oplot([cpo.xmin, cpo.xmax], [0, 0]);
  }

  popt.color = popt.color+1;
  set_plot_options(popt); % reset options
}
%}}}

%%%%%%%%%%%%%%%%%
define plot_xyfit(data_id)
%%%%%%%%%%%%%%%%%
%!%+
%\function{plot_xyfit}
%\synopsis{plot xy-data and its current xy-model}
%#c%{{{
%\usage{plot_xyfit(Integer_Type data_id);}
%\description
%    A simple plot function of the xy-data (given by \code{define_xydata})
%    and -model (last evaluation of the xy-fit-function, see \code{xyfit_fun}).
%
%    This functions basically calls:\n
%        \code{plot_with_err( get_xydata(data_id) );}\n
%        \code{oplot( get_xymodel(data_id) );}
%\seealso{get_xydata, get_xymodel, define_xydata, xyfit_fun, plot_with_err}
%!%-
{
  variable dcol=qualifier("dcol", 4);
  variable decol=qualifier("decol", dcol);
  variable mcol=qualifier("mcol", 2);
  variable rcol=qualifier("rcol", dcol);
  variable recol=qualifier("recol", dcol);
  variable dsym=qualifier("dsym", 4);
  variable rsym=qualifier("rsym", dsym);
  variable res=qualifier("res", 0);
  variable xlbl=qualifier("xlabel", "x"), ylbl=qualifier("ylabel", "y");

  variable cdcol,cdecol,cmcol,cdsym;
  variable hasmodel = (get_xyfit_fun!=NULL)?1:0;

  variable popt = get_plot_options;

  if (res && hasmodel) {
    multiplot([3,1]);
  }
  ylabel(ylbl); xlabel(xlbl);
  variable i, id, md, noticed, xasym, yasym;
  % set plot range if not set already
  % this requires the full data to be in one array
  variable all_x = Double_Type[0];
  variable all_y = Double_Type[0];
  variable all_xloerr = Double_Type[0];
  variable all_xhierr = Double_Type[0];
  variable all_yloerr = Double_Type[0];
  variable all_yhierr = Double_Type[0];
  _for i (0, length(data_id)-1, 1) {
    md = get_dataset_metadata(data_id[i]);
    noticed = get_data_info(data_id[i]).notice_list;
    xasym = length(md.xy_data.xerr) != md.xy_data.n ? _isnull(md.xy_data.xerr) ? -1 : 1 : 0;
    yasym = length(md.xy_data.yerr) != md.xy_data.n ? _isnull(md.xy_data.yerr) ? -1 : 1 : 0;
    all_x = [all_x, md.xy_data.x[noticed]];
    all_y = [all_y, md.xy_data.y[noticed]];
    if (xasym < 0) {
      all_xloerr = [all_xloerr, Double_Type[md.xy_data.n][noticed]];
      all_xhierr = [all_xhierr, Double_Type[md.xy_data.n][noticed]];
    } else if (xasym > 0) {
      all_xloerr = [all_xloerr, md.xy_data.xerr[0,*][noticed]];
      all_xhierr = [all_xhierr, md.xy_data.xerr[1,*][noticed]];
    } else {
      all_xloerr = [all_xloerr, -md.xy_data.xerr[noticed]];
      all_xhierr = [all_xhierr, md.xy_data.xerr[noticed]];
    }
    if (yasym < 0) {
      all_yloerr = [all_yloerr, Double_Type[md.xy_data.n][noticed]];
      all_yhierr = [all_yhierr, Double_Type[md.xy_data.n][noticed]];
    } else if (yasym > 0) {
      all_yloerr = [all_yloerr, md.xy_data.yerr[0,*][noticed]];
      all_yhierr = [all_yhierr, md.xy_data.yerr[1,*][noticed]];
    } else {
      all_yloerr = [all_yloerr, -md.xy_data.yerr[noticed]];
      all_yhierr = [all_yhierr, md.xy_data.yerr[noticed]];
    }
  }
  % Mikes private function
  variable cp = get_plot_options.connect_points;
  oplt = 0; % is undefined when none of Mikes plotting functions is called
  () = start_plot(all_x+all_xloerr,
		  all_x+all_xhierr,
		  all_y+all_yloerr,
		  all_y+all_yhierr);
  set_plot_options(popt);

  _for i (0, length(data_id)-1, 1) {
    cdcol = length(dcol)>i ? dcol[i] : 4;
    cdecol = length(decol)>i ? decol[i] : cdcol;
    cmcol = length(mcol)>i ? mcol[i] : 2;
    cdsym = length(dsym)>i ? dsym[i] : 4;

    id = data_id[i];
    color(cdcol);
    pointstyle(cdsym);
    %if (i)
      oplot_with_err( get_xydata(id; noticed); error_color=cdecol );
    %else
    %  plot_with_err( get_xydata(id; noticed); error_color=cdecol );
    connect_points(-1);
    color(cmcol);
    if (hasmodel) {
      oplot( get_xymodel(id) );
    }
    connect_points(cp);
  }
  if (res && hasmodel) {
    plot_xyresiduals(data_id;; __qualifiers);
  }
  set_plot_options(popt);
}
%}}}

%
% Code from src/plot/xylabel_in_box.sl
%
define xylabel_in_box()
%!%+
%\function{xylabel_in_box}
%\synopsis{places a text label at a relative position in the plot box}
%\usage{xylabel_in_box(Double_Type x_rel, y_rel, String_Type label[, angle[, justify]]);}
%\description
%    \code{xylabel_in_box} uses relative coordinates \code{x_rel} and \code{y_rel}:
%    the plot box corresponds to \code{0<=x_rel<=1} and \code{0<=y_rel<=1},
%    but the \code{label} can also be placed outside of this area.
%    In order to calculate the world coordinates (with \code{coords_in_box}),
%    the \code{x}- and \code{yrange} has to be set before the first plot command.
%
%    The world coordinates and the optional parameters are passed to
%    \code{xylabel}, so see \code{xylabel} for a description of \code{angle} and \code{justify}.
%\seealso{coords_in_box, xylabel}
%!%-
{
  variable relX, relY, s, angle=0, justify=0;
  switch(_NARGS)
  { case 3: (relX, relY, s                ) = (); }
  { case 4: (relX, relY, s, angle         ) = (); }
  { case 5: (relX, relY, s, angle, justify) = (); }
  { help(_function_name()); return; }

  xylabel(coords_in_box(relX, relY), s, angle, justify);
}

%
% Code from src/plot/y1axis.sl
%

define y1axis()
%!%+
%\function{y1axis}
%\synopsis{(re)draws a first y-axis}
%\usage{y1axis(Double_Type min_value, Double_Type max_value[, Double_Type step]);}
%\qualifiers{
%\qualifier{nsub}{[=0] number of minor tick marks within each major divison (nsub=0 => no minor ticks)}
%\qualifier{maji}{[=0.5] length of major tick marks, drawn inwards (in units of the character height)}
%\qualifier{majo}{[=0] length of major tick marks, drawn outwards (in units of the character height)}
%\qualifier{fmin}{[=0.5] length of minor tick marks (as fraction of the major tick marks)}
%\qualifier{disp}{[=0.5] displacement of baseline of tick labels to the axis (in units of the character height)}
%\qualifier{angle}{[=0] orientation of the text (in degrees)}
%\qualifier{color}{[=1] color of the axis}
%\qualifier{linewidth}{[=1] line width of the axis}
%\qualifier{opt}{[="N"/"LN"] pgaxis-options: "L" (log), "N" (numbers), "1" (force decimal), "2" (force EE)}
%\qualifier{xrel}{[=0] relative position of the y1axis (0=left, 1=right)}
%\qualifier{ymin}{[=ymin] absolute y-position of the axis}
%\qualifier{ymax}{[=ymax] absolute y-position of the axis}
%}
%\seealso{x2axis (and references therein)}
%!%-
{
  variable po = get_plot_options();
  variable min_value, max_value, step=0;
  switch(_NARGS)
  { case 2: (min_value, max_value) = (); }
  { case 3: (min_value, max_value, step) = (); }
  { help(_function_name()); return; }

  variable opt = "N";
  variable ymin = qualifier("ymin", po.ymin);
  variable ymax = qualifier("ymax", po.ymax);
  if(po.logy)
  { ymin = log10(ymin);
    ymax = log10(ymax);
    opt = "LN";
    min_value = log10(min_value);
    max_value = log10(max_value);
  }
  variable x = coordX_in_box( qualifier("xrel", 0) );
  if(po.logx)  x = log10(x);
  _pgsci( qualifier("color", 1) );
  _pgslw( qualifier("linewidth", 1) );
  _pgaxis(qualifier("opt", opt), x, ymin, x, ymax,
          min_value, max_value,  step, qualifier("nsub", 0),
          qualifier("majo", 0), qualifier("maji", 0.5), qualifier("fmin", 0.5), -qualifier("disp", 0.5),
          qualifier("angle", 0)
         );
}

%
% Code from src/plot/y2axis.sl
%

define y2axis()
%!%+
%\function{y2axis}
%\synopsis{draws a second y-axis}
%\usage{y2axis(Double_Type min_value, Double_Type max_value[, Double_Type step]);}
%\qualifiers{
%\qualifier{nsub}{[=0] number of minor tick marks within each major divison (nsub=0 => no minor ticks)}
%\qualifier{maji}{[=0.5] length of major tick marks, drawn inwards (in units of the character height)}
%\qualifier{majo}{[=0] length of major tick marks, drawn outwards (in units of the character height)}
%\qualifier{fmin}{[=0.5] length of minor tick marks (as fraction of the major tick marks)}
%\qualifier{disp}{[=0.5] displacement of baseline of tick labels to the axis (in units of the character height)}
%\qualifier{angle}{[=0] orientation of the text (in degrees)}
%\qualifier{color}{[=1] color of the axis}
%\qualifier{linewidth}{[=1] line width of the axis}
%\qualifier{opt}{[="N"/"LN"] pgaxis-options: "L" (log), "N" (numbers), "1" (force decimal), "2" (force EE)}
%\qualifier{xrel}{[=1] relative position of the y2axis (0=left, 1=right)}
%\qualifier{ymin}{[=ymin] absolute y-position of the axis}
%\qualifier{ymax}{[=ymax] absolute y-position of the axis}
%}
%\seealso{x2axis (and references therein)}
%!%-
{
  variable po = get_plot_options();
  variable min_value, max_value, step=0;
  switch(_NARGS)
  { case 2: (min_value, max_value) = (); }
  { case 3: (min_value, max_value, step) = (); }
  { help(_function_name()); return; }

  variable opt = "N";
  variable ymin = qualifier("ymin", po.ymin);
  variable ymax = qualifier("ymax", po.ymax);
  if(po.logy)
  { ymin = log10(ymin);
    ymax = log10(ymax);
    opt = "LN";
    min_value = log10(min_value);
    max_value = log10(max_value);
  }
  variable x = coordX_in_box( qualifier("xrel", 1) );
  if(po.logx)  x = log10(x);
  _pgsci( qualifier("color", 1) );
  _pgslw( qualifier("linewidth", 1) );
  _pgaxis(qualifier("opt", opt), x, ymin, x, ymax,
          min_value, max_value,  step, qualifier("nsub", 0),
          qualifier("maji", 0.5), qualifier("majo", 0), qualifier("fmin", 0.5), qualifier("disp", 0.5),
          qualifier("angle", 0)
         );
}

%
% Code from src/plot/y2label.sl
%
define y2label()
%!%+
%\function{y2label}
%\synopsis{labels the second y-axis}
%\usage{y2label(String_Type s);}
%\qualifiers{
%\qualifier{color}{[=1] textcolor}
%\qualifier{f}{[=0.03] fraction of the plot-area's width that the label is to the right}
%}
%\seealso{xylabel}
%!%-
{
  variable s;
  switch(_NARGS)
  { case 1: s = (); }
  { help(_function_name()); return; }

  variable plot_opt = get_plot_options();
  color( qualifier("color", 1) );
  xylabel(coords_in_box(1+qualifier("f", 0.03), 0.5), s, 90, 0.5);
}

%
% Code from src/slang/date/yearOfMJD.sl
%
%%%%%%%%%%%%%%%%
define yearOfMJD()
%%%%%%%%%%%%%%%%
%!%+
%\function{yearOfMJD}
%\synopsis{transforms a date given as MJD into the year (with fractional parts)}
%\usage{Double_Type year = yearOfMJD(Double_Type MJD);}
%\seealso{jd2year}
%!%-
{
    variable MJD;
    switch(_NARGS)
    { case 1: MJD = (); }
    { help(_function_name()); return; }

    return jd2year(MJD;mjd);
}

%
% Code from src/data/timing/z2fold.sl
%
%
%%%%%%%%%%%%%%%%%%%%%%%%
define z2fold ()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{z2fold}
%\synopsis{peforms Z^2_m search on a lightcurve or event data in a given period
%interval}
%\usage{(Struct_Type res) = z2fold(Double_Type t, r, pstart, pstop);
%or (Struct_Type res) = z2fold(Double_Type t, pstart, pstop) ; (event data)}
%
%\qualifiers{
%\qualifier{sampling} {how many periods per peak to use (default=10)}
%\qualifier{nsrch} {how many periods to search in a linear grid (default not set)}
%\qualifier{dp} {delta period of linear period grid (default  not set)}
%\qualifier{m} {number of harmonics used (default =2)}
%\qualifier{chatty} {set the verbosity, (default=0)}
%\qualifier{gti}{GTIs for event data, given as struct{start=Double_Type, stop=Double_Type}} 
%}
%
%\description
%   Performs Z^2_m test  on a given lightcurve or event data between the periods
%   pstart and pstop. The function is based on epfold.sl.
%   GTI correction is implemented only  for event-data.
%
%   By default, periods are sampled according to the triangular rule
%   for estimating the period error, using "sampling" periods per peak.
%   If a linear grid is to be used, either "dp" for a given distance
%   between two consecutive periods or "nsrch" for a given number of
%   periods can be given. These qualifiers are mutually exclusive. 
%
%   The returned structure "res" contains four fields: "p" for the
%   evaluated period and "stat" for the value of the statistic used.
%
%   Please see Buccheri et al., 1983, Astron. Astrophys. 128, 245 for
%   more information on the Z-square statistics. 
%   
%\seealso{zsquare, epfold, pfold}   
%!%-
{
   variable t,r,pstart, pstop ;

  variable eventdata = 0;
  % disable chatty if not set
   variable chatty = qualifier("chatty",0) ;
  
    switch(_NARGS)
    {case 3: (t,  pstart, pstop) = ();
      eventdata = 1 ;
      if (chatty) {print("using event data"); } 
    }
    {case 4: (t,r,pstart, pstop) = (); }
    {help(_function_name()); return; }
  
   variable t0 = qualifier ("t0", t[0]);
   variable nbins = qualifier ("nbins", 20) ;
   variable sampl = qualifier("sampling", 10)  ; %periods per peak
   variable dp = qualifier("dp", -10) ; %delta p for linear period sampling
   variable nsrch = qualifier("nsrch", -10) ; %number of periods in linear period grid

   variable m = qualifier("m", 2) ; % using Z^2_2 as standard method
   if (chatty) {vmessage("Using Z^2_%d statistic\n", m) ; } ;
   
   nsrch = int(nsrch) ;

  % ifnot(qualifier_exists("dt") or eventdata==1){
  %   message("WARNING (epfold): no dt qualifier given, creating differential grid! Might lead to unexpected results!") ;
  % }
   variable dt = qualifier("dt" , make_hi_grid(t)-t) ; %exposure times

  % ifnot (qualifier_exists("gti") or eventdata==0){
  %   message("WARNING (epfold): using eventdata without gti qualifier, taking time of first and last event. This is most likely NOT what you want!") ;
  % }
  % variable gti = qualifier("gti", struct{start=min(t), stop=max(t)}) ;

   variable tges = sum(dt) ; %effective observation time
  
   % if (tges < 0) {message("ERROR (epfold): Please provide lightcurve with continously increasing time (tges was < 0)."); return;}
  
   if (chatty) {message(">> Starting data analysis <<") ;}
       
   if (chatty) {
      ()=printf("Determining number of trial periods between %f and %f.\n", pstart, pstop) ;
   }

   variable ergdim=1 ;
   variable p = pstart ;
   variable i = 1 ;
   
   if (dp > 0 && nsrch == -10)
     {	
	if (chatty) {()=printf("Making linear period grid with delta P = %f.\n", dp) ;}
	p = [pstart:pstop:dp] ;
	ergdim = length(p) ;
     }
   else if (dp == -10  && nsrch > 0)
     {
	if (chatty) {()=printf("Making linear period grid with %d entries.\n", nsrch) ;}
	p = [pstart:pstop:#nsrch] ;
	ergdim = nsrch ;
     }
   else if (dp > 0 && nsrch > 0)
     {
	()=printf("Please use either dp or nsrch!\n") ;
	return NULL ;
     }
   else { 
      if (chatty) ()=printf("Making intelligent grid!\n") ; 
      while (p[-1] <= pstop) 
	{
	   ergdim++ ;
	   p=[p, p[i-1]+p[i-1]*p[i-1]/(tges*sampl) ];
	   i++ ;
   	}
   }
 
   if (chatty) {()=printf("number of periods = %d\n....done\n",ergdim);}
   
   if (chatty) {variable tstart = _time(); variable tlast = _time() ; }
   
   variable ptest = pstart ;
   variable z2 = Double_Type[ergdim] ;
   
   variable chierg = Double_Type[ergdim] ;
   
   variable titer, remain ;

   % %susbstract mean from rate, as we need only the
   % %difference for the statistics and it does not change
   % %the generality
  % ifnot(eventdata){
  % variable rm = moment(r) ;
   % r -= rm.ave ;
      
   % if (chatty)  {
   %    ()=printf("mean = %f \n", rm.ave) ;
   %    % ()=printf("effective observation time: %f\n", tges  );
   % }
  % }
  
   _for i (0, length(p)-1, 1)
     {
	ptest = p[i] ;
        if(eventdata) {z2[i] = zsquare(t,ptest ; m=m) ;}
	else { z2[i] = zsquare(t,ptest  ; lc=r, m=m ) ;}

     if(chatty) 
	  {
	     if (_time()-tlast > 10)
	       {
		  tlast=_time() ;
		  titer = (tlast-tstart)/double(i) ;
		  remain=(ergdim-i)*titer ;
		  
		  ()=printf("%.2f%% done\n", i*100./ergdim ) ;
		  ()=printf("its taking %.2fsec per iteration\n", titer) ;
		  
		  if (remain <= 600)  ()=printf("%.2f sec remaining...\n", remain) ;
		  else if (remain <= 7200) ()=printf("%.2f min remaining...\n", remain/60.) ;
		  else ()=printf("%.2f h remaining...\n", remain/3600.) ;
	       }
	  }
     }

  %structure containing searched periods, statistics value,
  %phasebins used and indices of periods with gaps in the pulse profile
   return struct{p=p, stat=z2 } ;
   
}

%
% Code from src/slang/astronomy/zams.sl
%
% helper function for a fast polynomial
private define tout_poly(a,b,c,d,e,logz) {
    return (((e*logz+d)*logz+c)*logz+b)*logz+a;
}

define zams () 
%
%!%+
%\function{zams}
%\synopsis{return properties of the ZAMS per Tout et al., 1996, MNRAS 281, 257}
%\usage{(radius,luminosity,temperature)=zams(mass;z=metallicity);}
%\qualifiers{
%    \qualifier{z}{metallicity (between 0.0001 and 0.03; default: 0.02 (solar))}
%}
%\description
%  This function uses the rational function fits of Tout et al. (1996, MNRAS 281, 257)
%  and calculates the radius, luminosity, and temperature for zero age main sequence stars
%  of a given mass (in solar units;  can be an array, must be between 0.1 and 100 Msun).
%  The radius and luminosity are returned in solar units, the temperature is in K.
%  Tout et al. state that in general the fits are better than 7.5per cent in luminosity and
%  5 per cent in radius, for solar metallicity they are good to 3 per cent in L and
%  1.2 per cent in R.
%!%-
%  
{
    variable mass;
    switch(_NARGS)
    { case 1: mass=(); }
    { help(_function_name()); return;}

    % metallicity
    variable zz=qualifier("z",0.02); 
    
    % check parameter validity
    if (zz<0.0001 || zz>0.03) {
	vmessage("Metallicity must be between 0.0001 and 0.03");
    }
    variable mmin=min(mass);
    variable mmax=max(mass);

    if (mmin<0.1 || mmin>100. || mmax<0.1 || mmax>100.) {
	vmessage("Mass must be between 0.1 and 100 solar masses");
    }
    
    % variable xx=0.76-3.*zz; % Hydrogen
    % variable yy=0.24+2.*zz; % Helium

    variable zsun=0.02; % solar metallicity
    variable logz=log10(zz/zsun);

    % Tout, Table 1 
    variable alpha  =tout_poly(0.39704170, -0.32913574, 0.34776688, 0.37470851,0.09011915,logz);
    variable beta   =tout_poly(8.52762600,-24.41225973,56.43597107,37.06152575,5.45624060,logz);
    variable gamma  =tout_poly(0.00025546, -0.00123461,-0.00023246, 0.00045519,0.00016176,logz);
    variable deltaT =tout_poly(5.43288900, -8.62157806,13.44202049,14.51584135,3.39793084,logz);
    variable epsilon=tout_poly(5.56357900,-10.32345224,19.44322980,18.97361347,4.16903097,logz);
    variable zeta   =tout_poly(0.78866060, -2.90870942, 6.54713531, 4.05606657,0.53287322,logz);
    variable eta    =tout_poly(0.00586685, -0.01704237, 0.03872348, 0.02570041,0.00383376,logz);

    % Tout, Table 2
    variable theta  =tout_poly( 1.71535900,  0.62246212, -0.92557761, -1.16996966,-0.30631491,logz);
    variable iota   =tout_poly( 6.59778800, -0.42450044,-12.13339427,-10.73509484,-2.51487077,logz);
    variable kappa  =tout_poly(10.08855000, -7.11727086,-31.67119479,-24.24848322,-5.33608972,logz);
    variable lambda =tout_poly( 1.01249500,  0.32699690, -0.00923418, -0.03876858,-0.00412750,logz);
    variable mu     =tout_poly( 0.07490166,  0.02410413,  0.07233664,  0.03040467, 0.00197741,logz);
    variable nu     =tout_poly( 0.01077422,  0.00000000,  0.00000000,  0.00000000, 0.00000000,logz);
    variable xi     =tout_poly( 3.08223400,  0.94472050, -2.15200882, -2.49219496,-0.63848738,logz);
    variable omikron=tout_poly(17.84778000, -7.45345690,-48.96066856,-40.05386135,-9.09331816,logz);
    variable pi     =tout_poly( 0.00022582, -0.00186899,  0.00388783,  0.00142402,-0.00007671,logz);


    % Tout, Eq. (1)
    % should probably implement parts here using a horner scheme
    variable lum=(alpha*mass^5.5 + beta*mass^11.)/(gamma+mass^3+deltaT*mass^5+epsilon*mass^7+zeta*mass^8+eta*mass^9.5);

    % Tout, Eq. (2)
    % dito
    variable radius=(theta*mass^2.5+iota*mass^6.5+kappa*mass^11.+lambda*mass^19.+mu*mass^19.5)/(nu+xi*mass^2.0+omikron*mass^8.5+mass^18.5+pi*mass^19.5);
    
    % L = 4pi r^2 sigma t^4
    variable lsun=3.828e26; % in Watt(urgh)
    variable sigma_sb=5.670367e-8; % w/m^2 k^4 (urgh)
    variable rsun=695500e3; % m
    variable temperature=( lsun*lum/(4*PI*(rsun*radius)^2.*sigma_sb) )^(1./4.);
    
    return(radius,lum,temperature);
}

%
% Code from src/slang/astronomy/zang.sl
%
%%%%%%%%%%%%%%
define zang ()
%%%%%%%%%%%%%%
%!%+
%\function{zang}
%\synopsis{Determine the angular size of an object as a function of redshift}
%\usage{Double_Type angsize = zang (Double_Type size, Double_Type redshift);}
%\description
%    Requires an input size in kpc and returns an angular size
%    in arc seconds. (Analog to the function \code{zsize},
%    which converts angular size to projected size size given the
%    redshift.)
%    The default cosmology according is used. The function passes
%    qualifiers to the functions \code{lumdist} and \code{cosmo_param} allowing
%    to change the cosmology.
%\example
%    variable size = 50.; % kpc
%    variable    z = 1.5; % redshift
%    zang(50,1.5; omega_m = 0.3, omega_lambda = 0.0, silent);
%    % ---> 6.58 arc seconds
%\seealso{lumdist, cosmo_param}
%!%-
{
  variable dl, z;

  switch(_NARGS)
  { case 2: (dl,z) = ();}
  { help(_function_name()); return; }
  
  variable d =  lumdist(z;; __qualifiers);

  % Angular distance is equal to the luminosity distance times (1+z)^2
  return 180./PI*3600.*dl*(1.+z)^2/(1000.*d);
}

%
% Code from src/fitting/fit-functions/zpc_tbnew_simple.sl
%

define zpc_tbnew_simple_fit(lo,hi,par)
%!%+
%\function{zpc_tbnew_simple (fit-function)}
%\synopsis{partial covering absorption}
%\description
%    This function describes two partial coverers with
%    covering fraction f, according to the formula:
%    \code{ (1-f) +  f*exp(-nH*sigma) }
%    This definition is equal to how, e.g., zpcfabs is defined.
%\seealso{tbnew_simple,tbnew,zpcfabs}
%!%-
{
   variable v1 = tbnew_simple_z_fit(lo,hi,[par[0],par[2]]);
   variable c=par[1];

   return  ((1-c) + c*v1);
}
add_slang_function("zpc_tbnew_simple", ["nH [10^22/cm^2]","C [Covering Fraction]","z [redshift]"]);

private define zpc_tbnew_simple_defaults(i)
{
   switch(i)
   {case 0:  return (1, 0, 0, 100); }
   {case 1:  return (0.5, 0, 0, 1); }   
   {case 2:  return (0, 1, 0, 10); }   
}
set_param_default_hook("zpc_tbnew_simple", &zpc_tbnew_simple_defaults);


%
% Code from src/slang/astronomy/zsize.sl
%
define zsize ()
%!%+
%\function{zsize}
%\synopsis{Determine the projected size of an object as a function of redshift}
%\usage{Double_Type proj_size = zsize (Double_Type angle, Double_Type redshift);}
%\description
%    Requires an input angular size in arc seconds and returns
%    the projected size in kpc. (Analog to the function \code{zang},
%    which converts projected size to angular size given the
%    redshift.)
%    The default cosmology according is used. The function passes
%    qualifiers to the functions \code{lumdist} and \code{cosmo_param} allowing
%    to change the cosmology.
%\example
%    variable angsize = 1.; % arc seconds
%    variable    z = 1.5; % redshift
%    zsize(angsize,z);
%    % ---> 8.46 kpc
%\seealso{zang, lumdist, cosmo_param}
%!%-
{
  variable ang, z;

  switch(_NARGS)
  { case 2: (ang,z) = ();}
  { help(_function_name()); return; }
  
  variable d =  lumdist(z;; __qualifiers);

  % Angular distance is equal to the luminosity distance times (1+z)^2
  return ang*PI/(180.0*3600)*(1.0+z)^(-2) * 1000.0*d;
}

%
% Code from src/data/timing/zsquare.sl
%
%%%%%%%%%%%%%%%%%%%%%%%
define zsquare ()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{zsquare}
%\synopsis{calculate Z^2 statistics for pulse period search}
%\usage{Double_Type Z2 = zsquare(Double_Type times, Double_Type testperiod}
%
%\description
%    Calculates the Z^2 statistics of an event list or light-curve for
%    a given period. By itself not very useful, but is used in z2fold
%    to search for periodicities using the Z^2 statistics.
%
%\qualifiers{
%\qualifier{lc}{=Double_Type rate; allows the use of a lightcurve instead of event files}
%\qualifier{m}{= Integer_Type; determines the highest order of summations (default 2)}
%}
%!%-
{
   variable t, P ; 
  
   switch(_NARGS)
    {case 2: (t, P) = ();}
    {help(_function_name()); return; }

   variable phis = (t mod P)/P ;
   
   variable j, k ;
   variable cossq, sinsq ;

   variable lc = qualifier("lc", 1) ; %% if not provided assume event data
   variable m = qualifier("m", 2) ; %% standard test should be Z^2_2

   variable norm = qualifier_exists("lc") ? sum(lc) : length(phis) ;
   
   variable zsqtmp = Double_Type[m] ;

   _for j (1, m, 1)
     {
	cossq = (sum(lc* cos(j*phis*2*PI)))^2. ;
	sinsq = (sum(lc* sin(j*phis*2*PI)))^2. ;
	zsqtmp[j-1] = cossq + sinsq ;
     }

   return sum(zsqtmp) * 2./norm ;
}
%
% initialization routines for isisscripts to be executed
% AFTER all functions have been defined
%


